;;; phps-mode-parser.el --- Exported Emacs Parser Generator -*- lexical-binding: t -*-

;; Copyright (C) 2018-2026  Free Software Foundation, Inc.

;; This file is not part of GNU Emacs.


;;; Commentary:


;;; Code:


(require 'phps-mode-lexer)

(define-error
 'phps-parser-error
 "PHPs Parser Error")


;;; Variables:


(defvar
  phps-mode-parser--action-tables
  #s(hash-table test equal data (0 0 1 1 2 2 3 3 4 4 5 2 6 2 7 2 8 8 9 2 10 10 11 11 12 2 13 13 14 2 15 15 16 16 17 17 18 2 19 16 20 20 21 15 22 22 23 11 24 24 25 25 26 26 27 2 28 2 29 11 30 16 31 11 32 32 33 33 34 34 35 35 36 11 37 11 38 38 39 39 40 40 41 16 42 11 43 11 44 22 45 2 46 2 47 47 48 16 49 2 50 11 51 51 52 11 53 53 54 11 55 55 56 56 57 57 58 58 59 59 60 60 61 61 62 2 63 2 64 64 65 65 66 2 67 2 68 15 69 69 70 70 71 71 72 2 73 11 74 2 75 16 76 76 77 77 78 11 79 2 80 80 81 81 82 2 83 11 84 84 85 2 86 86 87 87 88 88 89 89 90 90 91 91 92 92 93 93 94 94 95 11 96 96 97 97 98 98 99 99 100 100 101 101 102 102 103 103 104 104 105 105 106 106 107 107 108 108 109 109 110 110 111 111 112 112 113 113 114 114 115 115 116 116 117 117 118 118 119 119 120 120 121 121 122 122 123 123 124 124 125 125 126 126 127 101 128 128 129 2 130 13 131 131 132 11 133 133 134 134 135 135 136 136 137 137 138 106 139 139 140 140 141 141 142 142 143 11 144 144 145 145 146 13 147 147 148 148 149 149 150 150 151 151 152 152 153 153 154 154 155 155 156 156 157 157 158 158 159 159 160 160 161 161 162 162 163 163 164 164 165 165 166 11 167 167 168 168 169 169 170 77 171 128 172 172 173 173 174 13 175 15 176 16 177 15 178 11 179 26 180 2 181 16 182 11 183 11 184 40 185 16 186 11 187 11 188 188 189 16 190 15 191 70 192 71 193 11 194 16 195 77 196 11 197 2 198 11 199 199 200 89 201 201 202 202 203 203 204 204 205 100 206 206 207 105 208 108 209 209 210 210 211 211 212 212 213 213 214 214 215 215 216 128 217 217 218 218 219 172 220 220 221 26 222 11 223 2 224 2 225 3 226 4 227 2 228 2 229 2 230 2 231 11 232 2 233 2 234 234 235 2 236 236 237 22 238 238 239 239 240 2 241 11 242 11 243 243 244 244 245 245 246 22 247 2 248 2 249 2 250 11 251 251 252 11 253 253 254 11 255 255 256 256 257 257 258 258 259 60 260 260 261 2 262 2 263 263 264 11 265 2 266 2 267 69 268 133 269 269 270 2 271 2 272 272 273 2 274 274 275 275 276 2 277 86 278 87 279 90 280 136 281 11 282 282 283 283 284 101 285 285 286 286 287 287 288 106 289 106 290 290 291 291 292 292 293 293 294 294 295 295 296 296 297 297 298 298 299 299 300 300 301 101 302 2 303 303 304 2 305 2 306 2 307 2 308 2 309 2 310 2 311 2 312 312 313 2 314 2 315 2 316 2 317 2 318 318 319 2 320 2 321 2 322 2 323 2 324 2 325 2 326 2 327 2 328 2 329 2 330 2 331 2 332 2 333 2 334 2 335 335 336 336 337 337 338 338 339 339 340 340 341 341 342 342 343 343 344 344 345 345 346 346 347 347 348 348 349 349 350 350 351 4 352 2 353 353 354 354 355 355 356 356 357 357 358 358 359 359 360 360 361 361 362 362 363 363 364 364 365 364 366 40 367 367 368 2 369 3 370 4 371 2 372 2 373 2 374 2 375 11 376 2 377 2 378 378 379 2 380 380 381 22 382 382 383 383 384 2 385 11 386 11 387 387 388 388 389 389 390 22 391 2 392 2 393 2 394 11 395 395 396 11 397 397 398 11 399 399 400 400 401 401 402 402 403 60 404 404 405 2 406 2 407 407 408 11 409 2 410 2 411 69 412 133 413 413 414 2 415 2 416 416 417 2 418 418 419 419 420 2 421 86 422 87 423 90 424 136 425 11 426 426 427 427 428 101 429 429 430 430 431 431 432 106 433 106 434 434 435 435 436 436 437 437 438 438 439 439 440 440 441 441 442 442 443 443 444 444 445 445 446 101 447 2 448 448 449 2 450 2 451 2 452 2 453 2 454 2 455 2 456 2 457 312 458 2 459 2 460 2 461 2 462 2 463 318 464 2 465 2 466 2 467 2 468 2 469 2 470 2 471 2 472 2 473 2 474 2 475 2 476 2 477 2 478 2 479 2 480 480 481 481 482 482 483 483 484 484 485 485 486 486 487 487 488 488 489 489 490 490 491 491 492 492 493 493 494 494 495 495 496 4 497 2 498 498 499 499 500 500 501 501 502 502 503 503 504 504 505 505 506 506 507 507 508 508 509 364 510 364 511 40 512 367 513 441 514 514 515 515 516 516 517 517 518 518 519 2 520 2 521 3 522 4 523 2 524 2 525 2 526 2 527 11 528 2 529 2 530 530 531 2 532 532 533 22 534 534 535 535 536 2 537 11 538 11 539 539 540 540 541 541 542 22 543 2 544 2 545 2 546 11 547 547 548 11 549 549 550 11 551 551 552 552 553 553 554 554 555 60 556 556 557 2 558 2 559 559 560 11 561 2 562 2 563 69 564 133 565 565 566 2 567 2 568 568 569 2 570 570 571 571 572 2 573 86 574 87 575 90 576 136 577 11 578 578 579 579 580 101 581 581 582 582 583 583 584 106 585 106 586 586 587 587 588 588 589 589 590 590 591 591 592 592 593 593 594 594 595 595 596 596 597 101 598 2 599 599 600 2 601 2 602 2 603 2 604 2 605 2 606 2 607 2 608 312 609 2 610 2 611 2 612 2 613 2 614 318 615 2 616 2 617 2 618 2 619 2 620 2 621 2 622 2 623 2 624 2 625 2 626 2 627 2 628 2 629 2 630 2 631 631 632 632 633 633 634 634 635 635 636 636 637 637 638 638 639 639 640 640 641 641 642 642 643 643 644 644 645 645 646 646 647 4 648 2 649 649 650 650 651 651 652 652 653 653 654 654 655 655 656 656 657 657 658 658 659 659 660 364 661 364 662 40 663 367 664 441 665 665 666 666 667 667 668 668 669 669 670 2 671 583 672 672 673 673 674 40 675 675 676 287 677 677 678 678 679 2 680 583 681 681 682 682 683 683 684 684 685 685 686 686 687 2 688 3 689 4 690 2 691 2 692 2 693 2 694 2 695 11 696 2 697 2 698 698 699 2 700 700 701 22 702 702 703 703 704 2 705 11 706 11 707 707 708 708 709 709 710 22 711 2 712 2 713 2 714 11 715 715 716 11 717 717 718 11 719 719 720 720 721 721 722 722 723 60 724 724 725 2 726 2 727 727 728 11 729 2 730 2 731 69 732 133 733 733 734 2 735 2 736 736 737 2 738 738 739 739 740 2 741 86 742 87 743 90 744 136 745 11 746 746 747 747 748 101 749 749 750 750 751 751 752 106 753 106 754 754 755 755 756 756 757 757 758 758 759 759 760 760 761 761 762 762 763 763 764 764 765 101 766 2 767 767 768 2 769 2 770 2 771 2 772 2 773 2 774 2 775 2 776 312 777 2 778 2 779 2 780 2 781 2 782 318 783 2 784 2 785 2 786 2 787 2 788 2 789 2 790 2 791 2 792 2 793 2 794 2 795 2 796 2 797 2 798 2 799 799 800 800 801 801 802 802 803 803 804 804 805 805 806 806 807 807 808 808 809 809 810 810 811 811 812 812 813 813 814 814 815 4 816 2 817 817 818 818 819 819 820 820 821 821 822 822 823 823 824 824 825 825 826 826 827 827 828 364 829 364 830 40 831 367 832 441 833 833 834 834 835 835 836 836 837 837 838 2 839 583 840 840 841 841 842 40 843 843 844 287 845 845 846 846 847 2 848 583 849 849 850 850 851 851 852 852 853 853 854 854 855 2 856 751 857 2 858 858 859 859 860 860 861 861 862 862 863 863 864 864 865 865 866 866 867 867 868 868 869 869 870 870 871 871 872 872 873 873 874 874 875 875 876 876 877 877 878 878 879 879 880 880 881 881 882 882 883 883 884 884 885 885 886 886 887 887 888 888 889 889 890 890 891 891 892 892 893 893 894 894 895 895 896 896 897 897 898 898 899 899 900 900 901 901 902 902 903 903 904 904 905 905 906 906 907 907 908 908 909 909 910 910 911 911 912 912 913 913 914 914 915 915 916 916 917 917 918 918 919 919 920 920 921 921 922 922 923 923 924 924 925 925 926 926 927 927 928 928 929 929 930 930 931 931 932 932 933 933 934 934 935 935 936 936 937 937 938 938 939 939 940 940 941 941 942 942 943 943 944 944 945 945 946 946 947 947 948 948 949 11 950 950 951 951 952 952 953 2 954 583 955 955 956 956 957 957 958 2 959 3 960 4 961 2 962 962 963 2 964 2 965 2 966 966 967 967 968 2 969 969 970 2 971 971 972 972 973 973 974 974 975 975 976 976 977 977 978 978 979 979 980 980 981 22 982 982 983 983 984 984 985 985 986 986 987 2 988 988 989 989 990 990 991 991 992 992 993 993 994 994 995 995 996 996 997 997 998 998 999 999 1000 1000 1001 1001 1002 1002 1003 1003 1004 1004 1005 1005 1006 1006 1007 1007 1008 1008 1009 1009 1010 1010 1011 1011 1012 1012 1013 1013 1014 1014 1015 22 1016 1016 1017 1017 1018 1018 1019 1019 1020 1020 1021 2 1022 1022 1023 1023 1024 1024 1025 1025 1026 1026 1027 1027 1028 1028 1029 1029 1030 1030 1031 1031 1032 1032 1033 1033 1034 1034 1035 1035 1036 1036 1037 2 1038 1038 1039 1039 1040 1040 1041 1041 1042 1042 1043 1043 1044 1044 1045 1045 1046 1046 1047 69 1048 1048 1049 1049 1050 2 1051 1051 1052 1052 1053 1053 1054 1054 1055 1055 1056 1056 1057 2 1058 1058 1059 1059 1060 1060 1061 1061 1062 1062 1063 2 1064 86 1065 87 1066 1066 1067 1067 1068 90 1069 136 1070 11 1071 1071 1072 1072 1073 101 1074 1074 1075 1075 1076 1076 1077 106 1078 106 1079 1079 1080 1080 1081 1081 1082 1082 1083 1083 1084 1084 1085 1085 1086 1086 1087 161 1088 1088 1089 1089 1090 1090 1091 1091 1092 1092 1093 1093 1094 101 1095 2 1096 11 1097 1097 1098 2 1099 1099 1100 11 1101 11 1102 1102 1103 1103 1104 1104 1105 2 1106 2 1107 11 1108 1108 1109 11 1110 11 1111 1111 1112 60 1113 1113 1114 2 1115 1115 1116 11 1117 2 1118 2 1119 133 1120 1120 1121 2 1122 1122 1123 1123 1124 1124 1125 2 1126 2 1127 2 1128 2 1129 2 1130 2 1131 2 1132 2 1133 312 1134 2 1135 2 1136 2 1137 2 1138 2 1139 318 1140 2 1141 2 1142 2 1143 2 1144 2 1145 2 1146 2 1147 2 1148 2 1149 2 1150 2 1151 2 1152 2 1153 2 1154 2 1155 2 1156 1156 1157 1157 1158 1158 1159 1159 1160 1160 1161 1161 1162 1162 1163 1163 1164 1164 1165 1165 1166 1166 1167 1167 1168 1168 1169 1169 1170 1170 1171 1171 1172 4 1173 2 1174 1174 1175 1175 1176 1176 1177 1177 1178 1178 1179 1179 1180 1180 1181 1181 1182 1182 1183 1183 1184 1184 1185 364 1186 364 1187 40 1188 367 1189 441 1190 1190 1191 1191 1192 1192 1193 1193 1194 1194 1195 2 1196 583 1197 1197 1198 1198 1199 40 1200 1200 1201 287 1202 1202 1203 1203 1204 2 1205 583 1206 1206 1207 1207 1208 1208 1209 1209 1210 1210 1211 1211 1212 2 1213 751 1214 2 1215 1215 1216 1216 1217 1217 1218 1218 1219 1219 1220 1220 1221 1221 1222 1222 1223 1223 1224 1224 1225 2 1226 3 1227 4 1228 2 1229 2 1230 2 1231 2 1232 11 1233 2 1234 2 1235 1235 1236 2 1237 1237 1238 22 1239 1239 1240 1240 1241 2 1242 11 1243 11 1244 1244 1245 1245 1246 1246 1247 22 1248 2 1249 2 1250 2 1251 11 1252 1252 1253 11 1254 1254 1255 11 1256 1256 1257 1257 1258 1258 1259 1259 1260 60 1261 1261 1262 2 1263 2 1264 1264 1265 11 1266 2 1267 2 1268 69 1269 133 1270 1270 1271 2 1272 2 1273 1273 1274 2 1275 1275 1276 1276 1277 2 1278 86 1279 87 1280 90 1281 136 1282 11 1283 1283 1284 1284 1285 101 1286 1286 1287 1287 1288 1288 1289 106 1290 106 1291 1291 1292 1292 1293 1293 1294 1294 1295 1295 1296 1296 1297 1297 1298 1298 1299 1299 1300 1300 1301 1301 1302 101 1303 2 1304 1304 1305 2 1306 2 1307 2 1308 2 1309 2 1310 2 1311 2 1312 2 1313 312 1314 2 1315 2 1316 2 1317 2 1318 2 1319 318 1320 2 1321 2 1322 2 1323 2 1324 2 1325 2 1326 2 1327 2 1328 2 1329 2 1330 2 1331 2 1332 2 1333 2 1334 2 1335 2 1336 1336 1337 1337 1338 1338 1339 1339 1340 1340 1341 1341 1342 1342 1343 1343 1344 1344 1345 1345 1346 1346 1347 1347 1348 1348 1349 1349 1350 1350 1351 1351 1352 4 1353 2 1354 1354 1355 1355 1356 1356 1357 1357 1358 1358 1359 1359 1360 1360 1361 1361 1362 1362 1363 1363 1364 1364 1365 364 1366 364 1367 40 1368 367 1369 441 1370 1370 1371 1371 1372 1372 1373 1373 1374 1374 1375 2 1376 583 1377 1377 1378 1378 1379 40 1380 1380 1381 287 1382 1382 1383 1383 1384 2 1385 583 1386 1386 1387 1387 1388 1388 1389 1389 1390 1390 1391 1391 1392 2 1393 751 1394 2 1395 1395 1396 1396 1397 1397 1398 1398 1399 1399 1400 1400 1401 1401 1402 1402 1403 1403 1404 1404 1405 868 1406 1406 1407 1407 1408 1408 1409 1409 1410 1410 1411 1411 1412 1412 1413 1413 1414 1414 1415 1415 1416 1416 1417 1417 1418 1418 1419 1419 1420 1420 1421 1421 1422 1422 1423 1423 1424 1424 1425 1425 1426 1426 1427 1427 1428 1428 1429 1429 1430 1430 1431 1431 1432 1432 1433 1433 1434 1434 1435 1435 1436 1436 1437 1437 1438 1438 1439 1439 1440 1440 1441 1441 1442 1442 1443 1443 1444 1444 1445 1445 1446 1446 1447 1447 1448 1448 1449 1449 1450 1450 1451 1451 1452 1452 1453 1453 1454 1454 1455 1455 1456 1456 1457 1457 1458 1458 1459 1459 1460 1460 1461 1461 1462 1462 1463 1463 1464 1464 1465 1465 1466 1466 1467 1467 1468 1468 1469 1469 1470 1470 1471 1471 1472 1472 1473 1473 1474 1474 1475 1475 1476 1476 1477 1477 1478 1478 1479 1479 1480 1480 1481 1481 1482 1482 1483 1483 1484 1484 1485 1485 1486 11 1487 1487 1488 1488 1489 1489 1490 2 1491 583 1492 1492 1493 956 1494 1494 1495 1495 1496 989 1497 161 1498 1498 1499 164 1500 1500 1501 2 1502 1502 1503 1503 1504 1504 1505 1505 1506 2 1507 2 1508 2 1509 1509 1510 2 1511 1511 1512 2 1513 2 1514 2 1515 2 1516 2 1517 2 1518 2 1519 2 1520 2 1521 1521 1522 1522 1523 1523 1524 1524 1525 1525 1526 1526 1527 1527 1528 1528 1529 1529 1530 1530 1531 1531 1532 1532 1533 1533 1534 1534 1535 1535 1536 22 1537 1537 1538 3 1539 4 1540 2 1541 11 1542 1542 1543 1543 1544 1544 1545 1545 1546 1546 1547 1547 1548 1548 1549 1549 1550 1550 1551 1551 1552 60 1553 1553 1554 1554 1555 11 1556 1556 1557 1557 1558 1558 1559 1559 1560 86 1561 90 1562 11 1563 1563 1564 1564 1565 101 1566 1566 1567 1567 1568 1568 1569 1569 1570 1570 1571 1571 1572 1572 1573 1573 1574 101 1575 868 1576 1576 1577 1577 1578 1578 1579 1579 1580 1580 1581 1581 1582 1582 1583 1583 1584 1584 1585 1585 1586 1586 1587 1587 1588 1588 1589 1589 1590 1590 1591 1591 1592 1592 1593 1593 1594 1594 1595 1595 1596 1596 1597 1597 1598 1598 1599 1599 1600 1600 1601 1601 1602 1602 1603 1603 1604 1604 1605 1605 1606 1606 1607 1607 1608 1608 1609 1609 1610 1610 1611 1611 1612 1612 1613 1613 1614 1614 1615 1615 1616 1616 1617 1617 1618 1618 1619 1619 1620 1620 1621 1621 1622 1622 1623 1623 1624 1624 1625 1625 1626 1626 1627 1627 1628 1628 1629 1629 1630 1630 1631 1631 1632 1632 1633 1633 1634 1634 1635 1635 1636 1636 1637 1637 1638 1638 1639 1639 1640 1640 1641 1641 1642 1642 1643 1643 1644 1644 1645 1645 1646 1646 1647 1647 1648 1648 1649 1649 1650 1650 1651 1651 1652 1652 1653 1653 1654 1654 1655 1655 1656 11 1657 1657 1658 1658 1659 1659 1660 2 1661 583 1662 1662 1663 956 1664 1664 1665 1665 1666 989 1667 161 1668 164 1669 1669 1670 1670 1671 1671 1672 868 1673 1673 1674 11 1675 1675 1676 2 1677 583 1678 1678 1679 1679 1680 1680 1681 364 1682 364 1683 367 1684 441 1685 1685 1686 1686 1687 1687 1688 1688 1689 2 1690 583 1691 1691 1692 1692 1693 1693 1694 1694 1695 2 1696 3 1697 4 1698 2 1699 2 1700 2 1701 2 1702 11 1703 2 1704 2 1705 1705 1706 2 1707 1707 1708 22 1709 1709 1710 1710 1711 2 1712 2 1713 11 1714 11 1715 1715 1716 1716 1717 1717 1718 22 1719 2 1720 2 1721 2 1722 11 1723 1723 1724 11 1725 1725 1726 11 1727 1727 1728 1728 1729 1729 1730 1730 1731 60 1732 1732 1733 2 1734 2 1735 1735 1736 11 1737 2 1738 2 1739 69 1740 133 1741 1741 1742 2 1743 2 1744 1744 1745 2 1746 1746 1747 1747 1748 2 1749 86 1750 87 1751 22 1752 90 1753 1753 1754 441 1755 136 1756 11 1757 1757 1758 1758 1759 101 1760 1760 1761 1761 1762 1762 1763 106 1764 106 1765 1765 1766 1766 1767 1767 1768 1768 1769 1769 1770 1770 1771 1771 1772 1772 1773 1773 1774 1774 1775 1775 1776 1776 1777 1777 1778 101 1779 2 1780 11 1781 1781 1782 2 1783 2 1784 2 1785 2 1786 2 1787 2 1788 2 1789 2 1790 312 1791 2 1792 2 1793 2 1794 2 1795 2 1796 318 1797 2 1798 2 1799 2 1800 2 1801 2 1802 2 1803 2 1804 2 1805 2 1806 2 1807 2 1808 2 1809 2 1810 2 1811 2 1812 2 1813 1813 1814 1814 1815 1815 1816 1816 1817 1817 1818 1818 1819 1819 1820 1820 1821 1821 1822 1822 1823 1823 1824 1824 1825 1825 1826 1826 1827 1827 1828 1828 1829 4 1830 2 1831 1831 1832 1832 1833 1833 1834 1834 1835 1835 1836 1836 1837 1837 1838 1838 1839 1839 1840 1840 1841 1841 1842 364 1843 364 1844 40 1845 367 1846 441 1847 1847 1848 1848 1849 1849 1850 1850 1851 1851 1852 2 1853 583 1854 1854 1855 1855 1856 40 1857 1857 1858 287 1859 1859 1860 1860 1861 2 1862 583 1863 1863 1864 1864 1865 1865 1866 1866 1867 1867 1868 1868 1869 2 1870 751 1871 2 1872 1872 1873 1873 1874 1874 1875 1875 1876 1876 1877 1877 1878 1878 1879 1879 1880 1880 1881 1881 1882 1882 1883 2 1884 3 1885 4 1886 2 1887 2 1888 2 1889 2 1890 11 1891 2 1892 2 1893 1235 1894 2 1895 1237 1896 22 1897 1239 1898 1240 1899 2 1900 2 1901 11 1902 11 1903 1244 1904 1245 1905 1246 1906 22 1907 2 1908 2 1909 2 1910 11 1911 1252 1912 11 1913 1254 1914 11 1915 1256 1916 1257 1917 1258 1918 1259 1919 60 1920 1261 1921 2 1922 2 1923 1264 1924 11 1925 2 1926 2 1927 69 1928 133 1929 1270 1930 2 1931 2 1932 1273 1933 2 1934 1275 1935 1276 1936 2 1937 86 1938 87 1939 22 1940 90 1941 1941 1942 164 1943 136 1944 11 1945 1283 1946 1284 1947 101 1948 1286 1949 1287 1950 1950 1951 106 1952 106 1953 1291 1954 1292 1955 1293 1956 1294 1957 1295 1958 1296 1959 1297 1960 1960 1961 1961 1962 1298 1963 1299 1964 1300 1965 1301 1966 101 1967 2 1968 11 1969 1304 1970 2 1971 2 1972 2 1973 2 1974 2 1975 2 1976 2 1977 2 1978 312 1979 2 1980 2 1981 2 1982 2 1983 2 1984 318 1985 2 1986 2 1987 2 1988 2 1989 2 1990 2 1991 2 1992 2 1993 2 1994 2 1995 2 1996 2 1997 2 1998 2 1999 2 2000 2 2001 1336 2002 1337 2003 1338 2004 1339 2005 1340 2006 1341 2007 1342 2008 1343 2009 1344 2010 1345 2011 1346 2012 1347 2013 1348 2014 1349 2015 1350 2016 1351 2017 4 2018 2 2019 1354 2020 1355 2021 1356 2022 1357 2023 1358 2024 1359 2025 1360 2026 1361 2027 1362 2028 1363 2029 1364 2030 364 2031 364 2032 40 2033 367 2034 441 2035 1370 2036 1371 2037 1372 2038 1373 2039 1374 2040 2 2041 583 2042 1377 2043 1378 2044 40 2045 1380 2046 287 2047 1382 2048 1383 2049 2 2050 583 2051 1386 2052 1387 2053 1388 2054 1389 2055 1390 2056 1391 2057 2 2058 751 2059 2 2060 1395 2061 1396 2062 1397 2063 1398 2064 1399 2065 1400 2066 1401 2067 1402 2068 1403 2069 1404 2070 1882 2071 164 2072 2072 2073 2 2074 2074 2075 868 2076 1406 2077 1407 2078 1408 2079 1409 2080 1410 2081 1411 2082 1412 2083 1413 2084 1414 2085 1415 2086 1416 2087 1417 2088 1418 2089 1419 2090 1420 2091 1421 2092 1422 2093 1423 2094 1424 2095 1425 2096 1426 2097 1427 2098 1428 2099 1429 2100 1430 2101 1431 2102 1432 2103 1433 2104 1434 2105 1435 2106 1436 2107 1437 2108 1438 2109 1439 2110 1440 2111 1441 2112 1442 2113 1443 2114 1444 2115 1445 2116 1446 2117 1447 2118 1448 2119 1449 2120 1450 2121 1451 2122 1452 2123 1453 2124 1454 2125 1455 2126 1456 2127 1457 2128 1458 2129 1459 2130 1460 2131 1461 2132 1462 2133 1463 2134 1464 2135 1465 2136 1466 2137 1467 2138 1468 2139 1469 2140 1470 2141 1471 2142 1472 2143 1473 2144 1474 2145 1475 2146 1476 2147 1477 2148 1478 2149 1479 2150 1480 2151 1481 2152 1482 2153 1483 2154 1484 2155 1485 2156 11 2157 1487 2158 1488 2159 1489 2160 2 2161 583 2162 1492 2163 956 2164 1494 2165 1495 2166 989 2167 161 2168 164 2169 1500 2170 1504 2171 1505 2172 2 2173 2 2174 2 2175 1509 2176 2 2177 1511 2178 2 2179 2 2180 2 2181 2 2182 2 2183 2 2184 2 2185 2 2186 2 2187 1521 2188 1522 2189 1523 2190 1524 2191 1525 2192 1526 2193 1527 2194 1528 2195 1529 2196 1530 2197 1531 2198 1532 2199 1533 2200 22 2201 1537 2202 4 2203 11 2204 1559 2205 90 2206 11 2207 1563 2208 101 2209 1568 2210 1569 2211 1571 2212 1572 2213 1573 2214 101 2215 868 2216 11 2217 1659 2218 956 2219 1664 2220 1665 2221 989 2222 161 2223 164 2224 1669 2225 1670 2226 1671 2227 868 2228 11 2229 1675 2230 1679 2231 1680 2232 364 2233 364 2234 367 2235 441 2236 1685 2237 1686 2238 1687 2239 1688 2240 2 2241 583 2242 1691 2243 1692 2244 1693 2245 1694 2246 2246 2247 2247 2248 2 2249 583 2250 2250 2251 1882 2252 2252 2253 2253 2254 142 2255 11 2256 144 2257 164 2258 165 2259 167 2260 169 2261 77 2262 128 2263 172 2264 2264 2265 2265 2266 142 2267 11 2268 144 2269 164 2270 2270 2271 168 2272 2272 2273 2273 2274 2274 2275 2 2276 2276 2277 2277 2278 2278 2279 2278 2280 2280 2281 2281 2282 2282 2283 2283 2284 2284 2285 2285 2286 2286 2287 2287 2288 2288 2289 2289 2290 2290 2291 2291 2292 2292 2293 2293 2294 2294 2295 2278 2296 2296 2297 2297 2298 2298 2299 2299 2300 2300 2301 2301 2302 2302 2303 2303 2304 2304 2305 2305 2306 2306 2307 2307 2308 2308 2309 2309 2310 2310 2311 2311 2312 2312 2313 2313 2314 2314 2315 2315 2316 2316 2317 2317 2318 2278 2319 2319 2320 2320 2321 2321 2322 2322 2323 2323 2324 2324 2325 2325 2326 2326 2327 2327 2328 2328 2329 2329 2330 2278 2331 2331 2332 2294 2333 2333 2334 2278 2335 2335 2336 2336 2337 2337 2338 2338 2339 2339 2340 2340 2341 2341 2342 2342 2343 2343 2344 2344 2345 2278 2346 2346 2347 2347 2348 2348 2349 2349 2350 2350 2351 2351 2352 2352 2353 2353 2354 2354 2355 2355 2356 2356 2357 2314 2358 2358 2359 1505 2360 2 2361 3 2362 4 2363 2 2364 2 2365 2 2366 2 2367 11 2368 2 2369 2 2370 1097 2371 2 2372 979 2373 22 2374 1099 2375 985 2376 2 2377 11 2378 11 2379 1102 2380 1103 2381 1104 2382 22 2383 2 2384 2 2385 2 2386 11 2387 1108 2388 11 2389 1025 2390 11 2391 1111 2392 1032 2393 1033 2394 1034 2395 60 2396 1113 2397 2 2398 2 2399 1115 2400 11 2401 2 2402 2 2403 69 2404 133 2405 1120 2406 2 2407 2 2408 1122 2409 2 2410 1060 2411 1123 2412 2 2413 86 2414 87 2415 22 2416 90 2417 136 2418 11 2419 1071 2420 1072 2421 101 2422 1074 2423 1075 2424 2424 2425 106 2426 106 2427 1079 2428 1081 2429 1082 2430 1083 2431 1084 2432 1085 2433 1086 2434 1089 2435 1091 2436 1092 2437 1093 2438 101 2439 2 2440 11 2441 1124 2442 2 2443 2 2444 2 2445 2 2446 2 2447 2 2448 2 2449 2 2450 312 2451 2 2452 2 2453 2 2454 2 2455 2 2456 318 2457 2 2458 2 2459 2 2460 2 2461 2 2462 2 2463 2 2464 2 2465 2 2466 2 2467 2 2468 2 2469 2 2470 2 2471 2 2472 2 2473 1156 2474 1157 2475 1158 2476 1159 2477 1160 2478 1161 2479 1162 2480 1163 2481 1164 2482 1165 2483 1166 2484 1167 2485 1168 2486 1169 2487 1170 2488 1171 2489 4 2490 2 2491 1174 2492 1175 2493 1176 2494 1177 2495 1178 2496 1179 2497 1180 2498 1181 2499 1182 2500 1183 2501 1184 2502 364 2503 364 2504 40 2505 367 2506 441 2507 1190 2508 1191 2509 1192 2510 1193 2511 1194 2512 2 2513 583 2514 1197 2515 1198 2516 40 2517 1200 2518 287 2519 1202 2520 1203 2521 2 2522 583 2523 1206 2524 1207 2525 1208 2526 1209 2527 1210 2528 1211 2529 2 2530 751 2531 2 2532 1215 2533 1216 2534 1217 2535 1218 2536 1219 2537 1220 2538 1221 2539 1222 2540 1223 2541 1224 2542 1882 2543 164 2544 2072 2545 2 2546 2546 2547 868 2548 2548 2549 2549 2550 2550 2551 2551 2552 2552 2553 2553 2554 2554 2555 2555 2556 2556 2557 2557 2558 2558 2559 2559 2560 2560 2561 2561 2562 2562 2563 2563 2564 2564 2565 2565 2566 2566 2567 2567 2568 2568 2569 2569 2570 2570 2571 2571 2572 2572 2573 2573 2574 2574 2575 2575 2576 2576 2577 2577 2578 2578 2579 2579 2580 2580 2581 2581 2582 2582 2583 2583 2584 2584 2585 2585 2586 2586 2587 2587 2588 2588 2589 2589 2590 2590 2591 2591 2592 2592 2593 2593 2594 2594 2595 2595 2596 2596 2597 2597 2598 2598 2599 2599 2600 2600 2601 2601 2602 2602 2603 2603 2604 2604 2605 2605 2606 2606 2607 2607 2608 2608 2609 2609 2610 2610 2611 2611 2612 2612 2613 2613 2614 2614 2615 2615 2616 2616 2617 2617 2618 2618 2619 2619 2620 2620 2621 2621 2622 2622 2623 2623 2624 2624 2625 2625 2626 2626 2627 2627 2628 11 2629 2629 2630 2630 2631 2631 2632 2 2633 583 2634 2634 2635 956 2636 2636 2637 2637 2638 989 2639 161 2640 164 2641 2641 2642 2642 2643 1505 2644 2 2645 2 2646 2 2647 2647 2648 2 2649 2649 2650 2 2651 2 2652 2 2653 2 2654 2 2655 2 2656 2 2657 2 2658 2 2659 2659 2660 2660 2661 2661 2662 2662 2663 2663 2664 2664 2665 2665 2666 2666 2667 2667 2668 2668 2669 2669 2670 2670 2671 2671 2672 22 2673 2673 2674 4 2675 11 2676 2676 2677 90 2678 11 2679 2679 2680 101 2681 2681 2682 1569 2683 2683 2684 2684 2685 2685 2686 101 2687 868 2688 11 2689 2689 2690 956 2691 2691 2692 2692 2693 989 2694 161 2695 164 2696 2696 2697 2697 2698 2698 2699 868 2700 11 2701 2701 2702 2702 2703 2703 2704 364 2705 364 2706 367 2707 441 2708 2708 2709 2709 2710 2710 2711 2711 2712 2 2713 583 2714 2714 2715 2715 2716 2716 2717 2717 2718 2718 2719 2719 2720 2 2721 583 2722 2722 2723 2723 2724 142 2725 11 2726 144 2727 164 2728 165 2729 167 2730 169 2731 77 2732 128 2733 172 2734 2734 2735 2735 2736 142 2737 11 2738 144 2739 164 2740 2270 2741 2272 2742 2273 2743 2274 2744 2 2745 2745 2746 2746 2747 868 2748 2748 2749 11 2750 2750 2751 2 2752 583 2753 2753 2754 2754 2755 2755 2756 2756 2757 2757 2758 2758 2759 2759 2760 364 2761 364 2762 367 2763 441 2764 2764 2765 2765 2766 2766 2767 2767 2768 2 2769 583 2770 2770 2771 2771 2772 2772 2773 2773 2774 4 2775 11 2776 2776 2777 90 2778 11 2779 2779 2780 101 2781 2781 2782 1569 2783 2783 2784 2784 2785 2785 2786 101 2787 868 2788 11 2789 2789 2790 956 2791 2791 2792 2792 2793 989 2794 161 2795 164 2796 2796 2797 2797 2798 2798 2799 868 2800 11 2801 2801 2802 2802 2803 2803 2804 364 2805 364 2806 367 2807 441 2808 2808 2809 2809 2810 2810 2811 2811 2812 2 2813 583 2814 2814 2815 2815 2816 2816 2817 2817 2818 2818 2819 2819 2820 2 2821 583 2822 2822 2823 22 2824 2824 2825 2825 2826 2826 2827 2827 2828 2828 2829 2829 2830 2830 2831 2831 2832 2832 2833 16 2834 16 2835 2835 2836 2836 2837 2837 2838 2838 2839 2839 2840 441 2841 2841 2842 2842 2843 2843 2844 2844 2845 2845 2846 2846 2847 583 2848 2848 2849 2 2850 2850 2851 2851 2852 2852 2853 2853 2854 4 2855 11 2856 2856 2857 90 2858 11 2859 2859 2860 101 2861 2861 2862 1569 2863 2863 2864 2864 2865 2865 2866 101 2867 868 2868 11 2869 2869 2870 956 2871 2871 2872 2872 2873 989 2874 161 2875 164 2876 2876 2877 2877 2878 2878 2879 2879 2880 868 2881 11 2882 2882 2883 2883 2884 2884 2885 364 2886 364 2887 367 2888 441 2889 2889 2890 2890 2891 2891 2892 2892 2893 2 2894 583 2895 2895 2896 2896 2897 2897 2898 2898 2899 2899 2900 2900 2901 2 2902 583 2903 2903 2904 441 2905 2905 2906 2 2907 2907 2908 2908 2909 1288 2910 2 2911 2911 2912 2912 2913 2913 2914 2914 2915 2915 2916 22 2917 2824 2918 2918 2919 2919 2920 2920 2921 69 2922 2922 2923 2923 2924 2924 2925 2925 2926 16 2927 16 2928 2835 2929 441 2930 2930 2931 2931 2932 2932 2933 2933 2934 2934 2935 2846 2936 583 2937 2937 2938 2 2939 2939 2940 2851 2941 2852 2942 2942 2943 2865 2944 2944 2945 2945 2946 2946 2947 2947 2948 2948 2949 2949 2950 2 2951 13 2952 2952 2953 2953 2954 2954 2955 2955 2956 2956 2957 2957 2958 2958 2959 2959 2960 2960 2961 2961 2962 2962 2963 2963 2964 2964 2965 2965 2966 2966 2967 2967 2968 2968 2969 364 2970 364 2971 40 2972 367 2973 441 2974 2974 2975 2975 2976 2976 2977 2977 2978 2978 2979 40 2980 2980 2981 2981 2982 2982 2983 2983 2984 956 2985 2985 2986 2986 2987 13 2988 2988 2989 13 2990 2990 2991 77 2992 2992 2993 2993 2994 2994 2995 13 2996 2996 2997 2997 2998 2998 2999 2999 3000 3000 3001 3001 3002 3002 3003 3003 3004 3004 3005 3005 3006 13 3007 3007 3008 3008 3009 3009 3010 3010 3011 3011 3012 3012 3013 3013 3014 3014 3015 3015 3016 3016 3017 3017 3018 3018 3019 3019 3020 3019 3021 3021 3022 3022 3023 3023 3024 3024 3025 3025 3026 3026 3027 3027 3028 3028 3029 3029 3030 3030 3031 3031 3032 3032 3033 2293 3034 3034 3035 3019 3036 3036 3037 3037 3038 3038 3039 3039 3040 3040 3041 3041 3042 3042 3043 3043 3044 3044 3045 2314 3046 2316 3047 3019 3048 3048 3049 3049 3050 3019 3051 3051 3052 3034 3053 3053 3054 3019 3055 3055 3056 3056 3057 3057 3058 3058 3059 3059 3060 3060 3061 3061 3062 3062 3063 3063 3064 3064 3065 3065 3066 3066 3067 3029 3068 3068 3069 3069 3070 3070 3071 2 3072 3072 3073 2 3074 3 3075 4 3076 2 3077 2 3078 2 3079 2 3080 11 3081 2 3082 2 3083 3083 3084 2 3085 3085 3086 22 3087 3087 3088 3088 3089 2 3090 11 3091 11 3092 3092 3093 3093 3094 3094 3095 22 3096 2 3097 2 3098 2 3099 11 3100 3100 3101 11 3102 3102 3103 11 3104 3104 3105 3105 3106 3106 3107 3107 3108 60 3109 3109 3110 2 3111 2 3112 3112 3113 11 3114 2 3115 2 3116 69 3117 133 3118 3118 3119 2 3120 2 3121 3121 3122 2 3123 3123 3124 3124 3125 2 3126 86 3127 87 3128 90 3129 136 3130 11 3131 3131 3132 3132 3133 101 3134 3134 3135 3135 3136 3136 3137 106 3138 106 3139 3139 3140 3140 3141 3141 3142 3142 3143 3143 3144 3144 3145 3145 3146 3146 3147 3147 3148 3148 3149 3149 3150 101 3151 2 3152 3152 3153 2 3154 2 3155 2 3156 2 3157 2 3158 2 3159 2 3160 2 3161 312 3162 2 3163 2 3164 2 3165 2 3166 2 3167 318 3168 2 3169 2 3170 2 3171 2 3172 2 3173 2 3174 2 3175 2 3176 2 3177 2 3178 2 3179 2 3180 2 3181 2 3182 2 3183 2 3184 3184 3185 3185 3186 3186 3187 3187 3188 3188 3189 3189 3190 3190 3191 3191 3192 3192 3193 3193 3194 3194 3195 3195 3196 3196 3197 3197 3198 3198 3199 3199 3200 4 3201 2 3202 3202 3203 3203 3204 3204 3205 3205 3206 3206 3207 3207 3208 3208 3209 3209 3210 3210 3211 3211 3212 3212 3213 364 3214 364 3215 40 3216 367 3217 441 3218 3218 3219 3219 3220 3220 3221 3221 3222 3222 3223 2 3224 583 3225 3225 3226 3226 3227 40 3228 3228 3229 287 3230 3230 3231 3231 3232 2 3233 583 3234 3234 3235 3235 3236 3236 3237 3237 3238 3238 3239 3239 3240 2 3241 751 3242 2 3243 3243 3244 3244 3245 3245 3246 3246 3247 3247 3248 3248 3249 3249 3250 3250 3251 3251 3252 3252 3253 868 3254 3254 3255 3255 3256 3256 3257 3257 3258 3258 3259 3259 3260 3260 3261 3261 3262 3262 3263 3263 3264 3264 3265 3265 3266 3266 3267 3267 3268 3268 3269 3269 3270 3270 3271 3271 3272 3272 3273 3273 3274 3274 3275 3275 3276 3276 3277 3277 3278 3278 3279 3279 3280 3280 3281 3281 3282 3282 3283 3283 3284 3284 3285 3285 3286 3286 3287 3287 3288 3288 3289 3289 3290 3290 3291 3291 3292 3292 3293 3293 3294 3294 3295 3295 3296 3296 3297 3297 3298 3298 3299 3299 3300 3300 3301 3301 3302 3302 3303 3303 3304 3304 3305 3305 3306 3306 3307 3307 3308 3308 3309 3309 3310 3310 3311 3311 3312 3312 3313 3313 3314 3314 3315 3315 3316 3316 3317 3317 3318 3318 3319 3319 3320 3320 3321 3321 3322 3322 3323 3323 3324 3324 3325 3325 3326 3326 3327 3327 3328 3328 3329 3329 3330 3330 3331 3331 3332 3332 3333 3333 3334 11 3335 3335 3336 3336 3337 3337 3338 2 3339 583 3340 3340 3341 956 3342 3342 3343 3343 3344 989 3345 161 3346 164 3347 3347 3348 3348 3349 1505 3350 2 3351 2 3352 2 3353 3353 3354 2 3355 3355 3356 2 3357 2 3358 2 3359 2 3360 2 3361 2 3362 2 3363 2 3364 2 3365 3365 3366 3366 3367 3367 3368 3368 3369 3369 3370 3370 3371 3371 3372 3372 3373 3373 3374 3374 3375 3375 3376 3376 3377 3377 3378 22 3379 3379 3380 4 3381 11 3382 3382 3383 90 3384 11 3385 3385 3386 101 3387 3387 3388 1569 3389 3389 3390 3390 3391 3391 3392 101 3393 868 3394 11 3395 3395 3396 956 3397 3397 3398 3398 3399 989 3400 161 3401 164 3402 3402 3403 3403 3404 3404 3405 868 3406 11 3407 3407 3408 3408 3409 3409 3410 364 3411 364 3412 367 3413 441 3414 3414 3415 3415 3416 3416 3417 3417 3418 2 3419 583 3420 3420 3421 3421 3422 3422 3423 3423 3424 3424 3425 3425 3426 2 3427 583 3428 3428 3429 3429 3430 142 3431 11 3432 144 3433 164 3434 165 3435 167 3436 169 3437 77 3438 128 3439 172 3440 3440 3441 3441 3442 142 3443 11 3444 144 3445 164 3446 2270 3447 2272 3448 2273 3449 2274 3450 2 3451 3451 3452 3452 3453 3453 3454 868 3455 3455 3456 11 3457 3457 3458 2 3459 583 3460 3460 3461 3461 3462 3462 3463 2756 3464 3464 3465 3465 3466 364 3467 364 3468 367 3469 441 3470 3470 3471 3471 3472 3472 3473 3473 3474 2 3475 583 3476 3476 3477 3477 3478 3478 3479 3479 3480 2827 3481 3481 3482 441 3483 3483 3484 2 3485 3485 3486 3486 3487 2 3488 3 3489 4 3490 2 3491 2 3492 2 3493 2 3494 11 3495 2 3496 2 3497 3497 3498 2 3499 3499 3500 22 3501 3501 3502 3502 3503 2 3504 11 3505 11 3506 3506 3507 3507 3508 3508 3509 22 3510 2 3511 2 3512 2 3513 11 3514 3514 3515 11 3516 3516 3517 11 3518 3518 3519 3519 3520 3520 3521 3521 3522 60 3523 3523 3524 2 3525 2 3526 3526 3527 11 3528 2 3529 2 3530 69 3531 133 3532 3532 3533 2 3534 2 3535 3535 3536 2 3537 3537 3538 3538 3539 2 3540 86 3541 87 3542 90 3543 136 3544 11 3545 3545 3546 3546 3547 101 3548 3548 3549 3549 3550 3550 3551 106 3552 106 3553 3553 3554 3554 3555 3555 3556 3556 3557 3557 3558 3558 3559 3559 3560 3560 3561 3561 3562 3562 3563 3563 3564 101 3565 2 3566 3566 3567 2 3568 2 3569 2 3570 2 3571 2 3572 2 3573 2 3574 2 3575 312 3576 2 3577 2 3578 2 3579 2 3580 2 3581 318 3582 2 3583 2 3584 2 3585 2 3586 2 3587 2 3588 2 3589 2 3590 2 3591 2 3592 2 3593 2 3594 2 3595 2 3596 2 3597 2 3598 3598 3599 3599 3600 3600 3601 3601 3602 3602 3603 3603 3604 3604 3605 3605 3606 3606 3607 3607 3608 3608 3609 3609 3610 3610 3611 3611 3612 3612 3613 3613 3614 4 3615 2 3616 3616 3617 3617 3618 3618 3619 3619 3620 3620 3621 3621 3622 3622 3623 3623 3624 3624 3625 3625 3626 3626 3627 364 3628 364 3629 40 3630 367 3631 441 3632 3632 3633 3633 3634 3634 3635 3635 3636 3636 3637 2 3638 583 3639 3639 3640 3640 3641 40 3642 3642 3643 287 3644 3644 3645 3645 3646 2 3647 583 3648 3648 3649 3649 3650 3650 3651 3651 3652 3652 3653 3653 3654 2 3655 751 3656 2 3657 3657 3658 3658 3659 3659 3660 3660 3661 3661 3662 3662 3663 3663 3664 3664 3665 3665 3666 3666 3667 868 3668 3668 3669 3669 3670 3670 3671 3671 3672 3672 3673 3673 3674 3674 3675 3675 3676 3676 3677 3677 3678 3678 3679 3679 3680 3680 3681 3681 3682 3682 3683 3683 3684 3684 3685 3685 3686 3686 3687 3687 3688 3688 3689 3689 3690 3690 3691 3691 3692 3692 3693 3693 3694 3694 3695 3695 3696 3696 3697 3697 3698 3698 3699 3699 3700 3700 3701 3701 3702 3702 3703 3703 3704 3704 3705 3705 3706 3706 3707 3707 3708 3708 3709 3709 3710 3710 3711 3711 3712 3712 3713 3713 3714 3714 3715 3715 3716 3716 3717 3717 3718 3718 3719 3719 3720 3720 3721 3721 3722 3722 3723 3723 3724 3724 3725 3725 3726 3726 3727 3727 3728 3728 3729 3729 3730 3730 3731 3731 3732 3732 3733 3733 3734 3734 3735 3735 3736 3736 3737 3737 3738 3738 3739 3739 3740 3740 3741 3741 3742 3742 3743 3743 3744 3744 3745 3745 3746 3746 3747 3747 3748 11 3749 3749 3750 3750 3751 3751 3752 2 3753 583 3754 3754 3755 956 3756 3756 3757 3757 3758 989 3759 161 3760 164 3761 3761 3762 3762 3763 1505 3764 2 3765 2 3766 2 3767 3767 3768 2 3769 3769 3770 2 3771 2 3772 2 3773 2 3774 2 3775 2 3776 2 3777 2 3778 2 3779 3779 3780 3780 3781 3781 3782 3782 3783 3783 3784 3784 3785 3785 3786 3786 3787 3787 3788 3788 3789 3789 3790 3790 3791 3791 3792 22 3793 3793 3794 4 3795 11 3796 3796 3797 90 3798 11 3799 3799 3800 101 3801 3801 3802 1569 3803 3803 3804 3804 3805 3805 3806 101 3807 868 3808 11 3809 3809 3810 956 3811 3811 3812 3812 3813 989 3814 161 3815 164 3816 3816 3817 3817 3818 3818 3819 868 3820 11 3821 3821 3822 3822 3823 3823 3824 364 3825 364 3826 367 3827 441 3828 3828 3829 3829 3830 3830 3831 3831 3832 2 3833 583 3834 3834 3835 3835 3836 3836 3837 3837 3838 3838 3839 3839 3840 2 3841 583 3842 3842 3843 3843 3844 142 3845 11 3846 144 3847 164 3848 165 3849 167 3850 169 3851 77 3852 128 3853 172 3854 3854 3855 3855 3856 142 3857 11 3858 144 3859 164 3860 2270 3861 2272 3862 2273 3863 2274 3864 2 3865 3865 3866 3866 3867 2 3868 3868 3869 868 3870 3870 3871 11 3872 3872 3873 2 3874 583 3875 3875 3876 3876 3877 3877 3878 2756 3879 3879 3880 3880 3881 364 3882 364 3883 367 3884 441 3885 3885 3886 3886 3887 3887 3888 3888 3889 2 3890 583 3891 3891 3892 3892 3893 3893 3894 3894 3895 2827 3896 3896 3897 441 3898 3898 3899 2 3900 3900 3901 3901 3902 3550 3903 2 3904 3904 3905 3905 3906 3906 3907 3907 3908 3908 3909 2918 3910 3910 3911 69 3912 3912 3913 3913 3914 3914 3915 3915 3916 3916 3917 3917 3918 3918 3919 2 3920 3920 3921 3921 3922 3922 3923 3923 3924 3924 3925 3925 3926 2960 3927 2962 3928 3928 3929 3929 3930 3930 3931 3931 3932 3932 3933 364 3934 364 3935 40 3936 367 3937 441 3938 3938 3939 3939 3940 3940 3941 3941 3942 3942 3943 40 3944 3944 3945 3945 3946 2983 3947 2986 3948 2988 3949 2990 3950 77 3951 2992 3952 2993 3953 3953 3954 287 3955 3955 3956 2 3957 287 3958 77 3959 3959 3960 2 3961 3 3962 4 3963 2 3964 2 3965 2 3966 2 3967 11 3968 2 3969 2 3970 3970 3971 2 3972 3972 3973 22 3974 3974 3975 3975 3976 3976 3977 2 3978 11 3979 11 3980 3980 3981 3981 3982 3982 3983 22 3984 2 3985 2 3986 2 3987 11 3988 3988 3989 11 3990 3990 3991 11 3992 3992 3993 3993 3994 3994 3995 3995 3996 60 3997 3997 3998 2 3999 2 4000 4000 4001 11 4002 2 4003 2 4004 69 4005 133 4006 4006 4007 2 4008 2 4009 4009 4010 2 4011 4011 4012 4012 4013 2 4014 86 4015 87 4016 90 4017 136 4018 11 4019 4019 4020 4020 4021 101 4022 4022 4023 4023 4024 4024 4025 106 4026 106 4027 4027 4028 4028 4029 4029 4030 4030 4031 4031 4032 3974 4033 2852 4034 4034 4035 4035 4036 4036 4037 4037 4038 4038 4039 4039 4040 4040 4041 4041 4042 101 4043 2 4044 4044 4045 2 4046 2 4047 2 4048 2 4049 2 4050 2 4051 2 4052 2 4053 312 4054 2 4055 2 4056 2 4057 2 4058 2 4059 318 4060 2 4061 2 4062 2 4063 2 4064 2 4065 2 4066 2 4067 2 4068 2 4069 2 4070 2 4071 2 4072 2 4073 2 4074 2 4075 2 4076 4076 4077 4077 4078 4078 4079 4079 4080 4080 4081 4081 4082 4082 4083 4083 4084 4084 4085 4085 4086 4086 4087 4087 4088 4088 4089 4089 4090 4090 4091 4091 4092 4 4093 2 4094 4094 4095 4095 4096 4096 4097 4097 4098 4098 4099 4099 4100 4100 4101 4101 4102 4102 4103 4103 4104 4104 4105 364 4106 364 4107 40 4108 367 4109 441 4110 4110 4111 4111 4112 4112 4113 4113 4114 4114 4115 2 4116 583 4117 4117 4118 4118 4119 40 4120 4120 4121 287 4122 4122 4123 4123 4124 2 4125 583 4126 4126 4127 4127 4128 4128 4129 4129 4130 4130 4131 4131 4132 2 4133 751 4134 2 4135 4135 4136 4136 4137 4137 4138 4138 4139 4139 4140 4140 4141 4141 4142 4142 4143 4143 4144 4144 4145 868 4146 4146 4147 4147 4148 4148 4149 4149 4150 4150 4151 4151 4152 4152 4153 4153 4154 4154 4155 4155 4156 4156 4157 4157 4158 4158 4159 4159 4160 4160 4161 4161 4162 4162 4163 4163 4164 4164 4165 4165 4166 4166 4167 4167 4168 4168 4169 4169 4170 4170 4171 4171 4172 4172 4173 4173 4174 4174 4175 4175 4176 4176 4177 4177 4178 4178 4179 4179 4180 4180 4181 4181 4182 4182 4183 4183 4184 4184 4185 4185 4186 4186 4187 4187 4188 4188 4189 4189 4190 4190 4191 4191 4192 4192 4193 4193 4194 4194 4195 4195 4196 4196 4197 4197 4198 4198 4199 4199 4200 4200 4201 4201 4202 4202 4203 4203 4204 4204 4205 4205 4206 4206 4207 4207 4208 4208 4209 4209 4210 4210 4211 4211 4212 4212 4213 4213 4214 4214 4215 4215 4216 4216 4217 4217 4218 4218 4219 4219 4220 4220 4221 4221 4222 4222 4223 4223 4224 4224 4225 4225 4226 11 4227 4227 4228 4228 4229 4229 4230 2 4231 583 4232 4232 4233 956 4234 4234 4235 4235 4236 989 4237 161 4238 164 4239 4239 4240 4240 4241 1505 4242 2 4243 2 4244 2 4245 4245 4246 2 4247 4247 4248 2 4249 2 4250 2 4251 2 4252 2 4253 2 4254 2 4255 2 4256 2 4257 4257 4258 4258 4259 4259 4260 4260 4261 4261 4262 4262 4263 4263 4264 4264 4265 4265 4266 4266 4267 4267 4268 4268 4269 4269 4270 22 4271 4271 4272 4 4273 11 4274 4274 4275 90 4276 11 4277 4277 4278 101 4279 4279 4280 1569 4281 4281 4282 4282 4283 4283 4284 101 4285 868 4286 11 4287 4287 4288 956 4289 4289 4290 4290 4291 989 4292 161 4293 164 4294 4294 4295 4295 4296 4296 4297 868 4298 11 4299 4299 4300 4300 4301 4301 4302 364 4303 364 4304 367 4305 441 4306 4306 4307 4307 4308 4308 4309 4309 4310 2 4311 583 4312 4312 4313 4313 4314 4314 4315 4315 4316 4316 4317 4317 4318 2 4319 583 4320 4320 4321 4321 4322 4322 4323 4323 4324 4324 4325 4325 4326 4326 4327 2272 4328 2 4329 2 4330 3 4331 4 4332 2 4333 2 4334 2 4335 2 4336 11 4337 2 4338 2 4339 4339 4340 2 4341 4341 4342 22 4343 4343 4344 4344 4345 2 4346 11 4347 11 4348 4348 4349 4349 4350 4350 4351 22 4352 2 4353 2 4354 2 4355 11 4356 4356 4357 11 4358 4358 4359 11 4360 4360 4361 4361 4362 4362 4363 4363 4364 60 4365 4365 4366 2 4367 2 4368 4368 4369 11 4370 2 4371 2 4372 69 4373 133 4374 4374 4375 2 4376 2 4377 4377 4378 2 4379 4379 4380 4380 4381 2 4382 86 4383 87 4384 90 4385 136 4386 11 4387 4387 4388 4388 4389 101 4390 4390 4391 4391 4392 4392 4393 106 4394 106 4395 4395 4396 4396 4397 4397 4398 4398 4399 4399 4400 4400 4401 4401 4402 4402 4403 4403 4404 4404 4405 4405 4406 101 4407 2 4408 4408 4409 2 4410 2 4411 2 4412 2 4413 2 4414 2 4415 2 4416 2 4417 312 4418 2 4419 2 4420 2 4421 2 4422 2 4423 318 4424 2 4425 2 4426 2 4427 2 4428 2 4429 2 4430 2 4431 2 4432 2 4433 2 4434 2 4435 2 4436 2 4437 2 4438 2 4439 2 4440 4440 4441 4441 4442 4442 4443 4443 4444 4444 4445 4445 4446 4446 4447 4447 4448 4448 4449 4449 4450 4450 4451 4451 4452 4452 4453 4453 4454 4454 4455 4455 4456 4 4457 2 4458 4458 4459 4459 4460 4460 4461 4461 4462 4462 4463 4463 4464 4464 4465 4465 4466 4466 4467 4467 4468 4468 4469 364 4470 364 4471 40 4472 367 4473 441 4474 4474 4475 4475 4476 4476 4477 4477 4478 4478 4479 2 4480 583 4481 4481 4482 4482 4483 40 4484 4484 4485 287 4486 4486 4487 4487 4488 2 4489 583 4490 4490 4491 4491 4492 4492 4493 4493 4494 4494 4495 4495 4496 2 4497 751 4498 2 4499 4499 4500 4500 4501 4501 4502 4502 4503 4503 4504 4504 4505 4505 4506 4506 4507 4507 4508 4508 4509 868 4510 4510 4511 4511 4512 4512 4513 4513 4514 4514 4515 4515 4516 4516 4517 4517 4518 4518 4519 4519 4520 4520 4521 4521 4522 4522 4523 4523 4524 4524 4525 4525 4526 4526 4527 4527 4528 4528 4529 4529 4530 4530 4531 4531 4532 4532 4533 4533 4534 4534 4535 4535 4536 4536 4537 4537 4538 4538 4539 4539 4540 4540 4541 4541 4542 4542 4543 4543 4544 4544 4545 4545 4546 4546 4547 4547 4548 4548 4549 4549 4550 4550 4551 4551 4552 4552 4553 4553 4554 4554 4555 4555 4556 4556 4557 4557 4558 4558 4559 4559 4560 4560 4561 4561 4562 4562 4563 4563 4564 4564 4565 4565 4566 4566 4567 4567 4568 4568 4569 4569 4570 4570 4571 4571 4572 4572 4573 4573 4574 4574 4575 4575 4576 4576 4577 4577 4578 4578 4579 4579 4580 4580 4581 4581 4582 4582 4583 4583 4584 4584 4585 4585 4586 4586 4587 4587 4588 4588 4589 4589 4590 11 4591 4591 4592 4592 4593 4593 4594 2 4595 583 4596 4596 4597 956 4598 4598 4599 4599 4600 989 4601 161 4602 164 4603 4603 4604 4604 4605 1505 4606 2 4607 2 4608 2 4609 4609 4610 2 4611 4611 4612 2 4613 2 4614 2 4615 2 4616 2 4617 2 4618 2 4619 2 4620 2 4621 4621 4622 4622 4623 4623 4624 4624 4625 4625 4626 4626 4627 4627 4628 4628 4629 4629 4630 4630 4631 4631 4632 4632 4633 4633 4634 22 4635 4635 4636 4 4637 11 4638 4638 4639 90 4640 11 4641 4641 4642 101 4643 4643 4644 1569 4645 4645 4646 4646 4647 4647 4648 101 4649 868 4650 11 4651 4651 4652 956 4653 4653 4654 4654 4655 989 4656 161 4657 164 4658 4658 4659 4659 4660 4660 4661 868 4662 11 4663 4663 4664 4664 4665 4665 4666 364 4667 364 4668 367 4669 441 4670 4670 4671 4671 4672 4672 4673 4673 4674 2 4675 583 4676 4676 4677 4677 4678 4678 4679 4679 4680 4680 4681 4681 4682 2 4683 583 4684 4684 4685 4685 4686 142 4687 11 4688 144 4689 164 4690 165 4691 167 4692 169 4693 77 4694 128 4695 172 4696 4696 4697 4697 4698 142 4699 11 4700 144 4701 164 4702 2270 4703 2272 4704 2273 4705 2274 4706 2 4707 4707 4708 4708 4709 868 4710 4710 4711 11 4712 4712 4713 2 4714 583 4715 4715 4716 4716 4717 4717 4718 2756 4719 4719 4720 4720 4721 364 4722 364 4723 367 4724 441 4725 4725 4726 4726 4727 4727 4728 4728 4729 2 4730 583 4731 4731 4732 4732 4733 4733 4734 4734 4735 2827 4736 4736 4737 441 4738 4738 4739 2 4740 4740 4741 4741 4742 2 4743 3 4744 4 4745 2 4746 2 4747 2 4748 2 4749 11 4750 2 4751 2 4752 4752 4753 2 4754 4754 4755 22 4756 4756 4757 4757 4758 2 4759 11 4760 11 4761 4761 4762 4762 4763 4763 4764 22 4765 2 4766 2 4767 2 4768 11 4769 4769 4770 11 4771 4771 4772 11 4773 4773 4774 4774 4775 4775 4776 4776 4777 60 4778 4778 4779 2 4780 2 4781 4781 4782 11 4783 2 4784 2 4785 69 4786 133 4787 4787 4788 2 4789 2 4790 4790 4791 2 4792 4792 4793 4793 4794 2 4795 86 4796 87 4797 90 4798 136 4799 11 4800 4800 4801 4801 4802 101 4803 4803 4804 4804 4805 4805 4806 106 4807 106 4808 4808 4809 4809 4810 4810 4811 4811 4812 4812 4813 4813 4814 4814 4815 4815 4816 4816 4817 4817 4818 4818 4819 101 4820 2 4821 4821 4822 2 4823 2 4824 2 4825 2 4826 2 4827 2 4828 2 4829 2 4830 312 4831 2 4832 2 4833 2 4834 2 4835 2 4836 318 4837 2 4838 2 4839 2 4840 2 4841 2 4842 2 4843 2 4844 2 4845 2 4846 2 4847 2 4848 2 4849 2 4850 2 4851 2 4852 2 4853 4853 4854 4854 4855 4855 4856 4856 4857 4857 4858 4858 4859 4859 4860 4860 4861 4861 4862 4862 4863 4863 4864 4864 4865 4865 4866 4866 4867 4867 4868 4868 4869 4 4870 2 4871 4871 4872 4872 4873 4873 4874 4874 4875 4875 4876 4876 4877 4877 4878 4878 4879 4879 4880 4880 4881 4881 4882 364 4883 364 4884 40 4885 367 4886 441 4887 4887 4888 4888 4889 4889 4890 4890 4891 4891 4892 2 4893 583 4894 4894 4895 4895 4896 40 4897 4897 4898 287 4899 4899 4900 4900 4901 2 4902 583 4903 4903 4904 4904 4905 4905 4906 4906 4907 4907 4908 4908 4909 2 4910 751 4911 2 4912 4912 4913 4913 4914 4914 4915 4915 4916 4916 4917 4917 4918 4918 4919 4919 4920 4920 4921 4921 4922 868 4923 4923 4924 4924 4925 4925 4926 4926 4927 4927 4928 4928 4929 4929 4930 4930 4931 4931 4932 4932 4933 4933 4934 4934 4935 4935 4936 4936 4937 4937 4938 4938 4939 4939 4940 4940 4941 4941 4942 4942 4943 4943 4944 4944 4945 4945 4946 4946 4947 4947 4948 4948 4949 4949 4950 4950 4951 4951 4952 4952 4953 4953 4954 4954 4955 4955 4956 4956 4957 4957 4958 4958 4959 4959 4960 4960 4961 4961 4962 4962 4963 4963 4964 4964 4965 4965 4966 4966 4967 4967 4968 4968 4969 4969 4970 4970 4971 4971 4972 4972 4973 4973 4974 4974 4975 4975 4976 4976 4977 4977 4978 4978 4979 4979 4980 4980 4981 4981 4982 4982 4983 4983 4984 4984 4985 4985 4986 4986 4987 4987 4988 4988 4989 4989 4990 4990 4991 4991 4992 4992 4993 4993 4994 4994 4995 4995 4996 4996 4997 4997 4998 4998 4999 4999 5000 5000 5001 5001 5002 5002 5003 11 5004 5004 5005 5005 5006 5006 5007 2 5008 583 5009 5009 5010 956 5011 5011 5012 5012 5013 989 5014 161 5015 164 5016 5016 5017 5017 5018 1505 5019 2 5020 2 5021 2 5022 5022 5023 2 5024 5024 5025 2 5026 2 5027 2 5028 2 5029 2 5030 2 5031 2 5032 2 5033 2 5034 5034 5035 5035 5036 5036 5037 5037 5038 5038 5039 5039 5040 5040 5041 5041 5042 5042 5043 5043 5044 5044 5045 5045 5046 5046 5047 22 5048 5048 5049 4 5050 11 5051 5051 5052 90 5053 11 5054 5054 5055 101 5056 5056 5057 1569 5058 5058 5059 5059 5060 5060 5061 101 5062 868 5063 11 5064 5064 5065 956 5066 5066 5067 5067 5068 989 5069 161 5070 164 5071 5071 5072 5072 5073 5073 5074 868 5075 11 5076 5076 5077 5077 5078 5078 5079 364 5080 364 5081 367 5082 441 5083 5083 5084 5084 5085 5085 5086 5086 5087 2 5088 583 5089 5089 5090 5090 5091 5091 5092 5092 5093 5093 5094 5094 5095 2 5096 583 5097 5097 5098 5098 5099 142 5100 11 5101 144 5102 164 5103 165 5104 167 5105 169 5106 77 5107 128 5108 172 5109 5109 5110 5110 5111 142 5112 11 5113 144 5114 164 5115 2270 5116 2272 5117 2273 5118 2274 5119 2 5120 5120 5121 5121 5122 2 5123 5123 5124 868 5125 5125 5126 11 5127 5127 5128 2 5129 583 5130 5130 5131 5131 5132 5132 5133 2756 5134 5134 5135 5135 5136 364 5137 364 5138 367 5139 441 5140 5140 5141 5141 5142 5142 5143 5143 5144 2 5145 583 5146 5146 5147 5147 5148 5148 5149 5149 5150 2827 5151 5151 5152 441 5153 5153 5154 2 5155 5155 5156 5156 5157 4805 5158 2 5159 5159 5160 5160 5161 5161 5162 5162 5163 5163 5164 2918 5165 5165 5166 69 5167 5167 5168 5168 5169 5169 5170 5170 5171 5171 5172 5172 5173 5173 5174 2 5175 5175 5176 5176 5177 5177 5178 5178 5179 5179 5180 5180 5181 2960 5182 2962 5183 5183 5184 5184 5185 5185 5186 5186 5187 5187 5188 364 5189 364 5190 40 5191 367 5192 441 5193 5193 5194 5194 5195 5195 5196 5196 5197 5197 5198 40 5199 5199 5200 5200 5201 2983 5202 2986 5203 2988 5204 2990 5205 77 5206 2992 5207 2993 5208 5208 5209 287 5210 5210 5211 2 5212 287 5213 77 5214 3959 5215 2852 5216 5216 5217 1882 5218 164 5219 2072 5220 2 5221 5221 5222 2 5223 5223 5224 5224 5225 161 5226 5226 5227 164 5228 5228 5229 5229 5230 5230 5231 5231 5232 5232 5233 5233 5234 956 5235 5235 5236 5236 5237 5237 5238 989 5239 161 5240 164 5241 5241 5242 5242 5243 2 5244 287 5245 5245 5246 2 5247 287 5248 5248 5249 5249 5250 5250 5251 956 5252 5252 5253 5253 5254 5254 5255 967 5256 976 5257 977 5258 984 5259 989 5260 992 5261 1000 5262 1001 5263 1003 5264 1010 5265 1016 5266 1017 5267 1022 5268 1023 5269 1024 5270 1029 5271 1030 5272 1035 5273 1036 5274 1038 5275 1040 5276 1043 5277 1044 5278 1045 5279 1048 5280 1049 5281 1052 5282 1054 5283 1062 5284 5284 5285 5285 5286 161 5287 1498 5288 164 5289 5289 5290 5290 5291 5291 5292 956 5293 5293 5294 5294 5295 2 5296 287 5297 77 5298 3959 5299 2852 5300 5300 5301 2 5302 161 5303 164 5304 5304 5305 5305 5306 5306 5307 956 5308 5308 5309 5309 5310 5310 5311 989 5312 161 5313 164 5314 5314 5315 5315 5316 2 5317 287 5318 5318 5319 2 5320 287 5321 5321 5322 5322 5323 956 5324 5324 5325 5325 5326 5326 5327 989 5328 5285 5329 161 5330 164 5331 5331 5332 5332 5333 956 5334 5334 5335 5335 5336 1882 5337 164 5338 5338 5339 5339 5340 5340 5341 1882 5342 164 5343 5343 5344 5344 5345 5345 5346 5346 5347 287 5348 5348 5349 2 5350 583 5351 5351 5352 22 5353 2824 5354 5354 5355 5355 5356 5356 5357 5357 5358 5358 5359 5359 5360 5360 5361 16 5362 16 5363 2835 5364 441 5365 5365 5366 5366 5367 5367 5368 5368 5369 2846 5370 583 5371 5371 5372 2 5373 5373 5374 2851 5375 2852 5376 5376 5377 2865 5378 5378 5379 5379 5380 5380 5381 5381 5382 5382 5383 5383 5384 2918 5385 5385 5386 69 5387 5387 5388 5388 5389 5389 5390 5390 5391 5391 5392 5392 5393 5393 5394 2 5395 5395 5396 5396 5397 5397 5398 5398 5399 5399 5400 5400 5401 2960 5402 2962 5403 5403 5404 5404 5405 5405 5406 5406 5407 5407 5408 364 5409 364 5410 40 5411 367 5412 441 5413 5413 5414 5414 5415 5415 5416 5416 5417 5417 5418 40 5419 5419 5420 5420 5421 2983 5422 2986 5423 2988 5424 2990 5425 77 5426 2992 5427 2993 5428 5428 5429 287 5430 5430 5431 2 5432 287 5433 77 5434 3959 5435 2852 5436 5436 5437 1882 5438 164 5439 2072 5440 2 5441 5441 5442 2 5443 161 5444 164 5445 5445 5446 5446 5447 5447 5448 5448 5449 5449 5450 956 5451 5451 5452 5452 5453 5453 5454 989 5455 161 5456 164 5457 5457 5458 5458 5459 2 5460 287 5461 5461 5462 2 5463 287 5464 5464 5465 5465 5466 5466 5467 956 5468 5468 5469 5469 5470 5470 5471 989 5472 5285 5473 161 5474 164 5475 5475 5476 5476 5477 956 5478 5478 5479 5479 5480 5480 5481 5481 5482 1882 5483 164 5484 5484 5485 5485 5486 5486 5487 5487 5488 287 5489 5489 5490 2 5491 583 5492 5492 5493 5356 5494 5494 5495 5495 5496 5496 5497 142 5498 11 5499 144 5500 164 5501 165 5502 167 5503 169 5504 77 5505 128 5506 172 5507 5507 5508 5508 5509 142 5510 11 5511 144 5512 164 5513 2270 5514 2272 5515 2273 5516 2274 5517 2 5518 5518 5519 5519 5520 868 5521 5521 5522 11 5523 5523 5524 2 5525 583 5526 5526 5527 5527 5528 5528 5529 2756 5530 5530 5531 5531 5532 364 5533 364 5534 367 5535 441 5536 5536 5537 5537 5538 5538 5539 5539 5540 2 5541 583 5542 5542 5543 5543 5544 5544 5545 5545 5546 2827 5547 5547 5548 441 5549 5549 5550 2 5551 5551 5552 5552 5553 5553 5554 2 5555 5555 5556 5556 5557 5557 5558 5558 5559 5559 5560 2918 5561 5561 5562 69 5563 5563 5564 5564 5565 5565 5566 5566 5567 5567 5568 5568 5569 5569 5570 2 5571 5571 5572 5572 5573 5573 5574 5574 5575 5575 5576 5576 5577 2960 5578 2962 5579 5579 5580 5580 5581 5581 5582 5582 5583 5583 5584 364 5585 364 5586 40 5587 367 5588 441 5589 5589 5590 5590 5591 5591 5592 5592 5593 5593 5594 40 5595 5595 5596 5596 5597 2983 5598 2986 5599 2988 5600 2990 5601 77 5602 2992 5603 2993 5604 5604 5605 287 5606 5606 5607 2 5608 287 5609 77 5610 3959 5611 2852 5612 5612 5613 1882 5614 164 5615 2072 5616 2 5617 5617 5618 2 5619 161 5620 164 5621 5621 5622 5622 5623 5623 5624 5624 5625 5625 5626 956 5627 5627 5628 5628 5629 5629 5630 989 5631 161 5632 164 5633 5633 5634 5634 5635 2 5636 287 5637 5637 5638 2 5639 287 5640 5640 5641 5641 5642 5642 5643 2272 5644 2 5645 5645 5646 5646 5647 956 5648 5648 5649 5649 5650 5650 5651 989 5652 5285 5653 161 5654 164 5655 5655 5656 5656 5657 956 5658 5658 5659 5659 5660 5660 5661 5661 5662 1882 5663 164 5664 5664 5665 5665 5666 5666 5667 5667 5668 287 5669 5669 5670 2 5671 583 5672 5672 5673 5356 5674 5674 5675 5675 5676 1882 5677 164 5678 2072 5679 2 5680 5680 5681 2 5682 161 5683 164 5684 5684 5685 5685 5686 5686 5687 5687 5688 5688 5689 956 5690 5690 5691 5691 5692 5692 5693 989 5694 161 5695 164 5696 5696 5697 5697 5698 2 5699 287 5700 5700 5701 2 5702 287 5703 5703 5704 5704 5705 5705 5706 956 5707 5707 5708 5708 5709 5709 5710 989 5711 5285 5712 161 5713 164 5714 5714 5715 5715 5716 956 5717 5717 5718 5718 5719 5719 5720 5720 5721 1882 5722 164 5723 5723 5724 5724 5725 5725 5726 5726 5727 287 5728 5728 5729 2 5730 583 5731 5731 5732 5356 5733 5733 5734 5734 5735 5735 5736 5736 5737 5737 5738 5738 5739 2918 5740 5740 5741 69 5742 5742 5743 5743 5744 5744 5745 5745 5746 5746 5747 5747 5748 5748 5749 2 5750 5750 5751 5751 5752 5752 5753 5753 5754 5754 5755 5755 5756 2960 5757 2962 5758 5758 5759 5759 5760 5760 5761 5761 5762 5762 5763 364 5764 364 5765 40 5766 367 5767 441 5768 5768 5769 5769 5770 5770 5771 5771 5772 5772 5773 40 5774 5774 5775 5775 5776 2983 5777 2986 5778 2988 5779 2990 5780 77 5781 2992 5782 2993 5783 5783 5784 287 5785 5785 5786 2 5787 287 5788 77 5789 3959 5790 2852 5791 5791 5792 1882 5793 164 5794 2072 5795 2 5796 5796 5797 2 5798 161 5799 164 5800 5800 5801 5801 5802 5802 5803 5803 5804 5804 5805 956 5806 5806 5807 5807 5808 5808 5809 989 5810 161 5811 164 5812 5812 5813 5813 5814 2 5815 287 5816 5816 5817 2 5818 287 5819 5819 5820 5820 5821 5821 5822 956 5823 5823 5824 5824 5825 5825 5826 989 5827 5285 5828 161 5829 164 5830 5830 5831 5831 5832 956 5833 5833 5834 5834 5835 5835 5836 5836 5837 1882 5838 164 5839 5839 5840 5840 5841 5841 5842 5842 5843 287 5844 5844 5845 2 5846 583 5847 5847 5848 5356 5849 5849 5850 5850 5851 2 5852 5852 5853 5853 5854 5854 5855 5855 5856 13 5857 5857 5858 5858 5859 5859 5860 5860 5861 5861 5862 5862 5863 5863 5864 5864 5865 5865 5866 5866 5867 5867 5868 5868 5869 5869 5870 5870 5871 5871 5872 5872 5873 5873 5874 5874 5875 5875 5876 16 5877 5877 5878 5878 5879 144 5880 5880 5881 5881 5882 5882 5883 2 5884 5884 5885 128 5886 172 5887 5887 5888 5888 5889 105 5890 2 5891 2 5892 2 5893 2 5894 2 5895 2 5896 5896 5897 2 5898 2 5899 312 5900 2 5901 2 5902 2 5903 2 5904 2 5905 318 5906 2 5907 2 5908 2 5909 2 5910 2 5911 2 5912 2 5913 2 5914 2 5915 2 5916 2 5917 2 5918 2 5919 2 5920 2 5921 2 5922 5922 5923 5923 5924 5924 5925 5925 5926 5926 5927 5927 5928 5928 5929 5929 5930 5930 5931 5931 5932 5932 5933 5933 5934 5934 5935 5935 5936 5936 5937 5937 5938 4 5939 2 5940 5940 5941 5941 5942 5942 5943 5943 5944 5944 5945 5945 5946 5946 5947 5947 5948 5948 5949 5949 5950 5950 5951 364 5952 364 5953 40 5954 367 5955 441 5956 5956 5957 5957 5958 5958 5959 5959 5960 5960 5961 2 5962 583 5963 5963 5964 5964 5965 40 5966 5966 5967 287 5968 5968 5969 5969 5970 2 5971 583 5972 5972 5973 5973 5974 5974 5975 5975 5976 5976 5977 5977 5978 2 5979 751 5980 2 5981 5981 5982 5982 5983 5983 5984 5984 5985 5985 5986 5986 5987 5987 5988 5988 5989 5989 5990 5990 5991 164 5992 5992 5993 5993 5994 5994 5995 5995 5996 5996 5997 5997 5998 5998 5999 5999 6000 6000 6001 6001 6002 6002 6003 6003 6004 6004 6005 956 6006 6006 6007 6007 6008 989 6009 161 6010 164 6011 6011 6012 6012 6013 6013 6014 441 6015 6015 6016 6016 6017 2 6018 3 6019 4 6020 2 6021 2 6022 2 6023 2 6024 11 6025 2 6026 2 6027 6027 6028 2 6029 6029 6030 22 6031 6031 6032 6032 6033 2 6034 11 6035 11 6036 6036 6037 6037 6038 6038 6039 22 6040 2 6041 2 6042 2 6043 11 6044 6044 6045 11 6046 6046 6047 11 6048 6048 6049 6049 6050 6050 6051 6051 6052 60 6053 6053 6054 2 6055 2 6056 6056 6057 11 6058 2 6059 2 6060 69 6061 133 6062 6062 6063 2 6064 2 6065 6065 6066 2 6067 6067 6068 6068 6069 2 6070 86 6071 87 6072 90 6073 136 6074 11 6075 6075 6076 6076 6077 101 6078 6078 6079 6079 6080 6080 6081 106 6082 106 6083 6083 6084 6084 6085 6085 6086 6086 6087 6087 6088 6088 6089 6089 6090 6090 6091 6091 6092 6092 6093 6093 6094 101 6095 2 6096 6096 6097 2 6098 2 6099 2 6100 2 6101 2 6102 2 6103 2 6104 2 6105 312 6106 2 6107 2 6108 2 6109 2 6110 2 6111 318 6112 2 6113 2 6114 2 6115 2 6116 2 6117 2 6118 2 6119 2 6120 2 6121 2 6122 2 6123 2 6124 2 6125 2 6126 2 6127 2 6128 6128 6129 6129 6130 6130 6131 6131 6132 6132 6133 6133 6134 6134 6135 6135 6136 6136 6137 6137 6138 6138 6139 6139 6140 6140 6141 6141 6142 6142 6143 6143 6144 4 6145 2 6146 6146 6147 6147 6148 6148 6149 6149 6150 6150 6151 6151 6152 6152 6153 6153 6154 6154 6155 6155 6156 6156 6157 364 6158 364 6159 40 6160 367 6161 441 6162 6162 6163 6163 6164 6164 6165 6165 6166 6166 6167 2 6168 583 6169 6169 6170 6170 6171 40 6172 6172 6173 287 6174 6174 6175 6175 6176 2 6177 583 6178 6178 6179 6179 6180 6180 6181 6181 6182 6182 6183 6183 6184 2 6185 751 6186 2 6187 6187 6188 6188 6189 6189 6190 6190 6191 6191 6192 6192 6193 6193 6194 6194 6195 6195 6196 6196 6197 868 6198 6198 6199 6199 6200 6200 6201 6201 6202 6202 6203 6203 6204 6204 6205 6205 6206 6206 6207 6207 6208 6208 6209 6209 6210 6210 6211 6211 6212 6212 6213 6213 6214 6214 6215 6215 6216 6216 6217 6217 6218 6218 6219 6219 6220 6220 6221 6221 6222 6222 6223 6223 6224 6224 6225 6225 6226 6226 6227 6227 6228 6228 6229 6229 6230 6230 6231 6231 6232 6232 6233 6233 6234 6234 6235 6235 6236 6236 6237 6237 6238 6238 6239 6239 6240 6240 6241 6241 6242 6242 6243 6243 6244 6244 6245 6245 6246 6246 6247 6247 6248 6248 6249 6249 6250 6250 6251 6251 6252 6252 6253 6253 6254 6254 6255 6255 6256 6256 6257 6257 6258 6258 6259 6259 6260 6260 6261 6261 6262 6262 6263 6263 6264 6264 6265 6265 6266 6266 6267 6267 6268 6268 6269 6269 6270 6270 6271 6271 6272 6272 6273 6273 6274 6274 6275 6275 6276 6276 6277 6277 6278 11 6279 6279 6280 6280 6281 6281 6282 2 6283 583 6284 6284 6285 956 6286 6286 6287 6287 6288 989 6289 161 6290 164 6291 6291 6292 6292 6293 1505 6294 2 6295 2 6296 2 6297 6297 6298 2 6299 6299 6300 2 6301 2 6302 2 6303 2 6304 2 6305 2 6306 2 6307 2 6308 2 6309 6309 6310 6310 6311 6311 6312 6312 6313 6313 6314 6314 6315 6315 6316 6316 6317 6317 6318 6318 6319 6319 6320 6320 6321 6321 6322 22 6323 6323 6324 4 6325 11 6326 6326 6327 90 6328 11 6329 6329 6330 101 6331 6331 6332 1569 6333 6333 6334 6334 6335 6335 6336 101 6337 868 6338 11 6339 6339 6340 956 6341 6341 6342 6342 6343 989 6344 161 6345 164 6346 6346 6347 6347 6348 6348 6349 868 6350 11 6351 6351 6352 6352 6353 6353 6354 364 6355 364 6356 367 6357 441 6358 6358 6359 6359 6360 6360 6361 6361 6362 2 6363 583 6364 6364 6365 6365 6366 6366 6367 6367 6368 6368 6369 6369 6370 2 6371 583 6372 6372 6373 6373 6374 142 6375 11 6376 144 6377 164 6378 165 6379 167 6380 169 6381 77 6382 128 6383 172 6384 6384 6385 6385 6386 142 6387 11 6388 144 6389 164 6390 2270 6391 2272 6392 2273 6393 2274 6394 2 6395 6395 6396 6396 6397 6397 6398 5853 6399 5854 6400 6400 6401 868 6402 6402 6403 11 6404 6404 6405 2 6406 583 6407 6407 6408 6408 6409 6409 6410 2756 6411 6411 6412 6412 6413 364 6414 364 6415 367 6416 441 6417 6417 6418 6418 6419 6419 6420 6420 6421 2 6422 583 6423 6423 6424 6424 6425 6425 6426 6426 6427 2827 6428 6428 6429 441 6430 6430 6431 2 6432 6432 6433 6433 6434 2 6435 3 6436 4 6437 2 6438 2 6439 2 6440 2 6441 11 6442 2 6443 2 6444 6444 6445 2 6446 6446 6447 22 6448 6448 6449 6449 6450 2 6451 11 6452 11 6453 6453 6454 6454 6455 6455 6456 22 6457 2 6458 2 6459 2 6460 11 6461 6461 6462 11 6463 6463 6464 11 6465 6465 6466 6466 6467 6467 6468 6468 6469 60 6470 6470 6471 2 6472 2 6473 6473 6474 11 6475 2 6476 2 6477 69 6478 133 6479 6479 6480 2 6481 2 6482 6482 6483 2 6484 6484 6485 6485 6486 2 6487 86 6488 87 6489 90 6490 136 6491 11 6492 6492 6493 6493 6494 101 6495 6495 6496 6496 6497 6497 6498 106 6499 106 6500 6500 6501 6501 6502 6502 6503 6503 6504 6504 6505 6505 6506 6506 6507 6507 6508 6508 6509 6509 6510 6510 6511 101 6512 2 6513 6513 6514 2 6515 2 6516 2 6517 2 6518 2 6519 2 6520 2 6521 2 6522 312 6523 2 6524 2 6525 2 6526 2 6527 2 6528 318 6529 2 6530 2 6531 2 6532 2 6533 2 6534 2 6535 2 6536 2 6537 2 6538 2 6539 2 6540 2 6541 2 6542 2 6543 2 6544 2 6545 6545 6546 6546 6547 6547 6548 6548 6549 6549 6550 6550 6551 6551 6552 6552 6553 6553 6554 6554 6555 6555 6556 6556 6557 6557 6558 6558 6559 6559 6560 6560 6561 4 6562 2 6563 6563 6564 6564 6565 6565 6566 6566 6567 6567 6568 6568 6569 6569 6570 6570 6571 6571 6572 6572 6573 6573 6574 364 6575 364 6576 40 6577 367 6578 441 6579 6579 6580 6580 6581 6581 6582 6582 6583 6583 6584 2 6585 583 6586 6586 6587 6587 6588 40 6589 6589 6590 287 6591 6591 6592 6592 6593 2 6594 583 6595 6595 6596 6596 6597 6597 6598 6598 6599 6599 6600 6600 6601 2 6602 751 6603 2 6604 6604 6605 6605 6606 6606 6607 6607 6608 6608 6609 6609 6610 6610 6611 6611 6612 6612 6613 6613 6614 868 6615 6615 6616 6616 6617 6617 6618 6618 6619 6619 6620 6620 6621 6621 6622 6622 6623 6623 6624 6624 6625 6625 6626 6626 6627 6627 6628 6628 6629 6629 6630 6630 6631 6631 6632 6632 6633 6633 6634 6634 6635 6635 6636 6636 6637 6637 6638 6638 6639 6639 6640 6640 6641 6641 6642 6642 6643 6643 6644 6644 6645 6645 6646 6646 6647 6647 6648 6648 6649 6649 6650 6650 6651 6651 6652 6652 6653 6653 6654 6654 6655 6655 6656 6656 6657 6657 6658 6658 6659 6659 6660 6660 6661 6661 6662 6662 6663 6663 6664 6664 6665 6665 6666 6666 6667 6667 6668 6668 6669 6669 6670 6670 6671 6671 6672 6672 6673 6673 6674 6674 6675 6675 6676 6676 6677 6677 6678 6678 6679 6679 6680 6680 6681 6681 6682 6682 6683 6683 6684 6684 6685 6685 6686 6686 6687 6687 6688 6688 6689 6689 6690 6690 6691 6691 6692 6692 6693 6693 6694 6694 6695 11 6696 6696 6697 6697 6698 6698 6699 2 6700 583 6701 6701 6702 956 6703 6703 6704 6704 6705 989 6706 161 6707 164 6708 6708 6709 6709 6710 1505 6711 2 6712 2 6713 2 6714 6714 6715 2 6716 6716 6717 2 6718 2 6719 2 6720 2 6721 2 6722 2 6723 2 6724 2 6725 2 6726 6726 6727 6727 6728 6728 6729 6729 6730 6730 6731 6731 6732 6732 6733 6733 6734 6734 6735 6735 6736 6736 6737 6737 6738 6738 6739 22 6740 6740 6741 4 6742 11 6743 6743 6744 90 6745 11 6746 6746 6747 101 6748 6748 6749 1569 6750 6750 6751 6751 6752 6752 6753 101 6754 868 6755 11 6756 6756 6757 956 6758 6758 6759 6759 6760 989 6761 161 6762 164 6763 6763 6764 6764 6765 6765 6766 868 6767 11 6768 6768 6769 6769 6770 6770 6771 364 6772 364 6773 367 6774 441 6775 6775 6776 6776 6777 6777 6778 6778 6779 2 6780 583 6781 6781 6782 6782 6783 6783 6784 6784 6785 6785 6786 6786 6787 2 6788 583 6789 6789 6790 6790 6791 142 6792 11 6793 144 6794 164 6795 165 6796 167 6797 169 6798 77 6799 128 6800 172 6801 6801 6802 6802 6803 142 6804 11 6805 144 6806 164 6807 2270 6808 2272 6809 2273 6810 2274 6811 2 6812 6812 6813 6813 6814 2 6815 6815 6816 868 6817 6817 6818 11 6819 6819 6820 2 6821 583 6822 6822 6823 6823 6824 6824 6825 2756 6826 6826 6827 6827 6828 364 6829 364 6830 367 6831 441 6832 6832 6833 6833 6834 6834 6835 6835 6836 2 6837 583 6838 6838 6839 6839 6840 6840 6841 6841 6842 2827 6843 6843 6844 441 6845 6845 6846 2 6847 6847 6848 6848 6849 6497 6850 2 6851 6851 6852 6852 6853 6853 6854 6854 6855 6855 6856 2918 6857 6857 6858 69 6859 6859 6860 6860 6861 6861 6862 6862 6863 6863 6864 6864 6865 6865 6866 2 6867 6867 6868 6868 6869 6869 6870 6870 6871 6871 6872 6872 6873 2960 6874 2962 6875 6875 6876 6876 6877 6877 6878 6878 6879 6879 6880 364 6881 364 6882 40 6883 367 6884 441 6885 6885 6886 6886 6887 6887 6888 6888 6889 6889 6890 40 6891 6891 6892 6892 6893 2983 6894 2986 6895 2988 6896 2990 6897 77 6898 2992 6899 2993 6900 6900 6901 287 6902 6902 6903 2 6904 287 6905 77 6906 3959 6907 2852 6908 6908 6909 1882 6910 164 6911 2072 6912 2 6913 6913 6914 2 6915 161 6916 164 6917 6917 6918 6918 6919 6919 6920 6920 6921 6921 6922 956 6923 6923 6924 6924 6925 6925 6926 989 6927 161 6928 164 6929 6929 6930 6930 6931 2 6932 287 6933 6933 6934 2 6935 287 6936 6936 6937 6937 6938 6938 6939 956 6940 6940 6941 6941 6942 6942 6943 989 6944 5285 6945 161 6946 164 6947 6947 6948 6948 6949 956 6950 6950 6951 6951 6952 6952 6953 6953 6954 1882 6955 164 6956 6956 6957 6957 6958 6958 6959 6959 6960 287 6961 6961 6962 2 6963 583 6964 6964 6965 5356 6966 6966 6967 6967 6968 6968 6969 6969 6970 6970 6971 6971 6972 2918 6973 6973 6974 69 6975 6975 6976 6976 6977 6977 6978 6978 6979 6979 6980 6980 6981 6981 6982 2 6983 6983 6984 6984 6985 6985 6986 6986 6987 6987 6988 6988 6989 2960 6990 2962 6991 6991 6992 6992 6993 6993 6994 6994 6995 6995 6996 364 6997 364 6998 40 6999 367 7000 441 7001 7001 7002 7002 7003 7003 7004 7004 7005 7005 7006 40 7007 7007 7008 7008 7009 2983 7010 2986 7011 2988 7012 2990 7013 77 7014 2992 7015 2993 7016 7016 7017 287 7018 7018 7019 2 7020 287 7021 77 7022 3959 7023 2852 7024 7024 7025 1882 7026 164 7027 2072 7028 2 7029 7029 7030 2 7031 161 7032 164 7033 7033 7034 7034 7035 7035 7036 7036 7037 7037 7038 956 7039 7039 7040 7040 7041 7041 7042 989 7043 161 7044 164 7045 7045 7046 7046 7047 2 7048 287 7049 7049 7050 2 7051 287 7052 7052 7053 7053 7054 7054 7055 956 7056 7056 7057 7057 7058 7058 7059 989 7060 5285 7061 161 7062 164 7063 7063 7064 7064 7065 956 7066 7066 7067 7067 7068 7068 7069 7069 7070 1882 7071 164 7072 7072 7073 7073 7074 7074 7075 7075 7076 287 7077 7077 7078 2 7079 583 7080 7080 7081 5356 7082 7082 7083 7083 7084 3019 7085 7085 7086 7086 7087 7087 7088 7088 7089 7089 7090 7090 7091 7091 7092 7092 7093 7093 7094 2314 7095 7095 7096 7096 7097 7097 7098 7098 7099 7099 7100 7100 7101 7101 7102 7102 7103 7103 7104 7104 7105 7105 7106 7106 7107 7107 7108 7108 7109 7109 7110 7110 7111 7111 7112 7112 7113 7113 7114 7114 7115 7115 7116 7116 7117 7117 7118 7118 7119 7119 7120 7120 7121 7121 7122 7122 7123 7123 7124 7124 7125 7125 7126 7126 7127 7127 7128 7128 7129 7129 7130 7130 7131 7131 7132 7132 7133 7133 7134 7134 7135 7135 7136 7136 7137 7137 7138 7138 7139 7139 7140 7140 7141 7141 7142 7142 7143 7143 7144 7144 7145 7145 7146 7146 7147 7147 7148 7148 7149 7149 7150 7150 7151 7151 7152 7152 7153 7153 7154 7154 7155 7155 7156 7156 7157 7157 7158 7158 7159 7159 7160 7160 7161 7161 7162 7162 7163 7163 7164 7164 7165 7165 7166 7166 7167 7167 7168 7168 7169 7169 7170 7170 7171 7171 7172 7172 7173 7173 7174 7174 7175 7175 7176 7176 7177 7177 7178 7178 7179 7179 7180 7180 7181 7181 7182 142 7183 7183 7184 7184 7185 11 7186 144 7187 164 7188 7188 7189 168 7190 7190 7191 7191 7192 7192 7193 7193 7194 128 7195 172 7196 7196 7197 7197 7198 2278 7199 7199 7200 7200 7201 7201 7202 7202 7203 7203 7204 7204 7205 7205 7206 7206 7207 7207 7208 7208 7209 7209 7210 7210 7211 7211 7212 2293 7213 2294 7214 2278 7215 7215 7216 7216 7217 7217 7218 7218 7219 7219 7220 7220 7221 7221 7222 7222 7223 7223 7224 7224 7225 7225 7226 2314 7227 7227 7228 2294 7229 7229 7230 2278 7231 7231 7232 7232 7233 7233 7234 7234 7235 7235 7236 7236 7237 7237 7238 7238 7239 7239 7240 7240 7241 2278 7242 7242 7243 7243 7244 7244 7245 7245 7246 7246 7247 7247 7248 7248 7249 7249 7250 7250 7251 7251 7252 7252 7253 7253 7254 2278 7255 7255 7256 7256 7257 7257 7258 7258 7259 7259 7260 7260 7261 7261 7262 7262 7263 7263 7264 7264 7265 7265 7266 7266 7267 7267 7268 7268 7269 7269 7270 7270 7271 7271 7272 7272 7273 7273 7274 7274 7275 7275 7276 7276 7277 7277 7278 7278 7279 7279 7280 7280 7281 7281 7282 7282 7283 7283 7284 7284 7285 7285 7286 7286 7287 7287 7288 7288 7289 7289 7290 7290 7291 7291 7292 7292 7293 7293 7294 7294 7295 7295 7296 7296 7297 7297 7298 7298 7299 7299 7300 7300 7301 7301 7302 7302 7303 7303 7304 7304 7305 7305 7306 7306 7307 7307 7308 7308 7309 7309 7310 7310 7311 7311 7312 7312 7313 7313 7314 7314 7315 7315 7316 7316 7317 7317 7318 7318 7319 7319 7320 7320 7321 7321 7322 7322 7323 7323 7324 7324 7325 7325 7326 7326 7327 7327 7328 7328 7329 7329 7330 7330 7331 7331 7332 7332 7333 7333 7334 7334 7335 7335 7336 7336 7337 7337 7338 3066 7339 7339 7340 7340 7341 7341 7342 2072 7343 7343 7344 7102 7345 7345 7346 7346 7347 2293 7348 2294 7349 2278 7350 7350 7351 7351 7352 7352 7353 7353 7354 7354 7355 7355 7356 7356 7357 7357 7358 7358 7359 7359 7360 7360 7361 2314 7362 7362 7363 2294 7364 7364 7365 7365 7366 7366 7367 7367 7368 2072 7369 3066 7370 7102 7371 7371 7372 7372 7373 2 7374 3136 7375 7375 7376 2278 7377 7377 7378 7378 7379 7379 7380 7380 7381 7381 7382 7382 7383 7383 7384 7384 7385 7385 7386 7386 7387 2 7388 3136 7389 7389 7390 2278 7391 7391 7392 7392 7393 7393 7394 7394 7395 7395 7396 7396 7397 7397 7398 7398 7399 7399 7400 7400 7401 7401 7402 7402 7403 7403 7404 7404 7405 7405 7406 7406 7407 7407 7408 7408 7409 7409 7410 7410 7411 7411 7412 7412 7413 13 7414 7414 7415 7415 7416 7416 7417 7417 7418 7418 7419 7419 7420 7420 7421 7421 7422 7422 7423 7423 7424 7424 7425 7425 7426 7426 7427 7427 7428 7428 7429 7429 7430 7430 7431 7431 7432 7432 7433 7433 7434 7434 7435 7435 7436 7436 7437 7437 7438 7438 7439 7439 7440 7440 7441 7441 7442 7442 7443 7443 7444 7444 7445 7445 7446 7446 7447 7447 7448 7448 7449 7449 7450 7450 7451 7451 7452 7452 7453 7453 7454 7454 7455 7455 7456 7456 7457 7457 7458 7458 7459 7459 7460 7460 7461 7461 7462 7462 7463 7463 7464 7464 7465 7465 7466 7466 7467 7467 7468 7468 7469 7469 7470 7470 7471 7471 7472 7472 7473 7473 7474 7474 7475 7475 7476 7476 7477 7477 7478 7478 7479 7479 7480 7480 7481 7481 7482 7482 7483 7483 7484 7484 7485 7485 7486 7486 7487 7487 7488 7488 7489 7489 7490 7490 7491 7491 7492 7492 7493 7493 7494 7494 7495 7495 7496 7496 7497 7497 7498 7498 7499 7499 7500 101 7501 7501 7502 7502 7503 7503 7504 7504 7505 7505 7506 7416 7507 7507 7508 7508 7509 7507 7510 7510 7511 7511 7512 7512 7513 7513 7514 7514 7515 7515 7516 7516 7517 7517 7518 7518 7519 7519 7520 7520 7521 7521 7522 7522 7523 7523 7524 7524 7525 7525 7526 7526 7527 7527 7528 7528 7529 7529 7530 7530 7531 7531 7532 7532 7533 7533 7534 7534 7535 7535 7536 7536 7537 7537 7538 7538 7539 7539 7540 7540 7541 7541 7542 7542 7543 7543 7544 7544 7545 7545 7546 7546 7547 7547 7548 7548 7549 7549 7550 7550 7551 7551 7552 7552 7553 7553 7554 7554 7555 7555 7556 7556 7557 7557 7558 7558 7559 7559 7560 7560 7561 7561 7562 7562 7563 7563 7564 7564 7565 7565 7566 7566 7567 7567 7568 7568 7569 7569 7570 7570 7571 7571 7572 7572 7573 7573 7574 7574 7575 7575 7576 7576 7577 7577 7578 7578 7579 7579 7580 7580 7581 7581 7582 7582 7583 7583 7584 7584 7585 7585 7586 7586 7587 7587 7588 7588 7589 7589 7590 7590 7591 7591 7592 7592 7593 7593 7594 7594 7595 7595 7596 7596 7597 7597 7598 7598 7599 7599 7600 7600 7601 7601 7602 7602 7603 7603 7604 7604 7605 7605 7606 7606 7607 7607 7608 7608 7609 7609 7610 7610 7611 7102 7612 7612 7613 7613 7614 7614 7615 7615 7616 7616 7617 7617 7618 7618 7619 7619 7620 7620 7621 7621 7622 7622 7623 7623 7624 7624 7625 7625 7626 7626 7627 7627 7628 7628 7629 7629 7630 7630 7631 7631 7632 7632 7633 7633 7634 7634 7635 7635 7636 7636 7637 7637 7638 7638 7639 7639 7640 7640 7641 7641 7642 7642 7643 7643 7644 7644 7645 7645 7646 7646 7647 7647 7648 7648 7649 7649 7650 7650 7651 7651 7652 7652 7653 7653 7654 7654 7655 7655 7656 7656 7657 7657 7658 7658 7659 7659 7660 7660 7661 7661 7662 7662 7663 7663 7664 7664 7665 7665 7666 7666 7667 7667 7668 7668 7669 7669 7670 7670 7671 7671 7672 7672 7673 7673 7674 7674 7675 7675 7676 7676 7677 7677 7678 7678 7679 7679 7680 7680 7681 7681 7682 7682 7683 7683 7684 7684 7685 7685 7686 7686 7687 7687 7688 7688 7689 7689 7690 7690 7691 7691 7692 7692 7693 7693 7694 13 7695 7695 7696 7696 7697 7697 7698 7698 7699 7699 7700 7700 7701 7701 7702 7702 7703 13 7704 7704 7705 7705 7706 7102 7707 7707 7708 7708 7709 7709 7710 7710 7711 7711 7712 7712 7713 7713 7714 7714 7715 7715 7716 7716 7717 7717 7718 7718 7719 7719 7720 7720 7721 7721 7722 7722 7723 7723 7724 7724 7725 7725 7726 7726 7727 7727 7728 7728 7729 7729 7730 7730 7731 7731 7732 7732 7733 7733 7734 7734 7735 7735 7736 7736 7737 7737 7738 7738 7739 7739 7740 7740 7741 7741 7742 7742 7743 7743 7744 7744 7745 7745 7746 7746 7747 7747 7748 7748 7749 7749 7750 7750 7751 7751 7752 7752 7753 7753 7754 7754 7755 7755 7756 7756 7757 7757 7758 7758 7759 7759 7760 7760 7761 7761 7762 7762 7763 7763 7764 7764 7765 7765 7766 7766 7767 7767 7768 7768 7769 7769 7770 7770 7771 7771 7772 7772 7773 7773 7774 7774 7775 7775 7776 7776 7777 7777 7778 7778 7779 7779 7780 7780 7781 7781 7782 7782 7783 7783 7784 7784 7785 7785 7786 7786 7787 7787 7788 7788 7789 2 7790 7507 7791 7791 7792 7792 7793 6002 7794 441 7795 7795 7796 7796 7797 7797 7798 7798 7799 7799 7800 7800 7801 7801 7802 7802 7803 7803 7804 13 7805 7805 7806 7806 7807 7807 7808 7808 7809 7809 7810 7810 7811 7811 7812 7812 7813 7813 7814 989 7815 161 7816 164 7817 7817 7818 7818 7819 7819 7820 6002 7821 441 7822 7822 7823 287 7824 7824 7825 1882 7826 164 7827 7827 7828 7828 7829 7829 7830 7830 7831 7831 7832 7832 7833 7833 7834 7834 7835 7835 7836 7836 7837 287 7838 7838 7839 2 7840 583 7841 7841 7842 5356 7843 7843 7844 7844 7845 868 7846 7846 7847 11 7848 7848 7849 2 7850 583 7851 7851 7852 7852 7853 7853 7854 2756 7855 7855 7856 7856 7857 2072 7858 2 7859 7859 7860 364 7861 364 7862 367 7863 441 7864 7864 7865 7865 7866 7866 7867 7867 7868 2 7869 583 7870 7870 7871 7871 7872 7872 7873 7873 7874 7874 7875 2827 7876 7876 7877 441 7878 7878 7879 2 7880 7880 7881 7881 7882 1288 7883 2 7884 7884 7885 7885 7886 7886 7887 7887 7888 7888 7889 2918 7890 7890 7891 69 7892 7892 7893 7893 7894 7894 7895 7895 7896 7896 7897 7897 7898 7898 7899 2 7900 7900 7901 7901 7902 7902 7903 7903 7904 7904 7905 7905 7906 2960 7907 2962 7908 7908 7909 7909 7910 7910 7911 7911 7912 7912 7913 364 7914 364 7915 40 7916 367 7917 441 7918 7918 7919 7919 7920 7920 7921 7921 7922 7922 7923 40 7924 7924 7925 7925 7926 2983 7927 2986 7928 2988 7929 2990 7930 77 7931 2992 7932 2993 7933 7933 7934 287 7935 7935 7936 2 7937 287 7938 77 7939 3959 7940 2852 7941 7941 7942 1882 7943 164 7944 7944 7945 2 7946 161 7947 164 7948 7948 7949 7949 7950 7950 7951 7951 7952 7952 7953 956 7954 7954 7955 7955 7956 7956 7957 989 7958 161 7959 164 7960 7960 7961 7961 7962 2 7963 287 7964 7964 7965 2 7966 287 7967 7967 7968 7968 7969 7969 7970 7970 7971 956 7972 7972 7973 7973 7974 7974 7975 989 7976 5285 7977 161 7978 164 7979 7979 7980 7980 7981 956 7982 7982 7983 7983 7984 7984 7985 7985 7986 1882 7987 164 7988 7988 7989 7989 7990 7990 7991 7991 7992 287 7993 7993 7994 2 7995 583 7996 7996 7997 5356 7998 7998 7999 7999 8000 868 8001 8001 8002 8002 8003 8003 8004 8004 8005 8005 8006 8006 8007 8007 8008 8008 8009 8009 8010 8010 8011 8011 8012 8012 8013 8013 8014 8014 8015 8015 8016 8016 8017 8017 8018 8018 8019 8019 8020 8020 8021 8021 8022 8022 8023 8023 8024 8024 8025 8025 8026 8026 8027 8027 8028 8028 8029 8029 8030 8030 8031 8031 8032 8032 8033 8033 8034 8034 8035 8035 8036 8036 8037 8037 8038 8038 8039 8039 8040 8040 8041 8041 8042 8042 8043 8043 8044 8044 8045 8045 8046 8046 8047 8047 8048 8048 8049 8049 8050 8050 8051 8051 8052 8052 8053 8053 8054 8054 8055 8055 8056 8056 8057 8057 8058 8058 8059 8059 8060 8060 8061 8061 8062 8062 8063 8063 8064 8064 8065 8065 8066 8066 8067 8067 8068 8068 8069 8069 8070 8070 8071 8071 8072 8072 8073 8073 8074 8074 8075 8075 8076 8076 8077 8077 8078 8078 8079 8079 8080 8080 8081 11 8082 8082 8083 8083 8084 8084 8085 2 8086 583 8087 8087 8088 956 8089 8089 8090 8090 8091 989 8092 161 8093 164 8094 8094 8095 8095 8096 1505 8097 2 8098 2 8099 2 8100 8100 8101 2 8102 8102 8103 2 8104 2 8105 2 8106 2 8107 2 8108 2 8109 2 8110 2 8111 2 8112 8112 8113 8113 8114 8114 8115 8115 8116 8116 8117 8117 8118 8118 8119 8119 8120 8120 8121 8121 8122 8122 8123 8123 8124 8124 8125 22 8126 8126 8127 4 8128 11 8129 8129 8130 90 8131 11 8132 8132 8133 101 8134 8134 8135 1569 8136 8136 8137 8137 8138 8138 8139 101 8140 868 8141 11 8142 8142 8143 956 8144 8144 8145 8145 8146 989 8147 161 8148 164 8149 8149 8150 8150 8151 8151 8152 868 8153 11 8154 8154 8155 8155 8156 8156 8157 364 8158 364 8159 367 8160 441 8161 8161 8162 8162 8163 8163 8164 8164 8165 2 8166 583 8167 8167 8168 8168 8169 8169 8170 8170 8171 8171 8172 8172 8173 2 8174 583 8175 8175 8176 86 8177 8177 8178 8178 8179 142 8180 11 8181 144 8182 164 8183 165 8184 167 8185 169 8186 77 8187 128 8188 172 8189 8189 8190 8190 8191 142 8192 11 8193 144 8194 164 8195 2270 8196 2272 8197 2273 8198 2274 8199 2 8200 2 8201 3 8202 4 8203 2 8204 2 8205 2 8206 2 8207 11 8208 2 8209 2 8210 1705 8211 2 8212 1707 8213 22 8214 1709 8215 1710 8216 2 8217 11 8218 11 8219 1715 8220 1716 8221 1717 8222 22 8223 2 8224 2 8225 2 8226 11 8227 1723 8228 11 8229 1725 8230 11 8231 1727 8232 1728 8233 1729 8234 1730 8235 60 8236 1732 8237 2 8238 2 8239 1735 8240 11 8241 2 8242 2 8243 69 8244 133 8245 1741 8246 2 8247 2 8248 1744 8249 2 8250 1746 8251 1747 8252 2 8253 86 8254 87 8255 90 8256 136 8257 11 8258 1757 8259 1758 8260 101 8261 1760 8262 1761 8263 8263 8264 106 8265 106 8266 1765 8267 1766 8268 1767 8269 1768 8270 1769 8271 1770 8272 1771 8273 1774 8274 1775 8275 1776 8276 1777 8277 101 8278 2 8279 1781 8280 2 8281 2 8282 2 8283 2 8284 2 8285 2 8286 2 8287 2 8288 312 8289 2 8290 2 8291 2 8292 2 8293 2 8294 318 8295 2 8296 2 8297 2 8298 2 8299 2 8300 2 8301 2 8302 2 8303 2 8304 2 8305 2 8306 2 8307 2 8308 2 8309 2 8310 2 8311 1813 8312 1814 8313 1815 8314 1816 8315 1817 8316 1818 8317 1819 8318 1820 8319 1821 8320 1822 8321 1823 8322 1824 8323 1825 8324 1826 8325 1827 8326 1828 8327 4 8328 2 8329 1831 8330 1832 8331 1833 8332 1834 8333 1835 8334 1836 8335 1837 8336 1838 8337 1839 8338 1840 8339 1841 8340 364 8341 364 8342 40 8343 367 8344 441 8345 1847 8346 1848 8347 1849 8348 1850 8349 1851 8350 2 8351 583 8352 1854 8353 1855 8354 40 8355 1857 8356 287 8357 1859 8358 1860 8359 2 8360 583 8361 1863 8362 1864 8363 1865 8364 1866 8365 1867 8366 1868 8367 2 8368 751 8369 2 8370 1872 8371 1873 8372 1874 8373 1875 8374 1876 8375 1877 8376 1878 8377 1879 8378 1880 8379 1881 8380 868 8381 8001 8382 8002 8383 8003 8384 8004 8385 8005 8386 8006 8387 8007 8388 8008 8389 8009 8390 8010 8391 8011 8392 8012 8393 8013 8394 8014 8395 8015 8396 8016 8397 8017 8398 8018 8399 8019 8400 8020 8401 8021 8402 8022 8403 8023 8404 8024 8405 8025 8406 8026 8407 8027 8408 8028 8409 8029 8410 8030 8411 8031 8412 8032 8413 8033 8414 8034 8415 8035 8416 8036 8417 8037 8418 8038 8419 8039 8420 8040 8421 8041 8422 8042 8423 8043 8424 8044 8425 8045 8426 8046 8427 8047 8428 8048 8429 8049 8430 8050 8431 8051 8432 8052 8433 8053 8434 8054 8435 8055 8436 8056 8437 8057 8438 8058 8439 8059 8440 8060 8441 8061 8442 8062 8443 8063 8444 8064 8445 8065 8446 8066 8447 8067 8448 8068 8449 8069 8450 8070 8451 8071 8452 8072 8453 8073 8454 8074 8455 8075 8456 8076 8457 8077 8458 8078 8459 8079 8460 8080 8461 11 8462 8082 8463 8083 8464 8084 8465 2 8466 583 8467 8087 8468 956 8469 8089 8470 8090 8471 989 8472 161 8473 164 8474 8094 8475 8095 8476 1505 8477 2 8478 2 8479 2 8480 8100 8481 2 8482 8102 8483 2 8484 2 8485 2 8486 2 8487 2 8488 2 8489 2 8490 2 8491 2 8492 8112 8493 8113 8494 8114 8495 8115 8496 8116 8497 8117 8498 8118 8499 8119 8500 8120 8501 8121 8502 8122 8503 8123 8504 8124 8505 22 8506 8126 8507 4 8508 11 8509 8129 8510 90 8511 11 8512 8132 8513 101 8514 8134 8515 1569 8516 8136 8517 8137 8518 8138 8519 101 8520 868 8521 11 8522 8142 8523 956 8524 8144 8525 8145 8526 989 8527 161 8528 164 8529 8149 8530 8150 8531 8151 8532 868 8533 11 8534 8154 8535 8155 8536 8156 8537 364 8538 364 8539 367 8540 441 8541 8161 8542 8162 8543 8163 8544 8164 8545 2 8546 583 8547 8167 8548 8168 8549 8169 8550 8170 8551 8171 8552 8172 8553 2 8554 583 8555 8175 8556 8178 8557 142 8558 11 8559 144 8560 164 8561 165 8562 167 8563 169 8564 77 8565 128 8566 172 8567 8189 8568 8190 8569 142 8570 11 8571 144 8572 164 8573 2270 8574 2272 8575 2273 8576 2274 8577 2 8578 8263 8579 8579 8580 8579 8581 868 8582 8582 8583 11 8584 8584 8585 2 8586 583 8587 8587 8588 8588 8589 8589 8590 2756 8591 8591 8592 8592 8593 364 8594 364 8595 367 8596 441 8597 8597 8598 8598 8599 8599 8600 8600 8601 2 8602 583 8603 8603 8604 8604 8605 8605 8606 8606 8607 2827 8608 8608 8609 441 8610 8610 8611 2 8612 8612 8613 8613 8614 8614 8615 2 8616 8616 8617 8617 8618 8618 8619 8619 8620 8620 8621 2918 8622 8622 8623 69 8624 8624 8625 8625 8626 8626 8627 8627 8628 8628 8629 8629 8630 8630 8631 2 8632 8632 8633 8633 8634 8634 8635 8635 8636 8636 8637 8637 8638 2960 8639 2962 8640 8640 8641 8641 8642 8642 8643 8643 8644 8644 8645 364 8646 364 8647 40 8648 367 8649 441 8650 8650 8651 8651 8652 8652 8653 8653 8654 8654 8655 40 8656 8656 8657 8657 8658 2983 8659 2986 8660 2988 8661 2990 8662 77 8663 2992 8664 2993 8665 8665 8666 287 8667 8667 8668 2 8669 287 8670 77 8671 3959 8672 2852 8673 8673 8674 1882 8675 164 8676 2072 8677 2 8678 7859 8679 2 8680 161 8681 164 8682 8682 8683 8683 8684 8684 8685 8685 8686 8686 8687 956 8688 8688 8689 8689 8690 8690 8691 989 8692 161 8693 164 8694 8694 8695 8695 8696 2 8697 287 8698 8698 8699 2 8700 287 8701 8701 8702 8702 8703 8703 8704 956 8705 8705 8706 8706 8707 8707 8708 989 8709 5285 8710 161 8711 164 8712 8712 8713 8713 8714 956 8715 8715 8716 8716 8717 8717 8718 8718 8719 1882 8720 164 8721 8721 8722 8722 8723 8723 8724 8724 8725 287 8726 8726 8727 2 8728 583 8729 8729 8730 5356 8731 8731 8732 8732 8733 1505 8734 2 8735 3 8736 4 8737 2 8738 2 8739 2 8740 2 8741 11 8742 2 8743 2 8744 8744 8745 2 8746 8746 8747 22 8748 8748 8749 8749 8750 2 8751 11 8752 11 8753 8753 8754 8754 8755 8755 8756 22 8757 2 8758 2 8759 2 8760 11 8761 8761 8762 11 8763 8763 8764 11 8765 8765 8766 8766 8767 8767 8768 8768 8769 60 8770 8770 8771 2 8772 2 8773 8773 8774 11 8775 2 8776 2 8777 69 8778 133 8779 8779 8780 2 8781 2 8782 8782 8783 2 8784 8784 8785 8785 8786 2 8787 86 8788 87 8789 22 8790 90 8791 136 8792 11 8793 8793 8794 8794 8795 101 8796 8796 8797 8797 8798 8798 8799 106 8800 106 8801 8801 8802 8802 8803 8803 8804 8804 8805 8805 8806 8806 8807 8807 8808 8808 8809 8809 8810 8810 8811 8811 8812 101 8813 2 8814 11 8815 8815 8816 2 8817 2 8818 2 8819 2 8820 2 8821 2 8822 2 8823 2 8824 312 8825 2 8826 2 8827 2 8828 2 8829 2 8830 318 8831 2 8832 2 8833 2 8834 2 8835 2 8836 2 8837 2 8838 2 8839 2 8840 2 8841 2 8842 2 8843 2 8844 2 8845 2 8846 2 8847 8847 8848 8848 8849 8849 8850 8850 8851 8851 8852 8852 8853 8853 8854 8854 8855 8855 8856 8856 8857 8857 8858 8858 8859 8859 8860 8860 8861 8861 8862 8862 8863 4 8864 2 8865 8865 8866 8866 8867 8867 8868 8868 8869 8869 8870 8870 8871 8871 8872 8872 8873 8873 8874 8874 8875 8875 8876 364 8877 364 8878 40 8879 367 8880 441 8881 8881 8882 8882 8883 8883 8884 8884 8885 8885 8886 2 8887 583 8888 8888 8889 8889 8890 40 8891 8891 8892 287 8893 8893 8894 8894 8895 2 8896 583 8897 8897 8898 8898 8899 8899 8900 8900 8901 8901 8902 8902 8903 2 8904 751 8905 2 8906 8906 8907 8907 8908 8908 8909 8909 8910 8910 8911 8911 8912 8912 8913 8913 8914 8914 8915 8915 8916 1882 8917 164 8918 2072 8919 2 8920 8920 8921 868 8922 8922 8923 8923 8924 8924 8925 8925 8926 8926 8927 8927 8928 8928 8929 8929 8930 8930 8931 8931 8932 8932 8933 8933 8934 8934 8935 8935 8936 8936 8937 8937 8938 8938 8939 8939 8940 8940 8941 8941 8942 8942 8943 8943 8944 8944 8945 8945 8946 8946 8947 8947 8948 8948 8949 8949 8950 8950 8951 8951 8952 8952 8953 8953 8954 8954 8955 8955 8956 8956 8957 8957 8958 8958 8959 8959 8960 8960 8961 8961 8962 8962 8963 8963 8964 8964 8965 8965 8966 8966 8967 8967 8968 8968 8969 8969 8970 8970 8971 8971 8972 8972 8973 8973 8974 8974 8975 8975 8976 8976 8977 8977 8978 8978 8979 8979 8980 8980 8981 8981 8982 8982 8983 8983 8984 8984 8985 8985 8986 8986 8987 8987 8988 8988 8989 8989 8990 8990 8991 8991 8992 8992 8993 8993 8994 8994 8995 8995 8996 8996 8997 8997 8998 8998 8999 8999 9000 9000 9001 9001 9002 11 9003 9003 9004 9004 9005 9005 9006 2 9007 583 9008 9008 9009 956 9010 9010 9011 9011 9012 989 9013 161 9014 164 9015 9015 9016 9016 9017 1505 9018 2 9019 2 9020 2 9021 9021 9022 2 9023 9023 9024 2 9025 2 9026 2 9027 2 9028 2 9029 2 9030 2 9031 2 9032 2 9033 9033 9034 9034 9035 9035 9036 9036 9037 9037 9038 9038 9039 9039 9040 9040 9041 9041 9042 9042 9043 9043 9044 9044 9045 9045 9046 22 9047 9047 9048 4 9049 11 9050 9050 9051 90 9052 11 9053 9053 9054 101 9055 9055 9056 1569 9057 9057 9058 9058 9059 9059 9060 101 9061 868 9062 11 9063 9063 9064 956 9065 9065 9066 9066 9067 989 9068 161 9069 164 9070 9070 9071 9071 9072 9072 9073 868 9074 11 9075 9075 9076 9076 9077 9077 9078 364 9079 364 9080 367 9081 441 9082 9082 9083 9083 9084 9084 9085 9085 9086 2 9087 583 9088 9088 9089 9089 9090 9090 9091 9091 9092 9092 9093 9093 9094 2 9095 583 9096 9096 9097 9097 9098 142 9099 11 9100 144 9101 164 9102 165 9103 167 9104 169 9105 77 9106 128 9107 172 9108 9108 9109 9109 9110 142 9111 11 9112 144 9113 164 9114 2270 9115 2272 9116 2273 9117 2274 9118 2 9119 2 9120 3 9121 4 9122 2 9123 2 9124 2 9125 2 9126 11 9127 2 9128 2 9129 8744 9130 2 9131 8746 9132 22 9133 8748 9134 8749 9135 2 9136 11 9137 11 9138 8753 9139 8754 9140 8755 9141 22 9142 2 9143 2 9144 2 9145 11 9146 8761 9147 11 9148 8763 9149 11 9150 8765 9151 8766 9152 8767 9153 8768 9154 60 9155 8770 9156 2 9157 2 9158 8773 9159 11 9160 2 9161 2 9162 69 9163 133 9164 8779 9165 2 9166 2 9167 8782 9168 2 9169 8784 9170 8785 9171 2 9172 86 9173 87 9174 90 9175 136 9176 11 9177 8793 9178 8794 9179 101 9180 8796 9181 8797 9182 9182 9183 106 9184 106 9185 8801 9186 8802 9187 8803 9188 8804 9189 8805 9190 8806 9191 8807 9192 8808 9193 8809 9194 8810 9195 8811 9196 101 9197 2 9198 8815 9199 2 9200 2 9201 2 9202 2 9203 2 9204 2 9205 2 9206 2 9207 312 9208 2 9209 2 9210 2 9211 2 9212 2 9213 318 9214 2 9215 2 9216 2 9217 2 9218 2 9219 2 9220 2 9221 2 9222 2 9223 2 9224 2 9225 2 9226 2 9227 2 9228 2 9229 2 9230 8847 9231 8848 9232 8849 9233 8850 9234 8851 9235 8852 9236 8853 9237 8854 9238 8855 9239 8856 9240 8857 9241 8858 9242 8859 9243 8860 9244 8861 9245 8862 9246 4 9247 2 9248 8865 9249 8866 9250 8867 9251 8868 9252 8869 9253 8870 9254 8871 9255 8872 9256 8873 9257 8874 9258 8875 9259 364 9260 364 9261 40 9262 367 9263 441 9264 8881 9265 8882 9266 8883 9267 8884 9268 8885 9269 2 9270 583 9271 8888 9272 8889 9273 40 9274 8891 9275 287 9276 8893 9277 8894 9278 2 9279 583 9280 8897 9281 8898 9282 8899 9283 8900 9284 8901 9285 8902 9286 2 9287 751 9288 2 9289 8906 9290 8907 9291 8908 9292 8909 9293 8910 9294 8911 9295 8912 9296 8913 9297 8914 9298 8915 9299 868 9300 8922 9301 8923 9302 8924 9303 8925 9304 8926 9305 8927 9306 8928 9307 8929 9308 8930 9309 8931 9310 8932 9311 8933 9312 8934 9313 8935 9314 8936 9315 8937 9316 8938 9317 8939 9318 8940 9319 8941 9320 8942 9321 8943 9322 8944 9323 8945 9324 8946 9325 8947 9326 8948 9327 8949 9328 8950 9329 8951 9330 8952 9331 8953 9332 8954 9333 8955 9334 8956 9335 8957 9336 8958 9337 8959 9338 8960 9339 8961 9340 8962 9341 8963 9342 8964 9343 8965 9344 8966 9345 8967 9346 8968 9347 8969 9348 8970 9349 8971 9350 8972 9351 8973 9352 8974 9353 8975 9354 8976 9355 8977 9356 8978 9357 8979 9358 8980 9359 8981 9360 8982 9361 8983 9362 8984 9363 8985 9364 8986 9365 8987 9366 8988 9367 8989 9368 8990 9369 8991 9370 8992 9371 8993 9372 8994 9373 8995 9374 8996 9375 8997 9376 8998 9377 8999 9378 9000 9379 9001 9380 11 9381 9003 9382 9004 9383 9005 9384 2 9385 583 9386 9008 9387 956 9388 9010 9389 9011 9390 989 9391 161 9392 164 9393 9015 9394 9016 9395 1505 9396 2 9397 2 9398 2 9399 9021 9400 2 9401 9023 9402 2 9403 2 9404 2 9405 2 9406 2 9407 2 9408 2 9409 2 9410 2 9411 9033 9412 9034 9413 9035 9414 9036 9415 9037 9416 9038 9417 9039 9418 9040 9419 9041 9420 9042 9421 9043 9422 9044 9423 9045 9424 22 9425 9047 9426 4 9427 11 9428 9050 9429 90 9430 11 9431 9053 9432 101 9433 9055 9434 1569 9435 9057 9436 9058 9437 9059 9438 101 9439 868 9440 11 9441 9063 9442 956 9443 9065 9444 9066 9445 989 9446 161 9447 164 9448 9070 9449 9071 9450 9072 9451 868 9452 11 9453 9075 9454 9076 9455 9077 9456 364 9457 364 9458 367 9459 441 9460 9082 9461 9083 9462 9084 9463 9085 9464 2 9465 583 9466 9088 9467 9089 9468 9090 9469 9091 9470 9092 9471 9093 9472 2 9473 583 9474 9096 9475 9097 9476 142 9477 11 9478 144 9479 164 9480 165 9481 167 9482 169 9483 77 9484 128 9485 172 9486 9108 9487 9109 9488 142 9489 11 9490 144 9491 164 9492 2270 9493 2272 9494 2273 9495 2274 9496 2 9497 9182 9498 9498 9499 9498 9500 868 9501 9501 9502 11 9503 9503 9504 2 9505 583 9506 9506 9507 9507 9508 9508 9509 2756 9510 9510 9511 9511 9512 364 9513 364 9514 367 9515 441 9516 9516 9517 9517 9518 9518 9519 9519 9520 2 9521 583 9522 9522 9523 9523 9524 9524 9525 9525 9526 2827 9527 9527 9528 441 9529 9529 9530 2 9531 9531 9532 9532 9533 8614 9534 2 9535 9535 9536 9536 9537 9537 9538 9538 9539 9539 9540 2918 9541 9541 9542 69 9543 9543 9544 9544 9545 9545 9546 9546 9547 9547 9548 9548 9549 9549 9550 2 9551 9551 9552 9552 9553 9553 9554 9554 9555 9555 9556 9556 9557 2960 9558 2962 9559 9559 9560 9560 9561 9561 9562 9562 9563 9563 9564 364 9565 364 9566 40 9567 367 9568 441 9569 9569 9570 9570 9571 9571 9572 9572 9573 9573 9574 40 9575 9575 9576 9576 9577 2983 9578 2986 9579 2988 9580 2990 9581 77 9582 2992 9583 2993 9584 9584 9585 287 9586 9586 9587 2 9588 287 9589 77 9590 3959 9591 2852 9592 9592 9593 1882 9594 164 9595 2072 9596 2 9597 8920 9598 2 9599 161 9600 164 9601 9601 9602 9602 9603 9603 9604 9604 9605 9605 9606 956 9607 9607 9608 9608 9609 9609 9610 989 9611 161 9612 164 9613 9613 9614 9614 9615 2 9616 287 9617 9617 9618 2 9619 287 9620 9620 9621 9621 9622 9622 9623 956 9624 9624 9625 9625 9626 9626 9627 989 9628 5285 9629 161 9630 164 9631 9631 9632 9632 9633 956 9634 9634 9635 9635 9636 9636 9637 9637 9638 1882 9639 164 9640 9640 9641 9641 9642 9642 9643 9643 9644 287 9645 9645 9646 2 9647 583 9648 9648 9649 5356 9650 9650 9651 9651 9652 868 9653 9501 9654 11 9655 9503 9656 2 9657 583 9658 9506 9659 9507 9660 9508 9661 2756 9662 9510 9663 9511 9664 364 9665 364 9666 367 9667 441 9668 9516 9669 9517 9670 9518 9671 9519 9672 2 9673 583 9674 9522 9675 9523 9676 9524 9677 9525 9678 4 9679 11 9680 9680 9681 90 9682 11 9683 9683 9684 101 9685 9685 9686 1569 9687 9687 9688 9688 9689 9689 9690 101 9691 868 9692 11 9693 9693 9694 956 9695 9695 9696 9696 9697 989 9698 161 9699 164 9700 9700 9701 9701 9702 9702 9703 868 9704 11 9705 9705 9706 9706 9707 9707 9708 364 9709 364 9710 367 9711 441 9712 9712 9713 9713 9714 9714 9715 9715 9716 2 9717 583 9718 9718 9719 9719 9720 9720 9721 9721 9722 9722 9723 9723 9724 2 9725 583 9726 9726 9727 2827 9728 9527 9729 441 9730 9529 9731 2 9732 9531 9733 9532 9734 8614 9735 2 9736 9535 9737 9536 9738 9537 9739 9538 9740 9539 9741 2918 9742 9541 9743 69 9744 9543 9745 9544 9746 9545 9747 9546 9748 9547 9749 9548 9750 9549 9751 2 9752 9551 9753 9552 9754 9553 9755 9554 9756 9555 9757 9556 9758 2960 9759 2962 9760 9559 9761 9560 9762 9561 9763 9562 9764 9563 9765 364 9766 364 9767 40 9768 367 9769 441 9770 9569 9771 9570 9772 9571 9773 9572 9774 9573 9775 40 9776 9575 9777 9576 9778 2983 9779 2986 9780 2988 9781 2990 9782 77 9783 2992 9784 2993 9785 9584 9786 287 9787 9586 9788 2 9789 287 9790 77 9791 3959 9792 2852 9793 9592 9794 1882 9795 164 9796 9796 9797 2 9798 161 9799 164 9800 9601 9801 9602 9802 9603 9803 9604 9804 9605 9805 956 9806 9607 9807 9608 9808 9609 9809 989 9810 161 9811 164 9812 9613 9813 9614 9814 2 9815 287 9816 9617 9817 2 9818 287 9819 9620 9820 9621 9821 9622 9822 956 9823 9624 9824 9625 9825 9626 9826 989 9827 5285 9828 161 9829 164 9830 9631 9831 9632 9832 956 9833 9634 9834 9635 9835 9636 9836 9637 9837 1882 9838 164 9839 9640 9840 9641 9841 9642 9842 9643 9843 287 9844 9645 9845 2 9846 583 9847 9648 9848 5356 9849 9650 9850 9651 9851 868 9852 8582 9853 11 9854 8584 9855 2 9856 583 9857 8587 9858 8588 9859 8589 9860 2756 9861 8591 9862 8592 9863 9863 9864 364 9865 364 9866 367 9867 441 9868 8597 9869 8598 9870 8599 9871 8600 9872 2 9873 583 9874 8603 9875 8604 9876 8605 9877 8606 9878 9878 9879 2827 9880 8608 9881 441 9882 8610 9883 2 9884 8612 9885 8613 9886 8614 9887 2 9888 8616 9889 8617 9890 8618 9891 8619 9892 8620 9893 2918 9894 8622 9895 69 9896 8624 9897 8625 9898 8626 9899 8627 9900 8628 9901 8629 9902 8630 9903 2 9904 8632 9905 8633 9906 8634 9907 8635 9908 8636 9909 8637 9910 2960 9911 2962 9912 8640 9913 8641 9914 8642 9915 8643 9916 8644 9917 364 9918 364 9919 40 9920 367 9921 441 9922 8650 9923 8651 9924 8652 9925 8653 9926 8654 9927 40 9928 8656 9929 8657 9930 2983 9931 2986 9932 2988 9933 2990 9934 77 9935 2992 9936 2993 9937 8665 9938 287 9939 8667 9940 2 9941 287 9942 77 9943 3959 9944 2852 9945 8673 9946 1882 9947 164 9948 9948 9949 2 9950 161 9951 164 9952 8682 9953 8683 9954 8684 9955 8685 9956 8686 9957 956 9958 8688 9959 8689 9960 8690 9961 989 9962 161 9963 164 9964 8694 9965 8695 9966 2 9967 287 9968 8698 9969 2 9970 287 9971 8701 9972 9972 9973 8702 9974 8703 9975 956 9976 8705 9977 8706 9978 8707 9979 989 9980 5285 9981 161 9982 164 9983 8712 9984 8713 9985 956 9986 8715 9987 8716 9988 8717 9989 8718 9990 1882 9991 164 9992 8721 9993 8722 9994 8723 9995 8724 9996 287 9997 8726 9998 2 9999 583 10000 8729 10001 5356 10002 8731 10003 8732 10004 2246 10005 4 10006 2 10007 10007 10008 10008 10009 10009 10010 10010 10011 10011 10012 10012 10013 10013 10014 2960 10015 2962 10016 10016 10017 11 10018 10018 10019 10019 10020 10020 10021 364 10022 364 10023 40 10024 367 10025 441 10026 10026 10027 10027 10028 10028 10029 10029 10030 10030 10031 2 10032 583 10033 10033 10034 10034 10035 956 10036 10036 10037 10037 10038 989 10039 161 10040 164 10041 10041 10042 10042 10043 40 10044 10044 10045 10045 10046 2983 10047 2986 10048 2988 10049 2990 10050 77 10051 2992 10052 2993 10053 10053 10054 287 10055 10055 10056 10056 10057 2 10058 583 10059 10059 10060 1882 10061 164 10062 10062 10063 287 10064 10064 10065 2247 10066 2 10067 583 10068 2250 10069 5356 10070 10070 10071 2253 10072 142 10073 11 10074 144 10075 164 10076 165 10077 167 10078 169 10079 77 10080 128 10081 172 10082 2264 10083 2265 10084 142 10085 11 10086 144 10087 164 10088 2270 10089 2272 10090 2273 10091 2274 10092 2 10093 2276 10094 2277 10095 2 10096 2911 10097 868 10098 7846 10099 11 10100 7848 10101 2 10102 583 10103 7851 10104 7852 10105 7853 10106 2756 10107 7855 10108 7856 10109 364 10110 364 10111 367 10112 441 10113 7864 10114 7865 10115 7866 10116 7867 10117 2 10118 583 10119 7870 10120 7871 10121 7872 10122 7873 10123 2827 10124 7876 10125 441 10126 7878 10127 2 10128 7880 10129 7881 10130 1288 10131 2 10132 7884 10133 7885 10134 7886 10135 7887 10136 7888 10137 2918 10138 7890 10139 69 10140 7892 10141 7893 10142 7894 10143 7895 10144 7896 10145 7897 10146 7898 10147 2 10148 7900 10149 7901 10150 7902 10151 7903 10152 7904 10153 7905 10154 2960 10155 2962 10156 7908 10157 7909 10158 7910 10159 7911 10160 7912 10161 364 10162 364 10163 40 10164 367 10165 441 10166 7918 10167 7919 10168 7920 10169 7921 10170 7922 10171 40 10172 7924 10173 7925 10174 2983 10175 2986 10176 2988 10177 2990 10178 77 10179 2992 10180 2993 10181 7933 10182 287 10183 7935 10184 2 10185 287 10186 77 10187 3959 10188 2852 10189 7941 10190 1882 10191 164 10192 2072 10193 2 10194 2074 10195 2 10196 161 10197 164 10198 7948 10199 7949 10200 7950 10201 7951 10202 7952 10203 956 10204 7954 10205 7955 10206 7956 10207 989 10208 161 10209 164 10210 7960 10211 7961 10212 2 10213 287 10214 7964 10215 2 10216 287 10217 7967 10218 7969 10219 7970 10220 956 10221 7972 10222 7973 10223 7974 10224 989 10225 5285 10226 161 10227 164 10228 7979 10229 7980 10230 956 10231 7982 10232 7983 10233 7984 10234 7985 10235 1882 10236 164 10237 7988 10238 7989 10239 7990 10240 7991 10241 287 10242 7993 10243 2 10244 583 10245 7996 10246 5356 10247 7998 10248 7999 10249 2913 10250 2915 10251 2945 10252 2946 10253 956 10254 2947 10255 2637 10256 989 10257 161 10258 164 10259 2641 10260 2642 10261 2948 10262 4 10263 2 10264 2952 10265 2953 10266 2954 10267 2956 10268 2957 10269 2676 10270 2958 10271 2960 10272 2962 10273 2964 10274 2965 10275 2966 10276 2967 10277 2968 10278 364 10279 364 10280 40 10281 367 10282 441 10283 2974 10284 2975 10285 2709 10286 2976 10287 2711 10288 2 10289 583 10290 2714 10291 2977 10292 956 10293 2978 10294 2692 10295 989 10296 161 10297 164 10298 2696 10299 2697 10300 40 10301 2980 10302 2981 10303 2983 10304 2986 10305 2988 10306 2990 10307 77 10308 2992 10309 2993 10310 3018 10311 287 10312 7824 10313 2719 10314 2 10315 583 10316 2722 10317 2 10318 287 10319 77 10320 3959 10321 2852 10322 5300 10323 1882 10324 164 10325 2072 10326 2 10327 2546 10328 2 10329 161 10330 164 10331 5304 10332 5305 10333 5306 10334 956 10335 5308 10336 5309 10337 5310 10338 989 10339 161 10340 164 10341 5314 10342 5315 10343 2 10344 287 10345 5318 10346 2 10347 287 10348 5321 10349 956 10350 5324 10351 5325 10352 5326 10353 989 10354 5285 10355 161 10356 164 10357 5331 10358 5332 10359 956 10360 5334 10361 5335 10362 1882 10363 164 10364 5338 10365 868 10366 2548 10367 2549 10368 2550 10369 2551 10370 2552 10371 2553 10372 2554 10373 2555 10374 2556 10375 2557 10376 2558 10377 2559 10378 2560 10379 2561 10380 2562 10381 2563 10382 2564 10383 2565 10384 2566 10385 2567 10386 2568 10387 2569 10388 2570 10389 2571 10390 2572 10391 2573 10392 2574 10393 2575 10394 2576 10395 2577 10396 2578 10397 2579 10398 2580 10399 2581 10400 2582 10401 2583 10402 2584 10403 2585 10404 2586 10405 2587 10406 2588 10407 2589 10408 2590 10409 2591 10410 2592 10411 2593 10412 2594 10413 2595 10414 2596 10415 2597 10416 2598 10417 2599 10418 2600 10419 2601 10420 2602 10421 2603 10422 2604 10423 2605 10424 2606 10425 2607 10426 2608 10427 2609 10428 2610 10429 2611 10430 2612 10431 2613 10432 2614 10433 2615 10434 2616 10435 2617 10436 2618 10437 2619 10438 2620 10439 2621 10440 2622 10441 2623 10442 2624 10443 2625 10444 2626 10445 2627 10446 11 10447 2629 10448 2630 10449 2631 10450 2 10451 583 10452 2634 10453 2636 10454 1505 10455 2 10456 2 10457 2 10458 2647 10459 2 10460 2649 10461 2 10462 2 10463 2 10464 2 10465 2 10466 2 10467 2 10468 2 10469 2 10470 2659 10471 2660 10472 2661 10473 2662 10474 2663 10475 2664 10476 2665 10477 2666 10478 2667 10479 2668 10480 2669 10481 2670 10482 2671 10483 22 10484 2673 10485 11 10486 90 10487 11 10488 2679 10489 101 10490 2681 10491 1569 10492 2683 10493 2684 10494 2685 10495 101 10496 868 10497 11 10498 2689 10499 2691 10500 2698 10501 868 10502 11 10503 2701 10504 2702 10505 2703 10506 364 10507 364 10508 367 10509 441 10510 2708 10511 2710 10512 2715 10513 2716 10514 2717 10515 2718 10516 164 10517 10517 10518 2723 10519 2 10520 10520 10521 142 10522 11 10523 144 10524 164 10525 165 10526 167 10527 169 10528 77 10529 128 10530 172 10531 2734 10532 2735 10533 142 10534 11 10535 144 10536 164 10537 2270 10538 2272 10539 2273 10540 2274 10541 2 10542 2745 10543 2746 10544 868 10545 2748 10546 11 10547 2750 10548 2 10549 583 10550 2753 10551 2754 10552 2755 10553 2756 10554 2758 10555 2759 10556 364 10557 364 10558 367 10559 441 10560 2764 10561 2765 10562 2766 10563 2767 10564 2 10565 583 10566 2770 10567 2771 10568 2772 10569 2773 10570 2827 10571 2832 10572 441 10573 2905 10574 2 10575 2907 10576 2908 10577 2912 10578 2914 10579 2918 10580 2919 10581 69 10582 2924 10583 2933 10584 2949 10585 7828 10586 7829 10587 10587 10588 7830 10589 7831 10590 7832 10591 7833 10592 7834 10593 7835 10594 7836 10595 287 10596 7838 10597 2 10598 583 10599 7841 10600 5356 10601 7843 10602 7844 10603 1505 10604 2 10605 2 10606 2 10607 10607 10608 2 10609 10609 10610 2 10611 2 10612 2 10613 2 10614 2 10615 2 10616 2 10617 2 10618 2 10619 10619 10620 10620 10621 10621 10622 10622 10623 10623 10624 10624 10625 10625 10626 10626 10627 10627 10628 10628 10629 10629 10630 10630 10631 10631 10632 22 10633 10633 10634 4 10635 11 10636 10636 10637 90 10638 11 10639 10639 10640 101 10641 10641 10642 1569 10643 10643 10644 10644 10645 10645 10646 101 10647 868 10648 11 10649 10649 10650 956 10651 10651 10652 10652 10653 989 10654 161 10655 164 10656 10656 10657 10657 10658 10658 10659 868 10660 11 10661 10661 10662 10662 10663 10663 10664 364 10665 364 10666 367 10667 441 10668 10668 10669 10669 10670 10670 10671 10671 10672 2 10673 583 10674 10674 10675 10675 10676 10676 10677 10677 10678 10678 10679 10679 10680 2 10681 583 10682 10682 10683 10683 10684 142 10685 11 10686 144 10687 164 10688 165 10689 167 10690 169 10691 77 10692 128 10693 172 10694 10694 10695 10695 10696 142 10697 11 10698 144 10699 164 10700 2270 10701 2272 10702 2273 10703 2274 10704 2 10705 2 10706 3 10707 4 10708 2 10709 2 10710 2 10711 2 10712 11 10713 2 10714 2 10715 698 10716 2 10717 700 10718 22 10719 702 10720 703 10721 2 10722 11 10723 11 10724 707 10725 708 10726 709 10727 22 10728 2 10729 2 10730 2 10731 11 10732 715 10733 11 10734 717 10735 11 10736 719 10737 720 10738 721 10739 722 10740 60 10741 724 10742 2 10743 2 10744 727 10745 11 10746 2 10747 2 10748 69 10749 133 10750 733 10751 2 10752 2 10753 736 10754 2 10755 738 10756 739 10757 2 10758 86 10759 87 10760 90 10761 136 10762 11 10763 746 10764 747 10765 101 10766 749 10767 750 10768 10768 10769 106 10770 106 10771 754 10772 755 10773 756 10774 757 10775 758 10776 759 10777 760 10778 761 10779 762 10780 763 10781 764 10782 101 10783 2 10784 767 10785 2 10786 2 10787 2 10788 2 10789 2 10790 2 10791 2 10792 2 10793 312 10794 2 10795 2 10796 2 10797 2 10798 2 10799 318 10800 2 10801 2 10802 2 10803 2 10804 2 10805 2 10806 2 10807 2 10808 2 10809 2 10810 2 10811 2 10812 2 10813 2 10814 2 10815 2 10816 799 10817 800 10818 801 10819 802 10820 803 10821 804 10822 805 10823 806 10824 807 10825 808 10826 809 10827 810 10828 811 10829 812 10830 813 10831 814 10832 4 10833 2 10834 817 10835 818 10836 819 10837 820 10838 821 10839 822 10840 823 10841 824 10842 825 10843 826 10844 827 10845 364 10846 364 10847 40 10848 367 10849 441 10850 833 10851 834 10852 835 10853 836 10854 837 10855 2 10856 583 10857 840 10858 841 10859 40 10860 843 10861 287 10862 845 10863 846 10864 2 10865 583 10866 849 10867 850 10868 851 10869 852 10870 853 10871 854 10872 2 10873 751 10874 2 10875 858 10876 859 10877 860 10878 861 10879 862 10880 863 10881 864 10882 865 10883 866 10884 867 10885 868 10886 869 10887 870 10888 871 10889 872 10890 873 10891 874 10892 875 10893 876 10894 877 10895 878 10896 879 10897 880 10898 881 10899 882 10900 883 10901 884 10902 885 10903 886 10904 887 10905 888 10906 889 10907 890 10908 891 10909 892 10910 893 10911 894 10912 895 10913 896 10914 897 10915 898 10916 899 10917 900 10918 901 10919 902 10920 903 10921 904 10922 905 10923 906 10924 907 10925 908 10926 909 10927 910 10928 911 10929 912 10930 913 10931 914 10932 915 10933 916 10934 917 10935 918 10936 919 10937 920 10938 921 10939 922 10940 923 10941 924 10942 925 10943 926 10944 927 10945 928 10946 929 10947 930 10948 931 10949 932 10950 933 10951 934 10952 935 10953 936 10954 937 10955 938 10956 939 10957 940 10958 941 10959 942 10960 943 10961 944 10962 945 10963 946 10964 947 10965 948 10966 11 10967 950 10968 951 10969 952 10970 2 10971 583 10972 955 10973 956 10974 957 10975 962 10976 989 10977 161 10978 164 10979 10517 10980 10587 10981 1505 10982 2 10983 2 10984 2 10985 10607 10986 2 10987 10609 10988 2 10989 2 10990 2 10991 2 10992 2 10993 2 10994 2 10995 2 10996 2 10997 10619 10998 10620 10999 10621 11000 10622 11001 10623 11002 10624 11003 10625 11004 10626 11005 10627 11006 10628 11007 10629 11008 10630 11009 10631 11010 22 11011 10633 11012 4 11013 11 11014 10636 11015 90 11016 11 11017 10639 11018 101 11019 10641 11020 1569 11021 10643 11022 10644 11023 10645 11024 101 11025 868 11026 11 11027 10649 11028 956 11029 10651 11030 10652 11031 989 11032 161 11033 164 11034 10656 11035 10657 11036 10658 11037 868 11038 11 11039 10661 11040 10662 11041 10663 11042 364 11043 364 11044 367 11045 441 11046 10668 11047 10669 11048 10670 11049 10671 11050 2 11051 583 11052 10674 11053 10675 11054 10676 11055 10677 11056 10678 11057 10679 11058 2 11059 583 11060 10682 11061 10683 11062 142 11063 11 11064 144 11065 164 11066 165 11067 167 11068 169 11069 77 11070 128 11071 172 11072 10694 11073 10695 11074 142 11075 11 11076 144 11077 164 11078 2270 11079 2272 11080 2273 11081 2274 11082 2 11083 10768 11084 11084 11085 11084 11086 868 11087 11087 11088 11 11089 11089 11090 2 11091 583 11092 11092 11093 11093 11094 11094 11095 2756 11096 11096 11097 11097 11098 364 11099 364 11100 367 11101 441 11102 11102 11103 11103 11104 11104 11105 11105 11106 2 11107 583 11108 11108 11109 11109 11110 11110 11111 11111 11112 2827 11113 11113 11114 441 11115 11115 11116 2 11117 11117 11118 11118 11119 2 11120 3 11121 4 11122 2 11123 2 11124 2 11125 2 11126 11 11127 2 11128 2 11129 11129 11130 2 11131 11131 11132 22 11133 11133 11134 11134 11135 2 11136 11 11137 11 11138 11138 11139 11139 11140 11140 11141 22 11142 2 11143 2 11144 2 11145 11 11146 11146 11147 11 11148 11148 11149 11 11150 11150 11151 11151 11152 11152 11153 11153 11154 60 11155 11155 11156 2 11157 2 11158 11158 11159 11 11160 2 11161 2 11162 69 11163 133 11164 11164 11165 2 11166 2 11167 11167 11168 2 11169 11169 11170 11170 11171 2 11172 86 11173 87 11174 90 11175 136 11176 11 11177 11177 11178 11178 11179 101 11180 11180 11181 11181 11182 11182 11183 106 11184 106 11185 11185 11186 11186 11187 11187 11188 11188 11189 11189 11190 11190 11191 11191 11192 11192 11193 11193 11194 11194 11195 11195 11196 101 11197 2 11198 11198 11199 2 11200 2 11201 2 11202 2 11203 2 11204 2 11205 2 11206 2 11207 312 11208 2 11209 2 11210 2 11211 2 11212 2 11213 318 11214 2 11215 2 11216 2 11217 2 11218 2 11219 2 11220 2 11221 2 11222 2 11223 2 11224 2 11225 2 11226 2 11227 2 11228 2 11229 2 11230 11230 11231 11231 11232 11232 11233 11233 11234 11234 11235 11235 11236 11236 11237 11237 11238 11238 11239 11239 11240 11240 11241 11241 11242 11242 11243 11243 11244 11244 11245 11245 11246 4 11247 2 11248 11248 11249 11249 11250 11250 11251 11251 11252 11252 11253 11253 11254 11254 11255 11255 11256 11256 11257 11257 11258 11258 11259 364 11260 364 11261 40 11262 367 11263 441 11264 11264 11265 11265 11266 11266 11267 11267 11268 11268 11269 2 11270 583 11271 11271 11272 11272 11273 40 11274 11274 11275 287 11276 11276 11277 11277 11278 2 11279 583 11280 11280 11281 11281 11282 11282 11283 11283 11284 11284 11285 11285 11286 2 11287 751 11288 2 11289 11289 11290 11290 11291 11291 11292 11292 11293 11293 11294 11294 11295 11295 11296 11296 11297 11297 11298 11298 11299 868 11300 11300 11301 11301 11302 11302 11303 11303 11304 11304 11305 11305 11306 11306 11307 11307 11308 11308 11309 11309 11310 11310 11311 11311 11312 11312 11313 11313 11314 11314 11315 11315 11316 11316 11317 11317 11318 11318 11319 11319 11320 11320 11321 11321 11322 11322 11323 11323 11324 11324 11325 11325 11326 11326 11327 11327 11328 11328 11329 11329 11330 11330 11331 11331 11332 11332 11333 11333 11334 11334 11335 11335 11336 11336 11337 11337 11338 11338 11339 11339 11340 11340 11341 11341 11342 11342 11343 11343 11344 11344 11345 11345 11346 11346 11347 11347 11348 11348 11349 11349 11350 11350 11351 11351 11352 11352 11353 11353 11354 11354 11355 11355 11356 11356 11357 11357 11358 11358 11359 11359 11360 11360 11361 11361 11362 11362 11363 11363 11364 11364 11365 11365 11366 11366 11367 11367 11368 11368 11369 11369 11370 11370 11371 11371 11372 11372 11373 11373 11374 11374 11375 11375 11376 11376 11377 11377 11378 11378 11379 11379 11380 11 11381 11381 11382 11382 11383 11383 11384 2 11385 583 11386 11386 11387 956 11388 11388 11389 11389 11390 989 11391 161 11392 164 11393 11393 11394 11394 11395 1505 11396 2 11397 2 11398 2 11399 11399 11400 2 11401 11401 11402 2 11403 2 11404 2 11405 2 11406 2 11407 2 11408 2 11409 2 11410 2 11411 11411 11412 11412 11413 11413 11414 11414 11415 11415 11416 11416 11417 11417 11418 11418 11419 11419 11420 11420 11421 11421 11422 11422 11423 11423 11424 22 11425 11425 11426 4 11427 11 11428 11428 11429 90 11430 11 11431 11431 11432 101 11433 11433 11434 1569 11435 11435 11436 11436 11437 11437 11438 101 11439 868 11440 11 11441 11441 11442 956 11443 11443 11444 11444 11445 989 11446 161 11447 164 11448 11448 11449 11449 11450 11450 11451 868 11452 11 11453 11453 11454 11454 11455 11455 11456 364 11457 364 11458 367 11459 441 11460 11460 11461 11461 11462 11462 11463 11463 11464 2 11465 583 11466 11466 11467 11467 11468 11468 11469 11469 11470 11470 11471 11471 11472 2 11473 583 11474 11474 11475 11475 11476 142 11477 11 11478 144 11479 164 11480 165 11481 167 11482 169 11483 77 11484 128 11485 172 11486 11486 11487 11487 11488 142 11489 11 11490 144 11491 164 11492 2270 11493 2272 11494 2273 11495 2274 11496 2 11497 11497 11498 11498 11499 2 11500 11500 11501 868 11502 11502 11503 11 11504 11504 11505 2 11506 583 11507 11507 11508 11508 11509 11509 11510 2756 11511 11511 11512 11512 11513 364 11514 364 11515 367 11516 441 11517 11517 11518 11518 11519 11519 11520 11520 11521 2 11522 583 11523 11523 11524 11524 11525 11525 11526 11526 11527 2827 11528 11528 11529 441 11530 11530 11531 2 11532 11532 11533 11533 11534 11182 11535 2 11536 11536 11537 11537 11538 11538 11539 11539 11540 11540 11541 2918 11542 11542 11543 69 11544 11544 11545 11545 11546 11546 11547 11547 11548 11548 11549 11549 11550 11550 11551 2 11552 11552 11553 11553 11554 11554 11555 11555 11556 11556 11557 11557 11558 2960 11559 2962 11560 11560 11561 11561 11562 11562 11563 11563 11564 11564 11565 364 11566 364 11567 40 11568 367 11569 441 11570 11570 11571 11571 11572 11572 11573 11573 11574 11574 11575 40 11576 11576 11577 11577 11578 2983 11579 2986 11580 2988 11581 2990 11582 77 11583 2992 11584 2993 11585 11585 11586 287 11587 11587 11588 2 11589 287 11590 77 11591 3959 11592 2852 11593 11593 11594 1882 11595 164 11596 2072 11597 2 11598 11598 11599 2 11600 161 11601 164 11602 11602 11603 11603 11604 11604 11605 11605 11606 11606 11607 956 11608 11608 11609 11609 11610 11610 11611 989 11612 161 11613 164 11614 11614 11615 11615 11616 2 11617 287 11618 11618 11619 2 11620 287 11621 11621 11622 11622 11623 11623 11624 956 11625 11625 11626 11626 11627 11627 11628 989 11629 5285 11630 161 11631 164 11632 11632 11633 11633 11634 956 11635 11635 11636 11636 11637 11637 11638 11638 11639 1882 11640 164 11641 11641 11642 11642 11643 11643 11644 11644 11645 287 11646 11646 11647 2 11648 583 11649 11649 11650 5356 11651 11651 11652 11652 11653 11653 11654 11654 11655 11655 11656 11656 11657 2918 11658 11658 11659 69 11660 11660 11661 11661 11662 11662 11663 11663 11664 11664 11665 11665 11666 11666 11667 2 11668 11668 11669 11669 11670 11670 11671 11671 11672 11672 11673 11673 11674 2960 11675 2962 11676 11676 11677 11677 11678 11678 11679 11679 11680 11680 11681 364 11682 364 11683 40 11684 367 11685 441 11686 11686 11687 11687 11688 11688 11689 11689 11690 11690 11691 40 11692 11692 11693 11693 11694 2983 11695 2986 11696 2988 11697 2990 11698 77 11699 2992 11700 2993 11701 11701 11702 287 11703 11703 11704 2 11705 287 11706 77 11707 3959 11708 2852 11709 11709 11710 1882 11711 164 11712 2072 11713 2 11714 11714 11715 2 11716 161 11717 164 11718 11718 11719 11719 11720 11720 11721 11721 11722 11722 11723 956 11724 11724 11725 11725 11726 11726 11727 989 11728 161 11729 164 11730 11730 11731 11731 11732 2 11733 287 11734 11734 11735 2 11736 287 11737 11737 11738 11738 11739 11739 11740 956 11741 11741 11742 11742 11743 11743 11744 989 11745 5285 11746 161 11747 164 11748 11748 11749 11749 11750 956 11751 11751 11752 11752 11753 11753 11754 11754 11755 1882 11756 164 11757 11757 11758 11758 11759 11759 11760 11760 11761 287 11762 11762 11763 2 11764 583 11765 11765 11766 5356 11767 11767 11768 11768 11769 2 11770 11770 11771 868 11772 11087 11773 11 11774 11089 11775 2 11776 583 11777 11092 11778 11093 11779 11094 11780 2756 11781 11096 11782 11097 11783 364 11784 364 11785 367 11786 441 11787 11102 11788 11103 11789 11104 11790 11105 11791 2 11792 583 11793 11108 11794 11109 11795 11110 11796 11111 11797 2827 11798 11113 11799 441 11800 11115 11801 2 11802 11117 11803 11118 11804 2 11805 3 11806 4 11807 2 11808 2 11809 2 11810 2 11811 11 11812 2 11813 2 11814 11129 11815 2 11816 11131 11817 22 11818 11133 11819 11134 11820 2 11821 11 11822 11 11823 11138 11824 11139 11825 11140 11826 22 11827 2 11828 2 11829 2 11830 11 11831 11146 11832 11 11833 11148 11834 11 11835 11150 11836 11151 11837 11152 11838 11153 11839 60 11840 11155 11841 2 11842 2 11843 11158 11844 11 11845 2 11846 2 11847 69 11848 133 11849 11164 11850 2 11851 2 11852 11167 11853 2 11854 11169 11855 11170 11856 2 11857 86 11858 87 11859 90 11860 136 11861 11 11862 11177 11863 11178 11864 101 11865 11180 11866 11181 11867 11182 11868 106 11869 106 11870 11185 11871 11186 11872 11187 11873 11188 11874 11189 11875 11190 11876 11191 11877 11192 11878 11193 11879 11194 11880 11195 11881 101 11882 2 11883 11198 11884 2 11885 2 11886 2 11887 2 11888 2 11889 2 11890 2 11891 2 11892 312 11893 2 11894 2 11895 2 11896 2 11897 2 11898 318 11899 2 11900 2 11901 2 11902 2 11903 2 11904 2 11905 2 11906 2 11907 2 11908 2 11909 2 11910 2 11911 2 11912 2 11913 2 11914 2 11915 11230 11916 11231 11917 11232 11918 11233 11919 11234 11920 11235 11921 11236 11922 11237 11923 11238 11924 11239 11925 11240 11926 11241 11927 11242 11928 11243 11929 11244 11930 11245 11931 4 11932 2 11933 11248 11934 11249 11935 11250 11936 11251 11937 11252 11938 11253 11939 11254 11940 11255 11941 11256 11942 11257 11943 11258 11944 364 11945 364 11946 40 11947 367 11948 441 11949 11264 11950 11265 11951 11266 11952 11267 11953 11268 11954 2 11955 583 11956 11271 11957 11272 11958 40 11959 11274 11960 287 11961 11276 11962 11277 11963 2 11964 583 11965 11280 11966 11281 11967 11282 11968 11283 11969 11284 11970 11285 11971 2 11972 751 11973 2 11974 11289 11975 11290 11976 11291 11977 11292 11978 11293 11979 11294 11980 11295 11981 11296 11982 11297 11983 11298 11984 868 11985 11300 11986 11301 11987 11302 11988 11303 11989 11304 11990 11305 11991 11306 11992 11307 11993 11308 11994 11309 11995 11310 11996 11311 11997 11312 11998 11313 11999 11314 12000 11315 12001 11316 12002 11317 12003 11318 12004 11319 12005 11320 12006 11321 12007 11322 12008 11323 12009 11324 12010 11325 12011 11326 12012 11327 12013 11328 12014 11329 12015 11330 12016 11331 12017 11332 12018 11333 12019 11334 12020 11335 12021 11336 12022 11337 12023 11338 12024 11339 12025 11340 12026 11341 12027 11342 12028 11343 12029 11344 12030 11345 12031 11346 12032 11347 12033 11348 12034 11349 12035 11350 12036 11351 12037 11352 12038 11353 12039 11354 12040 11355 12041 11356 12042 11357 12043 11358 12044 11359 12045 11360 12046 11361 12047 11362 12048 11363 12049 11364 12050 11365 12051 11366 12052 11367 12053 11368 12054 11369 12055 11370 12056 11371 12057 11372 12058 11373 12059 11374 12060 11375 12061 11376 12062 11377 12063 11378 12064 11379 12065 11 12066 11381 12067 11382 12068 11383 12069 2 12070 583 12071 11386 12072 956 12073 11388 12074 11389 12075 989 12076 161 12077 164 12078 11393 12079 11394 12080 1505 12081 2 12082 2 12083 2 12084 11399 12085 2 12086 11401 12087 2 12088 2 12089 2 12090 2 12091 2 12092 2 12093 2 12094 2 12095 2 12096 11411 12097 11412 12098 11413 12099 11414 12100 11415 12101 11416 12102 11417 12103 11418 12104 11419 12105 11420 12106 11421 12107 11422 12108 11423 12109 22 12110 11425 12111 4 12112 11 12113 11428 12114 90 12115 11 12116 11431 12117 101 12118 11433 12119 1569 12120 11435 12121 11436 12122 11437 12123 101 12124 868 12125 11 12126 11441 12127 956 12128 11443 12129 11444 12130 989 12131 161 12132 164 12133 11448 12134 11449 12135 11450 12136 868 12137 11 12138 11453 12139 11454 12140 11455 12141 364 12142 364 12143 367 12144 441 12145 11460 12146 11461 12147 11462 12148 11463 12149 2 12150 583 12151 11466 12152 11467 12153 11468 12154 11469 12155 11470 12156 11471 12157 2 12158 583 12159 11474 12160 11475 12161 142 12162 11 12163 144 12164 164 12165 165 12166 167 12167 169 12168 77 12169 128 12170 172 12171 11486 12172 11487 12173 142 12174 11 12175 144 12176 164 12177 2270 12178 2272 12179 2273 12180 2274 12181 2 12182 11497 12183 11498 12184 2 12185 11500 12186 868 12187 11502 12188 11 12189 11504 12190 2 12191 583 12192 11507 12193 11508 12194 11509 12195 2756 12196 11511 12197 11512 12198 364 12199 364 12200 367 12201 441 12202 11517 12203 11518 12204 11519 12205 11520 12206 2 12207 583 12208 11523 12209 11524 12210 11525 12211 11526 12212 2827 12213 11528 12214 441 12215 11530 12216 2 12217 11532 12218 11533 12219 11182 12220 2 12221 11536 12222 11537 12223 11538 12224 11539 12225 11540 12226 2918 12227 11542 12228 69 12229 11544 12230 11545 12231 11546 12232 11547 12233 11548 12234 11549 12235 11550 12236 2 12237 11552 12238 11553 12239 11554 12240 11555 12241 11556 12242 11557 12243 2960 12244 2962 12245 11560 12246 11561 12247 11562 12248 11563 12249 11564 12250 364 12251 364 12252 40 12253 367 12254 441 12255 11570 12256 11571 12257 11572 12258 11573 12259 11574 12260 40 12261 11576 12262 11577 12263 2983 12264 2986 12265 2988 12266 2990 12267 77 12268 2992 12269 2993 12270 11585 12271 287 12272 11587 12273 2 12274 287 12275 77 12276 3959 12277 2852 12278 11593 12279 1882 12280 164 12281 2072 12282 2 12283 11598 12284 2 12285 161 12286 164 12287 11602 12288 11603 12289 11604 12290 11605 12291 11606 12292 956 12293 11608 12294 11609 12295 11610 12296 989 12297 161 12298 164 12299 11614 12300 11615 12301 2 12302 287 12303 11618 12304 2 12305 287 12306 11621 12307 11622 12308 11623 12309 956 12310 11625 12311 11626 12312 11627 12313 989 12314 5285 12315 161 12316 164 12317 11632 12318 11633 12319 956 12320 11635 12321 11636 12322 11637 12323 11638 12324 1882 12325 164 12326 11641 12327 11642 12328 11643 12329 11644 12330 287 12331 11646 12332 2 12333 583 12334 11649 12335 5356 12336 11651 12337 11652 12338 11653 12339 11654 12340 11655 12341 11656 12342 2918 12343 11658 12344 69 12345 11660 12346 11661 12347 11662 12348 11663 12349 11664 12350 11665 12351 11666 12352 2 12353 11668 12354 11669 12355 11670 12356 11671 12357 11672 12358 11673 12359 2960 12360 2962 12361 11676 12362 11677 12363 11678 12364 11679 12365 11680 12366 364 12367 364 12368 40 12369 367 12370 441 12371 11686 12372 11687 12373 11688 12374 11689 12375 11690 12376 40 12377 11692 12378 11693 12379 2983 12380 2986 12381 2988 12382 2990 12383 77 12384 2992 12385 2993 12386 11701 12387 287 12388 11703 12389 2 12390 287 12391 77 12392 3959 12393 2852 12394 11709 12395 1882 12396 164 12397 2072 12398 2 12399 11714 12400 2 12401 161 12402 164 12403 11718 12404 11719 12405 11720 12406 11721 12407 11722 12408 956 12409 11724 12410 11725 12411 11726 12412 989 12413 161 12414 164 12415 11730 12416 11731 12417 2 12418 287 12419 11734 12420 2 12421 287 12422 11737 12423 11738 12424 11739 12425 956 12426 11741 12427 11742 12428 11743 12429 989 12430 5285 12431 161 12432 164 12433 11748 12434 11749 12435 956 12436 11751 12437 11752 12438 11753 12439 11754 12440 1882 12441 164 12442 11757 12443 11758 12444 12444 12445 11759 12446 11760 12447 287 12448 11762 12449 2 12450 583 12451 11765 12452 5356 12453 11767 12454 11768 12455 12455 12456 12456 12457 12457 12458 12458 12459 12459 12460 12460 12461 12461 12462 12462 12463 868 12464 12464 12465 12465 12466 12466 12467 12467 12468 12468 12469 12469 12470 12470 12471 12471 12472 12472 12473 12473 12474 12474 12475 12475 12476 12476 12477 12477 12478 12478 12479 12479 12480 12480 12481 12481 12482 12482 12483 12483 12484 12484 12485 12485 12486 12486 12487 12487 12488 12488 12489 12489 12490 12490 12491 12491 12492 12492 12493 12493 12494 12494 12495 12495 12496 12496 12497 12497 12498 12498 12499 12499 12500 12500 12501 12501 12502 12502 12503 12503 12504 12504 12505 12505 12506 12506 12507 12507 12508 12508 12509 12509 12510 12510 12511 12511 12512 12512 12513 12513 12514 12514 12515 12515 12516 12516 12517 12517 12518 12518 12519 12519 12520 12520 12521 12521 12522 12522 12523 12523 12524 12524 12525 12525 12526 12526 12527 12527 12528 12528 12529 12529 12530 12530 12531 12531 12532 12532 12533 12533 12534 12534 12535 12535 12536 12536 12537 12537 12538 12538 12539 12539 12540 12540 12541 12541 12542 12542 12543 12543 12544 11 12545 12545 12546 12546 12547 12547 12548 2 12549 583 12550 12550 12551 956 12552 12552 12553 12553 12554 989 12555 161 12556 164 12557 12557 12558 12558 12559 1505 12560 2 12561 2 12562 2 12563 12563 12564 2 12565 12565 12566 2 12567 2 12568 2 12569 2 12570 2 12571 2 12572 2 12573 2 12574 2 12575 12575 12576 12576 12577 12577 12578 12578 12579 12579 12580 12580 12581 12581 12582 12582 12583 12583 12584 12584 12585 12585 12586 12586 12587 12587 12588 22 12589 12589 12590 4 12591 11 12592 12592 12593 90 12594 11 12595 12595 12596 101 12597 12597 12598 1569 12599 12599 12600 12600 12601 12601 12602 101 12603 868 12604 11 12605 12605 12606 956 12607 12607 12608 12608 12609 989 12610 161 12611 164 12612 12612 12613 12613 12614 12614 12615 868 12616 11 12617 12617 12618 12618 12619 12619 12620 364 12621 364 12622 367 12623 441 12624 12624 12625 12625 12626 12626 12627 12627 12628 2 12629 583 12630 12630 12631 12631 12632 12632 12633 12633 12634 12634 12635 12635 12636 2 12637 583 12638 12638 12639 12639 12640 142 12641 11 12642 144 12643 164 12644 165 12645 167 12646 169 12647 77 12648 128 12649 172 12650 12650 12651 12651 12652 142 12653 11 12654 144 12655 164 12656 2270 12657 2272 12658 2273 12659 2274 12660 2 12661 2 12662 3 12663 4 12664 2 12665 2 12666 2 12667 2 12668 11 12669 2 12670 2 12671 530 12672 2 12673 532 12674 22 12675 534 12676 535 12677 2 12678 11 12679 11 12680 539 12681 540 12682 541 12683 22 12684 2 12685 2 12686 2 12687 11 12688 547 12689 11 12690 549 12691 11 12692 551 12693 552 12694 553 12695 554 12696 60 12697 556 12698 2 12699 2 12700 559 12701 11 12702 2 12703 2 12704 69 12705 133 12706 565 12707 2 12708 2 12709 568 12710 2 12711 570 12712 571 12713 2 12714 86 12715 87 12716 90 12717 136 12718 11 12719 578 12720 579 12721 101 12722 581 12723 582 12724 12724 12725 106 12726 106 12727 586 12728 587 12729 588 12730 589 12731 590 12732 591 12733 592 12734 593 12735 594 12736 595 12737 596 12738 101 12739 2 12740 599 12741 2 12742 2 12743 2 12744 2 12745 2 12746 2 12747 2 12748 2 12749 312 12750 2 12751 2 12752 2 12753 2 12754 2 12755 318 12756 2 12757 2 12758 2 12759 2 12760 2 12761 2 12762 2 12763 2 12764 2 12765 2 12766 2 12767 2 12768 2 12769 2 12770 2 12771 2 12772 631 12773 632 12774 633 12775 634 12776 635 12777 636 12778 637 12779 638 12780 639 12781 640 12782 641 12783 642 12784 643 12785 644 12786 645 12787 646 12788 4 12789 2 12790 649 12791 650 12792 651 12793 652 12794 653 12795 654 12796 655 12797 656 12798 657 12799 658 12800 659 12801 364 12802 364 12803 40 12804 367 12805 441 12806 665 12807 666 12808 667 12809 668 12810 669 12811 2 12812 583 12813 672 12814 673 12815 40 12816 675 12817 287 12818 677 12819 678 12820 2 12821 583 12822 681 12823 682 12824 683 12825 684 12826 685 12827 686 12828 2 12829 751 12830 2 12831 11770 12832 12444 12833 12455 12834 12456 12835 12457 12836 12458 12837 12459 12838 12460 12839 12461 12840 12462 12841 868 12842 12464 12843 12465 12844 12466 12845 12467 12846 12468 12847 12469 12848 12470 12849 12471 12850 12472 12851 12473 12852 12474 12853 12475 12854 12476 12855 12477 12856 12478 12857 12479 12858 12480 12859 12481 12860 12482 12861 12483 12862 12484 12863 12485 12864 12486 12865 12487 12866 12488 12867 12489 12868 12490 12869 12491 12870 12492 12871 12493 12872 12494 12873 12495 12874 12496 12875 12497 12876 12498 12877 12499 12878 12500 12879 12501 12880 12502 12881 12503 12882 12504 12883 12505 12884 12506 12885 12507 12886 12508 12887 12509 12888 12510 12889 12511 12890 12512 12891 12513 12892 12514 12893 12515 12894 12516 12895 12517 12896 12518 12897 12519 12898 12520 12899 12521 12900 12522 12901 12523 12902 12524 12903 12525 12904 12526 12905 12527 12906 12528 12907 12529 12908 12530 12909 12531 12910 12532 12911 12533 12912 12534 12913 12535 12914 12536 12915 12537 12916 12538 12917 12539 12918 12540 12919 12541 12920 12542 12921 12543 12922 11 12923 12545 12924 12546 12925 12547 12926 2 12927 583 12928 12550 12929 956 12930 12552 12931 12553 12932 989 12933 161 12934 164 12935 12557 12936 12558 12937 1505 12938 2 12939 2 12940 2 12941 12563 12942 2 12943 12565 12944 2 12945 2 12946 2 12947 2 12948 2 12949 2 12950 2 12951 2 12952 2 12953 12575 12954 12576 12955 12577 12956 12578 12957 12579 12958 12580 12959 12581 12960 12582 12961 12583 12962 12584 12963 12585 12964 12586 12965 12587 12966 22 12967 12589 12968 4 12969 11 12970 12592 12971 90 12972 11 12973 12595 12974 101 12975 12597 12976 1569 12977 12599 12978 12600 12979 12601 12980 101 12981 868 12982 11 12983 12605 12984 956 12985 12607 12986 12608 12987 989 12988 161 12989 164 12990 12612 12991 12613 12992 12614 12993 868 12994 11 12995 12617 12996 12618 12997 12619 12998 364 12999 364 13000 367 13001 441 13002 12624 13003 12625 13004 12626 13005 12627 13006 2 13007 583 13008 12630 13009 12631 13010 12632 13011 12633 13012 12634 13013 12635 13014 2 13015 583 13016 12638 13017 12639 13018 142 13019 11 13020 144 13021 164 13022 165 13023 167 13024 169 13025 77 13026 128 13027 172 13028 12650 13029 12651 13030 142 13031 11 13032 144 13033 164 13034 2270 13035 2272 13036 2273 13037 2274 13038 2 13039 12724 13040 13040 13041 13040 13042 868 13043 13043 13044 11 13045 13045 13046 2 13047 583 13048 13048 13049 13049 13050 13050 13051 2756 13052 13052 13053 13053 13054 364 13055 364 13056 367 13057 441 13058 13058 13059 13059 13060 13060 13061 13061 13062 2 13063 583 13064 13064 13065 13065 13066 13066 13067 13067 13068 2827 13069 13069 13070 441 13071 13071 13072 2 13073 13073 13074 13074 13075 2 13076 3 13077 4 13078 2 13079 2 13080 2 13081 2 13082 11 13083 2 13084 2 13085 13085 13086 2 13087 13087 13088 22 13089 13089 13090 13090 13091 2 13092 11 13093 11 13094 13094 13095 13095 13096 13096 13097 22 13098 2 13099 2 13100 2 13101 11 13102 13102 13103 11 13104 13104 13105 11 13106 13106 13107 13107 13108 13108 13109 13109 13110 60 13111 13111 13112 2 13113 2 13114 13114 13115 11 13116 2 13117 2 13118 69 13119 133 13120 13120 13121 2 13122 2 13123 13123 13124 2 13125 13125 13126 13126 13127 2 13128 86 13129 87 13130 90 13131 136 13132 11 13133 13133 13134 13134 13135 101 13136 13136 13137 13137 13138 13138 13139 106 13140 106 13141 13141 13142 13142 13143 13143 13144 13144 13145 13145 13146 13146 13147 13147 13148 13148 13149 13149 13150 13150 13151 13151 13152 101 13153 2 13154 13154 13155 2 13156 2 13157 2 13158 2 13159 2 13160 2 13161 2 13162 2 13163 312 13164 2 13165 2 13166 2 13167 2 13168 2 13169 318 13170 2 13171 2 13172 2 13173 2 13174 2 13175 2 13176 2 13177 2 13178 2 13179 2 13180 2 13181 2 13182 2 13183 2 13184 2 13185 2 13186 13186 13187 13187 13188 13188 13189 13189 13190 13190 13191 13191 13192 13192 13193 13193 13194 13194 13195 13195 13196 13196 13197 13197 13198 13198 13199 13199 13200 13200 13201 13201 13202 4 13203 2 13204 13204 13205 13205 13206 13206 13207 13207 13208 13208 13209 13209 13210 13210 13211 13211 13212 13212 13213 13213 13214 13214 13215 364 13216 364 13217 40 13218 367 13219 441 13220 13220 13221 13221 13222 13222 13223 13223 13224 13224 13225 2 13226 583 13227 13227 13228 13228 13229 40 13230 13230 13231 287 13232 13232 13233 13233 13234 2 13235 583 13236 13236 13237 13237 13238 13238 13239 13239 13240 13240 13241 13241 13242 2 13243 751 13244 2 13245 13245 13246 13246 13247 13247 13248 13248 13249 13249 13250 13250 13251 13251 13252 13252 13253 13253 13254 13254 13255 868 13256 13256 13257 13257 13258 13258 13259 13259 13260 13260 13261 13261 13262 13262 13263 13263 13264 13264 13265 13265 13266 13266 13267 13267 13268 13268 13269 13269 13270 13270 13271 13271 13272 13272 13273 13273 13274 13274 13275 13275 13276 13276 13277 13277 13278 13278 13279 13279 13280 13280 13281 13281 13282 13282 13283 13283 13284 13284 13285 13285 13286 13286 13287 13287 13288 13288 13289 13289 13290 13290 13291 13291 13292 13292 13293 13293 13294 13294 13295 13295 13296 13296 13297 13297 13298 13298 13299 13299 13300 13300 13301 13301 13302 13302 13303 13303 13304 13304 13305 13305 13306 13306 13307 13307 13308 13308 13309 13309 13310 13310 13311 13311 13312 13312 13313 13313 13314 13314 13315 13315 13316 13316 13317 13317 13318 13318 13319 13319 13320 13320 13321 13321 13322 13322 13323 13323 13324 13324 13325 13325 13326 13326 13327 13327 13328 13328 13329 13329 13330 13330 13331 13331 13332 13332 13333 13333 13334 13334 13335 13335 13336 11 13337 13337 13338 13338 13339 13339 13340 2 13341 583 13342 13342 13343 956 13344 13344 13345 13345 13346 989 13347 161 13348 164 13349 13349 13350 13350 13351 1505 13352 2 13353 2 13354 2 13355 13355 13356 2 13357 13357 13358 2 13359 2 13360 2 13361 2 13362 2 13363 2 13364 2 13365 2 13366 2 13367 13367 13368 13368 13369 13369 13370 13370 13371 13371 13372 13372 13373 13373 13374 13374 13375 13375 13376 13376 13377 13377 13378 13378 13379 13379 13380 22 13381 13381 13382 4 13383 11 13384 13384 13385 90 13386 11 13387 13387 13388 101 13389 13389 13390 1569 13391 13391 13392 13392 13393 13393 13394 101 13395 868 13396 11 13397 13397 13398 956 13399 13399 13400 13400 13401 989 13402 161 13403 164 13404 13404 13405 13405 13406 13406 13407 868 13408 11 13409 13409 13410 13410 13411 13411 13412 364 13413 364 13414 367 13415 441 13416 13416 13417 13417 13418 13418 13419 13419 13420 2 13421 583 13422 13422 13423 13423 13424 13424 13425 13425 13426 13426 13427 13427 13428 2 13429 583 13430 13430 13431 13431 13432 142 13433 11 13434 144 13435 164 13436 165 13437 167 13438 169 13439 77 13440 128 13441 172 13442 13442 13443 13443 13444 142 13445 11 13446 144 13447 164 13448 2270 13449 2272 13450 2273 13451 2274 13452 2 13453 13453 13454 13454 13455 2 13456 13456 13457 868 13458 13458 13459 11 13460 13460 13461 2 13462 583 13463 13463 13464 13464 13465 13465 13466 2756 13467 13467 13468 13468 13469 364 13470 364 13471 367 13472 441 13473 13473 13474 13474 13475 13475 13476 13476 13477 2 13478 583 13479 13479 13480 13480 13481 13481 13482 13482 13483 2827 13484 13484 13485 441 13486 13486 13487 2 13488 13488 13489 13489 13490 13138 13491 2 13492 13492 13493 13493 13494 13494 13495 13495 13496 13496 13497 2918 13498 13498 13499 69 13500 13500 13501 13501 13502 13502 13503 13503 13504 13504 13505 13505 13506 13506 13507 2 13508 13508 13509 13509 13510 13510 13511 13511 13512 13512 13513 13513 13514 2960 13515 2962 13516 13516 13517 13517 13518 13518 13519 13519 13520 13520 13521 364 13522 364 13523 40 13524 367 13525 441 13526 13526 13527 13527 13528 13528 13529 13529 13530 13530 13531 40 13532 13532 13533 13533 13534 2983 13535 2986 13536 2988 13537 2990 13538 77 13539 2992 13540 2993 13541 13541 13542 287 13543 13543 13544 2 13545 287 13546 77 13547 3959 13548 2852 13549 13549 13550 1882 13551 164 13552 2072 13553 2 13554 13554 13555 2 13556 161 13557 164 13558 13558 13559 13559 13560 13560 13561 13561 13562 13562 13563 956 13564 13564 13565 13565 13566 13566 13567 989 13568 161 13569 164 13570 13570 13571 13571 13572 2 13573 287 13574 13574 13575 2 13576 287 13577 13577 13578 13578 13579 13579 13580 956 13581 13581 13582 13582 13583 13583 13584 989 13585 5285 13586 161 13587 164 13588 13588 13589 13589 13590 956 13591 13591 13592 13592 13593 13593 13594 13594 13595 1882 13596 164 13597 13597 13598 13598 13599 13599 13600 13600 13601 287 13602 13602 13603 2 13604 583 13605 13605 13606 5356 13607 13607 13608 13608 13609 13609 13610 13610 13611 13611 13612 13612 13613 2918 13614 13614 13615 69 13616 13616 13617 13617 13618 13618 13619 13619 13620 13620 13621 13621 13622 13622 13623 2 13624 13624 13625 13625 13626 13626 13627 13627 13628 13628 13629 13629 13630 2960 13631 2962 13632 13632 13633 13633 13634 13634 13635 13635 13636 13636 13637 364 13638 364 13639 40 13640 367 13641 441 13642 13642 13643 13643 13644 13644 13645 13645 13646 13646 13647 40 13648 13648 13649 13649 13650 2983 13651 2986 13652 2988 13653 2990 13654 77 13655 2992 13656 2993 13657 13657 13658 287 13659 13659 13660 2 13661 287 13662 77 13663 3959 13664 2852 13665 13665 13666 1882 13667 164 13668 2072 13669 2 13670 13670 13671 2 13672 161 13673 164 13674 13674 13675 13675 13676 13676 13677 13677 13678 13678 13679 956 13680 13680 13681 13681 13682 13682 13683 989 13684 161 13685 164 13686 13686 13687 13687 13688 2 13689 287 13690 13690 13691 2 13692 287 13693 13693 13694 13694 13695 13695 13696 956 13697 13697 13698 13698 13699 13699 13700 989 13701 5285 13702 161 13703 164 13704 13704 13705 13705 13706 956 13707 13707 13708 13708 13709 13709 13710 13710 13711 1882 13712 164 13713 13713 13714 13714 13715 13715 13716 13716 13717 287 13718 13718 13719 2 13720 583 13721 13721 13722 5356 13723 13723 13724 13724 13725 13725 13726 868 13727 13043 13728 11 13729 13045 13730 2 13731 583 13732 13048 13733 13049 13734 13050 13735 2756 13736 13052 13737 13053 13738 364 13739 364 13740 367 13741 441 13742 13058 13743 13059 13744 13060 13745 13061 13746 2 13747 583 13748 13064 13749 13065 13750 13066 13751 13067 13752 2827 13753 13069 13754 441 13755 13071 13756 2 13757 13073 13758 13074 13759 2 13760 3 13761 4 13762 2 13763 2 13764 2 13765 2 13766 11 13767 2 13768 2 13769 13085 13770 2 13771 13087 13772 22 13773 13089 13774 13090 13775 2 13776 11 13777 11 13778 13094 13779 13095 13780 13096 13781 22 13782 2 13783 2 13784 2 13785 11 13786 13102 13787 11 13788 13104 13789 11 13790 13106 13791 13107 13792 13108 13793 13109 13794 60 13795 13111 13796 2 13797 2 13798 13114 13799 11 13800 2 13801 2 13802 69 13803 133 13804 13120 13805 2 13806 2 13807 13123 13808 2 13809 13125 13810 13126 13811 2 13812 86 13813 87 13814 90 13815 136 13816 11 13817 13133 13818 13134 13819 101 13820 13136 13821 13137 13822 13138 13823 106 13824 106 13825 13141 13826 13142 13827 13143 13828 13144 13829 13145 13830 13146 13831 13147 13832 13148 13833 13149 13834 13150 13835 13151 13836 101 13837 2 13838 13154 13839 2 13840 2 13841 2 13842 2 13843 2 13844 2 13845 2 13846 2 13847 312 13848 2 13849 2 13850 2 13851 2 13852 2 13853 318 13854 2 13855 2 13856 2 13857 2 13858 2 13859 2 13860 2 13861 2 13862 2 13863 2 13864 2 13865 2 13866 2 13867 2 13868 2 13869 2 13870 13186 13871 13187 13872 13188 13873 13189 13874 13190 13875 13191 13876 13192 13877 13193 13878 13194 13879 13195 13880 13196 13881 13197 13882 13198 13883 13199 13884 13200 13885 13201 13886 4 13887 2 13888 13204 13889 13205 13890 13206 13891 13207 13892 13208 13893 13209 13894 13210 13895 13211 13896 13212 13897 13213 13898 13214 13899 364 13900 364 13901 40 13902 367 13903 441 13904 13220 13905 13221 13906 13222 13907 13223 13908 13224 13909 2 13910 583 13911 13227 13912 13228 13913 40 13914 13230 13915 287 13916 13232 13917 13233 13918 2 13919 583 13920 13236 13921 13237 13922 13238 13923 13239 13924 13240 13925 13241 13926 2 13927 751 13928 2 13929 13245 13930 13246 13931 13247 13932 13248 13933 13249 13934 13250 13935 13251 13936 13252 13937 13253 13938 13254 13939 868 13940 13256 13941 13257 13942 13258 13943 13259 13944 13260 13945 13261 13946 13262 13947 13263 13948 13264 13949 13265 13950 13266 13951 13267 13952 13268 13953 13269 13954 13270 13955 13271 13956 13272 13957 13273 13958 13274 13959 13275 13960 13276 13961 13277 13962 13278 13963 13279 13964 13280 13965 13281 13966 13282 13967 13283 13968 13284 13969 13285 13970 13286 13971 13287 13972 13288 13973 13289 13974 13290 13975 13291 13976 13292 13977 13293 13978 13294 13979 13295 13980 13296 13981 13297 13982 13298 13983 13299 13984 13300 13985 13301 13986 13302 13987 13303 13988 13304 13989 13305 13990 13306 13991 13307 13992 13308 13993 13309 13994 13310 13995 13311 13996 13312 13997 13313 13998 13314 13999 13315 14000 13316 14001 13317 14002 13318 14003 13319 14004 13320 14005 13321 14006 13322 14007 13323 14008 13324 14009 13325 14010 13326 14011 13327 14012 13328 14013 13329 14014 13330 14015 13331 14016 13332 14017 13333 14018 13334 14019 13335 14020 11 14021 13337 14022 13338 14023 13339 14024 2 14025 583 14026 13342 14027 956 14028 13344 14029 13345 14030 989 14031 161 14032 164 14033 13349 14034 13350 14035 1505 14036 2 14037 2 14038 2 14039 13355 14040 2 14041 13357 14042 2 14043 2 14044 2 14045 2 14046 2 14047 2 14048 2 14049 2 14050 2 14051 13367 14052 13368 14053 13369 14054 13370 14055 13371 14056 13372 14057 13373 14058 13374 14059 13375 14060 13376 14061 13377 14062 13378 14063 13379 14064 22 14065 13381 14066 4 14067 11 14068 13384 14069 90 14070 11 14071 13387 14072 101 14073 13389 14074 1569 14075 13391 14076 13392 14077 13393 14078 101 14079 868 14080 11 14081 13397 14082 956 14083 13399 14084 13400 14085 989 14086 161 14087 164 14088 13404 14089 13405 14090 13406 14091 868 14092 11 14093 13409 14094 13410 14095 13411 14096 364 14097 364 14098 367 14099 441 14100 13416 14101 13417 14102 13418 14103 13419 14104 2 14105 583 14106 13422 14107 13423 14108 13424 14109 13425 14110 13426 14111 13427 14112 2 14113 583 14114 13430 14115 13431 14116 142 14117 11 14118 144 14119 164 14120 165 14121 167 14122 169 14123 77 14124 128 14125 172 14126 13442 14127 13443 14128 142 14129 11 14130 144 14131 164 14132 2270 14133 2272 14134 2273 14135 2274 14136 2 14137 13453 14138 13454 14139 2 14140 13456 14141 868 14142 13458 14143 11 14144 13460 14145 2 14146 583 14147 13463 14148 13464 14149 13465 14150 2756 14151 13467 14152 13468 14153 364 14154 364 14155 367 14156 441 14157 13473 14158 13474 14159 13475 14160 13476 14161 2 14162 583 14163 13479 14164 13480 14165 13481 14166 13482 14167 2827 14168 13484 14169 441 14170 13486 14171 2 14172 13488 14173 13489 14174 13138 14175 2 14176 13492 14177 13493 14178 13494 14179 13495 14180 13496 14181 2918 14182 13498 14183 69 14184 13500 14185 13501 14186 13502 14187 13503 14188 13504 14189 13505 14190 13506 14191 2 14192 13508 14193 13509 14194 13510 14195 13511 14196 13512 14197 13513 14198 2960 14199 2962 14200 13516 14201 13517 14202 13518 14203 13519 14204 13520 14205 364 14206 364 14207 40 14208 367 14209 441 14210 13526 14211 13527 14212 13528 14213 13529 14214 13530 14215 40 14216 13532 14217 13533 14218 2983 14219 2986 14220 2988 14221 2990 14222 77 14223 2992 14224 2993 14225 13541 14226 287 14227 13543 14228 2 14229 287 14230 77 14231 3959 14232 2852 14233 13549 14234 1882 14235 164 14236 2072 14237 2 14238 13554 14239 2 14240 161 14241 164 14242 13558 14243 13559 14244 13560 14245 13561 14246 13562 14247 956 14248 13564 14249 13565 14250 13566 14251 989 14252 161 14253 164 14254 13570 14255 13571 14256 2 14257 287 14258 13574 14259 2 14260 287 14261 13577 14262 13578 14263 13579 14264 956 14265 13581 14266 13582 14267 13583 14268 989 14269 5285 14270 161 14271 164 14272 13588 14273 13589 14274 956 14275 13591 14276 13592 14277 13593 14278 13594 14279 1882 14280 164 14281 13597 14282 13598 14283 13599 14284 13600 14285 287 14286 13602 14287 2 14288 583 14289 13605 14290 5356 14291 13607 14292 13608 14293 13609 14294 13610 14295 13611 14296 13612 14297 2918 14298 13614 14299 69 14300 13616 14301 13617 14302 13618 14303 13619 14304 13620 14305 13621 14306 13622 14307 2 14308 13624 14309 13625 14310 13626 14311 13627 14312 13628 14313 13629 14314 2960 14315 2962 14316 13632 14317 13633 14318 13634 14319 13635 14320 13636 14321 364 14322 364 14323 40 14324 367 14325 441 14326 13642 14327 13643 14328 13644 14329 13645 14330 13646 14331 40 14332 13648 14333 13649 14334 2983 14335 2986 14336 2988 14337 2990 14338 77 14339 2992 14340 2993 14341 13657 14342 287 14343 13659 14344 2 14345 287 14346 77 14347 3959 14348 2852 14349 13665 14350 1882 14351 164 14352 2072 14353 2 14354 13670 14355 2 14356 161 14357 164 14358 13674 14359 13675 14360 13676 14361 13677 14362 13678 14363 956 14364 13680 14365 13681 14366 13682 14367 989 14368 161 14369 164 14370 13686 14371 13687 14372 2 14373 287 14374 13690 14375 2 14376 287 14377 13693 14378 13694 14379 13695 14380 956 14381 13697 14382 13698 14383 13699 14384 989 14385 5285 14386 161 14387 164 14388 13704 14389 13705 14390 956 14391 13707 14392 13708 14393 13709 14394 13710 14395 1882 14396 164 14397 13713 14398 13714 14399 13715 14400 13716 14401 287 14402 13718 14403 2 14404 583 14405 13721 14406 5356 14407 13723 14408 13724 14409 14409 14410 40 14411 14411 14412 287 14413 14413 14414 14414 14415 2 14416 583 14417 14417 14418 14418 14419 14419 14420 14420 14421 14421 14422 14422 14423 2 14424 751 14425 2 14426 14426 14427 14427 14428 14428 14429 14429 14430 14430 14431 14431 14432 14432 14433 14433 14434 14434 14435 14435 14436 868 14437 14437 14438 14438 14439 14439 14440 14440 14441 14441 14442 14442 14443 14443 14444 14444 14445 14445 14446 14446 14447 14447 14448 14448 14449 14449 14450 14450 14451 14451 14452 14452 14453 14453 14454 14454 14455 14455 14456 14456 14457 14457 14458 14458 14459 14459 14460 14460 14461 14461 14462 14462 14463 14463 14464 14464 14465 14465 14466 14466 14467 14467 14468 14468 14469 14469 14470 14470 14471 14471 14472 14472 14473 14473 14474 14474 14475 14475 14476 14476 14477 14477 14478 14478 14479 14479 14480 14480 14481 14481 14482 14482 14483 14483 14484 14484 14485 14485 14486 14486 14487 14487 14488 14488 14489 14489 14490 14490 14491 14491 14492 14492 14493 14493 14494 14494 14495 14495 14496 14496 14497 14497 14498 14498 14499 14499 14500 14500 14501 14501 14502 14502 14503 14503 14504 14504 14505 14505 14506 14506 14507 14507 14508 14508 14509 14509 14510 14510 14511 14511 14512 14512 14513 14513 14514 14514 14515 14515 14516 14516 14517 11 14518 14518 14519 14519 14520 14520 14521 2 14522 583 14523 14523 14524 956 14525 14525 14526 14526 14527 989 14528 161 14529 164 14530 14530 14531 14531 14532 1505 14533 2 14534 2 14535 2 14536 14536 14537 2 14538 14538 14539 2 14540 2 14541 2 14542 2 14543 2 14544 2 14545 2 14546 2 14547 2 14548 14548 14549 14549 14550 14550 14551 14551 14552 14552 14553 14553 14554 14554 14555 14555 14556 14556 14557 14557 14558 14558 14559 14559 14560 14560 14561 22 14562 14562 14563 4 14564 11 14565 14565 14566 90 14567 11 14568 14568 14569 101 14570 14570 14571 1569 14572 14572 14573 14573 14574 14574 14575 101 14576 868 14577 11 14578 14578 14579 956 14580 14580 14581 14581 14582 989 14583 161 14584 164 14585 14585 14586 14586 14587 14587 14588 868 14589 11 14590 14590 14591 14591 14592 14592 14593 364 14594 364 14595 367 14596 441 14597 14597 14598 14598 14599 14599 14600 14600 14601 2 14602 583 14603 14603 14604 14604 14605 14605 14606 14606 14607 14607 14608 14608 14609 2 14610 583 14611 14611 14612 14612 14613 14613 14614 142 14615 11 14616 144 14617 164 14618 165 14619 167 14620 169 14621 77 14622 128 14623 172 14624 14624 14625 14625 14626 142 14627 11 14628 144 14629 164 14630 2270 14631 2272 14632 2273 14633 2274 14634 2 14635 2 14636 3 14637 4 14638 2 14639 2 14640 2 14641 2 14642 11 14643 2 14644 2 14645 378 14646 2 14647 380 14648 22 14649 382 14650 383 14651 2 14652 11 14653 11 14654 387 14655 388 14656 389 14657 22 14658 2 14659 2 14660 2 14661 11 14662 395 14663 11 14664 397 14665 11 14666 399 14667 400 14668 401 14669 402 14670 60 14671 404 14672 2 14673 2 14674 407 14675 11 14676 2 14677 2 14678 69 14679 133 14680 413 14681 2 14682 2 14683 416 14684 2 14685 418 14686 419 14687 2 14688 86 14689 87 14690 90 14691 136 14692 11 14693 426 14694 427 14695 101 14696 429 14697 430 14698 14698 14699 106 14700 106 14701 434 14702 435 14703 436 14704 437 14705 438 14706 439 14707 440 14708 442 14709 443 14710 444 14711 445 14712 101 14713 2 14714 448 14715 2 14716 2 14717 2 14718 2 14719 2 14720 2 14721 2 14722 2 14723 312 14724 2 14725 2 14726 2 14727 2 14728 2 14729 318 14730 2 14731 2 14732 2 14733 2 14734 2 14735 2 14736 2 14737 2 14738 2 14739 2 14740 2 14741 2 14742 2 14743 2 14744 2 14745 2 14746 480 14747 481 14748 482 14749 483 14750 484 14751 485 14752 486 14753 487 14754 488 14755 489 14756 490 14757 491 14758 492 14759 493 14760 494 14761 495 14762 4 14763 2 14764 498 14765 499 14766 500 14767 501 14768 502 14769 503 14770 504 14771 505 14772 506 14773 507 14774 508 14775 364 14776 364 14777 40 14778 367 14779 441 14780 514 14781 515 14782 516 14783 517 14784 518 14785 2 14786 583 14787 13725 14788 14409 14789 40 14790 14411 14791 287 14792 14413 14793 14414 14794 2 14795 583 14796 14417 14797 14418 14798 14419 14799 14420 14800 14421 14801 14422 14802 2 14803 751 14804 2 14805 14426 14806 14427 14807 14428 14808 14429 14809 14430 14810 14431 14811 14432 14812 14433 14813 14434 14814 14435 14815 868 14816 14437 14817 14438 14818 14439 14819 14440 14820 14441 14821 14442 14822 14443 14823 14444 14824 14445 14825 14446 14826 14447 14827 14448 14828 14449 14829 14450 14830 14451 14831 14452 14832 14453 14833 14454 14834 14455 14835 14456 14836 14457 14837 14458 14838 14459 14839 14460 14840 14461 14841 14462 14842 14463 14843 14464 14844 14465 14845 14466 14846 14467 14847 14468 14848 14469 14849 14470 14850 14471 14851 14472 14852 14473 14853 14474 14854 14475 14855 14476 14856 14477 14857 14478 14858 14479 14859 14480 14860 14481 14861 14482 14862 14483 14863 14484 14864 14485 14865 14486 14866 14487 14867 14488 14868 14489 14869 14490 14870 14491 14871 14492 14872 14493 14873 14494 14874 14495 14875 14496 14876 14497 14877 14498 14878 14499 14879 14500 14880 14501 14881 14502 14882 14503 14883 14504 14884 14505 14885 14506 14886 14507 14887 14508 14888 14509 14889 14510 14890 14511 14891 14512 14892 14513 14893 14514 14894 14515 14895 14516 14896 11 14897 14518 14898 14519 14899 14520 14900 2 14901 583 14902 14523 14903 956 14904 14525 14905 14526 14906 989 14907 161 14908 164 14909 14530 14910 14531 14911 1505 14912 2 14913 2 14914 2 14915 14536 14916 2 14917 14538 14918 2 14919 2 14920 2 14921 2 14922 2 14923 2 14924 2 14925 2 14926 2 14927 14548 14928 14549 14929 14550 14930 14551 14931 14552 14932 14553 14933 14554 14934 14555 14935 14556 14936 14557 14937 14558 14938 14559 14939 14560 14940 22 14941 14562 14942 4 14943 11 14944 14565 14945 90 14946 11 14947 14568 14948 101 14949 14570 14950 1569 14951 14572 14952 14573 14953 14574 14954 101 14955 868 14956 11 14957 14578 14958 956 14959 14580 14960 14581 14961 989 14962 161 14963 164 14964 14585 14965 14586 14966 14587 14967 868 14968 11 14969 14590 14970 14591 14971 14592 14972 364 14973 364 14974 367 14975 441 14976 14597 14977 14598 14978 14599 14979 14600 14980 2 14981 583 14982 14603 14983 14604 14984 14605 14985 14606 14986 14607 14987 14608 14988 2 14989 583 14990 14611 14991 14613 14992 142 14993 11 14994 144 14995 164 14996 165 14997 167 14998 169 14999 77 15000 128 15001 172 15002 14624 15003 14625 15004 142 15005 11 15006 144 15007 164 15008 2270 15009 2272 15010 2273 15011 2274 15012 2 15013 14698 15014 15014 15015 15014 15016 868 15017 15017 15018 11 15019 15019 15020 2 15021 583 15022 15022 15023 15023 15024 15024 15025 2756 15026 15026 15027 15027 15028 364 15029 364 15030 367 15031 441 15032 15032 15033 15033 15034 15034 15035 15035 15036 2 15037 583 15038 15038 15039 15039 15040 15040 15041 15041 15042 2827 15043 15043 15044 441 15045 15045 15046 2 15047 15047 15048 15048 15049 2 15050 3 15051 4 15052 2 15053 2 15054 2 15055 2 15056 11 15057 2 15058 2 15059 15059 15060 2 15061 15061 15062 22 15063 15063 15064 15064 15065 2 15066 11 15067 11 15068 15068 15069 15069 15070 15070 15071 22 15072 2 15073 2 15074 2 15075 11 15076 15076 15077 11 15078 15078 15079 11 15080 15080 15081 15081 15082 15082 15083 15083 15084 60 15085 15085 15086 2 15087 2 15088 15088 15089 11 15090 2 15091 2 15092 69 15093 133 15094 15094 15095 2 15096 2 15097 15097 15098 2 15099 15099 15100 15100 15101 2 15102 86 15103 87 15104 90 15105 136 15106 11 15107 15107 15108 15108 15109 101 15110 15110 15111 15111 15112 15112 15113 106 15114 106 15115 15115 15116 15116 15117 15117 15118 15118 15119 15119 15120 15120 15121 15121 15122 15122 15123 15123 15124 15124 15125 15125 15126 101 15127 2 15128 15128 15129 2 15130 2 15131 2 15132 2 15133 2 15134 2 15135 2 15136 2 15137 312 15138 2 15139 2 15140 2 15141 2 15142 2 15143 318 15144 2 15145 2 15146 2 15147 2 15148 2 15149 2 15150 2 15151 2 15152 2 15153 2 15154 2 15155 2 15156 2 15157 2 15158 2 15159 2 15160 15160 15161 15161 15162 15162 15163 15163 15164 15164 15165 15165 15166 15166 15167 15167 15168 15168 15169 15169 15170 15170 15171 15171 15172 15172 15173 15173 15174 15174 15175 15175 15176 4 15177 2 15178 15178 15179 15179 15180 15180 15181 15181 15182 15182 15183 15183 15184 15184 15185 15185 15186 15186 15187 15187 15188 15188 15189 364 15190 364 15191 40 15192 367 15193 441 15194 15194 15195 15195 15196 15196 15197 15197 15198 15198 15199 2 15200 583 15201 15201 15202 15202 15203 40 15204 15204 15205 287 15206 15206 15207 15207 15208 2 15209 583 15210 15210 15211 15211 15212 15212 15213 15213 15214 15214 15215 15215 15216 2 15217 751 15218 2 15219 15219 15220 15220 15221 15221 15222 15222 15223 15223 15224 15224 15225 15225 15226 15226 15227 15227 15228 15228 15229 868 15230 15230 15231 15231 15232 15232 15233 15233 15234 15234 15235 15235 15236 15236 15237 15237 15238 15238 15239 15239 15240 15240 15241 15241 15242 15242 15243 15243 15244 15244 15245 15245 15246 15246 15247 15247 15248 15248 15249 15249 15250 15250 15251 15251 15252 15252 15253 15253 15254 15254 15255 15255 15256 15256 15257 15257 15258 15258 15259 15259 15260 15260 15261 15261 15262 15262 15263 15263 15264 15264 15265 15265 15266 15266 15267 15267 15268 15268 15269 15269 15270 15270 15271 15271 15272 15272 15273 15273 15274 15274 15275 15275 15276 15276 15277 15277 15278 15278 15279 15279 15280 15280 15281 15281 15282 15282 15283 15283 15284 15284 15285 15285 15286 15286 15287 15287 15288 15288 15289 15289 15290 15290 15291 15291 15292 15292 15293 15293 15294 15294 15295 15295 15296 15296 15297 15297 15298 15298 15299 15299 15300 15300 15301 15301 15302 15302 15303 15303 15304 15304 15305 15305 15306 15306 15307 15307 15308 15308 15309 15309 15310 11 15311 15311 15312 15312 15313 15313 15314 2 15315 583 15316 15316 15317 956 15318 15318 15319 15319 15320 989 15321 161 15322 164 15323 15323 15324 15324 15325 1505 15326 2 15327 2 15328 2 15329 15329 15330 2 15331 15331 15332 2 15333 2 15334 2 15335 2 15336 2 15337 2 15338 2 15339 2 15340 2 15341 15341 15342 15342 15343 15343 15344 15344 15345 15345 15346 15346 15347 15347 15348 15348 15349 15349 15350 15350 15351 15351 15352 15352 15353 15353 15354 22 15355 15355 15356 4 15357 11 15358 15358 15359 90 15360 11 15361 15361 15362 101 15363 15363 15364 1569 15365 15365 15366 15366 15367 15367 15368 101 15369 868 15370 11 15371 15371 15372 956 15373 15373 15374 15374 15375 989 15376 161 15377 164 15378 15378 15379 15379 15380 15380 15381 868 15382 11 15383 15383 15384 15384 15385 15385 15386 364 15387 364 15388 367 15389 441 15390 15390 15391 15391 15392 15392 15393 15393 15394 2 15395 583 15396 15396 15397 15397 15398 15398 15399 15399 15400 15400 15401 15401 15402 2 15403 583 15404 15404 15405 15405 15406 142 15407 11 15408 144 15409 164 15410 165 15411 167 15412 169 15413 77 15414 128 15415 172 15416 15416 15417 15417 15418 142 15419 11 15420 144 15421 164 15422 2270 15423 2272 15424 2273 15425 2274 15426 2 15427 15427 15428 15428 15429 2 15430 15430 15431 868 15432 15432 15433 11 15434 15434 15435 2 15436 583 15437 15437 15438 15438 15439 15439 15440 2756 15441 15441 15442 15442 15443 364 15444 364 15445 367 15446 441 15447 15447 15448 15448 15449 15449 15450 15450 15451 2 15452 583 15453 15453 15454 15454 15455 15455 15456 15456 15457 2827 15458 15458 15459 441 15460 15460 15461 2 15462 15462 15463 15463 15464 15112 15465 2 15466 15466 15467 15467 15468 15468 15469 15469 15470 15470 15471 2918 15472 15472 15473 69 15474 15474 15475 15475 15476 15476 15477 15477 15478 15478 15479 15479 15480 15480 15481 2 15482 15482 15483 15483 15484 15484 15485 15485 15486 15486 15487 15487 15488 2960 15489 2962 15490 15490 15491 15491 15492 15492 15493 15493 15494 15494 15495 364 15496 364 15497 40 15498 367 15499 441 15500 15500 15501 15501 15502 15502 15503 15503 15504 15504 15505 40 15506 15506 15507 15507 15508 2983 15509 2986 15510 2988 15511 2990 15512 77 15513 2992 15514 2993 15515 15515 15516 287 15517 15517 15518 2 15519 287 15520 77 15521 3959 15522 2852 15523 15523 15524 1882 15525 164 15526 2072 15527 2 15528 15528 15529 2 15530 161 15531 164 15532 15532 15533 15533 15534 15534 15535 15535 15536 15536 15537 956 15538 15538 15539 15539 15540 15540 15541 989 15542 161 15543 164 15544 15544 15545 15545 15546 2 15547 287 15548 15548 15549 2 15550 287 15551 15551 15552 15552 15553 15553 15554 956 15555 15555 15556 15556 15557 15557 15558 989 15559 5285 15560 161 15561 164 15562 15562 15563 15563 15564 956 15565 15565 15566 15566 15567 15567 15568 15568 15569 1882 15570 164 15571 15571 15572 15572 15573 15573 15574 15574 15575 287 15576 15576 15577 2 15578 583 15579 15579 15580 5356 15581 15581 15582 15582 15583 15583 15584 15584 15585 15585 15586 15586 15587 2918 15588 15588 15589 69 15590 15590 15591 15591 15592 15592 15593 15593 15594 15594 15595 15595 15596 15596 15597 2 15598 15598 15599 15599 15600 15600 15601 15601 15602 15602 15603 15603 15604 2960 15605 2962 15606 15606 15607 15607 15608 15608 15609 15609 15610 15610 15611 364 15612 364 15613 40 15614 367 15615 441 15616 15616 15617 15617 15618 15618 15619 15619 15620 15620 15621 40 15622 15622 15623 15623 15624 2983 15625 2986 15626 2988 15627 2990 15628 77 15629 2992 15630 2993 15631 15631 15632 287 15633 15633 15634 2 15635 287 15636 77 15637 3959 15638 2852 15639 15639 15640 1882 15641 164 15642 2072 15643 2 15644 15644 15645 2 15646 161 15647 164 15648 15648 15649 15649 15650 15650 15651 15651 15652 15652 15653 956 15654 15654 15655 15655 15656 15656 15657 989 15658 161 15659 164 15660 15660 15661 15661 15662 2 15663 287 15664 15664 15665 2 15666 287 15667 15667 15668 15668 15669 15669 15670 956 15671 15671 15672 15672 15673 15673 15674 989 15675 5285 15676 161 15677 164 15678 15678 15679 15679 15680 956 15681 15681 15682 15682 15683 15683 15684 15684 15685 1882 15686 164 15687 15687 15688 15688 15689 15689 15690 15690 15691 287 15692 15692 15693 2 15694 583 15695 15695 15696 5356 15697 15697 15698 15698 15699 868 15700 15017 15701 11 15702 15019 15703 2 15704 583 15705 15022 15706 15023 15707 15024 15708 2756 15709 15026 15710 15027 15711 364 15712 364 15713 367 15714 441 15715 15032 15716 15033 15717 15034 15718 15035 15719 2 15720 583 15721 15038 15722 15039 15723 15040 15724 15041 15725 2827 15726 15043 15727 441 15728 15045 15729 2 15730 15047 15731 15048 15732 2 15733 3 15734 4 15735 2 15736 2 15737 2 15738 2 15739 11 15740 2 15741 2 15742 15059 15743 2 15744 15061 15745 22 15746 15063 15747 15064 15748 2 15749 11 15750 11 15751 15068 15752 15069 15753 15070 15754 22 15755 2 15756 2 15757 2 15758 11 15759 15076 15760 11 15761 15078 15762 11 15763 15080 15764 15081 15765 15082 15766 15083 15767 60 15768 15085 15769 2 15770 2 15771 15088 15772 11 15773 2 15774 2 15775 69 15776 133 15777 15094 15778 2 15779 2 15780 15097 15781 2 15782 15099 15783 15100 15784 2 15785 86 15786 87 15787 90 15788 136 15789 11 15790 15107 15791 15108 15792 101 15793 15110 15794 15111 15795 15112 15796 106 15797 106 15798 15115 15799 15116 15800 15117 15801 15118 15802 15119 15803 15120 15804 15121 15805 15122 15806 15123 15807 15124 15808 15125 15809 101 15810 2 15811 15128 15812 2 15813 2 15814 2 15815 2 15816 2 15817 2 15818 2 15819 2 15820 312 15821 2 15822 2 15823 2 15824 2 15825 2 15826 318 15827 2 15828 2 15829 2 15830 2 15831 2 15832 2 15833 2 15834 2 15835 2 15836 2 15837 2 15838 2 15839 2 15840 2 15841 2 15842 2 15843 15160 15844 15161 15845 15162 15846 15163 15847 15164 15848 15165 15849 15166 15850 15167 15851 15168 15852 15169 15853 15170 15854 15171 15855 15172 15856 15173 15857 15174 15858 15175 15859 4 15860 2 15861 15178 15862 15179 15863 15180 15864 15181 15865 15182 15866 15183 15867 15184 15868 15185 15869 15186 15870 15187 15871 15188 15872 364 15873 364 15874 40 15875 367 15876 441 15877 15194 15878 15195 15879 15196 15880 15197 15881 15198 15882 2 15883 583 15884 15201 15885 15202 15886 40 15887 15204 15888 287 15889 15206 15890 15207 15891 2 15892 583 15893 15210 15894 15211 15895 15212 15896 15213 15897 15214 15898 15215 15899 2 15900 751 15901 2 15902 15219 15903 15220 15904 15221 15905 15222 15906 15223 15907 15224 15908 15225 15909 15226 15910 15227 15911 15228 15912 868 15913 15230 15914 15231 15915 15232 15916 15233 15917 15234 15918 15235 15919 15236 15920 15237 15921 15238 15922 15239 15923 15240 15924 15241 15925 15242 15926 15243 15927 15244 15928 15245 15929 15246 15930 15247 15931 15248 15932 15249 15933 15250 15934 15251 15935 15252 15936 15253 15937 15254 15938 15255 15939 15256 15940 15257 15941 15258 15942 15259 15943 15260 15944 15261 15945 15262 15946 15263 15947 15264 15948 15265 15949 15266 15950 15267 15951 15268 15952 15269 15953 15270 15954 15271 15955 15272 15956 15273 15957 15274 15958 15275 15959 15276 15960 15277 15961 15278 15962 15279 15963 15280 15964 15281 15965 15282 15966 15283 15967 15284 15968 15285 15969 15286 15970 15287 15971 15288 15972 15289 15973 15290 15974 15291 15975 15292 15976 15293 15977 15294 15978 15295 15979 15296 15980 15297 15981 15298 15982 15299 15983 15300 15984 15301 15985 15302 15986 15303 15987 15304 15988 15305 15989 15306 15990 15307 15991 15308 15992 15309 15993 11 15994 15311 15995 15312 15996 15313 15997 2 15998 583 15999 15316 16000 956 16001 15318 16002 15319 16003 989 16004 161 16005 164 16006 15323 16007 15324 16008 1505 16009 2 16010 2 16011 2 16012 15329 16013 2 16014 15331 16015 2 16016 2 16017 2 16018 2 16019 2 16020 2 16021 2 16022 2 16023 2 16024 15341 16025 15342 16026 15343 16027 15344 16028 15345 16029 15346 16030 15347 16031 15348 16032 15349 16033 15350 16034 15351 16035 15352 16036 15353 16037 22 16038 15355 16039 4 16040 11 16041 15358 16042 90 16043 11 16044 15361 16045 101 16046 15363 16047 1569 16048 15365 16049 15366 16050 15367 16051 101 16052 868 16053 11 16054 15371 16055 956 16056 15373 16057 15374 16058 989 16059 161 16060 164 16061 15378 16062 15379 16063 15380 16064 868 16065 11 16066 15383 16067 15384 16068 15385 16069 364 16070 364 16071 367 16072 441 16073 15390 16074 15391 16075 15392 16076 15393 16077 2 16078 583 16079 15396 16080 15397 16081 15398 16082 15399 16083 15400 16084 15401 16085 2 16086 583 16087 15404 16088 15405 16089 142 16090 11 16091 144 16092 164 16093 165 16094 167 16095 169 16096 77 16097 128 16098 172 16099 15416 16100 15417 16101 142 16102 11 16103 144 16104 164 16105 2270 16106 2272 16107 2273 16108 2274 16109 2 16110 15427 16111 15428 16112 2 16113 15430 16114 868 16115 15432 16116 11 16117 15434 16118 2 16119 583 16120 15437 16121 15438 16122 15439 16123 2756 16124 15441 16125 15442 16126 364 16127 364 16128 367 16129 441 16130 15447 16131 15448 16132 15449 16133 15450 16134 2 16135 583 16136 15453 16137 15454 16138 15455 16139 15456 16140 2827 16141 15458 16142 441 16143 15460 16144 2 16145 15462 16146 15463 16147 15112 16148 2 16149 15466 16150 15467 16151 15468 16152 15469 16153 15470 16154 2918 16155 15472 16156 69 16157 15474 16158 15475 16159 15476 16160 15477 16161 15478 16162 15479 16163 15480 16164 2 16165 15482 16166 15483 16167 15484 16168 15485 16169 15486 16170 15487 16171 2960 16172 2962 16173 15490 16174 15491 16175 15492 16176 15493 16177 15494 16178 364 16179 364 16180 40 16181 367 16182 441 16183 15500 16184 15501 16185 15502 16186 15503 16187 15504 16188 40 16189 15506 16190 15507 16191 2983 16192 2986 16193 2988 16194 2990 16195 77 16196 2992 16197 2993 16198 15515 16199 287 16200 15517 16201 2 16202 287 16203 77 16204 3959 16205 2852 16206 15523 16207 1882 16208 164 16209 2072 16210 2 16211 15528 16212 2 16213 161 16214 164 16215 15532 16216 15533 16217 15534 16218 15535 16219 15536 16220 956 16221 15538 16222 15539 16223 15540 16224 989 16225 161 16226 164 16227 15544 16228 15545 16229 2 16230 287 16231 15548 16232 2 16233 287 16234 15551 16235 15552 16236 15553 16237 956 16238 15555 16239 15556 16240 15557 16241 989 16242 5285 16243 161 16244 164 16245 15562 16246 15563 16247 956 16248 15565 16249 15566 16250 15567 16251 15568 16252 1882 16253 164 16254 15571 16255 15572 16256 15573 16257 15574 16258 287 16259 15576 16260 2 16261 583 16262 15579 16263 5356 16264 15581 16265 15582 16266 15583 16267 15584 16268 15585 16269 15586 16270 2918 16271 15588 16272 69 16273 15590 16274 15591 16275 15592 16276 15593 16277 15594 16278 15595 16279 15596 16280 2 16281 15598 16282 15599 16283 15600 16284 15601 16285 15602 16286 15603 16287 2960 16288 2962 16289 15606 16290 15607 16291 15608 16292 15609 16293 15610 16294 364 16295 364 16296 40 16297 367 16298 441 16299 15616 16300 15617 16301 15618 16302 15619 16303 15620 16304 40 16305 15622 16306 15623 16307 2983 16308 2986 16309 2988 16310 2990 16311 77 16312 2992 16313 2993 16314 15631 16315 287 16316 15633 16317 2 16318 287 16319 77 16320 3959 16321 2852 16322 15639 16323 1882 16324 164 16325 2072 16326 2 16327 15644 16328 2 16329 161 16330 164 16331 15648 16332 15649 16333 15650 16334 15651 16335 15652 16336 956 16337 15654 16338 15655 16339 15656 16340 989 16341 161 16342 164 16343 15660 16344 15661 16345 2 16346 287 16347 15664 16348 2 16349 287 16350 15667 16351 15668 16352 15669 16353 956 16354 15671 16355 15672 16356 15673 16357 989 16358 5285 16359 161 16360 164 16361 15678 16362 15679 16363 956 16364 15681 16365 15682 16366 15683 16367 15684 16368 1882 16369 164 16370 15687 16371 15688 16372 15689 16373 15690 16374 287 16375 15692 16376 2 16377 583 16378 15695 16379 5356 16380 15697 16381 15698 16382 16382 16383 16383 16384 16384 16385 16385 16386 2 16387 583 16388 16388 16389 16389 16390 40 16391 16391 16392 287 16393 16393 16394 16394 16395 2 16396 583 16397 16397 16398 16398 16399 16399 16400 16400 16401 16401 16402 16402 16403 2 16404 751 16405 2 16406 16406 16407 16407 16408 16408 16409 16409 16410 16410 16411 16411 16412 16412 16413 16413 16414 16414 16415 16415 16416 868 16417 16417 16418 16418 16419 16419 16420 16420 16421 16421 16422 16422 16423 16423 16424 16424 16425 16425 16426 16426 16427 16427 16428 16428 16429 16429 16430 16430 16431 16431 16432 16432 16433 16433 16434 16434 16435 16435 16436 16436 16437 16437 16438 16438 16439 16439 16440 16440 16441 16441 16442 16442 16443 16443 16444 16444 16445 16445 16446 16446 16447 16447 16448 16448 16449 16449 16450 16450 16451 16451 16452 16452 16453 16453 16454 16454 16455 16455 16456 16456 16457 16457 16458 16458 16459 16459 16460 16460 16461 16461 16462 16462 16463 16463 16464 16464 16465 16465 16466 16466 16467 16467 16468 16468 16469 16469 16470 16470 16471 16471 16472 16472 16473 16473 16474 16474 16475 16475 16476 16476 16477 16477 16478 16478 16479 16479 16480 16480 16481 16481 16482 16482 16483 16483 16484 16484 16485 16485 16486 16486 16487 16487 16488 16488 16489 16489 16490 16490 16491 16491 16492 16492 16493 16493 16494 16494 16495 16495 16496 16496 16497 11 16498 16498 16499 16499 16500 16500 16501 2 16502 583 16503 16503 16504 956 16505 16505 16506 16506 16507 989 16508 161 16509 164 16510 16510 16511 16511 16512 1505 16513 2 16514 2 16515 2 16516 16516 16517 2 16518 16518 16519 2 16520 2 16521 2 16522 2 16523 2 16524 2 16525 2 16526 2 16527 2 16528 16528 16529 16529 16530 16530 16531 16531 16532 16532 16533 16533 16534 16534 16535 16535 16536 16536 16537 16537 16538 16538 16539 16539 16540 16540 16541 22 16542 16542 16543 4 16544 11 16545 16545 16546 90 16547 11 16548 16548 16549 101 16550 16550 16551 1569 16552 16552 16553 16553 16554 16554 16555 101 16556 868 16557 11 16558 16558 16559 956 16560 16560 16561 16561 16562 989 16563 161 16564 164 16565 16565 16566 16566 16567 16567 16568 868 16569 11 16570 16570 16571 16571 16572 16572 16573 364 16574 364 16575 367 16576 441 16577 16577 16578 16578 16579 16579 16580 16580 16581 2 16582 583 16583 16583 16584 16584 16585 16585 16586 16586 16587 16587 16588 16588 16589 2 16590 583 16591 16591 16592 16592 16593 142 16594 11 16595 144 16596 164 16597 165 16598 167 16599 169 16600 77 16601 128 16602 172 16603 16603 16604 16604 16605 142 16606 11 16607 144 16608 164 16609 2270 16610 2272 16611 2273 16612 2274 16613 2 16614 2 16615 3 16616 4 16617 2 16618 2 16619 2 16620 2 16621 11 16622 2 16623 2 16624 234 16625 2 16626 236 16627 22 16628 238 16629 239 16630 2 16631 11 16632 11 16633 243 16634 244 16635 245 16636 22 16637 2 16638 2 16639 2 16640 11 16641 251 16642 11 16643 253 16644 11 16645 255 16646 256 16647 257 16648 258 16649 60 16650 260 16651 2 16652 2 16653 263 16654 11 16655 2 16656 2 16657 69 16658 133 16659 269 16660 2 16661 2 16662 272 16663 2 16664 274 16665 275 16666 2 16667 86 16668 87 16669 90 16670 136 16671 11 16672 282 16673 283 16674 101 16675 285 16676 286 16677 16677 16678 106 16679 106 16680 290 16681 291 16682 292 16683 293 16684 294 16685 295 16686 296 16687 297 16688 298 16689 299 16690 300 16691 101 16692 2 16693 303 16694 2 16695 2 16696 2 16697 2 16698 2 16699 2 16700 2 16701 2 16702 312 16703 2 16704 2 16705 2 16706 2 16707 2 16708 318 16709 2 16710 2 16711 2 16712 2 16713 2 16714 2 16715 2 16716 2 16717 2 16718 2 16719 2 16720 2 16721 2 16722 2 16723 2 16724 2 16725 335 16726 336 16727 337 16728 338 16729 339 16730 340 16731 341 16732 342 16733 343 16734 344 16735 345 16736 346 16737 347 16738 348 16739 349 16740 350 16741 4 16742 2 16743 353 16744 354 16745 355 16746 356 16747 357 16748 358 16749 359 16750 360 16751 361 16752 362 16753 363 16754 364 16755 364 16756 40 16757 367 16758 441 16759 14612 16760 16382 16761 16383 16762 16384 16763 16385 16764 2 16765 583 16766 16388 16767 16389 16768 40 16769 16391 16770 287 16771 16393 16772 16394 16773 2 16774 583 16775 16397 16776 16398 16777 16399 16778 16400 16779 16401 16780 16402 16781 2 16782 751 16783 2 16784 16406 16785 16407 16786 16408 16787 16409 16788 16410 16789 16411 16790 16412 16791 16413 16792 16414 16793 16415 16794 868 16795 16417 16796 16418 16797 16419 16798 16420 16799 16421 16800 16422 16801 16423 16802 16424 16803 16425 16804 16426 16805 16427 16806 16428 16807 16429 16808 16430 16809 16431 16810 16432 16811 16433 16812 16434 16813 16435 16814 16436 16815 16437 16816 16438 16817 16439 16818 16440 16819 16441 16820 16442 16821 16443 16822 16444 16823 16445 16824 16446 16825 16447 16826 16448 16827 16449 16828 16450 16829 16451 16830 16452 16831 16453 16832 16454 16833 16455 16834 16456 16835 16457 16836 16458 16837 16459 16838 16460 16839 16461 16840 16462 16841 16463 16842 16464 16843 16465 16844 16466 16845 16467 16846 16468 16847 16469 16848 16470 16849 16471 16850 16472 16851 16473 16852 16474 16853 16475 16854 16476 16855 16477 16856 16478 16857 16479 16858 16480 16859 16481 16860 16482 16861 16483 16862 16484 16863 16485 16864 16486 16865 16487 16866 16488 16867 16489 16868 16490 16869 16491 16870 16492 16871 16493 16872 16494 16873 16495 16874 16496 16875 11 16876 16498 16877 16499 16878 16500 16879 2 16880 583 16881 16503 16882 956 16883 16505 16884 16506 16885 989 16886 161 16887 164 16888 16510 16889 16511 16890 1505 16891 2 16892 2 16893 2 16894 16516 16895 2 16896 16518 16897 2 16898 2 16899 2 16900 2 16901 2 16902 2 16903 2 16904 2 16905 2 16906 16528 16907 16529 16908 16530 16909 16531 16910 16532 16911 16533 16912 16534 16913 16535 16914 16536 16915 16537 16916 16538 16917 16539 16918 16540 16919 22 16920 16542 16921 4 16922 11 16923 16545 16924 90 16925 11 16926 16548 16927 101 16928 16550 16929 1569 16930 16552 16931 16553 16932 16554 16933 101 16934 868 16935 11 16936 16558 16937 956 16938 16560 16939 16561 16940 989 16941 161 16942 164 16943 16565 16944 16566 16945 16567 16946 868 16947 11 16948 16570 16949 16571 16950 16572 16951 364 16952 364 16953 367 16954 441 16955 16577 16956 16578 16957 16579 16958 16580 16959 2 16960 583 16961 16583 16962 16584 16963 16585 16964 16586 16965 16587 16966 16588 16967 2 16968 583 16969 16591 16970 16592 16971 142 16972 11 16973 144 16974 164 16975 165 16976 167 16977 169 16978 77 16979 128 16980 172 16981 16603 16982 16604 16983 142 16984 11 16985 144 16986 164 16987 2270 16988 2272 16989 2273 16990 2274 16991 2 16992 16677 16993 16993 16994 16993 16995 868 16996 16996 16997 11 16998 16998 16999 2 17000 583 17001 17001 17002 17002 17003 17003 17004 2756 17005 17005 17006 17006 17007 364 17008 364 17009 367 17010 441 17011 17011 17012 17012 17013 17013 17014 17014 17015 2 17016 583 17017 17017 17018 17018 17019 17019 17020 17020 17021 2827 17022 17022 17023 441 17024 17024 17025 2 17026 17026 17027 17027 17028 2 17029 3 17030 4 17031 2 17032 2 17033 2 17034 2 17035 11 17036 2 17037 2 17038 17038 17039 2 17040 17040 17041 22 17042 17042 17043 17043 17044 2 17045 11 17046 11 17047 17047 17048 17048 17049 17049 17050 22 17051 2 17052 2 17053 2 17054 11 17055 17055 17056 11 17057 17057 17058 11 17059 17059 17060 17060 17061 17061 17062 17062 17063 60 17064 17064 17065 2 17066 2 17067 17067 17068 11 17069 2 17070 2 17071 69 17072 133 17073 17073 17074 2 17075 2 17076 17076 17077 2 17078 17078 17079 17079 17080 2 17081 86 17082 87 17083 90 17084 136 17085 11 17086 17086 17087 17087 17088 101 17089 17089 17090 17090 17091 17091 17092 106 17093 106 17094 17094 17095 17095 17096 17096 17097 17097 17098 17098 17099 17099 17100 17100 17101 17101 17102 17102 17103 17103 17104 17104 17105 101 17106 2 17107 17107 17108 2 17109 2 17110 2 17111 2 17112 2 17113 2 17114 2 17115 2 17116 312 17117 2 17118 2 17119 2 17120 2 17121 2 17122 318 17123 2 17124 2 17125 2 17126 2 17127 2 17128 2 17129 2 17130 2 17131 2 17132 2 17133 2 17134 2 17135 2 17136 2 17137 2 17138 2 17139 17139 17140 17140 17141 17141 17142 17142 17143 17143 17144 17144 17145 17145 17146 17146 17147 17147 17148 17148 17149 17149 17150 17150 17151 17151 17152 17152 17153 17153 17154 17154 17155 4 17156 2 17157 17157 17158 17158 17159 17159 17160 17160 17161 17161 17162 17162 17163 17163 17164 17164 17165 17165 17166 17166 17167 17167 17168 364 17169 364 17170 40 17171 367 17172 441 17173 17173 17174 17174 17175 17175 17176 17176 17177 17177 17178 2 17179 583 17180 17180 17181 17181 17182 40 17183 17183 17184 287 17185 17185 17186 17186 17187 2 17188 583 17189 17189 17190 17190 17191 17191 17192 17192 17193 17193 17194 17194 17195 2 17196 751 17197 2 17198 17198 17199 17199 17200 17200 17201 17201 17202 17202 17203 17203 17204 17204 17205 17205 17206 17206 17207 17207 17208 868 17209 17209 17210 17210 17211 17211 17212 17212 17213 17213 17214 17214 17215 17215 17216 17216 17217 17217 17218 17218 17219 17219 17220 17220 17221 17221 17222 17222 17223 17223 17224 17224 17225 17225 17226 17226 17227 17227 17228 17228 17229 17229 17230 17230 17231 17231 17232 17232 17233 17233 17234 17234 17235 17235 17236 17236 17237 17237 17238 17238 17239 17239 17240 17240 17241 17241 17242 17242 17243 17243 17244 17244 17245 17245 17246 17246 17247 17247 17248 17248 17249 17249 17250 17250 17251 17251 17252 17252 17253 17253 17254 17254 17255 17255 17256 17256 17257 17257 17258 17258 17259 17259 17260 17260 17261 17261 17262 17262 17263 17263 17264 17264 17265 17265 17266 17266 17267 17267 17268 17268 17269 17269 17270 17270 17271 17271 17272 17272 17273 17273 17274 17274 17275 17275 17276 17276 17277 17277 17278 17278 17279 17279 17280 17280 17281 17281 17282 17282 17283 17283 17284 17284 17285 17285 17286 17286 17287 17287 17288 17288 17289 11 17290 17290 17291 17291 17292 17292 17293 2 17294 583 17295 17295 17296 956 17297 17297 17298 17298 17299 989 17300 161 17301 164 17302 17302 17303 17303 17304 1505 17305 2 17306 2 17307 2 17308 17308 17309 2 17310 17310 17311 2 17312 2 17313 2 17314 2 17315 2 17316 2 17317 2 17318 2 17319 2 17320 17320 17321 17321 17322 17322 17323 17323 17324 17324 17325 17325 17326 17326 17327 17327 17328 17328 17329 17329 17330 17330 17331 17331 17332 17332 17333 22 17334 17334 17335 4 17336 11 17337 17337 17338 90 17339 11 17340 17340 17341 101 17342 17342 17343 1569 17344 17344 17345 17345 17346 17346 17347 101 17348 868 17349 11 17350 17350 17351 956 17352 17352 17353 17353 17354 989 17355 161 17356 164 17357 17357 17358 17358 17359 17359 17360 868 17361 11 17362 17362 17363 17363 17364 17364 17365 364 17366 364 17367 367 17368 441 17369 17369 17370 17370 17371 17371 17372 17372 17373 2 17374 583 17375 17375 17376 17376 17377 17377 17378 17378 17379 17379 17380 17380 17381 2 17382 583 17383 17383 17384 17384 17385 142 17386 11 17387 144 17388 164 17389 165 17390 167 17391 169 17392 77 17393 128 17394 172 17395 17395 17396 17396 17397 142 17398 11 17399 144 17400 164 17401 2270 17402 2272 17403 2273 17404 2274 17405 2 17406 17406 17407 17407 17408 2 17409 17409 17410 868 17411 17411 17412 11 17413 17413 17414 2 17415 583 17416 17416 17417 17417 17418 17418 17419 2756 17420 17420 17421 17421 17422 364 17423 364 17424 367 17425 441 17426 17426 17427 17427 17428 17428 17429 17429 17430 2 17431 583 17432 17432 17433 17433 17434 17434 17435 17435 17436 2827 17437 17437 17438 441 17439 17439 17440 2 17441 17441 17442 17442 17443 17091 17444 2 17445 17445 17446 17446 17447 17447 17448 17448 17449 17449 17450 2918 17451 17451 17452 69 17453 17453 17454 17454 17455 17455 17456 17456 17457 17457 17458 17458 17459 17459 17460 2 17461 17461 17462 17462 17463 17463 17464 17464 17465 17465 17466 17466 17467 2960 17468 2962 17469 17469 17470 17470 17471 17471 17472 17472 17473 17473 17474 364 17475 364 17476 40 17477 367 17478 441 17479 17479 17480 17480 17481 17481 17482 17482 17483 17483 17484 40 17485 17485 17486 17486 17487 2983 17488 2986 17489 2988 17490 2990 17491 77 17492 2992 17493 2993 17494 17494 17495 287 17496 17496 17497 2 17498 287 17499 77 17500 3959 17501 2852 17502 17502 17503 1882 17504 164 17505 2072 17506 2 17507 17507 17508 2 17509 161 17510 164 17511 17511 17512 17512 17513 17513 17514 17514 17515 17515 17516 956 17517 17517 17518 17518 17519 17519 17520 989 17521 161 17522 164 17523 17523 17524 17524 17525 2 17526 287 17527 17527 17528 2 17529 287 17530 17530 17531 17531 17532 17532 17533 956 17534 17534 17535 17535 17536 17536 17537 989 17538 5285 17539 161 17540 164 17541 17541 17542 17542 17543 956 17544 17544 17545 17545 17546 17546 17547 17547 17548 1882 17549 164 17550 17550 17551 17551 17552 17552 17553 17553 17554 287 17555 17555 17556 2 17557 583 17558 17558 17559 5356 17560 17560 17561 17561 17562 17562 17563 17563 17564 17564 17565 17565 17566 2918 17567 17567 17568 69 17569 17569 17570 17570 17571 17571 17572 17572 17573 17573 17574 17574 17575 17575 17576 2 17577 17577 17578 17578 17579 17579 17580 17580 17581 17581 17582 17582 17583 2960 17584 2962 17585 17585 17586 17586 17587 17587 17588 17588 17589 17589 17590 364 17591 364 17592 40 17593 367 17594 441 17595 17595 17596 17596 17597 17597 17598 17598 17599 17599 17600 40 17601 17601 17602 17602 17603 2983 17604 2986 17605 2988 17606 2990 17607 77 17608 2992 17609 2993 17610 17610 17611 287 17612 17612 17613 2 17614 287 17615 77 17616 3959 17617 2852 17618 17618 17619 1882 17620 164 17621 2072 17622 2 17623 17623 17624 2 17625 161 17626 164 17627 17627 17628 17628 17629 17629 17630 17630 17631 17631 17632 956 17633 17633 17634 17634 17635 17635 17636 989 17637 161 17638 164 17639 17639 17640 17640 17641 2 17642 287 17643 17643 17644 2 17645 287 17646 17646 17647 17647 17648 17648 17649 956 17650 17650 17651 17651 17652 17652 17653 989 17654 5285 17655 161 17656 164 17657 17657 17658 17658 17659 956 17660 17660 17661 17661 17662 17662 17663 17663 17664 1882 17665 164 17666 17666 17667 17667 17668 17668 17669 17669 17670 287 17671 17671 17672 2 17673 583 17674 17674 17675 5356 17676 17676 17677 17677 17678 26 17679 17679 17680 15 17681 15 17682 11 17683 26 17684 2 17685 11 17686 11 17687 40 17688 16 17689 17689 17690 15 17691 70 17692 71 17693 11 17694 77 17695 11 17696 2 17697 11 17698 17698 17699 89 17700 105 17701 17701 17702 211 17703 17703 17704 128 17705 172 17706 17706 17707 26 17708 17708 17709 17709 17710 1080 17711 11 17712 7507 17713 17713 17714 2 17715 287 17716 1080 17717 17717 17718 17718 17719 17719 17720 15 17721 16 17722 15 17723 11 17724 26 17725 2 17726 16 17727 11 17728 11 17729 40 17730 16 17731 11 17732 11 17733 17733 17734 16 17735 15 17736 70 17737 71 17738 11 17739 16 17740 77 17741 11 17742 2 17743 11 17744 17744 17745 89 17746 17746 17747 203 17748 17748 17749 100 17750 17750 17751 105 17752 108 17753 17753 17754 17754 17755 17755 17756 17756 17757 17757 17758 17758 17759 17759 17760 128 17761 172 17762 17762 17763 26 17764 11 17765 2 17766 287 17767 26 17768 17768 17769 17769 17770 17770 17771 17771 17772 17772 17773 17773 17774 17774 17775 17775 17776 142 17777 11 17778 144 17779 164 17780 167 17781 169 17782 77 17783 128 17784 172 17785 17785 17786 17786 17787 17787 17788 16 17789 17789 17790 2986 17791 2988 17792 2990 17793 77 17794 2992 17795 2993 17796 17796 17797 2756 17798 17798 17799 17799 17800 17800 17801 17801 17802 1080 17803 7507 17804 17804 17805 17805 17806 17806 17807 17807 17808 15 17809 16 17810 15 17811 11 17812 26 17813 2 17814 7507 17815 16 17816 11 17817 11 17818 40 17819 16 17820 11 17821 17821 17822 16 17823 15 17824 70 17825 71 17826 11 17827 16 17828 77 17829 11 17830 2 17831 11 17832 17832 17833 89 17834 17834 17835 203 17836 17836 17837 100 17838 17838 17839 105 17840 108 17841 17841 17842 17842 17843 17755 17844 17844 17845 17845 17846 17846 17847 17847 17848 128 17849 172 17850 17850 17851 26 17852 17852 17853 17773 17854 142 17855 11 17856 144 17857 164 17858 167 17859 169 17860 77 17861 128 17862 172 17863 17863 17864 17864 17865 17865 17866 16 17867 2986 17868 2988 17869 2990 17870 77 17871 2992 17872 2993 17873 17873 17874 17874 17875 1080 17876 7507 17877 17877 17878 17805 17879 17806 17880 7507 17881 17881 17882 2 17883 287 17884 17884 17885 17885 17886 17886 17887 17887 17888 17888 17889 17889 17890 15 17891 16 17892 15 17893 11 17894 26 17895 2 17896 7507 17897 16 17898 11 17899 11 17900 40 17901 16 17902 11 17903 11 17904 17904 17905 16 17906 15 17907 70 17908 71 17909 11 17910 16 17911 77 17912 11 17913 2 17914 11 17915 17915 17916 89 17917 17917 17918 203 17919 17919 17920 100 17921 17921 17922 105 17923 108 17924 17924 17925 17925 17926 17926 17927 17927 17928 17928 17929 17929 17930 17930 17931 128 17932 172 17933 17933 17934 26 17935 11 17936 2 17937 287 17938 26 17939 17939 17940 15 17941 15 17942 11 17943 26 17944 2 17945 11 17946 11 17947 40 17948 16 17949 17949 17950 15 17951 70 17952 71 17953 11 17954 77 17955 11 17956 2 17957 11 17958 17958 17959 89 17960 105 17961 17961 17962 17926 17963 17963 17964 128 17965 172 17966 17966 17967 26 17968 17968 17969 17969 17970 1080 17971 7507 17972 17972 17973 17805 17974 17806 17975 7507 17976 17976 17977 2 17978 287 17979 17884 17980 17885 17981 17981 17982 17982 17983 17888 17984 7507 17985 17985 17986 105 17987 17987 17988 22 17989 4 17990 11 17991 2776 17992 90 17993 11 17994 2779 17995 101 17996 2781 17997 1569 17998 2783 17999 2784 18000 18000 18001 161 18002 18002 18003 101 18004 868 18005 11 18006 2789 18007 956 18008 2791 18009 2792 18010 989 18011 161 18012 164 18013 2796 18014 2797 18015 18015 18016 164 18017 7507 18018 18018 18019 18019 18020 2798 18021 868 18022 11 18023 2801 18024 2802 18025 2803 18026 364 18027 364 18028 367 18029 441 18030 2808 18031 2809 18032 2810 18033 2811 18034 2 18035 583 18036 2814 18037 2815 18038 2816 18039 2817 18040 2818 18041 2819 18042 2 18043 583 18044 2822 18045 128 18046 172 18047 18047 18048 18048 18049 11 18050 77 18051 18051 18052 128 18053 172 18054 18054 18055 13 18056 18056 18057 18057 18058 18058 18059 18059 18060 18060 18061 18061 18062 18062 18063 18063 18064 18064 18065 164 18066 13 18067 18067 18068 77 18069 128 18070 172 18071 18071 18072 2 18073 287 18074 18074 18075 18075 18076 18076 18077 18077 18078 18078 18079 18079 18080 2 18081 18081 18082 18082 18083 18083 18084 18083 18085 18085 18086 18086 18087 15 18088 16 18089 15 18090 11 18091 26 18092 2 18093 16 18094 11 18095 11 18096 40 18097 16 18098 11 18099 11 18100 18100 18101 16 18102 15 18103 70 18104 71 18105 11 18106 16 18107 77 18108 11 18109 2 18110 11 18111 18111 18112 89 18113 18113 18114 203 18115 18115 18116 100 18117 18117 18118 105 18119 108 18120 18120 18121 18121 18122 18122 18123 18123 18124 18124 18125 18125 18126 18126 18127 128 18128 172 18129 18129 18130 26 18131 11 18132 2 18133 287 18134 26 18135 18135 18136 15 18137 15 18138 11 18139 26 18140 2 18141 11 18142 11 18143 40 18144 16 18145 18145 18146 15 18147 70 18148 71 18149 11 18150 77 18151 11 18152 2 18153 11 18154 18154 18155 89 18156 105 18157 18157 18158 18122 18159 18159 18160 128 18161 172 18162 18162 18163 26 18164 18164 18165 18165 18166 1080 18167 7507 18168 18168 18169 17805 18170 17806 18171 7507 18172 18172 18173 2 18174 287 18175 17884 18176 17885 18177 18177 18178 18178 18179 17888 18180 7507 18181 18181 18182 105 18183 18183 18184 22 18185 161 18186 164 18187 7507 18188 18188 18189 128 18190 172 18191 18191 18192 18192 18193 11 18194 77 18195 18195 18196 128 18197 172 18198 18198 18199 13 18200 18061 18201 164 18202 77 18203 128 18204 172 18205 18205 18206 2 18207 287 18208 18074 18209 18075 18210 18210 18211 18077 18212 18078 18213 18079 18214 18214 18215 18079 18216 18216 18217 18217 18218 18218 18219 2 18220 18081 18221 7507 18222 18222 18223 18223 18224 18223 18225 18225 18226 18226 18227 15 18228 16 18229 15 18230 11 18231 26 18232 2 18233 16 18234 11 18235 11 18236 40 18237 16 18238 11 18239 11 18240 18240 18241 16 18242 15 18243 70 18244 71 18245 11 18246 16 18247 77 18248 11 18249 2 18250 11 18251 18251 18252 89 18253 18253 18254 203 18255 18255 18256 100 18257 18257 18258 105 18259 108 18260 18260 18261 18261 18262 18262 18263 18263 18264 18264 18265 18265 18266 18266 18267 128 18268 172 18269 18269 18270 26 18271 11 18272 2 18273 287 18274 26 18275 18275 18276 15 18277 15 18278 11 18279 26 18280 2 18281 11 18282 11 18283 40 18284 16 18285 18285 18286 15 18287 70 18288 71 18289 11 18290 77 18291 11 18292 2 18293 11 18294 18294 18295 89 18296 105 18297 18297 18298 18262 18299 18299 18300 128 18301 172 18302 18302 18303 26 18304 18304 18305 18305 18306 1080 18307 7507 18308 18308 18309 17805 18310 17806 18311 7507 18312 18312 18313 2 18314 287 18315 17884 18316 17885 18317 18317 18318 18318 18319 17888 18320 7507 18321 18321 18322 105 18323 18323 18324 22 18325 161 18326 164 18327 7507 18328 18328 18329 128 18330 172 18331 18331 18332 18332 18333 11 18334 77 18335 18335 18336 128 18337 172 18338 18338 18339 13 18340 18061 18341 164 18342 77 18343 128 18344 172 18345 18345 18346 2 18347 287 18348 18074 18349 18075 18350 18350 18351 18077 18352 18078 18353 18079 18354 18354 18355 18079 18356 18356 18357 18217 18358 18218 18359 7507 18360 18360 18361 18218 18362 7507 18363 18363 18364 18364 18365 18365 18366 18366 18367 18367 18368 3066 18369 3029 18370 18370 18371 18371 18372 2 18373 2 18374 3 18375 4 18376 2 18377 2 18378 2 18379 2 18380 11 18381 2 18382 2 18383 3083 18384 2 18385 3085 18386 22 18387 3087 18388 3088 18389 2 18390 11 18391 11 18392 3092 18393 3093 18394 3094 18395 22 18396 2 18397 2 18398 2 18399 11 18400 3100 18401 11 18402 3102 18403 11 18404 3104 18405 3105 18406 3106 18407 3107 18408 60 18409 3109 18410 2 18411 2 18412 3112 18413 11 18414 2 18415 2 18416 69 18417 133 18418 3118 18419 2 18420 2 18421 3121 18422 2 18423 3123 18424 3124 18425 2 18426 86 18427 87 18428 90 18429 136 18430 11 18431 3131 18432 3132 18433 101 18434 3134 18435 3135 18436 18436 18437 106 18438 106 18439 3139 18440 3140 18441 3141 18442 3142 18443 3143 18444 3144 18445 3145 18446 3146 18447 3147 18448 3148 18449 3149 18450 101 18451 2 18452 3152 18453 2 18454 2 18455 2 18456 2 18457 2 18458 2 18459 2 18460 2 18461 312 18462 2 18463 2 18464 2 18465 2 18466 2 18467 318 18468 2 18469 2 18470 2 18471 2 18472 2 18473 2 18474 2 18475 2 18476 2 18477 2 18478 2 18479 2 18480 2 18481 2 18482 2 18483 2 18484 3184 18485 3185 18486 3186 18487 3187 18488 3188 18489 3189 18490 3190 18491 3191 18492 3192 18493 3193 18494 3194 18495 3195 18496 3196 18497 3197 18498 3198 18499 3199 18500 4 18501 2 18502 3202 18503 3203 18504 3204 18505 3205 18506 3206 18507 3207 18508 3208 18509 3209 18510 3210 18511 3211 18512 3212 18513 364 18514 364 18515 40 18516 367 18517 441 18518 3218 18519 3219 18520 3220 18521 3221 18522 3222 18523 2 18524 583 18525 3225 18526 3226 18527 40 18528 3228 18529 287 18530 3230 18531 3231 18532 2 18533 583 18534 3234 18535 3235 18536 3236 18537 3237 18538 3238 18539 3239 18540 2 18541 751 18542 2 18543 3243 18544 3244 18545 3245 18546 3246 18547 3247 18548 3248 18549 3249 18550 3250 18551 3251 18552 3252 18553 868 18554 3254 18555 3255 18556 3256 18557 3257 18558 3258 18559 3259 18560 3260 18561 3261 18562 3262 18563 3263 18564 3264 18565 3265 18566 3266 18567 3267 18568 3268 18569 3269 18570 3270 18571 3271 18572 3272 18573 3273 18574 3274 18575 3275 18576 3276 18577 3277 18578 3278 18579 3279 18580 3280 18581 3281 18582 3282 18583 3283 18584 3284 18585 3285 18586 3286 18587 3287 18588 3288 18589 3289 18590 3290 18591 3291 18592 3292 18593 3293 18594 3294 18595 3295 18596 3296 18597 3297 18598 3298 18599 3299 18600 3300 18601 3301 18602 3302 18603 3303 18604 3304 18605 3305 18606 3306 18607 3307 18608 3308 18609 3309 18610 3310 18611 3311 18612 3312 18613 3313 18614 3314 18615 3315 18616 3316 18617 3317 18618 3318 18619 3319 18620 3320 18621 3321 18622 3322 18623 3323 18624 3324 18625 3325 18626 3326 18627 3327 18628 3328 18629 3329 18630 3330 18631 3331 18632 3332 18633 3333 18634 11 18635 3335 18636 3336 18637 3337 18638 2 18639 583 18640 3340 18641 956 18642 3342 18643 3343 18644 989 18645 161 18646 164 18647 3347 18648 3348 18649 1505 18650 2 18651 2 18652 2 18653 3353 18654 2 18655 3355 18656 2 18657 2 18658 2 18659 2 18660 2 18661 2 18662 2 18663 2 18664 2 18665 3365 18666 3366 18667 3367 18668 3368 18669 3369 18670 3370 18671 3371 18672 3372 18673 3373 18674 3374 18675 3375 18676 3376 18677 3377 18678 22 18679 3379 18680 4 18681 11 18682 3382 18683 90 18684 11 18685 3385 18686 101 18687 3387 18688 1569 18689 3389 18690 3390 18691 3391 18692 101 18693 868 18694 11 18695 3395 18696 956 18697 3397 18698 3398 18699 989 18700 161 18701 164 18702 3402 18703 3403 18704 3404 18705 868 18706 11 18707 3407 18708 3408 18709 3409 18710 364 18711 364 18712 367 18713 441 18714 3414 18715 3415 18716 3416 18717 3417 18718 2 18719 583 18720 3420 18721 3421 18722 3422 18723 3423 18724 3424 18725 3425 18726 2 18727 583 18728 3428 18729 3429 18730 142 18731 11 18732 144 18733 164 18734 165 18735 167 18736 169 18737 77 18738 128 18739 172 18740 3440 18741 3441 18742 142 18743 11 18744 144 18745 164 18746 2270 18747 2272 18748 2273 18749 2274 18750 2 18751 3451 18752 3452 18753 868 18754 3455 18755 11 18756 3457 18757 2 18758 583 18759 3460 18760 3461 18761 3462 18762 2756 18763 3464 18764 3465 18765 364 18766 364 18767 367 18768 441 18769 3470 18770 3471 18771 3472 18772 3473 18773 2 18774 583 18775 3476 18776 3477 18777 3478 18778 3479 18779 2827 18780 3481 18781 441 18782 3483 18783 2 18784 3485 18785 3486 18786 2 18787 3 18788 4 18789 2 18790 2 18791 2 18792 2 18793 11 18794 2 18795 2 18796 3497 18797 2 18798 3499 18799 22 18800 3501 18801 3502 18802 2 18803 11 18804 11 18805 3506 18806 3507 18807 3508 18808 22 18809 2 18810 2 18811 2 18812 11 18813 3514 18814 11 18815 3516 18816 11 18817 3518 18818 3519 18819 3520 18820 3521 18821 60 18822 3523 18823 2 18824 2 18825 3526 18826 11 18827 2 18828 2 18829 69 18830 133 18831 3532 18832 2 18833 2 18834 3535 18835 2 18836 3537 18837 3538 18838 2 18839 86 18840 87 18841 90 18842 136 18843 11 18844 3545 18845 3546 18846 101 18847 3548 18848 3549 18849 3550 18850 106 18851 106 18852 3553 18853 3554 18854 3555 18855 3556 18856 3557 18857 3558 18858 3559 18859 3560 18860 3561 18861 3562 18862 3563 18863 101 18864 2 18865 3566 18866 2 18867 2 18868 2 18869 2 18870 2 18871 2 18872 2 18873 2 18874 312 18875 2 18876 2 18877 2 18878 2 18879 2 18880 318 18881 2 18882 2 18883 2 18884 2 18885 2 18886 2 18887 2 18888 2 18889 2 18890 2 18891 2 18892 2 18893 2 18894 2 18895 2 18896 2 18897 3598 18898 3599 18899 3600 18900 3601 18901 3602 18902 3603 18903 3604 18904 3605 18905 3606 18906 3607 18907 3608 18908 3609 18909 3610 18910 3611 18911 3612 18912 3613 18913 4 18914 2 18915 3616 18916 3617 18917 3618 18918 3619 18919 3620 18920 3621 18921 3622 18922 3623 18923 3624 18924 3625 18925 3626 18926 364 18927 364 18928 40 18929 367 18930 441 18931 3632 18932 3633 18933 3634 18934 3635 18935 3636 18936 2 18937 583 18938 3639 18939 3640 18940 40 18941 3642 18942 287 18943 3644 18944 3645 18945 2 18946 583 18947 3648 18948 3649 18949 3650 18950 3651 18951 3652 18952 3653 18953 2 18954 751 18955 2 18956 3657 18957 3658 18958 3659 18959 3660 18960 3661 18961 3662 18962 3663 18963 3664 18964 3665 18965 3666 18966 868 18967 3668 18968 3669 18969 3670 18970 3671 18971 3672 18972 3673 18973 3674 18974 3675 18975 3676 18976 3677 18977 3678 18978 3679 18979 3680 18980 3681 18981 3682 18982 3683 18983 3684 18984 3685 18985 3686 18986 3687 18987 3688 18988 3689 18989 3690 18990 3691 18991 3692 18992 3693 18993 3694 18994 3695 18995 3696 18996 3697 18997 3698 18998 3699 18999 3700 19000 3701 19001 3702 19002 3703 19003 3704 19004 3705 19005 3706 19006 3707 19007 3708 19008 3709 19009 3710 19010 3711 19011 3712 19012 3713 19013 3714 19014 3715 19015 3716 19016 3717 19017 3718 19018 3719 19019 3720 19020 3721 19021 3722 19022 3723 19023 3724 19024 3725 19025 3726 19026 3727 19027 3728 19028 3729 19029 3730 19030 3731 19031 3732 19032 3733 19033 3734 19034 3735 19035 3736 19036 3737 19037 3738 19038 3739 19039 3740 19040 3741 19041 3742 19042 3743 19043 3744 19044 3745 19045 3746 19046 3747 19047 11 19048 3749 19049 3750 19050 3751 19051 2 19052 583 19053 3754 19054 956 19055 3756 19056 3757 19057 989 19058 161 19059 164 19060 3761 19061 3762 19062 1505 19063 2 19064 2 19065 2 19066 3767 19067 2 19068 3769 19069 2 19070 2 19071 2 19072 2 19073 2 19074 2 19075 2 19076 2 19077 2 19078 3779 19079 3780 19080 3781 19081 3782 19082 3783 19083 3784 19084 3785 19085 3786 19086 3787 19087 3788 19088 3789 19089 3790 19090 3791 19091 22 19092 3793 19093 4 19094 11 19095 3796 19096 90 19097 11 19098 3799 19099 101 19100 3801 19101 1569 19102 3803 19103 3804 19104 3805 19105 101 19106 868 19107 11 19108 3809 19109 956 19110 3811 19111 3812 19112 989 19113 161 19114 164 19115 3816 19116 3817 19117 3818 19118 868 19119 11 19120 3821 19121 3822 19122 3823 19123 364 19124 364 19125 367 19126 441 19127 3828 19128 3829 19129 3830 19130 3831 19131 2 19132 583 19133 3834 19134 3835 19135 3836 19136 3837 19137 3838 19138 3839 19139 2 19140 583 19141 3842 19142 3843 19143 142 19144 11 19145 144 19146 164 19147 165 19148 167 19149 169 19150 77 19151 128 19152 172 19153 3854 19154 3855 19155 142 19156 11 19157 144 19158 164 19159 2270 19160 2272 19161 2273 19162 2274 19163 2 19164 3865 19165 3866 19166 2 19167 3868 19168 868 19169 3870 19170 11 19171 3872 19172 2 19173 583 19174 3875 19175 3876 19176 3877 19177 2756 19178 3879 19179 3880 19180 364 19181 364 19182 367 19183 441 19184 3885 19185 3886 19186 3887 19187 3888 19188 2 19189 583 19190 3891 19191 3892 19192 3893 19193 3894 19194 2827 19195 3896 19196 441 19197 3898 19198 2 19199 3900 19200 3901 19201 3550 19202 2 19203 3904 19204 3905 19205 3906 19206 3907 19207 3908 19208 2918 19209 3910 19210 69 19211 3912 19212 3913 19213 3914 19214 3915 19215 3916 19216 3917 19217 3918 19218 2 19219 3920 19220 3921 19221 3922 19222 3923 19223 3924 19224 3925 19225 2960 19226 2962 19227 3928 19228 3929 19229 3930 19230 3931 19231 3932 19232 364 19233 364 19234 40 19235 367 19236 441 19237 3938 19238 3939 19239 3940 19240 3941 19241 3942 19242 40 19243 3944 19244 3945 19245 2983 19246 2986 19247 2988 19248 2990 19249 77 19250 2992 19251 2993 19252 3953 19253 287 19254 3955 19255 2 19256 287 19257 77 19258 3959 19259 2852 19260 4325 19261 1882 19262 164 19263 2072 19264 2 19265 5680 19266 2 19267 161 19268 164 19269 5684 19270 5685 19271 5686 19272 5687 19273 5688 19274 956 19275 5690 19276 5691 19277 5692 19278 989 19279 161 19280 164 19281 5696 19282 5697 19283 2 19284 287 19285 5700 19286 2 19287 287 19288 5703 19289 5704 19290 5705 19291 956 19292 5707 19293 5708 19294 5709 19295 989 19296 5285 19297 161 19298 164 19299 5714 19300 5715 19301 956 19302 5717 19303 5718 19304 5719 19305 5720 19306 1882 19307 164 19308 5723 19309 5724 19310 5725 19311 5726 19312 287 19313 5728 19314 2 19315 583 19316 5731 19317 5356 19318 5733 19319 5734 19320 5735 19321 5736 19322 5737 19323 5738 19324 2918 19325 5740 19326 69 19327 5742 19328 5743 19329 5744 19330 5745 19331 5746 19332 5747 19333 5748 19334 2 19335 5750 19336 5751 19337 5752 19338 5753 19339 5754 19340 5755 19341 2960 19342 2962 19343 5758 19344 5759 19345 5760 19346 5761 19347 5762 19348 364 19349 364 19350 40 19351 367 19352 441 19353 5768 19354 5769 19355 5770 19356 5771 19357 5772 19358 40 19359 5774 19360 5775 19361 2983 19362 2986 19363 2988 19364 2990 19365 77 19366 2992 19367 2993 19368 5783 19369 287 19370 5785 19371 2 19372 287 19373 77 19374 3959 19375 2852 19376 5791 19377 1882 19378 164 19379 2072 19380 2 19381 5796 19382 2 19383 161 19384 164 19385 5800 19386 5801 19387 5802 19388 5803 19389 5804 19390 956 19391 5806 19392 5807 19393 5808 19394 989 19395 161 19396 164 19397 5812 19398 5813 19399 2 19400 287 19401 5816 19402 2 19403 287 19404 5819 19405 5820 19406 5821 19407 956 19408 5823 19409 5824 19410 5825 19411 989 19412 5285 19413 161 19414 164 19415 5830 19416 5831 19417 956 19418 5833 19419 5834 19420 5835 19421 5836 19422 1882 19423 164 19424 5839 19425 5840 19426 5841 19427 5842 19428 287 19429 5844 19430 2 19431 583 19432 5847 19433 5356 19434 5849 19435 5850 19436 19436 19437 7507 19438 19438 19439 7507 19440 19440 19441 4 19442 19442 19443 19443 19444 3066 19445 19445 19446 40 19447 19447 19448 19448 19449 19449 19450 2 19451 583 19452 19452 19453 2 19454 2 19455 3 19456 4 19457 2 19458 2 19459 2 19460 2 19461 11 19462 2 19463 2 19464 19464 19465 2 19466 19466 19467 22 19468 19468 19469 19469 19470 2 19471 11 19472 11 19473 19473 19474 19474 19475 19475 19476 22 19477 2 19478 2 19479 2 19480 11 19481 19481 19482 11 19483 19483 19484 11 19485 19485 19486 19486 19487 19487 19488 19488 19489 60 19490 19490 19491 2 19492 2 19493 19493 19494 11 19495 2 19496 2 19497 69 19498 133 19499 19499 19500 2 19501 2 19502 19502 19503 2 19504 19504 19505 19505 19506 2 19507 86 19508 87 19509 90 19510 136 19511 11 19512 19512 19513 19513 19514 101 19515 19515 19516 19516 19517 19517 19518 106 19519 106 19520 19520 19521 19521 19522 19522 19523 19523 19524 19524 19525 19525 19526 19526 19527 19527 19528 19528 19529 19529 19530 19530 19531 101 19532 2 19533 19533 19534 2 19535 2 19536 2 19537 2 19538 2 19539 2 19540 2 19541 2 19542 312 19543 2 19544 2 19545 2 19546 2 19547 2 19548 318 19549 2 19550 2 19551 2 19552 2 19553 2 19554 2 19555 2 19556 2 19557 2 19558 2 19559 2 19560 2 19561 2 19562 2 19563 2 19564 2 19565 19565 19566 19566 19567 19567 19568 19568 19569 19569 19570 19570 19571 19571 19572 19572 19573 19573 19574 19574 19575 19575 19576 19576 19577 19577 19578 19578 19579 19579 19580 19580 19581 4 19582 2 19583 19583 19584 19584 19585 19585 19586 19586 19587 19587 19588 19588 19589 19589 19590 19590 19591 19591 19592 19592 19593 19593 19594 364 19595 364 19596 40 19597 367 19598 441 19599 19599 19600 19600 19601 19601 19602 19602 19603 19603 19604 2 19605 583 19606 19606 19607 19607 19608 40 19609 19609 19610 287 19611 19611 19612 19612 19613 2 19614 583 19615 19615 19616 19616 19617 19617 19618 19618 19619 19619 19620 19620 19621 2 19622 751 19623 2 19624 19624 19625 19625 19626 19626 19627 19627 19628 19628 19629 19629 19630 19630 19631 19631 19632 19632 19633 19633 19634 868 19635 19635 19636 19636 19637 19637 19638 19638 19639 19639 19640 19640 19641 19641 19642 19642 19643 19643 19644 19644 19645 19645 19646 19646 19647 19647 19648 19648 19649 19649 19650 19650 19651 19651 19652 19652 19653 19653 19654 19654 19655 19655 19656 19656 19657 19657 19658 19658 19659 19659 19660 19660 19661 19661 19662 19662 19663 19663 19664 19664 19665 19665 19666 19666 19667 19667 19668 19668 19669 19669 19670 19670 19671 19671 19672 19672 19673 19673 19674 19674 19675 19675 19676 19676 19677 19677 19678 19678 19679 19679 19680 19680 19681 19681 19682 19682 19683 19683 19684 19684 19685 19685 19686 19686 19687 19687 19688 19688 19689 19689 19690 19690 19691 19691 19692 19692 19693 19693 19694 19694 19695 19695 19696 19696 19697 19697 19698 19698 19699 19699 19700 19700 19701 19701 19702 19702 19703 19703 19704 19704 19705 19705 19706 19706 19707 19707 19708 19708 19709 19709 19710 19710 19711 19711 19712 19712 19713 19713 19714 19714 19715 11 19716 19716 19717 19717 19718 19718 19719 2 19720 583 19721 19721 19722 956 19723 19723 19724 19724 19725 989 19726 161 19727 164 19728 19728 19729 19729 19730 1505 19731 2 19732 2 19733 2 19734 19734 19735 2 19736 19736 19737 2 19738 2 19739 2 19740 2 19741 2 19742 2 19743 2 19744 2 19745 2 19746 19746 19747 19747 19748 19748 19749 19749 19750 19750 19751 19751 19752 19752 19753 19753 19754 19754 19755 19755 19756 19756 19757 19757 19758 19758 19759 22 19760 19760 19761 4 19762 11 19763 19763 19764 90 19765 11 19766 19766 19767 101 19768 19768 19769 1569 19770 19770 19771 19771 19772 19772 19773 101 19774 868 19775 11 19776 19776 19777 956 19778 19778 19779 19779 19780 989 19781 161 19782 164 19783 19783 19784 19784 19785 19785 19786 868 19787 11 19788 19788 19789 19789 19790 19790 19791 364 19792 364 19793 367 19794 441 19795 19795 19796 19796 19797 19797 19798 19798 19799 2 19800 583 19801 19801 19802 19802 19803 19803 19804 19804 19805 19805 19806 19806 19807 2 19808 583 19809 19809 19810 19810 19811 142 19812 11 19813 144 19814 164 19815 165 19816 167 19817 169 19818 77 19819 128 19820 172 19821 19821 19822 19822 19823 142 19824 11 19825 144 19826 164 19827 2270 19828 2272 19829 2273 19830 2274 19831 2 19832 2 19833 3 19834 4 19835 2 19836 2 19837 2 19838 2 19839 11 19840 2 19841 2 19842 19464 19843 2 19844 19466 19845 22 19846 19468 19847 19469 19848 2 19849 11 19850 11 19851 19473 19852 19474 19853 19475 19854 22 19855 2 19856 2 19857 2 19858 11 19859 19481 19860 11 19861 19483 19862 11 19863 19485 19864 19486 19865 19487 19866 19488 19867 60 19868 19490 19869 2 19870 2 19871 19493 19872 11 19873 2 19874 2 19875 69 19876 133 19877 19499 19878 2 19879 2 19880 19502 19881 2 19882 19504 19883 19505 19884 2 19885 86 19886 87 19887 90 19888 136 19889 11 19890 19512 19891 19513 19892 101 19893 19515 19894 19516 19895 19895 19896 106 19897 106 19898 19520 19899 19521 19900 19522 19901 19523 19902 19524 19903 19525 19904 19526 19905 19527 19906 19528 19907 19529 19908 19530 19909 101 19910 2 19911 19533 19912 2 19913 2 19914 2 19915 2 19916 2 19917 2 19918 2 19919 2 19920 312 19921 2 19922 2 19923 2 19924 2 19925 2 19926 318 19927 2 19928 2 19929 2 19930 2 19931 2 19932 2 19933 2 19934 2 19935 2 19936 2 19937 2 19938 2 19939 2 19940 2 19941 2 19942 2 19943 19565 19944 19566 19945 19567 19946 19568 19947 19569 19948 19570 19949 19571 19950 19572 19951 19573 19952 19574 19953 19575 19954 19576 19955 19577 19956 19578 19957 19579 19958 19580 19959 4 19960 2 19961 19583 19962 19584 19963 19585 19964 19586 19965 19587 19966 19588 19967 19589 19968 19590 19969 19591 19970 19592 19971 19593 19972 364 19973 364 19974 40 19975 367 19976 441 19977 19599 19978 19600 19979 19601 19980 19602 19981 19603 19982 2 19983 583 19984 19606 19985 19607 19986 40 19987 19609 19988 287 19989 19611 19990 19612 19991 2 19992 583 19993 19615 19994 19616 19995 19617 19996 19618 19997 19619 19998 19620 19999 2 20000 751 20001 2 20002 19624 20003 19625 20004 19626 20005 19627 20006 19628 20007 19629 20008 19630 20009 19631 20010 19632 20011 19633 20012 868 20013 19635 20014 19636 20015 19637 20016 19638 20017 19639 20018 19640 20019 19641 20020 19642 20021 19643 20022 19644 20023 19645 20024 19646 20025 19647 20026 19648 20027 19649 20028 19650 20029 19651 20030 19652 20031 19653 20032 19654 20033 19655 20034 19656 20035 19657 20036 19658 20037 19659 20038 19660 20039 19661 20040 19662 20041 19663 20042 19664 20043 19665 20044 19666 20045 19667 20046 19668 20047 19669 20048 19670 20049 19671 20050 19672 20051 19673 20052 19674 20053 19675 20054 19676 20055 19677 20056 19678 20057 19679 20058 19680 20059 19681 20060 19682 20061 19683 20062 19684 20063 19685 20064 19686 20065 19687 20066 19688 20067 19689 20068 19690 20069 19691 20070 19692 20071 19693 20072 19694 20073 19695 20074 19696 20075 19697 20076 19698 20077 19699 20078 19700 20079 19701 20080 19702 20081 19703 20082 19704 20083 19705 20084 19706 20085 19707 20086 19708 20087 19709 20088 19710 20089 19711 20090 19712 20091 19713 20092 19714 20093 11 20094 19716 20095 19717 20096 19718 20097 2 20098 583 20099 19721 20100 956 20101 19723 20102 19724 20103 989 20104 161 20105 164 20106 19728 20107 19729 20108 1505 20109 2 20110 2 20111 2 20112 19734 20113 2 20114 19736 20115 2 20116 2 20117 2 20118 2 20119 2 20120 2 20121 2 20122 2 20123 2 20124 19746 20125 19747 20126 19748 20127 19749 20128 19750 20129 19751 20130 19752 20131 19753 20132 19754 20133 19755 20134 19756 20135 19757 20136 19758 20137 22 20138 19760 20139 4 20140 11 20141 19763 20142 90 20143 11 20144 19766 20145 101 20146 19768 20147 1569 20148 19770 20149 19771 20150 19772 20151 101 20152 868 20153 11 20154 19776 20155 956 20156 19778 20157 19779 20158 989 20159 161 20160 164 20161 19783 20162 19784 20163 19785 20164 868 20165 11 20166 19788 20167 19789 20168 19790 20169 364 20170 364 20171 367 20172 441 20173 19795 20174 19796 20175 19797 20176 19798 20177 2 20178 583 20179 19801 20180 19802 20181 19803 20182 19804 20183 19805 20184 19806 20185 2 20186 583 20187 19809 20188 19810 20189 142 20190 11 20191 144 20192 164 20193 165 20194 167 20195 169 20196 77 20197 128 20198 172 20199 19821 20200 19822 20201 142 20202 11 20203 144 20204 164 20205 2270 20206 2272 20207 2273 20208 2274 20209 2 20210 19895 20211 20211 20212 20211 20213 868 20214 20214 20215 11 20216 20216 20217 2 20218 583 20219 20219 20220 20220 20221 20221 20222 2756 20223 20223 20224 20224 20225 364 20226 364 20227 367 20228 441 20229 20229 20230 20230 20231 20231 20232 20232 20233 2 20234 583 20235 20235 20236 20236 20237 20237 20238 20238 20239 2827 20240 20240 20241 441 20242 20242 20243 2 20244 20244 20245 20245 20246 2 20247 3 20248 4 20249 2 20250 2 20251 2 20252 2 20253 11 20254 2 20255 2 20256 20256 20257 2 20258 20258 20259 22 20260 20260 20261 20261 20262 2 20263 11 20264 11 20265 20265 20266 20266 20267 20267 20268 22 20269 2 20270 2 20271 2 20272 11 20273 20273 20274 11 20275 20275 20276 11 20277 20277 20278 20278 20279 20279 20280 20280 20281 60 20282 20282 20283 2 20284 2 20285 20285 20286 11 20287 2 20288 2 20289 69 20290 133 20291 20291 20292 2 20293 2 20294 20294 20295 2 20296 20296 20297 20297 20298 2 20299 86 20300 87 20301 90 20302 136 20303 11 20304 20304 20305 20305 20306 101 20307 20307 20308 20308 20309 20309 20310 106 20311 106 20312 20312 20313 20313 20314 20314 20315 20315 20316 20316 20317 20317 20318 20318 20319 20319 20320 20320 20321 20321 20322 20322 20323 101 20324 2 20325 20325 20326 2 20327 2 20328 2 20329 2 20330 2 20331 2 20332 2 20333 2 20334 312 20335 2 20336 2 20337 2 20338 2 20339 2 20340 318 20341 2 20342 2 20343 2 20344 2 20345 2 20346 2 20347 2 20348 2 20349 2 20350 2 20351 2 20352 2 20353 2 20354 2 20355 2 20356 2 20357 20357 20358 20358 20359 20359 20360 20360 20361 20361 20362 20362 20363 20363 20364 20364 20365 20365 20366 20366 20367 20367 20368 20368 20369 20369 20370 20370 20371 20371 20372 20372 20373 4 20374 2 20375 20375 20376 20376 20377 20377 20378 20378 20379 20379 20380 20380 20381 20381 20382 20382 20383 20383 20384 20384 20385 20385 20386 364 20387 364 20388 40 20389 367 20390 441 20391 20391 20392 20392 20393 20393 20394 20394 20395 20395 20396 2 20397 583 20398 20398 20399 20399 20400 40 20401 20401 20402 287 20403 20403 20404 20404 20405 2 20406 583 20407 20407 20408 20408 20409 20409 20410 20410 20411 20411 20412 20412 20413 2 20414 751 20415 2 20416 20416 20417 20417 20418 20418 20419 20419 20420 20420 20421 20421 20422 20422 20423 20423 20424 20424 20425 20425 20426 868 20427 20427 20428 20428 20429 20429 20430 20430 20431 20431 20432 20432 20433 20433 20434 20434 20435 20435 20436 20436 20437 20437 20438 20438 20439 20439 20440 20440 20441 20441 20442 20442 20443 20443 20444 20444 20445 20445 20446 20446 20447 20447 20448 20448 20449 20449 20450 20450 20451 20451 20452 20452 20453 20453 20454 20454 20455 20455 20456 20456 20457 20457 20458 20458 20459 20459 20460 20460 20461 20461 20462 20462 20463 20463 20464 20464 20465 20465 20466 20466 20467 20467 20468 20468 20469 20469 20470 20470 20471 20471 20472 20472 20473 20473 20474 20474 20475 20475 20476 20476 20477 20477 20478 20478 20479 20479 20480 20480 20481 20481 20482 20482 20483 20483 20484 20484 20485 20485 20486 20486 20487 20487 20488 20488 20489 20489 20490 20490 20491 20491 20492 20492 20493 20493 20494 20494 20495 20495 20496 20496 20497 20497 20498 20498 20499 20499 20500 20500 20501 20501 20502 20502 20503 20503 20504 20504 20505 20505 20506 20506 20507 11 20508 20508 20509 20509 20510 20510 20511 2 20512 583 20513 20513 20514 956 20515 20515 20516 20516 20517 989 20518 161 20519 164 20520 20520 20521 20521 20522 1505 20523 2 20524 2 20525 2 20526 20526 20527 2 20528 20528 20529 2 20530 2 20531 2 20532 2 20533 2 20534 2 20535 2 20536 2 20537 2 20538 20538 20539 20539 20540 20540 20541 20541 20542 20542 20543 20543 20544 20544 20545 20545 20546 20546 20547 20547 20548 20548 20549 20549 20550 20550 20551 22 20552 20552 20553 4 20554 11 20555 20555 20556 90 20557 11 20558 20558 20559 101 20560 20560 20561 1569 20562 20562 20563 20563 20564 20564 20565 101 20566 868 20567 11 20568 20568 20569 956 20570 20570 20571 20571 20572 989 20573 161 20574 164 20575 20575 20576 20576 20577 20577 20578 868 20579 11 20580 20580 20581 20581 20582 20582 20583 364 20584 364 20585 367 20586 441 20587 20587 20588 20588 20589 20589 20590 20590 20591 2 20592 583 20593 20593 20594 20594 20595 20595 20596 20596 20597 20597 20598 20598 20599 2 20600 583 20601 20601 20602 20602 20603 142 20604 11 20605 144 20606 164 20607 165 20608 167 20609 169 20610 77 20611 128 20612 172 20613 20613 20614 20614 20615 142 20616 11 20617 144 20618 164 20619 2270 20620 2272 20621 2273 20622 2274 20623 2 20624 20624 20625 20625 20626 2 20627 20627 20628 868 20629 20629 20630 11 20631 20631 20632 2 20633 583 20634 20634 20635 20635 20636 20636 20637 2756 20638 20638 20639 20639 20640 364 20641 364 20642 367 20643 441 20644 20644 20645 20645 20646 20646 20647 20647 20648 2 20649 583 20650 20650 20651 20651 20652 20652 20653 20653 20654 2827 20655 20655 20656 441 20657 20657 20658 2 20659 20659 20660 20660 20661 20309 20662 2 20663 20663 20664 20664 20665 20665 20666 20666 20667 20667 20668 2918 20669 20669 20670 69 20671 20671 20672 20672 20673 20673 20674 20674 20675 20675 20676 20676 20677 20677 20678 2 20679 20679 20680 20680 20681 20681 20682 20682 20683 20683 20684 20684 20685 2960 20686 2962 20687 20687 20688 20688 20689 20689 20690 20690 20691 20691 20692 364 20693 364 20694 40 20695 367 20696 441 20697 20697 20698 20698 20699 20699 20700 20700 20701 20701 20702 40 20703 20703 20704 20704 20705 2983 20706 2986 20707 2988 20708 2990 20709 77 20710 2992 20711 2993 20712 20712 20713 287 20714 20714 20715 2 20716 287 20717 77 20718 3959 20719 2852 20720 20720 20721 1882 20722 164 20723 2072 20724 2 20725 20725 20726 2 20727 161 20728 164 20729 20729 20730 20730 20731 20731 20732 20732 20733 20733 20734 956 20735 20735 20736 20736 20737 20737 20738 989 20739 161 20740 164 20741 20741 20742 20742 20743 2 20744 287 20745 20745 20746 2 20747 287 20748 20748 20749 20749 20750 20750 20751 956 20752 20752 20753 20753 20754 20754 20755 989 20756 5285 20757 161 20758 164 20759 20759 20760 20760 20761 956 20762 20762 20763 20763 20764 20764 20765 20765 20766 1882 20767 164 20768 20768 20769 20769 20770 20770 20771 20771 20772 287 20773 20773 20774 2 20775 583 20776 20776 20777 5356 20778 20778 20779 20779 20780 20780 20781 20781 20782 20782 20783 20783 20784 2918 20785 20785 20786 69 20787 20787 20788 20788 20789 20789 20790 20790 20791 20791 20792 20792 20793 20793 20794 2 20795 20795 20796 20796 20797 20797 20798 20798 20799 20799 20800 20800 20801 2960 20802 2962 20803 20803 20804 20804 20805 20805 20806 20806 20807 20807 20808 364 20809 364 20810 40 20811 367 20812 441 20813 20813 20814 20814 20815 20815 20816 20816 20817 20817 20818 40 20819 20819 20820 20820 20821 2983 20822 2986 20823 2988 20824 2990 20825 77 20826 2992 20827 2993 20828 20828 20829 287 20830 20830 20831 2 20832 287 20833 77 20834 3959 20835 2852 20836 20836 20837 1882 20838 164 20839 2072 20840 2 20841 20841 20842 2 20843 161 20844 164 20845 20845 20846 20846 20847 20847 20848 20848 20849 20849 20850 956 20851 20851 20852 20852 20853 20853 20854 989 20855 161 20856 164 20857 20857 20858 20858 20859 2 20860 287 20861 20861 20862 2 20863 287 20864 20864 20865 20865 20866 20866 20867 956 20868 20868 20869 20869 20870 20870 20871 989 20872 5285 20873 161 20874 164 20875 20875 20876 20876 20877 956 20878 20878 20879 20879 20880 20880 20881 20881 20882 1882 20883 164 20884 20884 20885 20885 20886 20886 20887 20887 20888 287 20889 20889 20890 2 20891 583 20892 20892 20893 5356 20894 20894 20895 20895 20896 20896 20897 4 20898 11 20899 11 20900 20900 20901 86 20902 22 20903 90 20904 11 20905 20905 20906 101 20907 20907 20908 20908 20909 1569 20910 20910 20911 20911 20912 20912 20913 101 20914 868 20915 11 20916 20916 20917 956 20918 20918 20919 20919 20920 989 20921 161 20922 164 20923 20923 20924 20924 20925 20925 20926 17884 20927 20896 20928 4 20929 11 20930 11 20931 20931 20932 86 20933 22 20934 90 20935 11 20936 20936 20937 101 20938 164 20939 20939 20940 1569 20941 20941 20942 20942 20943 20943 20944 101 20945 868 20946 11 20947 20947 20948 956 20949 20949 20950 20950 20951 989 20952 161 20953 164 20954 20954 20955 20955 20956 20956 20957 17884 20958 20958 20959 20959 20960 20960 20961 20961 20962 20962 20963 15 20964 16 20965 15 20966 11 20967 26 20968 2 20969 7507 20970 16 20971 11 20972 11 20973 40 20974 16 20975 11 20976 11 20977 20977 20978 16 20979 15 20980 70 20981 71 20982 11 20983 16 20984 77 20985 11 20986 2 20987 11 20988 20988 20989 89 20990 20990 20991 203 20992 20992 20993 100 20994 20994 20995 105 20996 108 20997 20997 20998 20998 20999 20999 21000 21000 21001 21001 21002 21002 21003 21003 21004 128 21005 172 21006 21006 21007 26 21008 11 21009 2 21010 287 21011 26 21012 21012 21013 15 21014 15 21015 11 21016 26 21017 2 21018 11 21019 11 21020 40 21021 16 21022 21022 21023 15 21024 70 21025 71 21026 11 21027 77 21028 11 21029 2 21030 11 21031 21031 21032 89 21033 105 21034 21034 21035 20999 21036 21036 21037 128 21038 172 21039 21039 21040 26 21041 21041 21042 21042 21043 1080 21044 7507 21045 21045 21046 17805 21047 17806 21048 7507 21049 21049 21050 2 21051 287 21052 17884 21053 17885 21054 21054 21055 21055 21056 17888 21057 7507 21058 21058 21059 105 21060 21060 21061 22 21062 161 21063 164 21064 7507 21065 21065 21066 128 21067 172 21068 21068 21069 21069 21070 11 21071 77 21072 21072 21073 128 21074 172 21075 21075 21076 13 21077 18061 21078 164 21079 77 21080 128 21081 172 21082 21082 21083 2 21084 287 21085 18074 21086 18075 21087 21087 21088 18077 21089 18078 21090 18079 21091 21091 21092 18079 21093 21093 21094 18217 21095 18218 21096 7507 21097 21097 21098 18218 21099 7507 21100 21100 21101 21101 21102 3066 21103 21103 21104 7507 21105 21105 21106 7507 21107 21107 21108 3066 21109 21109 21110 2 21111 19517 21112 20896 21113 20907 21114 17884 21115 20896 21116 164 21117 17884 21118 20958 21119 21119 21120 21120 21121 20961 21122 7507 21123 21123 21124 21124 21125 21125 21126 2 21127 21127 21128 7507 21129 21129 21130 21130 21131 2 21132 21132 21133 21133 21134 21134 21135 2 21136 21136 21137 7507 21138 21138 21139 21130 21140 2 21141 21141 21142 2 21143 3 21144 4 21145 2 21146 2 21147 2 21148 2 21149 11 21150 2 21151 2 21152 21152 21153 2 21154 21154 21155 22 21156 21156 21157 21157 21158 2 21159 11 21160 11 21161 21161 21162 21162 21163 21163 21164 22 21165 2 21166 2 21167 2 21168 11 21169 21169 21170 11 21171 21171 21172 11 21173 21173 21174 21174 21175 21175 21176 21176 21177 60 21178 21178 21179 2 21180 2 21181 21181 21182 11 21183 2 21184 2 21185 69 21186 133 21187 21187 21188 2 21189 2 21190 21190 21191 2 21192 21192 21193 21193 21194 2 21195 86 21196 87 21197 90 21198 136 21199 11 21200 21200 21201 21201 21202 101 21203 21203 21204 21204 21205 21205 21206 106 21207 106 21208 21208 21209 21209 21210 21210 21211 21211 21212 21212 21213 21213 21214 21214 21215 21215 21216 21216 21217 21217 21218 21218 21219 101 21220 2 21221 21221 21222 2 21223 2 21224 2 21225 2 21226 2 21227 2 21228 2 21229 2 21230 312 21231 2 21232 2 21233 2 21234 2 21235 2 21236 318 21237 2 21238 2 21239 2 21240 2 21241 2 21242 2 21243 2 21244 2 21245 2 21246 2 21247 2 21248 2 21249 2 21250 2 21251 2 21252 2 21253 21253 21254 21254 21255 21255 21256 21256 21257 21257 21258 21258 21259 21259 21260 21260 21261 21261 21262 21262 21263 21263 21264 21264 21265 21265 21266 21266 21267 21267 21268 21268 21269 4 21270 2 21271 21271 21272 21272 21273 21273 21274 21274 21275 21275 21276 21276 21277 21277 21278 21278 21279 21279 21280 21280 21281 21281 21282 364 21283 364 21284 40 21285 367 21286 441 21287 21287 21288 21288 21289 21289 21290 21290 21291 21291 21292 2 21293 583 21294 21294 21295 21295 21296 40 21297 21297 21298 287 21299 21299 21300 21300 21301 2 21302 583 21303 21303 21304 21304 21305 21305 21306 21306 21307 21307 21308 21308 21309 2 21310 751 21311 2 21312 21312 21313 21313 21314 21314 21315 21315 21316 21316 21317 21317 21318 21318 21319 21319 21320 21320 21321 21321 21322 868 21323 21323 21324 21324 21325 21325 21326 21326 21327 21327 21328 21328 21329 21329 21330 21330 21331 21331 21332 21332 21333 21333 21334 21334 21335 21335 21336 21336 21337 21337 21338 21338 21339 21339 21340 21340 21341 21341 21342 21342 21343 21343 21344 21344 21345 21345 21346 21346 21347 21347 21348 21348 21349 21349 21350 21350 21351 21351 21352 21352 21353 21353 21354 21354 21355 21355 21356 21356 21357 21357 21358 21358 21359 21359 21360 21360 21361 21361 21362 21362 21363 21363 21364 21364 21365 21365 21366 21366 21367 21367 21368 21368 21369 21369 21370 21370 21371 21371 21372 21372 21373 21373 21374 21374 21375 21375 21376 21376 21377 21377 21378 21378 21379 21379 21380 21380 21381 21381 21382 21382 21383 21383 21384 21384 21385 21385 21386 21386 21387 21387 21388 21388 21389 21389 21390 21390 21391 21391 21392 21392 21393 21393 21394 21394 21395 21395 21396 21396 21397 21397 21398 21398 21399 21399 21400 21400 21401 21401 21402 21402 21403 11 21404 21404 21405 21405 21406 21406 21407 2 21408 583 21409 21409 21410 956 21411 21411 21412 21412 21413 989 21414 161 21415 164 21416 21416 21417 21417 21418 1505 21419 2 21420 2 21421 2 21422 21422 21423 2 21424 21424 21425 2 21426 2 21427 2 21428 2 21429 2 21430 2 21431 2 21432 2 21433 2 21434 21434 21435 21435 21436 21436 21437 21437 21438 21438 21439 21439 21440 21440 21441 21441 21442 21442 21443 21443 21444 21444 21445 21445 21446 21446 21447 22 21448 21448 21449 4 21450 11 21451 21451 21452 90 21453 11 21454 21454 21455 101 21456 21456 21457 1569 21458 21458 21459 21459 21460 21460 21461 101 21462 868 21463 11 21464 21464 21465 956 21466 21466 21467 21467 21468 989 21469 161 21470 164 21471 21471 21472 21472 21473 21473 21474 868 21475 11 21476 21476 21477 21477 21478 21478 21479 364 21480 364 21481 367 21482 441 21483 21483 21484 21484 21485 21485 21486 21486 21487 2 21488 583 21489 21489 21490 21490 21491 21491 21492 21492 21493 21493 21494 21494 21495 2 21496 583 21497 21497 21498 21498 21499 142 21500 11 21501 144 21502 164 21503 165 21504 167 21505 169 21506 77 21507 128 21508 172 21509 21509 21510 21510 21511 142 21512 11 21513 144 21514 164 21515 2270 21516 2272 21517 2273 21518 2274 21519 2 21520 2 21521 3 21522 4 21523 2 21524 2 21525 2 21526 2 21527 11 21528 2 21529 2 21530 21152 21531 2 21532 21154 21533 22 21534 21156 21535 21157 21536 2 21537 11 21538 11 21539 21161 21540 21162 21541 21163 21542 22 21543 2 21544 2 21545 2 21546 11 21547 21169 21548 11 21549 21171 21550 11 21551 21173 21552 21174 21553 21175 21554 21176 21555 60 21556 21178 21557 2 21558 2 21559 21181 21560 11 21561 2 21562 2 21563 69 21564 133 21565 21187 21566 2 21567 2 21568 21190 21569 2 21570 21192 21571 21193 21572 2 21573 86 21574 87 21575 90 21576 136 21577 11 21578 21200 21579 21201 21580 101 21581 21203 21582 21204 21583 21583 21584 106 21585 106 21586 21208 21587 21209 21588 21210 21589 21211 21590 21212 21591 21213 21592 21214 21593 21215 21594 21216 21595 21217 21596 21218 21597 101 21598 2 21599 21221 21600 2 21601 2 21602 2 21603 2 21604 2 21605 2 21606 2 21607 2 21608 312 21609 2 21610 2 21611 2 21612 2 21613 2 21614 318 21615 2 21616 2 21617 2 21618 2 21619 2 21620 2 21621 2 21622 2 21623 2 21624 2 21625 2 21626 2 21627 2 21628 2 21629 2 21630 2 21631 21253 21632 21254 21633 21255 21634 21256 21635 21257 21636 21258 21637 21259 21638 21260 21639 21261 21640 21262 21641 21263 21642 21264 21643 21265 21644 21266 21645 21267 21646 21268 21647 4 21648 2 21649 21271 21650 21272 21651 21273 21652 21274 21653 21275 21654 21276 21655 21277 21656 21278 21657 21279 21658 21280 21659 21281 21660 364 21661 364 21662 40 21663 367 21664 441 21665 21287 21666 21288 21667 21289 21668 21290 21669 21291 21670 2 21671 583 21672 21294 21673 21295 21674 40 21675 21297 21676 287 21677 21299 21678 21300 21679 2 21680 583 21681 21303 21682 21304 21683 21305 21684 21306 21685 21307 21686 21308 21687 2 21688 751 21689 2 21690 21312 21691 21313 21692 21314 21693 21315 21694 21316 21695 21317 21696 21318 21697 21319 21698 21320 21699 21321 21700 868 21701 21323 21702 21324 21703 21325 21704 21326 21705 21327 21706 21328 21707 21329 21708 21330 21709 21331 21710 21332 21711 21333 21712 21334 21713 21335 21714 21336 21715 21337 21716 21338 21717 21339 21718 21340 21719 21341 21720 21342 21721 21343 21722 21344 21723 21345 21724 21346 21725 21347 21726 21348 21727 21349 21728 21350 21729 21351 21730 21352 21731 21353 21732 21354 21733 21355 21734 21356 21735 21357 21736 21358 21737 21359 21738 21360 21739 21361 21740 21362 21741 21363 21742 21364 21743 21365 21744 21366 21745 21367 21746 21368 21747 21369 21748 21370 21749 21371 21750 21372 21751 21373 21752 21374 21753 21375 21754 21376 21755 21377 21756 21378 21757 21379 21758 21380 21759 21381 21760 21382 21761 21383 21762 21384 21763 21385 21764 21386 21765 21387 21766 21388 21767 21389 21768 21390 21769 21391 21770 21392 21771 21393 21772 21394 21773 21395 21774 21396 21775 21397 21776 21398 21777 21399 21778 21400 21779 21401 21780 21402 21781 11 21782 21404 21783 21405 21784 21406 21785 2 21786 583 21787 21409 21788 956 21789 21411 21790 21412 21791 989 21792 161 21793 164 21794 21416 21795 21417 21796 1505 21797 2 21798 2 21799 2 21800 21422 21801 2 21802 21424 21803 2 21804 2 21805 2 21806 2 21807 2 21808 2 21809 2 21810 2 21811 2 21812 21434 21813 21435 21814 21436 21815 21437 21816 21438 21817 21439 21818 21440 21819 21441 21820 21442 21821 21443 21822 21444 21823 21445 21824 21446 21825 22 21826 21448 21827 4 21828 11 21829 21451 21830 90 21831 11 21832 21454 21833 101 21834 21456 21835 1569 21836 21458 21837 21459 21838 21460 21839 101 21840 868 21841 11 21842 21464 21843 956 21844 21466 21845 21467 21846 989 21847 161 21848 164 21849 21471 21850 21472 21851 21473 21852 868 21853 11 21854 21476 21855 21477 21856 21478 21857 364 21858 364 21859 367 21860 441 21861 21483 21862 21484 21863 21485 21864 21486 21865 2 21866 583 21867 21489 21868 21490 21869 21491 21870 21492 21871 21493 21872 21494 21873 2 21874 583 21875 21497 21876 21498 21877 142 21878 11 21879 144 21880 164 21881 165 21882 167 21883 169 21884 77 21885 128 21886 172 21887 21509 21888 21510 21889 142 21890 11 21891 144 21892 164 21893 2270 21894 2272 21895 2273 21896 2274 21897 2 21898 21583 21899 21899 21900 21899 21901 868 21902 21902 21903 11 21904 21904 21905 2 21906 583 21907 21907 21908 21908 21909 21909 21910 2756 21911 21911 21912 21912 21913 364 21914 364 21915 367 21916 441 21917 21917 21918 21918 21919 21919 21920 21920 21921 2 21922 583 21923 21923 21924 21924 21925 21925 21926 21926 21927 2827 21928 21928 21929 441 21930 21930 21931 2 21932 21932 21933 21933 21934 5553 21935 2 21936 21936 21937 21937 21938 21938 21939 21939 21940 21940 21941 2918 21942 21942 21943 69 21944 21944 21945 21945 21946 21946 21947 21947 21948 21948 21949 21949 21950 21950 21951 2 21952 21952 21953 21953 21954 21954 21955 21955 21956 21956 21957 21957 21958 2960 21959 2962 21960 21960 21961 21961 21962 21962 21963 21963 21964 21964 21965 364 21966 364 21967 40 21968 367 21969 441 21970 21970 21971 21971 21972 21972 21973 21973 21974 21974 21975 40 21976 21976 21977 21977 21978 2983 21979 2986 21980 2988 21981 2990 21982 77 21983 2992 21984 2993 21985 21985 21986 287 21987 21987 21988 2 21989 287 21990 77 21991 3959 21992 2852 21993 21993 21994 1882 21995 164 21996 2072 21997 2 21998 21998 21999 2 22000 161 22001 164 22002 22002 22003 22003 22004 22004 22005 22005 22006 22006 22007 956 22008 22008 22009 22009 22010 22010 22011 989 22012 161 22013 164 22014 22014 22015 22015 22016 2 22017 287 22018 22018 22019 2 22020 287 22021 22021 22022 22022 22023 22023 22024 956 22025 22025 22026 22026 22027 22027 22028 989 22029 5285 22030 161 22031 164 22032 22032 22033 22033 22034 956 22035 22035 22036 22036 22037 22037 22038 22038 22039 1882 22040 164 22041 22041 22042 22042 22043 22043 22044 22044 22045 287 22046 22046 22047 2 22048 583 22049 22049 22050 5356 22051 22051 22052 22052 22053 868 22054 21902 22055 11 22056 21904 22057 2 22058 583 22059 21907 22060 21908 22061 21909 22062 2756 22063 21911 22064 21912 22065 364 22066 364 22067 367 22068 441 22069 21917 22070 21918 22071 21919 22072 21920 22073 2 22074 583 22075 21923 22076 21924 22077 21925 22078 21926 22079 2827 22080 21928 22081 441 22082 21930 22083 2 22084 21932 22085 21933 22086 2 22087 3 22088 4 22089 2 22090 2 22091 2 22092 2 22093 11 22094 2 22095 2 22096 3970 22097 2 22098 3972 22099 22 22100 3975 22101 3976 22102 2 22103 11 22104 11 22105 3980 22106 3981 22107 3982 22108 22 22109 2 22110 2 22111 2 22112 11 22113 3988 22114 11 22115 3990 22116 11 22117 3992 22118 3993 22119 3994 22120 3995 22121 60 22122 3997 22123 2 22124 2 22125 4000 22126 11 22127 2 22128 2 22129 69 22130 133 22131 4006 22132 2 22133 2 22134 4009 22135 2 22136 4011 22137 4012 22138 2 22139 86 22140 87 22141 90 22142 136 22143 11 22144 4019 22145 4020 22146 101 22147 4022 22148 4023 22149 5553 22150 106 22151 106 22152 4027 22153 4028 22154 4029 22155 4030 22156 4034 22157 4035 22158 4036 22159 4038 22160 4039 22161 4040 22162 4041 22163 101 22164 2 22165 4044 22166 2 22167 2 22168 2 22169 2 22170 2 22171 2 22172 2 22173 2 22174 312 22175 2 22176 2 22177 2 22178 2 22179 2 22180 318 22181 2 22182 2 22183 2 22184 2 22185 2 22186 2 22187 2 22188 2 22189 2 22190 2 22191 2 22192 2 22193 2 22194 2 22195 2 22196 2 22197 4076 22198 4077 22199 4078 22200 4079 22201 4080 22202 4081 22203 4082 22204 4083 22205 4084 22206 4085 22207 4086 22208 4087 22209 4088 22210 4089 22211 4090 22212 4091 22213 4 22214 2 22215 4094 22216 4095 22217 4096 22218 4097 22219 4098 22220 4099 22221 4100 22222 4101 22223 4102 22224 4103 22225 4104 22226 364 22227 364 22228 40 22229 367 22230 441 22231 4110 22232 4111 22233 4112 22234 4113 22235 4114 22236 2 22237 583 22238 4117 22239 4118 22240 40 22241 4120 22242 287 22243 4122 22244 4123 22245 2 22246 583 22247 4126 22248 4127 22249 4128 22250 4129 22251 4130 22252 4131 22253 2 22254 751 22255 2 22256 4135 22257 4136 22258 4137 22259 4138 22260 4139 22261 4140 22262 4141 22263 4142 22264 4143 22265 4144 22266 868 22267 4146 22268 4147 22269 4148 22270 4149 22271 4150 22272 4151 22273 4152 22274 4153 22275 4154 22276 4155 22277 4156 22278 4157 22279 4158 22280 4159 22281 4160 22282 4161 22283 4162 22284 4163 22285 4164 22286 4165 22287 4166 22288 4167 22289 4168 22290 4169 22291 4170 22292 4171 22293 4172 22294 4173 22295 4174 22296 4175 22297 4176 22298 4177 22299 4178 22300 4179 22301 4180 22302 4181 22303 4182 22304 4183 22305 4184 22306 4185 22307 4186 22308 4187 22309 4188 22310 4189 22311 4190 22312 4191 22313 4192 22314 4193 22315 4194 22316 4195 22317 4196 22318 4197 22319 4198 22320 4199 22321 4200 22322 4201 22323 4202 22324 4203 22325 4204 22326 4205 22327 4206 22328 4207 22329 4208 22330 4209 22331 4210 22332 4211 22333 4212 22334 4213 22335 4214 22336 4215 22337 4216 22338 4217 22339 4218 22340 4219 22341 4220 22342 4221 22343 4222 22344 4223 22345 4224 22346 4225 22347 11 22348 4227 22349 4228 22350 4229 22351 2 22352 583 22353 4232 22354 956 22355 4234 22356 4235 22357 989 22358 161 22359 164 22360 4239 22361 4240 22362 1505 22363 2 22364 2 22365 2 22366 4245 22367 2 22368 4247 22369 2 22370 2 22371 2 22372 2 22373 2 22374 2 22375 2 22376 2 22377 2 22378 4257 22379 4258 22380 4259 22381 4260 22382 4261 22383 4262 22384 4263 22385 4264 22386 4265 22387 4266 22388 4267 22389 4268 22390 4269 22391 22 22392 4271 22393 4 22394 11 22395 4274 22396 90 22397 11 22398 4277 22399 101 22400 4279 22401 1569 22402 4281 22403 4282 22404 4283 22405 101 22406 868 22407 11 22408 4287 22409 956 22410 4289 22411 4290 22412 989 22413 161 22414 164 22415 4294 22416 4295 22417 4296 22418 868 22419 11 22420 4299 22421 4300 22422 4301 22423 364 22424 364 22425 367 22426 441 22427 4306 22428 4307 22429 4308 22430 4309 22431 2 22432 583 22433 4312 22434 4313 22435 4314 22436 4315 22437 4316 22438 4317 22439 2 22440 583 22441 4320 22442 4324 22443 142 22444 11 22445 144 22446 164 22447 165 22448 167 22449 169 22450 77 22451 128 22452 172 22453 5507 22454 5508 22455 142 22456 11 22457 144 22458 164 22459 2270 22460 2272 22461 2273 22462 2274 22463 2 22464 5518 22465 5519 22466 2 22467 21936 22468 868 22469 5521 22470 11 22471 5523 22472 2 22473 583 22474 5526 22475 5527 22476 5528 22477 2756 22478 5530 22479 5531 22480 364 22481 364 22482 367 22483 441 22484 5536 22485 5537 22486 5538 22487 5539 22488 2 22489 583 22490 5542 22491 5543 22492 5544 22493 5545 22494 2827 22495 5547 22496 441 22497 5549 22498 2 22499 5551 22500 5552 22501 5553 22502 2 22503 5555 22504 5556 22505 5557 22506 5558 22507 5559 22508 2918 22509 5561 22510 69 22511 5563 22512 5564 22513 5565 22514 5566 22515 5567 22516 5568 22517 5569 22518 2 22519 5571 22520 5572 22521 5573 22522 5574 22523 5575 22524 5576 22525 2960 22526 2962 22527 5579 22528 5580 22529 5581 22530 5582 22531 5583 22532 364 22533 364 22534 40 22535 367 22536 441 22537 5589 22538 5590 22539 5591 22540 5592 22541 5593 22542 40 22543 5595 22544 5596 22545 2983 22546 2986 22547 2988 22548 2990 22549 77 22550 2992 22551 2993 22552 5604 22553 287 22554 5606 22555 2 22556 287 22557 77 22558 3959 22559 2852 22560 5612 22561 1882 22562 164 22563 2072 22564 2 22565 5617 22566 2 22567 161 22568 164 22569 5621 22570 5622 22571 5623 22572 5624 22573 5625 22574 956 22575 5627 22576 5628 22577 5629 22578 989 22579 161 22580 164 22581 5633 22582 5634 22583 2 22584 287 22585 5637 22586 2 22587 287 22588 5640 22589 5641 22590 5646 22591 956 22592 5648 22593 5649 22594 5650 22595 989 22596 5285 22597 161 22598 164 22599 5655 22600 5656 22601 956 22602 5658 22603 5659 22604 5660 22605 5661 22606 1882 22607 164 22608 5664 22609 5665 22610 5666 22611 5667 22612 287 22613 5669 22614 2 22615 583 22616 5672 22617 5356 22618 5674 22619 5675 22620 21937 22621 21938 22622 21939 22623 21940 22624 2918 22625 21942 22626 69 22627 21944 22628 21945 22629 21946 22630 21947 22631 21948 22632 21949 22633 21950 22634 2 22635 21952 22636 21953 22637 21954 22638 21955 22639 21956 22640 21957 22641 2960 22642 2962 22643 21960 22644 21961 22645 21962 22646 21963 22647 21964 22648 364 22649 364 22650 40 22651 367 22652 441 22653 21970 22654 21971 22655 21972 22656 21973 22657 21974 22658 40 22659 21976 22660 21977 22661 2983 22662 2986 22663 2988 22664 2990 22665 77 22666 2992 22667 2993 22668 21985 22669 287 22670 21987 22671 2 22672 287 22673 77 22674 3959 22675 2852 22676 21993 22677 1882 22678 164 22679 2072 22680 2 22681 21998 22682 2 22683 161 22684 164 22685 22002 22686 22003 22687 22004 22688 22005 22689 22006 22690 956 22691 22008 22692 22009 22693 22010 22694 989 22695 161 22696 164 22697 22014 22698 22015 22699 2 22700 287 22701 22018 22702 2 22703 287 22704 22021 22705 22022 22706 22023 22707 956 22708 22025 22709 22026 22710 22027 22711 989 22712 5285 22713 161 22714 164 22715 22032 22716 22033 22717 956 22718 22035 22719 22036 22720 22037 22721 22038 22722 1882 22723 164 22724 22041 22725 22042 22726 22043 22727 22044 22728 287 22729 22046 22730 2 22731 583 22732 22049 22733 5356 22734 22051 22735 22052 22736 22736 22737 2 22738 22738 22739 164 22740 22740 22741 21130 22742 2 22743 22743 22744 22744 22745 17884 22746 22746 22747 22747 22748 22748 22749 22749 22750 22750 22751 15 22752 16 22753 15 22754 11 22755 26 22756 2 22757 7507 22758 16 22759 11 22760 11 22761 40 22762 16 22763 11 22764 11 22765 22765 22766 16 22767 15 22768 70 22769 71 22770 11 22771 16 22772 77 22773 11 22774 2 22775 11 22776 22776 22777 89 22778 22778 22779 203 22780 22780 22781 100 22782 22782 22783 105 22784 108 22785 22785 22786 22786 22787 22787 22788 22788 22789 22789 22790 22790 22791 22791 22792 128 22793 172 22794 22794 22795 26 22796 11 22797 2 22798 287 22799 26 22800 22800 22801 15 22802 15 22803 11 22804 26 22805 2 22806 11 22807 11 22808 40 22809 16 22810 22810 22811 15 22812 70 22813 71 22814 11 22815 77 22816 11 22817 2 22818 11 22819 22819 22820 89 22821 105 22822 22822 22823 22787 22824 22824 22825 128 22826 172 22827 22827 22828 26 22829 22829 22830 22830 22831 1080 22832 7507 22833 22833 22834 17805 22835 17806 22836 7507 22837 22837 22838 2 22839 287 22840 17884 22841 17885 22842 22842 22843 22843 22844 17888 22845 7507 22846 22846 22847 105 22848 22848 22849 22 22850 161 22851 164 22852 7507 22853 22853 22854 128 22855 172 22856 22856 22857 22857 22858 11 22859 77 22860 22860 22861 128 22862 172 22863 22863 22864 13 22865 18061 22866 164 22867 77 22868 128 22869 172 22870 22870 22871 2 22872 287 22873 18074 22874 18075 22875 22875 22876 18077 22877 18078 22878 18079 22879 22879 22880 18079 22881 22881 22882 18217 22883 18218 22884 7507 22885 22885 22886 18218 22887 7507 22888 22888 22889 22889 22890 3066 22891 22891 22892 7507 22893 22893 22894 7507 22895 22895 22896 3066 22897 22897 22898 2 22899 19517 22900 20896 22901 20907 22902 17884 22903 20896 22904 164 22905 17884 22906 20958 22907 22907 22908 22908 22909 20961 22910 7507 22911 22911 22912 22912 22913 21125 22914 7507 22915 21134 22916 7507 22917 22736 22918 164 22919 17884 22920 22746 22921 22921 22922 22922 22923 22749 22924 7507 22925 22925 22926 22926 22927 3066 22928 22928 22929 2 22930 22930 22931 22931 22932 22932 22933 15 22934 15 22935 11 22936 26 22937 2 22938 11 22939 11 22940 40 22941 16 22942 11 22943 22943 22944 15 22945 70 22946 71 22947 11 22948 77 22949 11 22950 2 22951 11 22952 22952 22953 89 22954 105 22955 22955 22956 22956 22957 22957 22958 128 22959 172 22960 22960 22961 11 22962 2 22963 287 22964 7507 22965 22965 22966 26 22967 11 22968 2 22969 287 22970 26 22971 22971 22972 15 22973 15 22974 11 22975 26 22976 2 22977 11 22978 11 22979 40 22980 16 22981 22981 22982 15 22983 70 22984 71 22985 11 22986 77 22987 11 22988 2 22989 11 22990 22990 22991 89 22992 105 22993 22993 22994 22956 22995 22995 22996 128 22997 172 22998 22998 22999 26 23000 23000 23001 23001 23002 1080 23003 7507 23004 23004 23005 17805 23006 17806 23007 7507 23008 23008 23009 2 23010 287 23011 17884 23012 17885 23013 23013 23014 23014 23015 17888 23016 7507 23017 23017 23018 105 23019 23019 23020 22 23021 161 23022 164 23023 7507 23024 23024 23025 128 23026 172 23027 23027 23028 23028 23029 11 23030 77 23031 23031 23032 128 23033 172 23034 23034 23035 13 23036 18061 23037 164 23038 77 23039 128 23040 172 23041 23041 23042 2 23043 287 23044 18074 23045 18075 23046 23046 23047 18077 23048 18078 23049 18079 23050 23050 23051 18079 23052 23052 23053 18217 23054 18218 23055 7507 23056 23056 23057 18218 23058 7507 23059 23059 23060 23060 23061 3066 23062 23062 23063 7507 23064 23064 23065 7507 23066 23066 23067 3066 23068 23068 23069 2 23070 19517 23071 20896 23072 20907 23073 17884 23074 20896 23075 164 23076 17884 23077 20958 23078 23078 23079 23079 23080 20961 23081 7507 23082 23082 23083 23083 23084 21125 23085 7507 23086 21134 23087 7507 23088 22736 23089 164 23090 17884 23091 22746 23092 23092 23093 23093 23094 22749 23095 7507 23096 23096 23097 3066 23098 23098 23099 22957 23100 11 23101 2 23102 287 23103 7507 23104 23104 23105 16 23106 2072 23107 23107 23108 23108 23109 17884 23110 16 23111 23111 23112 23112 23113 23113 23114 23114 23115 23115 23116 23116 23117 15 23118 16 23119 15 23120 11 23121 26 23122 2 23123 7507 23124 16 23125 11 23126 11 23127 40 23128 16 23129 11 23130 11 23131 23131 23132 16 23133 15 23134 70 23135 71 23136 11 23137 16 23138 77 23139 11 23140 2 23141 11 23142 23142 23143 89 23144 23144 23145 203 23146 23146 23147 100 23148 23148 23149 105 23150 108 23151 23151 23152 23152 23153 23153 23154 23154 23155 23155 23156 23156 23157 23157 23158 128 23159 172 23160 23160 23161 26 23162 11 23163 2 23164 287 23165 26 23166 23166 23167 15 23168 15 23169 11 23170 26 23171 2 23172 11 23173 11 23174 40 23175 16 23176 23176 23177 15 23178 70 23179 71 23180 11 23181 77 23182 11 23183 2 23184 11 23185 23185 23186 89 23187 105 23188 23188 23189 23153 23190 23190 23191 128 23192 172 23193 23193 23194 26 23195 23195 23196 23196 23197 1080 23198 7507 23199 23199 23200 17805 23201 17806 23202 7507 23203 23203 23204 2 23205 287 23206 17884 23207 17885 23208 23208 23209 23209 23210 17888 23211 7507 23212 23212 23213 105 23214 23214 23215 22 23216 161 23217 164 23218 7507 23219 23219 23220 128 23221 172 23222 23222 23223 23223 23224 11 23225 77 23226 23226 23227 128 23228 172 23229 23229 23230 13 23231 18061 23232 164 23233 77 23234 128 23235 172 23236 23236 23237 2 23238 287 23239 18074 23240 18075 23241 23241 23242 18077 23243 18078 23244 18079 23245 23245 23246 18079 23247 23247 23248 18217 23249 18218 23250 7507 23251 23251 23252 18218 23253 7507 23254 23254 23255 23255 23256 3066 23257 23257 23258 7507 23259 23259 23260 7507 23261 23261 23262 3066 23263 23263 23264 2 23265 19517 23266 20896 23267 20907 23268 17884 23269 20896 23270 164 23271 17884 23272 20958 23273 23273 23274 23274 23275 20961 23276 7507 23277 23277 23278 23278 23279 21125 23280 7507 23281 21134 23282 7507 23283 22736 23284 164 23285 17884 23286 22746 23287 23287 23288 23288 23289 22749 23290 7507 23291 23291 23292 3066 23293 23293 23294 22957 23295 11 23296 2 23297 287 23298 7507 23299 23299 23300 16 23301 23108 23302 17884 23303 23112 23304 23304 23305 23305 23306 23115 23307 7507 23308 23308 23309 7507 23310 23310 23311 7507 23312 23312 23313 23313 23314 17773 23315 142 23316 11 23317 144 23318 164 23319 167 23320 169 23321 77 23322 128 23323 172 23324 23324 23325 23325 23326 23326 23327 16 23328 2986 23329 2988 23330 2990 23331 77 23332 2992 23333 2993 23334 23334 23335 23335 23336 1080 23337 7507 23338 23338 23339 17805 23340 17806 23341 7507 23342 23342 23343 2 23344 287 23345 17884 23346 17885 23347 23347 23348 23348 23349 17888 23350 7507 23351 23351 23352 105 23353 23353 23354 22 23355 161 23356 164 23357 7507 23358 23358 23359 128 23360 172 23361 23361 23362 23362 23363 11 23364 77 23365 23365 23366 128 23367 172 23368 23368 23369 13 23370 18061 23371 164 23372 77 23373 128 23374 172 23375 23375 23376 2990 23377 77 23378 2992 23379 2993 23380 23380 23381 2 23382 287 23383 18074 23384 18075 23385 23385 23386 18077 23387 18078 23388 18079 23389 23389 23390 18079 23391 23391 23392 18217 23393 18218 23394 7507 23395 23395 23396 18218 23397 7507 23398 23398 23399 23399 23400 3066 23401 23401 23402 7507 23403 23403 23404 23404 23405 13 23406 2990 23407 77 23408 2992 23409 2993 23410 23410 23411 23411 23412 2 23413 287 23414 17884 23415 17717 23416 23416 23417 23417 23418 164 23419 7507 23420 23420 23421 7507 23422 23422 23423 3066 23424 23424 23425 2 23426 19517 23427 20896 23428 20907 23429 17884 23430 20896 23431 164 23432 17884 23433 20958 23434 23434 23435 23435 23436 20961 23437 7507 23438 23438 23439 23439 23440 21125 23441 7507 23442 21134 23443 7507 23444 22736 23445 164 23446 17884 23447 22746 23448 23448 23449 23449 23450 22749 23451 7507 23452 23452 23453 23453 23454 168 23455 2988 23456 2990 23457 77 23458 2992 23459 2993 23460 23460 23461 2278 23462 23462 23463 23463 23464 23464 23465 23465 23466 23466 23467 23467 23468 23468 23469 23469 23470 23470 23471 23471 23472 23472 23473 23473 23474 23474 23475 2293 23476 2294 23477 2278 23478 23478 23479 23479 23480 23480 23481 23481 23482 23482 23483 23483 23484 23484 23485 23485 23486 23486 23487 23487 23488 23488 23489 2314 23490 23490 23491 2294 23492 23492 23493 2278 23494 23494 23495 23495 23496 23496 23497 23497 23498 23498 23499 23499 23500 23500 23501 23501 23502 23502 23503 23503 23504 2278 23505 23505 23506 23506 23507 23507 23508 23508 23509 23509 23510 23510 23511 23511 23512 23512 23513 3066 23514 23514 23515 22957 23516 11 23517 2 23518 287 23519 7507 23520 23520 23521 16 23522 23108 23523 17884 23524 23112 23525 23525 23526 23526 23527 23115 23528 7507 23529 23529 23530 7507 23531 23531 23532 2986 23533 2988 23534 2990 23535 77 23536 2992 23537 2993 23538 23538 23539 7507 23540 23540 23541 2 23542 23542 23543 23543 23544 7507 23545 23545 23546 7507 23547 23547 23548 23548 23549 23549 23550 1080 23551 7507 23552 23552 23553 17805 23554 17806 23555 7507 23556 23556 23557 2 23558 287 23559 17884 23560 17885 23561 23561 23562 23562 23563 17888 23564 7507 23565 23565 23566 105 23567 23567 23568 22 23569 161 23570 164 23571 7507 23572 23572 23573 128 23574 172 23575 23575 23576 23576 23577 11 23578 77 23579 23579 23580 128 23581 172 23582 23582 23583 13 23584 18061 23585 164 23586 77 23587 128 23588 172 23589 23589 23590 2 23591 287 23592 18074 23593 18075 23594 23594 23595 18077 23596 18078 23597 18079 23598 23598 23599 18079 23600 23600 23601 18217 23602 18218 23603 7507 23604 23604 23605 18218 23606 7507 23607 23607 23608 23608 23609 3066 23610 23610 23611 7507 23612 23612 23613 2 23614 287 23615 17884 23616 23616 23617 7507 23618 23618 23619 3066 23620 23620 23621 2 23622 19517 23623 20896 23624 20907 23625 17884 23626 20896 23627 164 23628 17884 23629 20958 23630 23630 23631 23631 23632 20961 23633 7507 23634 23634 23635 23635 23636 21125 23637 7507 23638 21134 23639 7507 23640 22736 23641 164 23642 17884 23643 22746 23644 23644 23645 23645 23646 22749 23647 7507 23648 23648 23649 3066 23650 23650 23651 22957 23652 11 23653 2 23654 287 23655 7507 23656 23656 23657 16 23658 23108 23659 17884 23660 23112 23661 23661 23662 23662 23663 23115 23664 7507 23665 23665 23666 7507 23667 23667 23668 7507 23669 23669 23670 16 23671 23108 23672 17884 23673 23112 23674 23674 23675 23675 23676 23115 23677 7507 23678 23678 23679 7507 23680 23680 23681 7507 23682 23682 23683 23683 23684 17773 23685 142 23686 11 23687 144 23688 164 23689 167 23690 169 23691 77 23692 128 23693 172 23694 23694 23695 23695 23696 23696 23697 16 23698 2986 23699 2988 23700 2990 23701 77 23702 2992 23703 2993 23704 23704 23705 23705 23706 1080 23707 7507 23708 23708 23709 17805 23710 17806 23711 7507 23712 23712 23713 2 23714 287 23715 17884 23716 17885 23717 23717 23718 23718 23719 17888 23720 7507 23721 23721 23722 105 23723 23723 23724 22 23725 161 23726 164 23727 7507 23728 23728 23729 128 23730 172 23731 23731 23732 23732 23733 11 23734 77 23735 23735 23736 128 23737 172 23738 23738 23739 13 23740 18061 23741 164 23742 77 23743 128 23744 172 23745 23745 23746 2990 23747 77 23748 2992 23749 2993 23750 23750 23751 2 23752 287 23753 18074 23754 18075 23755 23755 23756 18077 23757 18078 23758 18079 23759 23759 23760 18079 23761 23761 23762 18217 23763 18218 23764 7507 23765 23765 23766 18218 23767 7507 23768 23768 23769 23769 23770 3066 23771 23771 23772 7507 23773 23773 23774 23404 23775 2990 23776 77 23777 2992 23778 2993 23779 23779 23780 2 23781 287 23782 17884 23783 23783 23784 164 23785 7507 23786 23786 23787 7507 23788 23788 23789 3066 23790 23790 23791 2 23792 19517 23793 20896 23794 20907 23795 17884 23796 20896 23797 164 23798 17884 23799 20958 23800 23800 23801 23801 23802 20961 23803 7507 23804 23804 23805 23805 23806 21125 23807 7507 23808 21134 23809 7507 23810 22736 23811 164 23812 17884 23813 22746 23814 23814 23815 23815 23816 22749 23817 7507 23818 23818 23819 23453 23820 2988 23821 2990 23822 77 23823 2992 23824 2993 23825 23825 23826 23826 23827 3066 23828 23828 23829 22957 23830 11 23831 2 23832 287 23833 7507 23834 23834 23835 16 23836 23108 23837 17884 23838 23112 23839 23839 23840 23840 23841 23115 23842 7507 23843 23843 23844 7507 23845 23845 23846 2986 23847 2988 23848 2990 23849 77 23850 2992 23851 2993 23852 23852 23853 7507 23854 23854 23855 23855 23856 23856 23857 23857 23858 3066 23859 23859 23860 22957 23861 11 23862 2 23863 287 23864 7507 23865 23865 23866 16 23867 23108 23868 17884 23869 23112 23870 23870 23871 23871 23872 23115 23873 7507 23874 23874 23875 7507 23876 23876 23877 7507 23878 23878 23879 23879 23880 17773 23881 142 23882 11 23883 144 23884 164 23885 167 23886 169 23887 77 23888 128 23889 172 23890 23890 23891 23891 23892 23892 23893 16 23894 2986 23895 2988 23896 2990 23897 77 23898 2992 23899 2993 23900 23900 23901 23901 23902 1080 23903 7507 23904 23904 23905 17805 23906 17806 23907 7507 23908 23908 23909 2 23910 287 23911 17884 23912 17885 23913 23913 23914 23914 23915 17888 23916 7507 23917 23917 23918 105 23919 23919 23920 22 23921 161 23922 164 23923 7507 23924 23924 23925 128 23926 172 23927 23927 23928 23928 23929 11 23930 77 23931 23931 23932 128 23933 172 23934 23934 23935 13 23936 18061 23937 164 23938 77 23939 128 23940 172 23941 23941 23942 2990 23943 77 23944 2992 23945 2993 23946 23946 23947 2 23948 287 23949 18074 23950 18075 23951 23951 23952 18077 23953 18078 23954 18079 23955 23955 23956 18079 23957 23957 23958 18217 23959 18218 23960 7507 23961 23961 23962 18218 23963 7507 23964 23964 23965 23965 23966 3066 23967 23967 23968 7507 23969 23969 23970 23404 23971 2990 23972 77 23973 2992 23974 2993 23975 23975 23976 2 23977 287 23978 17884 23979 23979 23980 164 23981 7507 23982 23982 23983 7507 23984 23984 23985 3066 23986 23986 23987 2 23988 19517 23989 20896 23990 20907 23991 17884 23992 20896 23993 164 23994 17884 23995 20958 23996 23996 23997 23997 23998 20961 23999 7507 24000 24000 24001 24001 24002 21125 24003 7507 24004 21134 24005 7507 24006 22736 24007 164 24008 17884 24009 22746 24010 24010 24011 24011 24012 22749 24013 7507 24014 24014 24015 23453 24016 2988 24017 2990 24018 77 24019 2992 24020 2993 24021 24021 24022 24022 24023 3066 24024 24024 24025 22957 24026 11 24027 2 24028 287 24029 7507 24030 24030 24031 16 24032 23108 24033 17884 24034 23112 24035 24035 24036 24036 24037 23115 24038 7507 24039 24039 24040 7507 24041 24041 24042 2986 24043 2988 24044 2990 24045 77 24046 2992 24047 2993 24048 24048 24049 7507 24050 24050 24051 868 24052 11 24053 24053 24054 24054 24055 24055 24056 364 24057 364 24058 367 24059 441 24060 24060 24061 24061 24062 24062 24063 24063 24064 2 24065 583 24066 24066 24067 24067 24068 24068 24069 24069 24070 24070 24071 441 24072 24072 24073 24073 24074 1882 24075 164 24076 24076 24077 24077 24078 2 24079 583 24080 24080 24081 24081 24082 868 24083 11 24084 24084 24085 24085 24086 24086 24087 364 24088 364 24089 367 24090 441 24091 24091 24092 24092 24093 24093 24094 24094 24095 2 24096 583 24097 24097 24098 24098 24099 24099 24100 24100 24101 24101 24102 441 24103 24103 24104 24104 24105 1882 24106 164 24107 24107 24108 24108 24109 2 24110 583 24111 24111 24112 868 24113 20214 24114 11 24115 20216 24116 2 24117 583 24118 20219 24119 20220 24120 20221 24121 2756 24122 20223 24123 20224 24124 364 24125 364 24126 367 24127 441 24128 20229 24129 20230 24130 20231 24131 20232 24132 2 24133 583 24134 20235 24135 20236 24136 20237 24137 20238 24138 2827 24139 20240 24140 441 24141 20242 24142 2 24143 20244 24144 20245 24145 2 24146 3 24147 4 24148 2 24149 2 24150 2 24151 2 24152 11 24153 2 24154 2 24155 20256 24156 2 24157 20258 24158 22 24159 20260 24160 20261 24161 2 24162 11 24163 11 24164 20265 24165 20266 24166 20267 24167 22 24168 2 24169 2 24170 2 24171 11 24172 20273 24173 11 24174 20275 24175 11 24176 20277 24177 20278 24178 20279 24179 20280 24180 60 24181 20282 24182 2 24183 2 24184 20285 24185 11 24186 2 24187 2 24188 69 24189 133 24190 20291 24191 2 24192 2 24193 20294 24194 2 24195 20296 24196 20297 24197 2 24198 86 24199 87 24200 90 24201 136 24202 11 24203 20304 24204 20305 24205 101 24206 20307 24207 20308 24208 20309 24209 106 24210 106 24211 20312 24212 20313 24213 20314 24214 20315 24215 20316 24216 20317 24217 20318 24218 20319 24219 20320 24220 20321 24221 20322 24222 101 24223 2 24224 20325 24225 2 24226 2 24227 2 24228 2 24229 2 24230 2 24231 2 24232 2 24233 312 24234 2 24235 2 24236 2 24237 2 24238 2 24239 318 24240 2 24241 2 24242 2 24243 2 24244 2 24245 2 24246 2 24247 2 24248 2 24249 2 24250 2 24251 2 24252 2 24253 2 24254 2 24255 2 24256 20357 24257 20358 24258 20359 24259 20360 24260 20361 24261 20362 24262 20363 24263 20364 24264 20365 24265 20366 24266 20367 24267 20368 24268 20369 24269 20370 24270 20371 24271 20372 24272 4 24273 2 24274 20375 24275 20376 24276 20377 24277 20378 24278 20379 24279 20380 24280 20381 24281 20382 24282 20383 24283 20384 24284 20385 24285 364 24286 364 24287 40 24288 367 24289 441 24290 20391 24291 20392 24292 20393 24293 20394 24294 20395 24295 2 24296 583 24297 20398 24298 20399 24299 40 24300 20401 24301 287 24302 20403 24303 20404 24304 2 24305 583 24306 20407 24307 20408 24308 20409 24309 20410 24310 20411 24311 20412 24312 2 24313 751 24314 2 24315 20416 24316 20417 24317 20418 24318 20419 24319 20420 24320 20421 24321 20422 24322 20423 24323 20424 24324 20425 24325 868 24326 20427 24327 20428 24328 20429 24329 20430 24330 20431 24331 20432 24332 20433 24333 20434 24334 20435 24335 20436 24336 20437 24337 20438 24338 20439 24339 20440 24340 20441 24341 20442 24342 20443 24343 20444 24344 20445 24345 20446 24346 20447 24347 20448 24348 20449 24349 20450 24350 20451 24351 20452 24352 20453 24353 20454 24354 20455 24355 20456 24356 20457 24357 20458 24358 20459 24359 20460 24360 20461 24361 20462 24362 20463 24363 20464 24364 20465 24365 20466 24366 20467 24367 20468 24368 20469 24369 20470 24370 20471 24371 20472 24372 20473 24373 20474 24374 20475 24375 20476 24376 20477 24377 20478 24378 20479 24379 20480 24380 20481 24381 20482 24382 20483 24383 20484 24384 20485 24385 20486 24386 20487 24387 20488 24388 20489 24389 20490 24390 20491 24391 20492 24392 20493 24393 20494 24394 20495 24395 20496 24396 20497 24397 20498 24398 20499 24399 20500 24400 20501 24401 20502 24402 20503 24403 20504 24404 20505 24405 20506 24406 11 24407 20508 24408 20509 24409 20510 24410 2 24411 583 24412 20513 24413 956 24414 20515 24415 20516 24416 989 24417 161 24418 164 24419 20520 24420 20521 24421 1505 24422 2 24423 2 24424 2 24425 20526 24426 2 24427 20528 24428 2 24429 2 24430 2 24431 2 24432 2 24433 2 24434 2 24435 2 24436 2 24437 20538 24438 20539 24439 20540 24440 20541 24441 20542 24442 20543 24443 20544 24444 20545 24445 20546 24446 20547 24447 20548 24448 20549 24449 20550 24450 22 24451 20552 24452 4 24453 11 24454 20555 24455 90 24456 11 24457 20558 24458 101 24459 20560 24460 1569 24461 20562 24462 20563 24463 20564 24464 101 24465 868 24466 11 24467 20568 24468 956 24469 20570 24470 20571 24471 989 24472 161 24473 164 24474 20575 24475 20576 24476 20577 24477 868 24478 11 24479 20580 24480 20581 24481 20582 24482 364 24483 364 24484 367 24485 441 24486 20587 24487 20588 24488 20589 24489 20590 24490 2 24491 583 24492 20593 24493 20594 24494 20595 24495 20596 24496 20597 24497 20598 24498 2 24499 583 24500 20601 24501 20602 24502 142 24503 11 24504 144 24505 164 24506 165 24507 167 24508 169 24509 77 24510 128 24511 172 24512 20613 24513 20614 24514 142 24515 11 24516 144 24517 164 24518 2270 24519 2272 24520 2273 24521 2274 24522 2 24523 20624 24524 20625 24525 2 24526 20627 24527 868 24528 20629 24529 11 24530 20631 24531 2 24532 583 24533 20634 24534 20635 24535 20636 24536 2756 24537 20638 24538 20639 24539 364 24540 364 24541 367 24542 441 24543 20644 24544 20645 24545 20646 24546 20647 24547 2 24548 583 24549 20650 24550 20651 24551 20652 24552 20653 24553 2827 24554 20655 24555 441 24556 20657 24557 2 24558 20659 24559 20660 24560 20309 24561 2 24562 20663 24563 20664 24564 20665 24565 20666 24566 20667 24567 2918 24568 20669 24569 69 24570 20671 24571 20672 24572 20673 24573 20674 24574 20675 24575 20676 24576 20677 24577 2 24578 20679 24579 20680 24580 20681 24581 20682 24582 20683 24583 20684 24584 2960 24585 2962 24586 20687 24587 20688 24588 20689 24589 20690 24590 20691 24591 364 24592 364 24593 40 24594 367 24595 441 24596 20697 24597 20698 24598 20699 24599 20700 24600 20701 24601 40 24602 20703 24603 20704 24604 2983 24605 2986 24606 2988 24607 2990 24608 77 24609 2992 24610 2993 24611 20712 24612 287 24613 20714 24614 2 24615 287 24616 77 24617 3959 24618 2852 24619 20720 24620 1882 24621 164 24622 2072 24623 2 24624 20725 24625 2 24626 161 24627 164 24628 20729 24629 20730 24630 20731 24631 20732 24632 20733 24633 956 24634 20735 24635 20736 24636 20737 24637 989 24638 161 24639 164 24640 20741 24641 20742 24642 2 24643 287 24644 20745 24645 2 24646 287 24647 20748 24648 20749 24649 20750 24650 956 24651 20752 24652 20753 24653 20754 24654 989 24655 5285 24656 161 24657 164 24658 20759 24659 20760 24660 956 24661 20762 24662 20763 24663 20764 24664 20765 24665 1882 24666 164 24667 20768 24668 20769 24669 20770 24670 20771 24671 287 24672 20773 24673 2 24674 583 24675 20776 24676 5356 24677 20778 24678 20779 24679 20780 24680 20781 24681 20782 24682 20783 24683 2918 24684 20785 24685 69 24686 20787 24687 20788 24688 20789 24689 20790 24690 20791 24691 20792 24692 20793 24693 2 24694 20795 24695 20796 24696 20797 24697 20798 24698 20799 24699 20800 24700 2960 24701 2962 24702 20803 24703 20804 24704 20805 24705 20806 24706 20807 24707 364 24708 364 24709 40 24710 367 24711 441 24712 20813 24713 20814 24714 20815 24715 20816 24716 20817 24717 40 24718 20819 24719 20820 24720 2983 24721 2986 24722 2988 24723 2990 24724 77 24725 2992 24726 2993 24727 20828 24728 287 24729 20830 24730 2 24731 287 24732 77 24733 3959 24734 2852 24735 20836 24736 1882 24737 164 24738 2072 24739 2 24740 20841 24741 2 24742 161 24743 164 24744 20845 24745 20846 24746 20847 24747 20848 24748 20849 24749 956 24750 20851 24751 20852 24752 20853 24753 989 24754 161 24755 164 24756 20857 24757 20858 24758 2 24759 287 24760 20861 24761 2 24762 287 24763 20864 24764 20865 24765 20866 24766 956 24767 20868 24768 20869 24769 20870 24770 989 24771 5285 24772 161 24773 164 24774 20875 24775 20876 24776 956 24777 20878 24778 20879 24779 20880 24780 20881 24781 1882 24782 164 24783 20884 24784 20885 24785 20886 24786 20887 24787 287 24788 20889 24789 2 24790 583 24791 20892 24792 5356 24793 20894 24794 20895 24795 21125 24796 7507 24797 21134 24798 7507 24799 22736 24800 164 24801 17884 24802 22746 24803 24803 24804 24804 24805 22749 24806 7507 24807 24807 24808 3066 24809 24809 24810 22957 24811 11 24812 2 24813 287 24814 7507 24815 24815 24816 16 24817 23108 24818 17884 24819 23112 24820 24820 24821 24821 24822 23115 24823 7507 24824 24824 24825 7507 24826 24826 24827 7507 24828 24828 24829 24829 24830 17773 24831 142 24832 11 24833 144 24834 164 24835 167 24836 169 24837 77 24838 128 24839 172 24840 24840 24841 24841 24842 24842 24843 16 24844 2986 24845 2988 24846 2990 24847 77 24848 2992 24849 2993 24850 24850 24851 24851 24852 1080 24853 7507 24854 24854 24855 17805 24856 17806 24857 7507 24858 24858 24859 2 24860 287 24861 17884 24862 17885 24863 24863 24864 24864 24865 17888 24866 7507 24867 24867 24868 105 24869 24869 24870 22 24871 161 24872 164 24873 7507 24874 24874 24875 128 24876 172 24877 24877 24878 24878 24879 11 24880 77 24881 24881 24882 128 24883 172 24884 24884 24885 13 24886 18061 24887 164 24888 77 24889 128 24890 172 24891 24891 24892 2990 24893 77 24894 2992 24895 2993 24896 24896 24897 2 24898 287 24899 18074 24900 18075 24901 24901 24902 18077 24903 18078 24904 18079 24905 24905 24906 18079 24907 24907 24908 18217 24909 18218 24910 7507 24911 24911 24912 18218 24913 7507 24914 24914 24915 24915 24916 3066 24917 24917 24918 7507 24919 24919 24920 23404 24921 2990 24922 77 24923 2992 24924 2993 24925 24925 24926 2 24927 287 24928 17884 24929 24929 24930 164 24931 7507 24932 24932 24933 7507 24934 24934 24935 3066 24936 24936 24937 2 24938 19517 24939 20896 24940 20907 24941 17884 24942 20896 24943 164 24944 17884 24945 20958 24946 24946 24947 24947 24948 20961 24949 7507 24950 24950 24951 24951 24952 21125 24953 7507 24954 21134 24955 7507 24956 22736 24957 164 24958 17884 24959 22746 24960 24960 24961 24961 24962 22749 24963 7507 24964 24964 24965 23453 24966 2988 24967 2990 24968 77 24969 2992 24970 2993 24971 24971 24972 3066 24973 24973 24974 22957 24975 11 24976 2 24977 287 24978 7507 24979 24979 24980 16 24981 23108 24982 17884 24983 23112 24984 24984 24985 24985 24986 23115 24987 7507 24988 24988 24989 7507 24990 24990 24991 2986 24992 2988 24993 2990 24994 77 24995 2992 24996 2993 24997 24997 24998 7507 24999 24999 25000 25000 25001 2 25002 3 25003 4 25004 2 25005 2 25006 2 25007 2 25008 11 25009 2 25010 2 25011 25011 25012 2 25013 25013 25014 22 25015 25015 25016 25016 25017 2 25018 11 25019 11 25020 25020 25021 25021 25022 25022 25023 22 25024 2 25025 2 25026 2 25027 11 25028 25028 25029 11 25030 25030 25031 11 25032 25032 25033 25033 25034 25034 25035 25035 25036 60 25037 25037 25038 2 25039 2 25040 25040 25041 11 25042 2 25043 2 25044 69 25045 133 25046 25046 25047 2 25048 2 25049 25049 25050 2 25051 25051 25052 25052 25053 2 25054 86 25055 87 25056 90 25057 136 25058 11 25059 25059 25060 25060 25061 101 25062 25062 25063 25063 25064 25064 25065 106 25066 106 25067 25067 25068 25068 25069 25069 25070 25070 25071 25071 25072 25072 25073 25073 25074 25074 25075 25075 25076 25076 25077 25077 25078 101 25079 2 25080 25080 25081 2 25082 2 25083 2 25084 2 25085 2 25086 2 25087 2 25088 2 25089 312 25090 2 25091 2 25092 2 25093 2 25094 2 25095 318 25096 2 25097 2 25098 2 25099 2 25100 2 25101 2 25102 2 25103 2 25104 2 25105 2 25106 2 25107 2 25108 2 25109 2 25110 2 25111 2 25112 25112 25113 25113 25114 25114 25115 25115 25116 25116 25117 25117 25118 25118 25119 25119 25120 25120 25121 25121 25122 25122 25123 25123 25124 25124 25125 25125 25126 25126 25127 25127 25128 4 25129 2 25130 25130 25131 25131 25132 25132 25133 25133 25134 25134 25135 25135 25136 25136 25137 25137 25138 25138 25139 25139 25140 25140 25141 364 25142 364 25143 40 25144 367 25145 441 25146 25146 25147 25147 25148 25148 25149 25149 25150 25150 25151 2 25152 583 25153 25153 25154 25154 25155 40 25156 25156 25157 287 25158 25158 25159 25159 25160 2 25161 583 25162 25162 25163 25163 25164 25164 25165 25165 25166 25166 25167 25167 25168 2 25169 751 25170 2 25171 25171 25172 25172 25173 25173 25174 25174 25175 25175 25176 25176 25177 25177 25178 25178 25179 25179 25180 25180 25181 868 25182 25182 25183 25183 25184 25184 25185 25185 25186 25186 25187 25187 25188 25188 25189 25189 25190 25190 25191 25191 25192 25192 25193 25193 25194 25194 25195 25195 25196 25196 25197 25197 25198 25198 25199 25199 25200 25200 25201 25201 25202 25202 25203 25203 25204 25204 25205 25205 25206 25206 25207 25207 25208 25208 25209 25209 25210 25210 25211 25211 25212 25212 25213 25213 25214 25214 25215 25215 25216 25216 25217 25217 25218 25218 25219 25219 25220 25220 25221 25221 25222 25222 25223 25223 25224 25224 25225 25225 25226 25226 25227 25227 25228 25228 25229 25229 25230 25230 25231 25231 25232 25232 25233 25233 25234 25234 25235 25235 25236 25236 25237 25237 25238 25238 25239 25239 25240 25240 25241 25241 25242 25242 25243 25243 25244 25244 25245 25245 25246 25246 25247 25247 25248 25248 25249 25249 25250 25250 25251 25251 25252 25252 25253 25253 25254 25254 25255 25255 25256 25256 25257 25257 25258 25258 25259 25259 25260 25260 25261 25261 25262 11 25263 25263 25264 25264 25265 25265 25266 2 25267 583 25268 25268 25269 956 25270 25270 25271 25271 25272 989 25273 161 25274 164 25275 25275 25276 25276 25277 1505 25278 2 25279 2 25280 2 25281 25281 25282 2 25283 25283 25284 2 25285 2 25286 2 25287 2 25288 2 25289 2 25290 2 25291 2 25292 2 25293 25293 25294 25294 25295 25295 25296 25296 25297 25297 25298 25298 25299 25299 25300 25300 25301 25301 25302 25302 25303 25303 25304 25304 25305 25305 25306 22 25307 25307 25308 4 25309 11 25310 25310 25311 90 25312 11 25313 25313 25314 101 25315 25315 25316 1569 25317 25317 25318 25318 25319 25319 25320 101 25321 868 25322 11 25323 25323 25324 956 25325 25325 25326 25326 25327 989 25328 161 25329 164 25330 25330 25331 25331 25332 25332 25333 868 25334 11 25335 25335 25336 25336 25337 25337 25338 364 25339 364 25340 367 25341 441 25342 25342 25343 25343 25344 25344 25345 25345 25346 2 25347 583 25348 25348 25349 25349 25350 25350 25351 25351 25352 25352 25353 25353 25354 2 25355 583 25356 25356 25357 25357 25358 142 25359 11 25360 144 25361 164 25362 165 25363 167 25364 169 25365 77 25366 128 25367 172 25368 25368 25369 25369 25370 142 25371 11 25372 144 25373 164 25374 2270 25375 2272 25376 2273 25377 2274 25378 2 25379 2 25380 3 25381 4 25382 2 25383 2 25384 2 25385 2 25386 11 25387 2 25388 2 25389 25011 25390 2 25391 25013 25392 22 25393 25015 25394 25016 25395 2 25396 11 25397 11 25398 25020 25399 25021 25400 25022 25401 22 25402 2 25403 2 25404 2 25405 11 25406 25028 25407 11 25408 25030 25409 11 25410 25032 25411 25033 25412 25034 25413 25035 25414 60 25415 25037 25416 2 25417 2 25418 25040 25419 11 25420 2 25421 2 25422 69 25423 133 25424 25046 25425 2 25426 2 25427 25049 25428 2 25429 25051 25430 25052 25431 2 25432 86 25433 87 25434 90 25435 136 25436 11 25437 25059 25438 25060 25439 101 25440 25062 25441 25063 25442 25442 25443 106 25444 106 25445 25067 25446 25068 25447 25069 25448 25070 25449 25071 25450 25072 25451 25073 25452 25074 25453 25075 25454 25076 25455 25077 25456 101 25457 2 25458 25080 25459 2 25460 2 25461 2 25462 2 25463 2 25464 2 25465 2 25466 2 25467 312 25468 2 25469 2 25470 2 25471 2 25472 2 25473 318 25474 2 25475 2 25476 2 25477 2 25478 2 25479 2 25480 2 25481 2 25482 2 25483 2 25484 2 25485 2 25486 2 25487 2 25488 2 25489 2 25490 25112 25491 25113 25492 25114 25493 25115 25494 25116 25495 25117 25496 25118 25497 25119 25498 25120 25499 25121 25500 25122 25501 25123 25502 25124 25503 25125 25504 25126 25505 25127 25506 4 25507 2 25508 25130 25509 25131 25510 25132 25511 25133 25512 25134 25513 25135 25514 25136 25515 25137 25516 25138 25517 25139 25518 25140 25519 364 25520 364 25521 40 25522 367 25523 441 25524 25146 25525 25147 25526 25148 25527 25149 25528 25150 25529 2 25530 583 25531 25153 25532 25154 25533 40 25534 25156 25535 287 25536 25158 25537 25159 25538 2 25539 583 25540 25162 25541 25163 25542 25164 25543 25165 25544 25166 25545 25167 25546 2 25547 751 25548 2 25549 25171 25550 25172 25551 25173 25552 25174 25553 25175 25554 25176 25555 25177 25556 25178 25557 25179 25558 25180 25559 868 25560 25182 25561 25183 25562 25184 25563 25185 25564 25186 25565 25187 25566 25188 25567 25189 25568 25190 25569 25191 25570 25192 25571 25193 25572 25194 25573 25195 25574 25196 25575 25197 25576 25198 25577 25199 25578 25200 25579 25201 25580 25202 25581 25203 25582 25204 25583 25205 25584 25206 25585 25207 25586 25208 25587 25209 25588 25210 25589 25211 25590 25212 25591 25213 25592 25214 25593 25215 25594 25216 25595 25217 25596 25218 25597 25219 25598 25220 25599 25221 25600 25222 25601 25223 25602 25224 25603 25225 25604 25226 25605 25227 25606 25228 25607 25229 25608 25230 25609 25231 25610 25232 25611 25233 25612 25234 25613 25235 25614 25236 25615 25237 25616 25238 25617 25239 25618 25240 25619 25241 25620 25242 25621 25243 25622 25244 25623 25245 25624 25246 25625 25247 25626 25248 25627 25249 25628 25250 25629 25251 25630 25252 25631 25253 25632 25254 25633 25255 25634 25256 25635 25257 25636 25258 25637 25259 25638 25260 25639 25261 25640 11 25641 25263 25642 25264 25643 25265 25644 2 25645 583 25646 25268 25647 956 25648 25270 25649 25271 25650 989 25651 161 25652 164 25653 25275 25654 25276 25655 1505 25656 2 25657 2 25658 2 25659 25281 25660 2 25661 25283 25662 2 25663 2 25664 2 25665 2 25666 2 25667 2 25668 2 25669 2 25670 2 25671 25293 25672 25294 25673 25295 25674 25296 25675 25297 25676 25298 25677 25299 25678 25300 25679 25301 25680 25302 25681 25303 25682 25304 25683 25305 25684 22 25685 25307 25686 4 25687 11 25688 25310 25689 90 25690 11 25691 25313 25692 101 25693 25315 25694 1569 25695 25317 25696 25318 25697 25319 25698 101 25699 868 25700 11 25701 25323 25702 956 25703 25325 25704 25326 25705 989 25706 161 25707 164 25708 25330 25709 25331 25710 25332 25711 868 25712 11 25713 25335 25714 25336 25715 25337 25716 364 25717 364 25718 367 25719 441 25720 25342 25721 25343 25722 25344 25723 25345 25724 2 25725 583 25726 25348 25727 25349 25728 25350 25729 25351 25730 25352 25731 25353 25732 2 25733 583 25734 25356 25735 25357 25736 142 25737 11 25738 144 25739 164 25740 165 25741 167 25742 169 25743 77 25744 128 25745 172 25746 25368 25747 25369 25748 142 25749 11 25750 144 25751 164 25752 2270 25753 2272 25754 2273 25755 2274 25756 2 25757 25442 25758 25758 25759 25758 25760 868 25761 25761 25762 11 25763 25763 25764 2 25765 583 25766 25766 25767 25767 25768 25768 25769 2756 25770 25770 25771 25771 25772 364 25773 364 25774 367 25775 441 25776 25776 25777 25777 25778 25778 25779 25779 25780 2 25781 583 25782 25782 25783 25783 25784 25784 25785 25785 25786 2827 25787 25787 25788 441 25789 25789 25790 2 25791 25791 25792 25792 25793 2 25794 3 25795 4 25796 2 25797 2 25798 2 25799 2 25800 11 25801 2 25802 2 25803 25803 25804 2 25805 25805 25806 22 25807 25807 25808 25808 25809 2 25810 11 25811 11 25812 25812 25813 25813 25814 25814 25815 22 25816 2 25817 2 25818 2 25819 11 25820 25820 25821 11 25822 25822 25823 11 25824 25824 25825 25825 25826 25826 25827 25827 25828 60 25829 25829 25830 2 25831 2 25832 25832 25833 11 25834 2 25835 2 25836 69 25837 133 25838 25838 25839 2 25840 2 25841 25841 25842 2 25843 25843 25844 25844 25845 2 25846 86 25847 87 25848 90 25849 136 25850 11 25851 25851 25852 25852 25853 101 25854 25854 25855 25855 25856 25856 25857 106 25858 106 25859 25859 25860 25860 25861 25861 25862 25862 25863 25863 25864 25864 25865 25865 25866 25866 25867 25867 25868 25868 25869 25869 25870 101 25871 2 25872 25872 25873 2 25874 2 25875 2 25876 2 25877 2 25878 2 25879 2 25880 2 25881 312 25882 2 25883 2 25884 2 25885 2 25886 2 25887 318 25888 2 25889 2 25890 2 25891 2 25892 2 25893 2 25894 2 25895 2 25896 2 25897 2 25898 2 25899 2 25900 2 25901 2 25902 2 25903 2 25904 25904 25905 25905 25906 25906 25907 25907 25908 25908 25909 25909 25910 25910 25911 25911 25912 25912 25913 25913 25914 25914 25915 25915 25916 25916 25917 25917 25918 25918 25919 25919 25920 4 25921 2 25922 25922 25923 25923 25924 25924 25925 25925 25926 25926 25927 25927 25928 25928 25929 25929 25930 25930 25931 25931 25932 25932 25933 364 25934 364 25935 40 25936 367 25937 441 25938 25938 25939 25939 25940 25940 25941 25941 25942 25942 25943 2 25944 583 25945 25945 25946 25946 25947 40 25948 25948 25949 287 25950 25950 25951 25951 25952 2 25953 583 25954 25954 25955 25955 25956 25956 25957 25957 25958 25958 25959 25959 25960 2 25961 751 25962 2 25963 25963 25964 25964 25965 25965 25966 25966 25967 25967 25968 25968 25969 25969 25970 25970 25971 25971 25972 25972 25973 868 25974 25974 25975 25975 25976 25976 25977 25977 25978 25978 25979 25979 25980 25980 25981 25981 25982 25982 25983 25983 25984 25984 25985 25985 25986 25986 25987 25987 25988 25988 25989 25989 25990 25990 25991 25991 25992 25992 25993 25993 25994 25994 25995 25995 25996 25996 25997 25997 25998 25998 25999 25999 26000 26000 26001 26001 26002 26002 26003 26003 26004 26004 26005 26005 26006 26006 26007 26007 26008 26008 26009 26009 26010 26010 26011 26011 26012 26012 26013 26013 26014 26014 26015 26015 26016 26016 26017 26017 26018 26018 26019 26019 26020 26020 26021 26021 26022 26022 26023 26023 26024 26024 26025 26025 26026 26026 26027 26027 26028 26028 26029 26029 26030 26030 26031 26031 26032 26032 26033 26033 26034 26034 26035 26035 26036 26036 26037 26037 26038 26038 26039 26039 26040 26040 26041 26041 26042 26042 26043 26043 26044 26044 26045 26045 26046 26046 26047 26047 26048 26048 26049 26049 26050 26050 26051 26051 26052 26052 26053 26053 26054 11 26055 26055 26056 26056 26057 26057 26058 2 26059 583 26060 26060 26061 956 26062 26062 26063 26063 26064 989 26065 161 26066 164 26067 26067 26068 26068 26069 1505 26070 2 26071 2 26072 2 26073 26073 26074 2 26075 26075 26076 2 26077 2 26078 2 26079 2 26080 2 26081 2 26082 2 26083 2 26084 2 26085 26085 26086 26086 26087 26087 26088 26088 26089 26089 26090 26090 26091 26091 26092 26092 26093 26093 26094 26094 26095 26095 26096 26096 26097 26097 26098 22 26099 26099 26100 4 26101 11 26102 26102 26103 90 26104 11 26105 26105 26106 101 26107 26107 26108 1569 26109 26109 26110 26110 26111 26111 26112 101 26113 868 26114 11 26115 26115 26116 956 26117 26117 26118 26118 26119 989 26120 161 26121 164 26122 26122 26123 26123 26124 26124 26125 868 26126 11 26127 26127 26128 26128 26129 26129 26130 364 26131 364 26132 367 26133 441 26134 26134 26135 26135 26136 26136 26137 26137 26138 2 26139 583 26140 26140 26141 26141 26142 26142 26143 26143 26144 26144 26145 26145 26146 2 26147 583 26148 26148 26149 26149 26150 142 26151 11 26152 144 26153 164 26154 165 26155 167 26156 169 26157 77 26158 128 26159 172 26160 26160 26161 26161 26162 142 26163 11 26164 144 26165 164 26166 2270 26167 2272 26168 2273 26169 2274 26170 2 26171 26171 26172 26172 26173 2 26174 26174 26175 868 26176 26176 26177 11 26178 26178 26179 2 26180 583 26181 26181 26182 26182 26183 26183 26184 2756 26185 26185 26186 26186 26187 364 26188 364 26189 367 26190 441 26191 26191 26192 26192 26193 26193 26194 26194 26195 2 26196 583 26197 26197 26198 26198 26199 26199 26200 26200 26201 2827 26202 26202 26203 441 26204 26204 26205 2 26206 26206 26207 26207 26208 25856 26209 2 26210 26210 26211 26211 26212 26212 26213 26213 26214 26214 26215 2918 26216 26216 26217 69 26218 26218 26219 26219 26220 26220 26221 26221 26222 26222 26223 26223 26224 26224 26225 2 26226 26226 26227 26227 26228 26228 26229 26229 26230 26230 26231 26231 26232 2960 26233 2962 26234 26234 26235 26235 26236 26236 26237 26237 26238 26238 26239 364 26240 364 26241 40 26242 367 26243 441 26244 26244 26245 26245 26246 26246 26247 26247 26248 26248 26249 40 26250 26250 26251 26251 26252 2983 26253 2986 26254 2988 26255 2990 26256 77 26257 2992 26258 2993 26259 26259 26260 287 26261 26261 26262 2 26263 287 26264 77 26265 3959 26266 2852 26267 26267 26268 1882 26269 164 26270 2072 26271 2 26272 26272 26273 2 26274 161 26275 164 26276 26276 26277 26277 26278 26278 26279 26279 26280 26280 26281 956 26282 26282 26283 26283 26284 26284 26285 989 26286 161 26287 164 26288 26288 26289 26289 26290 2 26291 287 26292 26292 26293 2 26294 287 26295 26295 26296 26296 26297 26297 26298 956 26299 26299 26300 26300 26301 26301 26302 989 26303 5285 26304 161 26305 164 26306 26306 26307 26307 26308 956 26309 26309 26310 26310 26311 26311 26312 26312 26313 1882 26314 164 26315 26315 26316 26316 26317 26317 26318 26318 26319 287 26320 26320 26321 2 26322 583 26323 26323 26324 5356 26325 26325 26326 26326 26327 26327 26328 26328 26329 26329 26330 26330 26331 2918 26332 26332 26333 69 26334 26334 26335 26335 26336 26336 26337 26337 26338 26338 26339 26339 26340 26340 26341 2 26342 26342 26343 26343 26344 26344 26345 26345 26346 26346 26347 26347 26348 2960 26349 2962 26350 26350 26351 26351 26352 26352 26353 26353 26354 26354 26355 364 26356 364 26357 40 26358 367 26359 441 26360 26360 26361 26361 26362 26362 26363 26363 26364 26364 26365 40 26366 26366 26367 26367 26368 2983 26369 2986 26370 2988 26371 2990 26372 77 26373 2992 26374 2993 26375 26375 26376 287 26377 26377 26378 2 26379 287 26380 77 26381 3959 26382 2852 26383 26383 26384 1882 26385 164 26386 2072 26387 2 26388 26388 26389 2 26390 161 26391 164 26392 26392 26393 26393 26394 26394 26395 26395 26396 26396 26397 956 26398 26398 26399 26399 26400 26400 26401 989 26402 161 26403 164 26404 26404 26405 26405 26406 2 26407 287 26408 26408 26409 2 26410 287 26411 26411 26412 26412 26413 26413 26414 956 26415 26415 26416 26416 26417 26417 26418 989 26419 5285 26420 161 26421 164 26422 26422 26423 26423 26424 956 26425 26425 26426 26426 26427 26427 26428 26428 26429 1882 26430 164 26431 26431 26432 26432 26433 26433 26434 26434 26435 287 26436 26436 26437 2 26438 583 26439 26439 26440 5356 26441 26441 26442 26442 26443 18223 26444 18223 26445 26445 26446 26446 26447 868 26448 25761 26449 11 26450 25763 26451 2 26452 583 26453 25766 26454 25767 26455 25768 26456 2756 26457 25770 26458 25771 26459 364 26460 364 26461 367 26462 441 26463 25776 26464 25777 26465 25778 26466 25779 26467 2 26468 583 26469 25782 26470 25783 26471 25784 26472 25785 26473 2827 26474 25787 26475 441 26476 25789 26477 2 26478 25791 26479 25792 26480 2 26481 3 26482 4 26483 2 26484 2 26485 2 26486 2 26487 11 26488 2 26489 2 26490 25803 26491 2 26492 25805 26493 22 26494 25807 26495 25808 26496 2 26497 11 26498 11 26499 25812 26500 25813 26501 25814 26502 22 26503 2 26504 2 26505 2 26506 11 26507 25820 26508 11 26509 25822 26510 11 26511 25824 26512 25825 26513 25826 26514 25827 26515 60 26516 25829 26517 2 26518 2 26519 25832 26520 11 26521 2 26522 2 26523 69 26524 133 26525 25838 26526 2 26527 2 26528 25841 26529 2 26530 25843 26531 25844 26532 2 26533 86 26534 87 26535 90 26536 136 26537 11 26538 25851 26539 25852 26540 101 26541 25854 26542 25855 26543 25856 26544 106 26545 106 26546 25859 26547 25860 26548 25861 26549 25862 26550 25863 26551 25864 26552 25865 26553 25866 26554 25867 26555 25868 26556 25869 26557 101 26558 2 26559 25872 26560 2 26561 2 26562 2 26563 2 26564 2 26565 2 26566 2 26567 2 26568 312 26569 2 26570 2 26571 2 26572 2 26573 2 26574 318 26575 2 26576 2 26577 2 26578 2 26579 2 26580 2 26581 2 26582 2 26583 2 26584 2 26585 2 26586 2 26587 2 26588 2 26589 2 26590 2 26591 25904 26592 25905 26593 25906 26594 25907 26595 25908 26596 25909 26597 25910 26598 25911 26599 25912 26600 25913 26601 25914 26602 25915 26603 25916 26604 25917 26605 25918 26606 25919 26607 4 26608 2 26609 25922 26610 25923 26611 25924 26612 25925 26613 25926 26614 25927 26615 25928 26616 25929 26617 25930 26618 25931 26619 25932 26620 364 26621 364 26622 40 26623 367 26624 441 26625 25938 26626 25939 26627 25940 26628 25941 26629 25942 26630 2 26631 583 26632 25945 26633 25946 26634 40 26635 25948 26636 287 26637 25950 26638 25951 26639 2 26640 583 26641 25954 26642 25955 26643 25956 26644 25957 26645 25958 26646 25959 26647 2 26648 751 26649 2 26650 25963 26651 25964 26652 25965 26653 25966 26654 25967 26655 25968 26656 25969 26657 25970 26658 25971 26659 25972 26660 868 26661 25974 26662 25975 26663 25976 26664 25977 26665 25978 26666 25979 26667 25980 26668 25981 26669 25982 26670 25983 26671 25984 26672 25985 26673 25986 26674 25987 26675 25988 26676 25989 26677 25990 26678 25991 26679 25992 26680 25993 26681 25994 26682 25995 26683 25996 26684 25997 26685 25998 26686 25999 26687 26000 26688 26001 26689 26002 26690 26003 26691 26004 26692 26005 26693 26006 26694 26007 26695 26008 26696 26009 26697 26010 26698 26011 26699 26012 26700 26013 26701 26014 26702 26015 26703 26016 26704 26017 26705 26018 26706 26019 26707 26020 26708 26021 26709 26022 26710 26023 26711 26024 26712 26025 26713 26026 26714 26027 26715 26028 26716 26029 26717 26030 26718 26031 26719 26032 26720 26033 26721 26034 26722 26035 26723 26036 26724 26037 26725 26038 26726 26039 26727 26040 26728 26041 26729 26042 26730 26043 26731 26044 26732 26045 26733 26046 26734 26047 26735 26048 26736 26049 26737 26050 26738 26051 26739 26052 26740 26053 26741 11 26742 26055 26743 26056 26744 26057 26745 2 26746 583 26747 26060 26748 956 26749 26062 26750 26063 26751 989 26752 161 26753 164 26754 26067 26755 26068 26756 1505 26757 2 26758 2 26759 2 26760 26073 26761 2 26762 26075 26763 2 26764 2 26765 2 26766 2 26767 2 26768 2 26769 2 26770 2 26771 2 26772 26085 26773 26086 26774 26087 26775 26088 26776 26089 26777 26090 26778 26091 26779 26092 26780 26093 26781 26094 26782 26095 26783 26096 26784 26097 26785 22 26786 26099 26787 4 26788 11 26789 26102 26790 90 26791 11 26792 26105 26793 101 26794 26107 26795 1569 26796 26109 26797 26110 26798 26111 26799 101 26800 868 26801 11 26802 26115 26803 956 26804 26117 26805 26118 26806 989 26807 161 26808 164 26809 26122 26810 26123 26811 26124 26812 868 26813 11 26814 26127 26815 26128 26816 26129 26817 364 26818 364 26819 367 26820 441 26821 26134 26822 26135 26823 26136 26824 26137 26825 2 26826 583 26827 26140 26828 26141 26829 26142 26830 26143 26831 26144 26832 26145 26833 2 26834 583 26835 26148 26836 26149 26837 142 26838 11 26839 144 26840 164 26841 165 26842 167 26843 169 26844 77 26845 128 26846 172 26847 26160 26848 26161 26849 142 26850 11 26851 144 26852 164 26853 2270 26854 2272 26855 2273 26856 2274 26857 2 26858 26171 26859 26172 26860 2 26861 26174 26862 868 26863 26176 26864 11 26865 26178 26866 2 26867 583 26868 26181 26869 26182 26870 26183 26871 2756 26872 26185 26873 26186 26874 364 26875 364 26876 367 26877 441 26878 26191 26879 26192 26880 26193 26881 26194 26882 2 26883 583 26884 26197 26885 26198 26886 26199 26887 26200 26888 2827 26889 26202 26890 441 26891 26204 26892 2 26893 26206 26894 26207 26895 25856 26896 2 26897 26210 26898 26211 26899 26212 26900 26213 26901 26214 26902 2918 26903 26216 26904 69 26905 26218 26906 26219 26907 26220 26908 26221 26909 26222 26910 26223 26911 26224 26912 2 26913 26226 26914 26227 26915 26228 26916 26229 26917 26230 26918 26231 26919 2960 26920 2962 26921 26234 26922 26235 26923 26236 26924 26237 26925 26238 26926 364 26927 364 26928 40 26929 367 26930 441 26931 26244 26932 26245 26933 26246 26934 26247 26935 26248 26936 40 26937 26250 26938 26251 26939 2983 26940 2986 26941 2988 26942 2990 26943 77 26944 2992 26945 2993 26946 26259 26947 287 26948 26261 26949 2 26950 287 26951 77 26952 3959 26953 2852 26954 26267 26955 1882 26956 164 26957 2072 26958 2 26959 26272 26960 2 26961 161 26962 164 26963 26276 26964 26277 26965 26278 26966 26279 26967 26280 26968 956 26969 26282 26970 26283 26971 26284 26972 989 26973 161 26974 164 26975 26288 26976 26289 26977 2 26978 287 26979 26292 26980 2 26981 287 26982 26295 26983 26296 26984 26297 26985 956 26986 26299 26987 26300 26988 26301 26989 989 26990 5285 26991 161 26992 164 26993 26306 26994 26307 26995 956 26996 26309 26997 26310 26998 26311 26999 26312 27000 1882 27001 164 27002 26315 27003 26316 27004 26317 27005 26318 27006 287 27007 26320 27008 2 27009 583 27010 26323 27011 5356 27012 26325 27013 26326 27014 26327 27015 26328 27016 26329 27017 26330 27018 2918 27019 26332 27020 69 27021 26334 27022 26335 27023 26336 27024 26337 27025 26338 27026 26339 27027 26340 27028 2 27029 26342 27030 26343 27031 26344 27032 26345 27033 26346 27034 26347 27035 2960 27036 2962 27037 26350 27038 26351 27039 26352 27040 26353 27041 26354 27042 364 27043 364 27044 40 27045 367 27046 441 27047 26360 27048 26361 27049 26362 27050 26363 27051 26364 27052 40 27053 26366 27054 26367 27055 2983 27056 2986 27057 2988 27058 2990 27059 77 27060 2992 27061 2993 27062 26375 27063 287 27064 26377 27065 2 27066 287 27067 77 27068 3959 27069 2852 27070 26383 27071 1882 27072 164 27073 2072 27074 2 27075 26388 27076 2 27077 161 27078 164 27079 26392 27080 26393 27081 26394 27082 26395 27083 26396 27084 956 27085 26398 27086 26399 27087 26400 27088 989 27089 161 27090 164 27091 26404 27092 26405 27093 2 27094 287 27095 26408 27096 2 27097 287 27098 26411 27099 26412 27100 26413 27101 956 27102 26415 27103 26416 27104 26417 27105 989 27106 5285 27107 161 27108 164 27109 26422 27110 26423 27111 956 27112 26425 27113 26426 27114 26427 27115 26428 27116 1882 27117 164 27118 26431 27119 26432 27120 26433 27121 26434 27122 287 27123 26436 27124 2 27125 583 27126 26439 27127 5356 27128 26441 27129 26442 27130 18218 27131 7507 27132 27132 27133 27133 27134 3066 27135 27135 27136 7507 27137 27137 27138 7507 27139 27139 27140 3066 27141 27141 27142 2 27143 19517 27144 20896 27145 20907 27146 17884 27147 20896 27148 164 27149 17884 27150 20958 27151 27151 27152 27152 27153 20961 27154 7507 27155 27155 27156 27156 27157 21125 27158 7507 27159 21134 27160 7507 27161 22736 27162 164 27163 17884 27164 22746 27165 27165 27166 27166 27167 22749 27168 7507 27169 27169 27170 3066 27171 27171 27172 22957 27173 11 27174 2 27175 287 27176 7507 27177 27177 27178 16 27179 23108 27180 17884 27181 23112 27182 27182 27183 27183 27184 23115 27185 7507 27186 27186 27187 7507 27188 27188 27189 7507 27190 27190 27191 27191 27192 17773 27193 142 27194 11 27195 144 27196 164 27197 167 27198 169 27199 77 27200 128 27201 172 27202 27202 27203 27203 27204 27204 27205 16 27206 2986 27207 2988 27208 2990 27209 77 27210 2992 27211 2993 27212 27212 27213 27213 27214 1080 27215 7507 27216 27216 27217 17805 27218 17806 27219 7507 27220 27220 27221 2 27222 287 27223 17884 27224 17885 27225 27225 27226 27226 27227 17888 27228 7507 27229 27229 27230 105 27231 27231 27232 22 27233 161 27234 164 27235 7507 27236 27236 27237 128 27238 172 27239 27239 27240 27240 27241 11 27242 77 27243 27243 27244 128 27245 172 27246 27246 27247 13 27248 18061 27249 164 27250 77 27251 128 27252 172 27253 27253 27254 2990 27255 77 27256 2992 27257 2993 27258 27258 27259 2 27260 287 27261 18074 27262 18075 27263 27263 27264 18077 27265 18078 27266 18079 27267 27267 27268 18079 27269 27269 27270 18217 27271 18218 27272 7507 27273 27273 27274 18218 27275 7507 27276 27276 27277 27277 27278 3066 27279 27279 27280 7507 27281 27281 27282 23404 27283 2990 27284 77 27285 2992 27286 2993 27287 27287 27288 2 27289 287 27290 17884 27291 27291 27292 164 27293 7507 27294 27294 27295 7507 27296 27296 27297 3066 27298 27298 27299 2 27300 19517 27301 20896 27302 20907 27303 17884 27304 20896 27305 164 27306 17884 27307 20958 27308 27308 27309 27309 27310 20961 27311 7507 27312 27312 27313 27313 27314 21125 27315 7507 27316 21134 27317 7507 27318 22736 27319 164 27320 17884 27321 22746 27322 27322 27323 27323 27324 22749 27325 7507 27326 27326 27327 23453 27328 2988 27329 2990 27330 77 27331 2992 27332 2993 27333 27333 27334 3066 27335 27335 27336 22957 27337 11 27338 2 27339 287 27340 7507 27341 27341 27342 16 27343 23108 27344 17884 27345 23112 27346 27346 27347 27347 27348 23115 27349 7507 27350 27350 27351 7507 27352 27352 27353 2986 27354 2988 27355 2990 27356 77 27357 2992 27358 2993 27359 27359 27360 7507 27361 27361 27362 27362 27363 25064 27364 18083 27365 18083 27366 27366 27367 27367 27368 18079 27369 27369 27370 18217 27371 18218 27372 7507 27373 27373 27374 18218 27375 7507 27376 27376 27377 27377 27378 3066 27379 27379 27380 7507 27381 27381 27382 7507 27383 27383 27384 3066 27385 27385 27386 2 27387 19517 27388 20896 27389 20907 27390 17884 27391 20896 27392 164 27393 17884 27394 20958 27395 27395 27396 27396 27397 20961 27398 7507 27399 27399 27400 27400 27401 21125 27402 7507 27403 21134 27404 7507 27405 22736 27406 164 27407 17884 27408 22746 27409 27409 27410 27410 27411 22749 27412 7507 27413 27413 27414 3066 27415 27415 27416 22957 27417 11 27418 2 27419 287 27420 7507 27421 27421 27422 16 27423 23108 27424 17884 27425 23112 27426 27426 27427 27427 27428 23115 27429 7507 27430 27430 27431 7507 27432 27432 27433 7507 27434 27434 27435 27435 27436 17773 27437 142 27438 11 27439 144 27440 164 27441 167 27442 169 27443 77 27444 128 27445 172 27446 27446 27447 27447 27448 27448 27449 16 27450 2986 27451 2988 27452 2990 27453 77 27454 2992 27455 2993 27456 27456 27457 27457 27458 1080 27459 7507 27460 27460 27461 17805 27462 17806 27463 7507 27464 27464 27465 2 27466 287 27467 17884 27468 17885 27469 27469 27470 27470 27471 17888 27472 7507 27473 27473 27474 105 27475 27475 27476 22 27477 161 27478 164 27479 7507 27480 27480 27481 128 27482 172 27483 27483 27484 27484 27485 11 27486 77 27487 27487 27488 128 27489 172 27490 27490 27491 13 27492 18061 27493 164 27494 77 27495 128 27496 172 27497 27497 27498 2990 27499 77 27500 2992 27501 2993 27502 27502 27503 2 27504 287 27505 18074 27506 18075 27507 27507 27508 18077 27509 18078 27510 18079 27511 27511 27512 18079 27513 27513 27514 18217 27515 18218 27516 7507 27517 27517 27518 18218 27519 7507 27520 27520 27521 27521 27522 3066 27523 27523 27524 7507 27525 27525 27526 23404 27527 2990 27528 77 27529 2992 27530 2993 27531 27531 27532 2 27533 287 27534 17884 27535 27535 27536 164 27537 7507 27538 27538 27539 7507 27540 27540 27541 3066 27542 27542 27543 2 27544 19517 27545 20896 27546 20907 27547 17884 27548 20896 27549 164 27550 17884 27551 20958 27552 27552 27553 27553 27554 20961 27555 7507 27556 27556 27557 27557 27558 21125 27559 7507 27560 21134 27561 7507 27562 22736 27563 164 27564 17884 27565 22746 27566 27566 27567 27567 27568 22749 27569 7507 27570 27570 27571 23453 27572 2988 27573 2990 27574 77 27575 2992 27576 2993 27577 27577 27578 27578 27579 3066 27580 27580 27581 22957 27582 11 27583 2 27584 287 27585 7507 27586 27586 27587 16 27588 23108 27589 17884 27590 23112 27591 27591 27592 27592 27593 23115 27594 7507 27595 27595 27596 7507 27597 27597 27598 2986 27599 2988 27600 2990 27601 77 27602 2992 27603 2993 27604 27604 27605 7507 27606 27606 27607 105 27608 27608 27609 22 27610 161 27611 164 27612 7507 27613 27613 27614 128 27615 172 27616 27616 27617 27617 27618 11 27619 77 27620 27620 27621 128 27622 172 27623 27623 27624 13 27625 18061 27626 164 27627 77 27628 128 27629 172 27630 27630 27631 2990 27632 77 27633 2992 27634 2993 27635 27635 27636 2 27637 287 27638 18074 27639 18075 27640 27640 27641 18077 27642 18078 27643 18079 27644 27644 27645 18079 27646 27646 27647 18217 27648 18218 27649 7507 27650 27650 27651 18218 27652 7507 27653 27653 27654 27654 27655 3066 27656 27656 27657 7507 27658 27658 27659 23404 27660 2990 27661 77 27662 2992 27663 2993 27664 27664 27665 164 27666 7507 27667 27667 27668 7507 27669 27669 27670 3066 27671 27671 27672 2 27673 19517 27674 20896 27675 20907 27676 17884 27677 20896 27678 164 27679 17884 27680 20958 27681 27681 27682 27682 27683 20961 27684 7507 27685 27685 27686 27686 27687 21125 27688 7507 27689 21134 27690 7507 27691 22736 27692 164 27693 17884 27694 22746 27695 27695 27696 27696 27697 22749 27698 7507 27699 27699 27700 23453 27701 2988 27702 2990 27703 77 27704 2992 27705 2993 27706 27706 27707 27707 27708 3066 27709 27709 27710 22957 27711 11 27712 2 27713 287 27714 7507 27715 27715 27716 16 27717 23108 27718 17884 27719 23112 27720 27720 27721 27721 27722 23115 27723 7507 27724 27724 27725 7507 27726 27726 27727 2986 27728 2988 27729 2990 27730 77 27731 2992 27732 2993 27733 27733 27734 7507 27735 27735 27736 2 27737 287 27738 17884 27739 17885 27740 27740 27741 27741 27742 17888 27743 7507 27744 27744 27745 105 27746 27746 27747 22 27748 161 27749 164 27750 7507 27751 27751 27752 128 27753 172 27754 27754 27755 27755 27756 11 27757 77 27758 27758 27759 128 27760 172 27761 27761 27762 13 27763 18061 27764 164 27765 77 27766 128 27767 172 27768 27768 27769 2990 27770 77 27771 2992 27772 2993 27773 27773 27774 2 27775 287 27776 18074 27777 18075 27778 27778 27779 18077 27780 18078 27781 18079 27782 27782 27783 18079 27784 27784 27785 18217 27786 18218 27787 7507 27788 27788 27789 18218 27790 7507 27791 27791 27792 27792 27793 3066 27794 27794 27795 7507 27796 27796 27797 23404 27798 2990 27799 77 27800 2992 27801 2993 27802 27802 27803 2 27804 287 27805 17884 27806 27806 27807 164 27808 7507 27809 27809 27810 7507 27811 27811 27812 3066 27813 27813 27814 2 27815 19517 27816 20896 27817 20907 27818 17884 27819 20896 27820 164 27821 17884 27822 20958 27823 27823 27824 27824 27825 20961 27826 7507 27827 27827 27828 27828 27829 21125 27830 7507 27831 21134 27832 7507 27833 22736 27834 164 27835 17884 27836 22746 27837 27837 27838 27838 27839 22749 27840 7507 27841 27841 27842 23453 27843 2988 27844 2990 27845 77 27846 2992 27847 2993 27848 27848 27849 3066 27850 27850 27851 22957 27852 11 27853 2 27854 287 27855 7507 27856 27856 27857 16 27858 23108 27859 17884 27860 23112 27861 27861 27862 27862 27863 23115 27864 7507 27865 27865 27866 7507 27867 27867 27868 2986 27869 2988 27870 2990 27871 77 27872 2992 27873 2993 27874 27874 27875 7507 27876 27876 27877 17805 27878 17806 27879 7507 27880 27880 27881 2 27882 287 27883 17884 27884 17885 27885 27885 27886 27886 27887 17888 27888 7507 27889 27889 27890 105 27891 27891 27892 22 27893 161 27894 164 27895 7507 27896 27896 27897 128 27898 172 27899 27899 27900 27900 27901 11 27902 77 27903 27903 27904 128 27905 172 27906 27906 27907 13 27908 18061 27909 164 27910 77 27911 128 27912 172 27913 27913 27914 2 27915 287 27916 18074 27917 18075 27918 27918 27919 18077 27920 18078 27921 18079 27922 27922 27923 18079 27924 27924 27925 18217 27926 18218 27927 7507 27928 27928 27929 18218 27930 7507 27931 27931 27932 27932 27933 3066 27934 27934 27935 7507 27936 27936 27937 7507 27938 27938 27939 3066 27940 27940 27941 2 27942 19517 27943 20896 27944 20907 27945 17884 27946 20896 27947 164 27948 17884 27949 20958 27950 27950 27951 27951 27952 20961 27953 7507 27954 27954 27955 27955 27956 21125 27957 7507 27958 21134 27959 7507 27960 22736 27961 164 27962 17884 27963 22746 27964 27964 27965 27965 27966 22749 27967 7507 27968 27968 27969 3066 27970 27970 27971 22957 27972 11 27973 2 27974 287 27975 7507 27976 27976 27977 16 27978 23108 27979 17884 27980 23112 27981 27981 27982 27982 27983 23115 27984 7507 27985 27985 27986 7507 27987 27987 27988 7507 27989 27989 27990 868 27991 16996 27992 11 27993 16998 27994 2 27995 583 27996 17001 27997 17002 27998 17003 27999 2756 28000 17005 28001 17006 28002 364 28003 364 28004 367 28005 441 28006 17011 28007 17012 28008 17013 28009 17014 28010 2 28011 583 28012 17017 28013 17018 28014 17019 28015 17020 28016 2827 28017 17022 28018 441 28019 17024 28020 2 28021 17026 28022 17027 28023 2 28024 3 28025 4 28026 2 28027 2 28028 2 28029 2 28030 11 28031 2 28032 2 28033 17038 28034 2 28035 17040 28036 22 28037 17042 28038 17043 28039 2 28040 11 28041 11 28042 17047 28043 17048 28044 17049 28045 22 28046 2 28047 2 28048 2 28049 11 28050 17055 28051 11 28052 17057 28053 11 28054 17059 28055 17060 28056 17061 28057 17062 28058 60 28059 17064 28060 2 28061 2 28062 17067 28063 11 28064 2 28065 2 28066 69 28067 133 28068 17073 28069 2 28070 2 28071 17076 28072 2 28073 17078 28074 17079 28075 2 28076 86 28077 87 28078 90 28079 136 28080 11 28081 17086 28082 17087 28083 101 28084 17089 28085 17090 28086 17091 28087 106 28088 106 28089 17094 28090 17095 28091 17096 28092 17097 28093 17098 28094 17099 28095 17100 28096 17101 28097 17102 28098 17103 28099 17104 28100 101 28101 2 28102 17107 28103 2 28104 2 28105 2 28106 2 28107 2 28108 2 28109 2 28110 2 28111 312 28112 2 28113 2 28114 2 28115 2 28116 2 28117 318 28118 2 28119 2 28120 2 28121 2 28122 2 28123 2 28124 2 28125 2 28126 2 28127 2 28128 2 28129 2 28130 2 28131 2 28132 2 28133 2 28134 17139 28135 17140 28136 17141 28137 17142 28138 17143 28139 17144 28140 17145 28141 17146 28142 17147 28143 17148 28144 17149 28145 17150 28146 17151 28147 17152 28148 17153 28149 17154 28150 4 28151 2 28152 17157 28153 17158 28154 17159 28155 17160 28156 17161 28157 17162 28158 17163 28159 17164 28160 17165 28161 17166 28162 17167 28163 364 28164 364 28165 40 28166 367 28167 441 28168 17173 28169 17174 28170 17175 28171 17176 28172 17177 28173 2 28174 583 28175 17180 28176 17181 28177 40 28178 17183 28179 287 28180 17185 28181 17186 28182 2 28183 583 28184 17189 28185 17190 28186 17191 28187 17192 28188 17193 28189 17194 28190 2 28191 751 28192 2 28193 17198 28194 17199 28195 17200 28196 17201 28197 17202 28198 17203 28199 17204 28200 17205 28201 17206 28202 17207 28203 868 28204 17209 28205 17210 28206 17211 28207 17212 28208 17213 28209 17214 28210 17215 28211 17216 28212 17217 28213 17218 28214 17219 28215 17220 28216 17221 28217 17222 28218 17223 28219 17224 28220 17225 28221 17226 28222 17227 28223 17228 28224 17229 28225 17230 28226 17231 28227 17232 28228 17233 28229 17234 28230 17235 28231 17236 28232 17237 28233 17238 28234 17239 28235 17240 28236 17241 28237 17242 28238 17243 28239 17244 28240 17245 28241 17246 28242 17247 28243 17248 28244 17249 28245 17250 28246 17251 28247 17252 28248 17253 28249 17254 28250 17255 28251 17256 28252 17257 28253 17258 28254 17259 28255 17260 28256 17261 28257 17262 28258 17263 28259 17264 28260 17265 28261 17266 28262 17267 28263 17268 28264 17269 28265 17270 28266 17271 28267 17272 28268 17273 28269 17274 28270 17275 28271 17276 28272 17277 28273 17278 28274 17279 28275 17280 28276 17281 28277 17282 28278 17283 28279 17284 28280 17285 28281 17286 28282 17287 28283 17288 28284 11 28285 17290 28286 17291 28287 17292 28288 2 28289 583 28290 17295 28291 956 28292 17297 28293 17298 28294 989 28295 161 28296 164 28297 17302 28298 17303 28299 1505 28300 2 28301 2 28302 2 28303 17308 28304 2 28305 17310 28306 2 28307 2 28308 2 28309 2 28310 2 28311 2 28312 2 28313 2 28314 2 28315 17320 28316 17321 28317 17322 28318 17323 28319 17324 28320 17325 28321 17326 28322 17327 28323 17328 28324 17329 28325 17330 28326 17331 28327 17332 28328 22 28329 17334 28330 4 28331 11 28332 17337 28333 90 28334 11 28335 17340 28336 101 28337 17342 28338 1569 28339 17344 28340 17345 28341 17346 28342 101 28343 868 28344 11 28345 17350 28346 956 28347 17352 28348 17353 28349 989 28350 161 28351 164 28352 17357 28353 17358 28354 17359 28355 868 28356 11 28357 17362 28358 17363 28359 17364 28360 364 28361 364 28362 367 28363 441 28364 17369 28365 17370 28366 17371 28367 17372 28368 2 28369 583 28370 17375 28371 17376 28372 17377 28373 17378 28374 17379 28375 17380 28376 2 28377 583 28378 17383 28379 17384 28380 142 28381 11 28382 144 28383 164 28384 165 28385 167 28386 169 28387 77 28388 128 28389 172 28390 17395 28391 17396 28392 142 28393 11 28394 144 28395 164 28396 2270 28397 2272 28398 2273 28399 2274 28400 2 28401 17406 28402 17407 28403 2 28404 17409 28405 868 28406 17411 28407 11 28408 17413 28409 2 28410 583 28411 17416 28412 17417 28413 17418 28414 2756 28415 17420 28416 17421 28417 364 28418 364 28419 367 28420 441 28421 17426 28422 17427 28423 17428 28424 17429 28425 2 28426 583 28427 17432 28428 17433 28429 17434 28430 17435 28431 2827 28432 17437 28433 441 28434 17439 28435 2 28436 17441 28437 17442 28438 17091 28439 2 28440 17445 28441 17446 28442 17447 28443 17448 28444 17449 28445 2918 28446 17451 28447 69 28448 17453 28449 17454 28450 17455 28451 17456 28452 17457 28453 17458 28454 17459 28455 2 28456 17461 28457 17462 28458 17463 28459 17464 28460 17465 28461 17466 28462 2960 28463 2962 28464 17469 28465 17470 28466 17471 28467 17472 28468 17473 28469 364 28470 364 28471 40 28472 367 28473 441 28474 17479 28475 17480 28476 17481 28477 17482 28478 17483 28479 40 28480 17485 28481 17486 28482 2983 28483 2986 28484 2988 28485 2990 28486 77 28487 2992 28488 2993 28489 17494 28490 287 28491 17496 28492 2 28493 287 28494 77 28495 3959 28496 2852 28497 17502 28498 1882 28499 164 28500 2072 28501 2 28502 17507 28503 2 28504 161 28505 164 28506 17511 28507 17512 28508 17513 28509 17514 28510 17515 28511 956 28512 17517 28513 17518 28514 17519 28515 989 28516 161 28517 164 28518 17523 28519 17524 28520 2 28521 287 28522 17527 28523 2 28524 287 28525 17530 28526 17531 28527 17532 28528 956 28529 17534 28530 17535 28531 17536 28532 989 28533 5285 28534 161 28535 164 28536 17541 28537 17542 28538 956 28539 17544 28540 17545 28541 17546 28542 17547 28543 1882 28544 164 28545 17550 28546 17551 28547 17552 28548 17553 28549 287 28550 17555 28551 2 28552 583 28553 17558 28554 5356 28555 17560 28556 17561 28557 17562 28558 17563 28559 17564 28560 17565 28561 2918 28562 17567 28563 69 28564 17569 28565 17570 28566 17571 28567 17572 28568 17573 28569 17574 28570 17575 28571 2 28572 17577 28573 17578 28574 17579 28575 17580 28576 17581 28577 17582 28578 2960 28579 2962 28580 17585 28581 17586 28582 17587 28583 17588 28584 17589 28585 364 28586 364 28587 40 28588 367 28589 441 28590 17595 28591 17596 28592 17597 28593 17598 28594 17599 28595 40 28596 17601 28597 17602 28598 2983 28599 2986 28600 2988 28601 2990 28602 77 28603 2992 28604 2993 28605 17610 28606 287 28607 17612 28608 2 28609 287 28610 77 28611 3959 28612 2852 28613 17618 28614 1882 28615 164 28616 2072 28617 2 28618 17623 28619 2 28620 161 28621 164 28622 17627 28623 17628 28624 17629 28625 17630 28626 17631 28627 956 28628 17633 28629 17634 28630 17635 28631 989 28632 161 28633 164 28634 17639 28635 17640 28636 2 28637 287 28638 17643 28639 2 28640 287 28641 17646 28642 17647 28643 17648 28644 956 28645 17650 28646 17651 28647 17652 28648 989 28649 5285 28650 161 28651 164 28652 17657 28653 17658 28654 956 28655 17660 28656 17661 28657 17662 28658 17663 28659 1882 28660 164 28661 17666 28662 17667 28663 17668 28664 17669 28665 287 28666 17671 28667 2 28668 583 28669 17674 28670 5356 28671 17676 28672 17677 28673 28673 28674 17773 28675 142 28676 11 28677 144 28678 164 28679 167 28680 169 28681 77 28682 128 28683 172 28684 28684 28685 28685 28686 28686 28687 16 28688 2986 28689 2988 28690 2990 28691 77 28692 2992 28693 2993 28694 28694 28695 28695 28696 1080 28697 7507 28698 28698 28699 17805 28700 17806 28701 7507 28702 28702 28703 2 28704 287 28705 17884 28706 17885 28707 28707 28708 28708 28709 17888 28710 7507 28711 28711 28712 105 28713 28713 28714 22 28715 161 28716 164 28717 7507 28718 28718 28719 128 28720 172 28721 28721 28722 28722 28723 11 28724 77 28725 28725 28726 128 28727 172 28728 28728 28729 13 28730 18061 28731 164 28732 77 28733 128 28734 172 28735 28735 28736 2990 28737 77 28738 2992 28739 2993 28740 28740 28741 2 28742 287 28743 18074 28744 18075 28745 28745 28746 18077 28747 18078 28748 18079 28749 28749 28750 18079 28751 28751 28752 18217 28753 18218 28754 7507 28755 28755 28756 18218 28757 7507 28758 28758 28759 28759 28760 3066 28761 28761 28762 7507 28763 28763 28764 23404 28765 2990 28766 77 28767 2992 28768 2993 28769 28769 28770 2 28771 287 28772 17884 28773 28773 28774 164 28775 7507 28776 28776 28777 7507 28778 28778 28779 3066 28780 28780 28781 2 28782 19517 28783 20896 28784 20907 28785 17884 28786 20896 28787 164 28788 17884 28789 20958 28790 28790 28791 28791 28792 20961 28793 7507 28794 28794 28795 28795 28796 21125 28797 7507 28798 21134 28799 7507 28800 22736 28801 164 28802 17884 28803 22746 28804 28804 28805 28805 28806 22749 28807 7507 28808 28808 28809 23453 28810 2988 28811 2990 28812 77 28813 2992 28814 2993 28815 28815 28816 3066 28817 28817 28818 22957 28819 11 28820 2 28821 287 28822 7507 28823 28823 28824 16 28825 23108 28826 17884 28827 23112 28828 28828 28829 28829 28830 23115 28831 7507 28832 28832 28833 7507 28834 28834 28835 2986 28836 2988 28837 2990 28838 77 28839 2992 28840 2993 28841 28841 28842 7507 28843 28843 28844 6002 28845 441 28846 28846 28847 2278 28848 28848 28849 28849 28850 28850 28851 28851 28852 28852 28853 28853 28854 28854 28855 28855 28856 28856 28857 28857 28858 28858 28859 28859 28860 28860 28861 2293 28862 2294 28863 2278 28864 28864 28865 28865 28866 28866 28867 28867 28868 28868 28869 28869 28870 28870 28871 28871 28872 28872 28873 28873 28874 28874 28875 2314 28876 28876 28877 2294 28878 28878 28879 2278 28880 28880 28881 28881 28882 28882 28883 28883 28884 28884 28885 28885 28886 28886 28887 28887 28888 28888 28889 28889 28890 2278 28891 28891 28892 28892 28893 28893 28894 28894 28895 28895 28896 28896 28897 28897 28898 28898 28899 28899 28900 28900 28901 28901 28902 28902 28903 28903 28904 28904 28905 28905 28906 3029 28907 28907 28908 161 28909 28909 28910 164 28911 28911 28912 28912 28913 28913 28914 3019 28915 28915 28916 28916 28917 28917 28918 28918 28919 28919 28920 28920 28921 28921 28922 28922 28923 28923 28924 28924 28925 28925 28926 28926 28927 2293 28928 3034 28929 3019 28930 28930 28931 28931 28932 28932 28933 28933 28934 28934 28935 28935 28936 28936 28937 28937 28938 28938 28939 2314 28940 28940 28941 3034 28942 28942 28943 3019 28944 28944 28945 28945 28946 28946 28947 28947 28948 28948 28949 28949 28950 28950 28951 28951 28952 28952 28953 28953 28954 28954 28955 3029 28956 28956 28957 28957 28958 2 28959 28959 28960 5853 28961 5854 28962 28962 28963 2 28964 3 28965 4 28966 2 28967 2 28968 2 28969 2 28970 11 28971 2 28972 2 28973 28973 28974 2 28975 28975 28976 22 28977 28977 28978 28978 28979 2 28980 11 28981 11 28982 28982 28983 28983 28984 28984 28985 22 28986 2 28987 2 28988 2 28989 11 28990 28990 28991 11 28992 28992 28993 11 28994 28994 28995 28995 28996 28996 28997 28997 28998 60 28999 28999 29000 2 29001 2 29002 29002 29003 11 29004 2 29005 2 29006 69 29007 133 29008 29008 29009 2 29010 2 29011 29011 29012 2 29013 29013 29014 29014 29015 2 29016 86 29017 87 29018 90 29019 136 29020 11 29021 29021 29022 29022 29023 101 29024 29024 29025 29025 29026 29026 29027 106 29028 106 29029 29029 29030 29030 29031 29031 29032 29032 29033 29033 29034 29034 29035 29035 29036 29036 29037 29037 29038 29038 29039 29039 29040 101 29041 2 29042 29042 29043 2 29044 2 29045 2 29046 2 29047 2 29048 2 29049 2 29050 2 29051 312 29052 2 29053 2 29054 2 29055 2 29056 2 29057 318 29058 2 29059 2 29060 2 29061 2 29062 2 29063 2 29064 2 29065 2 29066 2 29067 2 29068 2 29069 2 29070 2 29071 2 29072 2 29073 2 29074 29074 29075 29075 29076 29076 29077 29077 29078 29078 29079 29079 29080 29080 29081 29081 29082 29082 29083 29083 29084 29084 29085 29085 29086 29086 29087 29087 29088 29088 29089 29089 29090 4 29091 2 29092 29092 29093 29093 29094 29094 29095 29095 29096 29096 29097 29097 29098 29098 29099 29099 29100 29100 29101 29101 29102 29102 29103 364 29104 364 29105 40 29106 367 29107 441 29108 29108 29109 29109 29110 29110 29111 29111 29112 29112 29113 2 29114 583 29115 29115 29116 29116 29117 40 29118 29118 29119 287 29120 29120 29121 29121 29122 2 29123 583 29124 29124 29125 29125 29126 29126 29127 29127 29128 29128 29129 29129 29130 2 29131 751 29132 2 29133 29133 29134 29134 29135 29135 29136 29136 29137 29137 29138 29138 29139 29139 29140 29140 29141 29141 29142 29142 29143 868 29144 29144 29145 29145 29146 29146 29147 29147 29148 29148 29149 29149 29150 29150 29151 29151 29152 29152 29153 29153 29154 29154 29155 29155 29156 29156 29157 29157 29158 29158 29159 29159 29160 29160 29161 29161 29162 29162 29163 29163 29164 29164 29165 29165 29166 29166 29167 29167 29168 29168 29169 29169 29170 29170 29171 29171 29172 29172 29173 29173 29174 29174 29175 29175 29176 29176 29177 29177 29178 29178 29179 29179 29180 29180 29181 29181 29182 29182 29183 29183 29184 29184 29185 29185 29186 29186 29187 29187 29188 29188 29189 29189 29190 29190 29191 29191 29192 29192 29193 29193 29194 29194 29195 29195 29196 29196 29197 29197 29198 29198 29199 29199 29200 29200 29201 29201 29202 29202 29203 29203 29204 29204 29205 29205 29206 29206 29207 29207 29208 29208 29209 29209 29210 29210 29211 29211 29212 29212 29213 29213 29214 29214 29215 29215 29216 29216 29217 29217 29218 29218 29219 29219 29220 29220 29221 29221 29222 29222 29223 29223 29224 11 29225 29225 29226 29226 29227 29227 29228 2 29229 583 29230 29230 29231 956 29232 29232 29233 29233 29234 989 29235 161 29236 164 29237 29237 29238 29238 29239 1505 29240 2 29241 2 29242 2 29243 29243 29244 2 29245 29245 29246 2 29247 2 29248 2 29249 2 29250 2 29251 2 29252 2 29253 2 29254 2 29255 29255 29256 29256 29257 29257 29258 29258 29259 29259 29260 29260 29261 29261 29262 29262 29263 29263 29264 29264 29265 29265 29266 29266 29267 29267 29268 22 29269 29269 29270 4 29271 11 29272 29272 29273 90 29274 11 29275 29275 29276 101 29277 29277 29278 1569 29279 29279 29280 29280 29281 29281 29282 101 29283 868 29284 11 29285 29285 29286 956 29287 29287 29288 29288 29289 989 29290 161 29291 164 29292 29292 29293 29293 29294 29294 29295 868 29296 11 29297 29297 29298 29298 29299 29299 29300 364 29301 364 29302 367 29303 441 29304 29304 29305 29305 29306 29306 29307 29307 29308 2 29309 583 29310 29310 29311 29311 29312 29312 29313 29313 29314 29314 29315 29315 29316 2 29317 583 29318 29318 29319 29319 29320 142 29321 11 29322 144 29323 164 29324 165 29325 167 29326 169 29327 77 29328 128 29329 172 29330 29330 29331 29331 29332 142 29333 11 29334 144 29335 164 29336 2270 29337 2272 29338 2273 29339 2274 29340 2 29341 29341 29342 29342 29343 29343 29344 868 29345 29345 29346 11 29347 29347 29348 2 29349 583 29350 29350 29351 29351 29352 29352 29353 2756 29354 29354 29355 29355 29356 364 29357 364 29358 367 29359 441 29360 29360 29361 29361 29362 29362 29363 29363 29364 2 29365 583 29366 29366 29367 29367 29368 29368 29369 29369 29370 2827 29371 29371 29372 441 29373 29373 29374 2 29375 29375 29376 29376 29377 2 29378 3 29379 4 29380 2 29381 2 29382 2 29383 2 29384 11 29385 2 29386 2 29387 29387 29388 2 29389 29389 29390 22 29391 29391 29392 29392 29393 2 29394 11 29395 11 29396 29396 29397 29397 29398 29398 29399 22 29400 2 29401 2 29402 2 29403 11 29404 29404 29405 11 29406 29406 29407 11 29408 29408 29409 29409 29410 29410 29411 29411 29412 60 29413 29413 29414 2 29415 2 29416 29416 29417 11 29418 2 29419 2 29420 69 29421 133 29422 29422 29423 2 29424 2 29425 29425 29426 2 29427 29427 29428 29428 29429 2 29430 86 29431 87 29432 90 29433 136 29434 11 29435 29435 29436 29436 29437 101 29438 29438 29439 29439 29440 29440 29441 106 29442 106 29443 29443 29444 29444 29445 29445 29446 29446 29447 29447 29448 29448 29449 29449 29450 29450 29451 29451 29452 29452 29453 29453 29454 101 29455 2 29456 29456 29457 2 29458 2 29459 2 29460 2 29461 2 29462 2 29463 2 29464 2 29465 312 29466 2 29467 2 29468 2 29469 2 29470 2 29471 318 29472 2 29473 2 29474 2 29475 2 29476 2 29477 2 29478 2 29479 2 29480 2 29481 2 29482 2 29483 2 29484 2 29485 2 29486 2 29487 2 29488 29488 29489 29489 29490 29490 29491 29491 29492 29492 29493 29493 29494 29494 29495 29495 29496 29496 29497 29497 29498 29498 29499 29499 29500 29500 29501 29501 29502 29502 29503 29503 29504 4 29505 2 29506 29506 29507 29507 29508 29508 29509 29509 29510 29510 29511 29511 29512 29512 29513 29513 29514 29514 29515 29515 29516 29516 29517 364 29518 364 29519 40 29520 367 29521 441 29522 29522 29523 29523 29524 29524 29525 29525 29526 29526 29527 2 29528 583 29529 29529 29530 29530 29531 40 29532 29532 29533 287 29534 29534 29535 29535 29536 2 29537 583 29538 29538 29539 29539 29540 29540 29541 29541 29542 29542 29543 29543 29544 2 29545 751 29546 2 29547 29547 29548 29548 29549 29549 29550 29550 29551 29551 29552 29552 29553 29553 29554 29554 29555 29555 29556 29556 29557 868 29558 29558 29559 29559 29560 29560 29561 29561 29562 29562 29563 29563 29564 29564 29565 29565 29566 29566 29567 29567 29568 29568 29569 29569 29570 29570 29571 29571 29572 29572 29573 29573 29574 29574 29575 29575 29576 29576 29577 29577 29578 29578 29579 29579 29580 29580 29581 29581 29582 29582 29583 29583 29584 29584 29585 29585 29586 29586 29587 29587 29588 29588 29589 29589 29590 29590 29591 29591 29592 29592 29593 29593 29594 29594 29595 29595 29596 29596 29597 29597 29598 29598 29599 29599 29600 29600 29601 29601 29602 29602 29603 29603 29604 29604 29605 29605 29606 29606 29607 29607 29608 29608 29609 29609 29610 29610 29611 29611 29612 29612 29613 29613 29614 29614 29615 29615 29616 29616 29617 29617 29618 29618 29619 29619 29620 29620 29621 29621 29622 29622 29623 29623 29624 29624 29625 29625 29626 29626 29627 29627 29628 29628 29629 29629 29630 29630 29631 29631 29632 29632 29633 29633 29634 29634 29635 29635 29636 29636 29637 29637 29638 11 29639 29639 29640 29640 29641 29641 29642 2 29643 583 29644 29644 29645 956 29646 29646 29647 29647 29648 989 29649 161 29650 164 29651 29651 29652 29652 29653 1505 29654 2 29655 2 29656 2 29657 29657 29658 2 29659 29659 29660 2 29661 2 29662 2 29663 2 29664 2 29665 2 29666 2 29667 2 29668 2 29669 29669 29670 29670 29671 29671 29672 29672 29673 29673 29674 29674 29675 29675 29676 29676 29677 29677 29678 29678 29679 29679 29680 29680 29681 29681 29682 22 29683 29683 29684 4 29685 11 29686 29686 29687 90 29688 11 29689 29689 29690 101 29691 29691 29692 1569 29693 29693 29694 29694 29695 29695 29696 101 29697 868 29698 11 29699 29699 29700 956 29701 29701 29702 29702 29703 989 29704 161 29705 164 29706 29706 29707 29707 29708 29708 29709 868 29710 11 29711 29711 29712 29712 29713 29713 29714 364 29715 364 29716 367 29717 441 29718 29718 29719 29719 29720 29720 29721 29721 29722 2 29723 583 29724 29724 29725 29725 29726 29726 29727 29727 29728 29728 29729 29729 29730 2 29731 583 29732 29732 29733 29733 29734 142 29735 11 29736 144 29737 164 29738 165 29739 167 29740 169 29741 77 29742 128 29743 172 29744 29744 29745 29745 29746 142 29747 11 29748 144 29749 164 29750 2270 29751 2272 29752 2273 29753 2274 29754 2 29755 29755 29756 29756 29757 2 29758 29758 29759 868 29760 29760 29761 11 29762 29762 29763 2 29764 583 29765 29765 29766 29766 29767 29767 29768 2756 29769 29769 29770 29770 29771 364 29772 364 29773 367 29774 441 29775 29775 29776 29776 29777 29777 29778 29778 29779 2 29780 583 29781 29781 29782 29782 29783 29783 29784 29784 29785 2827 29786 29786 29787 441 29788 29788 29789 2 29790 29790 29791 29791 29792 29440 29793 2 29794 29794 29795 29795 29796 29796 29797 29797 29798 29798 29799 2918 29800 29800 29801 69 29802 29802 29803 29803 29804 29804 29805 29805 29806 29806 29807 29807 29808 29808 29809 2 29810 29810 29811 29811 29812 29812 29813 29813 29814 29814 29815 29815 29816 2960 29817 2962 29818 29818 29819 29819 29820 29820 29821 29821 29822 29822 29823 364 29824 364 29825 40 29826 367 29827 441 29828 29828 29829 29829 29830 29830 29831 29831 29832 29832 29833 40 29834 29834 29835 29835 29836 2983 29837 2986 29838 2988 29839 2990 29840 77 29841 2992 29842 2993 29843 29843 29844 287 29845 29845 29846 2 29847 287 29848 77 29849 3959 29850 2852 29851 29851 29852 1882 29853 164 29854 2072 29855 2 29856 29856 29857 2 29858 161 29859 164 29860 29860 29861 29861 29862 29862 29863 29863 29864 29864 29865 956 29866 29866 29867 29867 29868 29868 29869 989 29870 161 29871 164 29872 29872 29873 29873 29874 2 29875 287 29876 29876 29877 2 29878 287 29879 29879 29880 29880 29881 29881 29882 956 29883 29883 29884 29884 29885 29885 29886 989 29887 5285 29888 161 29889 164 29890 29890 29891 29891 29892 956 29893 29893 29894 29894 29895 29895 29896 29896 29897 1882 29898 164 29899 29899 29900 29900 29901 29901 29902 29902 29903 287 29904 29904 29905 2 29906 583 29907 29907 29908 5356 29909 29909 29910 29910 29911 29911 29912 29912 29913 29913 29914 29914 29915 2918 29916 29916 29917 69 29918 29918 29919 29919 29920 29920 29921 29921 29922 29922 29923 29923 29924 29924 29925 2 29926 29926 29927 29927 29928 29928 29929 29929 29930 29930 29931 29931 29932 2960 29933 2962 29934 29934 29935 29935 29936 29936 29937 29937 29938 29938 29939 364 29940 364 29941 40 29942 367 29943 441 29944 29944 29945 29945 29946 29946 29947 29947 29948 29948 29949 40 29950 29950 29951 29951 29952 2983 29953 2986 29954 2988 29955 2990 29956 77 29957 2992 29958 2993 29959 29959 29960 287 29961 29961 29962 2 29963 287 29964 77 29965 3959 29966 2852 29967 29967 29968 1882 29969 164 29970 2072 29971 2 29972 29972 29973 2 29974 161 29975 164 29976 29976 29977 29977 29978 29978 29979 29979 29980 29980 29981 956 29982 29982 29983 29983 29984 29984 29985 989 29986 161 29987 164 29988 29988 29989 29989 29990 2 29991 287 29992 29992 29993 2 29994 287 29995 29995 29996 29996 29997 29997 29998 956 29999 29999 30000 30000 30001 30001 30002 989 30003 5285 30004 161 30005 164 30006 30006 30007 30007 30008 956 30009 30009 30010 30010 30011 30011 30012 30012 30013 1882 30014 164 30015 30015 30016 30016 30017 30017 30018 30018 30019 287 30020 30020 30021 2 30022 583 30023 30023 30024 5356 30025 30025 30026 30026 30027 3019 30028 30028 30029 30029 30030 30030 30031 30031 30032 30032 30033 30033 30034 30034 30035 30035 30036 30036 30037 30037 30038 30038 30039 30039 30040 30040 30041 30041 30042 30042 30043 6002 30044 441 30045 30045 30046 956 30047 30047 30048 30048 30049 989 30050 161 30051 164 30052 30052 30053 30053 30054 6002 30055 441 30056 30056 30057 172 30058 30058 30059 868 30060 30060 30061 30061 30062 30062 30063 30063 30064 30064 30065 30065 30066 30066 30067 30067 30068 30068 30069 30069 30070 30070 30071 30071 30072 30072 30073 30073 30074 30074 30075 30075 30076 30076 30077 30077 30078 30078 30079 30079 30080 30080 30081 30081 30082 30082 30083 30083 30084 30084 30085 30085 30086 30086 30087 30087 30088 30088 30089 30089 30090 30090 30091 30091 30092 30092 30093 30093 30094 30094 30095 30095 30096 30096 30097 30097 30098 30098 30099 30099 30100 30100 30101 30101 30102 30102 30103 30103 30104 30104 30105 30105 30106 30106 30107 30107 30108 30108 30109 30109 30110 30110 30111 30111 30112 30112 30113 30113 30114 30114 30115 30115 30116 30116 30117 30117 30118 30118 30119 30119 30120 30120 30121 30121 30122 30122 30123 30123 30124 30124 30125 30125 30126 30126 30127 30127 30128 30128 30129 30129 30130 30130 30131 30131 30132 30132 30133 30133 30134 30134 30135 30135 30136 30136 30137 30137 30138 30138 30139 30139 30140 11 30141 30141 30142 30142 30143 30143 30144 2 30145 583 30146 30146 30147 30147 30148 1505 30149 2 30150 2 30151 2 30152 30152 30153 2 30154 30154 30155 2 30156 2 30157 2 30158 2 30159 2 30160 2 30161 2 30162 2 30163 2 30164 30164 30165 30165 30166 30166 30167 30167 30168 30168 30169 30169 30170 30170 30171 30171 30172 30172 30173 30173 30174 30174 30175 30175 30176 30176 30177 22 30178 30178 30179 4 30180 11 30181 30181 30182 90 30183 11 30184 30184 30185 101 30186 30186 30187 1569 30188 30188 30189 30189 30190 30190 30191 101 30192 868 30193 11 30194 30194 30195 956 30196 30196 30197 30197 30198 989 30199 161 30200 164 30201 30201 30202 30202 30203 30203 30204 868 30205 11 30206 30206 30207 30207 30208 30208 30209 364 30210 364 30211 367 30212 441 30213 30213 30214 30214 30215 30215 30216 30216 30217 2 30218 583 30219 30219 30220 30220 30221 30221 30222 30222 30223 30223 30224 30224 30225 2 30226 583 30227 30227 30228 30228 30229 26 30230 11 30231 2 30232 287 30233 26 30234 30234 30235 15 30236 15 30237 11 30238 26 30239 2 30240 11 30241 11 30242 40 30243 16 30244 30244 30245 15 30246 70 30247 71 30248 11 30249 77 30250 11 30251 2 30252 11 30253 30253 30254 89 30255 105 30256 30256 30257 112 30258 30258 30259 128 30260 172 30261 30261 30262 26 30263 30263 30264 30264 30265 1080 30266 7507 30267 30267 30268 17805 30269 17806 30270 7507 30271 30271 30272 2 30273 287 30274 17884 30275 17885 30276 30276 30277 30277 30278 17888 30279 7507 30280 30280 30281 105 30282 30282 30283 22 30284 161 30285 164 30286 7507 30287 30287 30288 128 30289 172 30290 30290 30291 30291 30292 11 30293 77 30294 30294 30295 128 30296 172 30297 30297 30298 13 30299 18061 30300 164 30301 77 30302 128 30303 172 30304 30304 30305 2 30306 287 30307 18074 30308 18075 30309 30309 30310 18077 30311 18078 30312 18079 30313 30313 30314 18079 30315 30315 30316 18217 30317 18218 30318 7507 30319 30319 30320 18218 30321 7507 30322 30322 30323 30323 30324 3066 30325 30325 30326 7507 30327 30327 30328 7507 30329 30329 30330 3066 30331 30331 30332 2 30333 19517 30334 20896 30335 20907 30336 17884 30337 20896 30338 164 30339 17884 30340 20958 30341 30341 30342 30342 30343 20961 30344 7507 30345 30345 30346 30346 30347 21125 30348 7507 30349 21134 30350 7507 30351 22736 30352 164 30353 17884 30354 22746 30355 30355 30356 30356 30357 22749 30358 7507 30359 30359 30360 3066 30361 30361 30362 22957 30363 11 30364 2 30365 287 30366 7507 30367 30367 30368 16 30369 23108 30370 17884 30371 23112 30372 30372 30373 30373 30374 23115 30375 7507 30376 30376 30377 7507 30378 30378 30379 7507 30380 30380 30381 30381 30382 17773 30383 142 30384 11 30385 144 30386 164 30387 167 30388 169 30389 77 30390 128 30391 172 30392 30392 30393 30393 30394 142 30395 11 30396 144 30397 164 30398 2270 30399 2272 30400 2273 30401 2274 30402 2 30403 2 30404 3 30405 4 30406 2 30407 2 30408 2 30409 2 30410 11 30411 2 30412 2 30413 17 30414 2 30415 20 30416 22 30417 24 30418 25 30419 2 30420 11 30421 11 30422 32 30423 33 30424 39 30425 22 30426 2 30427 2 30428 2 30429 11 30430 51 30431 11 30432 53 30433 11 30434 55 30435 57 30436 58 30437 59 30438 60 30439 61 30440 2 30441 2 30442 64 30443 11 30444 2 30445 2 30446 69 30447 133 30448 134 30449 2 30450 2 30451 76 30452 2 30453 81 30454 84 30455 2 30456 86 30457 87 30458 90 30459 136 30460 11 30461 96 30462 97 30463 101 30464 102 30465 103 30466 30466 30467 106 30468 106 30469 109 30470 113 30471 115 30472 116 30473 117 30474 118 30475 119 30476 120 30477 121 30478 123 30479 126 30480 101 30481 2 30482 137 30483 2 30484 2 30485 2 30486 2 30487 2 30488 2 30489 2 30490 2 30491 312 30492 2 30493 2 30494 2 30495 2 30496 2 30497 318 30498 2 30499 2 30500 2 30501 2 30502 2 30503 2 30504 2 30505 2 30506 2 30507 2 30508 2 30509 2 30510 2 30511 2 30512 2 30513 2 30514 5922 30515 5923 30516 5924 30517 5925 30518 5926 30519 5927 30520 5928 30521 5929 30522 5930 30523 5931 30524 5932 30525 5933 30526 5934 30527 5935 30528 5936 30529 5937 30530 4 30531 2 30532 5940 30533 5941 30534 5942 30535 5943 30536 5944 30537 5945 30538 5946 30539 5947 30540 5948 30541 5949 30542 5950 30543 364 30544 364 30545 40 30546 367 30547 441 30548 5956 30549 5957 30550 5958 30551 5959 30552 5960 30553 2 30554 583 30555 5963 30556 5964 30557 40 30558 5966 30559 287 30560 5968 30561 5969 30562 2 30563 583 30564 5972 30565 5973 30566 5974 30567 5975 30568 5976 30569 5977 30570 2 30571 751 30572 2 30573 5981 30574 5982 30575 5983 30576 5984 30577 5985 30578 5986 30579 5987 30580 5988 30581 5989 30582 5990 30583 868 30584 30060 30585 30061 30586 30062 30587 30063 30588 30064 30589 30065 30590 30066 30591 30067 30592 30068 30593 30069 30594 30070 30595 30071 30596 30072 30597 30073 30598 30074 30599 30075 30600 30076 30601 30077 30602 30078 30603 30079 30604 30080 30605 30081 30606 30082 30607 30083 30608 30084 30609 30085 30610 30086 30611 30087 30612 30088 30613 30089 30614 30090 30615 30091 30616 30092 30617 30093 30618 30094 30619 30095 30620 30096 30621 30097 30622 30098 30623 30099 30624 30100 30625 30101 30626 30102 30627 30103 30628 30104 30629 30105 30630 30106 30631 30107 30632 30108 30633 30109 30634 30110 30635 30111 30636 30112 30637 30113 30638 30114 30639 30115 30640 30116 30641 30117 30642 30118 30643 30119 30644 30120 30645 30121 30646 30122 30647 30123 30648 30124 30649 30125 30650 30126 30651 30127 30652 30128 30653 30129 30654 30130 30655 30131 30656 30132 30657 30133 30658 30134 30659 30135 30660 30136 30661 30137 30662 30138 30663 30139 30664 11 30665 30141 30666 30142 30667 30143 30668 2 30669 583 30670 30146 30671 956 30672 30147 30673 30048 30674 989 30675 161 30676 164 30677 30052 30678 30053 30679 1505 30680 2 30681 2 30682 2 30683 30152 30684 2 30685 30154 30686 2 30687 2 30688 2 30689 2 30690 2 30691 2 30692 2 30693 2 30694 2 30695 30164 30696 30165 30697 30166 30698 30167 30699 30168 30700 30169 30701 30170 30702 30171 30703 30172 30704 30173 30705 30174 30706 30175 30707 30176 30708 22 30709 30178 30710 4 30711 11 30712 30181 30713 90 30714 11 30715 30184 30716 101 30717 30186 30718 1569 30719 30188 30720 30189 30721 30190 30722 101 30723 868 30724 11 30725 30194 30726 956 30727 30196 30728 30197 30729 989 30730 161 30731 164 30732 30201 30733 30202 30734 30203 30735 868 30736 11 30737 30206 30738 30207 30739 30208 30740 364 30741 364 30742 367 30743 441 30744 30213 30745 30214 30746 30215 30747 30216 30748 2 30749 583 30750 30219 30751 30220 30752 30221 30753 30222 30754 30223 30755 30224 30756 2 30757 583 30758 30227 30759 30228 30760 142 30761 11 30762 144 30763 164 30764 165 30765 167 30766 169 30767 77 30768 128 30769 172 30770 217 30771 218 30772 142 30773 11 30774 144 30775 164 30776 2270 30777 2272 30778 2273 30779 2274 30780 2 30781 30466 30782 30782 30783 30782 30784 868 30785 30785 30786 11 30787 30787 30788 2 30789 583 30790 30790 30791 30791 30792 30792 30793 2756 30794 17800 30795 17801 30796 364 30797 364 30798 367 30799 441 30800 30800 30801 30801 30802 30802 30803 30803 30804 2 30805 583 30806 30806 30807 30807 30808 30808 30809 30809 30810 2827 30811 30811 30812 441 30813 30813 30814 2 30815 30815 30816 30816 30817 2 30818 3 30819 4 30820 2 30821 2 30822 2 30823 2 30824 11 30825 2 30826 2 30827 30827 30828 2 30829 30829 30830 22 30831 30831 30832 30832 30833 2 30834 11 30835 11 30836 30836 30837 30837 30838 30838 30839 22 30840 2 30841 2 30842 2 30843 11 30844 30844 30845 11 30846 30846 30847 11 30848 30848 30849 30849 30850 30850 30851 30851 30852 60 30853 30853 30854 2 30855 2 30856 30856 30857 11 30858 2 30859 2 30860 69 30861 133 30862 30862 30863 2 30864 2 30865 30865 30866 2 30867 30867 30868 30868 30869 2 30870 86 30871 87 30872 90 30873 136 30874 11 30875 30875 30876 30876 30877 101 30878 30878 30879 30879 30880 30880 30881 106 30882 106 30883 30883 30884 30884 30885 30885 30886 30886 30887 30887 30888 30888 30889 30889 30890 30890 30891 30891 30892 30892 30893 30893 30894 101 30895 2 30896 30896 30897 2 30898 2 30899 2 30900 2 30901 2 30902 2 30903 2 30904 2 30905 312 30906 2 30907 2 30908 2 30909 2 30910 2 30911 318 30912 2 30913 2 30914 2 30915 2 30916 2 30917 2 30918 2 30919 2 30920 2 30921 2 30922 2 30923 2 30924 2 30925 2 30926 2 30927 2 30928 30928 30929 30929 30930 30930 30931 30931 30932 30932 30933 30933 30934 30934 30935 30935 30936 30936 30937 30937 30938 30938 30939 30939 30940 30940 30941 30941 30942 30942 30943 30943 30944 4 30945 2 30946 30946 30947 30947 30948 30948 30949 30949 30950 30950 30951 30951 30952 30952 30953 30953 30954 30954 30955 30955 30956 30956 30957 364 30958 364 30959 40 30960 367 30961 441 30962 30962 30963 30963 30964 30964 30965 30965 30966 30966 30967 2 30968 583 30969 30969 30970 30970 30971 40 30972 30972 30973 287 30974 30974 30975 30975 30976 2 30977 583 30978 30978 30979 30979 30980 30980 30981 30981 30982 30982 30983 30983 30984 2 30985 751 30986 2 30987 30987 30988 30988 30989 30989 30990 30990 30991 30991 30992 30992 30993 30993 30994 30994 30995 30995 30996 30996 30997 868 30998 30998 30999 30999 31000 31000 31001 31001 31002 31002 31003 31003 31004 31004 31005 31005 31006 31006 31007 31007 31008 31008 31009 31009 31010 31010 31011 31011 31012 31012 31013 31013 31014 31014 31015 31015 31016 31016 31017 31017 31018 31018 31019 31019 31020 31020 31021 31021 31022 31022 31023 31023 31024 31024 31025 31025 31026 31026 31027 31027 31028 31028 31029 31029 31030 31030 31031 31031 31032 31032 31033 31033 31034 31034 31035 31035 31036 31036 31037 31037 31038 31038 31039 31039 31040 31040 31041 31041 31042 31042 31043 31043 31044 31044 31045 31045 31046 31046 31047 31047 31048 31048 31049 31049 31050 31050 31051 31051 31052 31052 31053 31053 31054 31054 31055 31055 31056 31056 31057 31057 31058 31058 31059 31059 31060 31060 31061 31061 31062 31062 31063 31063 31064 31064 31065 31065 31066 31066 31067 31067 31068 31068 31069 31069 31070 31070 31071 31071 31072 31072 31073 31073 31074 31074 31075 31075 31076 31076 31077 31077 31078 11 31079 31079 31080 31080 31081 31081 31082 2 31083 583 31084 31084 31085 956 31086 31086 31087 31087 31088 989 31089 161 31090 164 31091 31091 31092 31092 31093 1505 31094 2 31095 2 31096 2 31097 31097 31098 2 31099 31099 31100 2 31101 2 31102 2 31103 2 31104 2 31105 2 31106 2 31107 2 31108 2 31109 31109 31110 31110 31111 31111 31112 31112 31113 31113 31114 31114 31115 31115 31116 31116 31117 31117 31118 31118 31119 31119 31120 31120 31121 31121 31122 22 31123 31123 31124 4 31125 11 31126 31126 31127 90 31128 11 31129 31129 31130 101 31131 31131 31132 1569 31133 31133 31134 31134 31135 31135 31136 101 31137 868 31138 11 31139 31139 31140 956 31141 31141 31142 31142 31143 989 31144 161 31145 164 31146 31146 31147 31147 31148 31148 31149 868 31150 11 31151 31151 31152 31152 31153 31153 31154 364 31155 364 31156 367 31157 441 31158 31158 31159 31159 31160 31160 31161 31161 31162 2 31163 583 31164 31164 31165 31165 31166 31166 31167 31167 31168 31168 31169 31169 31170 2 31171 583 31172 31172 31173 31173 31174 142 31175 11 31176 144 31177 164 31178 165 31179 167 31180 169 31181 77 31182 128 31183 172 31184 31184 31185 31185 31186 142 31187 11 31188 144 31189 164 31190 2270 31191 2272 31192 2273 31193 2274 31194 2 31195 31195 31196 31196 31197 2 31198 31198 31199 868 31200 31200 31201 11 31202 31202 31203 2 31204 583 31205 31205 31206 31206 31207 31207 31208 2756 31209 31209 31210 31210 31211 364 31212 364 31213 367 31214 441 31215 31215 31216 31216 31217 31217 31218 31218 31219 2 31220 583 31221 31221 31222 31222 31223 31223 31224 31224 31225 2827 31226 31226 31227 441 31228 31228 31229 2 31230 31230 31231 31231 31232 30880 31233 2 31234 31234 31235 31235 31236 31236 31237 31237 31238 31238 31239 2918 31240 31240 31241 69 31242 31242 31243 31243 31244 31244 31245 31245 31246 31246 31247 31247 31248 31248 31249 2 31250 31250 31251 31251 31252 31252 31253 31253 31254 31254 31255 31255 31256 2960 31257 2962 31258 31258 31259 31259 31260 31260 31261 31261 31262 31262 31263 364 31264 364 31265 40 31266 367 31267 441 31268 31268 31269 31269 31270 31270 31271 31271 31272 31272 31273 40 31274 31274 31275 31275 31276 2983 31277 2986 31278 2988 31279 2990 31280 77 31281 2992 31282 2993 31283 31283 31284 287 31285 31285 31286 2 31287 287 31288 77 31289 3959 31290 2852 31291 31291 31292 1882 31293 164 31294 2072 31295 2 31296 31296 31297 2 31298 161 31299 164 31300 31300 31301 31301 31302 31302 31303 31303 31304 31304 31305 956 31306 31306 31307 31307 31308 31308 31309 989 31310 161 31311 164 31312 31312 31313 31313 31314 2 31315 287 31316 31316 31317 2 31318 287 31319 31319 31320 31320 31321 31321 31322 956 31323 31323 31324 31324 31325 31325 31326 989 31327 5285 31328 161 31329 164 31330 31330 31331 31331 31332 956 31333 31333 31334 31334 31335 31335 31336 31336 31337 1882 31338 164 31339 31339 31340 31340 31341 31341 31342 31342 31343 287 31344 31344 31345 2 31346 583 31347 31347 31348 5356 31349 31349 31350 31350 31351 31351 31352 31352 31353 31353 31354 18366 31355 2918 31356 31356 31357 69 31358 31358 31359 31359 31360 31360 31361 31361 31362 30047 31363 31363 31364 31364 31365 2 31366 31366 31367 31367 31368 31368 31369 31369 31370 31370 31371 31371 31372 2960 31373 2962 31374 31374 31375 31375 31376 31376 31377 31377 31378 31378 31379 364 31380 364 31381 40 31382 367 31383 441 31384 31384 31385 31385 31386 31386 31387 31387 31388 31388 31389 40 31390 31390 31391 31391 31392 2983 31393 2986 31394 2988 31395 2990 31396 77 31397 2992 31398 2993 31399 31399 31400 287 31401 31401 31402 2 31403 287 31404 77 31405 3959 31406 2852 31407 31407 31408 1882 31409 164 31410 2072 31411 2 31412 31412 31413 2 31414 161 31415 164 31416 31416 31417 31417 31418 31418 31419 31419 31420 31420 31421 956 31422 31422 31423 31423 31424 31424 31425 989 31426 161 31427 164 31428 31428 31429 31429 31430 2 31431 287 31432 31432 31433 2 31434 287 31435 31435 31436 31436 31437 31437 31438 956 31439 31439 31440 31440 31441 31441 31442 989 31443 5285 31444 161 31445 164 31446 31446 31447 31447 31448 956 31449 31449 31450 31450 31451 31451 31452 31452 31453 1882 31454 164 31455 31455 31456 31456 31457 31457 31458 31458 31459 287 31460 31460 31461 2 31462 583 31463 31463 31464 5356 31465 31465 31466 31466 31467 31467 31468 868 31469 30785 31470 11 31471 30787 31472 2 31473 583 31474 30790 31475 30791 31476 16 31477 2986 31478 2988 31479 2990 31480 77 31481 2992 31482 2993 31483 31483 31484 30792 31485 16 31486 16 31487 16 31488 16 31489 16 31490 31490 31491 31491 31492 31492 31493 31493 31494 100 31495 31495 31496 108 31497 31497 31498 31498 31499 31499 31500 17773 31501 142 31502 11 31503 144 31504 164 31505 167 31506 169 31507 77 31508 128 31509 172 31510 31510 31511 31511 31512 16 31513 2986 31514 2988 31515 2990 31516 77 31517 2992 31518 2993 31519 31519 31520 2990 31521 77 31522 2992 31523 2993 31524 31524 31525 23404 31526 2990 31527 77 31528 2992 31529 2993 31530 31530 31531 23453 31532 2988 31533 2990 31534 77 31535 2992 31536 2993 31537 31537 31538 2072 31539 31539 31540 3066 31541 16 31542 31542 31543 31543 31544 2 31545 3136 31546 31546 31547 2986 31548 2988 31549 2990 31550 77 31551 2992 31552 2993 31553 31553 31554 364 31555 364 31556 367 31557 441 31558 30800 31559 30801 31560 30802 31561 30803 31562 2 31563 583 31564 30806 31565 30807 31566 30808 31567 30809 31568 1080 31569 7507 31570 31570 31571 17805 31572 17806 31573 7507 31574 31574 31575 2827 31576 30811 31577 441 31578 30813 31579 2 31580 30815 31581 30816 31582 2 31583 3 31584 4 31585 2 31586 2 31587 2 31588 2 31589 11 31590 2 31591 2 31592 30827 31593 2 31594 30829 31595 22 31596 30831 31597 30832 31598 2 31599 11 31600 11 31601 30836 31602 30837 31603 30838 31604 22 31605 2 31606 2 31607 2 31608 11 31609 30844 31610 11 31611 30846 31612 11 31613 30848 31614 30849 31615 30850 31616 30851 31617 60 31618 30853 31619 2 31620 2 31621 30856 31622 11 31623 2 31624 2 31625 69 31626 133 31627 30862 31628 2 31629 2 31630 30865 31631 2 31632 30867 31633 30868 31634 2 31635 86 31636 87 31637 90 31638 136 31639 11 31640 30875 31641 30876 31642 101 31643 30878 31644 30879 31645 30880 31646 106 31647 106 31648 30883 31649 30884 31650 30885 31651 30886 31652 30887 31653 30888 31654 30889 31655 30890 31656 30891 31657 30892 31658 30893 31659 101 31660 2 31661 30896 31662 2 31663 2 31664 2 31665 2 31666 2 31667 2 31668 2 31669 2 31670 312 31671 2 31672 2 31673 2 31674 2 31675 2 31676 318 31677 2 31678 2 31679 2 31680 2 31681 2 31682 2 31683 2 31684 2 31685 2 31686 2 31687 2 31688 2 31689 2 31690 2 31691 2 31692 2 31693 30928 31694 30929 31695 30930 31696 30931 31697 30932 31698 30933 31699 30934 31700 30935 31701 30936 31702 30937 31703 30938 31704 30939 31705 30940 31706 30941 31707 30942 31708 30943 31709 4 31710 2 31711 30946 31712 30947 31713 30948 31714 30949 31715 30950 31716 30951 31717 30952 31718 30953 31719 30954 31720 30955 31721 30956 31722 364 31723 364 31724 40 31725 367 31726 441 31727 30962 31728 30963 31729 30964 31730 30965 31731 30966 31732 2 31733 583 31734 30969 31735 30970 31736 40 31737 30972 31738 287 31739 30974 31740 30975 31741 2 31742 583 31743 30978 31744 30979 31745 30980 31746 30981 31747 30982 31748 30983 31749 2 31750 751 31751 2 31752 30987 31753 30988 31754 30989 31755 30990 31756 30991 31757 30992 31758 30993 31759 30994 31760 30995 31761 30996 31762 868 31763 30998 31764 30999 31765 31000 31766 31001 31767 31002 31768 31003 31769 31004 31770 31005 31771 31006 31772 31007 31773 31008 31774 31009 31775 31010 31776 31011 31777 31012 31778 31013 31779 31014 31780 31015 31781 31016 31782 31017 31783 31018 31784 31019 31785 31020 31786 31021 31787 31022 31788 31023 31789 31024 31790 31025 31791 31026 31792 31027 31793 31028 31794 31029 31795 31030 31796 31031 31797 31032 31798 31033 31799 31034 31800 31035 31801 31036 31802 31037 31803 31038 31804 31039 31805 31040 31806 31041 31807 31042 31808 31043 31809 31044 31810 31045 31811 31046 31812 31047 31813 31048 31814 31049 31815 31050 31816 31051 31817 31052 31818 31053 31819 31054 31820 31055 31821 31056 31822 31057 31823 31058 31824 31059 31825 31060 31826 31061 31827 31062 31828 31063 31829 31064 31830 31065 31831 31066 31832 31067 31833 31068 31834 31069 31835 31070 31836 31071 31837 31072 31838 31073 31839 31074 31840 31075 31841 31076 31842 31077 31843 11 31844 31079 31845 31080 31846 31081 31847 2 31848 583 31849 31084 31850 956 31851 31086 31852 31087 31853 989 31854 161 31855 164 31856 31091 31857 31092 31858 1505 31859 2 31860 2 31861 2 31862 31097 31863 2 31864 31099 31865 2 31866 2 31867 2 31868 2 31869 2 31870 2 31871 2 31872 2 31873 2 31874 31109 31875 31110 31876 31111 31877 31112 31878 31113 31879 31114 31880 31115 31881 31116 31882 31117 31883 31118 31884 31119 31885 31120 31886 31121 31887 22 31888 31123 31889 4 31890 11 31891 31126 31892 90 31893 11 31894 31129 31895 101 31896 31131 31897 1569 31898 31133 31899 31134 31900 31135 31901 101 31902 868 31903 11 31904 31139 31905 956 31906 31141 31907 31142 31908 989 31909 161 31910 164 31911 31146 31912 31147 31913 31148 31914 868 31915 11 31916 31151 31917 31152 31918 31153 31919 364 31920 364 31921 367 31922 441 31923 31158 31924 31159 31925 31160 31926 31161 31927 2 31928 583 31929 31164 31930 31165 31931 31166 31932 31167 31933 31168 31934 31169 31935 2 31936 583 31937 31172 31938 31173 31939 142 31940 11 31941 144 31942 164 31943 165 31944 167 31945 169 31946 77 31947 128 31948 172 31949 31184 31950 31185 31951 142 31952 11 31953 144 31954 164 31955 2270 31956 2272 31957 2273 31958 2274 31959 2 31960 31195 31961 31196 31962 2 31963 31198 31964 868 31965 31200 31966 11 31967 31202 31968 2 31969 583 31970 31205 31971 31206 31972 31207 31973 2756 31974 31209 31975 31210 31976 364 31977 364 31978 367 31979 441 31980 31215 31981 31216 31982 31217 31983 31218 31984 2 31985 583 31986 31221 31987 31222 31988 31223 31989 31224 31990 2827 31991 31226 31992 441 31993 31228 31994 2 31995 31230 31996 31231 31997 30880 31998 2 31999 31234 32000 31235 32001 31236 32002 31237 32003 31238 32004 2918 32005 31240 32006 69 32007 31242 32008 31243 32009 31244 32010 31245 32011 31246 32012 31247 32013 31248 32014 2 32015 31250 32016 31251 32017 31252 32018 31253 32019 31254 32020 31255 32021 2960 32022 2962 32023 31258 32024 31259 32025 31260 32026 31261 32027 31262 32028 364 32029 364 32030 40 32031 367 32032 441 32033 31268 32034 31269 32035 31270 32036 31271 32037 31272 32038 40 32039 31274 32040 31275 32041 2983 32042 2986 32043 2988 32044 2990 32045 77 32046 2992 32047 2993 32048 31283 32049 287 32050 31285 32051 2 32052 287 32053 77 32054 3959 32055 2852 32056 31291 32057 1882 32058 164 32059 2072 32060 2 32061 31296 32062 2 32063 161 32064 164 32065 31300 32066 31301 32067 31302 32068 31303 32069 31304 32070 956 32071 31306 32072 31307 32073 31308 32074 989 32075 161 32076 164 32077 31312 32078 31313 32079 2 32080 287 32081 31316 32082 2 32083 287 32084 31319 32085 31320 32086 31321 32087 956 32088 31323 32089 31324 32090 31325 32091 989 32092 5285 32093 161 32094 164 32095 31330 32096 31331 32097 956 32098 31333 32099 31334 32100 31335 32101 31336 32102 1882 32103 164 32104 31339 32105 31340 32106 31341 32107 31342 32108 287 32109 31344 32110 2 32111 583 32112 31347 32113 5356 32114 31349 32115 31350 32116 2 32117 287 32118 17884 32119 17885 32120 32120 32121 32121 32122 17888 32123 7507 32124 32124 32125 105 32126 32126 32127 32127 32128 32128 32129 32129 32130 32130 32131 32131 32132 32132 32133 7507 32134 32134 32135 32135 32136 3066 32137 32137 32138 7507 32139 32132 32140 3066 32141 32137 32142 32142 32143 32143 32144 32144 32145 32145 32146 32146 32147 32147 32148 32148 32149 16 32150 32150 32151 77 32152 32152 32153 32153 32154 32154 32155 32155 32156 32156 32157 4037 32158 32158 32159 2852 32160 32160 32161 32161 32162 16 32163 32163 32164 32164 32165 32165 32166 32166 32167 77 32168 32168 32169 32169 32170 32170 32171 32171 32172 4037 32173 32173 32174 32152 32175 32175 32176 32176 32177 2852 32178 32178 32179 32179 32180 31351 32181 22 32182 161 32183 164 32184 7507 32185 32185 32186 128 32187 172 32188 32188 32189 32189 32190 11 32191 77 32192 32192 32193 128 32194 172 32195 32195 32196 13 32197 18061 32198 164 32199 77 32200 128 32201 172 32202 32202 32203 2990 32204 77 32205 2992 32206 2993 32207 32207 32208 31352 32209 2 32210 287 32211 18074 32212 18075 32213 32213 32214 18077 32215 18078 32216 18079 32217 32217 32218 18079 32219 32219 32220 18217 32221 18218 32222 7507 32223 32223 32224 18218 32225 7507 32226 32226 32227 31353 32228 32228 32229 3066 32230 32230 32231 2918 32232 31356 32233 69 32234 31358 32235 31359 32236 7507 32237 32237 32238 31360 32239 31361 32240 31363 32241 31364 32242 2 32243 31366 32244 31367 32245 31368 32246 31369 32247 31370 32248 31371 32249 2960 32250 2962 32251 31374 32252 31375 32253 31376 32254 31377 32255 31378 32256 364 32257 364 32258 40 32259 367 32260 441 32261 31384 32262 31385 32263 31386 32264 31387 32265 31388 32266 40 32267 31390 32268 31391 32269 2983 32270 2986 32271 2988 32272 2990 32273 77 32274 2992 32275 2993 32276 31399 32277 287 32278 31401 32279 32279 32280 32280 32281 32281 32282 32282 32283 32283 32284 32284 32285 32285 32286 32286 32287 32287 32288 32288 32289 32289 32290 32290 32291 32291 32292 32292 32293 32293 32294 32294 32295 32295 32296 32296 32297 32297 32298 32298 32299 32299 32300 32300 32301 32301 32302 32302 32303 32303 32304 32304 32305 32305 32306 32306 32307 32307 32308 32308 32309 32309 32310 32310 32311 32311 32312 32312 32313 32313 32314 32314 32315 32315 32316 32316 32317 32317 32318 32318 32319 32319 32320 32320 32321 32321 32322 32322 32323 32323 32324 32324 32325 32325 32326 32326 32327 32327 32328 32328 32329 32329 32330 32330 32331 32331 32332 32332 32333 32333 32334 32334 32335 32335 32336 32336 32337 32337 32338 32338 32339 32339 32340 32340 32341 32341 32342 32342 32343 32343 32344 32344 32345 32345 32346 32346 32347 32347 32348 32348 32349 32349 32350 32350 32351 32351 32352 32352 32353 32353 32354 32354 32355 32355 32356 32356 32357 32357 32358 32358 32359 32359 32360 7191 32361 32361 32362 32362 32363 32363 32364 32364 32365 32365 32366 15 32367 16 32368 16 32369 15 32370 11 32371 26 32372 2 32373 16 32374 11 32375 11 32376 40 32377 16 32378 11 32379 11 32380 32380 32381 16 32382 56 32383 15 32384 70 32385 71 32386 11 32387 16 32388 77 32389 11 32390 80 32391 2 32392 11 32393 32393 32394 89 32395 32395 32396 32396 32397 94 32398 32398 32399 100 32400 32400 32401 105 32402 108 32403 32403 32404 32404 32405 32405 32406 32406 32407 32407 32408 32408 32409 32409 32410 128 32411 32411 32412 172 32413 32413 32414 26 32415 11 32416 2 32417 287 32418 26 32419 32419 32420 15 32421 15 32422 11 32423 26 32424 2 32425 11 32426 11 32427 40 32428 16 32429 32429 32430 15 32431 70 32432 71 32433 11 32434 77 32435 11 32436 2 32437 11 32438 32438 32439 89 32440 105 32441 32441 32442 32405 32443 32443 32444 128 32445 172 32446 32446 32447 26 32448 32448 32449 32449 32450 1080 32451 7507 32452 32452 32453 17805 32454 17806 32455 7507 32456 32456 32457 2 32458 287 32459 17884 32460 17885 32461 32461 32462 32462 32463 17888 32464 7507 32465 32465 32466 105 32467 32467 32468 22 32469 161 32470 164 32471 7507 32472 32472 32473 128 32474 172 32475 32475 32476 32476 32477 11 32478 77 32479 32479 32480 128 32481 172 32482 32482 32483 13 32484 18061 32485 164 32486 77 32487 128 32488 172 32489 32489 32490 2 32491 287 32492 18074 32493 18075 32494 32494 32495 18077 32496 18078 32497 18079 32498 32498 32499 18079 32500 32500 32501 18217 32502 18218 32503 7507 32504 32504 32505 18218 32506 7507 32507 32507 32508 32508 32509 3066 32510 32510 32511 7507 32512 32512 32513 7507 32514 32514 32515 3066 32516 32516 32517 2 32518 19517 32519 20896 32520 20907 32521 17884 32522 20896 32523 164 32524 17884 32525 20958 32526 32526 32527 32527 32528 20961 32529 7507 32530 32530 32531 32531 32532 21125 32533 7507 32534 21134 32535 7507 32536 22736 32537 164 32538 17884 32539 22746 32540 32540 32541 32541 32542 22749 32543 7507 32544 32544 32545 3066 32546 32546 32547 22957 32548 11 32549 2 32550 287 32551 7507 32552 32552 32553 16 32554 23108 32555 17884 32556 23112 32557 32557 32558 32558 32559 23115 32560 7507 32561 32561 32562 7507 32563 32563 32564 7507 32565 32565 32566 32566 32567 17773 32568 142 32569 11 32570 144 32571 164 32572 167 32573 169 32574 77 32575 128 32576 172 32577 32577 32578 32578 32579 32579 32580 16 32581 2986 32582 2988 32583 2990 32584 77 32585 2992 32586 2993 32587 32587 32588 32588 32589 1080 32590 7507 32591 32591 32592 17805 32593 17806 32594 7507 32595 32595 32596 2 32597 287 32598 17884 32599 17885 32600 32600 32601 32601 32602 17888 32603 7507 32604 32604 32605 105 32606 32606 32607 7507 32608 3066 32609 32137 32610 7507 32611 3066 32612 32612 32613 32613 32614 32614 32615 32615 32616 22 32617 161 32618 164 32619 7507 32620 32620 32621 128 32622 172 32623 32623 32624 32624 32625 11 32626 77 32627 32627 32628 128 32629 172 32630 32630 32631 13 32632 18061 32633 164 32634 77 32635 128 32636 172 32637 32637 32638 2990 32639 77 32640 2992 32641 2993 32642 32642 32643 2 32644 287 32645 18074 32646 18075 32647 32647 32648 18077 32649 18078 32650 18079 32651 32651 32652 18079 32653 32653 32654 18217 32655 18218 32656 7507 32657 32657 32658 18218 32659 7507 32660 32660 32661 32661 32662 3066 32663 32663 32664 7507 32665 32665 32666 7191 32667 32363 32668 32364 32669 32669 32670 32670 32671 32363 32672 32364 32673 32673 32674 23404 32675 2990 32676 77 32677 2992 32678 2993 32679 32679 32680 2 32681 287 32682 17884 32683 32683 32684 164 32685 7507 32686 32686 32687 7507 32688 32688 32689 3066 32690 32690 32691 2 32692 19517 32693 20896 32694 20907 32695 17884 32696 20896 32697 164 32698 17884 32699 20958 32700 32700 32701 32701 32702 20961 32703 7507 32704 32704 32705 32705 32706 21125 32707 7507 32708 21134 32709 7507 32710 22736 32711 164 32712 17884 32713 22746 32714 32714 32715 32715 32716 22749 32717 7507 32718 32718 32719 23453 32720 2988 32721 2990 32722 77 32723 2992 32724 2993 32725 32725 32726 3066 32727 32727 32728 22957 32729 11 32730 2 32731 287 32732 7507 32733 32733 32734 16 32735 23108 32736 17884 32737 23112 32738 32738 32739 32739 32740 23115 32741 7507 32742 32742 32743 7507 32744 32744 32745 3066 32746 32746 32747 2986 32748 2988 32749 2990 32750 77 32751 2992 32752 2993 32753 32753 32754 7507 32755 32755 32756 32756 32757 32363 32758 32364 32759 32759 32760 2 32761 287 32762 77 32763 3959 32764 2852 32765 31407 32766 1882 32767 164 32768 2072 32769 2 32770 31412 32771 2 32772 161 32773 164 32774 31416 32775 31417 32776 23404 32777 2990 32778 77 32779 2992 32780 2993 32781 32781 32782 31418 32783 31419 32784 31420 32785 2 32786 287 32787 17884 32788 32788 32789 15 32790 15 32791 11 32792 26 32793 2 32794 11 32795 11 32796 40 32797 16 32798 11 32799 32799 32800 15 32801 70 32802 71 32803 11 32804 77 32805 11 32806 2 32807 11 32808 32808 32809 89 32810 105 32811 32811 32812 32812 32813 32813 32814 128 32815 172 32816 32816 32817 26 32818 11 32819 2 32820 287 32821 26 32822 32822 32823 32823 32824 32824 32825 1080 32826 7507 32827 32827 32828 17805 32829 17806 32830 7507 32831 32831 32832 2 32833 287 32834 17884 32835 17885 32836 32836 32837 32837 32838 17888 32839 7507 32840 32840 32841 105 32842 32842 32843 22 32844 161 32845 164 32846 7507 32847 32847 32848 128 32849 172 32850 32850 32851 32851 32852 11 32853 77 32854 32854 32855 128 32856 172 32857 32857 32858 13 32859 18061 32860 164 32861 77 32862 128 32863 172 32864 32864 32865 2 32866 287 32867 18074 32868 18075 32869 32869 32870 18077 32871 18078 32872 18079 32873 32873 32874 18079 32875 32875 32876 18217 32877 18218 32878 7507 32879 32879 32880 18218 32881 7507 32882 32882 32883 32883 32884 3066 32885 32885 32886 7507 32887 32887 32888 2 32889 287 32890 17884 32891 32891 32892 7507 32893 32893 32894 3066 32895 32895 32896 2 32897 19517 32898 20896 32899 20907 32900 17884 32901 20896 32902 164 32903 17884 32904 20958 32905 32905 32906 32906 32907 20961 32908 7507 32909 32909 32910 32910 32911 21125 32912 7507 32913 21134 32914 7507 32915 22736 32916 164 32917 17884 32918 22746 32919 32919 32920 32920 32921 22749 32922 7507 32923 32923 32924 3066 32925 32925 32926 22957 32927 11 32928 2 32929 287 32930 7507 32931 32931 32932 16 32933 23108 32934 17884 32935 23112 32936 32936 32937 32937 32938 23115 32939 7507 32940 32940 32941 7507 32942 32942 32943 7507 32944 32944 32945 164 32946 7507 32947 32947 32948 7507 32949 32949 32950 3066 32951 32951 32952 2 32953 19517 32954 20896 32955 20907 32956 17884 32957 20896 32958 164 32959 17884 32960 20958 32961 32961 32962 32962 32963 20961 32964 7507 32965 32965 32966 32966 32967 21125 32968 7507 32969 21134 32970 7507 32971 22736 32972 164 32973 17884 32974 22746 32975 32975 32976 32976 32977 22749 32978 7507 32979 32979 32980 956 32981 31422 32982 31423 32983 31424 32984 989 32985 161 32986 164 32987 31428 32988 31429 32989 2 32990 287 32991 31432 32992 23453 32993 2988 32994 2990 32995 77 32996 2992 32997 2993 32998 32998 32999 2 33000 287 33001 31435 33002 3066 33003 33003 33004 31436 33005 22957 33006 11 33007 2 33008 287 33009 7507 33010 33010 33011 16 33012 23108 33013 17884 33014 23112 33015 33015 33016 33016 33017 23115 33018 7507 33019 33019 33020 31437 33021 7507 33022 33022 33023 3066 33024 33024 33025 956 33026 31439 33027 31440 33028 31441 33029 989 33030 5285 33031 161 33032 164 33033 31446 33034 31447 33035 956 33036 31449 33037 31450 33038 2986 33039 2988 33040 2990 33041 77 33042 2992 33043 2993 33044 33044 33045 7507 33046 33046 33047 31451 33048 6002 33049 441 33050 33050 33051 31452 33052 1882 33053 164 33054 31455 33055 31456 33056 31457 33057 31458 33058 287 33059 31460 33060 2 33061 583 33062 31463 33063 5356 33064 31465 33065 31466))
  "The generated action-tables.")

(defvar
  phps-mode-parser--distinct-action-tables
  #s(hash-table test equal data (0 ((("!") reduce 85) (("\"") reduce 85) (($) reduce 85) (("$") reduce 85) (("(") reduce 85) (("+") reduce 85) (("-") reduce 85) ((";") reduce 85) (("@") reduce 85) ((T_ABSTRACT) reduce 85) ((T_ARRAY) reduce 85) ((T_ARRAY_CAST) reduce 85) ((T_ATTRIBUTE) reduce 85) ((T_BOOL_CAST) reduce 85) ((T_BREAK) reduce 85) ((T_CLASS) reduce 85) ((T_CLASS_C) reduce 85) ((T_CLONE) reduce 85) ((T_CONST) reduce 85) ((T_CONSTANT_ENCAPSED_STRING) reduce 85) ((T_CONTINUE) reduce 85) ((T_DEC) reduce 85) ((T_DECLARE) reduce 85) ((T_DIR) reduce 85) ((T_DNUMBER) reduce 85) ((T_DO) reduce 85) ((T_DOUBLE_CAST) reduce 85) ((T_ECHO) reduce 85) ((T_EMPTY) reduce 85) ((T_ENUM) reduce 85) ((T_EVAL) reduce 85) ((T_EXIT) reduce 85) ((T_FILE) reduce 85) ((T_FINAL) reduce 85) ((T_FN) reduce 85) ((T_FOR) reduce 85) ((T_FOREACH) reduce 85) ((T_FUNCTION) reduce 85) ((T_FUNC_C) reduce 85) ((T_GLOBAL) reduce 85) ((T_GOTO) reduce 85) ((T_HALT_COMPILER) reduce 85) ((T_IF) reduce 85) ((T_INC) reduce 85) ((T_INCLUDE) reduce 85) ((T_INCLUDE_ONCE) reduce 85) ((T_INLINE_HTML) reduce 85) ((T_INTERFACE) reduce 85) ((T_INT_CAST) reduce 85) ((T_ISSET) reduce 85) ((T_LINE) reduce 85) ((T_LIST) reduce 85) ((T_LNUMBER) reduce 85) ((T_MATCH) reduce 85) ((T_METHOD_C) reduce 85) ((T_NAMESPACE) reduce 85) ((T_NAME_FULLY_QUALIFIED) reduce 85) ((T_NAME_QUALIFIED) reduce 85) ((T_NAME_RELATIVE) reduce 85) ((T_NEW) reduce 85) ((T_NS_C) reduce 85) ((T_OBJECT_CAST) reduce 85) ((T_PRINT) reduce 85) ((T_PROPERTY_C) reduce 85) ((T_READONLY) reduce 85) ((T_REQUIRE) reduce 85) ((T_REQUIRE_ONCE) reduce 85) ((T_RETURN) reduce 85) ((T_START_HEREDOC) reduce 85) ((T_STATIC) reduce 85) ((T_STRING) reduce 85) ((T_STRING_CAST) reduce 85) ((T_SWITCH) reduce 85) ((T_THROW) reduce 85) ((T_TRAIT) reduce 85) ((T_TRAIT_C) reduce 85) ((T_TRY) reduce 85) ((T_UNSET) reduce 85) ((T_UNSET_CAST) reduce 85) ((T_USE) reduce 85) ((T_VARIABLE) reduce 85) ((T_VOID_CAST) reduce 85) ((T_WHILE) reduce 85) ((T_YIELD) reduce 85) ((T_YIELD_FROM) reduce 85) (("[") reduce 85) (("`") reduce 85) (("{") reduce 85) (("~") reduce 85)) 1 ((("!") shift) (("\"") shift) (($) accept) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONST) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAMESPACE) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_USE) shift) ((T_VARIABLE) shift) ((T_VOID_CAST) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("~") shift)) 2 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 3 (((T_CURLY_OPEN) shift) ((T_DOLLAR_OPEN_CURLY_BRACES) shift) ((T_ENCAPSED_AND_WHITESPACE) shift) ((T_VARIABLE) shift)) 4 ((("$") shift) ((T_VARIABLE) shift) (("{") shift)) 8 ((("!") reduce 166) (("\"") reduce 166) (($) reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONST) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_EMPTY) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAMESPACE) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_USE) reduce 166) ((T_VARIABLE) reduce 166) ((T_VOID_CAST) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 10 (((T_ABSTRACT) reduce 197) ((T_CLASS) reduce 197) ((T_FINAL) reduce 197) ((T_READONLY) reduce 197)) 11 ((("(") shift)) 13 (((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_STATIC) shift) ((T_STRING) shift)) 15 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") reduce 545) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 16 (((T_STRING) shift)) 17 ((("%") reduce 540) (("*") reduce 540) (("+") reduce 540) (("-") reduce 540) ((".") reduce 540) (("/") reduce 540) ((";") reduce 540) (("<") reduce 540) ((">") reduce 540) (("?") reduce 540) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_BOOLEAN_AND) reduce 540) ((T_BOOLEAN_OR) reduce 540) ((T_COALESCE) reduce 540) ((T_INSTANCEOF) reduce 540) ((T_IS_EQUAL) reduce 540) ((T_IS_GREATER_OR_EQUAL) reduce 540) ((T_IS_IDENTICAL) reduce 540) ((T_IS_NOT_EQUAL) reduce 540) ((T_IS_NOT_IDENTICAL) reduce 540) ((T_IS_SMALLER_OR_EQUAL) reduce 540) ((T_LOGICAL_AND) reduce 540) ((T_LOGICAL_OR) reduce 540) ((T_LOGICAL_XOR) reduce 540) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 540) ((T_OBJECT_OPERATOR) reduce 540) ((T_PIPE) reduce 540) ((T_POW) reduce 540) ((T_SL) reduce 540) ((T_SPACESHIP) reduce 540) ((T_SR) reduce 540) (("[") reduce 540) (("^") reduce 540) (("|") reduce 540)) 20 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) ((";") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_PIPE) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("|") reduce 521)) 22 ((("\"") shift) (("$") shift) (("(") shift) ((T_ARRAY) shift) ((T_CLASS_C) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DIR) shift) ((T_FILE) shift) ((T_FUNC_C) shift) ((T_LINE) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_TRAIT_C) shift) ((T_VARIABLE) shift) (("[") shift)) 24 ((("%") reduce 534) (("*") reduce 534) (("+") reduce 534) (("-") reduce 534) ((".") reduce 534) (("/") reduce 534) ((";") reduce 534) (("<") reduce 534) ((">") reduce 534) (("?") reduce 534) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_BOOLEAN_AND) reduce 534) ((T_BOOLEAN_OR) reduce 534) ((T_COALESCE) reduce 534) ((T_INSTANCEOF) reduce 534) ((T_IS_EQUAL) reduce 534) ((T_IS_GREATER_OR_EQUAL) reduce 534) ((T_IS_IDENTICAL) reduce 534) ((T_IS_NOT_EQUAL) reduce 534) ((T_IS_NOT_IDENTICAL) reduce 534) ((T_IS_SMALLER_OR_EQUAL) reduce 534) ((T_LOGICAL_AND) reduce 534) ((T_LOGICAL_OR) reduce 534) ((T_LOGICAL_XOR) reduce 534) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PIPE) reduce 534) ((T_POW) reduce 534) ((T_SL) reduce 534) ((T_SPACESHIP) reduce 534) ((T_SR) reduce 534) (("[") reduce 534) (("^") reduce 534) (("|") reduce 534)) 25 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) ((";") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_PIPE) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("^") reduce 524) (("|") reduce 524)) 26 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_VOID_CAST) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("~") shift)) 32 ((("%") reduce 517) (("(") shift) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((";") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_PIPE) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("^") reduce 517) (("|") reduce 517)) 33 ((("%") reduce 533) (("*") reduce 533) (("+") reduce 533) (("-") reduce 533) ((".") reduce 533) (("/") reduce 533) ((";") reduce 533) (("<") reduce 533) ((">") reduce 533) (("?") reduce 533) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_BOOLEAN_AND) reduce 533) ((T_BOOLEAN_OR) reduce 533) ((T_COALESCE) reduce 533) ((T_INSTANCEOF) reduce 533) ((T_IS_EQUAL) reduce 533) ((T_IS_GREATER_OR_EQUAL) reduce 533) ((T_IS_IDENTICAL) reduce 533) ((T_IS_NOT_EQUAL) reduce 533) ((T_IS_NOT_IDENTICAL) reduce 533) ((T_IS_SMALLER_OR_EQUAL) reduce 533) ((T_LOGICAL_AND) reduce 533) ((T_LOGICAL_OR) reduce 533) ((T_LOGICAL_XOR) reduce 533) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PIPE) reduce 533) ((T_POW) reduce 533) ((T_SL) reduce 533) ((T_SPACESHIP) reduce 533) ((T_SR) reduce 533) (("[") reduce 533) (("^") reduce 533) (("|") reduce 533)) 34 (((T_ABSTRACT) reduce 198) ((T_CLASS) reduce 198) ((T_FINAL) reduce 198) ((T_READONLY) reduce 198)) 35 ((("(") reduce 491) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491)) 38 ((("(") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_READONLY) reduce 492) ((T_STRING) reduce 492)) 39 ((("%") reduce 537) (("*") reduce 537) (("+") reduce 537) (("-") reduce 537) ((".") reduce 537) (("/") reduce 537) ((";") reduce 537) (("<") reduce 537) ((">") reduce 537) (("?") reduce 537) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_BOOLEAN_AND) reduce 537) ((T_BOOLEAN_OR) reduce 537) ((T_COALESCE) reduce 537) ((T_INSTANCEOF) reduce 537) ((T_IS_EQUAL) reduce 537) ((T_IS_GREATER_OR_EQUAL) reduce 537) ((T_IS_IDENTICAL) reduce 537) ((T_IS_NOT_EQUAL) reduce 537) ((T_IS_NOT_IDENTICAL) reduce 537) ((T_IS_SMALLER_OR_EQUAL) reduce 537) ((T_LOGICAL_AND) reduce 537) ((T_LOGICAL_OR) reduce 537) ((T_LOGICAL_XOR) reduce 537) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_PIPE) reduce 537) ((T_POW) reduce 537) ((T_SL) reduce 537) ((T_SPACESHIP) reduce 537) ((T_SR) reduce 537) (("[") reduce 537) (("^") reduce 537) (("|") reduce 537)) 40 ((("$") shift) ((T_VARIABLE) shift)) 47 ((("!") reduce 160) (("\"") reduce 160) (($) reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONST) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_EMPTY) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAMESPACE) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_USE) reduce 160) ((T_VARIABLE) reduce 160) ((T_VOID_CAST) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 51 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) ((";") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PIPE) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("|") reduce 532)) 53 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) ((";") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_PIPE) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("^") reduce 523) (("|") reduce 523)) 55 ((("%") reduce 536) (("*") reduce 536) (("+") reduce 536) (("-") reduce 536) ((".") reduce 536) (("/") reduce 536) ((";") reduce 536) (("<") reduce 536) ((">") reduce 536) (("?") reduce 536) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_BOOLEAN_AND) reduce 536) ((T_BOOLEAN_OR) reduce 536) ((T_COALESCE) reduce 536) ((T_INSTANCEOF) reduce 536) ((T_IS_EQUAL) reduce 536) ((T_IS_GREATER_OR_EQUAL) reduce 536) ((T_IS_IDENTICAL) reduce 536) ((T_IS_NOT_EQUAL) reduce 536) ((T_IS_NOT_IDENTICAL) reduce 536) ((T_IS_SMALLER_OR_EQUAL) reduce 536) ((T_LOGICAL_AND) reduce 536) ((T_LOGICAL_OR) reduce 536) ((T_LOGICAL_XOR) reduce 536) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 536) ((T_OBJECT_OPERATOR) reduce 536) ((T_PIPE) reduce 536) ((T_POW) reduce 536) ((T_SL) reduce 536) ((T_SPACESHIP) reduce 536) ((T_SR) reduce 536) (("[") reduce 536) (("^") reduce 536) (("|") reduce 536)) 56 (((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_AS) shift) ((T_BREAK) shift) ((T_CALLABLE) shift) ((T_CASE) shift) ((T_CATCH) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONST) shift) ((T_CONTINUE) shift) ((T_DECLARE) shift) ((T_DEFAULT) shift) ((T_DIR) shift) ((T_DO) shift) ((T_ECHO) shift) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) shift) ((T_ENDDECLARE) shift) ((T_ENDFOR) shift) ((T_ENDFOREACH) shift) ((T_ENDIF) shift) ((T_ENDSWITCH) shift) ((T_ENDWHILE) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_EXTENDS) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FINALLY) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_IF) shift) ((T_IMPLEMENTS) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) shift) ((T_INSTEADOF) shift) ((T_INTERFACE) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAMESPACE) shift) ((T_NAME_QUALIFIED) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_PRINT) shift) ((T_PRIVATE) shift) ((T_PROPERTY_C) shift) ((T_PROTECTED) shift) ((T_PUBLIC) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_USE) shift) ((T_VAR) shift) ((T_WHILE) shift) ((T_YIELD) shift) (("{") shift)) 57 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("|") reduce 94)) 58 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("|") reduce 93)) 59 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("|") reduce 95)) 60 ((("$") shift) (("(") shift) ((T_ABSTRACT) shift) ((T_ATTRIBUTE) shift) ((T_CLASS) reduce 195) ((T_FINAL) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_READONLY) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_VARIABLE) shift)) 61 ((("%") reduce 539) (("*") reduce 539) (("+") reduce 539) (("-") reduce 539) ((".") reduce 539) (("/") reduce 539) ((";") reduce 539) (("<") reduce 539) ((">") reduce 539) (("?") reduce 539) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_BOOLEAN_AND) reduce 539) ((T_BOOLEAN_OR) reduce 539) ((T_COALESCE) reduce 539) ((T_INSTANCEOF) reduce 539) ((T_IS_EQUAL) reduce 539) ((T_IS_GREATER_OR_EQUAL) reduce 539) ((T_IS_IDENTICAL) reduce 539) ((T_IS_NOT_EQUAL) reduce 539) ((T_IS_NOT_IDENTICAL) reduce 539) ((T_IS_SMALLER_OR_EQUAL) reduce 539) ((T_LOGICAL_AND) reduce 539) ((T_LOGICAL_OR) reduce 539) ((T_LOGICAL_XOR) reduce 539) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 539) ((T_OBJECT_OPERATOR) reduce 539) ((T_PIPE) reduce 539) ((T_POW) reduce 539) ((T_SL) reduce 539) ((T_SPACESHIP) reduce 539) ((T_SR) reduce 539) (("[") reduce 539) (("^") reduce 539) (("|") reduce 539)) 64 ((("%") reduce 538) (("*") reduce 538) (("+") reduce 538) (("-") reduce 538) ((".") reduce 538) (("/") reduce 538) ((";") reduce 538) (("<") reduce 538) ((">") reduce 538) (("?") reduce 538) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_BOOLEAN_AND) reduce 538) ((T_BOOLEAN_OR) reduce 538) ((T_COALESCE) reduce 538) ((T_INSTANCEOF) reduce 538) ((T_IS_EQUAL) reduce 538) ((T_IS_GREATER_OR_EQUAL) reduce 538) ((T_IS_IDENTICAL) reduce 538) ((T_IS_NOT_EQUAL) reduce 538) ((T_IS_NOT_IDENTICAL) reduce 538) ((T_IS_SMALLER_OR_EQUAL) reduce 538) ((T_LOGICAL_AND) reduce 538) ((T_LOGICAL_OR) reduce 538) ((T_LOGICAL_XOR) reduce 538) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 538) ((T_OBJECT_OPERATOR) reduce 538) ((T_PIPE) reduce 538) ((T_POW) reduce 538) ((T_SL) reduce 538) ((T_SPACESHIP) reduce 538) ((T_SR) reduce 538) (("[") reduce 538) (("^") reduce 538) (("|") reduce 538)) 65 ((("(") shift) ((T_ABSTRACT) reduce 199) ((T_CLASS) reduce 199) ((T_FINAL) reduce 199) ((T_READONLY) reduce 199)) 69 (((T_CURLY_OPEN) shift) ((T_DOLLAR_OPEN_CURLY_BRACES) shift) ((T_ENCAPSED_AND_WHITESPACE) shift) ((T_END_HEREDOC) shift) ((T_VARIABLE) shift)) 70 (((T_FN) shift) ((T_FUNCTION) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_VARIABLE) shift)) 71 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((":") shift) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92)) 76 ((("%") reduce 535) (("*") reduce 535) (("+") reduce 535) (("-") reduce 535) ((".") reduce 535) (("/") reduce 535) ((";") reduce 535) (("<") reduce 535) ((">") reduce 535) (("?") reduce 535) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_BOOLEAN_AND) reduce 535) ((T_BOOLEAN_OR) reduce 535) ((T_COALESCE) reduce 535) ((T_INSTANCEOF) reduce 535) ((T_IS_EQUAL) reduce 535) ((T_IS_GREATER_OR_EQUAL) reduce 535) ((T_IS_IDENTICAL) reduce 535) ((T_IS_NOT_EQUAL) reduce 535) ((T_IS_NOT_IDENTICAL) reduce 535) ((T_IS_SMALLER_OR_EQUAL) reduce 535) ((T_LOGICAL_AND) reduce 535) ((T_LOGICAL_OR) reduce 535) ((T_LOGICAL_XOR) reduce 535) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PIPE) reduce 535) ((T_POW) reduce 535) ((T_SL) reduce 535) ((T_SPACESHIP) reduce 535) ((T_SR) reduce 535) (("[") reduce 535) (("^") reduce 535) (("|") reduce 535)) 77 ((("{") shift)) 80 (((T_CONST) shift) ((T_FUNCTION) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_STRING) shift)) 81 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) ((";") reduce 568) (("<") reduce 568) (("=") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AND_EQUAL) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_COALESCE_EQUAL) reduce 568) ((T_CONCAT_EQUAL) reduce 568) ((T_DEC) reduce 568) ((T_DIV_EQUAL) reduce 568) ((T_INC) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_MINUS_EQUAL) reduce 568) ((T_MOD_EQUAL) reduce 568) ((T_MUL_EQUAL) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_OR_EQUAL) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_PLUS_EQUAL) reduce 568) ((T_POW) reduce 568) ((T_POW_EQUAL) reduce 568) ((T_SL) reduce 568) ((T_SL_EQUAL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) ((T_SR_EQUAL) reduce 568) ((T_XOR_EQUAL) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 84 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 479) (("(") shift) (("*") reduce 479) (("+") shift) (("-") shift) ((".") reduce 479) (("/") reduce 479) ((";") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 479) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 479) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PIPE) reduce 479) ((T_POW) reduce 479) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 479) (("`") shift) (("|") reduce 479) (("~") shift)) 86 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) ((",") reduce 586) (("-") shift) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_ELLIPSIS) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("]") reduce 586) (("`") shift) (("~") shift)) 87 (((T_CURLY_OPEN) shift) ((T_DOLLAR_OPEN_CURLY_BRACES) shift) ((T_ENCAPSED_AND_WHITESPACE) shift) ((T_VARIABLE) shift) (("`") reduce 514)) 88 ((("!") reduce 149) (("\"") reduce 149) (($) reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONST) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_EMPTY) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAMESPACE) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_USE) reduce 149) ((T_VARIABLE) reduce 149) ((T_VOID_CAST) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 89 (((T_ELSE) shift) ((T_ELSEIF) shift) ((T_ENDIF) shift)) 90 (((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) (("[") shift)) 91 (((T_ABSTRACT) reduce 101) ((T_ATTRIBUTE) reduce 101) ((T_CLASS) reduce 101) ((T_CONST) reduce 101) ((T_ENUM) reduce 101) ((T_FINAL) reduce 101) ((T_FN) reduce 101) ((T_FUNCTION) reduce 101) ((T_INTERFACE) reduce 101) ((T_READONLY) reduce 101) ((T_STATIC) reduce 101) ((T_TRAIT) reduce 101)) 92 ((("!") reduce 108) (("\"") reduce 108) (($) reduce 108) (("$") reduce 108) (("(") reduce 108) (("+") reduce 108) (("-") reduce 108) ((";") reduce 108) (("@") reduce 108) ((T_ABSTRACT) reduce 108) ((T_ARRAY) reduce 108) ((T_ARRAY_CAST) reduce 108) ((T_ATTRIBUTE) reduce 108) ((T_BOOL_CAST) reduce 108) ((T_BREAK) reduce 108) ((T_CLASS) reduce 108) ((T_CLASS_C) reduce 108) ((T_CLONE) reduce 108) ((T_CONST) reduce 108) ((T_CONSTANT_ENCAPSED_STRING) reduce 108) ((T_CONTINUE) reduce 108) ((T_DEC) reduce 108) ((T_DECLARE) reduce 108) ((T_DIR) reduce 108) ((T_DNUMBER) reduce 108) ((T_DO) reduce 108) ((T_DOUBLE_CAST) reduce 108) ((T_ECHO) reduce 108) ((T_EMPTY) reduce 108) ((T_ENUM) reduce 108) ((T_EVAL) reduce 108) ((T_EXIT) reduce 108) ((T_FILE) reduce 108) ((T_FINAL) reduce 108) ((T_FN) reduce 108) ((T_FOR) reduce 108) ((T_FOREACH) reduce 108) ((T_FUNCTION) reduce 108) ((T_FUNC_C) reduce 108) ((T_GLOBAL) reduce 108) ((T_GOTO) reduce 108) ((T_HALT_COMPILER) reduce 108) ((T_IF) reduce 108) ((T_INC) reduce 108) ((T_INCLUDE) reduce 108) ((T_INCLUDE_ONCE) reduce 108) ((T_INLINE_HTML) reduce 108) ((T_INTERFACE) reduce 108) ((T_INT_CAST) reduce 108) ((T_ISSET) reduce 108) ((T_LINE) reduce 108) ((T_LIST) reduce 108) ((T_LNUMBER) reduce 108) ((T_MATCH) reduce 108) ((T_METHOD_C) reduce 108) ((T_NAMESPACE) reduce 108) ((T_NAME_FULLY_QUALIFIED) reduce 108) ((T_NAME_QUALIFIED) reduce 108) ((T_NAME_RELATIVE) reduce 108) ((T_NEW) reduce 108) ((T_NS_C) reduce 108) ((T_OBJECT_CAST) reduce 108) ((T_PRINT) reduce 108) ((T_PROPERTY_C) reduce 108) ((T_READONLY) reduce 108) ((T_REQUIRE) reduce 108) ((T_REQUIRE_ONCE) reduce 108) ((T_RETURN) reduce 108) ((T_START_HEREDOC) reduce 108) ((T_STATIC) reduce 108) ((T_STRING) reduce 108) ((T_STRING_CAST) reduce 108) ((T_SWITCH) reduce 108) ((T_THROW) reduce 108) ((T_TRAIT) reduce 108) ((T_TRAIT_C) reduce 108) ((T_TRY) reduce 108) ((T_UNSET) reduce 108) ((T_UNSET_CAST) reduce 108) ((T_USE) reduce 108) ((T_VARIABLE) reduce 108) ((T_VOID_CAST) reduce 108) ((T_WHILE) reduce 108) ((T_YIELD) reduce 108) ((T_YIELD_FROM) reduce 108) (("[") reduce 108) (("`") reduce 108) (("{") reduce 108) (("~") reduce 108)) 93 ((("!") reduce 111) (("\"") reduce 111) (("$") reduce 111) (($) reduce 111) (("(") reduce 111) (("+") reduce 111) (("-") reduce 111) ((";") reduce 111) (("@") reduce 111) ((T_ABSTRACT) reduce 111) ((T_ARRAY) reduce 111) ((T_ARRAY_CAST) reduce 111) ((T_ATTRIBUTE) reduce 111) ((T_BOOL_CAST) reduce 111) ((T_BREAK) reduce 111) ((T_CLASS) reduce 111) ((T_CLASS_C) reduce 111) ((T_CLONE) reduce 111) ((T_CONST) reduce 111) ((T_CONSTANT_ENCAPSED_STRING) reduce 111) ((T_CONTINUE) reduce 111) ((T_DEC) reduce 111) ((T_DECLARE) reduce 111) ((T_DIR) reduce 111) ((T_DNUMBER) reduce 111) ((T_DO) reduce 111) ((T_DOUBLE_CAST) reduce 111) ((T_ECHO) reduce 111) ((T_EMPTY) reduce 111) ((T_ENUM) reduce 111) ((T_EVAL) reduce 111) ((T_EXIT) reduce 111) ((T_FILE) reduce 111) ((T_FINAL) reduce 111) ((T_FN) reduce 111) ((T_FOR) reduce 111) ((T_FOREACH) reduce 111) ((T_FUNCTION) reduce 111) ((T_FUNC_C) reduce 111) ((T_GLOBAL) reduce 111) ((T_GOTO) reduce 111) ((T_HALT_COMPILER) reduce 111) ((T_IF) reduce 111) ((T_INC) reduce 111) ((T_INCLUDE) reduce 111) ((T_INCLUDE_ONCE) reduce 111) ((T_INLINE_HTML) reduce 111) ((T_INTERFACE) reduce 111) ((T_INT_CAST) reduce 111) ((T_ISSET) reduce 111) ((T_LINE) reduce 111) ((T_LIST) reduce 111) ((T_LNUMBER) reduce 111) ((T_MATCH) reduce 111) ((T_METHOD_C) reduce 111) ((T_NAMESPACE) reduce 111) ((T_NAME_FULLY_QUALIFIED) reduce 111) ((T_NAME_QUALIFIED) reduce 111) ((T_NAME_RELATIVE) reduce 111) ((T_NEW) reduce 111) ((T_NS_C) reduce 111) ((T_OBJECT_CAST) reduce 111) ((T_PRINT) reduce 111) ((T_PROPERTY_C) reduce 111) ((T_READONLY) reduce 111) ((T_REQUIRE) reduce 111) ((T_REQUIRE_ONCE) reduce 111) ((T_RETURN) reduce 111) ((T_START_HEREDOC) reduce 111) ((T_STATIC) reduce 111) ((T_STRING) reduce 111) ((T_STRING_CAST) reduce 111) ((T_SWITCH) reduce 111) ((T_THROW) reduce 111) ((T_TRAIT) reduce 111) ((T_TRAIT_C) reduce 111) ((T_TRY) reduce 111) ((T_UNSET) reduce 111) ((T_UNSET_CAST) reduce 111) ((T_USE) reduce 111) ((T_VARIABLE) reduce 111) ((T_VOID_CAST) reduce 111) ((T_WHILE) reduce 111) ((T_YIELD) reduce 111) ((T_YIELD_FROM) reduce 111) (("[") reduce 111) (("`") reduce 111) (("{") reduce 111) (("~") reduce 111)) 94 (((T_ABSTRACT) shift) ((T_ATTRIBUTE) shift) ((T_CLASS) shift) ((T_CONST) shift) ((T_ENUM) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_INTERFACE) shift) ((T_READONLY) shift) ((T_STATIC) shift) ((T_TRAIT) shift)) 96 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) ((";") reduce 564) (("<") reduce 564) (("=") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AND_EQUAL) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_COALESCE_EQUAL) reduce 564) ((T_CONCAT_EQUAL) reduce 564) ((T_DEC) reduce 564) ((T_DIV_EQUAL) reduce 564) ((T_INC) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_MINUS_EQUAL) reduce 564) ((T_MOD_EQUAL) reduce 564) ((T_MUL_EQUAL) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_OR_EQUAL) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_PLUS_EQUAL) reduce 564) ((T_POW) reduce 564) ((T_POW_EQUAL) reduce 564) ((T_SL) reduce 564) ((T_SL_EQUAL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) ((T_SR_EQUAL) reduce 564) ((T_XOR_EQUAL) reduce 564) (("[") reduce 564) (("^") reduce 564) (("|") reduce 564)) 97 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) ((";") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PIPE) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 551) (("^") reduce 530) (("|") reduce 530)) 98 ((("!") reduce 104) (("\"") reduce 104) (("$") reduce 104) (($) reduce 104) (("(") reduce 104) (("+") reduce 104) (("-") reduce 104) ((";") reduce 104) (("@") reduce 104) ((T_ABSTRACT) reduce 104) ((T_ARRAY) reduce 104) ((T_ARRAY_CAST) reduce 104) ((T_ATTRIBUTE) reduce 104) ((T_BOOL_CAST) reduce 104) ((T_BREAK) reduce 104) ((T_CLASS) reduce 104) ((T_CLASS_C) reduce 104) ((T_CLONE) reduce 104) ((T_CONST) reduce 104) ((T_CONSTANT_ENCAPSED_STRING) reduce 104) ((T_CONTINUE) reduce 104) ((T_DEC) reduce 104) ((T_DECLARE) reduce 104) ((T_DIR) reduce 104) ((T_DNUMBER) reduce 104) ((T_DO) reduce 104) ((T_DOUBLE_CAST) reduce 104) ((T_ECHO) reduce 104) ((T_EMPTY) reduce 104) ((T_ENUM) reduce 104) ((T_EVAL) reduce 104) ((T_EXIT) reduce 104) ((T_FILE) reduce 104) ((T_FINAL) reduce 104) ((T_FN) reduce 104) ((T_FOR) reduce 104) ((T_FOREACH) reduce 104) ((T_FUNCTION) reduce 104) ((T_FUNC_C) reduce 104) ((T_GLOBAL) reduce 104) ((T_GOTO) reduce 104) ((T_HALT_COMPILER) reduce 104) ((T_IF) reduce 104) ((T_INC) reduce 104) ((T_INCLUDE) reduce 104) ((T_INCLUDE_ONCE) reduce 104) ((T_INLINE_HTML) reduce 104) ((T_INTERFACE) reduce 104) ((T_INT_CAST) reduce 104) ((T_ISSET) reduce 104) ((T_LINE) reduce 104) ((T_LIST) reduce 104) ((T_LNUMBER) reduce 104) ((T_MATCH) reduce 104) ((T_METHOD_C) reduce 104) ((T_NAMESPACE) reduce 104) ((T_NAME_FULLY_QUALIFIED) reduce 104) ((T_NAME_QUALIFIED) reduce 104) ((T_NAME_RELATIVE) reduce 104) ((T_NEW) reduce 104) ((T_NS_C) reduce 104) ((T_OBJECT_CAST) reduce 104) ((T_PRINT) reduce 104) ((T_PROPERTY_C) reduce 104) ((T_READONLY) reduce 104) ((T_REQUIRE) reduce 104) ((T_REQUIRE_ONCE) reduce 104) ((T_RETURN) reduce 104) ((T_START_HEREDOC) reduce 104) ((T_STATIC) reduce 104) ((T_STRING) reduce 104) ((T_STRING_CAST) reduce 104) ((T_SWITCH) reduce 104) ((T_THROW) reduce 104) ((T_TRAIT) reduce 104) ((T_TRAIT_C) reduce 104) ((T_TRY) reduce 104) ((T_UNSET) reduce 104) ((T_UNSET_CAST) reduce 104) ((T_USE) reduce 104) ((T_VARIABLE) reduce 104) ((T_VOID_CAST) reduce 104) ((T_WHILE) reduce 104) ((T_YIELD) reduce 104) ((T_YIELD_FROM) reduce 104) (("[") reduce 104) (("`") reduce 104) (("{") reduce 104) (("~") reduce 104)) 99 (((T_ABSTRACT) reduce 191) ((T_CLASS) reduce 191) ((T_FINAL) reduce 191) ((T_READONLY) reduce 191)) 100 (((T_ABSTRACT) shift) ((T_CLASS) shift) ((T_FINAL) shift) ((T_READONLY) shift)) 101 (((T_PAAMAYIM_NEKUDOTAYIM) shift)) 102 ((("%") reduce 529) (("*") reduce 529) (("+") reduce 529) (("-") reduce 529) ((".") reduce 529) (("/") reduce 529) ((";") reduce 529) (("<") reduce 529) ((">") reduce 529) (("?") reduce 529) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_BOOLEAN_AND) reduce 529) ((T_BOOLEAN_OR) reduce 529) ((T_COALESCE) reduce 529) ((T_INSTANCEOF) reduce 529) ((T_IS_EQUAL) reduce 529) ((T_IS_GREATER_OR_EQUAL) reduce 529) ((T_IS_IDENTICAL) reduce 529) ((T_IS_NOT_EQUAL) reduce 529) ((T_IS_NOT_IDENTICAL) reduce 529) ((T_IS_SMALLER_OR_EQUAL) reduce 529) ((T_LOGICAL_AND) reduce 529) ((T_LOGICAL_OR) reduce 529) ((T_LOGICAL_XOR) reduce 529) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PIPE) reduce 529) ((T_POW) reduce 529) ((T_SL) reduce 529) ((T_SPACESHIP) reduce 529) ((T_SR) reduce 529) (("[") reduce 554) (("^") reduce 529) (("|") reduce 529)) 103 ((("%") reduce 528) (("(") reduce 557) (("*") reduce 528) (("+") reduce 528) (("-") reduce 528) ((".") reduce 528) (("/") reduce 528) ((";") reduce 528) (("<") reduce 528) ((">") reduce 528) (("?") reduce 528) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_BOOLEAN_AND) reduce 528) ((T_BOOLEAN_OR) reduce 528) ((T_COALESCE) reduce 528) ((T_INSTANCEOF) reduce 528) ((T_IS_EQUAL) reduce 528) ((T_IS_GREATER_OR_EQUAL) reduce 528) ((T_IS_IDENTICAL) reduce 528) ((T_IS_NOT_EQUAL) reduce 528) ((T_IS_NOT_IDENTICAL) reduce 528) ((T_IS_SMALLER_OR_EQUAL) reduce 528) ((T_LOGICAL_AND) reduce 528) ((T_LOGICAL_OR) reduce 528) ((T_LOGICAL_XOR) reduce 528) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PIPE) reduce 528) ((T_POW) reduce 528) ((T_SL) reduce 528) ((T_SPACESHIP) reduce 528) ((T_SR) reduce 528) (("[") reduce 550) (("^") reduce 528) (("|") reduce 528)) 104 ((("!") reduce 107) (("\"") reduce 107) (("$") reduce 107) (($) reduce 107) (("(") reduce 107) (("+") reduce 107) (("-") reduce 107) ((";") reduce 107) (("@") reduce 107) ((T_ABSTRACT) reduce 107) ((T_ARRAY) reduce 107) ((T_ARRAY_CAST) reduce 107) ((T_ATTRIBUTE) reduce 107) ((T_BOOL_CAST) reduce 107) ((T_BREAK) reduce 107) ((T_CLASS) reduce 107) ((T_CLASS_C) reduce 107) ((T_CLONE) reduce 107) ((T_CONST) reduce 107) ((T_CONSTANT_ENCAPSED_STRING) reduce 107) ((T_CONTINUE) reduce 107) ((T_DEC) reduce 107) ((T_DECLARE) reduce 107) ((T_DIR) reduce 107) ((T_DNUMBER) reduce 107) ((T_DO) reduce 107) ((T_DOUBLE_CAST) reduce 107) ((T_ECHO) reduce 107) ((T_EMPTY) reduce 107) ((T_ENUM) reduce 107) ((T_EVAL) reduce 107) ((T_EXIT) reduce 107) ((T_FILE) reduce 107) ((T_FINAL) reduce 107) ((T_FN) reduce 107) ((T_FOR) reduce 107) ((T_FOREACH) reduce 107) ((T_FUNCTION) reduce 107) ((T_FUNC_C) reduce 107) ((T_GLOBAL) reduce 107) ((T_GOTO) reduce 107) ((T_HALT_COMPILER) reduce 107) ((T_IF) reduce 107) ((T_INC) reduce 107) ((T_INCLUDE) reduce 107) ((T_INCLUDE_ONCE) reduce 107) ((T_INLINE_HTML) reduce 107) ((T_INTERFACE) reduce 107) ((T_INT_CAST) reduce 107) ((T_ISSET) reduce 107) ((T_LINE) reduce 107) ((T_LIST) reduce 107) ((T_LNUMBER) reduce 107) ((T_MATCH) reduce 107) ((T_METHOD_C) reduce 107) ((T_NAMESPACE) reduce 107) ((T_NAME_FULLY_QUALIFIED) reduce 107) ((T_NAME_QUALIFIED) reduce 107) ((T_NAME_RELATIVE) reduce 107) ((T_NEW) reduce 107) ((T_NS_C) reduce 107) ((T_OBJECT_CAST) reduce 107) ((T_PRINT) reduce 107) ((T_PROPERTY_C) reduce 107) ((T_READONLY) reduce 107) ((T_REQUIRE) reduce 107) ((T_REQUIRE_ONCE) reduce 107) ((T_RETURN) reduce 107) ((T_START_HEREDOC) reduce 107) ((T_STATIC) reduce 107) ((T_STRING) reduce 107) ((T_STRING_CAST) reduce 107) ((T_SWITCH) reduce 107) ((T_THROW) reduce 107) ((T_TRAIT) reduce 107) ((T_TRAIT_C) reduce 107) ((T_TRY) reduce 107) ((T_UNSET) reduce 107) ((T_UNSET_CAST) reduce 107) ((T_USE) reduce 107) ((T_VARIABLE) reduce 107) ((T_VOID_CAST) reduce 107) ((T_WHILE) reduce 107) ((T_YIELD) reduce 107) ((T_YIELD_FROM) reduce 107) (("[") reduce 107) (("`") reduce 107) (("{") reduce 107) (("~") reduce 107)) 105 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 106 ((("(") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift)) 107 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 553) ((T_OBJECT_OPERATOR) reduce 553) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 547) (("[") reduce 553)) 108 ((("(") reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_READONLY) reduce 496) ((T_STRING) reduce 496)) 109 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) ((";") reduce 563) (("<") reduce 563) (("=") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AND_EQUAL) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_COALESCE_EQUAL) reduce 563) ((T_CONCAT_EQUAL) reduce 563) ((T_DEC) reduce 563) ((T_DIV_EQUAL) reduce 563) ((T_INC) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_MINUS_EQUAL) reduce 563) ((T_MOD_EQUAL) reduce 563) ((T_MUL_EQUAL) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_OR_EQUAL) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_PLUS_EQUAL) reduce 563) ((T_POW) reduce 563) ((T_POW_EQUAL) reduce 563) ((T_SL) reduce 563) ((T_SL_EQUAL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) ((T_SR_EQUAL) reduce 563) ((T_XOR_EQUAL) reduce 563) (("[") reduce 563) (("^") reduce 563) (("|") reduce 563)) 110 ((("!") reduce 103) (("\"") reduce 103) (("$") reduce 103) (($) reduce 103) (("(") reduce 103) (("+") reduce 103) (("-") reduce 103) ((";") reduce 103) (("@") reduce 103) ((T_ABSTRACT) reduce 103) ((T_ARRAY) reduce 103) ((T_ARRAY_CAST) reduce 103) ((T_ATTRIBUTE) reduce 103) ((T_BOOL_CAST) reduce 103) ((T_BREAK) reduce 103) ((T_CLASS) reduce 103) ((T_CLASS_C) reduce 103) ((T_CLONE) reduce 103) ((T_CONST) reduce 103) ((T_CONSTANT_ENCAPSED_STRING) reduce 103) ((T_CONTINUE) reduce 103) ((T_DEC) reduce 103) ((T_DECLARE) reduce 103) ((T_DIR) reduce 103) ((T_DNUMBER) reduce 103) ((T_DO) reduce 103) ((T_DOUBLE_CAST) reduce 103) ((T_ECHO) reduce 103) ((T_EMPTY) reduce 103) ((T_ENUM) reduce 103) ((T_EVAL) reduce 103) ((T_EXIT) reduce 103) ((T_FILE) reduce 103) ((T_FINAL) reduce 103) ((T_FN) reduce 103) ((T_FOR) reduce 103) ((T_FOREACH) reduce 103) ((T_FUNCTION) reduce 103) ((T_FUNC_C) reduce 103) ((T_GLOBAL) reduce 103) ((T_GOTO) reduce 103) ((T_HALT_COMPILER) reduce 103) ((T_IF) reduce 103) ((T_INC) reduce 103) ((T_INCLUDE) reduce 103) ((T_INCLUDE_ONCE) reduce 103) ((T_INLINE_HTML) reduce 103) ((T_INTERFACE) reduce 103) ((T_INT_CAST) reduce 103) ((T_ISSET) reduce 103) ((T_LINE) reduce 103) ((T_LIST) reduce 103) ((T_LNUMBER) reduce 103) ((T_MATCH) reduce 103) ((T_METHOD_C) reduce 103) ((T_NAMESPACE) reduce 103) ((T_NAME_FULLY_QUALIFIED) reduce 103) ((T_NAME_QUALIFIED) reduce 103) ((T_NAME_RELATIVE) reduce 103) ((T_NEW) reduce 103) ((T_NS_C) reduce 103) ((T_OBJECT_CAST) reduce 103) ((T_PRINT) reduce 103) ((T_PROPERTY_C) reduce 103) ((T_READONLY) reduce 103) ((T_REQUIRE) reduce 103) ((T_REQUIRE_ONCE) reduce 103) ((T_RETURN) reduce 103) ((T_START_HEREDOC) reduce 103) ((T_STATIC) reduce 103) ((T_STRING) reduce 103) ((T_STRING_CAST) reduce 103) ((T_SWITCH) reduce 103) ((T_THROW) reduce 103) ((T_TRAIT) reduce 103) ((T_TRAIT_C) reduce 103) ((T_TRY) reduce 103) ((T_UNSET) reduce 103) ((T_UNSET_CAST) reduce 103) ((T_USE) reduce 103) ((T_VARIABLE) reduce 103) ((T_VOID_CAST) reduce 103) ((T_WHILE) reduce 103) ((T_YIELD) reduce 103) ((T_YIELD_FROM) reduce 103) (("[") reduce 103) (("`") reduce 103) (("{") reduce 103) (("~") reduce 103)) 111 ((("!") reduce 148) (("\"") reduce 148) (($) reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONST) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_EMPTY) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAMESPACE) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_USE) reduce 148) ((T_VARIABLE) reduce 148) ((T_VOID_CAST) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 112 ((("!") reduce 246) (("\"") reduce 246) (("$") reduce 246) (($) reduce 246) (("(") reduce 246) (("+") reduce 246) (("-") reduce 246) ((";") reduce 246) (("@") reduce 246) ((T_ABSTRACT) reduce 246) ((T_ARRAY) reduce 246) ((T_ARRAY_CAST) reduce 246) ((T_ATTRIBUTE) reduce 246) ((T_BOOL_CAST) reduce 246) ((T_BREAK) reduce 246) ((T_CLASS) reduce 246) ((T_CLASS_C) reduce 246) ((T_CLONE) reduce 246) ((T_CONST) reduce 246) ((T_CONSTANT_ENCAPSED_STRING) reduce 246) ((T_CONTINUE) reduce 246) ((T_DEC) reduce 246) ((T_DECLARE) reduce 246) ((T_DIR) reduce 246) ((T_DNUMBER) reduce 246) ((T_DO) reduce 246) ((T_DOUBLE_CAST) reduce 246) ((T_ECHO) reduce 246) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) reduce 246) ((T_ENUM) reduce 246) ((T_EVAL) reduce 246) ((T_EXIT) reduce 246) ((T_FILE) reduce 246) ((T_FINAL) reduce 246) ((T_FN) reduce 246) ((T_FOR) reduce 246) ((T_FOREACH) reduce 246) ((T_FUNCTION) reduce 246) ((T_FUNC_C) reduce 246) ((T_GLOBAL) reduce 246) ((T_GOTO) reduce 246) ((T_HALT_COMPILER) reduce 246) ((T_IF) reduce 246) ((T_INC) reduce 246) ((T_INCLUDE) reduce 246) ((T_INCLUDE_ONCE) reduce 246) ((T_INLINE_HTML) reduce 246) ((T_INTERFACE) reduce 246) ((T_INT_CAST) reduce 246) ((T_ISSET) reduce 246) ((T_LINE) reduce 246) ((T_LIST) reduce 246) ((T_LNUMBER) reduce 246) ((T_MATCH) reduce 246) ((T_METHOD_C) reduce 246) ((T_NAMESPACE) reduce 246) ((T_NAME_FULLY_QUALIFIED) reduce 246) ((T_NAME_QUALIFIED) reduce 246) ((T_NAME_RELATIVE) reduce 246) ((T_NEW) reduce 246) ((T_NS_C) reduce 246) ((T_OBJECT_CAST) reduce 246) ((T_PRINT) reduce 246) ((T_PROPERTY_C) reduce 246) ((T_READONLY) reduce 246) ((T_REQUIRE) reduce 246) ((T_REQUIRE_ONCE) reduce 246) ((T_RETURN) reduce 246) ((T_START_HEREDOC) reduce 246) ((T_STATIC) reduce 246) ((T_STRING) reduce 246) ((T_STRING_CAST) reduce 246) ((T_SWITCH) reduce 246) ((T_THROW) reduce 246) ((T_TRAIT) reduce 246) ((T_TRAIT_C) reduce 246) ((T_TRY) reduce 246) ((T_UNSET) reduce 246) ((T_UNSET_CAST) reduce 246) ((T_USE) reduce 246) ((T_VARIABLE) reduce 246) ((T_VOID_CAST) reduce 246) ((T_WHILE) reduce 246) ((T_YIELD) reduce 246) ((T_YIELD_FROM) reduce 246) (("[") reduce 246) (("`") reduce 246) (("{") reduce 246) (("~") reduce 246)) 113 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) ((";") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_PIPE) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("^") reduce 484) (("|") reduce 484)) 114 ((("!") reduce 106) (("\"") reduce 106) (("$") reduce 106) (($) reduce 106) (("(") reduce 106) (("+") reduce 106) (("-") reduce 106) ((";") reduce 106) (("@") reduce 106) ((T_ABSTRACT) reduce 106) ((T_ARRAY) reduce 106) ((T_ARRAY_CAST) reduce 106) ((T_ATTRIBUTE) reduce 106) ((T_BOOL_CAST) reduce 106) ((T_BREAK) reduce 106) ((T_CLASS) reduce 106) ((T_CLASS_C) reduce 106) ((T_CLONE) reduce 106) ((T_CONST) reduce 106) ((T_CONSTANT_ENCAPSED_STRING) reduce 106) ((T_CONTINUE) reduce 106) ((T_DEC) reduce 106) ((T_DECLARE) reduce 106) ((T_DIR) reduce 106) ((T_DNUMBER) reduce 106) ((T_DO) reduce 106) ((T_DOUBLE_CAST) reduce 106) ((T_ECHO) reduce 106) ((T_EMPTY) reduce 106) ((T_ENUM) reduce 106) ((T_EVAL) reduce 106) ((T_EXIT) reduce 106) ((T_FILE) reduce 106) ((T_FINAL) reduce 106) ((T_FN) reduce 106) ((T_FOR) reduce 106) ((T_FOREACH) reduce 106) ((T_FUNCTION) reduce 106) ((T_FUNC_C) reduce 106) ((T_GLOBAL) reduce 106) ((T_GOTO) reduce 106) ((T_HALT_COMPILER) reduce 106) ((T_IF) reduce 106) ((T_INC) reduce 106) ((T_INCLUDE) reduce 106) ((T_INCLUDE_ONCE) reduce 106) ((T_INLINE_HTML) reduce 106) ((T_INTERFACE) reduce 106) ((T_INT_CAST) reduce 106) ((T_ISSET) reduce 106) ((T_LINE) reduce 106) ((T_LIST) reduce 106) ((T_LNUMBER) reduce 106) ((T_MATCH) reduce 106) ((T_METHOD_C) reduce 106) ((T_NAMESPACE) reduce 106) ((T_NAME_FULLY_QUALIFIED) reduce 106) ((T_NAME_QUALIFIED) reduce 106) ((T_NAME_RELATIVE) reduce 106) ((T_NEW) reduce 106) ((T_NS_C) reduce 106) ((T_OBJECT_CAST) reduce 106) ((T_PRINT) reduce 106) ((T_PROPERTY_C) reduce 106) ((T_READONLY) reduce 106) ((T_REQUIRE) reduce 106) ((T_REQUIRE_ONCE) reduce 106) ((T_RETURN) reduce 106) ((T_START_HEREDOC) reduce 106) ((T_STATIC) reduce 106) ((T_STRING) reduce 106) ((T_STRING_CAST) reduce 106) ((T_SWITCH) reduce 106) ((T_THROW) reduce 106) ((T_TRAIT) reduce 106) ((T_TRAIT_C) reduce 106) ((T_TRY) reduce 106) ((T_UNSET) reduce 106) ((T_UNSET_CAST) reduce 106) ((T_USE) reduce 106) ((T_VARIABLE) reduce 106) ((T_VOID_CAST) reduce 106) ((T_WHILE) reduce 106) ((T_YIELD) reduce 106) ((T_YIELD_FROM) reduce 106) (("[") reduce 106) (("`") reduce 106) (("{") reduce 106) (("~") reduce 106)) 115 ((("%") reduce 466) (("*") reduce 466) (("+") reduce 466) (("-") reduce 466) ((".") reduce 466) (("/") reduce 466) ((";") reduce 466) (("<") reduce 466) ((">") reduce 466) (("?") reduce 466) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_BOOLEAN_AND) reduce 466) ((T_BOOLEAN_OR) reduce 466) ((T_COALESCE) reduce 466) ((T_INSTANCEOF) reduce 466) ((T_IS_EQUAL) reduce 466) ((T_IS_GREATER_OR_EQUAL) reduce 466) ((T_IS_IDENTICAL) reduce 466) ((T_IS_NOT_EQUAL) reduce 466) ((T_IS_NOT_IDENTICAL) reduce 466) ((T_IS_SMALLER_OR_EQUAL) reduce 466) ((T_LOGICAL_AND) reduce 466) ((T_LOGICAL_OR) reduce 466) ((T_LOGICAL_XOR) reduce 466) ((T_PIPE) reduce 466) ((T_POW) reduce 466) ((T_SL) reduce 466) ((T_SPACESHIP) reduce 466) ((T_SR) reduce 466) (("^") reduce 466) (("|") reduce 466)) 116 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) ((";") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_PIPE) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("^") reduce 488) (("|") reduce 488)) 117 ((("%") reduce 531) (("(") shift) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) ((";") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("|") reduce 531)) 118 ((("%") reduce 461) (("(") reduce 558) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) ((";") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PIPE) reduce 461) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") reduce 552) (("^") reduce 461) (("|") reduce 461)) 119 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) ((";") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_PIPE) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 120 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) ((";") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_PIPE) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("|") reduce 476)) 121 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) ((";") reduce 559) (("<") reduce 559) (("=") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AND_EQUAL) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_COALESCE_EQUAL) reduce 559) ((T_CONCAT_EQUAL) reduce 559) ((T_DEC) reduce 559) ((T_DIV_EQUAL) reduce 559) ((T_INC) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_MINUS_EQUAL) reduce 559) ((T_MOD_EQUAL) reduce 559) ((T_MUL_EQUAL) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_OR_EQUAL) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_PLUS_EQUAL) reduce 559) ((T_POW) reduce 559) ((T_POW_EQUAL) reduce 559) ((T_SL) reduce 559) ((T_SL_EQUAL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) ((T_SR_EQUAL) reduce 559) ((T_XOR_EQUAL) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 122 ((("!") reduce 110) (("\"") reduce 110) (("$") reduce 110) (($) reduce 110) (("(") reduce 110) (("+") reduce 110) (("-") reduce 110) ((";") reduce 110) (("@") reduce 110) ((T_ABSTRACT) reduce 110) ((T_ARRAY) reduce 110) ((T_ARRAY_CAST) reduce 110) ((T_ATTRIBUTE) reduce 110) ((T_BOOL_CAST) reduce 110) ((T_BREAK) reduce 110) ((T_CLASS) reduce 110) ((T_CLASS_C) reduce 110) ((T_CLONE) reduce 110) ((T_CONST) reduce 110) ((T_CONSTANT_ENCAPSED_STRING) reduce 110) ((T_CONTINUE) reduce 110) ((T_DEC) reduce 110) ((T_DECLARE) reduce 110) ((T_DIR) reduce 110) ((T_DNUMBER) reduce 110) ((T_DO) reduce 110) ((T_DOUBLE_CAST) reduce 110) ((T_ECHO) reduce 110) ((T_EMPTY) reduce 110) ((T_ENUM) reduce 110) ((T_EVAL) reduce 110) ((T_EXIT) reduce 110) ((T_FILE) reduce 110) ((T_FINAL) reduce 110) ((T_FN) reduce 110) ((T_FOR) reduce 110) ((T_FOREACH) reduce 110) ((T_FUNCTION) reduce 110) ((T_FUNC_C) reduce 110) ((T_GLOBAL) reduce 110) ((T_GOTO) reduce 110) ((T_HALT_COMPILER) reduce 110) ((T_IF) reduce 110) ((T_INC) reduce 110) ((T_INCLUDE) reduce 110) ((T_INCLUDE_ONCE) reduce 110) ((T_INLINE_HTML) reduce 110) ((T_INTERFACE) reduce 110) ((T_INT_CAST) reduce 110) ((T_ISSET) reduce 110) ((T_LINE) reduce 110) ((T_LIST) reduce 110) ((T_LNUMBER) reduce 110) ((T_MATCH) reduce 110) ((T_METHOD_C) reduce 110) ((T_NAMESPACE) reduce 110) ((T_NAME_FULLY_QUALIFIED) reduce 110) ((T_NAME_QUALIFIED) reduce 110) ((T_NAME_RELATIVE) reduce 110) ((T_NEW) reduce 110) ((T_NS_C) reduce 110) ((T_OBJECT_CAST) reduce 110) ((T_PRINT) reduce 110) ((T_PROPERTY_C) reduce 110) ((T_READONLY) reduce 110) ((T_REQUIRE) reduce 110) ((T_REQUIRE_ONCE) reduce 110) ((T_RETURN) reduce 110) ((T_START_HEREDOC) reduce 110) ((T_STATIC) reduce 110) ((T_STRING) reduce 110) ((T_STRING_CAST) reduce 110) ((T_SWITCH) reduce 110) ((T_THROW) reduce 110) ((T_TRAIT) reduce 110) ((T_TRAIT_C) reduce 110) ((T_TRY) reduce 110) ((T_UNSET) reduce 110) ((T_UNSET_CAST) reduce 110) ((T_USE) reduce 110) ((T_VARIABLE) reduce 110) ((T_VOID_CAST) reduce 110) ((T_WHILE) reduce 110) ((T_YIELD) reduce 110) ((T_YIELD_FROM) reduce 110) (("[") reduce 110) (("`") reduce 110) (("{") reduce 110) (("~") reduce 110)) 123 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((";") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 124 ((("!") reduce 84) (("\"") reduce 84) (($) reduce 84) (("$") reduce 84) (("(") reduce 84) (("+") reduce 84) (("-") reduce 84) ((";") reduce 84) (("@") reduce 84) ((T_ABSTRACT) reduce 84) ((T_ARRAY) reduce 84) ((T_ARRAY_CAST) reduce 84) ((T_ATTRIBUTE) reduce 84) ((T_BOOL_CAST) reduce 84) ((T_BREAK) reduce 84) ((T_CLASS) reduce 84) ((T_CLASS_C) reduce 84) ((T_CLONE) reduce 84) ((T_CONST) reduce 84) ((T_CONSTANT_ENCAPSED_STRING) reduce 84) ((T_CONTINUE) reduce 84) ((T_DEC) reduce 84) ((T_DECLARE) reduce 84) ((T_DIR) reduce 84) ((T_DNUMBER) reduce 84) ((T_DO) reduce 84) ((T_DOUBLE_CAST) reduce 84) ((T_ECHO) reduce 84) ((T_EMPTY) reduce 84) ((T_ENUM) reduce 84) ((T_EVAL) reduce 84) ((T_EXIT) reduce 84) ((T_FILE) reduce 84) ((T_FINAL) reduce 84) ((T_FN) reduce 84) ((T_FOR) reduce 84) ((T_FOREACH) reduce 84) ((T_FUNCTION) reduce 84) ((T_FUNC_C) reduce 84) ((T_GLOBAL) reduce 84) ((T_GOTO) reduce 84) ((T_HALT_COMPILER) reduce 84) ((T_IF) reduce 84) ((T_INC) reduce 84) ((T_INCLUDE) reduce 84) ((T_INCLUDE_ONCE) reduce 84) ((T_INLINE_HTML) reduce 84) ((T_INTERFACE) reduce 84) ((T_INT_CAST) reduce 84) ((T_ISSET) reduce 84) ((T_LINE) reduce 84) ((T_LIST) reduce 84) ((T_LNUMBER) reduce 84) ((T_MATCH) reduce 84) ((T_METHOD_C) reduce 84) ((T_NAMESPACE) reduce 84) ((T_NAME_FULLY_QUALIFIED) reduce 84) ((T_NAME_QUALIFIED) reduce 84) ((T_NAME_RELATIVE) reduce 84) ((T_NEW) reduce 84) ((T_NS_C) reduce 84) ((T_OBJECT_CAST) reduce 84) ((T_PRINT) reduce 84) ((T_PROPERTY_C) reduce 84) ((T_READONLY) reduce 84) ((T_REQUIRE) reduce 84) ((T_REQUIRE_ONCE) reduce 84) ((T_RETURN) reduce 84) ((T_START_HEREDOC) reduce 84) ((T_STATIC) reduce 84) ((T_STRING) reduce 84) ((T_STRING_CAST) reduce 84) ((T_SWITCH) reduce 84) ((T_THROW) reduce 84) ((T_TRAIT) reduce 84) ((T_TRAIT_C) reduce 84) ((T_TRY) reduce 84) ((T_UNSET) reduce 84) ((T_UNSET_CAST) reduce 84) ((T_USE) reduce 84) ((T_VARIABLE) reduce 84) ((T_VOID_CAST) reduce 84) ((T_WHILE) reduce 84) ((T_YIELD) reduce 84) ((T_YIELD_FROM) reduce 84) (("[") reduce 84) (("`") reduce 84) (("{") reduce 84) (("~") reduce 84)) 125 ((("!") reduce 105) (("\"") reduce 105) (("$") reduce 105) (($) reduce 105) (("(") reduce 105) (("+") reduce 105) (("-") reduce 105) ((";") reduce 105) (("@") reduce 105) ((T_ABSTRACT) reduce 105) ((T_ARRAY) reduce 105) ((T_ARRAY_CAST) reduce 105) ((T_ATTRIBUTE) reduce 105) ((T_BOOL_CAST) reduce 105) ((T_BREAK) reduce 105) ((T_CLASS) reduce 105) ((T_CLASS_C) reduce 105) ((T_CLONE) reduce 105) ((T_CONST) reduce 105) ((T_CONSTANT_ENCAPSED_STRING) reduce 105) ((T_CONTINUE) reduce 105) ((T_DEC) reduce 105) ((T_DECLARE) reduce 105) ((T_DIR) reduce 105) ((T_DNUMBER) reduce 105) ((T_DO) reduce 105) ((T_DOUBLE_CAST) reduce 105) ((T_ECHO) reduce 105) ((T_EMPTY) reduce 105) ((T_ENUM) reduce 105) ((T_EVAL) reduce 105) ((T_EXIT) reduce 105) ((T_FILE) reduce 105) ((T_FINAL) reduce 105) ((T_FN) reduce 105) ((T_FOR) reduce 105) ((T_FOREACH) reduce 105) ((T_FUNCTION) reduce 105) ((T_FUNC_C) reduce 105) ((T_GLOBAL) reduce 105) ((T_GOTO) reduce 105) ((T_HALT_COMPILER) reduce 105) ((T_IF) reduce 105) ((T_INC) reduce 105) ((T_INCLUDE) reduce 105) ((T_INCLUDE_ONCE) reduce 105) ((T_INLINE_HTML) reduce 105) ((T_INTERFACE) reduce 105) ((T_INT_CAST) reduce 105) ((T_ISSET) reduce 105) ((T_LINE) reduce 105) ((T_LIST) reduce 105) ((T_LNUMBER) reduce 105) ((T_MATCH) reduce 105) ((T_METHOD_C) reduce 105) ((T_NAMESPACE) reduce 105) ((T_NAME_FULLY_QUALIFIED) reduce 105) ((T_NAME_QUALIFIED) reduce 105) ((T_NAME_RELATIVE) reduce 105) ((T_NEW) reduce 105) ((T_NS_C) reduce 105) ((T_OBJECT_CAST) reduce 105) ((T_PRINT) reduce 105) ((T_PROPERTY_C) reduce 105) ((T_READONLY) reduce 105) ((T_REQUIRE) reduce 105) ((T_REQUIRE_ONCE) reduce 105) ((T_RETURN) reduce 105) ((T_START_HEREDOC) reduce 105) ((T_STATIC) reduce 105) ((T_STRING) reduce 105) ((T_STRING_CAST) reduce 105) ((T_SWITCH) reduce 105) ((T_THROW) reduce 105) ((T_TRAIT) reduce 105) ((T_TRAIT_C) reduce 105) ((T_TRY) reduce 105) ((T_UNSET) reduce 105) ((T_UNSET_CAST) reduce 105) ((T_USE) reduce 105) ((T_VARIABLE) reduce 105) ((T_VOID_CAST) reduce 105) ((T_WHILE) reduce 105) ((T_YIELD) reduce 105) ((T_YIELD_FROM) reduce 105) (("[") reduce 105) (("`") reduce 105) (("{") reduce 105) (("~") reduce 105)) 126 ((("%") reduce 403) (("*") reduce 403) (("+") reduce 403) (("-") reduce 403) ((".") reduce 403) (("/") reduce 403) ((";") reduce 403) (("<") reduce 403) (("=") shift) ((">") reduce 403) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_INC) shift) ((T_INSTANCEOF) reduce 403) ((T_IS_EQUAL) reduce 403) ((T_IS_GREATER_OR_EQUAL) reduce 403) ((T_IS_IDENTICAL) reduce 403) ((T_IS_NOT_EQUAL) reduce 403) ((T_IS_NOT_IDENTICAL) reduce 403) ((T_IS_SMALLER_OR_EQUAL) reduce 403) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 403) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 403) ((T_POW_EQUAL) shift) ((T_SL) reduce 403) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 403) ((T_SR) reduce 403) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 548) (("^") reduce 403) (("|") reduce 403)) 128 ((("!") reduce 142) (("\"") reduce 142) (("$") reduce 142) (("(") reduce 142) (("+") reduce 142) (("-") reduce 142) ((";") reduce 142) (("@") reduce 142) ((T_ABSTRACT) reduce 142) ((T_ARRAY) reduce 142) ((T_ARRAY_CAST) reduce 142) ((T_ATTRIBUTE) reduce 142) ((T_BOOL_CAST) reduce 142) ((T_BREAK) reduce 142) ((T_CLASS) reduce 142) ((T_CLASS_C) reduce 142) ((T_CLONE) reduce 142) ((T_CONSTANT_ENCAPSED_STRING) reduce 142) ((T_CONTINUE) reduce 142) ((T_DEC) reduce 142) ((T_DECLARE) reduce 142) ((T_DIR) reduce 142) ((T_DNUMBER) reduce 142) ((T_DO) reduce 142) ((T_DOUBLE_CAST) reduce 142) ((T_ECHO) reduce 142) ((T_EMPTY) reduce 142) ((T_ENUM) reduce 142) ((T_EVAL) reduce 142) ((T_EXIT) reduce 142) ((T_FILE) reduce 142) ((T_FINAL) reduce 142) ((T_FN) reduce 142) ((T_FOR) reduce 142) ((T_FOREACH) reduce 142) ((T_FUNCTION) reduce 142) ((T_FUNC_C) reduce 142) ((T_GLOBAL) reduce 142) ((T_GOTO) reduce 142) ((T_HALT_COMPILER) reduce 142) ((T_IF) reduce 142) ((T_INC) reduce 142) ((T_INCLUDE) reduce 142) ((T_INCLUDE_ONCE) reduce 142) ((T_INLINE_HTML) reduce 142) ((T_INTERFACE) reduce 142) ((T_INT_CAST) reduce 142) ((T_ISSET) reduce 142) ((T_LINE) reduce 142) ((T_LIST) reduce 142) ((T_LNUMBER) reduce 142) ((T_MATCH) reduce 142) ((T_METHOD_C) reduce 142) ((T_NAME_FULLY_QUALIFIED) reduce 142) ((T_NAME_QUALIFIED) reduce 142) ((T_NAME_RELATIVE) reduce 142) ((T_NEW) reduce 142) ((T_NS_C) reduce 142) ((T_OBJECT_CAST) reduce 142) ((T_PRINT) reduce 142) ((T_PROPERTY_C) reduce 142) ((T_READONLY) reduce 142) ((T_REQUIRE) reduce 142) ((T_REQUIRE_ONCE) reduce 142) ((T_RETURN) reduce 142) ((T_START_HEREDOC) reduce 142) ((T_STATIC) reduce 142) ((T_STRING) reduce 142) ((T_STRING_CAST) reduce 142) ((T_SWITCH) reduce 142) ((T_THROW) reduce 142) ((T_TRAIT) reduce 142) ((T_TRAIT_C) reduce 142) ((T_TRY) reduce 142) ((T_UNSET) reduce 142) ((T_UNSET_CAST) reduce 142) ((T_VARIABLE) reduce 142) ((T_VOID_CAST) reduce 142) ((T_WHILE) reduce 142) ((T_YIELD) reduce 142) ((T_YIELD_FROM) reduce 142) (("[") reduce 142) (("`") reduce 142) (("{") reduce 142) (("}") reduce 142) (("~") reduce 142)) 131 ((("(") reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492)) 133 (((T_FN) shift) ((T_FUNCTION) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509)) 134 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92)) 135 (((T_ATTRIBUTE) reduce 101) ((T_FN) reduce 101) ((T_FUNCTION) reduce 101) ((T_STATIC) reduce 101)) 136 (((T_ATTRIBUTE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_STATIC) shift)) 137 ((("%") reduce 448) (("*") reduce 448) (("+") reduce 448) (("-") reduce 448) ((".") reduce 448) (("/") reduce 448) ((";") reduce 448) (("<") reduce 448) ((">") reduce 448) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_BOOLEAN_AND) reduce 448) ((T_BOOLEAN_OR) reduce 448) ((T_COALESCE) reduce 448) ((T_INSTANCEOF) reduce 448) ((T_IS_EQUAL) reduce 448) ((T_IS_GREATER_OR_EQUAL) reduce 448) ((T_IS_IDENTICAL) reduce 448) ((T_IS_NOT_EQUAL) reduce 448) ((T_IS_NOT_IDENTICAL) reduce 448) ((T_IS_SMALLER_OR_EQUAL) reduce 448) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_PIPE) reduce 448) ((T_POW) shift) ((T_SL) reduce 448) ((T_SPACESHIP) reduce 448) ((T_SR) reduce 448) (("^") reduce 448) (("|") reduce 448)) 139 ((("(") reduce 80)) 140 ((("(") reduce 81)) 141 ((("(") reduce 497)) 142 ((("(") reduce 493)) 144 ((("(") reduce 258) ((")") reduce 253) (("?") reduce 258) ((T_ABSTRACT) shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 258) ((T_ARRAY) reduce 258) ((T_ATTRIBUTE) shift) ((T_CALLABLE) reduce 258) ((T_ELLIPSIS) reduce 258) ((T_FINAL) shift) ((T_NAME_FULLY_QUALIFIED) reduce 258) ((T_NAME_QUALIFIED) reduce 258) ((T_NAME_RELATIVE) reduce 258) ((T_PRIVATE) shift) ((T_PRIVATE_SET) shift) ((T_PROTECTED) shift) ((T_PROTECTED_SET) shift) ((T_PUBLIC) shift) ((T_PUBLIC_SET) shift) ((T_READONLY) shift) ((T_STATIC) shift) ((T_STRING) reduce 258) ((T_VARIABLE) reduce 258)) 145 ((("(") reduce 359) (("?") reduce 359) ((T_ABSTRACT) reduce 359) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 359) ((T_ARRAY) reduce 359) ((T_CALLABLE) reduce 359) ((T_ELLIPSIS) reduce 359) ((T_FINAL) reduce 359) ((T_NAME_FULLY_QUALIFIED) reduce 359) ((T_NAME_QUALIFIED) reduce 359) ((T_NAME_RELATIVE) reduce 359) ((T_PRIVATE) reduce 359) ((T_PRIVATE_SET) reduce 359) ((T_PROTECTED) reduce 359) ((T_PROTECTED_SET) reduce 359) ((T_PUBLIC) reduce 359) ((T_PUBLIC_SET) reduce 359) ((T_READONLY) reduce 359) ((T_STATIC) reduce 359) ((T_STRING) reduce 359) ((T_VARIABLE) reduce 359)) 147 ((("(") reduce 360) (("?") reduce 360) ((T_ABSTRACT) reduce 360) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_ARRAY) reduce 360) ((T_CALLABLE) reduce 360) ((T_ELLIPSIS) reduce 360) ((T_FINAL) reduce 360) ((T_NAME_FULLY_QUALIFIED) reduce 360) ((T_NAME_QUALIFIED) reduce 360) ((T_NAME_RELATIVE) reduce 360) ((T_PRIVATE) reduce 360) ((T_PRIVATE_SET) reduce 360) ((T_PROTECTED) reduce 360) ((T_PROTECTED_SET) reduce 360) ((T_PUBLIC) reduce 360) ((T_PUBLIC_SET) reduce 360) ((T_READONLY) reduce 360) ((T_STATIC) reduce 360) ((T_STRING) reduce 360) ((T_VARIABLE) reduce 360)) 148 ((("(") reduce 354) (("?") reduce 354) ((T_ABSTRACT) reduce 354) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_ARRAY) reduce 354) ((T_CALLABLE) reduce 354) ((T_ELLIPSIS) reduce 354) ((T_FINAL) reduce 354) ((T_NAME_FULLY_QUALIFIED) reduce 354) ((T_NAME_QUALIFIED) reduce 354) ((T_NAME_RELATIVE) reduce 354) ((T_PRIVATE) reduce 354) ((T_PRIVATE_SET) reduce 354) ((T_PROTECTED) reduce 354) ((T_PROTECTED_SET) reduce 354) ((T_PUBLIC) reduce 354) ((T_PUBLIC_SET) reduce 354) ((T_READONLY) reduce 354) ((T_STATIC) reduce 354) ((T_STRING) reduce 354) ((T_VARIABLE) reduce 354)) 149 ((("(") reduce 357) (("?") reduce 357) ((T_ABSTRACT) reduce 357) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 357) ((T_ARRAY) reduce 357) ((T_CALLABLE) reduce 357) ((T_ELLIPSIS) reduce 357) ((T_FINAL) reduce 357) ((T_NAME_FULLY_QUALIFIED) reduce 357) ((T_NAME_QUALIFIED) reduce 357) ((T_NAME_RELATIVE) reduce 357) ((T_PRIVATE) reduce 357) ((T_PRIVATE_SET) reduce 357) ((T_PROTECTED) reduce 357) ((T_PROTECTED_SET) reduce 357) ((T_PUBLIC) reduce 357) ((T_PUBLIC_SET) reduce 357) ((T_READONLY) reduce 357) ((T_STATIC) reduce 357) ((T_STRING) reduce 357) ((T_VARIABLE) reduce 357)) 150 ((("(") reduce 353) (("?") reduce 353) ((T_ABSTRACT) reduce 353) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_ARRAY) reduce 353) ((T_CALLABLE) reduce 353) ((T_ELLIPSIS) reduce 353) ((T_FINAL) reduce 353) ((T_NAME_FULLY_QUALIFIED) reduce 353) ((T_NAME_QUALIFIED) reduce 353) ((T_NAME_RELATIVE) reduce 353) ((T_PRIVATE) reduce 353) ((T_PRIVATE_SET) reduce 353) ((T_PROTECTED) reduce 353) ((T_PROTECTED_SET) reduce 353) ((T_PUBLIC) reduce 353) ((T_PUBLIC_SET) reduce 353) ((T_READONLY) reduce 353) ((T_STATIC) reduce 353) ((T_STRING) reduce 353) ((T_VARIABLE) reduce 353)) 151 ((("(") reduce 356) (("?") reduce 356) ((T_ABSTRACT) reduce 356) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_ARRAY) reduce 356) ((T_CALLABLE) reduce 356) ((T_ELLIPSIS) reduce 356) ((T_FINAL) reduce 356) ((T_NAME_FULLY_QUALIFIED) reduce 356) ((T_NAME_QUALIFIED) reduce 356) ((T_NAME_RELATIVE) reduce 356) ((T_PRIVATE) reduce 356) ((T_PRIVATE_SET) reduce 356) ((T_PROTECTED) reduce 356) ((T_PROTECTED_SET) reduce 356) ((T_PUBLIC) reduce 356) ((T_PUBLIC_SET) reduce 356) ((T_READONLY) reduce 356) ((T_STATIC) reduce 356) ((T_STRING) reduce 356) ((T_VARIABLE) reduce 356)) 152 ((("(") reduce 352) (("?") reduce 352) ((T_ABSTRACT) reduce 352) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_ARRAY) reduce 352) ((T_CALLABLE) reduce 352) ((T_ELLIPSIS) reduce 352) ((T_FINAL) reduce 352) ((T_NAME_FULLY_QUALIFIED) reduce 352) ((T_NAME_QUALIFIED) reduce 352) ((T_NAME_RELATIVE) reduce 352) ((T_PRIVATE) reduce 352) ((T_PRIVATE_SET) reduce 352) ((T_PROTECTED) reduce 352) ((T_PROTECTED_SET) reduce 352) ((T_PUBLIC) reduce 352) ((T_PUBLIC_SET) reduce 352) ((T_READONLY) reduce 352) ((T_STATIC) reduce 352) ((T_STRING) reduce 352) ((T_VARIABLE) reduce 352)) 153 ((("(") reduce 355) (("?") reduce 355) ((T_ABSTRACT) reduce 355) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_ARRAY) reduce 355) ((T_CALLABLE) reduce 355) ((T_ELLIPSIS) reduce 355) ((T_FINAL) reduce 355) ((T_NAME_FULLY_QUALIFIED) reduce 355) ((T_NAME_QUALIFIED) reduce 355) ((T_NAME_RELATIVE) reduce 355) ((T_PRIVATE) reduce 355) ((T_PRIVATE_SET) reduce 355) ((T_PROTECTED) reduce 355) ((T_PROTECTED_SET) reduce 355) ((T_PUBLIC) reduce 355) ((T_PUBLIC_SET) reduce 355) ((T_READONLY) reduce 355) ((T_STATIC) reduce 355) ((T_STRING) reduce 355) ((T_VARIABLE) reduce 355)) 154 ((("(") reduce 361) (("?") reduce 361) ((T_ABSTRACT) reduce 361) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_ARRAY) reduce 361) ((T_CALLABLE) reduce 361) ((T_ELLIPSIS) reduce 361) ((T_FINAL) reduce 361) ((T_NAME_FULLY_QUALIFIED) reduce 361) ((T_NAME_QUALIFIED) reduce 361) ((T_NAME_RELATIVE) reduce 361) ((T_PRIVATE) reduce 361) ((T_PRIVATE_SET) reduce 361) ((T_PROTECTED) reduce 361) ((T_PROTECTED_SET) reduce 361) ((T_PUBLIC) reduce 361) ((T_PUBLIC_SET) reduce 361) ((T_READONLY) reduce 361) ((T_STATIC) reduce 361) ((T_STRING) reduce 361) ((T_VARIABLE) reduce 361)) 155 ((("(") reduce 358) (("?") reduce 358) ((T_ABSTRACT) reduce 358) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 358) ((T_ARRAY) reduce 358) ((T_CALLABLE) reduce 358) ((T_ELLIPSIS) reduce 358) ((T_FINAL) reduce 358) ((T_NAME_FULLY_QUALIFIED) reduce 358) ((T_NAME_QUALIFIED) reduce 358) ((T_NAME_RELATIVE) reduce 358) ((T_PRIVATE) reduce 358) ((T_PRIVATE_SET) reduce 358) ((T_PROTECTED) reduce 358) ((T_PROTECTED_SET) reduce 358) ((T_PUBLIC) reduce 358) ((T_PUBLIC_SET) reduce 358) ((T_READONLY) reduce 358) ((T_STATIC) reduce 358) ((T_STRING) reduce 358) ((T_VARIABLE) reduce 358)) 156 ((("(") reduce 101) (("?") reduce 101) ((T_ABSTRACT) reduce 101) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 101) ((T_ARRAY) reduce 101) ((T_ATTRIBUTE) reduce 101) ((T_CALLABLE) reduce 101) ((T_ELLIPSIS) reduce 101) ((T_FINAL) reduce 101) ((T_NAME_FULLY_QUALIFIED) reduce 101) ((T_NAME_QUALIFIED) reduce 101) ((T_NAME_RELATIVE) reduce 101) ((T_PRIVATE) reduce 101) ((T_PRIVATE_SET) reduce 101) ((T_PROTECTED) reduce 101) ((T_PROTECTED_SET) reduce 101) ((T_PUBLIC) reduce 101) ((T_PUBLIC_SET) reduce 101) ((T_READONLY) reduce 101) ((T_STATIC) reduce 101) ((T_STRING) reduce 101) ((T_VARIABLE) reduce 101)) 157 (((")") reduce 254) ((",") reduce 254)) 158 ((("(") reduce 258) (("?") reduce 258) ((T_ABSTRACT) shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 258) ((T_ARRAY) reduce 258) ((T_ATTRIBUTE) shift) ((T_CALLABLE) reduce 258) ((T_ELLIPSIS) reduce 258) ((T_FINAL) shift) ((T_NAME_FULLY_QUALIFIED) reduce 258) ((T_NAME_QUALIFIED) reduce 258) ((T_NAME_RELATIVE) reduce 258) ((T_PRIVATE) shift) ((T_PRIVATE_SET) shift) ((T_PROTECTED) shift) ((T_PROTECTED_SET) shift) ((T_PUBLIC) shift) ((T_PUBLIC_SET) shift) ((T_READONLY) shift) ((T_STATIC) shift) ((T_STRING) reduce 258) ((T_VARIABLE) reduce 258)) 159 ((("(") reduce 350) (("?") reduce 350) ((T_ABSTRACT) reduce 350) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 350) ((T_ARRAY) reduce 350) ((T_CALLABLE) reduce 350) ((T_ELLIPSIS) reduce 350) ((T_FINAL) reduce 350) ((T_NAME_FULLY_QUALIFIED) reduce 350) ((T_NAME_QUALIFIED) reduce 350) ((T_NAME_RELATIVE) reduce 350) ((T_PRIVATE) reduce 350) ((T_PRIVATE_SET) reduce 350) ((T_PROTECTED) reduce 350) ((T_PROTECTED_SET) reduce 350) ((T_PUBLIC) reduce 350) ((T_PUBLIC_SET) reduce 350) ((T_READONLY) reduce 350) ((T_STATIC) reduce 350) ((T_STRING) reduce 350) ((T_VARIABLE) reduce 350)) 160 ((("(") reduce 259) (("?") reduce 259) ((T_ABSTRACT) shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 259) ((T_ARRAY) reduce 259) ((T_CALLABLE) reduce 259) ((T_ELLIPSIS) reduce 259) ((T_FINAL) shift) ((T_NAME_FULLY_QUALIFIED) reduce 259) ((T_NAME_QUALIFIED) reduce 259) ((T_NAME_RELATIVE) reduce 259) ((T_PRIVATE) shift) ((T_PRIVATE_SET) shift) ((T_PROTECTED) shift) ((T_PROTECTED_SET) shift) ((T_PUBLIC) shift) ((T_PUBLIC_SET) shift) ((T_READONLY) shift) ((T_STATIC) shift) ((T_STRING) reduce 259) ((T_VARIABLE) reduce 259)) 161 (((")") reduce 125) ((",") shift)) 162 ((("(") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 262) ((T_ARRAY) shift) ((T_CALLABLE) shift) ((T_ELLIPSIS) reduce 262) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_STRING) shift) ((T_VARIABLE) reduce 262)) 163 (((")") reduce 257) ((",") reduce 257)) 164 (((")") shift)) 165 (((":") reduce 498) ((T_USE) shift) (("{") reduce 498)) 167 (((":") shift) (("{") reduce 289)) 168 ((("(") shift) (("?") shift) ((T_ARRAY) shift) ((T_CALLABLE) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_STATIC) shift) ((T_STRING) shift)) 169 ((("{") reduce 494)) 172 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_VOID_CAST) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("}") shift) (("~") shift)) 173 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_EMPTY) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_VOID_CAST) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("}") reduce 166) (("~") reduce 166)) 188 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_EMPTY) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_VOID_CAST) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("}") reduce 160) (("~") reduce 160)) 199 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_EMPTY) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_VOID_CAST) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("}") reduce 149) (("~") reduce 149)) 201 (((T_ABSTRACT) reduce 101) ((T_ATTRIBUTE) reduce 101) ((T_CLASS) reduce 101) ((T_ENUM) reduce 101) ((T_FINAL) reduce 101) ((T_FN) reduce 101) ((T_FUNCTION) reduce 101) ((T_INTERFACE) reduce 101) ((T_READONLY) reduce 101) ((T_STATIC) reduce 101) ((T_TRAIT) reduce 101)) 202 ((("!") reduce 144) (("\"") reduce 144) (("$") reduce 144) (("(") reduce 144) (("+") reduce 144) (("-") reduce 144) ((";") reduce 144) (("@") reduce 144) ((T_ABSTRACT) reduce 144) ((T_ARRAY) reduce 144) ((T_ARRAY_CAST) reduce 144) ((T_ATTRIBUTE) reduce 144) ((T_BOOL_CAST) reduce 144) ((T_BREAK) reduce 144) ((T_CLASS) reduce 144) ((T_CLASS_C) reduce 144) ((T_CLONE) reduce 144) ((T_CONSTANT_ENCAPSED_STRING) reduce 144) ((T_CONTINUE) reduce 144) ((T_DEC) reduce 144) ((T_DECLARE) reduce 144) ((T_DIR) reduce 144) ((T_DNUMBER) reduce 144) ((T_DO) reduce 144) ((T_DOUBLE_CAST) reduce 144) ((T_ECHO) reduce 144) ((T_EMPTY) reduce 144) ((T_ENUM) reduce 144) ((T_EVAL) reduce 144) ((T_EXIT) reduce 144) ((T_FILE) reduce 144) ((T_FINAL) reduce 144) ((T_FN) reduce 144) ((T_FOR) reduce 144) ((T_FOREACH) reduce 144) ((T_FUNCTION) reduce 144) ((T_FUNC_C) reduce 144) ((T_GLOBAL) reduce 144) ((T_GOTO) reduce 144) ((T_HALT_COMPILER) reduce 144) ((T_IF) reduce 144) ((T_INC) reduce 144) ((T_INCLUDE) reduce 144) ((T_INCLUDE_ONCE) reduce 144) ((T_INLINE_HTML) reduce 144) ((T_INTERFACE) reduce 144) ((T_INT_CAST) reduce 144) ((T_ISSET) reduce 144) ((T_LINE) reduce 144) ((T_LIST) reduce 144) ((T_LNUMBER) reduce 144) ((T_MATCH) reduce 144) ((T_METHOD_C) reduce 144) ((T_NAME_FULLY_QUALIFIED) reduce 144) ((T_NAME_QUALIFIED) reduce 144) ((T_NAME_RELATIVE) reduce 144) ((T_NEW) reduce 144) ((T_NS_C) reduce 144) ((T_OBJECT_CAST) reduce 144) ((T_PRINT) reduce 144) ((T_PROPERTY_C) reduce 144) ((T_READONLY) reduce 144) ((T_REQUIRE) reduce 144) ((T_REQUIRE_ONCE) reduce 144) ((T_RETURN) reduce 144) ((T_START_HEREDOC) reduce 144) ((T_STATIC) reduce 144) ((T_STRING) reduce 144) ((T_STRING_CAST) reduce 144) ((T_SWITCH) reduce 144) ((T_THROW) reduce 144) ((T_TRAIT) reduce 144) ((T_TRAIT_C) reduce 144) ((T_TRY) reduce 144) ((T_UNSET) reduce 144) ((T_UNSET_CAST) reduce 144) ((T_VARIABLE) reduce 144) ((T_VOID_CAST) reduce 144) ((T_WHILE) reduce 144) ((T_YIELD) reduce 144) ((T_YIELD_FROM) reduce 144) (("[") reduce 144) (("`") reduce 144) (("{") reduce 144) (("}") reduce 144) (("~") reduce 144)) 203 (((T_ABSTRACT) shift) ((T_ATTRIBUTE) shift) ((T_CLASS) shift) ((T_ENUM) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_INTERFACE) shift) ((T_READONLY) shift) ((T_STATIC) shift) ((T_TRAIT) shift)) 204 ((("!") reduce 104) (("\"") reduce 104) (("$") reduce 104) (("(") reduce 104) (("+") reduce 104) (("-") reduce 104) ((";") reduce 104) (("@") reduce 104) ((T_ABSTRACT) reduce 104) ((T_ARRAY) reduce 104) ((T_ARRAY_CAST) reduce 104) ((T_ATTRIBUTE) reduce 104) ((T_BOOL_CAST) reduce 104) ((T_BREAK) reduce 104) ((T_CLASS) reduce 104) ((T_CLASS_C) reduce 104) ((T_CLONE) reduce 104) ((T_CONSTANT_ENCAPSED_STRING) reduce 104) ((T_CONTINUE) reduce 104) ((T_DEC) reduce 104) ((T_DECLARE) reduce 104) ((T_DIR) reduce 104) ((T_DNUMBER) reduce 104) ((T_DO) reduce 104) ((T_DOUBLE_CAST) reduce 104) ((T_ECHO) reduce 104) ((T_EMPTY) reduce 104) ((T_ENUM) reduce 104) ((T_EVAL) reduce 104) ((T_EXIT) reduce 104) ((T_FILE) reduce 104) ((T_FINAL) reduce 104) ((T_FN) reduce 104) ((T_FOR) reduce 104) ((T_FOREACH) reduce 104) ((T_FUNCTION) reduce 104) ((T_FUNC_C) reduce 104) ((T_GLOBAL) reduce 104) ((T_GOTO) reduce 104) ((T_HALT_COMPILER) reduce 104) ((T_IF) reduce 104) ((T_INC) reduce 104) ((T_INCLUDE) reduce 104) ((T_INCLUDE_ONCE) reduce 104) ((T_INLINE_HTML) reduce 104) ((T_INTERFACE) reduce 104) ((T_INT_CAST) reduce 104) ((T_ISSET) reduce 104) ((T_LINE) reduce 104) ((T_LIST) reduce 104) ((T_LNUMBER) reduce 104) ((T_MATCH) reduce 104) ((T_METHOD_C) reduce 104) ((T_NAME_FULLY_QUALIFIED) reduce 104) ((T_NAME_QUALIFIED) reduce 104) ((T_NAME_RELATIVE) reduce 104) ((T_NEW) reduce 104) ((T_NS_C) reduce 104) ((T_OBJECT_CAST) reduce 104) ((T_PRINT) reduce 104) ((T_PROPERTY_C) reduce 104) ((T_READONLY) reduce 104) ((T_REQUIRE) reduce 104) ((T_REQUIRE_ONCE) reduce 104) ((T_RETURN) reduce 104) ((T_START_HEREDOC) reduce 104) ((T_STATIC) reduce 104) ((T_STRING) reduce 104) ((T_STRING_CAST) reduce 104) ((T_SWITCH) reduce 104) ((T_THROW) reduce 104) ((T_TRAIT) reduce 104) ((T_TRAIT_C) reduce 104) ((T_TRY) reduce 104) ((T_UNSET) reduce 104) ((T_UNSET_CAST) reduce 104) ((T_VARIABLE) reduce 104) ((T_VOID_CAST) reduce 104) ((T_WHILE) reduce 104) ((T_YIELD) reduce 104) ((T_YIELD_FROM) reduce 104) (("[") reduce 104) (("`") reduce 104) (("{") reduce 104) (("}") reduce 104) (("~") reduce 104)) 206 ((("!") reduce 107) (("\"") reduce 107) (("$") reduce 107) (("(") reduce 107) (("+") reduce 107) (("-") reduce 107) ((";") reduce 107) (("@") reduce 107) ((T_ABSTRACT) reduce 107) ((T_ARRAY) reduce 107) ((T_ARRAY_CAST) reduce 107) ((T_ATTRIBUTE) reduce 107) ((T_BOOL_CAST) reduce 107) ((T_BREAK) reduce 107) ((T_CLASS) reduce 107) ((T_CLASS_C) reduce 107) ((T_CLONE) reduce 107) ((T_CONSTANT_ENCAPSED_STRING) reduce 107) ((T_CONTINUE) reduce 107) ((T_DEC) reduce 107) ((T_DECLARE) reduce 107) ((T_DIR) reduce 107) ((T_DNUMBER) reduce 107) ((T_DO) reduce 107) ((T_DOUBLE_CAST) reduce 107) ((T_ECHO) reduce 107) ((T_EMPTY) reduce 107) ((T_ENUM) reduce 107) ((T_EVAL) reduce 107) ((T_EXIT) reduce 107) ((T_FILE) reduce 107) ((T_FINAL) reduce 107) ((T_FN) reduce 107) ((T_FOR) reduce 107) ((T_FOREACH) reduce 107) ((T_FUNCTION) reduce 107) ((T_FUNC_C) reduce 107) ((T_GLOBAL) reduce 107) ((T_GOTO) reduce 107) ((T_HALT_COMPILER) reduce 107) ((T_IF) reduce 107) ((T_INC) reduce 107) ((T_INCLUDE) reduce 107) ((T_INCLUDE_ONCE) reduce 107) ((T_INLINE_HTML) reduce 107) ((T_INTERFACE) reduce 107) ((T_INT_CAST) reduce 107) ((T_ISSET) reduce 107) ((T_LINE) reduce 107) ((T_LIST) reduce 107) ((T_LNUMBER) reduce 107) ((T_MATCH) reduce 107) ((T_METHOD_C) reduce 107) ((T_NAME_FULLY_QUALIFIED) reduce 107) ((T_NAME_QUALIFIED) reduce 107) ((T_NAME_RELATIVE) reduce 107) ((T_NEW) reduce 107) ((T_NS_C) reduce 107) ((T_OBJECT_CAST) reduce 107) ((T_PRINT) reduce 107) ((T_PROPERTY_C) reduce 107) ((T_READONLY) reduce 107) ((T_REQUIRE) reduce 107) ((T_REQUIRE_ONCE) reduce 107) ((T_RETURN) reduce 107) ((T_START_HEREDOC) reduce 107) ((T_STATIC) reduce 107) ((T_STRING) reduce 107) ((T_STRING_CAST) reduce 107) ((T_SWITCH) reduce 107) ((T_THROW) reduce 107) ((T_TRAIT) reduce 107) ((T_TRAIT_C) reduce 107) ((T_TRY) reduce 107) ((T_UNSET) reduce 107) ((T_UNSET_CAST) reduce 107) ((T_VARIABLE) reduce 107) ((T_VOID_CAST) reduce 107) ((T_WHILE) reduce 107) ((T_YIELD) reduce 107) ((T_YIELD_FROM) reduce 107) (("[") reduce 107) (("`") reduce 107) (("{") reduce 107) (("}") reduce 107) (("~") reduce 107)) 209 ((("!") reduce 103) (("\"") reduce 103) (("$") reduce 103) (("(") reduce 103) (("+") reduce 103) (("-") reduce 103) ((";") reduce 103) (("@") reduce 103) ((T_ABSTRACT) reduce 103) ((T_ARRAY) reduce 103) ((T_ARRAY_CAST) reduce 103) ((T_ATTRIBUTE) reduce 103) ((T_BOOL_CAST) reduce 103) ((T_BREAK) reduce 103) ((T_CLASS) reduce 103) ((T_CLASS_C) reduce 103) ((T_CLONE) reduce 103) ((T_CONSTANT_ENCAPSED_STRING) reduce 103) ((T_CONTINUE) reduce 103) ((T_DEC) reduce 103) ((T_DECLARE) reduce 103) ((T_DIR) reduce 103) ((T_DNUMBER) reduce 103) ((T_DO) reduce 103) ((T_DOUBLE_CAST) reduce 103) ((T_ECHO) reduce 103) ((T_EMPTY) reduce 103) ((T_ENUM) reduce 103) ((T_EVAL) reduce 103) ((T_EXIT) reduce 103) ((T_FILE) reduce 103) ((T_FINAL) reduce 103) ((T_FN) reduce 103) ((T_FOR) reduce 103) ((T_FOREACH) reduce 103) ((T_FUNCTION) reduce 103) ((T_FUNC_C) reduce 103) ((T_GLOBAL) reduce 103) ((T_GOTO) reduce 103) ((T_HALT_COMPILER) reduce 103) ((T_IF) reduce 103) ((T_INC) reduce 103) ((T_INCLUDE) reduce 103) ((T_INCLUDE_ONCE) reduce 103) ((T_INLINE_HTML) reduce 103) ((T_INTERFACE) reduce 103) ((T_INT_CAST) reduce 103) ((T_ISSET) reduce 103) ((T_LINE) reduce 103) ((T_LIST) reduce 103) ((T_LNUMBER) reduce 103) ((T_MATCH) reduce 103) ((T_METHOD_C) reduce 103) ((T_NAME_FULLY_QUALIFIED) reduce 103) ((T_NAME_QUALIFIED) reduce 103) ((T_NAME_RELATIVE) reduce 103) ((T_NEW) reduce 103) ((T_NS_C) reduce 103) ((T_OBJECT_CAST) reduce 103) ((T_PRINT) reduce 103) ((T_PROPERTY_C) reduce 103) ((T_READONLY) reduce 103) ((T_REQUIRE) reduce 103) ((T_REQUIRE_ONCE) reduce 103) ((T_RETURN) reduce 103) ((T_START_HEREDOC) reduce 103) ((T_STATIC) reduce 103) ((T_STRING) reduce 103) ((T_STRING_CAST) reduce 103) ((T_SWITCH) reduce 103) ((T_THROW) reduce 103) ((T_TRAIT) reduce 103) ((T_TRAIT_C) reduce 103) ((T_TRY) reduce 103) ((T_UNSET) reduce 103) ((T_UNSET_CAST) reduce 103) ((T_VARIABLE) reduce 103) ((T_VOID_CAST) reduce 103) ((T_WHILE) reduce 103) ((T_YIELD) reduce 103) ((T_YIELD_FROM) reduce 103) (("[") reduce 103) (("`") reduce 103) (("{") reduce 103) (("}") reduce 103) (("~") reduce 103)) 210 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_EMPTY) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_VOID_CAST) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("}") reduce 148) (("~") reduce 148)) 211 ((("!") reduce 246) (("\"") reduce 246) (("$") reduce 246) (("(") reduce 246) (("+") reduce 246) (("-") reduce 246) ((";") reduce 246) (("@") reduce 246) ((T_ABSTRACT) reduce 246) ((T_ARRAY) reduce 246) ((T_ARRAY_CAST) reduce 246) ((T_ATTRIBUTE) reduce 246) ((T_BOOL_CAST) reduce 246) ((T_BREAK) reduce 246) ((T_CLASS) reduce 246) ((T_CLASS_C) reduce 246) ((T_CLONE) reduce 246) ((T_CONSTANT_ENCAPSED_STRING) reduce 246) ((T_CONTINUE) reduce 246) ((T_DEC) reduce 246) ((T_DECLARE) reduce 246) ((T_DIR) reduce 246) ((T_DNUMBER) reduce 246) ((T_DO) reduce 246) ((T_DOUBLE_CAST) reduce 246) ((T_ECHO) reduce 246) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) reduce 246) ((T_ENUM) reduce 246) ((T_EVAL) reduce 246) ((T_EXIT) reduce 246) ((T_FILE) reduce 246) ((T_FINAL) reduce 246) ((T_FN) reduce 246) ((T_FOR) reduce 246) ((T_FOREACH) reduce 246) ((T_FUNCTION) reduce 246) ((T_FUNC_C) reduce 246) ((T_GLOBAL) reduce 246) ((T_GOTO) reduce 246) ((T_HALT_COMPILER) reduce 246) ((T_IF) reduce 246) ((T_INC) reduce 246) ((T_INCLUDE) reduce 246) ((T_INCLUDE_ONCE) reduce 246) ((T_INLINE_HTML) reduce 246) ((T_INTERFACE) reduce 246) ((T_INT_CAST) reduce 246) ((T_ISSET) reduce 246) ((T_LINE) reduce 246) ((T_LIST) reduce 246) ((T_LNUMBER) reduce 246) ((T_MATCH) reduce 246) ((T_METHOD_C) reduce 246) ((T_NAME_FULLY_QUALIFIED) reduce 246) ((T_NAME_QUALIFIED) reduce 246) ((T_NAME_RELATIVE) reduce 246) ((T_NEW) reduce 246) ((T_NS_C) reduce 246) ((T_OBJECT_CAST) reduce 246) ((T_PRINT) reduce 246) ((T_PROPERTY_C) reduce 246) ((T_READONLY) reduce 246) ((T_REQUIRE) reduce 246) ((T_REQUIRE_ONCE) reduce 246) ((T_RETURN) reduce 246) ((T_START_HEREDOC) reduce 246) ((T_STATIC) reduce 246) ((T_STRING) reduce 246) ((T_STRING_CAST) reduce 246) ((T_SWITCH) reduce 246) ((T_THROW) reduce 246) ((T_TRAIT) reduce 246) ((T_TRAIT_C) reduce 246) ((T_TRY) reduce 246) ((T_UNSET) reduce 246) ((T_UNSET_CAST) reduce 246) ((T_VARIABLE) reduce 246) ((T_VOID_CAST) reduce 246) ((T_WHILE) reduce 246) ((T_YIELD) reduce 246) ((T_YIELD_FROM) reduce 246) (("[") reduce 246) (("`") reduce 246) (("{") reduce 246) (("}") reduce 246) (("~") reduce 246)) 212 ((("!") reduce 141) (("\"") reduce 141) (("$") reduce 141) (("(") reduce 141) (("+") reduce 141) (("-") reduce 141) ((";") reduce 141) (("@") reduce 141) ((T_ABSTRACT) reduce 141) ((T_ARRAY) reduce 141) ((T_ARRAY_CAST) reduce 141) ((T_ATTRIBUTE) reduce 141) ((T_BOOL_CAST) reduce 141) ((T_BREAK) reduce 141) ((T_CLASS) reduce 141) ((T_CLASS_C) reduce 141) ((T_CLONE) reduce 141) ((T_CONSTANT_ENCAPSED_STRING) reduce 141) ((T_CONTINUE) reduce 141) ((T_DEC) reduce 141) ((T_DECLARE) reduce 141) ((T_DIR) reduce 141) ((T_DNUMBER) reduce 141) ((T_DO) reduce 141) ((T_DOUBLE_CAST) reduce 141) ((T_ECHO) reduce 141) ((T_EMPTY) reduce 141) ((T_ENUM) reduce 141) ((T_EVAL) reduce 141) ((T_EXIT) reduce 141) ((T_FILE) reduce 141) ((T_FINAL) reduce 141) ((T_FN) reduce 141) ((T_FOR) reduce 141) ((T_FOREACH) reduce 141) ((T_FUNCTION) reduce 141) ((T_FUNC_C) reduce 141) ((T_GLOBAL) reduce 141) ((T_GOTO) reduce 141) ((T_HALT_COMPILER) reduce 141) ((T_IF) reduce 141) ((T_INC) reduce 141) ((T_INCLUDE) reduce 141) ((T_INCLUDE_ONCE) reduce 141) ((T_INLINE_HTML) reduce 141) ((T_INTERFACE) reduce 141) ((T_INT_CAST) reduce 141) ((T_ISSET) reduce 141) ((T_LINE) reduce 141) ((T_LIST) reduce 141) ((T_LNUMBER) reduce 141) ((T_MATCH) reduce 141) ((T_METHOD_C) reduce 141) ((T_NAME_FULLY_QUALIFIED) reduce 141) ((T_NAME_QUALIFIED) reduce 141) ((T_NAME_RELATIVE) reduce 141) ((T_NEW) reduce 141) ((T_NS_C) reduce 141) ((T_OBJECT_CAST) reduce 141) ((T_PRINT) reduce 141) ((T_PROPERTY_C) reduce 141) ((T_READONLY) reduce 141) ((T_REQUIRE) reduce 141) ((T_REQUIRE_ONCE) reduce 141) ((T_RETURN) reduce 141) ((T_START_HEREDOC) reduce 141) ((T_STATIC) reduce 141) ((T_STRING) reduce 141) ((T_STRING_CAST) reduce 141) ((T_SWITCH) reduce 141) ((T_THROW) reduce 141) ((T_TRAIT) reduce 141) ((T_TRAIT_C) reduce 141) ((T_TRY) reduce 141) ((T_UNSET) reduce 141) ((T_UNSET_CAST) reduce 141) ((T_VARIABLE) reduce 141) ((T_VOID_CAST) reduce 141) ((T_WHILE) reduce 141) ((T_YIELD) reduce 141) ((T_YIELD_FROM) reduce 141) (("[") reduce 141) (("`") reduce 141) (("{") reduce 141) (("}") reduce 141) (("~") reduce 141)) 213 ((("!") reduce 106) (("\"") reduce 106) (("$") reduce 106) (("(") reduce 106) (("+") reduce 106) (("-") reduce 106) ((";") reduce 106) (("@") reduce 106) ((T_ABSTRACT) reduce 106) ((T_ARRAY) reduce 106) ((T_ARRAY_CAST) reduce 106) ((T_ATTRIBUTE) reduce 106) ((T_BOOL_CAST) reduce 106) ((T_BREAK) reduce 106) ((T_CLASS) reduce 106) ((T_CLASS_C) reduce 106) ((T_CLONE) reduce 106) ((T_CONSTANT_ENCAPSED_STRING) reduce 106) ((T_CONTINUE) reduce 106) ((T_DEC) reduce 106) ((T_DECLARE) reduce 106) ((T_DIR) reduce 106) ((T_DNUMBER) reduce 106) ((T_DO) reduce 106) ((T_DOUBLE_CAST) reduce 106) ((T_ECHO) reduce 106) ((T_EMPTY) reduce 106) ((T_ENUM) reduce 106) ((T_EVAL) reduce 106) ((T_EXIT) reduce 106) ((T_FILE) reduce 106) ((T_FINAL) reduce 106) ((T_FN) reduce 106) ((T_FOR) reduce 106) ((T_FOREACH) reduce 106) ((T_FUNCTION) reduce 106) ((T_FUNC_C) reduce 106) ((T_GLOBAL) reduce 106) ((T_GOTO) reduce 106) ((T_HALT_COMPILER) reduce 106) ((T_IF) reduce 106) ((T_INC) reduce 106) ((T_INCLUDE) reduce 106) ((T_INCLUDE_ONCE) reduce 106) ((T_INLINE_HTML) reduce 106) ((T_INTERFACE) reduce 106) ((T_INT_CAST) reduce 106) ((T_ISSET) reduce 106) ((T_LINE) reduce 106) ((T_LIST) reduce 106) ((T_LNUMBER) reduce 106) ((T_MATCH) reduce 106) ((T_METHOD_C) reduce 106) ((T_NAME_FULLY_QUALIFIED) reduce 106) ((T_NAME_QUALIFIED) reduce 106) ((T_NAME_RELATIVE) reduce 106) ((T_NEW) reduce 106) ((T_NS_C) reduce 106) ((T_OBJECT_CAST) reduce 106) ((T_PRINT) reduce 106) ((T_PROPERTY_C) reduce 106) ((T_READONLY) reduce 106) ((T_REQUIRE) reduce 106) ((T_REQUIRE_ONCE) reduce 106) ((T_RETURN) reduce 106) ((T_START_HEREDOC) reduce 106) ((T_STATIC) reduce 106) ((T_STRING) reduce 106) ((T_STRING_CAST) reduce 106) ((T_SWITCH) reduce 106) ((T_THROW) reduce 106) ((T_TRAIT) reduce 106) ((T_TRAIT_C) reduce 106) ((T_TRY) reduce 106) ((T_UNSET) reduce 106) ((T_UNSET_CAST) reduce 106) ((T_VARIABLE) reduce 106) ((T_VOID_CAST) reduce 106) ((T_WHILE) reduce 106) ((T_YIELD) reduce 106) ((T_YIELD_FROM) reduce 106) (("[") reduce 106) (("`") reduce 106) (("{") reduce 106) (("}") reduce 106) (("~") reduce 106)) 214 ((("!") reduce 143) (("\"") reduce 143) (("$") reduce 143) (("(") reduce 143) (("+") reduce 143) (("-") reduce 143) ((";") reduce 143) (("@") reduce 143) ((T_ABSTRACT) reduce 143) ((T_ARRAY) reduce 143) ((T_ARRAY_CAST) reduce 143) ((T_ATTRIBUTE) reduce 143) ((T_BOOL_CAST) reduce 143) ((T_BREAK) reduce 143) ((T_CLASS) reduce 143) ((T_CLASS_C) reduce 143) ((T_CLONE) reduce 143) ((T_CONSTANT_ENCAPSED_STRING) reduce 143) ((T_CONTINUE) reduce 143) ((T_DEC) reduce 143) ((T_DECLARE) reduce 143) ((T_DIR) reduce 143) ((T_DNUMBER) reduce 143) ((T_DO) reduce 143) ((T_DOUBLE_CAST) reduce 143) ((T_ECHO) reduce 143) ((T_EMPTY) reduce 143) ((T_ENUM) reduce 143) ((T_EVAL) reduce 143) ((T_EXIT) reduce 143) ((T_FILE) reduce 143) ((T_FINAL) reduce 143) ((T_FN) reduce 143) ((T_FOR) reduce 143) ((T_FOREACH) reduce 143) ((T_FUNCTION) reduce 143) ((T_FUNC_C) reduce 143) ((T_GLOBAL) reduce 143) ((T_GOTO) reduce 143) ((T_HALT_COMPILER) reduce 143) ((T_IF) reduce 143) ((T_INC) reduce 143) ((T_INCLUDE) reduce 143) ((T_INCLUDE_ONCE) reduce 143) ((T_INLINE_HTML) reduce 143) ((T_INTERFACE) reduce 143) ((T_INT_CAST) reduce 143) ((T_ISSET) reduce 143) ((T_LINE) reduce 143) ((T_LIST) reduce 143) ((T_LNUMBER) reduce 143) ((T_MATCH) reduce 143) ((T_METHOD_C) reduce 143) ((T_NAME_FULLY_QUALIFIED) reduce 143) ((T_NAME_QUALIFIED) reduce 143) ((T_NAME_RELATIVE) reduce 143) ((T_NEW) reduce 143) ((T_NS_C) reduce 143) ((T_OBJECT_CAST) reduce 143) ((T_PRINT) reduce 143) ((T_PROPERTY_C) reduce 143) ((T_READONLY) reduce 143) ((T_REQUIRE) reduce 143) ((T_REQUIRE_ONCE) reduce 143) ((T_RETURN) reduce 143) ((T_START_HEREDOC) reduce 143) ((T_STATIC) reduce 143) ((T_STRING) reduce 143) ((T_STRING_CAST) reduce 143) ((T_SWITCH) reduce 143) ((T_THROW) reduce 143) ((T_TRAIT) reduce 143) ((T_TRAIT_C) reduce 143) ((T_TRY) reduce 143) ((T_UNSET) reduce 143) ((T_UNSET_CAST) reduce 143) ((T_VARIABLE) reduce 143) ((T_VOID_CAST) reduce 143) ((T_WHILE) reduce 143) ((T_YIELD) reduce 143) ((T_YIELD_FROM) reduce 143) (("[") reduce 143) (("`") reduce 143) (("{") reduce 143) (("}") reduce 143) (("~") reduce 143)) 215 ((("!") reduce 105) (("\"") reduce 105) (("$") reduce 105) (("(") reduce 105) (("+") reduce 105) (("-") reduce 105) ((";") reduce 105) (("@") reduce 105) ((T_ABSTRACT) reduce 105) ((T_ARRAY) reduce 105) ((T_ARRAY_CAST) reduce 105) ((T_ATTRIBUTE) reduce 105) ((T_BOOL_CAST) reduce 105) ((T_BREAK) reduce 105) ((T_CLASS) reduce 105) ((T_CLASS_C) reduce 105) ((T_CLONE) reduce 105) ((T_CONSTANT_ENCAPSED_STRING) reduce 105) ((T_CONTINUE) reduce 105) ((T_DEC) reduce 105) ((T_DECLARE) reduce 105) ((T_DIR) reduce 105) ((T_DNUMBER) reduce 105) ((T_DO) reduce 105) ((T_DOUBLE_CAST) reduce 105) ((T_ECHO) reduce 105) ((T_EMPTY) reduce 105) ((T_ENUM) reduce 105) ((T_EVAL) reduce 105) ((T_EXIT) reduce 105) ((T_FILE) reduce 105) ((T_FINAL) reduce 105) ((T_FN) reduce 105) ((T_FOR) reduce 105) ((T_FOREACH) reduce 105) ((T_FUNCTION) reduce 105) ((T_FUNC_C) reduce 105) ((T_GLOBAL) reduce 105) ((T_GOTO) reduce 105) ((T_HALT_COMPILER) reduce 105) ((T_IF) reduce 105) ((T_INC) reduce 105) ((T_INCLUDE) reduce 105) ((T_INCLUDE_ONCE) reduce 105) ((T_INLINE_HTML) reduce 105) ((T_INTERFACE) reduce 105) ((T_INT_CAST) reduce 105) ((T_ISSET) reduce 105) ((T_LINE) reduce 105) ((T_LIST) reduce 105) ((T_LNUMBER) reduce 105) ((T_MATCH) reduce 105) ((T_METHOD_C) reduce 105) ((T_NAME_FULLY_QUALIFIED) reduce 105) ((T_NAME_QUALIFIED) reduce 105) ((T_NAME_RELATIVE) reduce 105) ((T_NEW) reduce 105) ((T_NS_C) reduce 105) ((T_OBJECT_CAST) reduce 105) ((T_PRINT) reduce 105) ((T_PROPERTY_C) reduce 105) ((T_READONLY) reduce 105) ((T_REQUIRE) reduce 105) ((T_REQUIRE_ONCE) reduce 105) ((T_RETURN) reduce 105) ((T_START_HEREDOC) reduce 105) ((T_STATIC) reduce 105) ((T_STRING) reduce 105) ((T_STRING_CAST) reduce 105) ((T_SWITCH) reduce 105) ((T_THROW) reduce 105) ((T_TRAIT) reduce 105) ((T_TRAIT_C) reduce 105) ((T_TRY) reduce 105) ((T_UNSET) reduce 105) ((T_UNSET_CAST) reduce 105) ((T_VARIABLE) reduce 105) ((T_VOID_CAST) reduce 105) ((T_WHILE) reduce 105) ((T_YIELD) reduce 105) ((T_YIELD_FROM) reduce 105) (("[") reduce 105) (("`") reduce 105) (("{") reduce 105) (("}") reduce 105) (("~") reduce 105)) 217 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) ((";") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PIPE) reduce 494) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 218 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) ((";") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_PIPE) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("^") reduce 489) (("|") reduce 489)) 220 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_EMPTY) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_VOID_CAST) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("}") reduce 147) (("~") reduce 147)) 234 ((("%") reduce 540) ((")") reduce 540) (("*") reduce 540) (("+") reduce 540) (("-") reduce 540) ((".") reduce 540) (("/") reduce 540) (("<") reduce 540) ((">") reduce 540) (("?") reduce 540) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_BOOLEAN_AND) reduce 540) ((T_BOOLEAN_OR) reduce 540) ((T_COALESCE) reduce 540) ((T_INSTANCEOF) reduce 540) ((T_IS_EQUAL) reduce 540) ((T_IS_GREATER_OR_EQUAL) reduce 540) ((T_IS_IDENTICAL) reduce 540) ((T_IS_NOT_EQUAL) reduce 540) ((T_IS_NOT_IDENTICAL) reduce 540) ((T_IS_SMALLER_OR_EQUAL) reduce 540) ((T_LOGICAL_AND) reduce 540) ((T_LOGICAL_OR) reduce 540) ((T_LOGICAL_XOR) reduce 540) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 540) ((T_OBJECT_OPERATOR) reduce 540) ((T_PIPE) reduce 540) ((T_POW) reduce 540) ((T_SL) reduce 540) ((T_SPACESHIP) reduce 540) ((T_SR) reduce 540) (("[") reduce 540) (("^") reduce 540) (("|") reduce 540)) 236 ((("%") reduce 521) (("(") reduce 521) ((")") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_PIPE) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("|") reduce 521)) 238 ((("%") reduce 534) ((")") reduce 534) (("*") reduce 534) (("+") reduce 534) (("-") reduce 534) ((".") reduce 534) (("/") reduce 534) (("<") reduce 534) ((">") reduce 534) (("?") reduce 534) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_BOOLEAN_AND) reduce 534) ((T_BOOLEAN_OR) reduce 534) ((T_COALESCE) reduce 534) ((T_INSTANCEOF) reduce 534) ((T_IS_EQUAL) reduce 534) ((T_IS_GREATER_OR_EQUAL) reduce 534) ((T_IS_IDENTICAL) reduce 534) ((T_IS_NOT_EQUAL) reduce 534) ((T_IS_NOT_IDENTICAL) reduce 534) ((T_IS_SMALLER_OR_EQUAL) reduce 534) ((T_LOGICAL_AND) reduce 534) ((T_LOGICAL_OR) reduce 534) ((T_LOGICAL_XOR) reduce 534) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PIPE) reduce 534) ((T_POW) reduce 534) ((T_SL) reduce 534) ((T_SPACESHIP) reduce 534) ((T_SR) reduce 534) (("[") reduce 534) (("^") reduce 534) (("|") reduce 534)) 239 ((("%") reduce 524) ((")") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_PIPE) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("^") reduce 524) (("|") reduce 524)) 243 ((("%") reduce 517) (("(") shift) ((")") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_PIPE) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("^") reduce 517) (("|") reduce 517)) 244 ((("%") reduce 533) ((")") reduce 533) (("*") reduce 533) (("+") reduce 533) (("-") reduce 533) ((".") reduce 533) (("/") reduce 533) (("<") reduce 533) ((">") reduce 533) (("?") reduce 533) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_BOOLEAN_AND) reduce 533) ((T_BOOLEAN_OR) reduce 533) ((T_COALESCE) reduce 533) ((T_INSTANCEOF) reduce 533) ((T_IS_EQUAL) reduce 533) ((T_IS_GREATER_OR_EQUAL) reduce 533) ((T_IS_IDENTICAL) reduce 533) ((T_IS_NOT_EQUAL) reduce 533) ((T_IS_NOT_IDENTICAL) reduce 533) ((T_IS_SMALLER_OR_EQUAL) reduce 533) ((T_LOGICAL_AND) reduce 533) ((T_LOGICAL_OR) reduce 533) ((T_LOGICAL_XOR) reduce 533) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PIPE) reduce 533) ((T_POW) reduce 533) ((T_SL) reduce 533) ((T_SPACESHIP) reduce 533) ((T_SR) reduce 533) (("[") reduce 533) (("^") reduce 533) (("|") reduce 533)) 245 ((("%") reduce 537) ((")") reduce 537) (("*") reduce 537) (("+") reduce 537) (("-") reduce 537) ((".") reduce 537) (("/") reduce 537) (("<") reduce 537) ((">") reduce 537) (("?") reduce 537) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_BOOLEAN_AND) reduce 537) ((T_BOOLEAN_OR) reduce 537) ((T_COALESCE) reduce 537) ((T_INSTANCEOF) reduce 537) ((T_IS_EQUAL) reduce 537) ((T_IS_GREATER_OR_EQUAL) reduce 537) ((T_IS_IDENTICAL) reduce 537) ((T_IS_NOT_EQUAL) reduce 537) ((T_IS_NOT_IDENTICAL) reduce 537) ((T_IS_SMALLER_OR_EQUAL) reduce 537) ((T_LOGICAL_AND) reduce 537) ((T_LOGICAL_OR) reduce 537) ((T_LOGICAL_XOR) reduce 537) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_PIPE) reduce 537) ((T_POW) reduce 537) ((T_SL) reduce 537) ((T_SPACESHIP) reduce 537) ((T_SR) reduce 537) (("[") reduce 537) (("^") reduce 537) (("|") reduce 537)) 251 ((("%") reduce 532) ((")") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PIPE) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("|") reduce 532)) 253 ((("%") reduce 523) ((")") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_PIPE) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("^") reduce 523) (("|") reduce 523)) 255 ((("%") reduce 536) ((")") reduce 536) (("*") reduce 536) (("+") reduce 536) (("-") reduce 536) ((".") reduce 536) (("/") reduce 536) (("<") reduce 536) ((">") reduce 536) (("?") reduce 536) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_BOOLEAN_AND) reduce 536) ((T_BOOLEAN_OR) reduce 536) ((T_COALESCE) reduce 536) ((T_INSTANCEOF) reduce 536) ((T_IS_EQUAL) reduce 536) ((T_IS_GREATER_OR_EQUAL) reduce 536) ((T_IS_IDENTICAL) reduce 536) ((T_IS_NOT_EQUAL) reduce 536) ((T_IS_NOT_IDENTICAL) reduce 536) ((T_IS_SMALLER_OR_EQUAL) reduce 536) ((T_LOGICAL_AND) reduce 536) ((T_LOGICAL_OR) reduce 536) ((T_LOGICAL_XOR) reduce 536) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 536) ((T_OBJECT_OPERATOR) reduce 536) ((T_PIPE) reduce 536) ((T_POW) reduce 536) ((T_SL) reduce 536) ((T_SPACESHIP) reduce 536) ((T_SR) reduce 536) (("[") reduce 536) (("^") reduce 536) (("|") reduce 536)) 256 ((("%") reduce 94) (("(") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("|") reduce 94)) 257 ((("%") reduce 93) (("(") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("|") reduce 93)) 258 ((("%") reduce 95) (("(") reduce 95) ((")") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("|") reduce 95)) 260 ((("%") reduce 539) ((")") reduce 539) (("*") reduce 539) (("+") reduce 539) (("-") reduce 539) ((".") reduce 539) (("/") reduce 539) (("<") reduce 539) ((">") reduce 539) (("?") reduce 539) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_BOOLEAN_AND) reduce 539) ((T_BOOLEAN_OR) reduce 539) ((T_COALESCE) reduce 539) ((T_INSTANCEOF) reduce 539) ((T_IS_EQUAL) reduce 539) ((T_IS_GREATER_OR_EQUAL) reduce 539) ((T_IS_IDENTICAL) reduce 539) ((T_IS_NOT_EQUAL) reduce 539) ((T_IS_NOT_IDENTICAL) reduce 539) ((T_IS_SMALLER_OR_EQUAL) reduce 539) ((T_LOGICAL_AND) reduce 539) ((T_LOGICAL_OR) reduce 539) ((T_LOGICAL_XOR) reduce 539) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 539) ((T_OBJECT_OPERATOR) reduce 539) ((T_PIPE) reduce 539) ((T_POW) reduce 539) ((T_SL) reduce 539) ((T_SPACESHIP) reduce 539) ((T_SR) reduce 539) (("[") reduce 539) (("^") reduce 539) (("|") reduce 539)) 263 ((("%") reduce 538) ((")") reduce 538) (("*") reduce 538) (("+") reduce 538) (("-") reduce 538) ((".") reduce 538) (("/") reduce 538) (("<") reduce 538) ((">") reduce 538) (("?") reduce 538) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_BOOLEAN_AND) reduce 538) ((T_BOOLEAN_OR) reduce 538) ((T_COALESCE) reduce 538) ((T_INSTANCEOF) reduce 538) ((T_IS_EQUAL) reduce 538) ((T_IS_GREATER_OR_EQUAL) reduce 538) ((T_IS_IDENTICAL) reduce 538) ((T_IS_NOT_EQUAL) reduce 538) ((T_IS_NOT_IDENTICAL) reduce 538) ((T_IS_SMALLER_OR_EQUAL) reduce 538) ((T_LOGICAL_AND) reduce 538) ((T_LOGICAL_OR) reduce 538) ((T_LOGICAL_XOR) reduce 538) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 538) ((T_OBJECT_OPERATOR) reduce 538) ((T_PIPE) reduce 538) ((T_POW) reduce 538) ((T_SL) reduce 538) ((T_SPACESHIP) reduce 538) ((T_SR) reduce 538) (("[") reduce 538) (("^") reduce 538) (("|") reduce 538)) 269 ((("%") reduce 92) (("(") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92)) 272 ((("%") reduce 535) ((")") reduce 535) (("*") reduce 535) (("+") reduce 535) (("-") reduce 535) ((".") reduce 535) (("/") reduce 535) (("<") reduce 535) ((">") reduce 535) (("?") reduce 535) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_BOOLEAN_AND) reduce 535) ((T_BOOLEAN_OR) reduce 535) ((T_COALESCE) reduce 535) ((T_INSTANCEOF) reduce 535) ((T_IS_EQUAL) reduce 535) ((T_IS_GREATER_OR_EQUAL) reduce 535) ((T_IS_IDENTICAL) reduce 535) ((T_IS_NOT_EQUAL) reduce 535) ((T_IS_NOT_IDENTICAL) reduce 535) ((T_IS_SMALLER_OR_EQUAL) reduce 535) ((T_LOGICAL_AND) reduce 535) ((T_LOGICAL_OR) reduce 535) ((T_LOGICAL_XOR) reduce 535) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PIPE) reduce 535) ((T_POW) reduce 535) ((T_SL) reduce 535) ((T_SPACESHIP) reduce 535) ((T_SR) reduce 535) (("[") reduce 535) (("^") reduce 535) (("|") reduce 535)) 274 ((("%") reduce 568) (("(") reduce 568) ((")") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) (("=") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AND_EQUAL) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_COALESCE_EQUAL) reduce 568) ((T_CONCAT_EQUAL) reduce 568) ((T_DEC) reduce 568) ((T_DIV_EQUAL) reduce 568) ((T_INC) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_MINUS_EQUAL) reduce 568) ((T_MOD_EQUAL) reduce 568) ((T_MUL_EQUAL) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_OR_EQUAL) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_PLUS_EQUAL) reduce 568) ((T_POW) reduce 568) ((T_POW_EQUAL) reduce 568) ((T_SL) reduce 568) ((T_SL_EQUAL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) ((T_SR_EQUAL) reduce 568) ((T_XOR_EQUAL) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 275 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 479) (("(") shift) ((")") reduce 479) (("*") reduce 479) (("+") shift) (("-") shift) ((".") reduce 479) (("/") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 479) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 479) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PIPE) reduce 479) ((T_POW) reduce 479) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 479) (("`") shift) (("|") reduce 479) (("~") shift)) 282 ((("%") reduce 564) (("(") reduce 555) ((")") reduce 564) (("*") reduce 564) (("+") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) (("<") reduce 564) (("=") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AND_EQUAL) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_COALESCE_EQUAL) reduce 564) ((T_CONCAT_EQUAL) reduce 564) ((T_DEC) reduce 564) ((T_DIV_EQUAL) reduce 564) ((T_INC) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_MINUS_EQUAL) reduce 564) ((T_MOD_EQUAL) reduce 564) ((T_MUL_EQUAL) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_OR_EQUAL) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_PLUS_EQUAL) reduce 564) ((T_POW) reduce 564) ((T_POW_EQUAL) reduce 564) ((T_SL) reduce 564) ((T_SL_EQUAL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) ((T_SR_EQUAL) reduce 564) ((T_XOR_EQUAL) reduce 564) (("[") reduce 564) (("^") reduce 564) (("|") reduce 564)) 283 ((("%") reduce 530) ((")") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PIPE) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 551) (("^") reduce 530) (("|") reduce 530)) 285 ((("%") reduce 529) ((")") reduce 529) (("*") reduce 529) (("+") reduce 529) (("-") reduce 529) ((".") reduce 529) (("/") reduce 529) (("<") reduce 529) ((">") reduce 529) (("?") reduce 529) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_BOOLEAN_AND) reduce 529) ((T_BOOLEAN_OR) reduce 529) ((T_COALESCE) reduce 529) ((T_INSTANCEOF) reduce 529) ((T_IS_EQUAL) reduce 529) ((T_IS_GREATER_OR_EQUAL) reduce 529) ((T_IS_IDENTICAL) reduce 529) ((T_IS_NOT_EQUAL) reduce 529) ((T_IS_NOT_IDENTICAL) reduce 529) ((T_IS_SMALLER_OR_EQUAL) reduce 529) ((T_LOGICAL_AND) reduce 529) ((T_LOGICAL_OR) reduce 529) ((T_LOGICAL_XOR) reduce 529) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PIPE) reduce 529) ((T_POW) reduce 529) ((T_SL) reduce 529) ((T_SPACESHIP) reduce 529) ((T_SR) reduce 529) (("[") reduce 554) (("^") reduce 529) (("|") reduce 529)) 286 ((("%") reduce 528) (("(") reduce 557) ((")") reduce 528) (("*") reduce 528) (("+") reduce 528) (("-") reduce 528) ((".") reduce 528) (("/") reduce 528) (("<") reduce 528) ((">") reduce 528) (("?") reduce 528) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_BOOLEAN_AND) reduce 528) ((T_BOOLEAN_OR) reduce 528) ((T_COALESCE) reduce 528) ((T_INSTANCEOF) reduce 528) ((T_IS_EQUAL) reduce 528) ((T_IS_GREATER_OR_EQUAL) reduce 528) ((T_IS_IDENTICAL) reduce 528) ((T_IS_NOT_EQUAL) reduce 528) ((T_IS_NOT_IDENTICAL) reduce 528) ((T_IS_SMALLER_OR_EQUAL) reduce 528) ((T_LOGICAL_AND) reduce 528) ((T_LOGICAL_OR) reduce 528) ((T_LOGICAL_XOR) reduce 528) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PIPE) reduce 528) ((T_POW) reduce 528) ((T_SL) reduce 528) ((T_SPACESHIP) reduce 528) ((T_SR) reduce 528) (("[") reduce 550) (("^") reduce 528) (("|") reduce 528)) 287 ((("%") shift) ((")") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 290 ((("%") reduce 563) (("(") reduce 563) ((")") reduce 563) (("*") reduce 563) (("+") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) (("<") reduce 563) (("=") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AND_EQUAL) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_COALESCE_EQUAL) reduce 563) ((T_CONCAT_EQUAL) reduce 563) ((T_DEC) reduce 563) ((T_DIV_EQUAL) reduce 563) ((T_INC) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_MINUS_EQUAL) reduce 563) ((T_MOD_EQUAL) reduce 563) ((T_MUL_EQUAL) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_OR_EQUAL) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_PLUS_EQUAL) reduce 563) ((T_POW) reduce 563) ((T_POW_EQUAL) reduce 563) ((T_SL) reduce 563) ((T_SL_EQUAL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) ((T_SR_EQUAL) reduce 563) ((T_XOR_EQUAL) reduce 563) (("[") reduce 563) (("^") reduce 563) (("|") reduce 563)) 291 ((("%") reduce 484) ((")") reduce 484) (("*") reduce 484) (("+") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_PIPE) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("^") reduce 484) (("|") reduce 484)) 292 ((("%") reduce 466) ((")") reduce 466) (("*") reduce 466) (("+") reduce 466) (("-") reduce 466) ((".") reduce 466) (("/") reduce 466) (("<") reduce 466) ((">") reduce 466) (("?") reduce 466) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_BOOLEAN_AND) reduce 466) ((T_BOOLEAN_OR) reduce 466) ((T_COALESCE) reduce 466) ((T_INSTANCEOF) reduce 466) ((T_IS_EQUAL) reduce 466) ((T_IS_GREATER_OR_EQUAL) reduce 466) ((T_IS_IDENTICAL) reduce 466) ((T_IS_NOT_EQUAL) reduce 466) ((T_IS_NOT_IDENTICAL) reduce 466) ((T_IS_SMALLER_OR_EQUAL) reduce 466) ((T_LOGICAL_AND) reduce 466) ((T_LOGICAL_OR) reduce 466) ((T_LOGICAL_XOR) reduce 466) ((T_PIPE) reduce 466) ((T_POW) reduce 466) ((T_SL) reduce 466) ((T_SPACESHIP) reduce 466) ((T_SR) reduce 466) (("^") reduce 466) (("|") reduce 466)) 293 ((("%") reduce 488) ((")") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_PIPE) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("^") reduce 488) (("|") reduce 488)) 294 ((("%") reduce 531) (("(") shift) ((")") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("|") reduce 531)) 295 ((("%") reduce 461) (("(") reduce 558) ((")") reduce 461) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PIPE) reduce 461) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") reduce 552) (("^") reduce 461) (("|") reduce 461)) 296 ((("%") reduce 462) ((")") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_PIPE) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 297 ((("%") reduce 476) ((")") reduce 476) (("*") reduce 476) (("+") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_PIPE) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("|") reduce 476)) 298 ((("%") reduce 559) (("(") reduce 559) ((")") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) (("=") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AND_EQUAL) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_COALESCE_EQUAL) reduce 559) ((T_CONCAT_EQUAL) reduce 559) ((T_DEC) reduce 559) ((T_DIV_EQUAL) reduce 559) ((T_INC) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_MINUS_EQUAL) reduce 559) ((T_MOD_EQUAL) reduce 559) ((T_MUL_EQUAL) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_OR_EQUAL) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_PLUS_EQUAL) reduce 559) ((T_POW) reduce 559) ((T_POW_EQUAL) reduce 559) ((T_SL) reduce 559) ((T_SL_EQUAL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) ((T_SR_EQUAL) reduce 559) ((T_XOR_EQUAL) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 299 ((("%") reduce 565) ((")") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 300 ((("%") reduce 403) ((")") reduce 403) (("*") reduce 403) (("+") reduce 403) (("-") reduce 403) ((".") reduce 403) (("/") reduce 403) (("<") reduce 403) (("=") shift) ((">") reduce 403) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_INC) shift) ((T_INSTANCEOF) reduce 403) ((T_IS_EQUAL) reduce 403) ((T_IS_GREATER_OR_EQUAL) reduce 403) ((T_IS_IDENTICAL) reduce 403) ((T_IS_NOT_EQUAL) reduce 403) ((T_IS_NOT_IDENTICAL) reduce 403) ((T_IS_SMALLER_OR_EQUAL) reduce 403) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 403) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 403) ((T_POW_EQUAL) shift) ((T_SL) reduce 403) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 403) ((T_SR) reduce 403) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 548) (("^") reduce 403) (("|") reduce 403)) 303 ((("%") reduce 448) ((")") reduce 448) (("*") reduce 448) (("+") reduce 448) (("-") reduce 448) ((".") reduce 448) (("/") reduce 448) (("<") reduce 448) ((">") reduce 448) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_BOOLEAN_AND) reduce 448) ((T_BOOLEAN_OR) reduce 448) ((T_COALESCE) reduce 448) ((T_INSTANCEOF) reduce 448) ((T_IS_EQUAL) reduce 448) ((T_IS_GREATER_OR_EQUAL) reduce 448) ((T_IS_IDENTICAL) reduce 448) ((T_IS_NOT_EQUAL) reduce 448) ((T_IS_NOT_IDENTICAL) reduce 448) ((T_IS_SMALLER_OR_EQUAL) reduce 448) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_PIPE) reduce 448) ((T_POW) shift) ((T_SL) reduce 448) ((T_SPACESHIP) reduce 448) ((T_SR) reduce 448) (("^") reduce 448) (("|") reduce 448)) 312 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((":") shift) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 318 ((("$") shift) (("(") shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_VARIABLE) shift)) 335 ((("%") shift) ((")") reduce 432) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 432)) 336 ((("%") shift) ((")") reduce 435) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("|") reduce 435)) 337 ((("%") shift) ((")") reduce 444) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 444) (("/") shift) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_PIPE) reduce 444) ((T_POW) shift) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("|") reduce 444)) 338 ((("%") shift) ((")") reduce 458) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 458) (("|") reduce 458)) 339 ((("%") shift) ((")") reduce 443) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 443) (("/") shift) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_PIPE) reduce 443) ((T_POW) shift) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("^") reduce 443) (("|") reduce 443)) 340 ((("%") reduce 440) ((")") reduce 440) (("*") reduce 440) (("+") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_PIPE) reduce 440) ((T_POW) shift) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("|") reduce 440)) 341 ((("%") shift) ((")") reduce 453) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_PIPE) reduce 453) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 453) ((T_SR) shift) (("^") reduce 453) (("|") reduce 453)) 342 ((("%") shift) ((")") reduce 431) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 343 ((("%") shift) ((")") reduce 429) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 344 ((("%") shift) ((")") reduce 430) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 345 ((("%") shift) ((")") reduce 455) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 455) ((T_SR) shift) (("^") reduce 455) (("|") reduce 455)) 346 ((("%") shift) ((")") reduce 450) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 450) (("|") reduce 450)) 347 ((("%") shift) ((")") reduce 452) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 452) (("|") reduce 452)) 348 ((("%") shift) ((")") reduce 449) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 449) (("|") reduce 449)) 349 ((("%") shift) ((")") reduce 457) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 457) ((T_SR) shift) (("^") reduce 457) (("|") reduce 457)) 350 ((("%") shift) ((")") reduce 451) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 451) (("|") reduce 451)) 353 ((("%") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 354 ((("%") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 355 ((("%") reduce 95) ((")") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 356 ((("%") reduce 509) ((")") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 357 ((("%") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 358 ((("%") reduce 568) ((")") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 359 ((("%") reduce 511) ((")") reduce 511) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("|") reduce 511)) 360 ((("%") reduce 459) ((")") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PIPE) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 361 ((("%") reduce 510) ((")") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 362 ((("%") reduce 512) ((")") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("|") reduce 512)) 363 ((("%") reduce 573) ((")") reduce 573) (("*") reduce 573) (("+") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("|") reduce 573)) 364 ((("$") shift) ((T_STRING) shift) ((T_VARIABLE) shift) (("{") shift)) 367 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("]") reduce 545) (("`") shift) (("~") shift)) 378 ((("%") reduce 540) (("*") reduce 540) (("+") reduce 540) (("-") reduce 540) ((".") reduce 540) (("/") reduce 540) (("<") reduce 540) ((">") reduce 540) (("?") reduce 540) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_BOOLEAN_AND) reduce 540) ((T_BOOLEAN_OR) reduce 540) ((T_COALESCE) reduce 540) ((T_INSTANCEOF) reduce 540) ((T_IS_EQUAL) reduce 540) ((T_IS_GREATER_OR_EQUAL) reduce 540) ((T_IS_IDENTICAL) reduce 540) ((T_IS_NOT_EQUAL) reduce 540) ((T_IS_NOT_IDENTICAL) reduce 540) ((T_IS_SMALLER_OR_EQUAL) reduce 540) ((T_LOGICAL_AND) reduce 540) ((T_LOGICAL_OR) reduce 540) ((T_LOGICAL_XOR) reduce 540) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 540) ((T_OBJECT_OPERATOR) reduce 540) ((T_PIPE) reduce 540) ((T_POW) reduce 540) ((T_SL) reduce 540) ((T_SPACESHIP) reduce 540) ((T_SR) reduce 540) (("[") reduce 540) (("]") reduce 540) (("^") reduce 540) (("|") reduce 540)) 380 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_PIPE) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("]") reduce 521) (("^") reduce 521) (("|") reduce 521)) 382 ((("%") reduce 534) (("*") reduce 534) (("+") reduce 534) (("-") reduce 534) ((".") reduce 534) (("/") reduce 534) (("<") reduce 534) ((">") reduce 534) (("?") reduce 534) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_BOOLEAN_AND) reduce 534) ((T_BOOLEAN_OR) reduce 534) ((T_COALESCE) reduce 534) ((T_INSTANCEOF) reduce 534) ((T_IS_EQUAL) reduce 534) ((T_IS_GREATER_OR_EQUAL) reduce 534) ((T_IS_IDENTICAL) reduce 534) ((T_IS_NOT_EQUAL) reduce 534) ((T_IS_NOT_IDENTICAL) reduce 534) ((T_IS_SMALLER_OR_EQUAL) reduce 534) ((T_LOGICAL_AND) reduce 534) ((T_LOGICAL_OR) reduce 534) ((T_LOGICAL_XOR) reduce 534) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PIPE) reduce 534) ((T_POW) reduce 534) ((T_SL) reduce 534) ((T_SPACESHIP) reduce 534) ((T_SR) reduce 534) (("[") reduce 534) (("]") reduce 534) (("^") reduce 534) (("|") reduce 534)) 383 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_PIPE) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("]") reduce 524) (("^") reduce 524) (("|") reduce 524)) 387 ((("%") reduce 517) (("(") shift) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_PIPE) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("]") reduce 517) (("^") reduce 517) (("|") reduce 517)) 388 ((("%") reduce 533) (("*") reduce 533) (("+") reduce 533) (("-") reduce 533) ((".") reduce 533) (("/") reduce 533) (("<") reduce 533) ((">") reduce 533) (("?") reduce 533) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_BOOLEAN_AND) reduce 533) ((T_BOOLEAN_OR) reduce 533) ((T_COALESCE) reduce 533) ((T_INSTANCEOF) reduce 533) ((T_IS_EQUAL) reduce 533) ((T_IS_GREATER_OR_EQUAL) reduce 533) ((T_IS_IDENTICAL) reduce 533) ((T_IS_NOT_EQUAL) reduce 533) ((T_IS_NOT_IDENTICAL) reduce 533) ((T_IS_SMALLER_OR_EQUAL) reduce 533) ((T_LOGICAL_AND) reduce 533) ((T_LOGICAL_OR) reduce 533) ((T_LOGICAL_XOR) reduce 533) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PIPE) reduce 533) ((T_POW) reduce 533) ((T_SL) reduce 533) ((T_SPACESHIP) reduce 533) ((T_SR) reduce 533) (("[") reduce 533) (("]") reduce 533) (("^") reduce 533) (("|") reduce 533)) 389 ((("%") reduce 537) (("*") reduce 537) (("+") reduce 537) (("-") reduce 537) ((".") reduce 537) (("/") reduce 537) (("<") reduce 537) ((">") reduce 537) (("?") reduce 537) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_BOOLEAN_AND) reduce 537) ((T_BOOLEAN_OR) reduce 537) ((T_COALESCE) reduce 537) ((T_INSTANCEOF) reduce 537) ((T_IS_EQUAL) reduce 537) ((T_IS_GREATER_OR_EQUAL) reduce 537) ((T_IS_IDENTICAL) reduce 537) ((T_IS_NOT_EQUAL) reduce 537) ((T_IS_NOT_IDENTICAL) reduce 537) ((T_IS_SMALLER_OR_EQUAL) reduce 537) ((T_LOGICAL_AND) reduce 537) ((T_LOGICAL_OR) reduce 537) ((T_LOGICAL_XOR) reduce 537) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_PIPE) reduce 537) ((T_POW) reduce 537) ((T_SL) reduce 537) ((T_SPACESHIP) reduce 537) ((T_SR) reduce 537) (("[") reduce 537) (("]") reduce 537) (("^") reduce 537) (("|") reduce 537)) 395 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PIPE) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("]") reduce 532) (("^") reduce 532) (("|") reduce 532)) 397 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_PIPE) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("]") reduce 523) (("^") reduce 523) (("|") reduce 523)) 399 ((("%") reduce 536) (("*") reduce 536) (("+") reduce 536) (("-") reduce 536) ((".") reduce 536) (("/") reduce 536) (("<") reduce 536) ((">") reduce 536) (("?") reduce 536) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_BOOLEAN_AND) reduce 536) ((T_BOOLEAN_OR) reduce 536) ((T_COALESCE) reduce 536) ((T_INSTANCEOF) reduce 536) ((T_IS_EQUAL) reduce 536) ((T_IS_GREATER_OR_EQUAL) reduce 536) ((T_IS_IDENTICAL) reduce 536) ((T_IS_NOT_EQUAL) reduce 536) ((T_IS_NOT_IDENTICAL) reduce 536) ((T_IS_SMALLER_OR_EQUAL) reduce 536) ((T_LOGICAL_AND) reduce 536) ((T_LOGICAL_OR) reduce 536) ((T_LOGICAL_XOR) reduce 536) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 536) ((T_OBJECT_OPERATOR) reduce 536) ((T_PIPE) reduce 536) ((T_POW) reduce 536) ((T_SL) reduce 536) ((T_SPACESHIP) reduce 536) ((T_SR) reduce 536) (("[") reduce 536) (("]") reduce 536) (("^") reduce 536) (("|") reduce 536)) 400 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("]") reduce 94) (("^") reduce 94) (("|") reduce 94)) 401 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("]") reduce 93) (("^") reduce 93) (("|") reduce 93)) 402 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("]") reduce 95) (("^") reduce 95) (("|") reduce 95)) 404 ((("%") reduce 539) (("*") reduce 539) (("+") reduce 539) (("-") reduce 539) ((".") reduce 539) (("/") reduce 539) (("<") reduce 539) ((">") reduce 539) (("?") reduce 539) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_BOOLEAN_AND) reduce 539) ((T_BOOLEAN_OR) reduce 539) ((T_COALESCE) reduce 539) ((T_INSTANCEOF) reduce 539) ((T_IS_EQUAL) reduce 539) ((T_IS_GREATER_OR_EQUAL) reduce 539) ((T_IS_IDENTICAL) reduce 539) ((T_IS_NOT_EQUAL) reduce 539) ((T_IS_NOT_IDENTICAL) reduce 539) ((T_IS_SMALLER_OR_EQUAL) reduce 539) ((T_LOGICAL_AND) reduce 539) ((T_LOGICAL_OR) reduce 539) ((T_LOGICAL_XOR) reduce 539) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 539) ((T_OBJECT_OPERATOR) reduce 539) ((T_PIPE) reduce 539) ((T_POW) reduce 539) ((T_SL) reduce 539) ((T_SPACESHIP) reduce 539) ((T_SR) reduce 539) (("[") reduce 539) (("]") reduce 539) (("^") reduce 539) (("|") reduce 539)) 407 ((("%") reduce 538) (("*") reduce 538) (("+") reduce 538) (("-") reduce 538) ((".") reduce 538) (("/") reduce 538) (("<") reduce 538) ((">") reduce 538) (("?") reduce 538) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_BOOLEAN_AND) reduce 538) ((T_BOOLEAN_OR) reduce 538) ((T_COALESCE) reduce 538) ((T_INSTANCEOF) reduce 538) ((T_IS_EQUAL) reduce 538) ((T_IS_GREATER_OR_EQUAL) reduce 538) ((T_IS_IDENTICAL) reduce 538) ((T_IS_NOT_EQUAL) reduce 538) ((T_IS_NOT_IDENTICAL) reduce 538) ((T_IS_SMALLER_OR_EQUAL) reduce 538) ((T_LOGICAL_AND) reduce 538) ((T_LOGICAL_OR) reduce 538) ((T_LOGICAL_XOR) reduce 538) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 538) ((T_OBJECT_OPERATOR) reduce 538) ((T_PIPE) reduce 538) ((T_POW) reduce 538) ((T_SL) reduce 538) ((T_SPACESHIP) reduce 538) ((T_SR) reduce 538) (("[") reduce 538) (("]") reduce 538) (("^") reduce 538) (("|") reduce 538)) 413 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("]") reduce 92) (("^") reduce 92) (("|") reduce 92)) 416 ((("%") reduce 535) (("*") reduce 535) (("+") reduce 535) (("-") reduce 535) ((".") reduce 535) (("/") reduce 535) (("<") reduce 535) ((">") reduce 535) (("?") reduce 535) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_BOOLEAN_AND) reduce 535) ((T_BOOLEAN_OR) reduce 535) ((T_COALESCE) reduce 535) ((T_INSTANCEOF) reduce 535) ((T_IS_EQUAL) reduce 535) ((T_IS_GREATER_OR_EQUAL) reduce 535) ((T_IS_IDENTICAL) reduce 535) ((T_IS_NOT_EQUAL) reduce 535) ((T_IS_NOT_IDENTICAL) reduce 535) ((T_IS_SMALLER_OR_EQUAL) reduce 535) ((T_LOGICAL_AND) reduce 535) ((T_LOGICAL_OR) reduce 535) ((T_LOGICAL_XOR) reduce 535) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PIPE) reduce 535) ((T_POW) reduce 535) ((T_SL) reduce 535) ((T_SPACESHIP) reduce 535) ((T_SR) reduce 535) (("[") reduce 535) (("]") reduce 535) (("^") reduce 535) (("|") reduce 535)) 418 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) (("=") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AND_EQUAL) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_COALESCE_EQUAL) reduce 568) ((T_CONCAT_EQUAL) reduce 568) ((T_DEC) reduce 568) ((T_DIV_EQUAL) reduce 568) ((T_INC) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_MINUS_EQUAL) reduce 568) ((T_MOD_EQUAL) reduce 568) ((T_MUL_EQUAL) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_OR_EQUAL) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_PLUS_EQUAL) reduce 568) ((T_POW) reduce 568) ((T_POW_EQUAL) reduce 568) ((T_SL) reduce 568) ((T_SL_EQUAL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) ((T_SR_EQUAL) reduce 568) ((T_XOR_EQUAL) reduce 568) (("[") reduce 568) (("]") reduce 568) (("^") reduce 568) (("|") reduce 568)) 419 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 479) (("(") shift) (("*") reduce 479) (("+") shift) (("-") shift) ((".") reduce 479) (("/") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 479) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 479) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PIPE) reduce 479) ((T_POW) reduce 479) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("]") reduce 479) (("^") reduce 479) (("`") shift) (("|") reduce 479) (("~") shift)) 426 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) (("<") reduce 564) (("=") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AND_EQUAL) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_COALESCE_EQUAL) reduce 564) ((T_CONCAT_EQUAL) reduce 564) ((T_DEC) reduce 564) ((T_DIV_EQUAL) reduce 564) ((T_INC) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_MINUS_EQUAL) reduce 564) ((T_MOD_EQUAL) reduce 564) ((T_MUL_EQUAL) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_OR_EQUAL) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_PLUS_EQUAL) reduce 564) ((T_POW) reduce 564) ((T_POW_EQUAL) reduce 564) ((T_SL) reduce 564) ((T_SL_EQUAL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) ((T_SR_EQUAL) reduce 564) ((T_XOR_EQUAL) reduce 564) (("[") reduce 564) (("]") reduce 564) (("^") reduce 564) (("|") reduce 564)) 427 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PIPE) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 551) (("]") reduce 530) (("^") reduce 530) (("|") reduce 530)) 429 ((("%") reduce 529) (("*") reduce 529) (("+") reduce 529) (("-") reduce 529) ((".") reduce 529) (("/") reduce 529) (("<") reduce 529) ((">") reduce 529) (("?") reduce 529) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_BOOLEAN_AND) reduce 529) ((T_BOOLEAN_OR) reduce 529) ((T_COALESCE) reduce 529) ((T_INSTANCEOF) reduce 529) ((T_IS_EQUAL) reduce 529) ((T_IS_GREATER_OR_EQUAL) reduce 529) ((T_IS_IDENTICAL) reduce 529) ((T_IS_NOT_EQUAL) reduce 529) ((T_IS_NOT_IDENTICAL) reduce 529) ((T_IS_SMALLER_OR_EQUAL) reduce 529) ((T_LOGICAL_AND) reduce 529) ((T_LOGICAL_OR) reduce 529) ((T_LOGICAL_XOR) reduce 529) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PIPE) reduce 529) ((T_POW) reduce 529) ((T_SL) reduce 529) ((T_SPACESHIP) reduce 529) ((T_SR) reduce 529) (("[") reduce 554) (("]") reduce 529) (("^") reduce 529) (("|") reduce 529)) 430 ((("%") reduce 528) (("(") reduce 557) (("*") reduce 528) (("+") reduce 528) (("-") reduce 528) ((".") reduce 528) (("/") reduce 528) (("<") reduce 528) ((">") reduce 528) (("?") reduce 528) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_BOOLEAN_AND) reduce 528) ((T_BOOLEAN_OR) reduce 528) ((T_COALESCE) reduce 528) ((T_INSTANCEOF) reduce 528) ((T_IS_EQUAL) reduce 528) ((T_IS_GREATER_OR_EQUAL) reduce 528) ((T_IS_IDENTICAL) reduce 528) ((T_IS_NOT_EQUAL) reduce 528) ((T_IS_NOT_IDENTICAL) reduce 528) ((T_IS_SMALLER_OR_EQUAL) reduce 528) ((T_LOGICAL_AND) reduce 528) ((T_LOGICAL_OR) reduce 528) ((T_LOGICAL_XOR) reduce 528) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PIPE) reduce 528) ((T_POW) reduce 528) ((T_SL) reduce 528) ((T_SPACESHIP) reduce 528) ((T_SR) reduce 528) (("[") reduce 550) (("]") reduce 528) (("^") reduce 528) (("|") reduce 528)) 431 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 546) (("^") shift) (("|") shift)) 434 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) (("<") reduce 563) (("=") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AND_EQUAL) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_COALESCE_EQUAL) reduce 563) ((T_CONCAT_EQUAL) reduce 563) ((T_DEC) reduce 563) ((T_DIV_EQUAL) reduce 563) ((T_INC) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_MINUS_EQUAL) reduce 563) ((T_MOD_EQUAL) reduce 563) ((T_MUL_EQUAL) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_OR_EQUAL) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_PLUS_EQUAL) reduce 563) ((T_POW) reduce 563) ((T_POW_EQUAL) reduce 563) ((T_SL) reduce 563) ((T_SL_EQUAL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) ((T_SR_EQUAL) reduce 563) ((T_XOR_EQUAL) reduce 563) (("[") reduce 563) (("]") reduce 563) (("^") reduce 563) (("|") reduce 563)) 435 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_PIPE) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("]") reduce 484) (("^") reduce 484) (("|") reduce 484)) 436 ((("%") reduce 466) (("*") reduce 466) (("+") reduce 466) (("-") reduce 466) ((".") reduce 466) (("/") reduce 466) (("<") reduce 466) ((">") reduce 466) (("?") reduce 466) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_BOOLEAN_AND) reduce 466) ((T_BOOLEAN_OR) reduce 466) ((T_COALESCE) reduce 466) ((T_INSTANCEOF) reduce 466) ((T_IS_EQUAL) reduce 466) ((T_IS_GREATER_OR_EQUAL) reduce 466) ((T_IS_IDENTICAL) reduce 466) ((T_IS_NOT_EQUAL) reduce 466) ((T_IS_NOT_IDENTICAL) reduce 466) ((T_IS_SMALLER_OR_EQUAL) reduce 466) ((T_LOGICAL_AND) reduce 466) ((T_LOGICAL_OR) reduce 466) ((T_LOGICAL_XOR) reduce 466) ((T_PIPE) reduce 466) ((T_POW) reduce 466) ((T_SL) reduce 466) ((T_SPACESHIP) reduce 466) ((T_SR) reduce 466) (("]") reduce 466) (("^") reduce 466) (("|") reduce 466)) 437 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_PIPE) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("]") reduce 488) (("^") reduce 488) (("|") reduce 488)) 438 ((("%") reduce 531) (("(") shift) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("]") reduce 531) (("^") reduce 531) (("|") reduce 531)) 439 ((("%") reduce 461) (("(") reduce 558) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PIPE) reduce 461) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") reduce 552) (("]") reduce 461) (("^") reduce 461) (("|") reduce 461)) 440 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_PIPE) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("]") reduce 462) (("^") reduce 462) (("|") reduce 462)) 441 ((("]") shift)) 442 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_PIPE) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("]") reduce 476) (("^") reduce 476) (("|") reduce 476)) 443 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) (("=") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AND_EQUAL) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_COALESCE_EQUAL) reduce 559) ((T_CONCAT_EQUAL) reduce 559) ((T_DEC) reduce 559) ((T_DIV_EQUAL) reduce 559) ((T_INC) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_MINUS_EQUAL) reduce 559) ((T_MOD_EQUAL) reduce 559) ((T_MUL_EQUAL) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_OR_EQUAL) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_PLUS_EQUAL) reduce 559) ((T_POW) reduce 559) ((T_POW_EQUAL) reduce 559) ((T_SL) reduce 559) ((T_SL_EQUAL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) ((T_SR_EQUAL) reduce 559) ((T_XOR_EQUAL) reduce 559) (("[") reduce 559) (("]") reduce 559) (("^") reduce 559) (("|") reduce 559)) 444 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("]") reduce 565) (("^") reduce 565) (("|") reduce 565)) 445 ((("%") reduce 403) (("*") reduce 403) (("+") reduce 403) (("-") reduce 403) ((".") reduce 403) (("/") reduce 403) (("<") reduce 403) (("=") shift) ((">") reduce 403) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_INC) shift) ((T_INSTANCEOF) reduce 403) ((T_IS_EQUAL) reduce 403) ((T_IS_GREATER_OR_EQUAL) reduce 403) ((T_IS_IDENTICAL) reduce 403) ((T_IS_NOT_EQUAL) reduce 403) ((T_IS_NOT_IDENTICAL) reduce 403) ((T_IS_SMALLER_OR_EQUAL) reduce 403) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 403) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 403) ((T_POW_EQUAL) shift) ((T_SL) reduce 403) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 403) ((T_SR) reduce 403) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 548) (("]") reduce 403) (("^") reduce 403) (("|") reduce 403)) 448 ((("%") reduce 448) (("*") reduce 448) (("+") reduce 448) (("-") reduce 448) ((".") reduce 448) (("/") reduce 448) (("<") reduce 448) ((">") reduce 448) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_BOOLEAN_AND) reduce 448) ((T_BOOLEAN_OR) reduce 448) ((T_COALESCE) reduce 448) ((T_INSTANCEOF) reduce 448) ((T_IS_EQUAL) reduce 448) ((T_IS_GREATER_OR_EQUAL) reduce 448) ((T_IS_IDENTICAL) reduce 448) ((T_IS_NOT_EQUAL) reduce 448) ((T_IS_NOT_IDENTICAL) reduce 448) ((T_IS_SMALLER_OR_EQUAL) reduce 448) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_PIPE) reduce 448) ((T_POW) shift) ((T_SL) reduce 448) ((T_SPACESHIP) reduce 448) ((T_SR) reduce 448) (("]") reduce 448) (("^") reduce 448) (("|") reduce 448)) 480 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 432) (("^") shift) (("|") reduce 432)) 481 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 435) (("^") reduce 435) (("|") reduce 435)) 482 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 444) (("/") shift) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_PIPE) reduce 444) ((T_POW) shift) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("]") reduce 444) (("^") reduce 444) (("|") reduce 444)) 483 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 458) (("^") reduce 458) (("|") reduce 458)) 484 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 443) (("/") shift) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_PIPE) reduce 443) ((T_POW) shift) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("]") reduce 443) (("^") reduce 443) (("|") reduce 443)) 485 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_PIPE) reduce 440) ((T_POW) shift) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("]") reduce 440) (("^") reduce 440) (("|") reduce 440)) 486 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_PIPE) reduce 453) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 453) ((T_SR) shift) (("]") reduce 453) (("^") reduce 453) (("|") reduce 453)) 487 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 431) (("^") shift) (("|") shift)) 488 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 429) (("^") shift) (("|") shift)) 489 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 430) (("^") shift) (("|") shift)) 490 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 455) ((T_SR) shift) (("]") reduce 455) (("^") reduce 455) (("|") reduce 455)) 491 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 450) (("^") reduce 450) (("|") reduce 450)) 492 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 452) (("^") reduce 452) (("|") reduce 452)) 493 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 449) (("^") reduce 449) (("|") reduce 449)) 494 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 457) ((T_SR) shift) (("]") reduce 457) (("^") reduce 457) (("|") reduce 457)) 495 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 451) (("^") reduce 451) (("|") reduce 451)) 498 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("]") reduce 94) (("^") reduce 94) (("|") reduce 94)) 499 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("]") reduce 93) (("^") reduce 93) (("|") reduce 93)) 500 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("]") reduce 95) (("^") reduce 95) (("|") reduce 95)) 501 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("]") reduce 509) (("^") reduce 509) (("|") reduce 509)) 502 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("]") reduce 92) (("^") reduce 92) (("|") reduce 92)) 503 ((("%") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("]") reduce 568) (("^") reduce 568) (("|") reduce 568)) 504 ((("%") reduce 511) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("]") reduce 511) (("^") reduce 511) (("|") reduce 511)) 505 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PIPE) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("]") reduce 459) (("^") reduce 459) (("|") reduce 459)) 506 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("]") reduce 510) (("^") reduce 510) (("|") reduce 510)) 507 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("]") reduce 512) (("^") reduce 512) (("|") reduce 512)) 508 ((("%") reduce 573) (("*") reduce 573) (("+") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("]") reduce 573) (("^") reduce 573) (("|") reduce 573)) 514 ((("%") reduce 574) (("*") reduce 574) (("+") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("]") reduce 574) (("^") reduce 574) (("|") reduce 574)) 515 ((("%") reduce 578) (("*") reduce 578) (("+") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("]") reduce 578) (("^") reduce 578) (("|") reduce 578)) 516 ((("%") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("]") reduce 582) (("^") reduce 582) (("|") reduce 582)) 517 ((("%") reduce 575) (("*") reduce 575) (("+") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("]") reduce 575) (("^") reduce 575) (("|") reduce 575)) 518 ((("%") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("]") reduce 584) (("^") reduce 584) (("|") reduce 584)) 530 ((("%") reduce 540) (("*") reduce 540) (("+") reduce 540) (("-") reduce 540) ((".") reduce 540) (("/") reduce 540) (("<") reduce 540) ((">") reduce 540) (("?") reduce 540) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_BOOLEAN_AND) reduce 540) ((T_BOOLEAN_OR) reduce 540) ((T_COALESCE) reduce 540) ((T_INSTANCEOF) reduce 540) ((T_IS_EQUAL) reduce 540) ((T_IS_GREATER_OR_EQUAL) reduce 540) ((T_IS_IDENTICAL) reduce 540) ((T_IS_NOT_EQUAL) reduce 540) ((T_IS_NOT_IDENTICAL) reduce 540) ((T_IS_SMALLER_OR_EQUAL) reduce 540) ((T_LOGICAL_AND) reduce 540) ((T_LOGICAL_OR) reduce 540) ((T_LOGICAL_XOR) reduce 540) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 540) ((T_OBJECT_OPERATOR) reduce 540) ((T_PIPE) reduce 540) ((T_POW) reduce 540) ((T_SL) reduce 540) ((T_SPACESHIP) reduce 540) ((T_SR) reduce 540) (("[") reduce 540) (("^") reduce 540) (("|") reduce 540) (("}") reduce 540)) 532 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_PIPE) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("|") reduce 521) (("}") reduce 521)) 534 ((("%") reduce 534) (("*") reduce 534) (("+") reduce 534) (("-") reduce 534) ((".") reduce 534) (("/") reduce 534) (("<") reduce 534) ((">") reduce 534) (("?") reduce 534) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_BOOLEAN_AND) reduce 534) ((T_BOOLEAN_OR) reduce 534) ((T_COALESCE) reduce 534) ((T_INSTANCEOF) reduce 534) ((T_IS_EQUAL) reduce 534) ((T_IS_GREATER_OR_EQUAL) reduce 534) ((T_IS_IDENTICAL) reduce 534) ((T_IS_NOT_EQUAL) reduce 534) ((T_IS_NOT_IDENTICAL) reduce 534) ((T_IS_SMALLER_OR_EQUAL) reduce 534) ((T_LOGICAL_AND) reduce 534) ((T_LOGICAL_OR) reduce 534) ((T_LOGICAL_XOR) reduce 534) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PIPE) reduce 534) ((T_POW) reduce 534) ((T_SL) reduce 534) ((T_SPACESHIP) reduce 534) ((T_SR) reduce 534) (("[") reduce 534) (("^") reduce 534) (("|") reduce 534) (("}") reduce 534)) 535 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_PIPE) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("^") reduce 524) (("|") reduce 524) (("}") reduce 524)) 539 ((("%") reduce 517) (("(") shift) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_PIPE) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("^") reduce 517) (("|") reduce 517) (("}") reduce 517)) 540 ((("%") reduce 533) (("*") reduce 533) (("+") reduce 533) (("-") reduce 533) ((".") reduce 533) (("/") reduce 533) (("<") reduce 533) ((">") reduce 533) (("?") reduce 533) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_BOOLEAN_AND) reduce 533) ((T_BOOLEAN_OR) reduce 533) ((T_COALESCE) reduce 533) ((T_INSTANCEOF) reduce 533) ((T_IS_EQUAL) reduce 533) ((T_IS_GREATER_OR_EQUAL) reduce 533) ((T_IS_IDENTICAL) reduce 533) ((T_IS_NOT_EQUAL) reduce 533) ((T_IS_NOT_IDENTICAL) reduce 533) ((T_IS_SMALLER_OR_EQUAL) reduce 533) ((T_LOGICAL_AND) reduce 533) ((T_LOGICAL_OR) reduce 533) ((T_LOGICAL_XOR) reduce 533) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PIPE) reduce 533) ((T_POW) reduce 533) ((T_SL) reduce 533) ((T_SPACESHIP) reduce 533) ((T_SR) reduce 533) (("[") reduce 533) (("^") reduce 533) (("|") reduce 533) (("}") reduce 533)) 541 ((("%") reduce 537) (("*") reduce 537) (("+") reduce 537) (("-") reduce 537) ((".") reduce 537) (("/") reduce 537) (("<") reduce 537) ((">") reduce 537) (("?") reduce 537) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_BOOLEAN_AND) reduce 537) ((T_BOOLEAN_OR) reduce 537) ((T_COALESCE) reduce 537) ((T_INSTANCEOF) reduce 537) ((T_IS_EQUAL) reduce 537) ((T_IS_GREATER_OR_EQUAL) reduce 537) ((T_IS_IDENTICAL) reduce 537) ((T_IS_NOT_EQUAL) reduce 537) ((T_IS_NOT_IDENTICAL) reduce 537) ((T_IS_SMALLER_OR_EQUAL) reduce 537) ((T_LOGICAL_AND) reduce 537) ((T_LOGICAL_OR) reduce 537) ((T_LOGICAL_XOR) reduce 537) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_PIPE) reduce 537) ((T_POW) reduce 537) ((T_SL) reduce 537) ((T_SPACESHIP) reduce 537) ((T_SR) reduce 537) (("[") reduce 537) (("^") reduce 537) (("|") reduce 537) (("}") reduce 537)) 547 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PIPE) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("|") reduce 532) (("}") reduce 532)) 549 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_PIPE) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("^") reduce 523) (("|") reduce 523) (("}") reduce 523)) 551 ((("%") reduce 536) (("*") reduce 536) (("+") reduce 536) (("-") reduce 536) ((".") reduce 536) (("/") reduce 536) (("<") reduce 536) ((">") reduce 536) (("?") reduce 536) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_BOOLEAN_AND) reduce 536) ((T_BOOLEAN_OR) reduce 536) ((T_COALESCE) reduce 536) ((T_INSTANCEOF) reduce 536) ((T_IS_EQUAL) reduce 536) ((T_IS_GREATER_OR_EQUAL) reduce 536) ((T_IS_IDENTICAL) reduce 536) ((T_IS_NOT_EQUAL) reduce 536) ((T_IS_NOT_IDENTICAL) reduce 536) ((T_IS_SMALLER_OR_EQUAL) reduce 536) ((T_LOGICAL_AND) reduce 536) ((T_LOGICAL_OR) reduce 536) ((T_LOGICAL_XOR) reduce 536) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 536) ((T_OBJECT_OPERATOR) reduce 536) ((T_PIPE) reduce 536) ((T_POW) reduce 536) ((T_SL) reduce 536) ((T_SPACESHIP) reduce 536) ((T_SR) reduce 536) (("[") reduce 536) (("^") reduce 536) (("|") reduce 536) (("}") reduce 536)) 552 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("|") reduce 94) (("}") reduce 94)) 553 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("|") reduce 93) (("}") reduce 93)) 554 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("|") reduce 95) (("}") reduce 95)) 556 ((("%") reduce 539) (("*") reduce 539) (("+") reduce 539) (("-") reduce 539) ((".") reduce 539) (("/") reduce 539) (("<") reduce 539) ((">") reduce 539) (("?") reduce 539) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_BOOLEAN_AND) reduce 539) ((T_BOOLEAN_OR) reduce 539) ((T_COALESCE) reduce 539) ((T_INSTANCEOF) reduce 539) ((T_IS_EQUAL) reduce 539) ((T_IS_GREATER_OR_EQUAL) reduce 539) ((T_IS_IDENTICAL) reduce 539) ((T_IS_NOT_EQUAL) reduce 539) ((T_IS_NOT_IDENTICAL) reduce 539) ((T_IS_SMALLER_OR_EQUAL) reduce 539) ((T_LOGICAL_AND) reduce 539) ((T_LOGICAL_OR) reduce 539) ((T_LOGICAL_XOR) reduce 539) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 539) ((T_OBJECT_OPERATOR) reduce 539) ((T_PIPE) reduce 539) ((T_POW) reduce 539) ((T_SL) reduce 539) ((T_SPACESHIP) reduce 539) ((T_SR) reduce 539) (("[") reduce 539) (("^") reduce 539) (("|") reduce 539) (("}") reduce 539)) 559 ((("%") reduce 538) (("*") reduce 538) (("+") reduce 538) (("-") reduce 538) ((".") reduce 538) (("/") reduce 538) (("<") reduce 538) ((">") reduce 538) (("?") reduce 538) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_BOOLEAN_AND) reduce 538) ((T_BOOLEAN_OR) reduce 538) ((T_COALESCE) reduce 538) ((T_INSTANCEOF) reduce 538) ((T_IS_EQUAL) reduce 538) ((T_IS_GREATER_OR_EQUAL) reduce 538) ((T_IS_IDENTICAL) reduce 538) ((T_IS_NOT_EQUAL) reduce 538) ((T_IS_NOT_IDENTICAL) reduce 538) ((T_IS_SMALLER_OR_EQUAL) reduce 538) ((T_LOGICAL_AND) reduce 538) ((T_LOGICAL_OR) reduce 538) ((T_LOGICAL_XOR) reduce 538) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 538) ((T_OBJECT_OPERATOR) reduce 538) ((T_PIPE) reduce 538) ((T_POW) reduce 538) ((T_SL) reduce 538) ((T_SPACESHIP) reduce 538) ((T_SR) reduce 538) (("[") reduce 538) (("^") reduce 538) (("|") reduce 538) (("}") reduce 538)) 565 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92) (("}") reduce 92)) 568 ((("%") reduce 535) (("*") reduce 535) (("+") reduce 535) (("-") reduce 535) ((".") reduce 535) (("/") reduce 535) (("<") reduce 535) ((">") reduce 535) (("?") reduce 535) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_BOOLEAN_AND) reduce 535) ((T_BOOLEAN_OR) reduce 535) ((T_COALESCE) reduce 535) ((T_INSTANCEOF) reduce 535) ((T_IS_EQUAL) reduce 535) ((T_IS_GREATER_OR_EQUAL) reduce 535) ((T_IS_IDENTICAL) reduce 535) ((T_IS_NOT_EQUAL) reduce 535) ((T_IS_NOT_IDENTICAL) reduce 535) ((T_IS_SMALLER_OR_EQUAL) reduce 535) ((T_LOGICAL_AND) reduce 535) ((T_LOGICAL_OR) reduce 535) ((T_LOGICAL_XOR) reduce 535) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PIPE) reduce 535) ((T_POW) reduce 535) ((T_SL) reduce 535) ((T_SPACESHIP) reduce 535) ((T_SR) reduce 535) (("[") reduce 535) (("^") reduce 535) (("|") reduce 535) (("}") reduce 535)) 570 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) (("=") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AND_EQUAL) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_COALESCE_EQUAL) reduce 568) ((T_CONCAT_EQUAL) reduce 568) ((T_DEC) reduce 568) ((T_DIV_EQUAL) reduce 568) ((T_INC) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_MINUS_EQUAL) reduce 568) ((T_MOD_EQUAL) reduce 568) ((T_MUL_EQUAL) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_OR_EQUAL) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_PLUS_EQUAL) reduce 568) ((T_POW) reduce 568) ((T_POW_EQUAL) reduce 568) ((T_SL) reduce 568) ((T_SL_EQUAL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) ((T_SR_EQUAL) reduce 568) ((T_XOR_EQUAL) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568) (("}") reduce 568)) 571 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 479) (("(") shift) (("*") reduce 479) (("+") shift) (("-") shift) ((".") reduce 479) (("/") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 479) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 479) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PIPE) reduce 479) ((T_POW) reduce 479) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 479) (("`") shift) (("|") reduce 479) (("}") reduce 479) (("~") shift)) 578 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) (("<") reduce 564) (("=") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AND_EQUAL) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_COALESCE_EQUAL) reduce 564) ((T_CONCAT_EQUAL) reduce 564) ((T_DEC) reduce 564) ((T_DIV_EQUAL) reduce 564) ((T_INC) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_MINUS_EQUAL) reduce 564) ((T_MOD_EQUAL) reduce 564) ((T_MUL_EQUAL) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_OR_EQUAL) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_PLUS_EQUAL) reduce 564) ((T_POW) reduce 564) ((T_POW_EQUAL) reduce 564) ((T_SL) reduce 564) ((T_SL_EQUAL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) ((T_SR_EQUAL) reduce 564) ((T_XOR_EQUAL) reduce 564) (("[") reduce 564) (("^") reduce 564) (("|") reduce 564) (("}") reduce 564)) 579 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PIPE) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 551) (("^") reduce 530) (("|") reduce 530) (("}") reduce 530)) 581 ((("%") reduce 529) (("*") reduce 529) (("+") reduce 529) (("-") reduce 529) ((".") reduce 529) (("/") reduce 529) (("<") reduce 529) ((">") reduce 529) (("?") reduce 529) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_BOOLEAN_AND) reduce 529) ((T_BOOLEAN_OR) reduce 529) ((T_COALESCE) reduce 529) ((T_INSTANCEOF) reduce 529) ((T_IS_EQUAL) reduce 529) ((T_IS_GREATER_OR_EQUAL) reduce 529) ((T_IS_IDENTICAL) reduce 529) ((T_IS_NOT_EQUAL) reduce 529) ((T_IS_NOT_IDENTICAL) reduce 529) ((T_IS_SMALLER_OR_EQUAL) reduce 529) ((T_LOGICAL_AND) reduce 529) ((T_LOGICAL_OR) reduce 529) ((T_LOGICAL_XOR) reduce 529) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PIPE) reduce 529) ((T_POW) reduce 529) ((T_SL) reduce 529) ((T_SPACESHIP) reduce 529) ((T_SR) reduce 529) (("[") reduce 554) (("^") reduce 529) (("|") reduce 529) (("}") reduce 529)) 582 ((("%") reduce 528) (("(") reduce 557) (("*") reduce 528) (("+") reduce 528) (("-") reduce 528) ((".") reduce 528) (("/") reduce 528) (("<") reduce 528) ((">") reduce 528) (("?") reduce 528) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_BOOLEAN_AND) reduce 528) ((T_BOOLEAN_OR) reduce 528) ((T_COALESCE) reduce 528) ((T_INSTANCEOF) reduce 528) ((T_IS_EQUAL) reduce 528) ((T_IS_GREATER_OR_EQUAL) reduce 528) ((T_IS_IDENTICAL) reduce 528) ((T_IS_NOT_EQUAL) reduce 528) ((T_IS_NOT_IDENTICAL) reduce 528) ((T_IS_SMALLER_OR_EQUAL) reduce 528) ((T_LOGICAL_AND) reduce 528) ((T_LOGICAL_OR) reduce 528) ((T_LOGICAL_XOR) reduce 528) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PIPE) reduce 528) ((T_POW) reduce 528) ((T_SL) reduce 528) ((T_SPACESHIP) reduce 528) ((T_SR) reduce 528) (("[") reduce 550) (("^") reduce 528) (("|") reduce 528) (("}") reduce 528)) 583 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") shift)) 586 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) (("<") reduce 563) (("=") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AND_EQUAL) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_COALESCE_EQUAL) reduce 563) ((T_CONCAT_EQUAL) reduce 563) ((T_DEC) reduce 563) ((T_DIV_EQUAL) reduce 563) ((T_INC) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_MINUS_EQUAL) reduce 563) ((T_MOD_EQUAL) reduce 563) ((T_MUL_EQUAL) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_OR_EQUAL) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_PLUS_EQUAL) reduce 563) ((T_POW) reduce 563) ((T_POW_EQUAL) reduce 563) ((T_SL) reduce 563) ((T_SL_EQUAL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) ((T_SR_EQUAL) reduce 563) ((T_XOR_EQUAL) reduce 563) (("[") reduce 563) (("^") reduce 563) (("|") reduce 563) (("}") reduce 563)) 587 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_PIPE) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("^") reduce 484) (("|") reduce 484) (("}") reduce 484)) 588 ((("%") reduce 466) (("*") reduce 466) (("+") reduce 466) (("-") reduce 466) ((".") reduce 466) (("/") reduce 466) (("<") reduce 466) ((">") reduce 466) (("?") reduce 466) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_BOOLEAN_AND) reduce 466) ((T_BOOLEAN_OR) reduce 466) ((T_COALESCE) reduce 466) ((T_INSTANCEOF) reduce 466) ((T_IS_EQUAL) reduce 466) ((T_IS_GREATER_OR_EQUAL) reduce 466) ((T_IS_IDENTICAL) reduce 466) ((T_IS_NOT_EQUAL) reduce 466) ((T_IS_NOT_IDENTICAL) reduce 466) ((T_IS_SMALLER_OR_EQUAL) reduce 466) ((T_LOGICAL_AND) reduce 466) ((T_LOGICAL_OR) reduce 466) ((T_LOGICAL_XOR) reduce 466) ((T_PIPE) reduce 466) ((T_POW) reduce 466) ((T_SL) reduce 466) ((T_SPACESHIP) reduce 466) ((T_SR) reduce 466) (("^") reduce 466) (("|") reduce 466) (("}") reduce 466)) 589 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_PIPE) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("^") reduce 488) (("|") reduce 488) (("}") reduce 488)) 590 ((("%") reduce 531) (("(") shift) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("|") reduce 531) (("}") reduce 531)) 591 ((("%") reduce 461) (("(") reduce 558) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PIPE) reduce 461) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") reduce 552) (("^") reduce 461) (("|") reduce 461) (("}") reduce 461)) 592 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_PIPE) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462) (("}") reduce 462)) 593 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_PIPE) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("|") reduce 476) (("}") reduce 476)) 594 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) (("=") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AND_EQUAL) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_COALESCE_EQUAL) reduce 559) ((T_CONCAT_EQUAL) reduce 559) ((T_DEC) reduce 559) ((T_DIV_EQUAL) reduce 559) ((T_INC) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_MINUS_EQUAL) reduce 559) ((T_MOD_EQUAL) reduce 559) ((T_MUL_EQUAL) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_OR_EQUAL) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_PLUS_EQUAL) reduce 559) ((T_POW) reduce 559) ((T_POW_EQUAL) reduce 559) ((T_SL) reduce 559) ((T_SL_EQUAL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) ((T_SR_EQUAL) reduce 559) ((T_XOR_EQUAL) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559) (("}") reduce 559)) 595 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565) (("}") reduce 565)) 596 ((("%") reduce 403) (("*") reduce 403) (("+") reduce 403) (("-") reduce 403) ((".") reduce 403) (("/") reduce 403) (("<") reduce 403) (("=") shift) ((">") reduce 403) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_INC) shift) ((T_INSTANCEOF) reduce 403) ((T_IS_EQUAL) reduce 403) ((T_IS_GREATER_OR_EQUAL) reduce 403) ((T_IS_IDENTICAL) reduce 403) ((T_IS_NOT_EQUAL) reduce 403) ((T_IS_NOT_IDENTICAL) reduce 403) ((T_IS_SMALLER_OR_EQUAL) reduce 403) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 403) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 403) ((T_POW_EQUAL) shift) ((T_SL) reduce 403) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 403) ((T_SR) reduce 403) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 548) (("^") reduce 403) (("|") reduce 403) (("}") reduce 403)) 599 ((("%") reduce 448) (("*") reduce 448) (("+") reduce 448) (("-") reduce 448) ((".") reduce 448) (("/") reduce 448) (("<") reduce 448) ((">") reduce 448) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_BOOLEAN_AND) reduce 448) ((T_BOOLEAN_OR) reduce 448) ((T_COALESCE) reduce 448) ((T_INSTANCEOF) reduce 448) ((T_IS_EQUAL) reduce 448) ((T_IS_GREATER_OR_EQUAL) reduce 448) ((T_IS_IDENTICAL) reduce 448) ((T_IS_NOT_EQUAL) reduce 448) ((T_IS_NOT_IDENTICAL) reduce 448) ((T_IS_SMALLER_OR_EQUAL) reduce 448) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_PIPE) reduce 448) ((T_POW) shift) ((T_SL) reduce 448) ((T_SPACESHIP) reduce 448) ((T_SR) reduce 448) (("^") reduce 448) (("|") reduce 448) (("}") reduce 448)) 631 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 432) (("}") reduce 432)) 632 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("|") reduce 435) (("}") reduce 435)) 633 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 444) (("/") shift) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_PIPE) reduce 444) ((T_POW) shift) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("|") reduce 444) (("}") reduce 444)) 634 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 458) (("|") reduce 458) (("}") reduce 458)) 635 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 443) (("/") shift) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_PIPE) reduce 443) ((T_POW) shift) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("^") reduce 443) (("|") reduce 443) (("}") reduce 443)) 636 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_PIPE) reduce 440) ((T_POW) shift) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("|") reduce 440) (("}") reduce 440)) 637 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_PIPE) reduce 453) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 453) ((T_SR) shift) (("^") reduce 453) (("|") reduce 453) (("}") reduce 453)) 638 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 431)) 639 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 429)) 640 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 430)) 641 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 455) ((T_SR) shift) (("^") reduce 455) (("|") reduce 455) (("}") reduce 455)) 642 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 450) (("|") reduce 450) (("}") reduce 450)) 643 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 452) (("|") reduce 452) (("}") reduce 452)) 644 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 449) (("|") reduce 449) (("}") reduce 449)) 645 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 457) ((T_SR) shift) (("^") reduce 457) (("|") reduce 457) (("}") reduce 457)) 646 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 451) (("|") reduce 451) (("}") reduce 451)) 649 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94) (("}") reduce 94)) 650 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93) (("}") reduce 93)) 651 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95) (("}") reduce 95)) 652 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509) (("}") reduce 509)) 653 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92) (("}") reduce 92)) 654 ((("%") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568) (("}") reduce 568)) 655 ((("%") reduce 511) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("|") reduce 511) (("}") reduce 511)) 656 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PIPE) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459) (("}") reduce 459)) 657 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510) (("}") reduce 510)) 658 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("|") reduce 512) (("}") reduce 512)) 659 ((("%") reduce 573) (("*") reduce 573) (("+") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("|") reduce 573) (("}") reduce 573)) 665 ((("%") reduce 574) (("*") reduce 574) (("+") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("|") reduce 574) (("}") reduce 574)) 666 ((("%") reduce 578) (("*") reduce 578) (("+") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("|") reduce 578) (("}") reduce 578)) 667 ((("%") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582) (("}") reduce 582)) 668 ((("%") reduce 575) (("*") reduce 575) (("+") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("|") reduce 575) (("}") reduce 575)) 669 ((("%") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584) (("}") reduce 584)) 672 ((("%") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583) (("}") reduce 583)) 673 ((("%") reduce 576) (("*") reduce 576) (("+") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("|") reduce 576) (("}") reduce 576)) 675 ((("%") reduce 577) (("*") reduce 577) (("+") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("|") reduce 577) (("}") reduce 577)) 677 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("|") reduce 513) (("}") reduce 513)) 678 ((("%") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570) (("}") reduce 570)) 681 ((("%") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569) (("}") reduce 569)) 682 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 465) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 465)) 683 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 427)) 684 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 428)) 685 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("|") reduce 433) (("}") reduce 433)) 686 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("|") reduce 434) (("}") reduce 434)) 698 ((("%") reduce 540) (("*") reduce 540) (("+") reduce 540) (("-") reduce 540) ((".") reduce 540) (("/") reduce 540) ((":") reduce 540) (("<") reduce 540) ((">") reduce 540) (("?") reduce 540) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_BOOLEAN_AND) reduce 540) ((T_BOOLEAN_OR) reduce 540) ((T_COALESCE) reduce 540) ((T_INSTANCEOF) reduce 540) ((T_IS_EQUAL) reduce 540) ((T_IS_GREATER_OR_EQUAL) reduce 540) ((T_IS_IDENTICAL) reduce 540) ((T_IS_NOT_EQUAL) reduce 540) ((T_IS_NOT_IDENTICAL) reduce 540) ((T_IS_SMALLER_OR_EQUAL) reduce 540) ((T_LOGICAL_AND) reduce 540) ((T_LOGICAL_OR) reduce 540) ((T_LOGICAL_XOR) reduce 540) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 540) ((T_OBJECT_OPERATOR) reduce 540) ((T_PIPE) reduce 540) ((T_POW) reduce 540) ((T_SL) reduce 540) ((T_SPACESHIP) reduce 540) ((T_SR) reduce 540) (("[") reduce 540) (("^") reduce 540) (("|") reduce 540)) 700 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) ((":") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_PIPE) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("|") reduce 521)) 702 ((("%") reduce 534) (("*") reduce 534) (("+") reduce 534) (("-") reduce 534) ((".") reduce 534) (("/") reduce 534) ((":") reduce 534) (("<") reduce 534) ((">") reduce 534) (("?") reduce 534) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_BOOLEAN_AND) reduce 534) ((T_BOOLEAN_OR) reduce 534) ((T_COALESCE) reduce 534) ((T_INSTANCEOF) reduce 534) ((T_IS_EQUAL) reduce 534) ((T_IS_GREATER_OR_EQUAL) reduce 534) ((T_IS_IDENTICAL) reduce 534) ((T_IS_NOT_EQUAL) reduce 534) ((T_IS_NOT_IDENTICAL) reduce 534) ((T_IS_SMALLER_OR_EQUAL) reduce 534) ((T_LOGICAL_AND) reduce 534) ((T_LOGICAL_OR) reduce 534) ((T_LOGICAL_XOR) reduce 534) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PIPE) reduce 534) ((T_POW) reduce 534) ((T_SL) reduce 534) ((T_SPACESHIP) reduce 534) ((T_SR) reduce 534) (("[") reduce 534) (("^") reduce 534) (("|") reduce 534)) 703 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) ((":") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_PIPE) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("^") reduce 524) (("|") reduce 524)) 707 ((("%") reduce 517) (("(") shift) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((":") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_PIPE) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("^") reduce 517) (("|") reduce 517)) 708 ((("%") reduce 533) (("*") reduce 533) (("+") reduce 533) (("-") reduce 533) ((".") reduce 533) (("/") reduce 533) ((":") reduce 533) (("<") reduce 533) ((">") reduce 533) (("?") reduce 533) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_BOOLEAN_AND) reduce 533) ((T_BOOLEAN_OR) reduce 533) ((T_COALESCE) reduce 533) ((T_INSTANCEOF) reduce 533) ((T_IS_EQUAL) reduce 533) ((T_IS_GREATER_OR_EQUAL) reduce 533) ((T_IS_IDENTICAL) reduce 533) ((T_IS_NOT_EQUAL) reduce 533) ((T_IS_NOT_IDENTICAL) reduce 533) ((T_IS_SMALLER_OR_EQUAL) reduce 533) ((T_LOGICAL_AND) reduce 533) ((T_LOGICAL_OR) reduce 533) ((T_LOGICAL_XOR) reduce 533) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PIPE) reduce 533) ((T_POW) reduce 533) ((T_SL) reduce 533) ((T_SPACESHIP) reduce 533) ((T_SR) reduce 533) (("[") reduce 533) (("^") reduce 533) (("|") reduce 533)) 709 ((("%") reduce 537) (("*") reduce 537) (("+") reduce 537) (("-") reduce 537) ((".") reduce 537) (("/") reduce 537) ((":") reduce 537) (("<") reduce 537) ((">") reduce 537) (("?") reduce 537) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_BOOLEAN_AND) reduce 537) ((T_BOOLEAN_OR) reduce 537) ((T_COALESCE) reduce 537) ((T_INSTANCEOF) reduce 537) ((T_IS_EQUAL) reduce 537) ((T_IS_GREATER_OR_EQUAL) reduce 537) ((T_IS_IDENTICAL) reduce 537) ((T_IS_NOT_EQUAL) reduce 537) ((T_IS_NOT_IDENTICAL) reduce 537) ((T_IS_SMALLER_OR_EQUAL) reduce 537) ((T_LOGICAL_AND) reduce 537) ((T_LOGICAL_OR) reduce 537) ((T_LOGICAL_XOR) reduce 537) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_PIPE) reduce 537) ((T_POW) reduce 537) ((T_SL) reduce 537) ((T_SPACESHIP) reduce 537) ((T_SR) reduce 537) (("[") reduce 537) (("^") reduce 537) (("|") reduce 537)) 715 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) ((":") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PIPE) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("|") reduce 532)) 717 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) ((":") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_PIPE) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("^") reduce 523) (("|") reduce 523)) 719 ((("%") reduce 536) (("*") reduce 536) (("+") reduce 536) (("-") reduce 536) ((".") reduce 536) (("/") reduce 536) ((":") reduce 536) (("<") reduce 536) ((">") reduce 536) (("?") reduce 536) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_BOOLEAN_AND) reduce 536) ((T_BOOLEAN_OR) reduce 536) ((T_COALESCE) reduce 536) ((T_INSTANCEOF) reduce 536) ((T_IS_EQUAL) reduce 536) ((T_IS_GREATER_OR_EQUAL) reduce 536) ((T_IS_IDENTICAL) reduce 536) ((T_IS_NOT_EQUAL) reduce 536) ((T_IS_NOT_IDENTICAL) reduce 536) ((T_IS_SMALLER_OR_EQUAL) reduce 536) ((T_LOGICAL_AND) reduce 536) ((T_LOGICAL_OR) reduce 536) ((T_LOGICAL_XOR) reduce 536) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 536) ((T_OBJECT_OPERATOR) reduce 536) ((T_PIPE) reduce 536) ((T_POW) reduce 536) ((T_SL) reduce 536) ((T_SPACESHIP) reduce 536) ((T_SR) reduce 536) (("[") reduce 536) (("^") reduce 536) (("|") reduce 536)) 720 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((":") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("|") reduce 94)) 721 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((":") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("|") reduce 93)) 722 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((":") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("|") reduce 95)) 724 ((("%") reduce 539) (("*") reduce 539) (("+") reduce 539) (("-") reduce 539) ((".") reduce 539) (("/") reduce 539) ((":") reduce 539) (("<") reduce 539) ((">") reduce 539) (("?") reduce 539) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_BOOLEAN_AND) reduce 539) ((T_BOOLEAN_OR) reduce 539) ((T_COALESCE) reduce 539) ((T_INSTANCEOF) reduce 539) ((T_IS_EQUAL) reduce 539) ((T_IS_GREATER_OR_EQUAL) reduce 539) ((T_IS_IDENTICAL) reduce 539) ((T_IS_NOT_EQUAL) reduce 539) ((T_IS_NOT_IDENTICAL) reduce 539) ((T_IS_SMALLER_OR_EQUAL) reduce 539) ((T_LOGICAL_AND) reduce 539) ((T_LOGICAL_OR) reduce 539) ((T_LOGICAL_XOR) reduce 539) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 539) ((T_OBJECT_OPERATOR) reduce 539) ((T_PIPE) reduce 539) ((T_POW) reduce 539) ((T_SL) reduce 539) ((T_SPACESHIP) reduce 539) ((T_SR) reduce 539) (("[") reduce 539) (("^") reduce 539) (("|") reduce 539)) 727 ((("%") reduce 538) (("*") reduce 538) (("+") reduce 538) (("-") reduce 538) ((".") reduce 538) (("/") reduce 538) ((":") reduce 538) (("<") reduce 538) ((">") reduce 538) (("?") reduce 538) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_BOOLEAN_AND) reduce 538) ((T_BOOLEAN_OR) reduce 538) ((T_COALESCE) reduce 538) ((T_INSTANCEOF) reduce 538) ((T_IS_EQUAL) reduce 538) ((T_IS_GREATER_OR_EQUAL) reduce 538) ((T_IS_IDENTICAL) reduce 538) ((T_IS_NOT_EQUAL) reduce 538) ((T_IS_NOT_IDENTICAL) reduce 538) ((T_IS_SMALLER_OR_EQUAL) reduce 538) ((T_LOGICAL_AND) reduce 538) ((T_LOGICAL_OR) reduce 538) ((T_LOGICAL_XOR) reduce 538) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 538) ((T_OBJECT_OPERATOR) reduce 538) ((T_PIPE) reduce 538) ((T_POW) reduce 538) ((T_SL) reduce 538) ((T_SPACESHIP) reduce 538) ((T_SR) reduce 538) (("[") reduce 538) (("^") reduce 538) (("|") reduce 538)) 733 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((":") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92)) 736 ((("%") reduce 535) (("*") reduce 535) (("+") reduce 535) (("-") reduce 535) ((".") reduce 535) (("/") reduce 535) ((":") reduce 535) (("<") reduce 535) ((">") reduce 535) (("?") reduce 535) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_BOOLEAN_AND) reduce 535) ((T_BOOLEAN_OR) reduce 535) ((T_COALESCE) reduce 535) ((T_INSTANCEOF) reduce 535) ((T_IS_EQUAL) reduce 535) ((T_IS_GREATER_OR_EQUAL) reduce 535) ((T_IS_IDENTICAL) reduce 535) ((T_IS_NOT_EQUAL) reduce 535) ((T_IS_NOT_IDENTICAL) reduce 535) ((T_IS_SMALLER_OR_EQUAL) reduce 535) ((T_LOGICAL_AND) reduce 535) ((T_LOGICAL_OR) reduce 535) ((T_LOGICAL_XOR) reduce 535) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PIPE) reduce 535) ((T_POW) reduce 535) ((T_SL) reduce 535) ((T_SPACESHIP) reduce 535) ((T_SR) reduce 535) (("[") reduce 535) (("^") reduce 535) (("|") reduce 535)) 738 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) ((":") reduce 568) (("<") reduce 568) (("=") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AND_EQUAL) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_COALESCE_EQUAL) reduce 568) ((T_CONCAT_EQUAL) reduce 568) ((T_DEC) reduce 568) ((T_DIV_EQUAL) reduce 568) ((T_INC) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_MINUS_EQUAL) reduce 568) ((T_MOD_EQUAL) reduce 568) ((T_MUL_EQUAL) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_OR_EQUAL) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_PLUS_EQUAL) reduce 568) ((T_POW) reduce 568) ((T_POW_EQUAL) reduce 568) ((T_SL) reduce 568) ((T_SL_EQUAL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) ((T_SR_EQUAL) reduce 568) ((T_XOR_EQUAL) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 739 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 479) (("(") shift) (("*") reduce 479) (("+") shift) (("-") shift) ((".") reduce 479) (("/") reduce 479) ((":") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 479) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 479) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PIPE) reduce 479) ((T_POW) reduce 479) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 479) (("`") shift) (("|") reduce 479) (("~") shift)) 746 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) ((":") reduce 564) (("<") reduce 564) (("=") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AND_EQUAL) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_COALESCE_EQUAL) reduce 564) ((T_CONCAT_EQUAL) reduce 564) ((T_DEC) reduce 564) ((T_DIV_EQUAL) reduce 564) ((T_INC) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_MINUS_EQUAL) reduce 564) ((T_MOD_EQUAL) reduce 564) ((T_MUL_EQUAL) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_OR_EQUAL) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_PLUS_EQUAL) reduce 564) ((T_POW) reduce 564) ((T_POW_EQUAL) reduce 564) ((T_SL) reduce 564) ((T_SL_EQUAL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) ((T_SR_EQUAL) reduce 564) ((T_XOR_EQUAL) reduce 564) (("[") reduce 564) (("^") reduce 564) (("|") reduce 564)) 747 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) ((":") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PIPE) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 551) (("^") reduce 530) (("|") reduce 530)) 749 ((("%") reduce 529) (("*") reduce 529) (("+") reduce 529) (("-") reduce 529) ((".") reduce 529) (("/") reduce 529) ((":") reduce 529) (("<") reduce 529) ((">") reduce 529) (("?") reduce 529) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_BOOLEAN_AND) reduce 529) ((T_BOOLEAN_OR) reduce 529) ((T_COALESCE) reduce 529) ((T_INSTANCEOF) reduce 529) ((T_IS_EQUAL) reduce 529) ((T_IS_GREATER_OR_EQUAL) reduce 529) ((T_IS_IDENTICAL) reduce 529) ((T_IS_NOT_EQUAL) reduce 529) ((T_IS_NOT_IDENTICAL) reduce 529) ((T_IS_SMALLER_OR_EQUAL) reduce 529) ((T_LOGICAL_AND) reduce 529) ((T_LOGICAL_OR) reduce 529) ((T_LOGICAL_XOR) reduce 529) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PIPE) reduce 529) ((T_POW) reduce 529) ((T_SL) reduce 529) ((T_SPACESHIP) reduce 529) ((T_SR) reduce 529) (("[") reduce 554) (("^") reduce 529) (("|") reduce 529)) 750 ((("%") reduce 528) (("(") reduce 557) (("*") reduce 528) (("+") reduce 528) (("-") reduce 528) ((".") reduce 528) (("/") reduce 528) ((":") reduce 528) (("<") reduce 528) ((">") reduce 528) (("?") reduce 528) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_BOOLEAN_AND) reduce 528) ((T_BOOLEAN_OR) reduce 528) ((T_COALESCE) reduce 528) ((T_INSTANCEOF) reduce 528) ((T_IS_EQUAL) reduce 528) ((T_IS_GREATER_OR_EQUAL) reduce 528) ((T_IS_IDENTICAL) reduce 528) ((T_IS_NOT_EQUAL) reduce 528) ((T_IS_NOT_IDENTICAL) reduce 528) ((T_IS_SMALLER_OR_EQUAL) reduce 528) ((T_LOGICAL_AND) reduce 528) ((T_LOGICAL_OR) reduce 528) ((T_LOGICAL_XOR) reduce 528) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PIPE) reduce 528) ((T_POW) reduce 528) ((T_SL) reduce 528) ((T_SPACESHIP) reduce 528) ((T_SR) reduce 528) (("[") reduce 550) (("^") reduce 528) (("|") reduce 528)) 751 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 754 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) ((":") reduce 563) (("<") reduce 563) (("=") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AND_EQUAL) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_COALESCE_EQUAL) reduce 563) ((T_CONCAT_EQUAL) reduce 563) ((T_DEC) reduce 563) ((T_DIV_EQUAL) reduce 563) ((T_INC) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_MINUS_EQUAL) reduce 563) ((T_MOD_EQUAL) reduce 563) ((T_MUL_EQUAL) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_OR_EQUAL) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_PLUS_EQUAL) reduce 563) ((T_POW) reduce 563) ((T_POW_EQUAL) reduce 563) ((T_SL) reduce 563) ((T_SL_EQUAL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) ((T_SR_EQUAL) reduce 563) ((T_XOR_EQUAL) reduce 563) (("[") reduce 563) (("^") reduce 563) (("|") reduce 563)) 755 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) ((":") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_PIPE) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("^") reduce 484) (("|") reduce 484)) 756 ((("%") reduce 466) (("*") reduce 466) (("+") reduce 466) (("-") reduce 466) ((".") reduce 466) (("/") reduce 466) ((":") reduce 466) (("<") reduce 466) ((">") reduce 466) (("?") reduce 466) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_BOOLEAN_AND) reduce 466) ((T_BOOLEAN_OR) reduce 466) ((T_COALESCE) reduce 466) ((T_INSTANCEOF) reduce 466) ((T_IS_EQUAL) reduce 466) ((T_IS_GREATER_OR_EQUAL) reduce 466) ((T_IS_IDENTICAL) reduce 466) ((T_IS_NOT_EQUAL) reduce 466) ((T_IS_NOT_IDENTICAL) reduce 466) ((T_IS_SMALLER_OR_EQUAL) reduce 466) ((T_LOGICAL_AND) reduce 466) ((T_LOGICAL_OR) reduce 466) ((T_LOGICAL_XOR) reduce 466) ((T_PIPE) reduce 466) ((T_POW) reduce 466) ((T_SL) reduce 466) ((T_SPACESHIP) reduce 466) ((T_SR) reduce 466) (("^") reduce 466) (("|") reduce 466)) 757 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) ((":") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_PIPE) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("^") reduce 488) (("|") reduce 488)) 758 ((("%") reduce 531) (("(") shift) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) ((":") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("|") reduce 531)) 759 ((("%") reduce 461) (("(") reduce 558) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) ((":") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PIPE) reduce 461) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") reduce 552) (("^") reduce 461) (("|") reduce 461)) 760 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) ((":") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_PIPE) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 761 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) ((":") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_PIPE) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("|") reduce 476)) 762 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) ((":") reduce 559) (("<") reduce 559) (("=") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AND_EQUAL) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_COALESCE_EQUAL) reduce 559) ((T_CONCAT_EQUAL) reduce 559) ((T_DEC) reduce 559) ((T_DIV_EQUAL) reduce 559) ((T_INC) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_MINUS_EQUAL) reduce 559) ((T_MOD_EQUAL) reduce 559) ((T_MUL_EQUAL) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_OR_EQUAL) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_PLUS_EQUAL) reduce 559) ((T_POW) reduce 559) ((T_POW_EQUAL) reduce 559) ((T_SL) reduce 559) ((T_SL_EQUAL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) ((T_SR_EQUAL) reduce 559) ((T_XOR_EQUAL) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 763 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((":") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 764 ((("%") reduce 403) (("*") reduce 403) (("+") reduce 403) (("-") reduce 403) ((".") reduce 403) (("/") reduce 403) ((":") reduce 403) (("<") reduce 403) (("=") shift) ((">") reduce 403) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_INC) shift) ((T_INSTANCEOF) reduce 403) ((T_IS_EQUAL) reduce 403) ((T_IS_GREATER_OR_EQUAL) reduce 403) ((T_IS_IDENTICAL) reduce 403) ((T_IS_NOT_EQUAL) reduce 403) ((T_IS_NOT_IDENTICAL) reduce 403) ((T_IS_SMALLER_OR_EQUAL) reduce 403) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 403) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 403) ((T_POW_EQUAL) shift) ((T_SL) reduce 403) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 403) ((T_SR) reduce 403) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 548) (("^") reduce 403) (("|") reduce 403)) 767 ((("%") reduce 448) (("*") reduce 448) (("+") reduce 448) (("-") reduce 448) ((".") reduce 448) (("/") reduce 448) ((":") reduce 448) (("<") reduce 448) ((">") reduce 448) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_BOOLEAN_AND) reduce 448) ((T_BOOLEAN_OR) reduce 448) ((T_COALESCE) reduce 448) ((T_INSTANCEOF) reduce 448) ((T_IS_EQUAL) reduce 448) ((T_IS_GREATER_OR_EQUAL) reduce 448) ((T_IS_IDENTICAL) reduce 448) ((T_IS_NOT_EQUAL) reduce 448) ((T_IS_NOT_IDENTICAL) reduce 448) ((T_IS_SMALLER_OR_EQUAL) reduce 448) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_PIPE) reduce 448) ((T_POW) shift) ((T_SL) reduce 448) ((T_SPACESHIP) reduce 448) ((T_SR) reduce 448) (("^") reduce 448) (("|") reduce 448)) 799 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 432) (("<") shift) ((">") shift) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 432)) 800 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 435) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("|") reduce 435)) 801 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 444) (("/") shift) ((":") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_PIPE) reduce 444) ((T_POW) shift) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("|") reduce 444)) 802 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 458) (("<") shift) ((">") shift) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 458) (("|") reduce 458)) 803 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 443) (("/") shift) ((":") reduce 443) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_PIPE) reduce 443) ((T_POW) shift) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("^") reduce 443) (("|") reduce 443)) 804 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) ((":") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_PIPE) reduce 440) ((T_POW) shift) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("|") reduce 440)) 805 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 453) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_PIPE) reduce 453) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 453) ((T_SR) shift) (("^") reduce 453) (("|") reduce 453)) 806 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 431) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 807 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 429) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 808 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 430) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 809 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 455) (("<") shift) ((">") shift) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 455) ((T_SR) shift) (("^") reduce 455) (("|") reduce 455)) 810 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 450) (("<") shift) ((">") shift) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 450) (("|") reduce 450)) 811 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 452) (("<") shift) ((">") shift) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 452) (("|") reduce 452)) 812 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 449) (("<") shift) ((">") shift) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 449) (("|") reduce 449)) 813 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 457) (("<") shift) ((">") shift) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 457) ((T_SR) shift) (("^") reduce 457) (("|") reduce 457)) 814 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 451) (("<") shift) ((">") shift) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 451) (("|") reduce 451)) 817 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((":") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 818 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((":") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 819 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((":") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 820 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((":") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 821 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((":") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 822 ((("%") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) ((":") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 823 ((("%") reduce 511) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((":") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("|") reduce 511)) 824 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) ((":") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PIPE) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 825 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((":") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 826 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((":") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("|") reduce 512)) 827 ((("%") reduce 573) (("*") reduce 573) (("+") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) ((":") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("|") reduce 573)) 833 ((("%") reduce 574) (("*") reduce 574) (("+") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) ((":") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("|") reduce 574)) 834 ((("%") reduce 578) (("*") reduce 578) (("+") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) ((":") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("|") reduce 578)) 835 ((("%") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) ((":") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 836 ((("%") reduce 575) (("*") reduce 575) (("+") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) ((":") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("|") reduce 575)) 837 ((("%") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) ((":") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 840 ((("%") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) ((":") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 841 ((("%") reduce 576) (("*") reduce 576) (("+") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) ((":") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("|") reduce 576)) 843 ((("%") reduce 577) (("*") reduce 577) (("+") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) ((":") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("|") reduce 577)) 845 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((":") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("|") reduce 513)) 846 ((("%") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) ((":") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 849 ((("%") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) ((":") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 850 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 465) (("<") shift) ((">") shift) (("?") reduce 465) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 851 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 427) (("<") shift) ((">") shift) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 852 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 428) (("<") shift) ((">") shift) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 853 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 433) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("|") reduce 433)) 854 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 434) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("|") reduce 434)) 858 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 463) (("<") shift) ((">") shift) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 859 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 464) (("<") shift) ((">") shift) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 860 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 456) (("<") shift) ((">") shift) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 456) ((T_SR) shift) (("^") reduce 456) (("|") reduce 456)) 861 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 454) (("<") shift) ((">") shift) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 454) ((T_SR) shift) (("^") reduce 454) (("|") reduce 454)) 862 ((("%") reduce 441) (("*") reduce 441) (("+") reduce 441) (("-") reduce 441) ((".") reduce 441) (("/") reduce 441) ((":") reduce 441) (("<") reduce 441) ((">") reduce 441) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 441) ((T_IS_GREATER_OR_EQUAL) reduce 441) ((T_IS_IDENTICAL) reduce 441) ((T_IS_NOT_EQUAL) reduce 441) ((T_IS_NOT_IDENTICAL) reduce 441) ((T_IS_SMALLER_OR_EQUAL) reduce 441) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_PIPE) reduce 441) ((T_POW) shift) ((T_SL) reduce 441) ((T_SPACESHIP) reduce 441) ((T_SR) reduce 441) (("^") reduce 441) (("|") reduce 441)) 863 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 436) (("/") shift) ((":") reduce 436) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_PIPE) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 436) ((T_SR) shift) (("^") reduce 436) (("|") reduce 436)) 864 ((("%") shift) (("*") shift) (("+") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") shift) ((":") reduce 438) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_PIPE) reduce 438) ((T_POW) shift) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("|") reduce 438)) 865 ((("%") shift) (("*") shift) (("+") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") shift) ((":") reduce 437) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_PIPE) reduce 437) ((T_POW) shift) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("|") reduce 437)) 866 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) ((":") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_PIPE) reduce 439) ((T_POW) shift) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("|") reduce 439)) 867 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) ((":") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_PIPE) reduce 442) ((T_POW) shift) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("|") reduce 442)) 868 ((("$") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_AS) shift) ((T_BREAK) shift) ((T_CALLABLE) shift) ((T_CASE) shift) ((T_CATCH) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONST) shift) ((T_CONTINUE) shift) ((T_DECLARE) shift) ((T_DEFAULT) shift) ((T_DIR) shift) ((T_DO) shift) ((T_ECHO) shift) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) shift) ((T_ENDDECLARE) shift) ((T_ENDFOR) shift) ((T_ENDFOREACH) shift) ((T_ENDIF) shift) ((T_ENDSWITCH) shift) ((T_ENDWHILE) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_EXTENDS) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FINALLY) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_IF) shift) ((T_IMPLEMENTS) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) shift) ((T_INSTEADOF) shift) ((T_INTERFACE) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAMESPACE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_PRINT) shift) ((T_PRIVATE) shift) ((T_PROPERTY_C) shift) ((T_PROTECTED) shift) ((T_PUBLIC) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_USE) shift) ((T_VAR) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) (("{") shift)) 869 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) ((":") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_PIPE) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("|") reduce 74)) 870 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) ((":") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_PIPE) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("|") reduce 52)) 871 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) ((":") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_PIPE) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("|") reduce 27)) 872 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) ((":") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_PIPE) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("|") reduce 51)) 873 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) ((":") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_PIPE) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("|") reduce 53)) 874 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) ((":") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_PIPE) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("|") reduce 49)) 875 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) ((":") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_PIPE) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("|") reduce 29)) 876 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) ((":") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_PIPE) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("|") reduce 59)) 877 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) ((":") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_PIPE) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("|") reduce 60)) 878 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) ((":") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_PIPE) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("|") reduce 11)) 879 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) ((":") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_PIPE) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("|") reduce 42)) 880 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) ((":") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_PIPE) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("|") reduce 39)) 881 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) ((":") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_PIPE) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("|") reduce 25)) 882 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) ((":") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_PIPE) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("|") reduce 50)) 883 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) ((":") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_PIPE) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("|") reduce 66)) 884 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) ((":") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_PIPE) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("|") reduce 18)) 885 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) ((":") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_PIPE) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("|") reduce 17)) 886 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) ((":") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_PIPE) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("|") reduce 15)) 887 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) ((":") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_PIPE) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("|") reduce 14)) 888 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) ((":") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_PIPE) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("|") reduce 38)) 889 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) ((":") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_PIPE) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("|") reduce 26)) 890 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) ((":") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_PIPE) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("|") reduce 22)) 891 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) ((":") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_PIPE) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("|") reduce 24)) 892 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) ((":") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_PIPE) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("|") reduce 16)) 893 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) ((":") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_PIPE) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("|") reduce 48)) 894 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) ((":") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_PIPE) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("|") reduce 20)) 895 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) ((":") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_PIPE) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("|") reduce 70)) 896 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) ((":") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_PIPE) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("|") reduce 3)) 897 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) ((":") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_PIPE) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("|") reduce 12)) 898 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) ((":") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_PIPE) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("|") reduce 54)) 899 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) ((":") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_PIPE) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("|") reduce 65)) 900 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) ((":") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_PIPE) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("|") reduce 75)) 901 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) ((":") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_PIPE) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("|") reduce 30)) 902 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) ((":") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_PIPE) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("|") reduce 68)) 903 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) ((":") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_PIPE) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("|") reduce 21)) 904 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) ((":") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_PIPE) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("|") reduce 23)) 905 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) ((":") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_PIPE) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("|") reduce 41)) 906 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) ((":") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_PIPE) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("|") reduce 62)) 907 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) ((":") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_PIPE) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("|") reduce 34)) 908 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) ((":") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_PIPE) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("|") reduce 40)) 909 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) ((":") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_PIPE) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("|") reduce 13)) 910 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) ((":") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_PIPE) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("|") reduce 55)) 911 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) ((":") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_PIPE) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("|") reduce 1)) 912 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) ((":") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_PIPE) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("|") reduce 2)) 913 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) ((":") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_PIPE) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("|") reduce 9)) 914 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) ((":") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_PIPE) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("|") reduce 33)) 915 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) ((":") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_PIPE) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("|") reduce 58)) 916 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) ((":") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_PIPE) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("|") reduce 37)) 917 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) ((":") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_PIPE) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("|") reduce 64)) 918 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) ((":") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_PIPE) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("|") reduce 46)) 919 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) ((":") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_PIPE) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("|") reduce 8)) 920 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) ((":") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_PIPE) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("|") reduce 6)) 921 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) ((":") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_PIPE) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("|") reduce 7)) 922 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) ((":") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_PIPE) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("|") reduce 69)) 923 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) ((":") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_PIPE) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("|") reduce 63)) 924 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) ((":") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_PIPE) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("|") reduce 56)) 925 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) ((":") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_PIPE) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("|") reduce 10)) 926 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) ((":") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_PIPE) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("|") reduce 67)) 927 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) ((":") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_PIPE) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("|") reduce 44)) 928 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) ((":") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_PIPE) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("|") reduce 76)) 929 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) ((":") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_PIPE) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("|") reduce 71)) 930 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) ((":") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_PIPE) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("|") reduce 77)) 931 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) ((":") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_PIPE) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("|") reduce 78)) 932 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) ((":") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_PIPE) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("|") reduce 79)) 933 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) ((":") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_PIPE) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("|") reduce 4)) 934 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) ((":") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_PIPE) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("|") reduce 5)) 935 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) ((":") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_PIPE) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("|") reduce 43)) 936 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) ((":") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_PIPE) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("|") reduce 73)) 937 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) ((":") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_PIPE) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("|") reduce 82)) 938 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) ((":") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_PIPE) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("|") reduce 47)) 939 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) ((":") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_PIPE) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("|") reduce 31)) 940 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) ((":") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_PIPE) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("|") reduce 57)) 941 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) ((":") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_PIPE) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("|") reduce 61)) 942 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) ((":") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_PIPE) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("|") reduce 28)) 943 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) ((":") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_PIPE) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("|") reduce 36)) 944 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) ((":") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_PIPE) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("|") reduce 32)) 945 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) ((":") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_PIPE) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("|") reduce 35)) 946 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) ((":") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_PIPE) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("|") reduce 19)) 947 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) ((":") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_PIPE) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("|") reduce 45)) 948 ((("%") reduce 542) (("(") reduce 579) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) ((":") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PIPE) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 950 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) ((":") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_PIPE) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("|") reduce 72)) 951 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) ((":") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_PIPE) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("|") reduce 83)) 952 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) ((":") reduce 572) (("<") reduce 572) (("=") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AND_EQUAL) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_COALESCE_EQUAL) reduce 572) ((T_CONCAT_EQUAL) reduce 572) ((T_DEC) reduce 572) ((T_DIV_EQUAL) reduce 572) ((T_INC) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_MINUS_EQUAL) reduce 572) ((T_MOD_EQUAL) reduce 572) ((T_MUL_EQUAL) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_OR_EQUAL) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_PLUS_EQUAL) reduce 572) ((T_POW) reduce 572) ((T_POW_EQUAL) reduce 572) ((T_SL) reduce 572) ((T_SL_EQUAL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) ((T_SR_EQUAL) reduce 572) ((T_XOR_EQUAL) reduce 572) (("[") reduce 572) (("^") reduce 572) (("|") reduce 572)) 955 ((("%") reduce 544) (("(") reduce 580) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) ((":") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PIPE) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 956 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) ((")") shift) (("+") shift) (("-") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_AS) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CALLABLE) shift) ((T_CASE) shift) ((T_CATCH) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONST) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DEFAULT) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_ELLIPSIS) shift) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) shift) ((T_ENDDECLARE) shift) ((T_ENDFOR) shift) ((T_ENDFOREACH) shift) ((T_ENDIF) shift) ((T_ENDSWITCH) shift) ((T_ENDWHILE) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_EXTENDS) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FINALLY) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_IF) shift) ((T_IMPLEMENTS) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) shift) ((T_INSTEADOF) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAMESPACE) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PRIVATE) shift) ((T_PROPERTY_C) shift) ((T_PROTECTED) shift) ((T_PUBLIC) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_USE) shift) ((T_VAR) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 957 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((":") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("|") reduce 507)) 962 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((":") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 966 (((":") reduce 74)) 967 ((("(") shift) ((":") reduce 52)) 969 (((":") reduce 27)) 971 (((":") reduce 51)) 972 (((":") reduce 53)) 973 (((":") reduce 49)) 974 (((":") reduce 29)) 975 (((":") reduce 59)) 976 ((("%") reduce 540) ((")") reduce 540) (("*") reduce 540) (("+") reduce 540) ((",") reduce 540) (("-") reduce 540) ((".") reduce 540) (("/") reduce 540) ((":") reduce 60) (("<") reduce 540) ((">") reduce 540) (("?") reduce 540) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_BOOLEAN_AND) reduce 540) ((T_BOOLEAN_OR) reduce 540) ((T_COALESCE) reduce 540) ((T_INSTANCEOF) reduce 540) ((T_IS_EQUAL) reduce 540) ((T_IS_GREATER_OR_EQUAL) reduce 540) ((T_IS_IDENTICAL) reduce 540) ((T_IS_NOT_EQUAL) reduce 540) ((T_IS_NOT_IDENTICAL) reduce 540) ((T_IS_SMALLER_OR_EQUAL) reduce 540) ((T_LOGICAL_AND) reduce 540) ((T_LOGICAL_OR) reduce 540) ((T_LOGICAL_XOR) reduce 540) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 540) ((T_OBJECT_OPERATOR) reduce 540) ((T_PIPE) reduce 540) ((T_POW) reduce 540) ((T_SL) reduce 540) ((T_SPACESHIP) reduce 540) ((T_SR) reduce 540) (("[") reduce 540) (("^") reduce 540) (("|") reduce 540)) 977 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((":") reduce 11) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 978 (((":") reduce 42)) 979 ((("%") reduce 521) (("(") reduce 521) ((")") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_PIPE) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("|") reduce 521)) 980 (((":") reduce 39)) 982 (((":") reduce 25)) 983 (((":") reduce 50)) 984 ((("%") reduce 534) ((")") reduce 534) (("*") reduce 534) (("+") reduce 534) ((",") reduce 534) (("-") reduce 534) ((".") reduce 534) (("/") reduce 534) ((":") reduce 66) (("<") reduce 534) ((">") reduce 534) (("?") reduce 534) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_BOOLEAN_AND) reduce 534) ((T_BOOLEAN_OR) reduce 534) ((T_COALESCE) reduce 534) ((T_INSTANCEOF) reduce 534) ((T_IS_EQUAL) reduce 534) ((T_IS_GREATER_OR_EQUAL) reduce 534) ((T_IS_IDENTICAL) reduce 534) ((T_IS_NOT_EQUAL) reduce 534) ((T_IS_NOT_IDENTICAL) reduce 534) ((T_IS_SMALLER_OR_EQUAL) reduce 534) ((T_LOGICAL_AND) reduce 534) ((T_LOGICAL_OR) reduce 534) ((T_LOGICAL_XOR) reduce 534) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PIPE) reduce 534) ((T_POW) reduce 534) ((T_SL) reduce 534) ((T_SPACESHIP) reduce 534) ((T_SR) reduce 534) (("[") reduce 534) (("^") reduce 534) (("|") reduce 534)) 985 ((("%") reduce 524) ((")") reduce 524) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_PIPE) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("^") reduce 524) (("|") reduce 524)) 986 (((":") reduce 18)) 988 (((":") reduce 17)) 989 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) ((")") shift) (("+") shift) (("-") shift) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 990 (((":") reduce 15)) 991 (((":") reduce 14)) 992 ((("(") shift) ((":") reduce 38)) 993 (((":") reduce 26)) 994 (((":") reduce 22)) 995 (((":") reduce 24)) 996 (((":") reduce 16)) 997 (((":") reduce 48)) 998 (((":") reduce 20)) 999 (((":") reduce 70)) 1000 ((("(") shift) ((":") reduce 3)) 1001 ((("%") reduce 517) (("(") shift) ((")") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((":") reduce 12) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_PIPE) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("^") reduce 517) (("|") reduce 517)) 1002 (((":") reduce 54)) 1003 ((("%") reduce 533) ((")") reduce 533) (("*") reduce 533) (("+") reduce 533) ((",") reduce 533) (("-") reduce 533) ((".") reduce 533) (("/") reduce 533) ((":") reduce 65) (("<") reduce 533) ((">") reduce 533) (("?") reduce 533) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_BOOLEAN_AND) reduce 533) ((T_BOOLEAN_OR) reduce 533) ((T_COALESCE) reduce 533) ((T_INSTANCEOF) reduce 533) ((T_IS_EQUAL) reduce 533) ((T_IS_GREATER_OR_EQUAL) reduce 533) ((T_IS_IDENTICAL) reduce 533) ((T_IS_NOT_EQUAL) reduce 533) ((T_IS_NOT_IDENTICAL) reduce 533) ((T_IS_SMALLER_OR_EQUAL) reduce 533) ((T_LOGICAL_AND) reduce 533) ((T_LOGICAL_OR) reduce 533) ((T_LOGICAL_XOR) reduce 533) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PIPE) reduce 533) ((T_POW) reduce 533) ((T_SL) reduce 533) ((T_SPACESHIP) reduce 533) ((T_SR) reduce 533) (("[") reduce 533) (("^") reduce 533) (("|") reduce 533)) 1004 (((":") reduce 75)) 1005 (((":") reduce 30)) 1006 ((("(") reduce 491) ((":") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 491) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 491)) 1007 (((":") reduce 21)) 1008 (((":") reduce 23)) 1009 ((("(") reduce 492) ((":") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492)) 1010 ((("%") reduce 537) ((")") reduce 537) (("*") reduce 537) (("+") reduce 537) ((",") reduce 537) (("-") reduce 537) ((".") reduce 537) (("/") reduce 537) ((":") reduce 62) (("<") reduce 537) ((">") reduce 537) (("?") reduce 537) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_BOOLEAN_AND) reduce 537) ((T_BOOLEAN_OR) reduce 537) ((T_COALESCE) reduce 537) ((T_INSTANCEOF) reduce 537) ((T_IS_EQUAL) reduce 537) ((T_IS_GREATER_OR_EQUAL) reduce 537) ((T_IS_IDENTICAL) reduce 537) ((T_IS_NOT_EQUAL) reduce 537) ((T_IS_NOT_IDENTICAL) reduce 537) ((T_IS_SMALLER_OR_EQUAL) reduce 537) ((T_LOGICAL_AND) reduce 537) ((T_LOGICAL_OR) reduce 537) ((T_LOGICAL_XOR) reduce 537) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_PIPE) reduce 537) ((T_POW) reduce 537) ((T_SL) reduce 537) ((T_SPACESHIP) reduce 537) ((T_SR) reduce 537) (("[") reduce 537) (("^") reduce 537) (("|") reduce 537)) 1011 (((":") reduce 34)) 1012 (((":") reduce 40)) 1013 (((":") reduce 13)) 1014 (((":") reduce 55)) 1016 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((":") reduce 1) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 1017 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((":") reduce 2) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 1018 (((":") reduce 9)) 1019 (((":") reduce 33)) 1020 (((":") reduce 58)) 1022 ((("(") shift) ((":") reduce 37)) 1023 ((("%") reduce 532) ((")") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) ((":") reduce 64) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PIPE) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("|") reduce 532)) 1024 ((("(") shift) ((":") reduce 46)) 1025 ((("%") reduce 523) ((")") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_PIPE) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("^") reduce 523) (("|") reduce 523)) 1026 (((":") reduce 8)) 1027 (((":") reduce 6)) 1028 (((":") reduce 7)) 1029 ((("(") shift) ((":") reduce 69)) 1030 ((("%") reduce 536) ((")") reduce 536) (("*") reduce 536) (("+") reduce 536) ((",") reduce 536) (("-") reduce 536) ((".") reduce 536) (("/") reduce 536) ((":") reduce 63) (("<") reduce 536) ((">") reduce 536) (("?") reduce 536) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_BOOLEAN_AND) reduce 536) ((T_BOOLEAN_OR) reduce 536) ((T_COALESCE) reduce 536) ((T_INSTANCEOF) reduce 536) ((T_IS_EQUAL) reduce 536) ((T_IS_GREATER_OR_EQUAL) reduce 536) ((T_IS_IDENTICAL) reduce 536) ((T_IS_NOT_EQUAL) reduce 536) ((T_IS_NOT_IDENTICAL) reduce 536) ((T_IS_SMALLER_OR_EQUAL) reduce 536) ((T_LOGICAL_AND) reduce 536) ((T_LOGICAL_OR) reduce 536) ((T_LOGICAL_XOR) reduce 536) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 536) ((T_OBJECT_OPERATOR) reduce 536) ((T_PIPE) reduce 536) ((T_POW) reduce 536) ((T_SL) reduce 536) ((T_SPACESHIP) reduce 536) ((T_SR) reduce 536) (("[") reduce 536) (("^") reduce 536) (("|") reduce 536)) 1031 (((":") reduce 56)) 1032 ((("%") reduce 94) (("(") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("|") reduce 94)) 1033 ((("%") reduce 93) (("(") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("|") reduce 93)) 1034 ((("%") reduce 95) (("(") reduce 95) ((")") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("|") reduce 95)) 1035 ((("$") shift) (("(") shift) ((":") reduce 10) ((T_ABSTRACT) shift) ((T_ATTRIBUTE) shift) ((T_CLASS) reduce 195) ((T_FINAL) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_READONLY) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_VARIABLE) shift)) 1036 ((("%") reduce 539) ((")") reduce 539) (("*") reduce 539) (("+") reduce 539) ((",") reduce 539) (("-") reduce 539) ((".") reduce 539) (("/") reduce 539) ((":") reduce 67) (("<") reduce 539) ((">") reduce 539) (("?") reduce 539) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_BOOLEAN_AND) reduce 539) ((T_BOOLEAN_OR) reduce 539) ((T_COALESCE) reduce 539) ((T_INSTANCEOF) reduce 539) ((T_IS_EQUAL) reduce 539) ((T_IS_GREATER_OR_EQUAL) reduce 539) ((T_IS_IDENTICAL) reduce 539) ((T_IS_NOT_EQUAL) reduce 539) ((T_IS_NOT_IDENTICAL) reduce 539) ((T_IS_SMALLER_OR_EQUAL) reduce 539) ((T_LOGICAL_AND) reduce 539) ((T_LOGICAL_OR) reduce 539) ((T_LOGICAL_XOR) reduce 539) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 539) ((T_OBJECT_OPERATOR) reduce 539) ((T_PIPE) reduce 539) ((T_POW) reduce 539) ((T_SL) reduce 539) ((T_SPACESHIP) reduce 539) ((T_SR) reduce 539) (("[") reduce 539) (("^") reduce 539) (("|") reduce 539)) 1038 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((":") reduce 44) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 1039 (((":") reduce 76)) 1040 ((("%") reduce 538) ((")") reduce 538) (("*") reduce 538) (("+") reduce 538) ((",") reduce 538) (("-") reduce 538) ((".") reduce 538) (("/") reduce 538) ((":") reduce 71) (("<") reduce 538) ((">") reduce 538) (("?") reduce 538) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_BOOLEAN_AND) reduce 538) ((T_BOOLEAN_OR) reduce 538) ((T_COALESCE) reduce 538) ((T_INSTANCEOF) reduce 538) ((T_IS_EQUAL) reduce 538) ((T_IS_GREATER_OR_EQUAL) reduce 538) ((T_IS_IDENTICAL) reduce 538) ((T_IS_NOT_EQUAL) reduce 538) ((T_IS_NOT_IDENTICAL) reduce 538) ((T_IS_SMALLER_OR_EQUAL) reduce 538) ((T_LOGICAL_AND) reduce 538) ((T_LOGICAL_OR) reduce 538) ((T_LOGICAL_XOR) reduce 538) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 538) ((T_OBJECT_OPERATOR) reduce 538) ((T_PIPE) reduce 538) ((T_POW) reduce 538) ((T_SL) reduce 538) ((T_SPACESHIP) reduce 538) ((T_SR) reduce 538) (("[") reduce 538) (("^") reduce 538) (("|") reduce 538)) 1041 (((":") reduce 77)) 1042 (((":") reduce 78)) 1043 ((("(") shift) ((":") reduce 79)) 1044 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((":") reduce 4) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 1045 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((":") reduce 5) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 1046 (((":") reduce 43)) 1048 (((":") reduce 73) ((T_FN) shift) ((T_FUNCTION) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509)) 1049 ((("%") reduce 92) (("(") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((":") reduce 82) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92)) 1051 (((":") reduce 47)) 1052 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((":") reduce 31) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 1053 (((":") reduce 57)) 1054 ((("%") reduce 535) ((")") reduce 535) (("*") reduce 535) (("+") reduce 535) ((",") reduce 535) (("-") reduce 535) ((".") reduce 535) (("/") reduce 535) ((":") reduce 61) (("<") reduce 535) ((">") reduce 535) (("?") reduce 535) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_BOOLEAN_AND) reduce 535) ((T_BOOLEAN_OR) reduce 535) ((T_COALESCE) reduce 535) ((T_INSTANCEOF) reduce 535) ((T_IS_EQUAL) reduce 535) ((T_IS_GREATER_OR_EQUAL) reduce 535) ((T_IS_IDENTICAL) reduce 535) ((T_IS_NOT_EQUAL) reduce 535) ((T_IS_NOT_IDENTICAL) reduce 535) ((T_IS_SMALLER_OR_EQUAL) reduce 535) ((T_LOGICAL_AND) reduce 535) ((T_LOGICAL_OR) reduce 535) ((T_LOGICAL_XOR) reduce 535) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PIPE) reduce 535) ((T_POW) reduce 535) ((T_SL) reduce 535) ((T_SPACESHIP) reduce 535) ((T_SR) reduce 535) (("[") reduce 535) (("^") reduce 535) (("|") reduce 535)) 1055 (((":") reduce 28)) 1056 (((":") reduce 36)) 1058 (((":") reduce 32)) 1059 (((":") reduce 35)) 1060 ((("%") reduce 568) (("(") reduce 568) ((")") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) (("=") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AND_EQUAL) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_COALESCE_EQUAL) reduce 568) ((T_CONCAT_EQUAL) reduce 568) ((T_DEC) reduce 568) ((T_DIV_EQUAL) reduce 568) ((T_INC) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_MINUS_EQUAL) reduce 568) ((T_MOD_EQUAL) reduce 568) ((T_MUL_EQUAL) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_OR_EQUAL) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_PLUS_EQUAL) reduce 568) ((T_POW) reduce 568) ((T_POW_EQUAL) reduce 568) ((T_SL) reduce 568) ((T_SL_EQUAL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) ((T_SR_EQUAL) reduce 568) ((T_XOR_EQUAL) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 1061 (((":") reduce 19)) 1062 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 479) (("(") shift) ((")") reduce 479) (("*") reduce 479) (("+") shift) ((",") reduce 479) (("-") shift) ((".") reduce 479) (("/") reduce 479) ((":") reduce 45) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 479) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 479) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PIPE) reduce 479) ((T_POW) reduce 479) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 479) (("`") shift) (("|") reduce 479) (("~") shift)) 1066 (((")") reduce 294) ((",") reduce 294)) 1067 (((")") reduce 306) ((",") reduce 306)) 1071 ((("%") reduce 564) (("(") reduce 555) ((")") reduce 564) (("*") reduce 564) (("+") reduce 564) ((",") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) (("<") reduce 564) (("=") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AND_EQUAL) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_COALESCE_EQUAL) reduce 564) ((T_CONCAT_EQUAL) reduce 564) ((T_DEC) reduce 564) ((T_DIV_EQUAL) reduce 564) ((T_INC) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_MINUS_EQUAL) reduce 564) ((T_MOD_EQUAL) reduce 564) ((T_MUL_EQUAL) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_OR_EQUAL) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_PLUS_EQUAL) reduce 564) ((T_POW) reduce 564) ((T_POW_EQUAL) reduce 564) ((T_SL) reduce 564) ((T_SL_EQUAL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) ((T_SR_EQUAL) reduce 564) ((T_XOR_EQUAL) reduce 564) (("[") reduce 564) (("^") reduce 564) (("|") reduce 564)) 1072 ((("%") reduce 530) ((")") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PIPE) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 551) (("^") reduce 530) (("|") reduce 530)) 1074 ((("%") reduce 529) ((")") reduce 529) (("*") reduce 529) (("+") reduce 529) ((",") reduce 529) (("-") reduce 529) ((".") reduce 529) (("/") reduce 529) (("<") reduce 529) ((">") reduce 529) (("?") reduce 529) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_BOOLEAN_AND) reduce 529) ((T_BOOLEAN_OR) reduce 529) ((T_COALESCE) reduce 529) ((T_INSTANCEOF) reduce 529) ((T_IS_EQUAL) reduce 529) ((T_IS_GREATER_OR_EQUAL) reduce 529) ((T_IS_IDENTICAL) reduce 529) ((T_IS_NOT_EQUAL) reduce 529) ((T_IS_NOT_IDENTICAL) reduce 529) ((T_IS_SMALLER_OR_EQUAL) reduce 529) ((T_LOGICAL_AND) reduce 529) ((T_LOGICAL_OR) reduce 529) ((T_LOGICAL_XOR) reduce 529) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PIPE) reduce 529) ((T_POW) reduce 529) ((T_SL) reduce 529) ((T_SPACESHIP) reduce 529) ((T_SR) reduce 529) (("[") reduce 554) (("^") reduce 529) (("|") reduce 529)) 1075 ((("%") reduce 528) (("(") reduce 557) ((")") reduce 528) (("*") reduce 528) (("+") reduce 528) ((",") reduce 528) (("-") reduce 528) ((".") reduce 528) (("/") reduce 528) (("<") reduce 528) ((">") reduce 528) (("?") reduce 528) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_BOOLEAN_AND) reduce 528) ((T_BOOLEAN_OR) reduce 528) ((T_COALESCE) reduce 528) ((T_INSTANCEOF) reduce 528) ((T_IS_EQUAL) reduce 528) ((T_IS_GREATER_OR_EQUAL) reduce 528) ((T_IS_IDENTICAL) reduce 528) ((T_IS_NOT_EQUAL) reduce 528) ((T_IS_NOT_IDENTICAL) reduce 528) ((T_IS_SMALLER_OR_EQUAL) reduce 528) ((T_LOGICAL_AND) reduce 528) ((T_LOGICAL_OR) reduce 528) ((T_LOGICAL_XOR) reduce 528) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PIPE) reduce 528) ((T_POW) reduce 528) ((T_SL) reduce 528) ((T_SPACESHIP) reduce 528) ((T_SR) reduce 528) (("[") reduce 550) (("^") reduce 528) (("|") reduce 528)) 1076 ((("%") shift) ((")") reduce 305) (("*") shift) (("+") shift) ((",") reduce 305) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1079 ((("%") reduce 563) (("(") reduce 563) ((")") reduce 563) (("*") reduce 563) (("+") reduce 563) ((",") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) (("<") reduce 563) (("=") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AND_EQUAL) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_COALESCE_EQUAL) reduce 563) ((T_CONCAT_EQUAL) reduce 563) ((T_DEC) reduce 563) ((T_DIV_EQUAL) reduce 563) ((T_INC) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_MINUS_EQUAL) reduce 563) ((T_MOD_EQUAL) reduce 563) ((T_MUL_EQUAL) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_OR_EQUAL) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_PLUS_EQUAL) reduce 563) ((T_POW) reduce 563) ((T_POW_EQUAL) reduce 563) ((T_SL) reduce 563) ((T_SL_EQUAL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) ((T_SR_EQUAL) reduce 563) ((T_XOR_EQUAL) reduce 563) (("[") reduce 563) (("^") reduce 563) (("|") reduce 563)) 1080 (((":") shift)) 1081 ((("%") reduce 484) ((")") reduce 484) (("*") reduce 484) (("+") reduce 484) ((",") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_PIPE) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("^") reduce 484) (("|") reduce 484)) 1082 ((("%") reduce 466) ((")") reduce 466) (("*") reduce 466) (("+") reduce 466) ((",") reduce 466) (("-") reduce 466) ((".") reduce 466) (("/") reduce 466) (("<") reduce 466) ((">") reduce 466) (("?") reduce 466) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_BOOLEAN_AND) reduce 466) ((T_BOOLEAN_OR) reduce 466) ((T_COALESCE) reduce 466) ((T_INSTANCEOF) reduce 466) ((T_IS_EQUAL) reduce 466) ((T_IS_GREATER_OR_EQUAL) reduce 466) ((T_IS_IDENTICAL) reduce 466) ((T_IS_NOT_EQUAL) reduce 466) ((T_IS_NOT_IDENTICAL) reduce 466) ((T_IS_SMALLER_OR_EQUAL) reduce 466) ((T_LOGICAL_AND) reduce 466) ((T_LOGICAL_OR) reduce 466) ((T_LOGICAL_XOR) reduce 466) ((T_PIPE) reduce 466) ((T_POW) reduce 466) ((T_SL) reduce 466) ((T_SPACESHIP) reduce 466) ((T_SR) reduce 466) (("^") reduce 466) (("|") reduce 466)) 1083 ((("%") reduce 488) ((")") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_PIPE) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("^") reduce 488) (("|") reduce 488)) 1084 ((("%") reduce 531) (("(") shift) ((")") reduce 531) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("|") reduce 531)) 1085 ((("%") reduce 461) (("(") reduce 558) ((")") reduce 461) (("*") reduce 461) (("+") reduce 461) ((",") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PIPE) reduce 461) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") reduce 552) (("^") reduce 461) (("|") reduce 461)) 1086 ((("%") reduce 462) ((")") reduce 462) (("*") reduce 462) (("+") reduce 462) ((",") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_PIPE) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 1088 (((":") reduce 72)) 1089 ((("%") reduce 476) ((")") reduce 476) (("*") reduce 476) (("+") reduce 476) ((",") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_PIPE) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("|") reduce 476)) 1090 (((":") reduce 83)) 1091 ((("%") reduce 559) (("(") reduce 559) ((")") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) (("=") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AND_EQUAL) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_COALESCE_EQUAL) reduce 559) ((T_CONCAT_EQUAL) reduce 559) ((T_DEC) reduce 559) ((T_DIV_EQUAL) reduce 559) ((T_INC) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_MINUS_EQUAL) reduce 559) ((T_MOD_EQUAL) reduce 559) ((T_MUL_EQUAL) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_OR_EQUAL) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_PLUS_EQUAL) reduce 559) ((T_POW) reduce 559) ((T_POW_EQUAL) reduce 559) ((T_SL) reduce 559) ((T_SL_EQUAL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) ((T_SR_EQUAL) reduce 559) ((T_XOR_EQUAL) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 1092 ((("%") reduce 565) ((")") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 1093 ((("%") reduce 403) ((")") reduce 403) (("*") reduce 403) (("+") reduce 403) ((",") reduce 403) (("-") reduce 403) ((".") reduce 403) (("/") reduce 403) (("<") reduce 403) (("=") shift) ((">") reduce 403) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_INC) shift) ((T_INSTANCEOF) reduce 403) ((T_IS_EQUAL) reduce 403) ((T_IS_GREATER_OR_EQUAL) reduce 403) ((T_IS_IDENTICAL) reduce 403) ((T_IS_NOT_EQUAL) reduce 403) ((T_IS_NOT_IDENTICAL) reduce 403) ((T_IS_SMALLER_OR_EQUAL) reduce 403) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 403) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 403) ((T_POW_EQUAL) shift) ((T_SL) reduce 403) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 403) ((T_SR) reduce 403) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 548) (("^") reduce 403) (("|") reduce 403)) 1097 ((("%") reduce 540) ((")") reduce 540) (("*") reduce 540) (("+") reduce 540) ((",") reduce 540) (("-") reduce 540) ((".") reduce 540) (("/") reduce 540) (("<") reduce 540) ((">") reduce 540) (("?") reduce 540) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_BOOLEAN_AND) reduce 540) ((T_BOOLEAN_OR) reduce 540) ((T_COALESCE) reduce 540) ((T_INSTANCEOF) reduce 540) ((T_IS_EQUAL) reduce 540) ((T_IS_GREATER_OR_EQUAL) reduce 540) ((T_IS_IDENTICAL) reduce 540) ((T_IS_NOT_EQUAL) reduce 540) ((T_IS_NOT_IDENTICAL) reduce 540) ((T_IS_SMALLER_OR_EQUAL) reduce 540) ((T_LOGICAL_AND) reduce 540) ((T_LOGICAL_OR) reduce 540) ((T_LOGICAL_XOR) reduce 540) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 540) ((T_OBJECT_OPERATOR) reduce 540) ((T_PIPE) reduce 540) ((T_POW) reduce 540) ((T_SL) reduce 540) ((T_SPACESHIP) reduce 540) ((T_SR) reduce 540) (("[") reduce 540) (("^") reduce 540) (("|") reduce 540)) 1099 ((("%") reduce 534) ((")") reduce 534) (("*") reduce 534) (("+") reduce 534) ((",") reduce 534) (("-") reduce 534) ((".") reduce 534) (("/") reduce 534) (("<") reduce 534) ((">") reduce 534) (("?") reduce 534) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_BOOLEAN_AND) reduce 534) ((T_BOOLEAN_OR) reduce 534) ((T_COALESCE) reduce 534) ((T_INSTANCEOF) reduce 534) ((T_IS_EQUAL) reduce 534) ((T_IS_GREATER_OR_EQUAL) reduce 534) ((T_IS_IDENTICAL) reduce 534) ((T_IS_NOT_EQUAL) reduce 534) ((T_IS_NOT_IDENTICAL) reduce 534) ((T_IS_SMALLER_OR_EQUAL) reduce 534) ((T_LOGICAL_AND) reduce 534) ((T_LOGICAL_OR) reduce 534) ((T_LOGICAL_XOR) reduce 534) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PIPE) reduce 534) ((T_POW) reduce 534) ((T_SL) reduce 534) ((T_SPACESHIP) reduce 534) ((T_SR) reduce 534) (("[") reduce 534) (("^") reduce 534) (("|") reduce 534)) 1102 ((("%") reduce 517) (("(") shift) ((")") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_PIPE) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("^") reduce 517) (("|") reduce 517)) 1103 ((("%") reduce 533) ((")") reduce 533) (("*") reduce 533) (("+") reduce 533) ((",") reduce 533) (("-") reduce 533) ((".") reduce 533) (("/") reduce 533) (("<") reduce 533) ((">") reduce 533) (("?") reduce 533) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_BOOLEAN_AND) reduce 533) ((T_BOOLEAN_OR) reduce 533) ((T_COALESCE) reduce 533) ((T_INSTANCEOF) reduce 533) ((T_IS_EQUAL) reduce 533) ((T_IS_GREATER_OR_EQUAL) reduce 533) ((T_IS_IDENTICAL) reduce 533) ((T_IS_NOT_EQUAL) reduce 533) ((T_IS_NOT_IDENTICAL) reduce 533) ((T_IS_SMALLER_OR_EQUAL) reduce 533) ((T_LOGICAL_AND) reduce 533) ((T_LOGICAL_OR) reduce 533) ((T_LOGICAL_XOR) reduce 533) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PIPE) reduce 533) ((T_POW) reduce 533) ((T_SL) reduce 533) ((T_SPACESHIP) reduce 533) ((T_SR) reduce 533) (("[") reduce 533) (("^") reduce 533) (("|") reduce 533)) 1104 ((("%") reduce 537) ((")") reduce 537) (("*") reduce 537) (("+") reduce 537) ((",") reduce 537) (("-") reduce 537) ((".") reduce 537) (("/") reduce 537) (("<") reduce 537) ((">") reduce 537) (("?") reduce 537) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_BOOLEAN_AND) reduce 537) ((T_BOOLEAN_OR) reduce 537) ((T_COALESCE) reduce 537) ((T_INSTANCEOF) reduce 537) ((T_IS_EQUAL) reduce 537) ((T_IS_GREATER_OR_EQUAL) reduce 537) ((T_IS_IDENTICAL) reduce 537) ((T_IS_NOT_EQUAL) reduce 537) ((T_IS_NOT_IDENTICAL) reduce 537) ((T_IS_SMALLER_OR_EQUAL) reduce 537) ((T_LOGICAL_AND) reduce 537) ((T_LOGICAL_OR) reduce 537) ((T_LOGICAL_XOR) reduce 537) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_PIPE) reduce 537) ((T_POW) reduce 537) ((T_SL) reduce 537) ((T_SPACESHIP) reduce 537) ((T_SR) reduce 537) (("[") reduce 537) (("^") reduce 537) (("|") reduce 537)) 1108 ((("%") reduce 532) ((")") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PIPE) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("|") reduce 532)) 1111 ((("%") reduce 536) ((")") reduce 536) (("*") reduce 536) (("+") reduce 536) ((",") reduce 536) (("-") reduce 536) ((".") reduce 536) (("/") reduce 536) (("<") reduce 536) ((">") reduce 536) (("?") reduce 536) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_BOOLEAN_AND) reduce 536) ((T_BOOLEAN_OR) reduce 536) ((T_COALESCE) reduce 536) ((T_INSTANCEOF) reduce 536) ((T_IS_EQUAL) reduce 536) ((T_IS_GREATER_OR_EQUAL) reduce 536) ((T_IS_IDENTICAL) reduce 536) ((T_IS_NOT_EQUAL) reduce 536) ((T_IS_NOT_IDENTICAL) reduce 536) ((T_IS_SMALLER_OR_EQUAL) reduce 536) ((T_LOGICAL_AND) reduce 536) ((T_LOGICAL_OR) reduce 536) ((T_LOGICAL_XOR) reduce 536) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 536) ((T_OBJECT_OPERATOR) reduce 536) ((T_PIPE) reduce 536) ((T_POW) reduce 536) ((T_SL) reduce 536) ((T_SPACESHIP) reduce 536) ((T_SR) reduce 536) (("[") reduce 536) (("^") reduce 536) (("|") reduce 536)) 1113 ((("%") reduce 539) ((")") reduce 539) (("*") reduce 539) (("+") reduce 539) ((",") reduce 539) (("-") reduce 539) ((".") reduce 539) (("/") reduce 539) (("<") reduce 539) ((">") reduce 539) (("?") reduce 539) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_BOOLEAN_AND) reduce 539) ((T_BOOLEAN_OR) reduce 539) ((T_COALESCE) reduce 539) ((T_INSTANCEOF) reduce 539) ((T_IS_EQUAL) reduce 539) ((T_IS_GREATER_OR_EQUAL) reduce 539) ((T_IS_IDENTICAL) reduce 539) ((T_IS_NOT_EQUAL) reduce 539) ((T_IS_NOT_IDENTICAL) reduce 539) ((T_IS_SMALLER_OR_EQUAL) reduce 539) ((T_LOGICAL_AND) reduce 539) ((T_LOGICAL_OR) reduce 539) ((T_LOGICAL_XOR) reduce 539) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 539) ((T_OBJECT_OPERATOR) reduce 539) ((T_PIPE) reduce 539) ((T_POW) reduce 539) ((T_SL) reduce 539) ((T_SPACESHIP) reduce 539) ((T_SR) reduce 539) (("[") reduce 539) (("^") reduce 539) (("|") reduce 539)) 1115 ((("%") reduce 538) ((")") reduce 538) (("*") reduce 538) (("+") reduce 538) ((",") reduce 538) (("-") reduce 538) ((".") reduce 538) (("/") reduce 538) (("<") reduce 538) ((">") reduce 538) (("?") reduce 538) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_BOOLEAN_AND) reduce 538) ((T_BOOLEAN_OR) reduce 538) ((T_COALESCE) reduce 538) ((T_INSTANCEOF) reduce 538) ((T_IS_EQUAL) reduce 538) ((T_IS_GREATER_OR_EQUAL) reduce 538) ((T_IS_IDENTICAL) reduce 538) ((T_IS_NOT_EQUAL) reduce 538) ((T_IS_NOT_IDENTICAL) reduce 538) ((T_IS_SMALLER_OR_EQUAL) reduce 538) ((T_LOGICAL_AND) reduce 538) ((T_LOGICAL_OR) reduce 538) ((T_LOGICAL_XOR) reduce 538) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 538) ((T_OBJECT_OPERATOR) reduce 538) ((T_PIPE) reduce 538) ((T_POW) reduce 538) ((T_SL) reduce 538) ((T_SPACESHIP) reduce 538) ((T_SR) reduce 538) (("[") reduce 538) (("^") reduce 538) (("|") reduce 538)) 1120 ((("%") reduce 92) (("(") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92)) 1122 ((("%") reduce 535) ((")") reduce 535) (("*") reduce 535) (("+") reduce 535) ((",") reduce 535) (("-") reduce 535) ((".") reduce 535) (("/") reduce 535) (("<") reduce 535) ((">") reduce 535) (("?") reduce 535) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_BOOLEAN_AND) reduce 535) ((T_BOOLEAN_OR) reduce 535) ((T_COALESCE) reduce 535) ((T_INSTANCEOF) reduce 535) ((T_IS_EQUAL) reduce 535) ((T_IS_GREATER_OR_EQUAL) reduce 535) ((T_IS_IDENTICAL) reduce 535) ((T_IS_NOT_EQUAL) reduce 535) ((T_IS_NOT_IDENTICAL) reduce 535) ((T_IS_SMALLER_OR_EQUAL) reduce 535) ((T_LOGICAL_AND) reduce 535) ((T_LOGICAL_OR) reduce 535) ((T_LOGICAL_XOR) reduce 535) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PIPE) reduce 535) ((T_POW) reduce 535) ((T_SL) reduce 535) ((T_SPACESHIP) reduce 535) ((T_SR) reduce 535) (("[") reduce 535) (("^") reduce 535) (("|") reduce 535)) 1123 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 479) (("(") shift) ((")") reduce 479) (("*") reduce 479) (("+") shift) ((",") reduce 479) (("-") shift) ((".") reduce 479) (("/") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 479) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 479) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PIPE) reduce 479) ((T_POW) reduce 479) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 479) (("`") shift) (("|") reduce 479) (("~") shift)) 1124 ((("%") reduce 448) ((")") reduce 448) (("*") reduce 448) (("+") reduce 448) ((",") reduce 448) (("-") reduce 448) ((".") reduce 448) (("/") reduce 448) (("<") reduce 448) ((">") reduce 448) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_BOOLEAN_AND) reduce 448) ((T_BOOLEAN_OR) reduce 448) ((T_COALESCE) reduce 448) ((T_INSTANCEOF) reduce 448) ((T_IS_EQUAL) reduce 448) ((T_IS_GREATER_OR_EQUAL) reduce 448) ((T_IS_IDENTICAL) reduce 448) ((T_IS_NOT_EQUAL) reduce 448) ((T_IS_NOT_IDENTICAL) reduce 448) ((T_IS_SMALLER_OR_EQUAL) reduce 448) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_PIPE) reduce 448) ((T_POW) shift) ((T_SL) reduce 448) ((T_SPACESHIP) reduce 448) ((T_SR) reduce 448) (("^") reduce 448) (("|") reduce 448)) 1156 ((("%") shift) ((")") reduce 432) (("*") shift) (("+") shift) ((",") reduce 432) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 432)) 1157 ((("%") shift) ((")") reduce 435) (("*") shift) (("+") shift) ((",") reduce 435) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("|") reduce 435)) 1158 ((("%") shift) ((")") reduce 444) (("*") shift) (("+") shift) ((",") reduce 444) (("-") shift) ((".") reduce 444) (("/") shift) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_PIPE) reduce 444) ((T_POW) shift) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("|") reduce 444)) 1159 ((("%") shift) ((")") reduce 458) (("*") shift) (("+") shift) ((",") reduce 458) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 458) (("|") reduce 458)) 1160 ((("%") shift) ((")") reduce 443) (("*") shift) (("+") shift) ((",") reduce 443) (("-") shift) ((".") reduce 443) (("/") shift) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_PIPE) reduce 443) ((T_POW) shift) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("^") reduce 443) (("|") reduce 443)) 1161 ((("%") reduce 440) ((")") reduce 440) (("*") reduce 440) (("+") reduce 440) ((",") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_PIPE) reduce 440) ((T_POW) shift) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("|") reduce 440)) 1162 ((("%") shift) ((")") reduce 453) (("*") shift) (("+") shift) ((",") reduce 453) (("-") shift) ((".") shift) (("/") shift) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_PIPE) reduce 453) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 453) ((T_SR) shift) (("^") reduce 453) (("|") reduce 453)) 1163 ((("%") shift) ((")") reduce 431) (("*") shift) (("+") shift) ((",") reduce 431) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1164 ((("%") shift) ((")") reduce 429) (("*") shift) (("+") shift) ((",") reduce 429) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1165 ((("%") shift) ((")") reduce 430) (("*") shift) (("+") shift) ((",") reduce 430) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1166 ((("%") shift) ((")") reduce 455) (("*") shift) (("+") shift) ((",") reduce 455) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 455) ((T_SR) shift) (("^") reduce 455) (("|") reduce 455)) 1167 ((("%") shift) ((")") reduce 450) (("*") shift) (("+") shift) ((",") reduce 450) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 450) (("|") reduce 450)) 1168 ((("%") shift) ((")") reduce 452) (("*") shift) (("+") shift) ((",") reduce 452) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 452) (("|") reduce 452)) 1169 ((("%") shift) ((")") reduce 449) (("*") shift) (("+") shift) ((",") reduce 449) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 449) (("|") reduce 449)) 1170 ((("%") shift) ((")") reduce 457) (("*") shift) (("+") shift) ((",") reduce 457) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 457) ((T_SR) shift) (("^") reduce 457) (("|") reduce 457)) 1171 ((("%") shift) ((")") reduce 451) (("*") shift) (("+") shift) ((",") reduce 451) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 451) (("|") reduce 451)) 1174 ((("%") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 1175 ((("%") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 1176 ((("%") reduce 95) ((")") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 1177 ((("%") reduce 509) ((")") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 1178 ((("%") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 1179 ((("%") reduce 568) ((")") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 1180 ((("%") reduce 511) ((")") reduce 511) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("|") reduce 511)) 1181 ((("%") reduce 459) ((")") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PIPE) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 1182 ((("%") reduce 510) ((")") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 1183 ((("%") reduce 512) ((")") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("|") reduce 512)) 1184 ((("%") reduce 573) ((")") reduce 573) (("*") reduce 573) (("+") reduce 573) ((",") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("|") reduce 573)) 1190 ((("%") reduce 574) ((")") reduce 574) (("*") reduce 574) (("+") reduce 574) ((",") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("|") reduce 574)) 1191 ((("%") reduce 578) ((")") reduce 578) (("*") reduce 578) (("+") reduce 578) ((",") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("|") reduce 578)) 1192 ((("%") reduce 582) ((")") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 1193 ((("%") reduce 575) ((")") reduce 575) (("*") reduce 575) (("+") reduce 575) ((",") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("|") reduce 575)) 1194 ((("%") reduce 584) ((")") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 1197 ((("%") reduce 583) ((")") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 1198 ((("%") reduce 576) ((")") reduce 576) (("*") reduce 576) (("+") reduce 576) ((",") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("|") reduce 576)) 1200 ((("%") reduce 577) ((")") reduce 577) (("*") reduce 577) (("+") reduce 577) ((",") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("|") reduce 577)) 1202 ((("%") reduce 513) ((")") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("|") reduce 513)) 1203 ((("%") reduce 570) ((")") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 1206 ((("%") reduce 569) ((")") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 1207 ((("%") shift) ((")") reduce 465) (("*") shift) (("+") shift) ((",") reduce 465) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 465) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1208 ((("%") shift) ((")") reduce 427) (("*") shift) (("+") shift) ((",") reduce 427) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1209 ((("%") shift) ((")") reduce 428) (("*") shift) (("+") shift) ((",") reduce 428) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1210 ((("%") shift) ((")") reduce 433) (("*") shift) (("+") shift) ((",") reduce 433) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("|") reduce 433)) 1211 ((("%") shift) ((")") reduce 434) (("*") shift) (("+") shift) ((",") reduce 434) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("|") reduce 434)) 1215 ((("%") shift) ((")") reduce 463) (("*") shift) (("+") shift) ((",") reduce 463) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1216 ((("%") shift) ((")") reduce 464) (("*") shift) (("+") shift) ((",") reduce 464) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1217 ((("%") shift) ((")") reduce 456) (("*") shift) (("+") shift) ((",") reduce 456) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 456) ((T_SR) shift) (("^") reduce 456) (("|") reduce 456)) 1218 ((("%") shift) ((")") reduce 454) (("*") shift) (("+") shift) ((",") reduce 454) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 454) ((T_SR) shift) (("^") reduce 454) (("|") reduce 454)) 1219 ((("%") reduce 441) ((")") reduce 441) (("*") reduce 441) (("+") reduce 441) ((",") reduce 441) (("-") reduce 441) ((".") reduce 441) (("/") reduce 441) (("<") reduce 441) ((">") reduce 441) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 441) ((T_IS_GREATER_OR_EQUAL) reduce 441) ((T_IS_IDENTICAL) reduce 441) ((T_IS_NOT_EQUAL) reduce 441) ((T_IS_NOT_IDENTICAL) reduce 441) ((T_IS_SMALLER_OR_EQUAL) reduce 441) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_PIPE) reduce 441) ((T_POW) shift) ((T_SL) reduce 441) ((T_SPACESHIP) reduce 441) ((T_SR) reduce 441) (("^") reduce 441) (("|") reduce 441)) 1220 ((("%") shift) ((")") reduce 436) (("*") shift) (("+") shift) ((",") reduce 436) (("-") shift) ((".") reduce 436) (("/") shift) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_PIPE) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 436) ((T_SR) shift) (("^") reduce 436) (("|") reduce 436)) 1221 ((("%") shift) ((")") reduce 438) (("*") shift) (("+") reduce 438) ((",") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") shift) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_PIPE) reduce 438) ((T_POW) shift) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("|") reduce 438)) 1222 ((("%") shift) ((")") reduce 437) (("*") shift) (("+") reduce 437) ((",") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") shift) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_PIPE) reduce 437) ((T_POW) shift) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("|") reduce 437)) 1223 ((("%") reduce 439) ((")") reduce 439) (("*") reduce 439) (("+") reduce 439) ((",") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_PIPE) reduce 439) ((T_POW) shift) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("|") reduce 439)) 1224 ((("%") reduce 442) ((")") reduce 442) (("*") reduce 442) (("+") reduce 442) ((",") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_PIPE) reduce 442) ((T_POW) shift) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("|") reduce 442)) 1235 ((("%") reduce 540) ((")") reduce 540) (("*") reduce 540) (("+") reduce 540) ((",") reduce 540) (("-") reduce 540) ((".") reduce 540) (("/") reduce 540) (("<") reduce 540) ((">") reduce 540) (("?") reduce 540) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_BOOLEAN_AND) reduce 540) ((T_BOOLEAN_OR) reduce 540) ((T_COALESCE) reduce 540) ((T_DOUBLE_ARROW) reduce 540) ((T_INSTANCEOF) reduce 540) ((T_IS_EQUAL) reduce 540) ((T_IS_GREATER_OR_EQUAL) reduce 540) ((T_IS_IDENTICAL) reduce 540) ((T_IS_NOT_EQUAL) reduce 540) ((T_IS_NOT_IDENTICAL) reduce 540) ((T_IS_SMALLER_OR_EQUAL) reduce 540) ((T_LOGICAL_AND) reduce 540) ((T_LOGICAL_OR) reduce 540) ((T_LOGICAL_XOR) reduce 540) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 540) ((T_OBJECT_OPERATOR) reduce 540) ((T_PIPE) reduce 540) ((T_POW) reduce 540) ((T_SL) reduce 540) ((T_SPACESHIP) reduce 540) ((T_SR) reduce 540) (("[") reduce 540) (("^") reduce 540) (("|") reduce 540)) 1237 ((("%") reduce 521) (("(") reduce 521) ((")") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_PIPE) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("|") reduce 521)) 1239 ((("%") reduce 534) ((")") reduce 534) (("*") reduce 534) (("+") reduce 534) ((",") reduce 534) (("-") reduce 534) ((".") reduce 534) (("/") reduce 534) (("<") reduce 534) ((">") reduce 534) (("?") reduce 534) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_BOOLEAN_AND) reduce 534) ((T_BOOLEAN_OR) reduce 534) ((T_COALESCE) reduce 534) ((T_DOUBLE_ARROW) reduce 534) ((T_INSTANCEOF) reduce 534) ((T_IS_EQUAL) reduce 534) ((T_IS_GREATER_OR_EQUAL) reduce 534) ((T_IS_IDENTICAL) reduce 534) ((T_IS_NOT_EQUAL) reduce 534) ((T_IS_NOT_IDENTICAL) reduce 534) ((T_IS_SMALLER_OR_EQUAL) reduce 534) ((T_LOGICAL_AND) reduce 534) ((T_LOGICAL_OR) reduce 534) ((T_LOGICAL_XOR) reduce 534) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PIPE) reduce 534) ((T_POW) reduce 534) ((T_SL) reduce 534) ((T_SPACESHIP) reduce 534) ((T_SR) reduce 534) (("[") reduce 534) (("^") reduce 534) (("|") reduce 534)) 1240 ((("%") reduce 524) ((")") reduce 524) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_PIPE) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("^") reduce 524) (("|") reduce 524)) 1244 ((("%") reduce 517) (("(") shift) ((")") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_PIPE) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("^") reduce 517) (("|") reduce 517)) 1245 ((("%") reduce 533) ((")") reduce 533) (("*") reduce 533) (("+") reduce 533) ((",") reduce 533) (("-") reduce 533) ((".") reduce 533) (("/") reduce 533) (("<") reduce 533) ((">") reduce 533) (("?") reduce 533) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_BOOLEAN_AND) reduce 533) ((T_BOOLEAN_OR) reduce 533) ((T_COALESCE) reduce 533) ((T_DOUBLE_ARROW) reduce 533) ((T_INSTANCEOF) reduce 533) ((T_IS_EQUAL) reduce 533) ((T_IS_GREATER_OR_EQUAL) reduce 533) ((T_IS_IDENTICAL) reduce 533) ((T_IS_NOT_EQUAL) reduce 533) ((T_IS_NOT_IDENTICAL) reduce 533) ((T_IS_SMALLER_OR_EQUAL) reduce 533) ((T_LOGICAL_AND) reduce 533) ((T_LOGICAL_OR) reduce 533) ((T_LOGICAL_XOR) reduce 533) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PIPE) reduce 533) ((T_POW) reduce 533) ((T_SL) reduce 533) ((T_SPACESHIP) reduce 533) ((T_SR) reduce 533) (("[") reduce 533) (("^") reduce 533) (("|") reduce 533)) 1246 ((("%") reduce 537) ((")") reduce 537) (("*") reduce 537) (("+") reduce 537) ((",") reduce 537) (("-") reduce 537) ((".") reduce 537) (("/") reduce 537) (("<") reduce 537) ((">") reduce 537) (("?") reduce 537) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_BOOLEAN_AND) reduce 537) ((T_BOOLEAN_OR) reduce 537) ((T_COALESCE) reduce 537) ((T_DOUBLE_ARROW) reduce 537) ((T_INSTANCEOF) reduce 537) ((T_IS_EQUAL) reduce 537) ((T_IS_GREATER_OR_EQUAL) reduce 537) ((T_IS_IDENTICAL) reduce 537) ((T_IS_NOT_EQUAL) reduce 537) ((T_IS_NOT_IDENTICAL) reduce 537) ((T_IS_SMALLER_OR_EQUAL) reduce 537) ((T_LOGICAL_AND) reduce 537) ((T_LOGICAL_OR) reduce 537) ((T_LOGICAL_XOR) reduce 537) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_PIPE) reduce 537) ((T_POW) reduce 537) ((T_SL) reduce 537) ((T_SPACESHIP) reduce 537) ((T_SR) reduce 537) (("[") reduce 537) (("^") reduce 537) (("|") reduce 537)) 1252 ((("%") reduce 532) ((")") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PIPE) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("|") reduce 532)) 1254 ((("%") reduce 523) ((")") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_PIPE) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("^") reduce 523) (("|") reduce 523)) 1256 ((("%") reduce 536) ((")") reduce 536) (("*") reduce 536) (("+") reduce 536) ((",") reduce 536) (("-") reduce 536) ((".") reduce 536) (("/") reduce 536) (("<") reduce 536) ((">") reduce 536) (("?") reduce 536) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_BOOLEAN_AND) reduce 536) ((T_BOOLEAN_OR) reduce 536) ((T_COALESCE) reduce 536) ((T_DOUBLE_ARROW) reduce 536) ((T_INSTANCEOF) reduce 536) ((T_IS_EQUAL) reduce 536) ((T_IS_GREATER_OR_EQUAL) reduce 536) ((T_IS_IDENTICAL) reduce 536) ((T_IS_NOT_EQUAL) reduce 536) ((T_IS_NOT_IDENTICAL) reduce 536) ((T_IS_SMALLER_OR_EQUAL) reduce 536) ((T_LOGICAL_AND) reduce 536) ((T_LOGICAL_OR) reduce 536) ((T_LOGICAL_XOR) reduce 536) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 536) ((T_OBJECT_OPERATOR) reduce 536) ((T_PIPE) reduce 536) ((T_POW) reduce 536) ((T_SL) reduce 536) ((T_SPACESHIP) reduce 536) ((T_SR) reduce 536) (("[") reduce 536) (("^") reduce 536) (("|") reduce 536)) 1257 ((("%") reduce 94) (("(") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("|") reduce 94)) 1258 ((("%") reduce 93) (("(") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("|") reduce 93)) 1259 ((("%") reduce 95) (("(") reduce 95) ((")") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("|") reduce 95)) 1261 ((("%") reduce 539) ((")") reduce 539) (("*") reduce 539) (("+") reduce 539) ((",") reduce 539) (("-") reduce 539) ((".") reduce 539) (("/") reduce 539) (("<") reduce 539) ((">") reduce 539) (("?") reduce 539) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_BOOLEAN_AND) reduce 539) ((T_BOOLEAN_OR) reduce 539) ((T_COALESCE) reduce 539) ((T_DOUBLE_ARROW) reduce 539) ((T_INSTANCEOF) reduce 539) ((T_IS_EQUAL) reduce 539) ((T_IS_GREATER_OR_EQUAL) reduce 539) ((T_IS_IDENTICAL) reduce 539) ((T_IS_NOT_EQUAL) reduce 539) ((T_IS_NOT_IDENTICAL) reduce 539) ((T_IS_SMALLER_OR_EQUAL) reduce 539) ((T_LOGICAL_AND) reduce 539) ((T_LOGICAL_OR) reduce 539) ((T_LOGICAL_XOR) reduce 539) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 539) ((T_OBJECT_OPERATOR) reduce 539) ((T_PIPE) reduce 539) ((T_POW) reduce 539) ((T_SL) reduce 539) ((T_SPACESHIP) reduce 539) ((T_SR) reduce 539) (("[") reduce 539) (("^") reduce 539) (("|") reduce 539)) 1264 ((("%") reduce 538) ((")") reduce 538) (("*") reduce 538) (("+") reduce 538) ((",") reduce 538) (("-") reduce 538) ((".") reduce 538) (("/") reduce 538) (("<") reduce 538) ((">") reduce 538) (("?") reduce 538) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_BOOLEAN_AND) reduce 538) ((T_BOOLEAN_OR) reduce 538) ((T_COALESCE) reduce 538) ((T_DOUBLE_ARROW) reduce 538) ((T_INSTANCEOF) reduce 538) ((T_IS_EQUAL) reduce 538) ((T_IS_GREATER_OR_EQUAL) reduce 538) ((T_IS_IDENTICAL) reduce 538) ((T_IS_NOT_EQUAL) reduce 538) ((T_IS_NOT_IDENTICAL) reduce 538) ((T_IS_SMALLER_OR_EQUAL) reduce 538) ((T_LOGICAL_AND) reduce 538) ((T_LOGICAL_OR) reduce 538) ((T_LOGICAL_XOR) reduce 538) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 538) ((T_OBJECT_OPERATOR) reduce 538) ((T_PIPE) reduce 538) ((T_POW) reduce 538) ((T_SL) reduce 538) ((T_SPACESHIP) reduce 538) ((T_SR) reduce 538) (("[") reduce 538) (("^") reduce 538) (("|") reduce 538)) 1270 ((("%") reduce 92) (("(") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92)) 1273 ((("%") reduce 535) ((")") reduce 535) (("*") reduce 535) (("+") reduce 535) ((",") reduce 535) (("-") reduce 535) ((".") reduce 535) (("/") reduce 535) (("<") reduce 535) ((">") reduce 535) (("?") reduce 535) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_BOOLEAN_AND) reduce 535) ((T_BOOLEAN_OR) reduce 535) ((T_COALESCE) reduce 535) ((T_DOUBLE_ARROW) reduce 535) ((T_INSTANCEOF) reduce 535) ((T_IS_EQUAL) reduce 535) ((T_IS_GREATER_OR_EQUAL) reduce 535) ((T_IS_IDENTICAL) reduce 535) ((T_IS_NOT_EQUAL) reduce 535) ((T_IS_NOT_IDENTICAL) reduce 535) ((T_IS_SMALLER_OR_EQUAL) reduce 535) ((T_LOGICAL_AND) reduce 535) ((T_LOGICAL_OR) reduce 535) ((T_LOGICAL_XOR) reduce 535) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PIPE) reduce 535) ((T_POW) reduce 535) ((T_SL) reduce 535) ((T_SPACESHIP) reduce 535) ((T_SR) reduce 535) (("[") reduce 535) (("^") reduce 535) (("|") reduce 535)) 1275 ((("%") reduce 568) (("(") reduce 568) ((")") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) (("=") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AND_EQUAL) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_COALESCE_EQUAL) reduce 568) ((T_CONCAT_EQUAL) reduce 568) ((T_DEC) reduce 568) ((T_DIV_EQUAL) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INC) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_MINUS_EQUAL) reduce 568) ((T_MOD_EQUAL) reduce 568) ((T_MUL_EQUAL) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_OR_EQUAL) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_PLUS_EQUAL) reduce 568) ((T_POW) reduce 568) ((T_POW_EQUAL) reduce 568) ((T_SL) reduce 568) ((T_SL_EQUAL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) ((T_SR_EQUAL) reduce 568) ((T_XOR_EQUAL) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 1276 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 479) (("(") shift) ((")") reduce 479) (("*") reduce 479) (("+") shift) ((",") reduce 479) (("-") shift) ((".") reduce 479) (("/") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 479) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 479) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 479) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PIPE) reduce 479) ((T_POW) reduce 479) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 479) (("`") shift) (("|") reduce 479) (("~") shift)) 1283 ((("%") reduce 564) (("(") reduce 555) ((")") reduce 564) (("*") reduce 564) (("+") reduce 564) ((",") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) (("<") reduce 564) (("=") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AND_EQUAL) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_COALESCE_EQUAL) reduce 564) ((T_CONCAT_EQUAL) reduce 564) ((T_DEC) reduce 564) ((T_DIV_EQUAL) reduce 564) ((T_DOUBLE_ARROW) reduce 564) ((T_INC) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_MINUS_EQUAL) reduce 564) ((T_MOD_EQUAL) reduce 564) ((T_MUL_EQUAL) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_OR_EQUAL) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_PLUS_EQUAL) reduce 564) ((T_POW) reduce 564) ((T_POW_EQUAL) reduce 564) ((T_SL) reduce 564) ((T_SL_EQUAL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) ((T_SR_EQUAL) reduce 564) ((T_XOR_EQUAL) reduce 564) (("[") reduce 564) (("^") reduce 564) (("|") reduce 564)) 1284 ((("%") reduce 530) ((")") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PIPE) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 551) (("^") reduce 530) (("|") reduce 530)) 1286 ((("%") reduce 529) ((")") reduce 529) (("*") reduce 529) (("+") reduce 529) ((",") reduce 529) (("-") reduce 529) ((".") reduce 529) (("/") reduce 529) (("<") reduce 529) ((">") reduce 529) (("?") reduce 529) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_BOOLEAN_AND) reduce 529) ((T_BOOLEAN_OR) reduce 529) ((T_COALESCE) reduce 529) ((T_DOUBLE_ARROW) reduce 529) ((T_INSTANCEOF) reduce 529) ((T_IS_EQUAL) reduce 529) ((T_IS_GREATER_OR_EQUAL) reduce 529) ((T_IS_IDENTICAL) reduce 529) ((T_IS_NOT_EQUAL) reduce 529) ((T_IS_NOT_IDENTICAL) reduce 529) ((T_IS_SMALLER_OR_EQUAL) reduce 529) ((T_LOGICAL_AND) reduce 529) ((T_LOGICAL_OR) reduce 529) ((T_LOGICAL_XOR) reduce 529) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PIPE) reduce 529) ((T_POW) reduce 529) ((T_SL) reduce 529) ((T_SPACESHIP) reduce 529) ((T_SR) reduce 529) (("[") reduce 554) (("^") reduce 529) (("|") reduce 529)) 1287 ((("%") reduce 528) (("(") reduce 557) ((")") reduce 528) (("*") reduce 528) (("+") reduce 528) ((",") reduce 528) (("-") reduce 528) ((".") reduce 528) (("/") reduce 528) (("<") reduce 528) ((">") reduce 528) (("?") reduce 528) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_BOOLEAN_AND) reduce 528) ((T_BOOLEAN_OR) reduce 528) ((T_COALESCE) reduce 528) ((T_DOUBLE_ARROW) reduce 528) ((T_INSTANCEOF) reduce 528) ((T_IS_EQUAL) reduce 528) ((T_IS_GREATER_OR_EQUAL) reduce 528) ((T_IS_IDENTICAL) reduce 528) ((T_IS_NOT_EQUAL) reduce 528) ((T_IS_NOT_IDENTICAL) reduce 528) ((T_IS_SMALLER_OR_EQUAL) reduce 528) ((T_LOGICAL_AND) reduce 528) ((T_LOGICAL_OR) reduce 528) ((T_LOGICAL_XOR) reduce 528) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PIPE) reduce 528) ((T_POW) reduce 528) ((T_SL) reduce 528) ((T_SPACESHIP) reduce 528) ((T_SR) reduce 528) (("[") reduce 550) (("^") reduce 528) (("|") reduce 528)) 1288 ((("%") shift) ((")") reduce 480) (("*") shift) (("+") shift) ((",") reduce 480) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 480) ((T_LOGICAL_OR) reduce 480) ((T_LOGICAL_XOR) reduce 480) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1291 ((("%") reduce 563) (("(") reduce 563) ((")") reduce 563) (("*") reduce 563) (("+") reduce 563) ((",") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) (("<") reduce 563) (("=") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AND_EQUAL) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_COALESCE_EQUAL) reduce 563) ((T_CONCAT_EQUAL) reduce 563) ((T_DEC) reduce 563) ((T_DIV_EQUAL) reduce 563) ((T_DOUBLE_ARROW) reduce 563) ((T_INC) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_MINUS_EQUAL) reduce 563) ((T_MOD_EQUAL) reduce 563) ((T_MUL_EQUAL) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_OR_EQUAL) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_PLUS_EQUAL) reduce 563) ((T_POW) reduce 563) ((T_POW_EQUAL) reduce 563) ((T_SL) reduce 563) ((T_SL_EQUAL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) ((T_SR_EQUAL) reduce 563) ((T_XOR_EQUAL) reduce 563) (("[") reduce 563) (("^") reduce 563) (("|") reduce 563)) 1292 ((("%") reduce 484) ((")") reduce 484) (("*") reduce 484) (("+") reduce 484) ((",") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_DOUBLE_ARROW) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_PIPE) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("^") reduce 484) (("|") reduce 484)) 1293 ((("%") reduce 466) ((")") reduce 466) (("*") reduce 466) (("+") reduce 466) ((",") reduce 466) (("-") reduce 466) ((".") reduce 466) (("/") reduce 466) (("<") reduce 466) ((">") reduce 466) (("?") reduce 466) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_BOOLEAN_AND) reduce 466) ((T_BOOLEAN_OR) reduce 466) ((T_COALESCE) reduce 466) ((T_DOUBLE_ARROW) reduce 466) ((T_INSTANCEOF) reduce 466) ((T_IS_EQUAL) reduce 466) ((T_IS_GREATER_OR_EQUAL) reduce 466) ((T_IS_IDENTICAL) reduce 466) ((T_IS_NOT_EQUAL) reduce 466) ((T_IS_NOT_IDENTICAL) reduce 466) ((T_IS_SMALLER_OR_EQUAL) reduce 466) ((T_LOGICAL_AND) reduce 466) ((T_LOGICAL_OR) reduce 466) ((T_LOGICAL_XOR) reduce 466) ((T_PIPE) reduce 466) ((T_POW) reduce 466) ((T_SL) reduce 466) ((T_SPACESHIP) reduce 466) ((T_SR) reduce 466) (("^") reduce 466) (("|") reduce 466)) 1294 ((("%") reduce 488) ((")") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_PIPE) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("^") reduce 488) (("|") reduce 488)) 1295 ((("%") reduce 531) (("(") shift) ((")") reduce 531) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("|") reduce 531)) 1296 ((("%") reduce 461) (("(") reduce 558) ((")") reduce 461) (("*") reduce 461) (("+") reduce 461) ((",") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PIPE) reduce 461) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") reduce 552) (("^") reduce 461) (("|") reduce 461)) 1297 ((("%") reduce 462) ((")") reduce 462) (("*") reduce 462) (("+") reduce 462) ((",") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_DOUBLE_ARROW) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_PIPE) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 1298 ((("%") reduce 476) ((")") reduce 476) (("*") reduce 476) (("+") reduce 476) ((",") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_DOUBLE_ARROW) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_PIPE) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("|") reduce 476)) 1299 ((("%") reduce 559) (("(") reduce 559) ((")") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) (("=") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AND_EQUAL) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_COALESCE_EQUAL) reduce 559) ((T_CONCAT_EQUAL) reduce 559) ((T_DEC) reduce 559) ((T_DIV_EQUAL) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INC) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_MINUS_EQUAL) reduce 559) ((T_MOD_EQUAL) reduce 559) ((T_MUL_EQUAL) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_OR_EQUAL) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_PLUS_EQUAL) reduce 559) ((T_POW) reduce 559) ((T_POW_EQUAL) reduce 559) ((T_SL) reduce 559) ((T_SL_EQUAL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) ((T_SR_EQUAL) reduce 559) ((T_XOR_EQUAL) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 1300 ((("%") reduce 565) ((")") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 1301 ((("%") reduce 403) ((")") reduce 403) (("*") reduce 403) (("+") reduce 403) ((",") reduce 403) (("-") reduce 403) ((".") reduce 403) (("/") reduce 403) (("<") reduce 403) (("=") shift) ((">") reduce 403) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_DOUBLE_ARROW) reduce 403) ((T_INC) shift) ((T_INSTANCEOF) reduce 403) ((T_IS_EQUAL) reduce 403) ((T_IS_GREATER_OR_EQUAL) reduce 403) ((T_IS_IDENTICAL) reduce 403) ((T_IS_NOT_EQUAL) reduce 403) ((T_IS_NOT_IDENTICAL) reduce 403) ((T_IS_SMALLER_OR_EQUAL) reduce 403) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 403) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 403) ((T_POW_EQUAL) shift) ((T_SL) reduce 403) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 403) ((T_SR) reduce 403) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 548) (("^") reduce 403) (("|") reduce 403)) 1304 ((("%") reduce 448) ((")") reduce 448) (("*") reduce 448) (("+") reduce 448) ((",") reduce 448) (("-") reduce 448) ((".") reduce 448) (("/") reduce 448) (("<") reduce 448) ((">") reduce 448) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_BOOLEAN_AND) reduce 448) ((T_BOOLEAN_OR) reduce 448) ((T_COALESCE) reduce 448) ((T_DOUBLE_ARROW) reduce 448) ((T_INSTANCEOF) reduce 448) ((T_IS_EQUAL) reduce 448) ((T_IS_GREATER_OR_EQUAL) reduce 448) ((T_IS_IDENTICAL) reduce 448) ((T_IS_NOT_EQUAL) reduce 448) ((T_IS_NOT_IDENTICAL) reduce 448) ((T_IS_SMALLER_OR_EQUAL) reduce 448) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_PIPE) reduce 448) ((T_POW) shift) ((T_SL) reduce 448) ((T_SPACESHIP) reduce 448) ((T_SR) reduce 448) (("^") reduce 448) (("|") reduce 448)) 1336 ((("%") shift) ((")") reduce 432) (("*") shift) (("+") shift) ((",") reduce 432) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_DOUBLE_ARROW) reduce 432) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 432)) 1337 ((("%") shift) ((")") reduce 435) (("*") shift) (("+") shift) ((",") reduce 435) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_DOUBLE_ARROW) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("|") reduce 435)) 1338 ((("%") shift) ((")") reduce 444) (("*") shift) (("+") shift) ((",") reduce 444) (("-") shift) ((".") reduce 444) (("/") shift) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_PIPE) reduce 444) ((T_POW) shift) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("|") reduce 444)) 1339 ((("%") shift) ((")") reduce 458) (("*") shift) (("+") shift) ((",") reduce 458) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 458) (("|") reduce 458)) 1340 ((("%") shift) ((")") reduce 443) (("*") shift) (("+") shift) ((",") reduce 443) (("-") shift) ((".") reduce 443) (("/") shift) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_DOUBLE_ARROW) reduce 443) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_PIPE) reduce 443) ((T_POW) shift) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("^") reduce 443) (("|") reduce 443)) 1341 ((("%") reduce 440) ((")") reduce 440) (("*") reduce 440) (("+") reduce 440) ((",") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_DOUBLE_ARROW) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_PIPE) reduce 440) ((T_POW) shift) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("|") reduce 440)) 1342 ((("%") shift) ((")") reduce 453) (("*") shift) (("+") shift) ((",") reduce 453) (("-") shift) ((".") shift) (("/") shift) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_DOUBLE_ARROW) reduce 453) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_PIPE) reduce 453) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 453) ((T_SR) shift) (("^") reduce 453) (("|") reduce 453)) 1343 ((("%") shift) ((")") reduce 431) (("*") shift) (("+") shift) ((",") reduce 431) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1344 ((("%") shift) ((")") reduce 429) (("*") shift) (("+") shift) ((",") reduce 429) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 429) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1345 ((("%") shift) ((")") reduce 430) (("*") shift) (("+") shift) ((",") reduce 430) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 430) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1346 ((("%") shift) ((")") reduce 455) (("*") shift) (("+") shift) ((",") reduce 455) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 455) ((T_SR) shift) (("^") reduce 455) (("|") reduce 455)) 1347 ((("%") shift) ((")") reduce 450) (("*") shift) (("+") shift) ((",") reduce 450) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_DOUBLE_ARROW) reduce 450) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 450) (("|") reduce 450)) 1348 ((("%") shift) ((")") reduce 452) (("*") shift) (("+") shift) ((",") reduce 452) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_DOUBLE_ARROW) reduce 452) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 452) (("|") reduce 452)) 1349 ((("%") shift) ((")") reduce 449) (("*") shift) (("+") shift) ((",") reduce 449) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_DOUBLE_ARROW) reduce 449) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 449) (("|") reduce 449)) 1350 ((("%") shift) ((")") reduce 457) (("*") shift) (("+") shift) ((",") reduce 457) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 457) ((T_SR) shift) (("^") reduce 457) (("|") reduce 457)) 1351 ((("%") shift) ((")") reduce 451) (("*") shift) (("+") shift) ((",") reduce 451) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_DOUBLE_ARROW) reduce 451) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 451) (("|") reduce 451)) 1354 ((("%") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 1355 ((("%") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 1356 ((("%") reduce 95) ((")") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 1357 ((("%") reduce 509) ((")") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 1358 ((("%") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 1359 ((("%") reduce 568) ((")") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 1360 ((("%") reduce 511) ((")") reduce 511) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("|") reduce 511)) 1361 ((("%") reduce 459) ((")") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PIPE) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 1362 ((("%") reduce 510) ((")") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 1363 ((("%") reduce 512) ((")") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("|") reduce 512)) 1364 ((("%") reduce 573) ((")") reduce 573) (("*") reduce 573) (("+") reduce 573) ((",") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_DOUBLE_ARROW) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("|") reduce 573)) 1370 ((("%") reduce 574) ((")") reduce 574) (("*") reduce 574) (("+") reduce 574) ((",") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_DOUBLE_ARROW) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("|") reduce 574)) 1371 ((("%") reduce 578) ((")") reduce 578) (("*") reduce 578) (("+") reduce 578) ((",") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_DOUBLE_ARROW) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("|") reduce 578)) 1372 ((("%") reduce 582) ((")") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 1373 ((("%") reduce 575) ((")") reduce 575) (("*") reduce 575) (("+") reduce 575) ((",") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_DOUBLE_ARROW) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("|") reduce 575)) 1374 ((("%") reduce 584) ((")") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 1377 ((("%") reduce 583) ((")") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 1378 ((("%") reduce 576) ((")") reduce 576) (("*") reduce 576) (("+") reduce 576) ((",") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_DOUBLE_ARROW) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("|") reduce 576)) 1380 ((("%") reduce 577) ((")") reduce 577) (("*") reduce 577) (("+") reduce 577) ((",") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_DOUBLE_ARROW) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("|") reduce 577)) 1382 ((("%") reduce 513) ((")") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("|") reduce 513)) 1383 ((("%") reduce 570) ((")") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 1386 ((("%") reduce 569) ((")") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 1387 ((("%") shift) ((")") reduce 465) (("*") shift) (("+") shift) ((",") reduce 465) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 465) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 465) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1388 ((("%") shift) ((")") reduce 427) (("*") shift) (("+") shift) ((",") reduce 427) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_DOUBLE_ARROW) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1389 ((("%") shift) ((")") reduce 428) (("*") shift) (("+") shift) ((",") reduce 428) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_DOUBLE_ARROW) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1390 ((("%") shift) ((")") reduce 433) (("*") shift) (("+") shift) ((",") reduce 433) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_DOUBLE_ARROW) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("|") reduce 433)) 1391 ((("%") shift) ((")") reduce 434) (("*") shift) (("+") shift) ((",") reduce 434) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_DOUBLE_ARROW) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("|") reduce 434)) 1395 ((("%") shift) ((")") reduce 463) (("*") shift) (("+") shift) ((",") reduce 463) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 463) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1396 ((("%") shift) ((")") reduce 464) (("*") shift) (("+") shift) ((",") reduce 464) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 464) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1397 ((("%") shift) ((")") reduce 456) (("*") shift) (("+") shift) ((",") reduce 456) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 456) ((T_SR) shift) (("^") reduce 456) (("|") reduce 456)) 1398 ((("%") shift) ((")") reduce 454) (("*") shift) (("+") shift) ((",") reduce 454) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 454) ((T_SR) shift) (("^") reduce 454) (("|") reduce 454)) 1399 ((("%") reduce 441) ((")") reduce 441) (("*") reduce 441) (("+") reduce 441) ((",") reduce 441) (("-") reduce 441) ((".") reduce 441) (("/") reduce 441) (("<") reduce 441) ((">") reduce 441) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_DOUBLE_ARROW) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 441) ((T_IS_GREATER_OR_EQUAL) reduce 441) ((T_IS_IDENTICAL) reduce 441) ((T_IS_NOT_EQUAL) reduce 441) ((T_IS_NOT_IDENTICAL) reduce 441) ((T_IS_SMALLER_OR_EQUAL) reduce 441) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_PIPE) reduce 441) ((T_POW) shift) ((T_SL) reduce 441) ((T_SPACESHIP) reduce 441) ((T_SR) reduce 441) (("^") reduce 441) (("|") reduce 441)) 1400 ((("%") shift) ((")") reduce 436) (("*") shift) (("+") shift) ((",") reduce 436) (("-") shift) ((".") reduce 436) (("/") shift) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_DOUBLE_ARROW) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_PIPE) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 436) ((T_SR) shift) (("^") reduce 436) (("|") reduce 436)) 1401 ((("%") shift) ((")") reduce 438) (("*") shift) (("+") reduce 438) ((",") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") shift) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_DOUBLE_ARROW) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_PIPE) reduce 438) ((T_POW) shift) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("|") reduce 438)) 1402 ((("%") shift) ((")") reduce 437) (("*") shift) (("+") reduce 437) ((",") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") shift) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_DOUBLE_ARROW) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_PIPE) reduce 437) ((T_POW) shift) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("|") reduce 437)) 1403 ((("%") reduce 439) ((")") reduce 439) (("*") reduce 439) (("+") reduce 439) ((",") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_DOUBLE_ARROW) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_PIPE) reduce 439) ((T_POW) shift) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("|") reduce 439)) 1404 ((("%") reduce 442) ((")") reduce 442) (("*") reduce 442) (("+") reduce 442) ((",") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_DOUBLE_ARROW) reduce 442) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_PIPE) reduce 442) ((T_POW) shift) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("|") reduce 442)) 1406 ((("%") reduce 74) (("(") reduce 74) ((")") reduce 74) (("*") reduce 74) (("+") reduce 74) ((",") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_DOUBLE_ARROW) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_PIPE) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("|") reduce 74)) 1407 ((("%") reduce 52) (("(") reduce 52) ((")") reduce 52) (("*") reduce 52) (("+") reduce 52) ((",") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_DOUBLE_ARROW) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_PIPE) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("|") reduce 52)) 1408 ((("%") reduce 27) (("(") reduce 27) ((")") reduce 27) (("*") reduce 27) (("+") reduce 27) ((",") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_DOUBLE_ARROW) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_PIPE) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("|") reduce 27)) 1409 ((("%") reduce 51) (("(") reduce 51) ((")") reduce 51) (("*") reduce 51) (("+") reduce 51) ((",") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_DOUBLE_ARROW) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_PIPE) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("|") reduce 51)) 1410 ((("%") reduce 53) (("(") reduce 53) ((")") reduce 53) (("*") reduce 53) (("+") reduce 53) ((",") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_DOUBLE_ARROW) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_PIPE) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("|") reduce 53)) 1411 ((("%") reduce 49) (("(") reduce 49) ((")") reduce 49) (("*") reduce 49) (("+") reduce 49) ((",") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_DOUBLE_ARROW) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_PIPE) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("|") reduce 49)) 1412 ((("%") reduce 29) (("(") reduce 29) ((")") reduce 29) (("*") reduce 29) (("+") reduce 29) ((",") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_DOUBLE_ARROW) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_PIPE) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("|") reduce 29)) 1413 ((("%") reduce 59) (("(") reduce 59) ((")") reduce 59) (("*") reduce 59) (("+") reduce 59) ((",") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_DOUBLE_ARROW) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_PIPE) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("|") reduce 59)) 1414 ((("%") reduce 60) (("(") reduce 60) ((")") reduce 60) (("*") reduce 60) (("+") reduce 60) ((",") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_DOUBLE_ARROW) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_PIPE) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("|") reduce 60)) 1415 ((("%") reduce 11) (("(") reduce 11) ((")") reduce 11) (("*") reduce 11) (("+") reduce 11) ((",") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_DOUBLE_ARROW) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_PIPE) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("|") reduce 11)) 1416 ((("%") reduce 42) (("(") reduce 42) ((")") reduce 42) (("*") reduce 42) (("+") reduce 42) ((",") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_DOUBLE_ARROW) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_PIPE) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("|") reduce 42)) 1417 ((("%") reduce 39) (("(") reduce 39) ((")") reduce 39) (("*") reduce 39) (("+") reduce 39) ((",") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_DOUBLE_ARROW) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_PIPE) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("|") reduce 39)) 1418 ((("%") reduce 25) (("(") reduce 25) ((")") reduce 25) (("*") reduce 25) (("+") reduce 25) ((",") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_DOUBLE_ARROW) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_PIPE) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("|") reduce 25)) 1419 ((("%") reduce 50) (("(") reduce 50) ((")") reduce 50) (("*") reduce 50) (("+") reduce 50) ((",") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_DOUBLE_ARROW) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_PIPE) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("|") reduce 50)) 1420 ((("%") reduce 66) (("(") reduce 66) ((")") reduce 66) (("*") reduce 66) (("+") reduce 66) ((",") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_DOUBLE_ARROW) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_PIPE) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("|") reduce 66)) 1421 ((("%") reduce 18) (("(") reduce 18) ((")") reduce 18) (("*") reduce 18) (("+") reduce 18) ((",") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_DOUBLE_ARROW) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_PIPE) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("|") reduce 18)) 1422 ((("%") reduce 17) (("(") reduce 17) ((")") reduce 17) (("*") reduce 17) (("+") reduce 17) ((",") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_DOUBLE_ARROW) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_PIPE) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("|") reduce 17)) 1423 ((("%") reduce 15) (("(") reduce 15) ((")") reduce 15) (("*") reduce 15) (("+") reduce 15) ((",") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_DOUBLE_ARROW) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_PIPE) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("|") reduce 15)) 1424 ((("%") reduce 14) (("(") reduce 14) ((")") reduce 14) (("*") reduce 14) (("+") reduce 14) ((",") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_DOUBLE_ARROW) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_PIPE) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("|") reduce 14)) 1425 ((("%") reduce 38) (("(") reduce 38) ((")") reduce 38) (("*") reduce 38) (("+") reduce 38) ((",") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_DOUBLE_ARROW) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_PIPE) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("|") reduce 38)) 1426 ((("%") reduce 26) (("(") reduce 26) ((")") reduce 26) (("*") reduce 26) (("+") reduce 26) ((",") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_DOUBLE_ARROW) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_PIPE) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("|") reduce 26)) 1427 ((("%") reduce 22) (("(") reduce 22) ((")") reduce 22) (("*") reduce 22) (("+") reduce 22) ((",") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_DOUBLE_ARROW) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_PIPE) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("|") reduce 22)) 1428 ((("%") reduce 24) (("(") reduce 24) ((")") reduce 24) (("*") reduce 24) (("+") reduce 24) ((",") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_DOUBLE_ARROW) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_PIPE) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("|") reduce 24)) 1429 ((("%") reduce 16) (("(") reduce 16) ((")") reduce 16) (("*") reduce 16) (("+") reduce 16) ((",") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_DOUBLE_ARROW) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_PIPE) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("|") reduce 16)) 1430 ((("%") reduce 48) (("(") reduce 48) ((")") reduce 48) (("*") reduce 48) (("+") reduce 48) ((",") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_DOUBLE_ARROW) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_PIPE) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("|") reduce 48)) 1431 ((("%") reduce 20) (("(") reduce 20) ((")") reduce 20) (("*") reduce 20) (("+") reduce 20) ((",") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_DOUBLE_ARROW) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_PIPE) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("|") reduce 20)) 1432 ((("%") reduce 70) (("(") reduce 70) ((")") reduce 70) (("*") reduce 70) (("+") reduce 70) ((",") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_DOUBLE_ARROW) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_PIPE) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("|") reduce 70)) 1433 ((("%") reduce 3) (("(") reduce 3) ((")") reduce 3) (("*") reduce 3) (("+") reduce 3) ((",") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_DOUBLE_ARROW) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_PIPE) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("|") reduce 3)) 1434 ((("%") reduce 12) (("(") reduce 12) ((")") reduce 12) (("*") reduce 12) (("+") reduce 12) ((",") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_DOUBLE_ARROW) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_PIPE) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("|") reduce 12)) 1435 ((("%") reduce 54) (("(") reduce 54) ((")") reduce 54) (("*") reduce 54) (("+") reduce 54) ((",") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_DOUBLE_ARROW) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_PIPE) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("|") reduce 54)) 1436 ((("%") reduce 65) (("(") reduce 65) ((")") reduce 65) (("*") reduce 65) (("+") reduce 65) ((",") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_DOUBLE_ARROW) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_PIPE) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("|") reduce 65)) 1437 ((("%") reduce 75) (("(") reduce 75) ((")") reduce 75) (("*") reduce 75) (("+") reduce 75) ((",") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_DOUBLE_ARROW) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_PIPE) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("|") reduce 75)) 1438 ((("%") reduce 30) (("(") reduce 30) ((")") reduce 30) (("*") reduce 30) (("+") reduce 30) ((",") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_DOUBLE_ARROW) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_PIPE) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("|") reduce 30)) 1439 ((("%") reduce 68) (("(") reduce 68) ((")") reduce 68) (("*") reduce 68) (("+") reduce 68) ((",") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_DOUBLE_ARROW) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_PIPE) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("|") reduce 68)) 1440 ((("%") reduce 21) (("(") reduce 21) ((")") reduce 21) (("*") reduce 21) (("+") reduce 21) ((",") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_DOUBLE_ARROW) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_PIPE) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("|") reduce 21)) 1441 ((("%") reduce 23) (("(") reduce 23) ((")") reduce 23) (("*") reduce 23) (("+") reduce 23) ((",") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_DOUBLE_ARROW) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_PIPE) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("|") reduce 23)) 1442 ((("%") reduce 41) (("(") reduce 41) ((")") reduce 41) (("*") reduce 41) (("+") reduce 41) ((",") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_DOUBLE_ARROW) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_PIPE) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("|") reduce 41)) 1443 ((("%") reduce 62) (("(") reduce 62) ((")") reduce 62) (("*") reduce 62) (("+") reduce 62) ((",") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_DOUBLE_ARROW) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_PIPE) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("|") reduce 62)) 1444 ((("%") reduce 34) (("(") reduce 34) ((")") reduce 34) (("*") reduce 34) (("+") reduce 34) ((",") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_DOUBLE_ARROW) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_PIPE) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("|") reduce 34)) 1445 ((("%") reduce 40) (("(") reduce 40) ((")") reduce 40) (("*") reduce 40) (("+") reduce 40) ((",") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_DOUBLE_ARROW) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_PIPE) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("|") reduce 40)) 1446 ((("%") reduce 13) (("(") reduce 13) ((")") reduce 13) (("*") reduce 13) (("+") reduce 13) ((",") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_DOUBLE_ARROW) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_PIPE) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("|") reduce 13)) 1447 ((("%") reduce 55) (("(") reduce 55) ((")") reduce 55) (("*") reduce 55) (("+") reduce 55) ((",") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_DOUBLE_ARROW) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_PIPE) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("|") reduce 55)) 1448 ((("%") reduce 1) (("(") reduce 1) ((")") reduce 1) (("*") reduce 1) (("+") reduce 1) ((",") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_DOUBLE_ARROW) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_PIPE) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("|") reduce 1)) 1449 ((("%") reduce 2) (("(") reduce 2) ((")") reduce 2) (("*") reduce 2) (("+") reduce 2) ((",") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_DOUBLE_ARROW) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_PIPE) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("|") reduce 2)) 1450 ((("%") reduce 9) (("(") reduce 9) ((")") reduce 9) (("*") reduce 9) (("+") reduce 9) ((",") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_DOUBLE_ARROW) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_PIPE) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("|") reduce 9)) 1451 ((("%") reduce 33) (("(") reduce 33) ((")") reduce 33) (("*") reduce 33) (("+") reduce 33) ((",") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_DOUBLE_ARROW) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_PIPE) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("|") reduce 33)) 1452 ((("%") reduce 58) (("(") reduce 58) ((")") reduce 58) (("*") reduce 58) (("+") reduce 58) ((",") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_DOUBLE_ARROW) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_PIPE) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("|") reduce 58)) 1453 ((("%") reduce 37) (("(") reduce 37) ((")") reduce 37) (("*") reduce 37) (("+") reduce 37) ((",") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_DOUBLE_ARROW) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_PIPE) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("|") reduce 37)) 1454 ((("%") reduce 64) (("(") reduce 64) ((")") reduce 64) (("*") reduce 64) (("+") reduce 64) ((",") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_DOUBLE_ARROW) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_PIPE) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("|") reduce 64)) 1455 ((("%") reduce 46) (("(") reduce 46) ((")") reduce 46) (("*") reduce 46) (("+") reduce 46) ((",") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_DOUBLE_ARROW) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_PIPE) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("|") reduce 46)) 1456 ((("%") reduce 8) (("(") reduce 8) ((")") reduce 8) (("*") reduce 8) (("+") reduce 8) ((",") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_DOUBLE_ARROW) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_PIPE) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("|") reduce 8)) 1457 ((("%") reduce 6) (("(") reduce 6) ((")") reduce 6) (("*") reduce 6) (("+") reduce 6) ((",") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_DOUBLE_ARROW) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_PIPE) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("|") reduce 6)) 1458 ((("%") reduce 7) (("(") reduce 7) ((")") reduce 7) (("*") reduce 7) (("+") reduce 7) ((",") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_DOUBLE_ARROW) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_PIPE) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("|") reduce 7)) 1459 ((("%") reduce 69) (("(") reduce 69) ((")") reduce 69) (("*") reduce 69) (("+") reduce 69) ((",") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_DOUBLE_ARROW) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_PIPE) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("|") reduce 69)) 1460 ((("%") reduce 63) (("(") reduce 63) ((")") reduce 63) (("*") reduce 63) (("+") reduce 63) ((",") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_DOUBLE_ARROW) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_PIPE) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("|") reduce 63)) 1461 ((("%") reduce 56) (("(") reduce 56) ((")") reduce 56) (("*") reduce 56) (("+") reduce 56) ((",") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_DOUBLE_ARROW) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_PIPE) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("|") reduce 56)) 1462 ((("%") reduce 10) (("(") reduce 10) ((")") reduce 10) (("*") reduce 10) (("+") reduce 10) ((",") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_DOUBLE_ARROW) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_PIPE) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("|") reduce 10)) 1463 ((("%") reduce 67) (("(") reduce 67) ((")") reduce 67) (("*") reduce 67) (("+") reduce 67) ((",") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_DOUBLE_ARROW) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_PIPE) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("|") reduce 67)) 1464 ((("%") reduce 44) (("(") reduce 44) ((")") reduce 44) (("*") reduce 44) (("+") reduce 44) ((",") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_DOUBLE_ARROW) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_PIPE) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("|") reduce 44)) 1465 ((("%") reduce 76) (("(") reduce 76) ((")") reduce 76) (("*") reduce 76) (("+") reduce 76) ((",") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_DOUBLE_ARROW) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_PIPE) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("|") reduce 76)) 1466 ((("%") reduce 71) (("(") reduce 71) ((")") reduce 71) (("*") reduce 71) (("+") reduce 71) ((",") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_DOUBLE_ARROW) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_PIPE) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("|") reduce 71)) 1467 ((("%") reduce 77) (("(") reduce 77) ((")") reduce 77) (("*") reduce 77) (("+") reduce 77) ((",") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_DOUBLE_ARROW) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_PIPE) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("|") reduce 77)) 1468 ((("%") reduce 78) (("(") reduce 78) ((")") reduce 78) (("*") reduce 78) (("+") reduce 78) ((",") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_DOUBLE_ARROW) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_PIPE) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("|") reduce 78)) 1469 ((("%") reduce 79) (("(") reduce 79) ((")") reduce 79) (("*") reduce 79) (("+") reduce 79) ((",") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_DOUBLE_ARROW) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_PIPE) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("|") reduce 79)) 1470 ((("%") reduce 4) (("(") reduce 4) ((")") reduce 4) (("*") reduce 4) (("+") reduce 4) ((",") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_DOUBLE_ARROW) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_PIPE) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("|") reduce 4)) 1471 ((("%") reduce 5) (("(") reduce 5) ((")") reduce 5) (("*") reduce 5) (("+") reduce 5) ((",") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_DOUBLE_ARROW) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_PIPE) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("|") reduce 5)) 1472 ((("%") reduce 43) (("(") reduce 43) ((")") reduce 43) (("*") reduce 43) (("+") reduce 43) ((",") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_DOUBLE_ARROW) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_PIPE) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("|") reduce 43)) 1473 ((("%") reduce 73) (("(") reduce 73) ((")") reduce 73) (("*") reduce 73) (("+") reduce 73) ((",") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_DOUBLE_ARROW) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_PIPE) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("|") reduce 73)) 1474 ((("%") reduce 82) (("(") reduce 82) ((")") reduce 82) (("*") reduce 82) (("+") reduce 82) ((",") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_DOUBLE_ARROW) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_PIPE) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("|") reduce 82)) 1475 ((("%") reduce 47) (("(") reduce 47) ((")") reduce 47) (("*") reduce 47) (("+") reduce 47) ((",") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_DOUBLE_ARROW) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_PIPE) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("|") reduce 47)) 1476 ((("%") reduce 31) (("(") reduce 31) ((")") reduce 31) (("*") reduce 31) (("+") reduce 31) ((",") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_DOUBLE_ARROW) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_PIPE) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("|") reduce 31)) 1477 ((("%") reduce 57) (("(") reduce 57) ((")") reduce 57) (("*") reduce 57) (("+") reduce 57) ((",") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_DOUBLE_ARROW) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_PIPE) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("|") reduce 57)) 1478 ((("%") reduce 61) (("(") reduce 61) ((")") reduce 61) (("*") reduce 61) (("+") reduce 61) ((",") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_DOUBLE_ARROW) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_PIPE) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("|") reduce 61)) 1479 ((("%") reduce 28) (("(") reduce 28) ((")") reduce 28) (("*") reduce 28) (("+") reduce 28) ((",") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_DOUBLE_ARROW) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_PIPE) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("|") reduce 28)) 1480 ((("%") reduce 36) (("(") reduce 36) ((")") reduce 36) (("*") reduce 36) (("+") reduce 36) ((",") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_DOUBLE_ARROW) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_PIPE) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("|") reduce 36)) 1481 ((("%") reduce 32) (("(") reduce 32) ((")") reduce 32) (("*") reduce 32) (("+") reduce 32) ((",") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_DOUBLE_ARROW) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_PIPE) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("|") reduce 32)) 1482 ((("%") reduce 35) (("(") reduce 35) ((")") reduce 35) (("*") reduce 35) (("+") reduce 35) ((",") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_DOUBLE_ARROW) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_PIPE) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("|") reduce 35)) 1483 ((("%") reduce 19) (("(") reduce 19) ((")") reduce 19) (("*") reduce 19) (("+") reduce 19) ((",") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_DOUBLE_ARROW) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_PIPE) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("|") reduce 19)) 1484 ((("%") reduce 45) (("(") reduce 45) ((")") reduce 45) (("*") reduce 45) (("+") reduce 45) ((",") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_DOUBLE_ARROW) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_PIPE) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("|") reduce 45)) 1485 ((("%") reduce 542) (("(") reduce 579) ((")") reduce 542) (("*") reduce 542) (("+") reduce 542) ((",") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PIPE) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 1487 ((("%") reduce 72) (("(") reduce 72) ((")") reduce 72) (("*") reduce 72) (("+") reduce 72) ((",") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_DOUBLE_ARROW) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_PIPE) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("|") reduce 72)) 1488 ((("%") reduce 83) (("(") reduce 83) ((")") reduce 83) (("*") reduce 83) (("+") reduce 83) ((",") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_DOUBLE_ARROW) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_PIPE) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("|") reduce 83)) 1489 ((("%") reduce 572) (("(") reduce 581) ((")") reduce 572) (("*") reduce 572) (("+") reduce 572) ((",") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) (("<") reduce 572) (("=") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AND_EQUAL) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_COALESCE_EQUAL) reduce 572) ((T_CONCAT_EQUAL) reduce 572) ((T_DEC) reduce 572) ((T_DIV_EQUAL) reduce 572) ((T_DOUBLE_ARROW) reduce 572) ((T_INC) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_MINUS_EQUAL) reduce 572) ((T_MOD_EQUAL) reduce 572) ((T_MUL_EQUAL) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_OR_EQUAL) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_PLUS_EQUAL) reduce 572) ((T_POW) reduce 572) ((T_POW_EQUAL) reduce 572) ((T_SL) reduce 572) ((T_SL_EQUAL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) ((T_SR_EQUAL) reduce 572) ((T_XOR_EQUAL) reduce 572) (("[") reduce 572) (("^") reduce 572) (("|") reduce 572)) 1492 ((("%") reduce 544) (("(") reduce 580) ((")") reduce 544) (("*") reduce 544) (("+") reduce 544) ((",") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PIPE) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 1494 ((("%") reduce 507) (("(") reduce 507) ((")") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("|") reduce 507)) 1495 ((("%") reduce 291) (("(") reduce 291) ((")") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 1498 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) ((")") reduce 126) (("+") shift) (("-") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_AS) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CALLABLE) shift) ((T_CASE) shift) ((T_CATCH) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONST) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DEFAULT) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_ELLIPSIS) shift) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) shift) ((T_ENDDECLARE) shift) ((T_ENDFOR) shift) ((T_ENDFOREACH) shift) ((T_ENDIF) shift) ((T_ENDSWITCH) shift) ((T_ENDWHILE) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_EXTENDS) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FINALLY) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_IF) shift) ((T_IMPLEMENTS) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) shift) ((T_INSTEADOF) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAMESPACE) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PRIVATE) shift) ((T_PROPERTY_C) shift) ((T_PROTECTED) shift) ((T_PUBLIC) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_USE) shift) ((T_VAR) shift) ((T_VARIABLE) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 1500 ((("%") reduce 292) (("(") reduce 292) ((")") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) (("=") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AND_EQUAL) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_COALESCE_EQUAL) reduce 292) ((T_CONCAT_EQUAL) reduce 292) ((T_DEC) reduce 292) ((T_DIV_EQUAL) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INC) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_MINUS_EQUAL) reduce 292) ((T_MOD_EQUAL) reduce 292) ((T_MUL_EQUAL) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_OR_EQUAL) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_PLUS_EQUAL) reduce 292) ((T_POW) reduce 292) ((T_POW_EQUAL) reduce 292) ((T_SL) reduce 292) ((T_SL_EQUAL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) ((T_SR_EQUAL) reduce 292) ((T_XOR_EQUAL) reduce 292) (("[") reduce 292) (("^") reduce 292) (("|") reduce 292)) 1502 (((")") reduce 295) ((",") reduce 295)) 1503 ((("%") shift) ((")") reduce 304) (("*") shift) (("+") shift) ((",") reduce 304) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1504 ((("%") reduce 293) (("(") reduce 293) ((")") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) (("=") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AND_EQUAL) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_COALESCE_EQUAL) reduce 293) ((T_CONCAT_EQUAL) reduce 293) ((T_DEC) reduce 293) ((T_DIV_EQUAL) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INC) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_MINUS_EQUAL) reduce 293) ((T_MOD_EQUAL) reduce 293) ((T_MUL_EQUAL) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_OR_EQUAL) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_PLUS_EQUAL) reduce 293) ((T_POW) reduce 293) ((T_POW_EQUAL) reduce 293) ((T_SL) reduce 293) ((T_SL_EQUAL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) ((T_SR_EQUAL) reduce 293) ((T_XOR_EQUAL) reduce 293) (("[") reduce 293) (("^") reduce 293) (("|") reduce 293)) 1505 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 1509 ((("%") reduce 425) ((")") reduce 425) (("*") reduce 425) (("+") reduce 425) ((",") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_DOUBLE_ARROW) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_PIPE) reduce 425) ((T_POW) reduce 425) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 1511 ((("%") reduce 423) ((")") reduce 423) (("*") reduce 423) (("+") reduce 423) ((",") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_DOUBLE_ARROW) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_PIPE) reduce 423) ((T_POW) reduce 423) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 1521 ((("%") shift) ((")") reduce 419) (("*") shift) (("+") shift) ((",") reduce 419) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1522 ((("%") shift) ((")") reduce 421) (("*") shift) (("+") shift) ((",") reduce 421) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1523 ((("%") shift) ((")") reduce 420) (("*") shift) (("+") shift) ((",") reduce 420) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1524 ((("%") shift) ((")") reduce 413) (("*") shift) (("+") shift) ((",") reduce 413) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 413) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1525 ((("%") shift) ((")") reduce 410) (("*") shift) (("+") shift) ((",") reduce 410) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 410) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1526 ((("%") shift) ((")") reduce 418) (("*") shift) (("+") shift) ((",") reduce 418) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1527 ((("%") shift) ((")") reduce 412) (("*") shift) (("+") shift) ((",") reduce 412) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1528 ((("%") shift) ((")") reduce 416) (("*") shift) (("+") shift) ((",") reduce 416) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1529 ((("%") shift) ((")") reduce 411) (("*") shift) (("+") shift) ((",") reduce 411) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1530 ((("%") shift) ((")") reduce 414) (("*") shift) (("+") shift) ((",") reduce 414) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 414) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1531 ((("%") shift) ((")") reduce 415) (("*") shift) (("+") shift) ((",") reduce 415) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 415) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1532 ((("%") shift) ((")") reduce 422) (("*") shift) (("+") shift) ((",") reduce 422) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1533 ((("%") shift) ((")") reduce 417) (("*") shift) (("+") shift) ((",") reduce 417) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1534 ((("\"") reduce 80) (("$") reduce 80) (("(") reduce 80) ((T_ARRAY) reduce 80) ((T_CLASS_C) reduce 80) ((T_CONSTANT_ENCAPSED_STRING) reduce 80) ((T_DIR) reduce 80) ((T_FILE) reduce 80) ((T_FUNC_C) reduce 80) ((T_LINE) reduce 80) ((T_METHOD_C) reduce 80) ((T_NAME_FULLY_QUALIFIED) reduce 80) ((T_NAME_QUALIFIED) reduce 80) ((T_NAME_RELATIVE) reduce 80) ((T_NEW) reduce 80) ((T_NS_C) reduce 80) ((T_PROPERTY_C) reduce 80) ((T_READONLY) reduce 80) ((T_STATIC) reduce 80) ((T_STRING) reduce 80) ((T_TRAIT_C) reduce 80) ((T_VARIABLE) reduce 80) (("[") reduce 80)) 1535 ((("\"") reduce 81) (("$") reduce 81) (("(") reduce 81) ((T_ARRAY) reduce 81) ((T_CLASS_C) reduce 81) ((T_CONSTANT_ENCAPSED_STRING) reduce 81) ((T_DIR) reduce 81) ((T_FILE) reduce 81) ((T_FUNC_C) reduce 81) ((T_LINE) reduce 81) ((T_METHOD_C) reduce 81) ((T_NAME_FULLY_QUALIFIED) reduce 81) ((T_NAME_QUALIFIED) reduce 81) ((T_NAME_RELATIVE) reduce 81) ((T_NEW) reduce 81) ((T_NS_C) reduce 81) ((T_PROPERTY_C) reduce 81) ((T_READONLY) reduce 81) ((T_STATIC) reduce 81) ((T_STRING) reduce 81) ((T_TRAIT_C) reduce 81) ((T_VARIABLE) reduce 81) (("[") reduce 81)) 1537 ((("%") shift) ((")") reduce 406) (("*") shift) (("+") shift) ((",") reduce 406) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1542 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 540) ((T_OBJECT_OPERATOR) reduce 540) (("[") reduce 540)) 1543 ((("(") reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) (("[") reduce 521)) 1544 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) (("[") reduce 534)) 1545 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) (("[") reduce 533)) 1546 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) (("[") reduce 537)) 1547 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) (("[") reduce 532)) 1548 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 536) ((T_OBJECT_OPERATOR) reduce 536) (("[") reduce 536)) 1549 ((("(") reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) (("[") reduce 94)) 1550 ((("(") reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) (("[") reduce 93)) 1551 ((("(") reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) (("[") reduce 95)) 1553 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 539) ((T_OBJECT_OPERATOR) reduce 539) (("[") reduce 539)) 1554 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 538) ((T_OBJECT_OPERATOR) reduce 538) (("[") reduce 538)) 1556 (((T_PAAMAYIM_NEKUDOTAYIM) reduce 509)) 1557 ((("(") reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) (("[") reduce 92)) 1558 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) (("[") reduce 535)) 1559 ((("%") reduce 568) (("(") reduce 568) ((")") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 1563 ((("%") reduce 564) (("(") reduce 555) ((")") reduce 564) (("*") reduce 564) (("+") reduce 564) ((",") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) (("<") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_DOUBLE_ARROW) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_POW) reduce 564) ((T_SL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) (("[") reduce 564) (("^") reduce 564) (("|") reduce 564)) 1564 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) (("[") reduce 551)) 1566 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) (("[") reduce 554)) 1567 ((("(") reduce 557) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) (("[") reduce 550)) 1568 ((("%") reduce 563) (("(") reduce 563) ((")") reduce 563) (("*") reduce 563) (("+") reduce 563) ((",") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) (("<") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_DOUBLE_ARROW) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_POW) reduce 563) ((T_SL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) (("[") reduce 563) (("^") reduce 563) (("|") reduce 563)) 1569 ((("(") shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) (("[") reduce 531)) 1570 ((("(") reduce 558) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) (("[") reduce 552)) 1571 ((("%") reduce 559) (("(") reduce 559) ((")") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 1572 ((("%") reduce 565) ((")") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 1573 ((("%") reduce 407) ((")") reduce 407) (("*") reduce 407) (("+") reduce 407) ((",") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_DOUBLE_ARROW) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("[") reduce 548) (("^") reduce 407) (("|") reduce 407)) 1576 ((("(") reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) (("[") reduce 74)) 1577 ((("(") reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) (("[") reduce 52)) 1578 ((("(") reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) (("[") reduce 27)) 1579 ((("(") reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) (("[") reduce 51)) 1580 ((("(") reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) (("[") reduce 53)) 1581 ((("(") reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) (("[") reduce 49)) 1582 ((("(") reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) (("[") reduce 29)) 1583 ((("(") reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) (("[") reduce 59)) 1584 ((("(") reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) (("[") reduce 60)) 1585 ((("(") reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) (("[") reduce 11)) 1586 ((("(") reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) (("[") reduce 42)) 1587 ((("(") reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) (("[") reduce 39)) 1588 ((("(") reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) (("[") reduce 25)) 1589 ((("(") reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) (("[") reduce 50)) 1590 ((("(") reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) (("[") reduce 66)) 1591 ((("(") reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) (("[") reduce 18)) 1592 ((("(") reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) (("[") reduce 17)) 1593 ((("(") reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) (("[") reduce 15)) 1594 ((("(") reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) (("[") reduce 14)) 1595 ((("(") reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) (("[") reduce 38)) 1596 ((("(") reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) (("[") reduce 26)) 1597 ((("(") reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) (("[") reduce 22)) 1598 ((("(") reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) (("[") reduce 24)) 1599 ((("(") reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) (("[") reduce 16)) 1600 ((("(") reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) (("[") reduce 48)) 1601 ((("(") reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) (("[") reduce 20)) 1602 ((("(") reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) (("[") reduce 70)) 1603 ((("(") reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) (("[") reduce 3)) 1604 ((("(") reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) (("[") reduce 12)) 1605 ((("(") reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) (("[") reduce 54)) 1606 ((("(") reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) (("[") reduce 65)) 1607 ((("(") reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) (("[") reduce 75)) 1608 ((("(") reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) (("[") reduce 30)) 1609 ((("(") reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) (("[") reduce 68)) 1610 ((("(") reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) (("[") reduce 21)) 1611 ((("(") reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) (("[") reduce 23)) 1612 ((("(") reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) (("[") reduce 41)) 1613 ((("(") reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) (("[") reduce 62)) 1614 ((("(") reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) (("[") reduce 34)) 1615 ((("(") reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) (("[") reduce 40)) 1616 ((("(") reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) (("[") reduce 13)) 1617 ((("(") reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) (("[") reduce 55)) 1618 ((("(") reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) (("[") reduce 1)) 1619 ((("(") reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) (("[") reduce 2)) 1620 ((("(") reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) (("[") reduce 9)) 1621 ((("(") reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) (("[") reduce 33)) 1622 ((("(") reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) (("[") reduce 58)) 1623 ((("(") reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) (("[") reduce 37)) 1624 ((("(") reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) (("[") reduce 64)) 1625 ((("(") reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) (("[") reduce 46)) 1626 ((("(") reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) (("[") reduce 8)) 1627 ((("(") reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) (("[") reduce 6)) 1628 ((("(") reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) (("[") reduce 7)) 1629 ((("(") reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) (("[") reduce 69)) 1630 ((("(") reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) (("[") reduce 63)) 1631 ((("(") reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) (("[") reduce 56)) 1632 ((("(") reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) (("[") reduce 10)) 1633 ((("(") reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) (("[") reduce 67)) 1634 ((("(") reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) (("[") reduce 44)) 1635 ((("(") reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) (("[") reduce 76)) 1636 ((("(") reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) (("[") reduce 71)) 1637 ((("(") reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) (("[") reduce 77)) 1638 ((("(") reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) (("[") reduce 78)) 1639 ((("(") reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) (("[") reduce 79)) 1640 ((("(") reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) (("[") reduce 4)) 1641 ((("(") reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) (("[") reduce 5)) 1642 ((("(") reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) (("[") reduce 43)) 1643 ((("(") reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) (("[") reduce 73)) 1644 ((("(") reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) (("[") reduce 82)) 1645 ((("(") reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) (("[") reduce 47)) 1646 ((("(") reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) (("[") reduce 31)) 1647 ((("(") reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) (("[") reduce 57)) 1648 ((("(") reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) (("[") reduce 61)) 1649 ((("(") reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) (("[") reduce 28)) 1650 ((("(") reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) (("[") reduce 36)) 1651 ((("(") reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) (("[") reduce 32)) 1652 ((("(") reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) (("[") reduce 35)) 1653 ((("(") reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) (("[") reduce 19)) 1654 ((("(") reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) (("[") reduce 45)) 1655 ((("(") reduce 579) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) (("[") reduce 542)) 1657 ((("(") reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) (("[") reduce 72)) 1658 ((("(") reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) (("[") reduce 83)) 1659 ((("%") reduce 572) (("(") reduce 581) ((")") reduce 572) (("*") reduce 572) (("+") reduce 572) ((",") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) (("<") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_DOUBLE_ARROW) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_POW) reduce 572) ((T_SL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) (("[") reduce 572) (("^") reduce 572) (("|") reduce 572)) 1662 ((("(") reduce 580) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) (("[") reduce 544)) 1664 ((("%") reduce 507) (("(") reduce 507) ((")") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("|") reduce 507)) 1665 ((("%") reduce 291) (("(") reduce 291) ((")") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 1669 ((("%") reduce 292) (("(") reduce 292) ((")") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("[") reduce 292) (("^") reduce 292) (("|") reduce 292)) 1670 ((("%") reduce 293) (("(") reduce 293) ((")") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("[") reduce 293) (("^") reduce 293) (("|") reduce 293)) 1671 ((("%") reduce 504) (("(") reduce 504) ((")") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 1673 ((("(") reduce 579) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 541) ((T_OBJECT_OPERATOR) reduce 541) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 541) (("[") reduce 541)) 1675 ((("%") reduce 571) (("(") reduce 581) ((")") reduce 571) (("*") reduce 571) (("+") reduce 571) ((",") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) (("<") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_DOUBLE_ARROW) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_POW) reduce 571) ((T_SL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) (("[") reduce 571) (("^") reduce 571) (("|") reduce 571)) 1678 ((("(") reduce 580) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) (("[") reduce 543)) 1679 ((("%") reduce 506) (("(") reduce 506) ((")") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("|") reduce 506)) 1680 ((("%") reduce 508) (("(") reduce 508) ((")") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("|") reduce 508)) 1685 ((("%") reduce 560) (("(") reduce 560) ((")") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 1686 ((("%") reduce 582) (("(") reduce 582) ((")") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 1687 ((("%") reduce 566) (("(") shift) ((")") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 1688 ((("%") reduce 584) (("(") reduce 584) ((")") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 1691 ((("%") reduce 583) (("(") reduce 583) ((")") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 1692 ((("%") reduce 561) (("(") reduce 561) ((")") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 1693 ((("%") reduce 567) (("(") shift) ((")") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 1694 ((("%") reduce 562) (("(") reduce 562) ((")") reduce 562) (("*") reduce 562) (("+") reduce 562) ((",") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("[") reduce 562) (("^") reduce 562) (("|") reduce 562)) 1705 ((("%") reduce 540) (("*") reduce 540) (("+") reduce 540) ((",") reduce 540) (("-") reduce 540) ((".") reduce 540) (("/") reduce 540) (("<") reduce 540) ((">") reduce 540) (("?") reduce 540) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_BOOLEAN_AND) reduce 540) ((T_BOOLEAN_OR) reduce 540) ((T_COALESCE) reduce 540) ((T_DOUBLE_ARROW) reduce 540) ((T_INSTANCEOF) reduce 540) ((T_IS_EQUAL) reduce 540) ((T_IS_GREATER_OR_EQUAL) reduce 540) ((T_IS_IDENTICAL) reduce 540) ((T_IS_NOT_EQUAL) reduce 540) ((T_IS_NOT_IDENTICAL) reduce 540) ((T_IS_SMALLER_OR_EQUAL) reduce 540) ((T_LOGICAL_AND) reduce 540) ((T_LOGICAL_OR) reduce 540) ((T_LOGICAL_XOR) reduce 540) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 540) ((T_OBJECT_OPERATOR) reduce 540) ((T_PIPE) reduce 540) ((T_POW) reduce 540) ((T_SL) reduce 540) ((T_SPACESHIP) reduce 540) ((T_SR) reduce 540) (("[") reduce 540) (("]") reduce 540) (("^") reduce 540) (("|") reduce 540)) 1707 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_PIPE) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("]") reduce 521) (("^") reduce 521) (("|") reduce 521)) 1709 ((("%") reduce 534) (("*") reduce 534) (("+") reduce 534) ((",") reduce 534) (("-") reduce 534) ((".") reduce 534) (("/") reduce 534) (("<") reduce 534) ((">") reduce 534) (("?") reduce 534) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_BOOLEAN_AND) reduce 534) ((T_BOOLEAN_OR) reduce 534) ((T_COALESCE) reduce 534) ((T_DOUBLE_ARROW) reduce 534) ((T_INSTANCEOF) reduce 534) ((T_IS_EQUAL) reduce 534) ((T_IS_GREATER_OR_EQUAL) reduce 534) ((T_IS_IDENTICAL) reduce 534) ((T_IS_NOT_EQUAL) reduce 534) ((T_IS_NOT_IDENTICAL) reduce 534) ((T_IS_SMALLER_OR_EQUAL) reduce 534) ((T_LOGICAL_AND) reduce 534) ((T_LOGICAL_OR) reduce 534) ((T_LOGICAL_XOR) reduce 534) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PIPE) reduce 534) ((T_POW) reduce 534) ((T_SL) reduce 534) ((T_SPACESHIP) reduce 534) ((T_SR) reduce 534) (("[") reduce 534) (("]") reduce 534) (("^") reduce 534) (("|") reduce 534)) 1710 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_PIPE) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("]") reduce 524) (("^") reduce 524) (("|") reduce 524)) 1715 ((("%") reduce 517) (("(") shift) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_PIPE) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("]") reduce 517) (("^") reduce 517) (("|") reduce 517)) 1716 ((("%") reduce 533) (("*") reduce 533) (("+") reduce 533) ((",") reduce 533) (("-") reduce 533) ((".") reduce 533) (("/") reduce 533) (("<") reduce 533) ((">") reduce 533) (("?") reduce 533) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_BOOLEAN_AND) reduce 533) ((T_BOOLEAN_OR) reduce 533) ((T_COALESCE) reduce 533) ((T_DOUBLE_ARROW) reduce 533) ((T_INSTANCEOF) reduce 533) ((T_IS_EQUAL) reduce 533) ((T_IS_GREATER_OR_EQUAL) reduce 533) ((T_IS_IDENTICAL) reduce 533) ((T_IS_NOT_EQUAL) reduce 533) ((T_IS_NOT_IDENTICAL) reduce 533) ((T_IS_SMALLER_OR_EQUAL) reduce 533) ((T_LOGICAL_AND) reduce 533) ((T_LOGICAL_OR) reduce 533) ((T_LOGICAL_XOR) reduce 533) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PIPE) reduce 533) ((T_POW) reduce 533) ((T_SL) reduce 533) ((T_SPACESHIP) reduce 533) ((T_SR) reduce 533) (("[") reduce 533) (("]") reduce 533) (("^") reduce 533) (("|") reduce 533)) 1717 ((("%") reduce 537) (("*") reduce 537) (("+") reduce 537) ((",") reduce 537) (("-") reduce 537) ((".") reduce 537) (("/") reduce 537) (("<") reduce 537) ((">") reduce 537) (("?") reduce 537) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_BOOLEAN_AND) reduce 537) ((T_BOOLEAN_OR) reduce 537) ((T_COALESCE) reduce 537) ((T_DOUBLE_ARROW) reduce 537) ((T_INSTANCEOF) reduce 537) ((T_IS_EQUAL) reduce 537) ((T_IS_GREATER_OR_EQUAL) reduce 537) ((T_IS_IDENTICAL) reduce 537) ((T_IS_NOT_EQUAL) reduce 537) ((T_IS_NOT_IDENTICAL) reduce 537) ((T_IS_SMALLER_OR_EQUAL) reduce 537) ((T_LOGICAL_AND) reduce 537) ((T_LOGICAL_OR) reduce 537) ((T_LOGICAL_XOR) reduce 537) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_PIPE) reduce 537) ((T_POW) reduce 537) ((T_SL) reduce 537) ((T_SPACESHIP) reduce 537) ((T_SR) reduce 537) (("[") reduce 537) (("]") reduce 537) (("^") reduce 537) (("|") reduce 537)) 1723 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PIPE) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("]") reduce 532) (("^") reduce 532) (("|") reduce 532)) 1725 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_PIPE) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("]") reduce 523) (("^") reduce 523) (("|") reduce 523)) 1727 ((("%") reduce 536) (("*") reduce 536) (("+") reduce 536) ((",") reduce 536) (("-") reduce 536) ((".") reduce 536) (("/") reduce 536) (("<") reduce 536) ((">") reduce 536) (("?") reduce 536) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_BOOLEAN_AND) reduce 536) ((T_BOOLEAN_OR) reduce 536) ((T_COALESCE) reduce 536) ((T_DOUBLE_ARROW) reduce 536) ((T_INSTANCEOF) reduce 536) ((T_IS_EQUAL) reduce 536) ((T_IS_GREATER_OR_EQUAL) reduce 536) ((T_IS_IDENTICAL) reduce 536) ((T_IS_NOT_EQUAL) reduce 536) ((T_IS_NOT_IDENTICAL) reduce 536) ((T_IS_SMALLER_OR_EQUAL) reduce 536) ((T_LOGICAL_AND) reduce 536) ((T_LOGICAL_OR) reduce 536) ((T_LOGICAL_XOR) reduce 536) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 536) ((T_OBJECT_OPERATOR) reduce 536) ((T_PIPE) reduce 536) ((T_POW) reduce 536) ((T_SL) reduce 536) ((T_SPACESHIP) reduce 536) ((T_SR) reduce 536) (("[") reduce 536) (("]") reduce 536) (("^") reduce 536) (("|") reduce 536)) 1728 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("]") reduce 94) (("^") reduce 94) (("|") reduce 94)) 1729 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("]") reduce 93) (("^") reduce 93) (("|") reduce 93)) 1730 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("]") reduce 95) (("^") reduce 95) (("|") reduce 95)) 1732 ((("%") reduce 539) (("*") reduce 539) (("+") reduce 539) ((",") reduce 539) (("-") reduce 539) ((".") reduce 539) (("/") reduce 539) (("<") reduce 539) ((">") reduce 539) (("?") reduce 539) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_BOOLEAN_AND) reduce 539) ((T_BOOLEAN_OR) reduce 539) ((T_COALESCE) reduce 539) ((T_DOUBLE_ARROW) reduce 539) ((T_INSTANCEOF) reduce 539) ((T_IS_EQUAL) reduce 539) ((T_IS_GREATER_OR_EQUAL) reduce 539) ((T_IS_IDENTICAL) reduce 539) ((T_IS_NOT_EQUAL) reduce 539) ((T_IS_NOT_IDENTICAL) reduce 539) ((T_IS_SMALLER_OR_EQUAL) reduce 539) ((T_LOGICAL_AND) reduce 539) ((T_LOGICAL_OR) reduce 539) ((T_LOGICAL_XOR) reduce 539) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 539) ((T_OBJECT_OPERATOR) reduce 539) ((T_PIPE) reduce 539) ((T_POW) reduce 539) ((T_SL) reduce 539) ((T_SPACESHIP) reduce 539) ((T_SR) reduce 539) (("[") reduce 539) (("]") reduce 539) (("^") reduce 539) (("|") reduce 539)) 1735 ((("%") reduce 538) (("*") reduce 538) (("+") reduce 538) ((",") reduce 538) (("-") reduce 538) ((".") reduce 538) (("/") reduce 538) (("<") reduce 538) ((">") reduce 538) (("?") reduce 538) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_BOOLEAN_AND) reduce 538) ((T_BOOLEAN_OR) reduce 538) ((T_COALESCE) reduce 538) ((T_DOUBLE_ARROW) reduce 538) ((T_INSTANCEOF) reduce 538) ((T_IS_EQUAL) reduce 538) ((T_IS_GREATER_OR_EQUAL) reduce 538) ((T_IS_IDENTICAL) reduce 538) ((T_IS_NOT_EQUAL) reduce 538) ((T_IS_NOT_IDENTICAL) reduce 538) ((T_IS_SMALLER_OR_EQUAL) reduce 538) ((T_LOGICAL_AND) reduce 538) ((T_LOGICAL_OR) reduce 538) ((T_LOGICAL_XOR) reduce 538) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 538) ((T_OBJECT_OPERATOR) reduce 538) ((T_PIPE) reduce 538) ((T_POW) reduce 538) ((T_SL) reduce 538) ((T_SPACESHIP) reduce 538) ((T_SR) reduce 538) (("[") reduce 538) (("]") reduce 538) (("^") reduce 538) (("|") reduce 538)) 1741 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("]") reduce 92) (("^") reduce 92) (("|") reduce 92)) 1744 ((("%") reduce 535) (("*") reduce 535) (("+") reduce 535) ((",") reduce 535) (("-") reduce 535) ((".") reduce 535) (("/") reduce 535) (("<") reduce 535) ((">") reduce 535) (("?") reduce 535) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_BOOLEAN_AND) reduce 535) ((T_BOOLEAN_OR) reduce 535) ((T_COALESCE) reduce 535) ((T_DOUBLE_ARROW) reduce 535) ((T_INSTANCEOF) reduce 535) ((T_IS_EQUAL) reduce 535) ((T_IS_GREATER_OR_EQUAL) reduce 535) ((T_IS_IDENTICAL) reduce 535) ((T_IS_NOT_EQUAL) reduce 535) ((T_IS_NOT_IDENTICAL) reduce 535) ((T_IS_SMALLER_OR_EQUAL) reduce 535) ((T_LOGICAL_AND) reduce 535) ((T_LOGICAL_OR) reduce 535) ((T_LOGICAL_XOR) reduce 535) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PIPE) reduce 535) ((T_POW) reduce 535) ((T_SL) reduce 535) ((T_SPACESHIP) reduce 535) ((T_SR) reduce 535) (("[") reduce 535) (("]") reduce 535) (("^") reduce 535) (("|") reduce 535)) 1746 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) (("=") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AND_EQUAL) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_COALESCE_EQUAL) reduce 568) ((T_CONCAT_EQUAL) reduce 568) ((T_DEC) reduce 568) ((T_DIV_EQUAL) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INC) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_MINUS_EQUAL) reduce 568) ((T_MOD_EQUAL) reduce 568) ((T_MUL_EQUAL) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_OR_EQUAL) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_PLUS_EQUAL) reduce 568) ((T_POW) reduce 568) ((T_POW_EQUAL) reduce 568) ((T_SL) reduce 568) ((T_SL_EQUAL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) ((T_SR_EQUAL) reduce 568) ((T_XOR_EQUAL) reduce 568) (("[") reduce 568) (("]") reduce 568) (("^") reduce 568) (("|") reduce 568)) 1747 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 479) (("(") shift) (("*") reduce 479) (("+") shift) ((",") reduce 479) (("-") shift) ((".") reduce 479) (("/") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 479) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 479) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 479) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PIPE) reduce 479) ((T_POW) reduce 479) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("]") reduce 479) (("^") reduce 479) (("`") shift) (("|") reduce 479) (("~") shift)) 1753 (((",") reduce 587) (("]") reduce 587)) 1757 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) ((",") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) (("<") reduce 564) (("=") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AND_EQUAL) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_COALESCE_EQUAL) reduce 564) ((T_CONCAT_EQUAL) reduce 564) ((T_DEC) reduce 564) ((T_DIV_EQUAL) reduce 564) ((T_DOUBLE_ARROW) reduce 564) ((T_INC) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_MINUS_EQUAL) reduce 564) ((T_MOD_EQUAL) reduce 564) ((T_MUL_EQUAL) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_OR_EQUAL) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_PLUS_EQUAL) reduce 564) ((T_POW) reduce 564) ((T_POW_EQUAL) reduce 564) ((T_SL) reduce 564) ((T_SL_EQUAL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) ((T_SR_EQUAL) reduce 564) ((T_XOR_EQUAL) reduce 564) (("[") reduce 564) (("]") reduce 564) (("^") reduce 564) (("|") reduce 564)) 1758 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PIPE) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 551) (("]") reduce 530) (("^") reduce 530) (("|") reduce 530)) 1760 ((("%") reduce 529) (("*") reduce 529) (("+") reduce 529) ((",") reduce 529) (("-") reduce 529) ((".") reduce 529) (("/") reduce 529) (("<") reduce 529) ((">") reduce 529) (("?") reduce 529) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_BOOLEAN_AND) reduce 529) ((T_BOOLEAN_OR) reduce 529) ((T_COALESCE) reduce 529) ((T_DOUBLE_ARROW) reduce 529) ((T_INSTANCEOF) reduce 529) ((T_IS_EQUAL) reduce 529) ((T_IS_GREATER_OR_EQUAL) reduce 529) ((T_IS_IDENTICAL) reduce 529) ((T_IS_NOT_EQUAL) reduce 529) ((T_IS_NOT_IDENTICAL) reduce 529) ((T_IS_SMALLER_OR_EQUAL) reduce 529) ((T_LOGICAL_AND) reduce 529) ((T_LOGICAL_OR) reduce 529) ((T_LOGICAL_XOR) reduce 529) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PIPE) reduce 529) ((T_POW) reduce 529) ((T_SL) reduce 529) ((T_SPACESHIP) reduce 529) ((T_SR) reduce 529) (("[") reduce 554) (("]") reduce 529) (("^") reduce 529) (("|") reduce 529)) 1761 ((("%") reduce 528) (("(") reduce 557) (("*") reduce 528) (("+") reduce 528) ((",") reduce 528) (("-") reduce 528) ((".") reduce 528) (("/") reduce 528) (("<") reduce 528) ((">") reduce 528) (("?") reduce 528) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_BOOLEAN_AND) reduce 528) ((T_BOOLEAN_OR) reduce 528) ((T_COALESCE) reduce 528) ((T_DOUBLE_ARROW) reduce 528) ((T_INSTANCEOF) reduce 528) ((T_IS_EQUAL) reduce 528) ((T_IS_GREATER_OR_EQUAL) reduce 528) ((T_IS_IDENTICAL) reduce 528) ((T_IS_NOT_EQUAL) reduce 528) ((T_IS_NOT_IDENTICAL) reduce 528) ((T_IS_SMALLER_OR_EQUAL) reduce 528) ((T_LOGICAL_AND) reduce 528) ((T_LOGICAL_OR) reduce 528) ((T_LOGICAL_XOR) reduce 528) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PIPE) reduce 528) ((T_POW) reduce 528) ((T_SL) reduce 528) ((T_SPACESHIP) reduce 528) ((T_SR) reduce 528) (("[") reduce 550) (("]") reduce 528) (("^") reduce 528) (("|") reduce 528)) 1762 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 591) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 591) (("^") shift) (("|") shift)) 1765 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) ((",") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) (("<") reduce 563) (("=") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AND_EQUAL) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_COALESCE_EQUAL) reduce 563) ((T_CONCAT_EQUAL) reduce 563) ((T_DEC) reduce 563) ((T_DIV_EQUAL) reduce 563) ((T_DOUBLE_ARROW) reduce 563) ((T_INC) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_MINUS_EQUAL) reduce 563) ((T_MOD_EQUAL) reduce 563) ((T_MUL_EQUAL) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_OR_EQUAL) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_PLUS_EQUAL) reduce 563) ((T_POW) reduce 563) ((T_POW_EQUAL) reduce 563) ((T_SL) reduce 563) ((T_SL_EQUAL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) ((T_SR_EQUAL) reduce 563) ((T_XOR_EQUAL) reduce 563) (("[") reduce 563) (("]") reduce 563) (("^") reduce 563) (("|") reduce 563)) 1766 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) ((",") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_DOUBLE_ARROW) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_PIPE) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("]") reduce 484) (("^") reduce 484) (("|") reduce 484)) 1767 ((("%") reduce 466) (("*") reduce 466) (("+") reduce 466) ((",") reduce 466) (("-") reduce 466) ((".") reduce 466) (("/") reduce 466) (("<") reduce 466) ((">") reduce 466) (("?") reduce 466) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_BOOLEAN_AND) reduce 466) ((T_BOOLEAN_OR) reduce 466) ((T_COALESCE) reduce 466) ((T_DOUBLE_ARROW) reduce 466) ((T_INSTANCEOF) reduce 466) ((T_IS_EQUAL) reduce 466) ((T_IS_GREATER_OR_EQUAL) reduce 466) ((T_IS_IDENTICAL) reduce 466) ((T_IS_NOT_EQUAL) reduce 466) ((T_IS_NOT_IDENTICAL) reduce 466) ((T_IS_SMALLER_OR_EQUAL) reduce 466) ((T_LOGICAL_AND) reduce 466) ((T_LOGICAL_OR) reduce 466) ((T_LOGICAL_XOR) reduce 466) ((T_PIPE) reduce 466) ((T_POW) reduce 466) ((T_SL) reduce 466) ((T_SPACESHIP) reduce 466) ((T_SR) reduce 466) (("]") reduce 466) (("^") reduce 466) (("|") reduce 466)) 1768 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_PIPE) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("]") reduce 488) (("^") reduce 488) (("|") reduce 488)) 1769 ((("%") reduce 531) (("(") shift) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("]") reduce 531) (("^") reduce 531) (("|") reduce 531)) 1770 ((("%") reduce 461) (("(") reduce 558) (("*") reduce 461) (("+") reduce 461) ((",") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PIPE) reduce 461) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") reduce 552) (("]") reduce 461) (("^") reduce 461) (("|") reduce 461)) 1771 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) ((",") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_DOUBLE_ARROW) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_PIPE) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("]") reduce 462) (("^") reduce 462) (("|") reduce 462)) 1772 (((",") shift) (("]") reduce 585)) 1773 (((",") reduce 589) (("]") reduce 589)) 1774 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) ((",") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_DOUBLE_ARROW) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_PIPE) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("]") reduce 476) (("^") reduce 476) (("|") reduce 476)) 1775 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) (("=") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AND_EQUAL) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_COALESCE_EQUAL) reduce 559) ((T_CONCAT_EQUAL) reduce 559) ((T_DEC) reduce 559) ((T_DIV_EQUAL) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INC) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_MINUS_EQUAL) reduce 559) ((T_MOD_EQUAL) reduce 559) ((T_MUL_EQUAL) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_OR_EQUAL) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_PLUS_EQUAL) reduce 559) ((T_POW) reduce 559) ((T_POW_EQUAL) reduce 559) ((T_SL) reduce 559) ((T_SL_EQUAL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) ((T_SR_EQUAL) reduce 559) ((T_XOR_EQUAL) reduce 559) (("[") reduce 559) (("]") reduce 559) (("^") reduce 559) (("|") reduce 559)) 1776 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("]") reduce 565) (("^") reduce 565) (("|") reduce 565)) 1777 ((("%") reduce 403) (("*") reduce 403) (("+") reduce 403) ((",") reduce 403) (("-") reduce 403) ((".") reduce 403) (("/") reduce 403) (("<") reduce 403) (("=") shift) ((">") reduce 403) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_DOUBLE_ARROW) reduce 403) ((T_INC) shift) ((T_INSTANCEOF) reduce 403) ((T_IS_EQUAL) reduce 403) ((T_IS_GREATER_OR_EQUAL) reduce 403) ((T_IS_IDENTICAL) reduce 403) ((T_IS_NOT_EQUAL) reduce 403) ((T_IS_NOT_IDENTICAL) reduce 403) ((T_IS_SMALLER_OR_EQUAL) reduce 403) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 403) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 403) ((T_POW_EQUAL) shift) ((T_SL) reduce 403) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 403) ((T_SR) reduce 403) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 548) (("]") reduce 403) (("^") reduce 403) (("|") reduce 403)) 1781 ((("%") reduce 448) (("*") reduce 448) (("+") reduce 448) ((",") reduce 448) (("-") reduce 448) ((".") reduce 448) (("/") reduce 448) (("<") reduce 448) ((">") reduce 448) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_BOOLEAN_AND) reduce 448) ((T_BOOLEAN_OR) reduce 448) ((T_COALESCE) reduce 448) ((T_DOUBLE_ARROW) reduce 448) ((T_INSTANCEOF) reduce 448) ((T_IS_EQUAL) reduce 448) ((T_IS_GREATER_OR_EQUAL) reduce 448) ((T_IS_IDENTICAL) reduce 448) ((T_IS_NOT_EQUAL) reduce 448) ((T_IS_NOT_IDENTICAL) reduce 448) ((T_IS_SMALLER_OR_EQUAL) reduce 448) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_PIPE) reduce 448) ((T_POW) shift) ((T_SL) reduce 448) ((T_SPACESHIP) reduce 448) ((T_SR) reduce 448) (("]") reduce 448) (("^") reduce 448) (("|") reduce 448)) 1813 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 432) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_DOUBLE_ARROW) reduce 432) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 432) (("^") shift) (("|") reduce 432)) 1814 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 435) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_DOUBLE_ARROW) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 435) (("^") reduce 435) (("|") reduce 435)) 1815 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 444) (("-") shift) ((".") reduce 444) (("/") shift) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_PIPE) reduce 444) ((T_POW) shift) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("]") reduce 444) (("^") reduce 444) (("|") reduce 444)) 1816 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 458) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 458) (("^") reduce 458) (("|") reduce 458)) 1817 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 443) (("-") shift) ((".") reduce 443) (("/") shift) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_DOUBLE_ARROW) reduce 443) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_PIPE) reduce 443) ((T_POW) shift) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("]") reduce 443) (("^") reduce 443) (("|") reduce 443)) 1818 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) ((",") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_DOUBLE_ARROW) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_PIPE) reduce 440) ((T_POW) shift) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("]") reduce 440) (("^") reduce 440) (("|") reduce 440)) 1819 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 453) (("-") shift) ((".") shift) (("/") shift) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_DOUBLE_ARROW) reduce 453) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_PIPE) reduce 453) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 453) ((T_SR) shift) (("]") reduce 453) (("^") reduce 453) (("|") reduce 453)) 1820 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 431) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 431) (("^") shift) (("|") shift)) 1821 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 429) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 429) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 429) (("^") shift) (("|") shift)) 1822 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 430) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 430) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 430) (("^") shift) (("|") shift)) 1823 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 455) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 455) ((T_SR) shift) (("]") reduce 455) (("^") reduce 455) (("|") reduce 455)) 1824 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 450) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_DOUBLE_ARROW) reduce 450) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 450) (("^") reduce 450) (("|") reduce 450)) 1825 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 452) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_DOUBLE_ARROW) reduce 452) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 452) (("^") reduce 452) (("|") reduce 452)) 1826 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 449) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_DOUBLE_ARROW) reduce 449) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 449) (("^") reduce 449) (("|") reduce 449)) 1827 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 457) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 457) ((T_SR) shift) (("]") reduce 457) (("^") reduce 457) (("|") reduce 457)) 1828 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 451) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_DOUBLE_ARROW) reduce 451) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 451) (("^") reduce 451) (("|") reduce 451)) 1831 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("]") reduce 94) (("^") reduce 94) (("|") reduce 94)) 1832 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("]") reduce 93) (("^") reduce 93) (("|") reduce 93)) 1833 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("]") reduce 95) (("^") reduce 95) (("|") reduce 95)) 1834 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("]") reduce 509) (("^") reduce 509) (("|") reduce 509)) 1835 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("]") reduce 92) (("^") reduce 92) (("|") reduce 92)) 1836 ((("%") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("]") reduce 568) (("^") reduce 568) (("|") reduce 568)) 1837 ((("%") reduce 511) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("]") reduce 511) (("^") reduce 511) (("|") reduce 511)) 1838 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PIPE) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("]") reduce 459) (("^") reduce 459) (("|") reduce 459)) 1839 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("]") reduce 510) (("^") reduce 510) (("|") reduce 510)) 1840 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("]") reduce 512) (("^") reduce 512) (("|") reduce 512)) 1841 ((("%") reduce 573) (("*") reduce 573) (("+") reduce 573) ((",") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_DOUBLE_ARROW) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("]") reduce 573) (("^") reduce 573) (("|") reduce 573)) 1847 ((("%") reduce 574) (("*") reduce 574) (("+") reduce 574) ((",") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_DOUBLE_ARROW) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("]") reduce 574) (("^") reduce 574) (("|") reduce 574)) 1848 ((("%") reduce 578) (("*") reduce 578) (("+") reduce 578) ((",") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_DOUBLE_ARROW) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("]") reduce 578) (("^") reduce 578) (("|") reduce 578)) 1849 ((("%") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("]") reduce 582) (("^") reduce 582) (("|") reduce 582)) 1850 ((("%") reduce 575) (("*") reduce 575) (("+") reduce 575) ((",") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_DOUBLE_ARROW) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("]") reduce 575) (("^") reduce 575) (("|") reduce 575)) 1851 ((("%") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("]") reduce 584) (("^") reduce 584) (("|") reduce 584)) 1854 ((("%") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("]") reduce 583) (("^") reduce 583) (("|") reduce 583)) 1855 ((("%") reduce 576) (("*") reduce 576) (("+") reduce 576) ((",") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_DOUBLE_ARROW) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("]") reduce 576) (("^") reduce 576) (("|") reduce 576)) 1857 ((("%") reduce 577) (("*") reduce 577) (("+") reduce 577) ((",") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_DOUBLE_ARROW) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("]") reduce 577) (("^") reduce 577) (("|") reduce 577)) 1859 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("]") reduce 513) (("^") reduce 513) (("|") reduce 513)) 1860 ((("%") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("]") reduce 570) (("^") reduce 570) (("|") reduce 570)) 1863 ((("%") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("]") reduce 569) (("^") reduce 569) (("|") reduce 569)) 1864 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 465) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 465) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 465) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 465) (("^") shift) (("|") shift)) 1865 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 427) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_DOUBLE_ARROW) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 427) (("^") shift) (("|") shift)) 1866 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 428) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_DOUBLE_ARROW) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 428) (("^") shift) (("|") shift)) 1867 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 433) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_DOUBLE_ARROW) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 433) (("^") reduce 433) (("|") reduce 433)) 1868 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 434) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_DOUBLE_ARROW) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 434) (("^") reduce 434) (("|") reduce 434)) 1872 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 463) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 463) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 463) (("^") shift) (("|") shift)) 1873 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 464) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 464) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 464) (("^") shift) (("|") shift)) 1874 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 456) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 456) ((T_SR) shift) (("]") reduce 456) (("^") reduce 456) (("|") reduce 456)) 1875 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 454) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 454) ((T_SR) shift) (("]") reduce 454) (("^") reduce 454) (("|") reduce 454)) 1876 ((("%") reduce 441) (("*") reduce 441) (("+") reduce 441) ((",") reduce 441) (("-") reduce 441) ((".") reduce 441) (("/") reduce 441) (("<") reduce 441) ((">") reduce 441) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_DOUBLE_ARROW) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 441) ((T_IS_GREATER_OR_EQUAL) reduce 441) ((T_IS_IDENTICAL) reduce 441) ((T_IS_NOT_EQUAL) reduce 441) ((T_IS_NOT_IDENTICAL) reduce 441) ((T_IS_SMALLER_OR_EQUAL) reduce 441) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_PIPE) reduce 441) ((T_POW) shift) ((T_SL) reduce 441) ((T_SPACESHIP) reduce 441) ((T_SR) reduce 441) (("]") reduce 441) (("^") reduce 441) (("|") reduce 441)) 1877 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 436) (("-") shift) ((".") reduce 436) (("/") shift) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_DOUBLE_ARROW) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_PIPE) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 436) ((T_SR) shift) (("]") reduce 436) (("^") reduce 436) (("|") reduce 436)) 1878 ((("%") shift) (("*") shift) (("+") reduce 438) ((",") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") shift) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_DOUBLE_ARROW) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_PIPE) reduce 438) ((T_POW) shift) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("]") reduce 438) (("^") reduce 438) (("|") reduce 438)) 1879 ((("%") shift) (("*") shift) (("+") reduce 437) ((",") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") shift) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_DOUBLE_ARROW) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_PIPE) reduce 437) ((T_POW) shift) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("]") reduce 437) (("^") reduce 437) (("|") reduce 437)) 1880 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) ((",") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_DOUBLE_ARROW) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_PIPE) reduce 439) ((T_POW) shift) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("]") reduce 439) (("^") reduce 439) (("|") reduce 439)) 1881 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) ((",") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_DOUBLE_ARROW) reduce 442) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_PIPE) reduce 442) ((T_POW) shift) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("]") reduce 442) (("^") reduce 442) (("|") reduce 442)) 1882 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) ((")") reduce 586) (("+") shift) ((",") reduce 586) (("-") shift) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_ELLIPSIS) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 1941 (((")") reduce 587) ((",") reduce 587)) 1950 ((("%") shift) ((")") reduce 591) (("*") shift) (("+") shift) ((",") reduce 591) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 1960 (((")") reduce 585) ((",") shift)) 1961 (((")") reduce 589) ((",") reduce 589)) 2072 ((("=") shift)) 2074 ((("%") shift) ((")") reduce 404) (("*") shift) (("+") shift) ((",") reduce 404) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2246 ((("%") reduce 505) (("(") reduce 505) ((")") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 2247 ((("%") reduce 570) (("(") reduce 570) ((")") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 2250 ((("%") reduce 569) (("(") reduce 569) ((")") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 2252 (((")") reduce 588) ((",") reduce 588)) 2253 ((("%") reduce 504) (("(") reduce 504) ((")") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) (("=") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AND_EQUAL) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_COALESCE_EQUAL) reduce 504) ((T_CONCAT_EQUAL) reduce 504) ((T_DEC) reduce 504) ((T_DIV_EQUAL) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INC) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_MINUS_EQUAL) reduce 504) ((T_MOD_EQUAL) reduce 504) ((T_MUL_EQUAL) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_OR_EQUAL) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_PLUS_EQUAL) reduce 504) ((T_POW) reduce 504) ((T_POW_EQUAL) reduce 504) ((T_SL) reduce 504) ((T_SL_EQUAL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) ((T_SR_EQUAL) reduce 504) ((T_XOR_EQUAL) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 2264 ((("%") reduce 494) ((")") reduce 494) (("*") reduce 494) (("+") reduce 494) ((",") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PIPE) reduce 494) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 2265 ((("%") reduce 489) ((")") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_PIPE) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("^") reduce 489) (("|") reduce 489)) 2270 (((":") shift) ((T_DOUBLE_ARROW) reduce 289)) 2272 (((T_DOUBLE_ARROW) shift)) 2273 ((("!") reduce 494) (("\"") reduce 494) (("$") reduce 494) (("(") reduce 494) (("+") reduce 494) (("-") reduce 494) (("@") reduce 494) ((T_ARRAY) reduce 494) ((T_ARRAY_CAST) reduce 494) ((T_ATTRIBUTE) reduce 494) ((T_BOOL_CAST) reduce 494) ((T_CLASS_C) reduce 494) ((T_CLONE) reduce 494) ((T_CONSTANT_ENCAPSED_STRING) reduce 494) ((T_DEC) reduce 494) ((T_DIR) reduce 494) ((T_DNUMBER) reduce 494) ((T_DOUBLE_CAST) reduce 494) ((T_EMPTY) reduce 494) ((T_EVAL) reduce 494) ((T_EXIT) reduce 494) ((T_FILE) reduce 494) ((T_FN) reduce 494) ((T_FUNCTION) reduce 494) ((T_FUNC_C) reduce 494) ((T_INC) reduce 494) ((T_INCLUDE) reduce 494) ((T_INCLUDE_ONCE) reduce 494) ((T_INT_CAST) reduce 494) ((T_ISSET) reduce 494) ((T_LINE) reduce 494) ((T_LIST) reduce 494) ((T_LNUMBER) reduce 494) ((T_MATCH) reduce 494) ((T_METHOD_C) reduce 494) ((T_NAME_FULLY_QUALIFIED) reduce 494) ((T_NAME_QUALIFIED) reduce 494) ((T_NAME_RELATIVE) reduce 494) ((T_NEW) reduce 494) ((T_NS_C) reduce 494) ((T_OBJECT_CAST) reduce 494) ((T_PRINT) reduce 494) ((T_PROPERTY_C) reduce 494) ((T_READONLY) reduce 494) ((T_REQUIRE) reduce 494) ((T_REQUIRE_ONCE) reduce 494) ((T_START_HEREDOC) reduce 494) ((T_STATIC) reduce 494) ((T_STRING) reduce 494) ((T_STRING_CAST) reduce 494) ((T_THROW) reduce 494) ((T_TRAIT_C) reduce 494) ((T_UNSET_CAST) reduce 494) ((T_VARIABLE) reduce 494) ((T_YIELD) reduce 494) ((T_YIELD_FROM) reduce 494) (("[") reduce 494) (("`") reduce 494) (("~") reduce 494)) 2274 ((("!") reduce 495) (("\"") reduce 495) (("$") reduce 495) (("(") reduce 495) (("+") reduce 495) (("-") reduce 495) (("@") reduce 495) ((T_ARRAY) reduce 495) ((T_ARRAY_CAST) reduce 495) ((T_ATTRIBUTE) reduce 495) ((T_BOOL_CAST) reduce 495) ((T_CLASS_C) reduce 495) ((T_CLONE) reduce 495) ((T_CONSTANT_ENCAPSED_STRING) reduce 495) ((T_DEC) reduce 495) ((T_DIR) reduce 495) ((T_DNUMBER) reduce 495) ((T_DOUBLE_CAST) reduce 495) ((T_EMPTY) reduce 495) ((T_EVAL) reduce 495) ((T_EXIT) reduce 495) ((T_FILE) reduce 495) ((T_FN) reduce 495) ((T_FUNCTION) reduce 495) ((T_FUNC_C) reduce 495) ((T_INC) reduce 495) ((T_INCLUDE) reduce 495) ((T_INCLUDE_ONCE) reduce 495) ((T_INT_CAST) reduce 495) ((T_ISSET) reduce 495) ((T_LINE) reduce 495) ((T_LIST) reduce 495) ((T_LNUMBER) reduce 495) ((T_MATCH) reduce 495) ((T_METHOD_C) reduce 495) ((T_NAME_FULLY_QUALIFIED) reduce 495) ((T_NAME_QUALIFIED) reduce 495) ((T_NAME_RELATIVE) reduce 495) ((T_NEW) reduce 495) ((T_NS_C) reduce 495) ((T_OBJECT_CAST) reduce 495) ((T_PRINT) reduce 495) ((T_PROPERTY_C) reduce 495) ((T_READONLY) reduce 495) ((T_REQUIRE) reduce 495) ((T_REQUIRE_ONCE) reduce 495) ((T_START_HEREDOC) reduce 495) ((T_STATIC) reduce 495) ((T_STRING) reduce 495) ((T_STRING_CAST) reduce 495) ((T_THROW) reduce 495) ((T_TRAIT_C) reduce 495) ((T_UNSET_CAST) reduce 495) ((T_VARIABLE) reduce 495) ((T_YIELD) reduce 495) ((T_YIELD_FROM) reduce 495) (("[") reduce 495) (("`") reduce 495) (("~") reduce 495)) 2276 ((("%") shift) ((")") reduce 494) (("*") shift) (("+") shift) ((",") reduce 494) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2277 ((("%") reduce 490) ((")") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_PIPE) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("^") reduce 490) (("|") reduce 490)) 2278 (((T_ARRAY) shift) ((T_CALLABLE) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_STATIC) shift) ((T_STRING) shift)) 2280 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 280) ((T_DOUBLE_ARROW) reduce 280) (("|") reduce 280)) 2281 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_DOUBLE_ARROW) reduce 281) (("|") reduce 281)) 2282 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_DOUBLE_ARROW) reduce 94) (("|") reduce 94)) 2283 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_DOUBLE_ARROW) reduce 93) (("|") reduce 93)) 2284 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_DOUBLE_ARROW) reduce 95) (("|") reduce 95)) 2285 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 269) ((T_DOUBLE_ARROW) reduce 269) (("|") reduce 269)) 2286 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_DOUBLE_ARROW) reduce 92) (("|") reduce 92)) 2287 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_DOUBLE_ARROW) reduce 267)) 2288 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_DOUBLE_ARROW) reduce 282) (("|") reduce 282)) 2289 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_DOUBLE_ARROW) reduce 264) (("|") reduce 270)) 2290 (((T_DOUBLE_ARROW) reduce 290)) 2291 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 268) ((T_DOUBLE_ARROW) reduce 268) (("|") reduce 268)) 2292 (((T_DOUBLE_ARROW) reduce 266) (("|") shift)) 2293 ((("|") shift)) 2294 ((("(") shift) ((T_ARRAY) shift) ((T_CALLABLE) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_STATIC) shift) ((T_STRING) shift)) 2296 (((T_DOUBLE_ARROW) reduce 280) (("|") reduce 280)) 2297 (((T_DOUBLE_ARROW) reduce 281) (("|") reduce 281)) 2298 (((T_DOUBLE_ARROW) reduce 94) (("|") reduce 94)) 2299 (((T_DOUBLE_ARROW) reduce 93) (("|") reduce 93)) 2300 (((T_DOUBLE_ARROW) reduce 95) (("|") reduce 95)) 2301 (((T_DOUBLE_ARROW) reduce 269) (("|") reduce 269)) 2302 (((T_DOUBLE_ARROW) reduce 92) (("|") reduce 92)) 2303 (((T_DOUBLE_ARROW) reduce 282) (("|") reduce 282)) 2304 (((T_DOUBLE_ARROW) reduce 270) (("|") reduce 270)) 2305 (((T_DOUBLE_ARROW) reduce 268) (("|") reduce 268)) 2306 (((T_DOUBLE_ARROW) reduce 272) (("|") reduce 272)) 2307 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 280)) 2308 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281)) 2309 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94)) 2310 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93)) 2311 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95)) 2312 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 269)) 2313 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92)) 2314 (((")") shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift)) 2315 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282)) 2316 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift)) 2317 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 268)) 2319 (((")") reduce 280) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 280)) 2320 (((")") reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281)) 2321 (((")") reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94)) 2322 (((")") reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93)) 2323 (((")") reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95)) 2324 (((")") reduce 269) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 269)) 2325 (((")") reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92)) 2326 (((")") reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282)) 2327 (((")") reduce 274) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 274)) 2328 (((")") reduce 268) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 268)) 2329 (((T_DOUBLE_ARROW) reduce 271) (("|") reduce 271)) 2331 (((")") reduce 275) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 275)) 2333 (((T_DOUBLE_ARROW) reduce 273) (("|") reduce 273)) 2335 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 280) ((T_DOUBLE_ARROW) reduce 280)) 2336 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_DOUBLE_ARROW) reduce 281)) 2337 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_DOUBLE_ARROW) reduce 94)) 2338 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_DOUBLE_ARROW) reduce 93)) 2339 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_DOUBLE_ARROW) reduce 95)) 2340 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 269) ((T_DOUBLE_ARROW) reduce 269)) 2341 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_DOUBLE_ARROW) reduce 92)) 2342 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_DOUBLE_ARROW) reduce 282)) 2343 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 274) ((T_DOUBLE_ARROW) reduce 274)) 2344 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 268) ((T_DOUBLE_ARROW) reduce 268)) 2346 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 275) ((T_DOUBLE_ARROW) reduce 275)) 2347 (((T_DOUBLE_ARROW) reduce 280)) 2348 (((T_DOUBLE_ARROW) reduce 281)) 2349 (((T_DOUBLE_ARROW) reduce 94)) 2350 (((T_DOUBLE_ARROW) reduce 93)) 2351 (((T_DOUBLE_ARROW) reduce 95)) 2352 (((T_DOUBLE_ARROW) reduce 269)) 2353 (((T_DOUBLE_ARROW) reduce 92)) 2354 (((T_DOUBLE_ARROW) reduce 282)) 2355 (((T_DOUBLE_ARROW) reduce 265)) 2356 (((T_DOUBLE_ARROW) reduce 268)) 2358 ((("|") reduce 271)) 2424 ((("%") shift) ((")") reduce 590) (("*") shift) (("+") shift) ((",") reduce 590) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2546 ((("%") shift) ((")") reduce 404) (("*") shift) (("+") shift) ((",") reduce 404) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2548 ((("%") reduce 74) (("(") reduce 74) ((")") reduce 74) (("*") reduce 74) (("+") reduce 74) ((",") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_PIPE) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("|") reduce 74)) 2549 ((("%") reduce 52) (("(") reduce 52) ((")") reduce 52) (("*") reduce 52) (("+") reduce 52) ((",") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_PIPE) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("|") reduce 52)) 2550 ((("%") reduce 27) (("(") reduce 27) ((")") reduce 27) (("*") reduce 27) (("+") reduce 27) ((",") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_PIPE) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("|") reduce 27)) 2551 ((("%") reduce 51) (("(") reduce 51) ((")") reduce 51) (("*") reduce 51) (("+") reduce 51) ((",") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_PIPE) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("|") reduce 51)) 2552 ((("%") reduce 53) (("(") reduce 53) ((")") reduce 53) (("*") reduce 53) (("+") reduce 53) ((",") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_PIPE) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("|") reduce 53)) 2553 ((("%") reduce 49) (("(") reduce 49) ((")") reduce 49) (("*") reduce 49) (("+") reduce 49) ((",") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_PIPE) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("|") reduce 49)) 2554 ((("%") reduce 29) (("(") reduce 29) ((")") reduce 29) (("*") reduce 29) (("+") reduce 29) ((",") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_PIPE) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("|") reduce 29)) 2555 ((("%") reduce 59) (("(") reduce 59) ((")") reduce 59) (("*") reduce 59) (("+") reduce 59) ((",") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_PIPE) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("|") reduce 59)) 2556 ((("%") reduce 60) (("(") reduce 60) ((")") reduce 60) (("*") reduce 60) (("+") reduce 60) ((",") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_PIPE) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("|") reduce 60)) 2557 ((("%") reduce 11) (("(") reduce 11) ((")") reduce 11) (("*") reduce 11) (("+") reduce 11) ((",") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_PIPE) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("|") reduce 11)) 2558 ((("%") reduce 42) (("(") reduce 42) ((")") reduce 42) (("*") reduce 42) (("+") reduce 42) ((",") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_PIPE) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("|") reduce 42)) 2559 ((("%") reduce 39) (("(") reduce 39) ((")") reduce 39) (("*") reduce 39) (("+") reduce 39) ((",") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_PIPE) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("|") reduce 39)) 2560 ((("%") reduce 25) (("(") reduce 25) ((")") reduce 25) (("*") reduce 25) (("+") reduce 25) ((",") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_PIPE) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("|") reduce 25)) 2561 ((("%") reduce 50) (("(") reduce 50) ((")") reduce 50) (("*") reduce 50) (("+") reduce 50) ((",") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_PIPE) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("|") reduce 50)) 2562 ((("%") reduce 66) (("(") reduce 66) ((")") reduce 66) (("*") reduce 66) (("+") reduce 66) ((",") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_PIPE) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("|") reduce 66)) 2563 ((("%") reduce 18) (("(") reduce 18) ((")") reduce 18) (("*") reduce 18) (("+") reduce 18) ((",") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_PIPE) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("|") reduce 18)) 2564 ((("%") reduce 17) (("(") reduce 17) ((")") reduce 17) (("*") reduce 17) (("+") reduce 17) ((",") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_PIPE) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("|") reduce 17)) 2565 ((("%") reduce 15) (("(") reduce 15) ((")") reduce 15) (("*") reduce 15) (("+") reduce 15) ((",") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_PIPE) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("|") reduce 15)) 2566 ((("%") reduce 14) (("(") reduce 14) ((")") reduce 14) (("*") reduce 14) (("+") reduce 14) ((",") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_PIPE) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("|") reduce 14)) 2567 ((("%") reduce 38) (("(") reduce 38) ((")") reduce 38) (("*") reduce 38) (("+") reduce 38) ((",") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_PIPE) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("|") reduce 38)) 2568 ((("%") reduce 26) (("(") reduce 26) ((")") reduce 26) (("*") reduce 26) (("+") reduce 26) ((",") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_PIPE) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("|") reduce 26)) 2569 ((("%") reduce 22) (("(") reduce 22) ((")") reduce 22) (("*") reduce 22) (("+") reduce 22) ((",") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_PIPE) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("|") reduce 22)) 2570 ((("%") reduce 24) (("(") reduce 24) ((")") reduce 24) (("*") reduce 24) (("+") reduce 24) ((",") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_PIPE) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("|") reduce 24)) 2571 ((("%") reduce 16) (("(") reduce 16) ((")") reduce 16) (("*") reduce 16) (("+") reduce 16) ((",") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_PIPE) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("|") reduce 16)) 2572 ((("%") reduce 48) (("(") reduce 48) ((")") reduce 48) (("*") reduce 48) (("+") reduce 48) ((",") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_PIPE) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("|") reduce 48)) 2573 ((("%") reduce 20) (("(") reduce 20) ((")") reduce 20) (("*") reduce 20) (("+") reduce 20) ((",") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_PIPE) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("|") reduce 20)) 2574 ((("%") reduce 70) (("(") reduce 70) ((")") reduce 70) (("*") reduce 70) (("+") reduce 70) ((",") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_PIPE) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("|") reduce 70)) 2575 ((("%") reduce 3) (("(") reduce 3) ((")") reduce 3) (("*") reduce 3) (("+") reduce 3) ((",") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_PIPE) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("|") reduce 3)) 2576 ((("%") reduce 12) (("(") reduce 12) ((")") reduce 12) (("*") reduce 12) (("+") reduce 12) ((",") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_PIPE) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("|") reduce 12)) 2577 ((("%") reduce 54) (("(") reduce 54) ((")") reduce 54) (("*") reduce 54) (("+") reduce 54) ((",") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_PIPE) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("|") reduce 54)) 2578 ((("%") reduce 65) (("(") reduce 65) ((")") reduce 65) (("*") reduce 65) (("+") reduce 65) ((",") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_PIPE) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("|") reduce 65)) 2579 ((("%") reduce 75) (("(") reduce 75) ((")") reduce 75) (("*") reduce 75) (("+") reduce 75) ((",") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_PIPE) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("|") reduce 75)) 2580 ((("%") reduce 30) (("(") reduce 30) ((")") reduce 30) (("*") reduce 30) (("+") reduce 30) ((",") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_PIPE) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("|") reduce 30)) 2581 ((("%") reduce 68) (("(") reduce 68) ((")") reduce 68) (("*") reduce 68) (("+") reduce 68) ((",") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_PIPE) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("|") reduce 68)) 2582 ((("%") reduce 21) (("(") reduce 21) ((")") reduce 21) (("*") reduce 21) (("+") reduce 21) ((",") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_PIPE) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("|") reduce 21)) 2583 ((("%") reduce 23) (("(") reduce 23) ((")") reduce 23) (("*") reduce 23) (("+") reduce 23) ((",") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_PIPE) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("|") reduce 23)) 2584 ((("%") reduce 41) (("(") reduce 41) ((")") reduce 41) (("*") reduce 41) (("+") reduce 41) ((",") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_PIPE) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("|") reduce 41)) 2585 ((("%") reduce 62) (("(") reduce 62) ((")") reduce 62) (("*") reduce 62) (("+") reduce 62) ((",") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_PIPE) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("|") reduce 62)) 2586 ((("%") reduce 34) (("(") reduce 34) ((")") reduce 34) (("*") reduce 34) (("+") reduce 34) ((",") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_PIPE) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("|") reduce 34)) 2587 ((("%") reduce 40) (("(") reduce 40) ((")") reduce 40) (("*") reduce 40) (("+") reduce 40) ((",") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_PIPE) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("|") reduce 40)) 2588 ((("%") reduce 13) (("(") reduce 13) ((")") reduce 13) (("*") reduce 13) (("+") reduce 13) ((",") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_PIPE) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("|") reduce 13)) 2589 ((("%") reduce 55) (("(") reduce 55) ((")") reduce 55) (("*") reduce 55) (("+") reduce 55) ((",") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_PIPE) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("|") reduce 55)) 2590 ((("%") reduce 1) (("(") reduce 1) ((")") reduce 1) (("*") reduce 1) (("+") reduce 1) ((",") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_PIPE) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("|") reduce 1)) 2591 ((("%") reduce 2) (("(") reduce 2) ((")") reduce 2) (("*") reduce 2) (("+") reduce 2) ((",") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_PIPE) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("|") reduce 2)) 2592 ((("%") reduce 9) (("(") reduce 9) ((")") reduce 9) (("*") reduce 9) (("+") reduce 9) ((",") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_PIPE) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("|") reduce 9)) 2593 ((("%") reduce 33) (("(") reduce 33) ((")") reduce 33) (("*") reduce 33) (("+") reduce 33) ((",") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_PIPE) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("|") reduce 33)) 2594 ((("%") reduce 58) (("(") reduce 58) ((")") reduce 58) (("*") reduce 58) (("+") reduce 58) ((",") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_PIPE) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("|") reduce 58)) 2595 ((("%") reduce 37) (("(") reduce 37) ((")") reduce 37) (("*") reduce 37) (("+") reduce 37) ((",") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_PIPE) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("|") reduce 37)) 2596 ((("%") reduce 64) (("(") reduce 64) ((")") reduce 64) (("*") reduce 64) (("+") reduce 64) ((",") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_PIPE) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("|") reduce 64)) 2597 ((("%") reduce 46) (("(") reduce 46) ((")") reduce 46) (("*") reduce 46) (("+") reduce 46) ((",") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_PIPE) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("|") reduce 46)) 2598 ((("%") reduce 8) (("(") reduce 8) ((")") reduce 8) (("*") reduce 8) (("+") reduce 8) ((",") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_PIPE) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("|") reduce 8)) 2599 ((("%") reduce 6) (("(") reduce 6) ((")") reduce 6) (("*") reduce 6) (("+") reduce 6) ((",") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_PIPE) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("|") reduce 6)) 2600 ((("%") reduce 7) (("(") reduce 7) ((")") reduce 7) (("*") reduce 7) (("+") reduce 7) ((",") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_PIPE) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("|") reduce 7)) 2601 ((("%") reduce 69) (("(") reduce 69) ((")") reduce 69) (("*") reduce 69) (("+") reduce 69) ((",") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_PIPE) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("|") reduce 69)) 2602 ((("%") reduce 63) (("(") reduce 63) ((")") reduce 63) (("*") reduce 63) (("+") reduce 63) ((",") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_PIPE) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("|") reduce 63)) 2603 ((("%") reduce 56) (("(") reduce 56) ((")") reduce 56) (("*") reduce 56) (("+") reduce 56) ((",") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_PIPE) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("|") reduce 56)) 2604 ((("%") reduce 10) (("(") reduce 10) ((")") reduce 10) (("*") reduce 10) (("+") reduce 10) ((",") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_PIPE) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("|") reduce 10)) 2605 ((("%") reduce 67) (("(") reduce 67) ((")") reduce 67) (("*") reduce 67) (("+") reduce 67) ((",") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_PIPE) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("|") reduce 67)) 2606 ((("%") reduce 44) (("(") reduce 44) ((")") reduce 44) (("*") reduce 44) (("+") reduce 44) ((",") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_PIPE) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("|") reduce 44)) 2607 ((("%") reduce 76) (("(") reduce 76) ((")") reduce 76) (("*") reduce 76) (("+") reduce 76) ((",") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_PIPE) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("|") reduce 76)) 2608 ((("%") reduce 71) (("(") reduce 71) ((")") reduce 71) (("*") reduce 71) (("+") reduce 71) ((",") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_PIPE) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("|") reduce 71)) 2609 ((("%") reduce 77) (("(") reduce 77) ((")") reduce 77) (("*") reduce 77) (("+") reduce 77) ((",") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_PIPE) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("|") reduce 77)) 2610 ((("%") reduce 78) (("(") reduce 78) ((")") reduce 78) (("*") reduce 78) (("+") reduce 78) ((",") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_PIPE) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("|") reduce 78)) 2611 ((("%") reduce 79) (("(") reduce 79) ((")") reduce 79) (("*") reduce 79) (("+") reduce 79) ((",") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_PIPE) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("|") reduce 79)) 2612 ((("%") reduce 4) (("(") reduce 4) ((")") reduce 4) (("*") reduce 4) (("+") reduce 4) ((",") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_PIPE) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("|") reduce 4)) 2613 ((("%") reduce 5) (("(") reduce 5) ((")") reduce 5) (("*") reduce 5) (("+") reduce 5) ((",") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_PIPE) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("|") reduce 5)) 2614 ((("%") reduce 43) (("(") reduce 43) ((")") reduce 43) (("*") reduce 43) (("+") reduce 43) ((",") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_PIPE) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("|") reduce 43)) 2615 ((("%") reduce 73) (("(") reduce 73) ((")") reduce 73) (("*") reduce 73) (("+") reduce 73) ((",") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_PIPE) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("|") reduce 73)) 2616 ((("%") reduce 82) (("(") reduce 82) ((")") reduce 82) (("*") reduce 82) (("+") reduce 82) ((",") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_PIPE) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("|") reduce 82)) 2617 ((("%") reduce 47) (("(") reduce 47) ((")") reduce 47) (("*") reduce 47) (("+") reduce 47) ((",") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_PIPE) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("|") reduce 47)) 2618 ((("%") reduce 31) (("(") reduce 31) ((")") reduce 31) (("*") reduce 31) (("+") reduce 31) ((",") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_PIPE) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("|") reduce 31)) 2619 ((("%") reduce 57) (("(") reduce 57) ((")") reduce 57) (("*") reduce 57) (("+") reduce 57) ((",") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_PIPE) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("|") reduce 57)) 2620 ((("%") reduce 61) (("(") reduce 61) ((")") reduce 61) (("*") reduce 61) (("+") reduce 61) ((",") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_PIPE) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("|") reduce 61)) 2621 ((("%") reduce 28) (("(") reduce 28) ((")") reduce 28) (("*") reduce 28) (("+") reduce 28) ((",") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_PIPE) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("|") reduce 28)) 2622 ((("%") reduce 36) (("(") reduce 36) ((")") reduce 36) (("*") reduce 36) (("+") reduce 36) ((",") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_PIPE) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("|") reduce 36)) 2623 ((("%") reduce 32) (("(") reduce 32) ((")") reduce 32) (("*") reduce 32) (("+") reduce 32) ((",") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_PIPE) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("|") reduce 32)) 2624 ((("%") reduce 35) (("(") reduce 35) ((")") reduce 35) (("*") reduce 35) (("+") reduce 35) ((",") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_PIPE) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("|") reduce 35)) 2625 ((("%") reduce 19) (("(") reduce 19) ((")") reduce 19) (("*") reduce 19) (("+") reduce 19) ((",") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_PIPE) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("|") reduce 19)) 2626 ((("%") reduce 45) (("(") reduce 45) ((")") reduce 45) (("*") reduce 45) (("+") reduce 45) ((",") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_PIPE) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("|") reduce 45)) 2627 ((("%") reduce 542) (("(") reduce 579) ((")") reduce 542) (("*") reduce 542) (("+") reduce 542) ((",") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PIPE) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 2629 ((("%") reduce 72) (("(") reduce 72) ((")") reduce 72) (("*") reduce 72) (("+") reduce 72) ((",") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_PIPE) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("|") reduce 72)) 2630 ((("%") reduce 83) (("(") reduce 83) ((")") reduce 83) (("*") reduce 83) (("+") reduce 83) ((",") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_PIPE) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("|") reduce 83)) 2631 ((("%") reduce 572) (("(") reduce 581) ((")") reduce 572) (("*") reduce 572) (("+") reduce 572) ((",") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) (("<") reduce 572) (("=") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AND_EQUAL) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_COALESCE_EQUAL) reduce 572) ((T_CONCAT_EQUAL) reduce 572) ((T_DEC) reduce 572) ((T_DIV_EQUAL) reduce 572) ((T_INC) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_MINUS_EQUAL) reduce 572) ((T_MOD_EQUAL) reduce 572) ((T_MUL_EQUAL) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_OR_EQUAL) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_PLUS_EQUAL) reduce 572) ((T_POW) reduce 572) ((T_POW_EQUAL) reduce 572) ((T_SL) reduce 572) ((T_SL_EQUAL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) ((T_SR_EQUAL) reduce 572) ((T_XOR_EQUAL) reduce 572) (("[") reduce 572) (("^") reduce 572) (("|") reduce 572)) 2634 ((("%") reduce 544) (("(") reduce 580) ((")") reduce 544) (("*") reduce 544) (("+") reduce 544) ((",") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PIPE) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 2636 ((("%") reduce 507) (("(") reduce 507) ((")") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("|") reduce 507)) 2637 ((("%") reduce 291) (("(") reduce 291) ((")") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 2641 ((("%") reduce 292) (("(") reduce 292) ((")") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) (("=") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AND_EQUAL) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_COALESCE_EQUAL) reduce 292) ((T_CONCAT_EQUAL) reduce 292) ((T_DEC) reduce 292) ((T_DIV_EQUAL) reduce 292) ((T_INC) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_MINUS_EQUAL) reduce 292) ((T_MOD_EQUAL) reduce 292) ((T_MUL_EQUAL) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_OR_EQUAL) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_PLUS_EQUAL) reduce 292) ((T_POW) reduce 292) ((T_POW_EQUAL) reduce 292) ((T_SL) reduce 292) ((T_SL_EQUAL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) ((T_SR_EQUAL) reduce 292) ((T_XOR_EQUAL) reduce 292) (("[") reduce 292) (("^") reduce 292) (("|") reduce 292)) 2642 ((("%") reduce 293) (("(") reduce 293) ((")") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) (("=") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AND_EQUAL) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_COALESCE_EQUAL) reduce 293) ((T_CONCAT_EQUAL) reduce 293) ((T_DEC) reduce 293) ((T_DIV_EQUAL) reduce 293) ((T_INC) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_MINUS_EQUAL) reduce 293) ((T_MOD_EQUAL) reduce 293) ((T_MUL_EQUAL) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_OR_EQUAL) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_PLUS_EQUAL) reduce 293) ((T_POW) reduce 293) ((T_POW_EQUAL) reduce 293) ((T_SL) reduce 293) ((T_SL_EQUAL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) ((T_SR_EQUAL) reduce 293) ((T_XOR_EQUAL) reduce 293) (("[") reduce 293) (("^") reduce 293) (("|") reduce 293)) 2647 ((("%") reduce 425) ((")") reduce 425) (("*") reduce 425) (("+") reduce 425) ((",") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_PIPE) reduce 425) ((T_POW) reduce 425) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 2649 ((("%") reduce 423) ((")") reduce 423) (("*") reduce 423) (("+") reduce 423) ((",") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_PIPE) reduce 423) ((T_POW) reduce 423) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 2659 ((("%") shift) ((")") reduce 419) (("*") shift) (("+") shift) ((",") reduce 419) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2660 ((("%") shift) ((")") reduce 421) (("*") shift) (("+") shift) ((",") reduce 421) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2661 ((("%") shift) ((")") reduce 420) (("*") shift) (("+") shift) ((",") reduce 420) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2662 ((("%") shift) ((")") reduce 413) (("*") shift) (("+") shift) ((",") reduce 413) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2663 ((("%") shift) ((")") reduce 410) (("*") shift) (("+") shift) ((",") reduce 410) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2664 ((("%") shift) ((")") reduce 418) (("*") shift) (("+") shift) ((",") reduce 418) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2665 ((("%") shift) ((")") reduce 412) (("*") shift) (("+") shift) ((",") reduce 412) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2666 ((("%") shift) ((")") reduce 416) (("*") shift) (("+") shift) ((",") reduce 416) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2667 ((("%") shift) ((")") reduce 411) (("*") shift) (("+") shift) ((",") reduce 411) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2668 ((("%") shift) ((")") reduce 414) (("*") shift) (("+") shift) ((",") reduce 414) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2669 ((("%") shift) ((")") reduce 415) (("*") shift) (("+") shift) ((",") reduce 415) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2670 ((("%") shift) ((")") reduce 422) (("*") shift) (("+") shift) ((",") reduce 422) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2671 ((("%") shift) ((")") reduce 417) (("*") shift) (("+") shift) ((",") reduce 417) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2673 ((("%") shift) ((")") reduce 406) (("*") shift) (("+") shift) ((",") reduce 406) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2676 ((("%") reduce 568) (("(") reduce 568) ((")") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 2679 ((("%") reduce 564) (("(") reduce 555) ((")") reduce 564) (("*") reduce 564) (("+") reduce 564) ((",") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) (("<") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_POW) reduce 564) ((T_SL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) (("[") reduce 564) (("^") reduce 564) (("|") reduce 564)) 2681 ((("%") reduce 563) (("(") reduce 563) ((")") reduce 563) (("*") reduce 563) (("+") reduce 563) ((",") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) (("<") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_POW) reduce 563) ((T_SL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) (("[") reduce 563) (("^") reduce 563) (("|") reduce 563)) 2683 ((("%") reduce 559) (("(") reduce 559) ((")") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 2684 ((("%") reduce 565) ((")") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 2685 ((("%") reduce 407) ((")") reduce 407) (("*") reduce 407) (("+") reduce 407) ((",") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("[") reduce 548) (("^") reduce 407) (("|") reduce 407)) 2689 ((("%") reduce 572) (("(") reduce 581) ((")") reduce 572) (("*") reduce 572) (("+") reduce 572) ((",") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) (("<") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_POW) reduce 572) ((T_SL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) (("[") reduce 572) (("^") reduce 572) (("|") reduce 572)) 2691 ((("%") reduce 507) (("(") reduce 507) ((")") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("|") reduce 507)) 2692 ((("%") reduce 291) (("(") reduce 291) ((")") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 2696 ((("%") reduce 292) (("(") reduce 292) ((")") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("[") reduce 292) (("^") reduce 292) (("|") reduce 292)) 2697 ((("%") reduce 293) (("(") reduce 293) ((")") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("[") reduce 293) (("^") reduce 293) (("|") reduce 293)) 2698 ((("%") reduce 504) (("(") reduce 504) ((")") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 2701 ((("%") reduce 571) (("(") reduce 581) ((")") reduce 571) (("*") reduce 571) (("+") reduce 571) ((",") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) (("<") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_POW) reduce 571) ((T_SL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) (("[") reduce 571) (("^") reduce 571) (("|") reduce 571)) 2702 ((("%") reduce 506) (("(") reduce 506) ((")") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("|") reduce 506)) 2703 ((("%") reduce 508) (("(") reduce 508) ((")") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("|") reduce 508)) 2708 ((("%") reduce 560) (("(") reduce 560) ((")") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 2709 ((("%") reduce 582) (("(") reduce 582) ((")") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 2710 ((("%") reduce 566) (("(") shift) ((")") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 2711 ((("%") reduce 584) (("(") reduce 584) ((")") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 2714 ((("%") reduce 583) (("(") reduce 583) ((")") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 2715 ((("%") reduce 561) (("(") reduce 561) ((")") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 2716 ((("%") reduce 567) (("(") shift) ((")") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 2717 ((("%") reduce 562) (("(") reduce 562) ((")") reduce 562) (("*") reduce 562) (("+") reduce 562) ((",") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("[") reduce 562) (("^") reduce 562) (("|") reduce 562)) 2718 ((("%") reduce 505) (("(") reduce 505) ((")") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 2719 ((("%") reduce 570) (("(") reduce 570) ((")") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 2722 ((("%") reduce 569) (("(") reduce 569) ((")") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 2723 ((("%") reduce 504) (("(") reduce 504) ((")") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) (("=") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AND_EQUAL) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_COALESCE_EQUAL) reduce 504) ((T_CONCAT_EQUAL) reduce 504) ((T_DEC) reduce 504) ((T_DIV_EQUAL) reduce 504) ((T_INC) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_MINUS_EQUAL) reduce 504) ((T_MOD_EQUAL) reduce 504) ((T_MUL_EQUAL) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_OR_EQUAL) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_PLUS_EQUAL) reduce 504) ((T_POW) reduce 504) ((T_POW_EQUAL) reduce 504) ((T_SL) reduce 504) ((T_SL_EQUAL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) ((T_SR_EQUAL) reduce 504) ((T_XOR_EQUAL) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 2734 ((("%") reduce 494) ((")") reduce 494) (("*") reduce 494) (("+") reduce 494) ((",") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PIPE) reduce 494) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 2735 ((("%") reduce 489) ((")") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_PIPE) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("^") reduce 489) (("|") reduce 489)) 2745 ((("%") shift) ((")") reduce 494) (("*") shift) (("+") shift) ((",") reduce 494) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2746 ((("%") reduce 490) ((")") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_PIPE) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("^") reduce 490) (("|") reduce 490)) 2748 ((("%") reduce 541) (("(") reduce 579) ((")") reduce 541) (("*") reduce 541) (("+") reduce 541) ((",") reduce 541) (("-") reduce 541) ((".") reduce 541) (("/") reduce 541) (("<") reduce 541) ((">") reduce 541) (("?") reduce 541) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_BOOLEAN_AND) reduce 541) ((T_BOOLEAN_OR) reduce 541) ((T_COALESCE) reduce 541) ((T_INSTANCEOF) reduce 541) ((T_IS_EQUAL) reduce 541) ((T_IS_GREATER_OR_EQUAL) reduce 541) ((T_IS_IDENTICAL) reduce 541) ((T_IS_NOT_EQUAL) reduce 541) ((T_IS_NOT_IDENTICAL) reduce 541) ((T_IS_SMALLER_OR_EQUAL) reduce 541) ((T_LOGICAL_AND) reduce 541) ((T_LOGICAL_OR) reduce 541) ((T_LOGICAL_XOR) reduce 541) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 541) ((T_OBJECT_OPERATOR) reduce 541) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 541) ((T_PIPE) reduce 541) ((T_POW) reduce 541) ((T_SL) reduce 541) ((T_SPACESHIP) reduce 541) ((T_SR) reduce 541) (("[") reduce 541) (("^") reduce 541) (("|") reduce 541)) 2750 ((("%") reduce 571) (("(") reduce 581) ((")") reduce 571) (("*") reduce 571) (("+") reduce 571) ((",") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) (("<") reduce 571) (("=") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AND_EQUAL) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_COALESCE_EQUAL) reduce 571) ((T_CONCAT_EQUAL) reduce 571) ((T_DEC) reduce 571) ((T_DIV_EQUAL) reduce 571) ((T_INC) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_MINUS_EQUAL) reduce 571) ((T_MOD_EQUAL) reduce 571) ((T_MUL_EQUAL) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_OR_EQUAL) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_PLUS_EQUAL) reduce 571) ((T_POW) reduce 571) ((T_POW_EQUAL) reduce 571) ((T_SL) reduce 571) ((T_SL_EQUAL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) ((T_SR_EQUAL) reduce 571) ((T_XOR_EQUAL) reduce 571) (("[") reduce 571) (("^") reduce 571) (("|") reduce 571)) 2753 ((("%") reduce 543) (("(") reduce 580) ((")") reduce 543) (("*") reduce 543) (("+") reduce 543) ((",") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PIPE) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 2754 ((("%") reduce 506) (("(") reduce 506) ((")") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("|") reduce 506)) 2755 ((("%") reduce 508) (("(") reduce 508) ((")") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("|") reduce 508)) 2756 (((T_FN) shift) ((T_FUNCTION) shift)) 2757 (((T_ATTRIBUTE) reduce 102) ((T_FN) reduce 102) ((T_FUNCTION) reduce 102) ((T_STATIC) reduce 102)) 2758 ((("%") reduce 485) ((")") reduce 485) (("*") reduce 485) (("+") reduce 485) ((",") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_PIPE) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("^") reduce 485) (("|") reduce 485)) 2759 ((("%") reduce 487) ((")") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_PIPE) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("^") reduce 487) (("|") reduce 487)) 2764 ((("%") reduce 560) (("(") reduce 560) ((")") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) (("=") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AND_EQUAL) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_COALESCE_EQUAL) reduce 560) ((T_CONCAT_EQUAL) reduce 560) ((T_DEC) reduce 560) ((T_DIV_EQUAL) reduce 560) ((T_INC) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_MINUS_EQUAL) reduce 560) ((T_MOD_EQUAL) reduce 560) ((T_MUL_EQUAL) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_OR_EQUAL) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_PLUS_EQUAL) reduce 560) ((T_POW) reduce 560) ((T_POW_EQUAL) reduce 560) ((T_SL) reduce 560) ((T_SL_EQUAL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) ((T_SR_EQUAL) reduce 560) ((T_XOR_EQUAL) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 2765 ((("%") reduce 582) (("(") reduce 582) ((")") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) (("=") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AND_EQUAL) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_COALESCE_EQUAL) reduce 582) ((T_CONCAT_EQUAL) reduce 582) ((T_DEC) reduce 582) ((T_DIV_EQUAL) reduce 582) ((T_INC) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_MINUS_EQUAL) reduce 582) ((T_MOD_EQUAL) reduce 582) ((T_MUL_EQUAL) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_OR_EQUAL) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_PLUS_EQUAL) reduce 582) ((T_POW) reduce 582) ((T_POW_EQUAL) reduce 582) ((T_SL) reduce 582) ((T_SL_EQUAL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) ((T_SR_EQUAL) reduce 582) ((T_XOR_EQUAL) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 2766 ((("%") reduce 566) (("(") shift) ((")") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 2767 ((("%") reduce 584) (("(") reduce 584) ((")") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) (("=") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AND_EQUAL) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_COALESCE_EQUAL) reduce 584) ((T_CONCAT_EQUAL) reduce 584) ((T_DEC) reduce 584) ((T_DIV_EQUAL) reduce 584) ((T_INC) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_MINUS_EQUAL) reduce 584) ((T_MOD_EQUAL) reduce 584) ((T_MUL_EQUAL) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_OR_EQUAL) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_PLUS_EQUAL) reduce 584) ((T_POW) reduce 584) ((T_POW_EQUAL) reduce 584) ((T_SL) reduce 584) ((T_SL_EQUAL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) ((T_SR_EQUAL) reduce 584) ((T_XOR_EQUAL) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 2770 ((("%") reduce 583) (("(") reduce 583) ((")") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) (("=") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AND_EQUAL) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_COALESCE_EQUAL) reduce 583) ((T_CONCAT_EQUAL) reduce 583) ((T_DEC) reduce 583) ((T_DIV_EQUAL) reduce 583) ((T_INC) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_MINUS_EQUAL) reduce 583) ((T_MOD_EQUAL) reduce 583) ((T_MUL_EQUAL) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_OR_EQUAL) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_PLUS_EQUAL) reduce 583) ((T_POW) reduce 583) ((T_POW_EQUAL) reduce 583) ((T_SL) reduce 583) ((T_SL_EQUAL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) ((T_SR_EQUAL) reduce 583) ((T_XOR_EQUAL) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 2771 ((("%") reduce 561) (("(") reduce 561) ((")") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) (("=") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AND_EQUAL) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_COALESCE_EQUAL) reduce 561) ((T_CONCAT_EQUAL) reduce 561) ((T_DEC) reduce 561) ((T_DIV_EQUAL) reduce 561) ((T_INC) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_MINUS_EQUAL) reduce 561) ((T_MOD_EQUAL) reduce 561) ((T_MUL_EQUAL) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_OR_EQUAL) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_PLUS_EQUAL) reduce 561) ((T_POW) reduce 561) ((T_POW_EQUAL) reduce 561) ((T_SL) reduce 561) ((T_SL_EQUAL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) ((T_SR_EQUAL) reduce 561) ((T_XOR_EQUAL) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 2772 ((("%") reduce 567) (("(") shift) ((")") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 2773 ((("%") reduce 562) (("(") reduce 562) ((")") reduce 562) (("*") reduce 562) (("+") reduce 562) ((",") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) (("=") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AND_EQUAL) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_COALESCE_EQUAL) reduce 562) ((T_CONCAT_EQUAL) reduce 562) ((T_DEC) reduce 562) ((T_DIV_EQUAL) reduce 562) ((T_INC) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_MINUS_EQUAL) reduce 562) ((T_MOD_EQUAL) reduce 562) ((T_MUL_EQUAL) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_OR_EQUAL) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_PLUS_EQUAL) reduce 562) ((T_POW) reduce 562) ((T_POW_EQUAL) reduce 562) ((T_SL) reduce 562) ((T_SL_EQUAL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) ((T_SR_EQUAL) reduce 562) ((T_XOR_EQUAL) reduce 562) (("[") reduce 562) (("^") reduce 562) (("|") reduce 562)) 2776 ((("(") reduce 568) ((")") reduce 568) ((",") reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) (("[") reduce 568)) 2779 ((("(") reduce 555) ((")") reduce 564) ((",") reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) (("[") reduce 564)) 2781 ((("(") reduce 563) ((")") reduce 563) ((",") reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) (("[") reduce 563)) 2783 ((("(") reduce 559) ((")") reduce 559) ((",") reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) (("[") reduce 559)) 2784 (((")") reduce 565) ((",") reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) (("[") reduce 565)) 2785 (((")") reduce 592) ((",") reduce 592) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) (("[") reduce 548)) 2789 ((("(") reduce 581) ((")") reduce 572) ((",") reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) (("[") reduce 572)) 2791 ((("(") reduce 507) ((")") reduce 507) ((",") reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) (("[") reduce 507)) 2792 ((("(") reduce 291) ((")") reduce 291) ((",") reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) (("[") reduce 291)) 2796 ((("(") reduce 292) ((")") reduce 292) ((",") reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) (("[") reduce 292)) 2797 ((("(") reduce 293) ((")") reduce 293) ((",") reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) (("[") reduce 293)) 2798 ((("(") reduce 504) ((")") reduce 504) ((",") reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) (("[") reduce 504)) 2801 ((("(") reduce 581) ((")") reduce 571) ((",") reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) (("[") reduce 571)) 2802 ((("(") reduce 506) ((")") reduce 506) ((",") reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) (("[") reduce 506)) 2803 ((("(") reduce 508) ((")") reduce 508) ((",") reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) (("[") reduce 508)) 2808 ((("(") reduce 560) ((")") reduce 560) ((",") reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) (("[") reduce 560)) 2809 ((("(") reduce 582) ((")") reduce 582) ((",") reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) (("[") reduce 582)) 2810 ((("(") shift) ((")") reduce 566) ((",") reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) (("[") reduce 566)) 2811 ((("(") reduce 584) ((")") reduce 584) ((",") reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) (("[") reduce 584)) 2814 ((("(") reduce 583) ((")") reduce 583) ((",") reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) (("[") reduce 583)) 2815 ((("(") reduce 561) ((")") reduce 561) ((",") reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) (("[") reduce 561)) 2816 ((("(") shift) ((")") reduce 567) ((",") reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) (("[") reduce 567)) 2817 ((("(") reduce 562) ((")") reduce 562) ((",") reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) (("[") reduce 562)) 2818 ((("(") reduce 505) ((")") reduce 505) ((",") reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) (("[") reduce 505)) 2819 ((("(") reduce 570) ((")") reduce 570) ((",") reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) (("[") reduce 570)) 2822 ((("(") reduce 569) ((")") reduce 569) ((",") reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) (("[") reduce 569)) 2824 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_STRING_VARNAME) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 2825 (((T_CURLY_OPEN) shift) ((T_DOLLAR_OPEN_CURLY_BRACES) shift) ((T_VARIABLE) shift) (("`") reduce 515)) 2826 (((T_CURLY_OPEN) reduce 601) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 601) ((T_ENCAPSED_AND_WHITESPACE) reduce 601) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_VARIABLE) reduce 601) (("[") shift) (("`") reduce 601)) 2827 ((("`") shift)) 2828 (((T_CURLY_OPEN) shift) ((T_DOLLAR_OPEN_CURLY_BRACES) shift) ((T_ENCAPSED_AND_WHITESPACE) shift) ((T_VARIABLE) shift) (("`") reduce 516)) 2829 (((T_CURLY_OPEN) reduce 599) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 599) ((T_ENCAPSED_AND_WHITESPACE) reduce 599) ((T_VARIABLE) reduce 599) (("`") reduce 599)) 2830 (((T_CURLY_OPEN) reduce 598) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 598) ((T_ENCAPSED_AND_WHITESPACE) reduce 598) ((T_VARIABLE) reduce 598) (("`") reduce 598)) 2831 (((T_CURLY_OPEN) reduce 597) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 597) ((T_ENCAPSED_AND_WHITESPACE) reduce 597) ((T_VARIABLE) reduce 597) (("`") reduce 597)) 2832 ((("%") reduce 477) ((")") reduce 477) (("*") reduce 477) (("+") reduce 477) ((",") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_PIPE) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 2835 ((("-") shift) ((T_NUM_STRING) shift) ((T_STRING) shift) ((T_VARIABLE) shift)) 2836 (((T_NUM_STRING) shift)) 2837 ((("]") reduce 610)) 2838 ((("]") reduce 609)) 2839 ((("]") reduce 612)) 2841 (((T_CURLY_OPEN) reduce 602) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 602) ((T_ENCAPSED_AND_WHITESPACE) reduce 602) ((T_VARIABLE) reduce 602) (("`") reduce 602)) 2842 ((("]") reduce 611)) 2843 (((T_CURLY_OPEN) reduce 603) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 603) ((T_ENCAPSED_AND_WHITESPACE) reduce 603) ((T_VARIABLE) reduce 603) (("`") reduce 603)) 2844 (((T_CURLY_OPEN) reduce 604) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 604) ((T_ENCAPSED_AND_WHITESPACE) reduce 604) ((T_VARIABLE) reduce 604) (("`") reduce 604)) 2845 (((T_CURLY_OPEN) reduce 600) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 600) ((T_ENCAPSED_AND_WHITESPACE) reduce 600) ((T_VARIABLE) reduce 600) (("`") reduce 600)) 2846 ((("[") shift) (("}") shift)) 2848 (((T_CURLY_OPEN) reduce 605) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 605) ((T_ENCAPSED_AND_WHITESPACE) reduce 605) ((T_VARIABLE) reduce 605) (("`") reduce 605)) 2850 (((T_CURLY_OPEN) reduce 606) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 606) ((T_ENCAPSED_AND_WHITESPACE) reduce 606) ((T_VARIABLE) reduce 606) (("`") reduce 606)) 2851 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") shift) (("^") shift) (("|") shift)) 2852 ((("}") shift)) 2853 (((T_CURLY_OPEN) reduce 607) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 607) ((T_ENCAPSED_AND_WHITESPACE) reduce 607) ((T_VARIABLE) reduce 607) (("`") reduce 607)) 2856 ((("(") reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) (("[") reduce 568) (("}") reduce 568)) 2859 ((("(") reduce 555) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) (("[") reduce 564) (("}") reduce 564)) 2861 ((("(") reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) (("[") reduce 563) (("}") reduce 563)) 2863 ((("(") reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) (("[") reduce 559) (("}") reduce 559)) 2864 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) (("[") reduce 565) (("}") reduce 565)) 2865 (((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) (("[") reduce 548) (("}") shift)) 2869 ((("(") reduce 581) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) (("[") reduce 572) (("}") reduce 572)) 2871 ((("(") reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) (("[") reduce 507) (("}") reduce 507)) 2872 ((("(") reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) (("[") reduce 291) (("}") reduce 291)) 2876 ((("(") reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) (("[") reduce 292) (("}") reduce 292)) 2877 ((("(") reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) (("[") reduce 293) (("}") reduce 293)) 2878 (((T_CURLY_OPEN) reduce 608) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 608) ((T_ENCAPSED_AND_WHITESPACE) reduce 608) ((T_VARIABLE) reduce 608) (("`") reduce 608)) 2879 ((("(") reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) (("[") reduce 504) (("}") reduce 504)) 2882 ((("(") reduce 581) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) (("[") reduce 571) (("}") reduce 571)) 2883 ((("(") reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) (("[") reduce 506) (("}") reduce 506)) 2884 ((("(") reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) (("[") reduce 508) (("}") reduce 508)) 2889 ((("(") reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) (("[") reduce 560) (("}") reduce 560)) 2890 ((("(") reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) (("[") reduce 582) (("}") reduce 582)) 2891 ((("(") shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) (("[") reduce 566) (("}") reduce 566)) 2892 ((("(") reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) (("[") reduce 584) (("}") reduce 584)) 2895 ((("(") reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) (("[") reduce 583) (("}") reduce 583)) 2896 ((("(") reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) (("[") reduce 561) (("}") reduce 561)) 2897 ((("(") shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) (("[") reduce 567) (("}") reduce 567)) 2898 ((("(") reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) (("[") reduce 562) (("}") reduce 562)) 2899 ((("(") reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) (("[") reduce 505) (("}") reduce 505)) 2900 ((("(") reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) (("[") reduce 570) (("}") reduce 570)) 2903 ((("(") reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) (("[") reduce 569) (("}") reduce 569)) 2905 ((("%") reduce 520) (("(") reduce 520) ((")") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) (("=") shift) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_PIPE) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("|") reduce 520)) 2907 ((("%") shift) ((")") reduce 405) (("*") shift) (("+") shift) ((",") reduce 405) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2908 ((("%") shift) ((")") reduce 482) (("*") shift) (("+") shift) ((",") reduce 482) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2911 ((("%") shift) ((")") reduce 481) (("*") shift) (("+") shift) ((",") reduce 481) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2912 ((("%") reduce 473) ((")") reduce 473) (("*") reduce 473) (("+") reduce 473) ((",") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_PIPE) reduce 473) ((T_POW) shift) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("|") reduce 473)) 2913 ((("%") shift) ((")") reduce 483) (("*") shift) (("+") shift) ((",") reduce 483) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2914 ((("%") reduce 469) ((")") reduce 469) (("*") reduce 469) (("+") reduce 469) ((",") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_PIPE) reduce 469) ((T_POW) shift) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 2915 ((("%") reduce 486) ((")") reduce 486) (("*") reduce 486) (("+") reduce 486) ((",") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_PIPE) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("^") reduce 486) (("|") reduce 486)) 2918 (((T_CURLY_OPEN) shift) ((T_DOLLAR_OPEN_CURLY_BRACES) shift) ((T_END_HEREDOC) shift) ((T_VARIABLE) shift)) 2919 ((("%") reduce 526) ((")") reduce 526) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_PIPE) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("^") reduce 526) (("|") reduce 526)) 2920 (((T_CURLY_OPEN) reduce 601) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 601) ((T_ENCAPSED_AND_WHITESPACE) reduce 601) ((T_END_HEREDOC) reduce 601) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_VARIABLE) reduce 601) (("[") shift)) 2922 (((T_CURLY_OPEN) reduce 599) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 599) ((T_ENCAPSED_AND_WHITESPACE) reduce 599) ((T_END_HEREDOC) reduce 599) ((T_VARIABLE) reduce 599)) 2923 (((T_CURLY_OPEN) reduce 598) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 598) ((T_ENCAPSED_AND_WHITESPACE) reduce 598) ((T_END_HEREDOC) reduce 598) ((T_VARIABLE) reduce 598)) 2924 ((("%") reduce 527) ((")") reduce 527) (("*") reduce 527) (("+") reduce 527) ((",") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_PIPE) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("^") reduce 527) (("|") reduce 527)) 2925 (((T_CURLY_OPEN) reduce 597) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 597) ((T_ENCAPSED_AND_WHITESPACE) reduce 597) ((T_END_HEREDOC) reduce 597) ((T_VARIABLE) reduce 597)) 2930 (((T_CURLY_OPEN) reduce 602) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 602) ((T_ENCAPSED_AND_WHITESPACE) reduce 602) ((T_END_HEREDOC) reduce 602) ((T_VARIABLE) reduce 602)) 2931 (((T_CURLY_OPEN) reduce 603) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 603) ((T_ENCAPSED_AND_WHITESPACE) reduce 603) ((T_END_HEREDOC) reduce 603) ((T_VARIABLE) reduce 603)) 2932 (((T_CURLY_OPEN) reduce 604) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 604) ((T_ENCAPSED_AND_WHITESPACE) reduce 604) ((T_END_HEREDOC) reduce 604) ((T_VARIABLE) reduce 604)) 2933 ((("%") reduce 525) ((")") reduce 525) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_PIPE) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("^") reduce 525) (("|") reduce 525)) 2934 (((T_CURLY_OPEN) reduce 600) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 600) ((T_ENCAPSED_AND_WHITESPACE) reduce 600) ((T_END_HEREDOC) reduce 600) ((T_VARIABLE) reduce 600)) 2937 (((T_CURLY_OPEN) reduce 605) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 605) ((T_ENCAPSED_AND_WHITESPACE) reduce 605) ((T_END_HEREDOC) reduce 605) ((T_VARIABLE) reduce 605)) 2939 (((T_CURLY_OPEN) reduce 606) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 606) ((T_ENCAPSED_AND_WHITESPACE) reduce 606) ((T_END_HEREDOC) reduce 606) ((T_VARIABLE) reduce 606)) 2942 (((T_CURLY_OPEN) reduce 607) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 607) ((T_ENCAPSED_AND_WHITESPACE) reduce 607) ((T_END_HEREDOC) reduce 607) ((T_VARIABLE) reduce 607)) 2944 (((T_CURLY_OPEN) reduce 608) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 608) ((T_ENCAPSED_AND_WHITESPACE) reduce 608) ((T_END_HEREDOC) reduce 608) ((T_VARIABLE) reduce 608)) 2945 ((("%") shift) ((")") reduce 619) (("*") shift) (("+") shift) ((",") reduce 619) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2946 ((("%") shift) ((")") reduce 618) (("*") shift) (("+") shift) ((",") reduce 618) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2947 ((("%") reduce 505) (("(") reduce 505) ((")") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 2948 ((("%") shift) ((")") reduce 478) (("*") shift) (("+") shift) ((",") reduce 478) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 2949 ((("%") reduce 471) ((")") reduce 471) (("*") reduce 471) (("+") reduce 471) ((",") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_PIPE) reduce 471) ((T_POW) shift) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("|") reduce 471)) 2952 ((("%") reduce 94) (("(") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 2953 ((("%") reduce 93) (("(") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 2954 ((("%") reduce 95) (("(") reduce 95) ((")") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 2955 (((T_ABSTRACT) reduce 199) ((T_CLASS) reduce 199) ((T_FINAL) reduce 199) ((T_READONLY) reduce 199)) 2956 ((("%") reduce 509) (("(") reduce 509) ((")") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 2957 ((("%") reduce 92) (("(") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 2958 ((("%") reduce 400) (("(") reduce 400) ((")") reduce 400) (("*") reduce 400) (("+") reduce 400) ((",") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 400) ((T_OBJECT_OPERATOR) reduce 400) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 400) ((T_PIPE) reduce 400) ((T_POW) reduce 400) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("[") reduce 400) (("^") reduce 400) (("|") reduce 400)) 2959 (((T_ABSTRACT) shift) ((T_CLASS) reduce 196) ((T_FINAL) shift) ((T_READONLY) shift)) 2960 (((T_CLASS) shift)) 2961 (((T_ABSTRACT) reduce 101) ((T_ATTRIBUTE) reduce 101) ((T_CLASS) reduce 101) ((T_FINAL) reduce 101) ((T_READONLY) reduce 101)) 2962 (((T_ABSTRACT) shift) ((T_ATTRIBUTE) shift) ((T_CLASS) reduce 195) ((T_FINAL) shift) ((T_READONLY) shift)) 2963 (((T_ABSTRACT) reduce 193) ((T_CLASS) reduce 193) ((T_FINAL) reduce 193) ((T_READONLY) reduce 193)) 2964 ((("%") reduce 511) (("(") reduce 511) ((")") reduce 511) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("|") reduce 511)) 2965 ((("%") reduce 402) (("(") shift) ((")") reduce 402) (("*") reduce 402) (("+") reduce 402) ((",") reduce 402) (("-") reduce 402) ((".") reduce 402) (("/") reduce 402) (("<") reduce 402) ((">") reduce 402) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_INSTANCEOF) reduce 402) ((T_IS_EQUAL) reduce 402) ((T_IS_GREATER_OR_EQUAL) reduce 402) ((T_IS_IDENTICAL) reduce 402) ((T_IS_NOT_EQUAL) reduce 402) ((T_IS_NOT_IDENTICAL) reduce 402) ((T_IS_SMALLER_OR_EQUAL) reduce 402) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_PIPE) reduce 402) ((T_POW) reduce 402) ((T_SL) reduce 402) ((T_SPACESHIP) reduce 402) ((T_SR) reduce 402) (("^") reduce 402) (("|") reduce 402)) 2966 ((("%") reduce 510) (("(") reduce 510) ((")") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 2967 ((("%") reduce 512) (("(") reduce 512) ((")") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("|") reduce 512)) 2968 ((("%") reduce 573) (("(") reduce 573) ((")") reduce 573) (("*") reduce 573) (("+") reduce 573) ((",") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("|") reduce 573)) 2974 ((("%") reduce 574) (("(") reduce 574) ((")") reduce 574) (("*") reduce 574) (("+") reduce 574) ((",") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("|") reduce 574)) 2975 ((("%") reduce 578) (("(") reduce 578) ((")") reduce 578) (("*") reduce 578) (("+") reduce 578) ((",") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("|") reduce 578)) 2976 ((("%") reduce 575) (("(") reduce 575) ((")") reduce 575) (("*") reduce 575) (("+") reduce 575) ((",") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("|") reduce 575)) 2977 ((("%") reduce 576) (("(") reduce 576) ((")") reduce 576) (("*") reduce 576) (("+") reduce 576) ((",") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("|") reduce 576)) 2978 ((("%") reduce 399) (("(") reduce 399) ((")") reduce 399) (("*") reduce 399) (("+") reduce 399) ((",") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_INSTANCEOF) reduce 399) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 399) ((T_OBJECT_OPERATOR) reduce 399) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 399) ((T_PIPE) reduce 399) ((T_POW) reduce 399) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("[") reduce 399) (("^") reduce 399) (("|") reduce 399)) 2980 ((("%") reduce 577) (("(") reduce 577) ((")") reduce 577) (("*") reduce 577) (("+") reduce 577) ((",") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("|") reduce 577)) 2981 ((("%") reduce 401) (("(") reduce 401) ((")") reduce 401) (("*") reduce 401) (("+") reduce 401) ((",") reduce 401) (("-") reduce 401) ((".") reduce 401) (("/") reduce 401) (("<") reduce 401) ((">") reduce 401) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_INSTANCEOF) reduce 401) ((T_IS_EQUAL) reduce 401) ((T_IS_GREATER_OR_EQUAL) reduce 401) ((T_IS_IDENTICAL) reduce 401) ((T_IS_NOT_EQUAL) reduce 401) ((T_IS_NOT_IDENTICAL) reduce 401) ((T_IS_SMALLER_OR_EQUAL) reduce 401) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 401) ((T_OBJECT_OPERATOR) reduce 401) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 401) ((T_PIPE) reduce 401) ((T_POW) reduce 401) ((T_SL) reduce 401) ((T_SPACESHIP) reduce 401) ((T_SR) reduce 401) (("[") reduce 401) (("^") reduce 401) (("|") reduce 401)) 2982 (((T_ABSTRACT) reduce 102) ((T_ATTRIBUTE) reduce 102) ((T_CLASS) reduce 102) ((T_FINAL) reduce 102) ((T_READONLY) reduce 102)) 2983 ((("(") shift) ((T_EXTENDS) reduce 517) ((T_IMPLEMENTS) reduce 517) (("{") reduce 517)) 2985 (((T_EXTENDS) reduce 518) ((T_IMPLEMENTS) reduce 518) (("{") reduce 518)) 2986 (((T_EXTENDS) shift) ((T_IMPLEMENTS) reduce 208) (("{") reduce 208)) 2988 (((T_IMPLEMENTS) shift) (("{") reduce 212)) 2990 ((("{") reduce 493)) 2992 (((T_ABSTRACT) reduce 315) ((T_ATTRIBUTE) reduce 315) ((T_CASE) reduce 315) ((T_CONST) reduce 315) ((T_FINAL) reduce 315) ((T_FUNCTION) reduce 315) ((T_PRIVATE) reduce 315) ((T_PRIVATE_SET) reduce 315) ((T_PROTECTED) reduce 315) ((T_PROTECTED_SET) reduce 315) ((T_PUBLIC) reduce 315) ((T_PUBLIC_SET) reduce 315) ((T_READONLY) reduce 315) ((T_STATIC) reduce 315) ((T_USE) reduce 315) ((T_VAR) reduce 315) (("}") reduce 315)) 2993 (((T_ABSTRACT) shift) ((T_ATTRIBUTE) shift) ((T_CASE) shift) ((T_CONST) reduce 348) ((T_FINAL) shift) ((T_FUNCTION) reduce 346) ((T_PRIVATE) shift) ((T_PRIVATE_SET) shift) ((T_PROTECTED) shift) ((T_PROTECTED_SET) shift) ((T_PUBLIC) shift) ((T_PUBLIC_SET) shift) ((T_READONLY) shift) ((T_STATIC) shift) ((T_USE) shift) ((T_VAR) shift) (("}") shift)) 2994 ((("(") reduce 359) (("?") reduce 359) ((T_ABSTRACT) reduce 359) ((T_ARRAY) reduce 359) ((T_CALLABLE) reduce 359) ((T_CONST) reduce 359) ((T_FINAL) reduce 359) ((T_FUNCTION) reduce 359) ((T_NAME_FULLY_QUALIFIED) reduce 359) ((T_NAME_QUALIFIED) reduce 359) ((T_NAME_RELATIVE) reduce 359) ((T_PRIVATE) reduce 359) ((T_PRIVATE_SET) reduce 359) ((T_PROTECTED) reduce 359) ((T_PROTECTED_SET) reduce 359) ((T_PUBLIC) reduce 359) ((T_PUBLIC_SET) reduce 359) ((T_READONLY) reduce 359) ((T_STATIC) reduce 359) ((T_STRING) reduce 359) ((T_VARIABLE) reduce 359)) 2996 (((T_ABSTRACT) reduce 493) ((T_ARRAY) reduce 493) ((T_AS) reduce 493) ((T_BREAK) reduce 493) ((T_CALLABLE) reduce 493) ((T_CASE) reduce 493) ((T_CATCH) reduce 493) ((T_CLASS) reduce 493) ((T_CLASS_C) reduce 493) ((T_CLONE) reduce 493) ((T_CONST) reduce 493) ((T_CONTINUE) reduce 493) ((T_DECLARE) reduce 493) ((T_DEFAULT) reduce 493) ((T_DIR) reduce 493) ((T_DO) reduce 493) ((T_ECHO) reduce 493) ((T_ELSE) reduce 493) ((T_ELSEIF) reduce 493) ((T_EMPTY) reduce 493) ((T_ENDDECLARE) reduce 493) ((T_ENDFOR) reduce 493) ((T_ENDFOREACH) reduce 493) ((T_ENDIF) reduce 493) ((T_ENDSWITCH) reduce 493) ((T_ENDWHILE) reduce 493) ((T_ENUM) reduce 493) ((T_EVAL) reduce 493) ((T_EXIT) reduce 493) ((T_EXTENDS) reduce 493) ((T_FILE) reduce 493) ((T_FINAL) reduce 493) ((T_FINALLY) reduce 493) ((T_FN) reduce 493) ((T_FOR) reduce 493) ((T_FOREACH) reduce 493) ((T_FUNCTION) reduce 493) ((T_FUNC_C) reduce 493) ((T_GLOBAL) reduce 493) ((T_GOTO) reduce 493) ((T_IF) reduce 493) ((T_IMPLEMENTS) reduce 493) ((T_INCLUDE) reduce 493) ((T_INCLUDE_ONCE) reduce 493) ((T_INSTANCEOF) reduce 493) ((T_INSTEADOF) reduce 493) ((T_INTERFACE) reduce 493) ((T_ISSET) reduce 493) ((T_LINE) reduce 493) ((T_LIST) reduce 493) ((T_LOGICAL_AND) reduce 493) ((T_LOGICAL_OR) reduce 493) ((T_LOGICAL_XOR) reduce 493) ((T_MATCH) reduce 493) ((T_METHOD_C) reduce 493) ((T_NAMESPACE) reduce 493) ((T_NEW) reduce 493) ((T_NS_C) reduce 493) ((T_PRINT) reduce 493) ((T_PRIVATE) reduce 493) ((T_PROPERTY_C) reduce 493) ((T_PROTECTED) reduce 493) ((T_PUBLIC) reduce 493) ((T_READONLY) reduce 493) ((T_REQUIRE) reduce 493) ((T_REQUIRE_ONCE) reduce 493) ((T_RETURN) reduce 493) ((T_STATIC) reduce 493) ((T_STRING) reduce 493) ((T_SWITCH) reduce 493) ((T_THROW) reduce 493) ((T_TRAIT) reduce 493) ((T_TRAIT_C) reduce 493) ((T_TRY) reduce 493) ((T_UNSET) reduce 493) ((T_USE) reduce 493) ((T_VAR) reduce 493) ((T_WHILE) reduce 493) ((T_YIELD) reduce 493)) 2997 ((("(") reduce 360) (("?") reduce 360) ((T_ABSTRACT) reduce 360) ((T_ARRAY) reduce 360) ((T_CALLABLE) reduce 360) ((T_CONST) reduce 360) ((T_FINAL) reduce 360) ((T_FUNCTION) reduce 360) ((T_NAME_FULLY_QUALIFIED) reduce 360) ((T_NAME_QUALIFIED) reduce 360) ((T_NAME_RELATIVE) reduce 360) ((T_PRIVATE) reduce 360) ((T_PRIVATE_SET) reduce 360) ((T_PROTECTED) reduce 360) ((T_PROTECTED_SET) reduce 360) ((T_PUBLIC) reduce 360) ((T_PUBLIC_SET) reduce 360) ((T_READONLY) reduce 360) ((T_STATIC) reduce 360) ((T_STRING) reduce 360) ((T_VARIABLE) reduce 360)) 2998 ((("(") reduce 354) (("?") reduce 354) ((T_ABSTRACT) reduce 354) ((T_ARRAY) reduce 354) ((T_CALLABLE) reduce 354) ((T_CONST) reduce 354) ((T_FINAL) reduce 354) ((T_FUNCTION) reduce 354) ((T_NAME_FULLY_QUALIFIED) reduce 354) ((T_NAME_QUALIFIED) reduce 354) ((T_NAME_RELATIVE) reduce 354) ((T_PRIVATE) reduce 354) ((T_PRIVATE_SET) reduce 354) ((T_PROTECTED) reduce 354) ((T_PROTECTED_SET) reduce 354) ((T_PUBLIC) reduce 354) ((T_PUBLIC_SET) reduce 354) ((T_READONLY) reduce 354) ((T_STATIC) reduce 354) ((T_STRING) reduce 354) ((T_VARIABLE) reduce 354)) 2999 ((("(") reduce 357) (("?") reduce 357) ((T_ABSTRACT) reduce 357) ((T_ARRAY) reduce 357) ((T_CALLABLE) reduce 357) ((T_CONST) reduce 357) ((T_FINAL) reduce 357) ((T_FUNCTION) reduce 357) ((T_NAME_FULLY_QUALIFIED) reduce 357) ((T_NAME_QUALIFIED) reduce 357) ((T_NAME_RELATIVE) reduce 357) ((T_PRIVATE) reduce 357) ((T_PRIVATE_SET) reduce 357) ((T_PROTECTED) reduce 357) ((T_PROTECTED_SET) reduce 357) ((T_PUBLIC) reduce 357) ((T_PUBLIC_SET) reduce 357) ((T_READONLY) reduce 357) ((T_STATIC) reduce 357) ((T_STRING) reduce 357) ((T_VARIABLE) reduce 357)) 3000 ((("(") reduce 353) (("?") reduce 353) ((T_ABSTRACT) reduce 353) ((T_ARRAY) reduce 353) ((T_CALLABLE) reduce 353) ((T_CONST) reduce 353) ((T_FINAL) reduce 353) ((T_FUNCTION) reduce 353) ((T_NAME_FULLY_QUALIFIED) reduce 353) ((T_NAME_QUALIFIED) reduce 353) ((T_NAME_RELATIVE) reduce 353) ((T_PRIVATE) reduce 353) ((T_PRIVATE_SET) reduce 353) ((T_PROTECTED) reduce 353) ((T_PROTECTED_SET) reduce 353) ((T_PUBLIC) reduce 353) ((T_PUBLIC_SET) reduce 353) ((T_READONLY) reduce 353) ((T_STATIC) reduce 353) ((T_STRING) reduce 353) ((T_VARIABLE) reduce 353)) 3001 ((("(") reduce 356) (("?") reduce 356) ((T_ABSTRACT) reduce 356) ((T_ARRAY) reduce 356) ((T_CALLABLE) reduce 356) ((T_CONST) reduce 356) ((T_FINAL) reduce 356) ((T_FUNCTION) reduce 356) ((T_NAME_FULLY_QUALIFIED) reduce 356) ((T_NAME_QUALIFIED) reduce 356) ((T_NAME_RELATIVE) reduce 356) ((T_PRIVATE) reduce 356) ((T_PRIVATE_SET) reduce 356) ((T_PROTECTED) reduce 356) ((T_PROTECTED_SET) reduce 356) ((T_PUBLIC) reduce 356) ((T_PUBLIC_SET) reduce 356) ((T_READONLY) reduce 356) ((T_STATIC) reduce 356) ((T_STRING) reduce 356) ((T_VARIABLE) reduce 356)) 3002 ((("(") reduce 352) (("?") reduce 352) ((T_ABSTRACT) reduce 352) ((T_ARRAY) reduce 352) ((T_CALLABLE) reduce 352) ((T_CONST) reduce 352) ((T_FINAL) reduce 352) ((T_FUNCTION) reduce 352) ((T_NAME_FULLY_QUALIFIED) reduce 352) ((T_NAME_QUALIFIED) reduce 352) ((T_NAME_RELATIVE) reduce 352) ((T_PRIVATE) reduce 352) ((T_PRIVATE_SET) reduce 352) ((T_PROTECTED) reduce 352) ((T_PROTECTED_SET) reduce 352) ((T_PUBLIC) reduce 352) ((T_PUBLIC_SET) reduce 352) ((T_READONLY) reduce 352) ((T_STATIC) reduce 352) ((T_STRING) reduce 352) ((T_VARIABLE) reduce 352)) 3003 ((("(") reduce 355) (("?") reduce 355) ((T_ABSTRACT) reduce 355) ((T_ARRAY) reduce 355) ((T_CALLABLE) reduce 355) ((T_CONST) reduce 355) ((T_FINAL) reduce 355) ((T_FUNCTION) reduce 355) ((T_NAME_FULLY_QUALIFIED) reduce 355) ((T_NAME_QUALIFIED) reduce 355) ((T_NAME_RELATIVE) reduce 355) ((T_PRIVATE) reduce 355) ((T_PRIVATE_SET) reduce 355) ((T_PROTECTED) reduce 355) ((T_PROTECTED_SET) reduce 355) ((T_PUBLIC) reduce 355) ((T_PUBLIC_SET) reduce 355) ((T_READONLY) reduce 355) ((T_STATIC) reduce 355) ((T_STRING) reduce 355) ((T_VARIABLE) reduce 355)) 3004 ((("(") reduce 361) (("?") reduce 361) ((T_ABSTRACT) reduce 361) ((T_ARRAY) reduce 361) ((T_CALLABLE) reduce 361) ((T_CONST) reduce 361) ((T_FINAL) reduce 361) ((T_FUNCTION) reduce 361) ((T_NAME_FULLY_QUALIFIED) reduce 361) ((T_NAME_QUALIFIED) reduce 361) ((T_NAME_RELATIVE) reduce 361) ((T_PRIVATE) reduce 361) ((T_PRIVATE_SET) reduce 361) ((T_PROTECTED) reduce 361) ((T_PROTECTED_SET) reduce 361) ((T_PUBLIC) reduce 361) ((T_PUBLIC_SET) reduce 361) ((T_READONLY) reduce 361) ((T_STATIC) reduce 361) ((T_STRING) reduce 361) ((T_VARIABLE) reduce 361)) 3005 ((("(") reduce 358) (("?") reduce 358) ((T_ABSTRACT) reduce 358) ((T_ARRAY) reduce 358) ((T_CALLABLE) reduce 358) ((T_CONST) reduce 358) ((T_FINAL) reduce 358) ((T_FUNCTION) reduce 358) ((T_NAME_FULLY_QUALIFIED) reduce 358) ((T_NAME_QUALIFIED) reduce 358) ((T_NAME_RELATIVE) reduce 358) ((T_PRIVATE) reduce 358) ((T_PRIVATE_SET) reduce 358) ((T_PROTECTED) reduce 358) ((T_PROTECTED_SET) reduce 358) ((T_PUBLIC) reduce 358) ((T_PUBLIC_SET) reduce 358) ((T_READONLY) reduce 358) ((T_STATIC) reduce 358) ((T_STRING) reduce 358) ((T_VARIABLE) reduce 358)) 3007 ((("(") reduce 345) (("?") reduce 345) ((T_ARRAY) reduce 345) ((T_CALLABLE) reduce 345) ((T_NAME_FULLY_QUALIFIED) reduce 345) ((T_NAME_QUALIFIED) reduce 345) ((T_NAME_RELATIVE) reduce 345) ((T_STRING) reduce 345) ((T_VARIABLE) reduce 345)) 3008 (((T_ABSTRACT) reduce 101) ((T_ATTRIBUTE) reduce 101) ((T_CASE) reduce 101) ((T_CONST) reduce 101) ((T_FINAL) reduce 101) ((T_FUNCTION) reduce 101) ((T_PRIVATE) reduce 101) ((T_PRIVATE_SET) reduce 101) ((T_PROTECTED) reduce 101) ((T_PROTECTED_SET) reduce 101) ((T_PUBLIC) reduce 101) ((T_PUBLIC_SET) reduce 101) ((T_READONLY) reduce 101) ((T_STATIC) reduce 101) ((T_VAR) reduce 101)) 3009 (((T_ABSTRACT) reduce 322) ((T_ATTRIBUTE) reduce 322) ((T_CASE) reduce 322) ((T_CONST) reduce 322) ((T_FINAL) reduce 322) ((T_FUNCTION) reduce 322) ((T_PRIVATE) reduce 322) ((T_PRIVATE_SET) reduce 322) ((T_PROTECTED) reduce 322) ((T_PROTECTED_SET) reduce 322) ((T_PUBLIC) reduce 322) ((T_PUBLIC_SET) reduce 322) ((T_READONLY) reduce 322) ((T_STATIC) reduce 322) ((T_USE) reduce 322) ((T_VAR) reduce 322) (("}") reduce 322)) 3010 (((T_ABSTRACT) shift) ((T_ATTRIBUTE) shift) ((T_CASE) shift) ((T_CONST) reduce 348) ((T_FINAL) shift) ((T_FUNCTION) reduce 346) ((T_PRIVATE) shift) ((T_PRIVATE_SET) shift) ((T_PROTECTED) shift) ((T_PROTECTED_SET) shift) ((T_PUBLIC) shift) ((T_PUBLIC_SET) shift) ((T_READONLY) shift) ((T_STATIC) shift) ((T_VAR) shift)) 3011 (((T_CONST) shift)) 3012 (((T_ABSTRACT) reduce 314) ((T_ATTRIBUTE) reduce 314) ((T_CASE) reduce 314) ((T_CONST) reduce 314) ((T_FINAL) reduce 314) ((T_FUNCTION) reduce 314) ((T_PRIVATE) reduce 314) ((T_PRIVATE_SET) reduce 314) ((T_PROTECTED) reduce 314) ((T_PROTECTED_SET) reduce 314) ((T_PUBLIC) reduce 314) ((T_PUBLIC_SET) reduce 314) ((T_READONLY) reduce 314) ((T_STATIC) reduce 314) ((T_USE) reduce 314) ((T_VAR) reduce 314) (("}") reduce 314)) 3013 (((T_ABSTRACT) reduce 321) ((T_ATTRIBUTE) reduce 321) ((T_CASE) reduce 321) ((T_CONST) reduce 321) ((T_FINAL) reduce 321) ((T_FUNCTION) reduce 321) ((T_PRIVATE) reduce 321) ((T_PRIVATE_SET) reduce 321) ((T_PROTECTED) reduce 321) ((T_PROTECTED_SET) reduce 321) ((T_PUBLIC) reduce 321) ((T_PUBLIC_SET) reduce 321) ((T_READONLY) reduce 321) ((T_STATIC) reduce 321) ((T_USE) reduce 321) ((T_VAR) reduce 321) (("}") reduce 321)) 3014 ((("(") reduce 350) (("?") reduce 350) ((T_ABSTRACT) reduce 350) ((T_ARRAY) reduce 350) ((T_CALLABLE) reduce 350) ((T_CONST) reduce 350) ((T_FINAL) reduce 350) ((T_FUNCTION) reduce 350) ((T_NAME_FULLY_QUALIFIED) reduce 350) ((T_NAME_QUALIFIED) reduce 350) ((T_NAME_RELATIVE) reduce 350) ((T_PRIVATE) reduce 350) ((T_PRIVATE_SET) reduce 350) ((T_PROTECTED) reduce 350) ((T_PROTECTED_SET) reduce 350) ((T_PUBLIC) reduce 350) ((T_PUBLIC_SET) reduce 350) ((T_READONLY) reduce 350) ((T_STATIC) reduce 350) ((T_STRING) reduce 350) ((T_VARIABLE) reduce 350)) 3015 (((T_FUNCTION) shift)) 3016 ((("(") reduce 344) (("?") reduce 344) ((T_ABSTRACT) shift) ((T_ARRAY) reduce 344) ((T_CALLABLE) reduce 344) ((T_CONST) reduce 349) ((T_FINAL) shift) ((T_FUNCTION) reduce 347) ((T_NAME_FULLY_QUALIFIED) reduce 344) ((T_NAME_QUALIFIED) reduce 344) ((T_NAME_RELATIVE) reduce 344) ((T_PRIVATE) shift) ((T_PRIVATE_SET) shift) ((T_PROTECTED) shift) ((T_PROTECTED_SET) shift) ((T_PUBLIC) shift) ((T_PUBLIC_SET) shift) ((T_READONLY) shift) ((T_STATIC) shift) ((T_STRING) reduce 344) ((T_VARIABLE) reduce 344)) 3017 ((("(") shift) (("?") shift) ((T_ARRAY) shift) ((T_CALLABLE) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_STRING) shift) ((T_VARIABLE) reduce 262)) 3018 ((("%") reduce 398) (("(") reduce 398) ((")") reduce 398) (("*") reduce 398) (("+") reduce 398) ((",") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 398) ((T_OBJECT_OPERATOR) reduce 398) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 398) ((T_PIPE) reduce 398) ((T_POW) reduce 398) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("[") reduce 398) (("^") reduce 398) (("|") reduce 398)) 3019 (((T_ARRAY) shift) ((T_CALLABLE) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_STRING) shift)) 3021 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 280) ((T_VARIABLE) reduce 280) (("|") reduce 280)) 3022 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_VARIABLE) reduce 281) (("|") reduce 281)) 3023 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_VARIABLE) reduce 94) (("|") reduce 94)) 3024 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_VARIABLE) reduce 93) (("|") reduce 93)) 3025 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_VARIABLE) reduce 95) (("|") reduce 95)) 3026 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_VARIABLE) reduce 92) (("|") reduce 92)) 3027 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_VARIABLE) reduce 279)) 3028 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_VARIABLE) reduce 282) (("|") reduce 282)) 3029 (((T_VARIABLE) shift)) 3030 (((T_VARIABLE) reduce 263)) 3031 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_VARIABLE) reduce 276) (("|") reduce 283)) 3032 (((T_VARIABLE) reduce 278) (("|") shift)) 3034 ((("(") shift) ((T_ARRAY) shift) ((T_CALLABLE) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_STRING) shift)) 3036 (((T_VARIABLE) reduce 280) (("|") reduce 280)) 3037 (((T_VARIABLE) reduce 281) (("|") reduce 281)) 3038 (((T_VARIABLE) reduce 94) (("|") reduce 94)) 3039 (((T_VARIABLE) reduce 93) (("|") reduce 93)) 3040 (((T_VARIABLE) reduce 95) (("|") reduce 95)) 3041 (((T_VARIABLE) reduce 92) (("|") reduce 92)) 3042 (((T_VARIABLE) reduce 282) (("|") reduce 282)) 3043 (((T_VARIABLE) reduce 283) (("|") reduce 283)) 3044 (((T_VARIABLE) reduce 285) (("|") reduce 285)) 3048 (((")") reduce 287) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 287)) 3049 (((T_VARIABLE) reduce 284) (("|") reduce 284)) 3051 (((")") reduce 288) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 288)) 3053 (((T_VARIABLE) reduce 286) (("|") reduce 286)) 3055 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 280) ((T_VARIABLE) reduce 280)) 3056 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_VARIABLE) reduce 281)) 3057 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_VARIABLE) reduce 94)) 3058 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_VARIABLE) reduce 93)) 3059 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_VARIABLE) reduce 95)) 3060 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_VARIABLE) reduce 92)) 3061 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_VARIABLE) reduce 282)) 3062 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 287) ((T_VARIABLE) reduce 287)) 3063 (((",") reduce 493) ((";") reduce 493) (("=") shift) (("{") reduce 493)) 3064 (((T_ABSTRACT) reduce 317) ((T_ATTRIBUTE) reduce 317) ((T_CASE) reduce 317) ((T_CONST) reduce 317) ((T_FINAL) reduce 317) ((T_FUNCTION) reduce 317) ((T_PRIVATE) reduce 317) ((T_PRIVATE_SET) reduce 317) ((T_PROTECTED) reduce 317) ((T_PROTECTED_SET) reduce 317) ((T_PUBLIC) reduce 317) ((T_PUBLIC_SET) reduce 317) ((T_READONLY) reduce 317) ((T_STATIC) reduce 317) ((T_USE) reduce 317) ((T_VAR) reduce 317) (("}") reduce 317)) 3065 (((",") reduce 363) ((";") reduce 363)) 3066 (((",") shift) ((";") shift)) 3068 (((T_ABSTRACT) reduce 316) ((T_ATTRIBUTE) reduce 316) ((T_CASE) reduce 316) ((T_CONST) reduce 316) ((T_FINAL) reduce 316) ((T_FUNCTION) reduce 316) ((T_PRIVATE) reduce 316) ((T_PRIVATE_SET) reduce 316) ((T_PROTECTED) reduce 316) ((T_PROTECTED_SET) reduce 316) ((T_PUBLIC) reduce 316) ((T_PUBLIC_SET) reduce 316) ((T_READONLY) reduce 316) ((T_STATIC) reduce 316) ((T_USE) reduce 316) ((T_VAR) reduce 316) (("}") reduce 316)) 3069 (((",") reduce 493) ((";") reduce 493) (("=") shift)) 3070 (((",") reduce 362) ((";") reduce 362)) 3072 (((",") reduce 364) ((";") reduce 364)) 3083 ((("%") reduce 540) (("*") reduce 540) (("+") reduce 540) ((",") reduce 540) (("-") reduce 540) ((".") reduce 540) (("/") reduce 540) ((";") reduce 540) (("<") reduce 540) ((">") reduce 540) (("?") reduce 540) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_BOOLEAN_AND) reduce 540) ((T_BOOLEAN_OR) reduce 540) ((T_COALESCE) reduce 540) ((T_INSTANCEOF) reduce 540) ((T_IS_EQUAL) reduce 540) ((T_IS_GREATER_OR_EQUAL) reduce 540) ((T_IS_IDENTICAL) reduce 540) ((T_IS_NOT_EQUAL) reduce 540) ((T_IS_NOT_IDENTICAL) reduce 540) ((T_IS_SMALLER_OR_EQUAL) reduce 540) ((T_LOGICAL_AND) reduce 540) ((T_LOGICAL_OR) reduce 540) ((T_LOGICAL_XOR) reduce 540) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 540) ((T_OBJECT_OPERATOR) reduce 540) ((T_PIPE) reduce 540) ((T_POW) reduce 540) ((T_SL) reduce 540) ((T_SPACESHIP) reduce 540) ((T_SR) reduce 540) (("[") reduce 540) (("^") reduce 540) (("|") reduce 540)) 3085 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) ((";") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_PIPE) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("|") reduce 521)) 3087 ((("%") reduce 534) (("*") reduce 534) (("+") reduce 534) ((",") reduce 534) (("-") reduce 534) ((".") reduce 534) (("/") reduce 534) ((";") reduce 534) (("<") reduce 534) ((">") reduce 534) (("?") reduce 534) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_BOOLEAN_AND) reduce 534) ((T_BOOLEAN_OR) reduce 534) ((T_COALESCE) reduce 534) ((T_INSTANCEOF) reduce 534) ((T_IS_EQUAL) reduce 534) ((T_IS_GREATER_OR_EQUAL) reduce 534) ((T_IS_IDENTICAL) reduce 534) ((T_IS_NOT_EQUAL) reduce 534) ((T_IS_NOT_IDENTICAL) reduce 534) ((T_IS_SMALLER_OR_EQUAL) reduce 534) ((T_LOGICAL_AND) reduce 534) ((T_LOGICAL_OR) reduce 534) ((T_LOGICAL_XOR) reduce 534) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PIPE) reduce 534) ((T_POW) reduce 534) ((T_SL) reduce 534) ((T_SPACESHIP) reduce 534) ((T_SR) reduce 534) (("[") reduce 534) (("^") reduce 534) (("|") reduce 534)) 3088 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) ((";") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_PIPE) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("^") reduce 524) (("|") reduce 524)) 3092 ((("%") reduce 517) (("(") shift) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((";") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_PIPE) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("^") reduce 517) (("|") reduce 517)) 3093 ((("%") reduce 533) (("*") reduce 533) (("+") reduce 533) ((",") reduce 533) (("-") reduce 533) ((".") reduce 533) (("/") reduce 533) ((";") reduce 533) (("<") reduce 533) ((">") reduce 533) (("?") reduce 533) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_BOOLEAN_AND) reduce 533) ((T_BOOLEAN_OR) reduce 533) ((T_COALESCE) reduce 533) ((T_INSTANCEOF) reduce 533) ((T_IS_EQUAL) reduce 533) ((T_IS_GREATER_OR_EQUAL) reduce 533) ((T_IS_IDENTICAL) reduce 533) ((T_IS_NOT_EQUAL) reduce 533) ((T_IS_NOT_IDENTICAL) reduce 533) ((T_IS_SMALLER_OR_EQUAL) reduce 533) ((T_LOGICAL_AND) reduce 533) ((T_LOGICAL_OR) reduce 533) ((T_LOGICAL_XOR) reduce 533) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PIPE) reduce 533) ((T_POW) reduce 533) ((T_SL) reduce 533) ((T_SPACESHIP) reduce 533) ((T_SR) reduce 533) (("[") reduce 533) (("^") reduce 533) (("|") reduce 533)) 3094 ((("%") reduce 537) (("*") reduce 537) (("+") reduce 537) ((",") reduce 537) (("-") reduce 537) ((".") reduce 537) (("/") reduce 537) ((";") reduce 537) (("<") reduce 537) ((">") reduce 537) (("?") reduce 537) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_BOOLEAN_AND) reduce 537) ((T_BOOLEAN_OR) reduce 537) ((T_COALESCE) reduce 537) ((T_INSTANCEOF) reduce 537) ((T_IS_EQUAL) reduce 537) ((T_IS_GREATER_OR_EQUAL) reduce 537) ((T_IS_IDENTICAL) reduce 537) ((T_IS_NOT_EQUAL) reduce 537) ((T_IS_NOT_IDENTICAL) reduce 537) ((T_IS_SMALLER_OR_EQUAL) reduce 537) ((T_LOGICAL_AND) reduce 537) ((T_LOGICAL_OR) reduce 537) ((T_LOGICAL_XOR) reduce 537) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_PIPE) reduce 537) ((T_POW) reduce 537) ((T_SL) reduce 537) ((T_SPACESHIP) reduce 537) ((T_SR) reduce 537) (("[") reduce 537) (("^") reduce 537) (("|") reduce 537)) 3100 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) ((";") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PIPE) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("|") reduce 532)) 3102 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) ((";") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_PIPE) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("^") reduce 523) (("|") reduce 523)) 3104 ((("%") reduce 536) (("*") reduce 536) (("+") reduce 536) ((",") reduce 536) (("-") reduce 536) ((".") reduce 536) (("/") reduce 536) ((";") reduce 536) (("<") reduce 536) ((">") reduce 536) (("?") reduce 536) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_BOOLEAN_AND) reduce 536) ((T_BOOLEAN_OR) reduce 536) ((T_COALESCE) reduce 536) ((T_INSTANCEOF) reduce 536) ((T_IS_EQUAL) reduce 536) ((T_IS_GREATER_OR_EQUAL) reduce 536) ((T_IS_IDENTICAL) reduce 536) ((T_IS_NOT_EQUAL) reduce 536) ((T_IS_NOT_IDENTICAL) reduce 536) ((T_IS_SMALLER_OR_EQUAL) reduce 536) ((T_LOGICAL_AND) reduce 536) ((T_LOGICAL_OR) reduce 536) ((T_LOGICAL_XOR) reduce 536) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 536) ((T_OBJECT_OPERATOR) reduce 536) ((T_PIPE) reduce 536) ((T_POW) reduce 536) ((T_SL) reduce 536) ((T_SPACESHIP) reduce 536) ((T_SR) reduce 536) (("[") reduce 536) (("^") reduce 536) (("|") reduce 536)) 3105 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("|") reduce 94)) 3106 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("|") reduce 93)) 3107 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("|") reduce 95)) 3109 ((("%") reduce 539) (("*") reduce 539) (("+") reduce 539) ((",") reduce 539) (("-") reduce 539) ((".") reduce 539) (("/") reduce 539) ((";") reduce 539) (("<") reduce 539) ((">") reduce 539) (("?") reduce 539) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_BOOLEAN_AND) reduce 539) ((T_BOOLEAN_OR) reduce 539) ((T_COALESCE) reduce 539) ((T_INSTANCEOF) reduce 539) ((T_IS_EQUAL) reduce 539) ((T_IS_GREATER_OR_EQUAL) reduce 539) ((T_IS_IDENTICAL) reduce 539) ((T_IS_NOT_EQUAL) reduce 539) ((T_IS_NOT_IDENTICAL) reduce 539) ((T_IS_SMALLER_OR_EQUAL) reduce 539) ((T_LOGICAL_AND) reduce 539) ((T_LOGICAL_OR) reduce 539) ((T_LOGICAL_XOR) reduce 539) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 539) ((T_OBJECT_OPERATOR) reduce 539) ((T_PIPE) reduce 539) ((T_POW) reduce 539) ((T_SL) reduce 539) ((T_SPACESHIP) reduce 539) ((T_SR) reduce 539) (("[") reduce 539) (("^") reduce 539) (("|") reduce 539)) 3112 ((("%") reduce 538) (("*") reduce 538) (("+") reduce 538) ((",") reduce 538) (("-") reduce 538) ((".") reduce 538) (("/") reduce 538) ((";") reduce 538) (("<") reduce 538) ((">") reduce 538) (("?") reduce 538) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_BOOLEAN_AND) reduce 538) ((T_BOOLEAN_OR) reduce 538) ((T_COALESCE) reduce 538) ((T_INSTANCEOF) reduce 538) ((T_IS_EQUAL) reduce 538) ((T_IS_GREATER_OR_EQUAL) reduce 538) ((T_IS_IDENTICAL) reduce 538) ((T_IS_NOT_EQUAL) reduce 538) ((T_IS_NOT_IDENTICAL) reduce 538) ((T_IS_SMALLER_OR_EQUAL) reduce 538) ((T_LOGICAL_AND) reduce 538) ((T_LOGICAL_OR) reduce 538) ((T_LOGICAL_XOR) reduce 538) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 538) ((T_OBJECT_OPERATOR) reduce 538) ((T_PIPE) reduce 538) ((T_POW) reduce 538) ((T_SL) reduce 538) ((T_SPACESHIP) reduce 538) ((T_SR) reduce 538) (("[") reduce 538) (("^") reduce 538) (("|") reduce 538)) 3118 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92)) 3121 ((("%") reduce 535) (("*") reduce 535) (("+") reduce 535) ((",") reduce 535) (("-") reduce 535) ((".") reduce 535) (("/") reduce 535) ((";") reduce 535) (("<") reduce 535) ((">") reduce 535) (("?") reduce 535) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_BOOLEAN_AND) reduce 535) ((T_BOOLEAN_OR) reduce 535) ((T_COALESCE) reduce 535) ((T_INSTANCEOF) reduce 535) ((T_IS_EQUAL) reduce 535) ((T_IS_GREATER_OR_EQUAL) reduce 535) ((T_IS_IDENTICAL) reduce 535) ((T_IS_NOT_EQUAL) reduce 535) ((T_IS_NOT_IDENTICAL) reduce 535) ((T_IS_SMALLER_OR_EQUAL) reduce 535) ((T_LOGICAL_AND) reduce 535) ((T_LOGICAL_OR) reduce 535) ((T_LOGICAL_XOR) reduce 535) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PIPE) reduce 535) ((T_POW) reduce 535) ((T_SL) reduce 535) ((T_SPACESHIP) reduce 535) ((T_SR) reduce 535) (("[") reduce 535) (("^") reduce 535) (("|") reduce 535)) 3123 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) ((";") reduce 568) (("<") reduce 568) (("=") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AND_EQUAL) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_COALESCE_EQUAL) reduce 568) ((T_CONCAT_EQUAL) reduce 568) ((T_DEC) reduce 568) ((T_DIV_EQUAL) reduce 568) ((T_INC) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_MINUS_EQUAL) reduce 568) ((T_MOD_EQUAL) reduce 568) ((T_MUL_EQUAL) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_OR_EQUAL) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_PLUS_EQUAL) reduce 568) ((T_POW) reduce 568) ((T_POW_EQUAL) reduce 568) ((T_SL) reduce 568) ((T_SL_EQUAL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) ((T_SR_EQUAL) reduce 568) ((T_XOR_EQUAL) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 3124 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 479) (("(") shift) (("*") reduce 479) (("+") shift) ((",") reduce 479) (("-") shift) ((".") reduce 479) (("/") reduce 479) ((";") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 479) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 479) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PIPE) reduce 479) ((T_POW) reduce 479) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 479) (("`") shift) (("|") reduce 479) (("~") shift)) 3131 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) ((",") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) ((";") reduce 564) (("<") reduce 564) (("=") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AND_EQUAL) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_COALESCE_EQUAL) reduce 564) ((T_CONCAT_EQUAL) reduce 564) ((T_DEC) reduce 564) ((T_DIV_EQUAL) reduce 564) ((T_INC) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_MINUS_EQUAL) reduce 564) ((T_MOD_EQUAL) reduce 564) ((T_MUL_EQUAL) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_OR_EQUAL) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_PLUS_EQUAL) reduce 564) ((T_POW) reduce 564) ((T_POW_EQUAL) reduce 564) ((T_SL) reduce 564) ((T_SL_EQUAL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) ((T_SR_EQUAL) reduce 564) ((T_XOR_EQUAL) reduce 564) (("[") reduce 564) (("^") reduce 564) (("|") reduce 564)) 3132 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) ((";") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PIPE) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 551) (("^") reduce 530) (("|") reduce 530)) 3134 ((("%") reduce 529) (("*") reduce 529) (("+") reduce 529) ((",") reduce 529) (("-") reduce 529) ((".") reduce 529) (("/") reduce 529) ((";") reduce 529) (("<") reduce 529) ((">") reduce 529) (("?") reduce 529) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_BOOLEAN_AND) reduce 529) ((T_BOOLEAN_OR) reduce 529) ((T_COALESCE) reduce 529) ((T_INSTANCEOF) reduce 529) ((T_IS_EQUAL) reduce 529) ((T_IS_GREATER_OR_EQUAL) reduce 529) ((T_IS_IDENTICAL) reduce 529) ((T_IS_NOT_EQUAL) reduce 529) ((T_IS_NOT_IDENTICAL) reduce 529) ((T_IS_SMALLER_OR_EQUAL) reduce 529) ((T_LOGICAL_AND) reduce 529) ((T_LOGICAL_OR) reduce 529) ((T_LOGICAL_XOR) reduce 529) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PIPE) reduce 529) ((T_POW) reduce 529) ((T_SL) reduce 529) ((T_SPACESHIP) reduce 529) ((T_SR) reduce 529) (("[") reduce 554) (("^") reduce 529) (("|") reduce 529)) 3135 ((("%") reduce 528) (("(") reduce 557) (("*") reduce 528) (("+") reduce 528) ((",") reduce 528) (("-") reduce 528) ((".") reduce 528) (("/") reduce 528) ((";") reduce 528) (("<") reduce 528) ((">") reduce 528) (("?") reduce 528) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_BOOLEAN_AND) reduce 528) ((T_BOOLEAN_OR) reduce 528) ((T_COALESCE) reduce 528) ((T_INSTANCEOF) reduce 528) ((T_IS_EQUAL) reduce 528) ((T_IS_GREATER_OR_EQUAL) reduce 528) ((T_IS_IDENTICAL) reduce 528) ((T_IS_NOT_EQUAL) reduce 528) ((T_IS_NOT_IDENTICAL) reduce 528) ((T_IS_SMALLER_OR_EQUAL) reduce 528) ((T_LOGICAL_AND) reduce 528) ((T_LOGICAL_OR) reduce 528) ((T_LOGICAL_XOR) reduce 528) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PIPE) reduce 528) ((T_POW) reduce 528) ((T_SL) reduce 528) ((T_SPACESHIP) reduce 528) ((T_SR) reduce 528) (("[") reduce 550) (("^") reduce 528) (("|") reduce 528)) 3136 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 493) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 493) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3139 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) ((",") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) ((";") reduce 563) (("<") reduce 563) (("=") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AND_EQUAL) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_COALESCE_EQUAL) reduce 563) ((T_CONCAT_EQUAL) reduce 563) ((T_DEC) reduce 563) ((T_DIV_EQUAL) reduce 563) ((T_INC) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_MINUS_EQUAL) reduce 563) ((T_MOD_EQUAL) reduce 563) ((T_MUL_EQUAL) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_OR_EQUAL) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_PLUS_EQUAL) reduce 563) ((T_POW) reduce 563) ((T_POW_EQUAL) reduce 563) ((T_SL) reduce 563) ((T_SL_EQUAL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) ((T_SR_EQUAL) reduce 563) ((T_XOR_EQUAL) reduce 563) (("[") reduce 563) (("^") reduce 563) (("|") reduce 563)) 3140 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) ((",") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) ((";") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_PIPE) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("^") reduce 484) (("|") reduce 484)) 3141 ((("%") reduce 466) (("*") reduce 466) (("+") reduce 466) ((",") reduce 466) (("-") reduce 466) ((".") reduce 466) (("/") reduce 466) ((";") reduce 466) (("<") reduce 466) ((">") reduce 466) (("?") reduce 466) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_BOOLEAN_AND) reduce 466) ((T_BOOLEAN_OR) reduce 466) ((T_COALESCE) reduce 466) ((T_INSTANCEOF) reduce 466) ((T_IS_EQUAL) reduce 466) ((T_IS_GREATER_OR_EQUAL) reduce 466) ((T_IS_IDENTICAL) reduce 466) ((T_IS_NOT_EQUAL) reduce 466) ((T_IS_NOT_IDENTICAL) reduce 466) ((T_IS_SMALLER_OR_EQUAL) reduce 466) ((T_LOGICAL_AND) reduce 466) ((T_LOGICAL_OR) reduce 466) ((T_LOGICAL_XOR) reduce 466) ((T_PIPE) reduce 466) ((T_POW) reduce 466) ((T_SL) reduce 466) ((T_SPACESHIP) reduce 466) ((T_SR) reduce 466) (("^") reduce 466) (("|") reduce 466)) 3142 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) ((";") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_PIPE) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("^") reduce 488) (("|") reduce 488)) 3143 ((("%") reduce 531) (("(") shift) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) ((";") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("|") reduce 531)) 3144 ((("%") reduce 461) (("(") reduce 558) (("*") reduce 461) (("+") reduce 461) ((",") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) ((";") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PIPE) reduce 461) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") reduce 552) (("^") reduce 461) (("|") reduce 461)) 3145 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) ((",") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) ((";") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_PIPE) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 3146 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) ((",") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) ((";") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_PIPE) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("|") reduce 476)) 3147 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) ((";") reduce 559) (("<") reduce 559) (("=") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AND_EQUAL) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_COALESCE_EQUAL) reduce 559) ((T_CONCAT_EQUAL) reduce 559) ((T_DEC) reduce 559) ((T_DIV_EQUAL) reduce 559) ((T_INC) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_MINUS_EQUAL) reduce 559) ((T_MOD_EQUAL) reduce 559) ((T_MUL_EQUAL) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_OR_EQUAL) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_PLUS_EQUAL) reduce 559) ((T_POW) reduce 559) ((T_POW_EQUAL) reduce 559) ((T_SL) reduce 559) ((T_SL_EQUAL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) ((T_SR_EQUAL) reduce 559) ((T_XOR_EQUAL) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 3148 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((";") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 3149 ((("%") reduce 403) (("*") reduce 403) (("+") reduce 403) ((",") reduce 403) (("-") reduce 403) ((".") reduce 403) (("/") reduce 403) ((";") reduce 403) (("<") reduce 403) (("=") shift) ((">") reduce 403) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_INC) shift) ((T_INSTANCEOF) reduce 403) ((T_IS_EQUAL) reduce 403) ((T_IS_GREATER_OR_EQUAL) reduce 403) ((T_IS_IDENTICAL) reduce 403) ((T_IS_NOT_EQUAL) reduce 403) ((T_IS_NOT_IDENTICAL) reduce 403) ((T_IS_SMALLER_OR_EQUAL) reduce 403) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 403) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 403) ((T_POW_EQUAL) shift) ((T_SL) reduce 403) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 403) ((T_SR) reduce 403) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 548) (("^") reduce 403) (("|") reduce 403)) 3152 ((("%") reduce 448) (("*") reduce 448) (("+") reduce 448) ((",") reduce 448) (("-") reduce 448) ((".") reduce 448) (("/") reduce 448) ((";") reduce 448) (("<") reduce 448) ((">") reduce 448) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_BOOLEAN_AND) reduce 448) ((T_BOOLEAN_OR) reduce 448) ((T_COALESCE) reduce 448) ((T_INSTANCEOF) reduce 448) ((T_IS_EQUAL) reduce 448) ((T_IS_GREATER_OR_EQUAL) reduce 448) ((T_IS_IDENTICAL) reduce 448) ((T_IS_NOT_EQUAL) reduce 448) ((T_IS_NOT_IDENTICAL) reduce 448) ((T_IS_SMALLER_OR_EQUAL) reduce 448) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_PIPE) reduce 448) ((T_POW) shift) ((T_SL) reduce 448) ((T_SPACESHIP) reduce 448) ((T_SR) reduce 448) (("^") reduce 448) (("|") reduce 448)) 3184 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 432) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 432) (("<") shift) ((">") shift) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 432)) 3185 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 435) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 435) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("|") reduce 435)) 3186 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 444) (("-") shift) ((".") reduce 444) (("/") shift) ((";") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_PIPE) reduce 444) ((T_POW) shift) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("|") reduce 444)) 3187 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 458) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 458) (("<") shift) ((">") shift) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 458) (("|") reduce 458)) 3188 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 443) (("-") shift) ((".") reduce 443) (("/") shift) ((";") reduce 443) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_PIPE) reduce 443) ((T_POW) shift) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("^") reduce 443) (("|") reduce 443)) 3189 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) ((",") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) ((";") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_PIPE) reduce 440) ((T_POW) shift) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("|") reduce 440)) 3190 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 453) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 453) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_PIPE) reduce 453) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 453) ((T_SR) shift) (("^") reduce 453) (("|") reduce 453)) 3191 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 431) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 431) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3192 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 429) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 429) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3193 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 430) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 430) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3194 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 455) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 455) (("<") shift) ((">") shift) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 455) ((T_SR) shift) (("^") reduce 455) (("|") reduce 455)) 3195 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 450) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 450) (("<") shift) ((">") shift) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 450) (("|") reduce 450)) 3196 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 452) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 452) (("<") shift) ((">") shift) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 452) (("|") reduce 452)) 3197 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 449) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 449) (("<") shift) ((">") shift) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 449) (("|") reduce 449)) 3198 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 457) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 457) (("<") shift) ((">") shift) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 457) ((T_SR) shift) (("^") reduce 457) (("|") reduce 457)) 3199 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 451) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 451) (("<") shift) ((">") shift) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 451) (("|") reduce 451)) 3202 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 3203 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 3204 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 3205 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((";") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 3206 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 3207 ((("%") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) ((";") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 3208 ((("%") reduce 511) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((";") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("|") reduce 511)) 3209 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) ((";") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PIPE) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 3210 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((";") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 3211 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((";") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("|") reduce 512)) 3212 ((("%") reduce 573) (("*") reduce 573) (("+") reduce 573) ((",") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) ((";") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("|") reduce 573)) 3218 ((("%") reduce 574) (("*") reduce 574) (("+") reduce 574) ((",") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) ((";") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("|") reduce 574)) 3219 ((("%") reduce 578) (("*") reduce 578) (("+") reduce 578) ((",") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) ((";") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("|") reduce 578)) 3220 ((("%") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) ((";") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 3221 ((("%") reduce 575) (("*") reduce 575) (("+") reduce 575) ((",") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) ((";") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("|") reduce 575)) 3222 ((("%") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) ((";") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 3225 ((("%") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) ((";") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 3226 ((("%") reduce 576) (("*") reduce 576) (("+") reduce 576) ((",") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) ((";") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("|") reduce 576)) 3228 ((("%") reduce 577) (("*") reduce 577) (("+") reduce 577) ((",") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) ((";") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("|") reduce 577)) 3230 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((";") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("|") reduce 513)) 3231 ((("%") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) ((";") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 3234 ((("%") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) ((";") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 3235 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 465) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 465) (("<") shift) ((">") shift) (("?") reduce 465) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3236 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 427) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 427) (("<") shift) ((">") shift) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3237 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 428) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 428) (("<") shift) ((">") shift) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3238 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 433) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 433) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("|") reduce 433)) 3239 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 434) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 434) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("|") reduce 434)) 3243 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 463) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 463) (("<") shift) ((">") shift) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3244 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 464) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 464) (("<") shift) ((">") shift) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3245 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 456) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 456) (("<") shift) ((">") shift) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 456) ((T_SR) shift) (("^") reduce 456) (("|") reduce 456)) 3246 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 454) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 454) (("<") shift) ((">") shift) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 454) ((T_SR) shift) (("^") reduce 454) (("|") reduce 454)) 3247 ((("%") reduce 441) (("*") reduce 441) (("+") reduce 441) ((",") reduce 441) (("-") reduce 441) ((".") reduce 441) (("/") reduce 441) ((";") reduce 441) (("<") reduce 441) ((">") reduce 441) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 441) ((T_IS_GREATER_OR_EQUAL) reduce 441) ((T_IS_IDENTICAL) reduce 441) ((T_IS_NOT_EQUAL) reduce 441) ((T_IS_NOT_IDENTICAL) reduce 441) ((T_IS_SMALLER_OR_EQUAL) reduce 441) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_PIPE) reduce 441) ((T_POW) shift) ((T_SL) reduce 441) ((T_SPACESHIP) reduce 441) ((T_SR) reduce 441) (("^") reduce 441) (("|") reduce 441)) 3248 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 436) (("-") shift) ((".") reduce 436) (("/") shift) ((";") reduce 436) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_PIPE) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 436) ((T_SR) shift) (("^") reduce 436) (("|") reduce 436)) 3249 ((("%") shift) (("*") shift) (("+") reduce 438) ((",") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") shift) ((";") reduce 438) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_PIPE) reduce 438) ((T_POW) shift) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("|") reduce 438)) 3250 ((("%") shift) (("*") shift) (("+") reduce 437) ((",") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") shift) ((";") reduce 437) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_PIPE) reduce 437) ((T_POW) shift) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("|") reduce 437)) 3251 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) ((",") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) ((";") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_PIPE) reduce 439) ((T_POW) shift) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("|") reduce 439)) 3252 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) ((",") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) ((";") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_PIPE) reduce 442) ((T_POW) shift) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("|") reduce 442)) 3254 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) ((",") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) ((";") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_PIPE) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("|") reduce 74)) 3255 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) ((",") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) ((";") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_PIPE) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("|") reduce 52)) 3256 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) ((",") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) ((";") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_PIPE) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("|") reduce 27)) 3257 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) ((",") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) ((";") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_PIPE) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("|") reduce 51)) 3258 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) ((",") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) ((";") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_PIPE) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("|") reduce 53)) 3259 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) ((",") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) ((";") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_PIPE) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("|") reduce 49)) 3260 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) ((",") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) ((";") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_PIPE) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("|") reduce 29)) 3261 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) ((",") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) ((";") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_PIPE) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("|") reduce 59)) 3262 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) ((",") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) ((";") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_PIPE) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("|") reduce 60)) 3263 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) ((",") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) ((";") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_PIPE) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("|") reduce 11)) 3264 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) ((",") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) ((";") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_PIPE) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("|") reduce 42)) 3265 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) ((",") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) ((";") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_PIPE) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("|") reduce 39)) 3266 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) ((",") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) ((";") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_PIPE) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("|") reduce 25)) 3267 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) ((",") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) ((";") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_PIPE) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("|") reduce 50)) 3268 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) ((",") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) ((";") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_PIPE) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("|") reduce 66)) 3269 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) ((",") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) ((";") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_PIPE) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("|") reduce 18)) 3270 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) ((",") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) ((";") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_PIPE) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("|") reduce 17)) 3271 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) ((",") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) ((";") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_PIPE) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("|") reduce 15)) 3272 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) ((",") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) ((";") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_PIPE) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("|") reduce 14)) 3273 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) ((",") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) ((";") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_PIPE) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("|") reduce 38)) 3274 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) ((",") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) ((";") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_PIPE) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("|") reduce 26)) 3275 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) ((",") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) ((";") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_PIPE) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("|") reduce 22)) 3276 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) ((",") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) ((";") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_PIPE) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("|") reduce 24)) 3277 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) ((",") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) ((";") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_PIPE) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("|") reduce 16)) 3278 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) ((",") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) ((";") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_PIPE) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("|") reduce 48)) 3279 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) ((",") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) ((";") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_PIPE) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("|") reduce 20)) 3280 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) ((",") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) ((";") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_PIPE) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("|") reduce 70)) 3281 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) ((",") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) ((";") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_PIPE) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("|") reduce 3)) 3282 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) ((",") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) ((";") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_PIPE) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("|") reduce 12)) 3283 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) ((",") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) ((";") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_PIPE) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("|") reduce 54)) 3284 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) ((",") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) ((";") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_PIPE) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("|") reduce 65)) 3285 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) ((",") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) ((";") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_PIPE) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("|") reduce 75)) 3286 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) ((",") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) ((";") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_PIPE) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("|") reduce 30)) 3287 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) ((",") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) ((";") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_PIPE) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("|") reduce 68)) 3288 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) ((",") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) ((";") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_PIPE) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("|") reduce 21)) 3289 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) ((",") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) ((";") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_PIPE) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("|") reduce 23)) 3290 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) ((",") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) ((";") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_PIPE) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("|") reduce 41)) 3291 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) ((",") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) ((";") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_PIPE) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("|") reduce 62)) 3292 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) ((",") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) ((";") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_PIPE) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("|") reduce 34)) 3293 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) ((",") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) ((";") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_PIPE) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("|") reduce 40)) 3294 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) ((",") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) ((";") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_PIPE) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("|") reduce 13)) 3295 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) ((",") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) ((";") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_PIPE) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("|") reduce 55)) 3296 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) ((",") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) ((";") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_PIPE) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("|") reduce 1)) 3297 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) ((",") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) ((";") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_PIPE) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("|") reduce 2)) 3298 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) ((",") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) ((";") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_PIPE) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("|") reduce 9)) 3299 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) ((",") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) ((";") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_PIPE) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("|") reduce 33)) 3300 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) ((",") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) ((";") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_PIPE) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("|") reduce 58)) 3301 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) ((",") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) ((";") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_PIPE) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("|") reduce 37)) 3302 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) ((",") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) ((";") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_PIPE) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("|") reduce 64)) 3303 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) ((",") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) ((";") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_PIPE) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("|") reduce 46)) 3304 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) ((",") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) ((";") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_PIPE) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("|") reduce 8)) 3305 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) ((",") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) ((";") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_PIPE) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("|") reduce 6)) 3306 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) ((",") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) ((";") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_PIPE) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("|") reduce 7)) 3307 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) ((",") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) ((";") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_PIPE) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("|") reduce 69)) 3308 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) ((",") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) ((";") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_PIPE) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("|") reduce 63)) 3309 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) ((",") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) ((";") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_PIPE) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("|") reduce 56)) 3310 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) ((",") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) ((";") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_PIPE) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("|") reduce 10)) 3311 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) ((",") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) ((";") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_PIPE) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("|") reduce 67)) 3312 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) ((",") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) ((";") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_PIPE) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("|") reduce 44)) 3313 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) ((",") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) ((";") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_PIPE) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("|") reduce 76)) 3314 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) ((",") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) ((";") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_PIPE) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("|") reduce 71)) 3315 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) ((",") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) ((";") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_PIPE) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("|") reduce 77)) 3316 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) ((",") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) ((";") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_PIPE) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("|") reduce 78)) 3317 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) ((",") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) ((";") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_PIPE) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("|") reduce 79)) 3318 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) ((",") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) ((";") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_PIPE) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("|") reduce 4)) 3319 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) ((",") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) ((";") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_PIPE) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("|") reduce 5)) 3320 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) ((",") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) ((";") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_PIPE) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("|") reduce 43)) 3321 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) ((",") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) ((";") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_PIPE) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("|") reduce 73)) 3322 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) ((",") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) ((";") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_PIPE) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("|") reduce 82)) 3323 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) ((",") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) ((";") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_PIPE) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("|") reduce 47)) 3324 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) ((",") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) ((";") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_PIPE) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("|") reduce 31)) 3325 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) ((",") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) ((";") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_PIPE) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("|") reduce 57)) 3326 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) ((",") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) ((";") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_PIPE) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("|") reduce 61)) 3327 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) ((",") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) ((";") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_PIPE) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("|") reduce 28)) 3328 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) ((",") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) ((";") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_PIPE) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("|") reduce 36)) 3329 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) ((",") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) ((";") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_PIPE) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("|") reduce 32)) 3330 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) ((",") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) ((";") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_PIPE) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("|") reduce 35)) 3331 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) ((",") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) ((";") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_PIPE) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("|") reduce 19)) 3332 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) ((",") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) ((";") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_PIPE) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("|") reduce 45)) 3333 ((("%") reduce 542) (("(") reduce 579) (("*") reduce 542) (("+") reduce 542) ((",") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) ((";") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PIPE) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 3335 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) ((",") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) ((";") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_PIPE) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("|") reduce 72)) 3336 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) ((",") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) ((";") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_PIPE) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("|") reduce 83)) 3337 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) ((",") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) ((";") reduce 572) (("<") reduce 572) (("=") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AND_EQUAL) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_COALESCE_EQUAL) reduce 572) ((T_CONCAT_EQUAL) reduce 572) ((T_DEC) reduce 572) ((T_DIV_EQUAL) reduce 572) ((T_INC) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_MINUS_EQUAL) reduce 572) ((T_MOD_EQUAL) reduce 572) ((T_MUL_EQUAL) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_OR_EQUAL) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_PLUS_EQUAL) reduce 572) ((T_POW) reduce 572) ((T_POW_EQUAL) reduce 572) ((T_SL) reduce 572) ((T_SL_EQUAL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) ((T_SR_EQUAL) reduce 572) ((T_XOR_EQUAL) reduce 572) (("[") reduce 572) (("^") reduce 572) (("|") reduce 572)) 3340 ((("%") reduce 544) (("(") reduce 580) (("*") reduce 544) (("+") reduce 544) ((",") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) ((";") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PIPE) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 3342 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((";") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("|") reduce 507)) 3343 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((";") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 3347 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) ((";") reduce 292) (("<") reduce 292) (("=") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AND_EQUAL) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_COALESCE_EQUAL) reduce 292) ((T_CONCAT_EQUAL) reduce 292) ((T_DEC) reduce 292) ((T_DIV_EQUAL) reduce 292) ((T_INC) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_MINUS_EQUAL) reduce 292) ((T_MOD_EQUAL) reduce 292) ((T_MUL_EQUAL) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_OR_EQUAL) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_PLUS_EQUAL) reduce 292) ((T_POW) reduce 292) ((T_POW_EQUAL) reduce 292) ((T_SL) reduce 292) ((T_SL_EQUAL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) ((T_SR_EQUAL) reduce 292) ((T_XOR_EQUAL) reduce 292) (("[") reduce 292) (("^") reduce 292) (("|") reduce 292)) 3348 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) ((";") reduce 293) (("<") reduce 293) (("=") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AND_EQUAL) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_COALESCE_EQUAL) reduce 293) ((T_CONCAT_EQUAL) reduce 293) ((T_DEC) reduce 293) ((T_DIV_EQUAL) reduce 293) ((T_INC) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_MINUS_EQUAL) reduce 293) ((T_MOD_EQUAL) reduce 293) ((T_MUL_EQUAL) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_OR_EQUAL) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_PLUS_EQUAL) reduce 293) ((T_POW) reduce 293) ((T_POW_EQUAL) reduce 293) ((T_SL) reduce 293) ((T_SL_EQUAL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) ((T_SR_EQUAL) reduce 293) ((T_XOR_EQUAL) reduce 293) (("[") reduce 293) (("^") reduce 293) (("|") reduce 293)) 3353 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) ((",") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) ((";") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_PIPE) reduce 425) ((T_POW) reduce 425) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 3355 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) ((",") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) ((";") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_PIPE) reduce 423) ((T_POW) reduce 423) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 3365 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 419) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 419) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3366 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 421) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 421) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3367 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 420) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 420) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3368 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 413) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 413) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3369 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 410) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 410) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3370 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 418) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 418) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3371 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 412) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 412) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3372 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 416) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 416) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3373 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 411) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 411) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3374 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 414) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 414) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3375 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 415) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 415) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3376 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 422) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 422) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3377 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 417) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 417) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3379 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 406) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 406) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3382 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) ((";") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 3385 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) ((",") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) ((";") reduce 564) (("<") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_POW) reduce 564) ((T_SL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) (("[") reduce 564) (("^") reduce 564) (("|") reduce 564)) 3387 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) ((",") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) ((";") reduce 563) (("<") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_POW) reduce 563) ((T_SL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) (("[") reduce 563) (("^") reduce 563) (("|") reduce 563)) 3389 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) ((";") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 3390 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((";") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 3391 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) ((",") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) ((";") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("[") reduce 548) (("^") reduce 407) (("|") reduce 407)) 3395 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) ((",") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) ((";") reduce 572) (("<") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_POW) reduce 572) ((T_SL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) (("[") reduce 572) (("^") reduce 572) (("|") reduce 572)) 3397 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((";") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("|") reduce 507)) 3398 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((";") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 3402 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) ((";") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("[") reduce 292) (("^") reduce 292) (("|") reduce 292)) 3403 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) ((";") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("[") reduce 293) (("^") reduce 293) (("|") reduce 293)) 3404 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) ((";") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 3407 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) ((",") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) ((";") reduce 571) (("<") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_POW) reduce 571) ((T_SL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) (("[") reduce 571) (("^") reduce 571) (("|") reduce 571)) 3408 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((";") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("|") reduce 506)) 3409 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((";") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("|") reduce 508)) 3414 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) ((";") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 3415 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) ((";") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 3416 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((";") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 3417 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) ((";") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 3420 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) ((";") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 3421 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((";") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 3422 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((";") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 3423 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) ((",") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) ((";") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("[") reduce 562) (("^") reduce 562) (("|") reduce 562)) 3424 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((";") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 3425 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) ((";") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 3428 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) ((";") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 3429 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) ((";") reduce 504) (("<") reduce 504) (("=") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AND_EQUAL) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_COALESCE_EQUAL) reduce 504) ((T_CONCAT_EQUAL) reduce 504) ((T_DEC) reduce 504) ((T_DIV_EQUAL) reduce 504) ((T_INC) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_MINUS_EQUAL) reduce 504) ((T_MOD_EQUAL) reduce 504) ((T_MUL_EQUAL) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_OR_EQUAL) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_PLUS_EQUAL) reduce 504) ((T_POW) reduce 504) ((T_POW_EQUAL) reduce 504) ((T_SL) reduce 504) ((T_SL_EQUAL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) ((T_SR_EQUAL) reduce 504) ((T_XOR_EQUAL) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 3440 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) ((",") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) ((";") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PIPE) reduce 494) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 3441 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) ((";") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_PIPE) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("^") reduce 489) (("|") reduce 489)) 3451 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 494) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 494) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3452 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) ((";") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_PIPE) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("^") reduce 490) (("|") reduce 490)) 3453 (((",") reduce 365) ((";") reduce 365)) 3455 ((("%") reduce 541) (("(") reduce 579) (("*") reduce 541) (("+") reduce 541) ((",") reduce 541) (("-") reduce 541) ((".") reduce 541) (("/") reduce 541) ((";") reduce 541) (("<") reduce 541) ((">") reduce 541) (("?") reduce 541) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_BOOLEAN_AND) reduce 541) ((T_BOOLEAN_OR) reduce 541) ((T_COALESCE) reduce 541) ((T_INSTANCEOF) reduce 541) ((T_IS_EQUAL) reduce 541) ((T_IS_GREATER_OR_EQUAL) reduce 541) ((T_IS_IDENTICAL) reduce 541) ((T_IS_NOT_EQUAL) reduce 541) ((T_IS_NOT_IDENTICAL) reduce 541) ((T_IS_SMALLER_OR_EQUAL) reduce 541) ((T_LOGICAL_AND) reduce 541) ((T_LOGICAL_OR) reduce 541) ((T_LOGICAL_XOR) reduce 541) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 541) ((T_OBJECT_OPERATOR) reduce 541) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 541) ((T_PIPE) reduce 541) ((T_POW) reduce 541) ((T_SL) reduce 541) ((T_SPACESHIP) reduce 541) ((T_SR) reduce 541) (("[") reduce 541) (("^") reduce 541) (("|") reduce 541)) 3457 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) ((",") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) ((";") reduce 571) (("<") reduce 571) (("=") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AND_EQUAL) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_COALESCE_EQUAL) reduce 571) ((T_CONCAT_EQUAL) reduce 571) ((T_DEC) reduce 571) ((T_DIV_EQUAL) reduce 571) ((T_INC) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_MINUS_EQUAL) reduce 571) ((T_MOD_EQUAL) reduce 571) ((T_MUL_EQUAL) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_OR_EQUAL) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_PLUS_EQUAL) reduce 571) ((T_POW) reduce 571) ((T_POW_EQUAL) reduce 571) ((T_SL) reduce 571) ((T_SL_EQUAL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) ((T_SR_EQUAL) reduce 571) ((T_XOR_EQUAL) reduce 571) (("[") reduce 571) (("^") reduce 571) (("|") reduce 571)) 3460 ((("%") reduce 543) (("(") reduce 580) (("*") reduce 543) (("+") reduce 543) ((",") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) ((";") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PIPE) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 3461 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((";") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("|") reduce 506)) 3462 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((";") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("|") reduce 508)) 3464 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) ((",") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) ((";") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_PIPE) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("^") reduce 485) (("|") reduce 485)) 3465 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) ((";") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_PIPE) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("^") reduce 487) (("|") reduce 487)) 3470 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) ((";") reduce 560) (("<") reduce 560) (("=") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AND_EQUAL) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_COALESCE_EQUAL) reduce 560) ((T_CONCAT_EQUAL) reduce 560) ((T_DEC) reduce 560) ((T_DIV_EQUAL) reduce 560) ((T_INC) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_MINUS_EQUAL) reduce 560) ((T_MOD_EQUAL) reduce 560) ((T_MUL_EQUAL) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_OR_EQUAL) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_PLUS_EQUAL) reduce 560) ((T_POW) reduce 560) ((T_POW_EQUAL) reduce 560) ((T_SL) reduce 560) ((T_SL_EQUAL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) ((T_SR_EQUAL) reduce 560) ((T_XOR_EQUAL) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 3471 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) ((";") reduce 582) (("<") reduce 582) (("=") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AND_EQUAL) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_COALESCE_EQUAL) reduce 582) ((T_CONCAT_EQUAL) reduce 582) ((T_DEC) reduce 582) ((T_DIV_EQUAL) reduce 582) ((T_INC) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_MINUS_EQUAL) reduce 582) ((T_MOD_EQUAL) reduce 582) ((T_MUL_EQUAL) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_OR_EQUAL) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_PLUS_EQUAL) reduce 582) ((T_POW) reduce 582) ((T_POW_EQUAL) reduce 582) ((T_SL) reduce 582) ((T_SL_EQUAL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) ((T_SR_EQUAL) reduce 582) ((T_XOR_EQUAL) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 3472 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((";") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 3473 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) ((";") reduce 584) (("<") reduce 584) (("=") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AND_EQUAL) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_COALESCE_EQUAL) reduce 584) ((T_CONCAT_EQUAL) reduce 584) ((T_DEC) reduce 584) ((T_DIV_EQUAL) reduce 584) ((T_INC) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_MINUS_EQUAL) reduce 584) ((T_MOD_EQUAL) reduce 584) ((T_MUL_EQUAL) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_OR_EQUAL) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_PLUS_EQUAL) reduce 584) ((T_POW) reduce 584) ((T_POW_EQUAL) reduce 584) ((T_SL) reduce 584) ((T_SL_EQUAL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) ((T_SR_EQUAL) reduce 584) ((T_XOR_EQUAL) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 3476 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) ((";") reduce 583) (("<") reduce 583) (("=") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AND_EQUAL) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_COALESCE_EQUAL) reduce 583) ((T_CONCAT_EQUAL) reduce 583) ((T_DEC) reduce 583) ((T_DIV_EQUAL) reduce 583) ((T_INC) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_MINUS_EQUAL) reduce 583) ((T_MOD_EQUAL) reduce 583) ((T_MUL_EQUAL) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_OR_EQUAL) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_PLUS_EQUAL) reduce 583) ((T_POW) reduce 583) ((T_POW_EQUAL) reduce 583) ((T_SL) reduce 583) ((T_SL_EQUAL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) ((T_SR_EQUAL) reduce 583) ((T_XOR_EQUAL) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 3477 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((";") reduce 561) (("<") reduce 561) (("=") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AND_EQUAL) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_COALESCE_EQUAL) reduce 561) ((T_CONCAT_EQUAL) reduce 561) ((T_DEC) reduce 561) ((T_DIV_EQUAL) reduce 561) ((T_INC) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_MINUS_EQUAL) reduce 561) ((T_MOD_EQUAL) reduce 561) ((T_MUL_EQUAL) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_OR_EQUAL) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_PLUS_EQUAL) reduce 561) ((T_POW) reduce 561) ((T_POW_EQUAL) reduce 561) ((T_SL) reduce 561) ((T_SL_EQUAL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) ((T_SR_EQUAL) reduce 561) ((T_XOR_EQUAL) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 3478 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((";") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 3479 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) ((",") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) ((";") reduce 562) (("<") reduce 562) (("=") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AND_EQUAL) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_COALESCE_EQUAL) reduce 562) ((T_CONCAT_EQUAL) reduce 562) ((T_DEC) reduce 562) ((T_DIV_EQUAL) reduce 562) ((T_INC) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_MINUS_EQUAL) reduce 562) ((T_MOD_EQUAL) reduce 562) ((T_MUL_EQUAL) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_OR_EQUAL) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_PLUS_EQUAL) reduce 562) ((T_POW) reduce 562) ((T_POW_EQUAL) reduce 562) ((T_SL) reduce 562) ((T_SL_EQUAL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) ((T_SR_EQUAL) reduce 562) ((T_XOR_EQUAL) reduce 562) (("[") reduce 562) (("^") reduce 562) (("|") reduce 562)) 3481 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) ((",") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) ((";") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_PIPE) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 3483 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) ((";") reduce 520) (("<") reduce 520) (("=") shift) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_PIPE) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("|") reduce 520)) 3485 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 405) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 405) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3486 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 482) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 482) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3497 ((("%") reduce 540) (("*") reduce 540) (("+") reduce 540) ((",") reduce 540) (("-") reduce 540) ((".") reduce 540) (("/") reduce 540) ((";") reduce 540) (("<") reduce 540) ((">") reduce 540) (("?") reduce 540) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_BOOLEAN_AND) reduce 540) ((T_BOOLEAN_OR) reduce 540) ((T_COALESCE) reduce 540) ((T_DOUBLE_ARROW) reduce 540) ((T_INSTANCEOF) reduce 540) ((T_IS_EQUAL) reduce 540) ((T_IS_GREATER_OR_EQUAL) reduce 540) ((T_IS_IDENTICAL) reduce 540) ((T_IS_NOT_EQUAL) reduce 540) ((T_IS_NOT_IDENTICAL) reduce 540) ((T_IS_SMALLER_OR_EQUAL) reduce 540) ((T_LOGICAL_AND) reduce 540) ((T_LOGICAL_OR) reduce 540) ((T_LOGICAL_XOR) reduce 540) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 540) ((T_OBJECT_OPERATOR) reduce 540) ((T_PIPE) reduce 540) ((T_POW) reduce 540) ((T_SL) reduce 540) ((T_SPACESHIP) reduce 540) ((T_SR) reduce 540) (("[") reduce 540) (("^") reduce 540) (("|") reduce 540)) 3499 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) ((";") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_PIPE) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("|") reduce 521)) 3501 ((("%") reduce 534) (("*") reduce 534) (("+") reduce 534) ((",") reduce 534) (("-") reduce 534) ((".") reduce 534) (("/") reduce 534) ((";") reduce 534) (("<") reduce 534) ((">") reduce 534) (("?") reduce 534) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_BOOLEAN_AND) reduce 534) ((T_BOOLEAN_OR) reduce 534) ((T_COALESCE) reduce 534) ((T_DOUBLE_ARROW) reduce 534) ((T_INSTANCEOF) reduce 534) ((T_IS_EQUAL) reduce 534) ((T_IS_GREATER_OR_EQUAL) reduce 534) ((T_IS_IDENTICAL) reduce 534) ((T_IS_NOT_EQUAL) reduce 534) ((T_IS_NOT_IDENTICAL) reduce 534) ((T_IS_SMALLER_OR_EQUAL) reduce 534) ((T_LOGICAL_AND) reduce 534) ((T_LOGICAL_OR) reduce 534) ((T_LOGICAL_XOR) reduce 534) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PIPE) reduce 534) ((T_POW) reduce 534) ((T_SL) reduce 534) ((T_SPACESHIP) reduce 534) ((T_SR) reduce 534) (("[") reduce 534) (("^") reduce 534) (("|") reduce 534)) 3502 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) ((";") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_PIPE) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("^") reduce 524) (("|") reduce 524)) 3506 ((("%") reduce 517) (("(") shift) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((";") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_PIPE) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("^") reduce 517) (("|") reduce 517)) 3507 ((("%") reduce 533) (("*") reduce 533) (("+") reduce 533) ((",") reduce 533) (("-") reduce 533) ((".") reduce 533) (("/") reduce 533) ((";") reduce 533) (("<") reduce 533) ((">") reduce 533) (("?") reduce 533) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_BOOLEAN_AND) reduce 533) ((T_BOOLEAN_OR) reduce 533) ((T_COALESCE) reduce 533) ((T_DOUBLE_ARROW) reduce 533) ((T_INSTANCEOF) reduce 533) ((T_IS_EQUAL) reduce 533) ((T_IS_GREATER_OR_EQUAL) reduce 533) ((T_IS_IDENTICAL) reduce 533) ((T_IS_NOT_EQUAL) reduce 533) ((T_IS_NOT_IDENTICAL) reduce 533) ((T_IS_SMALLER_OR_EQUAL) reduce 533) ((T_LOGICAL_AND) reduce 533) ((T_LOGICAL_OR) reduce 533) ((T_LOGICAL_XOR) reduce 533) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PIPE) reduce 533) ((T_POW) reduce 533) ((T_SL) reduce 533) ((T_SPACESHIP) reduce 533) ((T_SR) reduce 533) (("[") reduce 533) (("^") reduce 533) (("|") reduce 533)) 3508 ((("%") reduce 537) (("*") reduce 537) (("+") reduce 537) ((",") reduce 537) (("-") reduce 537) ((".") reduce 537) (("/") reduce 537) ((";") reduce 537) (("<") reduce 537) ((">") reduce 537) (("?") reduce 537) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_BOOLEAN_AND) reduce 537) ((T_BOOLEAN_OR) reduce 537) ((T_COALESCE) reduce 537) ((T_DOUBLE_ARROW) reduce 537) ((T_INSTANCEOF) reduce 537) ((T_IS_EQUAL) reduce 537) ((T_IS_GREATER_OR_EQUAL) reduce 537) ((T_IS_IDENTICAL) reduce 537) ((T_IS_NOT_EQUAL) reduce 537) ((T_IS_NOT_IDENTICAL) reduce 537) ((T_IS_SMALLER_OR_EQUAL) reduce 537) ((T_LOGICAL_AND) reduce 537) ((T_LOGICAL_OR) reduce 537) ((T_LOGICAL_XOR) reduce 537) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_PIPE) reduce 537) ((T_POW) reduce 537) ((T_SL) reduce 537) ((T_SPACESHIP) reduce 537) ((T_SR) reduce 537) (("[") reduce 537) (("^") reduce 537) (("|") reduce 537)) 3514 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) ((";") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PIPE) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("|") reduce 532)) 3516 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) ((";") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_PIPE) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("^") reduce 523) (("|") reduce 523)) 3518 ((("%") reduce 536) (("*") reduce 536) (("+") reduce 536) ((",") reduce 536) (("-") reduce 536) ((".") reduce 536) (("/") reduce 536) ((";") reduce 536) (("<") reduce 536) ((">") reduce 536) (("?") reduce 536) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_BOOLEAN_AND) reduce 536) ((T_BOOLEAN_OR) reduce 536) ((T_COALESCE) reduce 536) ((T_DOUBLE_ARROW) reduce 536) ((T_INSTANCEOF) reduce 536) ((T_IS_EQUAL) reduce 536) ((T_IS_GREATER_OR_EQUAL) reduce 536) ((T_IS_IDENTICAL) reduce 536) ((T_IS_NOT_EQUAL) reduce 536) ((T_IS_NOT_IDENTICAL) reduce 536) ((T_IS_SMALLER_OR_EQUAL) reduce 536) ((T_LOGICAL_AND) reduce 536) ((T_LOGICAL_OR) reduce 536) ((T_LOGICAL_XOR) reduce 536) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 536) ((T_OBJECT_OPERATOR) reduce 536) ((T_PIPE) reduce 536) ((T_POW) reduce 536) ((T_SL) reduce 536) ((T_SPACESHIP) reduce 536) ((T_SR) reduce 536) (("[") reduce 536) (("^") reduce 536) (("|") reduce 536)) 3519 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("|") reduce 94)) 3520 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("|") reduce 93)) 3521 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("|") reduce 95)) 3523 ((("%") reduce 539) (("*") reduce 539) (("+") reduce 539) ((",") reduce 539) (("-") reduce 539) ((".") reduce 539) (("/") reduce 539) ((";") reduce 539) (("<") reduce 539) ((">") reduce 539) (("?") reduce 539) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_BOOLEAN_AND) reduce 539) ((T_BOOLEAN_OR) reduce 539) ((T_COALESCE) reduce 539) ((T_DOUBLE_ARROW) reduce 539) ((T_INSTANCEOF) reduce 539) ((T_IS_EQUAL) reduce 539) ((T_IS_GREATER_OR_EQUAL) reduce 539) ((T_IS_IDENTICAL) reduce 539) ((T_IS_NOT_EQUAL) reduce 539) ((T_IS_NOT_IDENTICAL) reduce 539) ((T_IS_SMALLER_OR_EQUAL) reduce 539) ((T_LOGICAL_AND) reduce 539) ((T_LOGICAL_OR) reduce 539) ((T_LOGICAL_XOR) reduce 539) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 539) ((T_OBJECT_OPERATOR) reduce 539) ((T_PIPE) reduce 539) ((T_POW) reduce 539) ((T_SL) reduce 539) ((T_SPACESHIP) reduce 539) ((T_SR) reduce 539) (("[") reduce 539) (("^") reduce 539) (("|") reduce 539)) 3526 ((("%") reduce 538) (("*") reduce 538) (("+") reduce 538) ((",") reduce 538) (("-") reduce 538) ((".") reduce 538) (("/") reduce 538) ((";") reduce 538) (("<") reduce 538) ((">") reduce 538) (("?") reduce 538) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_BOOLEAN_AND) reduce 538) ((T_BOOLEAN_OR) reduce 538) ((T_COALESCE) reduce 538) ((T_DOUBLE_ARROW) reduce 538) ((T_INSTANCEOF) reduce 538) ((T_IS_EQUAL) reduce 538) ((T_IS_GREATER_OR_EQUAL) reduce 538) ((T_IS_IDENTICAL) reduce 538) ((T_IS_NOT_EQUAL) reduce 538) ((T_IS_NOT_IDENTICAL) reduce 538) ((T_IS_SMALLER_OR_EQUAL) reduce 538) ((T_LOGICAL_AND) reduce 538) ((T_LOGICAL_OR) reduce 538) ((T_LOGICAL_XOR) reduce 538) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 538) ((T_OBJECT_OPERATOR) reduce 538) ((T_PIPE) reduce 538) ((T_POW) reduce 538) ((T_SL) reduce 538) ((T_SPACESHIP) reduce 538) ((T_SR) reduce 538) (("[") reduce 538) (("^") reduce 538) (("|") reduce 538)) 3532 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92)) 3535 ((("%") reduce 535) (("*") reduce 535) (("+") reduce 535) ((",") reduce 535) (("-") reduce 535) ((".") reduce 535) (("/") reduce 535) ((";") reduce 535) (("<") reduce 535) ((">") reduce 535) (("?") reduce 535) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_BOOLEAN_AND) reduce 535) ((T_BOOLEAN_OR) reduce 535) ((T_COALESCE) reduce 535) ((T_DOUBLE_ARROW) reduce 535) ((T_INSTANCEOF) reduce 535) ((T_IS_EQUAL) reduce 535) ((T_IS_GREATER_OR_EQUAL) reduce 535) ((T_IS_IDENTICAL) reduce 535) ((T_IS_NOT_EQUAL) reduce 535) ((T_IS_NOT_IDENTICAL) reduce 535) ((T_IS_SMALLER_OR_EQUAL) reduce 535) ((T_LOGICAL_AND) reduce 535) ((T_LOGICAL_OR) reduce 535) ((T_LOGICAL_XOR) reduce 535) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PIPE) reduce 535) ((T_POW) reduce 535) ((T_SL) reduce 535) ((T_SPACESHIP) reduce 535) ((T_SR) reduce 535) (("[") reduce 535) (("^") reduce 535) (("|") reduce 535)) 3537 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) ((";") reduce 568) (("<") reduce 568) (("=") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AND_EQUAL) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_COALESCE_EQUAL) reduce 568) ((T_CONCAT_EQUAL) reduce 568) ((T_DEC) reduce 568) ((T_DIV_EQUAL) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INC) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_MINUS_EQUAL) reduce 568) ((T_MOD_EQUAL) reduce 568) ((T_MUL_EQUAL) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_OR_EQUAL) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_PLUS_EQUAL) reduce 568) ((T_POW) reduce 568) ((T_POW_EQUAL) reduce 568) ((T_SL) reduce 568) ((T_SL_EQUAL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) ((T_SR_EQUAL) reduce 568) ((T_XOR_EQUAL) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 3538 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 479) (("(") shift) (("*") reduce 479) (("+") shift) ((",") reduce 479) (("-") shift) ((".") reduce 479) (("/") reduce 479) ((";") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 479) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 479) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 479) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PIPE) reduce 479) ((T_POW) reduce 479) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 479) (("`") shift) (("|") reduce 479) (("~") shift)) 3545 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) ((",") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) ((";") reduce 564) (("<") reduce 564) (("=") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AND_EQUAL) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_COALESCE_EQUAL) reduce 564) ((T_CONCAT_EQUAL) reduce 564) ((T_DEC) reduce 564) ((T_DIV_EQUAL) reduce 564) ((T_DOUBLE_ARROW) reduce 564) ((T_INC) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_MINUS_EQUAL) reduce 564) ((T_MOD_EQUAL) reduce 564) ((T_MUL_EQUAL) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_OR_EQUAL) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_PLUS_EQUAL) reduce 564) ((T_POW) reduce 564) ((T_POW_EQUAL) reduce 564) ((T_SL) reduce 564) ((T_SL_EQUAL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) ((T_SR_EQUAL) reduce 564) ((T_XOR_EQUAL) reduce 564) (("[") reduce 564) (("^") reduce 564) (("|") reduce 564)) 3546 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) ((";") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PIPE) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 551) (("^") reduce 530) (("|") reduce 530)) 3548 ((("%") reduce 529) (("*") reduce 529) (("+") reduce 529) ((",") reduce 529) (("-") reduce 529) ((".") reduce 529) (("/") reduce 529) ((";") reduce 529) (("<") reduce 529) ((">") reduce 529) (("?") reduce 529) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_BOOLEAN_AND) reduce 529) ((T_BOOLEAN_OR) reduce 529) ((T_COALESCE) reduce 529) ((T_DOUBLE_ARROW) reduce 529) ((T_INSTANCEOF) reduce 529) ((T_IS_EQUAL) reduce 529) ((T_IS_GREATER_OR_EQUAL) reduce 529) ((T_IS_IDENTICAL) reduce 529) ((T_IS_NOT_EQUAL) reduce 529) ((T_IS_NOT_IDENTICAL) reduce 529) ((T_IS_SMALLER_OR_EQUAL) reduce 529) ((T_LOGICAL_AND) reduce 529) ((T_LOGICAL_OR) reduce 529) ((T_LOGICAL_XOR) reduce 529) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PIPE) reduce 529) ((T_POW) reduce 529) ((T_SL) reduce 529) ((T_SPACESHIP) reduce 529) ((T_SR) reduce 529) (("[") reduce 554) (("^") reduce 529) (("|") reduce 529)) 3549 ((("%") reduce 528) (("(") reduce 557) (("*") reduce 528) (("+") reduce 528) ((",") reduce 528) (("-") reduce 528) ((".") reduce 528) (("/") reduce 528) ((";") reduce 528) (("<") reduce 528) ((">") reduce 528) (("?") reduce 528) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_BOOLEAN_AND) reduce 528) ((T_BOOLEAN_OR) reduce 528) ((T_COALESCE) reduce 528) ((T_DOUBLE_ARROW) reduce 528) ((T_INSTANCEOF) reduce 528) ((T_IS_EQUAL) reduce 528) ((T_IS_GREATER_OR_EQUAL) reduce 528) ((T_IS_IDENTICAL) reduce 528) ((T_IS_NOT_EQUAL) reduce 528) ((T_IS_NOT_IDENTICAL) reduce 528) ((T_IS_SMALLER_OR_EQUAL) reduce 528) ((T_LOGICAL_AND) reduce 528) ((T_LOGICAL_OR) reduce 528) ((T_LOGICAL_XOR) reduce 528) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PIPE) reduce 528) ((T_POW) reduce 528) ((T_SL) reduce 528) ((T_SPACESHIP) reduce 528) ((T_SR) reduce 528) (("[") reduce 550) (("^") reduce 528) (("|") reduce 528)) 3550 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 480) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 480) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 480) ((T_LOGICAL_OR) reduce 480) ((T_LOGICAL_XOR) reduce 480) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3553 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) ((",") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) ((";") reduce 563) (("<") reduce 563) (("=") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AND_EQUAL) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_COALESCE_EQUAL) reduce 563) ((T_CONCAT_EQUAL) reduce 563) ((T_DEC) reduce 563) ((T_DIV_EQUAL) reduce 563) ((T_DOUBLE_ARROW) reduce 563) ((T_INC) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_MINUS_EQUAL) reduce 563) ((T_MOD_EQUAL) reduce 563) ((T_MUL_EQUAL) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_OR_EQUAL) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_PLUS_EQUAL) reduce 563) ((T_POW) reduce 563) ((T_POW_EQUAL) reduce 563) ((T_SL) reduce 563) ((T_SL_EQUAL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) ((T_SR_EQUAL) reduce 563) ((T_XOR_EQUAL) reduce 563) (("[") reduce 563) (("^") reduce 563) (("|") reduce 563)) 3554 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) ((",") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) ((";") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_DOUBLE_ARROW) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_PIPE) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("^") reduce 484) (("|") reduce 484)) 3555 ((("%") reduce 466) (("*") reduce 466) (("+") reduce 466) ((",") reduce 466) (("-") reduce 466) ((".") reduce 466) (("/") reduce 466) ((";") reduce 466) (("<") reduce 466) ((">") reduce 466) (("?") reduce 466) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_BOOLEAN_AND) reduce 466) ((T_BOOLEAN_OR) reduce 466) ((T_COALESCE) reduce 466) ((T_DOUBLE_ARROW) reduce 466) ((T_INSTANCEOF) reduce 466) ((T_IS_EQUAL) reduce 466) ((T_IS_GREATER_OR_EQUAL) reduce 466) ((T_IS_IDENTICAL) reduce 466) ((T_IS_NOT_EQUAL) reduce 466) ((T_IS_NOT_IDENTICAL) reduce 466) ((T_IS_SMALLER_OR_EQUAL) reduce 466) ((T_LOGICAL_AND) reduce 466) ((T_LOGICAL_OR) reduce 466) ((T_LOGICAL_XOR) reduce 466) ((T_PIPE) reduce 466) ((T_POW) reduce 466) ((T_SL) reduce 466) ((T_SPACESHIP) reduce 466) ((T_SR) reduce 466) (("^") reduce 466) (("|") reduce 466)) 3556 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) ((";") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_PIPE) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("^") reduce 488) (("|") reduce 488)) 3557 ((("%") reduce 531) (("(") shift) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) ((";") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("|") reduce 531)) 3558 ((("%") reduce 461) (("(") reduce 558) (("*") reduce 461) (("+") reduce 461) ((",") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) ((";") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PIPE) reduce 461) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") reduce 552) (("^") reduce 461) (("|") reduce 461)) 3559 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) ((",") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) ((";") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_DOUBLE_ARROW) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_PIPE) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 3560 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) ((",") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) ((";") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_DOUBLE_ARROW) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_PIPE) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("|") reduce 476)) 3561 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) ((";") reduce 559) (("<") reduce 559) (("=") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AND_EQUAL) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_COALESCE_EQUAL) reduce 559) ((T_CONCAT_EQUAL) reduce 559) ((T_DEC) reduce 559) ((T_DIV_EQUAL) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INC) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_MINUS_EQUAL) reduce 559) ((T_MOD_EQUAL) reduce 559) ((T_MUL_EQUAL) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_OR_EQUAL) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_PLUS_EQUAL) reduce 559) ((T_POW) reduce 559) ((T_POW_EQUAL) reduce 559) ((T_SL) reduce 559) ((T_SL_EQUAL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) ((T_SR_EQUAL) reduce 559) ((T_XOR_EQUAL) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 3562 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((";") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 3563 ((("%") reduce 403) (("*") reduce 403) (("+") reduce 403) ((",") reduce 403) (("-") reduce 403) ((".") reduce 403) (("/") reduce 403) ((";") reduce 403) (("<") reduce 403) (("=") shift) ((">") reduce 403) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_DOUBLE_ARROW) reduce 403) ((T_INC) shift) ((T_INSTANCEOF) reduce 403) ((T_IS_EQUAL) reduce 403) ((T_IS_GREATER_OR_EQUAL) reduce 403) ((T_IS_IDENTICAL) reduce 403) ((T_IS_NOT_EQUAL) reduce 403) ((T_IS_NOT_IDENTICAL) reduce 403) ((T_IS_SMALLER_OR_EQUAL) reduce 403) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 403) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 403) ((T_POW_EQUAL) shift) ((T_SL) reduce 403) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 403) ((T_SR) reduce 403) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 548) (("^") reduce 403) (("|") reduce 403)) 3566 ((("%") reduce 448) (("*") reduce 448) (("+") reduce 448) ((",") reduce 448) (("-") reduce 448) ((".") reduce 448) (("/") reduce 448) ((";") reduce 448) (("<") reduce 448) ((">") reduce 448) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_BOOLEAN_AND) reduce 448) ((T_BOOLEAN_OR) reduce 448) ((T_COALESCE) reduce 448) ((T_DOUBLE_ARROW) reduce 448) ((T_INSTANCEOF) reduce 448) ((T_IS_EQUAL) reduce 448) ((T_IS_GREATER_OR_EQUAL) reduce 448) ((T_IS_IDENTICAL) reduce 448) ((T_IS_NOT_EQUAL) reduce 448) ((T_IS_NOT_IDENTICAL) reduce 448) ((T_IS_SMALLER_OR_EQUAL) reduce 448) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_PIPE) reduce 448) ((T_POW) shift) ((T_SL) reduce 448) ((T_SPACESHIP) reduce 448) ((T_SR) reduce 448) (("^") reduce 448) (("|") reduce 448)) 3598 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 432) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 432) (("<") shift) ((">") shift) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_DOUBLE_ARROW) reduce 432) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 432)) 3599 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 435) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 435) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_DOUBLE_ARROW) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("|") reduce 435)) 3600 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 444) (("-") shift) ((".") reduce 444) (("/") shift) ((";") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_PIPE) reduce 444) ((T_POW) shift) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("|") reduce 444)) 3601 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 458) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 458) (("<") shift) ((">") shift) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 458) (("|") reduce 458)) 3602 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 443) (("-") shift) ((".") reduce 443) (("/") shift) ((";") reduce 443) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_DOUBLE_ARROW) reduce 443) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_PIPE) reduce 443) ((T_POW) shift) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("^") reduce 443) (("|") reduce 443)) 3603 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) ((",") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) ((";") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_DOUBLE_ARROW) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_PIPE) reduce 440) ((T_POW) shift) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("|") reduce 440)) 3604 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 453) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 453) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_DOUBLE_ARROW) reduce 453) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_PIPE) reduce 453) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 453) ((T_SR) shift) (("^") reduce 453) (("|") reduce 453)) 3605 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 431) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 431) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3606 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 429) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 429) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 429) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3607 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 430) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 430) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 430) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3608 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 455) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 455) (("<") shift) ((">") shift) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 455) ((T_SR) shift) (("^") reduce 455) (("|") reduce 455)) 3609 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 450) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 450) (("<") shift) ((">") shift) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_DOUBLE_ARROW) reduce 450) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 450) (("|") reduce 450)) 3610 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 452) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 452) (("<") shift) ((">") shift) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_DOUBLE_ARROW) reduce 452) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 452) (("|") reduce 452)) 3611 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 449) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 449) (("<") shift) ((">") shift) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_DOUBLE_ARROW) reduce 449) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 449) (("|") reduce 449)) 3612 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 457) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 457) (("<") shift) ((">") shift) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 457) ((T_SR) shift) (("^") reduce 457) (("|") reduce 457)) 3613 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 451) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 451) (("<") shift) ((">") shift) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_DOUBLE_ARROW) reduce 451) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 451) (("|") reduce 451)) 3616 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 3617 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 3618 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 3619 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((";") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 3620 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 3621 ((("%") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) ((";") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 3622 ((("%") reduce 511) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((";") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("|") reduce 511)) 3623 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) ((";") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PIPE) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 3624 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((";") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 3625 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((";") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("|") reduce 512)) 3626 ((("%") reduce 573) (("*") reduce 573) (("+") reduce 573) ((",") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) ((";") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_DOUBLE_ARROW) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("|") reduce 573)) 3632 ((("%") reduce 574) (("*") reduce 574) (("+") reduce 574) ((",") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) ((";") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_DOUBLE_ARROW) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("|") reduce 574)) 3633 ((("%") reduce 578) (("*") reduce 578) (("+") reduce 578) ((",") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) ((";") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_DOUBLE_ARROW) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("|") reduce 578)) 3634 ((("%") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) ((";") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 3635 ((("%") reduce 575) (("*") reduce 575) (("+") reduce 575) ((",") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) ((";") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_DOUBLE_ARROW) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("|") reduce 575)) 3636 ((("%") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) ((";") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 3639 ((("%") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) ((";") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 3640 ((("%") reduce 576) (("*") reduce 576) (("+") reduce 576) ((",") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) ((";") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_DOUBLE_ARROW) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("|") reduce 576)) 3642 ((("%") reduce 577) (("*") reduce 577) (("+") reduce 577) ((",") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) ((";") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_DOUBLE_ARROW) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("|") reduce 577)) 3644 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((";") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("|") reduce 513)) 3645 ((("%") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) ((";") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 3648 ((("%") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) ((";") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 3649 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 465) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 465) (("<") shift) ((">") shift) (("?") reduce 465) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 465) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3650 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 427) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 427) (("<") shift) ((">") shift) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_DOUBLE_ARROW) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3651 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 428) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 428) (("<") shift) ((">") shift) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_DOUBLE_ARROW) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3652 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 433) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 433) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_DOUBLE_ARROW) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("|") reduce 433)) 3653 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 434) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 434) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_DOUBLE_ARROW) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("|") reduce 434)) 3657 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 463) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 463) (("<") shift) ((">") shift) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 463) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3658 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 464) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 464) (("<") shift) ((">") shift) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 464) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3659 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 456) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 456) (("<") shift) ((">") shift) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 456) ((T_SR) shift) (("^") reduce 456) (("|") reduce 456)) 3660 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 454) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 454) (("<") shift) ((">") shift) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 454) ((T_SR) shift) (("^") reduce 454) (("|") reduce 454)) 3661 ((("%") reduce 441) (("*") reduce 441) (("+") reduce 441) ((",") reduce 441) (("-") reduce 441) ((".") reduce 441) (("/") reduce 441) ((";") reduce 441) (("<") reduce 441) ((">") reduce 441) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_DOUBLE_ARROW) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 441) ((T_IS_GREATER_OR_EQUAL) reduce 441) ((T_IS_IDENTICAL) reduce 441) ((T_IS_NOT_EQUAL) reduce 441) ((T_IS_NOT_IDENTICAL) reduce 441) ((T_IS_SMALLER_OR_EQUAL) reduce 441) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_PIPE) reduce 441) ((T_POW) shift) ((T_SL) reduce 441) ((T_SPACESHIP) reduce 441) ((T_SR) reduce 441) (("^") reduce 441) (("|") reduce 441)) 3662 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 436) (("-") shift) ((".") reduce 436) (("/") shift) ((";") reduce 436) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_DOUBLE_ARROW) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_PIPE) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 436) ((T_SR) shift) (("^") reduce 436) (("|") reduce 436)) 3663 ((("%") shift) (("*") shift) (("+") reduce 438) ((",") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") shift) ((";") reduce 438) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_DOUBLE_ARROW) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_PIPE) reduce 438) ((T_POW) shift) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("|") reduce 438)) 3664 ((("%") shift) (("*") shift) (("+") reduce 437) ((",") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") shift) ((";") reduce 437) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_DOUBLE_ARROW) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_PIPE) reduce 437) ((T_POW) shift) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("|") reduce 437)) 3665 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) ((",") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) ((";") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_DOUBLE_ARROW) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_PIPE) reduce 439) ((T_POW) shift) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("|") reduce 439)) 3666 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) ((",") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) ((";") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_DOUBLE_ARROW) reduce 442) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_PIPE) reduce 442) ((T_POW) shift) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("|") reduce 442)) 3668 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) ((",") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) ((";") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_DOUBLE_ARROW) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_PIPE) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("|") reduce 74)) 3669 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) ((",") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) ((";") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_DOUBLE_ARROW) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_PIPE) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("|") reduce 52)) 3670 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) ((",") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) ((";") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_DOUBLE_ARROW) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_PIPE) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("|") reduce 27)) 3671 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) ((",") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) ((";") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_DOUBLE_ARROW) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_PIPE) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("|") reduce 51)) 3672 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) ((",") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) ((";") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_DOUBLE_ARROW) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_PIPE) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("|") reduce 53)) 3673 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) ((",") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) ((";") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_DOUBLE_ARROW) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_PIPE) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("|") reduce 49)) 3674 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) ((",") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) ((";") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_DOUBLE_ARROW) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_PIPE) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("|") reduce 29)) 3675 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) ((",") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) ((";") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_DOUBLE_ARROW) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_PIPE) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("|") reduce 59)) 3676 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) ((",") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) ((";") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_DOUBLE_ARROW) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_PIPE) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("|") reduce 60)) 3677 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) ((",") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) ((";") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_DOUBLE_ARROW) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_PIPE) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("|") reduce 11)) 3678 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) ((",") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) ((";") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_DOUBLE_ARROW) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_PIPE) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("|") reduce 42)) 3679 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) ((",") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) ((";") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_DOUBLE_ARROW) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_PIPE) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("|") reduce 39)) 3680 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) ((",") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) ((";") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_DOUBLE_ARROW) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_PIPE) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("|") reduce 25)) 3681 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) ((",") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) ((";") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_DOUBLE_ARROW) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_PIPE) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("|") reduce 50)) 3682 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) ((",") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) ((";") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_DOUBLE_ARROW) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_PIPE) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("|") reduce 66)) 3683 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) ((",") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) ((";") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_DOUBLE_ARROW) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_PIPE) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("|") reduce 18)) 3684 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) ((",") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) ((";") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_DOUBLE_ARROW) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_PIPE) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("|") reduce 17)) 3685 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) ((",") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) ((";") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_DOUBLE_ARROW) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_PIPE) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("|") reduce 15)) 3686 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) ((",") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) ((";") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_DOUBLE_ARROW) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_PIPE) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("|") reduce 14)) 3687 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) ((",") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) ((";") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_DOUBLE_ARROW) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_PIPE) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("|") reduce 38)) 3688 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) ((",") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) ((";") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_DOUBLE_ARROW) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_PIPE) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("|") reduce 26)) 3689 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) ((",") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) ((";") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_DOUBLE_ARROW) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_PIPE) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("|") reduce 22)) 3690 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) ((",") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) ((";") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_DOUBLE_ARROW) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_PIPE) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("|") reduce 24)) 3691 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) ((",") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) ((";") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_DOUBLE_ARROW) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_PIPE) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("|") reduce 16)) 3692 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) ((",") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) ((";") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_DOUBLE_ARROW) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_PIPE) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("|") reduce 48)) 3693 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) ((",") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) ((";") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_DOUBLE_ARROW) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_PIPE) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("|") reduce 20)) 3694 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) ((",") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) ((";") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_DOUBLE_ARROW) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_PIPE) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("|") reduce 70)) 3695 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) ((",") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) ((";") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_DOUBLE_ARROW) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_PIPE) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("|") reduce 3)) 3696 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) ((",") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) ((";") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_DOUBLE_ARROW) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_PIPE) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("|") reduce 12)) 3697 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) ((",") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) ((";") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_DOUBLE_ARROW) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_PIPE) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("|") reduce 54)) 3698 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) ((",") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) ((";") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_DOUBLE_ARROW) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_PIPE) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("|") reduce 65)) 3699 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) ((",") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) ((";") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_DOUBLE_ARROW) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_PIPE) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("|") reduce 75)) 3700 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) ((",") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) ((";") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_DOUBLE_ARROW) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_PIPE) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("|") reduce 30)) 3701 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) ((",") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) ((";") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_DOUBLE_ARROW) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_PIPE) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("|") reduce 68)) 3702 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) ((",") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) ((";") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_DOUBLE_ARROW) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_PIPE) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("|") reduce 21)) 3703 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) ((",") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) ((";") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_DOUBLE_ARROW) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_PIPE) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("|") reduce 23)) 3704 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) ((",") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) ((";") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_DOUBLE_ARROW) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_PIPE) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("|") reduce 41)) 3705 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) ((",") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) ((";") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_DOUBLE_ARROW) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_PIPE) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("|") reduce 62)) 3706 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) ((",") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) ((";") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_DOUBLE_ARROW) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_PIPE) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("|") reduce 34)) 3707 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) ((",") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) ((";") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_DOUBLE_ARROW) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_PIPE) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("|") reduce 40)) 3708 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) ((",") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) ((";") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_DOUBLE_ARROW) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_PIPE) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("|") reduce 13)) 3709 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) ((",") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) ((";") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_DOUBLE_ARROW) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_PIPE) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("|") reduce 55)) 3710 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) ((",") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) ((";") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_DOUBLE_ARROW) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_PIPE) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("|") reduce 1)) 3711 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) ((",") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) ((";") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_DOUBLE_ARROW) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_PIPE) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("|") reduce 2)) 3712 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) ((",") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) ((";") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_DOUBLE_ARROW) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_PIPE) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("|") reduce 9)) 3713 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) ((",") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) ((";") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_DOUBLE_ARROW) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_PIPE) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("|") reduce 33)) 3714 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) ((",") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) ((";") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_DOUBLE_ARROW) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_PIPE) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("|") reduce 58)) 3715 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) ((",") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) ((";") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_DOUBLE_ARROW) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_PIPE) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("|") reduce 37)) 3716 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) ((",") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) ((";") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_DOUBLE_ARROW) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_PIPE) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("|") reduce 64)) 3717 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) ((",") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) ((";") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_DOUBLE_ARROW) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_PIPE) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("|") reduce 46)) 3718 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) ((",") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) ((";") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_DOUBLE_ARROW) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_PIPE) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("|") reduce 8)) 3719 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) ((",") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) ((";") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_DOUBLE_ARROW) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_PIPE) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("|") reduce 6)) 3720 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) ((",") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) ((";") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_DOUBLE_ARROW) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_PIPE) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("|") reduce 7)) 3721 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) ((",") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) ((";") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_DOUBLE_ARROW) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_PIPE) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("|") reduce 69)) 3722 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) ((",") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) ((";") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_DOUBLE_ARROW) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_PIPE) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("|") reduce 63)) 3723 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) ((",") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) ((";") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_DOUBLE_ARROW) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_PIPE) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("|") reduce 56)) 3724 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) ((",") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) ((";") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_DOUBLE_ARROW) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_PIPE) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("|") reduce 10)) 3725 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) ((",") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) ((";") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_DOUBLE_ARROW) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_PIPE) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("|") reduce 67)) 3726 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) ((",") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) ((";") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_DOUBLE_ARROW) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_PIPE) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("|") reduce 44)) 3727 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) ((",") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) ((";") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_DOUBLE_ARROW) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_PIPE) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("|") reduce 76)) 3728 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) ((",") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) ((";") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_DOUBLE_ARROW) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_PIPE) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("|") reduce 71)) 3729 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) ((",") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) ((";") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_DOUBLE_ARROW) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_PIPE) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("|") reduce 77)) 3730 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) ((",") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) ((";") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_DOUBLE_ARROW) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_PIPE) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("|") reduce 78)) 3731 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) ((",") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) ((";") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_DOUBLE_ARROW) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_PIPE) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("|") reduce 79)) 3732 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) ((",") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) ((";") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_DOUBLE_ARROW) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_PIPE) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("|") reduce 4)) 3733 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) ((",") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) ((";") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_DOUBLE_ARROW) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_PIPE) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("|") reduce 5)) 3734 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) ((",") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) ((";") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_DOUBLE_ARROW) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_PIPE) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("|") reduce 43)) 3735 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) ((",") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) ((";") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_DOUBLE_ARROW) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_PIPE) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("|") reduce 73)) 3736 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) ((",") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) ((";") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_DOUBLE_ARROW) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_PIPE) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("|") reduce 82)) 3737 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) ((",") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) ((";") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_DOUBLE_ARROW) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_PIPE) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("|") reduce 47)) 3738 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) ((",") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) ((";") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_DOUBLE_ARROW) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_PIPE) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("|") reduce 31)) 3739 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) ((",") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) ((";") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_DOUBLE_ARROW) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_PIPE) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("|") reduce 57)) 3740 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) ((",") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) ((";") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_DOUBLE_ARROW) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_PIPE) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("|") reduce 61)) 3741 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) ((",") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) ((";") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_DOUBLE_ARROW) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_PIPE) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("|") reduce 28)) 3742 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) ((",") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) ((";") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_DOUBLE_ARROW) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_PIPE) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("|") reduce 36)) 3743 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) ((",") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) ((";") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_DOUBLE_ARROW) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_PIPE) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("|") reduce 32)) 3744 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) ((",") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) ((";") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_DOUBLE_ARROW) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_PIPE) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("|") reduce 35)) 3745 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) ((",") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) ((";") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_DOUBLE_ARROW) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_PIPE) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("|") reduce 19)) 3746 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) ((",") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) ((";") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_DOUBLE_ARROW) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_PIPE) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("|") reduce 45)) 3747 ((("%") reduce 542) (("(") reduce 579) (("*") reduce 542) (("+") reduce 542) ((",") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) ((";") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PIPE) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 3749 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) ((",") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) ((";") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_DOUBLE_ARROW) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_PIPE) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("|") reduce 72)) 3750 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) ((",") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) ((";") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_DOUBLE_ARROW) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_PIPE) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("|") reduce 83)) 3751 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) ((",") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) ((";") reduce 572) (("<") reduce 572) (("=") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AND_EQUAL) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_COALESCE_EQUAL) reduce 572) ((T_CONCAT_EQUAL) reduce 572) ((T_DEC) reduce 572) ((T_DIV_EQUAL) reduce 572) ((T_DOUBLE_ARROW) reduce 572) ((T_INC) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_MINUS_EQUAL) reduce 572) ((T_MOD_EQUAL) reduce 572) ((T_MUL_EQUAL) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_OR_EQUAL) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_PLUS_EQUAL) reduce 572) ((T_POW) reduce 572) ((T_POW_EQUAL) reduce 572) ((T_SL) reduce 572) ((T_SL_EQUAL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) ((T_SR_EQUAL) reduce 572) ((T_XOR_EQUAL) reduce 572) (("[") reduce 572) (("^") reduce 572) (("|") reduce 572)) 3754 ((("%") reduce 544) (("(") reduce 580) (("*") reduce 544) (("+") reduce 544) ((",") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) ((";") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PIPE) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 3756 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((";") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("|") reduce 507)) 3757 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((";") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 3761 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) ((";") reduce 292) (("<") reduce 292) (("=") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AND_EQUAL) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_COALESCE_EQUAL) reduce 292) ((T_CONCAT_EQUAL) reduce 292) ((T_DEC) reduce 292) ((T_DIV_EQUAL) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INC) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_MINUS_EQUAL) reduce 292) ((T_MOD_EQUAL) reduce 292) ((T_MUL_EQUAL) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_OR_EQUAL) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_PLUS_EQUAL) reduce 292) ((T_POW) reduce 292) ((T_POW_EQUAL) reduce 292) ((T_SL) reduce 292) ((T_SL_EQUAL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) ((T_SR_EQUAL) reduce 292) ((T_XOR_EQUAL) reduce 292) (("[") reduce 292) (("^") reduce 292) (("|") reduce 292)) 3762 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) ((";") reduce 293) (("<") reduce 293) (("=") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AND_EQUAL) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_COALESCE_EQUAL) reduce 293) ((T_CONCAT_EQUAL) reduce 293) ((T_DEC) reduce 293) ((T_DIV_EQUAL) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INC) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_MINUS_EQUAL) reduce 293) ((T_MOD_EQUAL) reduce 293) ((T_MUL_EQUAL) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_OR_EQUAL) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_PLUS_EQUAL) reduce 293) ((T_POW) reduce 293) ((T_POW_EQUAL) reduce 293) ((T_SL) reduce 293) ((T_SL_EQUAL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) ((T_SR_EQUAL) reduce 293) ((T_XOR_EQUAL) reduce 293) (("[") reduce 293) (("^") reduce 293) (("|") reduce 293)) 3767 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) ((",") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) ((";") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_DOUBLE_ARROW) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_PIPE) reduce 425) ((T_POW) reduce 425) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 3769 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) ((",") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) ((";") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_DOUBLE_ARROW) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_PIPE) reduce 423) ((T_POW) reduce 423) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 3779 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 419) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 419) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3780 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 421) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 421) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3781 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 420) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 420) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3782 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 413) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 413) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 413) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3783 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 410) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 410) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 410) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3784 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 418) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 418) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3785 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 412) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 412) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3786 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 416) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 416) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3787 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 411) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 411) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3788 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 414) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 414) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 414) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3789 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 415) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 415) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 415) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3790 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 422) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 422) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3791 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 417) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 417) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3793 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 406) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 406) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3796 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) ((";") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 3799 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) ((",") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) ((";") reduce 564) (("<") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_DOUBLE_ARROW) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_POW) reduce 564) ((T_SL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) (("[") reduce 564) (("^") reduce 564) (("|") reduce 564)) 3801 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) ((",") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) ((";") reduce 563) (("<") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_DOUBLE_ARROW) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_POW) reduce 563) ((T_SL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) (("[") reduce 563) (("^") reduce 563) (("|") reduce 563)) 3803 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) ((";") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 3804 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((";") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 3805 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) ((",") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) ((";") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_DOUBLE_ARROW) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("[") reduce 548) (("^") reduce 407) (("|") reduce 407)) 3809 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) ((",") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) ((";") reduce 572) (("<") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_DOUBLE_ARROW) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_POW) reduce 572) ((T_SL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) (("[") reduce 572) (("^") reduce 572) (("|") reduce 572)) 3811 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((";") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("|") reduce 507)) 3812 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((";") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 3816 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) ((";") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("[") reduce 292) (("^") reduce 292) (("|") reduce 292)) 3817 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) ((";") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("[") reduce 293) (("^") reduce 293) (("|") reduce 293)) 3818 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) ((";") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 3821 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) ((",") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) ((";") reduce 571) (("<") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_DOUBLE_ARROW) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_POW) reduce 571) ((T_SL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) (("[") reduce 571) (("^") reduce 571) (("|") reduce 571)) 3822 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((";") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("|") reduce 506)) 3823 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((";") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("|") reduce 508)) 3828 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) ((";") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 3829 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) ((";") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 3830 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((";") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 3831 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) ((";") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 3834 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) ((";") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 3835 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((";") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 3836 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((";") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 3837 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) ((",") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) ((";") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("[") reduce 562) (("^") reduce 562) (("|") reduce 562)) 3838 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((";") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 3839 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) ((";") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 3842 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) ((";") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 3843 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) ((";") reduce 504) (("<") reduce 504) (("=") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AND_EQUAL) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_COALESCE_EQUAL) reduce 504) ((T_CONCAT_EQUAL) reduce 504) ((T_DEC) reduce 504) ((T_DIV_EQUAL) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INC) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_MINUS_EQUAL) reduce 504) ((T_MOD_EQUAL) reduce 504) ((T_MUL_EQUAL) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_OR_EQUAL) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_PLUS_EQUAL) reduce 504) ((T_POW) reduce 504) ((T_POW_EQUAL) reduce 504) ((T_SL) reduce 504) ((T_SL_EQUAL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) ((T_SR_EQUAL) reduce 504) ((T_XOR_EQUAL) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 3854 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) ((",") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) ((";") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PIPE) reduce 494) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 3855 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) ((";") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_PIPE) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("^") reduce 489) (("|") reduce 489)) 3865 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 494) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 494) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3866 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) ((";") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_PIPE) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("^") reduce 490) (("|") reduce 490)) 3868 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 481) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 481) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3870 ((("%") reduce 541) (("(") reduce 579) (("*") reduce 541) (("+") reduce 541) ((",") reduce 541) (("-") reduce 541) ((".") reduce 541) (("/") reduce 541) ((";") reduce 541) (("<") reduce 541) ((">") reduce 541) (("?") reduce 541) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_BOOLEAN_AND) reduce 541) ((T_BOOLEAN_OR) reduce 541) ((T_COALESCE) reduce 541) ((T_DOUBLE_ARROW) reduce 541) ((T_INSTANCEOF) reduce 541) ((T_IS_EQUAL) reduce 541) ((T_IS_GREATER_OR_EQUAL) reduce 541) ((T_IS_IDENTICAL) reduce 541) ((T_IS_NOT_EQUAL) reduce 541) ((T_IS_NOT_IDENTICAL) reduce 541) ((T_IS_SMALLER_OR_EQUAL) reduce 541) ((T_LOGICAL_AND) reduce 541) ((T_LOGICAL_OR) reduce 541) ((T_LOGICAL_XOR) reduce 541) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 541) ((T_OBJECT_OPERATOR) reduce 541) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 541) ((T_PIPE) reduce 541) ((T_POW) reduce 541) ((T_SL) reduce 541) ((T_SPACESHIP) reduce 541) ((T_SR) reduce 541) (("[") reduce 541) (("^") reduce 541) (("|") reduce 541)) 3872 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) ((",") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) ((";") reduce 571) (("<") reduce 571) (("=") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AND_EQUAL) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_COALESCE_EQUAL) reduce 571) ((T_CONCAT_EQUAL) reduce 571) ((T_DEC) reduce 571) ((T_DIV_EQUAL) reduce 571) ((T_DOUBLE_ARROW) reduce 571) ((T_INC) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_MINUS_EQUAL) reduce 571) ((T_MOD_EQUAL) reduce 571) ((T_MUL_EQUAL) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_OR_EQUAL) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_PLUS_EQUAL) reduce 571) ((T_POW) reduce 571) ((T_POW_EQUAL) reduce 571) ((T_SL) reduce 571) ((T_SL_EQUAL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) ((T_SR_EQUAL) reduce 571) ((T_XOR_EQUAL) reduce 571) (("[") reduce 571) (("^") reduce 571) (("|") reduce 571)) 3875 ((("%") reduce 543) (("(") reduce 580) (("*") reduce 543) (("+") reduce 543) ((",") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) ((";") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PIPE) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 3876 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((";") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("|") reduce 506)) 3877 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((";") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("|") reduce 508)) 3879 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) ((",") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) ((";") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_DOUBLE_ARROW) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_PIPE) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("^") reduce 485) (("|") reduce 485)) 3880 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) ((";") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_PIPE) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("^") reduce 487) (("|") reduce 487)) 3885 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) ((";") reduce 560) (("<") reduce 560) (("=") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AND_EQUAL) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_COALESCE_EQUAL) reduce 560) ((T_CONCAT_EQUAL) reduce 560) ((T_DEC) reduce 560) ((T_DIV_EQUAL) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INC) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_MINUS_EQUAL) reduce 560) ((T_MOD_EQUAL) reduce 560) ((T_MUL_EQUAL) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_OR_EQUAL) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_PLUS_EQUAL) reduce 560) ((T_POW) reduce 560) ((T_POW_EQUAL) reduce 560) ((T_SL) reduce 560) ((T_SL_EQUAL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) ((T_SR_EQUAL) reduce 560) ((T_XOR_EQUAL) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 3886 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) ((";") reduce 582) (("<") reduce 582) (("=") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AND_EQUAL) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_COALESCE_EQUAL) reduce 582) ((T_CONCAT_EQUAL) reduce 582) ((T_DEC) reduce 582) ((T_DIV_EQUAL) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INC) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_MINUS_EQUAL) reduce 582) ((T_MOD_EQUAL) reduce 582) ((T_MUL_EQUAL) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_OR_EQUAL) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_PLUS_EQUAL) reduce 582) ((T_POW) reduce 582) ((T_POW_EQUAL) reduce 582) ((T_SL) reduce 582) ((T_SL_EQUAL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) ((T_SR_EQUAL) reduce 582) ((T_XOR_EQUAL) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 3887 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((";") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 3888 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) ((";") reduce 584) (("<") reduce 584) (("=") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AND_EQUAL) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_COALESCE_EQUAL) reduce 584) ((T_CONCAT_EQUAL) reduce 584) ((T_DEC) reduce 584) ((T_DIV_EQUAL) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INC) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_MINUS_EQUAL) reduce 584) ((T_MOD_EQUAL) reduce 584) ((T_MUL_EQUAL) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_OR_EQUAL) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_PLUS_EQUAL) reduce 584) ((T_POW) reduce 584) ((T_POW_EQUAL) reduce 584) ((T_SL) reduce 584) ((T_SL_EQUAL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) ((T_SR_EQUAL) reduce 584) ((T_XOR_EQUAL) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 3891 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) ((";") reduce 583) (("<") reduce 583) (("=") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AND_EQUAL) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_COALESCE_EQUAL) reduce 583) ((T_CONCAT_EQUAL) reduce 583) ((T_DEC) reduce 583) ((T_DIV_EQUAL) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INC) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_MINUS_EQUAL) reduce 583) ((T_MOD_EQUAL) reduce 583) ((T_MUL_EQUAL) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_OR_EQUAL) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_PLUS_EQUAL) reduce 583) ((T_POW) reduce 583) ((T_POW_EQUAL) reduce 583) ((T_SL) reduce 583) ((T_SL_EQUAL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) ((T_SR_EQUAL) reduce 583) ((T_XOR_EQUAL) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 3892 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((";") reduce 561) (("<") reduce 561) (("=") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AND_EQUAL) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_COALESCE_EQUAL) reduce 561) ((T_CONCAT_EQUAL) reduce 561) ((T_DEC) reduce 561) ((T_DIV_EQUAL) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INC) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_MINUS_EQUAL) reduce 561) ((T_MOD_EQUAL) reduce 561) ((T_MUL_EQUAL) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_OR_EQUAL) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_PLUS_EQUAL) reduce 561) ((T_POW) reduce 561) ((T_POW_EQUAL) reduce 561) ((T_SL) reduce 561) ((T_SL_EQUAL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) ((T_SR_EQUAL) reduce 561) ((T_XOR_EQUAL) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 3893 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((";") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 3894 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) ((",") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) ((";") reduce 562) (("<") reduce 562) (("=") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AND_EQUAL) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_COALESCE_EQUAL) reduce 562) ((T_CONCAT_EQUAL) reduce 562) ((T_DEC) reduce 562) ((T_DIV_EQUAL) reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_INC) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_MINUS_EQUAL) reduce 562) ((T_MOD_EQUAL) reduce 562) ((T_MUL_EQUAL) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_OR_EQUAL) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_PLUS_EQUAL) reduce 562) ((T_POW) reduce 562) ((T_POW_EQUAL) reduce 562) ((T_SL) reduce 562) ((T_SL_EQUAL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) ((T_SR_EQUAL) reduce 562) ((T_XOR_EQUAL) reduce 562) (("[") reduce 562) (("^") reduce 562) (("|") reduce 562)) 3896 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) ((",") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) ((";") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_PIPE) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 3898 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) ((";") reduce 520) (("<") reduce 520) (("=") shift) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_PIPE) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("|") reduce 520)) 3900 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 405) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 405) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3901 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 482) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 482) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 482) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3904 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 481) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 481) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 481) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3905 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) ((",") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) ((";") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_DOUBLE_ARROW) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_PIPE) reduce 473) ((T_POW) shift) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("|") reduce 473)) 3906 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 483) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 483) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 483) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3907 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) ((",") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) ((";") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_DOUBLE_ARROW) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_PIPE) reduce 469) ((T_POW) shift) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 3908 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) ((",") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) ((";") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_DOUBLE_ARROW) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_PIPE) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("^") reduce 486) (("|") reduce 486)) 3910 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) ((";") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_PIPE) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("^") reduce 526) (("|") reduce 526)) 3912 ((("%") reduce 527) (("*") reduce 527) (("+") reduce 527) ((",") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) ((";") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_DOUBLE_ARROW) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_PIPE) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("^") reduce 527) (("|") reduce 527)) 3913 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) ((";") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_PIPE) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("^") reduce 525) (("|") reduce 525)) 3914 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 619) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 619) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 619) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3915 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 618) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 618) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 618) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3916 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((";") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 3917 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 478) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 478) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 478) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 3918 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) ((",") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) ((";") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_DOUBLE_ARROW) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_PIPE) reduce 471) ((T_POW) shift) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("|") reduce 471)) 3920 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 3921 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 3922 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 3923 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((";") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 3924 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 3925 ((("%") reduce 400) (("(") reduce 400) (("*") reduce 400) (("+") reduce 400) ((",") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) ((";") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_DOUBLE_ARROW) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 400) ((T_OBJECT_OPERATOR) reduce 400) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 400) ((T_PIPE) reduce 400) ((T_POW) reduce 400) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("[") reduce 400) (("^") reduce 400) (("|") reduce 400)) 3928 ((("%") reduce 511) (("(") reduce 511) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((";") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("|") reduce 511)) 3929 ((("%") reduce 402) (("(") shift) (("*") reduce 402) (("+") reduce 402) ((",") reduce 402) (("-") reduce 402) ((".") reduce 402) (("/") reduce 402) ((";") reduce 402) (("<") reduce 402) ((">") reduce 402) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_DOUBLE_ARROW) reduce 402) ((T_INSTANCEOF) reduce 402) ((T_IS_EQUAL) reduce 402) ((T_IS_GREATER_OR_EQUAL) reduce 402) ((T_IS_IDENTICAL) reduce 402) ((T_IS_NOT_EQUAL) reduce 402) ((T_IS_NOT_IDENTICAL) reduce 402) ((T_IS_SMALLER_OR_EQUAL) reduce 402) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_PIPE) reduce 402) ((T_POW) reduce 402) ((T_SL) reduce 402) ((T_SPACESHIP) reduce 402) ((T_SR) reduce 402) (("^") reduce 402) (("|") reduce 402)) 3930 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((";") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 3931 ((("%") reduce 512) (("(") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((";") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("|") reduce 512)) 3932 ((("%") reduce 573) (("(") reduce 573) (("*") reduce 573) (("+") reduce 573) ((",") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) ((";") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_DOUBLE_ARROW) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("|") reduce 573)) 3938 ((("%") reduce 574) (("(") reduce 574) (("*") reduce 574) (("+") reduce 574) ((",") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) ((";") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_DOUBLE_ARROW) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("|") reduce 574)) 3939 ((("%") reduce 578) (("(") reduce 578) (("*") reduce 578) (("+") reduce 578) ((",") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) ((";") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_DOUBLE_ARROW) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("|") reduce 578)) 3940 ((("%") reduce 575) (("(") reduce 575) (("*") reduce 575) (("+") reduce 575) ((",") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) ((";") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_DOUBLE_ARROW) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("|") reduce 575)) 3941 ((("%") reduce 576) (("(") reduce 576) (("*") reduce 576) (("+") reduce 576) ((",") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) ((";") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_DOUBLE_ARROW) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("|") reduce 576)) 3942 ((("%") reduce 399) (("(") reduce 399) (("*") reduce 399) (("+") reduce 399) ((",") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) ((";") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_DOUBLE_ARROW) reduce 399) ((T_INSTANCEOF) reduce 399) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 399) ((T_OBJECT_OPERATOR) reduce 399) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 399) ((T_PIPE) reduce 399) ((T_POW) reduce 399) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("[") reduce 399) (("^") reduce 399) (("|") reduce 399)) 3944 ((("%") reduce 577) (("(") reduce 577) (("*") reduce 577) (("+") reduce 577) ((",") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) ((";") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_DOUBLE_ARROW) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("|") reduce 577)) 3945 ((("%") reduce 401) (("(") reduce 401) (("*") reduce 401) (("+") reduce 401) ((",") reduce 401) (("-") reduce 401) ((".") reduce 401) (("/") reduce 401) ((";") reduce 401) (("<") reduce 401) ((">") reduce 401) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_DOUBLE_ARROW) reduce 401) ((T_INSTANCEOF) reduce 401) ((T_IS_EQUAL) reduce 401) ((T_IS_GREATER_OR_EQUAL) reduce 401) ((T_IS_IDENTICAL) reduce 401) ((T_IS_NOT_EQUAL) reduce 401) ((T_IS_NOT_IDENTICAL) reduce 401) ((T_IS_SMALLER_OR_EQUAL) reduce 401) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 401) ((T_OBJECT_OPERATOR) reduce 401) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 401) ((T_PIPE) reduce 401) ((T_POW) reduce 401) ((T_SL) reduce 401) ((T_SPACESHIP) reduce 401) ((T_SR) reduce 401) (("[") reduce 401) (("^") reduce 401) (("|") reduce 401)) 3953 ((("%") reduce 398) (("(") reduce 398) (("*") reduce 398) (("+") reduce 398) ((",") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) ((";") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_DOUBLE_ARROW) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 398) ((T_OBJECT_OPERATOR) reduce 398) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 398) ((T_PIPE) reduce 398) ((T_POW) reduce 398) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("[") reduce 398) (("^") reduce 398) (("|") reduce 398)) 3955 ((("%") reduce 513) (("(") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((";") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("|") reduce 513)) 3959 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DEFAULT) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("}") reduce 234) (("~") shift)) 3970 ((("%") reduce 540) (("*") reduce 540) (("+") reduce 540) ((",") reduce 540) (("-") reduce 540) ((".") reduce 540) (("/") reduce 540) (("<") reduce 540) ((">") reduce 540) (("?") reduce 540) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_BOOLEAN_AND) reduce 540) ((T_BOOLEAN_OR) reduce 540) ((T_COALESCE) reduce 540) ((T_DOUBLE_ARROW) reduce 540) ((T_INSTANCEOF) reduce 540) ((T_IS_EQUAL) reduce 540) ((T_IS_GREATER_OR_EQUAL) reduce 540) ((T_IS_IDENTICAL) reduce 540) ((T_IS_NOT_EQUAL) reduce 540) ((T_IS_NOT_IDENTICAL) reduce 540) ((T_IS_SMALLER_OR_EQUAL) reduce 540) ((T_LOGICAL_AND) reduce 540) ((T_LOGICAL_OR) reduce 540) ((T_LOGICAL_XOR) reduce 540) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 540) ((T_OBJECT_OPERATOR) reduce 540) ((T_PIPE) reduce 540) ((T_POW) reduce 540) ((T_SL) reduce 540) ((T_SPACESHIP) reduce 540) ((T_SR) reduce 540) (("[") reduce 540) (("^") reduce 540) (("|") reduce 540)) 3972 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_PIPE) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("|") reduce 521)) 3974 (((",") shift) ((T_DOUBLE_ARROW) reduce 125)) 3975 ((("%") reduce 534) (("*") reduce 534) (("+") reduce 534) ((",") reduce 534) (("-") reduce 534) ((".") reduce 534) (("/") reduce 534) (("<") reduce 534) ((">") reduce 534) (("?") reduce 534) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_BOOLEAN_AND) reduce 534) ((T_BOOLEAN_OR) reduce 534) ((T_COALESCE) reduce 534) ((T_DOUBLE_ARROW) reduce 534) ((T_INSTANCEOF) reduce 534) ((T_IS_EQUAL) reduce 534) ((T_IS_GREATER_OR_EQUAL) reduce 534) ((T_IS_IDENTICAL) reduce 534) ((T_IS_NOT_EQUAL) reduce 534) ((T_IS_NOT_IDENTICAL) reduce 534) ((T_IS_SMALLER_OR_EQUAL) reduce 534) ((T_LOGICAL_AND) reduce 534) ((T_LOGICAL_OR) reduce 534) ((T_LOGICAL_XOR) reduce 534) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PIPE) reduce 534) ((T_POW) reduce 534) ((T_SL) reduce 534) ((T_SPACESHIP) reduce 534) ((T_SR) reduce 534) (("[") reduce 534) (("^") reduce 534) (("|") reduce 534)) 3976 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_PIPE) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("^") reduce 524) (("|") reduce 524)) 3980 ((("%") reduce 517) (("(") shift) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_PIPE) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("^") reduce 517) (("|") reduce 517)) 3981 ((("%") reduce 533) (("*") reduce 533) (("+") reduce 533) ((",") reduce 533) (("-") reduce 533) ((".") reduce 533) (("/") reduce 533) (("<") reduce 533) ((">") reduce 533) (("?") reduce 533) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_BOOLEAN_AND) reduce 533) ((T_BOOLEAN_OR) reduce 533) ((T_COALESCE) reduce 533) ((T_DOUBLE_ARROW) reduce 533) ((T_INSTANCEOF) reduce 533) ((T_IS_EQUAL) reduce 533) ((T_IS_GREATER_OR_EQUAL) reduce 533) ((T_IS_IDENTICAL) reduce 533) ((T_IS_NOT_EQUAL) reduce 533) ((T_IS_NOT_IDENTICAL) reduce 533) ((T_IS_SMALLER_OR_EQUAL) reduce 533) ((T_LOGICAL_AND) reduce 533) ((T_LOGICAL_OR) reduce 533) ((T_LOGICAL_XOR) reduce 533) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PIPE) reduce 533) ((T_POW) reduce 533) ((T_SL) reduce 533) ((T_SPACESHIP) reduce 533) ((T_SR) reduce 533) (("[") reduce 533) (("^") reduce 533) (("|") reduce 533)) 3982 ((("%") reduce 537) (("*") reduce 537) (("+") reduce 537) ((",") reduce 537) (("-") reduce 537) ((".") reduce 537) (("/") reduce 537) (("<") reduce 537) ((">") reduce 537) (("?") reduce 537) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_BOOLEAN_AND) reduce 537) ((T_BOOLEAN_OR) reduce 537) ((T_COALESCE) reduce 537) ((T_DOUBLE_ARROW) reduce 537) ((T_INSTANCEOF) reduce 537) ((T_IS_EQUAL) reduce 537) ((T_IS_GREATER_OR_EQUAL) reduce 537) ((T_IS_IDENTICAL) reduce 537) ((T_IS_NOT_EQUAL) reduce 537) ((T_IS_NOT_IDENTICAL) reduce 537) ((T_IS_SMALLER_OR_EQUAL) reduce 537) ((T_LOGICAL_AND) reduce 537) ((T_LOGICAL_OR) reduce 537) ((T_LOGICAL_XOR) reduce 537) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_PIPE) reduce 537) ((T_POW) reduce 537) ((T_SL) reduce 537) ((T_SPACESHIP) reduce 537) ((T_SR) reduce 537) (("[") reduce 537) (("^") reduce 537) (("|") reduce 537)) 3988 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PIPE) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("|") reduce 532)) 3990 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_PIPE) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("^") reduce 523) (("|") reduce 523)) 3992 ((("%") reduce 536) (("*") reduce 536) (("+") reduce 536) ((",") reduce 536) (("-") reduce 536) ((".") reduce 536) (("/") reduce 536) (("<") reduce 536) ((">") reduce 536) (("?") reduce 536) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_BOOLEAN_AND) reduce 536) ((T_BOOLEAN_OR) reduce 536) ((T_COALESCE) reduce 536) ((T_DOUBLE_ARROW) reduce 536) ((T_INSTANCEOF) reduce 536) ((T_IS_EQUAL) reduce 536) ((T_IS_GREATER_OR_EQUAL) reduce 536) ((T_IS_IDENTICAL) reduce 536) ((T_IS_NOT_EQUAL) reduce 536) ((T_IS_NOT_IDENTICAL) reduce 536) ((T_IS_SMALLER_OR_EQUAL) reduce 536) ((T_LOGICAL_AND) reduce 536) ((T_LOGICAL_OR) reduce 536) ((T_LOGICAL_XOR) reduce 536) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 536) ((T_OBJECT_OPERATOR) reduce 536) ((T_PIPE) reduce 536) ((T_POW) reduce 536) ((T_SL) reduce 536) ((T_SPACESHIP) reduce 536) ((T_SR) reduce 536) (("[") reduce 536) (("^") reduce 536) (("|") reduce 536)) 3993 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("|") reduce 94)) 3994 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("|") reduce 93)) 3995 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("|") reduce 95)) 3997 ((("%") reduce 539) (("*") reduce 539) (("+") reduce 539) ((",") reduce 539) (("-") reduce 539) ((".") reduce 539) (("/") reduce 539) (("<") reduce 539) ((">") reduce 539) (("?") reduce 539) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_BOOLEAN_AND) reduce 539) ((T_BOOLEAN_OR) reduce 539) ((T_COALESCE) reduce 539) ((T_DOUBLE_ARROW) reduce 539) ((T_INSTANCEOF) reduce 539) ((T_IS_EQUAL) reduce 539) ((T_IS_GREATER_OR_EQUAL) reduce 539) ((T_IS_IDENTICAL) reduce 539) ((T_IS_NOT_EQUAL) reduce 539) ((T_IS_NOT_IDENTICAL) reduce 539) ((T_IS_SMALLER_OR_EQUAL) reduce 539) ((T_LOGICAL_AND) reduce 539) ((T_LOGICAL_OR) reduce 539) ((T_LOGICAL_XOR) reduce 539) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 539) ((T_OBJECT_OPERATOR) reduce 539) ((T_PIPE) reduce 539) ((T_POW) reduce 539) ((T_SL) reduce 539) ((T_SPACESHIP) reduce 539) ((T_SR) reduce 539) (("[") reduce 539) (("^") reduce 539) (("|") reduce 539)) 4000 ((("%") reduce 538) (("*") reduce 538) (("+") reduce 538) ((",") reduce 538) (("-") reduce 538) ((".") reduce 538) (("/") reduce 538) (("<") reduce 538) ((">") reduce 538) (("?") reduce 538) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_BOOLEAN_AND) reduce 538) ((T_BOOLEAN_OR) reduce 538) ((T_COALESCE) reduce 538) ((T_DOUBLE_ARROW) reduce 538) ((T_INSTANCEOF) reduce 538) ((T_IS_EQUAL) reduce 538) ((T_IS_GREATER_OR_EQUAL) reduce 538) ((T_IS_IDENTICAL) reduce 538) ((T_IS_NOT_EQUAL) reduce 538) ((T_IS_NOT_IDENTICAL) reduce 538) ((T_IS_SMALLER_OR_EQUAL) reduce 538) ((T_LOGICAL_AND) reduce 538) ((T_LOGICAL_OR) reduce 538) ((T_LOGICAL_XOR) reduce 538) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 538) ((T_OBJECT_OPERATOR) reduce 538) ((T_PIPE) reduce 538) ((T_POW) reduce 538) ((T_SL) reduce 538) ((T_SPACESHIP) reduce 538) ((T_SR) reduce 538) (("[") reduce 538) (("^") reduce 538) (("|") reduce 538)) 4006 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92)) 4009 ((("%") reduce 535) (("*") reduce 535) (("+") reduce 535) ((",") reduce 535) (("-") reduce 535) ((".") reduce 535) (("/") reduce 535) (("<") reduce 535) ((">") reduce 535) (("?") reduce 535) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_BOOLEAN_AND) reduce 535) ((T_BOOLEAN_OR) reduce 535) ((T_COALESCE) reduce 535) ((T_DOUBLE_ARROW) reduce 535) ((T_INSTANCEOF) reduce 535) ((T_IS_EQUAL) reduce 535) ((T_IS_GREATER_OR_EQUAL) reduce 535) ((T_IS_IDENTICAL) reduce 535) ((T_IS_NOT_EQUAL) reduce 535) ((T_IS_NOT_IDENTICAL) reduce 535) ((T_IS_SMALLER_OR_EQUAL) reduce 535) ((T_LOGICAL_AND) reduce 535) ((T_LOGICAL_OR) reduce 535) ((T_LOGICAL_XOR) reduce 535) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PIPE) reduce 535) ((T_POW) reduce 535) ((T_SL) reduce 535) ((T_SPACESHIP) reduce 535) ((T_SR) reduce 535) (("[") reduce 535) (("^") reduce 535) (("|") reduce 535)) 4011 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) (("=") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AND_EQUAL) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_COALESCE_EQUAL) reduce 568) ((T_CONCAT_EQUAL) reduce 568) ((T_DEC) reduce 568) ((T_DIV_EQUAL) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INC) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_MINUS_EQUAL) reduce 568) ((T_MOD_EQUAL) reduce 568) ((T_MUL_EQUAL) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_OR_EQUAL) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_PLUS_EQUAL) reduce 568) ((T_POW) reduce 568) ((T_POW_EQUAL) reduce 568) ((T_SL) reduce 568) ((T_SL_EQUAL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) ((T_SR_EQUAL) reduce 568) ((T_XOR_EQUAL) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 4012 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 479) (("(") shift) (("*") reduce 479) (("+") shift) ((",") reduce 479) (("-") shift) ((".") reduce 479) (("/") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 479) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 479) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 479) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PIPE) reduce 479) ((T_POW) reduce 479) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 479) (("`") shift) (("|") reduce 479) (("~") shift)) 4019 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) ((",") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) (("<") reduce 564) (("=") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AND_EQUAL) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_COALESCE_EQUAL) reduce 564) ((T_CONCAT_EQUAL) reduce 564) ((T_DEC) reduce 564) ((T_DIV_EQUAL) reduce 564) ((T_DOUBLE_ARROW) reduce 564) ((T_INC) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_MINUS_EQUAL) reduce 564) ((T_MOD_EQUAL) reduce 564) ((T_MUL_EQUAL) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_OR_EQUAL) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_PLUS_EQUAL) reduce 564) ((T_POW) reduce 564) ((T_POW_EQUAL) reduce 564) ((T_SL) reduce 564) ((T_SL_EQUAL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) ((T_SR_EQUAL) reduce 564) ((T_XOR_EQUAL) reduce 564) (("[") reduce 564) (("^") reduce 564) (("|") reduce 564)) 4020 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PIPE) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 551) (("^") reduce 530) (("|") reduce 530)) 4022 ((("%") reduce 529) (("*") reduce 529) (("+") reduce 529) ((",") reduce 529) (("-") reduce 529) ((".") reduce 529) (("/") reduce 529) (("<") reduce 529) ((">") reduce 529) (("?") reduce 529) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_BOOLEAN_AND) reduce 529) ((T_BOOLEAN_OR) reduce 529) ((T_COALESCE) reduce 529) ((T_DOUBLE_ARROW) reduce 529) ((T_INSTANCEOF) reduce 529) ((T_IS_EQUAL) reduce 529) ((T_IS_GREATER_OR_EQUAL) reduce 529) ((T_IS_IDENTICAL) reduce 529) ((T_IS_NOT_EQUAL) reduce 529) ((T_IS_NOT_IDENTICAL) reduce 529) ((T_IS_SMALLER_OR_EQUAL) reduce 529) ((T_LOGICAL_AND) reduce 529) ((T_LOGICAL_OR) reduce 529) ((T_LOGICAL_XOR) reduce 529) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PIPE) reduce 529) ((T_POW) reduce 529) ((T_SL) reduce 529) ((T_SPACESHIP) reduce 529) ((T_SR) reduce 529) (("[") reduce 554) (("^") reduce 529) (("|") reduce 529)) 4023 ((("%") reduce 528) (("(") reduce 557) (("*") reduce 528) (("+") reduce 528) ((",") reduce 528) (("-") reduce 528) ((".") reduce 528) (("/") reduce 528) (("<") reduce 528) ((">") reduce 528) (("?") reduce 528) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_BOOLEAN_AND) reduce 528) ((T_BOOLEAN_OR) reduce 528) ((T_COALESCE) reduce 528) ((T_DOUBLE_ARROW) reduce 528) ((T_INSTANCEOF) reduce 528) ((T_IS_EQUAL) reduce 528) ((T_IS_GREATER_OR_EQUAL) reduce 528) ((T_IS_IDENTICAL) reduce 528) ((T_IS_NOT_EQUAL) reduce 528) ((T_IS_NOT_IDENTICAL) reduce 528) ((T_IS_SMALLER_OR_EQUAL) reduce 528) ((T_LOGICAL_AND) reduce 528) ((T_LOGICAL_OR) reduce 528) ((T_LOGICAL_XOR) reduce 528) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PIPE) reduce 528) ((T_POW) reduce 528) ((T_SL) reduce 528) ((T_SPACESHIP) reduce 528) ((T_SR) reduce 528) (("[") reduce 550) (("^") reduce 528) (("|") reduce 528)) 4024 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 240) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 240) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4027 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) ((",") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) (("<") reduce 563) (("=") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AND_EQUAL) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_COALESCE_EQUAL) reduce 563) ((T_CONCAT_EQUAL) reduce 563) ((T_DEC) reduce 563) ((T_DIV_EQUAL) reduce 563) ((T_DOUBLE_ARROW) reduce 563) ((T_INC) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_MINUS_EQUAL) reduce 563) ((T_MOD_EQUAL) reduce 563) ((T_MUL_EQUAL) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_OR_EQUAL) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_PLUS_EQUAL) reduce 563) ((T_POW) reduce 563) ((T_POW_EQUAL) reduce 563) ((T_SL) reduce 563) ((T_SL_EQUAL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) ((T_SR_EQUAL) reduce 563) ((T_XOR_EQUAL) reduce 563) (("[") reduce 563) (("^") reduce 563) (("|") reduce 563)) 4028 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) ((",") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_DOUBLE_ARROW) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_PIPE) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("^") reduce 484) (("|") reduce 484)) 4029 ((("%") reduce 466) (("*") reduce 466) (("+") reduce 466) ((",") reduce 466) (("-") reduce 466) ((".") reduce 466) (("/") reduce 466) (("<") reduce 466) ((">") reduce 466) (("?") reduce 466) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_BOOLEAN_AND) reduce 466) ((T_BOOLEAN_OR) reduce 466) ((T_COALESCE) reduce 466) ((T_DOUBLE_ARROW) reduce 466) ((T_INSTANCEOF) reduce 466) ((T_IS_EQUAL) reduce 466) ((T_IS_GREATER_OR_EQUAL) reduce 466) ((T_IS_IDENTICAL) reduce 466) ((T_IS_NOT_EQUAL) reduce 466) ((T_IS_NOT_IDENTICAL) reduce 466) ((T_IS_SMALLER_OR_EQUAL) reduce 466) ((T_LOGICAL_AND) reduce 466) ((T_LOGICAL_OR) reduce 466) ((T_LOGICAL_XOR) reduce 466) ((T_PIPE) reduce 466) ((T_POW) reduce 466) ((T_SL) reduce 466) ((T_SPACESHIP) reduce 466) ((T_SR) reduce 466) (("^") reduce 466) (("|") reduce 466)) 4030 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_PIPE) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("^") reduce 488) (("|") reduce 488)) 4031 (((",") reduce 236) (("}") reduce 236)) 4034 ((("%") reduce 531) (("(") shift) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("|") reduce 531)) 4035 ((("%") reduce 461) (("(") reduce 558) (("*") reduce 461) (("+") reduce 461) ((",") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PIPE) reduce 461) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") reduce 552) (("^") reduce 461) (("|") reduce 461)) 4036 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) ((",") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_DOUBLE_ARROW) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_PIPE) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 4037 (((",") shift) (("}") reduce 125)) 4038 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) ((",") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_DOUBLE_ARROW) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_PIPE) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("|") reduce 476)) 4039 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) (("=") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AND_EQUAL) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_COALESCE_EQUAL) reduce 559) ((T_CONCAT_EQUAL) reduce 559) ((T_DEC) reduce 559) ((T_DIV_EQUAL) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INC) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_MINUS_EQUAL) reduce 559) ((T_MOD_EQUAL) reduce 559) ((T_MUL_EQUAL) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_OR_EQUAL) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_PLUS_EQUAL) reduce 559) ((T_POW) reduce 559) ((T_POW_EQUAL) reduce 559) ((T_SL) reduce 559) ((T_SL_EQUAL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) ((T_SR_EQUAL) reduce 559) ((T_XOR_EQUAL) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 4040 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 4041 ((("%") reduce 403) (("*") reduce 403) (("+") reduce 403) ((",") reduce 403) (("-") reduce 403) ((".") reduce 403) (("/") reduce 403) (("<") reduce 403) (("=") shift) ((">") reduce 403) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_DOUBLE_ARROW) reduce 403) ((T_INC) shift) ((T_INSTANCEOF) reduce 403) ((T_IS_EQUAL) reduce 403) ((T_IS_GREATER_OR_EQUAL) reduce 403) ((T_IS_IDENTICAL) reduce 403) ((T_IS_NOT_EQUAL) reduce 403) ((T_IS_NOT_IDENTICAL) reduce 403) ((T_IS_SMALLER_OR_EQUAL) reduce 403) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 403) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 403) ((T_POW_EQUAL) shift) ((T_SL) reduce 403) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 403) ((T_SR) reduce 403) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 548) (("^") reduce 403) (("|") reduce 403)) 4044 ((("%") reduce 448) (("*") reduce 448) (("+") reduce 448) ((",") reduce 448) (("-") reduce 448) ((".") reduce 448) (("/") reduce 448) (("<") reduce 448) ((">") reduce 448) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_BOOLEAN_AND) reduce 448) ((T_BOOLEAN_OR) reduce 448) ((T_COALESCE) reduce 448) ((T_DOUBLE_ARROW) reduce 448) ((T_INSTANCEOF) reduce 448) ((T_IS_EQUAL) reduce 448) ((T_IS_GREATER_OR_EQUAL) reduce 448) ((T_IS_IDENTICAL) reduce 448) ((T_IS_NOT_EQUAL) reduce 448) ((T_IS_NOT_IDENTICAL) reduce 448) ((T_IS_SMALLER_OR_EQUAL) reduce 448) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_PIPE) reduce 448) ((T_POW) shift) ((T_SL) reduce 448) ((T_SPACESHIP) reduce 448) ((T_SR) reduce 448) (("^") reduce 448) (("|") reduce 448)) 4076 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 432) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_DOUBLE_ARROW) reduce 432) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 432)) 4077 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 435) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_DOUBLE_ARROW) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("|") reduce 435)) 4078 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 444) (("-") shift) ((".") reduce 444) (("/") shift) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_PIPE) reduce 444) ((T_POW) shift) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("|") reduce 444)) 4079 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 458) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 458) (("|") reduce 458)) 4080 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 443) (("-") shift) ((".") reduce 443) (("/") shift) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_DOUBLE_ARROW) reduce 443) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_PIPE) reduce 443) ((T_POW) shift) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("^") reduce 443) (("|") reduce 443)) 4081 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) ((",") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_DOUBLE_ARROW) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_PIPE) reduce 440) ((T_POW) shift) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("|") reduce 440)) 4082 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 453) (("-") shift) ((".") shift) (("/") shift) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_DOUBLE_ARROW) reduce 453) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_PIPE) reduce 453) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 453) ((T_SR) shift) (("^") reduce 453) (("|") reduce 453)) 4083 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 431) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4084 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 429) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 429) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4085 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 430) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 430) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4086 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 455) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 455) ((T_SR) shift) (("^") reduce 455) (("|") reduce 455)) 4087 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 450) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_DOUBLE_ARROW) reduce 450) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 450) (("|") reduce 450)) 4088 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 452) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_DOUBLE_ARROW) reduce 452) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 452) (("|") reduce 452)) 4089 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 449) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_DOUBLE_ARROW) reduce 449) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 449) (("|") reduce 449)) 4090 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 457) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 457) ((T_SR) shift) (("^") reduce 457) (("|") reduce 457)) 4091 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 451) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_DOUBLE_ARROW) reduce 451) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 451) (("|") reduce 451)) 4094 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 4095 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 4096 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 4097 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 4098 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 4099 ((("%") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 4100 ((("%") reduce 511) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("|") reduce 511)) 4101 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PIPE) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 4102 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 4103 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("|") reduce 512)) 4104 ((("%") reduce 573) (("*") reduce 573) (("+") reduce 573) ((",") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_DOUBLE_ARROW) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("|") reduce 573)) 4110 ((("%") reduce 574) (("*") reduce 574) (("+") reduce 574) ((",") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_DOUBLE_ARROW) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("|") reduce 574)) 4111 ((("%") reduce 578) (("*") reduce 578) (("+") reduce 578) ((",") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_DOUBLE_ARROW) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("|") reduce 578)) 4112 ((("%") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 4113 ((("%") reduce 575) (("*") reduce 575) (("+") reduce 575) ((",") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_DOUBLE_ARROW) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("|") reduce 575)) 4114 ((("%") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 4117 ((("%") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 4118 ((("%") reduce 576) (("*") reduce 576) (("+") reduce 576) ((",") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_DOUBLE_ARROW) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("|") reduce 576)) 4120 ((("%") reduce 577) (("*") reduce 577) (("+") reduce 577) ((",") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_DOUBLE_ARROW) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("|") reduce 577)) 4122 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("|") reduce 513)) 4123 ((("%") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 4126 ((("%") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 4127 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 465) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 465) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 465) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4128 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 427) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_DOUBLE_ARROW) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4129 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 428) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_DOUBLE_ARROW) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4130 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 433) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_DOUBLE_ARROW) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("|") reduce 433)) 4131 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 434) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_DOUBLE_ARROW) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("|") reduce 434)) 4135 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 463) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 463) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4136 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 464) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 464) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4137 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 456) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 456) ((T_SR) shift) (("^") reduce 456) (("|") reduce 456)) 4138 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 454) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 454) ((T_SR) shift) (("^") reduce 454) (("|") reduce 454)) 4139 ((("%") reduce 441) (("*") reduce 441) (("+") reduce 441) ((",") reduce 441) (("-") reduce 441) ((".") reduce 441) (("/") reduce 441) (("<") reduce 441) ((">") reduce 441) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_DOUBLE_ARROW) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 441) ((T_IS_GREATER_OR_EQUAL) reduce 441) ((T_IS_IDENTICAL) reduce 441) ((T_IS_NOT_EQUAL) reduce 441) ((T_IS_NOT_IDENTICAL) reduce 441) ((T_IS_SMALLER_OR_EQUAL) reduce 441) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_PIPE) reduce 441) ((T_POW) shift) ((T_SL) reduce 441) ((T_SPACESHIP) reduce 441) ((T_SR) reduce 441) (("^") reduce 441) (("|") reduce 441)) 4140 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 436) (("-") shift) ((".") reduce 436) (("/") shift) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_DOUBLE_ARROW) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_PIPE) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 436) ((T_SR) shift) (("^") reduce 436) (("|") reduce 436)) 4141 ((("%") shift) (("*") shift) (("+") reduce 438) ((",") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") shift) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_DOUBLE_ARROW) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_PIPE) reduce 438) ((T_POW) shift) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("|") reduce 438)) 4142 ((("%") shift) (("*") shift) (("+") reduce 437) ((",") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") shift) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_DOUBLE_ARROW) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_PIPE) reduce 437) ((T_POW) shift) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("|") reduce 437)) 4143 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) ((",") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_DOUBLE_ARROW) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_PIPE) reduce 439) ((T_POW) shift) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("|") reduce 439)) 4144 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) ((",") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_DOUBLE_ARROW) reduce 442) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_PIPE) reduce 442) ((T_POW) shift) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("|") reduce 442)) 4146 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) ((",") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_DOUBLE_ARROW) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_PIPE) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("|") reduce 74)) 4147 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) ((",") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_DOUBLE_ARROW) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_PIPE) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("|") reduce 52)) 4148 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) ((",") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_DOUBLE_ARROW) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_PIPE) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("|") reduce 27)) 4149 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) ((",") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_DOUBLE_ARROW) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_PIPE) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("|") reduce 51)) 4150 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) ((",") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_DOUBLE_ARROW) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_PIPE) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("|") reduce 53)) 4151 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) ((",") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_DOUBLE_ARROW) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_PIPE) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("|") reduce 49)) 4152 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) ((",") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_DOUBLE_ARROW) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_PIPE) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("|") reduce 29)) 4153 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) ((",") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_DOUBLE_ARROW) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_PIPE) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("|") reduce 59)) 4154 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) ((",") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_DOUBLE_ARROW) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_PIPE) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("|") reduce 60)) 4155 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) ((",") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_DOUBLE_ARROW) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_PIPE) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("|") reduce 11)) 4156 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) ((",") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_DOUBLE_ARROW) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_PIPE) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("|") reduce 42)) 4157 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) ((",") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_DOUBLE_ARROW) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_PIPE) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("|") reduce 39)) 4158 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) ((",") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_DOUBLE_ARROW) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_PIPE) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("|") reduce 25)) 4159 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) ((",") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_DOUBLE_ARROW) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_PIPE) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("|") reduce 50)) 4160 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) ((",") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_DOUBLE_ARROW) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_PIPE) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("|") reduce 66)) 4161 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) ((",") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_DOUBLE_ARROW) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_PIPE) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("|") reduce 18)) 4162 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) ((",") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_DOUBLE_ARROW) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_PIPE) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("|") reduce 17)) 4163 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) ((",") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_DOUBLE_ARROW) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_PIPE) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("|") reduce 15)) 4164 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) ((",") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_DOUBLE_ARROW) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_PIPE) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("|") reduce 14)) 4165 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) ((",") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_DOUBLE_ARROW) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_PIPE) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("|") reduce 38)) 4166 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) ((",") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_DOUBLE_ARROW) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_PIPE) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("|") reduce 26)) 4167 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) ((",") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_DOUBLE_ARROW) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_PIPE) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("|") reduce 22)) 4168 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) ((",") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_DOUBLE_ARROW) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_PIPE) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("|") reduce 24)) 4169 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) ((",") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_DOUBLE_ARROW) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_PIPE) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("|") reduce 16)) 4170 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) ((",") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_DOUBLE_ARROW) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_PIPE) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("|") reduce 48)) 4171 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) ((",") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_DOUBLE_ARROW) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_PIPE) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("|") reduce 20)) 4172 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) ((",") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_DOUBLE_ARROW) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_PIPE) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("|") reduce 70)) 4173 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) ((",") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_DOUBLE_ARROW) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_PIPE) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("|") reduce 3)) 4174 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) ((",") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_DOUBLE_ARROW) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_PIPE) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("|") reduce 12)) 4175 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) ((",") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_DOUBLE_ARROW) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_PIPE) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("|") reduce 54)) 4176 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) ((",") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_DOUBLE_ARROW) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_PIPE) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("|") reduce 65)) 4177 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) ((",") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_DOUBLE_ARROW) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_PIPE) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("|") reduce 75)) 4178 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) ((",") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_DOUBLE_ARROW) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_PIPE) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("|") reduce 30)) 4179 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) ((",") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_DOUBLE_ARROW) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_PIPE) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("|") reduce 68)) 4180 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) ((",") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_DOUBLE_ARROW) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_PIPE) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("|") reduce 21)) 4181 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) ((",") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_DOUBLE_ARROW) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_PIPE) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("|") reduce 23)) 4182 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) ((",") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_DOUBLE_ARROW) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_PIPE) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("|") reduce 41)) 4183 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) ((",") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_DOUBLE_ARROW) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_PIPE) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("|") reduce 62)) 4184 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) ((",") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_DOUBLE_ARROW) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_PIPE) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("|") reduce 34)) 4185 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) ((",") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_DOUBLE_ARROW) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_PIPE) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("|") reduce 40)) 4186 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) ((",") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_DOUBLE_ARROW) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_PIPE) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("|") reduce 13)) 4187 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) ((",") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_DOUBLE_ARROW) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_PIPE) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("|") reduce 55)) 4188 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) ((",") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_DOUBLE_ARROW) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_PIPE) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("|") reduce 1)) 4189 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) ((",") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_DOUBLE_ARROW) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_PIPE) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("|") reduce 2)) 4190 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) ((",") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_DOUBLE_ARROW) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_PIPE) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("|") reduce 9)) 4191 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) ((",") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_DOUBLE_ARROW) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_PIPE) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("|") reduce 33)) 4192 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) ((",") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_DOUBLE_ARROW) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_PIPE) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("|") reduce 58)) 4193 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) ((",") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_DOUBLE_ARROW) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_PIPE) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("|") reduce 37)) 4194 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) ((",") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_DOUBLE_ARROW) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_PIPE) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("|") reduce 64)) 4195 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) ((",") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_DOUBLE_ARROW) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_PIPE) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("|") reduce 46)) 4196 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) ((",") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_DOUBLE_ARROW) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_PIPE) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("|") reduce 8)) 4197 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) ((",") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_DOUBLE_ARROW) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_PIPE) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("|") reduce 6)) 4198 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) ((",") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_DOUBLE_ARROW) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_PIPE) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("|") reduce 7)) 4199 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) ((",") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_DOUBLE_ARROW) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_PIPE) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("|") reduce 69)) 4200 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) ((",") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_DOUBLE_ARROW) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_PIPE) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("|") reduce 63)) 4201 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) ((",") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_DOUBLE_ARROW) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_PIPE) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("|") reduce 56)) 4202 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) ((",") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_DOUBLE_ARROW) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_PIPE) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("|") reduce 10)) 4203 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) ((",") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_DOUBLE_ARROW) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_PIPE) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("|") reduce 67)) 4204 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) ((",") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_DOUBLE_ARROW) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_PIPE) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("|") reduce 44)) 4205 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) ((",") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_DOUBLE_ARROW) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_PIPE) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("|") reduce 76)) 4206 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) ((",") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_DOUBLE_ARROW) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_PIPE) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("|") reduce 71)) 4207 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) ((",") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_DOUBLE_ARROW) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_PIPE) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("|") reduce 77)) 4208 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) ((",") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_DOUBLE_ARROW) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_PIPE) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("|") reduce 78)) 4209 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) ((",") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_DOUBLE_ARROW) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_PIPE) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("|") reduce 79)) 4210 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) ((",") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_DOUBLE_ARROW) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_PIPE) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("|") reduce 4)) 4211 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) ((",") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_DOUBLE_ARROW) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_PIPE) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("|") reduce 5)) 4212 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) ((",") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_DOUBLE_ARROW) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_PIPE) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("|") reduce 43)) 4213 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) ((",") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_DOUBLE_ARROW) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_PIPE) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("|") reduce 73)) 4214 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) ((",") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_DOUBLE_ARROW) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_PIPE) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("|") reduce 82)) 4215 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) ((",") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_DOUBLE_ARROW) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_PIPE) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("|") reduce 47)) 4216 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) ((",") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_DOUBLE_ARROW) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_PIPE) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("|") reduce 31)) 4217 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) ((",") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_DOUBLE_ARROW) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_PIPE) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("|") reduce 57)) 4218 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) ((",") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_DOUBLE_ARROW) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_PIPE) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("|") reduce 61)) 4219 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) ((",") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_DOUBLE_ARROW) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_PIPE) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("|") reduce 28)) 4220 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) ((",") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_DOUBLE_ARROW) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_PIPE) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("|") reduce 36)) 4221 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) ((",") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_DOUBLE_ARROW) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_PIPE) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("|") reduce 32)) 4222 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) ((",") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_DOUBLE_ARROW) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_PIPE) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("|") reduce 35)) 4223 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) ((",") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_DOUBLE_ARROW) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_PIPE) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("|") reduce 19)) 4224 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) ((",") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_DOUBLE_ARROW) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_PIPE) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("|") reduce 45)) 4225 ((("%") reduce 542) (("(") reduce 579) (("*") reduce 542) (("+") reduce 542) ((",") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PIPE) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 4227 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) ((",") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_DOUBLE_ARROW) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_PIPE) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("|") reduce 72)) 4228 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) ((",") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_DOUBLE_ARROW) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_PIPE) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("|") reduce 83)) 4229 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) ((",") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) (("<") reduce 572) (("=") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AND_EQUAL) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_COALESCE_EQUAL) reduce 572) ((T_CONCAT_EQUAL) reduce 572) ((T_DEC) reduce 572) ((T_DIV_EQUAL) reduce 572) ((T_DOUBLE_ARROW) reduce 572) ((T_INC) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_MINUS_EQUAL) reduce 572) ((T_MOD_EQUAL) reduce 572) ((T_MUL_EQUAL) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_OR_EQUAL) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_PLUS_EQUAL) reduce 572) ((T_POW) reduce 572) ((T_POW_EQUAL) reduce 572) ((T_SL) reduce 572) ((T_SL_EQUAL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) ((T_SR_EQUAL) reduce 572) ((T_XOR_EQUAL) reduce 572) (("[") reduce 572) (("^") reduce 572) (("|") reduce 572)) 4232 ((("%") reduce 544) (("(") reduce 580) (("*") reduce 544) (("+") reduce 544) ((",") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PIPE) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 4234 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("|") reduce 507)) 4235 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 4239 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) (("=") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AND_EQUAL) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_COALESCE_EQUAL) reduce 292) ((T_CONCAT_EQUAL) reduce 292) ((T_DEC) reduce 292) ((T_DIV_EQUAL) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INC) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_MINUS_EQUAL) reduce 292) ((T_MOD_EQUAL) reduce 292) ((T_MUL_EQUAL) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_OR_EQUAL) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_PLUS_EQUAL) reduce 292) ((T_POW) reduce 292) ((T_POW_EQUAL) reduce 292) ((T_SL) reduce 292) ((T_SL_EQUAL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) ((T_SR_EQUAL) reduce 292) ((T_XOR_EQUAL) reduce 292) (("[") reduce 292) (("^") reduce 292) (("|") reduce 292)) 4240 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) (("=") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AND_EQUAL) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_COALESCE_EQUAL) reduce 293) ((T_CONCAT_EQUAL) reduce 293) ((T_DEC) reduce 293) ((T_DIV_EQUAL) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INC) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_MINUS_EQUAL) reduce 293) ((T_MOD_EQUAL) reduce 293) ((T_MUL_EQUAL) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_OR_EQUAL) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_PLUS_EQUAL) reduce 293) ((T_POW) reduce 293) ((T_POW_EQUAL) reduce 293) ((T_SL) reduce 293) ((T_SL_EQUAL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) ((T_SR_EQUAL) reduce 293) ((T_XOR_EQUAL) reduce 293) (("[") reduce 293) (("^") reduce 293) (("|") reduce 293)) 4245 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) ((",") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_DOUBLE_ARROW) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_PIPE) reduce 425) ((T_POW) reduce 425) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 4247 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) ((",") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_DOUBLE_ARROW) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_PIPE) reduce 423) ((T_POW) reduce 423) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 4257 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 419) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4258 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 421) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4259 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 420) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4260 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 413) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 413) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4261 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 410) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 410) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4262 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 418) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4263 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 412) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4264 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 416) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4265 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 411) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4266 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 414) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 414) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4267 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 415) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 415) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4268 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 422) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4269 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 417) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4271 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 406) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 4274 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 4277 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) ((",") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) (("<") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_DOUBLE_ARROW) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_POW) reduce 564) ((T_SL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) (("[") reduce 564) (("^") reduce 564) (("|") reduce 564)) 4279 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) ((",") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) (("<") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_DOUBLE_ARROW) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_POW) reduce 563) ((T_SL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) (("[") reduce 563) (("^") reduce 563) (("|") reduce 563)) 4281 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 4282 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 4283 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) ((",") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_DOUBLE_ARROW) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("[") reduce 548) (("^") reduce 407) (("|") reduce 407)) 4287 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) ((",") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) (("<") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_DOUBLE_ARROW) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_POW) reduce 572) ((T_SL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) (("[") reduce 572) (("^") reduce 572) (("|") reduce 572)) 4289 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("|") reduce 507)) 4290 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 4294 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("[") reduce 292) (("^") reduce 292) (("|") reduce 292)) 4295 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("[") reduce 293) (("^") reduce 293) (("|") reduce 293)) 4296 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 4299 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) ((",") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) (("<") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_DOUBLE_ARROW) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_POW) reduce 571) ((T_SL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) (("[") reduce 571) (("^") reduce 571) (("|") reduce 571)) 4300 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("|") reduce 506)) 4301 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("|") reduce 508)) 4306 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 4307 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 4308 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 4309 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 4312 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 4313 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 4314 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 4315 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) ((",") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("[") reduce 562) (("^") reduce 562) (("|") reduce 562)) 4316 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 4317 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 4320 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 4321 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DEFAULT) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("}") reduce 126) (("~") shift)) 4322 ((("}") reduce 235)) 4323 (((",") reduce 237) (("}") reduce 237)) 4324 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) (("=") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AND_EQUAL) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_COALESCE_EQUAL) reduce 504) ((T_CONCAT_EQUAL) reduce 504) ((T_DEC) reduce 504) ((T_DIV_EQUAL) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INC) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_MINUS_EQUAL) reduce 504) ((T_MOD_EQUAL) reduce 504) ((T_MUL_EQUAL) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_OR_EQUAL) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_PLUS_EQUAL) reduce 504) ((T_POW) reduce 504) ((T_POW_EQUAL) reduce 504) ((T_SL) reduce 504) ((T_SL_EQUAL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) ((T_SR_EQUAL) reduce 504) ((T_XOR_EQUAL) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 4325 ((("%") reduce 233) (("*") reduce 233) (("+") reduce 233) ((",") reduce 233) (("-") reduce 233) ((".") reduce 233) (("/") reduce 233) ((";") reduce 233) (("<") reduce 233) ((">") reduce 233) (("?") reduce 233) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_BOOLEAN_AND) reduce 233) ((T_BOOLEAN_OR) reduce 233) ((T_COALESCE) reduce 233) ((T_DOUBLE_ARROW) reduce 233) ((T_INSTANCEOF) reduce 233) ((T_IS_EQUAL) reduce 233) ((T_IS_GREATER_OR_EQUAL) reduce 233) ((T_IS_IDENTICAL) reduce 233) ((T_IS_NOT_EQUAL) reduce 233) ((T_IS_NOT_IDENTICAL) reduce 233) ((T_IS_SMALLER_OR_EQUAL) reduce 233) ((T_LOGICAL_AND) reduce 233) ((T_LOGICAL_OR) reduce 233) ((T_LOGICAL_XOR) reduce 233) ((T_PIPE) reduce 233) ((T_POW) reduce 233) ((T_SL) reduce 233) ((T_SPACESHIP) reduce 233) ((T_SR) reduce 233) (("^") reduce 233) (("|") reduce 233)) 4326 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 126) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 4339 ((("%") reduce 540) (("*") reduce 540) (("+") reduce 540) ((",") reduce 540) (("-") reduce 540) ((".") reduce 540) (("/") reduce 540) (("<") reduce 540) ((">") reduce 540) (("?") reduce 540) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_BOOLEAN_AND) reduce 540) ((T_BOOLEAN_OR) reduce 540) ((T_COALESCE) reduce 540) ((T_INSTANCEOF) reduce 540) ((T_IS_EQUAL) reduce 540) ((T_IS_GREATER_OR_EQUAL) reduce 540) ((T_IS_IDENTICAL) reduce 540) ((T_IS_NOT_EQUAL) reduce 540) ((T_IS_NOT_IDENTICAL) reduce 540) ((T_IS_SMALLER_OR_EQUAL) reduce 540) ((T_LOGICAL_AND) reduce 540) ((T_LOGICAL_OR) reduce 540) ((T_LOGICAL_XOR) reduce 540) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 540) ((T_OBJECT_OPERATOR) reduce 540) ((T_PIPE) reduce 540) ((T_POW) reduce 540) ((T_SL) reduce 540) ((T_SPACESHIP) reduce 540) ((T_SR) reduce 540) (("[") reduce 540) (("^") reduce 540) (("|") reduce 540) (("}") reduce 540)) 4341 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_PIPE) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("|") reduce 521) (("}") reduce 521)) 4343 ((("%") reduce 534) (("*") reduce 534) (("+") reduce 534) ((",") reduce 534) (("-") reduce 534) ((".") reduce 534) (("/") reduce 534) (("<") reduce 534) ((">") reduce 534) (("?") reduce 534) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_BOOLEAN_AND) reduce 534) ((T_BOOLEAN_OR) reduce 534) ((T_COALESCE) reduce 534) ((T_INSTANCEOF) reduce 534) ((T_IS_EQUAL) reduce 534) ((T_IS_GREATER_OR_EQUAL) reduce 534) ((T_IS_IDENTICAL) reduce 534) ((T_IS_NOT_EQUAL) reduce 534) ((T_IS_NOT_IDENTICAL) reduce 534) ((T_IS_SMALLER_OR_EQUAL) reduce 534) ((T_LOGICAL_AND) reduce 534) ((T_LOGICAL_OR) reduce 534) ((T_LOGICAL_XOR) reduce 534) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PIPE) reduce 534) ((T_POW) reduce 534) ((T_SL) reduce 534) ((T_SPACESHIP) reduce 534) ((T_SR) reduce 534) (("[") reduce 534) (("^") reduce 534) (("|") reduce 534) (("}") reduce 534)) 4344 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_PIPE) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("^") reduce 524) (("|") reduce 524) (("}") reduce 524)) 4348 ((("%") reduce 517) (("(") shift) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_PIPE) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("^") reduce 517) (("|") reduce 517) (("}") reduce 517)) 4349 ((("%") reduce 533) (("*") reduce 533) (("+") reduce 533) ((",") reduce 533) (("-") reduce 533) ((".") reduce 533) (("/") reduce 533) (("<") reduce 533) ((">") reduce 533) (("?") reduce 533) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_BOOLEAN_AND) reduce 533) ((T_BOOLEAN_OR) reduce 533) ((T_COALESCE) reduce 533) ((T_INSTANCEOF) reduce 533) ((T_IS_EQUAL) reduce 533) ((T_IS_GREATER_OR_EQUAL) reduce 533) ((T_IS_IDENTICAL) reduce 533) ((T_IS_NOT_EQUAL) reduce 533) ((T_IS_NOT_IDENTICAL) reduce 533) ((T_IS_SMALLER_OR_EQUAL) reduce 533) ((T_LOGICAL_AND) reduce 533) ((T_LOGICAL_OR) reduce 533) ((T_LOGICAL_XOR) reduce 533) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PIPE) reduce 533) ((T_POW) reduce 533) ((T_SL) reduce 533) ((T_SPACESHIP) reduce 533) ((T_SR) reduce 533) (("[") reduce 533) (("^") reduce 533) (("|") reduce 533) (("}") reduce 533)) 4350 ((("%") reduce 537) (("*") reduce 537) (("+") reduce 537) ((",") reduce 537) (("-") reduce 537) ((".") reduce 537) (("/") reduce 537) (("<") reduce 537) ((">") reduce 537) (("?") reduce 537) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_BOOLEAN_AND) reduce 537) ((T_BOOLEAN_OR) reduce 537) ((T_COALESCE) reduce 537) ((T_INSTANCEOF) reduce 537) ((T_IS_EQUAL) reduce 537) ((T_IS_GREATER_OR_EQUAL) reduce 537) ((T_IS_IDENTICAL) reduce 537) ((T_IS_NOT_EQUAL) reduce 537) ((T_IS_NOT_IDENTICAL) reduce 537) ((T_IS_SMALLER_OR_EQUAL) reduce 537) ((T_LOGICAL_AND) reduce 537) ((T_LOGICAL_OR) reduce 537) ((T_LOGICAL_XOR) reduce 537) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_PIPE) reduce 537) ((T_POW) reduce 537) ((T_SL) reduce 537) ((T_SPACESHIP) reduce 537) ((T_SR) reduce 537) (("[") reduce 537) (("^") reduce 537) (("|") reduce 537) (("}") reduce 537)) 4356 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PIPE) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("|") reduce 532) (("}") reduce 532)) 4358 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_PIPE) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("^") reduce 523) (("|") reduce 523) (("}") reduce 523)) 4360 ((("%") reduce 536) (("*") reduce 536) (("+") reduce 536) ((",") reduce 536) (("-") reduce 536) ((".") reduce 536) (("/") reduce 536) (("<") reduce 536) ((">") reduce 536) (("?") reduce 536) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_BOOLEAN_AND) reduce 536) ((T_BOOLEAN_OR) reduce 536) ((T_COALESCE) reduce 536) ((T_INSTANCEOF) reduce 536) ((T_IS_EQUAL) reduce 536) ((T_IS_GREATER_OR_EQUAL) reduce 536) ((T_IS_IDENTICAL) reduce 536) ((T_IS_NOT_EQUAL) reduce 536) ((T_IS_NOT_IDENTICAL) reduce 536) ((T_IS_SMALLER_OR_EQUAL) reduce 536) ((T_LOGICAL_AND) reduce 536) ((T_LOGICAL_OR) reduce 536) ((T_LOGICAL_XOR) reduce 536) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 536) ((T_OBJECT_OPERATOR) reduce 536) ((T_PIPE) reduce 536) ((T_POW) reduce 536) ((T_SL) reduce 536) ((T_SPACESHIP) reduce 536) ((T_SR) reduce 536) (("[") reduce 536) (("^") reduce 536) (("|") reduce 536) (("}") reduce 536)) 4361 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("|") reduce 94) (("}") reduce 94)) 4362 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("|") reduce 93) (("}") reduce 93)) 4363 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("|") reduce 95) (("}") reduce 95)) 4365 ((("%") reduce 539) (("*") reduce 539) (("+") reduce 539) ((",") reduce 539) (("-") reduce 539) ((".") reduce 539) (("/") reduce 539) (("<") reduce 539) ((">") reduce 539) (("?") reduce 539) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_BOOLEAN_AND) reduce 539) ((T_BOOLEAN_OR) reduce 539) ((T_COALESCE) reduce 539) ((T_INSTANCEOF) reduce 539) ((T_IS_EQUAL) reduce 539) ((T_IS_GREATER_OR_EQUAL) reduce 539) ((T_IS_IDENTICAL) reduce 539) ((T_IS_NOT_EQUAL) reduce 539) ((T_IS_NOT_IDENTICAL) reduce 539) ((T_IS_SMALLER_OR_EQUAL) reduce 539) ((T_LOGICAL_AND) reduce 539) ((T_LOGICAL_OR) reduce 539) ((T_LOGICAL_XOR) reduce 539) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 539) ((T_OBJECT_OPERATOR) reduce 539) ((T_PIPE) reduce 539) ((T_POW) reduce 539) ((T_SL) reduce 539) ((T_SPACESHIP) reduce 539) ((T_SR) reduce 539) (("[") reduce 539) (("^") reduce 539) (("|") reduce 539) (("}") reduce 539)) 4368 ((("%") reduce 538) (("*") reduce 538) (("+") reduce 538) ((",") reduce 538) (("-") reduce 538) ((".") reduce 538) (("/") reduce 538) (("<") reduce 538) ((">") reduce 538) (("?") reduce 538) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_BOOLEAN_AND) reduce 538) ((T_BOOLEAN_OR) reduce 538) ((T_COALESCE) reduce 538) ((T_INSTANCEOF) reduce 538) ((T_IS_EQUAL) reduce 538) ((T_IS_GREATER_OR_EQUAL) reduce 538) ((T_IS_IDENTICAL) reduce 538) ((T_IS_NOT_EQUAL) reduce 538) ((T_IS_NOT_IDENTICAL) reduce 538) ((T_IS_SMALLER_OR_EQUAL) reduce 538) ((T_LOGICAL_AND) reduce 538) ((T_LOGICAL_OR) reduce 538) ((T_LOGICAL_XOR) reduce 538) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 538) ((T_OBJECT_OPERATOR) reduce 538) ((T_PIPE) reduce 538) ((T_POW) reduce 538) ((T_SL) reduce 538) ((T_SPACESHIP) reduce 538) ((T_SR) reduce 538) (("[") reduce 538) (("^") reduce 538) (("|") reduce 538) (("}") reduce 538)) 4374 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92) (("}") reduce 92)) 4377 ((("%") reduce 535) (("*") reduce 535) (("+") reduce 535) ((",") reduce 535) (("-") reduce 535) ((".") reduce 535) (("/") reduce 535) (("<") reduce 535) ((">") reduce 535) (("?") reduce 535) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_BOOLEAN_AND) reduce 535) ((T_BOOLEAN_OR) reduce 535) ((T_COALESCE) reduce 535) ((T_INSTANCEOF) reduce 535) ((T_IS_EQUAL) reduce 535) ((T_IS_GREATER_OR_EQUAL) reduce 535) ((T_IS_IDENTICAL) reduce 535) ((T_IS_NOT_EQUAL) reduce 535) ((T_IS_NOT_IDENTICAL) reduce 535) ((T_IS_SMALLER_OR_EQUAL) reduce 535) ((T_LOGICAL_AND) reduce 535) ((T_LOGICAL_OR) reduce 535) ((T_LOGICAL_XOR) reduce 535) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PIPE) reduce 535) ((T_POW) reduce 535) ((T_SL) reduce 535) ((T_SPACESHIP) reduce 535) ((T_SR) reduce 535) (("[") reduce 535) (("^") reduce 535) (("|") reduce 535) (("}") reduce 535)) 4379 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) (("=") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AND_EQUAL) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_COALESCE_EQUAL) reduce 568) ((T_CONCAT_EQUAL) reduce 568) ((T_DEC) reduce 568) ((T_DIV_EQUAL) reduce 568) ((T_INC) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_MINUS_EQUAL) reduce 568) ((T_MOD_EQUAL) reduce 568) ((T_MUL_EQUAL) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_OR_EQUAL) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_PLUS_EQUAL) reduce 568) ((T_POW) reduce 568) ((T_POW_EQUAL) reduce 568) ((T_SL) reduce 568) ((T_SL_EQUAL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) ((T_SR_EQUAL) reduce 568) ((T_XOR_EQUAL) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568) (("}") reduce 568)) 4380 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 479) (("(") shift) (("*") reduce 479) (("+") shift) ((",") reduce 479) (("-") shift) ((".") reduce 479) (("/") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 479) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 479) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PIPE) reduce 479) ((T_POW) reduce 479) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 479) (("`") shift) (("|") reduce 479) (("}") reduce 479) (("~") shift)) 4387 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) ((",") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) (("<") reduce 564) (("=") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AND_EQUAL) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_COALESCE_EQUAL) reduce 564) ((T_CONCAT_EQUAL) reduce 564) ((T_DEC) reduce 564) ((T_DIV_EQUAL) reduce 564) ((T_INC) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_MINUS_EQUAL) reduce 564) ((T_MOD_EQUAL) reduce 564) ((T_MUL_EQUAL) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_OR_EQUAL) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_PLUS_EQUAL) reduce 564) ((T_POW) reduce 564) ((T_POW_EQUAL) reduce 564) ((T_SL) reduce 564) ((T_SL_EQUAL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) ((T_SR_EQUAL) reduce 564) ((T_XOR_EQUAL) reduce 564) (("[") reduce 564) (("^") reduce 564) (("|") reduce 564) (("}") reduce 564)) 4388 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PIPE) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 551) (("^") reduce 530) (("|") reduce 530) (("}") reduce 530)) 4390 ((("%") reduce 529) (("*") reduce 529) (("+") reduce 529) ((",") reduce 529) (("-") reduce 529) ((".") reduce 529) (("/") reduce 529) (("<") reduce 529) ((">") reduce 529) (("?") reduce 529) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_BOOLEAN_AND) reduce 529) ((T_BOOLEAN_OR) reduce 529) ((T_COALESCE) reduce 529) ((T_INSTANCEOF) reduce 529) ((T_IS_EQUAL) reduce 529) ((T_IS_GREATER_OR_EQUAL) reduce 529) ((T_IS_IDENTICAL) reduce 529) ((T_IS_NOT_EQUAL) reduce 529) ((T_IS_NOT_IDENTICAL) reduce 529) ((T_IS_SMALLER_OR_EQUAL) reduce 529) ((T_LOGICAL_AND) reduce 529) ((T_LOGICAL_OR) reduce 529) ((T_LOGICAL_XOR) reduce 529) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PIPE) reduce 529) ((T_POW) reduce 529) ((T_SL) reduce 529) ((T_SPACESHIP) reduce 529) ((T_SR) reduce 529) (("[") reduce 554) (("^") reduce 529) (("|") reduce 529) (("}") reduce 529)) 4391 ((("%") reduce 528) (("(") reduce 557) (("*") reduce 528) (("+") reduce 528) ((",") reduce 528) (("-") reduce 528) ((".") reduce 528) (("/") reduce 528) (("<") reduce 528) ((">") reduce 528) (("?") reduce 528) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_BOOLEAN_AND) reduce 528) ((T_BOOLEAN_OR) reduce 528) ((T_COALESCE) reduce 528) ((T_INSTANCEOF) reduce 528) ((T_IS_EQUAL) reduce 528) ((T_IS_GREATER_OR_EQUAL) reduce 528) ((T_IS_IDENTICAL) reduce 528) ((T_IS_NOT_EQUAL) reduce 528) ((T_IS_NOT_IDENTICAL) reduce 528) ((T_IS_SMALLER_OR_EQUAL) reduce 528) ((T_LOGICAL_AND) reduce 528) ((T_LOGICAL_OR) reduce 528) ((T_LOGICAL_XOR) reduce 528) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PIPE) reduce 528) ((T_POW) reduce 528) ((T_SL) reduce 528) ((T_SPACESHIP) reduce 528) ((T_SR) reduce 528) (("[") reduce 550) (("^") reduce 528) (("|") reduce 528) (("}") reduce 528)) 4392 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 238) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 238)) 4395 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) ((",") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) (("<") reduce 563) (("=") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AND_EQUAL) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_COALESCE_EQUAL) reduce 563) ((T_CONCAT_EQUAL) reduce 563) ((T_DEC) reduce 563) ((T_DIV_EQUAL) reduce 563) ((T_INC) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_MINUS_EQUAL) reduce 563) ((T_MOD_EQUAL) reduce 563) ((T_MUL_EQUAL) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_OR_EQUAL) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_PLUS_EQUAL) reduce 563) ((T_POW) reduce 563) ((T_POW_EQUAL) reduce 563) ((T_SL) reduce 563) ((T_SL_EQUAL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) ((T_SR_EQUAL) reduce 563) ((T_XOR_EQUAL) reduce 563) (("[") reduce 563) (("^") reduce 563) (("|") reduce 563) (("}") reduce 563)) 4396 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) ((",") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_PIPE) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("^") reduce 484) (("|") reduce 484) (("}") reduce 484)) 4397 ((("%") reduce 466) (("*") reduce 466) (("+") reduce 466) ((",") reduce 466) (("-") reduce 466) ((".") reduce 466) (("/") reduce 466) (("<") reduce 466) ((">") reduce 466) (("?") reduce 466) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_BOOLEAN_AND) reduce 466) ((T_BOOLEAN_OR) reduce 466) ((T_COALESCE) reduce 466) ((T_INSTANCEOF) reduce 466) ((T_IS_EQUAL) reduce 466) ((T_IS_GREATER_OR_EQUAL) reduce 466) ((T_IS_IDENTICAL) reduce 466) ((T_IS_NOT_EQUAL) reduce 466) ((T_IS_NOT_IDENTICAL) reduce 466) ((T_IS_SMALLER_OR_EQUAL) reduce 466) ((T_LOGICAL_AND) reduce 466) ((T_LOGICAL_OR) reduce 466) ((T_LOGICAL_XOR) reduce 466) ((T_PIPE) reduce 466) ((T_POW) reduce 466) ((T_SL) reduce 466) ((T_SPACESHIP) reduce 466) ((T_SR) reduce 466) (("^") reduce 466) (("|") reduce 466) (("}") reduce 466)) 4398 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_PIPE) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("^") reduce 488) (("|") reduce 488) (("}") reduce 488)) 4399 ((("%") reduce 531) (("(") shift) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("|") reduce 531) (("}") reduce 531)) 4400 ((("%") reduce 461) (("(") reduce 558) (("*") reduce 461) (("+") reduce 461) ((",") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PIPE) reduce 461) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") reduce 552) (("^") reduce 461) (("|") reduce 461) (("}") reduce 461)) 4401 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) ((",") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_PIPE) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462) (("}") reduce 462)) 4402 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) ((",") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_PIPE) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("|") reduce 476) (("}") reduce 476)) 4403 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) (("=") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AND_EQUAL) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_COALESCE_EQUAL) reduce 559) ((T_CONCAT_EQUAL) reduce 559) ((T_DEC) reduce 559) ((T_DIV_EQUAL) reduce 559) ((T_INC) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_MINUS_EQUAL) reduce 559) ((T_MOD_EQUAL) reduce 559) ((T_MUL_EQUAL) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_OR_EQUAL) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_PLUS_EQUAL) reduce 559) ((T_POW) reduce 559) ((T_POW_EQUAL) reduce 559) ((T_SL) reduce 559) ((T_SL_EQUAL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) ((T_SR_EQUAL) reduce 559) ((T_XOR_EQUAL) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559) (("}") reduce 559)) 4404 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565) (("}") reduce 565)) 4405 ((("%") reduce 403) (("*") reduce 403) (("+") reduce 403) ((",") reduce 403) (("-") reduce 403) ((".") reduce 403) (("/") reduce 403) (("<") reduce 403) (("=") shift) ((">") reduce 403) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_INC) shift) ((T_INSTANCEOF) reduce 403) ((T_IS_EQUAL) reduce 403) ((T_IS_GREATER_OR_EQUAL) reduce 403) ((T_IS_IDENTICAL) reduce 403) ((T_IS_NOT_EQUAL) reduce 403) ((T_IS_NOT_IDENTICAL) reduce 403) ((T_IS_SMALLER_OR_EQUAL) reduce 403) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 403) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 403) ((T_POW_EQUAL) shift) ((T_SL) reduce 403) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 403) ((T_SR) reduce 403) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 548) (("^") reduce 403) (("|") reduce 403) (("}") reduce 403)) 4408 ((("%") reduce 448) (("*") reduce 448) (("+") reduce 448) ((",") reduce 448) (("-") reduce 448) ((".") reduce 448) (("/") reduce 448) (("<") reduce 448) ((">") reduce 448) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_BOOLEAN_AND) reduce 448) ((T_BOOLEAN_OR) reduce 448) ((T_COALESCE) reduce 448) ((T_INSTANCEOF) reduce 448) ((T_IS_EQUAL) reduce 448) ((T_IS_GREATER_OR_EQUAL) reduce 448) ((T_IS_IDENTICAL) reduce 448) ((T_IS_NOT_EQUAL) reduce 448) ((T_IS_NOT_IDENTICAL) reduce 448) ((T_IS_SMALLER_OR_EQUAL) reduce 448) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_PIPE) reduce 448) ((T_POW) shift) ((T_SL) reduce 448) ((T_SPACESHIP) reduce 448) ((T_SR) reduce 448) (("^") reduce 448) (("|") reduce 448) (("}") reduce 448)) 4440 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 432) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 432) (("}") reduce 432)) 4441 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 435) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("|") reduce 435) (("}") reduce 435)) 4442 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 444) (("-") shift) ((".") reduce 444) (("/") shift) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_PIPE) reduce 444) ((T_POW) shift) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("|") reduce 444) (("}") reduce 444)) 4443 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 458) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 458) (("|") reduce 458) (("}") reduce 458)) 4444 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 443) (("-") shift) ((".") reduce 443) (("/") shift) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_PIPE) reduce 443) ((T_POW) shift) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("^") reduce 443) (("|") reduce 443) (("}") reduce 443)) 4445 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) ((",") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_PIPE) reduce 440) ((T_POW) shift) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("|") reduce 440) (("}") reduce 440)) 4446 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 453) (("-") shift) ((".") shift) (("/") shift) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_PIPE) reduce 453) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 453) ((T_SR) shift) (("^") reduce 453) (("|") reduce 453) (("}") reduce 453)) 4447 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 431) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 431)) 4448 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 429) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 429)) 4449 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 430) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 430)) 4450 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 455) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 455) ((T_SR) shift) (("^") reduce 455) (("|") reduce 455) (("}") reduce 455)) 4451 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 450) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 450) (("|") reduce 450) (("}") reduce 450)) 4452 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 452) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 452) (("|") reduce 452) (("}") reduce 452)) 4453 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 449) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 449) (("|") reduce 449) (("}") reduce 449)) 4454 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 457) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 457) ((T_SR) shift) (("^") reduce 457) (("|") reduce 457) (("}") reduce 457)) 4455 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 451) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 451) (("|") reduce 451) (("}") reduce 451)) 4458 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94) (("}") reduce 94)) 4459 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93) (("}") reduce 93)) 4460 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95) (("}") reduce 95)) 4461 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509) (("}") reduce 509)) 4462 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92) (("}") reduce 92)) 4463 ((("%") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568) (("}") reduce 568)) 4464 ((("%") reduce 511) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("|") reduce 511) (("}") reduce 511)) 4465 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PIPE) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459) (("}") reduce 459)) 4466 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510) (("}") reduce 510)) 4467 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("|") reduce 512) (("}") reduce 512)) 4468 ((("%") reduce 573) (("*") reduce 573) (("+") reduce 573) ((",") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("|") reduce 573) (("}") reduce 573)) 4474 ((("%") reduce 574) (("*") reduce 574) (("+") reduce 574) ((",") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("|") reduce 574) (("}") reduce 574)) 4475 ((("%") reduce 578) (("*") reduce 578) (("+") reduce 578) ((",") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("|") reduce 578) (("}") reduce 578)) 4476 ((("%") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582) (("}") reduce 582)) 4477 ((("%") reduce 575) (("*") reduce 575) (("+") reduce 575) ((",") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("|") reduce 575) (("}") reduce 575)) 4478 ((("%") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584) (("}") reduce 584)) 4481 ((("%") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583) (("}") reduce 583)) 4482 ((("%") reduce 576) (("*") reduce 576) (("+") reduce 576) ((",") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("|") reduce 576) (("}") reduce 576)) 4484 ((("%") reduce 577) (("*") reduce 577) (("+") reduce 577) ((",") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("|") reduce 577) (("}") reduce 577)) 4486 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("|") reduce 513) (("}") reduce 513)) 4487 ((("%") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570) (("}") reduce 570)) 4490 ((("%") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569) (("}") reduce 569)) 4491 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 465) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 465) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 465)) 4492 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 427) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 427)) 4493 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 428) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 428)) 4494 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 433) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("|") reduce 433) (("}") reduce 433)) 4495 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 434) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("|") reduce 434) (("}") reduce 434)) 4499 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 463) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 463)) 4500 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 464) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 464)) 4501 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 456) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 456) ((T_SR) shift) (("^") reduce 456) (("|") reduce 456) (("}") reduce 456)) 4502 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 454) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 454) ((T_SR) shift) (("^") reduce 454) (("|") reduce 454) (("}") reduce 454)) 4503 ((("%") reduce 441) (("*") reduce 441) (("+") reduce 441) ((",") reduce 441) (("-") reduce 441) ((".") reduce 441) (("/") reduce 441) (("<") reduce 441) ((">") reduce 441) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 441) ((T_IS_GREATER_OR_EQUAL) reduce 441) ((T_IS_IDENTICAL) reduce 441) ((T_IS_NOT_EQUAL) reduce 441) ((T_IS_NOT_IDENTICAL) reduce 441) ((T_IS_SMALLER_OR_EQUAL) reduce 441) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_PIPE) reduce 441) ((T_POW) shift) ((T_SL) reduce 441) ((T_SPACESHIP) reduce 441) ((T_SR) reduce 441) (("^") reduce 441) (("|") reduce 441) (("}") reduce 441)) 4504 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 436) (("-") shift) ((".") reduce 436) (("/") shift) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_PIPE) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 436) ((T_SR) shift) (("^") reduce 436) (("|") reduce 436) (("}") reduce 436)) 4505 ((("%") shift) (("*") shift) (("+") reduce 438) ((",") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") shift) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_PIPE) reduce 438) ((T_POW) shift) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("|") reduce 438) (("}") reduce 438)) 4506 ((("%") shift) (("*") shift) (("+") reduce 437) ((",") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") shift) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_PIPE) reduce 437) ((T_POW) shift) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("|") reduce 437) (("}") reduce 437)) 4507 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) ((",") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_PIPE) reduce 439) ((T_POW) shift) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("|") reduce 439) (("}") reduce 439)) 4508 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) ((",") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_PIPE) reduce 442) ((T_POW) shift) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("|") reduce 442) (("}") reduce 442)) 4510 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) ((",") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_PIPE) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("|") reduce 74) (("}") reduce 74)) 4511 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) ((",") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_PIPE) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("|") reduce 52) (("}") reduce 52)) 4512 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) ((",") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_PIPE) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("|") reduce 27) (("}") reduce 27)) 4513 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) ((",") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_PIPE) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("|") reduce 51) (("}") reduce 51)) 4514 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) ((",") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_PIPE) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("|") reduce 53) (("}") reduce 53)) 4515 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) ((",") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_PIPE) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("|") reduce 49) (("}") reduce 49)) 4516 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) ((",") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_PIPE) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("|") reduce 29) (("}") reduce 29)) 4517 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) ((",") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_PIPE) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("|") reduce 59) (("}") reduce 59)) 4518 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) ((",") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_PIPE) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("|") reduce 60) (("}") reduce 60)) 4519 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) ((",") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_PIPE) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("|") reduce 11) (("}") reduce 11)) 4520 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) ((",") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_PIPE) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("|") reduce 42) (("}") reduce 42)) 4521 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) ((",") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_PIPE) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("|") reduce 39) (("}") reduce 39)) 4522 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) ((",") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_PIPE) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("|") reduce 25) (("}") reduce 25)) 4523 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) ((",") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_PIPE) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("|") reduce 50) (("}") reduce 50)) 4524 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) ((",") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_PIPE) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("|") reduce 66) (("}") reduce 66)) 4525 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) ((",") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_PIPE) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("|") reduce 18) (("}") reduce 18)) 4526 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) ((",") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_PIPE) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("|") reduce 17) (("}") reduce 17)) 4527 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) ((",") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_PIPE) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("|") reduce 15) (("}") reduce 15)) 4528 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) ((",") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_PIPE) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("|") reduce 14) (("}") reduce 14)) 4529 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) ((",") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_PIPE) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("|") reduce 38) (("}") reduce 38)) 4530 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) ((",") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_PIPE) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("|") reduce 26) (("}") reduce 26)) 4531 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) ((",") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_PIPE) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("|") reduce 22) (("}") reduce 22)) 4532 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) ((",") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_PIPE) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("|") reduce 24) (("}") reduce 24)) 4533 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) ((",") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_PIPE) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("|") reduce 16) (("}") reduce 16)) 4534 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) ((",") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_PIPE) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("|") reduce 48) (("}") reduce 48)) 4535 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) ((",") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_PIPE) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("|") reduce 20) (("}") reduce 20)) 4536 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) ((",") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_PIPE) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("|") reduce 70) (("}") reduce 70)) 4537 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) ((",") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_PIPE) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("|") reduce 3) (("}") reduce 3)) 4538 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) ((",") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_PIPE) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("|") reduce 12) (("}") reduce 12)) 4539 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) ((",") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_PIPE) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("|") reduce 54) (("}") reduce 54)) 4540 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) ((",") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_PIPE) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("|") reduce 65) (("}") reduce 65)) 4541 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) ((",") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_PIPE) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("|") reduce 75) (("}") reduce 75)) 4542 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) ((",") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_PIPE) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("|") reduce 30) (("}") reduce 30)) 4543 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) ((",") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_PIPE) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("|") reduce 68) (("}") reduce 68)) 4544 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) ((",") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_PIPE) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("|") reduce 21) (("}") reduce 21)) 4545 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) ((",") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_PIPE) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("|") reduce 23) (("}") reduce 23)) 4546 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) ((",") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_PIPE) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("|") reduce 41) (("}") reduce 41)) 4547 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) ((",") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_PIPE) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("|") reduce 62) (("}") reduce 62)) 4548 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) ((",") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_PIPE) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("|") reduce 34) (("}") reduce 34)) 4549 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) ((",") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_PIPE) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("|") reduce 40) (("}") reduce 40)) 4550 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) ((",") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_PIPE) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("|") reduce 13) (("}") reduce 13)) 4551 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) ((",") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_PIPE) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("|") reduce 55) (("}") reduce 55)) 4552 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) ((",") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_PIPE) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("|") reduce 1) (("}") reduce 1)) 4553 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) ((",") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_PIPE) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("|") reduce 2) (("}") reduce 2)) 4554 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) ((",") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_PIPE) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("|") reduce 9) (("}") reduce 9)) 4555 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) ((",") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_PIPE) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("|") reduce 33) (("}") reduce 33)) 4556 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) ((",") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_PIPE) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("|") reduce 58) (("}") reduce 58)) 4557 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) ((",") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_PIPE) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("|") reduce 37) (("}") reduce 37)) 4558 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) ((",") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_PIPE) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("|") reduce 64) (("}") reduce 64)) 4559 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) ((",") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_PIPE) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("|") reduce 46) (("}") reduce 46)) 4560 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) ((",") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_PIPE) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("|") reduce 8) (("}") reduce 8)) 4561 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) ((",") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_PIPE) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("|") reduce 6) (("}") reduce 6)) 4562 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) ((",") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_PIPE) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("|") reduce 7) (("}") reduce 7)) 4563 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) ((",") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_PIPE) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("|") reduce 69) (("}") reduce 69)) 4564 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) ((",") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_PIPE) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("|") reduce 63) (("}") reduce 63)) 4565 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) ((",") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_PIPE) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("|") reduce 56) (("}") reduce 56)) 4566 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) ((",") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_PIPE) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("|") reduce 10) (("}") reduce 10)) 4567 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) ((",") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_PIPE) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("|") reduce 67) (("}") reduce 67)) 4568 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) ((",") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_PIPE) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("|") reduce 44) (("}") reduce 44)) 4569 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) ((",") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_PIPE) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("|") reduce 76) (("}") reduce 76)) 4570 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) ((",") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_PIPE) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("|") reduce 71) (("}") reduce 71)) 4571 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) ((",") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_PIPE) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("|") reduce 77) (("}") reduce 77)) 4572 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) ((",") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_PIPE) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("|") reduce 78) (("}") reduce 78)) 4573 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) ((",") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_PIPE) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("|") reduce 79) (("}") reduce 79)) 4574 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) ((",") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_PIPE) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("|") reduce 4) (("}") reduce 4)) 4575 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) ((",") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_PIPE) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("|") reduce 5) (("}") reduce 5)) 4576 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) ((",") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_PIPE) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("|") reduce 43) (("}") reduce 43)) 4577 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) ((",") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_PIPE) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("|") reduce 73) (("}") reduce 73)) 4578 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) ((",") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_PIPE) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("|") reduce 82) (("}") reduce 82)) 4579 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) ((",") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_PIPE) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("|") reduce 47) (("}") reduce 47)) 4580 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) ((",") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_PIPE) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("|") reduce 31) (("}") reduce 31)) 4581 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) ((",") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_PIPE) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("|") reduce 57) (("}") reduce 57)) 4582 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) ((",") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_PIPE) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("|") reduce 61) (("}") reduce 61)) 4583 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) ((",") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_PIPE) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("|") reduce 28) (("}") reduce 28)) 4584 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) ((",") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_PIPE) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("|") reduce 36) (("}") reduce 36)) 4585 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) ((",") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_PIPE) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("|") reduce 32) (("}") reduce 32)) 4586 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) ((",") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_PIPE) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("|") reduce 35) (("}") reduce 35)) 4587 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) ((",") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_PIPE) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("|") reduce 19) (("}") reduce 19)) 4588 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) ((",") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_PIPE) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("|") reduce 45) (("}") reduce 45)) 4589 ((("%") reduce 542) (("(") reduce 579) (("*") reduce 542) (("+") reduce 542) ((",") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PIPE) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542) (("}") reduce 542)) 4591 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) ((",") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_PIPE) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("|") reduce 72) (("}") reduce 72)) 4592 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) ((",") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_PIPE) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("|") reduce 83) (("}") reduce 83)) 4593 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) ((",") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) (("<") reduce 572) (("=") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AND_EQUAL) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_COALESCE_EQUAL) reduce 572) ((T_CONCAT_EQUAL) reduce 572) ((T_DEC) reduce 572) ((T_DIV_EQUAL) reduce 572) ((T_INC) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_MINUS_EQUAL) reduce 572) ((T_MOD_EQUAL) reduce 572) ((T_MUL_EQUAL) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_OR_EQUAL) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_PLUS_EQUAL) reduce 572) ((T_POW) reduce 572) ((T_POW_EQUAL) reduce 572) ((T_SL) reduce 572) ((T_SL_EQUAL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) ((T_SR_EQUAL) reduce 572) ((T_XOR_EQUAL) reduce 572) (("[") reduce 572) (("^") reduce 572) (("|") reduce 572) (("}") reduce 572)) 4596 ((("%") reduce 544) (("(") reduce 580) (("*") reduce 544) (("+") reduce 544) ((",") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PIPE) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544) (("}") reduce 544)) 4598 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("|") reduce 507) (("}") reduce 507)) 4599 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291) (("}") reduce 291)) 4603 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) (("=") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AND_EQUAL) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_COALESCE_EQUAL) reduce 292) ((T_CONCAT_EQUAL) reduce 292) ((T_DEC) reduce 292) ((T_DIV_EQUAL) reduce 292) ((T_INC) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_MINUS_EQUAL) reduce 292) ((T_MOD_EQUAL) reduce 292) ((T_MUL_EQUAL) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_OR_EQUAL) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_PLUS_EQUAL) reduce 292) ((T_POW) reduce 292) ((T_POW_EQUAL) reduce 292) ((T_SL) reduce 292) ((T_SL_EQUAL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) ((T_SR_EQUAL) reduce 292) ((T_XOR_EQUAL) reduce 292) (("[") reduce 292) (("^") reduce 292) (("|") reduce 292) (("}") reduce 292)) 4604 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) (("=") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AND_EQUAL) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_COALESCE_EQUAL) reduce 293) ((T_CONCAT_EQUAL) reduce 293) ((T_DEC) reduce 293) ((T_DIV_EQUAL) reduce 293) ((T_INC) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_MINUS_EQUAL) reduce 293) ((T_MOD_EQUAL) reduce 293) ((T_MUL_EQUAL) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_OR_EQUAL) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_PLUS_EQUAL) reduce 293) ((T_POW) reduce 293) ((T_POW_EQUAL) reduce 293) ((T_SL) reduce 293) ((T_SL_EQUAL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) ((T_SR_EQUAL) reduce 293) ((T_XOR_EQUAL) reduce 293) (("[") reduce 293) (("^") reduce 293) (("|") reduce 293) (("}") reduce 293)) 4609 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) ((",") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_PIPE) reduce 425) ((T_POW) reduce 425) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425) (("}") reduce 425)) 4611 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) ((",") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_PIPE) reduce 423) ((T_POW) reduce 423) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423) (("}") reduce 423)) 4621 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 419) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 419)) 4622 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 421) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 421)) 4623 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 420) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 420)) 4624 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 413) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 413)) 4625 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 410) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 410)) 4626 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 418) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 418)) 4627 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 412) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 412)) 4628 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 416) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 416)) 4629 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 411) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 411)) 4630 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 414) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 414)) 4631 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 415) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 415)) 4632 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 422) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 422)) 4633 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 417) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 417)) 4635 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 406) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 406)) 4638 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568) (("}") reduce 568)) 4641 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) ((",") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) (("<") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_POW) reduce 564) ((T_SL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) (("[") reduce 564) (("^") reduce 564) (("|") reduce 564) (("}") reduce 564)) 4643 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) ((",") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) (("<") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_POW) reduce 563) ((T_SL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) (("[") reduce 563) (("^") reduce 563) (("|") reduce 563) (("}") reduce 563)) 4645 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559) (("}") reduce 559)) 4646 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565) (("}") reduce 565)) 4647 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) ((",") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("[") reduce 548) (("^") reduce 407) (("|") reduce 407) (("}") reduce 407)) 4651 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) ((",") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) (("<") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_POW) reduce 572) ((T_SL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) (("[") reduce 572) (("^") reduce 572) (("|") reduce 572) (("}") reduce 572)) 4653 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("|") reduce 507) (("}") reduce 507)) 4654 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291) (("}") reduce 291)) 4658 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("[") reduce 292) (("^") reduce 292) (("|") reduce 292) (("}") reduce 292)) 4659 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("[") reduce 293) (("^") reduce 293) (("|") reduce 293) (("}") reduce 293)) 4660 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504) (("}") reduce 504)) 4663 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) ((",") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) (("<") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_POW) reduce 571) ((T_SL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) (("[") reduce 571) (("^") reduce 571) (("|") reduce 571) (("}") reduce 571)) 4664 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("|") reduce 506) (("}") reduce 506)) 4665 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("|") reduce 508) (("}") reduce 508)) 4670 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560) (("}") reduce 560)) 4671 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582) (("}") reduce 582)) 4672 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566) (("}") reduce 566)) 4673 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584) (("}") reduce 584)) 4676 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583) (("}") reduce 583)) 4677 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561) (("}") reduce 561)) 4678 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567) (("}") reduce 567)) 4679 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) ((",") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("[") reduce 562) (("^") reduce 562) (("|") reduce 562) (("}") reduce 562)) 4680 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505) (("}") reduce 505)) 4681 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570) (("}") reduce 570)) 4684 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569) (("}") reduce 569)) 4685 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) (("=") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AND_EQUAL) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_COALESCE_EQUAL) reduce 504) ((T_CONCAT_EQUAL) reduce 504) ((T_DEC) reduce 504) ((T_DIV_EQUAL) reduce 504) ((T_INC) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_MINUS_EQUAL) reduce 504) ((T_MOD_EQUAL) reduce 504) ((T_MUL_EQUAL) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_OR_EQUAL) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_PLUS_EQUAL) reduce 504) ((T_POW) reduce 504) ((T_POW_EQUAL) reduce 504) ((T_SL) reduce 504) ((T_SL_EQUAL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) ((T_SR_EQUAL) reduce 504) ((T_XOR_EQUAL) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504) (("}") reduce 504)) 4696 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) ((",") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PIPE) reduce 494) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494) (("}") reduce 494)) 4697 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_PIPE) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("^") reduce 489) (("|") reduce 489) (("}") reduce 489)) 4707 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 494) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 494)) 4708 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_PIPE) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("^") reduce 490) (("|") reduce 490) (("}") reduce 490)) 4710 ((("%") reduce 541) (("(") reduce 579) (("*") reduce 541) (("+") reduce 541) ((",") reduce 541) (("-") reduce 541) ((".") reduce 541) (("/") reduce 541) (("<") reduce 541) ((">") reduce 541) (("?") reduce 541) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_BOOLEAN_AND) reduce 541) ((T_BOOLEAN_OR) reduce 541) ((T_COALESCE) reduce 541) ((T_INSTANCEOF) reduce 541) ((T_IS_EQUAL) reduce 541) ((T_IS_GREATER_OR_EQUAL) reduce 541) ((T_IS_IDENTICAL) reduce 541) ((T_IS_NOT_EQUAL) reduce 541) ((T_IS_NOT_IDENTICAL) reduce 541) ((T_IS_SMALLER_OR_EQUAL) reduce 541) ((T_LOGICAL_AND) reduce 541) ((T_LOGICAL_OR) reduce 541) ((T_LOGICAL_XOR) reduce 541) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 541) ((T_OBJECT_OPERATOR) reduce 541) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 541) ((T_PIPE) reduce 541) ((T_POW) reduce 541) ((T_SL) reduce 541) ((T_SPACESHIP) reduce 541) ((T_SR) reduce 541) (("[") reduce 541) (("^") reduce 541) (("|") reduce 541) (("}") reduce 541)) 4712 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) ((",") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) (("<") reduce 571) (("=") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AND_EQUAL) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_COALESCE_EQUAL) reduce 571) ((T_CONCAT_EQUAL) reduce 571) ((T_DEC) reduce 571) ((T_DIV_EQUAL) reduce 571) ((T_INC) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_MINUS_EQUAL) reduce 571) ((T_MOD_EQUAL) reduce 571) ((T_MUL_EQUAL) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_OR_EQUAL) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_PLUS_EQUAL) reduce 571) ((T_POW) reduce 571) ((T_POW_EQUAL) reduce 571) ((T_SL) reduce 571) ((T_SL_EQUAL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) ((T_SR_EQUAL) reduce 571) ((T_XOR_EQUAL) reduce 571) (("[") reduce 571) (("^") reduce 571) (("|") reduce 571) (("}") reduce 571)) 4715 ((("%") reduce 543) (("(") reduce 580) (("*") reduce 543) (("+") reduce 543) ((",") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PIPE) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543) (("}") reduce 543)) 4716 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("|") reduce 506) (("}") reduce 506)) 4717 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("|") reduce 508) (("}") reduce 508)) 4719 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) ((",") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_PIPE) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("^") reduce 485) (("|") reduce 485) (("}") reduce 485)) 4720 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_PIPE) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("^") reduce 487) (("|") reduce 487) (("}") reduce 487)) 4725 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) (("=") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AND_EQUAL) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_COALESCE_EQUAL) reduce 560) ((T_CONCAT_EQUAL) reduce 560) ((T_DEC) reduce 560) ((T_DIV_EQUAL) reduce 560) ((T_INC) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_MINUS_EQUAL) reduce 560) ((T_MOD_EQUAL) reduce 560) ((T_MUL_EQUAL) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_OR_EQUAL) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_PLUS_EQUAL) reduce 560) ((T_POW) reduce 560) ((T_POW_EQUAL) reduce 560) ((T_SL) reduce 560) ((T_SL_EQUAL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) ((T_SR_EQUAL) reduce 560) ((T_XOR_EQUAL) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560) (("}") reduce 560)) 4726 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) (("=") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AND_EQUAL) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_COALESCE_EQUAL) reduce 582) ((T_CONCAT_EQUAL) reduce 582) ((T_DEC) reduce 582) ((T_DIV_EQUAL) reduce 582) ((T_INC) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_MINUS_EQUAL) reduce 582) ((T_MOD_EQUAL) reduce 582) ((T_MUL_EQUAL) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_OR_EQUAL) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_PLUS_EQUAL) reduce 582) ((T_POW) reduce 582) ((T_POW_EQUAL) reduce 582) ((T_SL) reduce 582) ((T_SL_EQUAL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) ((T_SR_EQUAL) reduce 582) ((T_XOR_EQUAL) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582) (("}") reduce 582)) 4727 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566) (("}") reduce 566)) 4728 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) (("=") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AND_EQUAL) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_COALESCE_EQUAL) reduce 584) ((T_CONCAT_EQUAL) reduce 584) ((T_DEC) reduce 584) ((T_DIV_EQUAL) reduce 584) ((T_INC) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_MINUS_EQUAL) reduce 584) ((T_MOD_EQUAL) reduce 584) ((T_MUL_EQUAL) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_OR_EQUAL) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_PLUS_EQUAL) reduce 584) ((T_POW) reduce 584) ((T_POW_EQUAL) reduce 584) ((T_SL) reduce 584) ((T_SL_EQUAL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) ((T_SR_EQUAL) reduce 584) ((T_XOR_EQUAL) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584) (("}") reduce 584)) 4731 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) (("=") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AND_EQUAL) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_COALESCE_EQUAL) reduce 583) ((T_CONCAT_EQUAL) reduce 583) ((T_DEC) reduce 583) ((T_DIV_EQUAL) reduce 583) ((T_INC) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_MINUS_EQUAL) reduce 583) ((T_MOD_EQUAL) reduce 583) ((T_MUL_EQUAL) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_OR_EQUAL) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_PLUS_EQUAL) reduce 583) ((T_POW) reduce 583) ((T_POW_EQUAL) reduce 583) ((T_SL) reduce 583) ((T_SL_EQUAL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) ((T_SR_EQUAL) reduce 583) ((T_XOR_EQUAL) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583) (("}") reduce 583)) 4732 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) (("=") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AND_EQUAL) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_COALESCE_EQUAL) reduce 561) ((T_CONCAT_EQUAL) reduce 561) ((T_DEC) reduce 561) ((T_DIV_EQUAL) reduce 561) ((T_INC) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_MINUS_EQUAL) reduce 561) ((T_MOD_EQUAL) reduce 561) ((T_MUL_EQUAL) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_OR_EQUAL) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_PLUS_EQUAL) reduce 561) ((T_POW) reduce 561) ((T_POW_EQUAL) reduce 561) ((T_SL) reduce 561) ((T_SL_EQUAL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) ((T_SR_EQUAL) reduce 561) ((T_XOR_EQUAL) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561) (("}") reduce 561)) 4733 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567) (("}") reduce 567)) 4734 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) ((",") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) (("=") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AND_EQUAL) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_COALESCE_EQUAL) reduce 562) ((T_CONCAT_EQUAL) reduce 562) ((T_DEC) reduce 562) ((T_DIV_EQUAL) reduce 562) ((T_INC) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_MINUS_EQUAL) reduce 562) ((T_MOD_EQUAL) reduce 562) ((T_MUL_EQUAL) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_OR_EQUAL) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_PLUS_EQUAL) reduce 562) ((T_POW) reduce 562) ((T_POW_EQUAL) reduce 562) ((T_SL) reduce 562) ((T_SL_EQUAL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) ((T_SR_EQUAL) reduce 562) ((T_XOR_EQUAL) reduce 562) (("[") reduce 562) (("^") reduce 562) (("|") reduce 562) (("}") reduce 562)) 4736 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) ((",") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_PIPE) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477) (("}") reduce 477)) 4738 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) (("=") shift) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_PIPE) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("|") reduce 520) (("}") reduce 520)) 4740 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 405) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 405)) 4741 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 482) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 482)) 4752 ((("%") reduce 540) (("*") reduce 540) (("+") reduce 540) ((",") reduce 540) (("-") reduce 540) ((".") reduce 540) (("/") reduce 540) (("<") reduce 540) ((">") reduce 540) (("?") reduce 540) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_BOOLEAN_AND) reduce 540) ((T_BOOLEAN_OR) reduce 540) ((T_COALESCE) reduce 540) ((T_DOUBLE_ARROW) reduce 540) ((T_INSTANCEOF) reduce 540) ((T_IS_EQUAL) reduce 540) ((T_IS_GREATER_OR_EQUAL) reduce 540) ((T_IS_IDENTICAL) reduce 540) ((T_IS_NOT_EQUAL) reduce 540) ((T_IS_NOT_IDENTICAL) reduce 540) ((T_IS_SMALLER_OR_EQUAL) reduce 540) ((T_LOGICAL_AND) reduce 540) ((T_LOGICAL_OR) reduce 540) ((T_LOGICAL_XOR) reduce 540) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 540) ((T_OBJECT_OPERATOR) reduce 540) ((T_PIPE) reduce 540) ((T_POW) reduce 540) ((T_SL) reduce 540) ((T_SPACESHIP) reduce 540) ((T_SR) reduce 540) (("[") reduce 540) (("^") reduce 540) (("|") reduce 540) (("}") reduce 540)) 4754 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_PIPE) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("|") reduce 521) (("}") reduce 521)) 4756 ((("%") reduce 534) (("*") reduce 534) (("+") reduce 534) ((",") reduce 534) (("-") reduce 534) ((".") reduce 534) (("/") reduce 534) (("<") reduce 534) ((">") reduce 534) (("?") reduce 534) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_BOOLEAN_AND) reduce 534) ((T_BOOLEAN_OR) reduce 534) ((T_COALESCE) reduce 534) ((T_DOUBLE_ARROW) reduce 534) ((T_INSTANCEOF) reduce 534) ((T_IS_EQUAL) reduce 534) ((T_IS_GREATER_OR_EQUAL) reduce 534) ((T_IS_IDENTICAL) reduce 534) ((T_IS_NOT_EQUAL) reduce 534) ((T_IS_NOT_IDENTICAL) reduce 534) ((T_IS_SMALLER_OR_EQUAL) reduce 534) ((T_LOGICAL_AND) reduce 534) ((T_LOGICAL_OR) reduce 534) ((T_LOGICAL_XOR) reduce 534) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PIPE) reduce 534) ((T_POW) reduce 534) ((T_SL) reduce 534) ((T_SPACESHIP) reduce 534) ((T_SR) reduce 534) (("[") reduce 534) (("^") reduce 534) (("|") reduce 534) (("}") reduce 534)) 4757 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_PIPE) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("^") reduce 524) (("|") reduce 524) (("}") reduce 524)) 4761 ((("%") reduce 517) (("(") shift) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_PIPE) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("^") reduce 517) (("|") reduce 517) (("}") reduce 517)) 4762 ((("%") reduce 533) (("*") reduce 533) (("+") reduce 533) ((",") reduce 533) (("-") reduce 533) ((".") reduce 533) (("/") reduce 533) (("<") reduce 533) ((">") reduce 533) (("?") reduce 533) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_BOOLEAN_AND) reduce 533) ((T_BOOLEAN_OR) reduce 533) ((T_COALESCE) reduce 533) ((T_DOUBLE_ARROW) reduce 533) ((T_INSTANCEOF) reduce 533) ((T_IS_EQUAL) reduce 533) ((T_IS_GREATER_OR_EQUAL) reduce 533) ((T_IS_IDENTICAL) reduce 533) ((T_IS_NOT_EQUAL) reduce 533) ((T_IS_NOT_IDENTICAL) reduce 533) ((T_IS_SMALLER_OR_EQUAL) reduce 533) ((T_LOGICAL_AND) reduce 533) ((T_LOGICAL_OR) reduce 533) ((T_LOGICAL_XOR) reduce 533) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PIPE) reduce 533) ((T_POW) reduce 533) ((T_SL) reduce 533) ((T_SPACESHIP) reduce 533) ((T_SR) reduce 533) (("[") reduce 533) (("^") reduce 533) (("|") reduce 533) (("}") reduce 533)) 4763 ((("%") reduce 537) (("*") reduce 537) (("+") reduce 537) ((",") reduce 537) (("-") reduce 537) ((".") reduce 537) (("/") reduce 537) (("<") reduce 537) ((">") reduce 537) (("?") reduce 537) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_BOOLEAN_AND) reduce 537) ((T_BOOLEAN_OR) reduce 537) ((T_COALESCE) reduce 537) ((T_DOUBLE_ARROW) reduce 537) ((T_INSTANCEOF) reduce 537) ((T_IS_EQUAL) reduce 537) ((T_IS_GREATER_OR_EQUAL) reduce 537) ((T_IS_IDENTICAL) reduce 537) ((T_IS_NOT_EQUAL) reduce 537) ((T_IS_NOT_IDENTICAL) reduce 537) ((T_IS_SMALLER_OR_EQUAL) reduce 537) ((T_LOGICAL_AND) reduce 537) ((T_LOGICAL_OR) reduce 537) ((T_LOGICAL_XOR) reduce 537) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_PIPE) reduce 537) ((T_POW) reduce 537) ((T_SL) reduce 537) ((T_SPACESHIP) reduce 537) ((T_SR) reduce 537) (("[") reduce 537) (("^") reduce 537) (("|") reduce 537) (("}") reduce 537)) 4769 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PIPE) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("|") reduce 532) (("}") reduce 532)) 4771 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_PIPE) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("^") reduce 523) (("|") reduce 523) (("}") reduce 523)) 4773 ((("%") reduce 536) (("*") reduce 536) (("+") reduce 536) ((",") reduce 536) (("-") reduce 536) ((".") reduce 536) (("/") reduce 536) (("<") reduce 536) ((">") reduce 536) (("?") reduce 536) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_BOOLEAN_AND) reduce 536) ((T_BOOLEAN_OR) reduce 536) ((T_COALESCE) reduce 536) ((T_DOUBLE_ARROW) reduce 536) ((T_INSTANCEOF) reduce 536) ((T_IS_EQUAL) reduce 536) ((T_IS_GREATER_OR_EQUAL) reduce 536) ((T_IS_IDENTICAL) reduce 536) ((T_IS_NOT_EQUAL) reduce 536) ((T_IS_NOT_IDENTICAL) reduce 536) ((T_IS_SMALLER_OR_EQUAL) reduce 536) ((T_LOGICAL_AND) reduce 536) ((T_LOGICAL_OR) reduce 536) ((T_LOGICAL_XOR) reduce 536) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 536) ((T_OBJECT_OPERATOR) reduce 536) ((T_PIPE) reduce 536) ((T_POW) reduce 536) ((T_SL) reduce 536) ((T_SPACESHIP) reduce 536) ((T_SR) reduce 536) (("[") reduce 536) (("^") reduce 536) (("|") reduce 536) (("}") reduce 536)) 4774 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("|") reduce 94) (("}") reduce 94)) 4775 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("|") reduce 93) (("}") reduce 93)) 4776 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("|") reduce 95) (("}") reduce 95)) 4778 ((("%") reduce 539) (("*") reduce 539) (("+") reduce 539) ((",") reduce 539) (("-") reduce 539) ((".") reduce 539) (("/") reduce 539) (("<") reduce 539) ((">") reduce 539) (("?") reduce 539) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_BOOLEAN_AND) reduce 539) ((T_BOOLEAN_OR) reduce 539) ((T_COALESCE) reduce 539) ((T_DOUBLE_ARROW) reduce 539) ((T_INSTANCEOF) reduce 539) ((T_IS_EQUAL) reduce 539) ((T_IS_GREATER_OR_EQUAL) reduce 539) ((T_IS_IDENTICAL) reduce 539) ((T_IS_NOT_EQUAL) reduce 539) ((T_IS_NOT_IDENTICAL) reduce 539) ((T_IS_SMALLER_OR_EQUAL) reduce 539) ((T_LOGICAL_AND) reduce 539) ((T_LOGICAL_OR) reduce 539) ((T_LOGICAL_XOR) reduce 539) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 539) ((T_OBJECT_OPERATOR) reduce 539) ((T_PIPE) reduce 539) ((T_POW) reduce 539) ((T_SL) reduce 539) ((T_SPACESHIP) reduce 539) ((T_SR) reduce 539) (("[") reduce 539) (("^") reduce 539) (("|") reduce 539) (("}") reduce 539)) 4781 ((("%") reduce 538) (("*") reduce 538) (("+") reduce 538) ((",") reduce 538) (("-") reduce 538) ((".") reduce 538) (("/") reduce 538) (("<") reduce 538) ((">") reduce 538) (("?") reduce 538) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_BOOLEAN_AND) reduce 538) ((T_BOOLEAN_OR) reduce 538) ((T_COALESCE) reduce 538) ((T_DOUBLE_ARROW) reduce 538) ((T_INSTANCEOF) reduce 538) ((T_IS_EQUAL) reduce 538) ((T_IS_GREATER_OR_EQUAL) reduce 538) ((T_IS_IDENTICAL) reduce 538) ((T_IS_NOT_EQUAL) reduce 538) ((T_IS_NOT_IDENTICAL) reduce 538) ((T_IS_SMALLER_OR_EQUAL) reduce 538) ((T_LOGICAL_AND) reduce 538) ((T_LOGICAL_OR) reduce 538) ((T_LOGICAL_XOR) reduce 538) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 538) ((T_OBJECT_OPERATOR) reduce 538) ((T_PIPE) reduce 538) ((T_POW) reduce 538) ((T_SL) reduce 538) ((T_SPACESHIP) reduce 538) ((T_SR) reduce 538) (("[") reduce 538) (("^") reduce 538) (("|") reduce 538) (("}") reduce 538)) 4787 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92) (("}") reduce 92)) 4790 ((("%") reduce 535) (("*") reduce 535) (("+") reduce 535) ((",") reduce 535) (("-") reduce 535) ((".") reduce 535) (("/") reduce 535) (("<") reduce 535) ((">") reduce 535) (("?") reduce 535) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_BOOLEAN_AND) reduce 535) ((T_BOOLEAN_OR) reduce 535) ((T_COALESCE) reduce 535) ((T_DOUBLE_ARROW) reduce 535) ((T_INSTANCEOF) reduce 535) ((T_IS_EQUAL) reduce 535) ((T_IS_GREATER_OR_EQUAL) reduce 535) ((T_IS_IDENTICAL) reduce 535) ((T_IS_NOT_EQUAL) reduce 535) ((T_IS_NOT_IDENTICAL) reduce 535) ((T_IS_SMALLER_OR_EQUAL) reduce 535) ((T_LOGICAL_AND) reduce 535) ((T_LOGICAL_OR) reduce 535) ((T_LOGICAL_XOR) reduce 535) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PIPE) reduce 535) ((T_POW) reduce 535) ((T_SL) reduce 535) ((T_SPACESHIP) reduce 535) ((T_SR) reduce 535) (("[") reduce 535) (("^") reduce 535) (("|") reduce 535) (("}") reduce 535)) 4792 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) (("=") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AND_EQUAL) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_COALESCE_EQUAL) reduce 568) ((T_CONCAT_EQUAL) reduce 568) ((T_DEC) reduce 568) ((T_DIV_EQUAL) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INC) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_MINUS_EQUAL) reduce 568) ((T_MOD_EQUAL) reduce 568) ((T_MUL_EQUAL) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_OR_EQUAL) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_PLUS_EQUAL) reduce 568) ((T_POW) reduce 568) ((T_POW_EQUAL) reduce 568) ((T_SL) reduce 568) ((T_SL_EQUAL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) ((T_SR_EQUAL) reduce 568) ((T_XOR_EQUAL) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568) (("}") reduce 568)) 4793 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 479) (("(") shift) (("*") reduce 479) (("+") shift) ((",") reduce 479) (("-") shift) ((".") reduce 479) (("/") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 479) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 479) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 479) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PIPE) reduce 479) ((T_POW) reduce 479) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 479) (("`") shift) (("|") reduce 479) (("}") reduce 479) (("~") shift)) 4800 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) ((",") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) (("<") reduce 564) (("=") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AND_EQUAL) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_COALESCE_EQUAL) reduce 564) ((T_CONCAT_EQUAL) reduce 564) ((T_DEC) reduce 564) ((T_DIV_EQUAL) reduce 564) ((T_DOUBLE_ARROW) reduce 564) ((T_INC) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_MINUS_EQUAL) reduce 564) ((T_MOD_EQUAL) reduce 564) ((T_MUL_EQUAL) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_OR_EQUAL) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_PLUS_EQUAL) reduce 564) ((T_POW) reduce 564) ((T_POW_EQUAL) reduce 564) ((T_SL) reduce 564) ((T_SL_EQUAL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) ((T_SR_EQUAL) reduce 564) ((T_XOR_EQUAL) reduce 564) (("[") reduce 564) (("^") reduce 564) (("|") reduce 564) (("}") reduce 564)) 4801 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PIPE) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 551) (("^") reduce 530) (("|") reduce 530) (("}") reduce 530)) 4803 ((("%") reduce 529) (("*") reduce 529) (("+") reduce 529) ((",") reduce 529) (("-") reduce 529) ((".") reduce 529) (("/") reduce 529) (("<") reduce 529) ((">") reduce 529) (("?") reduce 529) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_BOOLEAN_AND) reduce 529) ((T_BOOLEAN_OR) reduce 529) ((T_COALESCE) reduce 529) ((T_DOUBLE_ARROW) reduce 529) ((T_INSTANCEOF) reduce 529) ((T_IS_EQUAL) reduce 529) ((T_IS_GREATER_OR_EQUAL) reduce 529) ((T_IS_IDENTICAL) reduce 529) ((T_IS_NOT_EQUAL) reduce 529) ((T_IS_NOT_IDENTICAL) reduce 529) ((T_IS_SMALLER_OR_EQUAL) reduce 529) ((T_LOGICAL_AND) reduce 529) ((T_LOGICAL_OR) reduce 529) ((T_LOGICAL_XOR) reduce 529) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PIPE) reduce 529) ((T_POW) reduce 529) ((T_SL) reduce 529) ((T_SPACESHIP) reduce 529) ((T_SR) reduce 529) (("[") reduce 554) (("^") reduce 529) (("|") reduce 529) (("}") reduce 529)) 4804 ((("%") reduce 528) (("(") reduce 557) (("*") reduce 528) (("+") reduce 528) ((",") reduce 528) (("-") reduce 528) ((".") reduce 528) (("/") reduce 528) (("<") reduce 528) ((">") reduce 528) (("?") reduce 528) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_BOOLEAN_AND) reduce 528) ((T_BOOLEAN_OR) reduce 528) ((T_COALESCE) reduce 528) ((T_DOUBLE_ARROW) reduce 528) ((T_INSTANCEOF) reduce 528) ((T_IS_EQUAL) reduce 528) ((T_IS_GREATER_OR_EQUAL) reduce 528) ((T_IS_IDENTICAL) reduce 528) ((T_IS_NOT_EQUAL) reduce 528) ((T_IS_NOT_IDENTICAL) reduce 528) ((T_IS_SMALLER_OR_EQUAL) reduce 528) ((T_LOGICAL_AND) reduce 528) ((T_LOGICAL_OR) reduce 528) ((T_LOGICAL_XOR) reduce 528) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PIPE) reduce 528) ((T_POW) reduce 528) ((T_SL) reduce 528) ((T_SPACESHIP) reduce 528) ((T_SR) reduce 528) (("[") reduce 550) (("^") reduce 528) (("|") reduce 528) (("}") reduce 528)) 4805 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 480) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 480) ((T_LOGICAL_OR) reduce 480) ((T_LOGICAL_XOR) reduce 480) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 480)) 4808 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) ((",") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) (("<") reduce 563) (("=") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AND_EQUAL) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_COALESCE_EQUAL) reduce 563) ((T_CONCAT_EQUAL) reduce 563) ((T_DEC) reduce 563) ((T_DIV_EQUAL) reduce 563) ((T_DOUBLE_ARROW) reduce 563) ((T_INC) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_MINUS_EQUAL) reduce 563) ((T_MOD_EQUAL) reduce 563) ((T_MUL_EQUAL) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_OR_EQUAL) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_PLUS_EQUAL) reduce 563) ((T_POW) reduce 563) ((T_POW_EQUAL) reduce 563) ((T_SL) reduce 563) ((T_SL_EQUAL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) ((T_SR_EQUAL) reduce 563) ((T_XOR_EQUAL) reduce 563) (("[") reduce 563) (("^") reduce 563) (("|") reduce 563) (("}") reduce 563)) 4809 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) ((",") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_DOUBLE_ARROW) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_PIPE) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("^") reduce 484) (("|") reduce 484) (("}") reduce 484)) 4810 ((("%") reduce 466) (("*") reduce 466) (("+") reduce 466) ((",") reduce 466) (("-") reduce 466) ((".") reduce 466) (("/") reduce 466) (("<") reduce 466) ((">") reduce 466) (("?") reduce 466) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_BOOLEAN_AND) reduce 466) ((T_BOOLEAN_OR) reduce 466) ((T_COALESCE) reduce 466) ((T_DOUBLE_ARROW) reduce 466) ((T_INSTANCEOF) reduce 466) ((T_IS_EQUAL) reduce 466) ((T_IS_GREATER_OR_EQUAL) reduce 466) ((T_IS_IDENTICAL) reduce 466) ((T_IS_NOT_EQUAL) reduce 466) ((T_IS_NOT_IDENTICAL) reduce 466) ((T_IS_SMALLER_OR_EQUAL) reduce 466) ((T_LOGICAL_AND) reduce 466) ((T_LOGICAL_OR) reduce 466) ((T_LOGICAL_XOR) reduce 466) ((T_PIPE) reduce 466) ((T_POW) reduce 466) ((T_SL) reduce 466) ((T_SPACESHIP) reduce 466) ((T_SR) reduce 466) (("^") reduce 466) (("|") reduce 466) (("}") reduce 466)) 4811 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_PIPE) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("^") reduce 488) (("|") reduce 488) (("}") reduce 488)) 4812 ((("%") reduce 531) (("(") shift) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("|") reduce 531) (("}") reduce 531)) 4813 ((("%") reduce 461) (("(") reduce 558) (("*") reduce 461) (("+") reduce 461) ((",") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PIPE) reduce 461) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") reduce 552) (("^") reduce 461) (("|") reduce 461) (("}") reduce 461)) 4814 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) ((",") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_DOUBLE_ARROW) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_PIPE) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462) (("}") reduce 462)) 4815 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) ((",") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_DOUBLE_ARROW) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_PIPE) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("|") reduce 476) (("}") reduce 476)) 4816 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) (("=") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AND_EQUAL) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_COALESCE_EQUAL) reduce 559) ((T_CONCAT_EQUAL) reduce 559) ((T_DEC) reduce 559) ((T_DIV_EQUAL) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INC) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_MINUS_EQUAL) reduce 559) ((T_MOD_EQUAL) reduce 559) ((T_MUL_EQUAL) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_OR_EQUAL) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_PLUS_EQUAL) reduce 559) ((T_POW) reduce 559) ((T_POW_EQUAL) reduce 559) ((T_SL) reduce 559) ((T_SL_EQUAL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) ((T_SR_EQUAL) reduce 559) ((T_XOR_EQUAL) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559) (("}") reduce 559)) 4817 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565) (("}") reduce 565)) 4818 ((("%") reduce 403) (("*") reduce 403) (("+") reduce 403) ((",") reduce 403) (("-") reduce 403) ((".") reduce 403) (("/") reduce 403) (("<") reduce 403) (("=") shift) ((">") reduce 403) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_DOUBLE_ARROW) reduce 403) ((T_INC) shift) ((T_INSTANCEOF) reduce 403) ((T_IS_EQUAL) reduce 403) ((T_IS_GREATER_OR_EQUAL) reduce 403) ((T_IS_IDENTICAL) reduce 403) ((T_IS_NOT_EQUAL) reduce 403) ((T_IS_NOT_IDENTICAL) reduce 403) ((T_IS_SMALLER_OR_EQUAL) reduce 403) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 403) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 403) ((T_POW_EQUAL) shift) ((T_SL) reduce 403) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 403) ((T_SR) reduce 403) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 548) (("^") reduce 403) (("|") reduce 403) (("}") reduce 403)) 4821 ((("%") reduce 448) (("*") reduce 448) (("+") reduce 448) ((",") reduce 448) (("-") reduce 448) ((".") reduce 448) (("/") reduce 448) (("<") reduce 448) ((">") reduce 448) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_BOOLEAN_AND) reduce 448) ((T_BOOLEAN_OR) reduce 448) ((T_COALESCE) reduce 448) ((T_DOUBLE_ARROW) reduce 448) ((T_INSTANCEOF) reduce 448) ((T_IS_EQUAL) reduce 448) ((T_IS_GREATER_OR_EQUAL) reduce 448) ((T_IS_IDENTICAL) reduce 448) ((T_IS_NOT_EQUAL) reduce 448) ((T_IS_NOT_IDENTICAL) reduce 448) ((T_IS_SMALLER_OR_EQUAL) reduce 448) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_PIPE) reduce 448) ((T_POW) shift) ((T_SL) reduce 448) ((T_SPACESHIP) reduce 448) ((T_SR) reduce 448) (("^") reduce 448) (("|") reduce 448) (("}") reduce 448)) 4853 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 432) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_DOUBLE_ARROW) reduce 432) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 432) (("}") reduce 432)) 4854 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 435) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_DOUBLE_ARROW) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("|") reduce 435) (("}") reduce 435)) 4855 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 444) (("-") shift) ((".") reduce 444) (("/") shift) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_PIPE) reduce 444) ((T_POW) shift) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("|") reduce 444) (("}") reduce 444)) 4856 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 458) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 458) (("|") reduce 458) (("}") reduce 458)) 4857 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 443) (("-") shift) ((".") reduce 443) (("/") shift) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_DOUBLE_ARROW) reduce 443) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_PIPE) reduce 443) ((T_POW) shift) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("^") reduce 443) (("|") reduce 443) (("}") reduce 443)) 4858 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) ((",") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_DOUBLE_ARROW) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_PIPE) reduce 440) ((T_POW) shift) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("|") reduce 440) (("}") reduce 440)) 4859 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 453) (("-") shift) ((".") shift) (("/") shift) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_DOUBLE_ARROW) reduce 453) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_PIPE) reduce 453) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 453) ((T_SR) shift) (("^") reduce 453) (("|") reduce 453) (("}") reduce 453)) 4860 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 431) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 431)) 4861 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 429) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 429) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 429)) 4862 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 430) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 430) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 430)) 4863 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 455) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 455) ((T_SR) shift) (("^") reduce 455) (("|") reduce 455) (("}") reduce 455)) 4864 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 450) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_DOUBLE_ARROW) reduce 450) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 450) (("|") reduce 450) (("}") reduce 450)) 4865 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 452) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_DOUBLE_ARROW) reduce 452) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 452) (("|") reduce 452) (("}") reduce 452)) 4866 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 449) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_DOUBLE_ARROW) reduce 449) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 449) (("|") reduce 449) (("}") reduce 449)) 4867 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 457) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 457) ((T_SR) shift) (("^") reduce 457) (("|") reduce 457) (("}") reduce 457)) 4868 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 451) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_DOUBLE_ARROW) reduce 451) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 451) (("|") reduce 451) (("}") reduce 451)) 4871 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94) (("}") reduce 94)) 4872 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93) (("}") reduce 93)) 4873 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95) (("}") reduce 95)) 4874 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509) (("}") reduce 509)) 4875 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92) (("}") reduce 92)) 4876 ((("%") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568) (("}") reduce 568)) 4877 ((("%") reduce 511) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("|") reduce 511) (("}") reduce 511)) 4878 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PIPE) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459) (("}") reduce 459)) 4879 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510) (("}") reduce 510)) 4880 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("|") reduce 512) (("}") reduce 512)) 4881 ((("%") reduce 573) (("*") reduce 573) (("+") reduce 573) ((",") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_DOUBLE_ARROW) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("|") reduce 573) (("}") reduce 573)) 4887 ((("%") reduce 574) (("*") reduce 574) (("+") reduce 574) ((",") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_DOUBLE_ARROW) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("|") reduce 574) (("}") reduce 574)) 4888 ((("%") reduce 578) (("*") reduce 578) (("+") reduce 578) ((",") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_DOUBLE_ARROW) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("|") reduce 578) (("}") reduce 578)) 4889 ((("%") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582) (("}") reduce 582)) 4890 ((("%") reduce 575) (("*") reduce 575) (("+") reduce 575) ((",") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_DOUBLE_ARROW) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("|") reduce 575) (("}") reduce 575)) 4891 ((("%") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584) (("}") reduce 584)) 4894 ((("%") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583) (("}") reduce 583)) 4895 ((("%") reduce 576) (("*") reduce 576) (("+") reduce 576) ((",") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_DOUBLE_ARROW) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("|") reduce 576) (("}") reduce 576)) 4897 ((("%") reduce 577) (("*") reduce 577) (("+") reduce 577) ((",") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_DOUBLE_ARROW) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("|") reduce 577) (("}") reduce 577)) 4899 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("|") reduce 513) (("}") reduce 513)) 4900 ((("%") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570) (("}") reduce 570)) 4903 ((("%") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569) (("}") reduce 569)) 4904 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 465) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 465) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 465) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 465)) 4905 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 427) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_DOUBLE_ARROW) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 427)) 4906 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 428) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_DOUBLE_ARROW) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 428)) 4907 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 433) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_DOUBLE_ARROW) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("|") reduce 433) (("}") reduce 433)) 4908 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 434) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_DOUBLE_ARROW) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("|") reduce 434) (("}") reduce 434)) 4912 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 463) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 463) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 463)) 4913 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 464) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 464) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 464)) 4914 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 456) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 456) ((T_SR) shift) (("^") reduce 456) (("|") reduce 456) (("}") reduce 456)) 4915 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 454) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 454) ((T_SR) shift) (("^") reduce 454) (("|") reduce 454) (("}") reduce 454)) 4916 ((("%") reduce 441) (("*") reduce 441) (("+") reduce 441) ((",") reduce 441) (("-") reduce 441) ((".") reduce 441) (("/") reduce 441) (("<") reduce 441) ((">") reduce 441) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_DOUBLE_ARROW) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 441) ((T_IS_GREATER_OR_EQUAL) reduce 441) ((T_IS_IDENTICAL) reduce 441) ((T_IS_NOT_EQUAL) reduce 441) ((T_IS_NOT_IDENTICAL) reduce 441) ((T_IS_SMALLER_OR_EQUAL) reduce 441) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_PIPE) reduce 441) ((T_POW) shift) ((T_SL) reduce 441) ((T_SPACESHIP) reduce 441) ((T_SR) reduce 441) (("^") reduce 441) (("|") reduce 441) (("}") reduce 441)) 4917 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 436) (("-") shift) ((".") reduce 436) (("/") shift) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_DOUBLE_ARROW) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_PIPE) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 436) ((T_SR) shift) (("^") reduce 436) (("|") reduce 436) (("}") reduce 436)) 4918 ((("%") shift) (("*") shift) (("+") reduce 438) ((",") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") shift) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_DOUBLE_ARROW) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_PIPE) reduce 438) ((T_POW) shift) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("|") reduce 438) (("}") reduce 438)) 4919 ((("%") shift) (("*") shift) (("+") reduce 437) ((",") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") shift) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_DOUBLE_ARROW) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_PIPE) reduce 437) ((T_POW) shift) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("|") reduce 437) (("}") reduce 437)) 4920 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) ((",") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_DOUBLE_ARROW) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_PIPE) reduce 439) ((T_POW) shift) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("|") reduce 439) (("}") reduce 439)) 4921 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) ((",") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_DOUBLE_ARROW) reduce 442) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_PIPE) reduce 442) ((T_POW) shift) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("|") reduce 442) (("}") reduce 442)) 4923 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) ((",") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_DOUBLE_ARROW) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_PIPE) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("|") reduce 74) (("}") reduce 74)) 4924 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) ((",") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_DOUBLE_ARROW) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_PIPE) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("|") reduce 52) (("}") reduce 52)) 4925 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) ((",") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_DOUBLE_ARROW) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_PIPE) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("|") reduce 27) (("}") reduce 27)) 4926 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) ((",") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_DOUBLE_ARROW) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_PIPE) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("|") reduce 51) (("}") reduce 51)) 4927 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) ((",") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_DOUBLE_ARROW) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_PIPE) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("|") reduce 53) (("}") reduce 53)) 4928 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) ((",") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_DOUBLE_ARROW) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_PIPE) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("|") reduce 49) (("}") reduce 49)) 4929 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) ((",") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_DOUBLE_ARROW) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_PIPE) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("|") reduce 29) (("}") reduce 29)) 4930 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) ((",") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_DOUBLE_ARROW) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_PIPE) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("|") reduce 59) (("}") reduce 59)) 4931 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) ((",") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_DOUBLE_ARROW) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_PIPE) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("|") reduce 60) (("}") reduce 60)) 4932 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) ((",") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_DOUBLE_ARROW) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_PIPE) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("|") reduce 11) (("}") reduce 11)) 4933 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) ((",") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_DOUBLE_ARROW) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_PIPE) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("|") reduce 42) (("}") reduce 42)) 4934 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) ((",") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_DOUBLE_ARROW) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_PIPE) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("|") reduce 39) (("}") reduce 39)) 4935 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) ((",") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_DOUBLE_ARROW) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_PIPE) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("|") reduce 25) (("}") reduce 25)) 4936 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) ((",") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_DOUBLE_ARROW) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_PIPE) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("|") reduce 50) (("}") reduce 50)) 4937 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) ((",") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_DOUBLE_ARROW) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_PIPE) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("|") reduce 66) (("}") reduce 66)) 4938 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) ((",") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_DOUBLE_ARROW) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_PIPE) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("|") reduce 18) (("}") reduce 18)) 4939 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) ((",") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_DOUBLE_ARROW) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_PIPE) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("|") reduce 17) (("}") reduce 17)) 4940 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) ((",") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_DOUBLE_ARROW) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_PIPE) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("|") reduce 15) (("}") reduce 15)) 4941 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) ((",") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_DOUBLE_ARROW) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_PIPE) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("|") reduce 14) (("}") reduce 14)) 4942 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) ((",") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_DOUBLE_ARROW) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_PIPE) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("|") reduce 38) (("}") reduce 38)) 4943 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) ((",") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_DOUBLE_ARROW) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_PIPE) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("|") reduce 26) (("}") reduce 26)) 4944 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) ((",") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_DOUBLE_ARROW) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_PIPE) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("|") reduce 22) (("}") reduce 22)) 4945 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) ((",") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_DOUBLE_ARROW) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_PIPE) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("|") reduce 24) (("}") reduce 24)) 4946 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) ((",") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_DOUBLE_ARROW) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_PIPE) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("|") reduce 16) (("}") reduce 16)) 4947 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) ((",") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_DOUBLE_ARROW) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_PIPE) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("|") reduce 48) (("}") reduce 48)) 4948 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) ((",") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_DOUBLE_ARROW) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_PIPE) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("|") reduce 20) (("}") reduce 20)) 4949 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) ((",") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_DOUBLE_ARROW) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_PIPE) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("|") reduce 70) (("}") reduce 70)) 4950 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) ((",") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_DOUBLE_ARROW) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_PIPE) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("|") reduce 3) (("}") reduce 3)) 4951 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) ((",") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_DOUBLE_ARROW) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_PIPE) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("|") reduce 12) (("}") reduce 12)) 4952 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) ((",") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_DOUBLE_ARROW) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_PIPE) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("|") reduce 54) (("}") reduce 54)) 4953 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) ((",") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_DOUBLE_ARROW) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_PIPE) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("|") reduce 65) (("}") reduce 65)) 4954 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) ((",") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_DOUBLE_ARROW) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_PIPE) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("|") reduce 75) (("}") reduce 75)) 4955 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) ((",") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_DOUBLE_ARROW) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_PIPE) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("|") reduce 30) (("}") reduce 30)) 4956 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) ((",") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_DOUBLE_ARROW) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_PIPE) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("|") reduce 68) (("}") reduce 68)) 4957 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) ((",") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_DOUBLE_ARROW) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_PIPE) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("|") reduce 21) (("}") reduce 21)) 4958 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) ((",") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_DOUBLE_ARROW) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_PIPE) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("|") reduce 23) (("}") reduce 23)) 4959 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) ((",") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_DOUBLE_ARROW) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_PIPE) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("|") reduce 41) (("}") reduce 41)) 4960 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) ((",") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_DOUBLE_ARROW) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_PIPE) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("|") reduce 62) (("}") reduce 62)) 4961 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) ((",") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_DOUBLE_ARROW) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_PIPE) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("|") reduce 34) (("}") reduce 34)) 4962 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) ((",") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_DOUBLE_ARROW) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_PIPE) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("|") reduce 40) (("}") reduce 40)) 4963 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) ((",") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_DOUBLE_ARROW) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_PIPE) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("|") reduce 13) (("}") reduce 13)) 4964 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) ((",") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_DOUBLE_ARROW) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_PIPE) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("|") reduce 55) (("}") reduce 55)) 4965 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) ((",") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_DOUBLE_ARROW) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_PIPE) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("|") reduce 1) (("}") reduce 1)) 4966 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) ((",") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_DOUBLE_ARROW) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_PIPE) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("|") reduce 2) (("}") reduce 2)) 4967 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) ((",") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_DOUBLE_ARROW) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_PIPE) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("|") reduce 9) (("}") reduce 9)) 4968 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) ((",") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_DOUBLE_ARROW) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_PIPE) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("|") reduce 33) (("}") reduce 33)) 4969 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) ((",") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_DOUBLE_ARROW) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_PIPE) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("|") reduce 58) (("}") reduce 58)) 4970 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) ((",") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_DOUBLE_ARROW) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_PIPE) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("|") reduce 37) (("}") reduce 37)) 4971 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) ((",") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_DOUBLE_ARROW) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_PIPE) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("|") reduce 64) (("}") reduce 64)) 4972 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) ((",") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_DOUBLE_ARROW) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_PIPE) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("|") reduce 46) (("}") reduce 46)) 4973 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) ((",") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_DOUBLE_ARROW) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_PIPE) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("|") reduce 8) (("}") reduce 8)) 4974 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) ((",") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_DOUBLE_ARROW) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_PIPE) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("|") reduce 6) (("}") reduce 6)) 4975 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) ((",") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_DOUBLE_ARROW) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_PIPE) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("|") reduce 7) (("}") reduce 7)) 4976 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) ((",") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_DOUBLE_ARROW) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_PIPE) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("|") reduce 69) (("}") reduce 69)) 4977 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) ((",") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_DOUBLE_ARROW) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_PIPE) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("|") reduce 63) (("}") reduce 63)) 4978 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) ((",") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_DOUBLE_ARROW) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_PIPE) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("|") reduce 56) (("}") reduce 56)) 4979 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) ((",") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_DOUBLE_ARROW) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_PIPE) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("|") reduce 10) (("}") reduce 10)) 4980 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) ((",") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_DOUBLE_ARROW) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_PIPE) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("|") reduce 67) (("}") reduce 67)) 4981 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) ((",") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_DOUBLE_ARROW) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_PIPE) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("|") reduce 44) (("}") reduce 44)) 4982 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) ((",") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_DOUBLE_ARROW) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_PIPE) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("|") reduce 76) (("}") reduce 76)) 4983 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) ((",") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_DOUBLE_ARROW) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_PIPE) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("|") reduce 71) (("}") reduce 71)) 4984 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) ((",") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_DOUBLE_ARROW) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_PIPE) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("|") reduce 77) (("}") reduce 77)) 4985 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) ((",") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_DOUBLE_ARROW) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_PIPE) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("|") reduce 78) (("}") reduce 78)) 4986 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) ((",") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_DOUBLE_ARROW) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_PIPE) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("|") reduce 79) (("}") reduce 79)) 4987 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) ((",") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_DOUBLE_ARROW) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_PIPE) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("|") reduce 4) (("}") reduce 4)) 4988 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) ((",") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_DOUBLE_ARROW) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_PIPE) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("|") reduce 5) (("}") reduce 5)) 4989 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) ((",") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_DOUBLE_ARROW) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_PIPE) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("|") reduce 43) (("}") reduce 43)) 4990 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) ((",") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_DOUBLE_ARROW) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_PIPE) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("|") reduce 73) (("}") reduce 73)) 4991 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) ((",") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_DOUBLE_ARROW) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_PIPE) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("|") reduce 82) (("}") reduce 82)) 4992 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) ((",") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_DOUBLE_ARROW) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_PIPE) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("|") reduce 47) (("}") reduce 47)) 4993 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) ((",") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_DOUBLE_ARROW) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_PIPE) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("|") reduce 31) (("}") reduce 31)) 4994 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) ((",") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_DOUBLE_ARROW) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_PIPE) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("|") reduce 57) (("}") reduce 57)) 4995 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) ((",") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_DOUBLE_ARROW) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_PIPE) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("|") reduce 61) (("}") reduce 61)) 4996 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) ((",") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_DOUBLE_ARROW) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_PIPE) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("|") reduce 28) (("}") reduce 28)) 4997 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) ((",") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_DOUBLE_ARROW) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_PIPE) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("|") reduce 36) (("}") reduce 36)) 4998 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) ((",") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_DOUBLE_ARROW) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_PIPE) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("|") reduce 32) (("}") reduce 32)) 4999 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) ((",") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_DOUBLE_ARROW) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_PIPE) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("|") reduce 35) (("}") reduce 35)) 5000 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) ((",") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_DOUBLE_ARROW) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_PIPE) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("|") reduce 19) (("}") reduce 19)) 5001 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) ((",") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_DOUBLE_ARROW) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_PIPE) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("|") reduce 45) (("}") reduce 45)) 5002 ((("%") reduce 542) (("(") reduce 579) (("*") reduce 542) (("+") reduce 542) ((",") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PIPE) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542) (("}") reduce 542)) 5004 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) ((",") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_DOUBLE_ARROW) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_PIPE) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("|") reduce 72) (("}") reduce 72)) 5005 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) ((",") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_DOUBLE_ARROW) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_PIPE) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("|") reduce 83) (("}") reduce 83)) 5006 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) ((",") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) (("<") reduce 572) (("=") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AND_EQUAL) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_COALESCE_EQUAL) reduce 572) ((T_CONCAT_EQUAL) reduce 572) ((T_DEC) reduce 572) ((T_DIV_EQUAL) reduce 572) ((T_DOUBLE_ARROW) reduce 572) ((T_INC) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_MINUS_EQUAL) reduce 572) ((T_MOD_EQUAL) reduce 572) ((T_MUL_EQUAL) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_OR_EQUAL) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_PLUS_EQUAL) reduce 572) ((T_POW) reduce 572) ((T_POW_EQUAL) reduce 572) ((T_SL) reduce 572) ((T_SL_EQUAL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) ((T_SR_EQUAL) reduce 572) ((T_XOR_EQUAL) reduce 572) (("[") reduce 572) (("^") reduce 572) (("|") reduce 572) (("}") reduce 572)) 5009 ((("%") reduce 544) (("(") reduce 580) (("*") reduce 544) (("+") reduce 544) ((",") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PIPE) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544) (("}") reduce 544)) 5011 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("|") reduce 507) (("}") reduce 507)) 5012 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291) (("}") reduce 291)) 5016 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) (("=") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AND_EQUAL) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_COALESCE_EQUAL) reduce 292) ((T_CONCAT_EQUAL) reduce 292) ((T_DEC) reduce 292) ((T_DIV_EQUAL) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INC) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_MINUS_EQUAL) reduce 292) ((T_MOD_EQUAL) reduce 292) ((T_MUL_EQUAL) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_OR_EQUAL) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_PLUS_EQUAL) reduce 292) ((T_POW) reduce 292) ((T_POW_EQUAL) reduce 292) ((T_SL) reduce 292) ((T_SL_EQUAL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) ((T_SR_EQUAL) reduce 292) ((T_XOR_EQUAL) reduce 292) (("[") reduce 292) (("^") reduce 292) (("|") reduce 292) (("}") reduce 292)) 5017 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) (("=") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AND_EQUAL) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_COALESCE_EQUAL) reduce 293) ((T_CONCAT_EQUAL) reduce 293) ((T_DEC) reduce 293) ((T_DIV_EQUAL) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INC) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_MINUS_EQUAL) reduce 293) ((T_MOD_EQUAL) reduce 293) ((T_MUL_EQUAL) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_OR_EQUAL) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_PLUS_EQUAL) reduce 293) ((T_POW) reduce 293) ((T_POW_EQUAL) reduce 293) ((T_SL) reduce 293) ((T_SL_EQUAL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) ((T_SR_EQUAL) reduce 293) ((T_XOR_EQUAL) reduce 293) (("[") reduce 293) (("^") reduce 293) (("|") reduce 293) (("}") reduce 293)) 5022 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) ((",") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_DOUBLE_ARROW) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_PIPE) reduce 425) ((T_POW) reduce 425) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425) (("}") reduce 425)) 5024 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) ((",") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_DOUBLE_ARROW) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_PIPE) reduce 423) ((T_POW) reduce 423) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423) (("}") reduce 423)) 5034 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 419) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 419)) 5035 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 421) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 421)) 5036 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 420) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 420)) 5037 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 413) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 413) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 413)) 5038 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 410) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 410) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 410)) 5039 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 418) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 418)) 5040 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 412) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 412)) 5041 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 416) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 416)) 5042 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 411) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 411)) 5043 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 414) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 414) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 414)) 5044 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 415) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 415) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 415)) 5045 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 422) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 422)) 5046 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 417) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 417)) 5048 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 406) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 406)) 5051 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568) (("}") reduce 568)) 5054 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) ((",") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) (("<") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_DOUBLE_ARROW) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_POW) reduce 564) ((T_SL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) (("[") reduce 564) (("^") reduce 564) (("|") reduce 564) (("}") reduce 564)) 5056 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) ((",") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) (("<") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_DOUBLE_ARROW) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_POW) reduce 563) ((T_SL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) (("[") reduce 563) (("^") reduce 563) (("|") reduce 563) (("}") reduce 563)) 5058 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559) (("}") reduce 559)) 5059 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565) (("}") reduce 565)) 5060 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) ((",") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_DOUBLE_ARROW) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("[") reduce 548) (("^") reduce 407) (("|") reduce 407) (("}") reduce 407)) 5064 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) ((",") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) (("<") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_DOUBLE_ARROW) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_POW) reduce 572) ((T_SL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) (("[") reduce 572) (("^") reduce 572) (("|") reduce 572) (("}") reduce 572)) 5066 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("|") reduce 507) (("}") reduce 507)) 5067 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291) (("}") reduce 291)) 5071 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("[") reduce 292) (("^") reduce 292) (("|") reduce 292) (("}") reduce 292)) 5072 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("[") reduce 293) (("^") reduce 293) (("|") reduce 293) (("}") reduce 293)) 5073 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504) (("}") reduce 504)) 5076 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) ((",") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) (("<") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_DOUBLE_ARROW) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_POW) reduce 571) ((T_SL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) (("[") reduce 571) (("^") reduce 571) (("|") reduce 571) (("}") reduce 571)) 5077 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("|") reduce 506) (("}") reduce 506)) 5078 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("|") reduce 508) (("}") reduce 508)) 5083 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560) (("}") reduce 560)) 5084 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582) (("}") reduce 582)) 5085 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566) (("}") reduce 566)) 5086 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584) (("}") reduce 584)) 5089 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583) (("}") reduce 583)) 5090 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561) (("}") reduce 561)) 5091 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567) (("}") reduce 567)) 5092 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) ((",") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("[") reduce 562) (("^") reduce 562) (("|") reduce 562) (("}") reduce 562)) 5093 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505) (("}") reduce 505)) 5094 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570) (("}") reduce 570)) 5097 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569) (("}") reduce 569)) 5098 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) (("=") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AND_EQUAL) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_COALESCE_EQUAL) reduce 504) ((T_CONCAT_EQUAL) reduce 504) ((T_DEC) reduce 504) ((T_DIV_EQUAL) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INC) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_MINUS_EQUAL) reduce 504) ((T_MOD_EQUAL) reduce 504) ((T_MUL_EQUAL) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_OR_EQUAL) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_PLUS_EQUAL) reduce 504) ((T_POW) reduce 504) ((T_POW_EQUAL) reduce 504) ((T_SL) reduce 504) ((T_SL_EQUAL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) ((T_SR_EQUAL) reduce 504) ((T_XOR_EQUAL) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504) (("}") reduce 504)) 5109 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) ((",") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PIPE) reduce 494) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494) (("}") reduce 494)) 5110 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_PIPE) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("^") reduce 489) (("|") reduce 489) (("}") reduce 489)) 5120 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 494) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 494)) 5121 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_PIPE) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("^") reduce 490) (("|") reduce 490) (("}") reduce 490)) 5123 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 481) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 481)) 5125 ((("%") reduce 541) (("(") reduce 579) (("*") reduce 541) (("+") reduce 541) ((",") reduce 541) (("-") reduce 541) ((".") reduce 541) (("/") reduce 541) (("<") reduce 541) ((">") reduce 541) (("?") reduce 541) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_BOOLEAN_AND) reduce 541) ((T_BOOLEAN_OR) reduce 541) ((T_COALESCE) reduce 541) ((T_DOUBLE_ARROW) reduce 541) ((T_INSTANCEOF) reduce 541) ((T_IS_EQUAL) reduce 541) ((T_IS_GREATER_OR_EQUAL) reduce 541) ((T_IS_IDENTICAL) reduce 541) ((T_IS_NOT_EQUAL) reduce 541) ((T_IS_NOT_IDENTICAL) reduce 541) ((T_IS_SMALLER_OR_EQUAL) reduce 541) ((T_LOGICAL_AND) reduce 541) ((T_LOGICAL_OR) reduce 541) ((T_LOGICAL_XOR) reduce 541) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 541) ((T_OBJECT_OPERATOR) reduce 541) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 541) ((T_PIPE) reduce 541) ((T_POW) reduce 541) ((T_SL) reduce 541) ((T_SPACESHIP) reduce 541) ((T_SR) reduce 541) (("[") reduce 541) (("^") reduce 541) (("|") reduce 541) (("}") reduce 541)) 5127 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) ((",") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) (("<") reduce 571) (("=") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AND_EQUAL) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_COALESCE_EQUAL) reduce 571) ((T_CONCAT_EQUAL) reduce 571) ((T_DEC) reduce 571) ((T_DIV_EQUAL) reduce 571) ((T_DOUBLE_ARROW) reduce 571) ((T_INC) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_MINUS_EQUAL) reduce 571) ((T_MOD_EQUAL) reduce 571) ((T_MUL_EQUAL) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_OR_EQUAL) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_PLUS_EQUAL) reduce 571) ((T_POW) reduce 571) ((T_POW_EQUAL) reduce 571) ((T_SL) reduce 571) ((T_SL_EQUAL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) ((T_SR_EQUAL) reduce 571) ((T_XOR_EQUAL) reduce 571) (("[") reduce 571) (("^") reduce 571) (("|") reduce 571) (("}") reduce 571)) 5130 ((("%") reduce 543) (("(") reduce 580) (("*") reduce 543) (("+") reduce 543) ((",") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PIPE) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543) (("}") reduce 543)) 5131 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("|") reduce 506) (("}") reduce 506)) 5132 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("|") reduce 508) (("}") reduce 508)) 5134 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) ((",") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_DOUBLE_ARROW) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_PIPE) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("^") reduce 485) (("|") reduce 485) (("}") reduce 485)) 5135 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_PIPE) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("^") reduce 487) (("|") reduce 487) (("}") reduce 487)) 5140 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) (("=") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AND_EQUAL) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_COALESCE_EQUAL) reduce 560) ((T_CONCAT_EQUAL) reduce 560) ((T_DEC) reduce 560) ((T_DIV_EQUAL) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INC) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_MINUS_EQUAL) reduce 560) ((T_MOD_EQUAL) reduce 560) ((T_MUL_EQUAL) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_OR_EQUAL) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_PLUS_EQUAL) reduce 560) ((T_POW) reduce 560) ((T_POW_EQUAL) reduce 560) ((T_SL) reduce 560) ((T_SL_EQUAL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) ((T_SR_EQUAL) reduce 560) ((T_XOR_EQUAL) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560) (("}") reduce 560)) 5141 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) (("=") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AND_EQUAL) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_COALESCE_EQUAL) reduce 582) ((T_CONCAT_EQUAL) reduce 582) ((T_DEC) reduce 582) ((T_DIV_EQUAL) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INC) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_MINUS_EQUAL) reduce 582) ((T_MOD_EQUAL) reduce 582) ((T_MUL_EQUAL) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_OR_EQUAL) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_PLUS_EQUAL) reduce 582) ((T_POW) reduce 582) ((T_POW_EQUAL) reduce 582) ((T_SL) reduce 582) ((T_SL_EQUAL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) ((T_SR_EQUAL) reduce 582) ((T_XOR_EQUAL) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582) (("}") reduce 582)) 5142 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566) (("}") reduce 566)) 5143 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) (("=") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AND_EQUAL) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_COALESCE_EQUAL) reduce 584) ((T_CONCAT_EQUAL) reduce 584) ((T_DEC) reduce 584) ((T_DIV_EQUAL) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INC) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_MINUS_EQUAL) reduce 584) ((T_MOD_EQUAL) reduce 584) ((T_MUL_EQUAL) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_OR_EQUAL) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_PLUS_EQUAL) reduce 584) ((T_POW) reduce 584) ((T_POW_EQUAL) reduce 584) ((T_SL) reduce 584) ((T_SL_EQUAL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) ((T_SR_EQUAL) reduce 584) ((T_XOR_EQUAL) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584) (("}") reduce 584)) 5146 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) (("=") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AND_EQUAL) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_COALESCE_EQUAL) reduce 583) ((T_CONCAT_EQUAL) reduce 583) ((T_DEC) reduce 583) ((T_DIV_EQUAL) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INC) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_MINUS_EQUAL) reduce 583) ((T_MOD_EQUAL) reduce 583) ((T_MUL_EQUAL) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_OR_EQUAL) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_PLUS_EQUAL) reduce 583) ((T_POW) reduce 583) ((T_POW_EQUAL) reduce 583) ((T_SL) reduce 583) ((T_SL_EQUAL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) ((T_SR_EQUAL) reduce 583) ((T_XOR_EQUAL) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583) (("}") reduce 583)) 5147 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) (("=") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AND_EQUAL) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_COALESCE_EQUAL) reduce 561) ((T_CONCAT_EQUAL) reduce 561) ((T_DEC) reduce 561) ((T_DIV_EQUAL) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INC) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_MINUS_EQUAL) reduce 561) ((T_MOD_EQUAL) reduce 561) ((T_MUL_EQUAL) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_OR_EQUAL) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_PLUS_EQUAL) reduce 561) ((T_POW) reduce 561) ((T_POW_EQUAL) reduce 561) ((T_SL) reduce 561) ((T_SL_EQUAL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) ((T_SR_EQUAL) reduce 561) ((T_XOR_EQUAL) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561) (("}") reduce 561)) 5148 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567) (("}") reduce 567)) 5149 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) ((",") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) (("=") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AND_EQUAL) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_COALESCE_EQUAL) reduce 562) ((T_CONCAT_EQUAL) reduce 562) ((T_DEC) reduce 562) ((T_DIV_EQUAL) reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_INC) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_MINUS_EQUAL) reduce 562) ((T_MOD_EQUAL) reduce 562) ((T_MUL_EQUAL) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_OR_EQUAL) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_PLUS_EQUAL) reduce 562) ((T_POW) reduce 562) ((T_POW_EQUAL) reduce 562) ((T_SL) reduce 562) ((T_SL_EQUAL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) ((T_SR_EQUAL) reduce 562) ((T_XOR_EQUAL) reduce 562) (("[") reduce 562) (("^") reduce 562) (("|") reduce 562) (("}") reduce 562)) 5151 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) ((",") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_PIPE) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477) (("}") reduce 477)) 5153 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) (("=") shift) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_PIPE) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("|") reduce 520) (("}") reduce 520)) 5155 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 405) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 405)) 5156 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 482) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 482) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 482)) 5159 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 481) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 481) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 481)) 5160 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) ((",") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_DOUBLE_ARROW) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_PIPE) reduce 473) ((T_POW) shift) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("|") reduce 473) (("}") reduce 473)) 5161 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 483) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 483) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 483)) 5162 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) ((",") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_DOUBLE_ARROW) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_PIPE) reduce 469) ((T_POW) shift) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469) (("}") reduce 469)) 5163 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) ((",") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_DOUBLE_ARROW) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_PIPE) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("^") reduce 486) (("|") reduce 486) (("}") reduce 486)) 5165 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_PIPE) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("^") reduce 526) (("|") reduce 526) (("}") reduce 526)) 5167 ((("%") reduce 527) (("*") reduce 527) (("+") reduce 527) ((",") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_DOUBLE_ARROW) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_PIPE) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("^") reduce 527) (("|") reduce 527) (("}") reduce 527)) 5168 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_PIPE) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("^") reduce 525) (("|") reduce 525) (("}") reduce 525)) 5169 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 619) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 619) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 619)) 5170 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 618) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 618) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 618)) 5171 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505) (("}") reduce 505)) 5172 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 478) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 478) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 478)) 5173 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) ((",") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_DOUBLE_ARROW) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_PIPE) reduce 471) ((T_POW) shift) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("|") reduce 471) (("}") reduce 471)) 5175 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94) (("}") reduce 94)) 5176 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93) (("}") reduce 93)) 5177 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95) (("}") reduce 95)) 5178 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509) (("}") reduce 509)) 5179 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92) (("}") reduce 92)) 5180 ((("%") reduce 400) (("(") reduce 400) (("*") reduce 400) (("+") reduce 400) ((",") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_DOUBLE_ARROW) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 400) ((T_OBJECT_OPERATOR) reduce 400) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 400) ((T_PIPE) reduce 400) ((T_POW) reduce 400) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("[") reduce 400) (("^") reduce 400) (("|") reduce 400) (("}") reduce 400)) 5183 ((("%") reduce 511) (("(") reduce 511) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("|") reduce 511) (("}") reduce 511)) 5184 ((("%") reduce 402) (("(") shift) (("*") reduce 402) (("+") reduce 402) ((",") reduce 402) (("-") reduce 402) ((".") reduce 402) (("/") reduce 402) (("<") reduce 402) ((">") reduce 402) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_DOUBLE_ARROW) reduce 402) ((T_INSTANCEOF) reduce 402) ((T_IS_EQUAL) reduce 402) ((T_IS_GREATER_OR_EQUAL) reduce 402) ((T_IS_IDENTICAL) reduce 402) ((T_IS_NOT_EQUAL) reduce 402) ((T_IS_NOT_IDENTICAL) reduce 402) ((T_IS_SMALLER_OR_EQUAL) reduce 402) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_PIPE) reduce 402) ((T_POW) reduce 402) ((T_SL) reduce 402) ((T_SPACESHIP) reduce 402) ((T_SR) reduce 402) (("^") reduce 402) (("|") reduce 402) (("}") reduce 402)) 5185 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510) (("}") reduce 510)) 5186 ((("%") reduce 512) (("(") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("|") reduce 512) (("}") reduce 512)) 5187 ((("%") reduce 573) (("(") reduce 573) (("*") reduce 573) (("+") reduce 573) ((",") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_DOUBLE_ARROW) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("|") reduce 573) (("}") reduce 573)) 5193 ((("%") reduce 574) (("(") reduce 574) (("*") reduce 574) (("+") reduce 574) ((",") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_DOUBLE_ARROW) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("|") reduce 574) (("}") reduce 574)) 5194 ((("%") reduce 578) (("(") reduce 578) (("*") reduce 578) (("+") reduce 578) ((",") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_DOUBLE_ARROW) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("|") reduce 578) (("}") reduce 578)) 5195 ((("%") reduce 575) (("(") reduce 575) (("*") reduce 575) (("+") reduce 575) ((",") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_DOUBLE_ARROW) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("|") reduce 575) (("}") reduce 575)) 5196 ((("%") reduce 576) (("(") reduce 576) (("*") reduce 576) (("+") reduce 576) ((",") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_DOUBLE_ARROW) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("|") reduce 576) (("}") reduce 576)) 5197 ((("%") reduce 399) (("(") reduce 399) (("*") reduce 399) (("+") reduce 399) ((",") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_DOUBLE_ARROW) reduce 399) ((T_INSTANCEOF) reduce 399) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 399) ((T_OBJECT_OPERATOR) reduce 399) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 399) ((T_PIPE) reduce 399) ((T_POW) reduce 399) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("[") reduce 399) (("^") reduce 399) (("|") reduce 399) (("}") reduce 399)) 5199 ((("%") reduce 577) (("(") reduce 577) (("*") reduce 577) (("+") reduce 577) ((",") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_DOUBLE_ARROW) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("|") reduce 577) (("}") reduce 577)) 5200 ((("%") reduce 401) (("(") reduce 401) (("*") reduce 401) (("+") reduce 401) ((",") reduce 401) (("-") reduce 401) ((".") reduce 401) (("/") reduce 401) (("<") reduce 401) ((">") reduce 401) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_DOUBLE_ARROW) reduce 401) ((T_INSTANCEOF) reduce 401) ((T_IS_EQUAL) reduce 401) ((T_IS_GREATER_OR_EQUAL) reduce 401) ((T_IS_IDENTICAL) reduce 401) ((T_IS_NOT_EQUAL) reduce 401) ((T_IS_NOT_IDENTICAL) reduce 401) ((T_IS_SMALLER_OR_EQUAL) reduce 401) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 401) ((T_OBJECT_OPERATOR) reduce 401) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 401) ((T_PIPE) reduce 401) ((T_POW) reduce 401) ((T_SL) reduce 401) ((T_SPACESHIP) reduce 401) ((T_SR) reduce 401) (("[") reduce 401) (("^") reduce 401) (("|") reduce 401) (("}") reduce 401)) 5208 ((("%") reduce 398) (("(") reduce 398) (("*") reduce 398) (("+") reduce 398) ((",") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_DOUBLE_ARROW) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 398) ((T_OBJECT_OPERATOR) reduce 398) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 398) ((T_PIPE) reduce 398) ((T_POW) reduce 398) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("[") reduce 398) (("^") reduce 398) (("|") reduce 398) (("}") reduce 398)) 5210 ((("%") reduce 513) (("(") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("|") reduce 513) (("}") reduce 513)) 5216 ((("%") reduce 233) (("*") reduce 233) (("+") reduce 233) ((",") reduce 233) (("-") reduce 233) ((".") reduce 233) (("/") reduce 233) (("<") reduce 233) ((">") reduce 233) (("?") reduce 233) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_BOOLEAN_AND) reduce 233) ((T_BOOLEAN_OR) reduce 233) ((T_COALESCE) reduce 233) ((T_DOUBLE_ARROW) reduce 233) ((T_INSTANCEOF) reduce 233) ((T_IS_EQUAL) reduce 233) ((T_IS_GREATER_OR_EQUAL) reduce 233) ((T_IS_IDENTICAL) reduce 233) ((T_IS_NOT_EQUAL) reduce 233) ((T_IS_NOT_IDENTICAL) reduce 233) ((T_IS_SMALLER_OR_EQUAL) reduce 233) ((T_LOGICAL_AND) reduce 233) ((T_LOGICAL_OR) reduce 233) ((T_LOGICAL_XOR) reduce 233) ((T_PIPE) reduce 233) ((T_POW) reduce 233) ((T_SL) reduce 233) ((T_SPACESHIP) reduce 233) ((T_SR) reduce 233) (("^") reduce 233) (("|") reduce 233) (("}") reduce 233)) 5221 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 404) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 404)) 5223 ((("%") shift) ((")") reduce 622) (("*") shift) (("+") shift) ((",") reduce 622) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5224 (((")") reduce 620) ((",") reduce 620)) 5226 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) ((")") reduce 126) (("+") shift) (("-") shift) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 5228 ((("%") reduce 613) (("*") reduce 613) (("+") reduce 613) ((",") reduce 613) (("-") reduce 613) ((".") reduce 613) (("/") reduce 613) (("<") reduce 613) ((">") reduce 613) (("?") reduce 613) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_BOOLEAN_AND) reduce 613) ((T_BOOLEAN_OR) reduce 613) ((T_COALESCE) reduce 613) ((T_DOUBLE_ARROW) reduce 613) ((T_INSTANCEOF) reduce 613) ((T_IS_EQUAL) reduce 613) ((T_IS_GREATER_OR_EQUAL) reduce 613) ((T_IS_IDENTICAL) reduce 613) ((T_IS_NOT_EQUAL) reduce 613) ((T_IS_NOT_IDENTICAL) reduce 613) ((T_IS_SMALLER_OR_EQUAL) reduce 613) ((T_LOGICAL_AND) reduce 613) ((T_LOGICAL_OR) reduce 613) ((T_LOGICAL_XOR) reduce 613) ((T_PIPE) reduce 613) ((T_POW) reduce 613) ((T_SL) reduce 613) ((T_SPACESHIP) reduce 613) ((T_SR) reduce 613) (("^") reduce 613) (("|") reduce 613) (("}") reduce 613)) 5229 (((")") reduce 621) ((",") reduce 621)) 5230 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) ((",") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_DOUBLE_ARROW) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_PIPE) reduce 467) ((T_POW) shift) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("|") reduce 467) (("}") reduce 467)) 5231 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 616) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 616) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 616)) 5232 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 615) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 615) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 615)) 5233 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) ((",") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_DOUBLE_ARROW) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("[") reduce 548) (("^") reduce 424) (("|") reduce 424) (("}") reduce 424)) 5235 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_PIPE) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("^") reduce 518) (("|") reduce 518) (("}") reduce 518)) 5236 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) ((",") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_DOUBLE_ARROW) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_PIPE) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("|") reduce 474) (("}") reduce 474)) 5237 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("|") reduce 291) (("}") reduce 291)) 5241 ((("%") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("^") reduce 292) (("|") reduce 292) (("}") reduce 292)) 5242 ((("%") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("^") reduce 293) (("|") reduce 293) (("}") reduce 293)) 5245 ((("%") reduce 617) (("*") reduce 617) (("+") reduce 617) ((",") reduce 617) (("-") reduce 617) ((".") reduce 617) (("/") reduce 617) (("<") reduce 617) ((">") reduce 617) (("?") reduce 617) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_BOOLEAN_AND) reduce 617) ((T_BOOLEAN_OR) reduce 617) ((T_COALESCE) reduce 617) ((T_DOUBLE_ARROW) reduce 617) ((T_INSTANCEOF) reduce 617) ((T_IS_EQUAL) reduce 617) ((T_IS_GREATER_OR_EQUAL) reduce 617) ((T_IS_IDENTICAL) reduce 617) ((T_IS_NOT_EQUAL) reduce 617) ((T_IS_NOT_IDENTICAL) reduce 617) ((T_IS_SMALLER_OR_EQUAL) reduce 617) ((T_LOGICAL_AND) reduce 617) ((T_LOGICAL_OR) reduce 617) ((T_LOGICAL_XOR) reduce 617) ((T_PIPE) reduce 617) ((T_POW) reduce 617) ((T_SL) reduce 617) ((T_SPACESHIP) reduce 617) ((T_SR) reduce 617) (("^") reduce 617) (("|") reduce 617) (("}") reduce 617)) 5248 ((("%") reduce 614) (("*") reduce 614) (("+") reduce 614) ((",") reduce 614) (("-") reduce 614) ((".") reduce 614) (("/") reduce 614) (("<") reduce 614) ((">") reduce 614) (("?") reduce 614) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_BOOLEAN_AND) reduce 614) ((T_BOOLEAN_OR) reduce 614) ((T_COALESCE) reduce 614) ((T_DOUBLE_ARROW) reduce 614) ((T_INSTANCEOF) reduce 614) ((T_IS_EQUAL) reduce 614) ((T_IS_GREATER_OR_EQUAL) reduce 614) ((T_IS_IDENTICAL) reduce 614) ((T_IS_NOT_EQUAL) reduce 614) ((T_IS_NOT_IDENTICAL) reduce 614) ((T_IS_SMALLER_OR_EQUAL) reduce 614) ((T_LOGICAL_AND) reduce 614) ((T_LOGICAL_OR) reduce 614) ((T_LOGICAL_XOR) reduce 614) ((T_PIPE) reduce 614) ((T_POW) reduce 614) ((T_SL) reduce 614) ((T_SPACESHIP) reduce 614) ((T_SR) reduce 614) (("^") reduce 614) (("|") reduce 614) (("}") reduce 614)) 5249 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) ((",") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_DOUBLE_ARROW) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_PIPE) reduce 468) ((T_POW) shift) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468) (("}") reduce 468)) 5250 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) ((",") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_DOUBLE_ARROW) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("[") reduce 548) (("^") reduce 426) (("|") reduce 426) (("}") reduce 426)) 5252 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) ((",") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_DOUBLE_ARROW) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_PIPE) reduce 408) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("^") reduce 408) (("|") reduce 408) (("}") reduce 408)) 5253 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) ((",") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_DOUBLE_ARROW) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_PIPE) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("|") reduce 409) (("}") reduce 409)) 5254 ((("%") reduce 296) (("*") reduce 296) (("+") reduce 296) ((",") reduce 296) (("-") reduce 296) ((".") reduce 296) (("/") reduce 296) (("<") reduce 296) ((">") reduce 296) (("?") reduce 296) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_BOOLEAN_AND) reduce 296) ((T_BOOLEAN_OR) reduce 296) ((T_COALESCE) reduce 296) ((T_DOUBLE_ARROW) reduce 296) ((T_INSTANCEOF) reduce 296) ((T_IS_EQUAL) reduce 296) ((T_IS_GREATER_OR_EQUAL) reduce 296) ((T_IS_IDENTICAL) reduce 296) ((T_IS_NOT_EQUAL) reduce 296) ((T_IS_NOT_IDENTICAL) reduce 296) ((T_IS_SMALLER_OR_EQUAL) reduce 296) ((T_LOGICAL_AND) reduce 296) ((T_LOGICAL_OR) reduce 296) ((T_LOGICAL_XOR) reduce 296) ((T_PIPE) reduce 296) ((T_POW) reduce 296) ((T_SL) reduce 296) ((T_SPACESHIP) reduce 296) ((T_SR) reduce 296) (("^") reduce 296) (("|") reduce 296) (("}") reduce 296)) 5284 (((")") reduce 301) ((",") reduce 301)) 5285 ((("%") shift) ((")") shift) (("*") shift) (("+") shift) ((",") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5289 ((("%") reduce 297) (("*") reduce 297) (("+") reduce 297) ((",") reduce 297) (("-") reduce 297) ((".") reduce 297) (("/") reduce 297) (("<") reduce 297) ((">") reduce 297) (("?") reduce 297) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_BOOLEAN_AND) reduce 297) ((T_BOOLEAN_OR) reduce 297) ((T_COALESCE) reduce 297) ((T_DOUBLE_ARROW) reduce 297) ((T_INSTANCEOF) reduce 297) ((T_IS_EQUAL) reduce 297) ((T_IS_GREATER_OR_EQUAL) reduce 297) ((T_IS_IDENTICAL) reduce 297) ((T_IS_NOT_EQUAL) reduce 297) ((T_IS_NOT_IDENTICAL) reduce 297) ((T_IS_SMALLER_OR_EQUAL) reduce 297) ((T_LOGICAL_AND) reduce 297) ((T_LOGICAL_OR) reduce 297) ((T_LOGICAL_XOR) reduce 297) ((T_PIPE) reduce 297) ((T_POW) reduce 297) ((T_SL) reduce 297) ((T_SPACESHIP) reduce 297) ((T_SR) reduce 297) (("^") reduce 297) (("|") reduce 297) (("}") reduce 297)) 5290 (((")") reduce 302) ((",") reduce 302)) 5291 ((("%") reduce 460) (("(") reduce 556) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PIPE) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("[") reduce 549) (("^") reduce 460) (("|") reduce 460) (("}") reduce 460)) 5293 ((("%") reduce 298) (("*") reduce 298) (("+") reduce 298) ((",") reduce 298) (("-") reduce 298) ((".") reduce 298) (("/") reduce 298) (("<") reduce 298) ((">") reduce 298) (("?") reduce 298) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_BOOLEAN_AND) reduce 298) ((T_BOOLEAN_OR) reduce 298) ((T_COALESCE) reduce 298) ((T_DOUBLE_ARROW) reduce 298) ((T_INSTANCEOF) reduce 298) ((T_IS_EQUAL) reduce 298) ((T_IS_GREATER_OR_EQUAL) reduce 298) ((T_IS_IDENTICAL) reduce 298) ((T_IS_NOT_EQUAL) reduce 298) ((T_IS_NOT_IDENTICAL) reduce 298) ((T_IS_SMALLER_OR_EQUAL) reduce 298) ((T_LOGICAL_AND) reduce 298) ((T_LOGICAL_OR) reduce 298) ((T_LOGICAL_XOR) reduce 298) ((T_PIPE) reduce 298) ((T_POW) reduce 298) ((T_SL) reduce 298) ((T_SPACESHIP) reduce 298) ((T_SR) reduce 298) (("^") reduce 298) (("|") reduce 298) (("}") reduce 298)) 5294 (((")") reduce 300) ((",") reduce 300)) 5300 ((("%") reduce 233) ((")") reduce 233) (("*") reduce 233) (("+") reduce 233) ((",") reduce 233) (("-") reduce 233) ((".") reduce 233) (("/") reduce 233) (("<") reduce 233) ((">") reduce 233) (("?") reduce 233) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_BOOLEAN_AND) reduce 233) ((T_BOOLEAN_OR) reduce 233) ((T_COALESCE) reduce 233) ((T_INSTANCEOF) reduce 233) ((T_IS_EQUAL) reduce 233) ((T_IS_GREATER_OR_EQUAL) reduce 233) ((T_IS_IDENTICAL) reduce 233) ((T_IS_NOT_EQUAL) reduce 233) ((T_IS_NOT_IDENTICAL) reduce 233) ((T_IS_SMALLER_OR_EQUAL) reduce 233) ((T_LOGICAL_AND) reduce 233) ((T_LOGICAL_OR) reduce 233) ((T_LOGICAL_XOR) reduce 233) ((T_PIPE) reduce 233) ((T_POW) reduce 233) ((T_SL) reduce 233) ((T_SPACESHIP) reduce 233) ((T_SR) reduce 233) (("^") reduce 233) (("|") reduce 233)) 5304 ((("%") reduce 613) ((")") reduce 613) (("*") reduce 613) (("+") reduce 613) ((",") reduce 613) (("-") reduce 613) ((".") reduce 613) (("/") reduce 613) (("<") reduce 613) ((">") reduce 613) (("?") reduce 613) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_BOOLEAN_AND) reduce 613) ((T_BOOLEAN_OR) reduce 613) ((T_COALESCE) reduce 613) ((T_INSTANCEOF) reduce 613) ((T_IS_EQUAL) reduce 613) ((T_IS_GREATER_OR_EQUAL) reduce 613) ((T_IS_IDENTICAL) reduce 613) ((T_IS_NOT_EQUAL) reduce 613) ((T_IS_NOT_IDENTICAL) reduce 613) ((T_IS_SMALLER_OR_EQUAL) reduce 613) ((T_LOGICAL_AND) reduce 613) ((T_LOGICAL_OR) reduce 613) ((T_LOGICAL_XOR) reduce 613) ((T_PIPE) reduce 613) ((T_POW) reduce 613) ((T_SL) reduce 613) ((T_SPACESHIP) reduce 613) ((T_SR) reduce 613) (("^") reduce 613) (("|") reduce 613)) 5305 ((("%") shift) ((")") reduce 616) (("*") shift) (("+") shift) ((",") reduce 616) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5306 ((("%") shift) ((")") reduce 615) (("*") shift) (("+") shift) ((",") reduce 615) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5308 ((("%") reduce 518) ((")") reduce 518) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_PIPE) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("^") reduce 518) (("|") reduce 518)) 5309 ((("%") reduce 474) ((")") reduce 474) (("*") reduce 474) (("+") reduce 474) ((",") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_PIPE) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("|") reduce 474)) 5310 ((("%") reduce 291) ((")") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("|") reduce 291)) 5314 ((("%") reduce 292) ((")") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("^") reduce 292) (("|") reduce 292)) 5315 ((("%") reduce 293) ((")") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("^") reduce 293) (("|") reduce 293)) 5318 ((("%") reduce 617) ((")") reduce 617) (("*") reduce 617) (("+") reduce 617) ((",") reduce 617) (("-") reduce 617) ((".") reduce 617) (("/") reduce 617) (("<") reduce 617) ((">") reduce 617) (("?") reduce 617) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_BOOLEAN_AND) reduce 617) ((T_BOOLEAN_OR) reduce 617) ((T_COALESCE) reduce 617) ((T_INSTANCEOF) reduce 617) ((T_IS_EQUAL) reduce 617) ((T_IS_GREATER_OR_EQUAL) reduce 617) ((T_IS_IDENTICAL) reduce 617) ((T_IS_NOT_EQUAL) reduce 617) ((T_IS_NOT_IDENTICAL) reduce 617) ((T_IS_SMALLER_OR_EQUAL) reduce 617) ((T_LOGICAL_AND) reduce 617) ((T_LOGICAL_OR) reduce 617) ((T_LOGICAL_XOR) reduce 617) ((T_PIPE) reduce 617) ((T_POW) reduce 617) ((T_SL) reduce 617) ((T_SPACESHIP) reduce 617) ((T_SR) reduce 617) (("^") reduce 617) (("|") reduce 617)) 5321 ((("%") reduce 614) ((")") reduce 614) (("*") reduce 614) (("+") reduce 614) ((",") reduce 614) (("-") reduce 614) ((".") reduce 614) (("/") reduce 614) (("<") reduce 614) ((">") reduce 614) (("?") reduce 614) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_BOOLEAN_AND) reduce 614) ((T_BOOLEAN_OR) reduce 614) ((T_COALESCE) reduce 614) ((T_INSTANCEOF) reduce 614) ((T_IS_EQUAL) reduce 614) ((T_IS_GREATER_OR_EQUAL) reduce 614) ((T_IS_IDENTICAL) reduce 614) ((T_IS_NOT_EQUAL) reduce 614) ((T_IS_NOT_IDENTICAL) reduce 614) ((T_IS_SMALLER_OR_EQUAL) reduce 614) ((T_LOGICAL_AND) reduce 614) ((T_LOGICAL_OR) reduce 614) ((T_LOGICAL_XOR) reduce 614) ((T_PIPE) reduce 614) ((T_POW) reduce 614) ((T_SL) reduce 614) ((T_SPACESHIP) reduce 614) ((T_SR) reduce 614) (("^") reduce 614) (("|") reduce 614)) 5322 ((("%") reduce 299) (("*") reduce 299) (("+") reduce 299) ((",") reduce 299) (("-") reduce 299) ((".") reduce 299) (("/") reduce 299) (("<") reduce 299) ((">") reduce 299) (("?") reduce 299) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_BOOLEAN_AND) reduce 299) ((T_BOOLEAN_OR) reduce 299) ((T_COALESCE) reduce 299) ((T_DOUBLE_ARROW) reduce 299) ((T_INSTANCEOF) reduce 299) ((T_IS_EQUAL) reduce 299) ((T_IS_GREATER_OR_EQUAL) reduce 299) ((T_IS_IDENTICAL) reduce 299) ((T_IS_NOT_EQUAL) reduce 299) ((T_IS_NOT_IDENTICAL) reduce 299) ((T_IS_SMALLER_OR_EQUAL) reduce 299) ((T_LOGICAL_AND) reduce 299) ((T_LOGICAL_OR) reduce 299) ((T_LOGICAL_XOR) reduce 299) ((T_PIPE) reduce 299) ((T_POW) reduce 299) ((T_SL) reduce 299) ((T_SPACESHIP) reduce 299) ((T_SR) reduce 299) (("^") reduce 299) (("|") reduce 299) (("}") reduce 299)) 5324 ((("%") reduce 408) ((")") reduce 408) (("*") reduce 408) (("+") reduce 408) ((",") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_PIPE) reduce 408) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("^") reduce 408) (("|") reduce 408)) 5325 ((("%") reduce 409) ((")") reduce 409) (("*") reduce 409) (("+") reduce 409) ((",") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_PIPE) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("|") reduce 409)) 5326 ((("%") reduce 296) ((")") reduce 296) (("*") reduce 296) (("+") reduce 296) ((",") reduce 296) (("-") reduce 296) ((".") reduce 296) (("/") reduce 296) (("<") reduce 296) ((">") reduce 296) (("?") reduce 296) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_BOOLEAN_AND) reduce 296) ((T_BOOLEAN_OR) reduce 296) ((T_COALESCE) reduce 296) ((T_INSTANCEOF) reduce 296) ((T_IS_EQUAL) reduce 296) ((T_IS_GREATER_OR_EQUAL) reduce 296) ((T_IS_IDENTICAL) reduce 296) ((T_IS_NOT_EQUAL) reduce 296) ((T_IS_NOT_IDENTICAL) reduce 296) ((T_IS_SMALLER_OR_EQUAL) reduce 296) ((T_LOGICAL_AND) reduce 296) ((T_LOGICAL_OR) reduce 296) ((T_LOGICAL_XOR) reduce 296) ((T_PIPE) reduce 296) ((T_POW) reduce 296) ((T_SL) reduce 296) ((T_SPACESHIP) reduce 296) ((T_SR) reduce 296) (("^") reduce 296) (("|") reduce 296)) 5331 ((("%") reduce 297) ((")") reduce 297) (("*") reduce 297) (("+") reduce 297) ((",") reduce 297) (("-") reduce 297) ((".") reduce 297) (("/") reduce 297) (("<") reduce 297) ((">") reduce 297) (("?") reduce 297) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_BOOLEAN_AND) reduce 297) ((T_BOOLEAN_OR) reduce 297) ((T_COALESCE) reduce 297) ((T_INSTANCEOF) reduce 297) ((T_IS_EQUAL) reduce 297) ((T_IS_GREATER_OR_EQUAL) reduce 297) ((T_IS_IDENTICAL) reduce 297) ((T_IS_NOT_EQUAL) reduce 297) ((T_IS_NOT_IDENTICAL) reduce 297) ((T_IS_SMALLER_OR_EQUAL) reduce 297) ((T_LOGICAL_AND) reduce 297) ((T_LOGICAL_OR) reduce 297) ((T_LOGICAL_XOR) reduce 297) ((T_PIPE) reduce 297) ((T_POW) reduce 297) ((T_SL) reduce 297) ((T_SPACESHIP) reduce 297) ((T_SR) reduce 297) (("^") reduce 297) (("|") reduce 297)) 5332 ((("%") reduce 460) (("(") reduce 556) ((")") reduce 460) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PIPE) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("[") reduce 549) (("^") reduce 460) (("|") reduce 460)) 5334 ((("%") reduce 298) ((")") reduce 298) (("*") reduce 298) (("+") reduce 298) ((",") reduce 298) (("-") reduce 298) ((".") reduce 298) (("/") reduce 298) (("<") reduce 298) ((">") reduce 298) (("?") reduce 298) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_BOOLEAN_AND) reduce 298) ((T_BOOLEAN_OR) reduce 298) ((T_COALESCE) reduce 298) ((T_INSTANCEOF) reduce 298) ((T_IS_EQUAL) reduce 298) ((T_IS_GREATER_OR_EQUAL) reduce 298) ((T_IS_IDENTICAL) reduce 298) ((T_IS_NOT_EQUAL) reduce 298) ((T_IS_NOT_IDENTICAL) reduce 298) ((T_IS_SMALLER_OR_EQUAL) reduce 298) ((T_LOGICAL_AND) reduce 298) ((T_LOGICAL_OR) reduce 298) ((T_LOGICAL_XOR) reduce 298) ((T_PIPE) reduce 298) ((T_POW) reduce 298) ((T_SL) reduce 298) ((T_SPACESHIP) reduce 298) ((T_SR) reduce 298) (("^") reduce 298) (("|") reduce 298)) 5335 ((("%") reduce 299) ((")") reduce 299) (("*") reduce 299) (("+") reduce 299) ((",") reduce 299) (("-") reduce 299) ((".") reduce 299) (("/") reduce 299) (("<") reduce 299) ((">") reduce 299) (("?") reduce 299) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_BOOLEAN_AND) reduce 299) ((T_BOOLEAN_OR) reduce 299) ((T_COALESCE) reduce 299) ((T_INSTANCEOF) reduce 299) ((T_IS_EQUAL) reduce 299) ((T_IS_GREATER_OR_EQUAL) reduce 299) ((T_IS_IDENTICAL) reduce 299) ((T_IS_NOT_EQUAL) reduce 299) ((T_IS_NOT_IDENTICAL) reduce 299) ((T_IS_SMALLER_OR_EQUAL) reduce 299) ((T_LOGICAL_AND) reduce 299) ((T_LOGICAL_OR) reduce 299) ((T_LOGICAL_XOR) reduce 299) ((T_PIPE) reduce 299) ((T_POW) reduce 299) ((T_SL) reduce 299) ((T_SPACESHIP) reduce 299) ((T_SR) reduce 299) (("^") reduce 299) (("|") reduce 299)) 5338 ((("%") reduce 519) (("(") reduce 519) ((")") reduce 519) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PIPE) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("|") reduce 519)) 5339 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) ((",") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_DOUBLE_ARROW) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_PIPE) reduce 472) ((T_POW) shift) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("|") reduce 472) (("}") reduce 472)) 5340 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) ((",") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_DOUBLE_ARROW) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_PIPE) reduce 470) ((T_POW) shift) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("|") reduce 470) (("}") reduce 470)) 5343 ((("%") reduce 519) (("(") reduce 519) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PIPE) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("|") reduce 519) (("}") reduce 519)) 5344 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) ((",") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_DOUBLE_ARROW) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_PIPE) reduce 475) ((T_POW) shift) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("|") reduce 475) (("}") reduce 475)) 5345 ((("%") reduce 446) (("*") reduce 446) (("+") reduce 446) ((",") reduce 446) (("-") reduce 446) ((".") reduce 446) (("/") reduce 446) (("<") reduce 446) ((">") reduce 446) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_BOOLEAN_AND) reduce 446) ((T_BOOLEAN_OR) reduce 446) ((T_COALESCE) reduce 446) ((T_DOUBLE_ARROW) reduce 446) ((T_INSTANCEOF) reduce 446) ((T_IS_EQUAL) reduce 446) ((T_IS_GREATER_OR_EQUAL) reduce 446) ((T_IS_IDENTICAL) reduce 446) ((T_IS_NOT_EQUAL) reduce 446) ((T_IS_NOT_IDENTICAL) reduce 446) ((T_IS_SMALLER_OR_EQUAL) reduce 446) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_PIPE) reduce 446) ((T_POW) shift) ((T_SL) reduce 446) ((T_SPACESHIP) reduce 446) ((T_SR) reduce 446) (("^") reduce 446) (("|") reduce 446) (("}") reduce 446)) 5346 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) ((",") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_DOUBLE_ARROW) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_PIPE) reduce 445) ((T_POW) shift) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("|") reduce 445) (("}") reduce 445)) 5348 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) (("=") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AND_EQUAL) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_COALESCE_EQUAL) reduce 570) ((T_CONCAT_EQUAL) reduce 570) ((T_DEC) reduce 570) ((T_DIV_EQUAL) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INC) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_MINUS_EQUAL) reduce 570) ((T_MOD_EQUAL) reduce 570) ((T_MUL_EQUAL) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_OR_EQUAL) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_PLUS_EQUAL) reduce 570) ((T_POW) reduce 570) ((T_POW_EQUAL) reduce 570) ((T_SL) reduce 570) ((T_SL_EQUAL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) ((T_SR_EQUAL) reduce 570) ((T_XOR_EQUAL) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570) (("}") reduce 570)) 5351 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) (("=") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AND_EQUAL) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_COALESCE_EQUAL) reduce 569) ((T_CONCAT_EQUAL) reduce 569) ((T_DEC) reduce 569) ((T_DIV_EQUAL) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INC) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_MINUS_EQUAL) reduce 569) ((T_MOD_EQUAL) reduce 569) ((T_MUL_EQUAL) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_OR_EQUAL) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_PLUS_EQUAL) reduce 569) ((T_POW) reduce 569) ((T_POW_EQUAL) reduce 569) ((T_SL) reduce 569) ((T_SL_EQUAL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) ((T_SR_EQUAL) reduce 569) ((T_XOR_EQUAL) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569) (("}") reduce 569)) 5354 (((T_CURLY_OPEN) shift) ((T_DOLLAR_OPEN_CURLY_BRACES) shift) ((T_VARIABLE) shift)) 5355 ((("\"") reduce 601) ((T_CURLY_OPEN) reduce 601) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 601) ((T_ENCAPSED_AND_WHITESPACE) reduce 601) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_VARIABLE) reduce 601) (("[") shift)) 5356 ((("\"") shift) ((T_CURLY_OPEN) shift) ((T_DOLLAR_OPEN_CURLY_BRACES) shift) ((T_ENCAPSED_AND_WHITESPACE) shift) ((T_VARIABLE) shift)) 5357 ((("\"") reduce 599) ((T_CURLY_OPEN) reduce 599) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 599) ((T_ENCAPSED_AND_WHITESPACE) reduce 599) ((T_VARIABLE) reduce 599)) 5358 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_PIPE) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("|") reduce 522) (("}") reduce 522)) 5359 ((("\"") reduce 598) ((T_CURLY_OPEN) reduce 598) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 598) ((T_ENCAPSED_AND_WHITESPACE) reduce 598) ((T_VARIABLE) reduce 598)) 5360 ((("\"") reduce 597) ((T_CURLY_OPEN) reduce 597) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 597) ((T_ENCAPSED_AND_WHITESPACE) reduce 597) ((T_VARIABLE) reduce 597)) 5365 ((("\"") reduce 602) ((T_CURLY_OPEN) reduce 602) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 602) ((T_ENCAPSED_AND_WHITESPACE) reduce 602) ((T_VARIABLE) reduce 602)) 5366 ((("\"") reduce 603) ((T_CURLY_OPEN) reduce 603) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 603) ((T_ENCAPSED_AND_WHITESPACE) reduce 603) ((T_VARIABLE) reduce 603)) 5367 ((("\"") reduce 604) ((T_CURLY_OPEN) reduce 604) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 604) ((T_ENCAPSED_AND_WHITESPACE) reduce 604) ((T_VARIABLE) reduce 604)) 5368 ((("\"") reduce 600) ((T_CURLY_OPEN) reduce 600) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 600) ((T_ENCAPSED_AND_WHITESPACE) reduce 600) ((T_VARIABLE) reduce 600)) 5371 ((("\"") reduce 605) ((T_CURLY_OPEN) reduce 605) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 605) ((T_ENCAPSED_AND_WHITESPACE) reduce 605) ((T_VARIABLE) reduce 605)) 5373 ((("\"") reduce 606) ((T_CURLY_OPEN) reduce 606) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 606) ((T_ENCAPSED_AND_WHITESPACE) reduce 606) ((T_VARIABLE) reduce 606)) 5376 ((("\"") reduce 607) ((T_CURLY_OPEN) reduce 607) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 607) ((T_ENCAPSED_AND_WHITESPACE) reduce 607) ((T_VARIABLE) reduce 607)) 5378 ((("\"") reduce 608) ((T_CURLY_OPEN) reduce 608) ((T_DOLLAR_OPEN_CURLY_BRACES) reduce 608) ((T_ENCAPSED_AND_WHITESPACE) reduce 608) ((T_VARIABLE) reduce 608)) 5379 ((("%") reduce 447) (("*") reduce 447) (("+") reduce 447) ((",") reduce 447) (("-") reduce 447) ((".") reduce 447) (("/") reduce 447) (("<") reduce 447) ((">") reduce 447) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_BOOLEAN_AND) reduce 447) ((T_BOOLEAN_OR) reduce 447) ((T_COALESCE) reduce 447) ((T_DOUBLE_ARROW) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 447) ((T_IS_GREATER_OR_EQUAL) reduce 447) ((T_IS_IDENTICAL) reduce 447) ((T_IS_NOT_EQUAL) reduce 447) ((T_IS_NOT_IDENTICAL) reduce 447) ((T_IS_SMALLER_OR_EQUAL) reduce 447) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_PIPE) reduce 447) ((T_POW) shift) ((T_SL) reduce 447) ((T_SPACESHIP) reduce 447) ((T_SR) reduce 447) (("^") reduce 447) (("|") reduce 447) (("}") reduce 447)) 5380 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) ((",") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_PIPE) reduce 473) ((T_POW) shift) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("|") reduce 473) (("}") reduce 473)) 5381 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 483) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 483)) 5382 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) ((",") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_PIPE) reduce 469) ((T_POW) shift) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469) (("}") reduce 469)) 5383 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) ((",") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_PIPE) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("^") reduce 486) (("|") reduce 486) (("}") reduce 486)) 5385 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_PIPE) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("^") reduce 526) (("|") reduce 526) (("}") reduce 526)) 5387 ((("%") reduce 527) (("*") reduce 527) (("+") reduce 527) ((",") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_PIPE) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("^") reduce 527) (("|") reduce 527) (("}") reduce 527)) 5388 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_PIPE) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("^") reduce 525) (("|") reduce 525) (("}") reduce 525)) 5389 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 619) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 619)) 5390 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 618) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 618)) 5391 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505) (("}") reduce 505)) 5392 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 478) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 478)) 5393 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) ((",") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_PIPE) reduce 471) ((T_POW) shift) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("|") reduce 471) (("}") reduce 471)) 5395 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94) (("}") reduce 94)) 5396 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93) (("}") reduce 93)) 5397 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95) (("}") reduce 95)) 5398 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509) (("}") reduce 509)) 5399 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92) (("}") reduce 92)) 5400 ((("%") reduce 400) (("(") reduce 400) (("*") reduce 400) (("+") reduce 400) ((",") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 400) ((T_OBJECT_OPERATOR) reduce 400) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 400) ((T_PIPE) reduce 400) ((T_POW) reduce 400) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("[") reduce 400) (("^") reduce 400) (("|") reduce 400) (("}") reduce 400)) 5403 ((("%") reduce 511) (("(") reduce 511) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("|") reduce 511) (("}") reduce 511)) 5404 ((("%") reduce 402) (("(") shift) (("*") reduce 402) (("+") reduce 402) ((",") reduce 402) (("-") reduce 402) ((".") reduce 402) (("/") reduce 402) (("<") reduce 402) ((">") reduce 402) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_INSTANCEOF) reduce 402) ((T_IS_EQUAL) reduce 402) ((T_IS_GREATER_OR_EQUAL) reduce 402) ((T_IS_IDENTICAL) reduce 402) ((T_IS_NOT_EQUAL) reduce 402) ((T_IS_NOT_IDENTICAL) reduce 402) ((T_IS_SMALLER_OR_EQUAL) reduce 402) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_PIPE) reduce 402) ((T_POW) reduce 402) ((T_SL) reduce 402) ((T_SPACESHIP) reduce 402) ((T_SR) reduce 402) (("^") reduce 402) (("|") reduce 402) (("}") reduce 402)) 5405 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510) (("}") reduce 510)) 5406 ((("%") reduce 512) (("(") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("|") reduce 512) (("}") reduce 512)) 5407 ((("%") reduce 573) (("(") reduce 573) (("*") reduce 573) (("+") reduce 573) ((",") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("|") reduce 573) (("}") reduce 573)) 5413 ((("%") reduce 574) (("(") reduce 574) (("*") reduce 574) (("+") reduce 574) ((",") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("|") reduce 574) (("}") reduce 574)) 5414 ((("%") reduce 578) (("(") reduce 578) (("*") reduce 578) (("+") reduce 578) ((",") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("|") reduce 578) (("}") reduce 578)) 5415 ((("%") reduce 575) (("(") reduce 575) (("*") reduce 575) (("+") reduce 575) ((",") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("|") reduce 575) (("}") reduce 575)) 5416 ((("%") reduce 576) (("(") reduce 576) (("*") reduce 576) (("+") reduce 576) ((",") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("|") reduce 576) (("}") reduce 576)) 5417 ((("%") reduce 399) (("(") reduce 399) (("*") reduce 399) (("+") reduce 399) ((",") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_INSTANCEOF) reduce 399) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 399) ((T_OBJECT_OPERATOR) reduce 399) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 399) ((T_PIPE) reduce 399) ((T_POW) reduce 399) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("[") reduce 399) (("^") reduce 399) (("|") reduce 399) (("}") reduce 399)) 5419 ((("%") reduce 577) (("(") reduce 577) (("*") reduce 577) (("+") reduce 577) ((",") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("|") reduce 577) (("}") reduce 577)) 5420 ((("%") reduce 401) (("(") reduce 401) (("*") reduce 401) (("+") reduce 401) ((",") reduce 401) (("-") reduce 401) ((".") reduce 401) (("/") reduce 401) (("<") reduce 401) ((">") reduce 401) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_INSTANCEOF) reduce 401) ((T_IS_EQUAL) reduce 401) ((T_IS_GREATER_OR_EQUAL) reduce 401) ((T_IS_IDENTICAL) reduce 401) ((T_IS_NOT_EQUAL) reduce 401) ((T_IS_NOT_IDENTICAL) reduce 401) ((T_IS_SMALLER_OR_EQUAL) reduce 401) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 401) ((T_OBJECT_OPERATOR) reduce 401) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 401) ((T_PIPE) reduce 401) ((T_POW) reduce 401) ((T_SL) reduce 401) ((T_SPACESHIP) reduce 401) ((T_SR) reduce 401) (("[") reduce 401) (("^") reduce 401) (("|") reduce 401) (("}") reduce 401)) 5428 ((("%") reduce 398) (("(") reduce 398) (("*") reduce 398) (("+") reduce 398) ((",") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 398) ((T_OBJECT_OPERATOR) reduce 398) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 398) ((T_PIPE) reduce 398) ((T_POW) reduce 398) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("[") reduce 398) (("^") reduce 398) (("|") reduce 398) (("}") reduce 398)) 5430 ((("%") reduce 513) (("(") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("|") reduce 513) (("}") reduce 513)) 5436 ((("%") reduce 233) (("*") reduce 233) (("+") reduce 233) ((",") reduce 233) (("-") reduce 233) ((".") reduce 233) (("/") reduce 233) (("<") reduce 233) ((">") reduce 233) (("?") reduce 233) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_BOOLEAN_AND) reduce 233) ((T_BOOLEAN_OR) reduce 233) ((T_COALESCE) reduce 233) ((T_INSTANCEOF) reduce 233) ((T_IS_EQUAL) reduce 233) ((T_IS_GREATER_OR_EQUAL) reduce 233) ((T_IS_IDENTICAL) reduce 233) ((T_IS_NOT_EQUAL) reduce 233) ((T_IS_NOT_IDENTICAL) reduce 233) ((T_IS_SMALLER_OR_EQUAL) reduce 233) ((T_LOGICAL_AND) reduce 233) ((T_LOGICAL_OR) reduce 233) ((T_LOGICAL_XOR) reduce 233) ((T_PIPE) reduce 233) ((T_POW) reduce 233) ((T_SL) reduce 233) ((T_SPACESHIP) reduce 233) ((T_SR) reduce 233) (("^") reduce 233) (("|") reduce 233) (("}") reduce 233)) 5441 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 404) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 404)) 5445 ((("%") reduce 613) (("*") reduce 613) (("+") reduce 613) ((",") reduce 613) (("-") reduce 613) ((".") reduce 613) (("/") reduce 613) (("<") reduce 613) ((">") reduce 613) (("?") reduce 613) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_BOOLEAN_AND) reduce 613) ((T_BOOLEAN_OR) reduce 613) ((T_COALESCE) reduce 613) ((T_INSTANCEOF) reduce 613) ((T_IS_EQUAL) reduce 613) ((T_IS_GREATER_OR_EQUAL) reduce 613) ((T_IS_IDENTICAL) reduce 613) ((T_IS_NOT_EQUAL) reduce 613) ((T_IS_NOT_IDENTICAL) reduce 613) ((T_IS_SMALLER_OR_EQUAL) reduce 613) ((T_LOGICAL_AND) reduce 613) ((T_LOGICAL_OR) reduce 613) ((T_LOGICAL_XOR) reduce 613) ((T_PIPE) reduce 613) ((T_POW) reduce 613) ((T_SL) reduce 613) ((T_SPACESHIP) reduce 613) ((T_SR) reduce 613) (("^") reduce 613) (("|") reduce 613) (("}") reduce 613)) 5446 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) ((",") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_PIPE) reduce 467) ((T_POW) shift) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("|") reduce 467) (("}") reduce 467)) 5447 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 616) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 616)) 5448 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 615) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 615)) 5449 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) ((",") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("[") reduce 548) (("^") reduce 424) (("|") reduce 424) (("}") reduce 424)) 5451 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_PIPE) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("^") reduce 518) (("|") reduce 518) (("}") reduce 518)) 5452 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) ((",") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_PIPE) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("|") reduce 474) (("}") reduce 474)) 5453 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("|") reduce 291) (("}") reduce 291)) 5457 ((("%") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("^") reduce 292) (("|") reduce 292) (("}") reduce 292)) 5458 ((("%") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("^") reduce 293) (("|") reduce 293) (("}") reduce 293)) 5461 ((("%") reduce 617) (("*") reduce 617) (("+") reduce 617) ((",") reduce 617) (("-") reduce 617) ((".") reduce 617) (("/") reduce 617) (("<") reduce 617) ((">") reduce 617) (("?") reduce 617) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_BOOLEAN_AND) reduce 617) ((T_BOOLEAN_OR) reduce 617) ((T_COALESCE) reduce 617) ((T_INSTANCEOF) reduce 617) ((T_IS_EQUAL) reduce 617) ((T_IS_GREATER_OR_EQUAL) reduce 617) ((T_IS_IDENTICAL) reduce 617) ((T_IS_NOT_EQUAL) reduce 617) ((T_IS_NOT_IDENTICAL) reduce 617) ((T_IS_SMALLER_OR_EQUAL) reduce 617) ((T_LOGICAL_AND) reduce 617) ((T_LOGICAL_OR) reduce 617) ((T_LOGICAL_XOR) reduce 617) ((T_PIPE) reduce 617) ((T_POW) reduce 617) ((T_SL) reduce 617) ((T_SPACESHIP) reduce 617) ((T_SR) reduce 617) (("^") reduce 617) (("|") reduce 617) (("}") reduce 617)) 5464 ((("%") reduce 614) (("*") reduce 614) (("+") reduce 614) ((",") reduce 614) (("-") reduce 614) ((".") reduce 614) (("/") reduce 614) (("<") reduce 614) ((">") reduce 614) (("?") reduce 614) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_BOOLEAN_AND) reduce 614) ((T_BOOLEAN_OR) reduce 614) ((T_COALESCE) reduce 614) ((T_INSTANCEOF) reduce 614) ((T_IS_EQUAL) reduce 614) ((T_IS_GREATER_OR_EQUAL) reduce 614) ((T_IS_IDENTICAL) reduce 614) ((T_IS_NOT_EQUAL) reduce 614) ((T_IS_NOT_IDENTICAL) reduce 614) ((T_IS_SMALLER_OR_EQUAL) reduce 614) ((T_LOGICAL_AND) reduce 614) ((T_LOGICAL_OR) reduce 614) ((T_LOGICAL_XOR) reduce 614) ((T_PIPE) reduce 614) ((T_POW) reduce 614) ((T_SL) reduce 614) ((T_SPACESHIP) reduce 614) ((T_SR) reduce 614) (("^") reduce 614) (("|") reduce 614) (("}") reduce 614)) 5465 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) ((",") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_PIPE) reduce 468) ((T_POW) shift) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468) (("}") reduce 468)) 5466 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) ((",") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("[") reduce 548) (("^") reduce 426) (("|") reduce 426) (("}") reduce 426)) 5468 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) ((",") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_PIPE) reduce 408) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("^") reduce 408) (("|") reduce 408) (("}") reduce 408)) 5469 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) ((",") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_PIPE) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("|") reduce 409) (("}") reduce 409)) 5470 ((("%") reduce 296) (("*") reduce 296) (("+") reduce 296) ((",") reduce 296) (("-") reduce 296) ((".") reduce 296) (("/") reduce 296) (("<") reduce 296) ((">") reduce 296) (("?") reduce 296) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_BOOLEAN_AND) reduce 296) ((T_BOOLEAN_OR) reduce 296) ((T_COALESCE) reduce 296) ((T_INSTANCEOF) reduce 296) ((T_IS_EQUAL) reduce 296) ((T_IS_GREATER_OR_EQUAL) reduce 296) ((T_IS_IDENTICAL) reduce 296) ((T_IS_NOT_EQUAL) reduce 296) ((T_IS_NOT_IDENTICAL) reduce 296) ((T_IS_SMALLER_OR_EQUAL) reduce 296) ((T_LOGICAL_AND) reduce 296) ((T_LOGICAL_OR) reduce 296) ((T_LOGICAL_XOR) reduce 296) ((T_PIPE) reduce 296) ((T_POW) reduce 296) ((T_SL) reduce 296) ((T_SPACESHIP) reduce 296) ((T_SR) reduce 296) (("^") reduce 296) (("|") reduce 296) (("}") reduce 296)) 5475 ((("%") reduce 297) (("*") reduce 297) (("+") reduce 297) ((",") reduce 297) (("-") reduce 297) ((".") reduce 297) (("/") reduce 297) (("<") reduce 297) ((">") reduce 297) (("?") reduce 297) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_BOOLEAN_AND) reduce 297) ((T_BOOLEAN_OR) reduce 297) ((T_COALESCE) reduce 297) ((T_INSTANCEOF) reduce 297) ((T_IS_EQUAL) reduce 297) ((T_IS_GREATER_OR_EQUAL) reduce 297) ((T_IS_IDENTICAL) reduce 297) ((T_IS_NOT_EQUAL) reduce 297) ((T_IS_NOT_IDENTICAL) reduce 297) ((T_IS_SMALLER_OR_EQUAL) reduce 297) ((T_LOGICAL_AND) reduce 297) ((T_LOGICAL_OR) reduce 297) ((T_LOGICAL_XOR) reduce 297) ((T_PIPE) reduce 297) ((T_POW) reduce 297) ((T_SL) reduce 297) ((T_SPACESHIP) reduce 297) ((T_SR) reduce 297) (("^") reduce 297) (("|") reduce 297) (("}") reduce 297)) 5476 ((("%") reduce 460) (("(") reduce 556) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PIPE) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("[") reduce 549) (("^") reduce 460) (("|") reduce 460) (("}") reduce 460)) 5478 ((("%") reduce 298) (("*") reduce 298) (("+") reduce 298) ((",") reduce 298) (("-") reduce 298) ((".") reduce 298) (("/") reduce 298) (("<") reduce 298) ((">") reduce 298) (("?") reduce 298) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_BOOLEAN_AND) reduce 298) ((T_BOOLEAN_OR) reduce 298) ((T_COALESCE) reduce 298) ((T_INSTANCEOF) reduce 298) ((T_IS_EQUAL) reduce 298) ((T_IS_GREATER_OR_EQUAL) reduce 298) ((T_IS_IDENTICAL) reduce 298) ((T_IS_NOT_EQUAL) reduce 298) ((T_IS_NOT_IDENTICAL) reduce 298) ((T_IS_SMALLER_OR_EQUAL) reduce 298) ((T_LOGICAL_AND) reduce 298) ((T_LOGICAL_OR) reduce 298) ((T_LOGICAL_XOR) reduce 298) ((T_PIPE) reduce 298) ((T_POW) reduce 298) ((T_SL) reduce 298) ((T_SPACESHIP) reduce 298) ((T_SR) reduce 298) (("^") reduce 298) (("|") reduce 298) (("}") reduce 298)) 5479 ((("%") reduce 299) (("*") reduce 299) (("+") reduce 299) ((",") reduce 299) (("-") reduce 299) ((".") reduce 299) (("/") reduce 299) (("<") reduce 299) ((">") reduce 299) (("?") reduce 299) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_BOOLEAN_AND) reduce 299) ((T_BOOLEAN_OR) reduce 299) ((T_COALESCE) reduce 299) ((T_INSTANCEOF) reduce 299) ((T_IS_EQUAL) reduce 299) ((T_IS_GREATER_OR_EQUAL) reduce 299) ((T_IS_IDENTICAL) reduce 299) ((T_IS_NOT_EQUAL) reduce 299) ((T_IS_NOT_IDENTICAL) reduce 299) ((T_IS_SMALLER_OR_EQUAL) reduce 299) ((T_LOGICAL_AND) reduce 299) ((T_LOGICAL_OR) reduce 299) ((T_LOGICAL_XOR) reduce 299) ((T_PIPE) reduce 299) ((T_POW) reduce 299) ((T_SL) reduce 299) ((T_SPACESHIP) reduce 299) ((T_SR) reduce 299) (("^") reduce 299) (("|") reduce 299) (("}") reduce 299)) 5480 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) ((",") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_PIPE) reduce 472) ((T_POW) shift) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("|") reduce 472) (("}") reduce 472)) 5481 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) ((",") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_PIPE) reduce 470) ((T_POW) shift) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("|") reduce 470) (("}") reduce 470)) 5484 ((("%") reduce 519) (("(") reduce 519) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PIPE) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("|") reduce 519) (("}") reduce 519)) 5485 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) ((",") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_PIPE) reduce 475) ((T_POW) shift) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("|") reduce 475) (("}") reduce 475)) 5486 ((("%") reduce 446) (("*") reduce 446) (("+") reduce 446) ((",") reduce 446) (("-") reduce 446) ((".") reduce 446) (("/") reduce 446) (("<") reduce 446) ((">") reduce 446) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_BOOLEAN_AND) reduce 446) ((T_BOOLEAN_OR) reduce 446) ((T_COALESCE) reduce 446) ((T_INSTANCEOF) reduce 446) ((T_IS_EQUAL) reduce 446) ((T_IS_GREATER_OR_EQUAL) reduce 446) ((T_IS_IDENTICAL) reduce 446) ((T_IS_NOT_EQUAL) reduce 446) ((T_IS_NOT_IDENTICAL) reduce 446) ((T_IS_SMALLER_OR_EQUAL) reduce 446) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_PIPE) reduce 446) ((T_POW) shift) ((T_SL) reduce 446) ((T_SPACESHIP) reduce 446) ((T_SR) reduce 446) (("^") reduce 446) (("|") reduce 446) (("}") reduce 446)) 5487 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) ((",") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_PIPE) reduce 445) ((T_POW) shift) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("|") reduce 445) (("}") reduce 445)) 5489 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) (("=") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AND_EQUAL) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_COALESCE_EQUAL) reduce 570) ((T_CONCAT_EQUAL) reduce 570) ((T_DEC) reduce 570) ((T_DIV_EQUAL) reduce 570) ((T_INC) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_MINUS_EQUAL) reduce 570) ((T_MOD_EQUAL) reduce 570) ((T_MUL_EQUAL) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_OR_EQUAL) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_PLUS_EQUAL) reduce 570) ((T_POW) reduce 570) ((T_POW_EQUAL) reduce 570) ((T_SL) reduce 570) ((T_SL_EQUAL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) ((T_SR_EQUAL) reduce 570) ((T_XOR_EQUAL) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570) (("}") reduce 570)) 5492 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) (("=") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AND_EQUAL) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_COALESCE_EQUAL) reduce 569) ((T_CONCAT_EQUAL) reduce 569) ((T_DEC) reduce 569) ((T_DIV_EQUAL) reduce 569) ((T_INC) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_MINUS_EQUAL) reduce 569) ((T_MOD_EQUAL) reduce 569) ((T_MUL_EQUAL) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_OR_EQUAL) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_PLUS_EQUAL) reduce 569) ((T_POW) reduce 569) ((T_POW_EQUAL) reduce 569) ((T_SL) reduce 569) ((T_SL_EQUAL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) ((T_SR_EQUAL) reduce 569) ((T_XOR_EQUAL) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569) (("}") reduce 569)) 5494 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_PIPE) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("|") reduce 522) (("}") reduce 522)) 5495 ((("%") reduce 447) (("*") reduce 447) (("+") reduce 447) ((",") reduce 447) (("-") reduce 447) ((".") reduce 447) (("/") reduce 447) (("<") reduce 447) ((">") reduce 447) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_BOOLEAN_AND) reduce 447) ((T_BOOLEAN_OR) reduce 447) ((T_COALESCE) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 447) ((T_IS_GREATER_OR_EQUAL) reduce 447) ((T_IS_IDENTICAL) reduce 447) ((T_IS_NOT_EQUAL) reduce 447) ((T_IS_NOT_IDENTICAL) reduce 447) ((T_IS_SMALLER_OR_EQUAL) reduce 447) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_PIPE) reduce 447) ((T_POW) shift) ((T_SL) reduce 447) ((T_SPACESHIP) reduce 447) ((T_SR) reduce 447) (("^") reduce 447) (("|") reduce 447) (("}") reduce 447)) 5496 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 241) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 241) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5507 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) ((",") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PIPE) reduce 494) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 5508 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_PIPE) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("^") reduce 489) (("|") reduce 489)) 5518 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 494) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5519 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_PIPE) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("^") reduce 490) (("|") reduce 490)) 5521 ((("%") reduce 541) (("(") reduce 579) (("*") reduce 541) (("+") reduce 541) ((",") reduce 541) (("-") reduce 541) ((".") reduce 541) (("/") reduce 541) (("<") reduce 541) ((">") reduce 541) (("?") reduce 541) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_BOOLEAN_AND) reduce 541) ((T_BOOLEAN_OR) reduce 541) ((T_COALESCE) reduce 541) ((T_DOUBLE_ARROW) reduce 541) ((T_INSTANCEOF) reduce 541) ((T_IS_EQUAL) reduce 541) ((T_IS_GREATER_OR_EQUAL) reduce 541) ((T_IS_IDENTICAL) reduce 541) ((T_IS_NOT_EQUAL) reduce 541) ((T_IS_NOT_IDENTICAL) reduce 541) ((T_IS_SMALLER_OR_EQUAL) reduce 541) ((T_LOGICAL_AND) reduce 541) ((T_LOGICAL_OR) reduce 541) ((T_LOGICAL_XOR) reduce 541) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 541) ((T_OBJECT_OPERATOR) reduce 541) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 541) ((T_PIPE) reduce 541) ((T_POW) reduce 541) ((T_SL) reduce 541) ((T_SPACESHIP) reduce 541) ((T_SR) reduce 541) (("[") reduce 541) (("^") reduce 541) (("|") reduce 541)) 5523 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) ((",") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) (("<") reduce 571) (("=") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AND_EQUAL) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_COALESCE_EQUAL) reduce 571) ((T_CONCAT_EQUAL) reduce 571) ((T_DEC) reduce 571) ((T_DIV_EQUAL) reduce 571) ((T_DOUBLE_ARROW) reduce 571) ((T_INC) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_MINUS_EQUAL) reduce 571) ((T_MOD_EQUAL) reduce 571) ((T_MUL_EQUAL) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_OR_EQUAL) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_PLUS_EQUAL) reduce 571) ((T_POW) reduce 571) ((T_POW_EQUAL) reduce 571) ((T_SL) reduce 571) ((T_SL_EQUAL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) ((T_SR_EQUAL) reduce 571) ((T_XOR_EQUAL) reduce 571) (("[") reduce 571) (("^") reduce 571) (("|") reduce 571)) 5526 ((("%") reduce 543) (("(") reduce 580) (("*") reduce 543) (("+") reduce 543) ((",") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PIPE) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 5527 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("|") reduce 506)) 5528 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("|") reduce 508)) 5530 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) ((",") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_DOUBLE_ARROW) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_PIPE) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("^") reduce 485) (("|") reduce 485)) 5531 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_PIPE) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("^") reduce 487) (("|") reduce 487)) 5536 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) (("=") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AND_EQUAL) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_COALESCE_EQUAL) reduce 560) ((T_CONCAT_EQUAL) reduce 560) ((T_DEC) reduce 560) ((T_DIV_EQUAL) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INC) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_MINUS_EQUAL) reduce 560) ((T_MOD_EQUAL) reduce 560) ((T_MUL_EQUAL) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_OR_EQUAL) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_PLUS_EQUAL) reduce 560) ((T_POW) reduce 560) ((T_POW_EQUAL) reduce 560) ((T_SL) reduce 560) ((T_SL_EQUAL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) ((T_SR_EQUAL) reduce 560) ((T_XOR_EQUAL) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 5537 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) (("=") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AND_EQUAL) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_COALESCE_EQUAL) reduce 582) ((T_CONCAT_EQUAL) reduce 582) ((T_DEC) reduce 582) ((T_DIV_EQUAL) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INC) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_MINUS_EQUAL) reduce 582) ((T_MOD_EQUAL) reduce 582) ((T_MUL_EQUAL) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_OR_EQUAL) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_PLUS_EQUAL) reduce 582) ((T_POW) reduce 582) ((T_POW_EQUAL) reduce 582) ((T_SL) reduce 582) ((T_SL_EQUAL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) ((T_SR_EQUAL) reduce 582) ((T_XOR_EQUAL) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 5538 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 5539 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) (("=") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AND_EQUAL) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_COALESCE_EQUAL) reduce 584) ((T_CONCAT_EQUAL) reduce 584) ((T_DEC) reduce 584) ((T_DIV_EQUAL) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INC) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_MINUS_EQUAL) reduce 584) ((T_MOD_EQUAL) reduce 584) ((T_MUL_EQUAL) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_OR_EQUAL) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_PLUS_EQUAL) reduce 584) ((T_POW) reduce 584) ((T_POW_EQUAL) reduce 584) ((T_SL) reduce 584) ((T_SL_EQUAL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) ((T_SR_EQUAL) reduce 584) ((T_XOR_EQUAL) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 5542 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) (("=") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AND_EQUAL) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_COALESCE_EQUAL) reduce 583) ((T_CONCAT_EQUAL) reduce 583) ((T_DEC) reduce 583) ((T_DIV_EQUAL) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INC) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_MINUS_EQUAL) reduce 583) ((T_MOD_EQUAL) reduce 583) ((T_MUL_EQUAL) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_OR_EQUAL) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_PLUS_EQUAL) reduce 583) ((T_POW) reduce 583) ((T_POW_EQUAL) reduce 583) ((T_SL) reduce 583) ((T_SL_EQUAL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) ((T_SR_EQUAL) reduce 583) ((T_XOR_EQUAL) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 5543 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) (("=") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AND_EQUAL) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_COALESCE_EQUAL) reduce 561) ((T_CONCAT_EQUAL) reduce 561) ((T_DEC) reduce 561) ((T_DIV_EQUAL) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INC) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_MINUS_EQUAL) reduce 561) ((T_MOD_EQUAL) reduce 561) ((T_MUL_EQUAL) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_OR_EQUAL) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_PLUS_EQUAL) reduce 561) ((T_POW) reduce 561) ((T_POW_EQUAL) reduce 561) ((T_SL) reduce 561) ((T_SL_EQUAL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) ((T_SR_EQUAL) reduce 561) ((T_XOR_EQUAL) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 5544 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 5545 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) ((",") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) (("=") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AND_EQUAL) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_COALESCE_EQUAL) reduce 562) ((T_CONCAT_EQUAL) reduce 562) ((T_DEC) reduce 562) ((T_DIV_EQUAL) reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_INC) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_MINUS_EQUAL) reduce 562) ((T_MOD_EQUAL) reduce 562) ((T_MUL_EQUAL) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_OR_EQUAL) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_PLUS_EQUAL) reduce 562) ((T_POW) reduce 562) ((T_POW_EQUAL) reduce 562) ((T_SL) reduce 562) ((T_SL_EQUAL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) ((T_SR_EQUAL) reduce 562) ((T_XOR_EQUAL) reduce 562) (("[") reduce 562) (("^") reduce 562) (("|") reduce 562)) 5547 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) ((",") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_PIPE) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 5549 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) (("=") shift) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_PIPE) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("|") reduce 520)) 5551 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 405) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5552 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 482) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 482) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5553 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 480) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 480) ((T_LOGICAL_OR) reduce 480) ((T_LOGICAL_XOR) reduce 480) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5555 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 481) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 481) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5556 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) ((",") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_DOUBLE_ARROW) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_PIPE) reduce 473) ((T_POW) shift) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("|") reduce 473)) 5557 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 483) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 483) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5558 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) ((",") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_DOUBLE_ARROW) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_PIPE) reduce 469) ((T_POW) shift) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 5559 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) ((",") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_DOUBLE_ARROW) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_PIPE) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("^") reduce 486) (("|") reduce 486)) 5561 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_PIPE) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("^") reduce 526) (("|") reduce 526)) 5563 ((("%") reduce 527) (("*") reduce 527) (("+") reduce 527) ((",") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_DOUBLE_ARROW) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_PIPE) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("^") reduce 527) (("|") reduce 527)) 5564 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_PIPE) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("^") reduce 525) (("|") reduce 525)) 5565 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 619) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 619) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5566 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 618) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 618) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5567 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 5568 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 478) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 478) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5569 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) ((",") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_DOUBLE_ARROW) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_PIPE) reduce 471) ((T_POW) shift) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("|") reduce 471)) 5571 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 5572 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 5573 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 5574 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 5575 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 5576 ((("%") reduce 400) (("(") reduce 400) (("*") reduce 400) (("+") reduce 400) ((",") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_DOUBLE_ARROW) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 400) ((T_OBJECT_OPERATOR) reduce 400) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 400) ((T_PIPE) reduce 400) ((T_POW) reduce 400) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("[") reduce 400) (("^") reduce 400) (("|") reduce 400)) 5579 ((("%") reduce 511) (("(") reduce 511) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("|") reduce 511)) 5580 ((("%") reduce 402) (("(") shift) (("*") reduce 402) (("+") reduce 402) ((",") reduce 402) (("-") reduce 402) ((".") reduce 402) (("/") reduce 402) (("<") reduce 402) ((">") reduce 402) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_DOUBLE_ARROW) reduce 402) ((T_INSTANCEOF) reduce 402) ((T_IS_EQUAL) reduce 402) ((T_IS_GREATER_OR_EQUAL) reduce 402) ((T_IS_IDENTICAL) reduce 402) ((T_IS_NOT_EQUAL) reduce 402) ((T_IS_NOT_IDENTICAL) reduce 402) ((T_IS_SMALLER_OR_EQUAL) reduce 402) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_PIPE) reduce 402) ((T_POW) reduce 402) ((T_SL) reduce 402) ((T_SPACESHIP) reduce 402) ((T_SR) reduce 402) (("^") reduce 402) (("|") reduce 402)) 5581 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 5582 ((("%") reduce 512) (("(") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("|") reduce 512)) 5583 ((("%") reduce 573) (("(") reduce 573) (("*") reduce 573) (("+") reduce 573) ((",") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_DOUBLE_ARROW) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("|") reduce 573)) 5589 ((("%") reduce 574) (("(") reduce 574) (("*") reduce 574) (("+") reduce 574) ((",") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_DOUBLE_ARROW) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("|") reduce 574)) 5590 ((("%") reduce 578) (("(") reduce 578) (("*") reduce 578) (("+") reduce 578) ((",") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_DOUBLE_ARROW) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("|") reduce 578)) 5591 ((("%") reduce 575) (("(") reduce 575) (("*") reduce 575) (("+") reduce 575) ((",") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_DOUBLE_ARROW) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("|") reduce 575)) 5592 ((("%") reduce 576) (("(") reduce 576) (("*") reduce 576) (("+") reduce 576) ((",") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_DOUBLE_ARROW) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("|") reduce 576)) 5593 ((("%") reduce 399) (("(") reduce 399) (("*") reduce 399) (("+") reduce 399) ((",") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_DOUBLE_ARROW) reduce 399) ((T_INSTANCEOF) reduce 399) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 399) ((T_OBJECT_OPERATOR) reduce 399) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 399) ((T_PIPE) reduce 399) ((T_POW) reduce 399) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("[") reduce 399) (("^") reduce 399) (("|") reduce 399)) 5595 ((("%") reduce 577) (("(") reduce 577) (("*") reduce 577) (("+") reduce 577) ((",") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_DOUBLE_ARROW) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("|") reduce 577)) 5596 ((("%") reduce 401) (("(") reduce 401) (("*") reduce 401) (("+") reduce 401) ((",") reduce 401) (("-") reduce 401) ((".") reduce 401) (("/") reduce 401) (("<") reduce 401) ((">") reduce 401) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_DOUBLE_ARROW) reduce 401) ((T_INSTANCEOF) reduce 401) ((T_IS_EQUAL) reduce 401) ((T_IS_GREATER_OR_EQUAL) reduce 401) ((T_IS_IDENTICAL) reduce 401) ((T_IS_NOT_EQUAL) reduce 401) ((T_IS_NOT_IDENTICAL) reduce 401) ((T_IS_SMALLER_OR_EQUAL) reduce 401) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 401) ((T_OBJECT_OPERATOR) reduce 401) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 401) ((T_PIPE) reduce 401) ((T_POW) reduce 401) ((T_SL) reduce 401) ((T_SPACESHIP) reduce 401) ((T_SR) reduce 401) (("[") reduce 401) (("^") reduce 401) (("|") reduce 401)) 5604 ((("%") reduce 398) (("(") reduce 398) (("*") reduce 398) (("+") reduce 398) ((",") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_DOUBLE_ARROW) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 398) ((T_OBJECT_OPERATOR) reduce 398) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 398) ((T_PIPE) reduce 398) ((T_POW) reduce 398) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("[") reduce 398) (("^") reduce 398) (("|") reduce 398)) 5606 ((("%") reduce 513) (("(") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("|") reduce 513)) 5612 ((("%") reduce 233) (("*") reduce 233) (("+") reduce 233) ((",") reduce 233) (("-") reduce 233) ((".") reduce 233) (("/") reduce 233) (("<") reduce 233) ((">") reduce 233) (("?") reduce 233) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_BOOLEAN_AND) reduce 233) ((T_BOOLEAN_OR) reduce 233) ((T_COALESCE) reduce 233) ((T_DOUBLE_ARROW) reduce 233) ((T_INSTANCEOF) reduce 233) ((T_IS_EQUAL) reduce 233) ((T_IS_GREATER_OR_EQUAL) reduce 233) ((T_IS_IDENTICAL) reduce 233) ((T_IS_NOT_EQUAL) reduce 233) ((T_IS_NOT_IDENTICAL) reduce 233) ((T_IS_SMALLER_OR_EQUAL) reduce 233) ((T_LOGICAL_AND) reduce 233) ((T_LOGICAL_OR) reduce 233) ((T_LOGICAL_XOR) reduce 233) ((T_PIPE) reduce 233) ((T_POW) reduce 233) ((T_SL) reduce 233) ((T_SPACESHIP) reduce 233) ((T_SR) reduce 233) (("^") reduce 233) (("|") reduce 233)) 5617 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 404) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5621 ((("%") reduce 613) (("*") reduce 613) (("+") reduce 613) ((",") reduce 613) (("-") reduce 613) ((".") reduce 613) (("/") reduce 613) (("<") reduce 613) ((">") reduce 613) (("?") reduce 613) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_BOOLEAN_AND) reduce 613) ((T_BOOLEAN_OR) reduce 613) ((T_COALESCE) reduce 613) ((T_DOUBLE_ARROW) reduce 613) ((T_INSTANCEOF) reduce 613) ((T_IS_EQUAL) reduce 613) ((T_IS_GREATER_OR_EQUAL) reduce 613) ((T_IS_IDENTICAL) reduce 613) ((T_IS_NOT_EQUAL) reduce 613) ((T_IS_NOT_IDENTICAL) reduce 613) ((T_IS_SMALLER_OR_EQUAL) reduce 613) ((T_LOGICAL_AND) reduce 613) ((T_LOGICAL_OR) reduce 613) ((T_LOGICAL_XOR) reduce 613) ((T_PIPE) reduce 613) ((T_POW) reduce 613) ((T_SL) reduce 613) ((T_SPACESHIP) reduce 613) ((T_SR) reduce 613) (("^") reduce 613) (("|") reduce 613)) 5622 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) ((",") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_DOUBLE_ARROW) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_PIPE) reduce 467) ((T_POW) shift) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("|") reduce 467)) 5623 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 616) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 616) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5624 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 615) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 615) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5625 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) ((",") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_DOUBLE_ARROW) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("[") reduce 548) (("^") reduce 424) (("|") reduce 424)) 5627 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_PIPE) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("^") reduce 518) (("|") reduce 518)) 5628 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) ((",") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_DOUBLE_ARROW) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_PIPE) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("|") reduce 474)) 5629 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("|") reduce 291)) 5633 ((("%") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("^") reduce 292) (("|") reduce 292)) 5634 ((("%") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("^") reduce 293) (("|") reduce 293)) 5637 ((("%") reduce 617) (("*") reduce 617) (("+") reduce 617) ((",") reduce 617) (("-") reduce 617) ((".") reduce 617) (("/") reduce 617) (("<") reduce 617) ((">") reduce 617) (("?") reduce 617) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_BOOLEAN_AND) reduce 617) ((T_BOOLEAN_OR) reduce 617) ((T_COALESCE) reduce 617) ((T_DOUBLE_ARROW) reduce 617) ((T_INSTANCEOF) reduce 617) ((T_IS_EQUAL) reduce 617) ((T_IS_GREATER_OR_EQUAL) reduce 617) ((T_IS_IDENTICAL) reduce 617) ((T_IS_NOT_EQUAL) reduce 617) ((T_IS_NOT_IDENTICAL) reduce 617) ((T_IS_SMALLER_OR_EQUAL) reduce 617) ((T_LOGICAL_AND) reduce 617) ((T_LOGICAL_OR) reduce 617) ((T_LOGICAL_XOR) reduce 617) ((T_PIPE) reduce 617) ((T_POW) reduce 617) ((T_SL) reduce 617) ((T_SPACESHIP) reduce 617) ((T_SR) reduce 617) (("^") reduce 617) (("|") reduce 617)) 5640 ((("%") reduce 614) (("*") reduce 614) (("+") reduce 614) ((",") reduce 614) (("-") reduce 614) ((".") reduce 614) (("/") reduce 614) (("<") reduce 614) ((">") reduce 614) (("?") reduce 614) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_BOOLEAN_AND) reduce 614) ((T_BOOLEAN_OR) reduce 614) ((T_COALESCE) reduce 614) ((T_DOUBLE_ARROW) reduce 614) ((T_INSTANCEOF) reduce 614) ((T_IS_EQUAL) reduce 614) ((T_IS_GREATER_OR_EQUAL) reduce 614) ((T_IS_IDENTICAL) reduce 614) ((T_IS_NOT_EQUAL) reduce 614) ((T_IS_NOT_IDENTICAL) reduce 614) ((T_IS_SMALLER_OR_EQUAL) reduce 614) ((T_LOGICAL_AND) reduce 614) ((T_LOGICAL_OR) reduce 614) ((T_LOGICAL_XOR) reduce 614) ((T_PIPE) reduce 614) ((T_POW) reduce 614) ((T_SL) reduce 614) ((T_SPACESHIP) reduce 614) ((T_SR) reduce 614) (("^") reduce 614) (("|") reduce 614)) 5641 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) ((",") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_DOUBLE_ARROW) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_PIPE) reduce 468) ((T_POW) shift) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 5642 (((T_DOUBLE_ARROW) reduce 126)) 5645 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 239) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 239)) 5646 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) ((",") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_DOUBLE_ARROW) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("[") reduce 548) (("^") reduce 426) (("|") reduce 426)) 5648 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) ((",") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_DOUBLE_ARROW) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_PIPE) reduce 408) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("^") reduce 408) (("|") reduce 408)) 5649 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) ((",") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_DOUBLE_ARROW) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_PIPE) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("|") reduce 409)) 5650 ((("%") reduce 296) (("*") reduce 296) (("+") reduce 296) ((",") reduce 296) (("-") reduce 296) ((".") reduce 296) (("/") reduce 296) (("<") reduce 296) ((">") reduce 296) (("?") reduce 296) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_BOOLEAN_AND) reduce 296) ((T_BOOLEAN_OR) reduce 296) ((T_COALESCE) reduce 296) ((T_DOUBLE_ARROW) reduce 296) ((T_INSTANCEOF) reduce 296) ((T_IS_EQUAL) reduce 296) ((T_IS_GREATER_OR_EQUAL) reduce 296) ((T_IS_IDENTICAL) reduce 296) ((T_IS_NOT_EQUAL) reduce 296) ((T_IS_NOT_IDENTICAL) reduce 296) ((T_IS_SMALLER_OR_EQUAL) reduce 296) ((T_LOGICAL_AND) reduce 296) ((T_LOGICAL_OR) reduce 296) ((T_LOGICAL_XOR) reduce 296) ((T_PIPE) reduce 296) ((T_POW) reduce 296) ((T_SL) reduce 296) ((T_SPACESHIP) reduce 296) ((T_SR) reduce 296) (("^") reduce 296) (("|") reduce 296)) 5655 ((("%") reduce 297) (("*") reduce 297) (("+") reduce 297) ((",") reduce 297) (("-") reduce 297) ((".") reduce 297) (("/") reduce 297) (("<") reduce 297) ((">") reduce 297) (("?") reduce 297) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_BOOLEAN_AND) reduce 297) ((T_BOOLEAN_OR) reduce 297) ((T_COALESCE) reduce 297) ((T_DOUBLE_ARROW) reduce 297) ((T_INSTANCEOF) reduce 297) ((T_IS_EQUAL) reduce 297) ((T_IS_GREATER_OR_EQUAL) reduce 297) ((T_IS_IDENTICAL) reduce 297) ((T_IS_NOT_EQUAL) reduce 297) ((T_IS_NOT_IDENTICAL) reduce 297) ((T_IS_SMALLER_OR_EQUAL) reduce 297) ((T_LOGICAL_AND) reduce 297) ((T_LOGICAL_OR) reduce 297) ((T_LOGICAL_XOR) reduce 297) ((T_PIPE) reduce 297) ((T_POW) reduce 297) ((T_SL) reduce 297) ((T_SPACESHIP) reduce 297) ((T_SR) reduce 297) (("^") reduce 297) (("|") reduce 297)) 5656 ((("%") reduce 460) (("(") reduce 556) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PIPE) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("[") reduce 549) (("^") reduce 460) (("|") reduce 460)) 5658 ((("%") reduce 298) (("*") reduce 298) (("+") reduce 298) ((",") reduce 298) (("-") reduce 298) ((".") reduce 298) (("/") reduce 298) (("<") reduce 298) ((">") reduce 298) (("?") reduce 298) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_BOOLEAN_AND) reduce 298) ((T_BOOLEAN_OR) reduce 298) ((T_COALESCE) reduce 298) ((T_DOUBLE_ARROW) reduce 298) ((T_INSTANCEOF) reduce 298) ((T_IS_EQUAL) reduce 298) ((T_IS_GREATER_OR_EQUAL) reduce 298) ((T_IS_IDENTICAL) reduce 298) ((T_IS_NOT_EQUAL) reduce 298) ((T_IS_NOT_IDENTICAL) reduce 298) ((T_IS_SMALLER_OR_EQUAL) reduce 298) ((T_LOGICAL_AND) reduce 298) ((T_LOGICAL_OR) reduce 298) ((T_LOGICAL_XOR) reduce 298) ((T_PIPE) reduce 298) ((T_POW) reduce 298) ((T_SL) reduce 298) ((T_SPACESHIP) reduce 298) ((T_SR) reduce 298) (("^") reduce 298) (("|") reduce 298)) 5659 ((("%") reduce 299) (("*") reduce 299) (("+") reduce 299) ((",") reduce 299) (("-") reduce 299) ((".") reduce 299) (("/") reduce 299) (("<") reduce 299) ((">") reduce 299) (("?") reduce 299) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_BOOLEAN_AND) reduce 299) ((T_BOOLEAN_OR) reduce 299) ((T_COALESCE) reduce 299) ((T_DOUBLE_ARROW) reduce 299) ((T_INSTANCEOF) reduce 299) ((T_IS_EQUAL) reduce 299) ((T_IS_GREATER_OR_EQUAL) reduce 299) ((T_IS_IDENTICAL) reduce 299) ((T_IS_NOT_EQUAL) reduce 299) ((T_IS_NOT_IDENTICAL) reduce 299) ((T_IS_SMALLER_OR_EQUAL) reduce 299) ((T_LOGICAL_AND) reduce 299) ((T_LOGICAL_OR) reduce 299) ((T_LOGICAL_XOR) reduce 299) ((T_PIPE) reduce 299) ((T_POW) reduce 299) ((T_SL) reduce 299) ((T_SPACESHIP) reduce 299) ((T_SR) reduce 299) (("^") reduce 299) (("|") reduce 299)) 5660 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) ((",") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_DOUBLE_ARROW) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_PIPE) reduce 472) ((T_POW) shift) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("|") reduce 472)) 5661 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) ((",") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_DOUBLE_ARROW) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_PIPE) reduce 470) ((T_POW) shift) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("|") reduce 470)) 5664 ((("%") reduce 519) (("(") reduce 519) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PIPE) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("|") reduce 519)) 5665 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) ((",") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_DOUBLE_ARROW) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_PIPE) reduce 475) ((T_POW) shift) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("|") reduce 475)) 5666 ((("%") reduce 446) (("*") reduce 446) (("+") reduce 446) ((",") reduce 446) (("-") reduce 446) ((".") reduce 446) (("/") reduce 446) (("<") reduce 446) ((">") reduce 446) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_BOOLEAN_AND) reduce 446) ((T_BOOLEAN_OR) reduce 446) ((T_COALESCE) reduce 446) ((T_DOUBLE_ARROW) reduce 446) ((T_INSTANCEOF) reduce 446) ((T_IS_EQUAL) reduce 446) ((T_IS_GREATER_OR_EQUAL) reduce 446) ((T_IS_IDENTICAL) reduce 446) ((T_IS_NOT_EQUAL) reduce 446) ((T_IS_NOT_IDENTICAL) reduce 446) ((T_IS_SMALLER_OR_EQUAL) reduce 446) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_PIPE) reduce 446) ((T_POW) shift) ((T_SL) reduce 446) ((T_SPACESHIP) reduce 446) ((T_SR) reduce 446) (("^") reduce 446) (("|") reduce 446)) 5667 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) ((",") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_DOUBLE_ARROW) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_PIPE) reduce 445) ((T_POW) shift) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("|") reduce 445)) 5669 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) (("=") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AND_EQUAL) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_COALESCE_EQUAL) reduce 570) ((T_CONCAT_EQUAL) reduce 570) ((T_DEC) reduce 570) ((T_DIV_EQUAL) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INC) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_MINUS_EQUAL) reduce 570) ((T_MOD_EQUAL) reduce 570) ((T_MUL_EQUAL) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_OR_EQUAL) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_PLUS_EQUAL) reduce 570) ((T_POW) reduce 570) ((T_POW_EQUAL) reduce 570) ((T_SL) reduce 570) ((T_SL_EQUAL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) ((T_SR_EQUAL) reduce 570) ((T_XOR_EQUAL) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 5672 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) (("=") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AND_EQUAL) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_COALESCE_EQUAL) reduce 569) ((T_CONCAT_EQUAL) reduce 569) ((T_DEC) reduce 569) ((T_DIV_EQUAL) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INC) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_MINUS_EQUAL) reduce 569) ((T_MOD_EQUAL) reduce 569) ((T_MUL_EQUAL) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_OR_EQUAL) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_PLUS_EQUAL) reduce 569) ((T_POW) reduce 569) ((T_POW_EQUAL) reduce 569) ((T_SL) reduce 569) ((T_SL_EQUAL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) ((T_SR_EQUAL) reduce 569) ((T_XOR_EQUAL) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 5674 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_PIPE) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("|") reduce 522)) 5675 ((("%") reduce 447) (("*") reduce 447) (("+") reduce 447) ((",") reduce 447) (("-") reduce 447) ((".") reduce 447) (("/") reduce 447) (("<") reduce 447) ((">") reduce 447) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_BOOLEAN_AND) reduce 447) ((T_BOOLEAN_OR) reduce 447) ((T_COALESCE) reduce 447) ((T_DOUBLE_ARROW) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 447) ((T_IS_GREATER_OR_EQUAL) reduce 447) ((T_IS_IDENTICAL) reduce 447) ((T_IS_NOT_EQUAL) reduce 447) ((T_IS_NOT_IDENTICAL) reduce 447) ((T_IS_SMALLER_OR_EQUAL) reduce 447) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_PIPE) reduce 447) ((T_POW) shift) ((T_SL) reduce 447) ((T_SPACESHIP) reduce 447) ((T_SR) reduce 447) (("^") reduce 447) (("|") reduce 447)) 5680 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 404) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 404) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5684 ((("%") reduce 613) (("*") reduce 613) (("+") reduce 613) ((",") reduce 613) (("-") reduce 613) ((".") reduce 613) (("/") reduce 613) ((";") reduce 613) (("<") reduce 613) ((">") reduce 613) (("?") reduce 613) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_BOOLEAN_AND) reduce 613) ((T_BOOLEAN_OR) reduce 613) ((T_COALESCE) reduce 613) ((T_DOUBLE_ARROW) reduce 613) ((T_INSTANCEOF) reduce 613) ((T_IS_EQUAL) reduce 613) ((T_IS_GREATER_OR_EQUAL) reduce 613) ((T_IS_IDENTICAL) reduce 613) ((T_IS_NOT_EQUAL) reduce 613) ((T_IS_NOT_IDENTICAL) reduce 613) ((T_IS_SMALLER_OR_EQUAL) reduce 613) ((T_LOGICAL_AND) reduce 613) ((T_LOGICAL_OR) reduce 613) ((T_LOGICAL_XOR) reduce 613) ((T_PIPE) reduce 613) ((T_POW) reduce 613) ((T_SL) reduce 613) ((T_SPACESHIP) reduce 613) ((T_SR) reduce 613) (("^") reduce 613) (("|") reduce 613)) 5685 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) ((",") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) ((";") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_DOUBLE_ARROW) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_PIPE) reduce 467) ((T_POW) shift) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("|") reduce 467)) 5686 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 616) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 616) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 616) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5687 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 615) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 615) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 615) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5688 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) ((",") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) ((";") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_DOUBLE_ARROW) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("[") reduce 548) (("^") reduce 424) (("|") reduce 424)) 5690 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) ((";") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_PIPE) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("^") reduce 518) (("|") reduce 518)) 5691 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) ((",") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) ((";") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_DOUBLE_ARROW) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_PIPE) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("|") reduce 474)) 5692 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((";") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("|") reduce 291)) 5696 ((("%") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) ((";") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("^") reduce 292) (("|") reduce 292)) 5697 ((("%") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) ((";") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("^") reduce 293) (("|") reduce 293)) 5700 ((("%") reduce 617) (("*") reduce 617) (("+") reduce 617) ((",") reduce 617) (("-") reduce 617) ((".") reduce 617) (("/") reduce 617) ((";") reduce 617) (("<") reduce 617) ((">") reduce 617) (("?") reduce 617) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_BOOLEAN_AND) reduce 617) ((T_BOOLEAN_OR) reduce 617) ((T_COALESCE) reduce 617) ((T_DOUBLE_ARROW) reduce 617) ((T_INSTANCEOF) reduce 617) ((T_IS_EQUAL) reduce 617) ((T_IS_GREATER_OR_EQUAL) reduce 617) ((T_IS_IDENTICAL) reduce 617) ((T_IS_NOT_EQUAL) reduce 617) ((T_IS_NOT_IDENTICAL) reduce 617) ((T_IS_SMALLER_OR_EQUAL) reduce 617) ((T_LOGICAL_AND) reduce 617) ((T_LOGICAL_OR) reduce 617) ((T_LOGICAL_XOR) reduce 617) ((T_PIPE) reduce 617) ((T_POW) reduce 617) ((T_SL) reduce 617) ((T_SPACESHIP) reduce 617) ((T_SR) reduce 617) (("^") reduce 617) (("|") reduce 617)) 5703 ((("%") reduce 614) (("*") reduce 614) (("+") reduce 614) ((",") reduce 614) (("-") reduce 614) ((".") reduce 614) (("/") reduce 614) ((";") reduce 614) (("<") reduce 614) ((">") reduce 614) (("?") reduce 614) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_BOOLEAN_AND) reduce 614) ((T_BOOLEAN_OR) reduce 614) ((T_COALESCE) reduce 614) ((T_DOUBLE_ARROW) reduce 614) ((T_INSTANCEOF) reduce 614) ((T_IS_EQUAL) reduce 614) ((T_IS_GREATER_OR_EQUAL) reduce 614) ((T_IS_IDENTICAL) reduce 614) ((T_IS_NOT_EQUAL) reduce 614) ((T_IS_NOT_IDENTICAL) reduce 614) ((T_IS_SMALLER_OR_EQUAL) reduce 614) ((T_LOGICAL_AND) reduce 614) ((T_LOGICAL_OR) reduce 614) ((T_LOGICAL_XOR) reduce 614) ((T_PIPE) reduce 614) ((T_POW) reduce 614) ((T_SL) reduce 614) ((T_SPACESHIP) reduce 614) ((T_SR) reduce 614) (("^") reduce 614) (("|") reduce 614)) 5704 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) ((",") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) ((";") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_DOUBLE_ARROW) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_PIPE) reduce 468) ((T_POW) shift) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 5705 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) ((",") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) ((";") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_DOUBLE_ARROW) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("[") reduce 548) (("^") reduce 426) (("|") reduce 426)) 5707 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) ((",") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) ((";") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_DOUBLE_ARROW) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_PIPE) reduce 408) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("^") reduce 408) (("|") reduce 408)) 5708 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) ((",") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) ((";") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_DOUBLE_ARROW) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_PIPE) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("|") reduce 409)) 5709 ((("%") reduce 296) (("*") reduce 296) (("+") reduce 296) ((",") reduce 296) (("-") reduce 296) ((".") reduce 296) (("/") reduce 296) ((";") reduce 296) (("<") reduce 296) ((">") reduce 296) (("?") reduce 296) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_BOOLEAN_AND) reduce 296) ((T_BOOLEAN_OR) reduce 296) ((T_COALESCE) reduce 296) ((T_DOUBLE_ARROW) reduce 296) ((T_INSTANCEOF) reduce 296) ((T_IS_EQUAL) reduce 296) ((T_IS_GREATER_OR_EQUAL) reduce 296) ((T_IS_IDENTICAL) reduce 296) ((T_IS_NOT_EQUAL) reduce 296) ((T_IS_NOT_IDENTICAL) reduce 296) ((T_IS_SMALLER_OR_EQUAL) reduce 296) ((T_LOGICAL_AND) reduce 296) ((T_LOGICAL_OR) reduce 296) ((T_LOGICAL_XOR) reduce 296) ((T_PIPE) reduce 296) ((T_POW) reduce 296) ((T_SL) reduce 296) ((T_SPACESHIP) reduce 296) ((T_SR) reduce 296) (("^") reduce 296) (("|") reduce 296)) 5714 ((("%") reduce 297) (("*") reduce 297) (("+") reduce 297) ((",") reduce 297) (("-") reduce 297) ((".") reduce 297) (("/") reduce 297) ((";") reduce 297) (("<") reduce 297) ((">") reduce 297) (("?") reduce 297) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_BOOLEAN_AND) reduce 297) ((T_BOOLEAN_OR) reduce 297) ((T_COALESCE) reduce 297) ((T_DOUBLE_ARROW) reduce 297) ((T_INSTANCEOF) reduce 297) ((T_IS_EQUAL) reduce 297) ((T_IS_GREATER_OR_EQUAL) reduce 297) ((T_IS_IDENTICAL) reduce 297) ((T_IS_NOT_EQUAL) reduce 297) ((T_IS_NOT_IDENTICAL) reduce 297) ((T_IS_SMALLER_OR_EQUAL) reduce 297) ((T_LOGICAL_AND) reduce 297) ((T_LOGICAL_OR) reduce 297) ((T_LOGICAL_XOR) reduce 297) ((T_PIPE) reduce 297) ((T_POW) reduce 297) ((T_SL) reduce 297) ((T_SPACESHIP) reduce 297) ((T_SR) reduce 297) (("^") reduce 297) (("|") reduce 297)) 5715 ((("%") reduce 460) (("(") reduce 556) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) ((";") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PIPE) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("[") reduce 549) (("^") reduce 460) (("|") reduce 460)) 5717 ((("%") reduce 298) (("*") reduce 298) (("+") reduce 298) ((",") reduce 298) (("-") reduce 298) ((".") reduce 298) (("/") reduce 298) ((";") reduce 298) (("<") reduce 298) ((">") reduce 298) (("?") reduce 298) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_BOOLEAN_AND) reduce 298) ((T_BOOLEAN_OR) reduce 298) ((T_COALESCE) reduce 298) ((T_DOUBLE_ARROW) reduce 298) ((T_INSTANCEOF) reduce 298) ((T_IS_EQUAL) reduce 298) ((T_IS_GREATER_OR_EQUAL) reduce 298) ((T_IS_IDENTICAL) reduce 298) ((T_IS_NOT_EQUAL) reduce 298) ((T_IS_NOT_IDENTICAL) reduce 298) ((T_IS_SMALLER_OR_EQUAL) reduce 298) ((T_LOGICAL_AND) reduce 298) ((T_LOGICAL_OR) reduce 298) ((T_LOGICAL_XOR) reduce 298) ((T_PIPE) reduce 298) ((T_POW) reduce 298) ((T_SL) reduce 298) ((T_SPACESHIP) reduce 298) ((T_SR) reduce 298) (("^") reduce 298) (("|") reduce 298)) 5718 ((("%") reduce 299) (("*") reduce 299) (("+") reduce 299) ((",") reduce 299) (("-") reduce 299) ((".") reduce 299) (("/") reduce 299) ((";") reduce 299) (("<") reduce 299) ((">") reduce 299) (("?") reduce 299) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_BOOLEAN_AND) reduce 299) ((T_BOOLEAN_OR) reduce 299) ((T_COALESCE) reduce 299) ((T_DOUBLE_ARROW) reduce 299) ((T_INSTANCEOF) reduce 299) ((T_IS_EQUAL) reduce 299) ((T_IS_GREATER_OR_EQUAL) reduce 299) ((T_IS_IDENTICAL) reduce 299) ((T_IS_NOT_EQUAL) reduce 299) ((T_IS_NOT_IDENTICAL) reduce 299) ((T_IS_SMALLER_OR_EQUAL) reduce 299) ((T_LOGICAL_AND) reduce 299) ((T_LOGICAL_OR) reduce 299) ((T_LOGICAL_XOR) reduce 299) ((T_PIPE) reduce 299) ((T_POW) reduce 299) ((T_SL) reduce 299) ((T_SPACESHIP) reduce 299) ((T_SR) reduce 299) (("^") reduce 299) (("|") reduce 299)) 5719 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) ((",") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) ((";") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_DOUBLE_ARROW) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_PIPE) reduce 472) ((T_POW) shift) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("|") reduce 472)) 5720 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) ((",") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) ((";") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_DOUBLE_ARROW) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_PIPE) reduce 470) ((T_POW) shift) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("|") reduce 470)) 5723 ((("%") reduce 519) (("(") reduce 519) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) ((";") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PIPE) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("|") reduce 519)) 5724 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) ((",") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) ((";") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_DOUBLE_ARROW) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_PIPE) reduce 475) ((T_POW) shift) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("|") reduce 475)) 5725 ((("%") reduce 446) (("*") reduce 446) (("+") reduce 446) ((",") reduce 446) (("-") reduce 446) ((".") reduce 446) (("/") reduce 446) ((";") reduce 446) (("<") reduce 446) ((">") reduce 446) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_BOOLEAN_AND) reduce 446) ((T_BOOLEAN_OR) reduce 446) ((T_COALESCE) reduce 446) ((T_DOUBLE_ARROW) reduce 446) ((T_INSTANCEOF) reduce 446) ((T_IS_EQUAL) reduce 446) ((T_IS_GREATER_OR_EQUAL) reduce 446) ((T_IS_IDENTICAL) reduce 446) ((T_IS_NOT_EQUAL) reduce 446) ((T_IS_NOT_IDENTICAL) reduce 446) ((T_IS_SMALLER_OR_EQUAL) reduce 446) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_PIPE) reduce 446) ((T_POW) shift) ((T_SL) reduce 446) ((T_SPACESHIP) reduce 446) ((T_SR) reduce 446) (("^") reduce 446) (("|") reduce 446)) 5726 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) ((",") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) ((";") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_DOUBLE_ARROW) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_PIPE) reduce 445) ((T_POW) shift) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("|") reduce 445)) 5728 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) ((";") reduce 570) (("<") reduce 570) (("=") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AND_EQUAL) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_COALESCE_EQUAL) reduce 570) ((T_CONCAT_EQUAL) reduce 570) ((T_DEC) reduce 570) ((T_DIV_EQUAL) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INC) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_MINUS_EQUAL) reduce 570) ((T_MOD_EQUAL) reduce 570) ((T_MUL_EQUAL) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_OR_EQUAL) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_PLUS_EQUAL) reduce 570) ((T_POW) reduce 570) ((T_POW_EQUAL) reduce 570) ((T_SL) reduce 570) ((T_SL_EQUAL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) ((T_SR_EQUAL) reduce 570) ((T_XOR_EQUAL) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 5731 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) ((";") reduce 569) (("<") reduce 569) (("=") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AND_EQUAL) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_COALESCE_EQUAL) reduce 569) ((T_CONCAT_EQUAL) reduce 569) ((T_DEC) reduce 569) ((T_DIV_EQUAL) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INC) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_MINUS_EQUAL) reduce 569) ((T_MOD_EQUAL) reduce 569) ((T_MUL_EQUAL) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_OR_EQUAL) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_PLUS_EQUAL) reduce 569) ((T_POW) reduce 569) ((T_POW_EQUAL) reduce 569) ((T_SL) reduce 569) ((T_SL_EQUAL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) ((T_SR_EQUAL) reduce 569) ((T_XOR_EQUAL) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 5733 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) ((";") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_PIPE) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("|") reduce 522)) 5734 ((("%") reduce 447) (("*") reduce 447) (("+") reduce 447) ((",") reduce 447) (("-") reduce 447) ((".") reduce 447) (("/") reduce 447) ((";") reduce 447) (("<") reduce 447) ((">") reduce 447) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_BOOLEAN_AND) reduce 447) ((T_BOOLEAN_OR) reduce 447) ((T_COALESCE) reduce 447) ((T_DOUBLE_ARROW) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 447) ((T_IS_GREATER_OR_EQUAL) reduce 447) ((T_IS_IDENTICAL) reduce 447) ((T_IS_NOT_EQUAL) reduce 447) ((T_IS_NOT_IDENTICAL) reduce 447) ((T_IS_SMALLER_OR_EQUAL) reduce 447) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_PIPE) reduce 447) ((T_POW) shift) ((T_SL) reduce 447) ((T_SPACESHIP) reduce 447) ((T_SR) reduce 447) (("^") reduce 447) (("|") reduce 447)) 5735 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) ((",") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) ((";") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_PIPE) reduce 473) ((T_POW) shift) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("|") reduce 473)) 5736 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 483) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 483) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5737 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) ((",") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) ((";") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_PIPE) reduce 469) ((T_POW) shift) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 5738 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) ((",") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) ((";") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_PIPE) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("^") reduce 486) (("|") reduce 486)) 5740 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) ((";") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_PIPE) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("^") reduce 526) (("|") reduce 526)) 5742 ((("%") reduce 527) (("*") reduce 527) (("+") reduce 527) ((",") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) ((";") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_PIPE) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("^") reduce 527) (("|") reduce 527)) 5743 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) ((";") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_PIPE) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("^") reduce 525) (("|") reduce 525)) 5744 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 619) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 619) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5745 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 618) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 618) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5746 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((";") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 5747 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 478) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 478) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5748 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) ((",") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) ((";") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_PIPE) reduce 471) ((T_POW) shift) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("|") reduce 471)) 5750 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 5751 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 5752 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 5753 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((";") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 5754 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 5755 ((("%") reduce 400) (("(") reduce 400) (("*") reduce 400) (("+") reduce 400) ((",") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) ((";") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 400) ((T_OBJECT_OPERATOR) reduce 400) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 400) ((T_PIPE) reduce 400) ((T_POW) reduce 400) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("[") reduce 400) (("^") reduce 400) (("|") reduce 400)) 5758 ((("%") reduce 511) (("(") reduce 511) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((";") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("|") reduce 511)) 5759 ((("%") reduce 402) (("(") shift) (("*") reduce 402) (("+") reduce 402) ((",") reduce 402) (("-") reduce 402) ((".") reduce 402) (("/") reduce 402) ((";") reduce 402) (("<") reduce 402) ((">") reduce 402) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_INSTANCEOF) reduce 402) ((T_IS_EQUAL) reduce 402) ((T_IS_GREATER_OR_EQUAL) reduce 402) ((T_IS_IDENTICAL) reduce 402) ((T_IS_NOT_EQUAL) reduce 402) ((T_IS_NOT_IDENTICAL) reduce 402) ((T_IS_SMALLER_OR_EQUAL) reduce 402) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_PIPE) reduce 402) ((T_POW) reduce 402) ((T_SL) reduce 402) ((T_SPACESHIP) reduce 402) ((T_SR) reduce 402) (("^") reduce 402) (("|") reduce 402)) 5760 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((";") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 5761 ((("%") reduce 512) (("(") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((";") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("|") reduce 512)) 5762 ((("%") reduce 573) (("(") reduce 573) (("*") reduce 573) (("+") reduce 573) ((",") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) ((";") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("|") reduce 573)) 5768 ((("%") reduce 574) (("(") reduce 574) (("*") reduce 574) (("+") reduce 574) ((",") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) ((";") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("|") reduce 574)) 5769 ((("%") reduce 578) (("(") reduce 578) (("*") reduce 578) (("+") reduce 578) ((",") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) ((";") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("|") reduce 578)) 5770 ((("%") reduce 575) (("(") reduce 575) (("*") reduce 575) (("+") reduce 575) ((",") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) ((";") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("|") reduce 575)) 5771 ((("%") reduce 576) (("(") reduce 576) (("*") reduce 576) (("+") reduce 576) ((",") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) ((";") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("|") reduce 576)) 5772 ((("%") reduce 399) (("(") reduce 399) (("*") reduce 399) (("+") reduce 399) ((",") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) ((";") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_INSTANCEOF) reduce 399) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 399) ((T_OBJECT_OPERATOR) reduce 399) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 399) ((T_PIPE) reduce 399) ((T_POW) reduce 399) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("[") reduce 399) (("^") reduce 399) (("|") reduce 399)) 5774 ((("%") reduce 577) (("(") reduce 577) (("*") reduce 577) (("+") reduce 577) ((",") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) ((";") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("|") reduce 577)) 5775 ((("%") reduce 401) (("(") reduce 401) (("*") reduce 401) (("+") reduce 401) ((",") reduce 401) (("-") reduce 401) ((".") reduce 401) (("/") reduce 401) ((";") reduce 401) (("<") reduce 401) ((">") reduce 401) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_INSTANCEOF) reduce 401) ((T_IS_EQUAL) reduce 401) ((T_IS_GREATER_OR_EQUAL) reduce 401) ((T_IS_IDENTICAL) reduce 401) ((T_IS_NOT_EQUAL) reduce 401) ((T_IS_NOT_IDENTICAL) reduce 401) ((T_IS_SMALLER_OR_EQUAL) reduce 401) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 401) ((T_OBJECT_OPERATOR) reduce 401) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 401) ((T_PIPE) reduce 401) ((T_POW) reduce 401) ((T_SL) reduce 401) ((T_SPACESHIP) reduce 401) ((T_SR) reduce 401) (("[") reduce 401) (("^") reduce 401) (("|") reduce 401)) 5783 ((("%") reduce 398) (("(") reduce 398) (("*") reduce 398) (("+") reduce 398) ((",") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) ((";") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 398) ((T_OBJECT_OPERATOR) reduce 398) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 398) ((T_PIPE) reduce 398) ((T_POW) reduce 398) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("[") reduce 398) (("^") reduce 398) (("|") reduce 398)) 5785 ((("%") reduce 513) (("(") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((";") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("|") reduce 513)) 5791 ((("%") reduce 233) (("*") reduce 233) (("+") reduce 233) ((",") reduce 233) (("-") reduce 233) ((".") reduce 233) (("/") reduce 233) ((";") reduce 233) (("<") reduce 233) ((">") reduce 233) (("?") reduce 233) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_BOOLEAN_AND) reduce 233) ((T_BOOLEAN_OR) reduce 233) ((T_COALESCE) reduce 233) ((T_INSTANCEOF) reduce 233) ((T_IS_EQUAL) reduce 233) ((T_IS_GREATER_OR_EQUAL) reduce 233) ((T_IS_IDENTICAL) reduce 233) ((T_IS_NOT_EQUAL) reduce 233) ((T_IS_NOT_IDENTICAL) reduce 233) ((T_IS_SMALLER_OR_EQUAL) reduce 233) ((T_LOGICAL_AND) reduce 233) ((T_LOGICAL_OR) reduce 233) ((T_LOGICAL_XOR) reduce 233) ((T_PIPE) reduce 233) ((T_POW) reduce 233) ((T_SL) reduce 233) ((T_SPACESHIP) reduce 233) ((T_SR) reduce 233) (("^") reduce 233) (("|") reduce 233)) 5796 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 404) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 404) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5800 ((("%") reduce 613) (("*") reduce 613) (("+") reduce 613) ((",") reduce 613) (("-") reduce 613) ((".") reduce 613) (("/") reduce 613) ((";") reduce 613) (("<") reduce 613) ((">") reduce 613) (("?") reduce 613) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_BOOLEAN_AND) reduce 613) ((T_BOOLEAN_OR) reduce 613) ((T_COALESCE) reduce 613) ((T_INSTANCEOF) reduce 613) ((T_IS_EQUAL) reduce 613) ((T_IS_GREATER_OR_EQUAL) reduce 613) ((T_IS_IDENTICAL) reduce 613) ((T_IS_NOT_EQUAL) reduce 613) ((T_IS_NOT_IDENTICAL) reduce 613) ((T_IS_SMALLER_OR_EQUAL) reduce 613) ((T_LOGICAL_AND) reduce 613) ((T_LOGICAL_OR) reduce 613) ((T_LOGICAL_XOR) reduce 613) ((T_PIPE) reduce 613) ((T_POW) reduce 613) ((T_SL) reduce 613) ((T_SPACESHIP) reduce 613) ((T_SR) reduce 613) (("^") reduce 613) (("|") reduce 613)) 5801 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) ((",") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) ((";") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_PIPE) reduce 467) ((T_POW) shift) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("|") reduce 467)) 5802 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 616) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 616) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5803 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 615) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 615) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5804 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) ((",") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) ((";") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("[") reduce 548) (("^") reduce 424) (("|") reduce 424)) 5806 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) ((";") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_PIPE) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("^") reduce 518) (("|") reduce 518)) 5807 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) ((",") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) ((";") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_PIPE) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("|") reduce 474)) 5808 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((";") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("|") reduce 291)) 5812 ((("%") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) ((";") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("^") reduce 292) (("|") reduce 292)) 5813 ((("%") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) ((";") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("^") reduce 293) (("|") reduce 293)) 5816 ((("%") reduce 617) (("*") reduce 617) (("+") reduce 617) ((",") reduce 617) (("-") reduce 617) ((".") reduce 617) (("/") reduce 617) ((";") reduce 617) (("<") reduce 617) ((">") reduce 617) (("?") reduce 617) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_BOOLEAN_AND) reduce 617) ((T_BOOLEAN_OR) reduce 617) ((T_COALESCE) reduce 617) ((T_INSTANCEOF) reduce 617) ((T_IS_EQUAL) reduce 617) ((T_IS_GREATER_OR_EQUAL) reduce 617) ((T_IS_IDENTICAL) reduce 617) ((T_IS_NOT_EQUAL) reduce 617) ((T_IS_NOT_IDENTICAL) reduce 617) ((T_IS_SMALLER_OR_EQUAL) reduce 617) ((T_LOGICAL_AND) reduce 617) ((T_LOGICAL_OR) reduce 617) ((T_LOGICAL_XOR) reduce 617) ((T_PIPE) reduce 617) ((T_POW) reduce 617) ((T_SL) reduce 617) ((T_SPACESHIP) reduce 617) ((T_SR) reduce 617) (("^") reduce 617) (("|") reduce 617)) 5819 ((("%") reduce 614) (("*") reduce 614) (("+") reduce 614) ((",") reduce 614) (("-") reduce 614) ((".") reduce 614) (("/") reduce 614) ((";") reduce 614) (("<") reduce 614) ((">") reduce 614) (("?") reduce 614) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_BOOLEAN_AND) reduce 614) ((T_BOOLEAN_OR) reduce 614) ((T_COALESCE) reduce 614) ((T_INSTANCEOF) reduce 614) ((T_IS_EQUAL) reduce 614) ((T_IS_GREATER_OR_EQUAL) reduce 614) ((T_IS_IDENTICAL) reduce 614) ((T_IS_NOT_EQUAL) reduce 614) ((T_IS_NOT_IDENTICAL) reduce 614) ((T_IS_SMALLER_OR_EQUAL) reduce 614) ((T_LOGICAL_AND) reduce 614) ((T_LOGICAL_OR) reduce 614) ((T_LOGICAL_XOR) reduce 614) ((T_PIPE) reduce 614) ((T_POW) reduce 614) ((T_SL) reduce 614) ((T_SPACESHIP) reduce 614) ((T_SR) reduce 614) (("^") reduce 614) (("|") reduce 614)) 5820 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) ((",") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) ((";") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_PIPE) reduce 468) ((T_POW) shift) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 5821 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) ((",") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) ((";") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("[") reduce 548) (("^") reduce 426) (("|") reduce 426)) 5823 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) ((",") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) ((";") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_PIPE) reduce 408) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("^") reduce 408) (("|") reduce 408)) 5824 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) ((",") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) ((";") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_PIPE) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("|") reduce 409)) 5825 ((("%") reduce 296) (("*") reduce 296) (("+") reduce 296) ((",") reduce 296) (("-") reduce 296) ((".") reduce 296) (("/") reduce 296) ((";") reduce 296) (("<") reduce 296) ((">") reduce 296) (("?") reduce 296) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_BOOLEAN_AND) reduce 296) ((T_BOOLEAN_OR) reduce 296) ((T_COALESCE) reduce 296) ((T_INSTANCEOF) reduce 296) ((T_IS_EQUAL) reduce 296) ((T_IS_GREATER_OR_EQUAL) reduce 296) ((T_IS_IDENTICAL) reduce 296) ((T_IS_NOT_EQUAL) reduce 296) ((T_IS_NOT_IDENTICAL) reduce 296) ((T_IS_SMALLER_OR_EQUAL) reduce 296) ((T_LOGICAL_AND) reduce 296) ((T_LOGICAL_OR) reduce 296) ((T_LOGICAL_XOR) reduce 296) ((T_PIPE) reduce 296) ((T_POW) reduce 296) ((T_SL) reduce 296) ((T_SPACESHIP) reduce 296) ((T_SR) reduce 296) (("^") reduce 296) (("|") reduce 296)) 5830 ((("%") reduce 297) (("*") reduce 297) (("+") reduce 297) ((",") reduce 297) (("-") reduce 297) ((".") reduce 297) (("/") reduce 297) ((";") reduce 297) (("<") reduce 297) ((">") reduce 297) (("?") reduce 297) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_BOOLEAN_AND) reduce 297) ((T_BOOLEAN_OR) reduce 297) ((T_COALESCE) reduce 297) ((T_INSTANCEOF) reduce 297) ((T_IS_EQUAL) reduce 297) ((T_IS_GREATER_OR_EQUAL) reduce 297) ((T_IS_IDENTICAL) reduce 297) ((T_IS_NOT_EQUAL) reduce 297) ((T_IS_NOT_IDENTICAL) reduce 297) ((T_IS_SMALLER_OR_EQUAL) reduce 297) ((T_LOGICAL_AND) reduce 297) ((T_LOGICAL_OR) reduce 297) ((T_LOGICAL_XOR) reduce 297) ((T_PIPE) reduce 297) ((T_POW) reduce 297) ((T_SL) reduce 297) ((T_SPACESHIP) reduce 297) ((T_SR) reduce 297) (("^") reduce 297) (("|") reduce 297)) 5831 ((("%") reduce 460) (("(") reduce 556) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) ((";") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PIPE) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("[") reduce 549) (("^") reduce 460) (("|") reduce 460)) 5833 ((("%") reduce 298) (("*") reduce 298) (("+") reduce 298) ((",") reduce 298) (("-") reduce 298) ((".") reduce 298) (("/") reduce 298) ((";") reduce 298) (("<") reduce 298) ((">") reduce 298) (("?") reduce 298) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_BOOLEAN_AND) reduce 298) ((T_BOOLEAN_OR) reduce 298) ((T_COALESCE) reduce 298) ((T_INSTANCEOF) reduce 298) ((T_IS_EQUAL) reduce 298) ((T_IS_GREATER_OR_EQUAL) reduce 298) ((T_IS_IDENTICAL) reduce 298) ((T_IS_NOT_EQUAL) reduce 298) ((T_IS_NOT_IDENTICAL) reduce 298) ((T_IS_SMALLER_OR_EQUAL) reduce 298) ((T_LOGICAL_AND) reduce 298) ((T_LOGICAL_OR) reduce 298) ((T_LOGICAL_XOR) reduce 298) ((T_PIPE) reduce 298) ((T_POW) reduce 298) ((T_SL) reduce 298) ((T_SPACESHIP) reduce 298) ((T_SR) reduce 298) (("^") reduce 298) (("|") reduce 298)) 5834 ((("%") reduce 299) (("*") reduce 299) (("+") reduce 299) ((",") reduce 299) (("-") reduce 299) ((".") reduce 299) (("/") reduce 299) ((";") reduce 299) (("<") reduce 299) ((">") reduce 299) (("?") reduce 299) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_BOOLEAN_AND) reduce 299) ((T_BOOLEAN_OR) reduce 299) ((T_COALESCE) reduce 299) ((T_INSTANCEOF) reduce 299) ((T_IS_EQUAL) reduce 299) ((T_IS_GREATER_OR_EQUAL) reduce 299) ((T_IS_IDENTICAL) reduce 299) ((T_IS_NOT_EQUAL) reduce 299) ((T_IS_NOT_IDENTICAL) reduce 299) ((T_IS_SMALLER_OR_EQUAL) reduce 299) ((T_LOGICAL_AND) reduce 299) ((T_LOGICAL_OR) reduce 299) ((T_LOGICAL_XOR) reduce 299) ((T_PIPE) reduce 299) ((T_POW) reduce 299) ((T_SL) reduce 299) ((T_SPACESHIP) reduce 299) ((T_SR) reduce 299) (("^") reduce 299) (("|") reduce 299)) 5835 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) ((",") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) ((";") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_PIPE) reduce 472) ((T_POW) shift) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("|") reduce 472)) 5836 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) ((",") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) ((";") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_PIPE) reduce 470) ((T_POW) shift) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("|") reduce 470)) 5839 ((("%") reduce 519) (("(") reduce 519) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) ((";") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PIPE) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("|") reduce 519)) 5840 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) ((",") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) ((";") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_PIPE) reduce 475) ((T_POW) shift) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("|") reduce 475)) 5841 ((("%") reduce 446) (("*") reduce 446) (("+") reduce 446) ((",") reduce 446) (("-") reduce 446) ((".") reduce 446) (("/") reduce 446) ((";") reduce 446) (("<") reduce 446) ((">") reduce 446) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_BOOLEAN_AND) reduce 446) ((T_BOOLEAN_OR) reduce 446) ((T_COALESCE) reduce 446) ((T_INSTANCEOF) reduce 446) ((T_IS_EQUAL) reduce 446) ((T_IS_GREATER_OR_EQUAL) reduce 446) ((T_IS_IDENTICAL) reduce 446) ((T_IS_NOT_EQUAL) reduce 446) ((T_IS_NOT_IDENTICAL) reduce 446) ((T_IS_SMALLER_OR_EQUAL) reduce 446) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_PIPE) reduce 446) ((T_POW) shift) ((T_SL) reduce 446) ((T_SPACESHIP) reduce 446) ((T_SR) reduce 446) (("^") reduce 446) (("|") reduce 446)) 5842 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) ((",") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) ((";") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_PIPE) reduce 445) ((T_POW) shift) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("|") reduce 445)) 5844 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) ((";") reduce 570) (("<") reduce 570) (("=") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AND_EQUAL) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_COALESCE_EQUAL) reduce 570) ((T_CONCAT_EQUAL) reduce 570) ((T_DEC) reduce 570) ((T_DIV_EQUAL) reduce 570) ((T_INC) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_MINUS_EQUAL) reduce 570) ((T_MOD_EQUAL) reduce 570) ((T_MUL_EQUAL) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_OR_EQUAL) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_PLUS_EQUAL) reduce 570) ((T_POW) reduce 570) ((T_POW_EQUAL) reduce 570) ((T_SL) reduce 570) ((T_SL_EQUAL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) ((T_SR_EQUAL) reduce 570) ((T_XOR_EQUAL) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 5847 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) ((";") reduce 569) (("<") reduce 569) (("=") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AND_EQUAL) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_COALESCE_EQUAL) reduce 569) ((T_CONCAT_EQUAL) reduce 569) ((T_DEC) reduce 569) ((T_DIV_EQUAL) reduce 569) ((T_INC) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_MINUS_EQUAL) reduce 569) ((T_MOD_EQUAL) reduce 569) ((T_MUL_EQUAL) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_OR_EQUAL) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_PLUS_EQUAL) reduce 569) ((T_POW) reduce 569) ((T_POW_EQUAL) reduce 569) ((T_SL) reduce 569) ((T_SL_EQUAL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) ((T_SR_EQUAL) reduce 569) ((T_XOR_EQUAL) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 5849 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) ((";") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_PIPE) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("|") reduce 522)) 5850 ((("%") reduce 447) (("*") reduce 447) (("+") reduce 447) ((",") reduce 447) (("-") reduce 447) ((".") reduce 447) (("/") reduce 447) ((";") reduce 447) (("<") reduce 447) ((">") reduce 447) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_BOOLEAN_AND) reduce 447) ((T_BOOLEAN_OR) reduce 447) ((T_COALESCE) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 447) ((T_IS_GREATER_OR_EQUAL) reduce 447) ((T_IS_IDENTICAL) reduce 447) ((T_IS_NOT_EQUAL) reduce 447) ((T_IS_NOT_IDENTICAL) reduce 447) ((T_IS_SMALLER_OR_EQUAL) reduce 447) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_PIPE) reduce 447) ((T_POW) shift) ((T_SL) reduce 447) ((T_SPACESHIP) reduce 447) ((T_SR) reduce 447) (("^") reduce 447) (("|") reduce 447)) 5852 (((",") reduce 364) ((";") reduce 364) (("{") shift)) 5853 (((T_ABSTRACT) reduce 368) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 368) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 368) ((T_ATTRIBUTE) reduce 368) ((T_FINAL) reduce 368) ((T_PRIVATE) reduce 368) ((T_PRIVATE_SET) reduce 368) ((T_PROTECTED) reduce 368) ((T_PROTECTED_SET) reduce 368) ((T_PUBLIC) reduce 368) ((T_PUBLIC_SET) reduce 368) ((T_READONLY) reduce 368) ((T_STATIC) reduce 368) ((T_STRING) reduce 368) (("}") reduce 368)) 5854 (((T_ABSTRACT) shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 373) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 373) ((T_ATTRIBUTE) shift) ((T_FINAL) shift) ((T_PRIVATE) shift) ((T_PRIVATE_SET) shift) ((T_PROTECTED) shift) ((T_PROTECTED_SET) shift) ((T_PUBLIC) shift) ((T_PUBLIC_SET) shift) ((T_READONLY) shift) ((T_STATIC) shift) ((T_STRING) reduce 373) (("}") shift)) 5855 (((T_ABSTRACT) reduce 359) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 359) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 359) ((T_FINAL) reduce 359) ((T_PRIVATE) reduce 359) ((T_PRIVATE_SET) reduce 359) ((T_PROTECTED) reduce 359) ((T_PROTECTED_SET) reduce 359) ((T_PUBLIC) reduce 359) ((T_PUBLIC_SET) reduce 359) ((T_READONLY) reduce 359) ((T_STATIC) reduce 359) ((T_STRING) reduce 359)) 5857 (((T_ABSTRACT) reduce 360) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 360) ((T_FINAL) reduce 360) ((T_PRIVATE) reduce 360) ((T_PRIVATE_SET) reduce 360) ((T_PROTECTED) reduce 360) ((T_PROTECTED_SET) reduce 360) ((T_PUBLIC) reduce 360) ((T_PUBLIC_SET) reduce 360) ((T_READONLY) reduce 360) ((T_STATIC) reduce 360) ((T_STRING) reduce 360)) 5858 (((T_ABSTRACT) reduce 354) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 354) ((T_FINAL) reduce 354) ((T_PRIVATE) reduce 354) ((T_PRIVATE_SET) reduce 354) ((T_PROTECTED) reduce 354) ((T_PROTECTED_SET) reduce 354) ((T_PUBLIC) reduce 354) ((T_PUBLIC_SET) reduce 354) ((T_READONLY) reduce 354) ((T_STATIC) reduce 354) ((T_STRING) reduce 354)) 5859 (((T_ABSTRACT) reduce 357) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 357) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 357) ((T_FINAL) reduce 357) ((T_PRIVATE) reduce 357) ((T_PRIVATE_SET) reduce 357) ((T_PROTECTED) reduce 357) ((T_PROTECTED_SET) reduce 357) ((T_PUBLIC) reduce 357) ((T_PUBLIC_SET) reduce 357) ((T_READONLY) reduce 357) ((T_STATIC) reduce 357) ((T_STRING) reduce 357)) 5860 (((T_ABSTRACT) reduce 353) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 353) ((T_FINAL) reduce 353) ((T_PRIVATE) reduce 353) ((T_PRIVATE_SET) reduce 353) ((T_PROTECTED) reduce 353) ((T_PROTECTED_SET) reduce 353) ((T_PUBLIC) reduce 353) ((T_PUBLIC_SET) reduce 353) ((T_READONLY) reduce 353) ((T_STATIC) reduce 353) ((T_STRING) reduce 353)) 5861 (((T_ABSTRACT) reduce 356) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 356) ((T_FINAL) reduce 356) ((T_PRIVATE) reduce 356) ((T_PRIVATE_SET) reduce 356) ((T_PROTECTED) reduce 356) ((T_PROTECTED_SET) reduce 356) ((T_PUBLIC) reduce 356) ((T_PUBLIC_SET) reduce 356) ((T_READONLY) reduce 356) ((T_STATIC) reduce 356) ((T_STRING) reduce 356)) 5862 (((T_ABSTRACT) reduce 352) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 352) ((T_FINAL) reduce 352) ((T_PRIVATE) reduce 352) ((T_PRIVATE_SET) reduce 352) ((T_PROTECTED) reduce 352) ((T_PROTECTED_SET) reduce 352) ((T_PUBLIC) reduce 352) ((T_PUBLIC_SET) reduce 352) ((T_READONLY) reduce 352) ((T_STATIC) reduce 352) ((T_STRING) reduce 352)) 5863 (((T_ABSTRACT) reduce 355) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 355) ((T_FINAL) reduce 355) ((T_PRIVATE) reduce 355) ((T_PRIVATE_SET) reduce 355) ((T_PROTECTED) reduce 355) ((T_PROTECTED_SET) reduce 355) ((T_PUBLIC) reduce 355) ((T_PUBLIC_SET) reduce 355) ((T_READONLY) reduce 355) ((T_STATIC) reduce 355) ((T_STRING) reduce 355)) 5864 (((T_ABSTRACT) reduce 361) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 361) ((T_FINAL) reduce 361) ((T_PRIVATE) reduce 361) ((T_PRIVATE_SET) reduce 361) ((T_PROTECTED) reduce 361) ((T_PROTECTED_SET) reduce 361) ((T_PUBLIC) reduce 361) ((T_PUBLIC_SET) reduce 361) ((T_READONLY) reduce 361) ((T_STATIC) reduce 361) ((T_STRING) reduce 361)) 5865 (((T_ABSTRACT) reduce 358) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 358) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 358) ((T_FINAL) reduce 358) ((T_PRIVATE) reduce 358) ((T_PRIVATE_SET) reduce 358) ((T_PROTECTED) reduce 358) ((T_PROTECTED_SET) reduce 358) ((T_PUBLIC) reduce 358) ((T_PUBLIC_SET) reduce 358) ((T_READONLY) reduce 358) ((T_STATIC) reduce 358) ((T_STRING) reduce 358)) 5866 (((T_ABSTRACT) reduce 101) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 101) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 101) ((T_ATTRIBUTE) reduce 101) ((T_FINAL) reduce 101) ((T_PRIVATE) reduce 101) ((T_PRIVATE_SET) reduce 101) ((T_PROTECTED) reduce 101) ((T_PROTECTED_SET) reduce 101) ((T_PUBLIC) reduce 101) ((T_PUBLIC_SET) reduce 101) ((T_READONLY) reduce 101) ((T_STATIC) reduce 101) ((T_STRING) reduce 101)) 5867 (((T_ABSTRACT) shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 373) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 373) ((T_ATTRIBUTE) shift) ((T_FINAL) shift) ((T_PRIVATE) shift) ((T_PRIVATE_SET) shift) ((T_PROTECTED) shift) ((T_PROTECTED_SET) shift) ((T_PUBLIC) shift) ((T_PUBLIC_SET) shift) ((T_READONLY) shift) ((T_STATIC) shift) ((T_STRING) reduce 373)) 5868 (((T_ABSTRACT) reduce 350) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 350) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 350) ((T_FINAL) reduce 350) ((T_PRIVATE) reduce 350) ((T_PRIVATE_SET) reduce 350) ((T_PROTECTED) reduce 350) ((T_PROTECTED_SET) reduce 350) ((T_PUBLIC) reduce 350) ((T_PUBLIC_SET) reduce 350) ((T_READONLY) reduce 350) ((T_STATIC) reduce 350) ((T_STRING) reduce 350)) 5869 (((T_ABSTRACT) shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 374) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 374) ((T_FINAL) shift) ((T_PRIVATE) shift) ((T_PRIVATE_SET) shift) ((T_PROTECTED) shift) ((T_PROTECTED_SET) shift) ((T_PUBLIC) shift) ((T_PUBLIC_SET) shift) ((T_READONLY) shift) ((T_STATIC) shift) ((T_STRING) reduce 374)) 5870 (((T_ABSTRACT) reduce 369) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 369) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 369) ((T_ATTRIBUTE) reduce 369) ((T_FINAL) reduce 369) ((T_PRIVATE) reduce 369) ((T_PRIVATE_SET) reduce 369) ((T_PROTECTED) reduce 369) ((T_PROTECTED_SET) reduce 369) ((T_PUBLIC) reduce 369) ((T_PUBLIC_SET) reduce 369) ((T_READONLY) reduce 369) ((T_STATIC) reduce 369) ((T_STRING) reduce 369) (("}") reduce 369)) 5871 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_STRING) reduce 496)) 5872 (((T_ABSTRACT) reduce 366) ((T_ATTRIBUTE) reduce 366) ((T_CASE) reduce 366) ((T_CONST) reduce 366) ((T_FINAL) reduce 366) ((T_FUNCTION) reduce 366) ((T_PRIVATE) reduce 366) ((T_PRIVATE_SET) reduce 366) ((T_PROTECTED) reduce 366) ((T_PROTECTED_SET) reduce 366) ((T_PUBLIC) reduce 366) ((T_PUBLIC_SET) reduce 366) ((T_READONLY) reduce 366) ((T_STATIC) reduce 366) ((T_USE) reduce 366) ((T_VAR) reduce 366) (("}") reduce 366)) 5873 (((T_STRING) reduce 80)) 5874 (((T_STRING) reduce 81)) 5875 (((T_STRING) reduce 497)) 5877 ((("(") reduce 493) ((";") reduce 493) ((T_DOUBLE_ARROW) reduce 493) (("{") reduce 493)) 5878 ((("(") shift) ((";") reduce 379) ((T_DOUBLE_ARROW) reduce 379) (("{") reduce 379)) 5880 (((";") reduce 494) ((T_DOUBLE_ARROW) reduce 494) (("{") reduce 494)) 5881 (((";") shift) ((T_DOUBLE_ARROW) shift) (("{") shift)) 5882 (((T_ABSTRACT) reduce 376) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 376) ((T_ATTRIBUTE) reduce 376) ((T_FINAL) reduce 376) ((T_PRIVATE) reduce 376) ((T_PRIVATE_SET) reduce 376) ((T_PROTECTED) reduce 376) ((T_PROTECTED_SET) reduce 376) ((T_PUBLIC) reduce 376) ((T_PUBLIC_SET) reduce 376) ((T_READONLY) reduce 376) ((T_STATIC) reduce 376) ((T_STRING) reduce 376) (("}") reduce 376)) 5884 (((T_ABSTRACT) reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_ATTRIBUTE) reduce 494) ((T_FINAL) reduce 494) ((T_PRIVATE) reduce 494) ((T_PRIVATE_SET) reduce 494) ((T_PROTECTED) reduce 494) ((T_PROTECTED_SET) reduce 494) ((T_PUBLIC) reduce 494) ((T_PUBLIC_SET) reduce 494) ((T_READONLY) reduce 494) ((T_STATIC) reduce 494) ((T_STRING) reduce 494) (("}") reduce 494)) 5887 (((T_ABSTRACT) reduce 377) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 377) ((T_ATTRIBUTE) reduce 377) ((T_FINAL) reduce 377) ((T_PRIVATE) reduce 377) ((T_PRIVATE_SET) reduce 377) ((T_PROTECTED) reduce 377) ((T_PROTECTED_SET) reduce 377) ((T_PUBLIC) reduce 377) ((T_PUBLIC_SET) reduce 377) ((T_READONLY) reduce 377) ((T_STATIC) reduce 377) ((T_STRING) reduce 377) (("}") reduce 377)) 5888 (((T_ABSTRACT) reduce 375) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 375) ((T_ATTRIBUTE) reduce 375) ((T_FINAL) reduce 375) ((T_PRIVATE) reduce 375) ((T_PRIVATE_SET) reduce 375) ((T_PROTECTED) reduce 375) ((T_PROTECTED_SET) reduce 375) ((T_PUBLIC) reduce 375) ((T_PUBLIC_SET) reduce 375) ((T_READONLY) reduce 375) ((T_STATIC) reduce 375) ((T_STRING) reduce 375) (("}") reduce 375)) 5896 (((T_ABSTRACT) reduce 378) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 378) ((T_ATTRIBUTE) reduce 378) ((T_FINAL) reduce 378) ((T_PRIVATE) reduce 378) ((T_PRIVATE_SET) reduce 378) ((T_PROTECTED) reduce 378) ((T_PROTECTED_SET) reduce 378) ((T_PUBLIC) reduce 378) ((T_PUBLIC_SET) reduce 378) ((T_READONLY) reduce 378) ((T_STATIC) reduce 378) ((T_STRING) reduce 378) (("}") reduce 378)) 5922 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 432) (("<") shift) ((">") shift) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 432)) 5923 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 435) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("|") reduce 435)) 5924 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 444) (("/") shift) ((";") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_PIPE) reduce 444) ((T_POW) shift) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("|") reduce 444)) 5925 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 458) (("<") shift) ((">") shift) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 458) (("|") reduce 458)) 5926 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 443) (("/") shift) ((";") reduce 443) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_PIPE) reduce 443) ((T_POW) shift) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("^") reduce 443) (("|") reduce 443)) 5927 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) ((";") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_PIPE) reduce 440) ((T_POW) shift) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("|") reduce 440)) 5928 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 453) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_PIPE) reduce 453) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 453) ((T_SR) shift) (("^") reduce 453) (("|") reduce 453)) 5929 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 431) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5930 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 429) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5931 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 430) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5932 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 455) (("<") shift) ((">") shift) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 455) ((T_SR) shift) (("^") reduce 455) (("|") reduce 455)) 5933 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 450) (("<") shift) ((">") shift) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 450) (("|") reduce 450)) 5934 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 452) (("<") shift) ((">") shift) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 452) (("|") reduce 452)) 5935 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 449) (("<") shift) ((">") shift) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 449) (("|") reduce 449)) 5936 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 457) (("<") shift) ((">") shift) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 457) ((T_SR) shift) (("^") reduce 457) (("|") reduce 457)) 5937 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 451) (("<") shift) ((">") shift) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 451) (("|") reduce 451)) 5940 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 5941 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 5942 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 5943 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((";") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 5944 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 5945 ((("%") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) ((";") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 5946 ((("%") reduce 511) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((";") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("|") reduce 511)) 5947 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) ((";") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PIPE) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 5948 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((";") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 5949 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((";") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("|") reduce 512)) 5950 ((("%") reduce 573) (("*") reduce 573) (("+") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) ((";") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("|") reduce 573)) 5956 ((("%") reduce 574) (("*") reduce 574) (("+") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) ((";") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("|") reduce 574)) 5957 ((("%") reduce 578) (("*") reduce 578) (("+") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) ((";") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("|") reduce 578)) 5958 ((("%") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) ((";") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 5959 ((("%") reduce 575) (("*") reduce 575) (("+") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) ((";") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("|") reduce 575)) 5960 ((("%") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) ((";") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 5963 ((("%") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) ((";") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 5964 ((("%") reduce 576) (("*") reduce 576) (("+") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) ((";") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("|") reduce 576)) 5966 ((("%") reduce 577) (("*") reduce 577) (("+") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) ((";") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("|") reduce 577)) 5968 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((";") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("|") reduce 513)) 5969 ((("%") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) ((";") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 5972 ((("%") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) ((";") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 5973 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 465) (("<") shift) ((">") shift) (("?") reduce 465) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5974 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 427) (("<") shift) ((">") shift) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5975 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 428) (("<") shift) ((">") shift) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5976 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 433) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("|") reduce 433)) 5977 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 434) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("|") reduce 434)) 5981 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 463) (("<") shift) ((">") shift) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5982 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 464) (("<") shift) ((">") shift) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 5983 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 456) (("<") shift) ((">") shift) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 456) ((T_SR) shift) (("^") reduce 456) (("|") reduce 456)) 5984 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 454) (("<") shift) ((">") shift) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 454) ((T_SR) shift) (("^") reduce 454) (("|") reduce 454)) 5985 ((("%") reduce 441) (("*") reduce 441) (("+") reduce 441) (("-") reduce 441) ((".") reduce 441) (("/") reduce 441) ((";") reduce 441) (("<") reduce 441) ((">") reduce 441) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 441) ((T_IS_GREATER_OR_EQUAL) reduce 441) ((T_IS_IDENTICAL) reduce 441) ((T_IS_NOT_EQUAL) reduce 441) ((T_IS_NOT_IDENTICAL) reduce 441) ((T_IS_SMALLER_OR_EQUAL) reduce 441) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_PIPE) reduce 441) ((T_POW) shift) ((T_SL) reduce 441) ((T_SPACESHIP) reduce 441) ((T_SR) reduce 441) (("^") reduce 441) (("|") reduce 441)) 5986 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 436) (("/") shift) ((";") reduce 436) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_PIPE) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 436) ((T_SR) shift) (("^") reduce 436) (("|") reduce 436)) 5987 ((("%") shift) (("*") shift) (("+") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") shift) ((";") reduce 438) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_PIPE) reduce 438) ((T_POW) shift) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("|") reduce 438)) 5988 ((("%") shift) (("*") shift) (("+") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") shift) ((";") reduce 437) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_PIPE) reduce 437) ((T_POW) shift) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("|") reduce 437)) 5989 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) ((";") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_PIPE) reduce 439) ((T_POW) shift) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("|") reduce 439)) 5990 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) ((";") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_PIPE) reduce 442) ((T_POW) shift) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("|") reduce 442)) 5992 (((";") reduce 380) ((T_DOUBLE_ARROW) reduce 380) (("{") reduce 380)) 5993 (((T_ABSTRACT) reduce 351) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 351) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 351) ((T_FINAL) reduce 351) ((T_PRIVATE) reduce 351) ((T_PRIVATE_SET) reduce 351) ((T_PROTECTED) reduce 351) ((T_PROTECTED_SET) reduce 351) ((T_PUBLIC) reduce 351) ((T_PUBLIC_SET) reduce 351) ((T_READONLY) reduce 351) ((T_STATIC) reduce 351) ((T_STRING) reduce 351)) 5994 (((T_ABSTRACT) reduce 102) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 102) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 102) ((T_ATTRIBUTE) reduce 102) ((T_FINAL) reduce 102) ((T_PRIVATE) reduce 102) ((T_PRIVATE_SET) reduce 102) ((T_PROTECTED) reduce 102) ((T_PROTECTED_SET) reduce 102) ((T_PUBLIC) reduce 102) ((T_PUBLIC_SET) reduce 102) ((T_READONLY) reduce 102) ((T_STATIC) reduce 102) ((T_STRING) reduce 102)) 5995 (((T_ABSTRACT) reduce 370) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 370) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 370) ((T_ATTRIBUTE) reduce 370) ((T_FINAL) reduce 370) ((T_PRIVATE) reduce 370) ((T_PRIVATE_SET) reduce 370) ((T_PROTECTED) reduce 370) ((T_PROTECTED_SET) reduce 370) ((T_PUBLIC) reduce 370) ((T_PUBLIC_SET) reduce 370) ((T_READONLY) reduce 370) ((T_STATIC) reduce 370) ((T_STRING) reduce 370) (("}") reduce 370)) 5996 ((("(") reduce 94) ((",") reduce 94) (("]") reduce 94)) 5997 ((("(") reduce 93) ((",") reduce 93) (("]") reduce 93)) 5998 ((("(") reduce 95) ((",") reduce 95) (("]") reduce 95)) 5999 ((("(") reduce 509) ((",") reduce 509) (("]") reduce 509)) 6000 ((("(") reduce 92) ((",") reduce 92) (("]") reduce 92)) 6001 (((",") reduce 98) (("]") reduce 98)) 6002 (((",") shift) (("]") reduce 125)) 6003 ((("(") shift) ((",") reduce 96) (("]") reduce 96)) 6004 ((("(") reduce 510) ((",") reduce 510) (("]") reduce 510)) 6006 (((",") reduce 97) (("]") reduce 97)) 6007 (((",") reduce 291) (("]") reduce 291)) 6011 (((",") reduce 292) (("]") reduce 292)) 6012 (((",") reduce 293) (("]") reduce 293)) 6013 (((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_STATIC) shift) ((T_STRING) shift) (("]") reduce 126)) 6015 (((T_ABSTRACT) reduce 100) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 100) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 100) ((T_ATTRIBUTE) reduce 100) ((T_FINAL) reduce 100) ((T_PRIVATE) reduce 100) ((T_PRIVATE_SET) reduce 100) ((T_PROTECTED) reduce 100) ((T_PROTECTED_SET) reduce 100) ((T_PUBLIC) reduce 100) ((T_PUBLIC_SET) reduce 100) ((T_READONLY) reduce 100) ((T_STATIC) reduce 100) ((T_STRING) reduce 100)) 6016 (((",") reduce 99) (("]") reduce 99)) 6027 ((("%") reduce 540) (("*") reduce 540) (("+") reduce 540) ((",") reduce 540) (("-") reduce 540) ((".") reduce 540) (("/") reduce 540) ((";") reduce 540) (("<") reduce 540) ((">") reduce 540) (("?") reduce 540) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_BOOLEAN_AND) reduce 540) ((T_BOOLEAN_OR) reduce 540) ((T_COALESCE) reduce 540) ((T_INSTANCEOF) reduce 540) ((T_IS_EQUAL) reduce 540) ((T_IS_GREATER_OR_EQUAL) reduce 540) ((T_IS_IDENTICAL) reduce 540) ((T_IS_NOT_EQUAL) reduce 540) ((T_IS_NOT_IDENTICAL) reduce 540) ((T_IS_SMALLER_OR_EQUAL) reduce 540) ((T_LOGICAL_AND) reduce 540) ((T_LOGICAL_OR) reduce 540) ((T_LOGICAL_XOR) reduce 540) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 540) ((T_OBJECT_OPERATOR) reduce 540) ((T_PIPE) reduce 540) ((T_POW) reduce 540) ((T_SL) reduce 540) ((T_SPACESHIP) reduce 540) ((T_SR) reduce 540) (("[") reduce 540) (("^") reduce 540) (("{") reduce 540) (("|") reduce 540)) 6029 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) ((";") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_PIPE) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 6031 ((("%") reduce 534) (("*") reduce 534) (("+") reduce 534) ((",") reduce 534) (("-") reduce 534) ((".") reduce 534) (("/") reduce 534) ((";") reduce 534) (("<") reduce 534) ((">") reduce 534) (("?") reduce 534) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_BOOLEAN_AND) reduce 534) ((T_BOOLEAN_OR) reduce 534) ((T_COALESCE) reduce 534) ((T_INSTANCEOF) reduce 534) ((T_IS_EQUAL) reduce 534) ((T_IS_GREATER_OR_EQUAL) reduce 534) ((T_IS_IDENTICAL) reduce 534) ((T_IS_NOT_EQUAL) reduce 534) ((T_IS_NOT_IDENTICAL) reduce 534) ((T_IS_SMALLER_OR_EQUAL) reduce 534) ((T_LOGICAL_AND) reduce 534) ((T_LOGICAL_OR) reduce 534) ((T_LOGICAL_XOR) reduce 534) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PIPE) reduce 534) ((T_POW) reduce 534) ((T_SL) reduce 534) ((T_SPACESHIP) reduce 534) ((T_SR) reduce 534) (("[") reduce 534) (("^") reduce 534) (("{") reduce 534) (("|") reduce 534)) 6032 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) ((";") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_PIPE) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 6036 ((("%") reduce 517) (("(") shift) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((";") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_PIPE) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 6037 ((("%") reduce 533) (("*") reduce 533) (("+") reduce 533) ((",") reduce 533) (("-") reduce 533) ((".") reduce 533) (("/") reduce 533) ((";") reduce 533) (("<") reduce 533) ((">") reduce 533) (("?") reduce 533) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_BOOLEAN_AND) reduce 533) ((T_BOOLEAN_OR) reduce 533) ((T_COALESCE) reduce 533) ((T_INSTANCEOF) reduce 533) ((T_IS_EQUAL) reduce 533) ((T_IS_GREATER_OR_EQUAL) reduce 533) ((T_IS_IDENTICAL) reduce 533) ((T_IS_NOT_EQUAL) reduce 533) ((T_IS_NOT_IDENTICAL) reduce 533) ((T_IS_SMALLER_OR_EQUAL) reduce 533) ((T_LOGICAL_AND) reduce 533) ((T_LOGICAL_OR) reduce 533) ((T_LOGICAL_XOR) reduce 533) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PIPE) reduce 533) ((T_POW) reduce 533) ((T_SL) reduce 533) ((T_SPACESHIP) reduce 533) ((T_SR) reduce 533) (("[") reduce 533) (("^") reduce 533) (("{") reduce 533) (("|") reduce 533)) 6038 ((("%") reduce 537) (("*") reduce 537) (("+") reduce 537) ((",") reduce 537) (("-") reduce 537) ((".") reduce 537) (("/") reduce 537) ((";") reduce 537) (("<") reduce 537) ((">") reduce 537) (("?") reduce 537) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_BOOLEAN_AND) reduce 537) ((T_BOOLEAN_OR) reduce 537) ((T_COALESCE) reduce 537) ((T_INSTANCEOF) reduce 537) ((T_IS_EQUAL) reduce 537) ((T_IS_GREATER_OR_EQUAL) reduce 537) ((T_IS_IDENTICAL) reduce 537) ((T_IS_NOT_EQUAL) reduce 537) ((T_IS_NOT_IDENTICAL) reduce 537) ((T_IS_SMALLER_OR_EQUAL) reduce 537) ((T_LOGICAL_AND) reduce 537) ((T_LOGICAL_OR) reduce 537) ((T_LOGICAL_XOR) reduce 537) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_PIPE) reduce 537) ((T_POW) reduce 537) ((T_SL) reduce 537) ((T_SPACESHIP) reduce 537) ((T_SR) reduce 537) (("[") reduce 537) (("^") reduce 537) (("{") reduce 537) (("|") reduce 537)) 6044 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) ((";") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PIPE) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 6046 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) ((";") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_PIPE) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 6048 ((("%") reduce 536) (("*") reduce 536) (("+") reduce 536) ((",") reduce 536) (("-") reduce 536) ((".") reduce 536) (("/") reduce 536) ((";") reduce 536) (("<") reduce 536) ((">") reduce 536) (("?") reduce 536) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_BOOLEAN_AND) reduce 536) ((T_BOOLEAN_OR) reduce 536) ((T_COALESCE) reduce 536) ((T_INSTANCEOF) reduce 536) ((T_IS_EQUAL) reduce 536) ((T_IS_GREATER_OR_EQUAL) reduce 536) ((T_IS_IDENTICAL) reduce 536) ((T_IS_NOT_EQUAL) reduce 536) ((T_IS_NOT_IDENTICAL) reduce 536) ((T_IS_SMALLER_OR_EQUAL) reduce 536) ((T_LOGICAL_AND) reduce 536) ((T_LOGICAL_OR) reduce 536) ((T_LOGICAL_XOR) reduce 536) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 536) ((T_OBJECT_OPERATOR) reduce 536) ((T_PIPE) reduce 536) ((T_POW) reduce 536) ((T_SL) reduce 536) ((T_SPACESHIP) reduce 536) ((T_SR) reduce 536) (("[") reduce 536) (("^") reduce 536) (("{") reduce 536) (("|") reduce 536)) 6049 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 6050 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 6051 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("{") reduce 95) (("|") reduce 95)) 6053 ((("%") reduce 539) (("*") reduce 539) (("+") reduce 539) ((",") reduce 539) (("-") reduce 539) ((".") reduce 539) (("/") reduce 539) ((";") reduce 539) (("<") reduce 539) ((">") reduce 539) (("?") reduce 539) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_BOOLEAN_AND) reduce 539) ((T_BOOLEAN_OR) reduce 539) ((T_COALESCE) reduce 539) ((T_INSTANCEOF) reduce 539) ((T_IS_EQUAL) reduce 539) ((T_IS_GREATER_OR_EQUAL) reduce 539) ((T_IS_IDENTICAL) reduce 539) ((T_IS_NOT_EQUAL) reduce 539) ((T_IS_NOT_IDENTICAL) reduce 539) ((T_IS_SMALLER_OR_EQUAL) reduce 539) ((T_LOGICAL_AND) reduce 539) ((T_LOGICAL_OR) reduce 539) ((T_LOGICAL_XOR) reduce 539) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 539) ((T_OBJECT_OPERATOR) reduce 539) ((T_PIPE) reduce 539) ((T_POW) reduce 539) ((T_SL) reduce 539) ((T_SPACESHIP) reduce 539) ((T_SR) reduce 539) (("[") reduce 539) (("^") reduce 539) (("{") reduce 539) (("|") reduce 539)) 6056 ((("%") reduce 538) (("*") reduce 538) (("+") reduce 538) ((",") reduce 538) (("-") reduce 538) ((".") reduce 538) (("/") reduce 538) ((";") reduce 538) (("<") reduce 538) ((">") reduce 538) (("?") reduce 538) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_BOOLEAN_AND) reduce 538) ((T_BOOLEAN_OR) reduce 538) ((T_COALESCE) reduce 538) ((T_INSTANCEOF) reduce 538) ((T_IS_EQUAL) reduce 538) ((T_IS_GREATER_OR_EQUAL) reduce 538) ((T_IS_IDENTICAL) reduce 538) ((T_IS_NOT_EQUAL) reduce 538) ((T_IS_NOT_IDENTICAL) reduce 538) ((T_IS_SMALLER_OR_EQUAL) reduce 538) ((T_LOGICAL_AND) reduce 538) ((T_LOGICAL_OR) reduce 538) ((T_LOGICAL_XOR) reduce 538) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 538) ((T_OBJECT_OPERATOR) reduce 538) ((T_PIPE) reduce 538) ((T_POW) reduce 538) ((T_SL) reduce 538) ((T_SPACESHIP) reduce 538) ((T_SR) reduce 538) (("[") reduce 538) (("^") reduce 538) (("{") reduce 538) (("|") reduce 538)) 6062 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 6065 ((("%") reduce 535) (("*") reduce 535) (("+") reduce 535) ((",") reduce 535) (("-") reduce 535) ((".") reduce 535) (("/") reduce 535) ((";") reduce 535) (("<") reduce 535) ((">") reduce 535) (("?") reduce 535) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_BOOLEAN_AND) reduce 535) ((T_BOOLEAN_OR) reduce 535) ((T_COALESCE) reduce 535) ((T_INSTANCEOF) reduce 535) ((T_IS_EQUAL) reduce 535) ((T_IS_GREATER_OR_EQUAL) reduce 535) ((T_IS_IDENTICAL) reduce 535) ((T_IS_NOT_EQUAL) reduce 535) ((T_IS_NOT_IDENTICAL) reduce 535) ((T_IS_SMALLER_OR_EQUAL) reduce 535) ((T_LOGICAL_AND) reduce 535) ((T_LOGICAL_OR) reduce 535) ((T_LOGICAL_XOR) reduce 535) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PIPE) reduce 535) ((T_POW) reduce 535) ((T_SL) reduce 535) ((T_SPACESHIP) reduce 535) ((T_SR) reduce 535) (("[") reduce 535) (("^") reduce 535) (("{") reduce 535) (("|") reduce 535)) 6067 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) ((";") reduce 568) (("<") reduce 568) (("=") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AND_EQUAL) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_COALESCE_EQUAL) reduce 568) ((T_CONCAT_EQUAL) reduce 568) ((T_DEC) reduce 568) ((T_DIV_EQUAL) reduce 568) ((T_INC) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_MINUS_EQUAL) reduce 568) ((T_MOD_EQUAL) reduce 568) ((T_MUL_EQUAL) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_OR_EQUAL) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_PLUS_EQUAL) reduce 568) ((T_POW) reduce 568) ((T_POW_EQUAL) reduce 568) ((T_SL) reduce 568) ((T_SL_EQUAL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) ((T_SR_EQUAL) reduce 568) ((T_XOR_EQUAL) reduce 568) (("[") reduce 568) (("^") reduce 568) (("{") reduce 568) (("|") reduce 568)) 6068 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 479) (("(") shift) (("*") reduce 479) (("+") shift) ((",") reduce 479) (("-") shift) ((".") reduce 479) (("/") reduce 479) ((";") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 479) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 479) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PIPE) reduce 479) ((T_POW) reduce 479) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 479) (("`") shift) (("{") reduce 479) (("|") reduce 479) (("~") shift)) 6075 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) ((",") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) ((";") reduce 564) (("<") reduce 564) (("=") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AND_EQUAL) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_COALESCE_EQUAL) reduce 564) ((T_CONCAT_EQUAL) reduce 564) ((T_DEC) reduce 564) ((T_DIV_EQUAL) reduce 564) ((T_INC) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_MINUS_EQUAL) reduce 564) ((T_MOD_EQUAL) reduce 564) ((T_MUL_EQUAL) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_OR_EQUAL) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_PLUS_EQUAL) reduce 564) ((T_POW) reduce 564) ((T_POW_EQUAL) reduce 564) ((T_SL) reduce 564) ((T_SL_EQUAL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) ((T_SR_EQUAL) reduce 564) ((T_XOR_EQUAL) reduce 564) (("[") reduce 564) (("^") reduce 564) (("{") reduce 564) (("|") reduce 564)) 6076 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) ((";") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PIPE) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 551) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 6078 ((("%") reduce 529) (("*") reduce 529) (("+") reduce 529) ((",") reduce 529) (("-") reduce 529) ((".") reduce 529) (("/") reduce 529) ((";") reduce 529) (("<") reduce 529) ((">") reduce 529) (("?") reduce 529) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_BOOLEAN_AND) reduce 529) ((T_BOOLEAN_OR) reduce 529) ((T_COALESCE) reduce 529) ((T_INSTANCEOF) reduce 529) ((T_IS_EQUAL) reduce 529) ((T_IS_GREATER_OR_EQUAL) reduce 529) ((T_IS_IDENTICAL) reduce 529) ((T_IS_NOT_EQUAL) reduce 529) ((T_IS_NOT_IDENTICAL) reduce 529) ((T_IS_SMALLER_OR_EQUAL) reduce 529) ((T_LOGICAL_AND) reduce 529) ((T_LOGICAL_OR) reduce 529) ((T_LOGICAL_XOR) reduce 529) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PIPE) reduce 529) ((T_POW) reduce 529) ((T_SL) reduce 529) ((T_SPACESHIP) reduce 529) ((T_SR) reduce 529) (("[") reduce 554) (("^") reduce 529) (("{") reduce 529) (("|") reduce 529)) 6079 ((("%") reduce 528) (("(") reduce 557) (("*") reduce 528) (("+") reduce 528) ((",") reduce 528) (("-") reduce 528) ((".") reduce 528) (("/") reduce 528) ((";") reduce 528) (("<") reduce 528) ((">") reduce 528) (("?") reduce 528) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_BOOLEAN_AND) reduce 528) ((T_BOOLEAN_OR) reduce 528) ((T_COALESCE) reduce 528) ((T_INSTANCEOF) reduce 528) ((T_IS_EQUAL) reduce 528) ((T_IS_GREATER_OR_EQUAL) reduce 528) ((T_IS_IDENTICAL) reduce 528) ((T_IS_NOT_EQUAL) reduce 528) ((T_IS_NOT_IDENTICAL) reduce 528) ((T_IS_SMALLER_OR_EQUAL) reduce 528) ((T_LOGICAL_AND) reduce 528) ((T_LOGICAL_OR) reduce 528) ((T_LOGICAL_XOR) reduce 528) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PIPE) reduce 528) ((T_POW) reduce 528) ((T_SL) reduce 528) ((T_SPACESHIP) reduce 528) ((T_SR) reduce 528) (("[") reduce 550) (("^") reduce 528) (("{") reduce 528) (("|") reduce 528)) 6080 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 493) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 493) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 493) (("|") shift)) 6083 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) ((",") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) ((";") reduce 563) (("<") reduce 563) (("=") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AND_EQUAL) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_COALESCE_EQUAL) reduce 563) ((T_CONCAT_EQUAL) reduce 563) ((T_DEC) reduce 563) ((T_DIV_EQUAL) reduce 563) ((T_INC) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_MINUS_EQUAL) reduce 563) ((T_MOD_EQUAL) reduce 563) ((T_MUL_EQUAL) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_OR_EQUAL) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_PLUS_EQUAL) reduce 563) ((T_POW) reduce 563) ((T_POW_EQUAL) reduce 563) ((T_SL) reduce 563) ((T_SL_EQUAL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) ((T_SR_EQUAL) reduce 563) ((T_XOR_EQUAL) reduce 563) (("[") reduce 563) (("^") reduce 563) (("{") reduce 563) (("|") reduce 563)) 6084 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) ((",") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) ((";") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_PIPE) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("^") reduce 484) (("{") reduce 484) (("|") reduce 484)) 6085 ((("%") reduce 466) (("*") reduce 466) (("+") reduce 466) ((",") reduce 466) (("-") reduce 466) ((".") reduce 466) (("/") reduce 466) ((";") reduce 466) (("<") reduce 466) ((">") reduce 466) (("?") reduce 466) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_BOOLEAN_AND) reduce 466) ((T_BOOLEAN_OR) reduce 466) ((T_COALESCE) reduce 466) ((T_INSTANCEOF) reduce 466) ((T_IS_EQUAL) reduce 466) ((T_IS_GREATER_OR_EQUAL) reduce 466) ((T_IS_IDENTICAL) reduce 466) ((T_IS_NOT_EQUAL) reduce 466) ((T_IS_NOT_IDENTICAL) reduce 466) ((T_IS_SMALLER_OR_EQUAL) reduce 466) ((T_LOGICAL_AND) reduce 466) ((T_LOGICAL_OR) reduce 466) ((T_LOGICAL_XOR) reduce 466) ((T_PIPE) reduce 466) ((T_POW) reduce 466) ((T_SL) reduce 466) ((T_SPACESHIP) reduce 466) ((T_SR) reduce 466) (("^") reduce 466) (("{") reduce 466) (("|") reduce 466)) 6086 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) ((";") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_PIPE) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("^") reduce 488) (("{") reduce 488) (("|") reduce 488)) 6087 ((("%") reduce 531) (("(") shift) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) ((";") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 6088 ((("%") reduce 461) (("(") reduce 558) (("*") reduce 461) (("+") reduce 461) ((",") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) ((";") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PIPE) reduce 461) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") reduce 552) (("^") reduce 461) (("{") reduce 461) (("|") reduce 461)) 6089 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) ((",") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) ((";") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_PIPE) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("{") reduce 462) (("|") reduce 462)) 6090 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) ((",") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) ((";") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_PIPE) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("{") reduce 476) (("|") reduce 476)) 6091 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) ((";") reduce 559) (("<") reduce 559) (("=") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AND_EQUAL) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_COALESCE_EQUAL) reduce 559) ((T_CONCAT_EQUAL) reduce 559) ((T_DEC) reduce 559) ((T_DIV_EQUAL) reduce 559) ((T_INC) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_MINUS_EQUAL) reduce 559) ((T_MOD_EQUAL) reduce 559) ((T_MUL_EQUAL) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_OR_EQUAL) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_PLUS_EQUAL) reduce 559) ((T_POW) reduce 559) ((T_POW_EQUAL) reduce 559) ((T_SL) reduce 559) ((T_SL_EQUAL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) ((T_SR_EQUAL) reduce 559) ((T_XOR_EQUAL) reduce 559) (("[") reduce 559) (("^") reduce 559) (("{") reduce 559) (("|") reduce 559)) 6092 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((";") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("{") reduce 565) (("|") reduce 565)) 6093 ((("%") reduce 403) (("*") reduce 403) (("+") reduce 403) ((",") reduce 403) (("-") reduce 403) ((".") reduce 403) (("/") reduce 403) ((";") reduce 403) (("<") reduce 403) (("=") shift) ((">") reduce 403) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_INC) shift) ((T_INSTANCEOF) reduce 403) ((T_IS_EQUAL) reduce 403) ((T_IS_GREATER_OR_EQUAL) reduce 403) ((T_IS_IDENTICAL) reduce 403) ((T_IS_NOT_EQUAL) reduce 403) ((T_IS_NOT_IDENTICAL) reduce 403) ((T_IS_SMALLER_OR_EQUAL) reduce 403) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 403) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 403) ((T_POW_EQUAL) shift) ((T_SL) reduce 403) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 403) ((T_SR) reduce 403) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 548) (("^") reduce 403) (("{") reduce 403) (("|") reduce 403)) 6096 ((("%") reduce 448) (("*") reduce 448) (("+") reduce 448) ((",") reduce 448) (("-") reduce 448) ((".") reduce 448) (("/") reduce 448) ((";") reduce 448) (("<") reduce 448) ((">") reduce 448) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_BOOLEAN_AND) reduce 448) ((T_BOOLEAN_OR) reduce 448) ((T_COALESCE) reduce 448) ((T_INSTANCEOF) reduce 448) ((T_IS_EQUAL) reduce 448) ((T_IS_GREATER_OR_EQUAL) reduce 448) ((T_IS_IDENTICAL) reduce 448) ((T_IS_NOT_EQUAL) reduce 448) ((T_IS_NOT_IDENTICAL) reduce 448) ((T_IS_SMALLER_OR_EQUAL) reduce 448) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_PIPE) reduce 448) ((T_POW) shift) ((T_SL) reduce 448) ((T_SPACESHIP) reduce 448) ((T_SR) reduce 448) (("^") reduce 448) (("{") reduce 448) (("|") reduce 448)) 6128 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 432) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 432) (("<") shift) ((">") shift) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 432) (("|") reduce 432)) 6129 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 435) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 435) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("{") reduce 435) (("|") reduce 435)) 6130 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 444) (("-") shift) ((".") reduce 444) (("/") shift) ((";") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_PIPE) reduce 444) ((T_POW) shift) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("{") reduce 444) (("|") reduce 444)) 6131 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 458) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 458) (("<") shift) ((">") shift) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 458) (("{") reduce 458) (("|") reduce 458)) 6132 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 443) (("-") shift) ((".") reduce 443) (("/") shift) ((";") reduce 443) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_PIPE) reduce 443) ((T_POW) shift) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("^") reduce 443) (("{") reduce 443) (("|") reduce 443)) 6133 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) ((",") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) ((";") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_PIPE) reduce 440) ((T_POW) shift) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("{") reduce 440) (("|") reduce 440)) 6134 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 453) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 453) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_PIPE) reduce 453) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 453) ((T_SR) shift) (("^") reduce 453) (("{") reduce 453) (("|") reduce 453)) 6135 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 431) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 431) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 431) (("|") shift)) 6136 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 429) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 429) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 429) (("|") shift)) 6137 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 430) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 430) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 430) (("|") shift)) 6138 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 455) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 455) (("<") shift) ((">") shift) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 455) ((T_SR) shift) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 6139 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 450) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 450) (("<") shift) ((">") shift) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 450) (("{") reduce 450) (("|") reduce 450)) 6140 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 452) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 452) (("<") shift) ((">") shift) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 452) (("{") reduce 452) (("|") reduce 452)) 6141 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 449) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 449) (("<") shift) ((">") shift) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 449) (("{") reduce 449) (("|") reduce 449)) 6142 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 457) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 457) (("<") shift) ((">") shift) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 457) ((T_SR) shift) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 6143 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 451) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 451) (("<") shift) ((">") shift) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 451) (("{") reduce 451) (("|") reduce 451)) 6146 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 6147 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 6148 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("{") reduce 95) (("|") reduce 95)) 6149 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((";") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 6150 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 6151 ((("%") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) ((";") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("{") reduce 568) (("|") reduce 568)) 6152 ((("%") reduce 511) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((";") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 6153 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) ((";") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PIPE) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("{") reduce 459) (("|") reduce 459)) 6154 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((";") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 6155 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((";") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 6156 ((("%") reduce 573) (("*") reduce 573) (("+") reduce 573) ((",") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) ((";") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("{") reduce 573) (("|") reduce 573)) 6162 ((("%") reduce 574) (("*") reduce 574) (("+") reduce 574) ((",") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) ((";") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("{") reduce 574) (("|") reduce 574)) 6163 ((("%") reduce 578) (("*") reduce 578) (("+") reduce 578) ((",") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) ((";") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("{") reduce 578) (("|") reduce 578)) 6164 ((("%") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) ((";") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("{") reduce 582) (("|") reduce 582)) 6165 ((("%") reduce 575) (("*") reduce 575) (("+") reduce 575) ((",") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) ((";") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("{") reduce 575) (("|") reduce 575)) 6166 ((("%") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) ((";") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("{") reduce 584) (("|") reduce 584)) 6169 ((("%") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) ((";") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("{") reduce 583) (("|") reduce 583)) 6170 ((("%") reduce 576) (("*") reduce 576) (("+") reduce 576) ((",") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) ((";") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("{") reduce 576) (("|") reduce 576)) 6172 ((("%") reduce 577) (("*") reduce 577) (("+") reduce 577) ((",") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) ((";") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("{") reduce 577) (("|") reduce 577)) 6174 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((";") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 6175 ((("%") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) ((";") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("{") reduce 570) (("|") reduce 570)) 6178 ((("%") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) ((";") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("{") reduce 569) (("|") reduce 569)) 6179 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 465) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 465) (("<") shift) ((">") shift) (("?") reduce 465) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 465) (("|") shift)) 6180 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 427) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 427) (("<") shift) ((">") shift) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 427) (("|") shift)) 6181 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 428) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 428) (("<") shift) ((">") shift) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 428) (("|") shift)) 6182 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 433) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 433) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("{") reduce 433) (("|") reduce 433)) 6183 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 434) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 434) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("{") reduce 434) (("|") reduce 434)) 6187 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 463) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 463) (("<") shift) ((">") shift) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 463) (("|") shift)) 6188 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 464) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 464) (("<") shift) ((">") shift) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 464) (("|") shift)) 6189 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 456) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 456) (("<") shift) ((">") shift) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 456) ((T_SR) shift) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 6190 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 454) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 454) (("<") shift) ((">") shift) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 454) ((T_SR) shift) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 6191 ((("%") reduce 441) (("*") reduce 441) (("+") reduce 441) ((",") reduce 441) (("-") reduce 441) ((".") reduce 441) (("/") reduce 441) ((";") reduce 441) (("<") reduce 441) ((">") reduce 441) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 441) ((T_IS_GREATER_OR_EQUAL) reduce 441) ((T_IS_IDENTICAL) reduce 441) ((T_IS_NOT_EQUAL) reduce 441) ((T_IS_NOT_IDENTICAL) reduce 441) ((T_IS_SMALLER_OR_EQUAL) reduce 441) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_PIPE) reduce 441) ((T_POW) shift) ((T_SL) reduce 441) ((T_SPACESHIP) reduce 441) ((T_SR) reduce 441) (("^") reduce 441) (("{") reduce 441) (("|") reduce 441)) 6192 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 436) (("-") shift) ((".") reduce 436) (("/") shift) ((";") reduce 436) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_PIPE) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 436) ((T_SR) shift) (("^") reduce 436) (("{") reduce 436) (("|") reduce 436)) 6193 ((("%") shift) (("*") shift) (("+") reduce 438) ((",") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") shift) ((";") reduce 438) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_PIPE) reduce 438) ((T_POW) shift) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("{") reduce 438) (("|") reduce 438)) 6194 ((("%") shift) (("*") shift) (("+") reduce 437) ((",") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") shift) ((";") reduce 437) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_PIPE) reduce 437) ((T_POW) shift) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("{") reduce 437) (("|") reduce 437)) 6195 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) ((",") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) ((";") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_PIPE) reduce 439) ((T_POW) shift) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("{") reduce 439) (("|") reduce 439)) 6196 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) ((",") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) ((";") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_PIPE) reduce 442) ((T_POW) shift) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("{") reduce 442) (("|") reduce 442)) 6198 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) ((",") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) ((";") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_PIPE) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("{") reduce 74) (("|") reduce 74)) 6199 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) ((",") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) ((";") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_PIPE) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("{") reduce 52) (("|") reduce 52)) 6200 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) ((",") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) ((";") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_PIPE) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("{") reduce 27) (("|") reduce 27)) 6201 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) ((",") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) ((";") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_PIPE) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("{") reduce 51) (("|") reduce 51)) 6202 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) ((",") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) ((";") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_PIPE) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("{") reduce 53) (("|") reduce 53)) 6203 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) ((",") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) ((";") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_PIPE) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("{") reduce 49) (("|") reduce 49)) 6204 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) ((",") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) ((";") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_PIPE) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("{") reduce 29) (("|") reduce 29)) 6205 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) ((",") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) ((";") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_PIPE) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("{") reduce 59) (("|") reduce 59)) 6206 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) ((",") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) ((";") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_PIPE) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("{") reduce 60) (("|") reduce 60)) 6207 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) ((",") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) ((";") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_PIPE) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("{") reduce 11) (("|") reduce 11)) 6208 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) ((",") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) ((";") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_PIPE) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("{") reduce 42) (("|") reduce 42)) 6209 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) ((",") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) ((";") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_PIPE) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("{") reduce 39) (("|") reduce 39)) 6210 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) ((",") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) ((";") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_PIPE) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("{") reduce 25) (("|") reduce 25)) 6211 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) ((",") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) ((";") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_PIPE) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("{") reduce 50) (("|") reduce 50)) 6212 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) ((",") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) ((";") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_PIPE) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("{") reduce 66) (("|") reduce 66)) 6213 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) ((",") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) ((";") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_PIPE) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("{") reduce 18) (("|") reduce 18)) 6214 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) ((",") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) ((";") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_PIPE) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("{") reduce 17) (("|") reduce 17)) 6215 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) ((",") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) ((";") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_PIPE) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("{") reduce 15) (("|") reduce 15)) 6216 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) ((",") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) ((";") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_PIPE) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("{") reduce 14) (("|") reduce 14)) 6217 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) ((",") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) ((";") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_PIPE) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("{") reduce 38) (("|") reduce 38)) 6218 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) ((",") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) ((";") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_PIPE) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("{") reduce 26) (("|") reduce 26)) 6219 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) ((",") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) ((";") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_PIPE) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("{") reduce 22) (("|") reduce 22)) 6220 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) ((",") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) ((";") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_PIPE) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("{") reduce 24) (("|") reduce 24)) 6221 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) ((",") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) ((";") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_PIPE) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("{") reduce 16) (("|") reduce 16)) 6222 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) ((",") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) ((";") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_PIPE) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("{") reduce 48) (("|") reduce 48)) 6223 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) ((",") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) ((";") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_PIPE) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("{") reduce 20) (("|") reduce 20)) 6224 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) ((",") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) ((";") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_PIPE) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("{") reduce 70) (("|") reduce 70)) 6225 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) ((",") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) ((";") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_PIPE) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("{") reduce 3) (("|") reduce 3)) 6226 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) ((",") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) ((";") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_PIPE) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("{") reduce 12) (("|") reduce 12)) 6227 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) ((",") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) ((";") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_PIPE) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("{") reduce 54) (("|") reduce 54)) 6228 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) ((",") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) ((";") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_PIPE) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("{") reduce 65) (("|") reduce 65)) 6229 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) ((",") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) ((";") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_PIPE) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("{") reduce 75) (("|") reduce 75)) 6230 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) ((",") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) ((";") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_PIPE) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("{") reduce 30) (("|") reduce 30)) 6231 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) ((",") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) ((";") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_PIPE) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("{") reduce 68) (("|") reduce 68)) 6232 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) ((",") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) ((";") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_PIPE) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("{") reduce 21) (("|") reduce 21)) 6233 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) ((",") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) ((";") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_PIPE) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("{") reduce 23) (("|") reduce 23)) 6234 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) ((",") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) ((";") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_PIPE) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("{") reduce 41) (("|") reduce 41)) 6235 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) ((",") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) ((";") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_PIPE) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("{") reduce 62) (("|") reduce 62)) 6236 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) ((",") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) ((";") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_PIPE) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("{") reduce 34) (("|") reduce 34)) 6237 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) ((",") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) ((";") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_PIPE) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("{") reduce 40) (("|") reduce 40)) 6238 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) ((",") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) ((";") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_PIPE) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("{") reduce 13) (("|") reduce 13)) 6239 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) ((",") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) ((";") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_PIPE) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("{") reduce 55) (("|") reduce 55)) 6240 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) ((",") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) ((";") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_PIPE) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("{") reduce 1) (("|") reduce 1)) 6241 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) ((",") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) ((";") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_PIPE) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("{") reduce 2) (("|") reduce 2)) 6242 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) ((",") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) ((";") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_PIPE) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("{") reduce 9) (("|") reduce 9)) 6243 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) ((",") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) ((";") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_PIPE) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("{") reduce 33) (("|") reduce 33)) 6244 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) ((",") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) ((";") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_PIPE) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("{") reduce 58) (("|") reduce 58)) 6245 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) ((",") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) ((";") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_PIPE) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("{") reduce 37) (("|") reduce 37)) 6246 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) ((",") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) ((";") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_PIPE) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("{") reduce 64) (("|") reduce 64)) 6247 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) ((",") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) ((";") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_PIPE) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("{") reduce 46) (("|") reduce 46)) 6248 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) ((",") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) ((";") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_PIPE) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("{") reduce 8) (("|") reduce 8)) 6249 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) ((",") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) ((";") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_PIPE) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("{") reduce 6) (("|") reduce 6)) 6250 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) ((",") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) ((";") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_PIPE) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("{") reduce 7) (("|") reduce 7)) 6251 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) ((",") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) ((";") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_PIPE) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("{") reduce 69) (("|") reduce 69)) 6252 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) ((",") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) ((";") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_PIPE) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("{") reduce 63) (("|") reduce 63)) 6253 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) ((",") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) ((";") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_PIPE) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("{") reduce 56) (("|") reduce 56)) 6254 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) ((",") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) ((";") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_PIPE) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("{") reduce 10) (("|") reduce 10)) 6255 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) ((",") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) ((";") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_PIPE) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("{") reduce 67) (("|") reduce 67)) 6256 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) ((",") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) ((";") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_PIPE) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("{") reduce 44) (("|") reduce 44)) 6257 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) ((",") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) ((";") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_PIPE) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("{") reduce 76) (("|") reduce 76)) 6258 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) ((",") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) ((";") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_PIPE) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("{") reduce 71) (("|") reduce 71)) 6259 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) ((",") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) ((";") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_PIPE) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("{") reduce 77) (("|") reduce 77)) 6260 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) ((",") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) ((";") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_PIPE) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("{") reduce 78) (("|") reduce 78)) 6261 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) ((",") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) ((";") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_PIPE) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("{") reduce 79) (("|") reduce 79)) 6262 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) ((",") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) ((";") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_PIPE) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("{") reduce 4) (("|") reduce 4)) 6263 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) ((",") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) ((";") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_PIPE) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("{") reduce 5) (("|") reduce 5)) 6264 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) ((",") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) ((";") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_PIPE) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("{") reduce 43) (("|") reduce 43)) 6265 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) ((",") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) ((";") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_PIPE) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("{") reduce 73) (("|") reduce 73)) 6266 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) ((",") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) ((";") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_PIPE) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("{") reduce 82) (("|") reduce 82)) 6267 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) ((",") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) ((";") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_PIPE) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("{") reduce 47) (("|") reduce 47)) 6268 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) ((",") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) ((";") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_PIPE) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("{") reduce 31) (("|") reduce 31)) 6269 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) ((",") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) ((";") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_PIPE) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("{") reduce 57) (("|") reduce 57)) 6270 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) ((",") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) ((";") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_PIPE) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("{") reduce 61) (("|") reduce 61)) 6271 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) ((",") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) ((";") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_PIPE) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("{") reduce 28) (("|") reduce 28)) 6272 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) ((",") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) ((";") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_PIPE) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("{") reduce 36) (("|") reduce 36)) 6273 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) ((",") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) ((";") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_PIPE) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("{") reduce 32) (("|") reduce 32)) 6274 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) ((",") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) ((";") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_PIPE) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("{") reduce 35) (("|") reduce 35)) 6275 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) ((",") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) ((";") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_PIPE) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("{") reduce 19) (("|") reduce 19)) 6276 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) ((",") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) ((";") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_PIPE) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("{") reduce 45) (("|") reduce 45)) 6277 ((("%") reduce 542) (("(") reduce 579) (("*") reduce 542) (("+") reduce 542) ((",") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) ((";") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PIPE) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("{") reduce 542) (("|") reduce 542)) 6279 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) ((",") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) ((";") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_PIPE) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("{") reduce 72) (("|") reduce 72)) 6280 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) ((",") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) ((";") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_PIPE) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("{") reduce 83) (("|") reduce 83)) 6281 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) ((",") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) ((";") reduce 572) (("<") reduce 572) (("=") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AND_EQUAL) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_COALESCE_EQUAL) reduce 572) ((T_CONCAT_EQUAL) reduce 572) ((T_DEC) reduce 572) ((T_DIV_EQUAL) reduce 572) ((T_INC) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_MINUS_EQUAL) reduce 572) ((T_MOD_EQUAL) reduce 572) ((T_MUL_EQUAL) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_OR_EQUAL) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_PLUS_EQUAL) reduce 572) ((T_POW) reduce 572) ((T_POW_EQUAL) reduce 572) ((T_SL) reduce 572) ((T_SL_EQUAL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) ((T_SR_EQUAL) reduce 572) ((T_XOR_EQUAL) reduce 572) (("[") reduce 572) (("^") reduce 572) (("{") reduce 572) (("|") reduce 572)) 6284 ((("%") reduce 544) (("(") reduce 580) (("*") reduce 544) (("+") reduce 544) ((",") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) ((";") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PIPE) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("{") reduce 544) (("|") reduce 544)) 6286 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((";") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 6287 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((";") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("{") reduce 291) (("|") reduce 291)) 6291 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) ((";") reduce 292) (("<") reduce 292) (("=") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AND_EQUAL) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_COALESCE_EQUAL) reduce 292) ((T_CONCAT_EQUAL) reduce 292) ((T_DEC) reduce 292) ((T_DIV_EQUAL) reduce 292) ((T_INC) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_MINUS_EQUAL) reduce 292) ((T_MOD_EQUAL) reduce 292) ((T_MUL_EQUAL) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_OR_EQUAL) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_PLUS_EQUAL) reduce 292) ((T_POW) reduce 292) ((T_POW_EQUAL) reduce 292) ((T_SL) reduce 292) ((T_SL_EQUAL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) ((T_SR_EQUAL) reduce 292) ((T_XOR_EQUAL) reduce 292) (("[") reduce 292) (("^") reduce 292) (("{") reduce 292) (("|") reduce 292)) 6292 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) ((";") reduce 293) (("<") reduce 293) (("=") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AND_EQUAL) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_COALESCE_EQUAL) reduce 293) ((T_CONCAT_EQUAL) reduce 293) ((T_DEC) reduce 293) ((T_DIV_EQUAL) reduce 293) ((T_INC) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_MINUS_EQUAL) reduce 293) ((T_MOD_EQUAL) reduce 293) ((T_MUL_EQUAL) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_OR_EQUAL) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_PLUS_EQUAL) reduce 293) ((T_POW) reduce 293) ((T_POW_EQUAL) reduce 293) ((T_SL) reduce 293) ((T_SL_EQUAL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) ((T_SR_EQUAL) reduce 293) ((T_XOR_EQUAL) reduce 293) (("[") reduce 293) (("^") reduce 293) (("{") reduce 293) (("|") reduce 293)) 6297 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) ((",") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) ((";") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_PIPE) reduce 425) ((T_POW) reduce 425) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("{") reduce 425) (("|") reduce 425)) 6299 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) ((",") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) ((";") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_PIPE) reduce 423) ((T_POW) reduce 423) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("{") reduce 423) (("|") reduce 423)) 6309 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 419) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 419) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 419) (("|") shift)) 6310 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 421) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 421) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 421) (("|") shift)) 6311 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 420) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 420) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 420) (("|") shift)) 6312 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 413) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 413) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 413) (("|") shift)) 6313 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 410) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 410) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 410) (("|") shift)) 6314 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 418) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 418) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 418) (("|") shift)) 6315 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 412) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 412) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 412) (("|") shift)) 6316 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 416) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 416) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 416) (("|") shift)) 6317 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 411) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 411) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 411) (("|") shift)) 6318 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 414) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 414) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 414) (("|") shift)) 6319 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 415) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 415) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 415) (("|") shift)) 6320 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 422) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 422) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 422) (("|") shift)) 6321 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 417) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 417) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 417) (("|") shift)) 6323 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 406) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 406) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 406) (("|") shift)) 6326 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) ((";") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("{") reduce 568) (("|") reduce 568)) 6329 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) ((",") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) ((";") reduce 564) (("<") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_POW) reduce 564) ((T_SL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) (("[") reduce 564) (("^") reduce 564) (("{") reduce 564) (("|") reduce 564)) 6331 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) ((",") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) ((";") reduce 563) (("<") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_POW) reduce 563) ((T_SL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) (("[") reduce 563) (("^") reduce 563) (("{") reduce 563) (("|") reduce 563)) 6333 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) ((";") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("{") reduce 559) (("|") reduce 559)) 6334 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((";") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("{") reduce 565) (("|") reduce 565)) 6335 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) ((",") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) ((";") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("[") reduce 548) (("^") reduce 407) (("{") reduce 407) (("|") reduce 407)) 6339 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) ((",") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) ((";") reduce 572) (("<") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_POW) reduce 572) ((T_SL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) (("[") reduce 572) (("^") reduce 572) (("{") reduce 572) (("|") reduce 572)) 6341 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((";") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 6342 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((";") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("{") reduce 291) (("|") reduce 291)) 6346 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) ((";") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("[") reduce 292) (("^") reduce 292) (("{") reduce 292) (("|") reduce 292)) 6347 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) ((";") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("[") reduce 293) (("^") reduce 293) (("{") reduce 293) (("|") reduce 293)) 6348 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) ((";") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("{") reduce 504) (("|") reduce 504)) 6351 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) ((",") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) ((";") reduce 571) (("<") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_POW) reduce 571) ((T_SL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) (("[") reduce 571) (("^") reduce 571) (("{") reduce 571) (("|") reduce 571)) 6352 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((";") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 6353 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((";") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 6358 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) ((";") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("{") reduce 560) (("|") reduce 560)) 6359 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) ((";") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("{") reduce 582) (("|") reduce 582)) 6360 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((";") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("{") reduce 566) (("|") reduce 566)) 6361 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) ((";") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("{") reduce 584) (("|") reduce 584)) 6364 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) ((";") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("{") reduce 583) (("|") reduce 583)) 6365 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((";") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("{") reduce 561) (("|") reduce 561)) 6366 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((";") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("{") reduce 567) (("|") reduce 567)) 6367 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) ((",") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) ((";") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("[") reduce 562) (("^") reduce 562) (("{") reduce 562) (("|") reduce 562)) 6368 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((";") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 6369 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) ((";") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("{") reduce 570) (("|") reduce 570)) 6372 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) ((";") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("{") reduce 569) (("|") reduce 569)) 6373 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) ((";") reduce 504) (("<") reduce 504) (("=") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AND_EQUAL) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_COALESCE_EQUAL) reduce 504) ((T_CONCAT_EQUAL) reduce 504) ((T_DEC) reduce 504) ((T_DIV_EQUAL) reduce 504) ((T_INC) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_MINUS_EQUAL) reduce 504) ((T_MOD_EQUAL) reduce 504) ((T_MUL_EQUAL) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_OR_EQUAL) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_PLUS_EQUAL) reduce 504) ((T_POW) reduce 504) ((T_POW_EQUAL) reduce 504) ((T_SL) reduce 504) ((T_SL_EQUAL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) ((T_SR_EQUAL) reduce 504) ((T_XOR_EQUAL) reduce 504) (("[") reduce 504) (("^") reduce 504) (("{") reduce 504) (("|") reduce 504)) 6384 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) ((",") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) ((";") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PIPE) reduce 494) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("{") reduce 494) (("|") reduce 494)) 6385 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) ((";") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_PIPE) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("^") reduce 489) (("{") reduce 489) (("|") reduce 489)) 6395 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 494) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 494) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 494) (("|") shift)) 6396 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) ((";") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_PIPE) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("^") reduce 490) (("{") reduce 490) (("|") reduce 490)) 6397 (((",") reduce 365) ((";") reduce 365) (("{") shift)) 6400 (((T_ABSTRACT) reduce 367) ((T_ATTRIBUTE) reduce 367) ((T_CASE) reduce 367) ((T_CONST) reduce 367) ((T_FINAL) reduce 367) ((T_FUNCTION) reduce 367) ((T_PRIVATE) reduce 367) ((T_PRIVATE_SET) reduce 367) ((T_PROTECTED) reduce 367) ((T_PROTECTED_SET) reduce 367) ((T_PUBLIC) reduce 367) ((T_PUBLIC_SET) reduce 367) ((T_READONLY) reduce 367) ((T_STATIC) reduce 367) ((T_USE) reduce 367) ((T_VAR) reduce 367) (("}") reduce 367)) 6402 ((("%") reduce 541) (("(") reduce 579) (("*") reduce 541) (("+") reduce 541) ((",") reduce 541) (("-") reduce 541) ((".") reduce 541) (("/") reduce 541) ((";") reduce 541) (("<") reduce 541) ((">") reduce 541) (("?") reduce 541) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_BOOLEAN_AND) reduce 541) ((T_BOOLEAN_OR) reduce 541) ((T_COALESCE) reduce 541) ((T_INSTANCEOF) reduce 541) ((T_IS_EQUAL) reduce 541) ((T_IS_GREATER_OR_EQUAL) reduce 541) ((T_IS_IDENTICAL) reduce 541) ((T_IS_NOT_EQUAL) reduce 541) ((T_IS_NOT_IDENTICAL) reduce 541) ((T_IS_SMALLER_OR_EQUAL) reduce 541) ((T_LOGICAL_AND) reduce 541) ((T_LOGICAL_OR) reduce 541) ((T_LOGICAL_XOR) reduce 541) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 541) ((T_OBJECT_OPERATOR) reduce 541) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 541) ((T_PIPE) reduce 541) ((T_POW) reduce 541) ((T_SL) reduce 541) ((T_SPACESHIP) reduce 541) ((T_SR) reduce 541) (("[") reduce 541) (("^") reduce 541) (("{") reduce 541) (("|") reduce 541)) 6404 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) ((",") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) ((";") reduce 571) (("<") reduce 571) (("=") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AND_EQUAL) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_COALESCE_EQUAL) reduce 571) ((T_CONCAT_EQUAL) reduce 571) ((T_DEC) reduce 571) ((T_DIV_EQUAL) reduce 571) ((T_INC) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_MINUS_EQUAL) reduce 571) ((T_MOD_EQUAL) reduce 571) ((T_MUL_EQUAL) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_OR_EQUAL) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_PLUS_EQUAL) reduce 571) ((T_POW) reduce 571) ((T_POW_EQUAL) reduce 571) ((T_SL) reduce 571) ((T_SL_EQUAL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) ((T_SR_EQUAL) reduce 571) ((T_XOR_EQUAL) reduce 571) (("[") reduce 571) (("^") reduce 571) (("{") reduce 571) (("|") reduce 571)) 6407 ((("%") reduce 543) (("(") reduce 580) (("*") reduce 543) (("+") reduce 543) ((",") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) ((";") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PIPE) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("{") reduce 543) (("|") reduce 543)) 6408 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((";") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 6409 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((";") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 6411 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) ((",") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) ((";") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_PIPE) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("^") reduce 485) (("{") reduce 485) (("|") reduce 485)) 6412 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) ((";") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_PIPE) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("^") reduce 487) (("{") reduce 487) (("|") reduce 487)) 6417 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) ((";") reduce 560) (("<") reduce 560) (("=") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AND_EQUAL) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_COALESCE_EQUAL) reduce 560) ((T_CONCAT_EQUAL) reduce 560) ((T_DEC) reduce 560) ((T_DIV_EQUAL) reduce 560) ((T_INC) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_MINUS_EQUAL) reduce 560) ((T_MOD_EQUAL) reduce 560) ((T_MUL_EQUAL) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_OR_EQUAL) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_PLUS_EQUAL) reduce 560) ((T_POW) reduce 560) ((T_POW_EQUAL) reduce 560) ((T_SL) reduce 560) ((T_SL_EQUAL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) ((T_SR_EQUAL) reduce 560) ((T_XOR_EQUAL) reduce 560) (("[") reduce 560) (("^") reduce 560) (("{") reduce 560) (("|") reduce 560)) 6418 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) ((";") reduce 582) (("<") reduce 582) (("=") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AND_EQUAL) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_COALESCE_EQUAL) reduce 582) ((T_CONCAT_EQUAL) reduce 582) ((T_DEC) reduce 582) ((T_DIV_EQUAL) reduce 582) ((T_INC) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_MINUS_EQUAL) reduce 582) ((T_MOD_EQUAL) reduce 582) ((T_MUL_EQUAL) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_OR_EQUAL) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_PLUS_EQUAL) reduce 582) ((T_POW) reduce 582) ((T_POW_EQUAL) reduce 582) ((T_SL) reduce 582) ((T_SL_EQUAL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) ((T_SR_EQUAL) reduce 582) ((T_XOR_EQUAL) reduce 582) (("[") reduce 582) (("^") reduce 582) (("{") reduce 582) (("|") reduce 582)) 6419 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((";") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("{") reduce 566) (("|") reduce 566)) 6420 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) ((";") reduce 584) (("<") reduce 584) (("=") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AND_EQUAL) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_COALESCE_EQUAL) reduce 584) ((T_CONCAT_EQUAL) reduce 584) ((T_DEC) reduce 584) ((T_DIV_EQUAL) reduce 584) ((T_INC) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_MINUS_EQUAL) reduce 584) ((T_MOD_EQUAL) reduce 584) ((T_MUL_EQUAL) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_OR_EQUAL) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_PLUS_EQUAL) reduce 584) ((T_POW) reduce 584) ((T_POW_EQUAL) reduce 584) ((T_SL) reduce 584) ((T_SL_EQUAL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) ((T_SR_EQUAL) reduce 584) ((T_XOR_EQUAL) reduce 584) (("[") reduce 584) (("^") reduce 584) (("{") reduce 584) (("|") reduce 584)) 6423 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) ((";") reduce 583) (("<") reduce 583) (("=") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AND_EQUAL) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_COALESCE_EQUAL) reduce 583) ((T_CONCAT_EQUAL) reduce 583) ((T_DEC) reduce 583) ((T_DIV_EQUAL) reduce 583) ((T_INC) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_MINUS_EQUAL) reduce 583) ((T_MOD_EQUAL) reduce 583) ((T_MUL_EQUAL) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_OR_EQUAL) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_PLUS_EQUAL) reduce 583) ((T_POW) reduce 583) ((T_POW_EQUAL) reduce 583) ((T_SL) reduce 583) ((T_SL_EQUAL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) ((T_SR_EQUAL) reduce 583) ((T_XOR_EQUAL) reduce 583) (("[") reduce 583) (("^") reduce 583) (("{") reduce 583) (("|") reduce 583)) 6424 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((";") reduce 561) (("<") reduce 561) (("=") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AND_EQUAL) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_COALESCE_EQUAL) reduce 561) ((T_CONCAT_EQUAL) reduce 561) ((T_DEC) reduce 561) ((T_DIV_EQUAL) reduce 561) ((T_INC) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_MINUS_EQUAL) reduce 561) ((T_MOD_EQUAL) reduce 561) ((T_MUL_EQUAL) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_OR_EQUAL) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_PLUS_EQUAL) reduce 561) ((T_POW) reduce 561) ((T_POW_EQUAL) reduce 561) ((T_SL) reduce 561) ((T_SL_EQUAL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) ((T_SR_EQUAL) reduce 561) ((T_XOR_EQUAL) reduce 561) (("[") reduce 561) (("^") reduce 561) (("{") reduce 561) (("|") reduce 561)) 6425 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((";") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("{") reduce 567) (("|") reduce 567)) 6426 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) ((",") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) ((";") reduce 562) (("<") reduce 562) (("=") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AND_EQUAL) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_COALESCE_EQUAL) reduce 562) ((T_CONCAT_EQUAL) reduce 562) ((T_DEC) reduce 562) ((T_DIV_EQUAL) reduce 562) ((T_INC) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_MINUS_EQUAL) reduce 562) ((T_MOD_EQUAL) reduce 562) ((T_MUL_EQUAL) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_OR_EQUAL) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_PLUS_EQUAL) reduce 562) ((T_POW) reduce 562) ((T_POW_EQUAL) reduce 562) ((T_SL) reduce 562) ((T_SL_EQUAL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) ((T_SR_EQUAL) reduce 562) ((T_XOR_EQUAL) reduce 562) (("[") reduce 562) (("^") reduce 562) (("{") reduce 562) (("|") reduce 562)) 6428 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) ((",") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) ((";") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_PIPE) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("{") reduce 477) (("|") reduce 477)) 6430 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) ((";") reduce 520) (("<") reduce 520) (("=") shift) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_PIPE) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 6432 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 405) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 405) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 405) (("|") shift)) 6433 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 482) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 482) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 482) (("|") shift)) 6444 ((("%") reduce 540) (("*") reduce 540) (("+") reduce 540) ((",") reduce 540) (("-") reduce 540) ((".") reduce 540) (("/") reduce 540) ((";") reduce 540) (("<") reduce 540) ((">") reduce 540) (("?") reduce 540) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_BOOLEAN_AND) reduce 540) ((T_BOOLEAN_OR) reduce 540) ((T_COALESCE) reduce 540) ((T_DOUBLE_ARROW) reduce 540) ((T_INSTANCEOF) reduce 540) ((T_IS_EQUAL) reduce 540) ((T_IS_GREATER_OR_EQUAL) reduce 540) ((T_IS_IDENTICAL) reduce 540) ((T_IS_NOT_EQUAL) reduce 540) ((T_IS_NOT_IDENTICAL) reduce 540) ((T_IS_SMALLER_OR_EQUAL) reduce 540) ((T_LOGICAL_AND) reduce 540) ((T_LOGICAL_OR) reduce 540) ((T_LOGICAL_XOR) reduce 540) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 540) ((T_OBJECT_OPERATOR) reduce 540) ((T_PIPE) reduce 540) ((T_POW) reduce 540) ((T_SL) reduce 540) ((T_SPACESHIP) reduce 540) ((T_SR) reduce 540) (("[") reduce 540) (("^") reduce 540) (("{") reduce 540) (("|") reduce 540)) 6446 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) ((";") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_PIPE) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 6448 ((("%") reduce 534) (("*") reduce 534) (("+") reduce 534) ((",") reduce 534) (("-") reduce 534) ((".") reduce 534) (("/") reduce 534) ((";") reduce 534) (("<") reduce 534) ((">") reduce 534) (("?") reduce 534) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_BOOLEAN_AND) reduce 534) ((T_BOOLEAN_OR) reduce 534) ((T_COALESCE) reduce 534) ((T_DOUBLE_ARROW) reduce 534) ((T_INSTANCEOF) reduce 534) ((T_IS_EQUAL) reduce 534) ((T_IS_GREATER_OR_EQUAL) reduce 534) ((T_IS_IDENTICAL) reduce 534) ((T_IS_NOT_EQUAL) reduce 534) ((T_IS_NOT_IDENTICAL) reduce 534) ((T_IS_SMALLER_OR_EQUAL) reduce 534) ((T_LOGICAL_AND) reduce 534) ((T_LOGICAL_OR) reduce 534) ((T_LOGICAL_XOR) reduce 534) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PIPE) reduce 534) ((T_POW) reduce 534) ((T_SL) reduce 534) ((T_SPACESHIP) reduce 534) ((T_SR) reduce 534) (("[") reduce 534) (("^") reduce 534) (("{") reduce 534) (("|") reduce 534)) 6449 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) ((";") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_PIPE) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 6453 ((("%") reduce 517) (("(") shift) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((";") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_PIPE) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 6454 ((("%") reduce 533) (("*") reduce 533) (("+") reduce 533) ((",") reduce 533) (("-") reduce 533) ((".") reduce 533) (("/") reduce 533) ((";") reduce 533) (("<") reduce 533) ((">") reduce 533) (("?") reduce 533) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_BOOLEAN_AND) reduce 533) ((T_BOOLEAN_OR) reduce 533) ((T_COALESCE) reduce 533) ((T_DOUBLE_ARROW) reduce 533) ((T_INSTANCEOF) reduce 533) ((T_IS_EQUAL) reduce 533) ((T_IS_GREATER_OR_EQUAL) reduce 533) ((T_IS_IDENTICAL) reduce 533) ((T_IS_NOT_EQUAL) reduce 533) ((T_IS_NOT_IDENTICAL) reduce 533) ((T_IS_SMALLER_OR_EQUAL) reduce 533) ((T_LOGICAL_AND) reduce 533) ((T_LOGICAL_OR) reduce 533) ((T_LOGICAL_XOR) reduce 533) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PIPE) reduce 533) ((T_POW) reduce 533) ((T_SL) reduce 533) ((T_SPACESHIP) reduce 533) ((T_SR) reduce 533) (("[") reduce 533) (("^") reduce 533) (("{") reduce 533) (("|") reduce 533)) 6455 ((("%") reduce 537) (("*") reduce 537) (("+") reduce 537) ((",") reduce 537) (("-") reduce 537) ((".") reduce 537) (("/") reduce 537) ((";") reduce 537) (("<") reduce 537) ((">") reduce 537) (("?") reduce 537) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_BOOLEAN_AND) reduce 537) ((T_BOOLEAN_OR) reduce 537) ((T_COALESCE) reduce 537) ((T_DOUBLE_ARROW) reduce 537) ((T_INSTANCEOF) reduce 537) ((T_IS_EQUAL) reduce 537) ((T_IS_GREATER_OR_EQUAL) reduce 537) ((T_IS_IDENTICAL) reduce 537) ((T_IS_NOT_EQUAL) reduce 537) ((T_IS_NOT_IDENTICAL) reduce 537) ((T_IS_SMALLER_OR_EQUAL) reduce 537) ((T_LOGICAL_AND) reduce 537) ((T_LOGICAL_OR) reduce 537) ((T_LOGICAL_XOR) reduce 537) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_PIPE) reduce 537) ((T_POW) reduce 537) ((T_SL) reduce 537) ((T_SPACESHIP) reduce 537) ((T_SR) reduce 537) (("[") reduce 537) (("^") reduce 537) (("{") reduce 537) (("|") reduce 537)) 6461 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) ((";") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PIPE) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 6463 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) ((";") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_PIPE) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 6465 ((("%") reduce 536) (("*") reduce 536) (("+") reduce 536) ((",") reduce 536) (("-") reduce 536) ((".") reduce 536) (("/") reduce 536) ((";") reduce 536) (("<") reduce 536) ((">") reduce 536) (("?") reduce 536) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_BOOLEAN_AND) reduce 536) ((T_BOOLEAN_OR) reduce 536) ((T_COALESCE) reduce 536) ((T_DOUBLE_ARROW) reduce 536) ((T_INSTANCEOF) reduce 536) ((T_IS_EQUAL) reduce 536) ((T_IS_GREATER_OR_EQUAL) reduce 536) ((T_IS_IDENTICAL) reduce 536) ((T_IS_NOT_EQUAL) reduce 536) ((T_IS_NOT_IDENTICAL) reduce 536) ((T_IS_SMALLER_OR_EQUAL) reduce 536) ((T_LOGICAL_AND) reduce 536) ((T_LOGICAL_OR) reduce 536) ((T_LOGICAL_XOR) reduce 536) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 536) ((T_OBJECT_OPERATOR) reduce 536) ((T_PIPE) reduce 536) ((T_POW) reduce 536) ((T_SL) reduce 536) ((T_SPACESHIP) reduce 536) ((T_SR) reduce 536) (("[") reduce 536) (("^") reduce 536) (("{") reduce 536) (("|") reduce 536)) 6466 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 6467 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 6468 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("{") reduce 95) (("|") reduce 95)) 6470 ((("%") reduce 539) (("*") reduce 539) (("+") reduce 539) ((",") reduce 539) (("-") reduce 539) ((".") reduce 539) (("/") reduce 539) ((";") reduce 539) (("<") reduce 539) ((">") reduce 539) (("?") reduce 539) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_BOOLEAN_AND) reduce 539) ((T_BOOLEAN_OR) reduce 539) ((T_COALESCE) reduce 539) ((T_DOUBLE_ARROW) reduce 539) ((T_INSTANCEOF) reduce 539) ((T_IS_EQUAL) reduce 539) ((T_IS_GREATER_OR_EQUAL) reduce 539) ((T_IS_IDENTICAL) reduce 539) ((T_IS_NOT_EQUAL) reduce 539) ((T_IS_NOT_IDENTICAL) reduce 539) ((T_IS_SMALLER_OR_EQUAL) reduce 539) ((T_LOGICAL_AND) reduce 539) ((T_LOGICAL_OR) reduce 539) ((T_LOGICAL_XOR) reduce 539) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 539) ((T_OBJECT_OPERATOR) reduce 539) ((T_PIPE) reduce 539) ((T_POW) reduce 539) ((T_SL) reduce 539) ((T_SPACESHIP) reduce 539) ((T_SR) reduce 539) (("[") reduce 539) (("^") reduce 539) (("{") reduce 539) (("|") reduce 539)) 6473 ((("%") reduce 538) (("*") reduce 538) (("+") reduce 538) ((",") reduce 538) (("-") reduce 538) ((".") reduce 538) (("/") reduce 538) ((";") reduce 538) (("<") reduce 538) ((">") reduce 538) (("?") reduce 538) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_BOOLEAN_AND) reduce 538) ((T_BOOLEAN_OR) reduce 538) ((T_COALESCE) reduce 538) ((T_DOUBLE_ARROW) reduce 538) ((T_INSTANCEOF) reduce 538) ((T_IS_EQUAL) reduce 538) ((T_IS_GREATER_OR_EQUAL) reduce 538) ((T_IS_IDENTICAL) reduce 538) ((T_IS_NOT_EQUAL) reduce 538) ((T_IS_NOT_IDENTICAL) reduce 538) ((T_IS_SMALLER_OR_EQUAL) reduce 538) ((T_LOGICAL_AND) reduce 538) ((T_LOGICAL_OR) reduce 538) ((T_LOGICAL_XOR) reduce 538) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 538) ((T_OBJECT_OPERATOR) reduce 538) ((T_PIPE) reduce 538) ((T_POW) reduce 538) ((T_SL) reduce 538) ((T_SPACESHIP) reduce 538) ((T_SR) reduce 538) (("[") reduce 538) (("^") reduce 538) (("{") reduce 538) (("|") reduce 538)) 6479 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 6482 ((("%") reduce 535) (("*") reduce 535) (("+") reduce 535) ((",") reduce 535) (("-") reduce 535) ((".") reduce 535) (("/") reduce 535) ((";") reduce 535) (("<") reduce 535) ((">") reduce 535) (("?") reduce 535) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_BOOLEAN_AND) reduce 535) ((T_BOOLEAN_OR) reduce 535) ((T_COALESCE) reduce 535) ((T_DOUBLE_ARROW) reduce 535) ((T_INSTANCEOF) reduce 535) ((T_IS_EQUAL) reduce 535) ((T_IS_GREATER_OR_EQUAL) reduce 535) ((T_IS_IDENTICAL) reduce 535) ((T_IS_NOT_EQUAL) reduce 535) ((T_IS_NOT_IDENTICAL) reduce 535) ((T_IS_SMALLER_OR_EQUAL) reduce 535) ((T_LOGICAL_AND) reduce 535) ((T_LOGICAL_OR) reduce 535) ((T_LOGICAL_XOR) reduce 535) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PIPE) reduce 535) ((T_POW) reduce 535) ((T_SL) reduce 535) ((T_SPACESHIP) reduce 535) ((T_SR) reduce 535) (("[") reduce 535) (("^") reduce 535) (("{") reduce 535) (("|") reduce 535)) 6484 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) ((";") reduce 568) (("<") reduce 568) (("=") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AND_EQUAL) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_COALESCE_EQUAL) reduce 568) ((T_CONCAT_EQUAL) reduce 568) ((T_DEC) reduce 568) ((T_DIV_EQUAL) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INC) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_MINUS_EQUAL) reduce 568) ((T_MOD_EQUAL) reduce 568) ((T_MUL_EQUAL) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_OR_EQUAL) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_PLUS_EQUAL) reduce 568) ((T_POW) reduce 568) ((T_POW_EQUAL) reduce 568) ((T_SL) reduce 568) ((T_SL_EQUAL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) ((T_SR_EQUAL) reduce 568) ((T_XOR_EQUAL) reduce 568) (("[") reduce 568) (("^") reduce 568) (("{") reduce 568) (("|") reduce 568)) 6485 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 479) (("(") shift) (("*") reduce 479) (("+") shift) ((",") reduce 479) (("-") shift) ((".") reduce 479) (("/") reduce 479) ((";") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 479) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 479) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 479) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PIPE) reduce 479) ((T_POW) reduce 479) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 479) (("`") shift) (("{") reduce 479) (("|") reduce 479) (("~") shift)) 6492 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) ((",") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) ((";") reduce 564) (("<") reduce 564) (("=") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AND_EQUAL) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_COALESCE_EQUAL) reduce 564) ((T_CONCAT_EQUAL) reduce 564) ((T_DEC) reduce 564) ((T_DIV_EQUAL) reduce 564) ((T_DOUBLE_ARROW) reduce 564) ((T_INC) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_MINUS_EQUAL) reduce 564) ((T_MOD_EQUAL) reduce 564) ((T_MUL_EQUAL) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_OR_EQUAL) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_PLUS_EQUAL) reduce 564) ((T_POW) reduce 564) ((T_POW_EQUAL) reduce 564) ((T_SL) reduce 564) ((T_SL_EQUAL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) ((T_SR_EQUAL) reduce 564) ((T_XOR_EQUAL) reduce 564) (("[") reduce 564) (("^") reduce 564) (("{") reduce 564) (("|") reduce 564)) 6493 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) ((";") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PIPE) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 551) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 6495 ((("%") reduce 529) (("*") reduce 529) (("+") reduce 529) ((",") reduce 529) (("-") reduce 529) ((".") reduce 529) (("/") reduce 529) ((";") reduce 529) (("<") reduce 529) ((">") reduce 529) (("?") reduce 529) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_BOOLEAN_AND) reduce 529) ((T_BOOLEAN_OR) reduce 529) ((T_COALESCE) reduce 529) ((T_DOUBLE_ARROW) reduce 529) ((T_INSTANCEOF) reduce 529) ((T_IS_EQUAL) reduce 529) ((T_IS_GREATER_OR_EQUAL) reduce 529) ((T_IS_IDENTICAL) reduce 529) ((T_IS_NOT_EQUAL) reduce 529) ((T_IS_NOT_IDENTICAL) reduce 529) ((T_IS_SMALLER_OR_EQUAL) reduce 529) ((T_LOGICAL_AND) reduce 529) ((T_LOGICAL_OR) reduce 529) ((T_LOGICAL_XOR) reduce 529) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PIPE) reduce 529) ((T_POW) reduce 529) ((T_SL) reduce 529) ((T_SPACESHIP) reduce 529) ((T_SR) reduce 529) (("[") reduce 554) (("^") reduce 529) (("{") reduce 529) (("|") reduce 529)) 6496 ((("%") reduce 528) (("(") reduce 557) (("*") reduce 528) (("+") reduce 528) ((",") reduce 528) (("-") reduce 528) ((".") reduce 528) (("/") reduce 528) ((";") reduce 528) (("<") reduce 528) ((">") reduce 528) (("?") reduce 528) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_BOOLEAN_AND) reduce 528) ((T_BOOLEAN_OR) reduce 528) ((T_COALESCE) reduce 528) ((T_DOUBLE_ARROW) reduce 528) ((T_INSTANCEOF) reduce 528) ((T_IS_EQUAL) reduce 528) ((T_IS_GREATER_OR_EQUAL) reduce 528) ((T_IS_IDENTICAL) reduce 528) ((T_IS_NOT_EQUAL) reduce 528) ((T_IS_NOT_IDENTICAL) reduce 528) ((T_IS_SMALLER_OR_EQUAL) reduce 528) ((T_LOGICAL_AND) reduce 528) ((T_LOGICAL_OR) reduce 528) ((T_LOGICAL_XOR) reduce 528) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PIPE) reduce 528) ((T_POW) reduce 528) ((T_SL) reduce 528) ((T_SPACESHIP) reduce 528) ((T_SR) reduce 528) (("[") reduce 550) (("^") reduce 528) (("{") reduce 528) (("|") reduce 528)) 6497 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 480) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 480) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 480) ((T_LOGICAL_OR) reduce 480) ((T_LOGICAL_XOR) reduce 480) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 480) (("|") shift)) 6500 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) ((",") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) ((";") reduce 563) (("<") reduce 563) (("=") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AND_EQUAL) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_COALESCE_EQUAL) reduce 563) ((T_CONCAT_EQUAL) reduce 563) ((T_DEC) reduce 563) ((T_DIV_EQUAL) reduce 563) ((T_DOUBLE_ARROW) reduce 563) ((T_INC) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_MINUS_EQUAL) reduce 563) ((T_MOD_EQUAL) reduce 563) ((T_MUL_EQUAL) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_OR_EQUAL) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_PLUS_EQUAL) reduce 563) ((T_POW) reduce 563) ((T_POW_EQUAL) reduce 563) ((T_SL) reduce 563) ((T_SL_EQUAL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) ((T_SR_EQUAL) reduce 563) ((T_XOR_EQUAL) reduce 563) (("[") reduce 563) (("^") reduce 563) (("{") reduce 563) (("|") reduce 563)) 6501 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) ((",") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) ((";") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_DOUBLE_ARROW) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_PIPE) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("^") reduce 484) (("{") reduce 484) (("|") reduce 484)) 6502 ((("%") reduce 466) (("*") reduce 466) (("+") reduce 466) ((",") reduce 466) (("-") reduce 466) ((".") reduce 466) (("/") reduce 466) ((";") reduce 466) (("<") reduce 466) ((">") reduce 466) (("?") reduce 466) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_BOOLEAN_AND) reduce 466) ((T_BOOLEAN_OR) reduce 466) ((T_COALESCE) reduce 466) ((T_DOUBLE_ARROW) reduce 466) ((T_INSTANCEOF) reduce 466) ((T_IS_EQUAL) reduce 466) ((T_IS_GREATER_OR_EQUAL) reduce 466) ((T_IS_IDENTICAL) reduce 466) ((T_IS_NOT_EQUAL) reduce 466) ((T_IS_NOT_IDENTICAL) reduce 466) ((T_IS_SMALLER_OR_EQUAL) reduce 466) ((T_LOGICAL_AND) reduce 466) ((T_LOGICAL_OR) reduce 466) ((T_LOGICAL_XOR) reduce 466) ((T_PIPE) reduce 466) ((T_POW) reduce 466) ((T_SL) reduce 466) ((T_SPACESHIP) reduce 466) ((T_SR) reduce 466) (("^") reduce 466) (("{") reduce 466) (("|") reduce 466)) 6503 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) ((";") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_PIPE) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("^") reduce 488) (("{") reduce 488) (("|") reduce 488)) 6504 ((("%") reduce 531) (("(") shift) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) ((";") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 6505 ((("%") reduce 461) (("(") reduce 558) (("*") reduce 461) (("+") reduce 461) ((",") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) ((";") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PIPE) reduce 461) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") reduce 552) (("^") reduce 461) (("{") reduce 461) (("|") reduce 461)) 6506 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) ((",") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) ((";") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_DOUBLE_ARROW) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_PIPE) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("{") reduce 462) (("|") reduce 462)) 6507 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) ((",") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) ((";") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_DOUBLE_ARROW) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_PIPE) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("{") reduce 476) (("|") reduce 476)) 6508 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) ((";") reduce 559) (("<") reduce 559) (("=") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AND_EQUAL) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_COALESCE_EQUAL) reduce 559) ((T_CONCAT_EQUAL) reduce 559) ((T_DEC) reduce 559) ((T_DIV_EQUAL) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INC) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_MINUS_EQUAL) reduce 559) ((T_MOD_EQUAL) reduce 559) ((T_MUL_EQUAL) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_OR_EQUAL) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_PLUS_EQUAL) reduce 559) ((T_POW) reduce 559) ((T_POW_EQUAL) reduce 559) ((T_SL) reduce 559) ((T_SL_EQUAL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) ((T_SR_EQUAL) reduce 559) ((T_XOR_EQUAL) reduce 559) (("[") reduce 559) (("^") reduce 559) (("{") reduce 559) (("|") reduce 559)) 6509 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((";") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("{") reduce 565) (("|") reduce 565)) 6510 ((("%") reduce 403) (("*") reduce 403) (("+") reduce 403) ((",") reduce 403) (("-") reduce 403) ((".") reduce 403) (("/") reduce 403) ((";") reduce 403) (("<") reduce 403) (("=") shift) ((">") reduce 403) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_DOUBLE_ARROW) reduce 403) ((T_INC) shift) ((T_INSTANCEOF) reduce 403) ((T_IS_EQUAL) reduce 403) ((T_IS_GREATER_OR_EQUAL) reduce 403) ((T_IS_IDENTICAL) reduce 403) ((T_IS_NOT_EQUAL) reduce 403) ((T_IS_NOT_IDENTICAL) reduce 403) ((T_IS_SMALLER_OR_EQUAL) reduce 403) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 403) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 403) ((T_POW_EQUAL) shift) ((T_SL) reduce 403) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 403) ((T_SR) reduce 403) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 548) (("^") reduce 403) (("{") reduce 403) (("|") reduce 403)) 6513 ((("%") reduce 448) (("*") reduce 448) (("+") reduce 448) ((",") reduce 448) (("-") reduce 448) ((".") reduce 448) (("/") reduce 448) ((";") reduce 448) (("<") reduce 448) ((">") reduce 448) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_BOOLEAN_AND) reduce 448) ((T_BOOLEAN_OR) reduce 448) ((T_COALESCE) reduce 448) ((T_DOUBLE_ARROW) reduce 448) ((T_INSTANCEOF) reduce 448) ((T_IS_EQUAL) reduce 448) ((T_IS_GREATER_OR_EQUAL) reduce 448) ((T_IS_IDENTICAL) reduce 448) ((T_IS_NOT_EQUAL) reduce 448) ((T_IS_NOT_IDENTICAL) reduce 448) ((T_IS_SMALLER_OR_EQUAL) reduce 448) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_PIPE) reduce 448) ((T_POW) shift) ((T_SL) reduce 448) ((T_SPACESHIP) reduce 448) ((T_SR) reduce 448) (("^") reduce 448) (("{") reduce 448) (("|") reduce 448)) 6545 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 432) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 432) (("<") shift) ((">") shift) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_DOUBLE_ARROW) reduce 432) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 432) (("|") reduce 432)) 6546 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 435) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 435) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_DOUBLE_ARROW) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("{") reduce 435) (("|") reduce 435)) 6547 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 444) (("-") shift) ((".") reduce 444) (("/") shift) ((";") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_PIPE) reduce 444) ((T_POW) shift) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("{") reduce 444) (("|") reduce 444)) 6548 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 458) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 458) (("<") shift) ((">") shift) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 458) (("{") reduce 458) (("|") reduce 458)) 6549 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 443) (("-") shift) ((".") reduce 443) (("/") shift) ((";") reduce 443) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_DOUBLE_ARROW) reduce 443) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_PIPE) reduce 443) ((T_POW) shift) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("^") reduce 443) (("{") reduce 443) (("|") reduce 443)) 6550 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) ((",") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) ((";") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_DOUBLE_ARROW) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_PIPE) reduce 440) ((T_POW) shift) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("{") reduce 440) (("|") reduce 440)) 6551 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 453) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 453) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_DOUBLE_ARROW) reduce 453) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_PIPE) reduce 453) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 453) ((T_SR) shift) (("^") reduce 453) (("{") reduce 453) (("|") reduce 453)) 6552 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 431) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 431) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 431) (("|") shift)) 6553 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 429) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 429) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 429) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 429) (("|") shift)) 6554 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 430) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 430) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 430) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 430) (("|") shift)) 6555 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 455) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 455) (("<") shift) ((">") shift) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 455) ((T_SR) shift) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 6556 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 450) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 450) (("<") shift) ((">") shift) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_DOUBLE_ARROW) reduce 450) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 450) (("{") reduce 450) (("|") reduce 450)) 6557 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 452) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 452) (("<") shift) ((">") shift) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_DOUBLE_ARROW) reduce 452) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 452) (("{") reduce 452) (("|") reduce 452)) 6558 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 449) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 449) (("<") shift) ((">") shift) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_DOUBLE_ARROW) reduce 449) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 449) (("{") reduce 449) (("|") reduce 449)) 6559 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 457) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 457) (("<") shift) ((">") shift) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 457) ((T_SR) shift) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 6560 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 451) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 451) (("<") shift) ((">") shift) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_DOUBLE_ARROW) reduce 451) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 451) (("{") reduce 451) (("|") reduce 451)) 6563 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 6564 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 6565 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("{") reduce 95) (("|") reduce 95)) 6566 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((";") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 6567 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 6568 ((("%") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) ((";") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("{") reduce 568) (("|") reduce 568)) 6569 ((("%") reduce 511) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((";") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 6570 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) ((";") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PIPE) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("{") reduce 459) (("|") reduce 459)) 6571 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((";") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 6572 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((";") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 6573 ((("%") reduce 573) (("*") reduce 573) (("+") reduce 573) ((",") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) ((";") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_DOUBLE_ARROW) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("{") reduce 573) (("|") reduce 573)) 6579 ((("%") reduce 574) (("*") reduce 574) (("+") reduce 574) ((",") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) ((";") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_DOUBLE_ARROW) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("{") reduce 574) (("|") reduce 574)) 6580 ((("%") reduce 578) (("*") reduce 578) (("+") reduce 578) ((",") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) ((";") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_DOUBLE_ARROW) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("{") reduce 578) (("|") reduce 578)) 6581 ((("%") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) ((";") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("{") reduce 582) (("|") reduce 582)) 6582 ((("%") reduce 575) (("*") reduce 575) (("+") reduce 575) ((",") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) ((";") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_DOUBLE_ARROW) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("{") reduce 575) (("|") reduce 575)) 6583 ((("%") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) ((";") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("{") reduce 584) (("|") reduce 584)) 6586 ((("%") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) ((";") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("{") reduce 583) (("|") reduce 583)) 6587 ((("%") reduce 576) (("*") reduce 576) (("+") reduce 576) ((",") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) ((";") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_DOUBLE_ARROW) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("{") reduce 576) (("|") reduce 576)) 6589 ((("%") reduce 577) (("*") reduce 577) (("+") reduce 577) ((",") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) ((";") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_DOUBLE_ARROW) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("{") reduce 577) (("|") reduce 577)) 6591 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((";") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 6592 ((("%") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) ((";") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("{") reduce 570) (("|") reduce 570)) 6595 ((("%") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) ((";") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("{") reduce 569) (("|") reduce 569)) 6596 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 465) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 465) (("<") shift) ((">") shift) (("?") reduce 465) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 465) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 465) (("|") shift)) 6597 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 427) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 427) (("<") shift) ((">") shift) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_DOUBLE_ARROW) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 427) (("|") shift)) 6598 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 428) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 428) (("<") shift) ((">") shift) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_DOUBLE_ARROW) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 428) (("|") shift)) 6599 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 433) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 433) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_DOUBLE_ARROW) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("{") reduce 433) (("|") reduce 433)) 6600 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 434) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 434) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_DOUBLE_ARROW) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("{") reduce 434) (("|") reduce 434)) 6604 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 463) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 463) (("<") shift) ((">") shift) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 463) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 463) (("|") shift)) 6605 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 464) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 464) (("<") shift) ((">") shift) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 464) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 464) (("|") shift)) 6606 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 456) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 456) (("<") shift) ((">") shift) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 456) ((T_SR) shift) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 6607 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 454) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 454) (("<") shift) ((">") shift) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 454) ((T_SR) shift) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 6608 ((("%") reduce 441) (("*") reduce 441) (("+") reduce 441) ((",") reduce 441) (("-") reduce 441) ((".") reduce 441) (("/") reduce 441) ((";") reduce 441) (("<") reduce 441) ((">") reduce 441) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_DOUBLE_ARROW) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 441) ((T_IS_GREATER_OR_EQUAL) reduce 441) ((T_IS_IDENTICAL) reduce 441) ((T_IS_NOT_EQUAL) reduce 441) ((T_IS_NOT_IDENTICAL) reduce 441) ((T_IS_SMALLER_OR_EQUAL) reduce 441) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_PIPE) reduce 441) ((T_POW) shift) ((T_SL) reduce 441) ((T_SPACESHIP) reduce 441) ((T_SR) reduce 441) (("^") reduce 441) (("{") reduce 441) (("|") reduce 441)) 6609 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 436) (("-") shift) ((".") reduce 436) (("/") shift) ((";") reduce 436) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_DOUBLE_ARROW) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_PIPE) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 436) ((T_SR) shift) (("^") reduce 436) (("{") reduce 436) (("|") reduce 436)) 6610 ((("%") shift) (("*") shift) (("+") reduce 438) ((",") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") shift) ((";") reduce 438) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_DOUBLE_ARROW) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_PIPE) reduce 438) ((T_POW) shift) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("{") reduce 438) (("|") reduce 438)) 6611 ((("%") shift) (("*") shift) (("+") reduce 437) ((",") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") shift) ((";") reduce 437) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_DOUBLE_ARROW) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_PIPE) reduce 437) ((T_POW) shift) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("{") reduce 437) (("|") reduce 437)) 6612 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) ((",") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) ((";") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_DOUBLE_ARROW) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_PIPE) reduce 439) ((T_POW) shift) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("{") reduce 439) (("|") reduce 439)) 6613 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) ((",") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) ((";") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_DOUBLE_ARROW) reduce 442) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_PIPE) reduce 442) ((T_POW) shift) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("{") reduce 442) (("|") reduce 442)) 6615 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) ((",") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) ((";") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_DOUBLE_ARROW) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_PIPE) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("{") reduce 74) (("|") reduce 74)) 6616 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) ((",") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) ((";") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_DOUBLE_ARROW) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_PIPE) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("{") reduce 52) (("|") reduce 52)) 6617 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) ((",") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) ((";") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_DOUBLE_ARROW) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_PIPE) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("{") reduce 27) (("|") reduce 27)) 6618 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) ((",") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) ((";") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_DOUBLE_ARROW) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_PIPE) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("{") reduce 51) (("|") reduce 51)) 6619 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) ((",") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) ((";") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_DOUBLE_ARROW) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_PIPE) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("{") reduce 53) (("|") reduce 53)) 6620 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) ((",") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) ((";") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_DOUBLE_ARROW) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_PIPE) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("{") reduce 49) (("|") reduce 49)) 6621 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) ((",") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) ((";") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_DOUBLE_ARROW) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_PIPE) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("{") reduce 29) (("|") reduce 29)) 6622 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) ((",") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) ((";") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_DOUBLE_ARROW) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_PIPE) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("{") reduce 59) (("|") reduce 59)) 6623 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) ((",") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) ((";") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_DOUBLE_ARROW) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_PIPE) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("{") reduce 60) (("|") reduce 60)) 6624 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) ((",") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) ((";") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_DOUBLE_ARROW) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_PIPE) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("{") reduce 11) (("|") reduce 11)) 6625 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) ((",") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) ((";") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_DOUBLE_ARROW) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_PIPE) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("{") reduce 42) (("|") reduce 42)) 6626 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) ((",") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) ((";") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_DOUBLE_ARROW) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_PIPE) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("{") reduce 39) (("|") reduce 39)) 6627 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) ((",") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) ((";") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_DOUBLE_ARROW) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_PIPE) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("{") reduce 25) (("|") reduce 25)) 6628 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) ((",") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) ((";") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_DOUBLE_ARROW) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_PIPE) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("{") reduce 50) (("|") reduce 50)) 6629 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) ((",") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) ((";") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_DOUBLE_ARROW) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_PIPE) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("{") reduce 66) (("|") reduce 66)) 6630 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) ((",") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) ((";") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_DOUBLE_ARROW) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_PIPE) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("{") reduce 18) (("|") reduce 18)) 6631 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) ((",") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) ((";") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_DOUBLE_ARROW) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_PIPE) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("{") reduce 17) (("|") reduce 17)) 6632 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) ((",") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) ((";") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_DOUBLE_ARROW) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_PIPE) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("{") reduce 15) (("|") reduce 15)) 6633 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) ((",") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) ((";") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_DOUBLE_ARROW) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_PIPE) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("{") reduce 14) (("|") reduce 14)) 6634 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) ((",") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) ((";") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_DOUBLE_ARROW) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_PIPE) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("{") reduce 38) (("|") reduce 38)) 6635 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) ((",") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) ((";") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_DOUBLE_ARROW) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_PIPE) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("{") reduce 26) (("|") reduce 26)) 6636 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) ((",") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) ((";") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_DOUBLE_ARROW) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_PIPE) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("{") reduce 22) (("|") reduce 22)) 6637 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) ((",") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) ((";") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_DOUBLE_ARROW) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_PIPE) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("{") reduce 24) (("|") reduce 24)) 6638 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) ((",") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) ((";") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_DOUBLE_ARROW) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_PIPE) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("{") reduce 16) (("|") reduce 16)) 6639 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) ((",") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) ((";") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_DOUBLE_ARROW) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_PIPE) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("{") reduce 48) (("|") reduce 48)) 6640 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) ((",") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) ((";") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_DOUBLE_ARROW) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_PIPE) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("{") reduce 20) (("|") reduce 20)) 6641 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) ((",") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) ((";") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_DOUBLE_ARROW) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_PIPE) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("{") reduce 70) (("|") reduce 70)) 6642 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) ((",") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) ((";") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_DOUBLE_ARROW) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_PIPE) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("{") reduce 3) (("|") reduce 3)) 6643 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) ((",") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) ((";") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_DOUBLE_ARROW) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_PIPE) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("{") reduce 12) (("|") reduce 12)) 6644 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) ((",") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) ((";") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_DOUBLE_ARROW) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_PIPE) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("{") reduce 54) (("|") reduce 54)) 6645 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) ((",") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) ((";") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_DOUBLE_ARROW) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_PIPE) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("{") reduce 65) (("|") reduce 65)) 6646 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) ((",") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) ((";") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_DOUBLE_ARROW) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_PIPE) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("{") reduce 75) (("|") reduce 75)) 6647 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) ((",") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) ((";") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_DOUBLE_ARROW) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_PIPE) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("{") reduce 30) (("|") reduce 30)) 6648 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) ((",") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) ((";") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_DOUBLE_ARROW) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_PIPE) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("{") reduce 68) (("|") reduce 68)) 6649 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) ((",") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) ((";") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_DOUBLE_ARROW) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_PIPE) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("{") reduce 21) (("|") reduce 21)) 6650 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) ((",") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) ((";") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_DOUBLE_ARROW) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_PIPE) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("{") reduce 23) (("|") reduce 23)) 6651 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) ((",") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) ((";") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_DOUBLE_ARROW) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_PIPE) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("{") reduce 41) (("|") reduce 41)) 6652 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) ((",") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) ((";") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_DOUBLE_ARROW) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_PIPE) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("{") reduce 62) (("|") reduce 62)) 6653 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) ((",") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) ((";") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_DOUBLE_ARROW) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_PIPE) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("{") reduce 34) (("|") reduce 34)) 6654 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) ((",") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) ((";") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_DOUBLE_ARROW) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_PIPE) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("{") reduce 40) (("|") reduce 40)) 6655 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) ((",") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) ((";") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_DOUBLE_ARROW) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_PIPE) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("{") reduce 13) (("|") reduce 13)) 6656 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) ((",") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) ((";") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_DOUBLE_ARROW) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_PIPE) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("{") reduce 55) (("|") reduce 55)) 6657 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) ((",") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) ((";") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_DOUBLE_ARROW) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_PIPE) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("{") reduce 1) (("|") reduce 1)) 6658 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) ((",") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) ((";") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_DOUBLE_ARROW) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_PIPE) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("{") reduce 2) (("|") reduce 2)) 6659 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) ((",") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) ((";") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_DOUBLE_ARROW) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_PIPE) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("{") reduce 9) (("|") reduce 9)) 6660 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) ((",") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) ((";") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_DOUBLE_ARROW) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_PIPE) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("{") reduce 33) (("|") reduce 33)) 6661 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) ((",") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) ((";") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_DOUBLE_ARROW) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_PIPE) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("{") reduce 58) (("|") reduce 58)) 6662 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) ((",") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) ((";") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_DOUBLE_ARROW) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_PIPE) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("{") reduce 37) (("|") reduce 37)) 6663 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) ((",") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) ((";") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_DOUBLE_ARROW) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_PIPE) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("{") reduce 64) (("|") reduce 64)) 6664 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) ((",") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) ((";") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_DOUBLE_ARROW) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_PIPE) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("{") reduce 46) (("|") reduce 46)) 6665 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) ((",") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) ((";") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_DOUBLE_ARROW) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_PIPE) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("{") reduce 8) (("|") reduce 8)) 6666 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) ((",") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) ((";") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_DOUBLE_ARROW) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_PIPE) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("{") reduce 6) (("|") reduce 6)) 6667 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) ((",") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) ((";") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_DOUBLE_ARROW) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_PIPE) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("{") reduce 7) (("|") reduce 7)) 6668 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) ((",") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) ((";") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_DOUBLE_ARROW) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_PIPE) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("{") reduce 69) (("|") reduce 69)) 6669 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) ((",") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) ((";") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_DOUBLE_ARROW) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_PIPE) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("{") reduce 63) (("|") reduce 63)) 6670 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) ((",") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) ((";") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_DOUBLE_ARROW) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_PIPE) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("{") reduce 56) (("|") reduce 56)) 6671 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) ((",") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) ((";") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_DOUBLE_ARROW) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_PIPE) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("{") reduce 10) (("|") reduce 10)) 6672 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) ((",") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) ((";") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_DOUBLE_ARROW) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_PIPE) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("{") reduce 67) (("|") reduce 67)) 6673 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) ((",") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) ((";") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_DOUBLE_ARROW) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_PIPE) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("{") reduce 44) (("|") reduce 44)) 6674 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) ((",") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) ((";") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_DOUBLE_ARROW) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_PIPE) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("{") reduce 76) (("|") reduce 76)) 6675 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) ((",") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) ((";") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_DOUBLE_ARROW) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_PIPE) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("{") reduce 71) (("|") reduce 71)) 6676 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) ((",") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) ((";") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_DOUBLE_ARROW) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_PIPE) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("{") reduce 77) (("|") reduce 77)) 6677 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) ((",") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) ((";") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_DOUBLE_ARROW) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_PIPE) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("{") reduce 78) (("|") reduce 78)) 6678 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) ((",") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) ((";") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_DOUBLE_ARROW) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_PIPE) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("{") reduce 79) (("|") reduce 79)) 6679 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) ((",") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) ((";") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_DOUBLE_ARROW) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_PIPE) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("{") reduce 4) (("|") reduce 4)) 6680 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) ((",") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) ((";") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_DOUBLE_ARROW) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_PIPE) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("{") reduce 5) (("|") reduce 5)) 6681 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) ((",") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) ((";") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_DOUBLE_ARROW) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_PIPE) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("{") reduce 43) (("|") reduce 43)) 6682 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) ((",") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) ((";") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_DOUBLE_ARROW) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_PIPE) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("{") reduce 73) (("|") reduce 73)) 6683 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) ((",") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) ((";") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_DOUBLE_ARROW) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_PIPE) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("{") reduce 82) (("|") reduce 82)) 6684 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) ((",") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) ((";") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_DOUBLE_ARROW) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_PIPE) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("{") reduce 47) (("|") reduce 47)) 6685 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) ((",") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) ((";") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_DOUBLE_ARROW) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_PIPE) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("{") reduce 31) (("|") reduce 31)) 6686 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) ((",") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) ((";") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_DOUBLE_ARROW) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_PIPE) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("{") reduce 57) (("|") reduce 57)) 6687 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) ((",") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) ((";") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_DOUBLE_ARROW) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_PIPE) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("{") reduce 61) (("|") reduce 61)) 6688 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) ((",") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) ((";") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_DOUBLE_ARROW) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_PIPE) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("{") reduce 28) (("|") reduce 28)) 6689 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) ((",") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) ((";") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_DOUBLE_ARROW) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_PIPE) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("{") reduce 36) (("|") reduce 36)) 6690 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) ((",") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) ((";") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_DOUBLE_ARROW) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_PIPE) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("{") reduce 32) (("|") reduce 32)) 6691 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) ((",") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) ((";") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_DOUBLE_ARROW) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_PIPE) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("{") reduce 35) (("|") reduce 35)) 6692 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) ((",") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) ((";") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_DOUBLE_ARROW) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_PIPE) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("{") reduce 19) (("|") reduce 19)) 6693 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) ((",") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) ((";") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_DOUBLE_ARROW) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_PIPE) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("{") reduce 45) (("|") reduce 45)) 6694 ((("%") reduce 542) (("(") reduce 579) (("*") reduce 542) (("+") reduce 542) ((",") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) ((";") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PIPE) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("{") reduce 542) (("|") reduce 542)) 6696 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) ((",") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) ((";") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_DOUBLE_ARROW) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_PIPE) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("{") reduce 72) (("|") reduce 72)) 6697 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) ((",") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) ((";") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_DOUBLE_ARROW) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_PIPE) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("{") reduce 83) (("|") reduce 83)) 6698 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) ((",") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) ((";") reduce 572) (("<") reduce 572) (("=") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AND_EQUAL) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_COALESCE_EQUAL) reduce 572) ((T_CONCAT_EQUAL) reduce 572) ((T_DEC) reduce 572) ((T_DIV_EQUAL) reduce 572) ((T_DOUBLE_ARROW) reduce 572) ((T_INC) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_MINUS_EQUAL) reduce 572) ((T_MOD_EQUAL) reduce 572) ((T_MUL_EQUAL) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_OR_EQUAL) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_PLUS_EQUAL) reduce 572) ((T_POW) reduce 572) ((T_POW_EQUAL) reduce 572) ((T_SL) reduce 572) ((T_SL_EQUAL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) ((T_SR_EQUAL) reduce 572) ((T_XOR_EQUAL) reduce 572) (("[") reduce 572) (("^") reduce 572) (("{") reduce 572) (("|") reduce 572)) 6701 ((("%") reduce 544) (("(") reduce 580) (("*") reduce 544) (("+") reduce 544) ((",") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) ((";") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PIPE) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("{") reduce 544) (("|") reduce 544)) 6703 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((";") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 6704 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((";") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("{") reduce 291) (("|") reduce 291)) 6708 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) ((";") reduce 292) (("<") reduce 292) (("=") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AND_EQUAL) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_COALESCE_EQUAL) reduce 292) ((T_CONCAT_EQUAL) reduce 292) ((T_DEC) reduce 292) ((T_DIV_EQUAL) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INC) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_MINUS_EQUAL) reduce 292) ((T_MOD_EQUAL) reduce 292) ((T_MUL_EQUAL) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_OR_EQUAL) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_PLUS_EQUAL) reduce 292) ((T_POW) reduce 292) ((T_POW_EQUAL) reduce 292) ((T_SL) reduce 292) ((T_SL_EQUAL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) ((T_SR_EQUAL) reduce 292) ((T_XOR_EQUAL) reduce 292) (("[") reduce 292) (("^") reduce 292) (("{") reduce 292) (("|") reduce 292)) 6709 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) ((";") reduce 293) (("<") reduce 293) (("=") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AND_EQUAL) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_COALESCE_EQUAL) reduce 293) ((T_CONCAT_EQUAL) reduce 293) ((T_DEC) reduce 293) ((T_DIV_EQUAL) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INC) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_MINUS_EQUAL) reduce 293) ((T_MOD_EQUAL) reduce 293) ((T_MUL_EQUAL) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_OR_EQUAL) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_PLUS_EQUAL) reduce 293) ((T_POW) reduce 293) ((T_POW_EQUAL) reduce 293) ((T_SL) reduce 293) ((T_SL_EQUAL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) ((T_SR_EQUAL) reduce 293) ((T_XOR_EQUAL) reduce 293) (("[") reduce 293) (("^") reduce 293) (("{") reduce 293) (("|") reduce 293)) 6714 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) ((",") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) ((";") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_DOUBLE_ARROW) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_PIPE) reduce 425) ((T_POW) reduce 425) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("{") reduce 425) (("|") reduce 425)) 6716 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) ((",") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) ((";") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_DOUBLE_ARROW) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_PIPE) reduce 423) ((T_POW) reduce 423) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("{") reduce 423) (("|") reduce 423)) 6726 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 419) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 419) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 419) (("|") shift)) 6727 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 421) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 421) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 421) (("|") shift)) 6728 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 420) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 420) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 420) (("|") shift)) 6729 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 413) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 413) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 413) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 413) (("|") shift)) 6730 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 410) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 410) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 410) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 410) (("|") shift)) 6731 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 418) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 418) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 418) (("|") shift)) 6732 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 412) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 412) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 412) (("|") shift)) 6733 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 416) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 416) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 416) (("|") shift)) 6734 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 411) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 411) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 411) (("|") shift)) 6735 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 414) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 414) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 414) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 414) (("|") shift)) 6736 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 415) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 415) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 415) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 415) (("|") shift)) 6737 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 422) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 422) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 422) (("|") shift)) 6738 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 417) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 417) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 417) (("|") shift)) 6740 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 406) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 406) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 406) (("|") shift)) 6743 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) ((";") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("{") reduce 568) (("|") reduce 568)) 6746 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) ((",") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) ((";") reduce 564) (("<") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_DOUBLE_ARROW) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_POW) reduce 564) ((T_SL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) (("[") reduce 564) (("^") reduce 564) (("{") reduce 564) (("|") reduce 564)) 6748 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) ((",") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) ((";") reduce 563) (("<") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_DOUBLE_ARROW) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_POW) reduce 563) ((T_SL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) (("[") reduce 563) (("^") reduce 563) (("{") reduce 563) (("|") reduce 563)) 6750 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) ((";") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("{") reduce 559) (("|") reduce 559)) 6751 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((";") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("{") reduce 565) (("|") reduce 565)) 6752 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) ((",") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) ((";") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_DOUBLE_ARROW) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("[") reduce 548) (("^") reduce 407) (("{") reduce 407) (("|") reduce 407)) 6756 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) ((",") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) ((";") reduce 572) (("<") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_DOUBLE_ARROW) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_POW) reduce 572) ((T_SL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) (("[") reduce 572) (("^") reduce 572) (("{") reduce 572) (("|") reduce 572)) 6758 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((";") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 6759 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((";") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("{") reduce 291) (("|") reduce 291)) 6763 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) ((";") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("[") reduce 292) (("^") reduce 292) (("{") reduce 292) (("|") reduce 292)) 6764 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) ((";") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("[") reduce 293) (("^") reduce 293) (("{") reduce 293) (("|") reduce 293)) 6765 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) ((";") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("{") reduce 504) (("|") reduce 504)) 6768 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) ((",") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) ((";") reduce 571) (("<") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_DOUBLE_ARROW) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_POW) reduce 571) ((T_SL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) (("[") reduce 571) (("^") reduce 571) (("{") reduce 571) (("|") reduce 571)) 6769 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((";") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 6770 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((";") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 6775 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) ((";") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("{") reduce 560) (("|") reduce 560)) 6776 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) ((";") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("{") reduce 582) (("|") reduce 582)) 6777 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((";") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("{") reduce 566) (("|") reduce 566)) 6778 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) ((";") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("{") reduce 584) (("|") reduce 584)) 6781 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) ((";") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("{") reduce 583) (("|") reduce 583)) 6782 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((";") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("{") reduce 561) (("|") reduce 561)) 6783 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((";") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("{") reduce 567) (("|") reduce 567)) 6784 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) ((",") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) ((";") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("[") reduce 562) (("^") reduce 562) (("{") reduce 562) (("|") reduce 562)) 6785 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((";") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 6786 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) ((";") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("{") reduce 570) (("|") reduce 570)) 6789 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) ((";") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("{") reduce 569) (("|") reduce 569)) 6790 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) ((";") reduce 504) (("<") reduce 504) (("=") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AND_EQUAL) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_COALESCE_EQUAL) reduce 504) ((T_CONCAT_EQUAL) reduce 504) ((T_DEC) reduce 504) ((T_DIV_EQUAL) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INC) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_MINUS_EQUAL) reduce 504) ((T_MOD_EQUAL) reduce 504) ((T_MUL_EQUAL) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_OR_EQUAL) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_PLUS_EQUAL) reduce 504) ((T_POW) reduce 504) ((T_POW_EQUAL) reduce 504) ((T_SL) reduce 504) ((T_SL_EQUAL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) ((T_SR_EQUAL) reduce 504) ((T_XOR_EQUAL) reduce 504) (("[") reduce 504) (("^") reduce 504) (("{") reduce 504) (("|") reduce 504)) 6801 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) ((",") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) ((";") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PIPE) reduce 494) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("{") reduce 494) (("|") reduce 494)) 6802 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) ((";") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_PIPE) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("^") reduce 489) (("{") reduce 489) (("|") reduce 489)) 6812 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 494) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 494) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 494) (("|") shift)) 6813 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) ((";") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_PIPE) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("^") reduce 490) (("{") reduce 490) (("|") reduce 490)) 6815 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 481) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 481) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 481) (("|") shift)) 6817 ((("%") reduce 541) (("(") reduce 579) (("*") reduce 541) (("+") reduce 541) ((",") reduce 541) (("-") reduce 541) ((".") reduce 541) (("/") reduce 541) ((";") reduce 541) (("<") reduce 541) ((">") reduce 541) (("?") reduce 541) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_BOOLEAN_AND) reduce 541) ((T_BOOLEAN_OR) reduce 541) ((T_COALESCE) reduce 541) ((T_DOUBLE_ARROW) reduce 541) ((T_INSTANCEOF) reduce 541) ((T_IS_EQUAL) reduce 541) ((T_IS_GREATER_OR_EQUAL) reduce 541) ((T_IS_IDENTICAL) reduce 541) ((T_IS_NOT_EQUAL) reduce 541) ((T_IS_NOT_IDENTICAL) reduce 541) ((T_IS_SMALLER_OR_EQUAL) reduce 541) ((T_LOGICAL_AND) reduce 541) ((T_LOGICAL_OR) reduce 541) ((T_LOGICAL_XOR) reduce 541) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 541) ((T_OBJECT_OPERATOR) reduce 541) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 541) ((T_PIPE) reduce 541) ((T_POW) reduce 541) ((T_SL) reduce 541) ((T_SPACESHIP) reduce 541) ((T_SR) reduce 541) (("[") reduce 541) (("^") reduce 541) (("{") reduce 541) (("|") reduce 541)) 6819 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) ((",") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) ((";") reduce 571) (("<") reduce 571) (("=") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AND_EQUAL) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_COALESCE_EQUAL) reduce 571) ((T_CONCAT_EQUAL) reduce 571) ((T_DEC) reduce 571) ((T_DIV_EQUAL) reduce 571) ((T_DOUBLE_ARROW) reduce 571) ((T_INC) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_MINUS_EQUAL) reduce 571) ((T_MOD_EQUAL) reduce 571) ((T_MUL_EQUAL) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_OR_EQUAL) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_PLUS_EQUAL) reduce 571) ((T_POW) reduce 571) ((T_POW_EQUAL) reduce 571) ((T_SL) reduce 571) ((T_SL_EQUAL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) ((T_SR_EQUAL) reduce 571) ((T_XOR_EQUAL) reduce 571) (("[") reduce 571) (("^") reduce 571) (("{") reduce 571) (("|") reduce 571)) 6822 ((("%") reduce 543) (("(") reduce 580) (("*") reduce 543) (("+") reduce 543) ((",") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) ((";") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PIPE) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("{") reduce 543) (("|") reduce 543)) 6823 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((";") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 6824 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((";") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 6826 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) ((",") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) ((";") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_DOUBLE_ARROW) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_PIPE) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("^") reduce 485) (("{") reduce 485) (("|") reduce 485)) 6827 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) ((";") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_PIPE) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("^") reduce 487) (("{") reduce 487) (("|") reduce 487)) 6832 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) ((";") reduce 560) (("<") reduce 560) (("=") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AND_EQUAL) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_COALESCE_EQUAL) reduce 560) ((T_CONCAT_EQUAL) reduce 560) ((T_DEC) reduce 560) ((T_DIV_EQUAL) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INC) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_MINUS_EQUAL) reduce 560) ((T_MOD_EQUAL) reduce 560) ((T_MUL_EQUAL) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_OR_EQUAL) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_PLUS_EQUAL) reduce 560) ((T_POW) reduce 560) ((T_POW_EQUAL) reduce 560) ((T_SL) reduce 560) ((T_SL_EQUAL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) ((T_SR_EQUAL) reduce 560) ((T_XOR_EQUAL) reduce 560) (("[") reduce 560) (("^") reduce 560) (("{") reduce 560) (("|") reduce 560)) 6833 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) ((";") reduce 582) (("<") reduce 582) (("=") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AND_EQUAL) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_COALESCE_EQUAL) reduce 582) ((T_CONCAT_EQUAL) reduce 582) ((T_DEC) reduce 582) ((T_DIV_EQUAL) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INC) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_MINUS_EQUAL) reduce 582) ((T_MOD_EQUAL) reduce 582) ((T_MUL_EQUAL) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_OR_EQUAL) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_PLUS_EQUAL) reduce 582) ((T_POW) reduce 582) ((T_POW_EQUAL) reduce 582) ((T_SL) reduce 582) ((T_SL_EQUAL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) ((T_SR_EQUAL) reduce 582) ((T_XOR_EQUAL) reduce 582) (("[") reduce 582) (("^") reduce 582) (("{") reduce 582) (("|") reduce 582)) 6834 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((";") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("{") reduce 566) (("|") reduce 566)) 6835 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) ((";") reduce 584) (("<") reduce 584) (("=") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AND_EQUAL) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_COALESCE_EQUAL) reduce 584) ((T_CONCAT_EQUAL) reduce 584) ((T_DEC) reduce 584) ((T_DIV_EQUAL) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INC) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_MINUS_EQUAL) reduce 584) ((T_MOD_EQUAL) reduce 584) ((T_MUL_EQUAL) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_OR_EQUAL) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_PLUS_EQUAL) reduce 584) ((T_POW) reduce 584) ((T_POW_EQUAL) reduce 584) ((T_SL) reduce 584) ((T_SL_EQUAL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) ((T_SR_EQUAL) reduce 584) ((T_XOR_EQUAL) reduce 584) (("[") reduce 584) (("^") reduce 584) (("{") reduce 584) (("|") reduce 584)) 6838 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) ((";") reduce 583) (("<") reduce 583) (("=") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AND_EQUAL) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_COALESCE_EQUAL) reduce 583) ((T_CONCAT_EQUAL) reduce 583) ((T_DEC) reduce 583) ((T_DIV_EQUAL) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INC) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_MINUS_EQUAL) reduce 583) ((T_MOD_EQUAL) reduce 583) ((T_MUL_EQUAL) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_OR_EQUAL) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_PLUS_EQUAL) reduce 583) ((T_POW) reduce 583) ((T_POW_EQUAL) reduce 583) ((T_SL) reduce 583) ((T_SL_EQUAL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) ((T_SR_EQUAL) reduce 583) ((T_XOR_EQUAL) reduce 583) (("[") reduce 583) (("^") reduce 583) (("{") reduce 583) (("|") reduce 583)) 6839 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((";") reduce 561) (("<") reduce 561) (("=") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AND_EQUAL) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_COALESCE_EQUAL) reduce 561) ((T_CONCAT_EQUAL) reduce 561) ((T_DEC) reduce 561) ((T_DIV_EQUAL) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INC) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_MINUS_EQUAL) reduce 561) ((T_MOD_EQUAL) reduce 561) ((T_MUL_EQUAL) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_OR_EQUAL) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_PLUS_EQUAL) reduce 561) ((T_POW) reduce 561) ((T_POW_EQUAL) reduce 561) ((T_SL) reduce 561) ((T_SL_EQUAL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) ((T_SR_EQUAL) reduce 561) ((T_XOR_EQUAL) reduce 561) (("[") reduce 561) (("^") reduce 561) (("{") reduce 561) (("|") reduce 561)) 6840 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((";") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("{") reduce 567) (("|") reduce 567)) 6841 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) ((",") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) ((";") reduce 562) (("<") reduce 562) (("=") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AND_EQUAL) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_COALESCE_EQUAL) reduce 562) ((T_CONCAT_EQUAL) reduce 562) ((T_DEC) reduce 562) ((T_DIV_EQUAL) reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_INC) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_MINUS_EQUAL) reduce 562) ((T_MOD_EQUAL) reduce 562) ((T_MUL_EQUAL) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_OR_EQUAL) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_PLUS_EQUAL) reduce 562) ((T_POW) reduce 562) ((T_POW_EQUAL) reduce 562) ((T_SL) reduce 562) ((T_SL_EQUAL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) ((T_SR_EQUAL) reduce 562) ((T_XOR_EQUAL) reduce 562) (("[") reduce 562) (("^") reduce 562) (("{") reduce 562) (("|") reduce 562)) 6843 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) ((",") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) ((";") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_PIPE) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("{") reduce 477) (("|") reduce 477)) 6845 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) ((";") reduce 520) (("<") reduce 520) (("=") shift) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_PIPE) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 6847 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 405) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 405) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 405) (("|") shift)) 6848 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 482) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 482) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 482) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 482) (("|") shift)) 6851 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 481) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 481) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 481) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 481) (("|") shift)) 6852 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) ((",") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) ((";") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_DOUBLE_ARROW) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_PIPE) reduce 473) ((T_POW) shift) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("{") reduce 473) (("|") reduce 473)) 6853 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 483) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 483) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 483) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 483) (("|") shift)) 6854 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) ((",") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) ((";") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_DOUBLE_ARROW) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_PIPE) reduce 469) ((T_POW) shift) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("{") reduce 469) (("|") reduce 469)) 6855 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) ((",") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) ((";") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_DOUBLE_ARROW) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_PIPE) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("^") reduce 486) (("{") reduce 486) (("|") reduce 486)) 6857 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) ((";") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_PIPE) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 6859 ((("%") reduce 527) (("*") reduce 527) (("+") reduce 527) ((",") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) ((";") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_DOUBLE_ARROW) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_PIPE) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("^") reduce 527) (("{") reduce 527) (("|") reduce 527)) 6860 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) ((";") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_PIPE) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 6861 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 619) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 619) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 619) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 619) (("|") shift)) 6862 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 618) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 618) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 618) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 618) (("|") shift)) 6863 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((";") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 6864 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 478) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 478) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 478) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 478) (("|") shift)) 6865 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) ((",") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) ((";") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_DOUBLE_ARROW) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_PIPE) reduce 471) ((T_POW) shift) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("{") reduce 471) (("|") reduce 471)) 6867 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 6868 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 6869 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("{") reduce 95) (("|") reduce 95)) 6870 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((";") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 6871 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 6872 ((("%") reduce 400) (("(") reduce 400) (("*") reduce 400) (("+") reduce 400) ((",") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) ((";") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_DOUBLE_ARROW) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 400) ((T_OBJECT_OPERATOR) reduce 400) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 400) ((T_PIPE) reduce 400) ((T_POW) reduce 400) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("[") reduce 400) (("^") reduce 400) (("{") reduce 400) (("|") reduce 400)) 6875 ((("%") reduce 511) (("(") reduce 511) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((";") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 6876 ((("%") reduce 402) (("(") shift) (("*") reduce 402) (("+") reduce 402) ((",") reduce 402) (("-") reduce 402) ((".") reduce 402) (("/") reduce 402) ((";") reduce 402) (("<") reduce 402) ((">") reduce 402) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_DOUBLE_ARROW) reduce 402) ((T_INSTANCEOF) reduce 402) ((T_IS_EQUAL) reduce 402) ((T_IS_GREATER_OR_EQUAL) reduce 402) ((T_IS_IDENTICAL) reduce 402) ((T_IS_NOT_EQUAL) reduce 402) ((T_IS_NOT_IDENTICAL) reduce 402) ((T_IS_SMALLER_OR_EQUAL) reduce 402) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_PIPE) reduce 402) ((T_POW) reduce 402) ((T_SL) reduce 402) ((T_SPACESHIP) reduce 402) ((T_SR) reduce 402) (("^") reduce 402) (("{") reduce 402) (("|") reduce 402)) 6877 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((";") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 6878 ((("%") reduce 512) (("(") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((";") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 6879 ((("%") reduce 573) (("(") reduce 573) (("*") reduce 573) (("+") reduce 573) ((",") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) ((";") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_DOUBLE_ARROW) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("{") reduce 573) (("|") reduce 573)) 6885 ((("%") reduce 574) (("(") reduce 574) (("*") reduce 574) (("+") reduce 574) ((",") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) ((";") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_DOUBLE_ARROW) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("{") reduce 574) (("|") reduce 574)) 6886 ((("%") reduce 578) (("(") reduce 578) (("*") reduce 578) (("+") reduce 578) ((",") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) ((";") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_DOUBLE_ARROW) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("{") reduce 578) (("|") reduce 578)) 6887 ((("%") reduce 575) (("(") reduce 575) (("*") reduce 575) (("+") reduce 575) ((",") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) ((";") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_DOUBLE_ARROW) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("{") reduce 575) (("|") reduce 575)) 6888 ((("%") reduce 576) (("(") reduce 576) (("*") reduce 576) (("+") reduce 576) ((",") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) ((";") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_DOUBLE_ARROW) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("{") reduce 576) (("|") reduce 576)) 6889 ((("%") reduce 399) (("(") reduce 399) (("*") reduce 399) (("+") reduce 399) ((",") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) ((";") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_DOUBLE_ARROW) reduce 399) ((T_INSTANCEOF) reduce 399) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 399) ((T_OBJECT_OPERATOR) reduce 399) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 399) ((T_PIPE) reduce 399) ((T_POW) reduce 399) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("[") reduce 399) (("^") reduce 399) (("{") reduce 399) (("|") reduce 399)) 6891 ((("%") reduce 577) (("(") reduce 577) (("*") reduce 577) (("+") reduce 577) ((",") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) ((";") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_DOUBLE_ARROW) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("{") reduce 577) (("|") reduce 577)) 6892 ((("%") reduce 401) (("(") reduce 401) (("*") reduce 401) (("+") reduce 401) ((",") reduce 401) (("-") reduce 401) ((".") reduce 401) (("/") reduce 401) ((";") reduce 401) (("<") reduce 401) ((">") reduce 401) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_DOUBLE_ARROW) reduce 401) ((T_INSTANCEOF) reduce 401) ((T_IS_EQUAL) reduce 401) ((T_IS_GREATER_OR_EQUAL) reduce 401) ((T_IS_IDENTICAL) reduce 401) ((T_IS_NOT_EQUAL) reduce 401) ((T_IS_NOT_IDENTICAL) reduce 401) ((T_IS_SMALLER_OR_EQUAL) reduce 401) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 401) ((T_OBJECT_OPERATOR) reduce 401) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 401) ((T_PIPE) reduce 401) ((T_POW) reduce 401) ((T_SL) reduce 401) ((T_SPACESHIP) reduce 401) ((T_SR) reduce 401) (("[") reduce 401) (("^") reduce 401) (("{") reduce 401) (("|") reduce 401)) 6900 ((("%") reduce 398) (("(") reduce 398) (("*") reduce 398) (("+") reduce 398) ((",") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) ((";") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_DOUBLE_ARROW) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 398) ((T_OBJECT_OPERATOR) reduce 398) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 398) ((T_PIPE) reduce 398) ((T_POW) reduce 398) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("[") reduce 398) (("^") reduce 398) (("{") reduce 398) (("|") reduce 398)) 6902 ((("%") reduce 513) (("(") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((";") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 6908 ((("%") reduce 233) (("*") reduce 233) (("+") reduce 233) ((",") reduce 233) (("-") reduce 233) ((".") reduce 233) (("/") reduce 233) ((";") reduce 233) (("<") reduce 233) ((">") reduce 233) (("?") reduce 233) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_BOOLEAN_AND) reduce 233) ((T_BOOLEAN_OR) reduce 233) ((T_COALESCE) reduce 233) ((T_DOUBLE_ARROW) reduce 233) ((T_INSTANCEOF) reduce 233) ((T_IS_EQUAL) reduce 233) ((T_IS_GREATER_OR_EQUAL) reduce 233) ((T_IS_IDENTICAL) reduce 233) ((T_IS_NOT_EQUAL) reduce 233) ((T_IS_NOT_IDENTICAL) reduce 233) ((T_IS_SMALLER_OR_EQUAL) reduce 233) ((T_LOGICAL_AND) reduce 233) ((T_LOGICAL_OR) reduce 233) ((T_LOGICAL_XOR) reduce 233) ((T_PIPE) reduce 233) ((T_POW) reduce 233) ((T_SL) reduce 233) ((T_SPACESHIP) reduce 233) ((T_SR) reduce 233) (("^") reduce 233) (("{") reduce 233) (("|") reduce 233)) 6913 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 404) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 404) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 404) (("|") shift)) 6917 ((("%") reduce 613) (("*") reduce 613) (("+") reduce 613) ((",") reduce 613) (("-") reduce 613) ((".") reduce 613) (("/") reduce 613) ((";") reduce 613) (("<") reduce 613) ((">") reduce 613) (("?") reduce 613) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_BOOLEAN_AND) reduce 613) ((T_BOOLEAN_OR) reduce 613) ((T_COALESCE) reduce 613) ((T_DOUBLE_ARROW) reduce 613) ((T_INSTANCEOF) reduce 613) ((T_IS_EQUAL) reduce 613) ((T_IS_GREATER_OR_EQUAL) reduce 613) ((T_IS_IDENTICAL) reduce 613) ((T_IS_NOT_EQUAL) reduce 613) ((T_IS_NOT_IDENTICAL) reduce 613) ((T_IS_SMALLER_OR_EQUAL) reduce 613) ((T_LOGICAL_AND) reduce 613) ((T_LOGICAL_OR) reduce 613) ((T_LOGICAL_XOR) reduce 613) ((T_PIPE) reduce 613) ((T_POW) reduce 613) ((T_SL) reduce 613) ((T_SPACESHIP) reduce 613) ((T_SR) reduce 613) (("^") reduce 613) (("{") reduce 613) (("|") reduce 613)) 6918 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) ((",") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) ((";") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_DOUBLE_ARROW) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_PIPE) reduce 467) ((T_POW) shift) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("{") reduce 467) (("|") reduce 467)) 6919 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 616) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 616) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 616) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 616) (("|") shift)) 6920 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 615) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 615) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 615) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 615) (("|") shift)) 6921 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) ((",") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) ((";") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_DOUBLE_ARROW) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("[") reduce 548) (("^") reduce 424) (("{") reduce 424) (("|") reduce 424)) 6923 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) ((";") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_PIPE) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 6924 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) ((",") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) ((";") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_DOUBLE_ARROW) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_PIPE) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("{") reduce 474) (("|") reduce 474)) 6925 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((";") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("{") reduce 291) (("|") reduce 291)) 6929 ((("%") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) ((";") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("^") reduce 292) (("{") reduce 292) (("|") reduce 292)) 6930 ((("%") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) ((";") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("^") reduce 293) (("{") reduce 293) (("|") reduce 293)) 6933 ((("%") reduce 617) (("*") reduce 617) (("+") reduce 617) ((",") reduce 617) (("-") reduce 617) ((".") reduce 617) (("/") reduce 617) ((";") reduce 617) (("<") reduce 617) ((">") reduce 617) (("?") reduce 617) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_BOOLEAN_AND) reduce 617) ((T_BOOLEAN_OR) reduce 617) ((T_COALESCE) reduce 617) ((T_DOUBLE_ARROW) reduce 617) ((T_INSTANCEOF) reduce 617) ((T_IS_EQUAL) reduce 617) ((T_IS_GREATER_OR_EQUAL) reduce 617) ((T_IS_IDENTICAL) reduce 617) ((T_IS_NOT_EQUAL) reduce 617) ((T_IS_NOT_IDENTICAL) reduce 617) ((T_IS_SMALLER_OR_EQUAL) reduce 617) ((T_LOGICAL_AND) reduce 617) ((T_LOGICAL_OR) reduce 617) ((T_LOGICAL_XOR) reduce 617) ((T_PIPE) reduce 617) ((T_POW) reduce 617) ((T_SL) reduce 617) ((T_SPACESHIP) reduce 617) ((T_SR) reduce 617) (("^") reduce 617) (("{") reduce 617) (("|") reduce 617)) 6936 ((("%") reduce 614) (("*") reduce 614) (("+") reduce 614) ((",") reduce 614) (("-") reduce 614) ((".") reduce 614) (("/") reduce 614) ((";") reduce 614) (("<") reduce 614) ((">") reduce 614) (("?") reduce 614) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_BOOLEAN_AND) reduce 614) ((T_BOOLEAN_OR) reduce 614) ((T_COALESCE) reduce 614) ((T_DOUBLE_ARROW) reduce 614) ((T_INSTANCEOF) reduce 614) ((T_IS_EQUAL) reduce 614) ((T_IS_GREATER_OR_EQUAL) reduce 614) ((T_IS_IDENTICAL) reduce 614) ((T_IS_NOT_EQUAL) reduce 614) ((T_IS_NOT_IDENTICAL) reduce 614) ((T_IS_SMALLER_OR_EQUAL) reduce 614) ((T_LOGICAL_AND) reduce 614) ((T_LOGICAL_OR) reduce 614) ((T_LOGICAL_XOR) reduce 614) ((T_PIPE) reduce 614) ((T_POW) reduce 614) ((T_SL) reduce 614) ((T_SPACESHIP) reduce 614) ((T_SR) reduce 614) (("^") reduce 614) (("{") reduce 614) (("|") reduce 614)) 6937 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) ((",") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) ((";") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_DOUBLE_ARROW) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_PIPE) reduce 468) ((T_POW) shift) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("{") reduce 468) (("|") reduce 468)) 6938 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) ((",") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) ((";") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_DOUBLE_ARROW) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("[") reduce 548) (("^") reduce 426) (("{") reduce 426) (("|") reduce 426)) 6940 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) ((",") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) ((";") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_DOUBLE_ARROW) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_PIPE) reduce 408) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("^") reduce 408) (("{") reduce 408) (("|") reduce 408)) 6941 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) ((",") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) ((";") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_DOUBLE_ARROW) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_PIPE) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("{") reduce 409) (("|") reduce 409)) 6942 ((("%") reduce 296) (("*") reduce 296) (("+") reduce 296) ((",") reduce 296) (("-") reduce 296) ((".") reduce 296) (("/") reduce 296) ((";") reduce 296) (("<") reduce 296) ((">") reduce 296) (("?") reduce 296) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_BOOLEAN_AND) reduce 296) ((T_BOOLEAN_OR) reduce 296) ((T_COALESCE) reduce 296) ((T_DOUBLE_ARROW) reduce 296) ((T_INSTANCEOF) reduce 296) ((T_IS_EQUAL) reduce 296) ((T_IS_GREATER_OR_EQUAL) reduce 296) ((T_IS_IDENTICAL) reduce 296) ((T_IS_NOT_EQUAL) reduce 296) ((T_IS_NOT_IDENTICAL) reduce 296) ((T_IS_SMALLER_OR_EQUAL) reduce 296) ((T_LOGICAL_AND) reduce 296) ((T_LOGICAL_OR) reduce 296) ((T_LOGICAL_XOR) reduce 296) ((T_PIPE) reduce 296) ((T_POW) reduce 296) ((T_SL) reduce 296) ((T_SPACESHIP) reduce 296) ((T_SR) reduce 296) (("^") reduce 296) (("{") reduce 296) (("|") reduce 296)) 6947 ((("%") reduce 297) (("*") reduce 297) (("+") reduce 297) ((",") reduce 297) (("-") reduce 297) ((".") reduce 297) (("/") reduce 297) ((";") reduce 297) (("<") reduce 297) ((">") reduce 297) (("?") reduce 297) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_BOOLEAN_AND) reduce 297) ((T_BOOLEAN_OR) reduce 297) ((T_COALESCE) reduce 297) ((T_DOUBLE_ARROW) reduce 297) ((T_INSTANCEOF) reduce 297) ((T_IS_EQUAL) reduce 297) ((T_IS_GREATER_OR_EQUAL) reduce 297) ((T_IS_IDENTICAL) reduce 297) ((T_IS_NOT_EQUAL) reduce 297) ((T_IS_NOT_IDENTICAL) reduce 297) ((T_IS_SMALLER_OR_EQUAL) reduce 297) ((T_LOGICAL_AND) reduce 297) ((T_LOGICAL_OR) reduce 297) ((T_LOGICAL_XOR) reduce 297) ((T_PIPE) reduce 297) ((T_POW) reduce 297) ((T_SL) reduce 297) ((T_SPACESHIP) reduce 297) ((T_SR) reduce 297) (("^") reduce 297) (("{") reduce 297) (("|") reduce 297)) 6948 ((("%") reduce 460) (("(") reduce 556) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) ((";") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PIPE) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("[") reduce 549) (("^") reduce 460) (("{") reduce 460) (("|") reduce 460)) 6950 ((("%") reduce 298) (("*") reduce 298) (("+") reduce 298) ((",") reduce 298) (("-") reduce 298) ((".") reduce 298) (("/") reduce 298) ((";") reduce 298) (("<") reduce 298) ((">") reduce 298) (("?") reduce 298) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_BOOLEAN_AND) reduce 298) ((T_BOOLEAN_OR) reduce 298) ((T_COALESCE) reduce 298) ((T_DOUBLE_ARROW) reduce 298) ((T_INSTANCEOF) reduce 298) ((T_IS_EQUAL) reduce 298) ((T_IS_GREATER_OR_EQUAL) reduce 298) ((T_IS_IDENTICAL) reduce 298) ((T_IS_NOT_EQUAL) reduce 298) ((T_IS_NOT_IDENTICAL) reduce 298) ((T_IS_SMALLER_OR_EQUAL) reduce 298) ((T_LOGICAL_AND) reduce 298) ((T_LOGICAL_OR) reduce 298) ((T_LOGICAL_XOR) reduce 298) ((T_PIPE) reduce 298) ((T_POW) reduce 298) ((T_SL) reduce 298) ((T_SPACESHIP) reduce 298) ((T_SR) reduce 298) (("^") reduce 298) (("{") reduce 298) (("|") reduce 298)) 6951 ((("%") reduce 299) (("*") reduce 299) (("+") reduce 299) ((",") reduce 299) (("-") reduce 299) ((".") reduce 299) (("/") reduce 299) ((";") reduce 299) (("<") reduce 299) ((">") reduce 299) (("?") reduce 299) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_BOOLEAN_AND) reduce 299) ((T_BOOLEAN_OR) reduce 299) ((T_COALESCE) reduce 299) ((T_DOUBLE_ARROW) reduce 299) ((T_INSTANCEOF) reduce 299) ((T_IS_EQUAL) reduce 299) ((T_IS_GREATER_OR_EQUAL) reduce 299) ((T_IS_IDENTICAL) reduce 299) ((T_IS_NOT_EQUAL) reduce 299) ((T_IS_NOT_IDENTICAL) reduce 299) ((T_IS_SMALLER_OR_EQUAL) reduce 299) ((T_LOGICAL_AND) reduce 299) ((T_LOGICAL_OR) reduce 299) ((T_LOGICAL_XOR) reduce 299) ((T_PIPE) reduce 299) ((T_POW) reduce 299) ((T_SL) reduce 299) ((T_SPACESHIP) reduce 299) ((T_SR) reduce 299) (("^") reduce 299) (("{") reduce 299) (("|") reduce 299)) 6952 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) ((",") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) ((";") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_DOUBLE_ARROW) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_PIPE) reduce 472) ((T_POW) shift) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("{") reduce 472) (("|") reduce 472)) 6953 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) ((",") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) ((";") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_DOUBLE_ARROW) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_PIPE) reduce 470) ((T_POW) shift) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("{") reduce 470) (("|") reduce 470)) 6956 ((("%") reduce 519) (("(") reduce 519) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) ((";") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PIPE) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 6957 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) ((",") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) ((";") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_DOUBLE_ARROW) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_PIPE) reduce 475) ((T_POW) shift) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("{") reduce 475) (("|") reduce 475)) 6958 ((("%") reduce 446) (("*") reduce 446) (("+") reduce 446) ((",") reduce 446) (("-") reduce 446) ((".") reduce 446) (("/") reduce 446) ((";") reduce 446) (("<") reduce 446) ((">") reduce 446) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_BOOLEAN_AND) reduce 446) ((T_BOOLEAN_OR) reduce 446) ((T_COALESCE) reduce 446) ((T_DOUBLE_ARROW) reduce 446) ((T_INSTANCEOF) reduce 446) ((T_IS_EQUAL) reduce 446) ((T_IS_GREATER_OR_EQUAL) reduce 446) ((T_IS_IDENTICAL) reduce 446) ((T_IS_NOT_EQUAL) reduce 446) ((T_IS_NOT_IDENTICAL) reduce 446) ((T_IS_SMALLER_OR_EQUAL) reduce 446) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_PIPE) reduce 446) ((T_POW) shift) ((T_SL) reduce 446) ((T_SPACESHIP) reduce 446) ((T_SR) reduce 446) (("^") reduce 446) (("{") reduce 446) (("|") reduce 446)) 6959 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) ((",") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) ((";") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_DOUBLE_ARROW) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_PIPE) reduce 445) ((T_POW) shift) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("{") reduce 445) (("|") reduce 445)) 6961 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) ((";") reduce 570) (("<") reduce 570) (("=") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AND_EQUAL) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_COALESCE_EQUAL) reduce 570) ((T_CONCAT_EQUAL) reduce 570) ((T_DEC) reduce 570) ((T_DIV_EQUAL) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INC) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_MINUS_EQUAL) reduce 570) ((T_MOD_EQUAL) reduce 570) ((T_MUL_EQUAL) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_OR_EQUAL) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_PLUS_EQUAL) reduce 570) ((T_POW) reduce 570) ((T_POW_EQUAL) reduce 570) ((T_SL) reduce 570) ((T_SL_EQUAL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) ((T_SR_EQUAL) reduce 570) ((T_XOR_EQUAL) reduce 570) (("[") reduce 570) (("^") reduce 570) (("{") reduce 570) (("|") reduce 570)) 6964 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) ((";") reduce 569) (("<") reduce 569) (("=") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AND_EQUAL) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_COALESCE_EQUAL) reduce 569) ((T_CONCAT_EQUAL) reduce 569) ((T_DEC) reduce 569) ((T_DIV_EQUAL) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INC) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_MINUS_EQUAL) reduce 569) ((T_MOD_EQUAL) reduce 569) ((T_MUL_EQUAL) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_OR_EQUAL) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_PLUS_EQUAL) reduce 569) ((T_POW) reduce 569) ((T_POW_EQUAL) reduce 569) ((T_SL) reduce 569) ((T_SL_EQUAL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) ((T_SR_EQUAL) reduce 569) ((T_XOR_EQUAL) reduce 569) (("[") reduce 569) (("^") reduce 569) (("{") reduce 569) (("|") reduce 569)) 6966 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) ((";") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_PIPE) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 6967 ((("%") reduce 447) (("*") reduce 447) (("+") reduce 447) ((",") reduce 447) (("-") reduce 447) ((".") reduce 447) (("/") reduce 447) ((";") reduce 447) (("<") reduce 447) ((">") reduce 447) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_BOOLEAN_AND) reduce 447) ((T_BOOLEAN_OR) reduce 447) ((T_COALESCE) reduce 447) ((T_DOUBLE_ARROW) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 447) ((T_IS_GREATER_OR_EQUAL) reduce 447) ((T_IS_IDENTICAL) reduce 447) ((T_IS_NOT_EQUAL) reduce 447) ((T_IS_NOT_IDENTICAL) reduce 447) ((T_IS_SMALLER_OR_EQUAL) reduce 447) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_PIPE) reduce 447) ((T_POW) shift) ((T_SL) reduce 447) ((T_SPACESHIP) reduce 447) ((T_SR) reduce 447) (("^") reduce 447) (("{") reduce 447) (("|") reduce 447)) 6968 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) ((",") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) ((";") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_PIPE) reduce 473) ((T_POW) shift) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("{") reduce 473) (("|") reduce 473)) 6969 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 483) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 483) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 483) (("|") shift)) 6970 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) ((",") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) ((";") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_PIPE) reduce 469) ((T_POW) shift) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("{") reduce 469) (("|") reduce 469)) 6971 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) ((",") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) ((";") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_PIPE) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("^") reduce 486) (("{") reduce 486) (("|") reduce 486)) 6973 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) ((";") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_PIPE) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 6975 ((("%") reduce 527) (("*") reduce 527) (("+") reduce 527) ((",") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) ((";") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_PIPE) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("^") reduce 527) (("{") reduce 527) (("|") reduce 527)) 6976 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) ((";") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_PIPE) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 6977 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 619) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 619) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 619) (("|") shift)) 6978 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 618) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 618) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 618) (("|") shift)) 6979 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((";") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 6980 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 478) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 478) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 478) (("|") shift)) 6981 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) ((",") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) ((";") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_PIPE) reduce 471) ((T_POW) shift) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("{") reduce 471) (("|") reduce 471)) 6983 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 6984 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 6985 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("{") reduce 95) (("|") reduce 95)) 6986 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((";") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 6987 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 6988 ((("%") reduce 400) (("(") reduce 400) (("*") reduce 400) (("+") reduce 400) ((",") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) ((";") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 400) ((T_OBJECT_OPERATOR) reduce 400) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 400) ((T_PIPE) reduce 400) ((T_POW) reduce 400) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("[") reduce 400) (("^") reduce 400) (("{") reduce 400) (("|") reduce 400)) 6991 ((("%") reduce 511) (("(") reduce 511) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((";") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 6992 ((("%") reduce 402) (("(") shift) (("*") reduce 402) (("+") reduce 402) ((",") reduce 402) (("-") reduce 402) ((".") reduce 402) (("/") reduce 402) ((";") reduce 402) (("<") reduce 402) ((">") reduce 402) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_INSTANCEOF) reduce 402) ((T_IS_EQUAL) reduce 402) ((T_IS_GREATER_OR_EQUAL) reduce 402) ((T_IS_IDENTICAL) reduce 402) ((T_IS_NOT_EQUAL) reduce 402) ((T_IS_NOT_IDENTICAL) reduce 402) ((T_IS_SMALLER_OR_EQUAL) reduce 402) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_PIPE) reduce 402) ((T_POW) reduce 402) ((T_SL) reduce 402) ((T_SPACESHIP) reduce 402) ((T_SR) reduce 402) (("^") reduce 402) (("{") reduce 402) (("|") reduce 402)) 6993 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((";") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 6994 ((("%") reduce 512) (("(") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((";") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 6995 ((("%") reduce 573) (("(") reduce 573) (("*") reduce 573) (("+") reduce 573) ((",") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) ((";") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("{") reduce 573) (("|") reduce 573)) 7001 ((("%") reduce 574) (("(") reduce 574) (("*") reduce 574) (("+") reduce 574) ((",") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) ((";") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("{") reduce 574) (("|") reduce 574)) 7002 ((("%") reduce 578) (("(") reduce 578) (("*") reduce 578) (("+") reduce 578) ((",") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) ((";") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("{") reduce 578) (("|") reduce 578)) 7003 ((("%") reduce 575) (("(") reduce 575) (("*") reduce 575) (("+") reduce 575) ((",") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) ((";") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("{") reduce 575) (("|") reduce 575)) 7004 ((("%") reduce 576) (("(") reduce 576) (("*") reduce 576) (("+") reduce 576) ((",") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) ((";") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("{") reduce 576) (("|") reduce 576)) 7005 ((("%") reduce 399) (("(") reduce 399) (("*") reduce 399) (("+") reduce 399) ((",") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) ((";") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_INSTANCEOF) reduce 399) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 399) ((T_OBJECT_OPERATOR) reduce 399) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 399) ((T_PIPE) reduce 399) ((T_POW) reduce 399) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("[") reduce 399) (("^") reduce 399) (("{") reduce 399) (("|") reduce 399)) 7007 ((("%") reduce 577) (("(") reduce 577) (("*") reduce 577) (("+") reduce 577) ((",") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) ((";") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("{") reduce 577) (("|") reduce 577)) 7008 ((("%") reduce 401) (("(") reduce 401) (("*") reduce 401) (("+") reduce 401) ((",") reduce 401) (("-") reduce 401) ((".") reduce 401) (("/") reduce 401) ((";") reduce 401) (("<") reduce 401) ((">") reduce 401) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_INSTANCEOF) reduce 401) ((T_IS_EQUAL) reduce 401) ((T_IS_GREATER_OR_EQUAL) reduce 401) ((T_IS_IDENTICAL) reduce 401) ((T_IS_NOT_EQUAL) reduce 401) ((T_IS_NOT_IDENTICAL) reduce 401) ((T_IS_SMALLER_OR_EQUAL) reduce 401) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 401) ((T_OBJECT_OPERATOR) reduce 401) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 401) ((T_PIPE) reduce 401) ((T_POW) reduce 401) ((T_SL) reduce 401) ((T_SPACESHIP) reduce 401) ((T_SR) reduce 401) (("[") reduce 401) (("^") reduce 401) (("{") reduce 401) (("|") reduce 401)) 7016 ((("%") reduce 398) (("(") reduce 398) (("*") reduce 398) (("+") reduce 398) ((",") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) ((";") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 398) ((T_OBJECT_OPERATOR) reduce 398) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 398) ((T_PIPE) reduce 398) ((T_POW) reduce 398) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("[") reduce 398) (("^") reduce 398) (("{") reduce 398) (("|") reduce 398)) 7018 ((("%") reduce 513) (("(") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((";") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 7024 ((("%") reduce 233) (("*") reduce 233) (("+") reduce 233) ((",") reduce 233) (("-") reduce 233) ((".") reduce 233) (("/") reduce 233) ((";") reduce 233) (("<") reduce 233) ((">") reduce 233) (("?") reduce 233) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_BOOLEAN_AND) reduce 233) ((T_BOOLEAN_OR) reduce 233) ((T_COALESCE) reduce 233) ((T_INSTANCEOF) reduce 233) ((T_IS_EQUAL) reduce 233) ((T_IS_GREATER_OR_EQUAL) reduce 233) ((T_IS_IDENTICAL) reduce 233) ((T_IS_NOT_EQUAL) reduce 233) ((T_IS_NOT_IDENTICAL) reduce 233) ((T_IS_SMALLER_OR_EQUAL) reduce 233) ((T_LOGICAL_AND) reduce 233) ((T_LOGICAL_OR) reduce 233) ((T_LOGICAL_XOR) reduce 233) ((T_PIPE) reduce 233) ((T_POW) reduce 233) ((T_SL) reduce 233) ((T_SPACESHIP) reduce 233) ((T_SR) reduce 233) (("^") reduce 233) (("{") reduce 233) (("|") reduce 233)) 7029 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 404) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 404) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 404) (("|") shift)) 7033 ((("%") reduce 613) (("*") reduce 613) (("+") reduce 613) ((",") reduce 613) (("-") reduce 613) ((".") reduce 613) (("/") reduce 613) ((";") reduce 613) (("<") reduce 613) ((">") reduce 613) (("?") reduce 613) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_BOOLEAN_AND) reduce 613) ((T_BOOLEAN_OR) reduce 613) ((T_COALESCE) reduce 613) ((T_INSTANCEOF) reduce 613) ((T_IS_EQUAL) reduce 613) ((T_IS_GREATER_OR_EQUAL) reduce 613) ((T_IS_IDENTICAL) reduce 613) ((T_IS_NOT_EQUAL) reduce 613) ((T_IS_NOT_IDENTICAL) reduce 613) ((T_IS_SMALLER_OR_EQUAL) reduce 613) ((T_LOGICAL_AND) reduce 613) ((T_LOGICAL_OR) reduce 613) ((T_LOGICAL_XOR) reduce 613) ((T_PIPE) reduce 613) ((T_POW) reduce 613) ((T_SL) reduce 613) ((T_SPACESHIP) reduce 613) ((T_SR) reduce 613) (("^") reduce 613) (("{") reduce 613) (("|") reduce 613)) 7034 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) ((",") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) ((";") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_PIPE) reduce 467) ((T_POW) shift) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("{") reduce 467) (("|") reduce 467)) 7035 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 616) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 616) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 616) (("|") shift)) 7036 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 615) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 615) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 615) (("|") shift)) 7037 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) ((",") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) ((";") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("[") reduce 548) (("^") reduce 424) (("{") reduce 424) (("|") reduce 424)) 7039 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) ((";") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_PIPE) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 7040 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) ((",") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) ((";") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_PIPE) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("{") reduce 474) (("|") reduce 474)) 7041 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((";") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("{") reduce 291) (("|") reduce 291)) 7045 ((("%") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) ((";") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("^") reduce 292) (("{") reduce 292) (("|") reduce 292)) 7046 ((("%") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) ((";") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("^") reduce 293) (("{") reduce 293) (("|") reduce 293)) 7049 ((("%") reduce 617) (("*") reduce 617) (("+") reduce 617) ((",") reduce 617) (("-") reduce 617) ((".") reduce 617) (("/") reduce 617) ((";") reduce 617) (("<") reduce 617) ((">") reduce 617) (("?") reduce 617) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_BOOLEAN_AND) reduce 617) ((T_BOOLEAN_OR) reduce 617) ((T_COALESCE) reduce 617) ((T_INSTANCEOF) reduce 617) ((T_IS_EQUAL) reduce 617) ((T_IS_GREATER_OR_EQUAL) reduce 617) ((T_IS_IDENTICAL) reduce 617) ((T_IS_NOT_EQUAL) reduce 617) ((T_IS_NOT_IDENTICAL) reduce 617) ((T_IS_SMALLER_OR_EQUAL) reduce 617) ((T_LOGICAL_AND) reduce 617) ((T_LOGICAL_OR) reduce 617) ((T_LOGICAL_XOR) reduce 617) ((T_PIPE) reduce 617) ((T_POW) reduce 617) ((T_SL) reduce 617) ((T_SPACESHIP) reduce 617) ((T_SR) reduce 617) (("^") reduce 617) (("{") reduce 617) (("|") reduce 617)) 7052 ((("%") reduce 614) (("*") reduce 614) (("+") reduce 614) ((",") reduce 614) (("-") reduce 614) ((".") reduce 614) (("/") reduce 614) ((";") reduce 614) (("<") reduce 614) ((">") reduce 614) (("?") reduce 614) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_BOOLEAN_AND) reduce 614) ((T_BOOLEAN_OR) reduce 614) ((T_COALESCE) reduce 614) ((T_INSTANCEOF) reduce 614) ((T_IS_EQUAL) reduce 614) ((T_IS_GREATER_OR_EQUAL) reduce 614) ((T_IS_IDENTICAL) reduce 614) ((T_IS_NOT_EQUAL) reduce 614) ((T_IS_NOT_IDENTICAL) reduce 614) ((T_IS_SMALLER_OR_EQUAL) reduce 614) ((T_LOGICAL_AND) reduce 614) ((T_LOGICAL_OR) reduce 614) ((T_LOGICAL_XOR) reduce 614) ((T_PIPE) reduce 614) ((T_POW) reduce 614) ((T_SL) reduce 614) ((T_SPACESHIP) reduce 614) ((T_SR) reduce 614) (("^") reduce 614) (("{") reduce 614) (("|") reduce 614)) 7053 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) ((",") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) ((";") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_PIPE) reduce 468) ((T_POW) shift) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("{") reduce 468) (("|") reduce 468)) 7054 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) ((",") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) ((";") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("[") reduce 548) (("^") reduce 426) (("{") reduce 426) (("|") reduce 426)) 7056 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) ((",") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) ((";") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_PIPE) reduce 408) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("^") reduce 408) (("{") reduce 408) (("|") reduce 408)) 7057 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) ((",") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) ((";") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_PIPE) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("{") reduce 409) (("|") reduce 409)) 7058 ((("%") reduce 296) (("*") reduce 296) (("+") reduce 296) ((",") reduce 296) (("-") reduce 296) ((".") reduce 296) (("/") reduce 296) ((";") reduce 296) (("<") reduce 296) ((">") reduce 296) (("?") reduce 296) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_BOOLEAN_AND) reduce 296) ((T_BOOLEAN_OR) reduce 296) ((T_COALESCE) reduce 296) ((T_INSTANCEOF) reduce 296) ((T_IS_EQUAL) reduce 296) ((T_IS_GREATER_OR_EQUAL) reduce 296) ((T_IS_IDENTICAL) reduce 296) ((T_IS_NOT_EQUAL) reduce 296) ((T_IS_NOT_IDENTICAL) reduce 296) ((T_IS_SMALLER_OR_EQUAL) reduce 296) ((T_LOGICAL_AND) reduce 296) ((T_LOGICAL_OR) reduce 296) ((T_LOGICAL_XOR) reduce 296) ((T_PIPE) reduce 296) ((T_POW) reduce 296) ((T_SL) reduce 296) ((T_SPACESHIP) reduce 296) ((T_SR) reduce 296) (("^") reduce 296) (("{") reduce 296) (("|") reduce 296)) 7063 ((("%") reduce 297) (("*") reduce 297) (("+") reduce 297) ((",") reduce 297) (("-") reduce 297) ((".") reduce 297) (("/") reduce 297) ((";") reduce 297) (("<") reduce 297) ((">") reduce 297) (("?") reduce 297) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_BOOLEAN_AND) reduce 297) ((T_BOOLEAN_OR) reduce 297) ((T_COALESCE) reduce 297) ((T_INSTANCEOF) reduce 297) ((T_IS_EQUAL) reduce 297) ((T_IS_GREATER_OR_EQUAL) reduce 297) ((T_IS_IDENTICAL) reduce 297) ((T_IS_NOT_EQUAL) reduce 297) ((T_IS_NOT_IDENTICAL) reduce 297) ((T_IS_SMALLER_OR_EQUAL) reduce 297) ((T_LOGICAL_AND) reduce 297) ((T_LOGICAL_OR) reduce 297) ((T_LOGICAL_XOR) reduce 297) ((T_PIPE) reduce 297) ((T_POW) reduce 297) ((T_SL) reduce 297) ((T_SPACESHIP) reduce 297) ((T_SR) reduce 297) (("^") reduce 297) (("{") reduce 297) (("|") reduce 297)) 7064 ((("%") reduce 460) (("(") reduce 556) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) ((";") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PIPE) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("[") reduce 549) (("^") reduce 460) (("{") reduce 460) (("|") reduce 460)) 7066 ((("%") reduce 298) (("*") reduce 298) (("+") reduce 298) ((",") reduce 298) (("-") reduce 298) ((".") reduce 298) (("/") reduce 298) ((";") reduce 298) (("<") reduce 298) ((">") reduce 298) (("?") reduce 298) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_BOOLEAN_AND) reduce 298) ((T_BOOLEAN_OR) reduce 298) ((T_COALESCE) reduce 298) ((T_INSTANCEOF) reduce 298) ((T_IS_EQUAL) reduce 298) ((T_IS_GREATER_OR_EQUAL) reduce 298) ((T_IS_IDENTICAL) reduce 298) ((T_IS_NOT_EQUAL) reduce 298) ((T_IS_NOT_IDENTICAL) reduce 298) ((T_IS_SMALLER_OR_EQUAL) reduce 298) ((T_LOGICAL_AND) reduce 298) ((T_LOGICAL_OR) reduce 298) ((T_LOGICAL_XOR) reduce 298) ((T_PIPE) reduce 298) ((T_POW) reduce 298) ((T_SL) reduce 298) ((T_SPACESHIP) reduce 298) ((T_SR) reduce 298) (("^") reduce 298) (("{") reduce 298) (("|") reduce 298)) 7067 ((("%") reduce 299) (("*") reduce 299) (("+") reduce 299) ((",") reduce 299) (("-") reduce 299) ((".") reduce 299) (("/") reduce 299) ((";") reduce 299) (("<") reduce 299) ((">") reduce 299) (("?") reduce 299) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_BOOLEAN_AND) reduce 299) ((T_BOOLEAN_OR) reduce 299) ((T_COALESCE) reduce 299) ((T_INSTANCEOF) reduce 299) ((T_IS_EQUAL) reduce 299) ((T_IS_GREATER_OR_EQUAL) reduce 299) ((T_IS_IDENTICAL) reduce 299) ((T_IS_NOT_EQUAL) reduce 299) ((T_IS_NOT_IDENTICAL) reduce 299) ((T_IS_SMALLER_OR_EQUAL) reduce 299) ((T_LOGICAL_AND) reduce 299) ((T_LOGICAL_OR) reduce 299) ((T_LOGICAL_XOR) reduce 299) ((T_PIPE) reduce 299) ((T_POW) reduce 299) ((T_SL) reduce 299) ((T_SPACESHIP) reduce 299) ((T_SR) reduce 299) (("^") reduce 299) (("{") reduce 299) (("|") reduce 299)) 7068 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) ((",") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) ((";") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_PIPE) reduce 472) ((T_POW) shift) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("{") reduce 472) (("|") reduce 472)) 7069 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) ((",") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) ((";") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_PIPE) reduce 470) ((T_POW) shift) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("{") reduce 470) (("|") reduce 470)) 7072 ((("%") reduce 519) (("(") reduce 519) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) ((";") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PIPE) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 7073 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) ((",") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) ((";") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_PIPE) reduce 475) ((T_POW) shift) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("{") reduce 475) (("|") reduce 475)) 7074 ((("%") reduce 446) (("*") reduce 446) (("+") reduce 446) ((",") reduce 446) (("-") reduce 446) ((".") reduce 446) (("/") reduce 446) ((";") reduce 446) (("<") reduce 446) ((">") reduce 446) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_BOOLEAN_AND) reduce 446) ((T_BOOLEAN_OR) reduce 446) ((T_COALESCE) reduce 446) ((T_INSTANCEOF) reduce 446) ((T_IS_EQUAL) reduce 446) ((T_IS_GREATER_OR_EQUAL) reduce 446) ((T_IS_IDENTICAL) reduce 446) ((T_IS_NOT_EQUAL) reduce 446) ((T_IS_NOT_IDENTICAL) reduce 446) ((T_IS_SMALLER_OR_EQUAL) reduce 446) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_PIPE) reduce 446) ((T_POW) shift) ((T_SL) reduce 446) ((T_SPACESHIP) reduce 446) ((T_SR) reduce 446) (("^") reduce 446) (("{") reduce 446) (("|") reduce 446)) 7075 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) ((",") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) ((";") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_PIPE) reduce 445) ((T_POW) shift) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("{") reduce 445) (("|") reduce 445)) 7077 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) ((";") reduce 570) (("<") reduce 570) (("=") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AND_EQUAL) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_COALESCE_EQUAL) reduce 570) ((T_CONCAT_EQUAL) reduce 570) ((T_DEC) reduce 570) ((T_DIV_EQUAL) reduce 570) ((T_INC) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_MINUS_EQUAL) reduce 570) ((T_MOD_EQUAL) reduce 570) ((T_MUL_EQUAL) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_OR_EQUAL) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_PLUS_EQUAL) reduce 570) ((T_POW) reduce 570) ((T_POW_EQUAL) reduce 570) ((T_SL) reduce 570) ((T_SL_EQUAL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) ((T_SR_EQUAL) reduce 570) ((T_XOR_EQUAL) reduce 570) (("[") reduce 570) (("^") reduce 570) (("{") reduce 570) (("|") reduce 570)) 7080 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) ((";") reduce 569) (("<") reduce 569) (("=") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AND_EQUAL) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_COALESCE_EQUAL) reduce 569) ((T_CONCAT_EQUAL) reduce 569) ((T_DEC) reduce 569) ((T_DIV_EQUAL) reduce 569) ((T_INC) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_MINUS_EQUAL) reduce 569) ((T_MOD_EQUAL) reduce 569) ((T_MUL_EQUAL) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_OR_EQUAL) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_PLUS_EQUAL) reduce 569) ((T_POW) reduce 569) ((T_POW_EQUAL) reduce 569) ((T_SL) reduce 569) ((T_SL_EQUAL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) ((T_SR_EQUAL) reduce 569) ((T_XOR_EQUAL) reduce 569) (("[") reduce 569) (("^") reduce 569) (("{") reduce 569) (("|") reduce 569)) 7082 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) ((";") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_PIPE) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 7083 ((("%") reduce 447) (("*") reduce 447) (("+") reduce 447) ((",") reduce 447) (("-") reduce 447) ((".") reduce 447) (("/") reduce 447) ((";") reduce 447) (("<") reduce 447) ((">") reduce 447) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_BOOLEAN_AND) reduce 447) ((T_BOOLEAN_OR) reduce 447) ((T_COALESCE) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 447) ((T_IS_GREATER_OR_EQUAL) reduce 447) ((T_IS_IDENTICAL) reduce 447) ((T_IS_NOT_EQUAL) reduce 447) ((T_IS_NOT_IDENTICAL) reduce 447) ((T_IS_SMALLER_OR_EQUAL) reduce 447) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_PIPE) reduce 447) ((T_POW) shift) ((T_SL) reduce 447) ((T_SPACESHIP) reduce 447) ((T_SR) reduce 447) (("^") reduce 447) (("{") reduce 447) (("|") reduce 447)) 7085 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 288) ((T_VARIABLE) reduce 288)) 7086 (((T_VARIABLE) reduce 280)) 7087 (((T_VARIABLE) reduce 281)) 7088 (((T_VARIABLE) reduce 94)) 7089 (((T_VARIABLE) reduce 93)) 7090 (((T_VARIABLE) reduce 95)) 7091 (((T_VARIABLE) reduce 92)) 7092 (((T_VARIABLE) reduce 282)) 7093 (((T_VARIABLE) reduce 277)) 7095 ((("|") reduce 284)) 7096 ((("(") reduce 351) (("?") reduce 351) ((T_ABSTRACT) reduce 351) ((T_ARRAY) reduce 351) ((T_CALLABLE) reduce 351) ((T_CONST) reduce 351) ((T_FINAL) reduce 351) ((T_FUNCTION) reduce 351) ((T_NAME_FULLY_QUALIFIED) reduce 351) ((T_NAME_QUALIFIED) reduce 351) ((T_NAME_RELATIVE) reduce 351) ((T_PRIVATE) reduce 351) ((T_PRIVATE_SET) reduce 351) ((T_PROTECTED) reduce 351) ((T_PROTECTED_SET) reduce 351) ((T_PUBLIC) reduce 351) ((T_PUBLIC_SET) reduce 351) ((T_READONLY) reduce 351) ((T_STATIC) reduce 351) ((T_STRING) reduce 351) ((T_VARIABLE) reduce 351)) 7097 (((T_ABSTRACT) reduce 492) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 492) ((T_ARRAY) reduce 492) ((T_AS) reduce 492) ((T_BREAK) reduce 492) ((T_CALLABLE) reduce 492) ((T_CASE) reduce 492) ((T_CATCH) reduce 492) ((T_CLASS) reduce 492) ((T_CLASS_C) reduce 492) ((T_CLONE) reduce 492) ((T_CONST) reduce 492) ((T_CONTINUE) reduce 492) ((T_DECLARE) reduce 492) ((T_DEFAULT) reduce 492) ((T_DIR) reduce 492) ((T_DO) reduce 492) ((T_ECHO) reduce 492) ((T_ELSE) reduce 492) ((T_ELSEIF) reduce 492) ((T_EMPTY) reduce 492) ((T_ENDDECLARE) reduce 492) ((T_ENDFOR) reduce 492) ((T_ENDFOREACH) reduce 492) ((T_ENDIF) reduce 492) ((T_ENDSWITCH) reduce 492) ((T_ENDWHILE) reduce 492) ((T_ENUM) reduce 492) ((T_EVAL) reduce 492) ((T_EXIT) reduce 492) ((T_EXTENDS) reduce 492) ((T_FILE) reduce 492) ((T_FINAL) reduce 492) ((T_FINALLY) reduce 492) ((T_FN) reduce 492) ((T_FOR) reduce 492) ((T_FOREACH) reduce 492) ((T_FUNCTION) reduce 492) ((T_FUNC_C) reduce 492) ((T_GLOBAL) reduce 492) ((T_GOTO) reduce 492) ((T_IF) reduce 492) ((T_IMPLEMENTS) reduce 492) ((T_INCLUDE) reduce 492) ((T_INCLUDE_ONCE) reduce 492) ((T_INSTANCEOF) reduce 492) ((T_INSTEADOF) reduce 492) ((T_INTERFACE) reduce 492) ((T_ISSET) reduce 492) ((T_LINE) reduce 492) ((T_LIST) reduce 492) ((T_LOGICAL_AND) reduce 492) ((T_LOGICAL_OR) reduce 492) ((T_LOGICAL_XOR) reduce 492) ((T_MATCH) reduce 492) ((T_METHOD_C) reduce 492) ((T_NAMESPACE) reduce 492) ((T_NEW) reduce 492) ((T_NS_C) reduce 492) ((T_PRINT) reduce 492) ((T_PRIVATE) reduce 492) ((T_PROPERTY_C) reduce 492) ((T_PROTECTED) reduce 492) ((T_PUBLIC) reduce 492) ((T_READONLY) reduce 492) ((T_REQUIRE) reduce 492) ((T_REQUIRE_ONCE) reduce 492) ((T_RETURN) reduce 492) ((T_STATIC) reduce 492) ((T_STRING) reduce 492) ((T_SWITCH) reduce 492) ((T_THROW) reduce 492) ((T_TRAIT) reduce 492) ((T_TRAIT_C) reduce 492) ((T_TRY) reduce 492) ((T_UNSET) reduce 492) ((T_USE) reduce 492) ((T_VAR) reduce 492) ((T_WHILE) reduce 492) ((T_YIELD) reduce 492)) 7098 (((T_ABSTRACT) reduce 496) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_ARRAY) reduce 496) ((T_AS) reduce 496) ((T_BREAK) reduce 496) ((T_CALLABLE) reduce 496) ((T_CASE) reduce 496) ((T_CATCH) reduce 496) ((T_CLASS) reduce 496) ((T_CLASS_C) reduce 496) ((T_CLONE) reduce 496) ((T_CONST) reduce 496) ((T_CONTINUE) reduce 496) ((T_DECLARE) reduce 496) ((T_DEFAULT) reduce 496) ((T_DIR) reduce 496) ((T_DO) reduce 496) ((T_ECHO) reduce 496) ((T_ELSE) reduce 496) ((T_ELSEIF) reduce 496) ((T_EMPTY) reduce 496) ((T_ENDDECLARE) reduce 496) ((T_ENDFOR) reduce 496) ((T_ENDFOREACH) reduce 496) ((T_ENDIF) reduce 496) ((T_ENDSWITCH) reduce 496) ((T_ENDWHILE) reduce 496) ((T_ENUM) reduce 496) ((T_EVAL) reduce 496) ((T_EXIT) reduce 496) ((T_EXTENDS) reduce 496) ((T_FILE) reduce 496) ((T_FINAL) reduce 496) ((T_FINALLY) reduce 496) ((T_FN) reduce 496) ((T_FOR) reduce 496) ((T_FOREACH) reduce 496) ((T_FUNCTION) reduce 496) ((T_FUNC_C) reduce 496) ((T_GLOBAL) reduce 496) ((T_GOTO) reduce 496) ((T_IF) reduce 496) ((T_IMPLEMENTS) reduce 496) ((T_INCLUDE) reduce 496) ((T_INCLUDE_ONCE) reduce 496) ((T_INSTANCEOF) reduce 496) ((T_INSTEADOF) reduce 496) ((T_INTERFACE) reduce 496) ((T_ISSET) reduce 496) ((T_LINE) reduce 496) ((T_LIST) reduce 496) ((T_LOGICAL_AND) reduce 496) ((T_LOGICAL_OR) reduce 496) ((T_LOGICAL_XOR) reduce 496) ((T_MATCH) reduce 496) ((T_METHOD_C) reduce 496) ((T_NAMESPACE) reduce 496) ((T_NEW) reduce 496) ((T_NS_C) reduce 496) ((T_PRINT) reduce 496) ((T_PRIVATE) reduce 496) ((T_PROPERTY_C) reduce 496) ((T_PROTECTED) reduce 496) ((T_PUBLIC) reduce 496) ((T_READONLY) reduce 496) ((T_REQUIRE) reduce 496) ((T_REQUIRE_ONCE) reduce 496) ((T_RETURN) reduce 496) ((T_STATIC) reduce 496) ((T_STRING) reduce 496) ((T_SWITCH) reduce 496) ((T_THROW) reduce 496) ((T_TRAIT) reduce 496) ((T_TRAIT_C) reduce 496) ((T_TRY) reduce 496) ((T_UNSET) reduce 496) ((T_USE) reduce 496) ((T_VAR) reduce 496) ((T_WHILE) reduce 496) ((T_YIELD) reduce 496)) 7099 (((T_ABSTRACT) reduce 80) ((T_ARRAY) reduce 80) ((T_AS) reduce 80) ((T_BREAK) reduce 80) ((T_CALLABLE) reduce 80) ((T_CASE) reduce 80) ((T_CATCH) reduce 80) ((T_CLASS) reduce 80) ((T_CLASS_C) reduce 80) ((T_CLONE) reduce 80) ((T_CONST) reduce 80) ((T_CONTINUE) reduce 80) ((T_DECLARE) reduce 80) ((T_DEFAULT) reduce 80) ((T_DIR) reduce 80) ((T_DO) reduce 80) ((T_ECHO) reduce 80) ((T_ELSE) reduce 80) ((T_ELSEIF) reduce 80) ((T_EMPTY) reduce 80) ((T_ENDDECLARE) reduce 80) ((T_ENDFOR) reduce 80) ((T_ENDFOREACH) reduce 80) ((T_ENDIF) reduce 80) ((T_ENDSWITCH) reduce 80) ((T_ENDWHILE) reduce 80) ((T_ENUM) reduce 80) ((T_EVAL) reduce 80) ((T_EXIT) reduce 80) ((T_EXTENDS) reduce 80) ((T_FILE) reduce 80) ((T_FINAL) reduce 80) ((T_FINALLY) reduce 80) ((T_FN) reduce 80) ((T_FOR) reduce 80) ((T_FOREACH) reduce 80) ((T_FUNCTION) reduce 80) ((T_FUNC_C) reduce 80) ((T_GLOBAL) reduce 80) ((T_GOTO) reduce 80) ((T_IF) reduce 80) ((T_IMPLEMENTS) reduce 80) ((T_INCLUDE) reduce 80) ((T_INCLUDE_ONCE) reduce 80) ((T_INSTANCEOF) reduce 80) ((T_INSTEADOF) reduce 80) ((T_INTERFACE) reduce 80) ((T_ISSET) reduce 80) ((T_LINE) reduce 80) ((T_LIST) reduce 80) ((T_LOGICAL_AND) reduce 80) ((T_LOGICAL_OR) reduce 80) ((T_LOGICAL_XOR) reduce 80) ((T_MATCH) reduce 80) ((T_METHOD_C) reduce 80) ((T_NAMESPACE) reduce 80) ((T_NEW) reduce 80) ((T_NS_C) reduce 80) ((T_PRINT) reduce 80) ((T_PRIVATE) reduce 80) ((T_PROPERTY_C) reduce 80) ((T_PROTECTED) reduce 80) ((T_PUBLIC) reduce 80) ((T_READONLY) reduce 80) ((T_REQUIRE) reduce 80) ((T_REQUIRE_ONCE) reduce 80) ((T_RETURN) reduce 80) ((T_STATIC) reduce 80) ((T_STRING) reduce 80) ((T_SWITCH) reduce 80) ((T_THROW) reduce 80) ((T_TRAIT) reduce 80) ((T_TRAIT_C) reduce 80) ((T_TRY) reduce 80) ((T_UNSET) reduce 80) ((T_USE) reduce 80) ((T_VAR) reduce 80) ((T_WHILE) reduce 80) ((T_YIELD) reduce 80)) 7100 (((T_ABSTRACT) reduce 81) ((T_ARRAY) reduce 81) ((T_AS) reduce 81) ((T_BREAK) reduce 81) ((T_CALLABLE) reduce 81) ((T_CASE) reduce 81) ((T_CATCH) reduce 81) ((T_CLASS) reduce 81) ((T_CLASS_C) reduce 81) ((T_CLONE) reduce 81) ((T_CONST) reduce 81) ((T_CONTINUE) reduce 81) ((T_DECLARE) reduce 81) ((T_DEFAULT) reduce 81) ((T_DIR) reduce 81) ((T_DO) reduce 81) ((T_ECHO) reduce 81) ((T_ELSE) reduce 81) ((T_ELSEIF) reduce 81) ((T_EMPTY) reduce 81) ((T_ENDDECLARE) reduce 81) ((T_ENDFOR) reduce 81) ((T_ENDFOREACH) reduce 81) ((T_ENDIF) reduce 81) ((T_ENDSWITCH) reduce 81) ((T_ENDWHILE) reduce 81) ((T_ENUM) reduce 81) ((T_EVAL) reduce 81) ((T_EXIT) reduce 81) ((T_EXTENDS) reduce 81) ((T_FILE) reduce 81) ((T_FINAL) reduce 81) ((T_FINALLY) reduce 81) ((T_FN) reduce 81) ((T_FOR) reduce 81) ((T_FOREACH) reduce 81) ((T_FUNCTION) reduce 81) ((T_FUNC_C) reduce 81) ((T_GLOBAL) reduce 81) ((T_GOTO) reduce 81) ((T_IF) reduce 81) ((T_IMPLEMENTS) reduce 81) ((T_INCLUDE) reduce 81) ((T_INCLUDE_ONCE) reduce 81) ((T_INSTANCEOF) reduce 81) ((T_INSTEADOF) reduce 81) ((T_INTERFACE) reduce 81) ((T_ISSET) reduce 81) ((T_LINE) reduce 81) ((T_LIST) reduce 81) ((T_LOGICAL_AND) reduce 81) ((T_LOGICAL_OR) reduce 81) ((T_LOGICAL_XOR) reduce 81) ((T_MATCH) reduce 81) ((T_METHOD_C) reduce 81) ((T_NAMESPACE) reduce 81) ((T_NEW) reduce 81) ((T_NS_C) reduce 81) ((T_PRINT) reduce 81) ((T_PRIVATE) reduce 81) ((T_PROPERTY_C) reduce 81) ((T_PROTECTED) reduce 81) ((T_PUBLIC) reduce 81) ((T_READONLY) reduce 81) ((T_REQUIRE) reduce 81) ((T_REQUIRE_ONCE) reduce 81) ((T_RETURN) reduce 81) ((T_STATIC) reduce 81) ((T_STRING) reduce 81) ((T_SWITCH) reduce 81) ((T_THROW) reduce 81) ((T_TRAIT) reduce 81) ((T_TRAIT_C) reduce 81) ((T_TRY) reduce 81) ((T_UNSET) reduce 81) ((T_USE) reduce 81) ((T_VAR) reduce 81) ((T_WHILE) reduce 81) ((T_YIELD) reduce 81)) 7101 (((T_ABSTRACT) reduce 497) ((T_ARRAY) reduce 497) ((T_AS) reduce 497) ((T_BREAK) reduce 497) ((T_CALLABLE) reduce 497) ((T_CASE) reduce 497) ((T_CATCH) reduce 497) ((T_CLASS) reduce 497) ((T_CLASS_C) reduce 497) ((T_CLONE) reduce 497) ((T_CONST) reduce 497) ((T_CONTINUE) reduce 497) ((T_DECLARE) reduce 497) ((T_DEFAULT) reduce 497) ((T_DIR) reduce 497) ((T_DO) reduce 497) ((T_ECHO) reduce 497) ((T_ELSE) reduce 497) ((T_ELSEIF) reduce 497) ((T_EMPTY) reduce 497) ((T_ENDDECLARE) reduce 497) ((T_ENDFOR) reduce 497) ((T_ENDFOREACH) reduce 497) ((T_ENDIF) reduce 497) ((T_ENDSWITCH) reduce 497) ((T_ENDWHILE) reduce 497) ((T_ENUM) reduce 497) ((T_EVAL) reduce 497) ((T_EXIT) reduce 497) ((T_EXTENDS) reduce 497) ((T_FILE) reduce 497) ((T_FINAL) reduce 497) ((T_FINALLY) reduce 497) ((T_FN) reduce 497) ((T_FOR) reduce 497) ((T_FOREACH) reduce 497) ((T_FUNCTION) reduce 497) ((T_FUNC_C) reduce 497) ((T_GLOBAL) reduce 497) ((T_GOTO) reduce 497) ((T_IF) reduce 497) ((T_IMPLEMENTS) reduce 497) ((T_INCLUDE) reduce 497) ((T_INCLUDE_ONCE) reduce 497) ((T_INSTANCEOF) reduce 497) ((T_INSTEADOF) reduce 497) ((T_INTERFACE) reduce 497) ((T_ISSET) reduce 497) ((T_LINE) reduce 497) ((T_LIST) reduce 497) ((T_LOGICAL_AND) reduce 497) ((T_LOGICAL_OR) reduce 497) ((T_LOGICAL_XOR) reduce 497) ((T_MATCH) reduce 497) ((T_METHOD_C) reduce 497) ((T_NAMESPACE) reduce 497) ((T_NEW) reduce 497) ((T_NS_C) reduce 497) ((T_PRINT) reduce 497) ((T_PRIVATE) reduce 497) ((T_PROPERTY_C) reduce 497) ((T_PROTECTED) reduce 497) ((T_PUBLIC) reduce 497) ((T_READONLY) reduce 497) ((T_REQUIRE) reduce 497) ((T_REQUIRE_ONCE) reduce 497) ((T_RETURN) reduce 497) ((T_STATIC) reduce 497) ((T_STRING) reduce 497) ((T_SWITCH) reduce 497) ((T_THROW) reduce 497) ((T_TRAIT) reduce 497) ((T_TRAIT_C) reduce 497) ((T_TRY) reduce 497) ((T_UNSET) reduce 497) ((T_USE) reduce 497) ((T_VAR) reduce 497) ((T_WHILE) reduce 497) ((T_YIELD) reduce 497)) 7102 (((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_AS) shift) ((T_BREAK) shift) ((T_CALLABLE) shift) ((T_CASE) shift) ((T_CATCH) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONST) shift) ((T_CONTINUE) shift) ((T_DECLARE) shift) ((T_DEFAULT) shift) ((T_DIR) shift) ((T_DO) shift) ((T_ECHO) shift) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) shift) ((T_ENDDECLARE) shift) ((T_ENDFOR) shift) ((T_ENDFOREACH) shift) ((T_ENDIF) shift) ((T_ENDSWITCH) shift) ((T_ENDWHILE) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_EXTENDS) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FINALLY) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_IF) shift) ((T_IMPLEMENTS) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) shift) ((T_INSTEADOF) shift) ((T_INTERFACE) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAMESPACE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_PRINT) shift) ((T_PRIVATE) shift) ((T_PROPERTY_C) shift) ((T_PROTECTED) shift) ((T_PUBLIC) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_USE) shift) ((T_VAR) shift) ((T_WHILE) shift) ((T_YIELD) shift)) 7103 ((("(") reduce 74)) 7104 ((("(") reduce 52)) 7105 ((("(") reduce 27)) 7106 ((("(") reduce 51)) 7107 ((("(") reduce 53)) 7108 ((("(") reduce 49)) 7109 ((("(") reduce 29)) 7110 ((("(") reduce 59)) 7111 ((("(") reduce 60)) 7112 ((("(") reduce 11)) 7113 ((("(") reduce 42)) 7114 ((("(") reduce 39)) 7115 ((("(") reduce 25)) 7116 ((("(") reduce 50)) 7117 ((("(") reduce 66)) 7118 ((("(") reduce 18)) 7119 ((("(") reduce 17)) 7120 ((("(") reduce 15)) 7121 ((("(") reduce 14)) 7122 ((("(") reduce 38)) 7123 ((("(") reduce 26)) 7124 ((("(") reduce 22)) 7125 ((("(") reduce 24)) 7126 ((("(") reduce 16)) 7127 ((("(") reduce 48)) 7128 ((("(") reduce 20)) 7129 ((("(") reduce 70)) 7130 ((("(") reduce 3)) 7131 ((("(") reduce 12)) 7132 ((("(") reduce 54)) 7133 ((("(") reduce 65)) 7134 ((("(") reduce 75)) 7135 ((("(") reduce 30)) 7136 ((("(") reduce 68)) 7137 ((("(") reduce 21)) 7138 ((("(") reduce 23)) 7139 ((("(") reduce 41)) 7140 ((("(") reduce 62)) 7141 ((("(") reduce 34)) 7142 ((("(") reduce 40)) 7143 ((("(") reduce 13)) 7144 ((("(") reduce 55)) 7145 ((("(") reduce 1)) 7146 ((("(") reduce 2)) 7147 ((("(") reduce 9)) 7148 ((("(") reduce 33)) 7149 ((("(") reduce 58)) 7150 ((("(") reduce 37)) 7151 ((("(") reduce 64)) 7152 ((("(") reduce 46)) 7153 ((("(") reduce 8)) 7154 ((("(") reduce 6)) 7155 ((("(") reduce 7)) 7156 ((("(") reduce 69)) 7157 ((("(") reduce 63)) 7158 ((("(") reduce 56)) 7159 ((("(") reduce 10)) 7160 ((("(") reduce 67)) 7161 ((("(") reduce 44)) 7162 ((("(") reduce 76)) 7163 ((("(") reduce 71)) 7164 ((("(") reduce 77)) 7165 ((("(") reduce 78)) 7166 ((("(") reduce 79)) 7167 ((("(") reduce 4)) 7168 ((("(") reduce 5)) 7169 ((("(") reduce 43)) 7170 ((("(") reduce 73)) 7171 ((("(") reduce 82)) 7172 ((("(") reduce 47)) 7173 ((("(") reduce 31)) 7174 ((("(") reduce 57)) 7175 ((("(") reduce 61)) 7176 ((("(") reduce 28)) 7177 ((("(") reduce 36)) 7178 ((("(") reduce 32)) 7179 ((("(") reduce 35)) 7180 ((("(") reduce 19)) 7181 ((("(") reduce 45)) 7183 ((("(") reduce 72)) 7184 ((("(") reduce 83)) 7188 (((":") shift) ((";") reduce 289) (("{") reduce 289)) 7190 (((";") reduce 494) (("{") reduce 494)) 7191 (((";") shift) (("{") shift)) 7192 (((T_ABSTRACT) reduce 342) ((T_ATTRIBUTE) reduce 342) ((T_CASE) reduce 342) ((T_CONST) reduce 342) ((T_FINAL) reduce 342) ((T_FUNCTION) reduce 342) ((T_PRIVATE) reduce 342) ((T_PRIVATE_SET) reduce 342) ((T_PROTECTED) reduce 342) ((T_PROTECTED_SET) reduce 342) ((T_PUBLIC) reduce 342) ((T_PUBLIC_SET) reduce 342) ((T_READONLY) reduce 342) ((T_STATIC) reduce 342) ((T_USE) reduce 342) ((T_VAR) reduce 342) (("}") reduce 342)) 7193 (((T_ABSTRACT) reduce 494) ((T_ATTRIBUTE) reduce 494) ((T_CASE) reduce 494) ((T_CONST) reduce 494) ((T_FINAL) reduce 494) ((T_FUNCTION) reduce 494) ((T_PRIVATE) reduce 494) ((T_PRIVATE_SET) reduce 494) ((T_PROTECTED) reduce 494) ((T_PROTECTED_SET) reduce 494) ((T_PUBLIC) reduce 494) ((T_PUBLIC_SET) reduce 494) ((T_READONLY) reduce 494) ((T_STATIC) reduce 494) ((T_USE) reduce 494) ((T_VAR) reduce 494) (("}") reduce 494)) 7196 (((T_ABSTRACT) reduce 343) ((T_ATTRIBUTE) reduce 343) ((T_CASE) reduce 343) ((T_CONST) reduce 343) ((T_FINAL) reduce 343) ((T_FUNCTION) reduce 343) ((T_PRIVATE) reduce 343) ((T_PRIVATE_SET) reduce 343) ((T_PROTECTED) reduce 343) ((T_PROTECTED_SET) reduce 343) ((T_PUBLIC) reduce 343) ((T_PUBLIC_SET) reduce 343) ((T_READONLY) reduce 343) ((T_STATIC) reduce 343) ((T_USE) reduce 343) ((T_VAR) reduce 343) (("}") reduce 343)) 7197 (((T_ABSTRACT) reduce 320) ((T_ATTRIBUTE) reduce 320) ((T_CASE) reduce 320) ((T_CONST) reduce 320) ((T_FINAL) reduce 320) ((T_FUNCTION) reduce 320) ((T_PRIVATE) reduce 320) ((T_PRIVATE_SET) reduce 320) ((T_PROTECTED) reduce 320) ((T_PROTECTED_SET) reduce 320) ((T_PUBLIC) reduce 320) ((T_PUBLIC_SET) reduce 320) ((T_READONLY) reduce 320) ((T_STATIC) reduce 320) ((T_USE) reduce 320) ((T_VAR) reduce 320) (("}") reduce 320)) 7199 (((";") reduce 280) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 280) (("{") reduce 280) (("|") reduce 280)) 7200 (((";") reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) (("{") reduce 281) (("|") reduce 281)) 7201 (((";") reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) (("{") reduce 94) (("|") reduce 94)) 7202 (((";") reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) (("{") reduce 93) (("|") reduce 93)) 7203 (((";") reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) (("{") reduce 95) (("|") reduce 95)) 7204 (((";") reduce 269) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 269) (("{") reduce 269) (("|") reduce 269)) 7205 (((";") reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) (("{") reduce 92) (("|") reduce 92)) 7206 (((";") reduce 267) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) (("{") reduce 267)) 7207 (((";") reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) (("{") reduce 282) (("|") reduce 282)) 7208 (((";") reduce 264) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) (("{") reduce 264) (("|") reduce 270)) 7209 (((";") reduce 290) (("{") reduce 290)) 7210 (((";") reduce 268) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 268) (("{") reduce 268) (("|") reduce 268)) 7211 (((";") reduce 266) (("{") reduce 266) (("|") shift)) 7215 (((";") reduce 280) (("{") reduce 280) (("|") reduce 280)) 7216 (((";") reduce 281) (("{") reduce 281) (("|") reduce 281)) 7217 (((";") reduce 94) (("{") reduce 94) (("|") reduce 94)) 7218 (((";") reduce 93) (("{") reduce 93) (("|") reduce 93)) 7219 (((";") reduce 95) (("{") reduce 95) (("|") reduce 95)) 7220 (((";") reduce 269) (("{") reduce 269) (("|") reduce 269)) 7221 (((";") reduce 92) (("{") reduce 92) (("|") reduce 92)) 7222 (((";") reduce 282) (("{") reduce 282) (("|") reduce 282)) 7223 (((";") reduce 270) (("{") reduce 270) (("|") reduce 270)) 7224 (((";") reduce 268) (("{") reduce 268) (("|") reduce 268)) 7225 (((";") reduce 272) (("{") reduce 272) (("|") reduce 272)) 7227 (((";") reduce 271) (("{") reduce 271) (("|") reduce 271)) 7229 (((";") reduce 273) (("{") reduce 273) (("|") reduce 273)) 7231 (((";") reduce 280) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 280) (("{") reduce 280)) 7232 (((";") reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) (("{") reduce 281)) 7233 (((";") reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) (("{") reduce 94)) 7234 (((";") reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) (("{") reduce 93)) 7235 (((";") reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) (("{") reduce 95)) 7236 (((";") reduce 269) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 269) (("{") reduce 269)) 7237 (((";") reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) (("{") reduce 92)) 7238 (((";") reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) (("{") reduce 282)) 7239 (((";") reduce 274) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 274) (("{") reduce 274)) 7240 (((";") reduce 268) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 268) (("{") reduce 268)) 7242 (((";") reduce 275) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 275) (("{") reduce 275)) 7243 (((";") reduce 280) (("{") reduce 280)) 7244 (((";") reduce 281) (("{") reduce 281)) 7245 (((";") reduce 94) (("{") reduce 94)) 7246 (((";") reduce 93) (("{") reduce 93)) 7247 (((";") reduce 95) (("{") reduce 95)) 7248 (((";") reduce 269) (("{") reduce 269)) 7249 (((";") reduce 92) (("{") reduce 92)) 7250 (((";") reduce 282) (("{") reduce 282)) 7251 (((";") reduce 265) (("{") reduce 265)) 7252 (((";") reduce 268) (("{") reduce 268)) 7253 ((("(") shift) (("?") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_AS) shift) ((T_BREAK) shift) ((T_CALLABLE) shift) ((T_CASE) shift) ((T_CATCH) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONST) shift) ((T_CONTINUE) shift) ((T_DECLARE) shift) ((T_DEFAULT) shift) ((T_DIR) shift) ((T_DO) shift) ((T_ECHO) shift) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) shift) ((T_ENDDECLARE) shift) ((T_ENDFOR) shift) ((T_ENDFOREACH) shift) ((T_ENDIF) shift) ((T_ENDSWITCH) shift) ((T_ENDWHILE) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_EXTENDS) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FINALLY) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_IF) shift) ((T_IMPLEMENTS) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) shift) ((T_INSTEADOF) shift) ((T_INTERFACE) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAMESPACE) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_PRINT) shift) ((T_PRIVATE) shift) ((T_PROPERTY_C) shift) ((T_PROTECTED) shift) ((T_PUBLIC) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_USE) shift) ((T_VAR) shift) ((T_WHILE) shift) ((T_YIELD) shift)) 7255 ((("=") reduce 74)) 7256 ((("=") reduce 52) ((T_ABSTRACT) reduce 280) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 280) ((T_ARRAY) reduce 280) ((T_AS) reduce 280) ((T_BREAK) reduce 280) ((T_CALLABLE) reduce 280) ((T_CASE) reduce 280) ((T_CATCH) reduce 280) ((T_CLASS) reduce 280) ((T_CLASS_C) reduce 280) ((T_CLONE) reduce 280) ((T_CONST) reduce 280) ((T_CONTINUE) reduce 280) ((T_DECLARE) reduce 280) ((T_DEFAULT) reduce 280) ((T_DIR) reduce 280) ((T_DO) reduce 280) ((T_ECHO) reduce 280) ((T_ELSE) reduce 280) ((T_ELSEIF) reduce 280) ((T_EMPTY) reduce 280) ((T_ENDDECLARE) reduce 280) ((T_ENDFOR) reduce 280) ((T_ENDFOREACH) reduce 280) ((T_ENDIF) reduce 280) ((T_ENDSWITCH) reduce 280) ((T_ENDWHILE) reduce 280) ((T_ENUM) reduce 280) ((T_EVAL) reduce 280) ((T_EXIT) reduce 280) ((T_EXTENDS) reduce 280) ((T_FILE) reduce 280) ((T_FINAL) reduce 280) ((T_FINALLY) reduce 280) ((T_FN) reduce 280) ((T_FOR) reduce 280) ((T_FOREACH) reduce 280) ((T_FUNCTION) reduce 280) ((T_FUNC_C) reduce 280) ((T_GLOBAL) reduce 280) ((T_GOTO) reduce 280) ((T_IF) reduce 280) ((T_IMPLEMENTS) reduce 280) ((T_INCLUDE) reduce 280) ((T_INCLUDE_ONCE) reduce 280) ((T_INSTANCEOF) reduce 280) ((T_INSTEADOF) reduce 280) ((T_INTERFACE) reduce 280) ((T_ISSET) reduce 280) ((T_LINE) reduce 280) ((T_LIST) reduce 280) ((T_LOGICAL_AND) reduce 280) ((T_LOGICAL_OR) reduce 280) ((T_LOGICAL_XOR) reduce 280) ((T_MATCH) reduce 280) ((T_METHOD_C) reduce 280) ((T_NAMESPACE) reduce 280) ((T_NEW) reduce 280) ((T_NS_C) reduce 280) ((T_PRINT) reduce 280) ((T_PRIVATE) reduce 280) ((T_PROPERTY_C) reduce 280) ((T_PROTECTED) reduce 280) ((T_PUBLIC) reduce 280) ((T_READONLY) reduce 280) ((T_REQUIRE) reduce 280) ((T_REQUIRE_ONCE) reduce 280) ((T_RETURN) reduce 280) ((T_STATIC) reduce 280) ((T_STRING) reduce 280) ((T_SWITCH) reduce 280) ((T_THROW) reduce 280) ((T_TRAIT) reduce 280) ((T_TRAIT_C) reduce 280) ((T_TRY) reduce 280) ((T_UNSET) reduce 280) ((T_USE) reduce 280) ((T_VAR) reduce 280) ((T_WHILE) reduce 280) ((T_YIELD) reduce 280) (("|") reduce 280)) 7257 ((("=") reduce 27)) 7258 ((("=") reduce 51)) 7259 ((("=") reduce 53) ((T_ABSTRACT) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_ARRAY) reduce 281) ((T_AS) reduce 281) ((T_BREAK) reduce 281) ((T_CALLABLE) reduce 281) ((T_CASE) reduce 281) ((T_CATCH) reduce 281) ((T_CLASS) reduce 281) ((T_CLASS_C) reduce 281) ((T_CLONE) reduce 281) ((T_CONST) reduce 281) ((T_CONTINUE) reduce 281) ((T_DECLARE) reduce 281) ((T_DEFAULT) reduce 281) ((T_DIR) reduce 281) ((T_DO) reduce 281) ((T_ECHO) reduce 281) ((T_ELSE) reduce 281) ((T_ELSEIF) reduce 281) ((T_EMPTY) reduce 281) ((T_ENDDECLARE) reduce 281) ((T_ENDFOR) reduce 281) ((T_ENDFOREACH) reduce 281) ((T_ENDIF) reduce 281) ((T_ENDSWITCH) reduce 281) ((T_ENDWHILE) reduce 281) ((T_ENUM) reduce 281) ((T_EVAL) reduce 281) ((T_EXIT) reduce 281) ((T_EXTENDS) reduce 281) ((T_FILE) reduce 281) ((T_FINAL) reduce 281) ((T_FINALLY) reduce 281) ((T_FN) reduce 281) ((T_FOR) reduce 281) ((T_FOREACH) reduce 281) ((T_FUNCTION) reduce 281) ((T_FUNC_C) reduce 281) ((T_GLOBAL) reduce 281) ((T_GOTO) reduce 281) ((T_IF) reduce 281) ((T_IMPLEMENTS) reduce 281) ((T_INCLUDE) reduce 281) ((T_INCLUDE_ONCE) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_INSTEADOF) reduce 281) ((T_INTERFACE) reduce 281) ((T_ISSET) reduce 281) ((T_LINE) reduce 281) ((T_LIST) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_MATCH) reduce 281) ((T_METHOD_C) reduce 281) ((T_NAMESPACE) reduce 281) ((T_NEW) reduce 281) ((T_NS_C) reduce 281) ((T_PRINT) reduce 281) ((T_PRIVATE) reduce 281) ((T_PROPERTY_C) reduce 281) ((T_PROTECTED) reduce 281) ((T_PUBLIC) reduce 281) ((T_READONLY) reduce 281) ((T_REQUIRE) reduce 281) ((T_REQUIRE_ONCE) reduce 281) ((T_RETURN) reduce 281) ((T_STATIC) reduce 281) ((T_STRING) reduce 281) ((T_SWITCH) reduce 281) ((T_THROW) reduce 281) ((T_TRAIT) reduce 281) ((T_TRAIT_C) reduce 281) ((T_TRY) reduce 281) ((T_UNSET) reduce 281) ((T_USE) reduce 281) ((T_VAR) reduce 281) ((T_WHILE) reduce 281) ((T_YIELD) reduce 281) (("|") reduce 281)) 7260 ((("=") reduce 49)) 7261 ((("=") reduce 29)) 7262 ((("=") reduce 59)) 7263 ((("=") reduce 60)) 7264 ((("=") reduce 11)) 7265 ((("=") reduce 42)) 7266 ((("=") reduce 39)) 7267 ((("=") reduce 25)) 7268 ((("=") reduce 50)) 7269 ((("=") reduce 66)) 7270 ((("=") reduce 18)) 7271 ((("=") reduce 17)) 7272 ((("=") reduce 15)) 7273 ((("=") reduce 14)) 7274 ((("=") reduce 38)) 7275 ((("=") reduce 26)) 7276 ((("=") reduce 22)) 7277 ((("=") reduce 24)) 7278 ((("=") reduce 16)) 7279 ((("=") reduce 48)) 7280 ((("=") reduce 20)) 7281 ((("=") reduce 70)) 7282 ((("=") reduce 3)) 7283 ((("=") reduce 12)) 7284 ((("=") reduce 54)) 7285 ((("=") reduce 65)) 7286 ((("=") reduce 75)) 7287 ((("=") reduce 30)) 7288 ((("=") reduce 68)) 7289 ((("=") reduce 21)) 7290 ((("=") reduce 23)) 7291 ((("=") reduce 41)) 7292 ((("=") reduce 62)) 7293 ((("=") reduce 34)) 7294 ((("=") reduce 40)) 7295 ((("=") reduce 13)) 7296 ((("=") reduce 55)) 7297 ((("=") reduce 1)) 7298 ((("=") reduce 2)) 7299 ((("=") reduce 9)) 7300 ((("=") reduce 33)) 7301 ((("=") reduce 58)) 7302 ((("=") reduce 37)) 7303 ((("=") reduce 64)) 7304 ((("=") reduce 46)) 7305 ((("=") reduce 8)) 7306 ((("=") reduce 6)) 7307 ((("=") reduce 7)) 7308 ((("=") reduce 69)) 7309 ((("=") reduce 63)) 7310 ((("=") reduce 56)) 7311 (((T_ABSTRACT) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_ARRAY) reduce 94) ((T_AS) reduce 94) ((T_BREAK) reduce 94) ((T_CALLABLE) reduce 94) ((T_CASE) reduce 94) ((T_CATCH) reduce 94) ((T_CLASS) reduce 94) ((T_CLASS_C) reduce 94) ((T_CLONE) reduce 94) ((T_CONST) reduce 94) ((T_CONTINUE) reduce 94) ((T_DECLARE) reduce 94) ((T_DEFAULT) reduce 94) ((T_DIR) reduce 94) ((T_DO) reduce 94) ((T_ECHO) reduce 94) ((T_ELSE) reduce 94) ((T_ELSEIF) reduce 94) ((T_EMPTY) reduce 94) ((T_ENDDECLARE) reduce 94) ((T_ENDFOR) reduce 94) ((T_ENDFOREACH) reduce 94) ((T_ENDIF) reduce 94) ((T_ENDSWITCH) reduce 94) ((T_ENDWHILE) reduce 94) ((T_ENUM) reduce 94) ((T_EVAL) reduce 94) ((T_EXIT) reduce 94) ((T_EXTENDS) reduce 94) ((T_FILE) reduce 94) ((T_FINAL) reduce 94) ((T_FINALLY) reduce 94) ((T_FN) reduce 94) ((T_FOR) reduce 94) ((T_FOREACH) reduce 94) ((T_FUNCTION) reduce 94) ((T_FUNC_C) reduce 94) ((T_GLOBAL) reduce 94) ((T_GOTO) reduce 94) ((T_IF) reduce 94) ((T_IMPLEMENTS) reduce 94) ((T_INCLUDE) reduce 94) ((T_INCLUDE_ONCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_INSTEADOF) reduce 94) ((T_INTERFACE) reduce 94) ((T_ISSET) reduce 94) ((T_LINE) reduce 94) ((T_LIST) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_MATCH) reduce 94) ((T_METHOD_C) reduce 94) ((T_NAMESPACE) reduce 94) ((T_NEW) reduce 94) ((T_NS_C) reduce 94) ((T_PRINT) reduce 94) ((T_PRIVATE) reduce 94) ((T_PROPERTY_C) reduce 94) ((T_PROTECTED) reduce 94) ((T_PUBLIC) reduce 94) ((T_READONLY) reduce 94) ((T_REQUIRE) reduce 94) ((T_REQUIRE_ONCE) reduce 94) ((T_RETURN) reduce 94) ((T_STATIC) reduce 94) ((T_STRING) reduce 94) ((T_SWITCH) reduce 94) ((T_THROW) reduce 94) ((T_TRAIT) reduce 94) ((T_TRAIT_C) reduce 94) ((T_TRY) reduce 94) ((T_UNSET) reduce 94) ((T_USE) reduce 94) ((T_VAR) reduce 94) ((T_WHILE) reduce 94) ((T_YIELD) reduce 94) (("|") reduce 94)) 7312 (((T_ABSTRACT) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_ARRAY) reduce 93) ((T_AS) reduce 93) ((T_BREAK) reduce 93) ((T_CALLABLE) reduce 93) ((T_CASE) reduce 93) ((T_CATCH) reduce 93) ((T_CLASS) reduce 93) ((T_CLASS_C) reduce 93) ((T_CLONE) reduce 93) ((T_CONST) reduce 93) ((T_CONTINUE) reduce 93) ((T_DECLARE) reduce 93) ((T_DEFAULT) reduce 93) ((T_DIR) reduce 93) ((T_DO) reduce 93) ((T_ECHO) reduce 93) ((T_ELSE) reduce 93) ((T_ELSEIF) reduce 93) ((T_EMPTY) reduce 93) ((T_ENDDECLARE) reduce 93) ((T_ENDFOR) reduce 93) ((T_ENDFOREACH) reduce 93) ((T_ENDIF) reduce 93) ((T_ENDSWITCH) reduce 93) ((T_ENDWHILE) reduce 93) ((T_ENUM) reduce 93) ((T_EVAL) reduce 93) ((T_EXIT) reduce 93) ((T_EXTENDS) reduce 93) ((T_FILE) reduce 93) ((T_FINAL) reduce 93) ((T_FINALLY) reduce 93) ((T_FN) reduce 93) ((T_FOR) reduce 93) ((T_FOREACH) reduce 93) ((T_FUNCTION) reduce 93) ((T_FUNC_C) reduce 93) ((T_GLOBAL) reduce 93) ((T_GOTO) reduce 93) ((T_IF) reduce 93) ((T_IMPLEMENTS) reduce 93) ((T_INCLUDE) reduce 93) ((T_INCLUDE_ONCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_INSTEADOF) reduce 93) ((T_INTERFACE) reduce 93) ((T_ISSET) reduce 93) ((T_LINE) reduce 93) ((T_LIST) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_MATCH) reduce 93) ((T_METHOD_C) reduce 93) ((T_NAMESPACE) reduce 93) ((T_NEW) reduce 93) ((T_NS_C) reduce 93) ((T_PRINT) reduce 93) ((T_PRIVATE) reduce 93) ((T_PROPERTY_C) reduce 93) ((T_PROTECTED) reduce 93) ((T_PUBLIC) reduce 93) ((T_READONLY) reduce 93) ((T_REQUIRE) reduce 93) ((T_REQUIRE_ONCE) reduce 93) ((T_RETURN) reduce 93) ((T_STATIC) reduce 93) ((T_STRING) reduce 93) ((T_SWITCH) reduce 93) ((T_THROW) reduce 93) ((T_TRAIT) reduce 93) ((T_TRAIT_C) reduce 93) ((T_TRY) reduce 93) ((T_UNSET) reduce 93) ((T_USE) reduce 93) ((T_VAR) reduce 93) ((T_WHILE) reduce 93) ((T_YIELD) reduce 93) (("|") reduce 93)) 7313 (((T_ABSTRACT) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_ARRAY) reduce 95) ((T_AS) reduce 95) ((T_BREAK) reduce 95) ((T_CALLABLE) reduce 95) ((T_CASE) reduce 95) ((T_CATCH) reduce 95) ((T_CLASS) reduce 95) ((T_CLASS_C) reduce 95) ((T_CLONE) reduce 95) ((T_CONST) reduce 95) ((T_CONTINUE) reduce 95) ((T_DECLARE) reduce 95) ((T_DEFAULT) reduce 95) ((T_DIR) reduce 95) ((T_DO) reduce 95) ((T_ECHO) reduce 95) ((T_ELSE) reduce 95) ((T_ELSEIF) reduce 95) ((T_EMPTY) reduce 95) ((T_ENDDECLARE) reduce 95) ((T_ENDFOR) reduce 95) ((T_ENDFOREACH) reduce 95) ((T_ENDIF) reduce 95) ((T_ENDSWITCH) reduce 95) ((T_ENDWHILE) reduce 95) ((T_ENUM) reduce 95) ((T_EVAL) reduce 95) ((T_EXIT) reduce 95) ((T_EXTENDS) reduce 95) ((T_FILE) reduce 95) ((T_FINAL) reduce 95) ((T_FINALLY) reduce 95) ((T_FN) reduce 95) ((T_FOR) reduce 95) ((T_FOREACH) reduce 95) ((T_FUNCTION) reduce 95) ((T_FUNC_C) reduce 95) ((T_GLOBAL) reduce 95) ((T_GOTO) reduce 95) ((T_IF) reduce 95) ((T_IMPLEMENTS) reduce 95) ((T_INCLUDE) reduce 95) ((T_INCLUDE_ONCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_INSTEADOF) reduce 95) ((T_INTERFACE) reduce 95) ((T_ISSET) reduce 95) ((T_LINE) reduce 95) ((T_LIST) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_MATCH) reduce 95) ((T_METHOD_C) reduce 95) ((T_NAMESPACE) reduce 95) ((T_NEW) reduce 95) ((T_NS_C) reduce 95) ((T_PRINT) reduce 95) ((T_PRIVATE) reduce 95) ((T_PROPERTY_C) reduce 95) ((T_PROTECTED) reduce 95) ((T_PUBLIC) reduce 95) ((T_READONLY) reduce 95) ((T_REQUIRE) reduce 95) ((T_REQUIRE_ONCE) reduce 95) ((T_RETURN) reduce 95) ((T_STATIC) reduce 95) ((T_STRING) reduce 95) ((T_SWITCH) reduce 95) ((T_THROW) reduce 95) ((T_TRAIT) reduce 95) ((T_TRAIT_C) reduce 95) ((T_TRY) reduce 95) ((T_UNSET) reduce 95) ((T_USE) reduce 95) ((T_VAR) reduce 95) ((T_WHILE) reduce 95) ((T_YIELD) reduce 95) (("|") reduce 95)) 7314 ((("=") reduce 10)) 7315 ((("=") reduce 67)) 7316 ((("=") reduce 44)) 7317 ((("=") reduce 76)) 7318 ((("=") reduce 71)) 7319 ((("=") reduce 77)) 7320 ((("=") reduce 78)) 7321 ((("=") reduce 79)) 7322 ((("=") reduce 4)) 7323 ((("=") reduce 5)) 7324 ((("=") reduce 43)) 7325 ((("=") reduce 73) ((T_ABSTRACT) reduce 269) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 269) ((T_ARRAY) reduce 269) ((T_AS) reduce 269) ((T_BREAK) reduce 269) ((T_CALLABLE) reduce 269) ((T_CASE) reduce 269) ((T_CATCH) reduce 269) ((T_CLASS) reduce 269) ((T_CLASS_C) reduce 269) ((T_CLONE) reduce 269) ((T_CONST) reduce 269) ((T_CONTINUE) reduce 269) ((T_DECLARE) reduce 269) ((T_DEFAULT) reduce 269) ((T_DIR) reduce 269) ((T_DO) reduce 269) ((T_ECHO) reduce 269) ((T_ELSE) reduce 269) ((T_ELSEIF) reduce 269) ((T_EMPTY) reduce 269) ((T_ENDDECLARE) reduce 269) ((T_ENDFOR) reduce 269) ((T_ENDFOREACH) reduce 269) ((T_ENDIF) reduce 269) ((T_ENDSWITCH) reduce 269) ((T_ENDWHILE) reduce 269) ((T_ENUM) reduce 269) ((T_EVAL) reduce 269) ((T_EXIT) reduce 269) ((T_EXTENDS) reduce 269) ((T_FILE) reduce 269) ((T_FINAL) reduce 269) ((T_FINALLY) reduce 269) ((T_FN) reduce 269) ((T_FOR) reduce 269) ((T_FOREACH) reduce 269) ((T_FUNCTION) reduce 269) ((T_FUNC_C) reduce 269) ((T_GLOBAL) reduce 269) ((T_GOTO) reduce 269) ((T_IF) reduce 269) ((T_IMPLEMENTS) reduce 269) ((T_INCLUDE) reduce 269) ((T_INCLUDE_ONCE) reduce 269) ((T_INSTANCEOF) reduce 269) ((T_INSTEADOF) reduce 269) ((T_INTERFACE) reduce 269) ((T_ISSET) reduce 269) ((T_LINE) reduce 269) ((T_LIST) reduce 269) ((T_LOGICAL_AND) reduce 269) ((T_LOGICAL_OR) reduce 269) ((T_LOGICAL_XOR) reduce 269) ((T_MATCH) reduce 269) ((T_METHOD_C) reduce 269) ((T_NAMESPACE) reduce 269) ((T_NEW) reduce 269) ((T_NS_C) reduce 269) ((T_PRINT) reduce 269) ((T_PRIVATE) reduce 269) ((T_PROPERTY_C) reduce 269) ((T_PROTECTED) reduce 269) ((T_PUBLIC) reduce 269) ((T_READONLY) reduce 269) ((T_REQUIRE) reduce 269) ((T_REQUIRE_ONCE) reduce 269) ((T_RETURN) reduce 269) ((T_STATIC) reduce 269) ((T_STRING) reduce 269) ((T_SWITCH) reduce 269) ((T_THROW) reduce 269) ((T_TRAIT) reduce 269) ((T_TRAIT_C) reduce 269) ((T_TRY) reduce 269) ((T_UNSET) reduce 269) ((T_USE) reduce 269) ((T_VAR) reduce 269) ((T_WHILE) reduce 269) ((T_YIELD) reduce 269) (("|") reduce 269)) 7326 ((("=") shift) ((T_ABSTRACT) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_ARRAY) reduce 92) ((T_AS) reduce 92) ((T_BREAK) reduce 92) ((T_CALLABLE) reduce 92) ((T_CASE) reduce 92) ((T_CATCH) reduce 92) ((T_CLASS) reduce 92) ((T_CLASS_C) reduce 92) ((T_CLONE) reduce 92) ((T_CONST) reduce 92) ((T_CONTINUE) reduce 92) ((T_DECLARE) reduce 92) ((T_DEFAULT) reduce 92) ((T_DIR) reduce 92) ((T_DO) reduce 92) ((T_ECHO) reduce 92) ((T_ELSE) reduce 92) ((T_ELSEIF) reduce 92) ((T_EMPTY) reduce 92) ((T_ENDDECLARE) reduce 92) ((T_ENDFOR) reduce 92) ((T_ENDFOREACH) reduce 92) ((T_ENDIF) reduce 92) ((T_ENDSWITCH) reduce 92) ((T_ENDWHILE) reduce 92) ((T_ENUM) reduce 92) ((T_EVAL) reduce 92) ((T_EXIT) reduce 92) ((T_EXTENDS) reduce 92) ((T_FILE) reduce 92) ((T_FINAL) reduce 92) ((T_FINALLY) reduce 92) ((T_FN) reduce 92) ((T_FOR) reduce 92) ((T_FOREACH) reduce 92) ((T_FUNCTION) reduce 92) ((T_FUNC_C) reduce 92) ((T_GLOBAL) reduce 92) ((T_GOTO) reduce 92) ((T_IF) reduce 92) ((T_IMPLEMENTS) reduce 92) ((T_INCLUDE) reduce 92) ((T_INCLUDE_ONCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_INSTEADOF) reduce 92) ((T_INTERFACE) reduce 92) ((T_ISSET) reduce 92) ((T_LINE) reduce 92) ((T_LIST) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_MATCH) reduce 92) ((T_METHOD_C) reduce 92) ((T_NAMESPACE) reduce 92) ((T_NEW) reduce 92) ((T_NS_C) reduce 92) ((T_PRINT) reduce 92) ((T_PRIVATE) reduce 92) ((T_PROPERTY_C) reduce 92) ((T_PROTECTED) reduce 92) ((T_PUBLIC) reduce 92) ((T_READONLY) reduce 92) ((T_REQUIRE) reduce 92) ((T_REQUIRE_ONCE) reduce 92) ((T_RETURN) reduce 92) ((T_STATIC) reduce 92) ((T_STRING) reduce 92) ((T_SWITCH) reduce 92) ((T_THROW) reduce 92) ((T_TRAIT) reduce 92) ((T_TRAIT_C) reduce 92) ((T_TRY) reduce 92) ((T_UNSET) reduce 92) ((T_USE) reduce 92) ((T_VAR) reduce 92) ((T_WHILE) reduce 92) ((T_YIELD) reduce 92) (("|") reduce 92)) 7327 ((("=") reduce 47)) 7328 ((("=") reduce 31)) 7329 ((("=") reduce 57)) 7330 ((("=") reduce 61)) 7331 ((("=") reduce 28)) 7332 ((("=") reduce 36)) 7333 ((("=") reduce 32)) 7334 ((("=") reduce 35)) 7335 ((("=") reduce 19)) 7336 ((("=") reduce 45)) 7337 (((",") reduce 382) ((";") reduce 382)) 7339 (((T_ABSTRACT) reduce 267) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_ARRAY) reduce 267) ((T_AS) reduce 267) ((T_BREAK) reduce 267) ((T_CALLABLE) reduce 267) ((T_CASE) reduce 267) ((T_CATCH) reduce 267) ((T_CLASS) reduce 267) ((T_CLASS_C) reduce 267) ((T_CLONE) reduce 267) ((T_CONST) reduce 267) ((T_CONTINUE) reduce 267) ((T_DECLARE) reduce 267) ((T_DEFAULT) reduce 267) ((T_DIR) reduce 267) ((T_DO) reduce 267) ((T_ECHO) reduce 267) ((T_ELSE) reduce 267) ((T_ELSEIF) reduce 267) ((T_EMPTY) reduce 267) ((T_ENDDECLARE) reduce 267) ((T_ENDFOR) reduce 267) ((T_ENDFOREACH) reduce 267) ((T_ENDIF) reduce 267) ((T_ENDSWITCH) reduce 267) ((T_ENDWHILE) reduce 267) ((T_ENUM) reduce 267) ((T_EVAL) reduce 267) ((T_EXIT) reduce 267) ((T_EXTENDS) reduce 267) ((T_FILE) reduce 267) ((T_FINAL) reduce 267) ((T_FINALLY) reduce 267) ((T_FN) reduce 267) ((T_FOR) reduce 267) ((T_FOREACH) reduce 267) ((T_FUNCTION) reduce 267) ((T_FUNC_C) reduce 267) ((T_GLOBAL) reduce 267) ((T_GOTO) reduce 267) ((T_IF) reduce 267) ((T_IMPLEMENTS) reduce 267) ((T_INCLUDE) reduce 267) ((T_INCLUDE_ONCE) reduce 267) ((T_INSTANCEOF) reduce 267) ((T_INSTEADOF) reduce 267) ((T_INTERFACE) reduce 267) ((T_ISSET) reduce 267) ((T_LINE) reduce 267) ((T_LIST) reduce 267) ((T_LOGICAL_AND) reduce 267) ((T_LOGICAL_OR) reduce 267) ((T_LOGICAL_XOR) reduce 267) ((T_MATCH) reduce 267) ((T_METHOD_C) reduce 267) ((T_NAMESPACE) reduce 267) ((T_NEW) reduce 267) ((T_NS_C) reduce 267) ((T_PRINT) reduce 267) ((T_PRIVATE) reduce 267) ((T_PROPERTY_C) reduce 267) ((T_PROTECTED) reduce 267) ((T_PUBLIC) reduce 267) ((T_READONLY) reduce 267) ((T_REQUIRE) reduce 267) ((T_REQUIRE_ONCE) reduce 267) ((T_RETURN) reduce 267) ((T_STATIC) reduce 267) ((T_STRING) reduce 267) ((T_SWITCH) reduce 267) ((T_THROW) reduce 267) ((T_TRAIT) reduce 267) ((T_TRAIT_C) reduce 267) ((T_TRY) reduce 267) ((T_UNSET) reduce 267) ((T_USE) reduce 267) ((T_VAR) reduce 267) ((T_WHILE) reduce 267) ((T_YIELD) reduce 267)) 7340 (((T_ABSTRACT) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_ARRAY) reduce 282) ((T_AS) reduce 282) ((T_BREAK) reduce 282) ((T_CALLABLE) reduce 282) ((T_CASE) reduce 282) ((T_CATCH) reduce 282) ((T_CLASS) reduce 282) ((T_CLASS_C) reduce 282) ((T_CLONE) reduce 282) ((T_CONST) reduce 282) ((T_CONTINUE) reduce 282) ((T_DECLARE) reduce 282) ((T_DEFAULT) reduce 282) ((T_DIR) reduce 282) ((T_DO) reduce 282) ((T_ECHO) reduce 282) ((T_ELSE) reduce 282) ((T_ELSEIF) reduce 282) ((T_EMPTY) reduce 282) ((T_ENDDECLARE) reduce 282) ((T_ENDFOR) reduce 282) ((T_ENDFOREACH) reduce 282) ((T_ENDIF) reduce 282) ((T_ENDSWITCH) reduce 282) ((T_ENDWHILE) reduce 282) ((T_ENUM) reduce 282) ((T_EVAL) reduce 282) ((T_EXIT) reduce 282) ((T_EXTENDS) reduce 282) ((T_FILE) reduce 282) ((T_FINAL) reduce 282) ((T_FINALLY) reduce 282) ((T_FN) reduce 282) ((T_FOR) reduce 282) ((T_FOREACH) reduce 282) ((T_FUNCTION) reduce 282) ((T_FUNC_C) reduce 282) ((T_GLOBAL) reduce 282) ((T_GOTO) reduce 282) ((T_IF) reduce 282) ((T_IMPLEMENTS) reduce 282) ((T_INCLUDE) reduce 282) ((T_INCLUDE_ONCE) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_INSTEADOF) reduce 282) ((T_INTERFACE) reduce 282) ((T_ISSET) reduce 282) ((T_LINE) reduce 282) ((T_LIST) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_MATCH) reduce 282) ((T_METHOD_C) reduce 282) ((T_NAMESPACE) reduce 282) ((T_NEW) reduce 282) ((T_NS_C) reduce 282) ((T_PRINT) reduce 282) ((T_PRIVATE) reduce 282) ((T_PROPERTY_C) reduce 282) ((T_PROTECTED) reduce 282) ((T_PUBLIC) reduce 282) ((T_READONLY) reduce 282) ((T_REQUIRE) reduce 282) ((T_REQUIRE_ONCE) reduce 282) ((T_RETURN) reduce 282) ((T_STATIC) reduce 282) ((T_STRING) reduce 282) ((T_SWITCH) reduce 282) ((T_THROW) reduce 282) ((T_TRAIT) reduce 282) ((T_TRAIT_C) reduce 282) ((T_TRY) reduce 282) ((T_UNSET) reduce 282) ((T_USE) reduce 282) ((T_VAR) reduce 282) ((T_WHILE) reduce 282) ((T_YIELD) reduce 282) (("|") reduce 282)) 7341 ((("=") reduce 72)) 7343 (((T_ABSTRACT) reduce 264) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_ARRAY) reduce 264) ((T_AS) reduce 264) ((T_BREAK) reduce 264) ((T_CALLABLE) reduce 264) ((T_CASE) reduce 264) ((T_CATCH) reduce 264) ((T_CLASS) reduce 264) ((T_CLASS_C) reduce 264) ((T_CLONE) reduce 264) ((T_CONST) reduce 264) ((T_CONTINUE) reduce 264) ((T_DECLARE) reduce 264) ((T_DEFAULT) reduce 264) ((T_DIR) reduce 264) ((T_DO) reduce 264) ((T_ECHO) reduce 264) ((T_ELSE) reduce 264) ((T_ELSEIF) reduce 264) ((T_EMPTY) reduce 264) ((T_ENDDECLARE) reduce 264) ((T_ENDFOR) reduce 264) ((T_ENDFOREACH) reduce 264) ((T_ENDIF) reduce 264) ((T_ENDSWITCH) reduce 264) ((T_ENDWHILE) reduce 264) ((T_ENUM) reduce 264) ((T_EVAL) reduce 264) ((T_EXIT) reduce 264) ((T_EXTENDS) reduce 264) ((T_FILE) reduce 264) ((T_FINAL) reduce 264) ((T_FINALLY) reduce 264) ((T_FN) reduce 264) ((T_FOR) reduce 264) ((T_FOREACH) reduce 264) ((T_FUNCTION) reduce 264) ((T_FUNC_C) reduce 264) ((T_GLOBAL) reduce 264) ((T_GOTO) reduce 264) ((T_IF) reduce 264) ((T_IMPLEMENTS) reduce 264) ((T_INCLUDE) reduce 264) ((T_INCLUDE_ONCE) reduce 264) ((T_INSTANCEOF) reduce 264) ((T_INSTEADOF) reduce 264) ((T_INTERFACE) reduce 264) ((T_ISSET) reduce 264) ((T_LINE) reduce 264) ((T_LIST) reduce 264) ((T_LOGICAL_AND) reduce 264) ((T_LOGICAL_OR) reduce 264) ((T_LOGICAL_XOR) reduce 264) ((T_MATCH) reduce 264) ((T_METHOD_C) reduce 264) ((T_NAMESPACE) reduce 264) ((T_NEW) reduce 264) ((T_NS_C) reduce 264) ((T_PRINT) reduce 264) ((T_PRIVATE) reduce 264) ((T_PROPERTY_C) reduce 264) ((T_PROTECTED) reduce 264) ((T_PUBLIC) reduce 264) ((T_READONLY) reduce 264) ((T_REQUIRE) reduce 264) ((T_REQUIRE_ONCE) reduce 264) ((T_RETURN) reduce 264) ((T_STATIC) reduce 264) ((T_STRING) reduce 264) ((T_SWITCH) reduce 264) ((T_THROW) reduce 264) ((T_TRAIT) reduce 264) ((T_TRAIT_C) reduce 264) ((T_TRY) reduce 264) ((T_UNSET) reduce 264) ((T_USE) reduce 264) ((T_VAR) reduce 264) ((T_WHILE) reduce 264) ((T_YIELD) reduce 264) (("|") reduce 270)) 7345 (((T_ABSTRACT) reduce 268) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 268) ((T_ARRAY) reduce 268) ((T_AS) reduce 268) ((T_BREAK) reduce 268) ((T_CALLABLE) reduce 268) ((T_CASE) reduce 268) ((T_CATCH) reduce 268) ((T_CLASS) reduce 268) ((T_CLASS_C) reduce 268) ((T_CLONE) reduce 268) ((T_CONST) reduce 268) ((T_CONTINUE) reduce 268) ((T_DECLARE) reduce 268) ((T_DEFAULT) reduce 268) ((T_DIR) reduce 268) ((T_DO) reduce 268) ((T_ECHO) reduce 268) ((T_ELSE) reduce 268) ((T_ELSEIF) reduce 268) ((T_EMPTY) reduce 268) ((T_ENDDECLARE) reduce 268) ((T_ENDFOR) reduce 268) ((T_ENDFOREACH) reduce 268) ((T_ENDIF) reduce 268) ((T_ENDSWITCH) reduce 268) ((T_ENDWHILE) reduce 268) ((T_ENUM) reduce 268) ((T_EVAL) reduce 268) ((T_EXIT) reduce 268) ((T_EXTENDS) reduce 268) ((T_FILE) reduce 268) ((T_FINAL) reduce 268) ((T_FINALLY) reduce 268) ((T_FN) reduce 268) ((T_FOR) reduce 268) ((T_FOREACH) reduce 268) ((T_FUNCTION) reduce 268) ((T_FUNC_C) reduce 268) ((T_GLOBAL) reduce 268) ((T_GOTO) reduce 268) ((T_IF) reduce 268) ((T_IMPLEMENTS) reduce 268) ((T_INCLUDE) reduce 268) ((T_INCLUDE_ONCE) reduce 268) ((T_INSTANCEOF) reduce 268) ((T_INSTEADOF) reduce 268) ((T_INTERFACE) reduce 268) ((T_ISSET) reduce 268) ((T_LINE) reduce 268) ((T_LIST) reduce 268) ((T_LOGICAL_AND) reduce 268) ((T_LOGICAL_OR) reduce 268) ((T_LOGICAL_XOR) reduce 268) ((T_MATCH) reduce 268) ((T_METHOD_C) reduce 268) ((T_NAMESPACE) reduce 268) ((T_NEW) reduce 268) ((T_NS_C) reduce 268) ((T_PRINT) reduce 268) ((T_PRIVATE) reduce 268) ((T_PROPERTY_C) reduce 268) ((T_PROTECTED) reduce 268) ((T_PUBLIC) reduce 268) ((T_READONLY) reduce 268) ((T_REQUIRE) reduce 268) ((T_REQUIRE_ONCE) reduce 268) ((T_RETURN) reduce 268) ((T_STATIC) reduce 268) ((T_STRING) reduce 268) ((T_SWITCH) reduce 268) ((T_THROW) reduce 268) ((T_TRAIT) reduce 268) ((T_TRAIT_C) reduce 268) ((T_TRY) reduce 268) ((T_UNSET) reduce 268) ((T_USE) reduce 268) ((T_VAR) reduce 268) ((T_WHILE) reduce 268) ((T_YIELD) reduce 268) (("|") reduce 268)) 7346 (((T_ABSTRACT) reduce 266) ((T_ARRAY) reduce 266) ((T_AS) reduce 266) ((T_BREAK) reduce 266) ((T_CALLABLE) reduce 266) ((T_CASE) reduce 266) ((T_CATCH) reduce 266) ((T_CLASS) reduce 266) ((T_CLASS_C) reduce 266) ((T_CLONE) reduce 266) ((T_CONST) reduce 266) ((T_CONTINUE) reduce 266) ((T_DECLARE) reduce 266) ((T_DEFAULT) reduce 266) ((T_DIR) reduce 266) ((T_DO) reduce 266) ((T_ECHO) reduce 266) ((T_ELSE) reduce 266) ((T_ELSEIF) reduce 266) ((T_EMPTY) reduce 266) ((T_ENDDECLARE) reduce 266) ((T_ENDFOR) reduce 266) ((T_ENDFOREACH) reduce 266) ((T_ENDIF) reduce 266) ((T_ENDSWITCH) reduce 266) ((T_ENDWHILE) reduce 266) ((T_ENUM) reduce 266) ((T_EVAL) reduce 266) ((T_EXIT) reduce 266) ((T_EXTENDS) reduce 266) ((T_FILE) reduce 266) ((T_FINAL) reduce 266) ((T_FINALLY) reduce 266) ((T_FN) reduce 266) ((T_FOR) reduce 266) ((T_FOREACH) reduce 266) ((T_FUNCTION) reduce 266) ((T_FUNC_C) reduce 266) ((T_GLOBAL) reduce 266) ((T_GOTO) reduce 266) ((T_IF) reduce 266) ((T_IMPLEMENTS) reduce 266) ((T_INCLUDE) reduce 266) ((T_INCLUDE_ONCE) reduce 266) ((T_INSTANCEOF) reduce 266) ((T_INSTEADOF) reduce 266) ((T_INTERFACE) reduce 266) ((T_ISSET) reduce 266) ((T_LINE) reduce 266) ((T_LIST) reduce 266) ((T_LOGICAL_AND) reduce 266) ((T_LOGICAL_OR) reduce 266) ((T_LOGICAL_XOR) reduce 266) ((T_MATCH) reduce 266) ((T_METHOD_C) reduce 266) ((T_NAMESPACE) reduce 266) ((T_NEW) reduce 266) ((T_NS_C) reduce 266) ((T_PRINT) reduce 266) ((T_PRIVATE) reduce 266) ((T_PROPERTY_C) reduce 266) ((T_PROTECTED) reduce 266) ((T_PUBLIC) reduce 266) ((T_READONLY) reduce 266) ((T_REQUIRE) reduce 266) ((T_REQUIRE_ONCE) reduce 266) ((T_RETURN) reduce 266) ((T_STATIC) reduce 266) ((T_STRING) reduce 266) ((T_SWITCH) reduce 266) ((T_THROW) reduce 266) ((T_TRAIT) reduce 266) ((T_TRAIT_C) reduce 266) ((T_TRY) reduce 266) ((T_UNSET) reduce 266) ((T_USE) reduce 266) ((T_VAR) reduce 266) ((T_WHILE) reduce 266) ((T_YIELD) reduce 266) (("|") shift)) 7350 (((T_ABSTRACT) reduce 280) ((T_ARRAY) reduce 280) ((T_AS) reduce 280) ((T_BREAK) reduce 280) ((T_CALLABLE) reduce 280) ((T_CASE) reduce 280) ((T_CATCH) reduce 280) ((T_CLASS) reduce 280) ((T_CLASS_C) reduce 280) ((T_CLONE) reduce 280) ((T_CONST) reduce 280) ((T_CONTINUE) reduce 280) ((T_DECLARE) reduce 280) ((T_DEFAULT) reduce 280) ((T_DIR) reduce 280) ((T_DO) reduce 280) ((T_ECHO) reduce 280) ((T_ELSE) reduce 280) ((T_ELSEIF) reduce 280) ((T_EMPTY) reduce 280) ((T_ENDDECLARE) reduce 280) ((T_ENDFOR) reduce 280) ((T_ENDFOREACH) reduce 280) ((T_ENDIF) reduce 280) ((T_ENDSWITCH) reduce 280) ((T_ENDWHILE) reduce 280) ((T_ENUM) reduce 280) ((T_EVAL) reduce 280) ((T_EXIT) reduce 280) ((T_EXTENDS) reduce 280) ((T_FILE) reduce 280) ((T_FINAL) reduce 280) ((T_FINALLY) reduce 280) ((T_FN) reduce 280) ((T_FOR) reduce 280) ((T_FOREACH) reduce 280) ((T_FUNCTION) reduce 280) ((T_FUNC_C) reduce 280) ((T_GLOBAL) reduce 280) ((T_GOTO) reduce 280) ((T_IF) reduce 280) ((T_IMPLEMENTS) reduce 280) ((T_INCLUDE) reduce 280) ((T_INCLUDE_ONCE) reduce 280) ((T_INSTANCEOF) reduce 280) ((T_INSTEADOF) reduce 280) ((T_INTERFACE) reduce 280) ((T_ISSET) reduce 280) ((T_LINE) reduce 280) ((T_LIST) reduce 280) ((T_LOGICAL_AND) reduce 280) ((T_LOGICAL_OR) reduce 280) ((T_LOGICAL_XOR) reduce 280) ((T_MATCH) reduce 280) ((T_METHOD_C) reduce 280) ((T_NAMESPACE) reduce 280) ((T_NEW) reduce 280) ((T_NS_C) reduce 280) ((T_PRINT) reduce 280) ((T_PRIVATE) reduce 280) ((T_PROPERTY_C) reduce 280) ((T_PROTECTED) reduce 280) ((T_PUBLIC) reduce 280) ((T_READONLY) reduce 280) ((T_REQUIRE) reduce 280) ((T_REQUIRE_ONCE) reduce 280) ((T_RETURN) reduce 280) ((T_STATIC) reduce 280) ((T_STRING) reduce 280) ((T_SWITCH) reduce 280) ((T_THROW) reduce 280) ((T_TRAIT) reduce 280) ((T_TRAIT_C) reduce 280) ((T_TRY) reduce 280) ((T_UNSET) reduce 280) ((T_USE) reduce 280) ((T_VAR) reduce 280) ((T_WHILE) reduce 280) ((T_YIELD) reduce 280) (("|") reduce 280)) 7351 (((T_ABSTRACT) reduce 281) ((T_ARRAY) reduce 281) ((T_AS) reduce 281) ((T_BREAK) reduce 281) ((T_CALLABLE) reduce 281) ((T_CASE) reduce 281) ((T_CATCH) reduce 281) ((T_CLASS) reduce 281) ((T_CLASS_C) reduce 281) ((T_CLONE) reduce 281) ((T_CONST) reduce 281) ((T_CONTINUE) reduce 281) ((T_DECLARE) reduce 281) ((T_DEFAULT) reduce 281) ((T_DIR) reduce 281) ((T_DO) reduce 281) ((T_ECHO) reduce 281) ((T_ELSE) reduce 281) ((T_ELSEIF) reduce 281) ((T_EMPTY) reduce 281) ((T_ENDDECLARE) reduce 281) ((T_ENDFOR) reduce 281) ((T_ENDFOREACH) reduce 281) ((T_ENDIF) reduce 281) ((T_ENDSWITCH) reduce 281) ((T_ENDWHILE) reduce 281) ((T_ENUM) reduce 281) ((T_EVAL) reduce 281) ((T_EXIT) reduce 281) ((T_EXTENDS) reduce 281) ((T_FILE) reduce 281) ((T_FINAL) reduce 281) ((T_FINALLY) reduce 281) ((T_FN) reduce 281) ((T_FOR) reduce 281) ((T_FOREACH) reduce 281) ((T_FUNCTION) reduce 281) ((T_FUNC_C) reduce 281) ((T_GLOBAL) reduce 281) ((T_GOTO) reduce 281) ((T_IF) reduce 281) ((T_IMPLEMENTS) reduce 281) ((T_INCLUDE) reduce 281) ((T_INCLUDE_ONCE) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_INSTEADOF) reduce 281) ((T_INTERFACE) reduce 281) ((T_ISSET) reduce 281) ((T_LINE) reduce 281) ((T_LIST) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_MATCH) reduce 281) ((T_METHOD_C) reduce 281) ((T_NAMESPACE) reduce 281) ((T_NEW) reduce 281) ((T_NS_C) reduce 281) ((T_PRINT) reduce 281) ((T_PRIVATE) reduce 281) ((T_PROPERTY_C) reduce 281) ((T_PROTECTED) reduce 281) ((T_PUBLIC) reduce 281) ((T_READONLY) reduce 281) ((T_REQUIRE) reduce 281) ((T_REQUIRE_ONCE) reduce 281) ((T_RETURN) reduce 281) ((T_STATIC) reduce 281) ((T_STRING) reduce 281) ((T_SWITCH) reduce 281) ((T_THROW) reduce 281) ((T_TRAIT) reduce 281) ((T_TRAIT_C) reduce 281) ((T_TRY) reduce 281) ((T_UNSET) reduce 281) ((T_USE) reduce 281) ((T_VAR) reduce 281) ((T_WHILE) reduce 281) ((T_YIELD) reduce 281) (("|") reduce 281)) 7352 (((T_ABSTRACT) reduce 94) ((T_ARRAY) reduce 94) ((T_AS) reduce 94) ((T_BREAK) reduce 94) ((T_CALLABLE) reduce 94) ((T_CASE) reduce 94) ((T_CATCH) reduce 94) ((T_CLASS) reduce 94) ((T_CLASS_C) reduce 94) ((T_CLONE) reduce 94) ((T_CONST) reduce 94) ((T_CONTINUE) reduce 94) ((T_DECLARE) reduce 94) ((T_DEFAULT) reduce 94) ((T_DIR) reduce 94) ((T_DO) reduce 94) ((T_ECHO) reduce 94) ((T_ELSE) reduce 94) ((T_ELSEIF) reduce 94) ((T_EMPTY) reduce 94) ((T_ENDDECLARE) reduce 94) ((T_ENDFOR) reduce 94) ((T_ENDFOREACH) reduce 94) ((T_ENDIF) reduce 94) ((T_ENDSWITCH) reduce 94) ((T_ENDWHILE) reduce 94) ((T_ENUM) reduce 94) ((T_EVAL) reduce 94) ((T_EXIT) reduce 94) ((T_EXTENDS) reduce 94) ((T_FILE) reduce 94) ((T_FINAL) reduce 94) ((T_FINALLY) reduce 94) ((T_FN) reduce 94) ((T_FOR) reduce 94) ((T_FOREACH) reduce 94) ((T_FUNCTION) reduce 94) ((T_FUNC_C) reduce 94) ((T_GLOBAL) reduce 94) ((T_GOTO) reduce 94) ((T_IF) reduce 94) ((T_IMPLEMENTS) reduce 94) ((T_INCLUDE) reduce 94) ((T_INCLUDE_ONCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_INSTEADOF) reduce 94) ((T_INTERFACE) reduce 94) ((T_ISSET) reduce 94) ((T_LINE) reduce 94) ((T_LIST) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_MATCH) reduce 94) ((T_METHOD_C) reduce 94) ((T_NAMESPACE) reduce 94) ((T_NEW) reduce 94) ((T_NS_C) reduce 94) ((T_PRINT) reduce 94) ((T_PRIVATE) reduce 94) ((T_PROPERTY_C) reduce 94) ((T_PROTECTED) reduce 94) ((T_PUBLIC) reduce 94) ((T_READONLY) reduce 94) ((T_REQUIRE) reduce 94) ((T_REQUIRE_ONCE) reduce 94) ((T_RETURN) reduce 94) ((T_STATIC) reduce 94) ((T_STRING) reduce 94) ((T_SWITCH) reduce 94) ((T_THROW) reduce 94) ((T_TRAIT) reduce 94) ((T_TRAIT_C) reduce 94) ((T_TRY) reduce 94) ((T_UNSET) reduce 94) ((T_USE) reduce 94) ((T_VAR) reduce 94) ((T_WHILE) reduce 94) ((T_YIELD) reduce 94) (("|") reduce 94)) 7353 (((T_ABSTRACT) reduce 93) ((T_ARRAY) reduce 93) ((T_AS) reduce 93) ((T_BREAK) reduce 93) ((T_CALLABLE) reduce 93) ((T_CASE) reduce 93) ((T_CATCH) reduce 93) ((T_CLASS) reduce 93) ((T_CLASS_C) reduce 93) ((T_CLONE) reduce 93) ((T_CONST) reduce 93) ((T_CONTINUE) reduce 93) ((T_DECLARE) reduce 93) ((T_DEFAULT) reduce 93) ((T_DIR) reduce 93) ((T_DO) reduce 93) ((T_ECHO) reduce 93) ((T_ELSE) reduce 93) ((T_ELSEIF) reduce 93) ((T_EMPTY) reduce 93) ((T_ENDDECLARE) reduce 93) ((T_ENDFOR) reduce 93) ((T_ENDFOREACH) reduce 93) ((T_ENDIF) reduce 93) ((T_ENDSWITCH) reduce 93) ((T_ENDWHILE) reduce 93) ((T_ENUM) reduce 93) ((T_EVAL) reduce 93) ((T_EXIT) reduce 93) ((T_EXTENDS) reduce 93) ((T_FILE) reduce 93) ((T_FINAL) reduce 93) ((T_FINALLY) reduce 93) ((T_FN) reduce 93) ((T_FOR) reduce 93) ((T_FOREACH) reduce 93) ((T_FUNCTION) reduce 93) ((T_FUNC_C) reduce 93) ((T_GLOBAL) reduce 93) ((T_GOTO) reduce 93) ((T_IF) reduce 93) ((T_IMPLEMENTS) reduce 93) ((T_INCLUDE) reduce 93) ((T_INCLUDE_ONCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_INSTEADOF) reduce 93) ((T_INTERFACE) reduce 93) ((T_ISSET) reduce 93) ((T_LINE) reduce 93) ((T_LIST) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_MATCH) reduce 93) ((T_METHOD_C) reduce 93) ((T_NAMESPACE) reduce 93) ((T_NEW) reduce 93) ((T_NS_C) reduce 93) ((T_PRINT) reduce 93) ((T_PRIVATE) reduce 93) ((T_PROPERTY_C) reduce 93) ((T_PROTECTED) reduce 93) ((T_PUBLIC) reduce 93) ((T_READONLY) reduce 93) ((T_REQUIRE) reduce 93) ((T_REQUIRE_ONCE) reduce 93) ((T_RETURN) reduce 93) ((T_STATIC) reduce 93) ((T_STRING) reduce 93) ((T_SWITCH) reduce 93) ((T_THROW) reduce 93) ((T_TRAIT) reduce 93) ((T_TRAIT_C) reduce 93) ((T_TRY) reduce 93) ((T_UNSET) reduce 93) ((T_USE) reduce 93) ((T_VAR) reduce 93) ((T_WHILE) reduce 93) ((T_YIELD) reduce 93) (("|") reduce 93)) 7354 (((T_ABSTRACT) reduce 95) ((T_ARRAY) reduce 95) ((T_AS) reduce 95) ((T_BREAK) reduce 95) ((T_CALLABLE) reduce 95) ((T_CASE) reduce 95) ((T_CATCH) reduce 95) ((T_CLASS) reduce 95) ((T_CLASS_C) reduce 95) ((T_CLONE) reduce 95) ((T_CONST) reduce 95) ((T_CONTINUE) reduce 95) ((T_DECLARE) reduce 95) ((T_DEFAULT) reduce 95) ((T_DIR) reduce 95) ((T_DO) reduce 95) ((T_ECHO) reduce 95) ((T_ELSE) reduce 95) ((T_ELSEIF) reduce 95) ((T_EMPTY) reduce 95) ((T_ENDDECLARE) reduce 95) ((T_ENDFOR) reduce 95) ((T_ENDFOREACH) reduce 95) ((T_ENDIF) reduce 95) ((T_ENDSWITCH) reduce 95) ((T_ENDWHILE) reduce 95) ((T_ENUM) reduce 95) ((T_EVAL) reduce 95) ((T_EXIT) reduce 95) ((T_EXTENDS) reduce 95) ((T_FILE) reduce 95) ((T_FINAL) reduce 95) ((T_FINALLY) reduce 95) ((T_FN) reduce 95) ((T_FOR) reduce 95) ((T_FOREACH) reduce 95) ((T_FUNCTION) reduce 95) ((T_FUNC_C) reduce 95) ((T_GLOBAL) reduce 95) ((T_GOTO) reduce 95) ((T_IF) reduce 95) ((T_IMPLEMENTS) reduce 95) ((T_INCLUDE) reduce 95) ((T_INCLUDE_ONCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_INSTEADOF) reduce 95) ((T_INTERFACE) reduce 95) ((T_ISSET) reduce 95) ((T_LINE) reduce 95) ((T_LIST) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_MATCH) reduce 95) ((T_METHOD_C) reduce 95) ((T_NAMESPACE) reduce 95) ((T_NEW) reduce 95) ((T_NS_C) reduce 95) ((T_PRINT) reduce 95) ((T_PRIVATE) reduce 95) ((T_PROPERTY_C) reduce 95) ((T_PROTECTED) reduce 95) ((T_PUBLIC) reduce 95) ((T_READONLY) reduce 95) ((T_REQUIRE) reduce 95) ((T_REQUIRE_ONCE) reduce 95) ((T_RETURN) reduce 95) ((T_STATIC) reduce 95) ((T_STRING) reduce 95) ((T_SWITCH) reduce 95) ((T_THROW) reduce 95) ((T_TRAIT) reduce 95) ((T_TRAIT_C) reduce 95) ((T_TRY) reduce 95) ((T_UNSET) reduce 95) ((T_USE) reduce 95) ((T_VAR) reduce 95) ((T_WHILE) reduce 95) ((T_YIELD) reduce 95) (("|") reduce 95)) 7355 (((T_ABSTRACT) reduce 269) ((T_ARRAY) reduce 269) ((T_AS) reduce 269) ((T_BREAK) reduce 269) ((T_CALLABLE) reduce 269) ((T_CASE) reduce 269) ((T_CATCH) reduce 269) ((T_CLASS) reduce 269) ((T_CLASS_C) reduce 269) ((T_CLONE) reduce 269) ((T_CONST) reduce 269) ((T_CONTINUE) reduce 269) ((T_DECLARE) reduce 269) ((T_DEFAULT) reduce 269) ((T_DIR) reduce 269) ((T_DO) reduce 269) ((T_ECHO) reduce 269) ((T_ELSE) reduce 269) ((T_ELSEIF) reduce 269) ((T_EMPTY) reduce 269) ((T_ENDDECLARE) reduce 269) ((T_ENDFOR) reduce 269) ((T_ENDFOREACH) reduce 269) ((T_ENDIF) reduce 269) ((T_ENDSWITCH) reduce 269) ((T_ENDWHILE) reduce 269) ((T_ENUM) reduce 269) ((T_EVAL) reduce 269) ((T_EXIT) reduce 269) ((T_EXTENDS) reduce 269) ((T_FILE) reduce 269) ((T_FINAL) reduce 269) ((T_FINALLY) reduce 269) ((T_FN) reduce 269) ((T_FOR) reduce 269) ((T_FOREACH) reduce 269) ((T_FUNCTION) reduce 269) ((T_FUNC_C) reduce 269) ((T_GLOBAL) reduce 269) ((T_GOTO) reduce 269) ((T_IF) reduce 269) ((T_IMPLEMENTS) reduce 269) ((T_INCLUDE) reduce 269) ((T_INCLUDE_ONCE) reduce 269) ((T_INSTANCEOF) reduce 269) ((T_INSTEADOF) reduce 269) ((T_INTERFACE) reduce 269) ((T_ISSET) reduce 269) ((T_LINE) reduce 269) ((T_LIST) reduce 269) ((T_LOGICAL_AND) reduce 269) ((T_LOGICAL_OR) reduce 269) ((T_LOGICAL_XOR) reduce 269) ((T_MATCH) reduce 269) ((T_METHOD_C) reduce 269) ((T_NAMESPACE) reduce 269) ((T_NEW) reduce 269) ((T_NS_C) reduce 269) ((T_PRINT) reduce 269) ((T_PRIVATE) reduce 269) ((T_PROPERTY_C) reduce 269) ((T_PROTECTED) reduce 269) ((T_PUBLIC) reduce 269) ((T_READONLY) reduce 269) ((T_REQUIRE) reduce 269) ((T_REQUIRE_ONCE) reduce 269) ((T_RETURN) reduce 269) ((T_STATIC) reduce 269) ((T_STRING) reduce 269) ((T_SWITCH) reduce 269) ((T_THROW) reduce 269) ((T_TRAIT) reduce 269) ((T_TRAIT_C) reduce 269) ((T_TRY) reduce 269) ((T_UNSET) reduce 269) ((T_USE) reduce 269) ((T_VAR) reduce 269) ((T_WHILE) reduce 269) ((T_YIELD) reduce 269) (("|") reduce 269)) 7356 (((T_ABSTRACT) reduce 92) ((T_ARRAY) reduce 92) ((T_AS) reduce 92) ((T_BREAK) reduce 92) ((T_CALLABLE) reduce 92) ((T_CASE) reduce 92) ((T_CATCH) reduce 92) ((T_CLASS) reduce 92) ((T_CLASS_C) reduce 92) ((T_CLONE) reduce 92) ((T_CONST) reduce 92) ((T_CONTINUE) reduce 92) ((T_DECLARE) reduce 92) ((T_DEFAULT) reduce 92) ((T_DIR) reduce 92) ((T_DO) reduce 92) ((T_ECHO) reduce 92) ((T_ELSE) reduce 92) ((T_ELSEIF) reduce 92) ((T_EMPTY) reduce 92) ((T_ENDDECLARE) reduce 92) ((T_ENDFOR) reduce 92) ((T_ENDFOREACH) reduce 92) ((T_ENDIF) reduce 92) ((T_ENDSWITCH) reduce 92) ((T_ENDWHILE) reduce 92) ((T_ENUM) reduce 92) ((T_EVAL) reduce 92) ((T_EXIT) reduce 92) ((T_EXTENDS) reduce 92) ((T_FILE) reduce 92) ((T_FINAL) reduce 92) ((T_FINALLY) reduce 92) ((T_FN) reduce 92) ((T_FOR) reduce 92) ((T_FOREACH) reduce 92) ((T_FUNCTION) reduce 92) ((T_FUNC_C) reduce 92) ((T_GLOBAL) reduce 92) ((T_GOTO) reduce 92) ((T_IF) reduce 92) ((T_IMPLEMENTS) reduce 92) ((T_INCLUDE) reduce 92) ((T_INCLUDE_ONCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_INSTEADOF) reduce 92) ((T_INTERFACE) reduce 92) ((T_ISSET) reduce 92) ((T_LINE) reduce 92) ((T_LIST) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_MATCH) reduce 92) ((T_METHOD_C) reduce 92) ((T_NAMESPACE) reduce 92) ((T_NEW) reduce 92) ((T_NS_C) reduce 92) ((T_PRINT) reduce 92) ((T_PRIVATE) reduce 92) ((T_PROPERTY_C) reduce 92) ((T_PROTECTED) reduce 92) ((T_PUBLIC) reduce 92) ((T_READONLY) reduce 92) ((T_REQUIRE) reduce 92) ((T_REQUIRE_ONCE) reduce 92) ((T_RETURN) reduce 92) ((T_STATIC) reduce 92) ((T_STRING) reduce 92) ((T_SWITCH) reduce 92) ((T_THROW) reduce 92) ((T_TRAIT) reduce 92) ((T_TRAIT_C) reduce 92) ((T_TRY) reduce 92) ((T_UNSET) reduce 92) ((T_USE) reduce 92) ((T_VAR) reduce 92) ((T_WHILE) reduce 92) ((T_YIELD) reduce 92) (("|") reduce 92)) 7357 (((T_ABSTRACT) reduce 282) ((T_ARRAY) reduce 282) ((T_AS) reduce 282) ((T_BREAK) reduce 282) ((T_CALLABLE) reduce 282) ((T_CASE) reduce 282) ((T_CATCH) reduce 282) ((T_CLASS) reduce 282) ((T_CLASS_C) reduce 282) ((T_CLONE) reduce 282) ((T_CONST) reduce 282) ((T_CONTINUE) reduce 282) ((T_DECLARE) reduce 282) ((T_DEFAULT) reduce 282) ((T_DIR) reduce 282) ((T_DO) reduce 282) ((T_ECHO) reduce 282) ((T_ELSE) reduce 282) ((T_ELSEIF) reduce 282) ((T_EMPTY) reduce 282) ((T_ENDDECLARE) reduce 282) ((T_ENDFOR) reduce 282) ((T_ENDFOREACH) reduce 282) ((T_ENDIF) reduce 282) ((T_ENDSWITCH) reduce 282) ((T_ENDWHILE) reduce 282) ((T_ENUM) reduce 282) ((T_EVAL) reduce 282) ((T_EXIT) reduce 282) ((T_EXTENDS) reduce 282) ((T_FILE) reduce 282) ((T_FINAL) reduce 282) ((T_FINALLY) reduce 282) ((T_FN) reduce 282) ((T_FOR) reduce 282) ((T_FOREACH) reduce 282) ((T_FUNCTION) reduce 282) ((T_FUNC_C) reduce 282) ((T_GLOBAL) reduce 282) ((T_GOTO) reduce 282) ((T_IF) reduce 282) ((T_IMPLEMENTS) reduce 282) ((T_INCLUDE) reduce 282) ((T_INCLUDE_ONCE) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_INSTEADOF) reduce 282) ((T_INTERFACE) reduce 282) ((T_ISSET) reduce 282) ((T_LINE) reduce 282) ((T_LIST) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_MATCH) reduce 282) ((T_METHOD_C) reduce 282) ((T_NAMESPACE) reduce 282) ((T_NEW) reduce 282) ((T_NS_C) reduce 282) ((T_PRINT) reduce 282) ((T_PRIVATE) reduce 282) ((T_PROPERTY_C) reduce 282) ((T_PROTECTED) reduce 282) ((T_PUBLIC) reduce 282) ((T_READONLY) reduce 282) ((T_REQUIRE) reduce 282) ((T_REQUIRE_ONCE) reduce 282) ((T_RETURN) reduce 282) ((T_STATIC) reduce 282) ((T_STRING) reduce 282) ((T_SWITCH) reduce 282) ((T_THROW) reduce 282) ((T_TRAIT) reduce 282) ((T_TRAIT_C) reduce 282) ((T_TRY) reduce 282) ((T_UNSET) reduce 282) ((T_USE) reduce 282) ((T_VAR) reduce 282) ((T_WHILE) reduce 282) ((T_YIELD) reduce 282) (("|") reduce 282)) 7358 (((T_ABSTRACT) reduce 270) ((T_ARRAY) reduce 270) ((T_AS) reduce 270) ((T_BREAK) reduce 270) ((T_CALLABLE) reduce 270) ((T_CASE) reduce 270) ((T_CATCH) reduce 270) ((T_CLASS) reduce 270) ((T_CLASS_C) reduce 270) ((T_CLONE) reduce 270) ((T_CONST) reduce 270) ((T_CONTINUE) reduce 270) ((T_DECLARE) reduce 270) ((T_DEFAULT) reduce 270) ((T_DIR) reduce 270) ((T_DO) reduce 270) ((T_ECHO) reduce 270) ((T_ELSE) reduce 270) ((T_ELSEIF) reduce 270) ((T_EMPTY) reduce 270) ((T_ENDDECLARE) reduce 270) ((T_ENDFOR) reduce 270) ((T_ENDFOREACH) reduce 270) ((T_ENDIF) reduce 270) ((T_ENDSWITCH) reduce 270) ((T_ENDWHILE) reduce 270) ((T_ENUM) reduce 270) ((T_EVAL) reduce 270) ((T_EXIT) reduce 270) ((T_EXTENDS) reduce 270) ((T_FILE) reduce 270) ((T_FINAL) reduce 270) ((T_FINALLY) reduce 270) ((T_FN) reduce 270) ((T_FOR) reduce 270) ((T_FOREACH) reduce 270) ((T_FUNCTION) reduce 270) ((T_FUNC_C) reduce 270) ((T_GLOBAL) reduce 270) ((T_GOTO) reduce 270) ((T_IF) reduce 270) ((T_IMPLEMENTS) reduce 270) ((T_INCLUDE) reduce 270) ((T_INCLUDE_ONCE) reduce 270) ((T_INSTANCEOF) reduce 270) ((T_INSTEADOF) reduce 270) ((T_INTERFACE) reduce 270) ((T_ISSET) reduce 270) ((T_LINE) reduce 270) ((T_LIST) reduce 270) ((T_LOGICAL_AND) reduce 270) ((T_LOGICAL_OR) reduce 270) ((T_LOGICAL_XOR) reduce 270) ((T_MATCH) reduce 270) ((T_METHOD_C) reduce 270) ((T_NAMESPACE) reduce 270) ((T_NEW) reduce 270) ((T_NS_C) reduce 270) ((T_PRINT) reduce 270) ((T_PRIVATE) reduce 270) ((T_PROPERTY_C) reduce 270) ((T_PROTECTED) reduce 270) ((T_PUBLIC) reduce 270) ((T_READONLY) reduce 270) ((T_REQUIRE) reduce 270) ((T_REQUIRE_ONCE) reduce 270) ((T_RETURN) reduce 270) ((T_STATIC) reduce 270) ((T_STRING) reduce 270) ((T_SWITCH) reduce 270) ((T_THROW) reduce 270) ((T_TRAIT) reduce 270) ((T_TRAIT_C) reduce 270) ((T_TRY) reduce 270) ((T_UNSET) reduce 270) ((T_USE) reduce 270) ((T_VAR) reduce 270) ((T_WHILE) reduce 270) ((T_YIELD) reduce 270) (("|") reduce 270)) 7359 (((T_ABSTRACT) reduce 268) ((T_ARRAY) reduce 268) ((T_AS) reduce 268) ((T_BREAK) reduce 268) ((T_CALLABLE) reduce 268) ((T_CASE) reduce 268) ((T_CATCH) reduce 268) ((T_CLASS) reduce 268) ((T_CLASS_C) reduce 268) ((T_CLONE) reduce 268) ((T_CONST) reduce 268) ((T_CONTINUE) reduce 268) ((T_DECLARE) reduce 268) ((T_DEFAULT) reduce 268) ((T_DIR) reduce 268) ((T_DO) reduce 268) ((T_ECHO) reduce 268) ((T_ELSE) reduce 268) ((T_ELSEIF) reduce 268) ((T_EMPTY) reduce 268) ((T_ENDDECLARE) reduce 268) ((T_ENDFOR) reduce 268) ((T_ENDFOREACH) reduce 268) ((T_ENDIF) reduce 268) ((T_ENDSWITCH) reduce 268) ((T_ENDWHILE) reduce 268) ((T_ENUM) reduce 268) ((T_EVAL) reduce 268) ((T_EXIT) reduce 268) ((T_EXTENDS) reduce 268) ((T_FILE) reduce 268) ((T_FINAL) reduce 268) ((T_FINALLY) reduce 268) ((T_FN) reduce 268) ((T_FOR) reduce 268) ((T_FOREACH) reduce 268) ((T_FUNCTION) reduce 268) ((T_FUNC_C) reduce 268) ((T_GLOBAL) reduce 268) ((T_GOTO) reduce 268) ((T_IF) reduce 268) ((T_IMPLEMENTS) reduce 268) ((T_INCLUDE) reduce 268) ((T_INCLUDE_ONCE) reduce 268) ((T_INSTANCEOF) reduce 268) ((T_INSTEADOF) reduce 268) ((T_INTERFACE) reduce 268) ((T_ISSET) reduce 268) ((T_LINE) reduce 268) ((T_LIST) reduce 268) ((T_LOGICAL_AND) reduce 268) ((T_LOGICAL_OR) reduce 268) ((T_LOGICAL_XOR) reduce 268) ((T_MATCH) reduce 268) ((T_METHOD_C) reduce 268) ((T_NAMESPACE) reduce 268) ((T_NEW) reduce 268) ((T_NS_C) reduce 268) ((T_PRINT) reduce 268) ((T_PRIVATE) reduce 268) ((T_PROPERTY_C) reduce 268) ((T_PROTECTED) reduce 268) ((T_PUBLIC) reduce 268) ((T_READONLY) reduce 268) ((T_REQUIRE) reduce 268) ((T_REQUIRE_ONCE) reduce 268) ((T_RETURN) reduce 268) ((T_STATIC) reduce 268) ((T_STRING) reduce 268) ((T_SWITCH) reduce 268) ((T_THROW) reduce 268) ((T_TRAIT) reduce 268) ((T_TRAIT_C) reduce 268) ((T_TRY) reduce 268) ((T_UNSET) reduce 268) ((T_USE) reduce 268) ((T_VAR) reduce 268) ((T_WHILE) reduce 268) ((T_YIELD) reduce 268) (("|") reduce 268)) 7360 (((T_ABSTRACT) reduce 272) ((T_ARRAY) reduce 272) ((T_AS) reduce 272) ((T_BREAK) reduce 272) ((T_CALLABLE) reduce 272) ((T_CASE) reduce 272) ((T_CATCH) reduce 272) ((T_CLASS) reduce 272) ((T_CLASS_C) reduce 272) ((T_CLONE) reduce 272) ((T_CONST) reduce 272) ((T_CONTINUE) reduce 272) ((T_DECLARE) reduce 272) ((T_DEFAULT) reduce 272) ((T_DIR) reduce 272) ((T_DO) reduce 272) ((T_ECHO) reduce 272) ((T_ELSE) reduce 272) ((T_ELSEIF) reduce 272) ((T_EMPTY) reduce 272) ((T_ENDDECLARE) reduce 272) ((T_ENDFOR) reduce 272) ((T_ENDFOREACH) reduce 272) ((T_ENDIF) reduce 272) ((T_ENDSWITCH) reduce 272) ((T_ENDWHILE) reduce 272) ((T_ENUM) reduce 272) ((T_EVAL) reduce 272) ((T_EXIT) reduce 272) ((T_EXTENDS) reduce 272) ((T_FILE) reduce 272) ((T_FINAL) reduce 272) ((T_FINALLY) reduce 272) ((T_FN) reduce 272) ((T_FOR) reduce 272) ((T_FOREACH) reduce 272) ((T_FUNCTION) reduce 272) ((T_FUNC_C) reduce 272) ((T_GLOBAL) reduce 272) ((T_GOTO) reduce 272) ((T_IF) reduce 272) ((T_IMPLEMENTS) reduce 272) ((T_INCLUDE) reduce 272) ((T_INCLUDE_ONCE) reduce 272) ((T_INSTANCEOF) reduce 272) ((T_INSTEADOF) reduce 272) ((T_INTERFACE) reduce 272) ((T_ISSET) reduce 272) ((T_LINE) reduce 272) ((T_LIST) reduce 272) ((T_LOGICAL_AND) reduce 272) ((T_LOGICAL_OR) reduce 272) ((T_LOGICAL_XOR) reduce 272) ((T_MATCH) reduce 272) ((T_METHOD_C) reduce 272) ((T_NAMESPACE) reduce 272) ((T_NEW) reduce 272) ((T_NS_C) reduce 272) ((T_PRINT) reduce 272) ((T_PRIVATE) reduce 272) ((T_PROPERTY_C) reduce 272) ((T_PROTECTED) reduce 272) ((T_PUBLIC) reduce 272) ((T_READONLY) reduce 272) ((T_REQUIRE) reduce 272) ((T_REQUIRE_ONCE) reduce 272) ((T_RETURN) reduce 272) ((T_STATIC) reduce 272) ((T_STRING) reduce 272) ((T_SWITCH) reduce 272) ((T_THROW) reduce 272) ((T_TRAIT) reduce 272) ((T_TRAIT_C) reduce 272) ((T_TRY) reduce 272) ((T_UNSET) reduce 272) ((T_USE) reduce 272) ((T_VAR) reduce 272) ((T_WHILE) reduce 272) ((T_YIELD) reduce 272) (("|") reduce 272)) 7362 (((T_ABSTRACT) reduce 271) ((T_ARRAY) reduce 271) ((T_AS) reduce 271) ((T_BREAK) reduce 271) ((T_CALLABLE) reduce 271) ((T_CASE) reduce 271) ((T_CATCH) reduce 271) ((T_CLASS) reduce 271) ((T_CLASS_C) reduce 271) ((T_CLONE) reduce 271) ((T_CONST) reduce 271) ((T_CONTINUE) reduce 271) ((T_DECLARE) reduce 271) ((T_DEFAULT) reduce 271) ((T_DIR) reduce 271) ((T_DO) reduce 271) ((T_ECHO) reduce 271) ((T_ELSE) reduce 271) ((T_ELSEIF) reduce 271) ((T_EMPTY) reduce 271) ((T_ENDDECLARE) reduce 271) ((T_ENDFOR) reduce 271) ((T_ENDFOREACH) reduce 271) ((T_ENDIF) reduce 271) ((T_ENDSWITCH) reduce 271) ((T_ENDWHILE) reduce 271) ((T_ENUM) reduce 271) ((T_EVAL) reduce 271) ((T_EXIT) reduce 271) ((T_EXTENDS) reduce 271) ((T_FILE) reduce 271) ((T_FINAL) reduce 271) ((T_FINALLY) reduce 271) ((T_FN) reduce 271) ((T_FOR) reduce 271) ((T_FOREACH) reduce 271) ((T_FUNCTION) reduce 271) ((T_FUNC_C) reduce 271) ((T_GLOBAL) reduce 271) ((T_GOTO) reduce 271) ((T_IF) reduce 271) ((T_IMPLEMENTS) reduce 271) ((T_INCLUDE) reduce 271) ((T_INCLUDE_ONCE) reduce 271) ((T_INSTANCEOF) reduce 271) ((T_INSTEADOF) reduce 271) ((T_INTERFACE) reduce 271) ((T_ISSET) reduce 271) ((T_LINE) reduce 271) ((T_LIST) reduce 271) ((T_LOGICAL_AND) reduce 271) ((T_LOGICAL_OR) reduce 271) ((T_LOGICAL_XOR) reduce 271) ((T_MATCH) reduce 271) ((T_METHOD_C) reduce 271) ((T_NAMESPACE) reduce 271) ((T_NEW) reduce 271) ((T_NS_C) reduce 271) ((T_PRINT) reduce 271) ((T_PRIVATE) reduce 271) ((T_PROPERTY_C) reduce 271) ((T_PROTECTED) reduce 271) ((T_PUBLIC) reduce 271) ((T_READONLY) reduce 271) ((T_REQUIRE) reduce 271) ((T_REQUIRE_ONCE) reduce 271) ((T_RETURN) reduce 271) ((T_STATIC) reduce 271) ((T_STRING) reduce 271) ((T_SWITCH) reduce 271) ((T_THROW) reduce 271) ((T_TRAIT) reduce 271) ((T_TRAIT_C) reduce 271) ((T_TRY) reduce 271) ((T_UNSET) reduce 271) ((T_USE) reduce 271) ((T_VAR) reduce 271) ((T_WHILE) reduce 271) ((T_YIELD) reduce 271) (("|") reduce 271)) 7364 (((T_ABSTRACT) reduce 273) ((T_ARRAY) reduce 273) ((T_AS) reduce 273) ((T_BREAK) reduce 273) ((T_CALLABLE) reduce 273) ((T_CASE) reduce 273) ((T_CATCH) reduce 273) ((T_CLASS) reduce 273) ((T_CLASS_C) reduce 273) ((T_CLONE) reduce 273) ((T_CONST) reduce 273) ((T_CONTINUE) reduce 273) ((T_DECLARE) reduce 273) ((T_DEFAULT) reduce 273) ((T_DIR) reduce 273) ((T_DO) reduce 273) ((T_ECHO) reduce 273) ((T_ELSE) reduce 273) ((T_ELSEIF) reduce 273) ((T_EMPTY) reduce 273) ((T_ENDDECLARE) reduce 273) ((T_ENDFOR) reduce 273) ((T_ENDFOREACH) reduce 273) ((T_ENDIF) reduce 273) ((T_ENDSWITCH) reduce 273) ((T_ENDWHILE) reduce 273) ((T_ENUM) reduce 273) ((T_EVAL) reduce 273) ((T_EXIT) reduce 273) ((T_EXTENDS) reduce 273) ((T_FILE) reduce 273) ((T_FINAL) reduce 273) ((T_FINALLY) reduce 273) ((T_FN) reduce 273) ((T_FOR) reduce 273) ((T_FOREACH) reduce 273) ((T_FUNCTION) reduce 273) ((T_FUNC_C) reduce 273) ((T_GLOBAL) reduce 273) ((T_GOTO) reduce 273) ((T_IF) reduce 273) ((T_IMPLEMENTS) reduce 273) ((T_INCLUDE) reduce 273) ((T_INCLUDE_ONCE) reduce 273) ((T_INSTANCEOF) reduce 273) ((T_INSTEADOF) reduce 273) ((T_INTERFACE) reduce 273) ((T_ISSET) reduce 273) ((T_LINE) reduce 273) ((T_LIST) reduce 273) ((T_LOGICAL_AND) reduce 273) ((T_LOGICAL_OR) reduce 273) ((T_LOGICAL_XOR) reduce 273) ((T_MATCH) reduce 273) ((T_METHOD_C) reduce 273) ((T_NAMESPACE) reduce 273) ((T_NEW) reduce 273) ((T_NS_C) reduce 273) ((T_PRINT) reduce 273) ((T_PRIVATE) reduce 273) ((T_PROPERTY_C) reduce 273) ((T_PROTECTED) reduce 273) ((T_PUBLIC) reduce 273) ((T_READONLY) reduce 273) ((T_REQUIRE) reduce 273) ((T_REQUIRE_ONCE) reduce 273) ((T_RETURN) reduce 273) ((T_STATIC) reduce 273) ((T_STRING) reduce 273) ((T_SWITCH) reduce 273) ((T_THROW) reduce 273) ((T_TRAIT) reduce 273) ((T_TRAIT_C) reduce 273) ((T_TRY) reduce 273) ((T_UNSET) reduce 273) ((T_USE) reduce 273) ((T_VAR) reduce 273) ((T_WHILE) reduce 273) ((T_YIELD) reduce 273) (("|") reduce 273)) 7365 ((("=") reduce 52)) 7366 ((("=") reduce 53)) 7367 ((("=") reduce 73)) 7371 (((T_ABSTRACT) reduce 319) ((T_ATTRIBUTE) reduce 319) ((T_CASE) reduce 319) ((T_CONST) reduce 319) ((T_FINAL) reduce 319) ((T_FUNCTION) reduce 319) ((T_PRIVATE) reduce 319) ((T_PRIVATE_SET) reduce 319) ((T_PROTECTED) reduce 319) ((T_PROTECTED_SET) reduce 319) ((T_PUBLIC) reduce 319) ((T_PUBLIC_SET) reduce 319) ((T_READONLY) reduce 319) ((T_STATIC) reduce 319) ((T_USE) reduce 319) ((T_VAR) reduce 319) (("}") reduce 319)) 7372 (((",") reduce 381) ((";") reduce 381)) 7375 (((",") reduce 383) ((";") reduce 383)) 7377 (((T_ABSTRACT) reduce 280) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 280) ((T_ARRAY) reduce 280) ((T_AS) reduce 280) ((T_BREAK) reduce 280) ((T_CALLABLE) reduce 280) ((T_CASE) reduce 280) ((T_CATCH) reduce 280) ((T_CLASS) reduce 280) ((T_CLASS_C) reduce 280) ((T_CLONE) reduce 280) ((T_CONST) reduce 280) ((T_CONTINUE) reduce 280) ((T_DECLARE) reduce 280) ((T_DEFAULT) reduce 280) ((T_DIR) reduce 280) ((T_DO) reduce 280) ((T_ECHO) reduce 280) ((T_ELSE) reduce 280) ((T_ELSEIF) reduce 280) ((T_EMPTY) reduce 280) ((T_ENDDECLARE) reduce 280) ((T_ENDFOR) reduce 280) ((T_ENDFOREACH) reduce 280) ((T_ENDIF) reduce 280) ((T_ENDSWITCH) reduce 280) ((T_ENDWHILE) reduce 280) ((T_ENUM) reduce 280) ((T_EVAL) reduce 280) ((T_EXIT) reduce 280) ((T_EXTENDS) reduce 280) ((T_FILE) reduce 280) ((T_FINAL) reduce 280) ((T_FINALLY) reduce 280) ((T_FN) reduce 280) ((T_FOR) reduce 280) ((T_FOREACH) reduce 280) ((T_FUNCTION) reduce 280) ((T_FUNC_C) reduce 280) ((T_GLOBAL) reduce 280) ((T_GOTO) reduce 280) ((T_IF) reduce 280) ((T_IMPLEMENTS) reduce 280) ((T_INCLUDE) reduce 280) ((T_INCLUDE_ONCE) reduce 280) ((T_INSTANCEOF) reduce 280) ((T_INSTEADOF) reduce 280) ((T_INTERFACE) reduce 280) ((T_ISSET) reduce 280) ((T_LINE) reduce 280) ((T_LIST) reduce 280) ((T_LOGICAL_AND) reduce 280) ((T_LOGICAL_OR) reduce 280) ((T_LOGICAL_XOR) reduce 280) ((T_MATCH) reduce 280) ((T_METHOD_C) reduce 280) ((T_NAMESPACE) reduce 280) ((T_NEW) reduce 280) ((T_NS_C) reduce 280) ((T_PRINT) reduce 280) ((T_PRIVATE) reduce 280) ((T_PROPERTY_C) reduce 280) ((T_PROTECTED) reduce 280) ((T_PUBLIC) reduce 280) ((T_READONLY) reduce 280) ((T_REQUIRE) reduce 280) ((T_REQUIRE_ONCE) reduce 280) ((T_RETURN) reduce 280) ((T_STATIC) reduce 280) ((T_STRING) reduce 280) ((T_SWITCH) reduce 280) ((T_THROW) reduce 280) ((T_TRAIT) reduce 280) ((T_TRAIT_C) reduce 280) ((T_TRY) reduce 280) ((T_UNSET) reduce 280) ((T_USE) reduce 280) ((T_VAR) reduce 280) ((T_WHILE) reduce 280) ((T_YIELD) reduce 280)) 7378 (((T_ABSTRACT) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_ARRAY) reduce 281) ((T_AS) reduce 281) ((T_BREAK) reduce 281) ((T_CALLABLE) reduce 281) ((T_CASE) reduce 281) ((T_CATCH) reduce 281) ((T_CLASS) reduce 281) ((T_CLASS_C) reduce 281) ((T_CLONE) reduce 281) ((T_CONST) reduce 281) ((T_CONTINUE) reduce 281) ((T_DECLARE) reduce 281) ((T_DEFAULT) reduce 281) ((T_DIR) reduce 281) ((T_DO) reduce 281) ((T_ECHO) reduce 281) ((T_ELSE) reduce 281) ((T_ELSEIF) reduce 281) ((T_EMPTY) reduce 281) ((T_ENDDECLARE) reduce 281) ((T_ENDFOR) reduce 281) ((T_ENDFOREACH) reduce 281) ((T_ENDIF) reduce 281) ((T_ENDSWITCH) reduce 281) ((T_ENDWHILE) reduce 281) ((T_ENUM) reduce 281) ((T_EVAL) reduce 281) ((T_EXIT) reduce 281) ((T_EXTENDS) reduce 281) ((T_FILE) reduce 281) ((T_FINAL) reduce 281) ((T_FINALLY) reduce 281) ((T_FN) reduce 281) ((T_FOR) reduce 281) ((T_FOREACH) reduce 281) ((T_FUNCTION) reduce 281) ((T_FUNC_C) reduce 281) ((T_GLOBAL) reduce 281) ((T_GOTO) reduce 281) ((T_IF) reduce 281) ((T_IMPLEMENTS) reduce 281) ((T_INCLUDE) reduce 281) ((T_INCLUDE_ONCE) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_INSTEADOF) reduce 281) ((T_INTERFACE) reduce 281) ((T_ISSET) reduce 281) ((T_LINE) reduce 281) ((T_LIST) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_MATCH) reduce 281) ((T_METHOD_C) reduce 281) ((T_NAMESPACE) reduce 281) ((T_NEW) reduce 281) ((T_NS_C) reduce 281) ((T_PRINT) reduce 281) ((T_PRIVATE) reduce 281) ((T_PROPERTY_C) reduce 281) ((T_PROTECTED) reduce 281) ((T_PUBLIC) reduce 281) ((T_READONLY) reduce 281) ((T_REQUIRE) reduce 281) ((T_REQUIRE_ONCE) reduce 281) ((T_RETURN) reduce 281) ((T_STATIC) reduce 281) ((T_STRING) reduce 281) ((T_SWITCH) reduce 281) ((T_THROW) reduce 281) ((T_TRAIT) reduce 281) ((T_TRAIT_C) reduce 281) ((T_TRY) reduce 281) ((T_UNSET) reduce 281) ((T_USE) reduce 281) ((T_VAR) reduce 281) ((T_WHILE) reduce 281) ((T_YIELD) reduce 281)) 7379 (((T_ABSTRACT) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_ARRAY) reduce 94) ((T_AS) reduce 94) ((T_BREAK) reduce 94) ((T_CALLABLE) reduce 94) ((T_CASE) reduce 94) ((T_CATCH) reduce 94) ((T_CLASS) reduce 94) ((T_CLASS_C) reduce 94) ((T_CLONE) reduce 94) ((T_CONST) reduce 94) ((T_CONTINUE) reduce 94) ((T_DECLARE) reduce 94) ((T_DEFAULT) reduce 94) ((T_DIR) reduce 94) ((T_DO) reduce 94) ((T_ECHO) reduce 94) ((T_ELSE) reduce 94) ((T_ELSEIF) reduce 94) ((T_EMPTY) reduce 94) ((T_ENDDECLARE) reduce 94) ((T_ENDFOR) reduce 94) ((T_ENDFOREACH) reduce 94) ((T_ENDIF) reduce 94) ((T_ENDSWITCH) reduce 94) ((T_ENDWHILE) reduce 94) ((T_ENUM) reduce 94) ((T_EVAL) reduce 94) ((T_EXIT) reduce 94) ((T_EXTENDS) reduce 94) ((T_FILE) reduce 94) ((T_FINAL) reduce 94) ((T_FINALLY) reduce 94) ((T_FN) reduce 94) ((T_FOR) reduce 94) ((T_FOREACH) reduce 94) ((T_FUNCTION) reduce 94) ((T_FUNC_C) reduce 94) ((T_GLOBAL) reduce 94) ((T_GOTO) reduce 94) ((T_IF) reduce 94) ((T_IMPLEMENTS) reduce 94) ((T_INCLUDE) reduce 94) ((T_INCLUDE_ONCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_INSTEADOF) reduce 94) ((T_INTERFACE) reduce 94) ((T_ISSET) reduce 94) ((T_LINE) reduce 94) ((T_LIST) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_MATCH) reduce 94) ((T_METHOD_C) reduce 94) ((T_NAMESPACE) reduce 94) ((T_NEW) reduce 94) ((T_NS_C) reduce 94) ((T_PRINT) reduce 94) ((T_PRIVATE) reduce 94) ((T_PROPERTY_C) reduce 94) ((T_PROTECTED) reduce 94) ((T_PUBLIC) reduce 94) ((T_READONLY) reduce 94) ((T_REQUIRE) reduce 94) ((T_REQUIRE_ONCE) reduce 94) ((T_RETURN) reduce 94) ((T_STATIC) reduce 94) ((T_STRING) reduce 94) ((T_SWITCH) reduce 94) ((T_THROW) reduce 94) ((T_TRAIT) reduce 94) ((T_TRAIT_C) reduce 94) ((T_TRY) reduce 94) ((T_UNSET) reduce 94) ((T_USE) reduce 94) ((T_VAR) reduce 94) ((T_WHILE) reduce 94) ((T_YIELD) reduce 94)) 7380 (((T_ABSTRACT) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_ARRAY) reduce 93) ((T_AS) reduce 93) ((T_BREAK) reduce 93) ((T_CALLABLE) reduce 93) ((T_CASE) reduce 93) ((T_CATCH) reduce 93) ((T_CLASS) reduce 93) ((T_CLASS_C) reduce 93) ((T_CLONE) reduce 93) ((T_CONST) reduce 93) ((T_CONTINUE) reduce 93) ((T_DECLARE) reduce 93) ((T_DEFAULT) reduce 93) ((T_DIR) reduce 93) ((T_DO) reduce 93) ((T_ECHO) reduce 93) ((T_ELSE) reduce 93) ((T_ELSEIF) reduce 93) ((T_EMPTY) reduce 93) ((T_ENDDECLARE) reduce 93) ((T_ENDFOR) reduce 93) ((T_ENDFOREACH) reduce 93) ((T_ENDIF) reduce 93) ((T_ENDSWITCH) reduce 93) ((T_ENDWHILE) reduce 93) ((T_ENUM) reduce 93) ((T_EVAL) reduce 93) ((T_EXIT) reduce 93) ((T_EXTENDS) reduce 93) ((T_FILE) reduce 93) ((T_FINAL) reduce 93) ((T_FINALLY) reduce 93) ((T_FN) reduce 93) ((T_FOR) reduce 93) ((T_FOREACH) reduce 93) ((T_FUNCTION) reduce 93) ((T_FUNC_C) reduce 93) ((T_GLOBAL) reduce 93) ((T_GOTO) reduce 93) ((T_IF) reduce 93) ((T_IMPLEMENTS) reduce 93) ((T_INCLUDE) reduce 93) ((T_INCLUDE_ONCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_INSTEADOF) reduce 93) ((T_INTERFACE) reduce 93) ((T_ISSET) reduce 93) ((T_LINE) reduce 93) ((T_LIST) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_MATCH) reduce 93) ((T_METHOD_C) reduce 93) ((T_NAMESPACE) reduce 93) ((T_NEW) reduce 93) ((T_NS_C) reduce 93) ((T_PRINT) reduce 93) ((T_PRIVATE) reduce 93) ((T_PROPERTY_C) reduce 93) ((T_PROTECTED) reduce 93) ((T_PUBLIC) reduce 93) ((T_READONLY) reduce 93) ((T_REQUIRE) reduce 93) ((T_REQUIRE_ONCE) reduce 93) ((T_RETURN) reduce 93) ((T_STATIC) reduce 93) ((T_STRING) reduce 93) ((T_SWITCH) reduce 93) ((T_THROW) reduce 93) ((T_TRAIT) reduce 93) ((T_TRAIT_C) reduce 93) ((T_TRY) reduce 93) ((T_UNSET) reduce 93) ((T_USE) reduce 93) ((T_VAR) reduce 93) ((T_WHILE) reduce 93) ((T_YIELD) reduce 93)) 7381 (((T_ABSTRACT) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_ARRAY) reduce 95) ((T_AS) reduce 95) ((T_BREAK) reduce 95) ((T_CALLABLE) reduce 95) ((T_CASE) reduce 95) ((T_CATCH) reduce 95) ((T_CLASS) reduce 95) ((T_CLASS_C) reduce 95) ((T_CLONE) reduce 95) ((T_CONST) reduce 95) ((T_CONTINUE) reduce 95) ((T_DECLARE) reduce 95) ((T_DEFAULT) reduce 95) ((T_DIR) reduce 95) ((T_DO) reduce 95) ((T_ECHO) reduce 95) ((T_ELSE) reduce 95) ((T_ELSEIF) reduce 95) ((T_EMPTY) reduce 95) ((T_ENDDECLARE) reduce 95) ((T_ENDFOR) reduce 95) ((T_ENDFOREACH) reduce 95) ((T_ENDIF) reduce 95) ((T_ENDSWITCH) reduce 95) ((T_ENDWHILE) reduce 95) ((T_ENUM) reduce 95) ((T_EVAL) reduce 95) ((T_EXIT) reduce 95) ((T_EXTENDS) reduce 95) ((T_FILE) reduce 95) ((T_FINAL) reduce 95) ((T_FINALLY) reduce 95) ((T_FN) reduce 95) ((T_FOR) reduce 95) ((T_FOREACH) reduce 95) ((T_FUNCTION) reduce 95) ((T_FUNC_C) reduce 95) ((T_GLOBAL) reduce 95) ((T_GOTO) reduce 95) ((T_IF) reduce 95) ((T_IMPLEMENTS) reduce 95) ((T_INCLUDE) reduce 95) ((T_INCLUDE_ONCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_INSTEADOF) reduce 95) ((T_INTERFACE) reduce 95) ((T_ISSET) reduce 95) ((T_LINE) reduce 95) ((T_LIST) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_MATCH) reduce 95) ((T_METHOD_C) reduce 95) ((T_NAMESPACE) reduce 95) ((T_NEW) reduce 95) ((T_NS_C) reduce 95) ((T_PRINT) reduce 95) ((T_PRIVATE) reduce 95) ((T_PROPERTY_C) reduce 95) ((T_PROTECTED) reduce 95) ((T_PUBLIC) reduce 95) ((T_READONLY) reduce 95) ((T_REQUIRE) reduce 95) ((T_REQUIRE_ONCE) reduce 95) ((T_RETURN) reduce 95) ((T_STATIC) reduce 95) ((T_STRING) reduce 95) ((T_SWITCH) reduce 95) ((T_THROW) reduce 95) ((T_TRAIT) reduce 95) ((T_TRAIT_C) reduce 95) ((T_TRY) reduce 95) ((T_UNSET) reduce 95) ((T_USE) reduce 95) ((T_VAR) reduce 95) ((T_WHILE) reduce 95) ((T_YIELD) reduce 95)) 7382 (((T_ABSTRACT) reduce 269) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 269) ((T_ARRAY) reduce 269) ((T_AS) reduce 269) ((T_BREAK) reduce 269) ((T_CALLABLE) reduce 269) ((T_CASE) reduce 269) ((T_CATCH) reduce 269) ((T_CLASS) reduce 269) ((T_CLASS_C) reduce 269) ((T_CLONE) reduce 269) ((T_CONST) reduce 269) ((T_CONTINUE) reduce 269) ((T_DECLARE) reduce 269) ((T_DEFAULT) reduce 269) ((T_DIR) reduce 269) ((T_DO) reduce 269) ((T_ECHO) reduce 269) ((T_ELSE) reduce 269) ((T_ELSEIF) reduce 269) ((T_EMPTY) reduce 269) ((T_ENDDECLARE) reduce 269) ((T_ENDFOR) reduce 269) ((T_ENDFOREACH) reduce 269) ((T_ENDIF) reduce 269) ((T_ENDSWITCH) reduce 269) ((T_ENDWHILE) reduce 269) ((T_ENUM) reduce 269) ((T_EVAL) reduce 269) ((T_EXIT) reduce 269) ((T_EXTENDS) reduce 269) ((T_FILE) reduce 269) ((T_FINAL) reduce 269) ((T_FINALLY) reduce 269) ((T_FN) reduce 269) ((T_FOR) reduce 269) ((T_FOREACH) reduce 269) ((T_FUNCTION) reduce 269) ((T_FUNC_C) reduce 269) ((T_GLOBAL) reduce 269) ((T_GOTO) reduce 269) ((T_IF) reduce 269) ((T_IMPLEMENTS) reduce 269) ((T_INCLUDE) reduce 269) ((T_INCLUDE_ONCE) reduce 269) ((T_INSTANCEOF) reduce 269) ((T_INSTEADOF) reduce 269) ((T_INTERFACE) reduce 269) ((T_ISSET) reduce 269) ((T_LINE) reduce 269) ((T_LIST) reduce 269) ((T_LOGICAL_AND) reduce 269) ((T_LOGICAL_OR) reduce 269) ((T_LOGICAL_XOR) reduce 269) ((T_MATCH) reduce 269) ((T_METHOD_C) reduce 269) ((T_NAMESPACE) reduce 269) ((T_NEW) reduce 269) ((T_NS_C) reduce 269) ((T_PRINT) reduce 269) ((T_PRIVATE) reduce 269) ((T_PROPERTY_C) reduce 269) ((T_PROTECTED) reduce 269) ((T_PUBLIC) reduce 269) ((T_READONLY) reduce 269) ((T_REQUIRE) reduce 269) ((T_REQUIRE_ONCE) reduce 269) ((T_RETURN) reduce 269) ((T_STATIC) reduce 269) ((T_STRING) reduce 269) ((T_SWITCH) reduce 269) ((T_THROW) reduce 269) ((T_TRAIT) reduce 269) ((T_TRAIT_C) reduce 269) ((T_TRY) reduce 269) ((T_UNSET) reduce 269) ((T_USE) reduce 269) ((T_VAR) reduce 269) ((T_WHILE) reduce 269) ((T_YIELD) reduce 269)) 7383 (((T_ABSTRACT) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_ARRAY) reduce 92) ((T_AS) reduce 92) ((T_BREAK) reduce 92) ((T_CALLABLE) reduce 92) ((T_CASE) reduce 92) ((T_CATCH) reduce 92) ((T_CLASS) reduce 92) ((T_CLASS_C) reduce 92) ((T_CLONE) reduce 92) ((T_CONST) reduce 92) ((T_CONTINUE) reduce 92) ((T_DECLARE) reduce 92) ((T_DEFAULT) reduce 92) ((T_DIR) reduce 92) ((T_DO) reduce 92) ((T_ECHO) reduce 92) ((T_ELSE) reduce 92) ((T_ELSEIF) reduce 92) ((T_EMPTY) reduce 92) ((T_ENDDECLARE) reduce 92) ((T_ENDFOR) reduce 92) ((T_ENDFOREACH) reduce 92) ((T_ENDIF) reduce 92) ((T_ENDSWITCH) reduce 92) ((T_ENDWHILE) reduce 92) ((T_ENUM) reduce 92) ((T_EVAL) reduce 92) ((T_EXIT) reduce 92) ((T_EXTENDS) reduce 92) ((T_FILE) reduce 92) ((T_FINAL) reduce 92) ((T_FINALLY) reduce 92) ((T_FN) reduce 92) ((T_FOR) reduce 92) ((T_FOREACH) reduce 92) ((T_FUNCTION) reduce 92) ((T_FUNC_C) reduce 92) ((T_GLOBAL) reduce 92) ((T_GOTO) reduce 92) ((T_IF) reduce 92) ((T_IMPLEMENTS) reduce 92) ((T_INCLUDE) reduce 92) ((T_INCLUDE_ONCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_INSTEADOF) reduce 92) ((T_INTERFACE) reduce 92) ((T_ISSET) reduce 92) ((T_LINE) reduce 92) ((T_LIST) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_MATCH) reduce 92) ((T_METHOD_C) reduce 92) ((T_NAMESPACE) reduce 92) ((T_NEW) reduce 92) ((T_NS_C) reduce 92) ((T_PRINT) reduce 92) ((T_PRIVATE) reduce 92) ((T_PROPERTY_C) reduce 92) ((T_PROTECTED) reduce 92) ((T_PUBLIC) reduce 92) ((T_READONLY) reduce 92) ((T_REQUIRE) reduce 92) ((T_REQUIRE_ONCE) reduce 92) ((T_RETURN) reduce 92) ((T_STATIC) reduce 92) ((T_STRING) reduce 92) ((T_SWITCH) reduce 92) ((T_THROW) reduce 92) ((T_TRAIT) reduce 92) ((T_TRAIT_C) reduce 92) ((T_TRY) reduce 92) ((T_UNSET) reduce 92) ((T_USE) reduce 92) ((T_VAR) reduce 92) ((T_WHILE) reduce 92) ((T_YIELD) reduce 92)) 7384 (((T_ABSTRACT) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_ARRAY) reduce 282) ((T_AS) reduce 282) ((T_BREAK) reduce 282) ((T_CALLABLE) reduce 282) ((T_CASE) reduce 282) ((T_CATCH) reduce 282) ((T_CLASS) reduce 282) ((T_CLASS_C) reduce 282) ((T_CLONE) reduce 282) ((T_CONST) reduce 282) ((T_CONTINUE) reduce 282) ((T_DECLARE) reduce 282) ((T_DEFAULT) reduce 282) ((T_DIR) reduce 282) ((T_DO) reduce 282) ((T_ECHO) reduce 282) ((T_ELSE) reduce 282) ((T_ELSEIF) reduce 282) ((T_EMPTY) reduce 282) ((T_ENDDECLARE) reduce 282) ((T_ENDFOR) reduce 282) ((T_ENDFOREACH) reduce 282) ((T_ENDIF) reduce 282) ((T_ENDSWITCH) reduce 282) ((T_ENDWHILE) reduce 282) ((T_ENUM) reduce 282) ((T_EVAL) reduce 282) ((T_EXIT) reduce 282) ((T_EXTENDS) reduce 282) ((T_FILE) reduce 282) ((T_FINAL) reduce 282) ((T_FINALLY) reduce 282) ((T_FN) reduce 282) ((T_FOR) reduce 282) ((T_FOREACH) reduce 282) ((T_FUNCTION) reduce 282) ((T_FUNC_C) reduce 282) ((T_GLOBAL) reduce 282) ((T_GOTO) reduce 282) ((T_IF) reduce 282) ((T_IMPLEMENTS) reduce 282) ((T_INCLUDE) reduce 282) ((T_INCLUDE_ONCE) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_INSTEADOF) reduce 282) ((T_INTERFACE) reduce 282) ((T_ISSET) reduce 282) ((T_LINE) reduce 282) ((T_LIST) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_MATCH) reduce 282) ((T_METHOD_C) reduce 282) ((T_NAMESPACE) reduce 282) ((T_NEW) reduce 282) ((T_NS_C) reduce 282) ((T_PRINT) reduce 282) ((T_PRIVATE) reduce 282) ((T_PROPERTY_C) reduce 282) ((T_PROTECTED) reduce 282) ((T_PUBLIC) reduce 282) ((T_READONLY) reduce 282) ((T_REQUIRE) reduce 282) ((T_REQUIRE_ONCE) reduce 282) ((T_RETURN) reduce 282) ((T_STATIC) reduce 282) ((T_STRING) reduce 282) ((T_SWITCH) reduce 282) ((T_THROW) reduce 282) ((T_TRAIT) reduce 282) ((T_TRAIT_C) reduce 282) ((T_TRY) reduce 282) ((T_UNSET) reduce 282) ((T_USE) reduce 282) ((T_VAR) reduce 282) ((T_WHILE) reduce 282) ((T_YIELD) reduce 282)) 7385 (((T_ABSTRACT) reduce 274) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 274) ((T_ARRAY) reduce 274) ((T_AS) reduce 274) ((T_BREAK) reduce 274) ((T_CALLABLE) reduce 274) ((T_CASE) reduce 274) ((T_CATCH) reduce 274) ((T_CLASS) reduce 274) ((T_CLASS_C) reduce 274) ((T_CLONE) reduce 274) ((T_CONST) reduce 274) ((T_CONTINUE) reduce 274) ((T_DECLARE) reduce 274) ((T_DEFAULT) reduce 274) ((T_DIR) reduce 274) ((T_DO) reduce 274) ((T_ECHO) reduce 274) ((T_ELSE) reduce 274) ((T_ELSEIF) reduce 274) ((T_EMPTY) reduce 274) ((T_ENDDECLARE) reduce 274) ((T_ENDFOR) reduce 274) ((T_ENDFOREACH) reduce 274) ((T_ENDIF) reduce 274) ((T_ENDSWITCH) reduce 274) ((T_ENDWHILE) reduce 274) ((T_ENUM) reduce 274) ((T_EVAL) reduce 274) ((T_EXIT) reduce 274) ((T_EXTENDS) reduce 274) ((T_FILE) reduce 274) ((T_FINAL) reduce 274) ((T_FINALLY) reduce 274) ((T_FN) reduce 274) ((T_FOR) reduce 274) ((T_FOREACH) reduce 274) ((T_FUNCTION) reduce 274) ((T_FUNC_C) reduce 274) ((T_GLOBAL) reduce 274) ((T_GOTO) reduce 274) ((T_IF) reduce 274) ((T_IMPLEMENTS) reduce 274) ((T_INCLUDE) reduce 274) ((T_INCLUDE_ONCE) reduce 274) ((T_INSTANCEOF) reduce 274) ((T_INSTEADOF) reduce 274) ((T_INTERFACE) reduce 274) ((T_ISSET) reduce 274) ((T_LINE) reduce 274) ((T_LIST) reduce 274) ((T_LOGICAL_AND) reduce 274) ((T_LOGICAL_OR) reduce 274) ((T_LOGICAL_XOR) reduce 274) ((T_MATCH) reduce 274) ((T_METHOD_C) reduce 274) ((T_NAMESPACE) reduce 274) ((T_NEW) reduce 274) ((T_NS_C) reduce 274) ((T_PRINT) reduce 274) ((T_PRIVATE) reduce 274) ((T_PROPERTY_C) reduce 274) ((T_PROTECTED) reduce 274) ((T_PUBLIC) reduce 274) ((T_READONLY) reduce 274) ((T_REQUIRE) reduce 274) ((T_REQUIRE_ONCE) reduce 274) ((T_RETURN) reduce 274) ((T_STATIC) reduce 274) ((T_STRING) reduce 274) ((T_SWITCH) reduce 274) ((T_THROW) reduce 274) ((T_TRAIT) reduce 274) ((T_TRAIT_C) reduce 274) ((T_TRY) reduce 274) ((T_UNSET) reduce 274) ((T_USE) reduce 274) ((T_VAR) reduce 274) ((T_WHILE) reduce 274) ((T_YIELD) reduce 274)) 7386 (((T_ABSTRACT) reduce 268) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 268) ((T_ARRAY) reduce 268) ((T_AS) reduce 268) ((T_BREAK) reduce 268) ((T_CALLABLE) reduce 268) ((T_CASE) reduce 268) ((T_CATCH) reduce 268) ((T_CLASS) reduce 268) ((T_CLASS_C) reduce 268) ((T_CLONE) reduce 268) ((T_CONST) reduce 268) ((T_CONTINUE) reduce 268) ((T_DECLARE) reduce 268) ((T_DEFAULT) reduce 268) ((T_DIR) reduce 268) ((T_DO) reduce 268) ((T_ECHO) reduce 268) ((T_ELSE) reduce 268) ((T_ELSEIF) reduce 268) ((T_EMPTY) reduce 268) ((T_ENDDECLARE) reduce 268) ((T_ENDFOR) reduce 268) ((T_ENDFOREACH) reduce 268) ((T_ENDIF) reduce 268) ((T_ENDSWITCH) reduce 268) ((T_ENDWHILE) reduce 268) ((T_ENUM) reduce 268) ((T_EVAL) reduce 268) ((T_EXIT) reduce 268) ((T_EXTENDS) reduce 268) ((T_FILE) reduce 268) ((T_FINAL) reduce 268) ((T_FINALLY) reduce 268) ((T_FN) reduce 268) ((T_FOR) reduce 268) ((T_FOREACH) reduce 268) ((T_FUNCTION) reduce 268) ((T_FUNC_C) reduce 268) ((T_GLOBAL) reduce 268) ((T_GOTO) reduce 268) ((T_IF) reduce 268) ((T_IMPLEMENTS) reduce 268) ((T_INCLUDE) reduce 268) ((T_INCLUDE_ONCE) reduce 268) ((T_INSTANCEOF) reduce 268) ((T_INSTEADOF) reduce 268) ((T_INTERFACE) reduce 268) ((T_ISSET) reduce 268) ((T_LINE) reduce 268) ((T_LIST) reduce 268) ((T_LOGICAL_AND) reduce 268) ((T_LOGICAL_OR) reduce 268) ((T_LOGICAL_XOR) reduce 268) ((T_MATCH) reduce 268) ((T_METHOD_C) reduce 268) ((T_NAMESPACE) reduce 268) ((T_NEW) reduce 268) ((T_NS_C) reduce 268) ((T_PRINT) reduce 268) ((T_PRIVATE) reduce 268) ((T_PROPERTY_C) reduce 268) ((T_PROTECTED) reduce 268) ((T_PUBLIC) reduce 268) ((T_READONLY) reduce 268) ((T_REQUIRE) reduce 268) ((T_REQUIRE_ONCE) reduce 268) ((T_RETURN) reduce 268) ((T_STATIC) reduce 268) ((T_STRING) reduce 268) ((T_SWITCH) reduce 268) ((T_THROW) reduce 268) ((T_TRAIT) reduce 268) ((T_TRAIT_C) reduce 268) ((T_TRY) reduce 268) ((T_UNSET) reduce 268) ((T_USE) reduce 268) ((T_VAR) reduce 268) ((T_WHILE) reduce 268) ((T_YIELD) reduce 268)) 7389 (((",") reduce 384) ((";") reduce 384)) 7391 (((T_ABSTRACT) reduce 275) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 275) ((T_ARRAY) reduce 275) ((T_AS) reduce 275) ((T_BREAK) reduce 275) ((T_CALLABLE) reduce 275) ((T_CASE) reduce 275) ((T_CATCH) reduce 275) ((T_CLASS) reduce 275) ((T_CLASS_C) reduce 275) ((T_CLONE) reduce 275) ((T_CONST) reduce 275) ((T_CONTINUE) reduce 275) ((T_DECLARE) reduce 275) ((T_DEFAULT) reduce 275) ((T_DIR) reduce 275) ((T_DO) reduce 275) ((T_ECHO) reduce 275) ((T_ELSE) reduce 275) ((T_ELSEIF) reduce 275) ((T_EMPTY) reduce 275) ((T_ENDDECLARE) reduce 275) ((T_ENDFOR) reduce 275) ((T_ENDFOREACH) reduce 275) ((T_ENDIF) reduce 275) ((T_ENDSWITCH) reduce 275) ((T_ENDWHILE) reduce 275) ((T_ENUM) reduce 275) ((T_EVAL) reduce 275) ((T_EXIT) reduce 275) ((T_EXTENDS) reduce 275) ((T_FILE) reduce 275) ((T_FINAL) reduce 275) ((T_FINALLY) reduce 275) ((T_FN) reduce 275) ((T_FOR) reduce 275) ((T_FOREACH) reduce 275) ((T_FUNCTION) reduce 275) ((T_FUNC_C) reduce 275) ((T_GLOBAL) reduce 275) ((T_GOTO) reduce 275) ((T_IF) reduce 275) ((T_IMPLEMENTS) reduce 275) ((T_INCLUDE) reduce 275) ((T_INCLUDE_ONCE) reduce 275) ((T_INSTANCEOF) reduce 275) ((T_INSTEADOF) reduce 275) ((T_INTERFACE) reduce 275) ((T_ISSET) reduce 275) ((T_LINE) reduce 275) ((T_LIST) reduce 275) ((T_LOGICAL_AND) reduce 275) ((T_LOGICAL_OR) reduce 275) ((T_LOGICAL_XOR) reduce 275) ((T_MATCH) reduce 275) ((T_METHOD_C) reduce 275) ((T_NAMESPACE) reduce 275) ((T_NEW) reduce 275) ((T_NS_C) reduce 275) ((T_PRINT) reduce 275) ((T_PRIVATE) reduce 275) ((T_PROPERTY_C) reduce 275) ((T_PROTECTED) reduce 275) ((T_PUBLIC) reduce 275) ((T_READONLY) reduce 275) ((T_REQUIRE) reduce 275) ((T_REQUIRE_ONCE) reduce 275) ((T_RETURN) reduce 275) ((T_STATIC) reduce 275) ((T_STRING) reduce 275) ((T_SWITCH) reduce 275) ((T_THROW) reduce 275) ((T_TRAIT) reduce 275) ((T_TRAIT_C) reduce 275) ((T_TRY) reduce 275) ((T_UNSET) reduce 275) ((T_USE) reduce 275) ((T_VAR) reduce 275) ((T_WHILE) reduce 275) ((T_YIELD) reduce 275)) 7392 (((T_ABSTRACT) reduce 318) ((T_ATTRIBUTE) reduce 318) ((T_CASE) reduce 318) ((T_CONST) reduce 318) ((T_FINAL) reduce 318) ((T_FUNCTION) reduce 318) ((T_PRIVATE) reduce 318) ((T_PRIVATE_SET) reduce 318) ((T_PROTECTED) reduce 318) ((T_PROTECTED_SET) reduce 318) ((T_PUBLIC) reduce 318) ((T_PUBLIC_SET) reduce 318) ((T_READONLY) reduce 318) ((T_STATIC) reduce 318) ((T_USE) reduce 318) ((T_VAR) reduce 318) (("}") reduce 318)) 7393 (((T_ABSTRACT) reduce 280) ((T_ARRAY) reduce 280) ((T_AS) reduce 280) ((T_BREAK) reduce 280) ((T_CALLABLE) reduce 280) ((T_CASE) reduce 280) ((T_CATCH) reduce 280) ((T_CLASS) reduce 280) ((T_CLASS_C) reduce 280) ((T_CLONE) reduce 280) ((T_CONST) reduce 280) ((T_CONTINUE) reduce 280) ((T_DECLARE) reduce 280) ((T_DEFAULT) reduce 280) ((T_DIR) reduce 280) ((T_DO) reduce 280) ((T_ECHO) reduce 280) ((T_ELSE) reduce 280) ((T_ELSEIF) reduce 280) ((T_EMPTY) reduce 280) ((T_ENDDECLARE) reduce 280) ((T_ENDFOR) reduce 280) ((T_ENDFOREACH) reduce 280) ((T_ENDIF) reduce 280) ((T_ENDSWITCH) reduce 280) ((T_ENDWHILE) reduce 280) ((T_ENUM) reduce 280) ((T_EVAL) reduce 280) ((T_EXIT) reduce 280) ((T_EXTENDS) reduce 280) ((T_FILE) reduce 280) ((T_FINAL) reduce 280) ((T_FINALLY) reduce 280) ((T_FN) reduce 280) ((T_FOR) reduce 280) ((T_FOREACH) reduce 280) ((T_FUNCTION) reduce 280) ((T_FUNC_C) reduce 280) ((T_GLOBAL) reduce 280) ((T_GOTO) reduce 280) ((T_IF) reduce 280) ((T_IMPLEMENTS) reduce 280) ((T_INCLUDE) reduce 280) ((T_INCLUDE_ONCE) reduce 280) ((T_INSTANCEOF) reduce 280) ((T_INSTEADOF) reduce 280) ((T_INTERFACE) reduce 280) ((T_ISSET) reduce 280) ((T_LINE) reduce 280) ((T_LIST) reduce 280) ((T_LOGICAL_AND) reduce 280) ((T_LOGICAL_OR) reduce 280) ((T_LOGICAL_XOR) reduce 280) ((T_MATCH) reduce 280) ((T_METHOD_C) reduce 280) ((T_NAMESPACE) reduce 280) ((T_NEW) reduce 280) ((T_NS_C) reduce 280) ((T_PRINT) reduce 280) ((T_PRIVATE) reduce 280) ((T_PROPERTY_C) reduce 280) ((T_PROTECTED) reduce 280) ((T_PUBLIC) reduce 280) ((T_READONLY) reduce 280) ((T_REQUIRE) reduce 280) ((T_REQUIRE_ONCE) reduce 280) ((T_RETURN) reduce 280) ((T_STATIC) reduce 280) ((T_STRING) reduce 280) ((T_SWITCH) reduce 280) ((T_THROW) reduce 280) ((T_TRAIT) reduce 280) ((T_TRAIT_C) reduce 280) ((T_TRY) reduce 280) ((T_UNSET) reduce 280) ((T_USE) reduce 280) ((T_VAR) reduce 280) ((T_WHILE) reduce 280) ((T_YIELD) reduce 280)) 7394 (((T_ABSTRACT) reduce 281) ((T_ARRAY) reduce 281) ((T_AS) reduce 281) ((T_BREAK) reduce 281) ((T_CALLABLE) reduce 281) ((T_CASE) reduce 281) ((T_CATCH) reduce 281) ((T_CLASS) reduce 281) ((T_CLASS_C) reduce 281) ((T_CLONE) reduce 281) ((T_CONST) reduce 281) ((T_CONTINUE) reduce 281) ((T_DECLARE) reduce 281) ((T_DEFAULT) reduce 281) ((T_DIR) reduce 281) ((T_DO) reduce 281) ((T_ECHO) reduce 281) ((T_ELSE) reduce 281) ((T_ELSEIF) reduce 281) ((T_EMPTY) reduce 281) ((T_ENDDECLARE) reduce 281) ((T_ENDFOR) reduce 281) ((T_ENDFOREACH) reduce 281) ((T_ENDIF) reduce 281) ((T_ENDSWITCH) reduce 281) ((T_ENDWHILE) reduce 281) ((T_ENUM) reduce 281) ((T_EVAL) reduce 281) ((T_EXIT) reduce 281) ((T_EXTENDS) reduce 281) ((T_FILE) reduce 281) ((T_FINAL) reduce 281) ((T_FINALLY) reduce 281) ((T_FN) reduce 281) ((T_FOR) reduce 281) ((T_FOREACH) reduce 281) ((T_FUNCTION) reduce 281) ((T_FUNC_C) reduce 281) ((T_GLOBAL) reduce 281) ((T_GOTO) reduce 281) ((T_IF) reduce 281) ((T_IMPLEMENTS) reduce 281) ((T_INCLUDE) reduce 281) ((T_INCLUDE_ONCE) reduce 281) ((T_INSTANCEOF) reduce 281) ((T_INSTEADOF) reduce 281) ((T_INTERFACE) reduce 281) ((T_ISSET) reduce 281) ((T_LINE) reduce 281) ((T_LIST) reduce 281) ((T_LOGICAL_AND) reduce 281) ((T_LOGICAL_OR) reduce 281) ((T_LOGICAL_XOR) reduce 281) ((T_MATCH) reduce 281) ((T_METHOD_C) reduce 281) ((T_NAMESPACE) reduce 281) ((T_NEW) reduce 281) ((T_NS_C) reduce 281) ((T_PRINT) reduce 281) ((T_PRIVATE) reduce 281) ((T_PROPERTY_C) reduce 281) ((T_PROTECTED) reduce 281) ((T_PUBLIC) reduce 281) ((T_READONLY) reduce 281) ((T_REQUIRE) reduce 281) ((T_REQUIRE_ONCE) reduce 281) ((T_RETURN) reduce 281) ((T_STATIC) reduce 281) ((T_STRING) reduce 281) ((T_SWITCH) reduce 281) ((T_THROW) reduce 281) ((T_TRAIT) reduce 281) ((T_TRAIT_C) reduce 281) ((T_TRY) reduce 281) ((T_UNSET) reduce 281) ((T_USE) reduce 281) ((T_VAR) reduce 281) ((T_WHILE) reduce 281) ((T_YIELD) reduce 281)) 7395 (((T_ABSTRACT) reduce 94) ((T_ARRAY) reduce 94) ((T_AS) reduce 94) ((T_BREAK) reduce 94) ((T_CALLABLE) reduce 94) ((T_CASE) reduce 94) ((T_CATCH) reduce 94) ((T_CLASS) reduce 94) ((T_CLASS_C) reduce 94) ((T_CLONE) reduce 94) ((T_CONST) reduce 94) ((T_CONTINUE) reduce 94) ((T_DECLARE) reduce 94) ((T_DEFAULT) reduce 94) ((T_DIR) reduce 94) ((T_DO) reduce 94) ((T_ECHO) reduce 94) ((T_ELSE) reduce 94) ((T_ELSEIF) reduce 94) ((T_EMPTY) reduce 94) ((T_ENDDECLARE) reduce 94) ((T_ENDFOR) reduce 94) ((T_ENDFOREACH) reduce 94) ((T_ENDIF) reduce 94) ((T_ENDSWITCH) reduce 94) ((T_ENDWHILE) reduce 94) ((T_ENUM) reduce 94) ((T_EVAL) reduce 94) ((T_EXIT) reduce 94) ((T_EXTENDS) reduce 94) ((T_FILE) reduce 94) ((T_FINAL) reduce 94) ((T_FINALLY) reduce 94) ((T_FN) reduce 94) ((T_FOR) reduce 94) ((T_FOREACH) reduce 94) ((T_FUNCTION) reduce 94) ((T_FUNC_C) reduce 94) ((T_GLOBAL) reduce 94) ((T_GOTO) reduce 94) ((T_IF) reduce 94) ((T_IMPLEMENTS) reduce 94) ((T_INCLUDE) reduce 94) ((T_INCLUDE_ONCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_INSTEADOF) reduce 94) ((T_INTERFACE) reduce 94) ((T_ISSET) reduce 94) ((T_LINE) reduce 94) ((T_LIST) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_MATCH) reduce 94) ((T_METHOD_C) reduce 94) ((T_NAMESPACE) reduce 94) ((T_NEW) reduce 94) ((T_NS_C) reduce 94) ((T_PRINT) reduce 94) ((T_PRIVATE) reduce 94) ((T_PROPERTY_C) reduce 94) ((T_PROTECTED) reduce 94) ((T_PUBLIC) reduce 94) ((T_READONLY) reduce 94) ((T_REQUIRE) reduce 94) ((T_REQUIRE_ONCE) reduce 94) ((T_RETURN) reduce 94) ((T_STATIC) reduce 94) ((T_STRING) reduce 94) ((T_SWITCH) reduce 94) ((T_THROW) reduce 94) ((T_TRAIT) reduce 94) ((T_TRAIT_C) reduce 94) ((T_TRY) reduce 94) ((T_UNSET) reduce 94) ((T_USE) reduce 94) ((T_VAR) reduce 94) ((T_WHILE) reduce 94) ((T_YIELD) reduce 94)) 7396 (((T_ABSTRACT) reduce 93) ((T_ARRAY) reduce 93) ((T_AS) reduce 93) ((T_BREAK) reduce 93) ((T_CALLABLE) reduce 93) ((T_CASE) reduce 93) ((T_CATCH) reduce 93) ((T_CLASS) reduce 93) ((T_CLASS_C) reduce 93) ((T_CLONE) reduce 93) ((T_CONST) reduce 93) ((T_CONTINUE) reduce 93) ((T_DECLARE) reduce 93) ((T_DEFAULT) reduce 93) ((T_DIR) reduce 93) ((T_DO) reduce 93) ((T_ECHO) reduce 93) ((T_ELSE) reduce 93) ((T_ELSEIF) reduce 93) ((T_EMPTY) reduce 93) ((T_ENDDECLARE) reduce 93) ((T_ENDFOR) reduce 93) ((T_ENDFOREACH) reduce 93) ((T_ENDIF) reduce 93) ((T_ENDSWITCH) reduce 93) ((T_ENDWHILE) reduce 93) ((T_ENUM) reduce 93) ((T_EVAL) reduce 93) ((T_EXIT) reduce 93) ((T_EXTENDS) reduce 93) ((T_FILE) reduce 93) ((T_FINAL) reduce 93) ((T_FINALLY) reduce 93) ((T_FN) reduce 93) ((T_FOR) reduce 93) ((T_FOREACH) reduce 93) ((T_FUNCTION) reduce 93) ((T_FUNC_C) reduce 93) ((T_GLOBAL) reduce 93) ((T_GOTO) reduce 93) ((T_IF) reduce 93) ((T_IMPLEMENTS) reduce 93) ((T_INCLUDE) reduce 93) ((T_INCLUDE_ONCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_INSTEADOF) reduce 93) ((T_INTERFACE) reduce 93) ((T_ISSET) reduce 93) ((T_LINE) reduce 93) ((T_LIST) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_MATCH) reduce 93) ((T_METHOD_C) reduce 93) ((T_NAMESPACE) reduce 93) ((T_NEW) reduce 93) ((T_NS_C) reduce 93) ((T_PRINT) reduce 93) ((T_PRIVATE) reduce 93) ((T_PROPERTY_C) reduce 93) ((T_PROTECTED) reduce 93) ((T_PUBLIC) reduce 93) ((T_READONLY) reduce 93) ((T_REQUIRE) reduce 93) ((T_REQUIRE_ONCE) reduce 93) ((T_RETURN) reduce 93) ((T_STATIC) reduce 93) ((T_STRING) reduce 93) ((T_SWITCH) reduce 93) ((T_THROW) reduce 93) ((T_TRAIT) reduce 93) ((T_TRAIT_C) reduce 93) ((T_TRY) reduce 93) ((T_UNSET) reduce 93) ((T_USE) reduce 93) ((T_VAR) reduce 93) ((T_WHILE) reduce 93) ((T_YIELD) reduce 93)) 7397 (((T_ABSTRACT) reduce 95) ((T_ARRAY) reduce 95) ((T_AS) reduce 95) ((T_BREAK) reduce 95) ((T_CALLABLE) reduce 95) ((T_CASE) reduce 95) ((T_CATCH) reduce 95) ((T_CLASS) reduce 95) ((T_CLASS_C) reduce 95) ((T_CLONE) reduce 95) ((T_CONST) reduce 95) ((T_CONTINUE) reduce 95) ((T_DECLARE) reduce 95) ((T_DEFAULT) reduce 95) ((T_DIR) reduce 95) ((T_DO) reduce 95) ((T_ECHO) reduce 95) ((T_ELSE) reduce 95) ((T_ELSEIF) reduce 95) ((T_EMPTY) reduce 95) ((T_ENDDECLARE) reduce 95) ((T_ENDFOR) reduce 95) ((T_ENDFOREACH) reduce 95) ((T_ENDIF) reduce 95) ((T_ENDSWITCH) reduce 95) ((T_ENDWHILE) reduce 95) ((T_ENUM) reduce 95) ((T_EVAL) reduce 95) ((T_EXIT) reduce 95) ((T_EXTENDS) reduce 95) ((T_FILE) reduce 95) ((T_FINAL) reduce 95) ((T_FINALLY) reduce 95) ((T_FN) reduce 95) ((T_FOR) reduce 95) ((T_FOREACH) reduce 95) ((T_FUNCTION) reduce 95) ((T_FUNC_C) reduce 95) ((T_GLOBAL) reduce 95) ((T_GOTO) reduce 95) ((T_IF) reduce 95) ((T_IMPLEMENTS) reduce 95) ((T_INCLUDE) reduce 95) ((T_INCLUDE_ONCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_INSTEADOF) reduce 95) ((T_INTERFACE) reduce 95) ((T_ISSET) reduce 95) ((T_LINE) reduce 95) ((T_LIST) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_MATCH) reduce 95) ((T_METHOD_C) reduce 95) ((T_NAMESPACE) reduce 95) ((T_NEW) reduce 95) ((T_NS_C) reduce 95) ((T_PRINT) reduce 95) ((T_PRIVATE) reduce 95) ((T_PROPERTY_C) reduce 95) ((T_PROTECTED) reduce 95) ((T_PUBLIC) reduce 95) ((T_READONLY) reduce 95) ((T_REQUIRE) reduce 95) ((T_REQUIRE_ONCE) reduce 95) ((T_RETURN) reduce 95) ((T_STATIC) reduce 95) ((T_STRING) reduce 95) ((T_SWITCH) reduce 95) ((T_THROW) reduce 95) ((T_TRAIT) reduce 95) ((T_TRAIT_C) reduce 95) ((T_TRY) reduce 95) ((T_UNSET) reduce 95) ((T_USE) reduce 95) ((T_VAR) reduce 95) ((T_WHILE) reduce 95) ((T_YIELD) reduce 95)) 7398 (((T_ABSTRACT) reduce 269) ((T_ARRAY) reduce 269) ((T_AS) reduce 269) ((T_BREAK) reduce 269) ((T_CALLABLE) reduce 269) ((T_CASE) reduce 269) ((T_CATCH) reduce 269) ((T_CLASS) reduce 269) ((T_CLASS_C) reduce 269) ((T_CLONE) reduce 269) ((T_CONST) reduce 269) ((T_CONTINUE) reduce 269) ((T_DECLARE) reduce 269) ((T_DEFAULT) reduce 269) ((T_DIR) reduce 269) ((T_DO) reduce 269) ((T_ECHO) reduce 269) ((T_ELSE) reduce 269) ((T_ELSEIF) reduce 269) ((T_EMPTY) reduce 269) ((T_ENDDECLARE) reduce 269) ((T_ENDFOR) reduce 269) ((T_ENDFOREACH) reduce 269) ((T_ENDIF) reduce 269) ((T_ENDSWITCH) reduce 269) ((T_ENDWHILE) reduce 269) ((T_ENUM) reduce 269) ((T_EVAL) reduce 269) ((T_EXIT) reduce 269) ((T_EXTENDS) reduce 269) ((T_FILE) reduce 269) ((T_FINAL) reduce 269) ((T_FINALLY) reduce 269) ((T_FN) reduce 269) ((T_FOR) reduce 269) ((T_FOREACH) reduce 269) ((T_FUNCTION) reduce 269) ((T_FUNC_C) reduce 269) ((T_GLOBAL) reduce 269) ((T_GOTO) reduce 269) ((T_IF) reduce 269) ((T_IMPLEMENTS) reduce 269) ((T_INCLUDE) reduce 269) ((T_INCLUDE_ONCE) reduce 269) ((T_INSTANCEOF) reduce 269) ((T_INSTEADOF) reduce 269) ((T_INTERFACE) reduce 269) ((T_ISSET) reduce 269) ((T_LINE) reduce 269) ((T_LIST) reduce 269) ((T_LOGICAL_AND) reduce 269) ((T_LOGICAL_OR) reduce 269) ((T_LOGICAL_XOR) reduce 269) ((T_MATCH) reduce 269) ((T_METHOD_C) reduce 269) ((T_NAMESPACE) reduce 269) ((T_NEW) reduce 269) ((T_NS_C) reduce 269) ((T_PRINT) reduce 269) ((T_PRIVATE) reduce 269) ((T_PROPERTY_C) reduce 269) ((T_PROTECTED) reduce 269) ((T_PUBLIC) reduce 269) ((T_READONLY) reduce 269) ((T_REQUIRE) reduce 269) ((T_REQUIRE_ONCE) reduce 269) ((T_RETURN) reduce 269) ((T_STATIC) reduce 269) ((T_STRING) reduce 269) ((T_SWITCH) reduce 269) ((T_THROW) reduce 269) ((T_TRAIT) reduce 269) ((T_TRAIT_C) reduce 269) ((T_TRY) reduce 269) ((T_UNSET) reduce 269) ((T_USE) reduce 269) ((T_VAR) reduce 269) ((T_WHILE) reduce 269) ((T_YIELD) reduce 269)) 7399 (((T_ABSTRACT) reduce 92) ((T_ARRAY) reduce 92) ((T_AS) reduce 92) ((T_BREAK) reduce 92) ((T_CALLABLE) reduce 92) ((T_CASE) reduce 92) ((T_CATCH) reduce 92) ((T_CLASS) reduce 92) ((T_CLASS_C) reduce 92) ((T_CLONE) reduce 92) ((T_CONST) reduce 92) ((T_CONTINUE) reduce 92) ((T_DECLARE) reduce 92) ((T_DEFAULT) reduce 92) ((T_DIR) reduce 92) ((T_DO) reduce 92) ((T_ECHO) reduce 92) ((T_ELSE) reduce 92) ((T_ELSEIF) reduce 92) ((T_EMPTY) reduce 92) ((T_ENDDECLARE) reduce 92) ((T_ENDFOR) reduce 92) ((T_ENDFOREACH) reduce 92) ((T_ENDIF) reduce 92) ((T_ENDSWITCH) reduce 92) ((T_ENDWHILE) reduce 92) ((T_ENUM) reduce 92) ((T_EVAL) reduce 92) ((T_EXIT) reduce 92) ((T_EXTENDS) reduce 92) ((T_FILE) reduce 92) ((T_FINAL) reduce 92) ((T_FINALLY) reduce 92) ((T_FN) reduce 92) ((T_FOR) reduce 92) ((T_FOREACH) reduce 92) ((T_FUNCTION) reduce 92) ((T_FUNC_C) reduce 92) ((T_GLOBAL) reduce 92) ((T_GOTO) reduce 92) ((T_IF) reduce 92) ((T_IMPLEMENTS) reduce 92) ((T_INCLUDE) reduce 92) ((T_INCLUDE_ONCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_INSTEADOF) reduce 92) ((T_INTERFACE) reduce 92) ((T_ISSET) reduce 92) ((T_LINE) reduce 92) ((T_LIST) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_MATCH) reduce 92) ((T_METHOD_C) reduce 92) ((T_NAMESPACE) reduce 92) ((T_NEW) reduce 92) ((T_NS_C) reduce 92) ((T_PRINT) reduce 92) ((T_PRIVATE) reduce 92) ((T_PROPERTY_C) reduce 92) ((T_PROTECTED) reduce 92) ((T_PUBLIC) reduce 92) ((T_READONLY) reduce 92) ((T_REQUIRE) reduce 92) ((T_REQUIRE_ONCE) reduce 92) ((T_RETURN) reduce 92) ((T_STATIC) reduce 92) ((T_STRING) reduce 92) ((T_SWITCH) reduce 92) ((T_THROW) reduce 92) ((T_TRAIT) reduce 92) ((T_TRAIT_C) reduce 92) ((T_TRY) reduce 92) ((T_UNSET) reduce 92) ((T_USE) reduce 92) ((T_VAR) reduce 92) ((T_WHILE) reduce 92) ((T_YIELD) reduce 92)) 7400 (((T_ABSTRACT) reduce 282) ((T_ARRAY) reduce 282) ((T_AS) reduce 282) ((T_BREAK) reduce 282) ((T_CALLABLE) reduce 282) ((T_CASE) reduce 282) ((T_CATCH) reduce 282) ((T_CLASS) reduce 282) ((T_CLASS_C) reduce 282) ((T_CLONE) reduce 282) ((T_CONST) reduce 282) ((T_CONTINUE) reduce 282) ((T_DECLARE) reduce 282) ((T_DEFAULT) reduce 282) ((T_DIR) reduce 282) ((T_DO) reduce 282) ((T_ECHO) reduce 282) ((T_ELSE) reduce 282) ((T_ELSEIF) reduce 282) ((T_EMPTY) reduce 282) ((T_ENDDECLARE) reduce 282) ((T_ENDFOR) reduce 282) ((T_ENDFOREACH) reduce 282) ((T_ENDIF) reduce 282) ((T_ENDSWITCH) reduce 282) ((T_ENDWHILE) reduce 282) ((T_ENUM) reduce 282) ((T_EVAL) reduce 282) ((T_EXIT) reduce 282) ((T_EXTENDS) reduce 282) ((T_FILE) reduce 282) ((T_FINAL) reduce 282) ((T_FINALLY) reduce 282) ((T_FN) reduce 282) ((T_FOR) reduce 282) ((T_FOREACH) reduce 282) ((T_FUNCTION) reduce 282) ((T_FUNC_C) reduce 282) ((T_GLOBAL) reduce 282) ((T_GOTO) reduce 282) ((T_IF) reduce 282) ((T_IMPLEMENTS) reduce 282) ((T_INCLUDE) reduce 282) ((T_INCLUDE_ONCE) reduce 282) ((T_INSTANCEOF) reduce 282) ((T_INSTEADOF) reduce 282) ((T_INTERFACE) reduce 282) ((T_ISSET) reduce 282) ((T_LINE) reduce 282) ((T_LIST) reduce 282) ((T_LOGICAL_AND) reduce 282) ((T_LOGICAL_OR) reduce 282) ((T_LOGICAL_XOR) reduce 282) ((T_MATCH) reduce 282) ((T_METHOD_C) reduce 282) ((T_NAMESPACE) reduce 282) ((T_NEW) reduce 282) ((T_NS_C) reduce 282) ((T_PRINT) reduce 282) ((T_PRIVATE) reduce 282) ((T_PROPERTY_C) reduce 282) ((T_PROTECTED) reduce 282) ((T_PUBLIC) reduce 282) ((T_READONLY) reduce 282) ((T_REQUIRE) reduce 282) ((T_REQUIRE_ONCE) reduce 282) ((T_RETURN) reduce 282) ((T_STATIC) reduce 282) ((T_STRING) reduce 282) ((T_SWITCH) reduce 282) ((T_THROW) reduce 282) ((T_TRAIT) reduce 282) ((T_TRAIT_C) reduce 282) ((T_TRY) reduce 282) ((T_UNSET) reduce 282) ((T_USE) reduce 282) ((T_VAR) reduce 282) ((T_WHILE) reduce 282) ((T_YIELD) reduce 282)) 7401 (((T_ABSTRACT) reduce 265) ((T_ARRAY) reduce 265) ((T_AS) reduce 265) ((T_BREAK) reduce 265) ((T_CALLABLE) reduce 265) ((T_CASE) reduce 265) ((T_CATCH) reduce 265) ((T_CLASS) reduce 265) ((T_CLASS_C) reduce 265) ((T_CLONE) reduce 265) ((T_CONST) reduce 265) ((T_CONTINUE) reduce 265) ((T_DECLARE) reduce 265) ((T_DEFAULT) reduce 265) ((T_DIR) reduce 265) ((T_DO) reduce 265) ((T_ECHO) reduce 265) ((T_ELSE) reduce 265) ((T_ELSEIF) reduce 265) ((T_EMPTY) reduce 265) ((T_ENDDECLARE) reduce 265) ((T_ENDFOR) reduce 265) ((T_ENDFOREACH) reduce 265) ((T_ENDIF) reduce 265) ((T_ENDSWITCH) reduce 265) ((T_ENDWHILE) reduce 265) ((T_ENUM) reduce 265) ((T_EVAL) reduce 265) ((T_EXIT) reduce 265) ((T_EXTENDS) reduce 265) ((T_FILE) reduce 265) ((T_FINAL) reduce 265) ((T_FINALLY) reduce 265) ((T_FN) reduce 265) ((T_FOR) reduce 265) ((T_FOREACH) reduce 265) ((T_FUNCTION) reduce 265) ((T_FUNC_C) reduce 265) ((T_GLOBAL) reduce 265) ((T_GOTO) reduce 265) ((T_IF) reduce 265) ((T_IMPLEMENTS) reduce 265) ((T_INCLUDE) reduce 265) ((T_INCLUDE_ONCE) reduce 265) ((T_INSTANCEOF) reduce 265) ((T_INSTEADOF) reduce 265) ((T_INTERFACE) reduce 265) ((T_ISSET) reduce 265) ((T_LINE) reduce 265) ((T_LIST) reduce 265) ((T_LOGICAL_AND) reduce 265) ((T_LOGICAL_OR) reduce 265) ((T_LOGICAL_XOR) reduce 265) ((T_MATCH) reduce 265) ((T_METHOD_C) reduce 265) ((T_NAMESPACE) reduce 265) ((T_NEW) reduce 265) ((T_NS_C) reduce 265) ((T_PRINT) reduce 265) ((T_PRIVATE) reduce 265) ((T_PROPERTY_C) reduce 265) ((T_PROTECTED) reduce 265) ((T_PUBLIC) reduce 265) ((T_READONLY) reduce 265) ((T_REQUIRE) reduce 265) ((T_REQUIRE_ONCE) reduce 265) ((T_RETURN) reduce 265) ((T_STATIC) reduce 265) ((T_STRING) reduce 265) ((T_SWITCH) reduce 265) ((T_THROW) reduce 265) ((T_TRAIT) reduce 265) ((T_TRAIT_C) reduce 265) ((T_TRY) reduce 265) ((T_UNSET) reduce 265) ((T_USE) reduce 265) ((T_VAR) reduce 265) ((T_WHILE) reduce 265) ((T_YIELD) reduce 265)) 7402 (((T_ABSTRACT) reduce 268) ((T_ARRAY) reduce 268) ((T_AS) reduce 268) ((T_BREAK) reduce 268) ((T_CALLABLE) reduce 268) ((T_CASE) reduce 268) ((T_CATCH) reduce 268) ((T_CLASS) reduce 268) ((T_CLASS_C) reduce 268) ((T_CLONE) reduce 268) ((T_CONST) reduce 268) ((T_CONTINUE) reduce 268) ((T_DECLARE) reduce 268) ((T_DEFAULT) reduce 268) ((T_DIR) reduce 268) ((T_DO) reduce 268) ((T_ECHO) reduce 268) ((T_ELSE) reduce 268) ((T_ELSEIF) reduce 268) ((T_EMPTY) reduce 268) ((T_ENDDECLARE) reduce 268) ((T_ENDFOR) reduce 268) ((T_ENDFOREACH) reduce 268) ((T_ENDIF) reduce 268) ((T_ENDSWITCH) reduce 268) ((T_ENDWHILE) reduce 268) ((T_ENUM) reduce 268) ((T_EVAL) reduce 268) ((T_EXIT) reduce 268) ((T_EXTENDS) reduce 268) ((T_FILE) reduce 268) ((T_FINAL) reduce 268) ((T_FINALLY) reduce 268) ((T_FN) reduce 268) ((T_FOR) reduce 268) ((T_FOREACH) reduce 268) ((T_FUNCTION) reduce 268) ((T_FUNC_C) reduce 268) ((T_GLOBAL) reduce 268) ((T_GOTO) reduce 268) ((T_IF) reduce 268) ((T_IMPLEMENTS) reduce 268) ((T_INCLUDE) reduce 268) ((T_INCLUDE_ONCE) reduce 268) ((T_INSTANCEOF) reduce 268) ((T_INSTEADOF) reduce 268) ((T_INTERFACE) reduce 268) ((T_ISSET) reduce 268) ((T_LINE) reduce 268) ((T_LIST) reduce 268) ((T_LOGICAL_AND) reduce 268) ((T_LOGICAL_OR) reduce 268) ((T_LOGICAL_XOR) reduce 268) ((T_MATCH) reduce 268) ((T_METHOD_C) reduce 268) ((T_NAMESPACE) reduce 268) ((T_NEW) reduce 268) ((T_NS_C) reduce 268) ((T_PRINT) reduce 268) ((T_PRIVATE) reduce 268) ((T_PROPERTY_C) reduce 268) ((T_PROTECTED) reduce 268) ((T_PUBLIC) reduce 268) ((T_READONLY) reduce 268) ((T_REQUIRE) reduce 268) ((T_REQUIRE_ONCE) reduce 268) ((T_RETURN) reduce 268) ((T_STATIC) reduce 268) ((T_STRING) reduce 268) ((T_SWITCH) reduce 268) ((T_THROW) reduce 268) ((T_TRAIT) reduce 268) ((T_TRAIT_C) reduce 268) ((T_TRY) reduce 268) ((T_UNSET) reduce 268) ((T_USE) reduce 268) ((T_VAR) reduce 268) ((T_WHILE) reduce 268) ((T_YIELD) reduce 268)) 7403 (((T_ABSTRACT) reduce 102) ((T_ATTRIBUTE) reduce 102) ((T_CASE) reduce 102) ((T_CONST) reduce 102) ((T_FINAL) reduce 102) ((T_FUNCTION) reduce 102) ((T_PRIVATE) reduce 102) ((T_PRIVATE_SET) reduce 102) ((T_PROTECTED) reduce 102) ((T_PROTECTED_SET) reduce 102) ((T_PUBLIC) reduce 102) ((T_PUBLIC_SET) reduce 102) ((T_READONLY) reduce 102) ((T_STATIC) reduce 102) ((T_VAR) reduce 102)) 7404 (((T_ABSTRACT) reduce 323) ((T_ATTRIBUTE) reduce 323) ((T_CASE) reduce 323) ((T_CONST) reduce 323) ((T_FINAL) reduce 323) ((T_FUNCTION) reduce 323) ((T_PRIVATE) reduce 323) ((T_PRIVATE_SET) reduce 323) ((T_PROTECTED) reduce 323) ((T_PROTECTED_SET) reduce 323) ((T_PUBLIC) reduce 323) ((T_PUBLIC_SET) reduce 323) ((T_READONLY) reduce 323) ((T_STATIC) reduce 323) ((T_USE) reduce 323) ((T_VAR) reduce 323) (("}") reduce 323)) 7405 (((",") reduce 94) ((";") reduce 94) (("{") reduce 94)) 7406 (((",") reduce 93) ((";") reduce 93) (("{") reduce 93)) 7407 (((",") reduce 95) ((";") reduce 95) (("{") reduce 95)) 7408 (((",") reduce 509) ((";") reduce 509) (("{") reduce 509)) 7409 (((",") reduce 92) ((";") reduce 92) (("{") reduce 92)) 7410 (((",") reduce 325) ((";") reduce 325) (("{") reduce 325)) 7411 (((",") shift) ((";") shift) (("{") shift)) 7412 (((",") reduce 510) ((";") reduce 510) (("{") reduce 510)) 7414 (((T_ABSTRACT) reduce 327) ((T_ATTRIBUTE) reduce 327) ((T_CASE) reduce 327) ((T_CONST) reduce 327) ((T_FINAL) reduce 327) ((T_FUNCTION) reduce 327) ((T_PRIVATE) reduce 327) ((T_PRIVATE_SET) reduce 327) ((T_PROTECTED) reduce 327) ((T_PROTECTED_SET) reduce 327) ((T_PUBLIC) reduce 327) ((T_PUBLIC_SET) reduce 327) ((T_READONLY) reduce 327) ((T_STATIC) reduce 327) ((T_USE) reduce 327) ((T_VAR) reduce 327) (("}") reduce 327)) 7415 (((T_ABSTRACT) reduce 324) ((T_ATTRIBUTE) reduce 324) ((T_CASE) reduce 324) ((T_CONST) reduce 324) ((T_FINAL) reduce 324) ((T_FUNCTION) reduce 324) ((T_PRIVATE) reduce 324) ((T_PRIVATE_SET) reduce 324) ((T_PROTECTED) reduce 324) ((T_PROTECTED_SET) reduce 324) ((T_PUBLIC) reduce 324) ((T_PUBLIC_SET) reduce 324) ((T_READONLY) reduce 324) ((T_STATIC) reduce 324) ((T_USE) reduce 324) ((T_VAR) reduce 324) (("}") reduce 324)) 7416 (((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_AS) shift) ((T_BREAK) shift) ((T_CALLABLE) shift) ((T_CASE) shift) ((T_CATCH) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONST) shift) ((T_CONTINUE) shift) ((T_DECLARE) shift) ((T_DEFAULT) shift) ((T_DIR) shift) ((T_DO) shift) ((T_ECHO) shift) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) shift) ((T_ENDDECLARE) shift) ((T_ENDFOR) shift) ((T_ENDFOREACH) shift) ((T_ENDIF) shift) ((T_ENDSWITCH) shift) ((T_ENDWHILE) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_EXTENDS) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FINALLY) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_IF) shift) ((T_IMPLEMENTS) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) shift) ((T_INSTEADOF) shift) ((T_INTERFACE) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAMESPACE) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_PRINT) shift) ((T_PRIVATE) shift) ((T_PROPERTY_C) shift) ((T_PROTECTED) shift) ((T_PUBLIC) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_USE) shift) ((T_VAR) shift) ((T_WHILE) shift) ((T_YIELD) shift) (("}") shift)) 7417 (((T_AS) reduce 74)) 7418 (((T_AS) reduce 52)) 7419 (((T_AS) reduce 27)) 7420 (((T_AS) reduce 51)) 7421 (((T_AS) reduce 53)) 7422 (((T_AS) reduce 49)) 7423 (((T_AS) reduce 29)) 7424 (((T_AS) reduce 59)) 7425 (((T_AS) reduce 60)) 7426 (((T_AS) reduce 11)) 7427 (((T_AS) reduce 42)) 7428 (((T_AS) reduce 39)) 7429 (((T_AS) reduce 25)) 7430 (((T_AS) reduce 50)) 7431 (((T_AS) reduce 66)) 7432 (((T_AS) reduce 18)) 7433 (((T_AS) reduce 17)) 7434 (((T_AS) reduce 15)) 7435 (((T_AS) reduce 14)) 7436 (((T_AS) reduce 38)) 7437 (((T_AS) reduce 26)) 7438 (((T_AS) reduce 22)) 7439 (((T_AS) reduce 24)) 7440 (((T_AS) reduce 16)) 7441 (((T_AS) reduce 48)) 7442 (((T_AS) reduce 20)) 7443 (((T_AS) reduce 70)) 7444 (((T_AS) reduce 3)) 7445 (((T_AS) reduce 12)) 7446 (((T_AS) reduce 54)) 7447 (((T_AS) reduce 65)) 7448 (((T_AS) reduce 75)) 7449 (((T_AS) reduce 30)) 7450 (((T_AS) reduce 68)) 7451 (((T_AS) reduce 21)) 7452 (((T_AS) reduce 23)) 7453 (((T_AS) reduce 41)) 7454 (((T_AS) reduce 62)) 7455 (((T_AS) reduce 34)) 7456 (((T_AS) reduce 40)) 7457 (((T_AS) reduce 13)) 7458 (((T_AS) reduce 55)) 7459 (((T_AS) reduce 1)) 7460 (((T_AS) reduce 2)) 7461 (((T_AS) reduce 9)) 7462 (((T_AS) reduce 33)) 7463 (((T_AS) reduce 58)) 7464 (((T_AS) reduce 37)) 7465 (((T_AS) reduce 64)) 7466 (((T_AS) reduce 46)) 7467 (((T_AS) reduce 8)) 7468 (((T_AS) reduce 6)) 7469 (((T_AS) reduce 7)) 7470 (((T_AS) reduce 69)) 7471 (((T_AS) reduce 63)) 7472 (((T_AS) reduce 56)) 7473 (((T_PAAMAYIM_NEKUDOTAYIM) reduce 94)) 7474 (((T_PAAMAYIM_NEKUDOTAYIM) reduce 93)) 7475 (((T_PAAMAYIM_NEKUDOTAYIM) reduce 95)) 7476 (((T_AS) reduce 10)) 7477 (((T_AS) reduce 67)) 7478 (((T_AS) reduce 44)) 7479 (((T_AS) reduce 76)) 7480 (((T_AS) reduce 71)) 7481 (((T_AS) reduce 77)) 7482 (((T_AS) reduce 78)) 7483 (((T_AS) reduce 79)) 7484 (((T_AS) reduce 4)) 7485 (((T_AS) reduce 5)) 7486 (((T_AS) reduce 43)) 7487 (((T_AS) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509)) 7488 (((T_AS) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92)) 7489 (((T_AS) reduce 47)) 7490 (((T_AS) reduce 31)) 7491 (((T_AS) reduce 57)) 7492 (((T_AS) reduce 61)) 7493 (((T_AS) reduce 28)) 7494 (((T_AS) reduce 36)) 7495 (((T_AS) reduce 32)) 7496 (((T_AS) reduce 35)) 7497 (((T_AS) reduce 19)) 7498 (((T_AS) reduce 45)) 7499 (((T_AS) reduce 340) ((T_INSTEADOF) shift)) 7501 (((T_AS) reduce 339)) 7502 (((T_PAAMAYIM_NEKUDOTAYIM) reduce 510)) 7503 (((T_AS) reduce 72)) 7504 (((T_AS) reduce 83)) 7505 (((T_ABSTRACT) reduce 330) ((T_ARRAY) reduce 330) ((T_AS) reduce 330) ((T_BREAK) reduce 330) ((T_CALLABLE) reduce 330) ((T_CASE) reduce 330) ((T_CATCH) reduce 330) ((T_CLASS) reduce 330) ((T_CLASS_C) reduce 330) ((T_CLONE) reduce 330) ((T_CONST) reduce 330) ((T_CONTINUE) reduce 330) ((T_DECLARE) reduce 330) ((T_DEFAULT) reduce 330) ((T_DIR) reduce 330) ((T_DO) reduce 330) ((T_ECHO) reduce 330) ((T_ELSE) reduce 330) ((T_ELSEIF) reduce 330) ((T_EMPTY) reduce 330) ((T_ENDDECLARE) reduce 330) ((T_ENDFOR) reduce 330) ((T_ENDFOREACH) reduce 330) ((T_ENDIF) reduce 330) ((T_ENDSWITCH) reduce 330) ((T_ENDWHILE) reduce 330) ((T_ENUM) reduce 330) ((T_EVAL) reduce 330) ((T_EXIT) reduce 330) ((T_EXTENDS) reduce 330) ((T_FILE) reduce 330) ((T_FINAL) reduce 330) ((T_FINALLY) reduce 330) ((T_FN) reduce 330) ((T_FOR) reduce 330) ((T_FOREACH) reduce 330) ((T_FUNCTION) reduce 330) ((T_FUNC_C) reduce 330) ((T_GLOBAL) reduce 330) ((T_GOTO) reduce 330) ((T_IF) reduce 330) ((T_IMPLEMENTS) reduce 330) ((T_INCLUDE) reduce 330) ((T_INCLUDE_ONCE) reduce 330) ((T_INSTANCEOF) reduce 330) ((T_INSTEADOF) reduce 330) ((T_INTERFACE) reduce 330) ((T_ISSET) reduce 330) ((T_LINE) reduce 330) ((T_LIST) reduce 330) ((T_LOGICAL_AND) reduce 330) ((T_LOGICAL_OR) reduce 330) ((T_LOGICAL_XOR) reduce 330) ((T_MATCH) reduce 330) ((T_METHOD_C) reduce 330) ((T_NAMESPACE) reduce 330) ((T_NAME_FULLY_QUALIFIED) reduce 330) ((T_NAME_QUALIFIED) reduce 330) ((T_NAME_RELATIVE) reduce 330) ((T_NEW) reduce 330) ((T_NS_C) reduce 330) ((T_PRINT) reduce 330) ((T_PRIVATE) reduce 330) ((T_PROPERTY_C) reduce 330) ((T_PROTECTED) reduce 330) ((T_PUBLIC) reduce 330) ((T_READONLY) reduce 330) ((T_REQUIRE) reduce 330) ((T_REQUIRE_ONCE) reduce 330) ((T_RETURN) reduce 330) ((T_STATIC) reduce 330) ((T_STRING) reduce 330) ((T_SWITCH) reduce 330) ((T_THROW) reduce 330) ((T_TRAIT) reduce 330) ((T_TRAIT_C) reduce 330) ((T_TRY) reduce 330) ((T_UNSET) reduce 330) ((T_USE) reduce 330) ((T_VAR) reduce 330) ((T_WHILE) reduce 330) ((T_YIELD) reduce 330) (("}") reduce 330)) 7507 (((";") shift)) 7508 (((T_AS) shift)) 7510 (((T_ABSTRACT) reduce 328) ((T_ATTRIBUTE) reduce 328) ((T_CASE) reduce 328) ((T_CONST) reduce 328) ((T_FINAL) reduce 328) ((T_FUNCTION) reduce 328) ((T_PRIVATE) reduce 328) ((T_PRIVATE_SET) reduce 328) ((T_PROTECTED) reduce 328) ((T_PROTECTED_SET) reduce 328) ((T_PUBLIC) reduce 328) ((T_PUBLIC_SET) reduce 328) ((T_READONLY) reduce 328) ((T_STATIC) reduce 328) ((T_USE) reduce 328) ((T_VAR) reduce 328) (("}") reduce 328)) 7511 (((T_ABSTRACT) reduce 332) ((T_ARRAY) reduce 332) ((T_AS) reduce 332) ((T_BREAK) reduce 332) ((T_CALLABLE) reduce 332) ((T_CASE) reduce 332) ((T_CATCH) reduce 332) ((T_CLASS) reduce 332) ((T_CLASS_C) reduce 332) ((T_CLONE) reduce 332) ((T_CONST) reduce 332) ((T_CONTINUE) reduce 332) ((T_DECLARE) reduce 332) ((T_DEFAULT) reduce 332) ((T_DIR) reduce 332) ((T_DO) reduce 332) ((T_ECHO) reduce 332) ((T_ELSE) reduce 332) ((T_ELSEIF) reduce 332) ((T_EMPTY) reduce 332) ((T_ENDDECLARE) reduce 332) ((T_ENDFOR) reduce 332) ((T_ENDFOREACH) reduce 332) ((T_ENDIF) reduce 332) ((T_ENDSWITCH) reduce 332) ((T_ENDWHILE) reduce 332) ((T_ENUM) reduce 332) ((T_EVAL) reduce 332) ((T_EXIT) reduce 332) ((T_EXTENDS) reduce 332) ((T_FILE) reduce 332) ((T_FINAL) reduce 332) ((T_FINALLY) reduce 332) ((T_FN) reduce 332) ((T_FOR) reduce 332) ((T_FOREACH) reduce 332) ((T_FUNCTION) reduce 332) ((T_FUNC_C) reduce 332) ((T_GLOBAL) reduce 332) ((T_GOTO) reduce 332) ((T_IF) reduce 332) ((T_IMPLEMENTS) reduce 332) ((T_INCLUDE) reduce 332) ((T_INCLUDE_ONCE) reduce 332) ((T_INSTANCEOF) reduce 332) ((T_INSTEADOF) reduce 332) ((T_INTERFACE) reduce 332) ((T_ISSET) reduce 332) ((T_LINE) reduce 332) ((T_LIST) reduce 332) ((T_LOGICAL_AND) reduce 332) ((T_LOGICAL_OR) reduce 332) ((T_LOGICAL_XOR) reduce 332) ((T_MATCH) reduce 332) ((T_METHOD_C) reduce 332) ((T_NAMESPACE) reduce 332) ((T_NAME_FULLY_QUALIFIED) reduce 332) ((T_NAME_QUALIFIED) reduce 332) ((T_NAME_RELATIVE) reduce 332) ((T_NEW) reduce 332) ((T_NS_C) reduce 332) ((T_PRINT) reduce 332) ((T_PRIVATE) reduce 332) ((T_PROPERTY_C) reduce 332) ((T_PROTECTED) reduce 332) ((T_PUBLIC) reduce 332) ((T_READONLY) reduce 332) ((T_REQUIRE) reduce 332) ((T_REQUIRE_ONCE) reduce 332) ((T_RETURN) reduce 332) ((T_STATIC) reduce 332) ((T_STRING) reduce 332) ((T_SWITCH) reduce 332) ((T_THROW) reduce 332) ((T_TRAIT) reduce 332) ((T_TRAIT_C) reduce 332) ((T_TRY) reduce 332) ((T_UNSET) reduce 332) ((T_USE) reduce 332) ((T_VAR) reduce 332) ((T_WHILE) reduce 332) ((T_YIELD) reduce 332) (("}") reduce 332)) 7512 (((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_AS) shift) ((T_BREAK) shift) ((T_CALLABLE) shift) ((T_CASE) shift) ((T_CATCH) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONST) shift) ((T_CONTINUE) shift) ((T_DECLARE) shift) ((T_DEFAULT) shift) ((T_DIR) shift) ((T_DO) shift) ((T_ECHO) shift) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) shift) ((T_ENDDECLARE) shift) ((T_ENDFOR) shift) ((T_ENDFOREACH) shift) ((T_ENDIF) shift) ((T_ENDSWITCH) shift) ((T_ENDWHILE) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_EXTENDS) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FINALLY) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_IF) shift) ((T_IMPLEMENTS) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) shift) ((T_INSTEADOF) shift) ((T_INTERFACE) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAMESPACE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_PRINT) shift) ((T_PRIVATE) shift) ((T_PRIVATE_SET) shift) ((T_PROPERTY_C) shift) ((T_PROTECTED) shift) ((T_PROTECTED_SET) shift) ((T_PUBLIC) shift) ((T_PUBLIC_SET) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_USE) shift) ((T_VAR) shift) ((T_WHILE) shift) ((T_YIELD) shift)) 7513 (((";") reduce 359) ((T_ABSTRACT) reduce 359) ((T_ARRAY) reduce 359) ((T_AS) reduce 359) ((T_BREAK) reduce 359) ((T_CALLABLE) reduce 359) ((T_CASE) reduce 359) ((T_CATCH) reduce 359) ((T_CLASS) reduce 359) ((T_CLASS_C) reduce 359) ((T_CLONE) reduce 359) ((T_CONST) reduce 359) ((T_CONTINUE) reduce 359) ((T_DECLARE) reduce 359) ((T_DEFAULT) reduce 359) ((T_DIR) reduce 359) ((T_DO) reduce 359) ((T_ECHO) reduce 359) ((T_ELSE) reduce 359) ((T_ELSEIF) reduce 359) ((T_EMPTY) reduce 359) ((T_ENDDECLARE) reduce 359) ((T_ENDFOR) reduce 359) ((T_ENDFOREACH) reduce 359) ((T_ENDIF) reduce 359) ((T_ENDSWITCH) reduce 359) ((T_ENDWHILE) reduce 359) ((T_ENUM) reduce 359) ((T_EVAL) reduce 359) ((T_EXIT) reduce 359) ((T_EXTENDS) reduce 359) ((T_FILE) reduce 359) ((T_FINAL) reduce 359) ((T_FINALLY) reduce 359) ((T_FN) reduce 359) ((T_FOR) reduce 359) ((T_FOREACH) reduce 359) ((T_FUNCTION) reduce 359) ((T_FUNC_C) reduce 359) ((T_GLOBAL) reduce 359) ((T_GOTO) reduce 359) ((T_IF) reduce 359) ((T_IMPLEMENTS) reduce 359) ((T_INCLUDE) reduce 359) ((T_INCLUDE_ONCE) reduce 359) ((T_INSTANCEOF) reduce 359) ((T_INSTEADOF) reduce 359) ((T_INTERFACE) reduce 359) ((T_ISSET) reduce 359) ((T_LINE) reduce 359) ((T_LIST) reduce 359) ((T_LOGICAL_AND) reduce 359) ((T_LOGICAL_OR) reduce 359) ((T_LOGICAL_XOR) reduce 359) ((T_MATCH) reduce 359) ((T_METHOD_C) reduce 359) ((T_NAMESPACE) reduce 359) ((T_NEW) reduce 359) ((T_NS_C) reduce 359) ((T_PRINT) reduce 359) ((T_PRIVATE) reduce 359) ((T_PROPERTY_C) reduce 359) ((T_PROTECTED) reduce 359) ((T_PUBLIC) reduce 359) ((T_READONLY) reduce 359) ((T_REQUIRE) reduce 359) ((T_REQUIRE_ONCE) reduce 359) ((T_RETURN) reduce 359) ((T_STATIC) reduce 359) ((T_STRING) reduce 359) ((T_SWITCH) reduce 359) ((T_THROW) reduce 359) ((T_TRAIT) reduce 359) ((T_TRAIT_C) reduce 359) ((T_TRY) reduce 359) ((T_UNSET) reduce 359) ((T_USE) reduce 359) ((T_VAR) reduce 359) ((T_WHILE) reduce 359) ((T_YIELD) reduce 359)) 7514 (((";") reduce 52)) 7515 (((";") reduce 27)) 7516 (((";") reduce 51)) 7517 (((";") reduce 53)) 7518 (((";") reduce 49)) 7519 (((";") reduce 29)) 7520 (((";") reduce 59)) 7521 (((";") reduce 60)) 7522 (((";") reduce 11)) 7523 (((";") reduce 42)) 7524 (((";") reduce 39)) 7525 (((";") reduce 25)) 7526 (((";") reduce 50)) 7527 (((";") reduce 66)) 7528 (((";") reduce 18)) 7529 (((";") reduce 17)) 7530 (((";") reduce 15)) 7531 (((";") reduce 14)) 7532 (((";") reduce 38)) 7533 (((";") reduce 26)) 7534 (((";") reduce 22)) 7535 (((";") reduce 24)) 7536 (((";") reduce 16)) 7537 (((";") reduce 48)) 7538 (((";") reduce 20)) 7539 (((";") reduce 70)) 7540 (((";") reduce 3)) 7541 (((";") reduce 12)) 7542 (((";") reduce 54)) 7543 (((";") reduce 65)) 7544 (((";") reduce 360) ((T_ABSTRACT) reduce 360) ((T_ARRAY) reduce 360) ((T_AS) reduce 360) ((T_BREAK) reduce 360) ((T_CALLABLE) reduce 360) ((T_CASE) reduce 360) ((T_CATCH) reduce 360) ((T_CLASS) reduce 360) ((T_CLASS_C) reduce 360) ((T_CLONE) reduce 360) ((T_CONST) reduce 360) ((T_CONTINUE) reduce 360) ((T_DECLARE) reduce 360) ((T_DEFAULT) reduce 360) ((T_DIR) reduce 360) ((T_DO) reduce 360) ((T_ECHO) reduce 360) ((T_ELSE) reduce 360) ((T_ELSEIF) reduce 360) ((T_EMPTY) reduce 360) ((T_ENDDECLARE) reduce 360) ((T_ENDFOR) reduce 360) ((T_ENDFOREACH) reduce 360) ((T_ENDIF) reduce 360) ((T_ENDSWITCH) reduce 360) ((T_ENDWHILE) reduce 360) ((T_ENUM) reduce 360) ((T_EVAL) reduce 360) ((T_EXIT) reduce 360) ((T_EXTENDS) reduce 360) ((T_FILE) reduce 360) ((T_FINAL) reduce 360) ((T_FINALLY) reduce 360) ((T_FN) reduce 360) ((T_FOR) reduce 360) ((T_FOREACH) reduce 360) ((T_FUNCTION) reduce 360) ((T_FUNC_C) reduce 360) ((T_GLOBAL) reduce 360) ((T_GOTO) reduce 360) ((T_IF) reduce 360) ((T_IMPLEMENTS) reduce 360) ((T_INCLUDE) reduce 360) ((T_INCLUDE_ONCE) reduce 360) ((T_INSTANCEOF) reduce 360) ((T_INSTEADOF) reduce 360) ((T_INTERFACE) reduce 360) ((T_ISSET) reduce 360) ((T_LINE) reduce 360) ((T_LIST) reduce 360) ((T_LOGICAL_AND) reduce 360) ((T_LOGICAL_OR) reduce 360) ((T_LOGICAL_XOR) reduce 360) ((T_MATCH) reduce 360) ((T_METHOD_C) reduce 360) ((T_NAMESPACE) reduce 360) ((T_NEW) reduce 360) ((T_NS_C) reduce 360) ((T_PRINT) reduce 360) ((T_PRIVATE) reduce 360) ((T_PROPERTY_C) reduce 360) ((T_PROTECTED) reduce 360) ((T_PUBLIC) reduce 360) ((T_READONLY) reduce 360) ((T_REQUIRE) reduce 360) ((T_REQUIRE_ONCE) reduce 360) ((T_RETURN) reduce 360) ((T_STATIC) reduce 360) ((T_STRING) reduce 360) ((T_SWITCH) reduce 360) ((T_THROW) reduce 360) ((T_TRAIT) reduce 360) ((T_TRAIT_C) reduce 360) ((T_TRY) reduce 360) ((T_UNSET) reduce 360) ((T_USE) reduce 360) ((T_VAR) reduce 360) ((T_WHILE) reduce 360) ((T_YIELD) reduce 360)) 7545 (((";") reduce 30)) 7546 (((";") reduce 68)) 7547 (((";") reduce 21)) 7548 (((";") reduce 23)) 7549 (((";") reduce 41)) 7550 (((";") reduce 62)) 7551 (((";") reduce 34)) 7552 (((";") reduce 40)) 7553 (((";") reduce 13)) 7554 (((";") reduce 55)) 7555 (((";") reduce 1)) 7556 (((";") reduce 2)) 7557 (((";") reduce 9)) 7558 (((";") reduce 33)) 7559 (((";") reduce 58)) 7560 (((";") reduce 37)) 7561 (((";") reduce 64)) 7562 (((";") reduce 46)) 7563 (((";") reduce 8)) 7564 (((";") reduce 6)) 7565 (((";") reduce 7)) 7566 (((";") reduce 69)) 7567 (((";") reduce 63)) 7568 (((";") reduce 56)) 7569 (((";") reduce 10)) 7570 (((";") reduce 67)) 7571 (((";") reduce 44)) 7572 (((";") reduce 354) ((T_ABSTRACT) reduce 354) ((T_ARRAY) reduce 354) ((T_AS) reduce 354) ((T_BREAK) reduce 354) ((T_CALLABLE) reduce 354) ((T_CASE) reduce 354) ((T_CATCH) reduce 354) ((T_CLASS) reduce 354) ((T_CLASS_C) reduce 354) ((T_CLONE) reduce 354) ((T_CONST) reduce 354) ((T_CONTINUE) reduce 354) ((T_DECLARE) reduce 354) ((T_DEFAULT) reduce 354) ((T_DIR) reduce 354) ((T_DO) reduce 354) ((T_ECHO) reduce 354) ((T_ELSE) reduce 354) ((T_ELSEIF) reduce 354) ((T_EMPTY) reduce 354) ((T_ENDDECLARE) reduce 354) ((T_ENDFOR) reduce 354) ((T_ENDFOREACH) reduce 354) ((T_ENDIF) reduce 354) ((T_ENDSWITCH) reduce 354) ((T_ENDWHILE) reduce 354) ((T_ENUM) reduce 354) ((T_EVAL) reduce 354) ((T_EXIT) reduce 354) ((T_EXTENDS) reduce 354) ((T_FILE) reduce 354) ((T_FINAL) reduce 354) ((T_FINALLY) reduce 354) ((T_FN) reduce 354) ((T_FOR) reduce 354) ((T_FOREACH) reduce 354) ((T_FUNCTION) reduce 354) ((T_FUNC_C) reduce 354) ((T_GLOBAL) reduce 354) ((T_GOTO) reduce 354) ((T_IF) reduce 354) ((T_IMPLEMENTS) reduce 354) ((T_INCLUDE) reduce 354) ((T_INCLUDE_ONCE) reduce 354) ((T_INSTANCEOF) reduce 354) ((T_INSTEADOF) reduce 354) ((T_INTERFACE) reduce 354) ((T_ISSET) reduce 354) ((T_LINE) reduce 354) ((T_LIST) reduce 354) ((T_LOGICAL_AND) reduce 354) ((T_LOGICAL_OR) reduce 354) ((T_LOGICAL_XOR) reduce 354) ((T_MATCH) reduce 354) ((T_METHOD_C) reduce 354) ((T_NAMESPACE) reduce 354) ((T_NEW) reduce 354) ((T_NS_C) reduce 354) ((T_PRINT) reduce 354) ((T_PRIVATE) reduce 354) ((T_PROPERTY_C) reduce 354) ((T_PROTECTED) reduce 354) ((T_PUBLIC) reduce 354) ((T_READONLY) reduce 354) ((T_REQUIRE) reduce 354) ((T_REQUIRE_ONCE) reduce 354) ((T_RETURN) reduce 354) ((T_STATIC) reduce 354) ((T_STRING) reduce 354) ((T_SWITCH) reduce 354) ((T_THROW) reduce 354) ((T_TRAIT) reduce 354) ((T_TRAIT_C) reduce 354) ((T_TRY) reduce 354) ((T_UNSET) reduce 354) ((T_USE) reduce 354) ((T_VAR) reduce 354) ((T_WHILE) reduce 354) ((T_YIELD) reduce 354)) 7573 (((";") reduce 357) ((T_ABSTRACT) reduce 357) ((T_ARRAY) reduce 357) ((T_AS) reduce 357) ((T_BREAK) reduce 357) ((T_CALLABLE) reduce 357) ((T_CASE) reduce 357) ((T_CATCH) reduce 357) ((T_CLASS) reduce 357) ((T_CLASS_C) reduce 357) ((T_CLONE) reduce 357) ((T_CONST) reduce 357) ((T_CONTINUE) reduce 357) ((T_DECLARE) reduce 357) ((T_DEFAULT) reduce 357) ((T_DIR) reduce 357) ((T_DO) reduce 357) ((T_ECHO) reduce 357) ((T_ELSE) reduce 357) ((T_ELSEIF) reduce 357) ((T_EMPTY) reduce 357) ((T_ENDDECLARE) reduce 357) ((T_ENDFOR) reduce 357) ((T_ENDFOREACH) reduce 357) ((T_ENDIF) reduce 357) ((T_ENDSWITCH) reduce 357) ((T_ENDWHILE) reduce 357) ((T_ENUM) reduce 357) ((T_EVAL) reduce 357) ((T_EXIT) reduce 357) ((T_EXTENDS) reduce 357) ((T_FILE) reduce 357) ((T_FINAL) reduce 357) ((T_FINALLY) reduce 357) ((T_FN) reduce 357) ((T_FOR) reduce 357) ((T_FOREACH) reduce 357) ((T_FUNCTION) reduce 357) ((T_FUNC_C) reduce 357) ((T_GLOBAL) reduce 357) ((T_GOTO) reduce 357) ((T_IF) reduce 357) ((T_IMPLEMENTS) reduce 357) ((T_INCLUDE) reduce 357) ((T_INCLUDE_ONCE) reduce 357) ((T_INSTANCEOF) reduce 357) ((T_INSTEADOF) reduce 357) ((T_INTERFACE) reduce 357) ((T_ISSET) reduce 357) ((T_LINE) reduce 357) ((T_LIST) reduce 357) ((T_LOGICAL_AND) reduce 357) ((T_LOGICAL_OR) reduce 357) ((T_LOGICAL_XOR) reduce 357) ((T_MATCH) reduce 357) ((T_METHOD_C) reduce 357) ((T_NAMESPACE) reduce 357) ((T_NEW) reduce 357) ((T_NS_C) reduce 357) ((T_PRINT) reduce 357) ((T_PRIVATE) reduce 357) ((T_PROPERTY_C) reduce 357) ((T_PROTECTED) reduce 357) ((T_PUBLIC) reduce 357) ((T_READONLY) reduce 357) ((T_REQUIRE) reduce 357) ((T_REQUIRE_ONCE) reduce 357) ((T_RETURN) reduce 357) ((T_STATIC) reduce 357) ((T_STRING) reduce 357) ((T_SWITCH) reduce 357) ((T_THROW) reduce 357) ((T_TRAIT) reduce 357) ((T_TRAIT_C) reduce 357) ((T_TRY) reduce 357) ((T_UNSET) reduce 357) ((T_USE) reduce 357) ((T_VAR) reduce 357) ((T_WHILE) reduce 357) ((T_YIELD) reduce 357)) 7574 (((";") reduce 71)) 7575 (((";") reduce 353) ((T_ABSTRACT) reduce 353) ((T_ARRAY) reduce 353) ((T_AS) reduce 353) ((T_BREAK) reduce 353) ((T_CALLABLE) reduce 353) ((T_CASE) reduce 353) ((T_CATCH) reduce 353) ((T_CLASS) reduce 353) ((T_CLASS_C) reduce 353) ((T_CLONE) reduce 353) ((T_CONST) reduce 353) ((T_CONTINUE) reduce 353) ((T_DECLARE) reduce 353) ((T_DEFAULT) reduce 353) ((T_DIR) reduce 353) ((T_DO) reduce 353) ((T_ECHO) reduce 353) ((T_ELSE) reduce 353) ((T_ELSEIF) reduce 353) ((T_EMPTY) reduce 353) ((T_ENDDECLARE) reduce 353) ((T_ENDFOR) reduce 353) ((T_ENDFOREACH) reduce 353) ((T_ENDIF) reduce 353) ((T_ENDSWITCH) reduce 353) ((T_ENDWHILE) reduce 353) ((T_ENUM) reduce 353) ((T_EVAL) reduce 353) ((T_EXIT) reduce 353) ((T_EXTENDS) reduce 353) ((T_FILE) reduce 353) ((T_FINAL) reduce 353) ((T_FINALLY) reduce 353) ((T_FN) reduce 353) ((T_FOR) reduce 353) ((T_FOREACH) reduce 353) ((T_FUNCTION) reduce 353) ((T_FUNC_C) reduce 353) ((T_GLOBAL) reduce 353) ((T_GOTO) reduce 353) ((T_IF) reduce 353) ((T_IMPLEMENTS) reduce 353) ((T_INCLUDE) reduce 353) ((T_INCLUDE_ONCE) reduce 353) ((T_INSTANCEOF) reduce 353) ((T_INSTEADOF) reduce 353) ((T_INTERFACE) reduce 353) ((T_ISSET) reduce 353) ((T_LINE) reduce 353) ((T_LIST) reduce 353) ((T_LOGICAL_AND) reduce 353) ((T_LOGICAL_OR) reduce 353) ((T_LOGICAL_XOR) reduce 353) ((T_MATCH) reduce 353) ((T_METHOD_C) reduce 353) ((T_NAMESPACE) reduce 353) ((T_NEW) reduce 353) ((T_NS_C) reduce 353) ((T_PRINT) reduce 353) ((T_PRIVATE) reduce 353) ((T_PROPERTY_C) reduce 353) ((T_PROTECTED) reduce 353) ((T_PUBLIC) reduce 353) ((T_READONLY) reduce 353) ((T_REQUIRE) reduce 353) ((T_REQUIRE_ONCE) reduce 353) ((T_RETURN) reduce 353) ((T_STATIC) reduce 353) ((T_STRING) reduce 353) ((T_SWITCH) reduce 353) ((T_THROW) reduce 353) ((T_TRAIT) reduce 353) ((T_TRAIT_C) reduce 353) ((T_TRY) reduce 353) ((T_UNSET) reduce 353) ((T_USE) reduce 353) ((T_VAR) reduce 353) ((T_WHILE) reduce 353) ((T_YIELD) reduce 353)) 7576 (((";") reduce 356) ((T_ABSTRACT) reduce 356) ((T_ARRAY) reduce 356) ((T_AS) reduce 356) ((T_BREAK) reduce 356) ((T_CALLABLE) reduce 356) ((T_CASE) reduce 356) ((T_CATCH) reduce 356) ((T_CLASS) reduce 356) ((T_CLASS_C) reduce 356) ((T_CLONE) reduce 356) ((T_CONST) reduce 356) ((T_CONTINUE) reduce 356) ((T_DECLARE) reduce 356) ((T_DEFAULT) reduce 356) ((T_DIR) reduce 356) ((T_DO) reduce 356) ((T_ECHO) reduce 356) ((T_ELSE) reduce 356) ((T_ELSEIF) reduce 356) ((T_EMPTY) reduce 356) ((T_ENDDECLARE) reduce 356) ((T_ENDFOR) reduce 356) ((T_ENDFOREACH) reduce 356) ((T_ENDIF) reduce 356) ((T_ENDSWITCH) reduce 356) ((T_ENDWHILE) reduce 356) ((T_ENUM) reduce 356) ((T_EVAL) reduce 356) ((T_EXIT) reduce 356) ((T_EXTENDS) reduce 356) ((T_FILE) reduce 356) ((T_FINAL) reduce 356) ((T_FINALLY) reduce 356) ((T_FN) reduce 356) ((T_FOR) reduce 356) ((T_FOREACH) reduce 356) ((T_FUNCTION) reduce 356) ((T_FUNC_C) reduce 356) ((T_GLOBAL) reduce 356) ((T_GOTO) reduce 356) ((T_IF) reduce 356) ((T_IMPLEMENTS) reduce 356) ((T_INCLUDE) reduce 356) ((T_INCLUDE_ONCE) reduce 356) ((T_INSTANCEOF) reduce 356) ((T_INSTEADOF) reduce 356) ((T_INTERFACE) reduce 356) ((T_ISSET) reduce 356) ((T_LINE) reduce 356) ((T_LIST) reduce 356) ((T_LOGICAL_AND) reduce 356) ((T_LOGICAL_OR) reduce 356) ((T_LOGICAL_XOR) reduce 356) ((T_MATCH) reduce 356) ((T_METHOD_C) reduce 356) ((T_NAMESPACE) reduce 356) ((T_NEW) reduce 356) ((T_NS_C) reduce 356) ((T_PRINT) reduce 356) ((T_PRIVATE) reduce 356) ((T_PROPERTY_C) reduce 356) ((T_PROTECTED) reduce 356) ((T_PUBLIC) reduce 356) ((T_READONLY) reduce 356) ((T_REQUIRE) reduce 356) ((T_REQUIRE_ONCE) reduce 356) ((T_RETURN) reduce 356) ((T_STATIC) reduce 356) ((T_STRING) reduce 356) ((T_SWITCH) reduce 356) ((T_THROW) reduce 356) ((T_TRAIT) reduce 356) ((T_TRAIT_C) reduce 356) ((T_TRY) reduce 356) ((T_UNSET) reduce 356) ((T_USE) reduce 356) ((T_VAR) reduce 356) ((T_WHILE) reduce 356) ((T_YIELD) reduce 356)) 7577 (((";") reduce 352) ((T_ABSTRACT) reduce 352) ((T_ARRAY) reduce 352) ((T_AS) reduce 352) ((T_BREAK) reduce 352) ((T_CALLABLE) reduce 352) ((T_CASE) reduce 352) ((T_CATCH) reduce 352) ((T_CLASS) reduce 352) ((T_CLASS_C) reduce 352) ((T_CLONE) reduce 352) ((T_CONST) reduce 352) ((T_CONTINUE) reduce 352) ((T_DECLARE) reduce 352) ((T_DEFAULT) reduce 352) ((T_DIR) reduce 352) ((T_DO) reduce 352) ((T_ECHO) reduce 352) ((T_ELSE) reduce 352) ((T_ELSEIF) reduce 352) ((T_EMPTY) reduce 352) ((T_ENDDECLARE) reduce 352) ((T_ENDFOR) reduce 352) ((T_ENDFOREACH) reduce 352) ((T_ENDIF) reduce 352) ((T_ENDSWITCH) reduce 352) ((T_ENDWHILE) reduce 352) ((T_ENUM) reduce 352) ((T_EVAL) reduce 352) ((T_EXIT) reduce 352) ((T_EXTENDS) reduce 352) ((T_FILE) reduce 352) ((T_FINAL) reduce 352) ((T_FINALLY) reduce 352) ((T_FN) reduce 352) ((T_FOR) reduce 352) ((T_FOREACH) reduce 352) ((T_FUNCTION) reduce 352) ((T_FUNC_C) reduce 352) ((T_GLOBAL) reduce 352) ((T_GOTO) reduce 352) ((T_IF) reduce 352) ((T_IMPLEMENTS) reduce 352) ((T_INCLUDE) reduce 352) ((T_INCLUDE_ONCE) reduce 352) ((T_INSTANCEOF) reduce 352) ((T_INSTEADOF) reduce 352) ((T_INTERFACE) reduce 352) ((T_ISSET) reduce 352) ((T_LINE) reduce 352) ((T_LIST) reduce 352) ((T_LOGICAL_AND) reduce 352) ((T_LOGICAL_OR) reduce 352) ((T_LOGICAL_XOR) reduce 352) ((T_MATCH) reduce 352) ((T_METHOD_C) reduce 352) ((T_NAMESPACE) reduce 352) ((T_NEW) reduce 352) ((T_NS_C) reduce 352) ((T_PRINT) reduce 352) ((T_PRIVATE) reduce 352) ((T_PROPERTY_C) reduce 352) ((T_PROTECTED) reduce 352) ((T_PUBLIC) reduce 352) ((T_READONLY) reduce 352) ((T_REQUIRE) reduce 352) ((T_REQUIRE_ONCE) reduce 352) ((T_RETURN) reduce 352) ((T_STATIC) reduce 352) ((T_STRING) reduce 352) ((T_SWITCH) reduce 352) ((T_THROW) reduce 352) ((T_TRAIT) reduce 352) ((T_TRAIT_C) reduce 352) ((T_TRY) reduce 352) ((T_UNSET) reduce 352) ((T_USE) reduce 352) ((T_VAR) reduce 352) ((T_WHILE) reduce 352) ((T_YIELD) reduce 352)) 7578 (((";") reduce 355) ((T_ABSTRACT) reduce 355) ((T_ARRAY) reduce 355) ((T_AS) reduce 355) ((T_BREAK) reduce 355) ((T_CALLABLE) reduce 355) ((T_CASE) reduce 355) ((T_CATCH) reduce 355) ((T_CLASS) reduce 355) ((T_CLASS_C) reduce 355) ((T_CLONE) reduce 355) ((T_CONST) reduce 355) ((T_CONTINUE) reduce 355) ((T_DECLARE) reduce 355) ((T_DEFAULT) reduce 355) ((T_DIR) reduce 355) ((T_DO) reduce 355) ((T_ECHO) reduce 355) ((T_ELSE) reduce 355) ((T_ELSEIF) reduce 355) ((T_EMPTY) reduce 355) ((T_ENDDECLARE) reduce 355) ((T_ENDFOR) reduce 355) ((T_ENDFOREACH) reduce 355) ((T_ENDIF) reduce 355) ((T_ENDSWITCH) reduce 355) ((T_ENDWHILE) reduce 355) ((T_ENUM) reduce 355) ((T_EVAL) reduce 355) ((T_EXIT) reduce 355) ((T_EXTENDS) reduce 355) ((T_FILE) reduce 355) ((T_FINAL) reduce 355) ((T_FINALLY) reduce 355) ((T_FN) reduce 355) ((T_FOR) reduce 355) ((T_FOREACH) reduce 355) ((T_FUNCTION) reduce 355) ((T_FUNC_C) reduce 355) ((T_GLOBAL) reduce 355) ((T_GOTO) reduce 355) ((T_IF) reduce 355) ((T_IMPLEMENTS) reduce 355) ((T_INCLUDE) reduce 355) ((T_INCLUDE_ONCE) reduce 355) ((T_INSTANCEOF) reduce 355) ((T_INSTEADOF) reduce 355) ((T_INTERFACE) reduce 355) ((T_ISSET) reduce 355) ((T_LINE) reduce 355) ((T_LIST) reduce 355) ((T_LOGICAL_AND) reduce 355) ((T_LOGICAL_OR) reduce 355) ((T_LOGICAL_XOR) reduce 355) ((T_MATCH) reduce 355) ((T_METHOD_C) reduce 355) ((T_NAMESPACE) reduce 355) ((T_NEW) reduce 355) ((T_NS_C) reduce 355) ((T_PRINT) reduce 355) ((T_PRIVATE) reduce 355) ((T_PROPERTY_C) reduce 355) ((T_PROTECTED) reduce 355) ((T_PUBLIC) reduce 355) ((T_READONLY) reduce 355) ((T_REQUIRE) reduce 355) ((T_REQUIRE_ONCE) reduce 355) ((T_RETURN) reduce 355) ((T_STATIC) reduce 355) ((T_STRING) reduce 355) ((T_SWITCH) reduce 355) ((T_THROW) reduce 355) ((T_TRAIT) reduce 355) ((T_TRAIT_C) reduce 355) ((T_TRY) reduce 355) ((T_UNSET) reduce 355) ((T_USE) reduce 355) ((T_VAR) reduce 355) ((T_WHILE) reduce 355) ((T_YIELD) reduce 355)) 7579 (((";") reduce 361) ((T_ABSTRACT) reduce 361) ((T_ARRAY) reduce 361) ((T_AS) reduce 361) ((T_BREAK) reduce 361) ((T_CALLABLE) reduce 361) ((T_CASE) reduce 361) ((T_CATCH) reduce 361) ((T_CLASS) reduce 361) ((T_CLASS_C) reduce 361) ((T_CLONE) reduce 361) ((T_CONST) reduce 361) ((T_CONTINUE) reduce 361) ((T_DECLARE) reduce 361) ((T_DEFAULT) reduce 361) ((T_DIR) reduce 361) ((T_DO) reduce 361) ((T_ECHO) reduce 361) ((T_ELSE) reduce 361) ((T_ELSEIF) reduce 361) ((T_EMPTY) reduce 361) ((T_ENDDECLARE) reduce 361) ((T_ENDFOR) reduce 361) ((T_ENDFOREACH) reduce 361) ((T_ENDIF) reduce 361) ((T_ENDSWITCH) reduce 361) ((T_ENDWHILE) reduce 361) ((T_ENUM) reduce 361) ((T_EVAL) reduce 361) ((T_EXIT) reduce 361) ((T_EXTENDS) reduce 361) ((T_FILE) reduce 361) ((T_FINAL) reduce 361) ((T_FINALLY) reduce 361) ((T_FN) reduce 361) ((T_FOR) reduce 361) ((T_FOREACH) reduce 361) ((T_FUNCTION) reduce 361) ((T_FUNC_C) reduce 361) ((T_GLOBAL) reduce 361) ((T_GOTO) reduce 361) ((T_IF) reduce 361) ((T_IMPLEMENTS) reduce 361) ((T_INCLUDE) reduce 361) ((T_INCLUDE_ONCE) reduce 361) ((T_INSTANCEOF) reduce 361) ((T_INSTEADOF) reduce 361) ((T_INTERFACE) reduce 361) ((T_ISSET) reduce 361) ((T_LINE) reduce 361) ((T_LIST) reduce 361) ((T_LOGICAL_AND) reduce 361) ((T_LOGICAL_OR) reduce 361) ((T_LOGICAL_XOR) reduce 361) ((T_MATCH) reduce 361) ((T_METHOD_C) reduce 361) ((T_NAMESPACE) reduce 361) ((T_NEW) reduce 361) ((T_NS_C) reduce 361) ((T_PRINT) reduce 361) ((T_PRIVATE) reduce 361) ((T_PROPERTY_C) reduce 361) ((T_PROTECTED) reduce 361) ((T_PUBLIC) reduce 361) ((T_READONLY) reduce 361) ((T_REQUIRE) reduce 361) ((T_REQUIRE_ONCE) reduce 361) ((T_RETURN) reduce 361) ((T_STATIC) reduce 361) ((T_STRING) reduce 361) ((T_SWITCH) reduce 361) ((T_THROW) reduce 361) ((T_TRAIT) reduce 361) ((T_TRAIT_C) reduce 361) ((T_TRY) reduce 361) ((T_UNSET) reduce 361) ((T_USE) reduce 361) ((T_VAR) reduce 361) ((T_WHILE) reduce 361) ((T_YIELD) reduce 361)) 7580 (((";") reduce 4)) 7581 (((";") reduce 5)) 7582 (((";") reduce 43)) 7583 (((";") reduce 358) ((T_ABSTRACT) reduce 358) ((T_ARRAY) reduce 358) ((T_AS) reduce 358) ((T_BREAK) reduce 358) ((T_CALLABLE) reduce 358) ((T_CASE) reduce 358) ((T_CATCH) reduce 358) ((T_CLASS) reduce 358) ((T_CLASS_C) reduce 358) ((T_CLONE) reduce 358) ((T_CONST) reduce 358) ((T_CONTINUE) reduce 358) ((T_DECLARE) reduce 358) ((T_DEFAULT) reduce 358) ((T_DIR) reduce 358) ((T_DO) reduce 358) ((T_ECHO) reduce 358) ((T_ELSE) reduce 358) ((T_ELSEIF) reduce 358) ((T_EMPTY) reduce 358) ((T_ENDDECLARE) reduce 358) ((T_ENDFOR) reduce 358) ((T_ENDFOREACH) reduce 358) ((T_ENDIF) reduce 358) ((T_ENDSWITCH) reduce 358) ((T_ENDWHILE) reduce 358) ((T_ENUM) reduce 358) ((T_EVAL) reduce 358) ((T_EXIT) reduce 358) ((T_EXTENDS) reduce 358) ((T_FILE) reduce 358) ((T_FINAL) reduce 358) ((T_FINALLY) reduce 358) ((T_FN) reduce 358) ((T_FOR) reduce 358) ((T_FOREACH) reduce 358) ((T_FUNCTION) reduce 358) ((T_FUNC_C) reduce 358) ((T_GLOBAL) reduce 358) ((T_GOTO) reduce 358) ((T_IF) reduce 358) ((T_IMPLEMENTS) reduce 358) ((T_INCLUDE) reduce 358) ((T_INCLUDE_ONCE) reduce 358) ((T_INSTANCEOF) reduce 358) ((T_INSTEADOF) reduce 358) ((T_INTERFACE) reduce 358) ((T_ISSET) reduce 358) ((T_LINE) reduce 358) ((T_LIST) reduce 358) ((T_LOGICAL_AND) reduce 358) ((T_LOGICAL_OR) reduce 358) ((T_LOGICAL_XOR) reduce 358) ((T_MATCH) reduce 358) ((T_METHOD_C) reduce 358) ((T_NAMESPACE) reduce 358) ((T_NEW) reduce 358) ((T_NS_C) reduce 358) ((T_PRINT) reduce 358) ((T_PRIVATE) reduce 358) ((T_PROPERTY_C) reduce 358) ((T_PROTECTED) reduce 358) ((T_PUBLIC) reduce 358) ((T_READONLY) reduce 358) ((T_REQUIRE) reduce 358) ((T_REQUIRE_ONCE) reduce 358) ((T_RETURN) reduce 358) ((T_STATIC) reduce 358) ((T_STRING) reduce 358) ((T_SWITCH) reduce 358) ((T_THROW) reduce 358) ((T_TRAIT) reduce 358) ((T_TRAIT_C) reduce 358) ((T_TRY) reduce 358) ((T_UNSET) reduce 358) ((T_USE) reduce 358) ((T_VAR) reduce 358) ((T_WHILE) reduce 358) ((T_YIELD) reduce 358)) 7584 (((";") reduce 335)) 7585 (((";") reduce 47)) 7586 (((";") reduce 31)) 7587 (((";") reduce 57)) 7588 (((";") reduce 61)) 7589 (((";") reduce 28)) 7590 (((";") reduce 36)) 7591 (((";") reduce 32)) 7592 (((";") reduce 35)) 7593 (((";") reduce 19)) 7594 (((";") reduce 45)) 7595 (((";") reduce 338) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_AS) shift) ((T_BREAK) shift) ((T_CALLABLE) shift) ((T_CASE) shift) ((T_CATCH) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONST) shift) ((T_CONTINUE) shift) ((T_DECLARE) shift) ((T_DEFAULT) shift) ((T_DIR) shift) ((T_DO) shift) ((T_ECHO) shift) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) shift) ((T_ENDDECLARE) shift) ((T_ENDFOR) shift) ((T_ENDFOREACH) shift) ((T_ENDIF) shift) ((T_ENDSWITCH) shift) ((T_ENDWHILE) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_EXTENDS) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FINALLY) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_IF) shift) ((T_IMPLEMENTS) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) shift) ((T_INSTEADOF) shift) ((T_INTERFACE) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAMESPACE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_PRINT) shift) ((T_PRIVATE) shift) ((T_PROPERTY_C) shift) ((T_PROTECTED) shift) ((T_PUBLIC) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_USE) shift) ((T_VAR) shift) ((T_WHILE) shift) ((T_YIELD) shift)) 7596 (((";") reduce 336)) 7597 (((";") reduce 74)) 7598 (((";") reduce 75)) 7599 (((";") reduce 76)) 7600 (((";") reduce 77)) 7601 (((";") reduce 78)) 7602 (((";") reduce 79)) 7603 (((";") reduce 73)) 7604 (((";") reduce 82)) 7605 (((";") reduce 337)) 7606 (((";") reduce 72)) 7607 (((";") reduce 83)) 7608 (((T_ABSTRACT) reduce 333) ((T_ARRAY) reduce 333) ((T_AS) reduce 333) ((T_BREAK) reduce 333) ((T_CALLABLE) reduce 333) ((T_CASE) reduce 333) ((T_CATCH) reduce 333) ((T_CLASS) reduce 333) ((T_CLASS_C) reduce 333) ((T_CLONE) reduce 333) ((T_CONST) reduce 333) ((T_CONTINUE) reduce 333) ((T_DECLARE) reduce 333) ((T_DEFAULT) reduce 333) ((T_DIR) reduce 333) ((T_DO) reduce 333) ((T_ECHO) reduce 333) ((T_ELSE) reduce 333) ((T_ELSEIF) reduce 333) ((T_EMPTY) reduce 333) ((T_ENDDECLARE) reduce 333) ((T_ENDFOR) reduce 333) ((T_ENDFOREACH) reduce 333) ((T_ENDIF) reduce 333) ((T_ENDSWITCH) reduce 333) ((T_ENDWHILE) reduce 333) ((T_ENUM) reduce 333) ((T_EVAL) reduce 333) ((T_EXIT) reduce 333) ((T_EXTENDS) reduce 333) ((T_FILE) reduce 333) ((T_FINAL) reduce 333) ((T_FINALLY) reduce 333) ((T_FN) reduce 333) ((T_FOR) reduce 333) ((T_FOREACH) reduce 333) ((T_FUNCTION) reduce 333) ((T_FUNC_C) reduce 333) ((T_GLOBAL) reduce 333) ((T_GOTO) reduce 333) ((T_IF) reduce 333) ((T_IMPLEMENTS) reduce 333) ((T_INCLUDE) reduce 333) ((T_INCLUDE_ONCE) reduce 333) ((T_INSTANCEOF) reduce 333) ((T_INSTEADOF) reduce 333) ((T_INTERFACE) reduce 333) ((T_ISSET) reduce 333) ((T_LINE) reduce 333) ((T_LIST) reduce 333) ((T_LOGICAL_AND) reduce 333) ((T_LOGICAL_OR) reduce 333) ((T_LOGICAL_XOR) reduce 333) ((T_MATCH) reduce 333) ((T_METHOD_C) reduce 333) ((T_NAMESPACE) reduce 333) ((T_NAME_FULLY_QUALIFIED) reduce 333) ((T_NAME_QUALIFIED) reduce 333) ((T_NAME_RELATIVE) reduce 333) ((T_NEW) reduce 333) ((T_NS_C) reduce 333) ((T_PRINT) reduce 333) ((T_PRIVATE) reduce 333) ((T_PROPERTY_C) reduce 333) ((T_PROTECTED) reduce 333) ((T_PUBLIC) reduce 333) ((T_READONLY) reduce 333) ((T_REQUIRE) reduce 333) ((T_REQUIRE_ONCE) reduce 333) ((T_RETURN) reduce 333) ((T_STATIC) reduce 333) ((T_STRING) reduce 333) ((T_SWITCH) reduce 333) ((T_THROW) reduce 333) ((T_TRAIT) reduce 333) ((T_TRAIT_C) reduce 333) ((T_TRY) reduce 333) ((T_UNSET) reduce 333) ((T_USE) reduce 333) ((T_VAR) reduce 333) ((T_WHILE) reduce 333) ((T_YIELD) reduce 333) (("}") reduce 333)) 7609 (((T_ABSTRACT) reduce 331) ((T_ARRAY) reduce 331) ((T_AS) reduce 331) ((T_BREAK) reduce 331) ((T_CALLABLE) reduce 331) ((T_CASE) reduce 331) ((T_CATCH) reduce 331) ((T_CLASS) reduce 331) ((T_CLASS_C) reduce 331) ((T_CLONE) reduce 331) ((T_CONST) reduce 331) ((T_CONTINUE) reduce 331) ((T_DECLARE) reduce 331) ((T_DEFAULT) reduce 331) ((T_DIR) reduce 331) ((T_DO) reduce 331) ((T_ECHO) reduce 331) ((T_ELSE) reduce 331) ((T_ELSEIF) reduce 331) ((T_EMPTY) reduce 331) ((T_ENDDECLARE) reduce 331) ((T_ENDFOR) reduce 331) ((T_ENDFOREACH) reduce 331) ((T_ENDIF) reduce 331) ((T_ENDSWITCH) reduce 331) ((T_ENDWHILE) reduce 331) ((T_ENUM) reduce 331) ((T_EVAL) reduce 331) ((T_EXIT) reduce 331) ((T_EXTENDS) reduce 331) ((T_FILE) reduce 331) ((T_FINAL) reduce 331) ((T_FINALLY) reduce 331) ((T_FN) reduce 331) ((T_FOR) reduce 331) ((T_FOREACH) reduce 331) ((T_FUNCTION) reduce 331) ((T_FUNC_C) reduce 331) ((T_GLOBAL) reduce 331) ((T_GOTO) reduce 331) ((T_IF) reduce 331) ((T_IMPLEMENTS) reduce 331) ((T_INCLUDE) reduce 331) ((T_INCLUDE_ONCE) reduce 331) ((T_INSTANCEOF) reduce 331) ((T_INSTEADOF) reduce 331) ((T_INTERFACE) reduce 331) ((T_ISSET) reduce 331) ((T_LINE) reduce 331) ((T_LIST) reduce 331) ((T_LOGICAL_AND) reduce 331) ((T_LOGICAL_OR) reduce 331) ((T_LOGICAL_XOR) reduce 331) ((T_MATCH) reduce 331) ((T_METHOD_C) reduce 331) ((T_NAMESPACE) reduce 331) ((T_NAME_FULLY_QUALIFIED) reduce 331) ((T_NAME_QUALIFIED) reduce 331) ((T_NAME_RELATIVE) reduce 331) ((T_NEW) reduce 331) ((T_NS_C) reduce 331) ((T_PRINT) reduce 331) ((T_PRIVATE) reduce 331) ((T_PROPERTY_C) reduce 331) ((T_PROTECTED) reduce 331) ((T_PUBLIC) reduce 331) ((T_READONLY) reduce 331) ((T_REQUIRE) reduce 331) ((T_REQUIRE_ONCE) reduce 331) ((T_RETURN) reduce 331) ((T_STATIC) reduce 331) ((T_STRING) reduce 331) ((T_SWITCH) reduce 331) ((T_THROW) reduce 331) ((T_TRAIT) reduce 331) ((T_TRAIT_C) reduce 331) ((T_TRY) reduce 331) ((T_UNSET) reduce 331) ((T_USE) reduce 331) ((T_VAR) reduce 331) ((T_WHILE) reduce 331) ((T_YIELD) reduce 331) (("}") reduce 331)) 7610 (((T_ABSTRACT) reduce 329) ((T_ATTRIBUTE) reduce 329) ((T_CASE) reduce 329) ((T_CONST) reduce 329) ((T_FINAL) reduce 329) ((T_FUNCTION) reduce 329) ((T_PRIVATE) reduce 329) ((T_PRIVATE_SET) reduce 329) ((T_PROTECTED) reduce 329) ((T_PROTECTED_SET) reduce 329) ((T_PUBLIC) reduce 329) ((T_PUBLIC_SET) reduce 329) ((T_READONLY) reduce 329) ((T_STATIC) reduce 329) ((T_USE) reduce 329) ((T_VAR) reduce 329) (("}") reduce 329)) 7612 (((T_AS) reduce 74) ((T_INSTEADOF) reduce 74)) 7613 (((T_AS) reduce 52) ((T_INSTEADOF) reduce 52)) 7614 (((T_AS) reduce 27) ((T_INSTEADOF) reduce 27)) 7615 (((T_AS) reduce 51) ((T_INSTEADOF) reduce 51)) 7616 (((T_AS) reduce 53) ((T_INSTEADOF) reduce 53)) 7617 (((T_AS) reduce 49) ((T_INSTEADOF) reduce 49)) 7618 (((T_AS) reduce 29) ((T_INSTEADOF) reduce 29)) 7619 (((T_AS) reduce 59) ((T_INSTEADOF) reduce 59)) 7620 (((T_AS) reduce 60) ((T_INSTEADOF) reduce 60)) 7621 (((T_AS) reduce 11) ((T_INSTEADOF) reduce 11)) 7622 (((T_AS) reduce 42) ((T_INSTEADOF) reduce 42)) 7623 (((T_AS) reduce 39) ((T_INSTEADOF) reduce 39)) 7624 (((T_AS) reduce 25) ((T_INSTEADOF) reduce 25)) 7625 (((T_AS) reduce 50) ((T_INSTEADOF) reduce 50)) 7626 (((T_AS) reduce 66) ((T_INSTEADOF) reduce 66)) 7627 (((T_AS) reduce 18) ((T_INSTEADOF) reduce 18)) 7628 (((T_AS) reduce 17) ((T_INSTEADOF) reduce 17)) 7629 (((T_AS) reduce 15) ((T_INSTEADOF) reduce 15)) 7630 (((T_AS) reduce 14) ((T_INSTEADOF) reduce 14)) 7631 (((T_AS) reduce 38) ((T_INSTEADOF) reduce 38)) 7632 (((T_AS) reduce 26) ((T_INSTEADOF) reduce 26)) 7633 (((T_AS) reduce 22) ((T_INSTEADOF) reduce 22)) 7634 (((T_AS) reduce 24) ((T_INSTEADOF) reduce 24)) 7635 (((T_AS) reduce 16) ((T_INSTEADOF) reduce 16)) 7636 (((T_AS) reduce 48) ((T_INSTEADOF) reduce 48)) 7637 (((T_AS) reduce 20) ((T_INSTEADOF) reduce 20)) 7638 (((T_AS) reduce 70) ((T_INSTEADOF) reduce 70)) 7639 (((T_AS) reduce 3) ((T_INSTEADOF) reduce 3)) 7640 (((T_AS) reduce 12) ((T_INSTEADOF) reduce 12)) 7641 (((T_AS) reduce 54) ((T_INSTEADOF) reduce 54)) 7642 (((T_AS) reduce 65) ((T_INSTEADOF) reduce 65)) 7643 (((T_AS) reduce 75) ((T_INSTEADOF) reduce 75)) 7644 (((T_AS) reduce 30) ((T_INSTEADOF) reduce 30)) 7645 (((T_AS) reduce 68) ((T_INSTEADOF) reduce 68)) 7646 (((T_AS) reduce 21) ((T_INSTEADOF) reduce 21)) 7647 (((T_AS) reduce 23) ((T_INSTEADOF) reduce 23)) 7648 (((T_AS) reduce 41) ((T_INSTEADOF) reduce 41)) 7649 (((T_AS) reduce 62) ((T_INSTEADOF) reduce 62)) 7650 (((T_AS) reduce 34) ((T_INSTEADOF) reduce 34)) 7651 (((T_AS) reduce 40) ((T_INSTEADOF) reduce 40)) 7652 (((T_AS) reduce 13) ((T_INSTEADOF) reduce 13)) 7653 (((T_AS) reduce 55) ((T_INSTEADOF) reduce 55)) 7654 (((T_AS) reduce 1) ((T_INSTEADOF) reduce 1)) 7655 (((T_AS) reduce 2) ((T_INSTEADOF) reduce 2)) 7656 (((T_AS) reduce 9) ((T_INSTEADOF) reduce 9)) 7657 (((T_AS) reduce 33) ((T_INSTEADOF) reduce 33)) 7658 (((T_AS) reduce 58) ((T_INSTEADOF) reduce 58)) 7659 (((T_AS) reduce 37) ((T_INSTEADOF) reduce 37)) 7660 (((T_AS) reduce 64) ((T_INSTEADOF) reduce 64)) 7661 (((T_AS) reduce 46) ((T_INSTEADOF) reduce 46)) 7662 (((T_AS) reduce 8) ((T_INSTEADOF) reduce 8)) 7663 (((T_AS) reduce 6) ((T_INSTEADOF) reduce 6)) 7664 (((T_AS) reduce 7) ((T_INSTEADOF) reduce 7)) 7665 (((T_AS) reduce 69) ((T_INSTEADOF) reduce 69)) 7666 (((T_AS) reduce 63) ((T_INSTEADOF) reduce 63)) 7667 (((T_AS) reduce 56) ((T_INSTEADOF) reduce 56)) 7668 (((T_AS) reduce 10) ((T_INSTEADOF) reduce 10)) 7669 (((T_AS) reduce 67) ((T_INSTEADOF) reduce 67)) 7670 (((T_AS) reduce 44) ((T_INSTEADOF) reduce 44)) 7671 (((T_AS) reduce 76) ((T_INSTEADOF) reduce 76)) 7672 (((T_AS) reduce 71) ((T_INSTEADOF) reduce 71)) 7673 (((T_AS) reduce 77) ((T_INSTEADOF) reduce 77)) 7674 (((T_AS) reduce 78) ((T_INSTEADOF) reduce 78)) 7675 (((T_AS) reduce 79) ((T_INSTEADOF) reduce 79)) 7676 (((T_AS) reduce 4) ((T_INSTEADOF) reduce 4)) 7677 (((T_AS) reduce 5) ((T_INSTEADOF) reduce 5)) 7678 (((T_AS) reduce 43) ((T_INSTEADOF) reduce 43)) 7679 (((T_AS) reduce 73) ((T_INSTEADOF) reduce 73)) 7680 (((T_AS) reduce 82) ((T_INSTEADOF) reduce 82)) 7681 (((T_AS) reduce 47) ((T_INSTEADOF) reduce 47)) 7682 (((T_AS) reduce 31) ((T_INSTEADOF) reduce 31)) 7683 (((T_AS) reduce 57) ((T_INSTEADOF) reduce 57)) 7684 (((T_AS) reduce 61) ((T_INSTEADOF) reduce 61)) 7685 (((T_AS) reduce 28) ((T_INSTEADOF) reduce 28)) 7686 (((T_AS) reduce 36) ((T_INSTEADOF) reduce 36)) 7687 (((T_AS) reduce 32) ((T_INSTEADOF) reduce 32)) 7688 (((T_AS) reduce 35) ((T_INSTEADOF) reduce 35)) 7689 (((T_AS) reduce 19) ((T_INSTEADOF) reduce 19)) 7690 (((T_AS) reduce 45) ((T_INSTEADOF) reduce 45)) 7691 (((T_AS) reduce 341) ((T_INSTEADOF) reduce 341)) 7692 (((T_AS) reduce 72) ((T_INSTEADOF) reduce 72)) 7693 (((T_AS) reduce 83) ((T_INSTEADOF) reduce 83)) 7695 (((",") reduce 94) ((";") reduce 94)) 7696 (((",") reduce 93) ((";") reduce 93)) 7697 (((",") reduce 95) ((";") reduce 95)) 7698 (((",") reduce 509) ((";") reduce 509)) 7699 (((",") reduce 92) ((";") reduce 92)) 7700 (((",") reduce 325) ((";") reduce 325)) 7701 (((",") shift) ((";") reduce 334)) 7702 (((",") reduce 510) ((";") reduce 510)) 7704 (((",") reduce 326) ((";") reduce 326)) 7705 (((",") reduce 326) ((";") reduce 326) (("{") reduce 326)) 7707 (((";") reduce 74) (("=") reduce 74)) 7708 (((";") reduce 52) (("=") reduce 52)) 7709 (((";") reduce 27) (("=") reduce 27)) 7710 (((";") reduce 51) (("=") reduce 51)) 7711 (((";") reduce 53) (("=") reduce 53)) 7712 (((";") reduce 49) (("=") reduce 49)) 7713 (((";") reduce 29) (("=") reduce 29)) 7714 (((";") reduce 59) (("=") reduce 59)) 7715 (((";") reduce 60) (("=") reduce 60)) 7716 (((";") reduce 11) (("=") reduce 11)) 7717 (((";") reduce 42) (("=") reduce 42)) 7718 (((";") reduce 39) (("=") reduce 39)) 7719 (((";") reduce 25) (("=") reduce 25)) 7720 (((";") reduce 50) (("=") reduce 50)) 7721 (((";") reduce 66) (("=") reduce 66)) 7722 (((";") reduce 18) (("=") reduce 18)) 7723 (((";") reduce 17) (("=") reduce 17)) 7724 (((";") reduce 15) (("=") reduce 15)) 7725 (((";") reduce 14) (("=") reduce 14)) 7726 (((";") reduce 38) (("=") reduce 38)) 7727 (((";") reduce 26) (("=") reduce 26)) 7728 (((";") reduce 22) (("=") reduce 22)) 7729 (((";") reduce 24) (("=") reduce 24)) 7730 (((";") reduce 16) (("=") reduce 16)) 7731 (((";") reduce 48) (("=") reduce 48)) 7732 (((";") reduce 20) (("=") reduce 20)) 7733 (((";") reduce 70) (("=") reduce 70)) 7734 (((";") reduce 3) (("=") reduce 3)) 7735 (((";") reduce 12) (("=") reduce 12)) 7736 (((";") reduce 54) (("=") reduce 54)) 7737 (((";") reduce 65) (("=") reduce 65)) 7738 (((";") reduce 75) (("=") reduce 75)) 7739 (((";") reduce 30) (("=") reduce 30)) 7740 (((";") reduce 68) (("=") reduce 68)) 7741 (((";") reduce 21) (("=") reduce 21)) 7742 (((";") reduce 23) (("=") reduce 23)) 7743 (((";") reduce 41) (("=") reduce 41)) 7744 (((";") reduce 62) (("=") reduce 62)) 7745 (((";") reduce 34) (("=") reduce 34)) 7746 (((";") reduce 40) (("=") reduce 40)) 7747 (((";") reduce 13) (("=") reduce 13)) 7748 (((";") reduce 55) (("=") reduce 55)) 7749 (((";") reduce 1) (("=") reduce 1)) 7750 (((";") reduce 2) (("=") reduce 2)) 7751 (((";") reduce 9) (("=") reduce 9)) 7752 (((";") reduce 33) (("=") reduce 33)) 7753 (((";") reduce 58) (("=") reduce 58)) 7754 (((";") reduce 37) (("=") reduce 37)) 7755 (((";") reduce 64) (("=") reduce 64)) 7756 (((";") reduce 46) (("=") reduce 46)) 7757 (((";") reduce 8) (("=") reduce 8)) 7758 (((";") reduce 6) (("=") reduce 6)) 7759 (((";") reduce 7) (("=") reduce 7)) 7760 (((";") reduce 69) (("=") reduce 69)) 7761 (((";") reduce 63) (("=") reduce 63)) 7762 (((";") reduce 56) (("=") reduce 56)) 7763 (((";") reduce 10) (("=") reduce 10)) 7764 (((";") reduce 67) (("=") reduce 67)) 7765 (((";") reduce 44) (("=") reduce 44)) 7766 (((";") reduce 76) (("=") reduce 76)) 7767 (((";") reduce 71) (("=") reduce 71)) 7768 (((";") reduce 77) (("=") reduce 77)) 7769 (((";") reduce 78) (("=") reduce 78)) 7770 (((";") reduce 79) (("=") reduce 79)) 7771 (((";") reduce 4) (("=") reduce 4)) 7772 (((";") reduce 5) (("=") reduce 5)) 7773 (((";") reduce 43) (("=") reduce 43)) 7774 (((";") reduce 73) (("=") reduce 73)) 7775 (((";") reduce 82) (("=") reduce 82)) 7776 (((";") reduce 47) (("=") reduce 47)) 7777 (((";") reduce 31) (("=") reduce 31)) 7778 (((";") reduce 57) (("=") reduce 57)) 7779 (((";") reduce 61) (("=") reduce 61)) 7780 (((";") reduce 28) (("=") reduce 28)) 7781 (((";") reduce 36) (("=") reduce 36)) 7782 (((";") reduce 32) (("=") reduce 32)) 7783 (((";") reduce 35) (("=") reduce 35)) 7784 (((";") reduce 19) (("=") reduce 19)) 7785 (((";") reduce 45) (("=") reduce 45)) 7786 (((";") reduce 206) (("=") shift)) 7787 (((";") reduce 72) (("=") reduce 72)) 7788 (((";") reduce 83) (("=") reduce 83)) 7791 (((T_ABSTRACT) reduce 205) ((T_ATTRIBUTE) reduce 205) ((T_CASE) reduce 205) ((T_CONST) reduce 205) ((T_FINAL) reduce 205) ((T_FUNCTION) reduce 205) ((T_PRIVATE) reduce 205) ((T_PRIVATE_SET) reduce 205) ((T_PROTECTED) reduce 205) ((T_PROTECTED_SET) reduce 205) ((T_PUBLIC) reduce 205) ((T_PUBLIC_SET) reduce 205) ((T_READONLY) reduce 205) ((T_STATIC) reduce 205) ((T_USE) reduce 205) ((T_VAR) reduce 205) (("}") reduce 205)) 7792 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 207) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 7795 (((T_ABSTRACT) reduce 100) ((T_ATTRIBUTE) reduce 100) ((T_CASE) reduce 100) ((T_CONST) reduce 100) ((T_FINAL) reduce 100) ((T_FUNCTION) reduce 100) ((T_PRIVATE) reduce 100) ((T_PRIVATE_SET) reduce 100) ((T_PROTECTED) reduce 100) ((T_PROTECTED_SET) reduce 100) ((T_PUBLIC) reduce 100) ((T_PUBLIC_SET) reduce 100) ((T_READONLY) reduce 100) ((T_STATIC) reduce 100) ((T_VAR) reduce 100)) 7796 (((",") reduce 94) (("{") reduce 94)) 7797 (((",") reduce 93) (("{") reduce 93)) 7798 (((",") reduce 95) (("{") reduce 95)) 7799 (((",") reduce 509) (("{") reduce 509)) 7800 (((",") reduce 92) (("{") reduce 92)) 7801 (((",") reduce 325) (("{") reduce 325)) 7802 (((",") shift) (("{") reduce 213)) 7803 (((",") reduce 510) (("{") reduce 510)) 7805 (((",") reduce 326) (("{") reduce 326)) 7806 (((T_IMPLEMENTS) reduce 94) (("{") reduce 94)) 7807 (((T_IMPLEMENTS) reduce 93) (("{") reduce 93)) 7808 (((T_IMPLEMENTS) reduce 95) (("{") reduce 95)) 7809 (((T_IMPLEMENTS) reduce 509) (("{") reduce 509)) 7810 (((T_IMPLEMENTS) reduce 92) (("{") reduce 92)) 7811 (((T_IMPLEMENTS) reduce 209) (("{") reduce 209)) 7812 (((T_IMPLEMENTS) reduce 510) (("{") reduce 510)) 7813 (((T_EXTENDS) reduce 291) ((T_IMPLEMENTS) reduce 291) (("{") reduce 291)) 7817 (((T_EXTENDS) reduce 292) ((T_IMPLEMENTS) reduce 292) (("{") reduce 292)) 7818 (((T_EXTENDS) reduce 293) ((T_IMPLEMENTS) reduce 293) (("{") reduce 293)) 7819 (((T_ABSTRACT) reduce 194) ((T_CLASS) reduce 194) ((T_FINAL) reduce 194) ((T_READONLY) reduce 194)) 7822 (((T_ABSTRACT) reduce 100) ((T_ATTRIBUTE) reduce 100) ((T_CLASS) reduce 100) ((T_FINAL) reduce 100) ((T_READONLY) reduce 100)) 7824 ((("%") reduce 513) (("(") reduce 513) ((")") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("|") reduce 513)) 7827 (((")") reduce 595) ((",") reduce 595) (("=") shift)) 7828 ((("%") reduce 467) ((")") reduce 467) (("*") reduce 467) (("+") reduce 467) ((",") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_PIPE) reduce 467) ((T_POW) shift) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("|") reduce 467)) 7829 ((("%") reduce 424) ((")") reduce 424) (("*") reduce 424) (("+") reduce 424) ((",") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("[") reduce 548) (("^") reduce 424) (("|") reduce 424)) 7830 ((("%") reduce 468) ((")") reduce 468) (("*") reduce 468) (("+") reduce 468) ((",") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_PIPE) reduce 468) ((T_POW) shift) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 7831 ((("%") reduce 426) ((")") reduce 426) (("*") reduce 426) (("+") reduce 426) ((",") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("[") reduce 548) (("^") reduce 426) (("|") reduce 426)) 7832 ((("%") reduce 472) ((")") reduce 472) (("*") reduce 472) (("+") reduce 472) ((",") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_PIPE) reduce 472) ((T_POW) shift) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("|") reduce 472)) 7833 ((("%") reduce 470) ((")") reduce 470) (("*") reduce 470) (("+") reduce 470) ((",") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_PIPE) reduce 470) ((T_POW) shift) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("|") reduce 470)) 7834 ((("%") reduce 475) ((")") reduce 475) (("*") reduce 475) (("+") reduce 475) ((",") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_PIPE) reduce 475) ((T_POW) shift) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("|") reduce 475)) 7835 ((("%") reduce 446) ((")") reduce 446) (("*") reduce 446) (("+") reduce 446) ((",") reduce 446) (("-") reduce 446) ((".") reduce 446) (("/") reduce 446) (("<") reduce 446) ((">") reduce 446) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_BOOLEAN_AND) reduce 446) ((T_BOOLEAN_OR) reduce 446) ((T_COALESCE) reduce 446) ((T_INSTANCEOF) reduce 446) ((T_IS_EQUAL) reduce 446) ((T_IS_GREATER_OR_EQUAL) reduce 446) ((T_IS_IDENTICAL) reduce 446) ((T_IS_NOT_EQUAL) reduce 446) ((T_IS_NOT_IDENTICAL) reduce 446) ((T_IS_SMALLER_OR_EQUAL) reduce 446) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_PIPE) reduce 446) ((T_POW) shift) ((T_SL) reduce 446) ((T_SPACESHIP) reduce 446) ((T_SR) reduce 446) (("^") reduce 446) (("|") reduce 446)) 7836 ((("%") reduce 445) ((")") reduce 445) (("*") reduce 445) (("+") reduce 445) ((",") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_PIPE) reduce 445) ((T_POW) shift) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("|") reduce 445)) 7838 ((("%") reduce 570) (("(") reduce 570) ((")") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) (("=") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AND_EQUAL) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_COALESCE_EQUAL) reduce 570) ((T_CONCAT_EQUAL) reduce 570) ((T_DEC) reduce 570) ((T_DIV_EQUAL) reduce 570) ((T_INC) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_MINUS_EQUAL) reduce 570) ((T_MOD_EQUAL) reduce 570) ((T_MUL_EQUAL) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_OR_EQUAL) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_PLUS_EQUAL) reduce 570) ((T_POW) reduce 570) ((T_POW_EQUAL) reduce 570) ((T_SL) reduce 570) ((T_SL_EQUAL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) ((T_SR_EQUAL) reduce 570) ((T_XOR_EQUAL) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 7841 ((("%") reduce 569) (("(") reduce 569) ((")") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) (("=") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AND_EQUAL) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_COALESCE_EQUAL) reduce 569) ((T_CONCAT_EQUAL) reduce 569) ((T_DEC) reduce 569) ((T_DIV_EQUAL) reduce 569) ((T_INC) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_MINUS_EQUAL) reduce 569) ((T_MOD_EQUAL) reduce 569) ((T_MUL_EQUAL) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_OR_EQUAL) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_PLUS_EQUAL) reduce 569) ((T_POW) reduce 569) ((T_POW_EQUAL) reduce 569) ((T_SL) reduce 569) ((T_SL_EQUAL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) ((T_SR_EQUAL) reduce 569) ((T_XOR_EQUAL) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 7843 ((("%") reduce 522) (("(") reduce 522) ((")") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_PIPE) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("|") reduce 522)) 7844 ((("%") reduce 447) ((")") reduce 447) (("*") reduce 447) (("+") reduce 447) ((",") reduce 447) (("-") reduce 447) ((".") reduce 447) (("/") reduce 447) (("<") reduce 447) ((">") reduce 447) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_BOOLEAN_AND) reduce 447) ((T_BOOLEAN_OR) reduce 447) ((T_COALESCE) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 447) ((T_IS_GREATER_OR_EQUAL) reduce 447) ((T_IS_IDENTICAL) reduce 447) ((T_IS_NOT_EQUAL) reduce 447) ((T_IS_NOT_IDENTICAL) reduce 447) ((T_IS_SMALLER_OR_EQUAL) reduce 447) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_PIPE) reduce 447) ((T_POW) shift) ((T_SL) reduce 447) ((T_SPACESHIP) reduce 447) ((T_SR) reduce 447) (("^") reduce 447) (("|") reduce 447)) 7846 ((("%") reduce 541) (("(") reduce 579) ((")") reduce 541) (("*") reduce 541) (("+") reduce 541) ((",") reduce 541) (("-") reduce 541) ((".") reduce 541) (("/") reduce 541) (("<") reduce 541) ((">") reduce 541) (("?") reduce 541) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_BOOLEAN_AND) reduce 541) ((T_BOOLEAN_OR) reduce 541) ((T_COALESCE) reduce 541) ((T_DOUBLE_ARROW) reduce 541) ((T_INSTANCEOF) reduce 541) ((T_IS_EQUAL) reduce 541) ((T_IS_GREATER_OR_EQUAL) reduce 541) ((T_IS_IDENTICAL) reduce 541) ((T_IS_NOT_EQUAL) reduce 541) ((T_IS_NOT_IDENTICAL) reduce 541) ((T_IS_SMALLER_OR_EQUAL) reduce 541) ((T_LOGICAL_AND) reduce 541) ((T_LOGICAL_OR) reduce 541) ((T_LOGICAL_XOR) reduce 541) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 541) ((T_OBJECT_OPERATOR) reduce 541) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 541) ((T_PIPE) reduce 541) ((T_POW) reduce 541) ((T_SL) reduce 541) ((T_SPACESHIP) reduce 541) ((T_SR) reduce 541) (("[") reduce 541) (("^") reduce 541) (("|") reduce 541)) 7848 ((("%") reduce 571) (("(") reduce 581) ((")") reduce 571) (("*") reduce 571) (("+") reduce 571) ((",") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) (("<") reduce 571) (("=") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AND_EQUAL) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_COALESCE_EQUAL) reduce 571) ((T_CONCAT_EQUAL) reduce 571) ((T_DEC) reduce 571) ((T_DIV_EQUAL) reduce 571) ((T_DOUBLE_ARROW) reduce 571) ((T_INC) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_MINUS_EQUAL) reduce 571) ((T_MOD_EQUAL) reduce 571) ((T_MUL_EQUAL) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_OR_EQUAL) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_PLUS_EQUAL) reduce 571) ((T_POW) reduce 571) ((T_POW_EQUAL) reduce 571) ((T_SL) reduce 571) ((T_SL_EQUAL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) ((T_SR_EQUAL) reduce 571) ((T_XOR_EQUAL) reduce 571) (("[") reduce 571) (("^") reduce 571) (("|") reduce 571)) 7851 ((("%") reduce 543) (("(") reduce 580) ((")") reduce 543) (("*") reduce 543) (("+") reduce 543) ((",") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PIPE) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 7852 ((("%") reduce 506) (("(") reduce 506) ((")") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("|") reduce 506)) 7853 ((("%") reduce 508) (("(") reduce 508) ((")") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("|") reduce 508)) 7855 ((("%") reduce 485) ((")") reduce 485) (("*") reduce 485) (("+") reduce 485) ((",") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_DOUBLE_ARROW) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_PIPE) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("^") reduce 485) (("|") reduce 485)) 7856 ((("%") reduce 487) ((")") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_PIPE) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("^") reduce 487) (("|") reduce 487)) 7859 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 404) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 404) (("^") shift) (("|") shift)) 7864 ((("%") reduce 560) (("(") reduce 560) ((")") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) (("=") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AND_EQUAL) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_COALESCE_EQUAL) reduce 560) ((T_CONCAT_EQUAL) reduce 560) ((T_DEC) reduce 560) ((T_DIV_EQUAL) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INC) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_MINUS_EQUAL) reduce 560) ((T_MOD_EQUAL) reduce 560) ((T_MUL_EQUAL) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_OR_EQUAL) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_PLUS_EQUAL) reduce 560) ((T_POW) reduce 560) ((T_POW_EQUAL) reduce 560) ((T_SL) reduce 560) ((T_SL_EQUAL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) ((T_SR_EQUAL) reduce 560) ((T_XOR_EQUAL) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 7865 ((("%") reduce 582) (("(") reduce 582) ((")") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) (("=") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AND_EQUAL) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_COALESCE_EQUAL) reduce 582) ((T_CONCAT_EQUAL) reduce 582) ((T_DEC) reduce 582) ((T_DIV_EQUAL) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INC) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_MINUS_EQUAL) reduce 582) ((T_MOD_EQUAL) reduce 582) ((T_MUL_EQUAL) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_OR_EQUAL) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_PLUS_EQUAL) reduce 582) ((T_POW) reduce 582) ((T_POW_EQUAL) reduce 582) ((T_SL) reduce 582) ((T_SL_EQUAL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) ((T_SR_EQUAL) reduce 582) ((T_XOR_EQUAL) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 7866 ((("%") reduce 566) (("(") shift) ((")") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 7867 ((("%") reduce 584) (("(") reduce 584) ((")") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) (("=") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AND_EQUAL) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_COALESCE_EQUAL) reduce 584) ((T_CONCAT_EQUAL) reduce 584) ((T_DEC) reduce 584) ((T_DIV_EQUAL) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INC) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_MINUS_EQUAL) reduce 584) ((T_MOD_EQUAL) reduce 584) ((T_MUL_EQUAL) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_OR_EQUAL) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_PLUS_EQUAL) reduce 584) ((T_POW) reduce 584) ((T_POW_EQUAL) reduce 584) ((T_SL) reduce 584) ((T_SL_EQUAL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) ((T_SR_EQUAL) reduce 584) ((T_XOR_EQUAL) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 7870 ((("%") reduce 583) (("(") reduce 583) ((")") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) (("=") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AND_EQUAL) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_COALESCE_EQUAL) reduce 583) ((T_CONCAT_EQUAL) reduce 583) ((T_DEC) reduce 583) ((T_DIV_EQUAL) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INC) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_MINUS_EQUAL) reduce 583) ((T_MOD_EQUAL) reduce 583) ((T_MUL_EQUAL) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_OR_EQUAL) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_PLUS_EQUAL) reduce 583) ((T_POW) reduce 583) ((T_POW_EQUAL) reduce 583) ((T_SL) reduce 583) ((T_SL_EQUAL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) ((T_SR_EQUAL) reduce 583) ((T_XOR_EQUAL) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 7871 ((("%") reduce 561) (("(") reduce 561) ((")") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) (("=") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AND_EQUAL) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_COALESCE_EQUAL) reduce 561) ((T_CONCAT_EQUAL) reduce 561) ((T_DEC) reduce 561) ((T_DIV_EQUAL) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INC) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_MINUS_EQUAL) reduce 561) ((T_MOD_EQUAL) reduce 561) ((T_MUL_EQUAL) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_OR_EQUAL) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_PLUS_EQUAL) reduce 561) ((T_POW) reduce 561) ((T_POW_EQUAL) reduce 561) ((T_SL) reduce 561) ((T_SL_EQUAL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) ((T_SR_EQUAL) reduce 561) ((T_XOR_EQUAL) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 7872 ((("%") reduce 567) (("(") shift) ((")") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 7873 ((("%") reduce 562) (("(") reduce 562) ((")") reduce 562) (("*") reduce 562) (("+") reduce 562) ((",") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) (("=") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AND_EQUAL) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_COALESCE_EQUAL) reduce 562) ((T_CONCAT_EQUAL) reduce 562) ((T_DEC) reduce 562) ((T_DIV_EQUAL) reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_INC) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_MINUS_EQUAL) reduce 562) ((T_MOD_EQUAL) reduce 562) ((T_MUL_EQUAL) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_OR_EQUAL) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_PLUS_EQUAL) reduce 562) ((T_POW) reduce 562) ((T_POW_EQUAL) reduce 562) ((T_SL) reduce 562) ((T_SL_EQUAL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) ((T_SR_EQUAL) reduce 562) ((T_XOR_EQUAL) reduce 562) (("[") reduce 562) (("^") reduce 562) (("|") reduce 562)) 7874 (((")") reduce 593) ((",") reduce 593) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) (("[") reduce 548)) 7876 ((("%") reduce 477) ((")") reduce 477) (("*") reduce 477) (("+") reduce 477) ((",") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_PIPE) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 7878 ((("%") reduce 520) (("(") reduce 520) ((")") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) (("=") shift) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_PIPE) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("|") reduce 520)) 7880 ((("%") shift) ((")") reduce 405) (("*") shift) (("+") shift) ((",") reduce 405) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 7881 ((("%") shift) ((")") reduce 482) (("*") shift) (("+") shift) ((",") reduce 482) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 482) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 7884 ((("%") shift) ((")") reduce 481) (("*") shift) (("+") shift) ((",") reduce 481) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 481) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 7885 ((("%") reduce 473) ((")") reduce 473) (("*") reduce 473) (("+") reduce 473) ((",") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_DOUBLE_ARROW) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_PIPE) reduce 473) ((T_POW) shift) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("|") reduce 473)) 7886 ((("%") shift) ((")") reduce 483) (("*") shift) (("+") shift) ((",") reduce 483) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 483) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 7887 ((("%") reduce 469) ((")") reduce 469) (("*") reduce 469) (("+") reduce 469) ((",") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_DOUBLE_ARROW) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_PIPE) reduce 469) ((T_POW) shift) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 7888 ((("%") reduce 486) ((")") reduce 486) (("*") reduce 486) (("+") reduce 486) ((",") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_DOUBLE_ARROW) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_PIPE) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("^") reduce 486) (("|") reduce 486)) 7890 ((("%") reduce 526) ((")") reduce 526) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_PIPE) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("^") reduce 526) (("|") reduce 526)) 7892 ((("%") reduce 527) ((")") reduce 527) (("*") reduce 527) (("+") reduce 527) ((",") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_DOUBLE_ARROW) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_PIPE) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("^") reduce 527) (("|") reduce 527)) 7893 ((("%") reduce 525) ((")") reduce 525) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_PIPE) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("^") reduce 525) (("|") reduce 525)) 7894 ((("%") shift) ((")") reduce 619) (("*") shift) (("+") shift) ((",") reduce 619) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 619) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 7895 ((("%") shift) ((")") reduce 618) (("*") shift) (("+") shift) ((",") reduce 618) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 618) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 7896 ((("%") reduce 505) (("(") reduce 505) ((")") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 7897 ((("%") shift) ((")") reduce 478) (("*") shift) (("+") shift) ((",") reduce 478) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 478) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 7898 ((("%") reduce 471) ((")") reduce 471) (("*") reduce 471) (("+") reduce 471) ((",") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_DOUBLE_ARROW) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_PIPE) reduce 471) ((T_POW) shift) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("|") reduce 471)) 7900 ((("%") reduce 94) (("(") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 7901 ((("%") reduce 93) (("(") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 7902 ((("%") reduce 95) (("(") reduce 95) ((")") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 7903 ((("%") reduce 509) (("(") reduce 509) ((")") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 7904 ((("%") reduce 92) (("(") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 7905 ((("%") reduce 400) (("(") reduce 400) ((")") reduce 400) (("*") reduce 400) (("+") reduce 400) ((",") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_DOUBLE_ARROW) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 400) ((T_OBJECT_OPERATOR) reduce 400) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 400) ((T_PIPE) reduce 400) ((T_POW) reduce 400) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("[") reduce 400) (("^") reduce 400) (("|") reduce 400)) 7908 ((("%") reduce 511) (("(") reduce 511) ((")") reduce 511) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("|") reduce 511)) 7909 ((("%") reduce 402) (("(") shift) ((")") reduce 402) (("*") reduce 402) (("+") reduce 402) ((",") reduce 402) (("-") reduce 402) ((".") reduce 402) (("/") reduce 402) (("<") reduce 402) ((">") reduce 402) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_DOUBLE_ARROW) reduce 402) ((T_INSTANCEOF) reduce 402) ((T_IS_EQUAL) reduce 402) ((T_IS_GREATER_OR_EQUAL) reduce 402) ((T_IS_IDENTICAL) reduce 402) ((T_IS_NOT_EQUAL) reduce 402) ((T_IS_NOT_IDENTICAL) reduce 402) ((T_IS_SMALLER_OR_EQUAL) reduce 402) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_PIPE) reduce 402) ((T_POW) reduce 402) ((T_SL) reduce 402) ((T_SPACESHIP) reduce 402) ((T_SR) reduce 402) (("^") reduce 402) (("|") reduce 402)) 7910 ((("%") reduce 510) (("(") reduce 510) ((")") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 7911 ((("%") reduce 512) (("(") reduce 512) ((")") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("|") reduce 512)) 7912 ((("%") reduce 573) (("(") reduce 573) ((")") reduce 573) (("*") reduce 573) (("+") reduce 573) ((",") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_DOUBLE_ARROW) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("|") reduce 573)) 7918 ((("%") reduce 574) (("(") reduce 574) ((")") reduce 574) (("*") reduce 574) (("+") reduce 574) ((",") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_DOUBLE_ARROW) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("|") reduce 574)) 7919 ((("%") reduce 578) (("(") reduce 578) ((")") reduce 578) (("*") reduce 578) (("+") reduce 578) ((",") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_DOUBLE_ARROW) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("|") reduce 578)) 7920 ((("%") reduce 575) (("(") reduce 575) ((")") reduce 575) (("*") reduce 575) (("+") reduce 575) ((",") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_DOUBLE_ARROW) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("|") reduce 575)) 7921 ((("%") reduce 576) (("(") reduce 576) ((")") reduce 576) (("*") reduce 576) (("+") reduce 576) ((",") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_DOUBLE_ARROW) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("|") reduce 576)) 7922 ((("%") reduce 399) (("(") reduce 399) ((")") reduce 399) (("*") reduce 399) (("+") reduce 399) ((",") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_DOUBLE_ARROW) reduce 399) ((T_INSTANCEOF) reduce 399) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 399) ((T_OBJECT_OPERATOR) reduce 399) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 399) ((T_PIPE) reduce 399) ((T_POW) reduce 399) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("[") reduce 399) (("^") reduce 399) (("|") reduce 399)) 7924 ((("%") reduce 577) (("(") reduce 577) ((")") reduce 577) (("*") reduce 577) (("+") reduce 577) ((",") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_DOUBLE_ARROW) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("|") reduce 577)) 7925 ((("%") reduce 401) (("(") reduce 401) ((")") reduce 401) (("*") reduce 401) (("+") reduce 401) ((",") reduce 401) (("-") reduce 401) ((".") reduce 401) (("/") reduce 401) (("<") reduce 401) ((">") reduce 401) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_DOUBLE_ARROW) reduce 401) ((T_INSTANCEOF) reduce 401) ((T_IS_EQUAL) reduce 401) ((T_IS_GREATER_OR_EQUAL) reduce 401) ((T_IS_IDENTICAL) reduce 401) ((T_IS_NOT_EQUAL) reduce 401) ((T_IS_NOT_IDENTICAL) reduce 401) ((T_IS_SMALLER_OR_EQUAL) reduce 401) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 401) ((T_OBJECT_OPERATOR) reduce 401) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 401) ((T_PIPE) reduce 401) ((T_POW) reduce 401) ((T_SL) reduce 401) ((T_SPACESHIP) reduce 401) ((T_SR) reduce 401) (("[") reduce 401) (("^") reduce 401) (("|") reduce 401)) 7933 ((("%") reduce 398) (("(") reduce 398) ((")") reduce 398) (("*") reduce 398) (("+") reduce 398) ((",") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_DOUBLE_ARROW) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 398) ((T_OBJECT_OPERATOR) reduce 398) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 398) ((T_PIPE) reduce 398) ((T_POW) reduce 398) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("[") reduce 398) (("^") reduce 398) (("|") reduce 398)) 7935 ((("%") reduce 513) (("(") reduce 513) ((")") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("|") reduce 513)) 7941 ((("%") reduce 233) ((")") reduce 233) (("*") reduce 233) (("+") reduce 233) ((",") reduce 233) (("-") reduce 233) ((".") reduce 233) (("/") reduce 233) (("<") reduce 233) ((">") reduce 233) (("?") reduce 233) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_BOOLEAN_AND) reduce 233) ((T_BOOLEAN_OR) reduce 233) ((T_COALESCE) reduce 233) ((T_DOUBLE_ARROW) reduce 233) ((T_INSTANCEOF) reduce 233) ((T_IS_EQUAL) reduce 233) ((T_IS_GREATER_OR_EQUAL) reduce 233) ((T_IS_IDENTICAL) reduce 233) ((T_IS_NOT_EQUAL) reduce 233) ((T_IS_NOT_IDENTICAL) reduce 233) ((T_IS_SMALLER_OR_EQUAL) reduce 233) ((T_LOGICAL_AND) reduce 233) ((T_LOGICAL_OR) reduce 233) ((T_LOGICAL_XOR) reduce 233) ((T_PIPE) reduce 233) ((T_POW) reduce 233) ((T_SL) reduce 233) ((T_SPACESHIP) reduce 233) ((T_SR) reduce 233) (("^") reduce 233) (("|") reduce 233)) 7944 (((")") reduce 596) ((",") reduce 596) (("=") shift)) 7948 ((("%") reduce 613) ((")") reduce 613) (("*") reduce 613) (("+") reduce 613) ((",") reduce 613) (("-") reduce 613) ((".") reduce 613) (("/") reduce 613) (("<") reduce 613) ((">") reduce 613) (("?") reduce 613) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_BOOLEAN_AND) reduce 613) ((T_BOOLEAN_OR) reduce 613) ((T_COALESCE) reduce 613) ((T_DOUBLE_ARROW) reduce 613) ((T_INSTANCEOF) reduce 613) ((T_IS_EQUAL) reduce 613) ((T_IS_GREATER_OR_EQUAL) reduce 613) ((T_IS_IDENTICAL) reduce 613) ((T_IS_NOT_EQUAL) reduce 613) ((T_IS_NOT_IDENTICAL) reduce 613) ((T_IS_SMALLER_OR_EQUAL) reduce 613) ((T_LOGICAL_AND) reduce 613) ((T_LOGICAL_OR) reduce 613) ((T_LOGICAL_XOR) reduce 613) ((T_PIPE) reduce 613) ((T_POW) reduce 613) ((T_SL) reduce 613) ((T_SPACESHIP) reduce 613) ((T_SR) reduce 613) (("^") reduce 613) (("|") reduce 613)) 7949 ((("%") reduce 467) ((")") reduce 467) (("*") reduce 467) (("+") reduce 467) ((",") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_DOUBLE_ARROW) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_PIPE) reduce 467) ((T_POW) shift) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("|") reduce 467)) 7950 ((("%") shift) ((")") reduce 616) (("*") shift) (("+") shift) ((",") reduce 616) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 616) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 7951 ((("%") shift) ((")") reduce 615) (("*") shift) (("+") shift) ((",") reduce 615) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 615) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 7952 ((("%") reduce 424) ((")") reduce 424) (("*") reduce 424) (("+") reduce 424) ((",") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_DOUBLE_ARROW) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("[") reduce 548) (("^") reduce 424) (("|") reduce 424)) 7954 ((("%") reduce 518) ((")") reduce 518) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_PIPE) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("^") reduce 518) (("|") reduce 518)) 7955 ((("%") reduce 474) ((")") reduce 474) (("*") reduce 474) (("+") reduce 474) ((",") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_DOUBLE_ARROW) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_PIPE) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("|") reduce 474)) 7956 ((("%") reduce 291) ((")") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("|") reduce 291)) 7960 ((("%") reduce 292) ((")") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("^") reduce 292) (("|") reduce 292)) 7961 ((("%") reduce 293) ((")") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("^") reduce 293) (("|") reduce 293)) 7964 ((("%") reduce 617) ((")") reduce 617) (("*") reduce 617) (("+") reduce 617) ((",") reduce 617) (("-") reduce 617) ((".") reduce 617) (("/") reduce 617) (("<") reduce 617) ((">") reduce 617) (("?") reduce 617) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_BOOLEAN_AND) reduce 617) ((T_BOOLEAN_OR) reduce 617) ((T_COALESCE) reduce 617) ((T_DOUBLE_ARROW) reduce 617) ((T_INSTANCEOF) reduce 617) ((T_IS_EQUAL) reduce 617) ((T_IS_GREATER_OR_EQUAL) reduce 617) ((T_IS_IDENTICAL) reduce 617) ((T_IS_NOT_EQUAL) reduce 617) ((T_IS_NOT_IDENTICAL) reduce 617) ((T_IS_SMALLER_OR_EQUAL) reduce 617) ((T_LOGICAL_AND) reduce 617) ((T_LOGICAL_OR) reduce 617) ((T_LOGICAL_XOR) reduce 617) ((T_PIPE) reduce 617) ((T_POW) reduce 617) ((T_SL) reduce 617) ((T_SPACESHIP) reduce 617) ((T_SR) reduce 617) (("^") reduce 617) (("|") reduce 617)) 7967 ((("%") reduce 614) ((")") reduce 614) (("*") reduce 614) (("+") reduce 614) ((",") reduce 614) (("-") reduce 614) ((".") reduce 614) (("/") reduce 614) (("<") reduce 614) ((">") reduce 614) (("?") reduce 614) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_BOOLEAN_AND) reduce 614) ((T_BOOLEAN_OR) reduce 614) ((T_COALESCE) reduce 614) ((T_DOUBLE_ARROW) reduce 614) ((T_INSTANCEOF) reduce 614) ((T_IS_EQUAL) reduce 614) ((T_IS_GREATER_OR_EQUAL) reduce 614) ((T_IS_IDENTICAL) reduce 614) ((T_IS_NOT_EQUAL) reduce 614) ((T_IS_NOT_IDENTICAL) reduce 614) ((T_IS_SMALLER_OR_EQUAL) reduce 614) ((T_LOGICAL_AND) reduce 614) ((T_LOGICAL_OR) reduce 614) ((T_LOGICAL_XOR) reduce 614) ((T_PIPE) reduce 614) ((T_POW) reduce 614) ((T_SL) reduce 614) ((T_SPACESHIP) reduce 614) ((T_SR) reduce 614) (("^") reduce 614) (("|") reduce 614)) 7968 ((("%") shift) ((")") reduce 594) (("*") shift) (("+") shift) ((",") reduce 594) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 7969 ((("%") reduce 468) ((")") reduce 468) (("*") reduce 468) (("+") reduce 468) ((",") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_DOUBLE_ARROW) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_PIPE) reduce 468) ((T_POW) shift) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 7970 ((("%") reduce 426) ((")") reduce 426) (("*") reduce 426) (("+") reduce 426) ((",") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_DOUBLE_ARROW) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("[") reduce 548) (("^") reduce 426) (("|") reduce 426)) 7972 ((("%") reduce 408) ((")") reduce 408) (("*") reduce 408) (("+") reduce 408) ((",") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_DOUBLE_ARROW) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_PIPE) reduce 408) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("^") reduce 408) (("|") reduce 408)) 7973 ((("%") reduce 409) ((")") reduce 409) (("*") reduce 409) (("+") reduce 409) ((",") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_DOUBLE_ARROW) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_PIPE) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("|") reduce 409)) 7974 ((("%") reduce 296) ((")") reduce 296) (("*") reduce 296) (("+") reduce 296) ((",") reduce 296) (("-") reduce 296) ((".") reduce 296) (("/") reduce 296) (("<") reduce 296) ((">") reduce 296) (("?") reduce 296) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_BOOLEAN_AND) reduce 296) ((T_BOOLEAN_OR) reduce 296) ((T_COALESCE) reduce 296) ((T_DOUBLE_ARROW) reduce 296) ((T_INSTANCEOF) reduce 296) ((T_IS_EQUAL) reduce 296) ((T_IS_GREATER_OR_EQUAL) reduce 296) ((T_IS_IDENTICAL) reduce 296) ((T_IS_NOT_EQUAL) reduce 296) ((T_IS_NOT_IDENTICAL) reduce 296) ((T_IS_SMALLER_OR_EQUAL) reduce 296) ((T_LOGICAL_AND) reduce 296) ((T_LOGICAL_OR) reduce 296) ((T_LOGICAL_XOR) reduce 296) ((T_PIPE) reduce 296) ((T_POW) reduce 296) ((T_SL) reduce 296) ((T_SPACESHIP) reduce 296) ((T_SR) reduce 296) (("^") reduce 296) (("|") reduce 296)) 7979 ((("%") reduce 297) ((")") reduce 297) (("*") reduce 297) (("+") reduce 297) ((",") reduce 297) (("-") reduce 297) ((".") reduce 297) (("/") reduce 297) (("<") reduce 297) ((">") reduce 297) (("?") reduce 297) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_BOOLEAN_AND) reduce 297) ((T_BOOLEAN_OR) reduce 297) ((T_COALESCE) reduce 297) ((T_DOUBLE_ARROW) reduce 297) ((T_INSTANCEOF) reduce 297) ((T_IS_EQUAL) reduce 297) ((T_IS_GREATER_OR_EQUAL) reduce 297) ((T_IS_IDENTICAL) reduce 297) ((T_IS_NOT_EQUAL) reduce 297) ((T_IS_NOT_IDENTICAL) reduce 297) ((T_IS_SMALLER_OR_EQUAL) reduce 297) ((T_LOGICAL_AND) reduce 297) ((T_LOGICAL_OR) reduce 297) ((T_LOGICAL_XOR) reduce 297) ((T_PIPE) reduce 297) ((T_POW) reduce 297) ((T_SL) reduce 297) ((T_SPACESHIP) reduce 297) ((T_SR) reduce 297) (("^") reduce 297) (("|") reduce 297)) 7980 ((("%") reduce 460) (("(") reduce 556) ((")") reduce 460) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PIPE) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("[") reduce 549) (("^") reduce 460) (("|") reduce 460)) 7982 ((("%") reduce 298) ((")") reduce 298) (("*") reduce 298) (("+") reduce 298) ((",") reduce 298) (("-") reduce 298) ((".") reduce 298) (("/") reduce 298) (("<") reduce 298) ((">") reduce 298) (("?") reduce 298) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_BOOLEAN_AND) reduce 298) ((T_BOOLEAN_OR) reduce 298) ((T_COALESCE) reduce 298) ((T_DOUBLE_ARROW) reduce 298) ((T_INSTANCEOF) reduce 298) ((T_IS_EQUAL) reduce 298) ((T_IS_GREATER_OR_EQUAL) reduce 298) ((T_IS_IDENTICAL) reduce 298) ((T_IS_NOT_EQUAL) reduce 298) ((T_IS_NOT_IDENTICAL) reduce 298) ((T_IS_SMALLER_OR_EQUAL) reduce 298) ((T_LOGICAL_AND) reduce 298) ((T_LOGICAL_OR) reduce 298) ((T_LOGICAL_XOR) reduce 298) ((T_PIPE) reduce 298) ((T_POW) reduce 298) ((T_SL) reduce 298) ((T_SPACESHIP) reduce 298) ((T_SR) reduce 298) (("^") reduce 298) (("|") reduce 298)) 7983 ((("%") reduce 299) ((")") reduce 299) (("*") reduce 299) (("+") reduce 299) ((",") reduce 299) (("-") reduce 299) ((".") reduce 299) (("/") reduce 299) (("<") reduce 299) ((">") reduce 299) (("?") reduce 299) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_BOOLEAN_AND) reduce 299) ((T_BOOLEAN_OR) reduce 299) ((T_COALESCE) reduce 299) ((T_DOUBLE_ARROW) reduce 299) ((T_INSTANCEOF) reduce 299) ((T_IS_EQUAL) reduce 299) ((T_IS_GREATER_OR_EQUAL) reduce 299) ((T_IS_IDENTICAL) reduce 299) ((T_IS_NOT_EQUAL) reduce 299) ((T_IS_NOT_IDENTICAL) reduce 299) ((T_IS_SMALLER_OR_EQUAL) reduce 299) ((T_LOGICAL_AND) reduce 299) ((T_LOGICAL_OR) reduce 299) ((T_LOGICAL_XOR) reduce 299) ((T_PIPE) reduce 299) ((T_POW) reduce 299) ((T_SL) reduce 299) ((T_SPACESHIP) reduce 299) ((T_SR) reduce 299) (("^") reduce 299) (("|") reduce 299)) 7984 ((("%") reduce 472) ((")") reduce 472) (("*") reduce 472) (("+") reduce 472) ((",") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_DOUBLE_ARROW) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_PIPE) reduce 472) ((T_POW) shift) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("|") reduce 472)) 7985 ((("%") reduce 470) ((")") reduce 470) (("*") reduce 470) (("+") reduce 470) ((",") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_DOUBLE_ARROW) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_PIPE) reduce 470) ((T_POW) shift) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("|") reduce 470)) 7988 ((("%") reduce 519) (("(") reduce 519) ((")") reduce 519) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PIPE) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("|") reduce 519)) 7989 ((("%") reduce 475) ((")") reduce 475) (("*") reduce 475) (("+") reduce 475) ((",") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_DOUBLE_ARROW) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_PIPE) reduce 475) ((T_POW) shift) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("|") reduce 475)) 7990 ((("%") reduce 446) ((")") reduce 446) (("*") reduce 446) (("+") reduce 446) ((",") reduce 446) (("-") reduce 446) ((".") reduce 446) (("/") reduce 446) (("<") reduce 446) ((">") reduce 446) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_BOOLEAN_AND) reduce 446) ((T_BOOLEAN_OR) reduce 446) ((T_COALESCE) reduce 446) ((T_DOUBLE_ARROW) reduce 446) ((T_INSTANCEOF) reduce 446) ((T_IS_EQUAL) reduce 446) ((T_IS_GREATER_OR_EQUAL) reduce 446) ((T_IS_IDENTICAL) reduce 446) ((T_IS_NOT_EQUAL) reduce 446) ((T_IS_NOT_IDENTICAL) reduce 446) ((T_IS_SMALLER_OR_EQUAL) reduce 446) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_PIPE) reduce 446) ((T_POW) shift) ((T_SL) reduce 446) ((T_SPACESHIP) reduce 446) ((T_SR) reduce 446) (("^") reduce 446) (("|") reduce 446)) 7991 ((("%") reduce 445) ((")") reduce 445) (("*") reduce 445) (("+") reduce 445) ((",") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_DOUBLE_ARROW) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_PIPE) reduce 445) ((T_POW) shift) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("|") reduce 445)) 7993 ((("%") reduce 570) (("(") reduce 570) ((")") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) (("=") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AND_EQUAL) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_COALESCE_EQUAL) reduce 570) ((T_CONCAT_EQUAL) reduce 570) ((T_DEC) reduce 570) ((T_DIV_EQUAL) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INC) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_MINUS_EQUAL) reduce 570) ((T_MOD_EQUAL) reduce 570) ((T_MUL_EQUAL) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_OR_EQUAL) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_PLUS_EQUAL) reduce 570) ((T_POW) reduce 570) ((T_POW_EQUAL) reduce 570) ((T_SL) reduce 570) ((T_SL_EQUAL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) ((T_SR_EQUAL) reduce 570) ((T_XOR_EQUAL) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 7996 ((("%") reduce 569) (("(") reduce 569) ((")") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) (("=") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AND_EQUAL) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_COALESCE_EQUAL) reduce 569) ((T_CONCAT_EQUAL) reduce 569) ((T_DEC) reduce 569) ((T_DIV_EQUAL) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INC) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_MINUS_EQUAL) reduce 569) ((T_MOD_EQUAL) reduce 569) ((T_MUL_EQUAL) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_OR_EQUAL) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_PLUS_EQUAL) reduce 569) ((T_POW) reduce 569) ((T_POW_EQUAL) reduce 569) ((T_SL) reduce 569) ((T_SL_EQUAL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) ((T_SR_EQUAL) reduce 569) ((T_XOR_EQUAL) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 7998 ((("%") reduce 522) (("(") reduce 522) ((")") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_PIPE) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("|") reduce 522)) 7999 ((("%") reduce 447) ((")") reduce 447) (("*") reduce 447) (("+") reduce 447) ((",") reduce 447) (("-") reduce 447) ((".") reduce 447) (("/") reduce 447) (("<") reduce 447) ((">") reduce 447) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_BOOLEAN_AND) reduce 447) ((T_BOOLEAN_OR) reduce 447) ((T_COALESCE) reduce 447) ((T_DOUBLE_ARROW) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 447) ((T_IS_GREATER_OR_EQUAL) reduce 447) ((T_IS_IDENTICAL) reduce 447) ((T_IS_NOT_EQUAL) reduce 447) ((T_IS_NOT_IDENTICAL) reduce 447) ((T_IS_SMALLER_OR_EQUAL) reduce 447) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_PIPE) reduce 447) ((T_POW) shift) ((T_SL) reduce 447) ((T_SPACESHIP) reduce 447) ((T_SR) reduce 447) (("^") reduce 447) (("|") reduce 447)) 8001 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) ((",") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_DOUBLE_ARROW) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_PIPE) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("]") reduce 74) (("^") reduce 74) (("|") reduce 74)) 8002 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) ((",") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_DOUBLE_ARROW) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_PIPE) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("]") reduce 52) (("^") reduce 52) (("|") reduce 52)) 8003 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) ((",") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_DOUBLE_ARROW) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_PIPE) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("]") reduce 27) (("^") reduce 27) (("|") reduce 27)) 8004 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) ((",") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_DOUBLE_ARROW) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_PIPE) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("]") reduce 51) (("^") reduce 51) (("|") reduce 51)) 8005 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) ((",") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_DOUBLE_ARROW) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_PIPE) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("]") reduce 53) (("^") reduce 53) (("|") reduce 53)) 8006 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) ((",") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_DOUBLE_ARROW) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_PIPE) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("]") reduce 49) (("^") reduce 49) (("|") reduce 49)) 8007 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) ((",") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_DOUBLE_ARROW) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_PIPE) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("]") reduce 29) (("^") reduce 29) (("|") reduce 29)) 8008 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) ((",") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_DOUBLE_ARROW) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_PIPE) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("]") reduce 59) (("^") reduce 59) (("|") reduce 59)) 8009 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) ((",") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_DOUBLE_ARROW) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_PIPE) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("]") reduce 60) (("^") reduce 60) (("|") reduce 60)) 8010 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) ((",") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_DOUBLE_ARROW) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_PIPE) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("]") reduce 11) (("^") reduce 11) (("|") reduce 11)) 8011 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) ((",") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_DOUBLE_ARROW) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_PIPE) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("]") reduce 42) (("^") reduce 42) (("|") reduce 42)) 8012 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) ((",") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_DOUBLE_ARROW) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_PIPE) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("]") reduce 39) (("^") reduce 39) (("|") reduce 39)) 8013 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) ((",") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_DOUBLE_ARROW) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_PIPE) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("]") reduce 25) (("^") reduce 25) (("|") reduce 25)) 8014 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) ((",") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_DOUBLE_ARROW) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_PIPE) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("]") reduce 50) (("^") reduce 50) (("|") reduce 50)) 8015 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) ((",") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_DOUBLE_ARROW) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_PIPE) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("]") reduce 66) (("^") reduce 66) (("|") reduce 66)) 8016 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) ((",") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_DOUBLE_ARROW) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_PIPE) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("]") reduce 18) (("^") reduce 18) (("|") reduce 18)) 8017 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) ((",") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_DOUBLE_ARROW) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_PIPE) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("]") reduce 17) (("^") reduce 17) (("|") reduce 17)) 8018 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) ((",") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_DOUBLE_ARROW) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_PIPE) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("]") reduce 15) (("^") reduce 15) (("|") reduce 15)) 8019 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) ((",") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_DOUBLE_ARROW) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_PIPE) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("]") reduce 14) (("^") reduce 14) (("|") reduce 14)) 8020 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) ((",") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_DOUBLE_ARROW) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_PIPE) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("]") reduce 38) (("^") reduce 38) (("|") reduce 38)) 8021 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) ((",") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_DOUBLE_ARROW) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_PIPE) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("]") reduce 26) (("^") reduce 26) (("|") reduce 26)) 8022 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) ((",") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_DOUBLE_ARROW) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_PIPE) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("]") reduce 22) (("^") reduce 22) (("|") reduce 22)) 8023 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) ((",") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_DOUBLE_ARROW) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_PIPE) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("]") reduce 24) (("^") reduce 24) (("|") reduce 24)) 8024 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) ((",") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_DOUBLE_ARROW) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_PIPE) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("]") reduce 16) (("^") reduce 16) (("|") reduce 16)) 8025 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) ((",") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_DOUBLE_ARROW) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_PIPE) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("]") reduce 48) (("^") reduce 48) (("|") reduce 48)) 8026 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) ((",") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_DOUBLE_ARROW) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_PIPE) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("]") reduce 20) (("^") reduce 20) (("|") reduce 20)) 8027 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) ((",") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_DOUBLE_ARROW) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_PIPE) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("]") reduce 70) (("^") reduce 70) (("|") reduce 70)) 8028 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) ((",") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_DOUBLE_ARROW) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_PIPE) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("]") reduce 3) (("^") reduce 3) (("|") reduce 3)) 8029 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) ((",") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_DOUBLE_ARROW) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_PIPE) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("]") reduce 12) (("^") reduce 12) (("|") reduce 12)) 8030 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) ((",") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_DOUBLE_ARROW) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_PIPE) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("]") reduce 54) (("^") reduce 54) (("|") reduce 54)) 8031 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) ((",") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_DOUBLE_ARROW) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_PIPE) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("]") reduce 65) (("^") reduce 65) (("|") reduce 65)) 8032 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) ((",") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_DOUBLE_ARROW) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_PIPE) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("]") reduce 75) (("^") reduce 75) (("|") reduce 75)) 8033 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) ((",") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_DOUBLE_ARROW) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_PIPE) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("]") reduce 30) (("^") reduce 30) (("|") reduce 30)) 8034 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) ((",") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_DOUBLE_ARROW) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_PIPE) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("]") reduce 68) (("^") reduce 68) (("|") reduce 68)) 8035 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) ((",") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_DOUBLE_ARROW) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_PIPE) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("]") reduce 21) (("^") reduce 21) (("|") reduce 21)) 8036 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) ((",") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_DOUBLE_ARROW) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_PIPE) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("]") reduce 23) (("^") reduce 23) (("|") reduce 23)) 8037 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) ((",") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_DOUBLE_ARROW) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_PIPE) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("]") reduce 41) (("^") reduce 41) (("|") reduce 41)) 8038 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) ((",") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_DOUBLE_ARROW) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_PIPE) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("]") reduce 62) (("^") reduce 62) (("|") reduce 62)) 8039 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) ((",") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_DOUBLE_ARROW) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_PIPE) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("]") reduce 34) (("^") reduce 34) (("|") reduce 34)) 8040 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) ((",") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_DOUBLE_ARROW) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_PIPE) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("]") reduce 40) (("^") reduce 40) (("|") reduce 40)) 8041 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) ((",") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_DOUBLE_ARROW) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_PIPE) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("]") reduce 13) (("^") reduce 13) (("|") reduce 13)) 8042 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) ((",") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_DOUBLE_ARROW) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_PIPE) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("]") reduce 55) (("^") reduce 55) (("|") reduce 55)) 8043 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) ((",") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_DOUBLE_ARROW) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_PIPE) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("]") reduce 1) (("^") reduce 1) (("|") reduce 1)) 8044 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) ((",") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_DOUBLE_ARROW) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_PIPE) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("]") reduce 2) (("^") reduce 2) (("|") reduce 2)) 8045 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) ((",") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_DOUBLE_ARROW) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_PIPE) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("]") reduce 9) (("^") reduce 9) (("|") reduce 9)) 8046 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) ((",") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_DOUBLE_ARROW) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_PIPE) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("]") reduce 33) (("^") reduce 33) (("|") reduce 33)) 8047 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) ((",") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_DOUBLE_ARROW) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_PIPE) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("]") reduce 58) (("^") reduce 58) (("|") reduce 58)) 8048 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) ((",") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_DOUBLE_ARROW) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_PIPE) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("]") reduce 37) (("^") reduce 37) (("|") reduce 37)) 8049 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) ((",") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_DOUBLE_ARROW) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_PIPE) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("]") reduce 64) (("^") reduce 64) (("|") reduce 64)) 8050 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) ((",") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_DOUBLE_ARROW) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_PIPE) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("]") reduce 46) (("^") reduce 46) (("|") reduce 46)) 8051 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) ((",") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_DOUBLE_ARROW) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_PIPE) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("]") reduce 8) (("^") reduce 8) (("|") reduce 8)) 8052 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) ((",") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_DOUBLE_ARROW) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_PIPE) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("]") reduce 6) (("^") reduce 6) (("|") reduce 6)) 8053 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) ((",") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_DOUBLE_ARROW) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_PIPE) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("]") reduce 7) (("^") reduce 7) (("|") reduce 7)) 8054 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) ((",") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_DOUBLE_ARROW) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_PIPE) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("]") reduce 69) (("^") reduce 69) (("|") reduce 69)) 8055 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) ((",") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_DOUBLE_ARROW) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_PIPE) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("]") reduce 63) (("^") reduce 63) (("|") reduce 63)) 8056 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) ((",") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_DOUBLE_ARROW) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_PIPE) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("]") reduce 56) (("^") reduce 56) (("|") reduce 56)) 8057 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) ((",") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_DOUBLE_ARROW) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_PIPE) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("]") reduce 10) (("^") reduce 10) (("|") reduce 10)) 8058 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) ((",") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_DOUBLE_ARROW) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_PIPE) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("]") reduce 67) (("^") reduce 67) (("|") reduce 67)) 8059 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) ((",") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_DOUBLE_ARROW) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_PIPE) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("]") reduce 44) (("^") reduce 44) (("|") reduce 44)) 8060 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) ((",") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_DOUBLE_ARROW) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_PIPE) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("]") reduce 76) (("^") reduce 76) (("|") reduce 76)) 8061 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) ((",") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_DOUBLE_ARROW) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_PIPE) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("]") reduce 71) (("^") reduce 71) (("|") reduce 71)) 8062 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) ((",") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_DOUBLE_ARROW) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_PIPE) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("]") reduce 77) (("^") reduce 77) (("|") reduce 77)) 8063 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) ((",") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_DOUBLE_ARROW) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_PIPE) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("]") reduce 78) (("^") reduce 78) (("|") reduce 78)) 8064 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) ((",") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_DOUBLE_ARROW) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_PIPE) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("]") reduce 79) (("^") reduce 79) (("|") reduce 79)) 8065 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) ((",") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_DOUBLE_ARROW) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_PIPE) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("]") reduce 4) (("^") reduce 4) (("|") reduce 4)) 8066 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) ((",") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_DOUBLE_ARROW) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_PIPE) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("]") reduce 5) (("^") reduce 5) (("|") reduce 5)) 8067 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) ((",") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_DOUBLE_ARROW) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_PIPE) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("]") reduce 43) (("^") reduce 43) (("|") reduce 43)) 8068 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) ((",") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_DOUBLE_ARROW) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_PIPE) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("]") reduce 73) (("^") reduce 73) (("|") reduce 73)) 8069 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) ((",") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_DOUBLE_ARROW) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_PIPE) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("]") reduce 82) (("^") reduce 82) (("|") reduce 82)) 8070 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) ((",") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_DOUBLE_ARROW) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_PIPE) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("]") reduce 47) (("^") reduce 47) (("|") reduce 47)) 8071 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) ((",") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_DOUBLE_ARROW) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_PIPE) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("]") reduce 31) (("^") reduce 31) (("|") reduce 31)) 8072 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) ((",") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_DOUBLE_ARROW) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_PIPE) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("]") reduce 57) (("^") reduce 57) (("|") reduce 57)) 8073 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) ((",") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_DOUBLE_ARROW) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_PIPE) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("]") reduce 61) (("^") reduce 61) (("|") reduce 61)) 8074 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) ((",") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_DOUBLE_ARROW) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_PIPE) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("]") reduce 28) (("^") reduce 28) (("|") reduce 28)) 8075 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) ((",") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_DOUBLE_ARROW) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_PIPE) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("]") reduce 36) (("^") reduce 36) (("|") reduce 36)) 8076 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) ((",") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_DOUBLE_ARROW) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_PIPE) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("]") reduce 32) (("^") reduce 32) (("|") reduce 32)) 8077 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) ((",") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_DOUBLE_ARROW) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_PIPE) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("]") reduce 35) (("^") reduce 35) (("|") reduce 35)) 8078 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) ((",") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_DOUBLE_ARROW) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_PIPE) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("]") reduce 19) (("^") reduce 19) (("|") reduce 19)) 8079 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) ((",") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_DOUBLE_ARROW) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_PIPE) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("]") reduce 45) (("^") reduce 45) (("|") reduce 45)) 8080 ((("%") reduce 542) (("(") reduce 579) (("*") reduce 542) (("+") reduce 542) ((",") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PIPE) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("]") reduce 542) (("^") reduce 542) (("|") reduce 542)) 8082 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) ((",") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_DOUBLE_ARROW) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_PIPE) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("]") reduce 72) (("^") reduce 72) (("|") reduce 72)) 8083 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) ((",") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_DOUBLE_ARROW) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_PIPE) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("]") reduce 83) (("^") reduce 83) (("|") reduce 83)) 8084 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) ((",") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) (("<") reduce 572) (("=") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AND_EQUAL) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_COALESCE_EQUAL) reduce 572) ((T_CONCAT_EQUAL) reduce 572) ((T_DEC) reduce 572) ((T_DIV_EQUAL) reduce 572) ((T_DOUBLE_ARROW) reduce 572) ((T_INC) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_MINUS_EQUAL) reduce 572) ((T_MOD_EQUAL) reduce 572) ((T_MUL_EQUAL) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_OR_EQUAL) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_PLUS_EQUAL) reduce 572) ((T_POW) reduce 572) ((T_POW_EQUAL) reduce 572) ((T_SL) reduce 572) ((T_SL_EQUAL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) ((T_SR_EQUAL) reduce 572) ((T_XOR_EQUAL) reduce 572) (("[") reduce 572) (("]") reduce 572) (("^") reduce 572) (("|") reduce 572)) 8087 ((("%") reduce 544) (("(") reduce 580) (("*") reduce 544) (("+") reduce 544) ((",") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PIPE) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("]") reduce 544) (("^") reduce 544) (("|") reduce 544)) 8089 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("]") reduce 507) (("^") reduce 507) (("|") reduce 507)) 8090 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("]") reduce 291) (("^") reduce 291) (("|") reduce 291)) 8094 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) (("=") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AND_EQUAL) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_COALESCE_EQUAL) reduce 292) ((T_CONCAT_EQUAL) reduce 292) ((T_DEC) reduce 292) ((T_DIV_EQUAL) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INC) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_MINUS_EQUAL) reduce 292) ((T_MOD_EQUAL) reduce 292) ((T_MUL_EQUAL) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_OR_EQUAL) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_PLUS_EQUAL) reduce 292) ((T_POW) reduce 292) ((T_POW_EQUAL) reduce 292) ((T_SL) reduce 292) ((T_SL_EQUAL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) ((T_SR_EQUAL) reduce 292) ((T_XOR_EQUAL) reduce 292) (("[") reduce 292) (("]") reduce 292) (("^") reduce 292) (("|") reduce 292)) 8095 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) (("=") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AND_EQUAL) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_COALESCE_EQUAL) reduce 293) ((T_CONCAT_EQUAL) reduce 293) ((T_DEC) reduce 293) ((T_DIV_EQUAL) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INC) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_MINUS_EQUAL) reduce 293) ((T_MOD_EQUAL) reduce 293) ((T_MUL_EQUAL) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_OR_EQUAL) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_PLUS_EQUAL) reduce 293) ((T_POW) reduce 293) ((T_POW_EQUAL) reduce 293) ((T_SL) reduce 293) ((T_SL_EQUAL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) ((T_SR_EQUAL) reduce 293) ((T_XOR_EQUAL) reduce 293) (("[") reduce 293) (("]") reduce 293) (("^") reduce 293) (("|") reduce 293)) 8100 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) ((",") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_DOUBLE_ARROW) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_PIPE) reduce 425) ((T_POW) reduce 425) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("]") reduce 425) (("^") reduce 425) (("|") reduce 425)) 8102 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) ((",") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_DOUBLE_ARROW) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_PIPE) reduce 423) ((T_POW) reduce 423) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("]") reduce 423) (("^") reduce 423) (("|") reduce 423)) 8112 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 419) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 419) (("^") shift) (("|") shift)) 8113 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 421) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 421) (("^") shift) (("|") shift)) 8114 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 420) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 420) (("^") shift) (("|") shift)) 8115 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 413) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 413) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 413) (("^") shift) (("|") shift)) 8116 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 410) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 410) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 410) (("^") shift) (("|") shift)) 8117 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 418) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 418) (("^") shift) (("|") shift)) 8118 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 412) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 412) (("^") shift) (("|") shift)) 8119 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 416) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 416) (("^") shift) (("|") shift)) 8120 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 411) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 411) (("^") shift) (("|") shift)) 8121 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 414) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 414) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 414) (("^") shift) (("|") shift)) 8122 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 415) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 415) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 415) (("^") shift) (("|") shift)) 8123 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 422) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 422) (("^") shift) (("|") shift)) 8124 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 417) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 417) (("^") shift) (("|") shift)) 8126 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 406) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 406) (("^") shift) (("|") shift)) 8129 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("]") reduce 568) (("^") reduce 568) (("|") reduce 568)) 8132 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) ((",") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) (("<") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_DOUBLE_ARROW) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_POW) reduce 564) ((T_SL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) (("[") reduce 564) (("]") reduce 564) (("^") reduce 564) (("|") reduce 564)) 8134 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) ((",") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) (("<") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_DOUBLE_ARROW) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_POW) reduce 563) ((T_SL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) (("[") reduce 563) (("]") reduce 563) (("^") reduce 563) (("|") reduce 563)) 8136 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("]") reduce 559) (("^") reduce 559) (("|") reduce 559)) 8137 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("]") reduce 565) (("^") reduce 565) (("|") reduce 565)) 8138 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) ((",") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_DOUBLE_ARROW) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("[") reduce 548) (("]") reduce 407) (("^") reduce 407) (("|") reduce 407)) 8142 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) ((",") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) (("<") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_DOUBLE_ARROW) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_POW) reduce 572) ((T_SL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) (("[") reduce 572) (("]") reduce 572) (("^") reduce 572) (("|") reduce 572)) 8144 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("]") reduce 507) (("^") reduce 507) (("|") reduce 507)) 8145 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("]") reduce 291) (("^") reduce 291) (("|") reduce 291)) 8149 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("[") reduce 292) (("]") reduce 292) (("^") reduce 292) (("|") reduce 292)) 8150 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("[") reduce 293) (("]") reduce 293) (("^") reduce 293) (("|") reduce 293)) 8151 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("]") reduce 504) (("^") reduce 504) (("|") reduce 504)) 8154 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) ((",") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) (("<") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_DOUBLE_ARROW) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_POW) reduce 571) ((T_SL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) (("[") reduce 571) (("]") reduce 571) (("^") reduce 571) (("|") reduce 571)) 8155 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("]") reduce 506) (("^") reduce 506) (("|") reduce 506)) 8156 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("]") reduce 508) (("^") reduce 508) (("|") reduce 508)) 8161 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("]") reduce 560) (("^") reduce 560) (("|") reduce 560)) 8162 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("]") reduce 582) (("^") reduce 582) (("|") reduce 582)) 8163 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("]") reduce 566) (("^") reduce 566) (("|") reduce 566)) 8164 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("]") reduce 584) (("^") reduce 584) (("|") reduce 584)) 8167 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("]") reduce 583) (("^") reduce 583) (("|") reduce 583)) 8168 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("]") reduce 561) (("^") reduce 561) (("|") reduce 561)) 8169 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("]") reduce 567) (("^") reduce 567) (("|") reduce 567)) 8170 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) ((",") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("[") reduce 562) (("]") reduce 562) (("^") reduce 562) (("|") reduce 562)) 8171 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("]") reduce 505) (("^") reduce 505) (("|") reduce 505)) 8172 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("]") reduce 570) (("^") reduce 570) (("|") reduce 570)) 8175 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("]") reduce 569) (("^") reduce 569) (("|") reduce 569)) 8177 (((",") reduce 588) (("]") reduce 588)) 8178 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) (("=") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AND_EQUAL) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_COALESCE_EQUAL) reduce 504) ((T_CONCAT_EQUAL) reduce 504) ((T_DEC) reduce 504) ((T_DIV_EQUAL) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INC) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_MINUS_EQUAL) reduce 504) ((T_MOD_EQUAL) reduce 504) ((T_MUL_EQUAL) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_OR_EQUAL) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_PLUS_EQUAL) reduce 504) ((T_POW) reduce 504) ((T_POW_EQUAL) reduce 504) ((T_SL) reduce 504) ((T_SL_EQUAL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) ((T_SR_EQUAL) reduce 504) ((T_XOR_EQUAL) reduce 504) (("[") reduce 504) (("]") reduce 504) (("^") reduce 504) (("|") reduce 504)) 8189 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) ((",") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PIPE) reduce 494) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("]") reduce 494) (("^") reduce 494) (("|") reduce 494)) 8190 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_PIPE) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("]") reduce 489) (("^") reduce 489) (("|") reduce 489)) 8263 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 494) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 494) (("^") shift) (("|") shift)) 8579 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_PIPE) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("]") reduce 490) (("^") reduce 490) (("|") reduce 490)) 8582 ((("%") reduce 541) (("(") reduce 579) (("*") reduce 541) (("+") reduce 541) ((",") reduce 541) (("-") reduce 541) ((".") reduce 541) (("/") reduce 541) (("<") reduce 541) ((">") reduce 541) (("?") reduce 541) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_BOOLEAN_AND) reduce 541) ((T_BOOLEAN_OR) reduce 541) ((T_COALESCE) reduce 541) ((T_DOUBLE_ARROW) reduce 541) ((T_INSTANCEOF) reduce 541) ((T_IS_EQUAL) reduce 541) ((T_IS_GREATER_OR_EQUAL) reduce 541) ((T_IS_IDENTICAL) reduce 541) ((T_IS_NOT_EQUAL) reduce 541) ((T_IS_NOT_IDENTICAL) reduce 541) ((T_IS_SMALLER_OR_EQUAL) reduce 541) ((T_LOGICAL_AND) reduce 541) ((T_LOGICAL_OR) reduce 541) ((T_LOGICAL_XOR) reduce 541) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 541) ((T_OBJECT_OPERATOR) reduce 541) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 541) ((T_PIPE) reduce 541) ((T_POW) reduce 541) ((T_SL) reduce 541) ((T_SPACESHIP) reduce 541) ((T_SR) reduce 541) (("[") reduce 541) (("]") reduce 541) (("^") reduce 541) (("|") reduce 541)) 8584 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) ((",") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) (("<") reduce 571) (("=") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AND_EQUAL) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_COALESCE_EQUAL) reduce 571) ((T_CONCAT_EQUAL) reduce 571) ((T_DEC) reduce 571) ((T_DIV_EQUAL) reduce 571) ((T_DOUBLE_ARROW) reduce 571) ((T_INC) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_MINUS_EQUAL) reduce 571) ((T_MOD_EQUAL) reduce 571) ((T_MUL_EQUAL) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_OR_EQUAL) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_PLUS_EQUAL) reduce 571) ((T_POW) reduce 571) ((T_POW_EQUAL) reduce 571) ((T_SL) reduce 571) ((T_SL_EQUAL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) ((T_SR_EQUAL) reduce 571) ((T_XOR_EQUAL) reduce 571) (("[") reduce 571) (("]") reduce 571) (("^") reduce 571) (("|") reduce 571)) 8587 ((("%") reduce 543) (("(") reduce 580) (("*") reduce 543) (("+") reduce 543) ((",") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PIPE) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("]") reduce 543) (("^") reduce 543) (("|") reduce 543)) 8588 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("]") reduce 506) (("^") reduce 506) (("|") reduce 506)) 8589 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("]") reduce 508) (("^") reduce 508) (("|") reduce 508)) 8591 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) ((",") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_DOUBLE_ARROW) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_PIPE) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("]") reduce 485) (("^") reduce 485) (("|") reduce 485)) 8592 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_PIPE) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("]") reduce 487) (("^") reduce 487) (("|") reduce 487)) 8597 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) (("=") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AND_EQUAL) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_COALESCE_EQUAL) reduce 560) ((T_CONCAT_EQUAL) reduce 560) ((T_DEC) reduce 560) ((T_DIV_EQUAL) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INC) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_MINUS_EQUAL) reduce 560) ((T_MOD_EQUAL) reduce 560) ((T_MUL_EQUAL) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_OR_EQUAL) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_PLUS_EQUAL) reduce 560) ((T_POW) reduce 560) ((T_POW_EQUAL) reduce 560) ((T_SL) reduce 560) ((T_SL_EQUAL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) ((T_SR_EQUAL) reduce 560) ((T_XOR_EQUAL) reduce 560) (("[") reduce 560) (("]") reduce 560) (("^") reduce 560) (("|") reduce 560)) 8598 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) (("=") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AND_EQUAL) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_COALESCE_EQUAL) reduce 582) ((T_CONCAT_EQUAL) reduce 582) ((T_DEC) reduce 582) ((T_DIV_EQUAL) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INC) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_MINUS_EQUAL) reduce 582) ((T_MOD_EQUAL) reduce 582) ((T_MUL_EQUAL) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_OR_EQUAL) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_PLUS_EQUAL) reduce 582) ((T_POW) reduce 582) ((T_POW_EQUAL) reduce 582) ((T_SL) reduce 582) ((T_SL_EQUAL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) ((T_SR_EQUAL) reduce 582) ((T_XOR_EQUAL) reduce 582) (("[") reduce 582) (("]") reduce 582) (("^") reduce 582) (("|") reduce 582)) 8599 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("]") reduce 566) (("^") reduce 566) (("|") reduce 566)) 8600 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) (("=") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AND_EQUAL) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_COALESCE_EQUAL) reduce 584) ((T_CONCAT_EQUAL) reduce 584) ((T_DEC) reduce 584) ((T_DIV_EQUAL) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INC) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_MINUS_EQUAL) reduce 584) ((T_MOD_EQUAL) reduce 584) ((T_MUL_EQUAL) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_OR_EQUAL) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_PLUS_EQUAL) reduce 584) ((T_POW) reduce 584) ((T_POW_EQUAL) reduce 584) ((T_SL) reduce 584) ((T_SL_EQUAL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) ((T_SR_EQUAL) reduce 584) ((T_XOR_EQUAL) reduce 584) (("[") reduce 584) (("]") reduce 584) (("^") reduce 584) (("|") reduce 584)) 8603 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) (("=") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AND_EQUAL) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_COALESCE_EQUAL) reduce 583) ((T_CONCAT_EQUAL) reduce 583) ((T_DEC) reduce 583) ((T_DIV_EQUAL) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INC) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_MINUS_EQUAL) reduce 583) ((T_MOD_EQUAL) reduce 583) ((T_MUL_EQUAL) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_OR_EQUAL) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_PLUS_EQUAL) reduce 583) ((T_POW) reduce 583) ((T_POW_EQUAL) reduce 583) ((T_SL) reduce 583) ((T_SL_EQUAL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) ((T_SR_EQUAL) reduce 583) ((T_XOR_EQUAL) reduce 583) (("[") reduce 583) (("]") reduce 583) (("^") reduce 583) (("|") reduce 583)) 8604 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) (("=") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AND_EQUAL) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_COALESCE_EQUAL) reduce 561) ((T_CONCAT_EQUAL) reduce 561) ((T_DEC) reduce 561) ((T_DIV_EQUAL) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INC) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_MINUS_EQUAL) reduce 561) ((T_MOD_EQUAL) reduce 561) ((T_MUL_EQUAL) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_OR_EQUAL) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_PLUS_EQUAL) reduce 561) ((T_POW) reduce 561) ((T_POW_EQUAL) reduce 561) ((T_SL) reduce 561) ((T_SL_EQUAL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) ((T_SR_EQUAL) reduce 561) ((T_XOR_EQUAL) reduce 561) (("[") reduce 561) (("]") reduce 561) (("^") reduce 561) (("|") reduce 561)) 8605 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("]") reduce 567) (("^") reduce 567) (("|") reduce 567)) 8606 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) ((",") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) (("=") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AND_EQUAL) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_COALESCE_EQUAL) reduce 562) ((T_CONCAT_EQUAL) reduce 562) ((T_DEC) reduce 562) ((T_DIV_EQUAL) reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_INC) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_MINUS_EQUAL) reduce 562) ((T_MOD_EQUAL) reduce 562) ((T_MUL_EQUAL) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_OR_EQUAL) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_PLUS_EQUAL) reduce 562) ((T_POW) reduce 562) ((T_POW_EQUAL) reduce 562) ((T_SL) reduce 562) ((T_SL_EQUAL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) ((T_SR_EQUAL) reduce 562) ((T_XOR_EQUAL) reduce 562) (("[") reduce 562) (("]") reduce 562) (("^") reduce 562) (("|") reduce 562)) 8608 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) ((",") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_PIPE) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("]") reduce 477) (("^") reduce 477) (("|") reduce 477)) 8610 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) (("=") shift) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_PIPE) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("]") reduce 520) (("^") reduce 520) (("|") reduce 520)) 8612 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 405) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 405) (("^") shift) (("|") shift)) 8613 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 482) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 482) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 482) (("^") shift) (("|") shift)) 8614 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 480) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 480) ((T_LOGICAL_OR) reduce 480) ((T_LOGICAL_XOR) reduce 480) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 480) (("^") shift) (("|") shift)) 8616 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 481) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 481) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 481) (("^") shift) (("|") shift)) 8617 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) ((",") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_DOUBLE_ARROW) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_PIPE) reduce 473) ((T_POW) shift) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("]") reduce 473) (("^") reduce 473) (("|") reduce 473)) 8618 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 483) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 483) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 483) (("^") shift) (("|") shift)) 8619 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) ((",") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_DOUBLE_ARROW) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_PIPE) reduce 469) ((T_POW) shift) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("]") reduce 469) (("^") reduce 469) (("|") reduce 469)) 8620 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) ((",") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_DOUBLE_ARROW) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_PIPE) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("]") reduce 486) (("^") reduce 486) (("|") reduce 486)) 8622 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_PIPE) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("]") reduce 526) (("^") reduce 526) (("|") reduce 526)) 8624 ((("%") reduce 527) (("*") reduce 527) (("+") reduce 527) ((",") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_DOUBLE_ARROW) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_PIPE) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("]") reduce 527) (("^") reduce 527) (("|") reduce 527)) 8625 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_PIPE) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("]") reduce 525) (("^") reduce 525) (("|") reduce 525)) 8626 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 619) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 619) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 619) (("^") shift) (("|") shift)) 8627 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 618) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 618) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 618) (("^") shift) (("|") shift)) 8628 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("]") reduce 505) (("^") reduce 505) (("|") reduce 505)) 8629 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 478) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 478) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 478) (("^") shift) (("|") shift)) 8630 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) ((",") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_DOUBLE_ARROW) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_PIPE) reduce 471) ((T_POW) shift) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("]") reduce 471) (("^") reduce 471) (("|") reduce 471)) 8632 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("]") reduce 94) (("^") reduce 94) (("|") reduce 94)) 8633 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("]") reduce 93) (("^") reduce 93) (("|") reduce 93)) 8634 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("]") reduce 95) (("^") reduce 95) (("|") reduce 95)) 8635 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("]") reduce 509) (("^") reduce 509) (("|") reduce 509)) 8636 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("]") reduce 92) (("^") reduce 92) (("|") reduce 92)) 8637 ((("%") reduce 400) (("(") reduce 400) (("*") reduce 400) (("+") reduce 400) ((",") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_DOUBLE_ARROW) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 400) ((T_OBJECT_OPERATOR) reduce 400) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 400) ((T_PIPE) reduce 400) ((T_POW) reduce 400) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("[") reduce 400) (("]") reduce 400) (("^") reduce 400) (("|") reduce 400)) 8640 ((("%") reduce 511) (("(") reduce 511) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("]") reduce 511) (("^") reduce 511) (("|") reduce 511)) 8641 ((("%") reduce 402) (("(") shift) (("*") reduce 402) (("+") reduce 402) ((",") reduce 402) (("-") reduce 402) ((".") reduce 402) (("/") reduce 402) (("<") reduce 402) ((">") reduce 402) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_DOUBLE_ARROW) reduce 402) ((T_INSTANCEOF) reduce 402) ((T_IS_EQUAL) reduce 402) ((T_IS_GREATER_OR_EQUAL) reduce 402) ((T_IS_IDENTICAL) reduce 402) ((T_IS_NOT_EQUAL) reduce 402) ((T_IS_NOT_IDENTICAL) reduce 402) ((T_IS_SMALLER_OR_EQUAL) reduce 402) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_PIPE) reduce 402) ((T_POW) reduce 402) ((T_SL) reduce 402) ((T_SPACESHIP) reduce 402) ((T_SR) reduce 402) (("]") reduce 402) (("^") reduce 402) (("|") reduce 402)) 8642 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("]") reduce 510) (("^") reduce 510) (("|") reduce 510)) 8643 ((("%") reduce 512) (("(") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("]") reduce 512) (("^") reduce 512) (("|") reduce 512)) 8644 ((("%") reduce 573) (("(") reduce 573) (("*") reduce 573) (("+") reduce 573) ((",") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_DOUBLE_ARROW) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("]") reduce 573) (("^") reduce 573) (("|") reduce 573)) 8650 ((("%") reduce 574) (("(") reduce 574) (("*") reduce 574) (("+") reduce 574) ((",") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_DOUBLE_ARROW) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("]") reduce 574) (("^") reduce 574) (("|") reduce 574)) 8651 ((("%") reduce 578) (("(") reduce 578) (("*") reduce 578) (("+") reduce 578) ((",") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_DOUBLE_ARROW) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("]") reduce 578) (("^") reduce 578) (("|") reduce 578)) 8652 ((("%") reduce 575) (("(") reduce 575) (("*") reduce 575) (("+") reduce 575) ((",") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_DOUBLE_ARROW) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("]") reduce 575) (("^") reduce 575) (("|") reduce 575)) 8653 ((("%") reduce 576) (("(") reduce 576) (("*") reduce 576) (("+") reduce 576) ((",") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_DOUBLE_ARROW) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("]") reduce 576) (("^") reduce 576) (("|") reduce 576)) 8654 ((("%") reduce 399) (("(") reduce 399) (("*") reduce 399) (("+") reduce 399) ((",") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_DOUBLE_ARROW) reduce 399) ((T_INSTANCEOF) reduce 399) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 399) ((T_OBJECT_OPERATOR) reduce 399) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 399) ((T_PIPE) reduce 399) ((T_POW) reduce 399) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("[") reduce 399) (("]") reduce 399) (("^") reduce 399) (("|") reduce 399)) 8656 ((("%") reduce 577) (("(") reduce 577) (("*") reduce 577) (("+") reduce 577) ((",") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_DOUBLE_ARROW) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("]") reduce 577) (("^") reduce 577) (("|") reduce 577)) 8657 ((("%") reduce 401) (("(") reduce 401) (("*") reduce 401) (("+") reduce 401) ((",") reduce 401) (("-") reduce 401) ((".") reduce 401) (("/") reduce 401) (("<") reduce 401) ((">") reduce 401) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_DOUBLE_ARROW) reduce 401) ((T_INSTANCEOF) reduce 401) ((T_IS_EQUAL) reduce 401) ((T_IS_GREATER_OR_EQUAL) reduce 401) ((T_IS_IDENTICAL) reduce 401) ((T_IS_NOT_EQUAL) reduce 401) ((T_IS_NOT_IDENTICAL) reduce 401) ((T_IS_SMALLER_OR_EQUAL) reduce 401) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 401) ((T_OBJECT_OPERATOR) reduce 401) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 401) ((T_PIPE) reduce 401) ((T_POW) reduce 401) ((T_SL) reduce 401) ((T_SPACESHIP) reduce 401) ((T_SR) reduce 401) (("[") reduce 401) (("]") reduce 401) (("^") reduce 401) (("|") reduce 401)) 8665 ((("%") reduce 398) (("(") reduce 398) (("*") reduce 398) (("+") reduce 398) ((",") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_DOUBLE_ARROW) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 398) ((T_OBJECT_OPERATOR) reduce 398) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 398) ((T_PIPE) reduce 398) ((T_POW) reduce 398) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("[") reduce 398) (("]") reduce 398) (("^") reduce 398) (("|") reduce 398)) 8667 ((("%") reduce 513) (("(") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("]") reduce 513) (("^") reduce 513) (("|") reduce 513)) 8673 ((("%") reduce 233) (("*") reduce 233) (("+") reduce 233) ((",") reduce 233) (("-") reduce 233) ((".") reduce 233) (("/") reduce 233) (("<") reduce 233) ((">") reduce 233) (("?") reduce 233) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_BOOLEAN_AND) reduce 233) ((T_BOOLEAN_OR) reduce 233) ((T_COALESCE) reduce 233) ((T_DOUBLE_ARROW) reduce 233) ((T_INSTANCEOF) reduce 233) ((T_IS_EQUAL) reduce 233) ((T_IS_GREATER_OR_EQUAL) reduce 233) ((T_IS_IDENTICAL) reduce 233) ((T_IS_NOT_EQUAL) reduce 233) ((T_IS_NOT_IDENTICAL) reduce 233) ((T_IS_SMALLER_OR_EQUAL) reduce 233) ((T_LOGICAL_AND) reduce 233) ((T_LOGICAL_OR) reduce 233) ((T_LOGICAL_XOR) reduce 233) ((T_PIPE) reduce 233) ((T_POW) reduce 233) ((T_SL) reduce 233) ((T_SPACESHIP) reduce 233) ((T_SR) reduce 233) (("]") reduce 233) (("^") reduce 233) (("|") reduce 233)) 8682 ((("%") reduce 613) (("*") reduce 613) (("+") reduce 613) ((",") reduce 613) (("-") reduce 613) ((".") reduce 613) (("/") reduce 613) (("<") reduce 613) ((">") reduce 613) (("?") reduce 613) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_BOOLEAN_AND) reduce 613) ((T_BOOLEAN_OR) reduce 613) ((T_COALESCE) reduce 613) ((T_DOUBLE_ARROW) reduce 613) ((T_INSTANCEOF) reduce 613) ((T_IS_EQUAL) reduce 613) ((T_IS_GREATER_OR_EQUAL) reduce 613) ((T_IS_IDENTICAL) reduce 613) ((T_IS_NOT_EQUAL) reduce 613) ((T_IS_NOT_IDENTICAL) reduce 613) ((T_IS_SMALLER_OR_EQUAL) reduce 613) ((T_LOGICAL_AND) reduce 613) ((T_LOGICAL_OR) reduce 613) ((T_LOGICAL_XOR) reduce 613) ((T_PIPE) reduce 613) ((T_POW) reduce 613) ((T_SL) reduce 613) ((T_SPACESHIP) reduce 613) ((T_SR) reduce 613) (("]") reduce 613) (("^") reduce 613) (("|") reduce 613)) 8683 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) ((",") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_DOUBLE_ARROW) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_PIPE) reduce 467) ((T_POW) shift) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("]") reduce 467) (("^") reduce 467) (("|") reduce 467)) 8684 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 616) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 616) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 616) (("^") shift) (("|") shift)) 8685 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 615) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 615) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 615) (("^") shift) (("|") shift)) 8686 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) ((",") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_DOUBLE_ARROW) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("[") reduce 548) (("]") reduce 424) (("^") reduce 424) (("|") reduce 424)) 8688 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_PIPE) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("]") reduce 518) (("^") reduce 518) (("|") reduce 518)) 8689 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) ((",") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_DOUBLE_ARROW) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_PIPE) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("]") reduce 474) (("^") reduce 474) (("|") reduce 474)) 8690 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("]") reduce 291) (("^") reduce 291) (("|") reduce 291)) 8694 ((("%") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("]") reduce 292) (("^") reduce 292) (("|") reduce 292)) 8695 ((("%") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("]") reduce 293) (("^") reduce 293) (("|") reduce 293)) 8698 ((("%") reduce 617) (("*") reduce 617) (("+") reduce 617) ((",") reduce 617) (("-") reduce 617) ((".") reduce 617) (("/") reduce 617) (("<") reduce 617) ((">") reduce 617) (("?") reduce 617) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_BOOLEAN_AND) reduce 617) ((T_BOOLEAN_OR) reduce 617) ((T_COALESCE) reduce 617) ((T_DOUBLE_ARROW) reduce 617) ((T_INSTANCEOF) reduce 617) ((T_IS_EQUAL) reduce 617) ((T_IS_GREATER_OR_EQUAL) reduce 617) ((T_IS_IDENTICAL) reduce 617) ((T_IS_NOT_EQUAL) reduce 617) ((T_IS_NOT_IDENTICAL) reduce 617) ((T_IS_SMALLER_OR_EQUAL) reduce 617) ((T_LOGICAL_AND) reduce 617) ((T_LOGICAL_OR) reduce 617) ((T_LOGICAL_XOR) reduce 617) ((T_PIPE) reduce 617) ((T_POW) reduce 617) ((T_SL) reduce 617) ((T_SPACESHIP) reduce 617) ((T_SR) reduce 617) (("]") reduce 617) (("^") reduce 617) (("|") reduce 617)) 8701 ((("%") reduce 614) (("*") reduce 614) (("+") reduce 614) ((",") reduce 614) (("-") reduce 614) ((".") reduce 614) (("/") reduce 614) (("<") reduce 614) ((">") reduce 614) (("?") reduce 614) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_BOOLEAN_AND) reduce 614) ((T_BOOLEAN_OR) reduce 614) ((T_COALESCE) reduce 614) ((T_DOUBLE_ARROW) reduce 614) ((T_INSTANCEOF) reduce 614) ((T_IS_EQUAL) reduce 614) ((T_IS_GREATER_OR_EQUAL) reduce 614) ((T_IS_IDENTICAL) reduce 614) ((T_IS_NOT_EQUAL) reduce 614) ((T_IS_NOT_IDENTICAL) reduce 614) ((T_IS_SMALLER_OR_EQUAL) reduce 614) ((T_LOGICAL_AND) reduce 614) ((T_LOGICAL_OR) reduce 614) ((T_LOGICAL_XOR) reduce 614) ((T_PIPE) reduce 614) ((T_POW) reduce 614) ((T_SL) reduce 614) ((T_SPACESHIP) reduce 614) ((T_SR) reduce 614) (("]") reduce 614) (("^") reduce 614) (("|") reduce 614)) 8702 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) ((",") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_DOUBLE_ARROW) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_PIPE) reduce 468) ((T_POW) shift) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("]") reduce 468) (("^") reduce 468) (("|") reduce 468)) 8703 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) ((",") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_DOUBLE_ARROW) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("[") reduce 548) (("]") reduce 426) (("^") reduce 426) (("|") reduce 426)) 8705 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) ((",") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_DOUBLE_ARROW) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_PIPE) reduce 408) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("]") reduce 408) (("^") reduce 408) (("|") reduce 408)) 8706 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) ((",") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_DOUBLE_ARROW) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_PIPE) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("]") reduce 409) (("^") reduce 409) (("|") reduce 409)) 8707 ((("%") reduce 296) (("*") reduce 296) (("+") reduce 296) ((",") reduce 296) (("-") reduce 296) ((".") reduce 296) (("/") reduce 296) (("<") reduce 296) ((">") reduce 296) (("?") reduce 296) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_BOOLEAN_AND) reduce 296) ((T_BOOLEAN_OR) reduce 296) ((T_COALESCE) reduce 296) ((T_DOUBLE_ARROW) reduce 296) ((T_INSTANCEOF) reduce 296) ((T_IS_EQUAL) reduce 296) ((T_IS_GREATER_OR_EQUAL) reduce 296) ((T_IS_IDENTICAL) reduce 296) ((T_IS_NOT_EQUAL) reduce 296) ((T_IS_NOT_IDENTICAL) reduce 296) ((T_IS_SMALLER_OR_EQUAL) reduce 296) ((T_LOGICAL_AND) reduce 296) ((T_LOGICAL_OR) reduce 296) ((T_LOGICAL_XOR) reduce 296) ((T_PIPE) reduce 296) ((T_POW) reduce 296) ((T_SL) reduce 296) ((T_SPACESHIP) reduce 296) ((T_SR) reduce 296) (("]") reduce 296) (("^") reduce 296) (("|") reduce 296)) 8712 ((("%") reduce 297) (("*") reduce 297) (("+") reduce 297) ((",") reduce 297) (("-") reduce 297) ((".") reduce 297) (("/") reduce 297) (("<") reduce 297) ((">") reduce 297) (("?") reduce 297) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_BOOLEAN_AND) reduce 297) ((T_BOOLEAN_OR) reduce 297) ((T_COALESCE) reduce 297) ((T_DOUBLE_ARROW) reduce 297) ((T_INSTANCEOF) reduce 297) ((T_IS_EQUAL) reduce 297) ((T_IS_GREATER_OR_EQUAL) reduce 297) ((T_IS_IDENTICAL) reduce 297) ((T_IS_NOT_EQUAL) reduce 297) ((T_IS_NOT_IDENTICAL) reduce 297) ((T_IS_SMALLER_OR_EQUAL) reduce 297) ((T_LOGICAL_AND) reduce 297) ((T_LOGICAL_OR) reduce 297) ((T_LOGICAL_XOR) reduce 297) ((T_PIPE) reduce 297) ((T_POW) reduce 297) ((T_SL) reduce 297) ((T_SPACESHIP) reduce 297) ((T_SR) reduce 297) (("]") reduce 297) (("^") reduce 297) (("|") reduce 297)) 8713 ((("%") reduce 460) (("(") reduce 556) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PIPE) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("[") reduce 549) (("]") reduce 460) (("^") reduce 460) (("|") reduce 460)) 8715 ((("%") reduce 298) (("*") reduce 298) (("+") reduce 298) ((",") reduce 298) (("-") reduce 298) ((".") reduce 298) (("/") reduce 298) (("<") reduce 298) ((">") reduce 298) (("?") reduce 298) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_BOOLEAN_AND) reduce 298) ((T_BOOLEAN_OR) reduce 298) ((T_COALESCE) reduce 298) ((T_DOUBLE_ARROW) reduce 298) ((T_INSTANCEOF) reduce 298) ((T_IS_EQUAL) reduce 298) ((T_IS_GREATER_OR_EQUAL) reduce 298) ((T_IS_IDENTICAL) reduce 298) ((T_IS_NOT_EQUAL) reduce 298) ((T_IS_NOT_IDENTICAL) reduce 298) ((T_IS_SMALLER_OR_EQUAL) reduce 298) ((T_LOGICAL_AND) reduce 298) ((T_LOGICAL_OR) reduce 298) ((T_LOGICAL_XOR) reduce 298) ((T_PIPE) reduce 298) ((T_POW) reduce 298) ((T_SL) reduce 298) ((T_SPACESHIP) reduce 298) ((T_SR) reduce 298) (("]") reduce 298) (("^") reduce 298) (("|") reduce 298)) 8716 ((("%") reduce 299) (("*") reduce 299) (("+") reduce 299) ((",") reduce 299) (("-") reduce 299) ((".") reduce 299) (("/") reduce 299) (("<") reduce 299) ((">") reduce 299) (("?") reduce 299) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_BOOLEAN_AND) reduce 299) ((T_BOOLEAN_OR) reduce 299) ((T_COALESCE) reduce 299) ((T_DOUBLE_ARROW) reduce 299) ((T_INSTANCEOF) reduce 299) ((T_IS_EQUAL) reduce 299) ((T_IS_GREATER_OR_EQUAL) reduce 299) ((T_IS_IDENTICAL) reduce 299) ((T_IS_NOT_EQUAL) reduce 299) ((T_IS_NOT_IDENTICAL) reduce 299) ((T_IS_SMALLER_OR_EQUAL) reduce 299) ((T_LOGICAL_AND) reduce 299) ((T_LOGICAL_OR) reduce 299) ((T_LOGICAL_XOR) reduce 299) ((T_PIPE) reduce 299) ((T_POW) reduce 299) ((T_SL) reduce 299) ((T_SPACESHIP) reduce 299) ((T_SR) reduce 299) (("]") reduce 299) (("^") reduce 299) (("|") reduce 299)) 8717 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) ((",") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_DOUBLE_ARROW) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_PIPE) reduce 472) ((T_POW) shift) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("]") reduce 472) (("^") reduce 472) (("|") reduce 472)) 8718 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) ((",") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_DOUBLE_ARROW) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_PIPE) reduce 470) ((T_POW) shift) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("]") reduce 470) (("^") reduce 470) (("|") reduce 470)) 8721 ((("%") reduce 519) (("(") reduce 519) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PIPE) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("]") reduce 519) (("^") reduce 519) (("|") reduce 519)) 8722 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) ((",") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_DOUBLE_ARROW) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_PIPE) reduce 475) ((T_POW) shift) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("]") reduce 475) (("^") reduce 475) (("|") reduce 475)) 8723 ((("%") reduce 446) (("*") reduce 446) (("+") reduce 446) ((",") reduce 446) (("-") reduce 446) ((".") reduce 446) (("/") reduce 446) (("<") reduce 446) ((">") reduce 446) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_BOOLEAN_AND) reduce 446) ((T_BOOLEAN_OR) reduce 446) ((T_COALESCE) reduce 446) ((T_DOUBLE_ARROW) reduce 446) ((T_INSTANCEOF) reduce 446) ((T_IS_EQUAL) reduce 446) ((T_IS_GREATER_OR_EQUAL) reduce 446) ((T_IS_IDENTICAL) reduce 446) ((T_IS_NOT_EQUAL) reduce 446) ((T_IS_NOT_IDENTICAL) reduce 446) ((T_IS_SMALLER_OR_EQUAL) reduce 446) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_PIPE) reduce 446) ((T_POW) shift) ((T_SL) reduce 446) ((T_SPACESHIP) reduce 446) ((T_SR) reduce 446) (("]") reduce 446) (("^") reduce 446) (("|") reduce 446)) 8724 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) ((",") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_DOUBLE_ARROW) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_PIPE) reduce 445) ((T_POW) shift) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("]") reduce 445) (("^") reduce 445) (("|") reduce 445)) 8726 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) (("=") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AND_EQUAL) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_COALESCE_EQUAL) reduce 570) ((T_CONCAT_EQUAL) reduce 570) ((T_DEC) reduce 570) ((T_DIV_EQUAL) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INC) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_MINUS_EQUAL) reduce 570) ((T_MOD_EQUAL) reduce 570) ((T_MUL_EQUAL) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_OR_EQUAL) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_PLUS_EQUAL) reduce 570) ((T_POW) reduce 570) ((T_POW_EQUAL) reduce 570) ((T_SL) reduce 570) ((T_SL_EQUAL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) ((T_SR_EQUAL) reduce 570) ((T_XOR_EQUAL) reduce 570) (("[") reduce 570) (("]") reduce 570) (("^") reduce 570) (("|") reduce 570)) 8729 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) (("=") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AND_EQUAL) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_COALESCE_EQUAL) reduce 569) ((T_CONCAT_EQUAL) reduce 569) ((T_DEC) reduce 569) ((T_DIV_EQUAL) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INC) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_MINUS_EQUAL) reduce 569) ((T_MOD_EQUAL) reduce 569) ((T_MUL_EQUAL) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_OR_EQUAL) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_PLUS_EQUAL) reduce 569) ((T_POW) reduce 569) ((T_POW_EQUAL) reduce 569) ((T_SL) reduce 569) ((T_SL_EQUAL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) ((T_SR_EQUAL) reduce 569) ((T_XOR_EQUAL) reduce 569) (("[") reduce 569) (("]") reduce 569) (("^") reduce 569) (("|") reduce 569)) 8731 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_PIPE) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("]") reduce 522) (("^") reduce 522) (("|") reduce 522)) 8732 ((("%") reduce 447) (("*") reduce 447) (("+") reduce 447) ((",") reduce 447) (("-") reduce 447) ((".") reduce 447) (("/") reduce 447) (("<") reduce 447) ((">") reduce 447) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_BOOLEAN_AND) reduce 447) ((T_BOOLEAN_OR) reduce 447) ((T_COALESCE) reduce 447) ((T_DOUBLE_ARROW) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 447) ((T_IS_GREATER_OR_EQUAL) reduce 447) ((T_IS_IDENTICAL) reduce 447) ((T_IS_NOT_EQUAL) reduce 447) ((T_IS_NOT_IDENTICAL) reduce 447) ((T_IS_SMALLER_OR_EQUAL) reduce 447) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_PIPE) reduce 447) ((T_POW) shift) ((T_SL) reduce 447) ((T_SPACESHIP) reduce 447) ((T_SR) reduce 447) (("]") reduce 447) (("^") reduce 447) (("|") reduce 447)) 8744 ((("%") reduce 540) (("*") reduce 540) (("+") reduce 540) ((",") reduce 540) (("-") reduce 540) ((".") reduce 540) (("/") reduce 540) (("<") reduce 540) ((">") reduce 540) (("?") reduce 540) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_BOOLEAN_AND) reduce 540) ((T_BOOLEAN_OR) reduce 540) ((T_COALESCE) reduce 540) ((T_INSTANCEOF) reduce 540) ((T_IS_EQUAL) reduce 540) ((T_IS_GREATER_OR_EQUAL) reduce 540) ((T_IS_IDENTICAL) reduce 540) ((T_IS_NOT_EQUAL) reduce 540) ((T_IS_NOT_IDENTICAL) reduce 540) ((T_IS_SMALLER_OR_EQUAL) reduce 540) ((T_LOGICAL_AND) reduce 540) ((T_LOGICAL_OR) reduce 540) ((T_LOGICAL_XOR) reduce 540) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 540) ((T_OBJECT_OPERATOR) reduce 540) ((T_PIPE) reduce 540) ((T_POW) reduce 540) ((T_SL) reduce 540) ((T_SPACESHIP) reduce 540) ((T_SR) reduce 540) (("[") reduce 540) (("]") reduce 540) (("^") reduce 540) (("|") reduce 540)) 8746 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_PIPE) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("]") reduce 521) (("^") reduce 521) (("|") reduce 521)) 8748 ((("%") reduce 534) (("*") reduce 534) (("+") reduce 534) ((",") reduce 534) (("-") reduce 534) ((".") reduce 534) (("/") reduce 534) (("<") reduce 534) ((">") reduce 534) (("?") reduce 534) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_BOOLEAN_AND) reduce 534) ((T_BOOLEAN_OR) reduce 534) ((T_COALESCE) reduce 534) ((T_INSTANCEOF) reduce 534) ((T_IS_EQUAL) reduce 534) ((T_IS_GREATER_OR_EQUAL) reduce 534) ((T_IS_IDENTICAL) reduce 534) ((T_IS_NOT_EQUAL) reduce 534) ((T_IS_NOT_IDENTICAL) reduce 534) ((T_IS_SMALLER_OR_EQUAL) reduce 534) ((T_LOGICAL_AND) reduce 534) ((T_LOGICAL_OR) reduce 534) ((T_LOGICAL_XOR) reduce 534) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PIPE) reduce 534) ((T_POW) reduce 534) ((T_SL) reduce 534) ((T_SPACESHIP) reduce 534) ((T_SR) reduce 534) (("[") reduce 534) (("]") reduce 534) (("^") reduce 534) (("|") reduce 534)) 8749 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_PIPE) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("]") reduce 524) (("^") reduce 524) (("|") reduce 524)) 8753 ((("%") reduce 517) (("(") shift) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_PIPE) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("]") reduce 517) (("^") reduce 517) (("|") reduce 517)) 8754 ((("%") reduce 533) (("*") reduce 533) (("+") reduce 533) ((",") reduce 533) (("-") reduce 533) ((".") reduce 533) (("/") reduce 533) (("<") reduce 533) ((">") reduce 533) (("?") reduce 533) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_BOOLEAN_AND) reduce 533) ((T_BOOLEAN_OR) reduce 533) ((T_COALESCE) reduce 533) ((T_INSTANCEOF) reduce 533) ((T_IS_EQUAL) reduce 533) ((T_IS_GREATER_OR_EQUAL) reduce 533) ((T_IS_IDENTICAL) reduce 533) ((T_IS_NOT_EQUAL) reduce 533) ((T_IS_NOT_IDENTICAL) reduce 533) ((T_IS_SMALLER_OR_EQUAL) reduce 533) ((T_LOGICAL_AND) reduce 533) ((T_LOGICAL_OR) reduce 533) ((T_LOGICAL_XOR) reduce 533) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PIPE) reduce 533) ((T_POW) reduce 533) ((T_SL) reduce 533) ((T_SPACESHIP) reduce 533) ((T_SR) reduce 533) (("[") reduce 533) (("]") reduce 533) (("^") reduce 533) (("|") reduce 533)) 8755 ((("%") reduce 537) (("*") reduce 537) (("+") reduce 537) ((",") reduce 537) (("-") reduce 537) ((".") reduce 537) (("/") reduce 537) (("<") reduce 537) ((">") reduce 537) (("?") reduce 537) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_BOOLEAN_AND) reduce 537) ((T_BOOLEAN_OR) reduce 537) ((T_COALESCE) reduce 537) ((T_INSTANCEOF) reduce 537) ((T_IS_EQUAL) reduce 537) ((T_IS_GREATER_OR_EQUAL) reduce 537) ((T_IS_IDENTICAL) reduce 537) ((T_IS_NOT_EQUAL) reduce 537) ((T_IS_NOT_IDENTICAL) reduce 537) ((T_IS_SMALLER_OR_EQUAL) reduce 537) ((T_LOGICAL_AND) reduce 537) ((T_LOGICAL_OR) reduce 537) ((T_LOGICAL_XOR) reduce 537) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_PIPE) reduce 537) ((T_POW) reduce 537) ((T_SL) reduce 537) ((T_SPACESHIP) reduce 537) ((T_SR) reduce 537) (("[") reduce 537) (("]") reduce 537) (("^") reduce 537) (("|") reduce 537)) 8761 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PIPE) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("]") reduce 532) (("^") reduce 532) (("|") reduce 532)) 8763 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_PIPE) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("]") reduce 523) (("^") reduce 523) (("|") reduce 523)) 8765 ((("%") reduce 536) (("*") reduce 536) (("+") reduce 536) ((",") reduce 536) (("-") reduce 536) ((".") reduce 536) (("/") reduce 536) (("<") reduce 536) ((">") reduce 536) (("?") reduce 536) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_BOOLEAN_AND) reduce 536) ((T_BOOLEAN_OR) reduce 536) ((T_COALESCE) reduce 536) ((T_INSTANCEOF) reduce 536) ((T_IS_EQUAL) reduce 536) ((T_IS_GREATER_OR_EQUAL) reduce 536) ((T_IS_IDENTICAL) reduce 536) ((T_IS_NOT_EQUAL) reduce 536) ((T_IS_NOT_IDENTICAL) reduce 536) ((T_IS_SMALLER_OR_EQUAL) reduce 536) ((T_LOGICAL_AND) reduce 536) ((T_LOGICAL_OR) reduce 536) ((T_LOGICAL_XOR) reduce 536) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 536) ((T_OBJECT_OPERATOR) reduce 536) ((T_PIPE) reduce 536) ((T_POW) reduce 536) ((T_SL) reduce 536) ((T_SPACESHIP) reduce 536) ((T_SR) reduce 536) (("[") reduce 536) (("]") reduce 536) (("^") reduce 536) (("|") reduce 536)) 8766 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("]") reduce 94) (("^") reduce 94) (("|") reduce 94)) 8767 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("]") reduce 93) (("^") reduce 93) (("|") reduce 93)) 8768 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("]") reduce 95) (("^") reduce 95) (("|") reduce 95)) 8770 ((("%") reduce 539) (("*") reduce 539) (("+") reduce 539) ((",") reduce 539) (("-") reduce 539) ((".") reduce 539) (("/") reduce 539) (("<") reduce 539) ((">") reduce 539) (("?") reduce 539) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_BOOLEAN_AND) reduce 539) ((T_BOOLEAN_OR) reduce 539) ((T_COALESCE) reduce 539) ((T_INSTANCEOF) reduce 539) ((T_IS_EQUAL) reduce 539) ((T_IS_GREATER_OR_EQUAL) reduce 539) ((T_IS_IDENTICAL) reduce 539) ((T_IS_NOT_EQUAL) reduce 539) ((T_IS_NOT_IDENTICAL) reduce 539) ((T_IS_SMALLER_OR_EQUAL) reduce 539) ((T_LOGICAL_AND) reduce 539) ((T_LOGICAL_OR) reduce 539) ((T_LOGICAL_XOR) reduce 539) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 539) ((T_OBJECT_OPERATOR) reduce 539) ((T_PIPE) reduce 539) ((T_POW) reduce 539) ((T_SL) reduce 539) ((T_SPACESHIP) reduce 539) ((T_SR) reduce 539) (("[") reduce 539) (("]") reduce 539) (("^") reduce 539) (("|") reduce 539)) 8773 ((("%") reduce 538) (("*") reduce 538) (("+") reduce 538) ((",") reduce 538) (("-") reduce 538) ((".") reduce 538) (("/") reduce 538) (("<") reduce 538) ((">") reduce 538) (("?") reduce 538) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_BOOLEAN_AND) reduce 538) ((T_BOOLEAN_OR) reduce 538) ((T_COALESCE) reduce 538) ((T_INSTANCEOF) reduce 538) ((T_IS_EQUAL) reduce 538) ((T_IS_GREATER_OR_EQUAL) reduce 538) ((T_IS_IDENTICAL) reduce 538) ((T_IS_NOT_EQUAL) reduce 538) ((T_IS_NOT_IDENTICAL) reduce 538) ((T_IS_SMALLER_OR_EQUAL) reduce 538) ((T_LOGICAL_AND) reduce 538) ((T_LOGICAL_OR) reduce 538) ((T_LOGICAL_XOR) reduce 538) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 538) ((T_OBJECT_OPERATOR) reduce 538) ((T_PIPE) reduce 538) ((T_POW) reduce 538) ((T_SL) reduce 538) ((T_SPACESHIP) reduce 538) ((T_SR) reduce 538) (("[") reduce 538) (("]") reduce 538) (("^") reduce 538) (("|") reduce 538)) 8779 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("]") reduce 92) (("^") reduce 92) (("|") reduce 92)) 8782 ((("%") reduce 535) (("*") reduce 535) (("+") reduce 535) ((",") reduce 535) (("-") reduce 535) ((".") reduce 535) (("/") reduce 535) (("<") reduce 535) ((">") reduce 535) (("?") reduce 535) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_BOOLEAN_AND) reduce 535) ((T_BOOLEAN_OR) reduce 535) ((T_COALESCE) reduce 535) ((T_INSTANCEOF) reduce 535) ((T_IS_EQUAL) reduce 535) ((T_IS_GREATER_OR_EQUAL) reduce 535) ((T_IS_IDENTICAL) reduce 535) ((T_IS_NOT_EQUAL) reduce 535) ((T_IS_NOT_IDENTICAL) reduce 535) ((T_IS_SMALLER_OR_EQUAL) reduce 535) ((T_LOGICAL_AND) reduce 535) ((T_LOGICAL_OR) reduce 535) ((T_LOGICAL_XOR) reduce 535) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PIPE) reduce 535) ((T_POW) reduce 535) ((T_SL) reduce 535) ((T_SPACESHIP) reduce 535) ((T_SR) reduce 535) (("[") reduce 535) (("]") reduce 535) (("^") reduce 535) (("|") reduce 535)) 8784 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) (("=") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AND_EQUAL) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_COALESCE_EQUAL) reduce 568) ((T_CONCAT_EQUAL) reduce 568) ((T_DEC) reduce 568) ((T_DIV_EQUAL) reduce 568) ((T_INC) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_MINUS_EQUAL) reduce 568) ((T_MOD_EQUAL) reduce 568) ((T_MUL_EQUAL) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_OR_EQUAL) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_PLUS_EQUAL) reduce 568) ((T_POW) reduce 568) ((T_POW_EQUAL) reduce 568) ((T_SL) reduce 568) ((T_SL_EQUAL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) ((T_SR_EQUAL) reduce 568) ((T_XOR_EQUAL) reduce 568) (("[") reduce 568) (("]") reduce 568) (("^") reduce 568) (("|") reduce 568)) 8785 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 479) (("(") shift) (("*") reduce 479) (("+") shift) ((",") reduce 479) (("-") shift) ((".") reduce 479) (("/") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 479) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 479) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PIPE) reduce 479) ((T_POW) reduce 479) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("]") reduce 479) (("^") reduce 479) (("`") shift) (("|") reduce 479) (("~") shift)) 8793 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) ((",") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) (("<") reduce 564) (("=") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AND_EQUAL) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_COALESCE_EQUAL) reduce 564) ((T_CONCAT_EQUAL) reduce 564) ((T_DEC) reduce 564) ((T_DIV_EQUAL) reduce 564) ((T_INC) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_MINUS_EQUAL) reduce 564) ((T_MOD_EQUAL) reduce 564) ((T_MUL_EQUAL) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_OR_EQUAL) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_PLUS_EQUAL) reduce 564) ((T_POW) reduce 564) ((T_POW_EQUAL) reduce 564) ((T_SL) reduce 564) ((T_SL_EQUAL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) ((T_SR_EQUAL) reduce 564) ((T_XOR_EQUAL) reduce 564) (("[") reduce 564) (("]") reduce 564) (("^") reduce 564) (("|") reduce 564)) 8794 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PIPE) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 551) (("]") reduce 530) (("^") reduce 530) (("|") reduce 530)) 8796 ((("%") reduce 529) (("*") reduce 529) (("+") reduce 529) ((",") reduce 529) (("-") reduce 529) ((".") reduce 529) (("/") reduce 529) (("<") reduce 529) ((">") reduce 529) (("?") reduce 529) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_BOOLEAN_AND) reduce 529) ((T_BOOLEAN_OR) reduce 529) ((T_COALESCE) reduce 529) ((T_INSTANCEOF) reduce 529) ((T_IS_EQUAL) reduce 529) ((T_IS_GREATER_OR_EQUAL) reduce 529) ((T_IS_IDENTICAL) reduce 529) ((T_IS_NOT_EQUAL) reduce 529) ((T_IS_NOT_IDENTICAL) reduce 529) ((T_IS_SMALLER_OR_EQUAL) reduce 529) ((T_LOGICAL_AND) reduce 529) ((T_LOGICAL_OR) reduce 529) ((T_LOGICAL_XOR) reduce 529) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PIPE) reduce 529) ((T_POW) reduce 529) ((T_SL) reduce 529) ((T_SPACESHIP) reduce 529) ((T_SR) reduce 529) (("[") reduce 554) (("]") reduce 529) (("^") reduce 529) (("|") reduce 529)) 8797 ((("%") reduce 528) (("(") reduce 557) (("*") reduce 528) (("+") reduce 528) ((",") reduce 528) (("-") reduce 528) ((".") reduce 528) (("/") reduce 528) (("<") reduce 528) ((">") reduce 528) (("?") reduce 528) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_BOOLEAN_AND) reduce 528) ((T_BOOLEAN_OR) reduce 528) ((T_COALESCE) reduce 528) ((T_INSTANCEOF) reduce 528) ((T_IS_EQUAL) reduce 528) ((T_IS_GREATER_OR_EQUAL) reduce 528) ((T_IS_IDENTICAL) reduce 528) ((T_IS_NOT_EQUAL) reduce 528) ((T_IS_NOT_IDENTICAL) reduce 528) ((T_IS_SMALLER_OR_EQUAL) reduce 528) ((T_LOGICAL_AND) reduce 528) ((T_LOGICAL_OR) reduce 528) ((T_LOGICAL_XOR) reduce 528) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PIPE) reduce 528) ((T_POW) reduce 528) ((T_SL) reduce 528) ((T_SPACESHIP) reduce 528) ((T_SR) reduce 528) (("[") reduce 550) (("]") reduce 528) (("^") reduce 528) (("|") reduce 528)) 8798 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 590) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 590) (("^") shift) (("|") shift)) 8801 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) ((",") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) (("<") reduce 563) (("=") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AND_EQUAL) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_COALESCE_EQUAL) reduce 563) ((T_CONCAT_EQUAL) reduce 563) ((T_DEC) reduce 563) ((T_DIV_EQUAL) reduce 563) ((T_INC) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_MINUS_EQUAL) reduce 563) ((T_MOD_EQUAL) reduce 563) ((T_MUL_EQUAL) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_OR_EQUAL) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_PLUS_EQUAL) reduce 563) ((T_POW) reduce 563) ((T_POW_EQUAL) reduce 563) ((T_SL) reduce 563) ((T_SL_EQUAL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) ((T_SR_EQUAL) reduce 563) ((T_XOR_EQUAL) reduce 563) (("[") reduce 563) (("]") reduce 563) (("^") reduce 563) (("|") reduce 563)) 8802 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) ((",") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_PIPE) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("]") reduce 484) (("^") reduce 484) (("|") reduce 484)) 8803 ((("%") reduce 466) (("*") reduce 466) (("+") reduce 466) ((",") reduce 466) (("-") reduce 466) ((".") reduce 466) (("/") reduce 466) (("<") reduce 466) ((">") reduce 466) (("?") reduce 466) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_BOOLEAN_AND) reduce 466) ((T_BOOLEAN_OR) reduce 466) ((T_COALESCE) reduce 466) ((T_INSTANCEOF) reduce 466) ((T_IS_EQUAL) reduce 466) ((T_IS_GREATER_OR_EQUAL) reduce 466) ((T_IS_IDENTICAL) reduce 466) ((T_IS_NOT_EQUAL) reduce 466) ((T_IS_NOT_IDENTICAL) reduce 466) ((T_IS_SMALLER_OR_EQUAL) reduce 466) ((T_LOGICAL_AND) reduce 466) ((T_LOGICAL_OR) reduce 466) ((T_LOGICAL_XOR) reduce 466) ((T_PIPE) reduce 466) ((T_POW) reduce 466) ((T_SL) reduce 466) ((T_SPACESHIP) reduce 466) ((T_SR) reduce 466) (("]") reduce 466) (("^") reduce 466) (("|") reduce 466)) 8804 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_PIPE) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("]") reduce 488) (("^") reduce 488) (("|") reduce 488)) 8805 ((("%") reduce 531) (("(") shift) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("]") reduce 531) (("^") reduce 531) (("|") reduce 531)) 8806 ((("%") reduce 461) (("(") reduce 558) (("*") reduce 461) (("+") reduce 461) ((",") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PIPE) reduce 461) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") reduce 552) (("]") reduce 461) (("^") reduce 461) (("|") reduce 461)) 8807 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) ((",") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_PIPE) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("]") reduce 462) (("^") reduce 462) (("|") reduce 462)) 8808 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) ((",") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_PIPE) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("]") reduce 476) (("^") reduce 476) (("|") reduce 476)) 8809 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) (("=") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AND_EQUAL) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_COALESCE_EQUAL) reduce 559) ((T_CONCAT_EQUAL) reduce 559) ((T_DEC) reduce 559) ((T_DIV_EQUAL) reduce 559) ((T_INC) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_MINUS_EQUAL) reduce 559) ((T_MOD_EQUAL) reduce 559) ((T_MUL_EQUAL) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_OR_EQUAL) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_PLUS_EQUAL) reduce 559) ((T_POW) reduce 559) ((T_POW_EQUAL) reduce 559) ((T_SL) reduce 559) ((T_SL_EQUAL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) ((T_SR_EQUAL) reduce 559) ((T_XOR_EQUAL) reduce 559) (("[") reduce 559) (("]") reduce 559) (("^") reduce 559) (("|") reduce 559)) 8810 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("]") reduce 565) (("^") reduce 565) (("|") reduce 565)) 8811 ((("%") reduce 403) (("*") reduce 403) (("+") reduce 403) ((",") reduce 403) (("-") reduce 403) ((".") reduce 403) (("/") reduce 403) (("<") reduce 403) (("=") shift) ((">") reduce 403) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_INC) shift) ((T_INSTANCEOF) reduce 403) ((T_IS_EQUAL) reduce 403) ((T_IS_GREATER_OR_EQUAL) reduce 403) ((T_IS_IDENTICAL) reduce 403) ((T_IS_NOT_EQUAL) reduce 403) ((T_IS_NOT_IDENTICAL) reduce 403) ((T_IS_SMALLER_OR_EQUAL) reduce 403) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 403) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 403) ((T_POW_EQUAL) shift) ((T_SL) reduce 403) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 403) ((T_SR) reduce 403) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 548) (("]") reduce 403) (("^") reduce 403) (("|") reduce 403)) 8815 ((("%") reduce 448) (("*") reduce 448) (("+") reduce 448) ((",") reduce 448) (("-") reduce 448) ((".") reduce 448) (("/") reduce 448) (("<") reduce 448) ((">") reduce 448) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_BOOLEAN_AND) reduce 448) ((T_BOOLEAN_OR) reduce 448) ((T_COALESCE) reduce 448) ((T_INSTANCEOF) reduce 448) ((T_IS_EQUAL) reduce 448) ((T_IS_GREATER_OR_EQUAL) reduce 448) ((T_IS_IDENTICAL) reduce 448) ((T_IS_NOT_EQUAL) reduce 448) ((T_IS_NOT_IDENTICAL) reduce 448) ((T_IS_SMALLER_OR_EQUAL) reduce 448) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_PIPE) reduce 448) ((T_POW) shift) ((T_SL) reduce 448) ((T_SPACESHIP) reduce 448) ((T_SR) reduce 448) (("]") reduce 448) (("^") reduce 448) (("|") reduce 448)) 8847 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 432) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 432) (("^") shift) (("|") reduce 432)) 8848 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 435) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 435) (("^") reduce 435) (("|") reduce 435)) 8849 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 444) (("-") shift) ((".") reduce 444) (("/") shift) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_PIPE) reduce 444) ((T_POW) shift) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("]") reduce 444) (("^") reduce 444) (("|") reduce 444)) 8850 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 458) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 458) (("^") reduce 458) (("|") reduce 458)) 8851 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 443) (("-") shift) ((".") reduce 443) (("/") shift) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_PIPE) reduce 443) ((T_POW) shift) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("]") reduce 443) (("^") reduce 443) (("|") reduce 443)) 8852 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) ((",") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_PIPE) reduce 440) ((T_POW) shift) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("]") reduce 440) (("^") reduce 440) (("|") reduce 440)) 8853 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 453) (("-") shift) ((".") shift) (("/") shift) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_PIPE) reduce 453) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 453) ((T_SR) shift) (("]") reduce 453) (("^") reduce 453) (("|") reduce 453)) 8854 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 431) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 431) (("^") shift) (("|") shift)) 8855 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 429) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 429) (("^") shift) (("|") shift)) 8856 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 430) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 430) (("^") shift) (("|") shift)) 8857 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 455) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 455) ((T_SR) shift) (("]") reduce 455) (("^") reduce 455) (("|") reduce 455)) 8858 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 450) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 450) (("^") reduce 450) (("|") reduce 450)) 8859 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 452) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 452) (("^") reduce 452) (("|") reduce 452)) 8860 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 449) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 449) (("^") reduce 449) (("|") reduce 449)) 8861 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 457) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 457) ((T_SR) shift) (("]") reduce 457) (("^") reduce 457) (("|") reduce 457)) 8862 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 451) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 451) (("^") reduce 451) (("|") reduce 451)) 8865 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("]") reduce 94) (("^") reduce 94) (("|") reduce 94)) 8866 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("]") reduce 93) (("^") reduce 93) (("|") reduce 93)) 8867 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("]") reduce 95) (("^") reduce 95) (("|") reduce 95)) 8868 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("]") reduce 509) (("^") reduce 509) (("|") reduce 509)) 8869 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("]") reduce 92) (("^") reduce 92) (("|") reduce 92)) 8870 ((("%") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("]") reduce 568) (("^") reduce 568) (("|") reduce 568)) 8871 ((("%") reduce 511) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("]") reduce 511) (("^") reduce 511) (("|") reduce 511)) 8872 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PIPE) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("]") reduce 459) (("^") reduce 459) (("|") reduce 459)) 8873 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("]") reduce 510) (("^") reduce 510) (("|") reduce 510)) 8874 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("]") reduce 512) (("^") reduce 512) (("|") reduce 512)) 8875 ((("%") reduce 573) (("*") reduce 573) (("+") reduce 573) ((",") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("]") reduce 573) (("^") reduce 573) (("|") reduce 573)) 8881 ((("%") reduce 574) (("*") reduce 574) (("+") reduce 574) ((",") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("]") reduce 574) (("^") reduce 574) (("|") reduce 574)) 8882 ((("%") reduce 578) (("*") reduce 578) (("+") reduce 578) ((",") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("]") reduce 578) (("^") reduce 578) (("|") reduce 578)) 8883 ((("%") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("]") reduce 582) (("^") reduce 582) (("|") reduce 582)) 8884 ((("%") reduce 575) (("*") reduce 575) (("+") reduce 575) ((",") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("]") reduce 575) (("^") reduce 575) (("|") reduce 575)) 8885 ((("%") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("]") reduce 584) (("^") reduce 584) (("|") reduce 584)) 8888 ((("%") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("]") reduce 583) (("^") reduce 583) (("|") reduce 583)) 8889 ((("%") reduce 576) (("*") reduce 576) (("+") reduce 576) ((",") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("]") reduce 576) (("^") reduce 576) (("|") reduce 576)) 8891 ((("%") reduce 577) (("*") reduce 577) (("+") reduce 577) ((",") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("]") reduce 577) (("^") reduce 577) (("|") reduce 577)) 8893 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("]") reduce 513) (("^") reduce 513) (("|") reduce 513)) 8894 ((("%") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("]") reduce 570) (("^") reduce 570) (("|") reduce 570)) 8897 ((("%") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("]") reduce 569) (("^") reduce 569) (("|") reduce 569)) 8898 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 465) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 465) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 465) (("^") shift) (("|") shift)) 8899 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 427) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 427) (("^") shift) (("|") shift)) 8900 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 428) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 428) (("^") shift) (("|") shift)) 8901 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 433) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 433) (("^") reduce 433) (("|") reduce 433)) 8902 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 434) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 434) (("^") reduce 434) (("|") reduce 434)) 8906 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 463) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 463) (("^") shift) (("|") shift)) 8907 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 464) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 464) (("^") shift) (("|") shift)) 8908 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 456) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 456) ((T_SR) shift) (("]") reduce 456) (("^") reduce 456) (("|") reduce 456)) 8909 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 454) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 454) ((T_SR) shift) (("]") reduce 454) (("^") reduce 454) (("|") reduce 454)) 8910 ((("%") reduce 441) (("*") reduce 441) (("+") reduce 441) ((",") reduce 441) (("-") reduce 441) ((".") reduce 441) (("/") reduce 441) (("<") reduce 441) ((">") reduce 441) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 441) ((T_IS_GREATER_OR_EQUAL) reduce 441) ((T_IS_IDENTICAL) reduce 441) ((T_IS_NOT_EQUAL) reduce 441) ((T_IS_NOT_IDENTICAL) reduce 441) ((T_IS_SMALLER_OR_EQUAL) reduce 441) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_PIPE) reduce 441) ((T_POW) shift) ((T_SL) reduce 441) ((T_SPACESHIP) reduce 441) ((T_SR) reduce 441) (("]") reduce 441) (("^") reduce 441) (("|") reduce 441)) 8911 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 436) (("-") shift) ((".") reduce 436) (("/") shift) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_PIPE) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 436) ((T_SR) shift) (("]") reduce 436) (("^") reduce 436) (("|") reduce 436)) 8912 ((("%") shift) (("*") shift) (("+") reduce 438) ((",") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") shift) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_PIPE) reduce 438) ((T_POW) shift) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("]") reduce 438) (("^") reduce 438) (("|") reduce 438)) 8913 ((("%") shift) (("*") shift) (("+") reduce 437) ((",") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") shift) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_PIPE) reduce 437) ((T_POW) shift) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("]") reduce 437) (("^") reduce 437) (("|") reduce 437)) 8914 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) ((",") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_PIPE) reduce 439) ((T_POW) shift) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("]") reduce 439) (("^") reduce 439) (("|") reduce 439)) 8915 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) ((",") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_PIPE) reduce 442) ((T_POW) shift) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("]") reduce 442) (("^") reduce 442) (("|") reduce 442)) 8920 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 404) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 404) (("^") shift) (("|") shift)) 8922 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) ((",") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_PIPE) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("]") reduce 74) (("^") reduce 74) (("|") reduce 74)) 8923 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) ((",") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_PIPE) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("]") reduce 52) (("^") reduce 52) (("|") reduce 52)) 8924 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) ((",") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_PIPE) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("]") reduce 27) (("^") reduce 27) (("|") reduce 27)) 8925 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) ((",") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_PIPE) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("]") reduce 51) (("^") reduce 51) (("|") reduce 51)) 8926 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) ((",") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_PIPE) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("]") reduce 53) (("^") reduce 53) (("|") reduce 53)) 8927 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) ((",") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_PIPE) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("]") reduce 49) (("^") reduce 49) (("|") reduce 49)) 8928 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) ((",") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_PIPE) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("]") reduce 29) (("^") reduce 29) (("|") reduce 29)) 8929 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) ((",") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_PIPE) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("]") reduce 59) (("^") reduce 59) (("|") reduce 59)) 8930 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) ((",") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_PIPE) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("]") reduce 60) (("^") reduce 60) (("|") reduce 60)) 8931 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) ((",") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_PIPE) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("]") reduce 11) (("^") reduce 11) (("|") reduce 11)) 8932 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) ((",") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_PIPE) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("]") reduce 42) (("^") reduce 42) (("|") reduce 42)) 8933 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) ((",") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_PIPE) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("]") reduce 39) (("^") reduce 39) (("|") reduce 39)) 8934 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) ((",") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_PIPE) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("]") reduce 25) (("^") reduce 25) (("|") reduce 25)) 8935 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) ((",") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_PIPE) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("]") reduce 50) (("^") reduce 50) (("|") reduce 50)) 8936 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) ((",") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_PIPE) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("]") reduce 66) (("^") reduce 66) (("|") reduce 66)) 8937 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) ((",") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_PIPE) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("]") reduce 18) (("^") reduce 18) (("|") reduce 18)) 8938 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) ((",") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_PIPE) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("]") reduce 17) (("^") reduce 17) (("|") reduce 17)) 8939 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) ((",") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_PIPE) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("]") reduce 15) (("^") reduce 15) (("|") reduce 15)) 8940 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) ((",") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_PIPE) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("]") reduce 14) (("^") reduce 14) (("|") reduce 14)) 8941 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) ((",") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_PIPE) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("]") reduce 38) (("^") reduce 38) (("|") reduce 38)) 8942 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) ((",") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_PIPE) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("]") reduce 26) (("^") reduce 26) (("|") reduce 26)) 8943 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) ((",") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_PIPE) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("]") reduce 22) (("^") reduce 22) (("|") reduce 22)) 8944 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) ((",") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_PIPE) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("]") reduce 24) (("^") reduce 24) (("|") reduce 24)) 8945 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) ((",") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_PIPE) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("]") reduce 16) (("^") reduce 16) (("|") reduce 16)) 8946 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) ((",") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_PIPE) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("]") reduce 48) (("^") reduce 48) (("|") reduce 48)) 8947 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) ((",") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_PIPE) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("]") reduce 20) (("^") reduce 20) (("|") reduce 20)) 8948 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) ((",") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_PIPE) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("]") reduce 70) (("^") reduce 70) (("|") reduce 70)) 8949 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) ((",") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_PIPE) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("]") reduce 3) (("^") reduce 3) (("|") reduce 3)) 8950 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) ((",") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_PIPE) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("]") reduce 12) (("^") reduce 12) (("|") reduce 12)) 8951 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) ((",") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_PIPE) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("]") reduce 54) (("^") reduce 54) (("|") reduce 54)) 8952 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) ((",") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_PIPE) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("]") reduce 65) (("^") reduce 65) (("|") reduce 65)) 8953 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) ((",") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_PIPE) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("]") reduce 75) (("^") reduce 75) (("|") reduce 75)) 8954 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) ((",") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_PIPE) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("]") reduce 30) (("^") reduce 30) (("|") reduce 30)) 8955 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) ((",") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_PIPE) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("]") reduce 68) (("^") reduce 68) (("|") reduce 68)) 8956 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) ((",") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_PIPE) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("]") reduce 21) (("^") reduce 21) (("|") reduce 21)) 8957 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) ((",") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_PIPE) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("]") reduce 23) (("^") reduce 23) (("|") reduce 23)) 8958 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) ((",") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_PIPE) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("]") reduce 41) (("^") reduce 41) (("|") reduce 41)) 8959 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) ((",") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_PIPE) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("]") reduce 62) (("^") reduce 62) (("|") reduce 62)) 8960 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) ((",") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_PIPE) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("]") reduce 34) (("^") reduce 34) (("|") reduce 34)) 8961 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) ((",") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_PIPE) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("]") reduce 40) (("^") reduce 40) (("|") reduce 40)) 8962 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) ((",") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_PIPE) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("]") reduce 13) (("^") reduce 13) (("|") reduce 13)) 8963 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) ((",") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_PIPE) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("]") reduce 55) (("^") reduce 55) (("|") reduce 55)) 8964 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) ((",") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_PIPE) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("]") reduce 1) (("^") reduce 1) (("|") reduce 1)) 8965 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) ((",") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_PIPE) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("]") reduce 2) (("^") reduce 2) (("|") reduce 2)) 8966 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) ((",") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_PIPE) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("]") reduce 9) (("^") reduce 9) (("|") reduce 9)) 8967 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) ((",") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_PIPE) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("]") reduce 33) (("^") reduce 33) (("|") reduce 33)) 8968 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) ((",") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_PIPE) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("]") reduce 58) (("^") reduce 58) (("|") reduce 58)) 8969 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) ((",") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_PIPE) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("]") reduce 37) (("^") reduce 37) (("|") reduce 37)) 8970 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) ((",") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_PIPE) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("]") reduce 64) (("^") reduce 64) (("|") reduce 64)) 8971 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) ((",") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_PIPE) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("]") reduce 46) (("^") reduce 46) (("|") reduce 46)) 8972 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) ((",") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_PIPE) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("]") reduce 8) (("^") reduce 8) (("|") reduce 8)) 8973 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) ((",") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_PIPE) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("]") reduce 6) (("^") reduce 6) (("|") reduce 6)) 8974 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) ((",") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_PIPE) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("]") reduce 7) (("^") reduce 7) (("|") reduce 7)) 8975 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) ((",") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_PIPE) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("]") reduce 69) (("^") reduce 69) (("|") reduce 69)) 8976 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) ((",") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_PIPE) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("]") reduce 63) (("^") reduce 63) (("|") reduce 63)) 8977 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) ((",") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_PIPE) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("]") reduce 56) (("^") reduce 56) (("|") reduce 56)) 8978 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) ((",") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_PIPE) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("]") reduce 10) (("^") reduce 10) (("|") reduce 10)) 8979 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) ((",") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_PIPE) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("]") reduce 67) (("^") reduce 67) (("|") reduce 67)) 8980 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) ((",") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_PIPE) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("]") reduce 44) (("^") reduce 44) (("|") reduce 44)) 8981 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) ((",") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_PIPE) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("]") reduce 76) (("^") reduce 76) (("|") reduce 76)) 8982 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) ((",") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_PIPE) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("]") reduce 71) (("^") reduce 71) (("|") reduce 71)) 8983 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) ((",") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_PIPE) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("]") reduce 77) (("^") reduce 77) (("|") reduce 77)) 8984 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) ((",") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_PIPE) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("]") reduce 78) (("^") reduce 78) (("|") reduce 78)) 8985 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) ((",") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_PIPE) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("]") reduce 79) (("^") reduce 79) (("|") reduce 79)) 8986 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) ((",") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_PIPE) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("]") reduce 4) (("^") reduce 4) (("|") reduce 4)) 8987 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) ((",") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_PIPE) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("]") reduce 5) (("^") reduce 5) (("|") reduce 5)) 8988 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) ((",") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_PIPE) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("]") reduce 43) (("^") reduce 43) (("|") reduce 43)) 8989 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) ((",") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_PIPE) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("]") reduce 73) (("^") reduce 73) (("|") reduce 73)) 8990 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) ((",") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_PIPE) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("]") reduce 82) (("^") reduce 82) (("|") reduce 82)) 8991 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) ((",") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_PIPE) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("]") reduce 47) (("^") reduce 47) (("|") reduce 47)) 8992 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) ((",") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_PIPE) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("]") reduce 31) (("^") reduce 31) (("|") reduce 31)) 8993 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) ((",") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_PIPE) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("]") reduce 57) (("^") reduce 57) (("|") reduce 57)) 8994 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) ((",") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_PIPE) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("]") reduce 61) (("^") reduce 61) (("|") reduce 61)) 8995 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) ((",") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_PIPE) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("]") reduce 28) (("^") reduce 28) (("|") reduce 28)) 8996 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) ((",") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_PIPE) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("]") reduce 36) (("^") reduce 36) (("|") reduce 36)) 8997 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) ((",") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_PIPE) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("]") reduce 32) (("^") reduce 32) (("|") reduce 32)) 8998 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) ((",") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_PIPE) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("]") reduce 35) (("^") reduce 35) (("|") reduce 35)) 8999 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) ((",") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_PIPE) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("]") reduce 19) (("^") reduce 19) (("|") reduce 19)) 9000 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) ((",") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_PIPE) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("]") reduce 45) (("^") reduce 45) (("|") reduce 45)) 9001 ((("%") reduce 542) (("(") reduce 579) (("*") reduce 542) (("+") reduce 542) ((",") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PIPE) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("]") reduce 542) (("^") reduce 542) (("|") reduce 542)) 9003 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) ((",") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_PIPE) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("]") reduce 72) (("^") reduce 72) (("|") reduce 72)) 9004 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) ((",") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_PIPE) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("]") reduce 83) (("^") reduce 83) (("|") reduce 83)) 9005 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) ((",") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) (("<") reduce 572) (("=") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AND_EQUAL) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_COALESCE_EQUAL) reduce 572) ((T_CONCAT_EQUAL) reduce 572) ((T_DEC) reduce 572) ((T_DIV_EQUAL) reduce 572) ((T_INC) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_MINUS_EQUAL) reduce 572) ((T_MOD_EQUAL) reduce 572) ((T_MUL_EQUAL) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_OR_EQUAL) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_PLUS_EQUAL) reduce 572) ((T_POW) reduce 572) ((T_POW_EQUAL) reduce 572) ((T_SL) reduce 572) ((T_SL_EQUAL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) ((T_SR_EQUAL) reduce 572) ((T_XOR_EQUAL) reduce 572) (("[") reduce 572) (("]") reduce 572) (("^") reduce 572) (("|") reduce 572)) 9008 ((("%") reduce 544) (("(") reduce 580) (("*") reduce 544) (("+") reduce 544) ((",") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PIPE) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("]") reduce 544) (("^") reduce 544) (("|") reduce 544)) 9010 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("]") reduce 507) (("^") reduce 507) (("|") reduce 507)) 9011 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("]") reduce 291) (("^") reduce 291) (("|") reduce 291)) 9015 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) (("=") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AND_EQUAL) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_COALESCE_EQUAL) reduce 292) ((T_CONCAT_EQUAL) reduce 292) ((T_DEC) reduce 292) ((T_DIV_EQUAL) reduce 292) ((T_INC) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_MINUS_EQUAL) reduce 292) ((T_MOD_EQUAL) reduce 292) ((T_MUL_EQUAL) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_OR_EQUAL) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_PLUS_EQUAL) reduce 292) ((T_POW) reduce 292) ((T_POW_EQUAL) reduce 292) ((T_SL) reduce 292) ((T_SL_EQUAL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) ((T_SR_EQUAL) reduce 292) ((T_XOR_EQUAL) reduce 292) (("[") reduce 292) (("]") reduce 292) (("^") reduce 292) (("|") reduce 292)) 9016 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) (("=") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AND_EQUAL) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_COALESCE_EQUAL) reduce 293) ((T_CONCAT_EQUAL) reduce 293) ((T_DEC) reduce 293) ((T_DIV_EQUAL) reduce 293) ((T_INC) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_MINUS_EQUAL) reduce 293) ((T_MOD_EQUAL) reduce 293) ((T_MUL_EQUAL) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_OR_EQUAL) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_PLUS_EQUAL) reduce 293) ((T_POW) reduce 293) ((T_POW_EQUAL) reduce 293) ((T_SL) reduce 293) ((T_SL_EQUAL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) ((T_SR_EQUAL) reduce 293) ((T_XOR_EQUAL) reduce 293) (("[") reduce 293) (("]") reduce 293) (("^") reduce 293) (("|") reduce 293)) 9021 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) ((",") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_PIPE) reduce 425) ((T_POW) reduce 425) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("]") reduce 425) (("^") reduce 425) (("|") reduce 425)) 9023 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) ((",") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_PIPE) reduce 423) ((T_POW) reduce 423) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("]") reduce 423) (("^") reduce 423) (("|") reduce 423)) 9033 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 419) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 419) (("^") shift) (("|") shift)) 9034 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 421) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 421) (("^") shift) (("|") shift)) 9035 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 420) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 420) (("^") shift) (("|") shift)) 9036 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 413) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 413) (("^") shift) (("|") shift)) 9037 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 410) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 410) (("^") shift) (("|") shift)) 9038 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 418) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 418) (("^") shift) (("|") shift)) 9039 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 412) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 412) (("^") shift) (("|") shift)) 9040 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 416) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 416) (("^") shift) (("|") shift)) 9041 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 411) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 411) (("^") shift) (("|") shift)) 9042 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 414) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 414) (("^") shift) (("|") shift)) 9043 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 415) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 415) (("^") shift) (("|") shift)) 9044 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 422) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 422) (("^") shift) (("|") shift)) 9045 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 417) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 417) (("^") shift) (("|") shift)) 9047 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 406) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 406) (("^") shift) (("|") shift)) 9050 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("]") reduce 568) (("^") reduce 568) (("|") reduce 568)) 9053 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) ((",") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) (("<") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_POW) reduce 564) ((T_SL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) (("[") reduce 564) (("]") reduce 564) (("^") reduce 564) (("|") reduce 564)) 9055 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) ((",") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) (("<") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_POW) reduce 563) ((T_SL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) (("[") reduce 563) (("]") reduce 563) (("^") reduce 563) (("|") reduce 563)) 9057 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("]") reduce 559) (("^") reduce 559) (("|") reduce 559)) 9058 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("]") reduce 565) (("^") reduce 565) (("|") reduce 565)) 9059 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) ((",") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("[") reduce 548) (("]") reduce 407) (("^") reduce 407) (("|") reduce 407)) 9063 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) ((",") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) (("<") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_POW) reduce 572) ((T_SL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) (("[") reduce 572) (("]") reduce 572) (("^") reduce 572) (("|") reduce 572)) 9065 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("]") reduce 507) (("^") reduce 507) (("|") reduce 507)) 9066 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("]") reduce 291) (("^") reduce 291) (("|") reduce 291)) 9070 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("[") reduce 292) (("]") reduce 292) (("^") reduce 292) (("|") reduce 292)) 9071 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("[") reduce 293) (("]") reduce 293) (("^") reduce 293) (("|") reduce 293)) 9072 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("]") reduce 504) (("^") reduce 504) (("|") reduce 504)) 9075 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) ((",") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) (("<") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_POW) reduce 571) ((T_SL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) (("[") reduce 571) (("]") reduce 571) (("^") reduce 571) (("|") reduce 571)) 9076 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("]") reduce 506) (("^") reduce 506) (("|") reduce 506)) 9077 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("]") reduce 508) (("^") reduce 508) (("|") reduce 508)) 9082 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("]") reduce 560) (("^") reduce 560) (("|") reduce 560)) 9083 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("]") reduce 582) (("^") reduce 582) (("|") reduce 582)) 9084 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("]") reduce 566) (("^") reduce 566) (("|") reduce 566)) 9085 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("]") reduce 584) (("^") reduce 584) (("|") reduce 584)) 9088 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("]") reduce 583) (("^") reduce 583) (("|") reduce 583)) 9089 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("]") reduce 561) (("^") reduce 561) (("|") reduce 561)) 9090 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("]") reduce 567) (("^") reduce 567) (("|") reduce 567)) 9091 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) ((",") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("[") reduce 562) (("]") reduce 562) (("^") reduce 562) (("|") reduce 562)) 9092 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("]") reduce 505) (("^") reduce 505) (("|") reduce 505)) 9093 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("]") reduce 570) (("^") reduce 570) (("|") reduce 570)) 9096 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("]") reduce 569) (("^") reduce 569) (("|") reduce 569)) 9097 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) (("=") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AND_EQUAL) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_COALESCE_EQUAL) reduce 504) ((T_CONCAT_EQUAL) reduce 504) ((T_DEC) reduce 504) ((T_DIV_EQUAL) reduce 504) ((T_INC) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_MINUS_EQUAL) reduce 504) ((T_MOD_EQUAL) reduce 504) ((T_MUL_EQUAL) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_OR_EQUAL) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_PLUS_EQUAL) reduce 504) ((T_POW) reduce 504) ((T_POW_EQUAL) reduce 504) ((T_SL) reduce 504) ((T_SL_EQUAL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) ((T_SR_EQUAL) reduce 504) ((T_XOR_EQUAL) reduce 504) (("[") reduce 504) (("]") reduce 504) (("^") reduce 504) (("|") reduce 504)) 9108 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) ((",") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PIPE) reduce 494) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("]") reduce 494) (("^") reduce 494) (("|") reduce 494)) 9109 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_PIPE) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("]") reduce 489) (("^") reduce 489) (("|") reduce 489)) 9182 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 494) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 494) (("^") shift) (("|") shift)) 9498 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_PIPE) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("]") reduce 490) (("^") reduce 490) (("|") reduce 490)) 9501 ((("%") reduce 541) (("(") reduce 579) (("*") reduce 541) (("+") reduce 541) ((",") reduce 541) (("-") reduce 541) ((".") reduce 541) (("/") reduce 541) (("<") reduce 541) ((">") reduce 541) (("?") reduce 541) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_BOOLEAN_AND) reduce 541) ((T_BOOLEAN_OR) reduce 541) ((T_COALESCE) reduce 541) ((T_INSTANCEOF) reduce 541) ((T_IS_EQUAL) reduce 541) ((T_IS_GREATER_OR_EQUAL) reduce 541) ((T_IS_IDENTICAL) reduce 541) ((T_IS_NOT_EQUAL) reduce 541) ((T_IS_NOT_IDENTICAL) reduce 541) ((T_IS_SMALLER_OR_EQUAL) reduce 541) ((T_LOGICAL_AND) reduce 541) ((T_LOGICAL_OR) reduce 541) ((T_LOGICAL_XOR) reduce 541) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 541) ((T_OBJECT_OPERATOR) reduce 541) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 541) ((T_PIPE) reduce 541) ((T_POW) reduce 541) ((T_SL) reduce 541) ((T_SPACESHIP) reduce 541) ((T_SR) reduce 541) (("[") reduce 541) (("]") reduce 541) (("^") reduce 541) (("|") reduce 541)) 9503 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) ((",") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) (("<") reduce 571) (("=") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AND_EQUAL) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_COALESCE_EQUAL) reduce 571) ((T_CONCAT_EQUAL) reduce 571) ((T_DEC) reduce 571) ((T_DIV_EQUAL) reduce 571) ((T_INC) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_MINUS_EQUAL) reduce 571) ((T_MOD_EQUAL) reduce 571) ((T_MUL_EQUAL) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_OR_EQUAL) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_PLUS_EQUAL) reduce 571) ((T_POW) reduce 571) ((T_POW_EQUAL) reduce 571) ((T_SL) reduce 571) ((T_SL_EQUAL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) ((T_SR_EQUAL) reduce 571) ((T_XOR_EQUAL) reduce 571) (("[") reduce 571) (("]") reduce 571) (("^") reduce 571) (("|") reduce 571)) 9506 ((("%") reduce 543) (("(") reduce 580) (("*") reduce 543) (("+") reduce 543) ((",") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PIPE) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("]") reduce 543) (("^") reduce 543) (("|") reduce 543)) 9507 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("]") reduce 506) (("^") reduce 506) (("|") reduce 506)) 9508 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("]") reduce 508) (("^") reduce 508) (("|") reduce 508)) 9510 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) ((",") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_PIPE) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("]") reduce 485) (("^") reduce 485) (("|") reduce 485)) 9511 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_PIPE) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("]") reduce 487) (("^") reduce 487) (("|") reduce 487)) 9516 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) (("=") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AND_EQUAL) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_COALESCE_EQUAL) reduce 560) ((T_CONCAT_EQUAL) reduce 560) ((T_DEC) reduce 560) ((T_DIV_EQUAL) reduce 560) ((T_INC) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_MINUS_EQUAL) reduce 560) ((T_MOD_EQUAL) reduce 560) ((T_MUL_EQUAL) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_OR_EQUAL) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_PLUS_EQUAL) reduce 560) ((T_POW) reduce 560) ((T_POW_EQUAL) reduce 560) ((T_SL) reduce 560) ((T_SL_EQUAL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) ((T_SR_EQUAL) reduce 560) ((T_XOR_EQUAL) reduce 560) (("[") reduce 560) (("]") reduce 560) (("^") reduce 560) (("|") reduce 560)) 9517 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) (("=") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AND_EQUAL) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_COALESCE_EQUAL) reduce 582) ((T_CONCAT_EQUAL) reduce 582) ((T_DEC) reduce 582) ((T_DIV_EQUAL) reduce 582) ((T_INC) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_MINUS_EQUAL) reduce 582) ((T_MOD_EQUAL) reduce 582) ((T_MUL_EQUAL) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_OR_EQUAL) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_PLUS_EQUAL) reduce 582) ((T_POW) reduce 582) ((T_POW_EQUAL) reduce 582) ((T_SL) reduce 582) ((T_SL_EQUAL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) ((T_SR_EQUAL) reduce 582) ((T_XOR_EQUAL) reduce 582) (("[") reduce 582) (("]") reduce 582) (("^") reduce 582) (("|") reduce 582)) 9518 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("]") reduce 566) (("^") reduce 566) (("|") reduce 566)) 9519 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) (("=") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AND_EQUAL) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_COALESCE_EQUAL) reduce 584) ((T_CONCAT_EQUAL) reduce 584) ((T_DEC) reduce 584) ((T_DIV_EQUAL) reduce 584) ((T_INC) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_MINUS_EQUAL) reduce 584) ((T_MOD_EQUAL) reduce 584) ((T_MUL_EQUAL) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_OR_EQUAL) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_PLUS_EQUAL) reduce 584) ((T_POW) reduce 584) ((T_POW_EQUAL) reduce 584) ((T_SL) reduce 584) ((T_SL_EQUAL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) ((T_SR_EQUAL) reduce 584) ((T_XOR_EQUAL) reduce 584) (("[") reduce 584) (("]") reduce 584) (("^") reduce 584) (("|") reduce 584)) 9522 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) (("=") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AND_EQUAL) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_COALESCE_EQUAL) reduce 583) ((T_CONCAT_EQUAL) reduce 583) ((T_DEC) reduce 583) ((T_DIV_EQUAL) reduce 583) ((T_INC) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_MINUS_EQUAL) reduce 583) ((T_MOD_EQUAL) reduce 583) ((T_MUL_EQUAL) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_OR_EQUAL) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_PLUS_EQUAL) reduce 583) ((T_POW) reduce 583) ((T_POW_EQUAL) reduce 583) ((T_SL) reduce 583) ((T_SL_EQUAL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) ((T_SR_EQUAL) reduce 583) ((T_XOR_EQUAL) reduce 583) (("[") reduce 583) (("]") reduce 583) (("^") reduce 583) (("|") reduce 583)) 9523 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) (("=") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AND_EQUAL) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_COALESCE_EQUAL) reduce 561) ((T_CONCAT_EQUAL) reduce 561) ((T_DEC) reduce 561) ((T_DIV_EQUAL) reduce 561) ((T_INC) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_MINUS_EQUAL) reduce 561) ((T_MOD_EQUAL) reduce 561) ((T_MUL_EQUAL) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_OR_EQUAL) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_PLUS_EQUAL) reduce 561) ((T_POW) reduce 561) ((T_POW_EQUAL) reduce 561) ((T_SL) reduce 561) ((T_SL_EQUAL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) ((T_SR_EQUAL) reduce 561) ((T_XOR_EQUAL) reduce 561) (("[") reduce 561) (("]") reduce 561) (("^") reduce 561) (("|") reduce 561)) 9524 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("]") reduce 567) (("^") reduce 567) (("|") reduce 567)) 9525 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) ((",") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) (("=") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AND_EQUAL) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_COALESCE_EQUAL) reduce 562) ((T_CONCAT_EQUAL) reduce 562) ((T_DEC) reduce 562) ((T_DIV_EQUAL) reduce 562) ((T_INC) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_MINUS_EQUAL) reduce 562) ((T_MOD_EQUAL) reduce 562) ((T_MUL_EQUAL) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_OR_EQUAL) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_PLUS_EQUAL) reduce 562) ((T_POW) reduce 562) ((T_POW_EQUAL) reduce 562) ((T_SL) reduce 562) ((T_SL_EQUAL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) ((T_SR_EQUAL) reduce 562) ((T_XOR_EQUAL) reduce 562) (("[") reduce 562) (("]") reduce 562) (("^") reduce 562) (("|") reduce 562)) 9527 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) ((",") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_PIPE) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("]") reduce 477) (("^") reduce 477) (("|") reduce 477)) 9529 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) (("=") shift) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_PIPE) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("]") reduce 520) (("^") reduce 520) (("|") reduce 520)) 9531 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 405) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 405) (("^") shift) (("|") shift)) 9532 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 482) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 482) (("^") shift) (("|") shift)) 9535 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 481) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 481) (("^") shift) (("|") shift)) 9536 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) ((",") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_PIPE) reduce 473) ((T_POW) shift) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("]") reduce 473) (("^") reduce 473) (("|") reduce 473)) 9537 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 483) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 483) (("^") shift) (("|") shift)) 9538 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) ((",") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_PIPE) reduce 469) ((T_POW) shift) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("]") reduce 469) (("^") reduce 469) (("|") reduce 469)) 9539 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) ((",") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_PIPE) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("]") reduce 486) (("^") reduce 486) (("|") reduce 486)) 9541 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_PIPE) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("]") reduce 526) (("^") reduce 526) (("|") reduce 526)) 9543 ((("%") reduce 527) (("*") reduce 527) (("+") reduce 527) ((",") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_PIPE) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("]") reduce 527) (("^") reduce 527) (("|") reduce 527)) 9544 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_PIPE) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("]") reduce 525) (("^") reduce 525) (("|") reduce 525)) 9545 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 619) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 619) (("^") shift) (("|") shift)) 9546 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 618) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 618) (("^") shift) (("|") shift)) 9547 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("]") reduce 505) (("^") reduce 505) (("|") reduce 505)) 9548 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 478) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 478) (("^") shift) (("|") shift)) 9549 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) ((",") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_PIPE) reduce 471) ((T_POW) shift) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("]") reduce 471) (("^") reduce 471) (("|") reduce 471)) 9551 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("]") reduce 94) (("^") reduce 94) (("|") reduce 94)) 9552 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("]") reduce 93) (("^") reduce 93) (("|") reduce 93)) 9553 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("]") reduce 95) (("^") reduce 95) (("|") reduce 95)) 9554 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("]") reduce 509) (("^") reduce 509) (("|") reduce 509)) 9555 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("]") reduce 92) (("^") reduce 92) (("|") reduce 92)) 9556 ((("%") reduce 400) (("(") reduce 400) (("*") reduce 400) (("+") reduce 400) ((",") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 400) ((T_OBJECT_OPERATOR) reduce 400) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 400) ((T_PIPE) reduce 400) ((T_POW) reduce 400) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("[") reduce 400) (("]") reduce 400) (("^") reduce 400) (("|") reduce 400)) 9559 ((("%") reduce 511) (("(") reduce 511) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("]") reduce 511) (("^") reduce 511) (("|") reduce 511)) 9560 ((("%") reduce 402) (("(") shift) (("*") reduce 402) (("+") reduce 402) ((",") reduce 402) (("-") reduce 402) ((".") reduce 402) (("/") reduce 402) (("<") reduce 402) ((">") reduce 402) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_INSTANCEOF) reduce 402) ((T_IS_EQUAL) reduce 402) ((T_IS_GREATER_OR_EQUAL) reduce 402) ((T_IS_IDENTICAL) reduce 402) ((T_IS_NOT_EQUAL) reduce 402) ((T_IS_NOT_IDENTICAL) reduce 402) ((T_IS_SMALLER_OR_EQUAL) reduce 402) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_PIPE) reduce 402) ((T_POW) reduce 402) ((T_SL) reduce 402) ((T_SPACESHIP) reduce 402) ((T_SR) reduce 402) (("]") reduce 402) (("^") reduce 402) (("|") reduce 402)) 9561 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("]") reduce 510) (("^") reduce 510) (("|") reduce 510)) 9562 ((("%") reduce 512) (("(") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("]") reduce 512) (("^") reduce 512) (("|") reduce 512)) 9563 ((("%") reduce 573) (("(") reduce 573) (("*") reduce 573) (("+") reduce 573) ((",") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("]") reduce 573) (("^") reduce 573) (("|") reduce 573)) 9569 ((("%") reduce 574) (("(") reduce 574) (("*") reduce 574) (("+") reduce 574) ((",") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("]") reduce 574) (("^") reduce 574) (("|") reduce 574)) 9570 ((("%") reduce 578) (("(") reduce 578) (("*") reduce 578) (("+") reduce 578) ((",") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("]") reduce 578) (("^") reduce 578) (("|") reduce 578)) 9571 ((("%") reduce 575) (("(") reduce 575) (("*") reduce 575) (("+") reduce 575) ((",") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("]") reduce 575) (("^") reduce 575) (("|") reduce 575)) 9572 ((("%") reduce 576) (("(") reduce 576) (("*") reduce 576) (("+") reduce 576) ((",") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("]") reduce 576) (("^") reduce 576) (("|") reduce 576)) 9573 ((("%") reduce 399) (("(") reduce 399) (("*") reduce 399) (("+") reduce 399) ((",") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_INSTANCEOF) reduce 399) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 399) ((T_OBJECT_OPERATOR) reduce 399) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 399) ((T_PIPE) reduce 399) ((T_POW) reduce 399) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("[") reduce 399) (("]") reduce 399) (("^") reduce 399) (("|") reduce 399)) 9575 ((("%") reduce 577) (("(") reduce 577) (("*") reduce 577) (("+") reduce 577) ((",") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("]") reduce 577) (("^") reduce 577) (("|") reduce 577)) 9576 ((("%") reduce 401) (("(") reduce 401) (("*") reduce 401) (("+") reduce 401) ((",") reduce 401) (("-") reduce 401) ((".") reduce 401) (("/") reduce 401) (("<") reduce 401) ((">") reduce 401) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_INSTANCEOF) reduce 401) ((T_IS_EQUAL) reduce 401) ((T_IS_GREATER_OR_EQUAL) reduce 401) ((T_IS_IDENTICAL) reduce 401) ((T_IS_NOT_EQUAL) reduce 401) ((T_IS_NOT_IDENTICAL) reduce 401) ((T_IS_SMALLER_OR_EQUAL) reduce 401) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 401) ((T_OBJECT_OPERATOR) reduce 401) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 401) ((T_PIPE) reduce 401) ((T_POW) reduce 401) ((T_SL) reduce 401) ((T_SPACESHIP) reduce 401) ((T_SR) reduce 401) (("[") reduce 401) (("]") reduce 401) (("^") reduce 401) (("|") reduce 401)) 9584 ((("%") reduce 398) (("(") reduce 398) (("*") reduce 398) (("+") reduce 398) ((",") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 398) ((T_OBJECT_OPERATOR) reduce 398) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 398) ((T_PIPE) reduce 398) ((T_POW) reduce 398) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("[") reduce 398) (("]") reduce 398) (("^") reduce 398) (("|") reduce 398)) 9586 ((("%") reduce 513) (("(") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("]") reduce 513) (("^") reduce 513) (("|") reduce 513)) 9592 ((("%") reduce 233) (("*") reduce 233) (("+") reduce 233) ((",") reduce 233) (("-") reduce 233) ((".") reduce 233) (("/") reduce 233) (("<") reduce 233) ((">") reduce 233) (("?") reduce 233) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_BOOLEAN_AND) reduce 233) ((T_BOOLEAN_OR) reduce 233) ((T_COALESCE) reduce 233) ((T_INSTANCEOF) reduce 233) ((T_IS_EQUAL) reduce 233) ((T_IS_GREATER_OR_EQUAL) reduce 233) ((T_IS_IDENTICAL) reduce 233) ((T_IS_NOT_EQUAL) reduce 233) ((T_IS_NOT_IDENTICAL) reduce 233) ((T_IS_SMALLER_OR_EQUAL) reduce 233) ((T_LOGICAL_AND) reduce 233) ((T_LOGICAL_OR) reduce 233) ((T_LOGICAL_XOR) reduce 233) ((T_PIPE) reduce 233) ((T_POW) reduce 233) ((T_SL) reduce 233) ((T_SPACESHIP) reduce 233) ((T_SR) reduce 233) (("]") reduce 233) (("^") reduce 233) (("|") reduce 233)) 9601 ((("%") reduce 613) (("*") reduce 613) (("+") reduce 613) ((",") reduce 613) (("-") reduce 613) ((".") reduce 613) (("/") reduce 613) (("<") reduce 613) ((">") reduce 613) (("?") reduce 613) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_BOOLEAN_AND) reduce 613) ((T_BOOLEAN_OR) reduce 613) ((T_COALESCE) reduce 613) ((T_INSTANCEOF) reduce 613) ((T_IS_EQUAL) reduce 613) ((T_IS_GREATER_OR_EQUAL) reduce 613) ((T_IS_IDENTICAL) reduce 613) ((T_IS_NOT_EQUAL) reduce 613) ((T_IS_NOT_IDENTICAL) reduce 613) ((T_IS_SMALLER_OR_EQUAL) reduce 613) ((T_LOGICAL_AND) reduce 613) ((T_LOGICAL_OR) reduce 613) ((T_LOGICAL_XOR) reduce 613) ((T_PIPE) reduce 613) ((T_POW) reduce 613) ((T_SL) reduce 613) ((T_SPACESHIP) reduce 613) ((T_SR) reduce 613) (("]") reduce 613) (("^") reduce 613) (("|") reduce 613)) 9602 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) ((",") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_PIPE) reduce 467) ((T_POW) shift) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("]") reduce 467) (("^") reduce 467) (("|") reduce 467)) 9603 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 616) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 616) (("^") shift) (("|") shift)) 9604 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 615) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 615) (("^") shift) (("|") shift)) 9605 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) ((",") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("[") reduce 548) (("]") reduce 424) (("^") reduce 424) (("|") reduce 424)) 9607 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_PIPE) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("]") reduce 518) (("^") reduce 518) (("|") reduce 518)) 9608 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) ((",") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_PIPE) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("]") reduce 474) (("^") reduce 474) (("|") reduce 474)) 9609 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("]") reduce 291) (("^") reduce 291) (("|") reduce 291)) 9613 ((("%") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("]") reduce 292) (("^") reduce 292) (("|") reduce 292)) 9614 ((("%") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("]") reduce 293) (("^") reduce 293) (("|") reduce 293)) 9617 ((("%") reduce 617) (("*") reduce 617) (("+") reduce 617) ((",") reduce 617) (("-") reduce 617) ((".") reduce 617) (("/") reduce 617) (("<") reduce 617) ((">") reduce 617) (("?") reduce 617) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_BOOLEAN_AND) reduce 617) ((T_BOOLEAN_OR) reduce 617) ((T_COALESCE) reduce 617) ((T_INSTANCEOF) reduce 617) ((T_IS_EQUAL) reduce 617) ((T_IS_GREATER_OR_EQUAL) reduce 617) ((T_IS_IDENTICAL) reduce 617) ((T_IS_NOT_EQUAL) reduce 617) ((T_IS_NOT_IDENTICAL) reduce 617) ((T_IS_SMALLER_OR_EQUAL) reduce 617) ((T_LOGICAL_AND) reduce 617) ((T_LOGICAL_OR) reduce 617) ((T_LOGICAL_XOR) reduce 617) ((T_PIPE) reduce 617) ((T_POW) reduce 617) ((T_SL) reduce 617) ((T_SPACESHIP) reduce 617) ((T_SR) reduce 617) (("]") reduce 617) (("^") reduce 617) (("|") reduce 617)) 9620 ((("%") reduce 614) (("*") reduce 614) (("+") reduce 614) ((",") reduce 614) (("-") reduce 614) ((".") reduce 614) (("/") reduce 614) (("<") reduce 614) ((">") reduce 614) (("?") reduce 614) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_BOOLEAN_AND) reduce 614) ((T_BOOLEAN_OR) reduce 614) ((T_COALESCE) reduce 614) ((T_INSTANCEOF) reduce 614) ((T_IS_EQUAL) reduce 614) ((T_IS_GREATER_OR_EQUAL) reduce 614) ((T_IS_IDENTICAL) reduce 614) ((T_IS_NOT_EQUAL) reduce 614) ((T_IS_NOT_IDENTICAL) reduce 614) ((T_IS_SMALLER_OR_EQUAL) reduce 614) ((T_LOGICAL_AND) reduce 614) ((T_LOGICAL_OR) reduce 614) ((T_LOGICAL_XOR) reduce 614) ((T_PIPE) reduce 614) ((T_POW) reduce 614) ((T_SL) reduce 614) ((T_SPACESHIP) reduce 614) ((T_SR) reduce 614) (("]") reduce 614) (("^") reduce 614) (("|") reduce 614)) 9621 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) ((",") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_PIPE) reduce 468) ((T_POW) shift) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("]") reduce 468) (("^") reduce 468) (("|") reduce 468)) 9622 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) ((",") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("[") reduce 548) (("]") reduce 426) (("^") reduce 426) (("|") reduce 426)) 9624 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) ((",") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_PIPE) reduce 408) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("]") reduce 408) (("^") reduce 408) (("|") reduce 408)) 9625 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) ((",") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_PIPE) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("]") reduce 409) (("^") reduce 409) (("|") reduce 409)) 9626 ((("%") reduce 296) (("*") reduce 296) (("+") reduce 296) ((",") reduce 296) (("-") reduce 296) ((".") reduce 296) (("/") reduce 296) (("<") reduce 296) ((">") reduce 296) (("?") reduce 296) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_BOOLEAN_AND) reduce 296) ((T_BOOLEAN_OR) reduce 296) ((T_COALESCE) reduce 296) ((T_INSTANCEOF) reduce 296) ((T_IS_EQUAL) reduce 296) ((T_IS_GREATER_OR_EQUAL) reduce 296) ((T_IS_IDENTICAL) reduce 296) ((T_IS_NOT_EQUAL) reduce 296) ((T_IS_NOT_IDENTICAL) reduce 296) ((T_IS_SMALLER_OR_EQUAL) reduce 296) ((T_LOGICAL_AND) reduce 296) ((T_LOGICAL_OR) reduce 296) ((T_LOGICAL_XOR) reduce 296) ((T_PIPE) reduce 296) ((T_POW) reduce 296) ((T_SL) reduce 296) ((T_SPACESHIP) reduce 296) ((T_SR) reduce 296) (("]") reduce 296) (("^") reduce 296) (("|") reduce 296)) 9631 ((("%") reduce 297) (("*") reduce 297) (("+") reduce 297) ((",") reduce 297) (("-") reduce 297) ((".") reduce 297) (("/") reduce 297) (("<") reduce 297) ((">") reduce 297) (("?") reduce 297) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_BOOLEAN_AND) reduce 297) ((T_BOOLEAN_OR) reduce 297) ((T_COALESCE) reduce 297) ((T_INSTANCEOF) reduce 297) ((T_IS_EQUAL) reduce 297) ((T_IS_GREATER_OR_EQUAL) reduce 297) ((T_IS_IDENTICAL) reduce 297) ((T_IS_NOT_EQUAL) reduce 297) ((T_IS_NOT_IDENTICAL) reduce 297) ((T_IS_SMALLER_OR_EQUAL) reduce 297) ((T_LOGICAL_AND) reduce 297) ((T_LOGICAL_OR) reduce 297) ((T_LOGICAL_XOR) reduce 297) ((T_PIPE) reduce 297) ((T_POW) reduce 297) ((T_SL) reduce 297) ((T_SPACESHIP) reduce 297) ((T_SR) reduce 297) (("]") reduce 297) (("^") reduce 297) (("|") reduce 297)) 9632 ((("%") reduce 460) (("(") reduce 556) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PIPE) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("[") reduce 549) (("]") reduce 460) (("^") reduce 460) (("|") reduce 460)) 9634 ((("%") reduce 298) (("*") reduce 298) (("+") reduce 298) ((",") reduce 298) (("-") reduce 298) ((".") reduce 298) (("/") reduce 298) (("<") reduce 298) ((">") reduce 298) (("?") reduce 298) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_BOOLEAN_AND) reduce 298) ((T_BOOLEAN_OR) reduce 298) ((T_COALESCE) reduce 298) ((T_INSTANCEOF) reduce 298) ((T_IS_EQUAL) reduce 298) ((T_IS_GREATER_OR_EQUAL) reduce 298) ((T_IS_IDENTICAL) reduce 298) ((T_IS_NOT_EQUAL) reduce 298) ((T_IS_NOT_IDENTICAL) reduce 298) ((T_IS_SMALLER_OR_EQUAL) reduce 298) ((T_LOGICAL_AND) reduce 298) ((T_LOGICAL_OR) reduce 298) ((T_LOGICAL_XOR) reduce 298) ((T_PIPE) reduce 298) ((T_POW) reduce 298) ((T_SL) reduce 298) ((T_SPACESHIP) reduce 298) ((T_SR) reduce 298) (("]") reduce 298) (("^") reduce 298) (("|") reduce 298)) 9635 ((("%") reduce 299) (("*") reduce 299) (("+") reduce 299) ((",") reduce 299) (("-") reduce 299) ((".") reduce 299) (("/") reduce 299) (("<") reduce 299) ((">") reduce 299) (("?") reduce 299) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_BOOLEAN_AND) reduce 299) ((T_BOOLEAN_OR) reduce 299) ((T_COALESCE) reduce 299) ((T_INSTANCEOF) reduce 299) ((T_IS_EQUAL) reduce 299) ((T_IS_GREATER_OR_EQUAL) reduce 299) ((T_IS_IDENTICAL) reduce 299) ((T_IS_NOT_EQUAL) reduce 299) ((T_IS_NOT_IDENTICAL) reduce 299) ((T_IS_SMALLER_OR_EQUAL) reduce 299) ((T_LOGICAL_AND) reduce 299) ((T_LOGICAL_OR) reduce 299) ((T_LOGICAL_XOR) reduce 299) ((T_PIPE) reduce 299) ((T_POW) reduce 299) ((T_SL) reduce 299) ((T_SPACESHIP) reduce 299) ((T_SR) reduce 299) (("]") reduce 299) (("^") reduce 299) (("|") reduce 299)) 9636 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) ((",") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_PIPE) reduce 472) ((T_POW) shift) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("]") reduce 472) (("^") reduce 472) (("|") reduce 472)) 9637 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) ((",") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_PIPE) reduce 470) ((T_POW) shift) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("]") reduce 470) (("^") reduce 470) (("|") reduce 470)) 9640 ((("%") reduce 519) (("(") reduce 519) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PIPE) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("]") reduce 519) (("^") reduce 519) (("|") reduce 519)) 9641 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) ((",") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_PIPE) reduce 475) ((T_POW) shift) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("]") reduce 475) (("^") reduce 475) (("|") reduce 475)) 9642 ((("%") reduce 446) (("*") reduce 446) (("+") reduce 446) ((",") reduce 446) (("-") reduce 446) ((".") reduce 446) (("/") reduce 446) (("<") reduce 446) ((">") reduce 446) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_BOOLEAN_AND) reduce 446) ((T_BOOLEAN_OR) reduce 446) ((T_COALESCE) reduce 446) ((T_INSTANCEOF) reduce 446) ((T_IS_EQUAL) reduce 446) ((T_IS_GREATER_OR_EQUAL) reduce 446) ((T_IS_IDENTICAL) reduce 446) ((T_IS_NOT_EQUAL) reduce 446) ((T_IS_NOT_IDENTICAL) reduce 446) ((T_IS_SMALLER_OR_EQUAL) reduce 446) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_PIPE) reduce 446) ((T_POW) shift) ((T_SL) reduce 446) ((T_SPACESHIP) reduce 446) ((T_SR) reduce 446) (("]") reduce 446) (("^") reduce 446) (("|") reduce 446)) 9643 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) ((",") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_PIPE) reduce 445) ((T_POW) shift) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("]") reduce 445) (("^") reduce 445) (("|") reduce 445)) 9645 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) (("=") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AND_EQUAL) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_COALESCE_EQUAL) reduce 570) ((T_CONCAT_EQUAL) reduce 570) ((T_DEC) reduce 570) ((T_DIV_EQUAL) reduce 570) ((T_INC) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_MINUS_EQUAL) reduce 570) ((T_MOD_EQUAL) reduce 570) ((T_MUL_EQUAL) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_OR_EQUAL) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_PLUS_EQUAL) reduce 570) ((T_POW) reduce 570) ((T_POW_EQUAL) reduce 570) ((T_SL) reduce 570) ((T_SL_EQUAL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) ((T_SR_EQUAL) reduce 570) ((T_XOR_EQUAL) reduce 570) (("[") reduce 570) (("]") reduce 570) (("^") reduce 570) (("|") reduce 570)) 9648 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) (("=") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AND_EQUAL) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_COALESCE_EQUAL) reduce 569) ((T_CONCAT_EQUAL) reduce 569) ((T_DEC) reduce 569) ((T_DIV_EQUAL) reduce 569) ((T_INC) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_MINUS_EQUAL) reduce 569) ((T_MOD_EQUAL) reduce 569) ((T_MUL_EQUAL) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_OR_EQUAL) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_PLUS_EQUAL) reduce 569) ((T_POW) reduce 569) ((T_POW_EQUAL) reduce 569) ((T_SL) reduce 569) ((T_SL_EQUAL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) ((T_SR_EQUAL) reduce 569) ((T_XOR_EQUAL) reduce 569) (("[") reduce 569) (("]") reduce 569) (("^") reduce 569) (("|") reduce 569)) 9650 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_PIPE) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("]") reduce 522) (("^") reduce 522) (("|") reduce 522)) 9651 ((("%") reduce 447) (("*") reduce 447) (("+") reduce 447) ((",") reduce 447) (("-") reduce 447) ((".") reduce 447) (("/") reduce 447) (("<") reduce 447) ((">") reduce 447) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_BOOLEAN_AND) reduce 447) ((T_BOOLEAN_OR) reduce 447) ((T_COALESCE) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 447) ((T_IS_GREATER_OR_EQUAL) reduce 447) ((T_IS_IDENTICAL) reduce 447) ((T_IS_NOT_EQUAL) reduce 447) ((T_IS_NOT_IDENTICAL) reduce 447) ((T_IS_SMALLER_OR_EQUAL) reduce 447) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_PIPE) reduce 447) ((T_POW) shift) ((T_SL) reduce 447) ((T_SPACESHIP) reduce 447) ((T_SR) reduce 447) (("]") reduce 447) (("^") reduce 447) (("|") reduce 447)) 9680 ((("(") reduce 568) ((",") reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) (("[") reduce 568) (("]") reduce 568)) 9683 ((("(") reduce 555) ((",") reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) (("[") reduce 564) (("]") reduce 564)) 9685 ((("(") reduce 563) ((",") reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) (("[") reduce 563) (("]") reduce 563)) 9687 ((("(") reduce 559) ((",") reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) (("[") reduce 559) (("]") reduce 559)) 9688 (((",") reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) (("[") reduce 565) (("]") reduce 565)) 9689 (((",") reduce 592) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) (("[") reduce 548) (("]") reduce 592)) 9693 ((("(") reduce 581) ((",") reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) (("[") reduce 572) (("]") reduce 572)) 9695 ((("(") reduce 507) ((",") reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) (("[") reduce 507) (("]") reduce 507)) 9696 ((("(") reduce 291) ((",") reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) (("[") reduce 291) (("]") reduce 291)) 9700 ((("(") reduce 292) ((",") reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) (("[") reduce 292) (("]") reduce 292)) 9701 ((("(") reduce 293) ((",") reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) (("[") reduce 293) (("]") reduce 293)) 9702 ((("(") reduce 504) ((",") reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) (("[") reduce 504) (("]") reduce 504)) 9705 ((("(") reduce 581) ((",") reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) (("[") reduce 571) (("]") reduce 571)) 9706 ((("(") reduce 506) ((",") reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) (("[") reduce 506) (("]") reduce 506)) 9707 ((("(") reduce 508) ((",") reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) (("[") reduce 508) (("]") reduce 508)) 9712 ((("(") reduce 560) ((",") reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) (("[") reduce 560) (("]") reduce 560)) 9713 ((("(") reduce 582) ((",") reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) (("[") reduce 582) (("]") reduce 582)) 9714 ((("(") shift) ((",") reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) (("[") reduce 566) (("]") reduce 566)) 9715 ((("(") reduce 584) ((",") reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) (("[") reduce 584) (("]") reduce 584)) 9718 ((("(") reduce 583) ((",") reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) (("[") reduce 583) (("]") reduce 583)) 9719 ((("(") reduce 561) ((",") reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) (("[") reduce 561) (("]") reduce 561)) 9720 ((("(") shift) ((",") reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) (("[") reduce 567) (("]") reduce 567)) 9721 ((("(") reduce 562) ((",") reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) (("[") reduce 562) (("]") reduce 562)) 9722 ((("(") reduce 505) ((",") reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) (("[") reduce 505) (("]") reduce 505)) 9723 ((("(") reduce 570) ((",") reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) (("[") reduce 570) (("]") reduce 570)) 9726 ((("(") reduce 569) ((",") reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) (("[") reduce 569) (("]") reduce 569)) 9796 (((",") reduce 595) (("=") shift) (("]") reduce 595)) 9863 ((("(") reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) (("[") reduce 520)) 9878 (((",") reduce 593) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) (("[") reduce 548) (("]") reduce 593)) 9948 (((",") reduce 596) (("=") shift) (("]") reduce 596)) 9972 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 594) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 594) (("^") shift) (("|") shift)) 10007 ((("(") reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94)) 10008 ((("(") reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93)) 10009 ((("(") reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95)) 10010 ((("(") reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509)) 10011 ((("(") reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92)) 10012 ((("(") reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) (("[") reduce 568)) 10013 ((("(") reduce 400) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 400) ((T_OBJECT_OPERATOR) reduce 400) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 400) (("[") reduce 400)) 10016 ((("(") reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift)) 10018 ((("(") reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510)) 10019 ((("(") reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) (("[") shift)) 10020 ((("(") reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) (("[") reduce 573)) 10026 ((("(") reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) (("[") reduce 574)) 10027 ((("(") reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) (("[") reduce 578)) 10028 ((("(") reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) (("[") reduce 582)) 10029 ((("(") reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) (("[") reduce 575)) 10030 ((("(") reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) (("[") reduce 584)) 10033 ((("(") reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) (("[") reduce 583)) 10034 ((("(") reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) (("[") reduce 576)) 10036 ((("(") reduce 399) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 399) ((T_OBJECT_OPERATOR) reduce 399) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 399) (("[") reduce 399)) 10037 ((("(") reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) (("[") reduce 291)) 10041 ((("(") reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) (("[") reduce 292)) 10042 ((("(") reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) (("[") reduce 293)) 10044 ((("(") reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) (("[") reduce 577)) 10045 ((("(") reduce 401) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 401) ((T_OBJECT_OPERATOR) reduce 401) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 401) (("[") reduce 401)) 10053 ((("(") reduce 398) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 398) ((T_OBJECT_OPERATOR) reduce 398) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 398) (("[") reduce 398)) 10055 ((("(") reduce 513)) 10056 ((("(") reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) (("[") reduce 570)) 10059 ((("(") reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) (("[") reduce 569)) 10062 ((("(") reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) (("[") reduce 519)) 10064 ((("(") reduce 556) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) (("[") reduce 549)) 10070 ((("(") reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) (("[") reduce 522)) 10517 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) ((":") reduce 292) (("<") reduce 292) (("=") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AND_EQUAL) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_COALESCE_EQUAL) reduce 292) ((T_CONCAT_EQUAL) reduce 292) ((T_DEC) reduce 292) ((T_DIV_EQUAL) reduce 292) ((T_INC) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_MINUS_EQUAL) reduce 292) ((T_MOD_EQUAL) reduce 292) ((T_MUL_EQUAL) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_OR_EQUAL) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_PLUS_EQUAL) reduce 292) ((T_POW) reduce 292) ((T_POW_EQUAL) reduce 292) ((T_SL) reduce 292) ((T_SL_EQUAL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) ((T_SR_EQUAL) reduce 292) ((T_XOR_EQUAL) reduce 292) (("[") reduce 292) (("^") reduce 292) (("|") reduce 292)) 10520 ((("%") shift) ((")") reduce 303) (("*") shift) (("+") shift) ((",") reduce 303) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10587 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) ((":") reduce 293) (("<") reduce 293) (("=") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AND_EQUAL) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_COALESCE_EQUAL) reduce 293) ((T_CONCAT_EQUAL) reduce 293) ((T_DEC) reduce 293) ((T_DIV_EQUAL) reduce 293) ((T_INC) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_MINUS_EQUAL) reduce 293) ((T_MOD_EQUAL) reduce 293) ((T_MUL_EQUAL) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_OR_EQUAL) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_PLUS_EQUAL) reduce 293) ((T_POW) reduce 293) ((T_POW_EQUAL) reduce 293) ((T_SL) reduce 293) ((T_SL_EQUAL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) ((T_SR_EQUAL) reduce 293) ((T_XOR_EQUAL) reduce 293) (("[") reduce 293) (("^") reduce 293) (("|") reduce 293)) 10607 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) ((":") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_PIPE) reduce 425) ((T_POW) reduce 425) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 10609 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) ((":") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_PIPE) reduce 423) ((T_POW) reduce 423) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 10619 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 419) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10620 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 421) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10621 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 420) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10622 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 413) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10623 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 410) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10624 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 418) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10625 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 412) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10626 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 416) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10627 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 411) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10628 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 414) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10629 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 415) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10630 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 422) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10631 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 417) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10633 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 406) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 10636 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) ((":") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 10639 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) ((":") reduce 564) (("<") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_POW) reduce 564) ((T_SL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) (("[") reduce 564) (("^") reduce 564) (("|") reduce 564)) 10641 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) ((":") reduce 563) (("<") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_POW) reduce 563) ((T_SL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) (("[") reduce 563) (("^") reduce 563) (("|") reduce 563)) 10643 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) ((":") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 10644 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((":") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 10645 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) ((":") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("[") reduce 548) (("^") reduce 407) (("|") reduce 407)) 10649 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) ((":") reduce 572) (("<") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_POW) reduce 572) ((T_SL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) (("[") reduce 572) (("^") reduce 572) (("|") reduce 572)) 10651 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((":") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("|") reduce 507)) 10652 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((":") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 10656 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) ((":") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("[") reduce 292) (("^") reduce 292) (("|") reduce 292)) 10657 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) ((":") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("[") reduce 293) (("^") reduce 293) (("|") reduce 293)) 10658 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) ((":") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 10661 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) ((":") reduce 571) (("<") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_POW) reduce 571) ((T_SL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) (("[") reduce 571) (("^") reduce 571) (("|") reduce 571)) 10662 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((":") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("|") reduce 506)) 10663 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((":") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("|") reduce 508)) 10668 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) ((":") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 10669 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) ((":") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 10670 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((":") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 10671 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) ((":") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 10674 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) ((":") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 10675 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((":") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 10676 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((":") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 10677 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) ((":") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("[") reduce 562) (("^") reduce 562) (("|") reduce 562)) 10678 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((":") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 10679 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) ((":") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 10682 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) ((":") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 10683 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) ((":") reduce 504) (("<") reduce 504) (("=") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AND_EQUAL) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_COALESCE_EQUAL) reduce 504) ((T_CONCAT_EQUAL) reduce 504) ((T_DEC) reduce 504) ((T_DIV_EQUAL) reduce 504) ((T_INC) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_MINUS_EQUAL) reduce 504) ((T_MOD_EQUAL) reduce 504) ((T_MUL_EQUAL) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_OR_EQUAL) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_PLUS_EQUAL) reduce 504) ((T_POW) reduce 504) ((T_POW_EQUAL) reduce 504) ((T_SL) reduce 504) ((T_SL_EQUAL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) ((T_SR_EQUAL) reduce 504) ((T_XOR_EQUAL) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 10694 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) ((":") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PIPE) reduce 494) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 10695 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) ((":") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_PIPE) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("^") reduce 489) (("|") reduce 489)) 10768 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 494) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11084 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) ((":") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_PIPE) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("^") reduce 490) (("|") reduce 490)) 11087 ((("%") reduce 541) (("(") reduce 579) (("*") reduce 541) (("+") reduce 541) (("-") reduce 541) ((".") reduce 541) (("/") reduce 541) ((":") reduce 541) (("<") reduce 541) ((">") reduce 541) (("?") reduce 541) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_BOOLEAN_AND) reduce 541) ((T_BOOLEAN_OR) reduce 541) ((T_COALESCE) reduce 541) ((T_INSTANCEOF) reduce 541) ((T_IS_EQUAL) reduce 541) ((T_IS_GREATER_OR_EQUAL) reduce 541) ((T_IS_IDENTICAL) reduce 541) ((T_IS_NOT_EQUAL) reduce 541) ((T_IS_NOT_IDENTICAL) reduce 541) ((T_IS_SMALLER_OR_EQUAL) reduce 541) ((T_LOGICAL_AND) reduce 541) ((T_LOGICAL_OR) reduce 541) ((T_LOGICAL_XOR) reduce 541) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 541) ((T_OBJECT_OPERATOR) reduce 541) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 541) ((T_PIPE) reduce 541) ((T_POW) reduce 541) ((T_SL) reduce 541) ((T_SPACESHIP) reduce 541) ((T_SR) reduce 541) (("[") reduce 541) (("^") reduce 541) (("|") reduce 541)) 11089 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) ((":") reduce 571) (("<") reduce 571) (("=") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AND_EQUAL) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_COALESCE_EQUAL) reduce 571) ((T_CONCAT_EQUAL) reduce 571) ((T_DEC) reduce 571) ((T_DIV_EQUAL) reduce 571) ((T_INC) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_MINUS_EQUAL) reduce 571) ((T_MOD_EQUAL) reduce 571) ((T_MUL_EQUAL) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_OR_EQUAL) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_PLUS_EQUAL) reduce 571) ((T_POW) reduce 571) ((T_POW_EQUAL) reduce 571) ((T_SL) reduce 571) ((T_SL_EQUAL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) ((T_SR_EQUAL) reduce 571) ((T_XOR_EQUAL) reduce 571) (("[") reduce 571) (("^") reduce 571) (("|") reduce 571)) 11092 ((("%") reduce 543) (("(") reduce 580) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) ((":") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PIPE) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 11093 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((":") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("|") reduce 506)) 11094 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((":") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("|") reduce 508)) 11096 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) ((":") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_PIPE) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("^") reduce 485) (("|") reduce 485)) 11097 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) ((":") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_PIPE) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("^") reduce 487) (("|") reduce 487)) 11102 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) ((":") reduce 560) (("<") reduce 560) (("=") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AND_EQUAL) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_COALESCE_EQUAL) reduce 560) ((T_CONCAT_EQUAL) reduce 560) ((T_DEC) reduce 560) ((T_DIV_EQUAL) reduce 560) ((T_INC) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_MINUS_EQUAL) reduce 560) ((T_MOD_EQUAL) reduce 560) ((T_MUL_EQUAL) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_OR_EQUAL) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_PLUS_EQUAL) reduce 560) ((T_POW) reduce 560) ((T_POW_EQUAL) reduce 560) ((T_SL) reduce 560) ((T_SL_EQUAL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) ((T_SR_EQUAL) reduce 560) ((T_XOR_EQUAL) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 11103 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) ((":") reduce 582) (("<") reduce 582) (("=") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AND_EQUAL) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_COALESCE_EQUAL) reduce 582) ((T_CONCAT_EQUAL) reduce 582) ((T_DEC) reduce 582) ((T_DIV_EQUAL) reduce 582) ((T_INC) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_MINUS_EQUAL) reduce 582) ((T_MOD_EQUAL) reduce 582) ((T_MUL_EQUAL) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_OR_EQUAL) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_PLUS_EQUAL) reduce 582) ((T_POW) reduce 582) ((T_POW_EQUAL) reduce 582) ((T_SL) reduce 582) ((T_SL_EQUAL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) ((T_SR_EQUAL) reduce 582) ((T_XOR_EQUAL) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 11104 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((":") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 11105 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) ((":") reduce 584) (("<") reduce 584) (("=") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AND_EQUAL) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_COALESCE_EQUAL) reduce 584) ((T_CONCAT_EQUAL) reduce 584) ((T_DEC) reduce 584) ((T_DIV_EQUAL) reduce 584) ((T_INC) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_MINUS_EQUAL) reduce 584) ((T_MOD_EQUAL) reduce 584) ((T_MUL_EQUAL) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_OR_EQUAL) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_PLUS_EQUAL) reduce 584) ((T_POW) reduce 584) ((T_POW_EQUAL) reduce 584) ((T_SL) reduce 584) ((T_SL_EQUAL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) ((T_SR_EQUAL) reduce 584) ((T_XOR_EQUAL) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 11108 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) ((":") reduce 583) (("<") reduce 583) (("=") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AND_EQUAL) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_COALESCE_EQUAL) reduce 583) ((T_CONCAT_EQUAL) reduce 583) ((T_DEC) reduce 583) ((T_DIV_EQUAL) reduce 583) ((T_INC) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_MINUS_EQUAL) reduce 583) ((T_MOD_EQUAL) reduce 583) ((T_MUL_EQUAL) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_OR_EQUAL) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_PLUS_EQUAL) reduce 583) ((T_POW) reduce 583) ((T_POW_EQUAL) reduce 583) ((T_SL) reduce 583) ((T_SL_EQUAL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) ((T_SR_EQUAL) reduce 583) ((T_XOR_EQUAL) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 11109 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((":") reduce 561) (("<") reduce 561) (("=") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AND_EQUAL) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_COALESCE_EQUAL) reduce 561) ((T_CONCAT_EQUAL) reduce 561) ((T_DEC) reduce 561) ((T_DIV_EQUAL) reduce 561) ((T_INC) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_MINUS_EQUAL) reduce 561) ((T_MOD_EQUAL) reduce 561) ((T_MUL_EQUAL) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_OR_EQUAL) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_PLUS_EQUAL) reduce 561) ((T_POW) reduce 561) ((T_POW_EQUAL) reduce 561) ((T_SL) reduce 561) ((T_SL_EQUAL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) ((T_SR_EQUAL) reduce 561) ((T_XOR_EQUAL) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 11110 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((":") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 11111 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) ((":") reduce 562) (("<") reduce 562) (("=") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AND_EQUAL) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_COALESCE_EQUAL) reduce 562) ((T_CONCAT_EQUAL) reduce 562) ((T_DEC) reduce 562) ((T_DIV_EQUAL) reduce 562) ((T_INC) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_MINUS_EQUAL) reduce 562) ((T_MOD_EQUAL) reduce 562) ((T_MUL_EQUAL) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_OR_EQUAL) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_PLUS_EQUAL) reduce 562) ((T_POW) reduce 562) ((T_POW_EQUAL) reduce 562) ((T_SL) reduce 562) ((T_SL_EQUAL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) ((T_SR_EQUAL) reduce 562) ((T_XOR_EQUAL) reduce 562) (("[") reduce 562) (("^") reduce 562) (("|") reduce 562)) 11113 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) ((":") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_PIPE) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 11115 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) ((":") reduce 520) (("<") reduce 520) (("=") shift) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_PIPE) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("|") reduce 520)) 11117 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 405) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11118 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 482) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11129 ((("%") reduce 540) (("*") reduce 540) (("+") reduce 540) (("-") reduce 540) ((".") reduce 540) (("/") reduce 540) ((":") reduce 540) (("<") reduce 540) ((">") reduce 540) (("?") reduce 540) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_BOOLEAN_AND) reduce 540) ((T_BOOLEAN_OR) reduce 540) ((T_COALESCE) reduce 540) ((T_DOUBLE_ARROW) reduce 540) ((T_INSTANCEOF) reduce 540) ((T_IS_EQUAL) reduce 540) ((T_IS_GREATER_OR_EQUAL) reduce 540) ((T_IS_IDENTICAL) reduce 540) ((T_IS_NOT_EQUAL) reduce 540) ((T_IS_NOT_IDENTICAL) reduce 540) ((T_IS_SMALLER_OR_EQUAL) reduce 540) ((T_LOGICAL_AND) reduce 540) ((T_LOGICAL_OR) reduce 540) ((T_LOGICAL_XOR) reduce 540) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 540) ((T_OBJECT_OPERATOR) reduce 540) ((T_PIPE) reduce 540) ((T_POW) reduce 540) ((T_SL) reduce 540) ((T_SPACESHIP) reduce 540) ((T_SR) reduce 540) (("[") reduce 540) (("^") reduce 540) (("|") reduce 540)) 11131 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) ((":") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_PIPE) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("|") reduce 521)) 11133 ((("%") reduce 534) (("*") reduce 534) (("+") reduce 534) (("-") reduce 534) ((".") reduce 534) (("/") reduce 534) ((":") reduce 534) (("<") reduce 534) ((">") reduce 534) (("?") reduce 534) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_BOOLEAN_AND) reduce 534) ((T_BOOLEAN_OR) reduce 534) ((T_COALESCE) reduce 534) ((T_DOUBLE_ARROW) reduce 534) ((T_INSTANCEOF) reduce 534) ((T_IS_EQUAL) reduce 534) ((T_IS_GREATER_OR_EQUAL) reduce 534) ((T_IS_IDENTICAL) reduce 534) ((T_IS_NOT_EQUAL) reduce 534) ((T_IS_NOT_IDENTICAL) reduce 534) ((T_IS_SMALLER_OR_EQUAL) reduce 534) ((T_LOGICAL_AND) reduce 534) ((T_LOGICAL_OR) reduce 534) ((T_LOGICAL_XOR) reduce 534) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PIPE) reduce 534) ((T_POW) reduce 534) ((T_SL) reduce 534) ((T_SPACESHIP) reduce 534) ((T_SR) reduce 534) (("[") reduce 534) (("^") reduce 534) (("|") reduce 534)) 11134 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) ((":") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_PIPE) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("^") reduce 524) (("|") reduce 524)) 11138 ((("%") reduce 517) (("(") shift) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((":") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_PIPE) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("^") reduce 517) (("|") reduce 517)) 11139 ((("%") reduce 533) (("*") reduce 533) (("+") reduce 533) (("-") reduce 533) ((".") reduce 533) (("/") reduce 533) ((":") reduce 533) (("<") reduce 533) ((">") reduce 533) (("?") reduce 533) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_BOOLEAN_AND) reduce 533) ((T_BOOLEAN_OR) reduce 533) ((T_COALESCE) reduce 533) ((T_DOUBLE_ARROW) reduce 533) ((T_INSTANCEOF) reduce 533) ((T_IS_EQUAL) reduce 533) ((T_IS_GREATER_OR_EQUAL) reduce 533) ((T_IS_IDENTICAL) reduce 533) ((T_IS_NOT_EQUAL) reduce 533) ((T_IS_NOT_IDENTICAL) reduce 533) ((T_IS_SMALLER_OR_EQUAL) reduce 533) ((T_LOGICAL_AND) reduce 533) ((T_LOGICAL_OR) reduce 533) ((T_LOGICAL_XOR) reduce 533) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PIPE) reduce 533) ((T_POW) reduce 533) ((T_SL) reduce 533) ((T_SPACESHIP) reduce 533) ((T_SR) reduce 533) (("[") reduce 533) (("^") reduce 533) (("|") reduce 533)) 11140 ((("%") reduce 537) (("*") reduce 537) (("+") reduce 537) (("-") reduce 537) ((".") reduce 537) (("/") reduce 537) ((":") reduce 537) (("<") reduce 537) ((">") reduce 537) (("?") reduce 537) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_BOOLEAN_AND) reduce 537) ((T_BOOLEAN_OR) reduce 537) ((T_COALESCE) reduce 537) ((T_DOUBLE_ARROW) reduce 537) ((T_INSTANCEOF) reduce 537) ((T_IS_EQUAL) reduce 537) ((T_IS_GREATER_OR_EQUAL) reduce 537) ((T_IS_IDENTICAL) reduce 537) ((T_IS_NOT_EQUAL) reduce 537) ((T_IS_NOT_IDENTICAL) reduce 537) ((T_IS_SMALLER_OR_EQUAL) reduce 537) ((T_LOGICAL_AND) reduce 537) ((T_LOGICAL_OR) reduce 537) ((T_LOGICAL_XOR) reduce 537) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_PIPE) reduce 537) ((T_POW) reduce 537) ((T_SL) reduce 537) ((T_SPACESHIP) reduce 537) ((T_SR) reduce 537) (("[") reduce 537) (("^") reduce 537) (("|") reduce 537)) 11146 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) ((":") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PIPE) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("|") reduce 532)) 11148 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) ((":") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_PIPE) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("^") reduce 523) (("|") reduce 523)) 11150 ((("%") reduce 536) (("*") reduce 536) (("+") reduce 536) (("-") reduce 536) ((".") reduce 536) (("/") reduce 536) ((":") reduce 536) (("<") reduce 536) ((">") reduce 536) (("?") reduce 536) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_BOOLEAN_AND) reduce 536) ((T_BOOLEAN_OR) reduce 536) ((T_COALESCE) reduce 536) ((T_DOUBLE_ARROW) reduce 536) ((T_INSTANCEOF) reduce 536) ((T_IS_EQUAL) reduce 536) ((T_IS_GREATER_OR_EQUAL) reduce 536) ((T_IS_IDENTICAL) reduce 536) ((T_IS_NOT_EQUAL) reduce 536) ((T_IS_NOT_IDENTICAL) reduce 536) ((T_IS_SMALLER_OR_EQUAL) reduce 536) ((T_LOGICAL_AND) reduce 536) ((T_LOGICAL_OR) reduce 536) ((T_LOGICAL_XOR) reduce 536) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 536) ((T_OBJECT_OPERATOR) reduce 536) ((T_PIPE) reduce 536) ((T_POW) reduce 536) ((T_SL) reduce 536) ((T_SPACESHIP) reduce 536) ((T_SR) reduce 536) (("[") reduce 536) (("^") reduce 536) (("|") reduce 536)) 11151 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((":") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("|") reduce 94)) 11152 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((":") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("|") reduce 93)) 11153 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((":") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("|") reduce 95)) 11155 ((("%") reduce 539) (("*") reduce 539) (("+") reduce 539) (("-") reduce 539) ((".") reduce 539) (("/") reduce 539) ((":") reduce 539) (("<") reduce 539) ((">") reduce 539) (("?") reduce 539) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_BOOLEAN_AND) reduce 539) ((T_BOOLEAN_OR) reduce 539) ((T_COALESCE) reduce 539) ((T_DOUBLE_ARROW) reduce 539) ((T_INSTANCEOF) reduce 539) ((T_IS_EQUAL) reduce 539) ((T_IS_GREATER_OR_EQUAL) reduce 539) ((T_IS_IDENTICAL) reduce 539) ((T_IS_NOT_EQUAL) reduce 539) ((T_IS_NOT_IDENTICAL) reduce 539) ((T_IS_SMALLER_OR_EQUAL) reduce 539) ((T_LOGICAL_AND) reduce 539) ((T_LOGICAL_OR) reduce 539) ((T_LOGICAL_XOR) reduce 539) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 539) ((T_OBJECT_OPERATOR) reduce 539) ((T_PIPE) reduce 539) ((T_POW) reduce 539) ((T_SL) reduce 539) ((T_SPACESHIP) reduce 539) ((T_SR) reduce 539) (("[") reduce 539) (("^") reduce 539) (("|") reduce 539)) 11158 ((("%") reduce 538) (("*") reduce 538) (("+") reduce 538) (("-") reduce 538) ((".") reduce 538) (("/") reduce 538) ((":") reduce 538) (("<") reduce 538) ((">") reduce 538) (("?") reduce 538) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_BOOLEAN_AND) reduce 538) ((T_BOOLEAN_OR) reduce 538) ((T_COALESCE) reduce 538) ((T_DOUBLE_ARROW) reduce 538) ((T_INSTANCEOF) reduce 538) ((T_IS_EQUAL) reduce 538) ((T_IS_GREATER_OR_EQUAL) reduce 538) ((T_IS_IDENTICAL) reduce 538) ((T_IS_NOT_EQUAL) reduce 538) ((T_IS_NOT_IDENTICAL) reduce 538) ((T_IS_SMALLER_OR_EQUAL) reduce 538) ((T_LOGICAL_AND) reduce 538) ((T_LOGICAL_OR) reduce 538) ((T_LOGICAL_XOR) reduce 538) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 538) ((T_OBJECT_OPERATOR) reduce 538) ((T_PIPE) reduce 538) ((T_POW) reduce 538) ((T_SL) reduce 538) ((T_SPACESHIP) reduce 538) ((T_SR) reduce 538) (("[") reduce 538) (("^") reduce 538) (("|") reduce 538)) 11164 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((":") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92)) 11167 ((("%") reduce 535) (("*") reduce 535) (("+") reduce 535) (("-") reduce 535) ((".") reduce 535) (("/") reduce 535) ((":") reduce 535) (("<") reduce 535) ((">") reduce 535) (("?") reduce 535) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_BOOLEAN_AND) reduce 535) ((T_BOOLEAN_OR) reduce 535) ((T_COALESCE) reduce 535) ((T_DOUBLE_ARROW) reduce 535) ((T_INSTANCEOF) reduce 535) ((T_IS_EQUAL) reduce 535) ((T_IS_GREATER_OR_EQUAL) reduce 535) ((T_IS_IDENTICAL) reduce 535) ((T_IS_NOT_EQUAL) reduce 535) ((T_IS_NOT_IDENTICAL) reduce 535) ((T_IS_SMALLER_OR_EQUAL) reduce 535) ((T_LOGICAL_AND) reduce 535) ((T_LOGICAL_OR) reduce 535) ((T_LOGICAL_XOR) reduce 535) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PIPE) reduce 535) ((T_POW) reduce 535) ((T_SL) reduce 535) ((T_SPACESHIP) reduce 535) ((T_SR) reduce 535) (("[") reduce 535) (("^") reduce 535) (("|") reduce 535)) 11169 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) ((":") reduce 568) (("<") reduce 568) (("=") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AND_EQUAL) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_COALESCE_EQUAL) reduce 568) ((T_CONCAT_EQUAL) reduce 568) ((T_DEC) reduce 568) ((T_DIV_EQUAL) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INC) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_MINUS_EQUAL) reduce 568) ((T_MOD_EQUAL) reduce 568) ((T_MUL_EQUAL) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_OR_EQUAL) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_PLUS_EQUAL) reduce 568) ((T_POW) reduce 568) ((T_POW_EQUAL) reduce 568) ((T_SL) reduce 568) ((T_SL_EQUAL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) ((T_SR_EQUAL) reduce 568) ((T_XOR_EQUAL) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 11170 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 479) (("(") shift) (("*") reduce 479) (("+") shift) (("-") shift) ((".") reduce 479) (("/") reduce 479) ((":") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 479) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 479) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 479) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PIPE) reduce 479) ((T_POW) reduce 479) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 479) (("`") shift) (("|") reduce 479) (("~") shift)) 11177 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) ((":") reduce 564) (("<") reduce 564) (("=") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AND_EQUAL) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_COALESCE_EQUAL) reduce 564) ((T_CONCAT_EQUAL) reduce 564) ((T_DEC) reduce 564) ((T_DIV_EQUAL) reduce 564) ((T_DOUBLE_ARROW) reduce 564) ((T_INC) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_MINUS_EQUAL) reduce 564) ((T_MOD_EQUAL) reduce 564) ((T_MUL_EQUAL) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_OR_EQUAL) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_PLUS_EQUAL) reduce 564) ((T_POW) reduce 564) ((T_POW_EQUAL) reduce 564) ((T_SL) reduce 564) ((T_SL_EQUAL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) ((T_SR_EQUAL) reduce 564) ((T_XOR_EQUAL) reduce 564) (("[") reduce 564) (("^") reduce 564) (("|") reduce 564)) 11178 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) ((":") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PIPE) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 551) (("^") reduce 530) (("|") reduce 530)) 11180 ((("%") reduce 529) (("*") reduce 529) (("+") reduce 529) (("-") reduce 529) ((".") reduce 529) (("/") reduce 529) ((":") reduce 529) (("<") reduce 529) ((">") reduce 529) (("?") reduce 529) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_BOOLEAN_AND) reduce 529) ((T_BOOLEAN_OR) reduce 529) ((T_COALESCE) reduce 529) ((T_DOUBLE_ARROW) reduce 529) ((T_INSTANCEOF) reduce 529) ((T_IS_EQUAL) reduce 529) ((T_IS_GREATER_OR_EQUAL) reduce 529) ((T_IS_IDENTICAL) reduce 529) ((T_IS_NOT_EQUAL) reduce 529) ((T_IS_NOT_IDENTICAL) reduce 529) ((T_IS_SMALLER_OR_EQUAL) reduce 529) ((T_LOGICAL_AND) reduce 529) ((T_LOGICAL_OR) reduce 529) ((T_LOGICAL_XOR) reduce 529) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PIPE) reduce 529) ((T_POW) reduce 529) ((T_SL) reduce 529) ((T_SPACESHIP) reduce 529) ((T_SR) reduce 529) (("[") reduce 554) (("^") reduce 529) (("|") reduce 529)) 11181 ((("%") reduce 528) (("(") reduce 557) (("*") reduce 528) (("+") reduce 528) (("-") reduce 528) ((".") reduce 528) (("/") reduce 528) ((":") reduce 528) (("<") reduce 528) ((">") reduce 528) (("?") reduce 528) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_BOOLEAN_AND) reduce 528) ((T_BOOLEAN_OR) reduce 528) ((T_COALESCE) reduce 528) ((T_DOUBLE_ARROW) reduce 528) ((T_INSTANCEOF) reduce 528) ((T_IS_EQUAL) reduce 528) ((T_IS_GREATER_OR_EQUAL) reduce 528) ((T_IS_IDENTICAL) reduce 528) ((T_IS_NOT_EQUAL) reduce 528) ((T_IS_NOT_IDENTICAL) reduce 528) ((T_IS_SMALLER_OR_EQUAL) reduce 528) ((T_LOGICAL_AND) reduce 528) ((T_LOGICAL_OR) reduce 528) ((T_LOGICAL_XOR) reduce 528) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PIPE) reduce 528) ((T_POW) reduce 528) ((T_SL) reduce 528) ((T_SPACESHIP) reduce 528) ((T_SR) reduce 528) (("[") reduce 550) (("^") reduce 528) (("|") reduce 528)) 11182 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 480) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 480) ((T_LOGICAL_OR) reduce 480) ((T_LOGICAL_XOR) reduce 480) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11185 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) ((":") reduce 563) (("<") reduce 563) (("=") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AND_EQUAL) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_COALESCE_EQUAL) reduce 563) ((T_CONCAT_EQUAL) reduce 563) ((T_DEC) reduce 563) ((T_DIV_EQUAL) reduce 563) ((T_DOUBLE_ARROW) reduce 563) ((T_INC) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_MINUS_EQUAL) reduce 563) ((T_MOD_EQUAL) reduce 563) ((T_MUL_EQUAL) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_OR_EQUAL) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_PLUS_EQUAL) reduce 563) ((T_POW) reduce 563) ((T_POW_EQUAL) reduce 563) ((T_SL) reduce 563) ((T_SL_EQUAL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) ((T_SR_EQUAL) reduce 563) ((T_XOR_EQUAL) reduce 563) (("[") reduce 563) (("^") reduce 563) (("|") reduce 563)) 11186 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) ((":") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_DOUBLE_ARROW) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_PIPE) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("^") reduce 484) (("|") reduce 484)) 11187 ((("%") reduce 466) (("*") reduce 466) (("+") reduce 466) (("-") reduce 466) ((".") reduce 466) (("/") reduce 466) ((":") reduce 466) (("<") reduce 466) ((">") reduce 466) (("?") reduce 466) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_BOOLEAN_AND) reduce 466) ((T_BOOLEAN_OR) reduce 466) ((T_COALESCE) reduce 466) ((T_DOUBLE_ARROW) reduce 466) ((T_INSTANCEOF) reduce 466) ((T_IS_EQUAL) reduce 466) ((T_IS_GREATER_OR_EQUAL) reduce 466) ((T_IS_IDENTICAL) reduce 466) ((T_IS_NOT_EQUAL) reduce 466) ((T_IS_NOT_IDENTICAL) reduce 466) ((T_IS_SMALLER_OR_EQUAL) reduce 466) ((T_LOGICAL_AND) reduce 466) ((T_LOGICAL_OR) reduce 466) ((T_LOGICAL_XOR) reduce 466) ((T_PIPE) reduce 466) ((T_POW) reduce 466) ((T_SL) reduce 466) ((T_SPACESHIP) reduce 466) ((T_SR) reduce 466) (("^") reduce 466) (("|") reduce 466)) 11188 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) ((":") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_PIPE) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("^") reduce 488) (("|") reduce 488)) 11189 ((("%") reduce 531) (("(") shift) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) ((":") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("|") reduce 531)) 11190 ((("%") reduce 461) (("(") reduce 558) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) ((":") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PIPE) reduce 461) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") reduce 552) (("^") reduce 461) (("|") reduce 461)) 11191 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) ((":") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_DOUBLE_ARROW) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_PIPE) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 11192 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) ((":") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_DOUBLE_ARROW) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_PIPE) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("|") reduce 476)) 11193 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) ((":") reduce 559) (("<") reduce 559) (("=") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AND_EQUAL) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_COALESCE_EQUAL) reduce 559) ((T_CONCAT_EQUAL) reduce 559) ((T_DEC) reduce 559) ((T_DIV_EQUAL) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INC) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_MINUS_EQUAL) reduce 559) ((T_MOD_EQUAL) reduce 559) ((T_MUL_EQUAL) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_OR_EQUAL) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_PLUS_EQUAL) reduce 559) ((T_POW) reduce 559) ((T_POW_EQUAL) reduce 559) ((T_SL) reduce 559) ((T_SL_EQUAL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) ((T_SR_EQUAL) reduce 559) ((T_XOR_EQUAL) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 11194 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((":") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 11195 ((("%") reduce 403) (("*") reduce 403) (("+") reduce 403) (("-") reduce 403) ((".") reduce 403) (("/") reduce 403) ((":") reduce 403) (("<") reduce 403) (("=") shift) ((">") reduce 403) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_DOUBLE_ARROW) reduce 403) ((T_INC) shift) ((T_INSTANCEOF) reduce 403) ((T_IS_EQUAL) reduce 403) ((T_IS_GREATER_OR_EQUAL) reduce 403) ((T_IS_IDENTICAL) reduce 403) ((T_IS_NOT_EQUAL) reduce 403) ((T_IS_NOT_IDENTICAL) reduce 403) ((T_IS_SMALLER_OR_EQUAL) reduce 403) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 403) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 403) ((T_POW_EQUAL) shift) ((T_SL) reduce 403) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 403) ((T_SR) reduce 403) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 548) (("^") reduce 403) (("|") reduce 403)) 11198 ((("%") reduce 448) (("*") reduce 448) (("+") reduce 448) (("-") reduce 448) ((".") reduce 448) (("/") reduce 448) ((":") reduce 448) (("<") reduce 448) ((">") reduce 448) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_BOOLEAN_AND) reduce 448) ((T_BOOLEAN_OR) reduce 448) ((T_COALESCE) reduce 448) ((T_DOUBLE_ARROW) reduce 448) ((T_INSTANCEOF) reduce 448) ((T_IS_EQUAL) reduce 448) ((T_IS_GREATER_OR_EQUAL) reduce 448) ((T_IS_IDENTICAL) reduce 448) ((T_IS_NOT_EQUAL) reduce 448) ((T_IS_NOT_IDENTICAL) reduce 448) ((T_IS_SMALLER_OR_EQUAL) reduce 448) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_PIPE) reduce 448) ((T_POW) shift) ((T_SL) reduce 448) ((T_SPACESHIP) reduce 448) ((T_SR) reduce 448) (("^") reduce 448) (("|") reduce 448)) 11230 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 432) (("<") shift) ((">") shift) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_DOUBLE_ARROW) reduce 432) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 432)) 11231 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 435) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_DOUBLE_ARROW) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("|") reduce 435)) 11232 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 444) (("/") shift) ((":") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_PIPE) reduce 444) ((T_POW) shift) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("|") reduce 444)) 11233 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 458) (("<") shift) ((">") shift) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 458) (("|") reduce 458)) 11234 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 443) (("/") shift) ((":") reduce 443) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_DOUBLE_ARROW) reduce 443) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_PIPE) reduce 443) ((T_POW) shift) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("^") reduce 443) (("|") reduce 443)) 11235 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) ((":") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_DOUBLE_ARROW) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_PIPE) reduce 440) ((T_POW) shift) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("|") reduce 440)) 11236 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 453) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_DOUBLE_ARROW) reduce 453) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_PIPE) reduce 453) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 453) ((T_SR) shift) (("^") reduce 453) (("|") reduce 453)) 11237 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 431) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11238 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 429) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 429) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11239 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 430) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 430) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11240 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 455) (("<") shift) ((">") shift) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 455) ((T_SR) shift) (("^") reduce 455) (("|") reduce 455)) 11241 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 450) (("<") shift) ((">") shift) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_DOUBLE_ARROW) reduce 450) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 450) (("|") reduce 450)) 11242 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 452) (("<") shift) ((">") shift) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_DOUBLE_ARROW) reduce 452) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 452) (("|") reduce 452)) 11243 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 449) (("<") shift) ((">") shift) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_DOUBLE_ARROW) reduce 449) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 449) (("|") reduce 449)) 11244 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 457) (("<") shift) ((">") shift) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 457) ((T_SR) shift) (("^") reduce 457) (("|") reduce 457)) 11245 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 451) (("<") shift) ((">") shift) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_DOUBLE_ARROW) reduce 451) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 451) (("|") reduce 451)) 11248 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((":") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 11249 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((":") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 11250 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((":") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 11251 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((":") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 11252 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((":") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 11253 ((("%") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) ((":") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 11254 ((("%") reduce 511) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((":") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("|") reduce 511)) 11255 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) ((":") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PIPE) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 11256 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((":") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 11257 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((":") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("|") reduce 512)) 11258 ((("%") reduce 573) (("*") reduce 573) (("+") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) ((":") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_DOUBLE_ARROW) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("|") reduce 573)) 11264 ((("%") reduce 574) (("*") reduce 574) (("+") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) ((":") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_DOUBLE_ARROW) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("|") reduce 574)) 11265 ((("%") reduce 578) (("*") reduce 578) (("+") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) ((":") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_DOUBLE_ARROW) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("|") reduce 578)) 11266 ((("%") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) ((":") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 11267 ((("%") reduce 575) (("*") reduce 575) (("+") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) ((":") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_DOUBLE_ARROW) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("|") reduce 575)) 11268 ((("%") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) ((":") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 11271 ((("%") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) ((":") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 11272 ((("%") reduce 576) (("*") reduce 576) (("+") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) ((":") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_DOUBLE_ARROW) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("|") reduce 576)) 11274 ((("%") reduce 577) (("*") reduce 577) (("+") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) ((":") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_DOUBLE_ARROW) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("|") reduce 577)) 11276 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((":") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("|") reduce 513)) 11277 ((("%") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) ((":") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 11280 ((("%") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) ((":") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 11281 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 465) (("<") shift) ((">") shift) (("?") reduce 465) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 465) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11282 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 427) (("<") shift) ((">") shift) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_DOUBLE_ARROW) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11283 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 428) (("<") shift) ((">") shift) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_DOUBLE_ARROW) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11284 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 433) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_DOUBLE_ARROW) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("|") reduce 433)) 11285 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 434) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_DOUBLE_ARROW) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("|") reduce 434)) 11289 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 463) (("<") shift) ((">") shift) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 463) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11290 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 464) (("<") shift) ((">") shift) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 464) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11291 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 456) (("<") shift) ((">") shift) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 456) ((T_SR) shift) (("^") reduce 456) (("|") reduce 456)) 11292 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 454) (("<") shift) ((">") shift) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 454) ((T_SR) shift) (("^") reduce 454) (("|") reduce 454)) 11293 ((("%") reduce 441) (("*") reduce 441) (("+") reduce 441) (("-") reduce 441) ((".") reduce 441) (("/") reduce 441) ((":") reduce 441) (("<") reduce 441) ((">") reduce 441) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_DOUBLE_ARROW) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 441) ((T_IS_GREATER_OR_EQUAL) reduce 441) ((T_IS_IDENTICAL) reduce 441) ((T_IS_NOT_EQUAL) reduce 441) ((T_IS_NOT_IDENTICAL) reduce 441) ((T_IS_SMALLER_OR_EQUAL) reduce 441) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_PIPE) reduce 441) ((T_POW) shift) ((T_SL) reduce 441) ((T_SPACESHIP) reduce 441) ((T_SR) reduce 441) (("^") reduce 441) (("|") reduce 441)) 11294 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 436) (("/") shift) ((":") reduce 436) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_DOUBLE_ARROW) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_PIPE) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 436) ((T_SR) shift) (("^") reduce 436) (("|") reduce 436)) 11295 ((("%") shift) (("*") shift) (("+") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") shift) ((":") reduce 438) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_DOUBLE_ARROW) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_PIPE) reduce 438) ((T_POW) shift) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("|") reduce 438)) 11296 ((("%") shift) (("*") shift) (("+") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") shift) ((":") reduce 437) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_DOUBLE_ARROW) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_PIPE) reduce 437) ((T_POW) shift) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("|") reduce 437)) 11297 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) ((":") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_DOUBLE_ARROW) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_PIPE) reduce 439) ((T_POW) shift) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("|") reduce 439)) 11298 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) ((":") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_DOUBLE_ARROW) reduce 442) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_PIPE) reduce 442) ((T_POW) shift) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("|") reduce 442)) 11300 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) ((":") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_DOUBLE_ARROW) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_PIPE) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("|") reduce 74)) 11301 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) ((":") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_DOUBLE_ARROW) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_PIPE) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("|") reduce 52)) 11302 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) ((":") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_DOUBLE_ARROW) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_PIPE) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("|") reduce 27)) 11303 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) ((":") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_DOUBLE_ARROW) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_PIPE) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("|") reduce 51)) 11304 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) ((":") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_DOUBLE_ARROW) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_PIPE) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("|") reduce 53)) 11305 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) ((":") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_DOUBLE_ARROW) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_PIPE) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("|") reduce 49)) 11306 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) ((":") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_DOUBLE_ARROW) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_PIPE) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("|") reduce 29)) 11307 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) ((":") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_DOUBLE_ARROW) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_PIPE) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("|") reduce 59)) 11308 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) ((":") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_DOUBLE_ARROW) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_PIPE) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("|") reduce 60)) 11309 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) ((":") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_DOUBLE_ARROW) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_PIPE) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("|") reduce 11)) 11310 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) ((":") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_DOUBLE_ARROW) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_PIPE) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("|") reduce 42)) 11311 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) ((":") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_DOUBLE_ARROW) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_PIPE) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("|") reduce 39)) 11312 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) ((":") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_DOUBLE_ARROW) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_PIPE) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("|") reduce 25)) 11313 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) ((":") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_DOUBLE_ARROW) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_PIPE) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("|") reduce 50)) 11314 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) ((":") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_DOUBLE_ARROW) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_PIPE) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("|") reduce 66)) 11315 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) ((":") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_DOUBLE_ARROW) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_PIPE) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("|") reduce 18)) 11316 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) ((":") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_DOUBLE_ARROW) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_PIPE) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("|") reduce 17)) 11317 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) ((":") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_DOUBLE_ARROW) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_PIPE) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("|") reduce 15)) 11318 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) ((":") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_DOUBLE_ARROW) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_PIPE) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("|") reduce 14)) 11319 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) ((":") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_DOUBLE_ARROW) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_PIPE) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("|") reduce 38)) 11320 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) ((":") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_DOUBLE_ARROW) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_PIPE) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("|") reduce 26)) 11321 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) ((":") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_DOUBLE_ARROW) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_PIPE) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("|") reduce 22)) 11322 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) ((":") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_DOUBLE_ARROW) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_PIPE) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("|") reduce 24)) 11323 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) ((":") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_DOUBLE_ARROW) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_PIPE) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("|") reduce 16)) 11324 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) ((":") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_DOUBLE_ARROW) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_PIPE) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("|") reduce 48)) 11325 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) ((":") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_DOUBLE_ARROW) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_PIPE) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("|") reduce 20)) 11326 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) ((":") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_DOUBLE_ARROW) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_PIPE) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("|") reduce 70)) 11327 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) ((":") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_DOUBLE_ARROW) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_PIPE) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("|") reduce 3)) 11328 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) ((":") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_DOUBLE_ARROW) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_PIPE) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("|") reduce 12)) 11329 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) ((":") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_DOUBLE_ARROW) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_PIPE) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("|") reduce 54)) 11330 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) ((":") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_DOUBLE_ARROW) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_PIPE) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("|") reduce 65)) 11331 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) ((":") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_DOUBLE_ARROW) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_PIPE) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("|") reduce 75)) 11332 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) ((":") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_DOUBLE_ARROW) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_PIPE) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("|") reduce 30)) 11333 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) ((":") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_DOUBLE_ARROW) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_PIPE) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("|") reduce 68)) 11334 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) ((":") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_DOUBLE_ARROW) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_PIPE) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("|") reduce 21)) 11335 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) ((":") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_DOUBLE_ARROW) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_PIPE) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("|") reduce 23)) 11336 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) ((":") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_DOUBLE_ARROW) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_PIPE) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("|") reduce 41)) 11337 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) ((":") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_DOUBLE_ARROW) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_PIPE) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("|") reduce 62)) 11338 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) ((":") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_DOUBLE_ARROW) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_PIPE) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("|") reduce 34)) 11339 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) ((":") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_DOUBLE_ARROW) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_PIPE) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("|") reduce 40)) 11340 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) ((":") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_DOUBLE_ARROW) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_PIPE) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("|") reduce 13)) 11341 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) ((":") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_DOUBLE_ARROW) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_PIPE) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("|") reduce 55)) 11342 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) ((":") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_DOUBLE_ARROW) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_PIPE) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("|") reduce 1)) 11343 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) ((":") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_DOUBLE_ARROW) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_PIPE) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("|") reduce 2)) 11344 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) ((":") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_DOUBLE_ARROW) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_PIPE) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("|") reduce 9)) 11345 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) ((":") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_DOUBLE_ARROW) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_PIPE) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("|") reduce 33)) 11346 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) ((":") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_DOUBLE_ARROW) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_PIPE) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("|") reduce 58)) 11347 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) ((":") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_DOUBLE_ARROW) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_PIPE) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("|") reduce 37)) 11348 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) ((":") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_DOUBLE_ARROW) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_PIPE) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("|") reduce 64)) 11349 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) ((":") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_DOUBLE_ARROW) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_PIPE) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("|") reduce 46)) 11350 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) ((":") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_DOUBLE_ARROW) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_PIPE) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("|") reduce 8)) 11351 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) ((":") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_DOUBLE_ARROW) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_PIPE) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("|") reduce 6)) 11352 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) ((":") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_DOUBLE_ARROW) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_PIPE) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("|") reduce 7)) 11353 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) ((":") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_DOUBLE_ARROW) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_PIPE) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("|") reduce 69)) 11354 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) ((":") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_DOUBLE_ARROW) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_PIPE) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("|") reduce 63)) 11355 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) ((":") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_DOUBLE_ARROW) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_PIPE) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("|") reduce 56)) 11356 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) ((":") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_DOUBLE_ARROW) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_PIPE) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("|") reduce 10)) 11357 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) ((":") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_DOUBLE_ARROW) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_PIPE) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("|") reduce 67)) 11358 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) ((":") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_DOUBLE_ARROW) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_PIPE) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("|") reduce 44)) 11359 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) ((":") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_DOUBLE_ARROW) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_PIPE) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("|") reduce 76)) 11360 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) ((":") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_DOUBLE_ARROW) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_PIPE) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("|") reduce 71)) 11361 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) ((":") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_DOUBLE_ARROW) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_PIPE) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("|") reduce 77)) 11362 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) ((":") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_DOUBLE_ARROW) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_PIPE) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("|") reduce 78)) 11363 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) ((":") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_DOUBLE_ARROW) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_PIPE) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("|") reduce 79)) 11364 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) ((":") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_DOUBLE_ARROW) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_PIPE) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("|") reduce 4)) 11365 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) ((":") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_DOUBLE_ARROW) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_PIPE) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("|") reduce 5)) 11366 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) ((":") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_DOUBLE_ARROW) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_PIPE) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("|") reduce 43)) 11367 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) ((":") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_DOUBLE_ARROW) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_PIPE) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("|") reduce 73)) 11368 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) ((":") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_DOUBLE_ARROW) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_PIPE) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("|") reduce 82)) 11369 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) ((":") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_DOUBLE_ARROW) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_PIPE) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("|") reduce 47)) 11370 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) ((":") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_DOUBLE_ARROW) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_PIPE) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("|") reduce 31)) 11371 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) ((":") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_DOUBLE_ARROW) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_PIPE) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("|") reduce 57)) 11372 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) ((":") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_DOUBLE_ARROW) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_PIPE) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("|") reduce 61)) 11373 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) ((":") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_DOUBLE_ARROW) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_PIPE) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("|") reduce 28)) 11374 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) ((":") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_DOUBLE_ARROW) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_PIPE) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("|") reduce 36)) 11375 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) ((":") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_DOUBLE_ARROW) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_PIPE) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("|") reduce 32)) 11376 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) ((":") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_DOUBLE_ARROW) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_PIPE) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("|") reduce 35)) 11377 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) ((":") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_DOUBLE_ARROW) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_PIPE) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("|") reduce 19)) 11378 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) ((":") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_DOUBLE_ARROW) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_PIPE) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("|") reduce 45)) 11379 ((("%") reduce 542) (("(") reduce 579) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) ((":") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PIPE) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 11381 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) ((":") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_DOUBLE_ARROW) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_PIPE) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("|") reduce 72)) 11382 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) ((":") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_DOUBLE_ARROW) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_PIPE) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("|") reduce 83)) 11383 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) ((":") reduce 572) (("<") reduce 572) (("=") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AND_EQUAL) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_COALESCE_EQUAL) reduce 572) ((T_CONCAT_EQUAL) reduce 572) ((T_DEC) reduce 572) ((T_DIV_EQUAL) reduce 572) ((T_DOUBLE_ARROW) reduce 572) ((T_INC) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_MINUS_EQUAL) reduce 572) ((T_MOD_EQUAL) reduce 572) ((T_MUL_EQUAL) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_OR_EQUAL) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_PLUS_EQUAL) reduce 572) ((T_POW) reduce 572) ((T_POW_EQUAL) reduce 572) ((T_SL) reduce 572) ((T_SL_EQUAL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) ((T_SR_EQUAL) reduce 572) ((T_XOR_EQUAL) reduce 572) (("[") reduce 572) (("^") reduce 572) (("|") reduce 572)) 11386 ((("%") reduce 544) (("(") reduce 580) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) ((":") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PIPE) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 11388 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((":") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("|") reduce 507)) 11389 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((":") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 11393 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) ((":") reduce 292) (("<") reduce 292) (("=") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AND_EQUAL) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_COALESCE_EQUAL) reduce 292) ((T_CONCAT_EQUAL) reduce 292) ((T_DEC) reduce 292) ((T_DIV_EQUAL) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INC) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_MINUS_EQUAL) reduce 292) ((T_MOD_EQUAL) reduce 292) ((T_MUL_EQUAL) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_OR_EQUAL) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_PLUS_EQUAL) reduce 292) ((T_POW) reduce 292) ((T_POW_EQUAL) reduce 292) ((T_SL) reduce 292) ((T_SL_EQUAL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) ((T_SR_EQUAL) reduce 292) ((T_XOR_EQUAL) reduce 292) (("[") reduce 292) (("^") reduce 292) (("|") reduce 292)) 11394 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) ((":") reduce 293) (("<") reduce 293) (("=") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AND_EQUAL) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_COALESCE_EQUAL) reduce 293) ((T_CONCAT_EQUAL) reduce 293) ((T_DEC) reduce 293) ((T_DIV_EQUAL) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INC) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_MINUS_EQUAL) reduce 293) ((T_MOD_EQUAL) reduce 293) ((T_MUL_EQUAL) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_OR_EQUAL) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_PLUS_EQUAL) reduce 293) ((T_POW) reduce 293) ((T_POW_EQUAL) reduce 293) ((T_SL) reduce 293) ((T_SL_EQUAL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) ((T_SR_EQUAL) reduce 293) ((T_XOR_EQUAL) reduce 293) (("[") reduce 293) (("^") reduce 293) (("|") reduce 293)) 11399 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) ((":") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_DOUBLE_ARROW) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_PIPE) reduce 425) ((T_POW) reduce 425) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 11401 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) ((":") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_DOUBLE_ARROW) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_PIPE) reduce 423) ((T_POW) reduce 423) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 11411 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 419) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11412 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 421) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11413 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 420) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11414 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 413) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 413) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11415 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 410) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 410) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11416 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 418) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11417 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 412) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11418 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 416) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11419 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 411) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11420 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 414) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 414) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11421 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 415) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 415) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11422 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 422) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11423 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 417) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11425 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 406) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11428 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) ((":") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 11431 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) ((":") reduce 564) (("<") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_DOUBLE_ARROW) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_POW) reduce 564) ((T_SL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) (("[") reduce 564) (("^") reduce 564) (("|") reduce 564)) 11433 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) ((":") reduce 563) (("<") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_DOUBLE_ARROW) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_POW) reduce 563) ((T_SL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) (("[") reduce 563) (("^") reduce 563) (("|") reduce 563)) 11435 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) ((":") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 11436 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((":") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 11437 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) ((":") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_DOUBLE_ARROW) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("[") reduce 548) (("^") reduce 407) (("|") reduce 407)) 11441 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) ((":") reduce 572) (("<") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_DOUBLE_ARROW) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_POW) reduce 572) ((T_SL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) (("[") reduce 572) (("^") reduce 572) (("|") reduce 572)) 11443 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((":") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("|") reduce 507)) 11444 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((":") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 11448 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) ((":") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("[") reduce 292) (("^") reduce 292) (("|") reduce 292)) 11449 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) ((":") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("[") reduce 293) (("^") reduce 293) (("|") reduce 293)) 11450 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) ((":") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 11453 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) ((":") reduce 571) (("<") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_DOUBLE_ARROW) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_POW) reduce 571) ((T_SL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) (("[") reduce 571) (("^") reduce 571) (("|") reduce 571)) 11454 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((":") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("|") reduce 506)) 11455 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((":") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("|") reduce 508)) 11460 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) ((":") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 11461 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) ((":") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 11462 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((":") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 11463 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) ((":") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 11466 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) ((":") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 11467 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((":") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 11468 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((":") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 11469 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) ((":") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("[") reduce 562) (("^") reduce 562) (("|") reduce 562)) 11470 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((":") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 11471 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) ((":") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 11474 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) ((":") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 11475 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) ((":") reduce 504) (("<") reduce 504) (("=") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AND_EQUAL) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_COALESCE_EQUAL) reduce 504) ((T_CONCAT_EQUAL) reduce 504) ((T_DEC) reduce 504) ((T_DIV_EQUAL) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INC) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_MINUS_EQUAL) reduce 504) ((T_MOD_EQUAL) reduce 504) ((T_MUL_EQUAL) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_OR_EQUAL) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_PLUS_EQUAL) reduce 504) ((T_POW) reduce 504) ((T_POW_EQUAL) reduce 504) ((T_SL) reduce 504) ((T_SL_EQUAL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) ((T_SR_EQUAL) reduce 504) ((T_XOR_EQUAL) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 11486 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) ((":") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PIPE) reduce 494) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 11487 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) ((":") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_PIPE) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("^") reduce 489) (("|") reduce 489)) 11497 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 494) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11498 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) ((":") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_PIPE) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("^") reduce 490) (("|") reduce 490)) 11500 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 481) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11502 ((("%") reduce 541) (("(") reduce 579) (("*") reduce 541) (("+") reduce 541) (("-") reduce 541) ((".") reduce 541) (("/") reduce 541) ((":") reduce 541) (("<") reduce 541) ((">") reduce 541) (("?") reduce 541) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_BOOLEAN_AND) reduce 541) ((T_BOOLEAN_OR) reduce 541) ((T_COALESCE) reduce 541) ((T_DOUBLE_ARROW) reduce 541) ((T_INSTANCEOF) reduce 541) ((T_IS_EQUAL) reduce 541) ((T_IS_GREATER_OR_EQUAL) reduce 541) ((T_IS_IDENTICAL) reduce 541) ((T_IS_NOT_EQUAL) reduce 541) ((T_IS_NOT_IDENTICAL) reduce 541) ((T_IS_SMALLER_OR_EQUAL) reduce 541) ((T_LOGICAL_AND) reduce 541) ((T_LOGICAL_OR) reduce 541) ((T_LOGICAL_XOR) reduce 541) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 541) ((T_OBJECT_OPERATOR) reduce 541) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 541) ((T_PIPE) reduce 541) ((T_POW) reduce 541) ((T_SL) reduce 541) ((T_SPACESHIP) reduce 541) ((T_SR) reduce 541) (("[") reduce 541) (("^") reduce 541) (("|") reduce 541)) 11504 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) ((":") reduce 571) (("<") reduce 571) (("=") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AND_EQUAL) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_COALESCE_EQUAL) reduce 571) ((T_CONCAT_EQUAL) reduce 571) ((T_DEC) reduce 571) ((T_DIV_EQUAL) reduce 571) ((T_DOUBLE_ARROW) reduce 571) ((T_INC) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_MINUS_EQUAL) reduce 571) ((T_MOD_EQUAL) reduce 571) ((T_MUL_EQUAL) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_OR_EQUAL) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_PLUS_EQUAL) reduce 571) ((T_POW) reduce 571) ((T_POW_EQUAL) reduce 571) ((T_SL) reduce 571) ((T_SL_EQUAL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) ((T_SR_EQUAL) reduce 571) ((T_XOR_EQUAL) reduce 571) (("[") reduce 571) (("^") reduce 571) (("|") reduce 571)) 11507 ((("%") reduce 543) (("(") reduce 580) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) ((":") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PIPE) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 11508 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((":") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("|") reduce 506)) 11509 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((":") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("|") reduce 508)) 11511 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) ((":") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_DOUBLE_ARROW) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_PIPE) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("^") reduce 485) (("|") reduce 485)) 11512 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) ((":") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_PIPE) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("^") reduce 487) (("|") reduce 487)) 11517 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) ((":") reduce 560) (("<") reduce 560) (("=") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AND_EQUAL) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_COALESCE_EQUAL) reduce 560) ((T_CONCAT_EQUAL) reduce 560) ((T_DEC) reduce 560) ((T_DIV_EQUAL) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INC) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_MINUS_EQUAL) reduce 560) ((T_MOD_EQUAL) reduce 560) ((T_MUL_EQUAL) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_OR_EQUAL) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_PLUS_EQUAL) reduce 560) ((T_POW) reduce 560) ((T_POW_EQUAL) reduce 560) ((T_SL) reduce 560) ((T_SL_EQUAL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) ((T_SR_EQUAL) reduce 560) ((T_XOR_EQUAL) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 11518 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) ((":") reduce 582) (("<") reduce 582) (("=") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AND_EQUAL) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_COALESCE_EQUAL) reduce 582) ((T_CONCAT_EQUAL) reduce 582) ((T_DEC) reduce 582) ((T_DIV_EQUAL) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INC) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_MINUS_EQUAL) reduce 582) ((T_MOD_EQUAL) reduce 582) ((T_MUL_EQUAL) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_OR_EQUAL) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_PLUS_EQUAL) reduce 582) ((T_POW) reduce 582) ((T_POW_EQUAL) reduce 582) ((T_SL) reduce 582) ((T_SL_EQUAL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) ((T_SR_EQUAL) reduce 582) ((T_XOR_EQUAL) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 11519 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((":") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 11520 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) ((":") reduce 584) (("<") reduce 584) (("=") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AND_EQUAL) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_COALESCE_EQUAL) reduce 584) ((T_CONCAT_EQUAL) reduce 584) ((T_DEC) reduce 584) ((T_DIV_EQUAL) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INC) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_MINUS_EQUAL) reduce 584) ((T_MOD_EQUAL) reduce 584) ((T_MUL_EQUAL) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_OR_EQUAL) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_PLUS_EQUAL) reduce 584) ((T_POW) reduce 584) ((T_POW_EQUAL) reduce 584) ((T_SL) reduce 584) ((T_SL_EQUAL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) ((T_SR_EQUAL) reduce 584) ((T_XOR_EQUAL) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 11523 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) ((":") reduce 583) (("<") reduce 583) (("=") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AND_EQUAL) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_COALESCE_EQUAL) reduce 583) ((T_CONCAT_EQUAL) reduce 583) ((T_DEC) reduce 583) ((T_DIV_EQUAL) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INC) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_MINUS_EQUAL) reduce 583) ((T_MOD_EQUAL) reduce 583) ((T_MUL_EQUAL) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_OR_EQUAL) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_PLUS_EQUAL) reduce 583) ((T_POW) reduce 583) ((T_POW_EQUAL) reduce 583) ((T_SL) reduce 583) ((T_SL_EQUAL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) ((T_SR_EQUAL) reduce 583) ((T_XOR_EQUAL) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 11524 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((":") reduce 561) (("<") reduce 561) (("=") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AND_EQUAL) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_COALESCE_EQUAL) reduce 561) ((T_CONCAT_EQUAL) reduce 561) ((T_DEC) reduce 561) ((T_DIV_EQUAL) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INC) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_MINUS_EQUAL) reduce 561) ((T_MOD_EQUAL) reduce 561) ((T_MUL_EQUAL) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_OR_EQUAL) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_PLUS_EQUAL) reduce 561) ((T_POW) reduce 561) ((T_POW_EQUAL) reduce 561) ((T_SL) reduce 561) ((T_SL_EQUAL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) ((T_SR_EQUAL) reduce 561) ((T_XOR_EQUAL) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 11525 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((":") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 11526 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) ((":") reduce 562) (("<") reduce 562) (("=") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AND_EQUAL) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_COALESCE_EQUAL) reduce 562) ((T_CONCAT_EQUAL) reduce 562) ((T_DEC) reduce 562) ((T_DIV_EQUAL) reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_INC) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_MINUS_EQUAL) reduce 562) ((T_MOD_EQUAL) reduce 562) ((T_MUL_EQUAL) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_OR_EQUAL) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_PLUS_EQUAL) reduce 562) ((T_POW) reduce 562) ((T_POW_EQUAL) reduce 562) ((T_SL) reduce 562) ((T_SL_EQUAL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) ((T_SR_EQUAL) reduce 562) ((T_XOR_EQUAL) reduce 562) (("[") reduce 562) (("^") reduce 562) (("|") reduce 562)) 11528 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) ((":") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_PIPE) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 11530 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) ((":") reduce 520) (("<") reduce 520) (("=") shift) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_PIPE) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("|") reduce 520)) 11532 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 405) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11533 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 482) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 482) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11536 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 481) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 481) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11537 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) ((":") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_DOUBLE_ARROW) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_PIPE) reduce 473) ((T_POW) shift) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("|") reduce 473)) 11538 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 483) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 483) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11539 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) ((":") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_DOUBLE_ARROW) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_PIPE) reduce 469) ((T_POW) shift) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 11540 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) ((":") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_DOUBLE_ARROW) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_PIPE) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("^") reduce 486) (("|") reduce 486)) 11542 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) ((":") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_PIPE) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("^") reduce 526) (("|") reduce 526)) 11544 ((("%") reduce 527) (("*") reduce 527) (("+") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) ((":") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_DOUBLE_ARROW) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_PIPE) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("^") reduce 527) (("|") reduce 527)) 11545 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) ((":") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_PIPE) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("^") reduce 525) (("|") reduce 525)) 11546 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 619) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 619) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11547 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 618) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 618) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11548 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((":") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 11549 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 478) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 478) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11550 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) ((":") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_DOUBLE_ARROW) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_PIPE) reduce 471) ((T_POW) shift) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("|") reduce 471)) 11552 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((":") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 11553 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((":") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 11554 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((":") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 11555 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((":") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 11556 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((":") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 11557 ((("%") reduce 400) (("(") reduce 400) (("*") reduce 400) (("+") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) ((":") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_DOUBLE_ARROW) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 400) ((T_OBJECT_OPERATOR) reduce 400) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 400) ((T_PIPE) reduce 400) ((T_POW) reduce 400) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("[") reduce 400) (("^") reduce 400) (("|") reduce 400)) 11560 ((("%") reduce 511) (("(") reduce 511) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((":") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("|") reduce 511)) 11561 ((("%") reduce 402) (("(") shift) (("*") reduce 402) (("+") reduce 402) (("-") reduce 402) ((".") reduce 402) (("/") reduce 402) ((":") reduce 402) (("<") reduce 402) ((">") reduce 402) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_DOUBLE_ARROW) reduce 402) ((T_INSTANCEOF) reduce 402) ((T_IS_EQUAL) reduce 402) ((T_IS_GREATER_OR_EQUAL) reduce 402) ((T_IS_IDENTICAL) reduce 402) ((T_IS_NOT_EQUAL) reduce 402) ((T_IS_NOT_IDENTICAL) reduce 402) ((T_IS_SMALLER_OR_EQUAL) reduce 402) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_PIPE) reduce 402) ((T_POW) reduce 402) ((T_SL) reduce 402) ((T_SPACESHIP) reduce 402) ((T_SR) reduce 402) (("^") reduce 402) (("|") reduce 402)) 11562 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((":") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 11563 ((("%") reduce 512) (("(") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((":") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("|") reduce 512)) 11564 ((("%") reduce 573) (("(") reduce 573) (("*") reduce 573) (("+") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) ((":") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_DOUBLE_ARROW) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("|") reduce 573)) 11570 ((("%") reduce 574) (("(") reduce 574) (("*") reduce 574) (("+") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) ((":") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_DOUBLE_ARROW) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("|") reduce 574)) 11571 ((("%") reduce 578) (("(") reduce 578) (("*") reduce 578) (("+") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) ((":") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_DOUBLE_ARROW) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("|") reduce 578)) 11572 ((("%") reduce 575) (("(") reduce 575) (("*") reduce 575) (("+") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) ((":") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_DOUBLE_ARROW) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("|") reduce 575)) 11573 ((("%") reduce 576) (("(") reduce 576) (("*") reduce 576) (("+") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) ((":") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_DOUBLE_ARROW) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("|") reduce 576)) 11574 ((("%") reduce 399) (("(") reduce 399) (("*") reduce 399) (("+") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) ((":") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_DOUBLE_ARROW) reduce 399) ((T_INSTANCEOF) reduce 399) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 399) ((T_OBJECT_OPERATOR) reduce 399) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 399) ((T_PIPE) reduce 399) ((T_POW) reduce 399) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("[") reduce 399) (("^") reduce 399) (("|") reduce 399)) 11576 ((("%") reduce 577) (("(") reduce 577) (("*") reduce 577) (("+") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) ((":") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_DOUBLE_ARROW) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("|") reduce 577)) 11577 ((("%") reduce 401) (("(") reduce 401) (("*") reduce 401) (("+") reduce 401) (("-") reduce 401) ((".") reduce 401) (("/") reduce 401) ((":") reduce 401) (("<") reduce 401) ((">") reduce 401) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_DOUBLE_ARROW) reduce 401) ((T_INSTANCEOF) reduce 401) ((T_IS_EQUAL) reduce 401) ((T_IS_GREATER_OR_EQUAL) reduce 401) ((T_IS_IDENTICAL) reduce 401) ((T_IS_NOT_EQUAL) reduce 401) ((T_IS_NOT_IDENTICAL) reduce 401) ((T_IS_SMALLER_OR_EQUAL) reduce 401) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 401) ((T_OBJECT_OPERATOR) reduce 401) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 401) ((T_PIPE) reduce 401) ((T_POW) reduce 401) ((T_SL) reduce 401) ((T_SPACESHIP) reduce 401) ((T_SR) reduce 401) (("[") reduce 401) (("^") reduce 401) (("|") reduce 401)) 11585 ((("%") reduce 398) (("(") reduce 398) (("*") reduce 398) (("+") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) ((":") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_DOUBLE_ARROW) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 398) ((T_OBJECT_OPERATOR) reduce 398) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 398) ((T_PIPE) reduce 398) ((T_POW) reduce 398) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("[") reduce 398) (("^") reduce 398) (("|") reduce 398)) 11587 ((("%") reduce 513) (("(") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((":") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("|") reduce 513)) 11593 ((("%") reduce 233) (("*") reduce 233) (("+") reduce 233) (("-") reduce 233) ((".") reduce 233) (("/") reduce 233) ((":") reduce 233) (("<") reduce 233) ((">") reduce 233) (("?") reduce 233) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_BOOLEAN_AND) reduce 233) ((T_BOOLEAN_OR) reduce 233) ((T_COALESCE) reduce 233) ((T_DOUBLE_ARROW) reduce 233) ((T_INSTANCEOF) reduce 233) ((T_IS_EQUAL) reduce 233) ((T_IS_GREATER_OR_EQUAL) reduce 233) ((T_IS_IDENTICAL) reduce 233) ((T_IS_NOT_EQUAL) reduce 233) ((T_IS_NOT_IDENTICAL) reduce 233) ((T_IS_SMALLER_OR_EQUAL) reduce 233) ((T_LOGICAL_AND) reduce 233) ((T_LOGICAL_OR) reduce 233) ((T_LOGICAL_XOR) reduce 233) ((T_PIPE) reduce 233) ((T_POW) reduce 233) ((T_SL) reduce 233) ((T_SPACESHIP) reduce 233) ((T_SR) reduce 233) (("^") reduce 233) (("|") reduce 233)) 11598 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 404) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11602 ((("%") reduce 613) (("*") reduce 613) (("+") reduce 613) (("-") reduce 613) ((".") reduce 613) (("/") reduce 613) ((":") reduce 613) (("<") reduce 613) ((">") reduce 613) (("?") reduce 613) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_BOOLEAN_AND) reduce 613) ((T_BOOLEAN_OR) reduce 613) ((T_COALESCE) reduce 613) ((T_DOUBLE_ARROW) reduce 613) ((T_INSTANCEOF) reduce 613) ((T_IS_EQUAL) reduce 613) ((T_IS_GREATER_OR_EQUAL) reduce 613) ((T_IS_IDENTICAL) reduce 613) ((T_IS_NOT_EQUAL) reduce 613) ((T_IS_NOT_IDENTICAL) reduce 613) ((T_IS_SMALLER_OR_EQUAL) reduce 613) ((T_LOGICAL_AND) reduce 613) ((T_LOGICAL_OR) reduce 613) ((T_LOGICAL_XOR) reduce 613) ((T_PIPE) reduce 613) ((T_POW) reduce 613) ((T_SL) reduce 613) ((T_SPACESHIP) reduce 613) ((T_SR) reduce 613) (("^") reduce 613) (("|") reduce 613)) 11603 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) ((":") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_DOUBLE_ARROW) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_PIPE) reduce 467) ((T_POW) shift) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("|") reduce 467)) 11604 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 616) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 616) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11605 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 615) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 615) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11606 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) ((":") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_DOUBLE_ARROW) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("[") reduce 548) (("^") reduce 424) (("|") reduce 424)) 11608 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) ((":") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_PIPE) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("^") reduce 518) (("|") reduce 518)) 11609 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) ((":") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_DOUBLE_ARROW) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_PIPE) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("|") reduce 474)) 11610 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((":") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("|") reduce 291)) 11614 ((("%") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) ((":") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("^") reduce 292) (("|") reduce 292)) 11615 ((("%") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) ((":") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("^") reduce 293) (("|") reduce 293)) 11618 ((("%") reduce 617) (("*") reduce 617) (("+") reduce 617) (("-") reduce 617) ((".") reduce 617) (("/") reduce 617) ((":") reduce 617) (("<") reduce 617) ((">") reduce 617) (("?") reduce 617) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_BOOLEAN_AND) reduce 617) ((T_BOOLEAN_OR) reduce 617) ((T_COALESCE) reduce 617) ((T_DOUBLE_ARROW) reduce 617) ((T_INSTANCEOF) reduce 617) ((T_IS_EQUAL) reduce 617) ((T_IS_GREATER_OR_EQUAL) reduce 617) ((T_IS_IDENTICAL) reduce 617) ((T_IS_NOT_EQUAL) reduce 617) ((T_IS_NOT_IDENTICAL) reduce 617) ((T_IS_SMALLER_OR_EQUAL) reduce 617) ((T_LOGICAL_AND) reduce 617) ((T_LOGICAL_OR) reduce 617) ((T_LOGICAL_XOR) reduce 617) ((T_PIPE) reduce 617) ((T_POW) reduce 617) ((T_SL) reduce 617) ((T_SPACESHIP) reduce 617) ((T_SR) reduce 617) (("^") reduce 617) (("|") reduce 617)) 11621 ((("%") reduce 614) (("*") reduce 614) (("+") reduce 614) (("-") reduce 614) ((".") reduce 614) (("/") reduce 614) ((":") reduce 614) (("<") reduce 614) ((">") reduce 614) (("?") reduce 614) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_BOOLEAN_AND) reduce 614) ((T_BOOLEAN_OR) reduce 614) ((T_COALESCE) reduce 614) ((T_DOUBLE_ARROW) reduce 614) ((T_INSTANCEOF) reduce 614) ((T_IS_EQUAL) reduce 614) ((T_IS_GREATER_OR_EQUAL) reduce 614) ((T_IS_IDENTICAL) reduce 614) ((T_IS_NOT_EQUAL) reduce 614) ((T_IS_NOT_IDENTICAL) reduce 614) ((T_IS_SMALLER_OR_EQUAL) reduce 614) ((T_LOGICAL_AND) reduce 614) ((T_LOGICAL_OR) reduce 614) ((T_LOGICAL_XOR) reduce 614) ((T_PIPE) reduce 614) ((T_POW) reduce 614) ((T_SL) reduce 614) ((T_SPACESHIP) reduce 614) ((T_SR) reduce 614) (("^") reduce 614) (("|") reduce 614)) 11622 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) ((":") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_DOUBLE_ARROW) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_PIPE) reduce 468) ((T_POW) shift) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 11623 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) ((":") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_DOUBLE_ARROW) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("[") reduce 548) (("^") reduce 426) (("|") reduce 426)) 11625 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) ((":") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_DOUBLE_ARROW) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_PIPE) reduce 408) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("^") reduce 408) (("|") reduce 408)) 11626 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) ((":") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_DOUBLE_ARROW) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_PIPE) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("|") reduce 409)) 11627 ((("%") reduce 296) (("*") reduce 296) (("+") reduce 296) (("-") reduce 296) ((".") reduce 296) (("/") reduce 296) ((":") reduce 296) (("<") reduce 296) ((">") reduce 296) (("?") reduce 296) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_BOOLEAN_AND) reduce 296) ((T_BOOLEAN_OR) reduce 296) ((T_COALESCE) reduce 296) ((T_DOUBLE_ARROW) reduce 296) ((T_INSTANCEOF) reduce 296) ((T_IS_EQUAL) reduce 296) ((T_IS_GREATER_OR_EQUAL) reduce 296) ((T_IS_IDENTICAL) reduce 296) ((T_IS_NOT_EQUAL) reduce 296) ((T_IS_NOT_IDENTICAL) reduce 296) ((T_IS_SMALLER_OR_EQUAL) reduce 296) ((T_LOGICAL_AND) reduce 296) ((T_LOGICAL_OR) reduce 296) ((T_LOGICAL_XOR) reduce 296) ((T_PIPE) reduce 296) ((T_POW) reduce 296) ((T_SL) reduce 296) ((T_SPACESHIP) reduce 296) ((T_SR) reduce 296) (("^") reduce 296) (("|") reduce 296)) 11632 ((("%") reduce 297) (("*") reduce 297) (("+") reduce 297) (("-") reduce 297) ((".") reduce 297) (("/") reduce 297) ((":") reduce 297) (("<") reduce 297) ((">") reduce 297) (("?") reduce 297) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_BOOLEAN_AND) reduce 297) ((T_BOOLEAN_OR) reduce 297) ((T_COALESCE) reduce 297) ((T_DOUBLE_ARROW) reduce 297) ((T_INSTANCEOF) reduce 297) ((T_IS_EQUAL) reduce 297) ((T_IS_GREATER_OR_EQUAL) reduce 297) ((T_IS_IDENTICAL) reduce 297) ((T_IS_NOT_EQUAL) reduce 297) ((T_IS_NOT_IDENTICAL) reduce 297) ((T_IS_SMALLER_OR_EQUAL) reduce 297) ((T_LOGICAL_AND) reduce 297) ((T_LOGICAL_OR) reduce 297) ((T_LOGICAL_XOR) reduce 297) ((T_PIPE) reduce 297) ((T_POW) reduce 297) ((T_SL) reduce 297) ((T_SPACESHIP) reduce 297) ((T_SR) reduce 297) (("^") reduce 297) (("|") reduce 297)) 11633 ((("%") reduce 460) (("(") reduce 556) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) ((":") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PIPE) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("[") reduce 549) (("^") reduce 460) (("|") reduce 460)) 11635 ((("%") reduce 298) (("*") reduce 298) (("+") reduce 298) (("-") reduce 298) ((".") reduce 298) (("/") reduce 298) ((":") reduce 298) (("<") reduce 298) ((">") reduce 298) (("?") reduce 298) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_BOOLEAN_AND) reduce 298) ((T_BOOLEAN_OR) reduce 298) ((T_COALESCE) reduce 298) ((T_DOUBLE_ARROW) reduce 298) ((T_INSTANCEOF) reduce 298) ((T_IS_EQUAL) reduce 298) ((T_IS_GREATER_OR_EQUAL) reduce 298) ((T_IS_IDENTICAL) reduce 298) ((T_IS_NOT_EQUAL) reduce 298) ((T_IS_NOT_IDENTICAL) reduce 298) ((T_IS_SMALLER_OR_EQUAL) reduce 298) ((T_LOGICAL_AND) reduce 298) ((T_LOGICAL_OR) reduce 298) ((T_LOGICAL_XOR) reduce 298) ((T_PIPE) reduce 298) ((T_POW) reduce 298) ((T_SL) reduce 298) ((T_SPACESHIP) reduce 298) ((T_SR) reduce 298) (("^") reduce 298) (("|") reduce 298)) 11636 ((("%") reduce 299) (("*") reduce 299) (("+") reduce 299) (("-") reduce 299) ((".") reduce 299) (("/") reduce 299) ((":") reduce 299) (("<") reduce 299) ((">") reduce 299) (("?") reduce 299) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_BOOLEAN_AND) reduce 299) ((T_BOOLEAN_OR) reduce 299) ((T_COALESCE) reduce 299) ((T_DOUBLE_ARROW) reduce 299) ((T_INSTANCEOF) reduce 299) ((T_IS_EQUAL) reduce 299) ((T_IS_GREATER_OR_EQUAL) reduce 299) ((T_IS_IDENTICAL) reduce 299) ((T_IS_NOT_EQUAL) reduce 299) ((T_IS_NOT_IDENTICAL) reduce 299) ((T_IS_SMALLER_OR_EQUAL) reduce 299) ((T_LOGICAL_AND) reduce 299) ((T_LOGICAL_OR) reduce 299) ((T_LOGICAL_XOR) reduce 299) ((T_PIPE) reduce 299) ((T_POW) reduce 299) ((T_SL) reduce 299) ((T_SPACESHIP) reduce 299) ((T_SR) reduce 299) (("^") reduce 299) (("|") reduce 299)) 11637 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) ((":") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_DOUBLE_ARROW) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_PIPE) reduce 472) ((T_POW) shift) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("|") reduce 472)) 11638 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) ((":") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_DOUBLE_ARROW) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_PIPE) reduce 470) ((T_POW) shift) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("|") reduce 470)) 11641 ((("%") reduce 519) (("(") reduce 519) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) ((":") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PIPE) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("|") reduce 519)) 11642 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) ((":") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_DOUBLE_ARROW) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_PIPE) reduce 475) ((T_POW) shift) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("|") reduce 475)) 11643 ((("%") reduce 446) (("*") reduce 446) (("+") reduce 446) (("-") reduce 446) ((".") reduce 446) (("/") reduce 446) ((":") reduce 446) (("<") reduce 446) ((">") reduce 446) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_BOOLEAN_AND) reduce 446) ((T_BOOLEAN_OR) reduce 446) ((T_COALESCE) reduce 446) ((T_DOUBLE_ARROW) reduce 446) ((T_INSTANCEOF) reduce 446) ((T_IS_EQUAL) reduce 446) ((T_IS_GREATER_OR_EQUAL) reduce 446) ((T_IS_IDENTICAL) reduce 446) ((T_IS_NOT_EQUAL) reduce 446) ((T_IS_NOT_IDENTICAL) reduce 446) ((T_IS_SMALLER_OR_EQUAL) reduce 446) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_PIPE) reduce 446) ((T_POW) shift) ((T_SL) reduce 446) ((T_SPACESHIP) reduce 446) ((T_SR) reduce 446) (("^") reduce 446) (("|") reduce 446)) 11644 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) ((":") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_DOUBLE_ARROW) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_PIPE) reduce 445) ((T_POW) shift) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("|") reduce 445)) 11646 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) ((":") reduce 570) (("<") reduce 570) (("=") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AND_EQUAL) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_COALESCE_EQUAL) reduce 570) ((T_CONCAT_EQUAL) reduce 570) ((T_DEC) reduce 570) ((T_DIV_EQUAL) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INC) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_MINUS_EQUAL) reduce 570) ((T_MOD_EQUAL) reduce 570) ((T_MUL_EQUAL) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_OR_EQUAL) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_PLUS_EQUAL) reduce 570) ((T_POW) reduce 570) ((T_POW_EQUAL) reduce 570) ((T_SL) reduce 570) ((T_SL_EQUAL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) ((T_SR_EQUAL) reduce 570) ((T_XOR_EQUAL) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 11649 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) ((":") reduce 569) (("<") reduce 569) (("=") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AND_EQUAL) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_COALESCE_EQUAL) reduce 569) ((T_CONCAT_EQUAL) reduce 569) ((T_DEC) reduce 569) ((T_DIV_EQUAL) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INC) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_MINUS_EQUAL) reduce 569) ((T_MOD_EQUAL) reduce 569) ((T_MUL_EQUAL) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_OR_EQUAL) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_PLUS_EQUAL) reduce 569) ((T_POW) reduce 569) ((T_POW_EQUAL) reduce 569) ((T_SL) reduce 569) ((T_SL_EQUAL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) ((T_SR_EQUAL) reduce 569) ((T_XOR_EQUAL) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 11651 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) ((":") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_PIPE) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("|") reduce 522)) 11652 ((("%") reduce 447) (("*") reduce 447) (("+") reduce 447) (("-") reduce 447) ((".") reduce 447) (("/") reduce 447) ((":") reduce 447) (("<") reduce 447) ((">") reduce 447) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_BOOLEAN_AND) reduce 447) ((T_BOOLEAN_OR) reduce 447) ((T_COALESCE) reduce 447) ((T_DOUBLE_ARROW) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 447) ((T_IS_GREATER_OR_EQUAL) reduce 447) ((T_IS_IDENTICAL) reduce 447) ((T_IS_NOT_EQUAL) reduce 447) ((T_IS_NOT_IDENTICAL) reduce 447) ((T_IS_SMALLER_OR_EQUAL) reduce 447) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_PIPE) reduce 447) ((T_POW) shift) ((T_SL) reduce 447) ((T_SPACESHIP) reduce 447) ((T_SR) reduce 447) (("^") reduce 447) (("|") reduce 447)) 11653 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) ((":") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_PIPE) reduce 473) ((T_POW) shift) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("|") reduce 473)) 11654 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 483) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11655 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) ((":") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_PIPE) reduce 469) ((T_POW) shift) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 11656 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) ((":") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_PIPE) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("^") reduce 486) (("|") reduce 486)) 11658 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) ((":") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_PIPE) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("^") reduce 526) (("|") reduce 526)) 11660 ((("%") reduce 527) (("*") reduce 527) (("+") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) ((":") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_PIPE) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("^") reduce 527) (("|") reduce 527)) 11661 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) ((":") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_PIPE) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("^") reduce 525) (("|") reduce 525)) 11662 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 619) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11663 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 618) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11664 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((":") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 11665 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 478) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11666 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) ((":") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_PIPE) reduce 471) ((T_POW) shift) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("|") reduce 471)) 11668 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((":") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 11669 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((":") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 11670 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((":") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 11671 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((":") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 11672 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((":") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 11673 ((("%") reduce 400) (("(") reduce 400) (("*") reduce 400) (("+") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) ((":") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 400) ((T_OBJECT_OPERATOR) reduce 400) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 400) ((T_PIPE) reduce 400) ((T_POW) reduce 400) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("[") reduce 400) (("^") reduce 400) (("|") reduce 400)) 11676 ((("%") reduce 511) (("(") reduce 511) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((":") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("|") reduce 511)) 11677 ((("%") reduce 402) (("(") shift) (("*") reduce 402) (("+") reduce 402) (("-") reduce 402) ((".") reduce 402) (("/") reduce 402) ((":") reduce 402) (("<") reduce 402) ((">") reduce 402) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_INSTANCEOF) reduce 402) ((T_IS_EQUAL) reduce 402) ((T_IS_GREATER_OR_EQUAL) reduce 402) ((T_IS_IDENTICAL) reduce 402) ((T_IS_NOT_EQUAL) reduce 402) ((T_IS_NOT_IDENTICAL) reduce 402) ((T_IS_SMALLER_OR_EQUAL) reduce 402) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_PIPE) reduce 402) ((T_POW) reduce 402) ((T_SL) reduce 402) ((T_SPACESHIP) reduce 402) ((T_SR) reduce 402) (("^") reduce 402) (("|") reduce 402)) 11678 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((":") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 11679 ((("%") reduce 512) (("(") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((":") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("|") reduce 512)) 11680 ((("%") reduce 573) (("(") reduce 573) (("*") reduce 573) (("+") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) ((":") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("|") reduce 573)) 11686 ((("%") reduce 574) (("(") reduce 574) (("*") reduce 574) (("+") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) ((":") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("|") reduce 574)) 11687 ((("%") reduce 578) (("(") reduce 578) (("*") reduce 578) (("+") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) ((":") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("|") reduce 578)) 11688 ((("%") reduce 575) (("(") reduce 575) (("*") reduce 575) (("+") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) ((":") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("|") reduce 575)) 11689 ((("%") reduce 576) (("(") reduce 576) (("*") reduce 576) (("+") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) ((":") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("|") reduce 576)) 11690 ((("%") reduce 399) (("(") reduce 399) (("*") reduce 399) (("+") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) ((":") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_INSTANCEOF) reduce 399) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 399) ((T_OBJECT_OPERATOR) reduce 399) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 399) ((T_PIPE) reduce 399) ((T_POW) reduce 399) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("[") reduce 399) (("^") reduce 399) (("|") reduce 399)) 11692 ((("%") reduce 577) (("(") reduce 577) (("*") reduce 577) (("+") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) ((":") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("|") reduce 577)) 11693 ((("%") reduce 401) (("(") reduce 401) (("*") reduce 401) (("+") reduce 401) (("-") reduce 401) ((".") reduce 401) (("/") reduce 401) ((":") reduce 401) (("<") reduce 401) ((">") reduce 401) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_INSTANCEOF) reduce 401) ((T_IS_EQUAL) reduce 401) ((T_IS_GREATER_OR_EQUAL) reduce 401) ((T_IS_IDENTICAL) reduce 401) ((T_IS_NOT_EQUAL) reduce 401) ((T_IS_NOT_IDENTICAL) reduce 401) ((T_IS_SMALLER_OR_EQUAL) reduce 401) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 401) ((T_OBJECT_OPERATOR) reduce 401) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 401) ((T_PIPE) reduce 401) ((T_POW) reduce 401) ((T_SL) reduce 401) ((T_SPACESHIP) reduce 401) ((T_SR) reduce 401) (("[") reduce 401) (("^") reduce 401) (("|") reduce 401)) 11701 ((("%") reduce 398) (("(") reduce 398) (("*") reduce 398) (("+") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) ((":") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 398) ((T_OBJECT_OPERATOR) reduce 398) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 398) ((T_PIPE) reduce 398) ((T_POW) reduce 398) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("[") reduce 398) (("^") reduce 398) (("|") reduce 398)) 11703 ((("%") reduce 513) (("(") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((":") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("|") reduce 513)) 11709 ((("%") reduce 233) (("*") reduce 233) (("+") reduce 233) (("-") reduce 233) ((".") reduce 233) (("/") reduce 233) ((":") reduce 233) (("<") reduce 233) ((">") reduce 233) (("?") reduce 233) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_BOOLEAN_AND) reduce 233) ((T_BOOLEAN_OR) reduce 233) ((T_COALESCE) reduce 233) ((T_INSTANCEOF) reduce 233) ((T_IS_EQUAL) reduce 233) ((T_IS_GREATER_OR_EQUAL) reduce 233) ((T_IS_IDENTICAL) reduce 233) ((T_IS_NOT_EQUAL) reduce 233) ((T_IS_NOT_IDENTICAL) reduce 233) ((T_IS_SMALLER_OR_EQUAL) reduce 233) ((T_LOGICAL_AND) reduce 233) ((T_LOGICAL_OR) reduce 233) ((T_LOGICAL_XOR) reduce 233) ((T_PIPE) reduce 233) ((T_POW) reduce 233) ((T_SL) reduce 233) ((T_SPACESHIP) reduce 233) ((T_SR) reduce 233) (("^") reduce 233) (("|") reduce 233)) 11714 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 404) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11718 ((("%") reduce 613) (("*") reduce 613) (("+") reduce 613) (("-") reduce 613) ((".") reduce 613) (("/") reduce 613) ((":") reduce 613) (("<") reduce 613) ((">") reduce 613) (("?") reduce 613) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_BOOLEAN_AND) reduce 613) ((T_BOOLEAN_OR) reduce 613) ((T_COALESCE) reduce 613) ((T_INSTANCEOF) reduce 613) ((T_IS_EQUAL) reduce 613) ((T_IS_GREATER_OR_EQUAL) reduce 613) ((T_IS_IDENTICAL) reduce 613) ((T_IS_NOT_EQUAL) reduce 613) ((T_IS_NOT_IDENTICAL) reduce 613) ((T_IS_SMALLER_OR_EQUAL) reduce 613) ((T_LOGICAL_AND) reduce 613) ((T_LOGICAL_OR) reduce 613) ((T_LOGICAL_XOR) reduce 613) ((T_PIPE) reduce 613) ((T_POW) reduce 613) ((T_SL) reduce 613) ((T_SPACESHIP) reduce 613) ((T_SR) reduce 613) (("^") reduce 613) (("|") reduce 613)) 11719 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) ((":") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_PIPE) reduce 467) ((T_POW) shift) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("|") reduce 467)) 11720 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 616) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11721 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 615) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 11722 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) ((":") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("[") reduce 548) (("^") reduce 424) (("|") reduce 424)) 11724 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) ((":") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_PIPE) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("^") reduce 518) (("|") reduce 518)) 11725 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) ((":") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_PIPE) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("|") reduce 474)) 11726 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((":") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("|") reduce 291)) 11730 ((("%") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) ((":") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("^") reduce 292) (("|") reduce 292)) 11731 ((("%") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) ((":") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("^") reduce 293) (("|") reduce 293)) 11734 ((("%") reduce 617) (("*") reduce 617) (("+") reduce 617) (("-") reduce 617) ((".") reduce 617) (("/") reduce 617) ((":") reduce 617) (("<") reduce 617) ((">") reduce 617) (("?") reduce 617) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_BOOLEAN_AND) reduce 617) ((T_BOOLEAN_OR) reduce 617) ((T_COALESCE) reduce 617) ((T_INSTANCEOF) reduce 617) ((T_IS_EQUAL) reduce 617) ((T_IS_GREATER_OR_EQUAL) reduce 617) ((T_IS_IDENTICAL) reduce 617) ((T_IS_NOT_EQUAL) reduce 617) ((T_IS_NOT_IDENTICAL) reduce 617) ((T_IS_SMALLER_OR_EQUAL) reduce 617) ((T_LOGICAL_AND) reduce 617) ((T_LOGICAL_OR) reduce 617) ((T_LOGICAL_XOR) reduce 617) ((T_PIPE) reduce 617) ((T_POW) reduce 617) ((T_SL) reduce 617) ((T_SPACESHIP) reduce 617) ((T_SR) reduce 617) (("^") reduce 617) (("|") reduce 617)) 11737 ((("%") reduce 614) (("*") reduce 614) (("+") reduce 614) (("-") reduce 614) ((".") reduce 614) (("/") reduce 614) ((":") reduce 614) (("<") reduce 614) ((">") reduce 614) (("?") reduce 614) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_BOOLEAN_AND) reduce 614) ((T_BOOLEAN_OR) reduce 614) ((T_COALESCE) reduce 614) ((T_INSTANCEOF) reduce 614) ((T_IS_EQUAL) reduce 614) ((T_IS_GREATER_OR_EQUAL) reduce 614) ((T_IS_IDENTICAL) reduce 614) ((T_IS_NOT_EQUAL) reduce 614) ((T_IS_NOT_IDENTICAL) reduce 614) ((T_IS_SMALLER_OR_EQUAL) reduce 614) ((T_LOGICAL_AND) reduce 614) ((T_LOGICAL_OR) reduce 614) ((T_LOGICAL_XOR) reduce 614) ((T_PIPE) reduce 614) ((T_POW) reduce 614) ((T_SL) reduce 614) ((T_SPACESHIP) reduce 614) ((T_SR) reduce 614) (("^") reduce 614) (("|") reduce 614)) 11738 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) ((":") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_PIPE) reduce 468) ((T_POW) shift) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 11739 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) ((":") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("[") reduce 548) (("^") reduce 426) (("|") reduce 426)) 11741 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) ((":") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_PIPE) reduce 408) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("^") reduce 408) (("|") reduce 408)) 11742 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) ((":") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_PIPE) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("|") reduce 409)) 11743 ((("%") reduce 296) (("*") reduce 296) (("+") reduce 296) (("-") reduce 296) ((".") reduce 296) (("/") reduce 296) ((":") reduce 296) (("<") reduce 296) ((">") reduce 296) (("?") reduce 296) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_BOOLEAN_AND) reduce 296) ((T_BOOLEAN_OR) reduce 296) ((T_COALESCE) reduce 296) ((T_INSTANCEOF) reduce 296) ((T_IS_EQUAL) reduce 296) ((T_IS_GREATER_OR_EQUAL) reduce 296) ((T_IS_IDENTICAL) reduce 296) ((T_IS_NOT_EQUAL) reduce 296) ((T_IS_NOT_IDENTICAL) reduce 296) ((T_IS_SMALLER_OR_EQUAL) reduce 296) ((T_LOGICAL_AND) reduce 296) ((T_LOGICAL_OR) reduce 296) ((T_LOGICAL_XOR) reduce 296) ((T_PIPE) reduce 296) ((T_POW) reduce 296) ((T_SL) reduce 296) ((T_SPACESHIP) reduce 296) ((T_SR) reduce 296) (("^") reduce 296) (("|") reduce 296)) 11748 ((("%") reduce 297) (("*") reduce 297) (("+") reduce 297) (("-") reduce 297) ((".") reduce 297) (("/") reduce 297) ((":") reduce 297) (("<") reduce 297) ((">") reduce 297) (("?") reduce 297) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_BOOLEAN_AND) reduce 297) ((T_BOOLEAN_OR) reduce 297) ((T_COALESCE) reduce 297) ((T_INSTANCEOF) reduce 297) ((T_IS_EQUAL) reduce 297) ((T_IS_GREATER_OR_EQUAL) reduce 297) ((T_IS_IDENTICAL) reduce 297) ((T_IS_NOT_EQUAL) reduce 297) ((T_IS_NOT_IDENTICAL) reduce 297) ((T_IS_SMALLER_OR_EQUAL) reduce 297) ((T_LOGICAL_AND) reduce 297) ((T_LOGICAL_OR) reduce 297) ((T_LOGICAL_XOR) reduce 297) ((T_PIPE) reduce 297) ((T_POW) reduce 297) ((T_SL) reduce 297) ((T_SPACESHIP) reduce 297) ((T_SR) reduce 297) (("^") reduce 297) (("|") reduce 297)) 11749 ((("%") reduce 460) (("(") reduce 556) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) ((":") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PIPE) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("[") reduce 549) (("^") reduce 460) (("|") reduce 460)) 11751 ((("%") reduce 298) (("*") reduce 298) (("+") reduce 298) (("-") reduce 298) ((".") reduce 298) (("/") reduce 298) ((":") reduce 298) (("<") reduce 298) ((">") reduce 298) (("?") reduce 298) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_BOOLEAN_AND) reduce 298) ((T_BOOLEAN_OR) reduce 298) ((T_COALESCE) reduce 298) ((T_INSTANCEOF) reduce 298) ((T_IS_EQUAL) reduce 298) ((T_IS_GREATER_OR_EQUAL) reduce 298) ((T_IS_IDENTICAL) reduce 298) ((T_IS_NOT_EQUAL) reduce 298) ((T_IS_NOT_IDENTICAL) reduce 298) ((T_IS_SMALLER_OR_EQUAL) reduce 298) ((T_LOGICAL_AND) reduce 298) ((T_LOGICAL_OR) reduce 298) ((T_LOGICAL_XOR) reduce 298) ((T_PIPE) reduce 298) ((T_POW) reduce 298) ((T_SL) reduce 298) ((T_SPACESHIP) reduce 298) ((T_SR) reduce 298) (("^") reduce 298) (("|") reduce 298)) 11752 ((("%") reduce 299) (("*") reduce 299) (("+") reduce 299) (("-") reduce 299) ((".") reduce 299) (("/") reduce 299) ((":") reduce 299) (("<") reduce 299) ((">") reduce 299) (("?") reduce 299) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_BOOLEAN_AND) reduce 299) ((T_BOOLEAN_OR) reduce 299) ((T_COALESCE) reduce 299) ((T_INSTANCEOF) reduce 299) ((T_IS_EQUAL) reduce 299) ((T_IS_GREATER_OR_EQUAL) reduce 299) ((T_IS_IDENTICAL) reduce 299) ((T_IS_NOT_EQUAL) reduce 299) ((T_IS_NOT_IDENTICAL) reduce 299) ((T_IS_SMALLER_OR_EQUAL) reduce 299) ((T_LOGICAL_AND) reduce 299) ((T_LOGICAL_OR) reduce 299) ((T_LOGICAL_XOR) reduce 299) ((T_PIPE) reduce 299) ((T_POW) reduce 299) ((T_SL) reduce 299) ((T_SPACESHIP) reduce 299) ((T_SR) reduce 299) (("^") reduce 299) (("|") reduce 299)) 11753 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) ((":") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_PIPE) reduce 472) ((T_POW) shift) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("|") reduce 472)) 11754 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) ((":") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_PIPE) reduce 470) ((T_POW) shift) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("|") reduce 470)) 11757 ((("%") reduce 519) (("(") reduce 519) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) ((":") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PIPE) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("|") reduce 519)) 11758 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) ((":") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_PIPE) reduce 475) ((T_POW) shift) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("|") reduce 475)) 11759 ((("%") reduce 446) (("*") reduce 446) (("+") reduce 446) (("-") reduce 446) ((".") reduce 446) (("/") reduce 446) ((":") reduce 446) (("<") reduce 446) ((">") reduce 446) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_BOOLEAN_AND) reduce 446) ((T_BOOLEAN_OR) reduce 446) ((T_COALESCE) reduce 446) ((T_INSTANCEOF) reduce 446) ((T_IS_EQUAL) reduce 446) ((T_IS_GREATER_OR_EQUAL) reduce 446) ((T_IS_IDENTICAL) reduce 446) ((T_IS_NOT_EQUAL) reduce 446) ((T_IS_NOT_IDENTICAL) reduce 446) ((T_IS_SMALLER_OR_EQUAL) reduce 446) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_PIPE) reduce 446) ((T_POW) shift) ((T_SL) reduce 446) ((T_SPACESHIP) reduce 446) ((T_SR) reduce 446) (("^") reduce 446) (("|") reduce 446)) 11760 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) ((":") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_PIPE) reduce 445) ((T_POW) shift) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("|") reduce 445)) 11762 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) ((":") reduce 570) (("<") reduce 570) (("=") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AND_EQUAL) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_COALESCE_EQUAL) reduce 570) ((T_CONCAT_EQUAL) reduce 570) ((T_DEC) reduce 570) ((T_DIV_EQUAL) reduce 570) ((T_INC) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_MINUS_EQUAL) reduce 570) ((T_MOD_EQUAL) reduce 570) ((T_MUL_EQUAL) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_OR_EQUAL) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_PLUS_EQUAL) reduce 570) ((T_POW) reduce 570) ((T_POW_EQUAL) reduce 570) ((T_SL) reduce 570) ((T_SL_EQUAL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) ((T_SR_EQUAL) reduce 570) ((T_XOR_EQUAL) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 11765 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) ((":") reduce 569) (("<") reduce 569) (("=") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AND_EQUAL) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_COALESCE_EQUAL) reduce 569) ((T_CONCAT_EQUAL) reduce 569) ((T_DEC) reduce 569) ((T_DIV_EQUAL) reduce 569) ((T_INC) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_MINUS_EQUAL) reduce 569) ((T_MOD_EQUAL) reduce 569) ((T_MUL_EQUAL) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_OR_EQUAL) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_PLUS_EQUAL) reduce 569) ((T_POW) reduce 569) ((T_POW_EQUAL) reduce 569) ((T_SL) reduce 569) ((T_SL_EQUAL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) ((T_SR_EQUAL) reduce 569) ((T_XOR_EQUAL) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 11767 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) ((":") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_PIPE) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("|") reduce 522)) 11768 ((("%") reduce 447) (("*") reduce 447) (("+") reduce 447) (("-") reduce 447) ((".") reduce 447) (("/") reduce 447) ((":") reduce 447) (("<") reduce 447) ((">") reduce 447) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_BOOLEAN_AND) reduce 447) ((T_BOOLEAN_OR) reduce 447) ((T_COALESCE) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 447) ((T_IS_GREATER_OR_EQUAL) reduce 447) ((T_IS_IDENTICAL) reduce 447) ((T_IS_NOT_EQUAL) reduce 447) ((T_IS_NOT_IDENTICAL) reduce 447) ((T_IS_SMALLER_OR_EQUAL) reduce 447) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_PIPE) reduce 447) ((T_POW) shift) ((T_SL) reduce 447) ((T_SPACESHIP) reduce 447) ((T_SR) reduce 447) (("^") reduce 447) (("|") reduce 447)) 11770 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 463)) 12444 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 464)) 12455 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 456) ((T_SR) shift) (("^") reduce 456) (("|") reduce 456) (("}") reduce 456)) 12456 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 454) ((T_SR) shift) (("^") reduce 454) (("|") reduce 454) (("}") reduce 454)) 12457 ((("%") reduce 441) (("*") reduce 441) (("+") reduce 441) (("-") reduce 441) ((".") reduce 441) (("/") reduce 441) (("<") reduce 441) ((">") reduce 441) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 441) ((T_IS_GREATER_OR_EQUAL) reduce 441) ((T_IS_IDENTICAL) reduce 441) ((T_IS_NOT_EQUAL) reduce 441) ((T_IS_NOT_IDENTICAL) reduce 441) ((T_IS_SMALLER_OR_EQUAL) reduce 441) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_PIPE) reduce 441) ((T_POW) shift) ((T_SL) reduce 441) ((T_SPACESHIP) reduce 441) ((T_SR) reduce 441) (("^") reduce 441) (("|") reduce 441) (("}") reduce 441)) 12458 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 436) (("/") shift) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_PIPE) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 436) ((T_SR) shift) (("^") reduce 436) (("|") reduce 436) (("}") reduce 436)) 12459 ((("%") shift) (("*") shift) (("+") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") shift) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_PIPE) reduce 438) ((T_POW) shift) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("|") reduce 438) (("}") reduce 438)) 12460 ((("%") shift) (("*") shift) (("+") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") shift) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_PIPE) reduce 437) ((T_POW) shift) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("|") reduce 437) (("}") reduce 437)) 12461 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_PIPE) reduce 439) ((T_POW) shift) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("|") reduce 439) (("}") reduce 439)) 12462 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_PIPE) reduce 442) ((T_POW) shift) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("|") reduce 442) (("}") reduce 442)) 12464 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_PIPE) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("|") reduce 74) (("}") reduce 74)) 12465 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_PIPE) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("|") reduce 52) (("}") reduce 52)) 12466 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_PIPE) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("|") reduce 27) (("}") reduce 27)) 12467 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_PIPE) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("|") reduce 51) (("}") reduce 51)) 12468 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_PIPE) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("|") reduce 53) (("}") reduce 53)) 12469 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_PIPE) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("|") reduce 49) (("}") reduce 49)) 12470 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_PIPE) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("|") reduce 29) (("}") reduce 29)) 12471 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_PIPE) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("|") reduce 59) (("}") reduce 59)) 12472 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_PIPE) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("|") reduce 60) (("}") reduce 60)) 12473 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_PIPE) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("|") reduce 11) (("}") reduce 11)) 12474 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_PIPE) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("|") reduce 42) (("}") reduce 42)) 12475 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_PIPE) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("|") reduce 39) (("}") reduce 39)) 12476 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_PIPE) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("|") reduce 25) (("}") reduce 25)) 12477 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_PIPE) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("|") reduce 50) (("}") reduce 50)) 12478 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_PIPE) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("|") reduce 66) (("}") reduce 66)) 12479 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_PIPE) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("|") reduce 18) (("}") reduce 18)) 12480 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_PIPE) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("|") reduce 17) (("}") reduce 17)) 12481 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_PIPE) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("|") reduce 15) (("}") reduce 15)) 12482 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_PIPE) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("|") reduce 14) (("}") reduce 14)) 12483 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_PIPE) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("|") reduce 38) (("}") reduce 38)) 12484 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_PIPE) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("|") reduce 26) (("}") reduce 26)) 12485 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_PIPE) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("|") reduce 22) (("}") reduce 22)) 12486 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_PIPE) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("|") reduce 24) (("}") reduce 24)) 12487 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_PIPE) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("|") reduce 16) (("}") reduce 16)) 12488 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_PIPE) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("|") reduce 48) (("}") reduce 48)) 12489 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_PIPE) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("|") reduce 20) (("}") reduce 20)) 12490 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_PIPE) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("|") reduce 70) (("}") reduce 70)) 12491 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_PIPE) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("|") reduce 3) (("}") reduce 3)) 12492 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_PIPE) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("|") reduce 12) (("}") reduce 12)) 12493 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_PIPE) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("|") reduce 54) (("}") reduce 54)) 12494 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_PIPE) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("|") reduce 65) (("}") reduce 65)) 12495 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_PIPE) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("|") reduce 75) (("}") reduce 75)) 12496 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_PIPE) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("|") reduce 30) (("}") reduce 30)) 12497 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_PIPE) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("|") reduce 68) (("}") reduce 68)) 12498 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_PIPE) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("|") reduce 21) (("}") reduce 21)) 12499 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_PIPE) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("|") reduce 23) (("}") reduce 23)) 12500 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_PIPE) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("|") reduce 41) (("}") reduce 41)) 12501 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_PIPE) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("|") reduce 62) (("}") reduce 62)) 12502 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_PIPE) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("|") reduce 34) (("}") reduce 34)) 12503 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_PIPE) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("|") reduce 40) (("}") reduce 40)) 12504 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_PIPE) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("|") reduce 13) (("}") reduce 13)) 12505 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_PIPE) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("|") reduce 55) (("}") reduce 55)) 12506 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_PIPE) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("|") reduce 1) (("}") reduce 1)) 12507 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_PIPE) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("|") reduce 2) (("}") reduce 2)) 12508 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_PIPE) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("|") reduce 9) (("}") reduce 9)) 12509 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_PIPE) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("|") reduce 33) (("}") reduce 33)) 12510 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_PIPE) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("|") reduce 58) (("}") reduce 58)) 12511 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_PIPE) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("|") reduce 37) (("}") reduce 37)) 12512 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_PIPE) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("|") reduce 64) (("}") reduce 64)) 12513 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_PIPE) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("|") reduce 46) (("}") reduce 46)) 12514 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_PIPE) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("|") reduce 8) (("}") reduce 8)) 12515 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_PIPE) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("|") reduce 6) (("}") reduce 6)) 12516 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_PIPE) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("|") reduce 7) (("}") reduce 7)) 12517 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_PIPE) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("|") reduce 69) (("}") reduce 69)) 12518 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_PIPE) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("|") reduce 63) (("}") reduce 63)) 12519 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_PIPE) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("|") reduce 56) (("}") reduce 56)) 12520 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_PIPE) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("|") reduce 10) (("}") reduce 10)) 12521 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_PIPE) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("|") reduce 67) (("}") reduce 67)) 12522 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_PIPE) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("|") reduce 44) (("}") reduce 44)) 12523 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_PIPE) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("|") reduce 76) (("}") reduce 76)) 12524 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_PIPE) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("|") reduce 71) (("}") reduce 71)) 12525 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_PIPE) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("|") reduce 77) (("}") reduce 77)) 12526 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_PIPE) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("|") reduce 78) (("}") reduce 78)) 12527 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_PIPE) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("|") reduce 79) (("}") reduce 79)) 12528 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_PIPE) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("|") reduce 4) (("}") reduce 4)) 12529 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_PIPE) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("|") reduce 5) (("}") reduce 5)) 12530 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_PIPE) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("|") reduce 43) (("}") reduce 43)) 12531 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_PIPE) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("|") reduce 73) (("}") reduce 73)) 12532 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_PIPE) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("|") reduce 82) (("}") reduce 82)) 12533 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_PIPE) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("|") reduce 47) (("}") reduce 47)) 12534 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_PIPE) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("|") reduce 31) (("}") reduce 31)) 12535 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_PIPE) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("|") reduce 57) (("}") reduce 57)) 12536 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_PIPE) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("|") reduce 61) (("}") reduce 61)) 12537 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_PIPE) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("|") reduce 28) (("}") reduce 28)) 12538 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_PIPE) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("|") reduce 36) (("}") reduce 36)) 12539 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_PIPE) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("|") reduce 32) (("}") reduce 32)) 12540 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_PIPE) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("|") reduce 35) (("}") reduce 35)) 12541 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_PIPE) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("|") reduce 19) (("}") reduce 19)) 12542 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_PIPE) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("|") reduce 45) (("}") reduce 45)) 12543 ((("%") reduce 542) (("(") reduce 579) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PIPE) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542) (("}") reduce 542)) 12545 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_PIPE) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("|") reduce 72) (("}") reduce 72)) 12546 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_PIPE) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("|") reduce 83) (("}") reduce 83)) 12547 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) (("<") reduce 572) (("=") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AND_EQUAL) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_COALESCE_EQUAL) reduce 572) ((T_CONCAT_EQUAL) reduce 572) ((T_DEC) reduce 572) ((T_DIV_EQUAL) reduce 572) ((T_INC) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_MINUS_EQUAL) reduce 572) ((T_MOD_EQUAL) reduce 572) ((T_MUL_EQUAL) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_OR_EQUAL) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_PLUS_EQUAL) reduce 572) ((T_POW) reduce 572) ((T_POW_EQUAL) reduce 572) ((T_SL) reduce 572) ((T_SL_EQUAL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) ((T_SR_EQUAL) reduce 572) ((T_XOR_EQUAL) reduce 572) (("[") reduce 572) (("^") reduce 572) (("|") reduce 572) (("}") reduce 572)) 12550 ((("%") reduce 544) (("(") reduce 580) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PIPE) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544) (("}") reduce 544)) 12552 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("|") reduce 507) (("}") reduce 507)) 12553 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291) (("}") reduce 291)) 12557 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) (("=") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AND_EQUAL) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_COALESCE_EQUAL) reduce 292) ((T_CONCAT_EQUAL) reduce 292) ((T_DEC) reduce 292) ((T_DIV_EQUAL) reduce 292) ((T_INC) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_MINUS_EQUAL) reduce 292) ((T_MOD_EQUAL) reduce 292) ((T_MUL_EQUAL) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_OR_EQUAL) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_PLUS_EQUAL) reduce 292) ((T_POW) reduce 292) ((T_POW_EQUAL) reduce 292) ((T_SL) reduce 292) ((T_SL_EQUAL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) ((T_SR_EQUAL) reduce 292) ((T_XOR_EQUAL) reduce 292) (("[") reduce 292) (("^") reduce 292) (("|") reduce 292) (("}") reduce 292)) 12558 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) (("=") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AND_EQUAL) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_COALESCE_EQUAL) reduce 293) ((T_CONCAT_EQUAL) reduce 293) ((T_DEC) reduce 293) ((T_DIV_EQUAL) reduce 293) ((T_INC) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_MINUS_EQUAL) reduce 293) ((T_MOD_EQUAL) reduce 293) ((T_MUL_EQUAL) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_OR_EQUAL) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_PLUS_EQUAL) reduce 293) ((T_POW) reduce 293) ((T_POW_EQUAL) reduce 293) ((T_SL) reduce 293) ((T_SL_EQUAL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) ((T_SR_EQUAL) reduce 293) ((T_XOR_EQUAL) reduce 293) (("[") reduce 293) (("^") reduce 293) (("|") reduce 293) (("}") reduce 293)) 12563 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_PIPE) reduce 425) ((T_POW) reduce 425) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425) (("}") reduce 425)) 12565 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_PIPE) reduce 423) ((T_POW) reduce 423) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423) (("}") reduce 423)) 12575 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 419)) 12576 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 421)) 12577 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 420)) 12578 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 413)) 12579 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 410)) 12580 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 418)) 12581 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 412)) 12582 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 416)) 12583 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 411)) 12584 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 414)) 12585 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 415)) 12586 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 422)) 12587 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 417)) 12589 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 406)) 12592 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568) (("}") reduce 568)) 12595 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) (("<") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_POW) reduce 564) ((T_SL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) (("[") reduce 564) (("^") reduce 564) (("|") reduce 564) (("}") reduce 564)) 12597 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) (("<") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_POW) reduce 563) ((T_SL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) (("[") reduce 563) (("^") reduce 563) (("|") reduce 563) (("}") reduce 563)) 12599 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559) (("}") reduce 559)) 12600 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565) (("}") reduce 565)) 12601 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("[") reduce 548) (("^") reduce 407) (("|") reduce 407) (("}") reduce 407)) 12605 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) (("<") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_POW) reduce 572) ((T_SL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) (("[") reduce 572) (("^") reduce 572) (("|") reduce 572) (("}") reduce 572)) 12607 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("|") reduce 507) (("}") reduce 507)) 12608 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291) (("}") reduce 291)) 12612 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("[") reduce 292) (("^") reduce 292) (("|") reduce 292) (("}") reduce 292)) 12613 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("[") reduce 293) (("^") reduce 293) (("|") reduce 293) (("}") reduce 293)) 12614 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504) (("}") reduce 504)) 12617 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) (("<") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_POW) reduce 571) ((T_SL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) (("[") reduce 571) (("^") reduce 571) (("|") reduce 571) (("}") reduce 571)) 12618 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("|") reduce 506) (("}") reduce 506)) 12619 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("|") reduce 508) (("}") reduce 508)) 12624 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560) (("}") reduce 560)) 12625 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582) (("}") reduce 582)) 12626 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566) (("}") reduce 566)) 12627 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584) (("}") reduce 584)) 12630 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583) (("}") reduce 583)) 12631 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561) (("}") reduce 561)) 12632 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567) (("}") reduce 567)) 12633 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("[") reduce 562) (("^") reduce 562) (("|") reduce 562) (("}") reduce 562)) 12634 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505) (("}") reduce 505)) 12635 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570) (("}") reduce 570)) 12638 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569) (("}") reduce 569)) 12639 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) (("=") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AND_EQUAL) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_COALESCE_EQUAL) reduce 504) ((T_CONCAT_EQUAL) reduce 504) ((T_DEC) reduce 504) ((T_DIV_EQUAL) reduce 504) ((T_INC) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_MINUS_EQUAL) reduce 504) ((T_MOD_EQUAL) reduce 504) ((T_MUL_EQUAL) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_OR_EQUAL) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_PLUS_EQUAL) reduce 504) ((T_POW) reduce 504) ((T_POW_EQUAL) reduce 504) ((T_SL) reduce 504) ((T_SL_EQUAL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) ((T_SR_EQUAL) reduce 504) ((T_XOR_EQUAL) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504) (("}") reduce 504)) 12650 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PIPE) reduce 494) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494) (("}") reduce 494)) 12651 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_PIPE) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("^") reduce 489) (("|") reduce 489) (("}") reduce 489)) 12724 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 494)) 13040 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_PIPE) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("^") reduce 490) (("|") reduce 490) (("}") reduce 490)) 13043 ((("%") reduce 541) (("(") reduce 579) (("*") reduce 541) (("+") reduce 541) (("-") reduce 541) ((".") reduce 541) (("/") reduce 541) (("<") reduce 541) ((">") reduce 541) (("?") reduce 541) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_BOOLEAN_AND) reduce 541) ((T_BOOLEAN_OR) reduce 541) ((T_COALESCE) reduce 541) ((T_INSTANCEOF) reduce 541) ((T_IS_EQUAL) reduce 541) ((T_IS_GREATER_OR_EQUAL) reduce 541) ((T_IS_IDENTICAL) reduce 541) ((T_IS_NOT_EQUAL) reduce 541) ((T_IS_NOT_IDENTICAL) reduce 541) ((T_IS_SMALLER_OR_EQUAL) reduce 541) ((T_LOGICAL_AND) reduce 541) ((T_LOGICAL_OR) reduce 541) ((T_LOGICAL_XOR) reduce 541) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 541) ((T_OBJECT_OPERATOR) reduce 541) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 541) ((T_PIPE) reduce 541) ((T_POW) reduce 541) ((T_SL) reduce 541) ((T_SPACESHIP) reduce 541) ((T_SR) reduce 541) (("[") reduce 541) (("^") reduce 541) (("|") reduce 541) (("}") reduce 541)) 13045 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) (("<") reduce 571) (("=") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AND_EQUAL) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_COALESCE_EQUAL) reduce 571) ((T_CONCAT_EQUAL) reduce 571) ((T_DEC) reduce 571) ((T_DIV_EQUAL) reduce 571) ((T_INC) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_MINUS_EQUAL) reduce 571) ((T_MOD_EQUAL) reduce 571) ((T_MUL_EQUAL) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_OR_EQUAL) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_PLUS_EQUAL) reduce 571) ((T_POW) reduce 571) ((T_POW_EQUAL) reduce 571) ((T_SL) reduce 571) ((T_SL_EQUAL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) ((T_SR_EQUAL) reduce 571) ((T_XOR_EQUAL) reduce 571) (("[") reduce 571) (("^") reduce 571) (("|") reduce 571) (("}") reduce 571)) 13048 ((("%") reduce 543) (("(") reduce 580) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PIPE) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543) (("}") reduce 543)) 13049 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("|") reduce 506) (("}") reduce 506)) 13050 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("|") reduce 508) (("}") reduce 508)) 13052 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_PIPE) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("^") reduce 485) (("|") reduce 485) (("}") reduce 485)) 13053 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_PIPE) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("^") reduce 487) (("|") reduce 487) (("}") reduce 487)) 13058 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) (("=") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AND_EQUAL) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_COALESCE_EQUAL) reduce 560) ((T_CONCAT_EQUAL) reduce 560) ((T_DEC) reduce 560) ((T_DIV_EQUAL) reduce 560) ((T_INC) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_MINUS_EQUAL) reduce 560) ((T_MOD_EQUAL) reduce 560) ((T_MUL_EQUAL) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_OR_EQUAL) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_PLUS_EQUAL) reduce 560) ((T_POW) reduce 560) ((T_POW_EQUAL) reduce 560) ((T_SL) reduce 560) ((T_SL_EQUAL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) ((T_SR_EQUAL) reduce 560) ((T_XOR_EQUAL) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560) (("}") reduce 560)) 13059 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) (("=") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AND_EQUAL) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_COALESCE_EQUAL) reduce 582) ((T_CONCAT_EQUAL) reduce 582) ((T_DEC) reduce 582) ((T_DIV_EQUAL) reduce 582) ((T_INC) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_MINUS_EQUAL) reduce 582) ((T_MOD_EQUAL) reduce 582) ((T_MUL_EQUAL) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_OR_EQUAL) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_PLUS_EQUAL) reduce 582) ((T_POW) reduce 582) ((T_POW_EQUAL) reduce 582) ((T_SL) reduce 582) ((T_SL_EQUAL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) ((T_SR_EQUAL) reduce 582) ((T_XOR_EQUAL) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582) (("}") reduce 582)) 13060 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566) (("}") reduce 566)) 13061 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) (("=") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AND_EQUAL) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_COALESCE_EQUAL) reduce 584) ((T_CONCAT_EQUAL) reduce 584) ((T_DEC) reduce 584) ((T_DIV_EQUAL) reduce 584) ((T_INC) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_MINUS_EQUAL) reduce 584) ((T_MOD_EQUAL) reduce 584) ((T_MUL_EQUAL) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_OR_EQUAL) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_PLUS_EQUAL) reduce 584) ((T_POW) reduce 584) ((T_POW_EQUAL) reduce 584) ((T_SL) reduce 584) ((T_SL_EQUAL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) ((T_SR_EQUAL) reduce 584) ((T_XOR_EQUAL) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584) (("}") reduce 584)) 13064 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) (("=") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AND_EQUAL) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_COALESCE_EQUAL) reduce 583) ((T_CONCAT_EQUAL) reduce 583) ((T_DEC) reduce 583) ((T_DIV_EQUAL) reduce 583) ((T_INC) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_MINUS_EQUAL) reduce 583) ((T_MOD_EQUAL) reduce 583) ((T_MUL_EQUAL) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_OR_EQUAL) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_PLUS_EQUAL) reduce 583) ((T_POW) reduce 583) ((T_POW_EQUAL) reduce 583) ((T_SL) reduce 583) ((T_SL_EQUAL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) ((T_SR_EQUAL) reduce 583) ((T_XOR_EQUAL) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583) (("}") reduce 583)) 13065 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) (("=") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AND_EQUAL) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_COALESCE_EQUAL) reduce 561) ((T_CONCAT_EQUAL) reduce 561) ((T_DEC) reduce 561) ((T_DIV_EQUAL) reduce 561) ((T_INC) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_MINUS_EQUAL) reduce 561) ((T_MOD_EQUAL) reduce 561) ((T_MUL_EQUAL) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_OR_EQUAL) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_PLUS_EQUAL) reduce 561) ((T_POW) reduce 561) ((T_POW_EQUAL) reduce 561) ((T_SL) reduce 561) ((T_SL_EQUAL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) ((T_SR_EQUAL) reduce 561) ((T_XOR_EQUAL) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561) (("}") reduce 561)) 13066 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567) (("}") reduce 567)) 13067 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) (("=") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AND_EQUAL) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_COALESCE_EQUAL) reduce 562) ((T_CONCAT_EQUAL) reduce 562) ((T_DEC) reduce 562) ((T_DIV_EQUAL) reduce 562) ((T_INC) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_MINUS_EQUAL) reduce 562) ((T_MOD_EQUAL) reduce 562) ((T_MUL_EQUAL) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_OR_EQUAL) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_PLUS_EQUAL) reduce 562) ((T_POW) reduce 562) ((T_POW_EQUAL) reduce 562) ((T_SL) reduce 562) ((T_SL_EQUAL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) ((T_SR_EQUAL) reduce 562) ((T_XOR_EQUAL) reduce 562) (("[") reduce 562) (("^") reduce 562) (("|") reduce 562) (("}") reduce 562)) 13069 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_PIPE) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477) (("}") reduce 477)) 13071 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) (("=") shift) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_PIPE) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("|") reduce 520) (("}") reduce 520)) 13073 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 405)) 13074 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 482)) 13085 ((("%") reduce 540) (("*") reduce 540) (("+") reduce 540) (("-") reduce 540) ((".") reduce 540) (("/") reduce 540) (("<") reduce 540) ((">") reduce 540) (("?") reduce 540) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_BOOLEAN_AND) reduce 540) ((T_BOOLEAN_OR) reduce 540) ((T_COALESCE) reduce 540) ((T_DOUBLE_ARROW) reduce 540) ((T_INSTANCEOF) reduce 540) ((T_IS_EQUAL) reduce 540) ((T_IS_GREATER_OR_EQUAL) reduce 540) ((T_IS_IDENTICAL) reduce 540) ((T_IS_NOT_EQUAL) reduce 540) ((T_IS_NOT_IDENTICAL) reduce 540) ((T_IS_SMALLER_OR_EQUAL) reduce 540) ((T_LOGICAL_AND) reduce 540) ((T_LOGICAL_OR) reduce 540) ((T_LOGICAL_XOR) reduce 540) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 540) ((T_OBJECT_OPERATOR) reduce 540) ((T_PIPE) reduce 540) ((T_POW) reduce 540) ((T_SL) reduce 540) ((T_SPACESHIP) reduce 540) ((T_SR) reduce 540) (("[") reduce 540) (("^") reduce 540) (("|") reduce 540) (("}") reduce 540)) 13087 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_PIPE) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("|") reduce 521) (("}") reduce 521)) 13089 ((("%") reduce 534) (("*") reduce 534) (("+") reduce 534) (("-") reduce 534) ((".") reduce 534) (("/") reduce 534) (("<") reduce 534) ((">") reduce 534) (("?") reduce 534) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_BOOLEAN_AND) reduce 534) ((T_BOOLEAN_OR) reduce 534) ((T_COALESCE) reduce 534) ((T_DOUBLE_ARROW) reduce 534) ((T_INSTANCEOF) reduce 534) ((T_IS_EQUAL) reduce 534) ((T_IS_GREATER_OR_EQUAL) reduce 534) ((T_IS_IDENTICAL) reduce 534) ((T_IS_NOT_EQUAL) reduce 534) ((T_IS_NOT_IDENTICAL) reduce 534) ((T_IS_SMALLER_OR_EQUAL) reduce 534) ((T_LOGICAL_AND) reduce 534) ((T_LOGICAL_OR) reduce 534) ((T_LOGICAL_XOR) reduce 534) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PIPE) reduce 534) ((T_POW) reduce 534) ((T_SL) reduce 534) ((T_SPACESHIP) reduce 534) ((T_SR) reduce 534) (("[") reduce 534) (("^") reduce 534) (("|") reduce 534) (("}") reduce 534)) 13090 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_PIPE) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("^") reduce 524) (("|") reduce 524) (("}") reduce 524)) 13094 ((("%") reduce 517) (("(") shift) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_PIPE) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("^") reduce 517) (("|") reduce 517) (("}") reduce 517)) 13095 ((("%") reduce 533) (("*") reduce 533) (("+") reduce 533) (("-") reduce 533) ((".") reduce 533) (("/") reduce 533) (("<") reduce 533) ((">") reduce 533) (("?") reduce 533) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_BOOLEAN_AND) reduce 533) ((T_BOOLEAN_OR) reduce 533) ((T_COALESCE) reduce 533) ((T_DOUBLE_ARROW) reduce 533) ((T_INSTANCEOF) reduce 533) ((T_IS_EQUAL) reduce 533) ((T_IS_GREATER_OR_EQUAL) reduce 533) ((T_IS_IDENTICAL) reduce 533) ((T_IS_NOT_EQUAL) reduce 533) ((T_IS_NOT_IDENTICAL) reduce 533) ((T_IS_SMALLER_OR_EQUAL) reduce 533) ((T_LOGICAL_AND) reduce 533) ((T_LOGICAL_OR) reduce 533) ((T_LOGICAL_XOR) reduce 533) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PIPE) reduce 533) ((T_POW) reduce 533) ((T_SL) reduce 533) ((T_SPACESHIP) reduce 533) ((T_SR) reduce 533) (("[") reduce 533) (("^") reduce 533) (("|") reduce 533) (("}") reduce 533)) 13096 ((("%") reduce 537) (("*") reduce 537) (("+") reduce 537) (("-") reduce 537) ((".") reduce 537) (("/") reduce 537) (("<") reduce 537) ((">") reduce 537) (("?") reduce 537) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_BOOLEAN_AND) reduce 537) ((T_BOOLEAN_OR) reduce 537) ((T_COALESCE) reduce 537) ((T_DOUBLE_ARROW) reduce 537) ((T_INSTANCEOF) reduce 537) ((T_IS_EQUAL) reduce 537) ((T_IS_GREATER_OR_EQUAL) reduce 537) ((T_IS_IDENTICAL) reduce 537) ((T_IS_NOT_EQUAL) reduce 537) ((T_IS_NOT_IDENTICAL) reduce 537) ((T_IS_SMALLER_OR_EQUAL) reduce 537) ((T_LOGICAL_AND) reduce 537) ((T_LOGICAL_OR) reduce 537) ((T_LOGICAL_XOR) reduce 537) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_PIPE) reduce 537) ((T_POW) reduce 537) ((T_SL) reduce 537) ((T_SPACESHIP) reduce 537) ((T_SR) reduce 537) (("[") reduce 537) (("^") reduce 537) (("|") reduce 537) (("}") reduce 537)) 13102 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PIPE) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("|") reduce 532) (("}") reduce 532)) 13104 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_PIPE) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("^") reduce 523) (("|") reduce 523) (("}") reduce 523)) 13106 ((("%") reduce 536) (("*") reduce 536) (("+") reduce 536) (("-") reduce 536) ((".") reduce 536) (("/") reduce 536) (("<") reduce 536) ((">") reduce 536) (("?") reduce 536) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_BOOLEAN_AND) reduce 536) ((T_BOOLEAN_OR) reduce 536) ((T_COALESCE) reduce 536) ((T_DOUBLE_ARROW) reduce 536) ((T_INSTANCEOF) reduce 536) ((T_IS_EQUAL) reduce 536) ((T_IS_GREATER_OR_EQUAL) reduce 536) ((T_IS_IDENTICAL) reduce 536) ((T_IS_NOT_EQUAL) reduce 536) ((T_IS_NOT_IDENTICAL) reduce 536) ((T_IS_SMALLER_OR_EQUAL) reduce 536) ((T_LOGICAL_AND) reduce 536) ((T_LOGICAL_OR) reduce 536) ((T_LOGICAL_XOR) reduce 536) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 536) ((T_OBJECT_OPERATOR) reduce 536) ((T_PIPE) reduce 536) ((T_POW) reduce 536) ((T_SL) reduce 536) ((T_SPACESHIP) reduce 536) ((T_SR) reduce 536) (("[") reduce 536) (("^") reduce 536) (("|") reduce 536) (("}") reduce 536)) 13107 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("|") reduce 94) (("}") reduce 94)) 13108 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("|") reduce 93) (("}") reduce 93)) 13109 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("|") reduce 95) (("}") reduce 95)) 13111 ((("%") reduce 539) (("*") reduce 539) (("+") reduce 539) (("-") reduce 539) ((".") reduce 539) (("/") reduce 539) (("<") reduce 539) ((">") reduce 539) (("?") reduce 539) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_BOOLEAN_AND) reduce 539) ((T_BOOLEAN_OR) reduce 539) ((T_COALESCE) reduce 539) ((T_DOUBLE_ARROW) reduce 539) ((T_INSTANCEOF) reduce 539) ((T_IS_EQUAL) reduce 539) ((T_IS_GREATER_OR_EQUAL) reduce 539) ((T_IS_IDENTICAL) reduce 539) ((T_IS_NOT_EQUAL) reduce 539) ((T_IS_NOT_IDENTICAL) reduce 539) ((T_IS_SMALLER_OR_EQUAL) reduce 539) ((T_LOGICAL_AND) reduce 539) ((T_LOGICAL_OR) reduce 539) ((T_LOGICAL_XOR) reduce 539) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 539) ((T_OBJECT_OPERATOR) reduce 539) ((T_PIPE) reduce 539) ((T_POW) reduce 539) ((T_SL) reduce 539) ((T_SPACESHIP) reduce 539) ((T_SR) reduce 539) (("[") reduce 539) (("^") reduce 539) (("|") reduce 539) (("}") reduce 539)) 13114 ((("%") reduce 538) (("*") reduce 538) (("+") reduce 538) (("-") reduce 538) ((".") reduce 538) (("/") reduce 538) (("<") reduce 538) ((">") reduce 538) (("?") reduce 538) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_BOOLEAN_AND) reduce 538) ((T_BOOLEAN_OR) reduce 538) ((T_COALESCE) reduce 538) ((T_DOUBLE_ARROW) reduce 538) ((T_INSTANCEOF) reduce 538) ((T_IS_EQUAL) reduce 538) ((T_IS_GREATER_OR_EQUAL) reduce 538) ((T_IS_IDENTICAL) reduce 538) ((T_IS_NOT_EQUAL) reduce 538) ((T_IS_NOT_IDENTICAL) reduce 538) ((T_IS_SMALLER_OR_EQUAL) reduce 538) ((T_LOGICAL_AND) reduce 538) ((T_LOGICAL_OR) reduce 538) ((T_LOGICAL_XOR) reduce 538) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 538) ((T_OBJECT_OPERATOR) reduce 538) ((T_PIPE) reduce 538) ((T_POW) reduce 538) ((T_SL) reduce 538) ((T_SPACESHIP) reduce 538) ((T_SR) reduce 538) (("[") reduce 538) (("^") reduce 538) (("|") reduce 538) (("}") reduce 538)) 13120 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92) (("}") reduce 92)) 13123 ((("%") reduce 535) (("*") reduce 535) (("+") reduce 535) (("-") reduce 535) ((".") reduce 535) (("/") reduce 535) (("<") reduce 535) ((">") reduce 535) (("?") reduce 535) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_BOOLEAN_AND) reduce 535) ((T_BOOLEAN_OR) reduce 535) ((T_COALESCE) reduce 535) ((T_DOUBLE_ARROW) reduce 535) ((T_INSTANCEOF) reduce 535) ((T_IS_EQUAL) reduce 535) ((T_IS_GREATER_OR_EQUAL) reduce 535) ((T_IS_IDENTICAL) reduce 535) ((T_IS_NOT_EQUAL) reduce 535) ((T_IS_NOT_IDENTICAL) reduce 535) ((T_IS_SMALLER_OR_EQUAL) reduce 535) ((T_LOGICAL_AND) reduce 535) ((T_LOGICAL_OR) reduce 535) ((T_LOGICAL_XOR) reduce 535) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PIPE) reduce 535) ((T_POW) reduce 535) ((T_SL) reduce 535) ((T_SPACESHIP) reduce 535) ((T_SR) reduce 535) (("[") reduce 535) (("^") reduce 535) (("|") reduce 535) (("}") reduce 535)) 13125 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) (("=") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AND_EQUAL) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_COALESCE_EQUAL) reduce 568) ((T_CONCAT_EQUAL) reduce 568) ((T_DEC) reduce 568) ((T_DIV_EQUAL) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INC) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_MINUS_EQUAL) reduce 568) ((T_MOD_EQUAL) reduce 568) ((T_MUL_EQUAL) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_OR_EQUAL) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_PLUS_EQUAL) reduce 568) ((T_POW) reduce 568) ((T_POW_EQUAL) reduce 568) ((T_SL) reduce 568) ((T_SL_EQUAL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) ((T_SR_EQUAL) reduce 568) ((T_XOR_EQUAL) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568) (("}") reduce 568)) 13126 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 479) (("(") shift) (("*") reduce 479) (("+") shift) (("-") shift) ((".") reduce 479) (("/") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 479) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 479) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 479) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PIPE) reduce 479) ((T_POW) reduce 479) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 479) (("`") shift) (("|") reduce 479) (("}") reduce 479) (("~") shift)) 13133 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) (("<") reduce 564) (("=") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AND_EQUAL) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_COALESCE_EQUAL) reduce 564) ((T_CONCAT_EQUAL) reduce 564) ((T_DEC) reduce 564) ((T_DIV_EQUAL) reduce 564) ((T_DOUBLE_ARROW) reduce 564) ((T_INC) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_MINUS_EQUAL) reduce 564) ((T_MOD_EQUAL) reduce 564) ((T_MUL_EQUAL) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_OR_EQUAL) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_PLUS_EQUAL) reduce 564) ((T_POW) reduce 564) ((T_POW_EQUAL) reduce 564) ((T_SL) reduce 564) ((T_SL_EQUAL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) ((T_SR_EQUAL) reduce 564) ((T_XOR_EQUAL) reduce 564) (("[") reduce 564) (("^") reduce 564) (("|") reduce 564) (("}") reduce 564)) 13134 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PIPE) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 551) (("^") reduce 530) (("|") reduce 530) (("}") reduce 530)) 13136 ((("%") reduce 529) (("*") reduce 529) (("+") reduce 529) (("-") reduce 529) ((".") reduce 529) (("/") reduce 529) (("<") reduce 529) ((">") reduce 529) (("?") reduce 529) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_BOOLEAN_AND) reduce 529) ((T_BOOLEAN_OR) reduce 529) ((T_COALESCE) reduce 529) ((T_DOUBLE_ARROW) reduce 529) ((T_INSTANCEOF) reduce 529) ((T_IS_EQUAL) reduce 529) ((T_IS_GREATER_OR_EQUAL) reduce 529) ((T_IS_IDENTICAL) reduce 529) ((T_IS_NOT_EQUAL) reduce 529) ((T_IS_NOT_IDENTICAL) reduce 529) ((T_IS_SMALLER_OR_EQUAL) reduce 529) ((T_LOGICAL_AND) reduce 529) ((T_LOGICAL_OR) reduce 529) ((T_LOGICAL_XOR) reduce 529) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PIPE) reduce 529) ((T_POW) reduce 529) ((T_SL) reduce 529) ((T_SPACESHIP) reduce 529) ((T_SR) reduce 529) (("[") reduce 554) (("^") reduce 529) (("|") reduce 529) (("}") reduce 529)) 13137 ((("%") reduce 528) (("(") reduce 557) (("*") reduce 528) (("+") reduce 528) (("-") reduce 528) ((".") reduce 528) (("/") reduce 528) (("<") reduce 528) ((">") reduce 528) (("?") reduce 528) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_BOOLEAN_AND) reduce 528) ((T_BOOLEAN_OR) reduce 528) ((T_COALESCE) reduce 528) ((T_DOUBLE_ARROW) reduce 528) ((T_INSTANCEOF) reduce 528) ((T_IS_EQUAL) reduce 528) ((T_IS_GREATER_OR_EQUAL) reduce 528) ((T_IS_IDENTICAL) reduce 528) ((T_IS_NOT_EQUAL) reduce 528) ((T_IS_NOT_IDENTICAL) reduce 528) ((T_IS_SMALLER_OR_EQUAL) reduce 528) ((T_LOGICAL_AND) reduce 528) ((T_LOGICAL_OR) reduce 528) ((T_LOGICAL_XOR) reduce 528) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PIPE) reduce 528) ((T_POW) reduce 528) ((T_SL) reduce 528) ((T_SPACESHIP) reduce 528) ((T_SR) reduce 528) (("[") reduce 550) (("^") reduce 528) (("|") reduce 528) (("}") reduce 528)) 13138 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 480) ((T_LOGICAL_OR) reduce 480) ((T_LOGICAL_XOR) reduce 480) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 480)) 13141 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) (("<") reduce 563) (("=") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AND_EQUAL) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_COALESCE_EQUAL) reduce 563) ((T_CONCAT_EQUAL) reduce 563) ((T_DEC) reduce 563) ((T_DIV_EQUAL) reduce 563) ((T_DOUBLE_ARROW) reduce 563) ((T_INC) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_MINUS_EQUAL) reduce 563) ((T_MOD_EQUAL) reduce 563) ((T_MUL_EQUAL) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_OR_EQUAL) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_PLUS_EQUAL) reduce 563) ((T_POW) reduce 563) ((T_POW_EQUAL) reduce 563) ((T_SL) reduce 563) ((T_SL_EQUAL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) ((T_SR_EQUAL) reduce 563) ((T_XOR_EQUAL) reduce 563) (("[") reduce 563) (("^") reduce 563) (("|") reduce 563) (("}") reduce 563)) 13142 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_DOUBLE_ARROW) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_PIPE) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("^") reduce 484) (("|") reduce 484) (("}") reduce 484)) 13143 ((("%") reduce 466) (("*") reduce 466) (("+") reduce 466) (("-") reduce 466) ((".") reduce 466) (("/") reduce 466) (("<") reduce 466) ((">") reduce 466) (("?") reduce 466) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_BOOLEAN_AND) reduce 466) ((T_BOOLEAN_OR) reduce 466) ((T_COALESCE) reduce 466) ((T_DOUBLE_ARROW) reduce 466) ((T_INSTANCEOF) reduce 466) ((T_IS_EQUAL) reduce 466) ((T_IS_GREATER_OR_EQUAL) reduce 466) ((T_IS_IDENTICAL) reduce 466) ((T_IS_NOT_EQUAL) reduce 466) ((T_IS_NOT_IDENTICAL) reduce 466) ((T_IS_SMALLER_OR_EQUAL) reduce 466) ((T_LOGICAL_AND) reduce 466) ((T_LOGICAL_OR) reduce 466) ((T_LOGICAL_XOR) reduce 466) ((T_PIPE) reduce 466) ((T_POW) reduce 466) ((T_SL) reduce 466) ((T_SPACESHIP) reduce 466) ((T_SR) reduce 466) (("^") reduce 466) (("|") reduce 466) (("}") reduce 466)) 13144 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_PIPE) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("^") reduce 488) (("|") reduce 488) (("}") reduce 488)) 13145 ((("%") reduce 531) (("(") shift) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("|") reduce 531) (("}") reduce 531)) 13146 ((("%") reduce 461) (("(") reduce 558) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PIPE) reduce 461) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") reduce 552) (("^") reduce 461) (("|") reduce 461) (("}") reduce 461)) 13147 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_DOUBLE_ARROW) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_PIPE) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462) (("}") reduce 462)) 13148 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_DOUBLE_ARROW) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_PIPE) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("|") reduce 476) (("}") reduce 476)) 13149 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) (("=") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AND_EQUAL) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_COALESCE_EQUAL) reduce 559) ((T_CONCAT_EQUAL) reduce 559) ((T_DEC) reduce 559) ((T_DIV_EQUAL) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INC) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_MINUS_EQUAL) reduce 559) ((T_MOD_EQUAL) reduce 559) ((T_MUL_EQUAL) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_OR_EQUAL) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_PLUS_EQUAL) reduce 559) ((T_POW) reduce 559) ((T_POW_EQUAL) reduce 559) ((T_SL) reduce 559) ((T_SL_EQUAL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) ((T_SR_EQUAL) reduce 559) ((T_XOR_EQUAL) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559) (("}") reduce 559)) 13150 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565) (("}") reduce 565)) 13151 ((("%") reduce 403) (("*") reduce 403) (("+") reduce 403) (("-") reduce 403) ((".") reduce 403) (("/") reduce 403) (("<") reduce 403) (("=") shift) ((">") reduce 403) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_DOUBLE_ARROW) reduce 403) ((T_INC) shift) ((T_INSTANCEOF) reduce 403) ((T_IS_EQUAL) reduce 403) ((T_IS_GREATER_OR_EQUAL) reduce 403) ((T_IS_IDENTICAL) reduce 403) ((T_IS_NOT_EQUAL) reduce 403) ((T_IS_NOT_IDENTICAL) reduce 403) ((T_IS_SMALLER_OR_EQUAL) reduce 403) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 403) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 403) ((T_POW_EQUAL) shift) ((T_SL) reduce 403) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 403) ((T_SR) reduce 403) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 548) (("^") reduce 403) (("|") reduce 403) (("}") reduce 403)) 13154 ((("%") reduce 448) (("*") reduce 448) (("+") reduce 448) (("-") reduce 448) ((".") reduce 448) (("/") reduce 448) (("<") reduce 448) ((">") reduce 448) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_BOOLEAN_AND) reduce 448) ((T_BOOLEAN_OR) reduce 448) ((T_COALESCE) reduce 448) ((T_DOUBLE_ARROW) reduce 448) ((T_INSTANCEOF) reduce 448) ((T_IS_EQUAL) reduce 448) ((T_IS_GREATER_OR_EQUAL) reduce 448) ((T_IS_IDENTICAL) reduce 448) ((T_IS_NOT_EQUAL) reduce 448) ((T_IS_NOT_IDENTICAL) reduce 448) ((T_IS_SMALLER_OR_EQUAL) reduce 448) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_PIPE) reduce 448) ((T_POW) shift) ((T_SL) reduce 448) ((T_SPACESHIP) reduce 448) ((T_SR) reduce 448) (("^") reduce 448) (("|") reduce 448) (("}") reduce 448)) 13186 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_DOUBLE_ARROW) reduce 432) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 432) (("}") reduce 432)) 13187 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_DOUBLE_ARROW) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("|") reduce 435) (("}") reduce 435)) 13188 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 444) (("/") shift) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_PIPE) reduce 444) ((T_POW) shift) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("|") reduce 444) (("}") reduce 444)) 13189 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 458) (("|") reduce 458) (("}") reduce 458)) 13190 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 443) (("/") shift) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_DOUBLE_ARROW) reduce 443) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_PIPE) reduce 443) ((T_POW) shift) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("^") reduce 443) (("|") reduce 443) (("}") reduce 443)) 13191 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_DOUBLE_ARROW) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_PIPE) reduce 440) ((T_POW) shift) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("|") reduce 440) (("}") reduce 440)) 13192 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_DOUBLE_ARROW) reduce 453) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_PIPE) reduce 453) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 453) ((T_SR) shift) (("^") reduce 453) (("|") reduce 453) (("}") reduce 453)) 13193 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 431)) 13194 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 429) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 429)) 13195 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 430) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 430)) 13196 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 455) ((T_SR) shift) (("^") reduce 455) (("|") reduce 455) (("}") reduce 455)) 13197 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_DOUBLE_ARROW) reduce 450) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 450) (("|") reduce 450) (("}") reduce 450)) 13198 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_DOUBLE_ARROW) reduce 452) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 452) (("|") reduce 452) (("}") reduce 452)) 13199 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_DOUBLE_ARROW) reduce 449) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 449) (("|") reduce 449) (("}") reduce 449)) 13200 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 457) ((T_SR) shift) (("^") reduce 457) (("|") reduce 457) (("}") reduce 457)) 13201 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_DOUBLE_ARROW) reduce 451) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 451) (("|") reduce 451) (("}") reduce 451)) 13204 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94) (("}") reduce 94)) 13205 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93) (("}") reduce 93)) 13206 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95) (("}") reduce 95)) 13207 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509) (("}") reduce 509)) 13208 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92) (("}") reduce 92)) 13209 ((("%") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568) (("}") reduce 568)) 13210 ((("%") reduce 511) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("|") reduce 511) (("}") reduce 511)) 13211 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PIPE) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459) (("}") reduce 459)) 13212 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510) (("}") reduce 510)) 13213 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("|") reduce 512) (("}") reduce 512)) 13214 ((("%") reduce 573) (("*") reduce 573) (("+") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_DOUBLE_ARROW) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("|") reduce 573) (("}") reduce 573)) 13220 ((("%") reduce 574) (("*") reduce 574) (("+") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_DOUBLE_ARROW) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("|") reduce 574) (("}") reduce 574)) 13221 ((("%") reduce 578) (("*") reduce 578) (("+") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_DOUBLE_ARROW) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("|") reduce 578) (("}") reduce 578)) 13222 ((("%") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582) (("}") reduce 582)) 13223 ((("%") reduce 575) (("*") reduce 575) (("+") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_DOUBLE_ARROW) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("|") reduce 575) (("}") reduce 575)) 13224 ((("%") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584) (("}") reduce 584)) 13227 ((("%") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583) (("}") reduce 583)) 13228 ((("%") reduce 576) (("*") reduce 576) (("+") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_DOUBLE_ARROW) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("|") reduce 576) (("}") reduce 576)) 13230 ((("%") reduce 577) (("*") reduce 577) (("+") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_DOUBLE_ARROW) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("|") reduce 577) (("}") reduce 577)) 13232 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("|") reduce 513) (("}") reduce 513)) 13233 ((("%") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570) (("}") reduce 570)) 13236 ((("%") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569) (("}") reduce 569)) 13237 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 465) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 465) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 465)) 13238 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_DOUBLE_ARROW) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 427)) 13239 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_DOUBLE_ARROW) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 428)) 13240 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_DOUBLE_ARROW) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("|") reduce 433) (("}") reduce 433)) 13241 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_DOUBLE_ARROW) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("|") reduce 434) (("}") reduce 434)) 13245 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 463) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 463)) 13246 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 464) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 464)) 13247 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 456) ((T_SR) shift) (("^") reduce 456) (("|") reduce 456) (("}") reduce 456)) 13248 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 454) ((T_SR) shift) (("^") reduce 454) (("|") reduce 454) (("}") reduce 454)) 13249 ((("%") reduce 441) (("*") reduce 441) (("+") reduce 441) (("-") reduce 441) ((".") reduce 441) (("/") reduce 441) (("<") reduce 441) ((">") reduce 441) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_DOUBLE_ARROW) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 441) ((T_IS_GREATER_OR_EQUAL) reduce 441) ((T_IS_IDENTICAL) reduce 441) ((T_IS_NOT_EQUAL) reduce 441) ((T_IS_NOT_IDENTICAL) reduce 441) ((T_IS_SMALLER_OR_EQUAL) reduce 441) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_PIPE) reduce 441) ((T_POW) shift) ((T_SL) reduce 441) ((T_SPACESHIP) reduce 441) ((T_SR) reduce 441) (("^") reduce 441) (("|") reduce 441) (("}") reduce 441)) 13250 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 436) (("/") shift) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_DOUBLE_ARROW) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_PIPE) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 436) ((T_SR) shift) (("^") reduce 436) (("|") reduce 436) (("}") reduce 436)) 13251 ((("%") shift) (("*") shift) (("+") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") shift) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_DOUBLE_ARROW) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_PIPE) reduce 438) ((T_POW) shift) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("|") reduce 438) (("}") reduce 438)) 13252 ((("%") shift) (("*") shift) (("+") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") shift) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_DOUBLE_ARROW) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_PIPE) reduce 437) ((T_POW) shift) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("|") reduce 437) (("}") reduce 437)) 13253 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_DOUBLE_ARROW) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_PIPE) reduce 439) ((T_POW) shift) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("|") reduce 439) (("}") reduce 439)) 13254 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_DOUBLE_ARROW) reduce 442) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_PIPE) reduce 442) ((T_POW) shift) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("|") reduce 442) (("}") reduce 442)) 13256 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_DOUBLE_ARROW) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_PIPE) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("|") reduce 74) (("}") reduce 74)) 13257 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_DOUBLE_ARROW) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_PIPE) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("|") reduce 52) (("}") reduce 52)) 13258 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_DOUBLE_ARROW) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_PIPE) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("|") reduce 27) (("}") reduce 27)) 13259 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_DOUBLE_ARROW) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_PIPE) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("|") reduce 51) (("}") reduce 51)) 13260 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_DOUBLE_ARROW) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_PIPE) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("|") reduce 53) (("}") reduce 53)) 13261 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_DOUBLE_ARROW) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_PIPE) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("|") reduce 49) (("}") reduce 49)) 13262 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_DOUBLE_ARROW) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_PIPE) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("|") reduce 29) (("}") reduce 29)) 13263 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_DOUBLE_ARROW) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_PIPE) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("|") reduce 59) (("}") reduce 59)) 13264 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_DOUBLE_ARROW) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_PIPE) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("|") reduce 60) (("}") reduce 60)) 13265 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_DOUBLE_ARROW) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_PIPE) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("|") reduce 11) (("}") reduce 11)) 13266 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_DOUBLE_ARROW) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_PIPE) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("|") reduce 42) (("}") reduce 42)) 13267 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_DOUBLE_ARROW) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_PIPE) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("|") reduce 39) (("}") reduce 39)) 13268 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_DOUBLE_ARROW) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_PIPE) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("|") reduce 25) (("}") reduce 25)) 13269 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_DOUBLE_ARROW) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_PIPE) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("|") reduce 50) (("}") reduce 50)) 13270 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_DOUBLE_ARROW) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_PIPE) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("|") reduce 66) (("}") reduce 66)) 13271 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_DOUBLE_ARROW) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_PIPE) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("|") reduce 18) (("}") reduce 18)) 13272 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_DOUBLE_ARROW) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_PIPE) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("|") reduce 17) (("}") reduce 17)) 13273 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_DOUBLE_ARROW) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_PIPE) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("|") reduce 15) (("}") reduce 15)) 13274 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_DOUBLE_ARROW) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_PIPE) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("|") reduce 14) (("}") reduce 14)) 13275 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_DOUBLE_ARROW) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_PIPE) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("|") reduce 38) (("}") reduce 38)) 13276 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_DOUBLE_ARROW) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_PIPE) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("|") reduce 26) (("}") reduce 26)) 13277 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_DOUBLE_ARROW) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_PIPE) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("|") reduce 22) (("}") reduce 22)) 13278 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_DOUBLE_ARROW) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_PIPE) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("|") reduce 24) (("}") reduce 24)) 13279 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_DOUBLE_ARROW) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_PIPE) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("|") reduce 16) (("}") reduce 16)) 13280 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_DOUBLE_ARROW) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_PIPE) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("|") reduce 48) (("}") reduce 48)) 13281 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_DOUBLE_ARROW) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_PIPE) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("|") reduce 20) (("}") reduce 20)) 13282 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_DOUBLE_ARROW) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_PIPE) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("|") reduce 70) (("}") reduce 70)) 13283 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_DOUBLE_ARROW) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_PIPE) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("|") reduce 3) (("}") reduce 3)) 13284 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_DOUBLE_ARROW) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_PIPE) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("|") reduce 12) (("}") reduce 12)) 13285 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_DOUBLE_ARROW) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_PIPE) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("|") reduce 54) (("}") reduce 54)) 13286 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_DOUBLE_ARROW) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_PIPE) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("|") reduce 65) (("}") reduce 65)) 13287 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_DOUBLE_ARROW) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_PIPE) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("|") reduce 75) (("}") reduce 75)) 13288 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_DOUBLE_ARROW) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_PIPE) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("|") reduce 30) (("}") reduce 30)) 13289 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_DOUBLE_ARROW) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_PIPE) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("|") reduce 68) (("}") reduce 68)) 13290 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_DOUBLE_ARROW) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_PIPE) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("|") reduce 21) (("}") reduce 21)) 13291 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_DOUBLE_ARROW) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_PIPE) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("|") reduce 23) (("}") reduce 23)) 13292 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_DOUBLE_ARROW) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_PIPE) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("|") reduce 41) (("}") reduce 41)) 13293 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_DOUBLE_ARROW) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_PIPE) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("|") reduce 62) (("}") reduce 62)) 13294 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_DOUBLE_ARROW) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_PIPE) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("|") reduce 34) (("}") reduce 34)) 13295 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_DOUBLE_ARROW) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_PIPE) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("|") reduce 40) (("}") reduce 40)) 13296 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_DOUBLE_ARROW) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_PIPE) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("|") reduce 13) (("}") reduce 13)) 13297 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_DOUBLE_ARROW) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_PIPE) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("|") reduce 55) (("}") reduce 55)) 13298 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_DOUBLE_ARROW) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_PIPE) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("|") reduce 1) (("}") reduce 1)) 13299 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_DOUBLE_ARROW) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_PIPE) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("|") reduce 2) (("}") reduce 2)) 13300 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_DOUBLE_ARROW) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_PIPE) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("|") reduce 9) (("}") reduce 9)) 13301 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_DOUBLE_ARROW) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_PIPE) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("|") reduce 33) (("}") reduce 33)) 13302 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_DOUBLE_ARROW) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_PIPE) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("|") reduce 58) (("}") reduce 58)) 13303 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_DOUBLE_ARROW) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_PIPE) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("|") reduce 37) (("}") reduce 37)) 13304 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_DOUBLE_ARROW) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_PIPE) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("|") reduce 64) (("}") reduce 64)) 13305 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_DOUBLE_ARROW) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_PIPE) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("|") reduce 46) (("}") reduce 46)) 13306 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_DOUBLE_ARROW) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_PIPE) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("|") reduce 8) (("}") reduce 8)) 13307 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_DOUBLE_ARROW) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_PIPE) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("|") reduce 6) (("}") reduce 6)) 13308 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_DOUBLE_ARROW) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_PIPE) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("|") reduce 7) (("}") reduce 7)) 13309 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_DOUBLE_ARROW) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_PIPE) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("|") reduce 69) (("}") reduce 69)) 13310 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_DOUBLE_ARROW) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_PIPE) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("|") reduce 63) (("}") reduce 63)) 13311 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_DOUBLE_ARROW) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_PIPE) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("|") reduce 56) (("}") reduce 56)) 13312 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_DOUBLE_ARROW) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_PIPE) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("|") reduce 10) (("}") reduce 10)) 13313 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_DOUBLE_ARROW) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_PIPE) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("|") reduce 67) (("}") reduce 67)) 13314 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_DOUBLE_ARROW) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_PIPE) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("|") reduce 44) (("}") reduce 44)) 13315 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_DOUBLE_ARROW) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_PIPE) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("|") reduce 76) (("}") reduce 76)) 13316 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_DOUBLE_ARROW) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_PIPE) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("|") reduce 71) (("}") reduce 71)) 13317 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_DOUBLE_ARROW) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_PIPE) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("|") reduce 77) (("}") reduce 77)) 13318 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_DOUBLE_ARROW) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_PIPE) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("|") reduce 78) (("}") reduce 78)) 13319 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_DOUBLE_ARROW) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_PIPE) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("|") reduce 79) (("}") reduce 79)) 13320 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_DOUBLE_ARROW) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_PIPE) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("|") reduce 4) (("}") reduce 4)) 13321 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_DOUBLE_ARROW) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_PIPE) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("|") reduce 5) (("}") reduce 5)) 13322 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_DOUBLE_ARROW) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_PIPE) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("|") reduce 43) (("}") reduce 43)) 13323 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_DOUBLE_ARROW) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_PIPE) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("|") reduce 73) (("}") reduce 73)) 13324 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_DOUBLE_ARROW) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_PIPE) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("|") reduce 82) (("}") reduce 82)) 13325 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_DOUBLE_ARROW) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_PIPE) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("|") reduce 47) (("}") reduce 47)) 13326 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_DOUBLE_ARROW) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_PIPE) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("|") reduce 31) (("}") reduce 31)) 13327 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_DOUBLE_ARROW) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_PIPE) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("|") reduce 57) (("}") reduce 57)) 13328 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_DOUBLE_ARROW) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_PIPE) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("|") reduce 61) (("}") reduce 61)) 13329 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_DOUBLE_ARROW) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_PIPE) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("|") reduce 28) (("}") reduce 28)) 13330 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_DOUBLE_ARROW) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_PIPE) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("|") reduce 36) (("}") reduce 36)) 13331 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_DOUBLE_ARROW) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_PIPE) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("|") reduce 32) (("}") reduce 32)) 13332 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_DOUBLE_ARROW) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_PIPE) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("|") reduce 35) (("}") reduce 35)) 13333 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_DOUBLE_ARROW) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_PIPE) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("|") reduce 19) (("}") reduce 19)) 13334 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_DOUBLE_ARROW) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_PIPE) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("|") reduce 45) (("}") reduce 45)) 13335 ((("%") reduce 542) (("(") reduce 579) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PIPE) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542) (("}") reduce 542)) 13337 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_DOUBLE_ARROW) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_PIPE) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("|") reduce 72) (("}") reduce 72)) 13338 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_DOUBLE_ARROW) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_PIPE) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("|") reduce 83) (("}") reduce 83)) 13339 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) (("<") reduce 572) (("=") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AND_EQUAL) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_COALESCE_EQUAL) reduce 572) ((T_CONCAT_EQUAL) reduce 572) ((T_DEC) reduce 572) ((T_DIV_EQUAL) reduce 572) ((T_DOUBLE_ARROW) reduce 572) ((T_INC) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_MINUS_EQUAL) reduce 572) ((T_MOD_EQUAL) reduce 572) ((T_MUL_EQUAL) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_OR_EQUAL) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_PLUS_EQUAL) reduce 572) ((T_POW) reduce 572) ((T_POW_EQUAL) reduce 572) ((T_SL) reduce 572) ((T_SL_EQUAL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) ((T_SR_EQUAL) reduce 572) ((T_XOR_EQUAL) reduce 572) (("[") reduce 572) (("^") reduce 572) (("|") reduce 572) (("}") reduce 572)) 13342 ((("%") reduce 544) (("(") reduce 580) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PIPE) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544) (("}") reduce 544)) 13344 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("|") reduce 507) (("}") reduce 507)) 13345 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291) (("}") reduce 291)) 13349 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) (("=") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AND_EQUAL) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_COALESCE_EQUAL) reduce 292) ((T_CONCAT_EQUAL) reduce 292) ((T_DEC) reduce 292) ((T_DIV_EQUAL) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INC) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_MINUS_EQUAL) reduce 292) ((T_MOD_EQUAL) reduce 292) ((T_MUL_EQUAL) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_OR_EQUAL) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_PLUS_EQUAL) reduce 292) ((T_POW) reduce 292) ((T_POW_EQUAL) reduce 292) ((T_SL) reduce 292) ((T_SL_EQUAL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) ((T_SR_EQUAL) reduce 292) ((T_XOR_EQUAL) reduce 292) (("[") reduce 292) (("^") reduce 292) (("|") reduce 292) (("}") reduce 292)) 13350 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) (("=") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AND_EQUAL) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_COALESCE_EQUAL) reduce 293) ((T_CONCAT_EQUAL) reduce 293) ((T_DEC) reduce 293) ((T_DIV_EQUAL) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INC) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_MINUS_EQUAL) reduce 293) ((T_MOD_EQUAL) reduce 293) ((T_MUL_EQUAL) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_OR_EQUAL) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_PLUS_EQUAL) reduce 293) ((T_POW) reduce 293) ((T_POW_EQUAL) reduce 293) ((T_SL) reduce 293) ((T_SL_EQUAL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) ((T_SR_EQUAL) reduce 293) ((T_XOR_EQUAL) reduce 293) (("[") reduce 293) (("^") reduce 293) (("|") reduce 293) (("}") reduce 293)) 13355 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_DOUBLE_ARROW) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_PIPE) reduce 425) ((T_POW) reduce 425) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425) (("}") reduce 425)) 13357 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_DOUBLE_ARROW) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_PIPE) reduce 423) ((T_POW) reduce 423) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423) (("}") reduce 423)) 13367 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 419)) 13368 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 421)) 13369 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 420)) 13370 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 413) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 413)) 13371 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 410) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 410)) 13372 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 418)) 13373 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 412)) 13374 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 416)) 13375 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 411)) 13376 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 414) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 414)) 13377 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 415) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 415)) 13378 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 422)) 13379 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 417)) 13381 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 406)) 13384 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568) (("}") reduce 568)) 13387 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) (("<") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_DOUBLE_ARROW) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_POW) reduce 564) ((T_SL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) (("[") reduce 564) (("^") reduce 564) (("|") reduce 564) (("}") reduce 564)) 13389 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) (("<") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_DOUBLE_ARROW) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_POW) reduce 563) ((T_SL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) (("[") reduce 563) (("^") reduce 563) (("|") reduce 563) (("}") reduce 563)) 13391 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559) (("}") reduce 559)) 13392 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565) (("}") reduce 565)) 13393 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_DOUBLE_ARROW) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("[") reduce 548) (("^") reduce 407) (("|") reduce 407) (("}") reduce 407)) 13397 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) (("<") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_DOUBLE_ARROW) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_POW) reduce 572) ((T_SL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) (("[") reduce 572) (("^") reduce 572) (("|") reduce 572) (("}") reduce 572)) 13399 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("|") reduce 507) (("}") reduce 507)) 13400 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291) (("}") reduce 291)) 13404 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("[") reduce 292) (("^") reduce 292) (("|") reduce 292) (("}") reduce 292)) 13405 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("[") reduce 293) (("^") reduce 293) (("|") reduce 293) (("}") reduce 293)) 13406 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504) (("}") reduce 504)) 13409 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) (("<") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_DOUBLE_ARROW) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_POW) reduce 571) ((T_SL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) (("[") reduce 571) (("^") reduce 571) (("|") reduce 571) (("}") reduce 571)) 13410 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("|") reduce 506) (("}") reduce 506)) 13411 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("|") reduce 508) (("}") reduce 508)) 13416 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560) (("}") reduce 560)) 13417 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582) (("}") reduce 582)) 13418 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566) (("}") reduce 566)) 13419 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584) (("}") reduce 584)) 13422 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583) (("}") reduce 583)) 13423 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561) (("}") reduce 561)) 13424 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567) (("}") reduce 567)) 13425 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("[") reduce 562) (("^") reduce 562) (("|") reduce 562) (("}") reduce 562)) 13426 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505) (("}") reduce 505)) 13427 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570) (("}") reduce 570)) 13430 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569) (("}") reduce 569)) 13431 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) (("=") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AND_EQUAL) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_COALESCE_EQUAL) reduce 504) ((T_CONCAT_EQUAL) reduce 504) ((T_DEC) reduce 504) ((T_DIV_EQUAL) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INC) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_MINUS_EQUAL) reduce 504) ((T_MOD_EQUAL) reduce 504) ((T_MUL_EQUAL) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_OR_EQUAL) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_PLUS_EQUAL) reduce 504) ((T_POW) reduce 504) ((T_POW_EQUAL) reduce 504) ((T_SL) reduce 504) ((T_SL_EQUAL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) ((T_SR_EQUAL) reduce 504) ((T_XOR_EQUAL) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504) (("}") reduce 504)) 13442 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PIPE) reduce 494) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494) (("}") reduce 494)) 13443 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_PIPE) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("^") reduce 489) (("|") reduce 489) (("}") reduce 489)) 13453 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 494)) 13454 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_PIPE) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("^") reduce 490) (("|") reduce 490) (("}") reduce 490)) 13456 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 481)) 13458 ((("%") reduce 541) (("(") reduce 579) (("*") reduce 541) (("+") reduce 541) (("-") reduce 541) ((".") reduce 541) (("/") reduce 541) (("<") reduce 541) ((">") reduce 541) (("?") reduce 541) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_BOOLEAN_AND) reduce 541) ((T_BOOLEAN_OR) reduce 541) ((T_COALESCE) reduce 541) ((T_DOUBLE_ARROW) reduce 541) ((T_INSTANCEOF) reduce 541) ((T_IS_EQUAL) reduce 541) ((T_IS_GREATER_OR_EQUAL) reduce 541) ((T_IS_IDENTICAL) reduce 541) ((T_IS_NOT_EQUAL) reduce 541) ((T_IS_NOT_IDENTICAL) reduce 541) ((T_IS_SMALLER_OR_EQUAL) reduce 541) ((T_LOGICAL_AND) reduce 541) ((T_LOGICAL_OR) reduce 541) ((T_LOGICAL_XOR) reduce 541) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 541) ((T_OBJECT_OPERATOR) reduce 541) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 541) ((T_PIPE) reduce 541) ((T_POW) reduce 541) ((T_SL) reduce 541) ((T_SPACESHIP) reduce 541) ((T_SR) reduce 541) (("[") reduce 541) (("^") reduce 541) (("|") reduce 541) (("}") reduce 541)) 13460 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) (("<") reduce 571) (("=") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AND_EQUAL) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_COALESCE_EQUAL) reduce 571) ((T_CONCAT_EQUAL) reduce 571) ((T_DEC) reduce 571) ((T_DIV_EQUAL) reduce 571) ((T_DOUBLE_ARROW) reduce 571) ((T_INC) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_MINUS_EQUAL) reduce 571) ((T_MOD_EQUAL) reduce 571) ((T_MUL_EQUAL) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_OR_EQUAL) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_PLUS_EQUAL) reduce 571) ((T_POW) reduce 571) ((T_POW_EQUAL) reduce 571) ((T_SL) reduce 571) ((T_SL_EQUAL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) ((T_SR_EQUAL) reduce 571) ((T_XOR_EQUAL) reduce 571) (("[") reduce 571) (("^") reduce 571) (("|") reduce 571) (("}") reduce 571)) 13463 ((("%") reduce 543) (("(") reduce 580) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PIPE) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543) (("}") reduce 543)) 13464 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("|") reduce 506) (("}") reduce 506)) 13465 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("|") reduce 508) (("}") reduce 508)) 13467 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_DOUBLE_ARROW) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_PIPE) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("^") reduce 485) (("|") reduce 485) (("}") reduce 485)) 13468 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_PIPE) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("^") reduce 487) (("|") reduce 487) (("}") reduce 487)) 13473 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) (("=") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AND_EQUAL) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_COALESCE_EQUAL) reduce 560) ((T_CONCAT_EQUAL) reduce 560) ((T_DEC) reduce 560) ((T_DIV_EQUAL) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INC) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_MINUS_EQUAL) reduce 560) ((T_MOD_EQUAL) reduce 560) ((T_MUL_EQUAL) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_OR_EQUAL) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_PLUS_EQUAL) reduce 560) ((T_POW) reduce 560) ((T_POW_EQUAL) reduce 560) ((T_SL) reduce 560) ((T_SL_EQUAL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) ((T_SR_EQUAL) reduce 560) ((T_XOR_EQUAL) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560) (("}") reduce 560)) 13474 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) (("=") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AND_EQUAL) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_COALESCE_EQUAL) reduce 582) ((T_CONCAT_EQUAL) reduce 582) ((T_DEC) reduce 582) ((T_DIV_EQUAL) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INC) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_MINUS_EQUAL) reduce 582) ((T_MOD_EQUAL) reduce 582) ((T_MUL_EQUAL) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_OR_EQUAL) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_PLUS_EQUAL) reduce 582) ((T_POW) reduce 582) ((T_POW_EQUAL) reduce 582) ((T_SL) reduce 582) ((T_SL_EQUAL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) ((T_SR_EQUAL) reduce 582) ((T_XOR_EQUAL) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582) (("}") reduce 582)) 13475 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566) (("}") reduce 566)) 13476 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) (("=") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AND_EQUAL) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_COALESCE_EQUAL) reduce 584) ((T_CONCAT_EQUAL) reduce 584) ((T_DEC) reduce 584) ((T_DIV_EQUAL) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INC) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_MINUS_EQUAL) reduce 584) ((T_MOD_EQUAL) reduce 584) ((T_MUL_EQUAL) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_OR_EQUAL) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_PLUS_EQUAL) reduce 584) ((T_POW) reduce 584) ((T_POW_EQUAL) reduce 584) ((T_SL) reduce 584) ((T_SL_EQUAL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) ((T_SR_EQUAL) reduce 584) ((T_XOR_EQUAL) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584) (("}") reduce 584)) 13479 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) (("=") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AND_EQUAL) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_COALESCE_EQUAL) reduce 583) ((T_CONCAT_EQUAL) reduce 583) ((T_DEC) reduce 583) ((T_DIV_EQUAL) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INC) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_MINUS_EQUAL) reduce 583) ((T_MOD_EQUAL) reduce 583) ((T_MUL_EQUAL) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_OR_EQUAL) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_PLUS_EQUAL) reduce 583) ((T_POW) reduce 583) ((T_POW_EQUAL) reduce 583) ((T_SL) reduce 583) ((T_SL_EQUAL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) ((T_SR_EQUAL) reduce 583) ((T_XOR_EQUAL) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583) (("}") reduce 583)) 13480 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) (("=") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AND_EQUAL) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_COALESCE_EQUAL) reduce 561) ((T_CONCAT_EQUAL) reduce 561) ((T_DEC) reduce 561) ((T_DIV_EQUAL) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INC) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_MINUS_EQUAL) reduce 561) ((T_MOD_EQUAL) reduce 561) ((T_MUL_EQUAL) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_OR_EQUAL) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_PLUS_EQUAL) reduce 561) ((T_POW) reduce 561) ((T_POW_EQUAL) reduce 561) ((T_SL) reduce 561) ((T_SL_EQUAL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) ((T_SR_EQUAL) reduce 561) ((T_XOR_EQUAL) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561) (("}") reduce 561)) 13481 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567) (("}") reduce 567)) 13482 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) (("=") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AND_EQUAL) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_COALESCE_EQUAL) reduce 562) ((T_CONCAT_EQUAL) reduce 562) ((T_DEC) reduce 562) ((T_DIV_EQUAL) reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_INC) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_MINUS_EQUAL) reduce 562) ((T_MOD_EQUAL) reduce 562) ((T_MUL_EQUAL) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_OR_EQUAL) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_PLUS_EQUAL) reduce 562) ((T_POW) reduce 562) ((T_POW_EQUAL) reduce 562) ((T_SL) reduce 562) ((T_SL_EQUAL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) ((T_SR_EQUAL) reduce 562) ((T_XOR_EQUAL) reduce 562) (("[") reduce 562) (("^") reduce 562) (("|") reduce 562) (("}") reduce 562)) 13484 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_PIPE) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477) (("}") reduce 477)) 13486 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) (("=") shift) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_PIPE) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("|") reduce 520) (("}") reduce 520)) 13488 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 405)) 13489 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 482) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 482)) 13492 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 481) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 481)) 13493 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_DOUBLE_ARROW) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_PIPE) reduce 473) ((T_POW) shift) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("|") reduce 473) (("}") reduce 473)) 13494 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 483) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 483)) 13495 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_DOUBLE_ARROW) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_PIPE) reduce 469) ((T_POW) shift) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469) (("}") reduce 469)) 13496 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_DOUBLE_ARROW) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_PIPE) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("^") reduce 486) (("|") reduce 486) (("}") reduce 486)) 13498 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_PIPE) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("^") reduce 526) (("|") reduce 526) (("}") reduce 526)) 13500 ((("%") reduce 527) (("*") reduce 527) (("+") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_DOUBLE_ARROW) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_PIPE) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("^") reduce 527) (("|") reduce 527) (("}") reduce 527)) 13501 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_PIPE) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("^") reduce 525) (("|") reduce 525) (("}") reduce 525)) 13502 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 619) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 619)) 13503 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 618) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 618)) 13504 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505) (("}") reduce 505)) 13505 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 478) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 478)) 13506 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_DOUBLE_ARROW) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_PIPE) reduce 471) ((T_POW) shift) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("|") reduce 471) (("}") reduce 471)) 13508 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94) (("}") reduce 94)) 13509 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93) (("}") reduce 93)) 13510 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95) (("}") reduce 95)) 13511 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509) (("}") reduce 509)) 13512 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92) (("}") reduce 92)) 13513 ((("%") reduce 400) (("(") reduce 400) (("*") reduce 400) (("+") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_DOUBLE_ARROW) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 400) ((T_OBJECT_OPERATOR) reduce 400) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 400) ((T_PIPE) reduce 400) ((T_POW) reduce 400) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("[") reduce 400) (("^") reduce 400) (("|") reduce 400) (("}") reduce 400)) 13516 ((("%") reduce 511) (("(") reduce 511) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("|") reduce 511) (("}") reduce 511)) 13517 ((("%") reduce 402) (("(") shift) (("*") reduce 402) (("+") reduce 402) (("-") reduce 402) ((".") reduce 402) (("/") reduce 402) (("<") reduce 402) ((">") reduce 402) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_DOUBLE_ARROW) reduce 402) ((T_INSTANCEOF) reduce 402) ((T_IS_EQUAL) reduce 402) ((T_IS_GREATER_OR_EQUAL) reduce 402) ((T_IS_IDENTICAL) reduce 402) ((T_IS_NOT_EQUAL) reduce 402) ((T_IS_NOT_IDENTICAL) reduce 402) ((T_IS_SMALLER_OR_EQUAL) reduce 402) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_PIPE) reduce 402) ((T_POW) reduce 402) ((T_SL) reduce 402) ((T_SPACESHIP) reduce 402) ((T_SR) reduce 402) (("^") reduce 402) (("|") reduce 402) (("}") reduce 402)) 13518 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510) (("}") reduce 510)) 13519 ((("%") reduce 512) (("(") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("|") reduce 512) (("}") reduce 512)) 13520 ((("%") reduce 573) (("(") reduce 573) (("*") reduce 573) (("+") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_DOUBLE_ARROW) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("|") reduce 573) (("}") reduce 573)) 13526 ((("%") reduce 574) (("(") reduce 574) (("*") reduce 574) (("+") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_DOUBLE_ARROW) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("|") reduce 574) (("}") reduce 574)) 13527 ((("%") reduce 578) (("(") reduce 578) (("*") reduce 578) (("+") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_DOUBLE_ARROW) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("|") reduce 578) (("}") reduce 578)) 13528 ((("%") reduce 575) (("(") reduce 575) (("*") reduce 575) (("+") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_DOUBLE_ARROW) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("|") reduce 575) (("}") reduce 575)) 13529 ((("%") reduce 576) (("(") reduce 576) (("*") reduce 576) (("+") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_DOUBLE_ARROW) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("|") reduce 576) (("}") reduce 576)) 13530 ((("%") reduce 399) (("(") reduce 399) (("*") reduce 399) (("+") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_DOUBLE_ARROW) reduce 399) ((T_INSTANCEOF) reduce 399) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 399) ((T_OBJECT_OPERATOR) reduce 399) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 399) ((T_PIPE) reduce 399) ((T_POW) reduce 399) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("[") reduce 399) (("^") reduce 399) (("|") reduce 399) (("}") reduce 399)) 13532 ((("%") reduce 577) (("(") reduce 577) (("*") reduce 577) (("+") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_DOUBLE_ARROW) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("|") reduce 577) (("}") reduce 577)) 13533 ((("%") reduce 401) (("(") reduce 401) (("*") reduce 401) (("+") reduce 401) (("-") reduce 401) ((".") reduce 401) (("/") reduce 401) (("<") reduce 401) ((">") reduce 401) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_DOUBLE_ARROW) reduce 401) ((T_INSTANCEOF) reduce 401) ((T_IS_EQUAL) reduce 401) ((T_IS_GREATER_OR_EQUAL) reduce 401) ((T_IS_IDENTICAL) reduce 401) ((T_IS_NOT_EQUAL) reduce 401) ((T_IS_NOT_IDENTICAL) reduce 401) ((T_IS_SMALLER_OR_EQUAL) reduce 401) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 401) ((T_OBJECT_OPERATOR) reduce 401) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 401) ((T_PIPE) reduce 401) ((T_POW) reduce 401) ((T_SL) reduce 401) ((T_SPACESHIP) reduce 401) ((T_SR) reduce 401) (("[") reduce 401) (("^") reduce 401) (("|") reduce 401) (("}") reduce 401)) 13541 ((("%") reduce 398) (("(") reduce 398) (("*") reduce 398) (("+") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_DOUBLE_ARROW) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 398) ((T_OBJECT_OPERATOR) reduce 398) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 398) ((T_PIPE) reduce 398) ((T_POW) reduce 398) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("[") reduce 398) (("^") reduce 398) (("|") reduce 398) (("}") reduce 398)) 13543 ((("%") reduce 513) (("(") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("|") reduce 513) (("}") reduce 513)) 13549 ((("%") reduce 233) (("*") reduce 233) (("+") reduce 233) (("-") reduce 233) ((".") reduce 233) (("/") reduce 233) (("<") reduce 233) ((">") reduce 233) (("?") reduce 233) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_BOOLEAN_AND) reduce 233) ((T_BOOLEAN_OR) reduce 233) ((T_COALESCE) reduce 233) ((T_DOUBLE_ARROW) reduce 233) ((T_INSTANCEOF) reduce 233) ((T_IS_EQUAL) reduce 233) ((T_IS_GREATER_OR_EQUAL) reduce 233) ((T_IS_IDENTICAL) reduce 233) ((T_IS_NOT_EQUAL) reduce 233) ((T_IS_NOT_IDENTICAL) reduce 233) ((T_IS_SMALLER_OR_EQUAL) reduce 233) ((T_LOGICAL_AND) reduce 233) ((T_LOGICAL_OR) reduce 233) ((T_LOGICAL_XOR) reduce 233) ((T_PIPE) reduce 233) ((T_POW) reduce 233) ((T_SL) reduce 233) ((T_SPACESHIP) reduce 233) ((T_SR) reduce 233) (("^") reduce 233) (("|") reduce 233) (("}") reduce 233)) 13554 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 404)) 13558 ((("%") reduce 613) (("*") reduce 613) (("+") reduce 613) (("-") reduce 613) ((".") reduce 613) (("/") reduce 613) (("<") reduce 613) ((">") reduce 613) (("?") reduce 613) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_BOOLEAN_AND) reduce 613) ((T_BOOLEAN_OR) reduce 613) ((T_COALESCE) reduce 613) ((T_DOUBLE_ARROW) reduce 613) ((T_INSTANCEOF) reduce 613) ((T_IS_EQUAL) reduce 613) ((T_IS_GREATER_OR_EQUAL) reduce 613) ((T_IS_IDENTICAL) reduce 613) ((T_IS_NOT_EQUAL) reduce 613) ((T_IS_NOT_IDENTICAL) reduce 613) ((T_IS_SMALLER_OR_EQUAL) reduce 613) ((T_LOGICAL_AND) reduce 613) ((T_LOGICAL_OR) reduce 613) ((T_LOGICAL_XOR) reduce 613) ((T_PIPE) reduce 613) ((T_POW) reduce 613) ((T_SL) reduce 613) ((T_SPACESHIP) reduce 613) ((T_SR) reduce 613) (("^") reduce 613) (("|") reduce 613) (("}") reduce 613)) 13559 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_DOUBLE_ARROW) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_PIPE) reduce 467) ((T_POW) shift) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("|") reduce 467) (("}") reduce 467)) 13560 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 616) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 616)) 13561 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 615) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 615)) 13562 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_DOUBLE_ARROW) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("[") reduce 548) (("^") reduce 424) (("|") reduce 424) (("}") reduce 424)) 13564 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_PIPE) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("^") reduce 518) (("|") reduce 518) (("}") reduce 518)) 13565 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_DOUBLE_ARROW) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_PIPE) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("|") reduce 474) (("}") reduce 474)) 13566 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("|") reduce 291) (("}") reduce 291)) 13570 ((("%") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("^") reduce 292) (("|") reduce 292) (("}") reduce 292)) 13571 ((("%") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("^") reduce 293) (("|") reduce 293) (("}") reduce 293)) 13574 ((("%") reduce 617) (("*") reduce 617) (("+") reduce 617) (("-") reduce 617) ((".") reduce 617) (("/") reduce 617) (("<") reduce 617) ((">") reduce 617) (("?") reduce 617) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_BOOLEAN_AND) reduce 617) ((T_BOOLEAN_OR) reduce 617) ((T_COALESCE) reduce 617) ((T_DOUBLE_ARROW) reduce 617) ((T_INSTANCEOF) reduce 617) ((T_IS_EQUAL) reduce 617) ((T_IS_GREATER_OR_EQUAL) reduce 617) ((T_IS_IDENTICAL) reduce 617) ((T_IS_NOT_EQUAL) reduce 617) ((T_IS_NOT_IDENTICAL) reduce 617) ((T_IS_SMALLER_OR_EQUAL) reduce 617) ((T_LOGICAL_AND) reduce 617) ((T_LOGICAL_OR) reduce 617) ((T_LOGICAL_XOR) reduce 617) ((T_PIPE) reduce 617) ((T_POW) reduce 617) ((T_SL) reduce 617) ((T_SPACESHIP) reduce 617) ((T_SR) reduce 617) (("^") reduce 617) (("|") reduce 617) (("}") reduce 617)) 13577 ((("%") reduce 614) (("*") reduce 614) (("+") reduce 614) (("-") reduce 614) ((".") reduce 614) (("/") reduce 614) (("<") reduce 614) ((">") reduce 614) (("?") reduce 614) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_BOOLEAN_AND) reduce 614) ((T_BOOLEAN_OR) reduce 614) ((T_COALESCE) reduce 614) ((T_DOUBLE_ARROW) reduce 614) ((T_INSTANCEOF) reduce 614) ((T_IS_EQUAL) reduce 614) ((T_IS_GREATER_OR_EQUAL) reduce 614) ((T_IS_IDENTICAL) reduce 614) ((T_IS_NOT_EQUAL) reduce 614) ((T_IS_NOT_IDENTICAL) reduce 614) ((T_IS_SMALLER_OR_EQUAL) reduce 614) ((T_LOGICAL_AND) reduce 614) ((T_LOGICAL_OR) reduce 614) ((T_LOGICAL_XOR) reduce 614) ((T_PIPE) reduce 614) ((T_POW) reduce 614) ((T_SL) reduce 614) ((T_SPACESHIP) reduce 614) ((T_SR) reduce 614) (("^") reduce 614) (("|") reduce 614) (("}") reduce 614)) 13578 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_DOUBLE_ARROW) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_PIPE) reduce 468) ((T_POW) shift) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468) (("}") reduce 468)) 13579 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_DOUBLE_ARROW) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("[") reduce 548) (("^") reduce 426) (("|") reduce 426) (("}") reduce 426)) 13581 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_DOUBLE_ARROW) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_PIPE) reduce 408) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("^") reduce 408) (("|") reduce 408) (("}") reduce 408)) 13582 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_DOUBLE_ARROW) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_PIPE) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("|") reduce 409) (("}") reduce 409)) 13583 ((("%") reduce 296) (("*") reduce 296) (("+") reduce 296) (("-") reduce 296) ((".") reduce 296) (("/") reduce 296) (("<") reduce 296) ((">") reduce 296) (("?") reduce 296) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_BOOLEAN_AND) reduce 296) ((T_BOOLEAN_OR) reduce 296) ((T_COALESCE) reduce 296) ((T_DOUBLE_ARROW) reduce 296) ((T_INSTANCEOF) reduce 296) ((T_IS_EQUAL) reduce 296) ((T_IS_GREATER_OR_EQUAL) reduce 296) ((T_IS_IDENTICAL) reduce 296) ((T_IS_NOT_EQUAL) reduce 296) ((T_IS_NOT_IDENTICAL) reduce 296) ((T_IS_SMALLER_OR_EQUAL) reduce 296) ((T_LOGICAL_AND) reduce 296) ((T_LOGICAL_OR) reduce 296) ((T_LOGICAL_XOR) reduce 296) ((T_PIPE) reduce 296) ((T_POW) reduce 296) ((T_SL) reduce 296) ((T_SPACESHIP) reduce 296) ((T_SR) reduce 296) (("^") reduce 296) (("|") reduce 296) (("}") reduce 296)) 13588 ((("%") reduce 297) (("*") reduce 297) (("+") reduce 297) (("-") reduce 297) ((".") reduce 297) (("/") reduce 297) (("<") reduce 297) ((">") reduce 297) (("?") reduce 297) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_BOOLEAN_AND) reduce 297) ((T_BOOLEAN_OR) reduce 297) ((T_COALESCE) reduce 297) ((T_DOUBLE_ARROW) reduce 297) ((T_INSTANCEOF) reduce 297) ((T_IS_EQUAL) reduce 297) ((T_IS_GREATER_OR_EQUAL) reduce 297) ((T_IS_IDENTICAL) reduce 297) ((T_IS_NOT_EQUAL) reduce 297) ((T_IS_NOT_IDENTICAL) reduce 297) ((T_IS_SMALLER_OR_EQUAL) reduce 297) ((T_LOGICAL_AND) reduce 297) ((T_LOGICAL_OR) reduce 297) ((T_LOGICAL_XOR) reduce 297) ((T_PIPE) reduce 297) ((T_POW) reduce 297) ((T_SL) reduce 297) ((T_SPACESHIP) reduce 297) ((T_SR) reduce 297) (("^") reduce 297) (("|") reduce 297) (("}") reduce 297)) 13589 ((("%") reduce 460) (("(") reduce 556) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PIPE) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("[") reduce 549) (("^") reduce 460) (("|") reduce 460) (("}") reduce 460)) 13591 ((("%") reduce 298) (("*") reduce 298) (("+") reduce 298) (("-") reduce 298) ((".") reduce 298) (("/") reduce 298) (("<") reduce 298) ((">") reduce 298) (("?") reduce 298) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_BOOLEAN_AND) reduce 298) ((T_BOOLEAN_OR) reduce 298) ((T_COALESCE) reduce 298) ((T_DOUBLE_ARROW) reduce 298) ((T_INSTANCEOF) reduce 298) ((T_IS_EQUAL) reduce 298) ((T_IS_GREATER_OR_EQUAL) reduce 298) ((T_IS_IDENTICAL) reduce 298) ((T_IS_NOT_EQUAL) reduce 298) ((T_IS_NOT_IDENTICAL) reduce 298) ((T_IS_SMALLER_OR_EQUAL) reduce 298) ((T_LOGICAL_AND) reduce 298) ((T_LOGICAL_OR) reduce 298) ((T_LOGICAL_XOR) reduce 298) ((T_PIPE) reduce 298) ((T_POW) reduce 298) ((T_SL) reduce 298) ((T_SPACESHIP) reduce 298) ((T_SR) reduce 298) (("^") reduce 298) (("|") reduce 298) (("}") reduce 298)) 13592 ((("%") reduce 299) (("*") reduce 299) (("+") reduce 299) (("-") reduce 299) ((".") reduce 299) (("/") reduce 299) (("<") reduce 299) ((">") reduce 299) (("?") reduce 299) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_BOOLEAN_AND) reduce 299) ((T_BOOLEAN_OR) reduce 299) ((T_COALESCE) reduce 299) ((T_DOUBLE_ARROW) reduce 299) ((T_INSTANCEOF) reduce 299) ((T_IS_EQUAL) reduce 299) ((T_IS_GREATER_OR_EQUAL) reduce 299) ((T_IS_IDENTICAL) reduce 299) ((T_IS_NOT_EQUAL) reduce 299) ((T_IS_NOT_IDENTICAL) reduce 299) ((T_IS_SMALLER_OR_EQUAL) reduce 299) ((T_LOGICAL_AND) reduce 299) ((T_LOGICAL_OR) reduce 299) ((T_LOGICAL_XOR) reduce 299) ((T_PIPE) reduce 299) ((T_POW) reduce 299) ((T_SL) reduce 299) ((T_SPACESHIP) reduce 299) ((T_SR) reduce 299) (("^") reduce 299) (("|") reduce 299) (("}") reduce 299)) 13593 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_DOUBLE_ARROW) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_PIPE) reduce 472) ((T_POW) shift) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("|") reduce 472) (("}") reduce 472)) 13594 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_DOUBLE_ARROW) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_PIPE) reduce 470) ((T_POW) shift) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("|") reduce 470) (("}") reduce 470)) 13597 ((("%") reduce 519) (("(") reduce 519) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PIPE) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("|") reduce 519) (("}") reduce 519)) 13598 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_DOUBLE_ARROW) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_PIPE) reduce 475) ((T_POW) shift) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("|") reduce 475) (("}") reduce 475)) 13599 ((("%") reduce 446) (("*") reduce 446) (("+") reduce 446) (("-") reduce 446) ((".") reduce 446) (("/") reduce 446) (("<") reduce 446) ((">") reduce 446) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_BOOLEAN_AND) reduce 446) ((T_BOOLEAN_OR) reduce 446) ((T_COALESCE) reduce 446) ((T_DOUBLE_ARROW) reduce 446) ((T_INSTANCEOF) reduce 446) ((T_IS_EQUAL) reduce 446) ((T_IS_GREATER_OR_EQUAL) reduce 446) ((T_IS_IDENTICAL) reduce 446) ((T_IS_NOT_EQUAL) reduce 446) ((T_IS_NOT_IDENTICAL) reduce 446) ((T_IS_SMALLER_OR_EQUAL) reduce 446) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_PIPE) reduce 446) ((T_POW) shift) ((T_SL) reduce 446) ((T_SPACESHIP) reduce 446) ((T_SR) reduce 446) (("^") reduce 446) (("|") reduce 446) (("}") reduce 446)) 13600 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_DOUBLE_ARROW) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_PIPE) reduce 445) ((T_POW) shift) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("|") reduce 445) (("}") reduce 445)) 13602 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) (("=") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AND_EQUAL) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_COALESCE_EQUAL) reduce 570) ((T_CONCAT_EQUAL) reduce 570) ((T_DEC) reduce 570) ((T_DIV_EQUAL) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INC) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_MINUS_EQUAL) reduce 570) ((T_MOD_EQUAL) reduce 570) ((T_MUL_EQUAL) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_OR_EQUAL) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_PLUS_EQUAL) reduce 570) ((T_POW) reduce 570) ((T_POW_EQUAL) reduce 570) ((T_SL) reduce 570) ((T_SL_EQUAL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) ((T_SR_EQUAL) reduce 570) ((T_XOR_EQUAL) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570) (("}") reduce 570)) 13605 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) (("=") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AND_EQUAL) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_COALESCE_EQUAL) reduce 569) ((T_CONCAT_EQUAL) reduce 569) ((T_DEC) reduce 569) ((T_DIV_EQUAL) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INC) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_MINUS_EQUAL) reduce 569) ((T_MOD_EQUAL) reduce 569) ((T_MUL_EQUAL) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_OR_EQUAL) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_PLUS_EQUAL) reduce 569) ((T_POW) reduce 569) ((T_POW_EQUAL) reduce 569) ((T_SL) reduce 569) ((T_SL_EQUAL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) ((T_SR_EQUAL) reduce 569) ((T_XOR_EQUAL) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569) (("}") reduce 569)) 13607 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_PIPE) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("|") reduce 522) (("}") reduce 522)) 13608 ((("%") reduce 447) (("*") reduce 447) (("+") reduce 447) (("-") reduce 447) ((".") reduce 447) (("/") reduce 447) (("<") reduce 447) ((">") reduce 447) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_BOOLEAN_AND) reduce 447) ((T_BOOLEAN_OR) reduce 447) ((T_COALESCE) reduce 447) ((T_DOUBLE_ARROW) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 447) ((T_IS_GREATER_OR_EQUAL) reduce 447) ((T_IS_IDENTICAL) reduce 447) ((T_IS_NOT_EQUAL) reduce 447) ((T_IS_NOT_IDENTICAL) reduce 447) ((T_IS_SMALLER_OR_EQUAL) reduce 447) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_PIPE) reduce 447) ((T_POW) shift) ((T_SL) reduce 447) ((T_SPACESHIP) reduce 447) ((T_SR) reduce 447) (("^") reduce 447) (("|") reduce 447) (("}") reduce 447)) 13609 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_PIPE) reduce 473) ((T_POW) shift) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("|") reduce 473) (("}") reduce 473)) 13610 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 483)) 13611 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_PIPE) reduce 469) ((T_POW) shift) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469) (("}") reduce 469)) 13612 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_PIPE) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("^") reduce 486) (("|") reduce 486) (("}") reduce 486)) 13614 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_PIPE) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("^") reduce 526) (("|") reduce 526) (("}") reduce 526)) 13616 ((("%") reduce 527) (("*") reduce 527) (("+") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_PIPE) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("^") reduce 527) (("|") reduce 527) (("}") reduce 527)) 13617 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_PIPE) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("^") reduce 525) (("|") reduce 525) (("}") reduce 525)) 13618 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 619)) 13619 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 618)) 13620 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505) (("}") reduce 505)) 13621 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 478)) 13622 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_PIPE) reduce 471) ((T_POW) shift) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("|") reduce 471) (("}") reduce 471)) 13624 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94) (("}") reduce 94)) 13625 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93) (("}") reduce 93)) 13626 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95) (("}") reduce 95)) 13627 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509) (("}") reduce 509)) 13628 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92) (("}") reduce 92)) 13629 ((("%") reduce 400) (("(") reduce 400) (("*") reduce 400) (("+") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 400) ((T_OBJECT_OPERATOR) reduce 400) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 400) ((T_PIPE) reduce 400) ((T_POW) reduce 400) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("[") reduce 400) (("^") reduce 400) (("|") reduce 400) (("}") reduce 400)) 13632 ((("%") reduce 511) (("(") reduce 511) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("|") reduce 511) (("}") reduce 511)) 13633 ((("%") reduce 402) (("(") shift) (("*") reduce 402) (("+") reduce 402) (("-") reduce 402) ((".") reduce 402) (("/") reduce 402) (("<") reduce 402) ((">") reduce 402) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_INSTANCEOF) reduce 402) ((T_IS_EQUAL) reduce 402) ((T_IS_GREATER_OR_EQUAL) reduce 402) ((T_IS_IDENTICAL) reduce 402) ((T_IS_NOT_EQUAL) reduce 402) ((T_IS_NOT_IDENTICAL) reduce 402) ((T_IS_SMALLER_OR_EQUAL) reduce 402) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_PIPE) reduce 402) ((T_POW) reduce 402) ((T_SL) reduce 402) ((T_SPACESHIP) reduce 402) ((T_SR) reduce 402) (("^") reduce 402) (("|") reduce 402) (("}") reduce 402)) 13634 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510) (("}") reduce 510)) 13635 ((("%") reduce 512) (("(") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("|") reduce 512) (("}") reduce 512)) 13636 ((("%") reduce 573) (("(") reduce 573) (("*") reduce 573) (("+") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("|") reduce 573) (("}") reduce 573)) 13642 ((("%") reduce 574) (("(") reduce 574) (("*") reduce 574) (("+") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("|") reduce 574) (("}") reduce 574)) 13643 ((("%") reduce 578) (("(") reduce 578) (("*") reduce 578) (("+") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("|") reduce 578) (("}") reduce 578)) 13644 ((("%") reduce 575) (("(") reduce 575) (("*") reduce 575) (("+") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("|") reduce 575) (("}") reduce 575)) 13645 ((("%") reduce 576) (("(") reduce 576) (("*") reduce 576) (("+") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("|") reduce 576) (("}") reduce 576)) 13646 ((("%") reduce 399) (("(") reduce 399) (("*") reduce 399) (("+") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_INSTANCEOF) reduce 399) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 399) ((T_OBJECT_OPERATOR) reduce 399) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 399) ((T_PIPE) reduce 399) ((T_POW) reduce 399) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("[") reduce 399) (("^") reduce 399) (("|") reduce 399) (("}") reduce 399)) 13648 ((("%") reduce 577) (("(") reduce 577) (("*") reduce 577) (("+") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("|") reduce 577) (("}") reduce 577)) 13649 ((("%") reduce 401) (("(") reduce 401) (("*") reduce 401) (("+") reduce 401) (("-") reduce 401) ((".") reduce 401) (("/") reduce 401) (("<") reduce 401) ((">") reduce 401) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_INSTANCEOF) reduce 401) ((T_IS_EQUAL) reduce 401) ((T_IS_GREATER_OR_EQUAL) reduce 401) ((T_IS_IDENTICAL) reduce 401) ((T_IS_NOT_EQUAL) reduce 401) ((T_IS_NOT_IDENTICAL) reduce 401) ((T_IS_SMALLER_OR_EQUAL) reduce 401) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 401) ((T_OBJECT_OPERATOR) reduce 401) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 401) ((T_PIPE) reduce 401) ((T_POW) reduce 401) ((T_SL) reduce 401) ((T_SPACESHIP) reduce 401) ((T_SR) reduce 401) (("[") reduce 401) (("^") reduce 401) (("|") reduce 401) (("}") reduce 401)) 13657 ((("%") reduce 398) (("(") reduce 398) (("*") reduce 398) (("+") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 398) ((T_OBJECT_OPERATOR) reduce 398) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 398) ((T_PIPE) reduce 398) ((T_POW) reduce 398) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("[") reduce 398) (("^") reduce 398) (("|") reduce 398) (("}") reduce 398)) 13659 ((("%") reduce 513) (("(") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("|") reduce 513) (("}") reduce 513)) 13665 ((("%") reduce 233) (("*") reduce 233) (("+") reduce 233) (("-") reduce 233) ((".") reduce 233) (("/") reduce 233) (("<") reduce 233) ((">") reduce 233) (("?") reduce 233) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_BOOLEAN_AND) reduce 233) ((T_BOOLEAN_OR) reduce 233) ((T_COALESCE) reduce 233) ((T_INSTANCEOF) reduce 233) ((T_IS_EQUAL) reduce 233) ((T_IS_GREATER_OR_EQUAL) reduce 233) ((T_IS_IDENTICAL) reduce 233) ((T_IS_NOT_EQUAL) reduce 233) ((T_IS_NOT_IDENTICAL) reduce 233) ((T_IS_SMALLER_OR_EQUAL) reduce 233) ((T_LOGICAL_AND) reduce 233) ((T_LOGICAL_OR) reduce 233) ((T_LOGICAL_XOR) reduce 233) ((T_PIPE) reduce 233) ((T_POW) reduce 233) ((T_SL) reduce 233) ((T_SPACESHIP) reduce 233) ((T_SR) reduce 233) (("^") reduce 233) (("|") reduce 233) (("}") reduce 233)) 13670 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 404)) 13674 ((("%") reduce 613) (("*") reduce 613) (("+") reduce 613) (("-") reduce 613) ((".") reduce 613) (("/") reduce 613) (("<") reduce 613) ((">") reduce 613) (("?") reduce 613) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_BOOLEAN_AND) reduce 613) ((T_BOOLEAN_OR) reduce 613) ((T_COALESCE) reduce 613) ((T_INSTANCEOF) reduce 613) ((T_IS_EQUAL) reduce 613) ((T_IS_GREATER_OR_EQUAL) reduce 613) ((T_IS_IDENTICAL) reduce 613) ((T_IS_NOT_EQUAL) reduce 613) ((T_IS_NOT_IDENTICAL) reduce 613) ((T_IS_SMALLER_OR_EQUAL) reduce 613) ((T_LOGICAL_AND) reduce 613) ((T_LOGICAL_OR) reduce 613) ((T_LOGICAL_XOR) reduce 613) ((T_PIPE) reduce 613) ((T_POW) reduce 613) ((T_SL) reduce 613) ((T_SPACESHIP) reduce 613) ((T_SR) reduce 613) (("^") reduce 613) (("|") reduce 613) (("}") reduce 613)) 13675 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_PIPE) reduce 467) ((T_POW) shift) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("|") reduce 467) (("}") reduce 467)) 13676 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 616)) 13677 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift) (("}") reduce 615)) 13678 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("[") reduce 548) (("^") reduce 424) (("|") reduce 424) (("}") reduce 424)) 13680 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_PIPE) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("^") reduce 518) (("|") reduce 518) (("}") reduce 518)) 13681 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_PIPE) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("|") reduce 474) (("}") reduce 474)) 13682 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("|") reduce 291) (("}") reduce 291)) 13686 ((("%") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("^") reduce 292) (("|") reduce 292) (("}") reduce 292)) 13687 ((("%") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("^") reduce 293) (("|") reduce 293) (("}") reduce 293)) 13690 ((("%") reduce 617) (("*") reduce 617) (("+") reduce 617) (("-") reduce 617) ((".") reduce 617) (("/") reduce 617) (("<") reduce 617) ((">") reduce 617) (("?") reduce 617) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_BOOLEAN_AND) reduce 617) ((T_BOOLEAN_OR) reduce 617) ((T_COALESCE) reduce 617) ((T_INSTANCEOF) reduce 617) ((T_IS_EQUAL) reduce 617) ((T_IS_GREATER_OR_EQUAL) reduce 617) ((T_IS_IDENTICAL) reduce 617) ((T_IS_NOT_EQUAL) reduce 617) ((T_IS_NOT_IDENTICAL) reduce 617) ((T_IS_SMALLER_OR_EQUAL) reduce 617) ((T_LOGICAL_AND) reduce 617) ((T_LOGICAL_OR) reduce 617) ((T_LOGICAL_XOR) reduce 617) ((T_PIPE) reduce 617) ((T_POW) reduce 617) ((T_SL) reduce 617) ((T_SPACESHIP) reduce 617) ((T_SR) reduce 617) (("^") reduce 617) (("|") reduce 617) (("}") reduce 617)) 13693 ((("%") reduce 614) (("*") reduce 614) (("+") reduce 614) (("-") reduce 614) ((".") reduce 614) (("/") reduce 614) (("<") reduce 614) ((">") reduce 614) (("?") reduce 614) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_BOOLEAN_AND) reduce 614) ((T_BOOLEAN_OR) reduce 614) ((T_COALESCE) reduce 614) ((T_INSTANCEOF) reduce 614) ((T_IS_EQUAL) reduce 614) ((T_IS_GREATER_OR_EQUAL) reduce 614) ((T_IS_IDENTICAL) reduce 614) ((T_IS_NOT_EQUAL) reduce 614) ((T_IS_NOT_IDENTICAL) reduce 614) ((T_IS_SMALLER_OR_EQUAL) reduce 614) ((T_LOGICAL_AND) reduce 614) ((T_LOGICAL_OR) reduce 614) ((T_LOGICAL_XOR) reduce 614) ((T_PIPE) reduce 614) ((T_POW) reduce 614) ((T_SL) reduce 614) ((T_SPACESHIP) reduce 614) ((T_SR) reduce 614) (("^") reduce 614) (("|") reduce 614) (("}") reduce 614)) 13694 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_PIPE) reduce 468) ((T_POW) shift) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468) (("}") reduce 468)) 13695 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("[") reduce 548) (("^") reduce 426) (("|") reduce 426) (("}") reduce 426)) 13697 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_PIPE) reduce 408) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("^") reduce 408) (("|") reduce 408) (("}") reduce 408)) 13698 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_PIPE) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("|") reduce 409) (("}") reduce 409)) 13699 ((("%") reduce 296) (("*") reduce 296) (("+") reduce 296) (("-") reduce 296) ((".") reduce 296) (("/") reduce 296) (("<") reduce 296) ((">") reduce 296) (("?") reduce 296) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_BOOLEAN_AND) reduce 296) ((T_BOOLEAN_OR) reduce 296) ((T_COALESCE) reduce 296) ((T_INSTANCEOF) reduce 296) ((T_IS_EQUAL) reduce 296) ((T_IS_GREATER_OR_EQUAL) reduce 296) ((T_IS_IDENTICAL) reduce 296) ((T_IS_NOT_EQUAL) reduce 296) ((T_IS_NOT_IDENTICAL) reduce 296) ((T_IS_SMALLER_OR_EQUAL) reduce 296) ((T_LOGICAL_AND) reduce 296) ((T_LOGICAL_OR) reduce 296) ((T_LOGICAL_XOR) reduce 296) ((T_PIPE) reduce 296) ((T_POW) reduce 296) ((T_SL) reduce 296) ((T_SPACESHIP) reduce 296) ((T_SR) reduce 296) (("^") reduce 296) (("|") reduce 296) (("}") reduce 296)) 13704 ((("%") reduce 297) (("*") reduce 297) (("+") reduce 297) (("-") reduce 297) ((".") reduce 297) (("/") reduce 297) (("<") reduce 297) ((">") reduce 297) (("?") reduce 297) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_BOOLEAN_AND) reduce 297) ((T_BOOLEAN_OR) reduce 297) ((T_COALESCE) reduce 297) ((T_INSTANCEOF) reduce 297) ((T_IS_EQUAL) reduce 297) ((T_IS_GREATER_OR_EQUAL) reduce 297) ((T_IS_IDENTICAL) reduce 297) ((T_IS_NOT_EQUAL) reduce 297) ((T_IS_NOT_IDENTICAL) reduce 297) ((T_IS_SMALLER_OR_EQUAL) reduce 297) ((T_LOGICAL_AND) reduce 297) ((T_LOGICAL_OR) reduce 297) ((T_LOGICAL_XOR) reduce 297) ((T_PIPE) reduce 297) ((T_POW) reduce 297) ((T_SL) reduce 297) ((T_SPACESHIP) reduce 297) ((T_SR) reduce 297) (("^") reduce 297) (("|") reduce 297) (("}") reduce 297)) 13705 ((("%") reduce 460) (("(") reduce 556) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PIPE) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("[") reduce 549) (("^") reduce 460) (("|") reduce 460) (("}") reduce 460)) 13707 ((("%") reduce 298) (("*") reduce 298) (("+") reduce 298) (("-") reduce 298) ((".") reduce 298) (("/") reduce 298) (("<") reduce 298) ((">") reduce 298) (("?") reduce 298) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_BOOLEAN_AND) reduce 298) ((T_BOOLEAN_OR) reduce 298) ((T_COALESCE) reduce 298) ((T_INSTANCEOF) reduce 298) ((T_IS_EQUAL) reduce 298) ((T_IS_GREATER_OR_EQUAL) reduce 298) ((T_IS_IDENTICAL) reduce 298) ((T_IS_NOT_EQUAL) reduce 298) ((T_IS_NOT_IDENTICAL) reduce 298) ((T_IS_SMALLER_OR_EQUAL) reduce 298) ((T_LOGICAL_AND) reduce 298) ((T_LOGICAL_OR) reduce 298) ((T_LOGICAL_XOR) reduce 298) ((T_PIPE) reduce 298) ((T_POW) reduce 298) ((T_SL) reduce 298) ((T_SPACESHIP) reduce 298) ((T_SR) reduce 298) (("^") reduce 298) (("|") reduce 298) (("}") reduce 298)) 13708 ((("%") reduce 299) (("*") reduce 299) (("+") reduce 299) (("-") reduce 299) ((".") reduce 299) (("/") reduce 299) (("<") reduce 299) ((">") reduce 299) (("?") reduce 299) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_BOOLEAN_AND) reduce 299) ((T_BOOLEAN_OR) reduce 299) ((T_COALESCE) reduce 299) ((T_INSTANCEOF) reduce 299) ((T_IS_EQUAL) reduce 299) ((T_IS_GREATER_OR_EQUAL) reduce 299) ((T_IS_IDENTICAL) reduce 299) ((T_IS_NOT_EQUAL) reduce 299) ((T_IS_NOT_IDENTICAL) reduce 299) ((T_IS_SMALLER_OR_EQUAL) reduce 299) ((T_LOGICAL_AND) reduce 299) ((T_LOGICAL_OR) reduce 299) ((T_LOGICAL_XOR) reduce 299) ((T_PIPE) reduce 299) ((T_POW) reduce 299) ((T_SL) reduce 299) ((T_SPACESHIP) reduce 299) ((T_SR) reduce 299) (("^") reduce 299) (("|") reduce 299) (("}") reduce 299)) 13709 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_PIPE) reduce 472) ((T_POW) shift) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("|") reduce 472) (("}") reduce 472)) 13710 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_PIPE) reduce 470) ((T_POW) shift) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("|") reduce 470) (("}") reduce 470)) 13713 ((("%") reduce 519) (("(") reduce 519) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PIPE) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("|") reduce 519) (("}") reduce 519)) 13714 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_PIPE) reduce 475) ((T_POW) shift) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("|") reduce 475) (("}") reduce 475)) 13715 ((("%") reduce 446) (("*") reduce 446) (("+") reduce 446) (("-") reduce 446) ((".") reduce 446) (("/") reduce 446) (("<") reduce 446) ((">") reduce 446) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_BOOLEAN_AND) reduce 446) ((T_BOOLEAN_OR) reduce 446) ((T_COALESCE) reduce 446) ((T_INSTANCEOF) reduce 446) ((T_IS_EQUAL) reduce 446) ((T_IS_GREATER_OR_EQUAL) reduce 446) ((T_IS_IDENTICAL) reduce 446) ((T_IS_NOT_EQUAL) reduce 446) ((T_IS_NOT_IDENTICAL) reduce 446) ((T_IS_SMALLER_OR_EQUAL) reduce 446) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_PIPE) reduce 446) ((T_POW) shift) ((T_SL) reduce 446) ((T_SPACESHIP) reduce 446) ((T_SR) reduce 446) (("^") reduce 446) (("|") reduce 446) (("}") reduce 446)) 13716 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_PIPE) reduce 445) ((T_POW) shift) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("|") reduce 445) (("}") reduce 445)) 13718 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) (("=") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AND_EQUAL) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_COALESCE_EQUAL) reduce 570) ((T_CONCAT_EQUAL) reduce 570) ((T_DEC) reduce 570) ((T_DIV_EQUAL) reduce 570) ((T_INC) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_MINUS_EQUAL) reduce 570) ((T_MOD_EQUAL) reduce 570) ((T_MUL_EQUAL) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_OR_EQUAL) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_PLUS_EQUAL) reduce 570) ((T_POW) reduce 570) ((T_POW_EQUAL) reduce 570) ((T_SL) reduce 570) ((T_SL_EQUAL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) ((T_SR_EQUAL) reduce 570) ((T_XOR_EQUAL) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570) (("}") reduce 570)) 13721 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) (("=") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AND_EQUAL) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_COALESCE_EQUAL) reduce 569) ((T_CONCAT_EQUAL) reduce 569) ((T_DEC) reduce 569) ((T_DIV_EQUAL) reduce 569) ((T_INC) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_MINUS_EQUAL) reduce 569) ((T_MOD_EQUAL) reduce 569) ((T_MUL_EQUAL) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_OR_EQUAL) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_PLUS_EQUAL) reduce 569) ((T_POW) reduce 569) ((T_POW_EQUAL) reduce 569) ((T_SL) reduce 569) ((T_SL_EQUAL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) ((T_SR_EQUAL) reduce 569) ((T_XOR_EQUAL) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569) (("}") reduce 569)) 13723 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_PIPE) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("|") reduce 522) (("}") reduce 522)) 13724 ((("%") reduce 447) (("*") reduce 447) (("+") reduce 447) (("-") reduce 447) ((".") reduce 447) (("/") reduce 447) (("<") reduce 447) ((">") reduce 447) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_BOOLEAN_AND) reduce 447) ((T_BOOLEAN_OR) reduce 447) ((T_COALESCE) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 447) ((T_IS_GREATER_OR_EQUAL) reduce 447) ((T_IS_IDENTICAL) reduce 447) ((T_IS_NOT_EQUAL) reduce 447) ((T_IS_NOT_IDENTICAL) reduce 447) ((T_IS_SMALLER_OR_EQUAL) reduce 447) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_PIPE) reduce 447) ((T_POW) shift) ((T_SL) reduce 447) ((T_SPACESHIP) reduce 447) ((T_SR) reduce 447) (("^") reduce 447) (("|") reduce 447) (("}") reduce 447)) 13725 ((("%") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("]") reduce 583) (("^") reduce 583) (("|") reduce 583)) 14409 ((("%") reduce 576) (("*") reduce 576) (("+") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("]") reduce 576) (("^") reduce 576) (("|") reduce 576)) 14411 ((("%") reduce 577) (("*") reduce 577) (("+") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("]") reduce 577) (("^") reduce 577) (("|") reduce 577)) 14413 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("]") reduce 513) (("^") reduce 513) (("|") reduce 513)) 14414 ((("%") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("]") reduce 570) (("^") reduce 570) (("|") reduce 570)) 14417 ((("%") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("]") reduce 569) (("^") reduce 569) (("|") reduce 569)) 14418 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 465) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 465) (("^") shift) (("|") shift)) 14419 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 427) (("^") shift) (("|") shift)) 14420 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 428) (("^") shift) (("|") shift)) 14421 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 433) (("^") reduce 433) (("|") reduce 433)) 14422 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 434) (("^") reduce 434) (("|") reduce 434)) 14426 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 463) (("^") shift) (("|") shift)) 14427 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 464) (("^") shift) (("|") shift)) 14428 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 456) ((T_SR) shift) (("]") reduce 456) (("^") reduce 456) (("|") reduce 456)) 14429 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 454) ((T_SR) shift) (("]") reduce 454) (("^") reduce 454) (("|") reduce 454)) 14430 ((("%") reduce 441) (("*") reduce 441) (("+") reduce 441) (("-") reduce 441) ((".") reduce 441) (("/") reduce 441) (("<") reduce 441) ((">") reduce 441) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 441) ((T_IS_GREATER_OR_EQUAL) reduce 441) ((T_IS_IDENTICAL) reduce 441) ((T_IS_NOT_EQUAL) reduce 441) ((T_IS_NOT_IDENTICAL) reduce 441) ((T_IS_SMALLER_OR_EQUAL) reduce 441) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_PIPE) reduce 441) ((T_POW) shift) ((T_SL) reduce 441) ((T_SPACESHIP) reduce 441) ((T_SR) reduce 441) (("]") reduce 441) (("^") reduce 441) (("|") reduce 441)) 14431 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 436) (("/") shift) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_PIPE) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 436) ((T_SR) shift) (("]") reduce 436) (("^") reduce 436) (("|") reduce 436)) 14432 ((("%") shift) (("*") shift) (("+") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") shift) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_PIPE) reduce 438) ((T_POW) shift) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("]") reduce 438) (("^") reduce 438) (("|") reduce 438)) 14433 ((("%") shift) (("*") shift) (("+") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") shift) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_PIPE) reduce 437) ((T_POW) shift) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("]") reduce 437) (("^") reduce 437) (("|") reduce 437)) 14434 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_PIPE) reduce 439) ((T_POW) shift) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("]") reduce 439) (("^") reduce 439) (("|") reduce 439)) 14435 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_PIPE) reduce 442) ((T_POW) shift) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("]") reduce 442) (("^") reduce 442) (("|") reduce 442)) 14437 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_PIPE) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("]") reduce 74) (("^") reduce 74) (("|") reduce 74)) 14438 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_PIPE) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("]") reduce 52) (("^") reduce 52) (("|") reduce 52)) 14439 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_PIPE) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("]") reduce 27) (("^") reduce 27) (("|") reduce 27)) 14440 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_PIPE) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("]") reduce 51) (("^") reduce 51) (("|") reduce 51)) 14441 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_PIPE) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("]") reduce 53) (("^") reduce 53) (("|") reduce 53)) 14442 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_PIPE) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("]") reduce 49) (("^") reduce 49) (("|") reduce 49)) 14443 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_PIPE) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("]") reduce 29) (("^") reduce 29) (("|") reduce 29)) 14444 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_PIPE) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("]") reduce 59) (("^") reduce 59) (("|") reduce 59)) 14445 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_PIPE) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("]") reduce 60) (("^") reduce 60) (("|") reduce 60)) 14446 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_PIPE) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("]") reduce 11) (("^") reduce 11) (("|") reduce 11)) 14447 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_PIPE) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("]") reduce 42) (("^") reduce 42) (("|") reduce 42)) 14448 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_PIPE) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("]") reduce 39) (("^") reduce 39) (("|") reduce 39)) 14449 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_PIPE) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("]") reduce 25) (("^") reduce 25) (("|") reduce 25)) 14450 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_PIPE) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("]") reduce 50) (("^") reduce 50) (("|") reduce 50)) 14451 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_PIPE) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("]") reduce 66) (("^") reduce 66) (("|") reduce 66)) 14452 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_PIPE) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("]") reduce 18) (("^") reduce 18) (("|") reduce 18)) 14453 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_PIPE) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("]") reduce 17) (("^") reduce 17) (("|") reduce 17)) 14454 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_PIPE) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("]") reduce 15) (("^") reduce 15) (("|") reduce 15)) 14455 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_PIPE) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("]") reduce 14) (("^") reduce 14) (("|") reduce 14)) 14456 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_PIPE) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("]") reduce 38) (("^") reduce 38) (("|") reduce 38)) 14457 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_PIPE) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("]") reduce 26) (("^") reduce 26) (("|") reduce 26)) 14458 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_PIPE) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("]") reduce 22) (("^") reduce 22) (("|") reduce 22)) 14459 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_PIPE) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("]") reduce 24) (("^") reduce 24) (("|") reduce 24)) 14460 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_PIPE) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("]") reduce 16) (("^") reduce 16) (("|") reduce 16)) 14461 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_PIPE) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("]") reduce 48) (("^") reduce 48) (("|") reduce 48)) 14462 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_PIPE) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("]") reduce 20) (("^") reduce 20) (("|") reduce 20)) 14463 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_PIPE) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("]") reduce 70) (("^") reduce 70) (("|") reduce 70)) 14464 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_PIPE) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("]") reduce 3) (("^") reduce 3) (("|") reduce 3)) 14465 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_PIPE) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("]") reduce 12) (("^") reduce 12) (("|") reduce 12)) 14466 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_PIPE) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("]") reduce 54) (("^") reduce 54) (("|") reduce 54)) 14467 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_PIPE) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("]") reduce 65) (("^") reduce 65) (("|") reduce 65)) 14468 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_PIPE) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("]") reduce 75) (("^") reduce 75) (("|") reduce 75)) 14469 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_PIPE) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("]") reduce 30) (("^") reduce 30) (("|") reduce 30)) 14470 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_PIPE) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("]") reduce 68) (("^") reduce 68) (("|") reduce 68)) 14471 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_PIPE) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("]") reduce 21) (("^") reduce 21) (("|") reduce 21)) 14472 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_PIPE) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("]") reduce 23) (("^") reduce 23) (("|") reduce 23)) 14473 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_PIPE) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("]") reduce 41) (("^") reduce 41) (("|") reduce 41)) 14474 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_PIPE) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("]") reduce 62) (("^") reduce 62) (("|") reduce 62)) 14475 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_PIPE) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("]") reduce 34) (("^") reduce 34) (("|") reduce 34)) 14476 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_PIPE) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("]") reduce 40) (("^") reduce 40) (("|") reduce 40)) 14477 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_PIPE) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("]") reduce 13) (("^") reduce 13) (("|") reduce 13)) 14478 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_PIPE) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("]") reduce 55) (("^") reduce 55) (("|") reduce 55)) 14479 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_PIPE) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("]") reduce 1) (("^") reduce 1) (("|") reduce 1)) 14480 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_PIPE) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("]") reduce 2) (("^") reduce 2) (("|") reduce 2)) 14481 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_PIPE) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("]") reduce 9) (("^") reduce 9) (("|") reduce 9)) 14482 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_PIPE) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("]") reduce 33) (("^") reduce 33) (("|") reduce 33)) 14483 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_PIPE) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("]") reduce 58) (("^") reduce 58) (("|") reduce 58)) 14484 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_PIPE) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("]") reduce 37) (("^") reduce 37) (("|") reduce 37)) 14485 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_PIPE) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("]") reduce 64) (("^") reduce 64) (("|") reduce 64)) 14486 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_PIPE) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("]") reduce 46) (("^") reduce 46) (("|") reduce 46)) 14487 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_PIPE) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("]") reduce 8) (("^") reduce 8) (("|") reduce 8)) 14488 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_PIPE) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("]") reduce 6) (("^") reduce 6) (("|") reduce 6)) 14489 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_PIPE) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("]") reduce 7) (("^") reduce 7) (("|") reduce 7)) 14490 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_PIPE) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("]") reduce 69) (("^") reduce 69) (("|") reduce 69)) 14491 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_PIPE) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("]") reduce 63) (("^") reduce 63) (("|") reduce 63)) 14492 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_PIPE) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("]") reduce 56) (("^") reduce 56) (("|") reduce 56)) 14493 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_PIPE) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("]") reduce 10) (("^") reduce 10) (("|") reduce 10)) 14494 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_PIPE) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("]") reduce 67) (("^") reduce 67) (("|") reduce 67)) 14495 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_PIPE) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("]") reduce 44) (("^") reduce 44) (("|") reduce 44)) 14496 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_PIPE) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("]") reduce 76) (("^") reduce 76) (("|") reduce 76)) 14497 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_PIPE) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("]") reduce 71) (("^") reduce 71) (("|") reduce 71)) 14498 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_PIPE) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("]") reduce 77) (("^") reduce 77) (("|") reduce 77)) 14499 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_PIPE) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("]") reduce 78) (("^") reduce 78) (("|") reduce 78)) 14500 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_PIPE) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("]") reduce 79) (("^") reduce 79) (("|") reduce 79)) 14501 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_PIPE) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("]") reduce 4) (("^") reduce 4) (("|") reduce 4)) 14502 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_PIPE) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("]") reduce 5) (("^") reduce 5) (("|") reduce 5)) 14503 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_PIPE) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("]") reduce 43) (("^") reduce 43) (("|") reduce 43)) 14504 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_PIPE) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("]") reduce 73) (("^") reduce 73) (("|") reduce 73)) 14505 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_PIPE) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("]") reduce 82) (("^") reduce 82) (("|") reduce 82)) 14506 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_PIPE) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("]") reduce 47) (("^") reduce 47) (("|") reduce 47)) 14507 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_PIPE) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("]") reduce 31) (("^") reduce 31) (("|") reduce 31)) 14508 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_PIPE) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("]") reduce 57) (("^") reduce 57) (("|") reduce 57)) 14509 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_PIPE) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("]") reduce 61) (("^") reduce 61) (("|") reduce 61)) 14510 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_PIPE) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("]") reduce 28) (("^") reduce 28) (("|") reduce 28)) 14511 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_PIPE) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("]") reduce 36) (("^") reduce 36) (("|") reduce 36)) 14512 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_PIPE) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("]") reduce 32) (("^") reduce 32) (("|") reduce 32)) 14513 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_PIPE) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("]") reduce 35) (("^") reduce 35) (("|") reduce 35)) 14514 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_PIPE) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("]") reduce 19) (("^") reduce 19) (("|") reduce 19)) 14515 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_PIPE) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("]") reduce 45) (("^") reduce 45) (("|") reduce 45)) 14516 ((("%") reduce 542) (("(") reduce 579) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PIPE) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("]") reduce 542) (("^") reduce 542) (("|") reduce 542)) 14518 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_PIPE) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("]") reduce 72) (("^") reduce 72) (("|") reduce 72)) 14519 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_PIPE) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("]") reduce 83) (("^") reduce 83) (("|") reduce 83)) 14520 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) (("<") reduce 572) (("=") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AND_EQUAL) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_COALESCE_EQUAL) reduce 572) ((T_CONCAT_EQUAL) reduce 572) ((T_DEC) reduce 572) ((T_DIV_EQUAL) reduce 572) ((T_INC) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_MINUS_EQUAL) reduce 572) ((T_MOD_EQUAL) reduce 572) ((T_MUL_EQUAL) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_OR_EQUAL) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_PLUS_EQUAL) reduce 572) ((T_POW) reduce 572) ((T_POW_EQUAL) reduce 572) ((T_SL) reduce 572) ((T_SL_EQUAL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) ((T_SR_EQUAL) reduce 572) ((T_XOR_EQUAL) reduce 572) (("[") reduce 572) (("]") reduce 572) (("^") reduce 572) (("|") reduce 572)) 14523 ((("%") reduce 544) (("(") reduce 580) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PIPE) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("]") reduce 544) (("^") reduce 544) (("|") reduce 544)) 14525 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("]") reduce 507) (("^") reduce 507) (("|") reduce 507)) 14526 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("]") reduce 291) (("^") reduce 291) (("|") reduce 291)) 14530 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) (("=") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AND_EQUAL) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_COALESCE_EQUAL) reduce 292) ((T_CONCAT_EQUAL) reduce 292) ((T_DEC) reduce 292) ((T_DIV_EQUAL) reduce 292) ((T_INC) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_MINUS_EQUAL) reduce 292) ((T_MOD_EQUAL) reduce 292) ((T_MUL_EQUAL) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_OR_EQUAL) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_PLUS_EQUAL) reduce 292) ((T_POW) reduce 292) ((T_POW_EQUAL) reduce 292) ((T_SL) reduce 292) ((T_SL_EQUAL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) ((T_SR_EQUAL) reduce 292) ((T_XOR_EQUAL) reduce 292) (("[") reduce 292) (("]") reduce 292) (("^") reduce 292) (("|") reduce 292)) 14531 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) (("=") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AND_EQUAL) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_COALESCE_EQUAL) reduce 293) ((T_CONCAT_EQUAL) reduce 293) ((T_DEC) reduce 293) ((T_DIV_EQUAL) reduce 293) ((T_INC) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_MINUS_EQUAL) reduce 293) ((T_MOD_EQUAL) reduce 293) ((T_MUL_EQUAL) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_OR_EQUAL) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_PLUS_EQUAL) reduce 293) ((T_POW) reduce 293) ((T_POW_EQUAL) reduce 293) ((T_SL) reduce 293) ((T_SL_EQUAL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) ((T_SR_EQUAL) reduce 293) ((T_XOR_EQUAL) reduce 293) (("[") reduce 293) (("]") reduce 293) (("^") reduce 293) (("|") reduce 293)) 14536 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_PIPE) reduce 425) ((T_POW) reduce 425) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("]") reduce 425) (("^") reduce 425) (("|") reduce 425)) 14538 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_PIPE) reduce 423) ((T_POW) reduce 423) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("]") reduce 423) (("^") reduce 423) (("|") reduce 423)) 14548 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 419) (("^") shift) (("|") shift)) 14549 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 421) (("^") shift) (("|") shift)) 14550 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 420) (("^") shift) (("|") shift)) 14551 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 413) (("^") shift) (("|") shift)) 14552 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 410) (("^") shift) (("|") shift)) 14553 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 418) (("^") shift) (("|") shift)) 14554 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 412) (("^") shift) (("|") shift)) 14555 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 416) (("^") shift) (("|") shift)) 14556 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 411) (("^") shift) (("|") shift)) 14557 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 414) (("^") shift) (("|") shift)) 14558 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 415) (("^") shift) (("|") shift)) 14559 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 422) (("^") shift) (("|") shift)) 14560 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 417) (("^") shift) (("|") shift)) 14562 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 406) (("^") shift) (("|") shift)) 14565 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("]") reduce 568) (("^") reduce 568) (("|") reduce 568)) 14568 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) (("<") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_POW) reduce 564) ((T_SL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) (("[") reduce 564) (("]") reduce 564) (("^") reduce 564) (("|") reduce 564)) 14570 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) (("<") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_POW) reduce 563) ((T_SL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) (("[") reduce 563) (("]") reduce 563) (("^") reduce 563) (("|") reduce 563)) 14572 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("]") reduce 559) (("^") reduce 559) (("|") reduce 559)) 14573 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("]") reduce 565) (("^") reduce 565) (("|") reduce 565)) 14574 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("[") reduce 548) (("]") reduce 407) (("^") reduce 407) (("|") reduce 407)) 14578 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) (("<") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_POW) reduce 572) ((T_SL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) (("[") reduce 572) (("]") reduce 572) (("^") reduce 572) (("|") reduce 572)) 14580 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("]") reduce 507) (("^") reduce 507) (("|") reduce 507)) 14581 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("]") reduce 291) (("^") reduce 291) (("|") reduce 291)) 14585 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("[") reduce 292) (("]") reduce 292) (("^") reduce 292) (("|") reduce 292)) 14586 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("[") reduce 293) (("]") reduce 293) (("^") reduce 293) (("|") reduce 293)) 14587 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("]") reduce 504) (("^") reduce 504) (("|") reduce 504)) 14590 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) (("<") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_POW) reduce 571) ((T_SL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) (("[") reduce 571) (("]") reduce 571) (("^") reduce 571) (("|") reduce 571)) 14591 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("]") reduce 506) (("^") reduce 506) (("|") reduce 506)) 14592 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("]") reduce 508) (("^") reduce 508) (("|") reduce 508)) 14597 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("]") reduce 560) (("^") reduce 560) (("|") reduce 560)) 14598 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("]") reduce 582) (("^") reduce 582) (("|") reduce 582)) 14599 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("]") reduce 566) (("^") reduce 566) (("|") reduce 566)) 14600 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("]") reduce 584) (("^") reduce 584) (("|") reduce 584)) 14603 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("]") reduce 583) (("^") reduce 583) (("|") reduce 583)) 14604 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("]") reduce 561) (("^") reduce 561) (("|") reduce 561)) 14605 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("]") reduce 567) (("^") reduce 567) (("|") reduce 567)) 14606 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("[") reduce 562) (("]") reduce 562) (("^") reduce 562) (("|") reduce 562)) 14607 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("]") reduce 505) (("^") reduce 505) (("|") reduce 505)) 14608 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("]") reduce 570) (("^") reduce 570) (("|") reduce 570)) 14611 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("]") reduce 569) (("^") reduce 569) (("|") reduce 569)) 14612 ((("%") reduce 574) ((")") reduce 574) (("*") reduce 574) (("+") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("|") reduce 574)) 14613 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) (("=") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AND_EQUAL) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_COALESCE_EQUAL) reduce 504) ((T_CONCAT_EQUAL) reduce 504) ((T_DEC) reduce 504) ((T_DIV_EQUAL) reduce 504) ((T_INC) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_MINUS_EQUAL) reduce 504) ((T_MOD_EQUAL) reduce 504) ((T_MUL_EQUAL) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_OR_EQUAL) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_PLUS_EQUAL) reduce 504) ((T_POW) reduce 504) ((T_POW_EQUAL) reduce 504) ((T_SL) reduce 504) ((T_SL_EQUAL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) ((T_SR_EQUAL) reduce 504) ((T_XOR_EQUAL) reduce 504) (("[") reduce 504) (("]") reduce 504) (("^") reduce 504) (("|") reduce 504)) 14624 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PIPE) reduce 494) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("]") reduce 494) (("^") reduce 494) (("|") reduce 494)) 14625 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_PIPE) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("]") reduce 489) (("^") reduce 489) (("|") reduce 489)) 14698 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 494) (("^") shift) (("|") shift)) 15014 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_PIPE) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("]") reduce 490) (("^") reduce 490) (("|") reduce 490)) 15017 ((("%") reduce 541) (("(") reduce 579) (("*") reduce 541) (("+") reduce 541) (("-") reduce 541) ((".") reduce 541) (("/") reduce 541) (("<") reduce 541) ((">") reduce 541) (("?") reduce 541) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_BOOLEAN_AND) reduce 541) ((T_BOOLEAN_OR) reduce 541) ((T_COALESCE) reduce 541) ((T_INSTANCEOF) reduce 541) ((T_IS_EQUAL) reduce 541) ((T_IS_GREATER_OR_EQUAL) reduce 541) ((T_IS_IDENTICAL) reduce 541) ((T_IS_NOT_EQUAL) reduce 541) ((T_IS_NOT_IDENTICAL) reduce 541) ((T_IS_SMALLER_OR_EQUAL) reduce 541) ((T_LOGICAL_AND) reduce 541) ((T_LOGICAL_OR) reduce 541) ((T_LOGICAL_XOR) reduce 541) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 541) ((T_OBJECT_OPERATOR) reduce 541) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 541) ((T_PIPE) reduce 541) ((T_POW) reduce 541) ((T_SL) reduce 541) ((T_SPACESHIP) reduce 541) ((T_SR) reduce 541) (("[") reduce 541) (("]") reduce 541) (("^") reduce 541) (("|") reduce 541)) 15019 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) (("<") reduce 571) (("=") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AND_EQUAL) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_COALESCE_EQUAL) reduce 571) ((T_CONCAT_EQUAL) reduce 571) ((T_DEC) reduce 571) ((T_DIV_EQUAL) reduce 571) ((T_INC) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_MINUS_EQUAL) reduce 571) ((T_MOD_EQUAL) reduce 571) ((T_MUL_EQUAL) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_OR_EQUAL) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_PLUS_EQUAL) reduce 571) ((T_POW) reduce 571) ((T_POW_EQUAL) reduce 571) ((T_SL) reduce 571) ((T_SL_EQUAL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) ((T_SR_EQUAL) reduce 571) ((T_XOR_EQUAL) reduce 571) (("[") reduce 571) (("]") reduce 571) (("^") reduce 571) (("|") reduce 571)) 15022 ((("%") reduce 543) (("(") reduce 580) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PIPE) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("]") reduce 543) (("^") reduce 543) (("|") reduce 543)) 15023 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("]") reduce 506) (("^") reduce 506) (("|") reduce 506)) 15024 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("]") reduce 508) (("^") reduce 508) (("|") reduce 508)) 15026 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_PIPE) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("]") reduce 485) (("^") reduce 485) (("|") reduce 485)) 15027 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_PIPE) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("]") reduce 487) (("^") reduce 487) (("|") reduce 487)) 15032 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) (("=") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AND_EQUAL) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_COALESCE_EQUAL) reduce 560) ((T_CONCAT_EQUAL) reduce 560) ((T_DEC) reduce 560) ((T_DIV_EQUAL) reduce 560) ((T_INC) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_MINUS_EQUAL) reduce 560) ((T_MOD_EQUAL) reduce 560) ((T_MUL_EQUAL) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_OR_EQUAL) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_PLUS_EQUAL) reduce 560) ((T_POW) reduce 560) ((T_POW_EQUAL) reduce 560) ((T_SL) reduce 560) ((T_SL_EQUAL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) ((T_SR_EQUAL) reduce 560) ((T_XOR_EQUAL) reduce 560) (("[") reduce 560) (("]") reduce 560) (("^") reduce 560) (("|") reduce 560)) 15033 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) (("=") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AND_EQUAL) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_COALESCE_EQUAL) reduce 582) ((T_CONCAT_EQUAL) reduce 582) ((T_DEC) reduce 582) ((T_DIV_EQUAL) reduce 582) ((T_INC) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_MINUS_EQUAL) reduce 582) ((T_MOD_EQUAL) reduce 582) ((T_MUL_EQUAL) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_OR_EQUAL) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_PLUS_EQUAL) reduce 582) ((T_POW) reduce 582) ((T_POW_EQUAL) reduce 582) ((T_SL) reduce 582) ((T_SL_EQUAL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) ((T_SR_EQUAL) reduce 582) ((T_XOR_EQUAL) reduce 582) (("[") reduce 582) (("]") reduce 582) (("^") reduce 582) (("|") reduce 582)) 15034 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("]") reduce 566) (("^") reduce 566) (("|") reduce 566)) 15035 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) (("=") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AND_EQUAL) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_COALESCE_EQUAL) reduce 584) ((T_CONCAT_EQUAL) reduce 584) ((T_DEC) reduce 584) ((T_DIV_EQUAL) reduce 584) ((T_INC) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_MINUS_EQUAL) reduce 584) ((T_MOD_EQUAL) reduce 584) ((T_MUL_EQUAL) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_OR_EQUAL) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_PLUS_EQUAL) reduce 584) ((T_POW) reduce 584) ((T_POW_EQUAL) reduce 584) ((T_SL) reduce 584) ((T_SL_EQUAL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) ((T_SR_EQUAL) reduce 584) ((T_XOR_EQUAL) reduce 584) (("[") reduce 584) (("]") reduce 584) (("^") reduce 584) (("|") reduce 584)) 15038 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) (("=") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AND_EQUAL) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_COALESCE_EQUAL) reduce 583) ((T_CONCAT_EQUAL) reduce 583) ((T_DEC) reduce 583) ((T_DIV_EQUAL) reduce 583) ((T_INC) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_MINUS_EQUAL) reduce 583) ((T_MOD_EQUAL) reduce 583) ((T_MUL_EQUAL) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_OR_EQUAL) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_PLUS_EQUAL) reduce 583) ((T_POW) reduce 583) ((T_POW_EQUAL) reduce 583) ((T_SL) reduce 583) ((T_SL_EQUAL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) ((T_SR_EQUAL) reduce 583) ((T_XOR_EQUAL) reduce 583) (("[") reduce 583) (("]") reduce 583) (("^") reduce 583) (("|") reduce 583)) 15039 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) (("=") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AND_EQUAL) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_COALESCE_EQUAL) reduce 561) ((T_CONCAT_EQUAL) reduce 561) ((T_DEC) reduce 561) ((T_DIV_EQUAL) reduce 561) ((T_INC) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_MINUS_EQUAL) reduce 561) ((T_MOD_EQUAL) reduce 561) ((T_MUL_EQUAL) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_OR_EQUAL) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_PLUS_EQUAL) reduce 561) ((T_POW) reduce 561) ((T_POW_EQUAL) reduce 561) ((T_SL) reduce 561) ((T_SL_EQUAL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) ((T_SR_EQUAL) reduce 561) ((T_XOR_EQUAL) reduce 561) (("[") reduce 561) (("]") reduce 561) (("^") reduce 561) (("|") reduce 561)) 15040 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("]") reduce 567) (("^") reduce 567) (("|") reduce 567)) 15041 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) (("=") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AND_EQUAL) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_COALESCE_EQUAL) reduce 562) ((T_CONCAT_EQUAL) reduce 562) ((T_DEC) reduce 562) ((T_DIV_EQUAL) reduce 562) ((T_INC) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_MINUS_EQUAL) reduce 562) ((T_MOD_EQUAL) reduce 562) ((T_MUL_EQUAL) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_OR_EQUAL) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_PLUS_EQUAL) reduce 562) ((T_POW) reduce 562) ((T_POW_EQUAL) reduce 562) ((T_SL) reduce 562) ((T_SL_EQUAL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) ((T_SR_EQUAL) reduce 562) ((T_XOR_EQUAL) reduce 562) (("[") reduce 562) (("]") reduce 562) (("^") reduce 562) (("|") reduce 562)) 15043 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_PIPE) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("]") reduce 477) (("^") reduce 477) (("|") reduce 477)) 15045 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) (("=") shift) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_PIPE) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("]") reduce 520) (("^") reduce 520) (("|") reduce 520)) 15047 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 405) (("^") shift) (("|") shift)) 15048 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 482) (("^") shift) (("|") shift)) 15059 ((("%") reduce 540) (("*") reduce 540) (("+") reduce 540) (("-") reduce 540) ((".") reduce 540) (("/") reduce 540) (("<") reduce 540) ((">") reduce 540) (("?") reduce 540) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_BOOLEAN_AND) reduce 540) ((T_BOOLEAN_OR) reduce 540) ((T_COALESCE) reduce 540) ((T_DOUBLE_ARROW) reduce 540) ((T_INSTANCEOF) reduce 540) ((T_IS_EQUAL) reduce 540) ((T_IS_GREATER_OR_EQUAL) reduce 540) ((T_IS_IDENTICAL) reduce 540) ((T_IS_NOT_EQUAL) reduce 540) ((T_IS_NOT_IDENTICAL) reduce 540) ((T_IS_SMALLER_OR_EQUAL) reduce 540) ((T_LOGICAL_AND) reduce 540) ((T_LOGICAL_OR) reduce 540) ((T_LOGICAL_XOR) reduce 540) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 540) ((T_OBJECT_OPERATOR) reduce 540) ((T_PIPE) reduce 540) ((T_POW) reduce 540) ((T_SL) reduce 540) ((T_SPACESHIP) reduce 540) ((T_SR) reduce 540) (("[") reduce 540) (("]") reduce 540) (("^") reduce 540) (("|") reduce 540)) 15061 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_PIPE) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("]") reduce 521) (("^") reduce 521) (("|") reduce 521)) 15063 ((("%") reduce 534) (("*") reduce 534) (("+") reduce 534) (("-") reduce 534) ((".") reduce 534) (("/") reduce 534) (("<") reduce 534) ((">") reduce 534) (("?") reduce 534) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_BOOLEAN_AND) reduce 534) ((T_BOOLEAN_OR) reduce 534) ((T_COALESCE) reduce 534) ((T_DOUBLE_ARROW) reduce 534) ((T_INSTANCEOF) reduce 534) ((T_IS_EQUAL) reduce 534) ((T_IS_GREATER_OR_EQUAL) reduce 534) ((T_IS_IDENTICAL) reduce 534) ((T_IS_NOT_EQUAL) reduce 534) ((T_IS_NOT_IDENTICAL) reduce 534) ((T_IS_SMALLER_OR_EQUAL) reduce 534) ((T_LOGICAL_AND) reduce 534) ((T_LOGICAL_OR) reduce 534) ((T_LOGICAL_XOR) reduce 534) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PIPE) reduce 534) ((T_POW) reduce 534) ((T_SL) reduce 534) ((T_SPACESHIP) reduce 534) ((T_SR) reduce 534) (("[") reduce 534) (("]") reduce 534) (("^") reduce 534) (("|") reduce 534)) 15064 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_PIPE) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("]") reduce 524) (("^") reduce 524) (("|") reduce 524)) 15068 ((("%") reduce 517) (("(") shift) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_PIPE) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("]") reduce 517) (("^") reduce 517) (("|") reduce 517)) 15069 ((("%") reduce 533) (("*") reduce 533) (("+") reduce 533) (("-") reduce 533) ((".") reduce 533) (("/") reduce 533) (("<") reduce 533) ((">") reduce 533) (("?") reduce 533) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_BOOLEAN_AND) reduce 533) ((T_BOOLEAN_OR) reduce 533) ((T_COALESCE) reduce 533) ((T_DOUBLE_ARROW) reduce 533) ((T_INSTANCEOF) reduce 533) ((T_IS_EQUAL) reduce 533) ((T_IS_GREATER_OR_EQUAL) reduce 533) ((T_IS_IDENTICAL) reduce 533) ((T_IS_NOT_EQUAL) reduce 533) ((T_IS_NOT_IDENTICAL) reduce 533) ((T_IS_SMALLER_OR_EQUAL) reduce 533) ((T_LOGICAL_AND) reduce 533) ((T_LOGICAL_OR) reduce 533) ((T_LOGICAL_XOR) reduce 533) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PIPE) reduce 533) ((T_POW) reduce 533) ((T_SL) reduce 533) ((T_SPACESHIP) reduce 533) ((T_SR) reduce 533) (("[") reduce 533) (("]") reduce 533) (("^") reduce 533) (("|") reduce 533)) 15070 ((("%") reduce 537) (("*") reduce 537) (("+") reduce 537) (("-") reduce 537) ((".") reduce 537) (("/") reduce 537) (("<") reduce 537) ((">") reduce 537) (("?") reduce 537) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_BOOLEAN_AND) reduce 537) ((T_BOOLEAN_OR) reduce 537) ((T_COALESCE) reduce 537) ((T_DOUBLE_ARROW) reduce 537) ((T_INSTANCEOF) reduce 537) ((T_IS_EQUAL) reduce 537) ((T_IS_GREATER_OR_EQUAL) reduce 537) ((T_IS_IDENTICAL) reduce 537) ((T_IS_NOT_EQUAL) reduce 537) ((T_IS_NOT_IDENTICAL) reduce 537) ((T_IS_SMALLER_OR_EQUAL) reduce 537) ((T_LOGICAL_AND) reduce 537) ((T_LOGICAL_OR) reduce 537) ((T_LOGICAL_XOR) reduce 537) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_PIPE) reduce 537) ((T_POW) reduce 537) ((T_SL) reduce 537) ((T_SPACESHIP) reduce 537) ((T_SR) reduce 537) (("[") reduce 537) (("]") reduce 537) (("^") reduce 537) (("|") reduce 537)) 15076 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PIPE) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("]") reduce 532) (("^") reduce 532) (("|") reduce 532)) 15078 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_PIPE) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("]") reduce 523) (("^") reduce 523) (("|") reduce 523)) 15080 ((("%") reduce 536) (("*") reduce 536) (("+") reduce 536) (("-") reduce 536) ((".") reduce 536) (("/") reduce 536) (("<") reduce 536) ((">") reduce 536) (("?") reduce 536) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_BOOLEAN_AND) reduce 536) ((T_BOOLEAN_OR) reduce 536) ((T_COALESCE) reduce 536) ((T_DOUBLE_ARROW) reduce 536) ((T_INSTANCEOF) reduce 536) ((T_IS_EQUAL) reduce 536) ((T_IS_GREATER_OR_EQUAL) reduce 536) ((T_IS_IDENTICAL) reduce 536) ((T_IS_NOT_EQUAL) reduce 536) ((T_IS_NOT_IDENTICAL) reduce 536) ((T_IS_SMALLER_OR_EQUAL) reduce 536) ((T_LOGICAL_AND) reduce 536) ((T_LOGICAL_OR) reduce 536) ((T_LOGICAL_XOR) reduce 536) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 536) ((T_OBJECT_OPERATOR) reduce 536) ((T_PIPE) reduce 536) ((T_POW) reduce 536) ((T_SL) reduce 536) ((T_SPACESHIP) reduce 536) ((T_SR) reduce 536) (("[") reduce 536) (("]") reduce 536) (("^") reduce 536) (("|") reduce 536)) 15081 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("]") reduce 94) (("^") reduce 94) (("|") reduce 94)) 15082 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("]") reduce 93) (("^") reduce 93) (("|") reduce 93)) 15083 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("]") reduce 95) (("^") reduce 95) (("|") reduce 95)) 15085 ((("%") reduce 539) (("*") reduce 539) (("+") reduce 539) (("-") reduce 539) ((".") reduce 539) (("/") reduce 539) (("<") reduce 539) ((">") reduce 539) (("?") reduce 539) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_BOOLEAN_AND) reduce 539) ((T_BOOLEAN_OR) reduce 539) ((T_COALESCE) reduce 539) ((T_DOUBLE_ARROW) reduce 539) ((T_INSTANCEOF) reduce 539) ((T_IS_EQUAL) reduce 539) ((T_IS_GREATER_OR_EQUAL) reduce 539) ((T_IS_IDENTICAL) reduce 539) ((T_IS_NOT_EQUAL) reduce 539) ((T_IS_NOT_IDENTICAL) reduce 539) ((T_IS_SMALLER_OR_EQUAL) reduce 539) ((T_LOGICAL_AND) reduce 539) ((T_LOGICAL_OR) reduce 539) ((T_LOGICAL_XOR) reduce 539) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 539) ((T_OBJECT_OPERATOR) reduce 539) ((T_PIPE) reduce 539) ((T_POW) reduce 539) ((T_SL) reduce 539) ((T_SPACESHIP) reduce 539) ((T_SR) reduce 539) (("[") reduce 539) (("]") reduce 539) (("^") reduce 539) (("|") reduce 539)) 15088 ((("%") reduce 538) (("*") reduce 538) (("+") reduce 538) (("-") reduce 538) ((".") reduce 538) (("/") reduce 538) (("<") reduce 538) ((">") reduce 538) (("?") reduce 538) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_BOOLEAN_AND) reduce 538) ((T_BOOLEAN_OR) reduce 538) ((T_COALESCE) reduce 538) ((T_DOUBLE_ARROW) reduce 538) ((T_INSTANCEOF) reduce 538) ((T_IS_EQUAL) reduce 538) ((T_IS_GREATER_OR_EQUAL) reduce 538) ((T_IS_IDENTICAL) reduce 538) ((T_IS_NOT_EQUAL) reduce 538) ((T_IS_NOT_IDENTICAL) reduce 538) ((T_IS_SMALLER_OR_EQUAL) reduce 538) ((T_LOGICAL_AND) reduce 538) ((T_LOGICAL_OR) reduce 538) ((T_LOGICAL_XOR) reduce 538) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 538) ((T_OBJECT_OPERATOR) reduce 538) ((T_PIPE) reduce 538) ((T_POW) reduce 538) ((T_SL) reduce 538) ((T_SPACESHIP) reduce 538) ((T_SR) reduce 538) (("[") reduce 538) (("]") reduce 538) (("^") reduce 538) (("|") reduce 538)) 15094 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("]") reduce 92) (("^") reduce 92) (("|") reduce 92)) 15097 ((("%") reduce 535) (("*") reduce 535) (("+") reduce 535) (("-") reduce 535) ((".") reduce 535) (("/") reduce 535) (("<") reduce 535) ((">") reduce 535) (("?") reduce 535) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_BOOLEAN_AND) reduce 535) ((T_BOOLEAN_OR) reduce 535) ((T_COALESCE) reduce 535) ((T_DOUBLE_ARROW) reduce 535) ((T_INSTANCEOF) reduce 535) ((T_IS_EQUAL) reduce 535) ((T_IS_GREATER_OR_EQUAL) reduce 535) ((T_IS_IDENTICAL) reduce 535) ((T_IS_NOT_EQUAL) reduce 535) ((T_IS_NOT_IDENTICAL) reduce 535) ((T_IS_SMALLER_OR_EQUAL) reduce 535) ((T_LOGICAL_AND) reduce 535) ((T_LOGICAL_OR) reduce 535) ((T_LOGICAL_XOR) reduce 535) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PIPE) reduce 535) ((T_POW) reduce 535) ((T_SL) reduce 535) ((T_SPACESHIP) reduce 535) ((T_SR) reduce 535) (("[") reduce 535) (("]") reduce 535) (("^") reduce 535) (("|") reduce 535)) 15099 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) (("=") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AND_EQUAL) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_COALESCE_EQUAL) reduce 568) ((T_CONCAT_EQUAL) reduce 568) ((T_DEC) reduce 568) ((T_DIV_EQUAL) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INC) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_MINUS_EQUAL) reduce 568) ((T_MOD_EQUAL) reduce 568) ((T_MUL_EQUAL) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_OR_EQUAL) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_PLUS_EQUAL) reduce 568) ((T_POW) reduce 568) ((T_POW_EQUAL) reduce 568) ((T_SL) reduce 568) ((T_SL_EQUAL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) ((T_SR_EQUAL) reduce 568) ((T_XOR_EQUAL) reduce 568) (("[") reduce 568) (("]") reduce 568) (("^") reduce 568) (("|") reduce 568)) 15100 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 479) (("(") shift) (("*") reduce 479) (("+") shift) (("-") shift) ((".") reduce 479) (("/") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 479) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 479) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 479) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PIPE) reduce 479) ((T_POW) reduce 479) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("]") reduce 479) (("^") reduce 479) (("`") shift) (("|") reduce 479) (("~") shift)) 15107 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) (("<") reduce 564) (("=") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AND_EQUAL) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_COALESCE_EQUAL) reduce 564) ((T_CONCAT_EQUAL) reduce 564) ((T_DEC) reduce 564) ((T_DIV_EQUAL) reduce 564) ((T_DOUBLE_ARROW) reduce 564) ((T_INC) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_MINUS_EQUAL) reduce 564) ((T_MOD_EQUAL) reduce 564) ((T_MUL_EQUAL) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_OR_EQUAL) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_PLUS_EQUAL) reduce 564) ((T_POW) reduce 564) ((T_POW_EQUAL) reduce 564) ((T_SL) reduce 564) ((T_SL_EQUAL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) ((T_SR_EQUAL) reduce 564) ((T_XOR_EQUAL) reduce 564) (("[") reduce 564) (("]") reduce 564) (("^") reduce 564) (("|") reduce 564)) 15108 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PIPE) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 551) (("]") reduce 530) (("^") reduce 530) (("|") reduce 530)) 15110 ((("%") reduce 529) (("*") reduce 529) (("+") reduce 529) (("-") reduce 529) ((".") reduce 529) (("/") reduce 529) (("<") reduce 529) ((">") reduce 529) (("?") reduce 529) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_BOOLEAN_AND) reduce 529) ((T_BOOLEAN_OR) reduce 529) ((T_COALESCE) reduce 529) ((T_DOUBLE_ARROW) reduce 529) ((T_INSTANCEOF) reduce 529) ((T_IS_EQUAL) reduce 529) ((T_IS_GREATER_OR_EQUAL) reduce 529) ((T_IS_IDENTICAL) reduce 529) ((T_IS_NOT_EQUAL) reduce 529) ((T_IS_NOT_IDENTICAL) reduce 529) ((T_IS_SMALLER_OR_EQUAL) reduce 529) ((T_LOGICAL_AND) reduce 529) ((T_LOGICAL_OR) reduce 529) ((T_LOGICAL_XOR) reduce 529) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PIPE) reduce 529) ((T_POW) reduce 529) ((T_SL) reduce 529) ((T_SPACESHIP) reduce 529) ((T_SR) reduce 529) (("[") reduce 554) (("]") reduce 529) (("^") reduce 529) (("|") reduce 529)) 15111 ((("%") reduce 528) (("(") reduce 557) (("*") reduce 528) (("+") reduce 528) (("-") reduce 528) ((".") reduce 528) (("/") reduce 528) (("<") reduce 528) ((">") reduce 528) (("?") reduce 528) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_BOOLEAN_AND) reduce 528) ((T_BOOLEAN_OR) reduce 528) ((T_COALESCE) reduce 528) ((T_DOUBLE_ARROW) reduce 528) ((T_INSTANCEOF) reduce 528) ((T_IS_EQUAL) reduce 528) ((T_IS_GREATER_OR_EQUAL) reduce 528) ((T_IS_IDENTICAL) reduce 528) ((T_IS_NOT_EQUAL) reduce 528) ((T_IS_NOT_IDENTICAL) reduce 528) ((T_IS_SMALLER_OR_EQUAL) reduce 528) ((T_LOGICAL_AND) reduce 528) ((T_LOGICAL_OR) reduce 528) ((T_LOGICAL_XOR) reduce 528) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PIPE) reduce 528) ((T_POW) reduce 528) ((T_SL) reduce 528) ((T_SPACESHIP) reduce 528) ((T_SR) reduce 528) (("[") reduce 550) (("]") reduce 528) (("^") reduce 528) (("|") reduce 528)) 15112 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 480) ((T_LOGICAL_OR) reduce 480) ((T_LOGICAL_XOR) reduce 480) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 480) (("^") shift) (("|") shift)) 15115 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) (("<") reduce 563) (("=") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AND_EQUAL) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_COALESCE_EQUAL) reduce 563) ((T_CONCAT_EQUAL) reduce 563) ((T_DEC) reduce 563) ((T_DIV_EQUAL) reduce 563) ((T_DOUBLE_ARROW) reduce 563) ((T_INC) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_MINUS_EQUAL) reduce 563) ((T_MOD_EQUAL) reduce 563) ((T_MUL_EQUAL) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_OR_EQUAL) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_PLUS_EQUAL) reduce 563) ((T_POW) reduce 563) ((T_POW_EQUAL) reduce 563) ((T_SL) reduce 563) ((T_SL_EQUAL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) ((T_SR_EQUAL) reduce 563) ((T_XOR_EQUAL) reduce 563) (("[") reduce 563) (("]") reduce 563) (("^") reduce 563) (("|") reduce 563)) 15116 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_DOUBLE_ARROW) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_PIPE) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("]") reduce 484) (("^") reduce 484) (("|") reduce 484)) 15117 ((("%") reduce 466) (("*") reduce 466) (("+") reduce 466) (("-") reduce 466) ((".") reduce 466) (("/") reduce 466) (("<") reduce 466) ((">") reduce 466) (("?") reduce 466) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_BOOLEAN_AND) reduce 466) ((T_BOOLEAN_OR) reduce 466) ((T_COALESCE) reduce 466) ((T_DOUBLE_ARROW) reduce 466) ((T_INSTANCEOF) reduce 466) ((T_IS_EQUAL) reduce 466) ((T_IS_GREATER_OR_EQUAL) reduce 466) ((T_IS_IDENTICAL) reduce 466) ((T_IS_NOT_EQUAL) reduce 466) ((T_IS_NOT_IDENTICAL) reduce 466) ((T_IS_SMALLER_OR_EQUAL) reduce 466) ((T_LOGICAL_AND) reduce 466) ((T_LOGICAL_OR) reduce 466) ((T_LOGICAL_XOR) reduce 466) ((T_PIPE) reduce 466) ((T_POW) reduce 466) ((T_SL) reduce 466) ((T_SPACESHIP) reduce 466) ((T_SR) reduce 466) (("]") reduce 466) (("^") reduce 466) (("|") reduce 466)) 15118 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_PIPE) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("]") reduce 488) (("^") reduce 488) (("|") reduce 488)) 15119 ((("%") reduce 531) (("(") shift) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("]") reduce 531) (("^") reduce 531) (("|") reduce 531)) 15120 ((("%") reduce 461) (("(") reduce 558) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PIPE) reduce 461) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") reduce 552) (("]") reduce 461) (("^") reduce 461) (("|") reduce 461)) 15121 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_DOUBLE_ARROW) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_PIPE) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("]") reduce 462) (("^") reduce 462) (("|") reduce 462)) 15122 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_DOUBLE_ARROW) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_PIPE) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("]") reduce 476) (("^") reduce 476) (("|") reduce 476)) 15123 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) (("=") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AND_EQUAL) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_COALESCE_EQUAL) reduce 559) ((T_CONCAT_EQUAL) reduce 559) ((T_DEC) reduce 559) ((T_DIV_EQUAL) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INC) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_MINUS_EQUAL) reduce 559) ((T_MOD_EQUAL) reduce 559) ((T_MUL_EQUAL) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_OR_EQUAL) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_PLUS_EQUAL) reduce 559) ((T_POW) reduce 559) ((T_POW_EQUAL) reduce 559) ((T_SL) reduce 559) ((T_SL_EQUAL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) ((T_SR_EQUAL) reduce 559) ((T_XOR_EQUAL) reduce 559) (("[") reduce 559) (("]") reduce 559) (("^") reduce 559) (("|") reduce 559)) 15124 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("]") reduce 565) (("^") reduce 565) (("|") reduce 565)) 15125 ((("%") reduce 403) (("*") reduce 403) (("+") reduce 403) (("-") reduce 403) ((".") reduce 403) (("/") reduce 403) (("<") reduce 403) (("=") shift) ((">") reduce 403) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_DOUBLE_ARROW) reduce 403) ((T_INC) shift) ((T_INSTANCEOF) reduce 403) ((T_IS_EQUAL) reduce 403) ((T_IS_GREATER_OR_EQUAL) reduce 403) ((T_IS_IDENTICAL) reduce 403) ((T_IS_NOT_EQUAL) reduce 403) ((T_IS_NOT_IDENTICAL) reduce 403) ((T_IS_SMALLER_OR_EQUAL) reduce 403) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 403) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 403) ((T_POW_EQUAL) shift) ((T_SL) reduce 403) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 403) ((T_SR) reduce 403) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 548) (("]") reduce 403) (("^") reduce 403) (("|") reduce 403)) 15128 ((("%") reduce 448) (("*") reduce 448) (("+") reduce 448) (("-") reduce 448) ((".") reduce 448) (("/") reduce 448) (("<") reduce 448) ((">") reduce 448) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_BOOLEAN_AND) reduce 448) ((T_BOOLEAN_OR) reduce 448) ((T_COALESCE) reduce 448) ((T_DOUBLE_ARROW) reduce 448) ((T_INSTANCEOF) reduce 448) ((T_IS_EQUAL) reduce 448) ((T_IS_GREATER_OR_EQUAL) reduce 448) ((T_IS_IDENTICAL) reduce 448) ((T_IS_NOT_EQUAL) reduce 448) ((T_IS_NOT_IDENTICAL) reduce 448) ((T_IS_SMALLER_OR_EQUAL) reduce 448) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_PIPE) reduce 448) ((T_POW) shift) ((T_SL) reduce 448) ((T_SPACESHIP) reduce 448) ((T_SR) reduce 448) (("]") reduce 448) (("^") reduce 448) (("|") reduce 448)) 15160 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_DOUBLE_ARROW) reduce 432) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 432) (("^") shift) (("|") reduce 432)) 15161 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_DOUBLE_ARROW) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 435) (("^") reduce 435) (("|") reduce 435)) 15162 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 444) (("/") shift) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_PIPE) reduce 444) ((T_POW) shift) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("]") reduce 444) (("^") reduce 444) (("|") reduce 444)) 15163 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 458) (("^") reduce 458) (("|") reduce 458)) 15164 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 443) (("/") shift) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_DOUBLE_ARROW) reduce 443) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_PIPE) reduce 443) ((T_POW) shift) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("]") reduce 443) (("^") reduce 443) (("|") reduce 443)) 15165 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_DOUBLE_ARROW) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_PIPE) reduce 440) ((T_POW) shift) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("]") reduce 440) (("^") reduce 440) (("|") reduce 440)) 15166 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_DOUBLE_ARROW) reduce 453) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_PIPE) reduce 453) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 453) ((T_SR) shift) (("]") reduce 453) (("^") reduce 453) (("|") reduce 453)) 15167 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 431) (("^") shift) (("|") shift)) 15168 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 429) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 429) (("^") shift) (("|") shift)) 15169 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 430) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 430) (("^") shift) (("|") shift)) 15170 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 455) ((T_SR) shift) (("]") reduce 455) (("^") reduce 455) (("|") reduce 455)) 15171 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_DOUBLE_ARROW) reduce 450) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 450) (("^") reduce 450) (("|") reduce 450)) 15172 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_DOUBLE_ARROW) reduce 452) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 452) (("^") reduce 452) (("|") reduce 452)) 15173 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_DOUBLE_ARROW) reduce 449) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 449) (("^") reduce 449) (("|") reduce 449)) 15174 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 457) ((T_SR) shift) (("]") reduce 457) (("^") reduce 457) (("|") reduce 457)) 15175 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_DOUBLE_ARROW) reduce 451) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 451) (("^") reduce 451) (("|") reduce 451)) 15178 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("]") reduce 94) (("^") reduce 94) (("|") reduce 94)) 15179 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("]") reduce 93) (("^") reduce 93) (("|") reduce 93)) 15180 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("]") reduce 95) (("^") reduce 95) (("|") reduce 95)) 15181 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("]") reduce 509) (("^") reduce 509) (("|") reduce 509)) 15182 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("]") reduce 92) (("^") reduce 92) (("|") reduce 92)) 15183 ((("%") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("]") reduce 568) (("^") reduce 568) (("|") reduce 568)) 15184 ((("%") reduce 511) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("]") reduce 511) (("^") reduce 511) (("|") reduce 511)) 15185 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PIPE) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("]") reduce 459) (("^") reduce 459) (("|") reduce 459)) 15186 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("]") reduce 510) (("^") reduce 510) (("|") reduce 510)) 15187 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("]") reduce 512) (("^") reduce 512) (("|") reduce 512)) 15188 ((("%") reduce 573) (("*") reduce 573) (("+") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_DOUBLE_ARROW) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("]") reduce 573) (("^") reduce 573) (("|") reduce 573)) 15194 ((("%") reduce 574) (("*") reduce 574) (("+") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_DOUBLE_ARROW) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("]") reduce 574) (("^") reduce 574) (("|") reduce 574)) 15195 ((("%") reduce 578) (("*") reduce 578) (("+") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_DOUBLE_ARROW) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("]") reduce 578) (("^") reduce 578) (("|") reduce 578)) 15196 ((("%") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("]") reduce 582) (("^") reduce 582) (("|") reduce 582)) 15197 ((("%") reduce 575) (("*") reduce 575) (("+") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_DOUBLE_ARROW) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("]") reduce 575) (("^") reduce 575) (("|") reduce 575)) 15198 ((("%") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("]") reduce 584) (("^") reduce 584) (("|") reduce 584)) 15201 ((("%") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("]") reduce 583) (("^") reduce 583) (("|") reduce 583)) 15202 ((("%") reduce 576) (("*") reduce 576) (("+") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_DOUBLE_ARROW) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("]") reduce 576) (("^") reduce 576) (("|") reduce 576)) 15204 ((("%") reduce 577) (("*") reduce 577) (("+") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_DOUBLE_ARROW) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("]") reduce 577) (("^") reduce 577) (("|") reduce 577)) 15206 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("]") reduce 513) (("^") reduce 513) (("|") reduce 513)) 15207 ((("%") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("]") reduce 570) (("^") reduce 570) (("|") reduce 570)) 15210 ((("%") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("]") reduce 569) (("^") reduce 569) (("|") reduce 569)) 15211 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 465) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 465) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 465) (("^") shift) (("|") shift)) 15212 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_DOUBLE_ARROW) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 427) (("^") shift) (("|") shift)) 15213 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_DOUBLE_ARROW) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 428) (("^") shift) (("|") shift)) 15214 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_DOUBLE_ARROW) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 433) (("^") reduce 433) (("|") reduce 433)) 15215 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_DOUBLE_ARROW) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 434) (("^") reduce 434) (("|") reduce 434)) 15219 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 463) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 463) (("^") shift) (("|") shift)) 15220 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 464) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 464) (("^") shift) (("|") shift)) 15221 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 456) ((T_SR) shift) (("]") reduce 456) (("^") reduce 456) (("|") reduce 456)) 15222 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 454) ((T_SR) shift) (("]") reduce 454) (("^") reduce 454) (("|") reduce 454)) 15223 ((("%") reduce 441) (("*") reduce 441) (("+") reduce 441) (("-") reduce 441) ((".") reduce 441) (("/") reduce 441) (("<") reduce 441) ((">") reduce 441) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_DOUBLE_ARROW) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 441) ((T_IS_GREATER_OR_EQUAL) reduce 441) ((T_IS_IDENTICAL) reduce 441) ((T_IS_NOT_EQUAL) reduce 441) ((T_IS_NOT_IDENTICAL) reduce 441) ((T_IS_SMALLER_OR_EQUAL) reduce 441) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_PIPE) reduce 441) ((T_POW) shift) ((T_SL) reduce 441) ((T_SPACESHIP) reduce 441) ((T_SR) reduce 441) (("]") reduce 441) (("^") reduce 441) (("|") reduce 441)) 15224 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 436) (("/") shift) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_DOUBLE_ARROW) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_PIPE) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 436) ((T_SR) shift) (("]") reduce 436) (("^") reduce 436) (("|") reduce 436)) 15225 ((("%") shift) (("*") shift) (("+") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") shift) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_DOUBLE_ARROW) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_PIPE) reduce 438) ((T_POW) shift) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("]") reduce 438) (("^") reduce 438) (("|") reduce 438)) 15226 ((("%") shift) (("*") shift) (("+") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") shift) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_DOUBLE_ARROW) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_PIPE) reduce 437) ((T_POW) shift) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("]") reduce 437) (("^") reduce 437) (("|") reduce 437)) 15227 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_DOUBLE_ARROW) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_PIPE) reduce 439) ((T_POW) shift) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("]") reduce 439) (("^") reduce 439) (("|") reduce 439)) 15228 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_DOUBLE_ARROW) reduce 442) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_PIPE) reduce 442) ((T_POW) shift) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("]") reduce 442) (("^") reduce 442) (("|") reduce 442)) 15230 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_DOUBLE_ARROW) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_PIPE) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("]") reduce 74) (("^") reduce 74) (("|") reduce 74)) 15231 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_DOUBLE_ARROW) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_PIPE) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("]") reduce 52) (("^") reduce 52) (("|") reduce 52)) 15232 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_DOUBLE_ARROW) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_PIPE) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("]") reduce 27) (("^") reduce 27) (("|") reduce 27)) 15233 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_DOUBLE_ARROW) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_PIPE) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("]") reduce 51) (("^") reduce 51) (("|") reduce 51)) 15234 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_DOUBLE_ARROW) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_PIPE) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("]") reduce 53) (("^") reduce 53) (("|") reduce 53)) 15235 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_DOUBLE_ARROW) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_PIPE) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("]") reduce 49) (("^") reduce 49) (("|") reduce 49)) 15236 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_DOUBLE_ARROW) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_PIPE) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("]") reduce 29) (("^") reduce 29) (("|") reduce 29)) 15237 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_DOUBLE_ARROW) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_PIPE) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("]") reduce 59) (("^") reduce 59) (("|") reduce 59)) 15238 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_DOUBLE_ARROW) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_PIPE) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("]") reduce 60) (("^") reduce 60) (("|") reduce 60)) 15239 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_DOUBLE_ARROW) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_PIPE) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("]") reduce 11) (("^") reduce 11) (("|") reduce 11)) 15240 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_DOUBLE_ARROW) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_PIPE) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("]") reduce 42) (("^") reduce 42) (("|") reduce 42)) 15241 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_DOUBLE_ARROW) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_PIPE) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("]") reduce 39) (("^") reduce 39) (("|") reduce 39)) 15242 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_DOUBLE_ARROW) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_PIPE) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("]") reduce 25) (("^") reduce 25) (("|") reduce 25)) 15243 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_DOUBLE_ARROW) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_PIPE) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("]") reduce 50) (("^") reduce 50) (("|") reduce 50)) 15244 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_DOUBLE_ARROW) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_PIPE) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("]") reduce 66) (("^") reduce 66) (("|") reduce 66)) 15245 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_DOUBLE_ARROW) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_PIPE) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("]") reduce 18) (("^") reduce 18) (("|") reduce 18)) 15246 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_DOUBLE_ARROW) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_PIPE) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("]") reduce 17) (("^") reduce 17) (("|") reduce 17)) 15247 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_DOUBLE_ARROW) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_PIPE) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("]") reduce 15) (("^") reduce 15) (("|") reduce 15)) 15248 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_DOUBLE_ARROW) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_PIPE) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("]") reduce 14) (("^") reduce 14) (("|") reduce 14)) 15249 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_DOUBLE_ARROW) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_PIPE) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("]") reduce 38) (("^") reduce 38) (("|") reduce 38)) 15250 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_DOUBLE_ARROW) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_PIPE) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("]") reduce 26) (("^") reduce 26) (("|") reduce 26)) 15251 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_DOUBLE_ARROW) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_PIPE) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("]") reduce 22) (("^") reduce 22) (("|") reduce 22)) 15252 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_DOUBLE_ARROW) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_PIPE) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("]") reduce 24) (("^") reduce 24) (("|") reduce 24)) 15253 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_DOUBLE_ARROW) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_PIPE) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("]") reduce 16) (("^") reduce 16) (("|") reduce 16)) 15254 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_DOUBLE_ARROW) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_PIPE) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("]") reduce 48) (("^") reduce 48) (("|") reduce 48)) 15255 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_DOUBLE_ARROW) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_PIPE) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("]") reduce 20) (("^") reduce 20) (("|") reduce 20)) 15256 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_DOUBLE_ARROW) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_PIPE) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("]") reduce 70) (("^") reduce 70) (("|") reduce 70)) 15257 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_DOUBLE_ARROW) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_PIPE) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("]") reduce 3) (("^") reduce 3) (("|") reduce 3)) 15258 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_DOUBLE_ARROW) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_PIPE) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("]") reduce 12) (("^") reduce 12) (("|") reduce 12)) 15259 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_DOUBLE_ARROW) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_PIPE) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("]") reduce 54) (("^") reduce 54) (("|") reduce 54)) 15260 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_DOUBLE_ARROW) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_PIPE) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("]") reduce 65) (("^") reduce 65) (("|") reduce 65)) 15261 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_DOUBLE_ARROW) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_PIPE) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("]") reduce 75) (("^") reduce 75) (("|") reduce 75)) 15262 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_DOUBLE_ARROW) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_PIPE) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("]") reduce 30) (("^") reduce 30) (("|") reduce 30)) 15263 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_DOUBLE_ARROW) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_PIPE) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("]") reduce 68) (("^") reduce 68) (("|") reduce 68)) 15264 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_DOUBLE_ARROW) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_PIPE) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("]") reduce 21) (("^") reduce 21) (("|") reduce 21)) 15265 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_DOUBLE_ARROW) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_PIPE) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("]") reduce 23) (("^") reduce 23) (("|") reduce 23)) 15266 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_DOUBLE_ARROW) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_PIPE) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("]") reduce 41) (("^") reduce 41) (("|") reduce 41)) 15267 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_DOUBLE_ARROW) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_PIPE) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("]") reduce 62) (("^") reduce 62) (("|") reduce 62)) 15268 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_DOUBLE_ARROW) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_PIPE) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("]") reduce 34) (("^") reduce 34) (("|") reduce 34)) 15269 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_DOUBLE_ARROW) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_PIPE) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("]") reduce 40) (("^") reduce 40) (("|") reduce 40)) 15270 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_DOUBLE_ARROW) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_PIPE) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("]") reduce 13) (("^") reduce 13) (("|") reduce 13)) 15271 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_DOUBLE_ARROW) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_PIPE) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("]") reduce 55) (("^") reduce 55) (("|") reduce 55)) 15272 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_DOUBLE_ARROW) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_PIPE) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("]") reduce 1) (("^") reduce 1) (("|") reduce 1)) 15273 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_DOUBLE_ARROW) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_PIPE) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("]") reduce 2) (("^") reduce 2) (("|") reduce 2)) 15274 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_DOUBLE_ARROW) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_PIPE) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("]") reduce 9) (("^") reduce 9) (("|") reduce 9)) 15275 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_DOUBLE_ARROW) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_PIPE) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("]") reduce 33) (("^") reduce 33) (("|") reduce 33)) 15276 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_DOUBLE_ARROW) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_PIPE) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("]") reduce 58) (("^") reduce 58) (("|") reduce 58)) 15277 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_DOUBLE_ARROW) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_PIPE) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("]") reduce 37) (("^") reduce 37) (("|") reduce 37)) 15278 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_DOUBLE_ARROW) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_PIPE) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("]") reduce 64) (("^") reduce 64) (("|") reduce 64)) 15279 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_DOUBLE_ARROW) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_PIPE) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("]") reduce 46) (("^") reduce 46) (("|") reduce 46)) 15280 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_DOUBLE_ARROW) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_PIPE) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("]") reduce 8) (("^") reduce 8) (("|") reduce 8)) 15281 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_DOUBLE_ARROW) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_PIPE) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("]") reduce 6) (("^") reduce 6) (("|") reduce 6)) 15282 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_DOUBLE_ARROW) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_PIPE) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("]") reduce 7) (("^") reduce 7) (("|") reduce 7)) 15283 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_DOUBLE_ARROW) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_PIPE) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("]") reduce 69) (("^") reduce 69) (("|") reduce 69)) 15284 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_DOUBLE_ARROW) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_PIPE) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("]") reduce 63) (("^") reduce 63) (("|") reduce 63)) 15285 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_DOUBLE_ARROW) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_PIPE) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("]") reduce 56) (("^") reduce 56) (("|") reduce 56)) 15286 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_DOUBLE_ARROW) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_PIPE) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("]") reduce 10) (("^") reduce 10) (("|") reduce 10)) 15287 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_DOUBLE_ARROW) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_PIPE) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("]") reduce 67) (("^") reduce 67) (("|") reduce 67)) 15288 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_DOUBLE_ARROW) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_PIPE) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("]") reduce 44) (("^") reduce 44) (("|") reduce 44)) 15289 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_DOUBLE_ARROW) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_PIPE) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("]") reduce 76) (("^") reduce 76) (("|") reduce 76)) 15290 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_DOUBLE_ARROW) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_PIPE) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("]") reduce 71) (("^") reduce 71) (("|") reduce 71)) 15291 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_DOUBLE_ARROW) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_PIPE) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("]") reduce 77) (("^") reduce 77) (("|") reduce 77)) 15292 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_DOUBLE_ARROW) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_PIPE) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("]") reduce 78) (("^") reduce 78) (("|") reduce 78)) 15293 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_DOUBLE_ARROW) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_PIPE) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("]") reduce 79) (("^") reduce 79) (("|") reduce 79)) 15294 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_DOUBLE_ARROW) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_PIPE) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("]") reduce 4) (("^") reduce 4) (("|") reduce 4)) 15295 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_DOUBLE_ARROW) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_PIPE) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("]") reduce 5) (("^") reduce 5) (("|") reduce 5)) 15296 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_DOUBLE_ARROW) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_PIPE) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("]") reduce 43) (("^") reduce 43) (("|") reduce 43)) 15297 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_DOUBLE_ARROW) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_PIPE) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("]") reduce 73) (("^") reduce 73) (("|") reduce 73)) 15298 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_DOUBLE_ARROW) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_PIPE) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("]") reduce 82) (("^") reduce 82) (("|") reduce 82)) 15299 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_DOUBLE_ARROW) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_PIPE) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("]") reduce 47) (("^") reduce 47) (("|") reduce 47)) 15300 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_DOUBLE_ARROW) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_PIPE) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("]") reduce 31) (("^") reduce 31) (("|") reduce 31)) 15301 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_DOUBLE_ARROW) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_PIPE) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("]") reduce 57) (("^") reduce 57) (("|") reduce 57)) 15302 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_DOUBLE_ARROW) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_PIPE) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("]") reduce 61) (("^") reduce 61) (("|") reduce 61)) 15303 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_DOUBLE_ARROW) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_PIPE) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("]") reduce 28) (("^") reduce 28) (("|") reduce 28)) 15304 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_DOUBLE_ARROW) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_PIPE) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("]") reduce 36) (("^") reduce 36) (("|") reduce 36)) 15305 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_DOUBLE_ARROW) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_PIPE) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("]") reduce 32) (("^") reduce 32) (("|") reduce 32)) 15306 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_DOUBLE_ARROW) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_PIPE) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("]") reduce 35) (("^") reduce 35) (("|") reduce 35)) 15307 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_DOUBLE_ARROW) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_PIPE) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("]") reduce 19) (("^") reduce 19) (("|") reduce 19)) 15308 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_DOUBLE_ARROW) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_PIPE) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("]") reduce 45) (("^") reduce 45) (("|") reduce 45)) 15309 ((("%") reduce 542) (("(") reduce 579) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PIPE) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("]") reduce 542) (("^") reduce 542) (("|") reduce 542)) 15311 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_DOUBLE_ARROW) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_PIPE) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("]") reduce 72) (("^") reduce 72) (("|") reduce 72)) 15312 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_DOUBLE_ARROW) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_PIPE) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("]") reduce 83) (("^") reduce 83) (("|") reduce 83)) 15313 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) (("<") reduce 572) (("=") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AND_EQUAL) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_COALESCE_EQUAL) reduce 572) ((T_CONCAT_EQUAL) reduce 572) ((T_DEC) reduce 572) ((T_DIV_EQUAL) reduce 572) ((T_DOUBLE_ARROW) reduce 572) ((T_INC) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_MINUS_EQUAL) reduce 572) ((T_MOD_EQUAL) reduce 572) ((T_MUL_EQUAL) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_OR_EQUAL) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_PLUS_EQUAL) reduce 572) ((T_POW) reduce 572) ((T_POW_EQUAL) reduce 572) ((T_SL) reduce 572) ((T_SL_EQUAL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) ((T_SR_EQUAL) reduce 572) ((T_XOR_EQUAL) reduce 572) (("[") reduce 572) (("]") reduce 572) (("^") reduce 572) (("|") reduce 572)) 15316 ((("%") reduce 544) (("(") reduce 580) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PIPE) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("]") reduce 544) (("^") reduce 544) (("|") reduce 544)) 15318 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("]") reduce 507) (("^") reduce 507) (("|") reduce 507)) 15319 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("]") reduce 291) (("^") reduce 291) (("|") reduce 291)) 15323 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) (("=") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AND_EQUAL) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_COALESCE_EQUAL) reduce 292) ((T_CONCAT_EQUAL) reduce 292) ((T_DEC) reduce 292) ((T_DIV_EQUAL) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INC) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_MINUS_EQUAL) reduce 292) ((T_MOD_EQUAL) reduce 292) ((T_MUL_EQUAL) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_OR_EQUAL) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_PLUS_EQUAL) reduce 292) ((T_POW) reduce 292) ((T_POW_EQUAL) reduce 292) ((T_SL) reduce 292) ((T_SL_EQUAL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) ((T_SR_EQUAL) reduce 292) ((T_XOR_EQUAL) reduce 292) (("[") reduce 292) (("]") reduce 292) (("^") reduce 292) (("|") reduce 292)) 15324 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) (("=") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AND_EQUAL) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_COALESCE_EQUAL) reduce 293) ((T_CONCAT_EQUAL) reduce 293) ((T_DEC) reduce 293) ((T_DIV_EQUAL) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INC) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_MINUS_EQUAL) reduce 293) ((T_MOD_EQUAL) reduce 293) ((T_MUL_EQUAL) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_OR_EQUAL) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_PLUS_EQUAL) reduce 293) ((T_POW) reduce 293) ((T_POW_EQUAL) reduce 293) ((T_SL) reduce 293) ((T_SL_EQUAL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) ((T_SR_EQUAL) reduce 293) ((T_XOR_EQUAL) reduce 293) (("[") reduce 293) (("]") reduce 293) (("^") reduce 293) (("|") reduce 293)) 15329 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_DOUBLE_ARROW) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_PIPE) reduce 425) ((T_POW) reduce 425) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("]") reduce 425) (("^") reduce 425) (("|") reduce 425)) 15331 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_DOUBLE_ARROW) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_PIPE) reduce 423) ((T_POW) reduce 423) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("]") reduce 423) (("^") reduce 423) (("|") reduce 423)) 15341 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 419) (("^") shift) (("|") shift)) 15342 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 421) (("^") shift) (("|") shift)) 15343 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 420) (("^") shift) (("|") shift)) 15344 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 413) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 413) (("^") shift) (("|") shift)) 15345 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 410) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 410) (("^") shift) (("|") shift)) 15346 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 418) (("^") shift) (("|") shift)) 15347 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 412) (("^") shift) (("|") shift)) 15348 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 416) (("^") shift) (("|") shift)) 15349 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 411) (("^") shift) (("|") shift)) 15350 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 414) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 414) (("^") shift) (("|") shift)) 15351 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 415) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 415) (("^") shift) (("|") shift)) 15352 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 422) (("^") shift) (("|") shift)) 15353 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 417) (("^") shift) (("|") shift)) 15355 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 406) (("^") shift) (("|") shift)) 15358 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("]") reduce 568) (("^") reduce 568) (("|") reduce 568)) 15361 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) (("<") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_DOUBLE_ARROW) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_POW) reduce 564) ((T_SL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) (("[") reduce 564) (("]") reduce 564) (("^") reduce 564) (("|") reduce 564)) 15363 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) (("<") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_DOUBLE_ARROW) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_POW) reduce 563) ((T_SL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) (("[") reduce 563) (("]") reduce 563) (("^") reduce 563) (("|") reduce 563)) 15365 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("]") reduce 559) (("^") reduce 559) (("|") reduce 559)) 15366 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("]") reduce 565) (("^") reduce 565) (("|") reduce 565)) 15367 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_DOUBLE_ARROW) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("[") reduce 548) (("]") reduce 407) (("^") reduce 407) (("|") reduce 407)) 15371 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) (("<") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_DOUBLE_ARROW) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_POW) reduce 572) ((T_SL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) (("[") reduce 572) (("]") reduce 572) (("^") reduce 572) (("|") reduce 572)) 15373 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("]") reduce 507) (("^") reduce 507) (("|") reduce 507)) 15374 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("]") reduce 291) (("^") reduce 291) (("|") reduce 291)) 15378 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("[") reduce 292) (("]") reduce 292) (("^") reduce 292) (("|") reduce 292)) 15379 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("[") reduce 293) (("]") reduce 293) (("^") reduce 293) (("|") reduce 293)) 15380 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("]") reduce 504) (("^") reduce 504) (("|") reduce 504)) 15383 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) (("<") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_DOUBLE_ARROW) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_POW) reduce 571) ((T_SL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) (("[") reduce 571) (("]") reduce 571) (("^") reduce 571) (("|") reduce 571)) 15384 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("]") reduce 506) (("^") reduce 506) (("|") reduce 506)) 15385 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("]") reduce 508) (("^") reduce 508) (("|") reduce 508)) 15390 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("]") reduce 560) (("^") reduce 560) (("|") reduce 560)) 15391 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("]") reduce 582) (("^") reduce 582) (("|") reduce 582)) 15392 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("]") reduce 566) (("^") reduce 566) (("|") reduce 566)) 15393 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("]") reduce 584) (("^") reduce 584) (("|") reduce 584)) 15396 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("]") reduce 583) (("^") reduce 583) (("|") reduce 583)) 15397 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("]") reduce 561) (("^") reduce 561) (("|") reduce 561)) 15398 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("]") reduce 567) (("^") reduce 567) (("|") reduce 567)) 15399 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("[") reduce 562) (("]") reduce 562) (("^") reduce 562) (("|") reduce 562)) 15400 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("]") reduce 505) (("^") reduce 505) (("|") reduce 505)) 15401 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("]") reduce 570) (("^") reduce 570) (("|") reduce 570)) 15404 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("]") reduce 569) (("^") reduce 569) (("|") reduce 569)) 15405 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) (("=") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AND_EQUAL) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_COALESCE_EQUAL) reduce 504) ((T_CONCAT_EQUAL) reduce 504) ((T_DEC) reduce 504) ((T_DIV_EQUAL) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INC) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_MINUS_EQUAL) reduce 504) ((T_MOD_EQUAL) reduce 504) ((T_MUL_EQUAL) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_OR_EQUAL) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_PLUS_EQUAL) reduce 504) ((T_POW) reduce 504) ((T_POW_EQUAL) reduce 504) ((T_SL) reduce 504) ((T_SL_EQUAL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) ((T_SR_EQUAL) reduce 504) ((T_XOR_EQUAL) reduce 504) (("[") reduce 504) (("]") reduce 504) (("^") reduce 504) (("|") reduce 504)) 15416 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PIPE) reduce 494) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("]") reduce 494) (("^") reduce 494) (("|") reduce 494)) 15417 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_PIPE) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("]") reduce 489) (("^") reduce 489) (("|") reduce 489)) 15427 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 494) (("^") shift) (("|") shift)) 15428 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_PIPE) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("]") reduce 490) (("^") reduce 490) (("|") reduce 490)) 15430 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 481) (("^") shift) (("|") shift)) 15432 ((("%") reduce 541) (("(") reduce 579) (("*") reduce 541) (("+") reduce 541) (("-") reduce 541) ((".") reduce 541) (("/") reduce 541) (("<") reduce 541) ((">") reduce 541) (("?") reduce 541) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_BOOLEAN_AND) reduce 541) ((T_BOOLEAN_OR) reduce 541) ((T_COALESCE) reduce 541) ((T_DOUBLE_ARROW) reduce 541) ((T_INSTANCEOF) reduce 541) ((T_IS_EQUAL) reduce 541) ((T_IS_GREATER_OR_EQUAL) reduce 541) ((T_IS_IDENTICAL) reduce 541) ((T_IS_NOT_EQUAL) reduce 541) ((T_IS_NOT_IDENTICAL) reduce 541) ((T_IS_SMALLER_OR_EQUAL) reduce 541) ((T_LOGICAL_AND) reduce 541) ((T_LOGICAL_OR) reduce 541) ((T_LOGICAL_XOR) reduce 541) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 541) ((T_OBJECT_OPERATOR) reduce 541) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 541) ((T_PIPE) reduce 541) ((T_POW) reduce 541) ((T_SL) reduce 541) ((T_SPACESHIP) reduce 541) ((T_SR) reduce 541) (("[") reduce 541) (("]") reduce 541) (("^") reduce 541) (("|") reduce 541)) 15434 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) (("<") reduce 571) (("=") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AND_EQUAL) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_COALESCE_EQUAL) reduce 571) ((T_CONCAT_EQUAL) reduce 571) ((T_DEC) reduce 571) ((T_DIV_EQUAL) reduce 571) ((T_DOUBLE_ARROW) reduce 571) ((T_INC) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_MINUS_EQUAL) reduce 571) ((T_MOD_EQUAL) reduce 571) ((T_MUL_EQUAL) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_OR_EQUAL) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_PLUS_EQUAL) reduce 571) ((T_POW) reduce 571) ((T_POW_EQUAL) reduce 571) ((T_SL) reduce 571) ((T_SL_EQUAL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) ((T_SR_EQUAL) reduce 571) ((T_XOR_EQUAL) reduce 571) (("[") reduce 571) (("]") reduce 571) (("^") reduce 571) (("|") reduce 571)) 15437 ((("%") reduce 543) (("(") reduce 580) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PIPE) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("]") reduce 543) (("^") reduce 543) (("|") reduce 543)) 15438 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("]") reduce 506) (("^") reduce 506) (("|") reduce 506)) 15439 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("]") reduce 508) (("^") reduce 508) (("|") reduce 508)) 15441 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_DOUBLE_ARROW) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_PIPE) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("]") reduce 485) (("^") reduce 485) (("|") reduce 485)) 15442 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_PIPE) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("]") reduce 487) (("^") reduce 487) (("|") reduce 487)) 15447 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) (("=") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AND_EQUAL) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_COALESCE_EQUAL) reduce 560) ((T_CONCAT_EQUAL) reduce 560) ((T_DEC) reduce 560) ((T_DIV_EQUAL) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INC) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_MINUS_EQUAL) reduce 560) ((T_MOD_EQUAL) reduce 560) ((T_MUL_EQUAL) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_OR_EQUAL) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_PLUS_EQUAL) reduce 560) ((T_POW) reduce 560) ((T_POW_EQUAL) reduce 560) ((T_SL) reduce 560) ((T_SL_EQUAL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) ((T_SR_EQUAL) reduce 560) ((T_XOR_EQUAL) reduce 560) (("[") reduce 560) (("]") reduce 560) (("^") reduce 560) (("|") reduce 560)) 15448 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) (("=") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AND_EQUAL) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_COALESCE_EQUAL) reduce 582) ((T_CONCAT_EQUAL) reduce 582) ((T_DEC) reduce 582) ((T_DIV_EQUAL) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INC) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_MINUS_EQUAL) reduce 582) ((T_MOD_EQUAL) reduce 582) ((T_MUL_EQUAL) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_OR_EQUAL) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_PLUS_EQUAL) reduce 582) ((T_POW) reduce 582) ((T_POW_EQUAL) reduce 582) ((T_SL) reduce 582) ((T_SL_EQUAL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) ((T_SR_EQUAL) reduce 582) ((T_XOR_EQUAL) reduce 582) (("[") reduce 582) (("]") reduce 582) (("^") reduce 582) (("|") reduce 582)) 15449 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("]") reduce 566) (("^") reduce 566) (("|") reduce 566)) 15450 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) (("=") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AND_EQUAL) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_COALESCE_EQUAL) reduce 584) ((T_CONCAT_EQUAL) reduce 584) ((T_DEC) reduce 584) ((T_DIV_EQUAL) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INC) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_MINUS_EQUAL) reduce 584) ((T_MOD_EQUAL) reduce 584) ((T_MUL_EQUAL) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_OR_EQUAL) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_PLUS_EQUAL) reduce 584) ((T_POW) reduce 584) ((T_POW_EQUAL) reduce 584) ((T_SL) reduce 584) ((T_SL_EQUAL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) ((T_SR_EQUAL) reduce 584) ((T_XOR_EQUAL) reduce 584) (("[") reduce 584) (("]") reduce 584) (("^") reduce 584) (("|") reduce 584)) 15453 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) (("=") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AND_EQUAL) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_COALESCE_EQUAL) reduce 583) ((T_CONCAT_EQUAL) reduce 583) ((T_DEC) reduce 583) ((T_DIV_EQUAL) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INC) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_MINUS_EQUAL) reduce 583) ((T_MOD_EQUAL) reduce 583) ((T_MUL_EQUAL) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_OR_EQUAL) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_PLUS_EQUAL) reduce 583) ((T_POW) reduce 583) ((T_POW_EQUAL) reduce 583) ((T_SL) reduce 583) ((T_SL_EQUAL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) ((T_SR_EQUAL) reduce 583) ((T_XOR_EQUAL) reduce 583) (("[") reduce 583) (("]") reduce 583) (("^") reduce 583) (("|") reduce 583)) 15454 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) (("=") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AND_EQUAL) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_COALESCE_EQUAL) reduce 561) ((T_CONCAT_EQUAL) reduce 561) ((T_DEC) reduce 561) ((T_DIV_EQUAL) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INC) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_MINUS_EQUAL) reduce 561) ((T_MOD_EQUAL) reduce 561) ((T_MUL_EQUAL) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_OR_EQUAL) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_PLUS_EQUAL) reduce 561) ((T_POW) reduce 561) ((T_POW_EQUAL) reduce 561) ((T_SL) reduce 561) ((T_SL_EQUAL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) ((T_SR_EQUAL) reduce 561) ((T_XOR_EQUAL) reduce 561) (("[") reduce 561) (("]") reduce 561) (("^") reduce 561) (("|") reduce 561)) 15455 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("]") reduce 567) (("^") reduce 567) (("|") reduce 567)) 15456 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) (("=") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AND_EQUAL) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_COALESCE_EQUAL) reduce 562) ((T_CONCAT_EQUAL) reduce 562) ((T_DEC) reduce 562) ((T_DIV_EQUAL) reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_INC) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_MINUS_EQUAL) reduce 562) ((T_MOD_EQUAL) reduce 562) ((T_MUL_EQUAL) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_OR_EQUAL) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_PLUS_EQUAL) reduce 562) ((T_POW) reduce 562) ((T_POW_EQUAL) reduce 562) ((T_SL) reduce 562) ((T_SL_EQUAL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) ((T_SR_EQUAL) reduce 562) ((T_XOR_EQUAL) reduce 562) (("[") reduce 562) (("]") reduce 562) (("^") reduce 562) (("|") reduce 562)) 15458 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_PIPE) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("]") reduce 477) (("^") reduce 477) (("|") reduce 477)) 15460 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) (("=") shift) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_PIPE) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("]") reduce 520) (("^") reduce 520) (("|") reduce 520)) 15462 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 405) (("^") shift) (("|") shift)) 15463 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 482) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 482) (("^") shift) (("|") shift)) 15466 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 481) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 481) (("^") shift) (("|") shift)) 15467 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_DOUBLE_ARROW) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_PIPE) reduce 473) ((T_POW) shift) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("]") reduce 473) (("^") reduce 473) (("|") reduce 473)) 15468 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 483) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 483) (("^") shift) (("|") shift)) 15469 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_DOUBLE_ARROW) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_PIPE) reduce 469) ((T_POW) shift) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("]") reduce 469) (("^") reduce 469) (("|") reduce 469)) 15470 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_DOUBLE_ARROW) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_PIPE) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("]") reduce 486) (("^") reduce 486) (("|") reduce 486)) 15472 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_PIPE) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("]") reduce 526) (("^") reduce 526) (("|") reduce 526)) 15474 ((("%") reduce 527) (("*") reduce 527) (("+") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_DOUBLE_ARROW) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_PIPE) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("]") reduce 527) (("^") reduce 527) (("|") reduce 527)) 15475 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_PIPE) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("]") reduce 525) (("^") reduce 525) (("|") reduce 525)) 15476 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 619) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 619) (("^") shift) (("|") shift)) 15477 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 618) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 618) (("^") shift) (("|") shift)) 15478 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("]") reduce 505) (("^") reduce 505) (("|") reduce 505)) 15479 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 478) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 478) (("^") shift) (("|") shift)) 15480 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_DOUBLE_ARROW) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_PIPE) reduce 471) ((T_POW) shift) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("]") reduce 471) (("^") reduce 471) (("|") reduce 471)) 15482 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("]") reduce 94) (("^") reduce 94) (("|") reduce 94)) 15483 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("]") reduce 93) (("^") reduce 93) (("|") reduce 93)) 15484 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("]") reduce 95) (("^") reduce 95) (("|") reduce 95)) 15485 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("]") reduce 509) (("^") reduce 509) (("|") reduce 509)) 15486 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("]") reduce 92) (("^") reduce 92) (("|") reduce 92)) 15487 ((("%") reduce 400) (("(") reduce 400) (("*") reduce 400) (("+") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_DOUBLE_ARROW) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 400) ((T_OBJECT_OPERATOR) reduce 400) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 400) ((T_PIPE) reduce 400) ((T_POW) reduce 400) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("[") reduce 400) (("]") reduce 400) (("^") reduce 400) (("|") reduce 400)) 15490 ((("%") reduce 511) (("(") reduce 511) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("]") reduce 511) (("^") reduce 511) (("|") reduce 511)) 15491 ((("%") reduce 402) (("(") shift) (("*") reduce 402) (("+") reduce 402) (("-") reduce 402) ((".") reduce 402) (("/") reduce 402) (("<") reduce 402) ((">") reduce 402) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_DOUBLE_ARROW) reduce 402) ((T_INSTANCEOF) reduce 402) ((T_IS_EQUAL) reduce 402) ((T_IS_GREATER_OR_EQUAL) reduce 402) ((T_IS_IDENTICAL) reduce 402) ((T_IS_NOT_EQUAL) reduce 402) ((T_IS_NOT_IDENTICAL) reduce 402) ((T_IS_SMALLER_OR_EQUAL) reduce 402) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_PIPE) reduce 402) ((T_POW) reduce 402) ((T_SL) reduce 402) ((T_SPACESHIP) reduce 402) ((T_SR) reduce 402) (("]") reduce 402) (("^") reduce 402) (("|") reduce 402)) 15492 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("]") reduce 510) (("^") reduce 510) (("|") reduce 510)) 15493 ((("%") reduce 512) (("(") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("]") reduce 512) (("^") reduce 512) (("|") reduce 512)) 15494 ((("%") reduce 573) (("(") reduce 573) (("*") reduce 573) (("+") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_DOUBLE_ARROW) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("]") reduce 573) (("^") reduce 573) (("|") reduce 573)) 15500 ((("%") reduce 574) (("(") reduce 574) (("*") reduce 574) (("+") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_DOUBLE_ARROW) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("]") reduce 574) (("^") reduce 574) (("|") reduce 574)) 15501 ((("%") reduce 578) (("(") reduce 578) (("*") reduce 578) (("+") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_DOUBLE_ARROW) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("]") reduce 578) (("^") reduce 578) (("|") reduce 578)) 15502 ((("%") reduce 575) (("(") reduce 575) (("*") reduce 575) (("+") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_DOUBLE_ARROW) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("]") reduce 575) (("^") reduce 575) (("|") reduce 575)) 15503 ((("%") reduce 576) (("(") reduce 576) (("*") reduce 576) (("+") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_DOUBLE_ARROW) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("]") reduce 576) (("^") reduce 576) (("|") reduce 576)) 15504 ((("%") reduce 399) (("(") reduce 399) (("*") reduce 399) (("+") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_DOUBLE_ARROW) reduce 399) ((T_INSTANCEOF) reduce 399) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 399) ((T_OBJECT_OPERATOR) reduce 399) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 399) ((T_PIPE) reduce 399) ((T_POW) reduce 399) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("[") reduce 399) (("]") reduce 399) (("^") reduce 399) (("|") reduce 399)) 15506 ((("%") reduce 577) (("(") reduce 577) (("*") reduce 577) (("+") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_DOUBLE_ARROW) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("]") reduce 577) (("^") reduce 577) (("|") reduce 577)) 15507 ((("%") reduce 401) (("(") reduce 401) (("*") reduce 401) (("+") reduce 401) (("-") reduce 401) ((".") reduce 401) (("/") reduce 401) (("<") reduce 401) ((">") reduce 401) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_DOUBLE_ARROW) reduce 401) ((T_INSTANCEOF) reduce 401) ((T_IS_EQUAL) reduce 401) ((T_IS_GREATER_OR_EQUAL) reduce 401) ((T_IS_IDENTICAL) reduce 401) ((T_IS_NOT_EQUAL) reduce 401) ((T_IS_NOT_IDENTICAL) reduce 401) ((T_IS_SMALLER_OR_EQUAL) reduce 401) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 401) ((T_OBJECT_OPERATOR) reduce 401) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 401) ((T_PIPE) reduce 401) ((T_POW) reduce 401) ((T_SL) reduce 401) ((T_SPACESHIP) reduce 401) ((T_SR) reduce 401) (("[") reduce 401) (("]") reduce 401) (("^") reduce 401) (("|") reduce 401)) 15515 ((("%") reduce 398) (("(") reduce 398) (("*") reduce 398) (("+") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_DOUBLE_ARROW) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 398) ((T_OBJECT_OPERATOR) reduce 398) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 398) ((T_PIPE) reduce 398) ((T_POW) reduce 398) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("[") reduce 398) (("]") reduce 398) (("^") reduce 398) (("|") reduce 398)) 15517 ((("%") reduce 513) (("(") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("]") reduce 513) (("^") reduce 513) (("|") reduce 513)) 15523 ((("%") reduce 233) (("*") reduce 233) (("+") reduce 233) (("-") reduce 233) ((".") reduce 233) (("/") reduce 233) (("<") reduce 233) ((">") reduce 233) (("?") reduce 233) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_BOOLEAN_AND) reduce 233) ((T_BOOLEAN_OR) reduce 233) ((T_COALESCE) reduce 233) ((T_DOUBLE_ARROW) reduce 233) ((T_INSTANCEOF) reduce 233) ((T_IS_EQUAL) reduce 233) ((T_IS_GREATER_OR_EQUAL) reduce 233) ((T_IS_IDENTICAL) reduce 233) ((T_IS_NOT_EQUAL) reduce 233) ((T_IS_NOT_IDENTICAL) reduce 233) ((T_IS_SMALLER_OR_EQUAL) reduce 233) ((T_LOGICAL_AND) reduce 233) ((T_LOGICAL_OR) reduce 233) ((T_LOGICAL_XOR) reduce 233) ((T_PIPE) reduce 233) ((T_POW) reduce 233) ((T_SL) reduce 233) ((T_SPACESHIP) reduce 233) ((T_SR) reduce 233) (("]") reduce 233) (("^") reduce 233) (("|") reduce 233)) 15528 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 404) (("^") shift) (("|") shift)) 15532 ((("%") reduce 613) (("*") reduce 613) (("+") reduce 613) (("-") reduce 613) ((".") reduce 613) (("/") reduce 613) (("<") reduce 613) ((">") reduce 613) (("?") reduce 613) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_BOOLEAN_AND) reduce 613) ((T_BOOLEAN_OR) reduce 613) ((T_COALESCE) reduce 613) ((T_DOUBLE_ARROW) reduce 613) ((T_INSTANCEOF) reduce 613) ((T_IS_EQUAL) reduce 613) ((T_IS_GREATER_OR_EQUAL) reduce 613) ((T_IS_IDENTICAL) reduce 613) ((T_IS_NOT_EQUAL) reduce 613) ((T_IS_NOT_IDENTICAL) reduce 613) ((T_IS_SMALLER_OR_EQUAL) reduce 613) ((T_LOGICAL_AND) reduce 613) ((T_LOGICAL_OR) reduce 613) ((T_LOGICAL_XOR) reduce 613) ((T_PIPE) reduce 613) ((T_POW) reduce 613) ((T_SL) reduce 613) ((T_SPACESHIP) reduce 613) ((T_SR) reduce 613) (("]") reduce 613) (("^") reduce 613) (("|") reduce 613)) 15533 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_DOUBLE_ARROW) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_PIPE) reduce 467) ((T_POW) shift) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("]") reduce 467) (("^") reduce 467) (("|") reduce 467)) 15534 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 616) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 616) (("^") shift) (("|") shift)) 15535 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 615) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 615) (("^") shift) (("|") shift)) 15536 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_DOUBLE_ARROW) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("[") reduce 548) (("]") reduce 424) (("^") reduce 424) (("|") reduce 424)) 15538 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_PIPE) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("]") reduce 518) (("^") reduce 518) (("|") reduce 518)) 15539 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_DOUBLE_ARROW) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_PIPE) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("]") reduce 474) (("^") reduce 474) (("|") reduce 474)) 15540 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("]") reduce 291) (("^") reduce 291) (("|") reduce 291)) 15544 ((("%") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("]") reduce 292) (("^") reduce 292) (("|") reduce 292)) 15545 ((("%") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("]") reduce 293) (("^") reduce 293) (("|") reduce 293)) 15548 ((("%") reduce 617) (("*") reduce 617) (("+") reduce 617) (("-") reduce 617) ((".") reduce 617) (("/") reduce 617) (("<") reduce 617) ((">") reduce 617) (("?") reduce 617) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_BOOLEAN_AND) reduce 617) ((T_BOOLEAN_OR) reduce 617) ((T_COALESCE) reduce 617) ((T_DOUBLE_ARROW) reduce 617) ((T_INSTANCEOF) reduce 617) ((T_IS_EQUAL) reduce 617) ((T_IS_GREATER_OR_EQUAL) reduce 617) ((T_IS_IDENTICAL) reduce 617) ((T_IS_NOT_EQUAL) reduce 617) ((T_IS_NOT_IDENTICAL) reduce 617) ((T_IS_SMALLER_OR_EQUAL) reduce 617) ((T_LOGICAL_AND) reduce 617) ((T_LOGICAL_OR) reduce 617) ((T_LOGICAL_XOR) reduce 617) ((T_PIPE) reduce 617) ((T_POW) reduce 617) ((T_SL) reduce 617) ((T_SPACESHIP) reduce 617) ((T_SR) reduce 617) (("]") reduce 617) (("^") reduce 617) (("|") reduce 617)) 15551 ((("%") reduce 614) (("*") reduce 614) (("+") reduce 614) (("-") reduce 614) ((".") reduce 614) (("/") reduce 614) (("<") reduce 614) ((">") reduce 614) (("?") reduce 614) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_BOOLEAN_AND) reduce 614) ((T_BOOLEAN_OR) reduce 614) ((T_COALESCE) reduce 614) ((T_DOUBLE_ARROW) reduce 614) ((T_INSTANCEOF) reduce 614) ((T_IS_EQUAL) reduce 614) ((T_IS_GREATER_OR_EQUAL) reduce 614) ((T_IS_IDENTICAL) reduce 614) ((T_IS_NOT_EQUAL) reduce 614) ((T_IS_NOT_IDENTICAL) reduce 614) ((T_IS_SMALLER_OR_EQUAL) reduce 614) ((T_LOGICAL_AND) reduce 614) ((T_LOGICAL_OR) reduce 614) ((T_LOGICAL_XOR) reduce 614) ((T_PIPE) reduce 614) ((T_POW) reduce 614) ((T_SL) reduce 614) ((T_SPACESHIP) reduce 614) ((T_SR) reduce 614) (("]") reduce 614) (("^") reduce 614) (("|") reduce 614)) 15552 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_DOUBLE_ARROW) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_PIPE) reduce 468) ((T_POW) shift) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("]") reduce 468) (("^") reduce 468) (("|") reduce 468)) 15553 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_DOUBLE_ARROW) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("[") reduce 548) (("]") reduce 426) (("^") reduce 426) (("|") reduce 426)) 15555 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_DOUBLE_ARROW) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_PIPE) reduce 408) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("]") reduce 408) (("^") reduce 408) (("|") reduce 408)) 15556 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_DOUBLE_ARROW) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_PIPE) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("]") reduce 409) (("^") reduce 409) (("|") reduce 409)) 15557 ((("%") reduce 296) (("*") reduce 296) (("+") reduce 296) (("-") reduce 296) ((".") reduce 296) (("/") reduce 296) (("<") reduce 296) ((">") reduce 296) (("?") reduce 296) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_BOOLEAN_AND) reduce 296) ((T_BOOLEAN_OR) reduce 296) ((T_COALESCE) reduce 296) ((T_DOUBLE_ARROW) reduce 296) ((T_INSTANCEOF) reduce 296) ((T_IS_EQUAL) reduce 296) ((T_IS_GREATER_OR_EQUAL) reduce 296) ((T_IS_IDENTICAL) reduce 296) ((T_IS_NOT_EQUAL) reduce 296) ((T_IS_NOT_IDENTICAL) reduce 296) ((T_IS_SMALLER_OR_EQUAL) reduce 296) ((T_LOGICAL_AND) reduce 296) ((T_LOGICAL_OR) reduce 296) ((T_LOGICAL_XOR) reduce 296) ((T_PIPE) reduce 296) ((T_POW) reduce 296) ((T_SL) reduce 296) ((T_SPACESHIP) reduce 296) ((T_SR) reduce 296) (("]") reduce 296) (("^") reduce 296) (("|") reduce 296)) 15562 ((("%") reduce 297) (("*") reduce 297) (("+") reduce 297) (("-") reduce 297) ((".") reduce 297) (("/") reduce 297) (("<") reduce 297) ((">") reduce 297) (("?") reduce 297) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_BOOLEAN_AND) reduce 297) ((T_BOOLEAN_OR) reduce 297) ((T_COALESCE) reduce 297) ((T_DOUBLE_ARROW) reduce 297) ((T_INSTANCEOF) reduce 297) ((T_IS_EQUAL) reduce 297) ((T_IS_GREATER_OR_EQUAL) reduce 297) ((T_IS_IDENTICAL) reduce 297) ((T_IS_NOT_EQUAL) reduce 297) ((T_IS_NOT_IDENTICAL) reduce 297) ((T_IS_SMALLER_OR_EQUAL) reduce 297) ((T_LOGICAL_AND) reduce 297) ((T_LOGICAL_OR) reduce 297) ((T_LOGICAL_XOR) reduce 297) ((T_PIPE) reduce 297) ((T_POW) reduce 297) ((T_SL) reduce 297) ((T_SPACESHIP) reduce 297) ((T_SR) reduce 297) (("]") reduce 297) (("^") reduce 297) (("|") reduce 297)) 15563 ((("%") reduce 460) (("(") reduce 556) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PIPE) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("[") reduce 549) (("]") reduce 460) (("^") reduce 460) (("|") reduce 460)) 15565 ((("%") reduce 298) (("*") reduce 298) (("+") reduce 298) (("-") reduce 298) ((".") reduce 298) (("/") reduce 298) (("<") reduce 298) ((">") reduce 298) (("?") reduce 298) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_BOOLEAN_AND) reduce 298) ((T_BOOLEAN_OR) reduce 298) ((T_COALESCE) reduce 298) ((T_DOUBLE_ARROW) reduce 298) ((T_INSTANCEOF) reduce 298) ((T_IS_EQUAL) reduce 298) ((T_IS_GREATER_OR_EQUAL) reduce 298) ((T_IS_IDENTICAL) reduce 298) ((T_IS_NOT_EQUAL) reduce 298) ((T_IS_NOT_IDENTICAL) reduce 298) ((T_IS_SMALLER_OR_EQUAL) reduce 298) ((T_LOGICAL_AND) reduce 298) ((T_LOGICAL_OR) reduce 298) ((T_LOGICAL_XOR) reduce 298) ((T_PIPE) reduce 298) ((T_POW) reduce 298) ((T_SL) reduce 298) ((T_SPACESHIP) reduce 298) ((T_SR) reduce 298) (("]") reduce 298) (("^") reduce 298) (("|") reduce 298)) 15566 ((("%") reduce 299) (("*") reduce 299) (("+") reduce 299) (("-") reduce 299) ((".") reduce 299) (("/") reduce 299) (("<") reduce 299) ((">") reduce 299) (("?") reduce 299) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_BOOLEAN_AND) reduce 299) ((T_BOOLEAN_OR) reduce 299) ((T_COALESCE) reduce 299) ((T_DOUBLE_ARROW) reduce 299) ((T_INSTANCEOF) reduce 299) ((T_IS_EQUAL) reduce 299) ((T_IS_GREATER_OR_EQUAL) reduce 299) ((T_IS_IDENTICAL) reduce 299) ((T_IS_NOT_EQUAL) reduce 299) ((T_IS_NOT_IDENTICAL) reduce 299) ((T_IS_SMALLER_OR_EQUAL) reduce 299) ((T_LOGICAL_AND) reduce 299) ((T_LOGICAL_OR) reduce 299) ((T_LOGICAL_XOR) reduce 299) ((T_PIPE) reduce 299) ((T_POW) reduce 299) ((T_SL) reduce 299) ((T_SPACESHIP) reduce 299) ((T_SR) reduce 299) (("]") reduce 299) (("^") reduce 299) (("|") reduce 299)) 15567 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_DOUBLE_ARROW) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_PIPE) reduce 472) ((T_POW) shift) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("]") reduce 472) (("^") reduce 472) (("|") reduce 472)) 15568 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_DOUBLE_ARROW) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_PIPE) reduce 470) ((T_POW) shift) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("]") reduce 470) (("^") reduce 470) (("|") reduce 470)) 15571 ((("%") reduce 519) (("(") reduce 519) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PIPE) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("]") reduce 519) (("^") reduce 519) (("|") reduce 519)) 15572 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_DOUBLE_ARROW) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_PIPE) reduce 475) ((T_POW) shift) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("]") reduce 475) (("^") reduce 475) (("|") reduce 475)) 15573 ((("%") reduce 446) (("*") reduce 446) (("+") reduce 446) (("-") reduce 446) ((".") reduce 446) (("/") reduce 446) (("<") reduce 446) ((">") reduce 446) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_BOOLEAN_AND) reduce 446) ((T_BOOLEAN_OR) reduce 446) ((T_COALESCE) reduce 446) ((T_DOUBLE_ARROW) reduce 446) ((T_INSTANCEOF) reduce 446) ((T_IS_EQUAL) reduce 446) ((T_IS_GREATER_OR_EQUAL) reduce 446) ((T_IS_IDENTICAL) reduce 446) ((T_IS_NOT_EQUAL) reduce 446) ((T_IS_NOT_IDENTICAL) reduce 446) ((T_IS_SMALLER_OR_EQUAL) reduce 446) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_PIPE) reduce 446) ((T_POW) shift) ((T_SL) reduce 446) ((T_SPACESHIP) reduce 446) ((T_SR) reduce 446) (("]") reduce 446) (("^") reduce 446) (("|") reduce 446)) 15574 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_DOUBLE_ARROW) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_PIPE) reduce 445) ((T_POW) shift) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("]") reduce 445) (("^") reduce 445) (("|") reduce 445)) 15576 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) (("=") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AND_EQUAL) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_COALESCE_EQUAL) reduce 570) ((T_CONCAT_EQUAL) reduce 570) ((T_DEC) reduce 570) ((T_DIV_EQUAL) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INC) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_MINUS_EQUAL) reduce 570) ((T_MOD_EQUAL) reduce 570) ((T_MUL_EQUAL) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_OR_EQUAL) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_PLUS_EQUAL) reduce 570) ((T_POW) reduce 570) ((T_POW_EQUAL) reduce 570) ((T_SL) reduce 570) ((T_SL_EQUAL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) ((T_SR_EQUAL) reduce 570) ((T_XOR_EQUAL) reduce 570) (("[") reduce 570) (("]") reduce 570) (("^") reduce 570) (("|") reduce 570)) 15579 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) (("=") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AND_EQUAL) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_COALESCE_EQUAL) reduce 569) ((T_CONCAT_EQUAL) reduce 569) ((T_DEC) reduce 569) ((T_DIV_EQUAL) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INC) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_MINUS_EQUAL) reduce 569) ((T_MOD_EQUAL) reduce 569) ((T_MUL_EQUAL) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_OR_EQUAL) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_PLUS_EQUAL) reduce 569) ((T_POW) reduce 569) ((T_POW_EQUAL) reduce 569) ((T_SL) reduce 569) ((T_SL_EQUAL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) ((T_SR_EQUAL) reduce 569) ((T_XOR_EQUAL) reduce 569) (("[") reduce 569) (("]") reduce 569) (("^") reduce 569) (("|") reduce 569)) 15581 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_PIPE) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("]") reduce 522) (("^") reduce 522) (("|") reduce 522)) 15582 ((("%") reduce 447) (("*") reduce 447) (("+") reduce 447) (("-") reduce 447) ((".") reduce 447) (("/") reduce 447) (("<") reduce 447) ((">") reduce 447) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_BOOLEAN_AND) reduce 447) ((T_BOOLEAN_OR) reduce 447) ((T_COALESCE) reduce 447) ((T_DOUBLE_ARROW) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 447) ((T_IS_GREATER_OR_EQUAL) reduce 447) ((T_IS_IDENTICAL) reduce 447) ((T_IS_NOT_EQUAL) reduce 447) ((T_IS_NOT_IDENTICAL) reduce 447) ((T_IS_SMALLER_OR_EQUAL) reduce 447) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_PIPE) reduce 447) ((T_POW) shift) ((T_SL) reduce 447) ((T_SPACESHIP) reduce 447) ((T_SR) reduce 447) (("]") reduce 447) (("^") reduce 447) (("|") reduce 447)) 15583 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_PIPE) reduce 473) ((T_POW) shift) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("]") reduce 473) (("^") reduce 473) (("|") reduce 473)) 15584 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 483) (("^") shift) (("|") shift)) 15585 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_PIPE) reduce 469) ((T_POW) shift) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("]") reduce 469) (("^") reduce 469) (("|") reduce 469)) 15586 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_PIPE) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("]") reduce 486) (("^") reduce 486) (("|") reduce 486)) 15588 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_PIPE) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("]") reduce 526) (("^") reduce 526) (("|") reduce 526)) 15590 ((("%") reduce 527) (("*") reduce 527) (("+") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_PIPE) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("]") reduce 527) (("^") reduce 527) (("|") reduce 527)) 15591 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_PIPE) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("]") reduce 525) (("^") reduce 525) (("|") reduce 525)) 15592 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 619) (("^") shift) (("|") shift)) 15593 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 618) (("^") shift) (("|") shift)) 15594 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("]") reduce 505) (("^") reduce 505) (("|") reduce 505)) 15595 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 478) (("^") shift) (("|") shift)) 15596 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_PIPE) reduce 471) ((T_POW) shift) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("]") reduce 471) (("^") reduce 471) (("|") reduce 471)) 15598 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("]") reduce 94) (("^") reduce 94) (("|") reduce 94)) 15599 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("]") reduce 93) (("^") reduce 93) (("|") reduce 93)) 15600 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("]") reduce 95) (("^") reduce 95) (("|") reduce 95)) 15601 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("]") reduce 509) (("^") reduce 509) (("|") reduce 509)) 15602 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("]") reduce 92) (("^") reduce 92) (("|") reduce 92)) 15603 ((("%") reduce 400) (("(") reduce 400) (("*") reduce 400) (("+") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 400) ((T_OBJECT_OPERATOR) reduce 400) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 400) ((T_PIPE) reduce 400) ((T_POW) reduce 400) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("[") reduce 400) (("]") reduce 400) (("^") reduce 400) (("|") reduce 400)) 15606 ((("%") reduce 511) (("(") reduce 511) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("]") reduce 511) (("^") reduce 511) (("|") reduce 511)) 15607 ((("%") reduce 402) (("(") shift) (("*") reduce 402) (("+") reduce 402) (("-") reduce 402) ((".") reduce 402) (("/") reduce 402) (("<") reduce 402) ((">") reduce 402) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_INSTANCEOF) reduce 402) ((T_IS_EQUAL) reduce 402) ((T_IS_GREATER_OR_EQUAL) reduce 402) ((T_IS_IDENTICAL) reduce 402) ((T_IS_NOT_EQUAL) reduce 402) ((T_IS_NOT_IDENTICAL) reduce 402) ((T_IS_SMALLER_OR_EQUAL) reduce 402) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_PIPE) reduce 402) ((T_POW) reduce 402) ((T_SL) reduce 402) ((T_SPACESHIP) reduce 402) ((T_SR) reduce 402) (("]") reduce 402) (("^") reduce 402) (("|") reduce 402)) 15608 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("]") reduce 510) (("^") reduce 510) (("|") reduce 510)) 15609 ((("%") reduce 512) (("(") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("]") reduce 512) (("^") reduce 512) (("|") reduce 512)) 15610 ((("%") reduce 573) (("(") reduce 573) (("*") reduce 573) (("+") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("]") reduce 573) (("^") reduce 573) (("|") reduce 573)) 15616 ((("%") reduce 574) (("(") reduce 574) (("*") reduce 574) (("+") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("]") reduce 574) (("^") reduce 574) (("|") reduce 574)) 15617 ((("%") reduce 578) (("(") reduce 578) (("*") reduce 578) (("+") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("]") reduce 578) (("^") reduce 578) (("|") reduce 578)) 15618 ((("%") reduce 575) (("(") reduce 575) (("*") reduce 575) (("+") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("]") reduce 575) (("^") reduce 575) (("|") reduce 575)) 15619 ((("%") reduce 576) (("(") reduce 576) (("*") reduce 576) (("+") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("]") reduce 576) (("^") reduce 576) (("|") reduce 576)) 15620 ((("%") reduce 399) (("(") reduce 399) (("*") reduce 399) (("+") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_INSTANCEOF) reduce 399) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 399) ((T_OBJECT_OPERATOR) reduce 399) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 399) ((T_PIPE) reduce 399) ((T_POW) reduce 399) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("[") reduce 399) (("]") reduce 399) (("^") reduce 399) (("|") reduce 399)) 15622 ((("%") reduce 577) (("(") reduce 577) (("*") reduce 577) (("+") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("]") reduce 577) (("^") reduce 577) (("|") reduce 577)) 15623 ((("%") reduce 401) (("(") reduce 401) (("*") reduce 401) (("+") reduce 401) (("-") reduce 401) ((".") reduce 401) (("/") reduce 401) (("<") reduce 401) ((">") reduce 401) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_INSTANCEOF) reduce 401) ((T_IS_EQUAL) reduce 401) ((T_IS_GREATER_OR_EQUAL) reduce 401) ((T_IS_IDENTICAL) reduce 401) ((T_IS_NOT_EQUAL) reduce 401) ((T_IS_NOT_IDENTICAL) reduce 401) ((T_IS_SMALLER_OR_EQUAL) reduce 401) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 401) ((T_OBJECT_OPERATOR) reduce 401) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 401) ((T_PIPE) reduce 401) ((T_POW) reduce 401) ((T_SL) reduce 401) ((T_SPACESHIP) reduce 401) ((T_SR) reduce 401) (("[") reduce 401) (("]") reduce 401) (("^") reduce 401) (("|") reduce 401)) 15631 ((("%") reduce 398) (("(") reduce 398) (("*") reduce 398) (("+") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 398) ((T_OBJECT_OPERATOR) reduce 398) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 398) ((T_PIPE) reduce 398) ((T_POW) reduce 398) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("[") reduce 398) (("]") reduce 398) (("^") reduce 398) (("|") reduce 398)) 15633 ((("%") reduce 513) (("(") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("]") reduce 513) (("^") reduce 513) (("|") reduce 513)) 15639 ((("%") reduce 233) (("*") reduce 233) (("+") reduce 233) (("-") reduce 233) ((".") reduce 233) (("/") reduce 233) (("<") reduce 233) ((">") reduce 233) (("?") reduce 233) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_BOOLEAN_AND) reduce 233) ((T_BOOLEAN_OR) reduce 233) ((T_COALESCE) reduce 233) ((T_INSTANCEOF) reduce 233) ((T_IS_EQUAL) reduce 233) ((T_IS_GREATER_OR_EQUAL) reduce 233) ((T_IS_IDENTICAL) reduce 233) ((T_IS_NOT_EQUAL) reduce 233) ((T_IS_NOT_IDENTICAL) reduce 233) ((T_IS_SMALLER_OR_EQUAL) reduce 233) ((T_LOGICAL_AND) reduce 233) ((T_LOGICAL_OR) reduce 233) ((T_LOGICAL_XOR) reduce 233) ((T_PIPE) reduce 233) ((T_POW) reduce 233) ((T_SL) reduce 233) ((T_SPACESHIP) reduce 233) ((T_SR) reduce 233) (("]") reduce 233) (("^") reduce 233) (("|") reduce 233)) 15644 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 404) (("^") shift) (("|") shift)) 15648 ((("%") reduce 613) (("*") reduce 613) (("+") reduce 613) (("-") reduce 613) ((".") reduce 613) (("/") reduce 613) (("<") reduce 613) ((">") reduce 613) (("?") reduce 613) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_BOOLEAN_AND) reduce 613) ((T_BOOLEAN_OR) reduce 613) ((T_COALESCE) reduce 613) ((T_INSTANCEOF) reduce 613) ((T_IS_EQUAL) reduce 613) ((T_IS_GREATER_OR_EQUAL) reduce 613) ((T_IS_IDENTICAL) reduce 613) ((T_IS_NOT_EQUAL) reduce 613) ((T_IS_NOT_IDENTICAL) reduce 613) ((T_IS_SMALLER_OR_EQUAL) reduce 613) ((T_LOGICAL_AND) reduce 613) ((T_LOGICAL_OR) reduce 613) ((T_LOGICAL_XOR) reduce 613) ((T_PIPE) reduce 613) ((T_POW) reduce 613) ((T_SL) reduce 613) ((T_SPACESHIP) reduce 613) ((T_SR) reduce 613) (("]") reduce 613) (("^") reduce 613) (("|") reduce 613)) 15649 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_PIPE) reduce 467) ((T_POW) shift) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("]") reduce 467) (("^") reduce 467) (("|") reduce 467)) 15650 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 616) (("^") shift) (("|") shift)) 15651 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("]") reduce 615) (("^") shift) (("|") shift)) 15652 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("[") reduce 548) (("]") reduce 424) (("^") reduce 424) (("|") reduce 424)) 15654 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_PIPE) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("]") reduce 518) (("^") reduce 518) (("|") reduce 518)) 15655 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_PIPE) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("]") reduce 474) (("^") reduce 474) (("|") reduce 474)) 15656 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("]") reduce 291) (("^") reduce 291) (("|") reduce 291)) 15660 ((("%") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("]") reduce 292) (("^") reduce 292) (("|") reduce 292)) 15661 ((("%") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("]") reduce 293) (("^") reduce 293) (("|") reduce 293)) 15664 ((("%") reduce 617) (("*") reduce 617) (("+") reduce 617) (("-") reduce 617) ((".") reduce 617) (("/") reduce 617) (("<") reduce 617) ((">") reduce 617) (("?") reduce 617) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_BOOLEAN_AND) reduce 617) ((T_BOOLEAN_OR) reduce 617) ((T_COALESCE) reduce 617) ((T_INSTANCEOF) reduce 617) ((T_IS_EQUAL) reduce 617) ((T_IS_GREATER_OR_EQUAL) reduce 617) ((T_IS_IDENTICAL) reduce 617) ((T_IS_NOT_EQUAL) reduce 617) ((T_IS_NOT_IDENTICAL) reduce 617) ((T_IS_SMALLER_OR_EQUAL) reduce 617) ((T_LOGICAL_AND) reduce 617) ((T_LOGICAL_OR) reduce 617) ((T_LOGICAL_XOR) reduce 617) ((T_PIPE) reduce 617) ((T_POW) reduce 617) ((T_SL) reduce 617) ((T_SPACESHIP) reduce 617) ((T_SR) reduce 617) (("]") reduce 617) (("^") reduce 617) (("|") reduce 617)) 15667 ((("%") reduce 614) (("*") reduce 614) (("+") reduce 614) (("-") reduce 614) ((".") reduce 614) (("/") reduce 614) (("<") reduce 614) ((">") reduce 614) (("?") reduce 614) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_BOOLEAN_AND) reduce 614) ((T_BOOLEAN_OR) reduce 614) ((T_COALESCE) reduce 614) ((T_INSTANCEOF) reduce 614) ((T_IS_EQUAL) reduce 614) ((T_IS_GREATER_OR_EQUAL) reduce 614) ((T_IS_IDENTICAL) reduce 614) ((T_IS_NOT_EQUAL) reduce 614) ((T_IS_NOT_IDENTICAL) reduce 614) ((T_IS_SMALLER_OR_EQUAL) reduce 614) ((T_LOGICAL_AND) reduce 614) ((T_LOGICAL_OR) reduce 614) ((T_LOGICAL_XOR) reduce 614) ((T_PIPE) reduce 614) ((T_POW) reduce 614) ((T_SL) reduce 614) ((T_SPACESHIP) reduce 614) ((T_SR) reduce 614) (("]") reduce 614) (("^") reduce 614) (("|") reduce 614)) 15668 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_PIPE) reduce 468) ((T_POW) shift) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("]") reduce 468) (("^") reduce 468) (("|") reduce 468)) 15669 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("[") reduce 548) (("]") reduce 426) (("^") reduce 426) (("|") reduce 426)) 15671 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_PIPE) reduce 408) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("]") reduce 408) (("^") reduce 408) (("|") reduce 408)) 15672 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_PIPE) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("]") reduce 409) (("^") reduce 409) (("|") reduce 409)) 15673 ((("%") reduce 296) (("*") reduce 296) (("+") reduce 296) (("-") reduce 296) ((".") reduce 296) (("/") reduce 296) (("<") reduce 296) ((">") reduce 296) (("?") reduce 296) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_BOOLEAN_AND) reduce 296) ((T_BOOLEAN_OR) reduce 296) ((T_COALESCE) reduce 296) ((T_INSTANCEOF) reduce 296) ((T_IS_EQUAL) reduce 296) ((T_IS_GREATER_OR_EQUAL) reduce 296) ((T_IS_IDENTICAL) reduce 296) ((T_IS_NOT_EQUAL) reduce 296) ((T_IS_NOT_IDENTICAL) reduce 296) ((T_IS_SMALLER_OR_EQUAL) reduce 296) ((T_LOGICAL_AND) reduce 296) ((T_LOGICAL_OR) reduce 296) ((T_LOGICAL_XOR) reduce 296) ((T_PIPE) reduce 296) ((T_POW) reduce 296) ((T_SL) reduce 296) ((T_SPACESHIP) reduce 296) ((T_SR) reduce 296) (("]") reduce 296) (("^") reduce 296) (("|") reduce 296)) 15678 ((("%") reduce 297) (("*") reduce 297) (("+") reduce 297) (("-") reduce 297) ((".") reduce 297) (("/") reduce 297) (("<") reduce 297) ((">") reduce 297) (("?") reduce 297) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_BOOLEAN_AND) reduce 297) ((T_BOOLEAN_OR) reduce 297) ((T_COALESCE) reduce 297) ((T_INSTANCEOF) reduce 297) ((T_IS_EQUAL) reduce 297) ((T_IS_GREATER_OR_EQUAL) reduce 297) ((T_IS_IDENTICAL) reduce 297) ((T_IS_NOT_EQUAL) reduce 297) ((T_IS_NOT_IDENTICAL) reduce 297) ((T_IS_SMALLER_OR_EQUAL) reduce 297) ((T_LOGICAL_AND) reduce 297) ((T_LOGICAL_OR) reduce 297) ((T_LOGICAL_XOR) reduce 297) ((T_PIPE) reduce 297) ((T_POW) reduce 297) ((T_SL) reduce 297) ((T_SPACESHIP) reduce 297) ((T_SR) reduce 297) (("]") reduce 297) (("^") reduce 297) (("|") reduce 297)) 15679 ((("%") reduce 460) (("(") reduce 556) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PIPE) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("[") reduce 549) (("]") reduce 460) (("^") reduce 460) (("|") reduce 460)) 15681 ((("%") reduce 298) (("*") reduce 298) (("+") reduce 298) (("-") reduce 298) ((".") reduce 298) (("/") reduce 298) (("<") reduce 298) ((">") reduce 298) (("?") reduce 298) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_BOOLEAN_AND) reduce 298) ((T_BOOLEAN_OR) reduce 298) ((T_COALESCE) reduce 298) ((T_INSTANCEOF) reduce 298) ((T_IS_EQUAL) reduce 298) ((T_IS_GREATER_OR_EQUAL) reduce 298) ((T_IS_IDENTICAL) reduce 298) ((T_IS_NOT_EQUAL) reduce 298) ((T_IS_NOT_IDENTICAL) reduce 298) ((T_IS_SMALLER_OR_EQUAL) reduce 298) ((T_LOGICAL_AND) reduce 298) ((T_LOGICAL_OR) reduce 298) ((T_LOGICAL_XOR) reduce 298) ((T_PIPE) reduce 298) ((T_POW) reduce 298) ((T_SL) reduce 298) ((T_SPACESHIP) reduce 298) ((T_SR) reduce 298) (("]") reduce 298) (("^") reduce 298) (("|") reduce 298)) 15682 ((("%") reduce 299) (("*") reduce 299) (("+") reduce 299) (("-") reduce 299) ((".") reduce 299) (("/") reduce 299) (("<") reduce 299) ((">") reduce 299) (("?") reduce 299) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_BOOLEAN_AND) reduce 299) ((T_BOOLEAN_OR) reduce 299) ((T_COALESCE) reduce 299) ((T_INSTANCEOF) reduce 299) ((T_IS_EQUAL) reduce 299) ((T_IS_GREATER_OR_EQUAL) reduce 299) ((T_IS_IDENTICAL) reduce 299) ((T_IS_NOT_EQUAL) reduce 299) ((T_IS_NOT_IDENTICAL) reduce 299) ((T_IS_SMALLER_OR_EQUAL) reduce 299) ((T_LOGICAL_AND) reduce 299) ((T_LOGICAL_OR) reduce 299) ((T_LOGICAL_XOR) reduce 299) ((T_PIPE) reduce 299) ((T_POW) reduce 299) ((T_SL) reduce 299) ((T_SPACESHIP) reduce 299) ((T_SR) reduce 299) (("]") reduce 299) (("^") reduce 299) (("|") reduce 299)) 15683 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_PIPE) reduce 472) ((T_POW) shift) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("]") reduce 472) (("^") reduce 472) (("|") reduce 472)) 15684 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_PIPE) reduce 470) ((T_POW) shift) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("]") reduce 470) (("^") reduce 470) (("|") reduce 470)) 15687 ((("%") reduce 519) (("(") reduce 519) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PIPE) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("]") reduce 519) (("^") reduce 519) (("|") reduce 519)) 15688 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_PIPE) reduce 475) ((T_POW) shift) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("]") reduce 475) (("^") reduce 475) (("|") reduce 475)) 15689 ((("%") reduce 446) (("*") reduce 446) (("+") reduce 446) (("-") reduce 446) ((".") reduce 446) (("/") reduce 446) (("<") reduce 446) ((">") reduce 446) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_BOOLEAN_AND) reduce 446) ((T_BOOLEAN_OR) reduce 446) ((T_COALESCE) reduce 446) ((T_INSTANCEOF) reduce 446) ((T_IS_EQUAL) reduce 446) ((T_IS_GREATER_OR_EQUAL) reduce 446) ((T_IS_IDENTICAL) reduce 446) ((T_IS_NOT_EQUAL) reduce 446) ((T_IS_NOT_IDENTICAL) reduce 446) ((T_IS_SMALLER_OR_EQUAL) reduce 446) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_PIPE) reduce 446) ((T_POW) shift) ((T_SL) reduce 446) ((T_SPACESHIP) reduce 446) ((T_SR) reduce 446) (("]") reduce 446) (("^") reduce 446) (("|") reduce 446)) 15690 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_PIPE) reduce 445) ((T_POW) shift) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("]") reduce 445) (("^") reduce 445) (("|") reduce 445)) 15692 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) (("=") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AND_EQUAL) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_COALESCE_EQUAL) reduce 570) ((T_CONCAT_EQUAL) reduce 570) ((T_DEC) reduce 570) ((T_DIV_EQUAL) reduce 570) ((T_INC) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_MINUS_EQUAL) reduce 570) ((T_MOD_EQUAL) reduce 570) ((T_MUL_EQUAL) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_OR_EQUAL) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_PLUS_EQUAL) reduce 570) ((T_POW) reduce 570) ((T_POW_EQUAL) reduce 570) ((T_SL) reduce 570) ((T_SL_EQUAL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) ((T_SR_EQUAL) reduce 570) ((T_XOR_EQUAL) reduce 570) (("[") reduce 570) (("]") reduce 570) (("^") reduce 570) (("|") reduce 570)) 15695 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) (("=") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AND_EQUAL) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_COALESCE_EQUAL) reduce 569) ((T_CONCAT_EQUAL) reduce 569) ((T_DEC) reduce 569) ((T_DIV_EQUAL) reduce 569) ((T_INC) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_MINUS_EQUAL) reduce 569) ((T_MOD_EQUAL) reduce 569) ((T_MUL_EQUAL) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_OR_EQUAL) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_PLUS_EQUAL) reduce 569) ((T_POW) reduce 569) ((T_POW_EQUAL) reduce 569) ((T_SL) reduce 569) ((T_SL_EQUAL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) ((T_SR_EQUAL) reduce 569) ((T_XOR_EQUAL) reduce 569) (("[") reduce 569) (("]") reduce 569) (("^") reduce 569) (("|") reduce 569)) 15697 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_PIPE) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("]") reduce 522) (("^") reduce 522) (("|") reduce 522)) 15698 ((("%") reduce 447) (("*") reduce 447) (("+") reduce 447) (("-") reduce 447) ((".") reduce 447) (("/") reduce 447) (("<") reduce 447) ((">") reduce 447) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_BOOLEAN_AND) reduce 447) ((T_BOOLEAN_OR) reduce 447) ((T_COALESCE) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 447) ((T_IS_GREATER_OR_EQUAL) reduce 447) ((T_IS_IDENTICAL) reduce 447) ((T_IS_NOT_EQUAL) reduce 447) ((T_IS_NOT_IDENTICAL) reduce 447) ((T_IS_SMALLER_OR_EQUAL) reduce 447) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_PIPE) reduce 447) ((T_POW) shift) ((T_SL) reduce 447) ((T_SPACESHIP) reduce 447) ((T_SR) reduce 447) (("]") reduce 447) (("^") reduce 447) (("|") reduce 447)) 16382 ((("%") reduce 578) ((")") reduce 578) (("*") reduce 578) (("+") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("|") reduce 578)) 16383 ((("%") reduce 582) ((")") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 16384 ((("%") reduce 575) ((")") reduce 575) (("*") reduce 575) (("+") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("|") reduce 575)) 16385 ((("%") reduce 584) ((")") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 16388 ((("%") reduce 583) ((")") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 16389 ((("%") reduce 576) ((")") reduce 576) (("*") reduce 576) (("+") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("|") reduce 576)) 16391 ((("%") reduce 577) ((")") reduce 577) (("*") reduce 577) (("+") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("|") reduce 577)) 16393 ((("%") reduce 513) ((")") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("|") reduce 513)) 16394 ((("%") reduce 570) ((")") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 16397 ((("%") reduce 569) ((")") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 16398 ((("%") shift) ((")") reduce 465) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 465) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16399 ((("%") shift) ((")") reduce 427) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16400 ((("%") shift) ((")") reduce 428) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16401 ((("%") shift) ((")") reduce 433) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("|") reduce 433)) 16402 ((("%") shift) ((")") reduce 434) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("|") reduce 434)) 16406 ((("%") shift) ((")") reduce 463) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16407 ((("%") shift) ((")") reduce 464) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16408 ((("%") shift) ((")") reduce 456) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 456) ((T_SR) shift) (("^") reduce 456) (("|") reduce 456)) 16409 ((("%") shift) ((")") reduce 454) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 454) ((T_SR) shift) (("^") reduce 454) (("|") reduce 454)) 16410 ((("%") reduce 441) ((")") reduce 441) (("*") reduce 441) (("+") reduce 441) (("-") reduce 441) ((".") reduce 441) (("/") reduce 441) (("<") reduce 441) ((">") reduce 441) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 441) ((T_IS_GREATER_OR_EQUAL) reduce 441) ((T_IS_IDENTICAL) reduce 441) ((T_IS_NOT_EQUAL) reduce 441) ((T_IS_NOT_IDENTICAL) reduce 441) ((T_IS_SMALLER_OR_EQUAL) reduce 441) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_PIPE) reduce 441) ((T_POW) shift) ((T_SL) reduce 441) ((T_SPACESHIP) reduce 441) ((T_SR) reduce 441) (("^") reduce 441) (("|") reduce 441)) 16411 ((("%") shift) ((")") reduce 436) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 436) (("/") shift) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_PIPE) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 436) ((T_SR) shift) (("^") reduce 436) (("|") reduce 436)) 16412 ((("%") shift) ((")") reduce 438) (("*") shift) (("+") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") shift) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_PIPE) reduce 438) ((T_POW) shift) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("|") reduce 438)) 16413 ((("%") shift) ((")") reduce 437) (("*") shift) (("+") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") shift) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_PIPE) reduce 437) ((T_POW) shift) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("|") reduce 437)) 16414 ((("%") reduce 439) ((")") reduce 439) (("*") reduce 439) (("+") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_PIPE) reduce 439) ((T_POW) shift) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("|") reduce 439)) 16415 ((("%") reduce 442) ((")") reduce 442) (("*") reduce 442) (("+") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_PIPE) reduce 442) ((T_POW) shift) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("|") reduce 442)) 16417 ((("%") reduce 74) (("(") reduce 74) ((")") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_PIPE) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("|") reduce 74)) 16418 ((("%") reduce 52) (("(") reduce 52) ((")") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_PIPE) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("|") reduce 52)) 16419 ((("%") reduce 27) (("(") reduce 27) ((")") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_PIPE) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("|") reduce 27)) 16420 ((("%") reduce 51) (("(") reduce 51) ((")") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_PIPE) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("|") reduce 51)) 16421 ((("%") reduce 53) (("(") reduce 53) ((")") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_PIPE) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("|") reduce 53)) 16422 ((("%") reduce 49) (("(") reduce 49) ((")") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_PIPE) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("|") reduce 49)) 16423 ((("%") reduce 29) (("(") reduce 29) ((")") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_PIPE) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("|") reduce 29)) 16424 ((("%") reduce 59) (("(") reduce 59) ((")") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_PIPE) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("|") reduce 59)) 16425 ((("%") reduce 60) (("(") reduce 60) ((")") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_PIPE) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("|") reduce 60)) 16426 ((("%") reduce 11) (("(") reduce 11) ((")") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_PIPE) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("|") reduce 11)) 16427 ((("%") reduce 42) (("(") reduce 42) ((")") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_PIPE) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("|") reduce 42)) 16428 ((("%") reduce 39) (("(") reduce 39) ((")") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_PIPE) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("|") reduce 39)) 16429 ((("%") reduce 25) (("(") reduce 25) ((")") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_PIPE) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("|") reduce 25)) 16430 ((("%") reduce 50) (("(") reduce 50) ((")") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_PIPE) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("|") reduce 50)) 16431 ((("%") reduce 66) (("(") reduce 66) ((")") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_PIPE) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("|") reduce 66)) 16432 ((("%") reduce 18) (("(") reduce 18) ((")") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_PIPE) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("|") reduce 18)) 16433 ((("%") reduce 17) (("(") reduce 17) ((")") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_PIPE) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("|") reduce 17)) 16434 ((("%") reduce 15) (("(") reduce 15) ((")") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_PIPE) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("|") reduce 15)) 16435 ((("%") reduce 14) (("(") reduce 14) ((")") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_PIPE) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("|") reduce 14)) 16436 ((("%") reduce 38) (("(") reduce 38) ((")") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_PIPE) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("|") reduce 38)) 16437 ((("%") reduce 26) (("(") reduce 26) ((")") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_PIPE) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("|") reduce 26)) 16438 ((("%") reduce 22) (("(") reduce 22) ((")") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_PIPE) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("|") reduce 22)) 16439 ((("%") reduce 24) (("(") reduce 24) ((")") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_PIPE) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("|") reduce 24)) 16440 ((("%") reduce 16) (("(") reduce 16) ((")") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_PIPE) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("|") reduce 16)) 16441 ((("%") reduce 48) (("(") reduce 48) ((")") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_PIPE) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("|") reduce 48)) 16442 ((("%") reduce 20) (("(") reduce 20) ((")") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_PIPE) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("|") reduce 20)) 16443 ((("%") reduce 70) (("(") reduce 70) ((")") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_PIPE) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("|") reduce 70)) 16444 ((("%") reduce 3) (("(") reduce 3) ((")") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_PIPE) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("|") reduce 3)) 16445 ((("%") reduce 12) (("(") reduce 12) ((")") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_PIPE) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("|") reduce 12)) 16446 ((("%") reduce 54) (("(") reduce 54) ((")") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_PIPE) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("|") reduce 54)) 16447 ((("%") reduce 65) (("(") reduce 65) ((")") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_PIPE) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("|") reduce 65)) 16448 ((("%") reduce 75) (("(") reduce 75) ((")") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_PIPE) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("|") reduce 75)) 16449 ((("%") reduce 30) (("(") reduce 30) ((")") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_PIPE) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("|") reduce 30)) 16450 ((("%") reduce 68) (("(") reduce 68) ((")") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_PIPE) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("|") reduce 68)) 16451 ((("%") reduce 21) (("(") reduce 21) ((")") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_PIPE) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("|") reduce 21)) 16452 ((("%") reduce 23) (("(") reduce 23) ((")") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_PIPE) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("|") reduce 23)) 16453 ((("%") reduce 41) (("(") reduce 41) ((")") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_PIPE) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("|") reduce 41)) 16454 ((("%") reduce 62) (("(") reduce 62) ((")") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_PIPE) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("|") reduce 62)) 16455 ((("%") reduce 34) (("(") reduce 34) ((")") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_PIPE) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("|") reduce 34)) 16456 ((("%") reduce 40) (("(") reduce 40) ((")") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_PIPE) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("|") reduce 40)) 16457 ((("%") reduce 13) (("(") reduce 13) ((")") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_PIPE) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("|") reduce 13)) 16458 ((("%") reduce 55) (("(") reduce 55) ((")") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_PIPE) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("|") reduce 55)) 16459 ((("%") reduce 1) (("(") reduce 1) ((")") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_PIPE) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("|") reduce 1)) 16460 ((("%") reduce 2) (("(") reduce 2) ((")") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_PIPE) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("|") reduce 2)) 16461 ((("%") reduce 9) (("(") reduce 9) ((")") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_PIPE) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("|") reduce 9)) 16462 ((("%") reduce 33) (("(") reduce 33) ((")") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_PIPE) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("|") reduce 33)) 16463 ((("%") reduce 58) (("(") reduce 58) ((")") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_PIPE) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("|") reduce 58)) 16464 ((("%") reduce 37) (("(") reduce 37) ((")") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_PIPE) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("|") reduce 37)) 16465 ((("%") reduce 64) (("(") reduce 64) ((")") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_PIPE) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("|") reduce 64)) 16466 ((("%") reduce 46) (("(") reduce 46) ((")") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_PIPE) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("|") reduce 46)) 16467 ((("%") reduce 8) (("(") reduce 8) ((")") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_PIPE) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("|") reduce 8)) 16468 ((("%") reduce 6) (("(") reduce 6) ((")") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_PIPE) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("|") reduce 6)) 16469 ((("%") reduce 7) (("(") reduce 7) ((")") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_PIPE) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("|") reduce 7)) 16470 ((("%") reduce 69) (("(") reduce 69) ((")") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_PIPE) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("|") reduce 69)) 16471 ((("%") reduce 63) (("(") reduce 63) ((")") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_PIPE) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("|") reduce 63)) 16472 ((("%") reduce 56) (("(") reduce 56) ((")") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_PIPE) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("|") reduce 56)) 16473 ((("%") reduce 10) (("(") reduce 10) ((")") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_PIPE) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("|") reduce 10)) 16474 ((("%") reduce 67) (("(") reduce 67) ((")") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_PIPE) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("|") reduce 67)) 16475 ((("%") reduce 44) (("(") reduce 44) ((")") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_PIPE) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("|") reduce 44)) 16476 ((("%") reduce 76) (("(") reduce 76) ((")") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_PIPE) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("|") reduce 76)) 16477 ((("%") reduce 71) (("(") reduce 71) ((")") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_PIPE) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("|") reduce 71)) 16478 ((("%") reduce 77) (("(") reduce 77) ((")") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_PIPE) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("|") reduce 77)) 16479 ((("%") reduce 78) (("(") reduce 78) ((")") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_PIPE) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("|") reduce 78)) 16480 ((("%") reduce 79) (("(") reduce 79) ((")") reduce 79) (("*") reduce 79) (("+") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_PIPE) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("|") reduce 79)) 16481 ((("%") reduce 4) (("(") reduce 4) ((")") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_PIPE) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("|") reduce 4)) 16482 ((("%") reduce 5) (("(") reduce 5) ((")") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_PIPE) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("|") reduce 5)) 16483 ((("%") reduce 43) (("(") reduce 43) ((")") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_PIPE) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("|") reduce 43)) 16484 ((("%") reduce 73) (("(") reduce 73) ((")") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_PIPE) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("|") reduce 73)) 16485 ((("%") reduce 82) (("(") reduce 82) ((")") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_PIPE) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("|") reduce 82)) 16486 ((("%") reduce 47) (("(") reduce 47) ((")") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_PIPE) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("|") reduce 47)) 16487 ((("%") reduce 31) (("(") reduce 31) ((")") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_PIPE) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("|") reduce 31)) 16488 ((("%") reduce 57) (("(") reduce 57) ((")") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_PIPE) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("|") reduce 57)) 16489 ((("%") reduce 61) (("(") reduce 61) ((")") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_PIPE) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("|") reduce 61)) 16490 ((("%") reduce 28) (("(") reduce 28) ((")") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_PIPE) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("|") reduce 28)) 16491 ((("%") reduce 36) (("(") reduce 36) ((")") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_PIPE) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("|") reduce 36)) 16492 ((("%") reduce 32) (("(") reduce 32) ((")") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_PIPE) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("|") reduce 32)) 16493 ((("%") reduce 35) (("(") reduce 35) ((")") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_PIPE) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("|") reduce 35)) 16494 ((("%") reduce 19) (("(") reduce 19) ((")") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_PIPE) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("|") reduce 19)) 16495 ((("%") reduce 45) (("(") reduce 45) ((")") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_PIPE) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("|") reduce 45)) 16496 ((("%") reduce 542) (("(") reduce 579) ((")") reduce 542) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PIPE) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 16498 ((("%") reduce 72) (("(") reduce 72) ((")") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_PIPE) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("|") reduce 72)) 16499 ((("%") reduce 83) (("(") reduce 83) ((")") reduce 83) (("*") reduce 83) (("+") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_PIPE) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("|") reduce 83)) 16500 ((("%") reduce 572) (("(") reduce 581) ((")") reduce 572) (("*") reduce 572) (("+") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) (("<") reduce 572) (("=") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AND_EQUAL) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_COALESCE_EQUAL) reduce 572) ((T_CONCAT_EQUAL) reduce 572) ((T_DEC) reduce 572) ((T_DIV_EQUAL) reduce 572) ((T_INC) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_MINUS_EQUAL) reduce 572) ((T_MOD_EQUAL) reduce 572) ((T_MUL_EQUAL) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_OR_EQUAL) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_PLUS_EQUAL) reduce 572) ((T_POW) reduce 572) ((T_POW_EQUAL) reduce 572) ((T_SL) reduce 572) ((T_SL_EQUAL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) ((T_SR_EQUAL) reduce 572) ((T_XOR_EQUAL) reduce 572) (("[") reduce 572) (("^") reduce 572) (("|") reduce 572)) 16503 ((("%") reduce 544) (("(") reduce 580) ((")") reduce 544) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PIPE) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 16505 ((("%") reduce 507) (("(") reduce 507) ((")") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("|") reduce 507)) 16506 ((("%") reduce 291) (("(") reduce 291) ((")") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 16510 ((("%") reduce 292) (("(") reduce 292) ((")") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) (("=") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AND_EQUAL) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_COALESCE_EQUAL) reduce 292) ((T_CONCAT_EQUAL) reduce 292) ((T_DEC) reduce 292) ((T_DIV_EQUAL) reduce 292) ((T_INC) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_MINUS_EQUAL) reduce 292) ((T_MOD_EQUAL) reduce 292) ((T_MUL_EQUAL) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_OR_EQUAL) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_PLUS_EQUAL) reduce 292) ((T_POW) reduce 292) ((T_POW_EQUAL) reduce 292) ((T_SL) reduce 292) ((T_SL_EQUAL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) ((T_SR_EQUAL) reduce 292) ((T_XOR_EQUAL) reduce 292) (("[") reduce 292) (("^") reduce 292) (("|") reduce 292)) 16511 ((("%") reduce 293) (("(") reduce 293) ((")") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) (("=") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AND_EQUAL) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_COALESCE_EQUAL) reduce 293) ((T_CONCAT_EQUAL) reduce 293) ((T_DEC) reduce 293) ((T_DIV_EQUAL) reduce 293) ((T_INC) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_MINUS_EQUAL) reduce 293) ((T_MOD_EQUAL) reduce 293) ((T_MUL_EQUAL) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_OR_EQUAL) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_PLUS_EQUAL) reduce 293) ((T_POW) reduce 293) ((T_POW_EQUAL) reduce 293) ((T_SL) reduce 293) ((T_SL_EQUAL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) ((T_SR_EQUAL) reduce 293) ((T_XOR_EQUAL) reduce 293) (("[") reduce 293) (("^") reduce 293) (("|") reduce 293)) 16516 ((("%") reduce 425) ((")") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_PIPE) reduce 425) ((T_POW) reduce 425) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 16518 ((("%") reduce 423) ((")") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_PIPE) reduce 423) ((T_POW) reduce 423) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 16528 ((("%") shift) ((")") reduce 419) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16529 ((("%") shift) ((")") reduce 421) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16530 ((("%") shift) ((")") reduce 420) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16531 ((("%") shift) ((")") reduce 413) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16532 ((("%") shift) ((")") reduce 410) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16533 ((("%") shift) ((")") reduce 418) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16534 ((("%") shift) ((")") reduce 412) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16535 ((("%") shift) ((")") reduce 416) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16536 ((("%") shift) ((")") reduce 411) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16537 ((("%") shift) ((")") reduce 414) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16538 ((("%") shift) ((")") reduce 415) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16539 ((("%") shift) ((")") reduce 422) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16540 ((("%") shift) ((")") reduce 417) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16542 ((("%") shift) ((")") reduce 406) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16545 ((("%") reduce 568) (("(") reduce 568) ((")") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 16548 ((("%") reduce 564) (("(") reduce 555) ((")") reduce 564) (("*") reduce 564) (("+") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) (("<") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_POW) reduce 564) ((T_SL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) (("[") reduce 564) (("^") reduce 564) (("|") reduce 564)) 16550 ((("%") reduce 563) (("(") reduce 563) ((")") reduce 563) (("*") reduce 563) (("+") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) (("<") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_POW) reduce 563) ((T_SL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) (("[") reduce 563) (("^") reduce 563) (("|") reduce 563)) 16552 ((("%") reduce 559) (("(") reduce 559) ((")") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 16553 ((("%") reduce 565) ((")") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 16554 ((("%") reduce 407) ((")") reduce 407) (("*") reduce 407) (("+") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("[") reduce 548) (("^") reduce 407) (("|") reduce 407)) 16558 ((("%") reduce 572) (("(") reduce 581) ((")") reduce 572) (("*") reduce 572) (("+") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) (("<") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_POW) reduce 572) ((T_SL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) (("[") reduce 572) (("^") reduce 572) (("|") reduce 572)) 16560 ((("%") reduce 507) (("(") reduce 507) ((")") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("|") reduce 507)) 16561 ((("%") reduce 291) (("(") reduce 291) ((")") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 16565 ((("%") reduce 292) (("(") reduce 292) ((")") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("[") reduce 292) (("^") reduce 292) (("|") reduce 292)) 16566 ((("%") reduce 293) (("(") reduce 293) ((")") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("[") reduce 293) (("^") reduce 293) (("|") reduce 293)) 16567 ((("%") reduce 504) (("(") reduce 504) ((")") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 16570 ((("%") reduce 571) (("(") reduce 581) ((")") reduce 571) (("*") reduce 571) (("+") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) (("<") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_POW) reduce 571) ((T_SL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) (("[") reduce 571) (("^") reduce 571) (("|") reduce 571)) 16571 ((("%") reduce 506) (("(") reduce 506) ((")") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("|") reduce 506)) 16572 ((("%") reduce 508) (("(") reduce 508) ((")") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("|") reduce 508)) 16577 ((("%") reduce 560) (("(") reduce 560) ((")") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 16578 ((("%") reduce 582) (("(") reduce 582) ((")") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 16579 ((("%") reduce 566) (("(") shift) ((")") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 16580 ((("%") reduce 584) (("(") reduce 584) ((")") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 16583 ((("%") reduce 583) (("(") reduce 583) ((")") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 16584 ((("%") reduce 561) (("(") reduce 561) ((")") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 16585 ((("%") reduce 567) (("(") shift) ((")") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 16586 ((("%") reduce 562) (("(") reduce 562) ((")") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("[") reduce 562) (("^") reduce 562) (("|") reduce 562)) 16587 ((("%") reduce 505) (("(") reduce 505) ((")") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 16588 ((("%") reduce 570) (("(") reduce 570) ((")") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 16591 ((("%") reduce 569) (("(") reduce 569) ((")") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 16592 ((("%") reduce 504) (("(") reduce 504) ((")") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) (("=") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AND_EQUAL) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_COALESCE_EQUAL) reduce 504) ((T_CONCAT_EQUAL) reduce 504) ((T_DEC) reduce 504) ((T_DIV_EQUAL) reduce 504) ((T_INC) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_MINUS_EQUAL) reduce 504) ((T_MOD_EQUAL) reduce 504) ((T_MUL_EQUAL) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_OR_EQUAL) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_PLUS_EQUAL) reduce 504) ((T_POW) reduce 504) ((T_POW_EQUAL) reduce 504) ((T_SL) reduce 504) ((T_SL_EQUAL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) ((T_SR_EQUAL) reduce 504) ((T_XOR_EQUAL) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 16603 ((("%") reduce 494) ((")") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PIPE) reduce 494) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 16604 ((("%") reduce 489) ((")") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_PIPE) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("^") reduce 489) (("|") reduce 489)) 16677 ((("%") shift) ((")") reduce 494) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 16993 ((("%") reduce 490) ((")") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_PIPE) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("^") reduce 490) (("|") reduce 490)) 16996 ((("%") reduce 541) (("(") reduce 579) ((")") reduce 541) (("*") reduce 541) (("+") reduce 541) (("-") reduce 541) ((".") reduce 541) (("/") reduce 541) (("<") reduce 541) ((">") reduce 541) (("?") reduce 541) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_BOOLEAN_AND) reduce 541) ((T_BOOLEAN_OR) reduce 541) ((T_COALESCE) reduce 541) ((T_INSTANCEOF) reduce 541) ((T_IS_EQUAL) reduce 541) ((T_IS_GREATER_OR_EQUAL) reduce 541) ((T_IS_IDENTICAL) reduce 541) ((T_IS_NOT_EQUAL) reduce 541) ((T_IS_NOT_IDENTICAL) reduce 541) ((T_IS_SMALLER_OR_EQUAL) reduce 541) ((T_LOGICAL_AND) reduce 541) ((T_LOGICAL_OR) reduce 541) ((T_LOGICAL_XOR) reduce 541) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 541) ((T_OBJECT_OPERATOR) reduce 541) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 541) ((T_PIPE) reduce 541) ((T_POW) reduce 541) ((T_SL) reduce 541) ((T_SPACESHIP) reduce 541) ((T_SR) reduce 541) (("[") reduce 541) (("^") reduce 541) (("|") reduce 541)) 16998 ((("%") reduce 571) (("(") reduce 581) ((")") reduce 571) (("*") reduce 571) (("+") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) (("<") reduce 571) (("=") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AND_EQUAL) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_COALESCE_EQUAL) reduce 571) ((T_CONCAT_EQUAL) reduce 571) ((T_DEC) reduce 571) ((T_DIV_EQUAL) reduce 571) ((T_INC) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_MINUS_EQUAL) reduce 571) ((T_MOD_EQUAL) reduce 571) ((T_MUL_EQUAL) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_OR_EQUAL) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_PLUS_EQUAL) reduce 571) ((T_POW) reduce 571) ((T_POW_EQUAL) reduce 571) ((T_SL) reduce 571) ((T_SL_EQUAL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) ((T_SR_EQUAL) reduce 571) ((T_XOR_EQUAL) reduce 571) (("[") reduce 571) (("^") reduce 571) (("|") reduce 571)) 17001 ((("%") reduce 543) (("(") reduce 580) ((")") reduce 543) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PIPE) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 17002 ((("%") reduce 506) (("(") reduce 506) ((")") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("|") reduce 506)) 17003 ((("%") reduce 508) (("(") reduce 508) ((")") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("|") reduce 508)) 17005 ((("%") reduce 485) ((")") reduce 485) (("*") reduce 485) (("+") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_PIPE) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("^") reduce 485) (("|") reduce 485)) 17006 ((("%") reduce 487) ((")") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_PIPE) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("^") reduce 487) (("|") reduce 487)) 17011 ((("%") reduce 560) (("(") reduce 560) ((")") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) (("=") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AND_EQUAL) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_COALESCE_EQUAL) reduce 560) ((T_CONCAT_EQUAL) reduce 560) ((T_DEC) reduce 560) ((T_DIV_EQUAL) reduce 560) ((T_INC) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_MINUS_EQUAL) reduce 560) ((T_MOD_EQUAL) reduce 560) ((T_MUL_EQUAL) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_OR_EQUAL) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_PLUS_EQUAL) reduce 560) ((T_POW) reduce 560) ((T_POW_EQUAL) reduce 560) ((T_SL) reduce 560) ((T_SL_EQUAL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) ((T_SR_EQUAL) reduce 560) ((T_XOR_EQUAL) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 17012 ((("%") reduce 582) (("(") reduce 582) ((")") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) (("=") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AND_EQUAL) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_COALESCE_EQUAL) reduce 582) ((T_CONCAT_EQUAL) reduce 582) ((T_DEC) reduce 582) ((T_DIV_EQUAL) reduce 582) ((T_INC) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_MINUS_EQUAL) reduce 582) ((T_MOD_EQUAL) reduce 582) ((T_MUL_EQUAL) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_OR_EQUAL) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_PLUS_EQUAL) reduce 582) ((T_POW) reduce 582) ((T_POW_EQUAL) reduce 582) ((T_SL) reduce 582) ((T_SL_EQUAL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) ((T_SR_EQUAL) reduce 582) ((T_XOR_EQUAL) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 17013 ((("%") reduce 566) (("(") shift) ((")") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 17014 ((("%") reduce 584) (("(") reduce 584) ((")") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) (("=") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AND_EQUAL) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_COALESCE_EQUAL) reduce 584) ((T_CONCAT_EQUAL) reduce 584) ((T_DEC) reduce 584) ((T_DIV_EQUAL) reduce 584) ((T_INC) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_MINUS_EQUAL) reduce 584) ((T_MOD_EQUAL) reduce 584) ((T_MUL_EQUAL) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_OR_EQUAL) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_PLUS_EQUAL) reduce 584) ((T_POW) reduce 584) ((T_POW_EQUAL) reduce 584) ((T_SL) reduce 584) ((T_SL_EQUAL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) ((T_SR_EQUAL) reduce 584) ((T_XOR_EQUAL) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 17017 ((("%") reduce 583) (("(") reduce 583) ((")") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) (("=") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AND_EQUAL) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_COALESCE_EQUAL) reduce 583) ((T_CONCAT_EQUAL) reduce 583) ((T_DEC) reduce 583) ((T_DIV_EQUAL) reduce 583) ((T_INC) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_MINUS_EQUAL) reduce 583) ((T_MOD_EQUAL) reduce 583) ((T_MUL_EQUAL) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_OR_EQUAL) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_PLUS_EQUAL) reduce 583) ((T_POW) reduce 583) ((T_POW_EQUAL) reduce 583) ((T_SL) reduce 583) ((T_SL_EQUAL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) ((T_SR_EQUAL) reduce 583) ((T_XOR_EQUAL) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 17018 ((("%") reduce 561) (("(") reduce 561) ((")") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) (("=") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AND_EQUAL) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_COALESCE_EQUAL) reduce 561) ((T_CONCAT_EQUAL) reduce 561) ((T_DEC) reduce 561) ((T_DIV_EQUAL) reduce 561) ((T_INC) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_MINUS_EQUAL) reduce 561) ((T_MOD_EQUAL) reduce 561) ((T_MUL_EQUAL) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_OR_EQUAL) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_PLUS_EQUAL) reduce 561) ((T_POW) reduce 561) ((T_POW_EQUAL) reduce 561) ((T_SL) reduce 561) ((T_SL_EQUAL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) ((T_SR_EQUAL) reduce 561) ((T_XOR_EQUAL) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 17019 ((("%") reduce 567) (("(") shift) ((")") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 17020 ((("%") reduce 562) (("(") reduce 562) ((")") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) (("=") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AND_EQUAL) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_COALESCE_EQUAL) reduce 562) ((T_CONCAT_EQUAL) reduce 562) ((T_DEC) reduce 562) ((T_DIV_EQUAL) reduce 562) ((T_INC) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_MINUS_EQUAL) reduce 562) ((T_MOD_EQUAL) reduce 562) ((T_MUL_EQUAL) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_OR_EQUAL) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_PLUS_EQUAL) reduce 562) ((T_POW) reduce 562) ((T_POW_EQUAL) reduce 562) ((T_SL) reduce 562) ((T_SL_EQUAL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) ((T_SR_EQUAL) reduce 562) ((T_XOR_EQUAL) reduce 562) (("[") reduce 562) (("^") reduce 562) (("|") reduce 562)) 17022 ((("%") reduce 477) ((")") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_PIPE) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 17024 ((("%") reduce 520) (("(") reduce 520) ((")") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) (("=") shift) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_PIPE) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("|") reduce 520)) 17026 ((("%") shift) ((")") reduce 405) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17027 ((("%") shift) ((")") reduce 482) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17038 ((("%") reduce 540) ((")") reduce 540) (("*") reduce 540) (("+") reduce 540) (("-") reduce 540) ((".") reduce 540) (("/") reduce 540) (("<") reduce 540) ((">") reduce 540) (("?") reduce 540) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_BOOLEAN_AND) reduce 540) ((T_BOOLEAN_OR) reduce 540) ((T_COALESCE) reduce 540) ((T_DOUBLE_ARROW) reduce 540) ((T_INSTANCEOF) reduce 540) ((T_IS_EQUAL) reduce 540) ((T_IS_GREATER_OR_EQUAL) reduce 540) ((T_IS_IDENTICAL) reduce 540) ((T_IS_NOT_EQUAL) reduce 540) ((T_IS_NOT_IDENTICAL) reduce 540) ((T_IS_SMALLER_OR_EQUAL) reduce 540) ((T_LOGICAL_AND) reduce 540) ((T_LOGICAL_OR) reduce 540) ((T_LOGICAL_XOR) reduce 540) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 540) ((T_OBJECT_OPERATOR) reduce 540) ((T_PIPE) reduce 540) ((T_POW) reduce 540) ((T_SL) reduce 540) ((T_SPACESHIP) reduce 540) ((T_SR) reduce 540) (("[") reduce 540) (("^") reduce 540) (("|") reduce 540)) 17040 ((("%") reduce 521) (("(") reduce 521) ((")") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_PIPE) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("|") reduce 521)) 17042 ((("%") reduce 534) ((")") reduce 534) (("*") reduce 534) (("+") reduce 534) (("-") reduce 534) ((".") reduce 534) (("/") reduce 534) (("<") reduce 534) ((">") reduce 534) (("?") reduce 534) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_BOOLEAN_AND) reduce 534) ((T_BOOLEAN_OR) reduce 534) ((T_COALESCE) reduce 534) ((T_DOUBLE_ARROW) reduce 534) ((T_INSTANCEOF) reduce 534) ((T_IS_EQUAL) reduce 534) ((T_IS_GREATER_OR_EQUAL) reduce 534) ((T_IS_IDENTICAL) reduce 534) ((T_IS_NOT_EQUAL) reduce 534) ((T_IS_NOT_IDENTICAL) reduce 534) ((T_IS_SMALLER_OR_EQUAL) reduce 534) ((T_LOGICAL_AND) reduce 534) ((T_LOGICAL_OR) reduce 534) ((T_LOGICAL_XOR) reduce 534) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PIPE) reduce 534) ((T_POW) reduce 534) ((T_SL) reduce 534) ((T_SPACESHIP) reduce 534) ((T_SR) reduce 534) (("[") reduce 534) (("^") reduce 534) (("|") reduce 534)) 17043 ((("%") reduce 524) ((")") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_PIPE) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("^") reduce 524) (("|") reduce 524)) 17047 ((("%") reduce 517) (("(") shift) ((")") reduce 517) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_PIPE) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("^") reduce 517) (("|") reduce 517)) 17048 ((("%") reduce 533) ((")") reduce 533) (("*") reduce 533) (("+") reduce 533) (("-") reduce 533) ((".") reduce 533) (("/") reduce 533) (("<") reduce 533) ((">") reduce 533) (("?") reduce 533) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_BOOLEAN_AND) reduce 533) ((T_BOOLEAN_OR) reduce 533) ((T_COALESCE) reduce 533) ((T_DOUBLE_ARROW) reduce 533) ((T_INSTANCEOF) reduce 533) ((T_IS_EQUAL) reduce 533) ((T_IS_GREATER_OR_EQUAL) reduce 533) ((T_IS_IDENTICAL) reduce 533) ((T_IS_NOT_EQUAL) reduce 533) ((T_IS_NOT_IDENTICAL) reduce 533) ((T_IS_SMALLER_OR_EQUAL) reduce 533) ((T_LOGICAL_AND) reduce 533) ((T_LOGICAL_OR) reduce 533) ((T_LOGICAL_XOR) reduce 533) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PIPE) reduce 533) ((T_POW) reduce 533) ((T_SL) reduce 533) ((T_SPACESHIP) reduce 533) ((T_SR) reduce 533) (("[") reduce 533) (("^") reduce 533) (("|") reduce 533)) 17049 ((("%") reduce 537) ((")") reduce 537) (("*") reduce 537) (("+") reduce 537) (("-") reduce 537) ((".") reduce 537) (("/") reduce 537) (("<") reduce 537) ((">") reduce 537) (("?") reduce 537) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_BOOLEAN_AND) reduce 537) ((T_BOOLEAN_OR) reduce 537) ((T_COALESCE) reduce 537) ((T_DOUBLE_ARROW) reduce 537) ((T_INSTANCEOF) reduce 537) ((T_IS_EQUAL) reduce 537) ((T_IS_GREATER_OR_EQUAL) reduce 537) ((T_IS_IDENTICAL) reduce 537) ((T_IS_NOT_EQUAL) reduce 537) ((T_IS_NOT_IDENTICAL) reduce 537) ((T_IS_SMALLER_OR_EQUAL) reduce 537) ((T_LOGICAL_AND) reduce 537) ((T_LOGICAL_OR) reduce 537) ((T_LOGICAL_XOR) reduce 537) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_PIPE) reduce 537) ((T_POW) reduce 537) ((T_SL) reduce 537) ((T_SPACESHIP) reduce 537) ((T_SR) reduce 537) (("[") reduce 537) (("^") reduce 537) (("|") reduce 537)) 17055 ((("%") reduce 532) ((")") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PIPE) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("|") reduce 532)) 17057 ((("%") reduce 523) ((")") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_PIPE) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("^") reduce 523) (("|") reduce 523)) 17059 ((("%") reduce 536) ((")") reduce 536) (("*") reduce 536) (("+") reduce 536) (("-") reduce 536) ((".") reduce 536) (("/") reduce 536) (("<") reduce 536) ((">") reduce 536) (("?") reduce 536) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_BOOLEAN_AND) reduce 536) ((T_BOOLEAN_OR) reduce 536) ((T_COALESCE) reduce 536) ((T_DOUBLE_ARROW) reduce 536) ((T_INSTANCEOF) reduce 536) ((T_IS_EQUAL) reduce 536) ((T_IS_GREATER_OR_EQUAL) reduce 536) ((T_IS_IDENTICAL) reduce 536) ((T_IS_NOT_EQUAL) reduce 536) ((T_IS_NOT_IDENTICAL) reduce 536) ((T_IS_SMALLER_OR_EQUAL) reduce 536) ((T_LOGICAL_AND) reduce 536) ((T_LOGICAL_OR) reduce 536) ((T_LOGICAL_XOR) reduce 536) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 536) ((T_OBJECT_OPERATOR) reduce 536) ((T_PIPE) reduce 536) ((T_POW) reduce 536) ((T_SL) reduce 536) ((T_SPACESHIP) reduce 536) ((T_SR) reduce 536) (("[") reduce 536) (("^") reduce 536) (("|") reduce 536)) 17060 ((("%") reduce 94) (("(") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("|") reduce 94)) 17061 ((("%") reduce 93) (("(") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("|") reduce 93)) 17062 ((("%") reduce 95) (("(") reduce 95) ((")") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("|") reduce 95)) 17064 ((("%") reduce 539) ((")") reduce 539) (("*") reduce 539) (("+") reduce 539) (("-") reduce 539) ((".") reduce 539) (("/") reduce 539) (("<") reduce 539) ((">") reduce 539) (("?") reduce 539) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_BOOLEAN_AND) reduce 539) ((T_BOOLEAN_OR) reduce 539) ((T_COALESCE) reduce 539) ((T_DOUBLE_ARROW) reduce 539) ((T_INSTANCEOF) reduce 539) ((T_IS_EQUAL) reduce 539) ((T_IS_GREATER_OR_EQUAL) reduce 539) ((T_IS_IDENTICAL) reduce 539) ((T_IS_NOT_EQUAL) reduce 539) ((T_IS_NOT_IDENTICAL) reduce 539) ((T_IS_SMALLER_OR_EQUAL) reduce 539) ((T_LOGICAL_AND) reduce 539) ((T_LOGICAL_OR) reduce 539) ((T_LOGICAL_XOR) reduce 539) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 539) ((T_OBJECT_OPERATOR) reduce 539) ((T_PIPE) reduce 539) ((T_POW) reduce 539) ((T_SL) reduce 539) ((T_SPACESHIP) reduce 539) ((T_SR) reduce 539) (("[") reduce 539) (("^") reduce 539) (("|") reduce 539)) 17067 ((("%") reduce 538) ((")") reduce 538) (("*") reduce 538) (("+") reduce 538) (("-") reduce 538) ((".") reduce 538) (("/") reduce 538) (("<") reduce 538) ((">") reduce 538) (("?") reduce 538) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_BOOLEAN_AND) reduce 538) ((T_BOOLEAN_OR) reduce 538) ((T_COALESCE) reduce 538) ((T_DOUBLE_ARROW) reduce 538) ((T_INSTANCEOF) reduce 538) ((T_IS_EQUAL) reduce 538) ((T_IS_GREATER_OR_EQUAL) reduce 538) ((T_IS_IDENTICAL) reduce 538) ((T_IS_NOT_EQUAL) reduce 538) ((T_IS_NOT_IDENTICAL) reduce 538) ((T_IS_SMALLER_OR_EQUAL) reduce 538) ((T_LOGICAL_AND) reduce 538) ((T_LOGICAL_OR) reduce 538) ((T_LOGICAL_XOR) reduce 538) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 538) ((T_OBJECT_OPERATOR) reduce 538) ((T_PIPE) reduce 538) ((T_POW) reduce 538) ((T_SL) reduce 538) ((T_SPACESHIP) reduce 538) ((T_SR) reduce 538) (("[") reduce 538) (("^") reduce 538) (("|") reduce 538)) 17073 ((("%") reduce 92) (("(") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92)) 17076 ((("%") reduce 535) ((")") reduce 535) (("*") reduce 535) (("+") reduce 535) (("-") reduce 535) ((".") reduce 535) (("/") reduce 535) (("<") reduce 535) ((">") reduce 535) (("?") reduce 535) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_BOOLEAN_AND) reduce 535) ((T_BOOLEAN_OR) reduce 535) ((T_COALESCE) reduce 535) ((T_DOUBLE_ARROW) reduce 535) ((T_INSTANCEOF) reduce 535) ((T_IS_EQUAL) reduce 535) ((T_IS_GREATER_OR_EQUAL) reduce 535) ((T_IS_IDENTICAL) reduce 535) ((T_IS_NOT_EQUAL) reduce 535) ((T_IS_NOT_IDENTICAL) reduce 535) ((T_IS_SMALLER_OR_EQUAL) reduce 535) ((T_LOGICAL_AND) reduce 535) ((T_LOGICAL_OR) reduce 535) ((T_LOGICAL_XOR) reduce 535) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PIPE) reduce 535) ((T_POW) reduce 535) ((T_SL) reduce 535) ((T_SPACESHIP) reduce 535) ((T_SR) reduce 535) (("[") reduce 535) (("^") reduce 535) (("|") reduce 535)) 17078 ((("%") reduce 568) (("(") reduce 568) ((")") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) (("=") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AND_EQUAL) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_COALESCE_EQUAL) reduce 568) ((T_CONCAT_EQUAL) reduce 568) ((T_DEC) reduce 568) ((T_DIV_EQUAL) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INC) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_MINUS_EQUAL) reduce 568) ((T_MOD_EQUAL) reduce 568) ((T_MUL_EQUAL) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_OR_EQUAL) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_PLUS_EQUAL) reduce 568) ((T_POW) reduce 568) ((T_POW_EQUAL) reduce 568) ((T_SL) reduce 568) ((T_SL_EQUAL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) ((T_SR_EQUAL) reduce 568) ((T_XOR_EQUAL) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 17079 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 479) (("(") shift) ((")") reduce 479) (("*") reduce 479) (("+") shift) (("-") shift) ((".") reduce 479) (("/") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 479) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 479) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 479) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PIPE) reduce 479) ((T_POW) reduce 479) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 479) (("`") shift) (("|") reduce 479) (("~") shift)) 17086 ((("%") reduce 564) (("(") reduce 555) ((")") reduce 564) (("*") reduce 564) (("+") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) (("<") reduce 564) (("=") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AND_EQUAL) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_COALESCE_EQUAL) reduce 564) ((T_CONCAT_EQUAL) reduce 564) ((T_DEC) reduce 564) ((T_DIV_EQUAL) reduce 564) ((T_DOUBLE_ARROW) reduce 564) ((T_INC) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_MINUS_EQUAL) reduce 564) ((T_MOD_EQUAL) reduce 564) ((T_MUL_EQUAL) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_OR_EQUAL) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_PLUS_EQUAL) reduce 564) ((T_POW) reduce 564) ((T_POW_EQUAL) reduce 564) ((T_SL) reduce 564) ((T_SL_EQUAL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) ((T_SR_EQUAL) reduce 564) ((T_XOR_EQUAL) reduce 564) (("[") reduce 564) (("^") reduce 564) (("|") reduce 564)) 17087 ((("%") reduce 530) ((")") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PIPE) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 551) (("^") reduce 530) (("|") reduce 530)) 17089 ((("%") reduce 529) ((")") reduce 529) (("*") reduce 529) (("+") reduce 529) (("-") reduce 529) ((".") reduce 529) (("/") reduce 529) (("<") reduce 529) ((">") reduce 529) (("?") reduce 529) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_BOOLEAN_AND) reduce 529) ((T_BOOLEAN_OR) reduce 529) ((T_COALESCE) reduce 529) ((T_DOUBLE_ARROW) reduce 529) ((T_INSTANCEOF) reduce 529) ((T_IS_EQUAL) reduce 529) ((T_IS_GREATER_OR_EQUAL) reduce 529) ((T_IS_IDENTICAL) reduce 529) ((T_IS_NOT_EQUAL) reduce 529) ((T_IS_NOT_IDENTICAL) reduce 529) ((T_IS_SMALLER_OR_EQUAL) reduce 529) ((T_LOGICAL_AND) reduce 529) ((T_LOGICAL_OR) reduce 529) ((T_LOGICAL_XOR) reduce 529) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PIPE) reduce 529) ((T_POW) reduce 529) ((T_SL) reduce 529) ((T_SPACESHIP) reduce 529) ((T_SR) reduce 529) (("[") reduce 554) (("^") reduce 529) (("|") reduce 529)) 17090 ((("%") reduce 528) (("(") reduce 557) ((")") reduce 528) (("*") reduce 528) (("+") reduce 528) (("-") reduce 528) ((".") reduce 528) (("/") reduce 528) (("<") reduce 528) ((">") reduce 528) (("?") reduce 528) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_BOOLEAN_AND) reduce 528) ((T_BOOLEAN_OR) reduce 528) ((T_COALESCE) reduce 528) ((T_DOUBLE_ARROW) reduce 528) ((T_INSTANCEOF) reduce 528) ((T_IS_EQUAL) reduce 528) ((T_IS_GREATER_OR_EQUAL) reduce 528) ((T_IS_IDENTICAL) reduce 528) ((T_IS_NOT_EQUAL) reduce 528) ((T_IS_NOT_IDENTICAL) reduce 528) ((T_IS_SMALLER_OR_EQUAL) reduce 528) ((T_LOGICAL_AND) reduce 528) ((T_LOGICAL_OR) reduce 528) ((T_LOGICAL_XOR) reduce 528) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PIPE) reduce 528) ((T_POW) reduce 528) ((T_SL) reduce 528) ((T_SPACESHIP) reduce 528) ((T_SR) reduce 528) (("[") reduce 550) (("^") reduce 528) (("|") reduce 528)) 17091 ((("%") shift) ((")") reduce 480) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 480) ((T_LOGICAL_OR) reduce 480) ((T_LOGICAL_XOR) reduce 480) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17094 ((("%") reduce 563) (("(") reduce 563) ((")") reduce 563) (("*") reduce 563) (("+") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) (("<") reduce 563) (("=") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AND_EQUAL) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_COALESCE_EQUAL) reduce 563) ((T_CONCAT_EQUAL) reduce 563) ((T_DEC) reduce 563) ((T_DIV_EQUAL) reduce 563) ((T_DOUBLE_ARROW) reduce 563) ((T_INC) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_MINUS_EQUAL) reduce 563) ((T_MOD_EQUAL) reduce 563) ((T_MUL_EQUAL) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_OR_EQUAL) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_PLUS_EQUAL) reduce 563) ((T_POW) reduce 563) ((T_POW_EQUAL) reduce 563) ((T_SL) reduce 563) ((T_SL_EQUAL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) ((T_SR_EQUAL) reduce 563) ((T_XOR_EQUAL) reduce 563) (("[") reduce 563) (("^") reduce 563) (("|") reduce 563)) 17095 ((("%") reduce 484) ((")") reduce 484) (("*") reduce 484) (("+") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_DOUBLE_ARROW) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_PIPE) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("^") reduce 484) (("|") reduce 484)) 17096 ((("%") reduce 466) ((")") reduce 466) (("*") reduce 466) (("+") reduce 466) (("-") reduce 466) ((".") reduce 466) (("/") reduce 466) (("<") reduce 466) ((">") reduce 466) (("?") reduce 466) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_BOOLEAN_AND) reduce 466) ((T_BOOLEAN_OR) reduce 466) ((T_COALESCE) reduce 466) ((T_DOUBLE_ARROW) reduce 466) ((T_INSTANCEOF) reduce 466) ((T_IS_EQUAL) reduce 466) ((T_IS_GREATER_OR_EQUAL) reduce 466) ((T_IS_IDENTICAL) reduce 466) ((T_IS_NOT_EQUAL) reduce 466) ((T_IS_NOT_IDENTICAL) reduce 466) ((T_IS_SMALLER_OR_EQUAL) reduce 466) ((T_LOGICAL_AND) reduce 466) ((T_LOGICAL_OR) reduce 466) ((T_LOGICAL_XOR) reduce 466) ((T_PIPE) reduce 466) ((T_POW) reduce 466) ((T_SL) reduce 466) ((T_SPACESHIP) reduce 466) ((T_SR) reduce 466) (("^") reduce 466) (("|") reduce 466)) 17097 ((("%") reduce 488) ((")") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_PIPE) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("^") reduce 488) (("|") reduce 488)) 17098 ((("%") reduce 531) (("(") shift) ((")") reduce 531) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("|") reduce 531)) 17099 ((("%") reduce 461) (("(") reduce 558) ((")") reduce 461) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PIPE) reduce 461) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") reduce 552) (("^") reduce 461) (("|") reduce 461)) 17100 ((("%") reduce 462) ((")") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_DOUBLE_ARROW) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_PIPE) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 17101 ((("%") reduce 476) ((")") reduce 476) (("*") reduce 476) (("+") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_DOUBLE_ARROW) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_PIPE) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("|") reduce 476)) 17102 ((("%") reduce 559) (("(") reduce 559) ((")") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) (("=") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AND_EQUAL) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_COALESCE_EQUAL) reduce 559) ((T_CONCAT_EQUAL) reduce 559) ((T_DEC) reduce 559) ((T_DIV_EQUAL) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INC) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_MINUS_EQUAL) reduce 559) ((T_MOD_EQUAL) reduce 559) ((T_MUL_EQUAL) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_OR_EQUAL) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_PLUS_EQUAL) reduce 559) ((T_POW) reduce 559) ((T_POW_EQUAL) reduce 559) ((T_SL) reduce 559) ((T_SL_EQUAL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) ((T_SR_EQUAL) reduce 559) ((T_XOR_EQUAL) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 17103 ((("%") reduce 565) ((")") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 17104 ((("%") reduce 403) ((")") reduce 403) (("*") reduce 403) (("+") reduce 403) (("-") reduce 403) ((".") reduce 403) (("/") reduce 403) (("<") reduce 403) (("=") shift) ((">") reduce 403) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_DOUBLE_ARROW) reduce 403) ((T_INC) shift) ((T_INSTANCEOF) reduce 403) ((T_IS_EQUAL) reduce 403) ((T_IS_GREATER_OR_EQUAL) reduce 403) ((T_IS_IDENTICAL) reduce 403) ((T_IS_NOT_EQUAL) reduce 403) ((T_IS_NOT_IDENTICAL) reduce 403) ((T_IS_SMALLER_OR_EQUAL) reduce 403) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 403) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 403) ((T_POW_EQUAL) shift) ((T_SL) reduce 403) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 403) ((T_SR) reduce 403) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 548) (("^") reduce 403) (("|") reduce 403)) 17107 ((("%") reduce 448) ((")") reduce 448) (("*") reduce 448) (("+") reduce 448) (("-") reduce 448) ((".") reduce 448) (("/") reduce 448) (("<") reduce 448) ((">") reduce 448) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_BOOLEAN_AND) reduce 448) ((T_BOOLEAN_OR) reduce 448) ((T_COALESCE) reduce 448) ((T_DOUBLE_ARROW) reduce 448) ((T_INSTANCEOF) reduce 448) ((T_IS_EQUAL) reduce 448) ((T_IS_GREATER_OR_EQUAL) reduce 448) ((T_IS_IDENTICAL) reduce 448) ((T_IS_NOT_EQUAL) reduce 448) ((T_IS_NOT_IDENTICAL) reduce 448) ((T_IS_SMALLER_OR_EQUAL) reduce 448) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_PIPE) reduce 448) ((T_POW) shift) ((T_SL) reduce 448) ((T_SPACESHIP) reduce 448) ((T_SR) reduce 448) (("^") reduce 448) (("|") reduce 448)) 17139 ((("%") shift) ((")") reduce 432) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_DOUBLE_ARROW) reduce 432) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 432)) 17140 ((("%") shift) ((")") reduce 435) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_DOUBLE_ARROW) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("|") reduce 435)) 17141 ((("%") shift) ((")") reduce 444) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 444) (("/") shift) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_PIPE) reduce 444) ((T_POW) shift) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("|") reduce 444)) 17142 ((("%") shift) ((")") reduce 458) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 458) (("|") reduce 458)) 17143 ((("%") shift) ((")") reduce 443) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 443) (("/") shift) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_DOUBLE_ARROW) reduce 443) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_PIPE) reduce 443) ((T_POW) shift) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("^") reduce 443) (("|") reduce 443)) 17144 ((("%") reduce 440) ((")") reduce 440) (("*") reduce 440) (("+") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_DOUBLE_ARROW) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_PIPE) reduce 440) ((T_POW) shift) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("|") reduce 440)) 17145 ((("%") shift) ((")") reduce 453) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_DOUBLE_ARROW) reduce 453) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_PIPE) reduce 453) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 453) ((T_SR) shift) (("^") reduce 453) (("|") reduce 453)) 17146 ((("%") shift) ((")") reduce 431) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17147 ((("%") shift) ((")") reduce 429) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 429) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17148 ((("%") shift) ((")") reduce 430) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 430) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17149 ((("%") shift) ((")") reduce 455) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 455) ((T_SR) shift) (("^") reduce 455) (("|") reduce 455)) 17150 ((("%") shift) ((")") reduce 450) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_DOUBLE_ARROW) reduce 450) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 450) (("|") reduce 450)) 17151 ((("%") shift) ((")") reduce 452) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_DOUBLE_ARROW) reduce 452) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 452) (("|") reduce 452)) 17152 ((("%") shift) ((")") reduce 449) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_DOUBLE_ARROW) reduce 449) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 449) (("|") reduce 449)) 17153 ((("%") shift) ((")") reduce 457) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 457) ((T_SR) shift) (("^") reduce 457) (("|") reduce 457)) 17154 ((("%") shift) ((")") reduce 451) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_DOUBLE_ARROW) reduce 451) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 451) (("|") reduce 451)) 17157 ((("%") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 17158 ((("%") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 17159 ((("%") reduce 95) ((")") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 17160 ((("%") reduce 509) ((")") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 17161 ((("%") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 17162 ((("%") reduce 568) ((")") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 17163 ((("%") reduce 511) ((")") reduce 511) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("|") reduce 511)) 17164 ((("%") reduce 459) ((")") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PIPE) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 17165 ((("%") reduce 510) ((")") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 17166 ((("%") reduce 512) ((")") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("|") reduce 512)) 17167 ((("%") reduce 573) ((")") reduce 573) (("*") reduce 573) (("+") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_DOUBLE_ARROW) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("|") reduce 573)) 17173 ((("%") reduce 574) ((")") reduce 574) (("*") reduce 574) (("+") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_DOUBLE_ARROW) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("|") reduce 574)) 17174 ((("%") reduce 578) ((")") reduce 578) (("*") reduce 578) (("+") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_DOUBLE_ARROW) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("|") reduce 578)) 17175 ((("%") reduce 582) ((")") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 17176 ((("%") reduce 575) ((")") reduce 575) (("*") reduce 575) (("+") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_DOUBLE_ARROW) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("|") reduce 575)) 17177 ((("%") reduce 584) ((")") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 17180 ((("%") reduce 583) ((")") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 17181 ((("%") reduce 576) ((")") reduce 576) (("*") reduce 576) (("+") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_DOUBLE_ARROW) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("|") reduce 576)) 17183 ((("%") reduce 577) ((")") reduce 577) (("*") reduce 577) (("+") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_DOUBLE_ARROW) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("|") reduce 577)) 17185 ((("%") reduce 513) ((")") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("|") reduce 513)) 17186 ((("%") reduce 570) ((")") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 17189 ((("%") reduce 569) ((")") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 17190 ((("%") shift) ((")") reduce 465) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 465) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 465) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17191 ((("%") shift) ((")") reduce 427) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_DOUBLE_ARROW) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17192 ((("%") shift) ((")") reduce 428) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_DOUBLE_ARROW) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17193 ((("%") shift) ((")") reduce 433) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_DOUBLE_ARROW) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("|") reduce 433)) 17194 ((("%") shift) ((")") reduce 434) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_DOUBLE_ARROW) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("|") reduce 434)) 17198 ((("%") shift) ((")") reduce 463) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 463) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17199 ((("%") shift) ((")") reduce 464) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 464) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17200 ((("%") shift) ((")") reduce 456) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 456) ((T_SR) shift) (("^") reduce 456) (("|") reduce 456)) 17201 ((("%") shift) ((")") reduce 454) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 454) ((T_SR) shift) (("^") reduce 454) (("|") reduce 454)) 17202 ((("%") reduce 441) ((")") reduce 441) (("*") reduce 441) (("+") reduce 441) (("-") reduce 441) ((".") reduce 441) (("/") reduce 441) (("<") reduce 441) ((">") reduce 441) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_DOUBLE_ARROW) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 441) ((T_IS_GREATER_OR_EQUAL) reduce 441) ((T_IS_IDENTICAL) reduce 441) ((T_IS_NOT_EQUAL) reduce 441) ((T_IS_NOT_IDENTICAL) reduce 441) ((T_IS_SMALLER_OR_EQUAL) reduce 441) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_PIPE) reduce 441) ((T_POW) shift) ((T_SL) reduce 441) ((T_SPACESHIP) reduce 441) ((T_SR) reduce 441) (("^") reduce 441) (("|") reduce 441)) 17203 ((("%") shift) ((")") reduce 436) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 436) (("/") shift) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_DOUBLE_ARROW) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_PIPE) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 436) ((T_SR) shift) (("^") reduce 436) (("|") reduce 436)) 17204 ((("%") shift) ((")") reduce 438) (("*") shift) (("+") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") shift) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_DOUBLE_ARROW) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_PIPE) reduce 438) ((T_POW) shift) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("|") reduce 438)) 17205 ((("%") shift) ((")") reduce 437) (("*") shift) (("+") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") shift) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_DOUBLE_ARROW) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_PIPE) reduce 437) ((T_POW) shift) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("|") reduce 437)) 17206 ((("%") reduce 439) ((")") reduce 439) (("*") reduce 439) (("+") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_DOUBLE_ARROW) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_PIPE) reduce 439) ((T_POW) shift) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("|") reduce 439)) 17207 ((("%") reduce 442) ((")") reduce 442) (("*") reduce 442) (("+") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_DOUBLE_ARROW) reduce 442) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_PIPE) reduce 442) ((T_POW) shift) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("|") reduce 442)) 17209 ((("%") reduce 74) (("(") reduce 74) ((")") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_DOUBLE_ARROW) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_PIPE) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("|") reduce 74)) 17210 ((("%") reduce 52) (("(") reduce 52) ((")") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_DOUBLE_ARROW) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_PIPE) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("|") reduce 52)) 17211 ((("%") reduce 27) (("(") reduce 27) ((")") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_DOUBLE_ARROW) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_PIPE) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("|") reduce 27)) 17212 ((("%") reduce 51) (("(") reduce 51) ((")") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_DOUBLE_ARROW) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_PIPE) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("|") reduce 51)) 17213 ((("%") reduce 53) (("(") reduce 53) ((")") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_DOUBLE_ARROW) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_PIPE) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("|") reduce 53)) 17214 ((("%") reduce 49) (("(") reduce 49) ((")") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_DOUBLE_ARROW) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_PIPE) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("|") reduce 49)) 17215 ((("%") reduce 29) (("(") reduce 29) ((")") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_DOUBLE_ARROW) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_PIPE) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("|") reduce 29)) 17216 ((("%") reduce 59) (("(") reduce 59) ((")") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_DOUBLE_ARROW) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_PIPE) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("|") reduce 59)) 17217 ((("%") reduce 60) (("(") reduce 60) ((")") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_DOUBLE_ARROW) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_PIPE) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("|") reduce 60)) 17218 ((("%") reduce 11) (("(") reduce 11) ((")") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_DOUBLE_ARROW) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_PIPE) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("|") reduce 11)) 17219 ((("%") reduce 42) (("(") reduce 42) ((")") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_DOUBLE_ARROW) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_PIPE) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("|") reduce 42)) 17220 ((("%") reduce 39) (("(") reduce 39) ((")") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_DOUBLE_ARROW) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_PIPE) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("|") reduce 39)) 17221 ((("%") reduce 25) (("(") reduce 25) ((")") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_DOUBLE_ARROW) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_PIPE) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("|") reduce 25)) 17222 ((("%") reduce 50) (("(") reduce 50) ((")") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_DOUBLE_ARROW) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_PIPE) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("|") reduce 50)) 17223 ((("%") reduce 66) (("(") reduce 66) ((")") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_DOUBLE_ARROW) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_PIPE) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("|") reduce 66)) 17224 ((("%") reduce 18) (("(") reduce 18) ((")") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_DOUBLE_ARROW) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_PIPE) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("|") reduce 18)) 17225 ((("%") reduce 17) (("(") reduce 17) ((")") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_DOUBLE_ARROW) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_PIPE) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("|") reduce 17)) 17226 ((("%") reduce 15) (("(") reduce 15) ((")") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_DOUBLE_ARROW) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_PIPE) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("|") reduce 15)) 17227 ((("%") reduce 14) (("(") reduce 14) ((")") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_DOUBLE_ARROW) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_PIPE) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("|") reduce 14)) 17228 ((("%") reduce 38) (("(") reduce 38) ((")") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_DOUBLE_ARROW) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_PIPE) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("|") reduce 38)) 17229 ((("%") reduce 26) (("(") reduce 26) ((")") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_DOUBLE_ARROW) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_PIPE) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("|") reduce 26)) 17230 ((("%") reduce 22) (("(") reduce 22) ((")") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_DOUBLE_ARROW) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_PIPE) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("|") reduce 22)) 17231 ((("%") reduce 24) (("(") reduce 24) ((")") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_DOUBLE_ARROW) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_PIPE) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("|") reduce 24)) 17232 ((("%") reduce 16) (("(") reduce 16) ((")") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_DOUBLE_ARROW) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_PIPE) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("|") reduce 16)) 17233 ((("%") reduce 48) (("(") reduce 48) ((")") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_DOUBLE_ARROW) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_PIPE) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("|") reduce 48)) 17234 ((("%") reduce 20) (("(") reduce 20) ((")") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_DOUBLE_ARROW) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_PIPE) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("|") reduce 20)) 17235 ((("%") reduce 70) (("(") reduce 70) ((")") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_DOUBLE_ARROW) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_PIPE) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("|") reduce 70)) 17236 ((("%") reduce 3) (("(") reduce 3) ((")") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_DOUBLE_ARROW) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_PIPE) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("|") reduce 3)) 17237 ((("%") reduce 12) (("(") reduce 12) ((")") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_DOUBLE_ARROW) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_PIPE) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("|") reduce 12)) 17238 ((("%") reduce 54) (("(") reduce 54) ((")") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_DOUBLE_ARROW) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_PIPE) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("|") reduce 54)) 17239 ((("%") reduce 65) (("(") reduce 65) ((")") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_DOUBLE_ARROW) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_PIPE) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("|") reduce 65)) 17240 ((("%") reduce 75) (("(") reduce 75) ((")") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_DOUBLE_ARROW) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_PIPE) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("|") reduce 75)) 17241 ((("%") reduce 30) (("(") reduce 30) ((")") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_DOUBLE_ARROW) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_PIPE) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("|") reduce 30)) 17242 ((("%") reduce 68) (("(") reduce 68) ((")") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_DOUBLE_ARROW) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_PIPE) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("|") reduce 68)) 17243 ((("%") reduce 21) (("(") reduce 21) ((")") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_DOUBLE_ARROW) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_PIPE) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("|") reduce 21)) 17244 ((("%") reduce 23) (("(") reduce 23) ((")") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_DOUBLE_ARROW) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_PIPE) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("|") reduce 23)) 17245 ((("%") reduce 41) (("(") reduce 41) ((")") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_DOUBLE_ARROW) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_PIPE) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("|") reduce 41)) 17246 ((("%") reduce 62) (("(") reduce 62) ((")") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_DOUBLE_ARROW) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_PIPE) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("|") reduce 62)) 17247 ((("%") reduce 34) (("(") reduce 34) ((")") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_DOUBLE_ARROW) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_PIPE) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("|") reduce 34)) 17248 ((("%") reduce 40) (("(") reduce 40) ((")") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_DOUBLE_ARROW) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_PIPE) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("|") reduce 40)) 17249 ((("%") reduce 13) (("(") reduce 13) ((")") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_DOUBLE_ARROW) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_PIPE) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("|") reduce 13)) 17250 ((("%") reduce 55) (("(") reduce 55) ((")") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_DOUBLE_ARROW) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_PIPE) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("|") reduce 55)) 17251 ((("%") reduce 1) (("(") reduce 1) ((")") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_DOUBLE_ARROW) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_PIPE) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("|") reduce 1)) 17252 ((("%") reduce 2) (("(") reduce 2) ((")") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_DOUBLE_ARROW) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_PIPE) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("|") reduce 2)) 17253 ((("%") reduce 9) (("(") reduce 9) ((")") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_DOUBLE_ARROW) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_PIPE) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("|") reduce 9)) 17254 ((("%") reduce 33) (("(") reduce 33) ((")") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_DOUBLE_ARROW) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_PIPE) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("|") reduce 33)) 17255 ((("%") reduce 58) (("(") reduce 58) ((")") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_DOUBLE_ARROW) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_PIPE) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("|") reduce 58)) 17256 ((("%") reduce 37) (("(") reduce 37) ((")") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_DOUBLE_ARROW) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_PIPE) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("|") reduce 37)) 17257 ((("%") reduce 64) (("(") reduce 64) ((")") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_DOUBLE_ARROW) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_PIPE) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("|") reduce 64)) 17258 ((("%") reduce 46) (("(") reduce 46) ((")") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_DOUBLE_ARROW) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_PIPE) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("|") reduce 46)) 17259 ((("%") reduce 8) (("(") reduce 8) ((")") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_DOUBLE_ARROW) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_PIPE) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("|") reduce 8)) 17260 ((("%") reduce 6) (("(") reduce 6) ((")") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_DOUBLE_ARROW) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_PIPE) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("|") reduce 6)) 17261 ((("%") reduce 7) (("(") reduce 7) ((")") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_DOUBLE_ARROW) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_PIPE) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("|") reduce 7)) 17262 ((("%") reduce 69) (("(") reduce 69) ((")") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_DOUBLE_ARROW) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_PIPE) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("|") reduce 69)) 17263 ((("%") reduce 63) (("(") reduce 63) ((")") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_DOUBLE_ARROW) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_PIPE) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("|") reduce 63)) 17264 ((("%") reduce 56) (("(") reduce 56) ((")") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_DOUBLE_ARROW) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_PIPE) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("|") reduce 56)) 17265 ((("%") reduce 10) (("(") reduce 10) ((")") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_DOUBLE_ARROW) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_PIPE) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("|") reduce 10)) 17266 ((("%") reduce 67) (("(") reduce 67) ((")") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_DOUBLE_ARROW) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_PIPE) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("|") reduce 67)) 17267 ((("%") reduce 44) (("(") reduce 44) ((")") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_DOUBLE_ARROW) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_PIPE) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("|") reduce 44)) 17268 ((("%") reduce 76) (("(") reduce 76) ((")") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_DOUBLE_ARROW) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_PIPE) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("|") reduce 76)) 17269 ((("%") reduce 71) (("(") reduce 71) ((")") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_DOUBLE_ARROW) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_PIPE) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("|") reduce 71)) 17270 ((("%") reduce 77) (("(") reduce 77) ((")") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_DOUBLE_ARROW) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_PIPE) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("|") reduce 77)) 17271 ((("%") reduce 78) (("(") reduce 78) ((")") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_DOUBLE_ARROW) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_PIPE) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("|") reduce 78)) 17272 ((("%") reduce 79) (("(") reduce 79) ((")") reduce 79) (("*") reduce 79) (("+") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_DOUBLE_ARROW) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_PIPE) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("|") reduce 79)) 17273 ((("%") reduce 4) (("(") reduce 4) ((")") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_DOUBLE_ARROW) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_PIPE) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("|") reduce 4)) 17274 ((("%") reduce 5) (("(") reduce 5) ((")") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_DOUBLE_ARROW) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_PIPE) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("|") reduce 5)) 17275 ((("%") reduce 43) (("(") reduce 43) ((")") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_DOUBLE_ARROW) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_PIPE) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("|") reduce 43)) 17276 ((("%") reduce 73) (("(") reduce 73) ((")") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_DOUBLE_ARROW) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_PIPE) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("|") reduce 73)) 17277 ((("%") reduce 82) (("(") reduce 82) ((")") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_DOUBLE_ARROW) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_PIPE) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("|") reduce 82)) 17278 ((("%") reduce 47) (("(") reduce 47) ((")") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_DOUBLE_ARROW) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_PIPE) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("|") reduce 47)) 17279 ((("%") reduce 31) (("(") reduce 31) ((")") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_DOUBLE_ARROW) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_PIPE) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("|") reduce 31)) 17280 ((("%") reduce 57) (("(") reduce 57) ((")") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_DOUBLE_ARROW) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_PIPE) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("|") reduce 57)) 17281 ((("%") reduce 61) (("(") reduce 61) ((")") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_DOUBLE_ARROW) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_PIPE) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("|") reduce 61)) 17282 ((("%") reduce 28) (("(") reduce 28) ((")") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_DOUBLE_ARROW) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_PIPE) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("|") reduce 28)) 17283 ((("%") reduce 36) (("(") reduce 36) ((")") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_DOUBLE_ARROW) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_PIPE) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("|") reduce 36)) 17284 ((("%") reduce 32) (("(") reduce 32) ((")") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_DOUBLE_ARROW) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_PIPE) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("|") reduce 32)) 17285 ((("%") reduce 35) (("(") reduce 35) ((")") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_DOUBLE_ARROW) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_PIPE) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("|") reduce 35)) 17286 ((("%") reduce 19) (("(") reduce 19) ((")") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_DOUBLE_ARROW) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_PIPE) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("|") reduce 19)) 17287 ((("%") reduce 45) (("(") reduce 45) ((")") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_DOUBLE_ARROW) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_PIPE) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("|") reduce 45)) 17288 ((("%") reduce 542) (("(") reduce 579) ((")") reduce 542) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PIPE) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 17290 ((("%") reduce 72) (("(") reduce 72) ((")") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_DOUBLE_ARROW) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_PIPE) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("|") reduce 72)) 17291 ((("%") reduce 83) (("(") reduce 83) ((")") reduce 83) (("*") reduce 83) (("+") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_DOUBLE_ARROW) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_PIPE) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("|") reduce 83)) 17292 ((("%") reduce 572) (("(") reduce 581) ((")") reduce 572) (("*") reduce 572) (("+") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) (("<") reduce 572) (("=") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AND_EQUAL) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_COALESCE_EQUAL) reduce 572) ((T_CONCAT_EQUAL) reduce 572) ((T_DEC) reduce 572) ((T_DIV_EQUAL) reduce 572) ((T_DOUBLE_ARROW) reduce 572) ((T_INC) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_MINUS_EQUAL) reduce 572) ((T_MOD_EQUAL) reduce 572) ((T_MUL_EQUAL) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_OR_EQUAL) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_PLUS_EQUAL) reduce 572) ((T_POW) reduce 572) ((T_POW_EQUAL) reduce 572) ((T_SL) reduce 572) ((T_SL_EQUAL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) ((T_SR_EQUAL) reduce 572) ((T_XOR_EQUAL) reduce 572) (("[") reduce 572) (("^") reduce 572) (("|") reduce 572)) 17295 ((("%") reduce 544) (("(") reduce 580) ((")") reduce 544) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PIPE) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 17297 ((("%") reduce 507) (("(") reduce 507) ((")") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("|") reduce 507)) 17298 ((("%") reduce 291) (("(") reduce 291) ((")") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 17302 ((("%") reduce 292) (("(") reduce 292) ((")") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) (("=") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AND_EQUAL) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_COALESCE_EQUAL) reduce 292) ((T_CONCAT_EQUAL) reduce 292) ((T_DEC) reduce 292) ((T_DIV_EQUAL) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INC) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_MINUS_EQUAL) reduce 292) ((T_MOD_EQUAL) reduce 292) ((T_MUL_EQUAL) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_OR_EQUAL) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_PLUS_EQUAL) reduce 292) ((T_POW) reduce 292) ((T_POW_EQUAL) reduce 292) ((T_SL) reduce 292) ((T_SL_EQUAL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) ((T_SR_EQUAL) reduce 292) ((T_XOR_EQUAL) reduce 292) (("[") reduce 292) (("^") reduce 292) (("|") reduce 292)) 17303 ((("%") reduce 293) (("(") reduce 293) ((")") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) (("=") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AND_EQUAL) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_COALESCE_EQUAL) reduce 293) ((T_CONCAT_EQUAL) reduce 293) ((T_DEC) reduce 293) ((T_DIV_EQUAL) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INC) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_MINUS_EQUAL) reduce 293) ((T_MOD_EQUAL) reduce 293) ((T_MUL_EQUAL) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_OR_EQUAL) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_PLUS_EQUAL) reduce 293) ((T_POW) reduce 293) ((T_POW_EQUAL) reduce 293) ((T_SL) reduce 293) ((T_SL_EQUAL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) ((T_SR_EQUAL) reduce 293) ((T_XOR_EQUAL) reduce 293) (("[") reduce 293) (("^") reduce 293) (("|") reduce 293)) 17308 ((("%") reduce 425) ((")") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_DOUBLE_ARROW) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_PIPE) reduce 425) ((T_POW) reduce 425) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 17310 ((("%") reduce 423) ((")") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_DOUBLE_ARROW) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_PIPE) reduce 423) ((T_POW) reduce 423) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 17320 ((("%") shift) ((")") reduce 419) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17321 ((("%") shift) ((")") reduce 421) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17322 ((("%") shift) ((")") reduce 420) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17323 ((("%") shift) ((")") reduce 413) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 413) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17324 ((("%") shift) ((")") reduce 410) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 410) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17325 ((("%") shift) ((")") reduce 418) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17326 ((("%") shift) ((")") reduce 412) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17327 ((("%") shift) ((")") reduce 416) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17328 ((("%") shift) ((")") reduce 411) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17329 ((("%") shift) ((")") reduce 414) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 414) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17330 ((("%") shift) ((")") reduce 415) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 415) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17331 ((("%") shift) ((")") reduce 422) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17332 ((("%") shift) ((")") reduce 417) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17334 ((("%") shift) ((")") reduce 406) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17337 ((("%") reduce 568) (("(") reduce 568) ((")") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 17340 ((("%") reduce 564) (("(") reduce 555) ((")") reduce 564) (("*") reduce 564) (("+") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) (("<") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_DOUBLE_ARROW) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_POW) reduce 564) ((T_SL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) (("[") reduce 564) (("^") reduce 564) (("|") reduce 564)) 17342 ((("%") reduce 563) (("(") reduce 563) ((")") reduce 563) (("*") reduce 563) (("+") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) (("<") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_DOUBLE_ARROW) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_POW) reduce 563) ((T_SL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) (("[") reduce 563) (("^") reduce 563) (("|") reduce 563)) 17344 ((("%") reduce 559) (("(") reduce 559) ((")") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 17345 ((("%") reduce 565) ((")") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 17346 ((("%") reduce 407) ((")") reduce 407) (("*") reduce 407) (("+") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_DOUBLE_ARROW) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("[") reduce 548) (("^") reduce 407) (("|") reduce 407)) 17350 ((("%") reduce 572) (("(") reduce 581) ((")") reduce 572) (("*") reduce 572) (("+") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) (("<") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_DOUBLE_ARROW) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_POW) reduce 572) ((T_SL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) (("[") reduce 572) (("^") reduce 572) (("|") reduce 572)) 17352 ((("%") reduce 507) (("(") reduce 507) ((")") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("|") reduce 507)) 17353 ((("%") reduce 291) (("(") reduce 291) ((")") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 17357 ((("%") reduce 292) (("(") reduce 292) ((")") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("[") reduce 292) (("^") reduce 292) (("|") reduce 292)) 17358 ((("%") reduce 293) (("(") reduce 293) ((")") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("[") reduce 293) (("^") reduce 293) (("|") reduce 293)) 17359 ((("%") reduce 504) (("(") reduce 504) ((")") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 17362 ((("%") reduce 571) (("(") reduce 581) ((")") reduce 571) (("*") reduce 571) (("+") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) (("<") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_DOUBLE_ARROW) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_POW) reduce 571) ((T_SL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) (("[") reduce 571) (("^") reduce 571) (("|") reduce 571)) 17363 ((("%") reduce 506) (("(") reduce 506) ((")") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("|") reduce 506)) 17364 ((("%") reduce 508) (("(") reduce 508) ((")") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("|") reduce 508)) 17369 ((("%") reduce 560) (("(") reduce 560) ((")") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 17370 ((("%") reduce 582) (("(") reduce 582) ((")") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 17371 ((("%") reduce 566) (("(") shift) ((")") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 17372 ((("%") reduce 584) (("(") reduce 584) ((")") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 17375 ((("%") reduce 583) (("(") reduce 583) ((")") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 17376 ((("%") reduce 561) (("(") reduce 561) ((")") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 17377 ((("%") reduce 567) (("(") shift) ((")") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 17378 ((("%") reduce 562) (("(") reduce 562) ((")") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("[") reduce 562) (("^") reduce 562) (("|") reduce 562)) 17379 ((("%") reduce 505) (("(") reduce 505) ((")") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 17380 ((("%") reduce 570) (("(") reduce 570) ((")") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 17383 ((("%") reduce 569) (("(") reduce 569) ((")") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 17384 ((("%") reduce 504) (("(") reduce 504) ((")") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) (("=") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AND_EQUAL) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_COALESCE_EQUAL) reduce 504) ((T_CONCAT_EQUAL) reduce 504) ((T_DEC) reduce 504) ((T_DIV_EQUAL) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INC) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_MINUS_EQUAL) reduce 504) ((T_MOD_EQUAL) reduce 504) ((T_MUL_EQUAL) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_OR_EQUAL) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_PLUS_EQUAL) reduce 504) ((T_POW) reduce 504) ((T_POW_EQUAL) reduce 504) ((T_SL) reduce 504) ((T_SL_EQUAL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) ((T_SR_EQUAL) reduce 504) ((T_XOR_EQUAL) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 17395 ((("%") reduce 494) ((")") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PIPE) reduce 494) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 17396 ((("%") reduce 489) ((")") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_PIPE) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("^") reduce 489) (("|") reduce 489)) 17406 ((("%") shift) ((")") reduce 494) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17407 ((("%") reduce 490) ((")") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_PIPE) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("^") reduce 490) (("|") reduce 490)) 17409 ((("%") shift) ((")") reduce 481) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17411 ((("%") reduce 541) (("(") reduce 579) ((")") reduce 541) (("*") reduce 541) (("+") reduce 541) (("-") reduce 541) ((".") reduce 541) (("/") reduce 541) (("<") reduce 541) ((">") reduce 541) (("?") reduce 541) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_BOOLEAN_AND) reduce 541) ((T_BOOLEAN_OR) reduce 541) ((T_COALESCE) reduce 541) ((T_DOUBLE_ARROW) reduce 541) ((T_INSTANCEOF) reduce 541) ((T_IS_EQUAL) reduce 541) ((T_IS_GREATER_OR_EQUAL) reduce 541) ((T_IS_IDENTICAL) reduce 541) ((T_IS_NOT_EQUAL) reduce 541) ((T_IS_NOT_IDENTICAL) reduce 541) ((T_IS_SMALLER_OR_EQUAL) reduce 541) ((T_LOGICAL_AND) reduce 541) ((T_LOGICAL_OR) reduce 541) ((T_LOGICAL_XOR) reduce 541) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 541) ((T_OBJECT_OPERATOR) reduce 541) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 541) ((T_PIPE) reduce 541) ((T_POW) reduce 541) ((T_SL) reduce 541) ((T_SPACESHIP) reduce 541) ((T_SR) reduce 541) (("[") reduce 541) (("^") reduce 541) (("|") reduce 541)) 17413 ((("%") reduce 571) (("(") reduce 581) ((")") reduce 571) (("*") reduce 571) (("+") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) (("<") reduce 571) (("=") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AND_EQUAL) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_COALESCE_EQUAL) reduce 571) ((T_CONCAT_EQUAL) reduce 571) ((T_DEC) reduce 571) ((T_DIV_EQUAL) reduce 571) ((T_DOUBLE_ARROW) reduce 571) ((T_INC) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_MINUS_EQUAL) reduce 571) ((T_MOD_EQUAL) reduce 571) ((T_MUL_EQUAL) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_OR_EQUAL) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_PLUS_EQUAL) reduce 571) ((T_POW) reduce 571) ((T_POW_EQUAL) reduce 571) ((T_SL) reduce 571) ((T_SL_EQUAL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) ((T_SR_EQUAL) reduce 571) ((T_XOR_EQUAL) reduce 571) (("[") reduce 571) (("^") reduce 571) (("|") reduce 571)) 17416 ((("%") reduce 543) (("(") reduce 580) ((")") reduce 543) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PIPE) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 17417 ((("%") reduce 506) (("(") reduce 506) ((")") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("|") reduce 506)) 17418 ((("%") reduce 508) (("(") reduce 508) ((")") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("|") reduce 508)) 17420 ((("%") reduce 485) ((")") reduce 485) (("*") reduce 485) (("+") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_DOUBLE_ARROW) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_PIPE) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("^") reduce 485) (("|") reduce 485)) 17421 ((("%") reduce 487) ((")") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_PIPE) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("^") reduce 487) (("|") reduce 487)) 17426 ((("%") reduce 560) (("(") reduce 560) ((")") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) (("=") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AND_EQUAL) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_COALESCE_EQUAL) reduce 560) ((T_CONCAT_EQUAL) reduce 560) ((T_DEC) reduce 560) ((T_DIV_EQUAL) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INC) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_MINUS_EQUAL) reduce 560) ((T_MOD_EQUAL) reduce 560) ((T_MUL_EQUAL) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_OR_EQUAL) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_PLUS_EQUAL) reduce 560) ((T_POW) reduce 560) ((T_POW_EQUAL) reduce 560) ((T_SL) reduce 560) ((T_SL_EQUAL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) ((T_SR_EQUAL) reduce 560) ((T_XOR_EQUAL) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 17427 ((("%") reduce 582) (("(") reduce 582) ((")") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) (("=") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AND_EQUAL) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_COALESCE_EQUAL) reduce 582) ((T_CONCAT_EQUAL) reduce 582) ((T_DEC) reduce 582) ((T_DIV_EQUAL) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INC) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_MINUS_EQUAL) reduce 582) ((T_MOD_EQUAL) reduce 582) ((T_MUL_EQUAL) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_OR_EQUAL) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_PLUS_EQUAL) reduce 582) ((T_POW) reduce 582) ((T_POW_EQUAL) reduce 582) ((T_SL) reduce 582) ((T_SL_EQUAL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) ((T_SR_EQUAL) reduce 582) ((T_XOR_EQUAL) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 17428 ((("%") reduce 566) (("(") shift) ((")") reduce 566) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 17429 ((("%") reduce 584) (("(") reduce 584) ((")") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) (("=") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AND_EQUAL) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_COALESCE_EQUAL) reduce 584) ((T_CONCAT_EQUAL) reduce 584) ((T_DEC) reduce 584) ((T_DIV_EQUAL) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INC) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_MINUS_EQUAL) reduce 584) ((T_MOD_EQUAL) reduce 584) ((T_MUL_EQUAL) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_OR_EQUAL) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_PLUS_EQUAL) reduce 584) ((T_POW) reduce 584) ((T_POW_EQUAL) reduce 584) ((T_SL) reduce 584) ((T_SL_EQUAL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) ((T_SR_EQUAL) reduce 584) ((T_XOR_EQUAL) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 17432 ((("%") reduce 583) (("(") reduce 583) ((")") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) (("=") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AND_EQUAL) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_COALESCE_EQUAL) reduce 583) ((T_CONCAT_EQUAL) reduce 583) ((T_DEC) reduce 583) ((T_DIV_EQUAL) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INC) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_MINUS_EQUAL) reduce 583) ((T_MOD_EQUAL) reduce 583) ((T_MUL_EQUAL) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_OR_EQUAL) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_PLUS_EQUAL) reduce 583) ((T_POW) reduce 583) ((T_POW_EQUAL) reduce 583) ((T_SL) reduce 583) ((T_SL_EQUAL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) ((T_SR_EQUAL) reduce 583) ((T_XOR_EQUAL) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 17433 ((("%") reduce 561) (("(") reduce 561) ((")") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) (("=") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AND_EQUAL) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_COALESCE_EQUAL) reduce 561) ((T_CONCAT_EQUAL) reduce 561) ((T_DEC) reduce 561) ((T_DIV_EQUAL) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INC) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_MINUS_EQUAL) reduce 561) ((T_MOD_EQUAL) reduce 561) ((T_MUL_EQUAL) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_OR_EQUAL) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_PLUS_EQUAL) reduce 561) ((T_POW) reduce 561) ((T_POW_EQUAL) reduce 561) ((T_SL) reduce 561) ((T_SL_EQUAL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) ((T_SR_EQUAL) reduce 561) ((T_XOR_EQUAL) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 17434 ((("%") reduce 567) (("(") shift) ((")") reduce 567) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 17435 ((("%") reduce 562) (("(") reduce 562) ((")") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) (("=") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AND_EQUAL) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_COALESCE_EQUAL) reduce 562) ((T_CONCAT_EQUAL) reduce 562) ((T_DEC) reduce 562) ((T_DIV_EQUAL) reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_INC) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_MINUS_EQUAL) reduce 562) ((T_MOD_EQUAL) reduce 562) ((T_MUL_EQUAL) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_OR_EQUAL) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_PLUS_EQUAL) reduce 562) ((T_POW) reduce 562) ((T_POW_EQUAL) reduce 562) ((T_SL) reduce 562) ((T_SL_EQUAL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) ((T_SR_EQUAL) reduce 562) ((T_XOR_EQUAL) reduce 562) (("[") reduce 562) (("^") reduce 562) (("|") reduce 562)) 17437 ((("%") reduce 477) ((")") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_PIPE) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 17439 ((("%") reduce 520) (("(") reduce 520) ((")") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) (("=") shift) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_PIPE) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("|") reduce 520)) 17441 ((("%") shift) ((")") reduce 405) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17442 ((("%") shift) ((")") reduce 482) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 482) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17445 ((("%") shift) ((")") reduce 481) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 481) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17446 ((("%") reduce 473) ((")") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_DOUBLE_ARROW) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_PIPE) reduce 473) ((T_POW) shift) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("|") reduce 473)) 17447 ((("%") shift) ((")") reduce 483) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 483) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17448 ((("%") reduce 469) ((")") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_DOUBLE_ARROW) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_PIPE) reduce 469) ((T_POW) shift) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 17449 ((("%") reduce 486) ((")") reduce 486) (("*") reduce 486) (("+") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_DOUBLE_ARROW) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_PIPE) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("^") reduce 486) (("|") reduce 486)) 17451 ((("%") reduce 526) ((")") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_PIPE) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("^") reduce 526) (("|") reduce 526)) 17453 ((("%") reduce 527) ((")") reduce 527) (("*") reduce 527) (("+") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_DOUBLE_ARROW) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_PIPE) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("^") reduce 527) (("|") reduce 527)) 17454 ((("%") reduce 525) ((")") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_PIPE) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("^") reduce 525) (("|") reduce 525)) 17455 ((("%") shift) ((")") reduce 619) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 619) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17456 ((("%") shift) ((")") reduce 618) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 618) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17457 ((("%") reduce 505) (("(") reduce 505) ((")") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 17458 ((("%") shift) ((")") reduce 478) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 478) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17459 ((("%") reduce 471) ((")") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_DOUBLE_ARROW) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_PIPE) reduce 471) ((T_POW) shift) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("|") reduce 471)) 17461 ((("%") reduce 94) (("(") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 17462 ((("%") reduce 93) (("(") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 17463 ((("%") reduce 95) (("(") reduce 95) ((")") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 17464 ((("%") reduce 509) (("(") reduce 509) ((")") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 17465 ((("%") reduce 92) (("(") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 17466 ((("%") reduce 400) (("(") reduce 400) ((")") reduce 400) (("*") reduce 400) (("+") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_DOUBLE_ARROW) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 400) ((T_OBJECT_OPERATOR) reduce 400) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 400) ((T_PIPE) reduce 400) ((T_POW) reduce 400) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("[") reduce 400) (("^") reduce 400) (("|") reduce 400)) 17469 ((("%") reduce 511) (("(") reduce 511) ((")") reduce 511) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("|") reduce 511)) 17470 ((("%") reduce 402) (("(") shift) ((")") reduce 402) (("*") reduce 402) (("+") reduce 402) (("-") reduce 402) ((".") reduce 402) (("/") reduce 402) (("<") reduce 402) ((">") reduce 402) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_DOUBLE_ARROW) reduce 402) ((T_INSTANCEOF) reduce 402) ((T_IS_EQUAL) reduce 402) ((T_IS_GREATER_OR_EQUAL) reduce 402) ((T_IS_IDENTICAL) reduce 402) ((T_IS_NOT_EQUAL) reduce 402) ((T_IS_NOT_IDENTICAL) reduce 402) ((T_IS_SMALLER_OR_EQUAL) reduce 402) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_PIPE) reduce 402) ((T_POW) reduce 402) ((T_SL) reduce 402) ((T_SPACESHIP) reduce 402) ((T_SR) reduce 402) (("^") reduce 402) (("|") reduce 402)) 17471 ((("%") reduce 510) (("(") reduce 510) ((")") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 17472 ((("%") reduce 512) (("(") reduce 512) ((")") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("|") reduce 512)) 17473 ((("%") reduce 573) (("(") reduce 573) ((")") reduce 573) (("*") reduce 573) (("+") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_DOUBLE_ARROW) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("|") reduce 573)) 17479 ((("%") reduce 574) (("(") reduce 574) ((")") reduce 574) (("*") reduce 574) (("+") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_DOUBLE_ARROW) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("|") reduce 574)) 17480 ((("%") reduce 578) (("(") reduce 578) ((")") reduce 578) (("*") reduce 578) (("+") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_DOUBLE_ARROW) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("|") reduce 578)) 17481 ((("%") reduce 575) (("(") reduce 575) ((")") reduce 575) (("*") reduce 575) (("+") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_DOUBLE_ARROW) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("|") reduce 575)) 17482 ((("%") reduce 576) (("(") reduce 576) ((")") reduce 576) (("*") reduce 576) (("+") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_DOUBLE_ARROW) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("|") reduce 576)) 17483 ((("%") reduce 399) (("(") reduce 399) ((")") reduce 399) (("*") reduce 399) (("+") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_DOUBLE_ARROW) reduce 399) ((T_INSTANCEOF) reduce 399) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 399) ((T_OBJECT_OPERATOR) reduce 399) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 399) ((T_PIPE) reduce 399) ((T_POW) reduce 399) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("[") reduce 399) (("^") reduce 399) (("|") reduce 399)) 17485 ((("%") reduce 577) (("(") reduce 577) ((")") reduce 577) (("*") reduce 577) (("+") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_DOUBLE_ARROW) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("|") reduce 577)) 17486 ((("%") reduce 401) (("(") reduce 401) ((")") reduce 401) (("*") reduce 401) (("+") reduce 401) (("-") reduce 401) ((".") reduce 401) (("/") reduce 401) (("<") reduce 401) ((">") reduce 401) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_DOUBLE_ARROW) reduce 401) ((T_INSTANCEOF) reduce 401) ((T_IS_EQUAL) reduce 401) ((T_IS_GREATER_OR_EQUAL) reduce 401) ((T_IS_IDENTICAL) reduce 401) ((T_IS_NOT_EQUAL) reduce 401) ((T_IS_NOT_IDENTICAL) reduce 401) ((T_IS_SMALLER_OR_EQUAL) reduce 401) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 401) ((T_OBJECT_OPERATOR) reduce 401) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 401) ((T_PIPE) reduce 401) ((T_POW) reduce 401) ((T_SL) reduce 401) ((T_SPACESHIP) reduce 401) ((T_SR) reduce 401) (("[") reduce 401) (("^") reduce 401) (("|") reduce 401)) 17494 ((("%") reduce 398) (("(") reduce 398) ((")") reduce 398) (("*") reduce 398) (("+") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_DOUBLE_ARROW) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 398) ((T_OBJECT_OPERATOR) reduce 398) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 398) ((T_PIPE) reduce 398) ((T_POW) reduce 398) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("[") reduce 398) (("^") reduce 398) (("|") reduce 398)) 17496 ((("%") reduce 513) (("(") reduce 513) ((")") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("|") reduce 513)) 17502 ((("%") reduce 233) ((")") reduce 233) (("*") reduce 233) (("+") reduce 233) (("-") reduce 233) ((".") reduce 233) (("/") reduce 233) (("<") reduce 233) ((">") reduce 233) (("?") reduce 233) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_BOOLEAN_AND) reduce 233) ((T_BOOLEAN_OR) reduce 233) ((T_COALESCE) reduce 233) ((T_DOUBLE_ARROW) reduce 233) ((T_INSTANCEOF) reduce 233) ((T_IS_EQUAL) reduce 233) ((T_IS_GREATER_OR_EQUAL) reduce 233) ((T_IS_IDENTICAL) reduce 233) ((T_IS_NOT_EQUAL) reduce 233) ((T_IS_NOT_IDENTICAL) reduce 233) ((T_IS_SMALLER_OR_EQUAL) reduce 233) ((T_LOGICAL_AND) reduce 233) ((T_LOGICAL_OR) reduce 233) ((T_LOGICAL_XOR) reduce 233) ((T_PIPE) reduce 233) ((T_POW) reduce 233) ((T_SL) reduce 233) ((T_SPACESHIP) reduce 233) ((T_SR) reduce 233) (("^") reduce 233) (("|") reduce 233)) 17507 ((("%") shift) ((")") reduce 404) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17511 ((("%") reduce 613) ((")") reduce 613) (("*") reduce 613) (("+") reduce 613) (("-") reduce 613) ((".") reduce 613) (("/") reduce 613) (("<") reduce 613) ((">") reduce 613) (("?") reduce 613) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_BOOLEAN_AND) reduce 613) ((T_BOOLEAN_OR) reduce 613) ((T_COALESCE) reduce 613) ((T_DOUBLE_ARROW) reduce 613) ((T_INSTANCEOF) reduce 613) ((T_IS_EQUAL) reduce 613) ((T_IS_GREATER_OR_EQUAL) reduce 613) ((T_IS_IDENTICAL) reduce 613) ((T_IS_NOT_EQUAL) reduce 613) ((T_IS_NOT_IDENTICAL) reduce 613) ((T_IS_SMALLER_OR_EQUAL) reduce 613) ((T_LOGICAL_AND) reduce 613) ((T_LOGICAL_OR) reduce 613) ((T_LOGICAL_XOR) reduce 613) ((T_PIPE) reduce 613) ((T_POW) reduce 613) ((T_SL) reduce 613) ((T_SPACESHIP) reduce 613) ((T_SR) reduce 613) (("^") reduce 613) (("|") reduce 613)) 17512 ((("%") reduce 467) ((")") reduce 467) (("*") reduce 467) (("+") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_DOUBLE_ARROW) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_PIPE) reduce 467) ((T_POW) shift) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("|") reduce 467)) 17513 ((("%") shift) ((")") reduce 616) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 616) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17514 ((("%") shift) ((")") reduce 615) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 615) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17515 ((("%") reduce 424) ((")") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_DOUBLE_ARROW) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("[") reduce 548) (("^") reduce 424) (("|") reduce 424)) 17517 ((("%") reduce 518) ((")") reduce 518) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_PIPE) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("^") reduce 518) (("|") reduce 518)) 17518 ((("%") reduce 474) ((")") reduce 474) (("*") reduce 474) (("+") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_DOUBLE_ARROW) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_PIPE) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("|") reduce 474)) 17519 ((("%") reduce 291) ((")") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("|") reduce 291)) 17523 ((("%") reduce 292) ((")") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("^") reduce 292) (("|") reduce 292)) 17524 ((("%") reduce 293) ((")") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("^") reduce 293) (("|") reduce 293)) 17527 ((("%") reduce 617) ((")") reduce 617) (("*") reduce 617) (("+") reduce 617) (("-") reduce 617) ((".") reduce 617) (("/") reduce 617) (("<") reduce 617) ((">") reduce 617) (("?") reduce 617) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_BOOLEAN_AND) reduce 617) ((T_BOOLEAN_OR) reduce 617) ((T_COALESCE) reduce 617) ((T_DOUBLE_ARROW) reduce 617) ((T_INSTANCEOF) reduce 617) ((T_IS_EQUAL) reduce 617) ((T_IS_GREATER_OR_EQUAL) reduce 617) ((T_IS_IDENTICAL) reduce 617) ((T_IS_NOT_EQUAL) reduce 617) ((T_IS_NOT_IDENTICAL) reduce 617) ((T_IS_SMALLER_OR_EQUAL) reduce 617) ((T_LOGICAL_AND) reduce 617) ((T_LOGICAL_OR) reduce 617) ((T_LOGICAL_XOR) reduce 617) ((T_PIPE) reduce 617) ((T_POW) reduce 617) ((T_SL) reduce 617) ((T_SPACESHIP) reduce 617) ((T_SR) reduce 617) (("^") reduce 617) (("|") reduce 617)) 17530 ((("%") reduce 614) ((")") reduce 614) (("*") reduce 614) (("+") reduce 614) (("-") reduce 614) ((".") reduce 614) (("/") reduce 614) (("<") reduce 614) ((">") reduce 614) (("?") reduce 614) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_BOOLEAN_AND) reduce 614) ((T_BOOLEAN_OR) reduce 614) ((T_COALESCE) reduce 614) ((T_DOUBLE_ARROW) reduce 614) ((T_INSTANCEOF) reduce 614) ((T_IS_EQUAL) reduce 614) ((T_IS_GREATER_OR_EQUAL) reduce 614) ((T_IS_IDENTICAL) reduce 614) ((T_IS_NOT_EQUAL) reduce 614) ((T_IS_NOT_IDENTICAL) reduce 614) ((T_IS_SMALLER_OR_EQUAL) reduce 614) ((T_LOGICAL_AND) reduce 614) ((T_LOGICAL_OR) reduce 614) ((T_LOGICAL_XOR) reduce 614) ((T_PIPE) reduce 614) ((T_POW) reduce 614) ((T_SL) reduce 614) ((T_SPACESHIP) reduce 614) ((T_SR) reduce 614) (("^") reduce 614) (("|") reduce 614)) 17531 ((("%") reduce 468) ((")") reduce 468) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_DOUBLE_ARROW) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_PIPE) reduce 468) ((T_POW) shift) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 17532 ((("%") reduce 426) ((")") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_DOUBLE_ARROW) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("[") reduce 548) (("^") reduce 426) (("|") reduce 426)) 17534 ((("%") reduce 408) ((")") reduce 408) (("*") reduce 408) (("+") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_DOUBLE_ARROW) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_PIPE) reduce 408) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("^") reduce 408) (("|") reduce 408)) 17535 ((("%") reduce 409) ((")") reduce 409) (("*") reduce 409) (("+") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_DOUBLE_ARROW) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_PIPE) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("|") reduce 409)) 17536 ((("%") reduce 296) ((")") reduce 296) (("*") reduce 296) (("+") reduce 296) (("-") reduce 296) ((".") reduce 296) (("/") reduce 296) (("<") reduce 296) ((">") reduce 296) (("?") reduce 296) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_BOOLEAN_AND) reduce 296) ((T_BOOLEAN_OR) reduce 296) ((T_COALESCE) reduce 296) ((T_DOUBLE_ARROW) reduce 296) ((T_INSTANCEOF) reduce 296) ((T_IS_EQUAL) reduce 296) ((T_IS_GREATER_OR_EQUAL) reduce 296) ((T_IS_IDENTICAL) reduce 296) ((T_IS_NOT_EQUAL) reduce 296) ((T_IS_NOT_IDENTICAL) reduce 296) ((T_IS_SMALLER_OR_EQUAL) reduce 296) ((T_LOGICAL_AND) reduce 296) ((T_LOGICAL_OR) reduce 296) ((T_LOGICAL_XOR) reduce 296) ((T_PIPE) reduce 296) ((T_POW) reduce 296) ((T_SL) reduce 296) ((T_SPACESHIP) reduce 296) ((T_SR) reduce 296) (("^") reduce 296) (("|") reduce 296)) 17541 ((("%") reduce 297) ((")") reduce 297) (("*") reduce 297) (("+") reduce 297) (("-") reduce 297) ((".") reduce 297) (("/") reduce 297) (("<") reduce 297) ((">") reduce 297) (("?") reduce 297) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_BOOLEAN_AND) reduce 297) ((T_BOOLEAN_OR) reduce 297) ((T_COALESCE) reduce 297) ((T_DOUBLE_ARROW) reduce 297) ((T_INSTANCEOF) reduce 297) ((T_IS_EQUAL) reduce 297) ((T_IS_GREATER_OR_EQUAL) reduce 297) ((T_IS_IDENTICAL) reduce 297) ((T_IS_NOT_EQUAL) reduce 297) ((T_IS_NOT_IDENTICAL) reduce 297) ((T_IS_SMALLER_OR_EQUAL) reduce 297) ((T_LOGICAL_AND) reduce 297) ((T_LOGICAL_OR) reduce 297) ((T_LOGICAL_XOR) reduce 297) ((T_PIPE) reduce 297) ((T_POW) reduce 297) ((T_SL) reduce 297) ((T_SPACESHIP) reduce 297) ((T_SR) reduce 297) (("^") reduce 297) (("|") reduce 297)) 17542 ((("%") reduce 460) (("(") reduce 556) ((")") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PIPE) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("[") reduce 549) (("^") reduce 460) (("|") reduce 460)) 17544 ((("%") reduce 298) ((")") reduce 298) (("*") reduce 298) (("+") reduce 298) (("-") reduce 298) ((".") reduce 298) (("/") reduce 298) (("<") reduce 298) ((">") reduce 298) (("?") reduce 298) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_BOOLEAN_AND) reduce 298) ((T_BOOLEAN_OR) reduce 298) ((T_COALESCE) reduce 298) ((T_DOUBLE_ARROW) reduce 298) ((T_INSTANCEOF) reduce 298) ((T_IS_EQUAL) reduce 298) ((T_IS_GREATER_OR_EQUAL) reduce 298) ((T_IS_IDENTICAL) reduce 298) ((T_IS_NOT_EQUAL) reduce 298) ((T_IS_NOT_IDENTICAL) reduce 298) ((T_IS_SMALLER_OR_EQUAL) reduce 298) ((T_LOGICAL_AND) reduce 298) ((T_LOGICAL_OR) reduce 298) ((T_LOGICAL_XOR) reduce 298) ((T_PIPE) reduce 298) ((T_POW) reduce 298) ((T_SL) reduce 298) ((T_SPACESHIP) reduce 298) ((T_SR) reduce 298) (("^") reduce 298) (("|") reduce 298)) 17545 ((("%") reduce 299) ((")") reduce 299) (("*") reduce 299) (("+") reduce 299) (("-") reduce 299) ((".") reduce 299) (("/") reduce 299) (("<") reduce 299) ((">") reduce 299) (("?") reduce 299) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_BOOLEAN_AND) reduce 299) ((T_BOOLEAN_OR) reduce 299) ((T_COALESCE) reduce 299) ((T_DOUBLE_ARROW) reduce 299) ((T_INSTANCEOF) reduce 299) ((T_IS_EQUAL) reduce 299) ((T_IS_GREATER_OR_EQUAL) reduce 299) ((T_IS_IDENTICAL) reduce 299) ((T_IS_NOT_EQUAL) reduce 299) ((T_IS_NOT_IDENTICAL) reduce 299) ((T_IS_SMALLER_OR_EQUAL) reduce 299) ((T_LOGICAL_AND) reduce 299) ((T_LOGICAL_OR) reduce 299) ((T_LOGICAL_XOR) reduce 299) ((T_PIPE) reduce 299) ((T_POW) reduce 299) ((T_SL) reduce 299) ((T_SPACESHIP) reduce 299) ((T_SR) reduce 299) (("^") reduce 299) (("|") reduce 299)) 17546 ((("%") reduce 472) ((")") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_DOUBLE_ARROW) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_PIPE) reduce 472) ((T_POW) shift) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("|") reduce 472)) 17547 ((("%") reduce 470) ((")") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_DOUBLE_ARROW) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_PIPE) reduce 470) ((T_POW) shift) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("|") reduce 470)) 17550 ((("%") reduce 519) (("(") reduce 519) ((")") reduce 519) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PIPE) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("|") reduce 519)) 17551 ((("%") reduce 475) ((")") reduce 475) (("*") reduce 475) (("+") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_DOUBLE_ARROW) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_PIPE) reduce 475) ((T_POW) shift) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("|") reduce 475)) 17552 ((("%") reduce 446) ((")") reduce 446) (("*") reduce 446) (("+") reduce 446) (("-") reduce 446) ((".") reduce 446) (("/") reduce 446) (("<") reduce 446) ((">") reduce 446) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_BOOLEAN_AND) reduce 446) ((T_BOOLEAN_OR) reduce 446) ((T_COALESCE) reduce 446) ((T_DOUBLE_ARROW) reduce 446) ((T_INSTANCEOF) reduce 446) ((T_IS_EQUAL) reduce 446) ((T_IS_GREATER_OR_EQUAL) reduce 446) ((T_IS_IDENTICAL) reduce 446) ((T_IS_NOT_EQUAL) reduce 446) ((T_IS_NOT_IDENTICAL) reduce 446) ((T_IS_SMALLER_OR_EQUAL) reduce 446) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_PIPE) reduce 446) ((T_POW) shift) ((T_SL) reduce 446) ((T_SPACESHIP) reduce 446) ((T_SR) reduce 446) (("^") reduce 446) (("|") reduce 446)) 17553 ((("%") reduce 445) ((")") reduce 445) (("*") reduce 445) (("+") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_DOUBLE_ARROW) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_PIPE) reduce 445) ((T_POW) shift) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("|") reduce 445)) 17555 ((("%") reduce 570) (("(") reduce 570) ((")") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) (("=") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AND_EQUAL) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_COALESCE_EQUAL) reduce 570) ((T_CONCAT_EQUAL) reduce 570) ((T_DEC) reduce 570) ((T_DIV_EQUAL) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INC) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_MINUS_EQUAL) reduce 570) ((T_MOD_EQUAL) reduce 570) ((T_MUL_EQUAL) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_OR_EQUAL) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_PLUS_EQUAL) reduce 570) ((T_POW) reduce 570) ((T_POW_EQUAL) reduce 570) ((T_SL) reduce 570) ((T_SL_EQUAL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) ((T_SR_EQUAL) reduce 570) ((T_XOR_EQUAL) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 17558 ((("%") reduce 569) (("(") reduce 569) ((")") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) (("=") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AND_EQUAL) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_COALESCE_EQUAL) reduce 569) ((T_CONCAT_EQUAL) reduce 569) ((T_DEC) reduce 569) ((T_DIV_EQUAL) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INC) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_MINUS_EQUAL) reduce 569) ((T_MOD_EQUAL) reduce 569) ((T_MUL_EQUAL) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_OR_EQUAL) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_PLUS_EQUAL) reduce 569) ((T_POW) reduce 569) ((T_POW_EQUAL) reduce 569) ((T_SL) reduce 569) ((T_SL_EQUAL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) ((T_SR_EQUAL) reduce 569) ((T_XOR_EQUAL) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 17560 ((("%") reduce 522) (("(") reduce 522) ((")") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_PIPE) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("|") reduce 522)) 17561 ((("%") reduce 447) ((")") reduce 447) (("*") reduce 447) (("+") reduce 447) (("-") reduce 447) ((".") reduce 447) (("/") reduce 447) (("<") reduce 447) ((">") reduce 447) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_BOOLEAN_AND) reduce 447) ((T_BOOLEAN_OR) reduce 447) ((T_COALESCE) reduce 447) ((T_DOUBLE_ARROW) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 447) ((T_IS_GREATER_OR_EQUAL) reduce 447) ((T_IS_IDENTICAL) reduce 447) ((T_IS_NOT_EQUAL) reduce 447) ((T_IS_NOT_IDENTICAL) reduce 447) ((T_IS_SMALLER_OR_EQUAL) reduce 447) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_PIPE) reduce 447) ((T_POW) shift) ((T_SL) reduce 447) ((T_SPACESHIP) reduce 447) ((T_SR) reduce 447) (("^") reduce 447) (("|") reduce 447)) 17562 ((("%") reduce 473) ((")") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_PIPE) reduce 473) ((T_POW) shift) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("|") reduce 473)) 17563 ((("%") shift) ((")") reduce 483) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17564 ((("%") reduce 469) ((")") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_PIPE) reduce 469) ((T_POW) shift) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 17565 ((("%") reduce 486) ((")") reduce 486) (("*") reduce 486) (("+") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_PIPE) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("^") reduce 486) (("|") reduce 486)) 17567 ((("%") reduce 526) ((")") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_PIPE) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("^") reduce 526) (("|") reduce 526)) 17569 ((("%") reduce 527) ((")") reduce 527) (("*") reduce 527) (("+") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_PIPE) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("^") reduce 527) (("|") reduce 527)) 17570 ((("%") reduce 525) ((")") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_PIPE) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("^") reduce 525) (("|") reduce 525)) 17571 ((("%") shift) ((")") reduce 619) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17572 ((("%") shift) ((")") reduce 618) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17573 ((("%") reduce 505) (("(") reduce 505) ((")") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 17574 ((("%") shift) ((")") reduce 478) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17575 ((("%") reduce 471) ((")") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_PIPE) reduce 471) ((T_POW) shift) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("|") reduce 471)) 17577 ((("%") reduce 94) (("(") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 17578 ((("%") reduce 93) (("(") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 17579 ((("%") reduce 95) (("(") reduce 95) ((")") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 17580 ((("%") reduce 509) (("(") reduce 509) ((")") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 17581 ((("%") reduce 92) (("(") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 17582 ((("%") reduce 400) (("(") reduce 400) ((")") reduce 400) (("*") reduce 400) (("+") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 400) ((T_OBJECT_OPERATOR) reduce 400) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 400) ((T_PIPE) reduce 400) ((T_POW) reduce 400) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("[") reduce 400) (("^") reduce 400) (("|") reduce 400)) 17585 ((("%") reduce 511) (("(") reduce 511) ((")") reduce 511) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("|") reduce 511)) 17586 ((("%") reduce 402) (("(") shift) ((")") reduce 402) (("*") reduce 402) (("+") reduce 402) (("-") reduce 402) ((".") reduce 402) (("/") reduce 402) (("<") reduce 402) ((">") reduce 402) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_INSTANCEOF) reduce 402) ((T_IS_EQUAL) reduce 402) ((T_IS_GREATER_OR_EQUAL) reduce 402) ((T_IS_IDENTICAL) reduce 402) ((T_IS_NOT_EQUAL) reduce 402) ((T_IS_NOT_IDENTICAL) reduce 402) ((T_IS_SMALLER_OR_EQUAL) reduce 402) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_PIPE) reduce 402) ((T_POW) reduce 402) ((T_SL) reduce 402) ((T_SPACESHIP) reduce 402) ((T_SR) reduce 402) (("^") reduce 402) (("|") reduce 402)) 17587 ((("%") reduce 510) (("(") reduce 510) ((")") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 17588 ((("%") reduce 512) (("(") reduce 512) ((")") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("|") reduce 512)) 17589 ((("%") reduce 573) (("(") reduce 573) ((")") reduce 573) (("*") reduce 573) (("+") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("|") reduce 573)) 17595 ((("%") reduce 574) (("(") reduce 574) ((")") reduce 574) (("*") reduce 574) (("+") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("|") reduce 574)) 17596 ((("%") reduce 578) (("(") reduce 578) ((")") reduce 578) (("*") reduce 578) (("+") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("|") reduce 578)) 17597 ((("%") reduce 575) (("(") reduce 575) ((")") reduce 575) (("*") reduce 575) (("+") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("|") reduce 575)) 17598 ((("%") reduce 576) (("(") reduce 576) ((")") reduce 576) (("*") reduce 576) (("+") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("|") reduce 576)) 17599 ((("%") reduce 399) (("(") reduce 399) ((")") reduce 399) (("*") reduce 399) (("+") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_INSTANCEOF) reduce 399) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 399) ((T_OBJECT_OPERATOR) reduce 399) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 399) ((T_PIPE) reduce 399) ((T_POW) reduce 399) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("[") reduce 399) (("^") reduce 399) (("|") reduce 399)) 17601 ((("%") reduce 577) (("(") reduce 577) ((")") reduce 577) (("*") reduce 577) (("+") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("|") reduce 577)) 17602 ((("%") reduce 401) (("(") reduce 401) ((")") reduce 401) (("*") reduce 401) (("+") reduce 401) (("-") reduce 401) ((".") reduce 401) (("/") reduce 401) (("<") reduce 401) ((">") reduce 401) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_INSTANCEOF) reduce 401) ((T_IS_EQUAL) reduce 401) ((T_IS_GREATER_OR_EQUAL) reduce 401) ((T_IS_IDENTICAL) reduce 401) ((T_IS_NOT_EQUAL) reduce 401) ((T_IS_NOT_IDENTICAL) reduce 401) ((T_IS_SMALLER_OR_EQUAL) reduce 401) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 401) ((T_OBJECT_OPERATOR) reduce 401) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 401) ((T_PIPE) reduce 401) ((T_POW) reduce 401) ((T_SL) reduce 401) ((T_SPACESHIP) reduce 401) ((T_SR) reduce 401) (("[") reduce 401) (("^") reduce 401) (("|") reduce 401)) 17610 ((("%") reduce 398) (("(") reduce 398) ((")") reduce 398) (("*") reduce 398) (("+") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 398) ((T_OBJECT_OPERATOR) reduce 398) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 398) ((T_PIPE) reduce 398) ((T_POW) reduce 398) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("[") reduce 398) (("^") reduce 398) (("|") reduce 398)) 17612 ((("%") reduce 513) (("(") reduce 513) ((")") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("|") reduce 513)) 17618 ((("%") reduce 233) ((")") reduce 233) (("*") reduce 233) (("+") reduce 233) (("-") reduce 233) ((".") reduce 233) (("/") reduce 233) (("<") reduce 233) ((">") reduce 233) (("?") reduce 233) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_BOOLEAN_AND) reduce 233) ((T_BOOLEAN_OR) reduce 233) ((T_COALESCE) reduce 233) ((T_INSTANCEOF) reduce 233) ((T_IS_EQUAL) reduce 233) ((T_IS_GREATER_OR_EQUAL) reduce 233) ((T_IS_IDENTICAL) reduce 233) ((T_IS_NOT_EQUAL) reduce 233) ((T_IS_NOT_IDENTICAL) reduce 233) ((T_IS_SMALLER_OR_EQUAL) reduce 233) ((T_LOGICAL_AND) reduce 233) ((T_LOGICAL_OR) reduce 233) ((T_LOGICAL_XOR) reduce 233) ((T_PIPE) reduce 233) ((T_POW) reduce 233) ((T_SL) reduce 233) ((T_SPACESHIP) reduce 233) ((T_SR) reduce 233) (("^") reduce 233) (("|") reduce 233)) 17623 ((("%") shift) ((")") reduce 404) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17627 ((("%") reduce 613) ((")") reduce 613) (("*") reduce 613) (("+") reduce 613) (("-") reduce 613) ((".") reduce 613) (("/") reduce 613) (("<") reduce 613) ((">") reduce 613) (("?") reduce 613) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_BOOLEAN_AND) reduce 613) ((T_BOOLEAN_OR) reduce 613) ((T_COALESCE) reduce 613) ((T_INSTANCEOF) reduce 613) ((T_IS_EQUAL) reduce 613) ((T_IS_GREATER_OR_EQUAL) reduce 613) ((T_IS_IDENTICAL) reduce 613) ((T_IS_NOT_EQUAL) reduce 613) ((T_IS_NOT_IDENTICAL) reduce 613) ((T_IS_SMALLER_OR_EQUAL) reduce 613) ((T_LOGICAL_AND) reduce 613) ((T_LOGICAL_OR) reduce 613) ((T_LOGICAL_XOR) reduce 613) ((T_PIPE) reduce 613) ((T_POW) reduce 613) ((T_SL) reduce 613) ((T_SPACESHIP) reduce 613) ((T_SR) reduce 613) (("^") reduce 613) (("|") reduce 613)) 17628 ((("%") reduce 467) ((")") reduce 467) (("*") reduce 467) (("+") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_PIPE) reduce 467) ((T_POW) shift) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("|") reduce 467)) 17629 ((("%") shift) ((")") reduce 616) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17630 ((("%") shift) ((")") reduce 615) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 17631 ((("%") reduce 424) ((")") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("[") reduce 548) (("^") reduce 424) (("|") reduce 424)) 17633 ((("%") reduce 518) ((")") reduce 518) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_PIPE) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("^") reduce 518) (("|") reduce 518)) 17634 ((("%") reduce 474) ((")") reduce 474) (("*") reduce 474) (("+") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_PIPE) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("|") reduce 474)) 17635 ((("%") reduce 291) ((")") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("|") reduce 291)) 17639 ((("%") reduce 292) ((")") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("^") reduce 292) (("|") reduce 292)) 17640 ((("%") reduce 293) ((")") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("^") reduce 293) (("|") reduce 293)) 17643 ((("%") reduce 617) ((")") reduce 617) (("*") reduce 617) (("+") reduce 617) (("-") reduce 617) ((".") reduce 617) (("/") reduce 617) (("<") reduce 617) ((">") reduce 617) (("?") reduce 617) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_BOOLEAN_AND) reduce 617) ((T_BOOLEAN_OR) reduce 617) ((T_COALESCE) reduce 617) ((T_INSTANCEOF) reduce 617) ((T_IS_EQUAL) reduce 617) ((T_IS_GREATER_OR_EQUAL) reduce 617) ((T_IS_IDENTICAL) reduce 617) ((T_IS_NOT_EQUAL) reduce 617) ((T_IS_NOT_IDENTICAL) reduce 617) ((T_IS_SMALLER_OR_EQUAL) reduce 617) ((T_LOGICAL_AND) reduce 617) ((T_LOGICAL_OR) reduce 617) ((T_LOGICAL_XOR) reduce 617) ((T_PIPE) reduce 617) ((T_POW) reduce 617) ((T_SL) reduce 617) ((T_SPACESHIP) reduce 617) ((T_SR) reduce 617) (("^") reduce 617) (("|") reduce 617)) 17646 ((("%") reduce 614) ((")") reduce 614) (("*") reduce 614) (("+") reduce 614) (("-") reduce 614) ((".") reduce 614) (("/") reduce 614) (("<") reduce 614) ((">") reduce 614) (("?") reduce 614) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_BOOLEAN_AND) reduce 614) ((T_BOOLEAN_OR) reduce 614) ((T_COALESCE) reduce 614) ((T_INSTANCEOF) reduce 614) ((T_IS_EQUAL) reduce 614) ((T_IS_GREATER_OR_EQUAL) reduce 614) ((T_IS_IDENTICAL) reduce 614) ((T_IS_NOT_EQUAL) reduce 614) ((T_IS_NOT_IDENTICAL) reduce 614) ((T_IS_SMALLER_OR_EQUAL) reduce 614) ((T_LOGICAL_AND) reduce 614) ((T_LOGICAL_OR) reduce 614) ((T_LOGICAL_XOR) reduce 614) ((T_PIPE) reduce 614) ((T_POW) reduce 614) ((T_SL) reduce 614) ((T_SPACESHIP) reduce 614) ((T_SR) reduce 614) (("^") reduce 614) (("|") reduce 614)) 17647 ((("%") reduce 468) ((")") reduce 468) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_PIPE) reduce 468) ((T_POW) shift) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 17648 ((("%") reduce 426) ((")") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("[") reduce 548) (("^") reduce 426) (("|") reduce 426)) 17650 ((("%") reduce 408) ((")") reduce 408) (("*") reduce 408) (("+") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_PIPE) reduce 408) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("^") reduce 408) (("|") reduce 408)) 17651 ((("%") reduce 409) ((")") reduce 409) (("*") reduce 409) (("+") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_PIPE) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("|") reduce 409)) 17652 ((("%") reduce 296) ((")") reduce 296) (("*") reduce 296) (("+") reduce 296) (("-") reduce 296) ((".") reduce 296) (("/") reduce 296) (("<") reduce 296) ((">") reduce 296) (("?") reduce 296) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_BOOLEAN_AND) reduce 296) ((T_BOOLEAN_OR) reduce 296) ((T_COALESCE) reduce 296) ((T_INSTANCEOF) reduce 296) ((T_IS_EQUAL) reduce 296) ((T_IS_GREATER_OR_EQUAL) reduce 296) ((T_IS_IDENTICAL) reduce 296) ((T_IS_NOT_EQUAL) reduce 296) ((T_IS_NOT_IDENTICAL) reduce 296) ((T_IS_SMALLER_OR_EQUAL) reduce 296) ((T_LOGICAL_AND) reduce 296) ((T_LOGICAL_OR) reduce 296) ((T_LOGICAL_XOR) reduce 296) ((T_PIPE) reduce 296) ((T_POW) reduce 296) ((T_SL) reduce 296) ((T_SPACESHIP) reduce 296) ((T_SR) reduce 296) (("^") reduce 296) (("|") reduce 296)) 17657 ((("%") reduce 297) ((")") reduce 297) (("*") reduce 297) (("+") reduce 297) (("-") reduce 297) ((".") reduce 297) (("/") reduce 297) (("<") reduce 297) ((">") reduce 297) (("?") reduce 297) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_BOOLEAN_AND) reduce 297) ((T_BOOLEAN_OR) reduce 297) ((T_COALESCE) reduce 297) ((T_INSTANCEOF) reduce 297) ((T_IS_EQUAL) reduce 297) ((T_IS_GREATER_OR_EQUAL) reduce 297) ((T_IS_IDENTICAL) reduce 297) ((T_IS_NOT_EQUAL) reduce 297) ((T_IS_NOT_IDENTICAL) reduce 297) ((T_IS_SMALLER_OR_EQUAL) reduce 297) ((T_LOGICAL_AND) reduce 297) ((T_LOGICAL_OR) reduce 297) ((T_LOGICAL_XOR) reduce 297) ((T_PIPE) reduce 297) ((T_POW) reduce 297) ((T_SL) reduce 297) ((T_SPACESHIP) reduce 297) ((T_SR) reduce 297) (("^") reduce 297) (("|") reduce 297)) 17658 ((("%") reduce 460) (("(") reduce 556) ((")") reduce 460) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PIPE) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("[") reduce 549) (("^") reduce 460) (("|") reduce 460)) 17660 ((("%") reduce 298) ((")") reduce 298) (("*") reduce 298) (("+") reduce 298) (("-") reduce 298) ((".") reduce 298) (("/") reduce 298) (("<") reduce 298) ((">") reduce 298) (("?") reduce 298) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_BOOLEAN_AND) reduce 298) ((T_BOOLEAN_OR) reduce 298) ((T_COALESCE) reduce 298) ((T_INSTANCEOF) reduce 298) ((T_IS_EQUAL) reduce 298) ((T_IS_GREATER_OR_EQUAL) reduce 298) ((T_IS_IDENTICAL) reduce 298) ((T_IS_NOT_EQUAL) reduce 298) ((T_IS_NOT_IDENTICAL) reduce 298) ((T_IS_SMALLER_OR_EQUAL) reduce 298) ((T_LOGICAL_AND) reduce 298) ((T_LOGICAL_OR) reduce 298) ((T_LOGICAL_XOR) reduce 298) ((T_PIPE) reduce 298) ((T_POW) reduce 298) ((T_SL) reduce 298) ((T_SPACESHIP) reduce 298) ((T_SR) reduce 298) (("^") reduce 298) (("|") reduce 298)) 17661 ((("%") reduce 299) ((")") reduce 299) (("*") reduce 299) (("+") reduce 299) (("-") reduce 299) ((".") reduce 299) (("/") reduce 299) (("<") reduce 299) ((">") reduce 299) (("?") reduce 299) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_BOOLEAN_AND) reduce 299) ((T_BOOLEAN_OR) reduce 299) ((T_COALESCE) reduce 299) ((T_INSTANCEOF) reduce 299) ((T_IS_EQUAL) reduce 299) ((T_IS_GREATER_OR_EQUAL) reduce 299) ((T_IS_IDENTICAL) reduce 299) ((T_IS_NOT_EQUAL) reduce 299) ((T_IS_NOT_IDENTICAL) reduce 299) ((T_IS_SMALLER_OR_EQUAL) reduce 299) ((T_LOGICAL_AND) reduce 299) ((T_LOGICAL_OR) reduce 299) ((T_LOGICAL_XOR) reduce 299) ((T_PIPE) reduce 299) ((T_POW) reduce 299) ((T_SL) reduce 299) ((T_SPACESHIP) reduce 299) ((T_SR) reduce 299) (("^") reduce 299) (("|") reduce 299)) 17662 ((("%") reduce 472) ((")") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_PIPE) reduce 472) ((T_POW) shift) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("|") reduce 472)) 17663 ((("%") reduce 470) ((")") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_PIPE) reduce 470) ((T_POW) shift) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("|") reduce 470)) 17666 ((("%") reduce 519) (("(") reduce 519) ((")") reduce 519) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PIPE) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("|") reduce 519)) 17667 ((("%") reduce 475) ((")") reduce 475) (("*") reduce 475) (("+") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_PIPE) reduce 475) ((T_POW) shift) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("|") reduce 475)) 17668 ((("%") reduce 446) ((")") reduce 446) (("*") reduce 446) (("+") reduce 446) (("-") reduce 446) ((".") reduce 446) (("/") reduce 446) (("<") reduce 446) ((">") reduce 446) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_BOOLEAN_AND) reduce 446) ((T_BOOLEAN_OR) reduce 446) ((T_COALESCE) reduce 446) ((T_INSTANCEOF) reduce 446) ((T_IS_EQUAL) reduce 446) ((T_IS_GREATER_OR_EQUAL) reduce 446) ((T_IS_IDENTICAL) reduce 446) ((T_IS_NOT_EQUAL) reduce 446) ((T_IS_NOT_IDENTICAL) reduce 446) ((T_IS_SMALLER_OR_EQUAL) reduce 446) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_PIPE) reduce 446) ((T_POW) shift) ((T_SL) reduce 446) ((T_SPACESHIP) reduce 446) ((T_SR) reduce 446) (("^") reduce 446) (("|") reduce 446)) 17669 ((("%") reduce 445) ((")") reduce 445) (("*") reduce 445) (("+") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_PIPE) reduce 445) ((T_POW) shift) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("|") reduce 445)) 17671 ((("%") reduce 570) (("(") reduce 570) ((")") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) (("=") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AND_EQUAL) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_COALESCE_EQUAL) reduce 570) ((T_CONCAT_EQUAL) reduce 570) ((T_DEC) reduce 570) ((T_DIV_EQUAL) reduce 570) ((T_INC) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_MINUS_EQUAL) reduce 570) ((T_MOD_EQUAL) reduce 570) ((T_MUL_EQUAL) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_OR_EQUAL) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_PLUS_EQUAL) reduce 570) ((T_POW) reduce 570) ((T_POW_EQUAL) reduce 570) ((T_SL) reduce 570) ((T_SL_EQUAL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) ((T_SR_EQUAL) reduce 570) ((T_XOR_EQUAL) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 17674 ((("%") reduce 569) (("(") reduce 569) ((")") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) (("=") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AND_EQUAL) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_COALESCE_EQUAL) reduce 569) ((T_CONCAT_EQUAL) reduce 569) ((T_DEC) reduce 569) ((T_DIV_EQUAL) reduce 569) ((T_INC) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_MINUS_EQUAL) reduce 569) ((T_MOD_EQUAL) reduce 569) ((T_MUL_EQUAL) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_OR_EQUAL) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_PLUS_EQUAL) reduce 569) ((T_POW) reduce 569) ((T_POW_EQUAL) reduce 569) ((T_SL) reduce 569) ((T_SL_EQUAL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) ((T_SR_EQUAL) reduce 569) ((T_XOR_EQUAL) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 17676 ((("%") reduce 522) (("(") reduce 522) ((")") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_PIPE) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("|") reduce 522)) 17677 ((("%") reduce 447) ((")") reduce 447) (("*") reduce 447) (("+") reduce 447) (("-") reduce 447) ((".") reduce 447) (("/") reduce 447) (("<") reduce 447) ((">") reduce 447) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_BOOLEAN_AND) reduce 447) ((T_BOOLEAN_OR) reduce 447) ((T_COALESCE) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 447) ((T_IS_GREATER_OR_EQUAL) reduce 447) ((T_IS_IDENTICAL) reduce 447) ((T_IS_NOT_EQUAL) reduce 447) ((T_IS_NOT_IDENTICAL) reduce 447) ((T_IS_SMALLER_OR_EQUAL) reduce 447) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_PIPE) reduce 447) ((T_POW) shift) ((T_SL) reduce 447) ((T_SPACESHIP) reduce 447) ((T_SR) reduce 447) (("^") reduce 447) (("|") reduce 447)) 17679 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_ELSE) reduce 166) ((T_ELSEIF) reduce 166) ((T_EMPTY) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_VOID_CAST) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("}") reduce 166) (("~") reduce 166)) 17689 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_ELSE) reduce 160) ((T_ELSEIF) reduce 160) ((T_EMPTY) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_VOID_CAST) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("}") reduce 160) (("~") reduce 160)) 17698 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_ELSE) reduce 149) ((T_ELSEIF) reduce 149) ((T_EMPTY) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_VOID_CAST) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("}") reduce 149) (("~") reduce 149)) 17701 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_ELSE) reduce 148) ((T_ELSEIF) reduce 148) ((T_EMPTY) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_VOID_CAST) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("}") reduce 148) (("~") reduce 148)) 17703 ((("!") reduce 245) (("\"") reduce 245) (("$") reduce 245) (("(") reduce 245) (("+") reduce 245) (("-") reduce 245) ((";") reduce 245) (("@") reduce 245) ((T_ABSTRACT) reduce 245) ((T_ARRAY) reduce 245) ((T_ARRAY_CAST) reduce 245) ((T_ATTRIBUTE) reduce 245) ((T_BOOL_CAST) reduce 245) ((T_BREAK) reduce 245) ((T_CLASS) reduce 245) ((T_CLASS_C) reduce 245) ((T_CLONE) reduce 245) ((T_CONSTANT_ENCAPSED_STRING) reduce 245) ((T_CONTINUE) reduce 245) ((T_DEC) reduce 245) ((T_DECLARE) reduce 245) ((T_DIR) reduce 245) ((T_DNUMBER) reduce 245) ((T_DO) reduce 245) ((T_DOUBLE_CAST) reduce 245) ((T_ECHO) reduce 245) ((T_ELSE) reduce 245) ((T_ELSEIF) reduce 245) ((T_EMPTY) reduce 245) ((T_ENUM) reduce 245) ((T_EVAL) reduce 245) ((T_EXIT) reduce 245) ((T_FILE) reduce 245) ((T_FINAL) reduce 245) ((T_FN) reduce 245) ((T_FOR) reduce 245) ((T_FOREACH) reduce 245) ((T_FUNCTION) reduce 245) ((T_FUNC_C) reduce 245) ((T_GLOBAL) reduce 245) ((T_GOTO) reduce 245) ((T_HALT_COMPILER) reduce 245) ((T_IF) reduce 245) ((T_INC) reduce 245) ((T_INCLUDE) reduce 245) ((T_INCLUDE_ONCE) reduce 245) ((T_INLINE_HTML) reduce 245) ((T_INTERFACE) reduce 245) ((T_INT_CAST) reduce 245) ((T_ISSET) reduce 245) ((T_LINE) reduce 245) ((T_LIST) reduce 245) ((T_LNUMBER) reduce 245) ((T_MATCH) reduce 245) ((T_METHOD_C) reduce 245) ((T_NAME_FULLY_QUALIFIED) reduce 245) ((T_NAME_QUALIFIED) reduce 245) ((T_NAME_RELATIVE) reduce 245) ((T_NEW) reduce 245) ((T_NS_C) reduce 245) ((T_OBJECT_CAST) reduce 245) ((T_PRINT) reduce 245) ((T_PROPERTY_C) reduce 245) ((T_READONLY) reduce 245) ((T_REQUIRE) reduce 245) ((T_REQUIRE_ONCE) reduce 245) ((T_RETURN) reduce 245) ((T_START_HEREDOC) reduce 245) ((T_STATIC) reduce 245) ((T_STRING) reduce 245) ((T_STRING_CAST) reduce 245) ((T_SWITCH) reduce 245) ((T_THROW) reduce 245) ((T_TRAIT) reduce 245) ((T_TRAIT_C) reduce 245) ((T_TRY) reduce 245) ((T_UNSET) reduce 245) ((T_UNSET_CAST) reduce 245) ((T_VARIABLE) reduce 245) ((T_VOID_CAST) reduce 245) ((T_WHILE) reduce 245) ((T_YIELD) reduce 245) ((T_YIELD_FROM) reduce 245) (("[") reduce 245) (("`") reduce 245) (("{") reduce 245) (("}") reduce 245) (("~") reduce 245)) 17706 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_ELSE) reduce 147) ((T_ELSEIF) reduce 147) ((T_EMPTY) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_VOID_CAST) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("}") reduce 147) (("~") reduce 147)) 17708 ((("!") reduce 247) (("\"") reduce 247) (("$") reduce 247) (("(") reduce 247) (("+") reduce 247) (("-") reduce 247) ((";") reduce 247) (("@") reduce 247) ((T_ABSTRACT) reduce 247) ((T_ARRAY) reduce 247) ((T_ARRAY_CAST) reduce 247) ((T_ATTRIBUTE) reduce 247) ((T_BOOL_CAST) reduce 247) ((T_BREAK) reduce 247) ((T_CLASS) reduce 247) ((T_CLASS_C) reduce 247) ((T_CLONE) reduce 247) ((T_CONSTANT_ENCAPSED_STRING) reduce 247) ((T_CONTINUE) reduce 247) ((T_DEC) reduce 247) ((T_DECLARE) reduce 247) ((T_DIR) reduce 247) ((T_DNUMBER) reduce 247) ((T_DO) reduce 247) ((T_DOUBLE_CAST) reduce 247) ((T_ECHO) reduce 247) ((T_ELSE) reduce 247) ((T_ELSEIF) reduce 247) ((T_EMPTY) reduce 247) ((T_ENUM) reduce 247) ((T_EVAL) reduce 247) ((T_EXIT) reduce 247) ((T_FILE) reduce 247) ((T_FINAL) reduce 247) ((T_FN) reduce 247) ((T_FOR) reduce 247) ((T_FOREACH) reduce 247) ((T_FUNCTION) reduce 247) ((T_FUNC_C) reduce 247) ((T_GLOBAL) reduce 247) ((T_GOTO) reduce 247) ((T_HALT_COMPILER) reduce 247) ((T_IF) reduce 247) ((T_INC) reduce 247) ((T_INCLUDE) reduce 247) ((T_INCLUDE_ONCE) reduce 247) ((T_INLINE_HTML) reduce 247) ((T_INTERFACE) reduce 247) ((T_INT_CAST) reduce 247) ((T_ISSET) reduce 247) ((T_LINE) reduce 247) ((T_LIST) reduce 247) ((T_LNUMBER) reduce 247) ((T_MATCH) reduce 247) ((T_METHOD_C) reduce 247) ((T_NAME_FULLY_QUALIFIED) reduce 247) ((T_NAME_QUALIFIED) reduce 247) ((T_NAME_RELATIVE) reduce 247) ((T_NEW) reduce 247) ((T_NS_C) reduce 247) ((T_OBJECT_CAST) reduce 247) ((T_PRINT) reduce 247) ((T_PROPERTY_C) reduce 247) ((T_READONLY) reduce 247) ((T_REQUIRE) reduce 247) ((T_REQUIRE_ONCE) reduce 247) ((T_RETURN) reduce 247) ((T_START_HEREDOC) reduce 247) ((T_STATIC) reduce 247) ((T_STRING) reduce 247) ((T_STRING_CAST) reduce 247) ((T_SWITCH) reduce 247) ((T_THROW) reduce 247) ((T_TRAIT) reduce 247) ((T_TRAIT_C) reduce 247) ((T_TRY) reduce 247) ((T_UNSET) reduce 247) ((T_UNSET_CAST) reduce 247) ((T_VARIABLE) reduce 247) ((T_VOID_CAST) reduce 247) ((T_WHILE) reduce 247) ((T_YIELD) reduce 247) ((T_YIELD_FROM) reduce 247) (("[") reduce 247) (("`") reduce 247) (("{") reduce 247) (("}") reduce 247) (("~") reduce 247)) 17709 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_ELSE) reduce 161) ((T_ELSEIF) reduce 161) ((T_EMPTY) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_VOID_CAST) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("}") reduce 161) (("~") reduce 161)) 17713 ((("!") reduce 250) (("\"") reduce 250) (("$") reduce 250) (("(") reduce 250) (("+") reduce 250) (("-") reduce 250) ((";") reduce 250) (("@") reduce 250) ((T_ABSTRACT) reduce 250) ((T_ARRAY) reduce 250) ((T_ARRAY_CAST) reduce 250) ((T_ATTRIBUTE) reduce 250) ((T_BOOL_CAST) reduce 250) ((T_BREAK) reduce 250) ((T_CLASS) reduce 250) ((T_CLASS_C) reduce 250) ((T_CLONE) reduce 250) ((T_CONSTANT_ENCAPSED_STRING) reduce 250) ((T_CONTINUE) reduce 250) ((T_DEC) reduce 250) ((T_DECLARE) reduce 250) ((T_DIR) reduce 250) ((T_DNUMBER) reduce 250) ((T_DO) reduce 250) ((T_DOUBLE_CAST) reduce 250) ((T_ECHO) reduce 250) ((T_ELSE) reduce 250) ((T_ELSEIF) reduce 250) ((T_EMPTY) reduce 250) ((T_ENUM) reduce 250) ((T_EVAL) reduce 250) ((T_EXIT) reduce 250) ((T_FILE) reduce 250) ((T_FINAL) reduce 250) ((T_FN) reduce 250) ((T_FOR) reduce 250) ((T_FOREACH) reduce 250) ((T_FUNCTION) reduce 250) ((T_FUNC_C) reduce 250) ((T_GLOBAL) reduce 250) ((T_GOTO) reduce 250) ((T_HALT_COMPILER) reduce 250) ((T_IF) reduce 250) ((T_INC) reduce 250) ((T_INCLUDE) reduce 250) ((T_INCLUDE_ONCE) reduce 250) ((T_INLINE_HTML) reduce 250) ((T_INTERFACE) reduce 250) ((T_INT_CAST) reduce 250) ((T_ISSET) reduce 250) ((T_LINE) reduce 250) ((T_LIST) reduce 250) ((T_LNUMBER) reduce 250) ((T_MATCH) reduce 250) ((T_METHOD_C) reduce 250) ((T_NAME_FULLY_QUALIFIED) reduce 250) ((T_NAME_QUALIFIED) reduce 250) ((T_NAME_RELATIVE) reduce 250) ((T_NEW) reduce 250) ((T_NS_C) reduce 250) ((T_OBJECT_CAST) reduce 250) ((T_PRINT) reduce 250) ((T_PROPERTY_C) reduce 250) ((T_READONLY) reduce 250) ((T_REQUIRE) reduce 250) ((T_REQUIRE_ONCE) reduce 250) ((T_RETURN) reduce 250) ((T_START_HEREDOC) reduce 250) ((T_STATIC) reduce 250) ((T_STRING) reduce 250) ((T_STRING_CAST) reduce 250) ((T_SWITCH) reduce 250) ((T_THROW) reduce 250) ((T_TRAIT) reduce 250) ((T_TRAIT_C) reduce 250) ((T_TRY) reduce 250) ((T_UNSET) reduce 250) ((T_UNSET_CAST) reduce 250) ((T_VARIABLE) reduce 250) ((T_VOID_CAST) reduce 250) ((T_WHILE) reduce 250) ((T_YIELD) reduce 250) ((T_YIELD_FROM) reduce 250) (("[") reduce 250) (("`") reduce 250) (("{") reduce 250) (("}") reduce 250) (("~") reduce 250)) 17717 ((("!") reduce 142) (("\"") reduce 142) (("$") reduce 142) (("(") reduce 142) (("+") reduce 142) (("-") reduce 142) ((";") reduce 142) (("@") reduce 142) ((T_ABSTRACT) reduce 142) ((T_ARRAY) reduce 142) ((T_ARRAY_CAST) reduce 142) ((T_ATTRIBUTE) reduce 142) ((T_BOOL_CAST) reduce 142) ((T_BREAK) reduce 142) ((T_CLASS) reduce 142) ((T_CLASS_C) reduce 142) ((T_CLONE) reduce 142) ((T_CONSTANT_ENCAPSED_STRING) reduce 142) ((T_CONTINUE) reduce 142) ((T_DEC) reduce 142) ((T_DECLARE) reduce 142) ((T_DIR) reduce 142) ((T_DNUMBER) reduce 142) ((T_DO) reduce 142) ((T_DOUBLE_CAST) reduce 142) ((T_ECHO) reduce 142) ((T_ELSE) reduce 142) ((T_ELSEIF) reduce 142) ((T_EMPTY) reduce 142) ((T_ENDIF) reduce 142) ((T_ENUM) reduce 142) ((T_EVAL) reduce 142) ((T_EXIT) reduce 142) ((T_FILE) reduce 142) ((T_FINAL) reduce 142) ((T_FN) reduce 142) ((T_FOR) reduce 142) ((T_FOREACH) reduce 142) ((T_FUNCTION) reduce 142) ((T_FUNC_C) reduce 142) ((T_GLOBAL) reduce 142) ((T_GOTO) reduce 142) ((T_HALT_COMPILER) reduce 142) ((T_IF) reduce 142) ((T_INC) reduce 142) ((T_INCLUDE) reduce 142) ((T_INCLUDE_ONCE) reduce 142) ((T_INLINE_HTML) reduce 142) ((T_INTERFACE) reduce 142) ((T_INT_CAST) reduce 142) ((T_ISSET) reduce 142) ((T_LINE) reduce 142) ((T_LIST) reduce 142) ((T_LNUMBER) reduce 142) ((T_MATCH) reduce 142) ((T_METHOD_C) reduce 142) ((T_NAME_FULLY_QUALIFIED) reduce 142) ((T_NAME_QUALIFIED) reduce 142) ((T_NAME_RELATIVE) reduce 142) ((T_NEW) reduce 142) ((T_NS_C) reduce 142) ((T_OBJECT_CAST) reduce 142) ((T_PRINT) reduce 142) ((T_PROPERTY_C) reduce 142) ((T_READONLY) reduce 142) ((T_REQUIRE) reduce 142) ((T_REQUIRE_ONCE) reduce 142) ((T_RETURN) reduce 142) ((T_START_HEREDOC) reduce 142) ((T_STATIC) reduce 142) ((T_STRING) reduce 142) ((T_STRING_CAST) reduce 142) ((T_SWITCH) reduce 142) ((T_THROW) reduce 142) ((T_TRAIT) reduce 142) ((T_TRAIT_C) reduce 142) ((T_TRY) reduce 142) ((T_UNSET) reduce 142) ((T_UNSET_CAST) reduce 142) ((T_VARIABLE) reduce 142) ((T_VOID_CAST) reduce 142) ((T_WHILE) reduce 142) ((T_YIELD) reduce 142) ((T_YIELD_FROM) reduce 142) (("[") reduce 142) (("`") reduce 142) (("{") reduce 142) (("~") reduce 142)) 17718 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_ELSE) reduce 249) ((T_ELSEIF) reduce 249) ((T_EMPTY) shift) ((T_ENDIF) reduce 249) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_VOID_CAST) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("~") shift)) 17719 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_ELSE) reduce 166) ((T_ELSEIF) reduce 166) ((T_EMPTY) reduce 166) ((T_ENDIF) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_VOID_CAST) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 17733 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_ELSE) reduce 160) ((T_ELSEIF) reduce 160) ((T_EMPTY) reduce 160) ((T_ENDIF) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_VOID_CAST) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 17744 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_ELSE) reduce 149) ((T_ELSEIF) reduce 149) ((T_EMPTY) reduce 149) ((T_ENDIF) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_VOID_CAST) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 17746 ((("!") reduce 144) (("\"") reduce 144) (("$") reduce 144) (("(") reduce 144) (("+") reduce 144) (("-") reduce 144) ((";") reduce 144) (("@") reduce 144) ((T_ABSTRACT) reduce 144) ((T_ARRAY) reduce 144) ((T_ARRAY_CAST) reduce 144) ((T_ATTRIBUTE) reduce 144) ((T_BOOL_CAST) reduce 144) ((T_BREAK) reduce 144) ((T_CLASS) reduce 144) ((T_CLASS_C) reduce 144) ((T_CLONE) reduce 144) ((T_CONSTANT_ENCAPSED_STRING) reduce 144) ((T_CONTINUE) reduce 144) ((T_DEC) reduce 144) ((T_DECLARE) reduce 144) ((T_DIR) reduce 144) ((T_DNUMBER) reduce 144) ((T_DO) reduce 144) ((T_DOUBLE_CAST) reduce 144) ((T_ECHO) reduce 144) ((T_ELSE) reduce 144) ((T_ELSEIF) reduce 144) ((T_EMPTY) reduce 144) ((T_ENDIF) reduce 144) ((T_ENUM) reduce 144) ((T_EVAL) reduce 144) ((T_EXIT) reduce 144) ((T_FILE) reduce 144) ((T_FINAL) reduce 144) ((T_FN) reduce 144) ((T_FOR) reduce 144) ((T_FOREACH) reduce 144) ((T_FUNCTION) reduce 144) ((T_FUNC_C) reduce 144) ((T_GLOBAL) reduce 144) ((T_GOTO) reduce 144) ((T_HALT_COMPILER) reduce 144) ((T_IF) reduce 144) ((T_INC) reduce 144) ((T_INCLUDE) reduce 144) ((T_INCLUDE_ONCE) reduce 144) ((T_INLINE_HTML) reduce 144) ((T_INTERFACE) reduce 144) ((T_INT_CAST) reduce 144) ((T_ISSET) reduce 144) ((T_LINE) reduce 144) ((T_LIST) reduce 144) ((T_LNUMBER) reduce 144) ((T_MATCH) reduce 144) ((T_METHOD_C) reduce 144) ((T_NAME_FULLY_QUALIFIED) reduce 144) ((T_NAME_QUALIFIED) reduce 144) ((T_NAME_RELATIVE) reduce 144) ((T_NEW) reduce 144) ((T_NS_C) reduce 144) ((T_OBJECT_CAST) reduce 144) ((T_PRINT) reduce 144) ((T_PROPERTY_C) reduce 144) ((T_READONLY) reduce 144) ((T_REQUIRE) reduce 144) ((T_REQUIRE_ONCE) reduce 144) ((T_RETURN) reduce 144) ((T_START_HEREDOC) reduce 144) ((T_STATIC) reduce 144) ((T_STRING) reduce 144) ((T_STRING_CAST) reduce 144) ((T_SWITCH) reduce 144) ((T_THROW) reduce 144) ((T_TRAIT) reduce 144) ((T_TRAIT_C) reduce 144) ((T_TRY) reduce 144) ((T_UNSET) reduce 144) ((T_UNSET_CAST) reduce 144) ((T_VARIABLE) reduce 144) ((T_VOID_CAST) reduce 144) ((T_WHILE) reduce 144) ((T_YIELD) reduce 144) ((T_YIELD_FROM) reduce 144) (("[") reduce 144) (("`") reduce 144) (("{") reduce 144) (("~") reduce 144)) 17748 ((("!") reduce 104) (("\"") reduce 104) (("$") reduce 104) (("(") reduce 104) (("+") reduce 104) (("-") reduce 104) ((";") reduce 104) (("@") reduce 104) ((T_ABSTRACT) reduce 104) ((T_ARRAY) reduce 104) ((T_ARRAY_CAST) reduce 104) ((T_ATTRIBUTE) reduce 104) ((T_BOOL_CAST) reduce 104) ((T_BREAK) reduce 104) ((T_CLASS) reduce 104) ((T_CLASS_C) reduce 104) ((T_CLONE) reduce 104) ((T_CONSTANT_ENCAPSED_STRING) reduce 104) ((T_CONTINUE) reduce 104) ((T_DEC) reduce 104) ((T_DECLARE) reduce 104) ((T_DIR) reduce 104) ((T_DNUMBER) reduce 104) ((T_DO) reduce 104) ((T_DOUBLE_CAST) reduce 104) ((T_ECHO) reduce 104) ((T_ELSE) reduce 104) ((T_ELSEIF) reduce 104) ((T_EMPTY) reduce 104) ((T_ENDIF) reduce 104) ((T_ENUM) reduce 104) ((T_EVAL) reduce 104) ((T_EXIT) reduce 104) ((T_FILE) reduce 104) ((T_FINAL) reduce 104) ((T_FN) reduce 104) ((T_FOR) reduce 104) ((T_FOREACH) reduce 104) ((T_FUNCTION) reduce 104) ((T_FUNC_C) reduce 104) ((T_GLOBAL) reduce 104) ((T_GOTO) reduce 104) ((T_HALT_COMPILER) reduce 104) ((T_IF) reduce 104) ((T_INC) reduce 104) ((T_INCLUDE) reduce 104) ((T_INCLUDE_ONCE) reduce 104) ((T_INLINE_HTML) reduce 104) ((T_INTERFACE) reduce 104) ((T_INT_CAST) reduce 104) ((T_ISSET) reduce 104) ((T_LINE) reduce 104) ((T_LIST) reduce 104) ((T_LNUMBER) reduce 104) ((T_MATCH) reduce 104) ((T_METHOD_C) reduce 104) ((T_NAME_FULLY_QUALIFIED) reduce 104) ((T_NAME_QUALIFIED) reduce 104) ((T_NAME_RELATIVE) reduce 104) ((T_NEW) reduce 104) ((T_NS_C) reduce 104) ((T_OBJECT_CAST) reduce 104) ((T_PRINT) reduce 104) ((T_PROPERTY_C) reduce 104) ((T_READONLY) reduce 104) ((T_REQUIRE) reduce 104) ((T_REQUIRE_ONCE) reduce 104) ((T_RETURN) reduce 104) ((T_START_HEREDOC) reduce 104) ((T_STATIC) reduce 104) ((T_STRING) reduce 104) ((T_STRING_CAST) reduce 104) ((T_SWITCH) reduce 104) ((T_THROW) reduce 104) ((T_TRAIT) reduce 104) ((T_TRAIT_C) reduce 104) ((T_TRY) reduce 104) ((T_UNSET) reduce 104) ((T_UNSET_CAST) reduce 104) ((T_VARIABLE) reduce 104) ((T_VOID_CAST) reduce 104) ((T_WHILE) reduce 104) ((T_YIELD) reduce 104) ((T_YIELD_FROM) reduce 104) (("[") reduce 104) (("`") reduce 104) (("{") reduce 104) (("~") reduce 104)) 17750 ((("!") reduce 107) (("\"") reduce 107) (("$") reduce 107) (("(") reduce 107) (("+") reduce 107) (("-") reduce 107) ((";") reduce 107) (("@") reduce 107) ((T_ABSTRACT) reduce 107) ((T_ARRAY) reduce 107) ((T_ARRAY_CAST) reduce 107) ((T_ATTRIBUTE) reduce 107) ((T_BOOL_CAST) reduce 107) ((T_BREAK) reduce 107) ((T_CLASS) reduce 107) ((T_CLASS_C) reduce 107) ((T_CLONE) reduce 107) ((T_CONSTANT_ENCAPSED_STRING) reduce 107) ((T_CONTINUE) reduce 107) ((T_DEC) reduce 107) ((T_DECLARE) reduce 107) ((T_DIR) reduce 107) ((T_DNUMBER) reduce 107) ((T_DO) reduce 107) ((T_DOUBLE_CAST) reduce 107) ((T_ECHO) reduce 107) ((T_ELSE) reduce 107) ((T_ELSEIF) reduce 107) ((T_EMPTY) reduce 107) ((T_ENDIF) reduce 107) ((T_ENUM) reduce 107) ((T_EVAL) reduce 107) ((T_EXIT) reduce 107) ((T_FILE) reduce 107) ((T_FINAL) reduce 107) ((T_FN) reduce 107) ((T_FOR) reduce 107) ((T_FOREACH) reduce 107) ((T_FUNCTION) reduce 107) ((T_FUNC_C) reduce 107) ((T_GLOBAL) reduce 107) ((T_GOTO) reduce 107) ((T_HALT_COMPILER) reduce 107) ((T_IF) reduce 107) ((T_INC) reduce 107) ((T_INCLUDE) reduce 107) ((T_INCLUDE_ONCE) reduce 107) ((T_INLINE_HTML) reduce 107) ((T_INTERFACE) reduce 107) ((T_INT_CAST) reduce 107) ((T_ISSET) reduce 107) ((T_LINE) reduce 107) ((T_LIST) reduce 107) ((T_LNUMBER) reduce 107) ((T_MATCH) reduce 107) ((T_METHOD_C) reduce 107) ((T_NAME_FULLY_QUALIFIED) reduce 107) ((T_NAME_QUALIFIED) reduce 107) ((T_NAME_RELATIVE) reduce 107) ((T_NEW) reduce 107) ((T_NS_C) reduce 107) ((T_OBJECT_CAST) reduce 107) ((T_PRINT) reduce 107) ((T_PROPERTY_C) reduce 107) ((T_READONLY) reduce 107) ((T_REQUIRE) reduce 107) ((T_REQUIRE_ONCE) reduce 107) ((T_RETURN) reduce 107) ((T_START_HEREDOC) reduce 107) ((T_STATIC) reduce 107) ((T_STRING) reduce 107) ((T_STRING_CAST) reduce 107) ((T_SWITCH) reduce 107) ((T_THROW) reduce 107) ((T_TRAIT) reduce 107) ((T_TRAIT_C) reduce 107) ((T_TRY) reduce 107) ((T_UNSET) reduce 107) ((T_UNSET_CAST) reduce 107) ((T_VARIABLE) reduce 107) ((T_VOID_CAST) reduce 107) ((T_WHILE) reduce 107) ((T_YIELD) reduce 107) ((T_YIELD_FROM) reduce 107) (("[") reduce 107) (("`") reduce 107) (("{") reduce 107) (("~") reduce 107)) 17753 ((("!") reduce 103) (("\"") reduce 103) (("$") reduce 103) (("(") reduce 103) (("+") reduce 103) (("-") reduce 103) ((";") reduce 103) (("@") reduce 103) ((T_ABSTRACT) reduce 103) ((T_ARRAY) reduce 103) ((T_ARRAY_CAST) reduce 103) ((T_ATTRIBUTE) reduce 103) ((T_BOOL_CAST) reduce 103) ((T_BREAK) reduce 103) ((T_CLASS) reduce 103) ((T_CLASS_C) reduce 103) ((T_CLONE) reduce 103) ((T_CONSTANT_ENCAPSED_STRING) reduce 103) ((T_CONTINUE) reduce 103) ((T_DEC) reduce 103) ((T_DECLARE) reduce 103) ((T_DIR) reduce 103) ((T_DNUMBER) reduce 103) ((T_DO) reduce 103) ((T_DOUBLE_CAST) reduce 103) ((T_ECHO) reduce 103) ((T_ELSE) reduce 103) ((T_ELSEIF) reduce 103) ((T_EMPTY) reduce 103) ((T_ENDIF) reduce 103) ((T_ENUM) reduce 103) ((T_EVAL) reduce 103) ((T_EXIT) reduce 103) ((T_FILE) reduce 103) ((T_FINAL) reduce 103) ((T_FN) reduce 103) ((T_FOR) reduce 103) ((T_FOREACH) reduce 103) ((T_FUNCTION) reduce 103) ((T_FUNC_C) reduce 103) ((T_GLOBAL) reduce 103) ((T_GOTO) reduce 103) ((T_HALT_COMPILER) reduce 103) ((T_IF) reduce 103) ((T_INC) reduce 103) ((T_INCLUDE) reduce 103) ((T_INCLUDE_ONCE) reduce 103) ((T_INLINE_HTML) reduce 103) ((T_INTERFACE) reduce 103) ((T_INT_CAST) reduce 103) ((T_ISSET) reduce 103) ((T_LINE) reduce 103) ((T_LIST) reduce 103) ((T_LNUMBER) reduce 103) ((T_MATCH) reduce 103) ((T_METHOD_C) reduce 103) ((T_NAME_FULLY_QUALIFIED) reduce 103) ((T_NAME_QUALIFIED) reduce 103) ((T_NAME_RELATIVE) reduce 103) ((T_NEW) reduce 103) ((T_NS_C) reduce 103) ((T_OBJECT_CAST) reduce 103) ((T_PRINT) reduce 103) ((T_PROPERTY_C) reduce 103) ((T_READONLY) reduce 103) ((T_REQUIRE) reduce 103) ((T_REQUIRE_ONCE) reduce 103) ((T_RETURN) reduce 103) ((T_START_HEREDOC) reduce 103) ((T_STATIC) reduce 103) ((T_STRING) reduce 103) ((T_STRING_CAST) reduce 103) ((T_SWITCH) reduce 103) ((T_THROW) reduce 103) ((T_TRAIT) reduce 103) ((T_TRAIT_C) reduce 103) ((T_TRY) reduce 103) ((T_UNSET) reduce 103) ((T_UNSET_CAST) reduce 103) ((T_VARIABLE) reduce 103) ((T_VOID_CAST) reduce 103) ((T_WHILE) reduce 103) ((T_YIELD) reduce 103) ((T_YIELD_FROM) reduce 103) (("[") reduce 103) (("`") reduce 103) (("{") reduce 103) (("~") reduce 103)) 17754 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_ELSE) reduce 148) ((T_ELSEIF) reduce 148) ((T_EMPTY) reduce 148) ((T_ENDIF) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_VOID_CAST) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 17755 ((("!") reduce 246) (("\"") reduce 246) (("$") reduce 246) (("(") reduce 246) (("+") reduce 246) (("-") reduce 246) ((";") reduce 246) (("@") reduce 246) ((T_ABSTRACT) reduce 246) ((T_ARRAY) reduce 246) ((T_ARRAY_CAST) reduce 246) ((T_ATTRIBUTE) reduce 246) ((T_BOOL_CAST) reduce 246) ((T_BREAK) reduce 246) ((T_CLASS) reduce 246) ((T_CLASS_C) reduce 246) ((T_CLONE) reduce 246) ((T_CONSTANT_ENCAPSED_STRING) reduce 246) ((T_CONTINUE) reduce 246) ((T_DEC) reduce 246) ((T_DECLARE) reduce 246) ((T_DIR) reduce 246) ((T_DNUMBER) reduce 246) ((T_DO) reduce 246) ((T_DOUBLE_CAST) reduce 246) ((T_ECHO) reduce 246) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) reduce 246) ((T_ENDIF) reduce 246) ((T_ENUM) reduce 246) ((T_EVAL) reduce 246) ((T_EXIT) reduce 246) ((T_FILE) reduce 246) ((T_FINAL) reduce 246) ((T_FN) reduce 246) ((T_FOR) reduce 246) ((T_FOREACH) reduce 246) ((T_FUNCTION) reduce 246) ((T_FUNC_C) reduce 246) ((T_GLOBAL) reduce 246) ((T_GOTO) reduce 246) ((T_HALT_COMPILER) reduce 246) ((T_IF) reduce 246) ((T_INC) reduce 246) ((T_INCLUDE) reduce 246) ((T_INCLUDE_ONCE) reduce 246) ((T_INLINE_HTML) reduce 246) ((T_INTERFACE) reduce 246) ((T_INT_CAST) reduce 246) ((T_ISSET) reduce 246) ((T_LINE) reduce 246) ((T_LIST) reduce 246) ((T_LNUMBER) reduce 246) ((T_MATCH) reduce 246) ((T_METHOD_C) reduce 246) ((T_NAME_FULLY_QUALIFIED) reduce 246) ((T_NAME_QUALIFIED) reduce 246) ((T_NAME_RELATIVE) reduce 246) ((T_NEW) reduce 246) ((T_NS_C) reduce 246) ((T_OBJECT_CAST) reduce 246) ((T_PRINT) reduce 246) ((T_PROPERTY_C) reduce 246) ((T_READONLY) reduce 246) ((T_REQUIRE) reduce 246) ((T_REQUIRE_ONCE) reduce 246) ((T_RETURN) reduce 246) ((T_START_HEREDOC) reduce 246) ((T_STATIC) reduce 246) ((T_STRING) reduce 246) ((T_STRING_CAST) reduce 246) ((T_SWITCH) reduce 246) ((T_THROW) reduce 246) ((T_TRAIT) reduce 246) ((T_TRAIT_C) reduce 246) ((T_TRY) reduce 246) ((T_UNSET) reduce 246) ((T_UNSET_CAST) reduce 246) ((T_VARIABLE) reduce 246) ((T_VOID_CAST) reduce 246) ((T_WHILE) reduce 246) ((T_YIELD) reduce 246) ((T_YIELD_FROM) reduce 246) (("[") reduce 246) (("`") reduce 246) (("{") reduce 246) (("~") reduce 246)) 17756 ((("!") reduce 141) (("\"") reduce 141) (("$") reduce 141) (("(") reduce 141) (("+") reduce 141) (("-") reduce 141) ((";") reduce 141) (("@") reduce 141) ((T_ABSTRACT) reduce 141) ((T_ARRAY) reduce 141) ((T_ARRAY_CAST) reduce 141) ((T_ATTRIBUTE) reduce 141) ((T_BOOL_CAST) reduce 141) ((T_BREAK) reduce 141) ((T_CLASS) reduce 141) ((T_CLASS_C) reduce 141) ((T_CLONE) reduce 141) ((T_CONSTANT_ENCAPSED_STRING) reduce 141) ((T_CONTINUE) reduce 141) ((T_DEC) reduce 141) ((T_DECLARE) reduce 141) ((T_DIR) reduce 141) ((T_DNUMBER) reduce 141) ((T_DO) reduce 141) ((T_DOUBLE_CAST) reduce 141) ((T_ECHO) reduce 141) ((T_ELSE) reduce 141) ((T_ELSEIF) reduce 141) ((T_EMPTY) reduce 141) ((T_ENDIF) reduce 141) ((T_ENUM) reduce 141) ((T_EVAL) reduce 141) ((T_EXIT) reduce 141) ((T_FILE) reduce 141) ((T_FINAL) reduce 141) ((T_FN) reduce 141) ((T_FOR) reduce 141) ((T_FOREACH) reduce 141) ((T_FUNCTION) reduce 141) ((T_FUNC_C) reduce 141) ((T_GLOBAL) reduce 141) ((T_GOTO) reduce 141) ((T_HALT_COMPILER) reduce 141) ((T_IF) reduce 141) ((T_INC) reduce 141) ((T_INCLUDE) reduce 141) ((T_INCLUDE_ONCE) reduce 141) ((T_INLINE_HTML) reduce 141) ((T_INTERFACE) reduce 141) ((T_INT_CAST) reduce 141) ((T_ISSET) reduce 141) ((T_LINE) reduce 141) ((T_LIST) reduce 141) ((T_LNUMBER) reduce 141) ((T_MATCH) reduce 141) ((T_METHOD_C) reduce 141) ((T_NAME_FULLY_QUALIFIED) reduce 141) ((T_NAME_QUALIFIED) reduce 141) ((T_NAME_RELATIVE) reduce 141) ((T_NEW) reduce 141) ((T_NS_C) reduce 141) ((T_OBJECT_CAST) reduce 141) ((T_PRINT) reduce 141) ((T_PROPERTY_C) reduce 141) ((T_READONLY) reduce 141) ((T_REQUIRE) reduce 141) ((T_REQUIRE_ONCE) reduce 141) ((T_RETURN) reduce 141) ((T_START_HEREDOC) reduce 141) ((T_STATIC) reduce 141) ((T_STRING) reduce 141) ((T_STRING_CAST) reduce 141) ((T_SWITCH) reduce 141) ((T_THROW) reduce 141) ((T_TRAIT) reduce 141) ((T_TRAIT_C) reduce 141) ((T_TRY) reduce 141) ((T_UNSET) reduce 141) ((T_UNSET_CAST) reduce 141) ((T_VARIABLE) reduce 141) ((T_VOID_CAST) reduce 141) ((T_WHILE) reduce 141) ((T_YIELD) reduce 141) ((T_YIELD_FROM) reduce 141) (("[") reduce 141) (("`") reduce 141) (("{") reduce 141) (("~") reduce 141)) 17757 ((("!") reduce 106) (("\"") reduce 106) (("$") reduce 106) (("(") reduce 106) (("+") reduce 106) (("-") reduce 106) ((";") reduce 106) (("@") reduce 106) ((T_ABSTRACT) reduce 106) ((T_ARRAY) reduce 106) ((T_ARRAY_CAST) reduce 106) ((T_ATTRIBUTE) reduce 106) ((T_BOOL_CAST) reduce 106) ((T_BREAK) reduce 106) ((T_CLASS) reduce 106) ((T_CLASS_C) reduce 106) ((T_CLONE) reduce 106) ((T_CONSTANT_ENCAPSED_STRING) reduce 106) ((T_CONTINUE) reduce 106) ((T_DEC) reduce 106) ((T_DECLARE) reduce 106) ((T_DIR) reduce 106) ((T_DNUMBER) reduce 106) ((T_DO) reduce 106) ((T_DOUBLE_CAST) reduce 106) ((T_ECHO) reduce 106) ((T_ELSE) reduce 106) ((T_ELSEIF) reduce 106) ((T_EMPTY) reduce 106) ((T_ENDIF) reduce 106) ((T_ENUM) reduce 106) ((T_EVAL) reduce 106) ((T_EXIT) reduce 106) ((T_FILE) reduce 106) ((T_FINAL) reduce 106) ((T_FN) reduce 106) ((T_FOR) reduce 106) ((T_FOREACH) reduce 106) ((T_FUNCTION) reduce 106) ((T_FUNC_C) reduce 106) ((T_GLOBAL) reduce 106) ((T_GOTO) reduce 106) ((T_HALT_COMPILER) reduce 106) ((T_IF) reduce 106) ((T_INC) reduce 106) ((T_INCLUDE) reduce 106) ((T_INCLUDE_ONCE) reduce 106) ((T_INLINE_HTML) reduce 106) ((T_INTERFACE) reduce 106) ((T_INT_CAST) reduce 106) ((T_ISSET) reduce 106) ((T_LINE) reduce 106) ((T_LIST) reduce 106) ((T_LNUMBER) reduce 106) ((T_MATCH) reduce 106) ((T_METHOD_C) reduce 106) ((T_NAME_FULLY_QUALIFIED) reduce 106) ((T_NAME_QUALIFIED) reduce 106) ((T_NAME_RELATIVE) reduce 106) ((T_NEW) reduce 106) ((T_NS_C) reduce 106) ((T_OBJECT_CAST) reduce 106) ((T_PRINT) reduce 106) ((T_PROPERTY_C) reduce 106) ((T_READONLY) reduce 106) ((T_REQUIRE) reduce 106) ((T_REQUIRE_ONCE) reduce 106) ((T_RETURN) reduce 106) ((T_START_HEREDOC) reduce 106) ((T_STATIC) reduce 106) ((T_STRING) reduce 106) ((T_STRING_CAST) reduce 106) ((T_SWITCH) reduce 106) ((T_THROW) reduce 106) ((T_TRAIT) reduce 106) ((T_TRAIT_C) reduce 106) ((T_TRY) reduce 106) ((T_UNSET) reduce 106) ((T_UNSET_CAST) reduce 106) ((T_VARIABLE) reduce 106) ((T_VOID_CAST) reduce 106) ((T_WHILE) reduce 106) ((T_YIELD) reduce 106) ((T_YIELD_FROM) reduce 106) (("[") reduce 106) (("`") reduce 106) (("{") reduce 106) (("~") reduce 106)) 17758 ((("!") reduce 143) (("\"") reduce 143) (("$") reduce 143) (("(") reduce 143) (("+") reduce 143) (("-") reduce 143) ((";") reduce 143) (("@") reduce 143) ((T_ABSTRACT) reduce 143) ((T_ARRAY) reduce 143) ((T_ARRAY_CAST) reduce 143) ((T_ATTRIBUTE) reduce 143) ((T_BOOL_CAST) reduce 143) ((T_BREAK) reduce 143) ((T_CLASS) reduce 143) ((T_CLASS_C) reduce 143) ((T_CLONE) reduce 143) ((T_CONSTANT_ENCAPSED_STRING) reduce 143) ((T_CONTINUE) reduce 143) ((T_DEC) reduce 143) ((T_DECLARE) reduce 143) ((T_DIR) reduce 143) ((T_DNUMBER) reduce 143) ((T_DO) reduce 143) ((T_DOUBLE_CAST) reduce 143) ((T_ECHO) reduce 143) ((T_ELSE) reduce 143) ((T_ELSEIF) reduce 143) ((T_EMPTY) reduce 143) ((T_ENDIF) reduce 143) ((T_ENUM) reduce 143) ((T_EVAL) reduce 143) ((T_EXIT) reduce 143) ((T_FILE) reduce 143) ((T_FINAL) reduce 143) ((T_FN) reduce 143) ((T_FOR) reduce 143) ((T_FOREACH) reduce 143) ((T_FUNCTION) reduce 143) ((T_FUNC_C) reduce 143) ((T_GLOBAL) reduce 143) ((T_GOTO) reduce 143) ((T_HALT_COMPILER) reduce 143) ((T_IF) reduce 143) ((T_INC) reduce 143) ((T_INCLUDE) reduce 143) ((T_INCLUDE_ONCE) reduce 143) ((T_INLINE_HTML) reduce 143) ((T_INTERFACE) reduce 143) ((T_INT_CAST) reduce 143) ((T_ISSET) reduce 143) ((T_LINE) reduce 143) ((T_LIST) reduce 143) ((T_LNUMBER) reduce 143) ((T_MATCH) reduce 143) ((T_METHOD_C) reduce 143) ((T_NAME_FULLY_QUALIFIED) reduce 143) ((T_NAME_QUALIFIED) reduce 143) ((T_NAME_RELATIVE) reduce 143) ((T_NEW) reduce 143) ((T_NS_C) reduce 143) ((T_OBJECT_CAST) reduce 143) ((T_PRINT) reduce 143) ((T_PROPERTY_C) reduce 143) ((T_READONLY) reduce 143) ((T_REQUIRE) reduce 143) ((T_REQUIRE_ONCE) reduce 143) ((T_RETURN) reduce 143) ((T_START_HEREDOC) reduce 143) ((T_STATIC) reduce 143) ((T_STRING) reduce 143) ((T_STRING_CAST) reduce 143) ((T_SWITCH) reduce 143) ((T_THROW) reduce 143) ((T_TRAIT) reduce 143) ((T_TRAIT_C) reduce 143) ((T_TRY) reduce 143) ((T_UNSET) reduce 143) ((T_UNSET_CAST) reduce 143) ((T_VARIABLE) reduce 143) ((T_VOID_CAST) reduce 143) ((T_WHILE) reduce 143) ((T_YIELD) reduce 143) ((T_YIELD_FROM) reduce 143) (("[") reduce 143) (("`") reduce 143) (("{") reduce 143) (("~") reduce 143)) 17759 ((("!") reduce 105) (("\"") reduce 105) (("$") reduce 105) (("(") reduce 105) (("+") reduce 105) (("-") reduce 105) ((";") reduce 105) (("@") reduce 105) ((T_ABSTRACT) reduce 105) ((T_ARRAY) reduce 105) ((T_ARRAY_CAST) reduce 105) ((T_ATTRIBUTE) reduce 105) ((T_BOOL_CAST) reduce 105) ((T_BREAK) reduce 105) ((T_CLASS) reduce 105) ((T_CLASS_C) reduce 105) ((T_CLONE) reduce 105) ((T_CONSTANT_ENCAPSED_STRING) reduce 105) ((T_CONTINUE) reduce 105) ((T_DEC) reduce 105) ((T_DECLARE) reduce 105) ((T_DIR) reduce 105) ((T_DNUMBER) reduce 105) ((T_DO) reduce 105) ((T_DOUBLE_CAST) reduce 105) ((T_ECHO) reduce 105) ((T_ELSE) reduce 105) ((T_ELSEIF) reduce 105) ((T_EMPTY) reduce 105) ((T_ENDIF) reduce 105) ((T_ENUM) reduce 105) ((T_EVAL) reduce 105) ((T_EXIT) reduce 105) ((T_FILE) reduce 105) ((T_FINAL) reduce 105) ((T_FN) reduce 105) ((T_FOR) reduce 105) ((T_FOREACH) reduce 105) ((T_FUNCTION) reduce 105) ((T_FUNC_C) reduce 105) ((T_GLOBAL) reduce 105) ((T_GOTO) reduce 105) ((T_HALT_COMPILER) reduce 105) ((T_IF) reduce 105) ((T_INC) reduce 105) ((T_INCLUDE) reduce 105) ((T_INCLUDE_ONCE) reduce 105) ((T_INLINE_HTML) reduce 105) ((T_INTERFACE) reduce 105) ((T_INT_CAST) reduce 105) ((T_ISSET) reduce 105) ((T_LINE) reduce 105) ((T_LIST) reduce 105) ((T_LNUMBER) reduce 105) ((T_MATCH) reduce 105) ((T_METHOD_C) reduce 105) ((T_NAME_FULLY_QUALIFIED) reduce 105) ((T_NAME_QUALIFIED) reduce 105) ((T_NAME_RELATIVE) reduce 105) ((T_NEW) reduce 105) ((T_NS_C) reduce 105) ((T_OBJECT_CAST) reduce 105) ((T_PRINT) reduce 105) ((T_PROPERTY_C) reduce 105) ((T_READONLY) reduce 105) ((T_REQUIRE) reduce 105) ((T_REQUIRE_ONCE) reduce 105) ((T_RETURN) reduce 105) ((T_START_HEREDOC) reduce 105) ((T_STATIC) reduce 105) ((T_STRING) reduce 105) ((T_STRING_CAST) reduce 105) ((T_SWITCH) reduce 105) ((T_THROW) reduce 105) ((T_TRAIT) reduce 105) ((T_TRAIT_C) reduce 105) ((T_TRY) reduce 105) ((T_UNSET) reduce 105) ((T_UNSET_CAST) reduce 105) ((T_VARIABLE) reduce 105) ((T_VOID_CAST) reduce 105) ((T_WHILE) reduce 105) ((T_YIELD) reduce 105) ((T_YIELD_FROM) reduce 105) (("[") reduce 105) (("`") reduce 105) (("{") reduce 105) (("~") reduce 105)) 17762 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_ELSE) reduce 147) ((T_ELSEIF) reduce 147) ((T_EMPTY) reduce 147) ((T_ENDIF) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_VOID_CAST) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 17768 ((("!") reduce 245) (("\"") reduce 245) (("$") reduce 245) (("(") reduce 245) (("+") reduce 245) (("-") reduce 245) ((";") reduce 245) (("@") reduce 245) ((T_ABSTRACT) reduce 245) ((T_ARRAY) reduce 245) ((T_ARRAY_CAST) reduce 245) ((T_ATTRIBUTE) reduce 245) ((T_BOOL_CAST) reduce 245) ((T_BREAK) reduce 245) ((T_CLASS) reduce 245) ((T_CLASS_C) reduce 245) ((T_CLONE) reduce 245) ((T_CONSTANT_ENCAPSED_STRING) reduce 245) ((T_CONTINUE) reduce 245) ((T_DEC) reduce 245) ((T_DECLARE) reduce 245) ((T_DIR) reduce 245) ((T_DNUMBER) reduce 245) ((T_DO) reduce 245) ((T_DOUBLE_CAST) reduce 245) ((T_ECHO) reduce 245) ((T_ELSE) reduce 245) ((T_ELSEIF) reduce 245) ((T_EMPTY) reduce 245) ((T_ENDIF) reduce 245) ((T_ENUM) reduce 245) ((T_EVAL) reduce 245) ((T_EXIT) reduce 245) ((T_FILE) reduce 245) ((T_FINAL) reduce 245) ((T_FN) reduce 245) ((T_FOR) reduce 245) ((T_FOREACH) reduce 245) ((T_FUNCTION) reduce 245) ((T_FUNC_C) reduce 245) ((T_GLOBAL) reduce 245) ((T_GOTO) reduce 245) ((T_HALT_COMPILER) reduce 245) ((T_IF) reduce 245) ((T_INC) reduce 245) ((T_INCLUDE) reduce 245) ((T_INCLUDE_ONCE) reduce 245) ((T_INLINE_HTML) reduce 245) ((T_INTERFACE) reduce 245) ((T_INT_CAST) reduce 245) ((T_ISSET) reduce 245) ((T_LINE) reduce 245) ((T_LIST) reduce 245) ((T_LNUMBER) reduce 245) ((T_MATCH) reduce 245) ((T_METHOD_C) reduce 245) ((T_NAME_FULLY_QUALIFIED) reduce 245) ((T_NAME_QUALIFIED) reduce 245) ((T_NAME_RELATIVE) reduce 245) ((T_NEW) reduce 245) ((T_NS_C) reduce 245) ((T_OBJECT_CAST) reduce 245) ((T_PRINT) reduce 245) ((T_PROPERTY_C) reduce 245) ((T_READONLY) reduce 245) ((T_REQUIRE) reduce 245) ((T_REQUIRE_ONCE) reduce 245) ((T_RETURN) reduce 245) ((T_START_HEREDOC) reduce 245) ((T_STATIC) reduce 245) ((T_STRING) reduce 245) ((T_STRING_CAST) reduce 245) ((T_SWITCH) reduce 245) ((T_THROW) reduce 245) ((T_TRAIT) reduce 245) ((T_TRAIT_C) reduce 245) ((T_TRY) reduce 245) ((T_UNSET) reduce 245) ((T_UNSET_CAST) reduce 245) ((T_VARIABLE) reduce 245) ((T_VOID_CAST) reduce 245) ((T_WHILE) reduce 245) ((T_YIELD) reduce 245) ((T_YIELD_FROM) reduce 245) (("[") reduce 245) (("`") reduce 245) (("{") reduce 245) (("~") reduce 245)) 17769 ((("!") reduce 247) (("\"") reduce 247) (("$") reduce 247) (("(") reduce 247) (("+") reduce 247) (("-") reduce 247) ((";") reduce 247) (("@") reduce 247) ((T_ABSTRACT) reduce 247) ((T_ARRAY) reduce 247) ((T_ARRAY_CAST) reduce 247) ((T_ATTRIBUTE) reduce 247) ((T_BOOL_CAST) reduce 247) ((T_BREAK) reduce 247) ((T_CLASS) reduce 247) ((T_CLASS_C) reduce 247) ((T_CLONE) reduce 247) ((T_CONSTANT_ENCAPSED_STRING) reduce 247) ((T_CONTINUE) reduce 247) ((T_DEC) reduce 247) ((T_DECLARE) reduce 247) ((T_DIR) reduce 247) ((T_DNUMBER) reduce 247) ((T_DO) reduce 247) ((T_DOUBLE_CAST) reduce 247) ((T_ECHO) reduce 247) ((T_ELSE) reduce 247) ((T_ELSEIF) reduce 247) ((T_EMPTY) reduce 247) ((T_ENDIF) reduce 247) ((T_ENUM) reduce 247) ((T_EVAL) reduce 247) ((T_EXIT) reduce 247) ((T_FILE) reduce 247) ((T_FINAL) reduce 247) ((T_FN) reduce 247) ((T_FOR) reduce 247) ((T_FOREACH) reduce 247) ((T_FUNCTION) reduce 247) ((T_FUNC_C) reduce 247) ((T_GLOBAL) reduce 247) ((T_GOTO) reduce 247) ((T_HALT_COMPILER) reduce 247) ((T_IF) reduce 247) ((T_INC) reduce 247) ((T_INCLUDE) reduce 247) ((T_INCLUDE_ONCE) reduce 247) ((T_INLINE_HTML) reduce 247) ((T_INTERFACE) reduce 247) ((T_INT_CAST) reduce 247) ((T_ISSET) reduce 247) ((T_LINE) reduce 247) ((T_LIST) reduce 247) ((T_LNUMBER) reduce 247) ((T_MATCH) reduce 247) ((T_METHOD_C) reduce 247) ((T_NAME_FULLY_QUALIFIED) reduce 247) ((T_NAME_QUALIFIED) reduce 247) ((T_NAME_RELATIVE) reduce 247) ((T_NEW) reduce 247) ((T_NS_C) reduce 247) ((T_OBJECT_CAST) reduce 247) ((T_PRINT) reduce 247) ((T_PROPERTY_C) reduce 247) ((T_READONLY) reduce 247) ((T_REQUIRE) reduce 247) ((T_REQUIRE_ONCE) reduce 247) ((T_RETURN) reduce 247) ((T_START_HEREDOC) reduce 247) ((T_STATIC) reduce 247) ((T_STRING) reduce 247) ((T_STRING_CAST) reduce 247) ((T_SWITCH) reduce 247) ((T_THROW) reduce 247) ((T_TRAIT) reduce 247) ((T_TRAIT_C) reduce 247) ((T_TRY) reduce 247) ((T_UNSET) reduce 247) ((T_UNSET_CAST) reduce 247) ((T_VARIABLE) reduce 247) ((T_VOID_CAST) reduce 247) ((T_WHILE) reduce 247) ((T_YIELD) reduce 247) ((T_YIELD_FROM) reduce 247) (("[") reduce 247) (("`") reduce 247) (("{") reduce 247) (("~") reduce 247)) 17770 ((("(") reduce 80) ((T_READONLY) reduce 80) ((T_STRING) reduce 80)) 17771 ((("(") reduce 81) ((T_READONLY) reduce 81) ((T_STRING) reduce 81)) 17772 ((("(") reduce 497) ((T_READONLY) reduce 497) ((T_STRING) reduce 497)) 17773 ((("(") reduce 493) ((T_READONLY) shift) ((T_STRING) shift)) 17774 ((("(") reduce 183)) 17775 ((("(") reduce 182)) 17785 ((("!") reduce 494) (("\"") reduce 494) (("$") reduce 494) (("(") reduce 494) (("+") reduce 494) (("-") reduce 494) ((";") reduce 494) (("@") reduce 494) ((T_ABSTRACT) reduce 494) ((T_ARRAY) reduce 494) ((T_ARRAY_CAST) reduce 494) ((T_ATTRIBUTE) reduce 494) ((T_BOOL_CAST) reduce 494) ((T_BREAK) reduce 494) ((T_CLASS) reduce 494) ((T_CLASS_C) reduce 494) ((T_CLONE) reduce 494) ((T_CONSTANT_ENCAPSED_STRING) reduce 494) ((T_CONTINUE) reduce 494) ((T_DEC) reduce 494) ((T_DECLARE) reduce 494) ((T_DIR) reduce 494) ((T_DNUMBER) reduce 494) ((T_DO) reduce 494) ((T_DOUBLE_CAST) reduce 494) ((T_ECHO) reduce 494) ((T_ELSE) reduce 494) ((T_ELSEIF) reduce 494) ((T_EMPTY) reduce 494) ((T_ENDIF) reduce 494) ((T_ENUM) reduce 494) ((T_EVAL) reduce 494) ((T_EXIT) reduce 494) ((T_FILE) reduce 494) ((T_FINAL) reduce 494) ((T_FN) reduce 494) ((T_FOR) reduce 494) ((T_FOREACH) reduce 494) ((T_FUNCTION) reduce 494) ((T_FUNC_C) reduce 494) ((T_GLOBAL) reduce 494) ((T_GOTO) reduce 494) ((T_HALT_COMPILER) reduce 494) ((T_IF) reduce 494) ((T_INC) reduce 494) ((T_INCLUDE) reduce 494) ((T_INCLUDE_ONCE) reduce 494) ((T_INLINE_HTML) reduce 494) ((T_INTERFACE) reduce 494) ((T_INT_CAST) reduce 494) ((T_ISSET) reduce 494) ((T_LINE) reduce 494) ((T_LIST) reduce 494) ((T_LNUMBER) reduce 494) ((T_MATCH) reduce 494) ((T_METHOD_C) reduce 494) ((T_NAME_FULLY_QUALIFIED) reduce 494) ((T_NAME_QUALIFIED) reduce 494) ((T_NAME_RELATIVE) reduce 494) ((T_NEW) reduce 494) ((T_NS_C) reduce 494) ((T_OBJECT_CAST) reduce 494) ((T_PRINT) reduce 494) ((T_PROPERTY_C) reduce 494) ((T_READONLY) reduce 494) ((T_REQUIRE) reduce 494) ((T_REQUIRE_ONCE) reduce 494) ((T_RETURN) reduce 494) ((T_START_HEREDOC) reduce 494) ((T_STATIC) reduce 494) ((T_STRING) reduce 494) ((T_STRING_CAST) reduce 494) ((T_SWITCH) reduce 494) ((T_THROW) reduce 494) ((T_TRAIT) reduce 494) ((T_TRAIT_C) reduce 494) ((T_TRY) reduce 494) ((T_UNSET) reduce 494) ((T_UNSET_CAST) reduce 494) ((T_VARIABLE) reduce 494) ((T_VOID_CAST) reduce 494) ((T_WHILE) reduce 494) ((T_YIELD) reduce 494) ((T_YIELD_FROM) reduce 494) (("[") reduce 494) (("`") reduce 494) (("{") reduce 494) (("~") reduce 494)) 17786 ((("!") reduce 184) (("\"") reduce 184) (("$") reduce 184) (("(") reduce 184) (("+") reduce 184) (("-") reduce 184) ((";") reduce 184) (("@") reduce 184) ((T_ABSTRACT) reduce 184) ((T_ARRAY) reduce 184) ((T_ARRAY_CAST) reduce 184) ((T_ATTRIBUTE) reduce 184) ((T_BOOL_CAST) reduce 184) ((T_BREAK) reduce 184) ((T_CLASS) reduce 184) ((T_CLASS_C) reduce 184) ((T_CLONE) reduce 184) ((T_CONSTANT_ENCAPSED_STRING) reduce 184) ((T_CONTINUE) reduce 184) ((T_DEC) reduce 184) ((T_DECLARE) reduce 184) ((T_DIR) reduce 184) ((T_DNUMBER) reduce 184) ((T_DO) reduce 184) ((T_DOUBLE_CAST) reduce 184) ((T_ECHO) reduce 184) ((T_ELSE) reduce 184) ((T_ELSEIF) reduce 184) ((T_EMPTY) reduce 184) ((T_ENDIF) reduce 184) ((T_ENUM) reduce 184) ((T_EVAL) reduce 184) ((T_EXIT) reduce 184) ((T_FILE) reduce 184) ((T_FINAL) reduce 184) ((T_FN) reduce 184) ((T_FOR) reduce 184) ((T_FOREACH) reduce 184) ((T_FUNCTION) reduce 184) ((T_FUNC_C) reduce 184) ((T_GLOBAL) reduce 184) ((T_GOTO) reduce 184) ((T_HALT_COMPILER) reduce 184) ((T_IF) reduce 184) ((T_INC) reduce 184) ((T_INCLUDE) reduce 184) ((T_INCLUDE_ONCE) reduce 184) ((T_INLINE_HTML) reduce 184) ((T_INTERFACE) reduce 184) ((T_INT_CAST) reduce 184) ((T_ISSET) reduce 184) ((T_LINE) reduce 184) ((T_LIST) reduce 184) ((T_LNUMBER) reduce 184) ((T_MATCH) reduce 184) ((T_METHOD_C) reduce 184) ((T_NAME_FULLY_QUALIFIED) reduce 184) ((T_NAME_QUALIFIED) reduce 184) ((T_NAME_RELATIVE) reduce 184) ((T_NEW) reduce 184) ((T_NS_C) reduce 184) ((T_OBJECT_CAST) reduce 184) ((T_PRINT) reduce 184) ((T_PROPERTY_C) reduce 184) ((T_READONLY) reduce 184) ((T_REQUIRE) reduce 184) ((T_REQUIRE_ONCE) reduce 184) ((T_RETURN) reduce 184) ((T_START_HEREDOC) reduce 184) ((T_STATIC) reduce 184) ((T_STRING) reduce 184) ((T_STRING_CAST) reduce 184) ((T_SWITCH) reduce 184) ((T_THROW) reduce 184) ((T_TRAIT) reduce 184) ((T_TRAIT_C) reduce 184) ((T_TRY) reduce 184) ((T_UNSET) reduce 184) ((T_UNSET_CAST) reduce 184) ((T_VARIABLE) reduce 184) ((T_VOID_CAST) reduce 184) ((T_WHILE) reduce 184) ((T_YIELD) reduce 184) ((T_YIELD_FROM) reduce 184) (("[") reduce 184) (("`") reduce 184) (("{") reduce 184) (("~") reduce 184)) 17787 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_ELSE) reduce 161) ((T_ELSEIF) reduce 161) ((T_EMPTY) reduce 161) ((T_ENDIF) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_VOID_CAST) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 17789 (((T_ABSTRACT) reduce 192) ((T_CLASS) reduce 192) ((T_FINAL) reduce 192) ((T_READONLY) reduce 192)) 17796 ((("!") reduce 189) (("\"") reduce 189) (("$") reduce 189) (("(") reduce 189) (("+") reduce 189) (("-") reduce 189) ((";") reduce 189) (("@") reduce 189) ((T_ABSTRACT) reduce 189) ((T_ARRAY) reduce 189) ((T_ARRAY_CAST) reduce 189) ((T_ATTRIBUTE) reduce 189) ((T_BOOL_CAST) reduce 189) ((T_BREAK) reduce 189) ((T_CLASS) reduce 189) ((T_CLASS_C) reduce 189) ((T_CLONE) reduce 189) ((T_CONSTANT_ENCAPSED_STRING) reduce 189) ((T_CONTINUE) reduce 189) ((T_DEC) reduce 189) ((T_DECLARE) reduce 189) ((T_DIR) reduce 189) ((T_DNUMBER) reduce 189) ((T_DO) reduce 189) ((T_DOUBLE_CAST) reduce 189) ((T_ECHO) reduce 189) ((T_ELSE) reduce 189) ((T_ELSEIF) reduce 189) ((T_EMPTY) reduce 189) ((T_ENDIF) reduce 189) ((T_ENUM) reduce 189) ((T_EVAL) reduce 189) ((T_EXIT) reduce 189) ((T_FILE) reduce 189) ((T_FINAL) reduce 189) ((T_FN) reduce 189) ((T_FOR) reduce 189) ((T_FOREACH) reduce 189) ((T_FUNCTION) reduce 189) ((T_FUNC_C) reduce 189) ((T_GLOBAL) reduce 189) ((T_GOTO) reduce 189) ((T_HALT_COMPILER) reduce 189) ((T_IF) reduce 189) ((T_INC) reduce 189) ((T_INCLUDE) reduce 189) ((T_INCLUDE_ONCE) reduce 189) ((T_INLINE_HTML) reduce 189) ((T_INTERFACE) reduce 189) ((T_INT_CAST) reduce 189) ((T_ISSET) reduce 189) ((T_LINE) reduce 189) ((T_LIST) reduce 189) ((T_LNUMBER) reduce 189) ((T_MATCH) reduce 189) ((T_METHOD_C) reduce 189) ((T_NAME_FULLY_QUALIFIED) reduce 189) ((T_NAME_QUALIFIED) reduce 189) ((T_NAME_RELATIVE) reduce 189) ((T_NEW) reduce 189) ((T_NS_C) reduce 189) ((T_OBJECT_CAST) reduce 189) ((T_PRINT) reduce 189) ((T_PROPERTY_C) reduce 189) ((T_READONLY) reduce 189) ((T_REQUIRE) reduce 189) ((T_REQUIRE_ONCE) reduce 189) ((T_RETURN) reduce 189) ((T_START_HEREDOC) reduce 189) ((T_STATIC) reduce 189) ((T_STRING) reduce 189) ((T_STRING_CAST) reduce 189) ((T_SWITCH) reduce 189) ((T_THROW) reduce 189) ((T_TRAIT) reduce 189) ((T_TRAIT_C) reduce 189) ((T_TRY) reduce 189) ((T_UNSET) reduce 189) ((T_UNSET_CAST) reduce 189) ((T_VARIABLE) reduce 189) ((T_VOID_CAST) reduce 189) ((T_WHILE) reduce 189) ((T_YIELD) reduce 189) ((T_YIELD_FROM) reduce 189) (("[") reduce 189) (("`") reduce 189) (("{") reduce 189) (("~") reduce 189)) 17798 (((T_ABSTRACT) reduce 102) ((T_ATTRIBUTE) reduce 102) ((T_CLASS) reduce 102) ((T_ENUM) reduce 102) ((T_FINAL) reduce 102) ((T_FN) reduce 102) ((T_FUNCTION) reduce 102) ((T_INTERFACE) reduce 102) ((T_READONLY) reduce 102) ((T_STATIC) reduce 102) ((T_TRAIT) reduce 102)) 17799 ((("!") reduce 145) (("\"") reduce 145) (("$") reduce 145) (("(") reduce 145) (("+") reduce 145) (("-") reduce 145) ((";") reduce 145) (("@") reduce 145) ((T_ABSTRACT) reduce 145) ((T_ARRAY) reduce 145) ((T_ARRAY_CAST) reduce 145) ((T_ATTRIBUTE) reduce 145) ((T_BOOL_CAST) reduce 145) ((T_BREAK) reduce 145) ((T_CLASS) reduce 145) ((T_CLASS_C) reduce 145) ((T_CLONE) reduce 145) ((T_CONSTANT_ENCAPSED_STRING) reduce 145) ((T_CONTINUE) reduce 145) ((T_DEC) reduce 145) ((T_DECLARE) reduce 145) ((T_DIR) reduce 145) ((T_DNUMBER) reduce 145) ((T_DO) reduce 145) ((T_DOUBLE_CAST) reduce 145) ((T_ECHO) reduce 145) ((T_ELSE) reduce 145) ((T_ELSEIF) reduce 145) ((T_EMPTY) reduce 145) ((T_ENDIF) reduce 145) ((T_ENUM) reduce 145) ((T_EVAL) reduce 145) ((T_EXIT) reduce 145) ((T_FILE) reduce 145) ((T_FINAL) reduce 145) ((T_FN) reduce 145) ((T_FOR) reduce 145) ((T_FOREACH) reduce 145) ((T_FUNCTION) reduce 145) ((T_FUNC_C) reduce 145) ((T_GLOBAL) reduce 145) ((T_GOTO) reduce 145) ((T_HALT_COMPILER) reduce 145) ((T_IF) reduce 145) ((T_INC) reduce 145) ((T_INCLUDE) reduce 145) ((T_INCLUDE_ONCE) reduce 145) ((T_INLINE_HTML) reduce 145) ((T_INTERFACE) reduce 145) ((T_INT_CAST) reduce 145) ((T_ISSET) reduce 145) ((T_LINE) reduce 145) ((T_LIST) reduce 145) ((T_LNUMBER) reduce 145) ((T_MATCH) reduce 145) ((T_METHOD_C) reduce 145) ((T_NAME_FULLY_QUALIFIED) reduce 145) ((T_NAME_QUALIFIED) reduce 145) ((T_NAME_RELATIVE) reduce 145) ((T_NEW) reduce 145) ((T_NS_C) reduce 145) ((T_OBJECT_CAST) reduce 145) ((T_PRINT) reduce 145) ((T_PROPERTY_C) reduce 145) ((T_READONLY) reduce 145) ((T_REQUIRE) reduce 145) ((T_REQUIRE_ONCE) reduce 145) ((T_RETURN) reduce 145) ((T_START_HEREDOC) reduce 145) ((T_STATIC) reduce 145) ((T_STRING) reduce 145) ((T_STRING_CAST) reduce 145) ((T_SWITCH) reduce 145) ((T_THROW) reduce 145) ((T_TRAIT) reduce 145) ((T_TRAIT_C) reduce 145) ((T_TRY) reduce 145) ((T_UNSET) reduce 145) ((T_UNSET_CAST) reduce 145) ((T_VARIABLE) reduce 145) ((T_VOID_CAST) reduce 145) ((T_WHILE) reduce 145) ((T_YIELD) reduce 145) ((T_YIELD_FROM) reduce 145) (("[") reduce 145) (("`") reduce 145) (("{") reduce 145) (("~") reduce 145)) 17800 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) ((";") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_PIPE) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("^") reduce 485) (("|") reduce 485)) 17801 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) ((";") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_PIPE) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("^") reduce 487) (("|") reduce 487)) 17804 ((("!") reduce 250) (("\"") reduce 250) (("$") reduce 250) (("(") reduce 250) (("+") reduce 250) (("-") reduce 250) ((";") reduce 250) (("@") reduce 250) ((T_ABSTRACT) reduce 250) ((T_ARRAY) reduce 250) ((T_ARRAY_CAST) reduce 250) ((T_ATTRIBUTE) reduce 250) ((T_BOOL_CAST) reduce 250) ((T_BREAK) reduce 250) ((T_CLASS) reduce 250) ((T_CLASS_C) reduce 250) ((T_CLONE) reduce 250) ((T_CONSTANT_ENCAPSED_STRING) reduce 250) ((T_CONTINUE) reduce 250) ((T_DEC) reduce 250) ((T_DECLARE) reduce 250) ((T_DIR) reduce 250) ((T_DNUMBER) reduce 250) ((T_DO) reduce 250) ((T_DOUBLE_CAST) reduce 250) ((T_ECHO) reduce 250) ((T_ELSE) reduce 250) ((T_ELSEIF) reduce 250) ((T_EMPTY) reduce 250) ((T_ENDIF) reduce 250) ((T_ENUM) reduce 250) ((T_EVAL) reduce 250) ((T_EXIT) reduce 250) ((T_FILE) reduce 250) ((T_FINAL) reduce 250) ((T_FN) reduce 250) ((T_FOR) reduce 250) ((T_FOREACH) reduce 250) ((T_FUNCTION) reduce 250) ((T_FUNC_C) reduce 250) ((T_GLOBAL) reduce 250) ((T_GOTO) reduce 250) ((T_HALT_COMPILER) reduce 250) ((T_IF) reduce 250) ((T_INC) reduce 250) ((T_INCLUDE) reduce 250) ((T_INCLUDE_ONCE) reduce 250) ((T_INLINE_HTML) reduce 250) ((T_INTERFACE) reduce 250) ((T_INT_CAST) reduce 250) ((T_ISSET) reduce 250) ((T_LINE) reduce 250) ((T_LIST) reduce 250) ((T_LNUMBER) reduce 250) ((T_MATCH) reduce 250) ((T_METHOD_C) reduce 250) ((T_NAME_FULLY_QUALIFIED) reduce 250) ((T_NAME_QUALIFIED) reduce 250) ((T_NAME_RELATIVE) reduce 250) ((T_NEW) reduce 250) ((T_NS_C) reduce 250) ((T_OBJECT_CAST) reduce 250) ((T_PRINT) reduce 250) ((T_PROPERTY_C) reduce 250) ((T_READONLY) reduce 250) ((T_REQUIRE) reduce 250) ((T_REQUIRE_ONCE) reduce 250) ((T_RETURN) reduce 250) ((T_START_HEREDOC) reduce 250) ((T_STATIC) reduce 250) ((T_STRING) reduce 250) ((T_STRING_CAST) reduce 250) ((T_SWITCH) reduce 250) ((T_THROW) reduce 250) ((T_TRAIT) reduce 250) ((T_TRAIT_C) reduce 250) ((T_TRY) reduce 250) ((T_UNSET) reduce 250) ((T_UNSET_CAST) reduce 250) ((T_VARIABLE) reduce 250) ((T_VOID_CAST) reduce 250) ((T_WHILE) reduce 250) ((T_YIELD) reduce 250) ((T_YIELD_FROM) reduce 250) (("[") reduce 250) (("`") reduce 250) (("{") reduce 250) (("~") reduce 250)) 17805 ((("!") reduce 142) (("\"") reduce 142) (("$") reduce 142) (("(") reduce 142) (("+") reduce 142) (("-") reduce 142) ((";") reduce 142) (("@") reduce 142) ((T_ABSTRACT) reduce 142) ((T_ARRAY) reduce 142) ((T_ARRAY_CAST) reduce 142) ((T_ATTRIBUTE) reduce 142) ((T_BOOL_CAST) reduce 142) ((T_BREAK) reduce 142) ((T_CLASS) reduce 142) ((T_CLASS_C) reduce 142) ((T_CLONE) reduce 142) ((T_CONSTANT_ENCAPSED_STRING) reduce 142) ((T_CONTINUE) reduce 142) ((T_DEC) reduce 142) ((T_DECLARE) reduce 142) ((T_DIR) reduce 142) ((T_DNUMBER) reduce 142) ((T_DO) reduce 142) ((T_DOUBLE_CAST) reduce 142) ((T_ECHO) reduce 142) ((T_EMPTY) reduce 142) ((T_ENDIF) reduce 142) ((T_ENUM) reduce 142) ((T_EVAL) reduce 142) ((T_EXIT) reduce 142) ((T_FILE) reduce 142) ((T_FINAL) reduce 142) ((T_FN) reduce 142) ((T_FOR) reduce 142) ((T_FOREACH) reduce 142) ((T_FUNCTION) reduce 142) ((T_FUNC_C) reduce 142) ((T_GLOBAL) reduce 142) ((T_GOTO) reduce 142) ((T_HALT_COMPILER) reduce 142) ((T_IF) reduce 142) ((T_INC) reduce 142) ((T_INCLUDE) reduce 142) ((T_INCLUDE_ONCE) reduce 142) ((T_INLINE_HTML) reduce 142) ((T_INTERFACE) reduce 142) ((T_INT_CAST) reduce 142) ((T_ISSET) reduce 142) ((T_LINE) reduce 142) ((T_LIST) reduce 142) ((T_LNUMBER) reduce 142) ((T_MATCH) reduce 142) ((T_METHOD_C) reduce 142) ((T_NAME_FULLY_QUALIFIED) reduce 142) ((T_NAME_QUALIFIED) reduce 142) ((T_NAME_RELATIVE) reduce 142) ((T_NEW) reduce 142) ((T_NS_C) reduce 142) ((T_OBJECT_CAST) reduce 142) ((T_PRINT) reduce 142) ((T_PROPERTY_C) reduce 142) ((T_READONLY) reduce 142) ((T_REQUIRE) reduce 142) ((T_REQUIRE_ONCE) reduce 142) ((T_RETURN) reduce 142) ((T_START_HEREDOC) reduce 142) ((T_STATIC) reduce 142) ((T_STRING) reduce 142) ((T_STRING_CAST) reduce 142) ((T_SWITCH) reduce 142) ((T_THROW) reduce 142) ((T_TRAIT) reduce 142) ((T_TRAIT_C) reduce 142) ((T_TRY) reduce 142) ((T_UNSET) reduce 142) ((T_UNSET_CAST) reduce 142) ((T_VARIABLE) reduce 142) ((T_VOID_CAST) reduce 142) ((T_WHILE) reduce 142) ((T_YIELD) reduce 142) ((T_YIELD_FROM) reduce 142) (("[") reduce 142) (("`") reduce 142) (("{") reduce 142) (("~") reduce 142)) 17806 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_ENDIF) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_VOID_CAST) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("~") shift)) 17807 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_EMPTY) reduce 166) ((T_ENDIF) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_VOID_CAST) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 17821 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_EMPTY) reduce 160) ((T_ENDIF) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_VOID_CAST) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 17832 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_EMPTY) reduce 149) ((T_ENDIF) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_VOID_CAST) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 17834 ((("!") reduce 144) (("\"") reduce 144) (("$") reduce 144) (("(") reduce 144) (("+") reduce 144) (("-") reduce 144) ((";") reduce 144) (("@") reduce 144) ((T_ABSTRACT) reduce 144) ((T_ARRAY) reduce 144) ((T_ARRAY_CAST) reduce 144) ((T_ATTRIBUTE) reduce 144) ((T_BOOL_CAST) reduce 144) ((T_BREAK) reduce 144) ((T_CLASS) reduce 144) ((T_CLASS_C) reduce 144) ((T_CLONE) reduce 144) ((T_CONSTANT_ENCAPSED_STRING) reduce 144) ((T_CONTINUE) reduce 144) ((T_DEC) reduce 144) ((T_DECLARE) reduce 144) ((T_DIR) reduce 144) ((T_DNUMBER) reduce 144) ((T_DO) reduce 144) ((T_DOUBLE_CAST) reduce 144) ((T_ECHO) reduce 144) ((T_EMPTY) reduce 144) ((T_ENDIF) reduce 144) ((T_ENUM) reduce 144) ((T_EVAL) reduce 144) ((T_EXIT) reduce 144) ((T_FILE) reduce 144) ((T_FINAL) reduce 144) ((T_FN) reduce 144) ((T_FOR) reduce 144) ((T_FOREACH) reduce 144) ((T_FUNCTION) reduce 144) ((T_FUNC_C) reduce 144) ((T_GLOBAL) reduce 144) ((T_GOTO) reduce 144) ((T_HALT_COMPILER) reduce 144) ((T_IF) reduce 144) ((T_INC) reduce 144) ((T_INCLUDE) reduce 144) ((T_INCLUDE_ONCE) reduce 144) ((T_INLINE_HTML) reduce 144) ((T_INTERFACE) reduce 144) ((T_INT_CAST) reduce 144) ((T_ISSET) reduce 144) ((T_LINE) reduce 144) ((T_LIST) reduce 144) ((T_LNUMBER) reduce 144) ((T_MATCH) reduce 144) ((T_METHOD_C) reduce 144) ((T_NAME_FULLY_QUALIFIED) reduce 144) ((T_NAME_QUALIFIED) reduce 144) ((T_NAME_RELATIVE) reduce 144) ((T_NEW) reduce 144) ((T_NS_C) reduce 144) ((T_OBJECT_CAST) reduce 144) ((T_PRINT) reduce 144) ((T_PROPERTY_C) reduce 144) ((T_READONLY) reduce 144) ((T_REQUIRE) reduce 144) ((T_REQUIRE_ONCE) reduce 144) ((T_RETURN) reduce 144) ((T_START_HEREDOC) reduce 144) ((T_STATIC) reduce 144) ((T_STRING) reduce 144) ((T_STRING_CAST) reduce 144) ((T_SWITCH) reduce 144) ((T_THROW) reduce 144) ((T_TRAIT) reduce 144) ((T_TRAIT_C) reduce 144) ((T_TRY) reduce 144) ((T_UNSET) reduce 144) ((T_UNSET_CAST) reduce 144) ((T_VARIABLE) reduce 144) ((T_VOID_CAST) reduce 144) ((T_WHILE) reduce 144) ((T_YIELD) reduce 144) ((T_YIELD_FROM) reduce 144) (("[") reduce 144) (("`") reduce 144) (("{") reduce 144) (("~") reduce 144)) 17836 ((("!") reduce 104) (("\"") reduce 104) (("$") reduce 104) (("(") reduce 104) (("+") reduce 104) (("-") reduce 104) ((";") reduce 104) (("@") reduce 104) ((T_ABSTRACT) reduce 104) ((T_ARRAY) reduce 104) ((T_ARRAY_CAST) reduce 104) ((T_ATTRIBUTE) reduce 104) ((T_BOOL_CAST) reduce 104) ((T_BREAK) reduce 104) ((T_CLASS) reduce 104) ((T_CLASS_C) reduce 104) ((T_CLONE) reduce 104) ((T_CONSTANT_ENCAPSED_STRING) reduce 104) ((T_CONTINUE) reduce 104) ((T_DEC) reduce 104) ((T_DECLARE) reduce 104) ((T_DIR) reduce 104) ((T_DNUMBER) reduce 104) ((T_DO) reduce 104) ((T_DOUBLE_CAST) reduce 104) ((T_ECHO) reduce 104) ((T_EMPTY) reduce 104) ((T_ENDIF) reduce 104) ((T_ENUM) reduce 104) ((T_EVAL) reduce 104) ((T_EXIT) reduce 104) ((T_FILE) reduce 104) ((T_FINAL) reduce 104) ((T_FN) reduce 104) ((T_FOR) reduce 104) ((T_FOREACH) reduce 104) ((T_FUNCTION) reduce 104) ((T_FUNC_C) reduce 104) ((T_GLOBAL) reduce 104) ((T_GOTO) reduce 104) ((T_HALT_COMPILER) reduce 104) ((T_IF) reduce 104) ((T_INC) reduce 104) ((T_INCLUDE) reduce 104) ((T_INCLUDE_ONCE) reduce 104) ((T_INLINE_HTML) reduce 104) ((T_INTERFACE) reduce 104) ((T_INT_CAST) reduce 104) ((T_ISSET) reduce 104) ((T_LINE) reduce 104) ((T_LIST) reduce 104) ((T_LNUMBER) reduce 104) ((T_MATCH) reduce 104) ((T_METHOD_C) reduce 104) ((T_NAME_FULLY_QUALIFIED) reduce 104) ((T_NAME_QUALIFIED) reduce 104) ((T_NAME_RELATIVE) reduce 104) ((T_NEW) reduce 104) ((T_NS_C) reduce 104) ((T_OBJECT_CAST) reduce 104) ((T_PRINT) reduce 104) ((T_PROPERTY_C) reduce 104) ((T_READONLY) reduce 104) ((T_REQUIRE) reduce 104) ((T_REQUIRE_ONCE) reduce 104) ((T_RETURN) reduce 104) ((T_START_HEREDOC) reduce 104) ((T_STATIC) reduce 104) ((T_STRING) reduce 104) ((T_STRING_CAST) reduce 104) ((T_SWITCH) reduce 104) ((T_THROW) reduce 104) ((T_TRAIT) reduce 104) ((T_TRAIT_C) reduce 104) ((T_TRY) reduce 104) ((T_UNSET) reduce 104) ((T_UNSET_CAST) reduce 104) ((T_VARIABLE) reduce 104) ((T_VOID_CAST) reduce 104) ((T_WHILE) reduce 104) ((T_YIELD) reduce 104) ((T_YIELD_FROM) reduce 104) (("[") reduce 104) (("`") reduce 104) (("{") reduce 104) (("~") reduce 104)) 17838 ((("!") reduce 107) (("\"") reduce 107) (("$") reduce 107) (("(") reduce 107) (("+") reduce 107) (("-") reduce 107) ((";") reduce 107) (("@") reduce 107) ((T_ABSTRACT) reduce 107) ((T_ARRAY) reduce 107) ((T_ARRAY_CAST) reduce 107) ((T_ATTRIBUTE) reduce 107) ((T_BOOL_CAST) reduce 107) ((T_BREAK) reduce 107) ((T_CLASS) reduce 107) ((T_CLASS_C) reduce 107) ((T_CLONE) reduce 107) ((T_CONSTANT_ENCAPSED_STRING) reduce 107) ((T_CONTINUE) reduce 107) ((T_DEC) reduce 107) ((T_DECLARE) reduce 107) ((T_DIR) reduce 107) ((T_DNUMBER) reduce 107) ((T_DO) reduce 107) ((T_DOUBLE_CAST) reduce 107) ((T_ECHO) reduce 107) ((T_EMPTY) reduce 107) ((T_ENDIF) reduce 107) ((T_ENUM) reduce 107) ((T_EVAL) reduce 107) ((T_EXIT) reduce 107) ((T_FILE) reduce 107) ((T_FINAL) reduce 107) ((T_FN) reduce 107) ((T_FOR) reduce 107) ((T_FOREACH) reduce 107) ((T_FUNCTION) reduce 107) ((T_FUNC_C) reduce 107) ((T_GLOBAL) reduce 107) ((T_GOTO) reduce 107) ((T_HALT_COMPILER) reduce 107) ((T_IF) reduce 107) ((T_INC) reduce 107) ((T_INCLUDE) reduce 107) ((T_INCLUDE_ONCE) reduce 107) ((T_INLINE_HTML) reduce 107) ((T_INTERFACE) reduce 107) ((T_INT_CAST) reduce 107) ((T_ISSET) reduce 107) ((T_LINE) reduce 107) ((T_LIST) reduce 107) ((T_LNUMBER) reduce 107) ((T_MATCH) reduce 107) ((T_METHOD_C) reduce 107) ((T_NAME_FULLY_QUALIFIED) reduce 107) ((T_NAME_QUALIFIED) reduce 107) ((T_NAME_RELATIVE) reduce 107) ((T_NEW) reduce 107) ((T_NS_C) reduce 107) ((T_OBJECT_CAST) reduce 107) ((T_PRINT) reduce 107) ((T_PROPERTY_C) reduce 107) ((T_READONLY) reduce 107) ((T_REQUIRE) reduce 107) ((T_REQUIRE_ONCE) reduce 107) ((T_RETURN) reduce 107) ((T_START_HEREDOC) reduce 107) ((T_STATIC) reduce 107) ((T_STRING) reduce 107) ((T_STRING_CAST) reduce 107) ((T_SWITCH) reduce 107) ((T_THROW) reduce 107) ((T_TRAIT) reduce 107) ((T_TRAIT_C) reduce 107) ((T_TRY) reduce 107) ((T_UNSET) reduce 107) ((T_UNSET_CAST) reduce 107) ((T_VARIABLE) reduce 107) ((T_VOID_CAST) reduce 107) ((T_WHILE) reduce 107) ((T_YIELD) reduce 107) ((T_YIELD_FROM) reduce 107) (("[") reduce 107) (("`") reduce 107) (("{") reduce 107) (("~") reduce 107)) 17841 ((("!") reduce 103) (("\"") reduce 103) (("$") reduce 103) (("(") reduce 103) (("+") reduce 103) (("-") reduce 103) ((";") reduce 103) (("@") reduce 103) ((T_ABSTRACT) reduce 103) ((T_ARRAY) reduce 103) ((T_ARRAY_CAST) reduce 103) ((T_ATTRIBUTE) reduce 103) ((T_BOOL_CAST) reduce 103) ((T_BREAK) reduce 103) ((T_CLASS) reduce 103) ((T_CLASS_C) reduce 103) ((T_CLONE) reduce 103) ((T_CONSTANT_ENCAPSED_STRING) reduce 103) ((T_CONTINUE) reduce 103) ((T_DEC) reduce 103) ((T_DECLARE) reduce 103) ((T_DIR) reduce 103) ((T_DNUMBER) reduce 103) ((T_DO) reduce 103) ((T_DOUBLE_CAST) reduce 103) ((T_ECHO) reduce 103) ((T_EMPTY) reduce 103) ((T_ENDIF) reduce 103) ((T_ENUM) reduce 103) ((T_EVAL) reduce 103) ((T_EXIT) reduce 103) ((T_FILE) reduce 103) ((T_FINAL) reduce 103) ((T_FN) reduce 103) ((T_FOR) reduce 103) ((T_FOREACH) reduce 103) ((T_FUNCTION) reduce 103) ((T_FUNC_C) reduce 103) ((T_GLOBAL) reduce 103) ((T_GOTO) reduce 103) ((T_HALT_COMPILER) reduce 103) ((T_IF) reduce 103) ((T_INC) reduce 103) ((T_INCLUDE) reduce 103) ((T_INCLUDE_ONCE) reduce 103) ((T_INLINE_HTML) reduce 103) ((T_INTERFACE) reduce 103) ((T_INT_CAST) reduce 103) ((T_ISSET) reduce 103) ((T_LINE) reduce 103) ((T_LIST) reduce 103) ((T_LNUMBER) reduce 103) ((T_MATCH) reduce 103) ((T_METHOD_C) reduce 103) ((T_NAME_FULLY_QUALIFIED) reduce 103) ((T_NAME_QUALIFIED) reduce 103) ((T_NAME_RELATIVE) reduce 103) ((T_NEW) reduce 103) ((T_NS_C) reduce 103) ((T_OBJECT_CAST) reduce 103) ((T_PRINT) reduce 103) ((T_PROPERTY_C) reduce 103) ((T_READONLY) reduce 103) ((T_REQUIRE) reduce 103) ((T_REQUIRE_ONCE) reduce 103) ((T_RETURN) reduce 103) ((T_START_HEREDOC) reduce 103) ((T_STATIC) reduce 103) ((T_STRING) reduce 103) ((T_STRING_CAST) reduce 103) ((T_SWITCH) reduce 103) ((T_THROW) reduce 103) ((T_TRAIT) reduce 103) ((T_TRAIT_C) reduce 103) ((T_TRY) reduce 103) ((T_UNSET) reduce 103) ((T_UNSET_CAST) reduce 103) ((T_VARIABLE) reduce 103) ((T_VOID_CAST) reduce 103) ((T_WHILE) reduce 103) ((T_YIELD) reduce 103) ((T_YIELD_FROM) reduce 103) (("[") reduce 103) (("`") reduce 103) (("{") reduce 103) (("~") reduce 103)) 17842 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_EMPTY) reduce 148) ((T_ENDIF) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_VOID_CAST) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 17844 ((("!") reduce 141) (("\"") reduce 141) (("$") reduce 141) (("(") reduce 141) (("+") reduce 141) (("-") reduce 141) ((";") reduce 141) (("@") reduce 141) ((T_ABSTRACT) reduce 141) ((T_ARRAY) reduce 141) ((T_ARRAY_CAST) reduce 141) ((T_ATTRIBUTE) reduce 141) ((T_BOOL_CAST) reduce 141) ((T_BREAK) reduce 141) ((T_CLASS) reduce 141) ((T_CLASS_C) reduce 141) ((T_CLONE) reduce 141) ((T_CONSTANT_ENCAPSED_STRING) reduce 141) ((T_CONTINUE) reduce 141) ((T_DEC) reduce 141) ((T_DECLARE) reduce 141) ((T_DIR) reduce 141) ((T_DNUMBER) reduce 141) ((T_DO) reduce 141) ((T_DOUBLE_CAST) reduce 141) ((T_ECHO) reduce 141) ((T_EMPTY) reduce 141) ((T_ENDIF) reduce 141) ((T_ENUM) reduce 141) ((T_EVAL) reduce 141) ((T_EXIT) reduce 141) ((T_FILE) reduce 141) ((T_FINAL) reduce 141) ((T_FN) reduce 141) ((T_FOR) reduce 141) ((T_FOREACH) reduce 141) ((T_FUNCTION) reduce 141) ((T_FUNC_C) reduce 141) ((T_GLOBAL) reduce 141) ((T_GOTO) reduce 141) ((T_HALT_COMPILER) reduce 141) ((T_IF) reduce 141) ((T_INC) reduce 141) ((T_INCLUDE) reduce 141) ((T_INCLUDE_ONCE) reduce 141) ((T_INLINE_HTML) reduce 141) ((T_INTERFACE) reduce 141) ((T_INT_CAST) reduce 141) ((T_ISSET) reduce 141) ((T_LINE) reduce 141) ((T_LIST) reduce 141) ((T_LNUMBER) reduce 141) ((T_MATCH) reduce 141) ((T_METHOD_C) reduce 141) ((T_NAME_FULLY_QUALIFIED) reduce 141) ((T_NAME_QUALIFIED) reduce 141) ((T_NAME_RELATIVE) reduce 141) ((T_NEW) reduce 141) ((T_NS_C) reduce 141) ((T_OBJECT_CAST) reduce 141) ((T_PRINT) reduce 141) ((T_PROPERTY_C) reduce 141) ((T_READONLY) reduce 141) ((T_REQUIRE) reduce 141) ((T_REQUIRE_ONCE) reduce 141) ((T_RETURN) reduce 141) ((T_START_HEREDOC) reduce 141) ((T_STATIC) reduce 141) ((T_STRING) reduce 141) ((T_STRING_CAST) reduce 141) ((T_SWITCH) reduce 141) ((T_THROW) reduce 141) ((T_TRAIT) reduce 141) ((T_TRAIT_C) reduce 141) ((T_TRY) reduce 141) ((T_UNSET) reduce 141) ((T_UNSET_CAST) reduce 141) ((T_VARIABLE) reduce 141) ((T_VOID_CAST) reduce 141) ((T_WHILE) reduce 141) ((T_YIELD) reduce 141) ((T_YIELD_FROM) reduce 141) (("[") reduce 141) (("`") reduce 141) (("{") reduce 141) (("~") reduce 141)) 17845 ((("!") reduce 106) (("\"") reduce 106) (("$") reduce 106) (("(") reduce 106) (("+") reduce 106) (("-") reduce 106) ((";") reduce 106) (("@") reduce 106) ((T_ABSTRACT) reduce 106) ((T_ARRAY) reduce 106) ((T_ARRAY_CAST) reduce 106) ((T_ATTRIBUTE) reduce 106) ((T_BOOL_CAST) reduce 106) ((T_BREAK) reduce 106) ((T_CLASS) reduce 106) ((T_CLASS_C) reduce 106) ((T_CLONE) reduce 106) ((T_CONSTANT_ENCAPSED_STRING) reduce 106) ((T_CONTINUE) reduce 106) ((T_DEC) reduce 106) ((T_DECLARE) reduce 106) ((T_DIR) reduce 106) ((T_DNUMBER) reduce 106) ((T_DO) reduce 106) ((T_DOUBLE_CAST) reduce 106) ((T_ECHO) reduce 106) ((T_EMPTY) reduce 106) ((T_ENDIF) reduce 106) ((T_ENUM) reduce 106) ((T_EVAL) reduce 106) ((T_EXIT) reduce 106) ((T_FILE) reduce 106) ((T_FINAL) reduce 106) ((T_FN) reduce 106) ((T_FOR) reduce 106) ((T_FOREACH) reduce 106) ((T_FUNCTION) reduce 106) ((T_FUNC_C) reduce 106) ((T_GLOBAL) reduce 106) ((T_GOTO) reduce 106) ((T_HALT_COMPILER) reduce 106) ((T_IF) reduce 106) ((T_INC) reduce 106) ((T_INCLUDE) reduce 106) ((T_INCLUDE_ONCE) reduce 106) ((T_INLINE_HTML) reduce 106) ((T_INTERFACE) reduce 106) ((T_INT_CAST) reduce 106) ((T_ISSET) reduce 106) ((T_LINE) reduce 106) ((T_LIST) reduce 106) ((T_LNUMBER) reduce 106) ((T_MATCH) reduce 106) ((T_METHOD_C) reduce 106) ((T_NAME_FULLY_QUALIFIED) reduce 106) ((T_NAME_QUALIFIED) reduce 106) ((T_NAME_RELATIVE) reduce 106) ((T_NEW) reduce 106) ((T_NS_C) reduce 106) ((T_OBJECT_CAST) reduce 106) ((T_PRINT) reduce 106) ((T_PROPERTY_C) reduce 106) ((T_READONLY) reduce 106) ((T_REQUIRE) reduce 106) ((T_REQUIRE_ONCE) reduce 106) ((T_RETURN) reduce 106) ((T_START_HEREDOC) reduce 106) ((T_STATIC) reduce 106) ((T_STRING) reduce 106) ((T_STRING_CAST) reduce 106) ((T_SWITCH) reduce 106) ((T_THROW) reduce 106) ((T_TRAIT) reduce 106) ((T_TRAIT_C) reduce 106) ((T_TRY) reduce 106) ((T_UNSET) reduce 106) ((T_UNSET_CAST) reduce 106) ((T_VARIABLE) reduce 106) ((T_VOID_CAST) reduce 106) ((T_WHILE) reduce 106) ((T_YIELD) reduce 106) ((T_YIELD_FROM) reduce 106) (("[") reduce 106) (("`") reduce 106) (("{") reduce 106) (("~") reduce 106)) 17846 ((("!") reduce 143) (("\"") reduce 143) (("$") reduce 143) (("(") reduce 143) (("+") reduce 143) (("-") reduce 143) ((";") reduce 143) (("@") reduce 143) ((T_ABSTRACT) reduce 143) ((T_ARRAY) reduce 143) ((T_ARRAY_CAST) reduce 143) ((T_ATTRIBUTE) reduce 143) ((T_BOOL_CAST) reduce 143) ((T_BREAK) reduce 143) ((T_CLASS) reduce 143) ((T_CLASS_C) reduce 143) ((T_CLONE) reduce 143) ((T_CONSTANT_ENCAPSED_STRING) reduce 143) ((T_CONTINUE) reduce 143) ((T_DEC) reduce 143) ((T_DECLARE) reduce 143) ((T_DIR) reduce 143) ((T_DNUMBER) reduce 143) ((T_DO) reduce 143) ((T_DOUBLE_CAST) reduce 143) ((T_ECHO) reduce 143) ((T_EMPTY) reduce 143) ((T_ENDIF) reduce 143) ((T_ENUM) reduce 143) ((T_EVAL) reduce 143) ((T_EXIT) reduce 143) ((T_FILE) reduce 143) ((T_FINAL) reduce 143) ((T_FN) reduce 143) ((T_FOR) reduce 143) ((T_FOREACH) reduce 143) ((T_FUNCTION) reduce 143) ((T_FUNC_C) reduce 143) ((T_GLOBAL) reduce 143) ((T_GOTO) reduce 143) ((T_HALT_COMPILER) reduce 143) ((T_IF) reduce 143) ((T_INC) reduce 143) ((T_INCLUDE) reduce 143) ((T_INCLUDE_ONCE) reduce 143) ((T_INLINE_HTML) reduce 143) ((T_INTERFACE) reduce 143) ((T_INT_CAST) reduce 143) ((T_ISSET) reduce 143) ((T_LINE) reduce 143) ((T_LIST) reduce 143) ((T_LNUMBER) reduce 143) ((T_MATCH) reduce 143) ((T_METHOD_C) reduce 143) ((T_NAME_FULLY_QUALIFIED) reduce 143) ((T_NAME_QUALIFIED) reduce 143) ((T_NAME_RELATIVE) reduce 143) ((T_NEW) reduce 143) ((T_NS_C) reduce 143) ((T_OBJECT_CAST) reduce 143) ((T_PRINT) reduce 143) ((T_PROPERTY_C) reduce 143) ((T_READONLY) reduce 143) ((T_REQUIRE) reduce 143) ((T_REQUIRE_ONCE) reduce 143) ((T_RETURN) reduce 143) ((T_START_HEREDOC) reduce 143) ((T_STATIC) reduce 143) ((T_STRING) reduce 143) ((T_STRING_CAST) reduce 143) ((T_SWITCH) reduce 143) ((T_THROW) reduce 143) ((T_TRAIT) reduce 143) ((T_TRAIT_C) reduce 143) ((T_TRY) reduce 143) ((T_UNSET) reduce 143) ((T_UNSET_CAST) reduce 143) ((T_VARIABLE) reduce 143) ((T_VOID_CAST) reduce 143) ((T_WHILE) reduce 143) ((T_YIELD) reduce 143) ((T_YIELD_FROM) reduce 143) (("[") reduce 143) (("`") reduce 143) (("{") reduce 143) (("~") reduce 143)) 17847 ((("!") reduce 105) (("\"") reduce 105) (("$") reduce 105) (("(") reduce 105) (("+") reduce 105) (("-") reduce 105) ((";") reduce 105) (("@") reduce 105) ((T_ABSTRACT) reduce 105) ((T_ARRAY) reduce 105) ((T_ARRAY_CAST) reduce 105) ((T_ATTRIBUTE) reduce 105) ((T_BOOL_CAST) reduce 105) ((T_BREAK) reduce 105) ((T_CLASS) reduce 105) ((T_CLASS_C) reduce 105) ((T_CLONE) reduce 105) ((T_CONSTANT_ENCAPSED_STRING) reduce 105) ((T_CONTINUE) reduce 105) ((T_DEC) reduce 105) ((T_DECLARE) reduce 105) ((T_DIR) reduce 105) ((T_DNUMBER) reduce 105) ((T_DO) reduce 105) ((T_DOUBLE_CAST) reduce 105) ((T_ECHO) reduce 105) ((T_EMPTY) reduce 105) ((T_ENDIF) reduce 105) ((T_ENUM) reduce 105) ((T_EVAL) reduce 105) ((T_EXIT) reduce 105) ((T_FILE) reduce 105) ((T_FINAL) reduce 105) ((T_FN) reduce 105) ((T_FOR) reduce 105) ((T_FOREACH) reduce 105) ((T_FUNCTION) reduce 105) ((T_FUNC_C) reduce 105) ((T_GLOBAL) reduce 105) ((T_GOTO) reduce 105) ((T_HALT_COMPILER) reduce 105) ((T_IF) reduce 105) ((T_INC) reduce 105) ((T_INCLUDE) reduce 105) ((T_INCLUDE_ONCE) reduce 105) ((T_INLINE_HTML) reduce 105) ((T_INTERFACE) reduce 105) ((T_INT_CAST) reduce 105) ((T_ISSET) reduce 105) ((T_LINE) reduce 105) ((T_LIST) reduce 105) ((T_LNUMBER) reduce 105) ((T_MATCH) reduce 105) ((T_METHOD_C) reduce 105) ((T_NAME_FULLY_QUALIFIED) reduce 105) ((T_NAME_QUALIFIED) reduce 105) ((T_NAME_RELATIVE) reduce 105) ((T_NEW) reduce 105) ((T_NS_C) reduce 105) ((T_OBJECT_CAST) reduce 105) ((T_PRINT) reduce 105) ((T_PROPERTY_C) reduce 105) ((T_READONLY) reduce 105) ((T_REQUIRE) reduce 105) ((T_REQUIRE_ONCE) reduce 105) ((T_RETURN) reduce 105) ((T_START_HEREDOC) reduce 105) ((T_STATIC) reduce 105) ((T_STRING) reduce 105) ((T_STRING_CAST) reduce 105) ((T_SWITCH) reduce 105) ((T_THROW) reduce 105) ((T_TRAIT) reduce 105) ((T_TRAIT_C) reduce 105) ((T_TRY) reduce 105) ((T_UNSET) reduce 105) ((T_UNSET_CAST) reduce 105) ((T_VARIABLE) reduce 105) ((T_VOID_CAST) reduce 105) ((T_WHILE) reduce 105) ((T_YIELD) reduce 105) ((T_YIELD_FROM) reduce 105) (("[") reduce 105) (("`") reduce 105) (("{") reduce 105) (("~") reduce 105)) 17850 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_EMPTY) reduce 147) ((T_ENDIF) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_VOID_CAST) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 17852 ((("!") reduce 247) (("\"") reduce 247) (("$") reduce 247) (("(") reduce 247) (("+") reduce 247) (("-") reduce 247) ((";") reduce 247) (("@") reduce 247) ((T_ABSTRACT) reduce 247) ((T_ARRAY) reduce 247) ((T_ARRAY_CAST) reduce 247) ((T_ATTRIBUTE) reduce 247) ((T_BOOL_CAST) reduce 247) ((T_BREAK) reduce 247) ((T_CLASS) reduce 247) ((T_CLASS_C) reduce 247) ((T_CLONE) reduce 247) ((T_CONSTANT_ENCAPSED_STRING) reduce 247) ((T_CONTINUE) reduce 247) ((T_DEC) reduce 247) ((T_DECLARE) reduce 247) ((T_DIR) reduce 247) ((T_DNUMBER) reduce 247) ((T_DO) reduce 247) ((T_DOUBLE_CAST) reduce 247) ((T_ECHO) reduce 247) ((T_EMPTY) reduce 247) ((T_ENDIF) reduce 247) ((T_ENUM) reduce 247) ((T_EVAL) reduce 247) ((T_EXIT) reduce 247) ((T_FILE) reduce 247) ((T_FINAL) reduce 247) ((T_FN) reduce 247) ((T_FOR) reduce 247) ((T_FOREACH) reduce 247) ((T_FUNCTION) reduce 247) ((T_FUNC_C) reduce 247) ((T_GLOBAL) reduce 247) ((T_GOTO) reduce 247) ((T_HALT_COMPILER) reduce 247) ((T_IF) reduce 247) ((T_INC) reduce 247) ((T_INCLUDE) reduce 247) ((T_INCLUDE_ONCE) reduce 247) ((T_INLINE_HTML) reduce 247) ((T_INTERFACE) reduce 247) ((T_INT_CAST) reduce 247) ((T_ISSET) reduce 247) ((T_LINE) reduce 247) ((T_LIST) reduce 247) ((T_LNUMBER) reduce 247) ((T_MATCH) reduce 247) ((T_METHOD_C) reduce 247) ((T_NAME_FULLY_QUALIFIED) reduce 247) ((T_NAME_QUALIFIED) reduce 247) ((T_NAME_RELATIVE) reduce 247) ((T_NEW) reduce 247) ((T_NS_C) reduce 247) ((T_OBJECT_CAST) reduce 247) ((T_PRINT) reduce 247) ((T_PROPERTY_C) reduce 247) ((T_READONLY) reduce 247) ((T_REQUIRE) reduce 247) ((T_REQUIRE_ONCE) reduce 247) ((T_RETURN) reduce 247) ((T_START_HEREDOC) reduce 247) ((T_STATIC) reduce 247) ((T_STRING) reduce 247) ((T_STRING_CAST) reduce 247) ((T_SWITCH) reduce 247) ((T_THROW) reduce 247) ((T_TRAIT) reduce 247) ((T_TRAIT_C) reduce 247) ((T_TRY) reduce 247) ((T_UNSET) reduce 247) ((T_UNSET_CAST) reduce 247) ((T_VARIABLE) reduce 247) ((T_VOID_CAST) reduce 247) ((T_WHILE) reduce 247) ((T_YIELD) reduce 247) ((T_YIELD_FROM) reduce 247) (("[") reduce 247) (("`") reduce 247) (("{") reduce 247) (("~") reduce 247)) 17863 ((("!") reduce 494) (("\"") reduce 494) (("$") reduce 494) (("(") reduce 494) (("+") reduce 494) (("-") reduce 494) ((";") reduce 494) (("@") reduce 494) ((T_ABSTRACT) reduce 494) ((T_ARRAY) reduce 494) ((T_ARRAY_CAST) reduce 494) ((T_ATTRIBUTE) reduce 494) ((T_BOOL_CAST) reduce 494) ((T_BREAK) reduce 494) ((T_CLASS) reduce 494) ((T_CLASS_C) reduce 494) ((T_CLONE) reduce 494) ((T_CONSTANT_ENCAPSED_STRING) reduce 494) ((T_CONTINUE) reduce 494) ((T_DEC) reduce 494) ((T_DECLARE) reduce 494) ((T_DIR) reduce 494) ((T_DNUMBER) reduce 494) ((T_DO) reduce 494) ((T_DOUBLE_CAST) reduce 494) ((T_ECHO) reduce 494) ((T_EMPTY) reduce 494) ((T_ENDIF) reduce 494) ((T_ENUM) reduce 494) ((T_EVAL) reduce 494) ((T_EXIT) reduce 494) ((T_FILE) reduce 494) ((T_FINAL) reduce 494) ((T_FN) reduce 494) ((T_FOR) reduce 494) ((T_FOREACH) reduce 494) ((T_FUNCTION) reduce 494) ((T_FUNC_C) reduce 494) ((T_GLOBAL) reduce 494) ((T_GOTO) reduce 494) ((T_HALT_COMPILER) reduce 494) ((T_IF) reduce 494) ((T_INC) reduce 494) ((T_INCLUDE) reduce 494) ((T_INCLUDE_ONCE) reduce 494) ((T_INLINE_HTML) reduce 494) ((T_INTERFACE) reduce 494) ((T_INT_CAST) reduce 494) ((T_ISSET) reduce 494) ((T_LINE) reduce 494) ((T_LIST) reduce 494) ((T_LNUMBER) reduce 494) ((T_MATCH) reduce 494) ((T_METHOD_C) reduce 494) ((T_NAME_FULLY_QUALIFIED) reduce 494) ((T_NAME_QUALIFIED) reduce 494) ((T_NAME_RELATIVE) reduce 494) ((T_NEW) reduce 494) ((T_NS_C) reduce 494) ((T_OBJECT_CAST) reduce 494) ((T_PRINT) reduce 494) ((T_PROPERTY_C) reduce 494) ((T_READONLY) reduce 494) ((T_REQUIRE) reduce 494) ((T_REQUIRE_ONCE) reduce 494) ((T_RETURN) reduce 494) ((T_START_HEREDOC) reduce 494) ((T_STATIC) reduce 494) ((T_STRING) reduce 494) ((T_STRING_CAST) reduce 494) ((T_SWITCH) reduce 494) ((T_THROW) reduce 494) ((T_TRAIT) reduce 494) ((T_TRAIT_C) reduce 494) ((T_TRY) reduce 494) ((T_UNSET) reduce 494) ((T_UNSET_CAST) reduce 494) ((T_VARIABLE) reduce 494) ((T_VOID_CAST) reduce 494) ((T_WHILE) reduce 494) ((T_YIELD) reduce 494) ((T_YIELD_FROM) reduce 494) (("[") reduce 494) (("`") reduce 494) (("{") reduce 494) (("~") reduce 494)) 17864 ((("!") reduce 184) (("\"") reduce 184) (("$") reduce 184) (("(") reduce 184) (("+") reduce 184) (("-") reduce 184) ((";") reduce 184) (("@") reduce 184) ((T_ABSTRACT) reduce 184) ((T_ARRAY) reduce 184) ((T_ARRAY_CAST) reduce 184) ((T_ATTRIBUTE) reduce 184) ((T_BOOL_CAST) reduce 184) ((T_BREAK) reduce 184) ((T_CLASS) reduce 184) ((T_CLASS_C) reduce 184) ((T_CLONE) reduce 184) ((T_CONSTANT_ENCAPSED_STRING) reduce 184) ((T_CONTINUE) reduce 184) ((T_DEC) reduce 184) ((T_DECLARE) reduce 184) ((T_DIR) reduce 184) ((T_DNUMBER) reduce 184) ((T_DO) reduce 184) ((T_DOUBLE_CAST) reduce 184) ((T_ECHO) reduce 184) ((T_EMPTY) reduce 184) ((T_ENDIF) reduce 184) ((T_ENUM) reduce 184) ((T_EVAL) reduce 184) ((T_EXIT) reduce 184) ((T_FILE) reduce 184) ((T_FINAL) reduce 184) ((T_FN) reduce 184) ((T_FOR) reduce 184) ((T_FOREACH) reduce 184) ((T_FUNCTION) reduce 184) ((T_FUNC_C) reduce 184) ((T_GLOBAL) reduce 184) ((T_GOTO) reduce 184) ((T_HALT_COMPILER) reduce 184) ((T_IF) reduce 184) ((T_INC) reduce 184) ((T_INCLUDE) reduce 184) ((T_INCLUDE_ONCE) reduce 184) ((T_INLINE_HTML) reduce 184) ((T_INTERFACE) reduce 184) ((T_INT_CAST) reduce 184) ((T_ISSET) reduce 184) ((T_LINE) reduce 184) ((T_LIST) reduce 184) ((T_LNUMBER) reduce 184) ((T_MATCH) reduce 184) ((T_METHOD_C) reduce 184) ((T_NAME_FULLY_QUALIFIED) reduce 184) ((T_NAME_QUALIFIED) reduce 184) ((T_NAME_RELATIVE) reduce 184) ((T_NEW) reduce 184) ((T_NS_C) reduce 184) ((T_OBJECT_CAST) reduce 184) ((T_PRINT) reduce 184) ((T_PROPERTY_C) reduce 184) ((T_READONLY) reduce 184) ((T_REQUIRE) reduce 184) ((T_REQUIRE_ONCE) reduce 184) ((T_RETURN) reduce 184) ((T_START_HEREDOC) reduce 184) ((T_STATIC) reduce 184) ((T_STRING) reduce 184) ((T_STRING_CAST) reduce 184) ((T_SWITCH) reduce 184) ((T_THROW) reduce 184) ((T_TRAIT) reduce 184) ((T_TRAIT_C) reduce 184) ((T_TRY) reduce 184) ((T_UNSET) reduce 184) ((T_UNSET_CAST) reduce 184) ((T_VARIABLE) reduce 184) ((T_VOID_CAST) reduce 184) ((T_WHILE) reduce 184) ((T_YIELD) reduce 184) ((T_YIELD_FROM) reduce 184) (("[") reduce 184) (("`") reduce 184) (("{") reduce 184) (("~") reduce 184)) 17865 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_EMPTY) reduce 161) ((T_ENDIF) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_VOID_CAST) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 17873 ((("!") reduce 189) (("\"") reduce 189) (("$") reduce 189) (("(") reduce 189) (("+") reduce 189) (("-") reduce 189) ((";") reduce 189) (("@") reduce 189) ((T_ABSTRACT) reduce 189) ((T_ARRAY) reduce 189) ((T_ARRAY_CAST) reduce 189) ((T_ATTRIBUTE) reduce 189) ((T_BOOL_CAST) reduce 189) ((T_BREAK) reduce 189) ((T_CLASS) reduce 189) ((T_CLASS_C) reduce 189) ((T_CLONE) reduce 189) ((T_CONSTANT_ENCAPSED_STRING) reduce 189) ((T_CONTINUE) reduce 189) ((T_DEC) reduce 189) ((T_DECLARE) reduce 189) ((T_DIR) reduce 189) ((T_DNUMBER) reduce 189) ((T_DO) reduce 189) ((T_DOUBLE_CAST) reduce 189) ((T_ECHO) reduce 189) ((T_EMPTY) reduce 189) ((T_ENDIF) reduce 189) ((T_ENUM) reduce 189) ((T_EVAL) reduce 189) ((T_EXIT) reduce 189) ((T_FILE) reduce 189) ((T_FINAL) reduce 189) ((T_FN) reduce 189) ((T_FOR) reduce 189) ((T_FOREACH) reduce 189) ((T_FUNCTION) reduce 189) ((T_FUNC_C) reduce 189) ((T_GLOBAL) reduce 189) ((T_GOTO) reduce 189) ((T_HALT_COMPILER) reduce 189) ((T_IF) reduce 189) ((T_INC) reduce 189) ((T_INCLUDE) reduce 189) ((T_INCLUDE_ONCE) reduce 189) ((T_INLINE_HTML) reduce 189) ((T_INTERFACE) reduce 189) ((T_INT_CAST) reduce 189) ((T_ISSET) reduce 189) ((T_LINE) reduce 189) ((T_LIST) reduce 189) ((T_LNUMBER) reduce 189) ((T_MATCH) reduce 189) ((T_METHOD_C) reduce 189) ((T_NAME_FULLY_QUALIFIED) reduce 189) ((T_NAME_QUALIFIED) reduce 189) ((T_NAME_RELATIVE) reduce 189) ((T_NEW) reduce 189) ((T_NS_C) reduce 189) ((T_OBJECT_CAST) reduce 189) ((T_PRINT) reduce 189) ((T_PROPERTY_C) reduce 189) ((T_READONLY) reduce 189) ((T_REQUIRE) reduce 189) ((T_REQUIRE_ONCE) reduce 189) ((T_RETURN) reduce 189) ((T_START_HEREDOC) reduce 189) ((T_STATIC) reduce 189) ((T_STRING) reduce 189) ((T_STRING_CAST) reduce 189) ((T_SWITCH) reduce 189) ((T_THROW) reduce 189) ((T_TRAIT) reduce 189) ((T_TRAIT_C) reduce 189) ((T_TRY) reduce 189) ((T_UNSET) reduce 189) ((T_UNSET_CAST) reduce 189) ((T_VARIABLE) reduce 189) ((T_VOID_CAST) reduce 189) ((T_WHILE) reduce 189) ((T_YIELD) reduce 189) ((T_YIELD_FROM) reduce 189) (("[") reduce 189) (("`") reduce 189) (("{") reduce 189) (("~") reduce 189)) 17874 ((("!") reduce 145) (("\"") reduce 145) (("$") reduce 145) (("(") reduce 145) (("+") reduce 145) (("-") reduce 145) ((";") reduce 145) (("@") reduce 145) ((T_ABSTRACT) reduce 145) ((T_ARRAY) reduce 145) ((T_ARRAY_CAST) reduce 145) ((T_ATTRIBUTE) reduce 145) ((T_BOOL_CAST) reduce 145) ((T_BREAK) reduce 145) ((T_CLASS) reduce 145) ((T_CLASS_C) reduce 145) ((T_CLONE) reduce 145) ((T_CONSTANT_ENCAPSED_STRING) reduce 145) ((T_CONTINUE) reduce 145) ((T_DEC) reduce 145) ((T_DECLARE) reduce 145) ((T_DIR) reduce 145) ((T_DNUMBER) reduce 145) ((T_DO) reduce 145) ((T_DOUBLE_CAST) reduce 145) ((T_ECHO) reduce 145) ((T_EMPTY) reduce 145) ((T_ENDIF) reduce 145) ((T_ENUM) reduce 145) ((T_EVAL) reduce 145) ((T_EXIT) reduce 145) ((T_FILE) reduce 145) ((T_FINAL) reduce 145) ((T_FN) reduce 145) ((T_FOR) reduce 145) ((T_FOREACH) reduce 145) ((T_FUNCTION) reduce 145) ((T_FUNC_C) reduce 145) ((T_GLOBAL) reduce 145) ((T_GOTO) reduce 145) ((T_HALT_COMPILER) reduce 145) ((T_IF) reduce 145) ((T_INC) reduce 145) ((T_INCLUDE) reduce 145) ((T_INCLUDE_ONCE) reduce 145) ((T_INLINE_HTML) reduce 145) ((T_INTERFACE) reduce 145) ((T_INT_CAST) reduce 145) ((T_ISSET) reduce 145) ((T_LINE) reduce 145) ((T_LIST) reduce 145) ((T_LNUMBER) reduce 145) ((T_MATCH) reduce 145) ((T_METHOD_C) reduce 145) ((T_NAME_FULLY_QUALIFIED) reduce 145) ((T_NAME_QUALIFIED) reduce 145) ((T_NAME_RELATIVE) reduce 145) ((T_NEW) reduce 145) ((T_NS_C) reduce 145) ((T_OBJECT_CAST) reduce 145) ((T_PRINT) reduce 145) ((T_PROPERTY_C) reduce 145) ((T_READONLY) reduce 145) ((T_REQUIRE) reduce 145) ((T_REQUIRE_ONCE) reduce 145) ((T_RETURN) reduce 145) ((T_START_HEREDOC) reduce 145) ((T_STATIC) reduce 145) ((T_STRING) reduce 145) ((T_STRING_CAST) reduce 145) ((T_SWITCH) reduce 145) ((T_THROW) reduce 145) ((T_TRAIT) reduce 145) ((T_TRAIT_C) reduce 145) ((T_TRY) reduce 145) ((T_UNSET) reduce 145) ((T_UNSET_CAST) reduce 145) ((T_VARIABLE) reduce 145) ((T_VOID_CAST) reduce 145) ((T_WHILE) reduce 145) ((T_YIELD) reduce 145) ((T_YIELD_FROM) reduce 145) (("[") reduce 145) (("`") reduce 145) (("{") reduce 145) (("~") reduce 145)) 17877 ((("!") reduce 250) (("\"") reduce 250) (("$") reduce 250) (("(") reduce 250) (("+") reduce 250) (("-") reduce 250) ((";") reduce 250) (("@") reduce 250) ((T_ABSTRACT) reduce 250) ((T_ARRAY) reduce 250) ((T_ARRAY_CAST) reduce 250) ((T_ATTRIBUTE) reduce 250) ((T_BOOL_CAST) reduce 250) ((T_BREAK) reduce 250) ((T_CLASS) reduce 250) ((T_CLASS_C) reduce 250) ((T_CLONE) reduce 250) ((T_CONSTANT_ENCAPSED_STRING) reduce 250) ((T_CONTINUE) reduce 250) ((T_DEC) reduce 250) ((T_DECLARE) reduce 250) ((T_DIR) reduce 250) ((T_DNUMBER) reduce 250) ((T_DO) reduce 250) ((T_DOUBLE_CAST) reduce 250) ((T_ECHO) reduce 250) ((T_EMPTY) reduce 250) ((T_ENDIF) reduce 250) ((T_ENUM) reduce 250) ((T_EVAL) reduce 250) ((T_EXIT) reduce 250) ((T_FILE) reduce 250) ((T_FINAL) reduce 250) ((T_FN) reduce 250) ((T_FOR) reduce 250) ((T_FOREACH) reduce 250) ((T_FUNCTION) reduce 250) ((T_FUNC_C) reduce 250) ((T_GLOBAL) reduce 250) ((T_GOTO) reduce 250) ((T_HALT_COMPILER) reduce 250) ((T_IF) reduce 250) ((T_INC) reduce 250) ((T_INCLUDE) reduce 250) ((T_INCLUDE_ONCE) reduce 250) ((T_INLINE_HTML) reduce 250) ((T_INTERFACE) reduce 250) ((T_INT_CAST) reduce 250) ((T_ISSET) reduce 250) ((T_LINE) reduce 250) ((T_LIST) reduce 250) ((T_LNUMBER) reduce 250) ((T_MATCH) reduce 250) ((T_METHOD_C) reduce 250) ((T_NAME_FULLY_QUALIFIED) reduce 250) ((T_NAME_QUALIFIED) reduce 250) ((T_NAME_RELATIVE) reduce 250) ((T_NEW) reduce 250) ((T_NS_C) reduce 250) ((T_OBJECT_CAST) reduce 250) ((T_PRINT) reduce 250) ((T_PROPERTY_C) reduce 250) ((T_READONLY) reduce 250) ((T_REQUIRE) reduce 250) ((T_REQUIRE_ONCE) reduce 250) ((T_RETURN) reduce 250) ((T_START_HEREDOC) reduce 250) ((T_STATIC) reduce 250) ((T_STRING) reduce 250) ((T_STRING_CAST) reduce 250) ((T_SWITCH) reduce 250) ((T_THROW) reduce 250) ((T_TRAIT) reduce 250) ((T_TRAIT_C) reduce 250) ((T_TRY) reduce 250) ((T_UNSET) reduce 250) ((T_UNSET_CAST) reduce 250) ((T_VARIABLE) reduce 250) ((T_VOID_CAST) reduce 250) ((T_WHILE) reduce 250) ((T_YIELD) reduce 250) ((T_YIELD_FROM) reduce 250) (("[") reduce 250) (("`") reduce 250) (("{") reduce 250) (("~") reduce 250)) 17881 ((("!") reduce 251) (("\"") reduce 251) (("$") reduce 251) (("(") reduce 251) (("+") reduce 251) (("-") reduce 251) ((";") reduce 251) (("@") reduce 251) ((T_ABSTRACT) reduce 251) ((T_ARRAY) reduce 251) ((T_ARRAY_CAST) reduce 251) ((T_ATTRIBUTE) reduce 251) ((T_BOOL_CAST) reduce 251) ((T_BREAK) reduce 251) ((T_CLASS) reduce 251) ((T_CLASS_C) reduce 251) ((T_CLONE) reduce 251) ((T_CONSTANT_ENCAPSED_STRING) reduce 251) ((T_CONTINUE) reduce 251) ((T_DEC) reduce 251) ((T_DECLARE) reduce 251) ((T_DIR) reduce 251) ((T_DNUMBER) reduce 251) ((T_DO) reduce 251) ((T_DOUBLE_CAST) reduce 251) ((T_ECHO) reduce 251) ((T_EMPTY) reduce 251) ((T_ENDIF) reduce 251) ((T_ENUM) reduce 251) ((T_EVAL) reduce 251) ((T_EXIT) reduce 251) ((T_FILE) reduce 251) ((T_FINAL) reduce 251) ((T_FN) reduce 251) ((T_FOR) reduce 251) ((T_FOREACH) reduce 251) ((T_FUNCTION) reduce 251) ((T_FUNC_C) reduce 251) ((T_GLOBAL) reduce 251) ((T_GOTO) reduce 251) ((T_HALT_COMPILER) reduce 251) ((T_IF) reduce 251) ((T_INC) reduce 251) ((T_INCLUDE) reduce 251) ((T_INCLUDE_ONCE) reduce 251) ((T_INLINE_HTML) reduce 251) ((T_INTERFACE) reduce 251) ((T_INT_CAST) reduce 251) ((T_ISSET) reduce 251) ((T_LINE) reduce 251) ((T_LIST) reduce 251) ((T_LNUMBER) reduce 251) ((T_MATCH) reduce 251) ((T_METHOD_C) reduce 251) ((T_NAME_FULLY_QUALIFIED) reduce 251) ((T_NAME_QUALIFIED) reduce 251) ((T_NAME_RELATIVE) reduce 251) ((T_NEW) reduce 251) ((T_NS_C) reduce 251) ((T_OBJECT_CAST) reduce 251) ((T_PRINT) reduce 251) ((T_PROPERTY_C) reduce 251) ((T_READONLY) reduce 251) ((T_REQUIRE) reduce 251) ((T_REQUIRE_ONCE) reduce 251) ((T_RETURN) reduce 251) ((T_START_HEREDOC) reduce 251) ((T_STATIC) reduce 251) ((T_STRING) reduce 251) ((T_STRING_CAST) reduce 251) ((T_SWITCH) reduce 251) ((T_THROW) reduce 251) ((T_TRAIT) reduce 251) ((T_TRAIT_C) reduce 251) ((T_TRY) reduce 251) ((T_UNSET) reduce 251) ((T_UNSET_CAST) reduce 251) ((T_VARIABLE) reduce 251) ((T_VOID_CAST) reduce 251) ((T_WHILE) reduce 251) ((T_YIELD) reduce 251) ((T_YIELD_FROM) reduce 251) (("[") reduce 251) (("`") reduce 251) (("{") reduce 251) (("~") reduce 251)) 17884 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((":") shift) ((";") shift) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_VOID_CAST) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("~") shift)) 17885 ((("!") reduce 142) (("\"") reduce 142) (("$") reduce 142) (("(") reduce 142) (("+") reduce 142) (("-") reduce 142) ((";") reduce 142) (("@") reduce 142) ((T_ABSTRACT) reduce 142) ((T_ARRAY) reduce 142) ((T_ARRAY_CAST) reduce 142) ((T_ATTRIBUTE) reduce 142) ((T_BOOL_CAST) reduce 142) ((T_BREAK) reduce 142) ((T_CLASS) reduce 142) ((T_CLASS_C) reduce 142) ((T_CLONE) reduce 142) ((T_CONSTANT_ENCAPSED_STRING) reduce 142) ((T_CONTINUE) reduce 142) ((T_DEC) reduce 142) ((T_DECLARE) reduce 142) ((T_DIR) reduce 142) ((T_DNUMBER) reduce 142) ((T_DO) reduce 142) ((T_DOUBLE_CAST) reduce 142) ((T_ECHO) reduce 142) ((T_EMPTY) reduce 142) ((T_ENDWHILE) reduce 142) ((T_ENUM) reduce 142) ((T_EVAL) reduce 142) ((T_EXIT) reduce 142) ((T_FILE) reduce 142) ((T_FINAL) reduce 142) ((T_FN) reduce 142) ((T_FOR) reduce 142) ((T_FOREACH) reduce 142) ((T_FUNCTION) reduce 142) ((T_FUNC_C) reduce 142) ((T_GLOBAL) reduce 142) ((T_GOTO) reduce 142) ((T_HALT_COMPILER) reduce 142) ((T_IF) reduce 142) ((T_INC) reduce 142) ((T_INCLUDE) reduce 142) ((T_INCLUDE_ONCE) reduce 142) ((T_INLINE_HTML) reduce 142) ((T_INTERFACE) reduce 142) ((T_INT_CAST) reduce 142) ((T_ISSET) reduce 142) ((T_LINE) reduce 142) ((T_LIST) reduce 142) ((T_LNUMBER) reduce 142) ((T_MATCH) reduce 142) ((T_METHOD_C) reduce 142) ((T_NAME_FULLY_QUALIFIED) reduce 142) ((T_NAME_QUALIFIED) reduce 142) ((T_NAME_RELATIVE) reduce 142) ((T_NEW) reduce 142) ((T_NS_C) reduce 142) ((T_OBJECT_CAST) reduce 142) ((T_PRINT) reduce 142) ((T_PROPERTY_C) reduce 142) ((T_READONLY) reduce 142) ((T_REQUIRE) reduce 142) ((T_REQUIRE_ONCE) reduce 142) ((T_RETURN) reduce 142) ((T_START_HEREDOC) reduce 142) ((T_STATIC) reduce 142) ((T_STRING) reduce 142) ((T_STRING_CAST) reduce 142) ((T_SWITCH) reduce 142) ((T_THROW) reduce 142) ((T_TRAIT) reduce 142) ((T_TRAIT_C) reduce 142) ((T_TRY) reduce 142) ((T_UNSET) reduce 142) ((T_UNSET_CAST) reduce 142) ((T_VARIABLE) reduce 142) ((T_VOID_CAST) reduce 142) ((T_WHILE) reduce 142) ((T_YIELD) reduce 142) ((T_YIELD_FROM) reduce 142) (("[") reduce 142) (("`") reduce 142) (("{") reduce 142) (("~") reduce 142)) 17886 ((("!") reduce 242) (("\"") reduce 242) (("$") reduce 242) (("(") reduce 242) (("+") reduce 242) (("-") reduce 242) ((";") reduce 242) (("@") reduce 242) ((T_ABSTRACT) reduce 242) ((T_ARRAY) reduce 242) ((T_ARRAY_CAST) reduce 242) ((T_ATTRIBUTE) reduce 242) ((T_BOOL_CAST) reduce 242) ((T_BREAK) reduce 242) ((T_CLASS) reduce 242) ((T_CLASS_C) reduce 242) ((T_CLONE) reduce 242) ((T_CONSTANT_ENCAPSED_STRING) reduce 242) ((T_CONTINUE) reduce 242) ((T_DEC) reduce 242) ((T_DECLARE) reduce 242) ((T_DIR) reduce 242) ((T_DNUMBER) reduce 242) ((T_DO) reduce 242) ((T_DOUBLE_CAST) reduce 242) ((T_ECHO) reduce 242) ((T_EMPTY) reduce 242) ((T_ENDIF) reduce 242) ((T_ENUM) reduce 242) ((T_EVAL) reduce 242) ((T_EXIT) reduce 242) ((T_FILE) reduce 242) ((T_FINAL) reduce 242) ((T_FN) reduce 242) ((T_FOR) reduce 242) ((T_FOREACH) reduce 242) ((T_FUNCTION) reduce 242) ((T_FUNC_C) reduce 242) ((T_GLOBAL) reduce 242) ((T_GOTO) reduce 242) ((T_HALT_COMPILER) reduce 242) ((T_IF) reduce 242) ((T_INC) reduce 242) ((T_INCLUDE) reduce 242) ((T_INCLUDE_ONCE) reduce 242) ((T_INLINE_HTML) reduce 242) ((T_INTERFACE) reduce 242) ((T_INT_CAST) reduce 242) ((T_ISSET) reduce 242) ((T_LINE) reduce 242) ((T_LIST) reduce 242) ((T_LNUMBER) reduce 242) ((T_MATCH) reduce 242) ((T_METHOD_C) reduce 242) ((T_NAME_FULLY_QUALIFIED) reduce 242) ((T_NAME_QUALIFIED) reduce 242) ((T_NAME_RELATIVE) reduce 242) ((T_NEW) reduce 242) ((T_NS_C) reduce 242) ((T_OBJECT_CAST) reduce 242) ((T_PRINT) reduce 242) ((T_PROPERTY_C) reduce 242) ((T_READONLY) reduce 242) ((T_REQUIRE) reduce 242) ((T_REQUIRE_ONCE) reduce 242) ((T_RETURN) reduce 242) ((T_START_HEREDOC) reduce 242) ((T_STATIC) reduce 242) ((T_STRING) reduce 242) ((T_STRING_CAST) reduce 242) ((T_SWITCH) reduce 242) ((T_THROW) reduce 242) ((T_TRAIT) reduce 242) ((T_TRAIT_C) reduce 242) ((T_TRY) reduce 242) ((T_UNSET) reduce 242) ((T_UNSET_CAST) reduce 242) ((T_VARIABLE) reduce 242) ((T_VOID_CAST) reduce 242) ((T_WHILE) reduce 242) ((T_YIELD) reduce 242) ((T_YIELD_FROM) reduce 242) (("[") reduce 242) (("`") reduce 242) (("{") reduce 242) (("~") reduce 242)) 17887 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_EMPTY) reduce 150) ((T_ENDIF) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_VOID_CAST) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 17888 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_ENDWHILE) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_VOID_CAST) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("~") shift)) 17889 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_EMPTY) reduce 166) ((T_ENDWHILE) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_VOID_CAST) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 17904 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_EMPTY) reduce 160) ((T_ENDWHILE) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_VOID_CAST) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 17915 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_EMPTY) reduce 149) ((T_ENDWHILE) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_VOID_CAST) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 17917 ((("!") reduce 144) (("\"") reduce 144) (("$") reduce 144) (("(") reduce 144) (("+") reduce 144) (("-") reduce 144) ((";") reduce 144) (("@") reduce 144) ((T_ABSTRACT) reduce 144) ((T_ARRAY) reduce 144) ((T_ARRAY_CAST) reduce 144) ((T_ATTRIBUTE) reduce 144) ((T_BOOL_CAST) reduce 144) ((T_BREAK) reduce 144) ((T_CLASS) reduce 144) ((T_CLASS_C) reduce 144) ((T_CLONE) reduce 144) ((T_CONSTANT_ENCAPSED_STRING) reduce 144) ((T_CONTINUE) reduce 144) ((T_DEC) reduce 144) ((T_DECLARE) reduce 144) ((T_DIR) reduce 144) ((T_DNUMBER) reduce 144) ((T_DO) reduce 144) ((T_DOUBLE_CAST) reduce 144) ((T_ECHO) reduce 144) ((T_EMPTY) reduce 144) ((T_ENDWHILE) reduce 144) ((T_ENUM) reduce 144) ((T_EVAL) reduce 144) ((T_EXIT) reduce 144) ((T_FILE) reduce 144) ((T_FINAL) reduce 144) ((T_FN) reduce 144) ((T_FOR) reduce 144) ((T_FOREACH) reduce 144) ((T_FUNCTION) reduce 144) ((T_FUNC_C) reduce 144) ((T_GLOBAL) reduce 144) ((T_GOTO) reduce 144) ((T_HALT_COMPILER) reduce 144) ((T_IF) reduce 144) ((T_INC) reduce 144) ((T_INCLUDE) reduce 144) ((T_INCLUDE_ONCE) reduce 144) ((T_INLINE_HTML) reduce 144) ((T_INTERFACE) reduce 144) ((T_INT_CAST) reduce 144) ((T_ISSET) reduce 144) ((T_LINE) reduce 144) ((T_LIST) reduce 144) ((T_LNUMBER) reduce 144) ((T_MATCH) reduce 144) ((T_METHOD_C) reduce 144) ((T_NAME_FULLY_QUALIFIED) reduce 144) ((T_NAME_QUALIFIED) reduce 144) ((T_NAME_RELATIVE) reduce 144) ((T_NEW) reduce 144) ((T_NS_C) reduce 144) ((T_OBJECT_CAST) reduce 144) ((T_PRINT) reduce 144) ((T_PROPERTY_C) reduce 144) ((T_READONLY) reduce 144) ((T_REQUIRE) reduce 144) ((T_REQUIRE_ONCE) reduce 144) ((T_RETURN) reduce 144) ((T_START_HEREDOC) reduce 144) ((T_STATIC) reduce 144) ((T_STRING) reduce 144) ((T_STRING_CAST) reduce 144) ((T_SWITCH) reduce 144) ((T_THROW) reduce 144) ((T_TRAIT) reduce 144) ((T_TRAIT_C) reduce 144) ((T_TRY) reduce 144) ((T_UNSET) reduce 144) ((T_UNSET_CAST) reduce 144) ((T_VARIABLE) reduce 144) ((T_VOID_CAST) reduce 144) ((T_WHILE) reduce 144) ((T_YIELD) reduce 144) ((T_YIELD_FROM) reduce 144) (("[") reduce 144) (("`") reduce 144) (("{") reduce 144) (("~") reduce 144)) 17919 ((("!") reduce 104) (("\"") reduce 104) (("$") reduce 104) (("(") reduce 104) (("+") reduce 104) (("-") reduce 104) ((";") reduce 104) (("@") reduce 104) ((T_ABSTRACT) reduce 104) ((T_ARRAY) reduce 104) ((T_ARRAY_CAST) reduce 104) ((T_ATTRIBUTE) reduce 104) ((T_BOOL_CAST) reduce 104) ((T_BREAK) reduce 104) ((T_CLASS) reduce 104) ((T_CLASS_C) reduce 104) ((T_CLONE) reduce 104) ((T_CONSTANT_ENCAPSED_STRING) reduce 104) ((T_CONTINUE) reduce 104) ((T_DEC) reduce 104) ((T_DECLARE) reduce 104) ((T_DIR) reduce 104) ((T_DNUMBER) reduce 104) ((T_DO) reduce 104) ((T_DOUBLE_CAST) reduce 104) ((T_ECHO) reduce 104) ((T_EMPTY) reduce 104) ((T_ENDWHILE) reduce 104) ((T_ENUM) reduce 104) ((T_EVAL) reduce 104) ((T_EXIT) reduce 104) ((T_FILE) reduce 104) ((T_FINAL) reduce 104) ((T_FN) reduce 104) ((T_FOR) reduce 104) ((T_FOREACH) reduce 104) ((T_FUNCTION) reduce 104) ((T_FUNC_C) reduce 104) ((T_GLOBAL) reduce 104) ((T_GOTO) reduce 104) ((T_HALT_COMPILER) reduce 104) ((T_IF) reduce 104) ((T_INC) reduce 104) ((T_INCLUDE) reduce 104) ((T_INCLUDE_ONCE) reduce 104) ((T_INLINE_HTML) reduce 104) ((T_INTERFACE) reduce 104) ((T_INT_CAST) reduce 104) ((T_ISSET) reduce 104) ((T_LINE) reduce 104) ((T_LIST) reduce 104) ((T_LNUMBER) reduce 104) ((T_MATCH) reduce 104) ((T_METHOD_C) reduce 104) ((T_NAME_FULLY_QUALIFIED) reduce 104) ((T_NAME_QUALIFIED) reduce 104) ((T_NAME_RELATIVE) reduce 104) ((T_NEW) reduce 104) ((T_NS_C) reduce 104) ((T_OBJECT_CAST) reduce 104) ((T_PRINT) reduce 104) ((T_PROPERTY_C) reduce 104) ((T_READONLY) reduce 104) ((T_REQUIRE) reduce 104) ((T_REQUIRE_ONCE) reduce 104) ((T_RETURN) reduce 104) ((T_START_HEREDOC) reduce 104) ((T_STATIC) reduce 104) ((T_STRING) reduce 104) ((T_STRING_CAST) reduce 104) ((T_SWITCH) reduce 104) ((T_THROW) reduce 104) ((T_TRAIT) reduce 104) ((T_TRAIT_C) reduce 104) ((T_TRY) reduce 104) ((T_UNSET) reduce 104) ((T_UNSET_CAST) reduce 104) ((T_VARIABLE) reduce 104) ((T_VOID_CAST) reduce 104) ((T_WHILE) reduce 104) ((T_YIELD) reduce 104) ((T_YIELD_FROM) reduce 104) (("[") reduce 104) (("`") reduce 104) (("{") reduce 104) (("~") reduce 104)) 17921 ((("!") reduce 107) (("\"") reduce 107) (("$") reduce 107) (("(") reduce 107) (("+") reduce 107) (("-") reduce 107) ((";") reduce 107) (("@") reduce 107) ((T_ABSTRACT) reduce 107) ((T_ARRAY) reduce 107) ((T_ARRAY_CAST) reduce 107) ((T_ATTRIBUTE) reduce 107) ((T_BOOL_CAST) reduce 107) ((T_BREAK) reduce 107) ((T_CLASS) reduce 107) ((T_CLASS_C) reduce 107) ((T_CLONE) reduce 107) ((T_CONSTANT_ENCAPSED_STRING) reduce 107) ((T_CONTINUE) reduce 107) ((T_DEC) reduce 107) ((T_DECLARE) reduce 107) ((T_DIR) reduce 107) ((T_DNUMBER) reduce 107) ((T_DO) reduce 107) ((T_DOUBLE_CAST) reduce 107) ((T_ECHO) reduce 107) ((T_EMPTY) reduce 107) ((T_ENDWHILE) reduce 107) ((T_ENUM) reduce 107) ((T_EVAL) reduce 107) ((T_EXIT) reduce 107) ((T_FILE) reduce 107) ((T_FINAL) reduce 107) ((T_FN) reduce 107) ((T_FOR) reduce 107) ((T_FOREACH) reduce 107) ((T_FUNCTION) reduce 107) ((T_FUNC_C) reduce 107) ((T_GLOBAL) reduce 107) ((T_GOTO) reduce 107) ((T_HALT_COMPILER) reduce 107) ((T_IF) reduce 107) ((T_INC) reduce 107) ((T_INCLUDE) reduce 107) ((T_INCLUDE_ONCE) reduce 107) ((T_INLINE_HTML) reduce 107) ((T_INTERFACE) reduce 107) ((T_INT_CAST) reduce 107) ((T_ISSET) reduce 107) ((T_LINE) reduce 107) ((T_LIST) reduce 107) ((T_LNUMBER) reduce 107) ((T_MATCH) reduce 107) ((T_METHOD_C) reduce 107) ((T_NAME_FULLY_QUALIFIED) reduce 107) ((T_NAME_QUALIFIED) reduce 107) ((T_NAME_RELATIVE) reduce 107) ((T_NEW) reduce 107) ((T_NS_C) reduce 107) ((T_OBJECT_CAST) reduce 107) ((T_PRINT) reduce 107) ((T_PROPERTY_C) reduce 107) ((T_READONLY) reduce 107) ((T_REQUIRE) reduce 107) ((T_REQUIRE_ONCE) reduce 107) ((T_RETURN) reduce 107) ((T_START_HEREDOC) reduce 107) ((T_STATIC) reduce 107) ((T_STRING) reduce 107) ((T_STRING_CAST) reduce 107) ((T_SWITCH) reduce 107) ((T_THROW) reduce 107) ((T_TRAIT) reduce 107) ((T_TRAIT_C) reduce 107) ((T_TRY) reduce 107) ((T_UNSET) reduce 107) ((T_UNSET_CAST) reduce 107) ((T_VARIABLE) reduce 107) ((T_VOID_CAST) reduce 107) ((T_WHILE) reduce 107) ((T_YIELD) reduce 107) ((T_YIELD_FROM) reduce 107) (("[") reduce 107) (("`") reduce 107) (("{") reduce 107) (("~") reduce 107)) 17924 ((("!") reduce 103) (("\"") reduce 103) (("$") reduce 103) (("(") reduce 103) (("+") reduce 103) (("-") reduce 103) ((";") reduce 103) (("@") reduce 103) ((T_ABSTRACT) reduce 103) ((T_ARRAY) reduce 103) ((T_ARRAY_CAST) reduce 103) ((T_ATTRIBUTE) reduce 103) ((T_BOOL_CAST) reduce 103) ((T_BREAK) reduce 103) ((T_CLASS) reduce 103) ((T_CLASS_C) reduce 103) ((T_CLONE) reduce 103) ((T_CONSTANT_ENCAPSED_STRING) reduce 103) ((T_CONTINUE) reduce 103) ((T_DEC) reduce 103) ((T_DECLARE) reduce 103) ((T_DIR) reduce 103) ((T_DNUMBER) reduce 103) ((T_DO) reduce 103) ((T_DOUBLE_CAST) reduce 103) ((T_ECHO) reduce 103) ((T_EMPTY) reduce 103) ((T_ENDWHILE) reduce 103) ((T_ENUM) reduce 103) ((T_EVAL) reduce 103) ((T_EXIT) reduce 103) ((T_FILE) reduce 103) ((T_FINAL) reduce 103) ((T_FN) reduce 103) ((T_FOR) reduce 103) ((T_FOREACH) reduce 103) ((T_FUNCTION) reduce 103) ((T_FUNC_C) reduce 103) ((T_GLOBAL) reduce 103) ((T_GOTO) reduce 103) ((T_HALT_COMPILER) reduce 103) ((T_IF) reduce 103) ((T_INC) reduce 103) ((T_INCLUDE) reduce 103) ((T_INCLUDE_ONCE) reduce 103) ((T_INLINE_HTML) reduce 103) ((T_INTERFACE) reduce 103) ((T_INT_CAST) reduce 103) ((T_ISSET) reduce 103) ((T_LINE) reduce 103) ((T_LIST) reduce 103) ((T_LNUMBER) reduce 103) ((T_MATCH) reduce 103) ((T_METHOD_C) reduce 103) ((T_NAME_FULLY_QUALIFIED) reduce 103) ((T_NAME_QUALIFIED) reduce 103) ((T_NAME_RELATIVE) reduce 103) ((T_NEW) reduce 103) ((T_NS_C) reduce 103) ((T_OBJECT_CAST) reduce 103) ((T_PRINT) reduce 103) ((T_PROPERTY_C) reduce 103) ((T_READONLY) reduce 103) ((T_REQUIRE) reduce 103) ((T_REQUIRE_ONCE) reduce 103) ((T_RETURN) reduce 103) ((T_START_HEREDOC) reduce 103) ((T_STATIC) reduce 103) ((T_STRING) reduce 103) ((T_STRING_CAST) reduce 103) ((T_SWITCH) reduce 103) ((T_THROW) reduce 103) ((T_TRAIT) reduce 103) ((T_TRAIT_C) reduce 103) ((T_TRY) reduce 103) ((T_UNSET) reduce 103) ((T_UNSET_CAST) reduce 103) ((T_VARIABLE) reduce 103) ((T_VOID_CAST) reduce 103) ((T_WHILE) reduce 103) ((T_YIELD) reduce 103) ((T_YIELD_FROM) reduce 103) (("[") reduce 103) (("`") reduce 103) (("{") reduce 103) (("~") reduce 103)) 17925 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_EMPTY) reduce 148) ((T_ENDWHILE) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_VOID_CAST) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 17926 ((("!") reduce 246) (("\"") reduce 246) (("$") reduce 246) (("(") reduce 246) (("+") reduce 246) (("-") reduce 246) ((";") reduce 246) (("@") reduce 246) ((T_ABSTRACT) reduce 246) ((T_ARRAY) reduce 246) ((T_ARRAY_CAST) reduce 246) ((T_ATTRIBUTE) reduce 246) ((T_BOOL_CAST) reduce 246) ((T_BREAK) reduce 246) ((T_CLASS) reduce 246) ((T_CLASS_C) reduce 246) ((T_CLONE) reduce 246) ((T_CONSTANT_ENCAPSED_STRING) reduce 246) ((T_CONTINUE) reduce 246) ((T_DEC) reduce 246) ((T_DECLARE) reduce 246) ((T_DIR) reduce 246) ((T_DNUMBER) reduce 246) ((T_DO) reduce 246) ((T_DOUBLE_CAST) reduce 246) ((T_ECHO) reduce 246) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) reduce 246) ((T_ENDWHILE) reduce 246) ((T_ENUM) reduce 246) ((T_EVAL) reduce 246) ((T_EXIT) reduce 246) ((T_FILE) reduce 246) ((T_FINAL) reduce 246) ((T_FN) reduce 246) ((T_FOR) reduce 246) ((T_FOREACH) reduce 246) ((T_FUNCTION) reduce 246) ((T_FUNC_C) reduce 246) ((T_GLOBAL) reduce 246) ((T_GOTO) reduce 246) ((T_HALT_COMPILER) reduce 246) ((T_IF) reduce 246) ((T_INC) reduce 246) ((T_INCLUDE) reduce 246) ((T_INCLUDE_ONCE) reduce 246) ((T_INLINE_HTML) reduce 246) ((T_INTERFACE) reduce 246) ((T_INT_CAST) reduce 246) ((T_ISSET) reduce 246) ((T_LINE) reduce 246) ((T_LIST) reduce 246) ((T_LNUMBER) reduce 246) ((T_MATCH) reduce 246) ((T_METHOD_C) reduce 246) ((T_NAME_FULLY_QUALIFIED) reduce 246) ((T_NAME_QUALIFIED) reduce 246) ((T_NAME_RELATIVE) reduce 246) ((T_NEW) reduce 246) ((T_NS_C) reduce 246) ((T_OBJECT_CAST) reduce 246) ((T_PRINT) reduce 246) ((T_PROPERTY_C) reduce 246) ((T_READONLY) reduce 246) ((T_REQUIRE) reduce 246) ((T_REQUIRE_ONCE) reduce 246) ((T_RETURN) reduce 246) ((T_START_HEREDOC) reduce 246) ((T_STATIC) reduce 246) ((T_STRING) reduce 246) ((T_STRING_CAST) reduce 246) ((T_SWITCH) reduce 246) ((T_THROW) reduce 246) ((T_TRAIT) reduce 246) ((T_TRAIT_C) reduce 246) ((T_TRY) reduce 246) ((T_UNSET) reduce 246) ((T_UNSET_CAST) reduce 246) ((T_VARIABLE) reduce 246) ((T_VOID_CAST) reduce 246) ((T_WHILE) reduce 246) ((T_YIELD) reduce 246) ((T_YIELD_FROM) reduce 246) (("[") reduce 246) (("`") reduce 246) (("{") reduce 246) (("~") reduce 246)) 17927 ((("!") reduce 141) (("\"") reduce 141) (("$") reduce 141) (("(") reduce 141) (("+") reduce 141) (("-") reduce 141) ((";") reduce 141) (("@") reduce 141) ((T_ABSTRACT) reduce 141) ((T_ARRAY) reduce 141) ((T_ARRAY_CAST) reduce 141) ((T_ATTRIBUTE) reduce 141) ((T_BOOL_CAST) reduce 141) ((T_BREAK) reduce 141) ((T_CLASS) reduce 141) ((T_CLASS_C) reduce 141) ((T_CLONE) reduce 141) ((T_CONSTANT_ENCAPSED_STRING) reduce 141) ((T_CONTINUE) reduce 141) ((T_DEC) reduce 141) ((T_DECLARE) reduce 141) ((T_DIR) reduce 141) ((T_DNUMBER) reduce 141) ((T_DO) reduce 141) ((T_DOUBLE_CAST) reduce 141) ((T_ECHO) reduce 141) ((T_EMPTY) reduce 141) ((T_ENDWHILE) reduce 141) ((T_ENUM) reduce 141) ((T_EVAL) reduce 141) ((T_EXIT) reduce 141) ((T_FILE) reduce 141) ((T_FINAL) reduce 141) ((T_FN) reduce 141) ((T_FOR) reduce 141) ((T_FOREACH) reduce 141) ((T_FUNCTION) reduce 141) ((T_FUNC_C) reduce 141) ((T_GLOBAL) reduce 141) ((T_GOTO) reduce 141) ((T_HALT_COMPILER) reduce 141) ((T_IF) reduce 141) ((T_INC) reduce 141) ((T_INCLUDE) reduce 141) ((T_INCLUDE_ONCE) reduce 141) ((T_INLINE_HTML) reduce 141) ((T_INTERFACE) reduce 141) ((T_INT_CAST) reduce 141) ((T_ISSET) reduce 141) ((T_LINE) reduce 141) ((T_LIST) reduce 141) ((T_LNUMBER) reduce 141) ((T_MATCH) reduce 141) ((T_METHOD_C) reduce 141) ((T_NAME_FULLY_QUALIFIED) reduce 141) ((T_NAME_QUALIFIED) reduce 141) ((T_NAME_RELATIVE) reduce 141) ((T_NEW) reduce 141) ((T_NS_C) reduce 141) ((T_OBJECT_CAST) reduce 141) ((T_PRINT) reduce 141) ((T_PROPERTY_C) reduce 141) ((T_READONLY) reduce 141) ((T_REQUIRE) reduce 141) ((T_REQUIRE_ONCE) reduce 141) ((T_RETURN) reduce 141) ((T_START_HEREDOC) reduce 141) ((T_STATIC) reduce 141) ((T_STRING) reduce 141) ((T_STRING_CAST) reduce 141) ((T_SWITCH) reduce 141) ((T_THROW) reduce 141) ((T_TRAIT) reduce 141) ((T_TRAIT_C) reduce 141) ((T_TRY) reduce 141) ((T_UNSET) reduce 141) ((T_UNSET_CAST) reduce 141) ((T_VARIABLE) reduce 141) ((T_VOID_CAST) reduce 141) ((T_WHILE) reduce 141) ((T_YIELD) reduce 141) ((T_YIELD_FROM) reduce 141) (("[") reduce 141) (("`") reduce 141) (("{") reduce 141) (("~") reduce 141)) 17928 ((("!") reduce 106) (("\"") reduce 106) (("$") reduce 106) (("(") reduce 106) (("+") reduce 106) (("-") reduce 106) ((";") reduce 106) (("@") reduce 106) ((T_ABSTRACT) reduce 106) ((T_ARRAY) reduce 106) ((T_ARRAY_CAST) reduce 106) ((T_ATTRIBUTE) reduce 106) ((T_BOOL_CAST) reduce 106) ((T_BREAK) reduce 106) ((T_CLASS) reduce 106) ((T_CLASS_C) reduce 106) ((T_CLONE) reduce 106) ((T_CONSTANT_ENCAPSED_STRING) reduce 106) ((T_CONTINUE) reduce 106) ((T_DEC) reduce 106) ((T_DECLARE) reduce 106) ((T_DIR) reduce 106) ((T_DNUMBER) reduce 106) ((T_DO) reduce 106) ((T_DOUBLE_CAST) reduce 106) ((T_ECHO) reduce 106) ((T_EMPTY) reduce 106) ((T_ENDWHILE) reduce 106) ((T_ENUM) reduce 106) ((T_EVAL) reduce 106) ((T_EXIT) reduce 106) ((T_FILE) reduce 106) ((T_FINAL) reduce 106) ((T_FN) reduce 106) ((T_FOR) reduce 106) ((T_FOREACH) reduce 106) ((T_FUNCTION) reduce 106) ((T_FUNC_C) reduce 106) ((T_GLOBAL) reduce 106) ((T_GOTO) reduce 106) ((T_HALT_COMPILER) reduce 106) ((T_IF) reduce 106) ((T_INC) reduce 106) ((T_INCLUDE) reduce 106) ((T_INCLUDE_ONCE) reduce 106) ((T_INLINE_HTML) reduce 106) ((T_INTERFACE) reduce 106) ((T_INT_CAST) reduce 106) ((T_ISSET) reduce 106) ((T_LINE) reduce 106) ((T_LIST) reduce 106) ((T_LNUMBER) reduce 106) ((T_MATCH) reduce 106) ((T_METHOD_C) reduce 106) ((T_NAME_FULLY_QUALIFIED) reduce 106) ((T_NAME_QUALIFIED) reduce 106) ((T_NAME_RELATIVE) reduce 106) ((T_NEW) reduce 106) ((T_NS_C) reduce 106) ((T_OBJECT_CAST) reduce 106) ((T_PRINT) reduce 106) ((T_PROPERTY_C) reduce 106) ((T_READONLY) reduce 106) ((T_REQUIRE) reduce 106) ((T_REQUIRE_ONCE) reduce 106) ((T_RETURN) reduce 106) ((T_START_HEREDOC) reduce 106) ((T_STATIC) reduce 106) ((T_STRING) reduce 106) ((T_STRING_CAST) reduce 106) ((T_SWITCH) reduce 106) ((T_THROW) reduce 106) ((T_TRAIT) reduce 106) ((T_TRAIT_C) reduce 106) ((T_TRY) reduce 106) ((T_UNSET) reduce 106) ((T_UNSET_CAST) reduce 106) ((T_VARIABLE) reduce 106) ((T_VOID_CAST) reduce 106) ((T_WHILE) reduce 106) ((T_YIELD) reduce 106) ((T_YIELD_FROM) reduce 106) (("[") reduce 106) (("`") reduce 106) (("{") reduce 106) (("~") reduce 106)) 17929 ((("!") reduce 143) (("\"") reduce 143) (("$") reduce 143) (("(") reduce 143) (("+") reduce 143) (("-") reduce 143) ((";") reduce 143) (("@") reduce 143) ((T_ABSTRACT) reduce 143) ((T_ARRAY) reduce 143) ((T_ARRAY_CAST) reduce 143) ((T_ATTRIBUTE) reduce 143) ((T_BOOL_CAST) reduce 143) ((T_BREAK) reduce 143) ((T_CLASS) reduce 143) ((T_CLASS_C) reduce 143) ((T_CLONE) reduce 143) ((T_CONSTANT_ENCAPSED_STRING) reduce 143) ((T_CONTINUE) reduce 143) ((T_DEC) reduce 143) ((T_DECLARE) reduce 143) ((T_DIR) reduce 143) ((T_DNUMBER) reduce 143) ((T_DO) reduce 143) ((T_DOUBLE_CAST) reduce 143) ((T_ECHO) reduce 143) ((T_EMPTY) reduce 143) ((T_ENDWHILE) reduce 143) ((T_ENUM) reduce 143) ((T_EVAL) reduce 143) ((T_EXIT) reduce 143) ((T_FILE) reduce 143) ((T_FINAL) reduce 143) ((T_FN) reduce 143) ((T_FOR) reduce 143) ((T_FOREACH) reduce 143) ((T_FUNCTION) reduce 143) ((T_FUNC_C) reduce 143) ((T_GLOBAL) reduce 143) ((T_GOTO) reduce 143) ((T_HALT_COMPILER) reduce 143) ((T_IF) reduce 143) ((T_INC) reduce 143) ((T_INCLUDE) reduce 143) ((T_INCLUDE_ONCE) reduce 143) ((T_INLINE_HTML) reduce 143) ((T_INTERFACE) reduce 143) ((T_INT_CAST) reduce 143) ((T_ISSET) reduce 143) ((T_LINE) reduce 143) ((T_LIST) reduce 143) ((T_LNUMBER) reduce 143) ((T_MATCH) reduce 143) ((T_METHOD_C) reduce 143) ((T_NAME_FULLY_QUALIFIED) reduce 143) ((T_NAME_QUALIFIED) reduce 143) ((T_NAME_RELATIVE) reduce 143) ((T_NEW) reduce 143) ((T_NS_C) reduce 143) ((T_OBJECT_CAST) reduce 143) ((T_PRINT) reduce 143) ((T_PROPERTY_C) reduce 143) ((T_READONLY) reduce 143) ((T_REQUIRE) reduce 143) ((T_REQUIRE_ONCE) reduce 143) ((T_RETURN) reduce 143) ((T_START_HEREDOC) reduce 143) ((T_STATIC) reduce 143) ((T_STRING) reduce 143) ((T_STRING_CAST) reduce 143) ((T_SWITCH) reduce 143) ((T_THROW) reduce 143) ((T_TRAIT) reduce 143) ((T_TRAIT_C) reduce 143) ((T_TRY) reduce 143) ((T_UNSET) reduce 143) ((T_UNSET_CAST) reduce 143) ((T_VARIABLE) reduce 143) ((T_VOID_CAST) reduce 143) ((T_WHILE) reduce 143) ((T_YIELD) reduce 143) ((T_YIELD_FROM) reduce 143) (("[") reduce 143) (("`") reduce 143) (("{") reduce 143) (("~") reduce 143)) 17930 ((("!") reduce 105) (("\"") reduce 105) (("$") reduce 105) (("(") reduce 105) (("+") reduce 105) (("-") reduce 105) ((";") reduce 105) (("@") reduce 105) ((T_ABSTRACT) reduce 105) ((T_ARRAY) reduce 105) ((T_ARRAY_CAST) reduce 105) ((T_ATTRIBUTE) reduce 105) ((T_BOOL_CAST) reduce 105) ((T_BREAK) reduce 105) ((T_CLASS) reduce 105) ((T_CLASS_C) reduce 105) ((T_CLONE) reduce 105) ((T_CONSTANT_ENCAPSED_STRING) reduce 105) ((T_CONTINUE) reduce 105) ((T_DEC) reduce 105) ((T_DECLARE) reduce 105) ((T_DIR) reduce 105) ((T_DNUMBER) reduce 105) ((T_DO) reduce 105) ((T_DOUBLE_CAST) reduce 105) ((T_ECHO) reduce 105) ((T_EMPTY) reduce 105) ((T_ENDWHILE) reduce 105) ((T_ENUM) reduce 105) ((T_EVAL) reduce 105) ((T_EXIT) reduce 105) ((T_FILE) reduce 105) ((T_FINAL) reduce 105) ((T_FN) reduce 105) ((T_FOR) reduce 105) ((T_FOREACH) reduce 105) ((T_FUNCTION) reduce 105) ((T_FUNC_C) reduce 105) ((T_GLOBAL) reduce 105) ((T_GOTO) reduce 105) ((T_HALT_COMPILER) reduce 105) ((T_IF) reduce 105) ((T_INC) reduce 105) ((T_INCLUDE) reduce 105) ((T_INCLUDE_ONCE) reduce 105) ((T_INLINE_HTML) reduce 105) ((T_INTERFACE) reduce 105) ((T_INT_CAST) reduce 105) ((T_ISSET) reduce 105) ((T_LINE) reduce 105) ((T_LIST) reduce 105) ((T_LNUMBER) reduce 105) ((T_MATCH) reduce 105) ((T_METHOD_C) reduce 105) ((T_NAME_FULLY_QUALIFIED) reduce 105) ((T_NAME_QUALIFIED) reduce 105) ((T_NAME_RELATIVE) reduce 105) ((T_NEW) reduce 105) ((T_NS_C) reduce 105) ((T_OBJECT_CAST) reduce 105) ((T_PRINT) reduce 105) ((T_PROPERTY_C) reduce 105) ((T_READONLY) reduce 105) ((T_REQUIRE) reduce 105) ((T_REQUIRE_ONCE) reduce 105) ((T_RETURN) reduce 105) ((T_START_HEREDOC) reduce 105) ((T_STATIC) reduce 105) ((T_STRING) reduce 105) ((T_STRING_CAST) reduce 105) ((T_SWITCH) reduce 105) ((T_THROW) reduce 105) ((T_TRAIT) reduce 105) ((T_TRAIT_C) reduce 105) ((T_TRY) reduce 105) ((T_UNSET) reduce 105) ((T_UNSET_CAST) reduce 105) ((T_VARIABLE) reduce 105) ((T_VOID_CAST) reduce 105) ((T_WHILE) reduce 105) ((T_YIELD) reduce 105) ((T_YIELD_FROM) reduce 105) (("[") reduce 105) (("`") reduce 105) (("{") reduce 105) (("~") reduce 105)) 17933 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_EMPTY) reduce 147) ((T_ENDWHILE) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_VOID_CAST) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 17939 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_ELSE) reduce 166) ((T_ELSEIF) reduce 166) ((T_EMPTY) reduce 166) ((T_ENDWHILE) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_VOID_CAST) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 17949 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_ELSE) reduce 160) ((T_ELSEIF) reduce 160) ((T_EMPTY) reduce 160) ((T_ENDWHILE) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_VOID_CAST) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 17958 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_ELSE) reduce 149) ((T_ELSEIF) reduce 149) ((T_EMPTY) reduce 149) ((T_ENDWHILE) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_VOID_CAST) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 17961 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_ELSE) reduce 148) ((T_ELSEIF) reduce 148) ((T_EMPTY) reduce 148) ((T_ENDWHILE) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_VOID_CAST) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 17963 ((("!") reduce 245) (("\"") reduce 245) (("$") reduce 245) (("(") reduce 245) (("+") reduce 245) (("-") reduce 245) ((";") reduce 245) (("@") reduce 245) ((T_ABSTRACT) reduce 245) ((T_ARRAY) reduce 245) ((T_ARRAY_CAST) reduce 245) ((T_ATTRIBUTE) reduce 245) ((T_BOOL_CAST) reduce 245) ((T_BREAK) reduce 245) ((T_CLASS) reduce 245) ((T_CLASS_C) reduce 245) ((T_CLONE) reduce 245) ((T_CONSTANT_ENCAPSED_STRING) reduce 245) ((T_CONTINUE) reduce 245) ((T_DEC) reduce 245) ((T_DECLARE) reduce 245) ((T_DIR) reduce 245) ((T_DNUMBER) reduce 245) ((T_DO) reduce 245) ((T_DOUBLE_CAST) reduce 245) ((T_ECHO) reduce 245) ((T_ELSE) reduce 245) ((T_ELSEIF) reduce 245) ((T_EMPTY) reduce 245) ((T_ENDWHILE) reduce 245) ((T_ENUM) reduce 245) ((T_EVAL) reduce 245) ((T_EXIT) reduce 245) ((T_FILE) reduce 245) ((T_FINAL) reduce 245) ((T_FN) reduce 245) ((T_FOR) reduce 245) ((T_FOREACH) reduce 245) ((T_FUNCTION) reduce 245) ((T_FUNC_C) reduce 245) ((T_GLOBAL) reduce 245) ((T_GOTO) reduce 245) ((T_HALT_COMPILER) reduce 245) ((T_IF) reduce 245) ((T_INC) reduce 245) ((T_INCLUDE) reduce 245) ((T_INCLUDE_ONCE) reduce 245) ((T_INLINE_HTML) reduce 245) ((T_INTERFACE) reduce 245) ((T_INT_CAST) reduce 245) ((T_ISSET) reduce 245) ((T_LINE) reduce 245) ((T_LIST) reduce 245) ((T_LNUMBER) reduce 245) ((T_MATCH) reduce 245) ((T_METHOD_C) reduce 245) ((T_NAME_FULLY_QUALIFIED) reduce 245) ((T_NAME_QUALIFIED) reduce 245) ((T_NAME_RELATIVE) reduce 245) ((T_NEW) reduce 245) ((T_NS_C) reduce 245) ((T_OBJECT_CAST) reduce 245) ((T_PRINT) reduce 245) ((T_PROPERTY_C) reduce 245) ((T_READONLY) reduce 245) ((T_REQUIRE) reduce 245) ((T_REQUIRE_ONCE) reduce 245) ((T_RETURN) reduce 245) ((T_START_HEREDOC) reduce 245) ((T_STATIC) reduce 245) ((T_STRING) reduce 245) ((T_STRING_CAST) reduce 245) ((T_SWITCH) reduce 245) ((T_THROW) reduce 245) ((T_TRAIT) reduce 245) ((T_TRAIT_C) reduce 245) ((T_TRY) reduce 245) ((T_UNSET) reduce 245) ((T_UNSET_CAST) reduce 245) ((T_VARIABLE) reduce 245) ((T_VOID_CAST) reduce 245) ((T_WHILE) reduce 245) ((T_YIELD) reduce 245) ((T_YIELD_FROM) reduce 245) (("[") reduce 245) (("`") reduce 245) (("{") reduce 245) (("~") reduce 245)) 17966 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_ELSE) reduce 147) ((T_ELSEIF) reduce 147) ((T_EMPTY) reduce 147) ((T_ENDWHILE) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_VOID_CAST) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 17968 ((("!") reduce 247) (("\"") reduce 247) (("$") reduce 247) (("(") reduce 247) (("+") reduce 247) (("-") reduce 247) ((";") reduce 247) (("@") reduce 247) ((T_ABSTRACT) reduce 247) ((T_ARRAY) reduce 247) ((T_ARRAY_CAST) reduce 247) ((T_ATTRIBUTE) reduce 247) ((T_BOOL_CAST) reduce 247) ((T_BREAK) reduce 247) ((T_CLASS) reduce 247) ((T_CLASS_C) reduce 247) ((T_CLONE) reduce 247) ((T_CONSTANT_ENCAPSED_STRING) reduce 247) ((T_CONTINUE) reduce 247) ((T_DEC) reduce 247) ((T_DECLARE) reduce 247) ((T_DIR) reduce 247) ((T_DNUMBER) reduce 247) ((T_DO) reduce 247) ((T_DOUBLE_CAST) reduce 247) ((T_ECHO) reduce 247) ((T_ELSE) reduce 247) ((T_ELSEIF) reduce 247) ((T_EMPTY) reduce 247) ((T_ENDWHILE) reduce 247) ((T_ENUM) reduce 247) ((T_EVAL) reduce 247) ((T_EXIT) reduce 247) ((T_FILE) reduce 247) ((T_FINAL) reduce 247) ((T_FN) reduce 247) ((T_FOR) reduce 247) ((T_FOREACH) reduce 247) ((T_FUNCTION) reduce 247) ((T_FUNC_C) reduce 247) ((T_GLOBAL) reduce 247) ((T_GOTO) reduce 247) ((T_HALT_COMPILER) reduce 247) ((T_IF) reduce 247) ((T_INC) reduce 247) ((T_INCLUDE) reduce 247) ((T_INCLUDE_ONCE) reduce 247) ((T_INLINE_HTML) reduce 247) ((T_INTERFACE) reduce 247) ((T_INT_CAST) reduce 247) ((T_ISSET) reduce 247) ((T_LINE) reduce 247) ((T_LIST) reduce 247) ((T_LNUMBER) reduce 247) ((T_MATCH) reduce 247) ((T_METHOD_C) reduce 247) ((T_NAME_FULLY_QUALIFIED) reduce 247) ((T_NAME_QUALIFIED) reduce 247) ((T_NAME_RELATIVE) reduce 247) ((T_NEW) reduce 247) ((T_NS_C) reduce 247) ((T_OBJECT_CAST) reduce 247) ((T_PRINT) reduce 247) ((T_PROPERTY_C) reduce 247) ((T_READONLY) reduce 247) ((T_REQUIRE) reduce 247) ((T_REQUIRE_ONCE) reduce 247) ((T_RETURN) reduce 247) ((T_START_HEREDOC) reduce 247) ((T_STATIC) reduce 247) ((T_STRING) reduce 247) ((T_STRING_CAST) reduce 247) ((T_SWITCH) reduce 247) ((T_THROW) reduce 247) ((T_TRAIT) reduce 247) ((T_TRAIT_C) reduce 247) ((T_TRY) reduce 247) ((T_UNSET) reduce 247) ((T_UNSET_CAST) reduce 247) ((T_VARIABLE) reduce 247) ((T_VOID_CAST) reduce 247) ((T_WHILE) reduce 247) ((T_YIELD) reduce 247) ((T_YIELD_FROM) reduce 247) (("[") reduce 247) (("`") reduce 247) (("{") reduce 247) (("~") reduce 247)) 17969 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_ELSE) reduce 161) ((T_ELSEIF) reduce 161) ((T_EMPTY) reduce 161) ((T_ENDWHILE) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_VOID_CAST) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 17972 ((("!") reduce 250) (("\"") reduce 250) (("$") reduce 250) (("(") reduce 250) (("+") reduce 250) (("-") reduce 250) ((";") reduce 250) (("@") reduce 250) ((T_ABSTRACT) reduce 250) ((T_ARRAY) reduce 250) ((T_ARRAY_CAST) reduce 250) ((T_ATTRIBUTE) reduce 250) ((T_BOOL_CAST) reduce 250) ((T_BREAK) reduce 250) ((T_CLASS) reduce 250) ((T_CLASS_C) reduce 250) ((T_CLONE) reduce 250) ((T_CONSTANT_ENCAPSED_STRING) reduce 250) ((T_CONTINUE) reduce 250) ((T_DEC) reduce 250) ((T_DECLARE) reduce 250) ((T_DIR) reduce 250) ((T_DNUMBER) reduce 250) ((T_DO) reduce 250) ((T_DOUBLE_CAST) reduce 250) ((T_ECHO) reduce 250) ((T_ELSE) reduce 250) ((T_ELSEIF) reduce 250) ((T_EMPTY) reduce 250) ((T_ENDWHILE) reduce 250) ((T_ENUM) reduce 250) ((T_EVAL) reduce 250) ((T_EXIT) reduce 250) ((T_FILE) reduce 250) ((T_FINAL) reduce 250) ((T_FN) reduce 250) ((T_FOR) reduce 250) ((T_FOREACH) reduce 250) ((T_FUNCTION) reduce 250) ((T_FUNC_C) reduce 250) ((T_GLOBAL) reduce 250) ((T_GOTO) reduce 250) ((T_HALT_COMPILER) reduce 250) ((T_IF) reduce 250) ((T_INC) reduce 250) ((T_INCLUDE) reduce 250) ((T_INCLUDE_ONCE) reduce 250) ((T_INLINE_HTML) reduce 250) ((T_INTERFACE) reduce 250) ((T_INT_CAST) reduce 250) ((T_ISSET) reduce 250) ((T_LINE) reduce 250) ((T_LIST) reduce 250) ((T_LNUMBER) reduce 250) ((T_MATCH) reduce 250) ((T_METHOD_C) reduce 250) ((T_NAME_FULLY_QUALIFIED) reduce 250) ((T_NAME_QUALIFIED) reduce 250) ((T_NAME_RELATIVE) reduce 250) ((T_NEW) reduce 250) ((T_NS_C) reduce 250) ((T_OBJECT_CAST) reduce 250) ((T_PRINT) reduce 250) ((T_PROPERTY_C) reduce 250) ((T_READONLY) reduce 250) ((T_REQUIRE) reduce 250) ((T_REQUIRE_ONCE) reduce 250) ((T_RETURN) reduce 250) ((T_START_HEREDOC) reduce 250) ((T_STATIC) reduce 250) ((T_STRING) reduce 250) ((T_STRING_CAST) reduce 250) ((T_SWITCH) reduce 250) ((T_THROW) reduce 250) ((T_TRAIT) reduce 250) ((T_TRAIT_C) reduce 250) ((T_TRY) reduce 250) ((T_UNSET) reduce 250) ((T_UNSET_CAST) reduce 250) ((T_VARIABLE) reduce 250) ((T_VOID_CAST) reduce 250) ((T_WHILE) reduce 250) ((T_YIELD) reduce 250) ((T_YIELD_FROM) reduce 250) (("[") reduce 250) (("`") reduce 250) (("{") reduce 250) (("~") reduce 250)) 17976 ((("!") reduce 251) (("\"") reduce 251) (("$") reduce 251) (("(") reduce 251) (("+") reduce 251) (("-") reduce 251) ((";") reduce 251) (("@") reduce 251) ((T_ABSTRACT) reduce 251) ((T_ARRAY) reduce 251) ((T_ARRAY_CAST) reduce 251) ((T_ATTRIBUTE) reduce 251) ((T_BOOL_CAST) reduce 251) ((T_BREAK) reduce 251) ((T_CLASS) reduce 251) ((T_CLASS_C) reduce 251) ((T_CLONE) reduce 251) ((T_CONSTANT_ENCAPSED_STRING) reduce 251) ((T_CONTINUE) reduce 251) ((T_DEC) reduce 251) ((T_DECLARE) reduce 251) ((T_DIR) reduce 251) ((T_DNUMBER) reduce 251) ((T_DO) reduce 251) ((T_DOUBLE_CAST) reduce 251) ((T_ECHO) reduce 251) ((T_ELSE) reduce 251) ((T_ELSEIF) reduce 251) ((T_EMPTY) reduce 251) ((T_ENDWHILE) reduce 251) ((T_ENUM) reduce 251) ((T_EVAL) reduce 251) ((T_EXIT) reduce 251) ((T_FILE) reduce 251) ((T_FINAL) reduce 251) ((T_FN) reduce 251) ((T_FOR) reduce 251) ((T_FOREACH) reduce 251) ((T_FUNCTION) reduce 251) ((T_FUNC_C) reduce 251) ((T_GLOBAL) reduce 251) ((T_GOTO) reduce 251) ((T_HALT_COMPILER) reduce 251) ((T_IF) reduce 251) ((T_INC) reduce 251) ((T_INCLUDE) reduce 251) ((T_INCLUDE_ONCE) reduce 251) ((T_INLINE_HTML) reduce 251) ((T_INTERFACE) reduce 251) ((T_INT_CAST) reduce 251) ((T_ISSET) reduce 251) ((T_LINE) reduce 251) ((T_LIST) reduce 251) ((T_LNUMBER) reduce 251) ((T_MATCH) reduce 251) ((T_METHOD_C) reduce 251) ((T_NAME_FULLY_QUALIFIED) reduce 251) ((T_NAME_QUALIFIED) reduce 251) ((T_NAME_RELATIVE) reduce 251) ((T_NEW) reduce 251) ((T_NS_C) reduce 251) ((T_OBJECT_CAST) reduce 251) ((T_PRINT) reduce 251) ((T_PROPERTY_C) reduce 251) ((T_READONLY) reduce 251) ((T_REQUIRE) reduce 251) ((T_REQUIRE_ONCE) reduce 251) ((T_RETURN) reduce 251) ((T_START_HEREDOC) reduce 251) ((T_STATIC) reduce 251) ((T_STRING) reduce 251) ((T_STRING_CAST) reduce 251) ((T_SWITCH) reduce 251) ((T_THROW) reduce 251) ((T_TRAIT) reduce 251) ((T_TRAIT_C) reduce 251) ((T_TRY) reduce 251) ((T_UNSET) reduce 251) ((T_UNSET_CAST) reduce 251) ((T_VARIABLE) reduce 251) ((T_VOID_CAST) reduce 251) ((T_WHILE) reduce 251) ((T_YIELD) reduce 251) ((T_YIELD_FROM) reduce 251) (("[") reduce 251) (("`") reduce 251) (("{") reduce 251) (("~") reduce 251)) 17981 ((("!") reduce 242) (("\"") reduce 242) (("$") reduce 242) (("(") reduce 242) (("+") reduce 242) (("-") reduce 242) ((";") reduce 242) (("@") reduce 242) ((T_ABSTRACT) reduce 242) ((T_ARRAY) reduce 242) ((T_ARRAY_CAST) reduce 242) ((T_ATTRIBUTE) reduce 242) ((T_BOOL_CAST) reduce 242) ((T_BREAK) reduce 242) ((T_CLASS) reduce 242) ((T_CLASS_C) reduce 242) ((T_CLONE) reduce 242) ((T_CONSTANT_ENCAPSED_STRING) reduce 242) ((T_CONTINUE) reduce 242) ((T_DEC) reduce 242) ((T_DECLARE) reduce 242) ((T_DIR) reduce 242) ((T_DNUMBER) reduce 242) ((T_DO) reduce 242) ((T_DOUBLE_CAST) reduce 242) ((T_ECHO) reduce 242) ((T_ELSE) reduce 242) ((T_ELSEIF) reduce 242) ((T_EMPTY) reduce 242) ((T_ENDWHILE) reduce 242) ((T_ENUM) reduce 242) ((T_EVAL) reduce 242) ((T_EXIT) reduce 242) ((T_FILE) reduce 242) ((T_FINAL) reduce 242) ((T_FN) reduce 242) ((T_FOR) reduce 242) ((T_FOREACH) reduce 242) ((T_FUNCTION) reduce 242) ((T_FUNC_C) reduce 242) ((T_GLOBAL) reduce 242) ((T_GOTO) reduce 242) ((T_HALT_COMPILER) reduce 242) ((T_IF) reduce 242) ((T_INC) reduce 242) ((T_INCLUDE) reduce 242) ((T_INCLUDE_ONCE) reduce 242) ((T_INLINE_HTML) reduce 242) ((T_INTERFACE) reduce 242) ((T_INT_CAST) reduce 242) ((T_ISSET) reduce 242) ((T_LINE) reduce 242) ((T_LIST) reduce 242) ((T_LNUMBER) reduce 242) ((T_MATCH) reduce 242) ((T_METHOD_C) reduce 242) ((T_NAME_FULLY_QUALIFIED) reduce 242) ((T_NAME_QUALIFIED) reduce 242) ((T_NAME_RELATIVE) reduce 242) ((T_NEW) reduce 242) ((T_NS_C) reduce 242) ((T_OBJECT_CAST) reduce 242) ((T_PRINT) reduce 242) ((T_PROPERTY_C) reduce 242) ((T_READONLY) reduce 242) ((T_REQUIRE) reduce 242) ((T_REQUIRE_ONCE) reduce 242) ((T_RETURN) reduce 242) ((T_START_HEREDOC) reduce 242) ((T_STATIC) reduce 242) ((T_STRING) reduce 242) ((T_STRING_CAST) reduce 242) ((T_SWITCH) reduce 242) ((T_THROW) reduce 242) ((T_TRAIT) reduce 242) ((T_TRAIT_C) reduce 242) ((T_TRY) reduce 242) ((T_UNSET) reduce 242) ((T_UNSET_CAST) reduce 242) ((T_VARIABLE) reduce 242) ((T_VOID_CAST) reduce 242) ((T_WHILE) reduce 242) ((T_YIELD) reduce 242) ((T_YIELD_FROM) reduce 242) (("[") reduce 242) (("`") reduce 242) (("{") reduce 242) (("~") reduce 242)) 17982 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_ELSE) reduce 150) ((T_ELSEIF) reduce 150) ((T_EMPTY) reduce 150) ((T_ENDWHILE) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_VOID_CAST) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 17985 ((("!") reduce 243) (("\"") reduce 243) (("$") reduce 243) (("(") reduce 243) (("+") reduce 243) (("-") reduce 243) ((";") reduce 243) (("@") reduce 243) ((T_ABSTRACT) reduce 243) ((T_ARRAY) reduce 243) ((T_ARRAY_CAST) reduce 243) ((T_ATTRIBUTE) reduce 243) ((T_BOOL_CAST) reduce 243) ((T_BREAK) reduce 243) ((T_CLASS) reduce 243) ((T_CLASS_C) reduce 243) ((T_CLONE) reduce 243) ((T_CONSTANT_ENCAPSED_STRING) reduce 243) ((T_CONTINUE) reduce 243) ((T_DEC) reduce 243) ((T_DECLARE) reduce 243) ((T_DIR) reduce 243) ((T_DNUMBER) reduce 243) ((T_DO) reduce 243) ((T_DOUBLE_CAST) reduce 243) ((T_ECHO) reduce 243) ((T_ELSE) reduce 243) ((T_ELSEIF) reduce 243) ((T_EMPTY) reduce 243) ((T_ENDWHILE) reduce 243) ((T_ENUM) reduce 243) ((T_EVAL) reduce 243) ((T_EXIT) reduce 243) ((T_FILE) reduce 243) ((T_FINAL) reduce 243) ((T_FN) reduce 243) ((T_FOR) reduce 243) ((T_FOREACH) reduce 243) ((T_FUNCTION) reduce 243) ((T_FUNC_C) reduce 243) ((T_GLOBAL) reduce 243) ((T_GOTO) reduce 243) ((T_HALT_COMPILER) reduce 243) ((T_IF) reduce 243) ((T_INC) reduce 243) ((T_INCLUDE) reduce 243) ((T_INCLUDE_ONCE) reduce 243) ((T_INLINE_HTML) reduce 243) ((T_INTERFACE) reduce 243) ((T_INT_CAST) reduce 243) ((T_ISSET) reduce 243) ((T_LINE) reduce 243) ((T_LIST) reduce 243) ((T_LNUMBER) reduce 243) ((T_MATCH) reduce 243) ((T_METHOD_C) reduce 243) ((T_NAME_FULLY_QUALIFIED) reduce 243) ((T_NAME_QUALIFIED) reduce 243) ((T_NAME_RELATIVE) reduce 243) ((T_NEW) reduce 243) ((T_NS_C) reduce 243) ((T_OBJECT_CAST) reduce 243) ((T_PRINT) reduce 243) ((T_PROPERTY_C) reduce 243) ((T_READONLY) reduce 243) ((T_REQUIRE) reduce 243) ((T_REQUIRE_ONCE) reduce 243) ((T_RETURN) reduce 243) ((T_START_HEREDOC) reduce 243) ((T_STATIC) reduce 243) ((T_STRING) reduce 243) ((T_STRING_CAST) reduce 243) ((T_SWITCH) reduce 243) ((T_THROW) reduce 243) ((T_TRAIT) reduce 243) ((T_TRAIT_C) reduce 243) ((T_TRY) reduce 243) ((T_UNSET) reduce 243) ((T_UNSET_CAST) reduce 243) ((T_VARIABLE) reduce 243) ((T_VOID_CAST) reduce 243) ((T_WHILE) reduce 243) ((T_YIELD) reduce 243) ((T_YIELD_FROM) reduce 243) (("[") reduce 243) (("`") reduce 243) (("{") reduce 243) (("~") reduce 243)) 17987 ((("!") reduce 170) (("\"") reduce 170) (("$") reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_ELSE) reduce 170) ((T_ELSEIF) reduce 170) ((T_EMPTY) reduce 170) ((T_ENDWHILE) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_VARIABLE) reduce 170) ((T_VOID_CAST) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("~") reduce 170)) 18000 (((")") reduce 179) ((",") reduce 179)) 18002 (((")") reduce 181) ((",") reduce 181) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) (("[") reduce 548)) 18015 ((("\"") shift) (("$") shift) (("(") shift) ((")") reduce 126) ((T_ARRAY) shift) ((T_CLASS_C) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DIR) shift) ((T_FILE) shift) ((T_FUNC_C) shift) ((T_LINE) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_TRAIT_C) shift) ((T_VARIABLE) shift) (("[") shift)) 18018 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_ELSE) reduce 162) ((T_ELSEIF) reduce 162) ((T_EMPTY) reduce 162) ((T_ENDWHILE) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_VOID_CAST) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 18019 (((")") reduce 180) ((",") reduce 180)) 18047 ((("!") reduce 171) (("\"") reduce 171) (("$") reduce 171) (("(") reduce 171) (("+") reduce 171) (("-") reduce 171) ((";") reduce 171) (("@") reduce 171) ((T_ABSTRACT) reduce 171) ((T_ARRAY) reduce 171) ((T_ARRAY_CAST) reduce 171) ((T_ATTRIBUTE) reduce 171) ((T_BOOL_CAST) reduce 171) ((T_BREAK) reduce 171) ((T_CATCH) reduce 171) ((T_CLASS) reduce 171) ((T_CLASS_C) reduce 171) ((T_CLONE) reduce 171) ((T_CONSTANT_ENCAPSED_STRING) reduce 171) ((T_CONTINUE) reduce 171) ((T_DEC) reduce 171) ((T_DECLARE) reduce 171) ((T_DIR) reduce 171) ((T_DNUMBER) reduce 171) ((T_DO) reduce 171) ((T_DOUBLE_CAST) reduce 171) ((T_ECHO) reduce 171) ((T_ELSE) reduce 171) ((T_ELSEIF) reduce 171) ((T_EMPTY) reduce 171) ((T_ENDWHILE) reduce 171) ((T_ENUM) reduce 171) ((T_EVAL) reduce 171) ((T_EXIT) reduce 171) ((T_FILE) reduce 171) ((T_FINAL) reduce 171) ((T_FINALLY) reduce 171) ((T_FN) reduce 171) ((T_FOR) reduce 171) ((T_FOREACH) reduce 171) ((T_FUNCTION) reduce 171) ((T_FUNC_C) reduce 171) ((T_GLOBAL) reduce 171) ((T_GOTO) reduce 171) ((T_HALT_COMPILER) reduce 171) ((T_IF) reduce 171) ((T_INC) reduce 171) ((T_INCLUDE) reduce 171) ((T_INCLUDE_ONCE) reduce 171) ((T_INLINE_HTML) reduce 171) ((T_INTERFACE) reduce 171) ((T_INT_CAST) reduce 171) ((T_ISSET) reduce 171) ((T_LINE) reduce 171) ((T_LIST) reduce 171) ((T_LNUMBER) reduce 171) ((T_MATCH) reduce 171) ((T_METHOD_C) reduce 171) ((T_NAME_FULLY_QUALIFIED) reduce 171) ((T_NAME_QUALIFIED) reduce 171) ((T_NAME_RELATIVE) reduce 171) ((T_NEW) reduce 171) ((T_NS_C) reduce 171) ((T_OBJECT_CAST) reduce 171) ((T_PRINT) reduce 171) ((T_PROPERTY_C) reduce 171) ((T_READONLY) reduce 171) ((T_REQUIRE) reduce 171) ((T_REQUIRE_ONCE) reduce 171) ((T_RETURN) reduce 171) ((T_START_HEREDOC) reduce 171) ((T_STATIC) reduce 171) ((T_STRING) reduce 171) ((T_STRING_CAST) reduce 171) ((T_SWITCH) reduce 171) ((T_THROW) reduce 171) ((T_TRAIT) reduce 171) ((T_TRAIT_C) reduce 171) ((T_TRY) reduce 171) ((T_UNSET) reduce 171) ((T_UNSET_CAST) reduce 171) ((T_VARIABLE) reduce 171) ((T_VOID_CAST) reduce 171) ((T_WHILE) reduce 171) ((T_YIELD) reduce 171) ((T_YIELD_FROM) reduce 171) (("[") reduce 171) (("`") reduce 171) (("{") reduce 171) (("~") reduce 171)) 18048 ((("!") reduce 177) (("\"") reduce 177) (("$") reduce 177) (("(") reduce 177) (("+") reduce 177) (("-") reduce 177) ((";") reduce 177) (("@") reduce 177) ((T_ABSTRACT) reduce 177) ((T_ARRAY) reduce 177) ((T_ARRAY_CAST) reduce 177) ((T_ATTRIBUTE) reduce 177) ((T_BOOL_CAST) reduce 177) ((T_BREAK) reduce 177) ((T_CATCH) shift) ((T_CLASS) reduce 177) ((T_CLASS_C) reduce 177) ((T_CLONE) reduce 177) ((T_CONSTANT_ENCAPSED_STRING) reduce 177) ((T_CONTINUE) reduce 177) ((T_DEC) reduce 177) ((T_DECLARE) reduce 177) ((T_DIR) reduce 177) ((T_DNUMBER) reduce 177) ((T_DO) reduce 177) ((T_DOUBLE_CAST) reduce 177) ((T_ECHO) reduce 177) ((T_ELSE) reduce 177) ((T_ELSEIF) reduce 177) ((T_EMPTY) reduce 177) ((T_ENDWHILE) reduce 177) ((T_ENUM) reduce 177) ((T_EVAL) reduce 177) ((T_EXIT) reduce 177) ((T_FILE) reduce 177) ((T_FINAL) reduce 177) ((T_FINALLY) shift) ((T_FN) reduce 177) ((T_FOR) reduce 177) ((T_FOREACH) reduce 177) ((T_FUNCTION) reduce 177) ((T_FUNC_C) reduce 177) ((T_GLOBAL) reduce 177) ((T_GOTO) reduce 177) ((T_HALT_COMPILER) reduce 177) ((T_IF) reduce 177) ((T_INC) reduce 177) ((T_INCLUDE) reduce 177) ((T_INCLUDE_ONCE) reduce 177) ((T_INLINE_HTML) reduce 177) ((T_INTERFACE) reduce 177) ((T_INT_CAST) reduce 177) ((T_ISSET) reduce 177) ((T_LINE) reduce 177) ((T_LIST) reduce 177) ((T_LNUMBER) reduce 177) ((T_MATCH) reduce 177) ((T_METHOD_C) reduce 177) ((T_NAME_FULLY_QUALIFIED) reduce 177) ((T_NAME_QUALIFIED) reduce 177) ((T_NAME_RELATIVE) reduce 177) ((T_NEW) reduce 177) ((T_NS_C) reduce 177) ((T_OBJECT_CAST) reduce 177) ((T_PRINT) reduce 177) ((T_PROPERTY_C) reduce 177) ((T_READONLY) reduce 177) ((T_REQUIRE) reduce 177) ((T_REQUIRE_ONCE) reduce 177) ((T_RETURN) reduce 177) ((T_START_HEREDOC) reduce 177) ((T_STATIC) reduce 177) ((T_STRING) reduce 177) ((T_STRING_CAST) reduce 177) ((T_SWITCH) reduce 177) ((T_THROW) reduce 177) ((T_TRAIT) reduce 177) ((T_TRAIT_C) reduce 177) ((T_TRY) reduce 177) ((T_UNSET) reduce 177) ((T_UNSET_CAST) reduce 177) ((T_VARIABLE) reduce 177) ((T_VOID_CAST) reduce 177) ((T_WHILE) reduce 177) ((T_YIELD) reduce 177) ((T_YIELD_FROM) reduce 177) (("[") reduce 177) (("`") reduce 177) (("{") reduce 177) (("~") reduce 177)) 18051 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_ELSE) reduce 167) ((T_ELSEIF) reduce 167) ((T_EMPTY) reduce 167) ((T_ENDWHILE) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_VOID_CAST) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 18054 ((("!") reduce 178) (("\"") reduce 178) (("$") reduce 178) (("(") reduce 178) (("+") reduce 178) (("-") reduce 178) ((";") reduce 178) (("@") reduce 178) ((T_ABSTRACT) reduce 178) ((T_ARRAY) reduce 178) ((T_ARRAY_CAST) reduce 178) ((T_ATTRIBUTE) reduce 178) ((T_BOOL_CAST) reduce 178) ((T_BREAK) reduce 178) ((T_CLASS) reduce 178) ((T_CLASS_C) reduce 178) ((T_CLONE) reduce 178) ((T_CONSTANT_ENCAPSED_STRING) reduce 178) ((T_CONTINUE) reduce 178) ((T_DEC) reduce 178) ((T_DECLARE) reduce 178) ((T_DIR) reduce 178) ((T_DNUMBER) reduce 178) ((T_DO) reduce 178) ((T_DOUBLE_CAST) reduce 178) ((T_ECHO) reduce 178) ((T_ELSE) reduce 178) ((T_ELSEIF) reduce 178) ((T_EMPTY) reduce 178) ((T_ENDWHILE) reduce 178) ((T_ENUM) reduce 178) ((T_EVAL) reduce 178) ((T_EXIT) reduce 178) ((T_FILE) reduce 178) ((T_FINAL) reduce 178) ((T_FN) reduce 178) ((T_FOR) reduce 178) ((T_FOREACH) reduce 178) ((T_FUNCTION) reduce 178) ((T_FUNC_C) reduce 178) ((T_GLOBAL) reduce 178) ((T_GOTO) reduce 178) ((T_HALT_COMPILER) reduce 178) ((T_IF) reduce 178) ((T_INC) reduce 178) ((T_INCLUDE) reduce 178) ((T_INCLUDE_ONCE) reduce 178) ((T_INLINE_HTML) reduce 178) ((T_INTERFACE) reduce 178) ((T_INT_CAST) reduce 178) ((T_ISSET) reduce 178) ((T_LINE) reduce 178) ((T_LIST) reduce 178) ((T_LNUMBER) reduce 178) ((T_MATCH) reduce 178) ((T_METHOD_C) reduce 178) ((T_NAME_FULLY_QUALIFIED) reduce 178) ((T_NAME_QUALIFIED) reduce 178) ((T_NAME_RELATIVE) reduce 178) ((T_NEW) reduce 178) ((T_NS_C) reduce 178) ((T_OBJECT_CAST) reduce 178) ((T_PRINT) reduce 178) ((T_PROPERTY_C) reduce 178) ((T_READONLY) reduce 178) ((T_REQUIRE) reduce 178) ((T_REQUIRE_ONCE) reduce 178) ((T_RETURN) reduce 178) ((T_START_HEREDOC) reduce 178) ((T_STATIC) reduce 178) ((T_STRING) reduce 178) ((T_STRING_CAST) reduce 178) ((T_SWITCH) reduce 178) ((T_THROW) reduce 178) ((T_TRAIT) reduce 178) ((T_TRAIT_C) reduce 178) ((T_TRY) reduce 178) ((T_UNSET) reduce 178) ((T_UNSET_CAST) reduce 178) ((T_VARIABLE) reduce 178) ((T_VOID_CAST) reduce 178) ((T_WHILE) reduce 178) ((T_YIELD) reduce 178) ((T_YIELD_FROM) reduce 178) (("[") reduce 178) (("`") reduce 178) (("{") reduce 178) (("~") reduce 178)) 18056 (((")") reduce 94) ((T_VARIABLE) reduce 94) (("|") reduce 94)) 18057 (((")") reduce 93) ((T_VARIABLE) reduce 93) (("|") reduce 93)) 18058 (((")") reduce 95) ((T_VARIABLE) reduce 95) (("|") reduce 95)) 18059 (((")") reduce 509) ((T_VARIABLE) reduce 509) (("|") reduce 509)) 18060 (((")") reduce 92) ((T_VARIABLE) reduce 92) (("|") reduce 92)) 18061 (((")") reduce 175) ((T_VARIABLE) shift) (("|") shift)) 18062 (((")") reduce 173) ((T_VARIABLE) reduce 173) (("|") reduce 173)) 18063 (((")") reduce 510) ((T_VARIABLE) reduce 510) (("|") reduce 510)) 18064 (((")") reduce 176)) 18067 (((")") reduce 174) ((T_VARIABLE) reduce 174) (("|") reduce 174)) 18071 ((("!") reduce 172) (("\"") reduce 172) (("$") reduce 172) (("(") reduce 172) (("+") reduce 172) (("-") reduce 172) ((";") reduce 172) (("@") reduce 172) ((T_ABSTRACT) reduce 172) ((T_ARRAY) reduce 172) ((T_ARRAY_CAST) reduce 172) ((T_ATTRIBUTE) reduce 172) ((T_BOOL_CAST) reduce 172) ((T_BREAK) reduce 172) ((T_CATCH) reduce 172) ((T_CLASS) reduce 172) ((T_CLASS_C) reduce 172) ((T_CLONE) reduce 172) ((T_CONSTANT_ENCAPSED_STRING) reduce 172) ((T_CONTINUE) reduce 172) ((T_DEC) reduce 172) ((T_DECLARE) reduce 172) ((T_DIR) reduce 172) ((T_DNUMBER) reduce 172) ((T_DO) reduce 172) ((T_DOUBLE_CAST) reduce 172) ((T_ECHO) reduce 172) ((T_ELSE) reduce 172) ((T_ELSEIF) reduce 172) ((T_EMPTY) reduce 172) ((T_ENDWHILE) reduce 172) ((T_ENUM) reduce 172) ((T_EVAL) reduce 172) ((T_EXIT) reduce 172) ((T_FILE) reduce 172) ((T_FINAL) reduce 172) ((T_FINALLY) reduce 172) ((T_FN) reduce 172) ((T_FOR) reduce 172) ((T_FOREACH) reduce 172) ((T_FUNCTION) reduce 172) ((T_FUNC_C) reduce 172) ((T_GLOBAL) reduce 172) ((T_GOTO) reduce 172) ((T_HALT_COMPILER) reduce 172) ((T_IF) reduce 172) ((T_INC) reduce 172) ((T_INCLUDE) reduce 172) ((T_INCLUDE_ONCE) reduce 172) ((T_INLINE_HTML) reduce 172) ((T_INTERFACE) reduce 172) ((T_INT_CAST) reduce 172) ((T_ISSET) reduce 172) ((T_LINE) reduce 172) ((T_LIST) reduce 172) ((T_LNUMBER) reduce 172) ((T_MATCH) reduce 172) ((T_METHOD_C) reduce 172) ((T_NAME_FULLY_QUALIFIED) reduce 172) ((T_NAME_QUALIFIED) reduce 172) ((T_NAME_RELATIVE) reduce 172) ((T_NEW) reduce 172) ((T_NS_C) reduce 172) ((T_OBJECT_CAST) reduce 172) ((T_PRINT) reduce 172) ((T_PROPERTY_C) reduce 172) ((T_READONLY) reduce 172) ((T_REQUIRE) reduce 172) ((T_REQUIRE_ONCE) reduce 172) ((T_RETURN) reduce 172) ((T_START_HEREDOC) reduce 172) ((T_STATIC) reduce 172) ((T_STRING) reduce 172) ((T_STRING_CAST) reduce 172) ((T_SWITCH) reduce 172) ((T_THROW) reduce 172) ((T_TRAIT) reduce 172) ((T_TRAIT_C) reduce 172) ((T_TRY) reduce 172) ((T_UNSET) reduce 172) ((T_UNSET_CAST) reduce 172) ((T_VARIABLE) reduce 172) ((T_VOID_CAST) reduce 172) ((T_WHILE) reduce 172) ((T_YIELD) reduce 172) ((T_YIELD_FROM) reduce 172) (("[") reduce 172) (("`") reduce 172) (("{") reduce 172) (("~") reduce 172)) 18074 (((":") shift) (("{") shift)) 18075 (((";") shift) ((T_CASE) reduce 228) ((T_DEFAULT) reduce 228) ((T_ENDSWITCH) reduce 228)) 18076 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_ELSE) reduce 153) ((T_ELSEIF) reduce 153) ((T_EMPTY) reduce 153) ((T_ENDWHILE) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_VOID_CAST) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 18077 (((";") shift) ((T_CASE) reduce 228) ((T_DEFAULT) reduce 228) (("}") reduce 228)) 18078 (((T_CASE) reduce 228) ((T_DEFAULT) reduce 228) (("}") reduce 228)) 18079 (((T_CASE) shift) ((T_DEFAULT) shift) (("}") shift)) 18081 (((":") shift) ((";") shift)) 18082 ((("!") reduce 224) (("\"") reduce 224) (("$") reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_ELSE) reduce 224) ((T_ELSEIF) reduce 224) ((T_EMPTY) reduce 224) ((T_ENDWHILE) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_VARIABLE) reduce 224) ((T_VOID_CAST) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("~") reduce 224)) 18083 ((("!") reduce 142) (("\"") reduce 142) (("$") reduce 142) (("(") reduce 142) (("+") reduce 142) (("-") reduce 142) ((";") reduce 142) (("@") reduce 142) ((T_ABSTRACT) reduce 142) ((T_ARRAY) reduce 142) ((T_ARRAY_CAST) reduce 142) ((T_ATTRIBUTE) reduce 142) ((T_BOOL_CAST) reduce 142) ((T_BREAK) reduce 142) ((T_CASE) reduce 142) ((T_CLASS) reduce 142) ((T_CLASS_C) reduce 142) ((T_CLONE) reduce 142) ((T_CONSTANT_ENCAPSED_STRING) reduce 142) ((T_CONTINUE) reduce 142) ((T_DEC) reduce 142) ((T_DECLARE) reduce 142) ((T_DEFAULT) reduce 142) ((T_DIR) reduce 142) ((T_DNUMBER) reduce 142) ((T_DO) reduce 142) ((T_DOUBLE_CAST) reduce 142) ((T_ECHO) reduce 142) ((T_EMPTY) reduce 142) ((T_ENUM) reduce 142) ((T_EVAL) reduce 142) ((T_EXIT) reduce 142) ((T_FILE) reduce 142) ((T_FINAL) reduce 142) ((T_FN) reduce 142) ((T_FOR) reduce 142) ((T_FOREACH) reduce 142) ((T_FUNCTION) reduce 142) ((T_FUNC_C) reduce 142) ((T_GLOBAL) reduce 142) ((T_GOTO) reduce 142) ((T_HALT_COMPILER) reduce 142) ((T_IF) reduce 142) ((T_INC) reduce 142) ((T_INCLUDE) reduce 142) ((T_INCLUDE_ONCE) reduce 142) ((T_INLINE_HTML) reduce 142) ((T_INTERFACE) reduce 142) ((T_INT_CAST) reduce 142) ((T_ISSET) reduce 142) ((T_LINE) reduce 142) ((T_LIST) reduce 142) ((T_LNUMBER) reduce 142) ((T_MATCH) reduce 142) ((T_METHOD_C) reduce 142) ((T_NAME_FULLY_QUALIFIED) reduce 142) ((T_NAME_QUALIFIED) reduce 142) ((T_NAME_RELATIVE) reduce 142) ((T_NEW) reduce 142) ((T_NS_C) reduce 142) ((T_OBJECT_CAST) reduce 142) ((T_PRINT) reduce 142) ((T_PROPERTY_C) reduce 142) ((T_READONLY) reduce 142) ((T_REQUIRE) reduce 142) ((T_REQUIRE_ONCE) reduce 142) ((T_RETURN) reduce 142) ((T_START_HEREDOC) reduce 142) ((T_STATIC) reduce 142) ((T_STRING) reduce 142) ((T_STRING_CAST) reduce 142) ((T_SWITCH) reduce 142) ((T_THROW) reduce 142) ((T_TRAIT) reduce 142) ((T_TRAIT_C) reduce 142) ((T_TRY) reduce 142) ((T_UNSET) reduce 142) ((T_UNSET_CAST) reduce 142) ((T_VARIABLE) reduce 142) ((T_VOID_CAST) reduce 142) ((T_WHILE) reduce 142) ((T_YIELD) reduce 142) ((T_YIELD_FROM) reduce 142) (("[") reduce 142) (("`") reduce 142) (("{") reduce 142) (("}") reduce 142) (("~") reduce 142)) 18085 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CASE) reduce 232) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DEFAULT) reduce 232) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_VOID_CAST) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("}") reduce 232) (("~") shift)) 18086 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CASE) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DEFAULT) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_EMPTY) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_VOID_CAST) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("}") reduce 166) (("~") reduce 166)) 18100 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CASE) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DEFAULT) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_EMPTY) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_VOID_CAST) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("}") reduce 160) (("~") reduce 160)) 18111 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CASE) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DEFAULT) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_EMPTY) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_VOID_CAST) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("}") reduce 149) (("~") reduce 149)) 18113 ((("!") reduce 144) (("\"") reduce 144) (("$") reduce 144) (("(") reduce 144) (("+") reduce 144) (("-") reduce 144) ((";") reduce 144) (("@") reduce 144) ((T_ABSTRACT) reduce 144) ((T_ARRAY) reduce 144) ((T_ARRAY_CAST) reduce 144) ((T_ATTRIBUTE) reduce 144) ((T_BOOL_CAST) reduce 144) ((T_BREAK) reduce 144) ((T_CASE) reduce 144) ((T_CLASS) reduce 144) ((T_CLASS_C) reduce 144) ((T_CLONE) reduce 144) ((T_CONSTANT_ENCAPSED_STRING) reduce 144) ((T_CONTINUE) reduce 144) ((T_DEC) reduce 144) ((T_DECLARE) reduce 144) ((T_DEFAULT) reduce 144) ((T_DIR) reduce 144) ((T_DNUMBER) reduce 144) ((T_DO) reduce 144) ((T_DOUBLE_CAST) reduce 144) ((T_ECHO) reduce 144) ((T_EMPTY) reduce 144) ((T_ENUM) reduce 144) ((T_EVAL) reduce 144) ((T_EXIT) reduce 144) ((T_FILE) reduce 144) ((T_FINAL) reduce 144) ((T_FN) reduce 144) ((T_FOR) reduce 144) ((T_FOREACH) reduce 144) ((T_FUNCTION) reduce 144) ((T_FUNC_C) reduce 144) ((T_GLOBAL) reduce 144) ((T_GOTO) reduce 144) ((T_HALT_COMPILER) reduce 144) ((T_IF) reduce 144) ((T_INC) reduce 144) ((T_INCLUDE) reduce 144) ((T_INCLUDE_ONCE) reduce 144) ((T_INLINE_HTML) reduce 144) ((T_INTERFACE) reduce 144) ((T_INT_CAST) reduce 144) ((T_ISSET) reduce 144) ((T_LINE) reduce 144) ((T_LIST) reduce 144) ((T_LNUMBER) reduce 144) ((T_MATCH) reduce 144) ((T_METHOD_C) reduce 144) ((T_NAME_FULLY_QUALIFIED) reduce 144) ((T_NAME_QUALIFIED) reduce 144) ((T_NAME_RELATIVE) reduce 144) ((T_NEW) reduce 144) ((T_NS_C) reduce 144) ((T_OBJECT_CAST) reduce 144) ((T_PRINT) reduce 144) ((T_PROPERTY_C) reduce 144) ((T_READONLY) reduce 144) ((T_REQUIRE) reduce 144) ((T_REQUIRE_ONCE) reduce 144) ((T_RETURN) reduce 144) ((T_START_HEREDOC) reduce 144) ((T_STATIC) reduce 144) ((T_STRING) reduce 144) ((T_STRING_CAST) reduce 144) ((T_SWITCH) reduce 144) ((T_THROW) reduce 144) ((T_TRAIT) reduce 144) ((T_TRAIT_C) reduce 144) ((T_TRY) reduce 144) ((T_UNSET) reduce 144) ((T_UNSET_CAST) reduce 144) ((T_VARIABLE) reduce 144) ((T_VOID_CAST) reduce 144) ((T_WHILE) reduce 144) ((T_YIELD) reduce 144) ((T_YIELD_FROM) reduce 144) (("[") reduce 144) (("`") reduce 144) (("{") reduce 144) (("}") reduce 144) (("~") reduce 144)) 18115 ((("!") reduce 104) (("\"") reduce 104) (("$") reduce 104) (("(") reduce 104) (("+") reduce 104) (("-") reduce 104) ((";") reduce 104) (("@") reduce 104) ((T_ABSTRACT) reduce 104) ((T_ARRAY) reduce 104) ((T_ARRAY_CAST) reduce 104) ((T_ATTRIBUTE) reduce 104) ((T_BOOL_CAST) reduce 104) ((T_BREAK) reduce 104) ((T_CASE) reduce 104) ((T_CLASS) reduce 104) ((T_CLASS_C) reduce 104) ((T_CLONE) reduce 104) ((T_CONSTANT_ENCAPSED_STRING) reduce 104) ((T_CONTINUE) reduce 104) ((T_DEC) reduce 104) ((T_DECLARE) reduce 104) ((T_DEFAULT) reduce 104) ((T_DIR) reduce 104) ((T_DNUMBER) reduce 104) ((T_DO) reduce 104) ((T_DOUBLE_CAST) reduce 104) ((T_ECHO) reduce 104) ((T_EMPTY) reduce 104) ((T_ENUM) reduce 104) ((T_EVAL) reduce 104) ((T_EXIT) reduce 104) ((T_FILE) reduce 104) ((T_FINAL) reduce 104) ((T_FN) reduce 104) ((T_FOR) reduce 104) ((T_FOREACH) reduce 104) ((T_FUNCTION) reduce 104) ((T_FUNC_C) reduce 104) ((T_GLOBAL) reduce 104) ((T_GOTO) reduce 104) ((T_HALT_COMPILER) reduce 104) ((T_IF) reduce 104) ((T_INC) reduce 104) ((T_INCLUDE) reduce 104) ((T_INCLUDE_ONCE) reduce 104) ((T_INLINE_HTML) reduce 104) ((T_INTERFACE) reduce 104) ((T_INT_CAST) reduce 104) ((T_ISSET) reduce 104) ((T_LINE) reduce 104) ((T_LIST) reduce 104) ((T_LNUMBER) reduce 104) ((T_MATCH) reduce 104) ((T_METHOD_C) reduce 104) ((T_NAME_FULLY_QUALIFIED) reduce 104) ((T_NAME_QUALIFIED) reduce 104) ((T_NAME_RELATIVE) reduce 104) ((T_NEW) reduce 104) ((T_NS_C) reduce 104) ((T_OBJECT_CAST) reduce 104) ((T_PRINT) reduce 104) ((T_PROPERTY_C) reduce 104) ((T_READONLY) reduce 104) ((T_REQUIRE) reduce 104) ((T_REQUIRE_ONCE) reduce 104) ((T_RETURN) reduce 104) ((T_START_HEREDOC) reduce 104) ((T_STATIC) reduce 104) ((T_STRING) reduce 104) ((T_STRING_CAST) reduce 104) ((T_SWITCH) reduce 104) ((T_THROW) reduce 104) ((T_TRAIT) reduce 104) ((T_TRAIT_C) reduce 104) ((T_TRY) reduce 104) ((T_UNSET) reduce 104) ((T_UNSET_CAST) reduce 104) ((T_VARIABLE) reduce 104) ((T_VOID_CAST) reduce 104) ((T_WHILE) reduce 104) ((T_YIELD) reduce 104) ((T_YIELD_FROM) reduce 104) (("[") reduce 104) (("`") reduce 104) (("{") reduce 104) (("}") reduce 104) (("~") reduce 104)) 18117 ((("!") reduce 107) (("\"") reduce 107) (("$") reduce 107) (("(") reduce 107) (("+") reduce 107) (("-") reduce 107) ((";") reduce 107) (("@") reduce 107) ((T_ABSTRACT) reduce 107) ((T_ARRAY) reduce 107) ((T_ARRAY_CAST) reduce 107) ((T_ATTRIBUTE) reduce 107) ((T_BOOL_CAST) reduce 107) ((T_BREAK) reduce 107) ((T_CASE) reduce 107) ((T_CLASS) reduce 107) ((T_CLASS_C) reduce 107) ((T_CLONE) reduce 107) ((T_CONSTANT_ENCAPSED_STRING) reduce 107) ((T_CONTINUE) reduce 107) ((T_DEC) reduce 107) ((T_DECLARE) reduce 107) ((T_DEFAULT) reduce 107) ((T_DIR) reduce 107) ((T_DNUMBER) reduce 107) ((T_DO) reduce 107) ((T_DOUBLE_CAST) reduce 107) ((T_ECHO) reduce 107) ((T_EMPTY) reduce 107) ((T_ENUM) reduce 107) ((T_EVAL) reduce 107) ((T_EXIT) reduce 107) ((T_FILE) reduce 107) ((T_FINAL) reduce 107) ((T_FN) reduce 107) ((T_FOR) reduce 107) ((T_FOREACH) reduce 107) ((T_FUNCTION) reduce 107) ((T_FUNC_C) reduce 107) ((T_GLOBAL) reduce 107) ((T_GOTO) reduce 107) ((T_HALT_COMPILER) reduce 107) ((T_IF) reduce 107) ((T_INC) reduce 107) ((T_INCLUDE) reduce 107) ((T_INCLUDE_ONCE) reduce 107) ((T_INLINE_HTML) reduce 107) ((T_INTERFACE) reduce 107) ((T_INT_CAST) reduce 107) ((T_ISSET) reduce 107) ((T_LINE) reduce 107) ((T_LIST) reduce 107) ((T_LNUMBER) reduce 107) ((T_MATCH) reduce 107) ((T_METHOD_C) reduce 107) ((T_NAME_FULLY_QUALIFIED) reduce 107) ((T_NAME_QUALIFIED) reduce 107) ((T_NAME_RELATIVE) reduce 107) ((T_NEW) reduce 107) ((T_NS_C) reduce 107) ((T_OBJECT_CAST) reduce 107) ((T_PRINT) reduce 107) ((T_PROPERTY_C) reduce 107) ((T_READONLY) reduce 107) ((T_REQUIRE) reduce 107) ((T_REQUIRE_ONCE) reduce 107) ((T_RETURN) reduce 107) ((T_START_HEREDOC) reduce 107) ((T_STATIC) reduce 107) ((T_STRING) reduce 107) ((T_STRING_CAST) reduce 107) ((T_SWITCH) reduce 107) ((T_THROW) reduce 107) ((T_TRAIT) reduce 107) ((T_TRAIT_C) reduce 107) ((T_TRY) reduce 107) ((T_UNSET) reduce 107) ((T_UNSET_CAST) reduce 107) ((T_VARIABLE) reduce 107) ((T_VOID_CAST) reduce 107) ((T_WHILE) reduce 107) ((T_YIELD) reduce 107) ((T_YIELD_FROM) reduce 107) (("[") reduce 107) (("`") reduce 107) (("{") reduce 107) (("}") reduce 107) (("~") reduce 107)) 18120 ((("!") reduce 103) (("\"") reduce 103) (("$") reduce 103) (("(") reduce 103) (("+") reduce 103) (("-") reduce 103) ((";") reduce 103) (("@") reduce 103) ((T_ABSTRACT) reduce 103) ((T_ARRAY) reduce 103) ((T_ARRAY_CAST) reduce 103) ((T_ATTRIBUTE) reduce 103) ((T_BOOL_CAST) reduce 103) ((T_BREAK) reduce 103) ((T_CASE) reduce 103) ((T_CLASS) reduce 103) ((T_CLASS_C) reduce 103) ((T_CLONE) reduce 103) ((T_CONSTANT_ENCAPSED_STRING) reduce 103) ((T_CONTINUE) reduce 103) ((T_DEC) reduce 103) ((T_DECLARE) reduce 103) ((T_DEFAULT) reduce 103) ((T_DIR) reduce 103) ((T_DNUMBER) reduce 103) ((T_DO) reduce 103) ((T_DOUBLE_CAST) reduce 103) ((T_ECHO) reduce 103) ((T_EMPTY) reduce 103) ((T_ENUM) reduce 103) ((T_EVAL) reduce 103) ((T_EXIT) reduce 103) ((T_FILE) reduce 103) ((T_FINAL) reduce 103) ((T_FN) reduce 103) ((T_FOR) reduce 103) ((T_FOREACH) reduce 103) ((T_FUNCTION) reduce 103) ((T_FUNC_C) reduce 103) ((T_GLOBAL) reduce 103) ((T_GOTO) reduce 103) ((T_HALT_COMPILER) reduce 103) ((T_IF) reduce 103) ((T_INC) reduce 103) ((T_INCLUDE) reduce 103) ((T_INCLUDE_ONCE) reduce 103) ((T_INLINE_HTML) reduce 103) ((T_INTERFACE) reduce 103) ((T_INT_CAST) reduce 103) ((T_ISSET) reduce 103) ((T_LINE) reduce 103) ((T_LIST) reduce 103) ((T_LNUMBER) reduce 103) ((T_MATCH) reduce 103) ((T_METHOD_C) reduce 103) ((T_NAME_FULLY_QUALIFIED) reduce 103) ((T_NAME_QUALIFIED) reduce 103) ((T_NAME_RELATIVE) reduce 103) ((T_NEW) reduce 103) ((T_NS_C) reduce 103) ((T_OBJECT_CAST) reduce 103) ((T_PRINT) reduce 103) ((T_PROPERTY_C) reduce 103) ((T_READONLY) reduce 103) ((T_REQUIRE) reduce 103) ((T_REQUIRE_ONCE) reduce 103) ((T_RETURN) reduce 103) ((T_START_HEREDOC) reduce 103) ((T_STATIC) reduce 103) ((T_STRING) reduce 103) ((T_STRING_CAST) reduce 103) ((T_SWITCH) reduce 103) ((T_THROW) reduce 103) ((T_TRAIT) reduce 103) ((T_TRAIT_C) reduce 103) ((T_TRY) reduce 103) ((T_UNSET) reduce 103) ((T_UNSET_CAST) reduce 103) ((T_VARIABLE) reduce 103) ((T_VOID_CAST) reduce 103) ((T_WHILE) reduce 103) ((T_YIELD) reduce 103) ((T_YIELD_FROM) reduce 103) (("[") reduce 103) (("`") reduce 103) (("{") reduce 103) (("}") reduce 103) (("~") reduce 103)) 18121 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CASE) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DEFAULT) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_EMPTY) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_VOID_CAST) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("}") reduce 148) (("~") reduce 148)) 18122 ((("!") reduce 246) (("\"") reduce 246) (("$") reduce 246) (("(") reduce 246) (("+") reduce 246) (("-") reduce 246) ((";") reduce 246) (("@") reduce 246) ((T_ABSTRACT) reduce 246) ((T_ARRAY) reduce 246) ((T_ARRAY_CAST) reduce 246) ((T_ATTRIBUTE) reduce 246) ((T_BOOL_CAST) reduce 246) ((T_BREAK) reduce 246) ((T_CASE) reduce 246) ((T_CLASS) reduce 246) ((T_CLASS_C) reduce 246) ((T_CLONE) reduce 246) ((T_CONSTANT_ENCAPSED_STRING) reduce 246) ((T_CONTINUE) reduce 246) ((T_DEC) reduce 246) ((T_DECLARE) reduce 246) ((T_DEFAULT) reduce 246) ((T_DIR) reduce 246) ((T_DNUMBER) reduce 246) ((T_DO) reduce 246) ((T_DOUBLE_CAST) reduce 246) ((T_ECHO) reduce 246) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) reduce 246) ((T_ENUM) reduce 246) ((T_EVAL) reduce 246) ((T_EXIT) reduce 246) ((T_FILE) reduce 246) ((T_FINAL) reduce 246) ((T_FN) reduce 246) ((T_FOR) reduce 246) ((T_FOREACH) reduce 246) ((T_FUNCTION) reduce 246) ((T_FUNC_C) reduce 246) ((T_GLOBAL) reduce 246) ((T_GOTO) reduce 246) ((T_HALT_COMPILER) reduce 246) ((T_IF) reduce 246) ((T_INC) reduce 246) ((T_INCLUDE) reduce 246) ((T_INCLUDE_ONCE) reduce 246) ((T_INLINE_HTML) reduce 246) ((T_INTERFACE) reduce 246) ((T_INT_CAST) reduce 246) ((T_ISSET) reduce 246) ((T_LINE) reduce 246) ((T_LIST) reduce 246) ((T_LNUMBER) reduce 246) ((T_MATCH) reduce 246) ((T_METHOD_C) reduce 246) ((T_NAME_FULLY_QUALIFIED) reduce 246) ((T_NAME_QUALIFIED) reduce 246) ((T_NAME_RELATIVE) reduce 246) ((T_NEW) reduce 246) ((T_NS_C) reduce 246) ((T_OBJECT_CAST) reduce 246) ((T_PRINT) reduce 246) ((T_PROPERTY_C) reduce 246) ((T_READONLY) reduce 246) ((T_REQUIRE) reduce 246) ((T_REQUIRE_ONCE) reduce 246) ((T_RETURN) reduce 246) ((T_START_HEREDOC) reduce 246) ((T_STATIC) reduce 246) ((T_STRING) reduce 246) ((T_STRING_CAST) reduce 246) ((T_SWITCH) reduce 246) ((T_THROW) reduce 246) ((T_TRAIT) reduce 246) ((T_TRAIT_C) reduce 246) ((T_TRY) reduce 246) ((T_UNSET) reduce 246) ((T_UNSET_CAST) reduce 246) ((T_VARIABLE) reduce 246) ((T_VOID_CAST) reduce 246) ((T_WHILE) reduce 246) ((T_YIELD) reduce 246) ((T_YIELD_FROM) reduce 246) (("[") reduce 246) (("`") reduce 246) (("{") reduce 246) (("}") reduce 246) (("~") reduce 246)) 18123 ((("!") reduce 141) (("\"") reduce 141) (("$") reduce 141) (("(") reduce 141) (("+") reduce 141) (("-") reduce 141) ((";") reduce 141) (("@") reduce 141) ((T_ABSTRACT) reduce 141) ((T_ARRAY) reduce 141) ((T_ARRAY_CAST) reduce 141) ((T_ATTRIBUTE) reduce 141) ((T_BOOL_CAST) reduce 141) ((T_BREAK) reduce 141) ((T_CASE) reduce 141) ((T_CLASS) reduce 141) ((T_CLASS_C) reduce 141) ((T_CLONE) reduce 141) ((T_CONSTANT_ENCAPSED_STRING) reduce 141) ((T_CONTINUE) reduce 141) ((T_DEC) reduce 141) ((T_DECLARE) reduce 141) ((T_DEFAULT) reduce 141) ((T_DIR) reduce 141) ((T_DNUMBER) reduce 141) ((T_DO) reduce 141) ((T_DOUBLE_CAST) reduce 141) ((T_ECHO) reduce 141) ((T_EMPTY) reduce 141) ((T_ENUM) reduce 141) ((T_EVAL) reduce 141) ((T_EXIT) reduce 141) ((T_FILE) reduce 141) ((T_FINAL) reduce 141) ((T_FN) reduce 141) ((T_FOR) reduce 141) ((T_FOREACH) reduce 141) ((T_FUNCTION) reduce 141) ((T_FUNC_C) reduce 141) ((T_GLOBAL) reduce 141) ((T_GOTO) reduce 141) ((T_HALT_COMPILER) reduce 141) ((T_IF) reduce 141) ((T_INC) reduce 141) ((T_INCLUDE) reduce 141) ((T_INCLUDE_ONCE) reduce 141) ((T_INLINE_HTML) reduce 141) ((T_INTERFACE) reduce 141) ((T_INT_CAST) reduce 141) ((T_ISSET) reduce 141) ((T_LINE) reduce 141) ((T_LIST) reduce 141) ((T_LNUMBER) reduce 141) ((T_MATCH) reduce 141) ((T_METHOD_C) reduce 141) ((T_NAME_FULLY_QUALIFIED) reduce 141) ((T_NAME_QUALIFIED) reduce 141) ((T_NAME_RELATIVE) reduce 141) ((T_NEW) reduce 141) ((T_NS_C) reduce 141) ((T_OBJECT_CAST) reduce 141) ((T_PRINT) reduce 141) ((T_PROPERTY_C) reduce 141) ((T_READONLY) reduce 141) ((T_REQUIRE) reduce 141) ((T_REQUIRE_ONCE) reduce 141) ((T_RETURN) reduce 141) ((T_START_HEREDOC) reduce 141) ((T_STATIC) reduce 141) ((T_STRING) reduce 141) ((T_STRING_CAST) reduce 141) ((T_SWITCH) reduce 141) ((T_THROW) reduce 141) ((T_TRAIT) reduce 141) ((T_TRAIT_C) reduce 141) ((T_TRY) reduce 141) ((T_UNSET) reduce 141) ((T_UNSET_CAST) reduce 141) ((T_VARIABLE) reduce 141) ((T_VOID_CAST) reduce 141) ((T_WHILE) reduce 141) ((T_YIELD) reduce 141) ((T_YIELD_FROM) reduce 141) (("[") reduce 141) (("`") reduce 141) (("{") reduce 141) (("}") reduce 141) (("~") reduce 141)) 18124 ((("!") reduce 106) (("\"") reduce 106) (("$") reduce 106) (("(") reduce 106) (("+") reduce 106) (("-") reduce 106) ((";") reduce 106) (("@") reduce 106) ((T_ABSTRACT) reduce 106) ((T_ARRAY) reduce 106) ((T_ARRAY_CAST) reduce 106) ((T_ATTRIBUTE) reduce 106) ((T_BOOL_CAST) reduce 106) ((T_BREAK) reduce 106) ((T_CASE) reduce 106) ((T_CLASS) reduce 106) ((T_CLASS_C) reduce 106) ((T_CLONE) reduce 106) ((T_CONSTANT_ENCAPSED_STRING) reduce 106) ((T_CONTINUE) reduce 106) ((T_DEC) reduce 106) ((T_DECLARE) reduce 106) ((T_DEFAULT) reduce 106) ((T_DIR) reduce 106) ((T_DNUMBER) reduce 106) ((T_DO) reduce 106) ((T_DOUBLE_CAST) reduce 106) ((T_ECHO) reduce 106) ((T_EMPTY) reduce 106) ((T_ENUM) reduce 106) ((T_EVAL) reduce 106) ((T_EXIT) reduce 106) ((T_FILE) reduce 106) ((T_FINAL) reduce 106) ((T_FN) reduce 106) ((T_FOR) reduce 106) ((T_FOREACH) reduce 106) ((T_FUNCTION) reduce 106) ((T_FUNC_C) reduce 106) ((T_GLOBAL) reduce 106) ((T_GOTO) reduce 106) ((T_HALT_COMPILER) reduce 106) ((T_IF) reduce 106) ((T_INC) reduce 106) ((T_INCLUDE) reduce 106) ((T_INCLUDE_ONCE) reduce 106) ((T_INLINE_HTML) reduce 106) ((T_INTERFACE) reduce 106) ((T_INT_CAST) reduce 106) ((T_ISSET) reduce 106) ((T_LINE) reduce 106) ((T_LIST) reduce 106) ((T_LNUMBER) reduce 106) ((T_MATCH) reduce 106) ((T_METHOD_C) reduce 106) ((T_NAME_FULLY_QUALIFIED) reduce 106) ((T_NAME_QUALIFIED) reduce 106) ((T_NAME_RELATIVE) reduce 106) ((T_NEW) reduce 106) ((T_NS_C) reduce 106) ((T_OBJECT_CAST) reduce 106) ((T_PRINT) reduce 106) ((T_PROPERTY_C) reduce 106) ((T_READONLY) reduce 106) ((T_REQUIRE) reduce 106) ((T_REQUIRE_ONCE) reduce 106) ((T_RETURN) reduce 106) ((T_START_HEREDOC) reduce 106) ((T_STATIC) reduce 106) ((T_STRING) reduce 106) ((T_STRING_CAST) reduce 106) ((T_SWITCH) reduce 106) ((T_THROW) reduce 106) ((T_TRAIT) reduce 106) ((T_TRAIT_C) reduce 106) ((T_TRY) reduce 106) ((T_UNSET) reduce 106) ((T_UNSET_CAST) reduce 106) ((T_VARIABLE) reduce 106) ((T_VOID_CAST) reduce 106) ((T_WHILE) reduce 106) ((T_YIELD) reduce 106) ((T_YIELD_FROM) reduce 106) (("[") reduce 106) (("`") reduce 106) (("{") reduce 106) (("}") reduce 106) (("~") reduce 106)) 18125 ((("!") reduce 143) (("\"") reduce 143) (("$") reduce 143) (("(") reduce 143) (("+") reduce 143) (("-") reduce 143) ((";") reduce 143) (("@") reduce 143) ((T_ABSTRACT) reduce 143) ((T_ARRAY) reduce 143) ((T_ARRAY_CAST) reduce 143) ((T_ATTRIBUTE) reduce 143) ((T_BOOL_CAST) reduce 143) ((T_BREAK) reduce 143) ((T_CASE) reduce 143) ((T_CLASS) reduce 143) ((T_CLASS_C) reduce 143) ((T_CLONE) reduce 143) ((T_CONSTANT_ENCAPSED_STRING) reduce 143) ((T_CONTINUE) reduce 143) ((T_DEC) reduce 143) ((T_DECLARE) reduce 143) ((T_DEFAULT) reduce 143) ((T_DIR) reduce 143) ((T_DNUMBER) reduce 143) ((T_DO) reduce 143) ((T_DOUBLE_CAST) reduce 143) ((T_ECHO) reduce 143) ((T_EMPTY) reduce 143) ((T_ENUM) reduce 143) ((T_EVAL) reduce 143) ((T_EXIT) reduce 143) ((T_FILE) reduce 143) ((T_FINAL) reduce 143) ((T_FN) reduce 143) ((T_FOR) reduce 143) ((T_FOREACH) reduce 143) ((T_FUNCTION) reduce 143) ((T_FUNC_C) reduce 143) ((T_GLOBAL) reduce 143) ((T_GOTO) reduce 143) ((T_HALT_COMPILER) reduce 143) ((T_IF) reduce 143) ((T_INC) reduce 143) ((T_INCLUDE) reduce 143) ((T_INCLUDE_ONCE) reduce 143) ((T_INLINE_HTML) reduce 143) ((T_INTERFACE) reduce 143) ((T_INT_CAST) reduce 143) ((T_ISSET) reduce 143) ((T_LINE) reduce 143) ((T_LIST) reduce 143) ((T_LNUMBER) reduce 143) ((T_MATCH) reduce 143) ((T_METHOD_C) reduce 143) ((T_NAME_FULLY_QUALIFIED) reduce 143) ((T_NAME_QUALIFIED) reduce 143) ((T_NAME_RELATIVE) reduce 143) ((T_NEW) reduce 143) ((T_NS_C) reduce 143) ((T_OBJECT_CAST) reduce 143) ((T_PRINT) reduce 143) ((T_PROPERTY_C) reduce 143) ((T_READONLY) reduce 143) ((T_REQUIRE) reduce 143) ((T_REQUIRE_ONCE) reduce 143) ((T_RETURN) reduce 143) ((T_START_HEREDOC) reduce 143) ((T_STATIC) reduce 143) ((T_STRING) reduce 143) ((T_STRING_CAST) reduce 143) ((T_SWITCH) reduce 143) ((T_THROW) reduce 143) ((T_TRAIT) reduce 143) ((T_TRAIT_C) reduce 143) ((T_TRY) reduce 143) ((T_UNSET) reduce 143) ((T_UNSET_CAST) reduce 143) ((T_VARIABLE) reduce 143) ((T_VOID_CAST) reduce 143) ((T_WHILE) reduce 143) ((T_YIELD) reduce 143) ((T_YIELD_FROM) reduce 143) (("[") reduce 143) (("`") reduce 143) (("{") reduce 143) (("}") reduce 143) (("~") reduce 143)) 18126 ((("!") reduce 105) (("\"") reduce 105) (("$") reduce 105) (("(") reduce 105) (("+") reduce 105) (("-") reduce 105) ((";") reduce 105) (("@") reduce 105) ((T_ABSTRACT) reduce 105) ((T_ARRAY) reduce 105) ((T_ARRAY_CAST) reduce 105) ((T_ATTRIBUTE) reduce 105) ((T_BOOL_CAST) reduce 105) ((T_BREAK) reduce 105) ((T_CASE) reduce 105) ((T_CLASS) reduce 105) ((T_CLASS_C) reduce 105) ((T_CLONE) reduce 105) ((T_CONSTANT_ENCAPSED_STRING) reduce 105) ((T_CONTINUE) reduce 105) ((T_DEC) reduce 105) ((T_DECLARE) reduce 105) ((T_DEFAULT) reduce 105) ((T_DIR) reduce 105) ((T_DNUMBER) reduce 105) ((T_DO) reduce 105) ((T_DOUBLE_CAST) reduce 105) ((T_ECHO) reduce 105) ((T_EMPTY) reduce 105) ((T_ENUM) reduce 105) ((T_EVAL) reduce 105) ((T_EXIT) reduce 105) ((T_FILE) reduce 105) ((T_FINAL) reduce 105) ((T_FN) reduce 105) ((T_FOR) reduce 105) ((T_FOREACH) reduce 105) ((T_FUNCTION) reduce 105) ((T_FUNC_C) reduce 105) ((T_GLOBAL) reduce 105) ((T_GOTO) reduce 105) ((T_HALT_COMPILER) reduce 105) ((T_IF) reduce 105) ((T_INC) reduce 105) ((T_INCLUDE) reduce 105) ((T_INCLUDE_ONCE) reduce 105) ((T_INLINE_HTML) reduce 105) ((T_INTERFACE) reduce 105) ((T_INT_CAST) reduce 105) ((T_ISSET) reduce 105) ((T_LINE) reduce 105) ((T_LIST) reduce 105) ((T_LNUMBER) reduce 105) ((T_MATCH) reduce 105) ((T_METHOD_C) reduce 105) ((T_NAME_FULLY_QUALIFIED) reduce 105) ((T_NAME_QUALIFIED) reduce 105) ((T_NAME_RELATIVE) reduce 105) ((T_NEW) reduce 105) ((T_NS_C) reduce 105) ((T_OBJECT_CAST) reduce 105) ((T_PRINT) reduce 105) ((T_PROPERTY_C) reduce 105) ((T_READONLY) reduce 105) ((T_REQUIRE) reduce 105) ((T_REQUIRE_ONCE) reduce 105) ((T_RETURN) reduce 105) ((T_START_HEREDOC) reduce 105) ((T_STATIC) reduce 105) ((T_STRING) reduce 105) ((T_STRING_CAST) reduce 105) ((T_SWITCH) reduce 105) ((T_THROW) reduce 105) ((T_TRAIT) reduce 105) ((T_TRAIT_C) reduce 105) ((T_TRY) reduce 105) ((T_UNSET) reduce 105) ((T_UNSET_CAST) reduce 105) ((T_VARIABLE) reduce 105) ((T_VOID_CAST) reduce 105) ((T_WHILE) reduce 105) ((T_YIELD) reduce 105) ((T_YIELD_FROM) reduce 105) (("[") reduce 105) (("`") reduce 105) (("{") reduce 105) (("}") reduce 105) (("~") reduce 105)) 18129 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CASE) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DEFAULT) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_EMPTY) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_VOID_CAST) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("}") reduce 147) (("~") reduce 147)) 18135 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CASE) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DEFAULT) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_ELSE) reduce 166) ((T_ELSEIF) reduce 166) ((T_EMPTY) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_VOID_CAST) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("}") reduce 166) (("~") reduce 166)) 18145 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CASE) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DEFAULT) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_ELSE) reduce 160) ((T_ELSEIF) reduce 160) ((T_EMPTY) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_VOID_CAST) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("}") reduce 160) (("~") reduce 160)) 18154 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CASE) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DEFAULT) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_ELSE) reduce 149) ((T_ELSEIF) reduce 149) ((T_EMPTY) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_VOID_CAST) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("}") reduce 149) (("~") reduce 149)) 18157 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CASE) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DEFAULT) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_ELSE) reduce 148) ((T_ELSEIF) reduce 148) ((T_EMPTY) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_VOID_CAST) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("}") reduce 148) (("~") reduce 148)) 18159 ((("!") reduce 245) (("\"") reduce 245) (("$") reduce 245) (("(") reduce 245) (("+") reduce 245) (("-") reduce 245) ((";") reduce 245) (("@") reduce 245) ((T_ABSTRACT) reduce 245) ((T_ARRAY) reduce 245) ((T_ARRAY_CAST) reduce 245) ((T_ATTRIBUTE) reduce 245) ((T_BOOL_CAST) reduce 245) ((T_BREAK) reduce 245) ((T_CASE) reduce 245) ((T_CLASS) reduce 245) ((T_CLASS_C) reduce 245) ((T_CLONE) reduce 245) ((T_CONSTANT_ENCAPSED_STRING) reduce 245) ((T_CONTINUE) reduce 245) ((T_DEC) reduce 245) ((T_DECLARE) reduce 245) ((T_DEFAULT) reduce 245) ((T_DIR) reduce 245) ((T_DNUMBER) reduce 245) ((T_DO) reduce 245) ((T_DOUBLE_CAST) reduce 245) ((T_ECHO) reduce 245) ((T_ELSE) reduce 245) ((T_ELSEIF) reduce 245) ((T_EMPTY) reduce 245) ((T_ENUM) reduce 245) ((T_EVAL) reduce 245) ((T_EXIT) reduce 245) ((T_FILE) reduce 245) ((T_FINAL) reduce 245) ((T_FN) reduce 245) ((T_FOR) reduce 245) ((T_FOREACH) reduce 245) ((T_FUNCTION) reduce 245) ((T_FUNC_C) reduce 245) ((T_GLOBAL) reduce 245) ((T_GOTO) reduce 245) ((T_HALT_COMPILER) reduce 245) ((T_IF) reduce 245) ((T_INC) reduce 245) ((T_INCLUDE) reduce 245) ((T_INCLUDE_ONCE) reduce 245) ((T_INLINE_HTML) reduce 245) ((T_INTERFACE) reduce 245) ((T_INT_CAST) reduce 245) ((T_ISSET) reduce 245) ((T_LINE) reduce 245) ((T_LIST) reduce 245) ((T_LNUMBER) reduce 245) ((T_MATCH) reduce 245) ((T_METHOD_C) reduce 245) ((T_NAME_FULLY_QUALIFIED) reduce 245) ((T_NAME_QUALIFIED) reduce 245) ((T_NAME_RELATIVE) reduce 245) ((T_NEW) reduce 245) ((T_NS_C) reduce 245) ((T_OBJECT_CAST) reduce 245) ((T_PRINT) reduce 245) ((T_PROPERTY_C) reduce 245) ((T_READONLY) reduce 245) ((T_REQUIRE) reduce 245) ((T_REQUIRE_ONCE) reduce 245) ((T_RETURN) reduce 245) ((T_START_HEREDOC) reduce 245) ((T_STATIC) reduce 245) ((T_STRING) reduce 245) ((T_STRING_CAST) reduce 245) ((T_SWITCH) reduce 245) ((T_THROW) reduce 245) ((T_TRAIT) reduce 245) ((T_TRAIT_C) reduce 245) ((T_TRY) reduce 245) ((T_UNSET) reduce 245) ((T_UNSET_CAST) reduce 245) ((T_VARIABLE) reduce 245) ((T_VOID_CAST) reduce 245) ((T_WHILE) reduce 245) ((T_YIELD) reduce 245) ((T_YIELD_FROM) reduce 245) (("[") reduce 245) (("`") reduce 245) (("{") reduce 245) (("}") reduce 245) (("~") reduce 245)) 18162 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CASE) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DEFAULT) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_ELSE) reduce 147) ((T_ELSEIF) reduce 147) ((T_EMPTY) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_VOID_CAST) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("}") reduce 147) (("~") reduce 147)) 18164 ((("!") reduce 247) (("\"") reduce 247) (("$") reduce 247) (("(") reduce 247) (("+") reduce 247) (("-") reduce 247) ((";") reduce 247) (("@") reduce 247) ((T_ABSTRACT) reduce 247) ((T_ARRAY) reduce 247) ((T_ARRAY_CAST) reduce 247) ((T_ATTRIBUTE) reduce 247) ((T_BOOL_CAST) reduce 247) ((T_BREAK) reduce 247) ((T_CASE) reduce 247) ((T_CLASS) reduce 247) ((T_CLASS_C) reduce 247) ((T_CLONE) reduce 247) ((T_CONSTANT_ENCAPSED_STRING) reduce 247) ((T_CONTINUE) reduce 247) ((T_DEC) reduce 247) ((T_DECLARE) reduce 247) ((T_DEFAULT) reduce 247) ((T_DIR) reduce 247) ((T_DNUMBER) reduce 247) ((T_DO) reduce 247) ((T_DOUBLE_CAST) reduce 247) ((T_ECHO) reduce 247) ((T_ELSE) reduce 247) ((T_ELSEIF) reduce 247) ((T_EMPTY) reduce 247) ((T_ENUM) reduce 247) ((T_EVAL) reduce 247) ((T_EXIT) reduce 247) ((T_FILE) reduce 247) ((T_FINAL) reduce 247) ((T_FN) reduce 247) ((T_FOR) reduce 247) ((T_FOREACH) reduce 247) ((T_FUNCTION) reduce 247) ((T_FUNC_C) reduce 247) ((T_GLOBAL) reduce 247) ((T_GOTO) reduce 247) ((T_HALT_COMPILER) reduce 247) ((T_IF) reduce 247) ((T_INC) reduce 247) ((T_INCLUDE) reduce 247) ((T_INCLUDE_ONCE) reduce 247) ((T_INLINE_HTML) reduce 247) ((T_INTERFACE) reduce 247) ((T_INT_CAST) reduce 247) ((T_ISSET) reduce 247) ((T_LINE) reduce 247) ((T_LIST) reduce 247) ((T_LNUMBER) reduce 247) ((T_MATCH) reduce 247) ((T_METHOD_C) reduce 247) ((T_NAME_FULLY_QUALIFIED) reduce 247) ((T_NAME_QUALIFIED) reduce 247) ((T_NAME_RELATIVE) reduce 247) ((T_NEW) reduce 247) ((T_NS_C) reduce 247) ((T_OBJECT_CAST) reduce 247) ((T_PRINT) reduce 247) ((T_PROPERTY_C) reduce 247) ((T_READONLY) reduce 247) ((T_REQUIRE) reduce 247) ((T_REQUIRE_ONCE) reduce 247) ((T_RETURN) reduce 247) ((T_START_HEREDOC) reduce 247) ((T_STATIC) reduce 247) ((T_STRING) reduce 247) ((T_STRING_CAST) reduce 247) ((T_SWITCH) reduce 247) ((T_THROW) reduce 247) ((T_TRAIT) reduce 247) ((T_TRAIT_C) reduce 247) ((T_TRY) reduce 247) ((T_UNSET) reduce 247) ((T_UNSET_CAST) reduce 247) ((T_VARIABLE) reduce 247) ((T_VOID_CAST) reduce 247) ((T_WHILE) reduce 247) ((T_YIELD) reduce 247) ((T_YIELD_FROM) reduce 247) (("[") reduce 247) (("`") reduce 247) (("{") reduce 247) (("}") reduce 247) (("~") reduce 247)) 18165 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CASE) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DEFAULT) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_ELSE) reduce 161) ((T_ELSEIF) reduce 161) ((T_EMPTY) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_VOID_CAST) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("}") reduce 161) (("~") reduce 161)) 18168 ((("!") reduce 250) (("\"") reduce 250) (("$") reduce 250) (("(") reduce 250) (("+") reduce 250) (("-") reduce 250) ((";") reduce 250) (("@") reduce 250) ((T_ABSTRACT) reduce 250) ((T_ARRAY) reduce 250) ((T_ARRAY_CAST) reduce 250) ((T_ATTRIBUTE) reduce 250) ((T_BOOL_CAST) reduce 250) ((T_BREAK) reduce 250) ((T_CASE) reduce 250) ((T_CLASS) reduce 250) ((T_CLASS_C) reduce 250) ((T_CLONE) reduce 250) ((T_CONSTANT_ENCAPSED_STRING) reduce 250) ((T_CONTINUE) reduce 250) ((T_DEC) reduce 250) ((T_DECLARE) reduce 250) ((T_DEFAULT) reduce 250) ((T_DIR) reduce 250) ((T_DNUMBER) reduce 250) ((T_DO) reduce 250) ((T_DOUBLE_CAST) reduce 250) ((T_ECHO) reduce 250) ((T_ELSE) reduce 250) ((T_ELSEIF) reduce 250) ((T_EMPTY) reduce 250) ((T_ENUM) reduce 250) ((T_EVAL) reduce 250) ((T_EXIT) reduce 250) ((T_FILE) reduce 250) ((T_FINAL) reduce 250) ((T_FN) reduce 250) ((T_FOR) reduce 250) ((T_FOREACH) reduce 250) ((T_FUNCTION) reduce 250) ((T_FUNC_C) reduce 250) ((T_GLOBAL) reduce 250) ((T_GOTO) reduce 250) ((T_HALT_COMPILER) reduce 250) ((T_IF) reduce 250) ((T_INC) reduce 250) ((T_INCLUDE) reduce 250) ((T_INCLUDE_ONCE) reduce 250) ((T_INLINE_HTML) reduce 250) ((T_INTERFACE) reduce 250) ((T_INT_CAST) reduce 250) ((T_ISSET) reduce 250) ((T_LINE) reduce 250) ((T_LIST) reduce 250) ((T_LNUMBER) reduce 250) ((T_MATCH) reduce 250) ((T_METHOD_C) reduce 250) ((T_NAME_FULLY_QUALIFIED) reduce 250) ((T_NAME_QUALIFIED) reduce 250) ((T_NAME_RELATIVE) reduce 250) ((T_NEW) reduce 250) ((T_NS_C) reduce 250) ((T_OBJECT_CAST) reduce 250) ((T_PRINT) reduce 250) ((T_PROPERTY_C) reduce 250) ((T_READONLY) reduce 250) ((T_REQUIRE) reduce 250) ((T_REQUIRE_ONCE) reduce 250) ((T_RETURN) reduce 250) ((T_START_HEREDOC) reduce 250) ((T_STATIC) reduce 250) ((T_STRING) reduce 250) ((T_STRING_CAST) reduce 250) ((T_SWITCH) reduce 250) ((T_THROW) reduce 250) ((T_TRAIT) reduce 250) ((T_TRAIT_C) reduce 250) ((T_TRY) reduce 250) ((T_UNSET) reduce 250) ((T_UNSET_CAST) reduce 250) ((T_VARIABLE) reduce 250) ((T_VOID_CAST) reduce 250) ((T_WHILE) reduce 250) ((T_YIELD) reduce 250) ((T_YIELD_FROM) reduce 250) (("[") reduce 250) (("`") reduce 250) (("{") reduce 250) (("}") reduce 250) (("~") reduce 250)) 18172 ((("!") reduce 251) (("\"") reduce 251) (("$") reduce 251) (("(") reduce 251) (("+") reduce 251) (("-") reduce 251) ((";") reduce 251) (("@") reduce 251) ((T_ABSTRACT) reduce 251) ((T_ARRAY) reduce 251) ((T_ARRAY_CAST) reduce 251) ((T_ATTRIBUTE) reduce 251) ((T_BOOL_CAST) reduce 251) ((T_BREAK) reduce 251) ((T_CASE) reduce 251) ((T_CLASS) reduce 251) ((T_CLASS_C) reduce 251) ((T_CLONE) reduce 251) ((T_CONSTANT_ENCAPSED_STRING) reduce 251) ((T_CONTINUE) reduce 251) ((T_DEC) reduce 251) ((T_DECLARE) reduce 251) ((T_DEFAULT) reduce 251) ((T_DIR) reduce 251) ((T_DNUMBER) reduce 251) ((T_DO) reduce 251) ((T_DOUBLE_CAST) reduce 251) ((T_ECHO) reduce 251) ((T_ELSE) reduce 251) ((T_ELSEIF) reduce 251) ((T_EMPTY) reduce 251) ((T_ENUM) reduce 251) ((T_EVAL) reduce 251) ((T_EXIT) reduce 251) ((T_FILE) reduce 251) ((T_FINAL) reduce 251) ((T_FN) reduce 251) ((T_FOR) reduce 251) ((T_FOREACH) reduce 251) ((T_FUNCTION) reduce 251) ((T_FUNC_C) reduce 251) ((T_GLOBAL) reduce 251) ((T_GOTO) reduce 251) ((T_HALT_COMPILER) reduce 251) ((T_IF) reduce 251) ((T_INC) reduce 251) ((T_INCLUDE) reduce 251) ((T_INCLUDE_ONCE) reduce 251) ((T_INLINE_HTML) reduce 251) ((T_INTERFACE) reduce 251) ((T_INT_CAST) reduce 251) ((T_ISSET) reduce 251) ((T_LINE) reduce 251) ((T_LIST) reduce 251) ((T_LNUMBER) reduce 251) ((T_MATCH) reduce 251) ((T_METHOD_C) reduce 251) ((T_NAME_FULLY_QUALIFIED) reduce 251) ((T_NAME_QUALIFIED) reduce 251) ((T_NAME_RELATIVE) reduce 251) ((T_NEW) reduce 251) ((T_NS_C) reduce 251) ((T_OBJECT_CAST) reduce 251) ((T_PRINT) reduce 251) ((T_PROPERTY_C) reduce 251) ((T_READONLY) reduce 251) ((T_REQUIRE) reduce 251) ((T_REQUIRE_ONCE) reduce 251) ((T_RETURN) reduce 251) ((T_START_HEREDOC) reduce 251) ((T_STATIC) reduce 251) ((T_STRING) reduce 251) ((T_STRING_CAST) reduce 251) ((T_SWITCH) reduce 251) ((T_THROW) reduce 251) ((T_TRAIT) reduce 251) ((T_TRAIT_C) reduce 251) ((T_TRY) reduce 251) ((T_UNSET) reduce 251) ((T_UNSET_CAST) reduce 251) ((T_VARIABLE) reduce 251) ((T_VOID_CAST) reduce 251) ((T_WHILE) reduce 251) ((T_YIELD) reduce 251) ((T_YIELD_FROM) reduce 251) (("[") reduce 251) (("`") reduce 251) (("{") reduce 251) (("}") reduce 251) (("~") reduce 251)) 18177 ((("!") reduce 242) (("\"") reduce 242) (("$") reduce 242) (("(") reduce 242) (("+") reduce 242) (("-") reduce 242) ((";") reduce 242) (("@") reduce 242) ((T_ABSTRACT) reduce 242) ((T_ARRAY) reduce 242) ((T_ARRAY_CAST) reduce 242) ((T_ATTRIBUTE) reduce 242) ((T_BOOL_CAST) reduce 242) ((T_BREAK) reduce 242) ((T_CASE) reduce 242) ((T_CLASS) reduce 242) ((T_CLASS_C) reduce 242) ((T_CLONE) reduce 242) ((T_CONSTANT_ENCAPSED_STRING) reduce 242) ((T_CONTINUE) reduce 242) ((T_DEC) reduce 242) ((T_DECLARE) reduce 242) ((T_DEFAULT) reduce 242) ((T_DIR) reduce 242) ((T_DNUMBER) reduce 242) ((T_DO) reduce 242) ((T_DOUBLE_CAST) reduce 242) ((T_ECHO) reduce 242) ((T_ELSE) reduce 242) ((T_ELSEIF) reduce 242) ((T_EMPTY) reduce 242) ((T_ENUM) reduce 242) ((T_EVAL) reduce 242) ((T_EXIT) reduce 242) ((T_FILE) reduce 242) ((T_FINAL) reduce 242) ((T_FN) reduce 242) ((T_FOR) reduce 242) ((T_FOREACH) reduce 242) ((T_FUNCTION) reduce 242) ((T_FUNC_C) reduce 242) ((T_GLOBAL) reduce 242) ((T_GOTO) reduce 242) ((T_HALT_COMPILER) reduce 242) ((T_IF) reduce 242) ((T_INC) reduce 242) ((T_INCLUDE) reduce 242) ((T_INCLUDE_ONCE) reduce 242) ((T_INLINE_HTML) reduce 242) ((T_INTERFACE) reduce 242) ((T_INT_CAST) reduce 242) ((T_ISSET) reduce 242) ((T_LINE) reduce 242) ((T_LIST) reduce 242) ((T_LNUMBER) reduce 242) ((T_MATCH) reduce 242) ((T_METHOD_C) reduce 242) ((T_NAME_FULLY_QUALIFIED) reduce 242) ((T_NAME_QUALIFIED) reduce 242) ((T_NAME_RELATIVE) reduce 242) ((T_NEW) reduce 242) ((T_NS_C) reduce 242) ((T_OBJECT_CAST) reduce 242) ((T_PRINT) reduce 242) ((T_PROPERTY_C) reduce 242) ((T_READONLY) reduce 242) ((T_REQUIRE) reduce 242) ((T_REQUIRE_ONCE) reduce 242) ((T_RETURN) reduce 242) ((T_START_HEREDOC) reduce 242) ((T_STATIC) reduce 242) ((T_STRING) reduce 242) ((T_STRING_CAST) reduce 242) ((T_SWITCH) reduce 242) ((T_THROW) reduce 242) ((T_TRAIT) reduce 242) ((T_TRAIT_C) reduce 242) ((T_TRY) reduce 242) ((T_UNSET) reduce 242) ((T_UNSET_CAST) reduce 242) ((T_VARIABLE) reduce 242) ((T_VOID_CAST) reduce 242) ((T_WHILE) reduce 242) ((T_YIELD) reduce 242) ((T_YIELD_FROM) reduce 242) (("[") reduce 242) (("`") reduce 242) (("{") reduce 242) (("}") reduce 242) (("~") reduce 242)) 18178 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CASE) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DEFAULT) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_ELSE) reduce 150) ((T_ELSEIF) reduce 150) ((T_EMPTY) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_VOID_CAST) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("}") reduce 150) (("~") reduce 150)) 18181 ((("!") reduce 243) (("\"") reduce 243) (("$") reduce 243) (("(") reduce 243) (("+") reduce 243) (("-") reduce 243) ((";") reduce 243) (("@") reduce 243) ((T_ABSTRACT) reduce 243) ((T_ARRAY) reduce 243) ((T_ARRAY_CAST) reduce 243) ((T_ATTRIBUTE) reduce 243) ((T_BOOL_CAST) reduce 243) ((T_BREAK) reduce 243) ((T_CASE) reduce 243) ((T_CLASS) reduce 243) ((T_CLASS_C) reduce 243) ((T_CLONE) reduce 243) ((T_CONSTANT_ENCAPSED_STRING) reduce 243) ((T_CONTINUE) reduce 243) ((T_DEC) reduce 243) ((T_DECLARE) reduce 243) ((T_DEFAULT) reduce 243) ((T_DIR) reduce 243) ((T_DNUMBER) reduce 243) ((T_DO) reduce 243) ((T_DOUBLE_CAST) reduce 243) ((T_ECHO) reduce 243) ((T_ELSE) reduce 243) ((T_ELSEIF) reduce 243) ((T_EMPTY) reduce 243) ((T_ENUM) reduce 243) ((T_EVAL) reduce 243) ((T_EXIT) reduce 243) ((T_FILE) reduce 243) ((T_FINAL) reduce 243) ((T_FN) reduce 243) ((T_FOR) reduce 243) ((T_FOREACH) reduce 243) ((T_FUNCTION) reduce 243) ((T_FUNC_C) reduce 243) ((T_GLOBAL) reduce 243) ((T_GOTO) reduce 243) ((T_HALT_COMPILER) reduce 243) ((T_IF) reduce 243) ((T_INC) reduce 243) ((T_INCLUDE) reduce 243) ((T_INCLUDE_ONCE) reduce 243) ((T_INLINE_HTML) reduce 243) ((T_INTERFACE) reduce 243) ((T_INT_CAST) reduce 243) ((T_ISSET) reduce 243) ((T_LINE) reduce 243) ((T_LIST) reduce 243) ((T_LNUMBER) reduce 243) ((T_MATCH) reduce 243) ((T_METHOD_C) reduce 243) ((T_NAME_FULLY_QUALIFIED) reduce 243) ((T_NAME_QUALIFIED) reduce 243) ((T_NAME_RELATIVE) reduce 243) ((T_NEW) reduce 243) ((T_NS_C) reduce 243) ((T_OBJECT_CAST) reduce 243) ((T_PRINT) reduce 243) ((T_PROPERTY_C) reduce 243) ((T_READONLY) reduce 243) ((T_REQUIRE) reduce 243) ((T_REQUIRE_ONCE) reduce 243) ((T_RETURN) reduce 243) ((T_START_HEREDOC) reduce 243) ((T_STATIC) reduce 243) ((T_STRING) reduce 243) ((T_STRING_CAST) reduce 243) ((T_SWITCH) reduce 243) ((T_THROW) reduce 243) ((T_TRAIT) reduce 243) ((T_TRAIT_C) reduce 243) ((T_TRY) reduce 243) ((T_UNSET) reduce 243) ((T_UNSET_CAST) reduce 243) ((T_VARIABLE) reduce 243) ((T_VOID_CAST) reduce 243) ((T_WHILE) reduce 243) ((T_YIELD) reduce 243) ((T_YIELD_FROM) reduce 243) (("[") reduce 243) (("`") reduce 243) (("{") reduce 243) (("}") reduce 243) (("~") reduce 243)) 18183 ((("!") reduce 170) (("\"") reduce 170) (("$") reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CASE) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DEFAULT) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_ELSE) reduce 170) ((T_ELSEIF) reduce 170) ((T_EMPTY) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_VARIABLE) reduce 170) ((T_VOID_CAST) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("}") reduce 170) (("~") reduce 170)) 18188 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CASE) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DEFAULT) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_ELSE) reduce 162) ((T_ELSEIF) reduce 162) ((T_EMPTY) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_VOID_CAST) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("}") reduce 162) (("~") reduce 162)) 18191 ((("!") reduce 171) (("\"") reduce 171) (("$") reduce 171) (("(") reduce 171) (("+") reduce 171) (("-") reduce 171) ((";") reduce 171) (("@") reduce 171) ((T_ABSTRACT) reduce 171) ((T_ARRAY) reduce 171) ((T_ARRAY_CAST) reduce 171) ((T_ATTRIBUTE) reduce 171) ((T_BOOL_CAST) reduce 171) ((T_BREAK) reduce 171) ((T_CASE) reduce 171) ((T_CATCH) reduce 171) ((T_CLASS) reduce 171) ((T_CLASS_C) reduce 171) ((T_CLONE) reduce 171) ((T_CONSTANT_ENCAPSED_STRING) reduce 171) ((T_CONTINUE) reduce 171) ((T_DEC) reduce 171) ((T_DECLARE) reduce 171) ((T_DEFAULT) reduce 171) ((T_DIR) reduce 171) ((T_DNUMBER) reduce 171) ((T_DO) reduce 171) ((T_DOUBLE_CAST) reduce 171) ((T_ECHO) reduce 171) ((T_ELSE) reduce 171) ((T_ELSEIF) reduce 171) ((T_EMPTY) reduce 171) ((T_ENUM) reduce 171) ((T_EVAL) reduce 171) ((T_EXIT) reduce 171) ((T_FILE) reduce 171) ((T_FINAL) reduce 171) ((T_FINALLY) reduce 171) ((T_FN) reduce 171) ((T_FOR) reduce 171) ((T_FOREACH) reduce 171) ((T_FUNCTION) reduce 171) ((T_FUNC_C) reduce 171) ((T_GLOBAL) reduce 171) ((T_GOTO) reduce 171) ((T_HALT_COMPILER) reduce 171) ((T_IF) reduce 171) ((T_INC) reduce 171) ((T_INCLUDE) reduce 171) ((T_INCLUDE_ONCE) reduce 171) ((T_INLINE_HTML) reduce 171) ((T_INTERFACE) reduce 171) ((T_INT_CAST) reduce 171) ((T_ISSET) reduce 171) ((T_LINE) reduce 171) ((T_LIST) reduce 171) ((T_LNUMBER) reduce 171) ((T_MATCH) reduce 171) ((T_METHOD_C) reduce 171) ((T_NAME_FULLY_QUALIFIED) reduce 171) ((T_NAME_QUALIFIED) reduce 171) ((T_NAME_RELATIVE) reduce 171) ((T_NEW) reduce 171) ((T_NS_C) reduce 171) ((T_OBJECT_CAST) reduce 171) ((T_PRINT) reduce 171) ((T_PROPERTY_C) reduce 171) ((T_READONLY) reduce 171) ((T_REQUIRE) reduce 171) ((T_REQUIRE_ONCE) reduce 171) ((T_RETURN) reduce 171) ((T_START_HEREDOC) reduce 171) ((T_STATIC) reduce 171) ((T_STRING) reduce 171) ((T_STRING_CAST) reduce 171) ((T_SWITCH) reduce 171) ((T_THROW) reduce 171) ((T_TRAIT) reduce 171) ((T_TRAIT_C) reduce 171) ((T_TRY) reduce 171) ((T_UNSET) reduce 171) ((T_UNSET_CAST) reduce 171) ((T_VARIABLE) reduce 171) ((T_VOID_CAST) reduce 171) ((T_WHILE) reduce 171) ((T_YIELD) reduce 171) ((T_YIELD_FROM) reduce 171) (("[") reduce 171) (("`") reduce 171) (("{") reduce 171) (("}") reduce 171) (("~") reduce 171)) 18192 ((("!") reduce 177) (("\"") reduce 177) (("$") reduce 177) (("(") reduce 177) (("+") reduce 177) (("-") reduce 177) ((";") reduce 177) (("@") reduce 177) ((T_ABSTRACT) reduce 177) ((T_ARRAY) reduce 177) ((T_ARRAY_CAST) reduce 177) ((T_ATTRIBUTE) reduce 177) ((T_BOOL_CAST) reduce 177) ((T_BREAK) reduce 177) ((T_CASE) reduce 177) ((T_CATCH) shift) ((T_CLASS) reduce 177) ((T_CLASS_C) reduce 177) ((T_CLONE) reduce 177) ((T_CONSTANT_ENCAPSED_STRING) reduce 177) ((T_CONTINUE) reduce 177) ((T_DEC) reduce 177) ((T_DECLARE) reduce 177) ((T_DEFAULT) reduce 177) ((T_DIR) reduce 177) ((T_DNUMBER) reduce 177) ((T_DO) reduce 177) ((T_DOUBLE_CAST) reduce 177) ((T_ECHO) reduce 177) ((T_ELSE) reduce 177) ((T_ELSEIF) reduce 177) ((T_EMPTY) reduce 177) ((T_ENUM) reduce 177) ((T_EVAL) reduce 177) ((T_EXIT) reduce 177) ((T_FILE) reduce 177) ((T_FINAL) reduce 177) ((T_FINALLY) shift) ((T_FN) reduce 177) ((T_FOR) reduce 177) ((T_FOREACH) reduce 177) ((T_FUNCTION) reduce 177) ((T_FUNC_C) reduce 177) ((T_GLOBAL) reduce 177) ((T_GOTO) reduce 177) ((T_HALT_COMPILER) reduce 177) ((T_IF) reduce 177) ((T_INC) reduce 177) ((T_INCLUDE) reduce 177) ((T_INCLUDE_ONCE) reduce 177) ((T_INLINE_HTML) reduce 177) ((T_INTERFACE) reduce 177) ((T_INT_CAST) reduce 177) ((T_ISSET) reduce 177) ((T_LINE) reduce 177) ((T_LIST) reduce 177) ((T_LNUMBER) reduce 177) ((T_MATCH) reduce 177) ((T_METHOD_C) reduce 177) ((T_NAME_FULLY_QUALIFIED) reduce 177) ((T_NAME_QUALIFIED) reduce 177) ((T_NAME_RELATIVE) reduce 177) ((T_NEW) reduce 177) ((T_NS_C) reduce 177) ((T_OBJECT_CAST) reduce 177) ((T_PRINT) reduce 177) ((T_PROPERTY_C) reduce 177) ((T_READONLY) reduce 177) ((T_REQUIRE) reduce 177) ((T_REQUIRE_ONCE) reduce 177) ((T_RETURN) reduce 177) ((T_START_HEREDOC) reduce 177) ((T_STATIC) reduce 177) ((T_STRING) reduce 177) ((T_STRING_CAST) reduce 177) ((T_SWITCH) reduce 177) ((T_THROW) reduce 177) ((T_TRAIT) reduce 177) ((T_TRAIT_C) reduce 177) ((T_TRY) reduce 177) ((T_UNSET) reduce 177) ((T_UNSET_CAST) reduce 177) ((T_VARIABLE) reduce 177) ((T_VOID_CAST) reduce 177) ((T_WHILE) reduce 177) ((T_YIELD) reduce 177) ((T_YIELD_FROM) reduce 177) (("[") reduce 177) (("`") reduce 177) (("{") reduce 177) (("}") reduce 177) (("~") reduce 177)) 18195 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CASE) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DEFAULT) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_ELSE) reduce 167) ((T_ELSEIF) reduce 167) ((T_EMPTY) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_VOID_CAST) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("}") reduce 167) (("~") reduce 167)) 18198 ((("!") reduce 178) (("\"") reduce 178) (("$") reduce 178) (("(") reduce 178) (("+") reduce 178) (("-") reduce 178) ((";") reduce 178) (("@") reduce 178) ((T_ABSTRACT) reduce 178) ((T_ARRAY) reduce 178) ((T_ARRAY_CAST) reduce 178) ((T_ATTRIBUTE) reduce 178) ((T_BOOL_CAST) reduce 178) ((T_BREAK) reduce 178) ((T_CASE) reduce 178) ((T_CLASS) reduce 178) ((T_CLASS_C) reduce 178) ((T_CLONE) reduce 178) ((T_CONSTANT_ENCAPSED_STRING) reduce 178) ((T_CONTINUE) reduce 178) ((T_DEC) reduce 178) ((T_DECLARE) reduce 178) ((T_DEFAULT) reduce 178) ((T_DIR) reduce 178) ((T_DNUMBER) reduce 178) ((T_DO) reduce 178) ((T_DOUBLE_CAST) reduce 178) ((T_ECHO) reduce 178) ((T_ELSE) reduce 178) ((T_ELSEIF) reduce 178) ((T_EMPTY) reduce 178) ((T_ENUM) reduce 178) ((T_EVAL) reduce 178) ((T_EXIT) reduce 178) ((T_FILE) reduce 178) ((T_FINAL) reduce 178) ((T_FN) reduce 178) ((T_FOR) reduce 178) ((T_FOREACH) reduce 178) ((T_FUNCTION) reduce 178) ((T_FUNC_C) reduce 178) ((T_GLOBAL) reduce 178) ((T_GOTO) reduce 178) ((T_HALT_COMPILER) reduce 178) ((T_IF) reduce 178) ((T_INC) reduce 178) ((T_INCLUDE) reduce 178) ((T_INCLUDE_ONCE) reduce 178) ((T_INLINE_HTML) reduce 178) ((T_INTERFACE) reduce 178) ((T_INT_CAST) reduce 178) ((T_ISSET) reduce 178) ((T_LINE) reduce 178) ((T_LIST) reduce 178) ((T_LNUMBER) reduce 178) ((T_MATCH) reduce 178) ((T_METHOD_C) reduce 178) ((T_NAME_FULLY_QUALIFIED) reduce 178) ((T_NAME_QUALIFIED) reduce 178) ((T_NAME_RELATIVE) reduce 178) ((T_NEW) reduce 178) ((T_NS_C) reduce 178) ((T_OBJECT_CAST) reduce 178) ((T_PRINT) reduce 178) ((T_PROPERTY_C) reduce 178) ((T_READONLY) reduce 178) ((T_REQUIRE) reduce 178) ((T_REQUIRE_ONCE) reduce 178) ((T_RETURN) reduce 178) ((T_START_HEREDOC) reduce 178) ((T_STATIC) reduce 178) ((T_STRING) reduce 178) ((T_STRING_CAST) reduce 178) ((T_SWITCH) reduce 178) ((T_THROW) reduce 178) ((T_TRAIT) reduce 178) ((T_TRAIT_C) reduce 178) ((T_TRY) reduce 178) ((T_UNSET) reduce 178) ((T_UNSET_CAST) reduce 178) ((T_VARIABLE) reduce 178) ((T_VOID_CAST) reduce 178) ((T_WHILE) reduce 178) ((T_YIELD) reduce 178) ((T_YIELD_FROM) reduce 178) (("[") reduce 178) (("`") reduce 178) (("{") reduce 178) (("}") reduce 178) (("~") reduce 178)) 18205 ((("!") reduce 172) (("\"") reduce 172) (("$") reduce 172) (("(") reduce 172) (("+") reduce 172) (("-") reduce 172) ((";") reduce 172) (("@") reduce 172) ((T_ABSTRACT) reduce 172) ((T_ARRAY) reduce 172) ((T_ARRAY_CAST) reduce 172) ((T_ATTRIBUTE) reduce 172) ((T_BOOL_CAST) reduce 172) ((T_BREAK) reduce 172) ((T_CASE) reduce 172) ((T_CATCH) reduce 172) ((T_CLASS) reduce 172) ((T_CLASS_C) reduce 172) ((T_CLONE) reduce 172) ((T_CONSTANT_ENCAPSED_STRING) reduce 172) ((T_CONTINUE) reduce 172) ((T_DEC) reduce 172) ((T_DECLARE) reduce 172) ((T_DEFAULT) reduce 172) ((T_DIR) reduce 172) ((T_DNUMBER) reduce 172) ((T_DO) reduce 172) ((T_DOUBLE_CAST) reduce 172) ((T_ECHO) reduce 172) ((T_ELSE) reduce 172) ((T_ELSEIF) reduce 172) ((T_EMPTY) reduce 172) ((T_ENUM) reduce 172) ((T_EVAL) reduce 172) ((T_EXIT) reduce 172) ((T_FILE) reduce 172) ((T_FINAL) reduce 172) ((T_FINALLY) reduce 172) ((T_FN) reduce 172) ((T_FOR) reduce 172) ((T_FOREACH) reduce 172) ((T_FUNCTION) reduce 172) ((T_FUNC_C) reduce 172) ((T_GLOBAL) reduce 172) ((T_GOTO) reduce 172) ((T_HALT_COMPILER) reduce 172) ((T_IF) reduce 172) ((T_INC) reduce 172) ((T_INCLUDE) reduce 172) ((T_INCLUDE_ONCE) reduce 172) ((T_INLINE_HTML) reduce 172) ((T_INTERFACE) reduce 172) ((T_INT_CAST) reduce 172) ((T_ISSET) reduce 172) ((T_LINE) reduce 172) ((T_LIST) reduce 172) ((T_LNUMBER) reduce 172) ((T_MATCH) reduce 172) ((T_METHOD_C) reduce 172) ((T_NAME_FULLY_QUALIFIED) reduce 172) ((T_NAME_QUALIFIED) reduce 172) ((T_NAME_RELATIVE) reduce 172) ((T_NEW) reduce 172) ((T_NS_C) reduce 172) ((T_OBJECT_CAST) reduce 172) ((T_PRINT) reduce 172) ((T_PROPERTY_C) reduce 172) ((T_READONLY) reduce 172) ((T_REQUIRE) reduce 172) ((T_REQUIRE_ONCE) reduce 172) ((T_RETURN) reduce 172) ((T_START_HEREDOC) reduce 172) ((T_STATIC) reduce 172) ((T_STRING) reduce 172) ((T_STRING_CAST) reduce 172) ((T_SWITCH) reduce 172) ((T_THROW) reduce 172) ((T_TRAIT) reduce 172) ((T_TRAIT_C) reduce 172) ((T_TRY) reduce 172) ((T_UNSET) reduce 172) ((T_UNSET_CAST) reduce 172) ((T_VARIABLE) reduce 172) ((T_VOID_CAST) reduce 172) ((T_WHILE) reduce 172) ((T_YIELD) reduce 172) ((T_YIELD_FROM) reduce 172) (("[") reduce 172) (("`") reduce 172) (("{") reduce 172) (("}") reduce 172) (("~") reduce 172)) 18210 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CASE) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DEFAULT) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_ELSE) reduce 153) ((T_ELSEIF) reduce 153) ((T_EMPTY) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_VOID_CAST) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("}") reduce 153) (("~") reduce 153)) 18214 ((("!") reduce 224) (("\"") reduce 224) (("$") reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CASE) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DEFAULT) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_ELSE) reduce 224) ((T_ELSEIF) reduce 224) ((T_EMPTY) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_VARIABLE) reduce 224) ((T_VOID_CAST) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("}") reduce 224) (("~") reduce 224)) 18216 ((("!") reduce 225) (("\"") reduce 225) (("$") reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CASE) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DEFAULT) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_ELSE) reduce 225) ((T_ELSEIF) reduce 225) ((T_EMPTY) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_VARIABLE) reduce 225) ((T_VOID_CAST) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("}") reduce 225) (("~") reduce 225)) 18217 (((T_CASE) reduce 228) ((T_DEFAULT) reduce 228) ((T_ENDSWITCH) reduce 228)) 18218 (((T_CASE) shift) ((T_DEFAULT) shift) ((T_ENDSWITCH) shift)) 18222 ((("!") reduce 226) (("\"") reduce 226) (("$") reduce 226) (("(") reduce 226) (("+") reduce 226) (("-") reduce 226) ((";") reduce 226) (("@") reduce 226) ((T_ABSTRACT) reduce 226) ((T_ARRAY) reduce 226) ((T_ARRAY_CAST) reduce 226) ((T_ATTRIBUTE) reduce 226) ((T_BOOL_CAST) reduce 226) ((T_BREAK) reduce 226) ((T_CASE) reduce 226) ((T_CLASS) reduce 226) ((T_CLASS_C) reduce 226) ((T_CLONE) reduce 226) ((T_CONSTANT_ENCAPSED_STRING) reduce 226) ((T_CONTINUE) reduce 226) ((T_DEC) reduce 226) ((T_DECLARE) reduce 226) ((T_DEFAULT) reduce 226) ((T_DIR) reduce 226) ((T_DNUMBER) reduce 226) ((T_DO) reduce 226) ((T_DOUBLE_CAST) reduce 226) ((T_ECHO) reduce 226) ((T_ELSE) reduce 226) ((T_ELSEIF) reduce 226) ((T_EMPTY) reduce 226) ((T_ENUM) reduce 226) ((T_EVAL) reduce 226) ((T_EXIT) reduce 226) ((T_FILE) reduce 226) ((T_FINAL) reduce 226) ((T_FN) reduce 226) ((T_FOR) reduce 226) ((T_FOREACH) reduce 226) ((T_FUNCTION) reduce 226) ((T_FUNC_C) reduce 226) ((T_GLOBAL) reduce 226) ((T_GOTO) reduce 226) ((T_HALT_COMPILER) reduce 226) ((T_IF) reduce 226) ((T_INC) reduce 226) ((T_INCLUDE) reduce 226) ((T_INCLUDE_ONCE) reduce 226) ((T_INLINE_HTML) reduce 226) ((T_INTERFACE) reduce 226) ((T_INT_CAST) reduce 226) ((T_ISSET) reduce 226) ((T_LINE) reduce 226) ((T_LIST) reduce 226) ((T_LNUMBER) reduce 226) ((T_MATCH) reduce 226) ((T_METHOD_C) reduce 226) ((T_NAME_FULLY_QUALIFIED) reduce 226) ((T_NAME_QUALIFIED) reduce 226) ((T_NAME_RELATIVE) reduce 226) ((T_NEW) reduce 226) ((T_NS_C) reduce 226) ((T_OBJECT_CAST) reduce 226) ((T_PRINT) reduce 226) ((T_PROPERTY_C) reduce 226) ((T_READONLY) reduce 226) ((T_REQUIRE) reduce 226) ((T_REQUIRE_ONCE) reduce 226) ((T_RETURN) reduce 226) ((T_START_HEREDOC) reduce 226) ((T_STATIC) reduce 226) ((T_STRING) reduce 226) ((T_STRING_CAST) reduce 226) ((T_SWITCH) reduce 226) ((T_THROW) reduce 226) ((T_TRAIT) reduce 226) ((T_TRAIT_C) reduce 226) ((T_TRY) reduce 226) ((T_UNSET) reduce 226) ((T_UNSET_CAST) reduce 226) ((T_VARIABLE) reduce 226) ((T_VOID_CAST) reduce 226) ((T_WHILE) reduce 226) ((T_YIELD) reduce 226) ((T_YIELD_FROM) reduce 226) (("[") reduce 226) (("`") reduce 226) (("{") reduce 226) (("}") reduce 226) (("~") reduce 226)) 18223 ((("!") reduce 142) (("\"") reduce 142) (("$") reduce 142) (("(") reduce 142) (("+") reduce 142) (("-") reduce 142) ((";") reduce 142) (("@") reduce 142) ((T_ABSTRACT) reduce 142) ((T_ARRAY) reduce 142) ((T_ARRAY_CAST) reduce 142) ((T_ATTRIBUTE) reduce 142) ((T_BOOL_CAST) reduce 142) ((T_BREAK) reduce 142) ((T_CASE) reduce 142) ((T_CLASS) reduce 142) ((T_CLASS_C) reduce 142) ((T_CLONE) reduce 142) ((T_CONSTANT_ENCAPSED_STRING) reduce 142) ((T_CONTINUE) reduce 142) ((T_DEC) reduce 142) ((T_DECLARE) reduce 142) ((T_DEFAULT) reduce 142) ((T_DIR) reduce 142) ((T_DNUMBER) reduce 142) ((T_DO) reduce 142) ((T_DOUBLE_CAST) reduce 142) ((T_ECHO) reduce 142) ((T_EMPTY) reduce 142) ((T_ENDSWITCH) reduce 142) ((T_ENUM) reduce 142) ((T_EVAL) reduce 142) ((T_EXIT) reduce 142) ((T_FILE) reduce 142) ((T_FINAL) reduce 142) ((T_FN) reduce 142) ((T_FOR) reduce 142) ((T_FOREACH) reduce 142) ((T_FUNCTION) reduce 142) ((T_FUNC_C) reduce 142) ((T_GLOBAL) reduce 142) ((T_GOTO) reduce 142) ((T_HALT_COMPILER) reduce 142) ((T_IF) reduce 142) ((T_INC) reduce 142) ((T_INCLUDE) reduce 142) ((T_INCLUDE_ONCE) reduce 142) ((T_INLINE_HTML) reduce 142) ((T_INTERFACE) reduce 142) ((T_INT_CAST) reduce 142) ((T_ISSET) reduce 142) ((T_LINE) reduce 142) ((T_LIST) reduce 142) ((T_LNUMBER) reduce 142) ((T_MATCH) reduce 142) ((T_METHOD_C) reduce 142) ((T_NAME_FULLY_QUALIFIED) reduce 142) ((T_NAME_QUALIFIED) reduce 142) ((T_NAME_RELATIVE) reduce 142) ((T_NEW) reduce 142) ((T_NS_C) reduce 142) ((T_OBJECT_CAST) reduce 142) ((T_PRINT) reduce 142) ((T_PROPERTY_C) reduce 142) ((T_READONLY) reduce 142) ((T_REQUIRE) reduce 142) ((T_REQUIRE_ONCE) reduce 142) ((T_RETURN) reduce 142) ((T_START_HEREDOC) reduce 142) ((T_STATIC) reduce 142) ((T_STRING) reduce 142) ((T_STRING_CAST) reduce 142) ((T_SWITCH) reduce 142) ((T_THROW) reduce 142) ((T_TRAIT) reduce 142) ((T_TRAIT_C) reduce 142) ((T_TRY) reduce 142) ((T_UNSET) reduce 142) ((T_UNSET_CAST) reduce 142) ((T_VARIABLE) reduce 142) ((T_VOID_CAST) reduce 142) ((T_WHILE) reduce 142) ((T_YIELD) reduce 142) ((T_YIELD_FROM) reduce 142) (("[") reduce 142) (("`") reduce 142) (("{") reduce 142) (("~") reduce 142)) 18225 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CASE) reduce 232) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DEFAULT) reduce 232) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_ENDSWITCH) reduce 232) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_VOID_CAST) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("~") shift)) 18226 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CASE) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DEFAULT) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_EMPTY) reduce 166) ((T_ENDSWITCH) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_VOID_CAST) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 18240 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CASE) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DEFAULT) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_EMPTY) reduce 160) ((T_ENDSWITCH) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_VOID_CAST) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 18251 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CASE) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DEFAULT) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_EMPTY) reduce 149) ((T_ENDSWITCH) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_VOID_CAST) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 18253 ((("!") reduce 144) (("\"") reduce 144) (("$") reduce 144) (("(") reduce 144) (("+") reduce 144) (("-") reduce 144) ((";") reduce 144) (("@") reduce 144) ((T_ABSTRACT) reduce 144) ((T_ARRAY) reduce 144) ((T_ARRAY_CAST) reduce 144) ((T_ATTRIBUTE) reduce 144) ((T_BOOL_CAST) reduce 144) ((T_BREAK) reduce 144) ((T_CASE) reduce 144) ((T_CLASS) reduce 144) ((T_CLASS_C) reduce 144) ((T_CLONE) reduce 144) ((T_CONSTANT_ENCAPSED_STRING) reduce 144) ((T_CONTINUE) reduce 144) ((T_DEC) reduce 144) ((T_DECLARE) reduce 144) ((T_DEFAULT) reduce 144) ((T_DIR) reduce 144) ((T_DNUMBER) reduce 144) ((T_DO) reduce 144) ((T_DOUBLE_CAST) reduce 144) ((T_ECHO) reduce 144) ((T_EMPTY) reduce 144) ((T_ENDSWITCH) reduce 144) ((T_ENUM) reduce 144) ((T_EVAL) reduce 144) ((T_EXIT) reduce 144) ((T_FILE) reduce 144) ((T_FINAL) reduce 144) ((T_FN) reduce 144) ((T_FOR) reduce 144) ((T_FOREACH) reduce 144) ((T_FUNCTION) reduce 144) ((T_FUNC_C) reduce 144) ((T_GLOBAL) reduce 144) ((T_GOTO) reduce 144) ((T_HALT_COMPILER) reduce 144) ((T_IF) reduce 144) ((T_INC) reduce 144) ((T_INCLUDE) reduce 144) ((T_INCLUDE_ONCE) reduce 144) ((T_INLINE_HTML) reduce 144) ((T_INTERFACE) reduce 144) ((T_INT_CAST) reduce 144) ((T_ISSET) reduce 144) ((T_LINE) reduce 144) ((T_LIST) reduce 144) ((T_LNUMBER) reduce 144) ((T_MATCH) reduce 144) ((T_METHOD_C) reduce 144) ((T_NAME_FULLY_QUALIFIED) reduce 144) ((T_NAME_QUALIFIED) reduce 144) ((T_NAME_RELATIVE) reduce 144) ((T_NEW) reduce 144) ((T_NS_C) reduce 144) ((T_OBJECT_CAST) reduce 144) ((T_PRINT) reduce 144) ((T_PROPERTY_C) reduce 144) ((T_READONLY) reduce 144) ((T_REQUIRE) reduce 144) ((T_REQUIRE_ONCE) reduce 144) ((T_RETURN) reduce 144) ((T_START_HEREDOC) reduce 144) ((T_STATIC) reduce 144) ((T_STRING) reduce 144) ((T_STRING_CAST) reduce 144) ((T_SWITCH) reduce 144) ((T_THROW) reduce 144) ((T_TRAIT) reduce 144) ((T_TRAIT_C) reduce 144) ((T_TRY) reduce 144) ((T_UNSET) reduce 144) ((T_UNSET_CAST) reduce 144) ((T_VARIABLE) reduce 144) ((T_VOID_CAST) reduce 144) ((T_WHILE) reduce 144) ((T_YIELD) reduce 144) ((T_YIELD_FROM) reduce 144) (("[") reduce 144) (("`") reduce 144) (("{") reduce 144) (("~") reduce 144)) 18255 ((("!") reduce 104) (("\"") reduce 104) (("$") reduce 104) (("(") reduce 104) (("+") reduce 104) (("-") reduce 104) ((";") reduce 104) (("@") reduce 104) ((T_ABSTRACT) reduce 104) ((T_ARRAY) reduce 104) ((T_ARRAY_CAST) reduce 104) ((T_ATTRIBUTE) reduce 104) ((T_BOOL_CAST) reduce 104) ((T_BREAK) reduce 104) ((T_CASE) reduce 104) ((T_CLASS) reduce 104) ((T_CLASS_C) reduce 104) ((T_CLONE) reduce 104) ((T_CONSTANT_ENCAPSED_STRING) reduce 104) ((T_CONTINUE) reduce 104) ((T_DEC) reduce 104) ((T_DECLARE) reduce 104) ((T_DEFAULT) reduce 104) ((T_DIR) reduce 104) ((T_DNUMBER) reduce 104) ((T_DO) reduce 104) ((T_DOUBLE_CAST) reduce 104) ((T_ECHO) reduce 104) ((T_EMPTY) reduce 104) ((T_ENDSWITCH) reduce 104) ((T_ENUM) reduce 104) ((T_EVAL) reduce 104) ((T_EXIT) reduce 104) ((T_FILE) reduce 104) ((T_FINAL) reduce 104) ((T_FN) reduce 104) ((T_FOR) reduce 104) ((T_FOREACH) reduce 104) ((T_FUNCTION) reduce 104) ((T_FUNC_C) reduce 104) ((T_GLOBAL) reduce 104) ((T_GOTO) reduce 104) ((T_HALT_COMPILER) reduce 104) ((T_IF) reduce 104) ((T_INC) reduce 104) ((T_INCLUDE) reduce 104) ((T_INCLUDE_ONCE) reduce 104) ((T_INLINE_HTML) reduce 104) ((T_INTERFACE) reduce 104) ((T_INT_CAST) reduce 104) ((T_ISSET) reduce 104) ((T_LINE) reduce 104) ((T_LIST) reduce 104) ((T_LNUMBER) reduce 104) ((T_MATCH) reduce 104) ((T_METHOD_C) reduce 104) ((T_NAME_FULLY_QUALIFIED) reduce 104) ((T_NAME_QUALIFIED) reduce 104) ((T_NAME_RELATIVE) reduce 104) ((T_NEW) reduce 104) ((T_NS_C) reduce 104) ((T_OBJECT_CAST) reduce 104) ((T_PRINT) reduce 104) ((T_PROPERTY_C) reduce 104) ((T_READONLY) reduce 104) ((T_REQUIRE) reduce 104) ((T_REQUIRE_ONCE) reduce 104) ((T_RETURN) reduce 104) ((T_START_HEREDOC) reduce 104) ((T_STATIC) reduce 104) ((T_STRING) reduce 104) ((T_STRING_CAST) reduce 104) ((T_SWITCH) reduce 104) ((T_THROW) reduce 104) ((T_TRAIT) reduce 104) ((T_TRAIT_C) reduce 104) ((T_TRY) reduce 104) ((T_UNSET) reduce 104) ((T_UNSET_CAST) reduce 104) ((T_VARIABLE) reduce 104) ((T_VOID_CAST) reduce 104) ((T_WHILE) reduce 104) ((T_YIELD) reduce 104) ((T_YIELD_FROM) reduce 104) (("[") reduce 104) (("`") reduce 104) (("{") reduce 104) (("~") reduce 104)) 18257 ((("!") reduce 107) (("\"") reduce 107) (("$") reduce 107) (("(") reduce 107) (("+") reduce 107) (("-") reduce 107) ((";") reduce 107) (("@") reduce 107) ((T_ABSTRACT) reduce 107) ((T_ARRAY) reduce 107) ((T_ARRAY_CAST) reduce 107) ((T_ATTRIBUTE) reduce 107) ((T_BOOL_CAST) reduce 107) ((T_BREAK) reduce 107) ((T_CASE) reduce 107) ((T_CLASS) reduce 107) ((T_CLASS_C) reduce 107) ((T_CLONE) reduce 107) ((T_CONSTANT_ENCAPSED_STRING) reduce 107) ((T_CONTINUE) reduce 107) ((T_DEC) reduce 107) ((T_DECLARE) reduce 107) ((T_DEFAULT) reduce 107) ((T_DIR) reduce 107) ((T_DNUMBER) reduce 107) ((T_DO) reduce 107) ((T_DOUBLE_CAST) reduce 107) ((T_ECHO) reduce 107) ((T_EMPTY) reduce 107) ((T_ENDSWITCH) reduce 107) ((T_ENUM) reduce 107) ((T_EVAL) reduce 107) ((T_EXIT) reduce 107) ((T_FILE) reduce 107) ((T_FINAL) reduce 107) ((T_FN) reduce 107) ((T_FOR) reduce 107) ((T_FOREACH) reduce 107) ((T_FUNCTION) reduce 107) ((T_FUNC_C) reduce 107) ((T_GLOBAL) reduce 107) ((T_GOTO) reduce 107) ((T_HALT_COMPILER) reduce 107) ((T_IF) reduce 107) ((T_INC) reduce 107) ((T_INCLUDE) reduce 107) ((T_INCLUDE_ONCE) reduce 107) ((T_INLINE_HTML) reduce 107) ((T_INTERFACE) reduce 107) ((T_INT_CAST) reduce 107) ((T_ISSET) reduce 107) ((T_LINE) reduce 107) ((T_LIST) reduce 107) ((T_LNUMBER) reduce 107) ((T_MATCH) reduce 107) ((T_METHOD_C) reduce 107) ((T_NAME_FULLY_QUALIFIED) reduce 107) ((T_NAME_QUALIFIED) reduce 107) ((T_NAME_RELATIVE) reduce 107) ((T_NEW) reduce 107) ((T_NS_C) reduce 107) ((T_OBJECT_CAST) reduce 107) ((T_PRINT) reduce 107) ((T_PROPERTY_C) reduce 107) ((T_READONLY) reduce 107) ((T_REQUIRE) reduce 107) ((T_REQUIRE_ONCE) reduce 107) ((T_RETURN) reduce 107) ((T_START_HEREDOC) reduce 107) ((T_STATIC) reduce 107) ((T_STRING) reduce 107) ((T_STRING_CAST) reduce 107) ((T_SWITCH) reduce 107) ((T_THROW) reduce 107) ((T_TRAIT) reduce 107) ((T_TRAIT_C) reduce 107) ((T_TRY) reduce 107) ((T_UNSET) reduce 107) ((T_UNSET_CAST) reduce 107) ((T_VARIABLE) reduce 107) ((T_VOID_CAST) reduce 107) ((T_WHILE) reduce 107) ((T_YIELD) reduce 107) ((T_YIELD_FROM) reduce 107) (("[") reduce 107) (("`") reduce 107) (("{") reduce 107) (("~") reduce 107)) 18260 ((("!") reduce 103) (("\"") reduce 103) (("$") reduce 103) (("(") reduce 103) (("+") reduce 103) (("-") reduce 103) ((";") reduce 103) (("@") reduce 103) ((T_ABSTRACT) reduce 103) ((T_ARRAY) reduce 103) ((T_ARRAY_CAST) reduce 103) ((T_ATTRIBUTE) reduce 103) ((T_BOOL_CAST) reduce 103) ((T_BREAK) reduce 103) ((T_CASE) reduce 103) ((T_CLASS) reduce 103) ((T_CLASS_C) reduce 103) ((T_CLONE) reduce 103) ((T_CONSTANT_ENCAPSED_STRING) reduce 103) ((T_CONTINUE) reduce 103) ((T_DEC) reduce 103) ((T_DECLARE) reduce 103) ((T_DEFAULT) reduce 103) ((T_DIR) reduce 103) ((T_DNUMBER) reduce 103) ((T_DO) reduce 103) ((T_DOUBLE_CAST) reduce 103) ((T_ECHO) reduce 103) ((T_EMPTY) reduce 103) ((T_ENDSWITCH) reduce 103) ((T_ENUM) reduce 103) ((T_EVAL) reduce 103) ((T_EXIT) reduce 103) ((T_FILE) reduce 103) ((T_FINAL) reduce 103) ((T_FN) reduce 103) ((T_FOR) reduce 103) ((T_FOREACH) reduce 103) ((T_FUNCTION) reduce 103) ((T_FUNC_C) reduce 103) ((T_GLOBAL) reduce 103) ((T_GOTO) reduce 103) ((T_HALT_COMPILER) reduce 103) ((T_IF) reduce 103) ((T_INC) reduce 103) ((T_INCLUDE) reduce 103) ((T_INCLUDE_ONCE) reduce 103) ((T_INLINE_HTML) reduce 103) ((T_INTERFACE) reduce 103) ((T_INT_CAST) reduce 103) ((T_ISSET) reduce 103) ((T_LINE) reduce 103) ((T_LIST) reduce 103) ((T_LNUMBER) reduce 103) ((T_MATCH) reduce 103) ((T_METHOD_C) reduce 103) ((T_NAME_FULLY_QUALIFIED) reduce 103) ((T_NAME_QUALIFIED) reduce 103) ((T_NAME_RELATIVE) reduce 103) ((T_NEW) reduce 103) ((T_NS_C) reduce 103) ((T_OBJECT_CAST) reduce 103) ((T_PRINT) reduce 103) ((T_PROPERTY_C) reduce 103) ((T_READONLY) reduce 103) ((T_REQUIRE) reduce 103) ((T_REQUIRE_ONCE) reduce 103) ((T_RETURN) reduce 103) ((T_START_HEREDOC) reduce 103) ((T_STATIC) reduce 103) ((T_STRING) reduce 103) ((T_STRING_CAST) reduce 103) ((T_SWITCH) reduce 103) ((T_THROW) reduce 103) ((T_TRAIT) reduce 103) ((T_TRAIT_C) reduce 103) ((T_TRY) reduce 103) ((T_UNSET) reduce 103) ((T_UNSET_CAST) reduce 103) ((T_VARIABLE) reduce 103) ((T_VOID_CAST) reduce 103) ((T_WHILE) reduce 103) ((T_YIELD) reduce 103) ((T_YIELD_FROM) reduce 103) (("[") reduce 103) (("`") reduce 103) (("{") reduce 103) (("~") reduce 103)) 18261 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CASE) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DEFAULT) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_EMPTY) reduce 148) ((T_ENDSWITCH) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_VOID_CAST) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 18262 ((("!") reduce 246) (("\"") reduce 246) (("$") reduce 246) (("(") reduce 246) (("+") reduce 246) (("-") reduce 246) ((";") reduce 246) (("@") reduce 246) ((T_ABSTRACT) reduce 246) ((T_ARRAY) reduce 246) ((T_ARRAY_CAST) reduce 246) ((T_ATTRIBUTE) reduce 246) ((T_BOOL_CAST) reduce 246) ((T_BREAK) reduce 246) ((T_CASE) reduce 246) ((T_CLASS) reduce 246) ((T_CLASS_C) reduce 246) ((T_CLONE) reduce 246) ((T_CONSTANT_ENCAPSED_STRING) reduce 246) ((T_CONTINUE) reduce 246) ((T_DEC) reduce 246) ((T_DECLARE) reduce 246) ((T_DEFAULT) reduce 246) ((T_DIR) reduce 246) ((T_DNUMBER) reduce 246) ((T_DO) reduce 246) ((T_DOUBLE_CAST) reduce 246) ((T_ECHO) reduce 246) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) reduce 246) ((T_ENDSWITCH) reduce 246) ((T_ENUM) reduce 246) ((T_EVAL) reduce 246) ((T_EXIT) reduce 246) ((T_FILE) reduce 246) ((T_FINAL) reduce 246) ((T_FN) reduce 246) ((T_FOR) reduce 246) ((T_FOREACH) reduce 246) ((T_FUNCTION) reduce 246) ((T_FUNC_C) reduce 246) ((T_GLOBAL) reduce 246) ((T_GOTO) reduce 246) ((T_HALT_COMPILER) reduce 246) ((T_IF) reduce 246) ((T_INC) reduce 246) ((T_INCLUDE) reduce 246) ((T_INCLUDE_ONCE) reduce 246) ((T_INLINE_HTML) reduce 246) ((T_INTERFACE) reduce 246) ((T_INT_CAST) reduce 246) ((T_ISSET) reduce 246) ((T_LINE) reduce 246) ((T_LIST) reduce 246) ((T_LNUMBER) reduce 246) ((T_MATCH) reduce 246) ((T_METHOD_C) reduce 246) ((T_NAME_FULLY_QUALIFIED) reduce 246) ((T_NAME_QUALIFIED) reduce 246) ((T_NAME_RELATIVE) reduce 246) ((T_NEW) reduce 246) ((T_NS_C) reduce 246) ((T_OBJECT_CAST) reduce 246) ((T_PRINT) reduce 246) ((T_PROPERTY_C) reduce 246) ((T_READONLY) reduce 246) ((T_REQUIRE) reduce 246) ((T_REQUIRE_ONCE) reduce 246) ((T_RETURN) reduce 246) ((T_START_HEREDOC) reduce 246) ((T_STATIC) reduce 246) ((T_STRING) reduce 246) ((T_STRING_CAST) reduce 246) ((T_SWITCH) reduce 246) ((T_THROW) reduce 246) ((T_TRAIT) reduce 246) ((T_TRAIT_C) reduce 246) ((T_TRY) reduce 246) ((T_UNSET) reduce 246) ((T_UNSET_CAST) reduce 246) ((T_VARIABLE) reduce 246) ((T_VOID_CAST) reduce 246) ((T_WHILE) reduce 246) ((T_YIELD) reduce 246) ((T_YIELD_FROM) reduce 246) (("[") reduce 246) (("`") reduce 246) (("{") reduce 246) (("~") reduce 246)) 18263 ((("!") reduce 141) (("\"") reduce 141) (("$") reduce 141) (("(") reduce 141) (("+") reduce 141) (("-") reduce 141) ((";") reduce 141) (("@") reduce 141) ((T_ABSTRACT) reduce 141) ((T_ARRAY) reduce 141) ((T_ARRAY_CAST) reduce 141) ((T_ATTRIBUTE) reduce 141) ((T_BOOL_CAST) reduce 141) ((T_BREAK) reduce 141) ((T_CASE) reduce 141) ((T_CLASS) reduce 141) ((T_CLASS_C) reduce 141) ((T_CLONE) reduce 141) ((T_CONSTANT_ENCAPSED_STRING) reduce 141) ((T_CONTINUE) reduce 141) ((T_DEC) reduce 141) ((T_DECLARE) reduce 141) ((T_DEFAULT) reduce 141) ((T_DIR) reduce 141) ((T_DNUMBER) reduce 141) ((T_DO) reduce 141) ((T_DOUBLE_CAST) reduce 141) ((T_ECHO) reduce 141) ((T_EMPTY) reduce 141) ((T_ENDSWITCH) reduce 141) ((T_ENUM) reduce 141) ((T_EVAL) reduce 141) ((T_EXIT) reduce 141) ((T_FILE) reduce 141) ((T_FINAL) reduce 141) ((T_FN) reduce 141) ((T_FOR) reduce 141) ((T_FOREACH) reduce 141) ((T_FUNCTION) reduce 141) ((T_FUNC_C) reduce 141) ((T_GLOBAL) reduce 141) ((T_GOTO) reduce 141) ((T_HALT_COMPILER) reduce 141) ((T_IF) reduce 141) ((T_INC) reduce 141) ((T_INCLUDE) reduce 141) ((T_INCLUDE_ONCE) reduce 141) ((T_INLINE_HTML) reduce 141) ((T_INTERFACE) reduce 141) ((T_INT_CAST) reduce 141) ((T_ISSET) reduce 141) ((T_LINE) reduce 141) ((T_LIST) reduce 141) ((T_LNUMBER) reduce 141) ((T_MATCH) reduce 141) ((T_METHOD_C) reduce 141) ((T_NAME_FULLY_QUALIFIED) reduce 141) ((T_NAME_QUALIFIED) reduce 141) ((T_NAME_RELATIVE) reduce 141) ((T_NEW) reduce 141) ((T_NS_C) reduce 141) ((T_OBJECT_CAST) reduce 141) ((T_PRINT) reduce 141) ((T_PROPERTY_C) reduce 141) ((T_READONLY) reduce 141) ((T_REQUIRE) reduce 141) ((T_REQUIRE_ONCE) reduce 141) ((T_RETURN) reduce 141) ((T_START_HEREDOC) reduce 141) ((T_STATIC) reduce 141) ((T_STRING) reduce 141) ((T_STRING_CAST) reduce 141) ((T_SWITCH) reduce 141) ((T_THROW) reduce 141) ((T_TRAIT) reduce 141) ((T_TRAIT_C) reduce 141) ((T_TRY) reduce 141) ((T_UNSET) reduce 141) ((T_UNSET_CAST) reduce 141) ((T_VARIABLE) reduce 141) ((T_VOID_CAST) reduce 141) ((T_WHILE) reduce 141) ((T_YIELD) reduce 141) ((T_YIELD_FROM) reduce 141) (("[") reduce 141) (("`") reduce 141) (("{") reduce 141) (("~") reduce 141)) 18264 ((("!") reduce 106) (("\"") reduce 106) (("$") reduce 106) (("(") reduce 106) (("+") reduce 106) (("-") reduce 106) ((";") reduce 106) (("@") reduce 106) ((T_ABSTRACT) reduce 106) ((T_ARRAY) reduce 106) ((T_ARRAY_CAST) reduce 106) ((T_ATTRIBUTE) reduce 106) ((T_BOOL_CAST) reduce 106) ((T_BREAK) reduce 106) ((T_CASE) reduce 106) ((T_CLASS) reduce 106) ((T_CLASS_C) reduce 106) ((T_CLONE) reduce 106) ((T_CONSTANT_ENCAPSED_STRING) reduce 106) ((T_CONTINUE) reduce 106) ((T_DEC) reduce 106) ((T_DECLARE) reduce 106) ((T_DEFAULT) reduce 106) ((T_DIR) reduce 106) ((T_DNUMBER) reduce 106) ((T_DO) reduce 106) ((T_DOUBLE_CAST) reduce 106) ((T_ECHO) reduce 106) ((T_EMPTY) reduce 106) ((T_ENDSWITCH) reduce 106) ((T_ENUM) reduce 106) ((T_EVAL) reduce 106) ((T_EXIT) reduce 106) ((T_FILE) reduce 106) ((T_FINAL) reduce 106) ((T_FN) reduce 106) ((T_FOR) reduce 106) ((T_FOREACH) reduce 106) ((T_FUNCTION) reduce 106) ((T_FUNC_C) reduce 106) ((T_GLOBAL) reduce 106) ((T_GOTO) reduce 106) ((T_HALT_COMPILER) reduce 106) ((T_IF) reduce 106) ((T_INC) reduce 106) ((T_INCLUDE) reduce 106) ((T_INCLUDE_ONCE) reduce 106) ((T_INLINE_HTML) reduce 106) ((T_INTERFACE) reduce 106) ((T_INT_CAST) reduce 106) ((T_ISSET) reduce 106) ((T_LINE) reduce 106) ((T_LIST) reduce 106) ((T_LNUMBER) reduce 106) ((T_MATCH) reduce 106) ((T_METHOD_C) reduce 106) ((T_NAME_FULLY_QUALIFIED) reduce 106) ((T_NAME_QUALIFIED) reduce 106) ((T_NAME_RELATIVE) reduce 106) ((T_NEW) reduce 106) ((T_NS_C) reduce 106) ((T_OBJECT_CAST) reduce 106) ((T_PRINT) reduce 106) ((T_PROPERTY_C) reduce 106) ((T_READONLY) reduce 106) ((T_REQUIRE) reduce 106) ((T_REQUIRE_ONCE) reduce 106) ((T_RETURN) reduce 106) ((T_START_HEREDOC) reduce 106) ((T_STATIC) reduce 106) ((T_STRING) reduce 106) ((T_STRING_CAST) reduce 106) ((T_SWITCH) reduce 106) ((T_THROW) reduce 106) ((T_TRAIT) reduce 106) ((T_TRAIT_C) reduce 106) ((T_TRY) reduce 106) ((T_UNSET) reduce 106) ((T_UNSET_CAST) reduce 106) ((T_VARIABLE) reduce 106) ((T_VOID_CAST) reduce 106) ((T_WHILE) reduce 106) ((T_YIELD) reduce 106) ((T_YIELD_FROM) reduce 106) (("[") reduce 106) (("`") reduce 106) (("{") reduce 106) (("~") reduce 106)) 18265 ((("!") reduce 143) (("\"") reduce 143) (("$") reduce 143) (("(") reduce 143) (("+") reduce 143) (("-") reduce 143) ((";") reduce 143) (("@") reduce 143) ((T_ABSTRACT) reduce 143) ((T_ARRAY) reduce 143) ((T_ARRAY_CAST) reduce 143) ((T_ATTRIBUTE) reduce 143) ((T_BOOL_CAST) reduce 143) ((T_BREAK) reduce 143) ((T_CASE) reduce 143) ((T_CLASS) reduce 143) ((T_CLASS_C) reduce 143) ((T_CLONE) reduce 143) ((T_CONSTANT_ENCAPSED_STRING) reduce 143) ((T_CONTINUE) reduce 143) ((T_DEC) reduce 143) ((T_DECLARE) reduce 143) ((T_DEFAULT) reduce 143) ((T_DIR) reduce 143) ((T_DNUMBER) reduce 143) ((T_DO) reduce 143) ((T_DOUBLE_CAST) reduce 143) ((T_ECHO) reduce 143) ((T_EMPTY) reduce 143) ((T_ENDSWITCH) reduce 143) ((T_ENUM) reduce 143) ((T_EVAL) reduce 143) ((T_EXIT) reduce 143) ((T_FILE) reduce 143) ((T_FINAL) reduce 143) ((T_FN) reduce 143) ((T_FOR) reduce 143) ((T_FOREACH) reduce 143) ((T_FUNCTION) reduce 143) ((T_FUNC_C) reduce 143) ((T_GLOBAL) reduce 143) ((T_GOTO) reduce 143) ((T_HALT_COMPILER) reduce 143) ((T_IF) reduce 143) ((T_INC) reduce 143) ((T_INCLUDE) reduce 143) ((T_INCLUDE_ONCE) reduce 143) ((T_INLINE_HTML) reduce 143) ((T_INTERFACE) reduce 143) ((T_INT_CAST) reduce 143) ((T_ISSET) reduce 143) ((T_LINE) reduce 143) ((T_LIST) reduce 143) ((T_LNUMBER) reduce 143) ((T_MATCH) reduce 143) ((T_METHOD_C) reduce 143) ((T_NAME_FULLY_QUALIFIED) reduce 143) ((T_NAME_QUALIFIED) reduce 143) ((T_NAME_RELATIVE) reduce 143) ((T_NEW) reduce 143) ((T_NS_C) reduce 143) ((T_OBJECT_CAST) reduce 143) ((T_PRINT) reduce 143) ((T_PROPERTY_C) reduce 143) ((T_READONLY) reduce 143) ((T_REQUIRE) reduce 143) ((T_REQUIRE_ONCE) reduce 143) ((T_RETURN) reduce 143) ((T_START_HEREDOC) reduce 143) ((T_STATIC) reduce 143) ((T_STRING) reduce 143) ((T_STRING_CAST) reduce 143) ((T_SWITCH) reduce 143) ((T_THROW) reduce 143) ((T_TRAIT) reduce 143) ((T_TRAIT_C) reduce 143) ((T_TRY) reduce 143) ((T_UNSET) reduce 143) ((T_UNSET_CAST) reduce 143) ((T_VARIABLE) reduce 143) ((T_VOID_CAST) reduce 143) ((T_WHILE) reduce 143) ((T_YIELD) reduce 143) ((T_YIELD_FROM) reduce 143) (("[") reduce 143) (("`") reduce 143) (("{") reduce 143) (("~") reduce 143)) 18266 ((("!") reduce 105) (("\"") reduce 105) (("$") reduce 105) (("(") reduce 105) (("+") reduce 105) (("-") reduce 105) ((";") reduce 105) (("@") reduce 105) ((T_ABSTRACT) reduce 105) ((T_ARRAY) reduce 105) ((T_ARRAY_CAST) reduce 105) ((T_ATTRIBUTE) reduce 105) ((T_BOOL_CAST) reduce 105) ((T_BREAK) reduce 105) ((T_CASE) reduce 105) ((T_CLASS) reduce 105) ((T_CLASS_C) reduce 105) ((T_CLONE) reduce 105) ((T_CONSTANT_ENCAPSED_STRING) reduce 105) ((T_CONTINUE) reduce 105) ((T_DEC) reduce 105) ((T_DECLARE) reduce 105) ((T_DEFAULT) reduce 105) ((T_DIR) reduce 105) ((T_DNUMBER) reduce 105) ((T_DO) reduce 105) ((T_DOUBLE_CAST) reduce 105) ((T_ECHO) reduce 105) ((T_EMPTY) reduce 105) ((T_ENDSWITCH) reduce 105) ((T_ENUM) reduce 105) ((T_EVAL) reduce 105) ((T_EXIT) reduce 105) ((T_FILE) reduce 105) ((T_FINAL) reduce 105) ((T_FN) reduce 105) ((T_FOR) reduce 105) ((T_FOREACH) reduce 105) ((T_FUNCTION) reduce 105) ((T_FUNC_C) reduce 105) ((T_GLOBAL) reduce 105) ((T_GOTO) reduce 105) ((T_HALT_COMPILER) reduce 105) ((T_IF) reduce 105) ((T_INC) reduce 105) ((T_INCLUDE) reduce 105) ((T_INCLUDE_ONCE) reduce 105) ((T_INLINE_HTML) reduce 105) ((T_INTERFACE) reduce 105) ((T_INT_CAST) reduce 105) ((T_ISSET) reduce 105) ((T_LINE) reduce 105) ((T_LIST) reduce 105) ((T_LNUMBER) reduce 105) ((T_MATCH) reduce 105) ((T_METHOD_C) reduce 105) ((T_NAME_FULLY_QUALIFIED) reduce 105) ((T_NAME_QUALIFIED) reduce 105) ((T_NAME_RELATIVE) reduce 105) ((T_NEW) reduce 105) ((T_NS_C) reduce 105) ((T_OBJECT_CAST) reduce 105) ((T_PRINT) reduce 105) ((T_PROPERTY_C) reduce 105) ((T_READONLY) reduce 105) ((T_REQUIRE) reduce 105) ((T_REQUIRE_ONCE) reduce 105) ((T_RETURN) reduce 105) ((T_START_HEREDOC) reduce 105) ((T_STATIC) reduce 105) ((T_STRING) reduce 105) ((T_STRING_CAST) reduce 105) ((T_SWITCH) reduce 105) ((T_THROW) reduce 105) ((T_TRAIT) reduce 105) ((T_TRAIT_C) reduce 105) ((T_TRY) reduce 105) ((T_UNSET) reduce 105) ((T_UNSET_CAST) reduce 105) ((T_VARIABLE) reduce 105) ((T_VOID_CAST) reduce 105) ((T_WHILE) reduce 105) ((T_YIELD) reduce 105) ((T_YIELD_FROM) reduce 105) (("[") reduce 105) (("`") reduce 105) (("{") reduce 105) (("~") reduce 105)) 18269 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CASE) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DEFAULT) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_EMPTY) reduce 147) ((T_ENDSWITCH) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_VOID_CAST) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 18275 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CASE) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DEFAULT) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_ELSE) reduce 166) ((T_ELSEIF) reduce 166) ((T_EMPTY) reduce 166) ((T_ENDSWITCH) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_VOID_CAST) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 18285 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CASE) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DEFAULT) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_ELSE) reduce 160) ((T_ELSEIF) reduce 160) ((T_EMPTY) reduce 160) ((T_ENDSWITCH) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_VOID_CAST) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 18294 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CASE) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DEFAULT) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_ELSE) reduce 149) ((T_ELSEIF) reduce 149) ((T_EMPTY) reduce 149) ((T_ENDSWITCH) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_VOID_CAST) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 18297 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CASE) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DEFAULT) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_ELSE) reduce 148) ((T_ELSEIF) reduce 148) ((T_EMPTY) reduce 148) ((T_ENDSWITCH) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_VOID_CAST) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 18299 ((("!") reduce 245) (("\"") reduce 245) (("$") reduce 245) (("(") reduce 245) (("+") reduce 245) (("-") reduce 245) ((";") reduce 245) (("@") reduce 245) ((T_ABSTRACT) reduce 245) ((T_ARRAY) reduce 245) ((T_ARRAY_CAST) reduce 245) ((T_ATTRIBUTE) reduce 245) ((T_BOOL_CAST) reduce 245) ((T_BREAK) reduce 245) ((T_CASE) reduce 245) ((T_CLASS) reduce 245) ((T_CLASS_C) reduce 245) ((T_CLONE) reduce 245) ((T_CONSTANT_ENCAPSED_STRING) reduce 245) ((T_CONTINUE) reduce 245) ((T_DEC) reduce 245) ((T_DECLARE) reduce 245) ((T_DEFAULT) reduce 245) ((T_DIR) reduce 245) ((T_DNUMBER) reduce 245) ((T_DO) reduce 245) ((T_DOUBLE_CAST) reduce 245) ((T_ECHO) reduce 245) ((T_ELSE) reduce 245) ((T_ELSEIF) reduce 245) ((T_EMPTY) reduce 245) ((T_ENDSWITCH) reduce 245) ((T_ENUM) reduce 245) ((T_EVAL) reduce 245) ((T_EXIT) reduce 245) ((T_FILE) reduce 245) ((T_FINAL) reduce 245) ((T_FN) reduce 245) ((T_FOR) reduce 245) ((T_FOREACH) reduce 245) ((T_FUNCTION) reduce 245) ((T_FUNC_C) reduce 245) ((T_GLOBAL) reduce 245) ((T_GOTO) reduce 245) ((T_HALT_COMPILER) reduce 245) ((T_IF) reduce 245) ((T_INC) reduce 245) ((T_INCLUDE) reduce 245) ((T_INCLUDE_ONCE) reduce 245) ((T_INLINE_HTML) reduce 245) ((T_INTERFACE) reduce 245) ((T_INT_CAST) reduce 245) ((T_ISSET) reduce 245) ((T_LINE) reduce 245) ((T_LIST) reduce 245) ((T_LNUMBER) reduce 245) ((T_MATCH) reduce 245) ((T_METHOD_C) reduce 245) ((T_NAME_FULLY_QUALIFIED) reduce 245) ((T_NAME_QUALIFIED) reduce 245) ((T_NAME_RELATIVE) reduce 245) ((T_NEW) reduce 245) ((T_NS_C) reduce 245) ((T_OBJECT_CAST) reduce 245) ((T_PRINT) reduce 245) ((T_PROPERTY_C) reduce 245) ((T_READONLY) reduce 245) ((T_REQUIRE) reduce 245) ((T_REQUIRE_ONCE) reduce 245) ((T_RETURN) reduce 245) ((T_START_HEREDOC) reduce 245) ((T_STATIC) reduce 245) ((T_STRING) reduce 245) ((T_STRING_CAST) reduce 245) ((T_SWITCH) reduce 245) ((T_THROW) reduce 245) ((T_TRAIT) reduce 245) ((T_TRAIT_C) reduce 245) ((T_TRY) reduce 245) ((T_UNSET) reduce 245) ((T_UNSET_CAST) reduce 245) ((T_VARIABLE) reduce 245) ((T_VOID_CAST) reduce 245) ((T_WHILE) reduce 245) ((T_YIELD) reduce 245) ((T_YIELD_FROM) reduce 245) (("[") reduce 245) (("`") reduce 245) (("{") reduce 245) (("~") reduce 245)) 18302 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CASE) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DEFAULT) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_ELSE) reduce 147) ((T_ELSEIF) reduce 147) ((T_EMPTY) reduce 147) ((T_ENDSWITCH) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_VOID_CAST) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 18304 ((("!") reduce 247) (("\"") reduce 247) (("$") reduce 247) (("(") reduce 247) (("+") reduce 247) (("-") reduce 247) ((";") reduce 247) (("@") reduce 247) ((T_ABSTRACT) reduce 247) ((T_ARRAY) reduce 247) ((T_ARRAY_CAST) reduce 247) ((T_ATTRIBUTE) reduce 247) ((T_BOOL_CAST) reduce 247) ((T_BREAK) reduce 247) ((T_CASE) reduce 247) ((T_CLASS) reduce 247) ((T_CLASS_C) reduce 247) ((T_CLONE) reduce 247) ((T_CONSTANT_ENCAPSED_STRING) reduce 247) ((T_CONTINUE) reduce 247) ((T_DEC) reduce 247) ((T_DECLARE) reduce 247) ((T_DEFAULT) reduce 247) ((T_DIR) reduce 247) ((T_DNUMBER) reduce 247) ((T_DO) reduce 247) ((T_DOUBLE_CAST) reduce 247) ((T_ECHO) reduce 247) ((T_ELSE) reduce 247) ((T_ELSEIF) reduce 247) ((T_EMPTY) reduce 247) ((T_ENDSWITCH) reduce 247) ((T_ENUM) reduce 247) ((T_EVAL) reduce 247) ((T_EXIT) reduce 247) ((T_FILE) reduce 247) ((T_FINAL) reduce 247) ((T_FN) reduce 247) ((T_FOR) reduce 247) ((T_FOREACH) reduce 247) ((T_FUNCTION) reduce 247) ((T_FUNC_C) reduce 247) ((T_GLOBAL) reduce 247) ((T_GOTO) reduce 247) ((T_HALT_COMPILER) reduce 247) ((T_IF) reduce 247) ((T_INC) reduce 247) ((T_INCLUDE) reduce 247) ((T_INCLUDE_ONCE) reduce 247) ((T_INLINE_HTML) reduce 247) ((T_INTERFACE) reduce 247) ((T_INT_CAST) reduce 247) ((T_ISSET) reduce 247) ((T_LINE) reduce 247) ((T_LIST) reduce 247) ((T_LNUMBER) reduce 247) ((T_MATCH) reduce 247) ((T_METHOD_C) reduce 247) ((T_NAME_FULLY_QUALIFIED) reduce 247) ((T_NAME_QUALIFIED) reduce 247) ((T_NAME_RELATIVE) reduce 247) ((T_NEW) reduce 247) ((T_NS_C) reduce 247) ((T_OBJECT_CAST) reduce 247) ((T_PRINT) reduce 247) ((T_PROPERTY_C) reduce 247) ((T_READONLY) reduce 247) ((T_REQUIRE) reduce 247) ((T_REQUIRE_ONCE) reduce 247) ((T_RETURN) reduce 247) ((T_START_HEREDOC) reduce 247) ((T_STATIC) reduce 247) ((T_STRING) reduce 247) ((T_STRING_CAST) reduce 247) ((T_SWITCH) reduce 247) ((T_THROW) reduce 247) ((T_TRAIT) reduce 247) ((T_TRAIT_C) reduce 247) ((T_TRY) reduce 247) ((T_UNSET) reduce 247) ((T_UNSET_CAST) reduce 247) ((T_VARIABLE) reduce 247) ((T_VOID_CAST) reduce 247) ((T_WHILE) reduce 247) ((T_YIELD) reduce 247) ((T_YIELD_FROM) reduce 247) (("[") reduce 247) (("`") reduce 247) (("{") reduce 247) (("~") reduce 247)) 18305 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CASE) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DEFAULT) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_ELSE) reduce 161) ((T_ELSEIF) reduce 161) ((T_EMPTY) reduce 161) ((T_ENDSWITCH) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_VOID_CAST) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 18308 ((("!") reduce 250) (("\"") reduce 250) (("$") reduce 250) (("(") reduce 250) (("+") reduce 250) (("-") reduce 250) ((";") reduce 250) (("@") reduce 250) ((T_ABSTRACT) reduce 250) ((T_ARRAY) reduce 250) ((T_ARRAY_CAST) reduce 250) ((T_ATTRIBUTE) reduce 250) ((T_BOOL_CAST) reduce 250) ((T_BREAK) reduce 250) ((T_CASE) reduce 250) ((T_CLASS) reduce 250) ((T_CLASS_C) reduce 250) ((T_CLONE) reduce 250) ((T_CONSTANT_ENCAPSED_STRING) reduce 250) ((T_CONTINUE) reduce 250) ((T_DEC) reduce 250) ((T_DECLARE) reduce 250) ((T_DEFAULT) reduce 250) ((T_DIR) reduce 250) ((T_DNUMBER) reduce 250) ((T_DO) reduce 250) ((T_DOUBLE_CAST) reduce 250) ((T_ECHO) reduce 250) ((T_ELSE) reduce 250) ((T_ELSEIF) reduce 250) ((T_EMPTY) reduce 250) ((T_ENDSWITCH) reduce 250) ((T_ENUM) reduce 250) ((T_EVAL) reduce 250) ((T_EXIT) reduce 250) ((T_FILE) reduce 250) ((T_FINAL) reduce 250) ((T_FN) reduce 250) ((T_FOR) reduce 250) ((T_FOREACH) reduce 250) ((T_FUNCTION) reduce 250) ((T_FUNC_C) reduce 250) ((T_GLOBAL) reduce 250) ((T_GOTO) reduce 250) ((T_HALT_COMPILER) reduce 250) ((T_IF) reduce 250) ((T_INC) reduce 250) ((T_INCLUDE) reduce 250) ((T_INCLUDE_ONCE) reduce 250) ((T_INLINE_HTML) reduce 250) ((T_INTERFACE) reduce 250) ((T_INT_CAST) reduce 250) ((T_ISSET) reduce 250) ((T_LINE) reduce 250) ((T_LIST) reduce 250) ((T_LNUMBER) reduce 250) ((T_MATCH) reduce 250) ((T_METHOD_C) reduce 250) ((T_NAME_FULLY_QUALIFIED) reduce 250) ((T_NAME_QUALIFIED) reduce 250) ((T_NAME_RELATIVE) reduce 250) ((T_NEW) reduce 250) ((T_NS_C) reduce 250) ((T_OBJECT_CAST) reduce 250) ((T_PRINT) reduce 250) ((T_PROPERTY_C) reduce 250) ((T_READONLY) reduce 250) ((T_REQUIRE) reduce 250) ((T_REQUIRE_ONCE) reduce 250) ((T_RETURN) reduce 250) ((T_START_HEREDOC) reduce 250) ((T_STATIC) reduce 250) ((T_STRING) reduce 250) ((T_STRING_CAST) reduce 250) ((T_SWITCH) reduce 250) ((T_THROW) reduce 250) ((T_TRAIT) reduce 250) ((T_TRAIT_C) reduce 250) ((T_TRY) reduce 250) ((T_UNSET) reduce 250) ((T_UNSET_CAST) reduce 250) ((T_VARIABLE) reduce 250) ((T_VOID_CAST) reduce 250) ((T_WHILE) reduce 250) ((T_YIELD) reduce 250) ((T_YIELD_FROM) reduce 250) (("[") reduce 250) (("`") reduce 250) (("{") reduce 250) (("~") reduce 250)) 18312 ((("!") reduce 251) (("\"") reduce 251) (("$") reduce 251) (("(") reduce 251) (("+") reduce 251) (("-") reduce 251) ((";") reduce 251) (("@") reduce 251) ((T_ABSTRACT) reduce 251) ((T_ARRAY) reduce 251) ((T_ARRAY_CAST) reduce 251) ((T_ATTRIBUTE) reduce 251) ((T_BOOL_CAST) reduce 251) ((T_BREAK) reduce 251) ((T_CASE) reduce 251) ((T_CLASS) reduce 251) ((T_CLASS_C) reduce 251) ((T_CLONE) reduce 251) ((T_CONSTANT_ENCAPSED_STRING) reduce 251) ((T_CONTINUE) reduce 251) ((T_DEC) reduce 251) ((T_DECLARE) reduce 251) ((T_DEFAULT) reduce 251) ((T_DIR) reduce 251) ((T_DNUMBER) reduce 251) ((T_DO) reduce 251) ((T_DOUBLE_CAST) reduce 251) ((T_ECHO) reduce 251) ((T_ELSE) reduce 251) ((T_ELSEIF) reduce 251) ((T_EMPTY) reduce 251) ((T_ENDSWITCH) reduce 251) ((T_ENUM) reduce 251) ((T_EVAL) reduce 251) ((T_EXIT) reduce 251) ((T_FILE) reduce 251) ((T_FINAL) reduce 251) ((T_FN) reduce 251) ((T_FOR) reduce 251) ((T_FOREACH) reduce 251) ((T_FUNCTION) reduce 251) ((T_FUNC_C) reduce 251) ((T_GLOBAL) reduce 251) ((T_GOTO) reduce 251) ((T_HALT_COMPILER) reduce 251) ((T_IF) reduce 251) ((T_INC) reduce 251) ((T_INCLUDE) reduce 251) ((T_INCLUDE_ONCE) reduce 251) ((T_INLINE_HTML) reduce 251) ((T_INTERFACE) reduce 251) ((T_INT_CAST) reduce 251) ((T_ISSET) reduce 251) ((T_LINE) reduce 251) ((T_LIST) reduce 251) ((T_LNUMBER) reduce 251) ((T_MATCH) reduce 251) ((T_METHOD_C) reduce 251) ((T_NAME_FULLY_QUALIFIED) reduce 251) ((T_NAME_QUALIFIED) reduce 251) ((T_NAME_RELATIVE) reduce 251) ((T_NEW) reduce 251) ((T_NS_C) reduce 251) ((T_OBJECT_CAST) reduce 251) ((T_PRINT) reduce 251) ((T_PROPERTY_C) reduce 251) ((T_READONLY) reduce 251) ((T_REQUIRE) reduce 251) ((T_REQUIRE_ONCE) reduce 251) ((T_RETURN) reduce 251) ((T_START_HEREDOC) reduce 251) ((T_STATIC) reduce 251) ((T_STRING) reduce 251) ((T_STRING_CAST) reduce 251) ((T_SWITCH) reduce 251) ((T_THROW) reduce 251) ((T_TRAIT) reduce 251) ((T_TRAIT_C) reduce 251) ((T_TRY) reduce 251) ((T_UNSET) reduce 251) ((T_UNSET_CAST) reduce 251) ((T_VARIABLE) reduce 251) ((T_VOID_CAST) reduce 251) ((T_WHILE) reduce 251) ((T_YIELD) reduce 251) ((T_YIELD_FROM) reduce 251) (("[") reduce 251) (("`") reduce 251) (("{") reduce 251) (("~") reduce 251)) 18317 ((("!") reduce 242) (("\"") reduce 242) (("$") reduce 242) (("(") reduce 242) (("+") reduce 242) (("-") reduce 242) ((";") reduce 242) (("@") reduce 242) ((T_ABSTRACT) reduce 242) ((T_ARRAY) reduce 242) ((T_ARRAY_CAST) reduce 242) ((T_ATTRIBUTE) reduce 242) ((T_BOOL_CAST) reduce 242) ((T_BREAK) reduce 242) ((T_CASE) reduce 242) ((T_CLASS) reduce 242) ((T_CLASS_C) reduce 242) ((T_CLONE) reduce 242) ((T_CONSTANT_ENCAPSED_STRING) reduce 242) ((T_CONTINUE) reduce 242) ((T_DEC) reduce 242) ((T_DECLARE) reduce 242) ((T_DEFAULT) reduce 242) ((T_DIR) reduce 242) ((T_DNUMBER) reduce 242) ((T_DO) reduce 242) ((T_DOUBLE_CAST) reduce 242) ((T_ECHO) reduce 242) ((T_ELSE) reduce 242) ((T_ELSEIF) reduce 242) ((T_EMPTY) reduce 242) ((T_ENDSWITCH) reduce 242) ((T_ENUM) reduce 242) ((T_EVAL) reduce 242) ((T_EXIT) reduce 242) ((T_FILE) reduce 242) ((T_FINAL) reduce 242) ((T_FN) reduce 242) ((T_FOR) reduce 242) ((T_FOREACH) reduce 242) ((T_FUNCTION) reduce 242) ((T_FUNC_C) reduce 242) ((T_GLOBAL) reduce 242) ((T_GOTO) reduce 242) ((T_HALT_COMPILER) reduce 242) ((T_IF) reduce 242) ((T_INC) reduce 242) ((T_INCLUDE) reduce 242) ((T_INCLUDE_ONCE) reduce 242) ((T_INLINE_HTML) reduce 242) ((T_INTERFACE) reduce 242) ((T_INT_CAST) reduce 242) ((T_ISSET) reduce 242) ((T_LINE) reduce 242) ((T_LIST) reduce 242) ((T_LNUMBER) reduce 242) ((T_MATCH) reduce 242) ((T_METHOD_C) reduce 242) ((T_NAME_FULLY_QUALIFIED) reduce 242) ((T_NAME_QUALIFIED) reduce 242) ((T_NAME_RELATIVE) reduce 242) ((T_NEW) reduce 242) ((T_NS_C) reduce 242) ((T_OBJECT_CAST) reduce 242) ((T_PRINT) reduce 242) ((T_PROPERTY_C) reduce 242) ((T_READONLY) reduce 242) ((T_REQUIRE) reduce 242) ((T_REQUIRE_ONCE) reduce 242) ((T_RETURN) reduce 242) ((T_START_HEREDOC) reduce 242) ((T_STATIC) reduce 242) ((T_STRING) reduce 242) ((T_STRING_CAST) reduce 242) ((T_SWITCH) reduce 242) ((T_THROW) reduce 242) ((T_TRAIT) reduce 242) ((T_TRAIT_C) reduce 242) ((T_TRY) reduce 242) ((T_UNSET) reduce 242) ((T_UNSET_CAST) reduce 242) ((T_VARIABLE) reduce 242) ((T_VOID_CAST) reduce 242) ((T_WHILE) reduce 242) ((T_YIELD) reduce 242) ((T_YIELD_FROM) reduce 242) (("[") reduce 242) (("`") reduce 242) (("{") reduce 242) (("~") reduce 242)) 18318 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CASE) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DEFAULT) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_ELSE) reduce 150) ((T_ELSEIF) reduce 150) ((T_EMPTY) reduce 150) ((T_ENDSWITCH) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_VOID_CAST) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 18321 ((("!") reduce 243) (("\"") reduce 243) (("$") reduce 243) (("(") reduce 243) (("+") reduce 243) (("-") reduce 243) ((";") reduce 243) (("@") reduce 243) ((T_ABSTRACT) reduce 243) ((T_ARRAY) reduce 243) ((T_ARRAY_CAST) reduce 243) ((T_ATTRIBUTE) reduce 243) ((T_BOOL_CAST) reduce 243) ((T_BREAK) reduce 243) ((T_CASE) reduce 243) ((T_CLASS) reduce 243) ((T_CLASS_C) reduce 243) ((T_CLONE) reduce 243) ((T_CONSTANT_ENCAPSED_STRING) reduce 243) ((T_CONTINUE) reduce 243) ((T_DEC) reduce 243) ((T_DECLARE) reduce 243) ((T_DEFAULT) reduce 243) ((T_DIR) reduce 243) ((T_DNUMBER) reduce 243) ((T_DO) reduce 243) ((T_DOUBLE_CAST) reduce 243) ((T_ECHO) reduce 243) ((T_ELSE) reduce 243) ((T_ELSEIF) reduce 243) ((T_EMPTY) reduce 243) ((T_ENDSWITCH) reduce 243) ((T_ENUM) reduce 243) ((T_EVAL) reduce 243) ((T_EXIT) reduce 243) ((T_FILE) reduce 243) ((T_FINAL) reduce 243) ((T_FN) reduce 243) ((T_FOR) reduce 243) ((T_FOREACH) reduce 243) ((T_FUNCTION) reduce 243) ((T_FUNC_C) reduce 243) ((T_GLOBAL) reduce 243) ((T_GOTO) reduce 243) ((T_HALT_COMPILER) reduce 243) ((T_IF) reduce 243) ((T_INC) reduce 243) ((T_INCLUDE) reduce 243) ((T_INCLUDE_ONCE) reduce 243) ((T_INLINE_HTML) reduce 243) ((T_INTERFACE) reduce 243) ((T_INT_CAST) reduce 243) ((T_ISSET) reduce 243) ((T_LINE) reduce 243) ((T_LIST) reduce 243) ((T_LNUMBER) reduce 243) ((T_MATCH) reduce 243) ((T_METHOD_C) reduce 243) ((T_NAME_FULLY_QUALIFIED) reduce 243) ((T_NAME_QUALIFIED) reduce 243) ((T_NAME_RELATIVE) reduce 243) ((T_NEW) reduce 243) ((T_NS_C) reduce 243) ((T_OBJECT_CAST) reduce 243) ((T_PRINT) reduce 243) ((T_PROPERTY_C) reduce 243) ((T_READONLY) reduce 243) ((T_REQUIRE) reduce 243) ((T_REQUIRE_ONCE) reduce 243) ((T_RETURN) reduce 243) ((T_START_HEREDOC) reduce 243) ((T_STATIC) reduce 243) ((T_STRING) reduce 243) ((T_STRING_CAST) reduce 243) ((T_SWITCH) reduce 243) ((T_THROW) reduce 243) ((T_TRAIT) reduce 243) ((T_TRAIT_C) reduce 243) ((T_TRY) reduce 243) ((T_UNSET) reduce 243) ((T_UNSET_CAST) reduce 243) ((T_VARIABLE) reduce 243) ((T_VOID_CAST) reduce 243) ((T_WHILE) reduce 243) ((T_YIELD) reduce 243) ((T_YIELD_FROM) reduce 243) (("[") reduce 243) (("`") reduce 243) (("{") reduce 243) (("~") reduce 243)) 18323 ((("!") reduce 170) (("\"") reduce 170) (("$") reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CASE) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DEFAULT) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_ELSE) reduce 170) ((T_ELSEIF) reduce 170) ((T_EMPTY) reduce 170) ((T_ENDSWITCH) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_VARIABLE) reduce 170) ((T_VOID_CAST) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("~") reduce 170)) 18328 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CASE) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DEFAULT) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_ELSE) reduce 162) ((T_ELSEIF) reduce 162) ((T_EMPTY) reduce 162) ((T_ENDSWITCH) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_VOID_CAST) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 18331 ((("!") reduce 171) (("\"") reduce 171) (("$") reduce 171) (("(") reduce 171) (("+") reduce 171) (("-") reduce 171) ((";") reduce 171) (("@") reduce 171) ((T_ABSTRACT) reduce 171) ((T_ARRAY) reduce 171) ((T_ARRAY_CAST) reduce 171) ((T_ATTRIBUTE) reduce 171) ((T_BOOL_CAST) reduce 171) ((T_BREAK) reduce 171) ((T_CASE) reduce 171) ((T_CATCH) reduce 171) ((T_CLASS) reduce 171) ((T_CLASS_C) reduce 171) ((T_CLONE) reduce 171) ((T_CONSTANT_ENCAPSED_STRING) reduce 171) ((T_CONTINUE) reduce 171) ((T_DEC) reduce 171) ((T_DECLARE) reduce 171) ((T_DEFAULT) reduce 171) ((T_DIR) reduce 171) ((T_DNUMBER) reduce 171) ((T_DO) reduce 171) ((T_DOUBLE_CAST) reduce 171) ((T_ECHO) reduce 171) ((T_ELSE) reduce 171) ((T_ELSEIF) reduce 171) ((T_EMPTY) reduce 171) ((T_ENDSWITCH) reduce 171) ((T_ENUM) reduce 171) ((T_EVAL) reduce 171) ((T_EXIT) reduce 171) ((T_FILE) reduce 171) ((T_FINAL) reduce 171) ((T_FINALLY) reduce 171) ((T_FN) reduce 171) ((T_FOR) reduce 171) ((T_FOREACH) reduce 171) ((T_FUNCTION) reduce 171) ((T_FUNC_C) reduce 171) ((T_GLOBAL) reduce 171) ((T_GOTO) reduce 171) ((T_HALT_COMPILER) reduce 171) ((T_IF) reduce 171) ((T_INC) reduce 171) ((T_INCLUDE) reduce 171) ((T_INCLUDE_ONCE) reduce 171) ((T_INLINE_HTML) reduce 171) ((T_INTERFACE) reduce 171) ((T_INT_CAST) reduce 171) ((T_ISSET) reduce 171) ((T_LINE) reduce 171) ((T_LIST) reduce 171) ((T_LNUMBER) reduce 171) ((T_MATCH) reduce 171) ((T_METHOD_C) reduce 171) ((T_NAME_FULLY_QUALIFIED) reduce 171) ((T_NAME_QUALIFIED) reduce 171) ((T_NAME_RELATIVE) reduce 171) ((T_NEW) reduce 171) ((T_NS_C) reduce 171) ((T_OBJECT_CAST) reduce 171) ((T_PRINT) reduce 171) ((T_PROPERTY_C) reduce 171) ((T_READONLY) reduce 171) ((T_REQUIRE) reduce 171) ((T_REQUIRE_ONCE) reduce 171) ((T_RETURN) reduce 171) ((T_START_HEREDOC) reduce 171) ((T_STATIC) reduce 171) ((T_STRING) reduce 171) ((T_STRING_CAST) reduce 171) ((T_SWITCH) reduce 171) ((T_THROW) reduce 171) ((T_TRAIT) reduce 171) ((T_TRAIT_C) reduce 171) ((T_TRY) reduce 171) ((T_UNSET) reduce 171) ((T_UNSET_CAST) reduce 171) ((T_VARIABLE) reduce 171) ((T_VOID_CAST) reduce 171) ((T_WHILE) reduce 171) ((T_YIELD) reduce 171) ((T_YIELD_FROM) reduce 171) (("[") reduce 171) (("`") reduce 171) (("{") reduce 171) (("~") reduce 171)) 18332 ((("!") reduce 177) (("\"") reduce 177) (("$") reduce 177) (("(") reduce 177) (("+") reduce 177) (("-") reduce 177) ((";") reduce 177) (("@") reduce 177) ((T_ABSTRACT) reduce 177) ((T_ARRAY) reduce 177) ((T_ARRAY_CAST) reduce 177) ((T_ATTRIBUTE) reduce 177) ((T_BOOL_CAST) reduce 177) ((T_BREAK) reduce 177) ((T_CASE) reduce 177) ((T_CATCH) shift) ((T_CLASS) reduce 177) ((T_CLASS_C) reduce 177) ((T_CLONE) reduce 177) ((T_CONSTANT_ENCAPSED_STRING) reduce 177) ((T_CONTINUE) reduce 177) ((T_DEC) reduce 177) ((T_DECLARE) reduce 177) ((T_DEFAULT) reduce 177) ((T_DIR) reduce 177) ((T_DNUMBER) reduce 177) ((T_DO) reduce 177) ((T_DOUBLE_CAST) reduce 177) ((T_ECHO) reduce 177) ((T_ELSE) reduce 177) ((T_ELSEIF) reduce 177) ((T_EMPTY) reduce 177) ((T_ENDSWITCH) reduce 177) ((T_ENUM) reduce 177) ((T_EVAL) reduce 177) ((T_EXIT) reduce 177) ((T_FILE) reduce 177) ((T_FINAL) reduce 177) ((T_FINALLY) shift) ((T_FN) reduce 177) ((T_FOR) reduce 177) ((T_FOREACH) reduce 177) ((T_FUNCTION) reduce 177) ((T_FUNC_C) reduce 177) ((T_GLOBAL) reduce 177) ((T_GOTO) reduce 177) ((T_HALT_COMPILER) reduce 177) ((T_IF) reduce 177) ((T_INC) reduce 177) ((T_INCLUDE) reduce 177) ((T_INCLUDE_ONCE) reduce 177) ((T_INLINE_HTML) reduce 177) ((T_INTERFACE) reduce 177) ((T_INT_CAST) reduce 177) ((T_ISSET) reduce 177) ((T_LINE) reduce 177) ((T_LIST) reduce 177) ((T_LNUMBER) reduce 177) ((T_MATCH) reduce 177) ((T_METHOD_C) reduce 177) ((T_NAME_FULLY_QUALIFIED) reduce 177) ((T_NAME_QUALIFIED) reduce 177) ((T_NAME_RELATIVE) reduce 177) ((T_NEW) reduce 177) ((T_NS_C) reduce 177) ((T_OBJECT_CAST) reduce 177) ((T_PRINT) reduce 177) ((T_PROPERTY_C) reduce 177) ((T_READONLY) reduce 177) ((T_REQUIRE) reduce 177) ((T_REQUIRE_ONCE) reduce 177) ((T_RETURN) reduce 177) ((T_START_HEREDOC) reduce 177) ((T_STATIC) reduce 177) ((T_STRING) reduce 177) ((T_STRING_CAST) reduce 177) ((T_SWITCH) reduce 177) ((T_THROW) reduce 177) ((T_TRAIT) reduce 177) ((T_TRAIT_C) reduce 177) ((T_TRY) reduce 177) ((T_UNSET) reduce 177) ((T_UNSET_CAST) reduce 177) ((T_VARIABLE) reduce 177) ((T_VOID_CAST) reduce 177) ((T_WHILE) reduce 177) ((T_YIELD) reduce 177) ((T_YIELD_FROM) reduce 177) (("[") reduce 177) (("`") reduce 177) (("{") reduce 177) (("~") reduce 177)) 18335 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CASE) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DEFAULT) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_ELSE) reduce 167) ((T_ELSEIF) reduce 167) ((T_EMPTY) reduce 167) ((T_ENDSWITCH) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_VOID_CAST) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 18338 ((("!") reduce 178) (("\"") reduce 178) (("$") reduce 178) (("(") reduce 178) (("+") reduce 178) (("-") reduce 178) ((";") reduce 178) (("@") reduce 178) ((T_ABSTRACT) reduce 178) ((T_ARRAY) reduce 178) ((T_ARRAY_CAST) reduce 178) ((T_ATTRIBUTE) reduce 178) ((T_BOOL_CAST) reduce 178) ((T_BREAK) reduce 178) ((T_CASE) reduce 178) ((T_CLASS) reduce 178) ((T_CLASS_C) reduce 178) ((T_CLONE) reduce 178) ((T_CONSTANT_ENCAPSED_STRING) reduce 178) ((T_CONTINUE) reduce 178) ((T_DEC) reduce 178) ((T_DECLARE) reduce 178) ((T_DEFAULT) reduce 178) ((T_DIR) reduce 178) ((T_DNUMBER) reduce 178) ((T_DO) reduce 178) ((T_DOUBLE_CAST) reduce 178) ((T_ECHO) reduce 178) ((T_ELSE) reduce 178) ((T_ELSEIF) reduce 178) ((T_EMPTY) reduce 178) ((T_ENDSWITCH) reduce 178) ((T_ENUM) reduce 178) ((T_EVAL) reduce 178) ((T_EXIT) reduce 178) ((T_FILE) reduce 178) ((T_FINAL) reduce 178) ((T_FN) reduce 178) ((T_FOR) reduce 178) ((T_FOREACH) reduce 178) ((T_FUNCTION) reduce 178) ((T_FUNC_C) reduce 178) ((T_GLOBAL) reduce 178) ((T_GOTO) reduce 178) ((T_HALT_COMPILER) reduce 178) ((T_IF) reduce 178) ((T_INC) reduce 178) ((T_INCLUDE) reduce 178) ((T_INCLUDE_ONCE) reduce 178) ((T_INLINE_HTML) reduce 178) ((T_INTERFACE) reduce 178) ((T_INT_CAST) reduce 178) ((T_ISSET) reduce 178) ((T_LINE) reduce 178) ((T_LIST) reduce 178) ((T_LNUMBER) reduce 178) ((T_MATCH) reduce 178) ((T_METHOD_C) reduce 178) ((T_NAME_FULLY_QUALIFIED) reduce 178) ((T_NAME_QUALIFIED) reduce 178) ((T_NAME_RELATIVE) reduce 178) ((T_NEW) reduce 178) ((T_NS_C) reduce 178) ((T_OBJECT_CAST) reduce 178) ((T_PRINT) reduce 178) ((T_PROPERTY_C) reduce 178) ((T_READONLY) reduce 178) ((T_REQUIRE) reduce 178) ((T_REQUIRE_ONCE) reduce 178) ((T_RETURN) reduce 178) ((T_START_HEREDOC) reduce 178) ((T_STATIC) reduce 178) ((T_STRING) reduce 178) ((T_STRING_CAST) reduce 178) ((T_SWITCH) reduce 178) ((T_THROW) reduce 178) ((T_TRAIT) reduce 178) ((T_TRAIT_C) reduce 178) ((T_TRY) reduce 178) ((T_UNSET) reduce 178) ((T_UNSET_CAST) reduce 178) ((T_VARIABLE) reduce 178) ((T_VOID_CAST) reduce 178) ((T_WHILE) reduce 178) ((T_YIELD) reduce 178) ((T_YIELD_FROM) reduce 178) (("[") reduce 178) (("`") reduce 178) (("{") reduce 178) (("~") reduce 178)) 18345 ((("!") reduce 172) (("\"") reduce 172) (("$") reduce 172) (("(") reduce 172) (("+") reduce 172) (("-") reduce 172) ((";") reduce 172) (("@") reduce 172) ((T_ABSTRACT) reduce 172) ((T_ARRAY) reduce 172) ((T_ARRAY_CAST) reduce 172) ((T_ATTRIBUTE) reduce 172) ((T_BOOL_CAST) reduce 172) ((T_BREAK) reduce 172) ((T_CASE) reduce 172) ((T_CATCH) reduce 172) ((T_CLASS) reduce 172) ((T_CLASS_C) reduce 172) ((T_CLONE) reduce 172) ((T_CONSTANT_ENCAPSED_STRING) reduce 172) ((T_CONTINUE) reduce 172) ((T_DEC) reduce 172) ((T_DECLARE) reduce 172) ((T_DEFAULT) reduce 172) ((T_DIR) reduce 172) ((T_DNUMBER) reduce 172) ((T_DO) reduce 172) ((T_DOUBLE_CAST) reduce 172) ((T_ECHO) reduce 172) ((T_ELSE) reduce 172) ((T_ELSEIF) reduce 172) ((T_EMPTY) reduce 172) ((T_ENDSWITCH) reduce 172) ((T_ENUM) reduce 172) ((T_EVAL) reduce 172) ((T_EXIT) reduce 172) ((T_FILE) reduce 172) ((T_FINAL) reduce 172) ((T_FINALLY) reduce 172) ((T_FN) reduce 172) ((T_FOR) reduce 172) ((T_FOREACH) reduce 172) ((T_FUNCTION) reduce 172) ((T_FUNC_C) reduce 172) ((T_GLOBAL) reduce 172) ((T_GOTO) reduce 172) ((T_HALT_COMPILER) reduce 172) ((T_IF) reduce 172) ((T_INC) reduce 172) ((T_INCLUDE) reduce 172) ((T_INCLUDE_ONCE) reduce 172) ((T_INLINE_HTML) reduce 172) ((T_INTERFACE) reduce 172) ((T_INT_CAST) reduce 172) ((T_ISSET) reduce 172) ((T_LINE) reduce 172) ((T_LIST) reduce 172) ((T_LNUMBER) reduce 172) ((T_MATCH) reduce 172) ((T_METHOD_C) reduce 172) ((T_NAME_FULLY_QUALIFIED) reduce 172) ((T_NAME_QUALIFIED) reduce 172) ((T_NAME_RELATIVE) reduce 172) ((T_NEW) reduce 172) ((T_NS_C) reduce 172) ((T_OBJECT_CAST) reduce 172) ((T_PRINT) reduce 172) ((T_PROPERTY_C) reduce 172) ((T_READONLY) reduce 172) ((T_REQUIRE) reduce 172) ((T_REQUIRE_ONCE) reduce 172) ((T_RETURN) reduce 172) ((T_START_HEREDOC) reduce 172) ((T_STATIC) reduce 172) ((T_STRING) reduce 172) ((T_STRING_CAST) reduce 172) ((T_SWITCH) reduce 172) ((T_THROW) reduce 172) ((T_TRAIT) reduce 172) ((T_TRAIT_C) reduce 172) ((T_TRY) reduce 172) ((T_UNSET) reduce 172) ((T_UNSET_CAST) reduce 172) ((T_VARIABLE) reduce 172) ((T_VOID_CAST) reduce 172) ((T_WHILE) reduce 172) ((T_YIELD) reduce 172) ((T_YIELD_FROM) reduce 172) (("[") reduce 172) (("`") reduce 172) (("{") reduce 172) (("~") reduce 172)) 18350 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CASE) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DEFAULT) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_ELSE) reduce 153) ((T_ELSEIF) reduce 153) ((T_EMPTY) reduce 153) ((T_ENDSWITCH) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_VOID_CAST) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 18354 ((("!") reduce 224) (("\"") reduce 224) (("$") reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CASE) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DEFAULT) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_ELSE) reduce 224) ((T_ELSEIF) reduce 224) ((T_EMPTY) reduce 224) ((T_ENDSWITCH) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_VARIABLE) reduce 224) ((T_VOID_CAST) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("~") reduce 224)) 18356 ((("!") reduce 225) (("\"") reduce 225) (("$") reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CASE) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DEFAULT) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_ELSE) reduce 225) ((T_ELSEIF) reduce 225) ((T_EMPTY) reduce 225) ((T_ENDSWITCH) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_VARIABLE) reduce 225) ((T_VOID_CAST) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("~") reduce 225)) 18360 ((("!") reduce 226) (("\"") reduce 226) (("$") reduce 226) (("(") reduce 226) (("+") reduce 226) (("-") reduce 226) ((";") reduce 226) (("@") reduce 226) ((T_ABSTRACT) reduce 226) ((T_ARRAY) reduce 226) ((T_ARRAY_CAST) reduce 226) ((T_ATTRIBUTE) reduce 226) ((T_BOOL_CAST) reduce 226) ((T_BREAK) reduce 226) ((T_CASE) reduce 226) ((T_CLASS) reduce 226) ((T_CLASS_C) reduce 226) ((T_CLONE) reduce 226) ((T_CONSTANT_ENCAPSED_STRING) reduce 226) ((T_CONTINUE) reduce 226) ((T_DEC) reduce 226) ((T_DECLARE) reduce 226) ((T_DEFAULT) reduce 226) ((T_DIR) reduce 226) ((T_DNUMBER) reduce 226) ((T_DO) reduce 226) ((T_DOUBLE_CAST) reduce 226) ((T_ECHO) reduce 226) ((T_ELSE) reduce 226) ((T_ELSEIF) reduce 226) ((T_EMPTY) reduce 226) ((T_ENDSWITCH) reduce 226) ((T_ENUM) reduce 226) ((T_EVAL) reduce 226) ((T_EXIT) reduce 226) ((T_FILE) reduce 226) ((T_FINAL) reduce 226) ((T_FN) reduce 226) ((T_FOR) reduce 226) ((T_FOREACH) reduce 226) ((T_FUNCTION) reduce 226) ((T_FUNC_C) reduce 226) ((T_GLOBAL) reduce 226) ((T_GOTO) reduce 226) ((T_HALT_COMPILER) reduce 226) ((T_IF) reduce 226) ((T_INC) reduce 226) ((T_INCLUDE) reduce 226) ((T_INCLUDE_ONCE) reduce 226) ((T_INLINE_HTML) reduce 226) ((T_INTERFACE) reduce 226) ((T_INT_CAST) reduce 226) ((T_ISSET) reduce 226) ((T_LINE) reduce 226) ((T_LIST) reduce 226) ((T_LNUMBER) reduce 226) ((T_MATCH) reduce 226) ((T_METHOD_C) reduce 226) ((T_NAME_FULLY_QUALIFIED) reduce 226) ((T_NAME_QUALIFIED) reduce 226) ((T_NAME_RELATIVE) reduce 226) ((T_NEW) reduce 226) ((T_NS_C) reduce 226) ((T_OBJECT_CAST) reduce 226) ((T_PRINT) reduce 226) ((T_PROPERTY_C) reduce 226) ((T_READONLY) reduce 226) ((T_REQUIRE) reduce 226) ((T_REQUIRE_ONCE) reduce 226) ((T_RETURN) reduce 226) ((T_START_HEREDOC) reduce 226) ((T_STATIC) reduce 226) ((T_STRING) reduce 226) ((T_STRING_CAST) reduce 226) ((T_SWITCH) reduce 226) ((T_THROW) reduce 226) ((T_TRAIT) reduce 226) ((T_TRAIT_C) reduce 226) ((T_TRY) reduce 226) ((T_UNSET) reduce 226) ((T_UNSET_CAST) reduce 226) ((T_VARIABLE) reduce 226) ((T_VOID_CAST) reduce 226) ((T_WHILE) reduce 226) ((T_YIELD) reduce 226) ((T_YIELD_FROM) reduce 226) (("[") reduce 226) (("`") reduce 226) (("{") reduce 226) (("~") reduce 226)) 18363 ((("!") reduce 227) (("\"") reduce 227) (("$") reduce 227) (("(") reduce 227) (("+") reduce 227) (("-") reduce 227) ((";") reduce 227) (("@") reduce 227) ((T_ABSTRACT) reduce 227) ((T_ARRAY) reduce 227) ((T_ARRAY_CAST) reduce 227) ((T_ATTRIBUTE) reduce 227) ((T_BOOL_CAST) reduce 227) ((T_BREAK) reduce 227) ((T_CASE) reduce 227) ((T_CLASS) reduce 227) ((T_CLASS_C) reduce 227) ((T_CLONE) reduce 227) ((T_CONSTANT_ENCAPSED_STRING) reduce 227) ((T_CONTINUE) reduce 227) ((T_DEC) reduce 227) ((T_DECLARE) reduce 227) ((T_DEFAULT) reduce 227) ((T_DIR) reduce 227) ((T_DNUMBER) reduce 227) ((T_DO) reduce 227) ((T_DOUBLE_CAST) reduce 227) ((T_ECHO) reduce 227) ((T_ELSE) reduce 227) ((T_ELSEIF) reduce 227) ((T_EMPTY) reduce 227) ((T_ENDSWITCH) reduce 227) ((T_ENUM) reduce 227) ((T_EVAL) reduce 227) ((T_EXIT) reduce 227) ((T_FILE) reduce 227) ((T_FINAL) reduce 227) ((T_FN) reduce 227) ((T_FOR) reduce 227) ((T_FOREACH) reduce 227) ((T_FUNCTION) reduce 227) ((T_FUNC_C) reduce 227) ((T_GLOBAL) reduce 227) ((T_GOTO) reduce 227) ((T_HALT_COMPILER) reduce 227) ((T_IF) reduce 227) ((T_INC) reduce 227) ((T_INCLUDE) reduce 227) ((T_INCLUDE_ONCE) reduce 227) ((T_INLINE_HTML) reduce 227) ((T_INTERFACE) reduce 227) ((T_INT_CAST) reduce 227) ((T_ISSET) reduce 227) ((T_LINE) reduce 227) ((T_LIST) reduce 227) ((T_LNUMBER) reduce 227) ((T_MATCH) reduce 227) ((T_METHOD_C) reduce 227) ((T_NAME_FULLY_QUALIFIED) reduce 227) ((T_NAME_QUALIFIED) reduce 227) ((T_NAME_RELATIVE) reduce 227) ((T_NEW) reduce 227) ((T_NS_C) reduce 227) ((T_OBJECT_CAST) reduce 227) ((T_PRINT) reduce 227) ((T_PROPERTY_C) reduce 227) ((T_READONLY) reduce 227) ((T_REQUIRE) reduce 227) ((T_REQUIRE_ONCE) reduce 227) ((T_RETURN) reduce 227) ((T_START_HEREDOC) reduce 227) ((T_STATIC) reduce 227) ((T_STRING) reduce 227) ((T_STRING_CAST) reduce 227) ((T_SWITCH) reduce 227) ((T_THROW) reduce 227) ((T_TRAIT) reduce 227) ((T_TRAIT_C) reduce 227) ((T_TRY) reduce 227) ((T_UNSET) reduce 227) ((T_UNSET_CAST) reduce 227) ((T_VARIABLE) reduce 227) ((T_VOID_CAST) reduce 227) ((T_WHILE) reduce 227) ((T_YIELD) reduce 227) ((T_YIELD_FROM) reduce 227) (("[") reduce 227) (("`") reduce 227) (("{") reduce 227) (("~") reduce 227)) 18364 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CASE) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DEFAULT) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_ELSE) reduce 169) ((T_ELSEIF) reduce 169) ((T_EMPTY) reduce 169) ((T_ENDSWITCH) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_VOID_CAST) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 18365 (((",") reduce 312) ((";") reduce 312) (("=") shift)) 18366 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) ((";") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_PIPE) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("^") reduce 486) (("|") reduce 486)) 18367 (((",") reduce 311) ((";") reduce 311)) 18370 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CASE) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DEFAULT) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_ELSE) reduce 158) ((T_ELSEIF) reduce 158) ((T_EMPTY) reduce 158) ((T_ENDSWITCH) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_VOID_CAST) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 18371 (((",") reduce 310) ((";") reduce 310)) 18436 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 313) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 313) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19436 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 546) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19438 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CASE) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DEFAULT) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_ELSE) reduce 156) ((T_ELSEIF) reduce 156) ((T_EMPTY) reduce 156) ((T_ENDSWITCH) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_VOID_CAST) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 19440 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CASE) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DEFAULT) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_ELSE) reduce 168) ((T_ELSEIF) reduce 168) ((T_EMPTY) reduce 168) ((T_ENDSWITCH) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_VOID_CAST) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 19442 (((",") reduce 568) ((";") reduce 568)) 19443 (((",") reduce 308) ((";") reduce 308)) 19445 (((",") reduce 309) ((";") reduce 309)) 19447 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CASE) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DEFAULT) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_ELSE) reduce 157) ((T_ELSEIF) reduce 157) ((T_EMPTY) reduce 157) ((T_ENDSWITCH) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_VOID_CAST) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 19448 (((",") reduce 307) ((";") reduce 307)) 19449 (((",") reduce 570) ((";") reduce 570)) 19452 (((",") reduce 569) ((";") reduce 569)) 19464 ((("%") reduce 540) (("*") reduce 540) (("+") reduce 540) (("-") reduce 540) ((".") reduce 540) (("/") reduce 540) (("<") reduce 540) ((">") reduce 540) (("?") reduce 540) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_AS) reduce 540) ((T_BOOLEAN_AND) reduce 540) ((T_BOOLEAN_OR) reduce 540) ((T_COALESCE) reduce 540) ((T_INSTANCEOF) reduce 540) ((T_IS_EQUAL) reduce 540) ((T_IS_GREATER_OR_EQUAL) reduce 540) ((T_IS_IDENTICAL) reduce 540) ((T_IS_NOT_EQUAL) reduce 540) ((T_IS_NOT_IDENTICAL) reduce 540) ((T_IS_SMALLER_OR_EQUAL) reduce 540) ((T_LOGICAL_AND) reduce 540) ((T_LOGICAL_OR) reduce 540) ((T_LOGICAL_XOR) reduce 540) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 540) ((T_OBJECT_OPERATOR) reduce 540) ((T_PIPE) reduce 540) ((T_POW) reduce 540) ((T_SL) reduce 540) ((T_SPACESHIP) reduce 540) ((T_SR) reduce 540) (("[") reduce 540) (("^") reduce 540) (("|") reduce 540)) 19466 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AS) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_PIPE) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("|") reduce 521)) 19468 ((("%") reduce 534) (("*") reduce 534) (("+") reduce 534) (("-") reduce 534) ((".") reduce 534) (("/") reduce 534) (("<") reduce 534) ((">") reduce 534) (("?") reduce 534) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_AS) reduce 534) ((T_BOOLEAN_AND) reduce 534) ((T_BOOLEAN_OR) reduce 534) ((T_COALESCE) reduce 534) ((T_INSTANCEOF) reduce 534) ((T_IS_EQUAL) reduce 534) ((T_IS_GREATER_OR_EQUAL) reduce 534) ((T_IS_IDENTICAL) reduce 534) ((T_IS_NOT_EQUAL) reduce 534) ((T_IS_NOT_IDENTICAL) reduce 534) ((T_IS_SMALLER_OR_EQUAL) reduce 534) ((T_LOGICAL_AND) reduce 534) ((T_LOGICAL_OR) reduce 534) ((T_LOGICAL_XOR) reduce 534) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PIPE) reduce 534) ((T_POW) reduce 534) ((T_SL) reduce 534) ((T_SPACESHIP) reduce 534) ((T_SR) reduce 534) (("[") reduce 534) (("^") reduce 534) (("|") reduce 534)) 19469 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AS) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_PIPE) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("^") reduce 524) (("|") reduce 524)) 19473 ((("%") reduce 517) (("(") shift) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AS) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_PIPE) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("^") reduce 517) (("|") reduce 517)) 19474 ((("%") reduce 533) (("*") reduce 533) (("+") reduce 533) (("-") reduce 533) ((".") reduce 533) (("/") reduce 533) (("<") reduce 533) ((">") reduce 533) (("?") reduce 533) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_AS) reduce 533) ((T_BOOLEAN_AND) reduce 533) ((T_BOOLEAN_OR) reduce 533) ((T_COALESCE) reduce 533) ((T_INSTANCEOF) reduce 533) ((T_IS_EQUAL) reduce 533) ((T_IS_GREATER_OR_EQUAL) reduce 533) ((T_IS_IDENTICAL) reduce 533) ((T_IS_NOT_EQUAL) reduce 533) ((T_IS_NOT_IDENTICAL) reduce 533) ((T_IS_SMALLER_OR_EQUAL) reduce 533) ((T_LOGICAL_AND) reduce 533) ((T_LOGICAL_OR) reduce 533) ((T_LOGICAL_XOR) reduce 533) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PIPE) reduce 533) ((T_POW) reduce 533) ((T_SL) reduce 533) ((T_SPACESHIP) reduce 533) ((T_SR) reduce 533) (("[") reduce 533) (("^") reduce 533) (("|") reduce 533)) 19475 ((("%") reduce 537) (("*") reduce 537) (("+") reduce 537) (("-") reduce 537) ((".") reduce 537) (("/") reduce 537) (("<") reduce 537) ((">") reduce 537) (("?") reduce 537) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_AS) reduce 537) ((T_BOOLEAN_AND) reduce 537) ((T_BOOLEAN_OR) reduce 537) ((T_COALESCE) reduce 537) ((T_INSTANCEOF) reduce 537) ((T_IS_EQUAL) reduce 537) ((T_IS_GREATER_OR_EQUAL) reduce 537) ((T_IS_IDENTICAL) reduce 537) ((T_IS_NOT_EQUAL) reduce 537) ((T_IS_NOT_IDENTICAL) reduce 537) ((T_IS_SMALLER_OR_EQUAL) reduce 537) ((T_LOGICAL_AND) reduce 537) ((T_LOGICAL_OR) reduce 537) ((T_LOGICAL_XOR) reduce 537) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_PIPE) reduce 537) ((T_POW) reduce 537) ((T_SL) reduce 537) ((T_SPACESHIP) reduce 537) ((T_SR) reduce 537) (("[") reduce 537) (("^") reduce 537) (("|") reduce 537)) 19481 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AS) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PIPE) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("|") reduce 532)) 19483 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AS) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_PIPE) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("^") reduce 523) (("|") reduce 523)) 19485 ((("%") reduce 536) (("*") reduce 536) (("+") reduce 536) (("-") reduce 536) ((".") reduce 536) (("/") reduce 536) (("<") reduce 536) ((">") reduce 536) (("?") reduce 536) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_AS) reduce 536) ((T_BOOLEAN_AND) reduce 536) ((T_BOOLEAN_OR) reduce 536) ((T_COALESCE) reduce 536) ((T_INSTANCEOF) reduce 536) ((T_IS_EQUAL) reduce 536) ((T_IS_GREATER_OR_EQUAL) reduce 536) ((T_IS_IDENTICAL) reduce 536) ((T_IS_NOT_EQUAL) reduce 536) ((T_IS_NOT_IDENTICAL) reduce 536) ((T_IS_SMALLER_OR_EQUAL) reduce 536) ((T_LOGICAL_AND) reduce 536) ((T_LOGICAL_OR) reduce 536) ((T_LOGICAL_XOR) reduce 536) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 536) ((T_OBJECT_OPERATOR) reduce 536) ((T_PIPE) reduce 536) ((T_POW) reduce 536) ((T_SL) reduce 536) ((T_SPACESHIP) reduce 536) ((T_SR) reduce 536) (("[") reduce 536) (("^") reduce 536) (("|") reduce 536)) 19486 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AS) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("|") reduce 94)) 19487 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AS) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("|") reduce 93)) 19488 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AS) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("|") reduce 95)) 19490 ((("%") reduce 539) (("*") reduce 539) (("+") reduce 539) (("-") reduce 539) ((".") reduce 539) (("/") reduce 539) (("<") reduce 539) ((">") reduce 539) (("?") reduce 539) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_AS) reduce 539) ((T_BOOLEAN_AND) reduce 539) ((T_BOOLEAN_OR) reduce 539) ((T_COALESCE) reduce 539) ((T_INSTANCEOF) reduce 539) ((T_IS_EQUAL) reduce 539) ((T_IS_GREATER_OR_EQUAL) reduce 539) ((T_IS_IDENTICAL) reduce 539) ((T_IS_NOT_EQUAL) reduce 539) ((T_IS_NOT_IDENTICAL) reduce 539) ((T_IS_SMALLER_OR_EQUAL) reduce 539) ((T_LOGICAL_AND) reduce 539) ((T_LOGICAL_OR) reduce 539) ((T_LOGICAL_XOR) reduce 539) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 539) ((T_OBJECT_OPERATOR) reduce 539) ((T_PIPE) reduce 539) ((T_POW) reduce 539) ((T_SL) reduce 539) ((T_SPACESHIP) reduce 539) ((T_SR) reduce 539) (("[") reduce 539) (("^") reduce 539) (("|") reduce 539)) 19493 ((("%") reduce 538) (("*") reduce 538) (("+") reduce 538) (("-") reduce 538) ((".") reduce 538) (("/") reduce 538) (("<") reduce 538) ((">") reduce 538) (("?") reduce 538) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_AS) reduce 538) ((T_BOOLEAN_AND) reduce 538) ((T_BOOLEAN_OR) reduce 538) ((T_COALESCE) reduce 538) ((T_INSTANCEOF) reduce 538) ((T_IS_EQUAL) reduce 538) ((T_IS_GREATER_OR_EQUAL) reduce 538) ((T_IS_IDENTICAL) reduce 538) ((T_IS_NOT_EQUAL) reduce 538) ((T_IS_NOT_IDENTICAL) reduce 538) ((T_IS_SMALLER_OR_EQUAL) reduce 538) ((T_LOGICAL_AND) reduce 538) ((T_LOGICAL_OR) reduce 538) ((T_LOGICAL_XOR) reduce 538) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 538) ((T_OBJECT_OPERATOR) reduce 538) ((T_PIPE) reduce 538) ((T_POW) reduce 538) ((T_SL) reduce 538) ((T_SPACESHIP) reduce 538) ((T_SR) reduce 538) (("[") reduce 538) (("^") reduce 538) (("|") reduce 538)) 19499 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AS) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92)) 19502 ((("%") reduce 535) (("*") reduce 535) (("+") reduce 535) (("-") reduce 535) ((".") reduce 535) (("/") reduce 535) (("<") reduce 535) ((">") reduce 535) (("?") reduce 535) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_AS) reduce 535) ((T_BOOLEAN_AND) reduce 535) ((T_BOOLEAN_OR) reduce 535) ((T_COALESCE) reduce 535) ((T_INSTANCEOF) reduce 535) ((T_IS_EQUAL) reduce 535) ((T_IS_GREATER_OR_EQUAL) reduce 535) ((T_IS_IDENTICAL) reduce 535) ((T_IS_NOT_EQUAL) reduce 535) ((T_IS_NOT_IDENTICAL) reduce 535) ((T_IS_SMALLER_OR_EQUAL) reduce 535) ((T_LOGICAL_AND) reduce 535) ((T_LOGICAL_OR) reduce 535) ((T_LOGICAL_XOR) reduce 535) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PIPE) reduce 535) ((T_POW) reduce 535) ((T_SL) reduce 535) ((T_SPACESHIP) reduce 535) ((T_SR) reduce 535) (("[") reduce 535) (("^") reduce 535) (("|") reduce 535)) 19504 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) (("=") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AND_EQUAL) reduce 568) ((T_AS) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_COALESCE_EQUAL) reduce 568) ((T_CONCAT_EQUAL) reduce 568) ((T_DEC) reduce 568) ((T_DIV_EQUAL) reduce 568) ((T_INC) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_MINUS_EQUAL) reduce 568) ((T_MOD_EQUAL) reduce 568) ((T_MUL_EQUAL) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_OR_EQUAL) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_PLUS_EQUAL) reduce 568) ((T_POW) reduce 568) ((T_POW_EQUAL) reduce 568) ((T_SL) reduce 568) ((T_SL_EQUAL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) ((T_SR_EQUAL) reduce 568) ((T_XOR_EQUAL) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 19505 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 479) (("(") shift) (("*") reduce 479) (("+") shift) (("-") shift) ((".") reduce 479) (("/") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_AS) reduce 479) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 479) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 479) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PIPE) reduce 479) ((T_POW) reduce 479) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 479) (("`") shift) (("|") reduce 479) (("~") shift)) 19512 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) (("<") reduce 564) (("=") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AND_EQUAL) reduce 564) ((T_AS) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_COALESCE_EQUAL) reduce 564) ((T_CONCAT_EQUAL) reduce 564) ((T_DEC) reduce 564) ((T_DIV_EQUAL) reduce 564) ((T_INC) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_MINUS_EQUAL) reduce 564) ((T_MOD_EQUAL) reduce 564) ((T_MUL_EQUAL) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_OR_EQUAL) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_PLUS_EQUAL) reduce 564) ((T_POW) reduce 564) ((T_POW_EQUAL) reduce 564) ((T_SL) reduce 564) ((T_SL_EQUAL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) ((T_SR_EQUAL) reduce 564) ((T_XOR_EQUAL) reduce 564) (("[") reduce 564) (("^") reduce 564) (("|") reduce 564)) 19513 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AS) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PIPE) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 551) (("^") reduce 530) (("|") reduce 530)) 19515 ((("%") reduce 529) (("*") reduce 529) (("+") reduce 529) (("-") reduce 529) ((".") reduce 529) (("/") reduce 529) (("<") reduce 529) ((">") reduce 529) (("?") reduce 529) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_AS) reduce 529) ((T_BOOLEAN_AND) reduce 529) ((T_BOOLEAN_OR) reduce 529) ((T_COALESCE) reduce 529) ((T_INSTANCEOF) reduce 529) ((T_IS_EQUAL) reduce 529) ((T_IS_GREATER_OR_EQUAL) reduce 529) ((T_IS_IDENTICAL) reduce 529) ((T_IS_NOT_EQUAL) reduce 529) ((T_IS_NOT_IDENTICAL) reduce 529) ((T_IS_SMALLER_OR_EQUAL) reduce 529) ((T_LOGICAL_AND) reduce 529) ((T_LOGICAL_OR) reduce 529) ((T_LOGICAL_XOR) reduce 529) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PIPE) reduce 529) ((T_POW) reduce 529) ((T_SL) reduce 529) ((T_SPACESHIP) reduce 529) ((T_SR) reduce 529) (("[") reduce 554) (("^") reduce 529) (("|") reduce 529)) 19516 ((("%") reduce 528) (("(") reduce 557) (("*") reduce 528) (("+") reduce 528) (("-") reduce 528) ((".") reduce 528) (("/") reduce 528) (("<") reduce 528) ((">") reduce 528) (("?") reduce 528) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_AS) reduce 528) ((T_BOOLEAN_AND) reduce 528) ((T_BOOLEAN_OR) reduce 528) ((T_COALESCE) reduce 528) ((T_INSTANCEOF) reduce 528) ((T_IS_EQUAL) reduce 528) ((T_IS_GREATER_OR_EQUAL) reduce 528) ((T_IS_IDENTICAL) reduce 528) ((T_IS_NOT_EQUAL) reduce 528) ((T_IS_NOT_IDENTICAL) reduce 528) ((T_IS_SMALLER_OR_EQUAL) reduce 528) ((T_LOGICAL_AND) reduce 528) ((T_LOGICAL_OR) reduce 528) ((T_LOGICAL_XOR) reduce 528) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PIPE) reduce 528) ((T_POW) reduce 528) ((T_SL) reduce 528) ((T_SPACESHIP) reduce 528) ((T_SR) reduce 528) (("[") reduce 550) (("^") reduce 528) (("|") reduce 528)) 19517 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19520 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) (("<") reduce 563) (("=") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AND_EQUAL) reduce 563) ((T_AS) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_COALESCE_EQUAL) reduce 563) ((T_CONCAT_EQUAL) reduce 563) ((T_DEC) reduce 563) ((T_DIV_EQUAL) reduce 563) ((T_INC) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_MINUS_EQUAL) reduce 563) ((T_MOD_EQUAL) reduce 563) ((T_MUL_EQUAL) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_OR_EQUAL) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_PLUS_EQUAL) reduce 563) ((T_POW) reduce 563) ((T_POW_EQUAL) reduce 563) ((T_SL) reduce 563) ((T_SL_EQUAL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) ((T_SR_EQUAL) reduce 563) ((T_XOR_EQUAL) reduce 563) (("[") reduce 563) (("^") reduce 563) (("|") reduce 563)) 19521 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AS) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_PIPE) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("^") reduce 484) (("|") reduce 484)) 19522 ((("%") reduce 466) (("*") reduce 466) (("+") reduce 466) (("-") reduce 466) ((".") reduce 466) (("/") reduce 466) (("<") reduce 466) ((">") reduce 466) (("?") reduce 466) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_AS) reduce 466) ((T_BOOLEAN_AND) reduce 466) ((T_BOOLEAN_OR) reduce 466) ((T_COALESCE) reduce 466) ((T_INSTANCEOF) reduce 466) ((T_IS_EQUAL) reduce 466) ((T_IS_GREATER_OR_EQUAL) reduce 466) ((T_IS_IDENTICAL) reduce 466) ((T_IS_NOT_EQUAL) reduce 466) ((T_IS_NOT_IDENTICAL) reduce 466) ((T_IS_SMALLER_OR_EQUAL) reduce 466) ((T_LOGICAL_AND) reduce 466) ((T_LOGICAL_OR) reduce 466) ((T_LOGICAL_XOR) reduce 466) ((T_PIPE) reduce 466) ((T_POW) reduce 466) ((T_SL) reduce 466) ((T_SPACESHIP) reduce 466) ((T_SR) reduce 466) (("^") reduce 466) (("|") reduce 466)) 19523 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AS) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_PIPE) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("^") reduce 488) (("|") reduce 488)) 19524 ((("%") reduce 531) (("(") shift) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AS) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("|") reduce 531)) 19525 ((("%") reduce 461) (("(") reduce 558) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AS) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PIPE) reduce 461) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") reduce 552) (("^") reduce 461) (("|") reduce 461)) 19526 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AS) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_PIPE) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 19527 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AS) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_PIPE) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("|") reduce 476)) 19528 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) (("=") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AND_EQUAL) reduce 559) ((T_AS) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_COALESCE_EQUAL) reduce 559) ((T_CONCAT_EQUAL) reduce 559) ((T_DEC) reduce 559) ((T_DIV_EQUAL) reduce 559) ((T_INC) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_MINUS_EQUAL) reduce 559) ((T_MOD_EQUAL) reduce 559) ((T_MUL_EQUAL) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_OR_EQUAL) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_PLUS_EQUAL) reduce 559) ((T_POW) reduce 559) ((T_POW_EQUAL) reduce 559) ((T_SL) reduce 559) ((T_SL_EQUAL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) ((T_SR_EQUAL) reduce 559) ((T_XOR_EQUAL) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 19529 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_AS) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 19530 ((("%") reduce 403) (("*") reduce 403) (("+") reduce 403) (("-") reduce 403) ((".") reduce 403) (("/") reduce 403) (("<") reduce 403) (("=") shift) ((">") reduce 403) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AND_EQUAL) shift) ((T_AS) reduce 403) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_INC) shift) ((T_INSTANCEOF) reduce 403) ((T_IS_EQUAL) reduce 403) ((T_IS_GREATER_OR_EQUAL) reduce 403) ((T_IS_IDENTICAL) reduce 403) ((T_IS_NOT_EQUAL) reduce 403) ((T_IS_NOT_IDENTICAL) reduce 403) ((T_IS_SMALLER_OR_EQUAL) reduce 403) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 403) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 403) ((T_POW_EQUAL) shift) ((T_SL) reduce 403) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 403) ((T_SR) reduce 403) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 548) (("^") reduce 403) (("|") reduce 403)) 19533 ((("%") reduce 448) (("*") reduce 448) (("+") reduce 448) (("-") reduce 448) ((".") reduce 448) (("/") reduce 448) (("<") reduce 448) ((">") reduce 448) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_AS) reduce 448) ((T_BOOLEAN_AND) reduce 448) ((T_BOOLEAN_OR) reduce 448) ((T_COALESCE) reduce 448) ((T_INSTANCEOF) reduce 448) ((T_IS_EQUAL) reduce 448) ((T_IS_GREATER_OR_EQUAL) reduce 448) ((T_IS_IDENTICAL) reduce 448) ((T_IS_NOT_EQUAL) reduce 448) ((T_IS_NOT_IDENTICAL) reduce 448) ((T_IS_SMALLER_OR_EQUAL) reduce 448) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_PIPE) reduce 448) ((T_POW) shift) ((T_SL) reduce 448) ((T_SPACESHIP) reduce 448) ((T_SR) reduce 448) (("^") reduce 448) (("|") reduce 448)) 19565 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 432) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 432)) 19566 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("|") reduce 435)) 19567 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 444) (("/") shift) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AS) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_PIPE) reduce 444) ((T_POW) shift) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("|") reduce 444)) 19568 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AS) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 458) (("|") reduce 458)) 19569 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 443) (("/") shift) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AS) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_PIPE) reduce 443) ((T_POW) shift) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("^") reduce 443) (("|") reduce 443)) 19570 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AS) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_PIPE) reduce 440) ((T_POW) shift) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("|") reduce 440)) 19571 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AS) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_PIPE) reduce 453) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 453) ((T_SR) shift) (("^") reduce 453) (("|") reduce 453)) 19572 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 431) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19573 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 429) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19574 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 430) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19575 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AS) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 455) ((T_SR) shift) (("^") reduce 455) (("|") reduce 455)) 19576 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AS) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 450) (("|") reduce 450)) 19577 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AS) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 452) (("|") reduce 452)) 19578 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AS) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 449) (("|") reduce 449)) 19579 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AS) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 457) ((T_SR) shift) (("^") reduce 457) (("|") reduce 457)) 19580 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AS) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 451) (("|") reduce 451)) 19583 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AS) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 19584 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AS) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 19585 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AS) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 19586 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AS) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 19587 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AS) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 19588 ((("%") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AS) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 19589 ((("%") reduce 511) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AS) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("|") reduce 511)) 19590 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AS) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PIPE) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 19591 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AS) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 19592 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AS) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("|") reduce 512)) 19593 ((("%") reduce 573) (("*") reduce 573) (("+") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AS) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("|") reduce 573)) 19599 ((("%") reduce 574) (("*") reduce 574) (("+") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AS) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("|") reduce 574)) 19600 ((("%") reduce 578) (("*") reduce 578) (("+") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AS) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("|") reduce 578)) 19601 ((("%") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AS) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 19602 ((("%") reduce 575) (("*") reduce 575) (("+") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AS) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("|") reduce 575)) 19603 ((("%") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AS) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 19606 ((("%") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AS) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 19607 ((("%") reduce 576) (("*") reduce 576) (("+") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AS) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("|") reduce 576)) 19609 ((("%") reduce 577) (("*") reduce 577) (("+") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AS) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("|") reduce 577)) 19611 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AS) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("|") reduce 513)) 19612 ((("%") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AS) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 19615 ((("%") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AS) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 19616 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 465) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 465) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19617 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 427) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19618 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 428) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19619 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AS) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("|") reduce 433)) 19620 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AS) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("|") reduce 434)) 19624 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 463) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19625 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 464) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19626 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AS) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 456) ((T_SR) shift) (("^") reduce 456) (("|") reduce 456)) 19627 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AS) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 454) ((T_SR) shift) (("^") reduce 454) (("|") reduce 454)) 19628 ((("%") reduce 441) (("*") reduce 441) (("+") reduce 441) (("-") reduce 441) ((".") reduce 441) (("/") reduce 441) (("<") reduce 441) ((">") reduce 441) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AS) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 441) ((T_IS_GREATER_OR_EQUAL) reduce 441) ((T_IS_IDENTICAL) reduce 441) ((T_IS_NOT_EQUAL) reduce 441) ((T_IS_NOT_IDENTICAL) reduce 441) ((T_IS_SMALLER_OR_EQUAL) reduce 441) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_PIPE) reduce 441) ((T_POW) shift) ((T_SL) reduce 441) ((T_SPACESHIP) reduce 441) ((T_SR) reduce 441) (("^") reduce 441) (("|") reduce 441)) 19629 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 436) (("/") shift) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AS) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_PIPE) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 436) ((T_SR) shift) (("^") reduce 436) (("|") reduce 436)) 19630 ((("%") shift) (("*") shift) (("+") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") shift) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AS) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_PIPE) reduce 438) ((T_POW) shift) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("|") reduce 438)) 19631 ((("%") shift) (("*") shift) (("+") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") shift) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AS) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_PIPE) reduce 437) ((T_POW) shift) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("|") reduce 437)) 19632 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AS) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_PIPE) reduce 439) ((T_POW) shift) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("|") reduce 439)) 19633 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AS) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_PIPE) reduce 442) ((T_POW) shift) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("|") reduce 442)) 19635 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AS) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_PIPE) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("|") reduce 74)) 19636 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AS) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_PIPE) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("|") reduce 52)) 19637 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AS) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_PIPE) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("|") reduce 27)) 19638 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AS) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_PIPE) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("|") reduce 51)) 19639 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AS) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_PIPE) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("|") reduce 53)) 19640 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AS) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_PIPE) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("|") reduce 49)) 19641 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AS) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_PIPE) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("|") reduce 29)) 19642 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AS) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_PIPE) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("|") reduce 59)) 19643 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AS) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_PIPE) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("|") reduce 60)) 19644 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AS) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_PIPE) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("|") reduce 11)) 19645 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AS) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_PIPE) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("|") reduce 42)) 19646 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AS) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_PIPE) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("|") reduce 39)) 19647 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AS) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_PIPE) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("|") reduce 25)) 19648 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AS) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_PIPE) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("|") reduce 50)) 19649 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AS) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_PIPE) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("|") reduce 66)) 19650 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AS) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_PIPE) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("|") reduce 18)) 19651 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AS) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_PIPE) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("|") reduce 17)) 19652 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AS) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_PIPE) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("|") reduce 15)) 19653 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AS) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_PIPE) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("|") reduce 14)) 19654 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AS) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_PIPE) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("|") reduce 38)) 19655 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AS) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_PIPE) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("|") reduce 26)) 19656 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AS) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_PIPE) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("|") reduce 22)) 19657 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AS) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_PIPE) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("|") reduce 24)) 19658 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AS) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_PIPE) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("|") reduce 16)) 19659 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AS) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_PIPE) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("|") reduce 48)) 19660 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AS) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_PIPE) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("|") reduce 20)) 19661 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AS) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_PIPE) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("|") reduce 70)) 19662 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AS) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_PIPE) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("|") reduce 3)) 19663 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AS) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_PIPE) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("|") reduce 12)) 19664 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AS) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_PIPE) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("|") reduce 54)) 19665 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AS) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_PIPE) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("|") reduce 65)) 19666 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AS) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_PIPE) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("|") reduce 75)) 19667 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AS) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_PIPE) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("|") reduce 30)) 19668 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AS) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_PIPE) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("|") reduce 68)) 19669 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AS) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_PIPE) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("|") reduce 21)) 19670 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AS) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_PIPE) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("|") reduce 23)) 19671 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AS) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_PIPE) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("|") reduce 41)) 19672 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AS) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_PIPE) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("|") reduce 62)) 19673 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AS) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_PIPE) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("|") reduce 34)) 19674 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AS) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_PIPE) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("|") reduce 40)) 19675 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AS) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_PIPE) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("|") reduce 13)) 19676 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AS) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_PIPE) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("|") reduce 55)) 19677 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AS) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_PIPE) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("|") reduce 1)) 19678 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AS) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_PIPE) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("|") reduce 2)) 19679 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AS) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_PIPE) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("|") reduce 9)) 19680 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AS) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_PIPE) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("|") reduce 33)) 19681 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AS) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_PIPE) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("|") reduce 58)) 19682 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AS) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_PIPE) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("|") reduce 37)) 19683 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AS) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_PIPE) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("|") reduce 64)) 19684 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AS) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_PIPE) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("|") reduce 46)) 19685 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AS) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_PIPE) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("|") reduce 8)) 19686 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AS) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_PIPE) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("|") reduce 6)) 19687 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AS) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_PIPE) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("|") reduce 7)) 19688 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AS) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_PIPE) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("|") reduce 69)) 19689 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AS) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_PIPE) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("|") reduce 63)) 19690 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AS) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_PIPE) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("|") reduce 56)) 19691 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AS) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_PIPE) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("|") reduce 10)) 19692 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AS) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_PIPE) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("|") reduce 67)) 19693 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AS) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_PIPE) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("|") reduce 44)) 19694 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AS) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_PIPE) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("|") reduce 76)) 19695 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AS) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_PIPE) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("|") reduce 71)) 19696 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AS) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_PIPE) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("|") reduce 77)) 19697 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AS) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_PIPE) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("|") reduce 78)) 19698 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AS) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_PIPE) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("|") reduce 79)) 19699 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AS) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_PIPE) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("|") reduce 4)) 19700 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AS) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_PIPE) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("|") reduce 5)) 19701 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AS) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_PIPE) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("|") reduce 43)) 19702 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AS) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_PIPE) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("|") reduce 73)) 19703 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AS) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_PIPE) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("|") reduce 82)) 19704 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AS) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_PIPE) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("|") reduce 47)) 19705 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AS) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_PIPE) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("|") reduce 31)) 19706 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AS) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_PIPE) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("|") reduce 57)) 19707 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AS) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_PIPE) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("|") reduce 61)) 19708 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AS) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_PIPE) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("|") reduce 28)) 19709 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AS) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_PIPE) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("|") reduce 36)) 19710 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AS) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_PIPE) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("|") reduce 32)) 19711 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AS) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_PIPE) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("|") reduce 35)) 19712 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AS) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_PIPE) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("|") reduce 19)) 19713 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AS) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_PIPE) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("|") reduce 45)) 19714 ((("%") reduce 542) (("(") reduce 579) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AS) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PIPE) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 19716 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AS) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_PIPE) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("|") reduce 72)) 19717 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AS) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_PIPE) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("|") reduce 83)) 19718 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) (("<") reduce 572) (("=") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AND_EQUAL) reduce 572) ((T_AS) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_COALESCE_EQUAL) reduce 572) ((T_CONCAT_EQUAL) reduce 572) ((T_DEC) reduce 572) ((T_DIV_EQUAL) reduce 572) ((T_INC) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_MINUS_EQUAL) reduce 572) ((T_MOD_EQUAL) reduce 572) ((T_MUL_EQUAL) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_OR_EQUAL) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_PLUS_EQUAL) reduce 572) ((T_POW) reduce 572) ((T_POW_EQUAL) reduce 572) ((T_SL) reduce 572) ((T_SL_EQUAL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) ((T_SR_EQUAL) reduce 572) ((T_XOR_EQUAL) reduce 572) (("[") reduce 572) (("^") reduce 572) (("|") reduce 572)) 19721 ((("%") reduce 544) (("(") reduce 580) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AS) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PIPE) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 19723 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_AS) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("|") reduce 507)) 19724 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_AS) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 19728 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) (("=") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AND_EQUAL) reduce 292) ((T_AS) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_COALESCE_EQUAL) reduce 292) ((T_CONCAT_EQUAL) reduce 292) ((T_DEC) reduce 292) ((T_DIV_EQUAL) reduce 292) ((T_INC) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_MINUS_EQUAL) reduce 292) ((T_MOD_EQUAL) reduce 292) ((T_MUL_EQUAL) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_OR_EQUAL) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_PLUS_EQUAL) reduce 292) ((T_POW) reduce 292) ((T_POW_EQUAL) reduce 292) ((T_SL) reduce 292) ((T_SL_EQUAL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) ((T_SR_EQUAL) reduce 292) ((T_XOR_EQUAL) reduce 292) (("[") reduce 292) (("^") reduce 292) (("|") reduce 292)) 19729 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) (("=") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AND_EQUAL) reduce 293) ((T_AS) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_COALESCE_EQUAL) reduce 293) ((T_CONCAT_EQUAL) reduce 293) ((T_DEC) reduce 293) ((T_DIV_EQUAL) reduce 293) ((T_INC) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_MINUS_EQUAL) reduce 293) ((T_MOD_EQUAL) reduce 293) ((T_MUL_EQUAL) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_OR_EQUAL) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_PLUS_EQUAL) reduce 293) ((T_POW) reduce 293) ((T_POW_EQUAL) reduce 293) ((T_SL) reduce 293) ((T_SL_EQUAL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) ((T_SR_EQUAL) reduce 293) ((T_XOR_EQUAL) reduce 293) (("[") reduce 293) (("^") reduce 293) (("|") reduce 293)) 19734 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AS) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_PIPE) reduce 425) ((T_POW) reduce 425) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 19736 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AS) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_PIPE) reduce 423) ((T_POW) reduce 423) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 19746 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 419) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19747 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 421) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19748 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 420) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19749 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 413) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19750 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 410) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19751 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 418) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19752 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 412) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19753 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 416) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19754 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 411) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19755 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 414) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19756 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 415) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19757 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 422) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19758 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 417) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19760 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 406) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 19763 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AS) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 19766 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) (("<") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AS) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_POW) reduce 564) ((T_SL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) (("[") reduce 564) (("^") reduce 564) (("|") reduce 564)) 19768 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) (("<") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AS) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_POW) reduce 563) ((T_SL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) (("[") reduce 563) (("^") reduce 563) (("|") reduce 563)) 19770 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AS) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 19771 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AS) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 19772 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AS) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("[") reduce 548) (("^") reduce 407) (("|") reduce 407)) 19776 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) (("<") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AS) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_POW) reduce 572) ((T_SL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) (("[") reduce 572) (("^") reduce 572) (("|") reduce 572)) 19778 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AS) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("|") reduce 507)) 19779 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AS) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 19783 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AS) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("[") reduce 292) (("^") reduce 292) (("|") reduce 292)) 19784 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AS) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("[") reduce 293) (("^") reduce 293) (("|") reduce 293)) 19785 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AS) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 19788 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) (("<") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AS) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_POW) reduce 571) ((T_SL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) (("[") reduce 571) (("^") reduce 571) (("|") reduce 571)) 19789 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AS) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("|") reduce 506)) 19790 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AS) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("|") reduce 508)) 19795 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AS) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 19796 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AS) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 19797 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AS) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 19798 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AS) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 19801 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AS) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 19802 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AS) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 19803 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AS) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 19804 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AS) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("[") reduce 562) (("^") reduce 562) (("|") reduce 562)) 19805 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AS) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 19806 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AS) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 19809 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AS) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 19810 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) (("=") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AND_EQUAL) reduce 504) ((T_AS) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_COALESCE_EQUAL) reduce 504) ((T_CONCAT_EQUAL) reduce 504) ((T_DEC) reduce 504) ((T_DIV_EQUAL) reduce 504) ((T_INC) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_MINUS_EQUAL) reduce 504) ((T_MOD_EQUAL) reduce 504) ((T_MUL_EQUAL) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_OR_EQUAL) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_PLUS_EQUAL) reduce 504) ((T_POW) reduce 504) ((T_POW_EQUAL) reduce 504) ((T_SL) reduce 504) ((T_SL_EQUAL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) ((T_SR_EQUAL) reduce 504) ((T_XOR_EQUAL) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 19821 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AS) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PIPE) reduce 494) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 19822 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AS) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_PIPE) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("^") reduce 489) (("|") reduce 489)) 19895 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 494) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20211 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AS) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_PIPE) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("^") reduce 490) (("|") reduce 490)) 20214 ((("%") reduce 541) (("(") reduce 579) (("*") reduce 541) (("+") reduce 541) (("-") reduce 541) ((".") reduce 541) (("/") reduce 541) (("<") reduce 541) ((">") reduce 541) (("?") reduce 541) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_AS) reduce 541) ((T_BOOLEAN_AND) reduce 541) ((T_BOOLEAN_OR) reduce 541) ((T_COALESCE) reduce 541) ((T_INSTANCEOF) reduce 541) ((T_IS_EQUAL) reduce 541) ((T_IS_GREATER_OR_EQUAL) reduce 541) ((T_IS_IDENTICAL) reduce 541) ((T_IS_NOT_EQUAL) reduce 541) ((T_IS_NOT_IDENTICAL) reduce 541) ((T_IS_SMALLER_OR_EQUAL) reduce 541) ((T_LOGICAL_AND) reduce 541) ((T_LOGICAL_OR) reduce 541) ((T_LOGICAL_XOR) reduce 541) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 541) ((T_OBJECT_OPERATOR) reduce 541) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 541) ((T_PIPE) reduce 541) ((T_POW) reduce 541) ((T_SL) reduce 541) ((T_SPACESHIP) reduce 541) ((T_SR) reduce 541) (("[") reduce 541) (("^") reduce 541) (("|") reduce 541)) 20216 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) (("<") reduce 571) (("=") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AND_EQUAL) reduce 571) ((T_AS) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_COALESCE_EQUAL) reduce 571) ((T_CONCAT_EQUAL) reduce 571) ((T_DEC) reduce 571) ((T_DIV_EQUAL) reduce 571) ((T_INC) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_MINUS_EQUAL) reduce 571) ((T_MOD_EQUAL) reduce 571) ((T_MUL_EQUAL) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_OR_EQUAL) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_PLUS_EQUAL) reduce 571) ((T_POW) reduce 571) ((T_POW_EQUAL) reduce 571) ((T_SL) reduce 571) ((T_SL_EQUAL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) ((T_SR_EQUAL) reduce 571) ((T_XOR_EQUAL) reduce 571) (("[") reduce 571) (("^") reduce 571) (("|") reduce 571)) 20219 ((("%") reduce 543) (("(") reduce 580) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AS) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PIPE) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 20220 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_AS) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("|") reduce 506)) 20221 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_AS) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("|") reduce 508)) 20223 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AS) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_PIPE) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("^") reduce 485) (("|") reduce 485)) 20224 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AS) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_PIPE) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("^") reduce 487) (("|") reduce 487)) 20229 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) (("=") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AND_EQUAL) reduce 560) ((T_AS) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_COALESCE_EQUAL) reduce 560) ((T_CONCAT_EQUAL) reduce 560) ((T_DEC) reduce 560) ((T_DIV_EQUAL) reduce 560) ((T_INC) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_MINUS_EQUAL) reduce 560) ((T_MOD_EQUAL) reduce 560) ((T_MUL_EQUAL) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_OR_EQUAL) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_PLUS_EQUAL) reduce 560) ((T_POW) reduce 560) ((T_POW_EQUAL) reduce 560) ((T_SL) reduce 560) ((T_SL_EQUAL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) ((T_SR_EQUAL) reduce 560) ((T_XOR_EQUAL) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 20230 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) (("=") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AND_EQUAL) reduce 582) ((T_AS) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_COALESCE_EQUAL) reduce 582) ((T_CONCAT_EQUAL) reduce 582) ((T_DEC) reduce 582) ((T_DIV_EQUAL) reduce 582) ((T_INC) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_MINUS_EQUAL) reduce 582) ((T_MOD_EQUAL) reduce 582) ((T_MUL_EQUAL) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_OR_EQUAL) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_PLUS_EQUAL) reduce 582) ((T_POW) reduce 582) ((T_POW_EQUAL) reduce 582) ((T_SL) reduce 582) ((T_SL_EQUAL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) ((T_SR_EQUAL) reduce 582) ((T_XOR_EQUAL) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 20231 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_AS) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 20232 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) (("=") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AND_EQUAL) reduce 584) ((T_AS) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_COALESCE_EQUAL) reduce 584) ((T_CONCAT_EQUAL) reduce 584) ((T_DEC) reduce 584) ((T_DIV_EQUAL) reduce 584) ((T_INC) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_MINUS_EQUAL) reduce 584) ((T_MOD_EQUAL) reduce 584) ((T_MUL_EQUAL) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_OR_EQUAL) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_PLUS_EQUAL) reduce 584) ((T_POW) reduce 584) ((T_POW_EQUAL) reduce 584) ((T_SL) reduce 584) ((T_SL_EQUAL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) ((T_SR_EQUAL) reduce 584) ((T_XOR_EQUAL) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 20235 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) (("=") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AND_EQUAL) reduce 583) ((T_AS) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_COALESCE_EQUAL) reduce 583) ((T_CONCAT_EQUAL) reduce 583) ((T_DEC) reduce 583) ((T_DIV_EQUAL) reduce 583) ((T_INC) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_MINUS_EQUAL) reduce 583) ((T_MOD_EQUAL) reduce 583) ((T_MUL_EQUAL) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_OR_EQUAL) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_PLUS_EQUAL) reduce 583) ((T_POW) reduce 583) ((T_POW_EQUAL) reduce 583) ((T_SL) reduce 583) ((T_SL_EQUAL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) ((T_SR_EQUAL) reduce 583) ((T_XOR_EQUAL) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 20236 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) (("=") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AND_EQUAL) reduce 561) ((T_AS) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_COALESCE_EQUAL) reduce 561) ((T_CONCAT_EQUAL) reduce 561) ((T_DEC) reduce 561) ((T_DIV_EQUAL) reduce 561) ((T_INC) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_MINUS_EQUAL) reduce 561) ((T_MOD_EQUAL) reduce 561) ((T_MUL_EQUAL) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_OR_EQUAL) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_PLUS_EQUAL) reduce 561) ((T_POW) reduce 561) ((T_POW_EQUAL) reduce 561) ((T_SL) reduce 561) ((T_SL_EQUAL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) ((T_SR_EQUAL) reduce 561) ((T_XOR_EQUAL) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 20237 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_AS) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 20238 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) (("=") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AND_EQUAL) reduce 562) ((T_AS) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_COALESCE_EQUAL) reduce 562) ((T_CONCAT_EQUAL) reduce 562) ((T_DEC) reduce 562) ((T_DIV_EQUAL) reduce 562) ((T_INC) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_MINUS_EQUAL) reduce 562) ((T_MOD_EQUAL) reduce 562) ((T_MUL_EQUAL) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_OR_EQUAL) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_PLUS_EQUAL) reduce 562) ((T_POW) reduce 562) ((T_POW_EQUAL) reduce 562) ((T_SL) reduce 562) ((T_SL_EQUAL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) ((T_SR_EQUAL) reduce 562) ((T_XOR_EQUAL) reduce 562) (("[") reduce 562) (("^") reduce 562) (("|") reduce 562)) 20240 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AS) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_PIPE) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 20242 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) (("=") shift) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AS) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_PIPE) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("|") reduce 520)) 20244 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 405) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20245 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 482) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20256 ((("%") reduce 540) (("*") reduce 540) (("+") reduce 540) (("-") reduce 540) ((".") reduce 540) (("/") reduce 540) (("<") reduce 540) ((">") reduce 540) (("?") reduce 540) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_AS) reduce 540) ((T_BOOLEAN_AND) reduce 540) ((T_BOOLEAN_OR) reduce 540) ((T_COALESCE) reduce 540) ((T_DOUBLE_ARROW) reduce 540) ((T_INSTANCEOF) reduce 540) ((T_IS_EQUAL) reduce 540) ((T_IS_GREATER_OR_EQUAL) reduce 540) ((T_IS_IDENTICAL) reduce 540) ((T_IS_NOT_EQUAL) reduce 540) ((T_IS_NOT_IDENTICAL) reduce 540) ((T_IS_SMALLER_OR_EQUAL) reduce 540) ((T_LOGICAL_AND) reduce 540) ((T_LOGICAL_OR) reduce 540) ((T_LOGICAL_XOR) reduce 540) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 540) ((T_OBJECT_OPERATOR) reduce 540) ((T_PIPE) reduce 540) ((T_POW) reduce 540) ((T_SL) reduce 540) ((T_SPACESHIP) reduce 540) ((T_SR) reduce 540) (("[") reduce 540) (("^") reduce 540) (("|") reduce 540)) 20258 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AS) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_PIPE) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("|") reduce 521)) 20260 ((("%") reduce 534) (("*") reduce 534) (("+") reduce 534) (("-") reduce 534) ((".") reduce 534) (("/") reduce 534) (("<") reduce 534) ((">") reduce 534) (("?") reduce 534) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_AS) reduce 534) ((T_BOOLEAN_AND) reduce 534) ((T_BOOLEAN_OR) reduce 534) ((T_COALESCE) reduce 534) ((T_DOUBLE_ARROW) reduce 534) ((T_INSTANCEOF) reduce 534) ((T_IS_EQUAL) reduce 534) ((T_IS_GREATER_OR_EQUAL) reduce 534) ((T_IS_IDENTICAL) reduce 534) ((T_IS_NOT_EQUAL) reduce 534) ((T_IS_NOT_IDENTICAL) reduce 534) ((T_IS_SMALLER_OR_EQUAL) reduce 534) ((T_LOGICAL_AND) reduce 534) ((T_LOGICAL_OR) reduce 534) ((T_LOGICAL_XOR) reduce 534) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PIPE) reduce 534) ((T_POW) reduce 534) ((T_SL) reduce 534) ((T_SPACESHIP) reduce 534) ((T_SR) reduce 534) (("[") reduce 534) (("^") reduce 534) (("|") reduce 534)) 20261 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AS) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_PIPE) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("^") reduce 524) (("|") reduce 524)) 20265 ((("%") reduce 517) (("(") shift) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AS) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_PIPE) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("^") reduce 517) (("|") reduce 517)) 20266 ((("%") reduce 533) (("*") reduce 533) (("+") reduce 533) (("-") reduce 533) ((".") reduce 533) (("/") reduce 533) (("<") reduce 533) ((">") reduce 533) (("?") reduce 533) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_AS) reduce 533) ((T_BOOLEAN_AND) reduce 533) ((T_BOOLEAN_OR) reduce 533) ((T_COALESCE) reduce 533) ((T_DOUBLE_ARROW) reduce 533) ((T_INSTANCEOF) reduce 533) ((T_IS_EQUAL) reduce 533) ((T_IS_GREATER_OR_EQUAL) reduce 533) ((T_IS_IDENTICAL) reduce 533) ((T_IS_NOT_EQUAL) reduce 533) ((T_IS_NOT_IDENTICAL) reduce 533) ((T_IS_SMALLER_OR_EQUAL) reduce 533) ((T_LOGICAL_AND) reduce 533) ((T_LOGICAL_OR) reduce 533) ((T_LOGICAL_XOR) reduce 533) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PIPE) reduce 533) ((T_POW) reduce 533) ((T_SL) reduce 533) ((T_SPACESHIP) reduce 533) ((T_SR) reduce 533) (("[") reduce 533) (("^") reduce 533) (("|") reduce 533)) 20267 ((("%") reduce 537) (("*") reduce 537) (("+") reduce 537) (("-") reduce 537) ((".") reduce 537) (("/") reduce 537) (("<") reduce 537) ((">") reduce 537) (("?") reduce 537) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_AS) reduce 537) ((T_BOOLEAN_AND) reduce 537) ((T_BOOLEAN_OR) reduce 537) ((T_COALESCE) reduce 537) ((T_DOUBLE_ARROW) reduce 537) ((T_INSTANCEOF) reduce 537) ((T_IS_EQUAL) reduce 537) ((T_IS_GREATER_OR_EQUAL) reduce 537) ((T_IS_IDENTICAL) reduce 537) ((T_IS_NOT_EQUAL) reduce 537) ((T_IS_NOT_IDENTICAL) reduce 537) ((T_IS_SMALLER_OR_EQUAL) reduce 537) ((T_LOGICAL_AND) reduce 537) ((T_LOGICAL_OR) reduce 537) ((T_LOGICAL_XOR) reduce 537) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_PIPE) reduce 537) ((T_POW) reduce 537) ((T_SL) reduce 537) ((T_SPACESHIP) reduce 537) ((T_SR) reduce 537) (("[") reduce 537) (("^") reduce 537) (("|") reduce 537)) 20273 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AS) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PIPE) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("|") reduce 532)) 20275 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AS) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_PIPE) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("^") reduce 523) (("|") reduce 523)) 20277 ((("%") reduce 536) (("*") reduce 536) (("+") reduce 536) (("-") reduce 536) ((".") reduce 536) (("/") reduce 536) (("<") reduce 536) ((">") reduce 536) (("?") reduce 536) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_AS) reduce 536) ((T_BOOLEAN_AND) reduce 536) ((T_BOOLEAN_OR) reduce 536) ((T_COALESCE) reduce 536) ((T_DOUBLE_ARROW) reduce 536) ((T_INSTANCEOF) reduce 536) ((T_IS_EQUAL) reduce 536) ((T_IS_GREATER_OR_EQUAL) reduce 536) ((T_IS_IDENTICAL) reduce 536) ((T_IS_NOT_EQUAL) reduce 536) ((T_IS_NOT_IDENTICAL) reduce 536) ((T_IS_SMALLER_OR_EQUAL) reduce 536) ((T_LOGICAL_AND) reduce 536) ((T_LOGICAL_OR) reduce 536) ((T_LOGICAL_XOR) reduce 536) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 536) ((T_OBJECT_OPERATOR) reduce 536) ((T_PIPE) reduce 536) ((T_POW) reduce 536) ((T_SL) reduce 536) ((T_SPACESHIP) reduce 536) ((T_SR) reduce 536) (("[") reduce 536) (("^") reduce 536) (("|") reduce 536)) 20278 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AS) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("|") reduce 94)) 20279 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AS) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("|") reduce 93)) 20280 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AS) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("|") reduce 95)) 20282 ((("%") reduce 539) (("*") reduce 539) (("+") reduce 539) (("-") reduce 539) ((".") reduce 539) (("/") reduce 539) (("<") reduce 539) ((">") reduce 539) (("?") reduce 539) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_AS) reduce 539) ((T_BOOLEAN_AND) reduce 539) ((T_BOOLEAN_OR) reduce 539) ((T_COALESCE) reduce 539) ((T_DOUBLE_ARROW) reduce 539) ((T_INSTANCEOF) reduce 539) ((T_IS_EQUAL) reduce 539) ((T_IS_GREATER_OR_EQUAL) reduce 539) ((T_IS_IDENTICAL) reduce 539) ((T_IS_NOT_EQUAL) reduce 539) ((T_IS_NOT_IDENTICAL) reduce 539) ((T_IS_SMALLER_OR_EQUAL) reduce 539) ((T_LOGICAL_AND) reduce 539) ((T_LOGICAL_OR) reduce 539) ((T_LOGICAL_XOR) reduce 539) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 539) ((T_OBJECT_OPERATOR) reduce 539) ((T_PIPE) reduce 539) ((T_POW) reduce 539) ((T_SL) reduce 539) ((T_SPACESHIP) reduce 539) ((T_SR) reduce 539) (("[") reduce 539) (("^") reduce 539) (("|") reduce 539)) 20285 ((("%") reduce 538) (("*") reduce 538) (("+") reduce 538) (("-") reduce 538) ((".") reduce 538) (("/") reduce 538) (("<") reduce 538) ((">") reduce 538) (("?") reduce 538) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_AS) reduce 538) ((T_BOOLEAN_AND) reduce 538) ((T_BOOLEAN_OR) reduce 538) ((T_COALESCE) reduce 538) ((T_DOUBLE_ARROW) reduce 538) ((T_INSTANCEOF) reduce 538) ((T_IS_EQUAL) reduce 538) ((T_IS_GREATER_OR_EQUAL) reduce 538) ((T_IS_IDENTICAL) reduce 538) ((T_IS_NOT_EQUAL) reduce 538) ((T_IS_NOT_IDENTICAL) reduce 538) ((T_IS_SMALLER_OR_EQUAL) reduce 538) ((T_LOGICAL_AND) reduce 538) ((T_LOGICAL_OR) reduce 538) ((T_LOGICAL_XOR) reduce 538) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 538) ((T_OBJECT_OPERATOR) reduce 538) ((T_PIPE) reduce 538) ((T_POW) reduce 538) ((T_SL) reduce 538) ((T_SPACESHIP) reduce 538) ((T_SR) reduce 538) (("[") reduce 538) (("^") reduce 538) (("|") reduce 538)) 20291 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AS) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92)) 20294 ((("%") reduce 535) (("*") reduce 535) (("+") reduce 535) (("-") reduce 535) ((".") reduce 535) (("/") reduce 535) (("<") reduce 535) ((">") reduce 535) (("?") reduce 535) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_AS) reduce 535) ((T_BOOLEAN_AND) reduce 535) ((T_BOOLEAN_OR) reduce 535) ((T_COALESCE) reduce 535) ((T_DOUBLE_ARROW) reduce 535) ((T_INSTANCEOF) reduce 535) ((T_IS_EQUAL) reduce 535) ((T_IS_GREATER_OR_EQUAL) reduce 535) ((T_IS_IDENTICAL) reduce 535) ((T_IS_NOT_EQUAL) reduce 535) ((T_IS_NOT_IDENTICAL) reduce 535) ((T_IS_SMALLER_OR_EQUAL) reduce 535) ((T_LOGICAL_AND) reduce 535) ((T_LOGICAL_OR) reduce 535) ((T_LOGICAL_XOR) reduce 535) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PIPE) reduce 535) ((T_POW) reduce 535) ((T_SL) reduce 535) ((T_SPACESHIP) reduce 535) ((T_SR) reduce 535) (("[") reduce 535) (("^") reduce 535) (("|") reduce 535)) 20296 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) (("=") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AND_EQUAL) reduce 568) ((T_AS) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_COALESCE_EQUAL) reduce 568) ((T_CONCAT_EQUAL) reduce 568) ((T_DEC) reduce 568) ((T_DIV_EQUAL) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INC) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_MINUS_EQUAL) reduce 568) ((T_MOD_EQUAL) reduce 568) ((T_MUL_EQUAL) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_OR_EQUAL) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_PLUS_EQUAL) reduce 568) ((T_POW) reduce 568) ((T_POW_EQUAL) reduce 568) ((T_SL) reduce 568) ((T_SL_EQUAL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) ((T_SR_EQUAL) reduce 568) ((T_XOR_EQUAL) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 20297 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 479) (("(") shift) (("*") reduce 479) (("+") shift) (("-") shift) ((".") reduce 479) (("/") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_AS) reduce 479) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 479) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 479) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 479) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PIPE) reduce 479) ((T_POW) reduce 479) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 479) (("`") shift) (("|") reduce 479) (("~") shift)) 20304 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) (("<") reduce 564) (("=") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AND_EQUAL) reduce 564) ((T_AS) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_COALESCE_EQUAL) reduce 564) ((T_CONCAT_EQUAL) reduce 564) ((T_DEC) reduce 564) ((T_DIV_EQUAL) reduce 564) ((T_DOUBLE_ARROW) reduce 564) ((T_INC) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_MINUS_EQUAL) reduce 564) ((T_MOD_EQUAL) reduce 564) ((T_MUL_EQUAL) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_OR_EQUAL) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_PLUS_EQUAL) reduce 564) ((T_POW) reduce 564) ((T_POW_EQUAL) reduce 564) ((T_SL) reduce 564) ((T_SL_EQUAL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) ((T_SR_EQUAL) reduce 564) ((T_XOR_EQUAL) reduce 564) (("[") reduce 564) (("^") reduce 564) (("|") reduce 564)) 20305 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AS) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PIPE) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 551) (("^") reduce 530) (("|") reduce 530)) 20307 ((("%") reduce 529) (("*") reduce 529) (("+") reduce 529) (("-") reduce 529) ((".") reduce 529) (("/") reduce 529) (("<") reduce 529) ((">") reduce 529) (("?") reduce 529) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_AS) reduce 529) ((T_BOOLEAN_AND) reduce 529) ((T_BOOLEAN_OR) reduce 529) ((T_COALESCE) reduce 529) ((T_DOUBLE_ARROW) reduce 529) ((T_INSTANCEOF) reduce 529) ((T_IS_EQUAL) reduce 529) ((T_IS_GREATER_OR_EQUAL) reduce 529) ((T_IS_IDENTICAL) reduce 529) ((T_IS_NOT_EQUAL) reduce 529) ((T_IS_NOT_IDENTICAL) reduce 529) ((T_IS_SMALLER_OR_EQUAL) reduce 529) ((T_LOGICAL_AND) reduce 529) ((T_LOGICAL_OR) reduce 529) ((T_LOGICAL_XOR) reduce 529) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PIPE) reduce 529) ((T_POW) reduce 529) ((T_SL) reduce 529) ((T_SPACESHIP) reduce 529) ((T_SR) reduce 529) (("[") reduce 554) (("^") reduce 529) (("|") reduce 529)) 20308 ((("%") reduce 528) (("(") reduce 557) (("*") reduce 528) (("+") reduce 528) (("-") reduce 528) ((".") reduce 528) (("/") reduce 528) (("<") reduce 528) ((">") reduce 528) (("?") reduce 528) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_AS) reduce 528) ((T_BOOLEAN_AND) reduce 528) ((T_BOOLEAN_OR) reduce 528) ((T_COALESCE) reduce 528) ((T_DOUBLE_ARROW) reduce 528) ((T_INSTANCEOF) reduce 528) ((T_IS_EQUAL) reduce 528) ((T_IS_GREATER_OR_EQUAL) reduce 528) ((T_IS_IDENTICAL) reduce 528) ((T_IS_NOT_EQUAL) reduce 528) ((T_IS_NOT_IDENTICAL) reduce 528) ((T_IS_SMALLER_OR_EQUAL) reduce 528) ((T_LOGICAL_AND) reduce 528) ((T_LOGICAL_OR) reduce 528) ((T_LOGICAL_XOR) reduce 528) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PIPE) reduce 528) ((T_POW) reduce 528) ((T_SL) reduce 528) ((T_SPACESHIP) reduce 528) ((T_SR) reduce 528) (("[") reduce 550) (("^") reduce 528) (("|") reduce 528)) 20309 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 480) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 480) ((T_LOGICAL_OR) reduce 480) ((T_LOGICAL_XOR) reduce 480) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20312 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) (("<") reduce 563) (("=") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AND_EQUAL) reduce 563) ((T_AS) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_COALESCE_EQUAL) reduce 563) ((T_CONCAT_EQUAL) reduce 563) ((T_DEC) reduce 563) ((T_DIV_EQUAL) reduce 563) ((T_DOUBLE_ARROW) reduce 563) ((T_INC) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_MINUS_EQUAL) reduce 563) ((T_MOD_EQUAL) reduce 563) ((T_MUL_EQUAL) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_OR_EQUAL) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_PLUS_EQUAL) reduce 563) ((T_POW) reduce 563) ((T_POW_EQUAL) reduce 563) ((T_SL) reduce 563) ((T_SL_EQUAL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) ((T_SR_EQUAL) reduce 563) ((T_XOR_EQUAL) reduce 563) (("[") reduce 563) (("^") reduce 563) (("|") reduce 563)) 20313 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AS) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_DOUBLE_ARROW) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_PIPE) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("^") reduce 484) (("|") reduce 484)) 20314 ((("%") reduce 466) (("*") reduce 466) (("+") reduce 466) (("-") reduce 466) ((".") reduce 466) (("/") reduce 466) (("<") reduce 466) ((">") reduce 466) (("?") reduce 466) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_AS) reduce 466) ((T_BOOLEAN_AND) reduce 466) ((T_BOOLEAN_OR) reduce 466) ((T_COALESCE) reduce 466) ((T_DOUBLE_ARROW) reduce 466) ((T_INSTANCEOF) reduce 466) ((T_IS_EQUAL) reduce 466) ((T_IS_GREATER_OR_EQUAL) reduce 466) ((T_IS_IDENTICAL) reduce 466) ((T_IS_NOT_EQUAL) reduce 466) ((T_IS_NOT_IDENTICAL) reduce 466) ((T_IS_SMALLER_OR_EQUAL) reduce 466) ((T_LOGICAL_AND) reduce 466) ((T_LOGICAL_OR) reduce 466) ((T_LOGICAL_XOR) reduce 466) ((T_PIPE) reduce 466) ((T_POW) reduce 466) ((T_SL) reduce 466) ((T_SPACESHIP) reduce 466) ((T_SR) reduce 466) (("^") reduce 466) (("|") reduce 466)) 20315 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AS) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_PIPE) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("^") reduce 488) (("|") reduce 488)) 20316 ((("%") reduce 531) (("(") shift) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AS) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("|") reduce 531)) 20317 ((("%") reduce 461) (("(") reduce 558) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AS) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PIPE) reduce 461) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") reduce 552) (("^") reduce 461) (("|") reduce 461)) 20318 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AS) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_DOUBLE_ARROW) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_PIPE) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 20319 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AS) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_DOUBLE_ARROW) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_PIPE) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("|") reduce 476)) 20320 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) (("=") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AND_EQUAL) reduce 559) ((T_AS) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_COALESCE_EQUAL) reduce 559) ((T_CONCAT_EQUAL) reduce 559) ((T_DEC) reduce 559) ((T_DIV_EQUAL) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INC) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_MINUS_EQUAL) reduce 559) ((T_MOD_EQUAL) reduce 559) ((T_MUL_EQUAL) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_OR_EQUAL) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_PLUS_EQUAL) reduce 559) ((T_POW) reduce 559) ((T_POW_EQUAL) reduce 559) ((T_SL) reduce 559) ((T_SL_EQUAL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) ((T_SR_EQUAL) reduce 559) ((T_XOR_EQUAL) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 20321 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_AS) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 20322 ((("%") reduce 403) (("*") reduce 403) (("+") reduce 403) (("-") reduce 403) ((".") reduce 403) (("/") reduce 403) (("<") reduce 403) (("=") shift) ((">") reduce 403) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AND_EQUAL) shift) ((T_AS) reduce 403) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_DOUBLE_ARROW) reduce 403) ((T_INC) shift) ((T_INSTANCEOF) reduce 403) ((T_IS_EQUAL) reduce 403) ((T_IS_GREATER_OR_EQUAL) reduce 403) ((T_IS_IDENTICAL) reduce 403) ((T_IS_NOT_EQUAL) reduce 403) ((T_IS_NOT_IDENTICAL) reduce 403) ((T_IS_SMALLER_OR_EQUAL) reduce 403) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 403) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 403) ((T_POW_EQUAL) shift) ((T_SL) reduce 403) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 403) ((T_SR) reduce 403) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 548) (("^") reduce 403) (("|") reduce 403)) 20325 ((("%") reduce 448) (("*") reduce 448) (("+") reduce 448) (("-") reduce 448) ((".") reduce 448) (("/") reduce 448) (("<") reduce 448) ((">") reduce 448) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_AS) reduce 448) ((T_BOOLEAN_AND) reduce 448) ((T_BOOLEAN_OR) reduce 448) ((T_COALESCE) reduce 448) ((T_DOUBLE_ARROW) reduce 448) ((T_INSTANCEOF) reduce 448) ((T_IS_EQUAL) reduce 448) ((T_IS_GREATER_OR_EQUAL) reduce 448) ((T_IS_IDENTICAL) reduce 448) ((T_IS_NOT_EQUAL) reduce 448) ((T_IS_NOT_IDENTICAL) reduce 448) ((T_IS_SMALLER_OR_EQUAL) reduce 448) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_PIPE) reduce 448) ((T_POW) shift) ((T_SL) reduce 448) ((T_SPACESHIP) reduce 448) ((T_SR) reduce 448) (("^") reduce 448) (("|") reduce 448)) 20357 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 432) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_DOUBLE_ARROW) reduce 432) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 432)) 20358 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 435) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_DOUBLE_ARROW) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("|") reduce 435)) 20359 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 444) (("/") shift) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AS) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_PIPE) reduce 444) ((T_POW) shift) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("|") reduce 444)) 20360 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AS) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 458) (("|") reduce 458)) 20361 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 443) (("/") shift) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AS) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_DOUBLE_ARROW) reduce 443) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_PIPE) reduce 443) ((T_POW) shift) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("^") reduce 443) (("|") reduce 443)) 20362 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AS) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_DOUBLE_ARROW) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_PIPE) reduce 440) ((T_POW) shift) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("|") reduce 440)) 20363 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AS) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_DOUBLE_ARROW) reduce 453) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_PIPE) reduce 453) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 453) ((T_SR) shift) (("^") reduce 453) (("|") reduce 453)) 20364 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 431) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20365 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 429) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 429) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20366 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 430) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 430) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20367 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AS) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 455) ((T_SR) shift) (("^") reduce 455) (("|") reduce 455)) 20368 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AS) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_DOUBLE_ARROW) reduce 450) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 450) (("|") reduce 450)) 20369 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AS) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_DOUBLE_ARROW) reduce 452) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 452) (("|") reduce 452)) 20370 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AS) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_DOUBLE_ARROW) reduce 449) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 449) (("|") reduce 449)) 20371 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AS) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 457) ((T_SR) shift) (("^") reduce 457) (("|") reduce 457)) 20372 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AS) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_DOUBLE_ARROW) reduce 451) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 451) (("|") reduce 451)) 20375 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AS) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 20376 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AS) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 20377 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AS) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 20378 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AS) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 20379 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AS) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 20380 ((("%") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AS) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 20381 ((("%") reduce 511) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AS) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("|") reduce 511)) 20382 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AS) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PIPE) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 20383 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AS) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 20384 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AS) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("|") reduce 512)) 20385 ((("%") reduce 573) (("*") reduce 573) (("+") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AS) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_DOUBLE_ARROW) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("|") reduce 573)) 20391 ((("%") reduce 574) (("*") reduce 574) (("+") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AS) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_DOUBLE_ARROW) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("|") reduce 574)) 20392 ((("%") reduce 578) (("*") reduce 578) (("+") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AS) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_DOUBLE_ARROW) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("|") reduce 578)) 20393 ((("%") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AS) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 20394 ((("%") reduce 575) (("*") reduce 575) (("+") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AS) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_DOUBLE_ARROW) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("|") reduce 575)) 20395 ((("%") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AS) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 20398 ((("%") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AS) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 20399 ((("%") reduce 576) (("*") reduce 576) (("+") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AS) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_DOUBLE_ARROW) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("|") reduce 576)) 20401 ((("%") reduce 577) (("*") reduce 577) (("+") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AS) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_DOUBLE_ARROW) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("|") reduce 577)) 20403 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AS) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("|") reduce 513)) 20404 ((("%") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AS) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 20407 ((("%") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AS) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 20408 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 465) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 465) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 465) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20409 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 427) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_DOUBLE_ARROW) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20410 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 428) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_DOUBLE_ARROW) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20411 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AS) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_DOUBLE_ARROW) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("|") reduce 433)) 20412 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AS) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_DOUBLE_ARROW) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("|") reduce 434)) 20416 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 463) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 463) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20417 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 464) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 464) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20418 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AS) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 456) ((T_SR) shift) (("^") reduce 456) (("|") reduce 456)) 20419 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AS) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 454) ((T_SR) shift) (("^") reduce 454) (("|") reduce 454)) 20420 ((("%") reduce 441) (("*") reduce 441) (("+") reduce 441) (("-") reduce 441) ((".") reduce 441) (("/") reduce 441) (("<") reduce 441) ((">") reduce 441) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AS) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_DOUBLE_ARROW) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 441) ((T_IS_GREATER_OR_EQUAL) reduce 441) ((T_IS_IDENTICAL) reduce 441) ((T_IS_NOT_EQUAL) reduce 441) ((T_IS_NOT_IDENTICAL) reduce 441) ((T_IS_SMALLER_OR_EQUAL) reduce 441) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_PIPE) reduce 441) ((T_POW) shift) ((T_SL) reduce 441) ((T_SPACESHIP) reduce 441) ((T_SR) reduce 441) (("^") reduce 441) (("|") reduce 441)) 20421 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 436) (("/") shift) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AS) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_DOUBLE_ARROW) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_PIPE) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 436) ((T_SR) shift) (("^") reduce 436) (("|") reduce 436)) 20422 ((("%") shift) (("*") shift) (("+") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") shift) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AS) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_DOUBLE_ARROW) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_PIPE) reduce 438) ((T_POW) shift) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("|") reduce 438)) 20423 ((("%") shift) (("*") shift) (("+") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") shift) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AS) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_DOUBLE_ARROW) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_PIPE) reduce 437) ((T_POW) shift) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("|") reduce 437)) 20424 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AS) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_DOUBLE_ARROW) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_PIPE) reduce 439) ((T_POW) shift) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("|") reduce 439)) 20425 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AS) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_DOUBLE_ARROW) reduce 442) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_PIPE) reduce 442) ((T_POW) shift) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("|") reduce 442)) 20427 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AS) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_DOUBLE_ARROW) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_PIPE) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("|") reduce 74)) 20428 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AS) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_DOUBLE_ARROW) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_PIPE) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("|") reduce 52)) 20429 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AS) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_DOUBLE_ARROW) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_PIPE) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("|") reduce 27)) 20430 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AS) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_DOUBLE_ARROW) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_PIPE) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("|") reduce 51)) 20431 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AS) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_DOUBLE_ARROW) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_PIPE) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("|") reduce 53)) 20432 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AS) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_DOUBLE_ARROW) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_PIPE) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("|") reduce 49)) 20433 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AS) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_DOUBLE_ARROW) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_PIPE) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("|") reduce 29)) 20434 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AS) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_DOUBLE_ARROW) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_PIPE) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("|") reduce 59)) 20435 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AS) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_DOUBLE_ARROW) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_PIPE) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("|") reduce 60)) 20436 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AS) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_DOUBLE_ARROW) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_PIPE) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("|") reduce 11)) 20437 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AS) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_DOUBLE_ARROW) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_PIPE) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("|") reduce 42)) 20438 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AS) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_DOUBLE_ARROW) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_PIPE) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("|") reduce 39)) 20439 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AS) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_DOUBLE_ARROW) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_PIPE) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("|") reduce 25)) 20440 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AS) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_DOUBLE_ARROW) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_PIPE) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("|") reduce 50)) 20441 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AS) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_DOUBLE_ARROW) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_PIPE) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("|") reduce 66)) 20442 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AS) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_DOUBLE_ARROW) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_PIPE) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("|") reduce 18)) 20443 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AS) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_DOUBLE_ARROW) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_PIPE) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("|") reduce 17)) 20444 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AS) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_DOUBLE_ARROW) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_PIPE) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("|") reduce 15)) 20445 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AS) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_DOUBLE_ARROW) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_PIPE) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("|") reduce 14)) 20446 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AS) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_DOUBLE_ARROW) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_PIPE) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("|") reduce 38)) 20447 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AS) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_DOUBLE_ARROW) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_PIPE) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("|") reduce 26)) 20448 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AS) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_DOUBLE_ARROW) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_PIPE) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("|") reduce 22)) 20449 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AS) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_DOUBLE_ARROW) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_PIPE) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("|") reduce 24)) 20450 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AS) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_DOUBLE_ARROW) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_PIPE) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("|") reduce 16)) 20451 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AS) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_DOUBLE_ARROW) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_PIPE) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("|") reduce 48)) 20452 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AS) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_DOUBLE_ARROW) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_PIPE) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("|") reduce 20)) 20453 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AS) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_DOUBLE_ARROW) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_PIPE) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("|") reduce 70)) 20454 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AS) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_DOUBLE_ARROW) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_PIPE) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("|") reduce 3)) 20455 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AS) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_DOUBLE_ARROW) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_PIPE) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("|") reduce 12)) 20456 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AS) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_DOUBLE_ARROW) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_PIPE) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("|") reduce 54)) 20457 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AS) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_DOUBLE_ARROW) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_PIPE) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("|") reduce 65)) 20458 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AS) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_DOUBLE_ARROW) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_PIPE) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("|") reduce 75)) 20459 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AS) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_DOUBLE_ARROW) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_PIPE) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("|") reduce 30)) 20460 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AS) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_DOUBLE_ARROW) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_PIPE) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("|") reduce 68)) 20461 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AS) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_DOUBLE_ARROW) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_PIPE) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("|") reduce 21)) 20462 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AS) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_DOUBLE_ARROW) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_PIPE) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("|") reduce 23)) 20463 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AS) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_DOUBLE_ARROW) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_PIPE) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("|") reduce 41)) 20464 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AS) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_DOUBLE_ARROW) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_PIPE) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("|") reduce 62)) 20465 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AS) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_DOUBLE_ARROW) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_PIPE) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("|") reduce 34)) 20466 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AS) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_DOUBLE_ARROW) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_PIPE) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("|") reduce 40)) 20467 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AS) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_DOUBLE_ARROW) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_PIPE) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("|") reduce 13)) 20468 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AS) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_DOUBLE_ARROW) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_PIPE) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("|") reduce 55)) 20469 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AS) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_DOUBLE_ARROW) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_PIPE) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("|") reduce 1)) 20470 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AS) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_DOUBLE_ARROW) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_PIPE) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("|") reduce 2)) 20471 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AS) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_DOUBLE_ARROW) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_PIPE) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("|") reduce 9)) 20472 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AS) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_DOUBLE_ARROW) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_PIPE) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("|") reduce 33)) 20473 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AS) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_DOUBLE_ARROW) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_PIPE) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("|") reduce 58)) 20474 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AS) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_DOUBLE_ARROW) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_PIPE) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("|") reduce 37)) 20475 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AS) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_DOUBLE_ARROW) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_PIPE) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("|") reduce 64)) 20476 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AS) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_DOUBLE_ARROW) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_PIPE) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("|") reduce 46)) 20477 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AS) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_DOUBLE_ARROW) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_PIPE) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("|") reduce 8)) 20478 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AS) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_DOUBLE_ARROW) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_PIPE) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("|") reduce 6)) 20479 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AS) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_DOUBLE_ARROW) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_PIPE) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("|") reduce 7)) 20480 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AS) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_DOUBLE_ARROW) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_PIPE) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("|") reduce 69)) 20481 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AS) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_DOUBLE_ARROW) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_PIPE) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("|") reduce 63)) 20482 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AS) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_DOUBLE_ARROW) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_PIPE) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("|") reduce 56)) 20483 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AS) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_DOUBLE_ARROW) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_PIPE) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("|") reduce 10)) 20484 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AS) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_DOUBLE_ARROW) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_PIPE) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("|") reduce 67)) 20485 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AS) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_DOUBLE_ARROW) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_PIPE) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("|") reduce 44)) 20486 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AS) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_DOUBLE_ARROW) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_PIPE) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("|") reduce 76)) 20487 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AS) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_DOUBLE_ARROW) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_PIPE) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("|") reduce 71)) 20488 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AS) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_DOUBLE_ARROW) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_PIPE) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("|") reduce 77)) 20489 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AS) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_DOUBLE_ARROW) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_PIPE) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("|") reduce 78)) 20490 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AS) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_DOUBLE_ARROW) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_PIPE) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("|") reduce 79)) 20491 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AS) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_DOUBLE_ARROW) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_PIPE) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("|") reduce 4)) 20492 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AS) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_DOUBLE_ARROW) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_PIPE) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("|") reduce 5)) 20493 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AS) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_DOUBLE_ARROW) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_PIPE) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("|") reduce 43)) 20494 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AS) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_DOUBLE_ARROW) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_PIPE) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("|") reduce 73)) 20495 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AS) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_DOUBLE_ARROW) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_PIPE) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("|") reduce 82)) 20496 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AS) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_DOUBLE_ARROW) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_PIPE) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("|") reduce 47)) 20497 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AS) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_DOUBLE_ARROW) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_PIPE) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("|") reduce 31)) 20498 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AS) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_DOUBLE_ARROW) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_PIPE) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("|") reduce 57)) 20499 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AS) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_DOUBLE_ARROW) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_PIPE) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("|") reduce 61)) 20500 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AS) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_DOUBLE_ARROW) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_PIPE) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("|") reduce 28)) 20501 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AS) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_DOUBLE_ARROW) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_PIPE) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("|") reduce 36)) 20502 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AS) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_DOUBLE_ARROW) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_PIPE) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("|") reduce 32)) 20503 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AS) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_DOUBLE_ARROW) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_PIPE) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("|") reduce 35)) 20504 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AS) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_DOUBLE_ARROW) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_PIPE) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("|") reduce 19)) 20505 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AS) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_DOUBLE_ARROW) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_PIPE) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("|") reduce 45)) 20506 ((("%") reduce 542) (("(") reduce 579) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AS) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PIPE) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 20508 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AS) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_DOUBLE_ARROW) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_PIPE) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("|") reduce 72)) 20509 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AS) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_DOUBLE_ARROW) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_PIPE) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("|") reduce 83)) 20510 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) (("<") reduce 572) (("=") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AND_EQUAL) reduce 572) ((T_AS) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_COALESCE_EQUAL) reduce 572) ((T_CONCAT_EQUAL) reduce 572) ((T_DEC) reduce 572) ((T_DIV_EQUAL) reduce 572) ((T_DOUBLE_ARROW) reduce 572) ((T_INC) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_MINUS_EQUAL) reduce 572) ((T_MOD_EQUAL) reduce 572) ((T_MUL_EQUAL) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_OR_EQUAL) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_PLUS_EQUAL) reduce 572) ((T_POW) reduce 572) ((T_POW_EQUAL) reduce 572) ((T_SL) reduce 572) ((T_SL_EQUAL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) ((T_SR_EQUAL) reduce 572) ((T_XOR_EQUAL) reduce 572) (("[") reduce 572) (("^") reduce 572) (("|") reduce 572)) 20513 ((("%") reduce 544) (("(") reduce 580) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AS) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PIPE) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 20515 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_AS) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("|") reduce 507)) 20516 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_AS) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 20520 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) (("=") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AND_EQUAL) reduce 292) ((T_AS) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_COALESCE_EQUAL) reduce 292) ((T_CONCAT_EQUAL) reduce 292) ((T_DEC) reduce 292) ((T_DIV_EQUAL) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INC) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_MINUS_EQUAL) reduce 292) ((T_MOD_EQUAL) reduce 292) ((T_MUL_EQUAL) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_OR_EQUAL) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_PLUS_EQUAL) reduce 292) ((T_POW) reduce 292) ((T_POW_EQUAL) reduce 292) ((T_SL) reduce 292) ((T_SL_EQUAL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) ((T_SR_EQUAL) reduce 292) ((T_XOR_EQUAL) reduce 292) (("[") reduce 292) (("^") reduce 292) (("|") reduce 292)) 20521 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) (("=") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AND_EQUAL) reduce 293) ((T_AS) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_COALESCE_EQUAL) reduce 293) ((T_CONCAT_EQUAL) reduce 293) ((T_DEC) reduce 293) ((T_DIV_EQUAL) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INC) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_MINUS_EQUAL) reduce 293) ((T_MOD_EQUAL) reduce 293) ((T_MUL_EQUAL) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_OR_EQUAL) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_PLUS_EQUAL) reduce 293) ((T_POW) reduce 293) ((T_POW_EQUAL) reduce 293) ((T_SL) reduce 293) ((T_SL_EQUAL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) ((T_SR_EQUAL) reduce 293) ((T_XOR_EQUAL) reduce 293) (("[") reduce 293) (("^") reduce 293) (("|") reduce 293)) 20526 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AS) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_DOUBLE_ARROW) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_PIPE) reduce 425) ((T_POW) reduce 425) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 20528 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AS) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_DOUBLE_ARROW) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_PIPE) reduce 423) ((T_POW) reduce 423) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 20538 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 419) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20539 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 421) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20540 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 420) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20541 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 413) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 413) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20542 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 410) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 410) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20543 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 418) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20544 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 412) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20545 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 416) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20546 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 411) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20547 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 414) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 414) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20548 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 415) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 415) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20549 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 422) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20550 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 417) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20552 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 406) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20555 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AS) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 20558 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) (("<") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AS) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_DOUBLE_ARROW) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_POW) reduce 564) ((T_SL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) (("[") reduce 564) (("^") reduce 564) (("|") reduce 564)) 20560 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) (("<") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AS) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_DOUBLE_ARROW) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_POW) reduce 563) ((T_SL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) (("[") reduce 563) (("^") reduce 563) (("|") reduce 563)) 20562 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AS) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 20563 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AS) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 20564 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AS) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_DOUBLE_ARROW) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("[") reduce 548) (("^") reduce 407) (("|") reduce 407)) 20568 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) (("<") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AS) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_DOUBLE_ARROW) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_POW) reduce 572) ((T_SL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) (("[") reduce 572) (("^") reduce 572) (("|") reduce 572)) 20570 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AS) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("|") reduce 507)) 20571 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AS) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 20575 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AS) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("[") reduce 292) (("^") reduce 292) (("|") reduce 292)) 20576 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AS) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("[") reduce 293) (("^") reduce 293) (("|") reduce 293)) 20577 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AS) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 20580 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) (("<") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AS) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_DOUBLE_ARROW) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_POW) reduce 571) ((T_SL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) (("[") reduce 571) (("^") reduce 571) (("|") reduce 571)) 20581 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AS) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("|") reduce 506)) 20582 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AS) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("|") reduce 508)) 20587 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AS) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 20588 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AS) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 20589 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AS) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 20590 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AS) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 20593 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AS) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 20594 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AS) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 20595 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AS) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 20596 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AS) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("[") reduce 562) (("^") reduce 562) (("|") reduce 562)) 20597 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AS) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 20598 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AS) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 20601 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AS) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 20602 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) (("=") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AND_EQUAL) reduce 504) ((T_AS) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_COALESCE_EQUAL) reduce 504) ((T_CONCAT_EQUAL) reduce 504) ((T_DEC) reduce 504) ((T_DIV_EQUAL) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INC) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_MINUS_EQUAL) reduce 504) ((T_MOD_EQUAL) reduce 504) ((T_MUL_EQUAL) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_OR_EQUAL) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_PLUS_EQUAL) reduce 504) ((T_POW) reduce 504) ((T_POW_EQUAL) reduce 504) ((T_SL) reduce 504) ((T_SL_EQUAL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) ((T_SR_EQUAL) reduce 504) ((T_XOR_EQUAL) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 20613 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AS) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PIPE) reduce 494) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 20614 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AS) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_PIPE) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("^") reduce 489) (("|") reduce 489)) 20624 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 494) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20625 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AS) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_PIPE) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("^") reduce 490) (("|") reduce 490)) 20627 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 481) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20629 ((("%") reduce 541) (("(") reduce 579) (("*") reduce 541) (("+") reduce 541) (("-") reduce 541) ((".") reduce 541) (("/") reduce 541) (("<") reduce 541) ((">") reduce 541) (("?") reduce 541) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_AS) reduce 541) ((T_BOOLEAN_AND) reduce 541) ((T_BOOLEAN_OR) reduce 541) ((T_COALESCE) reduce 541) ((T_DOUBLE_ARROW) reduce 541) ((T_INSTANCEOF) reduce 541) ((T_IS_EQUAL) reduce 541) ((T_IS_GREATER_OR_EQUAL) reduce 541) ((T_IS_IDENTICAL) reduce 541) ((T_IS_NOT_EQUAL) reduce 541) ((T_IS_NOT_IDENTICAL) reduce 541) ((T_IS_SMALLER_OR_EQUAL) reduce 541) ((T_LOGICAL_AND) reduce 541) ((T_LOGICAL_OR) reduce 541) ((T_LOGICAL_XOR) reduce 541) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 541) ((T_OBJECT_OPERATOR) reduce 541) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 541) ((T_PIPE) reduce 541) ((T_POW) reduce 541) ((T_SL) reduce 541) ((T_SPACESHIP) reduce 541) ((T_SR) reduce 541) (("[") reduce 541) (("^") reduce 541) (("|") reduce 541)) 20631 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) (("<") reduce 571) (("=") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AND_EQUAL) reduce 571) ((T_AS) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_COALESCE_EQUAL) reduce 571) ((T_CONCAT_EQUAL) reduce 571) ((T_DEC) reduce 571) ((T_DIV_EQUAL) reduce 571) ((T_DOUBLE_ARROW) reduce 571) ((T_INC) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_MINUS_EQUAL) reduce 571) ((T_MOD_EQUAL) reduce 571) ((T_MUL_EQUAL) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_OR_EQUAL) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_PLUS_EQUAL) reduce 571) ((T_POW) reduce 571) ((T_POW_EQUAL) reduce 571) ((T_SL) reduce 571) ((T_SL_EQUAL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) ((T_SR_EQUAL) reduce 571) ((T_XOR_EQUAL) reduce 571) (("[") reduce 571) (("^") reduce 571) (("|") reduce 571)) 20634 ((("%") reduce 543) (("(") reduce 580) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AS) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PIPE) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 20635 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_AS) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("|") reduce 506)) 20636 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_AS) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("|") reduce 508)) 20638 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AS) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_DOUBLE_ARROW) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_PIPE) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("^") reduce 485) (("|") reduce 485)) 20639 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AS) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_PIPE) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("^") reduce 487) (("|") reduce 487)) 20644 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) (("=") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AND_EQUAL) reduce 560) ((T_AS) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_COALESCE_EQUAL) reduce 560) ((T_CONCAT_EQUAL) reduce 560) ((T_DEC) reduce 560) ((T_DIV_EQUAL) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INC) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_MINUS_EQUAL) reduce 560) ((T_MOD_EQUAL) reduce 560) ((T_MUL_EQUAL) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_OR_EQUAL) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_PLUS_EQUAL) reduce 560) ((T_POW) reduce 560) ((T_POW_EQUAL) reduce 560) ((T_SL) reduce 560) ((T_SL_EQUAL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) ((T_SR_EQUAL) reduce 560) ((T_XOR_EQUAL) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 20645 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) (("=") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AND_EQUAL) reduce 582) ((T_AS) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_COALESCE_EQUAL) reduce 582) ((T_CONCAT_EQUAL) reduce 582) ((T_DEC) reduce 582) ((T_DIV_EQUAL) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INC) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_MINUS_EQUAL) reduce 582) ((T_MOD_EQUAL) reduce 582) ((T_MUL_EQUAL) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_OR_EQUAL) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_PLUS_EQUAL) reduce 582) ((T_POW) reduce 582) ((T_POW_EQUAL) reduce 582) ((T_SL) reduce 582) ((T_SL_EQUAL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) ((T_SR_EQUAL) reduce 582) ((T_XOR_EQUAL) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 20646 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_AS) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 20647 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) (("=") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AND_EQUAL) reduce 584) ((T_AS) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_COALESCE_EQUAL) reduce 584) ((T_CONCAT_EQUAL) reduce 584) ((T_DEC) reduce 584) ((T_DIV_EQUAL) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INC) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_MINUS_EQUAL) reduce 584) ((T_MOD_EQUAL) reduce 584) ((T_MUL_EQUAL) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_OR_EQUAL) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_PLUS_EQUAL) reduce 584) ((T_POW) reduce 584) ((T_POW_EQUAL) reduce 584) ((T_SL) reduce 584) ((T_SL_EQUAL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) ((T_SR_EQUAL) reduce 584) ((T_XOR_EQUAL) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 20650 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) (("=") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AND_EQUAL) reduce 583) ((T_AS) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_COALESCE_EQUAL) reduce 583) ((T_CONCAT_EQUAL) reduce 583) ((T_DEC) reduce 583) ((T_DIV_EQUAL) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INC) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_MINUS_EQUAL) reduce 583) ((T_MOD_EQUAL) reduce 583) ((T_MUL_EQUAL) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_OR_EQUAL) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_PLUS_EQUAL) reduce 583) ((T_POW) reduce 583) ((T_POW_EQUAL) reduce 583) ((T_SL) reduce 583) ((T_SL_EQUAL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) ((T_SR_EQUAL) reduce 583) ((T_XOR_EQUAL) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 20651 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) (("=") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AND_EQUAL) reduce 561) ((T_AS) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_COALESCE_EQUAL) reduce 561) ((T_CONCAT_EQUAL) reduce 561) ((T_DEC) reduce 561) ((T_DIV_EQUAL) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INC) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_MINUS_EQUAL) reduce 561) ((T_MOD_EQUAL) reduce 561) ((T_MUL_EQUAL) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_OR_EQUAL) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_PLUS_EQUAL) reduce 561) ((T_POW) reduce 561) ((T_POW_EQUAL) reduce 561) ((T_SL) reduce 561) ((T_SL_EQUAL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) ((T_SR_EQUAL) reduce 561) ((T_XOR_EQUAL) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 20652 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_AS) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 20653 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) (("=") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AND_EQUAL) reduce 562) ((T_AS) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_COALESCE_EQUAL) reduce 562) ((T_CONCAT_EQUAL) reduce 562) ((T_DEC) reduce 562) ((T_DIV_EQUAL) reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_INC) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_MINUS_EQUAL) reduce 562) ((T_MOD_EQUAL) reduce 562) ((T_MUL_EQUAL) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_OR_EQUAL) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_PLUS_EQUAL) reduce 562) ((T_POW) reduce 562) ((T_POW_EQUAL) reduce 562) ((T_SL) reduce 562) ((T_SL_EQUAL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) ((T_SR_EQUAL) reduce 562) ((T_XOR_EQUAL) reduce 562) (("[") reduce 562) (("^") reduce 562) (("|") reduce 562)) 20655 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AS) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_PIPE) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 20657 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) (("=") shift) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AS) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_PIPE) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("|") reduce 520)) 20659 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 405) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20660 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 482) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 482) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20663 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 481) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 481) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20664 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AS) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_DOUBLE_ARROW) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_PIPE) reduce 473) ((T_POW) shift) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("|") reduce 473)) 20665 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 483) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 483) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20666 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AS) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_DOUBLE_ARROW) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_PIPE) reduce 469) ((T_POW) shift) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 20667 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AS) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_DOUBLE_ARROW) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_PIPE) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("^") reduce 486) (("|") reduce 486)) 20669 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AS) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_PIPE) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("^") reduce 526) (("|") reduce 526)) 20671 ((("%") reduce 527) (("*") reduce 527) (("+") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AS) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_DOUBLE_ARROW) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_PIPE) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("^") reduce 527) (("|") reduce 527)) 20672 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AS) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_PIPE) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("^") reduce 525) (("|") reduce 525)) 20673 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 619) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 619) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20674 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 618) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 618) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20675 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_AS) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 20676 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 478) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 478) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20677 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AS) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_DOUBLE_ARROW) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_PIPE) reduce 471) ((T_POW) shift) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("|") reduce 471)) 20679 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AS) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 20680 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AS) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 20681 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AS) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 20682 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AS) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 20683 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AS) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 20684 ((("%") reduce 400) (("(") reduce 400) (("*") reduce 400) (("+") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AS) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_DOUBLE_ARROW) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 400) ((T_OBJECT_OPERATOR) reduce 400) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 400) ((T_PIPE) reduce 400) ((T_POW) reduce 400) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("[") reduce 400) (("^") reduce 400) (("|") reduce 400)) 20687 ((("%") reduce 511) (("(") reduce 511) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AS) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("|") reduce 511)) 20688 ((("%") reduce 402) (("(") shift) (("*") reduce 402) (("+") reduce 402) (("-") reduce 402) ((".") reduce 402) (("/") reduce 402) (("<") reduce 402) ((">") reduce 402) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AS) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_DOUBLE_ARROW) reduce 402) ((T_INSTANCEOF) reduce 402) ((T_IS_EQUAL) reduce 402) ((T_IS_GREATER_OR_EQUAL) reduce 402) ((T_IS_IDENTICAL) reduce 402) ((T_IS_NOT_EQUAL) reduce 402) ((T_IS_NOT_IDENTICAL) reduce 402) ((T_IS_SMALLER_OR_EQUAL) reduce 402) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_PIPE) reduce 402) ((T_POW) reduce 402) ((T_SL) reduce 402) ((T_SPACESHIP) reduce 402) ((T_SR) reduce 402) (("^") reduce 402) (("|") reduce 402)) 20689 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AS) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 20690 ((("%") reduce 512) (("(") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AS) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("|") reduce 512)) 20691 ((("%") reduce 573) (("(") reduce 573) (("*") reduce 573) (("+") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AS) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_DOUBLE_ARROW) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("|") reduce 573)) 20697 ((("%") reduce 574) (("(") reduce 574) (("*") reduce 574) (("+") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AS) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_DOUBLE_ARROW) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("|") reduce 574)) 20698 ((("%") reduce 578) (("(") reduce 578) (("*") reduce 578) (("+") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AS) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_DOUBLE_ARROW) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("|") reduce 578)) 20699 ((("%") reduce 575) (("(") reduce 575) (("*") reduce 575) (("+") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AS) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_DOUBLE_ARROW) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("|") reduce 575)) 20700 ((("%") reduce 576) (("(") reduce 576) (("*") reduce 576) (("+") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AS) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_DOUBLE_ARROW) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("|") reduce 576)) 20701 ((("%") reduce 399) (("(") reduce 399) (("*") reduce 399) (("+") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AS) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_DOUBLE_ARROW) reduce 399) ((T_INSTANCEOF) reduce 399) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 399) ((T_OBJECT_OPERATOR) reduce 399) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 399) ((T_PIPE) reduce 399) ((T_POW) reduce 399) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("[") reduce 399) (("^") reduce 399) (("|") reduce 399)) 20703 ((("%") reduce 577) (("(") reduce 577) (("*") reduce 577) (("+") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AS) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_DOUBLE_ARROW) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("|") reduce 577)) 20704 ((("%") reduce 401) (("(") reduce 401) (("*") reduce 401) (("+") reduce 401) (("-") reduce 401) ((".") reduce 401) (("/") reduce 401) (("<") reduce 401) ((">") reduce 401) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AS) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_DOUBLE_ARROW) reduce 401) ((T_INSTANCEOF) reduce 401) ((T_IS_EQUAL) reduce 401) ((T_IS_GREATER_OR_EQUAL) reduce 401) ((T_IS_IDENTICAL) reduce 401) ((T_IS_NOT_EQUAL) reduce 401) ((T_IS_NOT_IDENTICAL) reduce 401) ((T_IS_SMALLER_OR_EQUAL) reduce 401) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 401) ((T_OBJECT_OPERATOR) reduce 401) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 401) ((T_PIPE) reduce 401) ((T_POW) reduce 401) ((T_SL) reduce 401) ((T_SPACESHIP) reduce 401) ((T_SR) reduce 401) (("[") reduce 401) (("^") reduce 401) (("|") reduce 401)) 20712 ((("%") reduce 398) (("(") reduce 398) (("*") reduce 398) (("+") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AS) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_DOUBLE_ARROW) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 398) ((T_OBJECT_OPERATOR) reduce 398) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 398) ((T_PIPE) reduce 398) ((T_POW) reduce 398) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("[") reduce 398) (("^") reduce 398) (("|") reduce 398)) 20714 ((("%") reduce 513) (("(") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AS) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("|") reduce 513)) 20720 ((("%") reduce 233) (("*") reduce 233) (("+") reduce 233) (("-") reduce 233) ((".") reduce 233) (("/") reduce 233) (("<") reduce 233) ((">") reduce 233) (("?") reduce 233) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_AS) reduce 233) ((T_BOOLEAN_AND) reduce 233) ((T_BOOLEAN_OR) reduce 233) ((T_COALESCE) reduce 233) ((T_DOUBLE_ARROW) reduce 233) ((T_INSTANCEOF) reduce 233) ((T_IS_EQUAL) reduce 233) ((T_IS_GREATER_OR_EQUAL) reduce 233) ((T_IS_IDENTICAL) reduce 233) ((T_IS_NOT_EQUAL) reduce 233) ((T_IS_NOT_IDENTICAL) reduce 233) ((T_IS_SMALLER_OR_EQUAL) reduce 233) ((T_LOGICAL_AND) reduce 233) ((T_LOGICAL_OR) reduce 233) ((T_LOGICAL_XOR) reduce 233) ((T_PIPE) reduce 233) ((T_POW) reduce 233) ((T_SL) reduce 233) ((T_SPACESHIP) reduce 233) ((T_SR) reduce 233) (("^") reduce 233) (("|") reduce 233)) 20725 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 404) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20729 ((("%") reduce 613) (("*") reduce 613) (("+") reduce 613) (("-") reduce 613) ((".") reduce 613) (("/") reduce 613) (("<") reduce 613) ((">") reduce 613) (("?") reduce 613) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_AS) reduce 613) ((T_BOOLEAN_AND) reduce 613) ((T_BOOLEAN_OR) reduce 613) ((T_COALESCE) reduce 613) ((T_DOUBLE_ARROW) reduce 613) ((T_INSTANCEOF) reduce 613) ((T_IS_EQUAL) reduce 613) ((T_IS_GREATER_OR_EQUAL) reduce 613) ((T_IS_IDENTICAL) reduce 613) ((T_IS_NOT_EQUAL) reduce 613) ((T_IS_NOT_IDENTICAL) reduce 613) ((T_IS_SMALLER_OR_EQUAL) reduce 613) ((T_LOGICAL_AND) reduce 613) ((T_LOGICAL_OR) reduce 613) ((T_LOGICAL_XOR) reduce 613) ((T_PIPE) reduce 613) ((T_POW) reduce 613) ((T_SL) reduce 613) ((T_SPACESHIP) reduce 613) ((T_SR) reduce 613) (("^") reduce 613) (("|") reduce 613)) 20730 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AS) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_DOUBLE_ARROW) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_PIPE) reduce 467) ((T_POW) shift) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("|") reduce 467)) 20731 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 616) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 616) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20732 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 615) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 615) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20733 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AS) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_DOUBLE_ARROW) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("[") reduce 548) (("^") reduce 424) (("|") reduce 424)) 20735 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AS) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_PIPE) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("^") reduce 518) (("|") reduce 518)) 20736 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AS) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_DOUBLE_ARROW) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_PIPE) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("|") reduce 474)) 20737 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AS) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("|") reduce 291)) 20741 ((("%") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AS) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("^") reduce 292) (("|") reduce 292)) 20742 ((("%") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AS) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("^") reduce 293) (("|") reduce 293)) 20745 ((("%") reduce 617) (("*") reduce 617) (("+") reduce 617) (("-") reduce 617) ((".") reduce 617) (("/") reduce 617) (("<") reduce 617) ((">") reduce 617) (("?") reduce 617) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_AS) reduce 617) ((T_BOOLEAN_AND) reduce 617) ((T_BOOLEAN_OR) reduce 617) ((T_COALESCE) reduce 617) ((T_DOUBLE_ARROW) reduce 617) ((T_INSTANCEOF) reduce 617) ((T_IS_EQUAL) reduce 617) ((T_IS_GREATER_OR_EQUAL) reduce 617) ((T_IS_IDENTICAL) reduce 617) ((T_IS_NOT_EQUAL) reduce 617) ((T_IS_NOT_IDENTICAL) reduce 617) ((T_IS_SMALLER_OR_EQUAL) reduce 617) ((T_LOGICAL_AND) reduce 617) ((T_LOGICAL_OR) reduce 617) ((T_LOGICAL_XOR) reduce 617) ((T_PIPE) reduce 617) ((T_POW) reduce 617) ((T_SL) reduce 617) ((T_SPACESHIP) reduce 617) ((T_SR) reduce 617) (("^") reduce 617) (("|") reduce 617)) 20748 ((("%") reduce 614) (("*") reduce 614) (("+") reduce 614) (("-") reduce 614) ((".") reduce 614) (("/") reduce 614) (("<") reduce 614) ((">") reduce 614) (("?") reduce 614) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_AS) reduce 614) ((T_BOOLEAN_AND) reduce 614) ((T_BOOLEAN_OR) reduce 614) ((T_COALESCE) reduce 614) ((T_DOUBLE_ARROW) reduce 614) ((T_INSTANCEOF) reduce 614) ((T_IS_EQUAL) reduce 614) ((T_IS_GREATER_OR_EQUAL) reduce 614) ((T_IS_IDENTICAL) reduce 614) ((T_IS_NOT_EQUAL) reduce 614) ((T_IS_NOT_IDENTICAL) reduce 614) ((T_IS_SMALLER_OR_EQUAL) reduce 614) ((T_LOGICAL_AND) reduce 614) ((T_LOGICAL_OR) reduce 614) ((T_LOGICAL_XOR) reduce 614) ((T_PIPE) reduce 614) ((T_POW) reduce 614) ((T_SL) reduce 614) ((T_SPACESHIP) reduce 614) ((T_SR) reduce 614) (("^") reduce 614) (("|") reduce 614)) 20749 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AS) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_DOUBLE_ARROW) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_PIPE) reduce 468) ((T_POW) shift) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 20750 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AS) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_DOUBLE_ARROW) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("[") reduce 548) (("^") reduce 426) (("|") reduce 426)) 20752 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AS) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_DOUBLE_ARROW) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_PIPE) reduce 408) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("^") reduce 408) (("|") reduce 408)) 20753 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AS) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_DOUBLE_ARROW) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_PIPE) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("|") reduce 409)) 20754 ((("%") reduce 296) (("*") reduce 296) (("+") reduce 296) (("-") reduce 296) ((".") reduce 296) (("/") reduce 296) (("<") reduce 296) ((">") reduce 296) (("?") reduce 296) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_AS) reduce 296) ((T_BOOLEAN_AND) reduce 296) ((T_BOOLEAN_OR) reduce 296) ((T_COALESCE) reduce 296) ((T_DOUBLE_ARROW) reduce 296) ((T_INSTANCEOF) reduce 296) ((T_IS_EQUAL) reduce 296) ((T_IS_GREATER_OR_EQUAL) reduce 296) ((T_IS_IDENTICAL) reduce 296) ((T_IS_NOT_EQUAL) reduce 296) ((T_IS_NOT_IDENTICAL) reduce 296) ((T_IS_SMALLER_OR_EQUAL) reduce 296) ((T_LOGICAL_AND) reduce 296) ((T_LOGICAL_OR) reduce 296) ((T_LOGICAL_XOR) reduce 296) ((T_PIPE) reduce 296) ((T_POW) reduce 296) ((T_SL) reduce 296) ((T_SPACESHIP) reduce 296) ((T_SR) reduce 296) (("^") reduce 296) (("|") reduce 296)) 20759 ((("%") reduce 297) (("*") reduce 297) (("+") reduce 297) (("-") reduce 297) ((".") reduce 297) (("/") reduce 297) (("<") reduce 297) ((">") reduce 297) (("?") reduce 297) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_AS) reduce 297) ((T_BOOLEAN_AND) reduce 297) ((T_BOOLEAN_OR) reduce 297) ((T_COALESCE) reduce 297) ((T_DOUBLE_ARROW) reduce 297) ((T_INSTANCEOF) reduce 297) ((T_IS_EQUAL) reduce 297) ((T_IS_GREATER_OR_EQUAL) reduce 297) ((T_IS_IDENTICAL) reduce 297) ((T_IS_NOT_EQUAL) reduce 297) ((T_IS_NOT_IDENTICAL) reduce 297) ((T_IS_SMALLER_OR_EQUAL) reduce 297) ((T_LOGICAL_AND) reduce 297) ((T_LOGICAL_OR) reduce 297) ((T_LOGICAL_XOR) reduce 297) ((T_PIPE) reduce 297) ((T_POW) reduce 297) ((T_SL) reduce 297) ((T_SPACESHIP) reduce 297) ((T_SR) reduce 297) (("^") reduce 297) (("|") reduce 297)) 20760 ((("%") reduce 460) (("(") reduce 556) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AS) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PIPE) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("[") reduce 549) (("^") reduce 460) (("|") reduce 460)) 20762 ((("%") reduce 298) (("*") reduce 298) (("+") reduce 298) (("-") reduce 298) ((".") reduce 298) (("/") reduce 298) (("<") reduce 298) ((">") reduce 298) (("?") reduce 298) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_AS) reduce 298) ((T_BOOLEAN_AND) reduce 298) ((T_BOOLEAN_OR) reduce 298) ((T_COALESCE) reduce 298) ((T_DOUBLE_ARROW) reduce 298) ((T_INSTANCEOF) reduce 298) ((T_IS_EQUAL) reduce 298) ((T_IS_GREATER_OR_EQUAL) reduce 298) ((T_IS_IDENTICAL) reduce 298) ((T_IS_NOT_EQUAL) reduce 298) ((T_IS_NOT_IDENTICAL) reduce 298) ((T_IS_SMALLER_OR_EQUAL) reduce 298) ((T_LOGICAL_AND) reduce 298) ((T_LOGICAL_OR) reduce 298) ((T_LOGICAL_XOR) reduce 298) ((T_PIPE) reduce 298) ((T_POW) reduce 298) ((T_SL) reduce 298) ((T_SPACESHIP) reduce 298) ((T_SR) reduce 298) (("^") reduce 298) (("|") reduce 298)) 20763 ((("%") reduce 299) (("*") reduce 299) (("+") reduce 299) (("-") reduce 299) ((".") reduce 299) (("/") reduce 299) (("<") reduce 299) ((">") reduce 299) (("?") reduce 299) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_AS) reduce 299) ((T_BOOLEAN_AND) reduce 299) ((T_BOOLEAN_OR) reduce 299) ((T_COALESCE) reduce 299) ((T_DOUBLE_ARROW) reduce 299) ((T_INSTANCEOF) reduce 299) ((T_IS_EQUAL) reduce 299) ((T_IS_GREATER_OR_EQUAL) reduce 299) ((T_IS_IDENTICAL) reduce 299) ((T_IS_NOT_EQUAL) reduce 299) ((T_IS_NOT_IDENTICAL) reduce 299) ((T_IS_SMALLER_OR_EQUAL) reduce 299) ((T_LOGICAL_AND) reduce 299) ((T_LOGICAL_OR) reduce 299) ((T_LOGICAL_XOR) reduce 299) ((T_PIPE) reduce 299) ((T_POW) reduce 299) ((T_SL) reduce 299) ((T_SPACESHIP) reduce 299) ((T_SR) reduce 299) (("^") reduce 299) (("|") reduce 299)) 20764 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AS) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_DOUBLE_ARROW) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_PIPE) reduce 472) ((T_POW) shift) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("|") reduce 472)) 20765 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AS) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_DOUBLE_ARROW) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_PIPE) reduce 470) ((T_POW) shift) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("|") reduce 470)) 20768 ((("%") reduce 519) (("(") reduce 519) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AS) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PIPE) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("|") reduce 519)) 20769 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AS) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_DOUBLE_ARROW) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_PIPE) reduce 475) ((T_POW) shift) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("|") reduce 475)) 20770 ((("%") reduce 446) (("*") reduce 446) (("+") reduce 446) (("-") reduce 446) ((".") reduce 446) (("/") reduce 446) (("<") reduce 446) ((">") reduce 446) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_AS) reduce 446) ((T_BOOLEAN_AND) reduce 446) ((T_BOOLEAN_OR) reduce 446) ((T_COALESCE) reduce 446) ((T_DOUBLE_ARROW) reduce 446) ((T_INSTANCEOF) reduce 446) ((T_IS_EQUAL) reduce 446) ((T_IS_GREATER_OR_EQUAL) reduce 446) ((T_IS_IDENTICAL) reduce 446) ((T_IS_NOT_EQUAL) reduce 446) ((T_IS_NOT_IDENTICAL) reduce 446) ((T_IS_SMALLER_OR_EQUAL) reduce 446) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_PIPE) reduce 446) ((T_POW) shift) ((T_SL) reduce 446) ((T_SPACESHIP) reduce 446) ((T_SR) reduce 446) (("^") reduce 446) (("|") reduce 446)) 20771 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AS) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_DOUBLE_ARROW) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_PIPE) reduce 445) ((T_POW) shift) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("|") reduce 445)) 20773 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) (("=") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AND_EQUAL) reduce 570) ((T_AS) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_COALESCE_EQUAL) reduce 570) ((T_CONCAT_EQUAL) reduce 570) ((T_DEC) reduce 570) ((T_DIV_EQUAL) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INC) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_MINUS_EQUAL) reduce 570) ((T_MOD_EQUAL) reduce 570) ((T_MUL_EQUAL) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_OR_EQUAL) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_PLUS_EQUAL) reduce 570) ((T_POW) reduce 570) ((T_POW_EQUAL) reduce 570) ((T_SL) reduce 570) ((T_SL_EQUAL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) ((T_SR_EQUAL) reduce 570) ((T_XOR_EQUAL) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 20776 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) (("=") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AND_EQUAL) reduce 569) ((T_AS) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_COALESCE_EQUAL) reduce 569) ((T_CONCAT_EQUAL) reduce 569) ((T_DEC) reduce 569) ((T_DIV_EQUAL) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INC) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_MINUS_EQUAL) reduce 569) ((T_MOD_EQUAL) reduce 569) ((T_MUL_EQUAL) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_OR_EQUAL) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_PLUS_EQUAL) reduce 569) ((T_POW) reduce 569) ((T_POW_EQUAL) reduce 569) ((T_SL) reduce 569) ((T_SL_EQUAL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) ((T_SR_EQUAL) reduce 569) ((T_XOR_EQUAL) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 20778 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AS) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_PIPE) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("|") reduce 522)) 20779 ((("%") reduce 447) (("*") reduce 447) (("+") reduce 447) (("-") reduce 447) ((".") reduce 447) (("/") reduce 447) (("<") reduce 447) ((">") reduce 447) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_AS) reduce 447) ((T_BOOLEAN_AND) reduce 447) ((T_BOOLEAN_OR) reduce 447) ((T_COALESCE) reduce 447) ((T_DOUBLE_ARROW) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 447) ((T_IS_GREATER_OR_EQUAL) reduce 447) ((T_IS_IDENTICAL) reduce 447) ((T_IS_NOT_EQUAL) reduce 447) ((T_IS_NOT_IDENTICAL) reduce 447) ((T_IS_SMALLER_OR_EQUAL) reduce 447) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_PIPE) reduce 447) ((T_POW) shift) ((T_SL) reduce 447) ((T_SPACESHIP) reduce 447) ((T_SR) reduce 447) (("^") reduce 447) (("|") reduce 447)) 20780 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AS) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_PIPE) reduce 473) ((T_POW) shift) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("|") reduce 473)) 20781 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 483) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20782 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AS) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_PIPE) reduce 469) ((T_POW) shift) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 20783 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AS) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_PIPE) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("^") reduce 486) (("|") reduce 486)) 20785 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AS) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_PIPE) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("^") reduce 526) (("|") reduce 526)) 20787 ((("%") reduce 527) (("*") reduce 527) (("+") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AS) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_PIPE) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("^") reduce 527) (("|") reduce 527)) 20788 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AS) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_PIPE) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("^") reduce 525) (("|") reduce 525)) 20789 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 619) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20790 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 618) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20791 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_AS) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 20792 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 478) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20793 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AS) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_PIPE) reduce 471) ((T_POW) shift) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("|") reduce 471)) 20795 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AS) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 20796 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AS) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 20797 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AS) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 20798 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AS) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 20799 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AS) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 20800 ((("%") reduce 400) (("(") reduce 400) (("*") reduce 400) (("+") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AS) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 400) ((T_OBJECT_OPERATOR) reduce 400) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 400) ((T_PIPE) reduce 400) ((T_POW) reduce 400) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("[") reduce 400) (("^") reduce 400) (("|") reduce 400)) 20803 ((("%") reduce 511) (("(") reduce 511) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AS) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("|") reduce 511)) 20804 ((("%") reduce 402) (("(") shift) (("*") reduce 402) (("+") reduce 402) (("-") reduce 402) ((".") reduce 402) (("/") reduce 402) (("<") reduce 402) ((">") reduce 402) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AS) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_INSTANCEOF) reduce 402) ((T_IS_EQUAL) reduce 402) ((T_IS_GREATER_OR_EQUAL) reduce 402) ((T_IS_IDENTICAL) reduce 402) ((T_IS_NOT_EQUAL) reduce 402) ((T_IS_NOT_IDENTICAL) reduce 402) ((T_IS_SMALLER_OR_EQUAL) reduce 402) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_PIPE) reduce 402) ((T_POW) reduce 402) ((T_SL) reduce 402) ((T_SPACESHIP) reduce 402) ((T_SR) reduce 402) (("^") reduce 402) (("|") reduce 402)) 20805 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AS) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 20806 ((("%") reduce 512) (("(") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AS) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("|") reduce 512)) 20807 ((("%") reduce 573) (("(") reduce 573) (("*") reduce 573) (("+") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AS) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("|") reduce 573)) 20813 ((("%") reduce 574) (("(") reduce 574) (("*") reduce 574) (("+") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AS) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("|") reduce 574)) 20814 ((("%") reduce 578) (("(") reduce 578) (("*") reduce 578) (("+") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AS) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("|") reduce 578)) 20815 ((("%") reduce 575) (("(") reduce 575) (("*") reduce 575) (("+") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AS) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("|") reduce 575)) 20816 ((("%") reduce 576) (("(") reduce 576) (("*") reduce 576) (("+") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AS) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("|") reduce 576)) 20817 ((("%") reduce 399) (("(") reduce 399) (("*") reduce 399) (("+") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AS) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_INSTANCEOF) reduce 399) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 399) ((T_OBJECT_OPERATOR) reduce 399) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 399) ((T_PIPE) reduce 399) ((T_POW) reduce 399) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("[") reduce 399) (("^") reduce 399) (("|") reduce 399)) 20819 ((("%") reduce 577) (("(") reduce 577) (("*") reduce 577) (("+") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AS) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("|") reduce 577)) 20820 ((("%") reduce 401) (("(") reduce 401) (("*") reduce 401) (("+") reduce 401) (("-") reduce 401) ((".") reduce 401) (("/") reduce 401) (("<") reduce 401) ((">") reduce 401) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AS) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_INSTANCEOF) reduce 401) ((T_IS_EQUAL) reduce 401) ((T_IS_GREATER_OR_EQUAL) reduce 401) ((T_IS_IDENTICAL) reduce 401) ((T_IS_NOT_EQUAL) reduce 401) ((T_IS_NOT_IDENTICAL) reduce 401) ((T_IS_SMALLER_OR_EQUAL) reduce 401) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 401) ((T_OBJECT_OPERATOR) reduce 401) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 401) ((T_PIPE) reduce 401) ((T_POW) reduce 401) ((T_SL) reduce 401) ((T_SPACESHIP) reduce 401) ((T_SR) reduce 401) (("[") reduce 401) (("^") reduce 401) (("|") reduce 401)) 20828 ((("%") reduce 398) (("(") reduce 398) (("*") reduce 398) (("+") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AS) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 398) ((T_OBJECT_OPERATOR) reduce 398) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 398) ((T_PIPE) reduce 398) ((T_POW) reduce 398) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("[") reduce 398) (("^") reduce 398) (("|") reduce 398)) 20830 ((("%") reduce 513) (("(") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AS) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("|") reduce 513)) 20836 ((("%") reduce 233) (("*") reduce 233) (("+") reduce 233) (("-") reduce 233) ((".") reduce 233) (("/") reduce 233) (("<") reduce 233) ((">") reduce 233) (("?") reduce 233) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_AS) reduce 233) ((T_BOOLEAN_AND) reduce 233) ((T_BOOLEAN_OR) reduce 233) ((T_COALESCE) reduce 233) ((T_INSTANCEOF) reduce 233) ((T_IS_EQUAL) reduce 233) ((T_IS_GREATER_OR_EQUAL) reduce 233) ((T_IS_IDENTICAL) reduce 233) ((T_IS_NOT_EQUAL) reduce 233) ((T_IS_NOT_IDENTICAL) reduce 233) ((T_IS_SMALLER_OR_EQUAL) reduce 233) ((T_LOGICAL_AND) reduce 233) ((T_LOGICAL_OR) reduce 233) ((T_LOGICAL_XOR) reduce 233) ((T_PIPE) reduce 233) ((T_POW) reduce 233) ((T_SL) reduce 233) ((T_SPACESHIP) reduce 233) ((T_SR) reduce 233) (("^") reduce 233) (("|") reduce 233)) 20841 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 404) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20845 ((("%") reduce 613) (("*") reduce 613) (("+") reduce 613) (("-") reduce 613) ((".") reduce 613) (("/") reduce 613) (("<") reduce 613) ((">") reduce 613) (("?") reduce 613) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_AS) reduce 613) ((T_BOOLEAN_AND) reduce 613) ((T_BOOLEAN_OR) reduce 613) ((T_COALESCE) reduce 613) ((T_INSTANCEOF) reduce 613) ((T_IS_EQUAL) reduce 613) ((T_IS_GREATER_OR_EQUAL) reduce 613) ((T_IS_IDENTICAL) reduce 613) ((T_IS_NOT_EQUAL) reduce 613) ((T_IS_NOT_IDENTICAL) reduce 613) ((T_IS_SMALLER_OR_EQUAL) reduce 613) ((T_LOGICAL_AND) reduce 613) ((T_LOGICAL_OR) reduce 613) ((T_LOGICAL_XOR) reduce 613) ((T_PIPE) reduce 613) ((T_POW) reduce 613) ((T_SL) reduce 613) ((T_SPACESHIP) reduce 613) ((T_SR) reduce 613) (("^") reduce 613) (("|") reduce 613)) 20846 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AS) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_PIPE) reduce 467) ((T_POW) shift) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("|") reduce 467)) 20847 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 616) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20848 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AS) reduce 615) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 20849 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AS) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("[") reduce 548) (("^") reduce 424) (("|") reduce 424)) 20851 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AS) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_PIPE) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("^") reduce 518) (("|") reduce 518)) 20852 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AS) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_PIPE) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("|") reduce 474)) 20853 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AS) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("|") reduce 291)) 20857 ((("%") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AS) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("^") reduce 292) (("|") reduce 292)) 20858 ((("%") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AS) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("^") reduce 293) (("|") reduce 293)) 20861 ((("%") reduce 617) (("*") reduce 617) (("+") reduce 617) (("-") reduce 617) ((".") reduce 617) (("/") reduce 617) (("<") reduce 617) ((">") reduce 617) (("?") reduce 617) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_AS) reduce 617) ((T_BOOLEAN_AND) reduce 617) ((T_BOOLEAN_OR) reduce 617) ((T_COALESCE) reduce 617) ((T_INSTANCEOF) reduce 617) ((T_IS_EQUAL) reduce 617) ((T_IS_GREATER_OR_EQUAL) reduce 617) ((T_IS_IDENTICAL) reduce 617) ((T_IS_NOT_EQUAL) reduce 617) ((T_IS_NOT_IDENTICAL) reduce 617) ((T_IS_SMALLER_OR_EQUAL) reduce 617) ((T_LOGICAL_AND) reduce 617) ((T_LOGICAL_OR) reduce 617) ((T_LOGICAL_XOR) reduce 617) ((T_PIPE) reduce 617) ((T_POW) reduce 617) ((T_SL) reduce 617) ((T_SPACESHIP) reduce 617) ((T_SR) reduce 617) (("^") reduce 617) (("|") reduce 617)) 20864 ((("%") reduce 614) (("*") reduce 614) (("+") reduce 614) (("-") reduce 614) ((".") reduce 614) (("/") reduce 614) (("<") reduce 614) ((">") reduce 614) (("?") reduce 614) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_AS) reduce 614) ((T_BOOLEAN_AND) reduce 614) ((T_BOOLEAN_OR) reduce 614) ((T_COALESCE) reduce 614) ((T_INSTANCEOF) reduce 614) ((T_IS_EQUAL) reduce 614) ((T_IS_GREATER_OR_EQUAL) reduce 614) ((T_IS_IDENTICAL) reduce 614) ((T_IS_NOT_EQUAL) reduce 614) ((T_IS_NOT_IDENTICAL) reduce 614) ((T_IS_SMALLER_OR_EQUAL) reduce 614) ((T_LOGICAL_AND) reduce 614) ((T_LOGICAL_OR) reduce 614) ((T_LOGICAL_XOR) reduce 614) ((T_PIPE) reduce 614) ((T_POW) reduce 614) ((T_SL) reduce 614) ((T_SPACESHIP) reduce 614) ((T_SR) reduce 614) (("^") reduce 614) (("|") reduce 614)) 20865 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AS) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_PIPE) reduce 468) ((T_POW) shift) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 20866 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AS) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("[") reduce 548) (("^") reduce 426) (("|") reduce 426)) 20868 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AS) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_PIPE) reduce 408) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("^") reduce 408) (("|") reduce 408)) 20869 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AS) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_PIPE) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("|") reduce 409)) 20870 ((("%") reduce 296) (("*") reduce 296) (("+") reduce 296) (("-") reduce 296) ((".") reduce 296) (("/") reduce 296) (("<") reduce 296) ((">") reduce 296) (("?") reduce 296) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_AS) reduce 296) ((T_BOOLEAN_AND) reduce 296) ((T_BOOLEAN_OR) reduce 296) ((T_COALESCE) reduce 296) ((T_INSTANCEOF) reduce 296) ((T_IS_EQUAL) reduce 296) ((T_IS_GREATER_OR_EQUAL) reduce 296) ((T_IS_IDENTICAL) reduce 296) ((T_IS_NOT_EQUAL) reduce 296) ((T_IS_NOT_IDENTICAL) reduce 296) ((T_IS_SMALLER_OR_EQUAL) reduce 296) ((T_LOGICAL_AND) reduce 296) ((T_LOGICAL_OR) reduce 296) ((T_LOGICAL_XOR) reduce 296) ((T_PIPE) reduce 296) ((T_POW) reduce 296) ((T_SL) reduce 296) ((T_SPACESHIP) reduce 296) ((T_SR) reduce 296) (("^") reduce 296) (("|") reduce 296)) 20875 ((("%") reduce 297) (("*") reduce 297) (("+") reduce 297) (("-") reduce 297) ((".") reduce 297) (("/") reduce 297) (("<") reduce 297) ((">") reduce 297) (("?") reduce 297) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_AS) reduce 297) ((T_BOOLEAN_AND) reduce 297) ((T_BOOLEAN_OR) reduce 297) ((T_COALESCE) reduce 297) ((T_INSTANCEOF) reduce 297) ((T_IS_EQUAL) reduce 297) ((T_IS_GREATER_OR_EQUAL) reduce 297) ((T_IS_IDENTICAL) reduce 297) ((T_IS_NOT_EQUAL) reduce 297) ((T_IS_NOT_IDENTICAL) reduce 297) ((T_IS_SMALLER_OR_EQUAL) reduce 297) ((T_LOGICAL_AND) reduce 297) ((T_LOGICAL_OR) reduce 297) ((T_LOGICAL_XOR) reduce 297) ((T_PIPE) reduce 297) ((T_POW) reduce 297) ((T_SL) reduce 297) ((T_SPACESHIP) reduce 297) ((T_SR) reduce 297) (("^") reduce 297) (("|") reduce 297)) 20876 ((("%") reduce 460) (("(") reduce 556) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AS) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PIPE) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("[") reduce 549) (("^") reduce 460) (("|") reduce 460)) 20878 ((("%") reduce 298) (("*") reduce 298) (("+") reduce 298) (("-") reduce 298) ((".") reduce 298) (("/") reduce 298) (("<") reduce 298) ((">") reduce 298) (("?") reduce 298) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_AS) reduce 298) ((T_BOOLEAN_AND) reduce 298) ((T_BOOLEAN_OR) reduce 298) ((T_COALESCE) reduce 298) ((T_INSTANCEOF) reduce 298) ((T_IS_EQUAL) reduce 298) ((T_IS_GREATER_OR_EQUAL) reduce 298) ((T_IS_IDENTICAL) reduce 298) ((T_IS_NOT_EQUAL) reduce 298) ((T_IS_NOT_IDENTICAL) reduce 298) ((T_IS_SMALLER_OR_EQUAL) reduce 298) ((T_LOGICAL_AND) reduce 298) ((T_LOGICAL_OR) reduce 298) ((T_LOGICAL_XOR) reduce 298) ((T_PIPE) reduce 298) ((T_POW) reduce 298) ((T_SL) reduce 298) ((T_SPACESHIP) reduce 298) ((T_SR) reduce 298) (("^") reduce 298) (("|") reduce 298)) 20879 ((("%") reduce 299) (("*") reduce 299) (("+") reduce 299) (("-") reduce 299) ((".") reduce 299) (("/") reduce 299) (("<") reduce 299) ((">") reduce 299) (("?") reduce 299) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_AS) reduce 299) ((T_BOOLEAN_AND) reduce 299) ((T_BOOLEAN_OR) reduce 299) ((T_COALESCE) reduce 299) ((T_INSTANCEOF) reduce 299) ((T_IS_EQUAL) reduce 299) ((T_IS_GREATER_OR_EQUAL) reduce 299) ((T_IS_IDENTICAL) reduce 299) ((T_IS_NOT_EQUAL) reduce 299) ((T_IS_NOT_IDENTICAL) reduce 299) ((T_IS_SMALLER_OR_EQUAL) reduce 299) ((T_LOGICAL_AND) reduce 299) ((T_LOGICAL_OR) reduce 299) ((T_LOGICAL_XOR) reduce 299) ((T_PIPE) reduce 299) ((T_POW) reduce 299) ((T_SL) reduce 299) ((T_SPACESHIP) reduce 299) ((T_SR) reduce 299) (("^") reduce 299) (("|") reduce 299)) 20880 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AS) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_PIPE) reduce 472) ((T_POW) shift) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("|") reduce 472)) 20881 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AS) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_PIPE) reduce 470) ((T_POW) shift) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("|") reduce 470)) 20884 ((("%") reduce 519) (("(") reduce 519) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AS) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PIPE) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("|") reduce 519)) 20885 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AS) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_PIPE) reduce 475) ((T_POW) shift) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("|") reduce 475)) 20886 ((("%") reduce 446) (("*") reduce 446) (("+") reduce 446) (("-") reduce 446) ((".") reduce 446) (("/") reduce 446) (("<") reduce 446) ((">") reduce 446) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_AS) reduce 446) ((T_BOOLEAN_AND) reduce 446) ((T_BOOLEAN_OR) reduce 446) ((T_COALESCE) reduce 446) ((T_INSTANCEOF) reduce 446) ((T_IS_EQUAL) reduce 446) ((T_IS_GREATER_OR_EQUAL) reduce 446) ((T_IS_IDENTICAL) reduce 446) ((T_IS_NOT_EQUAL) reduce 446) ((T_IS_NOT_IDENTICAL) reduce 446) ((T_IS_SMALLER_OR_EQUAL) reduce 446) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_PIPE) reduce 446) ((T_POW) shift) ((T_SL) reduce 446) ((T_SPACESHIP) reduce 446) ((T_SR) reduce 446) (("^") reduce 446) (("|") reduce 446)) 20887 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AS) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_PIPE) reduce 445) ((T_POW) shift) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("|") reduce 445)) 20889 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) (("=") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AND_EQUAL) reduce 570) ((T_AS) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_COALESCE_EQUAL) reduce 570) ((T_CONCAT_EQUAL) reduce 570) ((T_DEC) reduce 570) ((T_DIV_EQUAL) reduce 570) ((T_INC) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_MINUS_EQUAL) reduce 570) ((T_MOD_EQUAL) reduce 570) ((T_MUL_EQUAL) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_OR_EQUAL) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_PLUS_EQUAL) reduce 570) ((T_POW) reduce 570) ((T_POW_EQUAL) reduce 570) ((T_SL) reduce 570) ((T_SL_EQUAL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) ((T_SR_EQUAL) reduce 570) ((T_XOR_EQUAL) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 20892 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) (("=") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AND_EQUAL) reduce 569) ((T_AS) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_COALESCE_EQUAL) reduce 569) ((T_CONCAT_EQUAL) reduce 569) ((T_DEC) reduce 569) ((T_DIV_EQUAL) reduce 569) ((T_INC) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_MINUS_EQUAL) reduce 569) ((T_MOD_EQUAL) reduce 569) ((T_MUL_EQUAL) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_OR_EQUAL) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_PLUS_EQUAL) reduce 569) ((T_POW) reduce 569) ((T_POW_EQUAL) reduce 569) ((T_SL) reduce 569) ((T_SL_EQUAL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) ((T_SR_EQUAL) reduce 569) ((T_XOR_EQUAL) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 20894 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AS) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_PIPE) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("|") reduce 522)) 20895 ((("%") reduce 447) (("*") reduce 447) (("+") reduce 447) (("-") reduce 447) ((".") reduce 447) (("/") reduce 447) (("<") reduce 447) ((">") reduce 447) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_AS) reduce 447) ((T_BOOLEAN_AND) reduce 447) ((T_BOOLEAN_OR) reduce 447) ((T_COALESCE) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 447) ((T_IS_GREATER_OR_EQUAL) reduce 447) ((T_IS_IDENTICAL) reduce 447) ((T_IS_NOT_EQUAL) reduce 447) ((T_IS_NOT_IDENTICAL) reduce 447) ((T_IS_SMALLER_OR_EQUAL) reduce 447) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_PIPE) reduce 447) ((T_POW) shift) ((T_SL) reduce 447) ((T_SPACESHIP) reduce 447) ((T_SR) reduce 447) (("^") reduce 447) (("|") reduce 447)) 20896 ((("\"") shift) (("$") shift) (("(") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_ARRAY) shift) ((T_CLASS_C) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DIR) shift) ((T_FILE) shift) ((T_FUNC_C) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_TRAIT_C) shift) ((T_VARIABLE) shift) (("[") shift)) 20900 ((("(") reduce 568) ((")") reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) (("[") reduce 568)) 20905 ((("(") reduce 555) ((")") reduce 564) ((T_DOUBLE_ARROW) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) (("[") reduce 564)) 20907 (((")") shift) ((T_DOUBLE_ARROW) shift)) 20908 ((("(") reduce 563) ((")") reduce 563) ((T_DOUBLE_ARROW) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) (("[") reduce 563)) 20910 ((("(") reduce 559) ((")") reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) (("[") reduce 559)) 20911 (((")") reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) (("[") reduce 565)) 20912 (((")") reduce 214) ((T_DOUBLE_ARROW) reduce 214) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) (("[") reduce 548)) 20916 ((("(") reduce 581) ((")") reduce 572) ((T_DOUBLE_ARROW) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) (("[") reduce 572)) 20918 ((("(") reduce 507) ((")") reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) (("[") reduce 507)) 20919 ((("(") reduce 291) ((")") reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) (("[") reduce 291)) 20923 ((("(") reduce 292) ((")") reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) (("[") reduce 292)) 20924 ((("(") reduce 293) ((")") reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) (("[") reduce 293)) 20925 ((("(") reduce 504) ((")") reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) (("[") reduce 504)) 20931 ((("(") reduce 568) ((")") reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) (("[") reduce 568)) 20936 ((("(") reduce 555) ((")") reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) (("[") reduce 564)) 20939 ((("(") reduce 563) ((")") reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) (("[") reduce 563)) 20941 ((("(") reduce 559) ((")") reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) (("[") reduce 559)) 20942 (((")") reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) (("[") reduce 565)) 20943 (((")") reduce 214) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) (("[") reduce 548)) 20947 ((("(") reduce 581) ((")") reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) (("[") reduce 572)) 20949 ((("(") reduce 507) ((")") reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) (("[") reduce 507)) 20950 ((("(") reduce 291) ((")") reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) (("[") reduce 291)) 20954 ((("(") reduce 292) ((")") reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) (("[") reduce 292)) 20955 ((("(") reduce 293) ((")") reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) (("[") reduce 293)) 20956 ((("(") reduce 504) ((")") reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) (("[") reduce 504)) 20958 ((("!") reduce 142) (("\"") reduce 142) (("$") reduce 142) (("(") reduce 142) (("+") reduce 142) (("-") reduce 142) ((";") reduce 142) (("@") reduce 142) ((T_ABSTRACT) reduce 142) ((T_ARRAY) reduce 142) ((T_ARRAY_CAST) reduce 142) ((T_ATTRIBUTE) reduce 142) ((T_BOOL_CAST) reduce 142) ((T_BREAK) reduce 142) ((T_CLASS) reduce 142) ((T_CLASS_C) reduce 142) ((T_CLONE) reduce 142) ((T_CONSTANT_ENCAPSED_STRING) reduce 142) ((T_CONTINUE) reduce 142) ((T_DEC) reduce 142) ((T_DECLARE) reduce 142) ((T_DIR) reduce 142) ((T_DNUMBER) reduce 142) ((T_DO) reduce 142) ((T_DOUBLE_CAST) reduce 142) ((T_ECHO) reduce 142) ((T_EMPTY) reduce 142) ((T_ENDFOREACH) reduce 142) ((T_ENUM) reduce 142) ((T_EVAL) reduce 142) ((T_EXIT) reduce 142) ((T_FILE) reduce 142) ((T_FINAL) reduce 142) ((T_FN) reduce 142) ((T_FOR) reduce 142) ((T_FOREACH) reduce 142) ((T_FUNCTION) reduce 142) ((T_FUNC_C) reduce 142) ((T_GLOBAL) reduce 142) ((T_GOTO) reduce 142) ((T_HALT_COMPILER) reduce 142) ((T_IF) reduce 142) ((T_INC) reduce 142) ((T_INCLUDE) reduce 142) ((T_INCLUDE_ONCE) reduce 142) ((T_INLINE_HTML) reduce 142) ((T_INTERFACE) reduce 142) ((T_INT_CAST) reduce 142) ((T_ISSET) reduce 142) ((T_LINE) reduce 142) ((T_LIST) reduce 142) ((T_LNUMBER) reduce 142) ((T_MATCH) reduce 142) ((T_METHOD_C) reduce 142) ((T_NAME_FULLY_QUALIFIED) reduce 142) ((T_NAME_QUALIFIED) reduce 142) ((T_NAME_RELATIVE) reduce 142) ((T_NEW) reduce 142) ((T_NS_C) reduce 142) ((T_OBJECT_CAST) reduce 142) ((T_PRINT) reduce 142) ((T_PROPERTY_C) reduce 142) ((T_READONLY) reduce 142) ((T_REQUIRE) reduce 142) ((T_REQUIRE_ONCE) reduce 142) ((T_RETURN) reduce 142) ((T_START_HEREDOC) reduce 142) ((T_STATIC) reduce 142) ((T_STRING) reduce 142) ((T_STRING_CAST) reduce 142) ((T_SWITCH) reduce 142) ((T_THROW) reduce 142) ((T_TRAIT) reduce 142) ((T_TRAIT_C) reduce 142) ((T_TRY) reduce 142) ((T_UNSET) reduce 142) ((T_UNSET_CAST) reduce 142) ((T_VARIABLE) reduce 142) ((T_VOID_CAST) reduce 142) ((T_WHILE) reduce 142) ((T_YIELD) reduce 142) ((T_YIELD_FROM) reduce 142) (("[") reduce 142) (("`") reduce 142) (("{") reduce 142) (("~") reduce 142)) 20959 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CASE) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DEFAULT) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_ELSE) reduce 164) ((T_ELSEIF) reduce 164) ((T_EMPTY) reduce 164) ((T_ENDSWITCH) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_VOID_CAST) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 20960 ((("!") reduce 220) (("\"") reduce 220) (("$") reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CASE) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DEFAULT) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_ELSE) reduce 220) ((T_ELSEIF) reduce 220) ((T_EMPTY) reduce 220) ((T_ENDSWITCH) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_VARIABLE) reduce 220) ((T_VOID_CAST) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("~") reduce 220)) 20961 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_ENDFOREACH) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_VOID_CAST) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("~") shift)) 20962 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_EMPTY) reduce 166) ((T_ENDFOREACH) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_VOID_CAST) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 20977 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_EMPTY) reduce 160) ((T_ENDFOREACH) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_VOID_CAST) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 20988 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_EMPTY) reduce 149) ((T_ENDFOREACH) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_VOID_CAST) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 20990 ((("!") reduce 144) (("\"") reduce 144) (("$") reduce 144) (("(") reduce 144) (("+") reduce 144) (("-") reduce 144) ((";") reduce 144) (("@") reduce 144) ((T_ABSTRACT) reduce 144) ((T_ARRAY) reduce 144) ((T_ARRAY_CAST) reduce 144) ((T_ATTRIBUTE) reduce 144) ((T_BOOL_CAST) reduce 144) ((T_BREAK) reduce 144) ((T_CLASS) reduce 144) ((T_CLASS_C) reduce 144) ((T_CLONE) reduce 144) ((T_CONSTANT_ENCAPSED_STRING) reduce 144) ((T_CONTINUE) reduce 144) ((T_DEC) reduce 144) ((T_DECLARE) reduce 144) ((T_DIR) reduce 144) ((T_DNUMBER) reduce 144) ((T_DO) reduce 144) ((T_DOUBLE_CAST) reduce 144) ((T_ECHO) reduce 144) ((T_EMPTY) reduce 144) ((T_ENDFOREACH) reduce 144) ((T_ENUM) reduce 144) ((T_EVAL) reduce 144) ((T_EXIT) reduce 144) ((T_FILE) reduce 144) ((T_FINAL) reduce 144) ((T_FN) reduce 144) ((T_FOR) reduce 144) ((T_FOREACH) reduce 144) ((T_FUNCTION) reduce 144) ((T_FUNC_C) reduce 144) ((T_GLOBAL) reduce 144) ((T_GOTO) reduce 144) ((T_HALT_COMPILER) reduce 144) ((T_IF) reduce 144) ((T_INC) reduce 144) ((T_INCLUDE) reduce 144) ((T_INCLUDE_ONCE) reduce 144) ((T_INLINE_HTML) reduce 144) ((T_INTERFACE) reduce 144) ((T_INT_CAST) reduce 144) ((T_ISSET) reduce 144) ((T_LINE) reduce 144) ((T_LIST) reduce 144) ((T_LNUMBER) reduce 144) ((T_MATCH) reduce 144) ((T_METHOD_C) reduce 144) ((T_NAME_FULLY_QUALIFIED) reduce 144) ((T_NAME_QUALIFIED) reduce 144) ((T_NAME_RELATIVE) reduce 144) ((T_NEW) reduce 144) ((T_NS_C) reduce 144) ((T_OBJECT_CAST) reduce 144) ((T_PRINT) reduce 144) ((T_PROPERTY_C) reduce 144) ((T_READONLY) reduce 144) ((T_REQUIRE) reduce 144) ((T_REQUIRE_ONCE) reduce 144) ((T_RETURN) reduce 144) ((T_START_HEREDOC) reduce 144) ((T_STATIC) reduce 144) ((T_STRING) reduce 144) ((T_STRING_CAST) reduce 144) ((T_SWITCH) reduce 144) ((T_THROW) reduce 144) ((T_TRAIT) reduce 144) ((T_TRAIT_C) reduce 144) ((T_TRY) reduce 144) ((T_UNSET) reduce 144) ((T_UNSET_CAST) reduce 144) ((T_VARIABLE) reduce 144) ((T_VOID_CAST) reduce 144) ((T_WHILE) reduce 144) ((T_YIELD) reduce 144) ((T_YIELD_FROM) reduce 144) (("[") reduce 144) (("`") reduce 144) (("{") reduce 144) (("~") reduce 144)) 20992 ((("!") reduce 104) (("\"") reduce 104) (("$") reduce 104) (("(") reduce 104) (("+") reduce 104) (("-") reduce 104) ((";") reduce 104) (("@") reduce 104) ((T_ABSTRACT) reduce 104) ((T_ARRAY) reduce 104) ((T_ARRAY_CAST) reduce 104) ((T_ATTRIBUTE) reduce 104) ((T_BOOL_CAST) reduce 104) ((T_BREAK) reduce 104) ((T_CLASS) reduce 104) ((T_CLASS_C) reduce 104) ((T_CLONE) reduce 104) ((T_CONSTANT_ENCAPSED_STRING) reduce 104) ((T_CONTINUE) reduce 104) ((T_DEC) reduce 104) ((T_DECLARE) reduce 104) ((T_DIR) reduce 104) ((T_DNUMBER) reduce 104) ((T_DO) reduce 104) ((T_DOUBLE_CAST) reduce 104) ((T_ECHO) reduce 104) ((T_EMPTY) reduce 104) ((T_ENDFOREACH) reduce 104) ((T_ENUM) reduce 104) ((T_EVAL) reduce 104) ((T_EXIT) reduce 104) ((T_FILE) reduce 104) ((T_FINAL) reduce 104) ((T_FN) reduce 104) ((T_FOR) reduce 104) ((T_FOREACH) reduce 104) ((T_FUNCTION) reduce 104) ((T_FUNC_C) reduce 104) ((T_GLOBAL) reduce 104) ((T_GOTO) reduce 104) ((T_HALT_COMPILER) reduce 104) ((T_IF) reduce 104) ((T_INC) reduce 104) ((T_INCLUDE) reduce 104) ((T_INCLUDE_ONCE) reduce 104) ((T_INLINE_HTML) reduce 104) ((T_INTERFACE) reduce 104) ((T_INT_CAST) reduce 104) ((T_ISSET) reduce 104) ((T_LINE) reduce 104) ((T_LIST) reduce 104) ((T_LNUMBER) reduce 104) ((T_MATCH) reduce 104) ((T_METHOD_C) reduce 104) ((T_NAME_FULLY_QUALIFIED) reduce 104) ((T_NAME_QUALIFIED) reduce 104) ((T_NAME_RELATIVE) reduce 104) ((T_NEW) reduce 104) ((T_NS_C) reduce 104) ((T_OBJECT_CAST) reduce 104) ((T_PRINT) reduce 104) ((T_PROPERTY_C) reduce 104) ((T_READONLY) reduce 104) ((T_REQUIRE) reduce 104) ((T_REQUIRE_ONCE) reduce 104) ((T_RETURN) reduce 104) ((T_START_HEREDOC) reduce 104) ((T_STATIC) reduce 104) ((T_STRING) reduce 104) ((T_STRING_CAST) reduce 104) ((T_SWITCH) reduce 104) ((T_THROW) reduce 104) ((T_TRAIT) reduce 104) ((T_TRAIT_C) reduce 104) ((T_TRY) reduce 104) ((T_UNSET) reduce 104) ((T_UNSET_CAST) reduce 104) ((T_VARIABLE) reduce 104) ((T_VOID_CAST) reduce 104) ((T_WHILE) reduce 104) ((T_YIELD) reduce 104) ((T_YIELD_FROM) reduce 104) (("[") reduce 104) (("`") reduce 104) (("{") reduce 104) (("~") reduce 104)) 20994 ((("!") reduce 107) (("\"") reduce 107) (("$") reduce 107) (("(") reduce 107) (("+") reduce 107) (("-") reduce 107) ((";") reduce 107) (("@") reduce 107) ((T_ABSTRACT) reduce 107) ((T_ARRAY) reduce 107) ((T_ARRAY_CAST) reduce 107) ((T_ATTRIBUTE) reduce 107) ((T_BOOL_CAST) reduce 107) ((T_BREAK) reduce 107) ((T_CLASS) reduce 107) ((T_CLASS_C) reduce 107) ((T_CLONE) reduce 107) ((T_CONSTANT_ENCAPSED_STRING) reduce 107) ((T_CONTINUE) reduce 107) ((T_DEC) reduce 107) ((T_DECLARE) reduce 107) ((T_DIR) reduce 107) ((T_DNUMBER) reduce 107) ((T_DO) reduce 107) ((T_DOUBLE_CAST) reduce 107) ((T_ECHO) reduce 107) ((T_EMPTY) reduce 107) ((T_ENDFOREACH) reduce 107) ((T_ENUM) reduce 107) ((T_EVAL) reduce 107) ((T_EXIT) reduce 107) ((T_FILE) reduce 107) ((T_FINAL) reduce 107) ((T_FN) reduce 107) ((T_FOR) reduce 107) ((T_FOREACH) reduce 107) ((T_FUNCTION) reduce 107) ((T_FUNC_C) reduce 107) ((T_GLOBAL) reduce 107) ((T_GOTO) reduce 107) ((T_HALT_COMPILER) reduce 107) ((T_IF) reduce 107) ((T_INC) reduce 107) ((T_INCLUDE) reduce 107) ((T_INCLUDE_ONCE) reduce 107) ((T_INLINE_HTML) reduce 107) ((T_INTERFACE) reduce 107) ((T_INT_CAST) reduce 107) ((T_ISSET) reduce 107) ((T_LINE) reduce 107) ((T_LIST) reduce 107) ((T_LNUMBER) reduce 107) ((T_MATCH) reduce 107) ((T_METHOD_C) reduce 107) ((T_NAME_FULLY_QUALIFIED) reduce 107) ((T_NAME_QUALIFIED) reduce 107) ((T_NAME_RELATIVE) reduce 107) ((T_NEW) reduce 107) ((T_NS_C) reduce 107) ((T_OBJECT_CAST) reduce 107) ((T_PRINT) reduce 107) ((T_PROPERTY_C) reduce 107) ((T_READONLY) reduce 107) ((T_REQUIRE) reduce 107) ((T_REQUIRE_ONCE) reduce 107) ((T_RETURN) reduce 107) ((T_START_HEREDOC) reduce 107) ((T_STATIC) reduce 107) ((T_STRING) reduce 107) ((T_STRING_CAST) reduce 107) ((T_SWITCH) reduce 107) ((T_THROW) reduce 107) ((T_TRAIT) reduce 107) ((T_TRAIT_C) reduce 107) ((T_TRY) reduce 107) ((T_UNSET) reduce 107) ((T_UNSET_CAST) reduce 107) ((T_VARIABLE) reduce 107) ((T_VOID_CAST) reduce 107) ((T_WHILE) reduce 107) ((T_YIELD) reduce 107) ((T_YIELD_FROM) reduce 107) (("[") reduce 107) (("`") reduce 107) (("{") reduce 107) (("~") reduce 107)) 20997 ((("!") reduce 103) (("\"") reduce 103) (("$") reduce 103) (("(") reduce 103) (("+") reduce 103) (("-") reduce 103) ((";") reduce 103) (("@") reduce 103) ((T_ABSTRACT) reduce 103) ((T_ARRAY) reduce 103) ((T_ARRAY_CAST) reduce 103) ((T_ATTRIBUTE) reduce 103) ((T_BOOL_CAST) reduce 103) ((T_BREAK) reduce 103) ((T_CLASS) reduce 103) ((T_CLASS_C) reduce 103) ((T_CLONE) reduce 103) ((T_CONSTANT_ENCAPSED_STRING) reduce 103) ((T_CONTINUE) reduce 103) ((T_DEC) reduce 103) ((T_DECLARE) reduce 103) ((T_DIR) reduce 103) ((T_DNUMBER) reduce 103) ((T_DO) reduce 103) ((T_DOUBLE_CAST) reduce 103) ((T_ECHO) reduce 103) ((T_EMPTY) reduce 103) ((T_ENDFOREACH) reduce 103) ((T_ENUM) reduce 103) ((T_EVAL) reduce 103) ((T_EXIT) reduce 103) ((T_FILE) reduce 103) ((T_FINAL) reduce 103) ((T_FN) reduce 103) ((T_FOR) reduce 103) ((T_FOREACH) reduce 103) ((T_FUNCTION) reduce 103) ((T_FUNC_C) reduce 103) ((T_GLOBAL) reduce 103) ((T_GOTO) reduce 103) ((T_HALT_COMPILER) reduce 103) ((T_IF) reduce 103) ((T_INC) reduce 103) ((T_INCLUDE) reduce 103) ((T_INCLUDE_ONCE) reduce 103) ((T_INLINE_HTML) reduce 103) ((T_INTERFACE) reduce 103) ((T_INT_CAST) reduce 103) ((T_ISSET) reduce 103) ((T_LINE) reduce 103) ((T_LIST) reduce 103) ((T_LNUMBER) reduce 103) ((T_MATCH) reduce 103) ((T_METHOD_C) reduce 103) ((T_NAME_FULLY_QUALIFIED) reduce 103) ((T_NAME_QUALIFIED) reduce 103) ((T_NAME_RELATIVE) reduce 103) ((T_NEW) reduce 103) ((T_NS_C) reduce 103) ((T_OBJECT_CAST) reduce 103) ((T_PRINT) reduce 103) ((T_PROPERTY_C) reduce 103) ((T_READONLY) reduce 103) ((T_REQUIRE) reduce 103) ((T_REQUIRE_ONCE) reduce 103) ((T_RETURN) reduce 103) ((T_START_HEREDOC) reduce 103) ((T_STATIC) reduce 103) ((T_STRING) reduce 103) ((T_STRING_CAST) reduce 103) ((T_SWITCH) reduce 103) ((T_THROW) reduce 103) ((T_TRAIT) reduce 103) ((T_TRAIT_C) reduce 103) ((T_TRY) reduce 103) ((T_UNSET) reduce 103) ((T_UNSET_CAST) reduce 103) ((T_VARIABLE) reduce 103) ((T_VOID_CAST) reduce 103) ((T_WHILE) reduce 103) ((T_YIELD) reduce 103) ((T_YIELD_FROM) reduce 103) (("[") reduce 103) (("`") reduce 103) (("{") reduce 103) (("~") reduce 103)) 20998 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_EMPTY) reduce 148) ((T_ENDFOREACH) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_VOID_CAST) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 20999 ((("!") reduce 246) (("\"") reduce 246) (("$") reduce 246) (("(") reduce 246) (("+") reduce 246) (("-") reduce 246) ((";") reduce 246) (("@") reduce 246) ((T_ABSTRACT) reduce 246) ((T_ARRAY) reduce 246) ((T_ARRAY_CAST) reduce 246) ((T_ATTRIBUTE) reduce 246) ((T_BOOL_CAST) reduce 246) ((T_BREAK) reduce 246) ((T_CLASS) reduce 246) ((T_CLASS_C) reduce 246) ((T_CLONE) reduce 246) ((T_CONSTANT_ENCAPSED_STRING) reduce 246) ((T_CONTINUE) reduce 246) ((T_DEC) reduce 246) ((T_DECLARE) reduce 246) ((T_DIR) reduce 246) ((T_DNUMBER) reduce 246) ((T_DO) reduce 246) ((T_DOUBLE_CAST) reduce 246) ((T_ECHO) reduce 246) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) reduce 246) ((T_ENDFOREACH) reduce 246) ((T_ENUM) reduce 246) ((T_EVAL) reduce 246) ((T_EXIT) reduce 246) ((T_FILE) reduce 246) ((T_FINAL) reduce 246) ((T_FN) reduce 246) ((T_FOR) reduce 246) ((T_FOREACH) reduce 246) ((T_FUNCTION) reduce 246) ((T_FUNC_C) reduce 246) ((T_GLOBAL) reduce 246) ((T_GOTO) reduce 246) ((T_HALT_COMPILER) reduce 246) ((T_IF) reduce 246) ((T_INC) reduce 246) ((T_INCLUDE) reduce 246) ((T_INCLUDE_ONCE) reduce 246) ((T_INLINE_HTML) reduce 246) ((T_INTERFACE) reduce 246) ((T_INT_CAST) reduce 246) ((T_ISSET) reduce 246) ((T_LINE) reduce 246) ((T_LIST) reduce 246) ((T_LNUMBER) reduce 246) ((T_MATCH) reduce 246) ((T_METHOD_C) reduce 246) ((T_NAME_FULLY_QUALIFIED) reduce 246) ((T_NAME_QUALIFIED) reduce 246) ((T_NAME_RELATIVE) reduce 246) ((T_NEW) reduce 246) ((T_NS_C) reduce 246) ((T_OBJECT_CAST) reduce 246) ((T_PRINT) reduce 246) ((T_PROPERTY_C) reduce 246) ((T_READONLY) reduce 246) ((T_REQUIRE) reduce 246) ((T_REQUIRE_ONCE) reduce 246) ((T_RETURN) reduce 246) ((T_START_HEREDOC) reduce 246) ((T_STATIC) reduce 246) ((T_STRING) reduce 246) ((T_STRING_CAST) reduce 246) ((T_SWITCH) reduce 246) ((T_THROW) reduce 246) ((T_TRAIT) reduce 246) ((T_TRAIT_C) reduce 246) ((T_TRY) reduce 246) ((T_UNSET) reduce 246) ((T_UNSET_CAST) reduce 246) ((T_VARIABLE) reduce 246) ((T_VOID_CAST) reduce 246) ((T_WHILE) reduce 246) ((T_YIELD) reduce 246) ((T_YIELD_FROM) reduce 246) (("[") reduce 246) (("`") reduce 246) (("{") reduce 246) (("~") reduce 246)) 21000 ((("!") reduce 141) (("\"") reduce 141) (("$") reduce 141) (("(") reduce 141) (("+") reduce 141) (("-") reduce 141) ((";") reduce 141) (("@") reduce 141) ((T_ABSTRACT) reduce 141) ((T_ARRAY) reduce 141) ((T_ARRAY_CAST) reduce 141) ((T_ATTRIBUTE) reduce 141) ((T_BOOL_CAST) reduce 141) ((T_BREAK) reduce 141) ((T_CLASS) reduce 141) ((T_CLASS_C) reduce 141) ((T_CLONE) reduce 141) ((T_CONSTANT_ENCAPSED_STRING) reduce 141) ((T_CONTINUE) reduce 141) ((T_DEC) reduce 141) ((T_DECLARE) reduce 141) ((T_DIR) reduce 141) ((T_DNUMBER) reduce 141) ((T_DO) reduce 141) ((T_DOUBLE_CAST) reduce 141) ((T_ECHO) reduce 141) ((T_EMPTY) reduce 141) ((T_ENDFOREACH) reduce 141) ((T_ENUM) reduce 141) ((T_EVAL) reduce 141) ((T_EXIT) reduce 141) ((T_FILE) reduce 141) ((T_FINAL) reduce 141) ((T_FN) reduce 141) ((T_FOR) reduce 141) ((T_FOREACH) reduce 141) ((T_FUNCTION) reduce 141) ((T_FUNC_C) reduce 141) ((T_GLOBAL) reduce 141) ((T_GOTO) reduce 141) ((T_HALT_COMPILER) reduce 141) ((T_IF) reduce 141) ((T_INC) reduce 141) ((T_INCLUDE) reduce 141) ((T_INCLUDE_ONCE) reduce 141) ((T_INLINE_HTML) reduce 141) ((T_INTERFACE) reduce 141) ((T_INT_CAST) reduce 141) ((T_ISSET) reduce 141) ((T_LINE) reduce 141) ((T_LIST) reduce 141) ((T_LNUMBER) reduce 141) ((T_MATCH) reduce 141) ((T_METHOD_C) reduce 141) ((T_NAME_FULLY_QUALIFIED) reduce 141) ((T_NAME_QUALIFIED) reduce 141) ((T_NAME_RELATIVE) reduce 141) ((T_NEW) reduce 141) ((T_NS_C) reduce 141) ((T_OBJECT_CAST) reduce 141) ((T_PRINT) reduce 141) ((T_PROPERTY_C) reduce 141) ((T_READONLY) reduce 141) ((T_REQUIRE) reduce 141) ((T_REQUIRE_ONCE) reduce 141) ((T_RETURN) reduce 141) ((T_START_HEREDOC) reduce 141) ((T_STATIC) reduce 141) ((T_STRING) reduce 141) ((T_STRING_CAST) reduce 141) ((T_SWITCH) reduce 141) ((T_THROW) reduce 141) ((T_TRAIT) reduce 141) ((T_TRAIT_C) reduce 141) ((T_TRY) reduce 141) ((T_UNSET) reduce 141) ((T_UNSET_CAST) reduce 141) ((T_VARIABLE) reduce 141) ((T_VOID_CAST) reduce 141) ((T_WHILE) reduce 141) ((T_YIELD) reduce 141) ((T_YIELD_FROM) reduce 141) (("[") reduce 141) (("`") reduce 141) (("{") reduce 141) (("~") reduce 141)) 21001 ((("!") reduce 106) (("\"") reduce 106) (("$") reduce 106) (("(") reduce 106) (("+") reduce 106) (("-") reduce 106) ((";") reduce 106) (("@") reduce 106) ((T_ABSTRACT) reduce 106) ((T_ARRAY) reduce 106) ((T_ARRAY_CAST) reduce 106) ((T_ATTRIBUTE) reduce 106) ((T_BOOL_CAST) reduce 106) ((T_BREAK) reduce 106) ((T_CLASS) reduce 106) ((T_CLASS_C) reduce 106) ((T_CLONE) reduce 106) ((T_CONSTANT_ENCAPSED_STRING) reduce 106) ((T_CONTINUE) reduce 106) ((T_DEC) reduce 106) ((T_DECLARE) reduce 106) ((T_DIR) reduce 106) ((T_DNUMBER) reduce 106) ((T_DO) reduce 106) ((T_DOUBLE_CAST) reduce 106) ((T_ECHO) reduce 106) ((T_EMPTY) reduce 106) ((T_ENDFOREACH) reduce 106) ((T_ENUM) reduce 106) ((T_EVAL) reduce 106) ((T_EXIT) reduce 106) ((T_FILE) reduce 106) ((T_FINAL) reduce 106) ((T_FN) reduce 106) ((T_FOR) reduce 106) ((T_FOREACH) reduce 106) ((T_FUNCTION) reduce 106) ((T_FUNC_C) reduce 106) ((T_GLOBAL) reduce 106) ((T_GOTO) reduce 106) ((T_HALT_COMPILER) reduce 106) ((T_IF) reduce 106) ((T_INC) reduce 106) ((T_INCLUDE) reduce 106) ((T_INCLUDE_ONCE) reduce 106) ((T_INLINE_HTML) reduce 106) ((T_INTERFACE) reduce 106) ((T_INT_CAST) reduce 106) ((T_ISSET) reduce 106) ((T_LINE) reduce 106) ((T_LIST) reduce 106) ((T_LNUMBER) reduce 106) ((T_MATCH) reduce 106) ((T_METHOD_C) reduce 106) ((T_NAME_FULLY_QUALIFIED) reduce 106) ((T_NAME_QUALIFIED) reduce 106) ((T_NAME_RELATIVE) reduce 106) ((T_NEW) reduce 106) ((T_NS_C) reduce 106) ((T_OBJECT_CAST) reduce 106) ((T_PRINT) reduce 106) ((T_PROPERTY_C) reduce 106) ((T_READONLY) reduce 106) ((T_REQUIRE) reduce 106) ((T_REQUIRE_ONCE) reduce 106) ((T_RETURN) reduce 106) ((T_START_HEREDOC) reduce 106) ((T_STATIC) reduce 106) ((T_STRING) reduce 106) ((T_STRING_CAST) reduce 106) ((T_SWITCH) reduce 106) ((T_THROW) reduce 106) ((T_TRAIT) reduce 106) ((T_TRAIT_C) reduce 106) ((T_TRY) reduce 106) ((T_UNSET) reduce 106) ((T_UNSET_CAST) reduce 106) ((T_VARIABLE) reduce 106) ((T_VOID_CAST) reduce 106) ((T_WHILE) reduce 106) ((T_YIELD) reduce 106) ((T_YIELD_FROM) reduce 106) (("[") reduce 106) (("`") reduce 106) (("{") reduce 106) (("~") reduce 106)) 21002 ((("!") reduce 143) (("\"") reduce 143) (("$") reduce 143) (("(") reduce 143) (("+") reduce 143) (("-") reduce 143) ((";") reduce 143) (("@") reduce 143) ((T_ABSTRACT) reduce 143) ((T_ARRAY) reduce 143) ((T_ARRAY_CAST) reduce 143) ((T_ATTRIBUTE) reduce 143) ((T_BOOL_CAST) reduce 143) ((T_BREAK) reduce 143) ((T_CLASS) reduce 143) ((T_CLASS_C) reduce 143) ((T_CLONE) reduce 143) ((T_CONSTANT_ENCAPSED_STRING) reduce 143) ((T_CONTINUE) reduce 143) ((T_DEC) reduce 143) ((T_DECLARE) reduce 143) ((T_DIR) reduce 143) ((T_DNUMBER) reduce 143) ((T_DO) reduce 143) ((T_DOUBLE_CAST) reduce 143) ((T_ECHO) reduce 143) ((T_EMPTY) reduce 143) ((T_ENDFOREACH) reduce 143) ((T_ENUM) reduce 143) ((T_EVAL) reduce 143) ((T_EXIT) reduce 143) ((T_FILE) reduce 143) ((T_FINAL) reduce 143) ((T_FN) reduce 143) ((T_FOR) reduce 143) ((T_FOREACH) reduce 143) ((T_FUNCTION) reduce 143) ((T_FUNC_C) reduce 143) ((T_GLOBAL) reduce 143) ((T_GOTO) reduce 143) ((T_HALT_COMPILER) reduce 143) ((T_IF) reduce 143) ((T_INC) reduce 143) ((T_INCLUDE) reduce 143) ((T_INCLUDE_ONCE) reduce 143) ((T_INLINE_HTML) reduce 143) ((T_INTERFACE) reduce 143) ((T_INT_CAST) reduce 143) ((T_ISSET) reduce 143) ((T_LINE) reduce 143) ((T_LIST) reduce 143) ((T_LNUMBER) reduce 143) ((T_MATCH) reduce 143) ((T_METHOD_C) reduce 143) ((T_NAME_FULLY_QUALIFIED) reduce 143) ((T_NAME_QUALIFIED) reduce 143) ((T_NAME_RELATIVE) reduce 143) ((T_NEW) reduce 143) ((T_NS_C) reduce 143) ((T_OBJECT_CAST) reduce 143) ((T_PRINT) reduce 143) ((T_PROPERTY_C) reduce 143) ((T_READONLY) reduce 143) ((T_REQUIRE) reduce 143) ((T_REQUIRE_ONCE) reduce 143) ((T_RETURN) reduce 143) ((T_START_HEREDOC) reduce 143) ((T_STATIC) reduce 143) ((T_STRING) reduce 143) ((T_STRING_CAST) reduce 143) ((T_SWITCH) reduce 143) ((T_THROW) reduce 143) ((T_TRAIT) reduce 143) ((T_TRAIT_C) reduce 143) ((T_TRY) reduce 143) ((T_UNSET) reduce 143) ((T_UNSET_CAST) reduce 143) ((T_VARIABLE) reduce 143) ((T_VOID_CAST) reduce 143) ((T_WHILE) reduce 143) ((T_YIELD) reduce 143) ((T_YIELD_FROM) reduce 143) (("[") reduce 143) (("`") reduce 143) (("{") reduce 143) (("~") reduce 143)) 21003 ((("!") reduce 105) (("\"") reduce 105) (("$") reduce 105) (("(") reduce 105) (("+") reduce 105) (("-") reduce 105) ((";") reduce 105) (("@") reduce 105) ((T_ABSTRACT) reduce 105) ((T_ARRAY) reduce 105) ((T_ARRAY_CAST) reduce 105) ((T_ATTRIBUTE) reduce 105) ((T_BOOL_CAST) reduce 105) ((T_BREAK) reduce 105) ((T_CLASS) reduce 105) ((T_CLASS_C) reduce 105) ((T_CLONE) reduce 105) ((T_CONSTANT_ENCAPSED_STRING) reduce 105) ((T_CONTINUE) reduce 105) ((T_DEC) reduce 105) ((T_DECLARE) reduce 105) ((T_DIR) reduce 105) ((T_DNUMBER) reduce 105) ((T_DO) reduce 105) ((T_DOUBLE_CAST) reduce 105) ((T_ECHO) reduce 105) ((T_EMPTY) reduce 105) ((T_ENDFOREACH) reduce 105) ((T_ENUM) reduce 105) ((T_EVAL) reduce 105) ((T_EXIT) reduce 105) ((T_FILE) reduce 105) ((T_FINAL) reduce 105) ((T_FN) reduce 105) ((T_FOR) reduce 105) ((T_FOREACH) reduce 105) ((T_FUNCTION) reduce 105) ((T_FUNC_C) reduce 105) ((T_GLOBAL) reduce 105) ((T_GOTO) reduce 105) ((T_HALT_COMPILER) reduce 105) ((T_IF) reduce 105) ((T_INC) reduce 105) ((T_INCLUDE) reduce 105) ((T_INCLUDE_ONCE) reduce 105) ((T_INLINE_HTML) reduce 105) ((T_INTERFACE) reduce 105) ((T_INT_CAST) reduce 105) ((T_ISSET) reduce 105) ((T_LINE) reduce 105) ((T_LIST) reduce 105) ((T_LNUMBER) reduce 105) ((T_MATCH) reduce 105) ((T_METHOD_C) reduce 105) ((T_NAME_FULLY_QUALIFIED) reduce 105) ((T_NAME_QUALIFIED) reduce 105) ((T_NAME_RELATIVE) reduce 105) ((T_NEW) reduce 105) ((T_NS_C) reduce 105) ((T_OBJECT_CAST) reduce 105) ((T_PRINT) reduce 105) ((T_PROPERTY_C) reduce 105) ((T_READONLY) reduce 105) ((T_REQUIRE) reduce 105) ((T_REQUIRE_ONCE) reduce 105) ((T_RETURN) reduce 105) ((T_START_HEREDOC) reduce 105) ((T_STATIC) reduce 105) ((T_STRING) reduce 105) ((T_STRING_CAST) reduce 105) ((T_SWITCH) reduce 105) ((T_THROW) reduce 105) ((T_TRAIT) reduce 105) ((T_TRAIT_C) reduce 105) ((T_TRY) reduce 105) ((T_UNSET) reduce 105) ((T_UNSET_CAST) reduce 105) ((T_VARIABLE) reduce 105) ((T_VOID_CAST) reduce 105) ((T_WHILE) reduce 105) ((T_YIELD) reduce 105) ((T_YIELD_FROM) reduce 105) (("[") reduce 105) (("`") reduce 105) (("{") reduce 105) (("~") reduce 105)) 21006 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_EMPTY) reduce 147) ((T_ENDFOREACH) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_VOID_CAST) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 21012 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_ELSE) reduce 166) ((T_ELSEIF) reduce 166) ((T_EMPTY) reduce 166) ((T_ENDFOREACH) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_VOID_CAST) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 21022 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_ELSE) reduce 160) ((T_ELSEIF) reduce 160) ((T_EMPTY) reduce 160) ((T_ENDFOREACH) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_VOID_CAST) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 21031 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_ELSE) reduce 149) ((T_ELSEIF) reduce 149) ((T_EMPTY) reduce 149) ((T_ENDFOREACH) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_VOID_CAST) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 21034 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_ELSE) reduce 148) ((T_ELSEIF) reduce 148) ((T_EMPTY) reduce 148) ((T_ENDFOREACH) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_VOID_CAST) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 21036 ((("!") reduce 245) (("\"") reduce 245) (("$") reduce 245) (("(") reduce 245) (("+") reduce 245) (("-") reduce 245) ((";") reduce 245) (("@") reduce 245) ((T_ABSTRACT) reduce 245) ((T_ARRAY) reduce 245) ((T_ARRAY_CAST) reduce 245) ((T_ATTRIBUTE) reduce 245) ((T_BOOL_CAST) reduce 245) ((T_BREAK) reduce 245) ((T_CLASS) reduce 245) ((T_CLASS_C) reduce 245) ((T_CLONE) reduce 245) ((T_CONSTANT_ENCAPSED_STRING) reduce 245) ((T_CONTINUE) reduce 245) ((T_DEC) reduce 245) ((T_DECLARE) reduce 245) ((T_DIR) reduce 245) ((T_DNUMBER) reduce 245) ((T_DO) reduce 245) ((T_DOUBLE_CAST) reduce 245) ((T_ECHO) reduce 245) ((T_ELSE) reduce 245) ((T_ELSEIF) reduce 245) ((T_EMPTY) reduce 245) ((T_ENDFOREACH) reduce 245) ((T_ENUM) reduce 245) ((T_EVAL) reduce 245) ((T_EXIT) reduce 245) ((T_FILE) reduce 245) ((T_FINAL) reduce 245) ((T_FN) reduce 245) ((T_FOR) reduce 245) ((T_FOREACH) reduce 245) ((T_FUNCTION) reduce 245) ((T_FUNC_C) reduce 245) ((T_GLOBAL) reduce 245) ((T_GOTO) reduce 245) ((T_HALT_COMPILER) reduce 245) ((T_IF) reduce 245) ((T_INC) reduce 245) ((T_INCLUDE) reduce 245) ((T_INCLUDE_ONCE) reduce 245) ((T_INLINE_HTML) reduce 245) ((T_INTERFACE) reduce 245) ((T_INT_CAST) reduce 245) ((T_ISSET) reduce 245) ((T_LINE) reduce 245) ((T_LIST) reduce 245) ((T_LNUMBER) reduce 245) ((T_MATCH) reduce 245) ((T_METHOD_C) reduce 245) ((T_NAME_FULLY_QUALIFIED) reduce 245) ((T_NAME_QUALIFIED) reduce 245) ((T_NAME_RELATIVE) reduce 245) ((T_NEW) reduce 245) ((T_NS_C) reduce 245) ((T_OBJECT_CAST) reduce 245) ((T_PRINT) reduce 245) ((T_PROPERTY_C) reduce 245) ((T_READONLY) reduce 245) ((T_REQUIRE) reduce 245) ((T_REQUIRE_ONCE) reduce 245) ((T_RETURN) reduce 245) ((T_START_HEREDOC) reduce 245) ((T_STATIC) reduce 245) ((T_STRING) reduce 245) ((T_STRING_CAST) reduce 245) ((T_SWITCH) reduce 245) ((T_THROW) reduce 245) ((T_TRAIT) reduce 245) ((T_TRAIT_C) reduce 245) ((T_TRY) reduce 245) ((T_UNSET) reduce 245) ((T_UNSET_CAST) reduce 245) ((T_VARIABLE) reduce 245) ((T_VOID_CAST) reduce 245) ((T_WHILE) reduce 245) ((T_YIELD) reduce 245) ((T_YIELD_FROM) reduce 245) (("[") reduce 245) (("`") reduce 245) (("{") reduce 245) (("~") reduce 245)) 21039 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_ELSE) reduce 147) ((T_ELSEIF) reduce 147) ((T_EMPTY) reduce 147) ((T_ENDFOREACH) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_VOID_CAST) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 21041 ((("!") reduce 247) (("\"") reduce 247) (("$") reduce 247) (("(") reduce 247) (("+") reduce 247) (("-") reduce 247) ((";") reduce 247) (("@") reduce 247) ((T_ABSTRACT) reduce 247) ((T_ARRAY) reduce 247) ((T_ARRAY_CAST) reduce 247) ((T_ATTRIBUTE) reduce 247) ((T_BOOL_CAST) reduce 247) ((T_BREAK) reduce 247) ((T_CLASS) reduce 247) ((T_CLASS_C) reduce 247) ((T_CLONE) reduce 247) ((T_CONSTANT_ENCAPSED_STRING) reduce 247) ((T_CONTINUE) reduce 247) ((T_DEC) reduce 247) ((T_DECLARE) reduce 247) ((T_DIR) reduce 247) ((T_DNUMBER) reduce 247) ((T_DO) reduce 247) ((T_DOUBLE_CAST) reduce 247) ((T_ECHO) reduce 247) ((T_ELSE) reduce 247) ((T_ELSEIF) reduce 247) ((T_EMPTY) reduce 247) ((T_ENDFOREACH) reduce 247) ((T_ENUM) reduce 247) ((T_EVAL) reduce 247) ((T_EXIT) reduce 247) ((T_FILE) reduce 247) ((T_FINAL) reduce 247) ((T_FN) reduce 247) ((T_FOR) reduce 247) ((T_FOREACH) reduce 247) ((T_FUNCTION) reduce 247) ((T_FUNC_C) reduce 247) ((T_GLOBAL) reduce 247) ((T_GOTO) reduce 247) ((T_HALT_COMPILER) reduce 247) ((T_IF) reduce 247) ((T_INC) reduce 247) ((T_INCLUDE) reduce 247) ((T_INCLUDE_ONCE) reduce 247) ((T_INLINE_HTML) reduce 247) ((T_INTERFACE) reduce 247) ((T_INT_CAST) reduce 247) ((T_ISSET) reduce 247) ((T_LINE) reduce 247) ((T_LIST) reduce 247) ((T_LNUMBER) reduce 247) ((T_MATCH) reduce 247) ((T_METHOD_C) reduce 247) ((T_NAME_FULLY_QUALIFIED) reduce 247) ((T_NAME_QUALIFIED) reduce 247) ((T_NAME_RELATIVE) reduce 247) ((T_NEW) reduce 247) ((T_NS_C) reduce 247) ((T_OBJECT_CAST) reduce 247) ((T_PRINT) reduce 247) ((T_PROPERTY_C) reduce 247) ((T_READONLY) reduce 247) ((T_REQUIRE) reduce 247) ((T_REQUIRE_ONCE) reduce 247) ((T_RETURN) reduce 247) ((T_START_HEREDOC) reduce 247) ((T_STATIC) reduce 247) ((T_STRING) reduce 247) ((T_STRING_CAST) reduce 247) ((T_SWITCH) reduce 247) ((T_THROW) reduce 247) ((T_TRAIT) reduce 247) ((T_TRAIT_C) reduce 247) ((T_TRY) reduce 247) ((T_UNSET) reduce 247) ((T_UNSET_CAST) reduce 247) ((T_VARIABLE) reduce 247) ((T_VOID_CAST) reduce 247) ((T_WHILE) reduce 247) ((T_YIELD) reduce 247) ((T_YIELD_FROM) reduce 247) (("[") reduce 247) (("`") reduce 247) (("{") reduce 247) (("~") reduce 247)) 21042 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_ELSE) reduce 161) ((T_ELSEIF) reduce 161) ((T_EMPTY) reduce 161) ((T_ENDFOREACH) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_VOID_CAST) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 21045 ((("!") reduce 250) (("\"") reduce 250) (("$") reduce 250) (("(") reduce 250) (("+") reduce 250) (("-") reduce 250) ((";") reduce 250) (("@") reduce 250) ((T_ABSTRACT) reduce 250) ((T_ARRAY) reduce 250) ((T_ARRAY_CAST) reduce 250) ((T_ATTRIBUTE) reduce 250) ((T_BOOL_CAST) reduce 250) ((T_BREAK) reduce 250) ((T_CLASS) reduce 250) ((T_CLASS_C) reduce 250) ((T_CLONE) reduce 250) ((T_CONSTANT_ENCAPSED_STRING) reduce 250) ((T_CONTINUE) reduce 250) ((T_DEC) reduce 250) ((T_DECLARE) reduce 250) ((T_DIR) reduce 250) ((T_DNUMBER) reduce 250) ((T_DO) reduce 250) ((T_DOUBLE_CAST) reduce 250) ((T_ECHO) reduce 250) ((T_ELSE) reduce 250) ((T_ELSEIF) reduce 250) ((T_EMPTY) reduce 250) ((T_ENDFOREACH) reduce 250) ((T_ENUM) reduce 250) ((T_EVAL) reduce 250) ((T_EXIT) reduce 250) ((T_FILE) reduce 250) ((T_FINAL) reduce 250) ((T_FN) reduce 250) ((T_FOR) reduce 250) ((T_FOREACH) reduce 250) ((T_FUNCTION) reduce 250) ((T_FUNC_C) reduce 250) ((T_GLOBAL) reduce 250) ((T_GOTO) reduce 250) ((T_HALT_COMPILER) reduce 250) ((T_IF) reduce 250) ((T_INC) reduce 250) ((T_INCLUDE) reduce 250) ((T_INCLUDE_ONCE) reduce 250) ((T_INLINE_HTML) reduce 250) ((T_INTERFACE) reduce 250) ((T_INT_CAST) reduce 250) ((T_ISSET) reduce 250) ((T_LINE) reduce 250) ((T_LIST) reduce 250) ((T_LNUMBER) reduce 250) ((T_MATCH) reduce 250) ((T_METHOD_C) reduce 250) ((T_NAME_FULLY_QUALIFIED) reduce 250) ((T_NAME_QUALIFIED) reduce 250) ((T_NAME_RELATIVE) reduce 250) ((T_NEW) reduce 250) ((T_NS_C) reduce 250) ((T_OBJECT_CAST) reduce 250) ((T_PRINT) reduce 250) ((T_PROPERTY_C) reduce 250) ((T_READONLY) reduce 250) ((T_REQUIRE) reduce 250) ((T_REQUIRE_ONCE) reduce 250) ((T_RETURN) reduce 250) ((T_START_HEREDOC) reduce 250) ((T_STATIC) reduce 250) ((T_STRING) reduce 250) ((T_STRING_CAST) reduce 250) ((T_SWITCH) reduce 250) ((T_THROW) reduce 250) ((T_TRAIT) reduce 250) ((T_TRAIT_C) reduce 250) ((T_TRY) reduce 250) ((T_UNSET) reduce 250) ((T_UNSET_CAST) reduce 250) ((T_VARIABLE) reduce 250) ((T_VOID_CAST) reduce 250) ((T_WHILE) reduce 250) ((T_YIELD) reduce 250) ((T_YIELD_FROM) reduce 250) (("[") reduce 250) (("`") reduce 250) (("{") reduce 250) (("~") reduce 250)) 21049 ((("!") reduce 251) (("\"") reduce 251) (("$") reduce 251) (("(") reduce 251) (("+") reduce 251) (("-") reduce 251) ((";") reduce 251) (("@") reduce 251) ((T_ABSTRACT) reduce 251) ((T_ARRAY) reduce 251) ((T_ARRAY_CAST) reduce 251) ((T_ATTRIBUTE) reduce 251) ((T_BOOL_CAST) reduce 251) ((T_BREAK) reduce 251) ((T_CLASS) reduce 251) ((T_CLASS_C) reduce 251) ((T_CLONE) reduce 251) ((T_CONSTANT_ENCAPSED_STRING) reduce 251) ((T_CONTINUE) reduce 251) ((T_DEC) reduce 251) ((T_DECLARE) reduce 251) ((T_DIR) reduce 251) ((T_DNUMBER) reduce 251) ((T_DO) reduce 251) ((T_DOUBLE_CAST) reduce 251) ((T_ECHO) reduce 251) ((T_ELSE) reduce 251) ((T_ELSEIF) reduce 251) ((T_EMPTY) reduce 251) ((T_ENDFOREACH) reduce 251) ((T_ENUM) reduce 251) ((T_EVAL) reduce 251) ((T_EXIT) reduce 251) ((T_FILE) reduce 251) ((T_FINAL) reduce 251) ((T_FN) reduce 251) ((T_FOR) reduce 251) ((T_FOREACH) reduce 251) ((T_FUNCTION) reduce 251) ((T_FUNC_C) reduce 251) ((T_GLOBAL) reduce 251) ((T_GOTO) reduce 251) ((T_HALT_COMPILER) reduce 251) ((T_IF) reduce 251) ((T_INC) reduce 251) ((T_INCLUDE) reduce 251) ((T_INCLUDE_ONCE) reduce 251) ((T_INLINE_HTML) reduce 251) ((T_INTERFACE) reduce 251) ((T_INT_CAST) reduce 251) ((T_ISSET) reduce 251) ((T_LINE) reduce 251) ((T_LIST) reduce 251) ((T_LNUMBER) reduce 251) ((T_MATCH) reduce 251) ((T_METHOD_C) reduce 251) ((T_NAME_FULLY_QUALIFIED) reduce 251) ((T_NAME_QUALIFIED) reduce 251) ((T_NAME_RELATIVE) reduce 251) ((T_NEW) reduce 251) ((T_NS_C) reduce 251) ((T_OBJECT_CAST) reduce 251) ((T_PRINT) reduce 251) ((T_PROPERTY_C) reduce 251) ((T_READONLY) reduce 251) ((T_REQUIRE) reduce 251) ((T_REQUIRE_ONCE) reduce 251) ((T_RETURN) reduce 251) ((T_START_HEREDOC) reduce 251) ((T_STATIC) reduce 251) ((T_STRING) reduce 251) ((T_STRING_CAST) reduce 251) ((T_SWITCH) reduce 251) ((T_THROW) reduce 251) ((T_TRAIT) reduce 251) ((T_TRAIT_C) reduce 251) ((T_TRY) reduce 251) ((T_UNSET) reduce 251) ((T_UNSET_CAST) reduce 251) ((T_VARIABLE) reduce 251) ((T_VOID_CAST) reduce 251) ((T_WHILE) reduce 251) ((T_YIELD) reduce 251) ((T_YIELD_FROM) reduce 251) (("[") reduce 251) (("`") reduce 251) (("{") reduce 251) (("~") reduce 251)) 21054 ((("!") reduce 242) (("\"") reduce 242) (("$") reduce 242) (("(") reduce 242) (("+") reduce 242) (("-") reduce 242) ((";") reduce 242) (("@") reduce 242) ((T_ABSTRACT) reduce 242) ((T_ARRAY) reduce 242) ((T_ARRAY_CAST) reduce 242) ((T_ATTRIBUTE) reduce 242) ((T_BOOL_CAST) reduce 242) ((T_BREAK) reduce 242) ((T_CLASS) reduce 242) ((T_CLASS_C) reduce 242) ((T_CLONE) reduce 242) ((T_CONSTANT_ENCAPSED_STRING) reduce 242) ((T_CONTINUE) reduce 242) ((T_DEC) reduce 242) ((T_DECLARE) reduce 242) ((T_DIR) reduce 242) ((T_DNUMBER) reduce 242) ((T_DO) reduce 242) ((T_DOUBLE_CAST) reduce 242) ((T_ECHO) reduce 242) ((T_ELSE) reduce 242) ((T_ELSEIF) reduce 242) ((T_EMPTY) reduce 242) ((T_ENDFOREACH) reduce 242) ((T_ENUM) reduce 242) ((T_EVAL) reduce 242) ((T_EXIT) reduce 242) ((T_FILE) reduce 242) ((T_FINAL) reduce 242) ((T_FN) reduce 242) ((T_FOR) reduce 242) ((T_FOREACH) reduce 242) ((T_FUNCTION) reduce 242) ((T_FUNC_C) reduce 242) ((T_GLOBAL) reduce 242) ((T_GOTO) reduce 242) ((T_HALT_COMPILER) reduce 242) ((T_IF) reduce 242) ((T_INC) reduce 242) ((T_INCLUDE) reduce 242) ((T_INCLUDE_ONCE) reduce 242) ((T_INLINE_HTML) reduce 242) ((T_INTERFACE) reduce 242) ((T_INT_CAST) reduce 242) ((T_ISSET) reduce 242) ((T_LINE) reduce 242) ((T_LIST) reduce 242) ((T_LNUMBER) reduce 242) ((T_MATCH) reduce 242) ((T_METHOD_C) reduce 242) ((T_NAME_FULLY_QUALIFIED) reduce 242) ((T_NAME_QUALIFIED) reduce 242) ((T_NAME_RELATIVE) reduce 242) ((T_NEW) reduce 242) ((T_NS_C) reduce 242) ((T_OBJECT_CAST) reduce 242) ((T_PRINT) reduce 242) ((T_PROPERTY_C) reduce 242) ((T_READONLY) reduce 242) ((T_REQUIRE) reduce 242) ((T_REQUIRE_ONCE) reduce 242) ((T_RETURN) reduce 242) ((T_START_HEREDOC) reduce 242) ((T_STATIC) reduce 242) ((T_STRING) reduce 242) ((T_STRING_CAST) reduce 242) ((T_SWITCH) reduce 242) ((T_THROW) reduce 242) ((T_TRAIT) reduce 242) ((T_TRAIT_C) reduce 242) ((T_TRY) reduce 242) ((T_UNSET) reduce 242) ((T_UNSET_CAST) reduce 242) ((T_VARIABLE) reduce 242) ((T_VOID_CAST) reduce 242) ((T_WHILE) reduce 242) ((T_YIELD) reduce 242) ((T_YIELD_FROM) reduce 242) (("[") reduce 242) (("`") reduce 242) (("{") reduce 242) (("~") reduce 242)) 21055 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_ELSE) reduce 150) ((T_ELSEIF) reduce 150) ((T_EMPTY) reduce 150) ((T_ENDFOREACH) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_VOID_CAST) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 21058 ((("!") reduce 243) (("\"") reduce 243) (("$") reduce 243) (("(") reduce 243) (("+") reduce 243) (("-") reduce 243) ((";") reduce 243) (("@") reduce 243) ((T_ABSTRACT) reduce 243) ((T_ARRAY) reduce 243) ((T_ARRAY_CAST) reduce 243) ((T_ATTRIBUTE) reduce 243) ((T_BOOL_CAST) reduce 243) ((T_BREAK) reduce 243) ((T_CLASS) reduce 243) ((T_CLASS_C) reduce 243) ((T_CLONE) reduce 243) ((T_CONSTANT_ENCAPSED_STRING) reduce 243) ((T_CONTINUE) reduce 243) ((T_DEC) reduce 243) ((T_DECLARE) reduce 243) ((T_DIR) reduce 243) ((T_DNUMBER) reduce 243) ((T_DO) reduce 243) ((T_DOUBLE_CAST) reduce 243) ((T_ECHO) reduce 243) ((T_ELSE) reduce 243) ((T_ELSEIF) reduce 243) ((T_EMPTY) reduce 243) ((T_ENDFOREACH) reduce 243) ((T_ENUM) reduce 243) ((T_EVAL) reduce 243) ((T_EXIT) reduce 243) ((T_FILE) reduce 243) ((T_FINAL) reduce 243) ((T_FN) reduce 243) ((T_FOR) reduce 243) ((T_FOREACH) reduce 243) ((T_FUNCTION) reduce 243) ((T_FUNC_C) reduce 243) ((T_GLOBAL) reduce 243) ((T_GOTO) reduce 243) ((T_HALT_COMPILER) reduce 243) ((T_IF) reduce 243) ((T_INC) reduce 243) ((T_INCLUDE) reduce 243) ((T_INCLUDE_ONCE) reduce 243) ((T_INLINE_HTML) reduce 243) ((T_INTERFACE) reduce 243) ((T_INT_CAST) reduce 243) ((T_ISSET) reduce 243) ((T_LINE) reduce 243) ((T_LIST) reduce 243) ((T_LNUMBER) reduce 243) ((T_MATCH) reduce 243) ((T_METHOD_C) reduce 243) ((T_NAME_FULLY_QUALIFIED) reduce 243) ((T_NAME_QUALIFIED) reduce 243) ((T_NAME_RELATIVE) reduce 243) ((T_NEW) reduce 243) ((T_NS_C) reduce 243) ((T_OBJECT_CAST) reduce 243) ((T_PRINT) reduce 243) ((T_PROPERTY_C) reduce 243) ((T_READONLY) reduce 243) ((T_REQUIRE) reduce 243) ((T_REQUIRE_ONCE) reduce 243) ((T_RETURN) reduce 243) ((T_START_HEREDOC) reduce 243) ((T_STATIC) reduce 243) ((T_STRING) reduce 243) ((T_STRING_CAST) reduce 243) ((T_SWITCH) reduce 243) ((T_THROW) reduce 243) ((T_TRAIT) reduce 243) ((T_TRAIT_C) reduce 243) ((T_TRY) reduce 243) ((T_UNSET) reduce 243) ((T_UNSET_CAST) reduce 243) ((T_VARIABLE) reduce 243) ((T_VOID_CAST) reduce 243) ((T_WHILE) reduce 243) ((T_YIELD) reduce 243) ((T_YIELD_FROM) reduce 243) (("[") reduce 243) (("`") reduce 243) (("{") reduce 243) (("~") reduce 243)) 21060 ((("!") reduce 170) (("\"") reduce 170) (("$") reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_ELSE) reduce 170) ((T_ELSEIF) reduce 170) ((T_EMPTY) reduce 170) ((T_ENDFOREACH) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_VARIABLE) reduce 170) ((T_VOID_CAST) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("~") reduce 170)) 21065 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_ELSE) reduce 162) ((T_ELSEIF) reduce 162) ((T_EMPTY) reduce 162) ((T_ENDFOREACH) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_VOID_CAST) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 21068 ((("!") reduce 171) (("\"") reduce 171) (("$") reduce 171) (("(") reduce 171) (("+") reduce 171) (("-") reduce 171) ((";") reduce 171) (("@") reduce 171) ((T_ABSTRACT) reduce 171) ((T_ARRAY) reduce 171) ((T_ARRAY_CAST) reduce 171) ((T_ATTRIBUTE) reduce 171) ((T_BOOL_CAST) reduce 171) ((T_BREAK) reduce 171) ((T_CATCH) reduce 171) ((T_CLASS) reduce 171) ((T_CLASS_C) reduce 171) ((T_CLONE) reduce 171) ((T_CONSTANT_ENCAPSED_STRING) reduce 171) ((T_CONTINUE) reduce 171) ((T_DEC) reduce 171) ((T_DECLARE) reduce 171) ((T_DIR) reduce 171) ((T_DNUMBER) reduce 171) ((T_DO) reduce 171) ((T_DOUBLE_CAST) reduce 171) ((T_ECHO) reduce 171) ((T_ELSE) reduce 171) ((T_ELSEIF) reduce 171) ((T_EMPTY) reduce 171) ((T_ENDFOREACH) reduce 171) ((T_ENUM) reduce 171) ((T_EVAL) reduce 171) ((T_EXIT) reduce 171) ((T_FILE) reduce 171) ((T_FINAL) reduce 171) ((T_FINALLY) reduce 171) ((T_FN) reduce 171) ((T_FOR) reduce 171) ((T_FOREACH) reduce 171) ((T_FUNCTION) reduce 171) ((T_FUNC_C) reduce 171) ((T_GLOBAL) reduce 171) ((T_GOTO) reduce 171) ((T_HALT_COMPILER) reduce 171) ((T_IF) reduce 171) ((T_INC) reduce 171) ((T_INCLUDE) reduce 171) ((T_INCLUDE_ONCE) reduce 171) ((T_INLINE_HTML) reduce 171) ((T_INTERFACE) reduce 171) ((T_INT_CAST) reduce 171) ((T_ISSET) reduce 171) ((T_LINE) reduce 171) ((T_LIST) reduce 171) ((T_LNUMBER) reduce 171) ((T_MATCH) reduce 171) ((T_METHOD_C) reduce 171) ((T_NAME_FULLY_QUALIFIED) reduce 171) ((T_NAME_QUALIFIED) reduce 171) ((T_NAME_RELATIVE) reduce 171) ((T_NEW) reduce 171) ((T_NS_C) reduce 171) ((T_OBJECT_CAST) reduce 171) ((T_PRINT) reduce 171) ((T_PROPERTY_C) reduce 171) ((T_READONLY) reduce 171) ((T_REQUIRE) reduce 171) ((T_REQUIRE_ONCE) reduce 171) ((T_RETURN) reduce 171) ((T_START_HEREDOC) reduce 171) ((T_STATIC) reduce 171) ((T_STRING) reduce 171) ((T_STRING_CAST) reduce 171) ((T_SWITCH) reduce 171) ((T_THROW) reduce 171) ((T_TRAIT) reduce 171) ((T_TRAIT_C) reduce 171) ((T_TRY) reduce 171) ((T_UNSET) reduce 171) ((T_UNSET_CAST) reduce 171) ((T_VARIABLE) reduce 171) ((T_VOID_CAST) reduce 171) ((T_WHILE) reduce 171) ((T_YIELD) reduce 171) ((T_YIELD_FROM) reduce 171) (("[") reduce 171) (("`") reduce 171) (("{") reduce 171) (("~") reduce 171)) 21069 ((("!") reduce 177) (("\"") reduce 177) (("$") reduce 177) (("(") reduce 177) (("+") reduce 177) (("-") reduce 177) ((";") reduce 177) (("@") reduce 177) ((T_ABSTRACT) reduce 177) ((T_ARRAY) reduce 177) ((T_ARRAY_CAST) reduce 177) ((T_ATTRIBUTE) reduce 177) ((T_BOOL_CAST) reduce 177) ((T_BREAK) reduce 177) ((T_CATCH) shift) ((T_CLASS) reduce 177) ((T_CLASS_C) reduce 177) ((T_CLONE) reduce 177) ((T_CONSTANT_ENCAPSED_STRING) reduce 177) ((T_CONTINUE) reduce 177) ((T_DEC) reduce 177) ((T_DECLARE) reduce 177) ((T_DIR) reduce 177) ((T_DNUMBER) reduce 177) ((T_DO) reduce 177) ((T_DOUBLE_CAST) reduce 177) ((T_ECHO) reduce 177) ((T_ELSE) reduce 177) ((T_ELSEIF) reduce 177) ((T_EMPTY) reduce 177) ((T_ENDFOREACH) reduce 177) ((T_ENUM) reduce 177) ((T_EVAL) reduce 177) ((T_EXIT) reduce 177) ((T_FILE) reduce 177) ((T_FINAL) reduce 177) ((T_FINALLY) shift) ((T_FN) reduce 177) ((T_FOR) reduce 177) ((T_FOREACH) reduce 177) ((T_FUNCTION) reduce 177) ((T_FUNC_C) reduce 177) ((T_GLOBAL) reduce 177) ((T_GOTO) reduce 177) ((T_HALT_COMPILER) reduce 177) ((T_IF) reduce 177) ((T_INC) reduce 177) ((T_INCLUDE) reduce 177) ((T_INCLUDE_ONCE) reduce 177) ((T_INLINE_HTML) reduce 177) ((T_INTERFACE) reduce 177) ((T_INT_CAST) reduce 177) ((T_ISSET) reduce 177) ((T_LINE) reduce 177) ((T_LIST) reduce 177) ((T_LNUMBER) reduce 177) ((T_MATCH) reduce 177) ((T_METHOD_C) reduce 177) ((T_NAME_FULLY_QUALIFIED) reduce 177) ((T_NAME_QUALIFIED) reduce 177) ((T_NAME_RELATIVE) reduce 177) ((T_NEW) reduce 177) ((T_NS_C) reduce 177) ((T_OBJECT_CAST) reduce 177) ((T_PRINT) reduce 177) ((T_PROPERTY_C) reduce 177) ((T_READONLY) reduce 177) ((T_REQUIRE) reduce 177) ((T_REQUIRE_ONCE) reduce 177) ((T_RETURN) reduce 177) ((T_START_HEREDOC) reduce 177) ((T_STATIC) reduce 177) ((T_STRING) reduce 177) ((T_STRING_CAST) reduce 177) ((T_SWITCH) reduce 177) ((T_THROW) reduce 177) ((T_TRAIT) reduce 177) ((T_TRAIT_C) reduce 177) ((T_TRY) reduce 177) ((T_UNSET) reduce 177) ((T_UNSET_CAST) reduce 177) ((T_VARIABLE) reduce 177) ((T_VOID_CAST) reduce 177) ((T_WHILE) reduce 177) ((T_YIELD) reduce 177) ((T_YIELD_FROM) reduce 177) (("[") reduce 177) (("`") reduce 177) (("{") reduce 177) (("~") reduce 177)) 21072 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_ELSE) reduce 167) ((T_ELSEIF) reduce 167) ((T_EMPTY) reduce 167) ((T_ENDFOREACH) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_VOID_CAST) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 21075 ((("!") reduce 178) (("\"") reduce 178) (("$") reduce 178) (("(") reduce 178) (("+") reduce 178) (("-") reduce 178) ((";") reduce 178) (("@") reduce 178) ((T_ABSTRACT) reduce 178) ((T_ARRAY) reduce 178) ((T_ARRAY_CAST) reduce 178) ((T_ATTRIBUTE) reduce 178) ((T_BOOL_CAST) reduce 178) ((T_BREAK) reduce 178) ((T_CLASS) reduce 178) ((T_CLASS_C) reduce 178) ((T_CLONE) reduce 178) ((T_CONSTANT_ENCAPSED_STRING) reduce 178) ((T_CONTINUE) reduce 178) ((T_DEC) reduce 178) ((T_DECLARE) reduce 178) ((T_DIR) reduce 178) ((T_DNUMBER) reduce 178) ((T_DO) reduce 178) ((T_DOUBLE_CAST) reduce 178) ((T_ECHO) reduce 178) ((T_ELSE) reduce 178) ((T_ELSEIF) reduce 178) ((T_EMPTY) reduce 178) ((T_ENDFOREACH) reduce 178) ((T_ENUM) reduce 178) ((T_EVAL) reduce 178) ((T_EXIT) reduce 178) ((T_FILE) reduce 178) ((T_FINAL) reduce 178) ((T_FN) reduce 178) ((T_FOR) reduce 178) ((T_FOREACH) reduce 178) ((T_FUNCTION) reduce 178) ((T_FUNC_C) reduce 178) ((T_GLOBAL) reduce 178) ((T_GOTO) reduce 178) ((T_HALT_COMPILER) reduce 178) ((T_IF) reduce 178) ((T_INC) reduce 178) ((T_INCLUDE) reduce 178) ((T_INCLUDE_ONCE) reduce 178) ((T_INLINE_HTML) reduce 178) ((T_INTERFACE) reduce 178) ((T_INT_CAST) reduce 178) ((T_ISSET) reduce 178) ((T_LINE) reduce 178) ((T_LIST) reduce 178) ((T_LNUMBER) reduce 178) ((T_MATCH) reduce 178) ((T_METHOD_C) reduce 178) ((T_NAME_FULLY_QUALIFIED) reduce 178) ((T_NAME_QUALIFIED) reduce 178) ((T_NAME_RELATIVE) reduce 178) ((T_NEW) reduce 178) ((T_NS_C) reduce 178) ((T_OBJECT_CAST) reduce 178) ((T_PRINT) reduce 178) ((T_PROPERTY_C) reduce 178) ((T_READONLY) reduce 178) ((T_REQUIRE) reduce 178) ((T_REQUIRE_ONCE) reduce 178) ((T_RETURN) reduce 178) ((T_START_HEREDOC) reduce 178) ((T_STATIC) reduce 178) ((T_STRING) reduce 178) ((T_STRING_CAST) reduce 178) ((T_SWITCH) reduce 178) ((T_THROW) reduce 178) ((T_TRAIT) reduce 178) ((T_TRAIT_C) reduce 178) ((T_TRY) reduce 178) ((T_UNSET) reduce 178) ((T_UNSET_CAST) reduce 178) ((T_VARIABLE) reduce 178) ((T_VOID_CAST) reduce 178) ((T_WHILE) reduce 178) ((T_YIELD) reduce 178) ((T_YIELD_FROM) reduce 178) (("[") reduce 178) (("`") reduce 178) (("{") reduce 178) (("~") reduce 178)) 21082 ((("!") reduce 172) (("\"") reduce 172) (("$") reduce 172) (("(") reduce 172) (("+") reduce 172) (("-") reduce 172) ((";") reduce 172) (("@") reduce 172) ((T_ABSTRACT) reduce 172) ((T_ARRAY) reduce 172) ((T_ARRAY_CAST) reduce 172) ((T_ATTRIBUTE) reduce 172) ((T_BOOL_CAST) reduce 172) ((T_BREAK) reduce 172) ((T_CATCH) reduce 172) ((T_CLASS) reduce 172) ((T_CLASS_C) reduce 172) ((T_CLONE) reduce 172) ((T_CONSTANT_ENCAPSED_STRING) reduce 172) ((T_CONTINUE) reduce 172) ((T_DEC) reduce 172) ((T_DECLARE) reduce 172) ((T_DIR) reduce 172) ((T_DNUMBER) reduce 172) ((T_DO) reduce 172) ((T_DOUBLE_CAST) reduce 172) ((T_ECHO) reduce 172) ((T_ELSE) reduce 172) ((T_ELSEIF) reduce 172) ((T_EMPTY) reduce 172) ((T_ENDFOREACH) reduce 172) ((T_ENUM) reduce 172) ((T_EVAL) reduce 172) ((T_EXIT) reduce 172) ((T_FILE) reduce 172) ((T_FINAL) reduce 172) ((T_FINALLY) reduce 172) ((T_FN) reduce 172) ((T_FOR) reduce 172) ((T_FOREACH) reduce 172) ((T_FUNCTION) reduce 172) ((T_FUNC_C) reduce 172) ((T_GLOBAL) reduce 172) ((T_GOTO) reduce 172) ((T_HALT_COMPILER) reduce 172) ((T_IF) reduce 172) ((T_INC) reduce 172) ((T_INCLUDE) reduce 172) ((T_INCLUDE_ONCE) reduce 172) ((T_INLINE_HTML) reduce 172) ((T_INTERFACE) reduce 172) ((T_INT_CAST) reduce 172) ((T_ISSET) reduce 172) ((T_LINE) reduce 172) ((T_LIST) reduce 172) ((T_LNUMBER) reduce 172) ((T_MATCH) reduce 172) ((T_METHOD_C) reduce 172) ((T_NAME_FULLY_QUALIFIED) reduce 172) ((T_NAME_QUALIFIED) reduce 172) ((T_NAME_RELATIVE) reduce 172) ((T_NEW) reduce 172) ((T_NS_C) reduce 172) ((T_OBJECT_CAST) reduce 172) ((T_PRINT) reduce 172) ((T_PROPERTY_C) reduce 172) ((T_READONLY) reduce 172) ((T_REQUIRE) reduce 172) ((T_REQUIRE_ONCE) reduce 172) ((T_RETURN) reduce 172) ((T_START_HEREDOC) reduce 172) ((T_STATIC) reduce 172) ((T_STRING) reduce 172) ((T_STRING_CAST) reduce 172) ((T_SWITCH) reduce 172) ((T_THROW) reduce 172) ((T_TRAIT) reduce 172) ((T_TRAIT_C) reduce 172) ((T_TRY) reduce 172) ((T_UNSET) reduce 172) ((T_UNSET_CAST) reduce 172) ((T_VARIABLE) reduce 172) ((T_VOID_CAST) reduce 172) ((T_WHILE) reduce 172) ((T_YIELD) reduce 172) ((T_YIELD_FROM) reduce 172) (("[") reduce 172) (("`") reduce 172) (("{") reduce 172) (("~") reduce 172)) 21087 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_ELSE) reduce 153) ((T_ELSEIF) reduce 153) ((T_EMPTY) reduce 153) ((T_ENDFOREACH) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_VOID_CAST) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 21091 ((("!") reduce 224) (("\"") reduce 224) (("$") reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_ELSE) reduce 224) ((T_ELSEIF) reduce 224) ((T_EMPTY) reduce 224) ((T_ENDFOREACH) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_VARIABLE) reduce 224) ((T_VOID_CAST) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("~") reduce 224)) 21093 ((("!") reduce 225) (("\"") reduce 225) (("$") reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_ELSE) reduce 225) ((T_ELSEIF) reduce 225) ((T_EMPTY) reduce 225) ((T_ENDFOREACH) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_VARIABLE) reduce 225) ((T_VOID_CAST) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("~") reduce 225)) 21097 ((("!") reduce 226) (("\"") reduce 226) (("$") reduce 226) (("(") reduce 226) (("+") reduce 226) (("-") reduce 226) ((";") reduce 226) (("@") reduce 226) ((T_ABSTRACT) reduce 226) ((T_ARRAY) reduce 226) ((T_ARRAY_CAST) reduce 226) ((T_ATTRIBUTE) reduce 226) ((T_BOOL_CAST) reduce 226) ((T_BREAK) reduce 226) ((T_CLASS) reduce 226) ((T_CLASS_C) reduce 226) ((T_CLONE) reduce 226) ((T_CONSTANT_ENCAPSED_STRING) reduce 226) ((T_CONTINUE) reduce 226) ((T_DEC) reduce 226) ((T_DECLARE) reduce 226) ((T_DIR) reduce 226) ((T_DNUMBER) reduce 226) ((T_DO) reduce 226) ((T_DOUBLE_CAST) reduce 226) ((T_ECHO) reduce 226) ((T_ELSE) reduce 226) ((T_ELSEIF) reduce 226) ((T_EMPTY) reduce 226) ((T_ENDFOREACH) reduce 226) ((T_ENUM) reduce 226) ((T_EVAL) reduce 226) ((T_EXIT) reduce 226) ((T_FILE) reduce 226) ((T_FINAL) reduce 226) ((T_FN) reduce 226) ((T_FOR) reduce 226) ((T_FOREACH) reduce 226) ((T_FUNCTION) reduce 226) ((T_FUNC_C) reduce 226) ((T_GLOBAL) reduce 226) ((T_GOTO) reduce 226) ((T_HALT_COMPILER) reduce 226) ((T_IF) reduce 226) ((T_INC) reduce 226) ((T_INCLUDE) reduce 226) ((T_INCLUDE_ONCE) reduce 226) ((T_INLINE_HTML) reduce 226) ((T_INTERFACE) reduce 226) ((T_INT_CAST) reduce 226) ((T_ISSET) reduce 226) ((T_LINE) reduce 226) ((T_LIST) reduce 226) ((T_LNUMBER) reduce 226) ((T_MATCH) reduce 226) ((T_METHOD_C) reduce 226) ((T_NAME_FULLY_QUALIFIED) reduce 226) ((T_NAME_QUALIFIED) reduce 226) ((T_NAME_RELATIVE) reduce 226) ((T_NEW) reduce 226) ((T_NS_C) reduce 226) ((T_OBJECT_CAST) reduce 226) ((T_PRINT) reduce 226) ((T_PROPERTY_C) reduce 226) ((T_READONLY) reduce 226) ((T_REQUIRE) reduce 226) ((T_REQUIRE_ONCE) reduce 226) ((T_RETURN) reduce 226) ((T_START_HEREDOC) reduce 226) ((T_STATIC) reduce 226) ((T_STRING) reduce 226) ((T_STRING_CAST) reduce 226) ((T_SWITCH) reduce 226) ((T_THROW) reduce 226) ((T_TRAIT) reduce 226) ((T_TRAIT_C) reduce 226) ((T_TRY) reduce 226) ((T_UNSET) reduce 226) ((T_UNSET_CAST) reduce 226) ((T_VARIABLE) reduce 226) ((T_VOID_CAST) reduce 226) ((T_WHILE) reduce 226) ((T_YIELD) reduce 226) ((T_YIELD_FROM) reduce 226) (("[") reduce 226) (("`") reduce 226) (("{") reduce 226) (("~") reduce 226)) 21100 ((("!") reduce 227) (("\"") reduce 227) (("$") reduce 227) (("(") reduce 227) (("+") reduce 227) (("-") reduce 227) ((";") reduce 227) (("@") reduce 227) ((T_ABSTRACT) reduce 227) ((T_ARRAY) reduce 227) ((T_ARRAY_CAST) reduce 227) ((T_ATTRIBUTE) reduce 227) ((T_BOOL_CAST) reduce 227) ((T_BREAK) reduce 227) ((T_CLASS) reduce 227) ((T_CLASS_C) reduce 227) ((T_CLONE) reduce 227) ((T_CONSTANT_ENCAPSED_STRING) reduce 227) ((T_CONTINUE) reduce 227) ((T_DEC) reduce 227) ((T_DECLARE) reduce 227) ((T_DIR) reduce 227) ((T_DNUMBER) reduce 227) ((T_DO) reduce 227) ((T_DOUBLE_CAST) reduce 227) ((T_ECHO) reduce 227) ((T_ELSE) reduce 227) ((T_ELSEIF) reduce 227) ((T_EMPTY) reduce 227) ((T_ENDFOREACH) reduce 227) ((T_ENUM) reduce 227) ((T_EVAL) reduce 227) ((T_EXIT) reduce 227) ((T_FILE) reduce 227) ((T_FINAL) reduce 227) ((T_FN) reduce 227) ((T_FOR) reduce 227) ((T_FOREACH) reduce 227) ((T_FUNCTION) reduce 227) ((T_FUNC_C) reduce 227) ((T_GLOBAL) reduce 227) ((T_GOTO) reduce 227) ((T_HALT_COMPILER) reduce 227) ((T_IF) reduce 227) ((T_INC) reduce 227) ((T_INCLUDE) reduce 227) ((T_INCLUDE_ONCE) reduce 227) ((T_INLINE_HTML) reduce 227) ((T_INTERFACE) reduce 227) ((T_INT_CAST) reduce 227) ((T_ISSET) reduce 227) ((T_LINE) reduce 227) ((T_LIST) reduce 227) ((T_LNUMBER) reduce 227) ((T_MATCH) reduce 227) ((T_METHOD_C) reduce 227) ((T_NAME_FULLY_QUALIFIED) reduce 227) ((T_NAME_QUALIFIED) reduce 227) ((T_NAME_RELATIVE) reduce 227) ((T_NEW) reduce 227) ((T_NS_C) reduce 227) ((T_OBJECT_CAST) reduce 227) ((T_PRINT) reduce 227) ((T_PROPERTY_C) reduce 227) ((T_READONLY) reduce 227) ((T_REQUIRE) reduce 227) ((T_REQUIRE_ONCE) reduce 227) ((T_RETURN) reduce 227) ((T_START_HEREDOC) reduce 227) ((T_STATIC) reduce 227) ((T_STRING) reduce 227) ((T_STRING_CAST) reduce 227) ((T_SWITCH) reduce 227) ((T_THROW) reduce 227) ((T_TRAIT) reduce 227) ((T_TRAIT_C) reduce 227) ((T_TRY) reduce 227) ((T_UNSET) reduce 227) ((T_UNSET_CAST) reduce 227) ((T_VARIABLE) reduce 227) ((T_VOID_CAST) reduce 227) ((T_WHILE) reduce 227) ((T_YIELD) reduce 227) ((T_YIELD_FROM) reduce 227) (("[") reduce 227) (("`") reduce 227) (("{") reduce 227) (("~") reduce 227)) 21101 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_ELSE) reduce 169) ((T_ELSEIF) reduce 169) ((T_EMPTY) reduce 169) ((T_ENDFOREACH) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_VOID_CAST) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 21103 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_ELSE) reduce 158) ((T_ELSEIF) reduce 158) ((T_EMPTY) reduce 158) ((T_ENDFOREACH) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_VOID_CAST) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 21105 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_ELSE) reduce 156) ((T_ELSEIF) reduce 156) ((T_EMPTY) reduce 156) ((T_ENDFOREACH) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_VOID_CAST) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 21107 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_ELSE) reduce 168) ((T_ELSEIF) reduce 168) ((T_EMPTY) reduce 168) ((T_ENDFOREACH) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_VOID_CAST) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 21109 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_ELSE) reduce 157) ((T_ELSEIF) reduce 157) ((T_EMPTY) reduce 157) ((T_ENDFOREACH) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_VOID_CAST) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 21119 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_ELSE) reduce 164) ((T_ELSEIF) reduce 164) ((T_EMPTY) reduce 164) ((T_ENDFOREACH) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_VOID_CAST) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 21120 ((("!") reduce 220) (("\"") reduce 220) (("$") reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_ELSE) reduce 220) ((T_ELSEIF) reduce 220) ((T_EMPTY) reduce 220) ((T_ENDFOREACH) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_VARIABLE) reduce 220) ((T_VOID_CAST) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("~") reduce 220)) 21123 ((("!") reduce 221) (("\"") reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_ELSE) reduce 221) ((T_ELSEIF) reduce 221) ((T_EMPTY) reduce 221) ((T_ENDFOREACH) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_VARIABLE) reduce 221) ((T_VOID_CAST) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("~") reduce 221)) 21124 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_ELSE) reduce 163) ((T_ELSEIF) reduce 163) ((T_EMPTY) reduce 163) ((T_ENDFOREACH) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_VOID_CAST) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 21125 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") reduce 392) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_VOID_CAST) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 21127 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 397) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 397) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21129 (((",") shift) ((";") reduce 393)) 21130 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_VOID_CAST) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 21132 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 394) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 394) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21133 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 395) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 395) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21134 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") reduce 389) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_VOID_CAST) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 21136 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 397) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 391) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21138 (((",") shift)) 21141 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 394) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 390) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21152 ((("%") reduce 540) (("*") reduce 540) (("+") reduce 540) ((",") reduce 540) (("-") reduce 540) ((".") reduce 540) (("/") reduce 540) (("<") reduce 540) ((">") reduce 540) (("?") reduce 540) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_BOOLEAN_AND) reduce 540) ((T_BOOLEAN_OR) reduce 540) ((T_COALESCE) reduce 540) ((T_INSTANCEOF) reduce 540) ((T_IS_EQUAL) reduce 540) ((T_IS_GREATER_OR_EQUAL) reduce 540) ((T_IS_IDENTICAL) reduce 540) ((T_IS_NOT_EQUAL) reduce 540) ((T_IS_NOT_IDENTICAL) reduce 540) ((T_IS_SMALLER_OR_EQUAL) reduce 540) ((T_LOGICAL_AND) reduce 540) ((T_LOGICAL_OR) reduce 540) ((T_LOGICAL_XOR) reduce 540) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 540) ((T_OBJECT_OPERATOR) reduce 540) ((T_PIPE) reduce 540) ((T_POW) reduce 540) ((T_SL) reduce 540) ((T_SPACESHIP) reduce 540) ((T_SR) reduce 540) (("[") reduce 540) (("^") reduce 540) (("|") reduce 540)) 21154 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_PIPE) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("|") reduce 521)) 21156 ((("%") reduce 534) (("*") reduce 534) (("+") reduce 534) ((",") reduce 534) (("-") reduce 534) ((".") reduce 534) (("/") reduce 534) (("<") reduce 534) ((">") reduce 534) (("?") reduce 534) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_BOOLEAN_AND) reduce 534) ((T_BOOLEAN_OR) reduce 534) ((T_COALESCE) reduce 534) ((T_INSTANCEOF) reduce 534) ((T_IS_EQUAL) reduce 534) ((T_IS_GREATER_OR_EQUAL) reduce 534) ((T_IS_IDENTICAL) reduce 534) ((T_IS_NOT_EQUAL) reduce 534) ((T_IS_NOT_IDENTICAL) reduce 534) ((T_IS_SMALLER_OR_EQUAL) reduce 534) ((T_LOGICAL_AND) reduce 534) ((T_LOGICAL_OR) reduce 534) ((T_LOGICAL_XOR) reduce 534) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PIPE) reduce 534) ((T_POW) reduce 534) ((T_SL) reduce 534) ((T_SPACESHIP) reduce 534) ((T_SR) reduce 534) (("[") reduce 534) (("^") reduce 534) (("|") reduce 534)) 21157 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_PIPE) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("^") reduce 524) (("|") reduce 524)) 21161 ((("%") reduce 517) (("(") shift) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_PIPE) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("^") reduce 517) (("|") reduce 517)) 21162 ((("%") reduce 533) (("*") reduce 533) (("+") reduce 533) ((",") reduce 533) (("-") reduce 533) ((".") reduce 533) (("/") reduce 533) (("<") reduce 533) ((">") reduce 533) (("?") reduce 533) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_BOOLEAN_AND) reduce 533) ((T_BOOLEAN_OR) reduce 533) ((T_COALESCE) reduce 533) ((T_INSTANCEOF) reduce 533) ((T_IS_EQUAL) reduce 533) ((T_IS_GREATER_OR_EQUAL) reduce 533) ((T_IS_IDENTICAL) reduce 533) ((T_IS_NOT_EQUAL) reduce 533) ((T_IS_NOT_IDENTICAL) reduce 533) ((T_IS_SMALLER_OR_EQUAL) reduce 533) ((T_LOGICAL_AND) reduce 533) ((T_LOGICAL_OR) reduce 533) ((T_LOGICAL_XOR) reduce 533) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PIPE) reduce 533) ((T_POW) reduce 533) ((T_SL) reduce 533) ((T_SPACESHIP) reduce 533) ((T_SR) reduce 533) (("[") reduce 533) (("^") reduce 533) (("|") reduce 533)) 21163 ((("%") reduce 537) (("*") reduce 537) (("+") reduce 537) ((",") reduce 537) (("-") reduce 537) ((".") reduce 537) (("/") reduce 537) (("<") reduce 537) ((">") reduce 537) (("?") reduce 537) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_BOOLEAN_AND) reduce 537) ((T_BOOLEAN_OR) reduce 537) ((T_COALESCE) reduce 537) ((T_INSTANCEOF) reduce 537) ((T_IS_EQUAL) reduce 537) ((T_IS_GREATER_OR_EQUAL) reduce 537) ((T_IS_IDENTICAL) reduce 537) ((T_IS_NOT_EQUAL) reduce 537) ((T_IS_NOT_IDENTICAL) reduce 537) ((T_IS_SMALLER_OR_EQUAL) reduce 537) ((T_LOGICAL_AND) reduce 537) ((T_LOGICAL_OR) reduce 537) ((T_LOGICAL_XOR) reduce 537) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_PIPE) reduce 537) ((T_POW) reduce 537) ((T_SL) reduce 537) ((T_SPACESHIP) reduce 537) ((T_SR) reduce 537) (("[") reduce 537) (("^") reduce 537) (("|") reduce 537)) 21169 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PIPE) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("|") reduce 532)) 21171 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_PIPE) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("^") reduce 523) (("|") reduce 523)) 21173 ((("%") reduce 536) (("*") reduce 536) (("+") reduce 536) ((",") reduce 536) (("-") reduce 536) ((".") reduce 536) (("/") reduce 536) (("<") reduce 536) ((">") reduce 536) (("?") reduce 536) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_BOOLEAN_AND) reduce 536) ((T_BOOLEAN_OR) reduce 536) ((T_COALESCE) reduce 536) ((T_INSTANCEOF) reduce 536) ((T_IS_EQUAL) reduce 536) ((T_IS_GREATER_OR_EQUAL) reduce 536) ((T_IS_IDENTICAL) reduce 536) ((T_IS_NOT_EQUAL) reduce 536) ((T_IS_NOT_IDENTICAL) reduce 536) ((T_IS_SMALLER_OR_EQUAL) reduce 536) ((T_LOGICAL_AND) reduce 536) ((T_LOGICAL_OR) reduce 536) ((T_LOGICAL_XOR) reduce 536) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 536) ((T_OBJECT_OPERATOR) reduce 536) ((T_PIPE) reduce 536) ((T_POW) reduce 536) ((T_SL) reduce 536) ((T_SPACESHIP) reduce 536) ((T_SR) reduce 536) (("[") reduce 536) (("^") reduce 536) (("|") reduce 536)) 21174 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("|") reduce 94)) 21175 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("|") reduce 93)) 21176 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("|") reduce 95)) 21178 ((("%") reduce 539) (("*") reduce 539) (("+") reduce 539) ((",") reduce 539) (("-") reduce 539) ((".") reduce 539) (("/") reduce 539) (("<") reduce 539) ((">") reduce 539) (("?") reduce 539) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_BOOLEAN_AND) reduce 539) ((T_BOOLEAN_OR) reduce 539) ((T_COALESCE) reduce 539) ((T_INSTANCEOF) reduce 539) ((T_IS_EQUAL) reduce 539) ((T_IS_GREATER_OR_EQUAL) reduce 539) ((T_IS_IDENTICAL) reduce 539) ((T_IS_NOT_EQUAL) reduce 539) ((T_IS_NOT_IDENTICAL) reduce 539) ((T_IS_SMALLER_OR_EQUAL) reduce 539) ((T_LOGICAL_AND) reduce 539) ((T_LOGICAL_OR) reduce 539) ((T_LOGICAL_XOR) reduce 539) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 539) ((T_OBJECT_OPERATOR) reduce 539) ((T_PIPE) reduce 539) ((T_POW) reduce 539) ((T_SL) reduce 539) ((T_SPACESHIP) reduce 539) ((T_SR) reduce 539) (("[") reduce 539) (("^") reduce 539) (("|") reduce 539)) 21181 ((("%") reduce 538) (("*") reduce 538) (("+") reduce 538) ((",") reduce 538) (("-") reduce 538) ((".") reduce 538) (("/") reduce 538) (("<") reduce 538) ((">") reduce 538) (("?") reduce 538) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_BOOLEAN_AND) reduce 538) ((T_BOOLEAN_OR) reduce 538) ((T_COALESCE) reduce 538) ((T_INSTANCEOF) reduce 538) ((T_IS_EQUAL) reduce 538) ((T_IS_GREATER_OR_EQUAL) reduce 538) ((T_IS_IDENTICAL) reduce 538) ((T_IS_NOT_EQUAL) reduce 538) ((T_IS_NOT_IDENTICAL) reduce 538) ((T_IS_SMALLER_OR_EQUAL) reduce 538) ((T_LOGICAL_AND) reduce 538) ((T_LOGICAL_OR) reduce 538) ((T_LOGICAL_XOR) reduce 538) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 538) ((T_OBJECT_OPERATOR) reduce 538) ((T_PIPE) reduce 538) ((T_POW) reduce 538) ((T_SL) reduce 538) ((T_SPACESHIP) reduce 538) ((T_SR) reduce 538) (("[") reduce 538) (("^") reduce 538) (("|") reduce 538)) 21187 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92)) 21190 ((("%") reduce 535) (("*") reduce 535) (("+") reduce 535) ((",") reduce 535) (("-") reduce 535) ((".") reduce 535) (("/") reduce 535) (("<") reduce 535) ((">") reduce 535) (("?") reduce 535) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_BOOLEAN_AND) reduce 535) ((T_BOOLEAN_OR) reduce 535) ((T_COALESCE) reduce 535) ((T_INSTANCEOF) reduce 535) ((T_IS_EQUAL) reduce 535) ((T_IS_GREATER_OR_EQUAL) reduce 535) ((T_IS_IDENTICAL) reduce 535) ((T_IS_NOT_EQUAL) reduce 535) ((T_IS_NOT_IDENTICAL) reduce 535) ((T_IS_SMALLER_OR_EQUAL) reduce 535) ((T_LOGICAL_AND) reduce 535) ((T_LOGICAL_OR) reduce 535) ((T_LOGICAL_XOR) reduce 535) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PIPE) reduce 535) ((T_POW) reduce 535) ((T_SL) reduce 535) ((T_SPACESHIP) reduce 535) ((T_SR) reduce 535) (("[") reduce 535) (("^") reduce 535) (("|") reduce 535)) 21192 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) (("=") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AND_EQUAL) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_COALESCE_EQUAL) reduce 568) ((T_CONCAT_EQUAL) reduce 568) ((T_DEC) reduce 568) ((T_DIV_EQUAL) reduce 568) ((T_INC) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_MINUS_EQUAL) reduce 568) ((T_MOD_EQUAL) reduce 568) ((T_MUL_EQUAL) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_OR_EQUAL) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_PLUS_EQUAL) reduce 568) ((T_POW) reduce 568) ((T_POW_EQUAL) reduce 568) ((T_SL) reduce 568) ((T_SL_EQUAL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) ((T_SR_EQUAL) reduce 568) ((T_XOR_EQUAL) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 21193 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 479) (("(") shift) (("*") reduce 479) (("+") shift) ((",") reduce 479) (("-") shift) ((".") reduce 479) (("/") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 479) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 479) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PIPE) reduce 479) ((T_POW) reduce 479) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 479) (("`") shift) (("|") reduce 479) (("~") shift)) 21200 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) ((",") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) (("<") reduce 564) (("=") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AND_EQUAL) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_COALESCE_EQUAL) reduce 564) ((T_CONCAT_EQUAL) reduce 564) ((T_DEC) reduce 564) ((T_DIV_EQUAL) reduce 564) ((T_INC) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_MINUS_EQUAL) reduce 564) ((T_MOD_EQUAL) reduce 564) ((T_MUL_EQUAL) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_OR_EQUAL) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_PLUS_EQUAL) reduce 564) ((T_POW) reduce 564) ((T_POW_EQUAL) reduce 564) ((T_SL) reduce 564) ((T_SL_EQUAL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) ((T_SR_EQUAL) reduce 564) ((T_XOR_EQUAL) reduce 564) (("[") reduce 564) (("^") reduce 564) (("|") reduce 564)) 21201 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PIPE) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 551) (("^") reduce 530) (("|") reduce 530)) 21203 ((("%") reduce 529) (("*") reduce 529) (("+") reduce 529) ((",") reduce 529) (("-") reduce 529) ((".") reduce 529) (("/") reduce 529) (("<") reduce 529) ((">") reduce 529) (("?") reduce 529) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_BOOLEAN_AND) reduce 529) ((T_BOOLEAN_OR) reduce 529) ((T_COALESCE) reduce 529) ((T_INSTANCEOF) reduce 529) ((T_IS_EQUAL) reduce 529) ((T_IS_GREATER_OR_EQUAL) reduce 529) ((T_IS_IDENTICAL) reduce 529) ((T_IS_NOT_EQUAL) reduce 529) ((T_IS_NOT_IDENTICAL) reduce 529) ((T_IS_SMALLER_OR_EQUAL) reduce 529) ((T_LOGICAL_AND) reduce 529) ((T_LOGICAL_OR) reduce 529) ((T_LOGICAL_XOR) reduce 529) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PIPE) reduce 529) ((T_POW) reduce 529) ((T_SL) reduce 529) ((T_SPACESHIP) reduce 529) ((T_SR) reduce 529) (("[") reduce 554) (("^") reduce 529) (("|") reduce 529)) 21204 ((("%") reduce 528) (("(") reduce 557) (("*") reduce 528) (("+") reduce 528) ((",") reduce 528) (("-") reduce 528) ((".") reduce 528) (("/") reduce 528) (("<") reduce 528) ((">") reduce 528) (("?") reduce 528) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_BOOLEAN_AND) reduce 528) ((T_BOOLEAN_OR) reduce 528) ((T_COALESCE) reduce 528) ((T_INSTANCEOF) reduce 528) ((T_IS_EQUAL) reduce 528) ((T_IS_GREATER_OR_EQUAL) reduce 528) ((T_IS_IDENTICAL) reduce 528) ((T_IS_NOT_EQUAL) reduce 528) ((T_IS_NOT_IDENTICAL) reduce 528) ((T_IS_SMALLER_OR_EQUAL) reduce 528) ((T_LOGICAL_AND) reduce 528) ((T_LOGICAL_OR) reduce 528) ((T_LOGICAL_XOR) reduce 528) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PIPE) reduce 528) ((T_POW) reduce 528) ((T_SL) reduce 528) ((T_SPACESHIP) reduce 528) ((T_SR) reduce 528) (("[") reduce 550) (("^") reduce 528) (("|") reduce 528)) 21205 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 395) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21208 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) ((",") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) (("<") reduce 563) (("=") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AND_EQUAL) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_COALESCE_EQUAL) reduce 563) ((T_CONCAT_EQUAL) reduce 563) ((T_DEC) reduce 563) ((T_DIV_EQUAL) reduce 563) ((T_INC) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_MINUS_EQUAL) reduce 563) ((T_MOD_EQUAL) reduce 563) ((T_MUL_EQUAL) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_OR_EQUAL) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_PLUS_EQUAL) reduce 563) ((T_POW) reduce 563) ((T_POW_EQUAL) reduce 563) ((T_SL) reduce 563) ((T_SL_EQUAL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) ((T_SR_EQUAL) reduce 563) ((T_XOR_EQUAL) reduce 563) (("[") reduce 563) (("^") reduce 563) (("|") reduce 563)) 21209 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) ((",") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_PIPE) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("^") reduce 484) (("|") reduce 484)) 21210 ((("%") reduce 466) (("*") reduce 466) (("+") reduce 466) ((",") reduce 466) (("-") reduce 466) ((".") reduce 466) (("/") reduce 466) (("<") reduce 466) ((">") reduce 466) (("?") reduce 466) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_BOOLEAN_AND) reduce 466) ((T_BOOLEAN_OR) reduce 466) ((T_COALESCE) reduce 466) ((T_INSTANCEOF) reduce 466) ((T_IS_EQUAL) reduce 466) ((T_IS_GREATER_OR_EQUAL) reduce 466) ((T_IS_IDENTICAL) reduce 466) ((T_IS_NOT_EQUAL) reduce 466) ((T_IS_NOT_IDENTICAL) reduce 466) ((T_IS_SMALLER_OR_EQUAL) reduce 466) ((T_LOGICAL_AND) reduce 466) ((T_LOGICAL_OR) reduce 466) ((T_LOGICAL_XOR) reduce 466) ((T_PIPE) reduce 466) ((T_POW) reduce 466) ((T_SL) reduce 466) ((T_SPACESHIP) reduce 466) ((T_SR) reduce 466) (("^") reduce 466) (("|") reduce 466)) 21211 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_PIPE) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("^") reduce 488) (("|") reduce 488)) 21212 ((("%") reduce 531) (("(") shift) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("|") reduce 531)) 21213 ((("%") reduce 461) (("(") reduce 558) (("*") reduce 461) (("+") reduce 461) ((",") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PIPE) reduce 461) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") reduce 552) (("^") reduce 461) (("|") reduce 461)) 21214 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) ((",") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_PIPE) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 21215 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) ((",") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_PIPE) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("|") reduce 476)) 21216 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) (("=") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AND_EQUAL) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_COALESCE_EQUAL) reduce 559) ((T_CONCAT_EQUAL) reduce 559) ((T_DEC) reduce 559) ((T_DIV_EQUAL) reduce 559) ((T_INC) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_MINUS_EQUAL) reduce 559) ((T_MOD_EQUAL) reduce 559) ((T_MUL_EQUAL) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_OR_EQUAL) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_PLUS_EQUAL) reduce 559) ((T_POW) reduce 559) ((T_POW_EQUAL) reduce 559) ((T_SL) reduce 559) ((T_SL_EQUAL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) ((T_SR_EQUAL) reduce 559) ((T_XOR_EQUAL) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 21217 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 21218 ((("%") reduce 403) (("*") reduce 403) (("+") reduce 403) ((",") reduce 403) (("-") reduce 403) ((".") reduce 403) (("/") reduce 403) (("<") reduce 403) (("=") shift) ((">") reduce 403) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_INC) shift) ((T_INSTANCEOF) reduce 403) ((T_IS_EQUAL) reduce 403) ((T_IS_GREATER_OR_EQUAL) reduce 403) ((T_IS_IDENTICAL) reduce 403) ((T_IS_NOT_EQUAL) reduce 403) ((T_IS_NOT_IDENTICAL) reduce 403) ((T_IS_SMALLER_OR_EQUAL) reduce 403) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 403) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 403) ((T_POW_EQUAL) shift) ((T_SL) reduce 403) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 403) ((T_SR) reduce 403) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 548) (("^") reduce 403) (("|") reduce 403)) 21221 ((("%") reduce 448) (("*") reduce 448) (("+") reduce 448) ((",") reduce 448) (("-") reduce 448) ((".") reduce 448) (("/") reduce 448) (("<") reduce 448) ((">") reduce 448) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_BOOLEAN_AND) reduce 448) ((T_BOOLEAN_OR) reduce 448) ((T_COALESCE) reduce 448) ((T_INSTANCEOF) reduce 448) ((T_IS_EQUAL) reduce 448) ((T_IS_GREATER_OR_EQUAL) reduce 448) ((T_IS_IDENTICAL) reduce 448) ((T_IS_NOT_EQUAL) reduce 448) ((T_IS_NOT_IDENTICAL) reduce 448) ((T_IS_SMALLER_OR_EQUAL) reduce 448) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_PIPE) reduce 448) ((T_POW) shift) ((T_SL) reduce 448) ((T_SPACESHIP) reduce 448) ((T_SR) reduce 448) (("^") reduce 448) (("|") reduce 448)) 21253 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 432) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 432)) 21254 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 435) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("|") reduce 435)) 21255 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 444) (("-") shift) ((".") reduce 444) (("/") shift) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_PIPE) reduce 444) ((T_POW) shift) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("|") reduce 444)) 21256 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 458) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 458) (("|") reduce 458)) 21257 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 443) (("-") shift) ((".") reduce 443) (("/") shift) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_PIPE) reduce 443) ((T_POW) shift) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("^") reduce 443) (("|") reduce 443)) 21258 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) ((",") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_PIPE) reduce 440) ((T_POW) shift) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("|") reduce 440)) 21259 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 453) (("-") shift) ((".") shift) (("/") shift) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_PIPE) reduce 453) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 453) ((T_SR) shift) (("^") reduce 453) (("|") reduce 453)) 21260 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 431) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21261 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 429) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21262 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 430) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21263 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 455) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 455) ((T_SR) shift) (("^") reduce 455) (("|") reduce 455)) 21264 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 450) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 450) (("|") reduce 450)) 21265 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 452) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 452) (("|") reduce 452)) 21266 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 449) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 449) (("|") reduce 449)) 21267 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 457) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 457) ((T_SR) shift) (("^") reduce 457) (("|") reduce 457)) 21268 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 451) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 451) (("|") reduce 451)) 21271 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 21272 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 21273 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 21274 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 21275 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 21276 ((("%") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 21277 ((("%") reduce 511) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("|") reduce 511)) 21278 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PIPE) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 21279 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 21280 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("|") reduce 512)) 21281 ((("%") reduce 573) (("*") reduce 573) (("+") reduce 573) ((",") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("|") reduce 573)) 21287 ((("%") reduce 574) (("*") reduce 574) (("+") reduce 574) ((",") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("|") reduce 574)) 21288 ((("%") reduce 578) (("*") reduce 578) (("+") reduce 578) ((",") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("|") reduce 578)) 21289 ((("%") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 21290 ((("%") reduce 575) (("*") reduce 575) (("+") reduce 575) ((",") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("|") reduce 575)) 21291 ((("%") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 21294 ((("%") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 21295 ((("%") reduce 576) (("*") reduce 576) (("+") reduce 576) ((",") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("|") reduce 576)) 21297 ((("%") reduce 577) (("*") reduce 577) (("+") reduce 577) ((",") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("|") reduce 577)) 21299 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("|") reduce 513)) 21300 ((("%") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 21303 ((("%") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 21304 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 465) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 465) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21305 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 427) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21306 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 428) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21307 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 433) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("|") reduce 433)) 21308 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 434) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("|") reduce 434)) 21312 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 463) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21313 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 464) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21314 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 456) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 456) ((T_SR) shift) (("^") reduce 456) (("|") reduce 456)) 21315 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 454) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 454) ((T_SR) shift) (("^") reduce 454) (("|") reduce 454)) 21316 ((("%") reduce 441) (("*") reduce 441) (("+") reduce 441) ((",") reduce 441) (("-") reduce 441) ((".") reduce 441) (("/") reduce 441) (("<") reduce 441) ((">") reduce 441) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 441) ((T_IS_GREATER_OR_EQUAL) reduce 441) ((T_IS_IDENTICAL) reduce 441) ((T_IS_NOT_EQUAL) reduce 441) ((T_IS_NOT_IDENTICAL) reduce 441) ((T_IS_SMALLER_OR_EQUAL) reduce 441) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_PIPE) reduce 441) ((T_POW) shift) ((T_SL) reduce 441) ((T_SPACESHIP) reduce 441) ((T_SR) reduce 441) (("^") reduce 441) (("|") reduce 441)) 21317 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 436) (("-") shift) ((".") reduce 436) (("/") shift) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_PIPE) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 436) ((T_SR) shift) (("^") reduce 436) (("|") reduce 436)) 21318 ((("%") shift) (("*") shift) (("+") reduce 438) ((",") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") shift) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_PIPE) reduce 438) ((T_POW) shift) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("|") reduce 438)) 21319 ((("%") shift) (("*") shift) (("+") reduce 437) ((",") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") shift) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_PIPE) reduce 437) ((T_POW) shift) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("|") reduce 437)) 21320 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) ((",") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_PIPE) reduce 439) ((T_POW) shift) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("|") reduce 439)) 21321 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) ((",") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_PIPE) reduce 442) ((T_POW) shift) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("|") reduce 442)) 21323 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) ((",") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_PIPE) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("|") reduce 74)) 21324 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) ((",") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_PIPE) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("|") reduce 52)) 21325 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) ((",") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_PIPE) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("|") reduce 27)) 21326 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) ((",") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_PIPE) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("|") reduce 51)) 21327 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) ((",") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_PIPE) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("|") reduce 53)) 21328 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) ((",") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_PIPE) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("|") reduce 49)) 21329 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) ((",") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_PIPE) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("|") reduce 29)) 21330 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) ((",") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_PIPE) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("|") reduce 59)) 21331 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) ((",") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_PIPE) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("|") reduce 60)) 21332 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) ((",") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_PIPE) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("|") reduce 11)) 21333 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) ((",") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_PIPE) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("|") reduce 42)) 21334 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) ((",") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_PIPE) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("|") reduce 39)) 21335 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) ((",") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_PIPE) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("|") reduce 25)) 21336 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) ((",") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_PIPE) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("|") reduce 50)) 21337 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) ((",") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_PIPE) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("|") reduce 66)) 21338 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) ((",") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_PIPE) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("|") reduce 18)) 21339 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) ((",") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_PIPE) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("|") reduce 17)) 21340 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) ((",") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_PIPE) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("|") reduce 15)) 21341 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) ((",") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_PIPE) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("|") reduce 14)) 21342 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) ((",") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_PIPE) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("|") reduce 38)) 21343 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) ((",") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_PIPE) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("|") reduce 26)) 21344 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) ((",") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_PIPE) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("|") reduce 22)) 21345 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) ((",") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_PIPE) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("|") reduce 24)) 21346 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) ((",") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_PIPE) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("|") reduce 16)) 21347 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) ((",") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_PIPE) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("|") reduce 48)) 21348 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) ((",") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_PIPE) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("|") reduce 20)) 21349 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) ((",") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_PIPE) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("|") reduce 70)) 21350 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) ((",") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_PIPE) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("|") reduce 3)) 21351 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) ((",") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_PIPE) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("|") reduce 12)) 21352 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) ((",") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_PIPE) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("|") reduce 54)) 21353 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) ((",") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_PIPE) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("|") reduce 65)) 21354 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) ((",") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_PIPE) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("|") reduce 75)) 21355 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) ((",") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_PIPE) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("|") reduce 30)) 21356 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) ((",") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_PIPE) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("|") reduce 68)) 21357 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) ((",") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_PIPE) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("|") reduce 21)) 21358 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) ((",") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_PIPE) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("|") reduce 23)) 21359 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) ((",") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_PIPE) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("|") reduce 41)) 21360 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) ((",") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_PIPE) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("|") reduce 62)) 21361 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) ((",") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_PIPE) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("|") reduce 34)) 21362 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) ((",") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_PIPE) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("|") reduce 40)) 21363 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) ((",") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_PIPE) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("|") reduce 13)) 21364 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) ((",") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_PIPE) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("|") reduce 55)) 21365 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) ((",") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_PIPE) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("|") reduce 1)) 21366 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) ((",") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_PIPE) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("|") reduce 2)) 21367 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) ((",") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_PIPE) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("|") reduce 9)) 21368 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) ((",") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_PIPE) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("|") reduce 33)) 21369 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) ((",") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_PIPE) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("|") reduce 58)) 21370 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) ((",") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_PIPE) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("|") reduce 37)) 21371 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) ((",") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_PIPE) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("|") reduce 64)) 21372 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) ((",") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_PIPE) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("|") reduce 46)) 21373 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) ((",") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_PIPE) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("|") reduce 8)) 21374 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) ((",") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_PIPE) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("|") reduce 6)) 21375 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) ((",") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_PIPE) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("|") reduce 7)) 21376 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) ((",") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_PIPE) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("|") reduce 69)) 21377 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) ((",") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_PIPE) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("|") reduce 63)) 21378 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) ((",") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_PIPE) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("|") reduce 56)) 21379 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) ((",") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_PIPE) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("|") reduce 10)) 21380 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) ((",") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_PIPE) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("|") reduce 67)) 21381 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) ((",") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_PIPE) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("|") reduce 44)) 21382 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) ((",") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_PIPE) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("|") reduce 76)) 21383 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) ((",") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_PIPE) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("|") reduce 71)) 21384 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) ((",") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_PIPE) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("|") reduce 77)) 21385 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) ((",") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_PIPE) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("|") reduce 78)) 21386 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) ((",") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_PIPE) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("|") reduce 79)) 21387 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) ((",") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_PIPE) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("|") reduce 4)) 21388 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) ((",") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_PIPE) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("|") reduce 5)) 21389 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) ((",") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_PIPE) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("|") reduce 43)) 21390 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) ((",") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_PIPE) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("|") reduce 73)) 21391 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) ((",") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_PIPE) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("|") reduce 82)) 21392 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) ((",") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_PIPE) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("|") reduce 47)) 21393 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) ((",") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_PIPE) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("|") reduce 31)) 21394 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) ((",") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_PIPE) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("|") reduce 57)) 21395 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) ((",") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_PIPE) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("|") reduce 61)) 21396 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) ((",") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_PIPE) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("|") reduce 28)) 21397 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) ((",") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_PIPE) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("|") reduce 36)) 21398 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) ((",") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_PIPE) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("|") reduce 32)) 21399 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) ((",") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_PIPE) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("|") reduce 35)) 21400 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) ((",") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_PIPE) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("|") reduce 19)) 21401 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) ((",") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_PIPE) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("|") reduce 45)) 21402 ((("%") reduce 542) (("(") reduce 579) (("*") reduce 542) (("+") reduce 542) ((",") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PIPE) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 21404 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) ((",") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_PIPE) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("|") reduce 72)) 21405 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) ((",") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_PIPE) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("|") reduce 83)) 21406 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) ((",") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) (("<") reduce 572) (("=") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AND_EQUAL) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_COALESCE_EQUAL) reduce 572) ((T_CONCAT_EQUAL) reduce 572) ((T_DEC) reduce 572) ((T_DIV_EQUAL) reduce 572) ((T_INC) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_MINUS_EQUAL) reduce 572) ((T_MOD_EQUAL) reduce 572) ((T_MUL_EQUAL) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_OR_EQUAL) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_PLUS_EQUAL) reduce 572) ((T_POW) reduce 572) ((T_POW_EQUAL) reduce 572) ((T_SL) reduce 572) ((T_SL_EQUAL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) ((T_SR_EQUAL) reduce 572) ((T_XOR_EQUAL) reduce 572) (("[") reduce 572) (("^") reduce 572) (("|") reduce 572)) 21409 ((("%") reduce 544) (("(") reduce 580) (("*") reduce 544) (("+") reduce 544) ((",") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PIPE) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 21411 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("|") reduce 507)) 21412 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 21416 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) (("=") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AND_EQUAL) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_COALESCE_EQUAL) reduce 292) ((T_CONCAT_EQUAL) reduce 292) ((T_DEC) reduce 292) ((T_DIV_EQUAL) reduce 292) ((T_INC) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_MINUS_EQUAL) reduce 292) ((T_MOD_EQUAL) reduce 292) ((T_MUL_EQUAL) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_OR_EQUAL) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_PLUS_EQUAL) reduce 292) ((T_POW) reduce 292) ((T_POW_EQUAL) reduce 292) ((T_SL) reduce 292) ((T_SL_EQUAL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) ((T_SR_EQUAL) reduce 292) ((T_XOR_EQUAL) reduce 292) (("[") reduce 292) (("^") reduce 292) (("|") reduce 292)) 21417 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) (("=") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AND_EQUAL) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_COALESCE_EQUAL) reduce 293) ((T_CONCAT_EQUAL) reduce 293) ((T_DEC) reduce 293) ((T_DIV_EQUAL) reduce 293) ((T_INC) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_MINUS_EQUAL) reduce 293) ((T_MOD_EQUAL) reduce 293) ((T_MUL_EQUAL) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_OR_EQUAL) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_PLUS_EQUAL) reduce 293) ((T_POW) reduce 293) ((T_POW_EQUAL) reduce 293) ((T_SL) reduce 293) ((T_SL_EQUAL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) ((T_SR_EQUAL) reduce 293) ((T_XOR_EQUAL) reduce 293) (("[") reduce 293) (("^") reduce 293) (("|") reduce 293)) 21422 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) ((",") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_PIPE) reduce 425) ((T_POW) reduce 425) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 21424 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) ((",") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_PIPE) reduce 423) ((T_POW) reduce 423) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 21434 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 419) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21435 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 421) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21436 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 420) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21437 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 413) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21438 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 410) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21439 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 418) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21440 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 412) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21441 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 416) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21442 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 411) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21443 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 414) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21444 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 415) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21445 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 422) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21446 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 417) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21448 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 406) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21451 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 21454 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) ((",") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) (("<") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_POW) reduce 564) ((T_SL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) (("[") reduce 564) (("^") reduce 564) (("|") reduce 564)) 21456 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) ((",") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) (("<") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_POW) reduce 563) ((T_SL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) (("[") reduce 563) (("^") reduce 563) (("|") reduce 563)) 21458 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 21459 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 21460 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) ((",") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("[") reduce 548) (("^") reduce 407) (("|") reduce 407)) 21464 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) ((",") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) (("<") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_POW) reduce 572) ((T_SL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) (("[") reduce 572) (("^") reduce 572) (("|") reduce 572)) 21466 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("|") reduce 507)) 21467 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 21471 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("[") reduce 292) (("^") reduce 292) (("|") reduce 292)) 21472 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("[") reduce 293) (("^") reduce 293) (("|") reduce 293)) 21473 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 21476 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) ((",") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) (("<") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_POW) reduce 571) ((T_SL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) (("[") reduce 571) (("^") reduce 571) (("|") reduce 571)) 21477 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("|") reduce 506)) 21478 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("|") reduce 508)) 21483 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 21484 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 21485 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 21486 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 21489 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 21490 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 21491 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 21492 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) ((",") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("[") reduce 562) (("^") reduce 562) (("|") reduce 562)) 21493 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 21494 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 21497 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 21498 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) (("=") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AND_EQUAL) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_COALESCE_EQUAL) reduce 504) ((T_CONCAT_EQUAL) reduce 504) ((T_DEC) reduce 504) ((T_DIV_EQUAL) reduce 504) ((T_INC) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_MINUS_EQUAL) reduce 504) ((T_MOD_EQUAL) reduce 504) ((T_MUL_EQUAL) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_OR_EQUAL) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_PLUS_EQUAL) reduce 504) ((T_POW) reduce 504) ((T_POW_EQUAL) reduce 504) ((T_SL) reduce 504) ((T_SL_EQUAL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) ((T_SR_EQUAL) reduce 504) ((T_XOR_EQUAL) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 21509 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) ((",") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PIPE) reduce 494) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 21510 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_PIPE) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("^") reduce 489) (("|") reduce 489)) 21583 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 494) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21899 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_PIPE) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("^") reduce 490) (("|") reduce 490)) 21902 ((("%") reduce 541) (("(") reduce 579) (("*") reduce 541) (("+") reduce 541) ((",") reduce 541) (("-") reduce 541) ((".") reduce 541) (("/") reduce 541) (("<") reduce 541) ((">") reduce 541) (("?") reduce 541) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_BOOLEAN_AND) reduce 541) ((T_BOOLEAN_OR) reduce 541) ((T_COALESCE) reduce 541) ((T_INSTANCEOF) reduce 541) ((T_IS_EQUAL) reduce 541) ((T_IS_GREATER_OR_EQUAL) reduce 541) ((T_IS_IDENTICAL) reduce 541) ((T_IS_NOT_EQUAL) reduce 541) ((T_IS_NOT_IDENTICAL) reduce 541) ((T_IS_SMALLER_OR_EQUAL) reduce 541) ((T_LOGICAL_AND) reduce 541) ((T_LOGICAL_OR) reduce 541) ((T_LOGICAL_XOR) reduce 541) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 541) ((T_OBJECT_OPERATOR) reduce 541) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 541) ((T_PIPE) reduce 541) ((T_POW) reduce 541) ((T_SL) reduce 541) ((T_SPACESHIP) reduce 541) ((T_SR) reduce 541) (("[") reduce 541) (("^") reduce 541) (("|") reduce 541)) 21904 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) ((",") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) (("<") reduce 571) (("=") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AND_EQUAL) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_COALESCE_EQUAL) reduce 571) ((T_CONCAT_EQUAL) reduce 571) ((T_DEC) reduce 571) ((T_DIV_EQUAL) reduce 571) ((T_INC) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_MINUS_EQUAL) reduce 571) ((T_MOD_EQUAL) reduce 571) ((T_MUL_EQUAL) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_OR_EQUAL) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_PLUS_EQUAL) reduce 571) ((T_POW) reduce 571) ((T_POW_EQUAL) reduce 571) ((T_SL) reduce 571) ((T_SL_EQUAL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) ((T_SR_EQUAL) reduce 571) ((T_XOR_EQUAL) reduce 571) (("[") reduce 571) (("^") reduce 571) (("|") reduce 571)) 21907 ((("%") reduce 543) (("(") reduce 580) (("*") reduce 543) (("+") reduce 543) ((",") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PIPE) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 21908 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("|") reduce 506)) 21909 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("|") reduce 508)) 21911 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) ((",") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_PIPE) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("^") reduce 485) (("|") reduce 485)) 21912 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_PIPE) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("^") reduce 487) (("|") reduce 487)) 21917 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) (("=") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AND_EQUAL) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_COALESCE_EQUAL) reduce 560) ((T_CONCAT_EQUAL) reduce 560) ((T_DEC) reduce 560) ((T_DIV_EQUAL) reduce 560) ((T_INC) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_MINUS_EQUAL) reduce 560) ((T_MOD_EQUAL) reduce 560) ((T_MUL_EQUAL) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_OR_EQUAL) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_PLUS_EQUAL) reduce 560) ((T_POW) reduce 560) ((T_POW_EQUAL) reduce 560) ((T_SL) reduce 560) ((T_SL_EQUAL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) ((T_SR_EQUAL) reduce 560) ((T_XOR_EQUAL) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 21918 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) (("=") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AND_EQUAL) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_COALESCE_EQUAL) reduce 582) ((T_CONCAT_EQUAL) reduce 582) ((T_DEC) reduce 582) ((T_DIV_EQUAL) reduce 582) ((T_INC) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_MINUS_EQUAL) reduce 582) ((T_MOD_EQUAL) reduce 582) ((T_MUL_EQUAL) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_OR_EQUAL) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_PLUS_EQUAL) reduce 582) ((T_POW) reduce 582) ((T_POW_EQUAL) reduce 582) ((T_SL) reduce 582) ((T_SL_EQUAL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) ((T_SR_EQUAL) reduce 582) ((T_XOR_EQUAL) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 21919 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 21920 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) (("=") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AND_EQUAL) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_COALESCE_EQUAL) reduce 584) ((T_CONCAT_EQUAL) reduce 584) ((T_DEC) reduce 584) ((T_DIV_EQUAL) reduce 584) ((T_INC) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_MINUS_EQUAL) reduce 584) ((T_MOD_EQUAL) reduce 584) ((T_MUL_EQUAL) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_OR_EQUAL) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_PLUS_EQUAL) reduce 584) ((T_POW) reduce 584) ((T_POW_EQUAL) reduce 584) ((T_SL) reduce 584) ((T_SL_EQUAL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) ((T_SR_EQUAL) reduce 584) ((T_XOR_EQUAL) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 21923 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) (("=") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AND_EQUAL) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_COALESCE_EQUAL) reduce 583) ((T_CONCAT_EQUAL) reduce 583) ((T_DEC) reduce 583) ((T_DIV_EQUAL) reduce 583) ((T_INC) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_MINUS_EQUAL) reduce 583) ((T_MOD_EQUAL) reduce 583) ((T_MUL_EQUAL) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_OR_EQUAL) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_PLUS_EQUAL) reduce 583) ((T_POW) reduce 583) ((T_POW_EQUAL) reduce 583) ((T_SL) reduce 583) ((T_SL_EQUAL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) ((T_SR_EQUAL) reduce 583) ((T_XOR_EQUAL) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 21924 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) (("=") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AND_EQUAL) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_COALESCE_EQUAL) reduce 561) ((T_CONCAT_EQUAL) reduce 561) ((T_DEC) reduce 561) ((T_DIV_EQUAL) reduce 561) ((T_INC) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_MINUS_EQUAL) reduce 561) ((T_MOD_EQUAL) reduce 561) ((T_MUL_EQUAL) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_OR_EQUAL) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_PLUS_EQUAL) reduce 561) ((T_POW) reduce 561) ((T_POW_EQUAL) reduce 561) ((T_SL) reduce 561) ((T_SL_EQUAL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) ((T_SR_EQUAL) reduce 561) ((T_XOR_EQUAL) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 21925 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 21926 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) ((",") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) (("=") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AND_EQUAL) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_COALESCE_EQUAL) reduce 562) ((T_CONCAT_EQUAL) reduce 562) ((T_DEC) reduce 562) ((T_DIV_EQUAL) reduce 562) ((T_INC) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_MINUS_EQUAL) reduce 562) ((T_MOD_EQUAL) reduce 562) ((T_MUL_EQUAL) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_OR_EQUAL) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_PLUS_EQUAL) reduce 562) ((T_POW) reduce 562) ((T_POW_EQUAL) reduce 562) ((T_SL) reduce 562) ((T_SL_EQUAL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) ((T_SR_EQUAL) reduce 562) ((T_XOR_EQUAL) reduce 562) (("[") reduce 562) (("^") reduce 562) (("|") reduce 562)) 21928 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) ((",") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_PIPE) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 21930 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) (("=") shift) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_PIPE) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("|") reduce 520)) 21932 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 405) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21933 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 482) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21936 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 481) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21937 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) ((",") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_PIPE) reduce 473) ((T_POW) shift) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("|") reduce 473)) 21938 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 483) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21939 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) ((",") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_PIPE) reduce 469) ((T_POW) shift) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 21940 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) ((",") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_PIPE) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("^") reduce 486) (("|") reduce 486)) 21942 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_PIPE) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("^") reduce 526) (("|") reduce 526)) 21944 ((("%") reduce 527) (("*") reduce 527) (("+") reduce 527) ((",") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_PIPE) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("^") reduce 527) (("|") reduce 527)) 21945 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_PIPE) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("^") reduce 525) (("|") reduce 525)) 21946 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 619) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21947 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 618) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21948 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 21949 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 478) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 21950 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) ((",") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_PIPE) reduce 471) ((T_POW) shift) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("|") reduce 471)) 21952 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 21953 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 21954 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 21955 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 21956 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 21957 ((("%") reduce 400) (("(") reduce 400) (("*") reduce 400) (("+") reduce 400) ((",") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 400) ((T_OBJECT_OPERATOR) reduce 400) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 400) ((T_PIPE) reduce 400) ((T_POW) reduce 400) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("[") reduce 400) (("^") reduce 400) (("|") reduce 400)) 21960 ((("%") reduce 511) (("(") reduce 511) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("|") reduce 511)) 21961 ((("%") reduce 402) (("(") shift) (("*") reduce 402) (("+") reduce 402) ((",") reduce 402) (("-") reduce 402) ((".") reduce 402) (("/") reduce 402) (("<") reduce 402) ((">") reduce 402) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_INSTANCEOF) reduce 402) ((T_IS_EQUAL) reduce 402) ((T_IS_GREATER_OR_EQUAL) reduce 402) ((T_IS_IDENTICAL) reduce 402) ((T_IS_NOT_EQUAL) reduce 402) ((T_IS_NOT_IDENTICAL) reduce 402) ((T_IS_SMALLER_OR_EQUAL) reduce 402) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_PIPE) reduce 402) ((T_POW) reduce 402) ((T_SL) reduce 402) ((T_SPACESHIP) reduce 402) ((T_SR) reduce 402) (("^") reduce 402) (("|") reduce 402)) 21962 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 21963 ((("%") reduce 512) (("(") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("|") reduce 512)) 21964 ((("%") reduce 573) (("(") reduce 573) (("*") reduce 573) (("+") reduce 573) ((",") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("|") reduce 573)) 21970 ((("%") reduce 574) (("(") reduce 574) (("*") reduce 574) (("+") reduce 574) ((",") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("|") reduce 574)) 21971 ((("%") reduce 578) (("(") reduce 578) (("*") reduce 578) (("+") reduce 578) ((",") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("|") reduce 578)) 21972 ((("%") reduce 575) (("(") reduce 575) (("*") reduce 575) (("+") reduce 575) ((",") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("|") reduce 575)) 21973 ((("%") reduce 576) (("(") reduce 576) (("*") reduce 576) (("+") reduce 576) ((",") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("|") reduce 576)) 21974 ((("%") reduce 399) (("(") reduce 399) (("*") reduce 399) (("+") reduce 399) ((",") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_INSTANCEOF) reduce 399) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 399) ((T_OBJECT_OPERATOR) reduce 399) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 399) ((T_PIPE) reduce 399) ((T_POW) reduce 399) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("[") reduce 399) (("^") reduce 399) (("|") reduce 399)) 21976 ((("%") reduce 577) (("(") reduce 577) (("*") reduce 577) (("+") reduce 577) ((",") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("|") reduce 577)) 21977 ((("%") reduce 401) (("(") reduce 401) (("*") reduce 401) (("+") reduce 401) ((",") reduce 401) (("-") reduce 401) ((".") reduce 401) (("/") reduce 401) (("<") reduce 401) ((">") reduce 401) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_INSTANCEOF) reduce 401) ((T_IS_EQUAL) reduce 401) ((T_IS_GREATER_OR_EQUAL) reduce 401) ((T_IS_IDENTICAL) reduce 401) ((T_IS_NOT_EQUAL) reduce 401) ((T_IS_NOT_IDENTICAL) reduce 401) ((T_IS_SMALLER_OR_EQUAL) reduce 401) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 401) ((T_OBJECT_OPERATOR) reduce 401) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 401) ((T_PIPE) reduce 401) ((T_POW) reduce 401) ((T_SL) reduce 401) ((T_SPACESHIP) reduce 401) ((T_SR) reduce 401) (("[") reduce 401) (("^") reduce 401) (("|") reduce 401)) 21985 ((("%") reduce 398) (("(") reduce 398) (("*") reduce 398) (("+") reduce 398) ((",") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 398) ((T_OBJECT_OPERATOR) reduce 398) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 398) ((T_PIPE) reduce 398) ((T_POW) reduce 398) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("[") reduce 398) (("^") reduce 398) (("|") reduce 398)) 21987 ((("%") reduce 513) (("(") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("|") reduce 513)) 21993 ((("%") reduce 233) (("*") reduce 233) (("+") reduce 233) ((",") reduce 233) (("-") reduce 233) ((".") reduce 233) (("/") reduce 233) (("<") reduce 233) ((">") reduce 233) (("?") reduce 233) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_BOOLEAN_AND) reduce 233) ((T_BOOLEAN_OR) reduce 233) ((T_COALESCE) reduce 233) ((T_INSTANCEOF) reduce 233) ((T_IS_EQUAL) reduce 233) ((T_IS_GREATER_OR_EQUAL) reduce 233) ((T_IS_IDENTICAL) reduce 233) ((T_IS_NOT_EQUAL) reduce 233) ((T_IS_NOT_IDENTICAL) reduce 233) ((T_IS_SMALLER_OR_EQUAL) reduce 233) ((T_LOGICAL_AND) reduce 233) ((T_LOGICAL_OR) reduce 233) ((T_LOGICAL_XOR) reduce 233) ((T_PIPE) reduce 233) ((T_POW) reduce 233) ((T_SL) reduce 233) ((T_SPACESHIP) reduce 233) ((T_SR) reduce 233) (("^") reduce 233) (("|") reduce 233)) 21998 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 404) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22002 ((("%") reduce 613) (("*") reduce 613) (("+") reduce 613) ((",") reduce 613) (("-") reduce 613) ((".") reduce 613) (("/") reduce 613) (("<") reduce 613) ((">") reduce 613) (("?") reduce 613) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_BOOLEAN_AND) reduce 613) ((T_BOOLEAN_OR) reduce 613) ((T_COALESCE) reduce 613) ((T_INSTANCEOF) reduce 613) ((T_IS_EQUAL) reduce 613) ((T_IS_GREATER_OR_EQUAL) reduce 613) ((T_IS_IDENTICAL) reduce 613) ((T_IS_NOT_EQUAL) reduce 613) ((T_IS_NOT_IDENTICAL) reduce 613) ((T_IS_SMALLER_OR_EQUAL) reduce 613) ((T_LOGICAL_AND) reduce 613) ((T_LOGICAL_OR) reduce 613) ((T_LOGICAL_XOR) reduce 613) ((T_PIPE) reduce 613) ((T_POW) reduce 613) ((T_SL) reduce 613) ((T_SPACESHIP) reduce 613) ((T_SR) reduce 613) (("^") reduce 613) (("|") reduce 613)) 22003 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) ((",") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_PIPE) reduce 467) ((T_POW) shift) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("|") reduce 467)) 22004 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 616) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22005 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 615) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22006 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) ((",") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("[") reduce 548) (("^") reduce 424) (("|") reduce 424)) 22008 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_PIPE) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("^") reduce 518) (("|") reduce 518)) 22009 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) ((",") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_PIPE) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("|") reduce 474)) 22010 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("|") reduce 291)) 22014 ((("%") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("^") reduce 292) (("|") reduce 292)) 22015 ((("%") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("^") reduce 293) (("|") reduce 293)) 22018 ((("%") reduce 617) (("*") reduce 617) (("+") reduce 617) ((",") reduce 617) (("-") reduce 617) ((".") reduce 617) (("/") reduce 617) (("<") reduce 617) ((">") reduce 617) (("?") reduce 617) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_BOOLEAN_AND) reduce 617) ((T_BOOLEAN_OR) reduce 617) ((T_COALESCE) reduce 617) ((T_INSTANCEOF) reduce 617) ((T_IS_EQUAL) reduce 617) ((T_IS_GREATER_OR_EQUAL) reduce 617) ((T_IS_IDENTICAL) reduce 617) ((T_IS_NOT_EQUAL) reduce 617) ((T_IS_NOT_IDENTICAL) reduce 617) ((T_IS_SMALLER_OR_EQUAL) reduce 617) ((T_LOGICAL_AND) reduce 617) ((T_LOGICAL_OR) reduce 617) ((T_LOGICAL_XOR) reduce 617) ((T_PIPE) reduce 617) ((T_POW) reduce 617) ((T_SL) reduce 617) ((T_SPACESHIP) reduce 617) ((T_SR) reduce 617) (("^") reduce 617) (("|") reduce 617)) 22021 ((("%") reduce 614) (("*") reduce 614) (("+") reduce 614) ((",") reduce 614) (("-") reduce 614) ((".") reduce 614) (("/") reduce 614) (("<") reduce 614) ((">") reduce 614) (("?") reduce 614) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_BOOLEAN_AND) reduce 614) ((T_BOOLEAN_OR) reduce 614) ((T_COALESCE) reduce 614) ((T_INSTANCEOF) reduce 614) ((T_IS_EQUAL) reduce 614) ((T_IS_GREATER_OR_EQUAL) reduce 614) ((T_IS_IDENTICAL) reduce 614) ((T_IS_NOT_EQUAL) reduce 614) ((T_IS_NOT_IDENTICAL) reduce 614) ((T_IS_SMALLER_OR_EQUAL) reduce 614) ((T_LOGICAL_AND) reduce 614) ((T_LOGICAL_OR) reduce 614) ((T_LOGICAL_XOR) reduce 614) ((T_PIPE) reduce 614) ((T_POW) reduce 614) ((T_SL) reduce 614) ((T_SPACESHIP) reduce 614) ((T_SR) reduce 614) (("^") reduce 614) (("|") reduce 614)) 22022 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) ((",") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_PIPE) reduce 468) ((T_POW) shift) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 22023 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) ((",") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("[") reduce 548) (("^") reduce 426) (("|") reduce 426)) 22025 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) ((",") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_PIPE) reduce 408) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("^") reduce 408) (("|") reduce 408)) 22026 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) ((",") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_PIPE) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("|") reduce 409)) 22027 ((("%") reduce 296) (("*") reduce 296) (("+") reduce 296) ((",") reduce 296) (("-") reduce 296) ((".") reduce 296) (("/") reduce 296) (("<") reduce 296) ((">") reduce 296) (("?") reduce 296) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_BOOLEAN_AND) reduce 296) ((T_BOOLEAN_OR) reduce 296) ((T_COALESCE) reduce 296) ((T_INSTANCEOF) reduce 296) ((T_IS_EQUAL) reduce 296) ((T_IS_GREATER_OR_EQUAL) reduce 296) ((T_IS_IDENTICAL) reduce 296) ((T_IS_NOT_EQUAL) reduce 296) ((T_IS_NOT_IDENTICAL) reduce 296) ((T_IS_SMALLER_OR_EQUAL) reduce 296) ((T_LOGICAL_AND) reduce 296) ((T_LOGICAL_OR) reduce 296) ((T_LOGICAL_XOR) reduce 296) ((T_PIPE) reduce 296) ((T_POW) reduce 296) ((T_SL) reduce 296) ((T_SPACESHIP) reduce 296) ((T_SR) reduce 296) (("^") reduce 296) (("|") reduce 296)) 22032 ((("%") reduce 297) (("*") reduce 297) (("+") reduce 297) ((",") reduce 297) (("-") reduce 297) ((".") reduce 297) (("/") reduce 297) (("<") reduce 297) ((">") reduce 297) (("?") reduce 297) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_BOOLEAN_AND) reduce 297) ((T_BOOLEAN_OR) reduce 297) ((T_COALESCE) reduce 297) ((T_INSTANCEOF) reduce 297) ((T_IS_EQUAL) reduce 297) ((T_IS_GREATER_OR_EQUAL) reduce 297) ((T_IS_IDENTICAL) reduce 297) ((T_IS_NOT_EQUAL) reduce 297) ((T_IS_NOT_IDENTICAL) reduce 297) ((T_IS_SMALLER_OR_EQUAL) reduce 297) ((T_LOGICAL_AND) reduce 297) ((T_LOGICAL_OR) reduce 297) ((T_LOGICAL_XOR) reduce 297) ((T_PIPE) reduce 297) ((T_POW) reduce 297) ((T_SL) reduce 297) ((T_SPACESHIP) reduce 297) ((T_SR) reduce 297) (("^") reduce 297) (("|") reduce 297)) 22033 ((("%") reduce 460) (("(") reduce 556) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PIPE) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("[") reduce 549) (("^") reduce 460) (("|") reduce 460)) 22035 ((("%") reduce 298) (("*") reduce 298) (("+") reduce 298) ((",") reduce 298) (("-") reduce 298) ((".") reduce 298) (("/") reduce 298) (("<") reduce 298) ((">") reduce 298) (("?") reduce 298) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_BOOLEAN_AND) reduce 298) ((T_BOOLEAN_OR) reduce 298) ((T_COALESCE) reduce 298) ((T_INSTANCEOF) reduce 298) ((T_IS_EQUAL) reduce 298) ((T_IS_GREATER_OR_EQUAL) reduce 298) ((T_IS_IDENTICAL) reduce 298) ((T_IS_NOT_EQUAL) reduce 298) ((T_IS_NOT_IDENTICAL) reduce 298) ((T_IS_SMALLER_OR_EQUAL) reduce 298) ((T_LOGICAL_AND) reduce 298) ((T_LOGICAL_OR) reduce 298) ((T_LOGICAL_XOR) reduce 298) ((T_PIPE) reduce 298) ((T_POW) reduce 298) ((T_SL) reduce 298) ((T_SPACESHIP) reduce 298) ((T_SR) reduce 298) (("^") reduce 298) (("|") reduce 298)) 22036 ((("%") reduce 299) (("*") reduce 299) (("+") reduce 299) ((",") reduce 299) (("-") reduce 299) ((".") reduce 299) (("/") reduce 299) (("<") reduce 299) ((">") reduce 299) (("?") reduce 299) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_BOOLEAN_AND) reduce 299) ((T_BOOLEAN_OR) reduce 299) ((T_COALESCE) reduce 299) ((T_INSTANCEOF) reduce 299) ((T_IS_EQUAL) reduce 299) ((T_IS_GREATER_OR_EQUAL) reduce 299) ((T_IS_IDENTICAL) reduce 299) ((T_IS_NOT_EQUAL) reduce 299) ((T_IS_NOT_IDENTICAL) reduce 299) ((T_IS_SMALLER_OR_EQUAL) reduce 299) ((T_LOGICAL_AND) reduce 299) ((T_LOGICAL_OR) reduce 299) ((T_LOGICAL_XOR) reduce 299) ((T_PIPE) reduce 299) ((T_POW) reduce 299) ((T_SL) reduce 299) ((T_SPACESHIP) reduce 299) ((T_SR) reduce 299) (("^") reduce 299) (("|") reduce 299)) 22037 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) ((",") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_PIPE) reduce 472) ((T_POW) shift) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("|") reduce 472)) 22038 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) ((",") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_PIPE) reduce 470) ((T_POW) shift) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("|") reduce 470)) 22041 ((("%") reduce 519) (("(") reduce 519) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PIPE) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("|") reduce 519)) 22042 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) ((",") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_PIPE) reduce 475) ((T_POW) shift) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("|") reduce 475)) 22043 ((("%") reduce 446) (("*") reduce 446) (("+") reduce 446) ((",") reduce 446) (("-") reduce 446) ((".") reduce 446) (("/") reduce 446) (("<") reduce 446) ((">") reduce 446) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_BOOLEAN_AND) reduce 446) ((T_BOOLEAN_OR) reduce 446) ((T_COALESCE) reduce 446) ((T_INSTANCEOF) reduce 446) ((T_IS_EQUAL) reduce 446) ((T_IS_GREATER_OR_EQUAL) reduce 446) ((T_IS_IDENTICAL) reduce 446) ((T_IS_NOT_EQUAL) reduce 446) ((T_IS_NOT_IDENTICAL) reduce 446) ((T_IS_SMALLER_OR_EQUAL) reduce 446) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_PIPE) reduce 446) ((T_POW) shift) ((T_SL) reduce 446) ((T_SPACESHIP) reduce 446) ((T_SR) reduce 446) (("^") reduce 446) (("|") reduce 446)) 22044 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) ((",") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_PIPE) reduce 445) ((T_POW) shift) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("|") reduce 445)) 22046 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) (("=") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AND_EQUAL) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_COALESCE_EQUAL) reduce 570) ((T_CONCAT_EQUAL) reduce 570) ((T_DEC) reduce 570) ((T_DIV_EQUAL) reduce 570) ((T_INC) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_MINUS_EQUAL) reduce 570) ((T_MOD_EQUAL) reduce 570) ((T_MUL_EQUAL) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_OR_EQUAL) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_PLUS_EQUAL) reduce 570) ((T_POW) reduce 570) ((T_POW_EQUAL) reduce 570) ((T_SL) reduce 570) ((T_SL_EQUAL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) ((T_SR_EQUAL) reduce 570) ((T_XOR_EQUAL) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 22049 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) (("=") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AND_EQUAL) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_COALESCE_EQUAL) reduce 569) ((T_CONCAT_EQUAL) reduce 569) ((T_DEC) reduce 569) ((T_DIV_EQUAL) reduce 569) ((T_INC) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_MINUS_EQUAL) reduce 569) ((T_MOD_EQUAL) reduce 569) ((T_MUL_EQUAL) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_OR_EQUAL) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_PLUS_EQUAL) reduce 569) ((T_POW) reduce 569) ((T_POW_EQUAL) reduce 569) ((T_SL) reduce 569) ((T_SL_EQUAL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) ((T_SR_EQUAL) reduce 569) ((T_XOR_EQUAL) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 22051 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_PIPE) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("|") reduce 522)) 22052 ((("%") reduce 447) (("*") reduce 447) (("+") reduce 447) ((",") reduce 447) (("-") reduce 447) ((".") reduce 447) (("/") reduce 447) (("<") reduce 447) ((">") reduce 447) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_BOOLEAN_AND) reduce 447) ((T_BOOLEAN_OR) reduce 447) ((T_COALESCE) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 447) ((T_IS_GREATER_OR_EQUAL) reduce 447) ((T_IS_IDENTICAL) reduce 447) ((T_IS_NOT_EQUAL) reduce 447) ((T_IS_NOT_IDENTICAL) reduce 447) ((T_IS_SMALLER_OR_EQUAL) reduce 447) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_PIPE) reduce 447) ((T_POW) shift) ((T_SL) reduce 447) ((T_SPACESHIP) reduce 447) ((T_SR) reduce 447) (("^") reduce 447) (("|") reduce 447)) 22736 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) ((")") reduce 392) (("+") shift) (("-") shift) (("@") shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_VOID_CAST) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("~") shift)) 22738 ((("%") shift) ((")") reduce 397) (("*") shift) (("+") shift) ((",") reduce 397) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22740 (((")") reduce 393) ((",") shift)) 22743 ((("%") shift) ((")") reduce 394) (("*") shift) (("+") shift) ((",") reduce 394) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22744 ((("%") shift) ((")") reduce 395) (("*") shift) (("+") shift) ((",") reduce 395) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22746 ((("!") reduce 142) (("\"") reduce 142) (("$") reduce 142) (("(") reduce 142) (("+") reduce 142) (("-") reduce 142) ((";") reduce 142) (("@") reduce 142) ((T_ABSTRACT) reduce 142) ((T_ARRAY) reduce 142) ((T_ARRAY_CAST) reduce 142) ((T_ATTRIBUTE) reduce 142) ((T_BOOL_CAST) reduce 142) ((T_BREAK) reduce 142) ((T_CLASS) reduce 142) ((T_CLASS_C) reduce 142) ((T_CLONE) reduce 142) ((T_CONSTANT_ENCAPSED_STRING) reduce 142) ((T_CONTINUE) reduce 142) ((T_DEC) reduce 142) ((T_DECLARE) reduce 142) ((T_DIR) reduce 142) ((T_DNUMBER) reduce 142) ((T_DO) reduce 142) ((T_DOUBLE_CAST) reduce 142) ((T_ECHO) reduce 142) ((T_EMPTY) reduce 142) ((T_ENDFOR) reduce 142) ((T_ENUM) reduce 142) ((T_EVAL) reduce 142) ((T_EXIT) reduce 142) ((T_FILE) reduce 142) ((T_FINAL) reduce 142) ((T_FN) reduce 142) ((T_FOR) reduce 142) ((T_FOREACH) reduce 142) ((T_FUNCTION) reduce 142) ((T_FUNC_C) reduce 142) ((T_GLOBAL) reduce 142) ((T_GOTO) reduce 142) ((T_HALT_COMPILER) reduce 142) ((T_IF) reduce 142) ((T_INC) reduce 142) ((T_INCLUDE) reduce 142) ((T_INCLUDE_ONCE) reduce 142) ((T_INLINE_HTML) reduce 142) ((T_INTERFACE) reduce 142) ((T_INT_CAST) reduce 142) ((T_ISSET) reduce 142) ((T_LINE) reduce 142) ((T_LIST) reduce 142) ((T_LNUMBER) reduce 142) ((T_MATCH) reduce 142) ((T_METHOD_C) reduce 142) ((T_NAME_FULLY_QUALIFIED) reduce 142) ((T_NAME_QUALIFIED) reduce 142) ((T_NAME_RELATIVE) reduce 142) ((T_NEW) reduce 142) ((T_NS_C) reduce 142) ((T_OBJECT_CAST) reduce 142) ((T_PRINT) reduce 142) ((T_PROPERTY_C) reduce 142) ((T_READONLY) reduce 142) ((T_REQUIRE) reduce 142) ((T_REQUIRE_ONCE) reduce 142) ((T_RETURN) reduce 142) ((T_START_HEREDOC) reduce 142) ((T_STATIC) reduce 142) ((T_STRING) reduce 142) ((T_STRING_CAST) reduce 142) ((T_SWITCH) reduce 142) ((T_THROW) reduce 142) ((T_TRAIT) reduce 142) ((T_TRAIT_C) reduce 142) ((T_TRY) reduce 142) ((T_UNSET) reduce 142) ((T_UNSET_CAST) reduce 142) ((T_VARIABLE) reduce 142) ((T_VOID_CAST) reduce 142) ((T_WHILE) reduce 142) ((T_YIELD) reduce 142) ((T_YIELD_FROM) reduce 142) (("[") reduce 142) (("`") reduce 142) (("{") reduce 142) (("~") reduce 142)) 22747 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_ELSE) reduce 152) ((T_ELSEIF) reduce 152) ((T_EMPTY) reduce 152) ((T_ENDFOREACH) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_VOID_CAST) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 22748 ((("!") reduce 218) (("\"") reduce 218) (("$") reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_ELSE) reduce 218) ((T_ELSEIF) reduce 218) ((T_EMPTY) reduce 218) ((T_ENDFOREACH) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_VARIABLE) reduce 218) ((T_VOID_CAST) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("~") reduce 218)) 22749 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_ENDFOR) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_VOID_CAST) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("~") shift)) 22750 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_EMPTY) reduce 166) ((T_ENDFOR) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_VOID_CAST) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 22765 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_EMPTY) reduce 160) ((T_ENDFOR) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_VOID_CAST) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 22776 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_EMPTY) reduce 149) ((T_ENDFOR) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_VOID_CAST) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 22778 ((("!") reduce 144) (("\"") reduce 144) (("$") reduce 144) (("(") reduce 144) (("+") reduce 144) (("-") reduce 144) ((";") reduce 144) (("@") reduce 144) ((T_ABSTRACT) reduce 144) ((T_ARRAY) reduce 144) ((T_ARRAY_CAST) reduce 144) ((T_ATTRIBUTE) reduce 144) ((T_BOOL_CAST) reduce 144) ((T_BREAK) reduce 144) ((T_CLASS) reduce 144) ((T_CLASS_C) reduce 144) ((T_CLONE) reduce 144) ((T_CONSTANT_ENCAPSED_STRING) reduce 144) ((T_CONTINUE) reduce 144) ((T_DEC) reduce 144) ((T_DECLARE) reduce 144) ((T_DIR) reduce 144) ((T_DNUMBER) reduce 144) ((T_DO) reduce 144) ((T_DOUBLE_CAST) reduce 144) ((T_ECHO) reduce 144) ((T_EMPTY) reduce 144) ((T_ENDFOR) reduce 144) ((T_ENUM) reduce 144) ((T_EVAL) reduce 144) ((T_EXIT) reduce 144) ((T_FILE) reduce 144) ((T_FINAL) reduce 144) ((T_FN) reduce 144) ((T_FOR) reduce 144) ((T_FOREACH) reduce 144) ((T_FUNCTION) reduce 144) ((T_FUNC_C) reduce 144) ((T_GLOBAL) reduce 144) ((T_GOTO) reduce 144) ((T_HALT_COMPILER) reduce 144) ((T_IF) reduce 144) ((T_INC) reduce 144) ((T_INCLUDE) reduce 144) ((T_INCLUDE_ONCE) reduce 144) ((T_INLINE_HTML) reduce 144) ((T_INTERFACE) reduce 144) ((T_INT_CAST) reduce 144) ((T_ISSET) reduce 144) ((T_LINE) reduce 144) ((T_LIST) reduce 144) ((T_LNUMBER) reduce 144) ((T_MATCH) reduce 144) ((T_METHOD_C) reduce 144) ((T_NAME_FULLY_QUALIFIED) reduce 144) ((T_NAME_QUALIFIED) reduce 144) ((T_NAME_RELATIVE) reduce 144) ((T_NEW) reduce 144) ((T_NS_C) reduce 144) ((T_OBJECT_CAST) reduce 144) ((T_PRINT) reduce 144) ((T_PROPERTY_C) reduce 144) ((T_READONLY) reduce 144) ((T_REQUIRE) reduce 144) ((T_REQUIRE_ONCE) reduce 144) ((T_RETURN) reduce 144) ((T_START_HEREDOC) reduce 144) ((T_STATIC) reduce 144) ((T_STRING) reduce 144) ((T_STRING_CAST) reduce 144) ((T_SWITCH) reduce 144) ((T_THROW) reduce 144) ((T_TRAIT) reduce 144) ((T_TRAIT_C) reduce 144) ((T_TRY) reduce 144) ((T_UNSET) reduce 144) ((T_UNSET_CAST) reduce 144) ((T_VARIABLE) reduce 144) ((T_VOID_CAST) reduce 144) ((T_WHILE) reduce 144) ((T_YIELD) reduce 144) ((T_YIELD_FROM) reduce 144) (("[") reduce 144) (("`") reduce 144) (("{") reduce 144) (("~") reduce 144)) 22780 ((("!") reduce 104) (("\"") reduce 104) (("$") reduce 104) (("(") reduce 104) (("+") reduce 104) (("-") reduce 104) ((";") reduce 104) (("@") reduce 104) ((T_ABSTRACT) reduce 104) ((T_ARRAY) reduce 104) ((T_ARRAY_CAST) reduce 104) ((T_ATTRIBUTE) reduce 104) ((T_BOOL_CAST) reduce 104) ((T_BREAK) reduce 104) ((T_CLASS) reduce 104) ((T_CLASS_C) reduce 104) ((T_CLONE) reduce 104) ((T_CONSTANT_ENCAPSED_STRING) reduce 104) ((T_CONTINUE) reduce 104) ((T_DEC) reduce 104) ((T_DECLARE) reduce 104) ((T_DIR) reduce 104) ((T_DNUMBER) reduce 104) ((T_DO) reduce 104) ((T_DOUBLE_CAST) reduce 104) ((T_ECHO) reduce 104) ((T_EMPTY) reduce 104) ((T_ENDFOR) reduce 104) ((T_ENUM) reduce 104) ((T_EVAL) reduce 104) ((T_EXIT) reduce 104) ((T_FILE) reduce 104) ((T_FINAL) reduce 104) ((T_FN) reduce 104) ((T_FOR) reduce 104) ((T_FOREACH) reduce 104) ((T_FUNCTION) reduce 104) ((T_FUNC_C) reduce 104) ((T_GLOBAL) reduce 104) ((T_GOTO) reduce 104) ((T_HALT_COMPILER) reduce 104) ((T_IF) reduce 104) ((T_INC) reduce 104) ((T_INCLUDE) reduce 104) ((T_INCLUDE_ONCE) reduce 104) ((T_INLINE_HTML) reduce 104) ((T_INTERFACE) reduce 104) ((T_INT_CAST) reduce 104) ((T_ISSET) reduce 104) ((T_LINE) reduce 104) ((T_LIST) reduce 104) ((T_LNUMBER) reduce 104) ((T_MATCH) reduce 104) ((T_METHOD_C) reduce 104) ((T_NAME_FULLY_QUALIFIED) reduce 104) ((T_NAME_QUALIFIED) reduce 104) ((T_NAME_RELATIVE) reduce 104) ((T_NEW) reduce 104) ((T_NS_C) reduce 104) ((T_OBJECT_CAST) reduce 104) ((T_PRINT) reduce 104) ((T_PROPERTY_C) reduce 104) ((T_READONLY) reduce 104) ((T_REQUIRE) reduce 104) ((T_REQUIRE_ONCE) reduce 104) ((T_RETURN) reduce 104) ((T_START_HEREDOC) reduce 104) ((T_STATIC) reduce 104) ((T_STRING) reduce 104) ((T_STRING_CAST) reduce 104) ((T_SWITCH) reduce 104) ((T_THROW) reduce 104) ((T_TRAIT) reduce 104) ((T_TRAIT_C) reduce 104) ((T_TRY) reduce 104) ((T_UNSET) reduce 104) ((T_UNSET_CAST) reduce 104) ((T_VARIABLE) reduce 104) ((T_VOID_CAST) reduce 104) ((T_WHILE) reduce 104) ((T_YIELD) reduce 104) ((T_YIELD_FROM) reduce 104) (("[") reduce 104) (("`") reduce 104) (("{") reduce 104) (("~") reduce 104)) 22782 ((("!") reduce 107) (("\"") reduce 107) (("$") reduce 107) (("(") reduce 107) (("+") reduce 107) (("-") reduce 107) ((";") reduce 107) (("@") reduce 107) ((T_ABSTRACT) reduce 107) ((T_ARRAY) reduce 107) ((T_ARRAY_CAST) reduce 107) ((T_ATTRIBUTE) reduce 107) ((T_BOOL_CAST) reduce 107) ((T_BREAK) reduce 107) ((T_CLASS) reduce 107) ((T_CLASS_C) reduce 107) ((T_CLONE) reduce 107) ((T_CONSTANT_ENCAPSED_STRING) reduce 107) ((T_CONTINUE) reduce 107) ((T_DEC) reduce 107) ((T_DECLARE) reduce 107) ((T_DIR) reduce 107) ((T_DNUMBER) reduce 107) ((T_DO) reduce 107) ((T_DOUBLE_CAST) reduce 107) ((T_ECHO) reduce 107) ((T_EMPTY) reduce 107) ((T_ENDFOR) reduce 107) ((T_ENUM) reduce 107) ((T_EVAL) reduce 107) ((T_EXIT) reduce 107) ((T_FILE) reduce 107) ((T_FINAL) reduce 107) ((T_FN) reduce 107) ((T_FOR) reduce 107) ((T_FOREACH) reduce 107) ((T_FUNCTION) reduce 107) ((T_FUNC_C) reduce 107) ((T_GLOBAL) reduce 107) ((T_GOTO) reduce 107) ((T_HALT_COMPILER) reduce 107) ((T_IF) reduce 107) ((T_INC) reduce 107) ((T_INCLUDE) reduce 107) ((T_INCLUDE_ONCE) reduce 107) ((T_INLINE_HTML) reduce 107) ((T_INTERFACE) reduce 107) ((T_INT_CAST) reduce 107) ((T_ISSET) reduce 107) ((T_LINE) reduce 107) ((T_LIST) reduce 107) ((T_LNUMBER) reduce 107) ((T_MATCH) reduce 107) ((T_METHOD_C) reduce 107) ((T_NAME_FULLY_QUALIFIED) reduce 107) ((T_NAME_QUALIFIED) reduce 107) ((T_NAME_RELATIVE) reduce 107) ((T_NEW) reduce 107) ((T_NS_C) reduce 107) ((T_OBJECT_CAST) reduce 107) ((T_PRINT) reduce 107) ((T_PROPERTY_C) reduce 107) ((T_READONLY) reduce 107) ((T_REQUIRE) reduce 107) ((T_REQUIRE_ONCE) reduce 107) ((T_RETURN) reduce 107) ((T_START_HEREDOC) reduce 107) ((T_STATIC) reduce 107) ((T_STRING) reduce 107) ((T_STRING_CAST) reduce 107) ((T_SWITCH) reduce 107) ((T_THROW) reduce 107) ((T_TRAIT) reduce 107) ((T_TRAIT_C) reduce 107) ((T_TRY) reduce 107) ((T_UNSET) reduce 107) ((T_UNSET_CAST) reduce 107) ((T_VARIABLE) reduce 107) ((T_VOID_CAST) reduce 107) ((T_WHILE) reduce 107) ((T_YIELD) reduce 107) ((T_YIELD_FROM) reduce 107) (("[") reduce 107) (("`") reduce 107) (("{") reduce 107) (("~") reduce 107)) 22785 ((("!") reduce 103) (("\"") reduce 103) (("$") reduce 103) (("(") reduce 103) (("+") reduce 103) (("-") reduce 103) ((";") reduce 103) (("@") reduce 103) ((T_ABSTRACT) reduce 103) ((T_ARRAY) reduce 103) ((T_ARRAY_CAST) reduce 103) ((T_ATTRIBUTE) reduce 103) ((T_BOOL_CAST) reduce 103) ((T_BREAK) reduce 103) ((T_CLASS) reduce 103) ((T_CLASS_C) reduce 103) ((T_CLONE) reduce 103) ((T_CONSTANT_ENCAPSED_STRING) reduce 103) ((T_CONTINUE) reduce 103) ((T_DEC) reduce 103) ((T_DECLARE) reduce 103) ((T_DIR) reduce 103) ((T_DNUMBER) reduce 103) ((T_DO) reduce 103) ((T_DOUBLE_CAST) reduce 103) ((T_ECHO) reduce 103) ((T_EMPTY) reduce 103) ((T_ENDFOR) reduce 103) ((T_ENUM) reduce 103) ((T_EVAL) reduce 103) ((T_EXIT) reduce 103) ((T_FILE) reduce 103) ((T_FINAL) reduce 103) ((T_FN) reduce 103) ((T_FOR) reduce 103) ((T_FOREACH) reduce 103) ((T_FUNCTION) reduce 103) ((T_FUNC_C) reduce 103) ((T_GLOBAL) reduce 103) ((T_GOTO) reduce 103) ((T_HALT_COMPILER) reduce 103) ((T_IF) reduce 103) ((T_INC) reduce 103) ((T_INCLUDE) reduce 103) ((T_INCLUDE_ONCE) reduce 103) ((T_INLINE_HTML) reduce 103) ((T_INTERFACE) reduce 103) ((T_INT_CAST) reduce 103) ((T_ISSET) reduce 103) ((T_LINE) reduce 103) ((T_LIST) reduce 103) ((T_LNUMBER) reduce 103) ((T_MATCH) reduce 103) ((T_METHOD_C) reduce 103) ((T_NAME_FULLY_QUALIFIED) reduce 103) ((T_NAME_QUALIFIED) reduce 103) ((T_NAME_RELATIVE) reduce 103) ((T_NEW) reduce 103) ((T_NS_C) reduce 103) ((T_OBJECT_CAST) reduce 103) ((T_PRINT) reduce 103) ((T_PROPERTY_C) reduce 103) ((T_READONLY) reduce 103) ((T_REQUIRE) reduce 103) ((T_REQUIRE_ONCE) reduce 103) ((T_RETURN) reduce 103) ((T_START_HEREDOC) reduce 103) ((T_STATIC) reduce 103) ((T_STRING) reduce 103) ((T_STRING_CAST) reduce 103) ((T_SWITCH) reduce 103) ((T_THROW) reduce 103) ((T_TRAIT) reduce 103) ((T_TRAIT_C) reduce 103) ((T_TRY) reduce 103) ((T_UNSET) reduce 103) ((T_UNSET_CAST) reduce 103) ((T_VARIABLE) reduce 103) ((T_VOID_CAST) reduce 103) ((T_WHILE) reduce 103) ((T_YIELD) reduce 103) ((T_YIELD_FROM) reduce 103) (("[") reduce 103) (("`") reduce 103) (("{") reduce 103) (("~") reduce 103)) 22786 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_EMPTY) reduce 148) ((T_ENDFOR) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_VOID_CAST) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 22787 ((("!") reduce 246) (("\"") reduce 246) (("$") reduce 246) (("(") reduce 246) (("+") reduce 246) (("-") reduce 246) ((";") reduce 246) (("@") reduce 246) ((T_ABSTRACT) reduce 246) ((T_ARRAY) reduce 246) ((T_ARRAY_CAST) reduce 246) ((T_ATTRIBUTE) reduce 246) ((T_BOOL_CAST) reduce 246) ((T_BREAK) reduce 246) ((T_CLASS) reduce 246) ((T_CLASS_C) reduce 246) ((T_CLONE) reduce 246) ((T_CONSTANT_ENCAPSED_STRING) reduce 246) ((T_CONTINUE) reduce 246) ((T_DEC) reduce 246) ((T_DECLARE) reduce 246) ((T_DIR) reduce 246) ((T_DNUMBER) reduce 246) ((T_DO) reduce 246) ((T_DOUBLE_CAST) reduce 246) ((T_ECHO) reduce 246) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) reduce 246) ((T_ENDFOR) reduce 246) ((T_ENUM) reduce 246) ((T_EVAL) reduce 246) ((T_EXIT) reduce 246) ((T_FILE) reduce 246) ((T_FINAL) reduce 246) ((T_FN) reduce 246) ((T_FOR) reduce 246) ((T_FOREACH) reduce 246) ((T_FUNCTION) reduce 246) ((T_FUNC_C) reduce 246) ((T_GLOBAL) reduce 246) ((T_GOTO) reduce 246) ((T_HALT_COMPILER) reduce 246) ((T_IF) reduce 246) ((T_INC) reduce 246) ((T_INCLUDE) reduce 246) ((T_INCLUDE_ONCE) reduce 246) ((T_INLINE_HTML) reduce 246) ((T_INTERFACE) reduce 246) ((T_INT_CAST) reduce 246) ((T_ISSET) reduce 246) ((T_LINE) reduce 246) ((T_LIST) reduce 246) ((T_LNUMBER) reduce 246) ((T_MATCH) reduce 246) ((T_METHOD_C) reduce 246) ((T_NAME_FULLY_QUALIFIED) reduce 246) ((T_NAME_QUALIFIED) reduce 246) ((T_NAME_RELATIVE) reduce 246) ((T_NEW) reduce 246) ((T_NS_C) reduce 246) ((T_OBJECT_CAST) reduce 246) ((T_PRINT) reduce 246) ((T_PROPERTY_C) reduce 246) ((T_READONLY) reduce 246) ((T_REQUIRE) reduce 246) ((T_REQUIRE_ONCE) reduce 246) ((T_RETURN) reduce 246) ((T_START_HEREDOC) reduce 246) ((T_STATIC) reduce 246) ((T_STRING) reduce 246) ((T_STRING_CAST) reduce 246) ((T_SWITCH) reduce 246) ((T_THROW) reduce 246) ((T_TRAIT) reduce 246) ((T_TRAIT_C) reduce 246) ((T_TRY) reduce 246) ((T_UNSET) reduce 246) ((T_UNSET_CAST) reduce 246) ((T_VARIABLE) reduce 246) ((T_VOID_CAST) reduce 246) ((T_WHILE) reduce 246) ((T_YIELD) reduce 246) ((T_YIELD_FROM) reduce 246) (("[") reduce 246) (("`") reduce 246) (("{") reduce 246) (("~") reduce 246)) 22788 ((("!") reduce 141) (("\"") reduce 141) (("$") reduce 141) (("(") reduce 141) (("+") reduce 141) (("-") reduce 141) ((";") reduce 141) (("@") reduce 141) ((T_ABSTRACT) reduce 141) ((T_ARRAY) reduce 141) ((T_ARRAY_CAST) reduce 141) ((T_ATTRIBUTE) reduce 141) ((T_BOOL_CAST) reduce 141) ((T_BREAK) reduce 141) ((T_CLASS) reduce 141) ((T_CLASS_C) reduce 141) ((T_CLONE) reduce 141) ((T_CONSTANT_ENCAPSED_STRING) reduce 141) ((T_CONTINUE) reduce 141) ((T_DEC) reduce 141) ((T_DECLARE) reduce 141) ((T_DIR) reduce 141) ((T_DNUMBER) reduce 141) ((T_DO) reduce 141) ((T_DOUBLE_CAST) reduce 141) ((T_ECHO) reduce 141) ((T_EMPTY) reduce 141) ((T_ENDFOR) reduce 141) ((T_ENUM) reduce 141) ((T_EVAL) reduce 141) ((T_EXIT) reduce 141) ((T_FILE) reduce 141) ((T_FINAL) reduce 141) ((T_FN) reduce 141) ((T_FOR) reduce 141) ((T_FOREACH) reduce 141) ((T_FUNCTION) reduce 141) ((T_FUNC_C) reduce 141) ((T_GLOBAL) reduce 141) ((T_GOTO) reduce 141) ((T_HALT_COMPILER) reduce 141) ((T_IF) reduce 141) ((T_INC) reduce 141) ((T_INCLUDE) reduce 141) ((T_INCLUDE_ONCE) reduce 141) ((T_INLINE_HTML) reduce 141) ((T_INTERFACE) reduce 141) ((T_INT_CAST) reduce 141) ((T_ISSET) reduce 141) ((T_LINE) reduce 141) ((T_LIST) reduce 141) ((T_LNUMBER) reduce 141) ((T_MATCH) reduce 141) ((T_METHOD_C) reduce 141) ((T_NAME_FULLY_QUALIFIED) reduce 141) ((T_NAME_QUALIFIED) reduce 141) ((T_NAME_RELATIVE) reduce 141) ((T_NEW) reduce 141) ((T_NS_C) reduce 141) ((T_OBJECT_CAST) reduce 141) ((T_PRINT) reduce 141) ((T_PROPERTY_C) reduce 141) ((T_READONLY) reduce 141) ((T_REQUIRE) reduce 141) ((T_REQUIRE_ONCE) reduce 141) ((T_RETURN) reduce 141) ((T_START_HEREDOC) reduce 141) ((T_STATIC) reduce 141) ((T_STRING) reduce 141) ((T_STRING_CAST) reduce 141) ((T_SWITCH) reduce 141) ((T_THROW) reduce 141) ((T_TRAIT) reduce 141) ((T_TRAIT_C) reduce 141) ((T_TRY) reduce 141) ((T_UNSET) reduce 141) ((T_UNSET_CAST) reduce 141) ((T_VARIABLE) reduce 141) ((T_VOID_CAST) reduce 141) ((T_WHILE) reduce 141) ((T_YIELD) reduce 141) ((T_YIELD_FROM) reduce 141) (("[") reduce 141) (("`") reduce 141) (("{") reduce 141) (("~") reduce 141)) 22789 ((("!") reduce 106) (("\"") reduce 106) (("$") reduce 106) (("(") reduce 106) (("+") reduce 106) (("-") reduce 106) ((";") reduce 106) (("@") reduce 106) ((T_ABSTRACT) reduce 106) ((T_ARRAY) reduce 106) ((T_ARRAY_CAST) reduce 106) ((T_ATTRIBUTE) reduce 106) ((T_BOOL_CAST) reduce 106) ((T_BREAK) reduce 106) ((T_CLASS) reduce 106) ((T_CLASS_C) reduce 106) ((T_CLONE) reduce 106) ((T_CONSTANT_ENCAPSED_STRING) reduce 106) ((T_CONTINUE) reduce 106) ((T_DEC) reduce 106) ((T_DECLARE) reduce 106) ((T_DIR) reduce 106) ((T_DNUMBER) reduce 106) ((T_DO) reduce 106) ((T_DOUBLE_CAST) reduce 106) ((T_ECHO) reduce 106) ((T_EMPTY) reduce 106) ((T_ENDFOR) reduce 106) ((T_ENUM) reduce 106) ((T_EVAL) reduce 106) ((T_EXIT) reduce 106) ((T_FILE) reduce 106) ((T_FINAL) reduce 106) ((T_FN) reduce 106) ((T_FOR) reduce 106) ((T_FOREACH) reduce 106) ((T_FUNCTION) reduce 106) ((T_FUNC_C) reduce 106) ((T_GLOBAL) reduce 106) ((T_GOTO) reduce 106) ((T_HALT_COMPILER) reduce 106) ((T_IF) reduce 106) ((T_INC) reduce 106) ((T_INCLUDE) reduce 106) ((T_INCLUDE_ONCE) reduce 106) ((T_INLINE_HTML) reduce 106) ((T_INTERFACE) reduce 106) ((T_INT_CAST) reduce 106) ((T_ISSET) reduce 106) ((T_LINE) reduce 106) ((T_LIST) reduce 106) ((T_LNUMBER) reduce 106) ((T_MATCH) reduce 106) ((T_METHOD_C) reduce 106) ((T_NAME_FULLY_QUALIFIED) reduce 106) ((T_NAME_QUALIFIED) reduce 106) ((T_NAME_RELATIVE) reduce 106) ((T_NEW) reduce 106) ((T_NS_C) reduce 106) ((T_OBJECT_CAST) reduce 106) ((T_PRINT) reduce 106) ((T_PROPERTY_C) reduce 106) ((T_READONLY) reduce 106) ((T_REQUIRE) reduce 106) ((T_REQUIRE_ONCE) reduce 106) ((T_RETURN) reduce 106) ((T_START_HEREDOC) reduce 106) ((T_STATIC) reduce 106) ((T_STRING) reduce 106) ((T_STRING_CAST) reduce 106) ((T_SWITCH) reduce 106) ((T_THROW) reduce 106) ((T_TRAIT) reduce 106) ((T_TRAIT_C) reduce 106) ((T_TRY) reduce 106) ((T_UNSET) reduce 106) ((T_UNSET_CAST) reduce 106) ((T_VARIABLE) reduce 106) ((T_VOID_CAST) reduce 106) ((T_WHILE) reduce 106) ((T_YIELD) reduce 106) ((T_YIELD_FROM) reduce 106) (("[") reduce 106) (("`") reduce 106) (("{") reduce 106) (("~") reduce 106)) 22790 ((("!") reduce 143) (("\"") reduce 143) (("$") reduce 143) (("(") reduce 143) (("+") reduce 143) (("-") reduce 143) ((";") reduce 143) (("@") reduce 143) ((T_ABSTRACT) reduce 143) ((T_ARRAY) reduce 143) ((T_ARRAY_CAST) reduce 143) ((T_ATTRIBUTE) reduce 143) ((T_BOOL_CAST) reduce 143) ((T_BREAK) reduce 143) ((T_CLASS) reduce 143) ((T_CLASS_C) reduce 143) ((T_CLONE) reduce 143) ((T_CONSTANT_ENCAPSED_STRING) reduce 143) ((T_CONTINUE) reduce 143) ((T_DEC) reduce 143) ((T_DECLARE) reduce 143) ((T_DIR) reduce 143) ((T_DNUMBER) reduce 143) ((T_DO) reduce 143) ((T_DOUBLE_CAST) reduce 143) ((T_ECHO) reduce 143) ((T_EMPTY) reduce 143) ((T_ENDFOR) reduce 143) ((T_ENUM) reduce 143) ((T_EVAL) reduce 143) ((T_EXIT) reduce 143) ((T_FILE) reduce 143) ((T_FINAL) reduce 143) ((T_FN) reduce 143) ((T_FOR) reduce 143) ((T_FOREACH) reduce 143) ((T_FUNCTION) reduce 143) ((T_FUNC_C) reduce 143) ((T_GLOBAL) reduce 143) ((T_GOTO) reduce 143) ((T_HALT_COMPILER) reduce 143) ((T_IF) reduce 143) ((T_INC) reduce 143) ((T_INCLUDE) reduce 143) ((T_INCLUDE_ONCE) reduce 143) ((T_INLINE_HTML) reduce 143) ((T_INTERFACE) reduce 143) ((T_INT_CAST) reduce 143) ((T_ISSET) reduce 143) ((T_LINE) reduce 143) ((T_LIST) reduce 143) ((T_LNUMBER) reduce 143) ((T_MATCH) reduce 143) ((T_METHOD_C) reduce 143) ((T_NAME_FULLY_QUALIFIED) reduce 143) ((T_NAME_QUALIFIED) reduce 143) ((T_NAME_RELATIVE) reduce 143) ((T_NEW) reduce 143) ((T_NS_C) reduce 143) ((T_OBJECT_CAST) reduce 143) ((T_PRINT) reduce 143) ((T_PROPERTY_C) reduce 143) ((T_READONLY) reduce 143) ((T_REQUIRE) reduce 143) ((T_REQUIRE_ONCE) reduce 143) ((T_RETURN) reduce 143) ((T_START_HEREDOC) reduce 143) ((T_STATIC) reduce 143) ((T_STRING) reduce 143) ((T_STRING_CAST) reduce 143) ((T_SWITCH) reduce 143) ((T_THROW) reduce 143) ((T_TRAIT) reduce 143) ((T_TRAIT_C) reduce 143) ((T_TRY) reduce 143) ((T_UNSET) reduce 143) ((T_UNSET_CAST) reduce 143) ((T_VARIABLE) reduce 143) ((T_VOID_CAST) reduce 143) ((T_WHILE) reduce 143) ((T_YIELD) reduce 143) ((T_YIELD_FROM) reduce 143) (("[") reduce 143) (("`") reduce 143) (("{") reduce 143) (("~") reduce 143)) 22791 ((("!") reduce 105) (("\"") reduce 105) (("$") reduce 105) (("(") reduce 105) (("+") reduce 105) (("-") reduce 105) ((";") reduce 105) (("@") reduce 105) ((T_ABSTRACT) reduce 105) ((T_ARRAY) reduce 105) ((T_ARRAY_CAST) reduce 105) ((T_ATTRIBUTE) reduce 105) ((T_BOOL_CAST) reduce 105) ((T_BREAK) reduce 105) ((T_CLASS) reduce 105) ((T_CLASS_C) reduce 105) ((T_CLONE) reduce 105) ((T_CONSTANT_ENCAPSED_STRING) reduce 105) ((T_CONTINUE) reduce 105) ((T_DEC) reduce 105) ((T_DECLARE) reduce 105) ((T_DIR) reduce 105) ((T_DNUMBER) reduce 105) ((T_DO) reduce 105) ((T_DOUBLE_CAST) reduce 105) ((T_ECHO) reduce 105) ((T_EMPTY) reduce 105) ((T_ENDFOR) reduce 105) ((T_ENUM) reduce 105) ((T_EVAL) reduce 105) ((T_EXIT) reduce 105) ((T_FILE) reduce 105) ((T_FINAL) reduce 105) ((T_FN) reduce 105) ((T_FOR) reduce 105) ((T_FOREACH) reduce 105) ((T_FUNCTION) reduce 105) ((T_FUNC_C) reduce 105) ((T_GLOBAL) reduce 105) ((T_GOTO) reduce 105) ((T_HALT_COMPILER) reduce 105) ((T_IF) reduce 105) ((T_INC) reduce 105) ((T_INCLUDE) reduce 105) ((T_INCLUDE_ONCE) reduce 105) ((T_INLINE_HTML) reduce 105) ((T_INTERFACE) reduce 105) ((T_INT_CAST) reduce 105) ((T_ISSET) reduce 105) ((T_LINE) reduce 105) ((T_LIST) reduce 105) ((T_LNUMBER) reduce 105) ((T_MATCH) reduce 105) ((T_METHOD_C) reduce 105) ((T_NAME_FULLY_QUALIFIED) reduce 105) ((T_NAME_QUALIFIED) reduce 105) ((T_NAME_RELATIVE) reduce 105) ((T_NEW) reduce 105) ((T_NS_C) reduce 105) ((T_OBJECT_CAST) reduce 105) ((T_PRINT) reduce 105) ((T_PROPERTY_C) reduce 105) ((T_READONLY) reduce 105) ((T_REQUIRE) reduce 105) ((T_REQUIRE_ONCE) reduce 105) ((T_RETURN) reduce 105) ((T_START_HEREDOC) reduce 105) ((T_STATIC) reduce 105) ((T_STRING) reduce 105) ((T_STRING_CAST) reduce 105) ((T_SWITCH) reduce 105) ((T_THROW) reduce 105) ((T_TRAIT) reduce 105) ((T_TRAIT_C) reduce 105) ((T_TRY) reduce 105) ((T_UNSET) reduce 105) ((T_UNSET_CAST) reduce 105) ((T_VARIABLE) reduce 105) ((T_VOID_CAST) reduce 105) ((T_WHILE) reduce 105) ((T_YIELD) reduce 105) ((T_YIELD_FROM) reduce 105) (("[") reduce 105) (("`") reduce 105) (("{") reduce 105) (("~") reduce 105)) 22794 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_EMPTY) reduce 147) ((T_ENDFOR) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_VOID_CAST) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 22800 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_ELSE) reduce 166) ((T_ELSEIF) reduce 166) ((T_EMPTY) reduce 166) ((T_ENDFOR) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_VOID_CAST) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 22810 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_ELSE) reduce 160) ((T_ELSEIF) reduce 160) ((T_EMPTY) reduce 160) ((T_ENDFOR) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_VOID_CAST) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 22819 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_ELSE) reduce 149) ((T_ELSEIF) reduce 149) ((T_EMPTY) reduce 149) ((T_ENDFOR) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_VOID_CAST) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 22822 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_ELSE) reduce 148) ((T_ELSEIF) reduce 148) ((T_EMPTY) reduce 148) ((T_ENDFOR) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_VOID_CAST) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 22824 ((("!") reduce 245) (("\"") reduce 245) (("$") reduce 245) (("(") reduce 245) (("+") reduce 245) (("-") reduce 245) ((";") reduce 245) (("@") reduce 245) ((T_ABSTRACT) reduce 245) ((T_ARRAY) reduce 245) ((T_ARRAY_CAST) reduce 245) ((T_ATTRIBUTE) reduce 245) ((T_BOOL_CAST) reduce 245) ((T_BREAK) reduce 245) ((T_CLASS) reduce 245) ((T_CLASS_C) reduce 245) ((T_CLONE) reduce 245) ((T_CONSTANT_ENCAPSED_STRING) reduce 245) ((T_CONTINUE) reduce 245) ((T_DEC) reduce 245) ((T_DECLARE) reduce 245) ((T_DIR) reduce 245) ((T_DNUMBER) reduce 245) ((T_DO) reduce 245) ((T_DOUBLE_CAST) reduce 245) ((T_ECHO) reduce 245) ((T_ELSE) reduce 245) ((T_ELSEIF) reduce 245) ((T_EMPTY) reduce 245) ((T_ENDFOR) reduce 245) ((T_ENUM) reduce 245) ((T_EVAL) reduce 245) ((T_EXIT) reduce 245) ((T_FILE) reduce 245) ((T_FINAL) reduce 245) ((T_FN) reduce 245) ((T_FOR) reduce 245) ((T_FOREACH) reduce 245) ((T_FUNCTION) reduce 245) ((T_FUNC_C) reduce 245) ((T_GLOBAL) reduce 245) ((T_GOTO) reduce 245) ((T_HALT_COMPILER) reduce 245) ((T_IF) reduce 245) ((T_INC) reduce 245) ((T_INCLUDE) reduce 245) ((T_INCLUDE_ONCE) reduce 245) ((T_INLINE_HTML) reduce 245) ((T_INTERFACE) reduce 245) ((T_INT_CAST) reduce 245) ((T_ISSET) reduce 245) ((T_LINE) reduce 245) ((T_LIST) reduce 245) ((T_LNUMBER) reduce 245) ((T_MATCH) reduce 245) ((T_METHOD_C) reduce 245) ((T_NAME_FULLY_QUALIFIED) reduce 245) ((T_NAME_QUALIFIED) reduce 245) ((T_NAME_RELATIVE) reduce 245) ((T_NEW) reduce 245) ((T_NS_C) reduce 245) ((T_OBJECT_CAST) reduce 245) ((T_PRINT) reduce 245) ((T_PROPERTY_C) reduce 245) ((T_READONLY) reduce 245) ((T_REQUIRE) reduce 245) ((T_REQUIRE_ONCE) reduce 245) ((T_RETURN) reduce 245) ((T_START_HEREDOC) reduce 245) ((T_STATIC) reduce 245) ((T_STRING) reduce 245) ((T_STRING_CAST) reduce 245) ((T_SWITCH) reduce 245) ((T_THROW) reduce 245) ((T_TRAIT) reduce 245) ((T_TRAIT_C) reduce 245) ((T_TRY) reduce 245) ((T_UNSET) reduce 245) ((T_UNSET_CAST) reduce 245) ((T_VARIABLE) reduce 245) ((T_VOID_CAST) reduce 245) ((T_WHILE) reduce 245) ((T_YIELD) reduce 245) ((T_YIELD_FROM) reduce 245) (("[") reduce 245) (("`") reduce 245) (("{") reduce 245) (("~") reduce 245)) 22827 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_ELSE) reduce 147) ((T_ELSEIF) reduce 147) ((T_EMPTY) reduce 147) ((T_ENDFOR) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_VOID_CAST) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 22829 ((("!") reduce 247) (("\"") reduce 247) (("$") reduce 247) (("(") reduce 247) (("+") reduce 247) (("-") reduce 247) ((";") reduce 247) (("@") reduce 247) ((T_ABSTRACT) reduce 247) ((T_ARRAY) reduce 247) ((T_ARRAY_CAST) reduce 247) ((T_ATTRIBUTE) reduce 247) ((T_BOOL_CAST) reduce 247) ((T_BREAK) reduce 247) ((T_CLASS) reduce 247) ((T_CLASS_C) reduce 247) ((T_CLONE) reduce 247) ((T_CONSTANT_ENCAPSED_STRING) reduce 247) ((T_CONTINUE) reduce 247) ((T_DEC) reduce 247) ((T_DECLARE) reduce 247) ((T_DIR) reduce 247) ((T_DNUMBER) reduce 247) ((T_DO) reduce 247) ((T_DOUBLE_CAST) reduce 247) ((T_ECHO) reduce 247) ((T_ELSE) reduce 247) ((T_ELSEIF) reduce 247) ((T_EMPTY) reduce 247) ((T_ENDFOR) reduce 247) ((T_ENUM) reduce 247) ((T_EVAL) reduce 247) ((T_EXIT) reduce 247) ((T_FILE) reduce 247) ((T_FINAL) reduce 247) ((T_FN) reduce 247) ((T_FOR) reduce 247) ((T_FOREACH) reduce 247) ((T_FUNCTION) reduce 247) ((T_FUNC_C) reduce 247) ((T_GLOBAL) reduce 247) ((T_GOTO) reduce 247) ((T_HALT_COMPILER) reduce 247) ((T_IF) reduce 247) ((T_INC) reduce 247) ((T_INCLUDE) reduce 247) ((T_INCLUDE_ONCE) reduce 247) ((T_INLINE_HTML) reduce 247) ((T_INTERFACE) reduce 247) ((T_INT_CAST) reduce 247) ((T_ISSET) reduce 247) ((T_LINE) reduce 247) ((T_LIST) reduce 247) ((T_LNUMBER) reduce 247) ((T_MATCH) reduce 247) ((T_METHOD_C) reduce 247) ((T_NAME_FULLY_QUALIFIED) reduce 247) ((T_NAME_QUALIFIED) reduce 247) ((T_NAME_RELATIVE) reduce 247) ((T_NEW) reduce 247) ((T_NS_C) reduce 247) ((T_OBJECT_CAST) reduce 247) ((T_PRINT) reduce 247) ((T_PROPERTY_C) reduce 247) ((T_READONLY) reduce 247) ((T_REQUIRE) reduce 247) ((T_REQUIRE_ONCE) reduce 247) ((T_RETURN) reduce 247) ((T_START_HEREDOC) reduce 247) ((T_STATIC) reduce 247) ((T_STRING) reduce 247) ((T_STRING_CAST) reduce 247) ((T_SWITCH) reduce 247) ((T_THROW) reduce 247) ((T_TRAIT) reduce 247) ((T_TRAIT_C) reduce 247) ((T_TRY) reduce 247) ((T_UNSET) reduce 247) ((T_UNSET_CAST) reduce 247) ((T_VARIABLE) reduce 247) ((T_VOID_CAST) reduce 247) ((T_WHILE) reduce 247) ((T_YIELD) reduce 247) ((T_YIELD_FROM) reduce 247) (("[") reduce 247) (("`") reduce 247) (("{") reduce 247) (("~") reduce 247)) 22830 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_ELSE) reduce 161) ((T_ELSEIF) reduce 161) ((T_EMPTY) reduce 161) ((T_ENDFOR) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_VOID_CAST) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 22833 ((("!") reduce 250) (("\"") reduce 250) (("$") reduce 250) (("(") reduce 250) (("+") reduce 250) (("-") reduce 250) ((";") reduce 250) (("@") reduce 250) ((T_ABSTRACT) reduce 250) ((T_ARRAY) reduce 250) ((T_ARRAY_CAST) reduce 250) ((T_ATTRIBUTE) reduce 250) ((T_BOOL_CAST) reduce 250) ((T_BREAK) reduce 250) ((T_CLASS) reduce 250) ((T_CLASS_C) reduce 250) ((T_CLONE) reduce 250) ((T_CONSTANT_ENCAPSED_STRING) reduce 250) ((T_CONTINUE) reduce 250) ((T_DEC) reduce 250) ((T_DECLARE) reduce 250) ((T_DIR) reduce 250) ((T_DNUMBER) reduce 250) ((T_DO) reduce 250) ((T_DOUBLE_CAST) reduce 250) ((T_ECHO) reduce 250) ((T_ELSE) reduce 250) ((T_ELSEIF) reduce 250) ((T_EMPTY) reduce 250) ((T_ENDFOR) reduce 250) ((T_ENUM) reduce 250) ((T_EVAL) reduce 250) ((T_EXIT) reduce 250) ((T_FILE) reduce 250) ((T_FINAL) reduce 250) ((T_FN) reduce 250) ((T_FOR) reduce 250) ((T_FOREACH) reduce 250) ((T_FUNCTION) reduce 250) ((T_FUNC_C) reduce 250) ((T_GLOBAL) reduce 250) ((T_GOTO) reduce 250) ((T_HALT_COMPILER) reduce 250) ((T_IF) reduce 250) ((T_INC) reduce 250) ((T_INCLUDE) reduce 250) ((T_INCLUDE_ONCE) reduce 250) ((T_INLINE_HTML) reduce 250) ((T_INTERFACE) reduce 250) ((T_INT_CAST) reduce 250) ((T_ISSET) reduce 250) ((T_LINE) reduce 250) ((T_LIST) reduce 250) ((T_LNUMBER) reduce 250) ((T_MATCH) reduce 250) ((T_METHOD_C) reduce 250) ((T_NAME_FULLY_QUALIFIED) reduce 250) ((T_NAME_QUALIFIED) reduce 250) ((T_NAME_RELATIVE) reduce 250) ((T_NEW) reduce 250) ((T_NS_C) reduce 250) ((T_OBJECT_CAST) reduce 250) ((T_PRINT) reduce 250) ((T_PROPERTY_C) reduce 250) ((T_READONLY) reduce 250) ((T_REQUIRE) reduce 250) ((T_REQUIRE_ONCE) reduce 250) ((T_RETURN) reduce 250) ((T_START_HEREDOC) reduce 250) ((T_STATIC) reduce 250) ((T_STRING) reduce 250) ((T_STRING_CAST) reduce 250) ((T_SWITCH) reduce 250) ((T_THROW) reduce 250) ((T_TRAIT) reduce 250) ((T_TRAIT_C) reduce 250) ((T_TRY) reduce 250) ((T_UNSET) reduce 250) ((T_UNSET_CAST) reduce 250) ((T_VARIABLE) reduce 250) ((T_VOID_CAST) reduce 250) ((T_WHILE) reduce 250) ((T_YIELD) reduce 250) ((T_YIELD_FROM) reduce 250) (("[") reduce 250) (("`") reduce 250) (("{") reduce 250) (("~") reduce 250)) 22837 ((("!") reduce 251) (("\"") reduce 251) (("$") reduce 251) (("(") reduce 251) (("+") reduce 251) (("-") reduce 251) ((";") reduce 251) (("@") reduce 251) ((T_ABSTRACT) reduce 251) ((T_ARRAY) reduce 251) ((T_ARRAY_CAST) reduce 251) ((T_ATTRIBUTE) reduce 251) ((T_BOOL_CAST) reduce 251) ((T_BREAK) reduce 251) ((T_CLASS) reduce 251) ((T_CLASS_C) reduce 251) ((T_CLONE) reduce 251) ((T_CONSTANT_ENCAPSED_STRING) reduce 251) ((T_CONTINUE) reduce 251) ((T_DEC) reduce 251) ((T_DECLARE) reduce 251) ((T_DIR) reduce 251) ((T_DNUMBER) reduce 251) ((T_DO) reduce 251) ((T_DOUBLE_CAST) reduce 251) ((T_ECHO) reduce 251) ((T_ELSE) reduce 251) ((T_ELSEIF) reduce 251) ((T_EMPTY) reduce 251) ((T_ENDFOR) reduce 251) ((T_ENUM) reduce 251) ((T_EVAL) reduce 251) ((T_EXIT) reduce 251) ((T_FILE) reduce 251) ((T_FINAL) reduce 251) ((T_FN) reduce 251) ((T_FOR) reduce 251) ((T_FOREACH) reduce 251) ((T_FUNCTION) reduce 251) ((T_FUNC_C) reduce 251) ((T_GLOBAL) reduce 251) ((T_GOTO) reduce 251) ((T_HALT_COMPILER) reduce 251) ((T_IF) reduce 251) ((T_INC) reduce 251) ((T_INCLUDE) reduce 251) ((T_INCLUDE_ONCE) reduce 251) ((T_INLINE_HTML) reduce 251) ((T_INTERFACE) reduce 251) ((T_INT_CAST) reduce 251) ((T_ISSET) reduce 251) ((T_LINE) reduce 251) ((T_LIST) reduce 251) ((T_LNUMBER) reduce 251) ((T_MATCH) reduce 251) ((T_METHOD_C) reduce 251) ((T_NAME_FULLY_QUALIFIED) reduce 251) ((T_NAME_QUALIFIED) reduce 251) ((T_NAME_RELATIVE) reduce 251) ((T_NEW) reduce 251) ((T_NS_C) reduce 251) ((T_OBJECT_CAST) reduce 251) ((T_PRINT) reduce 251) ((T_PROPERTY_C) reduce 251) ((T_READONLY) reduce 251) ((T_REQUIRE) reduce 251) ((T_REQUIRE_ONCE) reduce 251) ((T_RETURN) reduce 251) ((T_START_HEREDOC) reduce 251) ((T_STATIC) reduce 251) ((T_STRING) reduce 251) ((T_STRING_CAST) reduce 251) ((T_SWITCH) reduce 251) ((T_THROW) reduce 251) ((T_TRAIT) reduce 251) ((T_TRAIT_C) reduce 251) ((T_TRY) reduce 251) ((T_UNSET) reduce 251) ((T_UNSET_CAST) reduce 251) ((T_VARIABLE) reduce 251) ((T_VOID_CAST) reduce 251) ((T_WHILE) reduce 251) ((T_YIELD) reduce 251) ((T_YIELD_FROM) reduce 251) (("[") reduce 251) (("`") reduce 251) (("{") reduce 251) (("~") reduce 251)) 22842 ((("!") reduce 242) (("\"") reduce 242) (("$") reduce 242) (("(") reduce 242) (("+") reduce 242) (("-") reduce 242) ((";") reduce 242) (("@") reduce 242) ((T_ABSTRACT) reduce 242) ((T_ARRAY) reduce 242) ((T_ARRAY_CAST) reduce 242) ((T_ATTRIBUTE) reduce 242) ((T_BOOL_CAST) reduce 242) ((T_BREAK) reduce 242) ((T_CLASS) reduce 242) ((T_CLASS_C) reduce 242) ((T_CLONE) reduce 242) ((T_CONSTANT_ENCAPSED_STRING) reduce 242) ((T_CONTINUE) reduce 242) ((T_DEC) reduce 242) ((T_DECLARE) reduce 242) ((T_DIR) reduce 242) ((T_DNUMBER) reduce 242) ((T_DO) reduce 242) ((T_DOUBLE_CAST) reduce 242) ((T_ECHO) reduce 242) ((T_ELSE) reduce 242) ((T_ELSEIF) reduce 242) ((T_EMPTY) reduce 242) ((T_ENDFOR) reduce 242) ((T_ENUM) reduce 242) ((T_EVAL) reduce 242) ((T_EXIT) reduce 242) ((T_FILE) reduce 242) ((T_FINAL) reduce 242) ((T_FN) reduce 242) ((T_FOR) reduce 242) ((T_FOREACH) reduce 242) ((T_FUNCTION) reduce 242) ((T_FUNC_C) reduce 242) ((T_GLOBAL) reduce 242) ((T_GOTO) reduce 242) ((T_HALT_COMPILER) reduce 242) ((T_IF) reduce 242) ((T_INC) reduce 242) ((T_INCLUDE) reduce 242) ((T_INCLUDE_ONCE) reduce 242) ((T_INLINE_HTML) reduce 242) ((T_INTERFACE) reduce 242) ((T_INT_CAST) reduce 242) ((T_ISSET) reduce 242) ((T_LINE) reduce 242) ((T_LIST) reduce 242) ((T_LNUMBER) reduce 242) ((T_MATCH) reduce 242) ((T_METHOD_C) reduce 242) ((T_NAME_FULLY_QUALIFIED) reduce 242) ((T_NAME_QUALIFIED) reduce 242) ((T_NAME_RELATIVE) reduce 242) ((T_NEW) reduce 242) ((T_NS_C) reduce 242) ((T_OBJECT_CAST) reduce 242) ((T_PRINT) reduce 242) ((T_PROPERTY_C) reduce 242) ((T_READONLY) reduce 242) ((T_REQUIRE) reduce 242) ((T_REQUIRE_ONCE) reduce 242) ((T_RETURN) reduce 242) ((T_START_HEREDOC) reduce 242) ((T_STATIC) reduce 242) ((T_STRING) reduce 242) ((T_STRING_CAST) reduce 242) ((T_SWITCH) reduce 242) ((T_THROW) reduce 242) ((T_TRAIT) reduce 242) ((T_TRAIT_C) reduce 242) ((T_TRY) reduce 242) ((T_UNSET) reduce 242) ((T_UNSET_CAST) reduce 242) ((T_VARIABLE) reduce 242) ((T_VOID_CAST) reduce 242) ((T_WHILE) reduce 242) ((T_YIELD) reduce 242) ((T_YIELD_FROM) reduce 242) (("[") reduce 242) (("`") reduce 242) (("{") reduce 242) (("~") reduce 242)) 22843 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_ELSE) reduce 150) ((T_ELSEIF) reduce 150) ((T_EMPTY) reduce 150) ((T_ENDFOR) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_VOID_CAST) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 22846 ((("!") reduce 243) (("\"") reduce 243) (("$") reduce 243) (("(") reduce 243) (("+") reduce 243) (("-") reduce 243) ((";") reduce 243) (("@") reduce 243) ((T_ABSTRACT) reduce 243) ((T_ARRAY) reduce 243) ((T_ARRAY_CAST) reduce 243) ((T_ATTRIBUTE) reduce 243) ((T_BOOL_CAST) reduce 243) ((T_BREAK) reduce 243) ((T_CLASS) reduce 243) ((T_CLASS_C) reduce 243) ((T_CLONE) reduce 243) ((T_CONSTANT_ENCAPSED_STRING) reduce 243) ((T_CONTINUE) reduce 243) ((T_DEC) reduce 243) ((T_DECLARE) reduce 243) ((T_DIR) reduce 243) ((T_DNUMBER) reduce 243) ((T_DO) reduce 243) ((T_DOUBLE_CAST) reduce 243) ((T_ECHO) reduce 243) ((T_ELSE) reduce 243) ((T_ELSEIF) reduce 243) ((T_EMPTY) reduce 243) ((T_ENDFOR) reduce 243) ((T_ENUM) reduce 243) ((T_EVAL) reduce 243) ((T_EXIT) reduce 243) ((T_FILE) reduce 243) ((T_FINAL) reduce 243) ((T_FN) reduce 243) ((T_FOR) reduce 243) ((T_FOREACH) reduce 243) ((T_FUNCTION) reduce 243) ((T_FUNC_C) reduce 243) ((T_GLOBAL) reduce 243) ((T_GOTO) reduce 243) ((T_HALT_COMPILER) reduce 243) ((T_IF) reduce 243) ((T_INC) reduce 243) ((T_INCLUDE) reduce 243) ((T_INCLUDE_ONCE) reduce 243) ((T_INLINE_HTML) reduce 243) ((T_INTERFACE) reduce 243) ((T_INT_CAST) reduce 243) ((T_ISSET) reduce 243) ((T_LINE) reduce 243) ((T_LIST) reduce 243) ((T_LNUMBER) reduce 243) ((T_MATCH) reduce 243) ((T_METHOD_C) reduce 243) ((T_NAME_FULLY_QUALIFIED) reduce 243) ((T_NAME_QUALIFIED) reduce 243) ((T_NAME_RELATIVE) reduce 243) ((T_NEW) reduce 243) ((T_NS_C) reduce 243) ((T_OBJECT_CAST) reduce 243) ((T_PRINT) reduce 243) ((T_PROPERTY_C) reduce 243) ((T_READONLY) reduce 243) ((T_REQUIRE) reduce 243) ((T_REQUIRE_ONCE) reduce 243) ((T_RETURN) reduce 243) ((T_START_HEREDOC) reduce 243) ((T_STATIC) reduce 243) ((T_STRING) reduce 243) ((T_STRING_CAST) reduce 243) ((T_SWITCH) reduce 243) ((T_THROW) reduce 243) ((T_TRAIT) reduce 243) ((T_TRAIT_C) reduce 243) ((T_TRY) reduce 243) ((T_UNSET) reduce 243) ((T_UNSET_CAST) reduce 243) ((T_VARIABLE) reduce 243) ((T_VOID_CAST) reduce 243) ((T_WHILE) reduce 243) ((T_YIELD) reduce 243) ((T_YIELD_FROM) reduce 243) (("[") reduce 243) (("`") reduce 243) (("{") reduce 243) (("~") reduce 243)) 22848 ((("!") reduce 170) (("\"") reduce 170) (("$") reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_ELSE) reduce 170) ((T_ELSEIF) reduce 170) ((T_EMPTY) reduce 170) ((T_ENDFOR) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_VARIABLE) reduce 170) ((T_VOID_CAST) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("~") reduce 170)) 22853 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_ELSE) reduce 162) ((T_ELSEIF) reduce 162) ((T_EMPTY) reduce 162) ((T_ENDFOR) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_VOID_CAST) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 22856 ((("!") reduce 171) (("\"") reduce 171) (("$") reduce 171) (("(") reduce 171) (("+") reduce 171) (("-") reduce 171) ((";") reduce 171) (("@") reduce 171) ((T_ABSTRACT) reduce 171) ((T_ARRAY) reduce 171) ((T_ARRAY_CAST) reduce 171) ((T_ATTRIBUTE) reduce 171) ((T_BOOL_CAST) reduce 171) ((T_BREAK) reduce 171) ((T_CATCH) reduce 171) ((T_CLASS) reduce 171) ((T_CLASS_C) reduce 171) ((T_CLONE) reduce 171) ((T_CONSTANT_ENCAPSED_STRING) reduce 171) ((T_CONTINUE) reduce 171) ((T_DEC) reduce 171) ((T_DECLARE) reduce 171) ((T_DIR) reduce 171) ((T_DNUMBER) reduce 171) ((T_DO) reduce 171) ((T_DOUBLE_CAST) reduce 171) ((T_ECHO) reduce 171) ((T_ELSE) reduce 171) ((T_ELSEIF) reduce 171) ((T_EMPTY) reduce 171) ((T_ENDFOR) reduce 171) ((T_ENUM) reduce 171) ((T_EVAL) reduce 171) ((T_EXIT) reduce 171) ((T_FILE) reduce 171) ((T_FINAL) reduce 171) ((T_FINALLY) reduce 171) ((T_FN) reduce 171) ((T_FOR) reduce 171) ((T_FOREACH) reduce 171) ((T_FUNCTION) reduce 171) ((T_FUNC_C) reduce 171) ((T_GLOBAL) reduce 171) ((T_GOTO) reduce 171) ((T_HALT_COMPILER) reduce 171) ((T_IF) reduce 171) ((T_INC) reduce 171) ((T_INCLUDE) reduce 171) ((T_INCLUDE_ONCE) reduce 171) ((T_INLINE_HTML) reduce 171) ((T_INTERFACE) reduce 171) ((T_INT_CAST) reduce 171) ((T_ISSET) reduce 171) ((T_LINE) reduce 171) ((T_LIST) reduce 171) ((T_LNUMBER) reduce 171) ((T_MATCH) reduce 171) ((T_METHOD_C) reduce 171) ((T_NAME_FULLY_QUALIFIED) reduce 171) ((T_NAME_QUALIFIED) reduce 171) ((T_NAME_RELATIVE) reduce 171) ((T_NEW) reduce 171) ((T_NS_C) reduce 171) ((T_OBJECT_CAST) reduce 171) ((T_PRINT) reduce 171) ((T_PROPERTY_C) reduce 171) ((T_READONLY) reduce 171) ((T_REQUIRE) reduce 171) ((T_REQUIRE_ONCE) reduce 171) ((T_RETURN) reduce 171) ((T_START_HEREDOC) reduce 171) ((T_STATIC) reduce 171) ((T_STRING) reduce 171) ((T_STRING_CAST) reduce 171) ((T_SWITCH) reduce 171) ((T_THROW) reduce 171) ((T_TRAIT) reduce 171) ((T_TRAIT_C) reduce 171) ((T_TRY) reduce 171) ((T_UNSET) reduce 171) ((T_UNSET_CAST) reduce 171) ((T_VARIABLE) reduce 171) ((T_VOID_CAST) reduce 171) ((T_WHILE) reduce 171) ((T_YIELD) reduce 171) ((T_YIELD_FROM) reduce 171) (("[") reduce 171) (("`") reduce 171) (("{") reduce 171) (("~") reduce 171)) 22857 ((("!") reduce 177) (("\"") reduce 177) (("$") reduce 177) (("(") reduce 177) (("+") reduce 177) (("-") reduce 177) ((";") reduce 177) (("@") reduce 177) ((T_ABSTRACT) reduce 177) ((T_ARRAY) reduce 177) ((T_ARRAY_CAST) reduce 177) ((T_ATTRIBUTE) reduce 177) ((T_BOOL_CAST) reduce 177) ((T_BREAK) reduce 177) ((T_CATCH) shift) ((T_CLASS) reduce 177) ((T_CLASS_C) reduce 177) ((T_CLONE) reduce 177) ((T_CONSTANT_ENCAPSED_STRING) reduce 177) ((T_CONTINUE) reduce 177) ((T_DEC) reduce 177) ((T_DECLARE) reduce 177) ((T_DIR) reduce 177) ((T_DNUMBER) reduce 177) ((T_DO) reduce 177) ((T_DOUBLE_CAST) reduce 177) ((T_ECHO) reduce 177) ((T_ELSE) reduce 177) ((T_ELSEIF) reduce 177) ((T_EMPTY) reduce 177) ((T_ENDFOR) reduce 177) ((T_ENUM) reduce 177) ((T_EVAL) reduce 177) ((T_EXIT) reduce 177) ((T_FILE) reduce 177) ((T_FINAL) reduce 177) ((T_FINALLY) shift) ((T_FN) reduce 177) ((T_FOR) reduce 177) ((T_FOREACH) reduce 177) ((T_FUNCTION) reduce 177) ((T_FUNC_C) reduce 177) ((T_GLOBAL) reduce 177) ((T_GOTO) reduce 177) ((T_HALT_COMPILER) reduce 177) ((T_IF) reduce 177) ((T_INC) reduce 177) ((T_INCLUDE) reduce 177) ((T_INCLUDE_ONCE) reduce 177) ((T_INLINE_HTML) reduce 177) ((T_INTERFACE) reduce 177) ((T_INT_CAST) reduce 177) ((T_ISSET) reduce 177) ((T_LINE) reduce 177) ((T_LIST) reduce 177) ((T_LNUMBER) reduce 177) ((T_MATCH) reduce 177) ((T_METHOD_C) reduce 177) ((T_NAME_FULLY_QUALIFIED) reduce 177) ((T_NAME_QUALIFIED) reduce 177) ((T_NAME_RELATIVE) reduce 177) ((T_NEW) reduce 177) ((T_NS_C) reduce 177) ((T_OBJECT_CAST) reduce 177) ((T_PRINT) reduce 177) ((T_PROPERTY_C) reduce 177) ((T_READONLY) reduce 177) ((T_REQUIRE) reduce 177) ((T_REQUIRE_ONCE) reduce 177) ((T_RETURN) reduce 177) ((T_START_HEREDOC) reduce 177) ((T_STATIC) reduce 177) ((T_STRING) reduce 177) ((T_STRING_CAST) reduce 177) ((T_SWITCH) reduce 177) ((T_THROW) reduce 177) ((T_TRAIT) reduce 177) ((T_TRAIT_C) reduce 177) ((T_TRY) reduce 177) ((T_UNSET) reduce 177) ((T_UNSET_CAST) reduce 177) ((T_VARIABLE) reduce 177) ((T_VOID_CAST) reduce 177) ((T_WHILE) reduce 177) ((T_YIELD) reduce 177) ((T_YIELD_FROM) reduce 177) (("[") reduce 177) (("`") reduce 177) (("{") reduce 177) (("~") reduce 177)) 22860 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_ELSE) reduce 167) ((T_ELSEIF) reduce 167) ((T_EMPTY) reduce 167) ((T_ENDFOR) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_VOID_CAST) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 22863 ((("!") reduce 178) (("\"") reduce 178) (("$") reduce 178) (("(") reduce 178) (("+") reduce 178) (("-") reduce 178) ((";") reduce 178) (("@") reduce 178) ((T_ABSTRACT) reduce 178) ((T_ARRAY) reduce 178) ((T_ARRAY_CAST) reduce 178) ((T_ATTRIBUTE) reduce 178) ((T_BOOL_CAST) reduce 178) ((T_BREAK) reduce 178) ((T_CLASS) reduce 178) ((T_CLASS_C) reduce 178) ((T_CLONE) reduce 178) ((T_CONSTANT_ENCAPSED_STRING) reduce 178) ((T_CONTINUE) reduce 178) ((T_DEC) reduce 178) ((T_DECLARE) reduce 178) ((T_DIR) reduce 178) ((T_DNUMBER) reduce 178) ((T_DO) reduce 178) ((T_DOUBLE_CAST) reduce 178) ((T_ECHO) reduce 178) ((T_ELSE) reduce 178) ((T_ELSEIF) reduce 178) ((T_EMPTY) reduce 178) ((T_ENDFOR) reduce 178) ((T_ENUM) reduce 178) ((T_EVAL) reduce 178) ((T_EXIT) reduce 178) ((T_FILE) reduce 178) ((T_FINAL) reduce 178) ((T_FN) reduce 178) ((T_FOR) reduce 178) ((T_FOREACH) reduce 178) ((T_FUNCTION) reduce 178) ((T_FUNC_C) reduce 178) ((T_GLOBAL) reduce 178) ((T_GOTO) reduce 178) ((T_HALT_COMPILER) reduce 178) ((T_IF) reduce 178) ((T_INC) reduce 178) ((T_INCLUDE) reduce 178) ((T_INCLUDE_ONCE) reduce 178) ((T_INLINE_HTML) reduce 178) ((T_INTERFACE) reduce 178) ((T_INT_CAST) reduce 178) ((T_ISSET) reduce 178) ((T_LINE) reduce 178) ((T_LIST) reduce 178) ((T_LNUMBER) reduce 178) ((T_MATCH) reduce 178) ((T_METHOD_C) reduce 178) ((T_NAME_FULLY_QUALIFIED) reduce 178) ((T_NAME_QUALIFIED) reduce 178) ((T_NAME_RELATIVE) reduce 178) ((T_NEW) reduce 178) ((T_NS_C) reduce 178) ((T_OBJECT_CAST) reduce 178) ((T_PRINT) reduce 178) ((T_PROPERTY_C) reduce 178) ((T_READONLY) reduce 178) ((T_REQUIRE) reduce 178) ((T_REQUIRE_ONCE) reduce 178) ((T_RETURN) reduce 178) ((T_START_HEREDOC) reduce 178) ((T_STATIC) reduce 178) ((T_STRING) reduce 178) ((T_STRING_CAST) reduce 178) ((T_SWITCH) reduce 178) ((T_THROW) reduce 178) ((T_TRAIT) reduce 178) ((T_TRAIT_C) reduce 178) ((T_TRY) reduce 178) ((T_UNSET) reduce 178) ((T_UNSET_CAST) reduce 178) ((T_VARIABLE) reduce 178) ((T_VOID_CAST) reduce 178) ((T_WHILE) reduce 178) ((T_YIELD) reduce 178) ((T_YIELD_FROM) reduce 178) (("[") reduce 178) (("`") reduce 178) (("{") reduce 178) (("~") reduce 178)) 22870 ((("!") reduce 172) (("\"") reduce 172) (("$") reduce 172) (("(") reduce 172) (("+") reduce 172) (("-") reduce 172) ((";") reduce 172) (("@") reduce 172) ((T_ABSTRACT) reduce 172) ((T_ARRAY) reduce 172) ((T_ARRAY_CAST) reduce 172) ((T_ATTRIBUTE) reduce 172) ((T_BOOL_CAST) reduce 172) ((T_BREAK) reduce 172) ((T_CATCH) reduce 172) ((T_CLASS) reduce 172) ((T_CLASS_C) reduce 172) ((T_CLONE) reduce 172) ((T_CONSTANT_ENCAPSED_STRING) reduce 172) ((T_CONTINUE) reduce 172) ((T_DEC) reduce 172) ((T_DECLARE) reduce 172) ((T_DIR) reduce 172) ((T_DNUMBER) reduce 172) ((T_DO) reduce 172) ((T_DOUBLE_CAST) reduce 172) ((T_ECHO) reduce 172) ((T_ELSE) reduce 172) ((T_ELSEIF) reduce 172) ((T_EMPTY) reduce 172) ((T_ENDFOR) reduce 172) ((T_ENUM) reduce 172) ((T_EVAL) reduce 172) ((T_EXIT) reduce 172) ((T_FILE) reduce 172) ((T_FINAL) reduce 172) ((T_FINALLY) reduce 172) ((T_FN) reduce 172) ((T_FOR) reduce 172) ((T_FOREACH) reduce 172) ((T_FUNCTION) reduce 172) ((T_FUNC_C) reduce 172) ((T_GLOBAL) reduce 172) ((T_GOTO) reduce 172) ((T_HALT_COMPILER) reduce 172) ((T_IF) reduce 172) ((T_INC) reduce 172) ((T_INCLUDE) reduce 172) ((T_INCLUDE_ONCE) reduce 172) ((T_INLINE_HTML) reduce 172) ((T_INTERFACE) reduce 172) ((T_INT_CAST) reduce 172) ((T_ISSET) reduce 172) ((T_LINE) reduce 172) ((T_LIST) reduce 172) ((T_LNUMBER) reduce 172) ((T_MATCH) reduce 172) ((T_METHOD_C) reduce 172) ((T_NAME_FULLY_QUALIFIED) reduce 172) ((T_NAME_QUALIFIED) reduce 172) ((T_NAME_RELATIVE) reduce 172) ((T_NEW) reduce 172) ((T_NS_C) reduce 172) ((T_OBJECT_CAST) reduce 172) ((T_PRINT) reduce 172) ((T_PROPERTY_C) reduce 172) ((T_READONLY) reduce 172) ((T_REQUIRE) reduce 172) ((T_REQUIRE_ONCE) reduce 172) ((T_RETURN) reduce 172) ((T_START_HEREDOC) reduce 172) ((T_STATIC) reduce 172) ((T_STRING) reduce 172) ((T_STRING_CAST) reduce 172) ((T_SWITCH) reduce 172) ((T_THROW) reduce 172) ((T_TRAIT) reduce 172) ((T_TRAIT_C) reduce 172) ((T_TRY) reduce 172) ((T_UNSET) reduce 172) ((T_UNSET_CAST) reduce 172) ((T_VARIABLE) reduce 172) ((T_VOID_CAST) reduce 172) ((T_WHILE) reduce 172) ((T_YIELD) reduce 172) ((T_YIELD_FROM) reduce 172) (("[") reduce 172) (("`") reduce 172) (("{") reduce 172) (("~") reduce 172)) 22875 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_ELSE) reduce 153) ((T_ELSEIF) reduce 153) ((T_EMPTY) reduce 153) ((T_ENDFOR) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_VOID_CAST) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 22879 ((("!") reduce 224) (("\"") reduce 224) (("$") reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_ELSE) reduce 224) ((T_ELSEIF) reduce 224) ((T_EMPTY) reduce 224) ((T_ENDFOR) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_VARIABLE) reduce 224) ((T_VOID_CAST) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("~") reduce 224)) 22881 ((("!") reduce 225) (("\"") reduce 225) (("$") reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_ELSE) reduce 225) ((T_ELSEIF) reduce 225) ((T_EMPTY) reduce 225) ((T_ENDFOR) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_VARIABLE) reduce 225) ((T_VOID_CAST) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("~") reduce 225)) 22885 ((("!") reduce 226) (("\"") reduce 226) (("$") reduce 226) (("(") reduce 226) (("+") reduce 226) (("-") reduce 226) ((";") reduce 226) (("@") reduce 226) ((T_ABSTRACT) reduce 226) ((T_ARRAY) reduce 226) ((T_ARRAY_CAST) reduce 226) ((T_ATTRIBUTE) reduce 226) ((T_BOOL_CAST) reduce 226) ((T_BREAK) reduce 226) ((T_CLASS) reduce 226) ((T_CLASS_C) reduce 226) ((T_CLONE) reduce 226) ((T_CONSTANT_ENCAPSED_STRING) reduce 226) ((T_CONTINUE) reduce 226) ((T_DEC) reduce 226) ((T_DECLARE) reduce 226) ((T_DIR) reduce 226) ((T_DNUMBER) reduce 226) ((T_DO) reduce 226) ((T_DOUBLE_CAST) reduce 226) ((T_ECHO) reduce 226) ((T_ELSE) reduce 226) ((T_ELSEIF) reduce 226) ((T_EMPTY) reduce 226) ((T_ENDFOR) reduce 226) ((T_ENUM) reduce 226) ((T_EVAL) reduce 226) ((T_EXIT) reduce 226) ((T_FILE) reduce 226) ((T_FINAL) reduce 226) ((T_FN) reduce 226) ((T_FOR) reduce 226) ((T_FOREACH) reduce 226) ((T_FUNCTION) reduce 226) ((T_FUNC_C) reduce 226) ((T_GLOBAL) reduce 226) ((T_GOTO) reduce 226) ((T_HALT_COMPILER) reduce 226) ((T_IF) reduce 226) ((T_INC) reduce 226) ((T_INCLUDE) reduce 226) ((T_INCLUDE_ONCE) reduce 226) ((T_INLINE_HTML) reduce 226) ((T_INTERFACE) reduce 226) ((T_INT_CAST) reduce 226) ((T_ISSET) reduce 226) ((T_LINE) reduce 226) ((T_LIST) reduce 226) ((T_LNUMBER) reduce 226) ((T_MATCH) reduce 226) ((T_METHOD_C) reduce 226) ((T_NAME_FULLY_QUALIFIED) reduce 226) ((T_NAME_QUALIFIED) reduce 226) ((T_NAME_RELATIVE) reduce 226) ((T_NEW) reduce 226) ((T_NS_C) reduce 226) ((T_OBJECT_CAST) reduce 226) ((T_PRINT) reduce 226) ((T_PROPERTY_C) reduce 226) ((T_READONLY) reduce 226) ((T_REQUIRE) reduce 226) ((T_REQUIRE_ONCE) reduce 226) ((T_RETURN) reduce 226) ((T_START_HEREDOC) reduce 226) ((T_STATIC) reduce 226) ((T_STRING) reduce 226) ((T_STRING_CAST) reduce 226) ((T_SWITCH) reduce 226) ((T_THROW) reduce 226) ((T_TRAIT) reduce 226) ((T_TRAIT_C) reduce 226) ((T_TRY) reduce 226) ((T_UNSET) reduce 226) ((T_UNSET_CAST) reduce 226) ((T_VARIABLE) reduce 226) ((T_VOID_CAST) reduce 226) ((T_WHILE) reduce 226) ((T_YIELD) reduce 226) ((T_YIELD_FROM) reduce 226) (("[") reduce 226) (("`") reduce 226) (("{") reduce 226) (("~") reduce 226)) 22888 ((("!") reduce 227) (("\"") reduce 227) (("$") reduce 227) (("(") reduce 227) (("+") reduce 227) (("-") reduce 227) ((";") reduce 227) (("@") reduce 227) ((T_ABSTRACT) reduce 227) ((T_ARRAY) reduce 227) ((T_ARRAY_CAST) reduce 227) ((T_ATTRIBUTE) reduce 227) ((T_BOOL_CAST) reduce 227) ((T_BREAK) reduce 227) ((T_CLASS) reduce 227) ((T_CLASS_C) reduce 227) ((T_CLONE) reduce 227) ((T_CONSTANT_ENCAPSED_STRING) reduce 227) ((T_CONTINUE) reduce 227) ((T_DEC) reduce 227) ((T_DECLARE) reduce 227) ((T_DIR) reduce 227) ((T_DNUMBER) reduce 227) ((T_DO) reduce 227) ((T_DOUBLE_CAST) reduce 227) ((T_ECHO) reduce 227) ((T_ELSE) reduce 227) ((T_ELSEIF) reduce 227) ((T_EMPTY) reduce 227) ((T_ENDFOR) reduce 227) ((T_ENUM) reduce 227) ((T_EVAL) reduce 227) ((T_EXIT) reduce 227) ((T_FILE) reduce 227) ((T_FINAL) reduce 227) ((T_FN) reduce 227) ((T_FOR) reduce 227) ((T_FOREACH) reduce 227) ((T_FUNCTION) reduce 227) ((T_FUNC_C) reduce 227) ((T_GLOBAL) reduce 227) ((T_GOTO) reduce 227) ((T_HALT_COMPILER) reduce 227) ((T_IF) reduce 227) ((T_INC) reduce 227) ((T_INCLUDE) reduce 227) ((T_INCLUDE_ONCE) reduce 227) ((T_INLINE_HTML) reduce 227) ((T_INTERFACE) reduce 227) ((T_INT_CAST) reduce 227) ((T_ISSET) reduce 227) ((T_LINE) reduce 227) ((T_LIST) reduce 227) ((T_LNUMBER) reduce 227) ((T_MATCH) reduce 227) ((T_METHOD_C) reduce 227) ((T_NAME_FULLY_QUALIFIED) reduce 227) ((T_NAME_QUALIFIED) reduce 227) ((T_NAME_RELATIVE) reduce 227) ((T_NEW) reduce 227) ((T_NS_C) reduce 227) ((T_OBJECT_CAST) reduce 227) ((T_PRINT) reduce 227) ((T_PROPERTY_C) reduce 227) ((T_READONLY) reduce 227) ((T_REQUIRE) reduce 227) ((T_REQUIRE_ONCE) reduce 227) ((T_RETURN) reduce 227) ((T_START_HEREDOC) reduce 227) ((T_STATIC) reduce 227) ((T_STRING) reduce 227) ((T_STRING_CAST) reduce 227) ((T_SWITCH) reduce 227) ((T_THROW) reduce 227) ((T_TRAIT) reduce 227) ((T_TRAIT_C) reduce 227) ((T_TRY) reduce 227) ((T_UNSET) reduce 227) ((T_UNSET_CAST) reduce 227) ((T_VARIABLE) reduce 227) ((T_VOID_CAST) reduce 227) ((T_WHILE) reduce 227) ((T_YIELD) reduce 227) ((T_YIELD_FROM) reduce 227) (("[") reduce 227) (("`") reduce 227) (("{") reduce 227) (("~") reduce 227)) 22889 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_ELSE) reduce 169) ((T_ELSEIF) reduce 169) ((T_EMPTY) reduce 169) ((T_ENDFOR) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_VOID_CAST) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 22891 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_ELSE) reduce 158) ((T_ELSEIF) reduce 158) ((T_EMPTY) reduce 158) ((T_ENDFOR) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_VOID_CAST) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 22893 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_ELSE) reduce 156) ((T_ELSEIF) reduce 156) ((T_EMPTY) reduce 156) ((T_ENDFOR) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_VOID_CAST) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 22895 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_ELSE) reduce 168) ((T_ELSEIF) reduce 168) ((T_EMPTY) reduce 168) ((T_ENDFOR) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_VOID_CAST) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 22897 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_ELSE) reduce 157) ((T_ELSEIF) reduce 157) ((T_EMPTY) reduce 157) ((T_ENDFOR) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_VOID_CAST) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 22907 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_ELSE) reduce 164) ((T_ELSEIF) reduce 164) ((T_EMPTY) reduce 164) ((T_ENDFOR) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_VOID_CAST) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 22908 ((("!") reduce 220) (("\"") reduce 220) (("$") reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_ELSE) reduce 220) ((T_ELSEIF) reduce 220) ((T_EMPTY) reduce 220) ((T_ENDFOR) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_VARIABLE) reduce 220) ((T_VOID_CAST) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("~") reduce 220)) 22911 ((("!") reduce 221) (("\"") reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_ELSE) reduce 221) ((T_ELSEIF) reduce 221) ((T_EMPTY) reduce 221) ((T_ENDFOR) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_VARIABLE) reduce 221) ((T_VOID_CAST) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("~") reduce 221)) 22912 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_ELSE) reduce 163) ((T_ELSEIF) reduce 163) ((T_EMPTY) reduce 163) ((T_ENDFOR) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_VOID_CAST) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 22921 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_ELSE) reduce 152) ((T_ELSEIF) reduce 152) ((T_EMPTY) reduce 152) ((T_ENDFOR) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_VOID_CAST) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 22922 ((("!") reduce 218) (("\"") reduce 218) (("$") reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_ELSE) reduce 218) ((T_ELSEIF) reduce 218) ((T_EMPTY) reduce 218) ((T_ENDFOR) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_VARIABLE) reduce 218) ((T_VOID_CAST) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("~") reduce 218)) 22925 ((("!") reduce 219) (("\"") reduce 219) (("$") reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_ELSE) reduce 219) ((T_ELSEIF) reduce 219) ((T_EMPTY) reduce 219) ((T_ENDFOR) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_VARIABLE) reduce 219) ((T_VOID_CAST) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("~") reduce 219)) 22926 (((",") reduce 387) ((";") reduce 387)) 22928 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 388) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 388) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 22930 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_ELSE) reduce 159) ((T_ELSEIF) reduce 159) ((T_EMPTY) reduce 159) ((T_ENDFOR) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_VOID_CAST) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 22931 (((",") reduce 386) ((";") reduce 386)) 22932 (((T_WHILE) reduce 166)) 22943 (((T_WHILE) reduce 160)) 22952 (((T_WHILE) reduce 149)) 22955 (((T_WHILE) reduce 148)) 22956 (((T_ELSE) shift) ((T_ELSEIF) shift) ((T_WHILE) reduce 246)) 22957 (((T_WHILE) shift)) 22960 (((T_WHILE) reduce 147)) 22965 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_ELSE) reduce 151) ((T_ELSEIF) reduce 151) ((T_EMPTY) reduce 151) ((T_ENDFOR) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_VOID_CAST) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 22971 (((T_ELSE) reduce 166) ((T_ELSEIF) reduce 166) ((T_WHILE) reduce 166)) 22981 (((T_ELSE) reduce 160) ((T_ELSEIF) reduce 160) ((T_WHILE) reduce 160)) 22990 (((T_ELSE) reduce 149) ((T_ELSEIF) reduce 149) ((T_WHILE) reduce 149)) 22993 (((T_ELSE) reduce 148) ((T_ELSEIF) reduce 148) ((T_WHILE) reduce 148)) 22995 (((T_ELSE) reduce 245) ((T_ELSEIF) reduce 245) ((T_WHILE) reduce 245)) 22998 (((T_ELSE) reduce 147) ((T_ELSEIF) reduce 147) ((T_WHILE) reduce 147)) 23000 (((T_ELSE) reduce 247) ((T_ELSEIF) reduce 247) ((T_WHILE) reduce 247)) 23001 (((T_ELSE) reduce 161) ((T_ELSEIF) reduce 161) ((T_WHILE) reduce 161)) 23004 (((T_ELSE) reduce 250) ((T_ELSEIF) reduce 250) ((T_WHILE) reduce 250)) 23008 (((T_ELSE) reduce 251) ((T_ELSEIF) reduce 251) ((T_WHILE) reduce 251)) 23013 (((T_ELSE) reduce 242) ((T_ELSEIF) reduce 242) ((T_WHILE) reduce 242)) 23014 (((T_ELSE) reduce 150) ((T_ELSEIF) reduce 150) ((T_WHILE) reduce 150)) 23017 (((T_ELSE) reduce 243) ((T_ELSEIF) reduce 243) ((T_WHILE) reduce 243)) 23019 (((T_ELSE) reduce 170) ((T_ELSEIF) reduce 170) ((T_WHILE) reduce 170)) 23024 (((T_ELSE) reduce 162) ((T_ELSEIF) reduce 162) ((T_WHILE) reduce 162)) 23027 (((T_CATCH) reduce 171) ((T_ELSE) reduce 171) ((T_ELSEIF) reduce 171) ((T_FINALLY) reduce 171) ((T_WHILE) reduce 171)) 23028 (((T_CATCH) shift) ((T_ELSE) reduce 177) ((T_ELSEIF) reduce 177) ((T_FINALLY) shift) ((T_WHILE) reduce 177)) 23031 (((T_ELSE) reduce 167) ((T_ELSEIF) reduce 167) ((T_WHILE) reduce 167)) 23034 (((T_ELSE) reduce 178) ((T_ELSEIF) reduce 178) ((T_WHILE) reduce 178)) 23041 (((T_CATCH) reduce 172) ((T_ELSE) reduce 172) ((T_ELSEIF) reduce 172) ((T_FINALLY) reduce 172) ((T_WHILE) reduce 172)) 23046 (((T_ELSE) reduce 153) ((T_ELSEIF) reduce 153) ((T_WHILE) reduce 153)) 23050 (((T_ELSE) reduce 224) ((T_ELSEIF) reduce 224) ((T_WHILE) reduce 224)) 23052 (((T_ELSE) reduce 225) ((T_ELSEIF) reduce 225) ((T_WHILE) reduce 225)) 23056 (((T_ELSE) reduce 226) ((T_ELSEIF) reduce 226) ((T_WHILE) reduce 226)) 23059 (((T_ELSE) reduce 227) ((T_ELSEIF) reduce 227) ((T_WHILE) reduce 227)) 23060 (((T_ELSE) reduce 169) ((T_ELSEIF) reduce 169) ((T_WHILE) reduce 169)) 23062 (((T_ELSE) reduce 158) ((T_ELSEIF) reduce 158) ((T_WHILE) reduce 158)) 23064 (((T_ELSE) reduce 156) ((T_ELSEIF) reduce 156) ((T_WHILE) reduce 156)) 23066 (((T_ELSE) reduce 168) ((T_ELSEIF) reduce 168) ((T_WHILE) reduce 168)) 23068 (((T_ELSE) reduce 157) ((T_ELSEIF) reduce 157) ((T_WHILE) reduce 157)) 23078 (((T_ELSE) reduce 164) ((T_ELSEIF) reduce 164) ((T_WHILE) reduce 164)) 23079 (((T_ELSE) reduce 220) ((T_ELSEIF) reduce 220) ((T_WHILE) reduce 220)) 23082 (((T_ELSE) reduce 221) ((T_ELSEIF) reduce 221) ((T_WHILE) reduce 221)) 23083 (((T_ELSE) reduce 163) ((T_ELSEIF) reduce 163) ((T_WHILE) reduce 163)) 23092 (((T_ELSE) reduce 152) ((T_ELSEIF) reduce 152) ((T_WHILE) reduce 152)) 23093 (((T_ELSE) reduce 218) ((T_ELSEIF) reduce 218) ((T_WHILE) reduce 218)) 23096 (((T_ELSE) reduce 219) ((T_ELSEIF) reduce 219) ((T_WHILE) reduce 219)) 23098 (((T_ELSE) reduce 159) ((T_ELSEIF) reduce 159) ((T_WHILE) reduce 159)) 23104 (((T_ELSE) reduce 151) ((T_ELSEIF) reduce 151) ((T_WHILE) reduce 151)) 23107 (((")") reduce 140) ((",") reduce 140)) 23108 (((")") shift) ((",") shift)) 23111 (((")") reduce 139) ((",") reduce 139)) 23112 ((("!") reduce 142) (("\"") reduce 142) (("$") reduce 142) (("(") reduce 142) (("+") reduce 142) (("-") reduce 142) ((";") reduce 142) (("@") reduce 142) ((T_ABSTRACT) reduce 142) ((T_ARRAY) reduce 142) ((T_ARRAY_CAST) reduce 142) ((T_ATTRIBUTE) reduce 142) ((T_BOOL_CAST) reduce 142) ((T_BREAK) reduce 142) ((T_CLASS) reduce 142) ((T_CLASS_C) reduce 142) ((T_CLONE) reduce 142) ((T_CONSTANT_ENCAPSED_STRING) reduce 142) ((T_CONTINUE) reduce 142) ((T_DEC) reduce 142) ((T_DECLARE) reduce 142) ((T_DIR) reduce 142) ((T_DNUMBER) reduce 142) ((T_DO) reduce 142) ((T_DOUBLE_CAST) reduce 142) ((T_ECHO) reduce 142) ((T_EMPTY) reduce 142) ((T_ENDDECLARE) reduce 142) ((T_ENUM) reduce 142) ((T_EVAL) reduce 142) ((T_EXIT) reduce 142) ((T_FILE) reduce 142) ((T_FINAL) reduce 142) ((T_FN) reduce 142) ((T_FOR) reduce 142) ((T_FOREACH) reduce 142) ((T_FUNCTION) reduce 142) ((T_FUNC_C) reduce 142) ((T_GLOBAL) reduce 142) ((T_GOTO) reduce 142) ((T_HALT_COMPILER) reduce 142) ((T_IF) reduce 142) ((T_INC) reduce 142) ((T_INCLUDE) reduce 142) ((T_INCLUDE_ONCE) reduce 142) ((T_INLINE_HTML) reduce 142) ((T_INTERFACE) reduce 142) ((T_INT_CAST) reduce 142) ((T_ISSET) reduce 142) ((T_LINE) reduce 142) ((T_LIST) reduce 142) ((T_LNUMBER) reduce 142) ((T_MATCH) reduce 142) ((T_METHOD_C) reduce 142) ((T_NAME_FULLY_QUALIFIED) reduce 142) ((T_NAME_QUALIFIED) reduce 142) ((T_NAME_RELATIVE) reduce 142) ((T_NEW) reduce 142) ((T_NS_C) reduce 142) ((T_OBJECT_CAST) reduce 142) ((T_PRINT) reduce 142) ((T_PROPERTY_C) reduce 142) ((T_READONLY) reduce 142) ((T_REQUIRE) reduce 142) ((T_REQUIRE_ONCE) reduce 142) ((T_RETURN) reduce 142) ((T_START_HEREDOC) reduce 142) ((T_STATIC) reduce 142) ((T_STRING) reduce 142) ((T_STRING_CAST) reduce 142) ((T_SWITCH) reduce 142) ((T_THROW) reduce 142) ((T_TRAIT) reduce 142) ((T_TRAIT_C) reduce 142) ((T_TRY) reduce 142) ((T_UNSET) reduce 142) ((T_UNSET_CAST) reduce 142) ((T_VARIABLE) reduce 142) ((T_VOID_CAST) reduce 142) ((T_WHILE) reduce 142) ((T_YIELD) reduce 142) ((T_YIELD_FROM) reduce 142) (("[") reduce 142) (("`") reduce 142) (("{") reduce 142) (("~") reduce 142)) 23113 (((T_ELSE) reduce 165) ((T_ELSEIF) reduce 165) ((T_WHILE) reduce 165)) 23114 (((T_ELSE) reduce 222) ((T_ELSEIF) reduce 222) ((T_WHILE) reduce 222)) 23115 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_ENDDECLARE) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_VOID_CAST) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("~") shift)) 23116 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_EMPTY) reduce 166) ((T_ENDDECLARE) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_VOID_CAST) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 23131 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_EMPTY) reduce 160) ((T_ENDDECLARE) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_VOID_CAST) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 23142 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_EMPTY) reduce 149) ((T_ENDDECLARE) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_VOID_CAST) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 23144 ((("!") reduce 144) (("\"") reduce 144) (("$") reduce 144) (("(") reduce 144) (("+") reduce 144) (("-") reduce 144) ((";") reduce 144) (("@") reduce 144) ((T_ABSTRACT) reduce 144) ((T_ARRAY) reduce 144) ((T_ARRAY_CAST) reduce 144) ((T_ATTRIBUTE) reduce 144) ((T_BOOL_CAST) reduce 144) ((T_BREAK) reduce 144) ((T_CLASS) reduce 144) ((T_CLASS_C) reduce 144) ((T_CLONE) reduce 144) ((T_CONSTANT_ENCAPSED_STRING) reduce 144) ((T_CONTINUE) reduce 144) ((T_DEC) reduce 144) ((T_DECLARE) reduce 144) ((T_DIR) reduce 144) ((T_DNUMBER) reduce 144) ((T_DO) reduce 144) ((T_DOUBLE_CAST) reduce 144) ((T_ECHO) reduce 144) ((T_EMPTY) reduce 144) ((T_ENDDECLARE) reduce 144) ((T_ENUM) reduce 144) ((T_EVAL) reduce 144) ((T_EXIT) reduce 144) ((T_FILE) reduce 144) ((T_FINAL) reduce 144) ((T_FN) reduce 144) ((T_FOR) reduce 144) ((T_FOREACH) reduce 144) ((T_FUNCTION) reduce 144) ((T_FUNC_C) reduce 144) ((T_GLOBAL) reduce 144) ((T_GOTO) reduce 144) ((T_HALT_COMPILER) reduce 144) ((T_IF) reduce 144) ((T_INC) reduce 144) ((T_INCLUDE) reduce 144) ((T_INCLUDE_ONCE) reduce 144) ((T_INLINE_HTML) reduce 144) ((T_INTERFACE) reduce 144) ((T_INT_CAST) reduce 144) ((T_ISSET) reduce 144) ((T_LINE) reduce 144) ((T_LIST) reduce 144) ((T_LNUMBER) reduce 144) ((T_MATCH) reduce 144) ((T_METHOD_C) reduce 144) ((T_NAME_FULLY_QUALIFIED) reduce 144) ((T_NAME_QUALIFIED) reduce 144) ((T_NAME_RELATIVE) reduce 144) ((T_NEW) reduce 144) ((T_NS_C) reduce 144) ((T_OBJECT_CAST) reduce 144) ((T_PRINT) reduce 144) ((T_PROPERTY_C) reduce 144) ((T_READONLY) reduce 144) ((T_REQUIRE) reduce 144) ((T_REQUIRE_ONCE) reduce 144) ((T_RETURN) reduce 144) ((T_START_HEREDOC) reduce 144) ((T_STATIC) reduce 144) ((T_STRING) reduce 144) ((T_STRING_CAST) reduce 144) ((T_SWITCH) reduce 144) ((T_THROW) reduce 144) ((T_TRAIT) reduce 144) ((T_TRAIT_C) reduce 144) ((T_TRY) reduce 144) ((T_UNSET) reduce 144) ((T_UNSET_CAST) reduce 144) ((T_VARIABLE) reduce 144) ((T_VOID_CAST) reduce 144) ((T_WHILE) reduce 144) ((T_YIELD) reduce 144) ((T_YIELD_FROM) reduce 144) (("[") reduce 144) (("`") reduce 144) (("{") reduce 144) (("~") reduce 144)) 23146 ((("!") reduce 104) (("\"") reduce 104) (("$") reduce 104) (("(") reduce 104) (("+") reduce 104) (("-") reduce 104) ((";") reduce 104) (("@") reduce 104) ((T_ABSTRACT) reduce 104) ((T_ARRAY) reduce 104) ((T_ARRAY_CAST) reduce 104) ((T_ATTRIBUTE) reduce 104) ((T_BOOL_CAST) reduce 104) ((T_BREAK) reduce 104) ((T_CLASS) reduce 104) ((T_CLASS_C) reduce 104) ((T_CLONE) reduce 104) ((T_CONSTANT_ENCAPSED_STRING) reduce 104) ((T_CONTINUE) reduce 104) ((T_DEC) reduce 104) ((T_DECLARE) reduce 104) ((T_DIR) reduce 104) ((T_DNUMBER) reduce 104) ((T_DO) reduce 104) ((T_DOUBLE_CAST) reduce 104) ((T_ECHO) reduce 104) ((T_EMPTY) reduce 104) ((T_ENDDECLARE) reduce 104) ((T_ENUM) reduce 104) ((T_EVAL) reduce 104) ((T_EXIT) reduce 104) ((T_FILE) reduce 104) ((T_FINAL) reduce 104) ((T_FN) reduce 104) ((T_FOR) reduce 104) ((T_FOREACH) reduce 104) ((T_FUNCTION) reduce 104) ((T_FUNC_C) reduce 104) ((T_GLOBAL) reduce 104) ((T_GOTO) reduce 104) ((T_HALT_COMPILER) reduce 104) ((T_IF) reduce 104) ((T_INC) reduce 104) ((T_INCLUDE) reduce 104) ((T_INCLUDE_ONCE) reduce 104) ((T_INLINE_HTML) reduce 104) ((T_INTERFACE) reduce 104) ((T_INT_CAST) reduce 104) ((T_ISSET) reduce 104) ((T_LINE) reduce 104) ((T_LIST) reduce 104) ((T_LNUMBER) reduce 104) ((T_MATCH) reduce 104) ((T_METHOD_C) reduce 104) ((T_NAME_FULLY_QUALIFIED) reduce 104) ((T_NAME_QUALIFIED) reduce 104) ((T_NAME_RELATIVE) reduce 104) ((T_NEW) reduce 104) ((T_NS_C) reduce 104) ((T_OBJECT_CAST) reduce 104) ((T_PRINT) reduce 104) ((T_PROPERTY_C) reduce 104) ((T_READONLY) reduce 104) ((T_REQUIRE) reduce 104) ((T_REQUIRE_ONCE) reduce 104) ((T_RETURN) reduce 104) ((T_START_HEREDOC) reduce 104) ((T_STATIC) reduce 104) ((T_STRING) reduce 104) ((T_STRING_CAST) reduce 104) ((T_SWITCH) reduce 104) ((T_THROW) reduce 104) ((T_TRAIT) reduce 104) ((T_TRAIT_C) reduce 104) ((T_TRY) reduce 104) ((T_UNSET) reduce 104) ((T_UNSET_CAST) reduce 104) ((T_VARIABLE) reduce 104) ((T_VOID_CAST) reduce 104) ((T_WHILE) reduce 104) ((T_YIELD) reduce 104) ((T_YIELD_FROM) reduce 104) (("[") reduce 104) (("`") reduce 104) (("{") reduce 104) (("~") reduce 104)) 23148 ((("!") reduce 107) (("\"") reduce 107) (("$") reduce 107) (("(") reduce 107) (("+") reduce 107) (("-") reduce 107) ((";") reduce 107) (("@") reduce 107) ((T_ABSTRACT) reduce 107) ((T_ARRAY) reduce 107) ((T_ARRAY_CAST) reduce 107) ((T_ATTRIBUTE) reduce 107) ((T_BOOL_CAST) reduce 107) ((T_BREAK) reduce 107) ((T_CLASS) reduce 107) ((T_CLASS_C) reduce 107) ((T_CLONE) reduce 107) ((T_CONSTANT_ENCAPSED_STRING) reduce 107) ((T_CONTINUE) reduce 107) ((T_DEC) reduce 107) ((T_DECLARE) reduce 107) ((T_DIR) reduce 107) ((T_DNUMBER) reduce 107) ((T_DO) reduce 107) ((T_DOUBLE_CAST) reduce 107) ((T_ECHO) reduce 107) ((T_EMPTY) reduce 107) ((T_ENDDECLARE) reduce 107) ((T_ENUM) reduce 107) ((T_EVAL) reduce 107) ((T_EXIT) reduce 107) ((T_FILE) reduce 107) ((T_FINAL) reduce 107) ((T_FN) reduce 107) ((T_FOR) reduce 107) ((T_FOREACH) reduce 107) ((T_FUNCTION) reduce 107) ((T_FUNC_C) reduce 107) ((T_GLOBAL) reduce 107) ((T_GOTO) reduce 107) ((T_HALT_COMPILER) reduce 107) ((T_IF) reduce 107) ((T_INC) reduce 107) ((T_INCLUDE) reduce 107) ((T_INCLUDE_ONCE) reduce 107) ((T_INLINE_HTML) reduce 107) ((T_INTERFACE) reduce 107) ((T_INT_CAST) reduce 107) ((T_ISSET) reduce 107) ((T_LINE) reduce 107) ((T_LIST) reduce 107) ((T_LNUMBER) reduce 107) ((T_MATCH) reduce 107) ((T_METHOD_C) reduce 107) ((T_NAME_FULLY_QUALIFIED) reduce 107) ((T_NAME_QUALIFIED) reduce 107) ((T_NAME_RELATIVE) reduce 107) ((T_NEW) reduce 107) ((T_NS_C) reduce 107) ((T_OBJECT_CAST) reduce 107) ((T_PRINT) reduce 107) ((T_PROPERTY_C) reduce 107) ((T_READONLY) reduce 107) ((T_REQUIRE) reduce 107) ((T_REQUIRE_ONCE) reduce 107) ((T_RETURN) reduce 107) ((T_START_HEREDOC) reduce 107) ((T_STATIC) reduce 107) ((T_STRING) reduce 107) ((T_STRING_CAST) reduce 107) ((T_SWITCH) reduce 107) ((T_THROW) reduce 107) ((T_TRAIT) reduce 107) ((T_TRAIT_C) reduce 107) ((T_TRY) reduce 107) ((T_UNSET) reduce 107) ((T_UNSET_CAST) reduce 107) ((T_VARIABLE) reduce 107) ((T_VOID_CAST) reduce 107) ((T_WHILE) reduce 107) ((T_YIELD) reduce 107) ((T_YIELD_FROM) reduce 107) (("[") reduce 107) (("`") reduce 107) (("{") reduce 107) (("~") reduce 107)) 23151 ((("!") reduce 103) (("\"") reduce 103) (("$") reduce 103) (("(") reduce 103) (("+") reduce 103) (("-") reduce 103) ((";") reduce 103) (("@") reduce 103) ((T_ABSTRACT) reduce 103) ((T_ARRAY) reduce 103) ((T_ARRAY_CAST) reduce 103) ((T_ATTRIBUTE) reduce 103) ((T_BOOL_CAST) reduce 103) ((T_BREAK) reduce 103) ((T_CLASS) reduce 103) ((T_CLASS_C) reduce 103) ((T_CLONE) reduce 103) ((T_CONSTANT_ENCAPSED_STRING) reduce 103) ((T_CONTINUE) reduce 103) ((T_DEC) reduce 103) ((T_DECLARE) reduce 103) ((T_DIR) reduce 103) ((T_DNUMBER) reduce 103) ((T_DO) reduce 103) ((T_DOUBLE_CAST) reduce 103) ((T_ECHO) reduce 103) ((T_EMPTY) reduce 103) ((T_ENDDECLARE) reduce 103) ((T_ENUM) reduce 103) ((T_EVAL) reduce 103) ((T_EXIT) reduce 103) ((T_FILE) reduce 103) ((T_FINAL) reduce 103) ((T_FN) reduce 103) ((T_FOR) reduce 103) ((T_FOREACH) reduce 103) ((T_FUNCTION) reduce 103) ((T_FUNC_C) reduce 103) ((T_GLOBAL) reduce 103) ((T_GOTO) reduce 103) ((T_HALT_COMPILER) reduce 103) ((T_IF) reduce 103) ((T_INC) reduce 103) ((T_INCLUDE) reduce 103) ((T_INCLUDE_ONCE) reduce 103) ((T_INLINE_HTML) reduce 103) ((T_INTERFACE) reduce 103) ((T_INT_CAST) reduce 103) ((T_ISSET) reduce 103) ((T_LINE) reduce 103) ((T_LIST) reduce 103) ((T_LNUMBER) reduce 103) ((T_MATCH) reduce 103) ((T_METHOD_C) reduce 103) ((T_NAME_FULLY_QUALIFIED) reduce 103) ((T_NAME_QUALIFIED) reduce 103) ((T_NAME_RELATIVE) reduce 103) ((T_NEW) reduce 103) ((T_NS_C) reduce 103) ((T_OBJECT_CAST) reduce 103) ((T_PRINT) reduce 103) ((T_PROPERTY_C) reduce 103) ((T_READONLY) reduce 103) ((T_REQUIRE) reduce 103) ((T_REQUIRE_ONCE) reduce 103) ((T_RETURN) reduce 103) ((T_START_HEREDOC) reduce 103) ((T_STATIC) reduce 103) ((T_STRING) reduce 103) ((T_STRING_CAST) reduce 103) ((T_SWITCH) reduce 103) ((T_THROW) reduce 103) ((T_TRAIT) reduce 103) ((T_TRAIT_C) reduce 103) ((T_TRY) reduce 103) ((T_UNSET) reduce 103) ((T_UNSET_CAST) reduce 103) ((T_VARIABLE) reduce 103) ((T_VOID_CAST) reduce 103) ((T_WHILE) reduce 103) ((T_YIELD) reduce 103) ((T_YIELD_FROM) reduce 103) (("[") reduce 103) (("`") reduce 103) (("{") reduce 103) (("~") reduce 103)) 23152 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_EMPTY) reduce 148) ((T_ENDDECLARE) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_VOID_CAST) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 23153 ((("!") reduce 246) (("\"") reduce 246) (("$") reduce 246) (("(") reduce 246) (("+") reduce 246) (("-") reduce 246) ((";") reduce 246) (("@") reduce 246) ((T_ABSTRACT) reduce 246) ((T_ARRAY) reduce 246) ((T_ARRAY_CAST) reduce 246) ((T_ATTRIBUTE) reduce 246) ((T_BOOL_CAST) reduce 246) ((T_BREAK) reduce 246) ((T_CLASS) reduce 246) ((T_CLASS_C) reduce 246) ((T_CLONE) reduce 246) ((T_CONSTANT_ENCAPSED_STRING) reduce 246) ((T_CONTINUE) reduce 246) ((T_DEC) reduce 246) ((T_DECLARE) reduce 246) ((T_DIR) reduce 246) ((T_DNUMBER) reduce 246) ((T_DO) reduce 246) ((T_DOUBLE_CAST) reduce 246) ((T_ECHO) reduce 246) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) reduce 246) ((T_ENDDECLARE) reduce 246) ((T_ENUM) reduce 246) ((T_EVAL) reduce 246) ((T_EXIT) reduce 246) ((T_FILE) reduce 246) ((T_FINAL) reduce 246) ((T_FN) reduce 246) ((T_FOR) reduce 246) ((T_FOREACH) reduce 246) ((T_FUNCTION) reduce 246) ((T_FUNC_C) reduce 246) ((T_GLOBAL) reduce 246) ((T_GOTO) reduce 246) ((T_HALT_COMPILER) reduce 246) ((T_IF) reduce 246) ((T_INC) reduce 246) ((T_INCLUDE) reduce 246) ((T_INCLUDE_ONCE) reduce 246) ((T_INLINE_HTML) reduce 246) ((T_INTERFACE) reduce 246) ((T_INT_CAST) reduce 246) ((T_ISSET) reduce 246) ((T_LINE) reduce 246) ((T_LIST) reduce 246) ((T_LNUMBER) reduce 246) ((T_MATCH) reduce 246) ((T_METHOD_C) reduce 246) ((T_NAME_FULLY_QUALIFIED) reduce 246) ((T_NAME_QUALIFIED) reduce 246) ((T_NAME_RELATIVE) reduce 246) ((T_NEW) reduce 246) ((T_NS_C) reduce 246) ((T_OBJECT_CAST) reduce 246) ((T_PRINT) reduce 246) ((T_PROPERTY_C) reduce 246) ((T_READONLY) reduce 246) ((T_REQUIRE) reduce 246) ((T_REQUIRE_ONCE) reduce 246) ((T_RETURN) reduce 246) ((T_START_HEREDOC) reduce 246) ((T_STATIC) reduce 246) ((T_STRING) reduce 246) ((T_STRING_CAST) reduce 246) ((T_SWITCH) reduce 246) ((T_THROW) reduce 246) ((T_TRAIT) reduce 246) ((T_TRAIT_C) reduce 246) ((T_TRY) reduce 246) ((T_UNSET) reduce 246) ((T_UNSET_CAST) reduce 246) ((T_VARIABLE) reduce 246) ((T_VOID_CAST) reduce 246) ((T_WHILE) reduce 246) ((T_YIELD) reduce 246) ((T_YIELD_FROM) reduce 246) (("[") reduce 246) (("`") reduce 246) (("{") reduce 246) (("~") reduce 246)) 23154 ((("!") reduce 141) (("\"") reduce 141) (("$") reduce 141) (("(") reduce 141) (("+") reduce 141) (("-") reduce 141) ((";") reduce 141) (("@") reduce 141) ((T_ABSTRACT) reduce 141) ((T_ARRAY) reduce 141) ((T_ARRAY_CAST) reduce 141) ((T_ATTRIBUTE) reduce 141) ((T_BOOL_CAST) reduce 141) ((T_BREAK) reduce 141) ((T_CLASS) reduce 141) ((T_CLASS_C) reduce 141) ((T_CLONE) reduce 141) ((T_CONSTANT_ENCAPSED_STRING) reduce 141) ((T_CONTINUE) reduce 141) ((T_DEC) reduce 141) ((T_DECLARE) reduce 141) ((T_DIR) reduce 141) ((T_DNUMBER) reduce 141) ((T_DO) reduce 141) ((T_DOUBLE_CAST) reduce 141) ((T_ECHO) reduce 141) ((T_EMPTY) reduce 141) ((T_ENDDECLARE) reduce 141) ((T_ENUM) reduce 141) ((T_EVAL) reduce 141) ((T_EXIT) reduce 141) ((T_FILE) reduce 141) ((T_FINAL) reduce 141) ((T_FN) reduce 141) ((T_FOR) reduce 141) ((T_FOREACH) reduce 141) ((T_FUNCTION) reduce 141) ((T_FUNC_C) reduce 141) ((T_GLOBAL) reduce 141) ((T_GOTO) reduce 141) ((T_HALT_COMPILER) reduce 141) ((T_IF) reduce 141) ((T_INC) reduce 141) ((T_INCLUDE) reduce 141) ((T_INCLUDE_ONCE) reduce 141) ((T_INLINE_HTML) reduce 141) ((T_INTERFACE) reduce 141) ((T_INT_CAST) reduce 141) ((T_ISSET) reduce 141) ((T_LINE) reduce 141) ((T_LIST) reduce 141) ((T_LNUMBER) reduce 141) ((T_MATCH) reduce 141) ((T_METHOD_C) reduce 141) ((T_NAME_FULLY_QUALIFIED) reduce 141) ((T_NAME_QUALIFIED) reduce 141) ((T_NAME_RELATIVE) reduce 141) ((T_NEW) reduce 141) ((T_NS_C) reduce 141) ((T_OBJECT_CAST) reduce 141) ((T_PRINT) reduce 141) ((T_PROPERTY_C) reduce 141) ((T_READONLY) reduce 141) ((T_REQUIRE) reduce 141) ((T_REQUIRE_ONCE) reduce 141) ((T_RETURN) reduce 141) ((T_START_HEREDOC) reduce 141) ((T_STATIC) reduce 141) ((T_STRING) reduce 141) ((T_STRING_CAST) reduce 141) ((T_SWITCH) reduce 141) ((T_THROW) reduce 141) ((T_TRAIT) reduce 141) ((T_TRAIT_C) reduce 141) ((T_TRY) reduce 141) ((T_UNSET) reduce 141) ((T_UNSET_CAST) reduce 141) ((T_VARIABLE) reduce 141) ((T_VOID_CAST) reduce 141) ((T_WHILE) reduce 141) ((T_YIELD) reduce 141) ((T_YIELD_FROM) reduce 141) (("[") reduce 141) (("`") reduce 141) (("{") reduce 141) (("~") reduce 141)) 23155 ((("!") reduce 106) (("\"") reduce 106) (("$") reduce 106) (("(") reduce 106) (("+") reduce 106) (("-") reduce 106) ((";") reduce 106) (("@") reduce 106) ((T_ABSTRACT) reduce 106) ((T_ARRAY) reduce 106) ((T_ARRAY_CAST) reduce 106) ((T_ATTRIBUTE) reduce 106) ((T_BOOL_CAST) reduce 106) ((T_BREAK) reduce 106) ((T_CLASS) reduce 106) ((T_CLASS_C) reduce 106) ((T_CLONE) reduce 106) ((T_CONSTANT_ENCAPSED_STRING) reduce 106) ((T_CONTINUE) reduce 106) ((T_DEC) reduce 106) ((T_DECLARE) reduce 106) ((T_DIR) reduce 106) ((T_DNUMBER) reduce 106) ((T_DO) reduce 106) ((T_DOUBLE_CAST) reduce 106) ((T_ECHO) reduce 106) ((T_EMPTY) reduce 106) ((T_ENDDECLARE) reduce 106) ((T_ENUM) reduce 106) ((T_EVAL) reduce 106) ((T_EXIT) reduce 106) ((T_FILE) reduce 106) ((T_FINAL) reduce 106) ((T_FN) reduce 106) ((T_FOR) reduce 106) ((T_FOREACH) reduce 106) ((T_FUNCTION) reduce 106) ((T_FUNC_C) reduce 106) ((T_GLOBAL) reduce 106) ((T_GOTO) reduce 106) ((T_HALT_COMPILER) reduce 106) ((T_IF) reduce 106) ((T_INC) reduce 106) ((T_INCLUDE) reduce 106) ((T_INCLUDE_ONCE) reduce 106) ((T_INLINE_HTML) reduce 106) ((T_INTERFACE) reduce 106) ((T_INT_CAST) reduce 106) ((T_ISSET) reduce 106) ((T_LINE) reduce 106) ((T_LIST) reduce 106) ((T_LNUMBER) reduce 106) ((T_MATCH) reduce 106) ((T_METHOD_C) reduce 106) ((T_NAME_FULLY_QUALIFIED) reduce 106) ((T_NAME_QUALIFIED) reduce 106) ((T_NAME_RELATIVE) reduce 106) ((T_NEW) reduce 106) ((T_NS_C) reduce 106) ((T_OBJECT_CAST) reduce 106) ((T_PRINT) reduce 106) ((T_PROPERTY_C) reduce 106) ((T_READONLY) reduce 106) ((T_REQUIRE) reduce 106) ((T_REQUIRE_ONCE) reduce 106) ((T_RETURN) reduce 106) ((T_START_HEREDOC) reduce 106) ((T_STATIC) reduce 106) ((T_STRING) reduce 106) ((T_STRING_CAST) reduce 106) ((T_SWITCH) reduce 106) ((T_THROW) reduce 106) ((T_TRAIT) reduce 106) ((T_TRAIT_C) reduce 106) ((T_TRY) reduce 106) ((T_UNSET) reduce 106) ((T_UNSET_CAST) reduce 106) ((T_VARIABLE) reduce 106) ((T_VOID_CAST) reduce 106) ((T_WHILE) reduce 106) ((T_YIELD) reduce 106) ((T_YIELD_FROM) reduce 106) (("[") reduce 106) (("`") reduce 106) (("{") reduce 106) (("~") reduce 106)) 23156 ((("!") reduce 143) (("\"") reduce 143) (("$") reduce 143) (("(") reduce 143) (("+") reduce 143) (("-") reduce 143) ((";") reduce 143) (("@") reduce 143) ((T_ABSTRACT) reduce 143) ((T_ARRAY) reduce 143) ((T_ARRAY_CAST) reduce 143) ((T_ATTRIBUTE) reduce 143) ((T_BOOL_CAST) reduce 143) ((T_BREAK) reduce 143) ((T_CLASS) reduce 143) ((T_CLASS_C) reduce 143) ((T_CLONE) reduce 143) ((T_CONSTANT_ENCAPSED_STRING) reduce 143) ((T_CONTINUE) reduce 143) ((T_DEC) reduce 143) ((T_DECLARE) reduce 143) ((T_DIR) reduce 143) ((T_DNUMBER) reduce 143) ((T_DO) reduce 143) ((T_DOUBLE_CAST) reduce 143) ((T_ECHO) reduce 143) ((T_EMPTY) reduce 143) ((T_ENDDECLARE) reduce 143) ((T_ENUM) reduce 143) ((T_EVAL) reduce 143) ((T_EXIT) reduce 143) ((T_FILE) reduce 143) ((T_FINAL) reduce 143) ((T_FN) reduce 143) ((T_FOR) reduce 143) ((T_FOREACH) reduce 143) ((T_FUNCTION) reduce 143) ((T_FUNC_C) reduce 143) ((T_GLOBAL) reduce 143) ((T_GOTO) reduce 143) ((T_HALT_COMPILER) reduce 143) ((T_IF) reduce 143) ((T_INC) reduce 143) ((T_INCLUDE) reduce 143) ((T_INCLUDE_ONCE) reduce 143) ((T_INLINE_HTML) reduce 143) ((T_INTERFACE) reduce 143) ((T_INT_CAST) reduce 143) ((T_ISSET) reduce 143) ((T_LINE) reduce 143) ((T_LIST) reduce 143) ((T_LNUMBER) reduce 143) ((T_MATCH) reduce 143) ((T_METHOD_C) reduce 143) ((T_NAME_FULLY_QUALIFIED) reduce 143) ((T_NAME_QUALIFIED) reduce 143) ((T_NAME_RELATIVE) reduce 143) ((T_NEW) reduce 143) ((T_NS_C) reduce 143) ((T_OBJECT_CAST) reduce 143) ((T_PRINT) reduce 143) ((T_PROPERTY_C) reduce 143) ((T_READONLY) reduce 143) ((T_REQUIRE) reduce 143) ((T_REQUIRE_ONCE) reduce 143) ((T_RETURN) reduce 143) ((T_START_HEREDOC) reduce 143) ((T_STATIC) reduce 143) ((T_STRING) reduce 143) ((T_STRING_CAST) reduce 143) ((T_SWITCH) reduce 143) ((T_THROW) reduce 143) ((T_TRAIT) reduce 143) ((T_TRAIT_C) reduce 143) ((T_TRY) reduce 143) ((T_UNSET) reduce 143) ((T_UNSET_CAST) reduce 143) ((T_VARIABLE) reduce 143) ((T_VOID_CAST) reduce 143) ((T_WHILE) reduce 143) ((T_YIELD) reduce 143) ((T_YIELD_FROM) reduce 143) (("[") reduce 143) (("`") reduce 143) (("{") reduce 143) (("~") reduce 143)) 23157 ((("!") reduce 105) (("\"") reduce 105) (("$") reduce 105) (("(") reduce 105) (("+") reduce 105) (("-") reduce 105) ((";") reduce 105) (("@") reduce 105) ((T_ABSTRACT) reduce 105) ((T_ARRAY) reduce 105) ((T_ARRAY_CAST) reduce 105) ((T_ATTRIBUTE) reduce 105) ((T_BOOL_CAST) reduce 105) ((T_BREAK) reduce 105) ((T_CLASS) reduce 105) ((T_CLASS_C) reduce 105) ((T_CLONE) reduce 105) ((T_CONSTANT_ENCAPSED_STRING) reduce 105) ((T_CONTINUE) reduce 105) ((T_DEC) reduce 105) ((T_DECLARE) reduce 105) ((T_DIR) reduce 105) ((T_DNUMBER) reduce 105) ((T_DO) reduce 105) ((T_DOUBLE_CAST) reduce 105) ((T_ECHO) reduce 105) ((T_EMPTY) reduce 105) ((T_ENDDECLARE) reduce 105) ((T_ENUM) reduce 105) ((T_EVAL) reduce 105) ((T_EXIT) reduce 105) ((T_FILE) reduce 105) ((T_FINAL) reduce 105) ((T_FN) reduce 105) ((T_FOR) reduce 105) ((T_FOREACH) reduce 105) ((T_FUNCTION) reduce 105) ((T_FUNC_C) reduce 105) ((T_GLOBAL) reduce 105) ((T_GOTO) reduce 105) ((T_HALT_COMPILER) reduce 105) ((T_IF) reduce 105) ((T_INC) reduce 105) ((T_INCLUDE) reduce 105) ((T_INCLUDE_ONCE) reduce 105) ((T_INLINE_HTML) reduce 105) ((T_INTERFACE) reduce 105) ((T_INT_CAST) reduce 105) ((T_ISSET) reduce 105) ((T_LINE) reduce 105) ((T_LIST) reduce 105) ((T_LNUMBER) reduce 105) ((T_MATCH) reduce 105) ((T_METHOD_C) reduce 105) ((T_NAME_FULLY_QUALIFIED) reduce 105) ((T_NAME_QUALIFIED) reduce 105) ((T_NAME_RELATIVE) reduce 105) ((T_NEW) reduce 105) ((T_NS_C) reduce 105) ((T_OBJECT_CAST) reduce 105) ((T_PRINT) reduce 105) ((T_PROPERTY_C) reduce 105) ((T_READONLY) reduce 105) ((T_REQUIRE) reduce 105) ((T_REQUIRE_ONCE) reduce 105) ((T_RETURN) reduce 105) ((T_START_HEREDOC) reduce 105) ((T_STATIC) reduce 105) ((T_STRING) reduce 105) ((T_STRING_CAST) reduce 105) ((T_SWITCH) reduce 105) ((T_THROW) reduce 105) ((T_TRAIT) reduce 105) ((T_TRAIT_C) reduce 105) ((T_TRY) reduce 105) ((T_UNSET) reduce 105) ((T_UNSET_CAST) reduce 105) ((T_VARIABLE) reduce 105) ((T_VOID_CAST) reduce 105) ((T_WHILE) reduce 105) ((T_YIELD) reduce 105) ((T_YIELD_FROM) reduce 105) (("[") reduce 105) (("`") reduce 105) (("{") reduce 105) (("~") reduce 105)) 23160 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_EMPTY) reduce 147) ((T_ENDDECLARE) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_VOID_CAST) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 23166 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_ELSE) reduce 166) ((T_ELSEIF) reduce 166) ((T_EMPTY) reduce 166) ((T_ENDDECLARE) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_VARIABLE) reduce 166) ((T_VOID_CAST) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 23176 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_ELSE) reduce 160) ((T_ELSEIF) reduce 160) ((T_EMPTY) reduce 160) ((T_ENDDECLARE) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_VARIABLE) reduce 160) ((T_VOID_CAST) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 23185 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_ELSE) reduce 149) ((T_ELSEIF) reduce 149) ((T_EMPTY) reduce 149) ((T_ENDDECLARE) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_VARIABLE) reduce 149) ((T_VOID_CAST) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 23188 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_ELSE) reduce 148) ((T_ELSEIF) reduce 148) ((T_EMPTY) reduce 148) ((T_ENDDECLARE) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_VARIABLE) reduce 148) ((T_VOID_CAST) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 23190 ((("!") reduce 245) (("\"") reduce 245) (("$") reduce 245) (("(") reduce 245) (("+") reduce 245) (("-") reduce 245) ((";") reduce 245) (("@") reduce 245) ((T_ABSTRACT) reduce 245) ((T_ARRAY) reduce 245) ((T_ARRAY_CAST) reduce 245) ((T_ATTRIBUTE) reduce 245) ((T_BOOL_CAST) reduce 245) ((T_BREAK) reduce 245) ((T_CLASS) reduce 245) ((T_CLASS_C) reduce 245) ((T_CLONE) reduce 245) ((T_CONSTANT_ENCAPSED_STRING) reduce 245) ((T_CONTINUE) reduce 245) ((T_DEC) reduce 245) ((T_DECLARE) reduce 245) ((T_DIR) reduce 245) ((T_DNUMBER) reduce 245) ((T_DO) reduce 245) ((T_DOUBLE_CAST) reduce 245) ((T_ECHO) reduce 245) ((T_ELSE) reduce 245) ((T_ELSEIF) reduce 245) ((T_EMPTY) reduce 245) ((T_ENDDECLARE) reduce 245) ((T_ENUM) reduce 245) ((T_EVAL) reduce 245) ((T_EXIT) reduce 245) ((T_FILE) reduce 245) ((T_FINAL) reduce 245) ((T_FN) reduce 245) ((T_FOR) reduce 245) ((T_FOREACH) reduce 245) ((T_FUNCTION) reduce 245) ((T_FUNC_C) reduce 245) ((T_GLOBAL) reduce 245) ((T_GOTO) reduce 245) ((T_HALT_COMPILER) reduce 245) ((T_IF) reduce 245) ((T_INC) reduce 245) ((T_INCLUDE) reduce 245) ((T_INCLUDE_ONCE) reduce 245) ((T_INLINE_HTML) reduce 245) ((T_INTERFACE) reduce 245) ((T_INT_CAST) reduce 245) ((T_ISSET) reduce 245) ((T_LINE) reduce 245) ((T_LIST) reduce 245) ((T_LNUMBER) reduce 245) ((T_MATCH) reduce 245) ((T_METHOD_C) reduce 245) ((T_NAME_FULLY_QUALIFIED) reduce 245) ((T_NAME_QUALIFIED) reduce 245) ((T_NAME_RELATIVE) reduce 245) ((T_NEW) reduce 245) ((T_NS_C) reduce 245) ((T_OBJECT_CAST) reduce 245) ((T_PRINT) reduce 245) ((T_PROPERTY_C) reduce 245) ((T_READONLY) reduce 245) ((T_REQUIRE) reduce 245) ((T_REQUIRE_ONCE) reduce 245) ((T_RETURN) reduce 245) ((T_START_HEREDOC) reduce 245) ((T_STATIC) reduce 245) ((T_STRING) reduce 245) ((T_STRING_CAST) reduce 245) ((T_SWITCH) reduce 245) ((T_THROW) reduce 245) ((T_TRAIT) reduce 245) ((T_TRAIT_C) reduce 245) ((T_TRY) reduce 245) ((T_UNSET) reduce 245) ((T_UNSET_CAST) reduce 245) ((T_VARIABLE) reduce 245) ((T_VOID_CAST) reduce 245) ((T_WHILE) reduce 245) ((T_YIELD) reduce 245) ((T_YIELD_FROM) reduce 245) (("[") reduce 245) (("`") reduce 245) (("{") reduce 245) (("~") reduce 245)) 23193 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_ELSE) reduce 147) ((T_ELSEIF) reduce 147) ((T_EMPTY) reduce 147) ((T_ENDDECLARE) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_VARIABLE) reduce 147) ((T_VOID_CAST) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 23195 ((("!") reduce 247) (("\"") reduce 247) (("$") reduce 247) (("(") reduce 247) (("+") reduce 247) (("-") reduce 247) ((";") reduce 247) (("@") reduce 247) ((T_ABSTRACT) reduce 247) ((T_ARRAY) reduce 247) ((T_ARRAY_CAST) reduce 247) ((T_ATTRIBUTE) reduce 247) ((T_BOOL_CAST) reduce 247) ((T_BREAK) reduce 247) ((T_CLASS) reduce 247) ((T_CLASS_C) reduce 247) ((T_CLONE) reduce 247) ((T_CONSTANT_ENCAPSED_STRING) reduce 247) ((T_CONTINUE) reduce 247) ((T_DEC) reduce 247) ((T_DECLARE) reduce 247) ((T_DIR) reduce 247) ((T_DNUMBER) reduce 247) ((T_DO) reduce 247) ((T_DOUBLE_CAST) reduce 247) ((T_ECHO) reduce 247) ((T_ELSE) reduce 247) ((T_ELSEIF) reduce 247) ((T_EMPTY) reduce 247) ((T_ENDDECLARE) reduce 247) ((T_ENUM) reduce 247) ((T_EVAL) reduce 247) ((T_EXIT) reduce 247) ((T_FILE) reduce 247) ((T_FINAL) reduce 247) ((T_FN) reduce 247) ((T_FOR) reduce 247) ((T_FOREACH) reduce 247) ((T_FUNCTION) reduce 247) ((T_FUNC_C) reduce 247) ((T_GLOBAL) reduce 247) ((T_GOTO) reduce 247) ((T_HALT_COMPILER) reduce 247) ((T_IF) reduce 247) ((T_INC) reduce 247) ((T_INCLUDE) reduce 247) ((T_INCLUDE_ONCE) reduce 247) ((T_INLINE_HTML) reduce 247) ((T_INTERFACE) reduce 247) ((T_INT_CAST) reduce 247) ((T_ISSET) reduce 247) ((T_LINE) reduce 247) ((T_LIST) reduce 247) ((T_LNUMBER) reduce 247) ((T_MATCH) reduce 247) ((T_METHOD_C) reduce 247) ((T_NAME_FULLY_QUALIFIED) reduce 247) ((T_NAME_QUALIFIED) reduce 247) ((T_NAME_RELATIVE) reduce 247) ((T_NEW) reduce 247) ((T_NS_C) reduce 247) ((T_OBJECT_CAST) reduce 247) ((T_PRINT) reduce 247) ((T_PROPERTY_C) reduce 247) ((T_READONLY) reduce 247) ((T_REQUIRE) reduce 247) ((T_REQUIRE_ONCE) reduce 247) ((T_RETURN) reduce 247) ((T_START_HEREDOC) reduce 247) ((T_STATIC) reduce 247) ((T_STRING) reduce 247) ((T_STRING_CAST) reduce 247) ((T_SWITCH) reduce 247) ((T_THROW) reduce 247) ((T_TRAIT) reduce 247) ((T_TRAIT_C) reduce 247) ((T_TRY) reduce 247) ((T_UNSET) reduce 247) ((T_UNSET_CAST) reduce 247) ((T_VARIABLE) reduce 247) ((T_VOID_CAST) reduce 247) ((T_WHILE) reduce 247) ((T_YIELD) reduce 247) ((T_YIELD_FROM) reduce 247) (("[") reduce 247) (("`") reduce 247) (("{") reduce 247) (("~") reduce 247)) 23196 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_ELSE) reduce 161) ((T_ELSEIF) reduce 161) ((T_EMPTY) reduce 161) ((T_ENDDECLARE) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_VOID_CAST) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 23199 ((("!") reduce 250) (("\"") reduce 250) (("$") reduce 250) (("(") reduce 250) (("+") reduce 250) (("-") reduce 250) ((";") reduce 250) (("@") reduce 250) ((T_ABSTRACT) reduce 250) ((T_ARRAY) reduce 250) ((T_ARRAY_CAST) reduce 250) ((T_ATTRIBUTE) reduce 250) ((T_BOOL_CAST) reduce 250) ((T_BREAK) reduce 250) ((T_CLASS) reduce 250) ((T_CLASS_C) reduce 250) ((T_CLONE) reduce 250) ((T_CONSTANT_ENCAPSED_STRING) reduce 250) ((T_CONTINUE) reduce 250) ((T_DEC) reduce 250) ((T_DECLARE) reduce 250) ((T_DIR) reduce 250) ((T_DNUMBER) reduce 250) ((T_DO) reduce 250) ((T_DOUBLE_CAST) reduce 250) ((T_ECHO) reduce 250) ((T_ELSE) reduce 250) ((T_ELSEIF) reduce 250) ((T_EMPTY) reduce 250) ((T_ENDDECLARE) reduce 250) ((T_ENUM) reduce 250) ((T_EVAL) reduce 250) ((T_EXIT) reduce 250) ((T_FILE) reduce 250) ((T_FINAL) reduce 250) ((T_FN) reduce 250) ((T_FOR) reduce 250) ((T_FOREACH) reduce 250) ((T_FUNCTION) reduce 250) ((T_FUNC_C) reduce 250) ((T_GLOBAL) reduce 250) ((T_GOTO) reduce 250) ((T_HALT_COMPILER) reduce 250) ((T_IF) reduce 250) ((T_INC) reduce 250) ((T_INCLUDE) reduce 250) ((T_INCLUDE_ONCE) reduce 250) ((T_INLINE_HTML) reduce 250) ((T_INTERFACE) reduce 250) ((T_INT_CAST) reduce 250) ((T_ISSET) reduce 250) ((T_LINE) reduce 250) ((T_LIST) reduce 250) ((T_LNUMBER) reduce 250) ((T_MATCH) reduce 250) ((T_METHOD_C) reduce 250) ((T_NAME_FULLY_QUALIFIED) reduce 250) ((T_NAME_QUALIFIED) reduce 250) ((T_NAME_RELATIVE) reduce 250) ((T_NEW) reduce 250) ((T_NS_C) reduce 250) ((T_OBJECT_CAST) reduce 250) ((T_PRINT) reduce 250) ((T_PROPERTY_C) reduce 250) ((T_READONLY) reduce 250) ((T_REQUIRE) reduce 250) ((T_REQUIRE_ONCE) reduce 250) ((T_RETURN) reduce 250) ((T_START_HEREDOC) reduce 250) ((T_STATIC) reduce 250) ((T_STRING) reduce 250) ((T_STRING_CAST) reduce 250) ((T_SWITCH) reduce 250) ((T_THROW) reduce 250) ((T_TRAIT) reduce 250) ((T_TRAIT_C) reduce 250) ((T_TRY) reduce 250) ((T_UNSET) reduce 250) ((T_UNSET_CAST) reduce 250) ((T_VARIABLE) reduce 250) ((T_VOID_CAST) reduce 250) ((T_WHILE) reduce 250) ((T_YIELD) reduce 250) ((T_YIELD_FROM) reduce 250) (("[") reduce 250) (("`") reduce 250) (("{") reduce 250) (("~") reduce 250)) 23203 ((("!") reduce 251) (("\"") reduce 251) (("$") reduce 251) (("(") reduce 251) (("+") reduce 251) (("-") reduce 251) ((";") reduce 251) (("@") reduce 251) ((T_ABSTRACT) reduce 251) ((T_ARRAY) reduce 251) ((T_ARRAY_CAST) reduce 251) ((T_ATTRIBUTE) reduce 251) ((T_BOOL_CAST) reduce 251) ((T_BREAK) reduce 251) ((T_CLASS) reduce 251) ((T_CLASS_C) reduce 251) ((T_CLONE) reduce 251) ((T_CONSTANT_ENCAPSED_STRING) reduce 251) ((T_CONTINUE) reduce 251) ((T_DEC) reduce 251) ((T_DECLARE) reduce 251) ((T_DIR) reduce 251) ((T_DNUMBER) reduce 251) ((T_DO) reduce 251) ((T_DOUBLE_CAST) reduce 251) ((T_ECHO) reduce 251) ((T_ELSE) reduce 251) ((T_ELSEIF) reduce 251) ((T_EMPTY) reduce 251) ((T_ENDDECLARE) reduce 251) ((T_ENUM) reduce 251) ((T_EVAL) reduce 251) ((T_EXIT) reduce 251) ((T_FILE) reduce 251) ((T_FINAL) reduce 251) ((T_FN) reduce 251) ((T_FOR) reduce 251) ((T_FOREACH) reduce 251) ((T_FUNCTION) reduce 251) ((T_FUNC_C) reduce 251) ((T_GLOBAL) reduce 251) ((T_GOTO) reduce 251) ((T_HALT_COMPILER) reduce 251) ((T_IF) reduce 251) ((T_INC) reduce 251) ((T_INCLUDE) reduce 251) ((T_INCLUDE_ONCE) reduce 251) ((T_INLINE_HTML) reduce 251) ((T_INTERFACE) reduce 251) ((T_INT_CAST) reduce 251) ((T_ISSET) reduce 251) ((T_LINE) reduce 251) ((T_LIST) reduce 251) ((T_LNUMBER) reduce 251) ((T_MATCH) reduce 251) ((T_METHOD_C) reduce 251) ((T_NAME_FULLY_QUALIFIED) reduce 251) ((T_NAME_QUALIFIED) reduce 251) ((T_NAME_RELATIVE) reduce 251) ((T_NEW) reduce 251) ((T_NS_C) reduce 251) ((T_OBJECT_CAST) reduce 251) ((T_PRINT) reduce 251) ((T_PROPERTY_C) reduce 251) ((T_READONLY) reduce 251) ((T_REQUIRE) reduce 251) ((T_REQUIRE_ONCE) reduce 251) ((T_RETURN) reduce 251) ((T_START_HEREDOC) reduce 251) ((T_STATIC) reduce 251) ((T_STRING) reduce 251) ((T_STRING_CAST) reduce 251) ((T_SWITCH) reduce 251) ((T_THROW) reduce 251) ((T_TRAIT) reduce 251) ((T_TRAIT_C) reduce 251) ((T_TRY) reduce 251) ((T_UNSET) reduce 251) ((T_UNSET_CAST) reduce 251) ((T_VARIABLE) reduce 251) ((T_VOID_CAST) reduce 251) ((T_WHILE) reduce 251) ((T_YIELD) reduce 251) ((T_YIELD_FROM) reduce 251) (("[") reduce 251) (("`") reduce 251) (("{") reduce 251) (("~") reduce 251)) 23208 ((("!") reduce 242) (("\"") reduce 242) (("$") reduce 242) (("(") reduce 242) (("+") reduce 242) (("-") reduce 242) ((";") reduce 242) (("@") reduce 242) ((T_ABSTRACT) reduce 242) ((T_ARRAY) reduce 242) ((T_ARRAY_CAST) reduce 242) ((T_ATTRIBUTE) reduce 242) ((T_BOOL_CAST) reduce 242) ((T_BREAK) reduce 242) ((T_CLASS) reduce 242) ((T_CLASS_C) reduce 242) ((T_CLONE) reduce 242) ((T_CONSTANT_ENCAPSED_STRING) reduce 242) ((T_CONTINUE) reduce 242) ((T_DEC) reduce 242) ((T_DECLARE) reduce 242) ((T_DIR) reduce 242) ((T_DNUMBER) reduce 242) ((T_DO) reduce 242) ((T_DOUBLE_CAST) reduce 242) ((T_ECHO) reduce 242) ((T_ELSE) reduce 242) ((T_ELSEIF) reduce 242) ((T_EMPTY) reduce 242) ((T_ENDDECLARE) reduce 242) ((T_ENUM) reduce 242) ((T_EVAL) reduce 242) ((T_EXIT) reduce 242) ((T_FILE) reduce 242) ((T_FINAL) reduce 242) ((T_FN) reduce 242) ((T_FOR) reduce 242) ((T_FOREACH) reduce 242) ((T_FUNCTION) reduce 242) ((T_FUNC_C) reduce 242) ((T_GLOBAL) reduce 242) ((T_GOTO) reduce 242) ((T_HALT_COMPILER) reduce 242) ((T_IF) reduce 242) ((T_INC) reduce 242) ((T_INCLUDE) reduce 242) ((T_INCLUDE_ONCE) reduce 242) ((T_INLINE_HTML) reduce 242) ((T_INTERFACE) reduce 242) ((T_INT_CAST) reduce 242) ((T_ISSET) reduce 242) ((T_LINE) reduce 242) ((T_LIST) reduce 242) ((T_LNUMBER) reduce 242) ((T_MATCH) reduce 242) ((T_METHOD_C) reduce 242) ((T_NAME_FULLY_QUALIFIED) reduce 242) ((T_NAME_QUALIFIED) reduce 242) ((T_NAME_RELATIVE) reduce 242) ((T_NEW) reduce 242) ((T_NS_C) reduce 242) ((T_OBJECT_CAST) reduce 242) ((T_PRINT) reduce 242) ((T_PROPERTY_C) reduce 242) ((T_READONLY) reduce 242) ((T_REQUIRE) reduce 242) ((T_REQUIRE_ONCE) reduce 242) ((T_RETURN) reduce 242) ((T_START_HEREDOC) reduce 242) ((T_STATIC) reduce 242) ((T_STRING) reduce 242) ((T_STRING_CAST) reduce 242) ((T_SWITCH) reduce 242) ((T_THROW) reduce 242) ((T_TRAIT) reduce 242) ((T_TRAIT_C) reduce 242) ((T_TRY) reduce 242) ((T_UNSET) reduce 242) ((T_UNSET_CAST) reduce 242) ((T_VARIABLE) reduce 242) ((T_VOID_CAST) reduce 242) ((T_WHILE) reduce 242) ((T_YIELD) reduce 242) ((T_YIELD_FROM) reduce 242) (("[") reduce 242) (("`") reduce 242) (("{") reduce 242) (("~") reduce 242)) 23209 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_ELSE) reduce 150) ((T_ELSEIF) reduce 150) ((T_EMPTY) reduce 150) ((T_ENDDECLARE) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_VOID_CAST) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 23212 ((("!") reduce 243) (("\"") reduce 243) (("$") reduce 243) (("(") reduce 243) (("+") reduce 243) (("-") reduce 243) ((";") reduce 243) (("@") reduce 243) ((T_ABSTRACT) reduce 243) ((T_ARRAY) reduce 243) ((T_ARRAY_CAST) reduce 243) ((T_ATTRIBUTE) reduce 243) ((T_BOOL_CAST) reduce 243) ((T_BREAK) reduce 243) ((T_CLASS) reduce 243) ((T_CLASS_C) reduce 243) ((T_CLONE) reduce 243) ((T_CONSTANT_ENCAPSED_STRING) reduce 243) ((T_CONTINUE) reduce 243) ((T_DEC) reduce 243) ((T_DECLARE) reduce 243) ((T_DIR) reduce 243) ((T_DNUMBER) reduce 243) ((T_DO) reduce 243) ((T_DOUBLE_CAST) reduce 243) ((T_ECHO) reduce 243) ((T_ELSE) reduce 243) ((T_ELSEIF) reduce 243) ((T_EMPTY) reduce 243) ((T_ENDDECLARE) reduce 243) ((T_ENUM) reduce 243) ((T_EVAL) reduce 243) ((T_EXIT) reduce 243) ((T_FILE) reduce 243) ((T_FINAL) reduce 243) ((T_FN) reduce 243) ((T_FOR) reduce 243) ((T_FOREACH) reduce 243) ((T_FUNCTION) reduce 243) ((T_FUNC_C) reduce 243) ((T_GLOBAL) reduce 243) ((T_GOTO) reduce 243) ((T_HALT_COMPILER) reduce 243) ((T_IF) reduce 243) ((T_INC) reduce 243) ((T_INCLUDE) reduce 243) ((T_INCLUDE_ONCE) reduce 243) ((T_INLINE_HTML) reduce 243) ((T_INTERFACE) reduce 243) ((T_INT_CAST) reduce 243) ((T_ISSET) reduce 243) ((T_LINE) reduce 243) ((T_LIST) reduce 243) ((T_LNUMBER) reduce 243) ((T_MATCH) reduce 243) ((T_METHOD_C) reduce 243) ((T_NAME_FULLY_QUALIFIED) reduce 243) ((T_NAME_QUALIFIED) reduce 243) ((T_NAME_RELATIVE) reduce 243) ((T_NEW) reduce 243) ((T_NS_C) reduce 243) ((T_OBJECT_CAST) reduce 243) ((T_PRINT) reduce 243) ((T_PROPERTY_C) reduce 243) ((T_READONLY) reduce 243) ((T_REQUIRE) reduce 243) ((T_REQUIRE_ONCE) reduce 243) ((T_RETURN) reduce 243) ((T_START_HEREDOC) reduce 243) ((T_STATIC) reduce 243) ((T_STRING) reduce 243) ((T_STRING_CAST) reduce 243) ((T_SWITCH) reduce 243) ((T_THROW) reduce 243) ((T_TRAIT) reduce 243) ((T_TRAIT_C) reduce 243) ((T_TRY) reduce 243) ((T_UNSET) reduce 243) ((T_UNSET_CAST) reduce 243) ((T_VARIABLE) reduce 243) ((T_VOID_CAST) reduce 243) ((T_WHILE) reduce 243) ((T_YIELD) reduce 243) ((T_YIELD_FROM) reduce 243) (("[") reduce 243) (("`") reduce 243) (("{") reduce 243) (("~") reduce 243)) 23214 ((("!") reduce 170) (("\"") reduce 170) (("$") reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_ELSE) reduce 170) ((T_ELSEIF) reduce 170) ((T_EMPTY) reduce 170) ((T_ENDDECLARE) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_VARIABLE) reduce 170) ((T_VOID_CAST) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("~") reduce 170)) 23219 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_ELSE) reduce 162) ((T_ELSEIF) reduce 162) ((T_EMPTY) reduce 162) ((T_ENDDECLARE) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_VOID_CAST) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 23222 ((("!") reduce 171) (("\"") reduce 171) (("$") reduce 171) (("(") reduce 171) (("+") reduce 171) (("-") reduce 171) ((";") reduce 171) (("@") reduce 171) ((T_ABSTRACT) reduce 171) ((T_ARRAY) reduce 171) ((T_ARRAY_CAST) reduce 171) ((T_ATTRIBUTE) reduce 171) ((T_BOOL_CAST) reduce 171) ((T_BREAK) reduce 171) ((T_CATCH) reduce 171) ((T_CLASS) reduce 171) ((T_CLASS_C) reduce 171) ((T_CLONE) reduce 171) ((T_CONSTANT_ENCAPSED_STRING) reduce 171) ((T_CONTINUE) reduce 171) ((T_DEC) reduce 171) ((T_DECLARE) reduce 171) ((T_DIR) reduce 171) ((T_DNUMBER) reduce 171) ((T_DO) reduce 171) ((T_DOUBLE_CAST) reduce 171) ((T_ECHO) reduce 171) ((T_ELSE) reduce 171) ((T_ELSEIF) reduce 171) ((T_EMPTY) reduce 171) ((T_ENDDECLARE) reduce 171) ((T_ENUM) reduce 171) ((T_EVAL) reduce 171) ((T_EXIT) reduce 171) ((T_FILE) reduce 171) ((T_FINAL) reduce 171) ((T_FINALLY) reduce 171) ((T_FN) reduce 171) ((T_FOR) reduce 171) ((T_FOREACH) reduce 171) ((T_FUNCTION) reduce 171) ((T_FUNC_C) reduce 171) ((T_GLOBAL) reduce 171) ((T_GOTO) reduce 171) ((T_HALT_COMPILER) reduce 171) ((T_IF) reduce 171) ((T_INC) reduce 171) ((T_INCLUDE) reduce 171) ((T_INCLUDE_ONCE) reduce 171) ((T_INLINE_HTML) reduce 171) ((T_INTERFACE) reduce 171) ((T_INT_CAST) reduce 171) ((T_ISSET) reduce 171) ((T_LINE) reduce 171) ((T_LIST) reduce 171) ((T_LNUMBER) reduce 171) ((T_MATCH) reduce 171) ((T_METHOD_C) reduce 171) ((T_NAME_FULLY_QUALIFIED) reduce 171) ((T_NAME_QUALIFIED) reduce 171) ((T_NAME_RELATIVE) reduce 171) ((T_NEW) reduce 171) ((T_NS_C) reduce 171) ((T_OBJECT_CAST) reduce 171) ((T_PRINT) reduce 171) ((T_PROPERTY_C) reduce 171) ((T_READONLY) reduce 171) ((T_REQUIRE) reduce 171) ((T_REQUIRE_ONCE) reduce 171) ((T_RETURN) reduce 171) ((T_START_HEREDOC) reduce 171) ((T_STATIC) reduce 171) ((T_STRING) reduce 171) ((T_STRING_CAST) reduce 171) ((T_SWITCH) reduce 171) ((T_THROW) reduce 171) ((T_TRAIT) reduce 171) ((T_TRAIT_C) reduce 171) ((T_TRY) reduce 171) ((T_UNSET) reduce 171) ((T_UNSET_CAST) reduce 171) ((T_VARIABLE) reduce 171) ((T_VOID_CAST) reduce 171) ((T_WHILE) reduce 171) ((T_YIELD) reduce 171) ((T_YIELD_FROM) reduce 171) (("[") reduce 171) (("`") reduce 171) (("{") reduce 171) (("~") reduce 171)) 23223 ((("!") reduce 177) (("\"") reduce 177) (("$") reduce 177) (("(") reduce 177) (("+") reduce 177) (("-") reduce 177) ((";") reduce 177) (("@") reduce 177) ((T_ABSTRACT) reduce 177) ((T_ARRAY) reduce 177) ((T_ARRAY_CAST) reduce 177) ((T_ATTRIBUTE) reduce 177) ((T_BOOL_CAST) reduce 177) ((T_BREAK) reduce 177) ((T_CATCH) shift) ((T_CLASS) reduce 177) ((T_CLASS_C) reduce 177) ((T_CLONE) reduce 177) ((T_CONSTANT_ENCAPSED_STRING) reduce 177) ((T_CONTINUE) reduce 177) ((T_DEC) reduce 177) ((T_DECLARE) reduce 177) ((T_DIR) reduce 177) ((T_DNUMBER) reduce 177) ((T_DO) reduce 177) ((T_DOUBLE_CAST) reduce 177) ((T_ECHO) reduce 177) ((T_ELSE) reduce 177) ((T_ELSEIF) reduce 177) ((T_EMPTY) reduce 177) ((T_ENDDECLARE) reduce 177) ((T_ENUM) reduce 177) ((T_EVAL) reduce 177) ((T_EXIT) reduce 177) ((T_FILE) reduce 177) ((T_FINAL) reduce 177) ((T_FINALLY) shift) ((T_FN) reduce 177) ((T_FOR) reduce 177) ((T_FOREACH) reduce 177) ((T_FUNCTION) reduce 177) ((T_FUNC_C) reduce 177) ((T_GLOBAL) reduce 177) ((T_GOTO) reduce 177) ((T_HALT_COMPILER) reduce 177) ((T_IF) reduce 177) ((T_INC) reduce 177) ((T_INCLUDE) reduce 177) ((T_INCLUDE_ONCE) reduce 177) ((T_INLINE_HTML) reduce 177) ((T_INTERFACE) reduce 177) ((T_INT_CAST) reduce 177) ((T_ISSET) reduce 177) ((T_LINE) reduce 177) ((T_LIST) reduce 177) ((T_LNUMBER) reduce 177) ((T_MATCH) reduce 177) ((T_METHOD_C) reduce 177) ((T_NAME_FULLY_QUALIFIED) reduce 177) ((T_NAME_QUALIFIED) reduce 177) ((T_NAME_RELATIVE) reduce 177) ((T_NEW) reduce 177) ((T_NS_C) reduce 177) ((T_OBJECT_CAST) reduce 177) ((T_PRINT) reduce 177) ((T_PROPERTY_C) reduce 177) ((T_READONLY) reduce 177) ((T_REQUIRE) reduce 177) ((T_REQUIRE_ONCE) reduce 177) ((T_RETURN) reduce 177) ((T_START_HEREDOC) reduce 177) ((T_STATIC) reduce 177) ((T_STRING) reduce 177) ((T_STRING_CAST) reduce 177) ((T_SWITCH) reduce 177) ((T_THROW) reduce 177) ((T_TRAIT) reduce 177) ((T_TRAIT_C) reduce 177) ((T_TRY) reduce 177) ((T_UNSET) reduce 177) ((T_UNSET_CAST) reduce 177) ((T_VARIABLE) reduce 177) ((T_VOID_CAST) reduce 177) ((T_WHILE) reduce 177) ((T_YIELD) reduce 177) ((T_YIELD_FROM) reduce 177) (("[") reduce 177) (("`") reduce 177) (("{") reduce 177) (("~") reduce 177)) 23226 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_ELSE) reduce 167) ((T_ELSEIF) reduce 167) ((T_EMPTY) reduce 167) ((T_ENDDECLARE) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_VOID_CAST) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 23229 ((("!") reduce 178) (("\"") reduce 178) (("$") reduce 178) (("(") reduce 178) (("+") reduce 178) (("-") reduce 178) ((";") reduce 178) (("@") reduce 178) ((T_ABSTRACT) reduce 178) ((T_ARRAY) reduce 178) ((T_ARRAY_CAST) reduce 178) ((T_ATTRIBUTE) reduce 178) ((T_BOOL_CAST) reduce 178) ((T_BREAK) reduce 178) ((T_CLASS) reduce 178) ((T_CLASS_C) reduce 178) ((T_CLONE) reduce 178) ((T_CONSTANT_ENCAPSED_STRING) reduce 178) ((T_CONTINUE) reduce 178) ((T_DEC) reduce 178) ((T_DECLARE) reduce 178) ((T_DIR) reduce 178) ((T_DNUMBER) reduce 178) ((T_DO) reduce 178) ((T_DOUBLE_CAST) reduce 178) ((T_ECHO) reduce 178) ((T_ELSE) reduce 178) ((T_ELSEIF) reduce 178) ((T_EMPTY) reduce 178) ((T_ENDDECLARE) reduce 178) ((T_ENUM) reduce 178) ((T_EVAL) reduce 178) ((T_EXIT) reduce 178) ((T_FILE) reduce 178) ((T_FINAL) reduce 178) ((T_FN) reduce 178) ((T_FOR) reduce 178) ((T_FOREACH) reduce 178) ((T_FUNCTION) reduce 178) ((T_FUNC_C) reduce 178) ((T_GLOBAL) reduce 178) ((T_GOTO) reduce 178) ((T_HALT_COMPILER) reduce 178) ((T_IF) reduce 178) ((T_INC) reduce 178) ((T_INCLUDE) reduce 178) ((T_INCLUDE_ONCE) reduce 178) ((T_INLINE_HTML) reduce 178) ((T_INTERFACE) reduce 178) ((T_INT_CAST) reduce 178) ((T_ISSET) reduce 178) ((T_LINE) reduce 178) ((T_LIST) reduce 178) ((T_LNUMBER) reduce 178) ((T_MATCH) reduce 178) ((T_METHOD_C) reduce 178) ((T_NAME_FULLY_QUALIFIED) reduce 178) ((T_NAME_QUALIFIED) reduce 178) ((T_NAME_RELATIVE) reduce 178) ((T_NEW) reduce 178) ((T_NS_C) reduce 178) ((T_OBJECT_CAST) reduce 178) ((T_PRINT) reduce 178) ((T_PROPERTY_C) reduce 178) ((T_READONLY) reduce 178) ((T_REQUIRE) reduce 178) ((T_REQUIRE_ONCE) reduce 178) ((T_RETURN) reduce 178) ((T_START_HEREDOC) reduce 178) ((T_STATIC) reduce 178) ((T_STRING) reduce 178) ((T_STRING_CAST) reduce 178) ((T_SWITCH) reduce 178) ((T_THROW) reduce 178) ((T_TRAIT) reduce 178) ((T_TRAIT_C) reduce 178) ((T_TRY) reduce 178) ((T_UNSET) reduce 178) ((T_UNSET_CAST) reduce 178) ((T_VARIABLE) reduce 178) ((T_VOID_CAST) reduce 178) ((T_WHILE) reduce 178) ((T_YIELD) reduce 178) ((T_YIELD_FROM) reduce 178) (("[") reduce 178) (("`") reduce 178) (("{") reduce 178) (("~") reduce 178)) 23236 ((("!") reduce 172) (("\"") reduce 172) (("$") reduce 172) (("(") reduce 172) (("+") reduce 172) (("-") reduce 172) ((";") reduce 172) (("@") reduce 172) ((T_ABSTRACT) reduce 172) ((T_ARRAY) reduce 172) ((T_ARRAY_CAST) reduce 172) ((T_ATTRIBUTE) reduce 172) ((T_BOOL_CAST) reduce 172) ((T_BREAK) reduce 172) ((T_CATCH) reduce 172) ((T_CLASS) reduce 172) ((T_CLASS_C) reduce 172) ((T_CLONE) reduce 172) ((T_CONSTANT_ENCAPSED_STRING) reduce 172) ((T_CONTINUE) reduce 172) ((T_DEC) reduce 172) ((T_DECLARE) reduce 172) ((T_DIR) reduce 172) ((T_DNUMBER) reduce 172) ((T_DO) reduce 172) ((T_DOUBLE_CAST) reduce 172) ((T_ECHO) reduce 172) ((T_ELSE) reduce 172) ((T_ELSEIF) reduce 172) ((T_EMPTY) reduce 172) ((T_ENDDECLARE) reduce 172) ((T_ENUM) reduce 172) ((T_EVAL) reduce 172) ((T_EXIT) reduce 172) ((T_FILE) reduce 172) ((T_FINAL) reduce 172) ((T_FINALLY) reduce 172) ((T_FN) reduce 172) ((T_FOR) reduce 172) ((T_FOREACH) reduce 172) ((T_FUNCTION) reduce 172) ((T_FUNC_C) reduce 172) ((T_GLOBAL) reduce 172) ((T_GOTO) reduce 172) ((T_HALT_COMPILER) reduce 172) ((T_IF) reduce 172) ((T_INC) reduce 172) ((T_INCLUDE) reduce 172) ((T_INCLUDE_ONCE) reduce 172) ((T_INLINE_HTML) reduce 172) ((T_INTERFACE) reduce 172) ((T_INT_CAST) reduce 172) ((T_ISSET) reduce 172) ((T_LINE) reduce 172) ((T_LIST) reduce 172) ((T_LNUMBER) reduce 172) ((T_MATCH) reduce 172) ((T_METHOD_C) reduce 172) ((T_NAME_FULLY_QUALIFIED) reduce 172) ((T_NAME_QUALIFIED) reduce 172) ((T_NAME_RELATIVE) reduce 172) ((T_NEW) reduce 172) ((T_NS_C) reduce 172) ((T_OBJECT_CAST) reduce 172) ((T_PRINT) reduce 172) ((T_PROPERTY_C) reduce 172) ((T_READONLY) reduce 172) ((T_REQUIRE) reduce 172) ((T_REQUIRE_ONCE) reduce 172) ((T_RETURN) reduce 172) ((T_START_HEREDOC) reduce 172) ((T_STATIC) reduce 172) ((T_STRING) reduce 172) ((T_STRING_CAST) reduce 172) ((T_SWITCH) reduce 172) ((T_THROW) reduce 172) ((T_TRAIT) reduce 172) ((T_TRAIT_C) reduce 172) ((T_TRY) reduce 172) ((T_UNSET) reduce 172) ((T_UNSET_CAST) reduce 172) ((T_VARIABLE) reduce 172) ((T_VOID_CAST) reduce 172) ((T_WHILE) reduce 172) ((T_YIELD) reduce 172) ((T_YIELD_FROM) reduce 172) (("[") reduce 172) (("`") reduce 172) (("{") reduce 172) (("~") reduce 172)) 23241 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_ELSE) reduce 153) ((T_ELSEIF) reduce 153) ((T_EMPTY) reduce 153) ((T_ENDDECLARE) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_VOID_CAST) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 23245 ((("!") reduce 224) (("\"") reduce 224) (("$") reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_ELSE) reduce 224) ((T_ELSEIF) reduce 224) ((T_EMPTY) reduce 224) ((T_ENDDECLARE) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_VARIABLE) reduce 224) ((T_VOID_CAST) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("~") reduce 224)) 23247 ((("!") reduce 225) (("\"") reduce 225) (("$") reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_ELSE) reduce 225) ((T_ELSEIF) reduce 225) ((T_EMPTY) reduce 225) ((T_ENDDECLARE) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_VARIABLE) reduce 225) ((T_VOID_CAST) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("~") reduce 225)) 23251 ((("!") reduce 226) (("\"") reduce 226) (("$") reduce 226) (("(") reduce 226) (("+") reduce 226) (("-") reduce 226) ((";") reduce 226) (("@") reduce 226) ((T_ABSTRACT) reduce 226) ((T_ARRAY) reduce 226) ((T_ARRAY_CAST) reduce 226) ((T_ATTRIBUTE) reduce 226) ((T_BOOL_CAST) reduce 226) ((T_BREAK) reduce 226) ((T_CLASS) reduce 226) ((T_CLASS_C) reduce 226) ((T_CLONE) reduce 226) ((T_CONSTANT_ENCAPSED_STRING) reduce 226) ((T_CONTINUE) reduce 226) ((T_DEC) reduce 226) ((T_DECLARE) reduce 226) ((T_DIR) reduce 226) ((T_DNUMBER) reduce 226) ((T_DO) reduce 226) ((T_DOUBLE_CAST) reduce 226) ((T_ECHO) reduce 226) ((T_ELSE) reduce 226) ((T_ELSEIF) reduce 226) ((T_EMPTY) reduce 226) ((T_ENDDECLARE) reduce 226) ((T_ENUM) reduce 226) ((T_EVAL) reduce 226) ((T_EXIT) reduce 226) ((T_FILE) reduce 226) ((T_FINAL) reduce 226) ((T_FN) reduce 226) ((T_FOR) reduce 226) ((T_FOREACH) reduce 226) ((T_FUNCTION) reduce 226) ((T_FUNC_C) reduce 226) ((T_GLOBAL) reduce 226) ((T_GOTO) reduce 226) ((T_HALT_COMPILER) reduce 226) ((T_IF) reduce 226) ((T_INC) reduce 226) ((T_INCLUDE) reduce 226) ((T_INCLUDE_ONCE) reduce 226) ((T_INLINE_HTML) reduce 226) ((T_INTERFACE) reduce 226) ((T_INT_CAST) reduce 226) ((T_ISSET) reduce 226) ((T_LINE) reduce 226) ((T_LIST) reduce 226) ((T_LNUMBER) reduce 226) ((T_MATCH) reduce 226) ((T_METHOD_C) reduce 226) ((T_NAME_FULLY_QUALIFIED) reduce 226) ((T_NAME_QUALIFIED) reduce 226) ((T_NAME_RELATIVE) reduce 226) ((T_NEW) reduce 226) ((T_NS_C) reduce 226) ((T_OBJECT_CAST) reduce 226) ((T_PRINT) reduce 226) ((T_PROPERTY_C) reduce 226) ((T_READONLY) reduce 226) ((T_REQUIRE) reduce 226) ((T_REQUIRE_ONCE) reduce 226) ((T_RETURN) reduce 226) ((T_START_HEREDOC) reduce 226) ((T_STATIC) reduce 226) ((T_STRING) reduce 226) ((T_STRING_CAST) reduce 226) ((T_SWITCH) reduce 226) ((T_THROW) reduce 226) ((T_TRAIT) reduce 226) ((T_TRAIT_C) reduce 226) ((T_TRY) reduce 226) ((T_UNSET) reduce 226) ((T_UNSET_CAST) reduce 226) ((T_VARIABLE) reduce 226) ((T_VOID_CAST) reduce 226) ((T_WHILE) reduce 226) ((T_YIELD) reduce 226) ((T_YIELD_FROM) reduce 226) (("[") reduce 226) (("`") reduce 226) (("{") reduce 226) (("~") reduce 226)) 23254 ((("!") reduce 227) (("\"") reduce 227) (("$") reduce 227) (("(") reduce 227) (("+") reduce 227) (("-") reduce 227) ((";") reduce 227) (("@") reduce 227) ((T_ABSTRACT) reduce 227) ((T_ARRAY) reduce 227) ((T_ARRAY_CAST) reduce 227) ((T_ATTRIBUTE) reduce 227) ((T_BOOL_CAST) reduce 227) ((T_BREAK) reduce 227) ((T_CLASS) reduce 227) ((T_CLASS_C) reduce 227) ((T_CLONE) reduce 227) ((T_CONSTANT_ENCAPSED_STRING) reduce 227) ((T_CONTINUE) reduce 227) ((T_DEC) reduce 227) ((T_DECLARE) reduce 227) ((T_DIR) reduce 227) ((T_DNUMBER) reduce 227) ((T_DO) reduce 227) ((T_DOUBLE_CAST) reduce 227) ((T_ECHO) reduce 227) ((T_ELSE) reduce 227) ((T_ELSEIF) reduce 227) ((T_EMPTY) reduce 227) ((T_ENDDECLARE) reduce 227) ((T_ENUM) reduce 227) ((T_EVAL) reduce 227) ((T_EXIT) reduce 227) ((T_FILE) reduce 227) ((T_FINAL) reduce 227) ((T_FN) reduce 227) ((T_FOR) reduce 227) ((T_FOREACH) reduce 227) ((T_FUNCTION) reduce 227) ((T_FUNC_C) reduce 227) ((T_GLOBAL) reduce 227) ((T_GOTO) reduce 227) ((T_HALT_COMPILER) reduce 227) ((T_IF) reduce 227) ((T_INC) reduce 227) ((T_INCLUDE) reduce 227) ((T_INCLUDE_ONCE) reduce 227) ((T_INLINE_HTML) reduce 227) ((T_INTERFACE) reduce 227) ((T_INT_CAST) reduce 227) ((T_ISSET) reduce 227) ((T_LINE) reduce 227) ((T_LIST) reduce 227) ((T_LNUMBER) reduce 227) ((T_MATCH) reduce 227) ((T_METHOD_C) reduce 227) ((T_NAME_FULLY_QUALIFIED) reduce 227) ((T_NAME_QUALIFIED) reduce 227) ((T_NAME_RELATIVE) reduce 227) ((T_NEW) reduce 227) ((T_NS_C) reduce 227) ((T_OBJECT_CAST) reduce 227) ((T_PRINT) reduce 227) ((T_PROPERTY_C) reduce 227) ((T_READONLY) reduce 227) ((T_REQUIRE) reduce 227) ((T_REQUIRE_ONCE) reduce 227) ((T_RETURN) reduce 227) ((T_START_HEREDOC) reduce 227) ((T_STATIC) reduce 227) ((T_STRING) reduce 227) ((T_STRING_CAST) reduce 227) ((T_SWITCH) reduce 227) ((T_THROW) reduce 227) ((T_TRAIT) reduce 227) ((T_TRAIT_C) reduce 227) ((T_TRY) reduce 227) ((T_UNSET) reduce 227) ((T_UNSET_CAST) reduce 227) ((T_VARIABLE) reduce 227) ((T_VOID_CAST) reduce 227) ((T_WHILE) reduce 227) ((T_YIELD) reduce 227) ((T_YIELD_FROM) reduce 227) (("[") reduce 227) (("`") reduce 227) (("{") reduce 227) (("~") reduce 227)) 23255 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_ELSE) reduce 169) ((T_ELSEIF) reduce 169) ((T_EMPTY) reduce 169) ((T_ENDDECLARE) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_VOID_CAST) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 23257 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_ELSE) reduce 158) ((T_ELSEIF) reduce 158) ((T_EMPTY) reduce 158) ((T_ENDDECLARE) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_VOID_CAST) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 23259 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_ELSE) reduce 156) ((T_ELSEIF) reduce 156) ((T_EMPTY) reduce 156) ((T_ENDDECLARE) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_VOID_CAST) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 23261 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_ELSE) reduce 168) ((T_ELSEIF) reduce 168) ((T_EMPTY) reduce 168) ((T_ENDDECLARE) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_VOID_CAST) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 23263 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_ELSE) reduce 157) ((T_ELSEIF) reduce 157) ((T_EMPTY) reduce 157) ((T_ENDDECLARE) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_VOID_CAST) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 23273 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_ELSE) reduce 164) ((T_ELSEIF) reduce 164) ((T_EMPTY) reduce 164) ((T_ENDDECLARE) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_VOID_CAST) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 23274 ((("!") reduce 220) (("\"") reduce 220) (("$") reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_ELSE) reduce 220) ((T_ELSEIF) reduce 220) ((T_EMPTY) reduce 220) ((T_ENDDECLARE) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_VARIABLE) reduce 220) ((T_VOID_CAST) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("~") reduce 220)) 23277 ((("!") reduce 221) (("\"") reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_ELSE) reduce 221) ((T_ELSEIF) reduce 221) ((T_EMPTY) reduce 221) ((T_ENDDECLARE) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_VARIABLE) reduce 221) ((T_VOID_CAST) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("~") reduce 221)) 23278 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_ELSE) reduce 163) ((T_ELSEIF) reduce 163) ((T_EMPTY) reduce 163) ((T_ENDDECLARE) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_VOID_CAST) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 23287 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_ELSE) reduce 152) ((T_ELSEIF) reduce 152) ((T_EMPTY) reduce 152) ((T_ENDDECLARE) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_VOID_CAST) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 23288 ((("!") reduce 218) (("\"") reduce 218) (("$") reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_ELSE) reduce 218) ((T_ELSEIF) reduce 218) ((T_EMPTY) reduce 218) ((T_ENDDECLARE) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_VARIABLE) reduce 218) ((T_VOID_CAST) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("~") reduce 218)) 23291 ((("!") reduce 219) (("\"") reduce 219) (("$") reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_ELSE) reduce 219) ((T_ELSEIF) reduce 219) ((T_EMPTY) reduce 219) ((T_ENDDECLARE) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_VARIABLE) reduce 219) ((T_VOID_CAST) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("~") reduce 219)) 23293 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_ELSE) reduce 159) ((T_ELSEIF) reduce 159) ((T_EMPTY) reduce 159) ((T_ENDDECLARE) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_VOID_CAST) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 23299 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_ELSE) reduce 151) ((T_ELSEIF) reduce 151) ((T_EMPTY) reduce 151) ((T_ENDDECLARE) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_VOID_CAST) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 23304 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_ELSE) reduce 165) ((T_ELSEIF) reduce 165) ((T_EMPTY) reduce 165) ((T_ENDDECLARE) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_VOID_CAST) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 23305 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_ELSE) reduce 222) ((T_ELSEIF) reduce 222) ((T_EMPTY) reduce 222) ((T_ENDDECLARE) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_VARIABLE) reduce 222) ((T_VOID_CAST) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("~") reduce 222)) 23308 ((("!") reduce 223) (("\"") reduce 223) (("$") reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_ELSE) reduce 223) ((T_ELSEIF) reduce 223) ((T_EMPTY) reduce 223) ((T_ENDDECLARE) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_VARIABLE) reduce 223) ((T_VOID_CAST) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("~") reduce 223)) 23310 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_ELSE) reduce 155) ((T_ELSEIF) reduce 155) ((T_EMPTY) reduce 155) ((T_ENDDECLARE) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_VOID_CAST) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 23312 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_ELSE) reduce 154) ((T_ELSEIF) reduce 154) ((T_EMPTY) reduce 154) ((T_ENDDECLARE) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_VOID_CAST) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 23313 ((("!") reduce 247) (("\"") reduce 247) (("$") reduce 247) (("(") reduce 247) (("+") reduce 247) (("-") reduce 247) ((";") reduce 247) (("@") reduce 247) ((T_ABSTRACT) reduce 247) ((T_ARRAY) reduce 247) ((T_ARRAY_CAST) reduce 247) ((T_ATTRIBUTE) reduce 247) ((T_BOOL_CAST) reduce 247) ((T_BREAK) reduce 247) ((T_CLASS) reduce 247) ((T_CLASS_C) reduce 247) ((T_CLONE) reduce 247) ((T_CONSTANT_ENCAPSED_STRING) reduce 247) ((T_CONTINUE) reduce 247) ((T_DEC) reduce 247) ((T_DECLARE) reduce 247) ((T_DIR) reduce 247) ((T_DNUMBER) reduce 247) ((T_DO) reduce 247) ((T_DOUBLE_CAST) reduce 247) ((T_ECHO) reduce 247) ((T_EMPTY) reduce 247) ((T_ENDDECLARE) reduce 247) ((T_ENUM) reduce 247) ((T_EVAL) reduce 247) ((T_EXIT) reduce 247) ((T_FILE) reduce 247) ((T_FINAL) reduce 247) ((T_FN) reduce 247) ((T_FOR) reduce 247) ((T_FOREACH) reduce 247) ((T_FUNCTION) reduce 247) ((T_FUNC_C) reduce 247) ((T_GLOBAL) reduce 247) ((T_GOTO) reduce 247) ((T_HALT_COMPILER) reduce 247) ((T_IF) reduce 247) ((T_INC) reduce 247) ((T_INCLUDE) reduce 247) ((T_INCLUDE_ONCE) reduce 247) ((T_INLINE_HTML) reduce 247) ((T_INTERFACE) reduce 247) ((T_INT_CAST) reduce 247) ((T_ISSET) reduce 247) ((T_LINE) reduce 247) ((T_LIST) reduce 247) ((T_LNUMBER) reduce 247) ((T_MATCH) reduce 247) ((T_METHOD_C) reduce 247) ((T_NAME_FULLY_QUALIFIED) reduce 247) ((T_NAME_QUALIFIED) reduce 247) ((T_NAME_RELATIVE) reduce 247) ((T_NEW) reduce 247) ((T_NS_C) reduce 247) ((T_OBJECT_CAST) reduce 247) ((T_PRINT) reduce 247) ((T_PROPERTY_C) reduce 247) ((T_READONLY) reduce 247) ((T_REQUIRE) reduce 247) ((T_REQUIRE_ONCE) reduce 247) ((T_RETURN) reduce 247) ((T_START_HEREDOC) reduce 247) ((T_STATIC) reduce 247) ((T_STRING) reduce 247) ((T_STRING_CAST) reduce 247) ((T_SWITCH) reduce 247) ((T_THROW) reduce 247) ((T_TRAIT) reduce 247) ((T_TRAIT_C) reduce 247) ((T_TRY) reduce 247) ((T_UNSET) reduce 247) ((T_UNSET_CAST) reduce 247) ((T_VARIABLE) reduce 247) ((T_VOID_CAST) reduce 247) ((T_WHILE) reduce 247) ((T_YIELD) reduce 247) ((T_YIELD_FROM) reduce 247) (("[") reduce 247) (("`") reduce 247) (("{") reduce 247) (("~") reduce 247)) 23324 ((("!") reduce 494) (("\"") reduce 494) (("$") reduce 494) (("(") reduce 494) (("+") reduce 494) (("-") reduce 494) ((";") reduce 494) (("@") reduce 494) ((T_ABSTRACT) reduce 494) ((T_ARRAY) reduce 494) ((T_ARRAY_CAST) reduce 494) ((T_ATTRIBUTE) reduce 494) ((T_BOOL_CAST) reduce 494) ((T_BREAK) reduce 494) ((T_CLASS) reduce 494) ((T_CLASS_C) reduce 494) ((T_CLONE) reduce 494) ((T_CONSTANT_ENCAPSED_STRING) reduce 494) ((T_CONTINUE) reduce 494) ((T_DEC) reduce 494) ((T_DECLARE) reduce 494) ((T_DIR) reduce 494) ((T_DNUMBER) reduce 494) ((T_DO) reduce 494) ((T_DOUBLE_CAST) reduce 494) ((T_ECHO) reduce 494) ((T_EMPTY) reduce 494) ((T_ENDDECLARE) reduce 494) ((T_ENUM) reduce 494) ((T_EVAL) reduce 494) ((T_EXIT) reduce 494) ((T_FILE) reduce 494) ((T_FINAL) reduce 494) ((T_FN) reduce 494) ((T_FOR) reduce 494) ((T_FOREACH) reduce 494) ((T_FUNCTION) reduce 494) ((T_FUNC_C) reduce 494) ((T_GLOBAL) reduce 494) ((T_GOTO) reduce 494) ((T_HALT_COMPILER) reduce 494) ((T_IF) reduce 494) ((T_INC) reduce 494) ((T_INCLUDE) reduce 494) ((T_INCLUDE_ONCE) reduce 494) ((T_INLINE_HTML) reduce 494) ((T_INTERFACE) reduce 494) ((T_INT_CAST) reduce 494) ((T_ISSET) reduce 494) ((T_LINE) reduce 494) ((T_LIST) reduce 494) ((T_LNUMBER) reduce 494) ((T_MATCH) reduce 494) ((T_METHOD_C) reduce 494) ((T_NAME_FULLY_QUALIFIED) reduce 494) ((T_NAME_QUALIFIED) reduce 494) ((T_NAME_RELATIVE) reduce 494) ((T_NEW) reduce 494) ((T_NS_C) reduce 494) ((T_OBJECT_CAST) reduce 494) ((T_PRINT) reduce 494) ((T_PROPERTY_C) reduce 494) ((T_READONLY) reduce 494) ((T_REQUIRE) reduce 494) ((T_REQUIRE_ONCE) reduce 494) ((T_RETURN) reduce 494) ((T_START_HEREDOC) reduce 494) ((T_STATIC) reduce 494) ((T_STRING) reduce 494) ((T_STRING_CAST) reduce 494) ((T_SWITCH) reduce 494) ((T_THROW) reduce 494) ((T_TRAIT) reduce 494) ((T_TRAIT_C) reduce 494) ((T_TRY) reduce 494) ((T_UNSET) reduce 494) ((T_UNSET_CAST) reduce 494) ((T_VARIABLE) reduce 494) ((T_VOID_CAST) reduce 494) ((T_WHILE) reduce 494) ((T_YIELD) reduce 494) ((T_YIELD_FROM) reduce 494) (("[") reduce 494) (("`") reduce 494) (("{") reduce 494) (("~") reduce 494)) 23325 ((("!") reduce 184) (("\"") reduce 184) (("$") reduce 184) (("(") reduce 184) (("+") reduce 184) (("-") reduce 184) ((";") reduce 184) (("@") reduce 184) ((T_ABSTRACT) reduce 184) ((T_ARRAY) reduce 184) ((T_ARRAY_CAST) reduce 184) ((T_ATTRIBUTE) reduce 184) ((T_BOOL_CAST) reduce 184) ((T_BREAK) reduce 184) ((T_CLASS) reduce 184) ((T_CLASS_C) reduce 184) ((T_CLONE) reduce 184) ((T_CONSTANT_ENCAPSED_STRING) reduce 184) ((T_CONTINUE) reduce 184) ((T_DEC) reduce 184) ((T_DECLARE) reduce 184) ((T_DIR) reduce 184) ((T_DNUMBER) reduce 184) ((T_DO) reduce 184) ((T_DOUBLE_CAST) reduce 184) ((T_ECHO) reduce 184) ((T_EMPTY) reduce 184) ((T_ENDDECLARE) reduce 184) ((T_ENUM) reduce 184) ((T_EVAL) reduce 184) ((T_EXIT) reduce 184) ((T_FILE) reduce 184) ((T_FINAL) reduce 184) ((T_FN) reduce 184) ((T_FOR) reduce 184) ((T_FOREACH) reduce 184) ((T_FUNCTION) reduce 184) ((T_FUNC_C) reduce 184) ((T_GLOBAL) reduce 184) ((T_GOTO) reduce 184) ((T_HALT_COMPILER) reduce 184) ((T_IF) reduce 184) ((T_INC) reduce 184) ((T_INCLUDE) reduce 184) ((T_INCLUDE_ONCE) reduce 184) ((T_INLINE_HTML) reduce 184) ((T_INTERFACE) reduce 184) ((T_INT_CAST) reduce 184) ((T_ISSET) reduce 184) ((T_LINE) reduce 184) ((T_LIST) reduce 184) ((T_LNUMBER) reduce 184) ((T_MATCH) reduce 184) ((T_METHOD_C) reduce 184) ((T_NAME_FULLY_QUALIFIED) reduce 184) ((T_NAME_QUALIFIED) reduce 184) ((T_NAME_RELATIVE) reduce 184) ((T_NEW) reduce 184) ((T_NS_C) reduce 184) ((T_OBJECT_CAST) reduce 184) ((T_PRINT) reduce 184) ((T_PROPERTY_C) reduce 184) ((T_READONLY) reduce 184) ((T_REQUIRE) reduce 184) ((T_REQUIRE_ONCE) reduce 184) ((T_RETURN) reduce 184) ((T_START_HEREDOC) reduce 184) ((T_STATIC) reduce 184) ((T_STRING) reduce 184) ((T_STRING_CAST) reduce 184) ((T_SWITCH) reduce 184) ((T_THROW) reduce 184) ((T_TRAIT) reduce 184) ((T_TRAIT_C) reduce 184) ((T_TRY) reduce 184) ((T_UNSET) reduce 184) ((T_UNSET_CAST) reduce 184) ((T_VARIABLE) reduce 184) ((T_VOID_CAST) reduce 184) ((T_WHILE) reduce 184) ((T_YIELD) reduce 184) ((T_YIELD_FROM) reduce 184) (("[") reduce 184) (("`") reduce 184) (("{") reduce 184) (("~") reduce 184)) 23326 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_EMPTY) reduce 161) ((T_ENDDECLARE) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_VOID_CAST) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 23334 ((("!") reduce 189) (("\"") reduce 189) (("$") reduce 189) (("(") reduce 189) (("+") reduce 189) (("-") reduce 189) ((";") reduce 189) (("@") reduce 189) ((T_ABSTRACT) reduce 189) ((T_ARRAY) reduce 189) ((T_ARRAY_CAST) reduce 189) ((T_ATTRIBUTE) reduce 189) ((T_BOOL_CAST) reduce 189) ((T_BREAK) reduce 189) ((T_CLASS) reduce 189) ((T_CLASS_C) reduce 189) ((T_CLONE) reduce 189) ((T_CONSTANT_ENCAPSED_STRING) reduce 189) ((T_CONTINUE) reduce 189) ((T_DEC) reduce 189) ((T_DECLARE) reduce 189) ((T_DIR) reduce 189) ((T_DNUMBER) reduce 189) ((T_DO) reduce 189) ((T_DOUBLE_CAST) reduce 189) ((T_ECHO) reduce 189) ((T_EMPTY) reduce 189) ((T_ENDDECLARE) reduce 189) ((T_ENUM) reduce 189) ((T_EVAL) reduce 189) ((T_EXIT) reduce 189) ((T_FILE) reduce 189) ((T_FINAL) reduce 189) ((T_FN) reduce 189) ((T_FOR) reduce 189) ((T_FOREACH) reduce 189) ((T_FUNCTION) reduce 189) ((T_FUNC_C) reduce 189) ((T_GLOBAL) reduce 189) ((T_GOTO) reduce 189) ((T_HALT_COMPILER) reduce 189) ((T_IF) reduce 189) ((T_INC) reduce 189) ((T_INCLUDE) reduce 189) ((T_INCLUDE_ONCE) reduce 189) ((T_INLINE_HTML) reduce 189) ((T_INTERFACE) reduce 189) ((T_INT_CAST) reduce 189) ((T_ISSET) reduce 189) ((T_LINE) reduce 189) ((T_LIST) reduce 189) ((T_LNUMBER) reduce 189) ((T_MATCH) reduce 189) ((T_METHOD_C) reduce 189) ((T_NAME_FULLY_QUALIFIED) reduce 189) ((T_NAME_QUALIFIED) reduce 189) ((T_NAME_RELATIVE) reduce 189) ((T_NEW) reduce 189) ((T_NS_C) reduce 189) ((T_OBJECT_CAST) reduce 189) ((T_PRINT) reduce 189) ((T_PROPERTY_C) reduce 189) ((T_READONLY) reduce 189) ((T_REQUIRE) reduce 189) ((T_REQUIRE_ONCE) reduce 189) ((T_RETURN) reduce 189) ((T_START_HEREDOC) reduce 189) ((T_STATIC) reduce 189) ((T_STRING) reduce 189) ((T_STRING_CAST) reduce 189) ((T_SWITCH) reduce 189) ((T_THROW) reduce 189) ((T_TRAIT) reduce 189) ((T_TRAIT_C) reduce 189) ((T_TRY) reduce 189) ((T_UNSET) reduce 189) ((T_UNSET_CAST) reduce 189) ((T_VARIABLE) reduce 189) ((T_VOID_CAST) reduce 189) ((T_WHILE) reduce 189) ((T_YIELD) reduce 189) ((T_YIELD_FROM) reduce 189) (("[") reduce 189) (("`") reduce 189) (("{") reduce 189) (("~") reduce 189)) 23335 ((("!") reduce 145) (("\"") reduce 145) (("$") reduce 145) (("(") reduce 145) (("+") reduce 145) (("-") reduce 145) ((";") reduce 145) (("@") reduce 145) ((T_ABSTRACT) reduce 145) ((T_ARRAY) reduce 145) ((T_ARRAY_CAST) reduce 145) ((T_ATTRIBUTE) reduce 145) ((T_BOOL_CAST) reduce 145) ((T_BREAK) reduce 145) ((T_CLASS) reduce 145) ((T_CLASS_C) reduce 145) ((T_CLONE) reduce 145) ((T_CONSTANT_ENCAPSED_STRING) reduce 145) ((T_CONTINUE) reduce 145) ((T_DEC) reduce 145) ((T_DECLARE) reduce 145) ((T_DIR) reduce 145) ((T_DNUMBER) reduce 145) ((T_DO) reduce 145) ((T_DOUBLE_CAST) reduce 145) ((T_ECHO) reduce 145) ((T_EMPTY) reduce 145) ((T_ENDDECLARE) reduce 145) ((T_ENUM) reduce 145) ((T_EVAL) reduce 145) ((T_EXIT) reduce 145) ((T_FILE) reduce 145) ((T_FINAL) reduce 145) ((T_FN) reduce 145) ((T_FOR) reduce 145) ((T_FOREACH) reduce 145) ((T_FUNCTION) reduce 145) ((T_FUNC_C) reduce 145) ((T_GLOBAL) reduce 145) ((T_GOTO) reduce 145) ((T_HALT_COMPILER) reduce 145) ((T_IF) reduce 145) ((T_INC) reduce 145) ((T_INCLUDE) reduce 145) ((T_INCLUDE_ONCE) reduce 145) ((T_INLINE_HTML) reduce 145) ((T_INTERFACE) reduce 145) ((T_INT_CAST) reduce 145) ((T_ISSET) reduce 145) ((T_LINE) reduce 145) ((T_LIST) reduce 145) ((T_LNUMBER) reduce 145) ((T_MATCH) reduce 145) ((T_METHOD_C) reduce 145) ((T_NAME_FULLY_QUALIFIED) reduce 145) ((T_NAME_QUALIFIED) reduce 145) ((T_NAME_RELATIVE) reduce 145) ((T_NEW) reduce 145) ((T_NS_C) reduce 145) ((T_OBJECT_CAST) reduce 145) ((T_PRINT) reduce 145) ((T_PROPERTY_C) reduce 145) ((T_READONLY) reduce 145) ((T_REQUIRE) reduce 145) ((T_REQUIRE_ONCE) reduce 145) ((T_RETURN) reduce 145) ((T_START_HEREDOC) reduce 145) ((T_STATIC) reduce 145) ((T_STRING) reduce 145) ((T_STRING_CAST) reduce 145) ((T_SWITCH) reduce 145) ((T_THROW) reduce 145) ((T_TRAIT) reduce 145) ((T_TRAIT_C) reduce 145) ((T_TRY) reduce 145) ((T_UNSET) reduce 145) ((T_UNSET_CAST) reduce 145) ((T_VARIABLE) reduce 145) ((T_VOID_CAST) reduce 145) ((T_WHILE) reduce 145) ((T_YIELD) reduce 145) ((T_YIELD_FROM) reduce 145) (("[") reduce 145) (("`") reduce 145) (("{") reduce 145) (("~") reduce 145)) 23338 ((("!") reduce 250) (("\"") reduce 250) (("$") reduce 250) (("(") reduce 250) (("+") reduce 250) (("-") reduce 250) ((";") reduce 250) (("@") reduce 250) ((T_ABSTRACT) reduce 250) ((T_ARRAY) reduce 250) ((T_ARRAY_CAST) reduce 250) ((T_ATTRIBUTE) reduce 250) ((T_BOOL_CAST) reduce 250) ((T_BREAK) reduce 250) ((T_CLASS) reduce 250) ((T_CLASS_C) reduce 250) ((T_CLONE) reduce 250) ((T_CONSTANT_ENCAPSED_STRING) reduce 250) ((T_CONTINUE) reduce 250) ((T_DEC) reduce 250) ((T_DECLARE) reduce 250) ((T_DIR) reduce 250) ((T_DNUMBER) reduce 250) ((T_DO) reduce 250) ((T_DOUBLE_CAST) reduce 250) ((T_ECHO) reduce 250) ((T_EMPTY) reduce 250) ((T_ENDDECLARE) reduce 250) ((T_ENUM) reduce 250) ((T_EVAL) reduce 250) ((T_EXIT) reduce 250) ((T_FILE) reduce 250) ((T_FINAL) reduce 250) ((T_FN) reduce 250) ((T_FOR) reduce 250) ((T_FOREACH) reduce 250) ((T_FUNCTION) reduce 250) ((T_FUNC_C) reduce 250) ((T_GLOBAL) reduce 250) ((T_GOTO) reduce 250) ((T_HALT_COMPILER) reduce 250) ((T_IF) reduce 250) ((T_INC) reduce 250) ((T_INCLUDE) reduce 250) ((T_INCLUDE_ONCE) reduce 250) ((T_INLINE_HTML) reduce 250) ((T_INTERFACE) reduce 250) ((T_INT_CAST) reduce 250) ((T_ISSET) reduce 250) ((T_LINE) reduce 250) ((T_LIST) reduce 250) ((T_LNUMBER) reduce 250) ((T_MATCH) reduce 250) ((T_METHOD_C) reduce 250) ((T_NAME_FULLY_QUALIFIED) reduce 250) ((T_NAME_QUALIFIED) reduce 250) ((T_NAME_RELATIVE) reduce 250) ((T_NEW) reduce 250) ((T_NS_C) reduce 250) ((T_OBJECT_CAST) reduce 250) ((T_PRINT) reduce 250) ((T_PROPERTY_C) reduce 250) ((T_READONLY) reduce 250) ((T_REQUIRE) reduce 250) ((T_REQUIRE_ONCE) reduce 250) ((T_RETURN) reduce 250) ((T_START_HEREDOC) reduce 250) ((T_STATIC) reduce 250) ((T_STRING) reduce 250) ((T_STRING_CAST) reduce 250) ((T_SWITCH) reduce 250) ((T_THROW) reduce 250) ((T_TRAIT) reduce 250) ((T_TRAIT_C) reduce 250) ((T_TRY) reduce 250) ((T_UNSET) reduce 250) ((T_UNSET_CAST) reduce 250) ((T_VARIABLE) reduce 250) ((T_VOID_CAST) reduce 250) ((T_WHILE) reduce 250) ((T_YIELD) reduce 250) ((T_YIELD_FROM) reduce 250) (("[") reduce 250) (("`") reduce 250) (("{") reduce 250) (("~") reduce 250)) 23342 ((("!") reduce 251) (("\"") reduce 251) (("$") reduce 251) (("(") reduce 251) (("+") reduce 251) (("-") reduce 251) ((";") reduce 251) (("@") reduce 251) ((T_ABSTRACT) reduce 251) ((T_ARRAY) reduce 251) ((T_ARRAY_CAST) reduce 251) ((T_ATTRIBUTE) reduce 251) ((T_BOOL_CAST) reduce 251) ((T_BREAK) reduce 251) ((T_CLASS) reduce 251) ((T_CLASS_C) reduce 251) ((T_CLONE) reduce 251) ((T_CONSTANT_ENCAPSED_STRING) reduce 251) ((T_CONTINUE) reduce 251) ((T_DEC) reduce 251) ((T_DECLARE) reduce 251) ((T_DIR) reduce 251) ((T_DNUMBER) reduce 251) ((T_DO) reduce 251) ((T_DOUBLE_CAST) reduce 251) ((T_ECHO) reduce 251) ((T_EMPTY) reduce 251) ((T_ENDDECLARE) reduce 251) ((T_ENUM) reduce 251) ((T_EVAL) reduce 251) ((T_EXIT) reduce 251) ((T_FILE) reduce 251) ((T_FINAL) reduce 251) ((T_FN) reduce 251) ((T_FOR) reduce 251) ((T_FOREACH) reduce 251) ((T_FUNCTION) reduce 251) ((T_FUNC_C) reduce 251) ((T_GLOBAL) reduce 251) ((T_GOTO) reduce 251) ((T_HALT_COMPILER) reduce 251) ((T_IF) reduce 251) ((T_INC) reduce 251) ((T_INCLUDE) reduce 251) ((T_INCLUDE_ONCE) reduce 251) ((T_INLINE_HTML) reduce 251) ((T_INTERFACE) reduce 251) ((T_INT_CAST) reduce 251) ((T_ISSET) reduce 251) ((T_LINE) reduce 251) ((T_LIST) reduce 251) ((T_LNUMBER) reduce 251) ((T_MATCH) reduce 251) ((T_METHOD_C) reduce 251) ((T_NAME_FULLY_QUALIFIED) reduce 251) ((T_NAME_QUALIFIED) reduce 251) ((T_NAME_RELATIVE) reduce 251) ((T_NEW) reduce 251) ((T_NS_C) reduce 251) ((T_OBJECT_CAST) reduce 251) ((T_PRINT) reduce 251) ((T_PROPERTY_C) reduce 251) ((T_READONLY) reduce 251) ((T_REQUIRE) reduce 251) ((T_REQUIRE_ONCE) reduce 251) ((T_RETURN) reduce 251) ((T_START_HEREDOC) reduce 251) ((T_STATIC) reduce 251) ((T_STRING) reduce 251) ((T_STRING_CAST) reduce 251) ((T_SWITCH) reduce 251) ((T_THROW) reduce 251) ((T_TRAIT) reduce 251) ((T_TRAIT_C) reduce 251) ((T_TRY) reduce 251) ((T_UNSET) reduce 251) ((T_UNSET_CAST) reduce 251) ((T_VARIABLE) reduce 251) ((T_VOID_CAST) reduce 251) ((T_WHILE) reduce 251) ((T_YIELD) reduce 251) ((T_YIELD_FROM) reduce 251) (("[") reduce 251) (("`") reduce 251) (("{") reduce 251) (("~") reduce 251)) 23347 ((("!") reduce 242) (("\"") reduce 242) (("$") reduce 242) (("(") reduce 242) (("+") reduce 242) (("-") reduce 242) ((";") reduce 242) (("@") reduce 242) ((T_ABSTRACT) reduce 242) ((T_ARRAY) reduce 242) ((T_ARRAY_CAST) reduce 242) ((T_ATTRIBUTE) reduce 242) ((T_BOOL_CAST) reduce 242) ((T_BREAK) reduce 242) ((T_CLASS) reduce 242) ((T_CLASS_C) reduce 242) ((T_CLONE) reduce 242) ((T_CONSTANT_ENCAPSED_STRING) reduce 242) ((T_CONTINUE) reduce 242) ((T_DEC) reduce 242) ((T_DECLARE) reduce 242) ((T_DIR) reduce 242) ((T_DNUMBER) reduce 242) ((T_DO) reduce 242) ((T_DOUBLE_CAST) reduce 242) ((T_ECHO) reduce 242) ((T_EMPTY) reduce 242) ((T_ENDDECLARE) reduce 242) ((T_ENUM) reduce 242) ((T_EVAL) reduce 242) ((T_EXIT) reduce 242) ((T_FILE) reduce 242) ((T_FINAL) reduce 242) ((T_FN) reduce 242) ((T_FOR) reduce 242) ((T_FOREACH) reduce 242) ((T_FUNCTION) reduce 242) ((T_FUNC_C) reduce 242) ((T_GLOBAL) reduce 242) ((T_GOTO) reduce 242) ((T_HALT_COMPILER) reduce 242) ((T_IF) reduce 242) ((T_INC) reduce 242) ((T_INCLUDE) reduce 242) ((T_INCLUDE_ONCE) reduce 242) ((T_INLINE_HTML) reduce 242) ((T_INTERFACE) reduce 242) ((T_INT_CAST) reduce 242) ((T_ISSET) reduce 242) ((T_LINE) reduce 242) ((T_LIST) reduce 242) ((T_LNUMBER) reduce 242) ((T_MATCH) reduce 242) ((T_METHOD_C) reduce 242) ((T_NAME_FULLY_QUALIFIED) reduce 242) ((T_NAME_QUALIFIED) reduce 242) ((T_NAME_RELATIVE) reduce 242) ((T_NEW) reduce 242) ((T_NS_C) reduce 242) ((T_OBJECT_CAST) reduce 242) ((T_PRINT) reduce 242) ((T_PROPERTY_C) reduce 242) ((T_READONLY) reduce 242) ((T_REQUIRE) reduce 242) ((T_REQUIRE_ONCE) reduce 242) ((T_RETURN) reduce 242) ((T_START_HEREDOC) reduce 242) ((T_STATIC) reduce 242) ((T_STRING) reduce 242) ((T_STRING_CAST) reduce 242) ((T_SWITCH) reduce 242) ((T_THROW) reduce 242) ((T_TRAIT) reduce 242) ((T_TRAIT_C) reduce 242) ((T_TRY) reduce 242) ((T_UNSET) reduce 242) ((T_UNSET_CAST) reduce 242) ((T_VARIABLE) reduce 242) ((T_VOID_CAST) reduce 242) ((T_WHILE) reduce 242) ((T_YIELD) reduce 242) ((T_YIELD_FROM) reduce 242) (("[") reduce 242) (("`") reduce 242) (("{") reduce 242) (("~") reduce 242)) 23348 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_EMPTY) reduce 150) ((T_ENDDECLARE) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_VOID_CAST) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 23351 ((("!") reduce 243) (("\"") reduce 243) (("$") reduce 243) (("(") reduce 243) (("+") reduce 243) (("-") reduce 243) ((";") reduce 243) (("@") reduce 243) ((T_ABSTRACT) reduce 243) ((T_ARRAY) reduce 243) ((T_ARRAY_CAST) reduce 243) ((T_ATTRIBUTE) reduce 243) ((T_BOOL_CAST) reduce 243) ((T_BREAK) reduce 243) ((T_CLASS) reduce 243) ((T_CLASS_C) reduce 243) ((T_CLONE) reduce 243) ((T_CONSTANT_ENCAPSED_STRING) reduce 243) ((T_CONTINUE) reduce 243) ((T_DEC) reduce 243) ((T_DECLARE) reduce 243) ((T_DIR) reduce 243) ((T_DNUMBER) reduce 243) ((T_DO) reduce 243) ((T_DOUBLE_CAST) reduce 243) ((T_ECHO) reduce 243) ((T_EMPTY) reduce 243) ((T_ENDDECLARE) reduce 243) ((T_ENUM) reduce 243) ((T_EVAL) reduce 243) ((T_EXIT) reduce 243) ((T_FILE) reduce 243) ((T_FINAL) reduce 243) ((T_FN) reduce 243) ((T_FOR) reduce 243) ((T_FOREACH) reduce 243) ((T_FUNCTION) reduce 243) ((T_FUNC_C) reduce 243) ((T_GLOBAL) reduce 243) ((T_GOTO) reduce 243) ((T_HALT_COMPILER) reduce 243) ((T_IF) reduce 243) ((T_INC) reduce 243) ((T_INCLUDE) reduce 243) ((T_INCLUDE_ONCE) reduce 243) ((T_INLINE_HTML) reduce 243) ((T_INTERFACE) reduce 243) ((T_INT_CAST) reduce 243) ((T_ISSET) reduce 243) ((T_LINE) reduce 243) ((T_LIST) reduce 243) ((T_LNUMBER) reduce 243) ((T_MATCH) reduce 243) ((T_METHOD_C) reduce 243) ((T_NAME_FULLY_QUALIFIED) reduce 243) ((T_NAME_QUALIFIED) reduce 243) ((T_NAME_RELATIVE) reduce 243) ((T_NEW) reduce 243) ((T_NS_C) reduce 243) ((T_OBJECT_CAST) reduce 243) ((T_PRINT) reduce 243) ((T_PROPERTY_C) reduce 243) ((T_READONLY) reduce 243) ((T_REQUIRE) reduce 243) ((T_REQUIRE_ONCE) reduce 243) ((T_RETURN) reduce 243) ((T_START_HEREDOC) reduce 243) ((T_STATIC) reduce 243) ((T_STRING) reduce 243) ((T_STRING_CAST) reduce 243) ((T_SWITCH) reduce 243) ((T_THROW) reduce 243) ((T_TRAIT) reduce 243) ((T_TRAIT_C) reduce 243) ((T_TRY) reduce 243) ((T_UNSET) reduce 243) ((T_UNSET_CAST) reduce 243) ((T_VARIABLE) reduce 243) ((T_VOID_CAST) reduce 243) ((T_WHILE) reduce 243) ((T_YIELD) reduce 243) ((T_YIELD_FROM) reduce 243) (("[") reduce 243) (("`") reduce 243) (("{") reduce 243) (("~") reduce 243)) 23353 ((("!") reduce 170) (("\"") reduce 170) (("$") reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_EMPTY) reduce 170) ((T_ENDDECLARE) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_VARIABLE) reduce 170) ((T_VOID_CAST) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("~") reduce 170)) 23358 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_EMPTY) reduce 162) ((T_ENDDECLARE) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_VOID_CAST) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 23361 ((("!") reduce 171) (("\"") reduce 171) (("$") reduce 171) (("(") reduce 171) (("+") reduce 171) (("-") reduce 171) ((";") reduce 171) (("@") reduce 171) ((T_ABSTRACT) reduce 171) ((T_ARRAY) reduce 171) ((T_ARRAY_CAST) reduce 171) ((T_ATTRIBUTE) reduce 171) ((T_BOOL_CAST) reduce 171) ((T_BREAK) reduce 171) ((T_CATCH) reduce 171) ((T_CLASS) reduce 171) ((T_CLASS_C) reduce 171) ((T_CLONE) reduce 171) ((T_CONSTANT_ENCAPSED_STRING) reduce 171) ((T_CONTINUE) reduce 171) ((T_DEC) reduce 171) ((T_DECLARE) reduce 171) ((T_DIR) reduce 171) ((T_DNUMBER) reduce 171) ((T_DO) reduce 171) ((T_DOUBLE_CAST) reduce 171) ((T_ECHO) reduce 171) ((T_EMPTY) reduce 171) ((T_ENDDECLARE) reduce 171) ((T_ENUM) reduce 171) ((T_EVAL) reduce 171) ((T_EXIT) reduce 171) ((T_FILE) reduce 171) ((T_FINAL) reduce 171) ((T_FINALLY) reduce 171) ((T_FN) reduce 171) ((T_FOR) reduce 171) ((T_FOREACH) reduce 171) ((T_FUNCTION) reduce 171) ((T_FUNC_C) reduce 171) ((T_GLOBAL) reduce 171) ((T_GOTO) reduce 171) ((T_HALT_COMPILER) reduce 171) ((T_IF) reduce 171) ((T_INC) reduce 171) ((T_INCLUDE) reduce 171) ((T_INCLUDE_ONCE) reduce 171) ((T_INLINE_HTML) reduce 171) ((T_INTERFACE) reduce 171) ((T_INT_CAST) reduce 171) ((T_ISSET) reduce 171) ((T_LINE) reduce 171) ((T_LIST) reduce 171) ((T_LNUMBER) reduce 171) ((T_MATCH) reduce 171) ((T_METHOD_C) reduce 171) ((T_NAME_FULLY_QUALIFIED) reduce 171) ((T_NAME_QUALIFIED) reduce 171) ((T_NAME_RELATIVE) reduce 171) ((T_NEW) reduce 171) ((T_NS_C) reduce 171) ((T_OBJECT_CAST) reduce 171) ((T_PRINT) reduce 171) ((T_PROPERTY_C) reduce 171) ((T_READONLY) reduce 171) ((T_REQUIRE) reduce 171) ((T_REQUIRE_ONCE) reduce 171) ((T_RETURN) reduce 171) ((T_START_HEREDOC) reduce 171) ((T_STATIC) reduce 171) ((T_STRING) reduce 171) ((T_STRING_CAST) reduce 171) ((T_SWITCH) reduce 171) ((T_THROW) reduce 171) ((T_TRAIT) reduce 171) ((T_TRAIT_C) reduce 171) ((T_TRY) reduce 171) ((T_UNSET) reduce 171) ((T_UNSET_CAST) reduce 171) ((T_VARIABLE) reduce 171) ((T_VOID_CAST) reduce 171) ((T_WHILE) reduce 171) ((T_YIELD) reduce 171) ((T_YIELD_FROM) reduce 171) (("[") reduce 171) (("`") reduce 171) (("{") reduce 171) (("~") reduce 171)) 23362 ((("!") reduce 177) (("\"") reduce 177) (("$") reduce 177) (("(") reduce 177) (("+") reduce 177) (("-") reduce 177) ((";") reduce 177) (("@") reduce 177) ((T_ABSTRACT) reduce 177) ((T_ARRAY) reduce 177) ((T_ARRAY_CAST) reduce 177) ((T_ATTRIBUTE) reduce 177) ((T_BOOL_CAST) reduce 177) ((T_BREAK) reduce 177) ((T_CATCH) shift) ((T_CLASS) reduce 177) ((T_CLASS_C) reduce 177) ((T_CLONE) reduce 177) ((T_CONSTANT_ENCAPSED_STRING) reduce 177) ((T_CONTINUE) reduce 177) ((T_DEC) reduce 177) ((T_DECLARE) reduce 177) ((T_DIR) reduce 177) ((T_DNUMBER) reduce 177) ((T_DO) reduce 177) ((T_DOUBLE_CAST) reduce 177) ((T_ECHO) reduce 177) ((T_EMPTY) reduce 177) ((T_ENDDECLARE) reduce 177) ((T_ENUM) reduce 177) ((T_EVAL) reduce 177) ((T_EXIT) reduce 177) ((T_FILE) reduce 177) ((T_FINAL) reduce 177) ((T_FINALLY) shift) ((T_FN) reduce 177) ((T_FOR) reduce 177) ((T_FOREACH) reduce 177) ((T_FUNCTION) reduce 177) ((T_FUNC_C) reduce 177) ((T_GLOBAL) reduce 177) ((T_GOTO) reduce 177) ((T_HALT_COMPILER) reduce 177) ((T_IF) reduce 177) ((T_INC) reduce 177) ((T_INCLUDE) reduce 177) ((T_INCLUDE_ONCE) reduce 177) ((T_INLINE_HTML) reduce 177) ((T_INTERFACE) reduce 177) ((T_INT_CAST) reduce 177) ((T_ISSET) reduce 177) ((T_LINE) reduce 177) ((T_LIST) reduce 177) ((T_LNUMBER) reduce 177) ((T_MATCH) reduce 177) ((T_METHOD_C) reduce 177) ((T_NAME_FULLY_QUALIFIED) reduce 177) ((T_NAME_QUALIFIED) reduce 177) ((T_NAME_RELATIVE) reduce 177) ((T_NEW) reduce 177) ((T_NS_C) reduce 177) ((T_OBJECT_CAST) reduce 177) ((T_PRINT) reduce 177) ((T_PROPERTY_C) reduce 177) ((T_READONLY) reduce 177) ((T_REQUIRE) reduce 177) ((T_REQUIRE_ONCE) reduce 177) ((T_RETURN) reduce 177) ((T_START_HEREDOC) reduce 177) ((T_STATIC) reduce 177) ((T_STRING) reduce 177) ((T_STRING_CAST) reduce 177) ((T_SWITCH) reduce 177) ((T_THROW) reduce 177) ((T_TRAIT) reduce 177) ((T_TRAIT_C) reduce 177) ((T_TRY) reduce 177) ((T_UNSET) reduce 177) ((T_UNSET_CAST) reduce 177) ((T_VARIABLE) reduce 177) ((T_VOID_CAST) reduce 177) ((T_WHILE) reduce 177) ((T_YIELD) reduce 177) ((T_YIELD_FROM) reduce 177) (("[") reduce 177) (("`") reduce 177) (("{") reduce 177) (("~") reduce 177)) 23365 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_EMPTY) reduce 167) ((T_ENDDECLARE) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_VOID_CAST) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 23368 ((("!") reduce 178) (("\"") reduce 178) (("$") reduce 178) (("(") reduce 178) (("+") reduce 178) (("-") reduce 178) ((";") reduce 178) (("@") reduce 178) ((T_ABSTRACT) reduce 178) ((T_ARRAY) reduce 178) ((T_ARRAY_CAST) reduce 178) ((T_ATTRIBUTE) reduce 178) ((T_BOOL_CAST) reduce 178) ((T_BREAK) reduce 178) ((T_CLASS) reduce 178) ((T_CLASS_C) reduce 178) ((T_CLONE) reduce 178) ((T_CONSTANT_ENCAPSED_STRING) reduce 178) ((T_CONTINUE) reduce 178) ((T_DEC) reduce 178) ((T_DECLARE) reduce 178) ((T_DIR) reduce 178) ((T_DNUMBER) reduce 178) ((T_DO) reduce 178) ((T_DOUBLE_CAST) reduce 178) ((T_ECHO) reduce 178) ((T_EMPTY) reduce 178) ((T_ENDDECLARE) reduce 178) ((T_ENUM) reduce 178) ((T_EVAL) reduce 178) ((T_EXIT) reduce 178) ((T_FILE) reduce 178) ((T_FINAL) reduce 178) ((T_FN) reduce 178) ((T_FOR) reduce 178) ((T_FOREACH) reduce 178) ((T_FUNCTION) reduce 178) ((T_FUNC_C) reduce 178) ((T_GLOBAL) reduce 178) ((T_GOTO) reduce 178) ((T_HALT_COMPILER) reduce 178) ((T_IF) reduce 178) ((T_INC) reduce 178) ((T_INCLUDE) reduce 178) ((T_INCLUDE_ONCE) reduce 178) ((T_INLINE_HTML) reduce 178) ((T_INTERFACE) reduce 178) ((T_INT_CAST) reduce 178) ((T_ISSET) reduce 178) ((T_LINE) reduce 178) ((T_LIST) reduce 178) ((T_LNUMBER) reduce 178) ((T_MATCH) reduce 178) ((T_METHOD_C) reduce 178) ((T_NAME_FULLY_QUALIFIED) reduce 178) ((T_NAME_QUALIFIED) reduce 178) ((T_NAME_RELATIVE) reduce 178) ((T_NEW) reduce 178) ((T_NS_C) reduce 178) ((T_OBJECT_CAST) reduce 178) ((T_PRINT) reduce 178) ((T_PROPERTY_C) reduce 178) ((T_READONLY) reduce 178) ((T_REQUIRE) reduce 178) ((T_REQUIRE_ONCE) reduce 178) ((T_RETURN) reduce 178) ((T_START_HEREDOC) reduce 178) ((T_STATIC) reduce 178) ((T_STRING) reduce 178) ((T_STRING_CAST) reduce 178) ((T_SWITCH) reduce 178) ((T_THROW) reduce 178) ((T_TRAIT) reduce 178) ((T_TRAIT_C) reduce 178) ((T_TRY) reduce 178) ((T_UNSET) reduce 178) ((T_UNSET_CAST) reduce 178) ((T_VARIABLE) reduce 178) ((T_VOID_CAST) reduce 178) ((T_WHILE) reduce 178) ((T_YIELD) reduce 178) ((T_YIELD_FROM) reduce 178) (("[") reduce 178) (("`") reduce 178) (("{") reduce 178) (("~") reduce 178)) 23375 ((("!") reduce 172) (("\"") reduce 172) (("$") reduce 172) (("(") reduce 172) (("+") reduce 172) (("-") reduce 172) ((";") reduce 172) (("@") reduce 172) ((T_ABSTRACT) reduce 172) ((T_ARRAY) reduce 172) ((T_ARRAY_CAST) reduce 172) ((T_ATTRIBUTE) reduce 172) ((T_BOOL_CAST) reduce 172) ((T_BREAK) reduce 172) ((T_CATCH) reduce 172) ((T_CLASS) reduce 172) ((T_CLASS_C) reduce 172) ((T_CLONE) reduce 172) ((T_CONSTANT_ENCAPSED_STRING) reduce 172) ((T_CONTINUE) reduce 172) ((T_DEC) reduce 172) ((T_DECLARE) reduce 172) ((T_DIR) reduce 172) ((T_DNUMBER) reduce 172) ((T_DO) reduce 172) ((T_DOUBLE_CAST) reduce 172) ((T_ECHO) reduce 172) ((T_EMPTY) reduce 172) ((T_ENDDECLARE) reduce 172) ((T_ENUM) reduce 172) ((T_EVAL) reduce 172) ((T_EXIT) reduce 172) ((T_FILE) reduce 172) ((T_FINAL) reduce 172) ((T_FINALLY) reduce 172) ((T_FN) reduce 172) ((T_FOR) reduce 172) ((T_FOREACH) reduce 172) ((T_FUNCTION) reduce 172) ((T_FUNC_C) reduce 172) ((T_GLOBAL) reduce 172) ((T_GOTO) reduce 172) ((T_HALT_COMPILER) reduce 172) ((T_IF) reduce 172) ((T_INC) reduce 172) ((T_INCLUDE) reduce 172) ((T_INCLUDE_ONCE) reduce 172) ((T_INLINE_HTML) reduce 172) ((T_INTERFACE) reduce 172) ((T_INT_CAST) reduce 172) ((T_ISSET) reduce 172) ((T_LINE) reduce 172) ((T_LIST) reduce 172) ((T_LNUMBER) reduce 172) ((T_MATCH) reduce 172) ((T_METHOD_C) reduce 172) ((T_NAME_FULLY_QUALIFIED) reduce 172) ((T_NAME_QUALIFIED) reduce 172) ((T_NAME_RELATIVE) reduce 172) ((T_NEW) reduce 172) ((T_NS_C) reduce 172) ((T_OBJECT_CAST) reduce 172) ((T_PRINT) reduce 172) ((T_PROPERTY_C) reduce 172) ((T_READONLY) reduce 172) ((T_REQUIRE) reduce 172) ((T_REQUIRE_ONCE) reduce 172) ((T_RETURN) reduce 172) ((T_START_HEREDOC) reduce 172) ((T_STATIC) reduce 172) ((T_STRING) reduce 172) ((T_STRING_CAST) reduce 172) ((T_SWITCH) reduce 172) ((T_THROW) reduce 172) ((T_TRAIT) reduce 172) ((T_TRAIT_C) reduce 172) ((T_TRY) reduce 172) ((T_UNSET) reduce 172) ((T_UNSET_CAST) reduce 172) ((T_VARIABLE) reduce 172) ((T_VOID_CAST) reduce 172) ((T_WHILE) reduce 172) ((T_YIELD) reduce 172) ((T_YIELD_FROM) reduce 172) (("[") reduce 172) (("`") reduce 172) (("{") reduce 172) (("~") reduce 172)) 23380 ((("!") reduce 200) (("\"") reduce 200) (("$") reduce 200) (("(") reduce 200) (("+") reduce 200) (("-") reduce 200) ((";") reduce 200) (("@") reduce 200) ((T_ABSTRACT) reduce 200) ((T_ARRAY) reduce 200) ((T_ARRAY_CAST) reduce 200) ((T_ATTRIBUTE) reduce 200) ((T_BOOL_CAST) reduce 200) ((T_BREAK) reduce 200) ((T_CLASS) reduce 200) ((T_CLASS_C) reduce 200) ((T_CLONE) reduce 200) ((T_CONSTANT_ENCAPSED_STRING) reduce 200) ((T_CONTINUE) reduce 200) ((T_DEC) reduce 200) ((T_DECLARE) reduce 200) ((T_DIR) reduce 200) ((T_DNUMBER) reduce 200) ((T_DO) reduce 200) ((T_DOUBLE_CAST) reduce 200) ((T_ECHO) reduce 200) ((T_EMPTY) reduce 200) ((T_ENDDECLARE) reduce 200) ((T_ENUM) reduce 200) ((T_EVAL) reduce 200) ((T_EXIT) reduce 200) ((T_FILE) reduce 200) ((T_FINAL) reduce 200) ((T_FN) reduce 200) ((T_FOR) reduce 200) ((T_FOREACH) reduce 200) ((T_FUNCTION) reduce 200) ((T_FUNC_C) reduce 200) ((T_GLOBAL) reduce 200) ((T_GOTO) reduce 200) ((T_HALT_COMPILER) reduce 200) ((T_IF) reduce 200) ((T_INC) reduce 200) ((T_INCLUDE) reduce 200) ((T_INCLUDE_ONCE) reduce 200) ((T_INLINE_HTML) reduce 200) ((T_INTERFACE) reduce 200) ((T_INT_CAST) reduce 200) ((T_ISSET) reduce 200) ((T_LINE) reduce 200) ((T_LIST) reduce 200) ((T_LNUMBER) reduce 200) ((T_MATCH) reduce 200) ((T_METHOD_C) reduce 200) ((T_NAME_FULLY_QUALIFIED) reduce 200) ((T_NAME_QUALIFIED) reduce 200) ((T_NAME_RELATIVE) reduce 200) ((T_NEW) reduce 200) ((T_NS_C) reduce 200) ((T_OBJECT_CAST) reduce 200) ((T_PRINT) reduce 200) ((T_PROPERTY_C) reduce 200) ((T_READONLY) reduce 200) ((T_REQUIRE) reduce 200) ((T_REQUIRE_ONCE) reduce 200) ((T_RETURN) reduce 200) ((T_START_HEREDOC) reduce 200) ((T_STATIC) reduce 200) ((T_STRING) reduce 200) ((T_STRING_CAST) reduce 200) ((T_SWITCH) reduce 200) ((T_THROW) reduce 200) ((T_TRAIT) reduce 200) ((T_TRAIT_C) reduce 200) ((T_TRY) reduce 200) ((T_UNSET) reduce 200) ((T_UNSET_CAST) reduce 200) ((T_VARIABLE) reduce 200) ((T_VOID_CAST) reduce 200) ((T_WHILE) reduce 200) ((T_YIELD) reduce 200) ((T_YIELD_FROM) reduce 200) (("[") reduce 200) (("`") reduce 200) (("{") reduce 200) (("~") reduce 200)) 23385 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_EMPTY) reduce 153) ((T_ENDDECLARE) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_VOID_CAST) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 23389 ((("!") reduce 224) (("\"") reduce 224) (("$") reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_EMPTY) reduce 224) ((T_ENDDECLARE) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_VARIABLE) reduce 224) ((T_VOID_CAST) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("~") reduce 224)) 23391 ((("!") reduce 225) (("\"") reduce 225) (("$") reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_EMPTY) reduce 225) ((T_ENDDECLARE) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_VARIABLE) reduce 225) ((T_VOID_CAST) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("~") reduce 225)) 23395 ((("!") reduce 226) (("\"") reduce 226) (("$") reduce 226) (("(") reduce 226) (("+") reduce 226) (("-") reduce 226) ((";") reduce 226) (("@") reduce 226) ((T_ABSTRACT) reduce 226) ((T_ARRAY) reduce 226) ((T_ARRAY_CAST) reduce 226) ((T_ATTRIBUTE) reduce 226) ((T_BOOL_CAST) reduce 226) ((T_BREAK) reduce 226) ((T_CLASS) reduce 226) ((T_CLASS_C) reduce 226) ((T_CLONE) reduce 226) ((T_CONSTANT_ENCAPSED_STRING) reduce 226) ((T_CONTINUE) reduce 226) ((T_DEC) reduce 226) ((T_DECLARE) reduce 226) ((T_DIR) reduce 226) ((T_DNUMBER) reduce 226) ((T_DO) reduce 226) ((T_DOUBLE_CAST) reduce 226) ((T_ECHO) reduce 226) ((T_EMPTY) reduce 226) ((T_ENDDECLARE) reduce 226) ((T_ENUM) reduce 226) ((T_EVAL) reduce 226) ((T_EXIT) reduce 226) ((T_FILE) reduce 226) ((T_FINAL) reduce 226) ((T_FN) reduce 226) ((T_FOR) reduce 226) ((T_FOREACH) reduce 226) ((T_FUNCTION) reduce 226) ((T_FUNC_C) reduce 226) ((T_GLOBAL) reduce 226) ((T_GOTO) reduce 226) ((T_HALT_COMPILER) reduce 226) ((T_IF) reduce 226) ((T_INC) reduce 226) ((T_INCLUDE) reduce 226) ((T_INCLUDE_ONCE) reduce 226) ((T_INLINE_HTML) reduce 226) ((T_INTERFACE) reduce 226) ((T_INT_CAST) reduce 226) ((T_ISSET) reduce 226) ((T_LINE) reduce 226) ((T_LIST) reduce 226) ((T_LNUMBER) reduce 226) ((T_MATCH) reduce 226) ((T_METHOD_C) reduce 226) ((T_NAME_FULLY_QUALIFIED) reduce 226) ((T_NAME_QUALIFIED) reduce 226) ((T_NAME_RELATIVE) reduce 226) ((T_NEW) reduce 226) ((T_NS_C) reduce 226) ((T_OBJECT_CAST) reduce 226) ((T_PRINT) reduce 226) ((T_PROPERTY_C) reduce 226) ((T_READONLY) reduce 226) ((T_REQUIRE) reduce 226) ((T_REQUIRE_ONCE) reduce 226) ((T_RETURN) reduce 226) ((T_START_HEREDOC) reduce 226) ((T_STATIC) reduce 226) ((T_STRING) reduce 226) ((T_STRING_CAST) reduce 226) ((T_SWITCH) reduce 226) ((T_THROW) reduce 226) ((T_TRAIT) reduce 226) ((T_TRAIT_C) reduce 226) ((T_TRY) reduce 226) ((T_UNSET) reduce 226) ((T_UNSET_CAST) reduce 226) ((T_VARIABLE) reduce 226) ((T_VOID_CAST) reduce 226) ((T_WHILE) reduce 226) ((T_YIELD) reduce 226) ((T_YIELD_FROM) reduce 226) (("[") reduce 226) (("`") reduce 226) (("{") reduce 226) (("~") reduce 226)) 23398 ((("!") reduce 227) (("\"") reduce 227) (("$") reduce 227) (("(") reduce 227) (("+") reduce 227) (("-") reduce 227) ((";") reduce 227) (("@") reduce 227) ((T_ABSTRACT) reduce 227) ((T_ARRAY) reduce 227) ((T_ARRAY_CAST) reduce 227) ((T_ATTRIBUTE) reduce 227) ((T_BOOL_CAST) reduce 227) ((T_BREAK) reduce 227) ((T_CLASS) reduce 227) ((T_CLASS_C) reduce 227) ((T_CLONE) reduce 227) ((T_CONSTANT_ENCAPSED_STRING) reduce 227) ((T_CONTINUE) reduce 227) ((T_DEC) reduce 227) ((T_DECLARE) reduce 227) ((T_DIR) reduce 227) ((T_DNUMBER) reduce 227) ((T_DO) reduce 227) ((T_DOUBLE_CAST) reduce 227) ((T_ECHO) reduce 227) ((T_EMPTY) reduce 227) ((T_ENDDECLARE) reduce 227) ((T_ENUM) reduce 227) ((T_EVAL) reduce 227) ((T_EXIT) reduce 227) ((T_FILE) reduce 227) ((T_FINAL) reduce 227) ((T_FN) reduce 227) ((T_FOR) reduce 227) ((T_FOREACH) reduce 227) ((T_FUNCTION) reduce 227) ((T_FUNC_C) reduce 227) ((T_GLOBAL) reduce 227) ((T_GOTO) reduce 227) ((T_HALT_COMPILER) reduce 227) ((T_IF) reduce 227) ((T_INC) reduce 227) ((T_INCLUDE) reduce 227) ((T_INCLUDE_ONCE) reduce 227) ((T_INLINE_HTML) reduce 227) ((T_INTERFACE) reduce 227) ((T_INT_CAST) reduce 227) ((T_ISSET) reduce 227) ((T_LINE) reduce 227) ((T_LIST) reduce 227) ((T_LNUMBER) reduce 227) ((T_MATCH) reduce 227) ((T_METHOD_C) reduce 227) ((T_NAME_FULLY_QUALIFIED) reduce 227) ((T_NAME_QUALIFIED) reduce 227) ((T_NAME_RELATIVE) reduce 227) ((T_NEW) reduce 227) ((T_NS_C) reduce 227) ((T_OBJECT_CAST) reduce 227) ((T_PRINT) reduce 227) ((T_PROPERTY_C) reduce 227) ((T_READONLY) reduce 227) ((T_REQUIRE) reduce 227) ((T_REQUIRE_ONCE) reduce 227) ((T_RETURN) reduce 227) ((T_START_HEREDOC) reduce 227) ((T_STATIC) reduce 227) ((T_STRING) reduce 227) ((T_STRING_CAST) reduce 227) ((T_SWITCH) reduce 227) ((T_THROW) reduce 227) ((T_TRAIT) reduce 227) ((T_TRAIT_C) reduce 227) ((T_TRY) reduce 227) ((T_UNSET) reduce 227) ((T_UNSET_CAST) reduce 227) ((T_VARIABLE) reduce 227) ((T_VOID_CAST) reduce 227) ((T_WHILE) reduce 227) ((T_YIELD) reduce 227) ((T_YIELD_FROM) reduce 227) (("[") reduce 227) (("`") reduce 227) (("{") reduce 227) (("~") reduce 227)) 23399 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_EMPTY) reduce 169) ((T_ENDDECLARE) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_VOID_CAST) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 23401 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_EMPTY) reduce 158) ((T_ENDDECLARE) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_VOID_CAST) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 23403 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_EMPTY) reduce 156) ((T_ENDDECLARE) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_VOID_CAST) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 23404 (((T_EXTENDS) shift) (("{") reduce 210)) 23410 ((("!") reduce 201) (("\"") reduce 201) (("$") reduce 201) (("(") reduce 201) (("+") reduce 201) (("-") reduce 201) ((";") reduce 201) (("@") reduce 201) ((T_ABSTRACT) reduce 201) ((T_ARRAY) reduce 201) ((T_ARRAY_CAST) reduce 201) ((T_ATTRIBUTE) reduce 201) ((T_BOOL_CAST) reduce 201) ((T_BREAK) reduce 201) ((T_CLASS) reduce 201) ((T_CLASS_C) reduce 201) ((T_CLONE) reduce 201) ((T_CONSTANT_ENCAPSED_STRING) reduce 201) ((T_CONTINUE) reduce 201) ((T_DEC) reduce 201) ((T_DECLARE) reduce 201) ((T_DIR) reduce 201) ((T_DNUMBER) reduce 201) ((T_DO) reduce 201) ((T_DOUBLE_CAST) reduce 201) ((T_ECHO) reduce 201) ((T_EMPTY) reduce 201) ((T_ENDDECLARE) reduce 201) ((T_ENUM) reduce 201) ((T_EVAL) reduce 201) ((T_EXIT) reduce 201) ((T_FILE) reduce 201) ((T_FINAL) reduce 201) ((T_FN) reduce 201) ((T_FOR) reduce 201) ((T_FOREACH) reduce 201) ((T_FUNCTION) reduce 201) ((T_FUNC_C) reduce 201) ((T_GLOBAL) reduce 201) ((T_GOTO) reduce 201) ((T_HALT_COMPILER) reduce 201) ((T_IF) reduce 201) ((T_INC) reduce 201) ((T_INCLUDE) reduce 201) ((T_INCLUDE_ONCE) reduce 201) ((T_INLINE_HTML) reduce 201) ((T_INTERFACE) reduce 201) ((T_INT_CAST) reduce 201) ((T_ISSET) reduce 201) ((T_LINE) reduce 201) ((T_LIST) reduce 201) ((T_LNUMBER) reduce 201) ((T_MATCH) reduce 201) ((T_METHOD_C) reduce 201) ((T_NAME_FULLY_QUALIFIED) reduce 201) ((T_NAME_QUALIFIED) reduce 201) ((T_NAME_RELATIVE) reduce 201) ((T_NEW) reduce 201) ((T_NS_C) reduce 201) ((T_OBJECT_CAST) reduce 201) ((T_PRINT) reduce 201) ((T_PROPERTY_C) reduce 201) ((T_READONLY) reduce 201) ((T_REQUIRE) reduce 201) ((T_REQUIRE_ONCE) reduce 201) ((T_RETURN) reduce 201) ((T_START_HEREDOC) reduce 201) ((T_STATIC) reduce 201) ((T_STRING) reduce 201) ((T_STRING_CAST) reduce 201) ((T_SWITCH) reduce 201) ((T_THROW) reduce 201) ((T_TRAIT) reduce 201) ((T_TRAIT_C) reduce 201) ((T_TRY) reduce 201) ((T_UNSET) reduce 201) ((T_UNSET_CAST) reduce 201) ((T_VARIABLE) reduce 201) ((T_VOID_CAST) reduce 201) ((T_WHILE) reduce 201) ((T_YIELD) reduce 201) ((T_YIELD_FROM) reduce 201) (("[") reduce 201) (("`") reduce 201) (("{") reduce 201) (("~") reduce 201)) 23411 (((",") shift) (("{") reduce 211)) 23416 ((("!") reduce 244) (("\"") reduce 244) (("$") reduce 244) (("(") reduce 244) (("+") reduce 244) (("-") reduce 244) ((";") reduce 244) (("@") reduce 244) ((T_ABSTRACT) reduce 244) ((T_ARRAY) reduce 244) ((T_ARRAY_CAST) reduce 244) ((T_ATTRIBUTE) reduce 244) ((T_BOOL_CAST) reduce 244) ((T_BREAK) reduce 244) ((T_CLASS) reduce 244) ((T_CLASS_C) reduce 244) ((T_CLONE) reduce 244) ((T_CONSTANT_ENCAPSED_STRING) reduce 244) ((T_CONTINUE) reduce 244) ((T_DEC) reduce 244) ((T_DECLARE) reduce 244) ((T_DIR) reduce 244) ((T_DNUMBER) reduce 244) ((T_DO) reduce 244) ((T_DOUBLE_CAST) reduce 244) ((T_ECHO) reduce 244) ((T_ELSE) reduce 244) ((T_ELSEIF) reduce 244) ((T_EMPTY) reduce 244) ((T_ENDDECLARE) reduce 244) ((T_ENUM) reduce 244) ((T_EVAL) reduce 244) ((T_EXIT) reduce 244) ((T_FILE) reduce 244) ((T_FINAL) reduce 244) ((T_FN) reduce 244) ((T_FOR) reduce 244) ((T_FOREACH) reduce 244) ((T_FUNCTION) reduce 244) ((T_FUNC_C) reduce 244) ((T_GLOBAL) reduce 244) ((T_GOTO) reduce 244) ((T_HALT_COMPILER) reduce 244) ((T_IF) reduce 244) ((T_INC) reduce 244) ((T_INCLUDE) reduce 244) ((T_INCLUDE_ONCE) reduce 244) ((T_INLINE_HTML) reduce 244) ((T_INTERFACE) reduce 244) ((T_INT_CAST) reduce 244) ((T_ISSET) reduce 244) ((T_LINE) reduce 244) ((T_LIST) reduce 244) ((T_LNUMBER) reduce 244) ((T_MATCH) reduce 244) ((T_METHOD_C) reduce 244) ((T_NAME_FULLY_QUALIFIED) reduce 244) ((T_NAME_QUALIFIED) reduce 244) ((T_NAME_RELATIVE) reduce 244) ((T_NEW) reduce 244) ((T_NS_C) reduce 244) ((T_OBJECT_CAST) reduce 244) ((T_PRINT) reduce 244) ((T_PROPERTY_C) reduce 244) ((T_READONLY) reduce 244) ((T_REQUIRE) reduce 244) ((T_REQUIRE_ONCE) reduce 244) ((T_RETURN) reduce 244) ((T_START_HEREDOC) reduce 244) ((T_STATIC) reduce 244) ((T_STRING) reduce 244) ((T_STRING_CAST) reduce 244) ((T_SWITCH) reduce 244) ((T_THROW) reduce 244) ((T_TRAIT) reduce 244) ((T_TRAIT_C) reduce 244) ((T_TRY) reduce 244) ((T_UNSET) reduce 244) ((T_UNSET_CAST) reduce 244) ((T_VARIABLE) reduce 244) ((T_VOID_CAST) reduce 244) ((T_WHILE) reduce 244) ((T_YIELD) reduce 244) ((T_YIELD_FROM) reduce 244) (("[") reduce 244) (("`") reduce 244) (("{") reduce 244) (("~") reduce 244)) 23417 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_ELSE) reduce 248) ((T_ELSEIF) reduce 248) ((T_EMPTY) shift) ((T_ENDIF) reduce 248) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_VOID_CAST) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("~") shift)) 23420 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_EMPTY) reduce 146) ((T_ENDDECLARE) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_PROPERTY_C) reduce 146) ((T_READONLY) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_VOID_CAST) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("~") reduce 146)) 23422 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_EMPTY) reduce 168) ((T_ENDDECLARE) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_VOID_CAST) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 23424 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_EMPTY) reduce 157) ((T_ENDDECLARE) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_VOID_CAST) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 23434 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_EMPTY) reduce 164) ((T_ENDDECLARE) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_VOID_CAST) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 23435 ((("!") reduce 220) (("\"") reduce 220) (("$") reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_EMPTY) reduce 220) ((T_ENDDECLARE) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_VARIABLE) reduce 220) ((T_VOID_CAST) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("~") reduce 220)) 23438 ((("!") reduce 221) (("\"") reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_EMPTY) reduce 221) ((T_ENDDECLARE) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_VARIABLE) reduce 221) ((T_VOID_CAST) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("~") reduce 221)) 23439 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_EMPTY) reduce 163) ((T_ENDDECLARE) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_VOID_CAST) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 23448 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_EMPTY) reduce 152) ((T_ENDDECLARE) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_VOID_CAST) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 23449 ((("!") reduce 218) (("\"") reduce 218) (("$") reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_EMPTY) reduce 218) ((T_ENDDECLARE) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_VARIABLE) reduce 218) ((T_VOID_CAST) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("~") reduce 218)) 23452 ((("!") reduce 219) (("\"") reduce 219) (("$") reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_EMPTY) reduce 219) ((T_ENDDECLARE) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_VARIABLE) reduce 219) ((T_VOID_CAST) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("~") reduce 219)) 23453 (((":") shift) ((T_IMPLEMENTS) reduce 203) (("{") reduce 203)) 23460 ((("!") reduce 202) (("\"") reduce 202) (("$") reduce 202) (("(") reduce 202) (("+") reduce 202) (("-") reduce 202) ((";") reduce 202) (("@") reduce 202) ((T_ABSTRACT) reduce 202) ((T_ARRAY) reduce 202) ((T_ARRAY_CAST) reduce 202) ((T_ATTRIBUTE) reduce 202) ((T_BOOL_CAST) reduce 202) ((T_BREAK) reduce 202) ((T_CLASS) reduce 202) ((T_CLASS_C) reduce 202) ((T_CLONE) reduce 202) ((T_CONSTANT_ENCAPSED_STRING) reduce 202) ((T_CONTINUE) reduce 202) ((T_DEC) reduce 202) ((T_DECLARE) reduce 202) ((T_DIR) reduce 202) ((T_DNUMBER) reduce 202) ((T_DO) reduce 202) ((T_DOUBLE_CAST) reduce 202) ((T_ECHO) reduce 202) ((T_EMPTY) reduce 202) ((T_ENDDECLARE) reduce 202) ((T_ENUM) reduce 202) ((T_EVAL) reduce 202) ((T_EXIT) reduce 202) ((T_FILE) reduce 202) ((T_FINAL) reduce 202) ((T_FN) reduce 202) ((T_FOR) reduce 202) ((T_FOREACH) reduce 202) ((T_FUNCTION) reduce 202) ((T_FUNC_C) reduce 202) ((T_GLOBAL) reduce 202) ((T_GOTO) reduce 202) ((T_HALT_COMPILER) reduce 202) ((T_IF) reduce 202) ((T_INC) reduce 202) ((T_INCLUDE) reduce 202) ((T_INCLUDE_ONCE) reduce 202) ((T_INLINE_HTML) reduce 202) ((T_INTERFACE) reduce 202) ((T_INT_CAST) reduce 202) ((T_ISSET) reduce 202) ((T_LINE) reduce 202) ((T_LIST) reduce 202) ((T_LNUMBER) reduce 202) ((T_MATCH) reduce 202) ((T_METHOD_C) reduce 202) ((T_NAME_FULLY_QUALIFIED) reduce 202) ((T_NAME_QUALIFIED) reduce 202) ((T_NAME_RELATIVE) reduce 202) ((T_NEW) reduce 202) ((T_NS_C) reduce 202) ((T_OBJECT_CAST) reduce 202) ((T_PRINT) reduce 202) ((T_PROPERTY_C) reduce 202) ((T_READONLY) reduce 202) ((T_REQUIRE) reduce 202) ((T_REQUIRE_ONCE) reduce 202) ((T_RETURN) reduce 202) ((T_START_HEREDOC) reduce 202) ((T_STATIC) reduce 202) ((T_STRING) reduce 202) ((T_STRING_CAST) reduce 202) ((T_SWITCH) reduce 202) ((T_THROW) reduce 202) ((T_TRAIT) reduce 202) ((T_TRAIT_C) reduce 202) ((T_TRY) reduce 202) ((T_UNSET) reduce 202) ((T_UNSET_CAST) reduce 202) ((T_VARIABLE) reduce 202) ((T_VOID_CAST) reduce 202) ((T_WHILE) reduce 202) ((T_YIELD) reduce 202) ((T_YIELD_FROM) reduce 202) (("[") reduce 202) (("`") reduce 202) (("{") reduce 202) (("~") reduce 202)) 23462 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 280) ((T_IMPLEMENTS) reduce 280) (("{") reduce 280) (("|") reduce 280)) 23463 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_IMPLEMENTS) reduce 281) (("{") reduce 281) (("|") reduce 281)) 23464 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_IMPLEMENTS) reduce 94) (("{") reduce 94) (("|") reduce 94)) 23465 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_IMPLEMENTS) reduce 93) (("{") reduce 93) (("|") reduce 93)) 23466 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_IMPLEMENTS) reduce 95) (("{") reduce 95) (("|") reduce 95)) 23467 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 269) ((T_IMPLEMENTS) reduce 269) (("{") reduce 269) (("|") reduce 269)) 23468 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_IMPLEMENTS) reduce 92) (("{") reduce 92) (("|") reduce 92)) 23469 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_IMPLEMENTS) reduce 267) (("{") reduce 267)) 23470 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_IMPLEMENTS) reduce 282) (("{") reduce 282) (("|") reduce 282)) 23471 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_IMPLEMENTS) reduce 264) (("{") reduce 264) (("|") reduce 270)) 23472 (((T_IMPLEMENTS) reduce 204) (("{") reduce 204)) 23473 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 268) ((T_IMPLEMENTS) reduce 268) (("{") reduce 268) (("|") reduce 268)) 23474 (((T_IMPLEMENTS) reduce 266) (("{") reduce 266) (("|") shift)) 23478 (((T_IMPLEMENTS) reduce 280) (("{") reduce 280) (("|") reduce 280)) 23479 (((T_IMPLEMENTS) reduce 281) (("{") reduce 281) (("|") reduce 281)) 23480 (((T_IMPLEMENTS) reduce 94) (("{") reduce 94) (("|") reduce 94)) 23481 (((T_IMPLEMENTS) reduce 93) (("{") reduce 93) (("|") reduce 93)) 23482 (((T_IMPLEMENTS) reduce 95) (("{") reduce 95) (("|") reduce 95)) 23483 (((T_IMPLEMENTS) reduce 269) (("{") reduce 269) (("|") reduce 269)) 23484 (((T_IMPLEMENTS) reduce 92) (("{") reduce 92) (("|") reduce 92)) 23485 (((T_IMPLEMENTS) reduce 282) (("{") reduce 282) (("|") reduce 282)) 23486 (((T_IMPLEMENTS) reduce 270) (("{") reduce 270) (("|") reduce 270)) 23487 (((T_IMPLEMENTS) reduce 268) (("{") reduce 268) (("|") reduce 268)) 23488 (((T_IMPLEMENTS) reduce 272) (("{") reduce 272) (("|") reduce 272)) 23490 (((T_IMPLEMENTS) reduce 271) (("{") reduce 271) (("|") reduce 271)) 23492 (((T_IMPLEMENTS) reduce 273) (("{") reduce 273) (("|") reduce 273)) 23494 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 280) ((T_IMPLEMENTS) reduce 280) (("{") reduce 280)) 23495 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_IMPLEMENTS) reduce 281) (("{") reduce 281)) 23496 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_IMPLEMENTS) reduce 94) (("{") reduce 94)) 23497 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_IMPLEMENTS) reduce 93) (("{") reduce 93)) 23498 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_IMPLEMENTS) reduce 95) (("{") reduce 95)) 23499 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 269) ((T_IMPLEMENTS) reduce 269) (("{") reduce 269)) 23500 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_IMPLEMENTS) reduce 92) (("{") reduce 92)) 23501 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_IMPLEMENTS) reduce 282) (("{") reduce 282)) 23502 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 274) ((T_IMPLEMENTS) reduce 274) (("{") reduce 274)) 23503 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 268) ((T_IMPLEMENTS) reduce 268) (("{") reduce 268)) 23505 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 275) ((T_IMPLEMENTS) reduce 275) (("{") reduce 275)) 23506 (((T_IMPLEMENTS) reduce 280) (("{") reduce 280)) 23507 (((T_IMPLEMENTS) reduce 281) (("{") reduce 281)) 23508 (((T_IMPLEMENTS) reduce 269) (("{") reduce 269)) 23509 (((T_IMPLEMENTS) reduce 282) (("{") reduce 282)) 23510 (((T_IMPLEMENTS) reduce 265) (("{") reduce 265)) 23511 (((T_IMPLEMENTS) reduce 268) (("{") reduce 268)) 23512 (((T_ELSE) reduce 223) ((T_ELSEIF) reduce 223) ((T_WHILE) reduce 223)) 23514 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_EMPTY) reduce 159) ((T_ENDDECLARE) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_VOID_CAST) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 23520 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_EMPTY) reduce 151) ((T_ENDDECLARE) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_VOID_CAST) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 23525 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_EMPTY) reduce 165) ((T_ENDDECLARE) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_VOID_CAST) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 23526 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_EMPTY) reduce 222) ((T_ENDDECLARE) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_VARIABLE) reduce 222) ((T_VOID_CAST) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("~") reduce 222)) 23529 ((("!") reduce 223) (("\"") reduce 223) (("$") reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_EMPTY) reduce 223) ((T_ENDDECLARE) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_VARIABLE) reduce 223) ((T_VOID_CAST) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("~") reduce 223)) 23531 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_EMPTY) reduce 155) ((T_ENDDECLARE) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_VOID_CAST) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 23538 ((("!") reduce 190) (("\"") reduce 190) (("$") reduce 190) (("(") reduce 190) (("+") reduce 190) (("-") reduce 190) ((";") reduce 190) (("@") reduce 190) ((T_ABSTRACT) reduce 190) ((T_ARRAY) reduce 190) ((T_ARRAY_CAST) reduce 190) ((T_ATTRIBUTE) reduce 190) ((T_BOOL_CAST) reduce 190) ((T_BREAK) reduce 190) ((T_CLASS) reduce 190) ((T_CLASS_C) reduce 190) ((T_CLONE) reduce 190) ((T_CONSTANT_ENCAPSED_STRING) reduce 190) ((T_CONTINUE) reduce 190) ((T_DEC) reduce 190) ((T_DECLARE) reduce 190) ((T_DIR) reduce 190) ((T_DNUMBER) reduce 190) ((T_DO) reduce 190) ((T_DOUBLE_CAST) reduce 190) ((T_ECHO) reduce 190) ((T_EMPTY) reduce 190) ((T_ENDDECLARE) reduce 190) ((T_ENUM) reduce 190) ((T_EVAL) reduce 190) ((T_EXIT) reduce 190) ((T_FILE) reduce 190) ((T_FINAL) reduce 190) ((T_FN) reduce 190) ((T_FOR) reduce 190) ((T_FOREACH) reduce 190) ((T_FUNCTION) reduce 190) ((T_FUNC_C) reduce 190) ((T_GLOBAL) reduce 190) ((T_GOTO) reduce 190) ((T_HALT_COMPILER) reduce 190) ((T_IF) reduce 190) ((T_INC) reduce 190) ((T_INCLUDE) reduce 190) ((T_INCLUDE_ONCE) reduce 190) ((T_INLINE_HTML) reduce 190) ((T_INTERFACE) reduce 190) ((T_INT_CAST) reduce 190) ((T_ISSET) reduce 190) ((T_LINE) reduce 190) ((T_LIST) reduce 190) ((T_LNUMBER) reduce 190) ((T_MATCH) reduce 190) ((T_METHOD_C) reduce 190) ((T_NAME_FULLY_QUALIFIED) reduce 190) ((T_NAME_QUALIFIED) reduce 190) ((T_NAME_RELATIVE) reduce 190) ((T_NEW) reduce 190) ((T_NS_C) reduce 190) ((T_OBJECT_CAST) reduce 190) ((T_PRINT) reduce 190) ((T_PROPERTY_C) reduce 190) ((T_READONLY) reduce 190) ((T_REQUIRE) reduce 190) ((T_REQUIRE_ONCE) reduce 190) ((T_RETURN) reduce 190) ((T_START_HEREDOC) reduce 190) ((T_STATIC) reduce 190) ((T_STRING) reduce 190) ((T_STRING_CAST) reduce 190) ((T_SWITCH) reduce 190) ((T_THROW) reduce 190) ((T_TRAIT) reduce 190) ((T_TRAIT_C) reduce 190) ((T_TRY) reduce 190) ((T_UNSET) reduce 190) ((T_UNSET_CAST) reduce 190) ((T_VARIABLE) reduce 190) ((T_VOID_CAST) reduce 190) ((T_WHILE) reduce 190) ((T_YIELD) reduce 190) ((T_YIELD_FROM) reduce 190) (("[") reduce 190) (("`") reduce 190) (("{") reduce 190) (("~") reduce 190)) 23540 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_EMPTY) reduce 154) ((T_ENDDECLARE) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_VOID_CAST) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 23542 ((("%") shift) ((")") reduce 493) (("*") shift) (("+") shift) ((",") reduce 493) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23543 (((")") reduce 385) ((",") reduce 385)) 23545 (((T_ELSE) reduce 155) ((T_ELSEIF) reduce 155) ((T_WHILE) reduce 155)) 23547 (((T_ELSE) reduce 154) ((T_ELSEIF) reduce 154) ((T_WHILE) reduce 154)) 23548 (((T_WHILE) reduce 247)) 23549 (((T_WHILE) reduce 161)) 23552 (((T_WHILE) reduce 250)) 23556 (((T_WHILE) reduce 251)) 23561 (((T_WHILE) reduce 242)) 23562 (((T_WHILE) reduce 150)) 23565 (((T_WHILE) reduce 243)) 23567 (((T_WHILE) reduce 170)) 23572 (((T_WHILE) reduce 162)) 23575 (((T_CATCH) reduce 171) ((T_FINALLY) reduce 171) ((T_WHILE) reduce 171)) 23576 (((T_CATCH) shift) ((T_FINALLY) shift) ((T_WHILE) reduce 177)) 23579 (((T_WHILE) reduce 167)) 23582 (((T_WHILE) reduce 178)) 23589 (((T_CATCH) reduce 172) ((T_FINALLY) reduce 172) ((T_WHILE) reduce 172)) 23594 (((T_WHILE) reduce 153)) 23598 (((T_WHILE) reduce 224)) 23600 (((T_WHILE) reduce 225)) 23604 (((T_WHILE) reduce 226)) 23607 (((T_WHILE) reduce 227)) 23608 (((T_WHILE) reduce 169)) 23610 (((T_WHILE) reduce 158)) 23612 (((T_WHILE) reduce 156)) 23616 (((T_ELSE) reduce 244) ((T_ELSEIF) reduce 244) ((T_WHILE) reduce 244)) 23618 (((T_WHILE) reduce 168)) 23620 (((T_WHILE) reduce 157)) 23630 (((T_WHILE) reduce 164)) 23631 (((T_WHILE) reduce 220)) 23634 (((T_WHILE) reduce 221)) 23635 (((T_WHILE) reduce 163)) 23644 (((T_WHILE) reduce 152)) 23645 (((T_WHILE) reduce 218)) 23648 (((T_WHILE) reduce 219)) 23650 (((T_WHILE) reduce 159)) 23656 (((T_WHILE) reduce 151)) 23661 (((T_WHILE) reduce 165)) 23662 (((T_WHILE) reduce 222)) 23665 (((T_WHILE) reduce 223)) 23667 (((T_WHILE) reduce 155)) 23669 (((T_WHILE) reduce 154)) 23674 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_ELSE) reduce 165) ((T_ELSEIF) reduce 165) ((T_EMPTY) reduce 165) ((T_ENDFOR) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_VOID_CAST) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 23675 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_ELSE) reduce 222) ((T_ELSEIF) reduce 222) ((T_EMPTY) reduce 222) ((T_ENDFOR) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_VARIABLE) reduce 222) ((T_VOID_CAST) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("~") reduce 222)) 23678 ((("!") reduce 223) (("\"") reduce 223) (("$") reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_ELSE) reduce 223) ((T_ELSEIF) reduce 223) ((T_EMPTY) reduce 223) ((T_ENDFOR) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_VARIABLE) reduce 223) ((T_VOID_CAST) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("~") reduce 223)) 23680 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_ELSE) reduce 155) ((T_ELSEIF) reduce 155) ((T_EMPTY) reduce 155) ((T_ENDFOR) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_VOID_CAST) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 23682 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_ELSE) reduce 154) ((T_ELSEIF) reduce 154) ((T_EMPTY) reduce 154) ((T_ENDFOR) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_VOID_CAST) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 23683 ((("!") reduce 247) (("\"") reduce 247) (("$") reduce 247) (("(") reduce 247) (("+") reduce 247) (("-") reduce 247) ((";") reduce 247) (("@") reduce 247) ((T_ABSTRACT) reduce 247) ((T_ARRAY) reduce 247) ((T_ARRAY_CAST) reduce 247) ((T_ATTRIBUTE) reduce 247) ((T_BOOL_CAST) reduce 247) ((T_BREAK) reduce 247) ((T_CLASS) reduce 247) ((T_CLASS_C) reduce 247) ((T_CLONE) reduce 247) ((T_CONSTANT_ENCAPSED_STRING) reduce 247) ((T_CONTINUE) reduce 247) ((T_DEC) reduce 247) ((T_DECLARE) reduce 247) ((T_DIR) reduce 247) ((T_DNUMBER) reduce 247) ((T_DO) reduce 247) ((T_DOUBLE_CAST) reduce 247) ((T_ECHO) reduce 247) ((T_EMPTY) reduce 247) ((T_ENDFOR) reduce 247) ((T_ENUM) reduce 247) ((T_EVAL) reduce 247) ((T_EXIT) reduce 247) ((T_FILE) reduce 247) ((T_FINAL) reduce 247) ((T_FN) reduce 247) ((T_FOR) reduce 247) ((T_FOREACH) reduce 247) ((T_FUNCTION) reduce 247) ((T_FUNC_C) reduce 247) ((T_GLOBAL) reduce 247) ((T_GOTO) reduce 247) ((T_HALT_COMPILER) reduce 247) ((T_IF) reduce 247) ((T_INC) reduce 247) ((T_INCLUDE) reduce 247) ((T_INCLUDE_ONCE) reduce 247) ((T_INLINE_HTML) reduce 247) ((T_INTERFACE) reduce 247) ((T_INT_CAST) reduce 247) ((T_ISSET) reduce 247) ((T_LINE) reduce 247) ((T_LIST) reduce 247) ((T_LNUMBER) reduce 247) ((T_MATCH) reduce 247) ((T_METHOD_C) reduce 247) ((T_NAME_FULLY_QUALIFIED) reduce 247) ((T_NAME_QUALIFIED) reduce 247) ((T_NAME_RELATIVE) reduce 247) ((T_NEW) reduce 247) ((T_NS_C) reduce 247) ((T_OBJECT_CAST) reduce 247) ((T_PRINT) reduce 247) ((T_PROPERTY_C) reduce 247) ((T_READONLY) reduce 247) ((T_REQUIRE) reduce 247) ((T_REQUIRE_ONCE) reduce 247) ((T_RETURN) reduce 247) ((T_START_HEREDOC) reduce 247) ((T_STATIC) reduce 247) ((T_STRING) reduce 247) ((T_STRING_CAST) reduce 247) ((T_SWITCH) reduce 247) ((T_THROW) reduce 247) ((T_TRAIT) reduce 247) ((T_TRAIT_C) reduce 247) ((T_TRY) reduce 247) ((T_UNSET) reduce 247) ((T_UNSET_CAST) reduce 247) ((T_VARIABLE) reduce 247) ((T_VOID_CAST) reduce 247) ((T_WHILE) reduce 247) ((T_YIELD) reduce 247) ((T_YIELD_FROM) reduce 247) (("[") reduce 247) (("`") reduce 247) (("{") reduce 247) (("~") reduce 247)) 23694 ((("!") reduce 494) (("\"") reduce 494) (("$") reduce 494) (("(") reduce 494) (("+") reduce 494) (("-") reduce 494) ((";") reduce 494) (("@") reduce 494) ((T_ABSTRACT) reduce 494) ((T_ARRAY) reduce 494) ((T_ARRAY_CAST) reduce 494) ((T_ATTRIBUTE) reduce 494) ((T_BOOL_CAST) reduce 494) ((T_BREAK) reduce 494) ((T_CLASS) reduce 494) ((T_CLASS_C) reduce 494) ((T_CLONE) reduce 494) ((T_CONSTANT_ENCAPSED_STRING) reduce 494) ((T_CONTINUE) reduce 494) ((T_DEC) reduce 494) ((T_DECLARE) reduce 494) ((T_DIR) reduce 494) ((T_DNUMBER) reduce 494) ((T_DO) reduce 494) ((T_DOUBLE_CAST) reduce 494) ((T_ECHO) reduce 494) ((T_EMPTY) reduce 494) ((T_ENDFOR) reduce 494) ((T_ENUM) reduce 494) ((T_EVAL) reduce 494) ((T_EXIT) reduce 494) ((T_FILE) reduce 494) ((T_FINAL) reduce 494) ((T_FN) reduce 494) ((T_FOR) reduce 494) ((T_FOREACH) reduce 494) ((T_FUNCTION) reduce 494) ((T_FUNC_C) reduce 494) ((T_GLOBAL) reduce 494) ((T_GOTO) reduce 494) ((T_HALT_COMPILER) reduce 494) ((T_IF) reduce 494) ((T_INC) reduce 494) ((T_INCLUDE) reduce 494) ((T_INCLUDE_ONCE) reduce 494) ((T_INLINE_HTML) reduce 494) ((T_INTERFACE) reduce 494) ((T_INT_CAST) reduce 494) ((T_ISSET) reduce 494) ((T_LINE) reduce 494) ((T_LIST) reduce 494) ((T_LNUMBER) reduce 494) ((T_MATCH) reduce 494) ((T_METHOD_C) reduce 494) ((T_NAME_FULLY_QUALIFIED) reduce 494) ((T_NAME_QUALIFIED) reduce 494) ((T_NAME_RELATIVE) reduce 494) ((T_NEW) reduce 494) ((T_NS_C) reduce 494) ((T_OBJECT_CAST) reduce 494) ((T_PRINT) reduce 494) ((T_PROPERTY_C) reduce 494) ((T_READONLY) reduce 494) ((T_REQUIRE) reduce 494) ((T_REQUIRE_ONCE) reduce 494) ((T_RETURN) reduce 494) ((T_START_HEREDOC) reduce 494) ((T_STATIC) reduce 494) ((T_STRING) reduce 494) ((T_STRING_CAST) reduce 494) ((T_SWITCH) reduce 494) ((T_THROW) reduce 494) ((T_TRAIT) reduce 494) ((T_TRAIT_C) reduce 494) ((T_TRY) reduce 494) ((T_UNSET) reduce 494) ((T_UNSET_CAST) reduce 494) ((T_VARIABLE) reduce 494) ((T_VOID_CAST) reduce 494) ((T_WHILE) reduce 494) ((T_YIELD) reduce 494) ((T_YIELD_FROM) reduce 494) (("[") reduce 494) (("`") reduce 494) (("{") reduce 494) (("~") reduce 494)) 23695 ((("!") reduce 184) (("\"") reduce 184) (("$") reduce 184) (("(") reduce 184) (("+") reduce 184) (("-") reduce 184) ((";") reduce 184) (("@") reduce 184) ((T_ABSTRACT) reduce 184) ((T_ARRAY) reduce 184) ((T_ARRAY_CAST) reduce 184) ((T_ATTRIBUTE) reduce 184) ((T_BOOL_CAST) reduce 184) ((T_BREAK) reduce 184) ((T_CLASS) reduce 184) ((T_CLASS_C) reduce 184) ((T_CLONE) reduce 184) ((T_CONSTANT_ENCAPSED_STRING) reduce 184) ((T_CONTINUE) reduce 184) ((T_DEC) reduce 184) ((T_DECLARE) reduce 184) ((T_DIR) reduce 184) ((T_DNUMBER) reduce 184) ((T_DO) reduce 184) ((T_DOUBLE_CAST) reduce 184) ((T_ECHO) reduce 184) ((T_EMPTY) reduce 184) ((T_ENDFOR) reduce 184) ((T_ENUM) reduce 184) ((T_EVAL) reduce 184) ((T_EXIT) reduce 184) ((T_FILE) reduce 184) ((T_FINAL) reduce 184) ((T_FN) reduce 184) ((T_FOR) reduce 184) ((T_FOREACH) reduce 184) ((T_FUNCTION) reduce 184) ((T_FUNC_C) reduce 184) ((T_GLOBAL) reduce 184) ((T_GOTO) reduce 184) ((T_HALT_COMPILER) reduce 184) ((T_IF) reduce 184) ((T_INC) reduce 184) ((T_INCLUDE) reduce 184) ((T_INCLUDE_ONCE) reduce 184) ((T_INLINE_HTML) reduce 184) ((T_INTERFACE) reduce 184) ((T_INT_CAST) reduce 184) ((T_ISSET) reduce 184) ((T_LINE) reduce 184) ((T_LIST) reduce 184) ((T_LNUMBER) reduce 184) ((T_MATCH) reduce 184) ((T_METHOD_C) reduce 184) ((T_NAME_FULLY_QUALIFIED) reduce 184) ((T_NAME_QUALIFIED) reduce 184) ((T_NAME_RELATIVE) reduce 184) ((T_NEW) reduce 184) ((T_NS_C) reduce 184) ((T_OBJECT_CAST) reduce 184) ((T_PRINT) reduce 184) ((T_PROPERTY_C) reduce 184) ((T_READONLY) reduce 184) ((T_REQUIRE) reduce 184) ((T_REQUIRE_ONCE) reduce 184) ((T_RETURN) reduce 184) ((T_START_HEREDOC) reduce 184) ((T_STATIC) reduce 184) ((T_STRING) reduce 184) ((T_STRING_CAST) reduce 184) ((T_SWITCH) reduce 184) ((T_THROW) reduce 184) ((T_TRAIT) reduce 184) ((T_TRAIT_C) reduce 184) ((T_TRY) reduce 184) ((T_UNSET) reduce 184) ((T_UNSET_CAST) reduce 184) ((T_VARIABLE) reduce 184) ((T_VOID_CAST) reduce 184) ((T_WHILE) reduce 184) ((T_YIELD) reduce 184) ((T_YIELD_FROM) reduce 184) (("[") reduce 184) (("`") reduce 184) (("{") reduce 184) (("~") reduce 184)) 23696 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_EMPTY) reduce 161) ((T_ENDFOR) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_VOID_CAST) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 23704 ((("!") reduce 189) (("\"") reduce 189) (("$") reduce 189) (("(") reduce 189) (("+") reduce 189) (("-") reduce 189) ((";") reduce 189) (("@") reduce 189) ((T_ABSTRACT) reduce 189) ((T_ARRAY) reduce 189) ((T_ARRAY_CAST) reduce 189) ((T_ATTRIBUTE) reduce 189) ((T_BOOL_CAST) reduce 189) ((T_BREAK) reduce 189) ((T_CLASS) reduce 189) ((T_CLASS_C) reduce 189) ((T_CLONE) reduce 189) ((T_CONSTANT_ENCAPSED_STRING) reduce 189) ((T_CONTINUE) reduce 189) ((T_DEC) reduce 189) ((T_DECLARE) reduce 189) ((T_DIR) reduce 189) ((T_DNUMBER) reduce 189) ((T_DO) reduce 189) ((T_DOUBLE_CAST) reduce 189) ((T_ECHO) reduce 189) ((T_EMPTY) reduce 189) ((T_ENDFOR) reduce 189) ((T_ENUM) reduce 189) ((T_EVAL) reduce 189) ((T_EXIT) reduce 189) ((T_FILE) reduce 189) ((T_FINAL) reduce 189) ((T_FN) reduce 189) ((T_FOR) reduce 189) ((T_FOREACH) reduce 189) ((T_FUNCTION) reduce 189) ((T_FUNC_C) reduce 189) ((T_GLOBAL) reduce 189) ((T_GOTO) reduce 189) ((T_HALT_COMPILER) reduce 189) ((T_IF) reduce 189) ((T_INC) reduce 189) ((T_INCLUDE) reduce 189) ((T_INCLUDE_ONCE) reduce 189) ((T_INLINE_HTML) reduce 189) ((T_INTERFACE) reduce 189) ((T_INT_CAST) reduce 189) ((T_ISSET) reduce 189) ((T_LINE) reduce 189) ((T_LIST) reduce 189) ((T_LNUMBER) reduce 189) ((T_MATCH) reduce 189) ((T_METHOD_C) reduce 189) ((T_NAME_FULLY_QUALIFIED) reduce 189) ((T_NAME_QUALIFIED) reduce 189) ((T_NAME_RELATIVE) reduce 189) ((T_NEW) reduce 189) ((T_NS_C) reduce 189) ((T_OBJECT_CAST) reduce 189) ((T_PRINT) reduce 189) ((T_PROPERTY_C) reduce 189) ((T_READONLY) reduce 189) ((T_REQUIRE) reduce 189) ((T_REQUIRE_ONCE) reduce 189) ((T_RETURN) reduce 189) ((T_START_HEREDOC) reduce 189) ((T_STATIC) reduce 189) ((T_STRING) reduce 189) ((T_STRING_CAST) reduce 189) ((T_SWITCH) reduce 189) ((T_THROW) reduce 189) ((T_TRAIT) reduce 189) ((T_TRAIT_C) reduce 189) ((T_TRY) reduce 189) ((T_UNSET) reduce 189) ((T_UNSET_CAST) reduce 189) ((T_VARIABLE) reduce 189) ((T_VOID_CAST) reduce 189) ((T_WHILE) reduce 189) ((T_YIELD) reduce 189) ((T_YIELD_FROM) reduce 189) (("[") reduce 189) (("`") reduce 189) (("{") reduce 189) (("~") reduce 189)) 23705 ((("!") reduce 145) (("\"") reduce 145) (("$") reduce 145) (("(") reduce 145) (("+") reduce 145) (("-") reduce 145) ((";") reduce 145) (("@") reduce 145) ((T_ABSTRACT) reduce 145) ((T_ARRAY) reduce 145) ((T_ARRAY_CAST) reduce 145) ((T_ATTRIBUTE) reduce 145) ((T_BOOL_CAST) reduce 145) ((T_BREAK) reduce 145) ((T_CLASS) reduce 145) ((T_CLASS_C) reduce 145) ((T_CLONE) reduce 145) ((T_CONSTANT_ENCAPSED_STRING) reduce 145) ((T_CONTINUE) reduce 145) ((T_DEC) reduce 145) ((T_DECLARE) reduce 145) ((T_DIR) reduce 145) ((T_DNUMBER) reduce 145) ((T_DO) reduce 145) ((T_DOUBLE_CAST) reduce 145) ((T_ECHO) reduce 145) ((T_EMPTY) reduce 145) ((T_ENDFOR) reduce 145) ((T_ENUM) reduce 145) ((T_EVAL) reduce 145) ((T_EXIT) reduce 145) ((T_FILE) reduce 145) ((T_FINAL) reduce 145) ((T_FN) reduce 145) ((T_FOR) reduce 145) ((T_FOREACH) reduce 145) ((T_FUNCTION) reduce 145) ((T_FUNC_C) reduce 145) ((T_GLOBAL) reduce 145) ((T_GOTO) reduce 145) ((T_HALT_COMPILER) reduce 145) ((T_IF) reduce 145) ((T_INC) reduce 145) ((T_INCLUDE) reduce 145) ((T_INCLUDE_ONCE) reduce 145) ((T_INLINE_HTML) reduce 145) ((T_INTERFACE) reduce 145) ((T_INT_CAST) reduce 145) ((T_ISSET) reduce 145) ((T_LINE) reduce 145) ((T_LIST) reduce 145) ((T_LNUMBER) reduce 145) ((T_MATCH) reduce 145) ((T_METHOD_C) reduce 145) ((T_NAME_FULLY_QUALIFIED) reduce 145) ((T_NAME_QUALIFIED) reduce 145) ((T_NAME_RELATIVE) reduce 145) ((T_NEW) reduce 145) ((T_NS_C) reduce 145) ((T_OBJECT_CAST) reduce 145) ((T_PRINT) reduce 145) ((T_PROPERTY_C) reduce 145) ((T_READONLY) reduce 145) ((T_REQUIRE) reduce 145) ((T_REQUIRE_ONCE) reduce 145) ((T_RETURN) reduce 145) ((T_START_HEREDOC) reduce 145) ((T_STATIC) reduce 145) ((T_STRING) reduce 145) ((T_STRING_CAST) reduce 145) ((T_SWITCH) reduce 145) ((T_THROW) reduce 145) ((T_TRAIT) reduce 145) ((T_TRAIT_C) reduce 145) ((T_TRY) reduce 145) ((T_UNSET) reduce 145) ((T_UNSET_CAST) reduce 145) ((T_VARIABLE) reduce 145) ((T_VOID_CAST) reduce 145) ((T_WHILE) reduce 145) ((T_YIELD) reduce 145) ((T_YIELD_FROM) reduce 145) (("[") reduce 145) (("`") reduce 145) (("{") reduce 145) (("~") reduce 145)) 23708 ((("!") reduce 250) (("\"") reduce 250) (("$") reduce 250) (("(") reduce 250) (("+") reduce 250) (("-") reduce 250) ((";") reduce 250) (("@") reduce 250) ((T_ABSTRACT) reduce 250) ((T_ARRAY) reduce 250) ((T_ARRAY_CAST) reduce 250) ((T_ATTRIBUTE) reduce 250) ((T_BOOL_CAST) reduce 250) ((T_BREAK) reduce 250) ((T_CLASS) reduce 250) ((T_CLASS_C) reduce 250) ((T_CLONE) reduce 250) ((T_CONSTANT_ENCAPSED_STRING) reduce 250) ((T_CONTINUE) reduce 250) ((T_DEC) reduce 250) ((T_DECLARE) reduce 250) ((T_DIR) reduce 250) ((T_DNUMBER) reduce 250) ((T_DO) reduce 250) ((T_DOUBLE_CAST) reduce 250) ((T_ECHO) reduce 250) ((T_EMPTY) reduce 250) ((T_ENDFOR) reduce 250) ((T_ENUM) reduce 250) ((T_EVAL) reduce 250) ((T_EXIT) reduce 250) ((T_FILE) reduce 250) ((T_FINAL) reduce 250) ((T_FN) reduce 250) ((T_FOR) reduce 250) ((T_FOREACH) reduce 250) ((T_FUNCTION) reduce 250) ((T_FUNC_C) reduce 250) ((T_GLOBAL) reduce 250) ((T_GOTO) reduce 250) ((T_HALT_COMPILER) reduce 250) ((T_IF) reduce 250) ((T_INC) reduce 250) ((T_INCLUDE) reduce 250) ((T_INCLUDE_ONCE) reduce 250) ((T_INLINE_HTML) reduce 250) ((T_INTERFACE) reduce 250) ((T_INT_CAST) reduce 250) ((T_ISSET) reduce 250) ((T_LINE) reduce 250) ((T_LIST) reduce 250) ((T_LNUMBER) reduce 250) ((T_MATCH) reduce 250) ((T_METHOD_C) reduce 250) ((T_NAME_FULLY_QUALIFIED) reduce 250) ((T_NAME_QUALIFIED) reduce 250) ((T_NAME_RELATIVE) reduce 250) ((T_NEW) reduce 250) ((T_NS_C) reduce 250) ((T_OBJECT_CAST) reduce 250) ((T_PRINT) reduce 250) ((T_PROPERTY_C) reduce 250) ((T_READONLY) reduce 250) ((T_REQUIRE) reduce 250) ((T_REQUIRE_ONCE) reduce 250) ((T_RETURN) reduce 250) ((T_START_HEREDOC) reduce 250) ((T_STATIC) reduce 250) ((T_STRING) reduce 250) ((T_STRING_CAST) reduce 250) ((T_SWITCH) reduce 250) ((T_THROW) reduce 250) ((T_TRAIT) reduce 250) ((T_TRAIT_C) reduce 250) ((T_TRY) reduce 250) ((T_UNSET) reduce 250) ((T_UNSET_CAST) reduce 250) ((T_VARIABLE) reduce 250) ((T_VOID_CAST) reduce 250) ((T_WHILE) reduce 250) ((T_YIELD) reduce 250) ((T_YIELD_FROM) reduce 250) (("[") reduce 250) (("`") reduce 250) (("{") reduce 250) (("~") reduce 250)) 23712 ((("!") reduce 251) (("\"") reduce 251) (("$") reduce 251) (("(") reduce 251) (("+") reduce 251) (("-") reduce 251) ((";") reduce 251) (("@") reduce 251) ((T_ABSTRACT) reduce 251) ((T_ARRAY) reduce 251) ((T_ARRAY_CAST) reduce 251) ((T_ATTRIBUTE) reduce 251) ((T_BOOL_CAST) reduce 251) ((T_BREAK) reduce 251) ((T_CLASS) reduce 251) ((T_CLASS_C) reduce 251) ((T_CLONE) reduce 251) ((T_CONSTANT_ENCAPSED_STRING) reduce 251) ((T_CONTINUE) reduce 251) ((T_DEC) reduce 251) ((T_DECLARE) reduce 251) ((T_DIR) reduce 251) ((T_DNUMBER) reduce 251) ((T_DO) reduce 251) ((T_DOUBLE_CAST) reduce 251) ((T_ECHO) reduce 251) ((T_EMPTY) reduce 251) ((T_ENDFOR) reduce 251) ((T_ENUM) reduce 251) ((T_EVAL) reduce 251) ((T_EXIT) reduce 251) ((T_FILE) reduce 251) ((T_FINAL) reduce 251) ((T_FN) reduce 251) ((T_FOR) reduce 251) ((T_FOREACH) reduce 251) ((T_FUNCTION) reduce 251) ((T_FUNC_C) reduce 251) ((T_GLOBAL) reduce 251) ((T_GOTO) reduce 251) ((T_HALT_COMPILER) reduce 251) ((T_IF) reduce 251) ((T_INC) reduce 251) ((T_INCLUDE) reduce 251) ((T_INCLUDE_ONCE) reduce 251) ((T_INLINE_HTML) reduce 251) ((T_INTERFACE) reduce 251) ((T_INT_CAST) reduce 251) ((T_ISSET) reduce 251) ((T_LINE) reduce 251) ((T_LIST) reduce 251) ((T_LNUMBER) reduce 251) ((T_MATCH) reduce 251) ((T_METHOD_C) reduce 251) ((T_NAME_FULLY_QUALIFIED) reduce 251) ((T_NAME_QUALIFIED) reduce 251) ((T_NAME_RELATIVE) reduce 251) ((T_NEW) reduce 251) ((T_NS_C) reduce 251) ((T_OBJECT_CAST) reduce 251) ((T_PRINT) reduce 251) ((T_PROPERTY_C) reduce 251) ((T_READONLY) reduce 251) ((T_REQUIRE) reduce 251) ((T_REQUIRE_ONCE) reduce 251) ((T_RETURN) reduce 251) ((T_START_HEREDOC) reduce 251) ((T_STATIC) reduce 251) ((T_STRING) reduce 251) ((T_STRING_CAST) reduce 251) ((T_SWITCH) reduce 251) ((T_THROW) reduce 251) ((T_TRAIT) reduce 251) ((T_TRAIT_C) reduce 251) ((T_TRY) reduce 251) ((T_UNSET) reduce 251) ((T_UNSET_CAST) reduce 251) ((T_VARIABLE) reduce 251) ((T_VOID_CAST) reduce 251) ((T_WHILE) reduce 251) ((T_YIELD) reduce 251) ((T_YIELD_FROM) reduce 251) (("[") reduce 251) (("`") reduce 251) (("{") reduce 251) (("~") reduce 251)) 23717 ((("!") reduce 242) (("\"") reduce 242) (("$") reduce 242) (("(") reduce 242) (("+") reduce 242) (("-") reduce 242) ((";") reduce 242) (("@") reduce 242) ((T_ABSTRACT) reduce 242) ((T_ARRAY) reduce 242) ((T_ARRAY_CAST) reduce 242) ((T_ATTRIBUTE) reduce 242) ((T_BOOL_CAST) reduce 242) ((T_BREAK) reduce 242) ((T_CLASS) reduce 242) ((T_CLASS_C) reduce 242) ((T_CLONE) reduce 242) ((T_CONSTANT_ENCAPSED_STRING) reduce 242) ((T_CONTINUE) reduce 242) ((T_DEC) reduce 242) ((T_DECLARE) reduce 242) ((T_DIR) reduce 242) ((T_DNUMBER) reduce 242) ((T_DO) reduce 242) ((T_DOUBLE_CAST) reduce 242) ((T_ECHO) reduce 242) ((T_EMPTY) reduce 242) ((T_ENDFOR) reduce 242) ((T_ENUM) reduce 242) ((T_EVAL) reduce 242) ((T_EXIT) reduce 242) ((T_FILE) reduce 242) ((T_FINAL) reduce 242) ((T_FN) reduce 242) ((T_FOR) reduce 242) ((T_FOREACH) reduce 242) ((T_FUNCTION) reduce 242) ((T_FUNC_C) reduce 242) ((T_GLOBAL) reduce 242) ((T_GOTO) reduce 242) ((T_HALT_COMPILER) reduce 242) ((T_IF) reduce 242) ((T_INC) reduce 242) ((T_INCLUDE) reduce 242) ((T_INCLUDE_ONCE) reduce 242) ((T_INLINE_HTML) reduce 242) ((T_INTERFACE) reduce 242) ((T_INT_CAST) reduce 242) ((T_ISSET) reduce 242) ((T_LINE) reduce 242) ((T_LIST) reduce 242) ((T_LNUMBER) reduce 242) ((T_MATCH) reduce 242) ((T_METHOD_C) reduce 242) ((T_NAME_FULLY_QUALIFIED) reduce 242) ((T_NAME_QUALIFIED) reduce 242) ((T_NAME_RELATIVE) reduce 242) ((T_NEW) reduce 242) ((T_NS_C) reduce 242) ((T_OBJECT_CAST) reduce 242) ((T_PRINT) reduce 242) ((T_PROPERTY_C) reduce 242) ((T_READONLY) reduce 242) ((T_REQUIRE) reduce 242) ((T_REQUIRE_ONCE) reduce 242) ((T_RETURN) reduce 242) ((T_START_HEREDOC) reduce 242) ((T_STATIC) reduce 242) ((T_STRING) reduce 242) ((T_STRING_CAST) reduce 242) ((T_SWITCH) reduce 242) ((T_THROW) reduce 242) ((T_TRAIT) reduce 242) ((T_TRAIT_C) reduce 242) ((T_TRY) reduce 242) ((T_UNSET) reduce 242) ((T_UNSET_CAST) reduce 242) ((T_VARIABLE) reduce 242) ((T_VOID_CAST) reduce 242) ((T_WHILE) reduce 242) ((T_YIELD) reduce 242) ((T_YIELD_FROM) reduce 242) (("[") reduce 242) (("`") reduce 242) (("{") reduce 242) (("~") reduce 242)) 23718 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_EMPTY) reduce 150) ((T_ENDFOR) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_VOID_CAST) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 23721 ((("!") reduce 243) (("\"") reduce 243) (("$") reduce 243) (("(") reduce 243) (("+") reduce 243) (("-") reduce 243) ((";") reduce 243) (("@") reduce 243) ((T_ABSTRACT) reduce 243) ((T_ARRAY) reduce 243) ((T_ARRAY_CAST) reduce 243) ((T_ATTRIBUTE) reduce 243) ((T_BOOL_CAST) reduce 243) ((T_BREAK) reduce 243) ((T_CLASS) reduce 243) ((T_CLASS_C) reduce 243) ((T_CLONE) reduce 243) ((T_CONSTANT_ENCAPSED_STRING) reduce 243) ((T_CONTINUE) reduce 243) ((T_DEC) reduce 243) ((T_DECLARE) reduce 243) ((T_DIR) reduce 243) ((T_DNUMBER) reduce 243) ((T_DO) reduce 243) ((T_DOUBLE_CAST) reduce 243) ((T_ECHO) reduce 243) ((T_EMPTY) reduce 243) ((T_ENDFOR) reduce 243) ((T_ENUM) reduce 243) ((T_EVAL) reduce 243) ((T_EXIT) reduce 243) ((T_FILE) reduce 243) ((T_FINAL) reduce 243) ((T_FN) reduce 243) ((T_FOR) reduce 243) ((T_FOREACH) reduce 243) ((T_FUNCTION) reduce 243) ((T_FUNC_C) reduce 243) ((T_GLOBAL) reduce 243) ((T_GOTO) reduce 243) ((T_HALT_COMPILER) reduce 243) ((T_IF) reduce 243) ((T_INC) reduce 243) ((T_INCLUDE) reduce 243) ((T_INCLUDE_ONCE) reduce 243) ((T_INLINE_HTML) reduce 243) ((T_INTERFACE) reduce 243) ((T_INT_CAST) reduce 243) ((T_ISSET) reduce 243) ((T_LINE) reduce 243) ((T_LIST) reduce 243) ((T_LNUMBER) reduce 243) ((T_MATCH) reduce 243) ((T_METHOD_C) reduce 243) ((T_NAME_FULLY_QUALIFIED) reduce 243) ((T_NAME_QUALIFIED) reduce 243) ((T_NAME_RELATIVE) reduce 243) ((T_NEW) reduce 243) ((T_NS_C) reduce 243) ((T_OBJECT_CAST) reduce 243) ((T_PRINT) reduce 243) ((T_PROPERTY_C) reduce 243) ((T_READONLY) reduce 243) ((T_REQUIRE) reduce 243) ((T_REQUIRE_ONCE) reduce 243) ((T_RETURN) reduce 243) ((T_START_HEREDOC) reduce 243) ((T_STATIC) reduce 243) ((T_STRING) reduce 243) ((T_STRING_CAST) reduce 243) ((T_SWITCH) reduce 243) ((T_THROW) reduce 243) ((T_TRAIT) reduce 243) ((T_TRAIT_C) reduce 243) ((T_TRY) reduce 243) ((T_UNSET) reduce 243) ((T_UNSET_CAST) reduce 243) ((T_VARIABLE) reduce 243) ((T_VOID_CAST) reduce 243) ((T_WHILE) reduce 243) ((T_YIELD) reduce 243) ((T_YIELD_FROM) reduce 243) (("[") reduce 243) (("`") reduce 243) (("{") reduce 243) (("~") reduce 243)) 23723 ((("!") reduce 170) (("\"") reduce 170) (("$") reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_EMPTY) reduce 170) ((T_ENDFOR) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_VARIABLE) reduce 170) ((T_VOID_CAST) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("~") reduce 170)) 23728 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_EMPTY) reduce 162) ((T_ENDFOR) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_VOID_CAST) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 23731 ((("!") reduce 171) (("\"") reduce 171) (("$") reduce 171) (("(") reduce 171) (("+") reduce 171) (("-") reduce 171) ((";") reduce 171) (("@") reduce 171) ((T_ABSTRACT) reduce 171) ((T_ARRAY) reduce 171) ((T_ARRAY_CAST) reduce 171) ((T_ATTRIBUTE) reduce 171) ((T_BOOL_CAST) reduce 171) ((T_BREAK) reduce 171) ((T_CATCH) reduce 171) ((T_CLASS) reduce 171) ((T_CLASS_C) reduce 171) ((T_CLONE) reduce 171) ((T_CONSTANT_ENCAPSED_STRING) reduce 171) ((T_CONTINUE) reduce 171) ((T_DEC) reduce 171) ((T_DECLARE) reduce 171) ((T_DIR) reduce 171) ((T_DNUMBER) reduce 171) ((T_DO) reduce 171) ((T_DOUBLE_CAST) reduce 171) ((T_ECHO) reduce 171) ((T_EMPTY) reduce 171) ((T_ENDFOR) reduce 171) ((T_ENUM) reduce 171) ((T_EVAL) reduce 171) ((T_EXIT) reduce 171) ((T_FILE) reduce 171) ((T_FINAL) reduce 171) ((T_FINALLY) reduce 171) ((T_FN) reduce 171) ((T_FOR) reduce 171) ((T_FOREACH) reduce 171) ((T_FUNCTION) reduce 171) ((T_FUNC_C) reduce 171) ((T_GLOBAL) reduce 171) ((T_GOTO) reduce 171) ((T_HALT_COMPILER) reduce 171) ((T_IF) reduce 171) ((T_INC) reduce 171) ((T_INCLUDE) reduce 171) ((T_INCLUDE_ONCE) reduce 171) ((T_INLINE_HTML) reduce 171) ((T_INTERFACE) reduce 171) ((T_INT_CAST) reduce 171) ((T_ISSET) reduce 171) ((T_LINE) reduce 171) ((T_LIST) reduce 171) ((T_LNUMBER) reduce 171) ((T_MATCH) reduce 171) ((T_METHOD_C) reduce 171) ((T_NAME_FULLY_QUALIFIED) reduce 171) ((T_NAME_QUALIFIED) reduce 171) ((T_NAME_RELATIVE) reduce 171) ((T_NEW) reduce 171) ((T_NS_C) reduce 171) ((T_OBJECT_CAST) reduce 171) ((T_PRINT) reduce 171) ((T_PROPERTY_C) reduce 171) ((T_READONLY) reduce 171) ((T_REQUIRE) reduce 171) ((T_REQUIRE_ONCE) reduce 171) ((T_RETURN) reduce 171) ((T_START_HEREDOC) reduce 171) ((T_STATIC) reduce 171) ((T_STRING) reduce 171) ((T_STRING_CAST) reduce 171) ((T_SWITCH) reduce 171) ((T_THROW) reduce 171) ((T_TRAIT) reduce 171) ((T_TRAIT_C) reduce 171) ((T_TRY) reduce 171) ((T_UNSET) reduce 171) ((T_UNSET_CAST) reduce 171) ((T_VARIABLE) reduce 171) ((T_VOID_CAST) reduce 171) ((T_WHILE) reduce 171) ((T_YIELD) reduce 171) ((T_YIELD_FROM) reduce 171) (("[") reduce 171) (("`") reduce 171) (("{") reduce 171) (("~") reduce 171)) 23732 ((("!") reduce 177) (("\"") reduce 177) (("$") reduce 177) (("(") reduce 177) (("+") reduce 177) (("-") reduce 177) ((";") reduce 177) (("@") reduce 177) ((T_ABSTRACT) reduce 177) ((T_ARRAY) reduce 177) ((T_ARRAY_CAST) reduce 177) ((T_ATTRIBUTE) reduce 177) ((T_BOOL_CAST) reduce 177) ((T_BREAK) reduce 177) ((T_CATCH) shift) ((T_CLASS) reduce 177) ((T_CLASS_C) reduce 177) ((T_CLONE) reduce 177) ((T_CONSTANT_ENCAPSED_STRING) reduce 177) ((T_CONTINUE) reduce 177) ((T_DEC) reduce 177) ((T_DECLARE) reduce 177) ((T_DIR) reduce 177) ((T_DNUMBER) reduce 177) ((T_DO) reduce 177) ((T_DOUBLE_CAST) reduce 177) ((T_ECHO) reduce 177) ((T_EMPTY) reduce 177) ((T_ENDFOR) reduce 177) ((T_ENUM) reduce 177) ((T_EVAL) reduce 177) ((T_EXIT) reduce 177) ((T_FILE) reduce 177) ((T_FINAL) reduce 177) ((T_FINALLY) shift) ((T_FN) reduce 177) ((T_FOR) reduce 177) ((T_FOREACH) reduce 177) ((T_FUNCTION) reduce 177) ((T_FUNC_C) reduce 177) ((T_GLOBAL) reduce 177) ((T_GOTO) reduce 177) ((T_HALT_COMPILER) reduce 177) ((T_IF) reduce 177) ((T_INC) reduce 177) ((T_INCLUDE) reduce 177) ((T_INCLUDE_ONCE) reduce 177) ((T_INLINE_HTML) reduce 177) ((T_INTERFACE) reduce 177) ((T_INT_CAST) reduce 177) ((T_ISSET) reduce 177) ((T_LINE) reduce 177) ((T_LIST) reduce 177) ((T_LNUMBER) reduce 177) ((T_MATCH) reduce 177) ((T_METHOD_C) reduce 177) ((T_NAME_FULLY_QUALIFIED) reduce 177) ((T_NAME_QUALIFIED) reduce 177) ((T_NAME_RELATIVE) reduce 177) ((T_NEW) reduce 177) ((T_NS_C) reduce 177) ((T_OBJECT_CAST) reduce 177) ((T_PRINT) reduce 177) ((T_PROPERTY_C) reduce 177) ((T_READONLY) reduce 177) ((T_REQUIRE) reduce 177) ((T_REQUIRE_ONCE) reduce 177) ((T_RETURN) reduce 177) ((T_START_HEREDOC) reduce 177) ((T_STATIC) reduce 177) ((T_STRING) reduce 177) ((T_STRING_CAST) reduce 177) ((T_SWITCH) reduce 177) ((T_THROW) reduce 177) ((T_TRAIT) reduce 177) ((T_TRAIT_C) reduce 177) ((T_TRY) reduce 177) ((T_UNSET) reduce 177) ((T_UNSET_CAST) reduce 177) ((T_VARIABLE) reduce 177) ((T_VOID_CAST) reduce 177) ((T_WHILE) reduce 177) ((T_YIELD) reduce 177) ((T_YIELD_FROM) reduce 177) (("[") reduce 177) (("`") reduce 177) (("{") reduce 177) (("~") reduce 177)) 23735 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_EMPTY) reduce 167) ((T_ENDFOR) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_VOID_CAST) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 23738 ((("!") reduce 178) (("\"") reduce 178) (("$") reduce 178) (("(") reduce 178) (("+") reduce 178) (("-") reduce 178) ((";") reduce 178) (("@") reduce 178) ((T_ABSTRACT) reduce 178) ((T_ARRAY) reduce 178) ((T_ARRAY_CAST) reduce 178) ((T_ATTRIBUTE) reduce 178) ((T_BOOL_CAST) reduce 178) ((T_BREAK) reduce 178) ((T_CLASS) reduce 178) ((T_CLASS_C) reduce 178) ((T_CLONE) reduce 178) ((T_CONSTANT_ENCAPSED_STRING) reduce 178) ((T_CONTINUE) reduce 178) ((T_DEC) reduce 178) ((T_DECLARE) reduce 178) ((T_DIR) reduce 178) ((T_DNUMBER) reduce 178) ((T_DO) reduce 178) ((T_DOUBLE_CAST) reduce 178) ((T_ECHO) reduce 178) ((T_EMPTY) reduce 178) ((T_ENDFOR) reduce 178) ((T_ENUM) reduce 178) ((T_EVAL) reduce 178) ((T_EXIT) reduce 178) ((T_FILE) reduce 178) ((T_FINAL) reduce 178) ((T_FN) reduce 178) ((T_FOR) reduce 178) ((T_FOREACH) reduce 178) ((T_FUNCTION) reduce 178) ((T_FUNC_C) reduce 178) ((T_GLOBAL) reduce 178) ((T_GOTO) reduce 178) ((T_HALT_COMPILER) reduce 178) ((T_IF) reduce 178) ((T_INC) reduce 178) ((T_INCLUDE) reduce 178) ((T_INCLUDE_ONCE) reduce 178) ((T_INLINE_HTML) reduce 178) ((T_INTERFACE) reduce 178) ((T_INT_CAST) reduce 178) ((T_ISSET) reduce 178) ((T_LINE) reduce 178) ((T_LIST) reduce 178) ((T_LNUMBER) reduce 178) ((T_MATCH) reduce 178) ((T_METHOD_C) reduce 178) ((T_NAME_FULLY_QUALIFIED) reduce 178) ((T_NAME_QUALIFIED) reduce 178) ((T_NAME_RELATIVE) reduce 178) ((T_NEW) reduce 178) ((T_NS_C) reduce 178) ((T_OBJECT_CAST) reduce 178) ((T_PRINT) reduce 178) ((T_PROPERTY_C) reduce 178) ((T_READONLY) reduce 178) ((T_REQUIRE) reduce 178) ((T_REQUIRE_ONCE) reduce 178) ((T_RETURN) reduce 178) ((T_START_HEREDOC) reduce 178) ((T_STATIC) reduce 178) ((T_STRING) reduce 178) ((T_STRING_CAST) reduce 178) ((T_SWITCH) reduce 178) ((T_THROW) reduce 178) ((T_TRAIT) reduce 178) ((T_TRAIT_C) reduce 178) ((T_TRY) reduce 178) ((T_UNSET) reduce 178) ((T_UNSET_CAST) reduce 178) ((T_VARIABLE) reduce 178) ((T_VOID_CAST) reduce 178) ((T_WHILE) reduce 178) ((T_YIELD) reduce 178) ((T_YIELD_FROM) reduce 178) (("[") reduce 178) (("`") reduce 178) (("{") reduce 178) (("~") reduce 178)) 23745 ((("!") reduce 172) (("\"") reduce 172) (("$") reduce 172) (("(") reduce 172) (("+") reduce 172) (("-") reduce 172) ((";") reduce 172) (("@") reduce 172) ((T_ABSTRACT) reduce 172) ((T_ARRAY) reduce 172) ((T_ARRAY_CAST) reduce 172) ((T_ATTRIBUTE) reduce 172) ((T_BOOL_CAST) reduce 172) ((T_BREAK) reduce 172) ((T_CATCH) reduce 172) ((T_CLASS) reduce 172) ((T_CLASS_C) reduce 172) ((T_CLONE) reduce 172) ((T_CONSTANT_ENCAPSED_STRING) reduce 172) ((T_CONTINUE) reduce 172) ((T_DEC) reduce 172) ((T_DECLARE) reduce 172) ((T_DIR) reduce 172) ((T_DNUMBER) reduce 172) ((T_DO) reduce 172) ((T_DOUBLE_CAST) reduce 172) ((T_ECHO) reduce 172) ((T_EMPTY) reduce 172) ((T_ENDFOR) reduce 172) ((T_ENUM) reduce 172) ((T_EVAL) reduce 172) ((T_EXIT) reduce 172) ((T_FILE) reduce 172) ((T_FINAL) reduce 172) ((T_FINALLY) reduce 172) ((T_FN) reduce 172) ((T_FOR) reduce 172) ((T_FOREACH) reduce 172) ((T_FUNCTION) reduce 172) ((T_FUNC_C) reduce 172) ((T_GLOBAL) reduce 172) ((T_GOTO) reduce 172) ((T_HALT_COMPILER) reduce 172) ((T_IF) reduce 172) ((T_INC) reduce 172) ((T_INCLUDE) reduce 172) ((T_INCLUDE_ONCE) reduce 172) ((T_INLINE_HTML) reduce 172) ((T_INTERFACE) reduce 172) ((T_INT_CAST) reduce 172) ((T_ISSET) reduce 172) ((T_LINE) reduce 172) ((T_LIST) reduce 172) ((T_LNUMBER) reduce 172) ((T_MATCH) reduce 172) ((T_METHOD_C) reduce 172) ((T_NAME_FULLY_QUALIFIED) reduce 172) ((T_NAME_QUALIFIED) reduce 172) ((T_NAME_RELATIVE) reduce 172) ((T_NEW) reduce 172) ((T_NS_C) reduce 172) ((T_OBJECT_CAST) reduce 172) ((T_PRINT) reduce 172) ((T_PROPERTY_C) reduce 172) ((T_READONLY) reduce 172) ((T_REQUIRE) reduce 172) ((T_REQUIRE_ONCE) reduce 172) ((T_RETURN) reduce 172) ((T_START_HEREDOC) reduce 172) ((T_STATIC) reduce 172) ((T_STRING) reduce 172) ((T_STRING_CAST) reduce 172) ((T_SWITCH) reduce 172) ((T_THROW) reduce 172) ((T_TRAIT) reduce 172) ((T_TRAIT_C) reduce 172) ((T_TRY) reduce 172) ((T_UNSET) reduce 172) ((T_UNSET_CAST) reduce 172) ((T_VARIABLE) reduce 172) ((T_VOID_CAST) reduce 172) ((T_WHILE) reduce 172) ((T_YIELD) reduce 172) ((T_YIELD_FROM) reduce 172) (("[") reduce 172) (("`") reduce 172) (("{") reduce 172) (("~") reduce 172)) 23750 ((("!") reduce 200) (("\"") reduce 200) (("$") reduce 200) (("(") reduce 200) (("+") reduce 200) (("-") reduce 200) ((";") reduce 200) (("@") reduce 200) ((T_ABSTRACT) reduce 200) ((T_ARRAY) reduce 200) ((T_ARRAY_CAST) reduce 200) ((T_ATTRIBUTE) reduce 200) ((T_BOOL_CAST) reduce 200) ((T_BREAK) reduce 200) ((T_CLASS) reduce 200) ((T_CLASS_C) reduce 200) ((T_CLONE) reduce 200) ((T_CONSTANT_ENCAPSED_STRING) reduce 200) ((T_CONTINUE) reduce 200) ((T_DEC) reduce 200) ((T_DECLARE) reduce 200) ((T_DIR) reduce 200) ((T_DNUMBER) reduce 200) ((T_DO) reduce 200) ((T_DOUBLE_CAST) reduce 200) ((T_ECHO) reduce 200) ((T_EMPTY) reduce 200) ((T_ENDFOR) reduce 200) ((T_ENUM) reduce 200) ((T_EVAL) reduce 200) ((T_EXIT) reduce 200) ((T_FILE) reduce 200) ((T_FINAL) reduce 200) ((T_FN) reduce 200) ((T_FOR) reduce 200) ((T_FOREACH) reduce 200) ((T_FUNCTION) reduce 200) ((T_FUNC_C) reduce 200) ((T_GLOBAL) reduce 200) ((T_GOTO) reduce 200) ((T_HALT_COMPILER) reduce 200) ((T_IF) reduce 200) ((T_INC) reduce 200) ((T_INCLUDE) reduce 200) ((T_INCLUDE_ONCE) reduce 200) ((T_INLINE_HTML) reduce 200) ((T_INTERFACE) reduce 200) ((T_INT_CAST) reduce 200) ((T_ISSET) reduce 200) ((T_LINE) reduce 200) ((T_LIST) reduce 200) ((T_LNUMBER) reduce 200) ((T_MATCH) reduce 200) ((T_METHOD_C) reduce 200) ((T_NAME_FULLY_QUALIFIED) reduce 200) ((T_NAME_QUALIFIED) reduce 200) ((T_NAME_RELATIVE) reduce 200) ((T_NEW) reduce 200) ((T_NS_C) reduce 200) ((T_OBJECT_CAST) reduce 200) ((T_PRINT) reduce 200) ((T_PROPERTY_C) reduce 200) ((T_READONLY) reduce 200) ((T_REQUIRE) reduce 200) ((T_REQUIRE_ONCE) reduce 200) ((T_RETURN) reduce 200) ((T_START_HEREDOC) reduce 200) ((T_STATIC) reduce 200) ((T_STRING) reduce 200) ((T_STRING_CAST) reduce 200) ((T_SWITCH) reduce 200) ((T_THROW) reduce 200) ((T_TRAIT) reduce 200) ((T_TRAIT_C) reduce 200) ((T_TRY) reduce 200) ((T_UNSET) reduce 200) ((T_UNSET_CAST) reduce 200) ((T_VARIABLE) reduce 200) ((T_VOID_CAST) reduce 200) ((T_WHILE) reduce 200) ((T_YIELD) reduce 200) ((T_YIELD_FROM) reduce 200) (("[") reduce 200) (("`") reduce 200) (("{") reduce 200) (("~") reduce 200)) 23755 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_EMPTY) reduce 153) ((T_ENDFOR) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_VOID_CAST) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 23759 ((("!") reduce 224) (("\"") reduce 224) (("$") reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_EMPTY) reduce 224) ((T_ENDFOR) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_VARIABLE) reduce 224) ((T_VOID_CAST) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("~") reduce 224)) 23761 ((("!") reduce 225) (("\"") reduce 225) (("$") reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_EMPTY) reduce 225) ((T_ENDFOR) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_VARIABLE) reduce 225) ((T_VOID_CAST) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("~") reduce 225)) 23765 ((("!") reduce 226) (("\"") reduce 226) (("$") reduce 226) (("(") reduce 226) (("+") reduce 226) (("-") reduce 226) ((";") reduce 226) (("@") reduce 226) ((T_ABSTRACT) reduce 226) ((T_ARRAY) reduce 226) ((T_ARRAY_CAST) reduce 226) ((T_ATTRIBUTE) reduce 226) ((T_BOOL_CAST) reduce 226) ((T_BREAK) reduce 226) ((T_CLASS) reduce 226) ((T_CLASS_C) reduce 226) ((T_CLONE) reduce 226) ((T_CONSTANT_ENCAPSED_STRING) reduce 226) ((T_CONTINUE) reduce 226) ((T_DEC) reduce 226) ((T_DECLARE) reduce 226) ((T_DIR) reduce 226) ((T_DNUMBER) reduce 226) ((T_DO) reduce 226) ((T_DOUBLE_CAST) reduce 226) ((T_ECHO) reduce 226) ((T_EMPTY) reduce 226) ((T_ENDFOR) reduce 226) ((T_ENUM) reduce 226) ((T_EVAL) reduce 226) ((T_EXIT) reduce 226) ((T_FILE) reduce 226) ((T_FINAL) reduce 226) ((T_FN) reduce 226) ((T_FOR) reduce 226) ((T_FOREACH) reduce 226) ((T_FUNCTION) reduce 226) ((T_FUNC_C) reduce 226) ((T_GLOBAL) reduce 226) ((T_GOTO) reduce 226) ((T_HALT_COMPILER) reduce 226) ((T_IF) reduce 226) ((T_INC) reduce 226) ((T_INCLUDE) reduce 226) ((T_INCLUDE_ONCE) reduce 226) ((T_INLINE_HTML) reduce 226) ((T_INTERFACE) reduce 226) ((T_INT_CAST) reduce 226) ((T_ISSET) reduce 226) ((T_LINE) reduce 226) ((T_LIST) reduce 226) ((T_LNUMBER) reduce 226) ((T_MATCH) reduce 226) ((T_METHOD_C) reduce 226) ((T_NAME_FULLY_QUALIFIED) reduce 226) ((T_NAME_QUALIFIED) reduce 226) ((T_NAME_RELATIVE) reduce 226) ((T_NEW) reduce 226) ((T_NS_C) reduce 226) ((T_OBJECT_CAST) reduce 226) ((T_PRINT) reduce 226) ((T_PROPERTY_C) reduce 226) ((T_READONLY) reduce 226) ((T_REQUIRE) reduce 226) ((T_REQUIRE_ONCE) reduce 226) ((T_RETURN) reduce 226) ((T_START_HEREDOC) reduce 226) ((T_STATIC) reduce 226) ((T_STRING) reduce 226) ((T_STRING_CAST) reduce 226) ((T_SWITCH) reduce 226) ((T_THROW) reduce 226) ((T_TRAIT) reduce 226) ((T_TRAIT_C) reduce 226) ((T_TRY) reduce 226) ((T_UNSET) reduce 226) ((T_UNSET_CAST) reduce 226) ((T_VARIABLE) reduce 226) ((T_VOID_CAST) reduce 226) ((T_WHILE) reduce 226) ((T_YIELD) reduce 226) ((T_YIELD_FROM) reduce 226) (("[") reduce 226) (("`") reduce 226) (("{") reduce 226) (("~") reduce 226)) 23768 ((("!") reduce 227) (("\"") reduce 227) (("$") reduce 227) (("(") reduce 227) (("+") reduce 227) (("-") reduce 227) ((";") reduce 227) (("@") reduce 227) ((T_ABSTRACT) reduce 227) ((T_ARRAY) reduce 227) ((T_ARRAY_CAST) reduce 227) ((T_ATTRIBUTE) reduce 227) ((T_BOOL_CAST) reduce 227) ((T_BREAK) reduce 227) ((T_CLASS) reduce 227) ((T_CLASS_C) reduce 227) ((T_CLONE) reduce 227) ((T_CONSTANT_ENCAPSED_STRING) reduce 227) ((T_CONTINUE) reduce 227) ((T_DEC) reduce 227) ((T_DECLARE) reduce 227) ((T_DIR) reduce 227) ((T_DNUMBER) reduce 227) ((T_DO) reduce 227) ((T_DOUBLE_CAST) reduce 227) ((T_ECHO) reduce 227) ((T_EMPTY) reduce 227) ((T_ENDFOR) reduce 227) ((T_ENUM) reduce 227) ((T_EVAL) reduce 227) ((T_EXIT) reduce 227) ((T_FILE) reduce 227) ((T_FINAL) reduce 227) ((T_FN) reduce 227) ((T_FOR) reduce 227) ((T_FOREACH) reduce 227) ((T_FUNCTION) reduce 227) ((T_FUNC_C) reduce 227) ((T_GLOBAL) reduce 227) ((T_GOTO) reduce 227) ((T_HALT_COMPILER) reduce 227) ((T_IF) reduce 227) ((T_INC) reduce 227) ((T_INCLUDE) reduce 227) ((T_INCLUDE_ONCE) reduce 227) ((T_INLINE_HTML) reduce 227) ((T_INTERFACE) reduce 227) ((T_INT_CAST) reduce 227) ((T_ISSET) reduce 227) ((T_LINE) reduce 227) ((T_LIST) reduce 227) ((T_LNUMBER) reduce 227) ((T_MATCH) reduce 227) ((T_METHOD_C) reduce 227) ((T_NAME_FULLY_QUALIFIED) reduce 227) ((T_NAME_QUALIFIED) reduce 227) ((T_NAME_RELATIVE) reduce 227) ((T_NEW) reduce 227) ((T_NS_C) reduce 227) ((T_OBJECT_CAST) reduce 227) ((T_PRINT) reduce 227) ((T_PROPERTY_C) reduce 227) ((T_READONLY) reduce 227) ((T_REQUIRE) reduce 227) ((T_REQUIRE_ONCE) reduce 227) ((T_RETURN) reduce 227) ((T_START_HEREDOC) reduce 227) ((T_STATIC) reduce 227) ((T_STRING) reduce 227) ((T_STRING_CAST) reduce 227) ((T_SWITCH) reduce 227) ((T_THROW) reduce 227) ((T_TRAIT) reduce 227) ((T_TRAIT_C) reduce 227) ((T_TRY) reduce 227) ((T_UNSET) reduce 227) ((T_UNSET_CAST) reduce 227) ((T_VARIABLE) reduce 227) ((T_VOID_CAST) reduce 227) ((T_WHILE) reduce 227) ((T_YIELD) reduce 227) ((T_YIELD_FROM) reduce 227) (("[") reduce 227) (("`") reduce 227) (("{") reduce 227) (("~") reduce 227)) 23769 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_EMPTY) reduce 169) ((T_ENDFOR) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_VOID_CAST) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 23771 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_EMPTY) reduce 158) ((T_ENDFOR) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_VOID_CAST) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 23773 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_EMPTY) reduce 156) ((T_ENDFOR) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_VOID_CAST) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 23779 ((("!") reduce 201) (("\"") reduce 201) (("$") reduce 201) (("(") reduce 201) (("+") reduce 201) (("-") reduce 201) ((";") reduce 201) (("@") reduce 201) ((T_ABSTRACT) reduce 201) ((T_ARRAY) reduce 201) ((T_ARRAY_CAST) reduce 201) ((T_ATTRIBUTE) reduce 201) ((T_BOOL_CAST) reduce 201) ((T_BREAK) reduce 201) ((T_CLASS) reduce 201) ((T_CLASS_C) reduce 201) ((T_CLONE) reduce 201) ((T_CONSTANT_ENCAPSED_STRING) reduce 201) ((T_CONTINUE) reduce 201) ((T_DEC) reduce 201) ((T_DECLARE) reduce 201) ((T_DIR) reduce 201) ((T_DNUMBER) reduce 201) ((T_DO) reduce 201) ((T_DOUBLE_CAST) reduce 201) ((T_ECHO) reduce 201) ((T_EMPTY) reduce 201) ((T_ENDFOR) reduce 201) ((T_ENUM) reduce 201) ((T_EVAL) reduce 201) ((T_EXIT) reduce 201) ((T_FILE) reduce 201) ((T_FINAL) reduce 201) ((T_FN) reduce 201) ((T_FOR) reduce 201) ((T_FOREACH) reduce 201) ((T_FUNCTION) reduce 201) ((T_FUNC_C) reduce 201) ((T_GLOBAL) reduce 201) ((T_GOTO) reduce 201) ((T_HALT_COMPILER) reduce 201) ((T_IF) reduce 201) ((T_INC) reduce 201) ((T_INCLUDE) reduce 201) ((T_INCLUDE_ONCE) reduce 201) ((T_INLINE_HTML) reduce 201) ((T_INTERFACE) reduce 201) ((T_INT_CAST) reduce 201) ((T_ISSET) reduce 201) ((T_LINE) reduce 201) ((T_LIST) reduce 201) ((T_LNUMBER) reduce 201) ((T_MATCH) reduce 201) ((T_METHOD_C) reduce 201) ((T_NAME_FULLY_QUALIFIED) reduce 201) ((T_NAME_QUALIFIED) reduce 201) ((T_NAME_RELATIVE) reduce 201) ((T_NEW) reduce 201) ((T_NS_C) reduce 201) ((T_OBJECT_CAST) reduce 201) ((T_PRINT) reduce 201) ((T_PROPERTY_C) reduce 201) ((T_READONLY) reduce 201) ((T_REQUIRE) reduce 201) ((T_REQUIRE_ONCE) reduce 201) ((T_RETURN) reduce 201) ((T_START_HEREDOC) reduce 201) ((T_STATIC) reduce 201) ((T_STRING) reduce 201) ((T_STRING_CAST) reduce 201) ((T_SWITCH) reduce 201) ((T_THROW) reduce 201) ((T_TRAIT) reduce 201) ((T_TRAIT_C) reduce 201) ((T_TRY) reduce 201) ((T_UNSET) reduce 201) ((T_UNSET_CAST) reduce 201) ((T_VARIABLE) reduce 201) ((T_VOID_CAST) reduce 201) ((T_WHILE) reduce 201) ((T_YIELD) reduce 201) ((T_YIELD_FROM) reduce 201) (("[") reduce 201) (("`") reduce 201) (("{") reduce 201) (("~") reduce 201)) 23783 ((("!") reduce 244) (("\"") reduce 244) (("$") reduce 244) (("(") reduce 244) (("+") reduce 244) (("-") reduce 244) ((";") reduce 244) (("@") reduce 244) ((T_ABSTRACT) reduce 244) ((T_ARRAY) reduce 244) ((T_ARRAY_CAST) reduce 244) ((T_ATTRIBUTE) reduce 244) ((T_BOOL_CAST) reduce 244) ((T_BREAK) reduce 244) ((T_CLASS) reduce 244) ((T_CLASS_C) reduce 244) ((T_CLONE) reduce 244) ((T_CONSTANT_ENCAPSED_STRING) reduce 244) ((T_CONTINUE) reduce 244) ((T_DEC) reduce 244) ((T_DECLARE) reduce 244) ((T_DIR) reduce 244) ((T_DNUMBER) reduce 244) ((T_DO) reduce 244) ((T_DOUBLE_CAST) reduce 244) ((T_ECHO) reduce 244) ((T_ELSE) reduce 244) ((T_ELSEIF) reduce 244) ((T_EMPTY) reduce 244) ((T_ENDFOR) reduce 244) ((T_ENUM) reduce 244) ((T_EVAL) reduce 244) ((T_EXIT) reduce 244) ((T_FILE) reduce 244) ((T_FINAL) reduce 244) ((T_FN) reduce 244) ((T_FOR) reduce 244) ((T_FOREACH) reduce 244) ((T_FUNCTION) reduce 244) ((T_FUNC_C) reduce 244) ((T_GLOBAL) reduce 244) ((T_GOTO) reduce 244) ((T_HALT_COMPILER) reduce 244) ((T_IF) reduce 244) ((T_INC) reduce 244) ((T_INCLUDE) reduce 244) ((T_INCLUDE_ONCE) reduce 244) ((T_INLINE_HTML) reduce 244) ((T_INTERFACE) reduce 244) ((T_INT_CAST) reduce 244) ((T_ISSET) reduce 244) ((T_LINE) reduce 244) ((T_LIST) reduce 244) ((T_LNUMBER) reduce 244) ((T_MATCH) reduce 244) ((T_METHOD_C) reduce 244) ((T_NAME_FULLY_QUALIFIED) reduce 244) ((T_NAME_QUALIFIED) reduce 244) ((T_NAME_RELATIVE) reduce 244) ((T_NEW) reduce 244) ((T_NS_C) reduce 244) ((T_OBJECT_CAST) reduce 244) ((T_PRINT) reduce 244) ((T_PROPERTY_C) reduce 244) ((T_READONLY) reduce 244) ((T_REQUIRE) reduce 244) ((T_REQUIRE_ONCE) reduce 244) ((T_RETURN) reduce 244) ((T_START_HEREDOC) reduce 244) ((T_STATIC) reduce 244) ((T_STRING) reduce 244) ((T_STRING_CAST) reduce 244) ((T_SWITCH) reduce 244) ((T_THROW) reduce 244) ((T_TRAIT) reduce 244) ((T_TRAIT_C) reduce 244) ((T_TRY) reduce 244) ((T_UNSET) reduce 244) ((T_UNSET_CAST) reduce 244) ((T_VARIABLE) reduce 244) ((T_VOID_CAST) reduce 244) ((T_WHILE) reduce 244) ((T_YIELD) reduce 244) ((T_YIELD_FROM) reduce 244) (("[") reduce 244) (("`") reduce 244) (("{") reduce 244) (("~") reduce 244)) 23786 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_EMPTY) reduce 146) ((T_ENDFOR) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_PROPERTY_C) reduce 146) ((T_READONLY) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_VOID_CAST) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("~") reduce 146)) 23788 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_EMPTY) reduce 168) ((T_ENDFOR) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_VOID_CAST) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 23790 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_EMPTY) reduce 157) ((T_ENDFOR) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_VOID_CAST) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 23800 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_EMPTY) reduce 164) ((T_ENDFOR) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_VOID_CAST) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 23801 ((("!") reduce 220) (("\"") reduce 220) (("$") reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_EMPTY) reduce 220) ((T_ENDFOR) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_VARIABLE) reduce 220) ((T_VOID_CAST) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("~") reduce 220)) 23804 ((("!") reduce 221) (("\"") reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_EMPTY) reduce 221) ((T_ENDFOR) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_VARIABLE) reduce 221) ((T_VOID_CAST) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("~") reduce 221)) 23805 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_EMPTY) reduce 163) ((T_ENDFOR) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_VOID_CAST) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 23814 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_EMPTY) reduce 152) ((T_ENDFOR) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_VOID_CAST) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 23815 ((("!") reduce 218) (("\"") reduce 218) (("$") reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_EMPTY) reduce 218) ((T_ENDFOR) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_VARIABLE) reduce 218) ((T_VOID_CAST) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("~") reduce 218)) 23818 ((("!") reduce 219) (("\"") reduce 219) (("$") reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_EMPTY) reduce 219) ((T_ENDFOR) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_VARIABLE) reduce 219) ((T_VOID_CAST) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("~") reduce 219)) 23825 ((("!") reduce 202) (("\"") reduce 202) (("$") reduce 202) (("(") reduce 202) (("+") reduce 202) (("-") reduce 202) ((";") reduce 202) (("@") reduce 202) ((T_ABSTRACT) reduce 202) ((T_ARRAY) reduce 202) ((T_ARRAY_CAST) reduce 202) ((T_ATTRIBUTE) reduce 202) ((T_BOOL_CAST) reduce 202) ((T_BREAK) reduce 202) ((T_CLASS) reduce 202) ((T_CLASS_C) reduce 202) ((T_CLONE) reduce 202) ((T_CONSTANT_ENCAPSED_STRING) reduce 202) ((T_CONTINUE) reduce 202) ((T_DEC) reduce 202) ((T_DECLARE) reduce 202) ((T_DIR) reduce 202) ((T_DNUMBER) reduce 202) ((T_DO) reduce 202) ((T_DOUBLE_CAST) reduce 202) ((T_ECHO) reduce 202) ((T_EMPTY) reduce 202) ((T_ENDFOR) reduce 202) ((T_ENUM) reduce 202) ((T_EVAL) reduce 202) ((T_EXIT) reduce 202) ((T_FILE) reduce 202) ((T_FINAL) reduce 202) ((T_FN) reduce 202) ((T_FOR) reduce 202) ((T_FOREACH) reduce 202) ((T_FUNCTION) reduce 202) ((T_FUNC_C) reduce 202) ((T_GLOBAL) reduce 202) ((T_GOTO) reduce 202) ((T_HALT_COMPILER) reduce 202) ((T_IF) reduce 202) ((T_INC) reduce 202) ((T_INCLUDE) reduce 202) ((T_INCLUDE_ONCE) reduce 202) ((T_INLINE_HTML) reduce 202) ((T_INTERFACE) reduce 202) ((T_INT_CAST) reduce 202) ((T_ISSET) reduce 202) ((T_LINE) reduce 202) ((T_LIST) reduce 202) ((T_LNUMBER) reduce 202) ((T_MATCH) reduce 202) ((T_METHOD_C) reduce 202) ((T_NAME_FULLY_QUALIFIED) reduce 202) ((T_NAME_QUALIFIED) reduce 202) ((T_NAME_RELATIVE) reduce 202) ((T_NEW) reduce 202) ((T_NS_C) reduce 202) ((T_OBJECT_CAST) reduce 202) ((T_PRINT) reduce 202) ((T_PROPERTY_C) reduce 202) ((T_READONLY) reduce 202) ((T_REQUIRE) reduce 202) ((T_REQUIRE_ONCE) reduce 202) ((T_RETURN) reduce 202) ((T_START_HEREDOC) reduce 202) ((T_STATIC) reduce 202) ((T_STRING) reduce 202) ((T_STRING_CAST) reduce 202) ((T_SWITCH) reduce 202) ((T_THROW) reduce 202) ((T_TRAIT) reduce 202) ((T_TRAIT_C) reduce 202) ((T_TRY) reduce 202) ((T_UNSET) reduce 202) ((T_UNSET_CAST) reduce 202) ((T_VARIABLE) reduce 202) ((T_VOID_CAST) reduce 202) ((T_WHILE) reduce 202) ((T_YIELD) reduce 202) ((T_YIELD_FROM) reduce 202) (("[") reduce 202) (("`") reduce 202) (("{") reduce 202) (("~") reduce 202)) 23826 ((("!") reduce 219) (("\"") reduce 219) (("$") reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_ELSE) reduce 219) ((T_ELSEIF) reduce 219) ((T_EMPTY) reduce 219) ((T_ENDFOREACH) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_VARIABLE) reduce 219) ((T_VOID_CAST) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("~") reduce 219)) 23828 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_EMPTY) reduce 159) ((T_ENDFOR) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_VOID_CAST) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 23834 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_EMPTY) reduce 151) ((T_ENDFOR) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_VOID_CAST) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 23839 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_EMPTY) reduce 165) ((T_ENDFOR) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_VOID_CAST) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 23840 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_EMPTY) reduce 222) ((T_ENDFOR) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_VARIABLE) reduce 222) ((T_VOID_CAST) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("~") reduce 222)) 23843 ((("!") reduce 223) (("\"") reduce 223) (("$") reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_EMPTY) reduce 223) ((T_ENDFOR) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_VARIABLE) reduce 223) ((T_VOID_CAST) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("~") reduce 223)) 23845 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_EMPTY) reduce 155) ((T_ENDFOR) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_VOID_CAST) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 23852 ((("!") reduce 190) (("\"") reduce 190) (("$") reduce 190) (("(") reduce 190) (("+") reduce 190) (("-") reduce 190) ((";") reduce 190) (("@") reduce 190) ((T_ABSTRACT) reduce 190) ((T_ARRAY) reduce 190) ((T_ARRAY_CAST) reduce 190) ((T_ATTRIBUTE) reduce 190) ((T_BOOL_CAST) reduce 190) ((T_BREAK) reduce 190) ((T_CLASS) reduce 190) ((T_CLASS_C) reduce 190) ((T_CLONE) reduce 190) ((T_CONSTANT_ENCAPSED_STRING) reduce 190) ((T_CONTINUE) reduce 190) ((T_DEC) reduce 190) ((T_DECLARE) reduce 190) ((T_DIR) reduce 190) ((T_DNUMBER) reduce 190) ((T_DO) reduce 190) ((T_DOUBLE_CAST) reduce 190) ((T_ECHO) reduce 190) ((T_EMPTY) reduce 190) ((T_ENDFOR) reduce 190) ((T_ENUM) reduce 190) ((T_EVAL) reduce 190) ((T_EXIT) reduce 190) ((T_FILE) reduce 190) ((T_FINAL) reduce 190) ((T_FN) reduce 190) ((T_FOR) reduce 190) ((T_FOREACH) reduce 190) ((T_FUNCTION) reduce 190) ((T_FUNC_C) reduce 190) ((T_GLOBAL) reduce 190) ((T_GOTO) reduce 190) ((T_HALT_COMPILER) reduce 190) ((T_IF) reduce 190) ((T_INC) reduce 190) ((T_INCLUDE) reduce 190) ((T_INCLUDE_ONCE) reduce 190) ((T_INLINE_HTML) reduce 190) ((T_INTERFACE) reduce 190) ((T_INT_CAST) reduce 190) ((T_ISSET) reduce 190) ((T_LINE) reduce 190) ((T_LIST) reduce 190) ((T_LNUMBER) reduce 190) ((T_MATCH) reduce 190) ((T_METHOD_C) reduce 190) ((T_NAME_FULLY_QUALIFIED) reduce 190) ((T_NAME_QUALIFIED) reduce 190) ((T_NAME_RELATIVE) reduce 190) ((T_NEW) reduce 190) ((T_NS_C) reduce 190) ((T_OBJECT_CAST) reduce 190) ((T_PRINT) reduce 190) ((T_PROPERTY_C) reduce 190) ((T_READONLY) reduce 190) ((T_REQUIRE) reduce 190) ((T_REQUIRE_ONCE) reduce 190) ((T_RETURN) reduce 190) ((T_START_HEREDOC) reduce 190) ((T_STATIC) reduce 190) ((T_STRING) reduce 190) ((T_STRING_CAST) reduce 190) ((T_SWITCH) reduce 190) ((T_THROW) reduce 190) ((T_TRAIT) reduce 190) ((T_TRAIT_C) reduce 190) ((T_TRY) reduce 190) ((T_UNSET) reduce 190) ((T_UNSET_CAST) reduce 190) ((T_VARIABLE) reduce 190) ((T_VOID_CAST) reduce 190) ((T_WHILE) reduce 190) ((T_YIELD) reduce 190) ((T_YIELD_FROM) reduce 190) (("[") reduce 190) (("`") reduce 190) (("{") reduce 190) (("~") reduce 190)) 23854 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_EMPTY) reduce 154) ((T_ENDFOR) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_VOID_CAST) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 23855 ((("%") shift) ((")") reduce 396) (("*") shift) (("+") shift) ((",") reduce 396) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23856 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 396) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23857 ((("%") shift) (("*") shift) (("+") shift) ((",") reduce 396) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 396) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 23859 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_ELSE) reduce 159) ((T_ELSEIF) reduce 159) ((T_EMPTY) reduce 159) ((T_ENDFOREACH) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_VOID_CAST) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 23865 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_ELSE) reduce 151) ((T_ELSEIF) reduce 151) ((T_EMPTY) reduce 151) ((T_ENDFOREACH) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_VOID_CAST) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 23870 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_ELSE) reduce 165) ((T_ELSEIF) reduce 165) ((T_EMPTY) reduce 165) ((T_ENDFOREACH) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_VOID_CAST) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 23871 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_ELSE) reduce 222) ((T_ELSEIF) reduce 222) ((T_EMPTY) reduce 222) ((T_ENDFOREACH) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_VARIABLE) reduce 222) ((T_VOID_CAST) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("~") reduce 222)) 23874 ((("!") reduce 223) (("\"") reduce 223) (("$") reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_ELSE) reduce 223) ((T_ELSEIF) reduce 223) ((T_EMPTY) reduce 223) ((T_ENDFOREACH) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_VARIABLE) reduce 223) ((T_VOID_CAST) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("~") reduce 223)) 23876 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_ELSE) reduce 155) ((T_ELSEIF) reduce 155) ((T_EMPTY) reduce 155) ((T_ENDFOREACH) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_VOID_CAST) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 23878 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_ELSE) reduce 154) ((T_ELSEIF) reduce 154) ((T_EMPTY) reduce 154) ((T_ENDFOREACH) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_VOID_CAST) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 23879 ((("!") reduce 247) (("\"") reduce 247) (("$") reduce 247) (("(") reduce 247) (("+") reduce 247) (("-") reduce 247) ((";") reduce 247) (("@") reduce 247) ((T_ABSTRACT) reduce 247) ((T_ARRAY) reduce 247) ((T_ARRAY_CAST) reduce 247) ((T_ATTRIBUTE) reduce 247) ((T_BOOL_CAST) reduce 247) ((T_BREAK) reduce 247) ((T_CLASS) reduce 247) ((T_CLASS_C) reduce 247) ((T_CLONE) reduce 247) ((T_CONSTANT_ENCAPSED_STRING) reduce 247) ((T_CONTINUE) reduce 247) ((T_DEC) reduce 247) ((T_DECLARE) reduce 247) ((T_DIR) reduce 247) ((T_DNUMBER) reduce 247) ((T_DO) reduce 247) ((T_DOUBLE_CAST) reduce 247) ((T_ECHO) reduce 247) ((T_EMPTY) reduce 247) ((T_ENDFOREACH) reduce 247) ((T_ENUM) reduce 247) ((T_EVAL) reduce 247) ((T_EXIT) reduce 247) ((T_FILE) reduce 247) ((T_FINAL) reduce 247) ((T_FN) reduce 247) ((T_FOR) reduce 247) ((T_FOREACH) reduce 247) ((T_FUNCTION) reduce 247) ((T_FUNC_C) reduce 247) ((T_GLOBAL) reduce 247) ((T_GOTO) reduce 247) ((T_HALT_COMPILER) reduce 247) ((T_IF) reduce 247) ((T_INC) reduce 247) ((T_INCLUDE) reduce 247) ((T_INCLUDE_ONCE) reduce 247) ((T_INLINE_HTML) reduce 247) ((T_INTERFACE) reduce 247) ((T_INT_CAST) reduce 247) ((T_ISSET) reduce 247) ((T_LINE) reduce 247) ((T_LIST) reduce 247) ((T_LNUMBER) reduce 247) ((T_MATCH) reduce 247) ((T_METHOD_C) reduce 247) ((T_NAME_FULLY_QUALIFIED) reduce 247) ((T_NAME_QUALIFIED) reduce 247) ((T_NAME_RELATIVE) reduce 247) ((T_NEW) reduce 247) ((T_NS_C) reduce 247) ((T_OBJECT_CAST) reduce 247) ((T_PRINT) reduce 247) ((T_PROPERTY_C) reduce 247) ((T_READONLY) reduce 247) ((T_REQUIRE) reduce 247) ((T_REQUIRE_ONCE) reduce 247) ((T_RETURN) reduce 247) ((T_START_HEREDOC) reduce 247) ((T_STATIC) reduce 247) ((T_STRING) reduce 247) ((T_STRING_CAST) reduce 247) ((T_SWITCH) reduce 247) ((T_THROW) reduce 247) ((T_TRAIT) reduce 247) ((T_TRAIT_C) reduce 247) ((T_TRY) reduce 247) ((T_UNSET) reduce 247) ((T_UNSET_CAST) reduce 247) ((T_VARIABLE) reduce 247) ((T_VOID_CAST) reduce 247) ((T_WHILE) reduce 247) ((T_YIELD) reduce 247) ((T_YIELD_FROM) reduce 247) (("[") reduce 247) (("`") reduce 247) (("{") reduce 247) (("~") reduce 247)) 23890 ((("!") reduce 494) (("\"") reduce 494) (("$") reduce 494) (("(") reduce 494) (("+") reduce 494) (("-") reduce 494) ((";") reduce 494) (("@") reduce 494) ((T_ABSTRACT) reduce 494) ((T_ARRAY) reduce 494) ((T_ARRAY_CAST) reduce 494) ((T_ATTRIBUTE) reduce 494) ((T_BOOL_CAST) reduce 494) ((T_BREAK) reduce 494) ((T_CLASS) reduce 494) ((T_CLASS_C) reduce 494) ((T_CLONE) reduce 494) ((T_CONSTANT_ENCAPSED_STRING) reduce 494) ((T_CONTINUE) reduce 494) ((T_DEC) reduce 494) ((T_DECLARE) reduce 494) ((T_DIR) reduce 494) ((T_DNUMBER) reduce 494) ((T_DO) reduce 494) ((T_DOUBLE_CAST) reduce 494) ((T_ECHO) reduce 494) ((T_EMPTY) reduce 494) ((T_ENDFOREACH) reduce 494) ((T_ENUM) reduce 494) ((T_EVAL) reduce 494) ((T_EXIT) reduce 494) ((T_FILE) reduce 494) ((T_FINAL) reduce 494) ((T_FN) reduce 494) ((T_FOR) reduce 494) ((T_FOREACH) reduce 494) ((T_FUNCTION) reduce 494) ((T_FUNC_C) reduce 494) ((T_GLOBAL) reduce 494) ((T_GOTO) reduce 494) ((T_HALT_COMPILER) reduce 494) ((T_IF) reduce 494) ((T_INC) reduce 494) ((T_INCLUDE) reduce 494) ((T_INCLUDE_ONCE) reduce 494) ((T_INLINE_HTML) reduce 494) ((T_INTERFACE) reduce 494) ((T_INT_CAST) reduce 494) ((T_ISSET) reduce 494) ((T_LINE) reduce 494) ((T_LIST) reduce 494) ((T_LNUMBER) reduce 494) ((T_MATCH) reduce 494) ((T_METHOD_C) reduce 494) ((T_NAME_FULLY_QUALIFIED) reduce 494) ((T_NAME_QUALIFIED) reduce 494) ((T_NAME_RELATIVE) reduce 494) ((T_NEW) reduce 494) ((T_NS_C) reduce 494) ((T_OBJECT_CAST) reduce 494) ((T_PRINT) reduce 494) ((T_PROPERTY_C) reduce 494) ((T_READONLY) reduce 494) ((T_REQUIRE) reduce 494) ((T_REQUIRE_ONCE) reduce 494) ((T_RETURN) reduce 494) ((T_START_HEREDOC) reduce 494) ((T_STATIC) reduce 494) ((T_STRING) reduce 494) ((T_STRING_CAST) reduce 494) ((T_SWITCH) reduce 494) ((T_THROW) reduce 494) ((T_TRAIT) reduce 494) ((T_TRAIT_C) reduce 494) ((T_TRY) reduce 494) ((T_UNSET) reduce 494) ((T_UNSET_CAST) reduce 494) ((T_VARIABLE) reduce 494) ((T_VOID_CAST) reduce 494) ((T_WHILE) reduce 494) ((T_YIELD) reduce 494) ((T_YIELD_FROM) reduce 494) (("[") reduce 494) (("`") reduce 494) (("{") reduce 494) (("~") reduce 494)) 23891 ((("!") reduce 184) (("\"") reduce 184) (("$") reduce 184) (("(") reduce 184) (("+") reduce 184) (("-") reduce 184) ((";") reduce 184) (("@") reduce 184) ((T_ABSTRACT) reduce 184) ((T_ARRAY) reduce 184) ((T_ARRAY_CAST) reduce 184) ((T_ATTRIBUTE) reduce 184) ((T_BOOL_CAST) reduce 184) ((T_BREAK) reduce 184) ((T_CLASS) reduce 184) ((T_CLASS_C) reduce 184) ((T_CLONE) reduce 184) ((T_CONSTANT_ENCAPSED_STRING) reduce 184) ((T_CONTINUE) reduce 184) ((T_DEC) reduce 184) ((T_DECLARE) reduce 184) ((T_DIR) reduce 184) ((T_DNUMBER) reduce 184) ((T_DO) reduce 184) ((T_DOUBLE_CAST) reduce 184) ((T_ECHO) reduce 184) ((T_EMPTY) reduce 184) ((T_ENDFOREACH) reduce 184) ((T_ENUM) reduce 184) ((T_EVAL) reduce 184) ((T_EXIT) reduce 184) ((T_FILE) reduce 184) ((T_FINAL) reduce 184) ((T_FN) reduce 184) ((T_FOR) reduce 184) ((T_FOREACH) reduce 184) ((T_FUNCTION) reduce 184) ((T_FUNC_C) reduce 184) ((T_GLOBAL) reduce 184) ((T_GOTO) reduce 184) ((T_HALT_COMPILER) reduce 184) ((T_IF) reduce 184) ((T_INC) reduce 184) ((T_INCLUDE) reduce 184) ((T_INCLUDE_ONCE) reduce 184) ((T_INLINE_HTML) reduce 184) ((T_INTERFACE) reduce 184) ((T_INT_CAST) reduce 184) ((T_ISSET) reduce 184) ((T_LINE) reduce 184) ((T_LIST) reduce 184) ((T_LNUMBER) reduce 184) ((T_MATCH) reduce 184) ((T_METHOD_C) reduce 184) ((T_NAME_FULLY_QUALIFIED) reduce 184) ((T_NAME_QUALIFIED) reduce 184) ((T_NAME_RELATIVE) reduce 184) ((T_NEW) reduce 184) ((T_NS_C) reduce 184) ((T_OBJECT_CAST) reduce 184) ((T_PRINT) reduce 184) ((T_PROPERTY_C) reduce 184) ((T_READONLY) reduce 184) ((T_REQUIRE) reduce 184) ((T_REQUIRE_ONCE) reduce 184) ((T_RETURN) reduce 184) ((T_START_HEREDOC) reduce 184) ((T_STATIC) reduce 184) ((T_STRING) reduce 184) ((T_STRING_CAST) reduce 184) ((T_SWITCH) reduce 184) ((T_THROW) reduce 184) ((T_TRAIT) reduce 184) ((T_TRAIT_C) reduce 184) ((T_TRY) reduce 184) ((T_UNSET) reduce 184) ((T_UNSET_CAST) reduce 184) ((T_VARIABLE) reduce 184) ((T_VOID_CAST) reduce 184) ((T_WHILE) reduce 184) ((T_YIELD) reduce 184) ((T_YIELD_FROM) reduce 184) (("[") reduce 184) (("`") reduce 184) (("{") reduce 184) (("~") reduce 184)) 23892 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_EMPTY) reduce 161) ((T_ENDFOREACH) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_VOID_CAST) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 23900 ((("!") reduce 189) (("\"") reduce 189) (("$") reduce 189) (("(") reduce 189) (("+") reduce 189) (("-") reduce 189) ((";") reduce 189) (("@") reduce 189) ((T_ABSTRACT) reduce 189) ((T_ARRAY) reduce 189) ((T_ARRAY_CAST) reduce 189) ((T_ATTRIBUTE) reduce 189) ((T_BOOL_CAST) reduce 189) ((T_BREAK) reduce 189) ((T_CLASS) reduce 189) ((T_CLASS_C) reduce 189) ((T_CLONE) reduce 189) ((T_CONSTANT_ENCAPSED_STRING) reduce 189) ((T_CONTINUE) reduce 189) ((T_DEC) reduce 189) ((T_DECLARE) reduce 189) ((T_DIR) reduce 189) ((T_DNUMBER) reduce 189) ((T_DO) reduce 189) ((T_DOUBLE_CAST) reduce 189) ((T_ECHO) reduce 189) ((T_EMPTY) reduce 189) ((T_ENDFOREACH) reduce 189) ((T_ENUM) reduce 189) ((T_EVAL) reduce 189) ((T_EXIT) reduce 189) ((T_FILE) reduce 189) ((T_FINAL) reduce 189) ((T_FN) reduce 189) ((T_FOR) reduce 189) ((T_FOREACH) reduce 189) ((T_FUNCTION) reduce 189) ((T_FUNC_C) reduce 189) ((T_GLOBAL) reduce 189) ((T_GOTO) reduce 189) ((T_HALT_COMPILER) reduce 189) ((T_IF) reduce 189) ((T_INC) reduce 189) ((T_INCLUDE) reduce 189) ((T_INCLUDE_ONCE) reduce 189) ((T_INLINE_HTML) reduce 189) ((T_INTERFACE) reduce 189) ((T_INT_CAST) reduce 189) ((T_ISSET) reduce 189) ((T_LINE) reduce 189) ((T_LIST) reduce 189) ((T_LNUMBER) reduce 189) ((T_MATCH) reduce 189) ((T_METHOD_C) reduce 189) ((T_NAME_FULLY_QUALIFIED) reduce 189) ((T_NAME_QUALIFIED) reduce 189) ((T_NAME_RELATIVE) reduce 189) ((T_NEW) reduce 189) ((T_NS_C) reduce 189) ((T_OBJECT_CAST) reduce 189) ((T_PRINT) reduce 189) ((T_PROPERTY_C) reduce 189) ((T_READONLY) reduce 189) ((T_REQUIRE) reduce 189) ((T_REQUIRE_ONCE) reduce 189) ((T_RETURN) reduce 189) ((T_START_HEREDOC) reduce 189) ((T_STATIC) reduce 189) ((T_STRING) reduce 189) ((T_STRING_CAST) reduce 189) ((T_SWITCH) reduce 189) ((T_THROW) reduce 189) ((T_TRAIT) reduce 189) ((T_TRAIT_C) reduce 189) ((T_TRY) reduce 189) ((T_UNSET) reduce 189) ((T_UNSET_CAST) reduce 189) ((T_VARIABLE) reduce 189) ((T_VOID_CAST) reduce 189) ((T_WHILE) reduce 189) ((T_YIELD) reduce 189) ((T_YIELD_FROM) reduce 189) (("[") reduce 189) (("`") reduce 189) (("{") reduce 189) (("~") reduce 189)) 23901 ((("!") reduce 145) (("\"") reduce 145) (("$") reduce 145) (("(") reduce 145) (("+") reduce 145) (("-") reduce 145) ((";") reduce 145) (("@") reduce 145) ((T_ABSTRACT) reduce 145) ((T_ARRAY) reduce 145) ((T_ARRAY_CAST) reduce 145) ((T_ATTRIBUTE) reduce 145) ((T_BOOL_CAST) reduce 145) ((T_BREAK) reduce 145) ((T_CLASS) reduce 145) ((T_CLASS_C) reduce 145) ((T_CLONE) reduce 145) ((T_CONSTANT_ENCAPSED_STRING) reduce 145) ((T_CONTINUE) reduce 145) ((T_DEC) reduce 145) ((T_DECLARE) reduce 145) ((T_DIR) reduce 145) ((T_DNUMBER) reduce 145) ((T_DO) reduce 145) ((T_DOUBLE_CAST) reduce 145) ((T_ECHO) reduce 145) ((T_EMPTY) reduce 145) ((T_ENDFOREACH) reduce 145) ((T_ENUM) reduce 145) ((T_EVAL) reduce 145) ((T_EXIT) reduce 145) ((T_FILE) reduce 145) ((T_FINAL) reduce 145) ((T_FN) reduce 145) ((T_FOR) reduce 145) ((T_FOREACH) reduce 145) ((T_FUNCTION) reduce 145) ((T_FUNC_C) reduce 145) ((T_GLOBAL) reduce 145) ((T_GOTO) reduce 145) ((T_HALT_COMPILER) reduce 145) ((T_IF) reduce 145) ((T_INC) reduce 145) ((T_INCLUDE) reduce 145) ((T_INCLUDE_ONCE) reduce 145) ((T_INLINE_HTML) reduce 145) ((T_INTERFACE) reduce 145) ((T_INT_CAST) reduce 145) ((T_ISSET) reduce 145) ((T_LINE) reduce 145) ((T_LIST) reduce 145) ((T_LNUMBER) reduce 145) ((T_MATCH) reduce 145) ((T_METHOD_C) reduce 145) ((T_NAME_FULLY_QUALIFIED) reduce 145) ((T_NAME_QUALIFIED) reduce 145) ((T_NAME_RELATIVE) reduce 145) ((T_NEW) reduce 145) ((T_NS_C) reduce 145) ((T_OBJECT_CAST) reduce 145) ((T_PRINT) reduce 145) ((T_PROPERTY_C) reduce 145) ((T_READONLY) reduce 145) ((T_REQUIRE) reduce 145) ((T_REQUIRE_ONCE) reduce 145) ((T_RETURN) reduce 145) ((T_START_HEREDOC) reduce 145) ((T_STATIC) reduce 145) ((T_STRING) reduce 145) ((T_STRING_CAST) reduce 145) ((T_SWITCH) reduce 145) ((T_THROW) reduce 145) ((T_TRAIT) reduce 145) ((T_TRAIT_C) reduce 145) ((T_TRY) reduce 145) ((T_UNSET) reduce 145) ((T_UNSET_CAST) reduce 145) ((T_VARIABLE) reduce 145) ((T_VOID_CAST) reduce 145) ((T_WHILE) reduce 145) ((T_YIELD) reduce 145) ((T_YIELD_FROM) reduce 145) (("[") reduce 145) (("`") reduce 145) (("{") reduce 145) (("~") reduce 145)) 23904 ((("!") reduce 250) (("\"") reduce 250) (("$") reduce 250) (("(") reduce 250) (("+") reduce 250) (("-") reduce 250) ((";") reduce 250) (("@") reduce 250) ((T_ABSTRACT) reduce 250) ((T_ARRAY) reduce 250) ((T_ARRAY_CAST) reduce 250) ((T_ATTRIBUTE) reduce 250) ((T_BOOL_CAST) reduce 250) ((T_BREAK) reduce 250) ((T_CLASS) reduce 250) ((T_CLASS_C) reduce 250) ((T_CLONE) reduce 250) ((T_CONSTANT_ENCAPSED_STRING) reduce 250) ((T_CONTINUE) reduce 250) ((T_DEC) reduce 250) ((T_DECLARE) reduce 250) ((T_DIR) reduce 250) ((T_DNUMBER) reduce 250) ((T_DO) reduce 250) ((T_DOUBLE_CAST) reduce 250) ((T_ECHO) reduce 250) ((T_EMPTY) reduce 250) ((T_ENDFOREACH) reduce 250) ((T_ENUM) reduce 250) ((T_EVAL) reduce 250) ((T_EXIT) reduce 250) ((T_FILE) reduce 250) ((T_FINAL) reduce 250) ((T_FN) reduce 250) ((T_FOR) reduce 250) ((T_FOREACH) reduce 250) ((T_FUNCTION) reduce 250) ((T_FUNC_C) reduce 250) ((T_GLOBAL) reduce 250) ((T_GOTO) reduce 250) ((T_HALT_COMPILER) reduce 250) ((T_IF) reduce 250) ((T_INC) reduce 250) ((T_INCLUDE) reduce 250) ((T_INCLUDE_ONCE) reduce 250) ((T_INLINE_HTML) reduce 250) ((T_INTERFACE) reduce 250) ((T_INT_CAST) reduce 250) ((T_ISSET) reduce 250) ((T_LINE) reduce 250) ((T_LIST) reduce 250) ((T_LNUMBER) reduce 250) ((T_MATCH) reduce 250) ((T_METHOD_C) reduce 250) ((T_NAME_FULLY_QUALIFIED) reduce 250) ((T_NAME_QUALIFIED) reduce 250) ((T_NAME_RELATIVE) reduce 250) ((T_NEW) reduce 250) ((T_NS_C) reduce 250) ((T_OBJECT_CAST) reduce 250) ((T_PRINT) reduce 250) ((T_PROPERTY_C) reduce 250) ((T_READONLY) reduce 250) ((T_REQUIRE) reduce 250) ((T_REQUIRE_ONCE) reduce 250) ((T_RETURN) reduce 250) ((T_START_HEREDOC) reduce 250) ((T_STATIC) reduce 250) ((T_STRING) reduce 250) ((T_STRING_CAST) reduce 250) ((T_SWITCH) reduce 250) ((T_THROW) reduce 250) ((T_TRAIT) reduce 250) ((T_TRAIT_C) reduce 250) ((T_TRY) reduce 250) ((T_UNSET) reduce 250) ((T_UNSET_CAST) reduce 250) ((T_VARIABLE) reduce 250) ((T_VOID_CAST) reduce 250) ((T_WHILE) reduce 250) ((T_YIELD) reduce 250) ((T_YIELD_FROM) reduce 250) (("[") reduce 250) (("`") reduce 250) (("{") reduce 250) (("~") reduce 250)) 23908 ((("!") reduce 251) (("\"") reduce 251) (("$") reduce 251) (("(") reduce 251) (("+") reduce 251) (("-") reduce 251) ((";") reduce 251) (("@") reduce 251) ((T_ABSTRACT) reduce 251) ((T_ARRAY) reduce 251) ((T_ARRAY_CAST) reduce 251) ((T_ATTRIBUTE) reduce 251) ((T_BOOL_CAST) reduce 251) ((T_BREAK) reduce 251) ((T_CLASS) reduce 251) ((T_CLASS_C) reduce 251) ((T_CLONE) reduce 251) ((T_CONSTANT_ENCAPSED_STRING) reduce 251) ((T_CONTINUE) reduce 251) ((T_DEC) reduce 251) ((T_DECLARE) reduce 251) ((T_DIR) reduce 251) ((T_DNUMBER) reduce 251) ((T_DO) reduce 251) ((T_DOUBLE_CAST) reduce 251) ((T_ECHO) reduce 251) ((T_EMPTY) reduce 251) ((T_ENDFOREACH) reduce 251) ((T_ENUM) reduce 251) ((T_EVAL) reduce 251) ((T_EXIT) reduce 251) ((T_FILE) reduce 251) ((T_FINAL) reduce 251) ((T_FN) reduce 251) ((T_FOR) reduce 251) ((T_FOREACH) reduce 251) ((T_FUNCTION) reduce 251) ((T_FUNC_C) reduce 251) ((T_GLOBAL) reduce 251) ((T_GOTO) reduce 251) ((T_HALT_COMPILER) reduce 251) ((T_IF) reduce 251) ((T_INC) reduce 251) ((T_INCLUDE) reduce 251) ((T_INCLUDE_ONCE) reduce 251) ((T_INLINE_HTML) reduce 251) ((T_INTERFACE) reduce 251) ((T_INT_CAST) reduce 251) ((T_ISSET) reduce 251) ((T_LINE) reduce 251) ((T_LIST) reduce 251) ((T_LNUMBER) reduce 251) ((T_MATCH) reduce 251) ((T_METHOD_C) reduce 251) ((T_NAME_FULLY_QUALIFIED) reduce 251) ((T_NAME_QUALIFIED) reduce 251) ((T_NAME_RELATIVE) reduce 251) ((T_NEW) reduce 251) ((T_NS_C) reduce 251) ((T_OBJECT_CAST) reduce 251) ((T_PRINT) reduce 251) ((T_PROPERTY_C) reduce 251) ((T_READONLY) reduce 251) ((T_REQUIRE) reduce 251) ((T_REQUIRE_ONCE) reduce 251) ((T_RETURN) reduce 251) ((T_START_HEREDOC) reduce 251) ((T_STATIC) reduce 251) ((T_STRING) reduce 251) ((T_STRING_CAST) reduce 251) ((T_SWITCH) reduce 251) ((T_THROW) reduce 251) ((T_TRAIT) reduce 251) ((T_TRAIT_C) reduce 251) ((T_TRY) reduce 251) ((T_UNSET) reduce 251) ((T_UNSET_CAST) reduce 251) ((T_VARIABLE) reduce 251) ((T_VOID_CAST) reduce 251) ((T_WHILE) reduce 251) ((T_YIELD) reduce 251) ((T_YIELD_FROM) reduce 251) (("[") reduce 251) (("`") reduce 251) (("{") reduce 251) (("~") reduce 251)) 23913 ((("!") reduce 242) (("\"") reduce 242) (("$") reduce 242) (("(") reduce 242) (("+") reduce 242) (("-") reduce 242) ((";") reduce 242) (("@") reduce 242) ((T_ABSTRACT) reduce 242) ((T_ARRAY) reduce 242) ((T_ARRAY_CAST) reduce 242) ((T_ATTRIBUTE) reduce 242) ((T_BOOL_CAST) reduce 242) ((T_BREAK) reduce 242) ((T_CLASS) reduce 242) ((T_CLASS_C) reduce 242) ((T_CLONE) reduce 242) ((T_CONSTANT_ENCAPSED_STRING) reduce 242) ((T_CONTINUE) reduce 242) ((T_DEC) reduce 242) ((T_DECLARE) reduce 242) ((T_DIR) reduce 242) ((T_DNUMBER) reduce 242) ((T_DO) reduce 242) ((T_DOUBLE_CAST) reduce 242) ((T_ECHO) reduce 242) ((T_EMPTY) reduce 242) ((T_ENDFOREACH) reduce 242) ((T_ENUM) reduce 242) ((T_EVAL) reduce 242) ((T_EXIT) reduce 242) ((T_FILE) reduce 242) ((T_FINAL) reduce 242) ((T_FN) reduce 242) ((T_FOR) reduce 242) ((T_FOREACH) reduce 242) ((T_FUNCTION) reduce 242) ((T_FUNC_C) reduce 242) ((T_GLOBAL) reduce 242) ((T_GOTO) reduce 242) ((T_HALT_COMPILER) reduce 242) ((T_IF) reduce 242) ((T_INC) reduce 242) ((T_INCLUDE) reduce 242) ((T_INCLUDE_ONCE) reduce 242) ((T_INLINE_HTML) reduce 242) ((T_INTERFACE) reduce 242) ((T_INT_CAST) reduce 242) ((T_ISSET) reduce 242) ((T_LINE) reduce 242) ((T_LIST) reduce 242) ((T_LNUMBER) reduce 242) ((T_MATCH) reduce 242) ((T_METHOD_C) reduce 242) ((T_NAME_FULLY_QUALIFIED) reduce 242) ((T_NAME_QUALIFIED) reduce 242) ((T_NAME_RELATIVE) reduce 242) ((T_NEW) reduce 242) ((T_NS_C) reduce 242) ((T_OBJECT_CAST) reduce 242) ((T_PRINT) reduce 242) ((T_PROPERTY_C) reduce 242) ((T_READONLY) reduce 242) ((T_REQUIRE) reduce 242) ((T_REQUIRE_ONCE) reduce 242) ((T_RETURN) reduce 242) ((T_START_HEREDOC) reduce 242) ((T_STATIC) reduce 242) ((T_STRING) reduce 242) ((T_STRING_CAST) reduce 242) ((T_SWITCH) reduce 242) ((T_THROW) reduce 242) ((T_TRAIT) reduce 242) ((T_TRAIT_C) reduce 242) ((T_TRY) reduce 242) ((T_UNSET) reduce 242) ((T_UNSET_CAST) reduce 242) ((T_VARIABLE) reduce 242) ((T_VOID_CAST) reduce 242) ((T_WHILE) reduce 242) ((T_YIELD) reduce 242) ((T_YIELD_FROM) reduce 242) (("[") reduce 242) (("`") reduce 242) (("{") reduce 242) (("~") reduce 242)) 23914 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_EMPTY) reduce 150) ((T_ENDFOREACH) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_VOID_CAST) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 23917 ((("!") reduce 243) (("\"") reduce 243) (("$") reduce 243) (("(") reduce 243) (("+") reduce 243) (("-") reduce 243) ((";") reduce 243) (("@") reduce 243) ((T_ABSTRACT) reduce 243) ((T_ARRAY) reduce 243) ((T_ARRAY_CAST) reduce 243) ((T_ATTRIBUTE) reduce 243) ((T_BOOL_CAST) reduce 243) ((T_BREAK) reduce 243) ((T_CLASS) reduce 243) ((T_CLASS_C) reduce 243) ((T_CLONE) reduce 243) ((T_CONSTANT_ENCAPSED_STRING) reduce 243) ((T_CONTINUE) reduce 243) ((T_DEC) reduce 243) ((T_DECLARE) reduce 243) ((T_DIR) reduce 243) ((T_DNUMBER) reduce 243) ((T_DO) reduce 243) ((T_DOUBLE_CAST) reduce 243) ((T_ECHO) reduce 243) ((T_EMPTY) reduce 243) ((T_ENDFOREACH) reduce 243) ((T_ENUM) reduce 243) ((T_EVAL) reduce 243) ((T_EXIT) reduce 243) ((T_FILE) reduce 243) ((T_FINAL) reduce 243) ((T_FN) reduce 243) ((T_FOR) reduce 243) ((T_FOREACH) reduce 243) ((T_FUNCTION) reduce 243) ((T_FUNC_C) reduce 243) ((T_GLOBAL) reduce 243) ((T_GOTO) reduce 243) ((T_HALT_COMPILER) reduce 243) ((T_IF) reduce 243) ((T_INC) reduce 243) ((T_INCLUDE) reduce 243) ((T_INCLUDE_ONCE) reduce 243) ((T_INLINE_HTML) reduce 243) ((T_INTERFACE) reduce 243) ((T_INT_CAST) reduce 243) ((T_ISSET) reduce 243) ((T_LINE) reduce 243) ((T_LIST) reduce 243) ((T_LNUMBER) reduce 243) ((T_MATCH) reduce 243) ((T_METHOD_C) reduce 243) ((T_NAME_FULLY_QUALIFIED) reduce 243) ((T_NAME_QUALIFIED) reduce 243) ((T_NAME_RELATIVE) reduce 243) ((T_NEW) reduce 243) ((T_NS_C) reduce 243) ((T_OBJECT_CAST) reduce 243) ((T_PRINT) reduce 243) ((T_PROPERTY_C) reduce 243) ((T_READONLY) reduce 243) ((T_REQUIRE) reduce 243) ((T_REQUIRE_ONCE) reduce 243) ((T_RETURN) reduce 243) ((T_START_HEREDOC) reduce 243) ((T_STATIC) reduce 243) ((T_STRING) reduce 243) ((T_STRING_CAST) reduce 243) ((T_SWITCH) reduce 243) ((T_THROW) reduce 243) ((T_TRAIT) reduce 243) ((T_TRAIT_C) reduce 243) ((T_TRY) reduce 243) ((T_UNSET) reduce 243) ((T_UNSET_CAST) reduce 243) ((T_VARIABLE) reduce 243) ((T_VOID_CAST) reduce 243) ((T_WHILE) reduce 243) ((T_YIELD) reduce 243) ((T_YIELD_FROM) reduce 243) (("[") reduce 243) (("`") reduce 243) (("{") reduce 243) (("~") reduce 243)) 23919 ((("!") reduce 170) (("\"") reduce 170) (("$") reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_EMPTY) reduce 170) ((T_ENDFOREACH) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_VARIABLE) reduce 170) ((T_VOID_CAST) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("~") reduce 170)) 23924 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_EMPTY) reduce 162) ((T_ENDFOREACH) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_VOID_CAST) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 23927 ((("!") reduce 171) (("\"") reduce 171) (("$") reduce 171) (("(") reduce 171) (("+") reduce 171) (("-") reduce 171) ((";") reduce 171) (("@") reduce 171) ((T_ABSTRACT) reduce 171) ((T_ARRAY) reduce 171) ((T_ARRAY_CAST) reduce 171) ((T_ATTRIBUTE) reduce 171) ((T_BOOL_CAST) reduce 171) ((T_BREAK) reduce 171) ((T_CATCH) reduce 171) ((T_CLASS) reduce 171) ((T_CLASS_C) reduce 171) ((T_CLONE) reduce 171) ((T_CONSTANT_ENCAPSED_STRING) reduce 171) ((T_CONTINUE) reduce 171) ((T_DEC) reduce 171) ((T_DECLARE) reduce 171) ((T_DIR) reduce 171) ((T_DNUMBER) reduce 171) ((T_DO) reduce 171) ((T_DOUBLE_CAST) reduce 171) ((T_ECHO) reduce 171) ((T_EMPTY) reduce 171) ((T_ENDFOREACH) reduce 171) ((T_ENUM) reduce 171) ((T_EVAL) reduce 171) ((T_EXIT) reduce 171) ((T_FILE) reduce 171) ((T_FINAL) reduce 171) ((T_FINALLY) reduce 171) ((T_FN) reduce 171) ((T_FOR) reduce 171) ((T_FOREACH) reduce 171) ((T_FUNCTION) reduce 171) ((T_FUNC_C) reduce 171) ((T_GLOBAL) reduce 171) ((T_GOTO) reduce 171) ((T_HALT_COMPILER) reduce 171) ((T_IF) reduce 171) ((T_INC) reduce 171) ((T_INCLUDE) reduce 171) ((T_INCLUDE_ONCE) reduce 171) ((T_INLINE_HTML) reduce 171) ((T_INTERFACE) reduce 171) ((T_INT_CAST) reduce 171) ((T_ISSET) reduce 171) ((T_LINE) reduce 171) ((T_LIST) reduce 171) ((T_LNUMBER) reduce 171) ((T_MATCH) reduce 171) ((T_METHOD_C) reduce 171) ((T_NAME_FULLY_QUALIFIED) reduce 171) ((T_NAME_QUALIFIED) reduce 171) ((T_NAME_RELATIVE) reduce 171) ((T_NEW) reduce 171) ((T_NS_C) reduce 171) ((T_OBJECT_CAST) reduce 171) ((T_PRINT) reduce 171) ((T_PROPERTY_C) reduce 171) ((T_READONLY) reduce 171) ((T_REQUIRE) reduce 171) ((T_REQUIRE_ONCE) reduce 171) ((T_RETURN) reduce 171) ((T_START_HEREDOC) reduce 171) ((T_STATIC) reduce 171) ((T_STRING) reduce 171) ((T_STRING_CAST) reduce 171) ((T_SWITCH) reduce 171) ((T_THROW) reduce 171) ((T_TRAIT) reduce 171) ((T_TRAIT_C) reduce 171) ((T_TRY) reduce 171) ((T_UNSET) reduce 171) ((T_UNSET_CAST) reduce 171) ((T_VARIABLE) reduce 171) ((T_VOID_CAST) reduce 171) ((T_WHILE) reduce 171) ((T_YIELD) reduce 171) ((T_YIELD_FROM) reduce 171) (("[") reduce 171) (("`") reduce 171) (("{") reduce 171) (("~") reduce 171)) 23928 ((("!") reduce 177) (("\"") reduce 177) (("$") reduce 177) (("(") reduce 177) (("+") reduce 177) (("-") reduce 177) ((";") reduce 177) (("@") reduce 177) ((T_ABSTRACT) reduce 177) ((T_ARRAY) reduce 177) ((T_ARRAY_CAST) reduce 177) ((T_ATTRIBUTE) reduce 177) ((T_BOOL_CAST) reduce 177) ((T_BREAK) reduce 177) ((T_CATCH) shift) ((T_CLASS) reduce 177) ((T_CLASS_C) reduce 177) ((T_CLONE) reduce 177) ((T_CONSTANT_ENCAPSED_STRING) reduce 177) ((T_CONTINUE) reduce 177) ((T_DEC) reduce 177) ((T_DECLARE) reduce 177) ((T_DIR) reduce 177) ((T_DNUMBER) reduce 177) ((T_DO) reduce 177) ((T_DOUBLE_CAST) reduce 177) ((T_ECHO) reduce 177) ((T_EMPTY) reduce 177) ((T_ENDFOREACH) reduce 177) ((T_ENUM) reduce 177) ((T_EVAL) reduce 177) ((T_EXIT) reduce 177) ((T_FILE) reduce 177) ((T_FINAL) reduce 177) ((T_FINALLY) shift) ((T_FN) reduce 177) ((T_FOR) reduce 177) ((T_FOREACH) reduce 177) ((T_FUNCTION) reduce 177) ((T_FUNC_C) reduce 177) ((T_GLOBAL) reduce 177) ((T_GOTO) reduce 177) ((T_HALT_COMPILER) reduce 177) ((T_IF) reduce 177) ((T_INC) reduce 177) ((T_INCLUDE) reduce 177) ((T_INCLUDE_ONCE) reduce 177) ((T_INLINE_HTML) reduce 177) ((T_INTERFACE) reduce 177) ((T_INT_CAST) reduce 177) ((T_ISSET) reduce 177) ((T_LINE) reduce 177) ((T_LIST) reduce 177) ((T_LNUMBER) reduce 177) ((T_MATCH) reduce 177) ((T_METHOD_C) reduce 177) ((T_NAME_FULLY_QUALIFIED) reduce 177) ((T_NAME_QUALIFIED) reduce 177) ((T_NAME_RELATIVE) reduce 177) ((T_NEW) reduce 177) ((T_NS_C) reduce 177) ((T_OBJECT_CAST) reduce 177) ((T_PRINT) reduce 177) ((T_PROPERTY_C) reduce 177) ((T_READONLY) reduce 177) ((T_REQUIRE) reduce 177) ((T_REQUIRE_ONCE) reduce 177) ((T_RETURN) reduce 177) ((T_START_HEREDOC) reduce 177) ((T_STATIC) reduce 177) ((T_STRING) reduce 177) ((T_STRING_CAST) reduce 177) ((T_SWITCH) reduce 177) ((T_THROW) reduce 177) ((T_TRAIT) reduce 177) ((T_TRAIT_C) reduce 177) ((T_TRY) reduce 177) ((T_UNSET) reduce 177) ((T_UNSET_CAST) reduce 177) ((T_VARIABLE) reduce 177) ((T_VOID_CAST) reduce 177) ((T_WHILE) reduce 177) ((T_YIELD) reduce 177) ((T_YIELD_FROM) reduce 177) (("[") reduce 177) (("`") reduce 177) (("{") reduce 177) (("~") reduce 177)) 23931 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_EMPTY) reduce 167) ((T_ENDFOREACH) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_VOID_CAST) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 23934 ((("!") reduce 178) (("\"") reduce 178) (("$") reduce 178) (("(") reduce 178) (("+") reduce 178) (("-") reduce 178) ((";") reduce 178) (("@") reduce 178) ((T_ABSTRACT) reduce 178) ((T_ARRAY) reduce 178) ((T_ARRAY_CAST) reduce 178) ((T_ATTRIBUTE) reduce 178) ((T_BOOL_CAST) reduce 178) ((T_BREAK) reduce 178) ((T_CLASS) reduce 178) ((T_CLASS_C) reduce 178) ((T_CLONE) reduce 178) ((T_CONSTANT_ENCAPSED_STRING) reduce 178) ((T_CONTINUE) reduce 178) ((T_DEC) reduce 178) ((T_DECLARE) reduce 178) ((T_DIR) reduce 178) ((T_DNUMBER) reduce 178) ((T_DO) reduce 178) ((T_DOUBLE_CAST) reduce 178) ((T_ECHO) reduce 178) ((T_EMPTY) reduce 178) ((T_ENDFOREACH) reduce 178) ((T_ENUM) reduce 178) ((T_EVAL) reduce 178) ((T_EXIT) reduce 178) ((T_FILE) reduce 178) ((T_FINAL) reduce 178) ((T_FN) reduce 178) ((T_FOR) reduce 178) ((T_FOREACH) reduce 178) ((T_FUNCTION) reduce 178) ((T_FUNC_C) reduce 178) ((T_GLOBAL) reduce 178) ((T_GOTO) reduce 178) ((T_HALT_COMPILER) reduce 178) ((T_IF) reduce 178) ((T_INC) reduce 178) ((T_INCLUDE) reduce 178) ((T_INCLUDE_ONCE) reduce 178) ((T_INLINE_HTML) reduce 178) ((T_INTERFACE) reduce 178) ((T_INT_CAST) reduce 178) ((T_ISSET) reduce 178) ((T_LINE) reduce 178) ((T_LIST) reduce 178) ((T_LNUMBER) reduce 178) ((T_MATCH) reduce 178) ((T_METHOD_C) reduce 178) ((T_NAME_FULLY_QUALIFIED) reduce 178) ((T_NAME_QUALIFIED) reduce 178) ((T_NAME_RELATIVE) reduce 178) ((T_NEW) reduce 178) ((T_NS_C) reduce 178) ((T_OBJECT_CAST) reduce 178) ((T_PRINT) reduce 178) ((T_PROPERTY_C) reduce 178) ((T_READONLY) reduce 178) ((T_REQUIRE) reduce 178) ((T_REQUIRE_ONCE) reduce 178) ((T_RETURN) reduce 178) ((T_START_HEREDOC) reduce 178) ((T_STATIC) reduce 178) ((T_STRING) reduce 178) ((T_STRING_CAST) reduce 178) ((T_SWITCH) reduce 178) ((T_THROW) reduce 178) ((T_TRAIT) reduce 178) ((T_TRAIT_C) reduce 178) ((T_TRY) reduce 178) ((T_UNSET) reduce 178) ((T_UNSET_CAST) reduce 178) ((T_VARIABLE) reduce 178) ((T_VOID_CAST) reduce 178) ((T_WHILE) reduce 178) ((T_YIELD) reduce 178) ((T_YIELD_FROM) reduce 178) (("[") reduce 178) (("`") reduce 178) (("{") reduce 178) (("~") reduce 178)) 23941 ((("!") reduce 172) (("\"") reduce 172) (("$") reduce 172) (("(") reduce 172) (("+") reduce 172) (("-") reduce 172) ((";") reduce 172) (("@") reduce 172) ((T_ABSTRACT) reduce 172) ((T_ARRAY) reduce 172) ((T_ARRAY_CAST) reduce 172) ((T_ATTRIBUTE) reduce 172) ((T_BOOL_CAST) reduce 172) ((T_BREAK) reduce 172) ((T_CATCH) reduce 172) ((T_CLASS) reduce 172) ((T_CLASS_C) reduce 172) ((T_CLONE) reduce 172) ((T_CONSTANT_ENCAPSED_STRING) reduce 172) ((T_CONTINUE) reduce 172) ((T_DEC) reduce 172) ((T_DECLARE) reduce 172) ((T_DIR) reduce 172) ((T_DNUMBER) reduce 172) ((T_DO) reduce 172) ((T_DOUBLE_CAST) reduce 172) ((T_ECHO) reduce 172) ((T_EMPTY) reduce 172) ((T_ENDFOREACH) reduce 172) ((T_ENUM) reduce 172) ((T_EVAL) reduce 172) ((T_EXIT) reduce 172) ((T_FILE) reduce 172) ((T_FINAL) reduce 172) ((T_FINALLY) reduce 172) ((T_FN) reduce 172) ((T_FOR) reduce 172) ((T_FOREACH) reduce 172) ((T_FUNCTION) reduce 172) ((T_FUNC_C) reduce 172) ((T_GLOBAL) reduce 172) ((T_GOTO) reduce 172) ((T_HALT_COMPILER) reduce 172) ((T_IF) reduce 172) ((T_INC) reduce 172) ((T_INCLUDE) reduce 172) ((T_INCLUDE_ONCE) reduce 172) ((T_INLINE_HTML) reduce 172) ((T_INTERFACE) reduce 172) ((T_INT_CAST) reduce 172) ((T_ISSET) reduce 172) ((T_LINE) reduce 172) ((T_LIST) reduce 172) ((T_LNUMBER) reduce 172) ((T_MATCH) reduce 172) ((T_METHOD_C) reduce 172) ((T_NAME_FULLY_QUALIFIED) reduce 172) ((T_NAME_QUALIFIED) reduce 172) ((T_NAME_RELATIVE) reduce 172) ((T_NEW) reduce 172) ((T_NS_C) reduce 172) ((T_OBJECT_CAST) reduce 172) ((T_PRINT) reduce 172) ((T_PROPERTY_C) reduce 172) ((T_READONLY) reduce 172) ((T_REQUIRE) reduce 172) ((T_REQUIRE_ONCE) reduce 172) ((T_RETURN) reduce 172) ((T_START_HEREDOC) reduce 172) ((T_STATIC) reduce 172) ((T_STRING) reduce 172) ((T_STRING_CAST) reduce 172) ((T_SWITCH) reduce 172) ((T_THROW) reduce 172) ((T_TRAIT) reduce 172) ((T_TRAIT_C) reduce 172) ((T_TRY) reduce 172) ((T_UNSET) reduce 172) ((T_UNSET_CAST) reduce 172) ((T_VARIABLE) reduce 172) ((T_VOID_CAST) reduce 172) ((T_WHILE) reduce 172) ((T_YIELD) reduce 172) ((T_YIELD_FROM) reduce 172) (("[") reduce 172) (("`") reduce 172) (("{") reduce 172) (("~") reduce 172)) 23946 ((("!") reduce 200) (("\"") reduce 200) (("$") reduce 200) (("(") reduce 200) (("+") reduce 200) (("-") reduce 200) ((";") reduce 200) (("@") reduce 200) ((T_ABSTRACT) reduce 200) ((T_ARRAY) reduce 200) ((T_ARRAY_CAST) reduce 200) ((T_ATTRIBUTE) reduce 200) ((T_BOOL_CAST) reduce 200) ((T_BREAK) reduce 200) ((T_CLASS) reduce 200) ((T_CLASS_C) reduce 200) ((T_CLONE) reduce 200) ((T_CONSTANT_ENCAPSED_STRING) reduce 200) ((T_CONTINUE) reduce 200) ((T_DEC) reduce 200) ((T_DECLARE) reduce 200) ((T_DIR) reduce 200) ((T_DNUMBER) reduce 200) ((T_DO) reduce 200) ((T_DOUBLE_CAST) reduce 200) ((T_ECHO) reduce 200) ((T_EMPTY) reduce 200) ((T_ENDFOREACH) reduce 200) ((T_ENUM) reduce 200) ((T_EVAL) reduce 200) ((T_EXIT) reduce 200) ((T_FILE) reduce 200) ((T_FINAL) reduce 200) ((T_FN) reduce 200) ((T_FOR) reduce 200) ((T_FOREACH) reduce 200) ((T_FUNCTION) reduce 200) ((T_FUNC_C) reduce 200) ((T_GLOBAL) reduce 200) ((T_GOTO) reduce 200) ((T_HALT_COMPILER) reduce 200) ((T_IF) reduce 200) ((T_INC) reduce 200) ((T_INCLUDE) reduce 200) ((T_INCLUDE_ONCE) reduce 200) ((T_INLINE_HTML) reduce 200) ((T_INTERFACE) reduce 200) ((T_INT_CAST) reduce 200) ((T_ISSET) reduce 200) ((T_LINE) reduce 200) ((T_LIST) reduce 200) ((T_LNUMBER) reduce 200) ((T_MATCH) reduce 200) ((T_METHOD_C) reduce 200) ((T_NAME_FULLY_QUALIFIED) reduce 200) ((T_NAME_QUALIFIED) reduce 200) ((T_NAME_RELATIVE) reduce 200) ((T_NEW) reduce 200) ((T_NS_C) reduce 200) ((T_OBJECT_CAST) reduce 200) ((T_PRINT) reduce 200) ((T_PROPERTY_C) reduce 200) ((T_READONLY) reduce 200) ((T_REQUIRE) reduce 200) ((T_REQUIRE_ONCE) reduce 200) ((T_RETURN) reduce 200) ((T_START_HEREDOC) reduce 200) ((T_STATIC) reduce 200) ((T_STRING) reduce 200) ((T_STRING_CAST) reduce 200) ((T_SWITCH) reduce 200) ((T_THROW) reduce 200) ((T_TRAIT) reduce 200) ((T_TRAIT_C) reduce 200) ((T_TRY) reduce 200) ((T_UNSET) reduce 200) ((T_UNSET_CAST) reduce 200) ((T_VARIABLE) reduce 200) ((T_VOID_CAST) reduce 200) ((T_WHILE) reduce 200) ((T_YIELD) reduce 200) ((T_YIELD_FROM) reduce 200) (("[") reduce 200) (("`") reduce 200) (("{") reduce 200) (("~") reduce 200)) 23951 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_EMPTY) reduce 153) ((T_ENDFOREACH) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_VOID_CAST) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 23955 ((("!") reduce 224) (("\"") reduce 224) (("$") reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_EMPTY) reduce 224) ((T_ENDFOREACH) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_VARIABLE) reduce 224) ((T_VOID_CAST) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("~") reduce 224)) 23957 ((("!") reduce 225) (("\"") reduce 225) (("$") reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_EMPTY) reduce 225) ((T_ENDFOREACH) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_VARIABLE) reduce 225) ((T_VOID_CAST) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("~") reduce 225)) 23961 ((("!") reduce 226) (("\"") reduce 226) (("$") reduce 226) (("(") reduce 226) (("+") reduce 226) (("-") reduce 226) ((";") reduce 226) (("@") reduce 226) ((T_ABSTRACT) reduce 226) ((T_ARRAY) reduce 226) ((T_ARRAY_CAST) reduce 226) ((T_ATTRIBUTE) reduce 226) ((T_BOOL_CAST) reduce 226) ((T_BREAK) reduce 226) ((T_CLASS) reduce 226) ((T_CLASS_C) reduce 226) ((T_CLONE) reduce 226) ((T_CONSTANT_ENCAPSED_STRING) reduce 226) ((T_CONTINUE) reduce 226) ((T_DEC) reduce 226) ((T_DECLARE) reduce 226) ((T_DIR) reduce 226) ((T_DNUMBER) reduce 226) ((T_DO) reduce 226) ((T_DOUBLE_CAST) reduce 226) ((T_ECHO) reduce 226) ((T_EMPTY) reduce 226) ((T_ENDFOREACH) reduce 226) ((T_ENUM) reduce 226) ((T_EVAL) reduce 226) ((T_EXIT) reduce 226) ((T_FILE) reduce 226) ((T_FINAL) reduce 226) ((T_FN) reduce 226) ((T_FOR) reduce 226) ((T_FOREACH) reduce 226) ((T_FUNCTION) reduce 226) ((T_FUNC_C) reduce 226) ((T_GLOBAL) reduce 226) ((T_GOTO) reduce 226) ((T_HALT_COMPILER) reduce 226) ((T_IF) reduce 226) ((T_INC) reduce 226) ((T_INCLUDE) reduce 226) ((T_INCLUDE_ONCE) reduce 226) ((T_INLINE_HTML) reduce 226) ((T_INTERFACE) reduce 226) ((T_INT_CAST) reduce 226) ((T_ISSET) reduce 226) ((T_LINE) reduce 226) ((T_LIST) reduce 226) ((T_LNUMBER) reduce 226) ((T_MATCH) reduce 226) ((T_METHOD_C) reduce 226) ((T_NAME_FULLY_QUALIFIED) reduce 226) ((T_NAME_QUALIFIED) reduce 226) ((T_NAME_RELATIVE) reduce 226) ((T_NEW) reduce 226) ((T_NS_C) reduce 226) ((T_OBJECT_CAST) reduce 226) ((T_PRINT) reduce 226) ((T_PROPERTY_C) reduce 226) ((T_READONLY) reduce 226) ((T_REQUIRE) reduce 226) ((T_REQUIRE_ONCE) reduce 226) ((T_RETURN) reduce 226) ((T_START_HEREDOC) reduce 226) ((T_STATIC) reduce 226) ((T_STRING) reduce 226) ((T_STRING_CAST) reduce 226) ((T_SWITCH) reduce 226) ((T_THROW) reduce 226) ((T_TRAIT) reduce 226) ((T_TRAIT_C) reduce 226) ((T_TRY) reduce 226) ((T_UNSET) reduce 226) ((T_UNSET_CAST) reduce 226) ((T_VARIABLE) reduce 226) ((T_VOID_CAST) reduce 226) ((T_WHILE) reduce 226) ((T_YIELD) reduce 226) ((T_YIELD_FROM) reduce 226) (("[") reduce 226) (("`") reduce 226) (("{") reduce 226) (("~") reduce 226)) 23964 ((("!") reduce 227) (("\"") reduce 227) (("$") reduce 227) (("(") reduce 227) (("+") reduce 227) (("-") reduce 227) ((";") reduce 227) (("@") reduce 227) ((T_ABSTRACT) reduce 227) ((T_ARRAY) reduce 227) ((T_ARRAY_CAST) reduce 227) ((T_ATTRIBUTE) reduce 227) ((T_BOOL_CAST) reduce 227) ((T_BREAK) reduce 227) ((T_CLASS) reduce 227) ((T_CLASS_C) reduce 227) ((T_CLONE) reduce 227) ((T_CONSTANT_ENCAPSED_STRING) reduce 227) ((T_CONTINUE) reduce 227) ((T_DEC) reduce 227) ((T_DECLARE) reduce 227) ((T_DIR) reduce 227) ((T_DNUMBER) reduce 227) ((T_DO) reduce 227) ((T_DOUBLE_CAST) reduce 227) ((T_ECHO) reduce 227) ((T_EMPTY) reduce 227) ((T_ENDFOREACH) reduce 227) ((T_ENUM) reduce 227) ((T_EVAL) reduce 227) ((T_EXIT) reduce 227) ((T_FILE) reduce 227) ((T_FINAL) reduce 227) ((T_FN) reduce 227) ((T_FOR) reduce 227) ((T_FOREACH) reduce 227) ((T_FUNCTION) reduce 227) ((T_FUNC_C) reduce 227) ((T_GLOBAL) reduce 227) ((T_GOTO) reduce 227) ((T_HALT_COMPILER) reduce 227) ((T_IF) reduce 227) ((T_INC) reduce 227) ((T_INCLUDE) reduce 227) ((T_INCLUDE_ONCE) reduce 227) ((T_INLINE_HTML) reduce 227) ((T_INTERFACE) reduce 227) ((T_INT_CAST) reduce 227) ((T_ISSET) reduce 227) ((T_LINE) reduce 227) ((T_LIST) reduce 227) ((T_LNUMBER) reduce 227) ((T_MATCH) reduce 227) ((T_METHOD_C) reduce 227) ((T_NAME_FULLY_QUALIFIED) reduce 227) ((T_NAME_QUALIFIED) reduce 227) ((T_NAME_RELATIVE) reduce 227) ((T_NEW) reduce 227) ((T_NS_C) reduce 227) ((T_OBJECT_CAST) reduce 227) ((T_PRINT) reduce 227) ((T_PROPERTY_C) reduce 227) ((T_READONLY) reduce 227) ((T_REQUIRE) reduce 227) ((T_REQUIRE_ONCE) reduce 227) ((T_RETURN) reduce 227) ((T_START_HEREDOC) reduce 227) ((T_STATIC) reduce 227) ((T_STRING) reduce 227) ((T_STRING_CAST) reduce 227) ((T_SWITCH) reduce 227) ((T_THROW) reduce 227) ((T_TRAIT) reduce 227) ((T_TRAIT_C) reduce 227) ((T_TRY) reduce 227) ((T_UNSET) reduce 227) ((T_UNSET_CAST) reduce 227) ((T_VARIABLE) reduce 227) ((T_VOID_CAST) reduce 227) ((T_WHILE) reduce 227) ((T_YIELD) reduce 227) ((T_YIELD_FROM) reduce 227) (("[") reduce 227) (("`") reduce 227) (("{") reduce 227) (("~") reduce 227)) 23965 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_EMPTY) reduce 169) ((T_ENDFOREACH) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_VOID_CAST) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 23967 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_EMPTY) reduce 158) ((T_ENDFOREACH) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_VOID_CAST) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 23969 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_EMPTY) reduce 156) ((T_ENDFOREACH) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_VOID_CAST) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 23975 ((("!") reduce 201) (("\"") reduce 201) (("$") reduce 201) (("(") reduce 201) (("+") reduce 201) (("-") reduce 201) ((";") reduce 201) (("@") reduce 201) ((T_ABSTRACT) reduce 201) ((T_ARRAY) reduce 201) ((T_ARRAY_CAST) reduce 201) ((T_ATTRIBUTE) reduce 201) ((T_BOOL_CAST) reduce 201) ((T_BREAK) reduce 201) ((T_CLASS) reduce 201) ((T_CLASS_C) reduce 201) ((T_CLONE) reduce 201) ((T_CONSTANT_ENCAPSED_STRING) reduce 201) ((T_CONTINUE) reduce 201) ((T_DEC) reduce 201) ((T_DECLARE) reduce 201) ((T_DIR) reduce 201) ((T_DNUMBER) reduce 201) ((T_DO) reduce 201) ((T_DOUBLE_CAST) reduce 201) ((T_ECHO) reduce 201) ((T_EMPTY) reduce 201) ((T_ENDFOREACH) reduce 201) ((T_ENUM) reduce 201) ((T_EVAL) reduce 201) ((T_EXIT) reduce 201) ((T_FILE) reduce 201) ((T_FINAL) reduce 201) ((T_FN) reduce 201) ((T_FOR) reduce 201) ((T_FOREACH) reduce 201) ((T_FUNCTION) reduce 201) ((T_FUNC_C) reduce 201) ((T_GLOBAL) reduce 201) ((T_GOTO) reduce 201) ((T_HALT_COMPILER) reduce 201) ((T_IF) reduce 201) ((T_INC) reduce 201) ((T_INCLUDE) reduce 201) ((T_INCLUDE_ONCE) reduce 201) ((T_INLINE_HTML) reduce 201) ((T_INTERFACE) reduce 201) ((T_INT_CAST) reduce 201) ((T_ISSET) reduce 201) ((T_LINE) reduce 201) ((T_LIST) reduce 201) ((T_LNUMBER) reduce 201) ((T_MATCH) reduce 201) ((T_METHOD_C) reduce 201) ((T_NAME_FULLY_QUALIFIED) reduce 201) ((T_NAME_QUALIFIED) reduce 201) ((T_NAME_RELATIVE) reduce 201) ((T_NEW) reduce 201) ((T_NS_C) reduce 201) ((T_OBJECT_CAST) reduce 201) ((T_PRINT) reduce 201) ((T_PROPERTY_C) reduce 201) ((T_READONLY) reduce 201) ((T_REQUIRE) reduce 201) ((T_REQUIRE_ONCE) reduce 201) ((T_RETURN) reduce 201) ((T_START_HEREDOC) reduce 201) ((T_STATIC) reduce 201) ((T_STRING) reduce 201) ((T_STRING_CAST) reduce 201) ((T_SWITCH) reduce 201) ((T_THROW) reduce 201) ((T_TRAIT) reduce 201) ((T_TRAIT_C) reduce 201) ((T_TRY) reduce 201) ((T_UNSET) reduce 201) ((T_UNSET_CAST) reduce 201) ((T_VARIABLE) reduce 201) ((T_VOID_CAST) reduce 201) ((T_WHILE) reduce 201) ((T_YIELD) reduce 201) ((T_YIELD_FROM) reduce 201) (("[") reduce 201) (("`") reduce 201) (("{") reduce 201) (("~") reduce 201)) 23979 ((("!") reduce 244) (("\"") reduce 244) (("$") reduce 244) (("(") reduce 244) (("+") reduce 244) (("-") reduce 244) ((";") reduce 244) (("@") reduce 244) ((T_ABSTRACT) reduce 244) ((T_ARRAY) reduce 244) ((T_ARRAY_CAST) reduce 244) ((T_ATTRIBUTE) reduce 244) ((T_BOOL_CAST) reduce 244) ((T_BREAK) reduce 244) ((T_CLASS) reduce 244) ((T_CLASS_C) reduce 244) ((T_CLONE) reduce 244) ((T_CONSTANT_ENCAPSED_STRING) reduce 244) ((T_CONTINUE) reduce 244) ((T_DEC) reduce 244) ((T_DECLARE) reduce 244) ((T_DIR) reduce 244) ((T_DNUMBER) reduce 244) ((T_DO) reduce 244) ((T_DOUBLE_CAST) reduce 244) ((T_ECHO) reduce 244) ((T_ELSE) reduce 244) ((T_ELSEIF) reduce 244) ((T_EMPTY) reduce 244) ((T_ENDFOREACH) reduce 244) ((T_ENUM) reduce 244) ((T_EVAL) reduce 244) ((T_EXIT) reduce 244) ((T_FILE) reduce 244) ((T_FINAL) reduce 244) ((T_FN) reduce 244) ((T_FOR) reduce 244) ((T_FOREACH) reduce 244) ((T_FUNCTION) reduce 244) ((T_FUNC_C) reduce 244) ((T_GLOBAL) reduce 244) ((T_GOTO) reduce 244) ((T_HALT_COMPILER) reduce 244) ((T_IF) reduce 244) ((T_INC) reduce 244) ((T_INCLUDE) reduce 244) ((T_INCLUDE_ONCE) reduce 244) ((T_INLINE_HTML) reduce 244) ((T_INTERFACE) reduce 244) ((T_INT_CAST) reduce 244) ((T_ISSET) reduce 244) ((T_LINE) reduce 244) ((T_LIST) reduce 244) ((T_LNUMBER) reduce 244) ((T_MATCH) reduce 244) ((T_METHOD_C) reduce 244) ((T_NAME_FULLY_QUALIFIED) reduce 244) ((T_NAME_QUALIFIED) reduce 244) ((T_NAME_RELATIVE) reduce 244) ((T_NEW) reduce 244) ((T_NS_C) reduce 244) ((T_OBJECT_CAST) reduce 244) ((T_PRINT) reduce 244) ((T_PROPERTY_C) reduce 244) ((T_READONLY) reduce 244) ((T_REQUIRE) reduce 244) ((T_REQUIRE_ONCE) reduce 244) ((T_RETURN) reduce 244) ((T_START_HEREDOC) reduce 244) ((T_STATIC) reduce 244) ((T_STRING) reduce 244) ((T_STRING_CAST) reduce 244) ((T_SWITCH) reduce 244) ((T_THROW) reduce 244) ((T_TRAIT) reduce 244) ((T_TRAIT_C) reduce 244) ((T_TRY) reduce 244) ((T_UNSET) reduce 244) ((T_UNSET_CAST) reduce 244) ((T_VARIABLE) reduce 244) ((T_VOID_CAST) reduce 244) ((T_WHILE) reduce 244) ((T_YIELD) reduce 244) ((T_YIELD_FROM) reduce 244) (("[") reduce 244) (("`") reduce 244) (("{") reduce 244) (("~") reduce 244)) 23982 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_EMPTY) reduce 146) ((T_ENDFOREACH) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_PROPERTY_C) reduce 146) ((T_READONLY) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_VOID_CAST) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("~") reduce 146)) 23984 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_EMPTY) reduce 168) ((T_ENDFOREACH) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_VOID_CAST) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 23986 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_EMPTY) reduce 157) ((T_ENDFOREACH) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_VOID_CAST) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 23996 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_EMPTY) reduce 164) ((T_ENDFOREACH) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_VOID_CAST) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 23997 ((("!") reduce 220) (("\"") reduce 220) (("$") reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_EMPTY) reduce 220) ((T_ENDFOREACH) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_VARIABLE) reduce 220) ((T_VOID_CAST) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("~") reduce 220)) 24000 ((("!") reduce 221) (("\"") reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_EMPTY) reduce 221) ((T_ENDFOREACH) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_VARIABLE) reduce 221) ((T_VOID_CAST) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("~") reduce 221)) 24001 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_EMPTY) reduce 163) ((T_ENDFOREACH) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_VOID_CAST) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 24010 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_EMPTY) reduce 152) ((T_ENDFOREACH) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_VOID_CAST) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 24011 ((("!") reduce 218) (("\"") reduce 218) (("$") reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_EMPTY) reduce 218) ((T_ENDFOREACH) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_VARIABLE) reduce 218) ((T_VOID_CAST) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("~") reduce 218)) 24014 ((("!") reduce 219) (("\"") reduce 219) (("$") reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_EMPTY) reduce 219) ((T_ENDFOREACH) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_VARIABLE) reduce 219) ((T_VOID_CAST) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("~") reduce 219)) 24021 ((("!") reduce 202) (("\"") reduce 202) (("$") reduce 202) (("(") reduce 202) (("+") reduce 202) (("-") reduce 202) ((";") reduce 202) (("@") reduce 202) ((T_ABSTRACT) reduce 202) ((T_ARRAY) reduce 202) ((T_ARRAY_CAST) reduce 202) ((T_ATTRIBUTE) reduce 202) ((T_BOOL_CAST) reduce 202) ((T_BREAK) reduce 202) ((T_CLASS) reduce 202) ((T_CLASS_C) reduce 202) ((T_CLONE) reduce 202) ((T_CONSTANT_ENCAPSED_STRING) reduce 202) ((T_CONTINUE) reduce 202) ((T_DEC) reduce 202) ((T_DECLARE) reduce 202) ((T_DIR) reduce 202) ((T_DNUMBER) reduce 202) ((T_DO) reduce 202) ((T_DOUBLE_CAST) reduce 202) ((T_ECHO) reduce 202) ((T_EMPTY) reduce 202) ((T_ENDFOREACH) reduce 202) ((T_ENUM) reduce 202) ((T_EVAL) reduce 202) ((T_EXIT) reduce 202) ((T_FILE) reduce 202) ((T_FINAL) reduce 202) ((T_FN) reduce 202) ((T_FOR) reduce 202) ((T_FOREACH) reduce 202) ((T_FUNCTION) reduce 202) ((T_FUNC_C) reduce 202) ((T_GLOBAL) reduce 202) ((T_GOTO) reduce 202) ((T_HALT_COMPILER) reduce 202) ((T_IF) reduce 202) ((T_INC) reduce 202) ((T_INCLUDE) reduce 202) ((T_INCLUDE_ONCE) reduce 202) ((T_INLINE_HTML) reduce 202) ((T_INTERFACE) reduce 202) ((T_INT_CAST) reduce 202) ((T_ISSET) reduce 202) ((T_LINE) reduce 202) ((T_LIST) reduce 202) ((T_LNUMBER) reduce 202) ((T_MATCH) reduce 202) ((T_METHOD_C) reduce 202) ((T_NAME_FULLY_QUALIFIED) reduce 202) ((T_NAME_QUALIFIED) reduce 202) ((T_NAME_RELATIVE) reduce 202) ((T_NEW) reduce 202) ((T_NS_C) reduce 202) ((T_OBJECT_CAST) reduce 202) ((T_PRINT) reduce 202) ((T_PROPERTY_C) reduce 202) ((T_READONLY) reduce 202) ((T_REQUIRE) reduce 202) ((T_REQUIRE_ONCE) reduce 202) ((T_RETURN) reduce 202) ((T_START_HEREDOC) reduce 202) ((T_STATIC) reduce 202) ((T_STRING) reduce 202) ((T_STRING_CAST) reduce 202) ((T_SWITCH) reduce 202) ((T_THROW) reduce 202) ((T_TRAIT) reduce 202) ((T_TRAIT_C) reduce 202) ((T_TRY) reduce 202) ((T_UNSET) reduce 202) ((T_UNSET_CAST) reduce 202) ((T_VARIABLE) reduce 202) ((T_VOID_CAST) reduce 202) ((T_WHILE) reduce 202) ((T_YIELD) reduce 202) ((T_YIELD_FROM) reduce 202) (("[") reduce 202) (("`") reduce 202) (("{") reduce 202) (("~") reduce 202)) 24022 ((("!") reduce 221) (("\"") reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CASE) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DEFAULT) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_ELSE) reduce 221) ((T_ELSEIF) reduce 221) ((T_EMPTY) reduce 221) ((T_ENDSWITCH) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_VARIABLE) reduce 221) ((T_VOID_CAST) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("~") reduce 221)) 24024 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_EMPTY) reduce 159) ((T_ENDFOREACH) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_VOID_CAST) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 24030 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_EMPTY) reduce 151) ((T_ENDFOREACH) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_VOID_CAST) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 24035 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_EMPTY) reduce 165) ((T_ENDFOREACH) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_VOID_CAST) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 24036 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_EMPTY) reduce 222) ((T_ENDFOREACH) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_VARIABLE) reduce 222) ((T_VOID_CAST) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("~") reduce 222)) 24039 ((("!") reduce 223) (("\"") reduce 223) (("$") reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_EMPTY) reduce 223) ((T_ENDFOREACH) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_VARIABLE) reduce 223) ((T_VOID_CAST) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("~") reduce 223)) 24041 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_EMPTY) reduce 155) ((T_ENDFOREACH) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_VOID_CAST) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 24048 ((("!") reduce 190) (("\"") reduce 190) (("$") reduce 190) (("(") reduce 190) (("+") reduce 190) (("-") reduce 190) ((";") reduce 190) (("@") reduce 190) ((T_ABSTRACT) reduce 190) ((T_ARRAY) reduce 190) ((T_ARRAY_CAST) reduce 190) ((T_ATTRIBUTE) reduce 190) ((T_BOOL_CAST) reduce 190) ((T_BREAK) reduce 190) ((T_CLASS) reduce 190) ((T_CLASS_C) reduce 190) ((T_CLONE) reduce 190) ((T_CONSTANT_ENCAPSED_STRING) reduce 190) ((T_CONTINUE) reduce 190) ((T_DEC) reduce 190) ((T_DECLARE) reduce 190) ((T_DIR) reduce 190) ((T_DNUMBER) reduce 190) ((T_DO) reduce 190) ((T_DOUBLE_CAST) reduce 190) ((T_ECHO) reduce 190) ((T_EMPTY) reduce 190) ((T_ENDFOREACH) reduce 190) ((T_ENUM) reduce 190) ((T_EVAL) reduce 190) ((T_EXIT) reduce 190) ((T_FILE) reduce 190) ((T_FINAL) reduce 190) ((T_FN) reduce 190) ((T_FOR) reduce 190) ((T_FOREACH) reduce 190) ((T_FUNCTION) reduce 190) ((T_FUNC_C) reduce 190) ((T_GLOBAL) reduce 190) ((T_GOTO) reduce 190) ((T_HALT_COMPILER) reduce 190) ((T_IF) reduce 190) ((T_INC) reduce 190) ((T_INCLUDE) reduce 190) ((T_INCLUDE_ONCE) reduce 190) ((T_INLINE_HTML) reduce 190) ((T_INTERFACE) reduce 190) ((T_INT_CAST) reduce 190) ((T_ISSET) reduce 190) ((T_LINE) reduce 190) ((T_LIST) reduce 190) ((T_LNUMBER) reduce 190) ((T_MATCH) reduce 190) ((T_METHOD_C) reduce 190) ((T_NAME_FULLY_QUALIFIED) reduce 190) ((T_NAME_QUALIFIED) reduce 190) ((T_NAME_RELATIVE) reduce 190) ((T_NEW) reduce 190) ((T_NS_C) reduce 190) ((T_OBJECT_CAST) reduce 190) ((T_PRINT) reduce 190) ((T_PROPERTY_C) reduce 190) ((T_READONLY) reduce 190) ((T_REQUIRE) reduce 190) ((T_REQUIRE_ONCE) reduce 190) ((T_RETURN) reduce 190) ((T_START_HEREDOC) reduce 190) ((T_STATIC) reduce 190) ((T_STRING) reduce 190) ((T_STRING_CAST) reduce 190) ((T_SWITCH) reduce 190) ((T_THROW) reduce 190) ((T_TRAIT) reduce 190) ((T_TRAIT_C) reduce 190) ((T_TRY) reduce 190) ((T_UNSET) reduce 190) ((T_UNSET_CAST) reduce 190) ((T_VARIABLE) reduce 190) ((T_VOID_CAST) reduce 190) ((T_WHILE) reduce 190) ((T_YIELD) reduce 190) ((T_YIELD_FROM) reduce 190) (("[") reduce 190) (("`") reduce 190) (("{") reduce 190) (("~") reduce 190)) 24050 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_EMPTY) reduce 154) ((T_ENDFOREACH) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_VOID_CAST) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 24053 ((("(") reduce 581) ((")") reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) (("[") reduce 571)) 24054 ((("(") reduce 506) ((")") reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) (("[") reduce 506)) 24055 ((("(") reduce 508) ((")") reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) (("[") reduce 508)) 24060 ((("(") reduce 560) ((")") reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) (("[") reduce 560)) 24061 ((("(") reduce 582) ((")") reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) (("[") reduce 582)) 24062 ((("(") shift) ((")") reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) (("[") reduce 566)) 24063 ((("(") reduce 584) ((")") reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) (("[") reduce 584)) 24066 ((("(") reduce 583) ((")") reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) (("[") reduce 583)) 24067 ((("(") reduce 561) ((")") reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) (("[") reduce 561)) 24068 ((("(") shift) ((")") reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) (("[") reduce 567)) 24069 ((("(") reduce 562) ((")") reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) (("[") reduce 562)) 24070 (((")") reduce 215) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) (("[") reduce 548)) 24072 ((("(") reduce 520) ((")") reduce 217) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) (("[") reduce 520)) 24073 ((("(") reduce 505) ((")") reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) (("[") reduce 505)) 24076 (((")") reduce 216)) 24077 ((("(") reduce 570) ((")") reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) (("[") reduce 570)) 24080 ((("(") reduce 569) ((")") reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) (("[") reduce 569)) 24081 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CASE) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DEFAULT) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_ELSE) reduce 163) ((T_ELSEIF) reduce 163) ((T_EMPTY) reduce 163) ((T_ENDSWITCH) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_VOID_CAST) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 24084 ((("(") reduce 581) ((")") reduce 571) ((T_DOUBLE_ARROW) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) (("[") reduce 571)) 24085 ((("(") reduce 506) ((")") reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) (("[") reduce 506)) 24086 ((("(") reduce 508) ((")") reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) (("[") reduce 508)) 24091 ((("(") reduce 560) ((")") reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) (("[") reduce 560)) 24092 ((("(") reduce 582) ((")") reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) (("[") reduce 582)) 24093 ((("(") shift) ((")") reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) (("[") reduce 566)) 24094 ((("(") reduce 584) ((")") reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) (("[") reduce 584)) 24097 ((("(") reduce 583) ((")") reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) (("[") reduce 583)) 24098 ((("(") reduce 561) ((")") reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) (("[") reduce 561)) 24099 ((("(") shift) ((")") reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) (("[") reduce 567)) 24100 ((("(") reduce 562) ((")") reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) (("[") reduce 562)) 24101 (((")") reduce 215) ((T_DOUBLE_ARROW) reduce 215) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) (("[") reduce 548)) 24103 ((("(") reduce 520) ((")") reduce 217) ((T_DOUBLE_ARROW) reduce 217) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) (("[") reduce 520)) 24104 ((("(") reduce 505) ((")") reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) (("[") reduce 505)) 24107 (((")") reduce 216) ((T_DOUBLE_ARROW) reduce 216)) 24108 ((("(") reduce 570) ((")") reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) (("[") reduce 570)) 24111 ((("(") reduce 569) ((")") reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) (("[") reduce 569)) 24803 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CASE) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DEFAULT) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_ELSE) reduce 152) ((T_ELSEIF) reduce 152) ((T_EMPTY) reduce 152) ((T_ENDSWITCH) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_VOID_CAST) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 24804 ((("!") reduce 218) (("\"") reduce 218) (("$") reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CASE) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DEFAULT) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_ELSE) reduce 218) ((T_ELSEIF) reduce 218) ((T_EMPTY) reduce 218) ((T_ENDSWITCH) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_VARIABLE) reduce 218) ((T_VOID_CAST) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("~") reduce 218)) 24807 ((("!") reduce 219) (("\"") reduce 219) (("$") reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CASE) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DEFAULT) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_ELSE) reduce 219) ((T_ELSEIF) reduce 219) ((T_EMPTY) reduce 219) ((T_ENDSWITCH) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_VARIABLE) reduce 219) ((T_VOID_CAST) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("~") reduce 219)) 24809 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CASE) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DEFAULT) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_ELSE) reduce 159) ((T_ELSEIF) reduce 159) ((T_EMPTY) reduce 159) ((T_ENDSWITCH) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_VOID_CAST) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 24815 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CASE) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DEFAULT) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_ELSE) reduce 151) ((T_ELSEIF) reduce 151) ((T_EMPTY) reduce 151) ((T_ENDSWITCH) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_VOID_CAST) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 24820 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CASE) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DEFAULT) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_ELSE) reduce 165) ((T_ELSEIF) reduce 165) ((T_EMPTY) reduce 165) ((T_ENDSWITCH) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_VOID_CAST) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 24821 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CASE) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DEFAULT) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_ELSE) reduce 222) ((T_ELSEIF) reduce 222) ((T_EMPTY) reduce 222) ((T_ENDSWITCH) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_VARIABLE) reduce 222) ((T_VOID_CAST) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("~") reduce 222)) 24824 ((("!") reduce 223) (("\"") reduce 223) (("$") reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CASE) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DEFAULT) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_ELSE) reduce 223) ((T_ELSEIF) reduce 223) ((T_EMPTY) reduce 223) ((T_ENDSWITCH) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_VARIABLE) reduce 223) ((T_VOID_CAST) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("~") reduce 223)) 24826 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CASE) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DEFAULT) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_ELSE) reduce 155) ((T_ELSEIF) reduce 155) ((T_EMPTY) reduce 155) ((T_ENDSWITCH) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_VOID_CAST) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 24828 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CASE) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DEFAULT) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_ELSE) reduce 154) ((T_ELSEIF) reduce 154) ((T_EMPTY) reduce 154) ((T_ENDSWITCH) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_VOID_CAST) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 24829 ((("!") reduce 247) (("\"") reduce 247) (("$") reduce 247) (("(") reduce 247) (("+") reduce 247) (("-") reduce 247) ((";") reduce 247) (("@") reduce 247) ((T_ABSTRACT) reduce 247) ((T_ARRAY) reduce 247) ((T_ARRAY_CAST) reduce 247) ((T_ATTRIBUTE) reduce 247) ((T_BOOL_CAST) reduce 247) ((T_BREAK) reduce 247) ((T_CASE) reduce 247) ((T_CLASS) reduce 247) ((T_CLASS_C) reduce 247) ((T_CLONE) reduce 247) ((T_CONSTANT_ENCAPSED_STRING) reduce 247) ((T_CONTINUE) reduce 247) ((T_DEC) reduce 247) ((T_DECLARE) reduce 247) ((T_DEFAULT) reduce 247) ((T_DIR) reduce 247) ((T_DNUMBER) reduce 247) ((T_DO) reduce 247) ((T_DOUBLE_CAST) reduce 247) ((T_ECHO) reduce 247) ((T_EMPTY) reduce 247) ((T_ENDSWITCH) reduce 247) ((T_ENUM) reduce 247) ((T_EVAL) reduce 247) ((T_EXIT) reduce 247) ((T_FILE) reduce 247) ((T_FINAL) reduce 247) ((T_FN) reduce 247) ((T_FOR) reduce 247) ((T_FOREACH) reduce 247) ((T_FUNCTION) reduce 247) ((T_FUNC_C) reduce 247) ((T_GLOBAL) reduce 247) ((T_GOTO) reduce 247) ((T_HALT_COMPILER) reduce 247) ((T_IF) reduce 247) ((T_INC) reduce 247) ((T_INCLUDE) reduce 247) ((T_INCLUDE_ONCE) reduce 247) ((T_INLINE_HTML) reduce 247) ((T_INTERFACE) reduce 247) ((T_INT_CAST) reduce 247) ((T_ISSET) reduce 247) ((T_LINE) reduce 247) ((T_LIST) reduce 247) ((T_LNUMBER) reduce 247) ((T_MATCH) reduce 247) ((T_METHOD_C) reduce 247) ((T_NAME_FULLY_QUALIFIED) reduce 247) ((T_NAME_QUALIFIED) reduce 247) ((T_NAME_RELATIVE) reduce 247) ((T_NEW) reduce 247) ((T_NS_C) reduce 247) ((T_OBJECT_CAST) reduce 247) ((T_PRINT) reduce 247) ((T_PROPERTY_C) reduce 247) ((T_READONLY) reduce 247) ((T_REQUIRE) reduce 247) ((T_REQUIRE_ONCE) reduce 247) ((T_RETURN) reduce 247) ((T_START_HEREDOC) reduce 247) ((T_STATIC) reduce 247) ((T_STRING) reduce 247) ((T_STRING_CAST) reduce 247) ((T_SWITCH) reduce 247) ((T_THROW) reduce 247) ((T_TRAIT) reduce 247) ((T_TRAIT_C) reduce 247) ((T_TRY) reduce 247) ((T_UNSET) reduce 247) ((T_UNSET_CAST) reduce 247) ((T_VARIABLE) reduce 247) ((T_VOID_CAST) reduce 247) ((T_WHILE) reduce 247) ((T_YIELD) reduce 247) ((T_YIELD_FROM) reduce 247) (("[") reduce 247) (("`") reduce 247) (("{") reduce 247) (("~") reduce 247)) 24840 ((("!") reduce 494) (("\"") reduce 494) (("$") reduce 494) (("(") reduce 494) (("+") reduce 494) (("-") reduce 494) ((";") reduce 494) (("@") reduce 494) ((T_ABSTRACT) reduce 494) ((T_ARRAY) reduce 494) ((T_ARRAY_CAST) reduce 494) ((T_ATTRIBUTE) reduce 494) ((T_BOOL_CAST) reduce 494) ((T_BREAK) reduce 494) ((T_CASE) reduce 494) ((T_CLASS) reduce 494) ((T_CLASS_C) reduce 494) ((T_CLONE) reduce 494) ((T_CONSTANT_ENCAPSED_STRING) reduce 494) ((T_CONTINUE) reduce 494) ((T_DEC) reduce 494) ((T_DECLARE) reduce 494) ((T_DEFAULT) reduce 494) ((T_DIR) reduce 494) ((T_DNUMBER) reduce 494) ((T_DO) reduce 494) ((T_DOUBLE_CAST) reduce 494) ((T_ECHO) reduce 494) ((T_EMPTY) reduce 494) ((T_ENDSWITCH) reduce 494) ((T_ENUM) reduce 494) ((T_EVAL) reduce 494) ((T_EXIT) reduce 494) ((T_FILE) reduce 494) ((T_FINAL) reduce 494) ((T_FN) reduce 494) ((T_FOR) reduce 494) ((T_FOREACH) reduce 494) ((T_FUNCTION) reduce 494) ((T_FUNC_C) reduce 494) ((T_GLOBAL) reduce 494) ((T_GOTO) reduce 494) ((T_HALT_COMPILER) reduce 494) ((T_IF) reduce 494) ((T_INC) reduce 494) ((T_INCLUDE) reduce 494) ((T_INCLUDE_ONCE) reduce 494) ((T_INLINE_HTML) reduce 494) ((T_INTERFACE) reduce 494) ((T_INT_CAST) reduce 494) ((T_ISSET) reduce 494) ((T_LINE) reduce 494) ((T_LIST) reduce 494) ((T_LNUMBER) reduce 494) ((T_MATCH) reduce 494) ((T_METHOD_C) reduce 494) ((T_NAME_FULLY_QUALIFIED) reduce 494) ((T_NAME_QUALIFIED) reduce 494) ((T_NAME_RELATIVE) reduce 494) ((T_NEW) reduce 494) ((T_NS_C) reduce 494) ((T_OBJECT_CAST) reduce 494) ((T_PRINT) reduce 494) ((T_PROPERTY_C) reduce 494) ((T_READONLY) reduce 494) ((T_REQUIRE) reduce 494) ((T_REQUIRE_ONCE) reduce 494) ((T_RETURN) reduce 494) ((T_START_HEREDOC) reduce 494) ((T_STATIC) reduce 494) ((T_STRING) reduce 494) ((T_STRING_CAST) reduce 494) ((T_SWITCH) reduce 494) ((T_THROW) reduce 494) ((T_TRAIT) reduce 494) ((T_TRAIT_C) reduce 494) ((T_TRY) reduce 494) ((T_UNSET) reduce 494) ((T_UNSET_CAST) reduce 494) ((T_VARIABLE) reduce 494) ((T_VOID_CAST) reduce 494) ((T_WHILE) reduce 494) ((T_YIELD) reduce 494) ((T_YIELD_FROM) reduce 494) (("[") reduce 494) (("`") reduce 494) (("{") reduce 494) (("~") reduce 494)) 24841 ((("!") reduce 184) (("\"") reduce 184) (("$") reduce 184) (("(") reduce 184) (("+") reduce 184) (("-") reduce 184) ((";") reduce 184) (("@") reduce 184) ((T_ABSTRACT) reduce 184) ((T_ARRAY) reduce 184) ((T_ARRAY_CAST) reduce 184) ((T_ATTRIBUTE) reduce 184) ((T_BOOL_CAST) reduce 184) ((T_BREAK) reduce 184) ((T_CASE) reduce 184) ((T_CLASS) reduce 184) ((T_CLASS_C) reduce 184) ((T_CLONE) reduce 184) ((T_CONSTANT_ENCAPSED_STRING) reduce 184) ((T_CONTINUE) reduce 184) ((T_DEC) reduce 184) ((T_DECLARE) reduce 184) ((T_DEFAULT) reduce 184) ((T_DIR) reduce 184) ((T_DNUMBER) reduce 184) ((T_DO) reduce 184) ((T_DOUBLE_CAST) reduce 184) ((T_ECHO) reduce 184) ((T_EMPTY) reduce 184) ((T_ENDSWITCH) reduce 184) ((T_ENUM) reduce 184) ((T_EVAL) reduce 184) ((T_EXIT) reduce 184) ((T_FILE) reduce 184) ((T_FINAL) reduce 184) ((T_FN) reduce 184) ((T_FOR) reduce 184) ((T_FOREACH) reduce 184) ((T_FUNCTION) reduce 184) ((T_FUNC_C) reduce 184) ((T_GLOBAL) reduce 184) ((T_GOTO) reduce 184) ((T_HALT_COMPILER) reduce 184) ((T_IF) reduce 184) ((T_INC) reduce 184) ((T_INCLUDE) reduce 184) ((T_INCLUDE_ONCE) reduce 184) ((T_INLINE_HTML) reduce 184) ((T_INTERFACE) reduce 184) ((T_INT_CAST) reduce 184) ((T_ISSET) reduce 184) ((T_LINE) reduce 184) ((T_LIST) reduce 184) ((T_LNUMBER) reduce 184) ((T_MATCH) reduce 184) ((T_METHOD_C) reduce 184) ((T_NAME_FULLY_QUALIFIED) reduce 184) ((T_NAME_QUALIFIED) reduce 184) ((T_NAME_RELATIVE) reduce 184) ((T_NEW) reduce 184) ((T_NS_C) reduce 184) ((T_OBJECT_CAST) reduce 184) ((T_PRINT) reduce 184) ((T_PROPERTY_C) reduce 184) ((T_READONLY) reduce 184) ((T_REQUIRE) reduce 184) ((T_REQUIRE_ONCE) reduce 184) ((T_RETURN) reduce 184) ((T_START_HEREDOC) reduce 184) ((T_STATIC) reduce 184) ((T_STRING) reduce 184) ((T_STRING_CAST) reduce 184) ((T_SWITCH) reduce 184) ((T_THROW) reduce 184) ((T_TRAIT) reduce 184) ((T_TRAIT_C) reduce 184) ((T_TRY) reduce 184) ((T_UNSET) reduce 184) ((T_UNSET_CAST) reduce 184) ((T_VARIABLE) reduce 184) ((T_VOID_CAST) reduce 184) ((T_WHILE) reduce 184) ((T_YIELD) reduce 184) ((T_YIELD_FROM) reduce 184) (("[") reduce 184) (("`") reduce 184) (("{") reduce 184) (("~") reduce 184)) 24842 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CASE) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DEFAULT) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_EMPTY) reduce 161) ((T_ENDSWITCH) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_VOID_CAST) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 24850 ((("!") reduce 189) (("\"") reduce 189) (("$") reduce 189) (("(") reduce 189) (("+") reduce 189) (("-") reduce 189) ((";") reduce 189) (("@") reduce 189) ((T_ABSTRACT) reduce 189) ((T_ARRAY) reduce 189) ((T_ARRAY_CAST) reduce 189) ((T_ATTRIBUTE) reduce 189) ((T_BOOL_CAST) reduce 189) ((T_BREAK) reduce 189) ((T_CASE) reduce 189) ((T_CLASS) reduce 189) ((T_CLASS_C) reduce 189) ((T_CLONE) reduce 189) ((T_CONSTANT_ENCAPSED_STRING) reduce 189) ((T_CONTINUE) reduce 189) ((T_DEC) reduce 189) ((T_DECLARE) reduce 189) ((T_DEFAULT) reduce 189) ((T_DIR) reduce 189) ((T_DNUMBER) reduce 189) ((T_DO) reduce 189) ((T_DOUBLE_CAST) reduce 189) ((T_ECHO) reduce 189) ((T_EMPTY) reduce 189) ((T_ENDSWITCH) reduce 189) ((T_ENUM) reduce 189) ((T_EVAL) reduce 189) ((T_EXIT) reduce 189) ((T_FILE) reduce 189) ((T_FINAL) reduce 189) ((T_FN) reduce 189) ((T_FOR) reduce 189) ((T_FOREACH) reduce 189) ((T_FUNCTION) reduce 189) ((T_FUNC_C) reduce 189) ((T_GLOBAL) reduce 189) ((T_GOTO) reduce 189) ((T_HALT_COMPILER) reduce 189) ((T_IF) reduce 189) ((T_INC) reduce 189) ((T_INCLUDE) reduce 189) ((T_INCLUDE_ONCE) reduce 189) ((T_INLINE_HTML) reduce 189) ((T_INTERFACE) reduce 189) ((T_INT_CAST) reduce 189) ((T_ISSET) reduce 189) ((T_LINE) reduce 189) ((T_LIST) reduce 189) ((T_LNUMBER) reduce 189) ((T_MATCH) reduce 189) ((T_METHOD_C) reduce 189) ((T_NAME_FULLY_QUALIFIED) reduce 189) ((T_NAME_QUALIFIED) reduce 189) ((T_NAME_RELATIVE) reduce 189) ((T_NEW) reduce 189) ((T_NS_C) reduce 189) ((T_OBJECT_CAST) reduce 189) ((T_PRINT) reduce 189) ((T_PROPERTY_C) reduce 189) ((T_READONLY) reduce 189) ((T_REQUIRE) reduce 189) ((T_REQUIRE_ONCE) reduce 189) ((T_RETURN) reduce 189) ((T_START_HEREDOC) reduce 189) ((T_STATIC) reduce 189) ((T_STRING) reduce 189) ((T_STRING_CAST) reduce 189) ((T_SWITCH) reduce 189) ((T_THROW) reduce 189) ((T_TRAIT) reduce 189) ((T_TRAIT_C) reduce 189) ((T_TRY) reduce 189) ((T_UNSET) reduce 189) ((T_UNSET_CAST) reduce 189) ((T_VARIABLE) reduce 189) ((T_VOID_CAST) reduce 189) ((T_WHILE) reduce 189) ((T_YIELD) reduce 189) ((T_YIELD_FROM) reduce 189) (("[") reduce 189) (("`") reduce 189) (("{") reduce 189) (("~") reduce 189)) 24851 ((("!") reduce 145) (("\"") reduce 145) (("$") reduce 145) (("(") reduce 145) (("+") reduce 145) (("-") reduce 145) ((";") reduce 145) (("@") reduce 145) ((T_ABSTRACT) reduce 145) ((T_ARRAY) reduce 145) ((T_ARRAY_CAST) reduce 145) ((T_ATTRIBUTE) reduce 145) ((T_BOOL_CAST) reduce 145) ((T_BREAK) reduce 145) ((T_CASE) reduce 145) ((T_CLASS) reduce 145) ((T_CLASS_C) reduce 145) ((T_CLONE) reduce 145) ((T_CONSTANT_ENCAPSED_STRING) reduce 145) ((T_CONTINUE) reduce 145) ((T_DEC) reduce 145) ((T_DECLARE) reduce 145) ((T_DEFAULT) reduce 145) ((T_DIR) reduce 145) ((T_DNUMBER) reduce 145) ((T_DO) reduce 145) ((T_DOUBLE_CAST) reduce 145) ((T_ECHO) reduce 145) ((T_EMPTY) reduce 145) ((T_ENDSWITCH) reduce 145) ((T_ENUM) reduce 145) ((T_EVAL) reduce 145) ((T_EXIT) reduce 145) ((T_FILE) reduce 145) ((T_FINAL) reduce 145) ((T_FN) reduce 145) ((T_FOR) reduce 145) ((T_FOREACH) reduce 145) ((T_FUNCTION) reduce 145) ((T_FUNC_C) reduce 145) ((T_GLOBAL) reduce 145) ((T_GOTO) reduce 145) ((T_HALT_COMPILER) reduce 145) ((T_IF) reduce 145) ((T_INC) reduce 145) ((T_INCLUDE) reduce 145) ((T_INCLUDE_ONCE) reduce 145) ((T_INLINE_HTML) reduce 145) ((T_INTERFACE) reduce 145) ((T_INT_CAST) reduce 145) ((T_ISSET) reduce 145) ((T_LINE) reduce 145) ((T_LIST) reduce 145) ((T_LNUMBER) reduce 145) ((T_MATCH) reduce 145) ((T_METHOD_C) reduce 145) ((T_NAME_FULLY_QUALIFIED) reduce 145) ((T_NAME_QUALIFIED) reduce 145) ((T_NAME_RELATIVE) reduce 145) ((T_NEW) reduce 145) ((T_NS_C) reduce 145) ((T_OBJECT_CAST) reduce 145) ((T_PRINT) reduce 145) ((T_PROPERTY_C) reduce 145) ((T_READONLY) reduce 145) ((T_REQUIRE) reduce 145) ((T_REQUIRE_ONCE) reduce 145) ((T_RETURN) reduce 145) ((T_START_HEREDOC) reduce 145) ((T_STATIC) reduce 145) ((T_STRING) reduce 145) ((T_STRING_CAST) reduce 145) ((T_SWITCH) reduce 145) ((T_THROW) reduce 145) ((T_TRAIT) reduce 145) ((T_TRAIT_C) reduce 145) ((T_TRY) reduce 145) ((T_UNSET) reduce 145) ((T_UNSET_CAST) reduce 145) ((T_VARIABLE) reduce 145) ((T_VOID_CAST) reduce 145) ((T_WHILE) reduce 145) ((T_YIELD) reduce 145) ((T_YIELD_FROM) reduce 145) (("[") reduce 145) (("`") reduce 145) (("{") reduce 145) (("~") reduce 145)) 24854 ((("!") reduce 250) (("\"") reduce 250) (("$") reduce 250) (("(") reduce 250) (("+") reduce 250) (("-") reduce 250) ((";") reduce 250) (("@") reduce 250) ((T_ABSTRACT) reduce 250) ((T_ARRAY) reduce 250) ((T_ARRAY_CAST) reduce 250) ((T_ATTRIBUTE) reduce 250) ((T_BOOL_CAST) reduce 250) ((T_BREAK) reduce 250) ((T_CASE) reduce 250) ((T_CLASS) reduce 250) ((T_CLASS_C) reduce 250) ((T_CLONE) reduce 250) ((T_CONSTANT_ENCAPSED_STRING) reduce 250) ((T_CONTINUE) reduce 250) ((T_DEC) reduce 250) ((T_DECLARE) reduce 250) ((T_DEFAULT) reduce 250) ((T_DIR) reduce 250) ((T_DNUMBER) reduce 250) ((T_DO) reduce 250) ((T_DOUBLE_CAST) reduce 250) ((T_ECHO) reduce 250) ((T_EMPTY) reduce 250) ((T_ENDSWITCH) reduce 250) ((T_ENUM) reduce 250) ((T_EVAL) reduce 250) ((T_EXIT) reduce 250) ((T_FILE) reduce 250) ((T_FINAL) reduce 250) ((T_FN) reduce 250) ((T_FOR) reduce 250) ((T_FOREACH) reduce 250) ((T_FUNCTION) reduce 250) ((T_FUNC_C) reduce 250) ((T_GLOBAL) reduce 250) ((T_GOTO) reduce 250) ((T_HALT_COMPILER) reduce 250) ((T_IF) reduce 250) ((T_INC) reduce 250) ((T_INCLUDE) reduce 250) ((T_INCLUDE_ONCE) reduce 250) ((T_INLINE_HTML) reduce 250) ((T_INTERFACE) reduce 250) ((T_INT_CAST) reduce 250) ((T_ISSET) reduce 250) ((T_LINE) reduce 250) ((T_LIST) reduce 250) ((T_LNUMBER) reduce 250) ((T_MATCH) reduce 250) ((T_METHOD_C) reduce 250) ((T_NAME_FULLY_QUALIFIED) reduce 250) ((T_NAME_QUALIFIED) reduce 250) ((T_NAME_RELATIVE) reduce 250) ((T_NEW) reduce 250) ((T_NS_C) reduce 250) ((T_OBJECT_CAST) reduce 250) ((T_PRINT) reduce 250) ((T_PROPERTY_C) reduce 250) ((T_READONLY) reduce 250) ((T_REQUIRE) reduce 250) ((T_REQUIRE_ONCE) reduce 250) ((T_RETURN) reduce 250) ((T_START_HEREDOC) reduce 250) ((T_STATIC) reduce 250) ((T_STRING) reduce 250) ((T_STRING_CAST) reduce 250) ((T_SWITCH) reduce 250) ((T_THROW) reduce 250) ((T_TRAIT) reduce 250) ((T_TRAIT_C) reduce 250) ((T_TRY) reduce 250) ((T_UNSET) reduce 250) ((T_UNSET_CAST) reduce 250) ((T_VARIABLE) reduce 250) ((T_VOID_CAST) reduce 250) ((T_WHILE) reduce 250) ((T_YIELD) reduce 250) ((T_YIELD_FROM) reduce 250) (("[") reduce 250) (("`") reduce 250) (("{") reduce 250) (("~") reduce 250)) 24858 ((("!") reduce 251) (("\"") reduce 251) (("$") reduce 251) (("(") reduce 251) (("+") reduce 251) (("-") reduce 251) ((";") reduce 251) (("@") reduce 251) ((T_ABSTRACT) reduce 251) ((T_ARRAY) reduce 251) ((T_ARRAY_CAST) reduce 251) ((T_ATTRIBUTE) reduce 251) ((T_BOOL_CAST) reduce 251) ((T_BREAK) reduce 251) ((T_CASE) reduce 251) ((T_CLASS) reduce 251) ((T_CLASS_C) reduce 251) ((T_CLONE) reduce 251) ((T_CONSTANT_ENCAPSED_STRING) reduce 251) ((T_CONTINUE) reduce 251) ((T_DEC) reduce 251) ((T_DECLARE) reduce 251) ((T_DEFAULT) reduce 251) ((T_DIR) reduce 251) ((T_DNUMBER) reduce 251) ((T_DO) reduce 251) ((T_DOUBLE_CAST) reduce 251) ((T_ECHO) reduce 251) ((T_EMPTY) reduce 251) ((T_ENDSWITCH) reduce 251) ((T_ENUM) reduce 251) ((T_EVAL) reduce 251) ((T_EXIT) reduce 251) ((T_FILE) reduce 251) ((T_FINAL) reduce 251) ((T_FN) reduce 251) ((T_FOR) reduce 251) ((T_FOREACH) reduce 251) ((T_FUNCTION) reduce 251) ((T_FUNC_C) reduce 251) ((T_GLOBAL) reduce 251) ((T_GOTO) reduce 251) ((T_HALT_COMPILER) reduce 251) ((T_IF) reduce 251) ((T_INC) reduce 251) ((T_INCLUDE) reduce 251) ((T_INCLUDE_ONCE) reduce 251) ((T_INLINE_HTML) reduce 251) ((T_INTERFACE) reduce 251) ((T_INT_CAST) reduce 251) ((T_ISSET) reduce 251) ((T_LINE) reduce 251) ((T_LIST) reduce 251) ((T_LNUMBER) reduce 251) ((T_MATCH) reduce 251) ((T_METHOD_C) reduce 251) ((T_NAME_FULLY_QUALIFIED) reduce 251) ((T_NAME_QUALIFIED) reduce 251) ((T_NAME_RELATIVE) reduce 251) ((T_NEW) reduce 251) ((T_NS_C) reduce 251) ((T_OBJECT_CAST) reduce 251) ((T_PRINT) reduce 251) ((T_PROPERTY_C) reduce 251) ((T_READONLY) reduce 251) ((T_REQUIRE) reduce 251) ((T_REQUIRE_ONCE) reduce 251) ((T_RETURN) reduce 251) ((T_START_HEREDOC) reduce 251) ((T_STATIC) reduce 251) ((T_STRING) reduce 251) ((T_STRING_CAST) reduce 251) ((T_SWITCH) reduce 251) ((T_THROW) reduce 251) ((T_TRAIT) reduce 251) ((T_TRAIT_C) reduce 251) ((T_TRY) reduce 251) ((T_UNSET) reduce 251) ((T_UNSET_CAST) reduce 251) ((T_VARIABLE) reduce 251) ((T_VOID_CAST) reduce 251) ((T_WHILE) reduce 251) ((T_YIELD) reduce 251) ((T_YIELD_FROM) reduce 251) (("[") reduce 251) (("`") reduce 251) (("{") reduce 251) (("~") reduce 251)) 24863 ((("!") reduce 242) (("\"") reduce 242) (("$") reduce 242) (("(") reduce 242) (("+") reduce 242) (("-") reduce 242) ((";") reduce 242) (("@") reduce 242) ((T_ABSTRACT) reduce 242) ((T_ARRAY) reduce 242) ((T_ARRAY_CAST) reduce 242) ((T_ATTRIBUTE) reduce 242) ((T_BOOL_CAST) reduce 242) ((T_BREAK) reduce 242) ((T_CASE) reduce 242) ((T_CLASS) reduce 242) ((T_CLASS_C) reduce 242) ((T_CLONE) reduce 242) ((T_CONSTANT_ENCAPSED_STRING) reduce 242) ((T_CONTINUE) reduce 242) ((T_DEC) reduce 242) ((T_DECLARE) reduce 242) ((T_DEFAULT) reduce 242) ((T_DIR) reduce 242) ((T_DNUMBER) reduce 242) ((T_DO) reduce 242) ((T_DOUBLE_CAST) reduce 242) ((T_ECHO) reduce 242) ((T_EMPTY) reduce 242) ((T_ENDSWITCH) reduce 242) ((T_ENUM) reduce 242) ((T_EVAL) reduce 242) ((T_EXIT) reduce 242) ((T_FILE) reduce 242) ((T_FINAL) reduce 242) ((T_FN) reduce 242) ((T_FOR) reduce 242) ((T_FOREACH) reduce 242) ((T_FUNCTION) reduce 242) ((T_FUNC_C) reduce 242) ((T_GLOBAL) reduce 242) ((T_GOTO) reduce 242) ((T_HALT_COMPILER) reduce 242) ((T_IF) reduce 242) ((T_INC) reduce 242) ((T_INCLUDE) reduce 242) ((T_INCLUDE_ONCE) reduce 242) ((T_INLINE_HTML) reduce 242) ((T_INTERFACE) reduce 242) ((T_INT_CAST) reduce 242) ((T_ISSET) reduce 242) ((T_LINE) reduce 242) ((T_LIST) reduce 242) ((T_LNUMBER) reduce 242) ((T_MATCH) reduce 242) ((T_METHOD_C) reduce 242) ((T_NAME_FULLY_QUALIFIED) reduce 242) ((T_NAME_QUALIFIED) reduce 242) ((T_NAME_RELATIVE) reduce 242) ((T_NEW) reduce 242) ((T_NS_C) reduce 242) ((T_OBJECT_CAST) reduce 242) ((T_PRINT) reduce 242) ((T_PROPERTY_C) reduce 242) ((T_READONLY) reduce 242) ((T_REQUIRE) reduce 242) ((T_REQUIRE_ONCE) reduce 242) ((T_RETURN) reduce 242) ((T_START_HEREDOC) reduce 242) ((T_STATIC) reduce 242) ((T_STRING) reduce 242) ((T_STRING_CAST) reduce 242) ((T_SWITCH) reduce 242) ((T_THROW) reduce 242) ((T_TRAIT) reduce 242) ((T_TRAIT_C) reduce 242) ((T_TRY) reduce 242) ((T_UNSET) reduce 242) ((T_UNSET_CAST) reduce 242) ((T_VARIABLE) reduce 242) ((T_VOID_CAST) reduce 242) ((T_WHILE) reduce 242) ((T_YIELD) reduce 242) ((T_YIELD_FROM) reduce 242) (("[") reduce 242) (("`") reduce 242) (("{") reduce 242) (("~") reduce 242)) 24864 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CASE) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DEFAULT) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_EMPTY) reduce 150) ((T_ENDSWITCH) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_VOID_CAST) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 24867 ((("!") reduce 243) (("\"") reduce 243) (("$") reduce 243) (("(") reduce 243) (("+") reduce 243) (("-") reduce 243) ((";") reduce 243) (("@") reduce 243) ((T_ABSTRACT) reduce 243) ((T_ARRAY) reduce 243) ((T_ARRAY_CAST) reduce 243) ((T_ATTRIBUTE) reduce 243) ((T_BOOL_CAST) reduce 243) ((T_BREAK) reduce 243) ((T_CASE) reduce 243) ((T_CLASS) reduce 243) ((T_CLASS_C) reduce 243) ((T_CLONE) reduce 243) ((T_CONSTANT_ENCAPSED_STRING) reduce 243) ((T_CONTINUE) reduce 243) ((T_DEC) reduce 243) ((T_DECLARE) reduce 243) ((T_DEFAULT) reduce 243) ((T_DIR) reduce 243) ((T_DNUMBER) reduce 243) ((T_DO) reduce 243) ((T_DOUBLE_CAST) reduce 243) ((T_ECHO) reduce 243) ((T_EMPTY) reduce 243) ((T_ENDSWITCH) reduce 243) ((T_ENUM) reduce 243) ((T_EVAL) reduce 243) ((T_EXIT) reduce 243) ((T_FILE) reduce 243) ((T_FINAL) reduce 243) ((T_FN) reduce 243) ((T_FOR) reduce 243) ((T_FOREACH) reduce 243) ((T_FUNCTION) reduce 243) ((T_FUNC_C) reduce 243) ((T_GLOBAL) reduce 243) ((T_GOTO) reduce 243) ((T_HALT_COMPILER) reduce 243) ((T_IF) reduce 243) ((T_INC) reduce 243) ((T_INCLUDE) reduce 243) ((T_INCLUDE_ONCE) reduce 243) ((T_INLINE_HTML) reduce 243) ((T_INTERFACE) reduce 243) ((T_INT_CAST) reduce 243) ((T_ISSET) reduce 243) ((T_LINE) reduce 243) ((T_LIST) reduce 243) ((T_LNUMBER) reduce 243) ((T_MATCH) reduce 243) ((T_METHOD_C) reduce 243) ((T_NAME_FULLY_QUALIFIED) reduce 243) ((T_NAME_QUALIFIED) reduce 243) ((T_NAME_RELATIVE) reduce 243) ((T_NEW) reduce 243) ((T_NS_C) reduce 243) ((T_OBJECT_CAST) reduce 243) ((T_PRINT) reduce 243) ((T_PROPERTY_C) reduce 243) ((T_READONLY) reduce 243) ((T_REQUIRE) reduce 243) ((T_REQUIRE_ONCE) reduce 243) ((T_RETURN) reduce 243) ((T_START_HEREDOC) reduce 243) ((T_STATIC) reduce 243) ((T_STRING) reduce 243) ((T_STRING_CAST) reduce 243) ((T_SWITCH) reduce 243) ((T_THROW) reduce 243) ((T_TRAIT) reduce 243) ((T_TRAIT_C) reduce 243) ((T_TRY) reduce 243) ((T_UNSET) reduce 243) ((T_UNSET_CAST) reduce 243) ((T_VARIABLE) reduce 243) ((T_VOID_CAST) reduce 243) ((T_WHILE) reduce 243) ((T_YIELD) reduce 243) ((T_YIELD_FROM) reduce 243) (("[") reduce 243) (("`") reduce 243) (("{") reduce 243) (("~") reduce 243)) 24869 ((("!") reduce 170) (("\"") reduce 170) (("$") reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CASE) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DEFAULT) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_EMPTY) reduce 170) ((T_ENDSWITCH) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_VARIABLE) reduce 170) ((T_VOID_CAST) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("~") reduce 170)) 24874 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CASE) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DEFAULT) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_EMPTY) reduce 162) ((T_ENDSWITCH) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_VOID_CAST) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 24877 ((("!") reduce 171) (("\"") reduce 171) (("$") reduce 171) (("(") reduce 171) (("+") reduce 171) (("-") reduce 171) ((";") reduce 171) (("@") reduce 171) ((T_ABSTRACT) reduce 171) ((T_ARRAY) reduce 171) ((T_ARRAY_CAST) reduce 171) ((T_ATTRIBUTE) reduce 171) ((T_BOOL_CAST) reduce 171) ((T_BREAK) reduce 171) ((T_CASE) reduce 171) ((T_CATCH) reduce 171) ((T_CLASS) reduce 171) ((T_CLASS_C) reduce 171) ((T_CLONE) reduce 171) ((T_CONSTANT_ENCAPSED_STRING) reduce 171) ((T_CONTINUE) reduce 171) ((T_DEC) reduce 171) ((T_DECLARE) reduce 171) ((T_DEFAULT) reduce 171) ((T_DIR) reduce 171) ((T_DNUMBER) reduce 171) ((T_DO) reduce 171) ((T_DOUBLE_CAST) reduce 171) ((T_ECHO) reduce 171) ((T_EMPTY) reduce 171) ((T_ENDSWITCH) reduce 171) ((T_ENUM) reduce 171) ((T_EVAL) reduce 171) ((T_EXIT) reduce 171) ((T_FILE) reduce 171) ((T_FINAL) reduce 171) ((T_FINALLY) reduce 171) ((T_FN) reduce 171) ((T_FOR) reduce 171) ((T_FOREACH) reduce 171) ((T_FUNCTION) reduce 171) ((T_FUNC_C) reduce 171) ((T_GLOBAL) reduce 171) ((T_GOTO) reduce 171) ((T_HALT_COMPILER) reduce 171) ((T_IF) reduce 171) ((T_INC) reduce 171) ((T_INCLUDE) reduce 171) ((T_INCLUDE_ONCE) reduce 171) ((T_INLINE_HTML) reduce 171) ((T_INTERFACE) reduce 171) ((T_INT_CAST) reduce 171) ((T_ISSET) reduce 171) ((T_LINE) reduce 171) ((T_LIST) reduce 171) ((T_LNUMBER) reduce 171) ((T_MATCH) reduce 171) ((T_METHOD_C) reduce 171) ((T_NAME_FULLY_QUALIFIED) reduce 171) ((T_NAME_QUALIFIED) reduce 171) ((T_NAME_RELATIVE) reduce 171) ((T_NEW) reduce 171) ((T_NS_C) reduce 171) ((T_OBJECT_CAST) reduce 171) ((T_PRINT) reduce 171) ((T_PROPERTY_C) reduce 171) ((T_READONLY) reduce 171) ((T_REQUIRE) reduce 171) ((T_REQUIRE_ONCE) reduce 171) ((T_RETURN) reduce 171) ((T_START_HEREDOC) reduce 171) ((T_STATIC) reduce 171) ((T_STRING) reduce 171) ((T_STRING_CAST) reduce 171) ((T_SWITCH) reduce 171) ((T_THROW) reduce 171) ((T_TRAIT) reduce 171) ((T_TRAIT_C) reduce 171) ((T_TRY) reduce 171) ((T_UNSET) reduce 171) ((T_UNSET_CAST) reduce 171) ((T_VARIABLE) reduce 171) ((T_VOID_CAST) reduce 171) ((T_WHILE) reduce 171) ((T_YIELD) reduce 171) ((T_YIELD_FROM) reduce 171) (("[") reduce 171) (("`") reduce 171) (("{") reduce 171) (("~") reduce 171)) 24878 ((("!") reduce 177) (("\"") reduce 177) (("$") reduce 177) (("(") reduce 177) (("+") reduce 177) (("-") reduce 177) ((";") reduce 177) (("@") reduce 177) ((T_ABSTRACT) reduce 177) ((T_ARRAY) reduce 177) ((T_ARRAY_CAST) reduce 177) ((T_ATTRIBUTE) reduce 177) ((T_BOOL_CAST) reduce 177) ((T_BREAK) reduce 177) ((T_CASE) reduce 177) ((T_CATCH) shift) ((T_CLASS) reduce 177) ((T_CLASS_C) reduce 177) ((T_CLONE) reduce 177) ((T_CONSTANT_ENCAPSED_STRING) reduce 177) ((T_CONTINUE) reduce 177) ((T_DEC) reduce 177) ((T_DECLARE) reduce 177) ((T_DEFAULT) reduce 177) ((T_DIR) reduce 177) ((T_DNUMBER) reduce 177) ((T_DO) reduce 177) ((T_DOUBLE_CAST) reduce 177) ((T_ECHO) reduce 177) ((T_EMPTY) reduce 177) ((T_ENDSWITCH) reduce 177) ((T_ENUM) reduce 177) ((T_EVAL) reduce 177) ((T_EXIT) reduce 177) ((T_FILE) reduce 177) ((T_FINAL) reduce 177) ((T_FINALLY) shift) ((T_FN) reduce 177) ((T_FOR) reduce 177) ((T_FOREACH) reduce 177) ((T_FUNCTION) reduce 177) ((T_FUNC_C) reduce 177) ((T_GLOBAL) reduce 177) ((T_GOTO) reduce 177) ((T_HALT_COMPILER) reduce 177) ((T_IF) reduce 177) ((T_INC) reduce 177) ((T_INCLUDE) reduce 177) ((T_INCLUDE_ONCE) reduce 177) ((T_INLINE_HTML) reduce 177) ((T_INTERFACE) reduce 177) ((T_INT_CAST) reduce 177) ((T_ISSET) reduce 177) ((T_LINE) reduce 177) ((T_LIST) reduce 177) ((T_LNUMBER) reduce 177) ((T_MATCH) reduce 177) ((T_METHOD_C) reduce 177) ((T_NAME_FULLY_QUALIFIED) reduce 177) ((T_NAME_QUALIFIED) reduce 177) ((T_NAME_RELATIVE) reduce 177) ((T_NEW) reduce 177) ((T_NS_C) reduce 177) ((T_OBJECT_CAST) reduce 177) ((T_PRINT) reduce 177) ((T_PROPERTY_C) reduce 177) ((T_READONLY) reduce 177) ((T_REQUIRE) reduce 177) ((T_REQUIRE_ONCE) reduce 177) ((T_RETURN) reduce 177) ((T_START_HEREDOC) reduce 177) ((T_STATIC) reduce 177) ((T_STRING) reduce 177) ((T_STRING_CAST) reduce 177) ((T_SWITCH) reduce 177) ((T_THROW) reduce 177) ((T_TRAIT) reduce 177) ((T_TRAIT_C) reduce 177) ((T_TRY) reduce 177) ((T_UNSET) reduce 177) ((T_UNSET_CAST) reduce 177) ((T_VARIABLE) reduce 177) ((T_VOID_CAST) reduce 177) ((T_WHILE) reduce 177) ((T_YIELD) reduce 177) ((T_YIELD_FROM) reduce 177) (("[") reduce 177) (("`") reduce 177) (("{") reduce 177) (("~") reduce 177)) 24881 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CASE) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DEFAULT) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_EMPTY) reduce 167) ((T_ENDSWITCH) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_VOID_CAST) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 24884 ((("!") reduce 178) (("\"") reduce 178) (("$") reduce 178) (("(") reduce 178) (("+") reduce 178) (("-") reduce 178) ((";") reduce 178) (("@") reduce 178) ((T_ABSTRACT) reduce 178) ((T_ARRAY) reduce 178) ((T_ARRAY_CAST) reduce 178) ((T_ATTRIBUTE) reduce 178) ((T_BOOL_CAST) reduce 178) ((T_BREAK) reduce 178) ((T_CASE) reduce 178) ((T_CLASS) reduce 178) ((T_CLASS_C) reduce 178) ((T_CLONE) reduce 178) ((T_CONSTANT_ENCAPSED_STRING) reduce 178) ((T_CONTINUE) reduce 178) ((T_DEC) reduce 178) ((T_DECLARE) reduce 178) ((T_DEFAULT) reduce 178) ((T_DIR) reduce 178) ((T_DNUMBER) reduce 178) ((T_DO) reduce 178) ((T_DOUBLE_CAST) reduce 178) ((T_ECHO) reduce 178) ((T_EMPTY) reduce 178) ((T_ENDSWITCH) reduce 178) ((T_ENUM) reduce 178) ((T_EVAL) reduce 178) ((T_EXIT) reduce 178) ((T_FILE) reduce 178) ((T_FINAL) reduce 178) ((T_FN) reduce 178) ((T_FOR) reduce 178) ((T_FOREACH) reduce 178) ((T_FUNCTION) reduce 178) ((T_FUNC_C) reduce 178) ((T_GLOBAL) reduce 178) ((T_GOTO) reduce 178) ((T_HALT_COMPILER) reduce 178) ((T_IF) reduce 178) ((T_INC) reduce 178) ((T_INCLUDE) reduce 178) ((T_INCLUDE_ONCE) reduce 178) ((T_INLINE_HTML) reduce 178) ((T_INTERFACE) reduce 178) ((T_INT_CAST) reduce 178) ((T_ISSET) reduce 178) ((T_LINE) reduce 178) ((T_LIST) reduce 178) ((T_LNUMBER) reduce 178) ((T_MATCH) reduce 178) ((T_METHOD_C) reduce 178) ((T_NAME_FULLY_QUALIFIED) reduce 178) ((T_NAME_QUALIFIED) reduce 178) ((T_NAME_RELATIVE) reduce 178) ((T_NEW) reduce 178) ((T_NS_C) reduce 178) ((T_OBJECT_CAST) reduce 178) ((T_PRINT) reduce 178) ((T_PROPERTY_C) reduce 178) ((T_READONLY) reduce 178) ((T_REQUIRE) reduce 178) ((T_REQUIRE_ONCE) reduce 178) ((T_RETURN) reduce 178) ((T_START_HEREDOC) reduce 178) ((T_STATIC) reduce 178) ((T_STRING) reduce 178) ((T_STRING_CAST) reduce 178) ((T_SWITCH) reduce 178) ((T_THROW) reduce 178) ((T_TRAIT) reduce 178) ((T_TRAIT_C) reduce 178) ((T_TRY) reduce 178) ((T_UNSET) reduce 178) ((T_UNSET_CAST) reduce 178) ((T_VARIABLE) reduce 178) ((T_VOID_CAST) reduce 178) ((T_WHILE) reduce 178) ((T_YIELD) reduce 178) ((T_YIELD_FROM) reduce 178) (("[") reduce 178) (("`") reduce 178) (("{") reduce 178) (("~") reduce 178)) 24891 ((("!") reduce 172) (("\"") reduce 172) (("$") reduce 172) (("(") reduce 172) (("+") reduce 172) (("-") reduce 172) ((";") reduce 172) (("@") reduce 172) ((T_ABSTRACT) reduce 172) ((T_ARRAY) reduce 172) ((T_ARRAY_CAST) reduce 172) ((T_ATTRIBUTE) reduce 172) ((T_BOOL_CAST) reduce 172) ((T_BREAK) reduce 172) ((T_CASE) reduce 172) ((T_CATCH) reduce 172) ((T_CLASS) reduce 172) ((T_CLASS_C) reduce 172) ((T_CLONE) reduce 172) ((T_CONSTANT_ENCAPSED_STRING) reduce 172) ((T_CONTINUE) reduce 172) ((T_DEC) reduce 172) ((T_DECLARE) reduce 172) ((T_DEFAULT) reduce 172) ((T_DIR) reduce 172) ((T_DNUMBER) reduce 172) ((T_DO) reduce 172) ((T_DOUBLE_CAST) reduce 172) ((T_ECHO) reduce 172) ((T_EMPTY) reduce 172) ((T_ENDSWITCH) reduce 172) ((T_ENUM) reduce 172) ((T_EVAL) reduce 172) ((T_EXIT) reduce 172) ((T_FILE) reduce 172) ((T_FINAL) reduce 172) ((T_FINALLY) reduce 172) ((T_FN) reduce 172) ((T_FOR) reduce 172) ((T_FOREACH) reduce 172) ((T_FUNCTION) reduce 172) ((T_FUNC_C) reduce 172) ((T_GLOBAL) reduce 172) ((T_GOTO) reduce 172) ((T_HALT_COMPILER) reduce 172) ((T_IF) reduce 172) ((T_INC) reduce 172) ((T_INCLUDE) reduce 172) ((T_INCLUDE_ONCE) reduce 172) ((T_INLINE_HTML) reduce 172) ((T_INTERFACE) reduce 172) ((T_INT_CAST) reduce 172) ((T_ISSET) reduce 172) ((T_LINE) reduce 172) ((T_LIST) reduce 172) ((T_LNUMBER) reduce 172) ((T_MATCH) reduce 172) ((T_METHOD_C) reduce 172) ((T_NAME_FULLY_QUALIFIED) reduce 172) ((T_NAME_QUALIFIED) reduce 172) ((T_NAME_RELATIVE) reduce 172) ((T_NEW) reduce 172) ((T_NS_C) reduce 172) ((T_OBJECT_CAST) reduce 172) ((T_PRINT) reduce 172) ((T_PROPERTY_C) reduce 172) ((T_READONLY) reduce 172) ((T_REQUIRE) reduce 172) ((T_REQUIRE_ONCE) reduce 172) ((T_RETURN) reduce 172) ((T_START_HEREDOC) reduce 172) ((T_STATIC) reduce 172) ((T_STRING) reduce 172) ((T_STRING_CAST) reduce 172) ((T_SWITCH) reduce 172) ((T_THROW) reduce 172) ((T_TRAIT) reduce 172) ((T_TRAIT_C) reduce 172) ((T_TRY) reduce 172) ((T_UNSET) reduce 172) ((T_UNSET_CAST) reduce 172) ((T_VARIABLE) reduce 172) ((T_VOID_CAST) reduce 172) ((T_WHILE) reduce 172) ((T_YIELD) reduce 172) ((T_YIELD_FROM) reduce 172) (("[") reduce 172) (("`") reduce 172) (("{") reduce 172) (("~") reduce 172)) 24896 ((("!") reduce 200) (("\"") reduce 200) (("$") reduce 200) (("(") reduce 200) (("+") reduce 200) (("-") reduce 200) ((";") reduce 200) (("@") reduce 200) ((T_ABSTRACT) reduce 200) ((T_ARRAY) reduce 200) ((T_ARRAY_CAST) reduce 200) ((T_ATTRIBUTE) reduce 200) ((T_BOOL_CAST) reduce 200) ((T_BREAK) reduce 200) ((T_CASE) reduce 200) ((T_CLASS) reduce 200) ((T_CLASS_C) reduce 200) ((T_CLONE) reduce 200) ((T_CONSTANT_ENCAPSED_STRING) reduce 200) ((T_CONTINUE) reduce 200) ((T_DEC) reduce 200) ((T_DECLARE) reduce 200) ((T_DEFAULT) reduce 200) ((T_DIR) reduce 200) ((T_DNUMBER) reduce 200) ((T_DO) reduce 200) ((T_DOUBLE_CAST) reduce 200) ((T_ECHO) reduce 200) ((T_EMPTY) reduce 200) ((T_ENDSWITCH) reduce 200) ((T_ENUM) reduce 200) ((T_EVAL) reduce 200) ((T_EXIT) reduce 200) ((T_FILE) reduce 200) ((T_FINAL) reduce 200) ((T_FN) reduce 200) ((T_FOR) reduce 200) ((T_FOREACH) reduce 200) ((T_FUNCTION) reduce 200) ((T_FUNC_C) reduce 200) ((T_GLOBAL) reduce 200) ((T_GOTO) reduce 200) ((T_HALT_COMPILER) reduce 200) ((T_IF) reduce 200) ((T_INC) reduce 200) ((T_INCLUDE) reduce 200) ((T_INCLUDE_ONCE) reduce 200) ((T_INLINE_HTML) reduce 200) ((T_INTERFACE) reduce 200) ((T_INT_CAST) reduce 200) ((T_ISSET) reduce 200) ((T_LINE) reduce 200) ((T_LIST) reduce 200) ((T_LNUMBER) reduce 200) ((T_MATCH) reduce 200) ((T_METHOD_C) reduce 200) ((T_NAME_FULLY_QUALIFIED) reduce 200) ((T_NAME_QUALIFIED) reduce 200) ((T_NAME_RELATIVE) reduce 200) ((T_NEW) reduce 200) ((T_NS_C) reduce 200) ((T_OBJECT_CAST) reduce 200) ((T_PRINT) reduce 200) ((T_PROPERTY_C) reduce 200) ((T_READONLY) reduce 200) ((T_REQUIRE) reduce 200) ((T_REQUIRE_ONCE) reduce 200) ((T_RETURN) reduce 200) ((T_START_HEREDOC) reduce 200) ((T_STATIC) reduce 200) ((T_STRING) reduce 200) ((T_STRING_CAST) reduce 200) ((T_SWITCH) reduce 200) ((T_THROW) reduce 200) ((T_TRAIT) reduce 200) ((T_TRAIT_C) reduce 200) ((T_TRY) reduce 200) ((T_UNSET) reduce 200) ((T_UNSET_CAST) reduce 200) ((T_VARIABLE) reduce 200) ((T_VOID_CAST) reduce 200) ((T_WHILE) reduce 200) ((T_YIELD) reduce 200) ((T_YIELD_FROM) reduce 200) (("[") reduce 200) (("`") reduce 200) (("{") reduce 200) (("~") reduce 200)) 24901 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CASE) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DEFAULT) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_EMPTY) reduce 153) ((T_ENDSWITCH) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_VOID_CAST) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 24905 ((("!") reduce 224) (("\"") reduce 224) (("$") reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CASE) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DEFAULT) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_EMPTY) reduce 224) ((T_ENDSWITCH) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_VARIABLE) reduce 224) ((T_VOID_CAST) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("~") reduce 224)) 24907 ((("!") reduce 225) (("\"") reduce 225) (("$") reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CASE) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DEFAULT) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_EMPTY) reduce 225) ((T_ENDSWITCH) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_VARIABLE) reduce 225) ((T_VOID_CAST) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("~") reduce 225)) 24911 ((("!") reduce 226) (("\"") reduce 226) (("$") reduce 226) (("(") reduce 226) (("+") reduce 226) (("-") reduce 226) ((";") reduce 226) (("@") reduce 226) ((T_ABSTRACT) reduce 226) ((T_ARRAY) reduce 226) ((T_ARRAY_CAST) reduce 226) ((T_ATTRIBUTE) reduce 226) ((T_BOOL_CAST) reduce 226) ((T_BREAK) reduce 226) ((T_CASE) reduce 226) ((T_CLASS) reduce 226) ((T_CLASS_C) reduce 226) ((T_CLONE) reduce 226) ((T_CONSTANT_ENCAPSED_STRING) reduce 226) ((T_CONTINUE) reduce 226) ((T_DEC) reduce 226) ((T_DECLARE) reduce 226) ((T_DEFAULT) reduce 226) ((T_DIR) reduce 226) ((T_DNUMBER) reduce 226) ((T_DO) reduce 226) ((T_DOUBLE_CAST) reduce 226) ((T_ECHO) reduce 226) ((T_EMPTY) reduce 226) ((T_ENDSWITCH) reduce 226) ((T_ENUM) reduce 226) ((T_EVAL) reduce 226) ((T_EXIT) reduce 226) ((T_FILE) reduce 226) ((T_FINAL) reduce 226) ((T_FN) reduce 226) ((T_FOR) reduce 226) ((T_FOREACH) reduce 226) ((T_FUNCTION) reduce 226) ((T_FUNC_C) reduce 226) ((T_GLOBAL) reduce 226) ((T_GOTO) reduce 226) ((T_HALT_COMPILER) reduce 226) ((T_IF) reduce 226) ((T_INC) reduce 226) ((T_INCLUDE) reduce 226) ((T_INCLUDE_ONCE) reduce 226) ((T_INLINE_HTML) reduce 226) ((T_INTERFACE) reduce 226) ((T_INT_CAST) reduce 226) ((T_ISSET) reduce 226) ((T_LINE) reduce 226) ((T_LIST) reduce 226) ((T_LNUMBER) reduce 226) ((T_MATCH) reduce 226) ((T_METHOD_C) reduce 226) ((T_NAME_FULLY_QUALIFIED) reduce 226) ((T_NAME_QUALIFIED) reduce 226) ((T_NAME_RELATIVE) reduce 226) ((T_NEW) reduce 226) ((T_NS_C) reduce 226) ((T_OBJECT_CAST) reduce 226) ((T_PRINT) reduce 226) ((T_PROPERTY_C) reduce 226) ((T_READONLY) reduce 226) ((T_REQUIRE) reduce 226) ((T_REQUIRE_ONCE) reduce 226) ((T_RETURN) reduce 226) ((T_START_HEREDOC) reduce 226) ((T_STATIC) reduce 226) ((T_STRING) reduce 226) ((T_STRING_CAST) reduce 226) ((T_SWITCH) reduce 226) ((T_THROW) reduce 226) ((T_TRAIT) reduce 226) ((T_TRAIT_C) reduce 226) ((T_TRY) reduce 226) ((T_UNSET) reduce 226) ((T_UNSET_CAST) reduce 226) ((T_VARIABLE) reduce 226) ((T_VOID_CAST) reduce 226) ((T_WHILE) reduce 226) ((T_YIELD) reduce 226) ((T_YIELD_FROM) reduce 226) (("[") reduce 226) (("`") reduce 226) (("{") reduce 226) (("~") reduce 226)) 24914 ((("!") reduce 227) (("\"") reduce 227) (("$") reduce 227) (("(") reduce 227) (("+") reduce 227) (("-") reduce 227) ((";") reduce 227) (("@") reduce 227) ((T_ABSTRACT) reduce 227) ((T_ARRAY) reduce 227) ((T_ARRAY_CAST) reduce 227) ((T_ATTRIBUTE) reduce 227) ((T_BOOL_CAST) reduce 227) ((T_BREAK) reduce 227) ((T_CASE) reduce 227) ((T_CLASS) reduce 227) ((T_CLASS_C) reduce 227) ((T_CLONE) reduce 227) ((T_CONSTANT_ENCAPSED_STRING) reduce 227) ((T_CONTINUE) reduce 227) ((T_DEC) reduce 227) ((T_DECLARE) reduce 227) ((T_DEFAULT) reduce 227) ((T_DIR) reduce 227) ((T_DNUMBER) reduce 227) ((T_DO) reduce 227) ((T_DOUBLE_CAST) reduce 227) ((T_ECHO) reduce 227) ((T_EMPTY) reduce 227) ((T_ENDSWITCH) reduce 227) ((T_ENUM) reduce 227) ((T_EVAL) reduce 227) ((T_EXIT) reduce 227) ((T_FILE) reduce 227) ((T_FINAL) reduce 227) ((T_FN) reduce 227) ((T_FOR) reduce 227) ((T_FOREACH) reduce 227) ((T_FUNCTION) reduce 227) ((T_FUNC_C) reduce 227) ((T_GLOBAL) reduce 227) ((T_GOTO) reduce 227) ((T_HALT_COMPILER) reduce 227) ((T_IF) reduce 227) ((T_INC) reduce 227) ((T_INCLUDE) reduce 227) ((T_INCLUDE_ONCE) reduce 227) ((T_INLINE_HTML) reduce 227) ((T_INTERFACE) reduce 227) ((T_INT_CAST) reduce 227) ((T_ISSET) reduce 227) ((T_LINE) reduce 227) ((T_LIST) reduce 227) ((T_LNUMBER) reduce 227) ((T_MATCH) reduce 227) ((T_METHOD_C) reduce 227) ((T_NAME_FULLY_QUALIFIED) reduce 227) ((T_NAME_QUALIFIED) reduce 227) ((T_NAME_RELATIVE) reduce 227) ((T_NEW) reduce 227) ((T_NS_C) reduce 227) ((T_OBJECT_CAST) reduce 227) ((T_PRINT) reduce 227) ((T_PROPERTY_C) reduce 227) ((T_READONLY) reduce 227) ((T_REQUIRE) reduce 227) ((T_REQUIRE_ONCE) reduce 227) ((T_RETURN) reduce 227) ((T_START_HEREDOC) reduce 227) ((T_STATIC) reduce 227) ((T_STRING) reduce 227) ((T_STRING_CAST) reduce 227) ((T_SWITCH) reduce 227) ((T_THROW) reduce 227) ((T_TRAIT) reduce 227) ((T_TRAIT_C) reduce 227) ((T_TRY) reduce 227) ((T_UNSET) reduce 227) ((T_UNSET_CAST) reduce 227) ((T_VARIABLE) reduce 227) ((T_VOID_CAST) reduce 227) ((T_WHILE) reduce 227) ((T_YIELD) reduce 227) ((T_YIELD_FROM) reduce 227) (("[") reduce 227) (("`") reduce 227) (("{") reduce 227) (("~") reduce 227)) 24915 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CASE) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DEFAULT) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_EMPTY) reduce 169) ((T_ENDSWITCH) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_VOID_CAST) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 24917 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CASE) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DEFAULT) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_EMPTY) reduce 158) ((T_ENDSWITCH) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_VOID_CAST) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 24919 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CASE) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DEFAULT) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_EMPTY) reduce 156) ((T_ENDSWITCH) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_VOID_CAST) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 24925 ((("!") reduce 201) (("\"") reduce 201) (("$") reduce 201) (("(") reduce 201) (("+") reduce 201) (("-") reduce 201) ((";") reduce 201) (("@") reduce 201) ((T_ABSTRACT) reduce 201) ((T_ARRAY) reduce 201) ((T_ARRAY_CAST) reduce 201) ((T_ATTRIBUTE) reduce 201) ((T_BOOL_CAST) reduce 201) ((T_BREAK) reduce 201) ((T_CASE) reduce 201) ((T_CLASS) reduce 201) ((T_CLASS_C) reduce 201) ((T_CLONE) reduce 201) ((T_CONSTANT_ENCAPSED_STRING) reduce 201) ((T_CONTINUE) reduce 201) ((T_DEC) reduce 201) ((T_DECLARE) reduce 201) ((T_DEFAULT) reduce 201) ((T_DIR) reduce 201) ((T_DNUMBER) reduce 201) ((T_DO) reduce 201) ((T_DOUBLE_CAST) reduce 201) ((T_ECHO) reduce 201) ((T_EMPTY) reduce 201) ((T_ENDSWITCH) reduce 201) ((T_ENUM) reduce 201) ((T_EVAL) reduce 201) ((T_EXIT) reduce 201) ((T_FILE) reduce 201) ((T_FINAL) reduce 201) ((T_FN) reduce 201) ((T_FOR) reduce 201) ((T_FOREACH) reduce 201) ((T_FUNCTION) reduce 201) ((T_FUNC_C) reduce 201) ((T_GLOBAL) reduce 201) ((T_GOTO) reduce 201) ((T_HALT_COMPILER) reduce 201) ((T_IF) reduce 201) ((T_INC) reduce 201) ((T_INCLUDE) reduce 201) ((T_INCLUDE_ONCE) reduce 201) ((T_INLINE_HTML) reduce 201) ((T_INTERFACE) reduce 201) ((T_INT_CAST) reduce 201) ((T_ISSET) reduce 201) ((T_LINE) reduce 201) ((T_LIST) reduce 201) ((T_LNUMBER) reduce 201) ((T_MATCH) reduce 201) ((T_METHOD_C) reduce 201) ((T_NAME_FULLY_QUALIFIED) reduce 201) ((T_NAME_QUALIFIED) reduce 201) ((T_NAME_RELATIVE) reduce 201) ((T_NEW) reduce 201) ((T_NS_C) reduce 201) ((T_OBJECT_CAST) reduce 201) ((T_PRINT) reduce 201) ((T_PROPERTY_C) reduce 201) ((T_READONLY) reduce 201) ((T_REQUIRE) reduce 201) ((T_REQUIRE_ONCE) reduce 201) ((T_RETURN) reduce 201) ((T_START_HEREDOC) reduce 201) ((T_STATIC) reduce 201) ((T_STRING) reduce 201) ((T_STRING_CAST) reduce 201) ((T_SWITCH) reduce 201) ((T_THROW) reduce 201) ((T_TRAIT) reduce 201) ((T_TRAIT_C) reduce 201) ((T_TRY) reduce 201) ((T_UNSET) reduce 201) ((T_UNSET_CAST) reduce 201) ((T_VARIABLE) reduce 201) ((T_VOID_CAST) reduce 201) ((T_WHILE) reduce 201) ((T_YIELD) reduce 201) ((T_YIELD_FROM) reduce 201) (("[") reduce 201) (("`") reduce 201) (("{") reduce 201) (("~") reduce 201)) 24929 ((("!") reduce 244) (("\"") reduce 244) (("$") reduce 244) (("(") reduce 244) (("+") reduce 244) (("-") reduce 244) ((";") reduce 244) (("@") reduce 244) ((T_ABSTRACT) reduce 244) ((T_ARRAY) reduce 244) ((T_ARRAY_CAST) reduce 244) ((T_ATTRIBUTE) reduce 244) ((T_BOOL_CAST) reduce 244) ((T_BREAK) reduce 244) ((T_CASE) reduce 244) ((T_CLASS) reduce 244) ((T_CLASS_C) reduce 244) ((T_CLONE) reduce 244) ((T_CONSTANT_ENCAPSED_STRING) reduce 244) ((T_CONTINUE) reduce 244) ((T_DEC) reduce 244) ((T_DECLARE) reduce 244) ((T_DEFAULT) reduce 244) ((T_DIR) reduce 244) ((T_DNUMBER) reduce 244) ((T_DO) reduce 244) ((T_DOUBLE_CAST) reduce 244) ((T_ECHO) reduce 244) ((T_ELSE) reduce 244) ((T_ELSEIF) reduce 244) ((T_EMPTY) reduce 244) ((T_ENDSWITCH) reduce 244) ((T_ENUM) reduce 244) ((T_EVAL) reduce 244) ((T_EXIT) reduce 244) ((T_FILE) reduce 244) ((T_FINAL) reduce 244) ((T_FN) reduce 244) ((T_FOR) reduce 244) ((T_FOREACH) reduce 244) ((T_FUNCTION) reduce 244) ((T_FUNC_C) reduce 244) ((T_GLOBAL) reduce 244) ((T_GOTO) reduce 244) ((T_HALT_COMPILER) reduce 244) ((T_IF) reduce 244) ((T_INC) reduce 244) ((T_INCLUDE) reduce 244) ((T_INCLUDE_ONCE) reduce 244) ((T_INLINE_HTML) reduce 244) ((T_INTERFACE) reduce 244) ((T_INT_CAST) reduce 244) ((T_ISSET) reduce 244) ((T_LINE) reduce 244) ((T_LIST) reduce 244) ((T_LNUMBER) reduce 244) ((T_MATCH) reduce 244) ((T_METHOD_C) reduce 244) ((T_NAME_FULLY_QUALIFIED) reduce 244) ((T_NAME_QUALIFIED) reduce 244) ((T_NAME_RELATIVE) reduce 244) ((T_NEW) reduce 244) ((T_NS_C) reduce 244) ((T_OBJECT_CAST) reduce 244) ((T_PRINT) reduce 244) ((T_PROPERTY_C) reduce 244) ((T_READONLY) reduce 244) ((T_REQUIRE) reduce 244) ((T_REQUIRE_ONCE) reduce 244) ((T_RETURN) reduce 244) ((T_START_HEREDOC) reduce 244) ((T_STATIC) reduce 244) ((T_STRING) reduce 244) ((T_STRING_CAST) reduce 244) ((T_SWITCH) reduce 244) ((T_THROW) reduce 244) ((T_TRAIT) reduce 244) ((T_TRAIT_C) reduce 244) ((T_TRY) reduce 244) ((T_UNSET) reduce 244) ((T_UNSET_CAST) reduce 244) ((T_VARIABLE) reduce 244) ((T_VOID_CAST) reduce 244) ((T_WHILE) reduce 244) ((T_YIELD) reduce 244) ((T_YIELD_FROM) reduce 244) (("[") reduce 244) (("`") reduce 244) (("{") reduce 244) (("~") reduce 244)) 24932 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CASE) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DEFAULT) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_EMPTY) reduce 146) ((T_ENDSWITCH) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_PROPERTY_C) reduce 146) ((T_READONLY) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_VOID_CAST) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("~") reduce 146)) 24934 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CASE) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DEFAULT) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_EMPTY) reduce 168) ((T_ENDSWITCH) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_VOID_CAST) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 24936 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CASE) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DEFAULT) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_EMPTY) reduce 157) ((T_ENDSWITCH) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_VOID_CAST) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 24946 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CASE) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DEFAULT) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_EMPTY) reduce 164) ((T_ENDSWITCH) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_VOID_CAST) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 24947 ((("!") reduce 220) (("\"") reduce 220) (("$") reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CASE) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DEFAULT) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_EMPTY) reduce 220) ((T_ENDSWITCH) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_VARIABLE) reduce 220) ((T_VOID_CAST) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("~") reduce 220)) 24950 ((("!") reduce 221) (("\"") reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CASE) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DEFAULT) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_EMPTY) reduce 221) ((T_ENDSWITCH) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_VARIABLE) reduce 221) ((T_VOID_CAST) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("~") reduce 221)) 24951 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CASE) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DEFAULT) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_EMPTY) reduce 163) ((T_ENDSWITCH) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_VOID_CAST) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 24960 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CASE) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DEFAULT) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_EMPTY) reduce 152) ((T_ENDSWITCH) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_VOID_CAST) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 24961 ((("!") reduce 218) (("\"") reduce 218) (("$") reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CASE) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DEFAULT) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_EMPTY) reduce 218) ((T_ENDSWITCH) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_VARIABLE) reduce 218) ((T_VOID_CAST) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("~") reduce 218)) 24964 ((("!") reduce 219) (("\"") reduce 219) (("$") reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CASE) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DEFAULT) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_EMPTY) reduce 219) ((T_ENDSWITCH) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_VARIABLE) reduce 219) ((T_VOID_CAST) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("~") reduce 219)) 24971 ((("!") reduce 202) (("\"") reduce 202) (("$") reduce 202) (("(") reduce 202) (("+") reduce 202) (("-") reduce 202) ((";") reduce 202) (("@") reduce 202) ((T_ABSTRACT) reduce 202) ((T_ARRAY) reduce 202) ((T_ARRAY_CAST) reduce 202) ((T_ATTRIBUTE) reduce 202) ((T_BOOL_CAST) reduce 202) ((T_BREAK) reduce 202) ((T_CASE) reduce 202) ((T_CLASS) reduce 202) ((T_CLASS_C) reduce 202) ((T_CLONE) reduce 202) ((T_CONSTANT_ENCAPSED_STRING) reduce 202) ((T_CONTINUE) reduce 202) ((T_DEC) reduce 202) ((T_DECLARE) reduce 202) ((T_DEFAULT) reduce 202) ((T_DIR) reduce 202) ((T_DNUMBER) reduce 202) ((T_DO) reduce 202) ((T_DOUBLE_CAST) reduce 202) ((T_ECHO) reduce 202) ((T_EMPTY) reduce 202) ((T_ENDSWITCH) reduce 202) ((T_ENUM) reduce 202) ((T_EVAL) reduce 202) ((T_EXIT) reduce 202) ((T_FILE) reduce 202) ((T_FINAL) reduce 202) ((T_FN) reduce 202) ((T_FOR) reduce 202) ((T_FOREACH) reduce 202) ((T_FUNCTION) reduce 202) ((T_FUNC_C) reduce 202) ((T_GLOBAL) reduce 202) ((T_GOTO) reduce 202) ((T_HALT_COMPILER) reduce 202) ((T_IF) reduce 202) ((T_INC) reduce 202) ((T_INCLUDE) reduce 202) ((T_INCLUDE_ONCE) reduce 202) ((T_INLINE_HTML) reduce 202) ((T_INTERFACE) reduce 202) ((T_INT_CAST) reduce 202) ((T_ISSET) reduce 202) ((T_LINE) reduce 202) ((T_LIST) reduce 202) ((T_LNUMBER) reduce 202) ((T_MATCH) reduce 202) ((T_METHOD_C) reduce 202) ((T_NAME_FULLY_QUALIFIED) reduce 202) ((T_NAME_QUALIFIED) reduce 202) ((T_NAME_RELATIVE) reduce 202) ((T_NEW) reduce 202) ((T_NS_C) reduce 202) ((T_OBJECT_CAST) reduce 202) ((T_PRINT) reduce 202) ((T_PROPERTY_C) reduce 202) ((T_READONLY) reduce 202) ((T_REQUIRE) reduce 202) ((T_REQUIRE_ONCE) reduce 202) ((T_RETURN) reduce 202) ((T_START_HEREDOC) reduce 202) ((T_STATIC) reduce 202) ((T_STRING) reduce 202) ((T_STRING_CAST) reduce 202) ((T_SWITCH) reduce 202) ((T_THROW) reduce 202) ((T_TRAIT) reduce 202) ((T_TRAIT_C) reduce 202) ((T_TRY) reduce 202) ((T_UNSET) reduce 202) ((T_UNSET_CAST) reduce 202) ((T_VARIABLE) reduce 202) ((T_VOID_CAST) reduce 202) ((T_WHILE) reduce 202) ((T_YIELD) reduce 202) ((T_YIELD_FROM) reduce 202) (("[") reduce 202) (("`") reduce 202) (("{") reduce 202) (("~") reduce 202)) 24973 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CASE) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DEFAULT) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_EMPTY) reduce 159) ((T_ENDSWITCH) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_VOID_CAST) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 24979 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CASE) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DEFAULT) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_EMPTY) reduce 151) ((T_ENDSWITCH) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_VOID_CAST) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 24984 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CASE) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DEFAULT) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_EMPTY) reduce 165) ((T_ENDSWITCH) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_VOID_CAST) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 24985 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CASE) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DEFAULT) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_EMPTY) reduce 222) ((T_ENDSWITCH) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_VARIABLE) reduce 222) ((T_VOID_CAST) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("~") reduce 222)) 24988 ((("!") reduce 223) (("\"") reduce 223) (("$") reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CASE) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DEFAULT) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_EMPTY) reduce 223) ((T_ENDSWITCH) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_VARIABLE) reduce 223) ((T_VOID_CAST) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("~") reduce 223)) 24990 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CASE) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DEFAULT) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_EMPTY) reduce 155) ((T_ENDSWITCH) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_VOID_CAST) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 24997 ((("!") reduce 190) (("\"") reduce 190) (("$") reduce 190) (("(") reduce 190) (("+") reduce 190) (("-") reduce 190) ((";") reduce 190) (("@") reduce 190) ((T_ABSTRACT) reduce 190) ((T_ARRAY) reduce 190) ((T_ARRAY_CAST) reduce 190) ((T_ATTRIBUTE) reduce 190) ((T_BOOL_CAST) reduce 190) ((T_BREAK) reduce 190) ((T_CASE) reduce 190) ((T_CLASS) reduce 190) ((T_CLASS_C) reduce 190) ((T_CLONE) reduce 190) ((T_CONSTANT_ENCAPSED_STRING) reduce 190) ((T_CONTINUE) reduce 190) ((T_DEC) reduce 190) ((T_DECLARE) reduce 190) ((T_DEFAULT) reduce 190) ((T_DIR) reduce 190) ((T_DNUMBER) reduce 190) ((T_DO) reduce 190) ((T_DOUBLE_CAST) reduce 190) ((T_ECHO) reduce 190) ((T_EMPTY) reduce 190) ((T_ENDSWITCH) reduce 190) ((T_ENUM) reduce 190) ((T_EVAL) reduce 190) ((T_EXIT) reduce 190) ((T_FILE) reduce 190) ((T_FINAL) reduce 190) ((T_FN) reduce 190) ((T_FOR) reduce 190) ((T_FOREACH) reduce 190) ((T_FUNCTION) reduce 190) ((T_FUNC_C) reduce 190) ((T_GLOBAL) reduce 190) ((T_GOTO) reduce 190) ((T_HALT_COMPILER) reduce 190) ((T_IF) reduce 190) ((T_INC) reduce 190) ((T_INCLUDE) reduce 190) ((T_INCLUDE_ONCE) reduce 190) ((T_INLINE_HTML) reduce 190) ((T_INTERFACE) reduce 190) ((T_INT_CAST) reduce 190) ((T_ISSET) reduce 190) ((T_LINE) reduce 190) ((T_LIST) reduce 190) ((T_LNUMBER) reduce 190) ((T_MATCH) reduce 190) ((T_METHOD_C) reduce 190) ((T_NAME_FULLY_QUALIFIED) reduce 190) ((T_NAME_QUALIFIED) reduce 190) ((T_NAME_RELATIVE) reduce 190) ((T_NEW) reduce 190) ((T_NS_C) reduce 190) ((T_OBJECT_CAST) reduce 190) ((T_PRINT) reduce 190) ((T_PROPERTY_C) reduce 190) ((T_READONLY) reduce 190) ((T_REQUIRE) reduce 190) ((T_REQUIRE_ONCE) reduce 190) ((T_RETURN) reduce 190) ((T_START_HEREDOC) reduce 190) ((T_STATIC) reduce 190) ((T_STRING) reduce 190) ((T_STRING_CAST) reduce 190) ((T_SWITCH) reduce 190) ((T_THROW) reduce 190) ((T_TRAIT) reduce 190) ((T_TRAIT_C) reduce 190) ((T_TRY) reduce 190) ((T_UNSET) reduce 190) ((T_UNSET_CAST) reduce 190) ((T_VARIABLE) reduce 190) ((T_VOID_CAST) reduce 190) ((T_WHILE) reduce 190) ((T_YIELD) reduce 190) ((T_YIELD_FROM) reduce 190) (("[") reduce 190) (("`") reduce 190) (("{") reduce 190) (("~") reduce 190)) 24999 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CASE) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DEFAULT) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_EMPTY) reduce 154) ((T_ENDSWITCH) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_VOID_CAST) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 25000 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CASE) reduce 231) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DEFAULT) reduce 231) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_ENDSWITCH) reduce 231) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_VOID_CAST) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("~") shift)) 25011 ((("%") reduce 540) (("*") reduce 540) (("+") reduce 540) (("-") reduce 540) ((".") reduce 540) (("/") reduce 540) ((":") reduce 540) ((";") reduce 540) (("<") reduce 540) ((">") reduce 540) (("?") reduce 540) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_BOOLEAN_AND) reduce 540) ((T_BOOLEAN_OR) reduce 540) ((T_COALESCE) reduce 540) ((T_INSTANCEOF) reduce 540) ((T_IS_EQUAL) reduce 540) ((T_IS_GREATER_OR_EQUAL) reduce 540) ((T_IS_IDENTICAL) reduce 540) ((T_IS_NOT_EQUAL) reduce 540) ((T_IS_NOT_IDENTICAL) reduce 540) ((T_IS_SMALLER_OR_EQUAL) reduce 540) ((T_LOGICAL_AND) reduce 540) ((T_LOGICAL_OR) reduce 540) ((T_LOGICAL_XOR) reduce 540) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 540) ((T_OBJECT_OPERATOR) reduce 540) ((T_PIPE) reduce 540) ((T_POW) reduce 540) ((T_SL) reduce 540) ((T_SPACESHIP) reduce 540) ((T_SR) reduce 540) (("[") reduce 540) (("^") reduce 540) (("|") reduce 540)) 25013 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) ((":") reduce 521) ((";") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_PIPE) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("|") reduce 521)) 25015 ((("%") reduce 534) (("*") reduce 534) (("+") reduce 534) (("-") reduce 534) ((".") reduce 534) (("/") reduce 534) ((":") reduce 534) ((";") reduce 534) (("<") reduce 534) ((">") reduce 534) (("?") reduce 534) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_BOOLEAN_AND) reduce 534) ((T_BOOLEAN_OR) reduce 534) ((T_COALESCE) reduce 534) ((T_INSTANCEOF) reduce 534) ((T_IS_EQUAL) reduce 534) ((T_IS_GREATER_OR_EQUAL) reduce 534) ((T_IS_IDENTICAL) reduce 534) ((T_IS_NOT_EQUAL) reduce 534) ((T_IS_NOT_IDENTICAL) reduce 534) ((T_IS_SMALLER_OR_EQUAL) reduce 534) ((T_LOGICAL_AND) reduce 534) ((T_LOGICAL_OR) reduce 534) ((T_LOGICAL_XOR) reduce 534) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PIPE) reduce 534) ((T_POW) reduce 534) ((T_SL) reduce 534) ((T_SPACESHIP) reduce 534) ((T_SR) reduce 534) (("[") reduce 534) (("^") reduce 534) (("|") reduce 534)) 25016 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) ((":") reduce 524) ((";") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_PIPE) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("^") reduce 524) (("|") reduce 524)) 25020 ((("%") reduce 517) (("(") shift) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((":") reduce 517) ((";") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_PIPE) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("^") reduce 517) (("|") reduce 517)) 25021 ((("%") reduce 533) (("*") reduce 533) (("+") reduce 533) (("-") reduce 533) ((".") reduce 533) (("/") reduce 533) ((":") reduce 533) ((";") reduce 533) (("<") reduce 533) ((">") reduce 533) (("?") reduce 533) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_BOOLEAN_AND) reduce 533) ((T_BOOLEAN_OR) reduce 533) ((T_COALESCE) reduce 533) ((T_INSTANCEOF) reduce 533) ((T_IS_EQUAL) reduce 533) ((T_IS_GREATER_OR_EQUAL) reduce 533) ((T_IS_IDENTICAL) reduce 533) ((T_IS_NOT_EQUAL) reduce 533) ((T_IS_NOT_IDENTICAL) reduce 533) ((T_IS_SMALLER_OR_EQUAL) reduce 533) ((T_LOGICAL_AND) reduce 533) ((T_LOGICAL_OR) reduce 533) ((T_LOGICAL_XOR) reduce 533) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PIPE) reduce 533) ((T_POW) reduce 533) ((T_SL) reduce 533) ((T_SPACESHIP) reduce 533) ((T_SR) reduce 533) (("[") reduce 533) (("^") reduce 533) (("|") reduce 533)) 25022 ((("%") reduce 537) (("*") reduce 537) (("+") reduce 537) (("-") reduce 537) ((".") reduce 537) (("/") reduce 537) ((":") reduce 537) ((";") reduce 537) (("<") reduce 537) ((">") reduce 537) (("?") reduce 537) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_BOOLEAN_AND) reduce 537) ((T_BOOLEAN_OR) reduce 537) ((T_COALESCE) reduce 537) ((T_INSTANCEOF) reduce 537) ((T_IS_EQUAL) reduce 537) ((T_IS_GREATER_OR_EQUAL) reduce 537) ((T_IS_IDENTICAL) reduce 537) ((T_IS_NOT_EQUAL) reduce 537) ((T_IS_NOT_IDENTICAL) reduce 537) ((T_IS_SMALLER_OR_EQUAL) reduce 537) ((T_LOGICAL_AND) reduce 537) ((T_LOGICAL_OR) reduce 537) ((T_LOGICAL_XOR) reduce 537) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_PIPE) reduce 537) ((T_POW) reduce 537) ((T_SL) reduce 537) ((T_SPACESHIP) reduce 537) ((T_SR) reduce 537) (("[") reduce 537) (("^") reduce 537) (("|") reduce 537)) 25028 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) ((":") reduce 532) ((";") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PIPE) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("|") reduce 532)) 25030 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) ((":") reduce 523) ((";") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_PIPE) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("^") reduce 523) (("|") reduce 523)) 25032 ((("%") reduce 536) (("*") reduce 536) (("+") reduce 536) (("-") reduce 536) ((".") reduce 536) (("/") reduce 536) ((":") reduce 536) ((";") reduce 536) (("<") reduce 536) ((">") reduce 536) (("?") reduce 536) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_BOOLEAN_AND) reduce 536) ((T_BOOLEAN_OR) reduce 536) ((T_COALESCE) reduce 536) ((T_INSTANCEOF) reduce 536) ((T_IS_EQUAL) reduce 536) ((T_IS_GREATER_OR_EQUAL) reduce 536) ((T_IS_IDENTICAL) reduce 536) ((T_IS_NOT_EQUAL) reduce 536) ((T_IS_NOT_IDENTICAL) reduce 536) ((T_IS_SMALLER_OR_EQUAL) reduce 536) ((T_LOGICAL_AND) reduce 536) ((T_LOGICAL_OR) reduce 536) ((T_LOGICAL_XOR) reduce 536) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 536) ((T_OBJECT_OPERATOR) reduce 536) ((T_PIPE) reduce 536) ((T_POW) reduce 536) ((T_SL) reduce 536) ((T_SPACESHIP) reduce 536) ((T_SR) reduce 536) (("[") reduce 536) (("^") reduce 536) (("|") reduce 536)) 25033 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((":") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("|") reduce 94)) 25034 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((":") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("|") reduce 93)) 25035 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((":") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("|") reduce 95)) 25037 ((("%") reduce 539) (("*") reduce 539) (("+") reduce 539) (("-") reduce 539) ((".") reduce 539) (("/") reduce 539) ((":") reduce 539) ((";") reduce 539) (("<") reduce 539) ((">") reduce 539) (("?") reduce 539) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_BOOLEAN_AND) reduce 539) ((T_BOOLEAN_OR) reduce 539) ((T_COALESCE) reduce 539) ((T_INSTANCEOF) reduce 539) ((T_IS_EQUAL) reduce 539) ((T_IS_GREATER_OR_EQUAL) reduce 539) ((T_IS_IDENTICAL) reduce 539) ((T_IS_NOT_EQUAL) reduce 539) ((T_IS_NOT_IDENTICAL) reduce 539) ((T_IS_SMALLER_OR_EQUAL) reduce 539) ((T_LOGICAL_AND) reduce 539) ((T_LOGICAL_OR) reduce 539) ((T_LOGICAL_XOR) reduce 539) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 539) ((T_OBJECT_OPERATOR) reduce 539) ((T_PIPE) reduce 539) ((T_POW) reduce 539) ((T_SL) reduce 539) ((T_SPACESHIP) reduce 539) ((T_SR) reduce 539) (("[") reduce 539) (("^") reduce 539) (("|") reduce 539)) 25040 ((("%") reduce 538) (("*") reduce 538) (("+") reduce 538) (("-") reduce 538) ((".") reduce 538) (("/") reduce 538) ((":") reduce 538) ((";") reduce 538) (("<") reduce 538) ((">") reduce 538) (("?") reduce 538) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_BOOLEAN_AND) reduce 538) ((T_BOOLEAN_OR) reduce 538) ((T_COALESCE) reduce 538) ((T_INSTANCEOF) reduce 538) ((T_IS_EQUAL) reduce 538) ((T_IS_GREATER_OR_EQUAL) reduce 538) ((T_IS_IDENTICAL) reduce 538) ((T_IS_NOT_EQUAL) reduce 538) ((T_IS_NOT_IDENTICAL) reduce 538) ((T_IS_SMALLER_OR_EQUAL) reduce 538) ((T_LOGICAL_AND) reduce 538) ((T_LOGICAL_OR) reduce 538) ((T_LOGICAL_XOR) reduce 538) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 538) ((T_OBJECT_OPERATOR) reduce 538) ((T_PIPE) reduce 538) ((T_POW) reduce 538) ((T_SL) reduce 538) ((T_SPACESHIP) reduce 538) ((T_SR) reduce 538) (("[") reduce 538) (("^") reduce 538) (("|") reduce 538)) 25046 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((":") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92)) 25049 ((("%") reduce 535) (("*") reduce 535) (("+") reduce 535) (("-") reduce 535) ((".") reduce 535) (("/") reduce 535) ((":") reduce 535) ((";") reduce 535) (("<") reduce 535) ((">") reduce 535) (("?") reduce 535) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_BOOLEAN_AND) reduce 535) ((T_BOOLEAN_OR) reduce 535) ((T_COALESCE) reduce 535) ((T_INSTANCEOF) reduce 535) ((T_IS_EQUAL) reduce 535) ((T_IS_GREATER_OR_EQUAL) reduce 535) ((T_IS_IDENTICAL) reduce 535) ((T_IS_NOT_EQUAL) reduce 535) ((T_IS_NOT_IDENTICAL) reduce 535) ((T_IS_SMALLER_OR_EQUAL) reduce 535) ((T_LOGICAL_AND) reduce 535) ((T_LOGICAL_OR) reduce 535) ((T_LOGICAL_XOR) reduce 535) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PIPE) reduce 535) ((T_POW) reduce 535) ((T_SL) reduce 535) ((T_SPACESHIP) reduce 535) ((T_SR) reduce 535) (("[") reduce 535) (("^") reduce 535) (("|") reduce 535)) 25051 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) ((":") reduce 568) ((";") reduce 568) (("<") reduce 568) (("=") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AND_EQUAL) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_COALESCE_EQUAL) reduce 568) ((T_CONCAT_EQUAL) reduce 568) ((T_DEC) reduce 568) ((T_DIV_EQUAL) reduce 568) ((T_INC) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_MINUS_EQUAL) reduce 568) ((T_MOD_EQUAL) reduce 568) ((T_MUL_EQUAL) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_OR_EQUAL) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_PLUS_EQUAL) reduce 568) ((T_POW) reduce 568) ((T_POW_EQUAL) reduce 568) ((T_SL) reduce 568) ((T_SL_EQUAL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) ((T_SR_EQUAL) reduce 568) ((T_XOR_EQUAL) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 25052 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 479) (("(") shift) (("*") reduce 479) (("+") shift) (("-") shift) ((".") reduce 479) (("/") reduce 479) ((":") reduce 479) ((";") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 479) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 479) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PIPE) reduce 479) ((T_POW) reduce 479) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 479) (("`") shift) (("|") reduce 479) (("~") shift)) 25059 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) ((":") reduce 564) ((";") reduce 564) (("<") reduce 564) (("=") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AND_EQUAL) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_COALESCE_EQUAL) reduce 564) ((T_CONCAT_EQUAL) reduce 564) ((T_DEC) reduce 564) ((T_DIV_EQUAL) reduce 564) ((T_INC) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_MINUS_EQUAL) reduce 564) ((T_MOD_EQUAL) reduce 564) ((T_MUL_EQUAL) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_OR_EQUAL) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_PLUS_EQUAL) reduce 564) ((T_POW) reduce 564) ((T_POW_EQUAL) reduce 564) ((T_SL) reduce 564) ((T_SL_EQUAL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) ((T_SR_EQUAL) reduce 564) ((T_XOR_EQUAL) reduce 564) (("[") reduce 564) (("^") reduce 564) (("|") reduce 564)) 25060 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) ((":") reduce 530) ((";") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PIPE) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 551) (("^") reduce 530) (("|") reduce 530)) 25062 ((("%") reduce 529) (("*") reduce 529) (("+") reduce 529) (("-") reduce 529) ((".") reduce 529) (("/") reduce 529) ((":") reduce 529) ((";") reduce 529) (("<") reduce 529) ((">") reduce 529) (("?") reduce 529) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_BOOLEAN_AND) reduce 529) ((T_BOOLEAN_OR) reduce 529) ((T_COALESCE) reduce 529) ((T_INSTANCEOF) reduce 529) ((T_IS_EQUAL) reduce 529) ((T_IS_GREATER_OR_EQUAL) reduce 529) ((T_IS_IDENTICAL) reduce 529) ((T_IS_NOT_EQUAL) reduce 529) ((T_IS_NOT_IDENTICAL) reduce 529) ((T_IS_SMALLER_OR_EQUAL) reduce 529) ((T_LOGICAL_AND) reduce 529) ((T_LOGICAL_OR) reduce 529) ((T_LOGICAL_XOR) reduce 529) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PIPE) reduce 529) ((T_POW) reduce 529) ((T_SL) reduce 529) ((T_SPACESHIP) reduce 529) ((T_SR) reduce 529) (("[") reduce 554) (("^") reduce 529) (("|") reduce 529)) 25063 ((("%") reduce 528) (("(") reduce 557) (("*") reduce 528) (("+") reduce 528) (("-") reduce 528) ((".") reduce 528) (("/") reduce 528) ((":") reduce 528) ((";") reduce 528) (("<") reduce 528) ((">") reduce 528) (("?") reduce 528) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_BOOLEAN_AND) reduce 528) ((T_BOOLEAN_OR) reduce 528) ((T_COALESCE) reduce 528) ((T_INSTANCEOF) reduce 528) ((T_IS_EQUAL) reduce 528) ((T_IS_GREATER_OR_EQUAL) reduce 528) ((T_IS_IDENTICAL) reduce 528) ((T_IS_NOT_EQUAL) reduce 528) ((T_IS_NOT_IDENTICAL) reduce 528) ((T_IS_SMALLER_OR_EQUAL) reduce 528) ((T_LOGICAL_AND) reduce 528) ((T_LOGICAL_OR) reduce 528) ((T_LOGICAL_XOR) reduce 528) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PIPE) reduce 528) ((T_POW) reduce 528) ((T_SL) reduce 528) ((T_SPACESHIP) reduce 528) ((T_SR) reduce 528) (("[") reduce 550) (("^") reduce 528) (("|") reduce 528)) 25064 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") shift) ((";") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25067 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) ((":") reduce 563) ((";") reduce 563) (("<") reduce 563) (("=") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AND_EQUAL) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_COALESCE_EQUAL) reduce 563) ((T_CONCAT_EQUAL) reduce 563) ((T_DEC) reduce 563) ((T_DIV_EQUAL) reduce 563) ((T_INC) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_MINUS_EQUAL) reduce 563) ((T_MOD_EQUAL) reduce 563) ((T_MUL_EQUAL) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_OR_EQUAL) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_PLUS_EQUAL) reduce 563) ((T_POW) reduce 563) ((T_POW_EQUAL) reduce 563) ((T_SL) reduce 563) ((T_SL_EQUAL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) ((T_SR_EQUAL) reduce 563) ((T_XOR_EQUAL) reduce 563) (("[") reduce 563) (("^") reduce 563) (("|") reduce 563)) 25068 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) ((":") reduce 484) ((";") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_PIPE) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("^") reduce 484) (("|") reduce 484)) 25069 ((("%") reduce 466) (("*") reduce 466) (("+") reduce 466) (("-") reduce 466) ((".") reduce 466) (("/") reduce 466) ((":") reduce 466) ((";") reduce 466) (("<") reduce 466) ((">") reduce 466) (("?") reduce 466) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_BOOLEAN_AND) reduce 466) ((T_BOOLEAN_OR) reduce 466) ((T_COALESCE) reduce 466) ((T_INSTANCEOF) reduce 466) ((T_IS_EQUAL) reduce 466) ((T_IS_GREATER_OR_EQUAL) reduce 466) ((T_IS_IDENTICAL) reduce 466) ((T_IS_NOT_EQUAL) reduce 466) ((T_IS_NOT_IDENTICAL) reduce 466) ((T_IS_SMALLER_OR_EQUAL) reduce 466) ((T_LOGICAL_AND) reduce 466) ((T_LOGICAL_OR) reduce 466) ((T_LOGICAL_XOR) reduce 466) ((T_PIPE) reduce 466) ((T_POW) reduce 466) ((T_SL) reduce 466) ((T_SPACESHIP) reduce 466) ((T_SR) reduce 466) (("^") reduce 466) (("|") reduce 466)) 25070 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) ((":") reduce 488) ((";") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_PIPE) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("^") reduce 488) (("|") reduce 488)) 25071 ((("%") reduce 531) (("(") shift) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) ((":") reduce 531) ((";") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("|") reduce 531)) 25072 ((("%") reduce 461) (("(") reduce 558) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) ((":") reduce 461) ((";") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PIPE) reduce 461) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") reduce 552) (("^") reduce 461) (("|") reduce 461)) 25073 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) ((":") reduce 462) ((";") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_PIPE) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 25074 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) ((":") reduce 476) ((";") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_PIPE) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("|") reduce 476)) 25075 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) ((":") reduce 559) ((";") reduce 559) (("<") reduce 559) (("=") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AND_EQUAL) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_COALESCE_EQUAL) reduce 559) ((T_CONCAT_EQUAL) reduce 559) ((T_DEC) reduce 559) ((T_DIV_EQUAL) reduce 559) ((T_INC) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_MINUS_EQUAL) reduce 559) ((T_MOD_EQUAL) reduce 559) ((T_MUL_EQUAL) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_OR_EQUAL) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_PLUS_EQUAL) reduce 559) ((T_POW) reduce 559) ((T_POW_EQUAL) reduce 559) ((T_SL) reduce 559) ((T_SL_EQUAL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) ((T_SR_EQUAL) reduce 559) ((T_XOR_EQUAL) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 25076 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((":") reduce 565) ((";") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 25077 ((("%") reduce 403) (("*") reduce 403) (("+") reduce 403) (("-") reduce 403) ((".") reduce 403) (("/") reduce 403) ((":") reduce 403) ((";") reduce 403) (("<") reduce 403) (("=") shift) ((">") reduce 403) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_INC) shift) ((T_INSTANCEOF) reduce 403) ((T_IS_EQUAL) reduce 403) ((T_IS_GREATER_OR_EQUAL) reduce 403) ((T_IS_IDENTICAL) reduce 403) ((T_IS_NOT_EQUAL) reduce 403) ((T_IS_NOT_IDENTICAL) reduce 403) ((T_IS_SMALLER_OR_EQUAL) reduce 403) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 403) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 403) ((T_POW_EQUAL) shift) ((T_SL) reduce 403) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 403) ((T_SR) reduce 403) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 548) (("^") reduce 403) (("|") reduce 403)) 25080 ((("%") reduce 448) (("*") reduce 448) (("+") reduce 448) (("-") reduce 448) ((".") reduce 448) (("/") reduce 448) ((":") reduce 448) ((";") reduce 448) (("<") reduce 448) ((">") reduce 448) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_BOOLEAN_AND) reduce 448) ((T_BOOLEAN_OR) reduce 448) ((T_COALESCE) reduce 448) ((T_INSTANCEOF) reduce 448) ((T_IS_EQUAL) reduce 448) ((T_IS_GREATER_OR_EQUAL) reduce 448) ((T_IS_IDENTICAL) reduce 448) ((T_IS_NOT_EQUAL) reduce 448) ((T_IS_NOT_IDENTICAL) reduce 448) ((T_IS_SMALLER_OR_EQUAL) reduce 448) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_PIPE) reduce 448) ((T_POW) shift) ((T_SL) reduce 448) ((T_SPACESHIP) reduce 448) ((T_SR) reduce 448) (("^") reduce 448) (("|") reduce 448)) 25112 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 432) ((";") reduce 432) (("<") shift) ((">") shift) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 432)) 25113 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 435) ((";") reduce 435) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("|") reduce 435)) 25114 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 444) (("/") shift) ((":") reduce 444) ((";") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_PIPE) reduce 444) ((T_POW) shift) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("|") reduce 444)) 25115 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 458) ((";") reduce 458) (("<") shift) ((">") shift) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 458) (("|") reduce 458)) 25116 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 443) (("/") shift) ((":") reduce 443) ((";") reduce 443) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_PIPE) reduce 443) ((T_POW) shift) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("^") reduce 443) (("|") reduce 443)) 25117 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) ((":") reduce 440) ((";") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_PIPE) reduce 440) ((T_POW) shift) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("|") reduce 440)) 25118 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 453) ((";") reduce 453) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_PIPE) reduce 453) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 453) ((T_SR) shift) (("^") reduce 453) (("|") reduce 453)) 25119 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 431) ((";") reduce 431) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25120 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 429) ((";") reduce 429) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25121 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 430) ((";") reduce 430) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25122 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 455) ((";") reduce 455) (("<") shift) ((">") shift) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 455) ((T_SR) shift) (("^") reduce 455) (("|") reduce 455)) 25123 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 450) ((";") reduce 450) (("<") shift) ((">") shift) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 450) (("|") reduce 450)) 25124 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 452) ((";") reduce 452) (("<") shift) ((">") shift) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 452) (("|") reduce 452)) 25125 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 449) ((";") reduce 449) (("<") shift) ((">") shift) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 449) (("|") reduce 449)) 25126 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 457) ((";") reduce 457) (("<") shift) ((">") shift) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 457) ((T_SR) shift) (("^") reduce 457) (("|") reduce 457)) 25127 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 451) ((";") reduce 451) (("<") shift) ((">") shift) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 451) (("|") reduce 451)) 25130 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((":") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 25131 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((":") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 25132 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((":") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 25133 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((":") reduce 509) ((";") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 25134 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((":") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 25135 ((("%") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) ((":") reduce 568) ((";") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 25136 ((("%") reduce 511) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((":") reduce 511) ((";") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("|") reduce 511)) 25137 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) ((":") reduce 459) ((";") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PIPE) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 25138 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((":") reduce 510) ((";") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 25139 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((":") reduce 512) ((";") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("|") reduce 512)) 25140 ((("%") reduce 573) (("*") reduce 573) (("+") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) ((":") reduce 573) ((";") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("|") reduce 573)) 25146 ((("%") reduce 574) (("*") reduce 574) (("+") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) ((":") reduce 574) ((";") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("|") reduce 574)) 25147 ((("%") reduce 578) (("*") reduce 578) (("+") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) ((":") reduce 578) ((";") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("|") reduce 578)) 25148 ((("%") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) ((":") reduce 582) ((";") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 25149 ((("%") reduce 575) (("*") reduce 575) (("+") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) ((":") reduce 575) ((";") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("|") reduce 575)) 25150 ((("%") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) ((":") reduce 584) ((";") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 25153 ((("%") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) ((":") reduce 583) ((";") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 25154 ((("%") reduce 576) (("*") reduce 576) (("+") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) ((":") reduce 576) ((";") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("|") reduce 576)) 25156 ((("%") reduce 577) (("*") reduce 577) (("+") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) ((":") reduce 577) ((";") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("|") reduce 577)) 25158 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((":") reduce 513) ((";") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("|") reduce 513)) 25159 ((("%") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) ((":") reduce 570) ((";") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 25162 ((("%") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) ((":") reduce 569) ((";") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 25163 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 465) ((";") reduce 465) (("<") shift) ((">") shift) (("?") reduce 465) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25164 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 427) ((";") reduce 427) (("<") shift) ((">") shift) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25165 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 428) ((";") reduce 428) (("<") shift) ((">") shift) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25166 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 433) ((";") reduce 433) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("|") reduce 433)) 25167 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 434) ((";") reduce 434) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("|") reduce 434)) 25171 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 463) ((";") reduce 463) (("<") shift) ((">") shift) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25172 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 464) ((";") reduce 464) (("<") shift) ((">") shift) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25173 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 456) ((";") reduce 456) (("<") shift) ((">") shift) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 456) ((T_SR) shift) (("^") reduce 456) (("|") reduce 456)) 25174 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 454) ((";") reduce 454) (("<") shift) ((">") shift) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 454) ((T_SR) shift) (("^") reduce 454) (("|") reduce 454)) 25175 ((("%") reduce 441) (("*") reduce 441) (("+") reduce 441) (("-") reduce 441) ((".") reduce 441) (("/") reduce 441) ((":") reduce 441) ((";") reduce 441) (("<") reduce 441) ((">") reduce 441) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 441) ((T_IS_GREATER_OR_EQUAL) reduce 441) ((T_IS_IDENTICAL) reduce 441) ((T_IS_NOT_EQUAL) reduce 441) ((T_IS_NOT_IDENTICAL) reduce 441) ((T_IS_SMALLER_OR_EQUAL) reduce 441) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_PIPE) reduce 441) ((T_POW) shift) ((T_SL) reduce 441) ((T_SPACESHIP) reduce 441) ((T_SR) reduce 441) (("^") reduce 441) (("|") reduce 441)) 25176 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 436) (("/") shift) ((":") reduce 436) ((";") reduce 436) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_PIPE) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 436) ((T_SR) shift) (("^") reduce 436) (("|") reduce 436)) 25177 ((("%") shift) (("*") shift) (("+") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") shift) ((":") reduce 438) ((";") reduce 438) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_PIPE) reduce 438) ((T_POW) shift) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("|") reduce 438)) 25178 ((("%") shift) (("*") shift) (("+") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") shift) ((":") reduce 437) ((";") reduce 437) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_PIPE) reduce 437) ((T_POW) shift) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("|") reduce 437)) 25179 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) ((":") reduce 439) ((";") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_PIPE) reduce 439) ((T_POW) shift) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("|") reduce 439)) 25180 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) ((":") reduce 442) ((";") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_PIPE) reduce 442) ((T_POW) shift) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("|") reduce 442)) 25182 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) ((":") reduce 74) ((";") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_PIPE) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("|") reduce 74)) 25183 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) ((":") reduce 52) ((";") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_PIPE) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("|") reduce 52)) 25184 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) ((":") reduce 27) ((";") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_PIPE) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("|") reduce 27)) 25185 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) ((":") reduce 51) ((";") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_PIPE) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("|") reduce 51)) 25186 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) ((":") reduce 53) ((";") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_PIPE) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("|") reduce 53)) 25187 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) ((":") reduce 49) ((";") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_PIPE) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("|") reduce 49)) 25188 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) ((":") reduce 29) ((";") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_PIPE) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("|") reduce 29)) 25189 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) ((":") reduce 59) ((";") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_PIPE) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("|") reduce 59)) 25190 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) ((":") reduce 60) ((";") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_PIPE) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("|") reduce 60)) 25191 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) ((":") reduce 11) ((";") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_PIPE) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("|") reduce 11)) 25192 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) ((":") reduce 42) ((";") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_PIPE) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("|") reduce 42)) 25193 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) ((":") reduce 39) ((";") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_PIPE) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("|") reduce 39)) 25194 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) ((":") reduce 25) ((";") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_PIPE) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("|") reduce 25)) 25195 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) ((":") reduce 50) ((";") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_PIPE) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("|") reduce 50)) 25196 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) ((":") reduce 66) ((";") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_PIPE) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("|") reduce 66)) 25197 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) ((":") reduce 18) ((";") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_PIPE) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("|") reduce 18)) 25198 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) ((":") reduce 17) ((";") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_PIPE) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("|") reduce 17)) 25199 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) ((":") reduce 15) ((";") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_PIPE) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("|") reduce 15)) 25200 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) ((":") reduce 14) ((";") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_PIPE) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("|") reduce 14)) 25201 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) ((":") reduce 38) ((";") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_PIPE) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("|") reduce 38)) 25202 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) ((":") reduce 26) ((";") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_PIPE) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("|") reduce 26)) 25203 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) ((":") reduce 22) ((";") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_PIPE) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("|") reduce 22)) 25204 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) ((":") reduce 24) ((";") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_PIPE) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("|") reduce 24)) 25205 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) ((":") reduce 16) ((";") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_PIPE) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("|") reduce 16)) 25206 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) ((":") reduce 48) ((";") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_PIPE) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("|") reduce 48)) 25207 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) ((":") reduce 20) ((";") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_PIPE) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("|") reduce 20)) 25208 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) ((":") reduce 70) ((";") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_PIPE) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("|") reduce 70)) 25209 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) ((":") reduce 3) ((";") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_PIPE) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("|") reduce 3)) 25210 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) ((":") reduce 12) ((";") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_PIPE) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("|") reduce 12)) 25211 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) ((":") reduce 54) ((";") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_PIPE) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("|") reduce 54)) 25212 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) ((":") reduce 65) ((";") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_PIPE) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("|") reduce 65)) 25213 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) ((":") reduce 75) ((";") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_PIPE) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("|") reduce 75)) 25214 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) ((":") reduce 30) ((";") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_PIPE) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("|") reduce 30)) 25215 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) ((":") reduce 68) ((";") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_PIPE) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("|") reduce 68)) 25216 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) ((":") reduce 21) ((";") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_PIPE) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("|") reduce 21)) 25217 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) ((":") reduce 23) ((";") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_PIPE) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("|") reduce 23)) 25218 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) ((":") reduce 41) ((";") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_PIPE) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("|") reduce 41)) 25219 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) ((":") reduce 62) ((";") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_PIPE) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("|") reduce 62)) 25220 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) ((":") reduce 34) ((";") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_PIPE) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("|") reduce 34)) 25221 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) ((":") reduce 40) ((";") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_PIPE) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("|") reduce 40)) 25222 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) ((":") reduce 13) ((";") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_PIPE) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("|") reduce 13)) 25223 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) ((":") reduce 55) ((";") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_PIPE) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("|") reduce 55)) 25224 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) ((":") reduce 1) ((";") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_PIPE) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("|") reduce 1)) 25225 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) ((":") reduce 2) ((";") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_PIPE) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("|") reduce 2)) 25226 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) ((":") reduce 9) ((";") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_PIPE) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("|") reduce 9)) 25227 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) ((":") reduce 33) ((";") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_PIPE) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("|") reduce 33)) 25228 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) ((":") reduce 58) ((";") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_PIPE) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("|") reduce 58)) 25229 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) ((":") reduce 37) ((";") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_PIPE) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("|") reduce 37)) 25230 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) ((":") reduce 64) ((";") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_PIPE) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("|") reduce 64)) 25231 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) ((":") reduce 46) ((";") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_PIPE) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("|") reduce 46)) 25232 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) ((":") reduce 8) ((";") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_PIPE) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("|") reduce 8)) 25233 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) ((":") reduce 6) ((";") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_PIPE) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("|") reduce 6)) 25234 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) ((":") reduce 7) ((";") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_PIPE) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("|") reduce 7)) 25235 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) ((":") reduce 69) ((";") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_PIPE) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("|") reduce 69)) 25236 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) ((":") reduce 63) ((";") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_PIPE) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("|") reduce 63)) 25237 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) ((":") reduce 56) ((";") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_PIPE) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("|") reduce 56)) 25238 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) ((":") reduce 10) ((";") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_PIPE) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("|") reduce 10)) 25239 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) ((":") reduce 67) ((";") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_PIPE) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("|") reduce 67)) 25240 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) ((":") reduce 44) ((";") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_PIPE) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("|") reduce 44)) 25241 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) ((":") reduce 76) ((";") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_PIPE) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("|") reduce 76)) 25242 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) ((":") reduce 71) ((";") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_PIPE) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("|") reduce 71)) 25243 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) ((":") reduce 77) ((";") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_PIPE) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("|") reduce 77)) 25244 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) ((":") reduce 78) ((";") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_PIPE) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("|") reduce 78)) 25245 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) ((":") reduce 79) ((";") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_PIPE) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("|") reduce 79)) 25246 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) ((":") reduce 4) ((";") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_PIPE) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("|") reduce 4)) 25247 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) ((":") reduce 5) ((";") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_PIPE) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("|") reduce 5)) 25248 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) ((":") reduce 43) ((";") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_PIPE) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("|") reduce 43)) 25249 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) ((":") reduce 73) ((";") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_PIPE) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("|") reduce 73)) 25250 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) ((":") reduce 82) ((";") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_PIPE) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("|") reduce 82)) 25251 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) ((":") reduce 47) ((";") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_PIPE) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("|") reduce 47)) 25252 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) ((":") reduce 31) ((";") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_PIPE) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("|") reduce 31)) 25253 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) ((":") reduce 57) ((";") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_PIPE) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("|") reduce 57)) 25254 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) ((":") reduce 61) ((";") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_PIPE) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("|") reduce 61)) 25255 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) ((":") reduce 28) ((";") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_PIPE) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("|") reduce 28)) 25256 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) ((":") reduce 36) ((";") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_PIPE) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("|") reduce 36)) 25257 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) ((":") reduce 32) ((";") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_PIPE) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("|") reduce 32)) 25258 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) ((":") reduce 35) ((";") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_PIPE) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("|") reduce 35)) 25259 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) ((":") reduce 19) ((";") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_PIPE) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("|") reduce 19)) 25260 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) ((":") reduce 45) ((";") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_PIPE) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("|") reduce 45)) 25261 ((("%") reduce 542) (("(") reduce 579) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) ((":") reduce 542) ((";") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PIPE) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 25263 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) ((":") reduce 72) ((";") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_PIPE) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("|") reduce 72)) 25264 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) ((":") reduce 83) ((";") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_PIPE) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("|") reduce 83)) 25265 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) ((":") reduce 572) ((";") reduce 572) (("<") reduce 572) (("=") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AND_EQUAL) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_COALESCE_EQUAL) reduce 572) ((T_CONCAT_EQUAL) reduce 572) ((T_DEC) reduce 572) ((T_DIV_EQUAL) reduce 572) ((T_INC) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_MINUS_EQUAL) reduce 572) ((T_MOD_EQUAL) reduce 572) ((T_MUL_EQUAL) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_OR_EQUAL) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_PLUS_EQUAL) reduce 572) ((T_POW) reduce 572) ((T_POW_EQUAL) reduce 572) ((T_SL) reduce 572) ((T_SL_EQUAL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) ((T_SR_EQUAL) reduce 572) ((T_XOR_EQUAL) reduce 572) (("[") reduce 572) (("^") reduce 572) (("|") reduce 572)) 25268 ((("%") reduce 544) (("(") reduce 580) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) ((":") reduce 544) ((";") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PIPE) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 25270 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((":") reduce 507) ((";") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("|") reduce 507)) 25271 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((":") reduce 291) ((";") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 25275 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) ((":") reduce 292) ((";") reduce 292) (("<") reduce 292) (("=") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AND_EQUAL) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_COALESCE_EQUAL) reduce 292) ((T_CONCAT_EQUAL) reduce 292) ((T_DEC) reduce 292) ((T_DIV_EQUAL) reduce 292) ((T_INC) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_MINUS_EQUAL) reduce 292) ((T_MOD_EQUAL) reduce 292) ((T_MUL_EQUAL) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_OR_EQUAL) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_PLUS_EQUAL) reduce 292) ((T_POW) reduce 292) ((T_POW_EQUAL) reduce 292) ((T_SL) reduce 292) ((T_SL_EQUAL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) ((T_SR_EQUAL) reduce 292) ((T_XOR_EQUAL) reduce 292) (("[") reduce 292) (("^") reduce 292) (("|") reduce 292)) 25276 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) ((":") reduce 293) ((";") reduce 293) (("<") reduce 293) (("=") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AND_EQUAL) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_COALESCE_EQUAL) reduce 293) ((T_CONCAT_EQUAL) reduce 293) ((T_DEC) reduce 293) ((T_DIV_EQUAL) reduce 293) ((T_INC) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_MINUS_EQUAL) reduce 293) ((T_MOD_EQUAL) reduce 293) ((T_MUL_EQUAL) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_OR_EQUAL) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_PLUS_EQUAL) reduce 293) ((T_POW) reduce 293) ((T_POW_EQUAL) reduce 293) ((T_SL) reduce 293) ((T_SL_EQUAL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) ((T_SR_EQUAL) reduce 293) ((T_XOR_EQUAL) reduce 293) (("[") reduce 293) (("^") reduce 293) (("|") reduce 293)) 25281 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) ((":") reduce 425) ((";") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_PIPE) reduce 425) ((T_POW) reduce 425) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 25283 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) ((":") reduce 423) ((";") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_PIPE) reduce 423) ((T_POW) reduce 423) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 25293 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 419) ((";") reduce 419) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25294 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 421) ((";") reduce 421) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25295 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 420) ((";") reduce 420) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25296 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 413) ((";") reduce 413) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25297 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 410) ((";") reduce 410) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25298 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 418) ((";") reduce 418) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25299 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 412) ((";") reduce 412) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25300 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 416) ((";") reduce 416) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25301 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 411) ((";") reduce 411) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25302 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 414) ((";") reduce 414) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25303 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 415) ((";") reduce 415) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25304 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 422) ((";") reduce 422) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25305 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 417) ((";") reduce 417) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25307 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 406) ((";") reduce 406) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25310 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) ((":") reduce 568) ((";") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 25313 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) ((":") reduce 564) ((";") reduce 564) (("<") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_POW) reduce 564) ((T_SL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) (("[") reduce 564) (("^") reduce 564) (("|") reduce 564)) 25315 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) ((":") reduce 563) ((";") reduce 563) (("<") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_POW) reduce 563) ((T_SL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) (("[") reduce 563) (("^") reduce 563) (("|") reduce 563)) 25317 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) ((":") reduce 559) ((";") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 25318 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((":") reduce 565) ((";") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 25319 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) ((":") reduce 407) ((";") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("[") reduce 548) (("^") reduce 407) (("|") reduce 407)) 25323 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) ((":") reduce 572) ((";") reduce 572) (("<") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_POW) reduce 572) ((T_SL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) (("[") reduce 572) (("^") reduce 572) (("|") reduce 572)) 25325 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((":") reduce 507) ((";") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("|") reduce 507)) 25326 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((":") reduce 291) ((";") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 25330 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) ((":") reduce 292) ((";") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("[") reduce 292) (("^") reduce 292) (("|") reduce 292)) 25331 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) ((":") reduce 293) ((";") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("[") reduce 293) (("^") reduce 293) (("|") reduce 293)) 25332 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) ((":") reduce 504) ((";") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 25335 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) ((":") reduce 571) ((";") reduce 571) (("<") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_POW) reduce 571) ((T_SL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) (("[") reduce 571) (("^") reduce 571) (("|") reduce 571)) 25336 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((":") reduce 506) ((";") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("|") reduce 506)) 25337 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((":") reduce 508) ((";") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("|") reduce 508)) 25342 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) ((":") reduce 560) ((";") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 25343 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) ((":") reduce 582) ((";") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 25344 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((":") reduce 566) ((";") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 25345 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) ((":") reduce 584) ((";") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 25348 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) ((":") reduce 583) ((";") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 25349 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((":") reduce 561) ((";") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 25350 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((":") reduce 567) ((";") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 25351 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) ((":") reduce 562) ((";") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("[") reduce 562) (("^") reduce 562) (("|") reduce 562)) 25352 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((":") reduce 505) ((";") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 25353 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) ((":") reduce 570) ((";") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 25356 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) ((":") reduce 569) ((";") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 25357 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) ((":") reduce 504) ((";") reduce 504) (("<") reduce 504) (("=") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AND_EQUAL) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_COALESCE_EQUAL) reduce 504) ((T_CONCAT_EQUAL) reduce 504) ((T_DEC) reduce 504) ((T_DIV_EQUAL) reduce 504) ((T_INC) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_MINUS_EQUAL) reduce 504) ((T_MOD_EQUAL) reduce 504) ((T_MUL_EQUAL) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_OR_EQUAL) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_PLUS_EQUAL) reduce 504) ((T_POW) reduce 504) ((T_POW_EQUAL) reduce 504) ((T_SL) reduce 504) ((T_SL_EQUAL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) ((T_SR_EQUAL) reduce 504) ((T_XOR_EQUAL) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 25368 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) ((":") reduce 494) ((";") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PIPE) reduce 494) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 25369 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) ((":") reduce 489) ((";") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_PIPE) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("^") reduce 489) (("|") reduce 489)) 25442 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 494) ((";") reduce 494) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25758 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) ((":") reduce 490) ((";") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_PIPE) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("^") reduce 490) (("|") reduce 490)) 25761 ((("%") reduce 541) (("(") reduce 579) (("*") reduce 541) (("+") reduce 541) (("-") reduce 541) ((".") reduce 541) (("/") reduce 541) ((":") reduce 541) ((";") reduce 541) (("<") reduce 541) ((">") reduce 541) (("?") reduce 541) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_BOOLEAN_AND) reduce 541) ((T_BOOLEAN_OR) reduce 541) ((T_COALESCE) reduce 541) ((T_INSTANCEOF) reduce 541) ((T_IS_EQUAL) reduce 541) ((T_IS_GREATER_OR_EQUAL) reduce 541) ((T_IS_IDENTICAL) reduce 541) ((T_IS_NOT_EQUAL) reduce 541) ((T_IS_NOT_IDENTICAL) reduce 541) ((T_IS_SMALLER_OR_EQUAL) reduce 541) ((T_LOGICAL_AND) reduce 541) ((T_LOGICAL_OR) reduce 541) ((T_LOGICAL_XOR) reduce 541) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 541) ((T_OBJECT_OPERATOR) reduce 541) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 541) ((T_PIPE) reduce 541) ((T_POW) reduce 541) ((T_SL) reduce 541) ((T_SPACESHIP) reduce 541) ((T_SR) reduce 541) (("[") reduce 541) (("^") reduce 541) (("|") reduce 541)) 25763 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) ((":") reduce 571) ((";") reduce 571) (("<") reduce 571) (("=") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AND_EQUAL) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_COALESCE_EQUAL) reduce 571) ((T_CONCAT_EQUAL) reduce 571) ((T_DEC) reduce 571) ((T_DIV_EQUAL) reduce 571) ((T_INC) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_MINUS_EQUAL) reduce 571) ((T_MOD_EQUAL) reduce 571) ((T_MUL_EQUAL) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_OR_EQUAL) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_PLUS_EQUAL) reduce 571) ((T_POW) reduce 571) ((T_POW_EQUAL) reduce 571) ((T_SL) reduce 571) ((T_SL_EQUAL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) ((T_SR_EQUAL) reduce 571) ((T_XOR_EQUAL) reduce 571) (("[") reduce 571) (("^") reduce 571) (("|") reduce 571)) 25766 ((("%") reduce 543) (("(") reduce 580) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) ((":") reduce 543) ((";") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PIPE) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 25767 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((":") reduce 506) ((";") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("|") reduce 506)) 25768 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((":") reduce 508) ((";") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("|") reduce 508)) 25770 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) ((":") reduce 485) ((";") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_PIPE) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("^") reduce 485) (("|") reduce 485)) 25771 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) ((":") reduce 487) ((";") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_PIPE) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("^") reduce 487) (("|") reduce 487)) 25776 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) ((":") reduce 560) ((";") reduce 560) (("<") reduce 560) (("=") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AND_EQUAL) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_COALESCE_EQUAL) reduce 560) ((T_CONCAT_EQUAL) reduce 560) ((T_DEC) reduce 560) ((T_DIV_EQUAL) reduce 560) ((T_INC) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_MINUS_EQUAL) reduce 560) ((T_MOD_EQUAL) reduce 560) ((T_MUL_EQUAL) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_OR_EQUAL) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_PLUS_EQUAL) reduce 560) ((T_POW) reduce 560) ((T_POW_EQUAL) reduce 560) ((T_SL) reduce 560) ((T_SL_EQUAL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) ((T_SR_EQUAL) reduce 560) ((T_XOR_EQUAL) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 25777 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) ((":") reduce 582) ((";") reduce 582) (("<") reduce 582) (("=") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AND_EQUAL) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_COALESCE_EQUAL) reduce 582) ((T_CONCAT_EQUAL) reduce 582) ((T_DEC) reduce 582) ((T_DIV_EQUAL) reduce 582) ((T_INC) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_MINUS_EQUAL) reduce 582) ((T_MOD_EQUAL) reduce 582) ((T_MUL_EQUAL) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_OR_EQUAL) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_PLUS_EQUAL) reduce 582) ((T_POW) reduce 582) ((T_POW_EQUAL) reduce 582) ((T_SL) reduce 582) ((T_SL_EQUAL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) ((T_SR_EQUAL) reduce 582) ((T_XOR_EQUAL) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 25778 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((":") reduce 566) ((";") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 25779 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) ((":") reduce 584) ((";") reduce 584) (("<") reduce 584) (("=") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AND_EQUAL) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_COALESCE_EQUAL) reduce 584) ((T_CONCAT_EQUAL) reduce 584) ((T_DEC) reduce 584) ((T_DIV_EQUAL) reduce 584) ((T_INC) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_MINUS_EQUAL) reduce 584) ((T_MOD_EQUAL) reduce 584) ((T_MUL_EQUAL) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_OR_EQUAL) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_PLUS_EQUAL) reduce 584) ((T_POW) reduce 584) ((T_POW_EQUAL) reduce 584) ((T_SL) reduce 584) ((T_SL_EQUAL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) ((T_SR_EQUAL) reduce 584) ((T_XOR_EQUAL) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 25782 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) ((":") reduce 583) ((";") reduce 583) (("<") reduce 583) (("=") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AND_EQUAL) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_COALESCE_EQUAL) reduce 583) ((T_CONCAT_EQUAL) reduce 583) ((T_DEC) reduce 583) ((T_DIV_EQUAL) reduce 583) ((T_INC) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_MINUS_EQUAL) reduce 583) ((T_MOD_EQUAL) reduce 583) ((T_MUL_EQUAL) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_OR_EQUAL) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_PLUS_EQUAL) reduce 583) ((T_POW) reduce 583) ((T_POW_EQUAL) reduce 583) ((T_SL) reduce 583) ((T_SL_EQUAL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) ((T_SR_EQUAL) reduce 583) ((T_XOR_EQUAL) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 25783 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((":") reduce 561) ((";") reduce 561) (("<") reduce 561) (("=") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AND_EQUAL) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_COALESCE_EQUAL) reduce 561) ((T_CONCAT_EQUAL) reduce 561) ((T_DEC) reduce 561) ((T_DIV_EQUAL) reduce 561) ((T_INC) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_MINUS_EQUAL) reduce 561) ((T_MOD_EQUAL) reduce 561) ((T_MUL_EQUAL) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_OR_EQUAL) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_PLUS_EQUAL) reduce 561) ((T_POW) reduce 561) ((T_POW_EQUAL) reduce 561) ((T_SL) reduce 561) ((T_SL_EQUAL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) ((T_SR_EQUAL) reduce 561) ((T_XOR_EQUAL) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 25784 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((":") reduce 567) ((";") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 25785 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) ((":") reduce 562) ((";") reduce 562) (("<") reduce 562) (("=") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AND_EQUAL) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_COALESCE_EQUAL) reduce 562) ((T_CONCAT_EQUAL) reduce 562) ((T_DEC) reduce 562) ((T_DIV_EQUAL) reduce 562) ((T_INC) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_MINUS_EQUAL) reduce 562) ((T_MOD_EQUAL) reduce 562) ((T_MUL_EQUAL) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_OR_EQUAL) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_PLUS_EQUAL) reduce 562) ((T_POW) reduce 562) ((T_POW_EQUAL) reduce 562) ((T_SL) reduce 562) ((T_SL_EQUAL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) ((T_SR_EQUAL) reduce 562) ((T_XOR_EQUAL) reduce 562) (("[") reduce 562) (("^") reduce 562) (("|") reduce 562)) 25787 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) ((":") reduce 477) ((";") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_PIPE) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 25789 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) ((":") reduce 520) ((";") reduce 520) (("<") reduce 520) (("=") shift) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_PIPE) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("|") reduce 520)) 25791 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 405) ((";") reduce 405) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25792 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 482) ((";") reduce 482) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25803 ((("%") reduce 540) (("*") reduce 540) (("+") reduce 540) (("-") reduce 540) ((".") reduce 540) (("/") reduce 540) ((":") reduce 540) ((";") reduce 540) (("<") reduce 540) ((">") reduce 540) (("?") reduce 540) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_BOOLEAN_AND) reduce 540) ((T_BOOLEAN_OR) reduce 540) ((T_COALESCE) reduce 540) ((T_DOUBLE_ARROW) reduce 540) ((T_INSTANCEOF) reduce 540) ((T_IS_EQUAL) reduce 540) ((T_IS_GREATER_OR_EQUAL) reduce 540) ((T_IS_IDENTICAL) reduce 540) ((T_IS_NOT_EQUAL) reduce 540) ((T_IS_NOT_IDENTICAL) reduce 540) ((T_IS_SMALLER_OR_EQUAL) reduce 540) ((T_LOGICAL_AND) reduce 540) ((T_LOGICAL_OR) reduce 540) ((T_LOGICAL_XOR) reduce 540) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 540) ((T_OBJECT_OPERATOR) reduce 540) ((T_PIPE) reduce 540) ((T_POW) reduce 540) ((T_SL) reduce 540) ((T_SPACESHIP) reduce 540) ((T_SR) reduce 540) (("[") reduce 540) (("^") reduce 540) (("|") reduce 540)) 25805 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) ((":") reduce 521) ((";") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_PIPE) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("|") reduce 521)) 25807 ((("%") reduce 534) (("*") reduce 534) (("+") reduce 534) (("-") reduce 534) ((".") reduce 534) (("/") reduce 534) ((":") reduce 534) ((";") reduce 534) (("<") reduce 534) ((">") reduce 534) (("?") reduce 534) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_BOOLEAN_AND) reduce 534) ((T_BOOLEAN_OR) reduce 534) ((T_COALESCE) reduce 534) ((T_DOUBLE_ARROW) reduce 534) ((T_INSTANCEOF) reduce 534) ((T_IS_EQUAL) reduce 534) ((T_IS_GREATER_OR_EQUAL) reduce 534) ((T_IS_IDENTICAL) reduce 534) ((T_IS_NOT_EQUAL) reduce 534) ((T_IS_NOT_IDENTICAL) reduce 534) ((T_IS_SMALLER_OR_EQUAL) reduce 534) ((T_LOGICAL_AND) reduce 534) ((T_LOGICAL_OR) reduce 534) ((T_LOGICAL_XOR) reduce 534) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PIPE) reduce 534) ((T_POW) reduce 534) ((T_SL) reduce 534) ((T_SPACESHIP) reduce 534) ((T_SR) reduce 534) (("[") reduce 534) (("^") reduce 534) (("|") reduce 534)) 25808 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) ((":") reduce 524) ((";") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_PIPE) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("^") reduce 524) (("|") reduce 524)) 25812 ((("%") reduce 517) (("(") shift) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((":") reduce 517) ((";") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_PIPE) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("^") reduce 517) (("|") reduce 517)) 25813 ((("%") reduce 533) (("*") reduce 533) (("+") reduce 533) (("-") reduce 533) ((".") reduce 533) (("/") reduce 533) ((":") reduce 533) ((";") reduce 533) (("<") reduce 533) ((">") reduce 533) (("?") reduce 533) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_BOOLEAN_AND) reduce 533) ((T_BOOLEAN_OR) reduce 533) ((T_COALESCE) reduce 533) ((T_DOUBLE_ARROW) reduce 533) ((T_INSTANCEOF) reduce 533) ((T_IS_EQUAL) reduce 533) ((T_IS_GREATER_OR_EQUAL) reduce 533) ((T_IS_IDENTICAL) reduce 533) ((T_IS_NOT_EQUAL) reduce 533) ((T_IS_NOT_IDENTICAL) reduce 533) ((T_IS_SMALLER_OR_EQUAL) reduce 533) ((T_LOGICAL_AND) reduce 533) ((T_LOGICAL_OR) reduce 533) ((T_LOGICAL_XOR) reduce 533) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PIPE) reduce 533) ((T_POW) reduce 533) ((T_SL) reduce 533) ((T_SPACESHIP) reduce 533) ((T_SR) reduce 533) (("[") reduce 533) (("^") reduce 533) (("|") reduce 533)) 25814 ((("%") reduce 537) (("*") reduce 537) (("+") reduce 537) (("-") reduce 537) ((".") reduce 537) (("/") reduce 537) ((":") reduce 537) ((";") reduce 537) (("<") reduce 537) ((">") reduce 537) (("?") reduce 537) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_BOOLEAN_AND) reduce 537) ((T_BOOLEAN_OR) reduce 537) ((T_COALESCE) reduce 537) ((T_DOUBLE_ARROW) reduce 537) ((T_INSTANCEOF) reduce 537) ((T_IS_EQUAL) reduce 537) ((T_IS_GREATER_OR_EQUAL) reduce 537) ((T_IS_IDENTICAL) reduce 537) ((T_IS_NOT_EQUAL) reduce 537) ((T_IS_NOT_IDENTICAL) reduce 537) ((T_IS_SMALLER_OR_EQUAL) reduce 537) ((T_LOGICAL_AND) reduce 537) ((T_LOGICAL_OR) reduce 537) ((T_LOGICAL_XOR) reduce 537) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_PIPE) reduce 537) ((T_POW) reduce 537) ((T_SL) reduce 537) ((T_SPACESHIP) reduce 537) ((T_SR) reduce 537) (("[") reduce 537) (("^") reduce 537) (("|") reduce 537)) 25820 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) ((":") reduce 532) ((";") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PIPE) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("|") reduce 532)) 25822 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) ((":") reduce 523) ((";") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_PIPE) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("^") reduce 523) (("|") reduce 523)) 25824 ((("%") reduce 536) (("*") reduce 536) (("+") reduce 536) (("-") reduce 536) ((".") reduce 536) (("/") reduce 536) ((":") reduce 536) ((";") reduce 536) (("<") reduce 536) ((">") reduce 536) (("?") reduce 536) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_BOOLEAN_AND) reduce 536) ((T_BOOLEAN_OR) reduce 536) ((T_COALESCE) reduce 536) ((T_DOUBLE_ARROW) reduce 536) ((T_INSTANCEOF) reduce 536) ((T_IS_EQUAL) reduce 536) ((T_IS_GREATER_OR_EQUAL) reduce 536) ((T_IS_IDENTICAL) reduce 536) ((T_IS_NOT_EQUAL) reduce 536) ((T_IS_NOT_IDENTICAL) reduce 536) ((T_IS_SMALLER_OR_EQUAL) reduce 536) ((T_LOGICAL_AND) reduce 536) ((T_LOGICAL_OR) reduce 536) ((T_LOGICAL_XOR) reduce 536) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 536) ((T_OBJECT_OPERATOR) reduce 536) ((T_PIPE) reduce 536) ((T_POW) reduce 536) ((T_SL) reduce 536) ((T_SPACESHIP) reduce 536) ((T_SR) reduce 536) (("[") reduce 536) (("^") reduce 536) (("|") reduce 536)) 25825 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((":") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("|") reduce 94)) 25826 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((":") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("|") reduce 93)) 25827 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((":") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("|") reduce 95)) 25829 ((("%") reduce 539) (("*") reduce 539) (("+") reduce 539) (("-") reduce 539) ((".") reduce 539) (("/") reduce 539) ((":") reduce 539) ((";") reduce 539) (("<") reduce 539) ((">") reduce 539) (("?") reduce 539) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_BOOLEAN_AND) reduce 539) ((T_BOOLEAN_OR) reduce 539) ((T_COALESCE) reduce 539) ((T_DOUBLE_ARROW) reduce 539) ((T_INSTANCEOF) reduce 539) ((T_IS_EQUAL) reduce 539) ((T_IS_GREATER_OR_EQUAL) reduce 539) ((T_IS_IDENTICAL) reduce 539) ((T_IS_NOT_EQUAL) reduce 539) ((T_IS_NOT_IDENTICAL) reduce 539) ((T_IS_SMALLER_OR_EQUAL) reduce 539) ((T_LOGICAL_AND) reduce 539) ((T_LOGICAL_OR) reduce 539) ((T_LOGICAL_XOR) reduce 539) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 539) ((T_OBJECT_OPERATOR) reduce 539) ((T_PIPE) reduce 539) ((T_POW) reduce 539) ((T_SL) reduce 539) ((T_SPACESHIP) reduce 539) ((T_SR) reduce 539) (("[") reduce 539) (("^") reduce 539) (("|") reduce 539)) 25832 ((("%") reduce 538) (("*") reduce 538) (("+") reduce 538) (("-") reduce 538) ((".") reduce 538) (("/") reduce 538) ((":") reduce 538) ((";") reduce 538) (("<") reduce 538) ((">") reduce 538) (("?") reduce 538) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_BOOLEAN_AND) reduce 538) ((T_BOOLEAN_OR) reduce 538) ((T_COALESCE) reduce 538) ((T_DOUBLE_ARROW) reduce 538) ((T_INSTANCEOF) reduce 538) ((T_IS_EQUAL) reduce 538) ((T_IS_GREATER_OR_EQUAL) reduce 538) ((T_IS_IDENTICAL) reduce 538) ((T_IS_NOT_EQUAL) reduce 538) ((T_IS_NOT_IDENTICAL) reduce 538) ((T_IS_SMALLER_OR_EQUAL) reduce 538) ((T_LOGICAL_AND) reduce 538) ((T_LOGICAL_OR) reduce 538) ((T_LOGICAL_XOR) reduce 538) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 538) ((T_OBJECT_OPERATOR) reduce 538) ((T_PIPE) reduce 538) ((T_POW) reduce 538) ((T_SL) reduce 538) ((T_SPACESHIP) reduce 538) ((T_SR) reduce 538) (("[") reduce 538) (("^") reduce 538) (("|") reduce 538)) 25838 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((":") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92)) 25841 ((("%") reduce 535) (("*") reduce 535) (("+") reduce 535) (("-") reduce 535) ((".") reduce 535) (("/") reduce 535) ((":") reduce 535) ((";") reduce 535) (("<") reduce 535) ((">") reduce 535) (("?") reduce 535) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_BOOLEAN_AND) reduce 535) ((T_BOOLEAN_OR) reduce 535) ((T_COALESCE) reduce 535) ((T_DOUBLE_ARROW) reduce 535) ((T_INSTANCEOF) reduce 535) ((T_IS_EQUAL) reduce 535) ((T_IS_GREATER_OR_EQUAL) reduce 535) ((T_IS_IDENTICAL) reduce 535) ((T_IS_NOT_EQUAL) reduce 535) ((T_IS_NOT_IDENTICAL) reduce 535) ((T_IS_SMALLER_OR_EQUAL) reduce 535) ((T_LOGICAL_AND) reduce 535) ((T_LOGICAL_OR) reduce 535) ((T_LOGICAL_XOR) reduce 535) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PIPE) reduce 535) ((T_POW) reduce 535) ((T_SL) reduce 535) ((T_SPACESHIP) reduce 535) ((T_SR) reduce 535) (("[") reduce 535) (("^") reduce 535) (("|") reduce 535)) 25843 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) ((":") reduce 568) ((";") reduce 568) (("<") reduce 568) (("=") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AND_EQUAL) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_COALESCE_EQUAL) reduce 568) ((T_CONCAT_EQUAL) reduce 568) ((T_DEC) reduce 568) ((T_DIV_EQUAL) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INC) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_MINUS_EQUAL) reduce 568) ((T_MOD_EQUAL) reduce 568) ((T_MUL_EQUAL) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_OR_EQUAL) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_PLUS_EQUAL) reduce 568) ((T_POW) reduce 568) ((T_POW_EQUAL) reduce 568) ((T_SL) reduce 568) ((T_SL_EQUAL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) ((T_SR_EQUAL) reduce 568) ((T_XOR_EQUAL) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 25844 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 479) (("(") shift) (("*") reduce 479) (("+") shift) (("-") shift) ((".") reduce 479) (("/") reduce 479) ((":") reduce 479) ((";") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 479) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 479) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 479) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PIPE) reduce 479) ((T_POW) reduce 479) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 479) (("`") shift) (("|") reduce 479) (("~") shift)) 25851 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) ((":") reduce 564) ((";") reduce 564) (("<") reduce 564) (("=") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AND_EQUAL) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_COALESCE_EQUAL) reduce 564) ((T_CONCAT_EQUAL) reduce 564) ((T_DEC) reduce 564) ((T_DIV_EQUAL) reduce 564) ((T_DOUBLE_ARROW) reduce 564) ((T_INC) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_MINUS_EQUAL) reduce 564) ((T_MOD_EQUAL) reduce 564) ((T_MUL_EQUAL) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_OR_EQUAL) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_PLUS_EQUAL) reduce 564) ((T_POW) reduce 564) ((T_POW_EQUAL) reduce 564) ((T_SL) reduce 564) ((T_SL_EQUAL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) ((T_SR_EQUAL) reduce 564) ((T_XOR_EQUAL) reduce 564) (("[") reduce 564) (("^") reduce 564) (("|") reduce 564)) 25852 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) ((":") reduce 530) ((";") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PIPE) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 551) (("^") reduce 530) (("|") reduce 530)) 25854 ((("%") reduce 529) (("*") reduce 529) (("+") reduce 529) (("-") reduce 529) ((".") reduce 529) (("/") reduce 529) ((":") reduce 529) ((";") reduce 529) (("<") reduce 529) ((">") reduce 529) (("?") reduce 529) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_BOOLEAN_AND) reduce 529) ((T_BOOLEAN_OR) reduce 529) ((T_COALESCE) reduce 529) ((T_DOUBLE_ARROW) reduce 529) ((T_INSTANCEOF) reduce 529) ((T_IS_EQUAL) reduce 529) ((T_IS_GREATER_OR_EQUAL) reduce 529) ((T_IS_IDENTICAL) reduce 529) ((T_IS_NOT_EQUAL) reduce 529) ((T_IS_NOT_IDENTICAL) reduce 529) ((T_IS_SMALLER_OR_EQUAL) reduce 529) ((T_LOGICAL_AND) reduce 529) ((T_LOGICAL_OR) reduce 529) ((T_LOGICAL_XOR) reduce 529) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PIPE) reduce 529) ((T_POW) reduce 529) ((T_SL) reduce 529) ((T_SPACESHIP) reduce 529) ((T_SR) reduce 529) (("[") reduce 554) (("^") reduce 529) (("|") reduce 529)) 25855 ((("%") reduce 528) (("(") reduce 557) (("*") reduce 528) (("+") reduce 528) (("-") reduce 528) ((".") reduce 528) (("/") reduce 528) ((":") reduce 528) ((";") reduce 528) (("<") reduce 528) ((">") reduce 528) (("?") reduce 528) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_BOOLEAN_AND) reduce 528) ((T_BOOLEAN_OR) reduce 528) ((T_COALESCE) reduce 528) ((T_DOUBLE_ARROW) reduce 528) ((T_INSTANCEOF) reduce 528) ((T_IS_EQUAL) reduce 528) ((T_IS_GREATER_OR_EQUAL) reduce 528) ((T_IS_IDENTICAL) reduce 528) ((T_IS_NOT_EQUAL) reduce 528) ((T_IS_NOT_IDENTICAL) reduce 528) ((T_IS_SMALLER_OR_EQUAL) reduce 528) ((T_LOGICAL_AND) reduce 528) ((T_LOGICAL_OR) reduce 528) ((T_LOGICAL_XOR) reduce 528) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PIPE) reduce 528) ((T_POW) reduce 528) ((T_SL) reduce 528) ((T_SPACESHIP) reduce 528) ((T_SR) reduce 528) (("[") reduce 550) (("^") reduce 528) (("|") reduce 528)) 25856 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 480) ((";") reduce 480) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 480) ((T_LOGICAL_OR) reduce 480) ((T_LOGICAL_XOR) reduce 480) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25859 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) ((":") reduce 563) ((";") reduce 563) (("<") reduce 563) (("=") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AND_EQUAL) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_COALESCE_EQUAL) reduce 563) ((T_CONCAT_EQUAL) reduce 563) ((T_DEC) reduce 563) ((T_DIV_EQUAL) reduce 563) ((T_DOUBLE_ARROW) reduce 563) ((T_INC) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_MINUS_EQUAL) reduce 563) ((T_MOD_EQUAL) reduce 563) ((T_MUL_EQUAL) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_OR_EQUAL) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_PLUS_EQUAL) reduce 563) ((T_POW) reduce 563) ((T_POW_EQUAL) reduce 563) ((T_SL) reduce 563) ((T_SL_EQUAL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) ((T_SR_EQUAL) reduce 563) ((T_XOR_EQUAL) reduce 563) (("[") reduce 563) (("^") reduce 563) (("|") reduce 563)) 25860 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) ((":") reduce 484) ((";") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_DOUBLE_ARROW) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_PIPE) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("^") reduce 484) (("|") reduce 484)) 25861 ((("%") reduce 466) (("*") reduce 466) (("+") reduce 466) (("-") reduce 466) ((".") reduce 466) (("/") reduce 466) ((":") reduce 466) ((";") reduce 466) (("<") reduce 466) ((">") reduce 466) (("?") reduce 466) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_BOOLEAN_AND) reduce 466) ((T_BOOLEAN_OR) reduce 466) ((T_COALESCE) reduce 466) ((T_DOUBLE_ARROW) reduce 466) ((T_INSTANCEOF) reduce 466) ((T_IS_EQUAL) reduce 466) ((T_IS_GREATER_OR_EQUAL) reduce 466) ((T_IS_IDENTICAL) reduce 466) ((T_IS_NOT_EQUAL) reduce 466) ((T_IS_NOT_IDENTICAL) reduce 466) ((T_IS_SMALLER_OR_EQUAL) reduce 466) ((T_LOGICAL_AND) reduce 466) ((T_LOGICAL_OR) reduce 466) ((T_LOGICAL_XOR) reduce 466) ((T_PIPE) reduce 466) ((T_POW) reduce 466) ((T_SL) reduce 466) ((T_SPACESHIP) reduce 466) ((T_SR) reduce 466) (("^") reduce 466) (("|") reduce 466)) 25862 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) ((":") reduce 488) ((";") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_PIPE) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("^") reduce 488) (("|") reduce 488)) 25863 ((("%") reduce 531) (("(") shift) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) ((":") reduce 531) ((";") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("|") reduce 531)) 25864 ((("%") reduce 461) (("(") reduce 558) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) ((":") reduce 461) ((";") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PIPE) reduce 461) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") reduce 552) (("^") reduce 461) (("|") reduce 461)) 25865 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) ((":") reduce 462) ((";") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_DOUBLE_ARROW) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_PIPE) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 25866 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) ((":") reduce 476) ((";") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_DOUBLE_ARROW) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_PIPE) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("|") reduce 476)) 25867 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) ((":") reduce 559) ((";") reduce 559) (("<") reduce 559) (("=") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AND_EQUAL) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_COALESCE_EQUAL) reduce 559) ((T_CONCAT_EQUAL) reduce 559) ((T_DEC) reduce 559) ((T_DIV_EQUAL) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INC) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_MINUS_EQUAL) reduce 559) ((T_MOD_EQUAL) reduce 559) ((T_MUL_EQUAL) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_OR_EQUAL) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_PLUS_EQUAL) reduce 559) ((T_POW) reduce 559) ((T_POW_EQUAL) reduce 559) ((T_SL) reduce 559) ((T_SL_EQUAL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) ((T_SR_EQUAL) reduce 559) ((T_XOR_EQUAL) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 25868 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((":") reduce 565) ((";") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 25869 ((("%") reduce 403) (("*") reduce 403) (("+") reduce 403) (("-") reduce 403) ((".") reduce 403) (("/") reduce 403) ((":") reduce 403) ((";") reduce 403) (("<") reduce 403) (("=") shift) ((">") reduce 403) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_DOUBLE_ARROW) reduce 403) ((T_INC) shift) ((T_INSTANCEOF) reduce 403) ((T_IS_EQUAL) reduce 403) ((T_IS_GREATER_OR_EQUAL) reduce 403) ((T_IS_IDENTICAL) reduce 403) ((T_IS_NOT_EQUAL) reduce 403) ((T_IS_NOT_IDENTICAL) reduce 403) ((T_IS_SMALLER_OR_EQUAL) reduce 403) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 403) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 403) ((T_POW_EQUAL) shift) ((T_SL) reduce 403) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 403) ((T_SR) reduce 403) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 548) (("^") reduce 403) (("|") reduce 403)) 25872 ((("%") reduce 448) (("*") reduce 448) (("+") reduce 448) (("-") reduce 448) ((".") reduce 448) (("/") reduce 448) ((":") reduce 448) ((";") reduce 448) (("<") reduce 448) ((">") reduce 448) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_BOOLEAN_AND) reduce 448) ((T_BOOLEAN_OR) reduce 448) ((T_COALESCE) reduce 448) ((T_DOUBLE_ARROW) reduce 448) ((T_INSTANCEOF) reduce 448) ((T_IS_EQUAL) reduce 448) ((T_IS_GREATER_OR_EQUAL) reduce 448) ((T_IS_IDENTICAL) reduce 448) ((T_IS_NOT_EQUAL) reduce 448) ((T_IS_NOT_IDENTICAL) reduce 448) ((T_IS_SMALLER_OR_EQUAL) reduce 448) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_PIPE) reduce 448) ((T_POW) shift) ((T_SL) reduce 448) ((T_SPACESHIP) reduce 448) ((T_SR) reduce 448) (("^") reduce 448) (("|") reduce 448)) 25904 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 432) ((";") reduce 432) (("<") shift) ((">") shift) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_DOUBLE_ARROW) reduce 432) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 432)) 25905 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 435) ((";") reduce 435) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_DOUBLE_ARROW) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("|") reduce 435)) 25906 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 444) (("/") shift) ((":") reduce 444) ((";") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_PIPE) reduce 444) ((T_POW) shift) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("|") reduce 444)) 25907 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 458) ((";") reduce 458) (("<") shift) ((">") shift) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 458) (("|") reduce 458)) 25908 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 443) (("/") shift) ((":") reduce 443) ((";") reduce 443) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_DOUBLE_ARROW) reduce 443) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_PIPE) reduce 443) ((T_POW) shift) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("^") reduce 443) (("|") reduce 443)) 25909 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) ((":") reduce 440) ((";") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_DOUBLE_ARROW) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_PIPE) reduce 440) ((T_POW) shift) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("|") reduce 440)) 25910 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 453) ((";") reduce 453) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_DOUBLE_ARROW) reduce 453) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_PIPE) reduce 453) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 453) ((T_SR) shift) (("^") reduce 453) (("|") reduce 453)) 25911 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 431) ((";") reduce 431) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25912 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 429) ((";") reduce 429) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 429) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25913 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 430) ((";") reduce 430) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 430) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25914 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 455) ((";") reduce 455) (("<") shift) ((">") shift) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 455) ((T_SR) shift) (("^") reduce 455) (("|") reduce 455)) 25915 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 450) ((";") reduce 450) (("<") shift) ((">") shift) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_DOUBLE_ARROW) reduce 450) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 450) (("|") reduce 450)) 25916 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 452) ((";") reduce 452) (("<") shift) ((">") shift) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_DOUBLE_ARROW) reduce 452) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 452) (("|") reduce 452)) 25917 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 449) ((";") reduce 449) (("<") shift) ((">") shift) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_DOUBLE_ARROW) reduce 449) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 449) (("|") reduce 449)) 25918 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 457) ((";") reduce 457) (("<") shift) ((">") shift) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 457) ((T_SR) shift) (("^") reduce 457) (("|") reduce 457)) 25919 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 451) ((";") reduce 451) (("<") shift) ((">") shift) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_DOUBLE_ARROW) reduce 451) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 451) (("|") reduce 451)) 25922 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((":") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 25923 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((":") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 25924 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((":") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 25925 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((":") reduce 509) ((";") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 25926 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((":") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 25927 ((("%") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) ((":") reduce 568) ((";") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 25928 ((("%") reduce 511) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((":") reduce 511) ((";") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("|") reduce 511)) 25929 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) ((":") reduce 459) ((";") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PIPE) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 25930 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((":") reduce 510) ((";") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 25931 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((":") reduce 512) ((";") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("|") reduce 512)) 25932 ((("%") reduce 573) (("*") reduce 573) (("+") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) ((":") reduce 573) ((";") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_DOUBLE_ARROW) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("|") reduce 573)) 25938 ((("%") reduce 574) (("*") reduce 574) (("+") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) ((":") reduce 574) ((";") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_DOUBLE_ARROW) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("|") reduce 574)) 25939 ((("%") reduce 578) (("*") reduce 578) (("+") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) ((":") reduce 578) ((";") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_DOUBLE_ARROW) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("|") reduce 578)) 25940 ((("%") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) ((":") reduce 582) ((";") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 25941 ((("%") reduce 575) (("*") reduce 575) (("+") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) ((":") reduce 575) ((";") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_DOUBLE_ARROW) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("|") reduce 575)) 25942 ((("%") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) ((":") reduce 584) ((";") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 25945 ((("%") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) ((":") reduce 583) ((";") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 25946 ((("%") reduce 576) (("*") reduce 576) (("+") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) ((":") reduce 576) ((";") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_DOUBLE_ARROW) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("|") reduce 576)) 25948 ((("%") reduce 577) (("*") reduce 577) (("+") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) ((":") reduce 577) ((";") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_DOUBLE_ARROW) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("|") reduce 577)) 25950 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((":") reduce 513) ((";") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("|") reduce 513)) 25951 ((("%") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) ((":") reduce 570) ((";") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 25954 ((("%") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) ((":") reduce 569) ((";") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 25955 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 465) ((";") reduce 465) (("<") shift) ((">") shift) (("?") reduce 465) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 465) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25956 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 427) ((";") reduce 427) (("<") shift) ((">") shift) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_DOUBLE_ARROW) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25957 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 428) ((";") reduce 428) (("<") shift) ((">") shift) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_DOUBLE_ARROW) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25958 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 433) ((";") reduce 433) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_DOUBLE_ARROW) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("|") reduce 433)) 25959 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 434) ((";") reduce 434) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_DOUBLE_ARROW) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("|") reduce 434)) 25963 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 463) ((";") reduce 463) (("<") shift) ((">") shift) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 463) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25964 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 464) ((";") reduce 464) (("<") shift) ((">") shift) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 464) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 25965 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 456) ((";") reduce 456) (("<") shift) ((">") shift) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 456) ((T_SR) shift) (("^") reduce 456) (("|") reduce 456)) 25966 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 454) ((";") reduce 454) (("<") shift) ((">") shift) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 454) ((T_SR) shift) (("^") reduce 454) (("|") reduce 454)) 25967 ((("%") reduce 441) (("*") reduce 441) (("+") reduce 441) (("-") reduce 441) ((".") reduce 441) (("/") reduce 441) ((":") reduce 441) ((";") reduce 441) (("<") reduce 441) ((">") reduce 441) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_DOUBLE_ARROW) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 441) ((T_IS_GREATER_OR_EQUAL) reduce 441) ((T_IS_IDENTICAL) reduce 441) ((T_IS_NOT_EQUAL) reduce 441) ((T_IS_NOT_IDENTICAL) reduce 441) ((T_IS_SMALLER_OR_EQUAL) reduce 441) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_PIPE) reduce 441) ((T_POW) shift) ((T_SL) reduce 441) ((T_SPACESHIP) reduce 441) ((T_SR) reduce 441) (("^") reduce 441) (("|") reduce 441)) 25968 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 436) (("/") shift) ((":") reduce 436) ((";") reduce 436) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_DOUBLE_ARROW) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_PIPE) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 436) ((T_SR) shift) (("^") reduce 436) (("|") reduce 436)) 25969 ((("%") shift) (("*") shift) (("+") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") shift) ((":") reduce 438) ((";") reduce 438) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_DOUBLE_ARROW) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_PIPE) reduce 438) ((T_POW) shift) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("|") reduce 438)) 25970 ((("%") shift) (("*") shift) (("+") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") shift) ((":") reduce 437) ((";") reduce 437) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_DOUBLE_ARROW) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_PIPE) reduce 437) ((T_POW) shift) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("|") reduce 437)) 25971 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) ((":") reduce 439) ((";") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_DOUBLE_ARROW) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_PIPE) reduce 439) ((T_POW) shift) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("|") reduce 439)) 25972 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) ((":") reduce 442) ((";") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_DOUBLE_ARROW) reduce 442) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_PIPE) reduce 442) ((T_POW) shift) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("|") reduce 442)) 25974 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) ((":") reduce 74) ((";") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_DOUBLE_ARROW) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_PIPE) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("|") reduce 74)) 25975 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) ((":") reduce 52) ((";") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_DOUBLE_ARROW) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_PIPE) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("|") reduce 52)) 25976 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) ((":") reduce 27) ((";") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_DOUBLE_ARROW) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_PIPE) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("|") reduce 27)) 25977 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) ((":") reduce 51) ((";") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_DOUBLE_ARROW) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_PIPE) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("|") reduce 51)) 25978 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) ((":") reduce 53) ((";") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_DOUBLE_ARROW) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_PIPE) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("|") reduce 53)) 25979 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) ((":") reduce 49) ((";") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_DOUBLE_ARROW) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_PIPE) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("|") reduce 49)) 25980 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) ((":") reduce 29) ((";") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_DOUBLE_ARROW) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_PIPE) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("|") reduce 29)) 25981 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) ((":") reduce 59) ((";") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_DOUBLE_ARROW) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_PIPE) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("|") reduce 59)) 25982 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) ((":") reduce 60) ((";") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_DOUBLE_ARROW) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_PIPE) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("|") reduce 60)) 25983 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) ((":") reduce 11) ((";") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_DOUBLE_ARROW) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_PIPE) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("|") reduce 11)) 25984 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) ((":") reduce 42) ((";") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_DOUBLE_ARROW) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_PIPE) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("|") reduce 42)) 25985 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) ((":") reduce 39) ((";") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_DOUBLE_ARROW) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_PIPE) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("|") reduce 39)) 25986 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) ((":") reduce 25) ((";") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_DOUBLE_ARROW) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_PIPE) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("|") reduce 25)) 25987 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) ((":") reduce 50) ((";") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_DOUBLE_ARROW) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_PIPE) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("|") reduce 50)) 25988 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) ((":") reduce 66) ((";") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_DOUBLE_ARROW) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_PIPE) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("|") reduce 66)) 25989 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) ((":") reduce 18) ((";") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_DOUBLE_ARROW) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_PIPE) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("|") reduce 18)) 25990 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) ((":") reduce 17) ((";") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_DOUBLE_ARROW) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_PIPE) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("|") reduce 17)) 25991 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) ((":") reduce 15) ((";") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_DOUBLE_ARROW) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_PIPE) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("|") reduce 15)) 25992 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) ((":") reduce 14) ((";") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_DOUBLE_ARROW) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_PIPE) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("|") reduce 14)) 25993 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) ((":") reduce 38) ((";") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_DOUBLE_ARROW) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_PIPE) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("|") reduce 38)) 25994 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) ((":") reduce 26) ((";") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_DOUBLE_ARROW) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_PIPE) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("|") reduce 26)) 25995 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) ((":") reduce 22) ((";") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_DOUBLE_ARROW) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_PIPE) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("|") reduce 22)) 25996 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) ((":") reduce 24) ((";") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_DOUBLE_ARROW) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_PIPE) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("|") reduce 24)) 25997 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) ((":") reduce 16) ((";") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_DOUBLE_ARROW) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_PIPE) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("|") reduce 16)) 25998 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) ((":") reduce 48) ((";") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_DOUBLE_ARROW) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_PIPE) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("|") reduce 48)) 25999 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) ((":") reduce 20) ((";") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_DOUBLE_ARROW) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_PIPE) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("|") reduce 20)) 26000 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) ((":") reduce 70) ((";") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_DOUBLE_ARROW) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_PIPE) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("|") reduce 70)) 26001 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) ((":") reduce 3) ((";") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_DOUBLE_ARROW) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_PIPE) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("|") reduce 3)) 26002 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) ((":") reduce 12) ((";") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_DOUBLE_ARROW) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_PIPE) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("|") reduce 12)) 26003 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) ((":") reduce 54) ((";") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_DOUBLE_ARROW) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_PIPE) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("|") reduce 54)) 26004 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) ((":") reduce 65) ((";") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_DOUBLE_ARROW) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_PIPE) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("|") reduce 65)) 26005 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) ((":") reduce 75) ((";") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_DOUBLE_ARROW) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_PIPE) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("|") reduce 75)) 26006 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) ((":") reduce 30) ((";") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_DOUBLE_ARROW) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_PIPE) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("|") reduce 30)) 26007 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) ((":") reduce 68) ((";") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_DOUBLE_ARROW) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_PIPE) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("|") reduce 68)) 26008 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) ((":") reduce 21) ((";") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_DOUBLE_ARROW) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_PIPE) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("|") reduce 21)) 26009 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) ((":") reduce 23) ((";") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_DOUBLE_ARROW) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_PIPE) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("|") reduce 23)) 26010 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) ((":") reduce 41) ((";") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_DOUBLE_ARROW) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_PIPE) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("|") reduce 41)) 26011 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) ((":") reduce 62) ((";") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_DOUBLE_ARROW) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_PIPE) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("|") reduce 62)) 26012 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) ((":") reduce 34) ((";") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_DOUBLE_ARROW) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_PIPE) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("|") reduce 34)) 26013 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) ((":") reduce 40) ((";") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_DOUBLE_ARROW) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_PIPE) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("|") reduce 40)) 26014 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) ((":") reduce 13) ((";") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_DOUBLE_ARROW) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_PIPE) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("|") reduce 13)) 26015 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) ((":") reduce 55) ((";") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_DOUBLE_ARROW) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_PIPE) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("|") reduce 55)) 26016 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) ((":") reduce 1) ((";") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_DOUBLE_ARROW) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_PIPE) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("|") reduce 1)) 26017 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) ((":") reduce 2) ((";") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_DOUBLE_ARROW) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_PIPE) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("|") reduce 2)) 26018 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) ((":") reduce 9) ((";") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_DOUBLE_ARROW) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_PIPE) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("|") reduce 9)) 26019 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) ((":") reduce 33) ((";") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_DOUBLE_ARROW) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_PIPE) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("|") reduce 33)) 26020 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) ((":") reduce 58) ((";") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_DOUBLE_ARROW) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_PIPE) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("|") reduce 58)) 26021 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) ((":") reduce 37) ((";") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_DOUBLE_ARROW) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_PIPE) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("|") reduce 37)) 26022 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) ((":") reduce 64) ((";") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_DOUBLE_ARROW) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_PIPE) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("|") reduce 64)) 26023 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) ((":") reduce 46) ((";") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_DOUBLE_ARROW) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_PIPE) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("|") reduce 46)) 26024 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) ((":") reduce 8) ((";") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_DOUBLE_ARROW) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_PIPE) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("|") reduce 8)) 26025 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) ((":") reduce 6) ((";") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_DOUBLE_ARROW) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_PIPE) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("|") reduce 6)) 26026 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) ((":") reduce 7) ((";") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_DOUBLE_ARROW) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_PIPE) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("|") reduce 7)) 26027 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) ((":") reduce 69) ((";") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_DOUBLE_ARROW) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_PIPE) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("|") reduce 69)) 26028 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) ((":") reduce 63) ((";") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_DOUBLE_ARROW) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_PIPE) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("|") reduce 63)) 26029 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) ((":") reduce 56) ((";") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_DOUBLE_ARROW) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_PIPE) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("|") reduce 56)) 26030 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) ((":") reduce 10) ((";") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_DOUBLE_ARROW) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_PIPE) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("|") reduce 10)) 26031 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) ((":") reduce 67) ((";") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_DOUBLE_ARROW) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_PIPE) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("|") reduce 67)) 26032 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) ((":") reduce 44) ((";") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_DOUBLE_ARROW) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_PIPE) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("|") reduce 44)) 26033 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) ((":") reduce 76) ((";") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_DOUBLE_ARROW) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_PIPE) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("|") reduce 76)) 26034 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) ((":") reduce 71) ((";") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_DOUBLE_ARROW) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_PIPE) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("|") reduce 71)) 26035 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) ((":") reduce 77) ((";") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_DOUBLE_ARROW) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_PIPE) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("|") reduce 77)) 26036 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) ((":") reduce 78) ((";") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_DOUBLE_ARROW) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_PIPE) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("|") reduce 78)) 26037 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) ((":") reduce 79) ((";") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_DOUBLE_ARROW) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_PIPE) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("|") reduce 79)) 26038 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) ((":") reduce 4) ((";") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_DOUBLE_ARROW) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_PIPE) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("|") reduce 4)) 26039 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) ((":") reduce 5) ((";") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_DOUBLE_ARROW) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_PIPE) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("|") reduce 5)) 26040 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) ((":") reduce 43) ((";") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_DOUBLE_ARROW) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_PIPE) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("|") reduce 43)) 26041 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) ((":") reduce 73) ((";") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_DOUBLE_ARROW) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_PIPE) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("|") reduce 73)) 26042 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) ((":") reduce 82) ((";") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_DOUBLE_ARROW) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_PIPE) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("|") reduce 82)) 26043 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) ((":") reduce 47) ((";") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_DOUBLE_ARROW) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_PIPE) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("|") reduce 47)) 26044 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) ((":") reduce 31) ((";") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_DOUBLE_ARROW) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_PIPE) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("|") reduce 31)) 26045 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) ((":") reduce 57) ((";") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_DOUBLE_ARROW) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_PIPE) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("|") reduce 57)) 26046 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) ((":") reduce 61) ((";") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_DOUBLE_ARROW) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_PIPE) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("|") reduce 61)) 26047 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) ((":") reduce 28) ((";") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_DOUBLE_ARROW) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_PIPE) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("|") reduce 28)) 26048 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) ((":") reduce 36) ((";") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_DOUBLE_ARROW) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_PIPE) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("|") reduce 36)) 26049 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) ((":") reduce 32) ((";") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_DOUBLE_ARROW) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_PIPE) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("|") reduce 32)) 26050 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) ((":") reduce 35) ((";") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_DOUBLE_ARROW) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_PIPE) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("|") reduce 35)) 26051 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) ((":") reduce 19) ((";") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_DOUBLE_ARROW) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_PIPE) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("|") reduce 19)) 26052 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) ((":") reduce 45) ((";") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_DOUBLE_ARROW) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_PIPE) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("|") reduce 45)) 26053 ((("%") reduce 542) (("(") reduce 579) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) ((":") reduce 542) ((";") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PIPE) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 26055 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) ((":") reduce 72) ((";") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_DOUBLE_ARROW) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_PIPE) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("|") reduce 72)) 26056 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) ((":") reduce 83) ((";") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_DOUBLE_ARROW) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_PIPE) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("|") reduce 83)) 26057 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) ((":") reduce 572) ((";") reduce 572) (("<") reduce 572) (("=") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AND_EQUAL) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_COALESCE_EQUAL) reduce 572) ((T_CONCAT_EQUAL) reduce 572) ((T_DEC) reduce 572) ((T_DIV_EQUAL) reduce 572) ((T_DOUBLE_ARROW) reduce 572) ((T_INC) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_MINUS_EQUAL) reduce 572) ((T_MOD_EQUAL) reduce 572) ((T_MUL_EQUAL) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_OR_EQUAL) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_PLUS_EQUAL) reduce 572) ((T_POW) reduce 572) ((T_POW_EQUAL) reduce 572) ((T_SL) reduce 572) ((T_SL_EQUAL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) ((T_SR_EQUAL) reduce 572) ((T_XOR_EQUAL) reduce 572) (("[") reduce 572) (("^") reduce 572) (("|") reduce 572)) 26060 ((("%") reduce 544) (("(") reduce 580) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) ((":") reduce 544) ((";") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PIPE) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 26062 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((":") reduce 507) ((";") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("|") reduce 507)) 26063 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((":") reduce 291) ((";") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 26067 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) ((":") reduce 292) ((";") reduce 292) (("<") reduce 292) (("=") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AND_EQUAL) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_COALESCE_EQUAL) reduce 292) ((T_CONCAT_EQUAL) reduce 292) ((T_DEC) reduce 292) ((T_DIV_EQUAL) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INC) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_MINUS_EQUAL) reduce 292) ((T_MOD_EQUAL) reduce 292) ((T_MUL_EQUAL) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_OR_EQUAL) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_PLUS_EQUAL) reduce 292) ((T_POW) reduce 292) ((T_POW_EQUAL) reduce 292) ((T_SL) reduce 292) ((T_SL_EQUAL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) ((T_SR_EQUAL) reduce 292) ((T_XOR_EQUAL) reduce 292) (("[") reduce 292) (("^") reduce 292) (("|") reduce 292)) 26068 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) ((":") reduce 293) ((";") reduce 293) (("<") reduce 293) (("=") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AND_EQUAL) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_COALESCE_EQUAL) reduce 293) ((T_CONCAT_EQUAL) reduce 293) ((T_DEC) reduce 293) ((T_DIV_EQUAL) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INC) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_MINUS_EQUAL) reduce 293) ((T_MOD_EQUAL) reduce 293) ((T_MUL_EQUAL) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_OR_EQUAL) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_PLUS_EQUAL) reduce 293) ((T_POW) reduce 293) ((T_POW_EQUAL) reduce 293) ((T_SL) reduce 293) ((T_SL_EQUAL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) ((T_SR_EQUAL) reduce 293) ((T_XOR_EQUAL) reduce 293) (("[") reduce 293) (("^") reduce 293) (("|") reduce 293)) 26073 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) ((":") reduce 425) ((";") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_DOUBLE_ARROW) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_PIPE) reduce 425) ((T_POW) reduce 425) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 26075 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) ((":") reduce 423) ((";") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_DOUBLE_ARROW) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_PIPE) reduce 423) ((T_POW) reduce 423) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 26085 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 419) ((";") reduce 419) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26086 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 421) ((";") reduce 421) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26087 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 420) ((";") reduce 420) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26088 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 413) ((";") reduce 413) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 413) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26089 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 410) ((";") reduce 410) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 410) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26090 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 418) ((";") reduce 418) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26091 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 412) ((";") reduce 412) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26092 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 416) ((";") reduce 416) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26093 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 411) ((";") reduce 411) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26094 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 414) ((";") reduce 414) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 414) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26095 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 415) ((";") reduce 415) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 415) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26096 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 422) ((";") reduce 422) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26097 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 417) ((";") reduce 417) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26099 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 406) ((";") reduce 406) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26102 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) ((":") reduce 568) ((";") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 26105 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) ((":") reduce 564) ((";") reduce 564) (("<") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_DOUBLE_ARROW) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_POW) reduce 564) ((T_SL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) (("[") reduce 564) (("^") reduce 564) (("|") reduce 564)) 26107 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) ((":") reduce 563) ((";") reduce 563) (("<") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_DOUBLE_ARROW) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_POW) reduce 563) ((T_SL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) (("[") reduce 563) (("^") reduce 563) (("|") reduce 563)) 26109 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) ((":") reduce 559) ((";") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 26110 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((":") reduce 565) ((";") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 26111 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) ((":") reduce 407) ((";") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_DOUBLE_ARROW) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("[") reduce 548) (("^") reduce 407) (("|") reduce 407)) 26115 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) ((":") reduce 572) ((";") reduce 572) (("<") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_DOUBLE_ARROW) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_POW) reduce 572) ((T_SL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) (("[") reduce 572) (("^") reduce 572) (("|") reduce 572)) 26117 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((":") reduce 507) ((";") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("|") reduce 507)) 26118 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((":") reduce 291) ((";") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 26122 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) ((":") reduce 292) ((";") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("[") reduce 292) (("^") reduce 292) (("|") reduce 292)) 26123 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) ((":") reduce 293) ((";") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("[") reduce 293) (("^") reduce 293) (("|") reduce 293)) 26124 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) ((":") reduce 504) ((";") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 26127 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) ((":") reduce 571) ((";") reduce 571) (("<") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_DOUBLE_ARROW) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_POW) reduce 571) ((T_SL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) (("[") reduce 571) (("^") reduce 571) (("|") reduce 571)) 26128 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((":") reduce 506) ((";") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("|") reduce 506)) 26129 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((":") reduce 508) ((";") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("|") reduce 508)) 26134 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) ((":") reduce 560) ((";") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 26135 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) ((":") reduce 582) ((";") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 26136 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((":") reduce 566) ((";") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 26137 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) ((":") reduce 584) ((";") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 26140 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) ((":") reduce 583) ((";") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 26141 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((":") reduce 561) ((";") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 26142 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((":") reduce 567) ((";") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 26143 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) ((":") reduce 562) ((";") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("[") reduce 562) (("^") reduce 562) (("|") reduce 562)) 26144 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((":") reduce 505) ((";") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 26145 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) ((":") reduce 570) ((";") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 26148 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) ((":") reduce 569) ((";") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 26149 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) ((":") reduce 504) ((";") reduce 504) (("<") reduce 504) (("=") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AND_EQUAL) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_COALESCE_EQUAL) reduce 504) ((T_CONCAT_EQUAL) reduce 504) ((T_DEC) reduce 504) ((T_DIV_EQUAL) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INC) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_MINUS_EQUAL) reduce 504) ((T_MOD_EQUAL) reduce 504) ((T_MUL_EQUAL) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_OR_EQUAL) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_PLUS_EQUAL) reduce 504) ((T_POW) reduce 504) ((T_POW_EQUAL) reduce 504) ((T_SL) reduce 504) ((T_SL_EQUAL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) ((T_SR_EQUAL) reduce 504) ((T_XOR_EQUAL) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 26160 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) ((":") reduce 494) ((";") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PIPE) reduce 494) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 26161 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) ((":") reduce 489) ((";") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_PIPE) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("^") reduce 489) (("|") reduce 489)) 26171 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 494) ((";") reduce 494) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26172 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) ((":") reduce 490) ((";") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_PIPE) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("^") reduce 490) (("|") reduce 490)) 26174 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 481) ((";") reduce 481) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26176 ((("%") reduce 541) (("(") reduce 579) (("*") reduce 541) (("+") reduce 541) (("-") reduce 541) ((".") reduce 541) (("/") reduce 541) ((":") reduce 541) ((";") reduce 541) (("<") reduce 541) ((">") reduce 541) (("?") reduce 541) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_BOOLEAN_AND) reduce 541) ((T_BOOLEAN_OR) reduce 541) ((T_COALESCE) reduce 541) ((T_DOUBLE_ARROW) reduce 541) ((T_INSTANCEOF) reduce 541) ((T_IS_EQUAL) reduce 541) ((T_IS_GREATER_OR_EQUAL) reduce 541) ((T_IS_IDENTICAL) reduce 541) ((T_IS_NOT_EQUAL) reduce 541) ((T_IS_NOT_IDENTICAL) reduce 541) ((T_IS_SMALLER_OR_EQUAL) reduce 541) ((T_LOGICAL_AND) reduce 541) ((T_LOGICAL_OR) reduce 541) ((T_LOGICAL_XOR) reduce 541) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 541) ((T_OBJECT_OPERATOR) reduce 541) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 541) ((T_PIPE) reduce 541) ((T_POW) reduce 541) ((T_SL) reduce 541) ((T_SPACESHIP) reduce 541) ((T_SR) reduce 541) (("[") reduce 541) (("^") reduce 541) (("|") reduce 541)) 26178 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) ((":") reduce 571) ((";") reduce 571) (("<") reduce 571) (("=") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AND_EQUAL) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_COALESCE_EQUAL) reduce 571) ((T_CONCAT_EQUAL) reduce 571) ((T_DEC) reduce 571) ((T_DIV_EQUAL) reduce 571) ((T_DOUBLE_ARROW) reduce 571) ((T_INC) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_MINUS_EQUAL) reduce 571) ((T_MOD_EQUAL) reduce 571) ((T_MUL_EQUAL) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_OR_EQUAL) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_PLUS_EQUAL) reduce 571) ((T_POW) reduce 571) ((T_POW_EQUAL) reduce 571) ((T_SL) reduce 571) ((T_SL_EQUAL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) ((T_SR_EQUAL) reduce 571) ((T_XOR_EQUAL) reduce 571) (("[") reduce 571) (("^") reduce 571) (("|") reduce 571)) 26181 ((("%") reduce 543) (("(") reduce 580) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) ((":") reduce 543) ((";") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PIPE) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 26182 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((":") reduce 506) ((";") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("|") reduce 506)) 26183 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((":") reduce 508) ((";") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("|") reduce 508)) 26185 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) ((":") reduce 485) ((";") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_DOUBLE_ARROW) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_PIPE) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("^") reduce 485) (("|") reduce 485)) 26186 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) ((":") reduce 487) ((";") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_PIPE) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("^") reduce 487) (("|") reduce 487)) 26191 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) ((":") reduce 560) ((";") reduce 560) (("<") reduce 560) (("=") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AND_EQUAL) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_COALESCE_EQUAL) reduce 560) ((T_CONCAT_EQUAL) reduce 560) ((T_DEC) reduce 560) ((T_DIV_EQUAL) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INC) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_MINUS_EQUAL) reduce 560) ((T_MOD_EQUAL) reduce 560) ((T_MUL_EQUAL) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_OR_EQUAL) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_PLUS_EQUAL) reduce 560) ((T_POW) reduce 560) ((T_POW_EQUAL) reduce 560) ((T_SL) reduce 560) ((T_SL_EQUAL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) ((T_SR_EQUAL) reduce 560) ((T_XOR_EQUAL) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 26192 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) ((":") reduce 582) ((";") reduce 582) (("<") reduce 582) (("=") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AND_EQUAL) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_COALESCE_EQUAL) reduce 582) ((T_CONCAT_EQUAL) reduce 582) ((T_DEC) reduce 582) ((T_DIV_EQUAL) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INC) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_MINUS_EQUAL) reduce 582) ((T_MOD_EQUAL) reduce 582) ((T_MUL_EQUAL) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_OR_EQUAL) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_PLUS_EQUAL) reduce 582) ((T_POW) reduce 582) ((T_POW_EQUAL) reduce 582) ((T_SL) reduce 582) ((T_SL_EQUAL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) ((T_SR_EQUAL) reduce 582) ((T_XOR_EQUAL) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 26193 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((":") reduce 566) ((";") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 26194 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) ((":") reduce 584) ((";") reduce 584) (("<") reduce 584) (("=") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AND_EQUAL) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_COALESCE_EQUAL) reduce 584) ((T_CONCAT_EQUAL) reduce 584) ((T_DEC) reduce 584) ((T_DIV_EQUAL) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INC) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_MINUS_EQUAL) reduce 584) ((T_MOD_EQUAL) reduce 584) ((T_MUL_EQUAL) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_OR_EQUAL) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_PLUS_EQUAL) reduce 584) ((T_POW) reduce 584) ((T_POW_EQUAL) reduce 584) ((T_SL) reduce 584) ((T_SL_EQUAL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) ((T_SR_EQUAL) reduce 584) ((T_XOR_EQUAL) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 26197 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) ((":") reduce 583) ((";") reduce 583) (("<") reduce 583) (("=") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AND_EQUAL) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_COALESCE_EQUAL) reduce 583) ((T_CONCAT_EQUAL) reduce 583) ((T_DEC) reduce 583) ((T_DIV_EQUAL) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INC) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_MINUS_EQUAL) reduce 583) ((T_MOD_EQUAL) reduce 583) ((T_MUL_EQUAL) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_OR_EQUAL) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_PLUS_EQUAL) reduce 583) ((T_POW) reduce 583) ((T_POW_EQUAL) reduce 583) ((T_SL) reduce 583) ((T_SL_EQUAL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) ((T_SR_EQUAL) reduce 583) ((T_XOR_EQUAL) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 26198 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((":") reduce 561) ((";") reduce 561) (("<") reduce 561) (("=") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AND_EQUAL) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_COALESCE_EQUAL) reduce 561) ((T_CONCAT_EQUAL) reduce 561) ((T_DEC) reduce 561) ((T_DIV_EQUAL) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INC) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_MINUS_EQUAL) reduce 561) ((T_MOD_EQUAL) reduce 561) ((T_MUL_EQUAL) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_OR_EQUAL) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_PLUS_EQUAL) reduce 561) ((T_POW) reduce 561) ((T_POW_EQUAL) reduce 561) ((T_SL) reduce 561) ((T_SL_EQUAL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) ((T_SR_EQUAL) reduce 561) ((T_XOR_EQUAL) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 26199 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((":") reduce 567) ((";") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 26200 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) ((":") reduce 562) ((";") reduce 562) (("<") reduce 562) (("=") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AND_EQUAL) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_COALESCE_EQUAL) reduce 562) ((T_CONCAT_EQUAL) reduce 562) ((T_DEC) reduce 562) ((T_DIV_EQUAL) reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_INC) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_MINUS_EQUAL) reduce 562) ((T_MOD_EQUAL) reduce 562) ((T_MUL_EQUAL) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_OR_EQUAL) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_PLUS_EQUAL) reduce 562) ((T_POW) reduce 562) ((T_POW_EQUAL) reduce 562) ((T_SL) reduce 562) ((T_SL_EQUAL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) ((T_SR_EQUAL) reduce 562) ((T_XOR_EQUAL) reduce 562) (("[") reduce 562) (("^") reduce 562) (("|") reduce 562)) 26202 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) ((":") reduce 477) ((";") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_PIPE) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 26204 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) ((":") reduce 520) ((";") reduce 520) (("<") reduce 520) (("=") shift) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_PIPE) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("|") reduce 520)) 26206 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 405) ((";") reduce 405) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26207 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 482) ((";") reduce 482) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 482) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26210 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 481) ((";") reduce 481) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 481) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26211 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) ((":") reduce 473) ((";") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_DOUBLE_ARROW) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_PIPE) reduce 473) ((T_POW) shift) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("|") reduce 473)) 26212 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 483) ((";") reduce 483) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 483) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26213 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) ((":") reduce 469) ((";") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_DOUBLE_ARROW) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_PIPE) reduce 469) ((T_POW) shift) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 26214 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) ((":") reduce 486) ((";") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_DOUBLE_ARROW) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_PIPE) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("^") reduce 486) (("|") reduce 486)) 26216 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) ((":") reduce 526) ((";") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_PIPE) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("^") reduce 526) (("|") reduce 526)) 26218 ((("%") reduce 527) (("*") reduce 527) (("+") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) ((":") reduce 527) ((";") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_DOUBLE_ARROW) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_PIPE) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("^") reduce 527) (("|") reduce 527)) 26219 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) ((":") reduce 525) ((";") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_PIPE) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("^") reduce 525) (("|") reduce 525)) 26220 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 619) ((";") reduce 619) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 619) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26221 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 618) ((";") reduce 618) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 618) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26222 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((":") reduce 505) ((";") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 26223 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 478) ((";") reduce 478) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 478) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26224 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) ((":") reduce 471) ((";") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_DOUBLE_ARROW) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_PIPE) reduce 471) ((T_POW) shift) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("|") reduce 471)) 26226 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((":") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 26227 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((":") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 26228 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((":") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 26229 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((":") reduce 509) ((";") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 26230 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((":") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 26231 ((("%") reduce 400) (("(") reduce 400) (("*") reduce 400) (("+") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) ((":") reduce 400) ((";") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_DOUBLE_ARROW) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 400) ((T_OBJECT_OPERATOR) reduce 400) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 400) ((T_PIPE) reduce 400) ((T_POW) reduce 400) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("[") reduce 400) (("^") reduce 400) (("|") reduce 400)) 26234 ((("%") reduce 511) (("(") reduce 511) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((":") reduce 511) ((";") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("|") reduce 511)) 26235 ((("%") reduce 402) (("(") shift) (("*") reduce 402) (("+") reduce 402) (("-") reduce 402) ((".") reduce 402) (("/") reduce 402) ((":") reduce 402) ((";") reduce 402) (("<") reduce 402) ((">") reduce 402) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_DOUBLE_ARROW) reduce 402) ((T_INSTANCEOF) reduce 402) ((T_IS_EQUAL) reduce 402) ((T_IS_GREATER_OR_EQUAL) reduce 402) ((T_IS_IDENTICAL) reduce 402) ((T_IS_NOT_EQUAL) reduce 402) ((T_IS_NOT_IDENTICAL) reduce 402) ((T_IS_SMALLER_OR_EQUAL) reduce 402) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_PIPE) reduce 402) ((T_POW) reduce 402) ((T_SL) reduce 402) ((T_SPACESHIP) reduce 402) ((T_SR) reduce 402) (("^") reduce 402) (("|") reduce 402)) 26236 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((":") reduce 510) ((";") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 26237 ((("%") reduce 512) (("(") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((":") reduce 512) ((";") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("|") reduce 512)) 26238 ((("%") reduce 573) (("(") reduce 573) (("*") reduce 573) (("+") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) ((":") reduce 573) ((";") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_DOUBLE_ARROW) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("|") reduce 573)) 26244 ((("%") reduce 574) (("(") reduce 574) (("*") reduce 574) (("+") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) ((":") reduce 574) ((";") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_DOUBLE_ARROW) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("|") reduce 574)) 26245 ((("%") reduce 578) (("(") reduce 578) (("*") reduce 578) (("+") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) ((":") reduce 578) ((";") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_DOUBLE_ARROW) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("|") reduce 578)) 26246 ((("%") reduce 575) (("(") reduce 575) (("*") reduce 575) (("+") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) ((":") reduce 575) ((";") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_DOUBLE_ARROW) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("|") reduce 575)) 26247 ((("%") reduce 576) (("(") reduce 576) (("*") reduce 576) (("+") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) ((":") reduce 576) ((";") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_DOUBLE_ARROW) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("|") reduce 576)) 26248 ((("%") reduce 399) (("(") reduce 399) (("*") reduce 399) (("+") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) ((":") reduce 399) ((";") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_DOUBLE_ARROW) reduce 399) ((T_INSTANCEOF) reduce 399) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 399) ((T_OBJECT_OPERATOR) reduce 399) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 399) ((T_PIPE) reduce 399) ((T_POW) reduce 399) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("[") reduce 399) (("^") reduce 399) (("|") reduce 399)) 26250 ((("%") reduce 577) (("(") reduce 577) (("*") reduce 577) (("+") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) ((":") reduce 577) ((";") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_DOUBLE_ARROW) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("|") reduce 577)) 26251 ((("%") reduce 401) (("(") reduce 401) (("*") reduce 401) (("+") reduce 401) (("-") reduce 401) ((".") reduce 401) (("/") reduce 401) ((":") reduce 401) ((";") reduce 401) (("<") reduce 401) ((">") reduce 401) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_DOUBLE_ARROW) reduce 401) ((T_INSTANCEOF) reduce 401) ((T_IS_EQUAL) reduce 401) ((T_IS_GREATER_OR_EQUAL) reduce 401) ((T_IS_IDENTICAL) reduce 401) ((T_IS_NOT_EQUAL) reduce 401) ((T_IS_NOT_IDENTICAL) reduce 401) ((T_IS_SMALLER_OR_EQUAL) reduce 401) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 401) ((T_OBJECT_OPERATOR) reduce 401) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 401) ((T_PIPE) reduce 401) ((T_POW) reduce 401) ((T_SL) reduce 401) ((T_SPACESHIP) reduce 401) ((T_SR) reduce 401) (("[") reduce 401) (("^") reduce 401) (("|") reduce 401)) 26259 ((("%") reduce 398) (("(") reduce 398) (("*") reduce 398) (("+") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) ((":") reduce 398) ((";") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_DOUBLE_ARROW) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 398) ((T_OBJECT_OPERATOR) reduce 398) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 398) ((T_PIPE) reduce 398) ((T_POW) reduce 398) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("[") reduce 398) (("^") reduce 398) (("|") reduce 398)) 26261 ((("%") reduce 513) (("(") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((":") reduce 513) ((";") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("|") reduce 513)) 26267 ((("%") reduce 233) (("*") reduce 233) (("+") reduce 233) (("-") reduce 233) ((".") reduce 233) (("/") reduce 233) ((":") reduce 233) ((";") reduce 233) (("<") reduce 233) ((">") reduce 233) (("?") reduce 233) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_BOOLEAN_AND) reduce 233) ((T_BOOLEAN_OR) reduce 233) ((T_COALESCE) reduce 233) ((T_DOUBLE_ARROW) reduce 233) ((T_INSTANCEOF) reduce 233) ((T_IS_EQUAL) reduce 233) ((T_IS_GREATER_OR_EQUAL) reduce 233) ((T_IS_IDENTICAL) reduce 233) ((T_IS_NOT_EQUAL) reduce 233) ((T_IS_NOT_IDENTICAL) reduce 233) ((T_IS_SMALLER_OR_EQUAL) reduce 233) ((T_LOGICAL_AND) reduce 233) ((T_LOGICAL_OR) reduce 233) ((T_LOGICAL_XOR) reduce 233) ((T_PIPE) reduce 233) ((T_POW) reduce 233) ((T_SL) reduce 233) ((T_SPACESHIP) reduce 233) ((T_SR) reduce 233) (("^") reduce 233) (("|") reduce 233)) 26272 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 404) ((";") reduce 404) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26276 ((("%") reduce 613) (("*") reduce 613) (("+") reduce 613) (("-") reduce 613) ((".") reduce 613) (("/") reduce 613) ((":") reduce 613) ((";") reduce 613) (("<") reduce 613) ((">") reduce 613) (("?") reduce 613) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_BOOLEAN_AND) reduce 613) ((T_BOOLEAN_OR) reduce 613) ((T_COALESCE) reduce 613) ((T_DOUBLE_ARROW) reduce 613) ((T_INSTANCEOF) reduce 613) ((T_IS_EQUAL) reduce 613) ((T_IS_GREATER_OR_EQUAL) reduce 613) ((T_IS_IDENTICAL) reduce 613) ((T_IS_NOT_EQUAL) reduce 613) ((T_IS_NOT_IDENTICAL) reduce 613) ((T_IS_SMALLER_OR_EQUAL) reduce 613) ((T_LOGICAL_AND) reduce 613) ((T_LOGICAL_OR) reduce 613) ((T_LOGICAL_XOR) reduce 613) ((T_PIPE) reduce 613) ((T_POW) reduce 613) ((T_SL) reduce 613) ((T_SPACESHIP) reduce 613) ((T_SR) reduce 613) (("^") reduce 613) (("|") reduce 613)) 26277 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) ((":") reduce 467) ((";") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_DOUBLE_ARROW) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_PIPE) reduce 467) ((T_POW) shift) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("|") reduce 467)) 26278 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 616) ((";") reduce 616) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 616) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26279 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 615) ((";") reduce 615) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 615) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26280 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) ((":") reduce 424) ((";") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_DOUBLE_ARROW) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("[") reduce 548) (("^") reduce 424) (("|") reduce 424)) 26282 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) ((":") reduce 518) ((";") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_PIPE) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("^") reduce 518) (("|") reduce 518)) 26283 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) ((":") reduce 474) ((";") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_DOUBLE_ARROW) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_PIPE) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("|") reduce 474)) 26284 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((":") reduce 291) ((";") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("|") reduce 291)) 26288 ((("%") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) ((":") reduce 292) ((";") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("^") reduce 292) (("|") reduce 292)) 26289 ((("%") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) ((":") reduce 293) ((";") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("^") reduce 293) (("|") reduce 293)) 26292 ((("%") reduce 617) (("*") reduce 617) (("+") reduce 617) (("-") reduce 617) ((".") reduce 617) (("/") reduce 617) ((":") reduce 617) ((";") reduce 617) (("<") reduce 617) ((">") reduce 617) (("?") reduce 617) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_BOOLEAN_AND) reduce 617) ((T_BOOLEAN_OR) reduce 617) ((T_COALESCE) reduce 617) ((T_DOUBLE_ARROW) reduce 617) ((T_INSTANCEOF) reduce 617) ((T_IS_EQUAL) reduce 617) ((T_IS_GREATER_OR_EQUAL) reduce 617) ((T_IS_IDENTICAL) reduce 617) ((T_IS_NOT_EQUAL) reduce 617) ((T_IS_NOT_IDENTICAL) reduce 617) ((T_IS_SMALLER_OR_EQUAL) reduce 617) ((T_LOGICAL_AND) reduce 617) ((T_LOGICAL_OR) reduce 617) ((T_LOGICAL_XOR) reduce 617) ((T_PIPE) reduce 617) ((T_POW) reduce 617) ((T_SL) reduce 617) ((T_SPACESHIP) reduce 617) ((T_SR) reduce 617) (("^") reduce 617) (("|") reduce 617)) 26295 ((("%") reduce 614) (("*") reduce 614) (("+") reduce 614) (("-") reduce 614) ((".") reduce 614) (("/") reduce 614) ((":") reduce 614) ((";") reduce 614) (("<") reduce 614) ((">") reduce 614) (("?") reduce 614) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_BOOLEAN_AND) reduce 614) ((T_BOOLEAN_OR) reduce 614) ((T_COALESCE) reduce 614) ((T_DOUBLE_ARROW) reduce 614) ((T_INSTANCEOF) reduce 614) ((T_IS_EQUAL) reduce 614) ((T_IS_GREATER_OR_EQUAL) reduce 614) ((T_IS_IDENTICAL) reduce 614) ((T_IS_NOT_EQUAL) reduce 614) ((T_IS_NOT_IDENTICAL) reduce 614) ((T_IS_SMALLER_OR_EQUAL) reduce 614) ((T_LOGICAL_AND) reduce 614) ((T_LOGICAL_OR) reduce 614) ((T_LOGICAL_XOR) reduce 614) ((T_PIPE) reduce 614) ((T_POW) reduce 614) ((T_SL) reduce 614) ((T_SPACESHIP) reduce 614) ((T_SR) reduce 614) (("^") reduce 614) (("|") reduce 614)) 26296 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) ((":") reduce 468) ((";") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_DOUBLE_ARROW) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_PIPE) reduce 468) ((T_POW) shift) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 26297 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) ((":") reduce 426) ((";") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_DOUBLE_ARROW) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("[") reduce 548) (("^") reduce 426) (("|") reduce 426)) 26299 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) ((":") reduce 408) ((";") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_DOUBLE_ARROW) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_PIPE) reduce 408) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("^") reduce 408) (("|") reduce 408)) 26300 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) ((":") reduce 409) ((";") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_DOUBLE_ARROW) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_PIPE) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("|") reduce 409)) 26301 ((("%") reduce 296) (("*") reduce 296) (("+") reduce 296) (("-") reduce 296) ((".") reduce 296) (("/") reduce 296) ((":") reduce 296) ((";") reduce 296) (("<") reduce 296) ((">") reduce 296) (("?") reduce 296) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_BOOLEAN_AND) reduce 296) ((T_BOOLEAN_OR) reduce 296) ((T_COALESCE) reduce 296) ((T_DOUBLE_ARROW) reduce 296) ((T_INSTANCEOF) reduce 296) ((T_IS_EQUAL) reduce 296) ((T_IS_GREATER_OR_EQUAL) reduce 296) ((T_IS_IDENTICAL) reduce 296) ((T_IS_NOT_EQUAL) reduce 296) ((T_IS_NOT_IDENTICAL) reduce 296) ((T_IS_SMALLER_OR_EQUAL) reduce 296) ((T_LOGICAL_AND) reduce 296) ((T_LOGICAL_OR) reduce 296) ((T_LOGICAL_XOR) reduce 296) ((T_PIPE) reduce 296) ((T_POW) reduce 296) ((T_SL) reduce 296) ((T_SPACESHIP) reduce 296) ((T_SR) reduce 296) (("^") reduce 296) (("|") reduce 296)) 26306 ((("%") reduce 297) (("*") reduce 297) (("+") reduce 297) (("-") reduce 297) ((".") reduce 297) (("/") reduce 297) ((":") reduce 297) ((";") reduce 297) (("<") reduce 297) ((">") reduce 297) (("?") reduce 297) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_BOOLEAN_AND) reduce 297) ((T_BOOLEAN_OR) reduce 297) ((T_COALESCE) reduce 297) ((T_DOUBLE_ARROW) reduce 297) ((T_INSTANCEOF) reduce 297) ((T_IS_EQUAL) reduce 297) ((T_IS_GREATER_OR_EQUAL) reduce 297) ((T_IS_IDENTICAL) reduce 297) ((T_IS_NOT_EQUAL) reduce 297) ((T_IS_NOT_IDENTICAL) reduce 297) ((T_IS_SMALLER_OR_EQUAL) reduce 297) ((T_LOGICAL_AND) reduce 297) ((T_LOGICAL_OR) reduce 297) ((T_LOGICAL_XOR) reduce 297) ((T_PIPE) reduce 297) ((T_POW) reduce 297) ((T_SL) reduce 297) ((T_SPACESHIP) reduce 297) ((T_SR) reduce 297) (("^") reduce 297) (("|") reduce 297)) 26307 ((("%") reduce 460) (("(") reduce 556) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) ((":") reduce 460) ((";") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PIPE) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("[") reduce 549) (("^") reduce 460) (("|") reduce 460)) 26309 ((("%") reduce 298) (("*") reduce 298) (("+") reduce 298) (("-") reduce 298) ((".") reduce 298) (("/") reduce 298) ((":") reduce 298) ((";") reduce 298) (("<") reduce 298) ((">") reduce 298) (("?") reduce 298) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_BOOLEAN_AND) reduce 298) ((T_BOOLEAN_OR) reduce 298) ((T_COALESCE) reduce 298) ((T_DOUBLE_ARROW) reduce 298) ((T_INSTANCEOF) reduce 298) ((T_IS_EQUAL) reduce 298) ((T_IS_GREATER_OR_EQUAL) reduce 298) ((T_IS_IDENTICAL) reduce 298) ((T_IS_NOT_EQUAL) reduce 298) ((T_IS_NOT_IDENTICAL) reduce 298) ((T_IS_SMALLER_OR_EQUAL) reduce 298) ((T_LOGICAL_AND) reduce 298) ((T_LOGICAL_OR) reduce 298) ((T_LOGICAL_XOR) reduce 298) ((T_PIPE) reduce 298) ((T_POW) reduce 298) ((T_SL) reduce 298) ((T_SPACESHIP) reduce 298) ((T_SR) reduce 298) (("^") reduce 298) (("|") reduce 298)) 26310 ((("%") reduce 299) (("*") reduce 299) (("+") reduce 299) (("-") reduce 299) ((".") reduce 299) (("/") reduce 299) ((":") reduce 299) ((";") reduce 299) (("<") reduce 299) ((">") reduce 299) (("?") reduce 299) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_BOOLEAN_AND) reduce 299) ((T_BOOLEAN_OR) reduce 299) ((T_COALESCE) reduce 299) ((T_DOUBLE_ARROW) reduce 299) ((T_INSTANCEOF) reduce 299) ((T_IS_EQUAL) reduce 299) ((T_IS_GREATER_OR_EQUAL) reduce 299) ((T_IS_IDENTICAL) reduce 299) ((T_IS_NOT_EQUAL) reduce 299) ((T_IS_NOT_IDENTICAL) reduce 299) ((T_IS_SMALLER_OR_EQUAL) reduce 299) ((T_LOGICAL_AND) reduce 299) ((T_LOGICAL_OR) reduce 299) ((T_LOGICAL_XOR) reduce 299) ((T_PIPE) reduce 299) ((T_POW) reduce 299) ((T_SL) reduce 299) ((T_SPACESHIP) reduce 299) ((T_SR) reduce 299) (("^") reduce 299) (("|") reduce 299)) 26311 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) ((":") reduce 472) ((";") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_DOUBLE_ARROW) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_PIPE) reduce 472) ((T_POW) shift) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("|") reduce 472)) 26312 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) ((":") reduce 470) ((";") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_DOUBLE_ARROW) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_PIPE) reduce 470) ((T_POW) shift) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("|") reduce 470)) 26315 ((("%") reduce 519) (("(") reduce 519) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) ((":") reduce 519) ((";") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PIPE) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("|") reduce 519)) 26316 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) ((":") reduce 475) ((";") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_DOUBLE_ARROW) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_PIPE) reduce 475) ((T_POW) shift) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("|") reduce 475)) 26317 ((("%") reduce 446) (("*") reduce 446) (("+") reduce 446) (("-") reduce 446) ((".") reduce 446) (("/") reduce 446) ((":") reduce 446) ((";") reduce 446) (("<") reduce 446) ((">") reduce 446) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_BOOLEAN_AND) reduce 446) ((T_BOOLEAN_OR) reduce 446) ((T_COALESCE) reduce 446) ((T_DOUBLE_ARROW) reduce 446) ((T_INSTANCEOF) reduce 446) ((T_IS_EQUAL) reduce 446) ((T_IS_GREATER_OR_EQUAL) reduce 446) ((T_IS_IDENTICAL) reduce 446) ((T_IS_NOT_EQUAL) reduce 446) ((T_IS_NOT_IDENTICAL) reduce 446) ((T_IS_SMALLER_OR_EQUAL) reduce 446) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_PIPE) reduce 446) ((T_POW) shift) ((T_SL) reduce 446) ((T_SPACESHIP) reduce 446) ((T_SR) reduce 446) (("^") reduce 446) (("|") reduce 446)) 26318 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) ((":") reduce 445) ((";") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_DOUBLE_ARROW) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_PIPE) reduce 445) ((T_POW) shift) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("|") reduce 445)) 26320 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) ((":") reduce 570) ((";") reduce 570) (("<") reduce 570) (("=") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AND_EQUAL) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_COALESCE_EQUAL) reduce 570) ((T_CONCAT_EQUAL) reduce 570) ((T_DEC) reduce 570) ((T_DIV_EQUAL) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INC) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_MINUS_EQUAL) reduce 570) ((T_MOD_EQUAL) reduce 570) ((T_MUL_EQUAL) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_OR_EQUAL) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_PLUS_EQUAL) reduce 570) ((T_POW) reduce 570) ((T_POW_EQUAL) reduce 570) ((T_SL) reduce 570) ((T_SL_EQUAL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) ((T_SR_EQUAL) reduce 570) ((T_XOR_EQUAL) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 26323 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) ((":") reduce 569) ((";") reduce 569) (("<") reduce 569) (("=") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AND_EQUAL) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_COALESCE_EQUAL) reduce 569) ((T_CONCAT_EQUAL) reduce 569) ((T_DEC) reduce 569) ((T_DIV_EQUAL) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INC) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_MINUS_EQUAL) reduce 569) ((T_MOD_EQUAL) reduce 569) ((T_MUL_EQUAL) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_OR_EQUAL) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_PLUS_EQUAL) reduce 569) ((T_POW) reduce 569) ((T_POW_EQUAL) reduce 569) ((T_SL) reduce 569) ((T_SL_EQUAL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) ((T_SR_EQUAL) reduce 569) ((T_XOR_EQUAL) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 26325 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) ((":") reduce 522) ((";") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_PIPE) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("|") reduce 522)) 26326 ((("%") reduce 447) (("*") reduce 447) (("+") reduce 447) (("-") reduce 447) ((".") reduce 447) (("/") reduce 447) ((":") reduce 447) ((";") reduce 447) (("<") reduce 447) ((">") reduce 447) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_BOOLEAN_AND) reduce 447) ((T_BOOLEAN_OR) reduce 447) ((T_COALESCE) reduce 447) ((T_DOUBLE_ARROW) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 447) ((T_IS_GREATER_OR_EQUAL) reduce 447) ((T_IS_IDENTICAL) reduce 447) ((T_IS_NOT_EQUAL) reduce 447) ((T_IS_NOT_IDENTICAL) reduce 447) ((T_IS_SMALLER_OR_EQUAL) reduce 447) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_PIPE) reduce 447) ((T_POW) shift) ((T_SL) reduce 447) ((T_SPACESHIP) reduce 447) ((T_SR) reduce 447) (("^") reduce 447) (("|") reduce 447)) 26327 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) ((":") reduce 473) ((";") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_PIPE) reduce 473) ((T_POW) shift) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("|") reduce 473)) 26328 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 483) ((";") reduce 483) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26329 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) ((":") reduce 469) ((";") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_PIPE) reduce 469) ((T_POW) shift) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 26330 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) ((":") reduce 486) ((";") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_PIPE) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("^") reduce 486) (("|") reduce 486)) 26332 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) ((":") reduce 526) ((";") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_PIPE) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("^") reduce 526) (("|") reduce 526)) 26334 ((("%") reduce 527) (("*") reduce 527) (("+") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) ((":") reduce 527) ((";") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_PIPE) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("^") reduce 527) (("|") reduce 527)) 26335 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) ((":") reduce 525) ((";") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_PIPE) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("^") reduce 525) (("|") reduce 525)) 26336 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 619) ((";") reduce 619) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26337 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 618) ((";") reduce 618) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26338 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((":") reduce 505) ((";") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 26339 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 478) ((";") reduce 478) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26340 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) ((":") reduce 471) ((";") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_PIPE) reduce 471) ((T_POW) shift) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("|") reduce 471)) 26342 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((":") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 26343 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((":") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 26344 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((":") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 26345 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((":") reduce 509) ((";") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 26346 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((":") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 26347 ((("%") reduce 400) (("(") reduce 400) (("*") reduce 400) (("+") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) ((":") reduce 400) ((";") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 400) ((T_OBJECT_OPERATOR) reduce 400) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 400) ((T_PIPE) reduce 400) ((T_POW) reduce 400) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("[") reduce 400) (("^") reduce 400) (("|") reduce 400)) 26350 ((("%") reduce 511) (("(") reduce 511) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((":") reduce 511) ((";") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("|") reduce 511)) 26351 ((("%") reduce 402) (("(") shift) (("*") reduce 402) (("+") reduce 402) (("-") reduce 402) ((".") reduce 402) (("/") reduce 402) ((":") reduce 402) ((";") reduce 402) (("<") reduce 402) ((">") reduce 402) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_INSTANCEOF) reduce 402) ((T_IS_EQUAL) reduce 402) ((T_IS_GREATER_OR_EQUAL) reduce 402) ((T_IS_IDENTICAL) reduce 402) ((T_IS_NOT_EQUAL) reduce 402) ((T_IS_NOT_IDENTICAL) reduce 402) ((T_IS_SMALLER_OR_EQUAL) reduce 402) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_PIPE) reduce 402) ((T_POW) reduce 402) ((T_SL) reduce 402) ((T_SPACESHIP) reduce 402) ((T_SR) reduce 402) (("^") reduce 402) (("|") reduce 402)) 26352 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((":") reduce 510) ((";") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 26353 ((("%") reduce 512) (("(") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((":") reduce 512) ((";") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("|") reduce 512)) 26354 ((("%") reduce 573) (("(") reduce 573) (("*") reduce 573) (("+") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) ((":") reduce 573) ((";") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("|") reduce 573)) 26360 ((("%") reduce 574) (("(") reduce 574) (("*") reduce 574) (("+") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) ((":") reduce 574) ((";") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("|") reduce 574)) 26361 ((("%") reduce 578) (("(") reduce 578) (("*") reduce 578) (("+") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) ((":") reduce 578) ((";") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("|") reduce 578)) 26362 ((("%") reduce 575) (("(") reduce 575) (("*") reduce 575) (("+") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) ((":") reduce 575) ((";") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("|") reduce 575)) 26363 ((("%") reduce 576) (("(") reduce 576) (("*") reduce 576) (("+") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) ((":") reduce 576) ((";") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("|") reduce 576)) 26364 ((("%") reduce 399) (("(") reduce 399) (("*") reduce 399) (("+") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) ((":") reduce 399) ((";") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_INSTANCEOF) reduce 399) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 399) ((T_OBJECT_OPERATOR) reduce 399) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 399) ((T_PIPE) reduce 399) ((T_POW) reduce 399) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("[") reduce 399) (("^") reduce 399) (("|") reduce 399)) 26366 ((("%") reduce 577) (("(") reduce 577) (("*") reduce 577) (("+") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) ((":") reduce 577) ((";") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("|") reduce 577)) 26367 ((("%") reduce 401) (("(") reduce 401) (("*") reduce 401) (("+") reduce 401) (("-") reduce 401) ((".") reduce 401) (("/") reduce 401) ((":") reduce 401) ((";") reduce 401) (("<") reduce 401) ((">") reduce 401) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_INSTANCEOF) reduce 401) ((T_IS_EQUAL) reduce 401) ((T_IS_GREATER_OR_EQUAL) reduce 401) ((T_IS_IDENTICAL) reduce 401) ((T_IS_NOT_EQUAL) reduce 401) ((T_IS_NOT_IDENTICAL) reduce 401) ((T_IS_SMALLER_OR_EQUAL) reduce 401) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 401) ((T_OBJECT_OPERATOR) reduce 401) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 401) ((T_PIPE) reduce 401) ((T_POW) reduce 401) ((T_SL) reduce 401) ((T_SPACESHIP) reduce 401) ((T_SR) reduce 401) (("[") reduce 401) (("^") reduce 401) (("|") reduce 401)) 26375 ((("%") reduce 398) (("(") reduce 398) (("*") reduce 398) (("+") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) ((":") reduce 398) ((";") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 398) ((T_OBJECT_OPERATOR) reduce 398) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 398) ((T_PIPE) reduce 398) ((T_POW) reduce 398) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("[") reduce 398) (("^") reduce 398) (("|") reduce 398)) 26377 ((("%") reduce 513) (("(") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((":") reduce 513) ((";") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("|") reduce 513)) 26383 ((("%") reduce 233) (("*") reduce 233) (("+") reduce 233) (("-") reduce 233) ((".") reduce 233) (("/") reduce 233) ((":") reduce 233) ((";") reduce 233) (("<") reduce 233) ((">") reduce 233) (("?") reduce 233) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_BOOLEAN_AND) reduce 233) ((T_BOOLEAN_OR) reduce 233) ((T_COALESCE) reduce 233) ((T_INSTANCEOF) reduce 233) ((T_IS_EQUAL) reduce 233) ((T_IS_GREATER_OR_EQUAL) reduce 233) ((T_IS_IDENTICAL) reduce 233) ((T_IS_NOT_EQUAL) reduce 233) ((T_IS_NOT_IDENTICAL) reduce 233) ((T_IS_SMALLER_OR_EQUAL) reduce 233) ((T_LOGICAL_AND) reduce 233) ((T_LOGICAL_OR) reduce 233) ((T_LOGICAL_XOR) reduce 233) ((T_PIPE) reduce 233) ((T_POW) reduce 233) ((T_SL) reduce 233) ((T_SPACESHIP) reduce 233) ((T_SR) reduce 233) (("^") reduce 233) (("|") reduce 233)) 26388 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 404) ((";") reduce 404) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26392 ((("%") reduce 613) (("*") reduce 613) (("+") reduce 613) (("-") reduce 613) ((".") reduce 613) (("/") reduce 613) ((":") reduce 613) ((";") reduce 613) (("<") reduce 613) ((">") reduce 613) (("?") reduce 613) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_BOOLEAN_AND) reduce 613) ((T_BOOLEAN_OR) reduce 613) ((T_COALESCE) reduce 613) ((T_INSTANCEOF) reduce 613) ((T_IS_EQUAL) reduce 613) ((T_IS_GREATER_OR_EQUAL) reduce 613) ((T_IS_IDENTICAL) reduce 613) ((T_IS_NOT_EQUAL) reduce 613) ((T_IS_NOT_IDENTICAL) reduce 613) ((T_IS_SMALLER_OR_EQUAL) reduce 613) ((T_LOGICAL_AND) reduce 613) ((T_LOGICAL_OR) reduce 613) ((T_LOGICAL_XOR) reduce 613) ((T_PIPE) reduce 613) ((T_POW) reduce 613) ((T_SL) reduce 613) ((T_SPACESHIP) reduce 613) ((T_SR) reduce 613) (("^") reduce 613) (("|") reduce 613)) 26393 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) ((":") reduce 467) ((";") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_PIPE) reduce 467) ((T_POW) shift) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("|") reduce 467)) 26394 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 616) ((";") reduce 616) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26395 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((":") reduce 615) ((";") reduce 615) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 26396 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) ((":") reduce 424) ((";") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("[") reduce 548) (("^") reduce 424) (("|") reduce 424)) 26398 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) ((":") reduce 518) ((";") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_PIPE) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("^") reduce 518) (("|") reduce 518)) 26399 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) ((":") reduce 474) ((";") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_PIPE) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("|") reduce 474)) 26400 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((":") reduce 291) ((";") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("|") reduce 291)) 26404 ((("%") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) ((":") reduce 292) ((";") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("^") reduce 292) (("|") reduce 292)) 26405 ((("%") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) ((":") reduce 293) ((";") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("^") reduce 293) (("|") reduce 293)) 26408 ((("%") reduce 617) (("*") reduce 617) (("+") reduce 617) (("-") reduce 617) ((".") reduce 617) (("/") reduce 617) ((":") reduce 617) ((";") reduce 617) (("<") reduce 617) ((">") reduce 617) (("?") reduce 617) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_BOOLEAN_AND) reduce 617) ((T_BOOLEAN_OR) reduce 617) ((T_COALESCE) reduce 617) ((T_INSTANCEOF) reduce 617) ((T_IS_EQUAL) reduce 617) ((T_IS_GREATER_OR_EQUAL) reduce 617) ((T_IS_IDENTICAL) reduce 617) ((T_IS_NOT_EQUAL) reduce 617) ((T_IS_NOT_IDENTICAL) reduce 617) ((T_IS_SMALLER_OR_EQUAL) reduce 617) ((T_LOGICAL_AND) reduce 617) ((T_LOGICAL_OR) reduce 617) ((T_LOGICAL_XOR) reduce 617) ((T_PIPE) reduce 617) ((T_POW) reduce 617) ((T_SL) reduce 617) ((T_SPACESHIP) reduce 617) ((T_SR) reduce 617) (("^") reduce 617) (("|") reduce 617)) 26411 ((("%") reduce 614) (("*") reduce 614) (("+") reduce 614) (("-") reduce 614) ((".") reduce 614) (("/") reduce 614) ((":") reduce 614) ((";") reduce 614) (("<") reduce 614) ((">") reduce 614) (("?") reduce 614) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_BOOLEAN_AND) reduce 614) ((T_BOOLEAN_OR) reduce 614) ((T_COALESCE) reduce 614) ((T_INSTANCEOF) reduce 614) ((T_IS_EQUAL) reduce 614) ((T_IS_GREATER_OR_EQUAL) reduce 614) ((T_IS_IDENTICAL) reduce 614) ((T_IS_NOT_EQUAL) reduce 614) ((T_IS_NOT_IDENTICAL) reduce 614) ((T_IS_SMALLER_OR_EQUAL) reduce 614) ((T_LOGICAL_AND) reduce 614) ((T_LOGICAL_OR) reduce 614) ((T_LOGICAL_XOR) reduce 614) ((T_PIPE) reduce 614) ((T_POW) reduce 614) ((T_SL) reduce 614) ((T_SPACESHIP) reduce 614) ((T_SR) reduce 614) (("^") reduce 614) (("|") reduce 614)) 26412 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) ((":") reduce 468) ((";") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_PIPE) reduce 468) ((T_POW) shift) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 26413 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) ((":") reduce 426) ((";") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("[") reduce 548) (("^") reduce 426) (("|") reduce 426)) 26415 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) ((":") reduce 408) ((";") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_PIPE) reduce 408) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("^") reduce 408) (("|") reduce 408)) 26416 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) ((":") reduce 409) ((";") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_PIPE) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("|") reduce 409)) 26417 ((("%") reduce 296) (("*") reduce 296) (("+") reduce 296) (("-") reduce 296) ((".") reduce 296) (("/") reduce 296) ((":") reduce 296) ((";") reduce 296) (("<") reduce 296) ((">") reduce 296) (("?") reduce 296) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_BOOLEAN_AND) reduce 296) ((T_BOOLEAN_OR) reduce 296) ((T_COALESCE) reduce 296) ((T_INSTANCEOF) reduce 296) ((T_IS_EQUAL) reduce 296) ((T_IS_GREATER_OR_EQUAL) reduce 296) ((T_IS_IDENTICAL) reduce 296) ((T_IS_NOT_EQUAL) reduce 296) ((T_IS_NOT_IDENTICAL) reduce 296) ((T_IS_SMALLER_OR_EQUAL) reduce 296) ((T_LOGICAL_AND) reduce 296) ((T_LOGICAL_OR) reduce 296) ((T_LOGICAL_XOR) reduce 296) ((T_PIPE) reduce 296) ((T_POW) reduce 296) ((T_SL) reduce 296) ((T_SPACESHIP) reduce 296) ((T_SR) reduce 296) (("^") reduce 296) (("|") reduce 296)) 26422 ((("%") reduce 297) (("*") reduce 297) (("+") reduce 297) (("-") reduce 297) ((".") reduce 297) (("/") reduce 297) ((":") reduce 297) ((";") reduce 297) (("<") reduce 297) ((">") reduce 297) (("?") reduce 297) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_BOOLEAN_AND) reduce 297) ((T_BOOLEAN_OR) reduce 297) ((T_COALESCE) reduce 297) ((T_INSTANCEOF) reduce 297) ((T_IS_EQUAL) reduce 297) ((T_IS_GREATER_OR_EQUAL) reduce 297) ((T_IS_IDENTICAL) reduce 297) ((T_IS_NOT_EQUAL) reduce 297) ((T_IS_NOT_IDENTICAL) reduce 297) ((T_IS_SMALLER_OR_EQUAL) reduce 297) ((T_LOGICAL_AND) reduce 297) ((T_LOGICAL_OR) reduce 297) ((T_LOGICAL_XOR) reduce 297) ((T_PIPE) reduce 297) ((T_POW) reduce 297) ((T_SL) reduce 297) ((T_SPACESHIP) reduce 297) ((T_SR) reduce 297) (("^") reduce 297) (("|") reduce 297)) 26423 ((("%") reduce 460) (("(") reduce 556) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) ((":") reduce 460) ((";") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PIPE) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("[") reduce 549) (("^") reduce 460) (("|") reduce 460)) 26425 ((("%") reduce 298) (("*") reduce 298) (("+") reduce 298) (("-") reduce 298) ((".") reduce 298) (("/") reduce 298) ((":") reduce 298) ((";") reduce 298) (("<") reduce 298) ((">") reduce 298) (("?") reduce 298) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_BOOLEAN_AND) reduce 298) ((T_BOOLEAN_OR) reduce 298) ((T_COALESCE) reduce 298) ((T_INSTANCEOF) reduce 298) ((T_IS_EQUAL) reduce 298) ((T_IS_GREATER_OR_EQUAL) reduce 298) ((T_IS_IDENTICAL) reduce 298) ((T_IS_NOT_EQUAL) reduce 298) ((T_IS_NOT_IDENTICAL) reduce 298) ((T_IS_SMALLER_OR_EQUAL) reduce 298) ((T_LOGICAL_AND) reduce 298) ((T_LOGICAL_OR) reduce 298) ((T_LOGICAL_XOR) reduce 298) ((T_PIPE) reduce 298) ((T_POW) reduce 298) ((T_SL) reduce 298) ((T_SPACESHIP) reduce 298) ((T_SR) reduce 298) (("^") reduce 298) (("|") reduce 298)) 26426 ((("%") reduce 299) (("*") reduce 299) (("+") reduce 299) (("-") reduce 299) ((".") reduce 299) (("/") reduce 299) ((":") reduce 299) ((";") reduce 299) (("<") reduce 299) ((">") reduce 299) (("?") reduce 299) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_BOOLEAN_AND) reduce 299) ((T_BOOLEAN_OR) reduce 299) ((T_COALESCE) reduce 299) ((T_INSTANCEOF) reduce 299) ((T_IS_EQUAL) reduce 299) ((T_IS_GREATER_OR_EQUAL) reduce 299) ((T_IS_IDENTICAL) reduce 299) ((T_IS_NOT_EQUAL) reduce 299) ((T_IS_NOT_IDENTICAL) reduce 299) ((T_IS_SMALLER_OR_EQUAL) reduce 299) ((T_LOGICAL_AND) reduce 299) ((T_LOGICAL_OR) reduce 299) ((T_LOGICAL_XOR) reduce 299) ((T_PIPE) reduce 299) ((T_POW) reduce 299) ((T_SL) reduce 299) ((T_SPACESHIP) reduce 299) ((T_SR) reduce 299) (("^") reduce 299) (("|") reduce 299)) 26427 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) ((":") reduce 472) ((";") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_PIPE) reduce 472) ((T_POW) shift) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("|") reduce 472)) 26428 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) ((":") reduce 470) ((";") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_PIPE) reduce 470) ((T_POW) shift) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("|") reduce 470)) 26431 ((("%") reduce 519) (("(") reduce 519) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) ((":") reduce 519) ((";") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PIPE) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("|") reduce 519)) 26432 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) ((":") reduce 475) ((";") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_PIPE) reduce 475) ((T_POW) shift) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("|") reduce 475)) 26433 ((("%") reduce 446) (("*") reduce 446) (("+") reduce 446) (("-") reduce 446) ((".") reduce 446) (("/") reduce 446) ((":") reduce 446) ((";") reduce 446) (("<") reduce 446) ((">") reduce 446) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_BOOLEAN_AND) reduce 446) ((T_BOOLEAN_OR) reduce 446) ((T_COALESCE) reduce 446) ((T_INSTANCEOF) reduce 446) ((T_IS_EQUAL) reduce 446) ((T_IS_GREATER_OR_EQUAL) reduce 446) ((T_IS_IDENTICAL) reduce 446) ((T_IS_NOT_EQUAL) reduce 446) ((T_IS_NOT_IDENTICAL) reduce 446) ((T_IS_SMALLER_OR_EQUAL) reduce 446) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_PIPE) reduce 446) ((T_POW) shift) ((T_SL) reduce 446) ((T_SPACESHIP) reduce 446) ((T_SR) reduce 446) (("^") reduce 446) (("|") reduce 446)) 26434 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) ((":") reduce 445) ((";") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_PIPE) reduce 445) ((T_POW) shift) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("|") reduce 445)) 26436 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) ((":") reduce 570) ((";") reduce 570) (("<") reduce 570) (("=") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AND_EQUAL) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_COALESCE_EQUAL) reduce 570) ((T_CONCAT_EQUAL) reduce 570) ((T_DEC) reduce 570) ((T_DIV_EQUAL) reduce 570) ((T_INC) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_MINUS_EQUAL) reduce 570) ((T_MOD_EQUAL) reduce 570) ((T_MUL_EQUAL) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_OR_EQUAL) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_PLUS_EQUAL) reduce 570) ((T_POW) reduce 570) ((T_POW_EQUAL) reduce 570) ((T_SL) reduce 570) ((T_SL_EQUAL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) ((T_SR_EQUAL) reduce 570) ((T_XOR_EQUAL) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 26439 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) ((":") reduce 569) ((";") reduce 569) (("<") reduce 569) (("=") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AND_EQUAL) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_COALESCE_EQUAL) reduce 569) ((T_CONCAT_EQUAL) reduce 569) ((T_DEC) reduce 569) ((T_DIV_EQUAL) reduce 569) ((T_INC) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_MINUS_EQUAL) reduce 569) ((T_MOD_EQUAL) reduce 569) ((T_MUL_EQUAL) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_OR_EQUAL) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_PLUS_EQUAL) reduce 569) ((T_POW) reduce 569) ((T_POW_EQUAL) reduce 569) ((T_SL) reduce 569) ((T_SL_EQUAL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) ((T_SR_EQUAL) reduce 569) ((T_XOR_EQUAL) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 26441 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) ((":") reduce 522) ((";") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_PIPE) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("|") reduce 522)) 26442 ((("%") reduce 447) (("*") reduce 447) (("+") reduce 447) (("-") reduce 447) ((".") reduce 447) (("/") reduce 447) ((":") reduce 447) ((";") reduce 447) (("<") reduce 447) ((">") reduce 447) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_BOOLEAN_AND) reduce 447) ((T_BOOLEAN_OR) reduce 447) ((T_COALESCE) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 447) ((T_IS_GREATER_OR_EQUAL) reduce 447) ((T_IS_IDENTICAL) reduce 447) ((T_IS_NOT_EQUAL) reduce 447) ((T_IS_NOT_IDENTICAL) reduce 447) ((T_IS_SMALLER_OR_EQUAL) reduce 447) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_PIPE) reduce 447) ((T_POW) shift) ((T_SL) reduce 447) ((T_SPACESHIP) reduce 447) ((T_SR) reduce 447) (("^") reduce 447) (("|") reduce 447)) 26445 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CASE) reduce 230) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DEFAULT) reduce 230) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_ENDSWITCH) reduce 230) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_VOID_CAST) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("~") shift)) 26446 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CASE) reduce 229) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DEFAULT) reduce 229) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_ENDSWITCH) reduce 229) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_VOID_CAST) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("~") shift)) 27132 ((("!") reduce 227) (("\"") reduce 227) (("$") reduce 227) (("(") reduce 227) (("+") reduce 227) (("-") reduce 227) ((";") reduce 227) (("@") reduce 227) ((T_ABSTRACT) reduce 227) ((T_ARRAY) reduce 227) ((T_ARRAY_CAST) reduce 227) ((T_ATTRIBUTE) reduce 227) ((T_BOOL_CAST) reduce 227) ((T_BREAK) reduce 227) ((T_CASE) reduce 227) ((T_CLASS) reduce 227) ((T_CLASS_C) reduce 227) ((T_CLONE) reduce 227) ((T_CONSTANT_ENCAPSED_STRING) reduce 227) ((T_CONTINUE) reduce 227) ((T_DEC) reduce 227) ((T_DECLARE) reduce 227) ((T_DEFAULT) reduce 227) ((T_DIR) reduce 227) ((T_DNUMBER) reduce 227) ((T_DO) reduce 227) ((T_DOUBLE_CAST) reduce 227) ((T_ECHO) reduce 227) ((T_ELSE) reduce 227) ((T_ELSEIF) reduce 227) ((T_EMPTY) reduce 227) ((T_ENUM) reduce 227) ((T_EVAL) reduce 227) ((T_EXIT) reduce 227) ((T_FILE) reduce 227) ((T_FINAL) reduce 227) ((T_FN) reduce 227) ((T_FOR) reduce 227) ((T_FOREACH) reduce 227) ((T_FUNCTION) reduce 227) ((T_FUNC_C) reduce 227) ((T_GLOBAL) reduce 227) ((T_GOTO) reduce 227) ((T_HALT_COMPILER) reduce 227) ((T_IF) reduce 227) ((T_INC) reduce 227) ((T_INCLUDE) reduce 227) ((T_INCLUDE_ONCE) reduce 227) ((T_INLINE_HTML) reduce 227) ((T_INTERFACE) reduce 227) ((T_INT_CAST) reduce 227) ((T_ISSET) reduce 227) ((T_LINE) reduce 227) ((T_LIST) reduce 227) ((T_LNUMBER) reduce 227) ((T_MATCH) reduce 227) ((T_METHOD_C) reduce 227) ((T_NAME_FULLY_QUALIFIED) reduce 227) ((T_NAME_QUALIFIED) reduce 227) ((T_NAME_RELATIVE) reduce 227) ((T_NEW) reduce 227) ((T_NS_C) reduce 227) ((T_OBJECT_CAST) reduce 227) ((T_PRINT) reduce 227) ((T_PROPERTY_C) reduce 227) ((T_READONLY) reduce 227) ((T_REQUIRE) reduce 227) ((T_REQUIRE_ONCE) reduce 227) ((T_RETURN) reduce 227) ((T_START_HEREDOC) reduce 227) ((T_STATIC) reduce 227) ((T_STRING) reduce 227) ((T_STRING_CAST) reduce 227) ((T_SWITCH) reduce 227) ((T_THROW) reduce 227) ((T_TRAIT) reduce 227) ((T_TRAIT_C) reduce 227) ((T_TRY) reduce 227) ((T_UNSET) reduce 227) ((T_UNSET_CAST) reduce 227) ((T_VARIABLE) reduce 227) ((T_VOID_CAST) reduce 227) ((T_WHILE) reduce 227) ((T_YIELD) reduce 227) ((T_YIELD_FROM) reduce 227) (("[") reduce 227) (("`") reduce 227) (("{") reduce 227) (("}") reduce 227) (("~") reduce 227)) 27133 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CASE) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DEFAULT) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_ELSE) reduce 169) ((T_ELSEIF) reduce 169) ((T_EMPTY) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_VOID_CAST) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("}") reduce 169) (("~") reduce 169)) 27135 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CASE) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DEFAULT) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_ELSE) reduce 158) ((T_ELSEIF) reduce 158) ((T_EMPTY) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_VOID_CAST) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("}") reduce 158) (("~") reduce 158)) 27137 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CASE) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DEFAULT) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_ELSE) reduce 156) ((T_ELSEIF) reduce 156) ((T_EMPTY) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_VOID_CAST) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("}") reduce 156) (("~") reduce 156)) 27139 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CASE) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DEFAULT) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_ELSE) reduce 168) ((T_ELSEIF) reduce 168) ((T_EMPTY) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_VOID_CAST) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("}") reduce 168) (("~") reduce 168)) 27141 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CASE) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DEFAULT) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_ELSE) reduce 157) ((T_ELSEIF) reduce 157) ((T_EMPTY) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_VOID_CAST) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("}") reduce 157) (("~") reduce 157)) 27151 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CASE) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DEFAULT) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_ELSE) reduce 164) ((T_ELSEIF) reduce 164) ((T_EMPTY) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_VOID_CAST) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("}") reduce 164) (("~") reduce 164)) 27152 ((("!") reduce 220) (("\"") reduce 220) (("$") reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CASE) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DEFAULT) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_ELSE) reduce 220) ((T_ELSEIF) reduce 220) ((T_EMPTY) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_VARIABLE) reduce 220) ((T_VOID_CAST) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("}") reduce 220) (("~") reduce 220)) 27155 ((("!") reduce 221) (("\"") reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CASE) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DEFAULT) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_ELSE) reduce 221) ((T_ELSEIF) reduce 221) ((T_EMPTY) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_VARIABLE) reduce 221) ((T_VOID_CAST) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("}") reduce 221) (("~") reduce 221)) 27156 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CASE) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DEFAULT) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_ELSE) reduce 163) ((T_ELSEIF) reduce 163) ((T_EMPTY) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_VOID_CAST) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("}") reduce 163) (("~") reduce 163)) 27165 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CASE) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DEFAULT) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_ELSE) reduce 152) ((T_ELSEIF) reduce 152) ((T_EMPTY) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_VOID_CAST) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("}") reduce 152) (("~") reduce 152)) 27166 ((("!") reduce 218) (("\"") reduce 218) (("$") reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CASE) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DEFAULT) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_ELSE) reduce 218) ((T_ELSEIF) reduce 218) ((T_EMPTY) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_VARIABLE) reduce 218) ((T_VOID_CAST) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("}") reduce 218) (("~") reduce 218)) 27169 ((("!") reduce 219) (("\"") reduce 219) (("$") reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CASE) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DEFAULT) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_ELSE) reduce 219) ((T_ELSEIF) reduce 219) ((T_EMPTY) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_VARIABLE) reduce 219) ((T_VOID_CAST) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("}") reduce 219) (("~") reduce 219)) 27171 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CASE) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DEFAULT) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_ELSE) reduce 159) ((T_ELSEIF) reduce 159) ((T_EMPTY) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_VOID_CAST) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("}") reduce 159) (("~") reduce 159)) 27177 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CASE) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DEFAULT) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_ELSE) reduce 151) ((T_ELSEIF) reduce 151) ((T_EMPTY) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_VOID_CAST) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("}") reduce 151) (("~") reduce 151)) 27182 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CASE) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DEFAULT) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_ELSE) reduce 165) ((T_ELSEIF) reduce 165) ((T_EMPTY) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_VOID_CAST) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("}") reduce 165) (("~") reduce 165)) 27183 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CASE) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DEFAULT) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_ELSE) reduce 222) ((T_ELSEIF) reduce 222) ((T_EMPTY) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_VARIABLE) reduce 222) ((T_VOID_CAST) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("}") reduce 222) (("~") reduce 222)) 27186 ((("!") reduce 223) (("\"") reduce 223) (("$") reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CASE) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DEFAULT) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_ELSE) reduce 223) ((T_ELSEIF) reduce 223) ((T_EMPTY) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_VARIABLE) reduce 223) ((T_VOID_CAST) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("}") reduce 223) (("~") reduce 223)) 27188 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CASE) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DEFAULT) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_ELSE) reduce 155) ((T_ELSEIF) reduce 155) ((T_EMPTY) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_VOID_CAST) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("}") reduce 155) (("~") reduce 155)) 27190 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CASE) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DEFAULT) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_ELSE) reduce 154) ((T_ELSEIF) reduce 154) ((T_EMPTY) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_VOID_CAST) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("}") reduce 154) (("~") reduce 154)) 27191 ((("!") reduce 247) (("\"") reduce 247) (("$") reduce 247) (("(") reduce 247) (("+") reduce 247) (("-") reduce 247) ((";") reduce 247) (("@") reduce 247) ((T_ABSTRACT) reduce 247) ((T_ARRAY) reduce 247) ((T_ARRAY_CAST) reduce 247) ((T_ATTRIBUTE) reduce 247) ((T_BOOL_CAST) reduce 247) ((T_BREAK) reduce 247) ((T_CASE) reduce 247) ((T_CLASS) reduce 247) ((T_CLASS_C) reduce 247) ((T_CLONE) reduce 247) ((T_CONSTANT_ENCAPSED_STRING) reduce 247) ((T_CONTINUE) reduce 247) ((T_DEC) reduce 247) ((T_DECLARE) reduce 247) ((T_DEFAULT) reduce 247) ((T_DIR) reduce 247) ((T_DNUMBER) reduce 247) ((T_DO) reduce 247) ((T_DOUBLE_CAST) reduce 247) ((T_ECHO) reduce 247) ((T_EMPTY) reduce 247) ((T_ENUM) reduce 247) ((T_EVAL) reduce 247) ((T_EXIT) reduce 247) ((T_FILE) reduce 247) ((T_FINAL) reduce 247) ((T_FN) reduce 247) ((T_FOR) reduce 247) ((T_FOREACH) reduce 247) ((T_FUNCTION) reduce 247) ((T_FUNC_C) reduce 247) ((T_GLOBAL) reduce 247) ((T_GOTO) reduce 247) ((T_HALT_COMPILER) reduce 247) ((T_IF) reduce 247) ((T_INC) reduce 247) ((T_INCLUDE) reduce 247) ((T_INCLUDE_ONCE) reduce 247) ((T_INLINE_HTML) reduce 247) ((T_INTERFACE) reduce 247) ((T_INT_CAST) reduce 247) ((T_ISSET) reduce 247) ((T_LINE) reduce 247) ((T_LIST) reduce 247) ((T_LNUMBER) reduce 247) ((T_MATCH) reduce 247) ((T_METHOD_C) reduce 247) ((T_NAME_FULLY_QUALIFIED) reduce 247) ((T_NAME_QUALIFIED) reduce 247) ((T_NAME_RELATIVE) reduce 247) ((T_NEW) reduce 247) ((T_NS_C) reduce 247) ((T_OBJECT_CAST) reduce 247) ((T_PRINT) reduce 247) ((T_PROPERTY_C) reduce 247) ((T_READONLY) reduce 247) ((T_REQUIRE) reduce 247) ((T_REQUIRE_ONCE) reduce 247) ((T_RETURN) reduce 247) ((T_START_HEREDOC) reduce 247) ((T_STATIC) reduce 247) ((T_STRING) reduce 247) ((T_STRING_CAST) reduce 247) ((T_SWITCH) reduce 247) ((T_THROW) reduce 247) ((T_TRAIT) reduce 247) ((T_TRAIT_C) reduce 247) ((T_TRY) reduce 247) ((T_UNSET) reduce 247) ((T_UNSET_CAST) reduce 247) ((T_VARIABLE) reduce 247) ((T_VOID_CAST) reduce 247) ((T_WHILE) reduce 247) ((T_YIELD) reduce 247) ((T_YIELD_FROM) reduce 247) (("[") reduce 247) (("`") reduce 247) (("{") reduce 247) (("}") reduce 247) (("~") reduce 247)) 27202 ((("!") reduce 494) (("\"") reduce 494) (("$") reduce 494) (("(") reduce 494) (("+") reduce 494) (("-") reduce 494) ((";") reduce 494) (("@") reduce 494) ((T_ABSTRACT) reduce 494) ((T_ARRAY) reduce 494) ((T_ARRAY_CAST) reduce 494) ((T_ATTRIBUTE) reduce 494) ((T_BOOL_CAST) reduce 494) ((T_BREAK) reduce 494) ((T_CASE) reduce 494) ((T_CLASS) reduce 494) ((T_CLASS_C) reduce 494) ((T_CLONE) reduce 494) ((T_CONSTANT_ENCAPSED_STRING) reduce 494) ((T_CONTINUE) reduce 494) ((T_DEC) reduce 494) ((T_DECLARE) reduce 494) ((T_DEFAULT) reduce 494) ((T_DIR) reduce 494) ((T_DNUMBER) reduce 494) ((T_DO) reduce 494) ((T_DOUBLE_CAST) reduce 494) ((T_ECHO) reduce 494) ((T_EMPTY) reduce 494) ((T_ENUM) reduce 494) ((T_EVAL) reduce 494) ((T_EXIT) reduce 494) ((T_FILE) reduce 494) ((T_FINAL) reduce 494) ((T_FN) reduce 494) ((T_FOR) reduce 494) ((T_FOREACH) reduce 494) ((T_FUNCTION) reduce 494) ((T_FUNC_C) reduce 494) ((T_GLOBAL) reduce 494) ((T_GOTO) reduce 494) ((T_HALT_COMPILER) reduce 494) ((T_IF) reduce 494) ((T_INC) reduce 494) ((T_INCLUDE) reduce 494) ((T_INCLUDE_ONCE) reduce 494) ((T_INLINE_HTML) reduce 494) ((T_INTERFACE) reduce 494) ((T_INT_CAST) reduce 494) ((T_ISSET) reduce 494) ((T_LINE) reduce 494) ((T_LIST) reduce 494) ((T_LNUMBER) reduce 494) ((T_MATCH) reduce 494) ((T_METHOD_C) reduce 494) ((T_NAME_FULLY_QUALIFIED) reduce 494) ((T_NAME_QUALIFIED) reduce 494) ((T_NAME_RELATIVE) reduce 494) ((T_NEW) reduce 494) ((T_NS_C) reduce 494) ((T_OBJECT_CAST) reduce 494) ((T_PRINT) reduce 494) ((T_PROPERTY_C) reduce 494) ((T_READONLY) reduce 494) ((T_REQUIRE) reduce 494) ((T_REQUIRE_ONCE) reduce 494) ((T_RETURN) reduce 494) ((T_START_HEREDOC) reduce 494) ((T_STATIC) reduce 494) ((T_STRING) reduce 494) ((T_STRING_CAST) reduce 494) ((T_SWITCH) reduce 494) ((T_THROW) reduce 494) ((T_TRAIT) reduce 494) ((T_TRAIT_C) reduce 494) ((T_TRY) reduce 494) ((T_UNSET) reduce 494) ((T_UNSET_CAST) reduce 494) ((T_VARIABLE) reduce 494) ((T_VOID_CAST) reduce 494) ((T_WHILE) reduce 494) ((T_YIELD) reduce 494) ((T_YIELD_FROM) reduce 494) (("[") reduce 494) (("`") reduce 494) (("{") reduce 494) (("}") reduce 494) (("~") reduce 494)) 27203 ((("!") reduce 184) (("\"") reduce 184) (("$") reduce 184) (("(") reduce 184) (("+") reduce 184) (("-") reduce 184) ((";") reduce 184) (("@") reduce 184) ((T_ABSTRACT) reduce 184) ((T_ARRAY) reduce 184) ((T_ARRAY_CAST) reduce 184) ((T_ATTRIBUTE) reduce 184) ((T_BOOL_CAST) reduce 184) ((T_BREAK) reduce 184) ((T_CASE) reduce 184) ((T_CLASS) reduce 184) ((T_CLASS_C) reduce 184) ((T_CLONE) reduce 184) ((T_CONSTANT_ENCAPSED_STRING) reduce 184) ((T_CONTINUE) reduce 184) ((T_DEC) reduce 184) ((T_DECLARE) reduce 184) ((T_DEFAULT) reduce 184) ((T_DIR) reduce 184) ((T_DNUMBER) reduce 184) ((T_DO) reduce 184) ((T_DOUBLE_CAST) reduce 184) ((T_ECHO) reduce 184) ((T_EMPTY) reduce 184) ((T_ENUM) reduce 184) ((T_EVAL) reduce 184) ((T_EXIT) reduce 184) ((T_FILE) reduce 184) ((T_FINAL) reduce 184) ((T_FN) reduce 184) ((T_FOR) reduce 184) ((T_FOREACH) reduce 184) ((T_FUNCTION) reduce 184) ((T_FUNC_C) reduce 184) ((T_GLOBAL) reduce 184) ((T_GOTO) reduce 184) ((T_HALT_COMPILER) reduce 184) ((T_IF) reduce 184) ((T_INC) reduce 184) ((T_INCLUDE) reduce 184) ((T_INCLUDE_ONCE) reduce 184) ((T_INLINE_HTML) reduce 184) ((T_INTERFACE) reduce 184) ((T_INT_CAST) reduce 184) ((T_ISSET) reduce 184) ((T_LINE) reduce 184) ((T_LIST) reduce 184) ((T_LNUMBER) reduce 184) ((T_MATCH) reduce 184) ((T_METHOD_C) reduce 184) ((T_NAME_FULLY_QUALIFIED) reduce 184) ((T_NAME_QUALIFIED) reduce 184) ((T_NAME_RELATIVE) reduce 184) ((T_NEW) reduce 184) ((T_NS_C) reduce 184) ((T_OBJECT_CAST) reduce 184) ((T_PRINT) reduce 184) ((T_PROPERTY_C) reduce 184) ((T_READONLY) reduce 184) ((T_REQUIRE) reduce 184) ((T_REQUIRE_ONCE) reduce 184) ((T_RETURN) reduce 184) ((T_START_HEREDOC) reduce 184) ((T_STATIC) reduce 184) ((T_STRING) reduce 184) ((T_STRING_CAST) reduce 184) ((T_SWITCH) reduce 184) ((T_THROW) reduce 184) ((T_TRAIT) reduce 184) ((T_TRAIT_C) reduce 184) ((T_TRY) reduce 184) ((T_UNSET) reduce 184) ((T_UNSET_CAST) reduce 184) ((T_VARIABLE) reduce 184) ((T_VOID_CAST) reduce 184) ((T_WHILE) reduce 184) ((T_YIELD) reduce 184) ((T_YIELD_FROM) reduce 184) (("[") reduce 184) (("`") reduce 184) (("{") reduce 184) (("}") reduce 184) (("~") reduce 184)) 27204 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CASE) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DEFAULT) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_EMPTY) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_VOID_CAST) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("}") reduce 161) (("~") reduce 161)) 27212 ((("!") reduce 189) (("\"") reduce 189) (("$") reduce 189) (("(") reduce 189) (("+") reduce 189) (("-") reduce 189) ((";") reduce 189) (("@") reduce 189) ((T_ABSTRACT) reduce 189) ((T_ARRAY) reduce 189) ((T_ARRAY_CAST) reduce 189) ((T_ATTRIBUTE) reduce 189) ((T_BOOL_CAST) reduce 189) ((T_BREAK) reduce 189) ((T_CASE) reduce 189) ((T_CLASS) reduce 189) ((T_CLASS_C) reduce 189) ((T_CLONE) reduce 189) ((T_CONSTANT_ENCAPSED_STRING) reduce 189) ((T_CONTINUE) reduce 189) ((T_DEC) reduce 189) ((T_DECLARE) reduce 189) ((T_DEFAULT) reduce 189) ((T_DIR) reduce 189) ((T_DNUMBER) reduce 189) ((T_DO) reduce 189) ((T_DOUBLE_CAST) reduce 189) ((T_ECHO) reduce 189) ((T_EMPTY) reduce 189) ((T_ENUM) reduce 189) ((T_EVAL) reduce 189) ((T_EXIT) reduce 189) ((T_FILE) reduce 189) ((T_FINAL) reduce 189) ((T_FN) reduce 189) ((T_FOR) reduce 189) ((T_FOREACH) reduce 189) ((T_FUNCTION) reduce 189) ((T_FUNC_C) reduce 189) ((T_GLOBAL) reduce 189) ((T_GOTO) reduce 189) ((T_HALT_COMPILER) reduce 189) ((T_IF) reduce 189) ((T_INC) reduce 189) ((T_INCLUDE) reduce 189) ((T_INCLUDE_ONCE) reduce 189) ((T_INLINE_HTML) reduce 189) ((T_INTERFACE) reduce 189) ((T_INT_CAST) reduce 189) ((T_ISSET) reduce 189) ((T_LINE) reduce 189) ((T_LIST) reduce 189) ((T_LNUMBER) reduce 189) ((T_MATCH) reduce 189) ((T_METHOD_C) reduce 189) ((T_NAME_FULLY_QUALIFIED) reduce 189) ((T_NAME_QUALIFIED) reduce 189) ((T_NAME_RELATIVE) reduce 189) ((T_NEW) reduce 189) ((T_NS_C) reduce 189) ((T_OBJECT_CAST) reduce 189) ((T_PRINT) reduce 189) ((T_PROPERTY_C) reduce 189) ((T_READONLY) reduce 189) ((T_REQUIRE) reduce 189) ((T_REQUIRE_ONCE) reduce 189) ((T_RETURN) reduce 189) ((T_START_HEREDOC) reduce 189) ((T_STATIC) reduce 189) ((T_STRING) reduce 189) ((T_STRING_CAST) reduce 189) ((T_SWITCH) reduce 189) ((T_THROW) reduce 189) ((T_TRAIT) reduce 189) ((T_TRAIT_C) reduce 189) ((T_TRY) reduce 189) ((T_UNSET) reduce 189) ((T_UNSET_CAST) reduce 189) ((T_VARIABLE) reduce 189) ((T_VOID_CAST) reduce 189) ((T_WHILE) reduce 189) ((T_YIELD) reduce 189) ((T_YIELD_FROM) reduce 189) (("[") reduce 189) (("`") reduce 189) (("{") reduce 189) (("}") reduce 189) (("~") reduce 189)) 27213 ((("!") reduce 145) (("\"") reduce 145) (("$") reduce 145) (("(") reduce 145) (("+") reduce 145) (("-") reduce 145) ((";") reduce 145) (("@") reduce 145) ((T_ABSTRACT) reduce 145) ((T_ARRAY) reduce 145) ((T_ARRAY_CAST) reduce 145) ((T_ATTRIBUTE) reduce 145) ((T_BOOL_CAST) reduce 145) ((T_BREAK) reduce 145) ((T_CASE) reduce 145) ((T_CLASS) reduce 145) ((T_CLASS_C) reduce 145) ((T_CLONE) reduce 145) ((T_CONSTANT_ENCAPSED_STRING) reduce 145) ((T_CONTINUE) reduce 145) ((T_DEC) reduce 145) ((T_DECLARE) reduce 145) ((T_DEFAULT) reduce 145) ((T_DIR) reduce 145) ((T_DNUMBER) reduce 145) ((T_DO) reduce 145) ((T_DOUBLE_CAST) reduce 145) ((T_ECHO) reduce 145) ((T_EMPTY) reduce 145) ((T_ENUM) reduce 145) ((T_EVAL) reduce 145) ((T_EXIT) reduce 145) ((T_FILE) reduce 145) ((T_FINAL) reduce 145) ((T_FN) reduce 145) ((T_FOR) reduce 145) ((T_FOREACH) reduce 145) ((T_FUNCTION) reduce 145) ((T_FUNC_C) reduce 145) ((T_GLOBAL) reduce 145) ((T_GOTO) reduce 145) ((T_HALT_COMPILER) reduce 145) ((T_IF) reduce 145) ((T_INC) reduce 145) ((T_INCLUDE) reduce 145) ((T_INCLUDE_ONCE) reduce 145) ((T_INLINE_HTML) reduce 145) ((T_INTERFACE) reduce 145) ((T_INT_CAST) reduce 145) ((T_ISSET) reduce 145) ((T_LINE) reduce 145) ((T_LIST) reduce 145) ((T_LNUMBER) reduce 145) ((T_MATCH) reduce 145) ((T_METHOD_C) reduce 145) ((T_NAME_FULLY_QUALIFIED) reduce 145) ((T_NAME_QUALIFIED) reduce 145) ((T_NAME_RELATIVE) reduce 145) ((T_NEW) reduce 145) ((T_NS_C) reduce 145) ((T_OBJECT_CAST) reduce 145) ((T_PRINT) reduce 145) ((T_PROPERTY_C) reduce 145) ((T_READONLY) reduce 145) ((T_REQUIRE) reduce 145) ((T_REQUIRE_ONCE) reduce 145) ((T_RETURN) reduce 145) ((T_START_HEREDOC) reduce 145) ((T_STATIC) reduce 145) ((T_STRING) reduce 145) ((T_STRING_CAST) reduce 145) ((T_SWITCH) reduce 145) ((T_THROW) reduce 145) ((T_TRAIT) reduce 145) ((T_TRAIT_C) reduce 145) ((T_TRY) reduce 145) ((T_UNSET) reduce 145) ((T_UNSET_CAST) reduce 145) ((T_VARIABLE) reduce 145) ((T_VOID_CAST) reduce 145) ((T_WHILE) reduce 145) ((T_YIELD) reduce 145) ((T_YIELD_FROM) reduce 145) (("[") reduce 145) (("`") reduce 145) (("{") reduce 145) (("}") reduce 145) (("~") reduce 145)) 27216 ((("!") reduce 250) (("\"") reduce 250) (("$") reduce 250) (("(") reduce 250) (("+") reduce 250) (("-") reduce 250) ((";") reduce 250) (("@") reduce 250) ((T_ABSTRACT) reduce 250) ((T_ARRAY) reduce 250) ((T_ARRAY_CAST) reduce 250) ((T_ATTRIBUTE) reduce 250) ((T_BOOL_CAST) reduce 250) ((T_BREAK) reduce 250) ((T_CASE) reduce 250) ((T_CLASS) reduce 250) ((T_CLASS_C) reduce 250) ((T_CLONE) reduce 250) ((T_CONSTANT_ENCAPSED_STRING) reduce 250) ((T_CONTINUE) reduce 250) ((T_DEC) reduce 250) ((T_DECLARE) reduce 250) ((T_DEFAULT) reduce 250) ((T_DIR) reduce 250) ((T_DNUMBER) reduce 250) ((T_DO) reduce 250) ((T_DOUBLE_CAST) reduce 250) ((T_ECHO) reduce 250) ((T_EMPTY) reduce 250) ((T_ENUM) reduce 250) ((T_EVAL) reduce 250) ((T_EXIT) reduce 250) ((T_FILE) reduce 250) ((T_FINAL) reduce 250) ((T_FN) reduce 250) ((T_FOR) reduce 250) ((T_FOREACH) reduce 250) ((T_FUNCTION) reduce 250) ((T_FUNC_C) reduce 250) ((T_GLOBAL) reduce 250) ((T_GOTO) reduce 250) ((T_HALT_COMPILER) reduce 250) ((T_IF) reduce 250) ((T_INC) reduce 250) ((T_INCLUDE) reduce 250) ((T_INCLUDE_ONCE) reduce 250) ((T_INLINE_HTML) reduce 250) ((T_INTERFACE) reduce 250) ((T_INT_CAST) reduce 250) ((T_ISSET) reduce 250) ((T_LINE) reduce 250) ((T_LIST) reduce 250) ((T_LNUMBER) reduce 250) ((T_MATCH) reduce 250) ((T_METHOD_C) reduce 250) ((T_NAME_FULLY_QUALIFIED) reduce 250) ((T_NAME_QUALIFIED) reduce 250) ((T_NAME_RELATIVE) reduce 250) ((T_NEW) reduce 250) ((T_NS_C) reduce 250) ((T_OBJECT_CAST) reduce 250) ((T_PRINT) reduce 250) ((T_PROPERTY_C) reduce 250) ((T_READONLY) reduce 250) ((T_REQUIRE) reduce 250) ((T_REQUIRE_ONCE) reduce 250) ((T_RETURN) reduce 250) ((T_START_HEREDOC) reduce 250) ((T_STATIC) reduce 250) ((T_STRING) reduce 250) ((T_STRING_CAST) reduce 250) ((T_SWITCH) reduce 250) ((T_THROW) reduce 250) ((T_TRAIT) reduce 250) ((T_TRAIT_C) reduce 250) ((T_TRY) reduce 250) ((T_UNSET) reduce 250) ((T_UNSET_CAST) reduce 250) ((T_VARIABLE) reduce 250) ((T_VOID_CAST) reduce 250) ((T_WHILE) reduce 250) ((T_YIELD) reduce 250) ((T_YIELD_FROM) reduce 250) (("[") reduce 250) (("`") reduce 250) (("{") reduce 250) (("}") reduce 250) (("~") reduce 250)) 27220 ((("!") reduce 251) (("\"") reduce 251) (("$") reduce 251) (("(") reduce 251) (("+") reduce 251) (("-") reduce 251) ((";") reduce 251) (("@") reduce 251) ((T_ABSTRACT) reduce 251) ((T_ARRAY) reduce 251) ((T_ARRAY_CAST) reduce 251) ((T_ATTRIBUTE) reduce 251) ((T_BOOL_CAST) reduce 251) ((T_BREAK) reduce 251) ((T_CASE) reduce 251) ((T_CLASS) reduce 251) ((T_CLASS_C) reduce 251) ((T_CLONE) reduce 251) ((T_CONSTANT_ENCAPSED_STRING) reduce 251) ((T_CONTINUE) reduce 251) ((T_DEC) reduce 251) ((T_DECLARE) reduce 251) ((T_DEFAULT) reduce 251) ((T_DIR) reduce 251) ((T_DNUMBER) reduce 251) ((T_DO) reduce 251) ((T_DOUBLE_CAST) reduce 251) ((T_ECHO) reduce 251) ((T_EMPTY) reduce 251) ((T_ENUM) reduce 251) ((T_EVAL) reduce 251) ((T_EXIT) reduce 251) ((T_FILE) reduce 251) ((T_FINAL) reduce 251) ((T_FN) reduce 251) ((T_FOR) reduce 251) ((T_FOREACH) reduce 251) ((T_FUNCTION) reduce 251) ((T_FUNC_C) reduce 251) ((T_GLOBAL) reduce 251) ((T_GOTO) reduce 251) ((T_HALT_COMPILER) reduce 251) ((T_IF) reduce 251) ((T_INC) reduce 251) ((T_INCLUDE) reduce 251) ((T_INCLUDE_ONCE) reduce 251) ((T_INLINE_HTML) reduce 251) ((T_INTERFACE) reduce 251) ((T_INT_CAST) reduce 251) ((T_ISSET) reduce 251) ((T_LINE) reduce 251) ((T_LIST) reduce 251) ((T_LNUMBER) reduce 251) ((T_MATCH) reduce 251) ((T_METHOD_C) reduce 251) ((T_NAME_FULLY_QUALIFIED) reduce 251) ((T_NAME_QUALIFIED) reduce 251) ((T_NAME_RELATIVE) reduce 251) ((T_NEW) reduce 251) ((T_NS_C) reduce 251) ((T_OBJECT_CAST) reduce 251) ((T_PRINT) reduce 251) ((T_PROPERTY_C) reduce 251) ((T_READONLY) reduce 251) ((T_REQUIRE) reduce 251) ((T_REQUIRE_ONCE) reduce 251) ((T_RETURN) reduce 251) ((T_START_HEREDOC) reduce 251) ((T_STATIC) reduce 251) ((T_STRING) reduce 251) ((T_STRING_CAST) reduce 251) ((T_SWITCH) reduce 251) ((T_THROW) reduce 251) ((T_TRAIT) reduce 251) ((T_TRAIT_C) reduce 251) ((T_TRY) reduce 251) ((T_UNSET) reduce 251) ((T_UNSET_CAST) reduce 251) ((T_VARIABLE) reduce 251) ((T_VOID_CAST) reduce 251) ((T_WHILE) reduce 251) ((T_YIELD) reduce 251) ((T_YIELD_FROM) reduce 251) (("[") reduce 251) (("`") reduce 251) (("{") reduce 251) (("}") reduce 251) (("~") reduce 251)) 27225 ((("!") reduce 242) (("\"") reduce 242) (("$") reduce 242) (("(") reduce 242) (("+") reduce 242) (("-") reduce 242) ((";") reduce 242) (("@") reduce 242) ((T_ABSTRACT) reduce 242) ((T_ARRAY) reduce 242) ((T_ARRAY_CAST) reduce 242) ((T_ATTRIBUTE) reduce 242) ((T_BOOL_CAST) reduce 242) ((T_BREAK) reduce 242) ((T_CASE) reduce 242) ((T_CLASS) reduce 242) ((T_CLASS_C) reduce 242) ((T_CLONE) reduce 242) ((T_CONSTANT_ENCAPSED_STRING) reduce 242) ((T_CONTINUE) reduce 242) ((T_DEC) reduce 242) ((T_DECLARE) reduce 242) ((T_DEFAULT) reduce 242) ((T_DIR) reduce 242) ((T_DNUMBER) reduce 242) ((T_DO) reduce 242) ((T_DOUBLE_CAST) reduce 242) ((T_ECHO) reduce 242) ((T_EMPTY) reduce 242) ((T_ENUM) reduce 242) ((T_EVAL) reduce 242) ((T_EXIT) reduce 242) ((T_FILE) reduce 242) ((T_FINAL) reduce 242) ((T_FN) reduce 242) ((T_FOR) reduce 242) ((T_FOREACH) reduce 242) ((T_FUNCTION) reduce 242) ((T_FUNC_C) reduce 242) ((T_GLOBAL) reduce 242) ((T_GOTO) reduce 242) ((T_HALT_COMPILER) reduce 242) ((T_IF) reduce 242) ((T_INC) reduce 242) ((T_INCLUDE) reduce 242) ((T_INCLUDE_ONCE) reduce 242) ((T_INLINE_HTML) reduce 242) ((T_INTERFACE) reduce 242) ((T_INT_CAST) reduce 242) ((T_ISSET) reduce 242) ((T_LINE) reduce 242) ((T_LIST) reduce 242) ((T_LNUMBER) reduce 242) ((T_MATCH) reduce 242) ((T_METHOD_C) reduce 242) ((T_NAME_FULLY_QUALIFIED) reduce 242) ((T_NAME_QUALIFIED) reduce 242) ((T_NAME_RELATIVE) reduce 242) ((T_NEW) reduce 242) ((T_NS_C) reduce 242) ((T_OBJECT_CAST) reduce 242) ((T_PRINT) reduce 242) ((T_PROPERTY_C) reduce 242) ((T_READONLY) reduce 242) ((T_REQUIRE) reduce 242) ((T_REQUIRE_ONCE) reduce 242) ((T_RETURN) reduce 242) ((T_START_HEREDOC) reduce 242) ((T_STATIC) reduce 242) ((T_STRING) reduce 242) ((T_STRING_CAST) reduce 242) ((T_SWITCH) reduce 242) ((T_THROW) reduce 242) ((T_TRAIT) reduce 242) ((T_TRAIT_C) reduce 242) ((T_TRY) reduce 242) ((T_UNSET) reduce 242) ((T_UNSET_CAST) reduce 242) ((T_VARIABLE) reduce 242) ((T_VOID_CAST) reduce 242) ((T_WHILE) reduce 242) ((T_YIELD) reduce 242) ((T_YIELD_FROM) reduce 242) (("[") reduce 242) (("`") reduce 242) (("{") reduce 242) (("}") reduce 242) (("~") reduce 242)) 27226 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CASE) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DEFAULT) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_EMPTY) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_VOID_CAST) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("}") reduce 150) (("~") reduce 150)) 27229 ((("!") reduce 243) (("\"") reduce 243) (("$") reduce 243) (("(") reduce 243) (("+") reduce 243) (("-") reduce 243) ((";") reduce 243) (("@") reduce 243) ((T_ABSTRACT) reduce 243) ((T_ARRAY) reduce 243) ((T_ARRAY_CAST) reduce 243) ((T_ATTRIBUTE) reduce 243) ((T_BOOL_CAST) reduce 243) ((T_BREAK) reduce 243) ((T_CASE) reduce 243) ((T_CLASS) reduce 243) ((T_CLASS_C) reduce 243) ((T_CLONE) reduce 243) ((T_CONSTANT_ENCAPSED_STRING) reduce 243) ((T_CONTINUE) reduce 243) ((T_DEC) reduce 243) ((T_DECLARE) reduce 243) ((T_DEFAULT) reduce 243) ((T_DIR) reduce 243) ((T_DNUMBER) reduce 243) ((T_DO) reduce 243) ((T_DOUBLE_CAST) reduce 243) ((T_ECHO) reduce 243) ((T_EMPTY) reduce 243) ((T_ENUM) reduce 243) ((T_EVAL) reduce 243) ((T_EXIT) reduce 243) ((T_FILE) reduce 243) ((T_FINAL) reduce 243) ((T_FN) reduce 243) ((T_FOR) reduce 243) ((T_FOREACH) reduce 243) ((T_FUNCTION) reduce 243) ((T_FUNC_C) reduce 243) ((T_GLOBAL) reduce 243) ((T_GOTO) reduce 243) ((T_HALT_COMPILER) reduce 243) ((T_IF) reduce 243) ((T_INC) reduce 243) ((T_INCLUDE) reduce 243) ((T_INCLUDE_ONCE) reduce 243) ((T_INLINE_HTML) reduce 243) ((T_INTERFACE) reduce 243) ((T_INT_CAST) reduce 243) ((T_ISSET) reduce 243) ((T_LINE) reduce 243) ((T_LIST) reduce 243) ((T_LNUMBER) reduce 243) ((T_MATCH) reduce 243) ((T_METHOD_C) reduce 243) ((T_NAME_FULLY_QUALIFIED) reduce 243) ((T_NAME_QUALIFIED) reduce 243) ((T_NAME_RELATIVE) reduce 243) ((T_NEW) reduce 243) ((T_NS_C) reduce 243) ((T_OBJECT_CAST) reduce 243) ((T_PRINT) reduce 243) ((T_PROPERTY_C) reduce 243) ((T_READONLY) reduce 243) ((T_REQUIRE) reduce 243) ((T_REQUIRE_ONCE) reduce 243) ((T_RETURN) reduce 243) ((T_START_HEREDOC) reduce 243) ((T_STATIC) reduce 243) ((T_STRING) reduce 243) ((T_STRING_CAST) reduce 243) ((T_SWITCH) reduce 243) ((T_THROW) reduce 243) ((T_TRAIT) reduce 243) ((T_TRAIT_C) reduce 243) ((T_TRY) reduce 243) ((T_UNSET) reduce 243) ((T_UNSET_CAST) reduce 243) ((T_VARIABLE) reduce 243) ((T_VOID_CAST) reduce 243) ((T_WHILE) reduce 243) ((T_YIELD) reduce 243) ((T_YIELD_FROM) reduce 243) (("[") reduce 243) (("`") reduce 243) (("{") reduce 243) (("}") reduce 243) (("~") reduce 243)) 27231 ((("!") reduce 170) (("\"") reduce 170) (("$") reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CASE) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DEFAULT) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_EMPTY) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_VARIABLE) reduce 170) ((T_VOID_CAST) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("}") reduce 170) (("~") reduce 170)) 27236 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CASE) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DEFAULT) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_EMPTY) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_VOID_CAST) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("}") reduce 162) (("~") reduce 162)) 27239 ((("!") reduce 171) (("\"") reduce 171) (("$") reduce 171) (("(") reduce 171) (("+") reduce 171) (("-") reduce 171) ((";") reduce 171) (("@") reduce 171) ((T_ABSTRACT) reduce 171) ((T_ARRAY) reduce 171) ((T_ARRAY_CAST) reduce 171) ((T_ATTRIBUTE) reduce 171) ((T_BOOL_CAST) reduce 171) ((T_BREAK) reduce 171) ((T_CASE) reduce 171) ((T_CATCH) reduce 171) ((T_CLASS) reduce 171) ((T_CLASS_C) reduce 171) ((T_CLONE) reduce 171) ((T_CONSTANT_ENCAPSED_STRING) reduce 171) ((T_CONTINUE) reduce 171) ((T_DEC) reduce 171) ((T_DECLARE) reduce 171) ((T_DEFAULT) reduce 171) ((T_DIR) reduce 171) ((T_DNUMBER) reduce 171) ((T_DO) reduce 171) ((T_DOUBLE_CAST) reduce 171) ((T_ECHO) reduce 171) ((T_EMPTY) reduce 171) ((T_ENUM) reduce 171) ((T_EVAL) reduce 171) ((T_EXIT) reduce 171) ((T_FILE) reduce 171) ((T_FINAL) reduce 171) ((T_FINALLY) reduce 171) ((T_FN) reduce 171) ((T_FOR) reduce 171) ((T_FOREACH) reduce 171) ((T_FUNCTION) reduce 171) ((T_FUNC_C) reduce 171) ((T_GLOBAL) reduce 171) ((T_GOTO) reduce 171) ((T_HALT_COMPILER) reduce 171) ((T_IF) reduce 171) ((T_INC) reduce 171) ((T_INCLUDE) reduce 171) ((T_INCLUDE_ONCE) reduce 171) ((T_INLINE_HTML) reduce 171) ((T_INTERFACE) reduce 171) ((T_INT_CAST) reduce 171) ((T_ISSET) reduce 171) ((T_LINE) reduce 171) ((T_LIST) reduce 171) ((T_LNUMBER) reduce 171) ((T_MATCH) reduce 171) ((T_METHOD_C) reduce 171) ((T_NAME_FULLY_QUALIFIED) reduce 171) ((T_NAME_QUALIFIED) reduce 171) ((T_NAME_RELATIVE) reduce 171) ((T_NEW) reduce 171) ((T_NS_C) reduce 171) ((T_OBJECT_CAST) reduce 171) ((T_PRINT) reduce 171) ((T_PROPERTY_C) reduce 171) ((T_READONLY) reduce 171) ((T_REQUIRE) reduce 171) ((T_REQUIRE_ONCE) reduce 171) ((T_RETURN) reduce 171) ((T_START_HEREDOC) reduce 171) ((T_STATIC) reduce 171) ((T_STRING) reduce 171) ((T_STRING_CAST) reduce 171) ((T_SWITCH) reduce 171) ((T_THROW) reduce 171) ((T_TRAIT) reduce 171) ((T_TRAIT_C) reduce 171) ((T_TRY) reduce 171) ((T_UNSET) reduce 171) ((T_UNSET_CAST) reduce 171) ((T_VARIABLE) reduce 171) ((T_VOID_CAST) reduce 171) ((T_WHILE) reduce 171) ((T_YIELD) reduce 171) ((T_YIELD_FROM) reduce 171) (("[") reduce 171) (("`") reduce 171) (("{") reduce 171) (("}") reduce 171) (("~") reduce 171)) 27240 ((("!") reduce 177) (("\"") reduce 177) (("$") reduce 177) (("(") reduce 177) (("+") reduce 177) (("-") reduce 177) ((";") reduce 177) (("@") reduce 177) ((T_ABSTRACT) reduce 177) ((T_ARRAY) reduce 177) ((T_ARRAY_CAST) reduce 177) ((T_ATTRIBUTE) reduce 177) ((T_BOOL_CAST) reduce 177) ((T_BREAK) reduce 177) ((T_CASE) reduce 177) ((T_CATCH) shift) ((T_CLASS) reduce 177) ((T_CLASS_C) reduce 177) ((T_CLONE) reduce 177) ((T_CONSTANT_ENCAPSED_STRING) reduce 177) ((T_CONTINUE) reduce 177) ((T_DEC) reduce 177) ((T_DECLARE) reduce 177) ((T_DEFAULT) reduce 177) ((T_DIR) reduce 177) ((T_DNUMBER) reduce 177) ((T_DO) reduce 177) ((T_DOUBLE_CAST) reduce 177) ((T_ECHO) reduce 177) ((T_EMPTY) reduce 177) ((T_ENUM) reduce 177) ((T_EVAL) reduce 177) ((T_EXIT) reduce 177) ((T_FILE) reduce 177) ((T_FINAL) reduce 177) ((T_FINALLY) shift) ((T_FN) reduce 177) ((T_FOR) reduce 177) ((T_FOREACH) reduce 177) ((T_FUNCTION) reduce 177) ((T_FUNC_C) reduce 177) ((T_GLOBAL) reduce 177) ((T_GOTO) reduce 177) ((T_HALT_COMPILER) reduce 177) ((T_IF) reduce 177) ((T_INC) reduce 177) ((T_INCLUDE) reduce 177) ((T_INCLUDE_ONCE) reduce 177) ((T_INLINE_HTML) reduce 177) ((T_INTERFACE) reduce 177) ((T_INT_CAST) reduce 177) ((T_ISSET) reduce 177) ((T_LINE) reduce 177) ((T_LIST) reduce 177) ((T_LNUMBER) reduce 177) ((T_MATCH) reduce 177) ((T_METHOD_C) reduce 177) ((T_NAME_FULLY_QUALIFIED) reduce 177) ((T_NAME_QUALIFIED) reduce 177) ((T_NAME_RELATIVE) reduce 177) ((T_NEW) reduce 177) ((T_NS_C) reduce 177) ((T_OBJECT_CAST) reduce 177) ((T_PRINT) reduce 177) ((T_PROPERTY_C) reduce 177) ((T_READONLY) reduce 177) ((T_REQUIRE) reduce 177) ((T_REQUIRE_ONCE) reduce 177) ((T_RETURN) reduce 177) ((T_START_HEREDOC) reduce 177) ((T_STATIC) reduce 177) ((T_STRING) reduce 177) ((T_STRING_CAST) reduce 177) ((T_SWITCH) reduce 177) ((T_THROW) reduce 177) ((T_TRAIT) reduce 177) ((T_TRAIT_C) reduce 177) ((T_TRY) reduce 177) ((T_UNSET) reduce 177) ((T_UNSET_CAST) reduce 177) ((T_VARIABLE) reduce 177) ((T_VOID_CAST) reduce 177) ((T_WHILE) reduce 177) ((T_YIELD) reduce 177) ((T_YIELD_FROM) reduce 177) (("[") reduce 177) (("`") reduce 177) (("{") reduce 177) (("}") reduce 177) (("~") reduce 177)) 27243 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CASE) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DEFAULT) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_EMPTY) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_VOID_CAST) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("}") reduce 167) (("~") reduce 167)) 27246 ((("!") reduce 178) (("\"") reduce 178) (("$") reduce 178) (("(") reduce 178) (("+") reduce 178) (("-") reduce 178) ((";") reduce 178) (("@") reduce 178) ((T_ABSTRACT) reduce 178) ((T_ARRAY) reduce 178) ((T_ARRAY_CAST) reduce 178) ((T_ATTRIBUTE) reduce 178) ((T_BOOL_CAST) reduce 178) ((T_BREAK) reduce 178) ((T_CASE) reduce 178) ((T_CLASS) reduce 178) ((T_CLASS_C) reduce 178) ((T_CLONE) reduce 178) ((T_CONSTANT_ENCAPSED_STRING) reduce 178) ((T_CONTINUE) reduce 178) ((T_DEC) reduce 178) ((T_DECLARE) reduce 178) ((T_DEFAULT) reduce 178) ((T_DIR) reduce 178) ((T_DNUMBER) reduce 178) ((T_DO) reduce 178) ((T_DOUBLE_CAST) reduce 178) ((T_ECHO) reduce 178) ((T_EMPTY) reduce 178) ((T_ENUM) reduce 178) ((T_EVAL) reduce 178) ((T_EXIT) reduce 178) ((T_FILE) reduce 178) ((T_FINAL) reduce 178) ((T_FN) reduce 178) ((T_FOR) reduce 178) ((T_FOREACH) reduce 178) ((T_FUNCTION) reduce 178) ((T_FUNC_C) reduce 178) ((T_GLOBAL) reduce 178) ((T_GOTO) reduce 178) ((T_HALT_COMPILER) reduce 178) ((T_IF) reduce 178) ((T_INC) reduce 178) ((T_INCLUDE) reduce 178) ((T_INCLUDE_ONCE) reduce 178) ((T_INLINE_HTML) reduce 178) ((T_INTERFACE) reduce 178) ((T_INT_CAST) reduce 178) ((T_ISSET) reduce 178) ((T_LINE) reduce 178) ((T_LIST) reduce 178) ((T_LNUMBER) reduce 178) ((T_MATCH) reduce 178) ((T_METHOD_C) reduce 178) ((T_NAME_FULLY_QUALIFIED) reduce 178) ((T_NAME_QUALIFIED) reduce 178) ((T_NAME_RELATIVE) reduce 178) ((T_NEW) reduce 178) ((T_NS_C) reduce 178) ((T_OBJECT_CAST) reduce 178) ((T_PRINT) reduce 178) ((T_PROPERTY_C) reduce 178) ((T_READONLY) reduce 178) ((T_REQUIRE) reduce 178) ((T_REQUIRE_ONCE) reduce 178) ((T_RETURN) reduce 178) ((T_START_HEREDOC) reduce 178) ((T_STATIC) reduce 178) ((T_STRING) reduce 178) ((T_STRING_CAST) reduce 178) ((T_SWITCH) reduce 178) ((T_THROW) reduce 178) ((T_TRAIT) reduce 178) ((T_TRAIT_C) reduce 178) ((T_TRY) reduce 178) ((T_UNSET) reduce 178) ((T_UNSET_CAST) reduce 178) ((T_VARIABLE) reduce 178) ((T_VOID_CAST) reduce 178) ((T_WHILE) reduce 178) ((T_YIELD) reduce 178) ((T_YIELD_FROM) reduce 178) (("[") reduce 178) (("`") reduce 178) (("{") reduce 178) (("}") reduce 178) (("~") reduce 178)) 27253 ((("!") reduce 172) (("\"") reduce 172) (("$") reduce 172) (("(") reduce 172) (("+") reduce 172) (("-") reduce 172) ((";") reduce 172) (("@") reduce 172) ((T_ABSTRACT) reduce 172) ((T_ARRAY) reduce 172) ((T_ARRAY_CAST) reduce 172) ((T_ATTRIBUTE) reduce 172) ((T_BOOL_CAST) reduce 172) ((T_BREAK) reduce 172) ((T_CASE) reduce 172) ((T_CATCH) reduce 172) ((T_CLASS) reduce 172) ((T_CLASS_C) reduce 172) ((T_CLONE) reduce 172) ((T_CONSTANT_ENCAPSED_STRING) reduce 172) ((T_CONTINUE) reduce 172) ((T_DEC) reduce 172) ((T_DECLARE) reduce 172) ((T_DEFAULT) reduce 172) ((T_DIR) reduce 172) ((T_DNUMBER) reduce 172) ((T_DO) reduce 172) ((T_DOUBLE_CAST) reduce 172) ((T_ECHO) reduce 172) ((T_EMPTY) reduce 172) ((T_ENUM) reduce 172) ((T_EVAL) reduce 172) ((T_EXIT) reduce 172) ((T_FILE) reduce 172) ((T_FINAL) reduce 172) ((T_FINALLY) reduce 172) ((T_FN) reduce 172) ((T_FOR) reduce 172) ((T_FOREACH) reduce 172) ((T_FUNCTION) reduce 172) ((T_FUNC_C) reduce 172) ((T_GLOBAL) reduce 172) ((T_GOTO) reduce 172) ((T_HALT_COMPILER) reduce 172) ((T_IF) reduce 172) ((T_INC) reduce 172) ((T_INCLUDE) reduce 172) ((T_INCLUDE_ONCE) reduce 172) ((T_INLINE_HTML) reduce 172) ((T_INTERFACE) reduce 172) ((T_INT_CAST) reduce 172) ((T_ISSET) reduce 172) ((T_LINE) reduce 172) ((T_LIST) reduce 172) ((T_LNUMBER) reduce 172) ((T_MATCH) reduce 172) ((T_METHOD_C) reduce 172) ((T_NAME_FULLY_QUALIFIED) reduce 172) ((T_NAME_QUALIFIED) reduce 172) ((T_NAME_RELATIVE) reduce 172) ((T_NEW) reduce 172) ((T_NS_C) reduce 172) ((T_OBJECT_CAST) reduce 172) ((T_PRINT) reduce 172) ((T_PROPERTY_C) reduce 172) ((T_READONLY) reduce 172) ((T_REQUIRE) reduce 172) ((T_REQUIRE_ONCE) reduce 172) ((T_RETURN) reduce 172) ((T_START_HEREDOC) reduce 172) ((T_STATIC) reduce 172) ((T_STRING) reduce 172) ((T_STRING_CAST) reduce 172) ((T_SWITCH) reduce 172) ((T_THROW) reduce 172) ((T_TRAIT) reduce 172) ((T_TRAIT_C) reduce 172) ((T_TRY) reduce 172) ((T_UNSET) reduce 172) ((T_UNSET_CAST) reduce 172) ((T_VARIABLE) reduce 172) ((T_VOID_CAST) reduce 172) ((T_WHILE) reduce 172) ((T_YIELD) reduce 172) ((T_YIELD_FROM) reduce 172) (("[") reduce 172) (("`") reduce 172) (("{") reduce 172) (("}") reduce 172) (("~") reduce 172)) 27258 ((("!") reduce 200) (("\"") reduce 200) (("$") reduce 200) (("(") reduce 200) (("+") reduce 200) (("-") reduce 200) ((";") reduce 200) (("@") reduce 200) ((T_ABSTRACT) reduce 200) ((T_ARRAY) reduce 200) ((T_ARRAY_CAST) reduce 200) ((T_ATTRIBUTE) reduce 200) ((T_BOOL_CAST) reduce 200) ((T_BREAK) reduce 200) ((T_CASE) reduce 200) ((T_CLASS) reduce 200) ((T_CLASS_C) reduce 200) ((T_CLONE) reduce 200) ((T_CONSTANT_ENCAPSED_STRING) reduce 200) ((T_CONTINUE) reduce 200) ((T_DEC) reduce 200) ((T_DECLARE) reduce 200) ((T_DEFAULT) reduce 200) ((T_DIR) reduce 200) ((T_DNUMBER) reduce 200) ((T_DO) reduce 200) ((T_DOUBLE_CAST) reduce 200) ((T_ECHO) reduce 200) ((T_EMPTY) reduce 200) ((T_ENUM) reduce 200) ((T_EVAL) reduce 200) ((T_EXIT) reduce 200) ((T_FILE) reduce 200) ((T_FINAL) reduce 200) ((T_FN) reduce 200) ((T_FOR) reduce 200) ((T_FOREACH) reduce 200) ((T_FUNCTION) reduce 200) ((T_FUNC_C) reduce 200) ((T_GLOBAL) reduce 200) ((T_GOTO) reduce 200) ((T_HALT_COMPILER) reduce 200) ((T_IF) reduce 200) ((T_INC) reduce 200) ((T_INCLUDE) reduce 200) ((T_INCLUDE_ONCE) reduce 200) ((T_INLINE_HTML) reduce 200) ((T_INTERFACE) reduce 200) ((T_INT_CAST) reduce 200) ((T_ISSET) reduce 200) ((T_LINE) reduce 200) ((T_LIST) reduce 200) ((T_LNUMBER) reduce 200) ((T_MATCH) reduce 200) ((T_METHOD_C) reduce 200) ((T_NAME_FULLY_QUALIFIED) reduce 200) ((T_NAME_QUALIFIED) reduce 200) ((T_NAME_RELATIVE) reduce 200) ((T_NEW) reduce 200) ((T_NS_C) reduce 200) ((T_OBJECT_CAST) reduce 200) ((T_PRINT) reduce 200) ((T_PROPERTY_C) reduce 200) ((T_READONLY) reduce 200) ((T_REQUIRE) reduce 200) ((T_REQUIRE_ONCE) reduce 200) ((T_RETURN) reduce 200) ((T_START_HEREDOC) reduce 200) ((T_STATIC) reduce 200) ((T_STRING) reduce 200) ((T_STRING_CAST) reduce 200) ((T_SWITCH) reduce 200) ((T_THROW) reduce 200) ((T_TRAIT) reduce 200) ((T_TRAIT_C) reduce 200) ((T_TRY) reduce 200) ((T_UNSET) reduce 200) ((T_UNSET_CAST) reduce 200) ((T_VARIABLE) reduce 200) ((T_VOID_CAST) reduce 200) ((T_WHILE) reduce 200) ((T_YIELD) reduce 200) ((T_YIELD_FROM) reduce 200) (("[") reduce 200) (("`") reduce 200) (("{") reduce 200) (("}") reduce 200) (("~") reduce 200)) 27263 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CASE) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DEFAULT) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_EMPTY) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_VOID_CAST) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("}") reduce 153) (("~") reduce 153)) 27267 ((("!") reduce 224) (("\"") reduce 224) (("$") reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CASE) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DEFAULT) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_EMPTY) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_VARIABLE) reduce 224) ((T_VOID_CAST) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("}") reduce 224) (("~") reduce 224)) 27269 ((("!") reduce 225) (("\"") reduce 225) (("$") reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CASE) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DEFAULT) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_EMPTY) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_VARIABLE) reduce 225) ((T_VOID_CAST) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("}") reduce 225) (("~") reduce 225)) 27273 ((("!") reduce 226) (("\"") reduce 226) (("$") reduce 226) (("(") reduce 226) (("+") reduce 226) (("-") reduce 226) ((";") reduce 226) (("@") reduce 226) ((T_ABSTRACT) reduce 226) ((T_ARRAY) reduce 226) ((T_ARRAY_CAST) reduce 226) ((T_ATTRIBUTE) reduce 226) ((T_BOOL_CAST) reduce 226) ((T_BREAK) reduce 226) ((T_CASE) reduce 226) ((T_CLASS) reduce 226) ((T_CLASS_C) reduce 226) ((T_CLONE) reduce 226) ((T_CONSTANT_ENCAPSED_STRING) reduce 226) ((T_CONTINUE) reduce 226) ((T_DEC) reduce 226) ((T_DECLARE) reduce 226) ((T_DEFAULT) reduce 226) ((T_DIR) reduce 226) ((T_DNUMBER) reduce 226) ((T_DO) reduce 226) ((T_DOUBLE_CAST) reduce 226) ((T_ECHO) reduce 226) ((T_EMPTY) reduce 226) ((T_ENUM) reduce 226) ((T_EVAL) reduce 226) ((T_EXIT) reduce 226) ((T_FILE) reduce 226) ((T_FINAL) reduce 226) ((T_FN) reduce 226) ((T_FOR) reduce 226) ((T_FOREACH) reduce 226) ((T_FUNCTION) reduce 226) ((T_FUNC_C) reduce 226) ((T_GLOBAL) reduce 226) ((T_GOTO) reduce 226) ((T_HALT_COMPILER) reduce 226) ((T_IF) reduce 226) ((T_INC) reduce 226) ((T_INCLUDE) reduce 226) ((T_INCLUDE_ONCE) reduce 226) ((T_INLINE_HTML) reduce 226) ((T_INTERFACE) reduce 226) ((T_INT_CAST) reduce 226) ((T_ISSET) reduce 226) ((T_LINE) reduce 226) ((T_LIST) reduce 226) ((T_LNUMBER) reduce 226) ((T_MATCH) reduce 226) ((T_METHOD_C) reduce 226) ((T_NAME_FULLY_QUALIFIED) reduce 226) ((T_NAME_QUALIFIED) reduce 226) ((T_NAME_RELATIVE) reduce 226) ((T_NEW) reduce 226) ((T_NS_C) reduce 226) ((T_OBJECT_CAST) reduce 226) ((T_PRINT) reduce 226) ((T_PROPERTY_C) reduce 226) ((T_READONLY) reduce 226) ((T_REQUIRE) reduce 226) ((T_REQUIRE_ONCE) reduce 226) ((T_RETURN) reduce 226) ((T_START_HEREDOC) reduce 226) ((T_STATIC) reduce 226) ((T_STRING) reduce 226) ((T_STRING_CAST) reduce 226) ((T_SWITCH) reduce 226) ((T_THROW) reduce 226) ((T_TRAIT) reduce 226) ((T_TRAIT_C) reduce 226) ((T_TRY) reduce 226) ((T_UNSET) reduce 226) ((T_UNSET_CAST) reduce 226) ((T_VARIABLE) reduce 226) ((T_VOID_CAST) reduce 226) ((T_WHILE) reduce 226) ((T_YIELD) reduce 226) ((T_YIELD_FROM) reduce 226) (("[") reduce 226) (("`") reduce 226) (("{") reduce 226) (("}") reduce 226) (("~") reduce 226)) 27276 ((("!") reduce 227) (("\"") reduce 227) (("$") reduce 227) (("(") reduce 227) (("+") reduce 227) (("-") reduce 227) ((";") reduce 227) (("@") reduce 227) ((T_ABSTRACT) reduce 227) ((T_ARRAY) reduce 227) ((T_ARRAY_CAST) reduce 227) ((T_ATTRIBUTE) reduce 227) ((T_BOOL_CAST) reduce 227) ((T_BREAK) reduce 227) ((T_CASE) reduce 227) ((T_CLASS) reduce 227) ((T_CLASS_C) reduce 227) ((T_CLONE) reduce 227) ((T_CONSTANT_ENCAPSED_STRING) reduce 227) ((T_CONTINUE) reduce 227) ((T_DEC) reduce 227) ((T_DECLARE) reduce 227) ((T_DEFAULT) reduce 227) ((T_DIR) reduce 227) ((T_DNUMBER) reduce 227) ((T_DO) reduce 227) ((T_DOUBLE_CAST) reduce 227) ((T_ECHO) reduce 227) ((T_EMPTY) reduce 227) ((T_ENUM) reduce 227) ((T_EVAL) reduce 227) ((T_EXIT) reduce 227) ((T_FILE) reduce 227) ((T_FINAL) reduce 227) ((T_FN) reduce 227) ((T_FOR) reduce 227) ((T_FOREACH) reduce 227) ((T_FUNCTION) reduce 227) ((T_FUNC_C) reduce 227) ((T_GLOBAL) reduce 227) ((T_GOTO) reduce 227) ((T_HALT_COMPILER) reduce 227) ((T_IF) reduce 227) ((T_INC) reduce 227) ((T_INCLUDE) reduce 227) ((T_INCLUDE_ONCE) reduce 227) ((T_INLINE_HTML) reduce 227) ((T_INTERFACE) reduce 227) ((T_INT_CAST) reduce 227) ((T_ISSET) reduce 227) ((T_LINE) reduce 227) ((T_LIST) reduce 227) ((T_LNUMBER) reduce 227) ((T_MATCH) reduce 227) ((T_METHOD_C) reduce 227) ((T_NAME_FULLY_QUALIFIED) reduce 227) ((T_NAME_QUALIFIED) reduce 227) ((T_NAME_RELATIVE) reduce 227) ((T_NEW) reduce 227) ((T_NS_C) reduce 227) ((T_OBJECT_CAST) reduce 227) ((T_PRINT) reduce 227) ((T_PROPERTY_C) reduce 227) ((T_READONLY) reduce 227) ((T_REQUIRE) reduce 227) ((T_REQUIRE_ONCE) reduce 227) ((T_RETURN) reduce 227) ((T_START_HEREDOC) reduce 227) ((T_STATIC) reduce 227) ((T_STRING) reduce 227) ((T_STRING_CAST) reduce 227) ((T_SWITCH) reduce 227) ((T_THROW) reduce 227) ((T_TRAIT) reduce 227) ((T_TRAIT_C) reduce 227) ((T_TRY) reduce 227) ((T_UNSET) reduce 227) ((T_UNSET_CAST) reduce 227) ((T_VARIABLE) reduce 227) ((T_VOID_CAST) reduce 227) ((T_WHILE) reduce 227) ((T_YIELD) reduce 227) ((T_YIELD_FROM) reduce 227) (("[") reduce 227) (("`") reduce 227) (("{") reduce 227) (("}") reduce 227) (("~") reduce 227)) 27277 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CASE) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DEFAULT) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_EMPTY) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_VOID_CAST) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("}") reduce 169) (("~") reduce 169)) 27279 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CASE) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DEFAULT) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_EMPTY) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_VOID_CAST) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("}") reduce 158) (("~") reduce 158)) 27281 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CASE) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DEFAULT) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_EMPTY) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_VOID_CAST) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("}") reduce 156) (("~") reduce 156)) 27287 ((("!") reduce 201) (("\"") reduce 201) (("$") reduce 201) (("(") reduce 201) (("+") reduce 201) (("-") reduce 201) ((";") reduce 201) (("@") reduce 201) ((T_ABSTRACT) reduce 201) ((T_ARRAY) reduce 201) ((T_ARRAY_CAST) reduce 201) ((T_ATTRIBUTE) reduce 201) ((T_BOOL_CAST) reduce 201) ((T_BREAK) reduce 201) ((T_CASE) reduce 201) ((T_CLASS) reduce 201) ((T_CLASS_C) reduce 201) ((T_CLONE) reduce 201) ((T_CONSTANT_ENCAPSED_STRING) reduce 201) ((T_CONTINUE) reduce 201) ((T_DEC) reduce 201) ((T_DECLARE) reduce 201) ((T_DEFAULT) reduce 201) ((T_DIR) reduce 201) ((T_DNUMBER) reduce 201) ((T_DO) reduce 201) ((T_DOUBLE_CAST) reduce 201) ((T_ECHO) reduce 201) ((T_EMPTY) reduce 201) ((T_ENUM) reduce 201) ((T_EVAL) reduce 201) ((T_EXIT) reduce 201) ((T_FILE) reduce 201) ((T_FINAL) reduce 201) ((T_FN) reduce 201) ((T_FOR) reduce 201) ((T_FOREACH) reduce 201) ((T_FUNCTION) reduce 201) ((T_FUNC_C) reduce 201) ((T_GLOBAL) reduce 201) ((T_GOTO) reduce 201) ((T_HALT_COMPILER) reduce 201) ((T_IF) reduce 201) ((T_INC) reduce 201) ((T_INCLUDE) reduce 201) ((T_INCLUDE_ONCE) reduce 201) ((T_INLINE_HTML) reduce 201) ((T_INTERFACE) reduce 201) ((T_INT_CAST) reduce 201) ((T_ISSET) reduce 201) ((T_LINE) reduce 201) ((T_LIST) reduce 201) ((T_LNUMBER) reduce 201) ((T_MATCH) reduce 201) ((T_METHOD_C) reduce 201) ((T_NAME_FULLY_QUALIFIED) reduce 201) ((T_NAME_QUALIFIED) reduce 201) ((T_NAME_RELATIVE) reduce 201) ((T_NEW) reduce 201) ((T_NS_C) reduce 201) ((T_OBJECT_CAST) reduce 201) ((T_PRINT) reduce 201) ((T_PROPERTY_C) reduce 201) ((T_READONLY) reduce 201) ((T_REQUIRE) reduce 201) ((T_REQUIRE_ONCE) reduce 201) ((T_RETURN) reduce 201) ((T_START_HEREDOC) reduce 201) ((T_STATIC) reduce 201) ((T_STRING) reduce 201) ((T_STRING_CAST) reduce 201) ((T_SWITCH) reduce 201) ((T_THROW) reduce 201) ((T_TRAIT) reduce 201) ((T_TRAIT_C) reduce 201) ((T_TRY) reduce 201) ((T_UNSET) reduce 201) ((T_UNSET_CAST) reduce 201) ((T_VARIABLE) reduce 201) ((T_VOID_CAST) reduce 201) ((T_WHILE) reduce 201) ((T_YIELD) reduce 201) ((T_YIELD_FROM) reduce 201) (("[") reduce 201) (("`") reduce 201) (("{") reduce 201) (("}") reduce 201) (("~") reduce 201)) 27291 ((("!") reduce 244) (("\"") reduce 244) (("$") reduce 244) (("(") reduce 244) (("+") reduce 244) (("-") reduce 244) ((";") reduce 244) (("@") reduce 244) ((T_ABSTRACT) reduce 244) ((T_ARRAY) reduce 244) ((T_ARRAY_CAST) reduce 244) ((T_ATTRIBUTE) reduce 244) ((T_BOOL_CAST) reduce 244) ((T_BREAK) reduce 244) ((T_CASE) reduce 244) ((T_CLASS) reduce 244) ((T_CLASS_C) reduce 244) ((T_CLONE) reduce 244) ((T_CONSTANT_ENCAPSED_STRING) reduce 244) ((T_CONTINUE) reduce 244) ((T_DEC) reduce 244) ((T_DECLARE) reduce 244) ((T_DEFAULT) reduce 244) ((T_DIR) reduce 244) ((T_DNUMBER) reduce 244) ((T_DO) reduce 244) ((T_DOUBLE_CAST) reduce 244) ((T_ECHO) reduce 244) ((T_ELSE) reduce 244) ((T_ELSEIF) reduce 244) ((T_EMPTY) reduce 244) ((T_ENUM) reduce 244) ((T_EVAL) reduce 244) ((T_EXIT) reduce 244) ((T_FILE) reduce 244) ((T_FINAL) reduce 244) ((T_FN) reduce 244) ((T_FOR) reduce 244) ((T_FOREACH) reduce 244) ((T_FUNCTION) reduce 244) ((T_FUNC_C) reduce 244) ((T_GLOBAL) reduce 244) ((T_GOTO) reduce 244) ((T_HALT_COMPILER) reduce 244) ((T_IF) reduce 244) ((T_INC) reduce 244) ((T_INCLUDE) reduce 244) ((T_INCLUDE_ONCE) reduce 244) ((T_INLINE_HTML) reduce 244) ((T_INTERFACE) reduce 244) ((T_INT_CAST) reduce 244) ((T_ISSET) reduce 244) ((T_LINE) reduce 244) ((T_LIST) reduce 244) ((T_LNUMBER) reduce 244) ((T_MATCH) reduce 244) ((T_METHOD_C) reduce 244) ((T_NAME_FULLY_QUALIFIED) reduce 244) ((T_NAME_QUALIFIED) reduce 244) ((T_NAME_RELATIVE) reduce 244) ((T_NEW) reduce 244) ((T_NS_C) reduce 244) ((T_OBJECT_CAST) reduce 244) ((T_PRINT) reduce 244) ((T_PROPERTY_C) reduce 244) ((T_READONLY) reduce 244) ((T_REQUIRE) reduce 244) ((T_REQUIRE_ONCE) reduce 244) ((T_RETURN) reduce 244) ((T_START_HEREDOC) reduce 244) ((T_STATIC) reduce 244) ((T_STRING) reduce 244) ((T_STRING_CAST) reduce 244) ((T_SWITCH) reduce 244) ((T_THROW) reduce 244) ((T_TRAIT) reduce 244) ((T_TRAIT_C) reduce 244) ((T_TRY) reduce 244) ((T_UNSET) reduce 244) ((T_UNSET_CAST) reduce 244) ((T_VARIABLE) reduce 244) ((T_VOID_CAST) reduce 244) ((T_WHILE) reduce 244) ((T_YIELD) reduce 244) ((T_YIELD_FROM) reduce 244) (("[") reduce 244) (("`") reduce 244) (("{") reduce 244) (("}") reduce 244) (("~") reduce 244)) 27294 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CASE) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DEFAULT) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_EMPTY) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_PROPERTY_C) reduce 146) ((T_READONLY) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_VOID_CAST) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("}") reduce 146) (("~") reduce 146)) 27296 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CASE) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DEFAULT) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_EMPTY) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_VOID_CAST) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("}") reduce 168) (("~") reduce 168)) 27298 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CASE) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DEFAULT) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_EMPTY) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_VOID_CAST) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("}") reduce 157) (("~") reduce 157)) 27308 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CASE) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DEFAULT) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_EMPTY) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_VOID_CAST) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("}") reduce 164) (("~") reduce 164)) 27309 ((("!") reduce 220) (("\"") reduce 220) (("$") reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CASE) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DEFAULT) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_EMPTY) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_VARIABLE) reduce 220) ((T_VOID_CAST) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("}") reduce 220) (("~") reduce 220)) 27312 ((("!") reduce 221) (("\"") reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CASE) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DEFAULT) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_EMPTY) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_VARIABLE) reduce 221) ((T_VOID_CAST) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("}") reduce 221) (("~") reduce 221)) 27313 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CASE) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DEFAULT) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_EMPTY) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_VOID_CAST) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("}") reduce 163) (("~") reduce 163)) 27322 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CASE) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DEFAULT) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_EMPTY) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_VOID_CAST) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("}") reduce 152) (("~") reduce 152)) 27323 ((("!") reduce 218) (("\"") reduce 218) (("$") reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CASE) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DEFAULT) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_EMPTY) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_VARIABLE) reduce 218) ((T_VOID_CAST) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("}") reduce 218) (("~") reduce 218)) 27326 ((("!") reduce 219) (("\"") reduce 219) (("$") reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CASE) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DEFAULT) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_EMPTY) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_VARIABLE) reduce 219) ((T_VOID_CAST) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("}") reduce 219) (("~") reduce 219)) 27333 ((("!") reduce 202) (("\"") reduce 202) (("$") reduce 202) (("(") reduce 202) (("+") reduce 202) (("-") reduce 202) ((";") reduce 202) (("@") reduce 202) ((T_ABSTRACT) reduce 202) ((T_ARRAY) reduce 202) ((T_ARRAY_CAST) reduce 202) ((T_ATTRIBUTE) reduce 202) ((T_BOOL_CAST) reduce 202) ((T_BREAK) reduce 202) ((T_CASE) reduce 202) ((T_CLASS) reduce 202) ((T_CLASS_C) reduce 202) ((T_CLONE) reduce 202) ((T_CONSTANT_ENCAPSED_STRING) reduce 202) ((T_CONTINUE) reduce 202) ((T_DEC) reduce 202) ((T_DECLARE) reduce 202) ((T_DEFAULT) reduce 202) ((T_DIR) reduce 202) ((T_DNUMBER) reduce 202) ((T_DO) reduce 202) ((T_DOUBLE_CAST) reduce 202) ((T_ECHO) reduce 202) ((T_EMPTY) reduce 202) ((T_ENUM) reduce 202) ((T_EVAL) reduce 202) ((T_EXIT) reduce 202) ((T_FILE) reduce 202) ((T_FINAL) reduce 202) ((T_FN) reduce 202) ((T_FOR) reduce 202) ((T_FOREACH) reduce 202) ((T_FUNCTION) reduce 202) ((T_FUNC_C) reduce 202) ((T_GLOBAL) reduce 202) ((T_GOTO) reduce 202) ((T_HALT_COMPILER) reduce 202) ((T_IF) reduce 202) ((T_INC) reduce 202) ((T_INCLUDE) reduce 202) ((T_INCLUDE_ONCE) reduce 202) ((T_INLINE_HTML) reduce 202) ((T_INTERFACE) reduce 202) ((T_INT_CAST) reduce 202) ((T_ISSET) reduce 202) ((T_LINE) reduce 202) ((T_LIST) reduce 202) ((T_LNUMBER) reduce 202) ((T_MATCH) reduce 202) ((T_METHOD_C) reduce 202) ((T_NAME_FULLY_QUALIFIED) reduce 202) ((T_NAME_QUALIFIED) reduce 202) ((T_NAME_RELATIVE) reduce 202) ((T_NEW) reduce 202) ((T_NS_C) reduce 202) ((T_OBJECT_CAST) reduce 202) ((T_PRINT) reduce 202) ((T_PROPERTY_C) reduce 202) ((T_READONLY) reduce 202) ((T_REQUIRE) reduce 202) ((T_REQUIRE_ONCE) reduce 202) ((T_RETURN) reduce 202) ((T_START_HEREDOC) reduce 202) ((T_STATIC) reduce 202) ((T_STRING) reduce 202) ((T_STRING_CAST) reduce 202) ((T_SWITCH) reduce 202) ((T_THROW) reduce 202) ((T_TRAIT) reduce 202) ((T_TRAIT_C) reduce 202) ((T_TRY) reduce 202) ((T_UNSET) reduce 202) ((T_UNSET_CAST) reduce 202) ((T_VARIABLE) reduce 202) ((T_VOID_CAST) reduce 202) ((T_WHILE) reduce 202) ((T_YIELD) reduce 202) ((T_YIELD_FROM) reduce 202) (("[") reduce 202) (("`") reduce 202) (("{") reduce 202) (("}") reduce 202) (("~") reduce 202)) 27335 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CASE) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DEFAULT) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_EMPTY) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_VOID_CAST) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("}") reduce 159) (("~") reduce 159)) 27341 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CASE) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DEFAULT) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_EMPTY) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_VOID_CAST) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("}") reduce 151) (("~") reduce 151)) 27346 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CASE) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DEFAULT) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_EMPTY) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_VOID_CAST) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("}") reduce 165) (("~") reduce 165)) 27347 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CASE) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DEFAULT) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_EMPTY) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_VARIABLE) reduce 222) ((T_VOID_CAST) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("}") reduce 222) (("~") reduce 222)) 27350 ((("!") reduce 223) (("\"") reduce 223) (("$") reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CASE) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DEFAULT) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_EMPTY) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_VARIABLE) reduce 223) ((T_VOID_CAST) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("}") reduce 223) (("~") reduce 223)) 27352 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CASE) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DEFAULT) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_EMPTY) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_VOID_CAST) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("}") reduce 155) (("~") reduce 155)) 27359 ((("!") reduce 190) (("\"") reduce 190) (("$") reduce 190) (("(") reduce 190) (("+") reduce 190) (("-") reduce 190) ((";") reduce 190) (("@") reduce 190) ((T_ABSTRACT) reduce 190) ((T_ARRAY) reduce 190) ((T_ARRAY_CAST) reduce 190) ((T_ATTRIBUTE) reduce 190) ((T_BOOL_CAST) reduce 190) ((T_BREAK) reduce 190) ((T_CASE) reduce 190) ((T_CLASS) reduce 190) ((T_CLASS_C) reduce 190) ((T_CLONE) reduce 190) ((T_CONSTANT_ENCAPSED_STRING) reduce 190) ((T_CONTINUE) reduce 190) ((T_DEC) reduce 190) ((T_DECLARE) reduce 190) ((T_DEFAULT) reduce 190) ((T_DIR) reduce 190) ((T_DNUMBER) reduce 190) ((T_DO) reduce 190) ((T_DOUBLE_CAST) reduce 190) ((T_ECHO) reduce 190) ((T_EMPTY) reduce 190) ((T_ENUM) reduce 190) ((T_EVAL) reduce 190) ((T_EXIT) reduce 190) ((T_FILE) reduce 190) ((T_FINAL) reduce 190) ((T_FN) reduce 190) ((T_FOR) reduce 190) ((T_FOREACH) reduce 190) ((T_FUNCTION) reduce 190) ((T_FUNC_C) reduce 190) ((T_GLOBAL) reduce 190) ((T_GOTO) reduce 190) ((T_HALT_COMPILER) reduce 190) ((T_IF) reduce 190) ((T_INC) reduce 190) ((T_INCLUDE) reduce 190) ((T_INCLUDE_ONCE) reduce 190) ((T_INLINE_HTML) reduce 190) ((T_INTERFACE) reduce 190) ((T_INT_CAST) reduce 190) ((T_ISSET) reduce 190) ((T_LINE) reduce 190) ((T_LIST) reduce 190) ((T_LNUMBER) reduce 190) ((T_MATCH) reduce 190) ((T_METHOD_C) reduce 190) ((T_NAME_FULLY_QUALIFIED) reduce 190) ((T_NAME_QUALIFIED) reduce 190) ((T_NAME_RELATIVE) reduce 190) ((T_NEW) reduce 190) ((T_NS_C) reduce 190) ((T_OBJECT_CAST) reduce 190) ((T_PRINT) reduce 190) ((T_PROPERTY_C) reduce 190) ((T_READONLY) reduce 190) ((T_REQUIRE) reduce 190) ((T_REQUIRE_ONCE) reduce 190) ((T_RETURN) reduce 190) ((T_START_HEREDOC) reduce 190) ((T_STATIC) reduce 190) ((T_STRING) reduce 190) ((T_STRING_CAST) reduce 190) ((T_SWITCH) reduce 190) ((T_THROW) reduce 190) ((T_TRAIT) reduce 190) ((T_TRAIT_C) reduce 190) ((T_TRY) reduce 190) ((T_UNSET) reduce 190) ((T_UNSET_CAST) reduce 190) ((T_VARIABLE) reduce 190) ((T_VOID_CAST) reduce 190) ((T_WHILE) reduce 190) ((T_YIELD) reduce 190) ((T_YIELD_FROM) reduce 190) (("[") reduce 190) (("`") reduce 190) (("{") reduce 190) (("}") reduce 190) (("~") reduce 190)) 27361 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CASE) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DEFAULT) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_EMPTY) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_VOID_CAST) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("}") reduce 154) (("~") reduce 154)) 27362 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CASE) reduce 231) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DEFAULT) reduce 231) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_VOID_CAST) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("}") reduce 231) (("~") shift)) 27366 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CASE) reduce 230) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DEFAULT) reduce 230) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_VOID_CAST) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("}") reduce 230) (("~") shift)) 27367 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CASE) reduce 229) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DEFAULT) reduce 229) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_VOID_CAST) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("}") reduce 229) (("~") shift)) 27369 ((("!") reduce 225) (("\"") reduce 225) (("$") reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_ELSE) reduce 225) ((T_ELSEIF) reduce 225) ((T_EMPTY) reduce 225) ((T_ENDWHILE) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_VARIABLE) reduce 225) ((T_VOID_CAST) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("~") reduce 225)) 27373 ((("!") reduce 226) (("\"") reduce 226) (("$") reduce 226) (("(") reduce 226) (("+") reduce 226) (("-") reduce 226) ((";") reduce 226) (("@") reduce 226) ((T_ABSTRACT) reduce 226) ((T_ARRAY) reduce 226) ((T_ARRAY_CAST) reduce 226) ((T_ATTRIBUTE) reduce 226) ((T_BOOL_CAST) reduce 226) ((T_BREAK) reduce 226) ((T_CLASS) reduce 226) ((T_CLASS_C) reduce 226) ((T_CLONE) reduce 226) ((T_CONSTANT_ENCAPSED_STRING) reduce 226) ((T_CONTINUE) reduce 226) ((T_DEC) reduce 226) ((T_DECLARE) reduce 226) ((T_DIR) reduce 226) ((T_DNUMBER) reduce 226) ((T_DO) reduce 226) ((T_DOUBLE_CAST) reduce 226) ((T_ECHO) reduce 226) ((T_ELSE) reduce 226) ((T_ELSEIF) reduce 226) ((T_EMPTY) reduce 226) ((T_ENDWHILE) reduce 226) ((T_ENUM) reduce 226) ((T_EVAL) reduce 226) ((T_EXIT) reduce 226) ((T_FILE) reduce 226) ((T_FINAL) reduce 226) ((T_FN) reduce 226) ((T_FOR) reduce 226) ((T_FOREACH) reduce 226) ((T_FUNCTION) reduce 226) ((T_FUNC_C) reduce 226) ((T_GLOBAL) reduce 226) ((T_GOTO) reduce 226) ((T_HALT_COMPILER) reduce 226) ((T_IF) reduce 226) ((T_INC) reduce 226) ((T_INCLUDE) reduce 226) ((T_INCLUDE_ONCE) reduce 226) ((T_INLINE_HTML) reduce 226) ((T_INTERFACE) reduce 226) ((T_INT_CAST) reduce 226) ((T_ISSET) reduce 226) ((T_LINE) reduce 226) ((T_LIST) reduce 226) ((T_LNUMBER) reduce 226) ((T_MATCH) reduce 226) ((T_METHOD_C) reduce 226) ((T_NAME_FULLY_QUALIFIED) reduce 226) ((T_NAME_QUALIFIED) reduce 226) ((T_NAME_RELATIVE) reduce 226) ((T_NEW) reduce 226) ((T_NS_C) reduce 226) ((T_OBJECT_CAST) reduce 226) ((T_PRINT) reduce 226) ((T_PROPERTY_C) reduce 226) ((T_READONLY) reduce 226) ((T_REQUIRE) reduce 226) ((T_REQUIRE_ONCE) reduce 226) ((T_RETURN) reduce 226) ((T_START_HEREDOC) reduce 226) ((T_STATIC) reduce 226) ((T_STRING) reduce 226) ((T_STRING_CAST) reduce 226) ((T_SWITCH) reduce 226) ((T_THROW) reduce 226) ((T_TRAIT) reduce 226) ((T_TRAIT_C) reduce 226) ((T_TRY) reduce 226) ((T_UNSET) reduce 226) ((T_UNSET_CAST) reduce 226) ((T_VARIABLE) reduce 226) ((T_VOID_CAST) reduce 226) ((T_WHILE) reduce 226) ((T_YIELD) reduce 226) ((T_YIELD_FROM) reduce 226) (("[") reduce 226) (("`") reduce 226) (("{") reduce 226) (("~") reduce 226)) 27376 ((("!") reduce 227) (("\"") reduce 227) (("$") reduce 227) (("(") reduce 227) (("+") reduce 227) (("-") reduce 227) ((";") reduce 227) (("@") reduce 227) ((T_ABSTRACT) reduce 227) ((T_ARRAY) reduce 227) ((T_ARRAY_CAST) reduce 227) ((T_ATTRIBUTE) reduce 227) ((T_BOOL_CAST) reduce 227) ((T_BREAK) reduce 227) ((T_CLASS) reduce 227) ((T_CLASS_C) reduce 227) ((T_CLONE) reduce 227) ((T_CONSTANT_ENCAPSED_STRING) reduce 227) ((T_CONTINUE) reduce 227) ((T_DEC) reduce 227) ((T_DECLARE) reduce 227) ((T_DIR) reduce 227) ((T_DNUMBER) reduce 227) ((T_DO) reduce 227) ((T_DOUBLE_CAST) reduce 227) ((T_ECHO) reduce 227) ((T_ELSE) reduce 227) ((T_ELSEIF) reduce 227) ((T_EMPTY) reduce 227) ((T_ENDWHILE) reduce 227) ((T_ENUM) reduce 227) ((T_EVAL) reduce 227) ((T_EXIT) reduce 227) ((T_FILE) reduce 227) ((T_FINAL) reduce 227) ((T_FN) reduce 227) ((T_FOR) reduce 227) ((T_FOREACH) reduce 227) ((T_FUNCTION) reduce 227) ((T_FUNC_C) reduce 227) ((T_GLOBAL) reduce 227) ((T_GOTO) reduce 227) ((T_HALT_COMPILER) reduce 227) ((T_IF) reduce 227) ((T_INC) reduce 227) ((T_INCLUDE) reduce 227) ((T_INCLUDE_ONCE) reduce 227) ((T_INLINE_HTML) reduce 227) ((T_INTERFACE) reduce 227) ((T_INT_CAST) reduce 227) ((T_ISSET) reduce 227) ((T_LINE) reduce 227) ((T_LIST) reduce 227) ((T_LNUMBER) reduce 227) ((T_MATCH) reduce 227) ((T_METHOD_C) reduce 227) ((T_NAME_FULLY_QUALIFIED) reduce 227) ((T_NAME_QUALIFIED) reduce 227) ((T_NAME_RELATIVE) reduce 227) ((T_NEW) reduce 227) ((T_NS_C) reduce 227) ((T_OBJECT_CAST) reduce 227) ((T_PRINT) reduce 227) ((T_PROPERTY_C) reduce 227) ((T_READONLY) reduce 227) ((T_REQUIRE) reduce 227) ((T_REQUIRE_ONCE) reduce 227) ((T_RETURN) reduce 227) ((T_START_HEREDOC) reduce 227) ((T_STATIC) reduce 227) ((T_STRING) reduce 227) ((T_STRING_CAST) reduce 227) ((T_SWITCH) reduce 227) ((T_THROW) reduce 227) ((T_TRAIT) reduce 227) ((T_TRAIT_C) reduce 227) ((T_TRY) reduce 227) ((T_UNSET) reduce 227) ((T_UNSET_CAST) reduce 227) ((T_VARIABLE) reduce 227) ((T_VOID_CAST) reduce 227) ((T_WHILE) reduce 227) ((T_YIELD) reduce 227) ((T_YIELD_FROM) reduce 227) (("[") reduce 227) (("`") reduce 227) (("{") reduce 227) (("~") reduce 227)) 27377 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_ELSE) reduce 169) ((T_ELSEIF) reduce 169) ((T_EMPTY) reduce 169) ((T_ENDWHILE) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_VOID_CAST) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 27379 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_ELSE) reduce 158) ((T_ELSEIF) reduce 158) ((T_EMPTY) reduce 158) ((T_ENDWHILE) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_VOID_CAST) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 27381 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_ELSE) reduce 156) ((T_ELSEIF) reduce 156) ((T_EMPTY) reduce 156) ((T_ENDWHILE) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_VOID_CAST) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 27383 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_ELSE) reduce 168) ((T_ELSEIF) reduce 168) ((T_EMPTY) reduce 168) ((T_ENDWHILE) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_VOID_CAST) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 27385 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_ELSE) reduce 157) ((T_ELSEIF) reduce 157) ((T_EMPTY) reduce 157) ((T_ENDWHILE) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_VOID_CAST) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 27395 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_ELSE) reduce 164) ((T_ELSEIF) reduce 164) ((T_EMPTY) reduce 164) ((T_ENDWHILE) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_VOID_CAST) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 27396 ((("!") reduce 220) (("\"") reduce 220) (("$") reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_ELSE) reduce 220) ((T_ELSEIF) reduce 220) ((T_EMPTY) reduce 220) ((T_ENDWHILE) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_VARIABLE) reduce 220) ((T_VOID_CAST) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("~") reduce 220)) 27399 ((("!") reduce 221) (("\"") reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_ELSE) reduce 221) ((T_ELSEIF) reduce 221) ((T_EMPTY) reduce 221) ((T_ENDWHILE) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_VARIABLE) reduce 221) ((T_VOID_CAST) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("~") reduce 221)) 27400 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_ELSE) reduce 163) ((T_ELSEIF) reduce 163) ((T_EMPTY) reduce 163) ((T_ENDWHILE) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_VOID_CAST) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 27409 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_ELSE) reduce 152) ((T_ELSEIF) reduce 152) ((T_EMPTY) reduce 152) ((T_ENDWHILE) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_VOID_CAST) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 27410 ((("!") reduce 218) (("\"") reduce 218) (("$") reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_ELSE) reduce 218) ((T_ELSEIF) reduce 218) ((T_EMPTY) reduce 218) ((T_ENDWHILE) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_VARIABLE) reduce 218) ((T_VOID_CAST) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("~") reduce 218)) 27413 ((("!") reduce 219) (("\"") reduce 219) (("$") reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_ELSE) reduce 219) ((T_ELSEIF) reduce 219) ((T_EMPTY) reduce 219) ((T_ENDWHILE) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_VARIABLE) reduce 219) ((T_VOID_CAST) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("~") reduce 219)) 27415 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_ELSE) reduce 159) ((T_ELSEIF) reduce 159) ((T_EMPTY) reduce 159) ((T_ENDWHILE) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_VOID_CAST) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 27421 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_ELSE) reduce 151) ((T_ELSEIF) reduce 151) ((T_EMPTY) reduce 151) ((T_ENDWHILE) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_VOID_CAST) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 27426 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_ELSE) reduce 165) ((T_ELSEIF) reduce 165) ((T_EMPTY) reduce 165) ((T_ENDWHILE) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_VOID_CAST) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 27427 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_ELSE) reduce 222) ((T_ELSEIF) reduce 222) ((T_EMPTY) reduce 222) ((T_ENDWHILE) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_VARIABLE) reduce 222) ((T_VOID_CAST) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("~") reduce 222)) 27430 ((("!") reduce 223) (("\"") reduce 223) (("$") reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_ELSE) reduce 223) ((T_ELSEIF) reduce 223) ((T_EMPTY) reduce 223) ((T_ENDWHILE) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_VARIABLE) reduce 223) ((T_VOID_CAST) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("~") reduce 223)) 27432 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_ELSE) reduce 155) ((T_ELSEIF) reduce 155) ((T_EMPTY) reduce 155) ((T_ENDWHILE) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_VOID_CAST) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 27434 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_ELSE) reduce 154) ((T_ELSEIF) reduce 154) ((T_EMPTY) reduce 154) ((T_ENDWHILE) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_VOID_CAST) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 27435 ((("!") reduce 247) (("\"") reduce 247) (("$") reduce 247) (("(") reduce 247) (("+") reduce 247) (("-") reduce 247) ((";") reduce 247) (("@") reduce 247) ((T_ABSTRACT) reduce 247) ((T_ARRAY) reduce 247) ((T_ARRAY_CAST) reduce 247) ((T_ATTRIBUTE) reduce 247) ((T_BOOL_CAST) reduce 247) ((T_BREAK) reduce 247) ((T_CLASS) reduce 247) ((T_CLASS_C) reduce 247) ((T_CLONE) reduce 247) ((T_CONSTANT_ENCAPSED_STRING) reduce 247) ((T_CONTINUE) reduce 247) ((T_DEC) reduce 247) ((T_DECLARE) reduce 247) ((T_DIR) reduce 247) ((T_DNUMBER) reduce 247) ((T_DO) reduce 247) ((T_DOUBLE_CAST) reduce 247) ((T_ECHO) reduce 247) ((T_EMPTY) reduce 247) ((T_ENDWHILE) reduce 247) ((T_ENUM) reduce 247) ((T_EVAL) reduce 247) ((T_EXIT) reduce 247) ((T_FILE) reduce 247) ((T_FINAL) reduce 247) ((T_FN) reduce 247) ((T_FOR) reduce 247) ((T_FOREACH) reduce 247) ((T_FUNCTION) reduce 247) ((T_FUNC_C) reduce 247) ((T_GLOBAL) reduce 247) ((T_GOTO) reduce 247) ((T_HALT_COMPILER) reduce 247) ((T_IF) reduce 247) ((T_INC) reduce 247) ((T_INCLUDE) reduce 247) ((T_INCLUDE_ONCE) reduce 247) ((T_INLINE_HTML) reduce 247) ((T_INTERFACE) reduce 247) ((T_INT_CAST) reduce 247) ((T_ISSET) reduce 247) ((T_LINE) reduce 247) ((T_LIST) reduce 247) ((T_LNUMBER) reduce 247) ((T_MATCH) reduce 247) ((T_METHOD_C) reduce 247) ((T_NAME_FULLY_QUALIFIED) reduce 247) ((T_NAME_QUALIFIED) reduce 247) ((T_NAME_RELATIVE) reduce 247) ((T_NEW) reduce 247) ((T_NS_C) reduce 247) ((T_OBJECT_CAST) reduce 247) ((T_PRINT) reduce 247) ((T_PROPERTY_C) reduce 247) ((T_READONLY) reduce 247) ((T_REQUIRE) reduce 247) ((T_REQUIRE_ONCE) reduce 247) ((T_RETURN) reduce 247) ((T_START_HEREDOC) reduce 247) ((T_STATIC) reduce 247) ((T_STRING) reduce 247) ((T_STRING_CAST) reduce 247) ((T_SWITCH) reduce 247) ((T_THROW) reduce 247) ((T_TRAIT) reduce 247) ((T_TRAIT_C) reduce 247) ((T_TRY) reduce 247) ((T_UNSET) reduce 247) ((T_UNSET_CAST) reduce 247) ((T_VARIABLE) reduce 247) ((T_VOID_CAST) reduce 247) ((T_WHILE) reduce 247) ((T_YIELD) reduce 247) ((T_YIELD_FROM) reduce 247) (("[") reduce 247) (("`") reduce 247) (("{") reduce 247) (("~") reduce 247)) 27446 ((("!") reduce 494) (("\"") reduce 494) (("$") reduce 494) (("(") reduce 494) (("+") reduce 494) (("-") reduce 494) ((";") reduce 494) (("@") reduce 494) ((T_ABSTRACT) reduce 494) ((T_ARRAY) reduce 494) ((T_ARRAY_CAST) reduce 494) ((T_ATTRIBUTE) reduce 494) ((T_BOOL_CAST) reduce 494) ((T_BREAK) reduce 494) ((T_CLASS) reduce 494) ((T_CLASS_C) reduce 494) ((T_CLONE) reduce 494) ((T_CONSTANT_ENCAPSED_STRING) reduce 494) ((T_CONTINUE) reduce 494) ((T_DEC) reduce 494) ((T_DECLARE) reduce 494) ((T_DIR) reduce 494) ((T_DNUMBER) reduce 494) ((T_DO) reduce 494) ((T_DOUBLE_CAST) reduce 494) ((T_ECHO) reduce 494) ((T_EMPTY) reduce 494) ((T_ENDWHILE) reduce 494) ((T_ENUM) reduce 494) ((T_EVAL) reduce 494) ((T_EXIT) reduce 494) ((T_FILE) reduce 494) ((T_FINAL) reduce 494) ((T_FN) reduce 494) ((T_FOR) reduce 494) ((T_FOREACH) reduce 494) ((T_FUNCTION) reduce 494) ((T_FUNC_C) reduce 494) ((T_GLOBAL) reduce 494) ((T_GOTO) reduce 494) ((T_HALT_COMPILER) reduce 494) ((T_IF) reduce 494) ((T_INC) reduce 494) ((T_INCLUDE) reduce 494) ((T_INCLUDE_ONCE) reduce 494) ((T_INLINE_HTML) reduce 494) ((T_INTERFACE) reduce 494) ((T_INT_CAST) reduce 494) ((T_ISSET) reduce 494) ((T_LINE) reduce 494) ((T_LIST) reduce 494) ((T_LNUMBER) reduce 494) ((T_MATCH) reduce 494) ((T_METHOD_C) reduce 494) ((T_NAME_FULLY_QUALIFIED) reduce 494) ((T_NAME_QUALIFIED) reduce 494) ((T_NAME_RELATIVE) reduce 494) ((T_NEW) reduce 494) ((T_NS_C) reduce 494) ((T_OBJECT_CAST) reduce 494) ((T_PRINT) reduce 494) ((T_PROPERTY_C) reduce 494) ((T_READONLY) reduce 494) ((T_REQUIRE) reduce 494) ((T_REQUIRE_ONCE) reduce 494) ((T_RETURN) reduce 494) ((T_START_HEREDOC) reduce 494) ((T_STATIC) reduce 494) ((T_STRING) reduce 494) ((T_STRING_CAST) reduce 494) ((T_SWITCH) reduce 494) ((T_THROW) reduce 494) ((T_TRAIT) reduce 494) ((T_TRAIT_C) reduce 494) ((T_TRY) reduce 494) ((T_UNSET) reduce 494) ((T_UNSET_CAST) reduce 494) ((T_VARIABLE) reduce 494) ((T_VOID_CAST) reduce 494) ((T_WHILE) reduce 494) ((T_YIELD) reduce 494) ((T_YIELD_FROM) reduce 494) (("[") reduce 494) (("`") reduce 494) (("{") reduce 494) (("~") reduce 494)) 27447 ((("!") reduce 184) (("\"") reduce 184) (("$") reduce 184) (("(") reduce 184) (("+") reduce 184) (("-") reduce 184) ((";") reduce 184) (("@") reduce 184) ((T_ABSTRACT) reduce 184) ((T_ARRAY) reduce 184) ((T_ARRAY_CAST) reduce 184) ((T_ATTRIBUTE) reduce 184) ((T_BOOL_CAST) reduce 184) ((T_BREAK) reduce 184) ((T_CLASS) reduce 184) ((T_CLASS_C) reduce 184) ((T_CLONE) reduce 184) ((T_CONSTANT_ENCAPSED_STRING) reduce 184) ((T_CONTINUE) reduce 184) ((T_DEC) reduce 184) ((T_DECLARE) reduce 184) ((T_DIR) reduce 184) ((T_DNUMBER) reduce 184) ((T_DO) reduce 184) ((T_DOUBLE_CAST) reduce 184) ((T_ECHO) reduce 184) ((T_EMPTY) reduce 184) ((T_ENDWHILE) reduce 184) ((T_ENUM) reduce 184) ((T_EVAL) reduce 184) ((T_EXIT) reduce 184) ((T_FILE) reduce 184) ((T_FINAL) reduce 184) ((T_FN) reduce 184) ((T_FOR) reduce 184) ((T_FOREACH) reduce 184) ((T_FUNCTION) reduce 184) ((T_FUNC_C) reduce 184) ((T_GLOBAL) reduce 184) ((T_GOTO) reduce 184) ((T_HALT_COMPILER) reduce 184) ((T_IF) reduce 184) ((T_INC) reduce 184) ((T_INCLUDE) reduce 184) ((T_INCLUDE_ONCE) reduce 184) ((T_INLINE_HTML) reduce 184) ((T_INTERFACE) reduce 184) ((T_INT_CAST) reduce 184) ((T_ISSET) reduce 184) ((T_LINE) reduce 184) ((T_LIST) reduce 184) ((T_LNUMBER) reduce 184) ((T_MATCH) reduce 184) ((T_METHOD_C) reduce 184) ((T_NAME_FULLY_QUALIFIED) reduce 184) ((T_NAME_QUALIFIED) reduce 184) ((T_NAME_RELATIVE) reduce 184) ((T_NEW) reduce 184) ((T_NS_C) reduce 184) ((T_OBJECT_CAST) reduce 184) ((T_PRINT) reduce 184) ((T_PROPERTY_C) reduce 184) ((T_READONLY) reduce 184) ((T_REQUIRE) reduce 184) ((T_REQUIRE_ONCE) reduce 184) ((T_RETURN) reduce 184) ((T_START_HEREDOC) reduce 184) ((T_STATIC) reduce 184) ((T_STRING) reduce 184) ((T_STRING_CAST) reduce 184) ((T_SWITCH) reduce 184) ((T_THROW) reduce 184) ((T_TRAIT) reduce 184) ((T_TRAIT_C) reduce 184) ((T_TRY) reduce 184) ((T_UNSET) reduce 184) ((T_UNSET_CAST) reduce 184) ((T_VARIABLE) reduce 184) ((T_VOID_CAST) reduce 184) ((T_WHILE) reduce 184) ((T_YIELD) reduce 184) ((T_YIELD_FROM) reduce 184) (("[") reduce 184) (("`") reduce 184) (("{") reduce 184) (("~") reduce 184)) 27448 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_EMPTY) reduce 161) ((T_ENDWHILE) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_VOID_CAST) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 27456 ((("!") reduce 189) (("\"") reduce 189) (("$") reduce 189) (("(") reduce 189) (("+") reduce 189) (("-") reduce 189) ((";") reduce 189) (("@") reduce 189) ((T_ABSTRACT) reduce 189) ((T_ARRAY) reduce 189) ((T_ARRAY_CAST) reduce 189) ((T_ATTRIBUTE) reduce 189) ((T_BOOL_CAST) reduce 189) ((T_BREAK) reduce 189) ((T_CLASS) reduce 189) ((T_CLASS_C) reduce 189) ((T_CLONE) reduce 189) ((T_CONSTANT_ENCAPSED_STRING) reduce 189) ((T_CONTINUE) reduce 189) ((T_DEC) reduce 189) ((T_DECLARE) reduce 189) ((T_DIR) reduce 189) ((T_DNUMBER) reduce 189) ((T_DO) reduce 189) ((T_DOUBLE_CAST) reduce 189) ((T_ECHO) reduce 189) ((T_EMPTY) reduce 189) ((T_ENDWHILE) reduce 189) ((T_ENUM) reduce 189) ((T_EVAL) reduce 189) ((T_EXIT) reduce 189) ((T_FILE) reduce 189) ((T_FINAL) reduce 189) ((T_FN) reduce 189) ((T_FOR) reduce 189) ((T_FOREACH) reduce 189) ((T_FUNCTION) reduce 189) ((T_FUNC_C) reduce 189) ((T_GLOBAL) reduce 189) ((T_GOTO) reduce 189) ((T_HALT_COMPILER) reduce 189) ((T_IF) reduce 189) ((T_INC) reduce 189) ((T_INCLUDE) reduce 189) ((T_INCLUDE_ONCE) reduce 189) ((T_INLINE_HTML) reduce 189) ((T_INTERFACE) reduce 189) ((T_INT_CAST) reduce 189) ((T_ISSET) reduce 189) ((T_LINE) reduce 189) ((T_LIST) reduce 189) ((T_LNUMBER) reduce 189) ((T_MATCH) reduce 189) ((T_METHOD_C) reduce 189) ((T_NAME_FULLY_QUALIFIED) reduce 189) ((T_NAME_QUALIFIED) reduce 189) ((T_NAME_RELATIVE) reduce 189) ((T_NEW) reduce 189) ((T_NS_C) reduce 189) ((T_OBJECT_CAST) reduce 189) ((T_PRINT) reduce 189) ((T_PROPERTY_C) reduce 189) ((T_READONLY) reduce 189) ((T_REQUIRE) reduce 189) ((T_REQUIRE_ONCE) reduce 189) ((T_RETURN) reduce 189) ((T_START_HEREDOC) reduce 189) ((T_STATIC) reduce 189) ((T_STRING) reduce 189) ((T_STRING_CAST) reduce 189) ((T_SWITCH) reduce 189) ((T_THROW) reduce 189) ((T_TRAIT) reduce 189) ((T_TRAIT_C) reduce 189) ((T_TRY) reduce 189) ((T_UNSET) reduce 189) ((T_UNSET_CAST) reduce 189) ((T_VARIABLE) reduce 189) ((T_VOID_CAST) reduce 189) ((T_WHILE) reduce 189) ((T_YIELD) reduce 189) ((T_YIELD_FROM) reduce 189) (("[") reduce 189) (("`") reduce 189) (("{") reduce 189) (("~") reduce 189)) 27457 ((("!") reduce 145) (("\"") reduce 145) (("$") reduce 145) (("(") reduce 145) (("+") reduce 145) (("-") reduce 145) ((";") reduce 145) (("@") reduce 145) ((T_ABSTRACT) reduce 145) ((T_ARRAY) reduce 145) ((T_ARRAY_CAST) reduce 145) ((T_ATTRIBUTE) reduce 145) ((T_BOOL_CAST) reduce 145) ((T_BREAK) reduce 145) ((T_CLASS) reduce 145) ((T_CLASS_C) reduce 145) ((T_CLONE) reduce 145) ((T_CONSTANT_ENCAPSED_STRING) reduce 145) ((T_CONTINUE) reduce 145) ((T_DEC) reduce 145) ((T_DECLARE) reduce 145) ((T_DIR) reduce 145) ((T_DNUMBER) reduce 145) ((T_DO) reduce 145) ((T_DOUBLE_CAST) reduce 145) ((T_ECHO) reduce 145) ((T_EMPTY) reduce 145) ((T_ENDWHILE) reduce 145) ((T_ENUM) reduce 145) ((T_EVAL) reduce 145) ((T_EXIT) reduce 145) ((T_FILE) reduce 145) ((T_FINAL) reduce 145) ((T_FN) reduce 145) ((T_FOR) reduce 145) ((T_FOREACH) reduce 145) ((T_FUNCTION) reduce 145) ((T_FUNC_C) reduce 145) ((T_GLOBAL) reduce 145) ((T_GOTO) reduce 145) ((T_HALT_COMPILER) reduce 145) ((T_IF) reduce 145) ((T_INC) reduce 145) ((T_INCLUDE) reduce 145) ((T_INCLUDE_ONCE) reduce 145) ((T_INLINE_HTML) reduce 145) ((T_INTERFACE) reduce 145) ((T_INT_CAST) reduce 145) ((T_ISSET) reduce 145) ((T_LINE) reduce 145) ((T_LIST) reduce 145) ((T_LNUMBER) reduce 145) ((T_MATCH) reduce 145) ((T_METHOD_C) reduce 145) ((T_NAME_FULLY_QUALIFIED) reduce 145) ((T_NAME_QUALIFIED) reduce 145) ((T_NAME_RELATIVE) reduce 145) ((T_NEW) reduce 145) ((T_NS_C) reduce 145) ((T_OBJECT_CAST) reduce 145) ((T_PRINT) reduce 145) ((T_PROPERTY_C) reduce 145) ((T_READONLY) reduce 145) ((T_REQUIRE) reduce 145) ((T_REQUIRE_ONCE) reduce 145) ((T_RETURN) reduce 145) ((T_START_HEREDOC) reduce 145) ((T_STATIC) reduce 145) ((T_STRING) reduce 145) ((T_STRING_CAST) reduce 145) ((T_SWITCH) reduce 145) ((T_THROW) reduce 145) ((T_TRAIT) reduce 145) ((T_TRAIT_C) reduce 145) ((T_TRY) reduce 145) ((T_UNSET) reduce 145) ((T_UNSET_CAST) reduce 145) ((T_VARIABLE) reduce 145) ((T_VOID_CAST) reduce 145) ((T_WHILE) reduce 145) ((T_YIELD) reduce 145) ((T_YIELD_FROM) reduce 145) (("[") reduce 145) (("`") reduce 145) (("{") reduce 145) (("~") reduce 145)) 27460 ((("!") reduce 250) (("\"") reduce 250) (("$") reduce 250) (("(") reduce 250) (("+") reduce 250) (("-") reduce 250) ((";") reduce 250) (("@") reduce 250) ((T_ABSTRACT) reduce 250) ((T_ARRAY) reduce 250) ((T_ARRAY_CAST) reduce 250) ((T_ATTRIBUTE) reduce 250) ((T_BOOL_CAST) reduce 250) ((T_BREAK) reduce 250) ((T_CLASS) reduce 250) ((T_CLASS_C) reduce 250) ((T_CLONE) reduce 250) ((T_CONSTANT_ENCAPSED_STRING) reduce 250) ((T_CONTINUE) reduce 250) ((T_DEC) reduce 250) ((T_DECLARE) reduce 250) ((T_DIR) reduce 250) ((T_DNUMBER) reduce 250) ((T_DO) reduce 250) ((T_DOUBLE_CAST) reduce 250) ((T_ECHO) reduce 250) ((T_EMPTY) reduce 250) ((T_ENDWHILE) reduce 250) ((T_ENUM) reduce 250) ((T_EVAL) reduce 250) ((T_EXIT) reduce 250) ((T_FILE) reduce 250) ((T_FINAL) reduce 250) ((T_FN) reduce 250) ((T_FOR) reduce 250) ((T_FOREACH) reduce 250) ((T_FUNCTION) reduce 250) ((T_FUNC_C) reduce 250) ((T_GLOBAL) reduce 250) ((T_GOTO) reduce 250) ((T_HALT_COMPILER) reduce 250) ((T_IF) reduce 250) ((T_INC) reduce 250) ((T_INCLUDE) reduce 250) ((T_INCLUDE_ONCE) reduce 250) ((T_INLINE_HTML) reduce 250) ((T_INTERFACE) reduce 250) ((T_INT_CAST) reduce 250) ((T_ISSET) reduce 250) ((T_LINE) reduce 250) ((T_LIST) reduce 250) ((T_LNUMBER) reduce 250) ((T_MATCH) reduce 250) ((T_METHOD_C) reduce 250) ((T_NAME_FULLY_QUALIFIED) reduce 250) ((T_NAME_QUALIFIED) reduce 250) ((T_NAME_RELATIVE) reduce 250) ((T_NEW) reduce 250) ((T_NS_C) reduce 250) ((T_OBJECT_CAST) reduce 250) ((T_PRINT) reduce 250) ((T_PROPERTY_C) reduce 250) ((T_READONLY) reduce 250) ((T_REQUIRE) reduce 250) ((T_REQUIRE_ONCE) reduce 250) ((T_RETURN) reduce 250) ((T_START_HEREDOC) reduce 250) ((T_STATIC) reduce 250) ((T_STRING) reduce 250) ((T_STRING_CAST) reduce 250) ((T_SWITCH) reduce 250) ((T_THROW) reduce 250) ((T_TRAIT) reduce 250) ((T_TRAIT_C) reduce 250) ((T_TRY) reduce 250) ((T_UNSET) reduce 250) ((T_UNSET_CAST) reduce 250) ((T_VARIABLE) reduce 250) ((T_VOID_CAST) reduce 250) ((T_WHILE) reduce 250) ((T_YIELD) reduce 250) ((T_YIELD_FROM) reduce 250) (("[") reduce 250) (("`") reduce 250) (("{") reduce 250) (("~") reduce 250)) 27464 ((("!") reduce 251) (("\"") reduce 251) (("$") reduce 251) (("(") reduce 251) (("+") reduce 251) (("-") reduce 251) ((";") reduce 251) (("@") reduce 251) ((T_ABSTRACT) reduce 251) ((T_ARRAY) reduce 251) ((T_ARRAY_CAST) reduce 251) ((T_ATTRIBUTE) reduce 251) ((T_BOOL_CAST) reduce 251) ((T_BREAK) reduce 251) ((T_CLASS) reduce 251) ((T_CLASS_C) reduce 251) ((T_CLONE) reduce 251) ((T_CONSTANT_ENCAPSED_STRING) reduce 251) ((T_CONTINUE) reduce 251) ((T_DEC) reduce 251) ((T_DECLARE) reduce 251) ((T_DIR) reduce 251) ((T_DNUMBER) reduce 251) ((T_DO) reduce 251) ((T_DOUBLE_CAST) reduce 251) ((T_ECHO) reduce 251) ((T_EMPTY) reduce 251) ((T_ENDWHILE) reduce 251) ((T_ENUM) reduce 251) ((T_EVAL) reduce 251) ((T_EXIT) reduce 251) ((T_FILE) reduce 251) ((T_FINAL) reduce 251) ((T_FN) reduce 251) ((T_FOR) reduce 251) ((T_FOREACH) reduce 251) ((T_FUNCTION) reduce 251) ((T_FUNC_C) reduce 251) ((T_GLOBAL) reduce 251) ((T_GOTO) reduce 251) ((T_HALT_COMPILER) reduce 251) ((T_IF) reduce 251) ((T_INC) reduce 251) ((T_INCLUDE) reduce 251) ((T_INCLUDE_ONCE) reduce 251) ((T_INLINE_HTML) reduce 251) ((T_INTERFACE) reduce 251) ((T_INT_CAST) reduce 251) ((T_ISSET) reduce 251) ((T_LINE) reduce 251) ((T_LIST) reduce 251) ((T_LNUMBER) reduce 251) ((T_MATCH) reduce 251) ((T_METHOD_C) reduce 251) ((T_NAME_FULLY_QUALIFIED) reduce 251) ((T_NAME_QUALIFIED) reduce 251) ((T_NAME_RELATIVE) reduce 251) ((T_NEW) reduce 251) ((T_NS_C) reduce 251) ((T_OBJECT_CAST) reduce 251) ((T_PRINT) reduce 251) ((T_PROPERTY_C) reduce 251) ((T_READONLY) reduce 251) ((T_REQUIRE) reduce 251) ((T_REQUIRE_ONCE) reduce 251) ((T_RETURN) reduce 251) ((T_START_HEREDOC) reduce 251) ((T_STATIC) reduce 251) ((T_STRING) reduce 251) ((T_STRING_CAST) reduce 251) ((T_SWITCH) reduce 251) ((T_THROW) reduce 251) ((T_TRAIT) reduce 251) ((T_TRAIT_C) reduce 251) ((T_TRY) reduce 251) ((T_UNSET) reduce 251) ((T_UNSET_CAST) reduce 251) ((T_VARIABLE) reduce 251) ((T_VOID_CAST) reduce 251) ((T_WHILE) reduce 251) ((T_YIELD) reduce 251) ((T_YIELD_FROM) reduce 251) (("[") reduce 251) (("`") reduce 251) (("{") reduce 251) (("~") reduce 251)) 27469 ((("!") reduce 242) (("\"") reduce 242) (("$") reduce 242) (("(") reduce 242) (("+") reduce 242) (("-") reduce 242) ((";") reduce 242) (("@") reduce 242) ((T_ABSTRACT) reduce 242) ((T_ARRAY) reduce 242) ((T_ARRAY_CAST) reduce 242) ((T_ATTRIBUTE) reduce 242) ((T_BOOL_CAST) reduce 242) ((T_BREAK) reduce 242) ((T_CLASS) reduce 242) ((T_CLASS_C) reduce 242) ((T_CLONE) reduce 242) ((T_CONSTANT_ENCAPSED_STRING) reduce 242) ((T_CONTINUE) reduce 242) ((T_DEC) reduce 242) ((T_DECLARE) reduce 242) ((T_DIR) reduce 242) ((T_DNUMBER) reduce 242) ((T_DO) reduce 242) ((T_DOUBLE_CAST) reduce 242) ((T_ECHO) reduce 242) ((T_EMPTY) reduce 242) ((T_ENDWHILE) reduce 242) ((T_ENUM) reduce 242) ((T_EVAL) reduce 242) ((T_EXIT) reduce 242) ((T_FILE) reduce 242) ((T_FINAL) reduce 242) ((T_FN) reduce 242) ((T_FOR) reduce 242) ((T_FOREACH) reduce 242) ((T_FUNCTION) reduce 242) ((T_FUNC_C) reduce 242) ((T_GLOBAL) reduce 242) ((T_GOTO) reduce 242) ((T_HALT_COMPILER) reduce 242) ((T_IF) reduce 242) ((T_INC) reduce 242) ((T_INCLUDE) reduce 242) ((T_INCLUDE_ONCE) reduce 242) ((T_INLINE_HTML) reduce 242) ((T_INTERFACE) reduce 242) ((T_INT_CAST) reduce 242) ((T_ISSET) reduce 242) ((T_LINE) reduce 242) ((T_LIST) reduce 242) ((T_LNUMBER) reduce 242) ((T_MATCH) reduce 242) ((T_METHOD_C) reduce 242) ((T_NAME_FULLY_QUALIFIED) reduce 242) ((T_NAME_QUALIFIED) reduce 242) ((T_NAME_RELATIVE) reduce 242) ((T_NEW) reduce 242) ((T_NS_C) reduce 242) ((T_OBJECT_CAST) reduce 242) ((T_PRINT) reduce 242) ((T_PROPERTY_C) reduce 242) ((T_READONLY) reduce 242) ((T_REQUIRE) reduce 242) ((T_REQUIRE_ONCE) reduce 242) ((T_RETURN) reduce 242) ((T_START_HEREDOC) reduce 242) ((T_STATIC) reduce 242) ((T_STRING) reduce 242) ((T_STRING_CAST) reduce 242) ((T_SWITCH) reduce 242) ((T_THROW) reduce 242) ((T_TRAIT) reduce 242) ((T_TRAIT_C) reduce 242) ((T_TRY) reduce 242) ((T_UNSET) reduce 242) ((T_UNSET_CAST) reduce 242) ((T_VARIABLE) reduce 242) ((T_VOID_CAST) reduce 242) ((T_WHILE) reduce 242) ((T_YIELD) reduce 242) ((T_YIELD_FROM) reduce 242) (("[") reduce 242) (("`") reduce 242) (("{") reduce 242) (("~") reduce 242)) 27470 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_EMPTY) reduce 150) ((T_ENDWHILE) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_VOID_CAST) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 27473 ((("!") reduce 243) (("\"") reduce 243) (("$") reduce 243) (("(") reduce 243) (("+") reduce 243) (("-") reduce 243) ((";") reduce 243) (("@") reduce 243) ((T_ABSTRACT) reduce 243) ((T_ARRAY) reduce 243) ((T_ARRAY_CAST) reduce 243) ((T_ATTRIBUTE) reduce 243) ((T_BOOL_CAST) reduce 243) ((T_BREAK) reduce 243) ((T_CLASS) reduce 243) ((T_CLASS_C) reduce 243) ((T_CLONE) reduce 243) ((T_CONSTANT_ENCAPSED_STRING) reduce 243) ((T_CONTINUE) reduce 243) ((T_DEC) reduce 243) ((T_DECLARE) reduce 243) ((T_DIR) reduce 243) ((T_DNUMBER) reduce 243) ((T_DO) reduce 243) ((T_DOUBLE_CAST) reduce 243) ((T_ECHO) reduce 243) ((T_EMPTY) reduce 243) ((T_ENDWHILE) reduce 243) ((T_ENUM) reduce 243) ((T_EVAL) reduce 243) ((T_EXIT) reduce 243) ((T_FILE) reduce 243) ((T_FINAL) reduce 243) ((T_FN) reduce 243) ((T_FOR) reduce 243) ((T_FOREACH) reduce 243) ((T_FUNCTION) reduce 243) ((T_FUNC_C) reduce 243) ((T_GLOBAL) reduce 243) ((T_GOTO) reduce 243) ((T_HALT_COMPILER) reduce 243) ((T_IF) reduce 243) ((T_INC) reduce 243) ((T_INCLUDE) reduce 243) ((T_INCLUDE_ONCE) reduce 243) ((T_INLINE_HTML) reduce 243) ((T_INTERFACE) reduce 243) ((T_INT_CAST) reduce 243) ((T_ISSET) reduce 243) ((T_LINE) reduce 243) ((T_LIST) reduce 243) ((T_LNUMBER) reduce 243) ((T_MATCH) reduce 243) ((T_METHOD_C) reduce 243) ((T_NAME_FULLY_QUALIFIED) reduce 243) ((T_NAME_QUALIFIED) reduce 243) ((T_NAME_RELATIVE) reduce 243) ((T_NEW) reduce 243) ((T_NS_C) reduce 243) ((T_OBJECT_CAST) reduce 243) ((T_PRINT) reduce 243) ((T_PROPERTY_C) reduce 243) ((T_READONLY) reduce 243) ((T_REQUIRE) reduce 243) ((T_REQUIRE_ONCE) reduce 243) ((T_RETURN) reduce 243) ((T_START_HEREDOC) reduce 243) ((T_STATIC) reduce 243) ((T_STRING) reduce 243) ((T_STRING_CAST) reduce 243) ((T_SWITCH) reduce 243) ((T_THROW) reduce 243) ((T_TRAIT) reduce 243) ((T_TRAIT_C) reduce 243) ((T_TRY) reduce 243) ((T_UNSET) reduce 243) ((T_UNSET_CAST) reduce 243) ((T_VARIABLE) reduce 243) ((T_VOID_CAST) reduce 243) ((T_WHILE) reduce 243) ((T_YIELD) reduce 243) ((T_YIELD_FROM) reduce 243) (("[") reduce 243) (("`") reduce 243) (("{") reduce 243) (("~") reduce 243)) 27475 ((("!") reduce 170) (("\"") reduce 170) (("$") reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_EMPTY) reduce 170) ((T_ENDWHILE) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_VARIABLE) reduce 170) ((T_VOID_CAST) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("~") reduce 170)) 27480 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_EMPTY) reduce 162) ((T_ENDWHILE) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_VOID_CAST) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 27483 ((("!") reduce 171) (("\"") reduce 171) (("$") reduce 171) (("(") reduce 171) (("+") reduce 171) (("-") reduce 171) ((";") reduce 171) (("@") reduce 171) ((T_ABSTRACT) reduce 171) ((T_ARRAY) reduce 171) ((T_ARRAY_CAST) reduce 171) ((T_ATTRIBUTE) reduce 171) ((T_BOOL_CAST) reduce 171) ((T_BREAK) reduce 171) ((T_CATCH) reduce 171) ((T_CLASS) reduce 171) ((T_CLASS_C) reduce 171) ((T_CLONE) reduce 171) ((T_CONSTANT_ENCAPSED_STRING) reduce 171) ((T_CONTINUE) reduce 171) ((T_DEC) reduce 171) ((T_DECLARE) reduce 171) ((T_DIR) reduce 171) ((T_DNUMBER) reduce 171) ((T_DO) reduce 171) ((T_DOUBLE_CAST) reduce 171) ((T_ECHO) reduce 171) ((T_EMPTY) reduce 171) ((T_ENDWHILE) reduce 171) ((T_ENUM) reduce 171) ((T_EVAL) reduce 171) ((T_EXIT) reduce 171) ((T_FILE) reduce 171) ((T_FINAL) reduce 171) ((T_FINALLY) reduce 171) ((T_FN) reduce 171) ((T_FOR) reduce 171) ((T_FOREACH) reduce 171) ((T_FUNCTION) reduce 171) ((T_FUNC_C) reduce 171) ((T_GLOBAL) reduce 171) ((T_GOTO) reduce 171) ((T_HALT_COMPILER) reduce 171) ((T_IF) reduce 171) ((T_INC) reduce 171) ((T_INCLUDE) reduce 171) ((T_INCLUDE_ONCE) reduce 171) ((T_INLINE_HTML) reduce 171) ((T_INTERFACE) reduce 171) ((T_INT_CAST) reduce 171) ((T_ISSET) reduce 171) ((T_LINE) reduce 171) ((T_LIST) reduce 171) ((T_LNUMBER) reduce 171) ((T_MATCH) reduce 171) ((T_METHOD_C) reduce 171) ((T_NAME_FULLY_QUALIFIED) reduce 171) ((T_NAME_QUALIFIED) reduce 171) ((T_NAME_RELATIVE) reduce 171) ((T_NEW) reduce 171) ((T_NS_C) reduce 171) ((T_OBJECT_CAST) reduce 171) ((T_PRINT) reduce 171) ((T_PROPERTY_C) reduce 171) ((T_READONLY) reduce 171) ((T_REQUIRE) reduce 171) ((T_REQUIRE_ONCE) reduce 171) ((T_RETURN) reduce 171) ((T_START_HEREDOC) reduce 171) ((T_STATIC) reduce 171) ((T_STRING) reduce 171) ((T_STRING_CAST) reduce 171) ((T_SWITCH) reduce 171) ((T_THROW) reduce 171) ((T_TRAIT) reduce 171) ((T_TRAIT_C) reduce 171) ((T_TRY) reduce 171) ((T_UNSET) reduce 171) ((T_UNSET_CAST) reduce 171) ((T_VARIABLE) reduce 171) ((T_VOID_CAST) reduce 171) ((T_WHILE) reduce 171) ((T_YIELD) reduce 171) ((T_YIELD_FROM) reduce 171) (("[") reduce 171) (("`") reduce 171) (("{") reduce 171) (("~") reduce 171)) 27484 ((("!") reduce 177) (("\"") reduce 177) (("$") reduce 177) (("(") reduce 177) (("+") reduce 177) (("-") reduce 177) ((";") reduce 177) (("@") reduce 177) ((T_ABSTRACT) reduce 177) ((T_ARRAY) reduce 177) ((T_ARRAY_CAST) reduce 177) ((T_ATTRIBUTE) reduce 177) ((T_BOOL_CAST) reduce 177) ((T_BREAK) reduce 177) ((T_CATCH) shift) ((T_CLASS) reduce 177) ((T_CLASS_C) reduce 177) ((T_CLONE) reduce 177) ((T_CONSTANT_ENCAPSED_STRING) reduce 177) ((T_CONTINUE) reduce 177) ((T_DEC) reduce 177) ((T_DECLARE) reduce 177) ((T_DIR) reduce 177) ((T_DNUMBER) reduce 177) ((T_DO) reduce 177) ((T_DOUBLE_CAST) reduce 177) ((T_ECHO) reduce 177) ((T_EMPTY) reduce 177) ((T_ENDWHILE) reduce 177) ((T_ENUM) reduce 177) ((T_EVAL) reduce 177) ((T_EXIT) reduce 177) ((T_FILE) reduce 177) ((T_FINAL) reduce 177) ((T_FINALLY) shift) ((T_FN) reduce 177) ((T_FOR) reduce 177) ((T_FOREACH) reduce 177) ((T_FUNCTION) reduce 177) ((T_FUNC_C) reduce 177) ((T_GLOBAL) reduce 177) ((T_GOTO) reduce 177) ((T_HALT_COMPILER) reduce 177) ((T_IF) reduce 177) ((T_INC) reduce 177) ((T_INCLUDE) reduce 177) ((T_INCLUDE_ONCE) reduce 177) ((T_INLINE_HTML) reduce 177) ((T_INTERFACE) reduce 177) ((T_INT_CAST) reduce 177) ((T_ISSET) reduce 177) ((T_LINE) reduce 177) ((T_LIST) reduce 177) ((T_LNUMBER) reduce 177) ((T_MATCH) reduce 177) ((T_METHOD_C) reduce 177) ((T_NAME_FULLY_QUALIFIED) reduce 177) ((T_NAME_QUALIFIED) reduce 177) ((T_NAME_RELATIVE) reduce 177) ((T_NEW) reduce 177) ((T_NS_C) reduce 177) ((T_OBJECT_CAST) reduce 177) ((T_PRINT) reduce 177) ((T_PROPERTY_C) reduce 177) ((T_READONLY) reduce 177) ((T_REQUIRE) reduce 177) ((T_REQUIRE_ONCE) reduce 177) ((T_RETURN) reduce 177) ((T_START_HEREDOC) reduce 177) ((T_STATIC) reduce 177) ((T_STRING) reduce 177) ((T_STRING_CAST) reduce 177) ((T_SWITCH) reduce 177) ((T_THROW) reduce 177) ((T_TRAIT) reduce 177) ((T_TRAIT_C) reduce 177) ((T_TRY) reduce 177) ((T_UNSET) reduce 177) ((T_UNSET_CAST) reduce 177) ((T_VARIABLE) reduce 177) ((T_VOID_CAST) reduce 177) ((T_WHILE) reduce 177) ((T_YIELD) reduce 177) ((T_YIELD_FROM) reduce 177) (("[") reduce 177) (("`") reduce 177) (("{") reduce 177) (("~") reduce 177)) 27487 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_EMPTY) reduce 167) ((T_ENDWHILE) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_VOID_CAST) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 27490 ((("!") reduce 178) (("\"") reduce 178) (("$") reduce 178) (("(") reduce 178) (("+") reduce 178) (("-") reduce 178) ((";") reduce 178) (("@") reduce 178) ((T_ABSTRACT) reduce 178) ((T_ARRAY) reduce 178) ((T_ARRAY_CAST) reduce 178) ((T_ATTRIBUTE) reduce 178) ((T_BOOL_CAST) reduce 178) ((T_BREAK) reduce 178) ((T_CLASS) reduce 178) ((T_CLASS_C) reduce 178) ((T_CLONE) reduce 178) ((T_CONSTANT_ENCAPSED_STRING) reduce 178) ((T_CONTINUE) reduce 178) ((T_DEC) reduce 178) ((T_DECLARE) reduce 178) ((T_DIR) reduce 178) ((T_DNUMBER) reduce 178) ((T_DO) reduce 178) ((T_DOUBLE_CAST) reduce 178) ((T_ECHO) reduce 178) ((T_EMPTY) reduce 178) ((T_ENDWHILE) reduce 178) ((T_ENUM) reduce 178) ((T_EVAL) reduce 178) ((T_EXIT) reduce 178) ((T_FILE) reduce 178) ((T_FINAL) reduce 178) ((T_FN) reduce 178) ((T_FOR) reduce 178) ((T_FOREACH) reduce 178) ((T_FUNCTION) reduce 178) ((T_FUNC_C) reduce 178) ((T_GLOBAL) reduce 178) ((T_GOTO) reduce 178) ((T_HALT_COMPILER) reduce 178) ((T_IF) reduce 178) ((T_INC) reduce 178) ((T_INCLUDE) reduce 178) ((T_INCLUDE_ONCE) reduce 178) ((T_INLINE_HTML) reduce 178) ((T_INTERFACE) reduce 178) ((T_INT_CAST) reduce 178) ((T_ISSET) reduce 178) ((T_LINE) reduce 178) ((T_LIST) reduce 178) ((T_LNUMBER) reduce 178) ((T_MATCH) reduce 178) ((T_METHOD_C) reduce 178) ((T_NAME_FULLY_QUALIFIED) reduce 178) ((T_NAME_QUALIFIED) reduce 178) ((T_NAME_RELATIVE) reduce 178) ((T_NEW) reduce 178) ((T_NS_C) reduce 178) ((T_OBJECT_CAST) reduce 178) ((T_PRINT) reduce 178) ((T_PROPERTY_C) reduce 178) ((T_READONLY) reduce 178) ((T_REQUIRE) reduce 178) ((T_REQUIRE_ONCE) reduce 178) ((T_RETURN) reduce 178) ((T_START_HEREDOC) reduce 178) ((T_STATIC) reduce 178) ((T_STRING) reduce 178) ((T_STRING_CAST) reduce 178) ((T_SWITCH) reduce 178) ((T_THROW) reduce 178) ((T_TRAIT) reduce 178) ((T_TRAIT_C) reduce 178) ((T_TRY) reduce 178) ((T_UNSET) reduce 178) ((T_UNSET_CAST) reduce 178) ((T_VARIABLE) reduce 178) ((T_VOID_CAST) reduce 178) ((T_WHILE) reduce 178) ((T_YIELD) reduce 178) ((T_YIELD_FROM) reduce 178) (("[") reduce 178) (("`") reduce 178) (("{") reduce 178) (("~") reduce 178)) 27497 ((("!") reduce 172) (("\"") reduce 172) (("$") reduce 172) (("(") reduce 172) (("+") reduce 172) (("-") reduce 172) ((";") reduce 172) (("@") reduce 172) ((T_ABSTRACT) reduce 172) ((T_ARRAY) reduce 172) ((T_ARRAY_CAST) reduce 172) ((T_ATTRIBUTE) reduce 172) ((T_BOOL_CAST) reduce 172) ((T_BREAK) reduce 172) ((T_CATCH) reduce 172) ((T_CLASS) reduce 172) ((T_CLASS_C) reduce 172) ((T_CLONE) reduce 172) ((T_CONSTANT_ENCAPSED_STRING) reduce 172) ((T_CONTINUE) reduce 172) ((T_DEC) reduce 172) ((T_DECLARE) reduce 172) ((T_DIR) reduce 172) ((T_DNUMBER) reduce 172) ((T_DO) reduce 172) ((T_DOUBLE_CAST) reduce 172) ((T_ECHO) reduce 172) ((T_EMPTY) reduce 172) ((T_ENDWHILE) reduce 172) ((T_ENUM) reduce 172) ((T_EVAL) reduce 172) ((T_EXIT) reduce 172) ((T_FILE) reduce 172) ((T_FINAL) reduce 172) ((T_FINALLY) reduce 172) ((T_FN) reduce 172) ((T_FOR) reduce 172) ((T_FOREACH) reduce 172) ((T_FUNCTION) reduce 172) ((T_FUNC_C) reduce 172) ((T_GLOBAL) reduce 172) ((T_GOTO) reduce 172) ((T_HALT_COMPILER) reduce 172) ((T_IF) reduce 172) ((T_INC) reduce 172) ((T_INCLUDE) reduce 172) ((T_INCLUDE_ONCE) reduce 172) ((T_INLINE_HTML) reduce 172) ((T_INTERFACE) reduce 172) ((T_INT_CAST) reduce 172) ((T_ISSET) reduce 172) ((T_LINE) reduce 172) ((T_LIST) reduce 172) ((T_LNUMBER) reduce 172) ((T_MATCH) reduce 172) ((T_METHOD_C) reduce 172) ((T_NAME_FULLY_QUALIFIED) reduce 172) ((T_NAME_QUALIFIED) reduce 172) ((T_NAME_RELATIVE) reduce 172) ((T_NEW) reduce 172) ((T_NS_C) reduce 172) ((T_OBJECT_CAST) reduce 172) ((T_PRINT) reduce 172) ((T_PROPERTY_C) reduce 172) ((T_READONLY) reduce 172) ((T_REQUIRE) reduce 172) ((T_REQUIRE_ONCE) reduce 172) ((T_RETURN) reduce 172) ((T_START_HEREDOC) reduce 172) ((T_STATIC) reduce 172) ((T_STRING) reduce 172) ((T_STRING_CAST) reduce 172) ((T_SWITCH) reduce 172) ((T_THROW) reduce 172) ((T_TRAIT) reduce 172) ((T_TRAIT_C) reduce 172) ((T_TRY) reduce 172) ((T_UNSET) reduce 172) ((T_UNSET_CAST) reduce 172) ((T_VARIABLE) reduce 172) ((T_VOID_CAST) reduce 172) ((T_WHILE) reduce 172) ((T_YIELD) reduce 172) ((T_YIELD_FROM) reduce 172) (("[") reduce 172) (("`") reduce 172) (("{") reduce 172) (("~") reduce 172)) 27502 ((("!") reduce 200) (("\"") reduce 200) (("$") reduce 200) (("(") reduce 200) (("+") reduce 200) (("-") reduce 200) ((";") reduce 200) (("@") reduce 200) ((T_ABSTRACT) reduce 200) ((T_ARRAY) reduce 200) ((T_ARRAY_CAST) reduce 200) ((T_ATTRIBUTE) reduce 200) ((T_BOOL_CAST) reduce 200) ((T_BREAK) reduce 200) ((T_CLASS) reduce 200) ((T_CLASS_C) reduce 200) ((T_CLONE) reduce 200) ((T_CONSTANT_ENCAPSED_STRING) reduce 200) ((T_CONTINUE) reduce 200) ((T_DEC) reduce 200) ((T_DECLARE) reduce 200) ((T_DIR) reduce 200) ((T_DNUMBER) reduce 200) ((T_DO) reduce 200) ((T_DOUBLE_CAST) reduce 200) ((T_ECHO) reduce 200) ((T_EMPTY) reduce 200) ((T_ENDWHILE) reduce 200) ((T_ENUM) reduce 200) ((T_EVAL) reduce 200) ((T_EXIT) reduce 200) ((T_FILE) reduce 200) ((T_FINAL) reduce 200) ((T_FN) reduce 200) ((T_FOR) reduce 200) ((T_FOREACH) reduce 200) ((T_FUNCTION) reduce 200) ((T_FUNC_C) reduce 200) ((T_GLOBAL) reduce 200) ((T_GOTO) reduce 200) ((T_HALT_COMPILER) reduce 200) ((T_IF) reduce 200) ((T_INC) reduce 200) ((T_INCLUDE) reduce 200) ((T_INCLUDE_ONCE) reduce 200) ((T_INLINE_HTML) reduce 200) ((T_INTERFACE) reduce 200) ((T_INT_CAST) reduce 200) ((T_ISSET) reduce 200) ((T_LINE) reduce 200) ((T_LIST) reduce 200) ((T_LNUMBER) reduce 200) ((T_MATCH) reduce 200) ((T_METHOD_C) reduce 200) ((T_NAME_FULLY_QUALIFIED) reduce 200) ((T_NAME_QUALIFIED) reduce 200) ((T_NAME_RELATIVE) reduce 200) ((T_NEW) reduce 200) ((T_NS_C) reduce 200) ((T_OBJECT_CAST) reduce 200) ((T_PRINT) reduce 200) ((T_PROPERTY_C) reduce 200) ((T_READONLY) reduce 200) ((T_REQUIRE) reduce 200) ((T_REQUIRE_ONCE) reduce 200) ((T_RETURN) reduce 200) ((T_START_HEREDOC) reduce 200) ((T_STATIC) reduce 200) ((T_STRING) reduce 200) ((T_STRING_CAST) reduce 200) ((T_SWITCH) reduce 200) ((T_THROW) reduce 200) ((T_TRAIT) reduce 200) ((T_TRAIT_C) reduce 200) ((T_TRY) reduce 200) ((T_UNSET) reduce 200) ((T_UNSET_CAST) reduce 200) ((T_VARIABLE) reduce 200) ((T_VOID_CAST) reduce 200) ((T_WHILE) reduce 200) ((T_YIELD) reduce 200) ((T_YIELD_FROM) reduce 200) (("[") reduce 200) (("`") reduce 200) (("{") reduce 200) (("~") reduce 200)) 27507 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_EMPTY) reduce 153) ((T_ENDWHILE) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_VOID_CAST) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 27511 ((("!") reduce 224) (("\"") reduce 224) (("$") reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_EMPTY) reduce 224) ((T_ENDWHILE) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_VARIABLE) reduce 224) ((T_VOID_CAST) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("~") reduce 224)) 27513 ((("!") reduce 225) (("\"") reduce 225) (("$") reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_EMPTY) reduce 225) ((T_ENDWHILE) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_VARIABLE) reduce 225) ((T_VOID_CAST) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("~") reduce 225)) 27517 ((("!") reduce 226) (("\"") reduce 226) (("$") reduce 226) (("(") reduce 226) (("+") reduce 226) (("-") reduce 226) ((";") reduce 226) (("@") reduce 226) ((T_ABSTRACT) reduce 226) ((T_ARRAY) reduce 226) ((T_ARRAY_CAST) reduce 226) ((T_ATTRIBUTE) reduce 226) ((T_BOOL_CAST) reduce 226) ((T_BREAK) reduce 226) ((T_CLASS) reduce 226) ((T_CLASS_C) reduce 226) ((T_CLONE) reduce 226) ((T_CONSTANT_ENCAPSED_STRING) reduce 226) ((T_CONTINUE) reduce 226) ((T_DEC) reduce 226) ((T_DECLARE) reduce 226) ((T_DIR) reduce 226) ((T_DNUMBER) reduce 226) ((T_DO) reduce 226) ((T_DOUBLE_CAST) reduce 226) ((T_ECHO) reduce 226) ((T_EMPTY) reduce 226) ((T_ENDWHILE) reduce 226) ((T_ENUM) reduce 226) ((T_EVAL) reduce 226) ((T_EXIT) reduce 226) ((T_FILE) reduce 226) ((T_FINAL) reduce 226) ((T_FN) reduce 226) ((T_FOR) reduce 226) ((T_FOREACH) reduce 226) ((T_FUNCTION) reduce 226) ((T_FUNC_C) reduce 226) ((T_GLOBAL) reduce 226) ((T_GOTO) reduce 226) ((T_HALT_COMPILER) reduce 226) ((T_IF) reduce 226) ((T_INC) reduce 226) ((T_INCLUDE) reduce 226) ((T_INCLUDE_ONCE) reduce 226) ((T_INLINE_HTML) reduce 226) ((T_INTERFACE) reduce 226) ((T_INT_CAST) reduce 226) ((T_ISSET) reduce 226) ((T_LINE) reduce 226) ((T_LIST) reduce 226) ((T_LNUMBER) reduce 226) ((T_MATCH) reduce 226) ((T_METHOD_C) reduce 226) ((T_NAME_FULLY_QUALIFIED) reduce 226) ((T_NAME_QUALIFIED) reduce 226) ((T_NAME_RELATIVE) reduce 226) ((T_NEW) reduce 226) ((T_NS_C) reduce 226) ((T_OBJECT_CAST) reduce 226) ((T_PRINT) reduce 226) ((T_PROPERTY_C) reduce 226) ((T_READONLY) reduce 226) ((T_REQUIRE) reduce 226) ((T_REQUIRE_ONCE) reduce 226) ((T_RETURN) reduce 226) ((T_START_HEREDOC) reduce 226) ((T_STATIC) reduce 226) ((T_STRING) reduce 226) ((T_STRING_CAST) reduce 226) ((T_SWITCH) reduce 226) ((T_THROW) reduce 226) ((T_TRAIT) reduce 226) ((T_TRAIT_C) reduce 226) ((T_TRY) reduce 226) ((T_UNSET) reduce 226) ((T_UNSET_CAST) reduce 226) ((T_VARIABLE) reduce 226) ((T_VOID_CAST) reduce 226) ((T_WHILE) reduce 226) ((T_YIELD) reduce 226) ((T_YIELD_FROM) reduce 226) (("[") reduce 226) (("`") reduce 226) (("{") reduce 226) (("~") reduce 226)) 27520 ((("!") reduce 227) (("\"") reduce 227) (("$") reduce 227) (("(") reduce 227) (("+") reduce 227) (("-") reduce 227) ((";") reduce 227) (("@") reduce 227) ((T_ABSTRACT) reduce 227) ((T_ARRAY) reduce 227) ((T_ARRAY_CAST) reduce 227) ((T_ATTRIBUTE) reduce 227) ((T_BOOL_CAST) reduce 227) ((T_BREAK) reduce 227) ((T_CLASS) reduce 227) ((T_CLASS_C) reduce 227) ((T_CLONE) reduce 227) ((T_CONSTANT_ENCAPSED_STRING) reduce 227) ((T_CONTINUE) reduce 227) ((T_DEC) reduce 227) ((T_DECLARE) reduce 227) ((T_DIR) reduce 227) ((T_DNUMBER) reduce 227) ((T_DO) reduce 227) ((T_DOUBLE_CAST) reduce 227) ((T_ECHO) reduce 227) ((T_EMPTY) reduce 227) ((T_ENDWHILE) reduce 227) ((T_ENUM) reduce 227) ((T_EVAL) reduce 227) ((T_EXIT) reduce 227) ((T_FILE) reduce 227) ((T_FINAL) reduce 227) ((T_FN) reduce 227) ((T_FOR) reduce 227) ((T_FOREACH) reduce 227) ((T_FUNCTION) reduce 227) ((T_FUNC_C) reduce 227) ((T_GLOBAL) reduce 227) ((T_GOTO) reduce 227) ((T_HALT_COMPILER) reduce 227) ((T_IF) reduce 227) ((T_INC) reduce 227) ((T_INCLUDE) reduce 227) ((T_INCLUDE_ONCE) reduce 227) ((T_INLINE_HTML) reduce 227) ((T_INTERFACE) reduce 227) ((T_INT_CAST) reduce 227) ((T_ISSET) reduce 227) ((T_LINE) reduce 227) ((T_LIST) reduce 227) ((T_LNUMBER) reduce 227) ((T_MATCH) reduce 227) ((T_METHOD_C) reduce 227) ((T_NAME_FULLY_QUALIFIED) reduce 227) ((T_NAME_QUALIFIED) reduce 227) ((T_NAME_RELATIVE) reduce 227) ((T_NEW) reduce 227) ((T_NS_C) reduce 227) ((T_OBJECT_CAST) reduce 227) ((T_PRINT) reduce 227) ((T_PROPERTY_C) reduce 227) ((T_READONLY) reduce 227) ((T_REQUIRE) reduce 227) ((T_REQUIRE_ONCE) reduce 227) ((T_RETURN) reduce 227) ((T_START_HEREDOC) reduce 227) ((T_STATIC) reduce 227) ((T_STRING) reduce 227) ((T_STRING_CAST) reduce 227) ((T_SWITCH) reduce 227) ((T_THROW) reduce 227) ((T_TRAIT) reduce 227) ((T_TRAIT_C) reduce 227) ((T_TRY) reduce 227) ((T_UNSET) reduce 227) ((T_UNSET_CAST) reduce 227) ((T_VARIABLE) reduce 227) ((T_VOID_CAST) reduce 227) ((T_WHILE) reduce 227) ((T_YIELD) reduce 227) ((T_YIELD_FROM) reduce 227) (("[") reduce 227) (("`") reduce 227) (("{") reduce 227) (("~") reduce 227)) 27521 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_EMPTY) reduce 169) ((T_ENDWHILE) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_VOID_CAST) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 27523 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_EMPTY) reduce 158) ((T_ENDWHILE) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_VOID_CAST) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 27525 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_EMPTY) reduce 156) ((T_ENDWHILE) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_VOID_CAST) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 27531 ((("!") reduce 201) (("\"") reduce 201) (("$") reduce 201) (("(") reduce 201) (("+") reduce 201) (("-") reduce 201) ((";") reduce 201) (("@") reduce 201) ((T_ABSTRACT) reduce 201) ((T_ARRAY) reduce 201) ((T_ARRAY_CAST) reduce 201) ((T_ATTRIBUTE) reduce 201) ((T_BOOL_CAST) reduce 201) ((T_BREAK) reduce 201) ((T_CLASS) reduce 201) ((T_CLASS_C) reduce 201) ((T_CLONE) reduce 201) ((T_CONSTANT_ENCAPSED_STRING) reduce 201) ((T_CONTINUE) reduce 201) ((T_DEC) reduce 201) ((T_DECLARE) reduce 201) ((T_DIR) reduce 201) ((T_DNUMBER) reduce 201) ((T_DO) reduce 201) ((T_DOUBLE_CAST) reduce 201) ((T_ECHO) reduce 201) ((T_EMPTY) reduce 201) ((T_ENDWHILE) reduce 201) ((T_ENUM) reduce 201) ((T_EVAL) reduce 201) ((T_EXIT) reduce 201) ((T_FILE) reduce 201) ((T_FINAL) reduce 201) ((T_FN) reduce 201) ((T_FOR) reduce 201) ((T_FOREACH) reduce 201) ((T_FUNCTION) reduce 201) ((T_FUNC_C) reduce 201) ((T_GLOBAL) reduce 201) ((T_GOTO) reduce 201) ((T_HALT_COMPILER) reduce 201) ((T_IF) reduce 201) ((T_INC) reduce 201) ((T_INCLUDE) reduce 201) ((T_INCLUDE_ONCE) reduce 201) ((T_INLINE_HTML) reduce 201) ((T_INTERFACE) reduce 201) ((T_INT_CAST) reduce 201) ((T_ISSET) reduce 201) ((T_LINE) reduce 201) ((T_LIST) reduce 201) ((T_LNUMBER) reduce 201) ((T_MATCH) reduce 201) ((T_METHOD_C) reduce 201) ((T_NAME_FULLY_QUALIFIED) reduce 201) ((T_NAME_QUALIFIED) reduce 201) ((T_NAME_RELATIVE) reduce 201) ((T_NEW) reduce 201) ((T_NS_C) reduce 201) ((T_OBJECT_CAST) reduce 201) ((T_PRINT) reduce 201) ((T_PROPERTY_C) reduce 201) ((T_READONLY) reduce 201) ((T_REQUIRE) reduce 201) ((T_REQUIRE_ONCE) reduce 201) ((T_RETURN) reduce 201) ((T_START_HEREDOC) reduce 201) ((T_STATIC) reduce 201) ((T_STRING) reduce 201) ((T_STRING_CAST) reduce 201) ((T_SWITCH) reduce 201) ((T_THROW) reduce 201) ((T_TRAIT) reduce 201) ((T_TRAIT_C) reduce 201) ((T_TRY) reduce 201) ((T_UNSET) reduce 201) ((T_UNSET_CAST) reduce 201) ((T_VARIABLE) reduce 201) ((T_VOID_CAST) reduce 201) ((T_WHILE) reduce 201) ((T_YIELD) reduce 201) ((T_YIELD_FROM) reduce 201) (("[") reduce 201) (("`") reduce 201) (("{") reduce 201) (("~") reduce 201)) 27535 ((("!") reduce 244) (("\"") reduce 244) (("$") reduce 244) (("(") reduce 244) (("+") reduce 244) (("-") reduce 244) ((";") reduce 244) (("@") reduce 244) ((T_ABSTRACT) reduce 244) ((T_ARRAY) reduce 244) ((T_ARRAY_CAST) reduce 244) ((T_ATTRIBUTE) reduce 244) ((T_BOOL_CAST) reduce 244) ((T_BREAK) reduce 244) ((T_CLASS) reduce 244) ((T_CLASS_C) reduce 244) ((T_CLONE) reduce 244) ((T_CONSTANT_ENCAPSED_STRING) reduce 244) ((T_CONTINUE) reduce 244) ((T_DEC) reduce 244) ((T_DECLARE) reduce 244) ((T_DIR) reduce 244) ((T_DNUMBER) reduce 244) ((T_DO) reduce 244) ((T_DOUBLE_CAST) reduce 244) ((T_ECHO) reduce 244) ((T_ELSE) reduce 244) ((T_ELSEIF) reduce 244) ((T_EMPTY) reduce 244) ((T_ENDWHILE) reduce 244) ((T_ENUM) reduce 244) ((T_EVAL) reduce 244) ((T_EXIT) reduce 244) ((T_FILE) reduce 244) ((T_FINAL) reduce 244) ((T_FN) reduce 244) ((T_FOR) reduce 244) ((T_FOREACH) reduce 244) ((T_FUNCTION) reduce 244) ((T_FUNC_C) reduce 244) ((T_GLOBAL) reduce 244) ((T_GOTO) reduce 244) ((T_HALT_COMPILER) reduce 244) ((T_IF) reduce 244) ((T_INC) reduce 244) ((T_INCLUDE) reduce 244) ((T_INCLUDE_ONCE) reduce 244) ((T_INLINE_HTML) reduce 244) ((T_INTERFACE) reduce 244) ((T_INT_CAST) reduce 244) ((T_ISSET) reduce 244) ((T_LINE) reduce 244) ((T_LIST) reduce 244) ((T_LNUMBER) reduce 244) ((T_MATCH) reduce 244) ((T_METHOD_C) reduce 244) ((T_NAME_FULLY_QUALIFIED) reduce 244) ((T_NAME_QUALIFIED) reduce 244) ((T_NAME_RELATIVE) reduce 244) ((T_NEW) reduce 244) ((T_NS_C) reduce 244) ((T_OBJECT_CAST) reduce 244) ((T_PRINT) reduce 244) ((T_PROPERTY_C) reduce 244) ((T_READONLY) reduce 244) ((T_REQUIRE) reduce 244) ((T_REQUIRE_ONCE) reduce 244) ((T_RETURN) reduce 244) ((T_START_HEREDOC) reduce 244) ((T_STATIC) reduce 244) ((T_STRING) reduce 244) ((T_STRING_CAST) reduce 244) ((T_SWITCH) reduce 244) ((T_THROW) reduce 244) ((T_TRAIT) reduce 244) ((T_TRAIT_C) reduce 244) ((T_TRY) reduce 244) ((T_UNSET) reduce 244) ((T_UNSET_CAST) reduce 244) ((T_VARIABLE) reduce 244) ((T_VOID_CAST) reduce 244) ((T_WHILE) reduce 244) ((T_YIELD) reduce 244) ((T_YIELD_FROM) reduce 244) (("[") reduce 244) (("`") reduce 244) (("{") reduce 244) (("~") reduce 244)) 27538 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_EMPTY) reduce 146) ((T_ENDWHILE) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_PROPERTY_C) reduce 146) ((T_READONLY) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_VOID_CAST) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("~") reduce 146)) 27540 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_EMPTY) reduce 168) ((T_ENDWHILE) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_VOID_CAST) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 27542 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_EMPTY) reduce 157) ((T_ENDWHILE) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_VOID_CAST) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 27552 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_EMPTY) reduce 164) ((T_ENDWHILE) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_VOID_CAST) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 27553 ((("!") reduce 220) (("\"") reduce 220) (("$") reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_EMPTY) reduce 220) ((T_ENDWHILE) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_VARIABLE) reduce 220) ((T_VOID_CAST) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("~") reduce 220)) 27556 ((("!") reduce 221) (("\"") reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_EMPTY) reduce 221) ((T_ENDWHILE) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_VARIABLE) reduce 221) ((T_VOID_CAST) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("~") reduce 221)) 27557 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_EMPTY) reduce 163) ((T_ENDWHILE) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_VOID_CAST) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 27566 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_EMPTY) reduce 152) ((T_ENDWHILE) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_VOID_CAST) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 27567 ((("!") reduce 218) (("\"") reduce 218) (("$") reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_EMPTY) reduce 218) ((T_ENDWHILE) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_VARIABLE) reduce 218) ((T_VOID_CAST) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("~") reduce 218)) 27570 ((("!") reduce 219) (("\"") reduce 219) (("$") reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_EMPTY) reduce 219) ((T_ENDWHILE) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_VARIABLE) reduce 219) ((T_VOID_CAST) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("~") reduce 219)) 27577 ((("!") reduce 202) (("\"") reduce 202) (("$") reduce 202) (("(") reduce 202) (("+") reduce 202) (("-") reduce 202) ((";") reduce 202) (("@") reduce 202) ((T_ABSTRACT) reduce 202) ((T_ARRAY) reduce 202) ((T_ARRAY_CAST) reduce 202) ((T_ATTRIBUTE) reduce 202) ((T_BOOL_CAST) reduce 202) ((T_BREAK) reduce 202) ((T_CLASS) reduce 202) ((T_CLASS_C) reduce 202) ((T_CLONE) reduce 202) ((T_CONSTANT_ENCAPSED_STRING) reduce 202) ((T_CONTINUE) reduce 202) ((T_DEC) reduce 202) ((T_DECLARE) reduce 202) ((T_DIR) reduce 202) ((T_DNUMBER) reduce 202) ((T_DO) reduce 202) ((T_DOUBLE_CAST) reduce 202) ((T_ECHO) reduce 202) ((T_EMPTY) reduce 202) ((T_ENDWHILE) reduce 202) ((T_ENUM) reduce 202) ((T_EVAL) reduce 202) ((T_EXIT) reduce 202) ((T_FILE) reduce 202) ((T_FINAL) reduce 202) ((T_FN) reduce 202) ((T_FOR) reduce 202) ((T_FOREACH) reduce 202) ((T_FUNCTION) reduce 202) ((T_FUNC_C) reduce 202) ((T_GLOBAL) reduce 202) ((T_GOTO) reduce 202) ((T_HALT_COMPILER) reduce 202) ((T_IF) reduce 202) ((T_INC) reduce 202) ((T_INCLUDE) reduce 202) ((T_INCLUDE_ONCE) reduce 202) ((T_INLINE_HTML) reduce 202) ((T_INTERFACE) reduce 202) ((T_INT_CAST) reduce 202) ((T_ISSET) reduce 202) ((T_LINE) reduce 202) ((T_LIST) reduce 202) ((T_LNUMBER) reduce 202) ((T_MATCH) reduce 202) ((T_METHOD_C) reduce 202) ((T_NAME_FULLY_QUALIFIED) reduce 202) ((T_NAME_QUALIFIED) reduce 202) ((T_NAME_RELATIVE) reduce 202) ((T_NEW) reduce 202) ((T_NS_C) reduce 202) ((T_OBJECT_CAST) reduce 202) ((T_PRINT) reduce 202) ((T_PROPERTY_C) reduce 202) ((T_READONLY) reduce 202) ((T_REQUIRE) reduce 202) ((T_REQUIRE_ONCE) reduce 202) ((T_RETURN) reduce 202) ((T_START_HEREDOC) reduce 202) ((T_STATIC) reduce 202) ((T_STRING) reduce 202) ((T_STRING_CAST) reduce 202) ((T_SWITCH) reduce 202) ((T_THROW) reduce 202) ((T_TRAIT) reduce 202) ((T_TRAIT_C) reduce 202) ((T_TRY) reduce 202) ((T_UNSET) reduce 202) ((T_UNSET_CAST) reduce 202) ((T_VARIABLE) reduce 202) ((T_VOID_CAST) reduce 202) ((T_WHILE) reduce 202) ((T_YIELD) reduce 202) ((T_YIELD_FROM) reduce 202) (("[") reduce 202) (("`") reduce 202) (("{") reduce 202) (("~") reduce 202)) 27578 ((("!") reduce 243) (("\"") reduce 243) (("$") reduce 243) (("(") reduce 243) (("+") reduce 243) (("-") reduce 243) ((";") reduce 243) (("@") reduce 243) ((T_ABSTRACT) reduce 243) ((T_ARRAY) reduce 243) ((T_ARRAY_CAST) reduce 243) ((T_ATTRIBUTE) reduce 243) ((T_BOOL_CAST) reduce 243) ((T_BREAK) reduce 243) ((T_CLASS) reduce 243) ((T_CLASS_C) reduce 243) ((T_CLONE) reduce 243) ((T_CONSTANT_ENCAPSED_STRING) reduce 243) ((T_CONTINUE) reduce 243) ((T_DEC) reduce 243) ((T_DECLARE) reduce 243) ((T_DIR) reduce 243) ((T_DNUMBER) reduce 243) ((T_DO) reduce 243) ((T_DOUBLE_CAST) reduce 243) ((T_ECHO) reduce 243) ((T_EMPTY) reduce 243) ((T_ENDIF) reduce 243) ((T_ENUM) reduce 243) ((T_EVAL) reduce 243) ((T_EXIT) reduce 243) ((T_FILE) reduce 243) ((T_FINAL) reduce 243) ((T_FN) reduce 243) ((T_FOR) reduce 243) ((T_FOREACH) reduce 243) ((T_FUNCTION) reduce 243) ((T_FUNC_C) reduce 243) ((T_GLOBAL) reduce 243) ((T_GOTO) reduce 243) ((T_HALT_COMPILER) reduce 243) ((T_IF) reduce 243) ((T_INC) reduce 243) ((T_INCLUDE) reduce 243) ((T_INCLUDE_ONCE) reduce 243) ((T_INLINE_HTML) reduce 243) ((T_INTERFACE) reduce 243) ((T_INT_CAST) reduce 243) ((T_ISSET) reduce 243) ((T_LINE) reduce 243) ((T_LIST) reduce 243) ((T_LNUMBER) reduce 243) ((T_MATCH) reduce 243) ((T_METHOD_C) reduce 243) ((T_NAME_FULLY_QUALIFIED) reduce 243) ((T_NAME_QUALIFIED) reduce 243) ((T_NAME_RELATIVE) reduce 243) ((T_NEW) reduce 243) ((T_NS_C) reduce 243) ((T_OBJECT_CAST) reduce 243) ((T_PRINT) reduce 243) ((T_PROPERTY_C) reduce 243) ((T_READONLY) reduce 243) ((T_REQUIRE) reduce 243) ((T_REQUIRE_ONCE) reduce 243) ((T_RETURN) reduce 243) ((T_START_HEREDOC) reduce 243) ((T_STATIC) reduce 243) ((T_STRING) reduce 243) ((T_STRING_CAST) reduce 243) ((T_SWITCH) reduce 243) ((T_THROW) reduce 243) ((T_TRAIT) reduce 243) ((T_TRAIT_C) reduce 243) ((T_TRY) reduce 243) ((T_UNSET) reduce 243) ((T_UNSET_CAST) reduce 243) ((T_VARIABLE) reduce 243) ((T_VOID_CAST) reduce 243) ((T_WHILE) reduce 243) ((T_YIELD) reduce 243) ((T_YIELD_FROM) reduce 243) (("[") reduce 243) (("`") reduce 243) (("{") reduce 243) (("~") reduce 243)) 27580 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_EMPTY) reduce 159) ((T_ENDWHILE) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_VOID_CAST) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 27586 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_EMPTY) reduce 151) ((T_ENDWHILE) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_VOID_CAST) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 27591 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_EMPTY) reduce 165) ((T_ENDWHILE) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_VOID_CAST) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 27592 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_EMPTY) reduce 222) ((T_ENDWHILE) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_VARIABLE) reduce 222) ((T_VOID_CAST) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("~") reduce 222)) 27595 ((("!") reduce 223) (("\"") reduce 223) (("$") reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_EMPTY) reduce 223) ((T_ENDWHILE) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_VARIABLE) reduce 223) ((T_VOID_CAST) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("~") reduce 223)) 27597 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_EMPTY) reduce 155) ((T_ENDWHILE) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_VOID_CAST) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 27604 ((("!") reduce 190) (("\"") reduce 190) (("$") reduce 190) (("(") reduce 190) (("+") reduce 190) (("-") reduce 190) ((";") reduce 190) (("@") reduce 190) ((T_ABSTRACT) reduce 190) ((T_ARRAY) reduce 190) ((T_ARRAY_CAST) reduce 190) ((T_ATTRIBUTE) reduce 190) ((T_BOOL_CAST) reduce 190) ((T_BREAK) reduce 190) ((T_CLASS) reduce 190) ((T_CLASS_C) reduce 190) ((T_CLONE) reduce 190) ((T_CONSTANT_ENCAPSED_STRING) reduce 190) ((T_CONTINUE) reduce 190) ((T_DEC) reduce 190) ((T_DECLARE) reduce 190) ((T_DIR) reduce 190) ((T_DNUMBER) reduce 190) ((T_DO) reduce 190) ((T_DOUBLE_CAST) reduce 190) ((T_ECHO) reduce 190) ((T_EMPTY) reduce 190) ((T_ENDWHILE) reduce 190) ((T_ENUM) reduce 190) ((T_EVAL) reduce 190) ((T_EXIT) reduce 190) ((T_FILE) reduce 190) ((T_FINAL) reduce 190) ((T_FN) reduce 190) ((T_FOR) reduce 190) ((T_FOREACH) reduce 190) ((T_FUNCTION) reduce 190) ((T_FUNC_C) reduce 190) ((T_GLOBAL) reduce 190) ((T_GOTO) reduce 190) ((T_HALT_COMPILER) reduce 190) ((T_IF) reduce 190) ((T_INC) reduce 190) ((T_INCLUDE) reduce 190) ((T_INCLUDE_ONCE) reduce 190) ((T_INLINE_HTML) reduce 190) ((T_INTERFACE) reduce 190) ((T_INT_CAST) reduce 190) ((T_ISSET) reduce 190) ((T_LINE) reduce 190) ((T_LIST) reduce 190) ((T_LNUMBER) reduce 190) ((T_MATCH) reduce 190) ((T_METHOD_C) reduce 190) ((T_NAME_FULLY_QUALIFIED) reduce 190) ((T_NAME_QUALIFIED) reduce 190) ((T_NAME_RELATIVE) reduce 190) ((T_NEW) reduce 190) ((T_NS_C) reduce 190) ((T_OBJECT_CAST) reduce 190) ((T_PRINT) reduce 190) ((T_PROPERTY_C) reduce 190) ((T_READONLY) reduce 190) ((T_REQUIRE) reduce 190) ((T_REQUIRE_ONCE) reduce 190) ((T_RETURN) reduce 190) ((T_START_HEREDOC) reduce 190) ((T_STATIC) reduce 190) ((T_STRING) reduce 190) ((T_STRING_CAST) reduce 190) ((T_SWITCH) reduce 190) ((T_THROW) reduce 190) ((T_TRAIT) reduce 190) ((T_TRAIT_C) reduce 190) ((T_TRY) reduce 190) ((T_UNSET) reduce 190) ((T_UNSET_CAST) reduce 190) ((T_VARIABLE) reduce 190) ((T_VOID_CAST) reduce 190) ((T_WHILE) reduce 190) ((T_YIELD) reduce 190) ((T_YIELD_FROM) reduce 190) (("[") reduce 190) (("`") reduce 190) (("{") reduce 190) (("~") reduce 190)) 27606 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_EMPTY) reduce 154) ((T_ENDWHILE) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_VOID_CAST) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 27608 ((("!") reduce 170) (("\"") reduce 170) (("$") reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_EMPTY) reduce 170) ((T_ENDIF) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_VARIABLE) reduce 170) ((T_VOID_CAST) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("~") reduce 170)) 27613 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_EMPTY) reduce 162) ((T_ENDIF) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_VOID_CAST) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 27616 ((("!") reduce 171) (("\"") reduce 171) (("$") reduce 171) (("(") reduce 171) (("+") reduce 171) (("-") reduce 171) ((";") reduce 171) (("@") reduce 171) ((T_ABSTRACT) reduce 171) ((T_ARRAY) reduce 171) ((T_ARRAY_CAST) reduce 171) ((T_ATTRIBUTE) reduce 171) ((T_BOOL_CAST) reduce 171) ((T_BREAK) reduce 171) ((T_CATCH) reduce 171) ((T_CLASS) reduce 171) ((T_CLASS_C) reduce 171) ((T_CLONE) reduce 171) ((T_CONSTANT_ENCAPSED_STRING) reduce 171) ((T_CONTINUE) reduce 171) ((T_DEC) reduce 171) ((T_DECLARE) reduce 171) ((T_DIR) reduce 171) ((T_DNUMBER) reduce 171) ((T_DO) reduce 171) ((T_DOUBLE_CAST) reduce 171) ((T_ECHO) reduce 171) ((T_EMPTY) reduce 171) ((T_ENDIF) reduce 171) ((T_ENUM) reduce 171) ((T_EVAL) reduce 171) ((T_EXIT) reduce 171) ((T_FILE) reduce 171) ((T_FINAL) reduce 171) ((T_FINALLY) reduce 171) ((T_FN) reduce 171) ((T_FOR) reduce 171) ((T_FOREACH) reduce 171) ((T_FUNCTION) reduce 171) ((T_FUNC_C) reduce 171) ((T_GLOBAL) reduce 171) ((T_GOTO) reduce 171) ((T_HALT_COMPILER) reduce 171) ((T_IF) reduce 171) ((T_INC) reduce 171) ((T_INCLUDE) reduce 171) ((T_INCLUDE_ONCE) reduce 171) ((T_INLINE_HTML) reduce 171) ((T_INTERFACE) reduce 171) ((T_INT_CAST) reduce 171) ((T_ISSET) reduce 171) ((T_LINE) reduce 171) ((T_LIST) reduce 171) ((T_LNUMBER) reduce 171) ((T_MATCH) reduce 171) ((T_METHOD_C) reduce 171) ((T_NAME_FULLY_QUALIFIED) reduce 171) ((T_NAME_QUALIFIED) reduce 171) ((T_NAME_RELATIVE) reduce 171) ((T_NEW) reduce 171) ((T_NS_C) reduce 171) ((T_OBJECT_CAST) reduce 171) ((T_PRINT) reduce 171) ((T_PROPERTY_C) reduce 171) ((T_READONLY) reduce 171) ((T_REQUIRE) reduce 171) ((T_REQUIRE_ONCE) reduce 171) ((T_RETURN) reduce 171) ((T_START_HEREDOC) reduce 171) ((T_STATIC) reduce 171) ((T_STRING) reduce 171) ((T_STRING_CAST) reduce 171) ((T_SWITCH) reduce 171) ((T_THROW) reduce 171) ((T_TRAIT) reduce 171) ((T_TRAIT_C) reduce 171) ((T_TRY) reduce 171) ((T_UNSET) reduce 171) ((T_UNSET_CAST) reduce 171) ((T_VARIABLE) reduce 171) ((T_VOID_CAST) reduce 171) ((T_WHILE) reduce 171) ((T_YIELD) reduce 171) ((T_YIELD_FROM) reduce 171) (("[") reduce 171) (("`") reduce 171) (("{") reduce 171) (("~") reduce 171)) 27617 ((("!") reduce 177) (("\"") reduce 177) (("$") reduce 177) (("(") reduce 177) (("+") reduce 177) (("-") reduce 177) ((";") reduce 177) (("@") reduce 177) ((T_ABSTRACT) reduce 177) ((T_ARRAY) reduce 177) ((T_ARRAY_CAST) reduce 177) ((T_ATTRIBUTE) reduce 177) ((T_BOOL_CAST) reduce 177) ((T_BREAK) reduce 177) ((T_CATCH) shift) ((T_CLASS) reduce 177) ((T_CLASS_C) reduce 177) ((T_CLONE) reduce 177) ((T_CONSTANT_ENCAPSED_STRING) reduce 177) ((T_CONTINUE) reduce 177) ((T_DEC) reduce 177) ((T_DECLARE) reduce 177) ((T_DIR) reduce 177) ((T_DNUMBER) reduce 177) ((T_DO) reduce 177) ((T_DOUBLE_CAST) reduce 177) ((T_ECHO) reduce 177) ((T_EMPTY) reduce 177) ((T_ENDIF) reduce 177) ((T_ENUM) reduce 177) ((T_EVAL) reduce 177) ((T_EXIT) reduce 177) ((T_FILE) reduce 177) ((T_FINAL) reduce 177) ((T_FINALLY) shift) ((T_FN) reduce 177) ((T_FOR) reduce 177) ((T_FOREACH) reduce 177) ((T_FUNCTION) reduce 177) ((T_FUNC_C) reduce 177) ((T_GLOBAL) reduce 177) ((T_GOTO) reduce 177) ((T_HALT_COMPILER) reduce 177) ((T_IF) reduce 177) ((T_INC) reduce 177) ((T_INCLUDE) reduce 177) ((T_INCLUDE_ONCE) reduce 177) ((T_INLINE_HTML) reduce 177) ((T_INTERFACE) reduce 177) ((T_INT_CAST) reduce 177) ((T_ISSET) reduce 177) ((T_LINE) reduce 177) ((T_LIST) reduce 177) ((T_LNUMBER) reduce 177) ((T_MATCH) reduce 177) ((T_METHOD_C) reduce 177) ((T_NAME_FULLY_QUALIFIED) reduce 177) ((T_NAME_QUALIFIED) reduce 177) ((T_NAME_RELATIVE) reduce 177) ((T_NEW) reduce 177) ((T_NS_C) reduce 177) ((T_OBJECT_CAST) reduce 177) ((T_PRINT) reduce 177) ((T_PROPERTY_C) reduce 177) ((T_READONLY) reduce 177) ((T_REQUIRE) reduce 177) ((T_REQUIRE_ONCE) reduce 177) ((T_RETURN) reduce 177) ((T_START_HEREDOC) reduce 177) ((T_STATIC) reduce 177) ((T_STRING) reduce 177) ((T_STRING_CAST) reduce 177) ((T_SWITCH) reduce 177) ((T_THROW) reduce 177) ((T_TRAIT) reduce 177) ((T_TRAIT_C) reduce 177) ((T_TRY) reduce 177) ((T_UNSET) reduce 177) ((T_UNSET_CAST) reduce 177) ((T_VARIABLE) reduce 177) ((T_VOID_CAST) reduce 177) ((T_WHILE) reduce 177) ((T_YIELD) reduce 177) ((T_YIELD_FROM) reduce 177) (("[") reduce 177) (("`") reduce 177) (("{") reduce 177) (("~") reduce 177)) 27620 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_EMPTY) reduce 167) ((T_ENDIF) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_VOID_CAST) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 27623 ((("!") reduce 178) (("\"") reduce 178) (("$") reduce 178) (("(") reduce 178) (("+") reduce 178) (("-") reduce 178) ((";") reduce 178) (("@") reduce 178) ((T_ABSTRACT) reduce 178) ((T_ARRAY) reduce 178) ((T_ARRAY_CAST) reduce 178) ((T_ATTRIBUTE) reduce 178) ((T_BOOL_CAST) reduce 178) ((T_BREAK) reduce 178) ((T_CLASS) reduce 178) ((T_CLASS_C) reduce 178) ((T_CLONE) reduce 178) ((T_CONSTANT_ENCAPSED_STRING) reduce 178) ((T_CONTINUE) reduce 178) ((T_DEC) reduce 178) ((T_DECLARE) reduce 178) ((T_DIR) reduce 178) ((T_DNUMBER) reduce 178) ((T_DO) reduce 178) ((T_DOUBLE_CAST) reduce 178) ((T_ECHO) reduce 178) ((T_EMPTY) reduce 178) ((T_ENDIF) reduce 178) ((T_ENUM) reduce 178) ((T_EVAL) reduce 178) ((T_EXIT) reduce 178) ((T_FILE) reduce 178) ((T_FINAL) reduce 178) ((T_FN) reduce 178) ((T_FOR) reduce 178) ((T_FOREACH) reduce 178) ((T_FUNCTION) reduce 178) ((T_FUNC_C) reduce 178) ((T_GLOBAL) reduce 178) ((T_GOTO) reduce 178) ((T_HALT_COMPILER) reduce 178) ((T_IF) reduce 178) ((T_INC) reduce 178) ((T_INCLUDE) reduce 178) ((T_INCLUDE_ONCE) reduce 178) ((T_INLINE_HTML) reduce 178) ((T_INTERFACE) reduce 178) ((T_INT_CAST) reduce 178) ((T_ISSET) reduce 178) ((T_LINE) reduce 178) ((T_LIST) reduce 178) ((T_LNUMBER) reduce 178) ((T_MATCH) reduce 178) ((T_METHOD_C) reduce 178) ((T_NAME_FULLY_QUALIFIED) reduce 178) ((T_NAME_QUALIFIED) reduce 178) ((T_NAME_RELATIVE) reduce 178) ((T_NEW) reduce 178) ((T_NS_C) reduce 178) ((T_OBJECT_CAST) reduce 178) ((T_PRINT) reduce 178) ((T_PROPERTY_C) reduce 178) ((T_READONLY) reduce 178) ((T_REQUIRE) reduce 178) ((T_REQUIRE_ONCE) reduce 178) ((T_RETURN) reduce 178) ((T_START_HEREDOC) reduce 178) ((T_STATIC) reduce 178) ((T_STRING) reduce 178) ((T_STRING_CAST) reduce 178) ((T_SWITCH) reduce 178) ((T_THROW) reduce 178) ((T_TRAIT) reduce 178) ((T_TRAIT_C) reduce 178) ((T_TRY) reduce 178) ((T_UNSET) reduce 178) ((T_UNSET_CAST) reduce 178) ((T_VARIABLE) reduce 178) ((T_VOID_CAST) reduce 178) ((T_WHILE) reduce 178) ((T_YIELD) reduce 178) ((T_YIELD_FROM) reduce 178) (("[") reduce 178) (("`") reduce 178) (("{") reduce 178) (("~") reduce 178)) 27630 ((("!") reduce 172) (("\"") reduce 172) (("$") reduce 172) (("(") reduce 172) (("+") reduce 172) (("-") reduce 172) ((";") reduce 172) (("@") reduce 172) ((T_ABSTRACT) reduce 172) ((T_ARRAY) reduce 172) ((T_ARRAY_CAST) reduce 172) ((T_ATTRIBUTE) reduce 172) ((T_BOOL_CAST) reduce 172) ((T_BREAK) reduce 172) ((T_CATCH) reduce 172) ((T_CLASS) reduce 172) ((T_CLASS_C) reduce 172) ((T_CLONE) reduce 172) ((T_CONSTANT_ENCAPSED_STRING) reduce 172) ((T_CONTINUE) reduce 172) ((T_DEC) reduce 172) ((T_DECLARE) reduce 172) ((T_DIR) reduce 172) ((T_DNUMBER) reduce 172) ((T_DO) reduce 172) ((T_DOUBLE_CAST) reduce 172) ((T_ECHO) reduce 172) ((T_EMPTY) reduce 172) ((T_ENDIF) reduce 172) ((T_ENUM) reduce 172) ((T_EVAL) reduce 172) ((T_EXIT) reduce 172) ((T_FILE) reduce 172) ((T_FINAL) reduce 172) ((T_FINALLY) reduce 172) ((T_FN) reduce 172) ((T_FOR) reduce 172) ((T_FOREACH) reduce 172) ((T_FUNCTION) reduce 172) ((T_FUNC_C) reduce 172) ((T_GLOBAL) reduce 172) ((T_GOTO) reduce 172) ((T_HALT_COMPILER) reduce 172) ((T_IF) reduce 172) ((T_INC) reduce 172) ((T_INCLUDE) reduce 172) ((T_INCLUDE_ONCE) reduce 172) ((T_INLINE_HTML) reduce 172) ((T_INTERFACE) reduce 172) ((T_INT_CAST) reduce 172) ((T_ISSET) reduce 172) ((T_LINE) reduce 172) ((T_LIST) reduce 172) ((T_LNUMBER) reduce 172) ((T_MATCH) reduce 172) ((T_METHOD_C) reduce 172) ((T_NAME_FULLY_QUALIFIED) reduce 172) ((T_NAME_QUALIFIED) reduce 172) ((T_NAME_RELATIVE) reduce 172) ((T_NEW) reduce 172) ((T_NS_C) reduce 172) ((T_OBJECT_CAST) reduce 172) ((T_PRINT) reduce 172) ((T_PROPERTY_C) reduce 172) ((T_READONLY) reduce 172) ((T_REQUIRE) reduce 172) ((T_REQUIRE_ONCE) reduce 172) ((T_RETURN) reduce 172) ((T_START_HEREDOC) reduce 172) ((T_STATIC) reduce 172) ((T_STRING) reduce 172) ((T_STRING_CAST) reduce 172) ((T_SWITCH) reduce 172) ((T_THROW) reduce 172) ((T_TRAIT) reduce 172) ((T_TRAIT_C) reduce 172) ((T_TRY) reduce 172) ((T_UNSET) reduce 172) ((T_UNSET_CAST) reduce 172) ((T_VARIABLE) reduce 172) ((T_VOID_CAST) reduce 172) ((T_WHILE) reduce 172) ((T_YIELD) reduce 172) ((T_YIELD_FROM) reduce 172) (("[") reduce 172) (("`") reduce 172) (("{") reduce 172) (("~") reduce 172)) 27635 ((("!") reduce 200) (("\"") reduce 200) (("$") reduce 200) (("(") reduce 200) (("+") reduce 200) (("-") reduce 200) ((";") reduce 200) (("@") reduce 200) ((T_ABSTRACT) reduce 200) ((T_ARRAY) reduce 200) ((T_ARRAY_CAST) reduce 200) ((T_ATTRIBUTE) reduce 200) ((T_BOOL_CAST) reduce 200) ((T_BREAK) reduce 200) ((T_CLASS) reduce 200) ((T_CLASS_C) reduce 200) ((T_CLONE) reduce 200) ((T_CONSTANT_ENCAPSED_STRING) reduce 200) ((T_CONTINUE) reduce 200) ((T_DEC) reduce 200) ((T_DECLARE) reduce 200) ((T_DIR) reduce 200) ((T_DNUMBER) reduce 200) ((T_DO) reduce 200) ((T_DOUBLE_CAST) reduce 200) ((T_ECHO) reduce 200) ((T_EMPTY) reduce 200) ((T_ENDIF) reduce 200) ((T_ENUM) reduce 200) ((T_EVAL) reduce 200) ((T_EXIT) reduce 200) ((T_FILE) reduce 200) ((T_FINAL) reduce 200) ((T_FN) reduce 200) ((T_FOR) reduce 200) ((T_FOREACH) reduce 200) ((T_FUNCTION) reduce 200) ((T_FUNC_C) reduce 200) ((T_GLOBAL) reduce 200) ((T_GOTO) reduce 200) ((T_HALT_COMPILER) reduce 200) ((T_IF) reduce 200) ((T_INC) reduce 200) ((T_INCLUDE) reduce 200) ((T_INCLUDE_ONCE) reduce 200) ((T_INLINE_HTML) reduce 200) ((T_INTERFACE) reduce 200) ((T_INT_CAST) reduce 200) ((T_ISSET) reduce 200) ((T_LINE) reduce 200) ((T_LIST) reduce 200) ((T_LNUMBER) reduce 200) ((T_MATCH) reduce 200) ((T_METHOD_C) reduce 200) ((T_NAME_FULLY_QUALIFIED) reduce 200) ((T_NAME_QUALIFIED) reduce 200) ((T_NAME_RELATIVE) reduce 200) ((T_NEW) reduce 200) ((T_NS_C) reduce 200) ((T_OBJECT_CAST) reduce 200) ((T_PRINT) reduce 200) ((T_PROPERTY_C) reduce 200) ((T_READONLY) reduce 200) ((T_REQUIRE) reduce 200) ((T_REQUIRE_ONCE) reduce 200) ((T_RETURN) reduce 200) ((T_START_HEREDOC) reduce 200) ((T_STATIC) reduce 200) ((T_STRING) reduce 200) ((T_STRING_CAST) reduce 200) ((T_SWITCH) reduce 200) ((T_THROW) reduce 200) ((T_TRAIT) reduce 200) ((T_TRAIT_C) reduce 200) ((T_TRY) reduce 200) ((T_UNSET) reduce 200) ((T_UNSET_CAST) reduce 200) ((T_VARIABLE) reduce 200) ((T_VOID_CAST) reduce 200) ((T_WHILE) reduce 200) ((T_YIELD) reduce 200) ((T_YIELD_FROM) reduce 200) (("[") reduce 200) (("`") reduce 200) (("{") reduce 200) (("~") reduce 200)) 27640 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_EMPTY) reduce 153) ((T_ENDIF) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_VOID_CAST) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 27644 ((("!") reduce 224) (("\"") reduce 224) (("$") reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_EMPTY) reduce 224) ((T_ENDIF) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_VARIABLE) reduce 224) ((T_VOID_CAST) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("~") reduce 224)) 27646 ((("!") reduce 225) (("\"") reduce 225) (("$") reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_EMPTY) reduce 225) ((T_ENDIF) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_VARIABLE) reduce 225) ((T_VOID_CAST) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("~") reduce 225)) 27650 ((("!") reduce 226) (("\"") reduce 226) (("$") reduce 226) (("(") reduce 226) (("+") reduce 226) (("-") reduce 226) ((";") reduce 226) (("@") reduce 226) ((T_ABSTRACT) reduce 226) ((T_ARRAY) reduce 226) ((T_ARRAY_CAST) reduce 226) ((T_ATTRIBUTE) reduce 226) ((T_BOOL_CAST) reduce 226) ((T_BREAK) reduce 226) ((T_CLASS) reduce 226) ((T_CLASS_C) reduce 226) ((T_CLONE) reduce 226) ((T_CONSTANT_ENCAPSED_STRING) reduce 226) ((T_CONTINUE) reduce 226) ((T_DEC) reduce 226) ((T_DECLARE) reduce 226) ((T_DIR) reduce 226) ((T_DNUMBER) reduce 226) ((T_DO) reduce 226) ((T_DOUBLE_CAST) reduce 226) ((T_ECHO) reduce 226) ((T_EMPTY) reduce 226) ((T_ENDIF) reduce 226) ((T_ENUM) reduce 226) ((T_EVAL) reduce 226) ((T_EXIT) reduce 226) ((T_FILE) reduce 226) ((T_FINAL) reduce 226) ((T_FN) reduce 226) ((T_FOR) reduce 226) ((T_FOREACH) reduce 226) ((T_FUNCTION) reduce 226) ((T_FUNC_C) reduce 226) ((T_GLOBAL) reduce 226) ((T_GOTO) reduce 226) ((T_HALT_COMPILER) reduce 226) ((T_IF) reduce 226) ((T_INC) reduce 226) ((T_INCLUDE) reduce 226) ((T_INCLUDE_ONCE) reduce 226) ((T_INLINE_HTML) reduce 226) ((T_INTERFACE) reduce 226) ((T_INT_CAST) reduce 226) ((T_ISSET) reduce 226) ((T_LINE) reduce 226) ((T_LIST) reduce 226) ((T_LNUMBER) reduce 226) ((T_MATCH) reduce 226) ((T_METHOD_C) reduce 226) ((T_NAME_FULLY_QUALIFIED) reduce 226) ((T_NAME_QUALIFIED) reduce 226) ((T_NAME_RELATIVE) reduce 226) ((T_NEW) reduce 226) ((T_NS_C) reduce 226) ((T_OBJECT_CAST) reduce 226) ((T_PRINT) reduce 226) ((T_PROPERTY_C) reduce 226) ((T_READONLY) reduce 226) ((T_REQUIRE) reduce 226) ((T_REQUIRE_ONCE) reduce 226) ((T_RETURN) reduce 226) ((T_START_HEREDOC) reduce 226) ((T_STATIC) reduce 226) ((T_STRING) reduce 226) ((T_STRING_CAST) reduce 226) ((T_SWITCH) reduce 226) ((T_THROW) reduce 226) ((T_TRAIT) reduce 226) ((T_TRAIT_C) reduce 226) ((T_TRY) reduce 226) ((T_UNSET) reduce 226) ((T_UNSET_CAST) reduce 226) ((T_VARIABLE) reduce 226) ((T_VOID_CAST) reduce 226) ((T_WHILE) reduce 226) ((T_YIELD) reduce 226) ((T_YIELD_FROM) reduce 226) (("[") reduce 226) (("`") reduce 226) (("{") reduce 226) (("~") reduce 226)) 27653 ((("!") reduce 227) (("\"") reduce 227) (("$") reduce 227) (("(") reduce 227) (("+") reduce 227) (("-") reduce 227) ((";") reduce 227) (("@") reduce 227) ((T_ABSTRACT) reduce 227) ((T_ARRAY) reduce 227) ((T_ARRAY_CAST) reduce 227) ((T_ATTRIBUTE) reduce 227) ((T_BOOL_CAST) reduce 227) ((T_BREAK) reduce 227) ((T_CLASS) reduce 227) ((T_CLASS_C) reduce 227) ((T_CLONE) reduce 227) ((T_CONSTANT_ENCAPSED_STRING) reduce 227) ((T_CONTINUE) reduce 227) ((T_DEC) reduce 227) ((T_DECLARE) reduce 227) ((T_DIR) reduce 227) ((T_DNUMBER) reduce 227) ((T_DO) reduce 227) ((T_DOUBLE_CAST) reduce 227) ((T_ECHO) reduce 227) ((T_EMPTY) reduce 227) ((T_ENDIF) reduce 227) ((T_ENUM) reduce 227) ((T_EVAL) reduce 227) ((T_EXIT) reduce 227) ((T_FILE) reduce 227) ((T_FINAL) reduce 227) ((T_FN) reduce 227) ((T_FOR) reduce 227) ((T_FOREACH) reduce 227) ((T_FUNCTION) reduce 227) ((T_FUNC_C) reduce 227) ((T_GLOBAL) reduce 227) ((T_GOTO) reduce 227) ((T_HALT_COMPILER) reduce 227) ((T_IF) reduce 227) ((T_INC) reduce 227) ((T_INCLUDE) reduce 227) ((T_INCLUDE_ONCE) reduce 227) ((T_INLINE_HTML) reduce 227) ((T_INTERFACE) reduce 227) ((T_INT_CAST) reduce 227) ((T_ISSET) reduce 227) ((T_LINE) reduce 227) ((T_LIST) reduce 227) ((T_LNUMBER) reduce 227) ((T_MATCH) reduce 227) ((T_METHOD_C) reduce 227) ((T_NAME_FULLY_QUALIFIED) reduce 227) ((T_NAME_QUALIFIED) reduce 227) ((T_NAME_RELATIVE) reduce 227) ((T_NEW) reduce 227) ((T_NS_C) reduce 227) ((T_OBJECT_CAST) reduce 227) ((T_PRINT) reduce 227) ((T_PROPERTY_C) reduce 227) ((T_READONLY) reduce 227) ((T_REQUIRE) reduce 227) ((T_REQUIRE_ONCE) reduce 227) ((T_RETURN) reduce 227) ((T_START_HEREDOC) reduce 227) ((T_STATIC) reduce 227) ((T_STRING) reduce 227) ((T_STRING_CAST) reduce 227) ((T_SWITCH) reduce 227) ((T_THROW) reduce 227) ((T_TRAIT) reduce 227) ((T_TRAIT_C) reduce 227) ((T_TRY) reduce 227) ((T_UNSET) reduce 227) ((T_UNSET_CAST) reduce 227) ((T_VARIABLE) reduce 227) ((T_VOID_CAST) reduce 227) ((T_WHILE) reduce 227) ((T_YIELD) reduce 227) ((T_YIELD_FROM) reduce 227) (("[") reduce 227) (("`") reduce 227) (("{") reduce 227) (("~") reduce 227)) 27654 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_EMPTY) reduce 169) ((T_ENDIF) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_VOID_CAST) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 27656 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_EMPTY) reduce 158) ((T_ENDIF) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_VOID_CAST) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 27658 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_EMPTY) reduce 156) ((T_ENDIF) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_VOID_CAST) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 27664 ((("!") reduce 201) (("\"") reduce 201) (("$") reduce 201) (("(") reduce 201) (("+") reduce 201) (("-") reduce 201) ((";") reduce 201) (("@") reduce 201) ((T_ABSTRACT) reduce 201) ((T_ARRAY) reduce 201) ((T_ARRAY_CAST) reduce 201) ((T_ATTRIBUTE) reduce 201) ((T_BOOL_CAST) reduce 201) ((T_BREAK) reduce 201) ((T_CLASS) reduce 201) ((T_CLASS_C) reduce 201) ((T_CLONE) reduce 201) ((T_CONSTANT_ENCAPSED_STRING) reduce 201) ((T_CONTINUE) reduce 201) ((T_DEC) reduce 201) ((T_DECLARE) reduce 201) ((T_DIR) reduce 201) ((T_DNUMBER) reduce 201) ((T_DO) reduce 201) ((T_DOUBLE_CAST) reduce 201) ((T_ECHO) reduce 201) ((T_EMPTY) reduce 201) ((T_ENDIF) reduce 201) ((T_ENUM) reduce 201) ((T_EVAL) reduce 201) ((T_EXIT) reduce 201) ((T_FILE) reduce 201) ((T_FINAL) reduce 201) ((T_FN) reduce 201) ((T_FOR) reduce 201) ((T_FOREACH) reduce 201) ((T_FUNCTION) reduce 201) ((T_FUNC_C) reduce 201) ((T_GLOBAL) reduce 201) ((T_GOTO) reduce 201) ((T_HALT_COMPILER) reduce 201) ((T_IF) reduce 201) ((T_INC) reduce 201) ((T_INCLUDE) reduce 201) ((T_INCLUDE_ONCE) reduce 201) ((T_INLINE_HTML) reduce 201) ((T_INTERFACE) reduce 201) ((T_INT_CAST) reduce 201) ((T_ISSET) reduce 201) ((T_LINE) reduce 201) ((T_LIST) reduce 201) ((T_LNUMBER) reduce 201) ((T_MATCH) reduce 201) ((T_METHOD_C) reduce 201) ((T_NAME_FULLY_QUALIFIED) reduce 201) ((T_NAME_QUALIFIED) reduce 201) ((T_NAME_RELATIVE) reduce 201) ((T_NEW) reduce 201) ((T_NS_C) reduce 201) ((T_OBJECT_CAST) reduce 201) ((T_PRINT) reduce 201) ((T_PROPERTY_C) reduce 201) ((T_READONLY) reduce 201) ((T_REQUIRE) reduce 201) ((T_REQUIRE_ONCE) reduce 201) ((T_RETURN) reduce 201) ((T_START_HEREDOC) reduce 201) ((T_STATIC) reduce 201) ((T_STRING) reduce 201) ((T_STRING_CAST) reduce 201) ((T_SWITCH) reduce 201) ((T_THROW) reduce 201) ((T_TRAIT) reduce 201) ((T_TRAIT_C) reduce 201) ((T_TRY) reduce 201) ((T_UNSET) reduce 201) ((T_UNSET_CAST) reduce 201) ((T_VARIABLE) reduce 201) ((T_VOID_CAST) reduce 201) ((T_WHILE) reduce 201) ((T_YIELD) reduce 201) ((T_YIELD_FROM) reduce 201) (("[") reduce 201) (("`") reduce 201) (("{") reduce 201) (("~") reduce 201)) 27667 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_EMPTY) reduce 146) ((T_ENDIF) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_PROPERTY_C) reduce 146) ((T_READONLY) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_VOID_CAST) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("~") reduce 146)) 27669 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_EMPTY) reduce 168) ((T_ENDIF) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_VOID_CAST) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 27671 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_EMPTY) reduce 157) ((T_ENDIF) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_VOID_CAST) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 27681 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_EMPTY) reduce 164) ((T_ENDIF) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_VOID_CAST) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 27682 ((("!") reduce 220) (("\"") reduce 220) (("$") reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_EMPTY) reduce 220) ((T_ENDIF) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_VARIABLE) reduce 220) ((T_VOID_CAST) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("~") reduce 220)) 27685 ((("!") reduce 221) (("\"") reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_EMPTY) reduce 221) ((T_ENDIF) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_VARIABLE) reduce 221) ((T_VOID_CAST) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("~") reduce 221)) 27686 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_EMPTY) reduce 163) ((T_ENDIF) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_VOID_CAST) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 27695 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_EMPTY) reduce 152) ((T_ENDIF) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_VOID_CAST) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 27696 ((("!") reduce 218) (("\"") reduce 218) (("$") reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_EMPTY) reduce 218) ((T_ENDIF) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_VARIABLE) reduce 218) ((T_VOID_CAST) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("~") reduce 218)) 27699 ((("!") reduce 219) (("\"") reduce 219) (("$") reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_EMPTY) reduce 219) ((T_ENDIF) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_VARIABLE) reduce 219) ((T_VOID_CAST) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("~") reduce 219)) 27706 ((("!") reduce 202) (("\"") reduce 202) (("$") reduce 202) (("(") reduce 202) (("+") reduce 202) (("-") reduce 202) ((";") reduce 202) (("@") reduce 202) ((T_ABSTRACT) reduce 202) ((T_ARRAY) reduce 202) ((T_ARRAY_CAST) reduce 202) ((T_ATTRIBUTE) reduce 202) ((T_BOOL_CAST) reduce 202) ((T_BREAK) reduce 202) ((T_CLASS) reduce 202) ((T_CLASS_C) reduce 202) ((T_CLONE) reduce 202) ((T_CONSTANT_ENCAPSED_STRING) reduce 202) ((T_CONTINUE) reduce 202) ((T_DEC) reduce 202) ((T_DECLARE) reduce 202) ((T_DIR) reduce 202) ((T_DNUMBER) reduce 202) ((T_DO) reduce 202) ((T_DOUBLE_CAST) reduce 202) ((T_ECHO) reduce 202) ((T_EMPTY) reduce 202) ((T_ENDIF) reduce 202) ((T_ENUM) reduce 202) ((T_EVAL) reduce 202) ((T_EXIT) reduce 202) ((T_FILE) reduce 202) ((T_FINAL) reduce 202) ((T_FN) reduce 202) ((T_FOR) reduce 202) ((T_FOREACH) reduce 202) ((T_FUNCTION) reduce 202) ((T_FUNC_C) reduce 202) ((T_GLOBAL) reduce 202) ((T_GOTO) reduce 202) ((T_HALT_COMPILER) reduce 202) ((T_IF) reduce 202) ((T_INC) reduce 202) ((T_INCLUDE) reduce 202) ((T_INCLUDE_ONCE) reduce 202) ((T_INLINE_HTML) reduce 202) ((T_INTERFACE) reduce 202) ((T_INT_CAST) reduce 202) ((T_ISSET) reduce 202) ((T_LINE) reduce 202) ((T_LIST) reduce 202) ((T_LNUMBER) reduce 202) ((T_MATCH) reduce 202) ((T_METHOD_C) reduce 202) ((T_NAME_FULLY_QUALIFIED) reduce 202) ((T_NAME_QUALIFIED) reduce 202) ((T_NAME_RELATIVE) reduce 202) ((T_NEW) reduce 202) ((T_NS_C) reduce 202) ((T_OBJECT_CAST) reduce 202) ((T_PRINT) reduce 202) ((T_PROPERTY_C) reduce 202) ((T_READONLY) reduce 202) ((T_REQUIRE) reduce 202) ((T_REQUIRE_ONCE) reduce 202) ((T_RETURN) reduce 202) ((T_START_HEREDOC) reduce 202) ((T_STATIC) reduce 202) ((T_STRING) reduce 202) ((T_STRING_CAST) reduce 202) ((T_SWITCH) reduce 202) ((T_THROW) reduce 202) ((T_TRAIT) reduce 202) ((T_TRAIT_C) reduce 202) ((T_TRY) reduce 202) ((T_UNSET) reduce 202) ((T_UNSET_CAST) reduce 202) ((T_VARIABLE) reduce 202) ((T_VOID_CAST) reduce 202) ((T_WHILE) reduce 202) ((T_YIELD) reduce 202) ((T_YIELD_FROM) reduce 202) (("[") reduce 202) (("`") reduce 202) (("{") reduce 202) (("~") reduce 202)) 27707 ((("!") reduce 251) (("\"") reduce 251) (("$") reduce 251) (("(") reduce 251) (("+") reduce 251) (("-") reduce 251) ((";") reduce 251) (("@") reduce 251) ((T_ABSTRACT) reduce 251) ((T_ARRAY) reduce 251) ((T_ARRAY_CAST) reduce 251) ((T_ATTRIBUTE) reduce 251) ((T_BOOL_CAST) reduce 251) ((T_BREAK) reduce 251) ((T_CLASS) reduce 251) ((T_CLASS_C) reduce 251) ((T_CLONE) reduce 251) ((T_CONSTANT_ENCAPSED_STRING) reduce 251) ((T_CONTINUE) reduce 251) ((T_DEC) reduce 251) ((T_DECLARE) reduce 251) ((T_DIR) reduce 251) ((T_DNUMBER) reduce 251) ((T_DO) reduce 251) ((T_DOUBLE_CAST) reduce 251) ((T_ECHO) reduce 251) ((T_ELSE) reduce 251) ((T_ELSEIF) reduce 251) ((T_EMPTY) reduce 251) ((T_ENDIF) reduce 251) ((T_ENUM) reduce 251) ((T_EVAL) reduce 251) ((T_EXIT) reduce 251) ((T_FILE) reduce 251) ((T_FINAL) reduce 251) ((T_FN) reduce 251) ((T_FOR) reduce 251) ((T_FOREACH) reduce 251) ((T_FUNCTION) reduce 251) ((T_FUNC_C) reduce 251) ((T_GLOBAL) reduce 251) ((T_GOTO) reduce 251) ((T_HALT_COMPILER) reduce 251) ((T_IF) reduce 251) ((T_INC) reduce 251) ((T_INCLUDE) reduce 251) ((T_INCLUDE_ONCE) reduce 251) ((T_INLINE_HTML) reduce 251) ((T_INTERFACE) reduce 251) ((T_INT_CAST) reduce 251) ((T_ISSET) reduce 251) ((T_LINE) reduce 251) ((T_LIST) reduce 251) ((T_LNUMBER) reduce 251) ((T_MATCH) reduce 251) ((T_METHOD_C) reduce 251) ((T_NAME_FULLY_QUALIFIED) reduce 251) ((T_NAME_QUALIFIED) reduce 251) ((T_NAME_RELATIVE) reduce 251) ((T_NEW) reduce 251) ((T_NS_C) reduce 251) ((T_OBJECT_CAST) reduce 251) ((T_PRINT) reduce 251) ((T_PROPERTY_C) reduce 251) ((T_READONLY) reduce 251) ((T_REQUIRE) reduce 251) ((T_REQUIRE_ONCE) reduce 251) ((T_RETURN) reduce 251) ((T_START_HEREDOC) reduce 251) ((T_STATIC) reduce 251) ((T_STRING) reduce 251) ((T_STRING_CAST) reduce 251) ((T_SWITCH) reduce 251) ((T_THROW) reduce 251) ((T_TRAIT) reduce 251) ((T_TRAIT_C) reduce 251) ((T_TRY) reduce 251) ((T_UNSET) reduce 251) ((T_UNSET_CAST) reduce 251) ((T_VARIABLE) reduce 251) ((T_VOID_CAST) reduce 251) ((T_WHILE) reduce 251) ((T_YIELD) reduce 251) ((T_YIELD_FROM) reduce 251) (("[") reduce 251) (("`") reduce 251) (("{") reduce 251) (("~") reduce 251)) 27709 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_EMPTY) reduce 159) ((T_ENDIF) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_VOID_CAST) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 27715 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_EMPTY) reduce 151) ((T_ENDIF) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_VOID_CAST) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 27720 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_EMPTY) reduce 165) ((T_ENDIF) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_VOID_CAST) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 27721 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_EMPTY) reduce 222) ((T_ENDIF) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_VARIABLE) reduce 222) ((T_VOID_CAST) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("~") reduce 222)) 27724 ((("!") reduce 223) (("\"") reduce 223) (("$") reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_EMPTY) reduce 223) ((T_ENDIF) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_VARIABLE) reduce 223) ((T_VOID_CAST) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("~") reduce 223)) 27726 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_EMPTY) reduce 155) ((T_ENDIF) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_VOID_CAST) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 27733 ((("!") reduce 190) (("\"") reduce 190) (("$") reduce 190) (("(") reduce 190) (("+") reduce 190) (("-") reduce 190) ((";") reduce 190) (("@") reduce 190) ((T_ABSTRACT) reduce 190) ((T_ARRAY) reduce 190) ((T_ARRAY_CAST) reduce 190) ((T_ATTRIBUTE) reduce 190) ((T_BOOL_CAST) reduce 190) ((T_BREAK) reduce 190) ((T_CLASS) reduce 190) ((T_CLASS_C) reduce 190) ((T_CLONE) reduce 190) ((T_CONSTANT_ENCAPSED_STRING) reduce 190) ((T_CONTINUE) reduce 190) ((T_DEC) reduce 190) ((T_DECLARE) reduce 190) ((T_DIR) reduce 190) ((T_DNUMBER) reduce 190) ((T_DO) reduce 190) ((T_DOUBLE_CAST) reduce 190) ((T_ECHO) reduce 190) ((T_EMPTY) reduce 190) ((T_ENDIF) reduce 190) ((T_ENUM) reduce 190) ((T_EVAL) reduce 190) ((T_EXIT) reduce 190) ((T_FILE) reduce 190) ((T_FINAL) reduce 190) ((T_FN) reduce 190) ((T_FOR) reduce 190) ((T_FOREACH) reduce 190) ((T_FUNCTION) reduce 190) ((T_FUNC_C) reduce 190) ((T_GLOBAL) reduce 190) ((T_GOTO) reduce 190) ((T_HALT_COMPILER) reduce 190) ((T_IF) reduce 190) ((T_INC) reduce 190) ((T_INCLUDE) reduce 190) ((T_INCLUDE_ONCE) reduce 190) ((T_INLINE_HTML) reduce 190) ((T_INTERFACE) reduce 190) ((T_INT_CAST) reduce 190) ((T_ISSET) reduce 190) ((T_LINE) reduce 190) ((T_LIST) reduce 190) ((T_LNUMBER) reduce 190) ((T_MATCH) reduce 190) ((T_METHOD_C) reduce 190) ((T_NAME_FULLY_QUALIFIED) reduce 190) ((T_NAME_QUALIFIED) reduce 190) ((T_NAME_RELATIVE) reduce 190) ((T_NEW) reduce 190) ((T_NS_C) reduce 190) ((T_OBJECT_CAST) reduce 190) ((T_PRINT) reduce 190) ((T_PROPERTY_C) reduce 190) ((T_READONLY) reduce 190) ((T_REQUIRE) reduce 190) ((T_REQUIRE_ONCE) reduce 190) ((T_RETURN) reduce 190) ((T_START_HEREDOC) reduce 190) ((T_STATIC) reduce 190) ((T_STRING) reduce 190) ((T_STRING_CAST) reduce 190) ((T_SWITCH) reduce 190) ((T_THROW) reduce 190) ((T_TRAIT) reduce 190) ((T_TRAIT_C) reduce 190) ((T_TRY) reduce 190) ((T_UNSET) reduce 190) ((T_UNSET_CAST) reduce 190) ((T_VARIABLE) reduce 190) ((T_VOID_CAST) reduce 190) ((T_WHILE) reduce 190) ((T_YIELD) reduce 190) ((T_YIELD_FROM) reduce 190) (("[") reduce 190) (("`") reduce 190) (("{") reduce 190) (("~") reduce 190)) 27735 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_EMPTY) reduce 154) ((T_ENDIF) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_VOID_CAST) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 27740 ((("!") reduce 242) (("\"") reduce 242) (("$") reduce 242) (("(") reduce 242) (("+") reduce 242) (("-") reduce 242) ((";") reduce 242) (("@") reduce 242) ((T_ABSTRACT) reduce 242) ((T_ARRAY) reduce 242) ((T_ARRAY_CAST) reduce 242) ((T_ATTRIBUTE) reduce 242) ((T_BOOL_CAST) reduce 242) ((T_BREAK) reduce 242) ((T_CLASS) reduce 242) ((T_CLASS_C) reduce 242) ((T_CLONE) reduce 242) ((T_CONSTANT_ENCAPSED_STRING) reduce 242) ((T_CONTINUE) reduce 242) ((T_DEC) reduce 242) ((T_DECLARE) reduce 242) ((T_DIR) reduce 242) ((T_DNUMBER) reduce 242) ((T_DO) reduce 242) ((T_DOUBLE_CAST) reduce 242) ((T_ECHO) reduce 242) ((T_ELSE) reduce 242) ((T_ELSEIF) reduce 242) ((T_EMPTY) reduce 242) ((T_ENDIF) reduce 242) ((T_ENUM) reduce 242) ((T_EVAL) reduce 242) ((T_EXIT) reduce 242) ((T_FILE) reduce 242) ((T_FINAL) reduce 242) ((T_FN) reduce 242) ((T_FOR) reduce 242) ((T_FOREACH) reduce 242) ((T_FUNCTION) reduce 242) ((T_FUNC_C) reduce 242) ((T_GLOBAL) reduce 242) ((T_GOTO) reduce 242) ((T_HALT_COMPILER) reduce 242) ((T_IF) reduce 242) ((T_INC) reduce 242) ((T_INCLUDE) reduce 242) ((T_INCLUDE_ONCE) reduce 242) ((T_INLINE_HTML) reduce 242) ((T_INTERFACE) reduce 242) ((T_INT_CAST) reduce 242) ((T_ISSET) reduce 242) ((T_LINE) reduce 242) ((T_LIST) reduce 242) ((T_LNUMBER) reduce 242) ((T_MATCH) reduce 242) ((T_METHOD_C) reduce 242) ((T_NAME_FULLY_QUALIFIED) reduce 242) ((T_NAME_QUALIFIED) reduce 242) ((T_NAME_RELATIVE) reduce 242) ((T_NEW) reduce 242) ((T_NS_C) reduce 242) ((T_OBJECT_CAST) reduce 242) ((T_PRINT) reduce 242) ((T_PROPERTY_C) reduce 242) ((T_READONLY) reduce 242) ((T_REQUIRE) reduce 242) ((T_REQUIRE_ONCE) reduce 242) ((T_RETURN) reduce 242) ((T_START_HEREDOC) reduce 242) ((T_STATIC) reduce 242) ((T_STRING) reduce 242) ((T_STRING_CAST) reduce 242) ((T_SWITCH) reduce 242) ((T_THROW) reduce 242) ((T_TRAIT) reduce 242) ((T_TRAIT_C) reduce 242) ((T_TRY) reduce 242) ((T_UNSET) reduce 242) ((T_UNSET_CAST) reduce 242) ((T_VARIABLE) reduce 242) ((T_VOID_CAST) reduce 242) ((T_WHILE) reduce 242) ((T_YIELD) reduce 242) ((T_YIELD_FROM) reduce 242) (("[") reduce 242) (("`") reduce 242) (("{") reduce 242) (("~") reduce 242)) 27741 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_ELSE) reduce 150) ((T_ELSEIF) reduce 150) ((T_EMPTY) reduce 150) ((T_ENDIF) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_VOID_CAST) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 27744 ((("!") reduce 243) (("\"") reduce 243) (("$") reduce 243) (("(") reduce 243) (("+") reduce 243) (("-") reduce 243) ((";") reduce 243) (("@") reduce 243) ((T_ABSTRACT) reduce 243) ((T_ARRAY) reduce 243) ((T_ARRAY_CAST) reduce 243) ((T_ATTRIBUTE) reduce 243) ((T_BOOL_CAST) reduce 243) ((T_BREAK) reduce 243) ((T_CLASS) reduce 243) ((T_CLASS_C) reduce 243) ((T_CLONE) reduce 243) ((T_CONSTANT_ENCAPSED_STRING) reduce 243) ((T_CONTINUE) reduce 243) ((T_DEC) reduce 243) ((T_DECLARE) reduce 243) ((T_DIR) reduce 243) ((T_DNUMBER) reduce 243) ((T_DO) reduce 243) ((T_DOUBLE_CAST) reduce 243) ((T_ECHO) reduce 243) ((T_ELSE) reduce 243) ((T_ELSEIF) reduce 243) ((T_EMPTY) reduce 243) ((T_ENDIF) reduce 243) ((T_ENUM) reduce 243) ((T_EVAL) reduce 243) ((T_EXIT) reduce 243) ((T_FILE) reduce 243) ((T_FINAL) reduce 243) ((T_FN) reduce 243) ((T_FOR) reduce 243) ((T_FOREACH) reduce 243) ((T_FUNCTION) reduce 243) ((T_FUNC_C) reduce 243) ((T_GLOBAL) reduce 243) ((T_GOTO) reduce 243) ((T_HALT_COMPILER) reduce 243) ((T_IF) reduce 243) ((T_INC) reduce 243) ((T_INCLUDE) reduce 243) ((T_INCLUDE_ONCE) reduce 243) ((T_INLINE_HTML) reduce 243) ((T_INTERFACE) reduce 243) ((T_INT_CAST) reduce 243) ((T_ISSET) reduce 243) ((T_LINE) reduce 243) ((T_LIST) reduce 243) ((T_LNUMBER) reduce 243) ((T_MATCH) reduce 243) ((T_METHOD_C) reduce 243) ((T_NAME_FULLY_QUALIFIED) reduce 243) ((T_NAME_QUALIFIED) reduce 243) ((T_NAME_RELATIVE) reduce 243) ((T_NEW) reduce 243) ((T_NS_C) reduce 243) ((T_OBJECT_CAST) reduce 243) ((T_PRINT) reduce 243) ((T_PROPERTY_C) reduce 243) ((T_READONLY) reduce 243) ((T_REQUIRE) reduce 243) ((T_REQUIRE_ONCE) reduce 243) ((T_RETURN) reduce 243) ((T_START_HEREDOC) reduce 243) ((T_STATIC) reduce 243) ((T_STRING) reduce 243) ((T_STRING_CAST) reduce 243) ((T_SWITCH) reduce 243) ((T_THROW) reduce 243) ((T_TRAIT) reduce 243) ((T_TRAIT_C) reduce 243) ((T_TRY) reduce 243) ((T_UNSET) reduce 243) ((T_UNSET_CAST) reduce 243) ((T_VARIABLE) reduce 243) ((T_VOID_CAST) reduce 243) ((T_WHILE) reduce 243) ((T_YIELD) reduce 243) ((T_YIELD_FROM) reduce 243) (("[") reduce 243) (("`") reduce 243) (("{") reduce 243) (("~") reduce 243)) 27746 ((("!") reduce 170) (("\"") reduce 170) (("$") reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_ELSE) reduce 170) ((T_ELSEIF) reduce 170) ((T_EMPTY) reduce 170) ((T_ENDIF) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_VARIABLE) reduce 170) ((T_VOID_CAST) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("~") reduce 170)) 27751 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_ELSE) reduce 162) ((T_ELSEIF) reduce 162) ((T_EMPTY) reduce 162) ((T_ENDIF) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_VOID_CAST) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 27754 ((("!") reduce 171) (("\"") reduce 171) (("$") reduce 171) (("(") reduce 171) (("+") reduce 171) (("-") reduce 171) ((";") reduce 171) (("@") reduce 171) ((T_ABSTRACT) reduce 171) ((T_ARRAY) reduce 171) ((T_ARRAY_CAST) reduce 171) ((T_ATTRIBUTE) reduce 171) ((T_BOOL_CAST) reduce 171) ((T_BREAK) reduce 171) ((T_CATCH) reduce 171) ((T_CLASS) reduce 171) ((T_CLASS_C) reduce 171) ((T_CLONE) reduce 171) ((T_CONSTANT_ENCAPSED_STRING) reduce 171) ((T_CONTINUE) reduce 171) ((T_DEC) reduce 171) ((T_DECLARE) reduce 171) ((T_DIR) reduce 171) ((T_DNUMBER) reduce 171) ((T_DO) reduce 171) ((T_DOUBLE_CAST) reduce 171) ((T_ECHO) reduce 171) ((T_ELSE) reduce 171) ((T_ELSEIF) reduce 171) ((T_EMPTY) reduce 171) ((T_ENDIF) reduce 171) ((T_ENUM) reduce 171) ((T_EVAL) reduce 171) ((T_EXIT) reduce 171) ((T_FILE) reduce 171) ((T_FINAL) reduce 171) ((T_FINALLY) reduce 171) ((T_FN) reduce 171) ((T_FOR) reduce 171) ((T_FOREACH) reduce 171) ((T_FUNCTION) reduce 171) ((T_FUNC_C) reduce 171) ((T_GLOBAL) reduce 171) ((T_GOTO) reduce 171) ((T_HALT_COMPILER) reduce 171) ((T_IF) reduce 171) ((T_INC) reduce 171) ((T_INCLUDE) reduce 171) ((T_INCLUDE_ONCE) reduce 171) ((T_INLINE_HTML) reduce 171) ((T_INTERFACE) reduce 171) ((T_INT_CAST) reduce 171) ((T_ISSET) reduce 171) ((T_LINE) reduce 171) ((T_LIST) reduce 171) ((T_LNUMBER) reduce 171) ((T_MATCH) reduce 171) ((T_METHOD_C) reduce 171) ((T_NAME_FULLY_QUALIFIED) reduce 171) ((T_NAME_QUALIFIED) reduce 171) ((T_NAME_RELATIVE) reduce 171) ((T_NEW) reduce 171) ((T_NS_C) reduce 171) ((T_OBJECT_CAST) reduce 171) ((T_PRINT) reduce 171) ((T_PROPERTY_C) reduce 171) ((T_READONLY) reduce 171) ((T_REQUIRE) reduce 171) ((T_REQUIRE_ONCE) reduce 171) ((T_RETURN) reduce 171) ((T_START_HEREDOC) reduce 171) ((T_STATIC) reduce 171) ((T_STRING) reduce 171) ((T_STRING_CAST) reduce 171) ((T_SWITCH) reduce 171) ((T_THROW) reduce 171) ((T_TRAIT) reduce 171) ((T_TRAIT_C) reduce 171) ((T_TRY) reduce 171) ((T_UNSET) reduce 171) ((T_UNSET_CAST) reduce 171) ((T_VARIABLE) reduce 171) ((T_VOID_CAST) reduce 171) ((T_WHILE) reduce 171) ((T_YIELD) reduce 171) ((T_YIELD_FROM) reduce 171) (("[") reduce 171) (("`") reduce 171) (("{") reduce 171) (("~") reduce 171)) 27755 ((("!") reduce 177) (("\"") reduce 177) (("$") reduce 177) (("(") reduce 177) (("+") reduce 177) (("-") reduce 177) ((";") reduce 177) (("@") reduce 177) ((T_ABSTRACT) reduce 177) ((T_ARRAY) reduce 177) ((T_ARRAY_CAST) reduce 177) ((T_ATTRIBUTE) reduce 177) ((T_BOOL_CAST) reduce 177) ((T_BREAK) reduce 177) ((T_CATCH) shift) ((T_CLASS) reduce 177) ((T_CLASS_C) reduce 177) ((T_CLONE) reduce 177) ((T_CONSTANT_ENCAPSED_STRING) reduce 177) ((T_CONTINUE) reduce 177) ((T_DEC) reduce 177) ((T_DECLARE) reduce 177) ((T_DIR) reduce 177) ((T_DNUMBER) reduce 177) ((T_DO) reduce 177) ((T_DOUBLE_CAST) reduce 177) ((T_ECHO) reduce 177) ((T_ELSE) reduce 177) ((T_ELSEIF) reduce 177) ((T_EMPTY) reduce 177) ((T_ENDIF) reduce 177) ((T_ENUM) reduce 177) ((T_EVAL) reduce 177) ((T_EXIT) reduce 177) ((T_FILE) reduce 177) ((T_FINAL) reduce 177) ((T_FINALLY) shift) ((T_FN) reduce 177) ((T_FOR) reduce 177) ((T_FOREACH) reduce 177) ((T_FUNCTION) reduce 177) ((T_FUNC_C) reduce 177) ((T_GLOBAL) reduce 177) ((T_GOTO) reduce 177) ((T_HALT_COMPILER) reduce 177) ((T_IF) reduce 177) ((T_INC) reduce 177) ((T_INCLUDE) reduce 177) ((T_INCLUDE_ONCE) reduce 177) ((T_INLINE_HTML) reduce 177) ((T_INTERFACE) reduce 177) ((T_INT_CAST) reduce 177) ((T_ISSET) reduce 177) ((T_LINE) reduce 177) ((T_LIST) reduce 177) ((T_LNUMBER) reduce 177) ((T_MATCH) reduce 177) ((T_METHOD_C) reduce 177) ((T_NAME_FULLY_QUALIFIED) reduce 177) ((T_NAME_QUALIFIED) reduce 177) ((T_NAME_RELATIVE) reduce 177) ((T_NEW) reduce 177) ((T_NS_C) reduce 177) ((T_OBJECT_CAST) reduce 177) ((T_PRINT) reduce 177) ((T_PROPERTY_C) reduce 177) ((T_READONLY) reduce 177) ((T_REQUIRE) reduce 177) ((T_REQUIRE_ONCE) reduce 177) ((T_RETURN) reduce 177) ((T_START_HEREDOC) reduce 177) ((T_STATIC) reduce 177) ((T_STRING) reduce 177) ((T_STRING_CAST) reduce 177) ((T_SWITCH) reduce 177) ((T_THROW) reduce 177) ((T_TRAIT) reduce 177) ((T_TRAIT_C) reduce 177) ((T_TRY) reduce 177) ((T_UNSET) reduce 177) ((T_UNSET_CAST) reduce 177) ((T_VARIABLE) reduce 177) ((T_VOID_CAST) reduce 177) ((T_WHILE) reduce 177) ((T_YIELD) reduce 177) ((T_YIELD_FROM) reduce 177) (("[") reduce 177) (("`") reduce 177) (("{") reduce 177) (("~") reduce 177)) 27758 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_ELSE) reduce 167) ((T_ELSEIF) reduce 167) ((T_EMPTY) reduce 167) ((T_ENDIF) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_VOID_CAST) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 27761 ((("!") reduce 178) (("\"") reduce 178) (("$") reduce 178) (("(") reduce 178) (("+") reduce 178) (("-") reduce 178) ((";") reduce 178) (("@") reduce 178) ((T_ABSTRACT) reduce 178) ((T_ARRAY) reduce 178) ((T_ARRAY_CAST) reduce 178) ((T_ATTRIBUTE) reduce 178) ((T_BOOL_CAST) reduce 178) ((T_BREAK) reduce 178) ((T_CLASS) reduce 178) ((T_CLASS_C) reduce 178) ((T_CLONE) reduce 178) ((T_CONSTANT_ENCAPSED_STRING) reduce 178) ((T_CONTINUE) reduce 178) ((T_DEC) reduce 178) ((T_DECLARE) reduce 178) ((T_DIR) reduce 178) ((T_DNUMBER) reduce 178) ((T_DO) reduce 178) ((T_DOUBLE_CAST) reduce 178) ((T_ECHO) reduce 178) ((T_ELSE) reduce 178) ((T_ELSEIF) reduce 178) ((T_EMPTY) reduce 178) ((T_ENDIF) reduce 178) ((T_ENUM) reduce 178) ((T_EVAL) reduce 178) ((T_EXIT) reduce 178) ((T_FILE) reduce 178) ((T_FINAL) reduce 178) ((T_FN) reduce 178) ((T_FOR) reduce 178) ((T_FOREACH) reduce 178) ((T_FUNCTION) reduce 178) ((T_FUNC_C) reduce 178) ((T_GLOBAL) reduce 178) ((T_GOTO) reduce 178) ((T_HALT_COMPILER) reduce 178) ((T_IF) reduce 178) ((T_INC) reduce 178) ((T_INCLUDE) reduce 178) ((T_INCLUDE_ONCE) reduce 178) ((T_INLINE_HTML) reduce 178) ((T_INTERFACE) reduce 178) ((T_INT_CAST) reduce 178) ((T_ISSET) reduce 178) ((T_LINE) reduce 178) ((T_LIST) reduce 178) ((T_LNUMBER) reduce 178) ((T_MATCH) reduce 178) ((T_METHOD_C) reduce 178) ((T_NAME_FULLY_QUALIFIED) reduce 178) ((T_NAME_QUALIFIED) reduce 178) ((T_NAME_RELATIVE) reduce 178) ((T_NEW) reduce 178) ((T_NS_C) reduce 178) ((T_OBJECT_CAST) reduce 178) ((T_PRINT) reduce 178) ((T_PROPERTY_C) reduce 178) ((T_READONLY) reduce 178) ((T_REQUIRE) reduce 178) ((T_REQUIRE_ONCE) reduce 178) ((T_RETURN) reduce 178) ((T_START_HEREDOC) reduce 178) ((T_STATIC) reduce 178) ((T_STRING) reduce 178) ((T_STRING_CAST) reduce 178) ((T_SWITCH) reduce 178) ((T_THROW) reduce 178) ((T_TRAIT) reduce 178) ((T_TRAIT_C) reduce 178) ((T_TRY) reduce 178) ((T_UNSET) reduce 178) ((T_UNSET_CAST) reduce 178) ((T_VARIABLE) reduce 178) ((T_VOID_CAST) reduce 178) ((T_WHILE) reduce 178) ((T_YIELD) reduce 178) ((T_YIELD_FROM) reduce 178) (("[") reduce 178) (("`") reduce 178) (("{") reduce 178) (("~") reduce 178)) 27768 ((("!") reduce 172) (("\"") reduce 172) (("$") reduce 172) (("(") reduce 172) (("+") reduce 172) (("-") reduce 172) ((";") reduce 172) (("@") reduce 172) ((T_ABSTRACT) reduce 172) ((T_ARRAY) reduce 172) ((T_ARRAY_CAST) reduce 172) ((T_ATTRIBUTE) reduce 172) ((T_BOOL_CAST) reduce 172) ((T_BREAK) reduce 172) ((T_CATCH) reduce 172) ((T_CLASS) reduce 172) ((T_CLASS_C) reduce 172) ((T_CLONE) reduce 172) ((T_CONSTANT_ENCAPSED_STRING) reduce 172) ((T_CONTINUE) reduce 172) ((T_DEC) reduce 172) ((T_DECLARE) reduce 172) ((T_DIR) reduce 172) ((T_DNUMBER) reduce 172) ((T_DO) reduce 172) ((T_DOUBLE_CAST) reduce 172) ((T_ECHO) reduce 172) ((T_ELSE) reduce 172) ((T_ELSEIF) reduce 172) ((T_EMPTY) reduce 172) ((T_ENDIF) reduce 172) ((T_ENUM) reduce 172) ((T_EVAL) reduce 172) ((T_EXIT) reduce 172) ((T_FILE) reduce 172) ((T_FINAL) reduce 172) ((T_FINALLY) reduce 172) ((T_FN) reduce 172) ((T_FOR) reduce 172) ((T_FOREACH) reduce 172) ((T_FUNCTION) reduce 172) ((T_FUNC_C) reduce 172) ((T_GLOBAL) reduce 172) ((T_GOTO) reduce 172) ((T_HALT_COMPILER) reduce 172) ((T_IF) reduce 172) ((T_INC) reduce 172) ((T_INCLUDE) reduce 172) ((T_INCLUDE_ONCE) reduce 172) ((T_INLINE_HTML) reduce 172) ((T_INTERFACE) reduce 172) ((T_INT_CAST) reduce 172) ((T_ISSET) reduce 172) ((T_LINE) reduce 172) ((T_LIST) reduce 172) ((T_LNUMBER) reduce 172) ((T_MATCH) reduce 172) ((T_METHOD_C) reduce 172) ((T_NAME_FULLY_QUALIFIED) reduce 172) ((T_NAME_QUALIFIED) reduce 172) ((T_NAME_RELATIVE) reduce 172) ((T_NEW) reduce 172) ((T_NS_C) reduce 172) ((T_OBJECT_CAST) reduce 172) ((T_PRINT) reduce 172) ((T_PROPERTY_C) reduce 172) ((T_READONLY) reduce 172) ((T_REQUIRE) reduce 172) ((T_REQUIRE_ONCE) reduce 172) ((T_RETURN) reduce 172) ((T_START_HEREDOC) reduce 172) ((T_STATIC) reduce 172) ((T_STRING) reduce 172) ((T_STRING_CAST) reduce 172) ((T_SWITCH) reduce 172) ((T_THROW) reduce 172) ((T_TRAIT) reduce 172) ((T_TRAIT_C) reduce 172) ((T_TRY) reduce 172) ((T_UNSET) reduce 172) ((T_UNSET_CAST) reduce 172) ((T_VARIABLE) reduce 172) ((T_VOID_CAST) reduce 172) ((T_WHILE) reduce 172) ((T_YIELD) reduce 172) ((T_YIELD_FROM) reduce 172) (("[") reduce 172) (("`") reduce 172) (("{") reduce 172) (("~") reduce 172)) 27773 ((("!") reduce 200) (("\"") reduce 200) (("$") reduce 200) (("(") reduce 200) (("+") reduce 200) (("-") reduce 200) ((";") reduce 200) (("@") reduce 200) ((T_ABSTRACT) reduce 200) ((T_ARRAY) reduce 200) ((T_ARRAY_CAST) reduce 200) ((T_ATTRIBUTE) reduce 200) ((T_BOOL_CAST) reduce 200) ((T_BREAK) reduce 200) ((T_CLASS) reduce 200) ((T_CLASS_C) reduce 200) ((T_CLONE) reduce 200) ((T_CONSTANT_ENCAPSED_STRING) reduce 200) ((T_CONTINUE) reduce 200) ((T_DEC) reduce 200) ((T_DECLARE) reduce 200) ((T_DIR) reduce 200) ((T_DNUMBER) reduce 200) ((T_DO) reduce 200) ((T_DOUBLE_CAST) reduce 200) ((T_ECHO) reduce 200) ((T_ELSE) reduce 200) ((T_ELSEIF) reduce 200) ((T_EMPTY) reduce 200) ((T_ENDIF) reduce 200) ((T_ENUM) reduce 200) ((T_EVAL) reduce 200) ((T_EXIT) reduce 200) ((T_FILE) reduce 200) ((T_FINAL) reduce 200) ((T_FN) reduce 200) ((T_FOR) reduce 200) ((T_FOREACH) reduce 200) ((T_FUNCTION) reduce 200) ((T_FUNC_C) reduce 200) ((T_GLOBAL) reduce 200) ((T_GOTO) reduce 200) ((T_HALT_COMPILER) reduce 200) ((T_IF) reduce 200) ((T_INC) reduce 200) ((T_INCLUDE) reduce 200) ((T_INCLUDE_ONCE) reduce 200) ((T_INLINE_HTML) reduce 200) ((T_INTERFACE) reduce 200) ((T_INT_CAST) reduce 200) ((T_ISSET) reduce 200) ((T_LINE) reduce 200) ((T_LIST) reduce 200) ((T_LNUMBER) reduce 200) ((T_MATCH) reduce 200) ((T_METHOD_C) reduce 200) ((T_NAME_FULLY_QUALIFIED) reduce 200) ((T_NAME_QUALIFIED) reduce 200) ((T_NAME_RELATIVE) reduce 200) ((T_NEW) reduce 200) ((T_NS_C) reduce 200) ((T_OBJECT_CAST) reduce 200) ((T_PRINT) reduce 200) ((T_PROPERTY_C) reduce 200) ((T_READONLY) reduce 200) ((T_REQUIRE) reduce 200) ((T_REQUIRE_ONCE) reduce 200) ((T_RETURN) reduce 200) ((T_START_HEREDOC) reduce 200) ((T_STATIC) reduce 200) ((T_STRING) reduce 200) ((T_STRING_CAST) reduce 200) ((T_SWITCH) reduce 200) ((T_THROW) reduce 200) ((T_TRAIT) reduce 200) ((T_TRAIT_C) reduce 200) ((T_TRY) reduce 200) ((T_UNSET) reduce 200) ((T_UNSET_CAST) reduce 200) ((T_VARIABLE) reduce 200) ((T_VOID_CAST) reduce 200) ((T_WHILE) reduce 200) ((T_YIELD) reduce 200) ((T_YIELD_FROM) reduce 200) (("[") reduce 200) (("`") reduce 200) (("{") reduce 200) (("~") reduce 200)) 27778 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_ELSE) reduce 153) ((T_ELSEIF) reduce 153) ((T_EMPTY) reduce 153) ((T_ENDIF) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_VOID_CAST) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 27782 ((("!") reduce 224) (("\"") reduce 224) (("$") reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_ELSE) reduce 224) ((T_ELSEIF) reduce 224) ((T_EMPTY) reduce 224) ((T_ENDIF) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_VARIABLE) reduce 224) ((T_VOID_CAST) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("~") reduce 224)) 27784 ((("!") reduce 225) (("\"") reduce 225) (("$") reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_ELSE) reduce 225) ((T_ELSEIF) reduce 225) ((T_EMPTY) reduce 225) ((T_ENDIF) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_VARIABLE) reduce 225) ((T_VOID_CAST) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("~") reduce 225)) 27788 ((("!") reduce 226) (("\"") reduce 226) (("$") reduce 226) (("(") reduce 226) (("+") reduce 226) (("-") reduce 226) ((";") reduce 226) (("@") reduce 226) ((T_ABSTRACT) reduce 226) ((T_ARRAY) reduce 226) ((T_ARRAY_CAST) reduce 226) ((T_ATTRIBUTE) reduce 226) ((T_BOOL_CAST) reduce 226) ((T_BREAK) reduce 226) ((T_CLASS) reduce 226) ((T_CLASS_C) reduce 226) ((T_CLONE) reduce 226) ((T_CONSTANT_ENCAPSED_STRING) reduce 226) ((T_CONTINUE) reduce 226) ((T_DEC) reduce 226) ((T_DECLARE) reduce 226) ((T_DIR) reduce 226) ((T_DNUMBER) reduce 226) ((T_DO) reduce 226) ((T_DOUBLE_CAST) reduce 226) ((T_ECHO) reduce 226) ((T_ELSE) reduce 226) ((T_ELSEIF) reduce 226) ((T_EMPTY) reduce 226) ((T_ENDIF) reduce 226) ((T_ENUM) reduce 226) ((T_EVAL) reduce 226) ((T_EXIT) reduce 226) ((T_FILE) reduce 226) ((T_FINAL) reduce 226) ((T_FN) reduce 226) ((T_FOR) reduce 226) ((T_FOREACH) reduce 226) ((T_FUNCTION) reduce 226) ((T_FUNC_C) reduce 226) ((T_GLOBAL) reduce 226) ((T_GOTO) reduce 226) ((T_HALT_COMPILER) reduce 226) ((T_IF) reduce 226) ((T_INC) reduce 226) ((T_INCLUDE) reduce 226) ((T_INCLUDE_ONCE) reduce 226) ((T_INLINE_HTML) reduce 226) ((T_INTERFACE) reduce 226) ((T_INT_CAST) reduce 226) ((T_ISSET) reduce 226) ((T_LINE) reduce 226) ((T_LIST) reduce 226) ((T_LNUMBER) reduce 226) ((T_MATCH) reduce 226) ((T_METHOD_C) reduce 226) ((T_NAME_FULLY_QUALIFIED) reduce 226) ((T_NAME_QUALIFIED) reduce 226) ((T_NAME_RELATIVE) reduce 226) ((T_NEW) reduce 226) ((T_NS_C) reduce 226) ((T_OBJECT_CAST) reduce 226) ((T_PRINT) reduce 226) ((T_PROPERTY_C) reduce 226) ((T_READONLY) reduce 226) ((T_REQUIRE) reduce 226) ((T_REQUIRE_ONCE) reduce 226) ((T_RETURN) reduce 226) ((T_START_HEREDOC) reduce 226) ((T_STATIC) reduce 226) ((T_STRING) reduce 226) ((T_STRING_CAST) reduce 226) ((T_SWITCH) reduce 226) ((T_THROW) reduce 226) ((T_TRAIT) reduce 226) ((T_TRAIT_C) reduce 226) ((T_TRY) reduce 226) ((T_UNSET) reduce 226) ((T_UNSET_CAST) reduce 226) ((T_VARIABLE) reduce 226) ((T_VOID_CAST) reduce 226) ((T_WHILE) reduce 226) ((T_YIELD) reduce 226) ((T_YIELD_FROM) reduce 226) (("[") reduce 226) (("`") reduce 226) (("{") reduce 226) (("~") reduce 226)) 27791 ((("!") reduce 227) (("\"") reduce 227) (("$") reduce 227) (("(") reduce 227) (("+") reduce 227) (("-") reduce 227) ((";") reduce 227) (("@") reduce 227) ((T_ABSTRACT) reduce 227) ((T_ARRAY) reduce 227) ((T_ARRAY_CAST) reduce 227) ((T_ATTRIBUTE) reduce 227) ((T_BOOL_CAST) reduce 227) ((T_BREAK) reduce 227) ((T_CLASS) reduce 227) ((T_CLASS_C) reduce 227) ((T_CLONE) reduce 227) ((T_CONSTANT_ENCAPSED_STRING) reduce 227) ((T_CONTINUE) reduce 227) ((T_DEC) reduce 227) ((T_DECLARE) reduce 227) ((T_DIR) reduce 227) ((T_DNUMBER) reduce 227) ((T_DO) reduce 227) ((T_DOUBLE_CAST) reduce 227) ((T_ECHO) reduce 227) ((T_ELSE) reduce 227) ((T_ELSEIF) reduce 227) ((T_EMPTY) reduce 227) ((T_ENDIF) reduce 227) ((T_ENUM) reduce 227) ((T_EVAL) reduce 227) ((T_EXIT) reduce 227) ((T_FILE) reduce 227) ((T_FINAL) reduce 227) ((T_FN) reduce 227) ((T_FOR) reduce 227) ((T_FOREACH) reduce 227) ((T_FUNCTION) reduce 227) ((T_FUNC_C) reduce 227) ((T_GLOBAL) reduce 227) ((T_GOTO) reduce 227) ((T_HALT_COMPILER) reduce 227) ((T_IF) reduce 227) ((T_INC) reduce 227) ((T_INCLUDE) reduce 227) ((T_INCLUDE_ONCE) reduce 227) ((T_INLINE_HTML) reduce 227) ((T_INTERFACE) reduce 227) ((T_INT_CAST) reduce 227) ((T_ISSET) reduce 227) ((T_LINE) reduce 227) ((T_LIST) reduce 227) ((T_LNUMBER) reduce 227) ((T_MATCH) reduce 227) ((T_METHOD_C) reduce 227) ((T_NAME_FULLY_QUALIFIED) reduce 227) ((T_NAME_QUALIFIED) reduce 227) ((T_NAME_RELATIVE) reduce 227) ((T_NEW) reduce 227) ((T_NS_C) reduce 227) ((T_OBJECT_CAST) reduce 227) ((T_PRINT) reduce 227) ((T_PROPERTY_C) reduce 227) ((T_READONLY) reduce 227) ((T_REQUIRE) reduce 227) ((T_REQUIRE_ONCE) reduce 227) ((T_RETURN) reduce 227) ((T_START_HEREDOC) reduce 227) ((T_STATIC) reduce 227) ((T_STRING) reduce 227) ((T_STRING_CAST) reduce 227) ((T_SWITCH) reduce 227) ((T_THROW) reduce 227) ((T_TRAIT) reduce 227) ((T_TRAIT_C) reduce 227) ((T_TRY) reduce 227) ((T_UNSET) reduce 227) ((T_UNSET_CAST) reduce 227) ((T_VARIABLE) reduce 227) ((T_VOID_CAST) reduce 227) ((T_WHILE) reduce 227) ((T_YIELD) reduce 227) ((T_YIELD_FROM) reduce 227) (("[") reduce 227) (("`") reduce 227) (("{") reduce 227) (("~") reduce 227)) 27792 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_ELSE) reduce 169) ((T_ELSEIF) reduce 169) ((T_EMPTY) reduce 169) ((T_ENDIF) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_VOID_CAST) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 27794 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_ELSE) reduce 158) ((T_ELSEIF) reduce 158) ((T_EMPTY) reduce 158) ((T_ENDIF) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_VOID_CAST) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 27796 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_ELSE) reduce 156) ((T_ELSEIF) reduce 156) ((T_EMPTY) reduce 156) ((T_ENDIF) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_VOID_CAST) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 27802 ((("!") reduce 201) (("\"") reduce 201) (("$") reduce 201) (("(") reduce 201) (("+") reduce 201) (("-") reduce 201) ((";") reduce 201) (("@") reduce 201) ((T_ABSTRACT) reduce 201) ((T_ARRAY) reduce 201) ((T_ARRAY_CAST) reduce 201) ((T_ATTRIBUTE) reduce 201) ((T_BOOL_CAST) reduce 201) ((T_BREAK) reduce 201) ((T_CLASS) reduce 201) ((T_CLASS_C) reduce 201) ((T_CLONE) reduce 201) ((T_CONSTANT_ENCAPSED_STRING) reduce 201) ((T_CONTINUE) reduce 201) ((T_DEC) reduce 201) ((T_DECLARE) reduce 201) ((T_DIR) reduce 201) ((T_DNUMBER) reduce 201) ((T_DO) reduce 201) ((T_DOUBLE_CAST) reduce 201) ((T_ECHO) reduce 201) ((T_ELSE) reduce 201) ((T_ELSEIF) reduce 201) ((T_EMPTY) reduce 201) ((T_ENDIF) reduce 201) ((T_ENUM) reduce 201) ((T_EVAL) reduce 201) ((T_EXIT) reduce 201) ((T_FILE) reduce 201) ((T_FINAL) reduce 201) ((T_FN) reduce 201) ((T_FOR) reduce 201) ((T_FOREACH) reduce 201) ((T_FUNCTION) reduce 201) ((T_FUNC_C) reduce 201) ((T_GLOBAL) reduce 201) ((T_GOTO) reduce 201) ((T_HALT_COMPILER) reduce 201) ((T_IF) reduce 201) ((T_INC) reduce 201) ((T_INCLUDE) reduce 201) ((T_INCLUDE_ONCE) reduce 201) ((T_INLINE_HTML) reduce 201) ((T_INTERFACE) reduce 201) ((T_INT_CAST) reduce 201) ((T_ISSET) reduce 201) ((T_LINE) reduce 201) ((T_LIST) reduce 201) ((T_LNUMBER) reduce 201) ((T_MATCH) reduce 201) ((T_METHOD_C) reduce 201) ((T_NAME_FULLY_QUALIFIED) reduce 201) ((T_NAME_QUALIFIED) reduce 201) ((T_NAME_RELATIVE) reduce 201) ((T_NEW) reduce 201) ((T_NS_C) reduce 201) ((T_OBJECT_CAST) reduce 201) ((T_PRINT) reduce 201) ((T_PROPERTY_C) reduce 201) ((T_READONLY) reduce 201) ((T_REQUIRE) reduce 201) ((T_REQUIRE_ONCE) reduce 201) ((T_RETURN) reduce 201) ((T_START_HEREDOC) reduce 201) ((T_STATIC) reduce 201) ((T_STRING) reduce 201) ((T_STRING_CAST) reduce 201) ((T_SWITCH) reduce 201) ((T_THROW) reduce 201) ((T_TRAIT) reduce 201) ((T_TRAIT_C) reduce 201) ((T_TRY) reduce 201) ((T_UNSET) reduce 201) ((T_UNSET_CAST) reduce 201) ((T_VARIABLE) reduce 201) ((T_VOID_CAST) reduce 201) ((T_WHILE) reduce 201) ((T_YIELD) reduce 201) ((T_YIELD_FROM) reduce 201) (("[") reduce 201) (("`") reduce 201) (("{") reduce 201) (("~") reduce 201)) 27806 ((("!") reduce 244) (("\"") reduce 244) (("$") reduce 244) (("(") reduce 244) (("+") reduce 244) (("-") reduce 244) ((";") reduce 244) (("@") reduce 244) ((T_ABSTRACT) reduce 244) ((T_ARRAY) reduce 244) ((T_ARRAY_CAST) reduce 244) ((T_ATTRIBUTE) reduce 244) ((T_BOOL_CAST) reduce 244) ((T_BREAK) reduce 244) ((T_CLASS) reduce 244) ((T_CLASS_C) reduce 244) ((T_CLONE) reduce 244) ((T_CONSTANT_ENCAPSED_STRING) reduce 244) ((T_CONTINUE) reduce 244) ((T_DEC) reduce 244) ((T_DECLARE) reduce 244) ((T_DIR) reduce 244) ((T_DNUMBER) reduce 244) ((T_DO) reduce 244) ((T_DOUBLE_CAST) reduce 244) ((T_ECHO) reduce 244) ((T_ELSE) reduce 244) ((T_ELSEIF) reduce 244) ((T_EMPTY) reduce 244) ((T_ENDIF) reduce 244) ((T_ENUM) reduce 244) ((T_EVAL) reduce 244) ((T_EXIT) reduce 244) ((T_FILE) reduce 244) ((T_FINAL) reduce 244) ((T_FN) reduce 244) ((T_FOR) reduce 244) ((T_FOREACH) reduce 244) ((T_FUNCTION) reduce 244) ((T_FUNC_C) reduce 244) ((T_GLOBAL) reduce 244) ((T_GOTO) reduce 244) ((T_HALT_COMPILER) reduce 244) ((T_IF) reduce 244) ((T_INC) reduce 244) ((T_INCLUDE) reduce 244) ((T_INCLUDE_ONCE) reduce 244) ((T_INLINE_HTML) reduce 244) ((T_INTERFACE) reduce 244) ((T_INT_CAST) reduce 244) ((T_ISSET) reduce 244) ((T_LINE) reduce 244) ((T_LIST) reduce 244) ((T_LNUMBER) reduce 244) ((T_MATCH) reduce 244) ((T_METHOD_C) reduce 244) ((T_NAME_FULLY_QUALIFIED) reduce 244) ((T_NAME_QUALIFIED) reduce 244) ((T_NAME_RELATIVE) reduce 244) ((T_NEW) reduce 244) ((T_NS_C) reduce 244) ((T_OBJECT_CAST) reduce 244) ((T_PRINT) reduce 244) ((T_PROPERTY_C) reduce 244) ((T_READONLY) reduce 244) ((T_REQUIRE) reduce 244) ((T_REQUIRE_ONCE) reduce 244) ((T_RETURN) reduce 244) ((T_START_HEREDOC) reduce 244) ((T_STATIC) reduce 244) ((T_STRING) reduce 244) ((T_STRING_CAST) reduce 244) ((T_SWITCH) reduce 244) ((T_THROW) reduce 244) ((T_TRAIT) reduce 244) ((T_TRAIT_C) reduce 244) ((T_TRY) reduce 244) ((T_UNSET) reduce 244) ((T_UNSET_CAST) reduce 244) ((T_VARIABLE) reduce 244) ((T_VOID_CAST) reduce 244) ((T_WHILE) reduce 244) ((T_YIELD) reduce 244) ((T_YIELD_FROM) reduce 244) (("[") reduce 244) (("`") reduce 244) (("{") reduce 244) (("~") reduce 244)) 27809 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_ELSE) reduce 146) ((T_ELSEIF) reduce 146) ((T_EMPTY) reduce 146) ((T_ENDIF) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_PROPERTY_C) reduce 146) ((T_READONLY) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_VOID_CAST) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("~") reduce 146)) 27811 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_ELSE) reduce 168) ((T_ELSEIF) reduce 168) ((T_EMPTY) reduce 168) ((T_ENDIF) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_VOID_CAST) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 27813 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_ELSE) reduce 157) ((T_ELSEIF) reduce 157) ((T_EMPTY) reduce 157) ((T_ENDIF) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_VOID_CAST) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 27823 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_ELSE) reduce 164) ((T_ELSEIF) reduce 164) ((T_EMPTY) reduce 164) ((T_ENDIF) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_VOID_CAST) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 27824 ((("!") reduce 220) (("\"") reduce 220) (("$") reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_ELSE) reduce 220) ((T_ELSEIF) reduce 220) ((T_EMPTY) reduce 220) ((T_ENDIF) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_VARIABLE) reduce 220) ((T_VOID_CAST) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("~") reduce 220)) 27827 ((("!") reduce 221) (("\"") reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_ELSE) reduce 221) ((T_ELSEIF) reduce 221) ((T_EMPTY) reduce 221) ((T_ENDIF) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_VARIABLE) reduce 221) ((T_VOID_CAST) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("~") reduce 221)) 27828 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_ELSE) reduce 163) ((T_ELSEIF) reduce 163) ((T_EMPTY) reduce 163) ((T_ENDIF) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_VOID_CAST) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 27837 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_ELSE) reduce 152) ((T_ELSEIF) reduce 152) ((T_EMPTY) reduce 152) ((T_ENDIF) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_VOID_CAST) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 27838 ((("!") reduce 218) (("\"") reduce 218) (("$") reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_ELSE) reduce 218) ((T_ELSEIF) reduce 218) ((T_EMPTY) reduce 218) ((T_ENDIF) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_VARIABLE) reduce 218) ((T_VOID_CAST) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("~") reduce 218)) 27841 ((("!") reduce 219) (("\"") reduce 219) (("$") reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_ELSE) reduce 219) ((T_ELSEIF) reduce 219) ((T_EMPTY) reduce 219) ((T_ENDIF) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_VARIABLE) reduce 219) ((T_VOID_CAST) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("~") reduce 219)) 27848 ((("!") reduce 202) (("\"") reduce 202) (("$") reduce 202) (("(") reduce 202) (("+") reduce 202) (("-") reduce 202) ((";") reduce 202) (("@") reduce 202) ((T_ABSTRACT) reduce 202) ((T_ARRAY) reduce 202) ((T_ARRAY_CAST) reduce 202) ((T_ATTRIBUTE) reduce 202) ((T_BOOL_CAST) reduce 202) ((T_BREAK) reduce 202) ((T_CLASS) reduce 202) ((T_CLASS_C) reduce 202) ((T_CLONE) reduce 202) ((T_CONSTANT_ENCAPSED_STRING) reduce 202) ((T_CONTINUE) reduce 202) ((T_DEC) reduce 202) ((T_DECLARE) reduce 202) ((T_DIR) reduce 202) ((T_DNUMBER) reduce 202) ((T_DO) reduce 202) ((T_DOUBLE_CAST) reduce 202) ((T_ECHO) reduce 202) ((T_ELSE) reduce 202) ((T_ELSEIF) reduce 202) ((T_EMPTY) reduce 202) ((T_ENDIF) reduce 202) ((T_ENUM) reduce 202) ((T_EVAL) reduce 202) ((T_EXIT) reduce 202) ((T_FILE) reduce 202) ((T_FINAL) reduce 202) ((T_FN) reduce 202) ((T_FOR) reduce 202) ((T_FOREACH) reduce 202) ((T_FUNCTION) reduce 202) ((T_FUNC_C) reduce 202) ((T_GLOBAL) reduce 202) ((T_GOTO) reduce 202) ((T_HALT_COMPILER) reduce 202) ((T_IF) reduce 202) ((T_INC) reduce 202) ((T_INCLUDE) reduce 202) ((T_INCLUDE_ONCE) reduce 202) ((T_INLINE_HTML) reduce 202) ((T_INTERFACE) reduce 202) ((T_INT_CAST) reduce 202) ((T_ISSET) reduce 202) ((T_LINE) reduce 202) ((T_LIST) reduce 202) ((T_LNUMBER) reduce 202) ((T_MATCH) reduce 202) ((T_METHOD_C) reduce 202) ((T_NAME_FULLY_QUALIFIED) reduce 202) ((T_NAME_QUALIFIED) reduce 202) ((T_NAME_RELATIVE) reduce 202) ((T_NEW) reduce 202) ((T_NS_C) reduce 202) ((T_OBJECT_CAST) reduce 202) ((T_PRINT) reduce 202) ((T_PROPERTY_C) reduce 202) ((T_READONLY) reduce 202) ((T_REQUIRE) reduce 202) ((T_REQUIRE_ONCE) reduce 202) ((T_RETURN) reduce 202) ((T_START_HEREDOC) reduce 202) ((T_STATIC) reduce 202) ((T_STRING) reduce 202) ((T_STRING_CAST) reduce 202) ((T_SWITCH) reduce 202) ((T_THROW) reduce 202) ((T_TRAIT) reduce 202) ((T_TRAIT_C) reduce 202) ((T_TRY) reduce 202) ((T_UNSET) reduce 202) ((T_UNSET_CAST) reduce 202) ((T_VARIABLE) reduce 202) ((T_VOID_CAST) reduce 202) ((T_WHILE) reduce 202) ((T_YIELD) reduce 202) ((T_YIELD_FROM) reduce 202) (("[") reduce 202) (("`") reduce 202) (("{") reduce 202) (("~") reduce 202)) 27850 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_ELSE) reduce 159) ((T_ELSEIF) reduce 159) ((T_EMPTY) reduce 159) ((T_ENDIF) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_VOID_CAST) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 27856 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_ELSE) reduce 151) ((T_ELSEIF) reduce 151) ((T_EMPTY) reduce 151) ((T_ENDIF) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_VOID_CAST) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 27861 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_ELSE) reduce 165) ((T_ELSEIF) reduce 165) ((T_EMPTY) reduce 165) ((T_ENDIF) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_VOID_CAST) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 27862 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_ELSE) reduce 222) ((T_ELSEIF) reduce 222) ((T_EMPTY) reduce 222) ((T_ENDIF) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_VARIABLE) reduce 222) ((T_VOID_CAST) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("~") reduce 222)) 27865 ((("!") reduce 223) (("\"") reduce 223) (("$") reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_ELSE) reduce 223) ((T_ELSEIF) reduce 223) ((T_EMPTY) reduce 223) ((T_ENDIF) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_VARIABLE) reduce 223) ((T_VOID_CAST) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("~") reduce 223)) 27867 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_ELSE) reduce 155) ((T_ELSEIF) reduce 155) ((T_EMPTY) reduce 155) ((T_ENDIF) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_VOID_CAST) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 27874 ((("!") reduce 190) (("\"") reduce 190) (("$") reduce 190) (("(") reduce 190) (("+") reduce 190) (("-") reduce 190) ((";") reduce 190) (("@") reduce 190) ((T_ABSTRACT) reduce 190) ((T_ARRAY) reduce 190) ((T_ARRAY_CAST) reduce 190) ((T_ATTRIBUTE) reduce 190) ((T_BOOL_CAST) reduce 190) ((T_BREAK) reduce 190) ((T_CLASS) reduce 190) ((T_CLASS_C) reduce 190) ((T_CLONE) reduce 190) ((T_CONSTANT_ENCAPSED_STRING) reduce 190) ((T_CONTINUE) reduce 190) ((T_DEC) reduce 190) ((T_DECLARE) reduce 190) ((T_DIR) reduce 190) ((T_DNUMBER) reduce 190) ((T_DO) reduce 190) ((T_DOUBLE_CAST) reduce 190) ((T_ECHO) reduce 190) ((T_ELSE) reduce 190) ((T_ELSEIF) reduce 190) ((T_EMPTY) reduce 190) ((T_ENDIF) reduce 190) ((T_ENUM) reduce 190) ((T_EVAL) reduce 190) ((T_EXIT) reduce 190) ((T_FILE) reduce 190) ((T_FINAL) reduce 190) ((T_FN) reduce 190) ((T_FOR) reduce 190) ((T_FOREACH) reduce 190) ((T_FUNCTION) reduce 190) ((T_FUNC_C) reduce 190) ((T_GLOBAL) reduce 190) ((T_GOTO) reduce 190) ((T_HALT_COMPILER) reduce 190) ((T_IF) reduce 190) ((T_INC) reduce 190) ((T_INCLUDE) reduce 190) ((T_INCLUDE_ONCE) reduce 190) ((T_INLINE_HTML) reduce 190) ((T_INTERFACE) reduce 190) ((T_INT_CAST) reduce 190) ((T_ISSET) reduce 190) ((T_LINE) reduce 190) ((T_LIST) reduce 190) ((T_LNUMBER) reduce 190) ((T_MATCH) reduce 190) ((T_METHOD_C) reduce 190) ((T_NAME_FULLY_QUALIFIED) reduce 190) ((T_NAME_QUALIFIED) reduce 190) ((T_NAME_RELATIVE) reduce 190) ((T_NEW) reduce 190) ((T_NS_C) reduce 190) ((T_OBJECT_CAST) reduce 190) ((T_PRINT) reduce 190) ((T_PROPERTY_C) reduce 190) ((T_READONLY) reduce 190) ((T_REQUIRE) reduce 190) ((T_REQUIRE_ONCE) reduce 190) ((T_RETURN) reduce 190) ((T_START_HEREDOC) reduce 190) ((T_STATIC) reduce 190) ((T_STRING) reduce 190) ((T_STRING_CAST) reduce 190) ((T_SWITCH) reduce 190) ((T_THROW) reduce 190) ((T_TRAIT) reduce 190) ((T_TRAIT_C) reduce 190) ((T_TRY) reduce 190) ((T_UNSET) reduce 190) ((T_UNSET_CAST) reduce 190) ((T_VARIABLE) reduce 190) ((T_VOID_CAST) reduce 190) ((T_WHILE) reduce 190) ((T_YIELD) reduce 190) ((T_YIELD_FROM) reduce 190) (("[") reduce 190) (("`") reduce 190) (("{") reduce 190) (("~") reduce 190)) 27876 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_ELSE) reduce 154) ((T_ELSEIF) reduce 154) ((T_EMPTY) reduce 154) ((T_ENDIF) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_VOID_CAST) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 27880 ((("!") reduce 251) (("\"") reduce 251) (("$") reduce 251) (("(") reduce 251) (("+") reduce 251) (("-") reduce 251) ((";") reduce 251) (("@") reduce 251) ((T_ABSTRACT) reduce 251) ((T_ARRAY) reduce 251) ((T_ARRAY_CAST) reduce 251) ((T_ATTRIBUTE) reduce 251) ((T_BOOL_CAST) reduce 251) ((T_BREAK) reduce 251) ((T_CLASS) reduce 251) ((T_CLASS_C) reduce 251) ((T_CLONE) reduce 251) ((T_CONSTANT_ENCAPSED_STRING) reduce 251) ((T_CONTINUE) reduce 251) ((T_DEC) reduce 251) ((T_DECLARE) reduce 251) ((T_DIR) reduce 251) ((T_DNUMBER) reduce 251) ((T_DO) reduce 251) ((T_DOUBLE_CAST) reduce 251) ((T_ECHO) reduce 251) ((T_ELSE) reduce 251) ((T_ELSEIF) reduce 251) ((T_EMPTY) reduce 251) ((T_ENUM) reduce 251) ((T_EVAL) reduce 251) ((T_EXIT) reduce 251) ((T_FILE) reduce 251) ((T_FINAL) reduce 251) ((T_FN) reduce 251) ((T_FOR) reduce 251) ((T_FOREACH) reduce 251) ((T_FUNCTION) reduce 251) ((T_FUNC_C) reduce 251) ((T_GLOBAL) reduce 251) ((T_GOTO) reduce 251) ((T_HALT_COMPILER) reduce 251) ((T_IF) reduce 251) ((T_INC) reduce 251) ((T_INCLUDE) reduce 251) ((T_INCLUDE_ONCE) reduce 251) ((T_INLINE_HTML) reduce 251) ((T_INTERFACE) reduce 251) ((T_INT_CAST) reduce 251) ((T_ISSET) reduce 251) ((T_LINE) reduce 251) ((T_LIST) reduce 251) ((T_LNUMBER) reduce 251) ((T_MATCH) reduce 251) ((T_METHOD_C) reduce 251) ((T_NAME_FULLY_QUALIFIED) reduce 251) ((T_NAME_QUALIFIED) reduce 251) ((T_NAME_RELATIVE) reduce 251) ((T_NEW) reduce 251) ((T_NS_C) reduce 251) ((T_OBJECT_CAST) reduce 251) ((T_PRINT) reduce 251) ((T_PROPERTY_C) reduce 251) ((T_READONLY) reduce 251) ((T_REQUIRE) reduce 251) ((T_REQUIRE_ONCE) reduce 251) ((T_RETURN) reduce 251) ((T_START_HEREDOC) reduce 251) ((T_STATIC) reduce 251) ((T_STRING) reduce 251) ((T_STRING_CAST) reduce 251) ((T_SWITCH) reduce 251) ((T_THROW) reduce 251) ((T_TRAIT) reduce 251) ((T_TRAIT_C) reduce 251) ((T_TRY) reduce 251) ((T_UNSET) reduce 251) ((T_UNSET_CAST) reduce 251) ((T_VARIABLE) reduce 251) ((T_VOID_CAST) reduce 251) ((T_WHILE) reduce 251) ((T_YIELD) reduce 251) ((T_YIELD_FROM) reduce 251) (("[") reduce 251) (("`") reduce 251) (("{") reduce 251) (("}") reduce 251) (("~") reduce 251)) 27885 ((("!") reduce 242) (("\"") reduce 242) (("$") reduce 242) (("(") reduce 242) (("+") reduce 242) (("-") reduce 242) ((";") reduce 242) (("@") reduce 242) ((T_ABSTRACT) reduce 242) ((T_ARRAY) reduce 242) ((T_ARRAY_CAST) reduce 242) ((T_ATTRIBUTE) reduce 242) ((T_BOOL_CAST) reduce 242) ((T_BREAK) reduce 242) ((T_CLASS) reduce 242) ((T_CLASS_C) reduce 242) ((T_CLONE) reduce 242) ((T_CONSTANT_ENCAPSED_STRING) reduce 242) ((T_CONTINUE) reduce 242) ((T_DEC) reduce 242) ((T_DECLARE) reduce 242) ((T_DIR) reduce 242) ((T_DNUMBER) reduce 242) ((T_DO) reduce 242) ((T_DOUBLE_CAST) reduce 242) ((T_ECHO) reduce 242) ((T_ELSE) reduce 242) ((T_ELSEIF) reduce 242) ((T_EMPTY) reduce 242) ((T_ENUM) reduce 242) ((T_EVAL) reduce 242) ((T_EXIT) reduce 242) ((T_FILE) reduce 242) ((T_FINAL) reduce 242) ((T_FN) reduce 242) ((T_FOR) reduce 242) ((T_FOREACH) reduce 242) ((T_FUNCTION) reduce 242) ((T_FUNC_C) reduce 242) ((T_GLOBAL) reduce 242) ((T_GOTO) reduce 242) ((T_HALT_COMPILER) reduce 242) ((T_IF) reduce 242) ((T_INC) reduce 242) ((T_INCLUDE) reduce 242) ((T_INCLUDE_ONCE) reduce 242) ((T_INLINE_HTML) reduce 242) ((T_INTERFACE) reduce 242) ((T_INT_CAST) reduce 242) ((T_ISSET) reduce 242) ((T_LINE) reduce 242) ((T_LIST) reduce 242) ((T_LNUMBER) reduce 242) ((T_MATCH) reduce 242) ((T_METHOD_C) reduce 242) ((T_NAME_FULLY_QUALIFIED) reduce 242) ((T_NAME_QUALIFIED) reduce 242) ((T_NAME_RELATIVE) reduce 242) ((T_NEW) reduce 242) ((T_NS_C) reduce 242) ((T_OBJECT_CAST) reduce 242) ((T_PRINT) reduce 242) ((T_PROPERTY_C) reduce 242) ((T_READONLY) reduce 242) ((T_REQUIRE) reduce 242) ((T_REQUIRE_ONCE) reduce 242) ((T_RETURN) reduce 242) ((T_START_HEREDOC) reduce 242) ((T_STATIC) reduce 242) ((T_STRING) reduce 242) ((T_STRING_CAST) reduce 242) ((T_SWITCH) reduce 242) ((T_THROW) reduce 242) ((T_TRAIT) reduce 242) ((T_TRAIT_C) reduce 242) ((T_TRY) reduce 242) ((T_UNSET) reduce 242) ((T_UNSET_CAST) reduce 242) ((T_VARIABLE) reduce 242) ((T_VOID_CAST) reduce 242) ((T_WHILE) reduce 242) ((T_YIELD) reduce 242) ((T_YIELD_FROM) reduce 242) (("[") reduce 242) (("`") reduce 242) (("{") reduce 242) (("}") reduce 242) (("~") reduce 242)) 27886 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_ELSE) reduce 150) ((T_ELSEIF) reduce 150) ((T_EMPTY) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_VOID_CAST) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("}") reduce 150) (("~") reduce 150)) 27889 ((("!") reduce 243) (("\"") reduce 243) (("$") reduce 243) (("(") reduce 243) (("+") reduce 243) (("-") reduce 243) ((";") reduce 243) (("@") reduce 243) ((T_ABSTRACT) reduce 243) ((T_ARRAY) reduce 243) ((T_ARRAY_CAST) reduce 243) ((T_ATTRIBUTE) reduce 243) ((T_BOOL_CAST) reduce 243) ((T_BREAK) reduce 243) ((T_CLASS) reduce 243) ((T_CLASS_C) reduce 243) ((T_CLONE) reduce 243) ((T_CONSTANT_ENCAPSED_STRING) reduce 243) ((T_CONTINUE) reduce 243) ((T_DEC) reduce 243) ((T_DECLARE) reduce 243) ((T_DIR) reduce 243) ((T_DNUMBER) reduce 243) ((T_DO) reduce 243) ((T_DOUBLE_CAST) reduce 243) ((T_ECHO) reduce 243) ((T_ELSE) reduce 243) ((T_ELSEIF) reduce 243) ((T_EMPTY) reduce 243) ((T_ENUM) reduce 243) ((T_EVAL) reduce 243) ((T_EXIT) reduce 243) ((T_FILE) reduce 243) ((T_FINAL) reduce 243) ((T_FN) reduce 243) ((T_FOR) reduce 243) ((T_FOREACH) reduce 243) ((T_FUNCTION) reduce 243) ((T_FUNC_C) reduce 243) ((T_GLOBAL) reduce 243) ((T_GOTO) reduce 243) ((T_HALT_COMPILER) reduce 243) ((T_IF) reduce 243) ((T_INC) reduce 243) ((T_INCLUDE) reduce 243) ((T_INCLUDE_ONCE) reduce 243) ((T_INLINE_HTML) reduce 243) ((T_INTERFACE) reduce 243) ((T_INT_CAST) reduce 243) ((T_ISSET) reduce 243) ((T_LINE) reduce 243) ((T_LIST) reduce 243) ((T_LNUMBER) reduce 243) ((T_MATCH) reduce 243) ((T_METHOD_C) reduce 243) ((T_NAME_FULLY_QUALIFIED) reduce 243) ((T_NAME_QUALIFIED) reduce 243) ((T_NAME_RELATIVE) reduce 243) ((T_NEW) reduce 243) ((T_NS_C) reduce 243) ((T_OBJECT_CAST) reduce 243) ((T_PRINT) reduce 243) ((T_PROPERTY_C) reduce 243) ((T_READONLY) reduce 243) ((T_REQUIRE) reduce 243) ((T_REQUIRE_ONCE) reduce 243) ((T_RETURN) reduce 243) ((T_START_HEREDOC) reduce 243) ((T_STATIC) reduce 243) ((T_STRING) reduce 243) ((T_STRING_CAST) reduce 243) ((T_SWITCH) reduce 243) ((T_THROW) reduce 243) ((T_TRAIT) reduce 243) ((T_TRAIT_C) reduce 243) ((T_TRY) reduce 243) ((T_UNSET) reduce 243) ((T_UNSET_CAST) reduce 243) ((T_VARIABLE) reduce 243) ((T_VOID_CAST) reduce 243) ((T_WHILE) reduce 243) ((T_YIELD) reduce 243) ((T_YIELD_FROM) reduce 243) (("[") reduce 243) (("`") reduce 243) (("{") reduce 243) (("}") reduce 243) (("~") reduce 243)) 27891 ((("!") reduce 170) (("\"") reduce 170) (("$") reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_ELSE) reduce 170) ((T_ELSEIF) reduce 170) ((T_EMPTY) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_VARIABLE) reduce 170) ((T_VOID_CAST) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("}") reduce 170) (("~") reduce 170)) 27896 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_ELSE) reduce 162) ((T_ELSEIF) reduce 162) ((T_EMPTY) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_VOID_CAST) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("}") reduce 162) (("~") reduce 162)) 27899 ((("!") reduce 171) (("\"") reduce 171) (("$") reduce 171) (("(") reduce 171) (("+") reduce 171) (("-") reduce 171) ((";") reduce 171) (("@") reduce 171) ((T_ABSTRACT) reduce 171) ((T_ARRAY) reduce 171) ((T_ARRAY_CAST) reduce 171) ((T_ATTRIBUTE) reduce 171) ((T_BOOL_CAST) reduce 171) ((T_BREAK) reduce 171) ((T_CATCH) reduce 171) ((T_CLASS) reduce 171) ((T_CLASS_C) reduce 171) ((T_CLONE) reduce 171) ((T_CONSTANT_ENCAPSED_STRING) reduce 171) ((T_CONTINUE) reduce 171) ((T_DEC) reduce 171) ((T_DECLARE) reduce 171) ((T_DIR) reduce 171) ((T_DNUMBER) reduce 171) ((T_DO) reduce 171) ((T_DOUBLE_CAST) reduce 171) ((T_ECHO) reduce 171) ((T_ELSE) reduce 171) ((T_ELSEIF) reduce 171) ((T_EMPTY) reduce 171) ((T_ENUM) reduce 171) ((T_EVAL) reduce 171) ((T_EXIT) reduce 171) ((T_FILE) reduce 171) ((T_FINAL) reduce 171) ((T_FINALLY) reduce 171) ((T_FN) reduce 171) ((T_FOR) reduce 171) ((T_FOREACH) reduce 171) ((T_FUNCTION) reduce 171) ((T_FUNC_C) reduce 171) ((T_GLOBAL) reduce 171) ((T_GOTO) reduce 171) ((T_HALT_COMPILER) reduce 171) ((T_IF) reduce 171) ((T_INC) reduce 171) ((T_INCLUDE) reduce 171) ((T_INCLUDE_ONCE) reduce 171) ((T_INLINE_HTML) reduce 171) ((T_INTERFACE) reduce 171) ((T_INT_CAST) reduce 171) ((T_ISSET) reduce 171) ((T_LINE) reduce 171) ((T_LIST) reduce 171) ((T_LNUMBER) reduce 171) ((T_MATCH) reduce 171) ((T_METHOD_C) reduce 171) ((T_NAME_FULLY_QUALIFIED) reduce 171) ((T_NAME_QUALIFIED) reduce 171) ((T_NAME_RELATIVE) reduce 171) ((T_NEW) reduce 171) ((T_NS_C) reduce 171) ((T_OBJECT_CAST) reduce 171) ((T_PRINT) reduce 171) ((T_PROPERTY_C) reduce 171) ((T_READONLY) reduce 171) ((T_REQUIRE) reduce 171) ((T_REQUIRE_ONCE) reduce 171) ((T_RETURN) reduce 171) ((T_START_HEREDOC) reduce 171) ((T_STATIC) reduce 171) ((T_STRING) reduce 171) ((T_STRING_CAST) reduce 171) ((T_SWITCH) reduce 171) ((T_THROW) reduce 171) ((T_TRAIT) reduce 171) ((T_TRAIT_C) reduce 171) ((T_TRY) reduce 171) ((T_UNSET) reduce 171) ((T_UNSET_CAST) reduce 171) ((T_VARIABLE) reduce 171) ((T_VOID_CAST) reduce 171) ((T_WHILE) reduce 171) ((T_YIELD) reduce 171) ((T_YIELD_FROM) reduce 171) (("[") reduce 171) (("`") reduce 171) (("{") reduce 171) (("}") reduce 171) (("~") reduce 171)) 27900 ((("!") reduce 177) (("\"") reduce 177) (("$") reduce 177) (("(") reduce 177) (("+") reduce 177) (("-") reduce 177) ((";") reduce 177) (("@") reduce 177) ((T_ABSTRACT) reduce 177) ((T_ARRAY) reduce 177) ((T_ARRAY_CAST) reduce 177) ((T_ATTRIBUTE) reduce 177) ((T_BOOL_CAST) reduce 177) ((T_BREAK) reduce 177) ((T_CATCH) shift) ((T_CLASS) reduce 177) ((T_CLASS_C) reduce 177) ((T_CLONE) reduce 177) ((T_CONSTANT_ENCAPSED_STRING) reduce 177) ((T_CONTINUE) reduce 177) ((T_DEC) reduce 177) ((T_DECLARE) reduce 177) ((T_DIR) reduce 177) ((T_DNUMBER) reduce 177) ((T_DO) reduce 177) ((T_DOUBLE_CAST) reduce 177) ((T_ECHO) reduce 177) ((T_ELSE) reduce 177) ((T_ELSEIF) reduce 177) ((T_EMPTY) reduce 177) ((T_ENUM) reduce 177) ((T_EVAL) reduce 177) ((T_EXIT) reduce 177) ((T_FILE) reduce 177) ((T_FINAL) reduce 177) ((T_FINALLY) shift) ((T_FN) reduce 177) ((T_FOR) reduce 177) ((T_FOREACH) reduce 177) ((T_FUNCTION) reduce 177) ((T_FUNC_C) reduce 177) ((T_GLOBAL) reduce 177) ((T_GOTO) reduce 177) ((T_HALT_COMPILER) reduce 177) ((T_IF) reduce 177) ((T_INC) reduce 177) ((T_INCLUDE) reduce 177) ((T_INCLUDE_ONCE) reduce 177) ((T_INLINE_HTML) reduce 177) ((T_INTERFACE) reduce 177) ((T_INT_CAST) reduce 177) ((T_ISSET) reduce 177) ((T_LINE) reduce 177) ((T_LIST) reduce 177) ((T_LNUMBER) reduce 177) ((T_MATCH) reduce 177) ((T_METHOD_C) reduce 177) ((T_NAME_FULLY_QUALIFIED) reduce 177) ((T_NAME_QUALIFIED) reduce 177) ((T_NAME_RELATIVE) reduce 177) ((T_NEW) reduce 177) ((T_NS_C) reduce 177) ((T_OBJECT_CAST) reduce 177) ((T_PRINT) reduce 177) ((T_PROPERTY_C) reduce 177) ((T_READONLY) reduce 177) ((T_REQUIRE) reduce 177) ((T_REQUIRE_ONCE) reduce 177) ((T_RETURN) reduce 177) ((T_START_HEREDOC) reduce 177) ((T_STATIC) reduce 177) ((T_STRING) reduce 177) ((T_STRING_CAST) reduce 177) ((T_SWITCH) reduce 177) ((T_THROW) reduce 177) ((T_TRAIT) reduce 177) ((T_TRAIT_C) reduce 177) ((T_TRY) reduce 177) ((T_UNSET) reduce 177) ((T_UNSET_CAST) reduce 177) ((T_VARIABLE) reduce 177) ((T_VOID_CAST) reduce 177) ((T_WHILE) reduce 177) ((T_YIELD) reduce 177) ((T_YIELD_FROM) reduce 177) (("[") reduce 177) (("`") reduce 177) (("{") reduce 177) (("}") reduce 177) (("~") reduce 177)) 27903 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_ELSE) reduce 167) ((T_ELSEIF) reduce 167) ((T_EMPTY) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_VOID_CAST) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("}") reduce 167) (("~") reduce 167)) 27906 ((("!") reduce 178) (("\"") reduce 178) (("$") reduce 178) (("(") reduce 178) (("+") reduce 178) (("-") reduce 178) ((";") reduce 178) (("@") reduce 178) ((T_ABSTRACT) reduce 178) ((T_ARRAY) reduce 178) ((T_ARRAY_CAST) reduce 178) ((T_ATTRIBUTE) reduce 178) ((T_BOOL_CAST) reduce 178) ((T_BREAK) reduce 178) ((T_CLASS) reduce 178) ((T_CLASS_C) reduce 178) ((T_CLONE) reduce 178) ((T_CONSTANT_ENCAPSED_STRING) reduce 178) ((T_CONTINUE) reduce 178) ((T_DEC) reduce 178) ((T_DECLARE) reduce 178) ((T_DIR) reduce 178) ((T_DNUMBER) reduce 178) ((T_DO) reduce 178) ((T_DOUBLE_CAST) reduce 178) ((T_ECHO) reduce 178) ((T_ELSE) reduce 178) ((T_ELSEIF) reduce 178) ((T_EMPTY) reduce 178) ((T_ENUM) reduce 178) ((T_EVAL) reduce 178) ((T_EXIT) reduce 178) ((T_FILE) reduce 178) ((T_FINAL) reduce 178) ((T_FN) reduce 178) ((T_FOR) reduce 178) ((T_FOREACH) reduce 178) ((T_FUNCTION) reduce 178) ((T_FUNC_C) reduce 178) ((T_GLOBAL) reduce 178) ((T_GOTO) reduce 178) ((T_HALT_COMPILER) reduce 178) ((T_IF) reduce 178) ((T_INC) reduce 178) ((T_INCLUDE) reduce 178) ((T_INCLUDE_ONCE) reduce 178) ((T_INLINE_HTML) reduce 178) ((T_INTERFACE) reduce 178) ((T_INT_CAST) reduce 178) ((T_ISSET) reduce 178) ((T_LINE) reduce 178) ((T_LIST) reduce 178) ((T_LNUMBER) reduce 178) ((T_MATCH) reduce 178) ((T_METHOD_C) reduce 178) ((T_NAME_FULLY_QUALIFIED) reduce 178) ((T_NAME_QUALIFIED) reduce 178) ((T_NAME_RELATIVE) reduce 178) ((T_NEW) reduce 178) ((T_NS_C) reduce 178) ((T_OBJECT_CAST) reduce 178) ((T_PRINT) reduce 178) ((T_PROPERTY_C) reduce 178) ((T_READONLY) reduce 178) ((T_REQUIRE) reduce 178) ((T_REQUIRE_ONCE) reduce 178) ((T_RETURN) reduce 178) ((T_START_HEREDOC) reduce 178) ((T_STATIC) reduce 178) ((T_STRING) reduce 178) ((T_STRING_CAST) reduce 178) ((T_SWITCH) reduce 178) ((T_THROW) reduce 178) ((T_TRAIT) reduce 178) ((T_TRAIT_C) reduce 178) ((T_TRY) reduce 178) ((T_UNSET) reduce 178) ((T_UNSET_CAST) reduce 178) ((T_VARIABLE) reduce 178) ((T_VOID_CAST) reduce 178) ((T_WHILE) reduce 178) ((T_YIELD) reduce 178) ((T_YIELD_FROM) reduce 178) (("[") reduce 178) (("`") reduce 178) (("{") reduce 178) (("}") reduce 178) (("~") reduce 178)) 27913 ((("!") reduce 172) (("\"") reduce 172) (("$") reduce 172) (("(") reduce 172) (("+") reduce 172) (("-") reduce 172) ((";") reduce 172) (("@") reduce 172) ((T_ABSTRACT) reduce 172) ((T_ARRAY) reduce 172) ((T_ARRAY_CAST) reduce 172) ((T_ATTRIBUTE) reduce 172) ((T_BOOL_CAST) reduce 172) ((T_BREAK) reduce 172) ((T_CATCH) reduce 172) ((T_CLASS) reduce 172) ((T_CLASS_C) reduce 172) ((T_CLONE) reduce 172) ((T_CONSTANT_ENCAPSED_STRING) reduce 172) ((T_CONTINUE) reduce 172) ((T_DEC) reduce 172) ((T_DECLARE) reduce 172) ((T_DIR) reduce 172) ((T_DNUMBER) reduce 172) ((T_DO) reduce 172) ((T_DOUBLE_CAST) reduce 172) ((T_ECHO) reduce 172) ((T_ELSE) reduce 172) ((T_ELSEIF) reduce 172) ((T_EMPTY) reduce 172) ((T_ENUM) reduce 172) ((T_EVAL) reduce 172) ((T_EXIT) reduce 172) ((T_FILE) reduce 172) ((T_FINAL) reduce 172) ((T_FINALLY) reduce 172) ((T_FN) reduce 172) ((T_FOR) reduce 172) ((T_FOREACH) reduce 172) ((T_FUNCTION) reduce 172) ((T_FUNC_C) reduce 172) ((T_GLOBAL) reduce 172) ((T_GOTO) reduce 172) ((T_HALT_COMPILER) reduce 172) ((T_IF) reduce 172) ((T_INC) reduce 172) ((T_INCLUDE) reduce 172) ((T_INCLUDE_ONCE) reduce 172) ((T_INLINE_HTML) reduce 172) ((T_INTERFACE) reduce 172) ((T_INT_CAST) reduce 172) ((T_ISSET) reduce 172) ((T_LINE) reduce 172) ((T_LIST) reduce 172) ((T_LNUMBER) reduce 172) ((T_MATCH) reduce 172) ((T_METHOD_C) reduce 172) ((T_NAME_FULLY_QUALIFIED) reduce 172) ((T_NAME_QUALIFIED) reduce 172) ((T_NAME_RELATIVE) reduce 172) ((T_NEW) reduce 172) ((T_NS_C) reduce 172) ((T_OBJECT_CAST) reduce 172) ((T_PRINT) reduce 172) ((T_PROPERTY_C) reduce 172) ((T_READONLY) reduce 172) ((T_REQUIRE) reduce 172) ((T_REQUIRE_ONCE) reduce 172) ((T_RETURN) reduce 172) ((T_START_HEREDOC) reduce 172) ((T_STATIC) reduce 172) ((T_STRING) reduce 172) ((T_STRING_CAST) reduce 172) ((T_SWITCH) reduce 172) ((T_THROW) reduce 172) ((T_TRAIT) reduce 172) ((T_TRAIT_C) reduce 172) ((T_TRY) reduce 172) ((T_UNSET) reduce 172) ((T_UNSET_CAST) reduce 172) ((T_VARIABLE) reduce 172) ((T_VOID_CAST) reduce 172) ((T_WHILE) reduce 172) ((T_YIELD) reduce 172) ((T_YIELD_FROM) reduce 172) (("[") reduce 172) (("`") reduce 172) (("{") reduce 172) (("}") reduce 172) (("~") reduce 172)) 27918 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_ELSE) reduce 153) ((T_ELSEIF) reduce 153) ((T_EMPTY) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_VOID_CAST) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("}") reduce 153) (("~") reduce 153)) 27922 ((("!") reduce 224) (("\"") reduce 224) (("$") reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_ELSE) reduce 224) ((T_ELSEIF) reduce 224) ((T_EMPTY) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_VARIABLE) reduce 224) ((T_VOID_CAST) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("}") reduce 224) (("~") reduce 224)) 27924 ((("!") reduce 225) (("\"") reduce 225) (("$") reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_ELSE) reduce 225) ((T_ELSEIF) reduce 225) ((T_EMPTY) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_VARIABLE) reduce 225) ((T_VOID_CAST) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("}") reduce 225) (("~") reduce 225)) 27928 ((("!") reduce 226) (("\"") reduce 226) (("$") reduce 226) (("(") reduce 226) (("+") reduce 226) (("-") reduce 226) ((";") reduce 226) (("@") reduce 226) ((T_ABSTRACT) reduce 226) ((T_ARRAY) reduce 226) ((T_ARRAY_CAST) reduce 226) ((T_ATTRIBUTE) reduce 226) ((T_BOOL_CAST) reduce 226) ((T_BREAK) reduce 226) ((T_CLASS) reduce 226) ((T_CLASS_C) reduce 226) ((T_CLONE) reduce 226) ((T_CONSTANT_ENCAPSED_STRING) reduce 226) ((T_CONTINUE) reduce 226) ((T_DEC) reduce 226) ((T_DECLARE) reduce 226) ((T_DIR) reduce 226) ((T_DNUMBER) reduce 226) ((T_DO) reduce 226) ((T_DOUBLE_CAST) reduce 226) ((T_ECHO) reduce 226) ((T_ELSE) reduce 226) ((T_ELSEIF) reduce 226) ((T_EMPTY) reduce 226) ((T_ENUM) reduce 226) ((T_EVAL) reduce 226) ((T_EXIT) reduce 226) ((T_FILE) reduce 226) ((T_FINAL) reduce 226) ((T_FN) reduce 226) ((T_FOR) reduce 226) ((T_FOREACH) reduce 226) ((T_FUNCTION) reduce 226) ((T_FUNC_C) reduce 226) ((T_GLOBAL) reduce 226) ((T_GOTO) reduce 226) ((T_HALT_COMPILER) reduce 226) ((T_IF) reduce 226) ((T_INC) reduce 226) ((T_INCLUDE) reduce 226) ((T_INCLUDE_ONCE) reduce 226) ((T_INLINE_HTML) reduce 226) ((T_INTERFACE) reduce 226) ((T_INT_CAST) reduce 226) ((T_ISSET) reduce 226) ((T_LINE) reduce 226) ((T_LIST) reduce 226) ((T_LNUMBER) reduce 226) ((T_MATCH) reduce 226) ((T_METHOD_C) reduce 226) ((T_NAME_FULLY_QUALIFIED) reduce 226) ((T_NAME_QUALIFIED) reduce 226) ((T_NAME_RELATIVE) reduce 226) ((T_NEW) reduce 226) ((T_NS_C) reduce 226) ((T_OBJECT_CAST) reduce 226) ((T_PRINT) reduce 226) ((T_PROPERTY_C) reduce 226) ((T_READONLY) reduce 226) ((T_REQUIRE) reduce 226) ((T_REQUIRE_ONCE) reduce 226) ((T_RETURN) reduce 226) ((T_START_HEREDOC) reduce 226) ((T_STATIC) reduce 226) ((T_STRING) reduce 226) ((T_STRING_CAST) reduce 226) ((T_SWITCH) reduce 226) ((T_THROW) reduce 226) ((T_TRAIT) reduce 226) ((T_TRAIT_C) reduce 226) ((T_TRY) reduce 226) ((T_UNSET) reduce 226) ((T_UNSET_CAST) reduce 226) ((T_VARIABLE) reduce 226) ((T_VOID_CAST) reduce 226) ((T_WHILE) reduce 226) ((T_YIELD) reduce 226) ((T_YIELD_FROM) reduce 226) (("[") reduce 226) (("`") reduce 226) (("{") reduce 226) (("}") reduce 226) (("~") reduce 226)) 27931 ((("!") reduce 227) (("\"") reduce 227) (("$") reduce 227) (("(") reduce 227) (("+") reduce 227) (("-") reduce 227) ((";") reduce 227) (("@") reduce 227) ((T_ABSTRACT) reduce 227) ((T_ARRAY) reduce 227) ((T_ARRAY_CAST) reduce 227) ((T_ATTRIBUTE) reduce 227) ((T_BOOL_CAST) reduce 227) ((T_BREAK) reduce 227) ((T_CLASS) reduce 227) ((T_CLASS_C) reduce 227) ((T_CLONE) reduce 227) ((T_CONSTANT_ENCAPSED_STRING) reduce 227) ((T_CONTINUE) reduce 227) ((T_DEC) reduce 227) ((T_DECLARE) reduce 227) ((T_DIR) reduce 227) ((T_DNUMBER) reduce 227) ((T_DO) reduce 227) ((T_DOUBLE_CAST) reduce 227) ((T_ECHO) reduce 227) ((T_ELSE) reduce 227) ((T_ELSEIF) reduce 227) ((T_EMPTY) reduce 227) ((T_ENUM) reduce 227) ((T_EVAL) reduce 227) ((T_EXIT) reduce 227) ((T_FILE) reduce 227) ((T_FINAL) reduce 227) ((T_FN) reduce 227) ((T_FOR) reduce 227) ((T_FOREACH) reduce 227) ((T_FUNCTION) reduce 227) ((T_FUNC_C) reduce 227) ((T_GLOBAL) reduce 227) ((T_GOTO) reduce 227) ((T_HALT_COMPILER) reduce 227) ((T_IF) reduce 227) ((T_INC) reduce 227) ((T_INCLUDE) reduce 227) ((T_INCLUDE_ONCE) reduce 227) ((T_INLINE_HTML) reduce 227) ((T_INTERFACE) reduce 227) ((T_INT_CAST) reduce 227) ((T_ISSET) reduce 227) ((T_LINE) reduce 227) ((T_LIST) reduce 227) ((T_LNUMBER) reduce 227) ((T_MATCH) reduce 227) ((T_METHOD_C) reduce 227) ((T_NAME_FULLY_QUALIFIED) reduce 227) ((T_NAME_QUALIFIED) reduce 227) ((T_NAME_RELATIVE) reduce 227) ((T_NEW) reduce 227) ((T_NS_C) reduce 227) ((T_OBJECT_CAST) reduce 227) ((T_PRINT) reduce 227) ((T_PROPERTY_C) reduce 227) ((T_READONLY) reduce 227) ((T_REQUIRE) reduce 227) ((T_REQUIRE_ONCE) reduce 227) ((T_RETURN) reduce 227) ((T_START_HEREDOC) reduce 227) ((T_STATIC) reduce 227) ((T_STRING) reduce 227) ((T_STRING_CAST) reduce 227) ((T_SWITCH) reduce 227) ((T_THROW) reduce 227) ((T_TRAIT) reduce 227) ((T_TRAIT_C) reduce 227) ((T_TRY) reduce 227) ((T_UNSET) reduce 227) ((T_UNSET_CAST) reduce 227) ((T_VARIABLE) reduce 227) ((T_VOID_CAST) reduce 227) ((T_WHILE) reduce 227) ((T_YIELD) reduce 227) ((T_YIELD_FROM) reduce 227) (("[") reduce 227) (("`") reduce 227) (("{") reduce 227) (("}") reduce 227) (("~") reduce 227)) 27932 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_ELSE) reduce 169) ((T_ELSEIF) reduce 169) ((T_EMPTY) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_VOID_CAST) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("}") reduce 169) (("~") reduce 169)) 27934 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_ELSE) reduce 158) ((T_ELSEIF) reduce 158) ((T_EMPTY) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_VOID_CAST) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("}") reduce 158) (("~") reduce 158)) 27936 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_ELSE) reduce 156) ((T_ELSEIF) reduce 156) ((T_EMPTY) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_VOID_CAST) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("}") reduce 156) (("~") reduce 156)) 27938 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_ELSE) reduce 168) ((T_ELSEIF) reduce 168) ((T_EMPTY) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_VOID_CAST) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("}") reduce 168) (("~") reduce 168)) 27940 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_ELSE) reduce 157) ((T_ELSEIF) reduce 157) ((T_EMPTY) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_VOID_CAST) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("}") reduce 157) (("~") reduce 157)) 27950 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_ELSE) reduce 164) ((T_ELSEIF) reduce 164) ((T_EMPTY) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_VOID_CAST) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("}") reduce 164) (("~") reduce 164)) 27951 ((("!") reduce 220) (("\"") reduce 220) (("$") reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_ELSE) reduce 220) ((T_ELSEIF) reduce 220) ((T_EMPTY) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_VARIABLE) reduce 220) ((T_VOID_CAST) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("}") reduce 220) (("~") reduce 220)) 27954 ((("!") reduce 221) (("\"") reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_ELSE) reduce 221) ((T_ELSEIF) reduce 221) ((T_EMPTY) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_VARIABLE) reduce 221) ((T_VOID_CAST) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("}") reduce 221) (("~") reduce 221)) 27955 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_ELSE) reduce 163) ((T_ELSEIF) reduce 163) ((T_EMPTY) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_VOID_CAST) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("}") reduce 163) (("~") reduce 163)) 27964 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_ELSE) reduce 152) ((T_ELSEIF) reduce 152) ((T_EMPTY) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_VOID_CAST) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("}") reduce 152) (("~") reduce 152)) 27965 ((("!") reduce 218) (("\"") reduce 218) (("$") reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_ELSE) reduce 218) ((T_ELSEIF) reduce 218) ((T_EMPTY) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_VARIABLE) reduce 218) ((T_VOID_CAST) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("}") reduce 218) (("~") reduce 218)) 27968 ((("!") reduce 219) (("\"") reduce 219) (("$") reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_ELSE) reduce 219) ((T_ELSEIF) reduce 219) ((T_EMPTY) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_VARIABLE) reduce 219) ((T_VOID_CAST) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("}") reduce 219) (("~") reduce 219)) 27970 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_ELSE) reduce 159) ((T_ELSEIF) reduce 159) ((T_EMPTY) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_VOID_CAST) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("}") reduce 159) (("~") reduce 159)) 27976 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_ELSE) reduce 151) ((T_ELSEIF) reduce 151) ((T_EMPTY) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_VOID_CAST) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("}") reduce 151) (("~") reduce 151)) 27981 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_ELSE) reduce 165) ((T_ELSEIF) reduce 165) ((T_EMPTY) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_VOID_CAST) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("}") reduce 165) (("~") reduce 165)) 27982 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_ELSE) reduce 222) ((T_ELSEIF) reduce 222) ((T_EMPTY) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_VARIABLE) reduce 222) ((T_VOID_CAST) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("}") reduce 222) (("~") reduce 222)) 27985 ((("!") reduce 223) (("\"") reduce 223) (("$") reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_ELSE) reduce 223) ((T_ELSEIF) reduce 223) ((T_EMPTY) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_VARIABLE) reduce 223) ((T_VOID_CAST) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("}") reduce 223) (("~") reduce 223)) 27987 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_ELSE) reduce 155) ((T_ELSEIF) reduce 155) ((T_EMPTY) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_VOID_CAST) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("}") reduce 155) (("~") reduce 155)) 27989 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_ELSE) reduce 154) ((T_ELSEIF) reduce 154) ((T_EMPTY) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_VOID_CAST) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("}") reduce 154) (("~") reduce 154)) 28673 ((("!") reduce 247) (("\"") reduce 247) (("$") reduce 247) (("(") reduce 247) (("+") reduce 247) (("-") reduce 247) ((";") reduce 247) (("@") reduce 247) ((T_ABSTRACT) reduce 247) ((T_ARRAY) reduce 247) ((T_ARRAY_CAST) reduce 247) ((T_ATTRIBUTE) reduce 247) ((T_BOOL_CAST) reduce 247) ((T_BREAK) reduce 247) ((T_CLASS) reduce 247) ((T_CLASS_C) reduce 247) ((T_CLONE) reduce 247) ((T_CONSTANT_ENCAPSED_STRING) reduce 247) ((T_CONTINUE) reduce 247) ((T_DEC) reduce 247) ((T_DECLARE) reduce 247) ((T_DIR) reduce 247) ((T_DNUMBER) reduce 247) ((T_DO) reduce 247) ((T_DOUBLE_CAST) reduce 247) ((T_ECHO) reduce 247) ((T_EMPTY) reduce 247) ((T_ENUM) reduce 247) ((T_EVAL) reduce 247) ((T_EXIT) reduce 247) ((T_FILE) reduce 247) ((T_FINAL) reduce 247) ((T_FN) reduce 247) ((T_FOR) reduce 247) ((T_FOREACH) reduce 247) ((T_FUNCTION) reduce 247) ((T_FUNC_C) reduce 247) ((T_GLOBAL) reduce 247) ((T_GOTO) reduce 247) ((T_HALT_COMPILER) reduce 247) ((T_IF) reduce 247) ((T_INC) reduce 247) ((T_INCLUDE) reduce 247) ((T_INCLUDE_ONCE) reduce 247) ((T_INLINE_HTML) reduce 247) ((T_INTERFACE) reduce 247) ((T_INT_CAST) reduce 247) ((T_ISSET) reduce 247) ((T_LINE) reduce 247) ((T_LIST) reduce 247) ((T_LNUMBER) reduce 247) ((T_MATCH) reduce 247) ((T_METHOD_C) reduce 247) ((T_NAME_FULLY_QUALIFIED) reduce 247) ((T_NAME_QUALIFIED) reduce 247) ((T_NAME_RELATIVE) reduce 247) ((T_NEW) reduce 247) ((T_NS_C) reduce 247) ((T_OBJECT_CAST) reduce 247) ((T_PRINT) reduce 247) ((T_PROPERTY_C) reduce 247) ((T_READONLY) reduce 247) ((T_REQUIRE) reduce 247) ((T_REQUIRE_ONCE) reduce 247) ((T_RETURN) reduce 247) ((T_START_HEREDOC) reduce 247) ((T_STATIC) reduce 247) ((T_STRING) reduce 247) ((T_STRING_CAST) reduce 247) ((T_SWITCH) reduce 247) ((T_THROW) reduce 247) ((T_TRAIT) reduce 247) ((T_TRAIT_C) reduce 247) ((T_TRY) reduce 247) ((T_UNSET) reduce 247) ((T_UNSET_CAST) reduce 247) ((T_VARIABLE) reduce 247) ((T_VOID_CAST) reduce 247) ((T_WHILE) reduce 247) ((T_YIELD) reduce 247) ((T_YIELD_FROM) reduce 247) (("[") reduce 247) (("`") reduce 247) (("{") reduce 247) (("}") reduce 247) (("~") reduce 247)) 28684 ((("!") reduce 494) (("\"") reduce 494) (("$") reduce 494) (("(") reduce 494) (("+") reduce 494) (("-") reduce 494) ((";") reduce 494) (("@") reduce 494) ((T_ABSTRACT) reduce 494) ((T_ARRAY) reduce 494) ((T_ARRAY_CAST) reduce 494) ((T_ATTRIBUTE) reduce 494) ((T_BOOL_CAST) reduce 494) ((T_BREAK) reduce 494) ((T_CLASS) reduce 494) ((T_CLASS_C) reduce 494) ((T_CLONE) reduce 494) ((T_CONSTANT_ENCAPSED_STRING) reduce 494) ((T_CONTINUE) reduce 494) ((T_DEC) reduce 494) ((T_DECLARE) reduce 494) ((T_DIR) reduce 494) ((T_DNUMBER) reduce 494) ((T_DO) reduce 494) ((T_DOUBLE_CAST) reduce 494) ((T_ECHO) reduce 494) ((T_EMPTY) reduce 494) ((T_ENUM) reduce 494) ((T_EVAL) reduce 494) ((T_EXIT) reduce 494) ((T_FILE) reduce 494) ((T_FINAL) reduce 494) ((T_FN) reduce 494) ((T_FOR) reduce 494) ((T_FOREACH) reduce 494) ((T_FUNCTION) reduce 494) ((T_FUNC_C) reduce 494) ((T_GLOBAL) reduce 494) ((T_GOTO) reduce 494) ((T_HALT_COMPILER) reduce 494) ((T_IF) reduce 494) ((T_INC) reduce 494) ((T_INCLUDE) reduce 494) ((T_INCLUDE_ONCE) reduce 494) ((T_INLINE_HTML) reduce 494) ((T_INTERFACE) reduce 494) ((T_INT_CAST) reduce 494) ((T_ISSET) reduce 494) ((T_LINE) reduce 494) ((T_LIST) reduce 494) ((T_LNUMBER) reduce 494) ((T_MATCH) reduce 494) ((T_METHOD_C) reduce 494) ((T_NAME_FULLY_QUALIFIED) reduce 494) ((T_NAME_QUALIFIED) reduce 494) ((T_NAME_RELATIVE) reduce 494) ((T_NEW) reduce 494) ((T_NS_C) reduce 494) ((T_OBJECT_CAST) reduce 494) ((T_PRINT) reduce 494) ((T_PROPERTY_C) reduce 494) ((T_READONLY) reduce 494) ((T_REQUIRE) reduce 494) ((T_REQUIRE_ONCE) reduce 494) ((T_RETURN) reduce 494) ((T_START_HEREDOC) reduce 494) ((T_STATIC) reduce 494) ((T_STRING) reduce 494) ((T_STRING_CAST) reduce 494) ((T_SWITCH) reduce 494) ((T_THROW) reduce 494) ((T_TRAIT) reduce 494) ((T_TRAIT_C) reduce 494) ((T_TRY) reduce 494) ((T_UNSET) reduce 494) ((T_UNSET_CAST) reduce 494) ((T_VARIABLE) reduce 494) ((T_VOID_CAST) reduce 494) ((T_WHILE) reduce 494) ((T_YIELD) reduce 494) ((T_YIELD_FROM) reduce 494) (("[") reduce 494) (("`") reduce 494) (("{") reduce 494) (("}") reduce 494) (("~") reduce 494)) 28685 ((("!") reduce 184) (("\"") reduce 184) (("$") reduce 184) (("(") reduce 184) (("+") reduce 184) (("-") reduce 184) ((";") reduce 184) (("@") reduce 184) ((T_ABSTRACT) reduce 184) ((T_ARRAY) reduce 184) ((T_ARRAY_CAST) reduce 184) ((T_ATTRIBUTE) reduce 184) ((T_BOOL_CAST) reduce 184) ((T_BREAK) reduce 184) ((T_CLASS) reduce 184) ((T_CLASS_C) reduce 184) ((T_CLONE) reduce 184) ((T_CONSTANT_ENCAPSED_STRING) reduce 184) ((T_CONTINUE) reduce 184) ((T_DEC) reduce 184) ((T_DECLARE) reduce 184) ((T_DIR) reduce 184) ((T_DNUMBER) reduce 184) ((T_DO) reduce 184) ((T_DOUBLE_CAST) reduce 184) ((T_ECHO) reduce 184) ((T_EMPTY) reduce 184) ((T_ENUM) reduce 184) ((T_EVAL) reduce 184) ((T_EXIT) reduce 184) ((T_FILE) reduce 184) ((T_FINAL) reduce 184) ((T_FN) reduce 184) ((T_FOR) reduce 184) ((T_FOREACH) reduce 184) ((T_FUNCTION) reduce 184) ((T_FUNC_C) reduce 184) ((T_GLOBAL) reduce 184) ((T_GOTO) reduce 184) ((T_HALT_COMPILER) reduce 184) ((T_IF) reduce 184) ((T_INC) reduce 184) ((T_INCLUDE) reduce 184) ((T_INCLUDE_ONCE) reduce 184) ((T_INLINE_HTML) reduce 184) ((T_INTERFACE) reduce 184) ((T_INT_CAST) reduce 184) ((T_ISSET) reduce 184) ((T_LINE) reduce 184) ((T_LIST) reduce 184) ((T_LNUMBER) reduce 184) ((T_MATCH) reduce 184) ((T_METHOD_C) reduce 184) ((T_NAME_FULLY_QUALIFIED) reduce 184) ((T_NAME_QUALIFIED) reduce 184) ((T_NAME_RELATIVE) reduce 184) ((T_NEW) reduce 184) ((T_NS_C) reduce 184) ((T_OBJECT_CAST) reduce 184) ((T_PRINT) reduce 184) ((T_PROPERTY_C) reduce 184) ((T_READONLY) reduce 184) ((T_REQUIRE) reduce 184) ((T_REQUIRE_ONCE) reduce 184) ((T_RETURN) reduce 184) ((T_START_HEREDOC) reduce 184) ((T_STATIC) reduce 184) ((T_STRING) reduce 184) ((T_STRING_CAST) reduce 184) ((T_SWITCH) reduce 184) ((T_THROW) reduce 184) ((T_TRAIT) reduce 184) ((T_TRAIT_C) reduce 184) ((T_TRY) reduce 184) ((T_UNSET) reduce 184) ((T_UNSET_CAST) reduce 184) ((T_VARIABLE) reduce 184) ((T_VOID_CAST) reduce 184) ((T_WHILE) reduce 184) ((T_YIELD) reduce 184) ((T_YIELD_FROM) reduce 184) (("[") reduce 184) (("`") reduce 184) (("{") reduce 184) (("}") reduce 184) (("~") reduce 184)) 28686 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_EMPTY) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_VARIABLE) reduce 161) ((T_VOID_CAST) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("}") reduce 161) (("~") reduce 161)) 28694 ((("!") reduce 189) (("\"") reduce 189) (("$") reduce 189) (("(") reduce 189) (("+") reduce 189) (("-") reduce 189) ((";") reduce 189) (("@") reduce 189) ((T_ABSTRACT) reduce 189) ((T_ARRAY) reduce 189) ((T_ARRAY_CAST) reduce 189) ((T_ATTRIBUTE) reduce 189) ((T_BOOL_CAST) reduce 189) ((T_BREAK) reduce 189) ((T_CLASS) reduce 189) ((T_CLASS_C) reduce 189) ((T_CLONE) reduce 189) ((T_CONSTANT_ENCAPSED_STRING) reduce 189) ((T_CONTINUE) reduce 189) ((T_DEC) reduce 189) ((T_DECLARE) reduce 189) ((T_DIR) reduce 189) ((T_DNUMBER) reduce 189) ((T_DO) reduce 189) ((T_DOUBLE_CAST) reduce 189) ((T_ECHO) reduce 189) ((T_EMPTY) reduce 189) ((T_ENUM) reduce 189) ((T_EVAL) reduce 189) ((T_EXIT) reduce 189) ((T_FILE) reduce 189) ((T_FINAL) reduce 189) ((T_FN) reduce 189) ((T_FOR) reduce 189) ((T_FOREACH) reduce 189) ((T_FUNCTION) reduce 189) ((T_FUNC_C) reduce 189) ((T_GLOBAL) reduce 189) ((T_GOTO) reduce 189) ((T_HALT_COMPILER) reduce 189) ((T_IF) reduce 189) ((T_INC) reduce 189) ((T_INCLUDE) reduce 189) ((T_INCLUDE_ONCE) reduce 189) ((T_INLINE_HTML) reduce 189) ((T_INTERFACE) reduce 189) ((T_INT_CAST) reduce 189) ((T_ISSET) reduce 189) ((T_LINE) reduce 189) ((T_LIST) reduce 189) ((T_LNUMBER) reduce 189) ((T_MATCH) reduce 189) ((T_METHOD_C) reduce 189) ((T_NAME_FULLY_QUALIFIED) reduce 189) ((T_NAME_QUALIFIED) reduce 189) ((T_NAME_RELATIVE) reduce 189) ((T_NEW) reduce 189) ((T_NS_C) reduce 189) ((T_OBJECT_CAST) reduce 189) ((T_PRINT) reduce 189) ((T_PROPERTY_C) reduce 189) ((T_READONLY) reduce 189) ((T_REQUIRE) reduce 189) ((T_REQUIRE_ONCE) reduce 189) ((T_RETURN) reduce 189) ((T_START_HEREDOC) reduce 189) ((T_STATIC) reduce 189) ((T_STRING) reduce 189) ((T_STRING_CAST) reduce 189) ((T_SWITCH) reduce 189) ((T_THROW) reduce 189) ((T_TRAIT) reduce 189) ((T_TRAIT_C) reduce 189) ((T_TRY) reduce 189) ((T_UNSET) reduce 189) ((T_UNSET_CAST) reduce 189) ((T_VARIABLE) reduce 189) ((T_VOID_CAST) reduce 189) ((T_WHILE) reduce 189) ((T_YIELD) reduce 189) ((T_YIELD_FROM) reduce 189) (("[") reduce 189) (("`") reduce 189) (("{") reduce 189) (("}") reduce 189) (("~") reduce 189)) 28695 ((("!") reduce 145) (("\"") reduce 145) (("$") reduce 145) (("(") reduce 145) (("+") reduce 145) (("-") reduce 145) ((";") reduce 145) (("@") reduce 145) ((T_ABSTRACT) reduce 145) ((T_ARRAY) reduce 145) ((T_ARRAY_CAST) reduce 145) ((T_ATTRIBUTE) reduce 145) ((T_BOOL_CAST) reduce 145) ((T_BREAK) reduce 145) ((T_CLASS) reduce 145) ((T_CLASS_C) reduce 145) ((T_CLONE) reduce 145) ((T_CONSTANT_ENCAPSED_STRING) reduce 145) ((T_CONTINUE) reduce 145) ((T_DEC) reduce 145) ((T_DECLARE) reduce 145) ((T_DIR) reduce 145) ((T_DNUMBER) reduce 145) ((T_DO) reduce 145) ((T_DOUBLE_CAST) reduce 145) ((T_ECHO) reduce 145) ((T_EMPTY) reduce 145) ((T_ENUM) reduce 145) ((T_EVAL) reduce 145) ((T_EXIT) reduce 145) ((T_FILE) reduce 145) ((T_FINAL) reduce 145) ((T_FN) reduce 145) ((T_FOR) reduce 145) ((T_FOREACH) reduce 145) ((T_FUNCTION) reduce 145) ((T_FUNC_C) reduce 145) ((T_GLOBAL) reduce 145) ((T_GOTO) reduce 145) ((T_HALT_COMPILER) reduce 145) ((T_IF) reduce 145) ((T_INC) reduce 145) ((T_INCLUDE) reduce 145) ((T_INCLUDE_ONCE) reduce 145) ((T_INLINE_HTML) reduce 145) ((T_INTERFACE) reduce 145) ((T_INT_CAST) reduce 145) ((T_ISSET) reduce 145) ((T_LINE) reduce 145) ((T_LIST) reduce 145) ((T_LNUMBER) reduce 145) ((T_MATCH) reduce 145) ((T_METHOD_C) reduce 145) ((T_NAME_FULLY_QUALIFIED) reduce 145) ((T_NAME_QUALIFIED) reduce 145) ((T_NAME_RELATIVE) reduce 145) ((T_NEW) reduce 145) ((T_NS_C) reduce 145) ((T_OBJECT_CAST) reduce 145) ((T_PRINT) reduce 145) ((T_PROPERTY_C) reduce 145) ((T_READONLY) reduce 145) ((T_REQUIRE) reduce 145) ((T_REQUIRE_ONCE) reduce 145) ((T_RETURN) reduce 145) ((T_START_HEREDOC) reduce 145) ((T_STATIC) reduce 145) ((T_STRING) reduce 145) ((T_STRING_CAST) reduce 145) ((T_SWITCH) reduce 145) ((T_THROW) reduce 145) ((T_TRAIT) reduce 145) ((T_TRAIT_C) reduce 145) ((T_TRY) reduce 145) ((T_UNSET) reduce 145) ((T_UNSET_CAST) reduce 145) ((T_VARIABLE) reduce 145) ((T_VOID_CAST) reduce 145) ((T_WHILE) reduce 145) ((T_YIELD) reduce 145) ((T_YIELD_FROM) reduce 145) (("[") reduce 145) (("`") reduce 145) (("{") reduce 145) (("}") reduce 145) (("~") reduce 145)) 28698 ((("!") reduce 250) (("\"") reduce 250) (("$") reduce 250) (("(") reduce 250) (("+") reduce 250) (("-") reduce 250) ((";") reduce 250) (("@") reduce 250) ((T_ABSTRACT) reduce 250) ((T_ARRAY) reduce 250) ((T_ARRAY_CAST) reduce 250) ((T_ATTRIBUTE) reduce 250) ((T_BOOL_CAST) reduce 250) ((T_BREAK) reduce 250) ((T_CLASS) reduce 250) ((T_CLASS_C) reduce 250) ((T_CLONE) reduce 250) ((T_CONSTANT_ENCAPSED_STRING) reduce 250) ((T_CONTINUE) reduce 250) ((T_DEC) reduce 250) ((T_DECLARE) reduce 250) ((T_DIR) reduce 250) ((T_DNUMBER) reduce 250) ((T_DO) reduce 250) ((T_DOUBLE_CAST) reduce 250) ((T_ECHO) reduce 250) ((T_EMPTY) reduce 250) ((T_ENUM) reduce 250) ((T_EVAL) reduce 250) ((T_EXIT) reduce 250) ((T_FILE) reduce 250) ((T_FINAL) reduce 250) ((T_FN) reduce 250) ((T_FOR) reduce 250) ((T_FOREACH) reduce 250) ((T_FUNCTION) reduce 250) ((T_FUNC_C) reduce 250) ((T_GLOBAL) reduce 250) ((T_GOTO) reduce 250) ((T_HALT_COMPILER) reduce 250) ((T_IF) reduce 250) ((T_INC) reduce 250) ((T_INCLUDE) reduce 250) ((T_INCLUDE_ONCE) reduce 250) ((T_INLINE_HTML) reduce 250) ((T_INTERFACE) reduce 250) ((T_INT_CAST) reduce 250) ((T_ISSET) reduce 250) ((T_LINE) reduce 250) ((T_LIST) reduce 250) ((T_LNUMBER) reduce 250) ((T_MATCH) reduce 250) ((T_METHOD_C) reduce 250) ((T_NAME_FULLY_QUALIFIED) reduce 250) ((T_NAME_QUALIFIED) reduce 250) ((T_NAME_RELATIVE) reduce 250) ((T_NEW) reduce 250) ((T_NS_C) reduce 250) ((T_OBJECT_CAST) reduce 250) ((T_PRINT) reduce 250) ((T_PROPERTY_C) reduce 250) ((T_READONLY) reduce 250) ((T_REQUIRE) reduce 250) ((T_REQUIRE_ONCE) reduce 250) ((T_RETURN) reduce 250) ((T_START_HEREDOC) reduce 250) ((T_STATIC) reduce 250) ((T_STRING) reduce 250) ((T_STRING_CAST) reduce 250) ((T_SWITCH) reduce 250) ((T_THROW) reduce 250) ((T_TRAIT) reduce 250) ((T_TRAIT_C) reduce 250) ((T_TRY) reduce 250) ((T_UNSET) reduce 250) ((T_UNSET_CAST) reduce 250) ((T_VARIABLE) reduce 250) ((T_VOID_CAST) reduce 250) ((T_WHILE) reduce 250) ((T_YIELD) reduce 250) ((T_YIELD_FROM) reduce 250) (("[") reduce 250) (("`") reduce 250) (("{") reduce 250) (("}") reduce 250) (("~") reduce 250)) 28702 ((("!") reduce 251) (("\"") reduce 251) (("$") reduce 251) (("(") reduce 251) (("+") reduce 251) (("-") reduce 251) ((";") reduce 251) (("@") reduce 251) ((T_ABSTRACT) reduce 251) ((T_ARRAY) reduce 251) ((T_ARRAY_CAST) reduce 251) ((T_ATTRIBUTE) reduce 251) ((T_BOOL_CAST) reduce 251) ((T_BREAK) reduce 251) ((T_CLASS) reduce 251) ((T_CLASS_C) reduce 251) ((T_CLONE) reduce 251) ((T_CONSTANT_ENCAPSED_STRING) reduce 251) ((T_CONTINUE) reduce 251) ((T_DEC) reduce 251) ((T_DECLARE) reduce 251) ((T_DIR) reduce 251) ((T_DNUMBER) reduce 251) ((T_DO) reduce 251) ((T_DOUBLE_CAST) reduce 251) ((T_ECHO) reduce 251) ((T_EMPTY) reduce 251) ((T_ENUM) reduce 251) ((T_EVAL) reduce 251) ((T_EXIT) reduce 251) ((T_FILE) reduce 251) ((T_FINAL) reduce 251) ((T_FN) reduce 251) ((T_FOR) reduce 251) ((T_FOREACH) reduce 251) ((T_FUNCTION) reduce 251) ((T_FUNC_C) reduce 251) ((T_GLOBAL) reduce 251) ((T_GOTO) reduce 251) ((T_HALT_COMPILER) reduce 251) ((T_IF) reduce 251) ((T_INC) reduce 251) ((T_INCLUDE) reduce 251) ((T_INCLUDE_ONCE) reduce 251) ((T_INLINE_HTML) reduce 251) ((T_INTERFACE) reduce 251) ((T_INT_CAST) reduce 251) ((T_ISSET) reduce 251) ((T_LINE) reduce 251) ((T_LIST) reduce 251) ((T_LNUMBER) reduce 251) ((T_MATCH) reduce 251) ((T_METHOD_C) reduce 251) ((T_NAME_FULLY_QUALIFIED) reduce 251) ((T_NAME_QUALIFIED) reduce 251) ((T_NAME_RELATIVE) reduce 251) ((T_NEW) reduce 251) ((T_NS_C) reduce 251) ((T_OBJECT_CAST) reduce 251) ((T_PRINT) reduce 251) ((T_PROPERTY_C) reduce 251) ((T_READONLY) reduce 251) ((T_REQUIRE) reduce 251) ((T_REQUIRE_ONCE) reduce 251) ((T_RETURN) reduce 251) ((T_START_HEREDOC) reduce 251) ((T_STATIC) reduce 251) ((T_STRING) reduce 251) ((T_STRING_CAST) reduce 251) ((T_SWITCH) reduce 251) ((T_THROW) reduce 251) ((T_TRAIT) reduce 251) ((T_TRAIT_C) reduce 251) ((T_TRY) reduce 251) ((T_UNSET) reduce 251) ((T_UNSET_CAST) reduce 251) ((T_VARIABLE) reduce 251) ((T_VOID_CAST) reduce 251) ((T_WHILE) reduce 251) ((T_YIELD) reduce 251) ((T_YIELD_FROM) reduce 251) (("[") reduce 251) (("`") reduce 251) (("{") reduce 251) (("}") reduce 251) (("~") reduce 251)) 28707 ((("!") reduce 242) (("\"") reduce 242) (("$") reduce 242) (("(") reduce 242) (("+") reduce 242) (("-") reduce 242) ((";") reduce 242) (("@") reduce 242) ((T_ABSTRACT) reduce 242) ((T_ARRAY) reduce 242) ((T_ARRAY_CAST) reduce 242) ((T_ATTRIBUTE) reduce 242) ((T_BOOL_CAST) reduce 242) ((T_BREAK) reduce 242) ((T_CLASS) reduce 242) ((T_CLASS_C) reduce 242) ((T_CLONE) reduce 242) ((T_CONSTANT_ENCAPSED_STRING) reduce 242) ((T_CONTINUE) reduce 242) ((T_DEC) reduce 242) ((T_DECLARE) reduce 242) ((T_DIR) reduce 242) ((T_DNUMBER) reduce 242) ((T_DO) reduce 242) ((T_DOUBLE_CAST) reduce 242) ((T_ECHO) reduce 242) ((T_EMPTY) reduce 242) ((T_ENUM) reduce 242) ((T_EVAL) reduce 242) ((T_EXIT) reduce 242) ((T_FILE) reduce 242) ((T_FINAL) reduce 242) ((T_FN) reduce 242) ((T_FOR) reduce 242) ((T_FOREACH) reduce 242) ((T_FUNCTION) reduce 242) ((T_FUNC_C) reduce 242) ((T_GLOBAL) reduce 242) ((T_GOTO) reduce 242) ((T_HALT_COMPILER) reduce 242) ((T_IF) reduce 242) ((T_INC) reduce 242) ((T_INCLUDE) reduce 242) ((T_INCLUDE_ONCE) reduce 242) ((T_INLINE_HTML) reduce 242) ((T_INTERFACE) reduce 242) ((T_INT_CAST) reduce 242) ((T_ISSET) reduce 242) ((T_LINE) reduce 242) ((T_LIST) reduce 242) ((T_LNUMBER) reduce 242) ((T_MATCH) reduce 242) ((T_METHOD_C) reduce 242) ((T_NAME_FULLY_QUALIFIED) reduce 242) ((T_NAME_QUALIFIED) reduce 242) ((T_NAME_RELATIVE) reduce 242) ((T_NEW) reduce 242) ((T_NS_C) reduce 242) ((T_OBJECT_CAST) reduce 242) ((T_PRINT) reduce 242) ((T_PROPERTY_C) reduce 242) ((T_READONLY) reduce 242) ((T_REQUIRE) reduce 242) ((T_REQUIRE_ONCE) reduce 242) ((T_RETURN) reduce 242) ((T_START_HEREDOC) reduce 242) ((T_STATIC) reduce 242) ((T_STRING) reduce 242) ((T_STRING_CAST) reduce 242) ((T_SWITCH) reduce 242) ((T_THROW) reduce 242) ((T_TRAIT) reduce 242) ((T_TRAIT_C) reduce 242) ((T_TRY) reduce 242) ((T_UNSET) reduce 242) ((T_UNSET_CAST) reduce 242) ((T_VARIABLE) reduce 242) ((T_VOID_CAST) reduce 242) ((T_WHILE) reduce 242) ((T_YIELD) reduce 242) ((T_YIELD_FROM) reduce 242) (("[") reduce 242) (("`") reduce 242) (("{") reduce 242) (("}") reduce 242) (("~") reduce 242)) 28708 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_EMPTY) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_VARIABLE) reduce 150) ((T_VOID_CAST) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("}") reduce 150) (("~") reduce 150)) 28711 ((("!") reduce 243) (("\"") reduce 243) (("$") reduce 243) (("(") reduce 243) (("+") reduce 243) (("-") reduce 243) ((";") reduce 243) (("@") reduce 243) ((T_ABSTRACT) reduce 243) ((T_ARRAY) reduce 243) ((T_ARRAY_CAST) reduce 243) ((T_ATTRIBUTE) reduce 243) ((T_BOOL_CAST) reduce 243) ((T_BREAK) reduce 243) ((T_CLASS) reduce 243) ((T_CLASS_C) reduce 243) ((T_CLONE) reduce 243) ((T_CONSTANT_ENCAPSED_STRING) reduce 243) ((T_CONTINUE) reduce 243) ((T_DEC) reduce 243) ((T_DECLARE) reduce 243) ((T_DIR) reduce 243) ((T_DNUMBER) reduce 243) ((T_DO) reduce 243) ((T_DOUBLE_CAST) reduce 243) ((T_ECHO) reduce 243) ((T_EMPTY) reduce 243) ((T_ENUM) reduce 243) ((T_EVAL) reduce 243) ((T_EXIT) reduce 243) ((T_FILE) reduce 243) ((T_FINAL) reduce 243) ((T_FN) reduce 243) ((T_FOR) reduce 243) ((T_FOREACH) reduce 243) ((T_FUNCTION) reduce 243) ((T_FUNC_C) reduce 243) ((T_GLOBAL) reduce 243) ((T_GOTO) reduce 243) ((T_HALT_COMPILER) reduce 243) ((T_IF) reduce 243) ((T_INC) reduce 243) ((T_INCLUDE) reduce 243) ((T_INCLUDE_ONCE) reduce 243) ((T_INLINE_HTML) reduce 243) ((T_INTERFACE) reduce 243) ((T_INT_CAST) reduce 243) ((T_ISSET) reduce 243) ((T_LINE) reduce 243) ((T_LIST) reduce 243) ((T_LNUMBER) reduce 243) ((T_MATCH) reduce 243) ((T_METHOD_C) reduce 243) ((T_NAME_FULLY_QUALIFIED) reduce 243) ((T_NAME_QUALIFIED) reduce 243) ((T_NAME_RELATIVE) reduce 243) ((T_NEW) reduce 243) ((T_NS_C) reduce 243) ((T_OBJECT_CAST) reduce 243) ((T_PRINT) reduce 243) ((T_PROPERTY_C) reduce 243) ((T_READONLY) reduce 243) ((T_REQUIRE) reduce 243) ((T_REQUIRE_ONCE) reduce 243) ((T_RETURN) reduce 243) ((T_START_HEREDOC) reduce 243) ((T_STATIC) reduce 243) ((T_STRING) reduce 243) ((T_STRING_CAST) reduce 243) ((T_SWITCH) reduce 243) ((T_THROW) reduce 243) ((T_TRAIT) reduce 243) ((T_TRAIT_C) reduce 243) ((T_TRY) reduce 243) ((T_UNSET) reduce 243) ((T_UNSET_CAST) reduce 243) ((T_VARIABLE) reduce 243) ((T_VOID_CAST) reduce 243) ((T_WHILE) reduce 243) ((T_YIELD) reduce 243) ((T_YIELD_FROM) reduce 243) (("[") reduce 243) (("`") reduce 243) (("{") reduce 243) (("}") reduce 243) (("~") reduce 243)) 28713 ((("!") reduce 170) (("\"") reduce 170) (("$") reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_EMPTY) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_VARIABLE) reduce 170) ((T_VOID_CAST) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("}") reduce 170) (("~") reduce 170)) 28718 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_EMPTY) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_VARIABLE) reduce 162) ((T_VOID_CAST) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("}") reduce 162) (("~") reduce 162)) 28721 ((("!") reduce 171) (("\"") reduce 171) (("$") reduce 171) (("(") reduce 171) (("+") reduce 171) (("-") reduce 171) ((";") reduce 171) (("@") reduce 171) ((T_ABSTRACT) reduce 171) ((T_ARRAY) reduce 171) ((T_ARRAY_CAST) reduce 171) ((T_ATTRIBUTE) reduce 171) ((T_BOOL_CAST) reduce 171) ((T_BREAK) reduce 171) ((T_CATCH) reduce 171) ((T_CLASS) reduce 171) ((T_CLASS_C) reduce 171) ((T_CLONE) reduce 171) ((T_CONSTANT_ENCAPSED_STRING) reduce 171) ((T_CONTINUE) reduce 171) ((T_DEC) reduce 171) ((T_DECLARE) reduce 171) ((T_DIR) reduce 171) ((T_DNUMBER) reduce 171) ((T_DO) reduce 171) ((T_DOUBLE_CAST) reduce 171) ((T_ECHO) reduce 171) ((T_EMPTY) reduce 171) ((T_ENUM) reduce 171) ((T_EVAL) reduce 171) ((T_EXIT) reduce 171) ((T_FILE) reduce 171) ((T_FINAL) reduce 171) ((T_FINALLY) reduce 171) ((T_FN) reduce 171) ((T_FOR) reduce 171) ((T_FOREACH) reduce 171) ((T_FUNCTION) reduce 171) ((T_FUNC_C) reduce 171) ((T_GLOBAL) reduce 171) ((T_GOTO) reduce 171) ((T_HALT_COMPILER) reduce 171) ((T_IF) reduce 171) ((T_INC) reduce 171) ((T_INCLUDE) reduce 171) ((T_INCLUDE_ONCE) reduce 171) ((T_INLINE_HTML) reduce 171) ((T_INTERFACE) reduce 171) ((T_INT_CAST) reduce 171) ((T_ISSET) reduce 171) ((T_LINE) reduce 171) ((T_LIST) reduce 171) ((T_LNUMBER) reduce 171) ((T_MATCH) reduce 171) ((T_METHOD_C) reduce 171) ((T_NAME_FULLY_QUALIFIED) reduce 171) ((T_NAME_QUALIFIED) reduce 171) ((T_NAME_RELATIVE) reduce 171) ((T_NEW) reduce 171) ((T_NS_C) reduce 171) ((T_OBJECT_CAST) reduce 171) ((T_PRINT) reduce 171) ((T_PROPERTY_C) reduce 171) ((T_READONLY) reduce 171) ((T_REQUIRE) reduce 171) ((T_REQUIRE_ONCE) reduce 171) ((T_RETURN) reduce 171) ((T_START_HEREDOC) reduce 171) ((T_STATIC) reduce 171) ((T_STRING) reduce 171) ((T_STRING_CAST) reduce 171) ((T_SWITCH) reduce 171) ((T_THROW) reduce 171) ((T_TRAIT) reduce 171) ((T_TRAIT_C) reduce 171) ((T_TRY) reduce 171) ((T_UNSET) reduce 171) ((T_UNSET_CAST) reduce 171) ((T_VARIABLE) reduce 171) ((T_VOID_CAST) reduce 171) ((T_WHILE) reduce 171) ((T_YIELD) reduce 171) ((T_YIELD_FROM) reduce 171) (("[") reduce 171) (("`") reduce 171) (("{") reduce 171) (("}") reduce 171) (("~") reduce 171)) 28722 ((("!") reduce 177) (("\"") reduce 177) (("$") reduce 177) (("(") reduce 177) (("+") reduce 177) (("-") reduce 177) ((";") reduce 177) (("@") reduce 177) ((T_ABSTRACT) reduce 177) ((T_ARRAY) reduce 177) ((T_ARRAY_CAST) reduce 177) ((T_ATTRIBUTE) reduce 177) ((T_BOOL_CAST) reduce 177) ((T_BREAK) reduce 177) ((T_CATCH) shift) ((T_CLASS) reduce 177) ((T_CLASS_C) reduce 177) ((T_CLONE) reduce 177) ((T_CONSTANT_ENCAPSED_STRING) reduce 177) ((T_CONTINUE) reduce 177) ((T_DEC) reduce 177) ((T_DECLARE) reduce 177) ((T_DIR) reduce 177) ((T_DNUMBER) reduce 177) ((T_DO) reduce 177) ((T_DOUBLE_CAST) reduce 177) ((T_ECHO) reduce 177) ((T_EMPTY) reduce 177) ((T_ENUM) reduce 177) ((T_EVAL) reduce 177) ((T_EXIT) reduce 177) ((T_FILE) reduce 177) ((T_FINAL) reduce 177) ((T_FINALLY) shift) ((T_FN) reduce 177) ((T_FOR) reduce 177) ((T_FOREACH) reduce 177) ((T_FUNCTION) reduce 177) ((T_FUNC_C) reduce 177) ((T_GLOBAL) reduce 177) ((T_GOTO) reduce 177) ((T_HALT_COMPILER) reduce 177) ((T_IF) reduce 177) ((T_INC) reduce 177) ((T_INCLUDE) reduce 177) ((T_INCLUDE_ONCE) reduce 177) ((T_INLINE_HTML) reduce 177) ((T_INTERFACE) reduce 177) ((T_INT_CAST) reduce 177) ((T_ISSET) reduce 177) ((T_LINE) reduce 177) ((T_LIST) reduce 177) ((T_LNUMBER) reduce 177) ((T_MATCH) reduce 177) ((T_METHOD_C) reduce 177) ((T_NAME_FULLY_QUALIFIED) reduce 177) ((T_NAME_QUALIFIED) reduce 177) ((T_NAME_RELATIVE) reduce 177) ((T_NEW) reduce 177) ((T_NS_C) reduce 177) ((T_OBJECT_CAST) reduce 177) ((T_PRINT) reduce 177) ((T_PROPERTY_C) reduce 177) ((T_READONLY) reduce 177) ((T_REQUIRE) reduce 177) ((T_REQUIRE_ONCE) reduce 177) ((T_RETURN) reduce 177) ((T_START_HEREDOC) reduce 177) ((T_STATIC) reduce 177) ((T_STRING) reduce 177) ((T_STRING_CAST) reduce 177) ((T_SWITCH) reduce 177) ((T_THROW) reduce 177) ((T_TRAIT) reduce 177) ((T_TRAIT_C) reduce 177) ((T_TRY) reduce 177) ((T_UNSET) reduce 177) ((T_UNSET_CAST) reduce 177) ((T_VARIABLE) reduce 177) ((T_VOID_CAST) reduce 177) ((T_WHILE) reduce 177) ((T_YIELD) reduce 177) ((T_YIELD_FROM) reduce 177) (("[") reduce 177) (("`") reduce 177) (("{") reduce 177) (("}") reduce 177) (("~") reduce 177)) 28725 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_EMPTY) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_VARIABLE) reduce 167) ((T_VOID_CAST) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("}") reduce 167) (("~") reduce 167)) 28728 ((("!") reduce 178) (("\"") reduce 178) (("$") reduce 178) (("(") reduce 178) (("+") reduce 178) (("-") reduce 178) ((";") reduce 178) (("@") reduce 178) ((T_ABSTRACT) reduce 178) ((T_ARRAY) reduce 178) ((T_ARRAY_CAST) reduce 178) ((T_ATTRIBUTE) reduce 178) ((T_BOOL_CAST) reduce 178) ((T_BREAK) reduce 178) ((T_CLASS) reduce 178) ((T_CLASS_C) reduce 178) ((T_CLONE) reduce 178) ((T_CONSTANT_ENCAPSED_STRING) reduce 178) ((T_CONTINUE) reduce 178) ((T_DEC) reduce 178) ((T_DECLARE) reduce 178) ((T_DIR) reduce 178) ((T_DNUMBER) reduce 178) ((T_DO) reduce 178) ((T_DOUBLE_CAST) reduce 178) ((T_ECHO) reduce 178) ((T_EMPTY) reduce 178) ((T_ENUM) reduce 178) ((T_EVAL) reduce 178) ((T_EXIT) reduce 178) ((T_FILE) reduce 178) ((T_FINAL) reduce 178) ((T_FN) reduce 178) ((T_FOR) reduce 178) ((T_FOREACH) reduce 178) ((T_FUNCTION) reduce 178) ((T_FUNC_C) reduce 178) ((T_GLOBAL) reduce 178) ((T_GOTO) reduce 178) ((T_HALT_COMPILER) reduce 178) ((T_IF) reduce 178) ((T_INC) reduce 178) ((T_INCLUDE) reduce 178) ((T_INCLUDE_ONCE) reduce 178) ((T_INLINE_HTML) reduce 178) ((T_INTERFACE) reduce 178) ((T_INT_CAST) reduce 178) ((T_ISSET) reduce 178) ((T_LINE) reduce 178) ((T_LIST) reduce 178) ((T_LNUMBER) reduce 178) ((T_MATCH) reduce 178) ((T_METHOD_C) reduce 178) ((T_NAME_FULLY_QUALIFIED) reduce 178) ((T_NAME_QUALIFIED) reduce 178) ((T_NAME_RELATIVE) reduce 178) ((T_NEW) reduce 178) ((T_NS_C) reduce 178) ((T_OBJECT_CAST) reduce 178) ((T_PRINT) reduce 178) ((T_PROPERTY_C) reduce 178) ((T_READONLY) reduce 178) ((T_REQUIRE) reduce 178) ((T_REQUIRE_ONCE) reduce 178) ((T_RETURN) reduce 178) ((T_START_HEREDOC) reduce 178) ((T_STATIC) reduce 178) ((T_STRING) reduce 178) ((T_STRING_CAST) reduce 178) ((T_SWITCH) reduce 178) ((T_THROW) reduce 178) ((T_TRAIT) reduce 178) ((T_TRAIT_C) reduce 178) ((T_TRY) reduce 178) ((T_UNSET) reduce 178) ((T_UNSET_CAST) reduce 178) ((T_VARIABLE) reduce 178) ((T_VOID_CAST) reduce 178) ((T_WHILE) reduce 178) ((T_YIELD) reduce 178) ((T_YIELD_FROM) reduce 178) (("[") reduce 178) (("`") reduce 178) (("{") reduce 178) (("}") reduce 178) (("~") reduce 178)) 28735 ((("!") reduce 172) (("\"") reduce 172) (("$") reduce 172) (("(") reduce 172) (("+") reduce 172) (("-") reduce 172) ((";") reduce 172) (("@") reduce 172) ((T_ABSTRACT) reduce 172) ((T_ARRAY) reduce 172) ((T_ARRAY_CAST) reduce 172) ((T_ATTRIBUTE) reduce 172) ((T_BOOL_CAST) reduce 172) ((T_BREAK) reduce 172) ((T_CATCH) reduce 172) ((T_CLASS) reduce 172) ((T_CLASS_C) reduce 172) ((T_CLONE) reduce 172) ((T_CONSTANT_ENCAPSED_STRING) reduce 172) ((T_CONTINUE) reduce 172) ((T_DEC) reduce 172) ((T_DECLARE) reduce 172) ((T_DIR) reduce 172) ((T_DNUMBER) reduce 172) ((T_DO) reduce 172) ((T_DOUBLE_CAST) reduce 172) ((T_ECHO) reduce 172) ((T_EMPTY) reduce 172) ((T_ENUM) reduce 172) ((T_EVAL) reduce 172) ((T_EXIT) reduce 172) ((T_FILE) reduce 172) ((T_FINAL) reduce 172) ((T_FINALLY) reduce 172) ((T_FN) reduce 172) ((T_FOR) reduce 172) ((T_FOREACH) reduce 172) ((T_FUNCTION) reduce 172) ((T_FUNC_C) reduce 172) ((T_GLOBAL) reduce 172) ((T_GOTO) reduce 172) ((T_HALT_COMPILER) reduce 172) ((T_IF) reduce 172) ((T_INC) reduce 172) ((T_INCLUDE) reduce 172) ((T_INCLUDE_ONCE) reduce 172) ((T_INLINE_HTML) reduce 172) ((T_INTERFACE) reduce 172) ((T_INT_CAST) reduce 172) ((T_ISSET) reduce 172) ((T_LINE) reduce 172) ((T_LIST) reduce 172) ((T_LNUMBER) reduce 172) ((T_MATCH) reduce 172) ((T_METHOD_C) reduce 172) ((T_NAME_FULLY_QUALIFIED) reduce 172) ((T_NAME_QUALIFIED) reduce 172) ((T_NAME_RELATIVE) reduce 172) ((T_NEW) reduce 172) ((T_NS_C) reduce 172) ((T_OBJECT_CAST) reduce 172) ((T_PRINT) reduce 172) ((T_PROPERTY_C) reduce 172) ((T_READONLY) reduce 172) ((T_REQUIRE) reduce 172) ((T_REQUIRE_ONCE) reduce 172) ((T_RETURN) reduce 172) ((T_START_HEREDOC) reduce 172) ((T_STATIC) reduce 172) ((T_STRING) reduce 172) ((T_STRING_CAST) reduce 172) ((T_SWITCH) reduce 172) ((T_THROW) reduce 172) ((T_TRAIT) reduce 172) ((T_TRAIT_C) reduce 172) ((T_TRY) reduce 172) ((T_UNSET) reduce 172) ((T_UNSET_CAST) reduce 172) ((T_VARIABLE) reduce 172) ((T_VOID_CAST) reduce 172) ((T_WHILE) reduce 172) ((T_YIELD) reduce 172) ((T_YIELD_FROM) reduce 172) (("[") reduce 172) (("`") reduce 172) (("{") reduce 172) (("}") reduce 172) (("~") reduce 172)) 28740 ((("!") reduce 200) (("\"") reduce 200) (("$") reduce 200) (("(") reduce 200) (("+") reduce 200) (("-") reduce 200) ((";") reduce 200) (("@") reduce 200) ((T_ABSTRACT) reduce 200) ((T_ARRAY) reduce 200) ((T_ARRAY_CAST) reduce 200) ((T_ATTRIBUTE) reduce 200) ((T_BOOL_CAST) reduce 200) ((T_BREAK) reduce 200) ((T_CLASS) reduce 200) ((T_CLASS_C) reduce 200) ((T_CLONE) reduce 200) ((T_CONSTANT_ENCAPSED_STRING) reduce 200) ((T_CONTINUE) reduce 200) ((T_DEC) reduce 200) ((T_DECLARE) reduce 200) ((T_DIR) reduce 200) ((T_DNUMBER) reduce 200) ((T_DO) reduce 200) ((T_DOUBLE_CAST) reduce 200) ((T_ECHO) reduce 200) ((T_EMPTY) reduce 200) ((T_ENUM) reduce 200) ((T_EVAL) reduce 200) ((T_EXIT) reduce 200) ((T_FILE) reduce 200) ((T_FINAL) reduce 200) ((T_FN) reduce 200) ((T_FOR) reduce 200) ((T_FOREACH) reduce 200) ((T_FUNCTION) reduce 200) ((T_FUNC_C) reduce 200) ((T_GLOBAL) reduce 200) ((T_GOTO) reduce 200) ((T_HALT_COMPILER) reduce 200) ((T_IF) reduce 200) ((T_INC) reduce 200) ((T_INCLUDE) reduce 200) ((T_INCLUDE_ONCE) reduce 200) ((T_INLINE_HTML) reduce 200) ((T_INTERFACE) reduce 200) ((T_INT_CAST) reduce 200) ((T_ISSET) reduce 200) ((T_LINE) reduce 200) ((T_LIST) reduce 200) ((T_LNUMBER) reduce 200) ((T_MATCH) reduce 200) ((T_METHOD_C) reduce 200) ((T_NAME_FULLY_QUALIFIED) reduce 200) ((T_NAME_QUALIFIED) reduce 200) ((T_NAME_RELATIVE) reduce 200) ((T_NEW) reduce 200) ((T_NS_C) reduce 200) ((T_OBJECT_CAST) reduce 200) ((T_PRINT) reduce 200) ((T_PROPERTY_C) reduce 200) ((T_READONLY) reduce 200) ((T_REQUIRE) reduce 200) ((T_REQUIRE_ONCE) reduce 200) ((T_RETURN) reduce 200) ((T_START_HEREDOC) reduce 200) ((T_STATIC) reduce 200) ((T_STRING) reduce 200) ((T_STRING_CAST) reduce 200) ((T_SWITCH) reduce 200) ((T_THROW) reduce 200) ((T_TRAIT) reduce 200) ((T_TRAIT_C) reduce 200) ((T_TRY) reduce 200) ((T_UNSET) reduce 200) ((T_UNSET_CAST) reduce 200) ((T_VARIABLE) reduce 200) ((T_VOID_CAST) reduce 200) ((T_WHILE) reduce 200) ((T_YIELD) reduce 200) ((T_YIELD_FROM) reduce 200) (("[") reduce 200) (("`") reduce 200) (("{") reduce 200) (("}") reduce 200) (("~") reduce 200)) 28745 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_EMPTY) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_VARIABLE) reduce 153) ((T_VOID_CAST) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("}") reduce 153) (("~") reduce 153)) 28749 ((("!") reduce 224) (("\"") reduce 224) (("$") reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_EMPTY) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_VARIABLE) reduce 224) ((T_VOID_CAST) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("}") reduce 224) (("~") reduce 224)) 28751 ((("!") reduce 225) (("\"") reduce 225) (("$") reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_EMPTY) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_VARIABLE) reduce 225) ((T_VOID_CAST) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("}") reduce 225) (("~") reduce 225)) 28755 ((("!") reduce 226) (("\"") reduce 226) (("$") reduce 226) (("(") reduce 226) (("+") reduce 226) (("-") reduce 226) ((";") reduce 226) (("@") reduce 226) ((T_ABSTRACT) reduce 226) ((T_ARRAY) reduce 226) ((T_ARRAY_CAST) reduce 226) ((T_ATTRIBUTE) reduce 226) ((T_BOOL_CAST) reduce 226) ((T_BREAK) reduce 226) ((T_CLASS) reduce 226) ((T_CLASS_C) reduce 226) ((T_CLONE) reduce 226) ((T_CONSTANT_ENCAPSED_STRING) reduce 226) ((T_CONTINUE) reduce 226) ((T_DEC) reduce 226) ((T_DECLARE) reduce 226) ((T_DIR) reduce 226) ((T_DNUMBER) reduce 226) ((T_DO) reduce 226) ((T_DOUBLE_CAST) reduce 226) ((T_ECHO) reduce 226) ((T_EMPTY) reduce 226) ((T_ENUM) reduce 226) ((T_EVAL) reduce 226) ((T_EXIT) reduce 226) ((T_FILE) reduce 226) ((T_FINAL) reduce 226) ((T_FN) reduce 226) ((T_FOR) reduce 226) ((T_FOREACH) reduce 226) ((T_FUNCTION) reduce 226) ((T_FUNC_C) reduce 226) ((T_GLOBAL) reduce 226) ((T_GOTO) reduce 226) ((T_HALT_COMPILER) reduce 226) ((T_IF) reduce 226) ((T_INC) reduce 226) ((T_INCLUDE) reduce 226) ((T_INCLUDE_ONCE) reduce 226) ((T_INLINE_HTML) reduce 226) ((T_INTERFACE) reduce 226) ((T_INT_CAST) reduce 226) ((T_ISSET) reduce 226) ((T_LINE) reduce 226) ((T_LIST) reduce 226) ((T_LNUMBER) reduce 226) ((T_MATCH) reduce 226) ((T_METHOD_C) reduce 226) ((T_NAME_FULLY_QUALIFIED) reduce 226) ((T_NAME_QUALIFIED) reduce 226) ((T_NAME_RELATIVE) reduce 226) ((T_NEW) reduce 226) ((T_NS_C) reduce 226) ((T_OBJECT_CAST) reduce 226) ((T_PRINT) reduce 226) ((T_PROPERTY_C) reduce 226) ((T_READONLY) reduce 226) ((T_REQUIRE) reduce 226) ((T_REQUIRE_ONCE) reduce 226) ((T_RETURN) reduce 226) ((T_START_HEREDOC) reduce 226) ((T_STATIC) reduce 226) ((T_STRING) reduce 226) ((T_STRING_CAST) reduce 226) ((T_SWITCH) reduce 226) ((T_THROW) reduce 226) ((T_TRAIT) reduce 226) ((T_TRAIT_C) reduce 226) ((T_TRY) reduce 226) ((T_UNSET) reduce 226) ((T_UNSET_CAST) reduce 226) ((T_VARIABLE) reduce 226) ((T_VOID_CAST) reduce 226) ((T_WHILE) reduce 226) ((T_YIELD) reduce 226) ((T_YIELD_FROM) reduce 226) (("[") reduce 226) (("`") reduce 226) (("{") reduce 226) (("}") reduce 226) (("~") reduce 226)) 28758 ((("!") reduce 227) (("\"") reduce 227) (("$") reduce 227) (("(") reduce 227) (("+") reduce 227) (("-") reduce 227) ((";") reduce 227) (("@") reduce 227) ((T_ABSTRACT) reduce 227) ((T_ARRAY) reduce 227) ((T_ARRAY_CAST) reduce 227) ((T_ATTRIBUTE) reduce 227) ((T_BOOL_CAST) reduce 227) ((T_BREAK) reduce 227) ((T_CLASS) reduce 227) ((T_CLASS_C) reduce 227) ((T_CLONE) reduce 227) ((T_CONSTANT_ENCAPSED_STRING) reduce 227) ((T_CONTINUE) reduce 227) ((T_DEC) reduce 227) ((T_DECLARE) reduce 227) ((T_DIR) reduce 227) ((T_DNUMBER) reduce 227) ((T_DO) reduce 227) ((T_DOUBLE_CAST) reduce 227) ((T_ECHO) reduce 227) ((T_EMPTY) reduce 227) ((T_ENUM) reduce 227) ((T_EVAL) reduce 227) ((T_EXIT) reduce 227) ((T_FILE) reduce 227) ((T_FINAL) reduce 227) ((T_FN) reduce 227) ((T_FOR) reduce 227) ((T_FOREACH) reduce 227) ((T_FUNCTION) reduce 227) ((T_FUNC_C) reduce 227) ((T_GLOBAL) reduce 227) ((T_GOTO) reduce 227) ((T_HALT_COMPILER) reduce 227) ((T_IF) reduce 227) ((T_INC) reduce 227) ((T_INCLUDE) reduce 227) ((T_INCLUDE_ONCE) reduce 227) ((T_INLINE_HTML) reduce 227) ((T_INTERFACE) reduce 227) ((T_INT_CAST) reduce 227) ((T_ISSET) reduce 227) ((T_LINE) reduce 227) ((T_LIST) reduce 227) ((T_LNUMBER) reduce 227) ((T_MATCH) reduce 227) ((T_METHOD_C) reduce 227) ((T_NAME_FULLY_QUALIFIED) reduce 227) ((T_NAME_QUALIFIED) reduce 227) ((T_NAME_RELATIVE) reduce 227) ((T_NEW) reduce 227) ((T_NS_C) reduce 227) ((T_OBJECT_CAST) reduce 227) ((T_PRINT) reduce 227) ((T_PROPERTY_C) reduce 227) ((T_READONLY) reduce 227) ((T_REQUIRE) reduce 227) ((T_REQUIRE_ONCE) reduce 227) ((T_RETURN) reduce 227) ((T_START_HEREDOC) reduce 227) ((T_STATIC) reduce 227) ((T_STRING) reduce 227) ((T_STRING_CAST) reduce 227) ((T_SWITCH) reduce 227) ((T_THROW) reduce 227) ((T_TRAIT) reduce 227) ((T_TRAIT_C) reduce 227) ((T_TRY) reduce 227) ((T_UNSET) reduce 227) ((T_UNSET_CAST) reduce 227) ((T_VARIABLE) reduce 227) ((T_VOID_CAST) reduce 227) ((T_WHILE) reduce 227) ((T_YIELD) reduce 227) ((T_YIELD_FROM) reduce 227) (("[") reduce 227) (("`") reduce 227) (("{") reduce 227) (("}") reduce 227) (("~") reduce 227)) 28759 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_EMPTY) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_VARIABLE) reduce 169) ((T_VOID_CAST) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("}") reduce 169) (("~") reduce 169)) 28761 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_EMPTY) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_VARIABLE) reduce 158) ((T_VOID_CAST) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("}") reduce 158) (("~") reduce 158)) 28763 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_EMPTY) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_VARIABLE) reduce 156) ((T_VOID_CAST) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("}") reduce 156) (("~") reduce 156)) 28769 ((("!") reduce 201) (("\"") reduce 201) (("$") reduce 201) (("(") reduce 201) (("+") reduce 201) (("-") reduce 201) ((";") reduce 201) (("@") reduce 201) ((T_ABSTRACT) reduce 201) ((T_ARRAY) reduce 201) ((T_ARRAY_CAST) reduce 201) ((T_ATTRIBUTE) reduce 201) ((T_BOOL_CAST) reduce 201) ((T_BREAK) reduce 201) ((T_CLASS) reduce 201) ((T_CLASS_C) reduce 201) ((T_CLONE) reduce 201) ((T_CONSTANT_ENCAPSED_STRING) reduce 201) ((T_CONTINUE) reduce 201) ((T_DEC) reduce 201) ((T_DECLARE) reduce 201) ((T_DIR) reduce 201) ((T_DNUMBER) reduce 201) ((T_DO) reduce 201) ((T_DOUBLE_CAST) reduce 201) ((T_ECHO) reduce 201) ((T_EMPTY) reduce 201) ((T_ENUM) reduce 201) ((T_EVAL) reduce 201) ((T_EXIT) reduce 201) ((T_FILE) reduce 201) ((T_FINAL) reduce 201) ((T_FN) reduce 201) ((T_FOR) reduce 201) ((T_FOREACH) reduce 201) ((T_FUNCTION) reduce 201) ((T_FUNC_C) reduce 201) ((T_GLOBAL) reduce 201) ((T_GOTO) reduce 201) ((T_HALT_COMPILER) reduce 201) ((T_IF) reduce 201) ((T_INC) reduce 201) ((T_INCLUDE) reduce 201) ((T_INCLUDE_ONCE) reduce 201) ((T_INLINE_HTML) reduce 201) ((T_INTERFACE) reduce 201) ((T_INT_CAST) reduce 201) ((T_ISSET) reduce 201) ((T_LINE) reduce 201) ((T_LIST) reduce 201) ((T_LNUMBER) reduce 201) ((T_MATCH) reduce 201) ((T_METHOD_C) reduce 201) ((T_NAME_FULLY_QUALIFIED) reduce 201) ((T_NAME_QUALIFIED) reduce 201) ((T_NAME_RELATIVE) reduce 201) ((T_NEW) reduce 201) ((T_NS_C) reduce 201) ((T_OBJECT_CAST) reduce 201) ((T_PRINT) reduce 201) ((T_PROPERTY_C) reduce 201) ((T_READONLY) reduce 201) ((T_REQUIRE) reduce 201) ((T_REQUIRE_ONCE) reduce 201) ((T_RETURN) reduce 201) ((T_START_HEREDOC) reduce 201) ((T_STATIC) reduce 201) ((T_STRING) reduce 201) ((T_STRING_CAST) reduce 201) ((T_SWITCH) reduce 201) ((T_THROW) reduce 201) ((T_TRAIT) reduce 201) ((T_TRAIT_C) reduce 201) ((T_TRY) reduce 201) ((T_UNSET) reduce 201) ((T_UNSET_CAST) reduce 201) ((T_VARIABLE) reduce 201) ((T_VOID_CAST) reduce 201) ((T_WHILE) reduce 201) ((T_YIELD) reduce 201) ((T_YIELD_FROM) reduce 201) (("[") reduce 201) (("`") reduce 201) (("{") reduce 201) (("}") reduce 201) (("~") reduce 201)) 28773 ((("!") reduce 244) (("\"") reduce 244) (("$") reduce 244) (("(") reduce 244) (("+") reduce 244) (("-") reduce 244) ((";") reduce 244) (("@") reduce 244) ((T_ABSTRACT) reduce 244) ((T_ARRAY) reduce 244) ((T_ARRAY_CAST) reduce 244) ((T_ATTRIBUTE) reduce 244) ((T_BOOL_CAST) reduce 244) ((T_BREAK) reduce 244) ((T_CLASS) reduce 244) ((T_CLASS_C) reduce 244) ((T_CLONE) reduce 244) ((T_CONSTANT_ENCAPSED_STRING) reduce 244) ((T_CONTINUE) reduce 244) ((T_DEC) reduce 244) ((T_DECLARE) reduce 244) ((T_DIR) reduce 244) ((T_DNUMBER) reduce 244) ((T_DO) reduce 244) ((T_DOUBLE_CAST) reduce 244) ((T_ECHO) reduce 244) ((T_ELSE) reduce 244) ((T_ELSEIF) reduce 244) ((T_EMPTY) reduce 244) ((T_ENUM) reduce 244) ((T_EVAL) reduce 244) ((T_EXIT) reduce 244) ((T_FILE) reduce 244) ((T_FINAL) reduce 244) ((T_FN) reduce 244) ((T_FOR) reduce 244) ((T_FOREACH) reduce 244) ((T_FUNCTION) reduce 244) ((T_FUNC_C) reduce 244) ((T_GLOBAL) reduce 244) ((T_GOTO) reduce 244) ((T_HALT_COMPILER) reduce 244) ((T_IF) reduce 244) ((T_INC) reduce 244) ((T_INCLUDE) reduce 244) ((T_INCLUDE_ONCE) reduce 244) ((T_INLINE_HTML) reduce 244) ((T_INTERFACE) reduce 244) ((T_INT_CAST) reduce 244) ((T_ISSET) reduce 244) ((T_LINE) reduce 244) ((T_LIST) reduce 244) ((T_LNUMBER) reduce 244) ((T_MATCH) reduce 244) ((T_METHOD_C) reduce 244) ((T_NAME_FULLY_QUALIFIED) reduce 244) ((T_NAME_QUALIFIED) reduce 244) ((T_NAME_RELATIVE) reduce 244) ((T_NEW) reduce 244) ((T_NS_C) reduce 244) ((T_OBJECT_CAST) reduce 244) ((T_PRINT) reduce 244) ((T_PROPERTY_C) reduce 244) ((T_READONLY) reduce 244) ((T_REQUIRE) reduce 244) ((T_REQUIRE_ONCE) reduce 244) ((T_RETURN) reduce 244) ((T_START_HEREDOC) reduce 244) ((T_STATIC) reduce 244) ((T_STRING) reduce 244) ((T_STRING_CAST) reduce 244) ((T_SWITCH) reduce 244) ((T_THROW) reduce 244) ((T_TRAIT) reduce 244) ((T_TRAIT_C) reduce 244) ((T_TRY) reduce 244) ((T_UNSET) reduce 244) ((T_UNSET_CAST) reduce 244) ((T_VARIABLE) reduce 244) ((T_VOID_CAST) reduce 244) ((T_WHILE) reduce 244) ((T_YIELD) reduce 244) ((T_YIELD_FROM) reduce 244) (("[") reduce 244) (("`") reduce 244) (("{") reduce 244) (("}") reduce 244) (("~") reduce 244)) 28776 ((("!") reduce 146) (("\"") reduce 146) (("$") reduce 146) (("(") reduce 146) (("+") reduce 146) (("-") reduce 146) ((";") reduce 146) (("@") reduce 146) ((T_ABSTRACT) reduce 146) ((T_ARRAY) reduce 146) ((T_ARRAY_CAST) reduce 146) ((T_ATTRIBUTE) reduce 146) ((T_BOOL_CAST) reduce 146) ((T_BREAK) reduce 146) ((T_CLASS) reduce 146) ((T_CLASS_C) reduce 146) ((T_CLONE) reduce 146) ((T_CONSTANT_ENCAPSED_STRING) reduce 146) ((T_CONTINUE) reduce 146) ((T_DEC) reduce 146) ((T_DECLARE) reduce 146) ((T_DIR) reduce 146) ((T_DNUMBER) reduce 146) ((T_DO) reduce 146) ((T_DOUBLE_CAST) reduce 146) ((T_ECHO) reduce 146) ((T_EMPTY) reduce 146) ((T_ENUM) reduce 146) ((T_EVAL) reduce 146) ((T_EXIT) reduce 146) ((T_FILE) reduce 146) ((T_FINAL) reduce 146) ((T_FN) reduce 146) ((T_FOR) reduce 146) ((T_FOREACH) reduce 146) ((T_FUNCTION) reduce 146) ((T_FUNC_C) reduce 146) ((T_GLOBAL) reduce 146) ((T_GOTO) reduce 146) ((T_HALT_COMPILER) reduce 146) ((T_IF) reduce 146) ((T_INC) reduce 146) ((T_INCLUDE) reduce 146) ((T_INCLUDE_ONCE) reduce 146) ((T_INLINE_HTML) reduce 146) ((T_INTERFACE) reduce 146) ((T_INT_CAST) reduce 146) ((T_ISSET) reduce 146) ((T_LINE) reduce 146) ((T_LIST) reduce 146) ((T_LNUMBER) reduce 146) ((T_MATCH) reduce 146) ((T_METHOD_C) reduce 146) ((T_NAME_FULLY_QUALIFIED) reduce 146) ((T_NAME_QUALIFIED) reduce 146) ((T_NAME_RELATIVE) reduce 146) ((T_NEW) reduce 146) ((T_NS_C) reduce 146) ((T_OBJECT_CAST) reduce 146) ((T_PRINT) reduce 146) ((T_PROPERTY_C) reduce 146) ((T_READONLY) reduce 146) ((T_REQUIRE) reduce 146) ((T_REQUIRE_ONCE) reduce 146) ((T_RETURN) reduce 146) ((T_START_HEREDOC) reduce 146) ((T_STATIC) reduce 146) ((T_STRING) reduce 146) ((T_STRING_CAST) reduce 146) ((T_SWITCH) reduce 146) ((T_THROW) reduce 146) ((T_TRAIT) reduce 146) ((T_TRAIT_C) reduce 146) ((T_TRY) reduce 146) ((T_UNSET) reduce 146) ((T_UNSET_CAST) reduce 146) ((T_VARIABLE) reduce 146) ((T_VOID_CAST) reduce 146) ((T_WHILE) reduce 146) ((T_YIELD) reduce 146) ((T_YIELD_FROM) reduce 146) (("[") reduce 146) (("`") reduce 146) (("{") reduce 146) (("}") reduce 146) (("~") reduce 146)) 28778 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_EMPTY) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_VARIABLE) reduce 168) ((T_VOID_CAST) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("}") reduce 168) (("~") reduce 168)) 28780 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_EMPTY) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_VARIABLE) reduce 157) ((T_VOID_CAST) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("}") reduce 157) (("~") reduce 157)) 28790 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_EMPTY) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_VARIABLE) reduce 164) ((T_VOID_CAST) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("}") reduce 164) (("~") reduce 164)) 28791 ((("!") reduce 220) (("\"") reduce 220) (("$") reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_EMPTY) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_VARIABLE) reduce 220) ((T_VOID_CAST) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("}") reduce 220) (("~") reduce 220)) 28794 ((("!") reduce 221) (("\"") reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_EMPTY) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_VARIABLE) reduce 221) ((T_VOID_CAST) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("}") reduce 221) (("~") reduce 221)) 28795 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_EMPTY) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_VARIABLE) reduce 163) ((T_VOID_CAST) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("}") reduce 163) (("~") reduce 163)) 28804 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_EMPTY) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_VARIABLE) reduce 152) ((T_VOID_CAST) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("}") reduce 152) (("~") reduce 152)) 28805 ((("!") reduce 218) (("\"") reduce 218) (("$") reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_EMPTY) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_VARIABLE) reduce 218) ((T_VOID_CAST) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("}") reduce 218) (("~") reduce 218)) 28808 ((("!") reduce 219) (("\"") reduce 219) (("$") reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_EMPTY) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_VARIABLE) reduce 219) ((T_VOID_CAST) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("}") reduce 219) (("~") reduce 219)) 28815 ((("!") reduce 202) (("\"") reduce 202) (("$") reduce 202) (("(") reduce 202) (("+") reduce 202) (("-") reduce 202) ((";") reduce 202) (("@") reduce 202) ((T_ABSTRACT) reduce 202) ((T_ARRAY) reduce 202) ((T_ARRAY_CAST) reduce 202) ((T_ATTRIBUTE) reduce 202) ((T_BOOL_CAST) reduce 202) ((T_BREAK) reduce 202) ((T_CLASS) reduce 202) ((T_CLASS_C) reduce 202) ((T_CLONE) reduce 202) ((T_CONSTANT_ENCAPSED_STRING) reduce 202) ((T_CONTINUE) reduce 202) ((T_DEC) reduce 202) ((T_DECLARE) reduce 202) ((T_DIR) reduce 202) ((T_DNUMBER) reduce 202) ((T_DO) reduce 202) ((T_DOUBLE_CAST) reduce 202) ((T_ECHO) reduce 202) ((T_EMPTY) reduce 202) ((T_ENUM) reduce 202) ((T_EVAL) reduce 202) ((T_EXIT) reduce 202) ((T_FILE) reduce 202) ((T_FINAL) reduce 202) ((T_FN) reduce 202) ((T_FOR) reduce 202) ((T_FOREACH) reduce 202) ((T_FUNCTION) reduce 202) ((T_FUNC_C) reduce 202) ((T_GLOBAL) reduce 202) ((T_GOTO) reduce 202) ((T_HALT_COMPILER) reduce 202) ((T_IF) reduce 202) ((T_INC) reduce 202) ((T_INCLUDE) reduce 202) ((T_INCLUDE_ONCE) reduce 202) ((T_INLINE_HTML) reduce 202) ((T_INTERFACE) reduce 202) ((T_INT_CAST) reduce 202) ((T_ISSET) reduce 202) ((T_LINE) reduce 202) ((T_LIST) reduce 202) ((T_LNUMBER) reduce 202) ((T_MATCH) reduce 202) ((T_METHOD_C) reduce 202) ((T_NAME_FULLY_QUALIFIED) reduce 202) ((T_NAME_QUALIFIED) reduce 202) ((T_NAME_RELATIVE) reduce 202) ((T_NEW) reduce 202) ((T_NS_C) reduce 202) ((T_OBJECT_CAST) reduce 202) ((T_PRINT) reduce 202) ((T_PROPERTY_C) reduce 202) ((T_READONLY) reduce 202) ((T_REQUIRE) reduce 202) ((T_REQUIRE_ONCE) reduce 202) ((T_RETURN) reduce 202) ((T_START_HEREDOC) reduce 202) ((T_STATIC) reduce 202) ((T_STRING) reduce 202) ((T_STRING_CAST) reduce 202) ((T_SWITCH) reduce 202) ((T_THROW) reduce 202) ((T_TRAIT) reduce 202) ((T_TRAIT_C) reduce 202) ((T_TRY) reduce 202) ((T_UNSET) reduce 202) ((T_UNSET_CAST) reduce 202) ((T_VARIABLE) reduce 202) ((T_VOID_CAST) reduce 202) ((T_WHILE) reduce 202) ((T_YIELD) reduce 202) ((T_YIELD_FROM) reduce 202) (("[") reduce 202) (("`") reduce 202) (("{") reduce 202) (("}") reduce 202) (("~") reduce 202)) 28817 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_EMPTY) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_VARIABLE) reduce 159) ((T_VOID_CAST) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("}") reduce 159) (("~") reduce 159)) 28823 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_EMPTY) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_VARIABLE) reduce 151) ((T_VOID_CAST) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("}") reduce 151) (("~") reduce 151)) 28828 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_EMPTY) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_VARIABLE) reduce 165) ((T_VOID_CAST) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("}") reduce 165) (("~") reduce 165)) 28829 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_EMPTY) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_VARIABLE) reduce 222) ((T_VOID_CAST) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("}") reduce 222) (("~") reduce 222)) 28832 ((("!") reduce 223) (("\"") reduce 223) (("$") reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_EMPTY) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_VARIABLE) reduce 223) ((T_VOID_CAST) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("}") reduce 223) (("~") reduce 223)) 28834 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_EMPTY) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_VARIABLE) reduce 155) ((T_VOID_CAST) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("}") reduce 155) (("~") reduce 155)) 28841 ((("!") reduce 190) (("\"") reduce 190) (("$") reduce 190) (("(") reduce 190) (("+") reduce 190) (("-") reduce 190) ((";") reduce 190) (("@") reduce 190) ((T_ABSTRACT) reduce 190) ((T_ARRAY) reduce 190) ((T_ARRAY_CAST) reduce 190) ((T_ATTRIBUTE) reduce 190) ((T_BOOL_CAST) reduce 190) ((T_BREAK) reduce 190) ((T_CLASS) reduce 190) ((T_CLASS_C) reduce 190) ((T_CLONE) reduce 190) ((T_CONSTANT_ENCAPSED_STRING) reduce 190) ((T_CONTINUE) reduce 190) ((T_DEC) reduce 190) ((T_DECLARE) reduce 190) ((T_DIR) reduce 190) ((T_DNUMBER) reduce 190) ((T_DO) reduce 190) ((T_DOUBLE_CAST) reduce 190) ((T_ECHO) reduce 190) ((T_EMPTY) reduce 190) ((T_ENUM) reduce 190) ((T_EVAL) reduce 190) ((T_EXIT) reduce 190) ((T_FILE) reduce 190) ((T_FINAL) reduce 190) ((T_FN) reduce 190) ((T_FOR) reduce 190) ((T_FOREACH) reduce 190) ((T_FUNCTION) reduce 190) ((T_FUNC_C) reduce 190) ((T_GLOBAL) reduce 190) ((T_GOTO) reduce 190) ((T_HALT_COMPILER) reduce 190) ((T_IF) reduce 190) ((T_INC) reduce 190) ((T_INCLUDE) reduce 190) ((T_INCLUDE_ONCE) reduce 190) ((T_INLINE_HTML) reduce 190) ((T_INTERFACE) reduce 190) ((T_INT_CAST) reduce 190) ((T_ISSET) reduce 190) ((T_LINE) reduce 190) ((T_LIST) reduce 190) ((T_LNUMBER) reduce 190) ((T_MATCH) reduce 190) ((T_METHOD_C) reduce 190) ((T_NAME_FULLY_QUALIFIED) reduce 190) ((T_NAME_QUALIFIED) reduce 190) ((T_NAME_RELATIVE) reduce 190) ((T_NEW) reduce 190) ((T_NS_C) reduce 190) ((T_OBJECT_CAST) reduce 190) ((T_PRINT) reduce 190) ((T_PROPERTY_C) reduce 190) ((T_READONLY) reduce 190) ((T_REQUIRE) reduce 190) ((T_REQUIRE_ONCE) reduce 190) ((T_RETURN) reduce 190) ((T_START_HEREDOC) reduce 190) ((T_STATIC) reduce 190) ((T_STRING) reduce 190) ((T_STRING_CAST) reduce 190) ((T_SWITCH) reduce 190) ((T_THROW) reduce 190) ((T_TRAIT) reduce 190) ((T_TRAIT_C) reduce 190) ((T_TRY) reduce 190) ((T_UNSET) reduce 190) ((T_UNSET_CAST) reduce 190) ((T_VARIABLE) reduce 190) ((T_VOID_CAST) reduce 190) ((T_WHILE) reduce 190) ((T_YIELD) reduce 190) ((T_YIELD_FROM) reduce 190) (("[") reduce 190) (("`") reduce 190) (("{") reduce 190) (("}") reduce 190) (("~") reduce 190)) 28843 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_EMPTY) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_VARIABLE) reduce 154) ((T_VOID_CAST) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("}") reduce 154) (("~") reduce 154)) 28846 (((T_ABSTRACT) reduce 100) ((T_ATTRIBUTE) reduce 100) ((T_CLASS) reduce 100) ((T_ENUM) reduce 100) ((T_FINAL) reduce 100) ((T_FN) reduce 100) ((T_FUNCTION) reduce 100) ((T_INTERFACE) reduce 100) ((T_READONLY) reduce 100) ((T_STATIC) reduce 100) ((T_TRAIT) reduce 100)) 28848 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 280) (("{") reduce 280) (("|") reduce 280)) 28849 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) (("{") reduce 281) (("|") reduce 281)) 28850 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) (("{") reduce 94) (("|") reduce 94)) 28851 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) (("{") reduce 93) (("|") reduce 93)) 28852 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) (("{") reduce 95) (("|") reduce 95)) 28853 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 269) (("{") reduce 269) (("|") reduce 269)) 28854 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) (("{") reduce 92) (("|") reduce 92)) 28855 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) (("{") reduce 267)) 28856 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) (("{") reduce 282) (("|") reduce 282)) 28857 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) (("{") reduce 264) (("|") reduce 270)) 28858 ((("{") reduce 290)) 28859 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 268) (("{") reduce 268) (("|") reduce 268)) 28860 ((("{") reduce 266) (("|") shift)) 28864 ((("{") reduce 280) (("|") reduce 280)) 28865 ((("{") reduce 281) (("|") reduce 281)) 28866 ((("{") reduce 94) (("|") reduce 94)) 28867 ((("{") reduce 93) (("|") reduce 93)) 28868 ((("{") reduce 95) (("|") reduce 95)) 28869 ((("{") reduce 269) (("|") reduce 269)) 28870 ((("{") reduce 92) (("|") reduce 92)) 28871 ((("{") reduce 282) (("|") reduce 282)) 28872 ((("{") reduce 270) (("|") reduce 270)) 28873 ((("{") reduce 268) (("|") reduce 268)) 28874 ((("{") reduce 272) (("|") reduce 272)) 28876 ((("{") reduce 271) (("|") reduce 271)) 28878 ((("{") reduce 273) (("|") reduce 273)) 28880 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 280) (("{") reduce 280)) 28881 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) (("{") reduce 281)) 28882 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) (("{") reduce 94)) 28883 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) (("{") reduce 93)) 28884 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) (("{") reduce 95)) 28885 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 269) (("{") reduce 269)) 28886 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) (("{") reduce 92)) 28887 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) (("{") reduce 282)) 28888 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 274) (("{") reduce 274)) 28889 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 268) (("{") reduce 268)) 28891 (((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 275) (("{") reduce 275)) 28892 ((("{") reduce 280)) 28893 ((("{") reduce 281)) 28894 ((("{") reduce 94)) 28895 ((("{") reduce 93)) 28896 ((("{") reduce 95)) 28897 ((("{") reduce 269)) 28898 ((("{") reduce 92)) 28899 ((("{") reduce 282)) 28900 ((("{") reduce 265)) 28901 ((("{") reduce 268)) 28902 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_VARIABLE) shift)) 28903 (((T_VARIABLE) reduce 80)) 28904 (((T_VARIABLE) reduce 81)) 28905 (((")") reduce 502) ((",") reduce 502)) 28907 (((")") reduce 501) ((",") reduce 501)) 28909 (((")") reduce 126) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_VARIABLE) shift)) 28911 (((":") reduce 499) (("{") reduce 499)) 28912 (((")") reduce 500) ((",") reduce 500)) 28913 (((")") reduce 503) ((",") reduce 503)) 28915 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 280) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 280) ((T_ELLIPSIS) reduce 280) ((T_VARIABLE) reduce 280) (("|") reduce 280)) 28916 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_ELLIPSIS) reduce 281) ((T_VARIABLE) reduce 281) (("|") reduce 281)) 28917 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_ELLIPSIS) reduce 94) ((T_VARIABLE) reduce 94) (("|") reduce 94)) 28918 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_ELLIPSIS) reduce 93) ((T_VARIABLE) reduce 93) (("|") reduce 93)) 28919 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_ELLIPSIS) reduce 95) ((T_VARIABLE) reduce 95) (("|") reduce 95)) 28920 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_ELLIPSIS) reduce 92) ((T_VARIABLE) reduce 92) (("|") reduce 92)) 28921 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 279) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_ELLIPSIS) reduce 279) ((T_VARIABLE) reduce 279)) 28922 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_ELLIPSIS) reduce 282) ((T_VARIABLE) reduce 282) (("|") reduce 282)) 28923 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_ELLIPSIS) reduce 185) ((T_VARIABLE) reduce 185)) 28924 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 263) ((T_ELLIPSIS) reduce 263) ((T_VARIABLE) reduce 263)) 28925 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 276) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_ELLIPSIS) reduce 276) ((T_VARIABLE) reduce 276) (("|") reduce 283)) 28926 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 278) ((T_ELLIPSIS) reduce 278) ((T_VARIABLE) reduce 278) (("|") shift)) 28930 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 280) ((T_ELLIPSIS) reduce 280) ((T_VARIABLE) reduce 280) (("|") reduce 280)) 28931 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_ELLIPSIS) reduce 281) ((T_VARIABLE) reduce 281) (("|") reduce 281)) 28932 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_ELLIPSIS) reduce 94) ((T_VARIABLE) reduce 94) (("|") reduce 94)) 28933 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_ELLIPSIS) reduce 93) ((T_VARIABLE) reduce 93) (("|") reduce 93)) 28934 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_ELLIPSIS) reduce 95) ((T_VARIABLE) reduce 95) (("|") reduce 95)) 28935 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_ELLIPSIS) reduce 92) ((T_VARIABLE) reduce 92) (("|") reduce 92)) 28936 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_ELLIPSIS) reduce 282) ((T_VARIABLE) reduce 282) (("|") reduce 282)) 28937 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 283) ((T_ELLIPSIS) reduce 283) ((T_VARIABLE) reduce 283) (("|") reduce 283)) 28938 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 285) ((T_ELLIPSIS) reduce 285) ((T_VARIABLE) reduce 285) (("|") reduce 285)) 28940 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 284) ((T_ELLIPSIS) reduce 284) ((T_VARIABLE) reduce 284) (("|") reduce 284)) 28942 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 286) ((T_ELLIPSIS) reduce 286) ((T_VARIABLE) reduce 286) (("|") reduce 286)) 28944 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 280) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 280) ((T_ELLIPSIS) reduce 280) ((T_VARIABLE) reduce 280)) 28945 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_ELLIPSIS) reduce 281) ((T_VARIABLE) reduce 281)) 28946 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_ELLIPSIS) reduce 94) ((T_VARIABLE) reduce 94)) 28947 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_ELLIPSIS) reduce 93) ((T_VARIABLE) reduce 93)) 28948 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_ELLIPSIS) reduce 95) ((T_VARIABLE) reduce 95)) 28949 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_ELLIPSIS) reduce 92) ((T_VARIABLE) reduce 92)) 28950 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_ELLIPSIS) reduce 282) ((T_VARIABLE) reduce 282)) 28951 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 287) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 287) ((T_ELLIPSIS) reduce 287) ((T_VARIABLE) reduce 287)) 28952 (((T_ELLIPSIS) reduce 186) ((T_VARIABLE) reduce 186)) 28953 (((T_ELLIPSIS) shift) ((T_VARIABLE) reduce 187)) 28954 (((T_VARIABLE) reduce 188)) 28956 (((")") reduce 493) ((",") reduce 493) (("=") reduce 493) (("{") reduce 493)) 28957 (((")") reduce 371) ((",") reduce 371) (("=") shift) (("{") shift)) 28959 (((")") reduce 260) ((",") reduce 260)) 28962 (((")") reduce 372) ((",") reduce 372)) 28973 ((("%") reduce 540) ((")") reduce 540) (("*") reduce 540) (("+") reduce 540) ((",") reduce 540) (("-") reduce 540) ((".") reduce 540) (("/") reduce 540) (("<") reduce 540) ((">") reduce 540) (("?") reduce 540) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_BOOLEAN_AND) reduce 540) ((T_BOOLEAN_OR) reduce 540) ((T_COALESCE) reduce 540) ((T_INSTANCEOF) reduce 540) ((T_IS_EQUAL) reduce 540) ((T_IS_GREATER_OR_EQUAL) reduce 540) ((T_IS_IDENTICAL) reduce 540) ((T_IS_NOT_EQUAL) reduce 540) ((T_IS_NOT_IDENTICAL) reduce 540) ((T_IS_SMALLER_OR_EQUAL) reduce 540) ((T_LOGICAL_AND) reduce 540) ((T_LOGICAL_OR) reduce 540) ((T_LOGICAL_XOR) reduce 540) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 540) ((T_OBJECT_OPERATOR) reduce 540) ((T_PIPE) reduce 540) ((T_POW) reduce 540) ((T_SL) reduce 540) ((T_SPACESHIP) reduce 540) ((T_SR) reduce 540) (("[") reduce 540) (("^") reduce 540) (("{") reduce 540) (("|") reduce 540)) 28975 ((("%") reduce 521) (("(") reduce 521) ((")") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_PIPE) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 28977 ((("%") reduce 534) ((")") reduce 534) (("*") reduce 534) (("+") reduce 534) ((",") reduce 534) (("-") reduce 534) ((".") reduce 534) (("/") reduce 534) (("<") reduce 534) ((">") reduce 534) (("?") reduce 534) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_BOOLEAN_AND) reduce 534) ((T_BOOLEAN_OR) reduce 534) ((T_COALESCE) reduce 534) ((T_INSTANCEOF) reduce 534) ((T_IS_EQUAL) reduce 534) ((T_IS_GREATER_OR_EQUAL) reduce 534) ((T_IS_IDENTICAL) reduce 534) ((T_IS_NOT_EQUAL) reduce 534) ((T_IS_NOT_IDENTICAL) reduce 534) ((T_IS_SMALLER_OR_EQUAL) reduce 534) ((T_LOGICAL_AND) reduce 534) ((T_LOGICAL_OR) reduce 534) ((T_LOGICAL_XOR) reduce 534) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PIPE) reduce 534) ((T_POW) reduce 534) ((T_SL) reduce 534) ((T_SPACESHIP) reduce 534) ((T_SR) reduce 534) (("[") reduce 534) (("^") reduce 534) (("{") reduce 534) (("|") reduce 534)) 28978 ((("%") reduce 524) ((")") reduce 524) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_PIPE) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 28982 ((("%") reduce 517) (("(") shift) ((")") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_PIPE) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 28983 ((("%") reduce 533) ((")") reduce 533) (("*") reduce 533) (("+") reduce 533) ((",") reduce 533) (("-") reduce 533) ((".") reduce 533) (("/") reduce 533) (("<") reduce 533) ((">") reduce 533) (("?") reduce 533) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_BOOLEAN_AND) reduce 533) ((T_BOOLEAN_OR) reduce 533) ((T_COALESCE) reduce 533) ((T_INSTANCEOF) reduce 533) ((T_IS_EQUAL) reduce 533) ((T_IS_GREATER_OR_EQUAL) reduce 533) ((T_IS_IDENTICAL) reduce 533) ((T_IS_NOT_EQUAL) reduce 533) ((T_IS_NOT_IDENTICAL) reduce 533) ((T_IS_SMALLER_OR_EQUAL) reduce 533) ((T_LOGICAL_AND) reduce 533) ((T_LOGICAL_OR) reduce 533) ((T_LOGICAL_XOR) reduce 533) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PIPE) reduce 533) ((T_POW) reduce 533) ((T_SL) reduce 533) ((T_SPACESHIP) reduce 533) ((T_SR) reduce 533) (("[") reduce 533) (("^") reduce 533) (("{") reduce 533) (("|") reduce 533)) 28984 ((("%") reduce 537) ((")") reduce 537) (("*") reduce 537) (("+") reduce 537) ((",") reduce 537) (("-") reduce 537) ((".") reduce 537) (("/") reduce 537) (("<") reduce 537) ((">") reduce 537) (("?") reduce 537) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_BOOLEAN_AND) reduce 537) ((T_BOOLEAN_OR) reduce 537) ((T_COALESCE) reduce 537) ((T_INSTANCEOF) reduce 537) ((T_IS_EQUAL) reduce 537) ((T_IS_GREATER_OR_EQUAL) reduce 537) ((T_IS_IDENTICAL) reduce 537) ((T_IS_NOT_EQUAL) reduce 537) ((T_IS_NOT_IDENTICAL) reduce 537) ((T_IS_SMALLER_OR_EQUAL) reduce 537) ((T_LOGICAL_AND) reduce 537) ((T_LOGICAL_OR) reduce 537) ((T_LOGICAL_XOR) reduce 537) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_PIPE) reduce 537) ((T_POW) reduce 537) ((T_SL) reduce 537) ((T_SPACESHIP) reduce 537) ((T_SR) reduce 537) (("[") reduce 537) (("^") reduce 537) (("{") reduce 537) (("|") reduce 537)) 28990 ((("%") reduce 532) ((")") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PIPE) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 28992 ((("%") reduce 523) ((")") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_PIPE) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 28994 ((("%") reduce 536) ((")") reduce 536) (("*") reduce 536) (("+") reduce 536) ((",") reduce 536) (("-") reduce 536) ((".") reduce 536) (("/") reduce 536) (("<") reduce 536) ((">") reduce 536) (("?") reduce 536) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_BOOLEAN_AND) reduce 536) ((T_BOOLEAN_OR) reduce 536) ((T_COALESCE) reduce 536) ((T_INSTANCEOF) reduce 536) ((T_IS_EQUAL) reduce 536) ((T_IS_GREATER_OR_EQUAL) reduce 536) ((T_IS_IDENTICAL) reduce 536) ((T_IS_NOT_EQUAL) reduce 536) ((T_IS_NOT_IDENTICAL) reduce 536) ((T_IS_SMALLER_OR_EQUAL) reduce 536) ((T_LOGICAL_AND) reduce 536) ((T_LOGICAL_OR) reduce 536) ((T_LOGICAL_XOR) reduce 536) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 536) ((T_OBJECT_OPERATOR) reduce 536) ((T_PIPE) reduce 536) ((T_POW) reduce 536) ((T_SL) reduce 536) ((T_SPACESHIP) reduce 536) ((T_SR) reduce 536) (("[") reduce 536) (("^") reduce 536) (("{") reduce 536) (("|") reduce 536)) 28995 ((("%") reduce 94) (("(") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 28996 ((("%") reduce 93) (("(") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 28997 ((("%") reduce 95) (("(") reduce 95) ((")") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("{") reduce 95) (("|") reduce 95)) 28999 ((("%") reduce 539) ((")") reduce 539) (("*") reduce 539) (("+") reduce 539) ((",") reduce 539) (("-") reduce 539) ((".") reduce 539) (("/") reduce 539) (("<") reduce 539) ((">") reduce 539) (("?") reduce 539) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_BOOLEAN_AND) reduce 539) ((T_BOOLEAN_OR) reduce 539) ((T_COALESCE) reduce 539) ((T_INSTANCEOF) reduce 539) ((T_IS_EQUAL) reduce 539) ((T_IS_GREATER_OR_EQUAL) reduce 539) ((T_IS_IDENTICAL) reduce 539) ((T_IS_NOT_EQUAL) reduce 539) ((T_IS_NOT_IDENTICAL) reduce 539) ((T_IS_SMALLER_OR_EQUAL) reduce 539) ((T_LOGICAL_AND) reduce 539) ((T_LOGICAL_OR) reduce 539) ((T_LOGICAL_XOR) reduce 539) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 539) ((T_OBJECT_OPERATOR) reduce 539) ((T_PIPE) reduce 539) ((T_POW) reduce 539) ((T_SL) reduce 539) ((T_SPACESHIP) reduce 539) ((T_SR) reduce 539) (("[") reduce 539) (("^") reduce 539) (("{") reduce 539) (("|") reduce 539)) 29002 ((("%") reduce 538) ((")") reduce 538) (("*") reduce 538) (("+") reduce 538) ((",") reduce 538) (("-") reduce 538) ((".") reduce 538) (("/") reduce 538) (("<") reduce 538) ((">") reduce 538) (("?") reduce 538) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_BOOLEAN_AND) reduce 538) ((T_BOOLEAN_OR) reduce 538) ((T_COALESCE) reduce 538) ((T_INSTANCEOF) reduce 538) ((T_IS_EQUAL) reduce 538) ((T_IS_GREATER_OR_EQUAL) reduce 538) ((T_IS_IDENTICAL) reduce 538) ((T_IS_NOT_EQUAL) reduce 538) ((T_IS_NOT_IDENTICAL) reduce 538) ((T_IS_SMALLER_OR_EQUAL) reduce 538) ((T_LOGICAL_AND) reduce 538) ((T_LOGICAL_OR) reduce 538) ((T_LOGICAL_XOR) reduce 538) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 538) ((T_OBJECT_OPERATOR) reduce 538) ((T_PIPE) reduce 538) ((T_POW) reduce 538) ((T_SL) reduce 538) ((T_SPACESHIP) reduce 538) ((T_SR) reduce 538) (("[") reduce 538) (("^") reduce 538) (("{") reduce 538) (("|") reduce 538)) 29008 ((("%") reduce 92) (("(") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 29011 ((("%") reduce 535) ((")") reduce 535) (("*") reduce 535) (("+") reduce 535) ((",") reduce 535) (("-") reduce 535) ((".") reduce 535) (("/") reduce 535) (("<") reduce 535) ((">") reduce 535) (("?") reduce 535) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_BOOLEAN_AND) reduce 535) ((T_BOOLEAN_OR) reduce 535) ((T_COALESCE) reduce 535) ((T_INSTANCEOF) reduce 535) ((T_IS_EQUAL) reduce 535) ((T_IS_GREATER_OR_EQUAL) reduce 535) ((T_IS_IDENTICAL) reduce 535) ((T_IS_NOT_EQUAL) reduce 535) ((T_IS_NOT_IDENTICAL) reduce 535) ((T_IS_SMALLER_OR_EQUAL) reduce 535) ((T_LOGICAL_AND) reduce 535) ((T_LOGICAL_OR) reduce 535) ((T_LOGICAL_XOR) reduce 535) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PIPE) reduce 535) ((T_POW) reduce 535) ((T_SL) reduce 535) ((T_SPACESHIP) reduce 535) ((T_SR) reduce 535) (("[") reduce 535) (("^") reduce 535) (("{") reduce 535) (("|") reduce 535)) 29013 ((("%") reduce 568) (("(") reduce 568) ((")") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) (("=") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AND_EQUAL) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_COALESCE_EQUAL) reduce 568) ((T_CONCAT_EQUAL) reduce 568) ((T_DEC) reduce 568) ((T_DIV_EQUAL) reduce 568) ((T_INC) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_MINUS_EQUAL) reduce 568) ((T_MOD_EQUAL) reduce 568) ((T_MUL_EQUAL) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_OR_EQUAL) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_PLUS_EQUAL) reduce 568) ((T_POW) reduce 568) ((T_POW_EQUAL) reduce 568) ((T_SL) reduce 568) ((T_SL_EQUAL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) ((T_SR_EQUAL) reduce 568) ((T_XOR_EQUAL) reduce 568) (("[") reduce 568) (("^") reduce 568) (("{") reduce 568) (("|") reduce 568)) 29014 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 479) (("(") shift) ((")") reduce 479) (("*") reduce 479) (("+") shift) ((",") reduce 479) (("-") shift) ((".") reduce 479) (("/") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 479) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 479) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PIPE) reduce 479) ((T_POW) reduce 479) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 479) (("`") shift) (("{") reduce 479) (("|") reduce 479) (("~") shift)) 29021 ((("%") reduce 564) (("(") reduce 555) ((")") reduce 564) (("*") reduce 564) (("+") reduce 564) ((",") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) (("<") reduce 564) (("=") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AND_EQUAL) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_COALESCE_EQUAL) reduce 564) ((T_CONCAT_EQUAL) reduce 564) ((T_DEC) reduce 564) ((T_DIV_EQUAL) reduce 564) ((T_INC) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_MINUS_EQUAL) reduce 564) ((T_MOD_EQUAL) reduce 564) ((T_MUL_EQUAL) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_OR_EQUAL) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_PLUS_EQUAL) reduce 564) ((T_POW) reduce 564) ((T_POW_EQUAL) reduce 564) ((T_SL) reduce 564) ((T_SL_EQUAL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) ((T_SR_EQUAL) reduce 564) ((T_XOR_EQUAL) reduce 564) (("[") reduce 564) (("^") reduce 564) (("{") reduce 564) (("|") reduce 564)) 29022 ((("%") reduce 530) ((")") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PIPE) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 551) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 29024 ((("%") reduce 529) ((")") reduce 529) (("*") reduce 529) (("+") reduce 529) ((",") reduce 529) (("-") reduce 529) ((".") reduce 529) (("/") reduce 529) (("<") reduce 529) ((">") reduce 529) (("?") reduce 529) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_BOOLEAN_AND) reduce 529) ((T_BOOLEAN_OR) reduce 529) ((T_COALESCE) reduce 529) ((T_INSTANCEOF) reduce 529) ((T_IS_EQUAL) reduce 529) ((T_IS_GREATER_OR_EQUAL) reduce 529) ((T_IS_IDENTICAL) reduce 529) ((T_IS_NOT_EQUAL) reduce 529) ((T_IS_NOT_IDENTICAL) reduce 529) ((T_IS_SMALLER_OR_EQUAL) reduce 529) ((T_LOGICAL_AND) reduce 529) ((T_LOGICAL_OR) reduce 529) ((T_LOGICAL_XOR) reduce 529) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PIPE) reduce 529) ((T_POW) reduce 529) ((T_SL) reduce 529) ((T_SPACESHIP) reduce 529) ((T_SR) reduce 529) (("[") reduce 554) (("^") reduce 529) (("{") reduce 529) (("|") reduce 529)) 29025 ((("%") reduce 528) (("(") reduce 557) ((")") reduce 528) (("*") reduce 528) (("+") reduce 528) ((",") reduce 528) (("-") reduce 528) ((".") reduce 528) (("/") reduce 528) (("<") reduce 528) ((">") reduce 528) (("?") reduce 528) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_BOOLEAN_AND) reduce 528) ((T_BOOLEAN_OR) reduce 528) ((T_COALESCE) reduce 528) ((T_INSTANCEOF) reduce 528) ((T_IS_EQUAL) reduce 528) ((T_IS_GREATER_OR_EQUAL) reduce 528) ((T_IS_IDENTICAL) reduce 528) ((T_IS_NOT_EQUAL) reduce 528) ((T_IS_NOT_IDENTICAL) reduce 528) ((T_IS_SMALLER_OR_EQUAL) reduce 528) ((T_LOGICAL_AND) reduce 528) ((T_LOGICAL_OR) reduce 528) ((T_LOGICAL_XOR) reduce 528) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PIPE) reduce 528) ((T_POW) reduce 528) ((T_SL) reduce 528) ((T_SPACESHIP) reduce 528) ((T_SR) reduce 528) (("[") reduce 550) (("^") reduce 528) (("{") reduce 528) (("|") reduce 528)) 29026 ((("%") shift) ((")") reduce 371) (("*") shift) (("+") shift) ((",") reduce 371) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") shift) (("|") shift)) 29029 ((("%") reduce 563) (("(") reduce 563) ((")") reduce 563) (("*") reduce 563) (("+") reduce 563) ((",") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) (("<") reduce 563) (("=") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AND_EQUAL) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_COALESCE_EQUAL) reduce 563) ((T_CONCAT_EQUAL) reduce 563) ((T_DEC) reduce 563) ((T_DIV_EQUAL) reduce 563) ((T_INC) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_MINUS_EQUAL) reduce 563) ((T_MOD_EQUAL) reduce 563) ((T_MUL_EQUAL) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_OR_EQUAL) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_PLUS_EQUAL) reduce 563) ((T_POW) reduce 563) ((T_POW_EQUAL) reduce 563) ((T_SL) reduce 563) ((T_SL_EQUAL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) ((T_SR_EQUAL) reduce 563) ((T_XOR_EQUAL) reduce 563) (("[") reduce 563) (("^") reduce 563) (("{") reduce 563) (("|") reduce 563)) 29030 ((("%") reduce 484) ((")") reduce 484) (("*") reduce 484) (("+") reduce 484) ((",") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_PIPE) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("^") reduce 484) (("{") reduce 484) (("|") reduce 484)) 29031 ((("%") reduce 466) ((")") reduce 466) (("*") reduce 466) (("+") reduce 466) ((",") reduce 466) (("-") reduce 466) ((".") reduce 466) (("/") reduce 466) (("<") reduce 466) ((">") reduce 466) (("?") reduce 466) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_BOOLEAN_AND) reduce 466) ((T_BOOLEAN_OR) reduce 466) ((T_COALESCE) reduce 466) ((T_INSTANCEOF) reduce 466) ((T_IS_EQUAL) reduce 466) ((T_IS_GREATER_OR_EQUAL) reduce 466) ((T_IS_IDENTICAL) reduce 466) ((T_IS_NOT_EQUAL) reduce 466) ((T_IS_NOT_IDENTICAL) reduce 466) ((T_IS_SMALLER_OR_EQUAL) reduce 466) ((T_LOGICAL_AND) reduce 466) ((T_LOGICAL_OR) reduce 466) ((T_LOGICAL_XOR) reduce 466) ((T_PIPE) reduce 466) ((T_POW) reduce 466) ((T_SL) reduce 466) ((T_SPACESHIP) reduce 466) ((T_SR) reduce 466) (("^") reduce 466) (("{") reduce 466) (("|") reduce 466)) 29032 ((("%") reduce 488) ((")") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_PIPE) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("^") reduce 488) (("{") reduce 488) (("|") reduce 488)) 29033 ((("%") reduce 531) (("(") shift) ((")") reduce 531) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 29034 ((("%") reduce 461) (("(") reduce 558) ((")") reduce 461) (("*") reduce 461) (("+") reduce 461) ((",") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PIPE) reduce 461) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") reduce 552) (("^") reduce 461) (("{") reduce 461) (("|") reduce 461)) 29035 ((("%") reduce 462) ((")") reduce 462) (("*") reduce 462) (("+") reduce 462) ((",") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_PIPE) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("{") reduce 462) (("|") reduce 462)) 29036 ((("%") reduce 476) ((")") reduce 476) (("*") reduce 476) (("+") reduce 476) ((",") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_PIPE) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("{") reduce 476) (("|") reduce 476)) 29037 ((("%") reduce 559) (("(") reduce 559) ((")") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) (("=") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AND_EQUAL) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_COALESCE_EQUAL) reduce 559) ((T_CONCAT_EQUAL) reduce 559) ((T_DEC) reduce 559) ((T_DIV_EQUAL) reduce 559) ((T_INC) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_MINUS_EQUAL) reduce 559) ((T_MOD_EQUAL) reduce 559) ((T_MUL_EQUAL) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_OR_EQUAL) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_PLUS_EQUAL) reduce 559) ((T_POW) reduce 559) ((T_POW_EQUAL) reduce 559) ((T_SL) reduce 559) ((T_SL_EQUAL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) ((T_SR_EQUAL) reduce 559) ((T_XOR_EQUAL) reduce 559) (("[") reduce 559) (("^") reduce 559) (("{") reduce 559) (("|") reduce 559)) 29038 ((("%") reduce 565) ((")") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("{") reduce 565) (("|") reduce 565)) 29039 ((("%") reduce 403) ((")") reduce 403) (("*") reduce 403) (("+") reduce 403) ((",") reduce 403) (("-") reduce 403) ((".") reduce 403) (("/") reduce 403) (("<") reduce 403) (("=") shift) ((">") reduce 403) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_INC) shift) ((T_INSTANCEOF) reduce 403) ((T_IS_EQUAL) reduce 403) ((T_IS_GREATER_OR_EQUAL) reduce 403) ((T_IS_IDENTICAL) reduce 403) ((T_IS_NOT_EQUAL) reduce 403) ((T_IS_NOT_IDENTICAL) reduce 403) ((T_IS_SMALLER_OR_EQUAL) reduce 403) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 403) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 403) ((T_POW_EQUAL) shift) ((T_SL) reduce 403) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 403) ((T_SR) reduce 403) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 548) (("^") reduce 403) (("{") reduce 403) (("|") reduce 403)) 29042 ((("%") reduce 448) ((")") reduce 448) (("*") reduce 448) (("+") reduce 448) ((",") reduce 448) (("-") reduce 448) ((".") reduce 448) (("/") reduce 448) (("<") reduce 448) ((">") reduce 448) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_BOOLEAN_AND) reduce 448) ((T_BOOLEAN_OR) reduce 448) ((T_COALESCE) reduce 448) ((T_INSTANCEOF) reduce 448) ((T_IS_EQUAL) reduce 448) ((T_IS_GREATER_OR_EQUAL) reduce 448) ((T_IS_IDENTICAL) reduce 448) ((T_IS_NOT_EQUAL) reduce 448) ((T_IS_NOT_IDENTICAL) reduce 448) ((T_IS_SMALLER_OR_EQUAL) reduce 448) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_PIPE) reduce 448) ((T_POW) shift) ((T_SL) reduce 448) ((T_SPACESHIP) reduce 448) ((T_SR) reduce 448) (("^") reduce 448) (("{") reduce 448) (("|") reduce 448)) 29074 ((("%") shift) ((")") reduce 432) (("*") shift) (("+") shift) ((",") reduce 432) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 432) (("|") reduce 432)) 29075 ((("%") shift) ((")") reduce 435) (("*") shift) (("+") shift) ((",") reduce 435) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("{") reduce 435) (("|") reduce 435)) 29076 ((("%") shift) ((")") reduce 444) (("*") shift) (("+") shift) ((",") reduce 444) (("-") shift) ((".") reduce 444) (("/") shift) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_PIPE) reduce 444) ((T_POW) shift) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("{") reduce 444) (("|") reduce 444)) 29077 ((("%") shift) ((")") reduce 458) (("*") shift) (("+") shift) ((",") reduce 458) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 458) (("{") reduce 458) (("|") reduce 458)) 29078 ((("%") shift) ((")") reduce 443) (("*") shift) (("+") shift) ((",") reduce 443) (("-") shift) ((".") reduce 443) (("/") shift) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_PIPE) reduce 443) ((T_POW) shift) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("^") reduce 443) (("{") reduce 443) (("|") reduce 443)) 29079 ((("%") reduce 440) ((")") reduce 440) (("*") reduce 440) (("+") reduce 440) ((",") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_PIPE) reduce 440) ((T_POW) shift) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("{") reduce 440) (("|") reduce 440)) 29080 ((("%") shift) ((")") reduce 453) (("*") shift) (("+") shift) ((",") reduce 453) (("-") shift) ((".") shift) (("/") shift) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_PIPE) reduce 453) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 453) ((T_SR) shift) (("^") reduce 453) (("{") reduce 453) (("|") reduce 453)) 29081 ((("%") shift) ((")") reduce 431) (("*") shift) (("+") shift) ((",") reduce 431) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 431) (("|") shift)) 29082 ((("%") shift) ((")") reduce 429) (("*") shift) (("+") shift) ((",") reduce 429) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 429) (("|") shift)) 29083 ((("%") shift) ((")") reduce 430) (("*") shift) (("+") shift) ((",") reduce 430) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 430) (("|") shift)) 29084 ((("%") shift) ((")") reduce 455) (("*") shift) (("+") shift) ((",") reduce 455) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 455) ((T_SR) shift) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 29085 ((("%") shift) ((")") reduce 450) (("*") shift) (("+") shift) ((",") reduce 450) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 450) (("{") reduce 450) (("|") reduce 450)) 29086 ((("%") shift) ((")") reduce 452) (("*") shift) (("+") shift) ((",") reduce 452) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 452) (("{") reduce 452) (("|") reduce 452)) 29087 ((("%") shift) ((")") reduce 449) (("*") shift) (("+") shift) ((",") reduce 449) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 449) (("{") reduce 449) (("|") reduce 449)) 29088 ((("%") shift) ((")") reduce 457) (("*") shift) (("+") shift) ((",") reduce 457) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 457) ((T_SR) shift) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 29089 ((("%") shift) ((")") reduce 451) (("*") shift) (("+") shift) ((",") reduce 451) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 451) (("{") reduce 451) (("|") reduce 451)) 29092 ((("%") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 29093 ((("%") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 29094 ((("%") reduce 95) ((")") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("{") reduce 95) (("|") reduce 95)) 29095 ((("%") reduce 509) ((")") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 29096 ((("%") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 29097 ((("%") reduce 568) ((")") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("{") reduce 568) (("|") reduce 568)) 29098 ((("%") reduce 511) ((")") reduce 511) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 29099 ((("%") reduce 459) ((")") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PIPE) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("{") reduce 459) (("|") reduce 459)) 29100 ((("%") reduce 510) ((")") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 29101 ((("%") reduce 512) ((")") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 29102 ((("%") reduce 573) ((")") reduce 573) (("*") reduce 573) (("+") reduce 573) ((",") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("{") reduce 573) (("|") reduce 573)) 29108 ((("%") reduce 574) ((")") reduce 574) (("*") reduce 574) (("+") reduce 574) ((",") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("{") reduce 574) (("|") reduce 574)) 29109 ((("%") reduce 578) ((")") reduce 578) (("*") reduce 578) (("+") reduce 578) ((",") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("{") reduce 578) (("|") reduce 578)) 29110 ((("%") reduce 582) ((")") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("{") reduce 582) (("|") reduce 582)) 29111 ((("%") reduce 575) ((")") reduce 575) (("*") reduce 575) (("+") reduce 575) ((",") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("{") reduce 575) (("|") reduce 575)) 29112 ((("%") reduce 584) ((")") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("{") reduce 584) (("|") reduce 584)) 29115 ((("%") reduce 583) ((")") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("{") reduce 583) (("|") reduce 583)) 29116 ((("%") reduce 576) ((")") reduce 576) (("*") reduce 576) (("+") reduce 576) ((",") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("{") reduce 576) (("|") reduce 576)) 29118 ((("%") reduce 577) ((")") reduce 577) (("*") reduce 577) (("+") reduce 577) ((",") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("{") reduce 577) (("|") reduce 577)) 29120 ((("%") reduce 513) ((")") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 29121 ((("%") reduce 570) ((")") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("{") reduce 570) (("|") reduce 570)) 29124 ((("%") reduce 569) ((")") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("{") reduce 569) (("|") reduce 569)) 29125 ((("%") shift) ((")") reduce 465) (("*") shift) (("+") shift) ((",") reduce 465) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 465) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 465) (("|") shift)) 29126 ((("%") shift) ((")") reduce 427) (("*") shift) (("+") shift) ((",") reduce 427) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 427) (("|") shift)) 29127 ((("%") shift) ((")") reduce 428) (("*") shift) (("+") shift) ((",") reduce 428) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 428) (("|") shift)) 29128 ((("%") shift) ((")") reduce 433) (("*") shift) (("+") shift) ((",") reduce 433) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("{") reduce 433) (("|") reduce 433)) 29129 ((("%") shift) ((")") reduce 434) (("*") shift) (("+") shift) ((",") reduce 434) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("{") reduce 434) (("|") reduce 434)) 29133 ((("%") shift) ((")") reduce 463) (("*") shift) (("+") shift) ((",") reduce 463) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 463) (("|") shift)) 29134 ((("%") shift) ((")") reduce 464) (("*") shift) (("+") shift) ((",") reduce 464) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 464) (("|") shift)) 29135 ((("%") shift) ((")") reduce 456) (("*") shift) (("+") shift) ((",") reduce 456) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 456) ((T_SR) shift) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 29136 ((("%") shift) ((")") reduce 454) (("*") shift) (("+") shift) ((",") reduce 454) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 454) ((T_SR) shift) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 29137 ((("%") reduce 441) ((")") reduce 441) (("*") reduce 441) (("+") reduce 441) ((",") reduce 441) (("-") reduce 441) ((".") reduce 441) (("/") reduce 441) (("<") reduce 441) ((">") reduce 441) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 441) ((T_IS_GREATER_OR_EQUAL) reduce 441) ((T_IS_IDENTICAL) reduce 441) ((T_IS_NOT_EQUAL) reduce 441) ((T_IS_NOT_IDENTICAL) reduce 441) ((T_IS_SMALLER_OR_EQUAL) reduce 441) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_PIPE) reduce 441) ((T_POW) shift) ((T_SL) reduce 441) ((T_SPACESHIP) reduce 441) ((T_SR) reduce 441) (("^") reduce 441) (("{") reduce 441) (("|") reduce 441)) 29138 ((("%") shift) ((")") reduce 436) (("*") shift) (("+") shift) ((",") reduce 436) (("-") shift) ((".") reduce 436) (("/") shift) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_PIPE) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 436) ((T_SR) shift) (("^") reduce 436) (("{") reduce 436) (("|") reduce 436)) 29139 ((("%") shift) ((")") reduce 438) (("*") shift) (("+") reduce 438) ((",") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") shift) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_PIPE) reduce 438) ((T_POW) shift) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("{") reduce 438) (("|") reduce 438)) 29140 ((("%") shift) ((")") reduce 437) (("*") shift) (("+") reduce 437) ((",") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") shift) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_PIPE) reduce 437) ((T_POW) shift) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("{") reduce 437) (("|") reduce 437)) 29141 ((("%") reduce 439) ((")") reduce 439) (("*") reduce 439) (("+") reduce 439) ((",") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_PIPE) reduce 439) ((T_POW) shift) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("{") reduce 439) (("|") reduce 439)) 29142 ((("%") reduce 442) ((")") reduce 442) (("*") reduce 442) (("+") reduce 442) ((",") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_PIPE) reduce 442) ((T_POW) shift) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("{") reduce 442) (("|") reduce 442)) 29144 ((("%") reduce 74) (("(") reduce 74) ((")") reduce 74) (("*") reduce 74) (("+") reduce 74) ((",") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_PIPE) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("{") reduce 74) (("|") reduce 74)) 29145 ((("%") reduce 52) (("(") reduce 52) ((")") reduce 52) (("*") reduce 52) (("+") reduce 52) ((",") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_PIPE) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("{") reduce 52) (("|") reduce 52)) 29146 ((("%") reduce 27) (("(") reduce 27) ((")") reduce 27) (("*") reduce 27) (("+") reduce 27) ((",") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_PIPE) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("{") reduce 27) (("|") reduce 27)) 29147 ((("%") reduce 51) (("(") reduce 51) ((")") reduce 51) (("*") reduce 51) (("+") reduce 51) ((",") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_PIPE) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("{") reduce 51) (("|") reduce 51)) 29148 ((("%") reduce 53) (("(") reduce 53) ((")") reduce 53) (("*") reduce 53) (("+") reduce 53) ((",") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_PIPE) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("{") reduce 53) (("|") reduce 53)) 29149 ((("%") reduce 49) (("(") reduce 49) ((")") reduce 49) (("*") reduce 49) (("+") reduce 49) ((",") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_PIPE) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("{") reduce 49) (("|") reduce 49)) 29150 ((("%") reduce 29) (("(") reduce 29) ((")") reduce 29) (("*") reduce 29) (("+") reduce 29) ((",") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_PIPE) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("{") reduce 29) (("|") reduce 29)) 29151 ((("%") reduce 59) (("(") reduce 59) ((")") reduce 59) (("*") reduce 59) (("+") reduce 59) ((",") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_PIPE) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("{") reduce 59) (("|") reduce 59)) 29152 ((("%") reduce 60) (("(") reduce 60) ((")") reduce 60) (("*") reduce 60) (("+") reduce 60) ((",") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_PIPE) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("{") reduce 60) (("|") reduce 60)) 29153 ((("%") reduce 11) (("(") reduce 11) ((")") reduce 11) (("*") reduce 11) (("+") reduce 11) ((",") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_PIPE) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("{") reduce 11) (("|") reduce 11)) 29154 ((("%") reduce 42) (("(") reduce 42) ((")") reduce 42) (("*") reduce 42) (("+") reduce 42) ((",") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_PIPE) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("{") reduce 42) (("|") reduce 42)) 29155 ((("%") reduce 39) (("(") reduce 39) ((")") reduce 39) (("*") reduce 39) (("+") reduce 39) ((",") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_PIPE) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("{") reduce 39) (("|") reduce 39)) 29156 ((("%") reduce 25) (("(") reduce 25) ((")") reduce 25) (("*") reduce 25) (("+") reduce 25) ((",") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_PIPE) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("{") reduce 25) (("|") reduce 25)) 29157 ((("%") reduce 50) (("(") reduce 50) ((")") reduce 50) (("*") reduce 50) (("+") reduce 50) ((",") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_PIPE) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("{") reduce 50) (("|") reduce 50)) 29158 ((("%") reduce 66) (("(") reduce 66) ((")") reduce 66) (("*") reduce 66) (("+") reduce 66) ((",") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_PIPE) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("{") reduce 66) (("|") reduce 66)) 29159 ((("%") reduce 18) (("(") reduce 18) ((")") reduce 18) (("*") reduce 18) (("+") reduce 18) ((",") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_PIPE) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("{") reduce 18) (("|") reduce 18)) 29160 ((("%") reduce 17) (("(") reduce 17) ((")") reduce 17) (("*") reduce 17) (("+") reduce 17) ((",") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_PIPE) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("{") reduce 17) (("|") reduce 17)) 29161 ((("%") reduce 15) (("(") reduce 15) ((")") reduce 15) (("*") reduce 15) (("+") reduce 15) ((",") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_PIPE) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("{") reduce 15) (("|") reduce 15)) 29162 ((("%") reduce 14) (("(") reduce 14) ((")") reduce 14) (("*") reduce 14) (("+") reduce 14) ((",") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_PIPE) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("{") reduce 14) (("|") reduce 14)) 29163 ((("%") reduce 38) (("(") reduce 38) ((")") reduce 38) (("*") reduce 38) (("+") reduce 38) ((",") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_PIPE) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("{") reduce 38) (("|") reduce 38)) 29164 ((("%") reduce 26) (("(") reduce 26) ((")") reduce 26) (("*") reduce 26) (("+") reduce 26) ((",") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_PIPE) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("{") reduce 26) (("|") reduce 26)) 29165 ((("%") reduce 22) (("(") reduce 22) ((")") reduce 22) (("*") reduce 22) (("+") reduce 22) ((",") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_PIPE) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("{") reduce 22) (("|") reduce 22)) 29166 ((("%") reduce 24) (("(") reduce 24) ((")") reduce 24) (("*") reduce 24) (("+") reduce 24) ((",") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_PIPE) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("{") reduce 24) (("|") reduce 24)) 29167 ((("%") reduce 16) (("(") reduce 16) ((")") reduce 16) (("*") reduce 16) (("+") reduce 16) ((",") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_PIPE) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("{") reduce 16) (("|") reduce 16)) 29168 ((("%") reduce 48) (("(") reduce 48) ((")") reduce 48) (("*") reduce 48) (("+") reduce 48) ((",") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_PIPE) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("{") reduce 48) (("|") reduce 48)) 29169 ((("%") reduce 20) (("(") reduce 20) ((")") reduce 20) (("*") reduce 20) (("+") reduce 20) ((",") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_PIPE) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("{") reduce 20) (("|") reduce 20)) 29170 ((("%") reduce 70) (("(") reduce 70) ((")") reduce 70) (("*") reduce 70) (("+") reduce 70) ((",") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_PIPE) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("{") reduce 70) (("|") reduce 70)) 29171 ((("%") reduce 3) (("(") reduce 3) ((")") reduce 3) (("*") reduce 3) (("+") reduce 3) ((",") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_PIPE) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("{") reduce 3) (("|") reduce 3)) 29172 ((("%") reduce 12) (("(") reduce 12) ((")") reduce 12) (("*") reduce 12) (("+") reduce 12) ((",") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_PIPE) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("{") reduce 12) (("|") reduce 12)) 29173 ((("%") reduce 54) (("(") reduce 54) ((")") reduce 54) (("*") reduce 54) (("+") reduce 54) ((",") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_PIPE) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("{") reduce 54) (("|") reduce 54)) 29174 ((("%") reduce 65) (("(") reduce 65) ((")") reduce 65) (("*") reduce 65) (("+") reduce 65) ((",") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_PIPE) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("{") reduce 65) (("|") reduce 65)) 29175 ((("%") reduce 75) (("(") reduce 75) ((")") reduce 75) (("*") reduce 75) (("+") reduce 75) ((",") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_PIPE) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("{") reduce 75) (("|") reduce 75)) 29176 ((("%") reduce 30) (("(") reduce 30) ((")") reduce 30) (("*") reduce 30) (("+") reduce 30) ((",") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_PIPE) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("{") reduce 30) (("|") reduce 30)) 29177 ((("%") reduce 68) (("(") reduce 68) ((")") reduce 68) (("*") reduce 68) (("+") reduce 68) ((",") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_PIPE) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("{") reduce 68) (("|") reduce 68)) 29178 ((("%") reduce 21) (("(") reduce 21) ((")") reduce 21) (("*") reduce 21) (("+") reduce 21) ((",") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_PIPE) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("{") reduce 21) (("|") reduce 21)) 29179 ((("%") reduce 23) (("(") reduce 23) ((")") reduce 23) (("*") reduce 23) (("+") reduce 23) ((",") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_PIPE) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("{") reduce 23) (("|") reduce 23)) 29180 ((("%") reduce 41) (("(") reduce 41) ((")") reduce 41) (("*") reduce 41) (("+") reduce 41) ((",") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_PIPE) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("{") reduce 41) (("|") reduce 41)) 29181 ((("%") reduce 62) (("(") reduce 62) ((")") reduce 62) (("*") reduce 62) (("+") reduce 62) ((",") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_PIPE) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("{") reduce 62) (("|") reduce 62)) 29182 ((("%") reduce 34) (("(") reduce 34) ((")") reduce 34) (("*") reduce 34) (("+") reduce 34) ((",") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_PIPE) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("{") reduce 34) (("|") reduce 34)) 29183 ((("%") reduce 40) (("(") reduce 40) ((")") reduce 40) (("*") reduce 40) (("+") reduce 40) ((",") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_PIPE) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("{") reduce 40) (("|") reduce 40)) 29184 ((("%") reduce 13) (("(") reduce 13) ((")") reduce 13) (("*") reduce 13) (("+") reduce 13) ((",") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_PIPE) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("{") reduce 13) (("|") reduce 13)) 29185 ((("%") reduce 55) (("(") reduce 55) ((")") reduce 55) (("*") reduce 55) (("+") reduce 55) ((",") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_PIPE) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("{") reduce 55) (("|") reduce 55)) 29186 ((("%") reduce 1) (("(") reduce 1) ((")") reduce 1) (("*") reduce 1) (("+") reduce 1) ((",") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_PIPE) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("{") reduce 1) (("|") reduce 1)) 29187 ((("%") reduce 2) (("(") reduce 2) ((")") reduce 2) (("*") reduce 2) (("+") reduce 2) ((",") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_PIPE) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("{") reduce 2) (("|") reduce 2)) 29188 ((("%") reduce 9) (("(") reduce 9) ((")") reduce 9) (("*") reduce 9) (("+") reduce 9) ((",") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_PIPE) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("{") reduce 9) (("|") reduce 9)) 29189 ((("%") reduce 33) (("(") reduce 33) ((")") reduce 33) (("*") reduce 33) (("+") reduce 33) ((",") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_PIPE) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("{") reduce 33) (("|") reduce 33)) 29190 ((("%") reduce 58) (("(") reduce 58) ((")") reduce 58) (("*") reduce 58) (("+") reduce 58) ((",") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_PIPE) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("{") reduce 58) (("|") reduce 58)) 29191 ((("%") reduce 37) (("(") reduce 37) ((")") reduce 37) (("*") reduce 37) (("+") reduce 37) ((",") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_PIPE) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("{") reduce 37) (("|") reduce 37)) 29192 ((("%") reduce 64) (("(") reduce 64) ((")") reduce 64) (("*") reduce 64) (("+") reduce 64) ((",") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_PIPE) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("{") reduce 64) (("|") reduce 64)) 29193 ((("%") reduce 46) (("(") reduce 46) ((")") reduce 46) (("*") reduce 46) (("+") reduce 46) ((",") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_PIPE) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("{") reduce 46) (("|") reduce 46)) 29194 ((("%") reduce 8) (("(") reduce 8) ((")") reduce 8) (("*") reduce 8) (("+") reduce 8) ((",") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_PIPE) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("{") reduce 8) (("|") reduce 8)) 29195 ((("%") reduce 6) (("(") reduce 6) ((")") reduce 6) (("*") reduce 6) (("+") reduce 6) ((",") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_PIPE) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("{") reduce 6) (("|") reduce 6)) 29196 ((("%") reduce 7) (("(") reduce 7) ((")") reduce 7) (("*") reduce 7) (("+") reduce 7) ((",") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_PIPE) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("{") reduce 7) (("|") reduce 7)) 29197 ((("%") reduce 69) (("(") reduce 69) ((")") reduce 69) (("*") reduce 69) (("+") reduce 69) ((",") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_PIPE) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("{") reduce 69) (("|") reduce 69)) 29198 ((("%") reduce 63) (("(") reduce 63) ((")") reduce 63) (("*") reduce 63) (("+") reduce 63) ((",") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_PIPE) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("{") reduce 63) (("|") reduce 63)) 29199 ((("%") reduce 56) (("(") reduce 56) ((")") reduce 56) (("*") reduce 56) (("+") reduce 56) ((",") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_PIPE) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("{") reduce 56) (("|") reduce 56)) 29200 ((("%") reduce 10) (("(") reduce 10) ((")") reduce 10) (("*") reduce 10) (("+") reduce 10) ((",") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_PIPE) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("{") reduce 10) (("|") reduce 10)) 29201 ((("%") reduce 67) (("(") reduce 67) ((")") reduce 67) (("*") reduce 67) (("+") reduce 67) ((",") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_PIPE) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("{") reduce 67) (("|") reduce 67)) 29202 ((("%") reduce 44) (("(") reduce 44) ((")") reduce 44) (("*") reduce 44) (("+") reduce 44) ((",") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_PIPE) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("{") reduce 44) (("|") reduce 44)) 29203 ((("%") reduce 76) (("(") reduce 76) ((")") reduce 76) (("*") reduce 76) (("+") reduce 76) ((",") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_PIPE) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("{") reduce 76) (("|") reduce 76)) 29204 ((("%") reduce 71) (("(") reduce 71) ((")") reduce 71) (("*") reduce 71) (("+") reduce 71) ((",") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_PIPE) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("{") reduce 71) (("|") reduce 71)) 29205 ((("%") reduce 77) (("(") reduce 77) ((")") reduce 77) (("*") reduce 77) (("+") reduce 77) ((",") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_PIPE) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("{") reduce 77) (("|") reduce 77)) 29206 ((("%") reduce 78) (("(") reduce 78) ((")") reduce 78) (("*") reduce 78) (("+") reduce 78) ((",") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_PIPE) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("{") reduce 78) (("|") reduce 78)) 29207 ((("%") reduce 79) (("(") reduce 79) ((")") reduce 79) (("*") reduce 79) (("+") reduce 79) ((",") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_PIPE) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("{") reduce 79) (("|") reduce 79)) 29208 ((("%") reduce 4) (("(") reduce 4) ((")") reduce 4) (("*") reduce 4) (("+") reduce 4) ((",") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_PIPE) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("{") reduce 4) (("|") reduce 4)) 29209 ((("%") reduce 5) (("(") reduce 5) ((")") reduce 5) (("*") reduce 5) (("+") reduce 5) ((",") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_PIPE) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("{") reduce 5) (("|") reduce 5)) 29210 ((("%") reduce 43) (("(") reduce 43) ((")") reduce 43) (("*") reduce 43) (("+") reduce 43) ((",") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_PIPE) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("{") reduce 43) (("|") reduce 43)) 29211 ((("%") reduce 73) (("(") reduce 73) ((")") reduce 73) (("*") reduce 73) (("+") reduce 73) ((",") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_PIPE) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("{") reduce 73) (("|") reduce 73)) 29212 ((("%") reduce 82) (("(") reduce 82) ((")") reduce 82) (("*") reduce 82) (("+") reduce 82) ((",") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_PIPE) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("{") reduce 82) (("|") reduce 82)) 29213 ((("%") reduce 47) (("(") reduce 47) ((")") reduce 47) (("*") reduce 47) (("+") reduce 47) ((",") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_PIPE) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("{") reduce 47) (("|") reduce 47)) 29214 ((("%") reduce 31) (("(") reduce 31) ((")") reduce 31) (("*") reduce 31) (("+") reduce 31) ((",") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_PIPE) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("{") reduce 31) (("|") reduce 31)) 29215 ((("%") reduce 57) (("(") reduce 57) ((")") reduce 57) (("*") reduce 57) (("+") reduce 57) ((",") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_PIPE) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("{") reduce 57) (("|") reduce 57)) 29216 ((("%") reduce 61) (("(") reduce 61) ((")") reduce 61) (("*") reduce 61) (("+") reduce 61) ((",") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_PIPE) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("{") reduce 61) (("|") reduce 61)) 29217 ((("%") reduce 28) (("(") reduce 28) ((")") reduce 28) (("*") reduce 28) (("+") reduce 28) ((",") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_PIPE) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("{") reduce 28) (("|") reduce 28)) 29218 ((("%") reduce 36) (("(") reduce 36) ((")") reduce 36) (("*") reduce 36) (("+") reduce 36) ((",") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_PIPE) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("{") reduce 36) (("|") reduce 36)) 29219 ((("%") reduce 32) (("(") reduce 32) ((")") reduce 32) (("*") reduce 32) (("+") reduce 32) ((",") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_PIPE) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("{") reduce 32) (("|") reduce 32)) 29220 ((("%") reduce 35) (("(") reduce 35) ((")") reduce 35) (("*") reduce 35) (("+") reduce 35) ((",") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_PIPE) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("{") reduce 35) (("|") reduce 35)) 29221 ((("%") reduce 19) (("(") reduce 19) ((")") reduce 19) (("*") reduce 19) (("+") reduce 19) ((",") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_PIPE) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("{") reduce 19) (("|") reduce 19)) 29222 ((("%") reduce 45) (("(") reduce 45) ((")") reduce 45) (("*") reduce 45) (("+") reduce 45) ((",") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_PIPE) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("{") reduce 45) (("|") reduce 45)) 29223 ((("%") reduce 542) (("(") reduce 579) ((")") reduce 542) (("*") reduce 542) (("+") reduce 542) ((",") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PIPE) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("{") reduce 542) (("|") reduce 542)) 29225 ((("%") reduce 72) (("(") reduce 72) ((")") reduce 72) (("*") reduce 72) (("+") reduce 72) ((",") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_PIPE) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("{") reduce 72) (("|") reduce 72)) 29226 ((("%") reduce 83) (("(") reduce 83) ((")") reduce 83) (("*") reduce 83) (("+") reduce 83) ((",") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_PIPE) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("{") reduce 83) (("|") reduce 83)) 29227 ((("%") reduce 572) (("(") reduce 581) ((")") reduce 572) (("*") reduce 572) (("+") reduce 572) ((",") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) (("<") reduce 572) (("=") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AND_EQUAL) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_COALESCE_EQUAL) reduce 572) ((T_CONCAT_EQUAL) reduce 572) ((T_DEC) reduce 572) ((T_DIV_EQUAL) reduce 572) ((T_INC) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_MINUS_EQUAL) reduce 572) ((T_MOD_EQUAL) reduce 572) ((T_MUL_EQUAL) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_OR_EQUAL) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_PLUS_EQUAL) reduce 572) ((T_POW) reduce 572) ((T_POW_EQUAL) reduce 572) ((T_SL) reduce 572) ((T_SL_EQUAL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) ((T_SR_EQUAL) reduce 572) ((T_XOR_EQUAL) reduce 572) (("[") reduce 572) (("^") reduce 572) (("{") reduce 572) (("|") reduce 572)) 29230 ((("%") reduce 544) (("(") reduce 580) ((")") reduce 544) (("*") reduce 544) (("+") reduce 544) ((",") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PIPE) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("{") reduce 544) (("|") reduce 544)) 29232 ((("%") reduce 507) (("(") reduce 507) ((")") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 29233 ((("%") reduce 291) (("(") reduce 291) ((")") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("{") reduce 291) (("|") reduce 291)) 29237 ((("%") reduce 292) (("(") reduce 292) ((")") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) (("=") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AND_EQUAL) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_COALESCE_EQUAL) reduce 292) ((T_CONCAT_EQUAL) reduce 292) ((T_DEC) reduce 292) ((T_DIV_EQUAL) reduce 292) ((T_INC) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_MINUS_EQUAL) reduce 292) ((T_MOD_EQUAL) reduce 292) ((T_MUL_EQUAL) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_OR_EQUAL) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_PLUS_EQUAL) reduce 292) ((T_POW) reduce 292) ((T_POW_EQUAL) reduce 292) ((T_SL) reduce 292) ((T_SL_EQUAL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) ((T_SR_EQUAL) reduce 292) ((T_XOR_EQUAL) reduce 292) (("[") reduce 292) (("^") reduce 292) (("{") reduce 292) (("|") reduce 292)) 29238 ((("%") reduce 293) (("(") reduce 293) ((")") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) (("=") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AND_EQUAL) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_COALESCE_EQUAL) reduce 293) ((T_CONCAT_EQUAL) reduce 293) ((T_DEC) reduce 293) ((T_DIV_EQUAL) reduce 293) ((T_INC) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_MINUS_EQUAL) reduce 293) ((T_MOD_EQUAL) reduce 293) ((T_MUL_EQUAL) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_OR_EQUAL) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_PLUS_EQUAL) reduce 293) ((T_POW) reduce 293) ((T_POW_EQUAL) reduce 293) ((T_SL) reduce 293) ((T_SL_EQUAL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) ((T_SR_EQUAL) reduce 293) ((T_XOR_EQUAL) reduce 293) (("[") reduce 293) (("^") reduce 293) (("{") reduce 293) (("|") reduce 293)) 29243 ((("%") reduce 425) ((")") reduce 425) (("*") reduce 425) (("+") reduce 425) ((",") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_PIPE) reduce 425) ((T_POW) reduce 425) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("{") reduce 425) (("|") reduce 425)) 29245 ((("%") reduce 423) ((")") reduce 423) (("*") reduce 423) (("+") reduce 423) ((",") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_PIPE) reduce 423) ((T_POW) reduce 423) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("{") reduce 423) (("|") reduce 423)) 29255 ((("%") shift) ((")") reduce 419) (("*") shift) (("+") shift) ((",") reduce 419) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 419) (("|") shift)) 29256 ((("%") shift) ((")") reduce 421) (("*") shift) (("+") shift) ((",") reduce 421) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 421) (("|") shift)) 29257 ((("%") shift) ((")") reduce 420) (("*") shift) (("+") shift) ((",") reduce 420) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 420) (("|") shift)) 29258 ((("%") shift) ((")") reduce 413) (("*") shift) (("+") shift) ((",") reduce 413) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 413) (("|") shift)) 29259 ((("%") shift) ((")") reduce 410) (("*") shift) (("+") shift) ((",") reduce 410) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 410) (("|") shift)) 29260 ((("%") shift) ((")") reduce 418) (("*") shift) (("+") shift) ((",") reduce 418) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 418) (("|") shift)) 29261 ((("%") shift) ((")") reduce 412) (("*") shift) (("+") shift) ((",") reduce 412) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 412) (("|") shift)) 29262 ((("%") shift) ((")") reduce 416) (("*") shift) (("+") shift) ((",") reduce 416) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 416) (("|") shift)) 29263 ((("%") shift) ((")") reduce 411) (("*") shift) (("+") shift) ((",") reduce 411) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 411) (("|") shift)) 29264 ((("%") shift) ((")") reduce 414) (("*") shift) (("+") shift) ((",") reduce 414) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 414) (("|") shift)) 29265 ((("%") shift) ((")") reduce 415) (("*") shift) (("+") shift) ((",") reduce 415) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 415) (("|") shift)) 29266 ((("%") shift) ((")") reduce 422) (("*") shift) (("+") shift) ((",") reduce 422) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 422) (("|") shift)) 29267 ((("%") shift) ((")") reduce 417) (("*") shift) (("+") shift) ((",") reduce 417) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 417) (("|") shift)) 29269 ((("%") shift) ((")") reduce 406) (("*") shift) (("+") shift) ((",") reduce 406) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 406) (("|") shift)) 29272 ((("%") reduce 568) (("(") reduce 568) ((")") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("{") reduce 568) (("|") reduce 568)) 29275 ((("%") reduce 564) (("(") reduce 555) ((")") reduce 564) (("*") reduce 564) (("+") reduce 564) ((",") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) (("<") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_POW) reduce 564) ((T_SL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) (("[") reduce 564) (("^") reduce 564) (("{") reduce 564) (("|") reduce 564)) 29277 ((("%") reduce 563) (("(") reduce 563) ((")") reduce 563) (("*") reduce 563) (("+") reduce 563) ((",") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) (("<") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_POW) reduce 563) ((T_SL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) (("[") reduce 563) (("^") reduce 563) (("{") reduce 563) (("|") reduce 563)) 29279 ((("%") reduce 559) (("(") reduce 559) ((")") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("{") reduce 559) (("|") reduce 559)) 29280 ((("%") reduce 565) ((")") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("{") reduce 565) (("|") reduce 565)) 29281 ((("%") reduce 407) ((")") reduce 407) (("*") reduce 407) (("+") reduce 407) ((",") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("[") reduce 548) (("^") reduce 407) (("{") reduce 407) (("|") reduce 407)) 29285 ((("%") reduce 572) (("(") reduce 581) ((")") reduce 572) (("*") reduce 572) (("+") reduce 572) ((",") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) (("<") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_POW) reduce 572) ((T_SL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) (("[") reduce 572) (("^") reduce 572) (("{") reduce 572) (("|") reduce 572)) 29287 ((("%") reduce 507) (("(") reduce 507) ((")") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 29288 ((("%") reduce 291) (("(") reduce 291) ((")") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("{") reduce 291) (("|") reduce 291)) 29292 ((("%") reduce 292) (("(") reduce 292) ((")") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("[") reduce 292) (("^") reduce 292) (("{") reduce 292) (("|") reduce 292)) 29293 ((("%") reduce 293) (("(") reduce 293) ((")") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("[") reduce 293) (("^") reduce 293) (("{") reduce 293) (("|") reduce 293)) 29294 ((("%") reduce 504) (("(") reduce 504) ((")") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("{") reduce 504) (("|") reduce 504)) 29297 ((("%") reduce 571) (("(") reduce 581) ((")") reduce 571) (("*") reduce 571) (("+") reduce 571) ((",") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) (("<") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_POW) reduce 571) ((T_SL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) (("[") reduce 571) (("^") reduce 571) (("{") reduce 571) (("|") reduce 571)) 29298 ((("%") reduce 506) (("(") reduce 506) ((")") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 29299 ((("%") reduce 508) (("(") reduce 508) ((")") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 29304 ((("%") reduce 560) (("(") reduce 560) ((")") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("{") reduce 560) (("|") reduce 560)) 29305 ((("%") reduce 582) (("(") reduce 582) ((")") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("{") reduce 582) (("|") reduce 582)) 29306 ((("%") reduce 566) (("(") shift) ((")") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("{") reduce 566) (("|") reduce 566)) 29307 ((("%") reduce 584) (("(") reduce 584) ((")") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("{") reduce 584) (("|") reduce 584)) 29310 ((("%") reduce 583) (("(") reduce 583) ((")") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("{") reduce 583) (("|") reduce 583)) 29311 ((("%") reduce 561) (("(") reduce 561) ((")") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("{") reduce 561) (("|") reduce 561)) 29312 ((("%") reduce 567) (("(") shift) ((")") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("{") reduce 567) (("|") reduce 567)) 29313 ((("%") reduce 562) (("(") reduce 562) ((")") reduce 562) (("*") reduce 562) (("+") reduce 562) ((",") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("[") reduce 562) (("^") reduce 562) (("{") reduce 562) (("|") reduce 562)) 29314 ((("%") reduce 505) (("(") reduce 505) ((")") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 29315 ((("%") reduce 570) (("(") reduce 570) ((")") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("{") reduce 570) (("|") reduce 570)) 29318 ((("%") reduce 569) (("(") reduce 569) ((")") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("{") reduce 569) (("|") reduce 569)) 29319 ((("%") reduce 504) (("(") reduce 504) ((")") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) (("=") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AND_EQUAL) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_COALESCE_EQUAL) reduce 504) ((T_CONCAT_EQUAL) reduce 504) ((T_DEC) reduce 504) ((T_DIV_EQUAL) reduce 504) ((T_INC) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_MINUS_EQUAL) reduce 504) ((T_MOD_EQUAL) reduce 504) ((T_MUL_EQUAL) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_OR_EQUAL) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_PLUS_EQUAL) reduce 504) ((T_POW) reduce 504) ((T_POW_EQUAL) reduce 504) ((T_SL) reduce 504) ((T_SL_EQUAL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) ((T_SR_EQUAL) reduce 504) ((T_XOR_EQUAL) reduce 504) (("[") reduce 504) (("^") reduce 504) (("{") reduce 504) (("|") reduce 504)) 29330 ((("%") reduce 494) ((")") reduce 494) (("*") reduce 494) (("+") reduce 494) ((",") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PIPE) reduce 494) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("{") reduce 494) (("|") reduce 494)) 29331 ((("%") reduce 489) ((")") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_PIPE) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("^") reduce 489) (("{") reduce 489) (("|") reduce 489)) 29341 ((("%") shift) ((")") reduce 494) (("*") shift) (("+") shift) ((",") reduce 494) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 494) (("|") shift)) 29342 ((("%") reduce 490) ((")") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_PIPE) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("^") reduce 490) (("{") reduce 490) (("|") reduce 490)) 29343 (((")") reduce 261) ((",") reduce 261)) 29345 ((("%") reduce 541) (("(") reduce 579) ((")") reduce 541) (("*") reduce 541) (("+") reduce 541) ((",") reduce 541) (("-") reduce 541) ((".") reduce 541) (("/") reduce 541) (("<") reduce 541) ((">") reduce 541) (("?") reduce 541) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_BOOLEAN_AND) reduce 541) ((T_BOOLEAN_OR) reduce 541) ((T_COALESCE) reduce 541) ((T_INSTANCEOF) reduce 541) ((T_IS_EQUAL) reduce 541) ((T_IS_GREATER_OR_EQUAL) reduce 541) ((T_IS_IDENTICAL) reduce 541) ((T_IS_NOT_EQUAL) reduce 541) ((T_IS_NOT_IDENTICAL) reduce 541) ((T_IS_SMALLER_OR_EQUAL) reduce 541) ((T_LOGICAL_AND) reduce 541) ((T_LOGICAL_OR) reduce 541) ((T_LOGICAL_XOR) reduce 541) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 541) ((T_OBJECT_OPERATOR) reduce 541) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 541) ((T_PIPE) reduce 541) ((T_POW) reduce 541) ((T_SL) reduce 541) ((T_SPACESHIP) reduce 541) ((T_SR) reduce 541) (("[") reduce 541) (("^") reduce 541) (("{") reduce 541) (("|") reduce 541)) 29347 ((("%") reduce 571) (("(") reduce 581) ((")") reduce 571) (("*") reduce 571) (("+") reduce 571) ((",") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) (("<") reduce 571) (("=") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AND_EQUAL) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_COALESCE_EQUAL) reduce 571) ((T_CONCAT_EQUAL) reduce 571) ((T_DEC) reduce 571) ((T_DIV_EQUAL) reduce 571) ((T_INC) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_MINUS_EQUAL) reduce 571) ((T_MOD_EQUAL) reduce 571) ((T_MUL_EQUAL) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_OR_EQUAL) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_PLUS_EQUAL) reduce 571) ((T_POW) reduce 571) ((T_POW_EQUAL) reduce 571) ((T_SL) reduce 571) ((T_SL_EQUAL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) ((T_SR_EQUAL) reduce 571) ((T_XOR_EQUAL) reduce 571) (("[") reduce 571) (("^") reduce 571) (("{") reduce 571) (("|") reduce 571)) 29350 ((("%") reduce 543) (("(") reduce 580) ((")") reduce 543) (("*") reduce 543) (("+") reduce 543) ((",") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PIPE) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("{") reduce 543) (("|") reduce 543)) 29351 ((("%") reduce 506) (("(") reduce 506) ((")") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 29352 ((("%") reduce 508) (("(") reduce 508) ((")") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 29354 ((("%") reduce 485) ((")") reduce 485) (("*") reduce 485) (("+") reduce 485) ((",") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_PIPE) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("^") reduce 485) (("{") reduce 485) (("|") reduce 485)) 29355 ((("%") reduce 487) ((")") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_PIPE) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("^") reduce 487) (("{") reduce 487) (("|") reduce 487)) 29360 ((("%") reduce 560) (("(") reduce 560) ((")") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) (("=") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AND_EQUAL) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_COALESCE_EQUAL) reduce 560) ((T_CONCAT_EQUAL) reduce 560) ((T_DEC) reduce 560) ((T_DIV_EQUAL) reduce 560) ((T_INC) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_MINUS_EQUAL) reduce 560) ((T_MOD_EQUAL) reduce 560) ((T_MUL_EQUAL) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_OR_EQUAL) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_PLUS_EQUAL) reduce 560) ((T_POW) reduce 560) ((T_POW_EQUAL) reduce 560) ((T_SL) reduce 560) ((T_SL_EQUAL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) ((T_SR_EQUAL) reduce 560) ((T_XOR_EQUAL) reduce 560) (("[") reduce 560) (("^") reduce 560) (("{") reduce 560) (("|") reduce 560)) 29361 ((("%") reduce 582) (("(") reduce 582) ((")") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) (("=") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AND_EQUAL) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_COALESCE_EQUAL) reduce 582) ((T_CONCAT_EQUAL) reduce 582) ((T_DEC) reduce 582) ((T_DIV_EQUAL) reduce 582) ((T_INC) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_MINUS_EQUAL) reduce 582) ((T_MOD_EQUAL) reduce 582) ((T_MUL_EQUAL) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_OR_EQUAL) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_PLUS_EQUAL) reduce 582) ((T_POW) reduce 582) ((T_POW_EQUAL) reduce 582) ((T_SL) reduce 582) ((T_SL_EQUAL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) ((T_SR_EQUAL) reduce 582) ((T_XOR_EQUAL) reduce 582) (("[") reduce 582) (("^") reduce 582) (("{") reduce 582) (("|") reduce 582)) 29362 ((("%") reduce 566) (("(") shift) ((")") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("{") reduce 566) (("|") reduce 566)) 29363 ((("%") reduce 584) (("(") reduce 584) ((")") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) (("=") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AND_EQUAL) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_COALESCE_EQUAL) reduce 584) ((T_CONCAT_EQUAL) reduce 584) ((T_DEC) reduce 584) ((T_DIV_EQUAL) reduce 584) ((T_INC) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_MINUS_EQUAL) reduce 584) ((T_MOD_EQUAL) reduce 584) ((T_MUL_EQUAL) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_OR_EQUAL) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_PLUS_EQUAL) reduce 584) ((T_POW) reduce 584) ((T_POW_EQUAL) reduce 584) ((T_SL) reduce 584) ((T_SL_EQUAL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) ((T_SR_EQUAL) reduce 584) ((T_XOR_EQUAL) reduce 584) (("[") reduce 584) (("^") reduce 584) (("{") reduce 584) (("|") reduce 584)) 29366 ((("%") reduce 583) (("(") reduce 583) ((")") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) (("=") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AND_EQUAL) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_COALESCE_EQUAL) reduce 583) ((T_CONCAT_EQUAL) reduce 583) ((T_DEC) reduce 583) ((T_DIV_EQUAL) reduce 583) ((T_INC) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_MINUS_EQUAL) reduce 583) ((T_MOD_EQUAL) reduce 583) ((T_MUL_EQUAL) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_OR_EQUAL) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_PLUS_EQUAL) reduce 583) ((T_POW) reduce 583) ((T_POW_EQUAL) reduce 583) ((T_SL) reduce 583) ((T_SL_EQUAL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) ((T_SR_EQUAL) reduce 583) ((T_XOR_EQUAL) reduce 583) (("[") reduce 583) (("^") reduce 583) (("{") reduce 583) (("|") reduce 583)) 29367 ((("%") reduce 561) (("(") reduce 561) ((")") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) (("=") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AND_EQUAL) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_COALESCE_EQUAL) reduce 561) ((T_CONCAT_EQUAL) reduce 561) ((T_DEC) reduce 561) ((T_DIV_EQUAL) reduce 561) ((T_INC) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_MINUS_EQUAL) reduce 561) ((T_MOD_EQUAL) reduce 561) ((T_MUL_EQUAL) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_OR_EQUAL) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_PLUS_EQUAL) reduce 561) ((T_POW) reduce 561) ((T_POW_EQUAL) reduce 561) ((T_SL) reduce 561) ((T_SL_EQUAL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) ((T_SR_EQUAL) reduce 561) ((T_XOR_EQUAL) reduce 561) (("[") reduce 561) (("^") reduce 561) (("{") reduce 561) (("|") reduce 561)) 29368 ((("%") reduce 567) (("(") shift) ((")") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("{") reduce 567) (("|") reduce 567)) 29369 ((("%") reduce 562) (("(") reduce 562) ((")") reduce 562) (("*") reduce 562) (("+") reduce 562) ((",") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) (("=") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AND_EQUAL) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_COALESCE_EQUAL) reduce 562) ((T_CONCAT_EQUAL) reduce 562) ((T_DEC) reduce 562) ((T_DIV_EQUAL) reduce 562) ((T_INC) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_MINUS_EQUAL) reduce 562) ((T_MOD_EQUAL) reduce 562) ((T_MUL_EQUAL) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_OR_EQUAL) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_PLUS_EQUAL) reduce 562) ((T_POW) reduce 562) ((T_POW_EQUAL) reduce 562) ((T_SL) reduce 562) ((T_SL_EQUAL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) ((T_SR_EQUAL) reduce 562) ((T_XOR_EQUAL) reduce 562) (("[") reduce 562) (("^") reduce 562) (("{") reduce 562) (("|") reduce 562)) 29371 ((("%") reduce 477) ((")") reduce 477) (("*") reduce 477) (("+") reduce 477) ((",") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_PIPE) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("{") reduce 477) (("|") reduce 477)) 29373 ((("%") reduce 520) (("(") reduce 520) ((")") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) (("=") shift) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_PIPE) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 29375 ((("%") shift) ((")") reduce 405) (("*") shift) (("+") shift) ((",") reduce 405) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 405) (("|") shift)) 29376 ((("%") shift) ((")") reduce 482) (("*") shift) (("+") shift) ((",") reduce 482) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 482) (("|") shift)) 29387 ((("%") reduce 540) ((")") reduce 540) (("*") reduce 540) (("+") reduce 540) ((",") reduce 540) (("-") reduce 540) ((".") reduce 540) (("/") reduce 540) (("<") reduce 540) ((">") reduce 540) (("?") reduce 540) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_BOOLEAN_AND) reduce 540) ((T_BOOLEAN_OR) reduce 540) ((T_COALESCE) reduce 540) ((T_DOUBLE_ARROW) reduce 540) ((T_INSTANCEOF) reduce 540) ((T_IS_EQUAL) reduce 540) ((T_IS_GREATER_OR_EQUAL) reduce 540) ((T_IS_IDENTICAL) reduce 540) ((T_IS_NOT_EQUAL) reduce 540) ((T_IS_NOT_IDENTICAL) reduce 540) ((T_IS_SMALLER_OR_EQUAL) reduce 540) ((T_LOGICAL_AND) reduce 540) ((T_LOGICAL_OR) reduce 540) ((T_LOGICAL_XOR) reduce 540) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 540) ((T_OBJECT_OPERATOR) reduce 540) ((T_PIPE) reduce 540) ((T_POW) reduce 540) ((T_SL) reduce 540) ((T_SPACESHIP) reduce 540) ((T_SR) reduce 540) (("[") reduce 540) (("^") reduce 540) (("{") reduce 540) (("|") reduce 540)) 29389 ((("%") reduce 521) (("(") reduce 521) ((")") reduce 521) (("*") reduce 521) (("+") reduce 521) ((",") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_PIPE) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("{") reduce 521) (("|") reduce 521)) 29391 ((("%") reduce 534) ((")") reduce 534) (("*") reduce 534) (("+") reduce 534) ((",") reduce 534) (("-") reduce 534) ((".") reduce 534) (("/") reduce 534) (("<") reduce 534) ((">") reduce 534) (("?") reduce 534) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_BOOLEAN_AND) reduce 534) ((T_BOOLEAN_OR) reduce 534) ((T_COALESCE) reduce 534) ((T_DOUBLE_ARROW) reduce 534) ((T_INSTANCEOF) reduce 534) ((T_IS_EQUAL) reduce 534) ((T_IS_GREATER_OR_EQUAL) reduce 534) ((T_IS_IDENTICAL) reduce 534) ((T_IS_NOT_EQUAL) reduce 534) ((T_IS_NOT_IDENTICAL) reduce 534) ((T_IS_SMALLER_OR_EQUAL) reduce 534) ((T_LOGICAL_AND) reduce 534) ((T_LOGICAL_OR) reduce 534) ((T_LOGICAL_XOR) reduce 534) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PIPE) reduce 534) ((T_POW) reduce 534) ((T_SL) reduce 534) ((T_SPACESHIP) reduce 534) ((T_SR) reduce 534) (("[") reduce 534) (("^") reduce 534) (("{") reduce 534) (("|") reduce 534)) 29392 ((("%") reduce 524) ((")") reduce 524) (("*") reduce 524) (("+") reduce 524) ((",") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_PIPE) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("^") reduce 524) (("{") reduce 524) (("|") reduce 524)) 29396 ((("%") reduce 517) (("(") shift) ((")") reduce 517) (("*") reduce 517) (("+") reduce 517) ((",") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_PIPE) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("^") reduce 517) (("{") reduce 517) (("|") reduce 517)) 29397 ((("%") reduce 533) ((")") reduce 533) (("*") reduce 533) (("+") reduce 533) ((",") reduce 533) (("-") reduce 533) ((".") reduce 533) (("/") reduce 533) (("<") reduce 533) ((">") reduce 533) (("?") reduce 533) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_BOOLEAN_AND) reduce 533) ((T_BOOLEAN_OR) reduce 533) ((T_COALESCE) reduce 533) ((T_DOUBLE_ARROW) reduce 533) ((T_INSTANCEOF) reduce 533) ((T_IS_EQUAL) reduce 533) ((T_IS_GREATER_OR_EQUAL) reduce 533) ((T_IS_IDENTICAL) reduce 533) ((T_IS_NOT_EQUAL) reduce 533) ((T_IS_NOT_IDENTICAL) reduce 533) ((T_IS_SMALLER_OR_EQUAL) reduce 533) ((T_LOGICAL_AND) reduce 533) ((T_LOGICAL_OR) reduce 533) ((T_LOGICAL_XOR) reduce 533) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PIPE) reduce 533) ((T_POW) reduce 533) ((T_SL) reduce 533) ((T_SPACESHIP) reduce 533) ((T_SR) reduce 533) (("[") reduce 533) (("^") reduce 533) (("{") reduce 533) (("|") reduce 533)) 29398 ((("%") reduce 537) ((")") reduce 537) (("*") reduce 537) (("+") reduce 537) ((",") reduce 537) (("-") reduce 537) ((".") reduce 537) (("/") reduce 537) (("<") reduce 537) ((">") reduce 537) (("?") reduce 537) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_BOOLEAN_AND) reduce 537) ((T_BOOLEAN_OR) reduce 537) ((T_COALESCE) reduce 537) ((T_DOUBLE_ARROW) reduce 537) ((T_INSTANCEOF) reduce 537) ((T_IS_EQUAL) reduce 537) ((T_IS_GREATER_OR_EQUAL) reduce 537) ((T_IS_IDENTICAL) reduce 537) ((T_IS_NOT_EQUAL) reduce 537) ((T_IS_NOT_IDENTICAL) reduce 537) ((T_IS_SMALLER_OR_EQUAL) reduce 537) ((T_LOGICAL_AND) reduce 537) ((T_LOGICAL_OR) reduce 537) ((T_LOGICAL_XOR) reduce 537) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_PIPE) reduce 537) ((T_POW) reduce 537) ((T_SL) reduce 537) ((T_SPACESHIP) reduce 537) ((T_SR) reduce 537) (("[") reduce 537) (("^") reduce 537) (("{") reduce 537) (("|") reduce 537)) 29404 ((("%") reduce 532) ((")") reduce 532) (("*") reduce 532) (("+") reduce 532) ((",") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PIPE) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("{") reduce 532) (("|") reduce 532)) 29406 ((("%") reduce 523) ((")") reduce 523) (("*") reduce 523) (("+") reduce 523) ((",") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_PIPE) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("^") reduce 523) (("{") reduce 523) (("|") reduce 523)) 29408 ((("%") reduce 536) ((")") reduce 536) (("*") reduce 536) (("+") reduce 536) ((",") reduce 536) (("-") reduce 536) ((".") reduce 536) (("/") reduce 536) (("<") reduce 536) ((">") reduce 536) (("?") reduce 536) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_BOOLEAN_AND) reduce 536) ((T_BOOLEAN_OR) reduce 536) ((T_COALESCE) reduce 536) ((T_DOUBLE_ARROW) reduce 536) ((T_INSTANCEOF) reduce 536) ((T_IS_EQUAL) reduce 536) ((T_IS_GREATER_OR_EQUAL) reduce 536) ((T_IS_IDENTICAL) reduce 536) ((T_IS_NOT_EQUAL) reduce 536) ((T_IS_NOT_IDENTICAL) reduce 536) ((T_IS_SMALLER_OR_EQUAL) reduce 536) ((T_LOGICAL_AND) reduce 536) ((T_LOGICAL_OR) reduce 536) ((T_LOGICAL_XOR) reduce 536) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 536) ((T_OBJECT_OPERATOR) reduce 536) ((T_PIPE) reduce 536) ((T_POW) reduce 536) ((T_SL) reduce 536) ((T_SPACESHIP) reduce 536) ((T_SR) reduce 536) (("[") reduce 536) (("^") reduce 536) (("{") reduce 536) (("|") reduce 536)) 29409 ((("%") reduce 94) (("(") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 29410 ((("%") reduce 93) (("(") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 29411 ((("%") reduce 95) (("(") reduce 95) ((")") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("{") reduce 95) (("|") reduce 95)) 29413 ((("%") reduce 539) ((")") reduce 539) (("*") reduce 539) (("+") reduce 539) ((",") reduce 539) (("-") reduce 539) ((".") reduce 539) (("/") reduce 539) (("<") reduce 539) ((">") reduce 539) (("?") reduce 539) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_BOOLEAN_AND) reduce 539) ((T_BOOLEAN_OR) reduce 539) ((T_COALESCE) reduce 539) ((T_DOUBLE_ARROW) reduce 539) ((T_INSTANCEOF) reduce 539) ((T_IS_EQUAL) reduce 539) ((T_IS_GREATER_OR_EQUAL) reduce 539) ((T_IS_IDENTICAL) reduce 539) ((T_IS_NOT_EQUAL) reduce 539) ((T_IS_NOT_IDENTICAL) reduce 539) ((T_IS_SMALLER_OR_EQUAL) reduce 539) ((T_LOGICAL_AND) reduce 539) ((T_LOGICAL_OR) reduce 539) ((T_LOGICAL_XOR) reduce 539) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 539) ((T_OBJECT_OPERATOR) reduce 539) ((T_PIPE) reduce 539) ((T_POW) reduce 539) ((T_SL) reduce 539) ((T_SPACESHIP) reduce 539) ((T_SR) reduce 539) (("[") reduce 539) (("^") reduce 539) (("{") reduce 539) (("|") reduce 539)) 29416 ((("%") reduce 538) ((")") reduce 538) (("*") reduce 538) (("+") reduce 538) ((",") reduce 538) (("-") reduce 538) ((".") reduce 538) (("/") reduce 538) (("<") reduce 538) ((">") reduce 538) (("?") reduce 538) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_BOOLEAN_AND) reduce 538) ((T_BOOLEAN_OR) reduce 538) ((T_COALESCE) reduce 538) ((T_DOUBLE_ARROW) reduce 538) ((T_INSTANCEOF) reduce 538) ((T_IS_EQUAL) reduce 538) ((T_IS_GREATER_OR_EQUAL) reduce 538) ((T_IS_IDENTICAL) reduce 538) ((T_IS_NOT_EQUAL) reduce 538) ((T_IS_NOT_IDENTICAL) reduce 538) ((T_IS_SMALLER_OR_EQUAL) reduce 538) ((T_LOGICAL_AND) reduce 538) ((T_LOGICAL_OR) reduce 538) ((T_LOGICAL_XOR) reduce 538) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 538) ((T_OBJECT_OPERATOR) reduce 538) ((T_PIPE) reduce 538) ((T_POW) reduce 538) ((T_SL) reduce 538) ((T_SPACESHIP) reduce 538) ((T_SR) reduce 538) (("[") reduce 538) (("^") reduce 538) (("{") reduce 538) (("|") reduce 538)) 29422 ((("%") reduce 92) (("(") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 29425 ((("%") reduce 535) ((")") reduce 535) (("*") reduce 535) (("+") reduce 535) ((",") reduce 535) (("-") reduce 535) ((".") reduce 535) (("/") reduce 535) (("<") reduce 535) ((">") reduce 535) (("?") reduce 535) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_BOOLEAN_AND) reduce 535) ((T_BOOLEAN_OR) reduce 535) ((T_COALESCE) reduce 535) ((T_DOUBLE_ARROW) reduce 535) ((T_INSTANCEOF) reduce 535) ((T_IS_EQUAL) reduce 535) ((T_IS_GREATER_OR_EQUAL) reduce 535) ((T_IS_IDENTICAL) reduce 535) ((T_IS_NOT_EQUAL) reduce 535) ((T_IS_NOT_IDENTICAL) reduce 535) ((T_IS_SMALLER_OR_EQUAL) reduce 535) ((T_LOGICAL_AND) reduce 535) ((T_LOGICAL_OR) reduce 535) ((T_LOGICAL_XOR) reduce 535) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PIPE) reduce 535) ((T_POW) reduce 535) ((T_SL) reduce 535) ((T_SPACESHIP) reduce 535) ((T_SR) reduce 535) (("[") reduce 535) (("^") reduce 535) (("{") reduce 535) (("|") reduce 535)) 29427 ((("%") reduce 568) (("(") reduce 568) ((")") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) (("=") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AND_EQUAL) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_COALESCE_EQUAL) reduce 568) ((T_CONCAT_EQUAL) reduce 568) ((T_DEC) reduce 568) ((T_DIV_EQUAL) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INC) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_MINUS_EQUAL) reduce 568) ((T_MOD_EQUAL) reduce 568) ((T_MUL_EQUAL) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_OR_EQUAL) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_PLUS_EQUAL) reduce 568) ((T_POW) reduce 568) ((T_POW_EQUAL) reduce 568) ((T_SL) reduce 568) ((T_SL_EQUAL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) ((T_SR_EQUAL) reduce 568) ((T_XOR_EQUAL) reduce 568) (("[") reduce 568) (("^") reduce 568) (("{") reduce 568) (("|") reduce 568)) 29428 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 479) (("(") shift) ((")") reduce 479) (("*") reduce 479) (("+") shift) ((",") reduce 479) (("-") shift) ((".") reduce 479) (("/") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 479) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 479) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 479) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PIPE) reduce 479) ((T_POW) reduce 479) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 479) (("`") shift) (("{") reduce 479) (("|") reduce 479) (("~") shift)) 29435 ((("%") reduce 564) (("(") reduce 555) ((")") reduce 564) (("*") reduce 564) (("+") reduce 564) ((",") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) (("<") reduce 564) (("=") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AND_EQUAL) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_COALESCE_EQUAL) reduce 564) ((T_CONCAT_EQUAL) reduce 564) ((T_DEC) reduce 564) ((T_DIV_EQUAL) reduce 564) ((T_DOUBLE_ARROW) reduce 564) ((T_INC) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_MINUS_EQUAL) reduce 564) ((T_MOD_EQUAL) reduce 564) ((T_MUL_EQUAL) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_OR_EQUAL) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_PLUS_EQUAL) reduce 564) ((T_POW) reduce 564) ((T_POW_EQUAL) reduce 564) ((T_SL) reduce 564) ((T_SL_EQUAL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) ((T_SR_EQUAL) reduce 564) ((T_XOR_EQUAL) reduce 564) (("[") reduce 564) (("^") reduce 564) (("{") reduce 564) (("|") reduce 564)) 29436 ((("%") reduce 530) ((")") reduce 530) (("*") reduce 530) (("+") reduce 530) ((",") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PIPE) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 551) (("^") reduce 530) (("{") reduce 530) (("|") reduce 530)) 29438 ((("%") reduce 529) ((")") reduce 529) (("*") reduce 529) (("+") reduce 529) ((",") reduce 529) (("-") reduce 529) ((".") reduce 529) (("/") reduce 529) (("<") reduce 529) ((">") reduce 529) (("?") reduce 529) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_BOOLEAN_AND) reduce 529) ((T_BOOLEAN_OR) reduce 529) ((T_COALESCE) reduce 529) ((T_DOUBLE_ARROW) reduce 529) ((T_INSTANCEOF) reduce 529) ((T_IS_EQUAL) reduce 529) ((T_IS_GREATER_OR_EQUAL) reduce 529) ((T_IS_IDENTICAL) reduce 529) ((T_IS_NOT_EQUAL) reduce 529) ((T_IS_NOT_IDENTICAL) reduce 529) ((T_IS_SMALLER_OR_EQUAL) reduce 529) ((T_LOGICAL_AND) reduce 529) ((T_LOGICAL_OR) reduce 529) ((T_LOGICAL_XOR) reduce 529) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PIPE) reduce 529) ((T_POW) reduce 529) ((T_SL) reduce 529) ((T_SPACESHIP) reduce 529) ((T_SR) reduce 529) (("[") reduce 554) (("^") reduce 529) (("{") reduce 529) (("|") reduce 529)) 29439 ((("%") reduce 528) (("(") reduce 557) ((")") reduce 528) (("*") reduce 528) (("+") reduce 528) ((",") reduce 528) (("-") reduce 528) ((".") reduce 528) (("/") reduce 528) (("<") reduce 528) ((">") reduce 528) (("?") reduce 528) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_BOOLEAN_AND) reduce 528) ((T_BOOLEAN_OR) reduce 528) ((T_COALESCE) reduce 528) ((T_DOUBLE_ARROW) reduce 528) ((T_INSTANCEOF) reduce 528) ((T_IS_EQUAL) reduce 528) ((T_IS_GREATER_OR_EQUAL) reduce 528) ((T_IS_IDENTICAL) reduce 528) ((T_IS_NOT_EQUAL) reduce 528) ((T_IS_NOT_IDENTICAL) reduce 528) ((T_IS_SMALLER_OR_EQUAL) reduce 528) ((T_LOGICAL_AND) reduce 528) ((T_LOGICAL_OR) reduce 528) ((T_LOGICAL_XOR) reduce 528) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PIPE) reduce 528) ((T_POW) reduce 528) ((T_SL) reduce 528) ((T_SPACESHIP) reduce 528) ((T_SR) reduce 528) (("[") reduce 550) (("^") reduce 528) (("{") reduce 528) (("|") reduce 528)) 29440 ((("%") shift) ((")") reduce 480) (("*") shift) (("+") shift) ((",") reduce 480) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 480) ((T_LOGICAL_OR) reduce 480) ((T_LOGICAL_XOR) reduce 480) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 480) (("|") shift)) 29443 ((("%") reduce 563) (("(") reduce 563) ((")") reduce 563) (("*") reduce 563) (("+") reduce 563) ((",") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) (("<") reduce 563) (("=") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AND_EQUAL) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_COALESCE_EQUAL) reduce 563) ((T_CONCAT_EQUAL) reduce 563) ((T_DEC) reduce 563) ((T_DIV_EQUAL) reduce 563) ((T_DOUBLE_ARROW) reduce 563) ((T_INC) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_MINUS_EQUAL) reduce 563) ((T_MOD_EQUAL) reduce 563) ((T_MUL_EQUAL) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_OR_EQUAL) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_PLUS_EQUAL) reduce 563) ((T_POW) reduce 563) ((T_POW_EQUAL) reduce 563) ((T_SL) reduce 563) ((T_SL_EQUAL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) ((T_SR_EQUAL) reduce 563) ((T_XOR_EQUAL) reduce 563) (("[") reduce 563) (("^") reduce 563) (("{") reduce 563) (("|") reduce 563)) 29444 ((("%") reduce 484) ((")") reduce 484) (("*") reduce 484) (("+") reduce 484) ((",") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_DOUBLE_ARROW) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_PIPE) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("^") reduce 484) (("{") reduce 484) (("|") reduce 484)) 29445 ((("%") reduce 466) ((")") reduce 466) (("*") reduce 466) (("+") reduce 466) ((",") reduce 466) (("-") reduce 466) ((".") reduce 466) (("/") reduce 466) (("<") reduce 466) ((">") reduce 466) (("?") reduce 466) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_BOOLEAN_AND) reduce 466) ((T_BOOLEAN_OR) reduce 466) ((T_COALESCE) reduce 466) ((T_DOUBLE_ARROW) reduce 466) ((T_INSTANCEOF) reduce 466) ((T_IS_EQUAL) reduce 466) ((T_IS_GREATER_OR_EQUAL) reduce 466) ((T_IS_IDENTICAL) reduce 466) ((T_IS_NOT_EQUAL) reduce 466) ((T_IS_NOT_IDENTICAL) reduce 466) ((T_IS_SMALLER_OR_EQUAL) reduce 466) ((T_LOGICAL_AND) reduce 466) ((T_LOGICAL_OR) reduce 466) ((T_LOGICAL_XOR) reduce 466) ((T_PIPE) reduce 466) ((T_POW) reduce 466) ((T_SL) reduce 466) ((T_SPACESHIP) reduce 466) ((T_SR) reduce 466) (("^") reduce 466) (("{") reduce 466) (("|") reduce 466)) 29446 ((("%") reduce 488) ((")") reduce 488) (("*") reduce 488) (("+") reduce 488) ((",") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_PIPE) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("^") reduce 488) (("{") reduce 488) (("|") reduce 488)) 29447 ((("%") reduce 531) (("(") shift) ((")") reduce 531) (("*") reduce 531) (("+") reduce 531) ((",") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("{") reduce 531) (("|") reduce 531)) 29448 ((("%") reduce 461) (("(") reduce 558) ((")") reduce 461) (("*") reduce 461) (("+") reduce 461) ((",") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PIPE) reduce 461) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") reduce 552) (("^") reduce 461) (("{") reduce 461) (("|") reduce 461)) 29449 ((("%") reduce 462) ((")") reduce 462) (("*") reduce 462) (("+") reduce 462) ((",") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_DOUBLE_ARROW) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_PIPE) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("{") reduce 462) (("|") reduce 462)) 29450 ((("%") reduce 476) ((")") reduce 476) (("*") reduce 476) (("+") reduce 476) ((",") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_DOUBLE_ARROW) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_PIPE) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("{") reduce 476) (("|") reduce 476)) 29451 ((("%") reduce 559) (("(") reduce 559) ((")") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) (("=") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AND_EQUAL) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_COALESCE_EQUAL) reduce 559) ((T_CONCAT_EQUAL) reduce 559) ((T_DEC) reduce 559) ((T_DIV_EQUAL) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INC) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_MINUS_EQUAL) reduce 559) ((T_MOD_EQUAL) reduce 559) ((T_MUL_EQUAL) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_OR_EQUAL) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_PLUS_EQUAL) reduce 559) ((T_POW) reduce 559) ((T_POW_EQUAL) reduce 559) ((T_SL) reduce 559) ((T_SL_EQUAL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) ((T_SR_EQUAL) reduce 559) ((T_XOR_EQUAL) reduce 559) (("[") reduce 559) (("^") reduce 559) (("{") reduce 559) (("|") reduce 559)) 29452 ((("%") reduce 565) ((")") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("{") reduce 565) (("|") reduce 565)) 29453 ((("%") reduce 403) ((")") reduce 403) (("*") reduce 403) (("+") reduce 403) ((",") reduce 403) (("-") reduce 403) ((".") reduce 403) (("/") reduce 403) (("<") reduce 403) (("=") shift) ((">") reduce 403) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_DOUBLE_ARROW) reduce 403) ((T_INC) shift) ((T_INSTANCEOF) reduce 403) ((T_IS_EQUAL) reduce 403) ((T_IS_GREATER_OR_EQUAL) reduce 403) ((T_IS_IDENTICAL) reduce 403) ((T_IS_NOT_EQUAL) reduce 403) ((T_IS_NOT_IDENTICAL) reduce 403) ((T_IS_SMALLER_OR_EQUAL) reduce 403) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 403) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 403) ((T_POW_EQUAL) shift) ((T_SL) reduce 403) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 403) ((T_SR) reduce 403) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 548) (("^") reduce 403) (("{") reduce 403) (("|") reduce 403)) 29456 ((("%") reduce 448) ((")") reduce 448) (("*") reduce 448) (("+") reduce 448) ((",") reduce 448) (("-") reduce 448) ((".") reduce 448) (("/") reduce 448) (("<") reduce 448) ((">") reduce 448) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_BOOLEAN_AND) reduce 448) ((T_BOOLEAN_OR) reduce 448) ((T_COALESCE) reduce 448) ((T_DOUBLE_ARROW) reduce 448) ((T_INSTANCEOF) reduce 448) ((T_IS_EQUAL) reduce 448) ((T_IS_GREATER_OR_EQUAL) reduce 448) ((T_IS_IDENTICAL) reduce 448) ((T_IS_NOT_EQUAL) reduce 448) ((T_IS_NOT_IDENTICAL) reduce 448) ((T_IS_SMALLER_OR_EQUAL) reduce 448) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_PIPE) reduce 448) ((T_POW) shift) ((T_SL) reduce 448) ((T_SPACESHIP) reduce 448) ((T_SR) reduce 448) (("^") reduce 448) (("{") reduce 448) (("|") reduce 448)) 29488 ((("%") shift) ((")") reduce 432) (("*") shift) (("+") shift) ((",") reduce 432) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_DOUBLE_ARROW) reduce 432) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 432) (("|") reduce 432)) 29489 ((("%") shift) ((")") reduce 435) (("*") shift) (("+") shift) ((",") reduce 435) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_DOUBLE_ARROW) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("{") reduce 435) (("|") reduce 435)) 29490 ((("%") shift) ((")") reduce 444) (("*") shift) (("+") shift) ((",") reduce 444) (("-") shift) ((".") reduce 444) (("/") shift) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_PIPE) reduce 444) ((T_POW) shift) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("{") reduce 444) (("|") reduce 444)) 29491 ((("%") shift) ((")") reduce 458) (("*") shift) (("+") shift) ((",") reduce 458) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 458) (("{") reduce 458) (("|") reduce 458)) 29492 ((("%") shift) ((")") reduce 443) (("*") shift) (("+") shift) ((",") reduce 443) (("-") shift) ((".") reduce 443) (("/") shift) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_DOUBLE_ARROW) reduce 443) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_PIPE) reduce 443) ((T_POW) shift) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("^") reduce 443) (("{") reduce 443) (("|") reduce 443)) 29493 ((("%") reduce 440) ((")") reduce 440) (("*") reduce 440) (("+") reduce 440) ((",") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_DOUBLE_ARROW) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_PIPE) reduce 440) ((T_POW) shift) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("{") reduce 440) (("|") reduce 440)) 29494 ((("%") shift) ((")") reduce 453) (("*") shift) (("+") shift) ((",") reduce 453) (("-") shift) ((".") shift) (("/") shift) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_DOUBLE_ARROW) reduce 453) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_PIPE) reduce 453) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 453) ((T_SR) shift) (("^") reduce 453) (("{") reduce 453) (("|") reduce 453)) 29495 ((("%") shift) ((")") reduce 431) (("*") shift) (("+") shift) ((",") reduce 431) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 431) (("|") shift)) 29496 ((("%") shift) ((")") reduce 429) (("*") shift) (("+") shift) ((",") reduce 429) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 429) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 429) (("|") shift)) 29497 ((("%") shift) ((")") reduce 430) (("*") shift) (("+") shift) ((",") reduce 430) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 430) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 430) (("|") shift)) 29498 ((("%") shift) ((")") reduce 455) (("*") shift) (("+") shift) ((",") reduce 455) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 455) ((T_SR) shift) (("^") reduce 455) (("{") reduce 455) (("|") reduce 455)) 29499 ((("%") shift) ((")") reduce 450) (("*") shift) (("+") shift) ((",") reduce 450) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_DOUBLE_ARROW) reduce 450) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 450) (("{") reduce 450) (("|") reduce 450)) 29500 ((("%") shift) ((")") reduce 452) (("*") shift) (("+") shift) ((",") reduce 452) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_DOUBLE_ARROW) reduce 452) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 452) (("{") reduce 452) (("|") reduce 452)) 29501 ((("%") shift) ((")") reduce 449) (("*") shift) (("+") shift) ((",") reduce 449) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_DOUBLE_ARROW) reduce 449) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 449) (("{") reduce 449) (("|") reduce 449)) 29502 ((("%") shift) ((")") reduce 457) (("*") shift) (("+") shift) ((",") reduce 457) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 457) ((T_SR) shift) (("^") reduce 457) (("{") reduce 457) (("|") reduce 457)) 29503 ((("%") shift) ((")") reduce 451) (("*") shift) (("+") shift) ((",") reduce 451) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_DOUBLE_ARROW) reduce 451) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 451) (("{") reduce 451) (("|") reduce 451)) 29506 ((("%") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 29507 ((("%") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 29508 ((("%") reduce 95) ((")") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("{") reduce 95) (("|") reduce 95)) 29509 ((("%") reduce 509) ((")") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 29510 ((("%") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 29511 ((("%") reduce 568) ((")") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("{") reduce 568) (("|") reduce 568)) 29512 ((("%") reduce 511) ((")") reduce 511) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 29513 ((("%") reduce 459) ((")") reduce 459) (("*") reduce 459) (("+") reduce 459) ((",") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PIPE) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("{") reduce 459) (("|") reduce 459)) 29514 ((("%") reduce 510) ((")") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 29515 ((("%") reduce 512) ((")") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 29516 ((("%") reduce 573) ((")") reduce 573) (("*") reduce 573) (("+") reduce 573) ((",") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_DOUBLE_ARROW) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("{") reduce 573) (("|") reduce 573)) 29522 ((("%") reduce 574) ((")") reduce 574) (("*") reduce 574) (("+") reduce 574) ((",") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_DOUBLE_ARROW) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("{") reduce 574) (("|") reduce 574)) 29523 ((("%") reduce 578) ((")") reduce 578) (("*") reduce 578) (("+") reduce 578) ((",") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_DOUBLE_ARROW) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("{") reduce 578) (("|") reduce 578)) 29524 ((("%") reduce 582) ((")") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("{") reduce 582) (("|") reduce 582)) 29525 ((("%") reduce 575) ((")") reduce 575) (("*") reduce 575) (("+") reduce 575) ((",") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_DOUBLE_ARROW) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("{") reduce 575) (("|") reduce 575)) 29526 ((("%") reduce 584) ((")") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("{") reduce 584) (("|") reduce 584)) 29529 ((("%") reduce 583) ((")") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("{") reduce 583) (("|") reduce 583)) 29530 ((("%") reduce 576) ((")") reduce 576) (("*") reduce 576) (("+") reduce 576) ((",") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_DOUBLE_ARROW) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("{") reduce 576) (("|") reduce 576)) 29532 ((("%") reduce 577) ((")") reduce 577) (("*") reduce 577) (("+") reduce 577) ((",") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_DOUBLE_ARROW) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("{") reduce 577) (("|") reduce 577)) 29534 ((("%") reduce 513) ((")") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 29535 ((("%") reduce 570) ((")") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("{") reduce 570) (("|") reduce 570)) 29538 ((("%") reduce 569) ((")") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("{") reduce 569) (("|") reduce 569)) 29539 ((("%") shift) ((")") reduce 465) (("*") shift) (("+") shift) ((",") reduce 465) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 465) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 465) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 465) (("|") shift)) 29540 ((("%") shift) ((")") reduce 427) (("*") shift) (("+") shift) ((",") reduce 427) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_DOUBLE_ARROW) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 427) (("|") shift)) 29541 ((("%") shift) ((")") reduce 428) (("*") shift) (("+") shift) ((",") reduce 428) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_DOUBLE_ARROW) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 428) (("|") shift)) 29542 ((("%") shift) ((")") reduce 433) (("*") shift) (("+") shift) ((",") reduce 433) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_DOUBLE_ARROW) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("{") reduce 433) (("|") reduce 433)) 29543 ((("%") shift) ((")") reduce 434) (("*") shift) (("+") shift) ((",") reduce 434) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_DOUBLE_ARROW) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("{") reduce 434) (("|") reduce 434)) 29547 ((("%") shift) ((")") reduce 463) (("*") shift) (("+") shift) ((",") reduce 463) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 463) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 463) (("|") shift)) 29548 ((("%") shift) ((")") reduce 464) (("*") shift) (("+") shift) ((",") reduce 464) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 464) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 464) (("|") shift)) 29549 ((("%") shift) ((")") reduce 456) (("*") shift) (("+") shift) ((",") reduce 456) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 456) ((T_SR) shift) (("^") reduce 456) (("{") reduce 456) (("|") reduce 456)) 29550 ((("%") shift) ((")") reduce 454) (("*") shift) (("+") shift) ((",") reduce 454) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 454) ((T_SR) shift) (("^") reduce 454) (("{") reduce 454) (("|") reduce 454)) 29551 ((("%") reduce 441) ((")") reduce 441) (("*") reduce 441) (("+") reduce 441) ((",") reduce 441) (("-") reduce 441) ((".") reduce 441) (("/") reduce 441) (("<") reduce 441) ((">") reduce 441) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_DOUBLE_ARROW) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 441) ((T_IS_GREATER_OR_EQUAL) reduce 441) ((T_IS_IDENTICAL) reduce 441) ((T_IS_NOT_EQUAL) reduce 441) ((T_IS_NOT_IDENTICAL) reduce 441) ((T_IS_SMALLER_OR_EQUAL) reduce 441) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_PIPE) reduce 441) ((T_POW) shift) ((T_SL) reduce 441) ((T_SPACESHIP) reduce 441) ((T_SR) reduce 441) (("^") reduce 441) (("{") reduce 441) (("|") reduce 441)) 29552 ((("%") shift) ((")") reduce 436) (("*") shift) (("+") shift) ((",") reduce 436) (("-") shift) ((".") reduce 436) (("/") shift) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_DOUBLE_ARROW) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_PIPE) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 436) ((T_SR) shift) (("^") reduce 436) (("{") reduce 436) (("|") reduce 436)) 29553 ((("%") shift) ((")") reduce 438) (("*") shift) (("+") reduce 438) ((",") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") shift) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_DOUBLE_ARROW) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_PIPE) reduce 438) ((T_POW) shift) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("{") reduce 438) (("|") reduce 438)) 29554 ((("%") shift) ((")") reduce 437) (("*") shift) (("+") reduce 437) ((",") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") shift) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_DOUBLE_ARROW) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_PIPE) reduce 437) ((T_POW) shift) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("{") reduce 437) (("|") reduce 437)) 29555 ((("%") reduce 439) ((")") reduce 439) (("*") reduce 439) (("+") reduce 439) ((",") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_DOUBLE_ARROW) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_PIPE) reduce 439) ((T_POW) shift) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("{") reduce 439) (("|") reduce 439)) 29556 ((("%") reduce 442) ((")") reduce 442) (("*") reduce 442) (("+") reduce 442) ((",") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_DOUBLE_ARROW) reduce 442) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_PIPE) reduce 442) ((T_POW) shift) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("{") reduce 442) (("|") reduce 442)) 29558 ((("%") reduce 74) (("(") reduce 74) ((")") reduce 74) (("*") reduce 74) (("+") reduce 74) ((",") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_DOUBLE_ARROW) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_PIPE) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("{") reduce 74) (("|") reduce 74)) 29559 ((("%") reduce 52) (("(") reduce 52) ((")") reduce 52) (("*") reduce 52) (("+") reduce 52) ((",") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_DOUBLE_ARROW) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_PIPE) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("{") reduce 52) (("|") reduce 52)) 29560 ((("%") reduce 27) (("(") reduce 27) ((")") reduce 27) (("*") reduce 27) (("+") reduce 27) ((",") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_DOUBLE_ARROW) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_PIPE) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("{") reduce 27) (("|") reduce 27)) 29561 ((("%") reduce 51) (("(") reduce 51) ((")") reduce 51) (("*") reduce 51) (("+") reduce 51) ((",") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_DOUBLE_ARROW) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_PIPE) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("{") reduce 51) (("|") reduce 51)) 29562 ((("%") reduce 53) (("(") reduce 53) ((")") reduce 53) (("*") reduce 53) (("+") reduce 53) ((",") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_DOUBLE_ARROW) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_PIPE) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("{") reduce 53) (("|") reduce 53)) 29563 ((("%") reduce 49) (("(") reduce 49) ((")") reduce 49) (("*") reduce 49) (("+") reduce 49) ((",") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_DOUBLE_ARROW) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_PIPE) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("{") reduce 49) (("|") reduce 49)) 29564 ((("%") reduce 29) (("(") reduce 29) ((")") reduce 29) (("*") reduce 29) (("+") reduce 29) ((",") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_DOUBLE_ARROW) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_PIPE) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("{") reduce 29) (("|") reduce 29)) 29565 ((("%") reduce 59) (("(") reduce 59) ((")") reduce 59) (("*") reduce 59) (("+") reduce 59) ((",") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_DOUBLE_ARROW) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_PIPE) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("{") reduce 59) (("|") reduce 59)) 29566 ((("%") reduce 60) (("(") reduce 60) ((")") reduce 60) (("*") reduce 60) (("+") reduce 60) ((",") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_DOUBLE_ARROW) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_PIPE) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("{") reduce 60) (("|") reduce 60)) 29567 ((("%") reduce 11) (("(") reduce 11) ((")") reduce 11) (("*") reduce 11) (("+") reduce 11) ((",") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_DOUBLE_ARROW) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_PIPE) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("{") reduce 11) (("|") reduce 11)) 29568 ((("%") reduce 42) (("(") reduce 42) ((")") reduce 42) (("*") reduce 42) (("+") reduce 42) ((",") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_DOUBLE_ARROW) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_PIPE) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("{") reduce 42) (("|") reduce 42)) 29569 ((("%") reduce 39) (("(") reduce 39) ((")") reduce 39) (("*") reduce 39) (("+") reduce 39) ((",") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_DOUBLE_ARROW) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_PIPE) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("{") reduce 39) (("|") reduce 39)) 29570 ((("%") reduce 25) (("(") reduce 25) ((")") reduce 25) (("*") reduce 25) (("+") reduce 25) ((",") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_DOUBLE_ARROW) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_PIPE) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("{") reduce 25) (("|") reduce 25)) 29571 ((("%") reduce 50) (("(") reduce 50) ((")") reduce 50) (("*") reduce 50) (("+") reduce 50) ((",") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_DOUBLE_ARROW) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_PIPE) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("{") reduce 50) (("|") reduce 50)) 29572 ((("%") reduce 66) (("(") reduce 66) ((")") reduce 66) (("*") reduce 66) (("+") reduce 66) ((",") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_DOUBLE_ARROW) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_PIPE) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("{") reduce 66) (("|") reduce 66)) 29573 ((("%") reduce 18) (("(") reduce 18) ((")") reduce 18) (("*") reduce 18) (("+") reduce 18) ((",") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_DOUBLE_ARROW) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_PIPE) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("{") reduce 18) (("|") reduce 18)) 29574 ((("%") reduce 17) (("(") reduce 17) ((")") reduce 17) (("*") reduce 17) (("+") reduce 17) ((",") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_DOUBLE_ARROW) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_PIPE) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("{") reduce 17) (("|") reduce 17)) 29575 ((("%") reduce 15) (("(") reduce 15) ((")") reduce 15) (("*") reduce 15) (("+") reduce 15) ((",") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_DOUBLE_ARROW) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_PIPE) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("{") reduce 15) (("|") reduce 15)) 29576 ((("%") reduce 14) (("(") reduce 14) ((")") reduce 14) (("*") reduce 14) (("+") reduce 14) ((",") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_DOUBLE_ARROW) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_PIPE) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("{") reduce 14) (("|") reduce 14)) 29577 ((("%") reduce 38) (("(") reduce 38) ((")") reduce 38) (("*") reduce 38) (("+") reduce 38) ((",") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_DOUBLE_ARROW) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_PIPE) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("{") reduce 38) (("|") reduce 38)) 29578 ((("%") reduce 26) (("(") reduce 26) ((")") reduce 26) (("*") reduce 26) (("+") reduce 26) ((",") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_DOUBLE_ARROW) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_PIPE) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("{") reduce 26) (("|") reduce 26)) 29579 ((("%") reduce 22) (("(") reduce 22) ((")") reduce 22) (("*") reduce 22) (("+") reduce 22) ((",") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_DOUBLE_ARROW) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_PIPE) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("{") reduce 22) (("|") reduce 22)) 29580 ((("%") reduce 24) (("(") reduce 24) ((")") reduce 24) (("*") reduce 24) (("+") reduce 24) ((",") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_DOUBLE_ARROW) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_PIPE) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("{") reduce 24) (("|") reduce 24)) 29581 ((("%") reduce 16) (("(") reduce 16) ((")") reduce 16) (("*") reduce 16) (("+") reduce 16) ((",") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_DOUBLE_ARROW) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_PIPE) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("{") reduce 16) (("|") reduce 16)) 29582 ((("%") reduce 48) (("(") reduce 48) ((")") reduce 48) (("*") reduce 48) (("+") reduce 48) ((",") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_DOUBLE_ARROW) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_PIPE) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("{") reduce 48) (("|") reduce 48)) 29583 ((("%") reduce 20) (("(") reduce 20) ((")") reduce 20) (("*") reduce 20) (("+") reduce 20) ((",") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_DOUBLE_ARROW) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_PIPE) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("{") reduce 20) (("|") reduce 20)) 29584 ((("%") reduce 70) (("(") reduce 70) ((")") reduce 70) (("*") reduce 70) (("+") reduce 70) ((",") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_DOUBLE_ARROW) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_PIPE) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("{") reduce 70) (("|") reduce 70)) 29585 ((("%") reduce 3) (("(") reduce 3) ((")") reduce 3) (("*") reduce 3) (("+") reduce 3) ((",") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_DOUBLE_ARROW) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_PIPE) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("{") reduce 3) (("|") reduce 3)) 29586 ((("%") reduce 12) (("(") reduce 12) ((")") reduce 12) (("*") reduce 12) (("+") reduce 12) ((",") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_DOUBLE_ARROW) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_PIPE) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("{") reduce 12) (("|") reduce 12)) 29587 ((("%") reduce 54) (("(") reduce 54) ((")") reduce 54) (("*") reduce 54) (("+") reduce 54) ((",") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_DOUBLE_ARROW) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_PIPE) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("{") reduce 54) (("|") reduce 54)) 29588 ((("%") reduce 65) (("(") reduce 65) ((")") reduce 65) (("*") reduce 65) (("+") reduce 65) ((",") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_DOUBLE_ARROW) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_PIPE) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("{") reduce 65) (("|") reduce 65)) 29589 ((("%") reduce 75) (("(") reduce 75) ((")") reduce 75) (("*") reduce 75) (("+") reduce 75) ((",") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_DOUBLE_ARROW) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_PIPE) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("{") reduce 75) (("|") reduce 75)) 29590 ((("%") reduce 30) (("(") reduce 30) ((")") reduce 30) (("*") reduce 30) (("+") reduce 30) ((",") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_DOUBLE_ARROW) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_PIPE) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("{") reduce 30) (("|") reduce 30)) 29591 ((("%") reduce 68) (("(") reduce 68) ((")") reduce 68) (("*") reduce 68) (("+") reduce 68) ((",") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_DOUBLE_ARROW) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_PIPE) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("{") reduce 68) (("|") reduce 68)) 29592 ((("%") reduce 21) (("(") reduce 21) ((")") reduce 21) (("*") reduce 21) (("+") reduce 21) ((",") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_DOUBLE_ARROW) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_PIPE) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("{") reduce 21) (("|") reduce 21)) 29593 ((("%") reduce 23) (("(") reduce 23) ((")") reduce 23) (("*") reduce 23) (("+") reduce 23) ((",") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_DOUBLE_ARROW) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_PIPE) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("{") reduce 23) (("|") reduce 23)) 29594 ((("%") reduce 41) (("(") reduce 41) ((")") reduce 41) (("*") reduce 41) (("+") reduce 41) ((",") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_DOUBLE_ARROW) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_PIPE) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("{") reduce 41) (("|") reduce 41)) 29595 ((("%") reduce 62) (("(") reduce 62) ((")") reduce 62) (("*") reduce 62) (("+") reduce 62) ((",") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_DOUBLE_ARROW) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_PIPE) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("{") reduce 62) (("|") reduce 62)) 29596 ((("%") reduce 34) (("(") reduce 34) ((")") reduce 34) (("*") reduce 34) (("+") reduce 34) ((",") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_DOUBLE_ARROW) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_PIPE) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("{") reduce 34) (("|") reduce 34)) 29597 ((("%") reduce 40) (("(") reduce 40) ((")") reduce 40) (("*") reduce 40) (("+") reduce 40) ((",") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_DOUBLE_ARROW) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_PIPE) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("{") reduce 40) (("|") reduce 40)) 29598 ((("%") reduce 13) (("(") reduce 13) ((")") reduce 13) (("*") reduce 13) (("+") reduce 13) ((",") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_DOUBLE_ARROW) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_PIPE) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("{") reduce 13) (("|") reduce 13)) 29599 ((("%") reduce 55) (("(") reduce 55) ((")") reduce 55) (("*") reduce 55) (("+") reduce 55) ((",") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_DOUBLE_ARROW) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_PIPE) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("{") reduce 55) (("|") reduce 55)) 29600 ((("%") reduce 1) (("(") reduce 1) ((")") reduce 1) (("*") reduce 1) (("+") reduce 1) ((",") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_DOUBLE_ARROW) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_PIPE) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("{") reduce 1) (("|") reduce 1)) 29601 ((("%") reduce 2) (("(") reduce 2) ((")") reduce 2) (("*") reduce 2) (("+") reduce 2) ((",") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_DOUBLE_ARROW) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_PIPE) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("{") reduce 2) (("|") reduce 2)) 29602 ((("%") reduce 9) (("(") reduce 9) ((")") reduce 9) (("*") reduce 9) (("+") reduce 9) ((",") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_DOUBLE_ARROW) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_PIPE) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("{") reduce 9) (("|") reduce 9)) 29603 ((("%") reduce 33) (("(") reduce 33) ((")") reduce 33) (("*") reduce 33) (("+") reduce 33) ((",") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_DOUBLE_ARROW) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_PIPE) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("{") reduce 33) (("|") reduce 33)) 29604 ((("%") reduce 58) (("(") reduce 58) ((")") reduce 58) (("*") reduce 58) (("+") reduce 58) ((",") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_DOUBLE_ARROW) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_PIPE) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("{") reduce 58) (("|") reduce 58)) 29605 ((("%") reduce 37) (("(") reduce 37) ((")") reduce 37) (("*") reduce 37) (("+") reduce 37) ((",") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_DOUBLE_ARROW) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_PIPE) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("{") reduce 37) (("|") reduce 37)) 29606 ((("%") reduce 64) (("(") reduce 64) ((")") reduce 64) (("*") reduce 64) (("+") reduce 64) ((",") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_DOUBLE_ARROW) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_PIPE) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("{") reduce 64) (("|") reduce 64)) 29607 ((("%") reduce 46) (("(") reduce 46) ((")") reduce 46) (("*") reduce 46) (("+") reduce 46) ((",") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_DOUBLE_ARROW) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_PIPE) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("{") reduce 46) (("|") reduce 46)) 29608 ((("%") reduce 8) (("(") reduce 8) ((")") reduce 8) (("*") reduce 8) (("+") reduce 8) ((",") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_DOUBLE_ARROW) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_PIPE) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("{") reduce 8) (("|") reduce 8)) 29609 ((("%") reduce 6) (("(") reduce 6) ((")") reduce 6) (("*") reduce 6) (("+") reduce 6) ((",") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_DOUBLE_ARROW) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_PIPE) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("{") reduce 6) (("|") reduce 6)) 29610 ((("%") reduce 7) (("(") reduce 7) ((")") reduce 7) (("*") reduce 7) (("+") reduce 7) ((",") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_DOUBLE_ARROW) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_PIPE) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("{") reduce 7) (("|") reduce 7)) 29611 ((("%") reduce 69) (("(") reduce 69) ((")") reduce 69) (("*") reduce 69) (("+") reduce 69) ((",") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_DOUBLE_ARROW) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_PIPE) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("{") reduce 69) (("|") reduce 69)) 29612 ((("%") reduce 63) (("(") reduce 63) ((")") reduce 63) (("*") reduce 63) (("+") reduce 63) ((",") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_DOUBLE_ARROW) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_PIPE) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("{") reduce 63) (("|") reduce 63)) 29613 ((("%") reduce 56) (("(") reduce 56) ((")") reduce 56) (("*") reduce 56) (("+") reduce 56) ((",") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_DOUBLE_ARROW) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_PIPE) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("{") reduce 56) (("|") reduce 56)) 29614 ((("%") reduce 10) (("(") reduce 10) ((")") reduce 10) (("*") reduce 10) (("+") reduce 10) ((",") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_DOUBLE_ARROW) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_PIPE) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("{") reduce 10) (("|") reduce 10)) 29615 ((("%") reduce 67) (("(") reduce 67) ((")") reduce 67) (("*") reduce 67) (("+") reduce 67) ((",") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_DOUBLE_ARROW) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_PIPE) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("{") reduce 67) (("|") reduce 67)) 29616 ((("%") reduce 44) (("(") reduce 44) ((")") reduce 44) (("*") reduce 44) (("+") reduce 44) ((",") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_DOUBLE_ARROW) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_PIPE) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("{") reduce 44) (("|") reduce 44)) 29617 ((("%") reduce 76) (("(") reduce 76) ((")") reduce 76) (("*") reduce 76) (("+") reduce 76) ((",") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_DOUBLE_ARROW) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_PIPE) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("{") reduce 76) (("|") reduce 76)) 29618 ((("%") reduce 71) (("(") reduce 71) ((")") reduce 71) (("*") reduce 71) (("+") reduce 71) ((",") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_DOUBLE_ARROW) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_PIPE) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("{") reduce 71) (("|") reduce 71)) 29619 ((("%") reduce 77) (("(") reduce 77) ((")") reduce 77) (("*") reduce 77) (("+") reduce 77) ((",") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_DOUBLE_ARROW) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_PIPE) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("{") reduce 77) (("|") reduce 77)) 29620 ((("%") reduce 78) (("(") reduce 78) ((")") reduce 78) (("*") reduce 78) (("+") reduce 78) ((",") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_DOUBLE_ARROW) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_PIPE) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("{") reduce 78) (("|") reduce 78)) 29621 ((("%") reduce 79) (("(") reduce 79) ((")") reduce 79) (("*") reduce 79) (("+") reduce 79) ((",") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_DOUBLE_ARROW) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_PIPE) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("{") reduce 79) (("|") reduce 79)) 29622 ((("%") reduce 4) (("(") reduce 4) ((")") reduce 4) (("*") reduce 4) (("+") reduce 4) ((",") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_DOUBLE_ARROW) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_PIPE) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("{") reduce 4) (("|") reduce 4)) 29623 ((("%") reduce 5) (("(") reduce 5) ((")") reduce 5) (("*") reduce 5) (("+") reduce 5) ((",") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_DOUBLE_ARROW) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_PIPE) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("{") reduce 5) (("|") reduce 5)) 29624 ((("%") reduce 43) (("(") reduce 43) ((")") reduce 43) (("*") reduce 43) (("+") reduce 43) ((",") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_DOUBLE_ARROW) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_PIPE) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("{") reduce 43) (("|") reduce 43)) 29625 ((("%") reduce 73) (("(") reduce 73) ((")") reduce 73) (("*") reduce 73) (("+") reduce 73) ((",") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_DOUBLE_ARROW) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_PIPE) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("{") reduce 73) (("|") reduce 73)) 29626 ((("%") reduce 82) (("(") reduce 82) ((")") reduce 82) (("*") reduce 82) (("+") reduce 82) ((",") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_DOUBLE_ARROW) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_PIPE) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("{") reduce 82) (("|") reduce 82)) 29627 ((("%") reduce 47) (("(") reduce 47) ((")") reduce 47) (("*") reduce 47) (("+") reduce 47) ((",") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_DOUBLE_ARROW) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_PIPE) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("{") reduce 47) (("|") reduce 47)) 29628 ((("%") reduce 31) (("(") reduce 31) ((")") reduce 31) (("*") reduce 31) (("+") reduce 31) ((",") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_DOUBLE_ARROW) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_PIPE) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("{") reduce 31) (("|") reduce 31)) 29629 ((("%") reduce 57) (("(") reduce 57) ((")") reduce 57) (("*") reduce 57) (("+") reduce 57) ((",") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_DOUBLE_ARROW) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_PIPE) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("{") reduce 57) (("|") reduce 57)) 29630 ((("%") reduce 61) (("(") reduce 61) ((")") reduce 61) (("*") reduce 61) (("+") reduce 61) ((",") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_DOUBLE_ARROW) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_PIPE) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("{") reduce 61) (("|") reduce 61)) 29631 ((("%") reduce 28) (("(") reduce 28) ((")") reduce 28) (("*") reduce 28) (("+") reduce 28) ((",") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_DOUBLE_ARROW) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_PIPE) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("{") reduce 28) (("|") reduce 28)) 29632 ((("%") reduce 36) (("(") reduce 36) ((")") reduce 36) (("*") reduce 36) (("+") reduce 36) ((",") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_DOUBLE_ARROW) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_PIPE) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("{") reduce 36) (("|") reduce 36)) 29633 ((("%") reduce 32) (("(") reduce 32) ((")") reduce 32) (("*") reduce 32) (("+") reduce 32) ((",") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_DOUBLE_ARROW) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_PIPE) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("{") reduce 32) (("|") reduce 32)) 29634 ((("%") reduce 35) (("(") reduce 35) ((")") reduce 35) (("*") reduce 35) (("+") reduce 35) ((",") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_DOUBLE_ARROW) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_PIPE) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("{") reduce 35) (("|") reduce 35)) 29635 ((("%") reduce 19) (("(") reduce 19) ((")") reduce 19) (("*") reduce 19) (("+") reduce 19) ((",") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_DOUBLE_ARROW) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_PIPE) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("{") reduce 19) (("|") reduce 19)) 29636 ((("%") reduce 45) (("(") reduce 45) ((")") reduce 45) (("*") reduce 45) (("+") reduce 45) ((",") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_DOUBLE_ARROW) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_PIPE) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("{") reduce 45) (("|") reduce 45)) 29637 ((("%") reduce 542) (("(") reduce 579) ((")") reduce 542) (("*") reduce 542) (("+") reduce 542) ((",") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PIPE) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("{") reduce 542) (("|") reduce 542)) 29639 ((("%") reduce 72) (("(") reduce 72) ((")") reduce 72) (("*") reduce 72) (("+") reduce 72) ((",") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_DOUBLE_ARROW) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_PIPE) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("{") reduce 72) (("|") reduce 72)) 29640 ((("%") reduce 83) (("(") reduce 83) ((")") reduce 83) (("*") reduce 83) (("+") reduce 83) ((",") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_DOUBLE_ARROW) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_PIPE) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("{") reduce 83) (("|") reduce 83)) 29641 ((("%") reduce 572) (("(") reduce 581) ((")") reduce 572) (("*") reduce 572) (("+") reduce 572) ((",") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) (("<") reduce 572) (("=") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AND_EQUAL) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_COALESCE_EQUAL) reduce 572) ((T_CONCAT_EQUAL) reduce 572) ((T_DEC) reduce 572) ((T_DIV_EQUAL) reduce 572) ((T_DOUBLE_ARROW) reduce 572) ((T_INC) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_MINUS_EQUAL) reduce 572) ((T_MOD_EQUAL) reduce 572) ((T_MUL_EQUAL) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_OR_EQUAL) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_PLUS_EQUAL) reduce 572) ((T_POW) reduce 572) ((T_POW_EQUAL) reduce 572) ((T_SL) reduce 572) ((T_SL_EQUAL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) ((T_SR_EQUAL) reduce 572) ((T_XOR_EQUAL) reduce 572) (("[") reduce 572) (("^") reduce 572) (("{") reduce 572) (("|") reduce 572)) 29644 ((("%") reduce 544) (("(") reduce 580) ((")") reduce 544) (("*") reduce 544) (("+") reduce 544) ((",") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PIPE) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("{") reduce 544) (("|") reduce 544)) 29646 ((("%") reduce 507) (("(") reduce 507) ((")") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 29647 ((("%") reduce 291) (("(") reduce 291) ((")") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("{") reduce 291) (("|") reduce 291)) 29651 ((("%") reduce 292) (("(") reduce 292) ((")") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) (("=") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AND_EQUAL) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_COALESCE_EQUAL) reduce 292) ((T_CONCAT_EQUAL) reduce 292) ((T_DEC) reduce 292) ((T_DIV_EQUAL) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INC) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_MINUS_EQUAL) reduce 292) ((T_MOD_EQUAL) reduce 292) ((T_MUL_EQUAL) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_OR_EQUAL) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_PLUS_EQUAL) reduce 292) ((T_POW) reduce 292) ((T_POW_EQUAL) reduce 292) ((T_SL) reduce 292) ((T_SL_EQUAL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) ((T_SR_EQUAL) reduce 292) ((T_XOR_EQUAL) reduce 292) (("[") reduce 292) (("^") reduce 292) (("{") reduce 292) (("|") reduce 292)) 29652 ((("%") reduce 293) (("(") reduce 293) ((")") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) (("=") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AND_EQUAL) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_COALESCE_EQUAL) reduce 293) ((T_CONCAT_EQUAL) reduce 293) ((T_DEC) reduce 293) ((T_DIV_EQUAL) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INC) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_MINUS_EQUAL) reduce 293) ((T_MOD_EQUAL) reduce 293) ((T_MUL_EQUAL) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_OR_EQUAL) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_PLUS_EQUAL) reduce 293) ((T_POW) reduce 293) ((T_POW_EQUAL) reduce 293) ((T_SL) reduce 293) ((T_SL_EQUAL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) ((T_SR_EQUAL) reduce 293) ((T_XOR_EQUAL) reduce 293) (("[") reduce 293) (("^") reduce 293) (("{") reduce 293) (("|") reduce 293)) 29657 ((("%") reduce 425) ((")") reduce 425) (("*") reduce 425) (("+") reduce 425) ((",") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_DOUBLE_ARROW) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_PIPE) reduce 425) ((T_POW) reduce 425) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("{") reduce 425) (("|") reduce 425)) 29659 ((("%") reduce 423) ((")") reduce 423) (("*") reduce 423) (("+") reduce 423) ((",") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_DOUBLE_ARROW) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_PIPE) reduce 423) ((T_POW) reduce 423) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("{") reduce 423) (("|") reduce 423)) 29669 ((("%") shift) ((")") reduce 419) (("*") shift) (("+") shift) ((",") reduce 419) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 419) (("|") shift)) 29670 ((("%") shift) ((")") reduce 421) (("*") shift) (("+") shift) ((",") reduce 421) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 421) (("|") shift)) 29671 ((("%") shift) ((")") reduce 420) (("*") shift) (("+") shift) ((",") reduce 420) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 420) (("|") shift)) 29672 ((("%") shift) ((")") reduce 413) (("*") shift) (("+") shift) ((",") reduce 413) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 413) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 413) (("|") shift)) 29673 ((("%") shift) ((")") reduce 410) (("*") shift) (("+") shift) ((",") reduce 410) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 410) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 410) (("|") shift)) 29674 ((("%") shift) ((")") reduce 418) (("*") shift) (("+") shift) ((",") reduce 418) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 418) (("|") shift)) 29675 ((("%") shift) ((")") reduce 412) (("*") shift) (("+") shift) ((",") reduce 412) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 412) (("|") shift)) 29676 ((("%") shift) ((")") reduce 416) (("*") shift) (("+") shift) ((",") reduce 416) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 416) (("|") shift)) 29677 ((("%") shift) ((")") reduce 411) (("*") shift) (("+") shift) ((",") reduce 411) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 411) (("|") shift)) 29678 ((("%") shift) ((")") reduce 414) (("*") shift) (("+") shift) ((",") reduce 414) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 414) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 414) (("|") shift)) 29679 ((("%") shift) ((")") reduce 415) (("*") shift) (("+") shift) ((",") reduce 415) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 415) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 415) (("|") shift)) 29680 ((("%") shift) ((")") reduce 422) (("*") shift) (("+") shift) ((",") reduce 422) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 422) (("|") shift)) 29681 ((("%") shift) ((")") reduce 417) (("*") shift) (("+") shift) ((",") reduce 417) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 417) (("|") shift)) 29683 ((("%") shift) ((")") reduce 406) (("*") shift) (("+") shift) ((",") reduce 406) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 406) (("|") shift)) 29686 ((("%") reduce 568) (("(") reduce 568) ((")") reduce 568) (("*") reduce 568) (("+") reduce 568) ((",") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("{") reduce 568) (("|") reduce 568)) 29689 ((("%") reduce 564) (("(") reduce 555) ((")") reduce 564) (("*") reduce 564) (("+") reduce 564) ((",") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) (("<") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_DOUBLE_ARROW) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_POW) reduce 564) ((T_SL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) (("[") reduce 564) (("^") reduce 564) (("{") reduce 564) (("|") reduce 564)) 29691 ((("%") reduce 563) (("(") reduce 563) ((")") reduce 563) (("*") reduce 563) (("+") reduce 563) ((",") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) (("<") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_DOUBLE_ARROW) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_POW) reduce 563) ((T_SL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) (("[") reduce 563) (("^") reduce 563) (("{") reduce 563) (("|") reduce 563)) 29693 ((("%") reduce 559) (("(") reduce 559) ((")") reduce 559) (("*") reduce 559) (("+") reduce 559) ((",") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("{") reduce 559) (("|") reduce 559)) 29694 ((("%") reduce 565) ((")") reduce 565) (("*") reduce 565) (("+") reduce 565) ((",") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("{") reduce 565) (("|") reduce 565)) 29695 ((("%") reduce 407) ((")") reduce 407) (("*") reduce 407) (("+") reduce 407) ((",") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_DOUBLE_ARROW) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("[") reduce 548) (("^") reduce 407) (("{") reduce 407) (("|") reduce 407)) 29699 ((("%") reduce 572) (("(") reduce 581) ((")") reduce 572) (("*") reduce 572) (("+") reduce 572) ((",") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) (("<") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_DOUBLE_ARROW) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_POW) reduce 572) ((T_SL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) (("[") reduce 572) (("^") reduce 572) (("{") reduce 572) (("|") reduce 572)) 29701 ((("%") reduce 507) (("(") reduce 507) ((")") reduce 507) (("*") reduce 507) (("+") reduce 507) ((",") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("{") reduce 507) (("|") reduce 507)) 29702 ((("%") reduce 291) (("(") reduce 291) ((")") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("{") reduce 291) (("|") reduce 291)) 29706 ((("%") reduce 292) (("(") reduce 292) ((")") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("[") reduce 292) (("^") reduce 292) (("{") reduce 292) (("|") reduce 292)) 29707 ((("%") reduce 293) (("(") reduce 293) ((")") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("[") reduce 293) (("^") reduce 293) (("{") reduce 293) (("|") reduce 293)) 29708 ((("%") reduce 504) (("(") reduce 504) ((")") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("{") reduce 504) (("|") reduce 504)) 29711 ((("%") reduce 571) (("(") reduce 581) ((")") reduce 571) (("*") reduce 571) (("+") reduce 571) ((",") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) (("<") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_DOUBLE_ARROW) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_POW) reduce 571) ((T_SL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) (("[") reduce 571) (("^") reduce 571) (("{") reduce 571) (("|") reduce 571)) 29712 ((("%") reduce 506) (("(") reduce 506) ((")") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 29713 ((("%") reduce 508) (("(") reduce 508) ((")") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 29718 ((("%") reduce 560) (("(") reduce 560) ((")") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("{") reduce 560) (("|") reduce 560)) 29719 ((("%") reduce 582) (("(") reduce 582) ((")") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("{") reduce 582) (("|") reduce 582)) 29720 ((("%") reduce 566) (("(") shift) ((")") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("{") reduce 566) (("|") reduce 566)) 29721 ((("%") reduce 584) (("(") reduce 584) ((")") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("{") reduce 584) (("|") reduce 584)) 29724 ((("%") reduce 583) (("(") reduce 583) ((")") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("{") reduce 583) (("|") reduce 583)) 29725 ((("%") reduce 561) (("(") reduce 561) ((")") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("{") reduce 561) (("|") reduce 561)) 29726 ((("%") reduce 567) (("(") shift) ((")") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("{") reduce 567) (("|") reduce 567)) 29727 ((("%") reduce 562) (("(") reduce 562) ((")") reduce 562) (("*") reduce 562) (("+") reduce 562) ((",") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("[") reduce 562) (("^") reduce 562) (("{") reduce 562) (("|") reduce 562)) 29728 ((("%") reduce 505) (("(") reduce 505) ((")") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 29729 ((("%") reduce 570) (("(") reduce 570) ((")") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("{") reduce 570) (("|") reduce 570)) 29732 ((("%") reduce 569) (("(") reduce 569) ((")") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("{") reduce 569) (("|") reduce 569)) 29733 ((("%") reduce 504) (("(") reduce 504) ((")") reduce 504) (("*") reduce 504) (("+") reduce 504) ((",") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) (("<") reduce 504) (("=") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AND_EQUAL) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_COALESCE_EQUAL) reduce 504) ((T_CONCAT_EQUAL) reduce 504) ((T_DEC) reduce 504) ((T_DIV_EQUAL) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INC) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_MINUS_EQUAL) reduce 504) ((T_MOD_EQUAL) reduce 504) ((T_MUL_EQUAL) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_OR_EQUAL) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_PLUS_EQUAL) reduce 504) ((T_POW) reduce 504) ((T_POW_EQUAL) reduce 504) ((T_SL) reduce 504) ((T_SL_EQUAL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) ((T_SR_EQUAL) reduce 504) ((T_XOR_EQUAL) reduce 504) (("[") reduce 504) (("^") reduce 504) (("{") reduce 504) (("|") reduce 504)) 29744 ((("%") reduce 494) ((")") reduce 494) (("*") reduce 494) (("+") reduce 494) ((",") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PIPE) reduce 494) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("{") reduce 494) (("|") reduce 494)) 29745 ((("%") reduce 489) ((")") reduce 489) (("*") reduce 489) (("+") reduce 489) ((",") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_PIPE) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("^") reduce 489) (("{") reduce 489) (("|") reduce 489)) 29755 ((("%") shift) ((")") reduce 494) (("*") shift) (("+") shift) ((",") reduce 494) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 494) (("|") shift)) 29756 ((("%") reduce 490) ((")") reduce 490) (("*") reduce 490) (("+") reduce 490) ((",") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_PIPE) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("^") reduce 490) (("{") reduce 490) (("|") reduce 490)) 29758 ((("%") shift) ((")") reduce 481) (("*") shift) (("+") shift) ((",") reduce 481) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 481) (("|") shift)) 29760 ((("%") reduce 541) (("(") reduce 579) ((")") reduce 541) (("*") reduce 541) (("+") reduce 541) ((",") reduce 541) (("-") reduce 541) ((".") reduce 541) (("/") reduce 541) (("<") reduce 541) ((">") reduce 541) (("?") reduce 541) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_BOOLEAN_AND) reduce 541) ((T_BOOLEAN_OR) reduce 541) ((T_COALESCE) reduce 541) ((T_DOUBLE_ARROW) reduce 541) ((T_INSTANCEOF) reduce 541) ((T_IS_EQUAL) reduce 541) ((T_IS_GREATER_OR_EQUAL) reduce 541) ((T_IS_IDENTICAL) reduce 541) ((T_IS_NOT_EQUAL) reduce 541) ((T_IS_NOT_IDENTICAL) reduce 541) ((T_IS_SMALLER_OR_EQUAL) reduce 541) ((T_LOGICAL_AND) reduce 541) ((T_LOGICAL_OR) reduce 541) ((T_LOGICAL_XOR) reduce 541) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 541) ((T_OBJECT_OPERATOR) reduce 541) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 541) ((T_PIPE) reduce 541) ((T_POW) reduce 541) ((T_SL) reduce 541) ((T_SPACESHIP) reduce 541) ((T_SR) reduce 541) (("[") reduce 541) (("^") reduce 541) (("{") reduce 541) (("|") reduce 541)) 29762 ((("%") reduce 571) (("(") reduce 581) ((")") reduce 571) (("*") reduce 571) (("+") reduce 571) ((",") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) (("<") reduce 571) (("=") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AND_EQUAL) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_COALESCE_EQUAL) reduce 571) ((T_CONCAT_EQUAL) reduce 571) ((T_DEC) reduce 571) ((T_DIV_EQUAL) reduce 571) ((T_DOUBLE_ARROW) reduce 571) ((T_INC) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_MINUS_EQUAL) reduce 571) ((T_MOD_EQUAL) reduce 571) ((T_MUL_EQUAL) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_OR_EQUAL) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_PLUS_EQUAL) reduce 571) ((T_POW) reduce 571) ((T_POW_EQUAL) reduce 571) ((T_SL) reduce 571) ((T_SL_EQUAL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) ((T_SR_EQUAL) reduce 571) ((T_XOR_EQUAL) reduce 571) (("[") reduce 571) (("^") reduce 571) (("{") reduce 571) (("|") reduce 571)) 29765 ((("%") reduce 543) (("(") reduce 580) ((")") reduce 543) (("*") reduce 543) (("+") reduce 543) ((",") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PIPE) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("{") reduce 543) (("|") reduce 543)) 29766 ((("%") reduce 506) (("(") reduce 506) ((")") reduce 506) (("*") reduce 506) (("+") reduce 506) ((",") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("{") reduce 506) (("|") reduce 506)) 29767 ((("%") reduce 508) (("(") reduce 508) ((")") reduce 508) (("*") reduce 508) (("+") reduce 508) ((",") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("{") reduce 508) (("|") reduce 508)) 29769 ((("%") reduce 485) ((")") reduce 485) (("*") reduce 485) (("+") reduce 485) ((",") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_DOUBLE_ARROW) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_PIPE) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("^") reduce 485) (("{") reduce 485) (("|") reduce 485)) 29770 ((("%") reduce 487) ((")") reduce 487) (("*") reduce 487) (("+") reduce 487) ((",") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_PIPE) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("^") reduce 487) (("{") reduce 487) (("|") reduce 487)) 29775 ((("%") reduce 560) (("(") reduce 560) ((")") reduce 560) (("*") reduce 560) (("+") reduce 560) ((",") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) (("<") reduce 560) (("=") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AND_EQUAL) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_COALESCE_EQUAL) reduce 560) ((T_CONCAT_EQUAL) reduce 560) ((T_DEC) reduce 560) ((T_DIV_EQUAL) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INC) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_MINUS_EQUAL) reduce 560) ((T_MOD_EQUAL) reduce 560) ((T_MUL_EQUAL) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_OR_EQUAL) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_PLUS_EQUAL) reduce 560) ((T_POW) reduce 560) ((T_POW_EQUAL) reduce 560) ((T_SL) reduce 560) ((T_SL_EQUAL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) ((T_SR_EQUAL) reduce 560) ((T_XOR_EQUAL) reduce 560) (("[") reduce 560) (("^") reduce 560) (("{") reduce 560) (("|") reduce 560)) 29776 ((("%") reduce 582) (("(") reduce 582) ((")") reduce 582) (("*") reduce 582) (("+") reduce 582) ((",") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) (("<") reduce 582) (("=") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AND_EQUAL) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_COALESCE_EQUAL) reduce 582) ((T_CONCAT_EQUAL) reduce 582) ((T_DEC) reduce 582) ((T_DIV_EQUAL) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INC) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_MINUS_EQUAL) reduce 582) ((T_MOD_EQUAL) reduce 582) ((T_MUL_EQUAL) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_OR_EQUAL) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_PLUS_EQUAL) reduce 582) ((T_POW) reduce 582) ((T_POW_EQUAL) reduce 582) ((T_SL) reduce 582) ((T_SL_EQUAL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) ((T_SR_EQUAL) reduce 582) ((T_XOR_EQUAL) reduce 582) (("[") reduce 582) (("^") reduce 582) (("{") reduce 582) (("|") reduce 582)) 29777 ((("%") reduce 566) (("(") shift) ((")") reduce 566) (("*") reduce 566) (("+") reduce 566) ((",") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("{") reduce 566) (("|") reduce 566)) 29778 ((("%") reduce 584) (("(") reduce 584) ((")") reduce 584) (("*") reduce 584) (("+") reduce 584) ((",") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) (("<") reduce 584) (("=") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AND_EQUAL) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_COALESCE_EQUAL) reduce 584) ((T_CONCAT_EQUAL) reduce 584) ((T_DEC) reduce 584) ((T_DIV_EQUAL) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INC) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_MINUS_EQUAL) reduce 584) ((T_MOD_EQUAL) reduce 584) ((T_MUL_EQUAL) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_OR_EQUAL) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_PLUS_EQUAL) reduce 584) ((T_POW) reduce 584) ((T_POW_EQUAL) reduce 584) ((T_SL) reduce 584) ((T_SL_EQUAL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) ((T_SR_EQUAL) reduce 584) ((T_XOR_EQUAL) reduce 584) (("[") reduce 584) (("^") reduce 584) (("{") reduce 584) (("|") reduce 584)) 29781 ((("%") reduce 583) (("(") reduce 583) ((")") reduce 583) (("*") reduce 583) (("+") reduce 583) ((",") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) (("<") reduce 583) (("=") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AND_EQUAL) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_COALESCE_EQUAL) reduce 583) ((T_CONCAT_EQUAL) reduce 583) ((T_DEC) reduce 583) ((T_DIV_EQUAL) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INC) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_MINUS_EQUAL) reduce 583) ((T_MOD_EQUAL) reduce 583) ((T_MUL_EQUAL) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_OR_EQUAL) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_PLUS_EQUAL) reduce 583) ((T_POW) reduce 583) ((T_POW_EQUAL) reduce 583) ((T_SL) reduce 583) ((T_SL_EQUAL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) ((T_SR_EQUAL) reduce 583) ((T_XOR_EQUAL) reduce 583) (("[") reduce 583) (("^") reduce 583) (("{") reduce 583) (("|") reduce 583)) 29782 ((("%") reduce 561) (("(") reduce 561) ((")") reduce 561) (("*") reduce 561) (("+") reduce 561) ((",") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) (("<") reduce 561) (("=") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AND_EQUAL) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_COALESCE_EQUAL) reduce 561) ((T_CONCAT_EQUAL) reduce 561) ((T_DEC) reduce 561) ((T_DIV_EQUAL) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INC) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_MINUS_EQUAL) reduce 561) ((T_MOD_EQUAL) reduce 561) ((T_MUL_EQUAL) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_OR_EQUAL) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_PLUS_EQUAL) reduce 561) ((T_POW) reduce 561) ((T_POW_EQUAL) reduce 561) ((T_SL) reduce 561) ((T_SL_EQUAL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) ((T_SR_EQUAL) reduce 561) ((T_XOR_EQUAL) reduce 561) (("[") reduce 561) (("^") reduce 561) (("{") reduce 561) (("|") reduce 561)) 29783 ((("%") reduce 567) (("(") shift) ((")") reduce 567) (("*") reduce 567) (("+") reduce 567) ((",") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("{") reduce 567) (("|") reduce 567)) 29784 ((("%") reduce 562) (("(") reduce 562) ((")") reduce 562) (("*") reduce 562) (("+") reduce 562) ((",") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) (("<") reduce 562) (("=") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AND_EQUAL) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_COALESCE_EQUAL) reduce 562) ((T_CONCAT_EQUAL) reduce 562) ((T_DEC) reduce 562) ((T_DIV_EQUAL) reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_INC) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_MINUS_EQUAL) reduce 562) ((T_MOD_EQUAL) reduce 562) ((T_MUL_EQUAL) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_OR_EQUAL) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_PLUS_EQUAL) reduce 562) ((T_POW) reduce 562) ((T_POW_EQUAL) reduce 562) ((T_SL) reduce 562) ((T_SL_EQUAL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) ((T_SR_EQUAL) reduce 562) ((T_XOR_EQUAL) reduce 562) (("[") reduce 562) (("^") reduce 562) (("{") reduce 562) (("|") reduce 562)) 29786 ((("%") reduce 477) ((")") reduce 477) (("*") reduce 477) (("+") reduce 477) ((",") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_PIPE) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("{") reduce 477) (("|") reduce 477)) 29788 ((("%") reduce 520) (("(") reduce 520) ((")") reduce 520) (("*") reduce 520) (("+") reduce 520) ((",") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) (("<") reduce 520) (("=") shift) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_PIPE) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("{") reduce 520) (("|") reduce 520)) 29790 ((("%") shift) ((")") reduce 405) (("*") shift) (("+") shift) ((",") reduce 405) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 405) (("|") shift)) 29791 ((("%") shift) ((")") reduce 482) (("*") shift) (("+") shift) ((",") reduce 482) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 482) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 482) (("|") shift)) 29794 ((("%") shift) ((")") reduce 481) (("*") shift) (("+") shift) ((",") reduce 481) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 481) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 481) (("|") shift)) 29795 ((("%") reduce 473) ((")") reduce 473) (("*") reduce 473) (("+") reduce 473) ((",") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_DOUBLE_ARROW) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_PIPE) reduce 473) ((T_POW) shift) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("{") reduce 473) (("|") reduce 473)) 29796 ((("%") shift) ((")") reduce 483) (("*") shift) (("+") shift) ((",") reduce 483) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 483) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 483) (("|") shift)) 29797 ((("%") reduce 469) ((")") reduce 469) (("*") reduce 469) (("+") reduce 469) ((",") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_DOUBLE_ARROW) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_PIPE) reduce 469) ((T_POW) shift) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("{") reduce 469) (("|") reduce 469)) 29798 ((("%") reduce 486) ((")") reduce 486) (("*") reduce 486) (("+") reduce 486) ((",") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_DOUBLE_ARROW) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_PIPE) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("^") reduce 486) (("{") reduce 486) (("|") reduce 486)) 29800 ((("%") reduce 526) ((")") reduce 526) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_PIPE) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 29802 ((("%") reduce 527) ((")") reduce 527) (("*") reduce 527) (("+") reduce 527) ((",") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_DOUBLE_ARROW) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_PIPE) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("^") reduce 527) (("{") reduce 527) (("|") reduce 527)) 29803 ((("%") reduce 525) ((")") reduce 525) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_PIPE) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 29804 ((("%") shift) ((")") reduce 619) (("*") shift) (("+") shift) ((",") reduce 619) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 619) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 619) (("|") shift)) 29805 ((("%") shift) ((")") reduce 618) (("*") shift) (("+") shift) ((",") reduce 618) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 618) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 618) (("|") shift)) 29806 ((("%") reduce 505) (("(") reduce 505) ((")") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 29807 ((("%") shift) ((")") reduce 478) (("*") shift) (("+") shift) ((",") reduce 478) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 478) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 478) (("|") shift)) 29808 ((("%") reduce 471) ((")") reduce 471) (("*") reduce 471) (("+") reduce 471) ((",") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_DOUBLE_ARROW) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_PIPE) reduce 471) ((T_POW) shift) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("{") reduce 471) (("|") reduce 471)) 29810 ((("%") reduce 94) (("(") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 29811 ((("%") reduce 93) (("(") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 29812 ((("%") reduce 95) (("(") reduce 95) ((")") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("{") reduce 95) (("|") reduce 95)) 29813 ((("%") reduce 509) (("(") reduce 509) ((")") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 29814 ((("%") reduce 92) (("(") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 29815 ((("%") reduce 400) (("(") reduce 400) ((")") reduce 400) (("*") reduce 400) (("+") reduce 400) ((",") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_DOUBLE_ARROW) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 400) ((T_OBJECT_OPERATOR) reduce 400) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 400) ((T_PIPE) reduce 400) ((T_POW) reduce 400) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("[") reduce 400) (("^") reduce 400) (("{") reduce 400) (("|") reduce 400)) 29818 ((("%") reduce 511) (("(") reduce 511) ((")") reduce 511) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 29819 ((("%") reduce 402) (("(") shift) ((")") reduce 402) (("*") reduce 402) (("+") reduce 402) ((",") reduce 402) (("-") reduce 402) ((".") reduce 402) (("/") reduce 402) (("<") reduce 402) ((">") reduce 402) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_DOUBLE_ARROW) reduce 402) ((T_INSTANCEOF) reduce 402) ((T_IS_EQUAL) reduce 402) ((T_IS_GREATER_OR_EQUAL) reduce 402) ((T_IS_IDENTICAL) reduce 402) ((T_IS_NOT_EQUAL) reduce 402) ((T_IS_NOT_IDENTICAL) reduce 402) ((T_IS_SMALLER_OR_EQUAL) reduce 402) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_PIPE) reduce 402) ((T_POW) reduce 402) ((T_SL) reduce 402) ((T_SPACESHIP) reduce 402) ((T_SR) reduce 402) (("^") reduce 402) (("{") reduce 402) (("|") reduce 402)) 29820 ((("%") reduce 510) (("(") reduce 510) ((")") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 29821 ((("%") reduce 512) (("(") reduce 512) ((")") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 29822 ((("%") reduce 573) (("(") reduce 573) ((")") reduce 573) (("*") reduce 573) (("+") reduce 573) ((",") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_DOUBLE_ARROW) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("{") reduce 573) (("|") reduce 573)) 29828 ((("%") reduce 574) (("(") reduce 574) ((")") reduce 574) (("*") reduce 574) (("+") reduce 574) ((",") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_DOUBLE_ARROW) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("{") reduce 574) (("|") reduce 574)) 29829 ((("%") reduce 578) (("(") reduce 578) ((")") reduce 578) (("*") reduce 578) (("+") reduce 578) ((",") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_DOUBLE_ARROW) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("{") reduce 578) (("|") reduce 578)) 29830 ((("%") reduce 575) (("(") reduce 575) ((")") reduce 575) (("*") reduce 575) (("+") reduce 575) ((",") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_DOUBLE_ARROW) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("{") reduce 575) (("|") reduce 575)) 29831 ((("%") reduce 576) (("(") reduce 576) ((")") reduce 576) (("*") reduce 576) (("+") reduce 576) ((",") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_DOUBLE_ARROW) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("{") reduce 576) (("|") reduce 576)) 29832 ((("%") reduce 399) (("(") reduce 399) ((")") reduce 399) (("*") reduce 399) (("+") reduce 399) ((",") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_DOUBLE_ARROW) reduce 399) ((T_INSTANCEOF) reduce 399) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 399) ((T_OBJECT_OPERATOR) reduce 399) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 399) ((T_PIPE) reduce 399) ((T_POW) reduce 399) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("[") reduce 399) (("^") reduce 399) (("{") reduce 399) (("|") reduce 399)) 29834 ((("%") reduce 577) (("(") reduce 577) ((")") reduce 577) (("*") reduce 577) (("+") reduce 577) ((",") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_DOUBLE_ARROW) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("{") reduce 577) (("|") reduce 577)) 29835 ((("%") reduce 401) (("(") reduce 401) ((")") reduce 401) (("*") reduce 401) (("+") reduce 401) ((",") reduce 401) (("-") reduce 401) ((".") reduce 401) (("/") reduce 401) (("<") reduce 401) ((">") reduce 401) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_DOUBLE_ARROW) reduce 401) ((T_INSTANCEOF) reduce 401) ((T_IS_EQUAL) reduce 401) ((T_IS_GREATER_OR_EQUAL) reduce 401) ((T_IS_IDENTICAL) reduce 401) ((T_IS_NOT_EQUAL) reduce 401) ((T_IS_NOT_IDENTICAL) reduce 401) ((T_IS_SMALLER_OR_EQUAL) reduce 401) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 401) ((T_OBJECT_OPERATOR) reduce 401) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 401) ((T_PIPE) reduce 401) ((T_POW) reduce 401) ((T_SL) reduce 401) ((T_SPACESHIP) reduce 401) ((T_SR) reduce 401) (("[") reduce 401) (("^") reduce 401) (("{") reduce 401) (("|") reduce 401)) 29843 ((("%") reduce 398) (("(") reduce 398) ((")") reduce 398) (("*") reduce 398) (("+") reduce 398) ((",") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_DOUBLE_ARROW) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 398) ((T_OBJECT_OPERATOR) reduce 398) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 398) ((T_PIPE) reduce 398) ((T_POW) reduce 398) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("[") reduce 398) (("^") reduce 398) (("{") reduce 398) (("|") reduce 398)) 29845 ((("%") reduce 513) (("(") reduce 513) ((")") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 29851 ((("%") reduce 233) ((")") reduce 233) (("*") reduce 233) (("+") reduce 233) ((",") reduce 233) (("-") reduce 233) ((".") reduce 233) (("/") reduce 233) (("<") reduce 233) ((">") reduce 233) (("?") reduce 233) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_BOOLEAN_AND) reduce 233) ((T_BOOLEAN_OR) reduce 233) ((T_COALESCE) reduce 233) ((T_DOUBLE_ARROW) reduce 233) ((T_INSTANCEOF) reduce 233) ((T_IS_EQUAL) reduce 233) ((T_IS_GREATER_OR_EQUAL) reduce 233) ((T_IS_IDENTICAL) reduce 233) ((T_IS_NOT_EQUAL) reduce 233) ((T_IS_NOT_IDENTICAL) reduce 233) ((T_IS_SMALLER_OR_EQUAL) reduce 233) ((T_LOGICAL_AND) reduce 233) ((T_LOGICAL_OR) reduce 233) ((T_LOGICAL_XOR) reduce 233) ((T_PIPE) reduce 233) ((T_POW) reduce 233) ((T_SL) reduce 233) ((T_SPACESHIP) reduce 233) ((T_SR) reduce 233) (("^") reduce 233) (("{") reduce 233) (("|") reduce 233)) 29856 ((("%") shift) ((")") reduce 404) (("*") shift) (("+") shift) ((",") reduce 404) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 404) (("|") shift)) 29860 ((("%") reduce 613) ((")") reduce 613) (("*") reduce 613) (("+") reduce 613) ((",") reduce 613) (("-") reduce 613) ((".") reduce 613) (("/") reduce 613) (("<") reduce 613) ((">") reduce 613) (("?") reduce 613) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_BOOLEAN_AND) reduce 613) ((T_BOOLEAN_OR) reduce 613) ((T_COALESCE) reduce 613) ((T_DOUBLE_ARROW) reduce 613) ((T_INSTANCEOF) reduce 613) ((T_IS_EQUAL) reduce 613) ((T_IS_GREATER_OR_EQUAL) reduce 613) ((T_IS_IDENTICAL) reduce 613) ((T_IS_NOT_EQUAL) reduce 613) ((T_IS_NOT_IDENTICAL) reduce 613) ((T_IS_SMALLER_OR_EQUAL) reduce 613) ((T_LOGICAL_AND) reduce 613) ((T_LOGICAL_OR) reduce 613) ((T_LOGICAL_XOR) reduce 613) ((T_PIPE) reduce 613) ((T_POW) reduce 613) ((T_SL) reduce 613) ((T_SPACESHIP) reduce 613) ((T_SR) reduce 613) (("^") reduce 613) (("{") reduce 613) (("|") reduce 613)) 29861 ((("%") reduce 467) ((")") reduce 467) (("*") reduce 467) (("+") reduce 467) ((",") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_DOUBLE_ARROW) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_PIPE) reduce 467) ((T_POW) shift) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("{") reduce 467) (("|") reduce 467)) 29862 ((("%") shift) ((")") reduce 616) (("*") shift) (("+") shift) ((",") reduce 616) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 616) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 616) (("|") shift)) 29863 ((("%") shift) ((")") reduce 615) (("*") shift) (("+") shift) ((",") reduce 615) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 615) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 615) (("|") shift)) 29864 ((("%") reduce 424) ((")") reduce 424) (("*") reduce 424) (("+") reduce 424) ((",") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_DOUBLE_ARROW) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("[") reduce 548) (("^") reduce 424) (("{") reduce 424) (("|") reduce 424)) 29866 ((("%") reduce 518) ((")") reduce 518) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_PIPE) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 29867 ((("%") reduce 474) ((")") reduce 474) (("*") reduce 474) (("+") reduce 474) ((",") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_DOUBLE_ARROW) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_PIPE) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("{") reduce 474) (("|") reduce 474)) 29868 ((("%") reduce 291) ((")") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("{") reduce 291) (("|") reduce 291)) 29872 ((("%") reduce 292) ((")") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("^") reduce 292) (("{") reduce 292) (("|") reduce 292)) 29873 ((("%") reduce 293) ((")") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("^") reduce 293) (("{") reduce 293) (("|") reduce 293)) 29876 ((("%") reduce 617) ((")") reduce 617) (("*") reduce 617) (("+") reduce 617) ((",") reduce 617) (("-") reduce 617) ((".") reduce 617) (("/") reduce 617) (("<") reduce 617) ((">") reduce 617) (("?") reduce 617) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_BOOLEAN_AND) reduce 617) ((T_BOOLEAN_OR) reduce 617) ((T_COALESCE) reduce 617) ((T_DOUBLE_ARROW) reduce 617) ((T_INSTANCEOF) reduce 617) ((T_IS_EQUAL) reduce 617) ((T_IS_GREATER_OR_EQUAL) reduce 617) ((T_IS_IDENTICAL) reduce 617) ((T_IS_NOT_EQUAL) reduce 617) ((T_IS_NOT_IDENTICAL) reduce 617) ((T_IS_SMALLER_OR_EQUAL) reduce 617) ((T_LOGICAL_AND) reduce 617) ((T_LOGICAL_OR) reduce 617) ((T_LOGICAL_XOR) reduce 617) ((T_PIPE) reduce 617) ((T_POW) reduce 617) ((T_SL) reduce 617) ((T_SPACESHIP) reduce 617) ((T_SR) reduce 617) (("^") reduce 617) (("{") reduce 617) (("|") reduce 617)) 29879 ((("%") reduce 614) ((")") reduce 614) (("*") reduce 614) (("+") reduce 614) ((",") reduce 614) (("-") reduce 614) ((".") reduce 614) (("/") reduce 614) (("<") reduce 614) ((">") reduce 614) (("?") reduce 614) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_BOOLEAN_AND) reduce 614) ((T_BOOLEAN_OR) reduce 614) ((T_COALESCE) reduce 614) ((T_DOUBLE_ARROW) reduce 614) ((T_INSTANCEOF) reduce 614) ((T_IS_EQUAL) reduce 614) ((T_IS_GREATER_OR_EQUAL) reduce 614) ((T_IS_IDENTICAL) reduce 614) ((T_IS_NOT_EQUAL) reduce 614) ((T_IS_NOT_IDENTICAL) reduce 614) ((T_IS_SMALLER_OR_EQUAL) reduce 614) ((T_LOGICAL_AND) reduce 614) ((T_LOGICAL_OR) reduce 614) ((T_LOGICAL_XOR) reduce 614) ((T_PIPE) reduce 614) ((T_POW) reduce 614) ((T_SL) reduce 614) ((T_SPACESHIP) reduce 614) ((T_SR) reduce 614) (("^") reduce 614) (("{") reduce 614) (("|") reduce 614)) 29880 ((("%") reduce 468) ((")") reduce 468) (("*") reduce 468) (("+") reduce 468) ((",") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_DOUBLE_ARROW) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_PIPE) reduce 468) ((T_POW) shift) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("{") reduce 468) (("|") reduce 468)) 29881 ((("%") reduce 426) ((")") reduce 426) (("*") reduce 426) (("+") reduce 426) ((",") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_DOUBLE_ARROW) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("[") reduce 548) (("^") reduce 426) (("{") reduce 426) (("|") reduce 426)) 29883 ((("%") reduce 408) ((")") reduce 408) (("*") reduce 408) (("+") reduce 408) ((",") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_DOUBLE_ARROW) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_PIPE) reduce 408) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("^") reduce 408) (("{") reduce 408) (("|") reduce 408)) 29884 ((("%") reduce 409) ((")") reduce 409) (("*") reduce 409) (("+") reduce 409) ((",") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_DOUBLE_ARROW) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_PIPE) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("{") reduce 409) (("|") reduce 409)) 29885 ((("%") reduce 296) ((")") reduce 296) (("*") reduce 296) (("+") reduce 296) ((",") reduce 296) (("-") reduce 296) ((".") reduce 296) (("/") reduce 296) (("<") reduce 296) ((">") reduce 296) (("?") reduce 296) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_BOOLEAN_AND) reduce 296) ((T_BOOLEAN_OR) reduce 296) ((T_COALESCE) reduce 296) ((T_DOUBLE_ARROW) reduce 296) ((T_INSTANCEOF) reduce 296) ((T_IS_EQUAL) reduce 296) ((T_IS_GREATER_OR_EQUAL) reduce 296) ((T_IS_IDENTICAL) reduce 296) ((T_IS_NOT_EQUAL) reduce 296) ((T_IS_NOT_IDENTICAL) reduce 296) ((T_IS_SMALLER_OR_EQUAL) reduce 296) ((T_LOGICAL_AND) reduce 296) ((T_LOGICAL_OR) reduce 296) ((T_LOGICAL_XOR) reduce 296) ((T_PIPE) reduce 296) ((T_POW) reduce 296) ((T_SL) reduce 296) ((T_SPACESHIP) reduce 296) ((T_SR) reduce 296) (("^") reduce 296) (("{") reduce 296) (("|") reduce 296)) 29890 ((("%") reduce 297) ((")") reduce 297) (("*") reduce 297) (("+") reduce 297) ((",") reduce 297) (("-") reduce 297) ((".") reduce 297) (("/") reduce 297) (("<") reduce 297) ((">") reduce 297) (("?") reduce 297) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_BOOLEAN_AND) reduce 297) ((T_BOOLEAN_OR) reduce 297) ((T_COALESCE) reduce 297) ((T_DOUBLE_ARROW) reduce 297) ((T_INSTANCEOF) reduce 297) ((T_IS_EQUAL) reduce 297) ((T_IS_GREATER_OR_EQUAL) reduce 297) ((T_IS_IDENTICAL) reduce 297) ((T_IS_NOT_EQUAL) reduce 297) ((T_IS_NOT_IDENTICAL) reduce 297) ((T_IS_SMALLER_OR_EQUAL) reduce 297) ((T_LOGICAL_AND) reduce 297) ((T_LOGICAL_OR) reduce 297) ((T_LOGICAL_XOR) reduce 297) ((T_PIPE) reduce 297) ((T_POW) reduce 297) ((T_SL) reduce 297) ((T_SPACESHIP) reduce 297) ((T_SR) reduce 297) (("^") reduce 297) (("{") reduce 297) (("|") reduce 297)) 29891 ((("%") reduce 460) (("(") reduce 556) ((")") reduce 460) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PIPE) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("[") reduce 549) (("^") reduce 460) (("{") reduce 460) (("|") reduce 460)) 29893 ((("%") reduce 298) ((")") reduce 298) (("*") reduce 298) (("+") reduce 298) ((",") reduce 298) (("-") reduce 298) ((".") reduce 298) (("/") reduce 298) (("<") reduce 298) ((">") reduce 298) (("?") reduce 298) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_BOOLEAN_AND) reduce 298) ((T_BOOLEAN_OR) reduce 298) ((T_COALESCE) reduce 298) ((T_DOUBLE_ARROW) reduce 298) ((T_INSTANCEOF) reduce 298) ((T_IS_EQUAL) reduce 298) ((T_IS_GREATER_OR_EQUAL) reduce 298) ((T_IS_IDENTICAL) reduce 298) ((T_IS_NOT_EQUAL) reduce 298) ((T_IS_NOT_IDENTICAL) reduce 298) ((T_IS_SMALLER_OR_EQUAL) reduce 298) ((T_LOGICAL_AND) reduce 298) ((T_LOGICAL_OR) reduce 298) ((T_LOGICAL_XOR) reduce 298) ((T_PIPE) reduce 298) ((T_POW) reduce 298) ((T_SL) reduce 298) ((T_SPACESHIP) reduce 298) ((T_SR) reduce 298) (("^") reduce 298) (("{") reduce 298) (("|") reduce 298)) 29894 ((("%") reduce 299) ((")") reduce 299) (("*") reduce 299) (("+") reduce 299) ((",") reduce 299) (("-") reduce 299) ((".") reduce 299) (("/") reduce 299) (("<") reduce 299) ((">") reduce 299) (("?") reduce 299) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_BOOLEAN_AND) reduce 299) ((T_BOOLEAN_OR) reduce 299) ((T_COALESCE) reduce 299) ((T_DOUBLE_ARROW) reduce 299) ((T_INSTANCEOF) reduce 299) ((T_IS_EQUAL) reduce 299) ((T_IS_GREATER_OR_EQUAL) reduce 299) ((T_IS_IDENTICAL) reduce 299) ((T_IS_NOT_EQUAL) reduce 299) ((T_IS_NOT_IDENTICAL) reduce 299) ((T_IS_SMALLER_OR_EQUAL) reduce 299) ((T_LOGICAL_AND) reduce 299) ((T_LOGICAL_OR) reduce 299) ((T_LOGICAL_XOR) reduce 299) ((T_PIPE) reduce 299) ((T_POW) reduce 299) ((T_SL) reduce 299) ((T_SPACESHIP) reduce 299) ((T_SR) reduce 299) (("^") reduce 299) (("{") reduce 299) (("|") reduce 299)) 29895 ((("%") reduce 472) ((")") reduce 472) (("*") reduce 472) (("+") reduce 472) ((",") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_DOUBLE_ARROW) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_PIPE) reduce 472) ((T_POW) shift) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("{") reduce 472) (("|") reduce 472)) 29896 ((("%") reduce 470) ((")") reduce 470) (("*") reduce 470) (("+") reduce 470) ((",") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_DOUBLE_ARROW) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_PIPE) reduce 470) ((T_POW) shift) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("{") reduce 470) (("|") reduce 470)) 29899 ((("%") reduce 519) (("(") reduce 519) ((")") reduce 519) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PIPE) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 29900 ((("%") reduce 475) ((")") reduce 475) (("*") reduce 475) (("+") reduce 475) ((",") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_DOUBLE_ARROW) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_PIPE) reduce 475) ((T_POW) shift) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("{") reduce 475) (("|") reduce 475)) 29901 ((("%") reduce 446) ((")") reduce 446) (("*") reduce 446) (("+") reduce 446) ((",") reduce 446) (("-") reduce 446) ((".") reduce 446) (("/") reduce 446) (("<") reduce 446) ((">") reduce 446) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_BOOLEAN_AND) reduce 446) ((T_BOOLEAN_OR) reduce 446) ((T_COALESCE) reduce 446) ((T_DOUBLE_ARROW) reduce 446) ((T_INSTANCEOF) reduce 446) ((T_IS_EQUAL) reduce 446) ((T_IS_GREATER_OR_EQUAL) reduce 446) ((T_IS_IDENTICAL) reduce 446) ((T_IS_NOT_EQUAL) reduce 446) ((T_IS_NOT_IDENTICAL) reduce 446) ((T_IS_SMALLER_OR_EQUAL) reduce 446) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_PIPE) reduce 446) ((T_POW) shift) ((T_SL) reduce 446) ((T_SPACESHIP) reduce 446) ((T_SR) reduce 446) (("^") reduce 446) (("{") reduce 446) (("|") reduce 446)) 29902 ((("%") reduce 445) ((")") reduce 445) (("*") reduce 445) (("+") reduce 445) ((",") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_DOUBLE_ARROW) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_PIPE) reduce 445) ((T_POW) shift) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("{") reduce 445) (("|") reduce 445)) 29904 ((("%") reduce 570) (("(") reduce 570) ((")") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) (("=") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AND_EQUAL) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_COALESCE_EQUAL) reduce 570) ((T_CONCAT_EQUAL) reduce 570) ((T_DEC) reduce 570) ((T_DIV_EQUAL) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INC) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_MINUS_EQUAL) reduce 570) ((T_MOD_EQUAL) reduce 570) ((T_MUL_EQUAL) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_OR_EQUAL) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_PLUS_EQUAL) reduce 570) ((T_POW) reduce 570) ((T_POW_EQUAL) reduce 570) ((T_SL) reduce 570) ((T_SL_EQUAL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) ((T_SR_EQUAL) reduce 570) ((T_XOR_EQUAL) reduce 570) (("[") reduce 570) (("^") reduce 570) (("{") reduce 570) (("|") reduce 570)) 29907 ((("%") reduce 569) (("(") reduce 569) ((")") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) (("=") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AND_EQUAL) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_COALESCE_EQUAL) reduce 569) ((T_CONCAT_EQUAL) reduce 569) ((T_DEC) reduce 569) ((T_DIV_EQUAL) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INC) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_MINUS_EQUAL) reduce 569) ((T_MOD_EQUAL) reduce 569) ((T_MUL_EQUAL) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_OR_EQUAL) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_PLUS_EQUAL) reduce 569) ((T_POW) reduce 569) ((T_POW_EQUAL) reduce 569) ((T_SL) reduce 569) ((T_SL_EQUAL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) ((T_SR_EQUAL) reduce 569) ((T_XOR_EQUAL) reduce 569) (("[") reduce 569) (("^") reduce 569) (("{") reduce 569) (("|") reduce 569)) 29909 ((("%") reduce 522) (("(") reduce 522) ((")") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_PIPE) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 29910 ((("%") reduce 447) ((")") reduce 447) (("*") reduce 447) (("+") reduce 447) ((",") reduce 447) (("-") reduce 447) ((".") reduce 447) (("/") reduce 447) (("<") reduce 447) ((">") reduce 447) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_BOOLEAN_AND) reduce 447) ((T_BOOLEAN_OR) reduce 447) ((T_COALESCE) reduce 447) ((T_DOUBLE_ARROW) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 447) ((T_IS_GREATER_OR_EQUAL) reduce 447) ((T_IS_IDENTICAL) reduce 447) ((T_IS_NOT_EQUAL) reduce 447) ((T_IS_NOT_IDENTICAL) reduce 447) ((T_IS_SMALLER_OR_EQUAL) reduce 447) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_PIPE) reduce 447) ((T_POW) shift) ((T_SL) reduce 447) ((T_SPACESHIP) reduce 447) ((T_SR) reduce 447) (("^") reduce 447) (("{") reduce 447) (("|") reduce 447)) 29911 ((("%") reduce 473) ((")") reduce 473) (("*") reduce 473) (("+") reduce 473) ((",") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_PIPE) reduce 473) ((T_POW) shift) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("{") reduce 473) (("|") reduce 473)) 29912 ((("%") shift) ((")") reduce 483) (("*") shift) (("+") shift) ((",") reduce 483) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 483) (("|") shift)) 29913 ((("%") reduce 469) ((")") reduce 469) (("*") reduce 469) (("+") reduce 469) ((",") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_PIPE) reduce 469) ((T_POW) shift) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("{") reduce 469) (("|") reduce 469)) 29914 ((("%") reduce 486) ((")") reduce 486) (("*") reduce 486) (("+") reduce 486) ((",") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_PIPE) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("^") reduce 486) (("{") reduce 486) (("|") reduce 486)) 29916 ((("%") reduce 526) ((")") reduce 526) (("*") reduce 526) (("+") reduce 526) ((",") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_PIPE) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("^") reduce 526) (("{") reduce 526) (("|") reduce 526)) 29918 ((("%") reduce 527) ((")") reduce 527) (("*") reduce 527) (("+") reduce 527) ((",") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_PIPE) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("^") reduce 527) (("{") reduce 527) (("|") reduce 527)) 29919 ((("%") reduce 525) ((")") reduce 525) (("*") reduce 525) (("+") reduce 525) ((",") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_PIPE) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("^") reduce 525) (("{") reduce 525) (("|") reduce 525)) 29920 ((("%") shift) ((")") reduce 619) (("*") shift) (("+") shift) ((",") reduce 619) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 619) (("|") shift)) 29921 ((("%") shift) ((")") reduce 618) (("*") shift) (("+") shift) ((",") reduce 618) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 618) (("|") shift)) 29922 ((("%") reduce 505) (("(") reduce 505) ((")") reduce 505) (("*") reduce 505) (("+") reduce 505) ((",") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("{") reduce 505) (("|") reduce 505)) 29923 ((("%") shift) ((")") reduce 478) (("*") shift) (("+") shift) ((",") reduce 478) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 478) (("|") shift)) 29924 ((("%") reduce 471) ((")") reduce 471) (("*") reduce 471) (("+") reduce 471) ((",") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_PIPE) reduce 471) ((T_POW) shift) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("{") reduce 471) (("|") reduce 471)) 29926 ((("%") reduce 94) (("(") reduce 94) ((")") reduce 94) (("*") reduce 94) (("+") reduce 94) ((",") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("{") reduce 94) (("|") reduce 94)) 29927 ((("%") reduce 93) (("(") reduce 93) ((")") reduce 93) (("*") reduce 93) (("+") reduce 93) ((",") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("{") reduce 93) (("|") reduce 93)) 29928 ((("%") reduce 95) (("(") reduce 95) ((")") reduce 95) (("*") reduce 95) (("+") reduce 95) ((",") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("{") reduce 95) (("|") reduce 95)) 29929 ((("%") reduce 509) (("(") reduce 509) ((")") reduce 509) (("*") reduce 509) (("+") reduce 509) ((",") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("{") reduce 509) (("|") reduce 509)) 29930 ((("%") reduce 92) (("(") reduce 92) ((")") reduce 92) (("*") reduce 92) (("+") reduce 92) ((",") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("{") reduce 92) (("|") reduce 92)) 29931 ((("%") reduce 400) (("(") reduce 400) ((")") reduce 400) (("*") reduce 400) (("+") reduce 400) ((",") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 400) ((T_OBJECT_OPERATOR) reduce 400) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 400) ((T_PIPE) reduce 400) ((T_POW) reduce 400) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("[") reduce 400) (("^") reduce 400) (("{") reduce 400) (("|") reduce 400)) 29934 ((("%") reduce 511) (("(") reduce 511) ((")") reduce 511) (("*") reduce 511) (("+") reduce 511) ((",") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("{") reduce 511) (("|") reduce 511)) 29935 ((("%") reduce 402) (("(") shift) ((")") reduce 402) (("*") reduce 402) (("+") reduce 402) ((",") reduce 402) (("-") reduce 402) ((".") reduce 402) (("/") reduce 402) (("<") reduce 402) ((">") reduce 402) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_INSTANCEOF) reduce 402) ((T_IS_EQUAL) reduce 402) ((T_IS_GREATER_OR_EQUAL) reduce 402) ((T_IS_IDENTICAL) reduce 402) ((T_IS_NOT_EQUAL) reduce 402) ((T_IS_NOT_IDENTICAL) reduce 402) ((T_IS_SMALLER_OR_EQUAL) reduce 402) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_PIPE) reduce 402) ((T_POW) reduce 402) ((T_SL) reduce 402) ((T_SPACESHIP) reduce 402) ((T_SR) reduce 402) (("^") reduce 402) (("{") reduce 402) (("|") reduce 402)) 29936 ((("%") reduce 510) (("(") reduce 510) ((")") reduce 510) (("*") reduce 510) (("+") reduce 510) ((",") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("{") reduce 510) (("|") reduce 510)) 29937 ((("%") reduce 512) (("(") reduce 512) ((")") reduce 512) (("*") reduce 512) (("+") reduce 512) ((",") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("{") reduce 512) (("|") reduce 512)) 29938 ((("%") reduce 573) (("(") reduce 573) ((")") reduce 573) (("*") reduce 573) (("+") reduce 573) ((",") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("{") reduce 573) (("|") reduce 573)) 29944 ((("%") reduce 574) (("(") reduce 574) ((")") reduce 574) (("*") reduce 574) (("+") reduce 574) ((",") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("{") reduce 574) (("|") reduce 574)) 29945 ((("%") reduce 578) (("(") reduce 578) ((")") reduce 578) (("*") reduce 578) (("+") reduce 578) ((",") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("{") reduce 578) (("|") reduce 578)) 29946 ((("%") reduce 575) (("(") reduce 575) ((")") reduce 575) (("*") reduce 575) (("+") reduce 575) ((",") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("{") reduce 575) (("|") reduce 575)) 29947 ((("%") reduce 576) (("(") reduce 576) ((")") reduce 576) (("*") reduce 576) (("+") reduce 576) ((",") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("{") reduce 576) (("|") reduce 576)) 29948 ((("%") reduce 399) (("(") reduce 399) ((")") reduce 399) (("*") reduce 399) (("+") reduce 399) ((",") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_INSTANCEOF) reduce 399) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 399) ((T_OBJECT_OPERATOR) reduce 399) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 399) ((T_PIPE) reduce 399) ((T_POW) reduce 399) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("[") reduce 399) (("^") reduce 399) (("{") reduce 399) (("|") reduce 399)) 29950 ((("%") reduce 577) (("(") reduce 577) ((")") reduce 577) (("*") reduce 577) (("+") reduce 577) ((",") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("{") reduce 577) (("|") reduce 577)) 29951 ((("%") reduce 401) (("(") reduce 401) ((")") reduce 401) (("*") reduce 401) (("+") reduce 401) ((",") reduce 401) (("-") reduce 401) ((".") reduce 401) (("/") reduce 401) (("<") reduce 401) ((">") reduce 401) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_INSTANCEOF) reduce 401) ((T_IS_EQUAL) reduce 401) ((T_IS_GREATER_OR_EQUAL) reduce 401) ((T_IS_IDENTICAL) reduce 401) ((T_IS_NOT_EQUAL) reduce 401) ((T_IS_NOT_IDENTICAL) reduce 401) ((T_IS_SMALLER_OR_EQUAL) reduce 401) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 401) ((T_OBJECT_OPERATOR) reduce 401) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 401) ((T_PIPE) reduce 401) ((T_POW) reduce 401) ((T_SL) reduce 401) ((T_SPACESHIP) reduce 401) ((T_SR) reduce 401) (("[") reduce 401) (("^") reduce 401) (("{") reduce 401) (("|") reduce 401)) 29959 ((("%") reduce 398) (("(") reduce 398) ((")") reduce 398) (("*") reduce 398) (("+") reduce 398) ((",") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 398) ((T_OBJECT_OPERATOR) reduce 398) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 398) ((T_PIPE) reduce 398) ((T_POW) reduce 398) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("[") reduce 398) (("^") reduce 398) (("{") reduce 398) (("|") reduce 398)) 29961 ((("%") reduce 513) (("(") reduce 513) ((")") reduce 513) (("*") reduce 513) (("+") reduce 513) ((",") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("{") reduce 513) (("|") reduce 513)) 29967 ((("%") reduce 233) ((")") reduce 233) (("*") reduce 233) (("+") reduce 233) ((",") reduce 233) (("-") reduce 233) ((".") reduce 233) (("/") reduce 233) (("<") reduce 233) ((">") reduce 233) (("?") reduce 233) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_BOOLEAN_AND) reduce 233) ((T_BOOLEAN_OR) reduce 233) ((T_COALESCE) reduce 233) ((T_INSTANCEOF) reduce 233) ((T_IS_EQUAL) reduce 233) ((T_IS_GREATER_OR_EQUAL) reduce 233) ((T_IS_IDENTICAL) reduce 233) ((T_IS_NOT_EQUAL) reduce 233) ((T_IS_NOT_IDENTICAL) reduce 233) ((T_IS_SMALLER_OR_EQUAL) reduce 233) ((T_LOGICAL_AND) reduce 233) ((T_LOGICAL_OR) reduce 233) ((T_LOGICAL_XOR) reduce 233) ((T_PIPE) reduce 233) ((T_POW) reduce 233) ((T_SL) reduce 233) ((T_SPACESHIP) reduce 233) ((T_SR) reduce 233) (("^") reduce 233) (("{") reduce 233) (("|") reduce 233)) 29972 ((("%") shift) ((")") reduce 404) (("*") shift) (("+") shift) ((",") reduce 404) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 404) (("|") shift)) 29976 ((("%") reduce 613) ((")") reduce 613) (("*") reduce 613) (("+") reduce 613) ((",") reduce 613) (("-") reduce 613) ((".") reduce 613) (("/") reduce 613) (("<") reduce 613) ((">") reduce 613) (("?") reduce 613) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_BOOLEAN_AND) reduce 613) ((T_BOOLEAN_OR) reduce 613) ((T_COALESCE) reduce 613) ((T_INSTANCEOF) reduce 613) ((T_IS_EQUAL) reduce 613) ((T_IS_GREATER_OR_EQUAL) reduce 613) ((T_IS_IDENTICAL) reduce 613) ((T_IS_NOT_EQUAL) reduce 613) ((T_IS_NOT_IDENTICAL) reduce 613) ((T_IS_SMALLER_OR_EQUAL) reduce 613) ((T_LOGICAL_AND) reduce 613) ((T_LOGICAL_OR) reduce 613) ((T_LOGICAL_XOR) reduce 613) ((T_PIPE) reduce 613) ((T_POW) reduce 613) ((T_SL) reduce 613) ((T_SPACESHIP) reduce 613) ((T_SR) reduce 613) (("^") reduce 613) (("{") reduce 613) (("|") reduce 613)) 29977 ((("%") reduce 467) ((")") reduce 467) (("*") reduce 467) (("+") reduce 467) ((",") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_PIPE) reduce 467) ((T_POW) shift) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("{") reduce 467) (("|") reduce 467)) 29978 ((("%") shift) ((")") reduce 616) (("*") shift) (("+") shift) ((",") reduce 616) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 616) (("|") shift)) 29979 ((("%") shift) ((")") reduce 615) (("*") shift) (("+") shift) ((",") reduce 615) (("-") shift) ((".") shift) (("/") shift) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("{") reduce 615) (("|") shift)) 29980 ((("%") reduce 424) ((")") reduce 424) (("*") reduce 424) (("+") reduce 424) ((",") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("[") reduce 548) (("^") reduce 424) (("{") reduce 424) (("|") reduce 424)) 29982 ((("%") reduce 518) ((")") reduce 518) (("*") reduce 518) (("+") reduce 518) ((",") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_PIPE) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("^") reduce 518) (("{") reduce 518) (("|") reduce 518)) 29983 ((("%") reduce 474) ((")") reduce 474) (("*") reduce 474) (("+") reduce 474) ((",") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_PIPE) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("{") reduce 474) (("|") reduce 474)) 29984 ((("%") reduce 291) ((")") reduce 291) (("*") reduce 291) (("+") reduce 291) ((",") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("{") reduce 291) (("|") reduce 291)) 29988 ((("%") reduce 292) ((")") reduce 292) (("*") reduce 292) (("+") reduce 292) ((",") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("^") reduce 292) (("{") reduce 292) (("|") reduce 292)) 29989 ((("%") reduce 293) ((")") reduce 293) (("*") reduce 293) (("+") reduce 293) ((",") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("^") reduce 293) (("{") reduce 293) (("|") reduce 293)) 29992 ((("%") reduce 617) ((")") reduce 617) (("*") reduce 617) (("+") reduce 617) ((",") reduce 617) (("-") reduce 617) ((".") reduce 617) (("/") reduce 617) (("<") reduce 617) ((">") reduce 617) (("?") reduce 617) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_BOOLEAN_AND) reduce 617) ((T_BOOLEAN_OR) reduce 617) ((T_COALESCE) reduce 617) ((T_INSTANCEOF) reduce 617) ((T_IS_EQUAL) reduce 617) ((T_IS_GREATER_OR_EQUAL) reduce 617) ((T_IS_IDENTICAL) reduce 617) ((T_IS_NOT_EQUAL) reduce 617) ((T_IS_NOT_IDENTICAL) reduce 617) ((T_IS_SMALLER_OR_EQUAL) reduce 617) ((T_LOGICAL_AND) reduce 617) ((T_LOGICAL_OR) reduce 617) ((T_LOGICAL_XOR) reduce 617) ((T_PIPE) reduce 617) ((T_POW) reduce 617) ((T_SL) reduce 617) ((T_SPACESHIP) reduce 617) ((T_SR) reduce 617) (("^") reduce 617) (("{") reduce 617) (("|") reduce 617)) 29995 ((("%") reduce 614) ((")") reduce 614) (("*") reduce 614) (("+") reduce 614) ((",") reduce 614) (("-") reduce 614) ((".") reduce 614) (("/") reduce 614) (("<") reduce 614) ((">") reduce 614) (("?") reduce 614) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_BOOLEAN_AND) reduce 614) ((T_BOOLEAN_OR) reduce 614) ((T_COALESCE) reduce 614) ((T_INSTANCEOF) reduce 614) ((T_IS_EQUAL) reduce 614) ((T_IS_GREATER_OR_EQUAL) reduce 614) ((T_IS_IDENTICAL) reduce 614) ((T_IS_NOT_EQUAL) reduce 614) ((T_IS_NOT_IDENTICAL) reduce 614) ((T_IS_SMALLER_OR_EQUAL) reduce 614) ((T_LOGICAL_AND) reduce 614) ((T_LOGICAL_OR) reduce 614) ((T_LOGICAL_XOR) reduce 614) ((T_PIPE) reduce 614) ((T_POW) reduce 614) ((T_SL) reduce 614) ((T_SPACESHIP) reduce 614) ((T_SR) reduce 614) (("^") reduce 614) (("{") reduce 614) (("|") reduce 614)) 29996 ((("%") reduce 468) ((")") reduce 468) (("*") reduce 468) (("+") reduce 468) ((",") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_PIPE) reduce 468) ((T_POW) shift) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("{") reduce 468) (("|") reduce 468)) 29997 ((("%") reduce 426) ((")") reduce 426) (("*") reduce 426) (("+") reduce 426) ((",") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("[") reduce 548) (("^") reduce 426) (("{") reduce 426) (("|") reduce 426)) 29999 ((("%") reduce 408) ((")") reduce 408) (("*") reduce 408) (("+") reduce 408) ((",") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_PIPE) reduce 408) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("^") reduce 408) (("{") reduce 408) (("|") reduce 408)) 30000 ((("%") reduce 409) ((")") reduce 409) (("*") reduce 409) (("+") reduce 409) ((",") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_PIPE) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("{") reduce 409) (("|") reduce 409)) 30001 ((("%") reduce 296) ((")") reduce 296) (("*") reduce 296) (("+") reduce 296) ((",") reduce 296) (("-") reduce 296) ((".") reduce 296) (("/") reduce 296) (("<") reduce 296) ((">") reduce 296) (("?") reduce 296) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_BOOLEAN_AND) reduce 296) ((T_BOOLEAN_OR) reduce 296) ((T_COALESCE) reduce 296) ((T_INSTANCEOF) reduce 296) ((T_IS_EQUAL) reduce 296) ((T_IS_GREATER_OR_EQUAL) reduce 296) ((T_IS_IDENTICAL) reduce 296) ((T_IS_NOT_EQUAL) reduce 296) ((T_IS_NOT_IDENTICAL) reduce 296) ((T_IS_SMALLER_OR_EQUAL) reduce 296) ((T_LOGICAL_AND) reduce 296) ((T_LOGICAL_OR) reduce 296) ((T_LOGICAL_XOR) reduce 296) ((T_PIPE) reduce 296) ((T_POW) reduce 296) ((T_SL) reduce 296) ((T_SPACESHIP) reduce 296) ((T_SR) reduce 296) (("^") reduce 296) (("{") reduce 296) (("|") reduce 296)) 30006 ((("%") reduce 297) ((")") reduce 297) (("*") reduce 297) (("+") reduce 297) ((",") reduce 297) (("-") reduce 297) ((".") reduce 297) (("/") reduce 297) (("<") reduce 297) ((">") reduce 297) (("?") reduce 297) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_BOOLEAN_AND) reduce 297) ((T_BOOLEAN_OR) reduce 297) ((T_COALESCE) reduce 297) ((T_INSTANCEOF) reduce 297) ((T_IS_EQUAL) reduce 297) ((T_IS_GREATER_OR_EQUAL) reduce 297) ((T_IS_IDENTICAL) reduce 297) ((T_IS_NOT_EQUAL) reduce 297) ((T_IS_NOT_IDENTICAL) reduce 297) ((T_IS_SMALLER_OR_EQUAL) reduce 297) ((T_LOGICAL_AND) reduce 297) ((T_LOGICAL_OR) reduce 297) ((T_LOGICAL_XOR) reduce 297) ((T_PIPE) reduce 297) ((T_POW) reduce 297) ((T_SL) reduce 297) ((T_SPACESHIP) reduce 297) ((T_SR) reduce 297) (("^") reduce 297) (("{") reduce 297) (("|") reduce 297)) 30007 ((("%") reduce 460) (("(") reduce 556) ((")") reduce 460) (("*") reduce 460) (("+") reduce 460) ((",") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PIPE) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("[") reduce 549) (("^") reduce 460) (("{") reduce 460) (("|") reduce 460)) 30009 ((("%") reduce 298) ((")") reduce 298) (("*") reduce 298) (("+") reduce 298) ((",") reduce 298) (("-") reduce 298) ((".") reduce 298) (("/") reduce 298) (("<") reduce 298) ((">") reduce 298) (("?") reduce 298) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_BOOLEAN_AND) reduce 298) ((T_BOOLEAN_OR) reduce 298) ((T_COALESCE) reduce 298) ((T_INSTANCEOF) reduce 298) ((T_IS_EQUAL) reduce 298) ((T_IS_GREATER_OR_EQUAL) reduce 298) ((T_IS_IDENTICAL) reduce 298) ((T_IS_NOT_EQUAL) reduce 298) ((T_IS_NOT_IDENTICAL) reduce 298) ((T_IS_SMALLER_OR_EQUAL) reduce 298) ((T_LOGICAL_AND) reduce 298) ((T_LOGICAL_OR) reduce 298) ((T_LOGICAL_XOR) reduce 298) ((T_PIPE) reduce 298) ((T_POW) reduce 298) ((T_SL) reduce 298) ((T_SPACESHIP) reduce 298) ((T_SR) reduce 298) (("^") reduce 298) (("{") reduce 298) (("|") reduce 298)) 30010 ((("%") reduce 299) ((")") reduce 299) (("*") reduce 299) (("+") reduce 299) ((",") reduce 299) (("-") reduce 299) ((".") reduce 299) (("/") reduce 299) (("<") reduce 299) ((">") reduce 299) (("?") reduce 299) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_BOOLEAN_AND) reduce 299) ((T_BOOLEAN_OR) reduce 299) ((T_COALESCE) reduce 299) ((T_INSTANCEOF) reduce 299) ((T_IS_EQUAL) reduce 299) ((T_IS_GREATER_OR_EQUAL) reduce 299) ((T_IS_IDENTICAL) reduce 299) ((T_IS_NOT_EQUAL) reduce 299) ((T_IS_NOT_IDENTICAL) reduce 299) ((T_IS_SMALLER_OR_EQUAL) reduce 299) ((T_LOGICAL_AND) reduce 299) ((T_LOGICAL_OR) reduce 299) ((T_LOGICAL_XOR) reduce 299) ((T_PIPE) reduce 299) ((T_POW) reduce 299) ((T_SL) reduce 299) ((T_SPACESHIP) reduce 299) ((T_SR) reduce 299) (("^") reduce 299) (("{") reduce 299) (("|") reduce 299)) 30011 ((("%") reduce 472) ((")") reduce 472) (("*") reduce 472) (("+") reduce 472) ((",") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_PIPE) reduce 472) ((T_POW) shift) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("{") reduce 472) (("|") reduce 472)) 30012 ((("%") reduce 470) ((")") reduce 470) (("*") reduce 470) (("+") reduce 470) ((",") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_PIPE) reduce 470) ((T_POW) shift) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("{") reduce 470) (("|") reduce 470)) 30015 ((("%") reduce 519) (("(") reduce 519) ((")") reduce 519) (("*") reduce 519) (("+") reduce 519) ((",") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PIPE) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("{") reduce 519) (("|") reduce 519)) 30016 ((("%") reduce 475) ((")") reduce 475) (("*") reduce 475) (("+") reduce 475) ((",") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_PIPE) reduce 475) ((T_POW) shift) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("{") reduce 475) (("|") reduce 475)) 30017 ((("%") reduce 446) ((")") reduce 446) (("*") reduce 446) (("+") reduce 446) ((",") reduce 446) (("-") reduce 446) ((".") reduce 446) (("/") reduce 446) (("<") reduce 446) ((">") reduce 446) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_BOOLEAN_AND) reduce 446) ((T_BOOLEAN_OR) reduce 446) ((T_COALESCE) reduce 446) ((T_INSTANCEOF) reduce 446) ((T_IS_EQUAL) reduce 446) ((T_IS_GREATER_OR_EQUAL) reduce 446) ((T_IS_IDENTICAL) reduce 446) ((T_IS_NOT_EQUAL) reduce 446) ((T_IS_NOT_IDENTICAL) reduce 446) ((T_IS_SMALLER_OR_EQUAL) reduce 446) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_PIPE) reduce 446) ((T_POW) shift) ((T_SL) reduce 446) ((T_SPACESHIP) reduce 446) ((T_SR) reduce 446) (("^") reduce 446) (("{") reduce 446) (("|") reduce 446)) 30018 ((("%") reduce 445) ((")") reduce 445) (("*") reduce 445) (("+") reduce 445) ((",") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_PIPE) reduce 445) ((T_POW) shift) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("{") reduce 445) (("|") reduce 445)) 30020 ((("%") reduce 570) (("(") reduce 570) ((")") reduce 570) (("*") reduce 570) (("+") reduce 570) ((",") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) (("<") reduce 570) (("=") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AND_EQUAL) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_COALESCE_EQUAL) reduce 570) ((T_CONCAT_EQUAL) reduce 570) ((T_DEC) reduce 570) ((T_DIV_EQUAL) reduce 570) ((T_INC) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_MINUS_EQUAL) reduce 570) ((T_MOD_EQUAL) reduce 570) ((T_MUL_EQUAL) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_OR_EQUAL) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_PLUS_EQUAL) reduce 570) ((T_POW) reduce 570) ((T_POW_EQUAL) reduce 570) ((T_SL) reduce 570) ((T_SL_EQUAL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) ((T_SR_EQUAL) reduce 570) ((T_XOR_EQUAL) reduce 570) (("[") reduce 570) (("^") reduce 570) (("{") reduce 570) (("|") reduce 570)) 30023 ((("%") reduce 569) (("(") reduce 569) ((")") reduce 569) (("*") reduce 569) (("+") reduce 569) ((",") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) (("<") reduce 569) (("=") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AND_EQUAL) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_COALESCE_EQUAL) reduce 569) ((T_CONCAT_EQUAL) reduce 569) ((T_DEC) reduce 569) ((T_DIV_EQUAL) reduce 569) ((T_INC) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_MINUS_EQUAL) reduce 569) ((T_MOD_EQUAL) reduce 569) ((T_MUL_EQUAL) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_OR_EQUAL) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_PLUS_EQUAL) reduce 569) ((T_POW) reduce 569) ((T_POW_EQUAL) reduce 569) ((T_SL) reduce 569) ((T_SL_EQUAL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) ((T_SR_EQUAL) reduce 569) ((T_XOR_EQUAL) reduce 569) (("[") reduce 569) (("^") reduce 569) (("{") reduce 569) (("|") reduce 569)) 30025 ((("%") reduce 522) (("(") reduce 522) ((")") reduce 522) (("*") reduce 522) (("+") reduce 522) ((",") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_PIPE) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("{") reduce 522) (("|") reduce 522)) 30026 ((("%") reduce 447) ((")") reduce 447) (("*") reduce 447) (("+") reduce 447) ((",") reduce 447) (("-") reduce 447) ((".") reduce 447) (("/") reduce 447) (("<") reduce 447) ((">") reduce 447) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_BOOLEAN_AND) reduce 447) ((T_BOOLEAN_OR) reduce 447) ((T_COALESCE) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 447) ((T_IS_GREATER_OR_EQUAL) reduce 447) ((T_IS_IDENTICAL) reduce 447) ((T_IS_NOT_EQUAL) reduce 447) ((T_IS_NOT_IDENTICAL) reduce 447) ((T_IS_SMALLER_OR_EQUAL) reduce 447) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_PIPE) reduce 447) ((T_POW) shift) ((T_SL) reduce 447) ((T_SPACESHIP) reduce 447) ((T_SR) reduce 447) (("^") reduce 447) (("{") reduce 447) (("|") reduce 447)) 30028 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 288) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 288) ((T_ELLIPSIS) reduce 288) ((T_VARIABLE) reduce 288)) 30029 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 280) ((T_ELLIPSIS) reduce 280) ((T_VARIABLE) reduce 280)) 30030 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 281) ((T_ELLIPSIS) reduce 281) ((T_VARIABLE) reduce 281)) 30031 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_ELLIPSIS) reduce 94) ((T_VARIABLE) reduce 94)) 30032 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_ELLIPSIS) reduce 93) ((T_VARIABLE) reduce 93)) 30033 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_ELLIPSIS) reduce 95) ((T_VARIABLE) reduce 95)) 30034 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_ELLIPSIS) reduce 92) ((T_VARIABLE) reduce 92)) 30035 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 282) ((T_ELLIPSIS) reduce 282) ((T_VARIABLE) reduce 282)) 30036 (((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 277) ((T_ELLIPSIS) reduce 277) ((T_VARIABLE) reduce 277)) 30037 ((("(") reduce 258) ((")") reduce 126) (("?") reduce 258) ((T_ABSTRACT) shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 258) ((T_ARRAY) reduce 258) ((T_ATTRIBUTE) shift) ((T_CALLABLE) reduce 258) ((T_ELLIPSIS) reduce 258) ((T_FINAL) shift) ((T_NAME_FULLY_QUALIFIED) reduce 258) ((T_NAME_QUALIFIED) reduce 258) ((T_NAME_RELATIVE) reduce 258) ((T_PRIVATE) shift) ((T_PRIVATE_SET) shift) ((T_PROTECTED) shift) ((T_PROTECTED_SET) shift) ((T_PUBLIC) shift) ((T_PUBLIC_SET) shift) ((T_READONLY) shift) ((T_STATIC) shift) ((T_STRING) reduce 258) ((T_VARIABLE) reduce 258)) 30038 (((")") reduce 252)) 30039 (((")") reduce 255) ((",") reduce 255)) 30040 ((("(") reduce 351) (("?") reduce 351) ((T_ABSTRACT) reduce 351) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 351) ((T_ARRAY) reduce 351) ((T_CALLABLE) reduce 351) ((T_ELLIPSIS) reduce 351) ((T_FINAL) reduce 351) ((T_NAME_FULLY_QUALIFIED) reduce 351) ((T_NAME_QUALIFIED) reduce 351) ((T_NAME_RELATIVE) reduce 351) ((T_PRIVATE) reduce 351) ((T_PRIVATE_SET) reduce 351) ((T_PROTECTED) reduce 351) ((T_PROTECTED_SET) reduce 351) ((T_PUBLIC) reduce 351) ((T_PUBLIC_SET) reduce 351) ((T_READONLY) reduce 351) ((T_STATIC) reduce 351) ((T_STRING) reduce 351) ((T_VARIABLE) reduce 351)) 30041 ((("(") reduce 102) (("?") reduce 102) ((T_ABSTRACT) reduce 102) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 102) ((T_ARRAY) reduce 102) ((T_ATTRIBUTE) reduce 102) ((T_CALLABLE) reduce 102) ((T_ELLIPSIS) reduce 102) ((T_FINAL) reduce 102) ((T_NAME_FULLY_QUALIFIED) reduce 102) ((T_NAME_QUALIFIED) reduce 102) ((T_NAME_RELATIVE) reduce 102) ((T_PRIVATE) reduce 102) ((T_PRIVATE_SET) reduce 102) ((T_PROTECTED) reduce 102) ((T_PROTECTED_SET) reduce 102) ((T_PUBLIC) reduce 102) ((T_PUBLIC_SET) reduce 102) ((T_READONLY) reduce 102) ((T_STATIC) reduce 102) ((T_STRING) reduce 102) ((T_VARIABLE) reduce 102)) 30042 (((")") reduce 256) ((",") reduce 256)) 30045 ((("(") reduce 100) (("?") reduce 100) ((T_ABSTRACT) reduce 100) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 100) ((T_ARRAY) reduce 100) ((T_ATTRIBUTE) reduce 100) ((T_CALLABLE) reduce 100) ((T_ELLIPSIS) reduce 100) ((T_FINAL) reduce 100) ((T_NAME_FULLY_QUALIFIED) reduce 100) ((T_NAME_QUALIFIED) reduce 100) ((T_NAME_RELATIVE) reduce 100) ((T_PRIVATE) reduce 100) ((T_PRIVATE_SET) reduce 100) ((T_PROTECTED) reduce 100) ((T_PROTECTED_SET) reduce 100) ((T_PUBLIC) reduce 100) ((T_PUBLIC_SET) reduce 100) ((T_READONLY) reduce 100) ((T_STATIC) reduce 100) ((T_STRING) reduce 100) ((T_VARIABLE) reduce 100)) 30047 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((";") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 30048 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((";") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 30052 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) ((";") reduce 292) (("<") reduce 292) (("=") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AND_EQUAL) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_COALESCE_EQUAL) reduce 292) ((T_CONCAT_EQUAL) reduce 292) ((T_DEC) reduce 292) ((T_DIV_EQUAL) reduce 292) ((T_INC) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_MINUS_EQUAL) reduce 292) ((T_MOD_EQUAL) reduce 292) ((T_MUL_EQUAL) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_OR_EQUAL) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_PLUS_EQUAL) reduce 292) ((T_POW) reduce 292) ((T_POW_EQUAL) reduce 292) ((T_SL) reduce 292) ((T_SL_EQUAL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) ((T_SR_EQUAL) reduce 292) ((T_XOR_EQUAL) reduce 292) (("[") reduce 292) (("^") reduce 292) (("|") reduce 292)) 30053 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) ((";") reduce 293) (("<") reduce 293) (("=") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AND_EQUAL) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_COALESCE_EQUAL) reduce 293) ((T_CONCAT_EQUAL) reduce 293) ((T_DEC) reduce 293) ((T_DIV_EQUAL) reduce 293) ((T_INC) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_MINUS_EQUAL) reduce 293) ((T_MOD_EQUAL) reduce 293) ((T_MUL_EQUAL) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_OR_EQUAL) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_PLUS_EQUAL) reduce 293) ((T_POW) reduce 293) ((T_POW_EQUAL) reduce 293) ((T_SL) reduce 293) ((T_SL_EQUAL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) ((T_SR_EQUAL) reduce 293) ((T_XOR_EQUAL) reduce 293) (("[") reduce 293) (("^") reduce 293) (("|") reduce 293)) 30056 (((T_ATTRIBUTE) reduce 100) ((T_FN) reduce 100) ((T_FUNCTION) reduce 100) ((T_STATIC) reduce 100)) 30058 ((("!") reduce 147) (("\"") reduce 147) (($) reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONST) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_EMPTY) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAMESPACE) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_USE) reduce 147) ((T_VARIABLE) reduce 147) ((T_VOID_CAST) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 30060 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) ((";") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_PIPE) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("|") reduce 74)) 30061 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) ((";") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_PIPE) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("|") reduce 52)) 30062 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) ((";") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_PIPE) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("|") reduce 27)) 30063 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) ((";") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_PIPE) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("|") reduce 51)) 30064 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) ((";") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_PIPE) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("|") reduce 53)) 30065 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) ((";") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_PIPE) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("|") reduce 49)) 30066 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) ((";") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_PIPE) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("|") reduce 29)) 30067 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) ((";") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_PIPE) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("|") reduce 59)) 30068 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) ((";") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_PIPE) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("|") reduce 60)) 30069 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) ((";") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_PIPE) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("|") reduce 11)) 30070 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) ((";") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_PIPE) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("|") reduce 42)) 30071 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) ((";") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_PIPE) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("|") reduce 39)) 30072 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) ((";") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_PIPE) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("|") reduce 25)) 30073 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) ((";") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_PIPE) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("|") reduce 50)) 30074 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) ((";") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_PIPE) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("|") reduce 66)) 30075 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) ((";") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_PIPE) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("|") reduce 18)) 30076 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) ((";") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_PIPE) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("|") reduce 17)) 30077 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) ((";") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_PIPE) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("|") reduce 15)) 30078 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) ((";") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_PIPE) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("|") reduce 14)) 30079 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) ((";") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_PIPE) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("|") reduce 38)) 30080 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) ((";") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_PIPE) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("|") reduce 26)) 30081 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) ((";") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_PIPE) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("|") reduce 22)) 30082 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) ((";") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_PIPE) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("|") reduce 24)) 30083 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) ((";") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_PIPE) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("|") reduce 16)) 30084 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) ((";") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_PIPE) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("|") reduce 48)) 30085 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) ((";") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_PIPE) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("|") reduce 20)) 30086 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) ((";") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_PIPE) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("|") reduce 70)) 30087 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) ((";") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_PIPE) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("|") reduce 3)) 30088 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) ((";") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_PIPE) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("|") reduce 12)) 30089 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) ((";") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_PIPE) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("|") reduce 54)) 30090 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) ((";") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_PIPE) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("|") reduce 65)) 30091 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) ((";") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_PIPE) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("|") reduce 75)) 30092 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) ((";") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_PIPE) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("|") reduce 30)) 30093 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) ((";") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_PIPE) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("|") reduce 68)) 30094 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) ((";") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_PIPE) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("|") reduce 21)) 30095 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) ((";") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_PIPE) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("|") reduce 23)) 30096 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) ((";") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_PIPE) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("|") reduce 41)) 30097 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) ((";") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_PIPE) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("|") reduce 62)) 30098 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) ((";") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_PIPE) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("|") reduce 34)) 30099 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) ((";") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_PIPE) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("|") reduce 40)) 30100 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) ((";") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_PIPE) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("|") reduce 13)) 30101 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) ((";") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_PIPE) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("|") reduce 55)) 30102 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) ((";") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_PIPE) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("|") reduce 1)) 30103 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) ((";") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_PIPE) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("|") reduce 2)) 30104 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) ((";") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_PIPE) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("|") reduce 9)) 30105 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) ((";") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_PIPE) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("|") reduce 33)) 30106 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) ((";") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_PIPE) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("|") reduce 58)) 30107 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) ((";") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_PIPE) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("|") reduce 37)) 30108 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) ((";") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_PIPE) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("|") reduce 64)) 30109 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) ((";") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_PIPE) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("|") reduce 46)) 30110 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) ((";") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_PIPE) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("|") reduce 8)) 30111 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) ((";") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_PIPE) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("|") reduce 6)) 30112 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) ((";") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_PIPE) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("|") reduce 7)) 30113 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) ((";") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_PIPE) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("|") reduce 69)) 30114 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) ((";") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_PIPE) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("|") reduce 63)) 30115 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) ((";") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_PIPE) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("|") reduce 56)) 30116 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) ((";") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_PIPE) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("|") reduce 10)) 30117 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) ((";") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_PIPE) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("|") reduce 67)) 30118 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) ((";") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_PIPE) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("|") reduce 44)) 30119 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) ((";") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_PIPE) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("|") reduce 76)) 30120 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) ((";") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_PIPE) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("|") reduce 71)) 30121 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) ((";") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_PIPE) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("|") reduce 77)) 30122 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) ((";") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_PIPE) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("|") reduce 78)) 30123 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) ((";") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_PIPE) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("|") reduce 79)) 30124 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) ((";") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_PIPE) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("|") reduce 4)) 30125 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) ((";") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_PIPE) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("|") reduce 5)) 30126 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) ((";") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_PIPE) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("|") reduce 43)) 30127 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) ((";") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_PIPE) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("|") reduce 73)) 30128 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) ((";") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_PIPE) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("|") reduce 82)) 30129 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) ((";") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_PIPE) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("|") reduce 47)) 30130 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) ((";") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_PIPE) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("|") reduce 31)) 30131 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) ((";") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_PIPE) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("|") reduce 57)) 30132 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) ((";") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_PIPE) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("|") reduce 61)) 30133 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) ((";") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_PIPE) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("|") reduce 28)) 30134 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) ((";") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_PIPE) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("|") reduce 36)) 30135 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) ((";") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_PIPE) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("|") reduce 32)) 30136 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) ((";") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_PIPE) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("|") reduce 35)) 30137 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) ((";") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_PIPE) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("|") reduce 19)) 30138 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) ((";") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_PIPE) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("|") reduce 45)) 30139 ((("%") reduce 542) (("(") reduce 579) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) ((";") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PIPE) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 30141 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) ((";") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_PIPE) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("|") reduce 72)) 30142 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) ((";") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_PIPE) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("|") reduce 83)) 30143 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) ((";") reduce 572) (("<") reduce 572) (("=") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AND_EQUAL) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_COALESCE_EQUAL) reduce 572) ((T_CONCAT_EQUAL) reduce 572) ((T_DEC) reduce 572) ((T_DIV_EQUAL) reduce 572) ((T_INC) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_MINUS_EQUAL) reduce 572) ((T_MOD_EQUAL) reduce 572) ((T_MUL_EQUAL) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_OR_EQUAL) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_PLUS_EQUAL) reduce 572) ((T_POW) reduce 572) ((T_POW_EQUAL) reduce 572) ((T_SL) reduce 572) ((T_SL_EQUAL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) ((T_SR_EQUAL) reduce 572) ((T_XOR_EQUAL) reduce 572) (("[") reduce 572) (("^") reduce 572) (("|") reduce 572)) 30146 ((("%") reduce 544) (("(") reduce 580) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) ((";") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PIPE) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 30147 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((";") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("|") reduce 507)) 30152 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) ((";") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_PIPE) reduce 425) ((T_POW) reduce 425) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 30154 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) ((";") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_PIPE) reduce 423) ((T_POW) reduce 423) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 30164 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 419) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 30165 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 421) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 30166 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 420) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 30167 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 413) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 30168 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 410) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 30169 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 418) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 30170 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 412) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 30171 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 416) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 30172 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 411) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 30173 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 414) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 30174 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 415) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 30175 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 422) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 30176 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 417) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 30178 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 406) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 30181 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) ((";") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 30184 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) ((";") reduce 564) (("<") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_POW) reduce 564) ((T_SL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) (("[") reduce 564) (("^") reduce 564) (("|") reduce 564)) 30186 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) ((";") reduce 563) (("<") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_POW) reduce 563) ((T_SL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) (("[") reduce 563) (("^") reduce 563) (("|") reduce 563)) 30188 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) ((";") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 30189 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((";") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 30190 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) ((";") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("[") reduce 548) (("^") reduce 407) (("|") reduce 407)) 30194 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) ((";") reduce 572) (("<") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_POW) reduce 572) ((T_SL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) (("[") reduce 572) (("^") reduce 572) (("|") reduce 572)) 30196 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((";") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("|") reduce 507)) 30197 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((";") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 30201 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) ((";") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("[") reduce 292) (("^") reduce 292) (("|") reduce 292)) 30202 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) ((";") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("[") reduce 293) (("^") reduce 293) (("|") reduce 293)) 30203 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) ((";") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 30206 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) ((";") reduce 571) (("<") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_POW) reduce 571) ((T_SL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) (("[") reduce 571) (("^") reduce 571) (("|") reduce 571)) 30207 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((";") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("|") reduce 506)) 30208 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((";") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("|") reduce 508)) 30213 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) ((";") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 30214 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) ((";") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 30215 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((";") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 30216 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) ((";") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 30219 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) ((";") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 30220 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((";") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 30221 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((";") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 30222 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) ((";") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("[") reduce 562) (("^") reduce 562) (("|") reduce 562)) 30223 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((";") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 30224 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) ((";") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 30227 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) ((";") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 30228 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) ((";") reduce 504) (("<") reduce 504) (("=") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AND_EQUAL) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_COALESCE_EQUAL) reduce 504) ((T_CONCAT_EQUAL) reduce 504) ((T_DEC) reduce 504) ((T_DIV_EQUAL) reduce 504) ((T_INC) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_MINUS_EQUAL) reduce 504) ((T_MOD_EQUAL) reduce 504) ((T_MUL_EQUAL) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_OR_EQUAL) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_PLUS_EQUAL) reduce 504) ((T_POW) reduce 504) ((T_POW_EQUAL) reduce 504) ((T_SL) reduce 504) ((T_SL_EQUAL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) ((T_SR_EQUAL) reduce 504) ((T_XOR_EQUAL) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 30234 ((("!") reduce 166) (("\"") reduce 166) (($) reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONST) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_ELSE) reduce 166) ((T_ELSEIF) reduce 166) ((T_EMPTY) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAMESPACE) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_USE) reduce 166) ((T_VARIABLE) reduce 166) ((T_VOID_CAST) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 30244 ((("!") reduce 160) (("\"") reduce 160) (($) reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONST) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_ELSE) reduce 160) ((T_ELSEIF) reduce 160) ((T_EMPTY) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAMESPACE) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_USE) reduce 160) ((T_VARIABLE) reduce 160) ((T_VOID_CAST) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 30253 ((("!") reduce 149) (("\"") reduce 149) (($) reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONST) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_ELSE) reduce 149) ((T_ELSEIF) reduce 149) ((T_EMPTY) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAMESPACE) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_USE) reduce 149) ((T_VARIABLE) reduce 149) ((T_VOID_CAST) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 30256 ((("!") reduce 148) (("\"") reduce 148) (($) reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONST) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_ELSE) reduce 148) ((T_ELSEIF) reduce 148) ((T_EMPTY) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAMESPACE) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_USE) reduce 148) ((T_VARIABLE) reduce 148) ((T_VOID_CAST) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 30258 ((("!") reduce 245) (("\"") reduce 245) (("$") reduce 245) (($) reduce 245) (("(") reduce 245) (("+") reduce 245) (("-") reduce 245) ((";") reduce 245) (("@") reduce 245) ((T_ABSTRACT) reduce 245) ((T_ARRAY) reduce 245) ((T_ARRAY_CAST) reduce 245) ((T_ATTRIBUTE) reduce 245) ((T_BOOL_CAST) reduce 245) ((T_BREAK) reduce 245) ((T_CLASS) reduce 245) ((T_CLASS_C) reduce 245) ((T_CLONE) reduce 245) ((T_CONST) reduce 245) ((T_CONSTANT_ENCAPSED_STRING) reduce 245) ((T_CONTINUE) reduce 245) ((T_DEC) reduce 245) ((T_DECLARE) reduce 245) ((T_DIR) reduce 245) ((T_DNUMBER) reduce 245) ((T_DO) reduce 245) ((T_DOUBLE_CAST) reduce 245) ((T_ECHO) reduce 245) ((T_ELSE) reduce 245) ((T_ELSEIF) reduce 245) ((T_EMPTY) reduce 245) ((T_ENUM) reduce 245) ((T_EVAL) reduce 245) ((T_EXIT) reduce 245) ((T_FILE) reduce 245) ((T_FINAL) reduce 245) ((T_FN) reduce 245) ((T_FOR) reduce 245) ((T_FOREACH) reduce 245) ((T_FUNCTION) reduce 245) ((T_FUNC_C) reduce 245) ((T_GLOBAL) reduce 245) ((T_GOTO) reduce 245) ((T_HALT_COMPILER) reduce 245) ((T_IF) reduce 245) ((T_INC) reduce 245) ((T_INCLUDE) reduce 245) ((T_INCLUDE_ONCE) reduce 245) ((T_INLINE_HTML) reduce 245) ((T_INTERFACE) reduce 245) ((T_INT_CAST) reduce 245) ((T_ISSET) reduce 245) ((T_LINE) reduce 245) ((T_LIST) reduce 245) ((T_LNUMBER) reduce 245) ((T_MATCH) reduce 245) ((T_METHOD_C) reduce 245) ((T_NAMESPACE) reduce 245) ((T_NAME_FULLY_QUALIFIED) reduce 245) ((T_NAME_QUALIFIED) reduce 245) ((T_NAME_RELATIVE) reduce 245) ((T_NEW) reduce 245) ((T_NS_C) reduce 245) ((T_OBJECT_CAST) reduce 245) ((T_PRINT) reduce 245) ((T_PROPERTY_C) reduce 245) ((T_READONLY) reduce 245) ((T_REQUIRE) reduce 245) ((T_REQUIRE_ONCE) reduce 245) ((T_RETURN) reduce 245) ((T_START_HEREDOC) reduce 245) ((T_STATIC) reduce 245) ((T_STRING) reduce 245) ((T_STRING_CAST) reduce 245) ((T_SWITCH) reduce 245) ((T_THROW) reduce 245) ((T_TRAIT) reduce 245) ((T_TRAIT_C) reduce 245) ((T_TRY) reduce 245) ((T_UNSET) reduce 245) ((T_UNSET_CAST) reduce 245) ((T_USE) reduce 245) ((T_VARIABLE) reduce 245) ((T_VOID_CAST) reduce 245) ((T_WHILE) reduce 245) ((T_YIELD) reduce 245) ((T_YIELD_FROM) reduce 245) (("[") reduce 245) (("`") reduce 245) (("{") reduce 245) (("~") reduce 245)) 30261 ((("!") reduce 147) (("\"") reduce 147) (($) reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONST) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_ELSE) reduce 147) ((T_ELSEIF) reduce 147) ((T_EMPTY) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAMESPACE) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_USE) reduce 147) ((T_VARIABLE) reduce 147) ((T_VOID_CAST) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 30263 ((("!") reduce 247) (("\"") reduce 247) (("$") reduce 247) (($) reduce 247) (("(") reduce 247) (("+") reduce 247) (("-") reduce 247) ((";") reduce 247) (("@") reduce 247) ((T_ABSTRACT) reduce 247) ((T_ARRAY) reduce 247) ((T_ARRAY_CAST) reduce 247) ((T_ATTRIBUTE) reduce 247) ((T_BOOL_CAST) reduce 247) ((T_BREAK) reduce 247) ((T_CLASS) reduce 247) ((T_CLASS_C) reduce 247) ((T_CLONE) reduce 247) ((T_CONST) reduce 247) ((T_CONSTANT_ENCAPSED_STRING) reduce 247) ((T_CONTINUE) reduce 247) ((T_DEC) reduce 247) ((T_DECLARE) reduce 247) ((T_DIR) reduce 247) ((T_DNUMBER) reduce 247) ((T_DO) reduce 247) ((T_DOUBLE_CAST) reduce 247) ((T_ECHO) reduce 247) ((T_ELSE) reduce 247) ((T_ELSEIF) reduce 247) ((T_EMPTY) reduce 247) ((T_ENUM) reduce 247) ((T_EVAL) reduce 247) ((T_EXIT) reduce 247) ((T_FILE) reduce 247) ((T_FINAL) reduce 247) ((T_FN) reduce 247) ((T_FOR) reduce 247) ((T_FOREACH) reduce 247) ((T_FUNCTION) reduce 247) ((T_FUNC_C) reduce 247) ((T_GLOBAL) reduce 247) ((T_GOTO) reduce 247) ((T_HALT_COMPILER) reduce 247) ((T_IF) reduce 247) ((T_INC) reduce 247) ((T_INCLUDE) reduce 247) ((T_INCLUDE_ONCE) reduce 247) ((T_INLINE_HTML) reduce 247) ((T_INTERFACE) reduce 247) ((T_INT_CAST) reduce 247) ((T_ISSET) reduce 247) ((T_LINE) reduce 247) ((T_LIST) reduce 247) ((T_LNUMBER) reduce 247) ((T_MATCH) reduce 247) ((T_METHOD_C) reduce 247) ((T_NAMESPACE) reduce 247) ((T_NAME_FULLY_QUALIFIED) reduce 247) ((T_NAME_QUALIFIED) reduce 247) ((T_NAME_RELATIVE) reduce 247) ((T_NEW) reduce 247) ((T_NS_C) reduce 247) ((T_OBJECT_CAST) reduce 247) ((T_PRINT) reduce 247) ((T_PROPERTY_C) reduce 247) ((T_READONLY) reduce 247) ((T_REQUIRE) reduce 247) ((T_REQUIRE_ONCE) reduce 247) ((T_RETURN) reduce 247) ((T_START_HEREDOC) reduce 247) ((T_STATIC) reduce 247) ((T_STRING) reduce 247) ((T_STRING_CAST) reduce 247) ((T_SWITCH) reduce 247) ((T_THROW) reduce 247) ((T_TRAIT) reduce 247) ((T_TRAIT_C) reduce 247) ((T_TRY) reduce 247) ((T_UNSET) reduce 247) ((T_UNSET_CAST) reduce 247) ((T_USE) reduce 247) ((T_VARIABLE) reduce 247) ((T_VOID_CAST) reduce 247) ((T_WHILE) reduce 247) ((T_YIELD) reduce 247) ((T_YIELD_FROM) reduce 247) (("[") reduce 247) (("`") reduce 247) (("{") reduce 247) (("~") reduce 247)) 30264 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (($) reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONST) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_ELSE) reduce 161) ((T_ELSEIF) reduce 161) ((T_EMPTY) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAMESPACE) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_USE) reduce 161) ((T_VARIABLE) reduce 161) ((T_VOID_CAST) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 30267 ((("!") reduce 250) (("\"") reduce 250) (("$") reduce 250) (($) reduce 250) (("(") reduce 250) (("+") reduce 250) (("-") reduce 250) ((";") reduce 250) (("@") reduce 250) ((T_ABSTRACT) reduce 250) ((T_ARRAY) reduce 250) ((T_ARRAY_CAST) reduce 250) ((T_ATTRIBUTE) reduce 250) ((T_BOOL_CAST) reduce 250) ((T_BREAK) reduce 250) ((T_CLASS) reduce 250) ((T_CLASS_C) reduce 250) ((T_CLONE) reduce 250) ((T_CONST) reduce 250) ((T_CONSTANT_ENCAPSED_STRING) reduce 250) ((T_CONTINUE) reduce 250) ((T_DEC) reduce 250) ((T_DECLARE) reduce 250) ((T_DIR) reduce 250) ((T_DNUMBER) reduce 250) ((T_DO) reduce 250) ((T_DOUBLE_CAST) reduce 250) ((T_ECHO) reduce 250) ((T_ELSE) reduce 250) ((T_ELSEIF) reduce 250) ((T_EMPTY) reduce 250) ((T_ENUM) reduce 250) ((T_EVAL) reduce 250) ((T_EXIT) reduce 250) ((T_FILE) reduce 250) ((T_FINAL) reduce 250) ((T_FN) reduce 250) ((T_FOR) reduce 250) ((T_FOREACH) reduce 250) ((T_FUNCTION) reduce 250) ((T_FUNC_C) reduce 250) ((T_GLOBAL) reduce 250) ((T_GOTO) reduce 250) ((T_HALT_COMPILER) reduce 250) ((T_IF) reduce 250) ((T_INC) reduce 250) ((T_INCLUDE) reduce 250) ((T_INCLUDE_ONCE) reduce 250) ((T_INLINE_HTML) reduce 250) ((T_INTERFACE) reduce 250) ((T_INT_CAST) reduce 250) ((T_ISSET) reduce 250) ((T_LINE) reduce 250) ((T_LIST) reduce 250) ((T_LNUMBER) reduce 250) ((T_MATCH) reduce 250) ((T_METHOD_C) reduce 250) ((T_NAMESPACE) reduce 250) ((T_NAME_FULLY_QUALIFIED) reduce 250) ((T_NAME_QUALIFIED) reduce 250) ((T_NAME_RELATIVE) reduce 250) ((T_NEW) reduce 250) ((T_NS_C) reduce 250) ((T_OBJECT_CAST) reduce 250) ((T_PRINT) reduce 250) ((T_PROPERTY_C) reduce 250) ((T_READONLY) reduce 250) ((T_REQUIRE) reduce 250) ((T_REQUIRE_ONCE) reduce 250) ((T_RETURN) reduce 250) ((T_START_HEREDOC) reduce 250) ((T_STATIC) reduce 250) ((T_STRING) reduce 250) ((T_STRING_CAST) reduce 250) ((T_SWITCH) reduce 250) ((T_THROW) reduce 250) ((T_TRAIT) reduce 250) ((T_TRAIT_C) reduce 250) ((T_TRY) reduce 250) ((T_UNSET) reduce 250) ((T_UNSET_CAST) reduce 250) ((T_USE) reduce 250) ((T_VARIABLE) reduce 250) ((T_VOID_CAST) reduce 250) ((T_WHILE) reduce 250) ((T_YIELD) reduce 250) ((T_YIELD_FROM) reduce 250) (("[") reduce 250) (("`") reduce 250) (("{") reduce 250) (("~") reduce 250)) 30271 ((("!") reduce 251) (("\"") reduce 251) (($) reduce 251) (("$") reduce 251) (("(") reduce 251) (("+") reduce 251) (("-") reduce 251) ((";") reduce 251) (("@") reduce 251) ((T_ABSTRACT) reduce 251) ((T_ARRAY) reduce 251) ((T_ARRAY_CAST) reduce 251) ((T_ATTRIBUTE) reduce 251) ((T_BOOL_CAST) reduce 251) ((T_BREAK) reduce 251) ((T_CLASS) reduce 251) ((T_CLASS_C) reduce 251) ((T_CLONE) reduce 251) ((T_CONST) reduce 251) ((T_CONSTANT_ENCAPSED_STRING) reduce 251) ((T_CONTINUE) reduce 251) ((T_DEC) reduce 251) ((T_DECLARE) reduce 251) ((T_DIR) reduce 251) ((T_DNUMBER) reduce 251) ((T_DO) reduce 251) ((T_DOUBLE_CAST) reduce 251) ((T_ECHO) reduce 251) ((T_ELSE) reduce 251) ((T_ELSEIF) reduce 251) ((T_EMPTY) reduce 251) ((T_ENUM) reduce 251) ((T_EVAL) reduce 251) ((T_EXIT) reduce 251) ((T_FILE) reduce 251) ((T_FINAL) reduce 251) ((T_FN) reduce 251) ((T_FOR) reduce 251) ((T_FOREACH) reduce 251) ((T_FUNCTION) reduce 251) ((T_FUNC_C) reduce 251) ((T_GLOBAL) reduce 251) ((T_GOTO) reduce 251) ((T_HALT_COMPILER) reduce 251) ((T_IF) reduce 251) ((T_INC) reduce 251) ((T_INCLUDE) reduce 251) ((T_INCLUDE_ONCE) reduce 251) ((T_INLINE_HTML) reduce 251) ((T_INTERFACE) reduce 251) ((T_INT_CAST) reduce 251) ((T_ISSET) reduce 251) ((T_LINE) reduce 251) ((T_LIST) reduce 251) ((T_LNUMBER) reduce 251) ((T_MATCH) reduce 251) ((T_METHOD_C) reduce 251) ((T_NAMESPACE) reduce 251) ((T_NAME_FULLY_QUALIFIED) reduce 251) ((T_NAME_QUALIFIED) reduce 251) ((T_NAME_RELATIVE) reduce 251) ((T_NEW) reduce 251) ((T_NS_C) reduce 251) ((T_OBJECT_CAST) reduce 251) ((T_PRINT) reduce 251) ((T_PROPERTY_C) reduce 251) ((T_READONLY) reduce 251) ((T_REQUIRE) reduce 251) ((T_REQUIRE_ONCE) reduce 251) ((T_RETURN) reduce 251) ((T_START_HEREDOC) reduce 251) ((T_STATIC) reduce 251) ((T_STRING) reduce 251) ((T_STRING_CAST) reduce 251) ((T_SWITCH) reduce 251) ((T_THROW) reduce 251) ((T_TRAIT) reduce 251) ((T_TRAIT_C) reduce 251) ((T_TRY) reduce 251) ((T_UNSET) reduce 251) ((T_UNSET_CAST) reduce 251) ((T_USE) reduce 251) ((T_VARIABLE) reduce 251) ((T_VOID_CAST) reduce 251) ((T_WHILE) reduce 251) ((T_YIELD) reduce 251) ((T_YIELD_FROM) reduce 251) (("[") reduce 251) (("`") reduce 251) (("{") reduce 251) (("~") reduce 251)) 30276 ((("!") reduce 242) (("\"") reduce 242) (("$") reduce 242) (($) reduce 242) (("(") reduce 242) (("+") reduce 242) (("-") reduce 242) ((";") reduce 242) (("@") reduce 242) ((T_ABSTRACT) reduce 242) ((T_ARRAY) reduce 242) ((T_ARRAY_CAST) reduce 242) ((T_ATTRIBUTE) reduce 242) ((T_BOOL_CAST) reduce 242) ((T_BREAK) reduce 242) ((T_CLASS) reduce 242) ((T_CLASS_C) reduce 242) ((T_CLONE) reduce 242) ((T_CONST) reduce 242) ((T_CONSTANT_ENCAPSED_STRING) reduce 242) ((T_CONTINUE) reduce 242) ((T_DEC) reduce 242) ((T_DECLARE) reduce 242) ((T_DIR) reduce 242) ((T_DNUMBER) reduce 242) ((T_DO) reduce 242) ((T_DOUBLE_CAST) reduce 242) ((T_ECHO) reduce 242) ((T_ELSE) reduce 242) ((T_ELSEIF) reduce 242) ((T_EMPTY) reduce 242) ((T_ENUM) reduce 242) ((T_EVAL) reduce 242) ((T_EXIT) reduce 242) ((T_FILE) reduce 242) ((T_FINAL) reduce 242) ((T_FN) reduce 242) ((T_FOR) reduce 242) ((T_FOREACH) reduce 242) ((T_FUNCTION) reduce 242) ((T_FUNC_C) reduce 242) ((T_GLOBAL) reduce 242) ((T_GOTO) reduce 242) ((T_HALT_COMPILER) reduce 242) ((T_IF) reduce 242) ((T_INC) reduce 242) ((T_INCLUDE) reduce 242) ((T_INCLUDE_ONCE) reduce 242) ((T_INLINE_HTML) reduce 242) ((T_INTERFACE) reduce 242) ((T_INT_CAST) reduce 242) ((T_ISSET) reduce 242) ((T_LINE) reduce 242) ((T_LIST) reduce 242) ((T_LNUMBER) reduce 242) ((T_MATCH) reduce 242) ((T_METHOD_C) reduce 242) ((T_NAMESPACE) reduce 242) ((T_NAME_FULLY_QUALIFIED) reduce 242) ((T_NAME_QUALIFIED) reduce 242) ((T_NAME_RELATIVE) reduce 242) ((T_NEW) reduce 242) ((T_NS_C) reduce 242) ((T_OBJECT_CAST) reduce 242) ((T_PRINT) reduce 242) ((T_PROPERTY_C) reduce 242) ((T_READONLY) reduce 242) ((T_REQUIRE) reduce 242) ((T_REQUIRE_ONCE) reduce 242) ((T_RETURN) reduce 242) ((T_START_HEREDOC) reduce 242) ((T_STATIC) reduce 242) ((T_STRING) reduce 242) ((T_STRING_CAST) reduce 242) ((T_SWITCH) reduce 242) ((T_THROW) reduce 242) ((T_TRAIT) reduce 242) ((T_TRAIT_C) reduce 242) ((T_TRY) reduce 242) ((T_UNSET) reduce 242) ((T_UNSET_CAST) reduce 242) ((T_USE) reduce 242) ((T_VARIABLE) reduce 242) ((T_VOID_CAST) reduce 242) ((T_WHILE) reduce 242) ((T_YIELD) reduce 242) ((T_YIELD_FROM) reduce 242) (("[") reduce 242) (("`") reduce 242) (("{") reduce 242) (("~") reduce 242)) 30277 ((("!") reduce 150) (("\"") reduce 150) (($) reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONST) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_ELSE) reduce 150) ((T_ELSEIF) reduce 150) ((T_EMPTY) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAMESPACE) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_USE) reduce 150) ((T_VARIABLE) reduce 150) ((T_VOID_CAST) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 30280 ((("!") reduce 243) (("\"") reduce 243) (($) reduce 243) (("$") reduce 243) (("(") reduce 243) (("+") reduce 243) (("-") reduce 243) ((";") reduce 243) (("@") reduce 243) ((T_ABSTRACT) reduce 243) ((T_ARRAY) reduce 243) ((T_ARRAY_CAST) reduce 243) ((T_ATTRIBUTE) reduce 243) ((T_BOOL_CAST) reduce 243) ((T_BREAK) reduce 243) ((T_CLASS) reduce 243) ((T_CLASS_C) reduce 243) ((T_CLONE) reduce 243) ((T_CONST) reduce 243) ((T_CONSTANT_ENCAPSED_STRING) reduce 243) ((T_CONTINUE) reduce 243) ((T_DEC) reduce 243) ((T_DECLARE) reduce 243) ((T_DIR) reduce 243) ((T_DNUMBER) reduce 243) ((T_DO) reduce 243) ((T_DOUBLE_CAST) reduce 243) ((T_ECHO) reduce 243) ((T_ELSE) reduce 243) ((T_ELSEIF) reduce 243) ((T_EMPTY) reduce 243) ((T_ENUM) reduce 243) ((T_EVAL) reduce 243) ((T_EXIT) reduce 243) ((T_FILE) reduce 243) ((T_FINAL) reduce 243) ((T_FN) reduce 243) ((T_FOR) reduce 243) ((T_FOREACH) reduce 243) ((T_FUNCTION) reduce 243) ((T_FUNC_C) reduce 243) ((T_GLOBAL) reduce 243) ((T_GOTO) reduce 243) ((T_HALT_COMPILER) reduce 243) ((T_IF) reduce 243) ((T_INC) reduce 243) ((T_INCLUDE) reduce 243) ((T_INCLUDE_ONCE) reduce 243) ((T_INLINE_HTML) reduce 243) ((T_INTERFACE) reduce 243) ((T_INT_CAST) reduce 243) ((T_ISSET) reduce 243) ((T_LINE) reduce 243) ((T_LIST) reduce 243) ((T_LNUMBER) reduce 243) ((T_MATCH) reduce 243) ((T_METHOD_C) reduce 243) ((T_NAMESPACE) reduce 243) ((T_NAME_FULLY_QUALIFIED) reduce 243) ((T_NAME_QUALIFIED) reduce 243) ((T_NAME_RELATIVE) reduce 243) ((T_NEW) reduce 243) ((T_NS_C) reduce 243) ((T_OBJECT_CAST) reduce 243) ((T_PRINT) reduce 243) ((T_PROPERTY_C) reduce 243) ((T_READONLY) reduce 243) ((T_REQUIRE) reduce 243) ((T_REQUIRE_ONCE) reduce 243) ((T_RETURN) reduce 243) ((T_START_HEREDOC) reduce 243) ((T_STATIC) reduce 243) ((T_STRING) reduce 243) ((T_STRING_CAST) reduce 243) ((T_SWITCH) reduce 243) ((T_THROW) reduce 243) ((T_TRAIT) reduce 243) ((T_TRAIT_C) reduce 243) ((T_TRY) reduce 243) ((T_UNSET) reduce 243) ((T_UNSET_CAST) reduce 243) ((T_USE) reduce 243) ((T_VARIABLE) reduce 243) ((T_VOID_CAST) reduce 243) ((T_WHILE) reduce 243) ((T_YIELD) reduce 243) ((T_YIELD_FROM) reduce 243) (("[") reduce 243) (("`") reduce 243) (("{") reduce 243) (("~") reduce 243)) 30282 ((("!") reduce 170) (("\"") reduce 170) (($) reduce 170) (("$") reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONST) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_ELSE) reduce 170) ((T_ELSEIF) reduce 170) ((T_EMPTY) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAMESPACE) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_USE) reduce 170) ((T_VARIABLE) reduce 170) ((T_VOID_CAST) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("~") reduce 170)) 30287 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (($) reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONST) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_ELSE) reduce 162) ((T_ELSEIF) reduce 162) ((T_EMPTY) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAMESPACE) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_USE) reduce 162) ((T_VARIABLE) reduce 162) ((T_VOID_CAST) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 30290 ((("!") reduce 171) (("\"") reduce 171) (($) reduce 171) (("$") reduce 171) (("(") reduce 171) (("+") reduce 171) (("-") reduce 171) ((";") reduce 171) (("@") reduce 171) ((T_ABSTRACT) reduce 171) ((T_ARRAY) reduce 171) ((T_ARRAY_CAST) reduce 171) ((T_ATTRIBUTE) reduce 171) ((T_BOOL_CAST) reduce 171) ((T_BREAK) reduce 171) ((T_CATCH) reduce 171) ((T_CLASS) reduce 171) ((T_CLASS_C) reduce 171) ((T_CLONE) reduce 171) ((T_CONST) reduce 171) ((T_CONSTANT_ENCAPSED_STRING) reduce 171) ((T_CONTINUE) reduce 171) ((T_DEC) reduce 171) ((T_DECLARE) reduce 171) ((T_DIR) reduce 171) ((T_DNUMBER) reduce 171) ((T_DO) reduce 171) ((T_DOUBLE_CAST) reduce 171) ((T_ECHO) reduce 171) ((T_ELSE) reduce 171) ((T_ELSEIF) reduce 171) ((T_EMPTY) reduce 171) ((T_ENUM) reduce 171) ((T_EVAL) reduce 171) ((T_EXIT) reduce 171) ((T_FILE) reduce 171) ((T_FINAL) reduce 171) ((T_FINALLY) reduce 171) ((T_FN) reduce 171) ((T_FOR) reduce 171) ((T_FOREACH) reduce 171) ((T_FUNCTION) reduce 171) ((T_FUNC_C) reduce 171) ((T_GLOBAL) reduce 171) ((T_GOTO) reduce 171) ((T_HALT_COMPILER) reduce 171) ((T_IF) reduce 171) ((T_INC) reduce 171) ((T_INCLUDE) reduce 171) ((T_INCLUDE_ONCE) reduce 171) ((T_INLINE_HTML) reduce 171) ((T_INTERFACE) reduce 171) ((T_INT_CAST) reduce 171) ((T_ISSET) reduce 171) ((T_LINE) reduce 171) ((T_LIST) reduce 171) ((T_LNUMBER) reduce 171) ((T_MATCH) reduce 171) ((T_METHOD_C) reduce 171) ((T_NAMESPACE) reduce 171) ((T_NAME_FULLY_QUALIFIED) reduce 171) ((T_NAME_QUALIFIED) reduce 171) ((T_NAME_RELATIVE) reduce 171) ((T_NEW) reduce 171) ((T_NS_C) reduce 171) ((T_OBJECT_CAST) reduce 171) ((T_PRINT) reduce 171) ((T_PROPERTY_C) reduce 171) ((T_READONLY) reduce 171) ((T_REQUIRE) reduce 171) ((T_REQUIRE_ONCE) reduce 171) ((T_RETURN) reduce 171) ((T_START_HEREDOC) reduce 171) ((T_STATIC) reduce 171) ((T_STRING) reduce 171) ((T_STRING_CAST) reduce 171) ((T_SWITCH) reduce 171) ((T_THROW) reduce 171) ((T_TRAIT) reduce 171) ((T_TRAIT_C) reduce 171) ((T_TRY) reduce 171) ((T_UNSET) reduce 171) ((T_UNSET_CAST) reduce 171) ((T_USE) reduce 171) ((T_VARIABLE) reduce 171) ((T_VOID_CAST) reduce 171) ((T_WHILE) reduce 171) ((T_YIELD) reduce 171) ((T_YIELD_FROM) reduce 171) (("[") reduce 171) (("`") reduce 171) (("{") reduce 171) (("~") reduce 171)) 30291 ((("!") reduce 177) (("\"") reduce 177) (("$") reduce 177) (($) reduce 177) (("(") reduce 177) (("+") reduce 177) (("-") reduce 177) ((";") reduce 177) (("@") reduce 177) ((T_ABSTRACT) reduce 177) ((T_ARRAY) reduce 177) ((T_ARRAY_CAST) reduce 177) ((T_ATTRIBUTE) reduce 177) ((T_BOOL_CAST) reduce 177) ((T_BREAK) reduce 177) ((T_CATCH) shift) ((T_CLASS) reduce 177) ((T_CLASS_C) reduce 177) ((T_CLONE) reduce 177) ((T_CONST) reduce 177) ((T_CONSTANT_ENCAPSED_STRING) reduce 177) ((T_CONTINUE) reduce 177) ((T_DEC) reduce 177) ((T_DECLARE) reduce 177) ((T_DIR) reduce 177) ((T_DNUMBER) reduce 177) ((T_DO) reduce 177) ((T_DOUBLE_CAST) reduce 177) ((T_ECHO) reduce 177) ((T_ELSE) reduce 177) ((T_ELSEIF) reduce 177) ((T_EMPTY) reduce 177) ((T_ENUM) reduce 177) ((T_EVAL) reduce 177) ((T_EXIT) reduce 177) ((T_FILE) reduce 177) ((T_FINAL) reduce 177) ((T_FINALLY) shift) ((T_FN) reduce 177) ((T_FOR) reduce 177) ((T_FOREACH) reduce 177) ((T_FUNCTION) reduce 177) ((T_FUNC_C) reduce 177) ((T_GLOBAL) reduce 177) ((T_GOTO) reduce 177) ((T_HALT_COMPILER) reduce 177) ((T_IF) reduce 177) ((T_INC) reduce 177) ((T_INCLUDE) reduce 177) ((T_INCLUDE_ONCE) reduce 177) ((T_INLINE_HTML) reduce 177) ((T_INTERFACE) reduce 177) ((T_INT_CAST) reduce 177) ((T_ISSET) reduce 177) ((T_LINE) reduce 177) ((T_LIST) reduce 177) ((T_LNUMBER) reduce 177) ((T_MATCH) reduce 177) ((T_METHOD_C) reduce 177) ((T_NAMESPACE) reduce 177) ((T_NAME_FULLY_QUALIFIED) reduce 177) ((T_NAME_QUALIFIED) reduce 177) ((T_NAME_RELATIVE) reduce 177) ((T_NEW) reduce 177) ((T_NS_C) reduce 177) ((T_OBJECT_CAST) reduce 177) ((T_PRINT) reduce 177) ((T_PROPERTY_C) reduce 177) ((T_READONLY) reduce 177) ((T_REQUIRE) reduce 177) ((T_REQUIRE_ONCE) reduce 177) ((T_RETURN) reduce 177) ((T_START_HEREDOC) reduce 177) ((T_STATIC) reduce 177) ((T_STRING) reduce 177) ((T_STRING_CAST) reduce 177) ((T_SWITCH) reduce 177) ((T_THROW) reduce 177) ((T_TRAIT) reduce 177) ((T_TRAIT_C) reduce 177) ((T_TRY) reduce 177) ((T_UNSET) reduce 177) ((T_UNSET_CAST) reduce 177) ((T_USE) reduce 177) ((T_VARIABLE) reduce 177) ((T_VOID_CAST) reduce 177) ((T_WHILE) reduce 177) ((T_YIELD) reduce 177) ((T_YIELD_FROM) reduce 177) (("[") reduce 177) (("`") reduce 177) (("{") reduce 177) (("~") reduce 177)) 30294 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (($) reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONST) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_ELSE) reduce 167) ((T_ELSEIF) reduce 167) ((T_EMPTY) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAMESPACE) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_USE) reduce 167) ((T_VARIABLE) reduce 167) ((T_VOID_CAST) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 30297 ((("!") reduce 178) (("\"") reduce 178) (("$") reduce 178) (($) reduce 178) (("(") reduce 178) (("+") reduce 178) (("-") reduce 178) ((";") reduce 178) (("@") reduce 178) ((T_ABSTRACT) reduce 178) ((T_ARRAY) reduce 178) ((T_ARRAY_CAST) reduce 178) ((T_ATTRIBUTE) reduce 178) ((T_BOOL_CAST) reduce 178) ((T_BREAK) reduce 178) ((T_CLASS) reduce 178) ((T_CLASS_C) reduce 178) ((T_CLONE) reduce 178) ((T_CONST) reduce 178) ((T_CONSTANT_ENCAPSED_STRING) reduce 178) ((T_CONTINUE) reduce 178) ((T_DEC) reduce 178) ((T_DECLARE) reduce 178) ((T_DIR) reduce 178) ((T_DNUMBER) reduce 178) ((T_DO) reduce 178) ((T_DOUBLE_CAST) reduce 178) ((T_ECHO) reduce 178) ((T_ELSE) reduce 178) ((T_ELSEIF) reduce 178) ((T_EMPTY) reduce 178) ((T_ENUM) reduce 178) ((T_EVAL) reduce 178) ((T_EXIT) reduce 178) ((T_FILE) reduce 178) ((T_FINAL) reduce 178) ((T_FN) reduce 178) ((T_FOR) reduce 178) ((T_FOREACH) reduce 178) ((T_FUNCTION) reduce 178) ((T_FUNC_C) reduce 178) ((T_GLOBAL) reduce 178) ((T_GOTO) reduce 178) ((T_HALT_COMPILER) reduce 178) ((T_IF) reduce 178) ((T_INC) reduce 178) ((T_INCLUDE) reduce 178) ((T_INCLUDE_ONCE) reduce 178) ((T_INLINE_HTML) reduce 178) ((T_INTERFACE) reduce 178) ((T_INT_CAST) reduce 178) ((T_ISSET) reduce 178) ((T_LINE) reduce 178) ((T_LIST) reduce 178) ((T_LNUMBER) reduce 178) ((T_MATCH) reduce 178) ((T_METHOD_C) reduce 178) ((T_NAMESPACE) reduce 178) ((T_NAME_FULLY_QUALIFIED) reduce 178) ((T_NAME_QUALIFIED) reduce 178) ((T_NAME_RELATIVE) reduce 178) ((T_NEW) reduce 178) ((T_NS_C) reduce 178) ((T_OBJECT_CAST) reduce 178) ((T_PRINT) reduce 178) ((T_PROPERTY_C) reduce 178) ((T_READONLY) reduce 178) ((T_REQUIRE) reduce 178) ((T_REQUIRE_ONCE) reduce 178) ((T_RETURN) reduce 178) ((T_START_HEREDOC) reduce 178) ((T_STATIC) reduce 178) ((T_STRING) reduce 178) ((T_STRING_CAST) reduce 178) ((T_SWITCH) reduce 178) ((T_THROW) reduce 178) ((T_TRAIT) reduce 178) ((T_TRAIT_C) reduce 178) ((T_TRY) reduce 178) ((T_UNSET) reduce 178) ((T_UNSET_CAST) reduce 178) ((T_USE) reduce 178) ((T_VARIABLE) reduce 178) ((T_VOID_CAST) reduce 178) ((T_WHILE) reduce 178) ((T_YIELD) reduce 178) ((T_YIELD_FROM) reduce 178) (("[") reduce 178) (("`") reduce 178) (("{") reduce 178) (("~") reduce 178)) 30304 ((("!") reduce 172) (("\"") reduce 172) (("$") reduce 172) (($) reduce 172) (("(") reduce 172) (("+") reduce 172) (("-") reduce 172) ((";") reduce 172) (("@") reduce 172) ((T_ABSTRACT) reduce 172) ((T_ARRAY) reduce 172) ((T_ARRAY_CAST) reduce 172) ((T_ATTRIBUTE) reduce 172) ((T_BOOL_CAST) reduce 172) ((T_BREAK) reduce 172) ((T_CATCH) reduce 172) ((T_CLASS) reduce 172) ((T_CLASS_C) reduce 172) ((T_CLONE) reduce 172) ((T_CONST) reduce 172) ((T_CONSTANT_ENCAPSED_STRING) reduce 172) ((T_CONTINUE) reduce 172) ((T_DEC) reduce 172) ((T_DECLARE) reduce 172) ((T_DIR) reduce 172) ((T_DNUMBER) reduce 172) ((T_DO) reduce 172) ((T_DOUBLE_CAST) reduce 172) ((T_ECHO) reduce 172) ((T_ELSE) reduce 172) ((T_ELSEIF) reduce 172) ((T_EMPTY) reduce 172) ((T_ENUM) reduce 172) ((T_EVAL) reduce 172) ((T_EXIT) reduce 172) ((T_FILE) reduce 172) ((T_FINAL) reduce 172) ((T_FINALLY) reduce 172) ((T_FN) reduce 172) ((T_FOR) reduce 172) ((T_FOREACH) reduce 172) ((T_FUNCTION) reduce 172) ((T_FUNC_C) reduce 172) ((T_GLOBAL) reduce 172) ((T_GOTO) reduce 172) ((T_HALT_COMPILER) reduce 172) ((T_IF) reduce 172) ((T_INC) reduce 172) ((T_INCLUDE) reduce 172) ((T_INCLUDE_ONCE) reduce 172) ((T_INLINE_HTML) reduce 172) ((T_INTERFACE) reduce 172) ((T_INT_CAST) reduce 172) ((T_ISSET) reduce 172) ((T_LINE) reduce 172) ((T_LIST) reduce 172) ((T_LNUMBER) reduce 172) ((T_MATCH) reduce 172) ((T_METHOD_C) reduce 172) ((T_NAMESPACE) reduce 172) ((T_NAME_FULLY_QUALIFIED) reduce 172) ((T_NAME_QUALIFIED) reduce 172) ((T_NAME_RELATIVE) reduce 172) ((T_NEW) reduce 172) ((T_NS_C) reduce 172) ((T_OBJECT_CAST) reduce 172) ((T_PRINT) reduce 172) ((T_PROPERTY_C) reduce 172) ((T_READONLY) reduce 172) ((T_REQUIRE) reduce 172) ((T_REQUIRE_ONCE) reduce 172) ((T_RETURN) reduce 172) ((T_START_HEREDOC) reduce 172) ((T_STATIC) reduce 172) ((T_STRING) reduce 172) ((T_STRING_CAST) reduce 172) ((T_SWITCH) reduce 172) ((T_THROW) reduce 172) ((T_TRAIT) reduce 172) ((T_TRAIT_C) reduce 172) ((T_TRY) reduce 172) ((T_UNSET) reduce 172) ((T_UNSET_CAST) reduce 172) ((T_USE) reduce 172) ((T_VARIABLE) reduce 172) ((T_VOID_CAST) reduce 172) ((T_WHILE) reduce 172) ((T_YIELD) reduce 172) ((T_YIELD_FROM) reduce 172) (("[") reduce 172) (("`") reduce 172) (("{") reduce 172) (("~") reduce 172)) 30309 ((("!") reduce 153) (("\"") reduce 153) (($) reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONST) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_ELSE) reduce 153) ((T_ELSEIF) reduce 153) ((T_EMPTY) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAMESPACE) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_USE) reduce 153) ((T_VARIABLE) reduce 153) ((T_VOID_CAST) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 30313 ((("!") reduce 224) (("\"") reduce 224) (("$") reduce 224) (($) reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONST) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_ELSE) reduce 224) ((T_ELSEIF) reduce 224) ((T_EMPTY) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAMESPACE) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_USE) reduce 224) ((T_VARIABLE) reduce 224) ((T_VOID_CAST) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("~") reduce 224)) 30315 ((("!") reduce 225) (("\"") reduce 225) (($) reduce 225) (("$") reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONST) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_ELSE) reduce 225) ((T_ELSEIF) reduce 225) ((T_EMPTY) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAMESPACE) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_USE) reduce 225) ((T_VARIABLE) reduce 225) ((T_VOID_CAST) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("~") reduce 225)) 30319 ((("!") reduce 226) (("\"") reduce 226) (($) reduce 226) (("$") reduce 226) (("(") reduce 226) (("+") reduce 226) (("-") reduce 226) ((";") reduce 226) (("@") reduce 226) ((T_ABSTRACT) reduce 226) ((T_ARRAY) reduce 226) ((T_ARRAY_CAST) reduce 226) ((T_ATTRIBUTE) reduce 226) ((T_BOOL_CAST) reduce 226) ((T_BREAK) reduce 226) ((T_CLASS) reduce 226) ((T_CLASS_C) reduce 226) ((T_CLONE) reduce 226) ((T_CONST) reduce 226) ((T_CONSTANT_ENCAPSED_STRING) reduce 226) ((T_CONTINUE) reduce 226) ((T_DEC) reduce 226) ((T_DECLARE) reduce 226) ((T_DIR) reduce 226) ((T_DNUMBER) reduce 226) ((T_DO) reduce 226) ((T_DOUBLE_CAST) reduce 226) ((T_ECHO) reduce 226) ((T_ELSE) reduce 226) ((T_ELSEIF) reduce 226) ((T_EMPTY) reduce 226) ((T_ENUM) reduce 226) ((T_EVAL) reduce 226) ((T_EXIT) reduce 226) ((T_FILE) reduce 226) ((T_FINAL) reduce 226) ((T_FN) reduce 226) ((T_FOR) reduce 226) ((T_FOREACH) reduce 226) ((T_FUNCTION) reduce 226) ((T_FUNC_C) reduce 226) ((T_GLOBAL) reduce 226) ((T_GOTO) reduce 226) ((T_HALT_COMPILER) reduce 226) ((T_IF) reduce 226) ((T_INC) reduce 226) ((T_INCLUDE) reduce 226) ((T_INCLUDE_ONCE) reduce 226) ((T_INLINE_HTML) reduce 226) ((T_INTERFACE) reduce 226) ((T_INT_CAST) reduce 226) ((T_ISSET) reduce 226) ((T_LINE) reduce 226) ((T_LIST) reduce 226) ((T_LNUMBER) reduce 226) ((T_MATCH) reduce 226) ((T_METHOD_C) reduce 226) ((T_NAMESPACE) reduce 226) ((T_NAME_FULLY_QUALIFIED) reduce 226) ((T_NAME_QUALIFIED) reduce 226) ((T_NAME_RELATIVE) reduce 226) ((T_NEW) reduce 226) ((T_NS_C) reduce 226) ((T_OBJECT_CAST) reduce 226) ((T_PRINT) reduce 226) ((T_PROPERTY_C) reduce 226) ((T_READONLY) reduce 226) ((T_REQUIRE) reduce 226) ((T_REQUIRE_ONCE) reduce 226) ((T_RETURN) reduce 226) ((T_START_HEREDOC) reduce 226) ((T_STATIC) reduce 226) ((T_STRING) reduce 226) ((T_STRING_CAST) reduce 226) ((T_SWITCH) reduce 226) ((T_THROW) reduce 226) ((T_TRAIT) reduce 226) ((T_TRAIT_C) reduce 226) ((T_TRY) reduce 226) ((T_UNSET) reduce 226) ((T_UNSET_CAST) reduce 226) ((T_USE) reduce 226) ((T_VARIABLE) reduce 226) ((T_VOID_CAST) reduce 226) ((T_WHILE) reduce 226) ((T_YIELD) reduce 226) ((T_YIELD_FROM) reduce 226) (("[") reduce 226) (("`") reduce 226) (("{") reduce 226) (("~") reduce 226)) 30322 ((("!") reduce 227) (("\"") reduce 227) (("$") reduce 227) (($) reduce 227) (("(") reduce 227) (("+") reduce 227) (("-") reduce 227) ((";") reduce 227) (("@") reduce 227) ((T_ABSTRACT) reduce 227) ((T_ARRAY) reduce 227) ((T_ARRAY_CAST) reduce 227) ((T_ATTRIBUTE) reduce 227) ((T_BOOL_CAST) reduce 227) ((T_BREAK) reduce 227) ((T_CLASS) reduce 227) ((T_CLASS_C) reduce 227) ((T_CLONE) reduce 227) ((T_CONST) reduce 227) ((T_CONSTANT_ENCAPSED_STRING) reduce 227) ((T_CONTINUE) reduce 227) ((T_DEC) reduce 227) ((T_DECLARE) reduce 227) ((T_DIR) reduce 227) ((T_DNUMBER) reduce 227) ((T_DO) reduce 227) ((T_DOUBLE_CAST) reduce 227) ((T_ECHO) reduce 227) ((T_ELSE) reduce 227) ((T_ELSEIF) reduce 227) ((T_EMPTY) reduce 227) ((T_ENUM) reduce 227) ((T_EVAL) reduce 227) ((T_EXIT) reduce 227) ((T_FILE) reduce 227) ((T_FINAL) reduce 227) ((T_FN) reduce 227) ((T_FOR) reduce 227) ((T_FOREACH) reduce 227) ((T_FUNCTION) reduce 227) ((T_FUNC_C) reduce 227) ((T_GLOBAL) reduce 227) ((T_GOTO) reduce 227) ((T_HALT_COMPILER) reduce 227) ((T_IF) reduce 227) ((T_INC) reduce 227) ((T_INCLUDE) reduce 227) ((T_INCLUDE_ONCE) reduce 227) ((T_INLINE_HTML) reduce 227) ((T_INTERFACE) reduce 227) ((T_INT_CAST) reduce 227) ((T_ISSET) reduce 227) ((T_LINE) reduce 227) ((T_LIST) reduce 227) ((T_LNUMBER) reduce 227) ((T_MATCH) reduce 227) ((T_METHOD_C) reduce 227) ((T_NAMESPACE) reduce 227) ((T_NAME_FULLY_QUALIFIED) reduce 227) ((T_NAME_QUALIFIED) reduce 227) ((T_NAME_RELATIVE) reduce 227) ((T_NEW) reduce 227) ((T_NS_C) reduce 227) ((T_OBJECT_CAST) reduce 227) ((T_PRINT) reduce 227) ((T_PROPERTY_C) reduce 227) ((T_READONLY) reduce 227) ((T_REQUIRE) reduce 227) ((T_REQUIRE_ONCE) reduce 227) ((T_RETURN) reduce 227) ((T_START_HEREDOC) reduce 227) ((T_STATIC) reduce 227) ((T_STRING) reduce 227) ((T_STRING_CAST) reduce 227) ((T_SWITCH) reduce 227) ((T_THROW) reduce 227) ((T_TRAIT) reduce 227) ((T_TRAIT_C) reduce 227) ((T_TRY) reduce 227) ((T_UNSET) reduce 227) ((T_UNSET_CAST) reduce 227) ((T_USE) reduce 227) ((T_VARIABLE) reduce 227) ((T_VOID_CAST) reduce 227) ((T_WHILE) reduce 227) ((T_YIELD) reduce 227) ((T_YIELD_FROM) reduce 227) (("[") reduce 227) (("`") reduce 227) (("{") reduce 227) (("~") reduce 227)) 30323 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (($) reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONST) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_ELSE) reduce 169) ((T_ELSEIF) reduce 169) ((T_EMPTY) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAMESPACE) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_USE) reduce 169) ((T_VARIABLE) reduce 169) ((T_VOID_CAST) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 30325 ((("!") reduce 158) (("\"") reduce 158) (($) reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONST) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_ELSE) reduce 158) ((T_ELSEIF) reduce 158) ((T_EMPTY) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAMESPACE) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_USE) reduce 158) ((T_VARIABLE) reduce 158) ((T_VOID_CAST) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 30327 ((("!") reduce 156) (("\"") reduce 156) (($) reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONST) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_ELSE) reduce 156) ((T_ELSEIF) reduce 156) ((T_EMPTY) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAMESPACE) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_USE) reduce 156) ((T_VARIABLE) reduce 156) ((T_VOID_CAST) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 30329 ((("!") reduce 168) (("\"") reduce 168) (($) reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONST) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_ELSE) reduce 168) ((T_ELSEIF) reduce 168) ((T_EMPTY) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAMESPACE) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_USE) reduce 168) ((T_VARIABLE) reduce 168) ((T_VOID_CAST) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 30331 ((("!") reduce 157) (("\"") reduce 157) (($) reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONST) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_ELSE) reduce 157) ((T_ELSEIF) reduce 157) ((T_EMPTY) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAMESPACE) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_USE) reduce 157) ((T_VARIABLE) reduce 157) ((T_VOID_CAST) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 30341 ((("!") reduce 164) (("\"") reduce 164) (($) reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONST) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_ELSE) reduce 164) ((T_ELSEIF) reduce 164) ((T_EMPTY) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAMESPACE) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_USE) reduce 164) ((T_VARIABLE) reduce 164) ((T_VOID_CAST) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 30342 ((("!") reduce 220) (("\"") reduce 220) (("$") reduce 220) (($) reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONST) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_ELSE) reduce 220) ((T_ELSEIF) reduce 220) ((T_EMPTY) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAMESPACE) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_USE) reduce 220) ((T_VARIABLE) reduce 220) ((T_VOID_CAST) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("~") reduce 220)) 30345 ((("!") reduce 221) (("\"") reduce 221) (($) reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONST) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_ELSE) reduce 221) ((T_ELSEIF) reduce 221) ((T_EMPTY) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAMESPACE) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_USE) reduce 221) ((T_VARIABLE) reduce 221) ((T_VOID_CAST) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("~") reduce 221)) 30346 ((("!") reduce 163) (("\"") reduce 163) (($) reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONST) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_ELSE) reduce 163) ((T_ELSEIF) reduce 163) ((T_EMPTY) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAMESPACE) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_USE) reduce 163) ((T_VARIABLE) reduce 163) ((T_VOID_CAST) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 30355 ((("!") reduce 152) (("\"") reduce 152) (($) reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONST) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_ELSE) reduce 152) ((T_ELSEIF) reduce 152) ((T_EMPTY) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAMESPACE) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_USE) reduce 152) ((T_VARIABLE) reduce 152) ((T_VOID_CAST) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 30356 ((("!") reduce 218) (("\"") reduce 218) (("$") reduce 218) (($) reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONST) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_ELSE) reduce 218) ((T_ELSEIF) reduce 218) ((T_EMPTY) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAMESPACE) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_USE) reduce 218) ((T_VARIABLE) reduce 218) ((T_VOID_CAST) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("~") reduce 218)) 30359 ((("!") reduce 219) (("\"") reduce 219) (($) reduce 219) (("$") reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONST) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_ELSE) reduce 219) ((T_ELSEIF) reduce 219) ((T_EMPTY) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAMESPACE) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_USE) reduce 219) ((T_VARIABLE) reduce 219) ((T_VOID_CAST) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("~") reduce 219)) 30361 ((("!") reduce 159) (("\"") reduce 159) (($) reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONST) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_ELSE) reduce 159) ((T_ELSEIF) reduce 159) ((T_EMPTY) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAMESPACE) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_USE) reduce 159) ((T_VARIABLE) reduce 159) ((T_VOID_CAST) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 30367 ((("!") reduce 151) (("\"") reduce 151) (($) reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONST) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_ELSE) reduce 151) ((T_ELSEIF) reduce 151) ((T_EMPTY) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAMESPACE) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_USE) reduce 151) ((T_VARIABLE) reduce 151) ((T_VOID_CAST) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 30372 ((("!") reduce 165) (("\"") reduce 165) (($) reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONST) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_ELSE) reduce 165) ((T_ELSEIF) reduce 165) ((T_EMPTY) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAMESPACE) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_USE) reduce 165) ((T_VARIABLE) reduce 165) ((T_VOID_CAST) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 30373 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (($) reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONST) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_ELSE) reduce 222) ((T_ELSEIF) reduce 222) ((T_EMPTY) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAMESPACE) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_USE) reduce 222) ((T_VARIABLE) reduce 222) ((T_VOID_CAST) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("~") reduce 222)) 30376 ((("!") reduce 223) (("\"") reduce 223) (($) reduce 223) (("$") reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONST) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_ELSE) reduce 223) ((T_ELSEIF) reduce 223) ((T_EMPTY) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAMESPACE) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_USE) reduce 223) ((T_VARIABLE) reduce 223) ((T_VOID_CAST) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("~") reduce 223)) 30378 ((("!") reduce 155) (("\"") reduce 155) (($) reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONST) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_ELSE) reduce 155) ((T_ELSEIF) reduce 155) ((T_EMPTY) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAMESPACE) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_USE) reduce 155) ((T_VARIABLE) reduce 155) ((T_VOID_CAST) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 30380 ((("!") reduce 154) (("\"") reduce 154) (($) reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONST) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_ELSE) reduce 154) ((T_ELSEIF) reduce 154) ((T_EMPTY) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAMESPACE) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_USE) reduce 154) ((T_VARIABLE) reduce 154) ((T_VOID_CAST) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 30381 ((("!") reduce 247) (("\"") reduce 247) (("$") reduce 247) (($) reduce 247) (("(") reduce 247) (("+") reduce 247) (("-") reduce 247) ((";") reduce 247) (("@") reduce 247) ((T_ABSTRACT) reduce 247) ((T_ARRAY) reduce 247) ((T_ARRAY_CAST) reduce 247) ((T_ATTRIBUTE) reduce 247) ((T_BOOL_CAST) reduce 247) ((T_BREAK) reduce 247) ((T_CLASS) reduce 247) ((T_CLASS_C) reduce 247) ((T_CLONE) reduce 247) ((T_CONST) reduce 247) ((T_CONSTANT_ENCAPSED_STRING) reduce 247) ((T_CONTINUE) reduce 247) ((T_DEC) reduce 247) ((T_DECLARE) reduce 247) ((T_DIR) reduce 247) ((T_DNUMBER) reduce 247) ((T_DO) reduce 247) ((T_DOUBLE_CAST) reduce 247) ((T_ECHO) reduce 247) ((T_EMPTY) reduce 247) ((T_ENUM) reduce 247) ((T_EVAL) reduce 247) ((T_EXIT) reduce 247) ((T_FILE) reduce 247) ((T_FINAL) reduce 247) ((T_FN) reduce 247) ((T_FOR) reduce 247) ((T_FOREACH) reduce 247) ((T_FUNCTION) reduce 247) ((T_FUNC_C) reduce 247) ((T_GLOBAL) reduce 247) ((T_GOTO) reduce 247) ((T_HALT_COMPILER) reduce 247) ((T_IF) reduce 247) ((T_INC) reduce 247) ((T_INCLUDE) reduce 247) ((T_INCLUDE_ONCE) reduce 247) ((T_INLINE_HTML) reduce 247) ((T_INTERFACE) reduce 247) ((T_INT_CAST) reduce 247) ((T_ISSET) reduce 247) ((T_LINE) reduce 247) ((T_LIST) reduce 247) ((T_LNUMBER) reduce 247) ((T_MATCH) reduce 247) ((T_METHOD_C) reduce 247) ((T_NAMESPACE) reduce 247) ((T_NAME_FULLY_QUALIFIED) reduce 247) ((T_NAME_QUALIFIED) reduce 247) ((T_NAME_RELATIVE) reduce 247) ((T_NEW) reduce 247) ((T_NS_C) reduce 247) ((T_OBJECT_CAST) reduce 247) ((T_PRINT) reduce 247) ((T_PROPERTY_C) reduce 247) ((T_READONLY) reduce 247) ((T_REQUIRE) reduce 247) ((T_REQUIRE_ONCE) reduce 247) ((T_RETURN) reduce 247) ((T_START_HEREDOC) reduce 247) ((T_STATIC) reduce 247) ((T_STRING) reduce 247) ((T_STRING_CAST) reduce 247) ((T_SWITCH) reduce 247) ((T_THROW) reduce 247) ((T_TRAIT) reduce 247) ((T_TRAIT_C) reduce 247) ((T_TRY) reduce 247) ((T_UNSET) reduce 247) ((T_UNSET_CAST) reduce 247) ((T_USE) reduce 247) ((T_VARIABLE) reduce 247) ((T_VOID_CAST) reduce 247) ((T_WHILE) reduce 247) ((T_YIELD) reduce 247) ((T_YIELD_FROM) reduce 247) (("[") reduce 247) (("`") reduce 247) (("{") reduce 247) (("~") reduce 247)) 30392 ((("!") reduce 494) (("\"") reduce 494) (($) reduce 494) (("$") reduce 494) (("(") reduce 494) (("+") reduce 494) (("-") reduce 494) ((";") reduce 494) (("@") reduce 494) ((T_ABSTRACT) reduce 494) ((T_ARRAY) reduce 494) ((T_ARRAY_CAST) reduce 494) ((T_ATTRIBUTE) reduce 494) ((T_BOOL_CAST) reduce 494) ((T_BREAK) reduce 494) ((T_CLASS) reduce 494) ((T_CLASS_C) reduce 494) ((T_CLONE) reduce 494) ((T_CONST) reduce 494) ((T_CONSTANT_ENCAPSED_STRING) reduce 494) ((T_CONTINUE) reduce 494) ((T_DEC) reduce 494) ((T_DECLARE) reduce 494) ((T_DIR) reduce 494) ((T_DNUMBER) reduce 494) ((T_DO) reduce 494) ((T_DOUBLE_CAST) reduce 494) ((T_ECHO) reduce 494) ((T_EMPTY) reduce 494) ((T_ENUM) reduce 494) ((T_EVAL) reduce 494) ((T_EXIT) reduce 494) ((T_FILE) reduce 494) ((T_FINAL) reduce 494) ((T_FN) reduce 494) ((T_FOR) reduce 494) ((T_FOREACH) reduce 494) ((T_FUNCTION) reduce 494) ((T_FUNC_C) reduce 494) ((T_GLOBAL) reduce 494) ((T_GOTO) reduce 494) ((T_HALT_COMPILER) reduce 494) ((T_IF) reduce 494) ((T_INC) reduce 494) ((T_INCLUDE) reduce 494) ((T_INCLUDE_ONCE) reduce 494) ((T_INLINE_HTML) reduce 494) ((T_INTERFACE) reduce 494) ((T_INT_CAST) reduce 494) ((T_ISSET) reduce 494) ((T_LINE) reduce 494) ((T_LIST) reduce 494) ((T_LNUMBER) reduce 494) ((T_MATCH) reduce 494) ((T_METHOD_C) reduce 494) ((T_NAMESPACE) reduce 494) ((T_NAME_FULLY_QUALIFIED) reduce 494) ((T_NAME_QUALIFIED) reduce 494) ((T_NAME_RELATIVE) reduce 494) ((T_NEW) reduce 494) ((T_NS_C) reduce 494) ((T_OBJECT_CAST) reduce 494) ((T_PRINT) reduce 494) ((T_PROPERTY_C) reduce 494) ((T_READONLY) reduce 494) ((T_REQUIRE) reduce 494) ((T_REQUIRE_ONCE) reduce 494) ((T_RETURN) reduce 494) ((T_START_HEREDOC) reduce 494) ((T_STATIC) reduce 494) ((T_STRING) reduce 494) ((T_STRING_CAST) reduce 494) ((T_SWITCH) reduce 494) ((T_THROW) reduce 494) ((T_TRAIT) reduce 494) ((T_TRAIT_C) reduce 494) ((T_TRY) reduce 494) ((T_UNSET) reduce 494) ((T_UNSET_CAST) reduce 494) ((T_USE) reduce 494) ((T_VARIABLE) reduce 494) ((T_VOID_CAST) reduce 494) ((T_WHILE) reduce 494) ((T_YIELD) reduce 494) ((T_YIELD_FROM) reduce 494) (("[") reduce 494) (("`") reduce 494) (("{") reduce 494) (("~") reduce 494)) 30393 ((("!") reduce 184) (("\"") reduce 184) (($) reduce 184) (("$") reduce 184) (("(") reduce 184) (("+") reduce 184) (("-") reduce 184) ((";") reduce 184) (("@") reduce 184) ((T_ABSTRACT) reduce 184) ((T_ARRAY) reduce 184) ((T_ARRAY_CAST) reduce 184) ((T_ATTRIBUTE) reduce 184) ((T_BOOL_CAST) reduce 184) ((T_BREAK) reduce 184) ((T_CLASS) reduce 184) ((T_CLASS_C) reduce 184) ((T_CLONE) reduce 184) ((T_CONST) reduce 184) ((T_CONSTANT_ENCAPSED_STRING) reduce 184) ((T_CONTINUE) reduce 184) ((T_DEC) reduce 184) ((T_DECLARE) reduce 184) ((T_DIR) reduce 184) ((T_DNUMBER) reduce 184) ((T_DO) reduce 184) ((T_DOUBLE_CAST) reduce 184) ((T_ECHO) reduce 184) ((T_EMPTY) reduce 184) ((T_ENUM) reduce 184) ((T_EVAL) reduce 184) ((T_EXIT) reduce 184) ((T_FILE) reduce 184) ((T_FINAL) reduce 184) ((T_FN) reduce 184) ((T_FOR) reduce 184) ((T_FOREACH) reduce 184) ((T_FUNCTION) reduce 184) ((T_FUNC_C) reduce 184) ((T_GLOBAL) reduce 184) ((T_GOTO) reduce 184) ((T_HALT_COMPILER) reduce 184) ((T_IF) reduce 184) ((T_INC) reduce 184) ((T_INCLUDE) reduce 184) ((T_INCLUDE_ONCE) reduce 184) ((T_INLINE_HTML) reduce 184) ((T_INTERFACE) reduce 184) ((T_INT_CAST) reduce 184) ((T_ISSET) reduce 184) ((T_LINE) reduce 184) ((T_LIST) reduce 184) ((T_LNUMBER) reduce 184) ((T_MATCH) reduce 184) ((T_METHOD_C) reduce 184) ((T_NAMESPACE) reduce 184) ((T_NAME_FULLY_QUALIFIED) reduce 184) ((T_NAME_QUALIFIED) reduce 184) ((T_NAME_RELATIVE) reduce 184) ((T_NEW) reduce 184) ((T_NS_C) reduce 184) ((T_OBJECT_CAST) reduce 184) ((T_PRINT) reduce 184) ((T_PROPERTY_C) reduce 184) ((T_READONLY) reduce 184) ((T_REQUIRE) reduce 184) ((T_REQUIRE_ONCE) reduce 184) ((T_RETURN) reduce 184) ((T_START_HEREDOC) reduce 184) ((T_STATIC) reduce 184) ((T_STRING) reduce 184) ((T_STRING_CAST) reduce 184) ((T_SWITCH) reduce 184) ((T_THROW) reduce 184) ((T_TRAIT) reduce 184) ((T_TRAIT_C) reduce 184) ((T_TRY) reduce 184) ((T_UNSET) reduce 184) ((T_UNSET_CAST) reduce 184) ((T_USE) reduce 184) ((T_VARIABLE) reduce 184) ((T_VOID_CAST) reduce 184) ((T_WHILE) reduce 184) ((T_YIELD) reduce 184) ((T_YIELD_FROM) reduce 184) (("[") reduce 184) (("`") reduce 184) (("{") reduce 184) (("~") reduce 184)) 30466 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 494) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 30782 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) ((";") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_PIPE) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("^") reduce 490) (("|") reduce 490)) 30785 ((("%") reduce 541) (("(") reduce 579) (("*") reduce 541) (("+") reduce 541) (("-") reduce 541) ((".") reduce 541) (("/") reduce 541) ((";") reduce 541) (("<") reduce 541) ((">") reduce 541) (("?") reduce 541) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_BOOLEAN_AND) reduce 541) ((T_BOOLEAN_OR) reduce 541) ((T_COALESCE) reduce 541) ((T_INSTANCEOF) reduce 541) ((T_IS_EQUAL) reduce 541) ((T_IS_GREATER_OR_EQUAL) reduce 541) ((T_IS_IDENTICAL) reduce 541) ((T_IS_NOT_EQUAL) reduce 541) ((T_IS_NOT_IDENTICAL) reduce 541) ((T_IS_SMALLER_OR_EQUAL) reduce 541) ((T_LOGICAL_AND) reduce 541) ((T_LOGICAL_OR) reduce 541) ((T_LOGICAL_XOR) reduce 541) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 541) ((T_OBJECT_OPERATOR) reduce 541) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 541) ((T_PIPE) reduce 541) ((T_POW) reduce 541) ((T_SL) reduce 541) ((T_SPACESHIP) reduce 541) ((T_SR) reduce 541) (("[") reduce 541) (("^") reduce 541) (("|") reduce 541)) 30787 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) ((";") reduce 571) (("<") reduce 571) (("=") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AND_EQUAL) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_COALESCE_EQUAL) reduce 571) ((T_CONCAT_EQUAL) reduce 571) ((T_DEC) reduce 571) ((T_DIV_EQUAL) reduce 571) ((T_INC) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_MINUS_EQUAL) reduce 571) ((T_MOD_EQUAL) reduce 571) ((T_MUL_EQUAL) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_OR_EQUAL) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_PLUS_EQUAL) reduce 571) ((T_POW) reduce 571) ((T_POW_EQUAL) reduce 571) ((T_SL) reduce 571) ((T_SL_EQUAL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) ((T_SR_EQUAL) reduce 571) ((T_XOR_EQUAL) reduce 571) (("[") reduce 571) (("^") reduce 571) (("|") reduce 571)) 30790 ((("%") reduce 543) (("(") reduce 580) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) ((";") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PIPE) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 30791 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((";") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("|") reduce 506)) 30792 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((";") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("|") reduce 508)) 30800 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) ((";") reduce 560) (("<") reduce 560) (("=") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AND_EQUAL) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_COALESCE_EQUAL) reduce 560) ((T_CONCAT_EQUAL) reduce 560) ((T_DEC) reduce 560) ((T_DIV_EQUAL) reduce 560) ((T_INC) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_MINUS_EQUAL) reduce 560) ((T_MOD_EQUAL) reduce 560) ((T_MUL_EQUAL) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_OR_EQUAL) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_PLUS_EQUAL) reduce 560) ((T_POW) reduce 560) ((T_POW_EQUAL) reduce 560) ((T_SL) reduce 560) ((T_SL_EQUAL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) ((T_SR_EQUAL) reduce 560) ((T_XOR_EQUAL) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 30801 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) ((";") reduce 582) (("<") reduce 582) (("=") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AND_EQUAL) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_COALESCE_EQUAL) reduce 582) ((T_CONCAT_EQUAL) reduce 582) ((T_DEC) reduce 582) ((T_DIV_EQUAL) reduce 582) ((T_INC) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_MINUS_EQUAL) reduce 582) ((T_MOD_EQUAL) reduce 582) ((T_MUL_EQUAL) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_OR_EQUAL) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_PLUS_EQUAL) reduce 582) ((T_POW) reduce 582) ((T_POW_EQUAL) reduce 582) ((T_SL) reduce 582) ((T_SL_EQUAL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) ((T_SR_EQUAL) reduce 582) ((T_XOR_EQUAL) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 30802 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((";") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 30803 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) ((";") reduce 584) (("<") reduce 584) (("=") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AND_EQUAL) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_COALESCE_EQUAL) reduce 584) ((T_CONCAT_EQUAL) reduce 584) ((T_DEC) reduce 584) ((T_DIV_EQUAL) reduce 584) ((T_INC) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_MINUS_EQUAL) reduce 584) ((T_MOD_EQUAL) reduce 584) ((T_MUL_EQUAL) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_OR_EQUAL) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_PLUS_EQUAL) reduce 584) ((T_POW) reduce 584) ((T_POW_EQUAL) reduce 584) ((T_SL) reduce 584) ((T_SL_EQUAL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) ((T_SR_EQUAL) reduce 584) ((T_XOR_EQUAL) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 30806 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) ((";") reduce 583) (("<") reduce 583) (("=") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AND_EQUAL) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_COALESCE_EQUAL) reduce 583) ((T_CONCAT_EQUAL) reduce 583) ((T_DEC) reduce 583) ((T_DIV_EQUAL) reduce 583) ((T_INC) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_MINUS_EQUAL) reduce 583) ((T_MOD_EQUAL) reduce 583) ((T_MUL_EQUAL) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_OR_EQUAL) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_PLUS_EQUAL) reduce 583) ((T_POW) reduce 583) ((T_POW_EQUAL) reduce 583) ((T_SL) reduce 583) ((T_SL_EQUAL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) ((T_SR_EQUAL) reduce 583) ((T_XOR_EQUAL) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 30807 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((";") reduce 561) (("<") reduce 561) (("=") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AND_EQUAL) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_COALESCE_EQUAL) reduce 561) ((T_CONCAT_EQUAL) reduce 561) ((T_DEC) reduce 561) ((T_DIV_EQUAL) reduce 561) ((T_INC) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_MINUS_EQUAL) reduce 561) ((T_MOD_EQUAL) reduce 561) ((T_MUL_EQUAL) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_OR_EQUAL) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_PLUS_EQUAL) reduce 561) ((T_POW) reduce 561) ((T_POW_EQUAL) reduce 561) ((T_SL) reduce 561) ((T_SL_EQUAL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) ((T_SR_EQUAL) reduce 561) ((T_XOR_EQUAL) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 30808 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((";") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 30809 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) ((";") reduce 562) (("<") reduce 562) (("=") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AND_EQUAL) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_COALESCE_EQUAL) reduce 562) ((T_CONCAT_EQUAL) reduce 562) ((T_DEC) reduce 562) ((T_DIV_EQUAL) reduce 562) ((T_INC) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_MINUS_EQUAL) reduce 562) ((T_MOD_EQUAL) reduce 562) ((T_MUL_EQUAL) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_OR_EQUAL) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_PLUS_EQUAL) reduce 562) ((T_POW) reduce 562) ((T_POW_EQUAL) reduce 562) ((T_SL) reduce 562) ((T_SL_EQUAL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) ((T_SR_EQUAL) reduce 562) ((T_XOR_EQUAL) reduce 562) (("[") reduce 562) (("^") reduce 562) (("|") reduce 562)) 30811 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) ((";") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_PIPE) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 30813 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) ((";") reduce 520) (("<") reduce 520) (("=") shift) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_PIPE) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("|") reduce 520)) 30815 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 405) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 30816 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 482) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 30827 ((("%") reduce 540) (("*") reduce 540) (("+") reduce 540) (("-") reduce 540) ((".") reduce 540) (("/") reduce 540) ((";") reduce 540) (("<") reduce 540) ((">") reduce 540) (("?") reduce 540) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 540) ((T_BOOLEAN_AND) reduce 540) ((T_BOOLEAN_OR) reduce 540) ((T_COALESCE) reduce 540) ((T_DOUBLE_ARROW) reduce 540) ((T_INSTANCEOF) reduce 540) ((T_IS_EQUAL) reduce 540) ((T_IS_GREATER_OR_EQUAL) reduce 540) ((T_IS_IDENTICAL) reduce 540) ((T_IS_NOT_EQUAL) reduce 540) ((T_IS_NOT_IDENTICAL) reduce 540) ((T_IS_SMALLER_OR_EQUAL) reduce 540) ((T_LOGICAL_AND) reduce 540) ((T_LOGICAL_OR) reduce 540) ((T_LOGICAL_XOR) reduce 540) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 540) ((T_OBJECT_OPERATOR) reduce 540) ((T_PIPE) reduce 540) ((T_POW) reduce 540) ((T_SL) reduce 540) ((T_SPACESHIP) reduce 540) ((T_SR) reduce 540) (("[") reduce 540) (("^") reduce 540) (("|") reduce 540)) 30829 ((("%") reduce 521) (("(") reduce 521) (("*") reduce 521) (("+") reduce 521) (("-") reduce 521) ((".") reduce 521) (("/") reduce 521) ((";") reduce 521) (("<") reduce 521) ((">") reduce 521) (("?") reduce 521) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 521) ((T_BOOLEAN_AND) reduce 521) ((T_BOOLEAN_OR) reduce 521) ((T_COALESCE) reduce 521) ((T_DOUBLE_ARROW) reduce 521) ((T_INSTANCEOF) reduce 521) ((T_IS_EQUAL) reduce 521) ((T_IS_GREATER_OR_EQUAL) reduce 521) ((T_IS_IDENTICAL) reduce 521) ((T_IS_NOT_EQUAL) reduce 521) ((T_IS_NOT_IDENTICAL) reduce 521) ((T_IS_SMALLER_OR_EQUAL) reduce 521) ((T_LOGICAL_AND) reduce 521) ((T_LOGICAL_OR) reduce 521) ((T_LOGICAL_XOR) reduce 521) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 521) ((T_OBJECT_OPERATOR) reduce 521) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 521) ((T_PIPE) reduce 521) ((T_POW) reduce 521) ((T_SL) reduce 521) ((T_SPACESHIP) reduce 521) ((T_SR) reduce 521) (("[") reduce 521) (("^") reduce 521) (("|") reduce 521)) 30831 ((("%") reduce 534) (("*") reduce 534) (("+") reduce 534) (("-") reduce 534) ((".") reduce 534) (("/") reduce 534) ((";") reduce 534) (("<") reduce 534) ((">") reduce 534) (("?") reduce 534) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 534) ((T_BOOLEAN_AND) reduce 534) ((T_BOOLEAN_OR) reduce 534) ((T_COALESCE) reduce 534) ((T_DOUBLE_ARROW) reduce 534) ((T_INSTANCEOF) reduce 534) ((T_IS_EQUAL) reduce 534) ((T_IS_GREATER_OR_EQUAL) reduce 534) ((T_IS_IDENTICAL) reduce 534) ((T_IS_NOT_EQUAL) reduce 534) ((T_IS_NOT_IDENTICAL) reduce 534) ((T_IS_SMALLER_OR_EQUAL) reduce 534) ((T_LOGICAL_AND) reduce 534) ((T_LOGICAL_OR) reduce 534) ((T_LOGICAL_XOR) reduce 534) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 534) ((T_OBJECT_OPERATOR) reduce 534) ((T_PIPE) reduce 534) ((T_POW) reduce 534) ((T_SL) reduce 534) ((T_SPACESHIP) reduce 534) ((T_SR) reduce 534) (("[") reduce 534) (("^") reduce 534) (("|") reduce 534)) 30832 ((("%") reduce 524) (("*") reduce 524) (("+") reduce 524) (("-") reduce 524) ((".") reduce 524) (("/") reduce 524) ((";") reduce 524) (("<") reduce 524) ((">") reduce 524) (("?") reduce 524) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 524) ((T_BOOLEAN_AND) reduce 524) ((T_BOOLEAN_OR) reduce 524) ((T_COALESCE) reduce 524) ((T_DOUBLE_ARROW) reduce 524) ((T_INSTANCEOF) reduce 524) ((T_IS_EQUAL) reduce 524) ((T_IS_GREATER_OR_EQUAL) reduce 524) ((T_IS_IDENTICAL) reduce 524) ((T_IS_NOT_EQUAL) reduce 524) ((T_IS_NOT_IDENTICAL) reduce 524) ((T_IS_SMALLER_OR_EQUAL) reduce 524) ((T_LOGICAL_AND) reduce 524) ((T_LOGICAL_OR) reduce 524) ((T_LOGICAL_XOR) reduce 524) ((T_PIPE) reduce 524) ((T_POW) reduce 524) ((T_SL) reduce 524) ((T_SPACESHIP) reduce 524) ((T_SR) reduce 524) (("^") reduce 524) (("|") reduce 524)) 30836 ((("%") reduce 517) (("(") shift) (("*") reduce 517) (("+") reduce 517) (("-") reduce 517) ((".") reduce 517) (("/") reduce 517) ((";") reduce 517) (("<") reduce 517) ((">") reduce 517) (("?") reduce 517) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 517) ((T_BOOLEAN_AND) reduce 517) ((T_BOOLEAN_OR) reduce 517) ((T_COALESCE) reduce 517) ((T_DOUBLE_ARROW) reduce 517) ((T_INSTANCEOF) reduce 517) ((T_IS_EQUAL) reduce 517) ((T_IS_GREATER_OR_EQUAL) reduce 517) ((T_IS_IDENTICAL) reduce 517) ((T_IS_NOT_EQUAL) reduce 517) ((T_IS_NOT_IDENTICAL) reduce 517) ((T_IS_SMALLER_OR_EQUAL) reduce 517) ((T_LOGICAL_AND) reduce 517) ((T_LOGICAL_OR) reduce 517) ((T_LOGICAL_XOR) reduce 517) ((T_PIPE) reduce 517) ((T_POW) reduce 517) ((T_SL) reduce 517) ((T_SPACESHIP) reduce 517) ((T_SR) reduce 517) (("^") reduce 517) (("|") reduce 517)) 30837 ((("%") reduce 533) (("*") reduce 533) (("+") reduce 533) (("-") reduce 533) ((".") reduce 533) (("/") reduce 533) ((";") reduce 533) (("<") reduce 533) ((">") reduce 533) (("?") reduce 533) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 533) ((T_BOOLEAN_AND) reduce 533) ((T_BOOLEAN_OR) reduce 533) ((T_COALESCE) reduce 533) ((T_DOUBLE_ARROW) reduce 533) ((T_INSTANCEOF) reduce 533) ((T_IS_EQUAL) reduce 533) ((T_IS_GREATER_OR_EQUAL) reduce 533) ((T_IS_IDENTICAL) reduce 533) ((T_IS_NOT_EQUAL) reduce 533) ((T_IS_NOT_IDENTICAL) reduce 533) ((T_IS_SMALLER_OR_EQUAL) reduce 533) ((T_LOGICAL_AND) reduce 533) ((T_LOGICAL_OR) reduce 533) ((T_LOGICAL_XOR) reduce 533) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 533) ((T_OBJECT_OPERATOR) reduce 533) ((T_PIPE) reduce 533) ((T_POW) reduce 533) ((T_SL) reduce 533) ((T_SPACESHIP) reduce 533) ((T_SR) reduce 533) (("[") reduce 533) (("^") reduce 533) (("|") reduce 533)) 30838 ((("%") reduce 537) (("*") reduce 537) (("+") reduce 537) (("-") reduce 537) ((".") reduce 537) (("/") reduce 537) ((";") reduce 537) (("<") reduce 537) ((">") reduce 537) (("?") reduce 537) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 537) ((T_BOOLEAN_AND) reduce 537) ((T_BOOLEAN_OR) reduce 537) ((T_COALESCE) reduce 537) ((T_DOUBLE_ARROW) reduce 537) ((T_INSTANCEOF) reduce 537) ((T_IS_EQUAL) reduce 537) ((T_IS_GREATER_OR_EQUAL) reduce 537) ((T_IS_IDENTICAL) reduce 537) ((T_IS_NOT_EQUAL) reduce 537) ((T_IS_NOT_IDENTICAL) reduce 537) ((T_IS_SMALLER_OR_EQUAL) reduce 537) ((T_LOGICAL_AND) reduce 537) ((T_LOGICAL_OR) reduce 537) ((T_LOGICAL_XOR) reduce 537) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 537) ((T_OBJECT_OPERATOR) reduce 537) ((T_PIPE) reduce 537) ((T_POW) reduce 537) ((T_SL) reduce 537) ((T_SPACESHIP) reduce 537) ((T_SR) reduce 537) (("[") reduce 537) (("^") reduce 537) (("|") reduce 537)) 30844 ((("%") reduce 532) (("*") reduce 532) (("+") reduce 532) (("-") reduce 532) ((".") reduce 532) (("/") reduce 532) ((";") reduce 532) (("<") reduce 532) ((">") reduce 532) (("?") reduce 532) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 532) ((T_BOOLEAN_AND) reduce 532) ((T_BOOLEAN_OR) reduce 532) ((T_COALESCE) reduce 532) ((T_DOUBLE_ARROW) reduce 532) ((T_INSTANCEOF) reduce 532) ((T_IS_EQUAL) reduce 532) ((T_IS_GREATER_OR_EQUAL) reduce 532) ((T_IS_IDENTICAL) reduce 532) ((T_IS_NOT_EQUAL) reduce 532) ((T_IS_NOT_IDENTICAL) reduce 532) ((T_IS_SMALLER_OR_EQUAL) reduce 532) ((T_LOGICAL_AND) reduce 532) ((T_LOGICAL_OR) reduce 532) ((T_LOGICAL_XOR) reduce 532) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 532) ((T_OBJECT_OPERATOR) reduce 532) ((T_PIPE) reduce 532) ((T_POW) reduce 532) ((T_SL) reduce 532) ((T_SPACESHIP) reduce 532) ((T_SR) reduce 532) (("[") reduce 532) (("^") reduce 532) (("|") reduce 532)) 30846 ((("%") reduce 523) (("*") reduce 523) (("+") reduce 523) (("-") reduce 523) ((".") reduce 523) (("/") reduce 523) ((";") reduce 523) (("<") reduce 523) ((">") reduce 523) (("?") reduce 523) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 523) ((T_BOOLEAN_AND) reduce 523) ((T_BOOLEAN_OR) reduce 523) ((T_COALESCE) reduce 523) ((T_DOUBLE_ARROW) reduce 523) ((T_INSTANCEOF) reduce 523) ((T_IS_EQUAL) reduce 523) ((T_IS_GREATER_OR_EQUAL) reduce 523) ((T_IS_IDENTICAL) reduce 523) ((T_IS_NOT_EQUAL) reduce 523) ((T_IS_NOT_IDENTICAL) reduce 523) ((T_IS_SMALLER_OR_EQUAL) reduce 523) ((T_LOGICAL_AND) reduce 523) ((T_LOGICAL_OR) reduce 523) ((T_LOGICAL_XOR) reduce 523) ((T_PIPE) reduce 523) ((T_POW) reduce 523) ((T_SL) reduce 523) ((T_SPACESHIP) reduce 523) ((T_SR) reduce 523) (("^") reduce 523) (("|") reduce 523)) 30848 ((("%") reduce 536) (("*") reduce 536) (("+") reduce 536) (("-") reduce 536) ((".") reduce 536) (("/") reduce 536) ((";") reduce 536) (("<") reduce 536) ((">") reduce 536) (("?") reduce 536) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 536) ((T_BOOLEAN_AND) reduce 536) ((T_BOOLEAN_OR) reduce 536) ((T_COALESCE) reduce 536) ((T_DOUBLE_ARROW) reduce 536) ((T_INSTANCEOF) reduce 536) ((T_IS_EQUAL) reduce 536) ((T_IS_GREATER_OR_EQUAL) reduce 536) ((T_IS_IDENTICAL) reduce 536) ((T_IS_NOT_EQUAL) reduce 536) ((T_IS_NOT_IDENTICAL) reduce 536) ((T_IS_SMALLER_OR_EQUAL) reduce 536) ((T_LOGICAL_AND) reduce 536) ((T_LOGICAL_OR) reduce 536) ((T_LOGICAL_XOR) reduce 536) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 536) ((T_OBJECT_OPERATOR) reduce 536) ((T_PIPE) reduce 536) ((T_POW) reduce 536) ((T_SL) reduce 536) ((T_SPACESHIP) reduce 536) ((T_SR) reduce 536) (("[") reduce 536) (("^") reduce 536) (("|") reduce 536)) 30849 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 94) ((T_OBJECT_OPERATOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("[") reduce 94) (("^") reduce 94) (("|") reduce 94)) 30850 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 93) ((T_OBJECT_OPERATOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("[") reduce 93) (("^") reduce 93) (("|") reduce 93)) 30851 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 95) ((T_OBJECT_OPERATOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("[") reduce 95) (("^") reduce 95) (("|") reduce 95)) 30853 ((("%") reduce 539) (("*") reduce 539) (("+") reduce 539) (("-") reduce 539) ((".") reduce 539) (("/") reduce 539) ((";") reduce 539) (("<") reduce 539) ((">") reduce 539) (("?") reduce 539) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 539) ((T_BOOLEAN_AND) reduce 539) ((T_BOOLEAN_OR) reduce 539) ((T_COALESCE) reduce 539) ((T_DOUBLE_ARROW) reduce 539) ((T_INSTANCEOF) reduce 539) ((T_IS_EQUAL) reduce 539) ((T_IS_GREATER_OR_EQUAL) reduce 539) ((T_IS_IDENTICAL) reduce 539) ((T_IS_NOT_EQUAL) reduce 539) ((T_IS_NOT_IDENTICAL) reduce 539) ((T_IS_SMALLER_OR_EQUAL) reduce 539) ((T_LOGICAL_AND) reduce 539) ((T_LOGICAL_OR) reduce 539) ((T_LOGICAL_XOR) reduce 539) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 539) ((T_OBJECT_OPERATOR) reduce 539) ((T_PIPE) reduce 539) ((T_POW) reduce 539) ((T_SL) reduce 539) ((T_SPACESHIP) reduce 539) ((T_SR) reduce 539) (("[") reduce 539) (("^") reduce 539) (("|") reduce 539)) 30856 ((("%") reduce 538) (("*") reduce 538) (("+") reduce 538) (("-") reduce 538) ((".") reduce 538) (("/") reduce 538) ((";") reduce 538) (("<") reduce 538) ((">") reduce 538) (("?") reduce 538) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 538) ((T_BOOLEAN_AND) reduce 538) ((T_BOOLEAN_OR) reduce 538) ((T_COALESCE) reduce 538) ((T_DOUBLE_ARROW) reduce 538) ((T_INSTANCEOF) reduce 538) ((T_IS_EQUAL) reduce 538) ((T_IS_GREATER_OR_EQUAL) reduce 538) ((T_IS_IDENTICAL) reduce 538) ((T_IS_NOT_EQUAL) reduce 538) ((T_IS_NOT_IDENTICAL) reduce 538) ((T_IS_SMALLER_OR_EQUAL) reduce 538) ((T_LOGICAL_AND) reduce 538) ((T_LOGICAL_OR) reduce 538) ((T_LOGICAL_XOR) reduce 538) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 538) ((T_OBJECT_OPERATOR) reduce 538) ((T_PIPE) reduce 538) ((T_POW) reduce 538) ((T_SL) reduce 538) ((T_SPACESHIP) reduce 538) ((T_SR) reduce 538) (("[") reduce 538) (("^") reduce 538) (("|") reduce 538)) 30862 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 92) ((T_OBJECT_OPERATOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("[") reduce 92) (("^") reduce 92) (("|") reduce 92)) 30865 ((("%") reduce 535) (("*") reduce 535) (("+") reduce 535) (("-") reduce 535) ((".") reduce 535) (("/") reduce 535) ((";") reduce 535) (("<") reduce 535) ((">") reduce 535) (("?") reduce 535) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 535) ((T_BOOLEAN_AND) reduce 535) ((T_BOOLEAN_OR) reduce 535) ((T_COALESCE) reduce 535) ((T_DOUBLE_ARROW) reduce 535) ((T_INSTANCEOF) reduce 535) ((T_IS_EQUAL) reduce 535) ((T_IS_GREATER_OR_EQUAL) reduce 535) ((T_IS_IDENTICAL) reduce 535) ((T_IS_NOT_EQUAL) reduce 535) ((T_IS_NOT_IDENTICAL) reduce 535) ((T_IS_SMALLER_OR_EQUAL) reduce 535) ((T_LOGICAL_AND) reduce 535) ((T_LOGICAL_OR) reduce 535) ((T_LOGICAL_XOR) reduce 535) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 535) ((T_OBJECT_OPERATOR) reduce 535) ((T_PIPE) reduce 535) ((T_POW) reduce 535) ((T_SL) reduce 535) ((T_SPACESHIP) reduce 535) ((T_SR) reduce 535) (("[") reduce 535) (("^") reduce 535) (("|") reduce 535)) 30867 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) ((";") reduce 568) (("<") reduce 568) (("=") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AND_EQUAL) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_COALESCE_EQUAL) reduce 568) ((T_CONCAT_EQUAL) reduce 568) ((T_DEC) reduce 568) ((T_DIV_EQUAL) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INC) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_MINUS_EQUAL) reduce 568) ((T_MOD_EQUAL) reduce 568) ((T_MUL_EQUAL) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_OR_EQUAL) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_PLUS_EQUAL) reduce 568) ((T_POW) reduce 568) ((T_POW_EQUAL) reduce 568) ((T_SL) reduce 568) ((T_SL_EQUAL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) ((T_SR_EQUAL) reduce 568) ((T_XOR_EQUAL) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 30868 ((("!") shift) (("\"") shift) (("$") shift) (("%") reduce 479) (("(") shift) (("*") reduce 479) (("+") shift) (("-") shift) ((".") reduce 479) (("/") reduce 479) ((";") reduce 479) (("<") reduce 479) ((">") reduce 479) (("?") reduce 479) (("@") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 479) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOLEAN_AND) reduce 479) ((T_BOOLEAN_OR) reduce 479) ((T_BOOL_CAST) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_COALESCE) reduce 479) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_DEC) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DOUBLE_ARROW) reduce 479) ((T_DOUBLE_CAST) shift) ((T_EMPTY) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FN) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INSTANCEOF) reduce 479) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_IS_EQUAL) reduce 479) ((T_IS_GREATER_OR_EQUAL) reduce 479) ((T_IS_IDENTICAL) reduce 479) ((T_IS_NOT_EQUAL) reduce 479) ((T_IS_NOT_IDENTICAL) reduce 479) ((T_IS_SMALLER_OR_EQUAL) reduce 479) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_LOGICAL_AND) reduce 479) ((T_LOGICAL_OR) reduce 479) ((T_LOGICAL_XOR) reduce 479) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PIPE) reduce 479) ((T_POW) reduce 479) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_SL) reduce 479) ((T_SPACESHIP) reduce 479) ((T_SR) reduce 479) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_THROW) shift) ((T_TRAIT_C) shift) ((T_UNSET_CAST) shift) ((T_VARIABLE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("^") reduce 479) (("`") shift) (("|") reduce 479) (("~") shift)) 30875 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) ((";") reduce 564) (("<") reduce 564) (("=") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AND_EQUAL) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_COALESCE_EQUAL) reduce 564) ((T_CONCAT_EQUAL) reduce 564) ((T_DEC) reduce 564) ((T_DIV_EQUAL) reduce 564) ((T_DOUBLE_ARROW) reduce 564) ((T_INC) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_MINUS_EQUAL) reduce 564) ((T_MOD_EQUAL) reduce 564) ((T_MUL_EQUAL) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_OR_EQUAL) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_PLUS_EQUAL) reduce 564) ((T_POW) reduce 564) ((T_POW_EQUAL) reduce 564) ((T_SL) reduce 564) ((T_SL_EQUAL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) ((T_SR_EQUAL) reduce 564) ((T_XOR_EQUAL) reduce 564) (("[") reduce 564) (("^") reduce 564) (("|") reduce 564)) 30876 ((("%") reduce 530) (("*") reduce 530) (("+") reduce 530) (("-") reduce 530) ((".") reduce 530) (("/") reduce 530) ((";") reduce 530) (("<") reduce 530) ((">") reduce 530) (("?") reduce 530) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 530) ((T_BOOLEAN_AND) reduce 530) ((T_BOOLEAN_OR) reduce 530) ((T_COALESCE) reduce 530) ((T_DOUBLE_ARROW) reduce 530) ((T_INSTANCEOF) reduce 530) ((T_IS_EQUAL) reduce 530) ((T_IS_GREATER_OR_EQUAL) reduce 530) ((T_IS_IDENTICAL) reduce 530) ((T_IS_NOT_EQUAL) reduce 530) ((T_IS_NOT_IDENTICAL) reduce 530) ((T_IS_SMALLER_OR_EQUAL) reduce 530) ((T_LOGICAL_AND) reduce 530) ((T_LOGICAL_OR) reduce 530) ((T_LOGICAL_XOR) reduce 530) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 551) ((T_OBJECT_OPERATOR) reduce 551) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 551) ((T_PIPE) reduce 530) ((T_POW) reduce 530) ((T_SL) reduce 530) ((T_SPACESHIP) reduce 530) ((T_SR) reduce 530) (("[") reduce 551) (("^") reduce 530) (("|") reduce 530)) 30878 ((("%") reduce 529) (("*") reduce 529) (("+") reduce 529) (("-") reduce 529) ((".") reduce 529) (("/") reduce 529) ((";") reduce 529) (("<") reduce 529) ((">") reduce 529) (("?") reduce 529) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 529) ((T_BOOLEAN_AND) reduce 529) ((T_BOOLEAN_OR) reduce 529) ((T_COALESCE) reduce 529) ((T_DOUBLE_ARROW) reduce 529) ((T_INSTANCEOF) reduce 529) ((T_IS_EQUAL) reduce 529) ((T_IS_GREATER_OR_EQUAL) reduce 529) ((T_IS_IDENTICAL) reduce 529) ((T_IS_NOT_EQUAL) reduce 529) ((T_IS_NOT_IDENTICAL) reduce 529) ((T_IS_SMALLER_OR_EQUAL) reduce 529) ((T_LOGICAL_AND) reduce 529) ((T_LOGICAL_OR) reduce 529) ((T_LOGICAL_XOR) reduce 529) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 554) ((T_OBJECT_OPERATOR) reduce 554) ((T_PIPE) reduce 529) ((T_POW) reduce 529) ((T_SL) reduce 529) ((T_SPACESHIP) reduce 529) ((T_SR) reduce 529) (("[") reduce 554) (("^") reduce 529) (("|") reduce 529)) 30879 ((("%") reduce 528) (("(") reduce 557) (("*") reduce 528) (("+") reduce 528) (("-") reduce 528) ((".") reduce 528) (("/") reduce 528) ((";") reduce 528) (("<") reduce 528) ((">") reduce 528) (("?") reduce 528) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 528) ((T_BOOLEAN_AND) reduce 528) ((T_BOOLEAN_OR) reduce 528) ((T_COALESCE) reduce 528) ((T_DOUBLE_ARROW) reduce 528) ((T_INSTANCEOF) reduce 528) ((T_IS_EQUAL) reduce 528) ((T_IS_GREATER_OR_EQUAL) reduce 528) ((T_IS_IDENTICAL) reduce 528) ((T_IS_NOT_EQUAL) reduce 528) ((T_IS_NOT_IDENTICAL) reduce 528) ((T_IS_SMALLER_OR_EQUAL) reduce 528) ((T_LOGICAL_AND) reduce 528) ((T_LOGICAL_OR) reduce 528) ((T_LOGICAL_XOR) reduce 528) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 550) ((T_OBJECT_OPERATOR) reduce 550) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 550) ((T_PIPE) reduce 528) ((T_POW) reduce 528) ((T_SL) reduce 528) ((T_SPACESHIP) reduce 528) ((T_SR) reduce 528) (("[") reduce 550) (("^") reduce 528) (("|") reduce 528)) 30880 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 480) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 480) ((T_LOGICAL_OR) reduce 480) ((T_LOGICAL_XOR) reduce 480) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 30883 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) ((";") reduce 563) (("<") reduce 563) (("=") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AND_EQUAL) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_COALESCE_EQUAL) reduce 563) ((T_CONCAT_EQUAL) reduce 563) ((T_DEC) reduce 563) ((T_DIV_EQUAL) reduce 563) ((T_DOUBLE_ARROW) reduce 563) ((T_INC) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_MINUS_EQUAL) reduce 563) ((T_MOD_EQUAL) reduce 563) ((T_MUL_EQUAL) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_OR_EQUAL) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_PLUS_EQUAL) reduce 563) ((T_POW) reduce 563) ((T_POW_EQUAL) reduce 563) ((T_SL) reduce 563) ((T_SL_EQUAL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) ((T_SR_EQUAL) reduce 563) ((T_XOR_EQUAL) reduce 563) (("[") reduce 563) (("^") reduce 563) (("|") reduce 563)) 30884 ((("%") reduce 484) (("*") reduce 484) (("+") reduce 484) (("-") reduce 484) ((".") reduce 484) (("/") reduce 484) ((";") reduce 484) (("<") reduce 484) ((">") reduce 484) (("?") reduce 484) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 484) ((T_BOOLEAN_AND) reduce 484) ((T_BOOLEAN_OR) reduce 484) ((T_COALESCE) reduce 484) ((T_DOUBLE_ARROW) reduce 484) ((T_INSTANCEOF) reduce 484) ((T_IS_EQUAL) reduce 484) ((T_IS_GREATER_OR_EQUAL) reduce 484) ((T_IS_IDENTICAL) reduce 484) ((T_IS_NOT_EQUAL) reduce 484) ((T_IS_NOT_IDENTICAL) reduce 484) ((T_IS_SMALLER_OR_EQUAL) reduce 484) ((T_LOGICAL_AND) reduce 484) ((T_LOGICAL_OR) reduce 484) ((T_LOGICAL_XOR) reduce 484) ((T_PIPE) reduce 484) ((T_POW) reduce 484) ((T_SL) reduce 484) ((T_SPACESHIP) reduce 484) ((T_SR) reduce 484) (("^") reduce 484) (("|") reduce 484)) 30885 ((("%") reduce 466) (("*") reduce 466) (("+") reduce 466) (("-") reduce 466) ((".") reduce 466) (("/") reduce 466) ((";") reduce 466) (("<") reduce 466) ((">") reduce 466) (("?") reduce 466) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 466) ((T_BOOLEAN_AND) reduce 466) ((T_BOOLEAN_OR) reduce 466) ((T_COALESCE) reduce 466) ((T_DOUBLE_ARROW) reduce 466) ((T_INSTANCEOF) reduce 466) ((T_IS_EQUAL) reduce 466) ((T_IS_GREATER_OR_EQUAL) reduce 466) ((T_IS_IDENTICAL) reduce 466) ((T_IS_NOT_EQUAL) reduce 466) ((T_IS_NOT_IDENTICAL) reduce 466) ((T_IS_SMALLER_OR_EQUAL) reduce 466) ((T_LOGICAL_AND) reduce 466) ((T_LOGICAL_OR) reduce 466) ((T_LOGICAL_XOR) reduce 466) ((T_PIPE) reduce 466) ((T_POW) reduce 466) ((T_SL) reduce 466) ((T_SPACESHIP) reduce 466) ((T_SR) reduce 466) (("^") reduce 466) (("|") reduce 466)) 30886 ((("%") reduce 488) (("*") reduce 488) (("+") reduce 488) (("-") reduce 488) ((".") reduce 488) (("/") reduce 488) ((";") reduce 488) (("<") reduce 488) ((">") reduce 488) (("?") reduce 488) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 488) ((T_BOOLEAN_AND) reduce 488) ((T_BOOLEAN_OR) reduce 488) ((T_COALESCE) reduce 488) ((T_DOUBLE_ARROW) reduce 488) ((T_INSTANCEOF) reduce 488) ((T_IS_EQUAL) reduce 488) ((T_IS_GREATER_OR_EQUAL) reduce 488) ((T_IS_IDENTICAL) reduce 488) ((T_IS_NOT_EQUAL) reduce 488) ((T_IS_NOT_IDENTICAL) reduce 488) ((T_IS_SMALLER_OR_EQUAL) reduce 488) ((T_LOGICAL_AND) reduce 488) ((T_LOGICAL_OR) reduce 488) ((T_LOGICAL_XOR) reduce 488) ((T_PIPE) reduce 488) ((T_POW) reduce 488) ((T_SL) reduce 488) ((T_SPACESHIP) reduce 488) ((T_SR) reduce 488) (("^") reduce 488) (("|") reduce 488)) 30887 ((("%") reduce 531) (("(") shift) (("*") reduce 531) (("+") reduce 531) (("-") reduce 531) ((".") reduce 531) (("/") reduce 531) ((";") reduce 531) (("<") reduce 531) ((">") reduce 531) (("?") reduce 531) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 531) ((T_BOOLEAN_AND) reduce 531) ((T_BOOLEAN_OR) reduce 531) ((T_COALESCE) reduce 531) ((T_DOUBLE_ARROW) reduce 531) ((T_INSTANCEOF) reduce 531) ((T_IS_EQUAL) reduce 531) ((T_IS_GREATER_OR_EQUAL) reduce 531) ((T_IS_IDENTICAL) reduce 531) ((T_IS_NOT_EQUAL) reduce 531) ((T_IS_NOT_IDENTICAL) reduce 531) ((T_IS_SMALLER_OR_EQUAL) reduce 531) ((T_LOGICAL_AND) reduce 531) ((T_LOGICAL_OR) reduce 531) ((T_LOGICAL_XOR) reduce 531) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 531) ((T_OBJECT_OPERATOR) reduce 531) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 531) ((T_POW) reduce 531) ((T_SL) reduce 531) ((T_SPACESHIP) reduce 531) ((T_SR) reduce 531) (("[") reduce 531) (("^") reduce 531) (("|") reduce 531)) 30888 ((("%") reduce 461) (("(") reduce 558) (("*") reduce 461) (("+") reduce 461) (("-") reduce 461) ((".") reduce 461) (("/") reduce 461) ((";") reduce 461) (("<") reduce 461) ((">") reduce 461) (("?") reduce 461) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 461) ((T_BOOLEAN_AND) reduce 461) ((T_BOOLEAN_OR) reduce 461) ((T_COALESCE) reduce 461) ((T_DOUBLE_ARROW) reduce 461) ((T_INSTANCEOF) reduce 461) ((T_IS_EQUAL) reduce 461) ((T_IS_GREATER_OR_EQUAL) reduce 461) ((T_IS_IDENTICAL) reduce 461) ((T_IS_NOT_EQUAL) reduce 461) ((T_IS_NOT_IDENTICAL) reduce 461) ((T_IS_SMALLER_OR_EQUAL) reduce 461) ((T_LOGICAL_AND) reduce 461) ((T_LOGICAL_OR) reduce 461) ((T_LOGICAL_XOR) reduce 461) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 552) ((T_OBJECT_OPERATOR) reduce 552) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 552) ((T_PIPE) reduce 461) ((T_POW) reduce 461) ((T_SL) reduce 461) ((T_SPACESHIP) reduce 461) ((T_SR) reduce 461) (("[") reduce 552) (("^") reduce 461) (("|") reduce 461)) 30889 ((("%") reduce 462) (("*") reduce 462) (("+") reduce 462) (("-") reduce 462) ((".") reduce 462) (("/") reduce 462) ((";") reduce 462) (("<") reduce 462) ((">") reduce 462) (("?") reduce 462) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 462) ((T_BOOLEAN_AND) reduce 462) ((T_BOOLEAN_OR) reduce 462) ((T_COALESCE) reduce 462) ((T_DOUBLE_ARROW) reduce 462) ((T_INSTANCEOF) reduce 462) ((T_IS_EQUAL) reduce 462) ((T_IS_GREATER_OR_EQUAL) reduce 462) ((T_IS_IDENTICAL) reduce 462) ((T_IS_NOT_EQUAL) reduce 462) ((T_IS_NOT_IDENTICAL) reduce 462) ((T_IS_SMALLER_OR_EQUAL) reduce 462) ((T_LOGICAL_AND) reduce 462) ((T_LOGICAL_OR) reduce 462) ((T_LOGICAL_XOR) reduce 462) ((T_PIPE) reduce 462) ((T_POW) reduce 462) ((T_SL) reduce 462) ((T_SPACESHIP) reduce 462) ((T_SR) reduce 462) (("^") reduce 462) (("|") reduce 462)) 30890 ((("%") reduce 476) (("*") reduce 476) (("+") reduce 476) (("-") reduce 476) ((".") reduce 476) (("/") reduce 476) ((";") reduce 476) (("<") reduce 476) ((">") reduce 476) (("?") reduce 476) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 476) ((T_BOOLEAN_AND) reduce 476) ((T_BOOLEAN_OR) reduce 476) ((T_COALESCE) reduce 476) ((T_DOUBLE_ARROW) reduce 476) ((T_INSTANCEOF) reduce 476) ((T_IS_EQUAL) reduce 476) ((T_IS_GREATER_OR_EQUAL) reduce 476) ((T_IS_IDENTICAL) reduce 476) ((T_IS_NOT_EQUAL) reduce 476) ((T_IS_NOT_IDENTICAL) reduce 476) ((T_IS_SMALLER_OR_EQUAL) reduce 476) ((T_LOGICAL_AND) reduce 476) ((T_LOGICAL_OR) reduce 476) ((T_LOGICAL_XOR) reduce 476) ((T_PIPE) reduce 476) ((T_POW) reduce 476) ((T_SL) reduce 476) ((T_SPACESHIP) reduce 476) ((T_SR) reduce 476) (("^") reduce 476) (("|") reduce 476)) 30891 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) ((";") reduce 559) (("<") reduce 559) (("=") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AND_EQUAL) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_COALESCE_EQUAL) reduce 559) ((T_CONCAT_EQUAL) reduce 559) ((T_DEC) reduce 559) ((T_DIV_EQUAL) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INC) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_MINUS_EQUAL) reduce 559) ((T_MOD_EQUAL) reduce 559) ((T_MUL_EQUAL) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_OR_EQUAL) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_PLUS_EQUAL) reduce 559) ((T_POW) reduce 559) ((T_POW_EQUAL) reduce 559) ((T_SL) reduce 559) ((T_SL_EQUAL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) ((T_SR_EQUAL) reduce 559) ((T_XOR_EQUAL) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 30892 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((";") reduce 565) (("<") reduce 565) (("=") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AND_EQUAL) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_COALESCE_EQUAL) reduce 565) ((T_CONCAT_EQUAL) reduce 565) ((T_DEC) reduce 565) ((T_DIV_EQUAL) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INC) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_MINUS_EQUAL) reduce 565) ((T_MOD_EQUAL) reduce 565) ((T_MUL_EQUAL) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_OR_EQUAL) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_PLUS_EQUAL) reduce 565) ((T_POW) reduce 565) ((T_POW_EQUAL) reduce 565) ((T_SL) reduce 565) ((T_SL_EQUAL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) ((T_SR_EQUAL) reduce 565) ((T_XOR_EQUAL) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 30893 ((("%") reduce 403) (("*") reduce 403) (("+") reduce 403) (("-") reduce 403) ((".") reduce 403) (("/") reduce 403) ((";") reduce 403) (("<") reduce 403) (("=") shift) ((">") reduce 403) (("?") reduce 403) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 403) ((T_AND_EQUAL) shift) ((T_BOOLEAN_AND) reduce 403) ((T_BOOLEAN_OR) reduce 403) ((T_COALESCE) reduce 403) ((T_COALESCE_EQUAL) shift) ((T_CONCAT_EQUAL) shift) ((T_DEC) shift) ((T_DIV_EQUAL) shift) ((T_DOUBLE_ARROW) reduce 403) ((T_INC) shift) ((T_INSTANCEOF) reduce 403) ((T_IS_EQUAL) reduce 403) ((T_IS_GREATER_OR_EQUAL) reduce 403) ((T_IS_IDENTICAL) reduce 403) ((T_IS_NOT_EQUAL) reduce 403) ((T_IS_NOT_IDENTICAL) reduce 403) ((T_IS_SMALLER_OR_EQUAL) reduce 403) ((T_LOGICAL_AND) reduce 403) ((T_LOGICAL_OR) reduce 403) ((T_LOGICAL_XOR) reduce 403) ((T_MINUS_EQUAL) shift) ((T_MOD_EQUAL) shift) ((T_MUL_EQUAL) shift) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_OR_EQUAL) shift) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 403) ((T_PLUS_EQUAL) shift) ((T_POW) reduce 403) ((T_POW_EQUAL) shift) ((T_SL) reduce 403) ((T_SL_EQUAL) shift) ((T_SPACESHIP) reduce 403) ((T_SR) reduce 403) ((T_SR_EQUAL) shift) ((T_XOR_EQUAL) shift) (("[") reduce 548) (("^") reduce 403) (("|") reduce 403)) 30896 ((("%") reduce 448) (("*") reduce 448) (("+") reduce 448) (("-") reduce 448) ((".") reduce 448) (("/") reduce 448) ((";") reduce 448) (("<") reduce 448) ((">") reduce 448) (("?") reduce 448) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 448) ((T_BOOLEAN_AND) reduce 448) ((T_BOOLEAN_OR) reduce 448) ((T_COALESCE) reduce 448) ((T_DOUBLE_ARROW) reduce 448) ((T_INSTANCEOF) reduce 448) ((T_IS_EQUAL) reduce 448) ((T_IS_GREATER_OR_EQUAL) reduce 448) ((T_IS_IDENTICAL) reduce 448) ((T_IS_NOT_EQUAL) reduce 448) ((T_IS_NOT_IDENTICAL) reduce 448) ((T_IS_SMALLER_OR_EQUAL) reduce 448) ((T_LOGICAL_AND) reduce 448) ((T_LOGICAL_OR) reduce 448) ((T_LOGICAL_XOR) reduce 448) ((T_PIPE) reduce 448) ((T_POW) shift) ((T_SL) reduce 448) ((T_SPACESHIP) reduce 448) ((T_SR) reduce 448) (("^") reduce 448) (("|") reduce 448)) 30928 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 432) (("<") shift) ((">") shift) (("?") reduce 432) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 432) ((T_BOOLEAN_OR) reduce 432) ((T_COALESCE) reduce 432) ((T_DOUBLE_ARROW) reduce 432) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 432) ((T_LOGICAL_OR) reduce 432) ((T_LOGICAL_XOR) reduce 432) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") reduce 432)) 30929 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 435) (("<") shift) ((">") shift) (("?") reduce 435) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 435) ((T_BOOLEAN_OR) reduce 435) ((T_COALESCE) reduce 435) ((T_DOUBLE_ARROW) reduce 435) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 435) ((T_LOGICAL_OR) reduce 435) ((T_LOGICAL_XOR) reduce 435) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 435) (("|") reduce 435)) 30930 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 444) (("/") shift) ((";") reduce 444) (("<") reduce 444) ((">") reduce 444) (("?") reduce 444) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 444) ((T_BOOLEAN_AND) reduce 444) ((T_BOOLEAN_OR) reduce 444) ((T_COALESCE) reduce 444) ((T_DOUBLE_ARROW) reduce 444) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 444) ((T_IS_GREATER_OR_EQUAL) reduce 444) ((T_IS_IDENTICAL) reduce 444) ((T_IS_NOT_EQUAL) reduce 444) ((T_IS_NOT_IDENTICAL) reduce 444) ((T_IS_SMALLER_OR_EQUAL) reduce 444) ((T_LOGICAL_AND) reduce 444) ((T_LOGICAL_OR) reduce 444) ((T_LOGICAL_XOR) reduce 444) ((T_PIPE) reduce 444) ((T_POW) shift) ((T_SL) reduce 444) ((T_SPACESHIP) reduce 444) ((T_SR) reduce 444) (("^") reduce 444) (("|") reduce 444)) 30931 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 458) (("<") shift) ((">") shift) (("?") reduce 458) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 458) ((T_BOOLEAN_AND) reduce 458) ((T_BOOLEAN_OR) reduce 458) ((T_COALESCE) reduce 458) ((T_DOUBLE_ARROW) reduce 458) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 458) ((T_LOGICAL_OR) reduce 458) ((T_LOGICAL_XOR) reduce 458) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 458) (("|") reduce 458)) 30932 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 443) (("/") shift) ((";") reduce 443) (("<") reduce 443) ((">") reduce 443) (("?") reduce 443) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 443) ((T_BOOLEAN_AND) reduce 443) ((T_BOOLEAN_OR) reduce 443) ((T_COALESCE) reduce 443) ((T_DOUBLE_ARROW) reduce 443) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 443) ((T_IS_GREATER_OR_EQUAL) reduce 443) ((T_IS_IDENTICAL) reduce 443) ((T_IS_NOT_EQUAL) reduce 443) ((T_IS_NOT_IDENTICAL) reduce 443) ((T_IS_SMALLER_OR_EQUAL) reduce 443) ((T_LOGICAL_AND) reduce 443) ((T_LOGICAL_OR) reduce 443) ((T_LOGICAL_XOR) reduce 443) ((T_PIPE) reduce 443) ((T_POW) shift) ((T_SL) reduce 443) ((T_SPACESHIP) reduce 443) ((T_SR) reduce 443) (("^") reduce 443) (("|") reduce 443)) 30933 ((("%") reduce 440) (("*") reduce 440) (("+") reduce 440) (("-") reduce 440) ((".") reduce 440) (("/") reduce 440) ((";") reduce 440) (("<") reduce 440) ((">") reduce 440) (("?") reduce 440) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 440) ((T_BOOLEAN_AND) reduce 440) ((T_BOOLEAN_OR) reduce 440) ((T_COALESCE) reduce 440) ((T_DOUBLE_ARROW) reduce 440) ((T_INSTANCEOF) reduce 440) ((T_IS_EQUAL) reduce 440) ((T_IS_GREATER_OR_EQUAL) reduce 440) ((T_IS_IDENTICAL) reduce 440) ((T_IS_NOT_EQUAL) reduce 440) ((T_IS_NOT_IDENTICAL) reduce 440) ((T_IS_SMALLER_OR_EQUAL) reduce 440) ((T_LOGICAL_AND) reduce 440) ((T_LOGICAL_OR) reduce 440) ((T_LOGICAL_XOR) reduce 440) ((T_PIPE) reduce 440) ((T_POW) shift) ((T_SL) reduce 440) ((T_SPACESHIP) reduce 440) ((T_SR) reduce 440) (("^") reduce 440) (("|") reduce 440)) 30934 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 453) (("<") reduce 453) ((">") reduce 453) (("?") reduce 453) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 453) ((T_BOOLEAN_AND) reduce 453) ((T_BOOLEAN_OR) reduce 453) ((T_COALESCE) reduce 453) ((T_DOUBLE_ARROW) reduce 453) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 453) ((T_IS_GREATER_OR_EQUAL) reduce 453) ((T_IS_IDENTICAL) reduce 453) ((T_IS_NOT_EQUAL) reduce 453) ((T_IS_NOT_IDENTICAL) reduce 453) ((T_IS_SMALLER_OR_EQUAL) reduce 453) ((T_LOGICAL_AND) reduce 453) ((T_LOGICAL_OR) reduce 453) ((T_LOGICAL_XOR) reduce 453) ((T_PIPE) reduce 453) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 453) ((T_SR) shift) (("^") reduce 453) (("|") reduce 453)) 30935 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 431) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 431) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 431) ((T_LOGICAL_XOR) reduce 431) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 30936 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 429) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 429) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) reduce 429) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 30937 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 430) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 430) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 430) ((T_LOGICAL_OR) reduce 430) ((T_LOGICAL_XOR) reduce 430) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 30938 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 455) (("<") shift) ((">") shift) (("?") reduce 455) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 455) ((T_BOOLEAN_AND) reduce 455) ((T_BOOLEAN_OR) reduce 455) ((T_COALESCE) reduce 455) ((T_DOUBLE_ARROW) reduce 455) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 455) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 455) ((T_IS_NOT_EQUAL) reduce 455) ((T_IS_NOT_IDENTICAL) reduce 455) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 455) ((T_LOGICAL_OR) reduce 455) ((T_LOGICAL_XOR) reduce 455) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 455) ((T_SR) shift) (("^") reduce 455) (("|") reduce 455)) 30939 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 450) (("<") shift) ((">") shift) (("?") reduce 450) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 450) ((T_BOOLEAN_AND) reduce 450) ((T_BOOLEAN_OR) reduce 450) ((T_COALESCE) reduce 450) ((T_DOUBLE_ARROW) reduce 450) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 450) ((T_LOGICAL_OR) reduce 450) ((T_LOGICAL_XOR) reduce 450) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 450) (("|") reduce 450)) 30940 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 452) (("<") shift) ((">") shift) (("?") reduce 452) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 452) ((T_BOOLEAN_AND) reduce 452) ((T_BOOLEAN_OR) reduce 452) ((T_COALESCE) reduce 452) ((T_DOUBLE_ARROW) reduce 452) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 452) ((T_LOGICAL_OR) reduce 452) ((T_LOGICAL_XOR) reduce 452) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 452) (("|") reduce 452)) 30941 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 449) (("<") shift) ((">") shift) (("?") reduce 449) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 449) ((T_BOOLEAN_AND) reduce 449) ((T_BOOLEAN_OR) reduce 449) ((T_COALESCE) reduce 449) ((T_DOUBLE_ARROW) reduce 449) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 449) ((T_LOGICAL_OR) reduce 449) ((T_LOGICAL_XOR) reduce 449) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 449) (("|") reduce 449)) 30942 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 457) (("<") shift) ((">") shift) (("?") reduce 457) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 457) ((T_BOOLEAN_AND) reduce 457) ((T_BOOLEAN_OR) reduce 457) ((T_COALESCE) reduce 457) ((T_DOUBLE_ARROW) reduce 457) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 457) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 457) ((T_IS_NOT_EQUAL) reduce 457) ((T_IS_NOT_IDENTICAL) reduce 457) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 457) ((T_LOGICAL_OR) reduce 457) ((T_LOGICAL_XOR) reduce 457) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 457) ((T_SR) shift) (("^") reduce 457) (("|") reduce 457)) 30943 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 451) (("<") shift) ((">") shift) (("?") reduce 451) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 451) ((T_BOOLEAN_AND) reduce 451) ((T_BOOLEAN_OR) reduce 451) ((T_COALESCE) reduce 451) ((T_DOUBLE_ARROW) reduce 451) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 451) ((T_LOGICAL_OR) reduce 451) ((T_LOGICAL_XOR) reduce 451) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 451) (("|") reduce 451)) 30946 ((("%") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 30947 ((("%") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 30948 ((("%") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 30949 ((("%") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((";") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 30950 ((("%") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 30951 ((("%") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) ((";") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 30952 ((("%") reduce 511) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((";") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("|") reduce 511)) 30953 ((("%") reduce 459) (("*") reduce 459) (("+") reduce 459) (("-") reduce 459) ((".") reduce 459) (("/") reduce 459) ((";") reduce 459) (("<") reduce 459) ((">") reduce 459) (("?") reduce 459) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 459) ((T_BOOLEAN_AND) reduce 459) ((T_BOOLEAN_OR) reduce 459) ((T_COALESCE) reduce 459) ((T_DOUBLE_ARROW) reduce 459) ((T_INSTANCEOF) reduce 459) ((T_IS_EQUAL) reduce 459) ((T_IS_GREATER_OR_EQUAL) reduce 459) ((T_IS_IDENTICAL) reduce 459) ((T_IS_NOT_EQUAL) reduce 459) ((T_IS_NOT_IDENTICAL) reduce 459) ((T_IS_SMALLER_OR_EQUAL) reduce 459) ((T_LOGICAL_AND) reduce 459) ((T_LOGICAL_OR) reduce 459) ((T_LOGICAL_XOR) reduce 459) ((T_PIPE) reduce 459) ((T_POW) reduce 459) ((T_SL) reduce 459) ((T_SPACESHIP) reduce 459) ((T_SR) reduce 459) (("^") reduce 459) (("|") reduce 459)) 30954 ((("%") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((";") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 30955 ((("%") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((";") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("|") reduce 512)) 30956 ((("%") reduce 573) (("*") reduce 573) (("+") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) ((";") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_DOUBLE_ARROW) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("|") reduce 573)) 30962 ((("%") reduce 574) (("*") reduce 574) (("+") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) ((";") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_DOUBLE_ARROW) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("|") reduce 574)) 30963 ((("%") reduce 578) (("*") reduce 578) (("+") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) ((";") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_DOUBLE_ARROW) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("|") reduce 578)) 30964 ((("%") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) ((";") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 30965 ((("%") reduce 575) (("*") reduce 575) (("+") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) ((";") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_DOUBLE_ARROW) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("|") reduce 575)) 30966 ((("%") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) ((";") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 30969 ((("%") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) ((";") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 30970 ((("%") reduce 576) (("*") reduce 576) (("+") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) ((";") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_DOUBLE_ARROW) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("|") reduce 576)) 30972 ((("%") reduce 577) (("*") reduce 577) (("+") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) ((";") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_DOUBLE_ARROW) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("|") reduce 577)) 30974 ((("%") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((";") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("|") reduce 513)) 30975 ((("%") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) ((";") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 30978 ((("%") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) ((";") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 30979 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 465) (("<") shift) ((">") shift) (("?") reduce 465) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 465) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 465) ((T_LOGICAL_OR) reduce 465) ((T_LOGICAL_XOR) reduce 465) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 30980 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 427) (("<") shift) ((">") shift) (("?") reduce 427) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) reduce 427) ((T_COALESCE) reduce 427) ((T_DOUBLE_ARROW) reduce 427) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 427) ((T_LOGICAL_OR) reduce 427) ((T_LOGICAL_XOR) reduce 427) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 30981 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 428) (("<") shift) ((">") shift) (("?") reduce 428) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) reduce 428) ((T_BOOLEAN_OR) reduce 428) ((T_COALESCE) reduce 428) ((T_DOUBLE_ARROW) reduce 428) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 428) ((T_LOGICAL_OR) reduce 428) ((T_LOGICAL_XOR) reduce 428) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 30982 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 433) (("<") shift) ((">") shift) (("?") reduce 433) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 433) ((T_BOOLEAN_AND) reduce 433) ((T_BOOLEAN_OR) reduce 433) ((T_COALESCE) reduce 433) ((T_DOUBLE_ARROW) reduce 433) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 433) ((T_LOGICAL_OR) reduce 433) ((T_LOGICAL_XOR) reduce 433) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 433) (("|") reduce 433)) 30983 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 434) (("<") shift) ((">") shift) (("?") reduce 434) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 434) ((T_BOOLEAN_AND) reduce 434) ((T_BOOLEAN_OR) reduce 434) ((T_COALESCE) reduce 434) ((T_DOUBLE_ARROW) reduce 434) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 434) ((T_LOGICAL_OR) reduce 434) ((T_LOGICAL_XOR) reduce 434) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") reduce 434) (("|") reduce 434)) 30987 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 463) (("<") shift) ((">") shift) (("?") reduce 463) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 463) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 463) ((T_LOGICAL_OR) reduce 463) ((T_LOGICAL_XOR) reduce 463) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 30988 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 464) (("<") shift) ((">") shift) (("?") reduce 464) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 464) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 464) ((T_LOGICAL_OR) reduce 464) ((T_LOGICAL_XOR) reduce 464) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 30989 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 456) (("<") shift) ((">") shift) (("?") reduce 456) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 456) ((T_BOOLEAN_AND) reduce 456) ((T_BOOLEAN_OR) reduce 456) ((T_COALESCE) reduce 456) ((T_DOUBLE_ARROW) reduce 456) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 456) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 456) ((T_IS_NOT_EQUAL) reduce 456) ((T_IS_NOT_IDENTICAL) reduce 456) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 456) ((T_LOGICAL_OR) reduce 456) ((T_LOGICAL_XOR) reduce 456) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 456) ((T_SR) shift) (("^") reduce 456) (("|") reduce 456)) 30990 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 454) (("<") shift) ((">") shift) (("?") reduce 454) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 454) ((T_BOOLEAN_AND) reduce 454) ((T_BOOLEAN_OR) reduce 454) ((T_COALESCE) reduce 454) ((T_DOUBLE_ARROW) reduce 454) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 454) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) reduce 454) ((T_IS_NOT_EQUAL) reduce 454) ((T_IS_NOT_IDENTICAL) reduce 454) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 454) ((T_LOGICAL_OR) reduce 454) ((T_LOGICAL_XOR) reduce 454) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 454) ((T_SR) shift) (("^") reduce 454) (("|") reduce 454)) 30991 ((("%") reduce 441) (("*") reduce 441) (("+") reduce 441) (("-") reduce 441) ((".") reduce 441) (("/") reduce 441) ((";") reduce 441) (("<") reduce 441) ((">") reduce 441) (("?") reduce 441) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 441) ((T_BOOLEAN_AND) reduce 441) ((T_BOOLEAN_OR) reduce 441) ((T_COALESCE) reduce 441) ((T_DOUBLE_ARROW) reduce 441) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 441) ((T_IS_GREATER_OR_EQUAL) reduce 441) ((T_IS_IDENTICAL) reduce 441) ((T_IS_NOT_EQUAL) reduce 441) ((T_IS_NOT_IDENTICAL) reduce 441) ((T_IS_SMALLER_OR_EQUAL) reduce 441) ((T_LOGICAL_AND) reduce 441) ((T_LOGICAL_OR) reduce 441) ((T_LOGICAL_XOR) reduce 441) ((T_PIPE) reduce 441) ((T_POW) shift) ((T_SL) reduce 441) ((T_SPACESHIP) reduce 441) ((T_SR) reduce 441) (("^") reduce 441) (("|") reduce 441)) 30992 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") reduce 436) (("/") shift) ((";") reduce 436) (("<") reduce 436) ((">") reduce 436) (("?") reduce 436) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 436) ((T_BOOLEAN_AND) reduce 436) ((T_BOOLEAN_OR) reduce 436) ((T_COALESCE) reduce 436) ((T_DOUBLE_ARROW) reduce 436) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 436) ((T_IS_GREATER_OR_EQUAL) reduce 436) ((T_IS_IDENTICAL) reduce 436) ((T_IS_NOT_EQUAL) reduce 436) ((T_IS_NOT_IDENTICAL) reduce 436) ((T_IS_SMALLER_OR_EQUAL) reduce 436) ((T_LOGICAL_AND) reduce 436) ((T_LOGICAL_OR) reduce 436) ((T_LOGICAL_XOR) reduce 436) ((T_PIPE) reduce 436) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) reduce 436) ((T_SR) shift) (("^") reduce 436) (("|") reduce 436)) 30993 ((("%") shift) (("*") shift) (("+") reduce 438) (("-") reduce 438) ((".") reduce 438) (("/") shift) ((";") reduce 438) (("<") reduce 438) ((">") reduce 438) (("?") reduce 438) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 438) ((T_BOOLEAN_AND) reduce 438) ((T_BOOLEAN_OR) reduce 438) ((T_COALESCE) reduce 438) ((T_DOUBLE_ARROW) reduce 438) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 438) ((T_IS_GREATER_OR_EQUAL) reduce 438) ((T_IS_IDENTICAL) reduce 438) ((T_IS_NOT_EQUAL) reduce 438) ((T_IS_NOT_IDENTICAL) reduce 438) ((T_IS_SMALLER_OR_EQUAL) reduce 438) ((T_LOGICAL_AND) reduce 438) ((T_LOGICAL_OR) reduce 438) ((T_LOGICAL_XOR) reduce 438) ((T_PIPE) reduce 438) ((T_POW) shift) ((T_SL) reduce 438) ((T_SPACESHIP) reduce 438) ((T_SR) reduce 438) (("^") reduce 438) (("|") reduce 438)) 30994 ((("%") shift) (("*") shift) (("+") reduce 437) (("-") reduce 437) ((".") reduce 437) (("/") shift) ((";") reduce 437) (("<") reduce 437) ((">") reduce 437) (("?") reduce 437) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 437) ((T_BOOLEAN_AND) reduce 437) ((T_BOOLEAN_OR) reduce 437) ((T_COALESCE) reduce 437) ((T_DOUBLE_ARROW) reduce 437) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 437) ((T_IS_GREATER_OR_EQUAL) reduce 437) ((T_IS_IDENTICAL) reduce 437) ((T_IS_NOT_EQUAL) reduce 437) ((T_IS_NOT_IDENTICAL) reduce 437) ((T_IS_SMALLER_OR_EQUAL) reduce 437) ((T_LOGICAL_AND) reduce 437) ((T_LOGICAL_OR) reduce 437) ((T_LOGICAL_XOR) reduce 437) ((T_PIPE) reduce 437) ((T_POW) shift) ((T_SL) reduce 437) ((T_SPACESHIP) reduce 437) ((T_SR) reduce 437) (("^") reduce 437) (("|") reduce 437)) 30995 ((("%") reduce 439) (("*") reduce 439) (("+") reduce 439) (("-") reduce 439) ((".") reduce 439) (("/") reduce 439) ((";") reduce 439) (("<") reduce 439) ((">") reduce 439) (("?") reduce 439) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 439) ((T_BOOLEAN_AND) reduce 439) ((T_BOOLEAN_OR) reduce 439) ((T_COALESCE) reduce 439) ((T_DOUBLE_ARROW) reduce 439) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 439) ((T_IS_GREATER_OR_EQUAL) reduce 439) ((T_IS_IDENTICAL) reduce 439) ((T_IS_NOT_EQUAL) reduce 439) ((T_IS_NOT_IDENTICAL) reduce 439) ((T_IS_SMALLER_OR_EQUAL) reduce 439) ((T_LOGICAL_AND) reduce 439) ((T_LOGICAL_OR) reduce 439) ((T_LOGICAL_XOR) reduce 439) ((T_PIPE) reduce 439) ((T_POW) shift) ((T_SL) reduce 439) ((T_SPACESHIP) reduce 439) ((T_SR) reduce 439) (("^") reduce 439) (("|") reduce 439)) 30996 ((("%") reduce 442) (("*") reduce 442) (("+") reduce 442) (("-") reduce 442) ((".") reduce 442) (("/") reduce 442) ((";") reduce 442) (("<") reduce 442) ((">") reduce 442) (("?") reduce 442) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 442) ((T_BOOLEAN_AND) reduce 442) ((T_BOOLEAN_OR) reduce 442) ((T_COALESCE) reduce 442) ((T_DOUBLE_ARROW) reduce 442) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 442) ((T_IS_GREATER_OR_EQUAL) reduce 442) ((T_IS_IDENTICAL) reduce 442) ((T_IS_NOT_EQUAL) reduce 442) ((T_IS_NOT_IDENTICAL) reduce 442) ((T_IS_SMALLER_OR_EQUAL) reduce 442) ((T_LOGICAL_AND) reduce 442) ((T_LOGICAL_OR) reduce 442) ((T_LOGICAL_XOR) reduce 442) ((T_PIPE) reduce 442) ((T_POW) shift) ((T_SL) reduce 442) ((T_SPACESHIP) reduce 442) ((T_SR) reduce 442) (("^") reduce 442) (("|") reduce 442)) 30998 ((("%") reduce 74) (("(") reduce 74) (("*") reduce 74) (("+") reduce 74) (("-") reduce 74) ((".") reduce 74) (("/") reduce 74) ((";") reduce 74) (("<") reduce 74) ((">") reduce 74) (("?") reduce 74) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 74) ((T_BOOLEAN_AND) reduce 74) ((T_BOOLEAN_OR) reduce 74) ((T_COALESCE) reduce 74) ((T_DOUBLE_ARROW) reduce 74) ((T_INSTANCEOF) reduce 74) ((T_IS_EQUAL) reduce 74) ((T_IS_GREATER_OR_EQUAL) reduce 74) ((T_IS_IDENTICAL) reduce 74) ((T_IS_NOT_EQUAL) reduce 74) ((T_IS_NOT_IDENTICAL) reduce 74) ((T_IS_SMALLER_OR_EQUAL) reduce 74) ((T_LOGICAL_AND) reduce 74) ((T_LOGICAL_OR) reduce 74) ((T_LOGICAL_XOR) reduce 74) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 74) ((T_OBJECT_OPERATOR) reduce 74) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 74) ((T_PIPE) reduce 74) ((T_POW) reduce 74) ((T_SL) reduce 74) ((T_SPACESHIP) reduce 74) ((T_SR) reduce 74) (("[") reduce 74) (("^") reduce 74) (("|") reduce 74)) 30999 ((("%") reduce 52) (("(") reduce 52) (("*") reduce 52) (("+") reduce 52) (("-") reduce 52) ((".") reduce 52) (("/") reduce 52) ((";") reduce 52) (("<") reduce 52) ((">") reduce 52) (("?") reduce 52) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 52) ((T_BOOLEAN_AND) reduce 52) ((T_BOOLEAN_OR) reduce 52) ((T_COALESCE) reduce 52) ((T_DOUBLE_ARROW) reduce 52) ((T_INSTANCEOF) reduce 52) ((T_IS_EQUAL) reduce 52) ((T_IS_GREATER_OR_EQUAL) reduce 52) ((T_IS_IDENTICAL) reduce 52) ((T_IS_NOT_EQUAL) reduce 52) ((T_IS_NOT_IDENTICAL) reduce 52) ((T_IS_SMALLER_OR_EQUAL) reduce 52) ((T_LOGICAL_AND) reduce 52) ((T_LOGICAL_OR) reduce 52) ((T_LOGICAL_XOR) reduce 52) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 52) ((T_OBJECT_OPERATOR) reduce 52) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 52) ((T_PIPE) reduce 52) ((T_POW) reduce 52) ((T_SL) reduce 52) ((T_SPACESHIP) reduce 52) ((T_SR) reduce 52) (("[") reduce 52) (("^") reduce 52) (("|") reduce 52)) 31000 ((("%") reduce 27) (("(") reduce 27) (("*") reduce 27) (("+") reduce 27) (("-") reduce 27) ((".") reduce 27) (("/") reduce 27) ((";") reduce 27) (("<") reduce 27) ((">") reduce 27) (("?") reduce 27) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 27) ((T_BOOLEAN_AND) reduce 27) ((T_BOOLEAN_OR) reduce 27) ((T_COALESCE) reduce 27) ((T_DOUBLE_ARROW) reduce 27) ((T_INSTANCEOF) reduce 27) ((T_IS_EQUAL) reduce 27) ((T_IS_GREATER_OR_EQUAL) reduce 27) ((T_IS_IDENTICAL) reduce 27) ((T_IS_NOT_EQUAL) reduce 27) ((T_IS_NOT_IDENTICAL) reduce 27) ((T_IS_SMALLER_OR_EQUAL) reduce 27) ((T_LOGICAL_AND) reduce 27) ((T_LOGICAL_OR) reduce 27) ((T_LOGICAL_XOR) reduce 27) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 27) ((T_OBJECT_OPERATOR) reduce 27) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 27) ((T_PIPE) reduce 27) ((T_POW) reduce 27) ((T_SL) reduce 27) ((T_SPACESHIP) reduce 27) ((T_SR) reduce 27) (("[") reduce 27) (("^") reduce 27) (("|") reduce 27)) 31001 ((("%") reduce 51) (("(") reduce 51) (("*") reduce 51) (("+") reduce 51) (("-") reduce 51) ((".") reduce 51) (("/") reduce 51) ((";") reduce 51) (("<") reduce 51) ((">") reduce 51) (("?") reduce 51) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 51) ((T_BOOLEAN_AND) reduce 51) ((T_BOOLEAN_OR) reduce 51) ((T_COALESCE) reduce 51) ((T_DOUBLE_ARROW) reduce 51) ((T_INSTANCEOF) reduce 51) ((T_IS_EQUAL) reduce 51) ((T_IS_GREATER_OR_EQUAL) reduce 51) ((T_IS_IDENTICAL) reduce 51) ((T_IS_NOT_EQUAL) reduce 51) ((T_IS_NOT_IDENTICAL) reduce 51) ((T_IS_SMALLER_OR_EQUAL) reduce 51) ((T_LOGICAL_AND) reduce 51) ((T_LOGICAL_OR) reduce 51) ((T_LOGICAL_XOR) reduce 51) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 51) ((T_OBJECT_OPERATOR) reduce 51) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 51) ((T_PIPE) reduce 51) ((T_POW) reduce 51) ((T_SL) reduce 51) ((T_SPACESHIP) reduce 51) ((T_SR) reduce 51) (("[") reduce 51) (("^") reduce 51) (("|") reduce 51)) 31002 ((("%") reduce 53) (("(") reduce 53) (("*") reduce 53) (("+") reduce 53) (("-") reduce 53) ((".") reduce 53) (("/") reduce 53) ((";") reduce 53) (("<") reduce 53) ((">") reduce 53) (("?") reduce 53) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 53) ((T_BOOLEAN_AND) reduce 53) ((T_BOOLEAN_OR) reduce 53) ((T_COALESCE) reduce 53) ((T_DOUBLE_ARROW) reduce 53) ((T_INSTANCEOF) reduce 53) ((T_IS_EQUAL) reduce 53) ((T_IS_GREATER_OR_EQUAL) reduce 53) ((T_IS_IDENTICAL) reduce 53) ((T_IS_NOT_EQUAL) reduce 53) ((T_IS_NOT_IDENTICAL) reduce 53) ((T_IS_SMALLER_OR_EQUAL) reduce 53) ((T_LOGICAL_AND) reduce 53) ((T_LOGICAL_OR) reduce 53) ((T_LOGICAL_XOR) reduce 53) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 53) ((T_OBJECT_OPERATOR) reduce 53) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 53) ((T_PIPE) reduce 53) ((T_POW) reduce 53) ((T_SL) reduce 53) ((T_SPACESHIP) reduce 53) ((T_SR) reduce 53) (("[") reduce 53) (("^") reduce 53) (("|") reduce 53)) 31003 ((("%") reduce 49) (("(") reduce 49) (("*") reduce 49) (("+") reduce 49) (("-") reduce 49) ((".") reduce 49) (("/") reduce 49) ((";") reduce 49) (("<") reduce 49) ((">") reduce 49) (("?") reduce 49) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 49) ((T_BOOLEAN_AND) reduce 49) ((T_BOOLEAN_OR) reduce 49) ((T_COALESCE) reduce 49) ((T_DOUBLE_ARROW) reduce 49) ((T_INSTANCEOF) reduce 49) ((T_IS_EQUAL) reduce 49) ((T_IS_GREATER_OR_EQUAL) reduce 49) ((T_IS_IDENTICAL) reduce 49) ((T_IS_NOT_EQUAL) reduce 49) ((T_IS_NOT_IDENTICAL) reduce 49) ((T_IS_SMALLER_OR_EQUAL) reduce 49) ((T_LOGICAL_AND) reduce 49) ((T_LOGICAL_OR) reduce 49) ((T_LOGICAL_XOR) reduce 49) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 49) ((T_OBJECT_OPERATOR) reduce 49) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 49) ((T_PIPE) reduce 49) ((T_POW) reduce 49) ((T_SL) reduce 49) ((T_SPACESHIP) reduce 49) ((T_SR) reduce 49) (("[") reduce 49) (("^") reduce 49) (("|") reduce 49)) 31004 ((("%") reduce 29) (("(") reduce 29) (("*") reduce 29) (("+") reduce 29) (("-") reduce 29) ((".") reduce 29) (("/") reduce 29) ((";") reduce 29) (("<") reduce 29) ((">") reduce 29) (("?") reduce 29) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 29) ((T_BOOLEAN_AND) reduce 29) ((T_BOOLEAN_OR) reduce 29) ((T_COALESCE) reduce 29) ((T_DOUBLE_ARROW) reduce 29) ((T_INSTANCEOF) reduce 29) ((T_IS_EQUAL) reduce 29) ((T_IS_GREATER_OR_EQUAL) reduce 29) ((T_IS_IDENTICAL) reduce 29) ((T_IS_NOT_EQUAL) reduce 29) ((T_IS_NOT_IDENTICAL) reduce 29) ((T_IS_SMALLER_OR_EQUAL) reduce 29) ((T_LOGICAL_AND) reduce 29) ((T_LOGICAL_OR) reduce 29) ((T_LOGICAL_XOR) reduce 29) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 29) ((T_OBJECT_OPERATOR) reduce 29) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 29) ((T_PIPE) reduce 29) ((T_POW) reduce 29) ((T_SL) reduce 29) ((T_SPACESHIP) reduce 29) ((T_SR) reduce 29) (("[") reduce 29) (("^") reduce 29) (("|") reduce 29)) 31005 ((("%") reduce 59) (("(") reduce 59) (("*") reduce 59) (("+") reduce 59) (("-") reduce 59) ((".") reduce 59) (("/") reduce 59) ((";") reduce 59) (("<") reduce 59) ((">") reduce 59) (("?") reduce 59) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 59) ((T_BOOLEAN_AND) reduce 59) ((T_BOOLEAN_OR) reduce 59) ((T_COALESCE) reduce 59) ((T_DOUBLE_ARROW) reduce 59) ((T_INSTANCEOF) reduce 59) ((T_IS_EQUAL) reduce 59) ((T_IS_GREATER_OR_EQUAL) reduce 59) ((T_IS_IDENTICAL) reduce 59) ((T_IS_NOT_EQUAL) reduce 59) ((T_IS_NOT_IDENTICAL) reduce 59) ((T_IS_SMALLER_OR_EQUAL) reduce 59) ((T_LOGICAL_AND) reduce 59) ((T_LOGICAL_OR) reduce 59) ((T_LOGICAL_XOR) reduce 59) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 59) ((T_OBJECT_OPERATOR) reduce 59) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 59) ((T_PIPE) reduce 59) ((T_POW) reduce 59) ((T_SL) reduce 59) ((T_SPACESHIP) reduce 59) ((T_SR) reduce 59) (("[") reduce 59) (("^") reduce 59) (("|") reduce 59)) 31006 ((("%") reduce 60) (("(") reduce 60) (("*") reduce 60) (("+") reduce 60) (("-") reduce 60) ((".") reduce 60) (("/") reduce 60) ((";") reduce 60) (("<") reduce 60) ((">") reduce 60) (("?") reduce 60) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 60) ((T_BOOLEAN_AND) reduce 60) ((T_BOOLEAN_OR) reduce 60) ((T_COALESCE) reduce 60) ((T_DOUBLE_ARROW) reduce 60) ((T_INSTANCEOF) reduce 60) ((T_IS_EQUAL) reduce 60) ((T_IS_GREATER_OR_EQUAL) reduce 60) ((T_IS_IDENTICAL) reduce 60) ((T_IS_NOT_EQUAL) reduce 60) ((T_IS_NOT_IDENTICAL) reduce 60) ((T_IS_SMALLER_OR_EQUAL) reduce 60) ((T_LOGICAL_AND) reduce 60) ((T_LOGICAL_OR) reduce 60) ((T_LOGICAL_XOR) reduce 60) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 60) ((T_OBJECT_OPERATOR) reduce 60) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 60) ((T_PIPE) reduce 60) ((T_POW) reduce 60) ((T_SL) reduce 60) ((T_SPACESHIP) reduce 60) ((T_SR) reduce 60) (("[") reduce 60) (("^") reduce 60) (("|") reduce 60)) 31007 ((("%") reduce 11) (("(") reduce 11) (("*") reduce 11) (("+") reduce 11) (("-") reduce 11) ((".") reduce 11) (("/") reduce 11) ((";") reduce 11) (("<") reduce 11) ((">") reduce 11) (("?") reduce 11) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 11) ((T_BOOLEAN_AND) reduce 11) ((T_BOOLEAN_OR) reduce 11) ((T_COALESCE) reduce 11) ((T_DOUBLE_ARROW) reduce 11) ((T_INSTANCEOF) reduce 11) ((T_IS_EQUAL) reduce 11) ((T_IS_GREATER_OR_EQUAL) reduce 11) ((T_IS_IDENTICAL) reduce 11) ((T_IS_NOT_EQUAL) reduce 11) ((T_IS_NOT_IDENTICAL) reduce 11) ((T_IS_SMALLER_OR_EQUAL) reduce 11) ((T_LOGICAL_AND) reduce 11) ((T_LOGICAL_OR) reduce 11) ((T_LOGICAL_XOR) reduce 11) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 11) ((T_OBJECT_OPERATOR) reduce 11) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 11) ((T_PIPE) reduce 11) ((T_POW) reduce 11) ((T_SL) reduce 11) ((T_SPACESHIP) reduce 11) ((T_SR) reduce 11) (("[") reduce 11) (("^") reduce 11) (("|") reduce 11)) 31008 ((("%") reduce 42) (("(") reduce 42) (("*") reduce 42) (("+") reduce 42) (("-") reduce 42) ((".") reduce 42) (("/") reduce 42) ((";") reduce 42) (("<") reduce 42) ((">") reduce 42) (("?") reduce 42) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 42) ((T_BOOLEAN_AND) reduce 42) ((T_BOOLEAN_OR) reduce 42) ((T_COALESCE) reduce 42) ((T_DOUBLE_ARROW) reduce 42) ((T_INSTANCEOF) reduce 42) ((T_IS_EQUAL) reduce 42) ((T_IS_GREATER_OR_EQUAL) reduce 42) ((T_IS_IDENTICAL) reduce 42) ((T_IS_NOT_EQUAL) reduce 42) ((T_IS_NOT_IDENTICAL) reduce 42) ((T_IS_SMALLER_OR_EQUAL) reduce 42) ((T_LOGICAL_AND) reduce 42) ((T_LOGICAL_OR) reduce 42) ((T_LOGICAL_XOR) reduce 42) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 42) ((T_OBJECT_OPERATOR) reduce 42) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 42) ((T_PIPE) reduce 42) ((T_POW) reduce 42) ((T_SL) reduce 42) ((T_SPACESHIP) reduce 42) ((T_SR) reduce 42) (("[") reduce 42) (("^") reduce 42) (("|") reduce 42)) 31009 ((("%") reduce 39) (("(") reduce 39) (("*") reduce 39) (("+") reduce 39) (("-") reduce 39) ((".") reduce 39) (("/") reduce 39) ((";") reduce 39) (("<") reduce 39) ((">") reduce 39) (("?") reduce 39) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 39) ((T_BOOLEAN_AND) reduce 39) ((T_BOOLEAN_OR) reduce 39) ((T_COALESCE) reduce 39) ((T_DOUBLE_ARROW) reduce 39) ((T_INSTANCEOF) reduce 39) ((T_IS_EQUAL) reduce 39) ((T_IS_GREATER_OR_EQUAL) reduce 39) ((T_IS_IDENTICAL) reduce 39) ((T_IS_NOT_EQUAL) reduce 39) ((T_IS_NOT_IDENTICAL) reduce 39) ((T_IS_SMALLER_OR_EQUAL) reduce 39) ((T_LOGICAL_AND) reduce 39) ((T_LOGICAL_OR) reduce 39) ((T_LOGICAL_XOR) reduce 39) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 39) ((T_OBJECT_OPERATOR) reduce 39) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 39) ((T_PIPE) reduce 39) ((T_POW) reduce 39) ((T_SL) reduce 39) ((T_SPACESHIP) reduce 39) ((T_SR) reduce 39) (("[") reduce 39) (("^") reduce 39) (("|") reduce 39)) 31010 ((("%") reduce 25) (("(") reduce 25) (("*") reduce 25) (("+") reduce 25) (("-") reduce 25) ((".") reduce 25) (("/") reduce 25) ((";") reduce 25) (("<") reduce 25) ((">") reduce 25) (("?") reduce 25) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 25) ((T_BOOLEAN_AND) reduce 25) ((T_BOOLEAN_OR) reduce 25) ((T_COALESCE) reduce 25) ((T_DOUBLE_ARROW) reduce 25) ((T_INSTANCEOF) reduce 25) ((T_IS_EQUAL) reduce 25) ((T_IS_GREATER_OR_EQUAL) reduce 25) ((T_IS_IDENTICAL) reduce 25) ((T_IS_NOT_EQUAL) reduce 25) ((T_IS_NOT_IDENTICAL) reduce 25) ((T_IS_SMALLER_OR_EQUAL) reduce 25) ((T_LOGICAL_AND) reduce 25) ((T_LOGICAL_OR) reduce 25) ((T_LOGICAL_XOR) reduce 25) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 25) ((T_OBJECT_OPERATOR) reduce 25) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 25) ((T_PIPE) reduce 25) ((T_POW) reduce 25) ((T_SL) reduce 25) ((T_SPACESHIP) reduce 25) ((T_SR) reduce 25) (("[") reduce 25) (("^") reduce 25) (("|") reduce 25)) 31011 ((("%") reduce 50) (("(") reduce 50) (("*") reduce 50) (("+") reduce 50) (("-") reduce 50) ((".") reduce 50) (("/") reduce 50) ((";") reduce 50) (("<") reduce 50) ((">") reduce 50) (("?") reduce 50) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 50) ((T_BOOLEAN_AND) reduce 50) ((T_BOOLEAN_OR) reduce 50) ((T_COALESCE) reduce 50) ((T_DOUBLE_ARROW) reduce 50) ((T_INSTANCEOF) reduce 50) ((T_IS_EQUAL) reduce 50) ((T_IS_GREATER_OR_EQUAL) reduce 50) ((T_IS_IDENTICAL) reduce 50) ((T_IS_NOT_EQUAL) reduce 50) ((T_IS_NOT_IDENTICAL) reduce 50) ((T_IS_SMALLER_OR_EQUAL) reduce 50) ((T_LOGICAL_AND) reduce 50) ((T_LOGICAL_OR) reduce 50) ((T_LOGICAL_XOR) reduce 50) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 50) ((T_OBJECT_OPERATOR) reduce 50) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 50) ((T_PIPE) reduce 50) ((T_POW) reduce 50) ((T_SL) reduce 50) ((T_SPACESHIP) reduce 50) ((T_SR) reduce 50) (("[") reduce 50) (("^") reduce 50) (("|") reduce 50)) 31012 ((("%") reduce 66) (("(") reduce 66) (("*") reduce 66) (("+") reduce 66) (("-") reduce 66) ((".") reduce 66) (("/") reduce 66) ((";") reduce 66) (("<") reduce 66) ((">") reduce 66) (("?") reduce 66) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 66) ((T_BOOLEAN_AND) reduce 66) ((T_BOOLEAN_OR) reduce 66) ((T_COALESCE) reduce 66) ((T_DOUBLE_ARROW) reduce 66) ((T_INSTANCEOF) reduce 66) ((T_IS_EQUAL) reduce 66) ((T_IS_GREATER_OR_EQUAL) reduce 66) ((T_IS_IDENTICAL) reduce 66) ((T_IS_NOT_EQUAL) reduce 66) ((T_IS_NOT_IDENTICAL) reduce 66) ((T_IS_SMALLER_OR_EQUAL) reduce 66) ((T_LOGICAL_AND) reduce 66) ((T_LOGICAL_OR) reduce 66) ((T_LOGICAL_XOR) reduce 66) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 66) ((T_OBJECT_OPERATOR) reduce 66) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 66) ((T_PIPE) reduce 66) ((T_POW) reduce 66) ((T_SL) reduce 66) ((T_SPACESHIP) reduce 66) ((T_SR) reduce 66) (("[") reduce 66) (("^") reduce 66) (("|") reduce 66)) 31013 ((("%") reduce 18) (("(") reduce 18) (("*") reduce 18) (("+") reduce 18) (("-") reduce 18) ((".") reduce 18) (("/") reduce 18) ((";") reduce 18) (("<") reduce 18) ((">") reduce 18) (("?") reduce 18) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 18) ((T_BOOLEAN_AND) reduce 18) ((T_BOOLEAN_OR) reduce 18) ((T_COALESCE) reduce 18) ((T_DOUBLE_ARROW) reduce 18) ((T_INSTANCEOF) reduce 18) ((T_IS_EQUAL) reduce 18) ((T_IS_GREATER_OR_EQUAL) reduce 18) ((T_IS_IDENTICAL) reduce 18) ((T_IS_NOT_EQUAL) reduce 18) ((T_IS_NOT_IDENTICAL) reduce 18) ((T_IS_SMALLER_OR_EQUAL) reduce 18) ((T_LOGICAL_AND) reduce 18) ((T_LOGICAL_OR) reduce 18) ((T_LOGICAL_XOR) reduce 18) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 18) ((T_OBJECT_OPERATOR) reduce 18) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 18) ((T_PIPE) reduce 18) ((T_POW) reduce 18) ((T_SL) reduce 18) ((T_SPACESHIP) reduce 18) ((T_SR) reduce 18) (("[") reduce 18) (("^") reduce 18) (("|") reduce 18)) 31014 ((("%") reduce 17) (("(") reduce 17) (("*") reduce 17) (("+") reduce 17) (("-") reduce 17) ((".") reduce 17) (("/") reduce 17) ((";") reduce 17) (("<") reduce 17) ((">") reduce 17) (("?") reduce 17) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 17) ((T_BOOLEAN_AND) reduce 17) ((T_BOOLEAN_OR) reduce 17) ((T_COALESCE) reduce 17) ((T_DOUBLE_ARROW) reduce 17) ((T_INSTANCEOF) reduce 17) ((T_IS_EQUAL) reduce 17) ((T_IS_GREATER_OR_EQUAL) reduce 17) ((T_IS_IDENTICAL) reduce 17) ((T_IS_NOT_EQUAL) reduce 17) ((T_IS_NOT_IDENTICAL) reduce 17) ((T_IS_SMALLER_OR_EQUAL) reduce 17) ((T_LOGICAL_AND) reduce 17) ((T_LOGICAL_OR) reduce 17) ((T_LOGICAL_XOR) reduce 17) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 17) ((T_OBJECT_OPERATOR) reduce 17) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 17) ((T_PIPE) reduce 17) ((T_POW) reduce 17) ((T_SL) reduce 17) ((T_SPACESHIP) reduce 17) ((T_SR) reduce 17) (("[") reduce 17) (("^") reduce 17) (("|") reduce 17)) 31015 ((("%") reduce 15) (("(") reduce 15) (("*") reduce 15) (("+") reduce 15) (("-") reduce 15) ((".") reduce 15) (("/") reduce 15) ((";") reduce 15) (("<") reduce 15) ((">") reduce 15) (("?") reduce 15) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 15) ((T_BOOLEAN_AND) reduce 15) ((T_BOOLEAN_OR) reduce 15) ((T_COALESCE) reduce 15) ((T_DOUBLE_ARROW) reduce 15) ((T_INSTANCEOF) reduce 15) ((T_IS_EQUAL) reduce 15) ((T_IS_GREATER_OR_EQUAL) reduce 15) ((T_IS_IDENTICAL) reduce 15) ((T_IS_NOT_EQUAL) reduce 15) ((T_IS_NOT_IDENTICAL) reduce 15) ((T_IS_SMALLER_OR_EQUAL) reduce 15) ((T_LOGICAL_AND) reduce 15) ((T_LOGICAL_OR) reduce 15) ((T_LOGICAL_XOR) reduce 15) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 15) ((T_OBJECT_OPERATOR) reduce 15) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 15) ((T_PIPE) reduce 15) ((T_POW) reduce 15) ((T_SL) reduce 15) ((T_SPACESHIP) reduce 15) ((T_SR) reduce 15) (("[") reduce 15) (("^") reduce 15) (("|") reduce 15)) 31016 ((("%") reduce 14) (("(") reduce 14) (("*") reduce 14) (("+") reduce 14) (("-") reduce 14) ((".") reduce 14) (("/") reduce 14) ((";") reduce 14) (("<") reduce 14) ((">") reduce 14) (("?") reduce 14) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 14) ((T_BOOLEAN_AND) reduce 14) ((T_BOOLEAN_OR) reduce 14) ((T_COALESCE) reduce 14) ((T_DOUBLE_ARROW) reduce 14) ((T_INSTANCEOF) reduce 14) ((T_IS_EQUAL) reduce 14) ((T_IS_GREATER_OR_EQUAL) reduce 14) ((T_IS_IDENTICAL) reduce 14) ((T_IS_NOT_EQUAL) reduce 14) ((T_IS_NOT_IDENTICAL) reduce 14) ((T_IS_SMALLER_OR_EQUAL) reduce 14) ((T_LOGICAL_AND) reduce 14) ((T_LOGICAL_OR) reduce 14) ((T_LOGICAL_XOR) reduce 14) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 14) ((T_OBJECT_OPERATOR) reduce 14) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 14) ((T_PIPE) reduce 14) ((T_POW) reduce 14) ((T_SL) reduce 14) ((T_SPACESHIP) reduce 14) ((T_SR) reduce 14) (("[") reduce 14) (("^") reduce 14) (("|") reduce 14)) 31017 ((("%") reduce 38) (("(") reduce 38) (("*") reduce 38) (("+") reduce 38) (("-") reduce 38) ((".") reduce 38) (("/") reduce 38) ((";") reduce 38) (("<") reduce 38) ((">") reduce 38) (("?") reduce 38) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 38) ((T_BOOLEAN_AND) reduce 38) ((T_BOOLEAN_OR) reduce 38) ((T_COALESCE) reduce 38) ((T_DOUBLE_ARROW) reduce 38) ((T_INSTANCEOF) reduce 38) ((T_IS_EQUAL) reduce 38) ((T_IS_GREATER_OR_EQUAL) reduce 38) ((T_IS_IDENTICAL) reduce 38) ((T_IS_NOT_EQUAL) reduce 38) ((T_IS_NOT_IDENTICAL) reduce 38) ((T_IS_SMALLER_OR_EQUAL) reduce 38) ((T_LOGICAL_AND) reduce 38) ((T_LOGICAL_OR) reduce 38) ((T_LOGICAL_XOR) reduce 38) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 38) ((T_OBJECT_OPERATOR) reduce 38) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 38) ((T_PIPE) reduce 38) ((T_POW) reduce 38) ((T_SL) reduce 38) ((T_SPACESHIP) reduce 38) ((T_SR) reduce 38) (("[") reduce 38) (("^") reduce 38) (("|") reduce 38)) 31018 ((("%") reduce 26) (("(") reduce 26) (("*") reduce 26) (("+") reduce 26) (("-") reduce 26) ((".") reduce 26) (("/") reduce 26) ((";") reduce 26) (("<") reduce 26) ((">") reduce 26) (("?") reduce 26) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 26) ((T_BOOLEAN_AND) reduce 26) ((T_BOOLEAN_OR) reduce 26) ((T_COALESCE) reduce 26) ((T_DOUBLE_ARROW) reduce 26) ((T_INSTANCEOF) reduce 26) ((T_IS_EQUAL) reduce 26) ((T_IS_GREATER_OR_EQUAL) reduce 26) ((T_IS_IDENTICAL) reduce 26) ((T_IS_NOT_EQUAL) reduce 26) ((T_IS_NOT_IDENTICAL) reduce 26) ((T_IS_SMALLER_OR_EQUAL) reduce 26) ((T_LOGICAL_AND) reduce 26) ((T_LOGICAL_OR) reduce 26) ((T_LOGICAL_XOR) reduce 26) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 26) ((T_OBJECT_OPERATOR) reduce 26) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 26) ((T_PIPE) reduce 26) ((T_POW) reduce 26) ((T_SL) reduce 26) ((T_SPACESHIP) reduce 26) ((T_SR) reduce 26) (("[") reduce 26) (("^") reduce 26) (("|") reduce 26)) 31019 ((("%") reduce 22) (("(") reduce 22) (("*") reduce 22) (("+") reduce 22) (("-") reduce 22) ((".") reduce 22) (("/") reduce 22) ((";") reduce 22) (("<") reduce 22) ((">") reduce 22) (("?") reduce 22) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 22) ((T_BOOLEAN_AND) reduce 22) ((T_BOOLEAN_OR) reduce 22) ((T_COALESCE) reduce 22) ((T_DOUBLE_ARROW) reduce 22) ((T_INSTANCEOF) reduce 22) ((T_IS_EQUAL) reduce 22) ((T_IS_GREATER_OR_EQUAL) reduce 22) ((T_IS_IDENTICAL) reduce 22) ((T_IS_NOT_EQUAL) reduce 22) ((T_IS_NOT_IDENTICAL) reduce 22) ((T_IS_SMALLER_OR_EQUAL) reduce 22) ((T_LOGICAL_AND) reduce 22) ((T_LOGICAL_OR) reduce 22) ((T_LOGICAL_XOR) reduce 22) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 22) ((T_OBJECT_OPERATOR) reduce 22) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 22) ((T_PIPE) reduce 22) ((T_POW) reduce 22) ((T_SL) reduce 22) ((T_SPACESHIP) reduce 22) ((T_SR) reduce 22) (("[") reduce 22) (("^") reduce 22) (("|") reduce 22)) 31020 ((("%") reduce 24) (("(") reduce 24) (("*") reduce 24) (("+") reduce 24) (("-") reduce 24) ((".") reduce 24) (("/") reduce 24) ((";") reduce 24) (("<") reduce 24) ((">") reduce 24) (("?") reduce 24) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 24) ((T_BOOLEAN_AND) reduce 24) ((T_BOOLEAN_OR) reduce 24) ((T_COALESCE) reduce 24) ((T_DOUBLE_ARROW) reduce 24) ((T_INSTANCEOF) reduce 24) ((T_IS_EQUAL) reduce 24) ((T_IS_GREATER_OR_EQUAL) reduce 24) ((T_IS_IDENTICAL) reduce 24) ((T_IS_NOT_EQUAL) reduce 24) ((T_IS_NOT_IDENTICAL) reduce 24) ((T_IS_SMALLER_OR_EQUAL) reduce 24) ((T_LOGICAL_AND) reduce 24) ((T_LOGICAL_OR) reduce 24) ((T_LOGICAL_XOR) reduce 24) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 24) ((T_OBJECT_OPERATOR) reduce 24) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 24) ((T_PIPE) reduce 24) ((T_POW) reduce 24) ((T_SL) reduce 24) ((T_SPACESHIP) reduce 24) ((T_SR) reduce 24) (("[") reduce 24) (("^") reduce 24) (("|") reduce 24)) 31021 ((("%") reduce 16) (("(") reduce 16) (("*") reduce 16) (("+") reduce 16) (("-") reduce 16) ((".") reduce 16) (("/") reduce 16) ((";") reduce 16) (("<") reduce 16) ((">") reduce 16) (("?") reduce 16) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 16) ((T_BOOLEAN_AND) reduce 16) ((T_BOOLEAN_OR) reduce 16) ((T_COALESCE) reduce 16) ((T_DOUBLE_ARROW) reduce 16) ((T_INSTANCEOF) reduce 16) ((T_IS_EQUAL) reduce 16) ((T_IS_GREATER_OR_EQUAL) reduce 16) ((T_IS_IDENTICAL) reduce 16) ((T_IS_NOT_EQUAL) reduce 16) ((T_IS_NOT_IDENTICAL) reduce 16) ((T_IS_SMALLER_OR_EQUAL) reduce 16) ((T_LOGICAL_AND) reduce 16) ((T_LOGICAL_OR) reduce 16) ((T_LOGICAL_XOR) reduce 16) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 16) ((T_OBJECT_OPERATOR) reduce 16) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 16) ((T_PIPE) reduce 16) ((T_POW) reduce 16) ((T_SL) reduce 16) ((T_SPACESHIP) reduce 16) ((T_SR) reduce 16) (("[") reduce 16) (("^") reduce 16) (("|") reduce 16)) 31022 ((("%") reduce 48) (("(") reduce 48) (("*") reduce 48) (("+") reduce 48) (("-") reduce 48) ((".") reduce 48) (("/") reduce 48) ((";") reduce 48) (("<") reduce 48) ((">") reduce 48) (("?") reduce 48) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 48) ((T_BOOLEAN_AND) reduce 48) ((T_BOOLEAN_OR) reduce 48) ((T_COALESCE) reduce 48) ((T_DOUBLE_ARROW) reduce 48) ((T_INSTANCEOF) reduce 48) ((T_IS_EQUAL) reduce 48) ((T_IS_GREATER_OR_EQUAL) reduce 48) ((T_IS_IDENTICAL) reduce 48) ((T_IS_NOT_EQUAL) reduce 48) ((T_IS_NOT_IDENTICAL) reduce 48) ((T_IS_SMALLER_OR_EQUAL) reduce 48) ((T_LOGICAL_AND) reduce 48) ((T_LOGICAL_OR) reduce 48) ((T_LOGICAL_XOR) reduce 48) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 48) ((T_OBJECT_OPERATOR) reduce 48) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 48) ((T_PIPE) reduce 48) ((T_POW) reduce 48) ((T_SL) reduce 48) ((T_SPACESHIP) reduce 48) ((T_SR) reduce 48) (("[") reduce 48) (("^") reduce 48) (("|") reduce 48)) 31023 ((("%") reduce 20) (("(") reduce 20) (("*") reduce 20) (("+") reduce 20) (("-") reduce 20) ((".") reduce 20) (("/") reduce 20) ((";") reduce 20) (("<") reduce 20) ((">") reduce 20) (("?") reduce 20) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 20) ((T_BOOLEAN_AND) reduce 20) ((T_BOOLEAN_OR) reduce 20) ((T_COALESCE) reduce 20) ((T_DOUBLE_ARROW) reduce 20) ((T_INSTANCEOF) reduce 20) ((T_IS_EQUAL) reduce 20) ((T_IS_GREATER_OR_EQUAL) reduce 20) ((T_IS_IDENTICAL) reduce 20) ((T_IS_NOT_EQUAL) reduce 20) ((T_IS_NOT_IDENTICAL) reduce 20) ((T_IS_SMALLER_OR_EQUAL) reduce 20) ((T_LOGICAL_AND) reduce 20) ((T_LOGICAL_OR) reduce 20) ((T_LOGICAL_XOR) reduce 20) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 20) ((T_OBJECT_OPERATOR) reduce 20) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 20) ((T_PIPE) reduce 20) ((T_POW) reduce 20) ((T_SL) reduce 20) ((T_SPACESHIP) reduce 20) ((T_SR) reduce 20) (("[") reduce 20) (("^") reduce 20) (("|") reduce 20)) 31024 ((("%") reduce 70) (("(") reduce 70) (("*") reduce 70) (("+") reduce 70) (("-") reduce 70) ((".") reduce 70) (("/") reduce 70) ((";") reduce 70) (("<") reduce 70) ((">") reduce 70) (("?") reduce 70) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 70) ((T_BOOLEAN_AND) reduce 70) ((T_BOOLEAN_OR) reduce 70) ((T_COALESCE) reduce 70) ((T_DOUBLE_ARROW) reduce 70) ((T_INSTANCEOF) reduce 70) ((T_IS_EQUAL) reduce 70) ((T_IS_GREATER_OR_EQUAL) reduce 70) ((T_IS_IDENTICAL) reduce 70) ((T_IS_NOT_EQUAL) reduce 70) ((T_IS_NOT_IDENTICAL) reduce 70) ((T_IS_SMALLER_OR_EQUAL) reduce 70) ((T_LOGICAL_AND) reduce 70) ((T_LOGICAL_OR) reduce 70) ((T_LOGICAL_XOR) reduce 70) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 70) ((T_OBJECT_OPERATOR) reduce 70) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 70) ((T_PIPE) reduce 70) ((T_POW) reduce 70) ((T_SL) reduce 70) ((T_SPACESHIP) reduce 70) ((T_SR) reduce 70) (("[") reduce 70) (("^") reduce 70) (("|") reduce 70)) 31025 ((("%") reduce 3) (("(") reduce 3) (("*") reduce 3) (("+") reduce 3) (("-") reduce 3) ((".") reduce 3) (("/") reduce 3) ((";") reduce 3) (("<") reduce 3) ((">") reduce 3) (("?") reduce 3) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 3) ((T_BOOLEAN_AND) reduce 3) ((T_BOOLEAN_OR) reduce 3) ((T_COALESCE) reduce 3) ((T_DOUBLE_ARROW) reduce 3) ((T_INSTANCEOF) reduce 3) ((T_IS_EQUAL) reduce 3) ((T_IS_GREATER_OR_EQUAL) reduce 3) ((T_IS_IDENTICAL) reduce 3) ((T_IS_NOT_EQUAL) reduce 3) ((T_IS_NOT_IDENTICAL) reduce 3) ((T_IS_SMALLER_OR_EQUAL) reduce 3) ((T_LOGICAL_AND) reduce 3) ((T_LOGICAL_OR) reduce 3) ((T_LOGICAL_XOR) reduce 3) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 3) ((T_OBJECT_OPERATOR) reduce 3) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 3) ((T_PIPE) reduce 3) ((T_POW) reduce 3) ((T_SL) reduce 3) ((T_SPACESHIP) reduce 3) ((T_SR) reduce 3) (("[") reduce 3) (("^") reduce 3) (("|") reduce 3)) 31026 ((("%") reduce 12) (("(") reduce 12) (("*") reduce 12) (("+") reduce 12) (("-") reduce 12) ((".") reduce 12) (("/") reduce 12) ((";") reduce 12) (("<") reduce 12) ((">") reduce 12) (("?") reduce 12) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 12) ((T_BOOLEAN_AND) reduce 12) ((T_BOOLEAN_OR) reduce 12) ((T_COALESCE) reduce 12) ((T_DOUBLE_ARROW) reduce 12) ((T_INSTANCEOF) reduce 12) ((T_IS_EQUAL) reduce 12) ((T_IS_GREATER_OR_EQUAL) reduce 12) ((T_IS_IDENTICAL) reduce 12) ((T_IS_NOT_EQUAL) reduce 12) ((T_IS_NOT_IDENTICAL) reduce 12) ((T_IS_SMALLER_OR_EQUAL) reduce 12) ((T_LOGICAL_AND) reduce 12) ((T_LOGICAL_OR) reduce 12) ((T_LOGICAL_XOR) reduce 12) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 12) ((T_OBJECT_OPERATOR) reduce 12) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 12) ((T_PIPE) reduce 12) ((T_POW) reduce 12) ((T_SL) reduce 12) ((T_SPACESHIP) reduce 12) ((T_SR) reduce 12) (("[") reduce 12) (("^") reduce 12) (("|") reduce 12)) 31027 ((("%") reduce 54) (("(") reduce 54) (("*") reduce 54) (("+") reduce 54) (("-") reduce 54) ((".") reduce 54) (("/") reduce 54) ((";") reduce 54) (("<") reduce 54) ((">") reduce 54) (("?") reduce 54) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 54) ((T_BOOLEAN_AND) reduce 54) ((T_BOOLEAN_OR) reduce 54) ((T_COALESCE) reduce 54) ((T_DOUBLE_ARROW) reduce 54) ((T_INSTANCEOF) reduce 54) ((T_IS_EQUAL) reduce 54) ((T_IS_GREATER_OR_EQUAL) reduce 54) ((T_IS_IDENTICAL) reduce 54) ((T_IS_NOT_EQUAL) reduce 54) ((T_IS_NOT_IDENTICAL) reduce 54) ((T_IS_SMALLER_OR_EQUAL) reduce 54) ((T_LOGICAL_AND) reduce 54) ((T_LOGICAL_OR) reduce 54) ((T_LOGICAL_XOR) reduce 54) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 54) ((T_OBJECT_OPERATOR) reduce 54) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 54) ((T_PIPE) reduce 54) ((T_POW) reduce 54) ((T_SL) reduce 54) ((T_SPACESHIP) reduce 54) ((T_SR) reduce 54) (("[") reduce 54) (("^") reduce 54) (("|") reduce 54)) 31028 ((("%") reduce 65) (("(") reduce 65) (("*") reduce 65) (("+") reduce 65) (("-") reduce 65) ((".") reduce 65) (("/") reduce 65) ((";") reduce 65) (("<") reduce 65) ((">") reduce 65) (("?") reduce 65) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 65) ((T_BOOLEAN_AND) reduce 65) ((T_BOOLEAN_OR) reduce 65) ((T_COALESCE) reduce 65) ((T_DOUBLE_ARROW) reduce 65) ((T_INSTANCEOF) reduce 65) ((T_IS_EQUAL) reduce 65) ((T_IS_GREATER_OR_EQUAL) reduce 65) ((T_IS_IDENTICAL) reduce 65) ((T_IS_NOT_EQUAL) reduce 65) ((T_IS_NOT_IDENTICAL) reduce 65) ((T_IS_SMALLER_OR_EQUAL) reduce 65) ((T_LOGICAL_AND) reduce 65) ((T_LOGICAL_OR) reduce 65) ((T_LOGICAL_XOR) reduce 65) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 65) ((T_OBJECT_OPERATOR) reduce 65) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 65) ((T_PIPE) reduce 65) ((T_POW) reduce 65) ((T_SL) reduce 65) ((T_SPACESHIP) reduce 65) ((T_SR) reduce 65) (("[") reduce 65) (("^") reduce 65) (("|") reduce 65)) 31029 ((("%") reduce 75) (("(") reduce 75) (("*") reduce 75) (("+") reduce 75) (("-") reduce 75) ((".") reduce 75) (("/") reduce 75) ((";") reduce 75) (("<") reduce 75) ((">") reduce 75) (("?") reduce 75) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 75) ((T_BOOLEAN_AND) reduce 75) ((T_BOOLEAN_OR) reduce 75) ((T_COALESCE) reduce 75) ((T_DOUBLE_ARROW) reduce 75) ((T_INSTANCEOF) reduce 75) ((T_IS_EQUAL) reduce 75) ((T_IS_GREATER_OR_EQUAL) reduce 75) ((T_IS_IDENTICAL) reduce 75) ((T_IS_NOT_EQUAL) reduce 75) ((T_IS_NOT_IDENTICAL) reduce 75) ((T_IS_SMALLER_OR_EQUAL) reduce 75) ((T_LOGICAL_AND) reduce 75) ((T_LOGICAL_OR) reduce 75) ((T_LOGICAL_XOR) reduce 75) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 75) ((T_OBJECT_OPERATOR) reduce 75) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 75) ((T_PIPE) reduce 75) ((T_POW) reduce 75) ((T_SL) reduce 75) ((T_SPACESHIP) reduce 75) ((T_SR) reduce 75) (("[") reduce 75) (("^") reduce 75) (("|") reduce 75)) 31030 ((("%") reduce 30) (("(") reduce 30) (("*") reduce 30) (("+") reduce 30) (("-") reduce 30) ((".") reduce 30) (("/") reduce 30) ((";") reduce 30) (("<") reduce 30) ((">") reduce 30) (("?") reduce 30) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 30) ((T_BOOLEAN_AND) reduce 30) ((T_BOOLEAN_OR) reduce 30) ((T_COALESCE) reduce 30) ((T_DOUBLE_ARROW) reduce 30) ((T_INSTANCEOF) reduce 30) ((T_IS_EQUAL) reduce 30) ((T_IS_GREATER_OR_EQUAL) reduce 30) ((T_IS_IDENTICAL) reduce 30) ((T_IS_NOT_EQUAL) reduce 30) ((T_IS_NOT_IDENTICAL) reduce 30) ((T_IS_SMALLER_OR_EQUAL) reduce 30) ((T_LOGICAL_AND) reduce 30) ((T_LOGICAL_OR) reduce 30) ((T_LOGICAL_XOR) reduce 30) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 30) ((T_OBJECT_OPERATOR) reduce 30) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 30) ((T_PIPE) reduce 30) ((T_POW) reduce 30) ((T_SL) reduce 30) ((T_SPACESHIP) reduce 30) ((T_SR) reduce 30) (("[") reduce 30) (("^") reduce 30) (("|") reduce 30)) 31031 ((("%") reduce 68) (("(") reduce 68) (("*") reduce 68) (("+") reduce 68) (("-") reduce 68) ((".") reduce 68) (("/") reduce 68) ((";") reduce 68) (("<") reduce 68) ((">") reduce 68) (("?") reduce 68) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 68) ((T_BOOLEAN_AND) reduce 68) ((T_BOOLEAN_OR) reduce 68) ((T_COALESCE) reduce 68) ((T_DOUBLE_ARROW) reduce 68) ((T_INSTANCEOF) reduce 68) ((T_IS_EQUAL) reduce 68) ((T_IS_GREATER_OR_EQUAL) reduce 68) ((T_IS_IDENTICAL) reduce 68) ((T_IS_NOT_EQUAL) reduce 68) ((T_IS_NOT_IDENTICAL) reduce 68) ((T_IS_SMALLER_OR_EQUAL) reduce 68) ((T_LOGICAL_AND) reduce 68) ((T_LOGICAL_OR) reduce 68) ((T_LOGICAL_XOR) reduce 68) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 68) ((T_OBJECT_OPERATOR) reduce 68) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 68) ((T_PIPE) reduce 68) ((T_POW) reduce 68) ((T_SL) reduce 68) ((T_SPACESHIP) reduce 68) ((T_SR) reduce 68) (("[") reduce 68) (("^") reduce 68) (("|") reduce 68)) 31032 ((("%") reduce 21) (("(") reduce 21) (("*") reduce 21) (("+") reduce 21) (("-") reduce 21) ((".") reduce 21) (("/") reduce 21) ((";") reduce 21) (("<") reduce 21) ((">") reduce 21) (("?") reduce 21) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 21) ((T_BOOLEAN_AND) reduce 21) ((T_BOOLEAN_OR) reduce 21) ((T_COALESCE) reduce 21) ((T_DOUBLE_ARROW) reduce 21) ((T_INSTANCEOF) reduce 21) ((T_IS_EQUAL) reduce 21) ((T_IS_GREATER_OR_EQUAL) reduce 21) ((T_IS_IDENTICAL) reduce 21) ((T_IS_NOT_EQUAL) reduce 21) ((T_IS_NOT_IDENTICAL) reduce 21) ((T_IS_SMALLER_OR_EQUAL) reduce 21) ((T_LOGICAL_AND) reduce 21) ((T_LOGICAL_OR) reduce 21) ((T_LOGICAL_XOR) reduce 21) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 21) ((T_OBJECT_OPERATOR) reduce 21) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 21) ((T_PIPE) reduce 21) ((T_POW) reduce 21) ((T_SL) reduce 21) ((T_SPACESHIP) reduce 21) ((T_SR) reduce 21) (("[") reduce 21) (("^") reduce 21) (("|") reduce 21)) 31033 ((("%") reduce 23) (("(") reduce 23) (("*") reduce 23) (("+") reduce 23) (("-") reduce 23) ((".") reduce 23) (("/") reduce 23) ((";") reduce 23) (("<") reduce 23) ((">") reduce 23) (("?") reduce 23) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 23) ((T_BOOLEAN_AND) reduce 23) ((T_BOOLEAN_OR) reduce 23) ((T_COALESCE) reduce 23) ((T_DOUBLE_ARROW) reduce 23) ((T_INSTANCEOF) reduce 23) ((T_IS_EQUAL) reduce 23) ((T_IS_GREATER_OR_EQUAL) reduce 23) ((T_IS_IDENTICAL) reduce 23) ((T_IS_NOT_EQUAL) reduce 23) ((T_IS_NOT_IDENTICAL) reduce 23) ((T_IS_SMALLER_OR_EQUAL) reduce 23) ((T_LOGICAL_AND) reduce 23) ((T_LOGICAL_OR) reduce 23) ((T_LOGICAL_XOR) reduce 23) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 23) ((T_OBJECT_OPERATOR) reduce 23) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 23) ((T_PIPE) reduce 23) ((T_POW) reduce 23) ((T_SL) reduce 23) ((T_SPACESHIP) reduce 23) ((T_SR) reduce 23) (("[") reduce 23) (("^") reduce 23) (("|") reduce 23)) 31034 ((("%") reduce 41) (("(") reduce 41) (("*") reduce 41) (("+") reduce 41) (("-") reduce 41) ((".") reduce 41) (("/") reduce 41) ((";") reduce 41) (("<") reduce 41) ((">") reduce 41) (("?") reduce 41) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 41) ((T_BOOLEAN_AND) reduce 41) ((T_BOOLEAN_OR) reduce 41) ((T_COALESCE) reduce 41) ((T_DOUBLE_ARROW) reduce 41) ((T_INSTANCEOF) reduce 41) ((T_IS_EQUAL) reduce 41) ((T_IS_GREATER_OR_EQUAL) reduce 41) ((T_IS_IDENTICAL) reduce 41) ((T_IS_NOT_EQUAL) reduce 41) ((T_IS_NOT_IDENTICAL) reduce 41) ((T_IS_SMALLER_OR_EQUAL) reduce 41) ((T_LOGICAL_AND) reduce 41) ((T_LOGICAL_OR) reduce 41) ((T_LOGICAL_XOR) reduce 41) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 41) ((T_OBJECT_OPERATOR) reduce 41) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 41) ((T_PIPE) reduce 41) ((T_POW) reduce 41) ((T_SL) reduce 41) ((T_SPACESHIP) reduce 41) ((T_SR) reduce 41) (("[") reduce 41) (("^") reduce 41) (("|") reduce 41)) 31035 ((("%") reduce 62) (("(") reduce 62) (("*") reduce 62) (("+") reduce 62) (("-") reduce 62) ((".") reduce 62) (("/") reduce 62) ((";") reduce 62) (("<") reduce 62) ((">") reduce 62) (("?") reduce 62) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 62) ((T_BOOLEAN_AND) reduce 62) ((T_BOOLEAN_OR) reduce 62) ((T_COALESCE) reduce 62) ((T_DOUBLE_ARROW) reduce 62) ((T_INSTANCEOF) reduce 62) ((T_IS_EQUAL) reduce 62) ((T_IS_GREATER_OR_EQUAL) reduce 62) ((T_IS_IDENTICAL) reduce 62) ((T_IS_NOT_EQUAL) reduce 62) ((T_IS_NOT_IDENTICAL) reduce 62) ((T_IS_SMALLER_OR_EQUAL) reduce 62) ((T_LOGICAL_AND) reduce 62) ((T_LOGICAL_OR) reduce 62) ((T_LOGICAL_XOR) reduce 62) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 62) ((T_OBJECT_OPERATOR) reduce 62) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 62) ((T_PIPE) reduce 62) ((T_POW) reduce 62) ((T_SL) reduce 62) ((T_SPACESHIP) reduce 62) ((T_SR) reduce 62) (("[") reduce 62) (("^") reduce 62) (("|") reduce 62)) 31036 ((("%") reduce 34) (("(") reduce 34) (("*") reduce 34) (("+") reduce 34) (("-") reduce 34) ((".") reduce 34) (("/") reduce 34) ((";") reduce 34) (("<") reduce 34) ((">") reduce 34) (("?") reduce 34) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 34) ((T_BOOLEAN_AND) reduce 34) ((T_BOOLEAN_OR) reduce 34) ((T_COALESCE) reduce 34) ((T_DOUBLE_ARROW) reduce 34) ((T_INSTANCEOF) reduce 34) ((T_IS_EQUAL) reduce 34) ((T_IS_GREATER_OR_EQUAL) reduce 34) ((T_IS_IDENTICAL) reduce 34) ((T_IS_NOT_EQUAL) reduce 34) ((T_IS_NOT_IDENTICAL) reduce 34) ((T_IS_SMALLER_OR_EQUAL) reduce 34) ((T_LOGICAL_AND) reduce 34) ((T_LOGICAL_OR) reduce 34) ((T_LOGICAL_XOR) reduce 34) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 34) ((T_OBJECT_OPERATOR) reduce 34) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 34) ((T_PIPE) reduce 34) ((T_POW) reduce 34) ((T_SL) reduce 34) ((T_SPACESHIP) reduce 34) ((T_SR) reduce 34) (("[") reduce 34) (("^") reduce 34) (("|") reduce 34)) 31037 ((("%") reduce 40) (("(") reduce 40) (("*") reduce 40) (("+") reduce 40) (("-") reduce 40) ((".") reduce 40) (("/") reduce 40) ((";") reduce 40) (("<") reduce 40) ((">") reduce 40) (("?") reduce 40) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 40) ((T_BOOLEAN_AND) reduce 40) ((T_BOOLEAN_OR) reduce 40) ((T_COALESCE) reduce 40) ((T_DOUBLE_ARROW) reduce 40) ((T_INSTANCEOF) reduce 40) ((T_IS_EQUAL) reduce 40) ((T_IS_GREATER_OR_EQUAL) reduce 40) ((T_IS_IDENTICAL) reduce 40) ((T_IS_NOT_EQUAL) reduce 40) ((T_IS_NOT_IDENTICAL) reduce 40) ((T_IS_SMALLER_OR_EQUAL) reduce 40) ((T_LOGICAL_AND) reduce 40) ((T_LOGICAL_OR) reduce 40) ((T_LOGICAL_XOR) reduce 40) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 40) ((T_OBJECT_OPERATOR) reduce 40) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 40) ((T_PIPE) reduce 40) ((T_POW) reduce 40) ((T_SL) reduce 40) ((T_SPACESHIP) reduce 40) ((T_SR) reduce 40) (("[") reduce 40) (("^") reduce 40) (("|") reduce 40)) 31038 ((("%") reduce 13) (("(") reduce 13) (("*") reduce 13) (("+") reduce 13) (("-") reduce 13) ((".") reduce 13) (("/") reduce 13) ((";") reduce 13) (("<") reduce 13) ((">") reduce 13) (("?") reduce 13) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 13) ((T_BOOLEAN_AND) reduce 13) ((T_BOOLEAN_OR) reduce 13) ((T_COALESCE) reduce 13) ((T_DOUBLE_ARROW) reduce 13) ((T_INSTANCEOF) reduce 13) ((T_IS_EQUAL) reduce 13) ((T_IS_GREATER_OR_EQUAL) reduce 13) ((T_IS_IDENTICAL) reduce 13) ((T_IS_NOT_EQUAL) reduce 13) ((T_IS_NOT_IDENTICAL) reduce 13) ((T_IS_SMALLER_OR_EQUAL) reduce 13) ((T_LOGICAL_AND) reduce 13) ((T_LOGICAL_OR) reduce 13) ((T_LOGICAL_XOR) reduce 13) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 13) ((T_OBJECT_OPERATOR) reduce 13) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 13) ((T_PIPE) reduce 13) ((T_POW) reduce 13) ((T_SL) reduce 13) ((T_SPACESHIP) reduce 13) ((T_SR) reduce 13) (("[") reduce 13) (("^") reduce 13) (("|") reduce 13)) 31039 ((("%") reduce 55) (("(") reduce 55) (("*") reduce 55) (("+") reduce 55) (("-") reduce 55) ((".") reduce 55) (("/") reduce 55) ((";") reduce 55) (("<") reduce 55) ((">") reduce 55) (("?") reduce 55) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 55) ((T_BOOLEAN_AND) reduce 55) ((T_BOOLEAN_OR) reduce 55) ((T_COALESCE) reduce 55) ((T_DOUBLE_ARROW) reduce 55) ((T_INSTANCEOF) reduce 55) ((T_IS_EQUAL) reduce 55) ((T_IS_GREATER_OR_EQUAL) reduce 55) ((T_IS_IDENTICAL) reduce 55) ((T_IS_NOT_EQUAL) reduce 55) ((T_IS_NOT_IDENTICAL) reduce 55) ((T_IS_SMALLER_OR_EQUAL) reduce 55) ((T_LOGICAL_AND) reduce 55) ((T_LOGICAL_OR) reduce 55) ((T_LOGICAL_XOR) reduce 55) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 55) ((T_OBJECT_OPERATOR) reduce 55) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 55) ((T_PIPE) reduce 55) ((T_POW) reduce 55) ((T_SL) reduce 55) ((T_SPACESHIP) reduce 55) ((T_SR) reduce 55) (("[") reduce 55) (("^") reduce 55) (("|") reduce 55)) 31040 ((("%") reduce 1) (("(") reduce 1) (("*") reduce 1) (("+") reduce 1) (("-") reduce 1) ((".") reduce 1) (("/") reduce 1) ((";") reduce 1) (("<") reduce 1) ((">") reduce 1) (("?") reduce 1) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 1) ((T_BOOLEAN_AND) reduce 1) ((T_BOOLEAN_OR) reduce 1) ((T_COALESCE) reduce 1) ((T_DOUBLE_ARROW) reduce 1) ((T_INSTANCEOF) reduce 1) ((T_IS_EQUAL) reduce 1) ((T_IS_GREATER_OR_EQUAL) reduce 1) ((T_IS_IDENTICAL) reduce 1) ((T_IS_NOT_EQUAL) reduce 1) ((T_IS_NOT_IDENTICAL) reduce 1) ((T_IS_SMALLER_OR_EQUAL) reduce 1) ((T_LOGICAL_AND) reduce 1) ((T_LOGICAL_OR) reduce 1) ((T_LOGICAL_XOR) reduce 1) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 1) ((T_OBJECT_OPERATOR) reduce 1) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 1) ((T_PIPE) reduce 1) ((T_POW) reduce 1) ((T_SL) reduce 1) ((T_SPACESHIP) reduce 1) ((T_SR) reduce 1) (("[") reduce 1) (("^") reduce 1) (("|") reduce 1)) 31041 ((("%") reduce 2) (("(") reduce 2) (("*") reduce 2) (("+") reduce 2) (("-") reduce 2) ((".") reduce 2) (("/") reduce 2) ((";") reduce 2) (("<") reduce 2) ((">") reduce 2) (("?") reduce 2) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 2) ((T_BOOLEAN_AND) reduce 2) ((T_BOOLEAN_OR) reduce 2) ((T_COALESCE) reduce 2) ((T_DOUBLE_ARROW) reduce 2) ((T_INSTANCEOF) reduce 2) ((T_IS_EQUAL) reduce 2) ((T_IS_GREATER_OR_EQUAL) reduce 2) ((T_IS_IDENTICAL) reduce 2) ((T_IS_NOT_EQUAL) reduce 2) ((T_IS_NOT_IDENTICAL) reduce 2) ((T_IS_SMALLER_OR_EQUAL) reduce 2) ((T_LOGICAL_AND) reduce 2) ((T_LOGICAL_OR) reduce 2) ((T_LOGICAL_XOR) reduce 2) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 2) ((T_OBJECT_OPERATOR) reduce 2) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 2) ((T_PIPE) reduce 2) ((T_POW) reduce 2) ((T_SL) reduce 2) ((T_SPACESHIP) reduce 2) ((T_SR) reduce 2) (("[") reduce 2) (("^") reduce 2) (("|") reduce 2)) 31042 ((("%") reduce 9) (("(") reduce 9) (("*") reduce 9) (("+") reduce 9) (("-") reduce 9) ((".") reduce 9) (("/") reduce 9) ((";") reduce 9) (("<") reduce 9) ((">") reduce 9) (("?") reduce 9) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 9) ((T_BOOLEAN_AND) reduce 9) ((T_BOOLEAN_OR) reduce 9) ((T_COALESCE) reduce 9) ((T_DOUBLE_ARROW) reduce 9) ((T_INSTANCEOF) reduce 9) ((T_IS_EQUAL) reduce 9) ((T_IS_GREATER_OR_EQUAL) reduce 9) ((T_IS_IDENTICAL) reduce 9) ((T_IS_NOT_EQUAL) reduce 9) ((T_IS_NOT_IDENTICAL) reduce 9) ((T_IS_SMALLER_OR_EQUAL) reduce 9) ((T_LOGICAL_AND) reduce 9) ((T_LOGICAL_OR) reduce 9) ((T_LOGICAL_XOR) reduce 9) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 9) ((T_OBJECT_OPERATOR) reduce 9) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 9) ((T_PIPE) reduce 9) ((T_POW) reduce 9) ((T_SL) reduce 9) ((T_SPACESHIP) reduce 9) ((T_SR) reduce 9) (("[") reduce 9) (("^") reduce 9) (("|") reduce 9)) 31043 ((("%") reduce 33) (("(") reduce 33) (("*") reduce 33) (("+") reduce 33) (("-") reduce 33) ((".") reduce 33) (("/") reduce 33) ((";") reduce 33) (("<") reduce 33) ((">") reduce 33) (("?") reduce 33) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 33) ((T_BOOLEAN_AND) reduce 33) ((T_BOOLEAN_OR) reduce 33) ((T_COALESCE) reduce 33) ((T_DOUBLE_ARROW) reduce 33) ((T_INSTANCEOF) reduce 33) ((T_IS_EQUAL) reduce 33) ((T_IS_GREATER_OR_EQUAL) reduce 33) ((T_IS_IDENTICAL) reduce 33) ((T_IS_NOT_EQUAL) reduce 33) ((T_IS_NOT_IDENTICAL) reduce 33) ((T_IS_SMALLER_OR_EQUAL) reduce 33) ((T_LOGICAL_AND) reduce 33) ((T_LOGICAL_OR) reduce 33) ((T_LOGICAL_XOR) reduce 33) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 33) ((T_OBJECT_OPERATOR) reduce 33) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 33) ((T_PIPE) reduce 33) ((T_POW) reduce 33) ((T_SL) reduce 33) ((T_SPACESHIP) reduce 33) ((T_SR) reduce 33) (("[") reduce 33) (("^") reduce 33) (("|") reduce 33)) 31044 ((("%") reduce 58) (("(") reduce 58) (("*") reduce 58) (("+") reduce 58) (("-") reduce 58) ((".") reduce 58) (("/") reduce 58) ((";") reduce 58) (("<") reduce 58) ((">") reduce 58) (("?") reduce 58) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 58) ((T_BOOLEAN_AND) reduce 58) ((T_BOOLEAN_OR) reduce 58) ((T_COALESCE) reduce 58) ((T_DOUBLE_ARROW) reduce 58) ((T_INSTANCEOF) reduce 58) ((T_IS_EQUAL) reduce 58) ((T_IS_GREATER_OR_EQUAL) reduce 58) ((T_IS_IDENTICAL) reduce 58) ((T_IS_NOT_EQUAL) reduce 58) ((T_IS_NOT_IDENTICAL) reduce 58) ((T_IS_SMALLER_OR_EQUAL) reduce 58) ((T_LOGICAL_AND) reduce 58) ((T_LOGICAL_OR) reduce 58) ((T_LOGICAL_XOR) reduce 58) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 58) ((T_OBJECT_OPERATOR) reduce 58) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 58) ((T_PIPE) reduce 58) ((T_POW) reduce 58) ((T_SL) reduce 58) ((T_SPACESHIP) reduce 58) ((T_SR) reduce 58) (("[") reduce 58) (("^") reduce 58) (("|") reduce 58)) 31045 ((("%") reduce 37) (("(") reduce 37) (("*") reduce 37) (("+") reduce 37) (("-") reduce 37) ((".") reduce 37) (("/") reduce 37) ((";") reduce 37) (("<") reduce 37) ((">") reduce 37) (("?") reduce 37) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 37) ((T_BOOLEAN_AND) reduce 37) ((T_BOOLEAN_OR) reduce 37) ((T_COALESCE) reduce 37) ((T_DOUBLE_ARROW) reduce 37) ((T_INSTANCEOF) reduce 37) ((T_IS_EQUAL) reduce 37) ((T_IS_GREATER_OR_EQUAL) reduce 37) ((T_IS_IDENTICAL) reduce 37) ((T_IS_NOT_EQUAL) reduce 37) ((T_IS_NOT_IDENTICAL) reduce 37) ((T_IS_SMALLER_OR_EQUAL) reduce 37) ((T_LOGICAL_AND) reduce 37) ((T_LOGICAL_OR) reduce 37) ((T_LOGICAL_XOR) reduce 37) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 37) ((T_OBJECT_OPERATOR) reduce 37) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 37) ((T_PIPE) reduce 37) ((T_POW) reduce 37) ((T_SL) reduce 37) ((T_SPACESHIP) reduce 37) ((T_SR) reduce 37) (("[") reduce 37) (("^") reduce 37) (("|") reduce 37)) 31046 ((("%") reduce 64) (("(") reduce 64) (("*") reduce 64) (("+") reduce 64) (("-") reduce 64) ((".") reduce 64) (("/") reduce 64) ((";") reduce 64) (("<") reduce 64) ((">") reduce 64) (("?") reduce 64) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 64) ((T_BOOLEAN_AND) reduce 64) ((T_BOOLEAN_OR) reduce 64) ((T_COALESCE) reduce 64) ((T_DOUBLE_ARROW) reduce 64) ((T_INSTANCEOF) reduce 64) ((T_IS_EQUAL) reduce 64) ((T_IS_GREATER_OR_EQUAL) reduce 64) ((T_IS_IDENTICAL) reduce 64) ((T_IS_NOT_EQUAL) reduce 64) ((T_IS_NOT_IDENTICAL) reduce 64) ((T_IS_SMALLER_OR_EQUAL) reduce 64) ((T_LOGICAL_AND) reduce 64) ((T_LOGICAL_OR) reduce 64) ((T_LOGICAL_XOR) reduce 64) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 64) ((T_OBJECT_OPERATOR) reduce 64) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 64) ((T_PIPE) reduce 64) ((T_POW) reduce 64) ((T_SL) reduce 64) ((T_SPACESHIP) reduce 64) ((T_SR) reduce 64) (("[") reduce 64) (("^") reduce 64) (("|") reduce 64)) 31047 ((("%") reduce 46) (("(") reduce 46) (("*") reduce 46) (("+") reduce 46) (("-") reduce 46) ((".") reduce 46) (("/") reduce 46) ((";") reduce 46) (("<") reduce 46) ((">") reduce 46) (("?") reduce 46) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 46) ((T_BOOLEAN_AND) reduce 46) ((T_BOOLEAN_OR) reduce 46) ((T_COALESCE) reduce 46) ((T_DOUBLE_ARROW) reduce 46) ((T_INSTANCEOF) reduce 46) ((T_IS_EQUAL) reduce 46) ((T_IS_GREATER_OR_EQUAL) reduce 46) ((T_IS_IDENTICAL) reduce 46) ((T_IS_NOT_EQUAL) reduce 46) ((T_IS_NOT_IDENTICAL) reduce 46) ((T_IS_SMALLER_OR_EQUAL) reduce 46) ((T_LOGICAL_AND) reduce 46) ((T_LOGICAL_OR) reduce 46) ((T_LOGICAL_XOR) reduce 46) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 46) ((T_OBJECT_OPERATOR) reduce 46) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 46) ((T_PIPE) reduce 46) ((T_POW) reduce 46) ((T_SL) reduce 46) ((T_SPACESHIP) reduce 46) ((T_SR) reduce 46) (("[") reduce 46) (("^") reduce 46) (("|") reduce 46)) 31048 ((("%") reduce 8) (("(") reduce 8) (("*") reduce 8) (("+") reduce 8) (("-") reduce 8) ((".") reduce 8) (("/") reduce 8) ((";") reduce 8) (("<") reduce 8) ((">") reduce 8) (("?") reduce 8) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 8) ((T_BOOLEAN_AND) reduce 8) ((T_BOOLEAN_OR) reduce 8) ((T_COALESCE) reduce 8) ((T_DOUBLE_ARROW) reduce 8) ((T_INSTANCEOF) reduce 8) ((T_IS_EQUAL) reduce 8) ((T_IS_GREATER_OR_EQUAL) reduce 8) ((T_IS_IDENTICAL) reduce 8) ((T_IS_NOT_EQUAL) reduce 8) ((T_IS_NOT_IDENTICAL) reduce 8) ((T_IS_SMALLER_OR_EQUAL) reduce 8) ((T_LOGICAL_AND) reduce 8) ((T_LOGICAL_OR) reduce 8) ((T_LOGICAL_XOR) reduce 8) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 8) ((T_OBJECT_OPERATOR) reduce 8) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 8) ((T_PIPE) reduce 8) ((T_POW) reduce 8) ((T_SL) reduce 8) ((T_SPACESHIP) reduce 8) ((T_SR) reduce 8) (("[") reduce 8) (("^") reduce 8) (("|") reduce 8)) 31049 ((("%") reduce 6) (("(") reduce 6) (("*") reduce 6) (("+") reduce 6) (("-") reduce 6) ((".") reduce 6) (("/") reduce 6) ((";") reduce 6) (("<") reduce 6) ((">") reduce 6) (("?") reduce 6) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 6) ((T_BOOLEAN_AND) reduce 6) ((T_BOOLEAN_OR) reduce 6) ((T_COALESCE) reduce 6) ((T_DOUBLE_ARROW) reduce 6) ((T_INSTANCEOF) reduce 6) ((T_IS_EQUAL) reduce 6) ((T_IS_GREATER_OR_EQUAL) reduce 6) ((T_IS_IDENTICAL) reduce 6) ((T_IS_NOT_EQUAL) reduce 6) ((T_IS_NOT_IDENTICAL) reduce 6) ((T_IS_SMALLER_OR_EQUAL) reduce 6) ((T_LOGICAL_AND) reduce 6) ((T_LOGICAL_OR) reduce 6) ((T_LOGICAL_XOR) reduce 6) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 6) ((T_OBJECT_OPERATOR) reduce 6) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 6) ((T_PIPE) reduce 6) ((T_POW) reduce 6) ((T_SL) reduce 6) ((T_SPACESHIP) reduce 6) ((T_SR) reduce 6) (("[") reduce 6) (("^") reduce 6) (("|") reduce 6)) 31050 ((("%") reduce 7) (("(") reduce 7) (("*") reduce 7) (("+") reduce 7) (("-") reduce 7) ((".") reduce 7) (("/") reduce 7) ((";") reduce 7) (("<") reduce 7) ((">") reduce 7) (("?") reduce 7) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 7) ((T_BOOLEAN_AND) reduce 7) ((T_BOOLEAN_OR) reduce 7) ((T_COALESCE) reduce 7) ((T_DOUBLE_ARROW) reduce 7) ((T_INSTANCEOF) reduce 7) ((T_IS_EQUAL) reduce 7) ((T_IS_GREATER_OR_EQUAL) reduce 7) ((T_IS_IDENTICAL) reduce 7) ((T_IS_NOT_EQUAL) reduce 7) ((T_IS_NOT_IDENTICAL) reduce 7) ((T_IS_SMALLER_OR_EQUAL) reduce 7) ((T_LOGICAL_AND) reduce 7) ((T_LOGICAL_OR) reduce 7) ((T_LOGICAL_XOR) reduce 7) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 7) ((T_OBJECT_OPERATOR) reduce 7) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 7) ((T_PIPE) reduce 7) ((T_POW) reduce 7) ((T_SL) reduce 7) ((T_SPACESHIP) reduce 7) ((T_SR) reduce 7) (("[") reduce 7) (("^") reduce 7) (("|") reduce 7)) 31051 ((("%") reduce 69) (("(") reduce 69) (("*") reduce 69) (("+") reduce 69) (("-") reduce 69) ((".") reduce 69) (("/") reduce 69) ((";") reduce 69) (("<") reduce 69) ((">") reduce 69) (("?") reduce 69) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 69) ((T_BOOLEAN_AND) reduce 69) ((T_BOOLEAN_OR) reduce 69) ((T_COALESCE) reduce 69) ((T_DOUBLE_ARROW) reduce 69) ((T_INSTANCEOF) reduce 69) ((T_IS_EQUAL) reduce 69) ((T_IS_GREATER_OR_EQUAL) reduce 69) ((T_IS_IDENTICAL) reduce 69) ((T_IS_NOT_EQUAL) reduce 69) ((T_IS_NOT_IDENTICAL) reduce 69) ((T_IS_SMALLER_OR_EQUAL) reduce 69) ((T_LOGICAL_AND) reduce 69) ((T_LOGICAL_OR) reduce 69) ((T_LOGICAL_XOR) reduce 69) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 69) ((T_OBJECT_OPERATOR) reduce 69) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 69) ((T_PIPE) reduce 69) ((T_POW) reduce 69) ((T_SL) reduce 69) ((T_SPACESHIP) reduce 69) ((T_SR) reduce 69) (("[") reduce 69) (("^") reduce 69) (("|") reduce 69)) 31052 ((("%") reduce 63) (("(") reduce 63) (("*") reduce 63) (("+") reduce 63) (("-") reduce 63) ((".") reduce 63) (("/") reduce 63) ((";") reduce 63) (("<") reduce 63) ((">") reduce 63) (("?") reduce 63) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 63) ((T_BOOLEAN_AND) reduce 63) ((T_BOOLEAN_OR) reduce 63) ((T_COALESCE) reduce 63) ((T_DOUBLE_ARROW) reduce 63) ((T_INSTANCEOF) reduce 63) ((T_IS_EQUAL) reduce 63) ((T_IS_GREATER_OR_EQUAL) reduce 63) ((T_IS_IDENTICAL) reduce 63) ((T_IS_NOT_EQUAL) reduce 63) ((T_IS_NOT_IDENTICAL) reduce 63) ((T_IS_SMALLER_OR_EQUAL) reduce 63) ((T_LOGICAL_AND) reduce 63) ((T_LOGICAL_OR) reduce 63) ((T_LOGICAL_XOR) reduce 63) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 63) ((T_OBJECT_OPERATOR) reduce 63) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 63) ((T_PIPE) reduce 63) ((T_POW) reduce 63) ((T_SL) reduce 63) ((T_SPACESHIP) reduce 63) ((T_SR) reduce 63) (("[") reduce 63) (("^") reduce 63) (("|") reduce 63)) 31053 ((("%") reduce 56) (("(") reduce 56) (("*") reduce 56) (("+") reduce 56) (("-") reduce 56) ((".") reduce 56) (("/") reduce 56) ((";") reduce 56) (("<") reduce 56) ((">") reduce 56) (("?") reduce 56) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 56) ((T_BOOLEAN_AND) reduce 56) ((T_BOOLEAN_OR) reduce 56) ((T_COALESCE) reduce 56) ((T_DOUBLE_ARROW) reduce 56) ((T_INSTANCEOF) reduce 56) ((T_IS_EQUAL) reduce 56) ((T_IS_GREATER_OR_EQUAL) reduce 56) ((T_IS_IDENTICAL) reduce 56) ((T_IS_NOT_EQUAL) reduce 56) ((T_IS_NOT_IDENTICAL) reduce 56) ((T_IS_SMALLER_OR_EQUAL) reduce 56) ((T_LOGICAL_AND) reduce 56) ((T_LOGICAL_OR) reduce 56) ((T_LOGICAL_XOR) reduce 56) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 56) ((T_OBJECT_OPERATOR) reduce 56) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 56) ((T_PIPE) reduce 56) ((T_POW) reduce 56) ((T_SL) reduce 56) ((T_SPACESHIP) reduce 56) ((T_SR) reduce 56) (("[") reduce 56) (("^") reduce 56) (("|") reduce 56)) 31054 ((("%") reduce 10) (("(") reduce 10) (("*") reduce 10) (("+") reduce 10) (("-") reduce 10) ((".") reduce 10) (("/") reduce 10) ((";") reduce 10) (("<") reduce 10) ((">") reduce 10) (("?") reduce 10) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 10) ((T_BOOLEAN_AND) reduce 10) ((T_BOOLEAN_OR) reduce 10) ((T_COALESCE) reduce 10) ((T_DOUBLE_ARROW) reduce 10) ((T_INSTANCEOF) reduce 10) ((T_IS_EQUAL) reduce 10) ((T_IS_GREATER_OR_EQUAL) reduce 10) ((T_IS_IDENTICAL) reduce 10) ((T_IS_NOT_EQUAL) reduce 10) ((T_IS_NOT_IDENTICAL) reduce 10) ((T_IS_SMALLER_OR_EQUAL) reduce 10) ((T_LOGICAL_AND) reduce 10) ((T_LOGICAL_OR) reduce 10) ((T_LOGICAL_XOR) reduce 10) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 10) ((T_OBJECT_OPERATOR) reduce 10) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 10) ((T_PIPE) reduce 10) ((T_POW) reduce 10) ((T_SL) reduce 10) ((T_SPACESHIP) reduce 10) ((T_SR) reduce 10) (("[") reduce 10) (("^") reduce 10) (("|") reduce 10)) 31055 ((("%") reduce 67) (("(") reduce 67) (("*") reduce 67) (("+") reduce 67) (("-") reduce 67) ((".") reduce 67) (("/") reduce 67) ((";") reduce 67) (("<") reduce 67) ((">") reduce 67) (("?") reduce 67) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 67) ((T_BOOLEAN_AND) reduce 67) ((T_BOOLEAN_OR) reduce 67) ((T_COALESCE) reduce 67) ((T_DOUBLE_ARROW) reduce 67) ((T_INSTANCEOF) reduce 67) ((T_IS_EQUAL) reduce 67) ((T_IS_GREATER_OR_EQUAL) reduce 67) ((T_IS_IDENTICAL) reduce 67) ((T_IS_NOT_EQUAL) reduce 67) ((T_IS_NOT_IDENTICAL) reduce 67) ((T_IS_SMALLER_OR_EQUAL) reduce 67) ((T_LOGICAL_AND) reduce 67) ((T_LOGICAL_OR) reduce 67) ((T_LOGICAL_XOR) reduce 67) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 67) ((T_OBJECT_OPERATOR) reduce 67) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 67) ((T_PIPE) reduce 67) ((T_POW) reduce 67) ((T_SL) reduce 67) ((T_SPACESHIP) reduce 67) ((T_SR) reduce 67) (("[") reduce 67) (("^") reduce 67) (("|") reduce 67)) 31056 ((("%") reduce 44) (("(") reduce 44) (("*") reduce 44) (("+") reduce 44) (("-") reduce 44) ((".") reduce 44) (("/") reduce 44) ((";") reduce 44) (("<") reduce 44) ((">") reduce 44) (("?") reduce 44) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 44) ((T_BOOLEAN_AND) reduce 44) ((T_BOOLEAN_OR) reduce 44) ((T_COALESCE) reduce 44) ((T_DOUBLE_ARROW) reduce 44) ((T_INSTANCEOF) reduce 44) ((T_IS_EQUAL) reduce 44) ((T_IS_GREATER_OR_EQUAL) reduce 44) ((T_IS_IDENTICAL) reduce 44) ((T_IS_NOT_EQUAL) reduce 44) ((T_IS_NOT_IDENTICAL) reduce 44) ((T_IS_SMALLER_OR_EQUAL) reduce 44) ((T_LOGICAL_AND) reduce 44) ((T_LOGICAL_OR) reduce 44) ((T_LOGICAL_XOR) reduce 44) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 44) ((T_OBJECT_OPERATOR) reduce 44) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 44) ((T_PIPE) reduce 44) ((T_POW) reduce 44) ((T_SL) reduce 44) ((T_SPACESHIP) reduce 44) ((T_SR) reduce 44) (("[") reduce 44) (("^") reduce 44) (("|") reduce 44)) 31057 ((("%") reduce 76) (("(") reduce 76) (("*") reduce 76) (("+") reduce 76) (("-") reduce 76) ((".") reduce 76) (("/") reduce 76) ((";") reduce 76) (("<") reduce 76) ((">") reduce 76) (("?") reduce 76) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 76) ((T_BOOLEAN_AND) reduce 76) ((T_BOOLEAN_OR) reduce 76) ((T_COALESCE) reduce 76) ((T_DOUBLE_ARROW) reduce 76) ((T_INSTANCEOF) reduce 76) ((T_IS_EQUAL) reduce 76) ((T_IS_GREATER_OR_EQUAL) reduce 76) ((T_IS_IDENTICAL) reduce 76) ((T_IS_NOT_EQUAL) reduce 76) ((T_IS_NOT_IDENTICAL) reduce 76) ((T_IS_SMALLER_OR_EQUAL) reduce 76) ((T_LOGICAL_AND) reduce 76) ((T_LOGICAL_OR) reduce 76) ((T_LOGICAL_XOR) reduce 76) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 76) ((T_OBJECT_OPERATOR) reduce 76) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 76) ((T_PIPE) reduce 76) ((T_POW) reduce 76) ((T_SL) reduce 76) ((T_SPACESHIP) reduce 76) ((T_SR) reduce 76) (("[") reduce 76) (("^") reduce 76) (("|") reduce 76)) 31058 ((("%") reduce 71) (("(") reduce 71) (("*") reduce 71) (("+") reduce 71) (("-") reduce 71) ((".") reduce 71) (("/") reduce 71) ((";") reduce 71) (("<") reduce 71) ((">") reduce 71) (("?") reduce 71) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 71) ((T_BOOLEAN_AND) reduce 71) ((T_BOOLEAN_OR) reduce 71) ((T_COALESCE) reduce 71) ((T_DOUBLE_ARROW) reduce 71) ((T_INSTANCEOF) reduce 71) ((T_IS_EQUAL) reduce 71) ((T_IS_GREATER_OR_EQUAL) reduce 71) ((T_IS_IDENTICAL) reduce 71) ((T_IS_NOT_EQUAL) reduce 71) ((T_IS_NOT_IDENTICAL) reduce 71) ((T_IS_SMALLER_OR_EQUAL) reduce 71) ((T_LOGICAL_AND) reduce 71) ((T_LOGICAL_OR) reduce 71) ((T_LOGICAL_XOR) reduce 71) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 71) ((T_OBJECT_OPERATOR) reduce 71) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 71) ((T_PIPE) reduce 71) ((T_POW) reduce 71) ((T_SL) reduce 71) ((T_SPACESHIP) reduce 71) ((T_SR) reduce 71) (("[") reduce 71) (("^") reduce 71) (("|") reduce 71)) 31059 ((("%") reduce 77) (("(") reduce 77) (("*") reduce 77) (("+") reduce 77) (("-") reduce 77) ((".") reduce 77) (("/") reduce 77) ((";") reduce 77) (("<") reduce 77) ((">") reduce 77) (("?") reduce 77) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 77) ((T_BOOLEAN_AND) reduce 77) ((T_BOOLEAN_OR) reduce 77) ((T_COALESCE) reduce 77) ((T_DOUBLE_ARROW) reduce 77) ((T_INSTANCEOF) reduce 77) ((T_IS_EQUAL) reduce 77) ((T_IS_GREATER_OR_EQUAL) reduce 77) ((T_IS_IDENTICAL) reduce 77) ((T_IS_NOT_EQUAL) reduce 77) ((T_IS_NOT_IDENTICAL) reduce 77) ((T_IS_SMALLER_OR_EQUAL) reduce 77) ((T_LOGICAL_AND) reduce 77) ((T_LOGICAL_OR) reduce 77) ((T_LOGICAL_XOR) reduce 77) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 77) ((T_OBJECT_OPERATOR) reduce 77) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 77) ((T_PIPE) reduce 77) ((T_POW) reduce 77) ((T_SL) reduce 77) ((T_SPACESHIP) reduce 77) ((T_SR) reduce 77) (("[") reduce 77) (("^") reduce 77) (("|") reduce 77)) 31060 ((("%") reduce 78) (("(") reduce 78) (("*") reduce 78) (("+") reduce 78) (("-") reduce 78) ((".") reduce 78) (("/") reduce 78) ((";") reduce 78) (("<") reduce 78) ((">") reduce 78) (("?") reduce 78) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 78) ((T_BOOLEAN_AND) reduce 78) ((T_BOOLEAN_OR) reduce 78) ((T_COALESCE) reduce 78) ((T_DOUBLE_ARROW) reduce 78) ((T_INSTANCEOF) reduce 78) ((T_IS_EQUAL) reduce 78) ((T_IS_GREATER_OR_EQUAL) reduce 78) ((T_IS_IDENTICAL) reduce 78) ((T_IS_NOT_EQUAL) reduce 78) ((T_IS_NOT_IDENTICAL) reduce 78) ((T_IS_SMALLER_OR_EQUAL) reduce 78) ((T_LOGICAL_AND) reduce 78) ((T_LOGICAL_OR) reduce 78) ((T_LOGICAL_XOR) reduce 78) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 78) ((T_OBJECT_OPERATOR) reduce 78) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 78) ((T_PIPE) reduce 78) ((T_POW) reduce 78) ((T_SL) reduce 78) ((T_SPACESHIP) reduce 78) ((T_SR) reduce 78) (("[") reduce 78) (("^") reduce 78) (("|") reduce 78)) 31061 ((("%") reduce 79) (("(") reduce 79) (("*") reduce 79) (("+") reduce 79) (("-") reduce 79) ((".") reduce 79) (("/") reduce 79) ((";") reduce 79) (("<") reduce 79) ((">") reduce 79) (("?") reduce 79) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 79) ((T_BOOLEAN_AND) reduce 79) ((T_BOOLEAN_OR) reduce 79) ((T_COALESCE) reduce 79) ((T_DOUBLE_ARROW) reduce 79) ((T_INSTANCEOF) reduce 79) ((T_IS_EQUAL) reduce 79) ((T_IS_GREATER_OR_EQUAL) reduce 79) ((T_IS_IDENTICAL) reduce 79) ((T_IS_NOT_EQUAL) reduce 79) ((T_IS_NOT_IDENTICAL) reduce 79) ((T_IS_SMALLER_OR_EQUAL) reduce 79) ((T_LOGICAL_AND) reduce 79) ((T_LOGICAL_OR) reduce 79) ((T_LOGICAL_XOR) reduce 79) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 79) ((T_OBJECT_OPERATOR) reduce 79) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 79) ((T_PIPE) reduce 79) ((T_POW) reduce 79) ((T_SL) reduce 79) ((T_SPACESHIP) reduce 79) ((T_SR) reduce 79) (("[") reduce 79) (("^") reduce 79) (("|") reduce 79)) 31062 ((("%") reduce 4) (("(") reduce 4) (("*") reduce 4) (("+") reduce 4) (("-") reduce 4) ((".") reduce 4) (("/") reduce 4) ((";") reduce 4) (("<") reduce 4) ((">") reduce 4) (("?") reduce 4) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 4) ((T_BOOLEAN_AND) reduce 4) ((T_BOOLEAN_OR) reduce 4) ((T_COALESCE) reduce 4) ((T_DOUBLE_ARROW) reduce 4) ((T_INSTANCEOF) reduce 4) ((T_IS_EQUAL) reduce 4) ((T_IS_GREATER_OR_EQUAL) reduce 4) ((T_IS_IDENTICAL) reduce 4) ((T_IS_NOT_EQUAL) reduce 4) ((T_IS_NOT_IDENTICAL) reduce 4) ((T_IS_SMALLER_OR_EQUAL) reduce 4) ((T_LOGICAL_AND) reduce 4) ((T_LOGICAL_OR) reduce 4) ((T_LOGICAL_XOR) reduce 4) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 4) ((T_OBJECT_OPERATOR) reduce 4) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 4) ((T_PIPE) reduce 4) ((T_POW) reduce 4) ((T_SL) reduce 4) ((T_SPACESHIP) reduce 4) ((T_SR) reduce 4) (("[") reduce 4) (("^") reduce 4) (("|") reduce 4)) 31063 ((("%") reduce 5) (("(") reduce 5) (("*") reduce 5) (("+") reduce 5) (("-") reduce 5) ((".") reduce 5) (("/") reduce 5) ((";") reduce 5) (("<") reduce 5) ((">") reduce 5) (("?") reduce 5) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 5) ((T_BOOLEAN_AND) reduce 5) ((T_BOOLEAN_OR) reduce 5) ((T_COALESCE) reduce 5) ((T_DOUBLE_ARROW) reduce 5) ((T_INSTANCEOF) reduce 5) ((T_IS_EQUAL) reduce 5) ((T_IS_GREATER_OR_EQUAL) reduce 5) ((T_IS_IDENTICAL) reduce 5) ((T_IS_NOT_EQUAL) reduce 5) ((T_IS_NOT_IDENTICAL) reduce 5) ((T_IS_SMALLER_OR_EQUAL) reduce 5) ((T_LOGICAL_AND) reduce 5) ((T_LOGICAL_OR) reduce 5) ((T_LOGICAL_XOR) reduce 5) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 5) ((T_OBJECT_OPERATOR) reduce 5) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 5) ((T_PIPE) reduce 5) ((T_POW) reduce 5) ((T_SL) reduce 5) ((T_SPACESHIP) reduce 5) ((T_SR) reduce 5) (("[") reduce 5) (("^") reduce 5) (("|") reduce 5)) 31064 ((("%") reduce 43) (("(") reduce 43) (("*") reduce 43) (("+") reduce 43) (("-") reduce 43) ((".") reduce 43) (("/") reduce 43) ((";") reduce 43) (("<") reduce 43) ((">") reduce 43) (("?") reduce 43) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 43) ((T_BOOLEAN_AND) reduce 43) ((T_BOOLEAN_OR) reduce 43) ((T_COALESCE) reduce 43) ((T_DOUBLE_ARROW) reduce 43) ((T_INSTANCEOF) reduce 43) ((T_IS_EQUAL) reduce 43) ((T_IS_GREATER_OR_EQUAL) reduce 43) ((T_IS_IDENTICAL) reduce 43) ((T_IS_NOT_EQUAL) reduce 43) ((T_IS_NOT_IDENTICAL) reduce 43) ((T_IS_SMALLER_OR_EQUAL) reduce 43) ((T_LOGICAL_AND) reduce 43) ((T_LOGICAL_OR) reduce 43) ((T_LOGICAL_XOR) reduce 43) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 43) ((T_OBJECT_OPERATOR) reduce 43) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 43) ((T_PIPE) reduce 43) ((T_POW) reduce 43) ((T_SL) reduce 43) ((T_SPACESHIP) reduce 43) ((T_SR) reduce 43) (("[") reduce 43) (("^") reduce 43) (("|") reduce 43)) 31065 ((("%") reduce 73) (("(") reduce 73) (("*") reduce 73) (("+") reduce 73) (("-") reduce 73) ((".") reduce 73) (("/") reduce 73) ((";") reduce 73) (("<") reduce 73) ((">") reduce 73) (("?") reduce 73) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 73) ((T_BOOLEAN_AND) reduce 73) ((T_BOOLEAN_OR) reduce 73) ((T_COALESCE) reduce 73) ((T_DOUBLE_ARROW) reduce 73) ((T_INSTANCEOF) reduce 73) ((T_IS_EQUAL) reduce 73) ((T_IS_GREATER_OR_EQUAL) reduce 73) ((T_IS_IDENTICAL) reduce 73) ((T_IS_NOT_EQUAL) reduce 73) ((T_IS_NOT_IDENTICAL) reduce 73) ((T_IS_SMALLER_OR_EQUAL) reduce 73) ((T_LOGICAL_AND) reduce 73) ((T_LOGICAL_OR) reduce 73) ((T_LOGICAL_XOR) reduce 73) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 73) ((T_OBJECT_OPERATOR) reduce 73) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 73) ((T_PIPE) reduce 73) ((T_POW) reduce 73) ((T_SL) reduce 73) ((T_SPACESHIP) reduce 73) ((T_SR) reduce 73) (("[") reduce 73) (("^") reduce 73) (("|") reduce 73)) 31066 ((("%") reduce 82) (("(") reduce 82) (("*") reduce 82) (("+") reduce 82) (("-") reduce 82) ((".") reduce 82) (("/") reduce 82) ((";") reduce 82) (("<") reduce 82) ((">") reduce 82) (("?") reduce 82) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 82) ((T_BOOLEAN_AND) reduce 82) ((T_BOOLEAN_OR) reduce 82) ((T_COALESCE) reduce 82) ((T_DOUBLE_ARROW) reduce 82) ((T_INSTANCEOF) reduce 82) ((T_IS_EQUAL) reduce 82) ((T_IS_GREATER_OR_EQUAL) reduce 82) ((T_IS_IDENTICAL) reduce 82) ((T_IS_NOT_EQUAL) reduce 82) ((T_IS_NOT_IDENTICAL) reduce 82) ((T_IS_SMALLER_OR_EQUAL) reduce 82) ((T_LOGICAL_AND) reduce 82) ((T_LOGICAL_OR) reduce 82) ((T_LOGICAL_XOR) reduce 82) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 82) ((T_OBJECT_OPERATOR) reduce 82) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 82) ((T_PIPE) reduce 82) ((T_POW) reduce 82) ((T_SL) reduce 82) ((T_SPACESHIP) reduce 82) ((T_SR) reduce 82) (("[") reduce 82) (("^") reduce 82) (("|") reduce 82)) 31067 ((("%") reduce 47) (("(") reduce 47) (("*") reduce 47) (("+") reduce 47) (("-") reduce 47) ((".") reduce 47) (("/") reduce 47) ((";") reduce 47) (("<") reduce 47) ((">") reduce 47) (("?") reduce 47) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 47) ((T_BOOLEAN_AND) reduce 47) ((T_BOOLEAN_OR) reduce 47) ((T_COALESCE) reduce 47) ((T_DOUBLE_ARROW) reduce 47) ((T_INSTANCEOF) reduce 47) ((T_IS_EQUAL) reduce 47) ((T_IS_GREATER_OR_EQUAL) reduce 47) ((T_IS_IDENTICAL) reduce 47) ((T_IS_NOT_EQUAL) reduce 47) ((T_IS_NOT_IDENTICAL) reduce 47) ((T_IS_SMALLER_OR_EQUAL) reduce 47) ((T_LOGICAL_AND) reduce 47) ((T_LOGICAL_OR) reduce 47) ((T_LOGICAL_XOR) reduce 47) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 47) ((T_OBJECT_OPERATOR) reduce 47) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 47) ((T_PIPE) reduce 47) ((T_POW) reduce 47) ((T_SL) reduce 47) ((T_SPACESHIP) reduce 47) ((T_SR) reduce 47) (("[") reduce 47) (("^") reduce 47) (("|") reduce 47)) 31068 ((("%") reduce 31) (("(") reduce 31) (("*") reduce 31) (("+") reduce 31) (("-") reduce 31) ((".") reduce 31) (("/") reduce 31) ((";") reduce 31) (("<") reduce 31) ((">") reduce 31) (("?") reduce 31) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 31) ((T_BOOLEAN_AND) reduce 31) ((T_BOOLEAN_OR) reduce 31) ((T_COALESCE) reduce 31) ((T_DOUBLE_ARROW) reduce 31) ((T_INSTANCEOF) reduce 31) ((T_IS_EQUAL) reduce 31) ((T_IS_GREATER_OR_EQUAL) reduce 31) ((T_IS_IDENTICAL) reduce 31) ((T_IS_NOT_EQUAL) reduce 31) ((T_IS_NOT_IDENTICAL) reduce 31) ((T_IS_SMALLER_OR_EQUAL) reduce 31) ((T_LOGICAL_AND) reduce 31) ((T_LOGICAL_OR) reduce 31) ((T_LOGICAL_XOR) reduce 31) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 31) ((T_OBJECT_OPERATOR) reduce 31) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 31) ((T_PIPE) reduce 31) ((T_POW) reduce 31) ((T_SL) reduce 31) ((T_SPACESHIP) reduce 31) ((T_SR) reduce 31) (("[") reduce 31) (("^") reduce 31) (("|") reduce 31)) 31069 ((("%") reduce 57) (("(") reduce 57) (("*") reduce 57) (("+") reduce 57) (("-") reduce 57) ((".") reduce 57) (("/") reduce 57) ((";") reduce 57) (("<") reduce 57) ((">") reduce 57) (("?") reduce 57) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 57) ((T_BOOLEAN_AND) reduce 57) ((T_BOOLEAN_OR) reduce 57) ((T_COALESCE) reduce 57) ((T_DOUBLE_ARROW) reduce 57) ((T_INSTANCEOF) reduce 57) ((T_IS_EQUAL) reduce 57) ((T_IS_GREATER_OR_EQUAL) reduce 57) ((T_IS_IDENTICAL) reduce 57) ((T_IS_NOT_EQUAL) reduce 57) ((T_IS_NOT_IDENTICAL) reduce 57) ((T_IS_SMALLER_OR_EQUAL) reduce 57) ((T_LOGICAL_AND) reduce 57) ((T_LOGICAL_OR) reduce 57) ((T_LOGICAL_XOR) reduce 57) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 57) ((T_OBJECT_OPERATOR) reduce 57) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 57) ((T_PIPE) reduce 57) ((T_POW) reduce 57) ((T_SL) reduce 57) ((T_SPACESHIP) reduce 57) ((T_SR) reduce 57) (("[") reduce 57) (("^") reduce 57) (("|") reduce 57)) 31070 ((("%") reduce 61) (("(") reduce 61) (("*") reduce 61) (("+") reduce 61) (("-") reduce 61) ((".") reduce 61) (("/") reduce 61) ((";") reduce 61) (("<") reduce 61) ((">") reduce 61) (("?") reduce 61) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 61) ((T_BOOLEAN_AND) reduce 61) ((T_BOOLEAN_OR) reduce 61) ((T_COALESCE) reduce 61) ((T_DOUBLE_ARROW) reduce 61) ((T_INSTANCEOF) reduce 61) ((T_IS_EQUAL) reduce 61) ((T_IS_GREATER_OR_EQUAL) reduce 61) ((T_IS_IDENTICAL) reduce 61) ((T_IS_NOT_EQUAL) reduce 61) ((T_IS_NOT_IDENTICAL) reduce 61) ((T_IS_SMALLER_OR_EQUAL) reduce 61) ((T_LOGICAL_AND) reduce 61) ((T_LOGICAL_OR) reduce 61) ((T_LOGICAL_XOR) reduce 61) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 61) ((T_OBJECT_OPERATOR) reduce 61) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 61) ((T_PIPE) reduce 61) ((T_POW) reduce 61) ((T_SL) reduce 61) ((T_SPACESHIP) reduce 61) ((T_SR) reduce 61) (("[") reduce 61) (("^") reduce 61) (("|") reduce 61)) 31071 ((("%") reduce 28) (("(") reduce 28) (("*") reduce 28) (("+") reduce 28) (("-") reduce 28) ((".") reduce 28) (("/") reduce 28) ((";") reduce 28) (("<") reduce 28) ((">") reduce 28) (("?") reduce 28) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 28) ((T_BOOLEAN_AND) reduce 28) ((T_BOOLEAN_OR) reduce 28) ((T_COALESCE) reduce 28) ((T_DOUBLE_ARROW) reduce 28) ((T_INSTANCEOF) reduce 28) ((T_IS_EQUAL) reduce 28) ((T_IS_GREATER_OR_EQUAL) reduce 28) ((T_IS_IDENTICAL) reduce 28) ((T_IS_NOT_EQUAL) reduce 28) ((T_IS_NOT_IDENTICAL) reduce 28) ((T_IS_SMALLER_OR_EQUAL) reduce 28) ((T_LOGICAL_AND) reduce 28) ((T_LOGICAL_OR) reduce 28) ((T_LOGICAL_XOR) reduce 28) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 28) ((T_OBJECT_OPERATOR) reduce 28) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 28) ((T_PIPE) reduce 28) ((T_POW) reduce 28) ((T_SL) reduce 28) ((T_SPACESHIP) reduce 28) ((T_SR) reduce 28) (("[") reduce 28) (("^") reduce 28) (("|") reduce 28)) 31072 ((("%") reduce 36) (("(") reduce 36) (("*") reduce 36) (("+") reduce 36) (("-") reduce 36) ((".") reduce 36) (("/") reduce 36) ((";") reduce 36) (("<") reduce 36) ((">") reduce 36) (("?") reduce 36) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 36) ((T_BOOLEAN_AND) reduce 36) ((T_BOOLEAN_OR) reduce 36) ((T_COALESCE) reduce 36) ((T_DOUBLE_ARROW) reduce 36) ((T_INSTANCEOF) reduce 36) ((T_IS_EQUAL) reduce 36) ((T_IS_GREATER_OR_EQUAL) reduce 36) ((T_IS_IDENTICAL) reduce 36) ((T_IS_NOT_EQUAL) reduce 36) ((T_IS_NOT_IDENTICAL) reduce 36) ((T_IS_SMALLER_OR_EQUAL) reduce 36) ((T_LOGICAL_AND) reduce 36) ((T_LOGICAL_OR) reduce 36) ((T_LOGICAL_XOR) reduce 36) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 36) ((T_OBJECT_OPERATOR) reduce 36) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 36) ((T_PIPE) reduce 36) ((T_POW) reduce 36) ((T_SL) reduce 36) ((T_SPACESHIP) reduce 36) ((T_SR) reduce 36) (("[") reduce 36) (("^") reduce 36) (("|") reduce 36)) 31073 ((("%") reduce 32) (("(") reduce 32) (("*") reduce 32) (("+") reduce 32) (("-") reduce 32) ((".") reduce 32) (("/") reduce 32) ((";") reduce 32) (("<") reduce 32) ((">") reduce 32) (("?") reduce 32) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 32) ((T_BOOLEAN_AND) reduce 32) ((T_BOOLEAN_OR) reduce 32) ((T_COALESCE) reduce 32) ((T_DOUBLE_ARROW) reduce 32) ((T_INSTANCEOF) reduce 32) ((T_IS_EQUAL) reduce 32) ((T_IS_GREATER_OR_EQUAL) reduce 32) ((T_IS_IDENTICAL) reduce 32) ((T_IS_NOT_EQUAL) reduce 32) ((T_IS_NOT_IDENTICAL) reduce 32) ((T_IS_SMALLER_OR_EQUAL) reduce 32) ((T_LOGICAL_AND) reduce 32) ((T_LOGICAL_OR) reduce 32) ((T_LOGICAL_XOR) reduce 32) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 32) ((T_OBJECT_OPERATOR) reduce 32) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 32) ((T_PIPE) reduce 32) ((T_POW) reduce 32) ((T_SL) reduce 32) ((T_SPACESHIP) reduce 32) ((T_SR) reduce 32) (("[") reduce 32) (("^") reduce 32) (("|") reduce 32)) 31074 ((("%") reduce 35) (("(") reduce 35) (("*") reduce 35) (("+") reduce 35) (("-") reduce 35) ((".") reduce 35) (("/") reduce 35) ((";") reduce 35) (("<") reduce 35) ((">") reduce 35) (("?") reduce 35) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 35) ((T_BOOLEAN_AND) reduce 35) ((T_BOOLEAN_OR) reduce 35) ((T_COALESCE) reduce 35) ((T_DOUBLE_ARROW) reduce 35) ((T_INSTANCEOF) reduce 35) ((T_IS_EQUAL) reduce 35) ((T_IS_GREATER_OR_EQUAL) reduce 35) ((T_IS_IDENTICAL) reduce 35) ((T_IS_NOT_EQUAL) reduce 35) ((T_IS_NOT_IDENTICAL) reduce 35) ((T_IS_SMALLER_OR_EQUAL) reduce 35) ((T_LOGICAL_AND) reduce 35) ((T_LOGICAL_OR) reduce 35) ((T_LOGICAL_XOR) reduce 35) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 35) ((T_OBJECT_OPERATOR) reduce 35) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 35) ((T_PIPE) reduce 35) ((T_POW) reduce 35) ((T_SL) reduce 35) ((T_SPACESHIP) reduce 35) ((T_SR) reduce 35) (("[") reduce 35) (("^") reduce 35) (("|") reduce 35)) 31075 ((("%") reduce 19) (("(") reduce 19) (("*") reduce 19) (("+") reduce 19) (("-") reduce 19) ((".") reduce 19) (("/") reduce 19) ((";") reduce 19) (("<") reduce 19) ((">") reduce 19) (("?") reduce 19) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 19) ((T_BOOLEAN_AND) reduce 19) ((T_BOOLEAN_OR) reduce 19) ((T_COALESCE) reduce 19) ((T_DOUBLE_ARROW) reduce 19) ((T_INSTANCEOF) reduce 19) ((T_IS_EQUAL) reduce 19) ((T_IS_GREATER_OR_EQUAL) reduce 19) ((T_IS_IDENTICAL) reduce 19) ((T_IS_NOT_EQUAL) reduce 19) ((T_IS_NOT_IDENTICAL) reduce 19) ((T_IS_SMALLER_OR_EQUAL) reduce 19) ((T_LOGICAL_AND) reduce 19) ((T_LOGICAL_OR) reduce 19) ((T_LOGICAL_XOR) reduce 19) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 19) ((T_OBJECT_OPERATOR) reduce 19) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 19) ((T_PIPE) reduce 19) ((T_POW) reduce 19) ((T_SL) reduce 19) ((T_SPACESHIP) reduce 19) ((T_SR) reduce 19) (("[") reduce 19) (("^") reduce 19) (("|") reduce 19)) 31076 ((("%") reduce 45) (("(") reduce 45) (("*") reduce 45) (("+") reduce 45) (("-") reduce 45) ((".") reduce 45) (("/") reduce 45) ((";") reduce 45) (("<") reduce 45) ((">") reduce 45) (("?") reduce 45) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 45) ((T_BOOLEAN_AND) reduce 45) ((T_BOOLEAN_OR) reduce 45) ((T_COALESCE) reduce 45) ((T_DOUBLE_ARROW) reduce 45) ((T_INSTANCEOF) reduce 45) ((T_IS_EQUAL) reduce 45) ((T_IS_GREATER_OR_EQUAL) reduce 45) ((T_IS_IDENTICAL) reduce 45) ((T_IS_NOT_EQUAL) reduce 45) ((T_IS_NOT_IDENTICAL) reduce 45) ((T_IS_SMALLER_OR_EQUAL) reduce 45) ((T_LOGICAL_AND) reduce 45) ((T_LOGICAL_OR) reduce 45) ((T_LOGICAL_XOR) reduce 45) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 45) ((T_OBJECT_OPERATOR) reduce 45) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 45) ((T_PIPE) reduce 45) ((T_POW) reduce 45) ((T_SL) reduce 45) ((T_SPACESHIP) reduce 45) ((T_SR) reduce 45) (("[") reduce 45) (("^") reduce 45) (("|") reduce 45)) 31077 ((("%") reduce 542) (("(") reduce 579) (("*") reduce 542) (("+") reduce 542) (("-") reduce 542) ((".") reduce 542) (("/") reduce 542) ((";") reduce 542) (("<") reduce 542) ((">") reduce 542) (("?") reduce 542) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 542) ((T_BOOLEAN_AND) reduce 542) ((T_BOOLEAN_OR) reduce 542) ((T_COALESCE) reduce 542) ((T_DOUBLE_ARROW) reduce 542) ((T_INSTANCEOF) reduce 542) ((T_IS_EQUAL) reduce 542) ((T_IS_GREATER_OR_EQUAL) reduce 542) ((T_IS_IDENTICAL) reduce 542) ((T_IS_NOT_EQUAL) reduce 542) ((T_IS_NOT_IDENTICAL) reduce 542) ((T_IS_SMALLER_OR_EQUAL) reduce 542) ((T_LOGICAL_AND) reduce 542) ((T_LOGICAL_OR) reduce 542) ((T_LOGICAL_XOR) reduce 542) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 542) ((T_OBJECT_OPERATOR) reduce 542) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 542) ((T_PIPE) reduce 542) ((T_POW) reduce 542) ((T_SL) reduce 542) ((T_SPACESHIP) reduce 542) ((T_SR) reduce 542) (("[") reduce 542) (("^") reduce 542) (("|") reduce 542)) 31079 ((("%") reduce 72) (("(") reduce 72) (("*") reduce 72) (("+") reduce 72) (("-") reduce 72) ((".") reduce 72) (("/") reduce 72) ((";") reduce 72) (("<") reduce 72) ((">") reduce 72) (("?") reduce 72) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 72) ((T_BOOLEAN_AND) reduce 72) ((T_BOOLEAN_OR) reduce 72) ((T_COALESCE) reduce 72) ((T_DOUBLE_ARROW) reduce 72) ((T_INSTANCEOF) reduce 72) ((T_IS_EQUAL) reduce 72) ((T_IS_GREATER_OR_EQUAL) reduce 72) ((T_IS_IDENTICAL) reduce 72) ((T_IS_NOT_EQUAL) reduce 72) ((T_IS_NOT_IDENTICAL) reduce 72) ((T_IS_SMALLER_OR_EQUAL) reduce 72) ((T_LOGICAL_AND) reduce 72) ((T_LOGICAL_OR) reduce 72) ((T_LOGICAL_XOR) reduce 72) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 72) ((T_OBJECT_OPERATOR) reduce 72) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 72) ((T_PIPE) reduce 72) ((T_POW) reduce 72) ((T_SL) reduce 72) ((T_SPACESHIP) reduce 72) ((T_SR) reduce 72) (("[") reduce 72) (("^") reduce 72) (("|") reduce 72)) 31080 ((("%") reduce 83) (("(") reduce 83) (("*") reduce 83) (("+") reduce 83) (("-") reduce 83) ((".") reduce 83) (("/") reduce 83) ((";") reduce 83) (("<") reduce 83) ((">") reduce 83) (("?") reduce 83) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 83) ((T_BOOLEAN_AND) reduce 83) ((T_BOOLEAN_OR) reduce 83) ((T_COALESCE) reduce 83) ((T_DOUBLE_ARROW) reduce 83) ((T_INSTANCEOF) reduce 83) ((T_IS_EQUAL) reduce 83) ((T_IS_GREATER_OR_EQUAL) reduce 83) ((T_IS_IDENTICAL) reduce 83) ((T_IS_NOT_EQUAL) reduce 83) ((T_IS_NOT_IDENTICAL) reduce 83) ((T_IS_SMALLER_OR_EQUAL) reduce 83) ((T_LOGICAL_AND) reduce 83) ((T_LOGICAL_OR) reduce 83) ((T_LOGICAL_XOR) reduce 83) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 83) ((T_OBJECT_OPERATOR) reduce 83) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 83) ((T_PIPE) reduce 83) ((T_POW) reduce 83) ((T_SL) reduce 83) ((T_SPACESHIP) reduce 83) ((T_SR) reduce 83) (("[") reduce 83) (("^") reduce 83) (("|") reduce 83)) 31081 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) ((";") reduce 572) (("<") reduce 572) (("=") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AND_EQUAL) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_COALESCE_EQUAL) reduce 572) ((T_CONCAT_EQUAL) reduce 572) ((T_DEC) reduce 572) ((T_DIV_EQUAL) reduce 572) ((T_DOUBLE_ARROW) reduce 572) ((T_INC) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_MINUS_EQUAL) reduce 572) ((T_MOD_EQUAL) reduce 572) ((T_MUL_EQUAL) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_OR_EQUAL) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_PLUS_EQUAL) reduce 572) ((T_POW) reduce 572) ((T_POW_EQUAL) reduce 572) ((T_SL) reduce 572) ((T_SL_EQUAL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) ((T_SR_EQUAL) reduce 572) ((T_XOR_EQUAL) reduce 572) (("[") reduce 572) (("^") reduce 572) (("|") reduce 572)) 31084 ((("%") reduce 544) (("(") reduce 580) (("*") reduce 544) (("+") reduce 544) (("-") reduce 544) ((".") reduce 544) (("/") reduce 544) ((";") reduce 544) (("<") reduce 544) ((">") reduce 544) (("?") reduce 544) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 544) ((T_BOOLEAN_AND) reduce 544) ((T_BOOLEAN_OR) reduce 544) ((T_COALESCE) reduce 544) ((T_DOUBLE_ARROW) reduce 544) ((T_INSTANCEOF) reduce 544) ((T_IS_EQUAL) reduce 544) ((T_IS_GREATER_OR_EQUAL) reduce 544) ((T_IS_IDENTICAL) reduce 544) ((T_IS_NOT_EQUAL) reduce 544) ((T_IS_NOT_IDENTICAL) reduce 544) ((T_IS_SMALLER_OR_EQUAL) reduce 544) ((T_LOGICAL_AND) reduce 544) ((T_LOGICAL_OR) reduce 544) ((T_LOGICAL_XOR) reduce 544) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 544) ((T_OBJECT_OPERATOR) reduce 544) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 544) ((T_PIPE) reduce 544) ((T_POW) reduce 544) ((T_SL) reduce 544) ((T_SPACESHIP) reduce 544) ((T_SR) reduce 544) (("[") reduce 544) (("^") reduce 544) (("|") reduce 544)) 31086 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((";") reduce 507) (("<") reduce 507) (("=") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AND_EQUAL) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_COALESCE_EQUAL) reduce 507) ((T_CONCAT_EQUAL) reduce 507) ((T_DEC) reduce 507) ((T_DIV_EQUAL) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INC) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_MINUS_EQUAL) reduce 507) ((T_MOD_EQUAL) reduce 507) ((T_MUL_EQUAL) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_OR_EQUAL) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_PLUS_EQUAL) reduce 507) ((T_POW) reduce 507) ((T_POW_EQUAL) reduce 507) ((T_SL) reduce 507) ((T_SL_EQUAL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) ((T_SR_EQUAL) reduce 507) ((T_XOR_EQUAL) reduce 507) (("[") reduce 507) (("^") reduce 507) (("|") reduce 507)) 31087 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((";") reduce 291) (("<") reduce 291) (("=") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AND_EQUAL) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_COALESCE_EQUAL) reduce 291) ((T_CONCAT_EQUAL) reduce 291) ((T_DEC) reduce 291) ((T_DIV_EQUAL) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INC) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_MINUS_EQUAL) reduce 291) ((T_MOD_EQUAL) reduce 291) ((T_MUL_EQUAL) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_OR_EQUAL) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_PLUS_EQUAL) reduce 291) ((T_POW) reduce 291) ((T_POW_EQUAL) reduce 291) ((T_SL) reduce 291) ((T_SL_EQUAL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) ((T_SR_EQUAL) reduce 291) ((T_XOR_EQUAL) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 31091 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) ((";") reduce 292) (("<") reduce 292) (("=") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AND_EQUAL) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_COALESCE_EQUAL) reduce 292) ((T_CONCAT_EQUAL) reduce 292) ((T_DEC) reduce 292) ((T_DIV_EQUAL) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INC) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_MINUS_EQUAL) reduce 292) ((T_MOD_EQUAL) reduce 292) ((T_MUL_EQUAL) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_OR_EQUAL) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_PLUS_EQUAL) reduce 292) ((T_POW) reduce 292) ((T_POW_EQUAL) reduce 292) ((T_SL) reduce 292) ((T_SL_EQUAL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) ((T_SR_EQUAL) reduce 292) ((T_XOR_EQUAL) reduce 292) (("[") reduce 292) (("^") reduce 292) (("|") reduce 292)) 31092 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) ((";") reduce 293) (("<") reduce 293) (("=") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AND_EQUAL) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_COALESCE_EQUAL) reduce 293) ((T_CONCAT_EQUAL) reduce 293) ((T_DEC) reduce 293) ((T_DIV_EQUAL) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INC) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_MINUS_EQUAL) reduce 293) ((T_MOD_EQUAL) reduce 293) ((T_MUL_EQUAL) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_OR_EQUAL) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_PLUS_EQUAL) reduce 293) ((T_POW) reduce 293) ((T_POW_EQUAL) reduce 293) ((T_SL) reduce 293) ((T_SL_EQUAL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) ((T_SR_EQUAL) reduce 293) ((T_XOR_EQUAL) reduce 293) (("[") reduce 293) (("^") reduce 293) (("|") reduce 293)) 31097 ((("%") reduce 425) (("*") reduce 425) (("+") reduce 425) (("-") reduce 425) ((".") reduce 425) (("/") reduce 425) ((";") reduce 425) (("<") reduce 425) ((">") reduce 425) (("?") reduce 425) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 425) ((T_BOOLEAN_AND) reduce 425) ((T_BOOLEAN_OR) reduce 425) ((T_COALESCE) reduce 425) ((T_DOUBLE_ARROW) reduce 425) ((T_INSTANCEOF) reduce 425) ((T_IS_EQUAL) reduce 425) ((T_IS_GREATER_OR_EQUAL) reduce 425) ((T_IS_IDENTICAL) reduce 425) ((T_IS_NOT_EQUAL) reduce 425) ((T_IS_NOT_IDENTICAL) reduce 425) ((T_IS_SMALLER_OR_EQUAL) reduce 425) ((T_LOGICAL_AND) reduce 425) ((T_LOGICAL_OR) reduce 425) ((T_LOGICAL_XOR) reduce 425) ((T_PIPE) reduce 425) ((T_POW) reduce 425) ((T_SL) reduce 425) ((T_SPACESHIP) reduce 425) ((T_SR) reduce 425) (("^") reduce 425) (("|") reduce 425)) 31099 ((("%") reduce 423) (("*") reduce 423) (("+") reduce 423) (("-") reduce 423) ((".") reduce 423) (("/") reduce 423) ((";") reduce 423) (("<") reduce 423) ((">") reduce 423) (("?") reduce 423) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 423) ((T_BOOLEAN_AND) reduce 423) ((T_BOOLEAN_OR) reduce 423) ((T_COALESCE) reduce 423) ((T_DOUBLE_ARROW) reduce 423) ((T_INSTANCEOF) reduce 423) ((T_IS_EQUAL) reduce 423) ((T_IS_GREATER_OR_EQUAL) reduce 423) ((T_IS_IDENTICAL) reduce 423) ((T_IS_NOT_EQUAL) reduce 423) ((T_IS_NOT_IDENTICAL) reduce 423) ((T_IS_SMALLER_OR_EQUAL) reduce 423) ((T_LOGICAL_AND) reduce 423) ((T_LOGICAL_OR) reduce 423) ((T_LOGICAL_XOR) reduce 423) ((T_PIPE) reduce 423) ((T_POW) reduce 423) ((T_SL) reduce 423) ((T_SPACESHIP) reduce 423) ((T_SR) reduce 423) (("^") reduce 423) (("|") reduce 423)) 31109 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 419) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 419) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 419) ((T_LOGICAL_OR) reduce 419) ((T_LOGICAL_XOR) reduce 419) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 31110 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 421) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 421) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 421) ((T_LOGICAL_OR) reduce 421) ((T_LOGICAL_XOR) reduce 421) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 31111 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 420) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 420) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 420) ((T_LOGICAL_OR) reduce 420) ((T_LOGICAL_XOR) reduce 420) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 31112 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 413) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 413) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 413) ((T_LOGICAL_OR) reduce 413) ((T_LOGICAL_XOR) reduce 413) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 31113 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 410) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 410) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 410) ((T_LOGICAL_OR) reduce 410) ((T_LOGICAL_XOR) reduce 410) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 31114 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 418) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 418) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 418) ((T_LOGICAL_OR) reduce 418) ((T_LOGICAL_XOR) reduce 418) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 31115 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 412) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 412) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 412) ((T_LOGICAL_OR) reduce 412) ((T_LOGICAL_XOR) reduce 412) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 31116 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 416) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 416) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 416) ((T_LOGICAL_OR) reduce 416) ((T_LOGICAL_XOR) reduce 416) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 31117 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 411) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 411) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 411) ((T_LOGICAL_OR) reduce 411) ((T_LOGICAL_XOR) reduce 411) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 31118 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 414) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 414) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 414) ((T_LOGICAL_OR) reduce 414) ((T_LOGICAL_XOR) reduce 414) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 31119 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 415) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 415) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 415) ((T_LOGICAL_OR) reduce 415) ((T_LOGICAL_XOR) reduce 415) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 31120 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 422) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 422) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 422) ((T_LOGICAL_OR) reduce 422) ((T_LOGICAL_XOR) reduce 422) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 31121 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 417) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 417) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 417) ((T_LOGICAL_OR) reduce 417) ((T_LOGICAL_XOR) reduce 417) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 31123 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 406) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 406) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 406) ((T_LOGICAL_OR) reduce 406) ((T_LOGICAL_XOR) reduce 406) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 31126 ((("%") reduce 568) (("(") reduce 568) (("*") reduce 568) (("+") reduce 568) (("-") reduce 568) ((".") reduce 568) (("/") reduce 568) ((";") reduce 568) (("<") reduce 568) ((">") reduce 568) (("?") reduce 568) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 568) ((T_BOOLEAN_AND) reduce 568) ((T_BOOLEAN_OR) reduce 568) ((T_COALESCE) reduce 568) ((T_DOUBLE_ARROW) reduce 568) ((T_INSTANCEOF) reduce 568) ((T_IS_EQUAL) reduce 568) ((T_IS_GREATER_OR_EQUAL) reduce 568) ((T_IS_IDENTICAL) reduce 568) ((T_IS_NOT_EQUAL) reduce 568) ((T_IS_NOT_IDENTICAL) reduce 568) ((T_IS_SMALLER_OR_EQUAL) reduce 568) ((T_LOGICAL_AND) reduce 568) ((T_LOGICAL_OR) reduce 568) ((T_LOGICAL_XOR) reduce 568) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 568) ((T_OBJECT_OPERATOR) reduce 568) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 568) ((T_PIPE) reduce 568) ((T_POW) reduce 568) ((T_SL) reduce 568) ((T_SPACESHIP) reduce 568) ((T_SR) reduce 568) (("[") reduce 568) (("^") reduce 568) (("|") reduce 568)) 31129 ((("%") reduce 564) (("(") reduce 555) (("*") reduce 564) (("+") reduce 564) (("-") reduce 564) ((".") reduce 564) (("/") reduce 564) ((";") reduce 564) (("<") reduce 564) ((">") reduce 564) (("?") reduce 564) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 564) ((T_BOOLEAN_AND) reduce 564) ((T_BOOLEAN_OR) reduce 564) ((T_COALESCE) reduce 564) ((T_DOUBLE_ARROW) reduce 564) ((T_INSTANCEOF) reduce 564) ((T_IS_EQUAL) reduce 564) ((T_IS_GREATER_OR_EQUAL) reduce 564) ((T_IS_IDENTICAL) reduce 564) ((T_IS_NOT_EQUAL) reduce 564) ((T_IS_NOT_IDENTICAL) reduce 564) ((T_IS_SMALLER_OR_EQUAL) reduce 564) ((T_LOGICAL_AND) reduce 564) ((T_LOGICAL_OR) reduce 564) ((T_LOGICAL_XOR) reduce 564) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 564) ((T_OBJECT_OPERATOR) reduce 564) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 564) ((T_PIPE) reduce 564) ((T_POW) reduce 564) ((T_SL) reduce 564) ((T_SPACESHIP) reduce 564) ((T_SR) reduce 564) (("[") reduce 564) (("^") reduce 564) (("|") reduce 564)) 31131 ((("%") reduce 563) (("(") reduce 563) (("*") reduce 563) (("+") reduce 563) (("-") reduce 563) ((".") reduce 563) (("/") reduce 563) ((";") reduce 563) (("<") reduce 563) ((">") reduce 563) (("?") reduce 563) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 563) ((T_BOOLEAN_AND) reduce 563) ((T_BOOLEAN_OR) reduce 563) ((T_COALESCE) reduce 563) ((T_DOUBLE_ARROW) reduce 563) ((T_INSTANCEOF) reduce 563) ((T_IS_EQUAL) reduce 563) ((T_IS_GREATER_OR_EQUAL) reduce 563) ((T_IS_IDENTICAL) reduce 563) ((T_IS_NOT_EQUAL) reduce 563) ((T_IS_NOT_IDENTICAL) reduce 563) ((T_IS_SMALLER_OR_EQUAL) reduce 563) ((T_LOGICAL_AND) reduce 563) ((T_LOGICAL_OR) reduce 563) ((T_LOGICAL_XOR) reduce 563) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 563) ((T_OBJECT_OPERATOR) reduce 563) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 563) ((T_PIPE) reduce 563) ((T_POW) reduce 563) ((T_SL) reduce 563) ((T_SPACESHIP) reduce 563) ((T_SR) reduce 563) (("[") reduce 563) (("^") reduce 563) (("|") reduce 563)) 31133 ((("%") reduce 559) (("(") reduce 559) (("*") reduce 559) (("+") reduce 559) (("-") reduce 559) ((".") reduce 559) (("/") reduce 559) ((";") reduce 559) (("<") reduce 559) ((">") reduce 559) (("?") reduce 559) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 559) ((T_BOOLEAN_AND) reduce 559) ((T_BOOLEAN_OR) reduce 559) ((T_COALESCE) reduce 559) ((T_DOUBLE_ARROW) reduce 559) ((T_INSTANCEOF) reduce 559) ((T_IS_EQUAL) reduce 559) ((T_IS_GREATER_OR_EQUAL) reduce 559) ((T_IS_IDENTICAL) reduce 559) ((T_IS_NOT_EQUAL) reduce 559) ((T_IS_NOT_IDENTICAL) reduce 559) ((T_IS_SMALLER_OR_EQUAL) reduce 559) ((T_LOGICAL_AND) reduce 559) ((T_LOGICAL_OR) reduce 559) ((T_LOGICAL_XOR) reduce 559) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 559) ((T_OBJECT_OPERATOR) reduce 559) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 559) ((T_PIPE) reduce 559) ((T_POW) reduce 559) ((T_SL) reduce 559) ((T_SPACESHIP) reduce 559) ((T_SR) reduce 559) (("[") reduce 559) (("^") reduce 559) (("|") reduce 559)) 31134 ((("%") reduce 565) (("*") reduce 565) (("+") reduce 565) (("-") reduce 565) ((".") reduce 565) (("/") reduce 565) ((";") reduce 565) (("<") reduce 565) ((">") reduce 565) (("?") reduce 565) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 565) ((T_BOOLEAN_AND) reduce 565) ((T_BOOLEAN_OR) reduce 565) ((T_COALESCE) reduce 565) ((T_DOUBLE_ARROW) reduce 565) ((T_INSTANCEOF) reduce 565) ((T_IS_EQUAL) reduce 565) ((T_IS_GREATER_OR_EQUAL) reduce 565) ((T_IS_IDENTICAL) reduce 565) ((T_IS_NOT_EQUAL) reduce 565) ((T_IS_NOT_IDENTICAL) reduce 565) ((T_IS_SMALLER_OR_EQUAL) reduce 565) ((T_LOGICAL_AND) reduce 565) ((T_LOGICAL_OR) reduce 565) ((T_LOGICAL_XOR) reduce 565) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 565) ((T_OBJECT_OPERATOR) reduce 565) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 565) ((T_PIPE) reduce 565) ((T_POW) reduce 565) ((T_SL) reduce 565) ((T_SPACESHIP) reduce 565) ((T_SR) reduce 565) (("[") reduce 565) (("^") reduce 565) (("|") reduce 565)) 31135 ((("%") reduce 407) (("*") reduce 407) (("+") reduce 407) (("-") reduce 407) ((".") reduce 407) (("/") reduce 407) ((";") reduce 407) (("<") reduce 407) ((">") reduce 407) (("?") reduce 407) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 407) ((T_BOOLEAN_AND) reduce 407) ((T_BOOLEAN_OR) reduce 407) ((T_COALESCE) reduce 407) ((T_DOUBLE_ARROW) reduce 407) ((T_INSTANCEOF) reduce 407) ((T_IS_EQUAL) reduce 407) ((T_IS_GREATER_OR_EQUAL) reduce 407) ((T_IS_IDENTICAL) reduce 407) ((T_IS_NOT_EQUAL) reduce 407) ((T_IS_NOT_IDENTICAL) reduce 407) ((T_IS_SMALLER_OR_EQUAL) reduce 407) ((T_LOGICAL_AND) reduce 407) ((T_LOGICAL_OR) reduce 407) ((T_LOGICAL_XOR) reduce 407) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 407) ((T_POW) reduce 407) ((T_SL) reduce 407) ((T_SPACESHIP) reduce 407) ((T_SR) reduce 407) (("[") reduce 548) (("^") reduce 407) (("|") reduce 407)) 31139 ((("%") reduce 572) (("(") reduce 581) (("*") reduce 572) (("+") reduce 572) (("-") reduce 572) ((".") reduce 572) (("/") reduce 572) ((";") reduce 572) (("<") reduce 572) ((">") reduce 572) (("?") reduce 572) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 572) ((T_BOOLEAN_AND) reduce 572) ((T_BOOLEAN_OR) reduce 572) ((T_COALESCE) reduce 572) ((T_DOUBLE_ARROW) reduce 572) ((T_INSTANCEOF) reduce 572) ((T_IS_EQUAL) reduce 572) ((T_IS_GREATER_OR_EQUAL) reduce 572) ((T_IS_IDENTICAL) reduce 572) ((T_IS_NOT_EQUAL) reduce 572) ((T_IS_NOT_IDENTICAL) reduce 572) ((T_IS_SMALLER_OR_EQUAL) reduce 572) ((T_LOGICAL_AND) reduce 572) ((T_LOGICAL_OR) reduce 572) ((T_LOGICAL_XOR) reduce 572) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 572) ((T_OBJECT_OPERATOR) reduce 572) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 572) ((T_PIPE) reduce 572) ((T_POW) reduce 572) ((T_SL) reduce 572) ((T_SPACESHIP) reduce 572) ((T_SR) reduce 572) (("[") reduce 572) (("^") reduce 572) (("|") reduce 572)) 31141 ((("%") reduce 507) (("(") reduce 507) (("*") reduce 507) (("+") reduce 507) (("-") reduce 507) ((".") reduce 507) (("/") reduce 507) ((";") reduce 507) (("<") reduce 507) ((">") reduce 507) (("?") reduce 507) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 507) ((T_BOOLEAN_AND) reduce 507) ((T_BOOLEAN_OR) reduce 507) ((T_COALESCE) reduce 507) ((T_DOUBLE_ARROW) reduce 507) ((T_INSTANCEOF) reduce 507) ((T_IS_EQUAL) reduce 507) ((T_IS_GREATER_OR_EQUAL) reduce 507) ((T_IS_IDENTICAL) reduce 507) ((T_IS_NOT_EQUAL) reduce 507) ((T_IS_NOT_IDENTICAL) reduce 507) ((T_IS_SMALLER_OR_EQUAL) reduce 507) ((T_LOGICAL_AND) reduce 507) ((T_LOGICAL_OR) reduce 507) ((T_LOGICAL_XOR) reduce 507) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 507) ((T_OBJECT_OPERATOR) reduce 507) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 507) ((T_PIPE) reduce 507) ((T_POW) reduce 507) ((T_SL) reduce 507) ((T_SPACESHIP) reduce 507) ((T_SR) reduce 507) (("[") reduce 507) (("^") reduce 507) (("|") reduce 507)) 31142 ((("%") reduce 291) (("(") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((";") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 291) ((T_OBJECT_OPERATOR) reduce 291) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("[") reduce 291) (("^") reduce 291) (("|") reduce 291)) 31146 ((("%") reduce 292) (("(") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) ((";") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 292) ((T_OBJECT_OPERATOR) reduce 292) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("[") reduce 292) (("^") reduce 292) (("|") reduce 292)) 31147 ((("%") reduce 293) (("(") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) ((";") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 293) ((T_OBJECT_OPERATOR) reduce 293) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("[") reduce 293) (("^") reduce 293) (("|") reduce 293)) 31148 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) ((";") reduce 504) (("<") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_POW) reduce 504) ((T_SL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 31151 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) ((";") reduce 571) (("<") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_DOUBLE_ARROW) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_POW) reduce 571) ((T_SL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) (("[") reduce 571) (("^") reduce 571) (("|") reduce 571)) 31152 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((";") reduce 506) (("<") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_POW) reduce 506) ((T_SL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) (("[") reduce 506) (("^") reduce 506) (("|") reduce 506)) 31153 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((";") reduce 508) (("<") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_POW) reduce 508) ((T_SL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) (("[") reduce 508) (("^") reduce 508) (("|") reduce 508)) 31158 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) ((";") reduce 560) (("<") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_POW) reduce 560) ((T_SL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 31159 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) ((";") reduce 582) (("<") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_POW) reduce 582) ((T_SL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 31160 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((";") reduce 566) (("<") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_POW) reduce 566) ((T_SL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 31161 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) ((";") reduce 584) (("<") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_POW) reduce 584) ((T_SL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 31164 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) ((";") reduce 583) (("<") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_POW) reduce 583) ((T_SL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 31165 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((";") reduce 561) (("<") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_POW) reduce 561) ((T_SL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 31166 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((";") reduce 567) (("<") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_POW) reduce 567) ((T_SL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 31167 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) ((";") reduce 562) (("<") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_POW) reduce 562) ((T_SL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) (("[") reduce 562) (("^") reduce 562) (("|") reduce 562)) 31168 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((";") reduce 505) (("<") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_POW) reduce 505) ((T_SL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 31169 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) ((";") reduce 570) (("<") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_POW) reduce 570) ((T_SL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 31172 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) ((";") reduce 569) (("<") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_POW) reduce 569) ((T_SL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 31173 ((("%") reduce 504) (("(") reduce 504) (("*") reduce 504) (("+") reduce 504) (("-") reduce 504) ((".") reduce 504) (("/") reduce 504) ((";") reduce 504) (("<") reduce 504) (("=") reduce 504) ((">") reduce 504) (("?") reduce 504) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 504) ((T_AND_EQUAL) reduce 504) ((T_BOOLEAN_AND) reduce 504) ((T_BOOLEAN_OR) reduce 504) ((T_COALESCE) reduce 504) ((T_COALESCE_EQUAL) reduce 504) ((T_CONCAT_EQUAL) reduce 504) ((T_DEC) reduce 504) ((T_DIV_EQUAL) reduce 504) ((T_DOUBLE_ARROW) reduce 504) ((T_INC) reduce 504) ((T_INSTANCEOF) reduce 504) ((T_IS_EQUAL) reduce 504) ((T_IS_GREATER_OR_EQUAL) reduce 504) ((T_IS_IDENTICAL) reduce 504) ((T_IS_NOT_EQUAL) reduce 504) ((T_IS_NOT_IDENTICAL) reduce 504) ((T_IS_SMALLER_OR_EQUAL) reduce 504) ((T_LOGICAL_AND) reduce 504) ((T_LOGICAL_OR) reduce 504) ((T_LOGICAL_XOR) reduce 504) ((T_MINUS_EQUAL) reduce 504) ((T_MOD_EQUAL) reduce 504) ((T_MUL_EQUAL) reduce 504) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 504) ((T_OBJECT_OPERATOR) reduce 504) ((T_OR_EQUAL) reduce 504) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 504) ((T_PIPE) reduce 504) ((T_PLUS_EQUAL) reduce 504) ((T_POW) reduce 504) ((T_POW_EQUAL) reduce 504) ((T_SL) reduce 504) ((T_SL_EQUAL) reduce 504) ((T_SPACESHIP) reduce 504) ((T_SR) reduce 504) ((T_SR_EQUAL) reduce 504) ((T_XOR_EQUAL) reduce 504) (("[") reduce 504) (("^") reduce 504) (("|") reduce 504)) 31184 ((("%") reduce 494) (("*") reduce 494) (("+") reduce 494) (("-") reduce 494) ((".") reduce 494) (("/") reduce 494) ((";") reduce 494) (("<") reduce 494) ((">") reduce 494) (("?") reduce 494) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 494) ((T_BOOLEAN_AND) reduce 494) ((T_BOOLEAN_OR) reduce 494) ((T_COALESCE) reduce 494) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) reduce 494) ((T_IS_EQUAL) reduce 494) ((T_IS_GREATER_OR_EQUAL) reduce 494) ((T_IS_IDENTICAL) reduce 494) ((T_IS_NOT_EQUAL) reduce 494) ((T_IS_NOT_IDENTICAL) reduce 494) ((T_IS_SMALLER_OR_EQUAL) reduce 494) ((T_LOGICAL_AND) reduce 494) ((T_LOGICAL_OR) reduce 494) ((T_LOGICAL_XOR) reduce 494) ((T_PIPE) reduce 494) ((T_POW) reduce 494) ((T_SL) reduce 494) ((T_SPACESHIP) reduce 494) ((T_SR) reduce 494) (("^") reduce 494) (("|") reduce 494)) 31185 ((("%") reduce 489) (("*") reduce 489) (("+") reduce 489) (("-") reduce 489) ((".") reduce 489) (("/") reduce 489) ((";") reduce 489) (("<") reduce 489) ((">") reduce 489) (("?") reduce 489) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 489) ((T_BOOLEAN_AND) reduce 489) ((T_BOOLEAN_OR) reduce 489) ((T_COALESCE) reduce 489) ((T_DOUBLE_ARROW) reduce 489) ((T_INSTANCEOF) reduce 489) ((T_IS_EQUAL) reduce 489) ((T_IS_GREATER_OR_EQUAL) reduce 489) ((T_IS_IDENTICAL) reduce 489) ((T_IS_NOT_EQUAL) reduce 489) ((T_IS_NOT_IDENTICAL) reduce 489) ((T_IS_SMALLER_OR_EQUAL) reduce 489) ((T_LOGICAL_AND) reduce 489) ((T_LOGICAL_OR) reduce 489) ((T_LOGICAL_XOR) reduce 489) ((T_PIPE) reduce 489) ((T_POW) reduce 489) ((T_SL) reduce 489) ((T_SPACESHIP) reduce 489) ((T_SR) reduce 489) (("^") reduce 489) (("|") reduce 489)) 31195 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 494) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 494) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 31196 ((("%") reduce 490) (("*") reduce 490) (("+") reduce 490) (("-") reduce 490) ((".") reduce 490) (("/") reduce 490) ((";") reduce 490) (("<") reduce 490) ((">") reduce 490) (("?") reduce 490) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 490) ((T_BOOLEAN_AND) reduce 490) ((T_BOOLEAN_OR) reduce 490) ((T_COALESCE) reduce 490) ((T_DOUBLE_ARROW) reduce 490) ((T_INSTANCEOF) reduce 490) ((T_IS_EQUAL) reduce 490) ((T_IS_GREATER_OR_EQUAL) reduce 490) ((T_IS_IDENTICAL) reduce 490) ((T_IS_NOT_EQUAL) reduce 490) ((T_IS_NOT_IDENTICAL) reduce 490) ((T_IS_SMALLER_OR_EQUAL) reduce 490) ((T_LOGICAL_AND) reduce 490) ((T_LOGICAL_OR) reduce 490) ((T_LOGICAL_XOR) reduce 490) ((T_PIPE) reduce 490) ((T_POW) reduce 490) ((T_SL) reduce 490) ((T_SPACESHIP) reduce 490) ((T_SR) reduce 490) (("^") reduce 490) (("|") reduce 490)) 31198 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 481) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 31200 ((("%") reduce 541) (("(") reduce 579) (("*") reduce 541) (("+") reduce 541) (("-") reduce 541) ((".") reduce 541) (("/") reduce 541) ((";") reduce 541) (("<") reduce 541) ((">") reduce 541) (("?") reduce 541) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 541) ((T_BOOLEAN_AND) reduce 541) ((T_BOOLEAN_OR) reduce 541) ((T_COALESCE) reduce 541) ((T_DOUBLE_ARROW) reduce 541) ((T_INSTANCEOF) reduce 541) ((T_IS_EQUAL) reduce 541) ((T_IS_GREATER_OR_EQUAL) reduce 541) ((T_IS_IDENTICAL) reduce 541) ((T_IS_NOT_EQUAL) reduce 541) ((T_IS_NOT_IDENTICAL) reduce 541) ((T_IS_SMALLER_OR_EQUAL) reduce 541) ((T_LOGICAL_AND) reduce 541) ((T_LOGICAL_OR) reduce 541) ((T_LOGICAL_XOR) reduce 541) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 541) ((T_OBJECT_OPERATOR) reduce 541) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 541) ((T_PIPE) reduce 541) ((T_POW) reduce 541) ((T_SL) reduce 541) ((T_SPACESHIP) reduce 541) ((T_SR) reduce 541) (("[") reduce 541) (("^") reduce 541) (("|") reduce 541)) 31202 ((("%") reduce 571) (("(") reduce 581) (("*") reduce 571) (("+") reduce 571) (("-") reduce 571) ((".") reduce 571) (("/") reduce 571) ((";") reduce 571) (("<") reduce 571) (("=") reduce 571) ((">") reduce 571) (("?") reduce 571) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 571) ((T_AND_EQUAL) reduce 571) ((T_BOOLEAN_AND) reduce 571) ((T_BOOLEAN_OR) reduce 571) ((T_COALESCE) reduce 571) ((T_COALESCE_EQUAL) reduce 571) ((T_CONCAT_EQUAL) reduce 571) ((T_DEC) reduce 571) ((T_DIV_EQUAL) reduce 571) ((T_DOUBLE_ARROW) reduce 571) ((T_INC) reduce 571) ((T_INSTANCEOF) reduce 571) ((T_IS_EQUAL) reduce 571) ((T_IS_GREATER_OR_EQUAL) reduce 571) ((T_IS_IDENTICAL) reduce 571) ((T_IS_NOT_EQUAL) reduce 571) ((T_IS_NOT_IDENTICAL) reduce 571) ((T_IS_SMALLER_OR_EQUAL) reduce 571) ((T_LOGICAL_AND) reduce 571) ((T_LOGICAL_OR) reduce 571) ((T_LOGICAL_XOR) reduce 571) ((T_MINUS_EQUAL) reduce 571) ((T_MOD_EQUAL) reduce 571) ((T_MUL_EQUAL) reduce 571) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 571) ((T_OBJECT_OPERATOR) reduce 571) ((T_OR_EQUAL) reduce 571) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 571) ((T_PIPE) reduce 571) ((T_PLUS_EQUAL) reduce 571) ((T_POW) reduce 571) ((T_POW_EQUAL) reduce 571) ((T_SL) reduce 571) ((T_SL_EQUAL) reduce 571) ((T_SPACESHIP) reduce 571) ((T_SR) reduce 571) ((T_SR_EQUAL) reduce 571) ((T_XOR_EQUAL) reduce 571) (("[") reduce 571) (("^") reduce 571) (("|") reduce 571)) 31205 ((("%") reduce 543) (("(") reduce 580) (("*") reduce 543) (("+") reduce 543) (("-") reduce 543) ((".") reduce 543) (("/") reduce 543) ((";") reduce 543) (("<") reduce 543) ((">") reduce 543) (("?") reduce 543) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 543) ((T_BOOLEAN_AND) reduce 543) ((T_BOOLEAN_OR) reduce 543) ((T_COALESCE) reduce 543) ((T_DOUBLE_ARROW) reduce 543) ((T_INSTANCEOF) reduce 543) ((T_IS_EQUAL) reduce 543) ((T_IS_GREATER_OR_EQUAL) reduce 543) ((T_IS_IDENTICAL) reduce 543) ((T_IS_NOT_EQUAL) reduce 543) ((T_IS_NOT_IDENTICAL) reduce 543) ((T_IS_SMALLER_OR_EQUAL) reduce 543) ((T_LOGICAL_AND) reduce 543) ((T_LOGICAL_OR) reduce 543) ((T_LOGICAL_XOR) reduce 543) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 543) ((T_OBJECT_OPERATOR) reduce 543) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 543) ((T_PIPE) reduce 543) ((T_POW) reduce 543) ((T_SL) reduce 543) ((T_SPACESHIP) reduce 543) ((T_SR) reduce 543) (("[") reduce 543) (("^") reduce 543) (("|") reduce 543)) 31206 ((("%") reduce 506) (("(") reduce 506) (("*") reduce 506) (("+") reduce 506) (("-") reduce 506) ((".") reduce 506) (("/") reduce 506) ((";") reduce 506) (("<") reduce 506) (("=") reduce 506) ((">") reduce 506) (("?") reduce 506) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 506) ((T_AND_EQUAL) reduce 506) ((T_BOOLEAN_AND) reduce 506) ((T_BOOLEAN_OR) reduce 506) ((T_COALESCE) reduce 506) ((T_COALESCE_EQUAL) reduce 506) ((T_CONCAT_EQUAL) reduce 506) ((T_DEC) reduce 506) ((T_DIV_EQUAL) reduce 506) ((T_DOUBLE_ARROW) reduce 506) ((T_INC) reduce 506) ((T_INSTANCEOF) reduce 506) ((T_IS_EQUAL) reduce 506) ((T_IS_GREATER_OR_EQUAL) reduce 506) ((T_IS_IDENTICAL) reduce 506) ((T_IS_NOT_EQUAL) reduce 506) ((T_IS_NOT_IDENTICAL) reduce 506) ((T_IS_SMALLER_OR_EQUAL) reduce 506) ((T_LOGICAL_AND) reduce 506) ((T_LOGICAL_OR) reduce 506) ((T_LOGICAL_XOR) reduce 506) ((T_MINUS_EQUAL) reduce 506) ((T_MOD_EQUAL) reduce 506) ((T_MUL_EQUAL) reduce 506) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 506) ((T_OBJECT_OPERATOR) reduce 506) ((T_OR_EQUAL) reduce 506) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 506) ((T_PIPE) reduce 506) ((T_PLUS_EQUAL) reduce 506) ((T_POW) reduce 506) ((T_POW_EQUAL) reduce 506) ((T_SL) reduce 506) ((T_SL_EQUAL) reduce 506) ((T_SPACESHIP) reduce 506) ((T_SR) reduce 506) ((T_SR_EQUAL) reduce 506) ((T_XOR_EQUAL) reduce 506) (("[") reduce 506) (("^") reduce 506) (("|") reduce 506)) 31207 ((("%") reduce 508) (("(") reduce 508) (("*") reduce 508) (("+") reduce 508) (("-") reduce 508) ((".") reduce 508) (("/") reduce 508) ((";") reduce 508) (("<") reduce 508) (("=") reduce 508) ((">") reduce 508) (("?") reduce 508) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 508) ((T_AND_EQUAL) reduce 508) ((T_BOOLEAN_AND) reduce 508) ((T_BOOLEAN_OR) reduce 508) ((T_COALESCE) reduce 508) ((T_COALESCE_EQUAL) reduce 508) ((T_CONCAT_EQUAL) reduce 508) ((T_DEC) reduce 508) ((T_DIV_EQUAL) reduce 508) ((T_DOUBLE_ARROW) reduce 508) ((T_INC) reduce 508) ((T_INSTANCEOF) reduce 508) ((T_IS_EQUAL) reduce 508) ((T_IS_GREATER_OR_EQUAL) reduce 508) ((T_IS_IDENTICAL) reduce 508) ((T_IS_NOT_EQUAL) reduce 508) ((T_IS_NOT_IDENTICAL) reduce 508) ((T_IS_SMALLER_OR_EQUAL) reduce 508) ((T_LOGICAL_AND) reduce 508) ((T_LOGICAL_OR) reduce 508) ((T_LOGICAL_XOR) reduce 508) ((T_MINUS_EQUAL) reduce 508) ((T_MOD_EQUAL) reduce 508) ((T_MUL_EQUAL) reduce 508) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 508) ((T_OBJECT_OPERATOR) reduce 508) ((T_OR_EQUAL) reduce 508) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 508) ((T_PIPE) reduce 508) ((T_PLUS_EQUAL) reduce 508) ((T_POW) reduce 508) ((T_POW_EQUAL) reduce 508) ((T_SL) reduce 508) ((T_SL_EQUAL) reduce 508) ((T_SPACESHIP) reduce 508) ((T_SR) reduce 508) ((T_SR_EQUAL) reduce 508) ((T_XOR_EQUAL) reduce 508) (("[") reduce 508) (("^") reduce 508) (("|") reduce 508)) 31209 ((("%") reduce 485) (("*") reduce 485) (("+") reduce 485) (("-") reduce 485) ((".") reduce 485) (("/") reduce 485) ((";") reduce 485) (("<") reduce 485) ((">") reduce 485) (("?") reduce 485) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 485) ((T_BOOLEAN_AND) reduce 485) ((T_BOOLEAN_OR) reduce 485) ((T_COALESCE) reduce 485) ((T_DOUBLE_ARROW) reduce 485) ((T_INSTANCEOF) reduce 485) ((T_IS_EQUAL) reduce 485) ((T_IS_GREATER_OR_EQUAL) reduce 485) ((T_IS_IDENTICAL) reduce 485) ((T_IS_NOT_EQUAL) reduce 485) ((T_IS_NOT_IDENTICAL) reduce 485) ((T_IS_SMALLER_OR_EQUAL) reduce 485) ((T_LOGICAL_AND) reduce 485) ((T_LOGICAL_OR) reduce 485) ((T_LOGICAL_XOR) reduce 485) ((T_PIPE) reduce 485) ((T_POW) reduce 485) ((T_SL) reduce 485) ((T_SPACESHIP) reduce 485) ((T_SR) reduce 485) (("^") reduce 485) (("|") reduce 485)) 31210 ((("%") reduce 487) (("*") reduce 487) (("+") reduce 487) (("-") reduce 487) ((".") reduce 487) (("/") reduce 487) ((";") reduce 487) (("<") reduce 487) ((">") reduce 487) (("?") reduce 487) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 487) ((T_BOOLEAN_AND) reduce 487) ((T_BOOLEAN_OR) reduce 487) ((T_COALESCE) reduce 487) ((T_DOUBLE_ARROW) reduce 487) ((T_INSTANCEOF) reduce 487) ((T_IS_EQUAL) reduce 487) ((T_IS_GREATER_OR_EQUAL) reduce 487) ((T_IS_IDENTICAL) reduce 487) ((T_IS_NOT_EQUAL) reduce 487) ((T_IS_NOT_IDENTICAL) reduce 487) ((T_IS_SMALLER_OR_EQUAL) reduce 487) ((T_LOGICAL_AND) reduce 487) ((T_LOGICAL_OR) reduce 487) ((T_LOGICAL_XOR) reduce 487) ((T_PIPE) reduce 487) ((T_POW) reduce 487) ((T_SL) reduce 487) ((T_SPACESHIP) reduce 487) ((T_SR) reduce 487) (("^") reduce 487) (("|") reduce 487)) 31215 ((("%") reduce 560) (("(") reduce 560) (("*") reduce 560) (("+") reduce 560) (("-") reduce 560) ((".") reduce 560) (("/") reduce 560) ((";") reduce 560) (("<") reduce 560) (("=") reduce 560) ((">") reduce 560) (("?") reduce 560) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 560) ((T_AND_EQUAL) reduce 560) ((T_BOOLEAN_AND) reduce 560) ((T_BOOLEAN_OR) reduce 560) ((T_COALESCE) reduce 560) ((T_COALESCE_EQUAL) reduce 560) ((T_CONCAT_EQUAL) reduce 560) ((T_DEC) reduce 560) ((T_DIV_EQUAL) reduce 560) ((T_DOUBLE_ARROW) reduce 560) ((T_INC) reduce 560) ((T_INSTANCEOF) reduce 560) ((T_IS_EQUAL) reduce 560) ((T_IS_GREATER_OR_EQUAL) reduce 560) ((T_IS_IDENTICAL) reduce 560) ((T_IS_NOT_EQUAL) reduce 560) ((T_IS_NOT_IDENTICAL) reduce 560) ((T_IS_SMALLER_OR_EQUAL) reduce 560) ((T_LOGICAL_AND) reduce 560) ((T_LOGICAL_OR) reduce 560) ((T_LOGICAL_XOR) reduce 560) ((T_MINUS_EQUAL) reduce 560) ((T_MOD_EQUAL) reduce 560) ((T_MUL_EQUAL) reduce 560) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 560) ((T_OBJECT_OPERATOR) reduce 560) ((T_OR_EQUAL) reduce 560) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 560) ((T_PIPE) reduce 560) ((T_PLUS_EQUAL) reduce 560) ((T_POW) reduce 560) ((T_POW_EQUAL) reduce 560) ((T_SL) reduce 560) ((T_SL_EQUAL) reduce 560) ((T_SPACESHIP) reduce 560) ((T_SR) reduce 560) ((T_SR_EQUAL) reduce 560) ((T_XOR_EQUAL) reduce 560) (("[") reduce 560) (("^") reduce 560) (("|") reduce 560)) 31216 ((("%") reduce 582) (("(") reduce 582) (("*") reduce 582) (("+") reduce 582) (("-") reduce 582) ((".") reduce 582) (("/") reduce 582) ((";") reduce 582) (("<") reduce 582) (("=") reduce 582) ((">") reduce 582) (("?") reduce 582) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 582) ((T_AND_EQUAL) reduce 582) ((T_BOOLEAN_AND) reduce 582) ((T_BOOLEAN_OR) reduce 582) ((T_COALESCE) reduce 582) ((T_COALESCE_EQUAL) reduce 582) ((T_CONCAT_EQUAL) reduce 582) ((T_DEC) reduce 582) ((T_DIV_EQUAL) reduce 582) ((T_DOUBLE_ARROW) reduce 582) ((T_INC) reduce 582) ((T_INSTANCEOF) reduce 582) ((T_IS_EQUAL) reduce 582) ((T_IS_GREATER_OR_EQUAL) reduce 582) ((T_IS_IDENTICAL) reduce 582) ((T_IS_NOT_EQUAL) reduce 582) ((T_IS_NOT_IDENTICAL) reduce 582) ((T_IS_SMALLER_OR_EQUAL) reduce 582) ((T_LOGICAL_AND) reduce 582) ((T_LOGICAL_OR) reduce 582) ((T_LOGICAL_XOR) reduce 582) ((T_MINUS_EQUAL) reduce 582) ((T_MOD_EQUAL) reduce 582) ((T_MUL_EQUAL) reduce 582) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 582) ((T_OBJECT_OPERATOR) reduce 582) ((T_OR_EQUAL) reduce 582) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 582) ((T_PIPE) reduce 582) ((T_PLUS_EQUAL) reduce 582) ((T_POW) reduce 582) ((T_POW_EQUAL) reduce 582) ((T_SL) reduce 582) ((T_SL_EQUAL) reduce 582) ((T_SPACESHIP) reduce 582) ((T_SR) reduce 582) ((T_SR_EQUAL) reduce 582) ((T_XOR_EQUAL) reduce 582) (("[") reduce 582) (("^") reduce 582) (("|") reduce 582)) 31217 ((("%") reduce 566) (("(") shift) (("*") reduce 566) (("+") reduce 566) (("-") reduce 566) ((".") reduce 566) (("/") reduce 566) ((";") reduce 566) (("<") reduce 566) (("=") reduce 566) ((">") reduce 566) (("?") reduce 566) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 566) ((T_AND_EQUAL) reduce 566) ((T_BOOLEAN_AND) reduce 566) ((T_BOOLEAN_OR) reduce 566) ((T_COALESCE) reduce 566) ((T_COALESCE_EQUAL) reduce 566) ((T_CONCAT_EQUAL) reduce 566) ((T_DEC) reduce 566) ((T_DIV_EQUAL) reduce 566) ((T_DOUBLE_ARROW) reduce 566) ((T_INC) reduce 566) ((T_INSTANCEOF) reduce 566) ((T_IS_EQUAL) reduce 566) ((T_IS_GREATER_OR_EQUAL) reduce 566) ((T_IS_IDENTICAL) reduce 566) ((T_IS_NOT_EQUAL) reduce 566) ((T_IS_NOT_IDENTICAL) reduce 566) ((T_IS_SMALLER_OR_EQUAL) reduce 566) ((T_LOGICAL_AND) reduce 566) ((T_LOGICAL_OR) reduce 566) ((T_LOGICAL_XOR) reduce 566) ((T_MINUS_EQUAL) reduce 566) ((T_MOD_EQUAL) reduce 566) ((T_MUL_EQUAL) reduce 566) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 566) ((T_OBJECT_OPERATOR) reduce 566) ((T_OR_EQUAL) reduce 566) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 566) ((T_PIPE) reduce 566) ((T_PLUS_EQUAL) reduce 566) ((T_POW) reduce 566) ((T_POW_EQUAL) reduce 566) ((T_SL) reduce 566) ((T_SL_EQUAL) reduce 566) ((T_SPACESHIP) reduce 566) ((T_SR) reduce 566) ((T_SR_EQUAL) reduce 566) ((T_XOR_EQUAL) reduce 566) (("[") reduce 566) (("^") reduce 566) (("|") reduce 566)) 31218 ((("%") reduce 584) (("(") reduce 584) (("*") reduce 584) (("+") reduce 584) (("-") reduce 584) ((".") reduce 584) (("/") reduce 584) ((";") reduce 584) (("<") reduce 584) (("=") reduce 584) ((">") reduce 584) (("?") reduce 584) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 584) ((T_AND_EQUAL) reduce 584) ((T_BOOLEAN_AND) reduce 584) ((T_BOOLEAN_OR) reduce 584) ((T_COALESCE) reduce 584) ((T_COALESCE_EQUAL) reduce 584) ((T_CONCAT_EQUAL) reduce 584) ((T_DEC) reduce 584) ((T_DIV_EQUAL) reduce 584) ((T_DOUBLE_ARROW) reduce 584) ((T_INC) reduce 584) ((T_INSTANCEOF) reduce 584) ((T_IS_EQUAL) reduce 584) ((T_IS_GREATER_OR_EQUAL) reduce 584) ((T_IS_IDENTICAL) reduce 584) ((T_IS_NOT_EQUAL) reduce 584) ((T_IS_NOT_IDENTICAL) reduce 584) ((T_IS_SMALLER_OR_EQUAL) reduce 584) ((T_LOGICAL_AND) reduce 584) ((T_LOGICAL_OR) reduce 584) ((T_LOGICAL_XOR) reduce 584) ((T_MINUS_EQUAL) reduce 584) ((T_MOD_EQUAL) reduce 584) ((T_MUL_EQUAL) reduce 584) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 584) ((T_OBJECT_OPERATOR) reduce 584) ((T_OR_EQUAL) reduce 584) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 584) ((T_PIPE) reduce 584) ((T_PLUS_EQUAL) reduce 584) ((T_POW) reduce 584) ((T_POW_EQUAL) reduce 584) ((T_SL) reduce 584) ((T_SL_EQUAL) reduce 584) ((T_SPACESHIP) reduce 584) ((T_SR) reduce 584) ((T_SR_EQUAL) reduce 584) ((T_XOR_EQUAL) reduce 584) (("[") reduce 584) (("^") reduce 584) (("|") reduce 584)) 31221 ((("%") reduce 583) (("(") reduce 583) (("*") reduce 583) (("+") reduce 583) (("-") reduce 583) ((".") reduce 583) (("/") reduce 583) ((";") reduce 583) (("<") reduce 583) (("=") reduce 583) ((">") reduce 583) (("?") reduce 583) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 583) ((T_AND_EQUAL) reduce 583) ((T_BOOLEAN_AND) reduce 583) ((T_BOOLEAN_OR) reduce 583) ((T_COALESCE) reduce 583) ((T_COALESCE_EQUAL) reduce 583) ((T_CONCAT_EQUAL) reduce 583) ((T_DEC) reduce 583) ((T_DIV_EQUAL) reduce 583) ((T_DOUBLE_ARROW) reduce 583) ((T_INC) reduce 583) ((T_INSTANCEOF) reduce 583) ((T_IS_EQUAL) reduce 583) ((T_IS_GREATER_OR_EQUAL) reduce 583) ((T_IS_IDENTICAL) reduce 583) ((T_IS_NOT_EQUAL) reduce 583) ((T_IS_NOT_IDENTICAL) reduce 583) ((T_IS_SMALLER_OR_EQUAL) reduce 583) ((T_LOGICAL_AND) reduce 583) ((T_LOGICAL_OR) reduce 583) ((T_LOGICAL_XOR) reduce 583) ((T_MINUS_EQUAL) reduce 583) ((T_MOD_EQUAL) reduce 583) ((T_MUL_EQUAL) reduce 583) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 583) ((T_OBJECT_OPERATOR) reduce 583) ((T_OR_EQUAL) reduce 583) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 583) ((T_PIPE) reduce 583) ((T_PLUS_EQUAL) reduce 583) ((T_POW) reduce 583) ((T_POW_EQUAL) reduce 583) ((T_SL) reduce 583) ((T_SL_EQUAL) reduce 583) ((T_SPACESHIP) reduce 583) ((T_SR) reduce 583) ((T_SR_EQUAL) reduce 583) ((T_XOR_EQUAL) reduce 583) (("[") reduce 583) (("^") reduce 583) (("|") reduce 583)) 31222 ((("%") reduce 561) (("(") reduce 561) (("*") reduce 561) (("+") reduce 561) (("-") reduce 561) ((".") reduce 561) (("/") reduce 561) ((";") reduce 561) (("<") reduce 561) (("=") reduce 561) ((">") reduce 561) (("?") reduce 561) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 561) ((T_AND_EQUAL) reduce 561) ((T_BOOLEAN_AND) reduce 561) ((T_BOOLEAN_OR) reduce 561) ((T_COALESCE) reduce 561) ((T_COALESCE_EQUAL) reduce 561) ((T_CONCAT_EQUAL) reduce 561) ((T_DEC) reduce 561) ((T_DIV_EQUAL) reduce 561) ((T_DOUBLE_ARROW) reduce 561) ((T_INC) reduce 561) ((T_INSTANCEOF) reduce 561) ((T_IS_EQUAL) reduce 561) ((T_IS_GREATER_OR_EQUAL) reduce 561) ((T_IS_IDENTICAL) reduce 561) ((T_IS_NOT_EQUAL) reduce 561) ((T_IS_NOT_IDENTICAL) reduce 561) ((T_IS_SMALLER_OR_EQUAL) reduce 561) ((T_LOGICAL_AND) reduce 561) ((T_LOGICAL_OR) reduce 561) ((T_LOGICAL_XOR) reduce 561) ((T_MINUS_EQUAL) reduce 561) ((T_MOD_EQUAL) reduce 561) ((T_MUL_EQUAL) reduce 561) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 561) ((T_OBJECT_OPERATOR) reduce 561) ((T_OR_EQUAL) reduce 561) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 561) ((T_PIPE) reduce 561) ((T_PLUS_EQUAL) reduce 561) ((T_POW) reduce 561) ((T_POW_EQUAL) reduce 561) ((T_SL) reduce 561) ((T_SL_EQUAL) reduce 561) ((T_SPACESHIP) reduce 561) ((T_SR) reduce 561) ((T_SR_EQUAL) reduce 561) ((T_XOR_EQUAL) reduce 561) (("[") reduce 561) (("^") reduce 561) (("|") reduce 561)) 31223 ((("%") reduce 567) (("(") shift) (("*") reduce 567) (("+") reduce 567) (("-") reduce 567) ((".") reduce 567) (("/") reduce 567) ((";") reduce 567) (("<") reduce 567) (("=") reduce 567) ((">") reduce 567) (("?") reduce 567) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 567) ((T_AND_EQUAL) reduce 567) ((T_BOOLEAN_AND) reduce 567) ((T_BOOLEAN_OR) reduce 567) ((T_COALESCE) reduce 567) ((T_COALESCE_EQUAL) reduce 567) ((T_CONCAT_EQUAL) reduce 567) ((T_DEC) reduce 567) ((T_DIV_EQUAL) reduce 567) ((T_DOUBLE_ARROW) reduce 567) ((T_INC) reduce 567) ((T_INSTANCEOF) reduce 567) ((T_IS_EQUAL) reduce 567) ((T_IS_GREATER_OR_EQUAL) reduce 567) ((T_IS_IDENTICAL) reduce 567) ((T_IS_NOT_EQUAL) reduce 567) ((T_IS_NOT_IDENTICAL) reduce 567) ((T_IS_SMALLER_OR_EQUAL) reduce 567) ((T_LOGICAL_AND) reduce 567) ((T_LOGICAL_OR) reduce 567) ((T_LOGICAL_XOR) reduce 567) ((T_MINUS_EQUAL) reduce 567) ((T_MOD_EQUAL) reduce 567) ((T_MUL_EQUAL) reduce 567) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 567) ((T_OBJECT_OPERATOR) reduce 567) ((T_OR_EQUAL) reduce 567) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 567) ((T_PIPE) reduce 567) ((T_PLUS_EQUAL) reduce 567) ((T_POW) reduce 567) ((T_POW_EQUAL) reduce 567) ((T_SL) reduce 567) ((T_SL_EQUAL) reduce 567) ((T_SPACESHIP) reduce 567) ((T_SR) reduce 567) ((T_SR_EQUAL) reduce 567) ((T_XOR_EQUAL) reduce 567) (("[") reduce 567) (("^") reduce 567) (("|") reduce 567)) 31224 ((("%") reduce 562) (("(") reduce 562) (("*") reduce 562) (("+") reduce 562) (("-") reduce 562) ((".") reduce 562) (("/") reduce 562) ((";") reduce 562) (("<") reduce 562) (("=") reduce 562) ((">") reduce 562) (("?") reduce 562) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 562) ((T_AND_EQUAL) reduce 562) ((T_BOOLEAN_AND) reduce 562) ((T_BOOLEAN_OR) reduce 562) ((T_COALESCE) reduce 562) ((T_COALESCE_EQUAL) reduce 562) ((T_CONCAT_EQUAL) reduce 562) ((T_DEC) reduce 562) ((T_DIV_EQUAL) reduce 562) ((T_DOUBLE_ARROW) reduce 562) ((T_INC) reduce 562) ((T_INSTANCEOF) reduce 562) ((T_IS_EQUAL) reduce 562) ((T_IS_GREATER_OR_EQUAL) reduce 562) ((T_IS_IDENTICAL) reduce 562) ((T_IS_NOT_EQUAL) reduce 562) ((T_IS_NOT_IDENTICAL) reduce 562) ((T_IS_SMALLER_OR_EQUAL) reduce 562) ((T_LOGICAL_AND) reduce 562) ((T_LOGICAL_OR) reduce 562) ((T_LOGICAL_XOR) reduce 562) ((T_MINUS_EQUAL) reduce 562) ((T_MOD_EQUAL) reduce 562) ((T_MUL_EQUAL) reduce 562) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 562) ((T_OBJECT_OPERATOR) reduce 562) ((T_OR_EQUAL) reduce 562) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 562) ((T_PIPE) reduce 562) ((T_PLUS_EQUAL) reduce 562) ((T_POW) reduce 562) ((T_POW_EQUAL) reduce 562) ((T_SL) reduce 562) ((T_SL_EQUAL) reduce 562) ((T_SPACESHIP) reduce 562) ((T_SR) reduce 562) ((T_SR_EQUAL) reduce 562) ((T_XOR_EQUAL) reduce 562) (("[") reduce 562) (("^") reduce 562) (("|") reduce 562)) 31226 ((("%") reduce 477) (("*") reduce 477) (("+") reduce 477) (("-") reduce 477) ((".") reduce 477) (("/") reduce 477) ((";") reduce 477) (("<") reduce 477) ((">") reduce 477) (("?") reduce 477) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 477) ((T_BOOLEAN_AND) reduce 477) ((T_BOOLEAN_OR) reduce 477) ((T_COALESCE) reduce 477) ((T_DOUBLE_ARROW) reduce 477) ((T_INSTANCEOF) reduce 477) ((T_IS_EQUAL) reduce 477) ((T_IS_GREATER_OR_EQUAL) reduce 477) ((T_IS_IDENTICAL) reduce 477) ((T_IS_NOT_EQUAL) reduce 477) ((T_IS_NOT_IDENTICAL) reduce 477) ((T_IS_SMALLER_OR_EQUAL) reduce 477) ((T_LOGICAL_AND) reduce 477) ((T_LOGICAL_OR) reduce 477) ((T_LOGICAL_XOR) reduce 477) ((T_PIPE) reduce 477) ((T_POW) reduce 477) ((T_SL) reduce 477) ((T_SPACESHIP) reduce 477) ((T_SR) reduce 477) (("^") reduce 477) (("|") reduce 477)) 31228 ((("%") reduce 520) (("(") reduce 520) (("*") reduce 520) (("+") reduce 520) (("-") reduce 520) ((".") reduce 520) (("/") reduce 520) ((";") reduce 520) (("<") reduce 520) (("=") shift) ((">") reduce 520) (("?") reduce 520) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 520) ((T_BOOLEAN_AND) reduce 520) ((T_BOOLEAN_OR) reduce 520) ((T_COALESCE) reduce 520) ((T_DOUBLE_ARROW) reduce 520) ((T_INSTANCEOF) reduce 520) ((T_IS_EQUAL) reduce 520) ((T_IS_GREATER_OR_EQUAL) reduce 520) ((T_IS_IDENTICAL) reduce 520) ((T_IS_NOT_EQUAL) reduce 520) ((T_IS_NOT_IDENTICAL) reduce 520) ((T_IS_SMALLER_OR_EQUAL) reduce 520) ((T_LOGICAL_AND) reduce 520) ((T_LOGICAL_OR) reduce 520) ((T_LOGICAL_XOR) reduce 520) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 520) ((T_OBJECT_OPERATOR) reduce 520) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 520) ((T_PIPE) reduce 520) ((T_POW) reduce 520) ((T_SL) reduce 520) ((T_SPACESHIP) reduce 520) ((T_SR) reduce 520) (("[") reduce 520) (("^") reduce 520) (("|") reduce 520)) 31230 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 405) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 405) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 405) ((T_LOGICAL_OR) reduce 405) ((T_LOGICAL_XOR) reduce 405) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 31231 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 482) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 482) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 482) ((T_LOGICAL_OR) reduce 482) ((T_LOGICAL_XOR) reduce 482) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 31234 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 481) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 481) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 481) ((T_LOGICAL_OR) reduce 481) ((T_LOGICAL_XOR) reduce 481) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 31235 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) ((";") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_DOUBLE_ARROW) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_PIPE) reduce 473) ((T_POW) shift) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("|") reduce 473)) 31236 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 483) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 483) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 31237 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) ((";") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_DOUBLE_ARROW) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_PIPE) reduce 469) ((T_POW) shift) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 31238 ((("%") reduce 486) (("*") reduce 486) (("+") reduce 486) (("-") reduce 486) ((".") reduce 486) (("/") reduce 486) ((";") reduce 486) (("<") reduce 486) ((">") reduce 486) (("?") reduce 486) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 486) ((T_BOOLEAN_AND) reduce 486) ((T_BOOLEAN_OR) reduce 486) ((T_COALESCE) reduce 486) ((T_DOUBLE_ARROW) reduce 486) ((T_INSTANCEOF) reduce 486) ((T_IS_EQUAL) reduce 486) ((T_IS_GREATER_OR_EQUAL) reduce 486) ((T_IS_IDENTICAL) reduce 486) ((T_IS_NOT_EQUAL) reduce 486) ((T_IS_NOT_IDENTICAL) reduce 486) ((T_IS_SMALLER_OR_EQUAL) reduce 486) ((T_LOGICAL_AND) reduce 486) ((T_LOGICAL_OR) reduce 486) ((T_LOGICAL_XOR) reduce 486) ((T_PIPE) reduce 486) ((T_POW) reduce 486) ((T_SL) reduce 486) ((T_SPACESHIP) reduce 486) ((T_SR) reduce 486) (("^") reduce 486) (("|") reduce 486)) 31240 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) ((";") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_DOUBLE_ARROW) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_PIPE) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("^") reduce 526) (("|") reduce 526)) 31242 ((("%") reduce 527) (("*") reduce 527) (("+") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) ((";") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_DOUBLE_ARROW) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_PIPE) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("^") reduce 527) (("|") reduce 527)) 31243 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) ((";") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_DOUBLE_ARROW) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_PIPE) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("^") reduce 525) (("|") reduce 525)) 31244 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 619) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 619) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 31245 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 618) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 618) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 31246 ((("%") reduce 505) (("(") reduce 505) (("*") reduce 505) (("+") reduce 505) (("-") reduce 505) ((".") reduce 505) (("/") reduce 505) ((";") reduce 505) (("<") reduce 505) (("=") reduce 505) ((">") reduce 505) (("?") reduce 505) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 505) ((T_AND_EQUAL) reduce 505) ((T_BOOLEAN_AND) reduce 505) ((T_BOOLEAN_OR) reduce 505) ((T_COALESCE) reduce 505) ((T_COALESCE_EQUAL) reduce 505) ((T_CONCAT_EQUAL) reduce 505) ((T_DEC) reduce 505) ((T_DIV_EQUAL) reduce 505) ((T_DOUBLE_ARROW) reduce 505) ((T_INC) reduce 505) ((T_INSTANCEOF) reduce 505) ((T_IS_EQUAL) reduce 505) ((T_IS_GREATER_OR_EQUAL) reduce 505) ((T_IS_IDENTICAL) reduce 505) ((T_IS_NOT_EQUAL) reduce 505) ((T_IS_NOT_IDENTICAL) reduce 505) ((T_IS_SMALLER_OR_EQUAL) reduce 505) ((T_LOGICAL_AND) reduce 505) ((T_LOGICAL_OR) reduce 505) ((T_LOGICAL_XOR) reduce 505) ((T_MINUS_EQUAL) reduce 505) ((T_MOD_EQUAL) reduce 505) ((T_MUL_EQUAL) reduce 505) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 505) ((T_OBJECT_OPERATOR) reduce 505) ((T_OR_EQUAL) reduce 505) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 505) ((T_PIPE) reduce 505) ((T_PLUS_EQUAL) reduce 505) ((T_POW) reduce 505) ((T_POW_EQUAL) reduce 505) ((T_SL) reduce 505) ((T_SL_EQUAL) reduce 505) ((T_SPACESHIP) reduce 505) ((T_SR) reduce 505) ((T_SR_EQUAL) reduce 505) ((T_XOR_EQUAL) reduce 505) (("[") reduce 505) (("^") reduce 505) (("|") reduce 505)) 31247 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 478) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 478) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 31248 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) ((";") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_DOUBLE_ARROW) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_PIPE) reduce 471) ((T_POW) shift) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("|") reduce 471)) 31250 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_DOUBLE_ARROW) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 31251 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_DOUBLE_ARROW) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 31252 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_DOUBLE_ARROW) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 31253 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((";") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_DOUBLE_ARROW) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 31254 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_DOUBLE_ARROW) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 31255 ((("%") reduce 400) (("(") reduce 400) (("*") reduce 400) (("+") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) ((";") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_DOUBLE_ARROW) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 400) ((T_OBJECT_OPERATOR) reduce 400) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 400) ((T_PIPE) reduce 400) ((T_POW) reduce 400) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("[") reduce 400) (("^") reduce 400) (("|") reduce 400)) 31258 ((("%") reduce 511) (("(") reduce 511) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((";") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_DOUBLE_ARROW) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("|") reduce 511)) 31259 ((("%") reduce 402) (("(") shift) (("*") reduce 402) (("+") reduce 402) (("-") reduce 402) ((".") reduce 402) (("/") reduce 402) ((";") reduce 402) (("<") reduce 402) ((">") reduce 402) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_DOUBLE_ARROW) reduce 402) ((T_INSTANCEOF) reduce 402) ((T_IS_EQUAL) reduce 402) ((T_IS_GREATER_OR_EQUAL) reduce 402) ((T_IS_IDENTICAL) reduce 402) ((T_IS_NOT_EQUAL) reduce 402) ((T_IS_NOT_IDENTICAL) reduce 402) ((T_IS_SMALLER_OR_EQUAL) reduce 402) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_PIPE) reduce 402) ((T_POW) reduce 402) ((T_SL) reduce 402) ((T_SPACESHIP) reduce 402) ((T_SR) reduce 402) (("^") reduce 402) (("|") reduce 402)) 31260 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((";") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_DOUBLE_ARROW) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 31261 ((("%") reduce 512) (("(") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((";") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_DOUBLE_ARROW) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("|") reduce 512)) 31262 ((("%") reduce 573) (("(") reduce 573) (("*") reduce 573) (("+") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) ((";") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_DOUBLE_ARROW) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("|") reduce 573)) 31268 ((("%") reduce 574) (("(") reduce 574) (("*") reduce 574) (("+") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) ((";") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_DOUBLE_ARROW) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("|") reduce 574)) 31269 ((("%") reduce 578) (("(") reduce 578) (("*") reduce 578) (("+") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) ((";") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_DOUBLE_ARROW) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("|") reduce 578)) 31270 ((("%") reduce 575) (("(") reduce 575) (("*") reduce 575) (("+") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) ((";") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_DOUBLE_ARROW) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("|") reduce 575)) 31271 ((("%") reduce 576) (("(") reduce 576) (("*") reduce 576) (("+") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) ((";") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_DOUBLE_ARROW) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("|") reduce 576)) 31272 ((("%") reduce 399) (("(") reduce 399) (("*") reduce 399) (("+") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) ((";") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_DOUBLE_ARROW) reduce 399) ((T_INSTANCEOF) reduce 399) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 399) ((T_OBJECT_OPERATOR) reduce 399) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 399) ((T_PIPE) reduce 399) ((T_POW) reduce 399) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("[") reduce 399) (("^") reduce 399) (("|") reduce 399)) 31274 ((("%") reduce 577) (("(") reduce 577) (("*") reduce 577) (("+") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) ((";") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_DOUBLE_ARROW) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("|") reduce 577)) 31275 ((("%") reduce 401) (("(") reduce 401) (("*") reduce 401) (("+") reduce 401) (("-") reduce 401) ((".") reduce 401) (("/") reduce 401) ((";") reduce 401) (("<") reduce 401) ((">") reduce 401) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_DOUBLE_ARROW) reduce 401) ((T_INSTANCEOF) reduce 401) ((T_IS_EQUAL) reduce 401) ((T_IS_GREATER_OR_EQUAL) reduce 401) ((T_IS_IDENTICAL) reduce 401) ((T_IS_NOT_EQUAL) reduce 401) ((T_IS_NOT_IDENTICAL) reduce 401) ((T_IS_SMALLER_OR_EQUAL) reduce 401) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 401) ((T_OBJECT_OPERATOR) reduce 401) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 401) ((T_PIPE) reduce 401) ((T_POW) reduce 401) ((T_SL) reduce 401) ((T_SPACESHIP) reduce 401) ((T_SR) reduce 401) (("[") reduce 401) (("^") reduce 401) (("|") reduce 401)) 31283 ((("%") reduce 398) (("(") reduce 398) (("*") reduce 398) (("+") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) ((";") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_DOUBLE_ARROW) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 398) ((T_OBJECT_OPERATOR) reduce 398) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 398) ((T_PIPE) reduce 398) ((T_POW) reduce 398) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("[") reduce 398) (("^") reduce 398) (("|") reduce 398)) 31285 ((("%") reduce 513) (("(") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((";") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_DOUBLE_ARROW) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("|") reduce 513)) 31291 ((("%") reduce 233) (("*") reduce 233) (("+") reduce 233) (("-") reduce 233) ((".") reduce 233) (("/") reduce 233) ((";") reduce 233) (("<") reduce 233) ((">") reduce 233) (("?") reduce 233) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_BOOLEAN_AND) reduce 233) ((T_BOOLEAN_OR) reduce 233) ((T_COALESCE) reduce 233) ((T_DOUBLE_ARROW) reduce 233) ((T_INSTANCEOF) reduce 233) ((T_IS_EQUAL) reduce 233) ((T_IS_GREATER_OR_EQUAL) reduce 233) ((T_IS_IDENTICAL) reduce 233) ((T_IS_NOT_EQUAL) reduce 233) ((T_IS_NOT_IDENTICAL) reduce 233) ((T_IS_SMALLER_OR_EQUAL) reduce 233) ((T_LOGICAL_AND) reduce 233) ((T_LOGICAL_OR) reduce 233) ((T_LOGICAL_XOR) reduce 233) ((T_PIPE) reduce 233) ((T_POW) reduce 233) ((T_SL) reduce 233) ((T_SPACESHIP) reduce 233) ((T_SR) reduce 233) (("^") reduce 233) (("|") reduce 233)) 31296 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 404) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 404) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 31300 ((("%") reduce 613) (("*") reduce 613) (("+") reduce 613) (("-") reduce 613) ((".") reduce 613) (("/") reduce 613) ((";") reduce 613) (("<") reduce 613) ((">") reduce 613) (("?") reduce 613) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_BOOLEAN_AND) reduce 613) ((T_BOOLEAN_OR) reduce 613) ((T_COALESCE) reduce 613) ((T_DOUBLE_ARROW) reduce 613) ((T_INSTANCEOF) reduce 613) ((T_IS_EQUAL) reduce 613) ((T_IS_GREATER_OR_EQUAL) reduce 613) ((T_IS_IDENTICAL) reduce 613) ((T_IS_NOT_EQUAL) reduce 613) ((T_IS_NOT_IDENTICAL) reduce 613) ((T_IS_SMALLER_OR_EQUAL) reduce 613) ((T_LOGICAL_AND) reduce 613) ((T_LOGICAL_OR) reduce 613) ((T_LOGICAL_XOR) reduce 613) ((T_PIPE) reduce 613) ((T_POW) reduce 613) ((T_SL) reduce 613) ((T_SPACESHIP) reduce 613) ((T_SR) reduce 613) (("^") reduce 613) (("|") reduce 613)) 31301 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) ((";") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_DOUBLE_ARROW) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_PIPE) reduce 467) ((T_POW) shift) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("|") reduce 467)) 31302 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 616) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 616) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 31303 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 615) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_DOUBLE_ARROW) reduce 615) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 31304 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) ((";") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_DOUBLE_ARROW) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("[") reduce 548) (("^") reduce 424) (("|") reduce 424)) 31306 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) ((";") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_DOUBLE_ARROW) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_PIPE) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("^") reduce 518) (("|") reduce 518)) 31307 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) ((";") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_DOUBLE_ARROW) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_PIPE) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("|") reduce 474)) 31308 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((";") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_DOUBLE_ARROW) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("|") reduce 291)) 31312 ((("%") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) ((";") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_DOUBLE_ARROW) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("^") reduce 292) (("|") reduce 292)) 31313 ((("%") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) ((";") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_DOUBLE_ARROW) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("^") reduce 293) (("|") reduce 293)) 31316 ((("%") reduce 617) (("*") reduce 617) (("+") reduce 617) (("-") reduce 617) ((".") reduce 617) (("/") reduce 617) ((";") reduce 617) (("<") reduce 617) ((">") reduce 617) (("?") reduce 617) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_BOOLEAN_AND) reduce 617) ((T_BOOLEAN_OR) reduce 617) ((T_COALESCE) reduce 617) ((T_DOUBLE_ARROW) reduce 617) ((T_INSTANCEOF) reduce 617) ((T_IS_EQUAL) reduce 617) ((T_IS_GREATER_OR_EQUAL) reduce 617) ((T_IS_IDENTICAL) reduce 617) ((T_IS_NOT_EQUAL) reduce 617) ((T_IS_NOT_IDENTICAL) reduce 617) ((T_IS_SMALLER_OR_EQUAL) reduce 617) ((T_LOGICAL_AND) reduce 617) ((T_LOGICAL_OR) reduce 617) ((T_LOGICAL_XOR) reduce 617) ((T_PIPE) reduce 617) ((T_POW) reduce 617) ((T_SL) reduce 617) ((T_SPACESHIP) reduce 617) ((T_SR) reduce 617) (("^") reduce 617) (("|") reduce 617)) 31319 ((("%") reduce 614) (("*") reduce 614) (("+") reduce 614) (("-") reduce 614) ((".") reduce 614) (("/") reduce 614) ((";") reduce 614) (("<") reduce 614) ((">") reduce 614) (("?") reduce 614) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_BOOLEAN_AND) reduce 614) ((T_BOOLEAN_OR) reduce 614) ((T_COALESCE) reduce 614) ((T_DOUBLE_ARROW) reduce 614) ((T_INSTANCEOF) reduce 614) ((T_IS_EQUAL) reduce 614) ((T_IS_GREATER_OR_EQUAL) reduce 614) ((T_IS_IDENTICAL) reduce 614) ((T_IS_NOT_EQUAL) reduce 614) ((T_IS_NOT_IDENTICAL) reduce 614) ((T_IS_SMALLER_OR_EQUAL) reduce 614) ((T_LOGICAL_AND) reduce 614) ((T_LOGICAL_OR) reduce 614) ((T_LOGICAL_XOR) reduce 614) ((T_PIPE) reduce 614) ((T_POW) reduce 614) ((T_SL) reduce 614) ((T_SPACESHIP) reduce 614) ((T_SR) reduce 614) (("^") reduce 614) (("|") reduce 614)) 31320 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) ((";") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_DOUBLE_ARROW) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_PIPE) reduce 468) ((T_POW) shift) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 31321 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) ((";") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_DOUBLE_ARROW) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("[") reduce 548) (("^") reduce 426) (("|") reduce 426)) 31323 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) ((";") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_DOUBLE_ARROW) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_PIPE) reduce 408) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("^") reduce 408) (("|") reduce 408)) 31324 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) ((";") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_DOUBLE_ARROW) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_PIPE) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("|") reduce 409)) 31325 ((("%") reduce 296) (("*") reduce 296) (("+") reduce 296) (("-") reduce 296) ((".") reduce 296) (("/") reduce 296) ((";") reduce 296) (("<") reduce 296) ((">") reduce 296) (("?") reduce 296) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_BOOLEAN_AND) reduce 296) ((T_BOOLEAN_OR) reduce 296) ((T_COALESCE) reduce 296) ((T_DOUBLE_ARROW) reduce 296) ((T_INSTANCEOF) reduce 296) ((T_IS_EQUAL) reduce 296) ((T_IS_GREATER_OR_EQUAL) reduce 296) ((T_IS_IDENTICAL) reduce 296) ((T_IS_NOT_EQUAL) reduce 296) ((T_IS_NOT_IDENTICAL) reduce 296) ((T_IS_SMALLER_OR_EQUAL) reduce 296) ((T_LOGICAL_AND) reduce 296) ((T_LOGICAL_OR) reduce 296) ((T_LOGICAL_XOR) reduce 296) ((T_PIPE) reduce 296) ((T_POW) reduce 296) ((T_SL) reduce 296) ((T_SPACESHIP) reduce 296) ((T_SR) reduce 296) (("^") reduce 296) (("|") reduce 296)) 31330 ((("%") reduce 297) (("*") reduce 297) (("+") reduce 297) (("-") reduce 297) ((".") reduce 297) (("/") reduce 297) ((";") reduce 297) (("<") reduce 297) ((">") reduce 297) (("?") reduce 297) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_BOOLEAN_AND) reduce 297) ((T_BOOLEAN_OR) reduce 297) ((T_COALESCE) reduce 297) ((T_DOUBLE_ARROW) reduce 297) ((T_INSTANCEOF) reduce 297) ((T_IS_EQUAL) reduce 297) ((T_IS_GREATER_OR_EQUAL) reduce 297) ((T_IS_IDENTICAL) reduce 297) ((T_IS_NOT_EQUAL) reduce 297) ((T_IS_NOT_IDENTICAL) reduce 297) ((T_IS_SMALLER_OR_EQUAL) reduce 297) ((T_LOGICAL_AND) reduce 297) ((T_LOGICAL_OR) reduce 297) ((T_LOGICAL_XOR) reduce 297) ((T_PIPE) reduce 297) ((T_POW) reduce 297) ((T_SL) reduce 297) ((T_SPACESHIP) reduce 297) ((T_SR) reduce 297) (("^") reduce 297) (("|") reduce 297)) 31331 ((("%") reduce 460) (("(") reduce 556) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) ((";") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_DOUBLE_ARROW) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PIPE) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("[") reduce 549) (("^") reduce 460) (("|") reduce 460)) 31333 ((("%") reduce 298) (("*") reduce 298) (("+") reduce 298) (("-") reduce 298) ((".") reduce 298) (("/") reduce 298) ((";") reduce 298) (("<") reduce 298) ((">") reduce 298) (("?") reduce 298) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_BOOLEAN_AND) reduce 298) ((T_BOOLEAN_OR) reduce 298) ((T_COALESCE) reduce 298) ((T_DOUBLE_ARROW) reduce 298) ((T_INSTANCEOF) reduce 298) ((T_IS_EQUAL) reduce 298) ((T_IS_GREATER_OR_EQUAL) reduce 298) ((T_IS_IDENTICAL) reduce 298) ((T_IS_NOT_EQUAL) reduce 298) ((T_IS_NOT_IDENTICAL) reduce 298) ((T_IS_SMALLER_OR_EQUAL) reduce 298) ((T_LOGICAL_AND) reduce 298) ((T_LOGICAL_OR) reduce 298) ((T_LOGICAL_XOR) reduce 298) ((T_PIPE) reduce 298) ((T_POW) reduce 298) ((T_SL) reduce 298) ((T_SPACESHIP) reduce 298) ((T_SR) reduce 298) (("^") reduce 298) (("|") reduce 298)) 31334 ((("%") reduce 299) (("*") reduce 299) (("+") reduce 299) (("-") reduce 299) ((".") reduce 299) (("/") reduce 299) ((";") reduce 299) (("<") reduce 299) ((">") reduce 299) (("?") reduce 299) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_BOOLEAN_AND) reduce 299) ((T_BOOLEAN_OR) reduce 299) ((T_COALESCE) reduce 299) ((T_DOUBLE_ARROW) reduce 299) ((T_INSTANCEOF) reduce 299) ((T_IS_EQUAL) reduce 299) ((T_IS_GREATER_OR_EQUAL) reduce 299) ((T_IS_IDENTICAL) reduce 299) ((T_IS_NOT_EQUAL) reduce 299) ((T_IS_NOT_IDENTICAL) reduce 299) ((T_IS_SMALLER_OR_EQUAL) reduce 299) ((T_LOGICAL_AND) reduce 299) ((T_LOGICAL_OR) reduce 299) ((T_LOGICAL_XOR) reduce 299) ((T_PIPE) reduce 299) ((T_POW) reduce 299) ((T_SL) reduce 299) ((T_SPACESHIP) reduce 299) ((T_SR) reduce 299) (("^") reduce 299) (("|") reduce 299)) 31335 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) ((";") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_DOUBLE_ARROW) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_PIPE) reduce 472) ((T_POW) shift) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("|") reduce 472)) 31336 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) ((";") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_DOUBLE_ARROW) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_PIPE) reduce 470) ((T_POW) shift) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("|") reduce 470)) 31339 ((("%") reduce 519) (("(") reduce 519) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) ((";") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_DOUBLE_ARROW) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PIPE) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("|") reduce 519)) 31340 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) ((";") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_DOUBLE_ARROW) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_PIPE) reduce 475) ((T_POW) shift) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("|") reduce 475)) 31341 ((("%") reduce 446) (("*") reduce 446) (("+") reduce 446) (("-") reduce 446) ((".") reduce 446) (("/") reduce 446) ((";") reduce 446) (("<") reduce 446) ((">") reduce 446) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_BOOLEAN_AND) reduce 446) ((T_BOOLEAN_OR) reduce 446) ((T_COALESCE) reduce 446) ((T_DOUBLE_ARROW) reduce 446) ((T_INSTANCEOF) reduce 446) ((T_IS_EQUAL) reduce 446) ((T_IS_GREATER_OR_EQUAL) reduce 446) ((T_IS_IDENTICAL) reduce 446) ((T_IS_NOT_EQUAL) reduce 446) ((T_IS_NOT_IDENTICAL) reduce 446) ((T_IS_SMALLER_OR_EQUAL) reduce 446) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_PIPE) reduce 446) ((T_POW) shift) ((T_SL) reduce 446) ((T_SPACESHIP) reduce 446) ((T_SR) reduce 446) (("^") reduce 446) (("|") reduce 446)) 31342 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) ((";") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_DOUBLE_ARROW) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_PIPE) reduce 445) ((T_POW) shift) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("|") reduce 445)) 31344 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) ((";") reduce 570) (("<") reduce 570) (("=") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AND_EQUAL) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_COALESCE_EQUAL) reduce 570) ((T_CONCAT_EQUAL) reduce 570) ((T_DEC) reduce 570) ((T_DIV_EQUAL) reduce 570) ((T_DOUBLE_ARROW) reduce 570) ((T_INC) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_MINUS_EQUAL) reduce 570) ((T_MOD_EQUAL) reduce 570) ((T_MUL_EQUAL) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_OR_EQUAL) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_PLUS_EQUAL) reduce 570) ((T_POW) reduce 570) ((T_POW_EQUAL) reduce 570) ((T_SL) reduce 570) ((T_SL_EQUAL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) ((T_SR_EQUAL) reduce 570) ((T_XOR_EQUAL) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 31347 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) ((";") reduce 569) (("<") reduce 569) (("=") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AND_EQUAL) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_COALESCE_EQUAL) reduce 569) ((T_CONCAT_EQUAL) reduce 569) ((T_DEC) reduce 569) ((T_DIV_EQUAL) reduce 569) ((T_DOUBLE_ARROW) reduce 569) ((T_INC) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_MINUS_EQUAL) reduce 569) ((T_MOD_EQUAL) reduce 569) ((T_MUL_EQUAL) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_OR_EQUAL) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_PLUS_EQUAL) reduce 569) ((T_POW) reduce 569) ((T_POW_EQUAL) reduce 569) ((T_SL) reduce 569) ((T_SL_EQUAL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) ((T_SR_EQUAL) reduce 569) ((T_XOR_EQUAL) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 31349 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) ((";") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_DOUBLE_ARROW) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_PIPE) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("|") reduce 522)) 31350 ((("%") reduce 447) (("*") reduce 447) (("+") reduce 447) (("-") reduce 447) ((".") reduce 447) (("/") reduce 447) ((";") reduce 447) (("<") reduce 447) ((">") reduce 447) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_BOOLEAN_AND) reduce 447) ((T_BOOLEAN_OR) reduce 447) ((T_COALESCE) reduce 447) ((T_DOUBLE_ARROW) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 447) ((T_IS_GREATER_OR_EQUAL) reduce 447) ((T_IS_IDENTICAL) reduce 447) ((T_IS_NOT_EQUAL) reduce 447) ((T_IS_NOT_IDENTICAL) reduce 447) ((T_IS_SMALLER_OR_EQUAL) reduce 447) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_PIPE) reduce 447) ((T_POW) shift) ((T_SL) reduce 447) ((T_SPACESHIP) reduce 447) ((T_SR) reduce 447) (("^") reduce 447) (("|") reduce 447)) 31351 ((("%") reduce 473) (("*") reduce 473) (("+") reduce 473) (("-") reduce 473) ((".") reduce 473) (("/") reduce 473) ((";") reduce 473) (("<") reduce 473) ((">") reduce 473) (("?") reduce 473) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 473) ((T_BOOLEAN_AND) reduce 473) ((T_BOOLEAN_OR) reduce 473) ((T_COALESCE) reduce 473) ((T_INSTANCEOF) reduce 473) ((T_IS_EQUAL) reduce 473) ((T_IS_GREATER_OR_EQUAL) reduce 473) ((T_IS_IDENTICAL) reduce 473) ((T_IS_NOT_EQUAL) reduce 473) ((T_IS_NOT_IDENTICAL) reduce 473) ((T_IS_SMALLER_OR_EQUAL) reduce 473) ((T_LOGICAL_AND) reduce 473) ((T_LOGICAL_OR) reduce 473) ((T_LOGICAL_XOR) reduce 473) ((T_PIPE) reduce 473) ((T_POW) shift) ((T_SL) reduce 473) ((T_SPACESHIP) reduce 473) ((T_SR) reduce 473) (("^") reduce 473) (("|") reduce 473)) 31352 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 483) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 31353 ((("%") reduce 469) (("*") reduce 469) (("+") reduce 469) (("-") reduce 469) ((".") reduce 469) (("/") reduce 469) ((";") reduce 469) (("<") reduce 469) ((">") reduce 469) (("?") reduce 469) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 469) ((T_BOOLEAN_AND) reduce 469) ((T_BOOLEAN_OR) reduce 469) ((T_COALESCE) reduce 469) ((T_INSTANCEOF) reduce 469) ((T_IS_EQUAL) reduce 469) ((T_IS_GREATER_OR_EQUAL) reduce 469) ((T_IS_IDENTICAL) reduce 469) ((T_IS_NOT_EQUAL) reduce 469) ((T_IS_NOT_IDENTICAL) reduce 469) ((T_IS_SMALLER_OR_EQUAL) reduce 469) ((T_LOGICAL_AND) reduce 469) ((T_LOGICAL_OR) reduce 469) ((T_LOGICAL_XOR) reduce 469) ((T_PIPE) reduce 469) ((T_POW) shift) ((T_SL) reduce 469) ((T_SPACESHIP) reduce 469) ((T_SR) reduce 469) (("^") reduce 469) (("|") reduce 469)) 31356 ((("%") reduce 526) (("*") reduce 526) (("+") reduce 526) (("-") reduce 526) ((".") reduce 526) (("/") reduce 526) ((";") reduce 526) (("<") reduce 526) ((">") reduce 526) (("?") reduce 526) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 526) ((T_BOOLEAN_AND) reduce 526) ((T_BOOLEAN_OR) reduce 526) ((T_COALESCE) reduce 526) ((T_INSTANCEOF) reduce 526) ((T_IS_EQUAL) reduce 526) ((T_IS_GREATER_OR_EQUAL) reduce 526) ((T_IS_IDENTICAL) reduce 526) ((T_IS_NOT_EQUAL) reduce 526) ((T_IS_NOT_IDENTICAL) reduce 526) ((T_IS_SMALLER_OR_EQUAL) reduce 526) ((T_LOGICAL_AND) reduce 526) ((T_LOGICAL_OR) reduce 526) ((T_LOGICAL_XOR) reduce 526) ((T_PIPE) reduce 526) ((T_POW) reduce 526) ((T_SL) reduce 526) ((T_SPACESHIP) reduce 526) ((T_SR) reduce 526) (("^") reduce 526) (("|") reduce 526)) 31358 ((("%") reduce 527) (("*") reduce 527) (("+") reduce 527) (("-") reduce 527) ((".") reduce 527) (("/") reduce 527) ((";") reduce 527) (("<") reduce 527) ((">") reduce 527) (("?") reduce 527) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 527) ((T_BOOLEAN_AND) reduce 527) ((T_BOOLEAN_OR) reduce 527) ((T_COALESCE) reduce 527) ((T_INSTANCEOF) reduce 527) ((T_IS_EQUAL) reduce 527) ((T_IS_GREATER_OR_EQUAL) reduce 527) ((T_IS_IDENTICAL) reduce 527) ((T_IS_NOT_EQUAL) reduce 527) ((T_IS_NOT_IDENTICAL) reduce 527) ((T_IS_SMALLER_OR_EQUAL) reduce 527) ((T_LOGICAL_AND) reduce 527) ((T_LOGICAL_OR) reduce 527) ((T_LOGICAL_XOR) reduce 527) ((T_PIPE) reduce 527) ((T_POW) reduce 527) ((T_SL) reduce 527) ((T_SPACESHIP) reduce 527) ((T_SR) reduce 527) (("^") reduce 527) (("|") reduce 527)) 31359 ((("%") reduce 525) (("*") reduce 525) (("+") reduce 525) (("-") reduce 525) ((".") reduce 525) (("/") reduce 525) ((";") reduce 525) (("<") reduce 525) ((">") reduce 525) (("?") reduce 525) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 525) ((T_BOOLEAN_AND) reduce 525) ((T_BOOLEAN_OR) reduce 525) ((T_COALESCE) reduce 525) ((T_INSTANCEOF) reduce 525) ((T_IS_EQUAL) reduce 525) ((T_IS_GREATER_OR_EQUAL) reduce 525) ((T_IS_IDENTICAL) reduce 525) ((T_IS_NOT_EQUAL) reduce 525) ((T_IS_NOT_IDENTICAL) reduce 525) ((T_IS_SMALLER_OR_EQUAL) reduce 525) ((T_LOGICAL_AND) reduce 525) ((T_LOGICAL_OR) reduce 525) ((T_LOGICAL_XOR) reduce 525) ((T_PIPE) reduce 525) ((T_POW) reduce 525) ((T_SL) reduce 525) ((T_SPACESHIP) reduce 525) ((T_SR) reduce 525) (("^") reduce 525) (("|") reduce 525)) 31360 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 619) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 31361 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 618) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 31363 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 478) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 478) ((T_LOGICAL_OR) reduce 478) ((T_LOGICAL_XOR) reduce 478) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 31364 ((("%") reduce 471) (("*") reduce 471) (("+") reduce 471) (("-") reduce 471) ((".") reduce 471) (("/") reduce 471) ((";") reduce 471) (("<") reduce 471) ((">") reduce 471) (("?") reduce 471) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 471) ((T_BOOLEAN_AND) reduce 471) ((T_BOOLEAN_OR) reduce 471) ((T_COALESCE) reduce 471) ((T_INSTANCEOF) reduce 471) ((T_IS_EQUAL) reduce 471) ((T_IS_GREATER_OR_EQUAL) reduce 471) ((T_IS_IDENTICAL) reduce 471) ((T_IS_NOT_EQUAL) reduce 471) ((T_IS_NOT_IDENTICAL) reduce 471) ((T_IS_SMALLER_OR_EQUAL) reduce 471) ((T_LOGICAL_AND) reduce 471) ((T_LOGICAL_OR) reduce 471) ((T_LOGICAL_XOR) reduce 471) ((T_PIPE) reduce 471) ((T_POW) shift) ((T_SL) reduce 471) ((T_SPACESHIP) reduce 471) ((T_SR) reduce 471) (("^") reduce 471) (("|") reduce 471)) 31366 ((("%") reduce 94) (("(") reduce 94) (("*") reduce 94) (("+") reduce 94) (("-") reduce 94) ((".") reduce 94) (("/") reduce 94) ((";") reduce 94) (("<") reduce 94) ((">") reduce 94) (("?") reduce 94) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 94) ((T_BOOLEAN_AND) reduce 94) ((T_BOOLEAN_OR) reduce 94) ((T_COALESCE) reduce 94) ((T_INSTANCEOF) reduce 94) ((T_IS_EQUAL) reduce 94) ((T_IS_GREATER_OR_EQUAL) reduce 94) ((T_IS_IDENTICAL) reduce 94) ((T_IS_NOT_EQUAL) reduce 94) ((T_IS_NOT_IDENTICAL) reduce 94) ((T_IS_SMALLER_OR_EQUAL) reduce 94) ((T_LOGICAL_AND) reduce 94) ((T_LOGICAL_OR) reduce 94) ((T_LOGICAL_XOR) reduce 94) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 94) ((T_PIPE) reduce 94) ((T_POW) reduce 94) ((T_SL) reduce 94) ((T_SPACESHIP) reduce 94) ((T_SR) reduce 94) (("^") reduce 94) (("|") reduce 94)) 31367 ((("%") reduce 93) (("(") reduce 93) (("*") reduce 93) (("+") reduce 93) (("-") reduce 93) ((".") reduce 93) (("/") reduce 93) ((";") reduce 93) (("<") reduce 93) ((">") reduce 93) (("?") reduce 93) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 93) ((T_BOOLEAN_AND) reduce 93) ((T_BOOLEAN_OR) reduce 93) ((T_COALESCE) reduce 93) ((T_INSTANCEOF) reduce 93) ((T_IS_EQUAL) reduce 93) ((T_IS_GREATER_OR_EQUAL) reduce 93) ((T_IS_IDENTICAL) reduce 93) ((T_IS_NOT_EQUAL) reduce 93) ((T_IS_NOT_IDENTICAL) reduce 93) ((T_IS_SMALLER_OR_EQUAL) reduce 93) ((T_LOGICAL_AND) reduce 93) ((T_LOGICAL_OR) reduce 93) ((T_LOGICAL_XOR) reduce 93) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 93) ((T_PIPE) reduce 93) ((T_POW) reduce 93) ((T_SL) reduce 93) ((T_SPACESHIP) reduce 93) ((T_SR) reduce 93) (("^") reduce 93) (("|") reduce 93)) 31368 ((("%") reduce 95) (("(") reduce 95) (("*") reduce 95) (("+") reduce 95) (("-") reduce 95) ((".") reduce 95) (("/") reduce 95) ((";") reduce 95) (("<") reduce 95) ((">") reduce 95) (("?") reduce 95) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 95) ((T_BOOLEAN_AND) reduce 95) ((T_BOOLEAN_OR) reduce 95) ((T_COALESCE) reduce 95) ((T_INSTANCEOF) reduce 95) ((T_IS_EQUAL) reduce 95) ((T_IS_GREATER_OR_EQUAL) reduce 95) ((T_IS_IDENTICAL) reduce 95) ((T_IS_NOT_EQUAL) reduce 95) ((T_IS_NOT_IDENTICAL) reduce 95) ((T_IS_SMALLER_OR_EQUAL) reduce 95) ((T_LOGICAL_AND) reduce 95) ((T_LOGICAL_OR) reduce 95) ((T_LOGICAL_XOR) reduce 95) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 95) ((T_PIPE) reduce 95) ((T_POW) reduce 95) ((T_SL) reduce 95) ((T_SPACESHIP) reduce 95) ((T_SR) reduce 95) (("^") reduce 95) (("|") reduce 95)) 31369 ((("%") reduce 509) (("(") reduce 509) (("*") reduce 509) (("+") reduce 509) (("-") reduce 509) ((".") reduce 509) (("/") reduce 509) ((";") reduce 509) (("<") reduce 509) ((">") reduce 509) (("?") reduce 509) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 509) ((T_BOOLEAN_AND) reduce 509) ((T_BOOLEAN_OR) reduce 509) ((T_COALESCE) reduce 509) ((T_INSTANCEOF) reduce 509) ((T_IS_EQUAL) reduce 509) ((T_IS_GREATER_OR_EQUAL) reduce 509) ((T_IS_IDENTICAL) reduce 509) ((T_IS_NOT_EQUAL) reduce 509) ((T_IS_NOT_IDENTICAL) reduce 509) ((T_IS_SMALLER_OR_EQUAL) reduce 509) ((T_LOGICAL_AND) reduce 509) ((T_LOGICAL_OR) reduce 509) ((T_LOGICAL_XOR) reduce 509) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 509) ((T_PIPE) reduce 509) ((T_POW) reduce 509) ((T_SL) reduce 509) ((T_SPACESHIP) reduce 509) ((T_SR) reduce 509) (("^") reduce 509) (("|") reduce 509)) 31370 ((("%") reduce 92) (("(") reduce 92) (("*") reduce 92) (("+") reduce 92) (("-") reduce 92) ((".") reduce 92) (("/") reduce 92) ((";") reduce 92) (("<") reduce 92) ((">") reduce 92) (("?") reduce 92) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 92) ((T_BOOLEAN_AND) reduce 92) ((T_BOOLEAN_OR) reduce 92) ((T_COALESCE) reduce 92) ((T_INSTANCEOF) reduce 92) ((T_IS_EQUAL) reduce 92) ((T_IS_GREATER_OR_EQUAL) reduce 92) ((T_IS_IDENTICAL) reduce 92) ((T_IS_NOT_EQUAL) reduce 92) ((T_IS_NOT_IDENTICAL) reduce 92) ((T_IS_SMALLER_OR_EQUAL) reduce 92) ((T_LOGICAL_AND) reduce 92) ((T_LOGICAL_OR) reduce 92) ((T_LOGICAL_XOR) reduce 92) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 92) ((T_PIPE) reduce 92) ((T_POW) reduce 92) ((T_SL) reduce 92) ((T_SPACESHIP) reduce 92) ((T_SR) reduce 92) (("^") reduce 92) (("|") reduce 92)) 31371 ((("%") reduce 400) (("(") reduce 400) (("*") reduce 400) (("+") reduce 400) (("-") reduce 400) ((".") reduce 400) (("/") reduce 400) ((";") reduce 400) (("<") reduce 400) ((">") reduce 400) (("?") reduce 400) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 400) ((T_BOOLEAN_AND) reduce 400) ((T_BOOLEAN_OR) reduce 400) ((T_COALESCE) reduce 400) ((T_INSTANCEOF) reduce 400) ((T_IS_EQUAL) reduce 400) ((T_IS_GREATER_OR_EQUAL) reduce 400) ((T_IS_IDENTICAL) reduce 400) ((T_IS_NOT_EQUAL) reduce 400) ((T_IS_NOT_IDENTICAL) reduce 400) ((T_IS_SMALLER_OR_EQUAL) reduce 400) ((T_LOGICAL_AND) reduce 400) ((T_LOGICAL_OR) reduce 400) ((T_LOGICAL_XOR) reduce 400) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 400) ((T_OBJECT_OPERATOR) reduce 400) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 400) ((T_PIPE) reduce 400) ((T_POW) reduce 400) ((T_SL) reduce 400) ((T_SPACESHIP) reduce 400) ((T_SR) reduce 400) (("[") reduce 400) (("^") reduce 400) (("|") reduce 400)) 31374 ((("%") reduce 511) (("(") reduce 511) (("*") reduce 511) (("+") reduce 511) (("-") reduce 511) ((".") reduce 511) (("/") reduce 511) ((";") reduce 511) (("<") reduce 511) ((">") reduce 511) (("?") reduce 511) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 511) ((T_BOOLEAN_AND) reduce 511) ((T_BOOLEAN_OR) reduce 511) ((T_COALESCE) reduce 511) ((T_INSTANCEOF) reduce 511) ((T_IS_EQUAL) reduce 511) ((T_IS_GREATER_OR_EQUAL) reduce 511) ((T_IS_IDENTICAL) reduce 511) ((T_IS_NOT_EQUAL) reduce 511) ((T_IS_NOT_IDENTICAL) reduce 511) ((T_IS_SMALLER_OR_EQUAL) reduce 511) ((T_LOGICAL_AND) reduce 511) ((T_LOGICAL_OR) reduce 511) ((T_LOGICAL_XOR) reduce 511) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 511) ((T_POW) reduce 511) ((T_SL) reduce 511) ((T_SPACESHIP) reduce 511) ((T_SR) reduce 511) (("^") reduce 511) (("|") reduce 511)) 31375 ((("%") reduce 402) (("(") shift) (("*") reduce 402) (("+") reduce 402) (("-") reduce 402) ((".") reduce 402) (("/") reduce 402) ((";") reduce 402) (("<") reduce 402) ((">") reduce 402) (("?") reduce 402) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 402) ((T_BOOLEAN_AND) reduce 402) ((T_BOOLEAN_OR) reduce 402) ((T_COALESCE) reduce 402) ((T_INSTANCEOF) reduce 402) ((T_IS_EQUAL) reduce 402) ((T_IS_GREATER_OR_EQUAL) reduce 402) ((T_IS_IDENTICAL) reduce 402) ((T_IS_NOT_EQUAL) reduce 402) ((T_IS_NOT_IDENTICAL) reduce 402) ((T_IS_SMALLER_OR_EQUAL) reduce 402) ((T_LOGICAL_AND) reduce 402) ((T_LOGICAL_OR) reduce 402) ((T_LOGICAL_XOR) reduce 402) ((T_PIPE) reduce 402) ((T_POW) reduce 402) ((T_SL) reduce 402) ((T_SPACESHIP) reduce 402) ((T_SR) reduce 402) (("^") reduce 402) (("|") reduce 402)) 31376 ((("%") reduce 510) (("(") reduce 510) (("*") reduce 510) (("+") reduce 510) (("-") reduce 510) ((".") reduce 510) (("/") reduce 510) ((";") reduce 510) (("<") reduce 510) ((">") reduce 510) (("?") reduce 510) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 510) ((T_BOOLEAN_AND) reduce 510) ((T_BOOLEAN_OR) reduce 510) ((T_COALESCE) reduce 510) ((T_INSTANCEOF) reduce 510) ((T_IS_EQUAL) reduce 510) ((T_IS_GREATER_OR_EQUAL) reduce 510) ((T_IS_IDENTICAL) reduce 510) ((T_IS_NOT_EQUAL) reduce 510) ((T_IS_NOT_IDENTICAL) reduce 510) ((T_IS_SMALLER_OR_EQUAL) reduce 510) ((T_LOGICAL_AND) reduce 510) ((T_LOGICAL_OR) reduce 510) ((T_LOGICAL_XOR) reduce 510) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 510) ((T_PIPE) reduce 510) ((T_POW) reduce 510) ((T_SL) reduce 510) ((T_SPACESHIP) reduce 510) ((T_SR) reduce 510) (("^") reduce 510) (("|") reduce 510)) 31377 ((("%") reduce 512) (("(") reduce 512) (("*") reduce 512) (("+") reduce 512) (("-") reduce 512) ((".") reduce 512) (("/") reduce 512) ((";") reduce 512) (("<") reduce 512) ((">") reduce 512) (("?") reduce 512) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 512) ((T_BOOLEAN_AND) reduce 512) ((T_BOOLEAN_OR) reduce 512) ((T_COALESCE) reduce 512) ((T_INSTANCEOF) reduce 512) ((T_IS_EQUAL) reduce 512) ((T_IS_GREATER_OR_EQUAL) reduce 512) ((T_IS_IDENTICAL) reduce 512) ((T_IS_NOT_EQUAL) reduce 512) ((T_IS_NOT_IDENTICAL) reduce 512) ((T_IS_SMALLER_OR_EQUAL) reduce 512) ((T_LOGICAL_AND) reduce 512) ((T_LOGICAL_OR) reduce 512) ((T_LOGICAL_XOR) reduce 512) ((T_NULLSAFE_OBJECT_OPERATOR) shift) ((T_OBJECT_OPERATOR) shift) ((T_PAAMAYIM_NEKUDOTAYIM) shift) ((T_PIPE) reduce 512) ((T_POW) reduce 512) ((T_SL) reduce 512) ((T_SPACESHIP) reduce 512) ((T_SR) reduce 512) (("[") shift) (("^") reduce 512) (("|") reduce 512)) 31378 ((("%") reduce 573) (("(") reduce 573) (("*") reduce 573) (("+") reduce 573) (("-") reduce 573) ((".") reduce 573) (("/") reduce 573) ((";") reduce 573) (("<") reduce 573) ((">") reduce 573) (("?") reduce 573) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 573) ((T_BOOLEAN_AND) reduce 573) ((T_BOOLEAN_OR) reduce 573) ((T_COALESCE) reduce 573) ((T_INSTANCEOF) reduce 573) ((T_IS_EQUAL) reduce 573) ((T_IS_GREATER_OR_EQUAL) reduce 573) ((T_IS_IDENTICAL) reduce 573) ((T_IS_NOT_EQUAL) reduce 573) ((T_IS_NOT_IDENTICAL) reduce 573) ((T_IS_SMALLER_OR_EQUAL) reduce 573) ((T_LOGICAL_AND) reduce 573) ((T_LOGICAL_OR) reduce 573) ((T_LOGICAL_XOR) reduce 573) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 573) ((T_OBJECT_OPERATOR) reduce 573) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 573) ((T_PIPE) reduce 573) ((T_POW) reduce 573) ((T_SL) reduce 573) ((T_SPACESHIP) reduce 573) ((T_SR) reduce 573) (("[") reduce 573) (("^") reduce 573) (("|") reduce 573)) 31384 ((("%") reduce 574) (("(") reduce 574) (("*") reduce 574) (("+") reduce 574) (("-") reduce 574) ((".") reduce 574) (("/") reduce 574) ((";") reduce 574) (("<") reduce 574) ((">") reduce 574) (("?") reduce 574) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 574) ((T_BOOLEAN_AND) reduce 574) ((T_BOOLEAN_OR) reduce 574) ((T_COALESCE) reduce 574) ((T_INSTANCEOF) reduce 574) ((T_IS_EQUAL) reduce 574) ((T_IS_GREATER_OR_EQUAL) reduce 574) ((T_IS_IDENTICAL) reduce 574) ((T_IS_NOT_EQUAL) reduce 574) ((T_IS_NOT_IDENTICAL) reduce 574) ((T_IS_SMALLER_OR_EQUAL) reduce 574) ((T_LOGICAL_AND) reduce 574) ((T_LOGICAL_OR) reduce 574) ((T_LOGICAL_XOR) reduce 574) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 574) ((T_OBJECT_OPERATOR) reduce 574) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 574) ((T_PIPE) reduce 574) ((T_POW) reduce 574) ((T_SL) reduce 574) ((T_SPACESHIP) reduce 574) ((T_SR) reduce 574) (("[") reduce 574) (("^") reduce 574) (("|") reduce 574)) 31385 ((("%") reduce 578) (("(") reduce 578) (("*") reduce 578) (("+") reduce 578) (("-") reduce 578) ((".") reduce 578) (("/") reduce 578) ((";") reduce 578) (("<") reduce 578) ((">") reduce 578) (("?") reduce 578) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 578) ((T_BOOLEAN_AND) reduce 578) ((T_BOOLEAN_OR) reduce 578) ((T_COALESCE) reduce 578) ((T_INSTANCEOF) reduce 578) ((T_IS_EQUAL) reduce 578) ((T_IS_GREATER_OR_EQUAL) reduce 578) ((T_IS_IDENTICAL) reduce 578) ((T_IS_NOT_EQUAL) reduce 578) ((T_IS_NOT_IDENTICAL) reduce 578) ((T_IS_SMALLER_OR_EQUAL) reduce 578) ((T_LOGICAL_AND) reduce 578) ((T_LOGICAL_OR) reduce 578) ((T_LOGICAL_XOR) reduce 578) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 578) ((T_OBJECT_OPERATOR) reduce 578) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 578) ((T_PIPE) reduce 578) ((T_POW) reduce 578) ((T_SL) reduce 578) ((T_SPACESHIP) reduce 578) ((T_SR) reduce 578) (("[") reduce 578) (("^") reduce 578) (("|") reduce 578)) 31386 ((("%") reduce 575) (("(") reduce 575) (("*") reduce 575) (("+") reduce 575) (("-") reduce 575) ((".") reduce 575) (("/") reduce 575) ((";") reduce 575) (("<") reduce 575) ((">") reduce 575) (("?") reduce 575) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 575) ((T_BOOLEAN_AND) reduce 575) ((T_BOOLEAN_OR) reduce 575) ((T_COALESCE) reduce 575) ((T_INSTANCEOF) reduce 575) ((T_IS_EQUAL) reduce 575) ((T_IS_GREATER_OR_EQUAL) reduce 575) ((T_IS_IDENTICAL) reduce 575) ((T_IS_NOT_EQUAL) reduce 575) ((T_IS_NOT_IDENTICAL) reduce 575) ((T_IS_SMALLER_OR_EQUAL) reduce 575) ((T_LOGICAL_AND) reduce 575) ((T_LOGICAL_OR) reduce 575) ((T_LOGICAL_XOR) reduce 575) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 575) ((T_OBJECT_OPERATOR) reduce 575) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 575) ((T_PIPE) reduce 575) ((T_POW) reduce 575) ((T_SL) reduce 575) ((T_SPACESHIP) reduce 575) ((T_SR) reduce 575) (("[") reduce 575) (("^") reduce 575) (("|") reduce 575)) 31387 ((("%") reduce 576) (("(") reduce 576) (("*") reduce 576) (("+") reduce 576) (("-") reduce 576) ((".") reduce 576) (("/") reduce 576) ((";") reduce 576) (("<") reduce 576) ((">") reduce 576) (("?") reduce 576) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 576) ((T_BOOLEAN_AND) reduce 576) ((T_BOOLEAN_OR) reduce 576) ((T_COALESCE) reduce 576) ((T_INSTANCEOF) reduce 576) ((T_IS_EQUAL) reduce 576) ((T_IS_GREATER_OR_EQUAL) reduce 576) ((T_IS_IDENTICAL) reduce 576) ((T_IS_NOT_EQUAL) reduce 576) ((T_IS_NOT_IDENTICAL) reduce 576) ((T_IS_SMALLER_OR_EQUAL) reduce 576) ((T_LOGICAL_AND) reduce 576) ((T_LOGICAL_OR) reduce 576) ((T_LOGICAL_XOR) reduce 576) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 576) ((T_OBJECT_OPERATOR) reduce 576) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 576) ((T_PIPE) reduce 576) ((T_POW) reduce 576) ((T_SL) reduce 576) ((T_SPACESHIP) reduce 576) ((T_SR) reduce 576) (("[") reduce 576) (("^") reduce 576) (("|") reduce 576)) 31388 ((("%") reduce 399) (("(") reduce 399) (("*") reduce 399) (("+") reduce 399) (("-") reduce 399) ((".") reduce 399) (("/") reduce 399) ((";") reduce 399) (("<") reduce 399) ((">") reduce 399) (("?") reduce 399) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 399) ((T_BOOLEAN_AND) reduce 399) ((T_BOOLEAN_OR) reduce 399) ((T_COALESCE) reduce 399) ((T_INSTANCEOF) reduce 399) ((T_IS_EQUAL) reduce 399) ((T_IS_GREATER_OR_EQUAL) reduce 399) ((T_IS_IDENTICAL) reduce 399) ((T_IS_NOT_EQUAL) reduce 399) ((T_IS_NOT_IDENTICAL) reduce 399) ((T_IS_SMALLER_OR_EQUAL) reduce 399) ((T_LOGICAL_AND) reduce 399) ((T_LOGICAL_OR) reduce 399) ((T_LOGICAL_XOR) reduce 399) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 399) ((T_OBJECT_OPERATOR) reduce 399) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 399) ((T_PIPE) reduce 399) ((T_POW) reduce 399) ((T_SL) reduce 399) ((T_SPACESHIP) reduce 399) ((T_SR) reduce 399) (("[") reduce 399) (("^") reduce 399) (("|") reduce 399)) 31390 ((("%") reduce 577) (("(") reduce 577) (("*") reduce 577) (("+") reduce 577) (("-") reduce 577) ((".") reduce 577) (("/") reduce 577) ((";") reduce 577) (("<") reduce 577) ((">") reduce 577) (("?") reduce 577) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 577) ((T_BOOLEAN_AND) reduce 577) ((T_BOOLEAN_OR) reduce 577) ((T_COALESCE) reduce 577) ((T_INSTANCEOF) reduce 577) ((T_IS_EQUAL) reduce 577) ((T_IS_GREATER_OR_EQUAL) reduce 577) ((T_IS_IDENTICAL) reduce 577) ((T_IS_NOT_EQUAL) reduce 577) ((T_IS_NOT_IDENTICAL) reduce 577) ((T_IS_SMALLER_OR_EQUAL) reduce 577) ((T_LOGICAL_AND) reduce 577) ((T_LOGICAL_OR) reduce 577) ((T_LOGICAL_XOR) reduce 577) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 577) ((T_OBJECT_OPERATOR) reduce 577) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 577) ((T_PIPE) reduce 577) ((T_POW) reduce 577) ((T_SL) reduce 577) ((T_SPACESHIP) reduce 577) ((T_SR) reduce 577) (("[") reduce 577) (("^") reduce 577) (("|") reduce 577)) 31391 ((("%") reduce 401) (("(") reduce 401) (("*") reduce 401) (("+") reduce 401) (("-") reduce 401) ((".") reduce 401) (("/") reduce 401) ((";") reduce 401) (("<") reduce 401) ((">") reduce 401) (("?") reduce 401) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 401) ((T_BOOLEAN_AND) reduce 401) ((T_BOOLEAN_OR) reduce 401) ((T_COALESCE) reduce 401) ((T_INSTANCEOF) reduce 401) ((T_IS_EQUAL) reduce 401) ((T_IS_GREATER_OR_EQUAL) reduce 401) ((T_IS_IDENTICAL) reduce 401) ((T_IS_NOT_EQUAL) reduce 401) ((T_IS_NOT_IDENTICAL) reduce 401) ((T_IS_SMALLER_OR_EQUAL) reduce 401) ((T_LOGICAL_AND) reduce 401) ((T_LOGICAL_OR) reduce 401) ((T_LOGICAL_XOR) reduce 401) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 401) ((T_OBJECT_OPERATOR) reduce 401) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 401) ((T_PIPE) reduce 401) ((T_POW) reduce 401) ((T_SL) reduce 401) ((T_SPACESHIP) reduce 401) ((T_SR) reduce 401) (("[") reduce 401) (("^") reduce 401) (("|") reduce 401)) 31399 ((("%") reduce 398) (("(") reduce 398) (("*") reduce 398) (("+") reduce 398) (("-") reduce 398) ((".") reduce 398) (("/") reduce 398) ((";") reduce 398) (("<") reduce 398) ((">") reduce 398) (("?") reduce 398) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 398) ((T_BOOLEAN_AND) reduce 398) ((T_BOOLEAN_OR) reduce 398) ((T_COALESCE) reduce 398) ((T_INSTANCEOF) reduce 398) ((T_IS_EQUAL) reduce 398) ((T_IS_GREATER_OR_EQUAL) reduce 398) ((T_IS_IDENTICAL) reduce 398) ((T_IS_NOT_EQUAL) reduce 398) ((T_IS_NOT_IDENTICAL) reduce 398) ((T_IS_SMALLER_OR_EQUAL) reduce 398) ((T_LOGICAL_AND) reduce 398) ((T_LOGICAL_OR) reduce 398) ((T_LOGICAL_XOR) reduce 398) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 398) ((T_OBJECT_OPERATOR) reduce 398) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 398) ((T_PIPE) reduce 398) ((T_POW) reduce 398) ((T_SL) reduce 398) ((T_SPACESHIP) reduce 398) ((T_SR) reduce 398) (("[") reduce 398) (("^") reduce 398) (("|") reduce 398)) 31401 ((("%") reduce 513) (("(") reduce 513) (("*") reduce 513) (("+") reduce 513) (("-") reduce 513) ((".") reduce 513) (("/") reduce 513) ((";") reduce 513) (("<") reduce 513) ((">") reduce 513) (("?") reduce 513) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 513) ((T_BOOLEAN_AND) reduce 513) ((T_BOOLEAN_OR) reduce 513) ((T_COALESCE) reduce 513) ((T_INSTANCEOF) reduce 513) ((T_IS_EQUAL) reduce 513) ((T_IS_GREATER_OR_EQUAL) reduce 513) ((T_IS_IDENTICAL) reduce 513) ((T_IS_NOT_EQUAL) reduce 513) ((T_IS_NOT_IDENTICAL) reduce 513) ((T_IS_SMALLER_OR_EQUAL) reduce 513) ((T_LOGICAL_AND) reduce 513) ((T_LOGICAL_OR) reduce 513) ((T_LOGICAL_XOR) reduce 513) ((T_PIPE) reduce 513) ((T_POW) reduce 513) ((T_SL) reduce 513) ((T_SPACESHIP) reduce 513) ((T_SR) reduce 513) (("^") reduce 513) (("|") reduce 513)) 31407 ((("%") reduce 233) (("*") reduce 233) (("+") reduce 233) (("-") reduce 233) ((".") reduce 233) (("/") reduce 233) ((";") reduce 233) (("<") reduce 233) ((">") reduce 233) (("?") reduce 233) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 233) ((T_BOOLEAN_AND) reduce 233) ((T_BOOLEAN_OR) reduce 233) ((T_COALESCE) reduce 233) ((T_INSTANCEOF) reduce 233) ((T_IS_EQUAL) reduce 233) ((T_IS_GREATER_OR_EQUAL) reduce 233) ((T_IS_IDENTICAL) reduce 233) ((T_IS_NOT_EQUAL) reduce 233) ((T_IS_NOT_IDENTICAL) reduce 233) ((T_IS_SMALLER_OR_EQUAL) reduce 233) ((T_LOGICAL_AND) reduce 233) ((T_LOGICAL_OR) reduce 233) ((T_LOGICAL_XOR) reduce 233) ((T_PIPE) reduce 233) ((T_POW) reduce 233) ((T_SL) reduce 233) ((T_SPACESHIP) reduce 233) ((T_SR) reduce 233) (("^") reduce 233) (("|") reduce 233)) 31412 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 404) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) reduce 404) ((T_LOGICAL_OR) reduce 404) ((T_LOGICAL_XOR) reduce 404) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 31416 ((("%") reduce 613) (("*") reduce 613) (("+") reduce 613) (("-") reduce 613) ((".") reduce 613) (("/") reduce 613) ((";") reduce 613) (("<") reduce 613) ((">") reduce 613) (("?") reduce 613) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 613) ((T_BOOLEAN_AND) reduce 613) ((T_BOOLEAN_OR) reduce 613) ((T_COALESCE) reduce 613) ((T_INSTANCEOF) reduce 613) ((T_IS_EQUAL) reduce 613) ((T_IS_GREATER_OR_EQUAL) reduce 613) ((T_IS_IDENTICAL) reduce 613) ((T_IS_NOT_EQUAL) reduce 613) ((T_IS_NOT_IDENTICAL) reduce 613) ((T_IS_SMALLER_OR_EQUAL) reduce 613) ((T_LOGICAL_AND) reduce 613) ((T_LOGICAL_OR) reduce 613) ((T_LOGICAL_XOR) reduce 613) ((T_PIPE) reduce 613) ((T_POW) reduce 613) ((T_SL) reduce 613) ((T_SPACESHIP) reduce 613) ((T_SR) reduce 613) (("^") reduce 613) (("|") reduce 613)) 31417 ((("%") reduce 467) (("*") reduce 467) (("+") reduce 467) (("-") reduce 467) ((".") reduce 467) (("/") reduce 467) ((";") reduce 467) (("<") reduce 467) ((">") reduce 467) (("?") reduce 467) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 467) ((T_BOOLEAN_AND) reduce 467) ((T_BOOLEAN_OR) reduce 467) ((T_COALESCE) reduce 467) ((T_INSTANCEOF) reduce 467) ((T_IS_EQUAL) reduce 467) ((T_IS_GREATER_OR_EQUAL) reduce 467) ((T_IS_IDENTICAL) reduce 467) ((T_IS_NOT_EQUAL) reduce 467) ((T_IS_NOT_IDENTICAL) reduce 467) ((T_IS_SMALLER_OR_EQUAL) reduce 467) ((T_LOGICAL_AND) reduce 467) ((T_LOGICAL_OR) reduce 467) ((T_LOGICAL_XOR) reduce 467) ((T_PIPE) reduce 467) ((T_POW) shift) ((T_SL) reduce 467) ((T_SPACESHIP) reduce 467) ((T_SR) reduce 467) (("^") reduce 467) (("|") reduce 467)) 31418 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 616) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 31419 ((("%") shift) (("*") shift) (("+") shift) (("-") shift) ((".") shift) (("/") shift) ((";") reduce 615) (("<") shift) ((">") shift) (("?") shift) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) shift) ((T_BOOLEAN_AND) shift) ((T_BOOLEAN_OR) shift) ((T_COALESCE) shift) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) shift) ((T_IS_GREATER_OR_EQUAL) shift) ((T_IS_IDENTICAL) shift) ((T_IS_NOT_EQUAL) shift) ((T_IS_NOT_IDENTICAL) shift) ((T_IS_SMALLER_OR_EQUAL) shift) ((T_LOGICAL_AND) shift) ((T_LOGICAL_OR) shift) ((T_LOGICAL_XOR) shift) ((T_PIPE) shift) ((T_POW) shift) ((T_SL) shift) ((T_SPACESHIP) shift) ((T_SR) shift) (("^") shift) (("|") shift)) 31420 ((("%") reduce 424) (("*") reduce 424) (("+") reduce 424) (("-") reduce 424) ((".") reduce 424) (("/") reduce 424) ((";") reduce 424) (("<") reduce 424) ((">") reduce 424) (("?") reduce 424) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 424) ((T_BOOLEAN_AND) reduce 424) ((T_BOOLEAN_OR) reduce 424) ((T_COALESCE) reduce 424) ((T_INSTANCEOF) reduce 424) ((T_IS_EQUAL) reduce 424) ((T_IS_GREATER_OR_EQUAL) reduce 424) ((T_IS_IDENTICAL) reduce 424) ((T_IS_NOT_EQUAL) reduce 424) ((T_IS_NOT_IDENTICAL) reduce 424) ((T_IS_SMALLER_OR_EQUAL) reduce 424) ((T_LOGICAL_AND) reduce 424) ((T_LOGICAL_OR) reduce 424) ((T_LOGICAL_XOR) reduce 424) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 424) ((T_POW) reduce 424) ((T_SL) reduce 424) ((T_SPACESHIP) reduce 424) ((T_SR) reduce 424) (("[") reduce 548) (("^") reduce 424) (("|") reduce 424)) 31422 ((("%") reduce 518) (("*") reduce 518) (("+") reduce 518) (("-") reduce 518) ((".") reduce 518) (("/") reduce 518) ((";") reduce 518) (("<") reduce 518) ((">") reduce 518) (("?") reduce 518) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 518) ((T_BOOLEAN_AND) reduce 518) ((T_BOOLEAN_OR) reduce 518) ((T_COALESCE) reduce 518) ((T_INSTANCEOF) reduce 518) ((T_IS_EQUAL) reduce 518) ((T_IS_GREATER_OR_EQUAL) reduce 518) ((T_IS_IDENTICAL) reduce 518) ((T_IS_NOT_EQUAL) reduce 518) ((T_IS_NOT_IDENTICAL) reduce 518) ((T_IS_SMALLER_OR_EQUAL) reduce 518) ((T_LOGICAL_AND) reduce 518) ((T_LOGICAL_OR) reduce 518) ((T_LOGICAL_XOR) reduce 518) ((T_PIPE) reduce 518) ((T_POW) reduce 518) ((T_SL) reduce 518) ((T_SPACESHIP) reduce 518) ((T_SR) reduce 518) (("^") reduce 518) (("|") reduce 518)) 31423 ((("%") reduce 474) (("*") reduce 474) (("+") reduce 474) (("-") reduce 474) ((".") reduce 474) (("/") reduce 474) ((";") reduce 474) (("<") reduce 474) ((">") reduce 474) (("?") reduce 474) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 474) ((T_BOOLEAN_AND) reduce 474) ((T_BOOLEAN_OR) reduce 474) ((T_COALESCE) reduce 474) ((T_INSTANCEOF) reduce 474) ((T_IS_EQUAL) reduce 474) ((T_IS_GREATER_OR_EQUAL) reduce 474) ((T_IS_IDENTICAL) reduce 474) ((T_IS_NOT_EQUAL) reduce 474) ((T_IS_NOT_IDENTICAL) reduce 474) ((T_IS_SMALLER_OR_EQUAL) reduce 474) ((T_LOGICAL_AND) reduce 474) ((T_LOGICAL_OR) reduce 474) ((T_LOGICAL_XOR) reduce 474) ((T_PIPE) reduce 474) ((T_POW) reduce 474) ((T_SL) reduce 474) ((T_SPACESHIP) reduce 474) ((T_SR) reduce 474) (("^") reduce 474) (("|") reduce 474)) 31424 ((("%") reduce 291) (("*") reduce 291) (("+") reduce 291) (("-") reduce 291) ((".") reduce 291) (("/") reduce 291) ((";") reduce 291) (("<") reduce 291) ((">") reduce 291) (("?") reduce 291) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 291) ((T_BOOLEAN_AND) reduce 291) ((T_BOOLEAN_OR) reduce 291) ((T_COALESCE) reduce 291) ((T_INSTANCEOF) reduce 291) ((T_IS_EQUAL) reduce 291) ((T_IS_GREATER_OR_EQUAL) reduce 291) ((T_IS_IDENTICAL) reduce 291) ((T_IS_NOT_EQUAL) reduce 291) ((T_IS_NOT_IDENTICAL) reduce 291) ((T_IS_SMALLER_OR_EQUAL) reduce 291) ((T_LOGICAL_AND) reduce 291) ((T_LOGICAL_OR) reduce 291) ((T_LOGICAL_XOR) reduce 291) ((T_PIPE) reduce 291) ((T_POW) reduce 291) ((T_SL) reduce 291) ((T_SPACESHIP) reduce 291) ((T_SR) reduce 291) (("^") reduce 291) (("|") reduce 291)) 31428 ((("%") reduce 292) (("*") reduce 292) (("+") reduce 292) (("-") reduce 292) ((".") reduce 292) (("/") reduce 292) ((";") reduce 292) (("<") reduce 292) ((">") reduce 292) (("?") reduce 292) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 292) ((T_BOOLEAN_AND) reduce 292) ((T_BOOLEAN_OR) reduce 292) ((T_COALESCE) reduce 292) ((T_INSTANCEOF) reduce 292) ((T_IS_EQUAL) reduce 292) ((T_IS_GREATER_OR_EQUAL) reduce 292) ((T_IS_IDENTICAL) reduce 292) ((T_IS_NOT_EQUAL) reduce 292) ((T_IS_NOT_IDENTICAL) reduce 292) ((T_IS_SMALLER_OR_EQUAL) reduce 292) ((T_LOGICAL_AND) reduce 292) ((T_LOGICAL_OR) reduce 292) ((T_LOGICAL_XOR) reduce 292) ((T_PIPE) reduce 292) ((T_POW) reduce 292) ((T_SL) reduce 292) ((T_SPACESHIP) reduce 292) ((T_SR) reduce 292) (("^") reduce 292) (("|") reduce 292)) 31429 ((("%") reduce 293) (("*") reduce 293) (("+") reduce 293) (("-") reduce 293) ((".") reduce 293) (("/") reduce 293) ((";") reduce 293) (("<") reduce 293) ((">") reduce 293) (("?") reduce 293) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 293) ((T_BOOLEAN_AND) reduce 293) ((T_BOOLEAN_OR) reduce 293) ((T_COALESCE) reduce 293) ((T_INSTANCEOF) reduce 293) ((T_IS_EQUAL) reduce 293) ((T_IS_GREATER_OR_EQUAL) reduce 293) ((T_IS_IDENTICAL) reduce 293) ((T_IS_NOT_EQUAL) reduce 293) ((T_IS_NOT_IDENTICAL) reduce 293) ((T_IS_SMALLER_OR_EQUAL) reduce 293) ((T_LOGICAL_AND) reduce 293) ((T_LOGICAL_OR) reduce 293) ((T_LOGICAL_XOR) reduce 293) ((T_PIPE) reduce 293) ((T_POW) reduce 293) ((T_SL) reduce 293) ((T_SPACESHIP) reduce 293) ((T_SR) reduce 293) (("^") reduce 293) (("|") reduce 293)) 31432 ((("%") reduce 617) (("*") reduce 617) (("+") reduce 617) (("-") reduce 617) ((".") reduce 617) (("/") reduce 617) ((";") reduce 617) (("<") reduce 617) ((">") reduce 617) (("?") reduce 617) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 617) ((T_BOOLEAN_AND) reduce 617) ((T_BOOLEAN_OR) reduce 617) ((T_COALESCE) reduce 617) ((T_INSTANCEOF) reduce 617) ((T_IS_EQUAL) reduce 617) ((T_IS_GREATER_OR_EQUAL) reduce 617) ((T_IS_IDENTICAL) reduce 617) ((T_IS_NOT_EQUAL) reduce 617) ((T_IS_NOT_IDENTICAL) reduce 617) ((T_IS_SMALLER_OR_EQUAL) reduce 617) ((T_LOGICAL_AND) reduce 617) ((T_LOGICAL_OR) reduce 617) ((T_LOGICAL_XOR) reduce 617) ((T_PIPE) reduce 617) ((T_POW) reduce 617) ((T_SL) reduce 617) ((T_SPACESHIP) reduce 617) ((T_SR) reduce 617) (("^") reduce 617) (("|") reduce 617)) 31435 ((("%") reduce 614) (("*") reduce 614) (("+") reduce 614) (("-") reduce 614) ((".") reduce 614) (("/") reduce 614) ((";") reduce 614) (("<") reduce 614) ((">") reduce 614) (("?") reduce 614) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 614) ((T_BOOLEAN_AND) reduce 614) ((T_BOOLEAN_OR) reduce 614) ((T_COALESCE) reduce 614) ((T_INSTANCEOF) reduce 614) ((T_IS_EQUAL) reduce 614) ((T_IS_GREATER_OR_EQUAL) reduce 614) ((T_IS_IDENTICAL) reduce 614) ((T_IS_NOT_EQUAL) reduce 614) ((T_IS_NOT_IDENTICAL) reduce 614) ((T_IS_SMALLER_OR_EQUAL) reduce 614) ((T_LOGICAL_AND) reduce 614) ((T_LOGICAL_OR) reduce 614) ((T_LOGICAL_XOR) reduce 614) ((T_PIPE) reduce 614) ((T_POW) reduce 614) ((T_SL) reduce 614) ((T_SPACESHIP) reduce 614) ((T_SR) reduce 614) (("^") reduce 614) (("|") reduce 614)) 31436 ((("%") reduce 468) (("*") reduce 468) (("+") reduce 468) (("-") reduce 468) ((".") reduce 468) (("/") reduce 468) ((";") reduce 468) (("<") reduce 468) ((">") reduce 468) (("?") reduce 468) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 468) ((T_BOOLEAN_AND) reduce 468) ((T_BOOLEAN_OR) reduce 468) ((T_COALESCE) reduce 468) ((T_INSTANCEOF) reduce 468) ((T_IS_EQUAL) reduce 468) ((T_IS_GREATER_OR_EQUAL) reduce 468) ((T_IS_IDENTICAL) reduce 468) ((T_IS_NOT_EQUAL) reduce 468) ((T_IS_NOT_IDENTICAL) reduce 468) ((T_IS_SMALLER_OR_EQUAL) reduce 468) ((T_LOGICAL_AND) reduce 468) ((T_LOGICAL_OR) reduce 468) ((T_LOGICAL_XOR) reduce 468) ((T_PIPE) reduce 468) ((T_POW) shift) ((T_SL) reduce 468) ((T_SPACESHIP) reduce 468) ((T_SR) reduce 468) (("^") reduce 468) (("|") reduce 468)) 31437 ((("%") reduce 426) (("*") reduce 426) (("+") reduce 426) (("-") reduce 426) ((".") reduce 426) (("/") reduce 426) ((";") reduce 426) (("<") reduce 426) ((">") reduce 426) (("?") reduce 426) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 426) ((T_BOOLEAN_AND) reduce 426) ((T_BOOLEAN_OR) reduce 426) ((T_COALESCE) reduce 426) ((T_INSTANCEOF) reduce 426) ((T_IS_EQUAL) reduce 426) ((T_IS_GREATER_OR_EQUAL) reduce 426) ((T_IS_IDENTICAL) reduce 426) ((T_IS_NOT_EQUAL) reduce 426) ((T_IS_NOT_IDENTICAL) reduce 426) ((T_IS_SMALLER_OR_EQUAL) reduce 426) ((T_LOGICAL_AND) reduce 426) ((T_LOGICAL_OR) reduce 426) ((T_LOGICAL_XOR) reduce 426) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 548) ((T_OBJECT_OPERATOR) reduce 548) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 548) ((T_PIPE) reduce 426) ((T_POW) reduce 426) ((T_SL) reduce 426) ((T_SPACESHIP) reduce 426) ((T_SR) reduce 426) (("[") reduce 548) (("^") reduce 426) (("|") reduce 426)) 31439 ((("%") reduce 408) (("*") reduce 408) (("+") reduce 408) (("-") reduce 408) ((".") reduce 408) (("/") reduce 408) ((";") reduce 408) (("<") reduce 408) ((">") reduce 408) (("?") reduce 408) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 408) ((T_BOOLEAN_AND) reduce 408) ((T_BOOLEAN_OR) reduce 408) ((T_COALESCE) reduce 408) ((T_INSTANCEOF) reduce 408) ((T_IS_EQUAL) reduce 408) ((T_IS_GREATER_OR_EQUAL) reduce 408) ((T_IS_IDENTICAL) reduce 408) ((T_IS_NOT_EQUAL) reduce 408) ((T_IS_NOT_IDENTICAL) reduce 408) ((T_IS_SMALLER_OR_EQUAL) reduce 408) ((T_LOGICAL_AND) reduce 408) ((T_LOGICAL_OR) reduce 408) ((T_LOGICAL_XOR) reduce 408) ((T_PIPE) reduce 408) ((T_POW) reduce 408) ((T_SL) reduce 408) ((T_SPACESHIP) reduce 408) ((T_SR) reduce 408) (("^") reduce 408) (("|") reduce 408)) 31440 ((("%") reduce 409) (("*") reduce 409) (("+") reduce 409) (("-") reduce 409) ((".") reduce 409) (("/") reduce 409) ((";") reduce 409) (("<") reduce 409) ((">") reduce 409) (("?") reduce 409) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 409) ((T_BOOLEAN_AND) reduce 409) ((T_BOOLEAN_OR) reduce 409) ((T_COALESCE) reduce 409) ((T_INSTANCEOF) reduce 409) ((T_IS_EQUAL) reduce 409) ((T_IS_GREATER_OR_EQUAL) reduce 409) ((T_IS_IDENTICAL) reduce 409) ((T_IS_NOT_EQUAL) reduce 409) ((T_IS_NOT_IDENTICAL) reduce 409) ((T_IS_SMALLER_OR_EQUAL) reduce 409) ((T_LOGICAL_AND) reduce 409) ((T_LOGICAL_OR) reduce 409) ((T_LOGICAL_XOR) reduce 409) ((T_PIPE) reduce 409) ((T_POW) reduce 409) ((T_SL) reduce 409) ((T_SPACESHIP) reduce 409) ((T_SR) reduce 409) (("^") reduce 409) (("|") reduce 409)) 31441 ((("%") reduce 296) (("*") reduce 296) (("+") reduce 296) (("-") reduce 296) ((".") reduce 296) (("/") reduce 296) ((";") reduce 296) (("<") reduce 296) ((">") reduce 296) (("?") reduce 296) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 296) ((T_BOOLEAN_AND) reduce 296) ((T_BOOLEAN_OR) reduce 296) ((T_COALESCE) reduce 296) ((T_INSTANCEOF) reduce 296) ((T_IS_EQUAL) reduce 296) ((T_IS_GREATER_OR_EQUAL) reduce 296) ((T_IS_IDENTICAL) reduce 296) ((T_IS_NOT_EQUAL) reduce 296) ((T_IS_NOT_IDENTICAL) reduce 296) ((T_IS_SMALLER_OR_EQUAL) reduce 296) ((T_LOGICAL_AND) reduce 296) ((T_LOGICAL_OR) reduce 296) ((T_LOGICAL_XOR) reduce 296) ((T_PIPE) reduce 296) ((T_POW) reduce 296) ((T_SL) reduce 296) ((T_SPACESHIP) reduce 296) ((T_SR) reduce 296) (("^") reduce 296) (("|") reduce 296)) 31446 ((("%") reduce 297) (("*") reduce 297) (("+") reduce 297) (("-") reduce 297) ((".") reduce 297) (("/") reduce 297) ((";") reduce 297) (("<") reduce 297) ((">") reduce 297) (("?") reduce 297) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 297) ((T_BOOLEAN_AND) reduce 297) ((T_BOOLEAN_OR) reduce 297) ((T_COALESCE) reduce 297) ((T_INSTANCEOF) reduce 297) ((T_IS_EQUAL) reduce 297) ((T_IS_GREATER_OR_EQUAL) reduce 297) ((T_IS_IDENTICAL) reduce 297) ((T_IS_NOT_EQUAL) reduce 297) ((T_IS_NOT_IDENTICAL) reduce 297) ((T_IS_SMALLER_OR_EQUAL) reduce 297) ((T_LOGICAL_AND) reduce 297) ((T_LOGICAL_OR) reduce 297) ((T_LOGICAL_XOR) reduce 297) ((T_PIPE) reduce 297) ((T_POW) reduce 297) ((T_SL) reduce 297) ((T_SPACESHIP) reduce 297) ((T_SR) reduce 297) (("^") reduce 297) (("|") reduce 297)) 31447 ((("%") reduce 460) (("(") reduce 556) (("*") reduce 460) (("+") reduce 460) (("-") reduce 460) ((".") reduce 460) (("/") reduce 460) ((";") reduce 460) (("<") reduce 460) ((">") reduce 460) (("?") reduce 460) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 460) ((T_BOOLEAN_AND) reduce 460) ((T_BOOLEAN_OR) reduce 460) ((T_COALESCE) reduce 460) ((T_INSTANCEOF) reduce 460) ((T_IS_EQUAL) reduce 460) ((T_IS_GREATER_OR_EQUAL) reduce 460) ((T_IS_IDENTICAL) reduce 460) ((T_IS_NOT_EQUAL) reduce 460) ((T_IS_NOT_IDENTICAL) reduce 460) ((T_IS_SMALLER_OR_EQUAL) reduce 460) ((T_LOGICAL_AND) reduce 460) ((T_LOGICAL_OR) reduce 460) ((T_LOGICAL_XOR) reduce 460) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 549) ((T_OBJECT_OPERATOR) reduce 549) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 549) ((T_PIPE) reduce 460) ((T_POW) reduce 460) ((T_SL) reduce 460) ((T_SPACESHIP) reduce 460) ((T_SR) reduce 460) (("[") reduce 549) (("^") reduce 460) (("|") reduce 460)) 31449 ((("%") reduce 298) (("*") reduce 298) (("+") reduce 298) (("-") reduce 298) ((".") reduce 298) (("/") reduce 298) ((";") reduce 298) (("<") reduce 298) ((">") reduce 298) (("?") reduce 298) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 298) ((T_BOOLEAN_AND) reduce 298) ((T_BOOLEAN_OR) reduce 298) ((T_COALESCE) reduce 298) ((T_INSTANCEOF) reduce 298) ((T_IS_EQUAL) reduce 298) ((T_IS_GREATER_OR_EQUAL) reduce 298) ((T_IS_IDENTICAL) reduce 298) ((T_IS_NOT_EQUAL) reduce 298) ((T_IS_NOT_IDENTICAL) reduce 298) ((T_IS_SMALLER_OR_EQUAL) reduce 298) ((T_LOGICAL_AND) reduce 298) ((T_LOGICAL_OR) reduce 298) ((T_LOGICAL_XOR) reduce 298) ((T_PIPE) reduce 298) ((T_POW) reduce 298) ((T_SL) reduce 298) ((T_SPACESHIP) reduce 298) ((T_SR) reduce 298) (("^") reduce 298) (("|") reduce 298)) 31450 ((("%") reduce 299) (("*") reduce 299) (("+") reduce 299) (("-") reduce 299) ((".") reduce 299) (("/") reduce 299) ((";") reduce 299) (("<") reduce 299) ((">") reduce 299) (("?") reduce 299) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 299) ((T_BOOLEAN_AND) reduce 299) ((T_BOOLEAN_OR) reduce 299) ((T_COALESCE) reduce 299) ((T_INSTANCEOF) reduce 299) ((T_IS_EQUAL) reduce 299) ((T_IS_GREATER_OR_EQUAL) reduce 299) ((T_IS_IDENTICAL) reduce 299) ((T_IS_NOT_EQUAL) reduce 299) ((T_IS_NOT_IDENTICAL) reduce 299) ((T_IS_SMALLER_OR_EQUAL) reduce 299) ((T_LOGICAL_AND) reduce 299) ((T_LOGICAL_OR) reduce 299) ((T_LOGICAL_XOR) reduce 299) ((T_PIPE) reduce 299) ((T_POW) reduce 299) ((T_SL) reduce 299) ((T_SPACESHIP) reduce 299) ((T_SR) reduce 299) (("^") reduce 299) (("|") reduce 299)) 31451 ((("%") reduce 472) (("*") reduce 472) (("+") reduce 472) (("-") reduce 472) ((".") reduce 472) (("/") reduce 472) ((";") reduce 472) (("<") reduce 472) ((">") reduce 472) (("?") reduce 472) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 472) ((T_BOOLEAN_AND) reduce 472) ((T_BOOLEAN_OR) reduce 472) ((T_COALESCE) reduce 472) ((T_INSTANCEOF) reduce 472) ((T_IS_EQUAL) reduce 472) ((T_IS_GREATER_OR_EQUAL) reduce 472) ((T_IS_IDENTICAL) reduce 472) ((T_IS_NOT_EQUAL) reduce 472) ((T_IS_NOT_IDENTICAL) reduce 472) ((T_IS_SMALLER_OR_EQUAL) reduce 472) ((T_LOGICAL_AND) reduce 472) ((T_LOGICAL_OR) reduce 472) ((T_LOGICAL_XOR) reduce 472) ((T_PIPE) reduce 472) ((T_POW) shift) ((T_SL) reduce 472) ((T_SPACESHIP) reduce 472) ((T_SR) reduce 472) (("^") reduce 472) (("|") reduce 472)) 31452 ((("%") reduce 470) (("*") reduce 470) (("+") reduce 470) (("-") reduce 470) ((".") reduce 470) (("/") reduce 470) ((";") reduce 470) (("<") reduce 470) ((">") reduce 470) (("?") reduce 470) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 470) ((T_BOOLEAN_AND) reduce 470) ((T_BOOLEAN_OR) reduce 470) ((T_COALESCE) reduce 470) ((T_INSTANCEOF) reduce 470) ((T_IS_EQUAL) reduce 470) ((T_IS_GREATER_OR_EQUAL) reduce 470) ((T_IS_IDENTICAL) reduce 470) ((T_IS_NOT_EQUAL) reduce 470) ((T_IS_NOT_IDENTICAL) reduce 470) ((T_IS_SMALLER_OR_EQUAL) reduce 470) ((T_LOGICAL_AND) reduce 470) ((T_LOGICAL_OR) reduce 470) ((T_LOGICAL_XOR) reduce 470) ((T_PIPE) reduce 470) ((T_POW) shift) ((T_SL) reduce 470) ((T_SPACESHIP) reduce 470) ((T_SR) reduce 470) (("^") reduce 470) (("|") reduce 470)) 31455 ((("%") reduce 519) (("(") reduce 519) (("*") reduce 519) (("+") reduce 519) (("-") reduce 519) ((".") reduce 519) (("/") reduce 519) ((";") reduce 519) (("<") reduce 519) ((">") reduce 519) (("?") reduce 519) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 519) ((T_BOOLEAN_AND) reduce 519) ((T_BOOLEAN_OR) reduce 519) ((T_COALESCE) reduce 519) ((T_INSTANCEOF) reduce 519) ((T_IS_EQUAL) reduce 519) ((T_IS_GREATER_OR_EQUAL) reduce 519) ((T_IS_IDENTICAL) reduce 519) ((T_IS_NOT_EQUAL) reduce 519) ((T_IS_NOT_IDENTICAL) reduce 519) ((T_IS_SMALLER_OR_EQUAL) reduce 519) ((T_LOGICAL_AND) reduce 519) ((T_LOGICAL_OR) reduce 519) ((T_LOGICAL_XOR) reduce 519) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 519) ((T_OBJECT_OPERATOR) reduce 519) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 519) ((T_PIPE) reduce 519) ((T_POW) reduce 519) ((T_SL) reduce 519) ((T_SPACESHIP) reduce 519) ((T_SR) reduce 519) (("[") reduce 519) (("^") reduce 519) (("|") reduce 519)) 31456 ((("%") reduce 475) (("*") reduce 475) (("+") reduce 475) (("-") reduce 475) ((".") reduce 475) (("/") reduce 475) ((";") reduce 475) (("<") reduce 475) ((">") reduce 475) (("?") reduce 475) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 475) ((T_BOOLEAN_AND) reduce 475) ((T_BOOLEAN_OR) reduce 475) ((T_COALESCE) reduce 475) ((T_INSTANCEOF) reduce 475) ((T_IS_EQUAL) reduce 475) ((T_IS_GREATER_OR_EQUAL) reduce 475) ((T_IS_IDENTICAL) reduce 475) ((T_IS_NOT_EQUAL) reduce 475) ((T_IS_NOT_IDENTICAL) reduce 475) ((T_IS_SMALLER_OR_EQUAL) reduce 475) ((T_LOGICAL_AND) reduce 475) ((T_LOGICAL_OR) reduce 475) ((T_LOGICAL_XOR) reduce 475) ((T_PIPE) reduce 475) ((T_POW) shift) ((T_SL) reduce 475) ((T_SPACESHIP) reduce 475) ((T_SR) reduce 475) (("^") reduce 475) (("|") reduce 475)) 31457 ((("%") reduce 446) (("*") reduce 446) (("+") reduce 446) (("-") reduce 446) ((".") reduce 446) (("/") reduce 446) ((";") reduce 446) (("<") reduce 446) ((">") reduce 446) (("?") reduce 446) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 446) ((T_BOOLEAN_AND) reduce 446) ((T_BOOLEAN_OR) reduce 446) ((T_COALESCE) reduce 446) ((T_INSTANCEOF) reduce 446) ((T_IS_EQUAL) reduce 446) ((T_IS_GREATER_OR_EQUAL) reduce 446) ((T_IS_IDENTICAL) reduce 446) ((T_IS_NOT_EQUAL) reduce 446) ((T_IS_NOT_IDENTICAL) reduce 446) ((T_IS_SMALLER_OR_EQUAL) reduce 446) ((T_LOGICAL_AND) reduce 446) ((T_LOGICAL_OR) reduce 446) ((T_LOGICAL_XOR) reduce 446) ((T_PIPE) reduce 446) ((T_POW) shift) ((T_SL) reduce 446) ((T_SPACESHIP) reduce 446) ((T_SR) reduce 446) (("^") reduce 446) (("|") reduce 446)) 31458 ((("%") reduce 445) (("*") reduce 445) (("+") reduce 445) (("-") reduce 445) ((".") reduce 445) (("/") reduce 445) ((";") reduce 445) (("<") reduce 445) ((">") reduce 445) (("?") reduce 445) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 445) ((T_BOOLEAN_AND) reduce 445) ((T_BOOLEAN_OR) reduce 445) ((T_COALESCE) reduce 445) ((T_INSTANCEOF) reduce 445) ((T_IS_EQUAL) reduce 445) ((T_IS_GREATER_OR_EQUAL) reduce 445) ((T_IS_IDENTICAL) reduce 445) ((T_IS_NOT_EQUAL) reduce 445) ((T_IS_NOT_IDENTICAL) reduce 445) ((T_IS_SMALLER_OR_EQUAL) reduce 445) ((T_LOGICAL_AND) reduce 445) ((T_LOGICAL_OR) reduce 445) ((T_LOGICAL_XOR) reduce 445) ((T_PIPE) reduce 445) ((T_POW) shift) ((T_SL) reduce 445) ((T_SPACESHIP) reduce 445) ((T_SR) reduce 445) (("^") reduce 445) (("|") reduce 445)) 31460 ((("%") reduce 570) (("(") reduce 570) (("*") reduce 570) (("+") reduce 570) (("-") reduce 570) ((".") reduce 570) (("/") reduce 570) ((";") reduce 570) (("<") reduce 570) (("=") reduce 570) ((">") reduce 570) (("?") reduce 570) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 570) ((T_AND_EQUAL) reduce 570) ((T_BOOLEAN_AND) reduce 570) ((T_BOOLEAN_OR) reduce 570) ((T_COALESCE) reduce 570) ((T_COALESCE_EQUAL) reduce 570) ((T_CONCAT_EQUAL) reduce 570) ((T_DEC) reduce 570) ((T_DIV_EQUAL) reduce 570) ((T_INC) reduce 570) ((T_INSTANCEOF) reduce 570) ((T_IS_EQUAL) reduce 570) ((T_IS_GREATER_OR_EQUAL) reduce 570) ((T_IS_IDENTICAL) reduce 570) ((T_IS_NOT_EQUAL) reduce 570) ((T_IS_NOT_IDENTICAL) reduce 570) ((T_IS_SMALLER_OR_EQUAL) reduce 570) ((T_LOGICAL_AND) reduce 570) ((T_LOGICAL_OR) reduce 570) ((T_LOGICAL_XOR) reduce 570) ((T_MINUS_EQUAL) reduce 570) ((T_MOD_EQUAL) reduce 570) ((T_MUL_EQUAL) reduce 570) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 570) ((T_OBJECT_OPERATOR) reduce 570) ((T_OR_EQUAL) reduce 570) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 570) ((T_PIPE) reduce 570) ((T_PLUS_EQUAL) reduce 570) ((T_POW) reduce 570) ((T_POW_EQUAL) reduce 570) ((T_SL) reduce 570) ((T_SL_EQUAL) reduce 570) ((T_SPACESHIP) reduce 570) ((T_SR) reduce 570) ((T_SR_EQUAL) reduce 570) ((T_XOR_EQUAL) reduce 570) (("[") reduce 570) (("^") reduce 570) (("|") reduce 570)) 31463 ((("%") reduce 569) (("(") reduce 569) (("*") reduce 569) (("+") reduce 569) (("-") reduce 569) ((".") reduce 569) (("/") reduce 569) ((";") reduce 569) (("<") reduce 569) (("=") reduce 569) ((">") reduce 569) (("?") reduce 569) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 569) ((T_AND_EQUAL) reduce 569) ((T_BOOLEAN_AND) reduce 569) ((T_BOOLEAN_OR) reduce 569) ((T_COALESCE) reduce 569) ((T_COALESCE_EQUAL) reduce 569) ((T_CONCAT_EQUAL) reduce 569) ((T_DEC) reduce 569) ((T_DIV_EQUAL) reduce 569) ((T_INC) reduce 569) ((T_INSTANCEOF) reduce 569) ((T_IS_EQUAL) reduce 569) ((T_IS_GREATER_OR_EQUAL) reduce 569) ((T_IS_IDENTICAL) reduce 569) ((T_IS_NOT_EQUAL) reduce 569) ((T_IS_NOT_IDENTICAL) reduce 569) ((T_IS_SMALLER_OR_EQUAL) reduce 569) ((T_LOGICAL_AND) reduce 569) ((T_LOGICAL_OR) reduce 569) ((T_LOGICAL_XOR) reduce 569) ((T_MINUS_EQUAL) reduce 569) ((T_MOD_EQUAL) reduce 569) ((T_MUL_EQUAL) reduce 569) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 569) ((T_OBJECT_OPERATOR) reduce 569) ((T_OR_EQUAL) reduce 569) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 569) ((T_PIPE) reduce 569) ((T_PLUS_EQUAL) reduce 569) ((T_POW) reduce 569) ((T_POW_EQUAL) reduce 569) ((T_SL) reduce 569) ((T_SL_EQUAL) reduce 569) ((T_SPACESHIP) reduce 569) ((T_SR) reduce 569) ((T_SR_EQUAL) reduce 569) ((T_XOR_EQUAL) reduce 569) (("[") reduce 569) (("^") reduce 569) (("|") reduce 569)) 31465 ((("%") reduce 522) (("(") reduce 522) (("*") reduce 522) (("+") reduce 522) (("-") reduce 522) ((".") reduce 522) (("/") reduce 522) ((";") reduce 522) (("<") reduce 522) ((">") reduce 522) (("?") reduce 522) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 522) ((T_BOOLEAN_AND) reduce 522) ((T_BOOLEAN_OR) reduce 522) ((T_COALESCE) reduce 522) ((T_INSTANCEOF) reduce 522) ((T_IS_EQUAL) reduce 522) ((T_IS_GREATER_OR_EQUAL) reduce 522) ((T_IS_IDENTICAL) reduce 522) ((T_IS_NOT_EQUAL) reduce 522) ((T_IS_NOT_IDENTICAL) reduce 522) ((T_IS_SMALLER_OR_EQUAL) reduce 522) ((T_LOGICAL_AND) reduce 522) ((T_LOGICAL_OR) reduce 522) ((T_LOGICAL_XOR) reduce 522) ((T_NULLSAFE_OBJECT_OPERATOR) reduce 522) ((T_OBJECT_OPERATOR) reduce 522) ((T_PAAMAYIM_NEKUDOTAYIM) reduce 522) ((T_PIPE) reduce 522) ((T_POW) reduce 522) ((T_SL) reduce 522) ((T_SPACESHIP) reduce 522) ((T_SR) reduce 522) (("[") reduce 522) (("^") reduce 522) (("|") reduce 522)) 31466 ((("%") reduce 447) (("*") reduce 447) (("+") reduce 447) (("-") reduce 447) ((".") reduce 447) (("/") reduce 447) ((";") reduce 447) (("<") reduce 447) ((">") reduce 447) (("?") reduce 447) ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) reduce 447) ((T_BOOLEAN_AND) reduce 447) ((T_BOOLEAN_OR) reduce 447) ((T_COALESCE) reduce 447) ((T_INSTANCEOF) shift) ((T_IS_EQUAL) reduce 447) ((T_IS_GREATER_OR_EQUAL) reduce 447) ((T_IS_IDENTICAL) reduce 447) ((T_IS_NOT_EQUAL) reduce 447) ((T_IS_NOT_IDENTICAL) reduce 447) ((T_IS_SMALLER_OR_EQUAL) reduce 447) ((T_LOGICAL_AND) reduce 447) ((T_LOGICAL_OR) reduce 447) ((T_LOGICAL_XOR) reduce 447) ((T_PIPE) reduce 447) ((T_POW) shift) ((T_SL) reduce 447) ((T_SPACESHIP) reduce 447) ((T_SR) reduce 447) (("^") reduce 447) (("|") reduce 447)) 31467 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (($) reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONST) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_EMPTY) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAMESPACE) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_USE) reduce 161) ((T_VARIABLE) reduce 161) ((T_VOID_CAST) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 31483 ((("!") reduce 189) (("\"") reduce 189) (($) reduce 189) (("$") reduce 189) (("(") reduce 189) (("+") reduce 189) (("-") reduce 189) ((";") reduce 189) (("@") reduce 189) ((T_ABSTRACT) reduce 189) ((T_ARRAY) reduce 189) ((T_ARRAY_CAST) reduce 189) ((T_ATTRIBUTE) reduce 189) ((T_BOOL_CAST) reduce 189) ((T_BREAK) reduce 189) ((T_CLASS) reduce 189) ((T_CLASS_C) reduce 189) ((T_CLONE) reduce 189) ((T_CONST) reduce 189) ((T_CONSTANT_ENCAPSED_STRING) reduce 189) ((T_CONTINUE) reduce 189) ((T_DEC) reduce 189) ((T_DECLARE) reduce 189) ((T_DIR) reduce 189) ((T_DNUMBER) reduce 189) ((T_DO) reduce 189) ((T_DOUBLE_CAST) reduce 189) ((T_ECHO) reduce 189) ((T_EMPTY) reduce 189) ((T_ENUM) reduce 189) ((T_EVAL) reduce 189) ((T_EXIT) reduce 189) ((T_FILE) reduce 189) ((T_FINAL) reduce 189) ((T_FN) reduce 189) ((T_FOR) reduce 189) ((T_FOREACH) reduce 189) ((T_FUNCTION) reduce 189) ((T_FUNC_C) reduce 189) ((T_GLOBAL) reduce 189) ((T_GOTO) reduce 189) ((T_HALT_COMPILER) reduce 189) ((T_IF) reduce 189) ((T_INC) reduce 189) ((T_INCLUDE) reduce 189) ((T_INCLUDE_ONCE) reduce 189) ((T_INLINE_HTML) reduce 189) ((T_INTERFACE) reduce 189) ((T_INT_CAST) reduce 189) ((T_ISSET) reduce 189) ((T_LINE) reduce 189) ((T_LIST) reduce 189) ((T_LNUMBER) reduce 189) ((T_MATCH) reduce 189) ((T_METHOD_C) reduce 189) ((T_NAMESPACE) reduce 189) ((T_NAME_FULLY_QUALIFIED) reduce 189) ((T_NAME_QUALIFIED) reduce 189) ((T_NAME_RELATIVE) reduce 189) ((T_NEW) reduce 189) ((T_NS_C) reduce 189) ((T_OBJECT_CAST) reduce 189) ((T_PRINT) reduce 189) ((T_PROPERTY_C) reduce 189) ((T_READONLY) reduce 189) ((T_REQUIRE) reduce 189) ((T_REQUIRE_ONCE) reduce 189) ((T_RETURN) reduce 189) ((T_START_HEREDOC) reduce 189) ((T_STATIC) reduce 189) ((T_STRING) reduce 189) ((T_STRING_CAST) reduce 189) ((T_SWITCH) reduce 189) ((T_THROW) reduce 189) ((T_TRAIT) reduce 189) ((T_TRAIT_C) reduce 189) ((T_TRY) reduce 189) ((T_UNSET) reduce 189) ((T_UNSET_CAST) reduce 189) ((T_USE) reduce 189) ((T_VARIABLE) reduce 189) ((T_VOID_CAST) reduce 189) ((T_WHILE) reduce 189) ((T_YIELD) reduce 189) ((T_YIELD_FROM) reduce 189) (("[") reduce 189) (("`") reduce 189) (("{") reduce 189) (("~") reduce 189)) 31490 (((T_ABSTRACT) reduce 102) ((T_ATTRIBUTE) reduce 102) ((T_CLASS) reduce 102) ((T_CONST) reduce 102) ((T_ENUM) reduce 102) ((T_FINAL) reduce 102) ((T_FN) reduce 102) ((T_FUNCTION) reduce 102) ((T_INTERFACE) reduce 102) ((T_READONLY) reduce 102) ((T_STATIC) reduce 102) ((T_TRAIT) reduce 102)) 31491 ((("!") reduce 108) (("\"") reduce 108) (("$") reduce 108) (($) reduce 108) (("(") reduce 108) (("+") reduce 108) (("-") reduce 108) ((";") reduce 108) (("@") reduce 108) ((T_ABSTRACT) reduce 108) ((T_ARRAY) reduce 108) ((T_ARRAY_CAST) reduce 108) ((T_ATTRIBUTE) reduce 108) ((T_BOOL_CAST) reduce 108) ((T_BREAK) reduce 108) ((T_CLASS) reduce 108) ((T_CLASS_C) reduce 108) ((T_CLONE) reduce 108) ((T_CONST) reduce 108) ((T_CONSTANT_ENCAPSED_STRING) reduce 108) ((T_CONTINUE) reduce 108) ((T_DEC) reduce 108) ((T_DECLARE) reduce 108) ((T_DIR) reduce 108) ((T_DNUMBER) reduce 108) ((T_DO) reduce 108) ((T_DOUBLE_CAST) reduce 108) ((T_ECHO) reduce 108) ((T_EMPTY) reduce 108) ((T_ENUM) reduce 108) ((T_EVAL) reduce 108) ((T_EXIT) reduce 108) ((T_FILE) reduce 108) ((T_FINAL) reduce 108) ((T_FN) reduce 108) ((T_FOR) reduce 108) ((T_FOREACH) reduce 108) ((T_FUNCTION) reduce 108) ((T_FUNC_C) reduce 108) ((T_GLOBAL) reduce 108) ((T_GOTO) reduce 108) ((T_HALT_COMPILER) reduce 108) ((T_IF) reduce 108) ((T_INC) reduce 108) ((T_INCLUDE) reduce 108) ((T_INCLUDE_ONCE) reduce 108) ((T_INLINE_HTML) reduce 108) ((T_INTERFACE) reduce 108) ((T_INT_CAST) reduce 108) ((T_ISSET) reduce 108) ((T_LINE) reduce 108) ((T_LIST) reduce 108) ((T_LNUMBER) reduce 108) ((T_MATCH) reduce 108) ((T_METHOD_C) reduce 108) ((T_NAMESPACE) reduce 108) ((T_NAME_FULLY_QUALIFIED) reduce 108) ((T_NAME_QUALIFIED) reduce 108) ((T_NAME_RELATIVE) reduce 108) ((T_NEW) reduce 108) ((T_NS_C) reduce 108) ((T_OBJECT_CAST) reduce 108) ((T_PRINT) reduce 108) ((T_PROPERTY_C) reduce 108) ((T_READONLY) reduce 108) ((T_REQUIRE) reduce 108) ((T_REQUIRE_ONCE) reduce 108) ((T_RETURN) reduce 108) ((T_START_HEREDOC) reduce 108) ((T_STATIC) reduce 108) ((T_STRING) reduce 108) ((T_STRING_CAST) reduce 108) ((T_SWITCH) reduce 108) ((T_THROW) reduce 108) ((T_TRAIT) reduce 108) ((T_TRAIT_C) reduce 108) ((T_TRY) reduce 108) ((T_UNSET) reduce 108) ((T_UNSET_CAST) reduce 108) ((T_USE) reduce 108) ((T_VARIABLE) reduce 108) ((T_VOID_CAST) reduce 108) ((T_WHILE) reduce 108) ((T_YIELD) reduce 108) ((T_YIELD_FROM) reduce 108) (("[") reduce 108) (("`") reduce 108) (("{") reduce 108) (("~") reduce 108)) 31492 ((("!") reduce 112) (("\"") reduce 112) (($) reduce 112) (("$") reduce 112) (("(") reduce 112) (("+") reduce 112) (("-") reduce 112) ((";") reduce 112) (("@") reduce 112) ((T_ABSTRACT) reduce 112) ((T_ARRAY) reduce 112) ((T_ARRAY_CAST) reduce 112) ((T_ATTRIBUTE) reduce 112) ((T_BOOL_CAST) reduce 112) ((T_BREAK) reduce 112) ((T_CLASS) reduce 112) ((T_CLASS_C) reduce 112) ((T_CLONE) reduce 112) ((T_CONST) reduce 112) ((T_CONSTANT_ENCAPSED_STRING) reduce 112) ((T_CONTINUE) reduce 112) ((T_DEC) reduce 112) ((T_DECLARE) reduce 112) ((T_DIR) reduce 112) ((T_DNUMBER) reduce 112) ((T_DO) reduce 112) ((T_DOUBLE_CAST) reduce 112) ((T_ECHO) reduce 112) ((T_EMPTY) reduce 112) ((T_ENUM) reduce 112) ((T_EVAL) reduce 112) ((T_EXIT) reduce 112) ((T_FILE) reduce 112) ((T_FINAL) reduce 112) ((T_FN) reduce 112) ((T_FOR) reduce 112) ((T_FOREACH) reduce 112) ((T_FUNCTION) reduce 112) ((T_FUNC_C) reduce 112) ((T_GLOBAL) reduce 112) ((T_GOTO) reduce 112) ((T_HALT_COMPILER) reduce 112) ((T_IF) reduce 112) ((T_INC) reduce 112) ((T_INCLUDE) reduce 112) ((T_INCLUDE_ONCE) reduce 112) ((T_INLINE_HTML) reduce 112) ((T_INTERFACE) reduce 112) ((T_INT_CAST) reduce 112) ((T_ISSET) reduce 112) ((T_LINE) reduce 112) ((T_LIST) reduce 112) ((T_LNUMBER) reduce 112) ((T_MATCH) reduce 112) ((T_METHOD_C) reduce 112) ((T_NAMESPACE) reduce 112) ((T_NAME_FULLY_QUALIFIED) reduce 112) ((T_NAME_QUALIFIED) reduce 112) ((T_NAME_RELATIVE) reduce 112) ((T_NEW) reduce 112) ((T_NS_C) reduce 112) ((T_OBJECT_CAST) reduce 112) ((T_PRINT) reduce 112) ((T_PROPERTY_C) reduce 112) ((T_READONLY) reduce 112) ((T_REQUIRE) reduce 112) ((T_REQUIRE_ONCE) reduce 112) ((T_RETURN) reduce 112) ((T_START_HEREDOC) reduce 112) ((T_STATIC) reduce 112) ((T_STRING) reduce 112) ((T_STRING_CAST) reduce 112) ((T_SWITCH) reduce 112) ((T_THROW) reduce 112) ((T_TRAIT) reduce 112) ((T_TRAIT_C) reduce 112) ((T_TRY) reduce 112) ((T_UNSET) reduce 112) ((T_UNSET_CAST) reduce 112) ((T_USE) reduce 112) ((T_VARIABLE) reduce 112) ((T_VOID_CAST) reduce 112) ((T_WHILE) reduce 112) ((T_YIELD) reduce 112) ((T_YIELD_FROM) reduce 112) (("[") reduce 112) (("`") reduce 112) (("{") reduce 112) (("~") reduce 112)) 31493 ((("!") reduce 104) (("\"") reduce 104) (($) reduce 104) (("$") reduce 104) (("(") reduce 104) (("+") reduce 104) (("-") reduce 104) ((";") reduce 104) (("@") reduce 104) ((T_ABSTRACT) reduce 104) ((T_ARRAY) reduce 104) ((T_ARRAY_CAST) reduce 104) ((T_ATTRIBUTE) reduce 104) ((T_BOOL_CAST) reduce 104) ((T_BREAK) reduce 104) ((T_CLASS) reduce 104) ((T_CLASS_C) reduce 104) ((T_CLONE) reduce 104) ((T_CONST) reduce 104) ((T_CONSTANT_ENCAPSED_STRING) reduce 104) ((T_CONTINUE) reduce 104) ((T_DEC) reduce 104) ((T_DECLARE) reduce 104) ((T_DIR) reduce 104) ((T_DNUMBER) reduce 104) ((T_DO) reduce 104) ((T_DOUBLE_CAST) reduce 104) ((T_ECHO) reduce 104) ((T_EMPTY) reduce 104) ((T_ENUM) reduce 104) ((T_EVAL) reduce 104) ((T_EXIT) reduce 104) ((T_FILE) reduce 104) ((T_FINAL) reduce 104) ((T_FN) reduce 104) ((T_FOR) reduce 104) ((T_FOREACH) reduce 104) ((T_FUNCTION) reduce 104) ((T_FUNC_C) reduce 104) ((T_GLOBAL) reduce 104) ((T_GOTO) reduce 104) ((T_HALT_COMPILER) reduce 104) ((T_IF) reduce 104) ((T_INC) reduce 104) ((T_INCLUDE) reduce 104) ((T_INCLUDE_ONCE) reduce 104) ((T_INLINE_HTML) reduce 104) ((T_INTERFACE) reduce 104) ((T_INT_CAST) reduce 104) ((T_ISSET) reduce 104) ((T_LINE) reduce 104) ((T_LIST) reduce 104) ((T_LNUMBER) reduce 104) ((T_MATCH) reduce 104) ((T_METHOD_C) reduce 104) ((T_NAMESPACE) reduce 104) ((T_NAME_FULLY_QUALIFIED) reduce 104) ((T_NAME_QUALIFIED) reduce 104) ((T_NAME_RELATIVE) reduce 104) ((T_NEW) reduce 104) ((T_NS_C) reduce 104) ((T_OBJECT_CAST) reduce 104) ((T_PRINT) reduce 104) ((T_PROPERTY_C) reduce 104) ((T_READONLY) reduce 104) ((T_REQUIRE) reduce 104) ((T_REQUIRE_ONCE) reduce 104) ((T_RETURN) reduce 104) ((T_START_HEREDOC) reduce 104) ((T_STATIC) reduce 104) ((T_STRING) reduce 104) ((T_STRING_CAST) reduce 104) ((T_SWITCH) reduce 104) ((T_THROW) reduce 104) ((T_TRAIT) reduce 104) ((T_TRAIT_C) reduce 104) ((T_TRY) reduce 104) ((T_UNSET) reduce 104) ((T_UNSET_CAST) reduce 104) ((T_USE) reduce 104) ((T_VARIABLE) reduce 104) ((T_VOID_CAST) reduce 104) ((T_WHILE) reduce 104) ((T_YIELD) reduce 104) ((T_YIELD_FROM) reduce 104) (("[") reduce 104) (("`") reduce 104) (("{") reduce 104) (("~") reduce 104)) 31495 ((("!") reduce 107) (("\"") reduce 107) (($) reduce 107) (("$") reduce 107) (("(") reduce 107) (("+") reduce 107) (("-") reduce 107) ((";") reduce 107) (("@") reduce 107) ((T_ABSTRACT) reduce 107) ((T_ARRAY) reduce 107) ((T_ARRAY_CAST) reduce 107) ((T_ATTRIBUTE) reduce 107) ((T_BOOL_CAST) reduce 107) ((T_BREAK) reduce 107) ((T_CLASS) reduce 107) ((T_CLASS_C) reduce 107) ((T_CLONE) reduce 107) ((T_CONST) reduce 107) ((T_CONSTANT_ENCAPSED_STRING) reduce 107) ((T_CONTINUE) reduce 107) ((T_DEC) reduce 107) ((T_DECLARE) reduce 107) ((T_DIR) reduce 107) ((T_DNUMBER) reduce 107) ((T_DO) reduce 107) ((T_DOUBLE_CAST) reduce 107) ((T_ECHO) reduce 107) ((T_EMPTY) reduce 107) ((T_ENUM) reduce 107) ((T_EVAL) reduce 107) ((T_EXIT) reduce 107) ((T_FILE) reduce 107) ((T_FINAL) reduce 107) ((T_FN) reduce 107) ((T_FOR) reduce 107) ((T_FOREACH) reduce 107) ((T_FUNCTION) reduce 107) ((T_FUNC_C) reduce 107) ((T_GLOBAL) reduce 107) ((T_GOTO) reduce 107) ((T_HALT_COMPILER) reduce 107) ((T_IF) reduce 107) ((T_INC) reduce 107) ((T_INCLUDE) reduce 107) ((T_INCLUDE_ONCE) reduce 107) ((T_INLINE_HTML) reduce 107) ((T_INTERFACE) reduce 107) ((T_INT_CAST) reduce 107) ((T_ISSET) reduce 107) ((T_LINE) reduce 107) ((T_LIST) reduce 107) ((T_LNUMBER) reduce 107) ((T_MATCH) reduce 107) ((T_METHOD_C) reduce 107) ((T_NAMESPACE) reduce 107) ((T_NAME_FULLY_QUALIFIED) reduce 107) ((T_NAME_QUALIFIED) reduce 107) ((T_NAME_RELATIVE) reduce 107) ((T_NEW) reduce 107) ((T_NS_C) reduce 107) ((T_OBJECT_CAST) reduce 107) ((T_PRINT) reduce 107) ((T_PROPERTY_C) reduce 107) ((T_READONLY) reduce 107) ((T_REQUIRE) reduce 107) ((T_REQUIRE_ONCE) reduce 107) ((T_RETURN) reduce 107) ((T_START_HEREDOC) reduce 107) ((T_STATIC) reduce 107) ((T_STRING) reduce 107) ((T_STRING_CAST) reduce 107) ((T_SWITCH) reduce 107) ((T_THROW) reduce 107) ((T_TRAIT) reduce 107) ((T_TRAIT_C) reduce 107) ((T_TRY) reduce 107) ((T_UNSET) reduce 107) ((T_UNSET_CAST) reduce 107) ((T_USE) reduce 107) ((T_VARIABLE) reduce 107) ((T_VOID_CAST) reduce 107) ((T_WHILE) reduce 107) ((T_YIELD) reduce 107) ((T_YIELD_FROM) reduce 107) (("[") reduce 107) (("`") reduce 107) (("{") reduce 107) (("~") reduce 107)) 31497 ((("!") reduce 103) (("\"") reduce 103) (($) reduce 103) (("$") reduce 103) (("(") reduce 103) (("+") reduce 103) (("-") reduce 103) ((";") reduce 103) (("@") reduce 103) ((T_ABSTRACT) reduce 103) ((T_ARRAY) reduce 103) ((T_ARRAY_CAST) reduce 103) ((T_ATTRIBUTE) reduce 103) ((T_BOOL_CAST) reduce 103) ((T_BREAK) reduce 103) ((T_CLASS) reduce 103) ((T_CLASS_C) reduce 103) ((T_CLONE) reduce 103) ((T_CONST) reduce 103) ((T_CONSTANT_ENCAPSED_STRING) reduce 103) ((T_CONTINUE) reduce 103) ((T_DEC) reduce 103) ((T_DECLARE) reduce 103) ((T_DIR) reduce 103) ((T_DNUMBER) reduce 103) ((T_DO) reduce 103) ((T_DOUBLE_CAST) reduce 103) ((T_ECHO) reduce 103) ((T_EMPTY) reduce 103) ((T_ENUM) reduce 103) ((T_EVAL) reduce 103) ((T_EXIT) reduce 103) ((T_FILE) reduce 103) ((T_FINAL) reduce 103) ((T_FN) reduce 103) ((T_FOR) reduce 103) ((T_FOREACH) reduce 103) ((T_FUNCTION) reduce 103) ((T_FUNC_C) reduce 103) ((T_GLOBAL) reduce 103) ((T_GOTO) reduce 103) ((T_HALT_COMPILER) reduce 103) ((T_IF) reduce 103) ((T_INC) reduce 103) ((T_INCLUDE) reduce 103) ((T_INCLUDE_ONCE) reduce 103) ((T_INLINE_HTML) reduce 103) ((T_INTERFACE) reduce 103) ((T_INT_CAST) reduce 103) ((T_ISSET) reduce 103) ((T_LINE) reduce 103) ((T_LIST) reduce 103) ((T_LNUMBER) reduce 103) ((T_MATCH) reduce 103) ((T_METHOD_C) reduce 103) ((T_NAMESPACE) reduce 103) ((T_NAME_FULLY_QUALIFIED) reduce 103) ((T_NAME_QUALIFIED) reduce 103) ((T_NAME_RELATIVE) reduce 103) ((T_NEW) reduce 103) ((T_NS_C) reduce 103) ((T_OBJECT_CAST) reduce 103) ((T_PRINT) reduce 103) ((T_PROPERTY_C) reduce 103) ((T_READONLY) reduce 103) ((T_REQUIRE) reduce 103) ((T_REQUIRE_ONCE) reduce 103) ((T_RETURN) reduce 103) ((T_START_HEREDOC) reduce 103) ((T_STATIC) reduce 103) ((T_STRING) reduce 103) ((T_STRING_CAST) reduce 103) ((T_SWITCH) reduce 103) ((T_THROW) reduce 103) ((T_TRAIT) reduce 103) ((T_TRAIT_C) reduce 103) ((T_TRY) reduce 103) ((T_UNSET) reduce 103) ((T_UNSET_CAST) reduce 103) ((T_USE) reduce 103) ((T_VARIABLE) reduce 103) ((T_VOID_CAST) reduce 103) ((T_WHILE) reduce 103) ((T_YIELD) reduce 103) ((T_YIELD_FROM) reduce 103) (("[") reduce 103) (("`") reduce 103) (("{") reduce 103) (("~") reduce 103)) 31498 ((("!") reduce 106) (("\"") reduce 106) (($) reduce 106) (("$") reduce 106) (("(") reduce 106) (("+") reduce 106) (("-") reduce 106) ((";") reduce 106) (("@") reduce 106) ((T_ABSTRACT) reduce 106) ((T_ARRAY) reduce 106) ((T_ARRAY_CAST) reduce 106) ((T_ATTRIBUTE) reduce 106) ((T_BOOL_CAST) reduce 106) ((T_BREAK) reduce 106) ((T_CLASS) reduce 106) ((T_CLASS_C) reduce 106) ((T_CLONE) reduce 106) ((T_CONST) reduce 106) ((T_CONSTANT_ENCAPSED_STRING) reduce 106) ((T_CONTINUE) reduce 106) ((T_DEC) reduce 106) ((T_DECLARE) reduce 106) ((T_DIR) reduce 106) ((T_DNUMBER) reduce 106) ((T_DO) reduce 106) ((T_DOUBLE_CAST) reduce 106) ((T_ECHO) reduce 106) ((T_EMPTY) reduce 106) ((T_ENUM) reduce 106) ((T_EVAL) reduce 106) ((T_EXIT) reduce 106) ((T_FILE) reduce 106) ((T_FINAL) reduce 106) ((T_FN) reduce 106) ((T_FOR) reduce 106) ((T_FOREACH) reduce 106) ((T_FUNCTION) reduce 106) ((T_FUNC_C) reduce 106) ((T_GLOBAL) reduce 106) ((T_GOTO) reduce 106) ((T_HALT_COMPILER) reduce 106) ((T_IF) reduce 106) ((T_INC) reduce 106) ((T_INCLUDE) reduce 106) ((T_INCLUDE_ONCE) reduce 106) ((T_INLINE_HTML) reduce 106) ((T_INTERFACE) reduce 106) ((T_INT_CAST) reduce 106) ((T_ISSET) reduce 106) ((T_LINE) reduce 106) ((T_LIST) reduce 106) ((T_LNUMBER) reduce 106) ((T_MATCH) reduce 106) ((T_METHOD_C) reduce 106) ((T_NAMESPACE) reduce 106) ((T_NAME_FULLY_QUALIFIED) reduce 106) ((T_NAME_QUALIFIED) reduce 106) ((T_NAME_RELATIVE) reduce 106) ((T_NEW) reduce 106) ((T_NS_C) reduce 106) ((T_OBJECT_CAST) reduce 106) ((T_PRINT) reduce 106) ((T_PROPERTY_C) reduce 106) ((T_READONLY) reduce 106) ((T_REQUIRE) reduce 106) ((T_REQUIRE_ONCE) reduce 106) ((T_RETURN) reduce 106) ((T_START_HEREDOC) reduce 106) ((T_STATIC) reduce 106) ((T_STRING) reduce 106) ((T_STRING_CAST) reduce 106) ((T_SWITCH) reduce 106) ((T_THROW) reduce 106) ((T_TRAIT) reduce 106) ((T_TRAIT_C) reduce 106) ((T_TRY) reduce 106) ((T_UNSET) reduce 106) ((T_UNSET_CAST) reduce 106) ((T_USE) reduce 106) ((T_VARIABLE) reduce 106) ((T_VOID_CAST) reduce 106) ((T_WHILE) reduce 106) ((T_YIELD) reduce 106) ((T_YIELD_FROM) reduce 106) (("[") reduce 106) (("`") reduce 106) (("{") reduce 106) (("~") reduce 106)) 31499 ((("!") reduce 105) (("\"") reduce 105) (($) reduce 105) (("$") reduce 105) (("(") reduce 105) (("+") reduce 105) (("-") reduce 105) ((";") reduce 105) (("@") reduce 105) ((T_ABSTRACT) reduce 105) ((T_ARRAY) reduce 105) ((T_ARRAY_CAST) reduce 105) ((T_ATTRIBUTE) reduce 105) ((T_BOOL_CAST) reduce 105) ((T_BREAK) reduce 105) ((T_CLASS) reduce 105) ((T_CLASS_C) reduce 105) ((T_CLONE) reduce 105) ((T_CONST) reduce 105) ((T_CONSTANT_ENCAPSED_STRING) reduce 105) ((T_CONTINUE) reduce 105) ((T_DEC) reduce 105) ((T_DECLARE) reduce 105) ((T_DIR) reduce 105) ((T_DNUMBER) reduce 105) ((T_DO) reduce 105) ((T_DOUBLE_CAST) reduce 105) ((T_ECHO) reduce 105) ((T_EMPTY) reduce 105) ((T_ENUM) reduce 105) ((T_EVAL) reduce 105) ((T_EXIT) reduce 105) ((T_FILE) reduce 105) ((T_FINAL) reduce 105) ((T_FN) reduce 105) ((T_FOR) reduce 105) ((T_FOREACH) reduce 105) ((T_FUNCTION) reduce 105) ((T_FUNC_C) reduce 105) ((T_GLOBAL) reduce 105) ((T_GOTO) reduce 105) ((T_HALT_COMPILER) reduce 105) ((T_IF) reduce 105) ((T_INC) reduce 105) ((T_INCLUDE) reduce 105) ((T_INCLUDE_ONCE) reduce 105) ((T_INLINE_HTML) reduce 105) ((T_INTERFACE) reduce 105) ((T_INT_CAST) reduce 105) ((T_ISSET) reduce 105) ((T_LINE) reduce 105) ((T_LIST) reduce 105) ((T_LNUMBER) reduce 105) ((T_MATCH) reduce 105) ((T_METHOD_C) reduce 105) ((T_NAMESPACE) reduce 105) ((T_NAME_FULLY_QUALIFIED) reduce 105) ((T_NAME_QUALIFIED) reduce 105) ((T_NAME_RELATIVE) reduce 105) ((T_NEW) reduce 105) ((T_NS_C) reduce 105) ((T_OBJECT_CAST) reduce 105) ((T_PRINT) reduce 105) ((T_PROPERTY_C) reduce 105) ((T_READONLY) reduce 105) ((T_REQUIRE) reduce 105) ((T_REQUIRE_ONCE) reduce 105) ((T_RETURN) reduce 105) ((T_START_HEREDOC) reduce 105) ((T_STATIC) reduce 105) ((T_STRING) reduce 105) ((T_STRING_CAST) reduce 105) ((T_SWITCH) reduce 105) ((T_THROW) reduce 105) ((T_TRAIT) reduce 105) ((T_TRAIT_C) reduce 105) ((T_TRY) reduce 105) ((T_UNSET) reduce 105) ((T_UNSET_CAST) reduce 105) ((T_USE) reduce 105) ((T_VARIABLE) reduce 105) ((T_VOID_CAST) reduce 105) ((T_WHILE) reduce 105) ((T_YIELD) reduce 105) ((T_YIELD_FROM) reduce 105) (("[") reduce 105) (("`") reduce 105) (("{") reduce 105) (("~") reduce 105)) 31510 ((("!") reduce 494) (("\"") reduce 494) (("$") reduce 494) (($) reduce 494) (("(") reduce 494) (("+") reduce 494) (("-") reduce 494) ((";") reduce 494) (("@") reduce 494) ((T_ABSTRACT) reduce 494) ((T_ARRAY) reduce 494) ((T_ARRAY_CAST) reduce 494) ((T_ATTRIBUTE) reduce 494) ((T_BOOL_CAST) reduce 494) ((T_BREAK) reduce 494) ((T_CLASS) reduce 494) ((T_CLASS_C) reduce 494) ((T_CLONE) reduce 494) ((T_CONST) reduce 494) ((T_CONSTANT_ENCAPSED_STRING) reduce 494) ((T_CONTINUE) reduce 494) ((T_DEC) reduce 494) ((T_DECLARE) reduce 494) ((T_DIR) reduce 494) ((T_DNUMBER) reduce 494) ((T_DO) reduce 494) ((T_DOUBLE_CAST) reduce 494) ((T_ECHO) reduce 494) ((T_EMPTY) reduce 494) ((T_ENUM) reduce 494) ((T_EVAL) reduce 494) ((T_EXIT) reduce 494) ((T_FILE) reduce 494) ((T_FINAL) reduce 494) ((T_FN) reduce 494) ((T_FOR) reduce 494) ((T_FOREACH) reduce 494) ((T_FUNCTION) reduce 494) ((T_FUNC_C) reduce 494) ((T_GLOBAL) reduce 494) ((T_GOTO) reduce 494) ((T_HALT_COMPILER) reduce 494) ((T_IF) reduce 494) ((T_INC) reduce 494) ((T_INCLUDE) reduce 494) ((T_INCLUDE_ONCE) reduce 494) ((T_INLINE_HTML) reduce 494) ((T_INTERFACE) reduce 494) ((T_INT_CAST) reduce 494) ((T_ISSET) reduce 494) ((T_LINE) reduce 494) ((T_LIST) reduce 494) ((T_LNUMBER) reduce 494) ((T_MATCH) reduce 494) ((T_METHOD_C) reduce 494) ((T_NAMESPACE) reduce 494) ((T_NAME_FULLY_QUALIFIED) reduce 494) ((T_NAME_QUALIFIED) reduce 494) ((T_NAME_RELATIVE) reduce 494) ((T_NEW) reduce 494) ((T_NS_C) reduce 494) ((T_OBJECT_CAST) reduce 494) ((T_PRINT) reduce 494) ((T_PROPERTY_C) reduce 494) ((T_READONLY) reduce 494) ((T_REQUIRE) reduce 494) ((T_REQUIRE_ONCE) reduce 494) ((T_RETURN) reduce 494) ((T_START_HEREDOC) reduce 494) ((T_STATIC) reduce 494) ((T_STRING) reduce 494) ((T_STRING_CAST) reduce 494) ((T_SWITCH) reduce 494) ((T_THROW) reduce 494) ((T_TRAIT) reduce 494) ((T_TRAIT_C) reduce 494) ((T_TRY) reduce 494) ((T_UNSET) reduce 494) ((T_UNSET_CAST) reduce 494) ((T_USE) reduce 494) ((T_VARIABLE) reduce 494) ((T_VOID_CAST) reduce 494) ((T_WHILE) reduce 494) ((T_YIELD) reduce 494) ((T_YIELD_FROM) reduce 494) (("[") reduce 494) (("`") reduce 494) (("{") reduce 494) (("~") reduce 494)) 31511 ((("!") reduce 184) (("\"") reduce 184) (("$") reduce 184) (($) reduce 184) (("(") reduce 184) (("+") reduce 184) (("-") reduce 184) ((";") reduce 184) (("@") reduce 184) ((T_ABSTRACT) reduce 184) ((T_ARRAY) reduce 184) ((T_ARRAY_CAST) reduce 184) ((T_ATTRIBUTE) reduce 184) ((T_BOOL_CAST) reduce 184) ((T_BREAK) reduce 184) ((T_CLASS) reduce 184) ((T_CLASS_C) reduce 184) ((T_CLONE) reduce 184) ((T_CONST) reduce 184) ((T_CONSTANT_ENCAPSED_STRING) reduce 184) ((T_CONTINUE) reduce 184) ((T_DEC) reduce 184) ((T_DECLARE) reduce 184) ((T_DIR) reduce 184) ((T_DNUMBER) reduce 184) ((T_DO) reduce 184) ((T_DOUBLE_CAST) reduce 184) ((T_ECHO) reduce 184) ((T_EMPTY) reduce 184) ((T_ENUM) reduce 184) ((T_EVAL) reduce 184) ((T_EXIT) reduce 184) ((T_FILE) reduce 184) ((T_FINAL) reduce 184) ((T_FN) reduce 184) ((T_FOR) reduce 184) ((T_FOREACH) reduce 184) ((T_FUNCTION) reduce 184) ((T_FUNC_C) reduce 184) ((T_GLOBAL) reduce 184) ((T_GOTO) reduce 184) ((T_HALT_COMPILER) reduce 184) ((T_IF) reduce 184) ((T_INC) reduce 184) ((T_INCLUDE) reduce 184) ((T_INCLUDE_ONCE) reduce 184) ((T_INLINE_HTML) reduce 184) ((T_INTERFACE) reduce 184) ((T_INT_CAST) reduce 184) ((T_ISSET) reduce 184) ((T_LINE) reduce 184) ((T_LIST) reduce 184) ((T_LNUMBER) reduce 184) ((T_MATCH) reduce 184) ((T_METHOD_C) reduce 184) ((T_NAMESPACE) reduce 184) ((T_NAME_FULLY_QUALIFIED) reduce 184) ((T_NAME_QUALIFIED) reduce 184) ((T_NAME_RELATIVE) reduce 184) ((T_NEW) reduce 184) ((T_NS_C) reduce 184) ((T_OBJECT_CAST) reduce 184) ((T_PRINT) reduce 184) ((T_PROPERTY_C) reduce 184) ((T_READONLY) reduce 184) ((T_REQUIRE) reduce 184) ((T_REQUIRE_ONCE) reduce 184) ((T_RETURN) reduce 184) ((T_START_HEREDOC) reduce 184) ((T_STATIC) reduce 184) ((T_STRING) reduce 184) ((T_STRING_CAST) reduce 184) ((T_SWITCH) reduce 184) ((T_THROW) reduce 184) ((T_TRAIT) reduce 184) ((T_TRAIT_C) reduce 184) ((T_TRY) reduce 184) ((T_UNSET) reduce 184) ((T_UNSET_CAST) reduce 184) ((T_USE) reduce 184) ((T_VARIABLE) reduce 184) ((T_VOID_CAST) reduce 184) ((T_WHILE) reduce 184) ((T_YIELD) reduce 184) ((T_YIELD_FROM) reduce 184) (("[") reduce 184) (("`") reduce 184) (("{") reduce 184) (("~") reduce 184)) 31519 ((("!") reduce 189) (("\"") reduce 189) (("$") reduce 189) (($) reduce 189) (("(") reduce 189) (("+") reduce 189) (("-") reduce 189) ((";") reduce 189) (("@") reduce 189) ((T_ABSTRACT) reduce 189) ((T_ARRAY) reduce 189) ((T_ARRAY_CAST) reduce 189) ((T_ATTRIBUTE) reduce 189) ((T_BOOL_CAST) reduce 189) ((T_BREAK) reduce 189) ((T_CLASS) reduce 189) ((T_CLASS_C) reduce 189) ((T_CLONE) reduce 189) ((T_CONST) reduce 189) ((T_CONSTANT_ENCAPSED_STRING) reduce 189) ((T_CONTINUE) reduce 189) ((T_DEC) reduce 189) ((T_DECLARE) reduce 189) ((T_DIR) reduce 189) ((T_DNUMBER) reduce 189) ((T_DO) reduce 189) ((T_DOUBLE_CAST) reduce 189) ((T_ECHO) reduce 189) ((T_EMPTY) reduce 189) ((T_ENUM) reduce 189) ((T_EVAL) reduce 189) ((T_EXIT) reduce 189) ((T_FILE) reduce 189) ((T_FINAL) reduce 189) ((T_FN) reduce 189) ((T_FOR) reduce 189) ((T_FOREACH) reduce 189) ((T_FUNCTION) reduce 189) ((T_FUNC_C) reduce 189) ((T_GLOBAL) reduce 189) ((T_GOTO) reduce 189) ((T_HALT_COMPILER) reduce 189) ((T_IF) reduce 189) ((T_INC) reduce 189) ((T_INCLUDE) reduce 189) ((T_INCLUDE_ONCE) reduce 189) ((T_INLINE_HTML) reduce 189) ((T_INTERFACE) reduce 189) ((T_INT_CAST) reduce 189) ((T_ISSET) reduce 189) ((T_LINE) reduce 189) ((T_LIST) reduce 189) ((T_LNUMBER) reduce 189) ((T_MATCH) reduce 189) ((T_METHOD_C) reduce 189) ((T_NAMESPACE) reduce 189) ((T_NAME_FULLY_QUALIFIED) reduce 189) ((T_NAME_QUALIFIED) reduce 189) ((T_NAME_RELATIVE) reduce 189) ((T_NEW) reduce 189) ((T_NS_C) reduce 189) ((T_OBJECT_CAST) reduce 189) ((T_PRINT) reduce 189) ((T_PROPERTY_C) reduce 189) ((T_READONLY) reduce 189) ((T_REQUIRE) reduce 189) ((T_REQUIRE_ONCE) reduce 189) ((T_RETURN) reduce 189) ((T_START_HEREDOC) reduce 189) ((T_STATIC) reduce 189) ((T_STRING) reduce 189) ((T_STRING_CAST) reduce 189) ((T_SWITCH) reduce 189) ((T_THROW) reduce 189) ((T_TRAIT) reduce 189) ((T_TRAIT_C) reduce 189) ((T_TRY) reduce 189) ((T_UNSET) reduce 189) ((T_UNSET_CAST) reduce 189) ((T_USE) reduce 189) ((T_VARIABLE) reduce 189) ((T_VOID_CAST) reduce 189) ((T_WHILE) reduce 189) ((T_YIELD) reduce 189) ((T_YIELD_FROM) reduce 189) (("[") reduce 189) (("`") reduce 189) (("{") reduce 189) (("~") reduce 189)) 31524 ((("!") reduce 200) (("\"") reduce 200) (($) reduce 200) (("$") reduce 200) (("(") reduce 200) (("+") reduce 200) (("-") reduce 200) ((";") reduce 200) (("@") reduce 200) ((T_ABSTRACT) reduce 200) ((T_ARRAY) reduce 200) ((T_ARRAY_CAST) reduce 200) ((T_ATTRIBUTE) reduce 200) ((T_BOOL_CAST) reduce 200) ((T_BREAK) reduce 200) ((T_CLASS) reduce 200) ((T_CLASS_C) reduce 200) ((T_CLONE) reduce 200) ((T_CONST) reduce 200) ((T_CONSTANT_ENCAPSED_STRING) reduce 200) ((T_CONTINUE) reduce 200) ((T_DEC) reduce 200) ((T_DECLARE) reduce 200) ((T_DIR) reduce 200) ((T_DNUMBER) reduce 200) ((T_DO) reduce 200) ((T_DOUBLE_CAST) reduce 200) ((T_ECHO) reduce 200) ((T_EMPTY) reduce 200) ((T_ENUM) reduce 200) ((T_EVAL) reduce 200) ((T_EXIT) reduce 200) ((T_FILE) reduce 200) ((T_FINAL) reduce 200) ((T_FN) reduce 200) ((T_FOR) reduce 200) ((T_FOREACH) reduce 200) ((T_FUNCTION) reduce 200) ((T_FUNC_C) reduce 200) ((T_GLOBAL) reduce 200) ((T_GOTO) reduce 200) ((T_HALT_COMPILER) reduce 200) ((T_IF) reduce 200) ((T_INC) reduce 200) ((T_INCLUDE) reduce 200) ((T_INCLUDE_ONCE) reduce 200) ((T_INLINE_HTML) reduce 200) ((T_INTERFACE) reduce 200) ((T_INT_CAST) reduce 200) ((T_ISSET) reduce 200) ((T_LINE) reduce 200) ((T_LIST) reduce 200) ((T_LNUMBER) reduce 200) ((T_MATCH) reduce 200) ((T_METHOD_C) reduce 200) ((T_NAMESPACE) reduce 200) ((T_NAME_FULLY_QUALIFIED) reduce 200) ((T_NAME_QUALIFIED) reduce 200) ((T_NAME_RELATIVE) reduce 200) ((T_NEW) reduce 200) ((T_NS_C) reduce 200) ((T_OBJECT_CAST) reduce 200) ((T_PRINT) reduce 200) ((T_PROPERTY_C) reduce 200) ((T_READONLY) reduce 200) ((T_REQUIRE) reduce 200) ((T_REQUIRE_ONCE) reduce 200) ((T_RETURN) reduce 200) ((T_START_HEREDOC) reduce 200) ((T_STATIC) reduce 200) ((T_STRING) reduce 200) ((T_STRING_CAST) reduce 200) ((T_SWITCH) reduce 200) ((T_THROW) reduce 200) ((T_TRAIT) reduce 200) ((T_TRAIT_C) reduce 200) ((T_TRY) reduce 200) ((T_UNSET) reduce 200) ((T_UNSET_CAST) reduce 200) ((T_USE) reduce 200) ((T_VARIABLE) reduce 200) ((T_VOID_CAST) reduce 200) ((T_WHILE) reduce 200) ((T_YIELD) reduce 200) ((T_YIELD_FROM) reduce 200) (("[") reduce 200) (("`") reduce 200) (("{") reduce 200) (("~") reduce 200)) 31530 ((("!") reduce 201) (("\"") reduce 201) (("$") reduce 201) (($) reduce 201) (("(") reduce 201) (("+") reduce 201) (("-") reduce 201) ((";") reduce 201) (("@") reduce 201) ((T_ABSTRACT) reduce 201) ((T_ARRAY) reduce 201) ((T_ARRAY_CAST) reduce 201) ((T_ATTRIBUTE) reduce 201) ((T_BOOL_CAST) reduce 201) ((T_BREAK) reduce 201) ((T_CLASS) reduce 201) ((T_CLASS_C) reduce 201) ((T_CLONE) reduce 201) ((T_CONST) reduce 201) ((T_CONSTANT_ENCAPSED_STRING) reduce 201) ((T_CONTINUE) reduce 201) ((T_DEC) reduce 201) ((T_DECLARE) reduce 201) ((T_DIR) reduce 201) ((T_DNUMBER) reduce 201) ((T_DO) reduce 201) ((T_DOUBLE_CAST) reduce 201) ((T_ECHO) reduce 201) ((T_EMPTY) reduce 201) ((T_ENUM) reduce 201) ((T_EVAL) reduce 201) ((T_EXIT) reduce 201) ((T_FILE) reduce 201) ((T_FINAL) reduce 201) ((T_FN) reduce 201) ((T_FOR) reduce 201) ((T_FOREACH) reduce 201) ((T_FUNCTION) reduce 201) ((T_FUNC_C) reduce 201) ((T_GLOBAL) reduce 201) ((T_GOTO) reduce 201) ((T_HALT_COMPILER) reduce 201) ((T_IF) reduce 201) ((T_INC) reduce 201) ((T_INCLUDE) reduce 201) ((T_INCLUDE_ONCE) reduce 201) ((T_INLINE_HTML) reduce 201) ((T_INTERFACE) reduce 201) ((T_INT_CAST) reduce 201) ((T_ISSET) reduce 201) ((T_LINE) reduce 201) ((T_LIST) reduce 201) ((T_LNUMBER) reduce 201) ((T_MATCH) reduce 201) ((T_METHOD_C) reduce 201) ((T_NAMESPACE) reduce 201) ((T_NAME_FULLY_QUALIFIED) reduce 201) ((T_NAME_QUALIFIED) reduce 201) ((T_NAME_RELATIVE) reduce 201) ((T_NEW) reduce 201) ((T_NS_C) reduce 201) ((T_OBJECT_CAST) reduce 201) ((T_PRINT) reduce 201) ((T_PROPERTY_C) reduce 201) ((T_READONLY) reduce 201) ((T_REQUIRE) reduce 201) ((T_REQUIRE_ONCE) reduce 201) ((T_RETURN) reduce 201) ((T_START_HEREDOC) reduce 201) ((T_STATIC) reduce 201) ((T_STRING) reduce 201) ((T_STRING_CAST) reduce 201) ((T_SWITCH) reduce 201) ((T_THROW) reduce 201) ((T_TRAIT) reduce 201) ((T_TRAIT_C) reduce 201) ((T_TRY) reduce 201) ((T_UNSET) reduce 201) ((T_UNSET_CAST) reduce 201) ((T_USE) reduce 201) ((T_VARIABLE) reduce 201) ((T_VOID_CAST) reduce 201) ((T_WHILE) reduce 201) ((T_YIELD) reduce 201) ((T_YIELD_FROM) reduce 201) (("[") reduce 201) (("`") reduce 201) (("{") reduce 201) (("~") reduce 201)) 31537 ((("!") reduce 202) (("\"") reduce 202) (($) reduce 202) (("$") reduce 202) (("(") reduce 202) (("+") reduce 202) (("-") reduce 202) ((";") reduce 202) (("@") reduce 202) ((T_ABSTRACT) reduce 202) ((T_ARRAY) reduce 202) ((T_ARRAY_CAST) reduce 202) ((T_ATTRIBUTE) reduce 202) ((T_BOOL_CAST) reduce 202) ((T_BREAK) reduce 202) ((T_CLASS) reduce 202) ((T_CLASS_C) reduce 202) ((T_CLONE) reduce 202) ((T_CONST) reduce 202) ((T_CONSTANT_ENCAPSED_STRING) reduce 202) ((T_CONTINUE) reduce 202) ((T_DEC) reduce 202) ((T_DECLARE) reduce 202) ((T_DIR) reduce 202) ((T_DNUMBER) reduce 202) ((T_DO) reduce 202) ((T_DOUBLE_CAST) reduce 202) ((T_ECHO) reduce 202) ((T_EMPTY) reduce 202) ((T_ENUM) reduce 202) ((T_EVAL) reduce 202) ((T_EXIT) reduce 202) ((T_FILE) reduce 202) ((T_FINAL) reduce 202) ((T_FN) reduce 202) ((T_FOR) reduce 202) ((T_FOREACH) reduce 202) ((T_FUNCTION) reduce 202) ((T_FUNC_C) reduce 202) ((T_GLOBAL) reduce 202) ((T_GOTO) reduce 202) ((T_HALT_COMPILER) reduce 202) ((T_IF) reduce 202) ((T_INC) reduce 202) ((T_INCLUDE) reduce 202) ((T_INCLUDE_ONCE) reduce 202) ((T_INLINE_HTML) reduce 202) ((T_INTERFACE) reduce 202) ((T_INT_CAST) reduce 202) ((T_ISSET) reduce 202) ((T_LINE) reduce 202) ((T_LIST) reduce 202) ((T_LNUMBER) reduce 202) ((T_MATCH) reduce 202) ((T_METHOD_C) reduce 202) ((T_NAMESPACE) reduce 202) ((T_NAME_FULLY_QUALIFIED) reduce 202) ((T_NAME_QUALIFIED) reduce 202) ((T_NAME_RELATIVE) reduce 202) ((T_NEW) reduce 202) ((T_NS_C) reduce 202) ((T_OBJECT_CAST) reduce 202) ((T_PRINT) reduce 202) ((T_PROPERTY_C) reduce 202) ((T_READONLY) reduce 202) ((T_REQUIRE) reduce 202) ((T_REQUIRE_ONCE) reduce 202) ((T_RETURN) reduce 202) ((T_START_HEREDOC) reduce 202) ((T_STATIC) reduce 202) ((T_STRING) reduce 202) ((T_STRING_CAST) reduce 202) ((T_SWITCH) reduce 202) ((T_THROW) reduce 202) ((T_TRAIT) reduce 202) ((T_TRAIT_C) reduce 202) ((T_TRY) reduce 202) ((T_UNSET) reduce 202) ((T_UNSET_CAST) reduce 202) ((T_USE) reduce 202) ((T_VARIABLE) reduce 202) ((T_VOID_CAST) reduce 202) ((T_WHILE) reduce 202) ((T_YIELD) reduce 202) ((T_YIELD_FROM) reduce 202) (("[") reduce 202) (("`") reduce 202) (("{") reduce 202) (("~") reduce 202)) 31539 (((",") reduce 140) ((";") reduce 140)) 31542 ((("!") reduce 109) (("\"") reduce 109) (("$") reduce 109) (($) reduce 109) (("(") reduce 109) (("+") reduce 109) (("-") reduce 109) ((";") reduce 109) (("@") reduce 109) ((T_ABSTRACT) reduce 109) ((T_ARRAY) reduce 109) ((T_ARRAY_CAST) reduce 109) ((T_ATTRIBUTE) reduce 109) ((T_BOOL_CAST) reduce 109) ((T_BREAK) reduce 109) ((T_CLASS) reduce 109) ((T_CLASS_C) reduce 109) ((T_CLONE) reduce 109) ((T_CONST) reduce 109) ((T_CONSTANT_ENCAPSED_STRING) reduce 109) ((T_CONTINUE) reduce 109) ((T_DEC) reduce 109) ((T_DECLARE) reduce 109) ((T_DIR) reduce 109) ((T_DNUMBER) reduce 109) ((T_DO) reduce 109) ((T_DOUBLE_CAST) reduce 109) ((T_ECHO) reduce 109) ((T_EMPTY) reduce 109) ((T_ENUM) reduce 109) ((T_EVAL) reduce 109) ((T_EXIT) reduce 109) ((T_FILE) reduce 109) ((T_FINAL) reduce 109) ((T_FN) reduce 109) ((T_FOR) reduce 109) ((T_FOREACH) reduce 109) ((T_FUNCTION) reduce 109) ((T_FUNC_C) reduce 109) ((T_GLOBAL) reduce 109) ((T_GOTO) reduce 109) ((T_HALT_COMPILER) reduce 109) ((T_IF) reduce 109) ((T_INC) reduce 109) ((T_INCLUDE) reduce 109) ((T_INCLUDE_ONCE) reduce 109) ((T_INLINE_HTML) reduce 109) ((T_INTERFACE) reduce 109) ((T_INT_CAST) reduce 109) ((T_ISSET) reduce 109) ((T_LINE) reduce 109) ((T_LIST) reduce 109) ((T_LNUMBER) reduce 109) ((T_MATCH) reduce 109) ((T_METHOD_C) reduce 109) ((T_NAMESPACE) reduce 109) ((T_NAME_FULLY_QUALIFIED) reduce 109) ((T_NAME_QUALIFIED) reduce 109) ((T_NAME_RELATIVE) reduce 109) ((T_NEW) reduce 109) ((T_NS_C) reduce 109) ((T_OBJECT_CAST) reduce 109) ((T_PRINT) reduce 109) ((T_PROPERTY_C) reduce 109) ((T_READONLY) reduce 109) ((T_REQUIRE) reduce 109) ((T_REQUIRE_ONCE) reduce 109) ((T_RETURN) reduce 109) ((T_START_HEREDOC) reduce 109) ((T_STATIC) reduce 109) ((T_STRING) reduce 109) ((T_STRING_CAST) reduce 109) ((T_SWITCH) reduce 109) ((T_THROW) reduce 109) ((T_TRAIT) reduce 109) ((T_TRAIT_C) reduce 109) ((T_TRY) reduce 109) ((T_UNSET) reduce 109) ((T_UNSET_CAST) reduce 109) ((T_USE) reduce 109) ((T_VARIABLE) reduce 109) ((T_VOID_CAST) reduce 109) ((T_WHILE) reduce 109) ((T_YIELD) reduce 109) ((T_YIELD_FROM) reduce 109) (("[") reduce 109) (("`") reduce 109) (("{") reduce 109) (("~") reduce 109)) 31543 (((",") reduce 139) ((";") reduce 139)) 31546 (((",") reduce 385) ((";") reduce 385)) 31553 ((("!") reduce 190) (("\"") reduce 190) (($) reduce 190) (("$") reduce 190) (("(") reduce 190) (("+") reduce 190) (("-") reduce 190) ((";") reduce 190) (("@") reduce 190) ((T_ABSTRACT) reduce 190) ((T_ARRAY) reduce 190) ((T_ARRAY_CAST) reduce 190) ((T_ATTRIBUTE) reduce 190) ((T_BOOL_CAST) reduce 190) ((T_BREAK) reduce 190) ((T_CLASS) reduce 190) ((T_CLASS_C) reduce 190) ((T_CLONE) reduce 190) ((T_CONST) reduce 190) ((T_CONSTANT_ENCAPSED_STRING) reduce 190) ((T_CONTINUE) reduce 190) ((T_DEC) reduce 190) ((T_DECLARE) reduce 190) ((T_DIR) reduce 190) ((T_DNUMBER) reduce 190) ((T_DO) reduce 190) ((T_DOUBLE_CAST) reduce 190) ((T_ECHO) reduce 190) ((T_EMPTY) reduce 190) ((T_ENUM) reduce 190) ((T_EVAL) reduce 190) ((T_EXIT) reduce 190) ((T_FILE) reduce 190) ((T_FINAL) reduce 190) ((T_FN) reduce 190) ((T_FOR) reduce 190) ((T_FOREACH) reduce 190) ((T_FUNCTION) reduce 190) ((T_FUNC_C) reduce 190) ((T_GLOBAL) reduce 190) ((T_GOTO) reduce 190) ((T_HALT_COMPILER) reduce 190) ((T_IF) reduce 190) ((T_INC) reduce 190) ((T_INCLUDE) reduce 190) ((T_INCLUDE_ONCE) reduce 190) ((T_INLINE_HTML) reduce 190) ((T_INTERFACE) reduce 190) ((T_INT_CAST) reduce 190) ((T_ISSET) reduce 190) ((T_LINE) reduce 190) ((T_LIST) reduce 190) ((T_LNUMBER) reduce 190) ((T_MATCH) reduce 190) ((T_METHOD_C) reduce 190) ((T_NAMESPACE) reduce 190) ((T_NAME_FULLY_QUALIFIED) reduce 190) ((T_NAME_QUALIFIED) reduce 190) ((T_NAME_RELATIVE) reduce 190) ((T_NEW) reduce 190) ((T_NS_C) reduce 190) ((T_OBJECT_CAST) reduce 190) ((T_PRINT) reduce 190) ((T_PROPERTY_C) reduce 190) ((T_READONLY) reduce 190) ((T_REQUIRE) reduce 190) ((T_REQUIRE_ONCE) reduce 190) ((T_RETURN) reduce 190) ((T_START_HEREDOC) reduce 190) ((T_STATIC) reduce 190) ((T_STRING) reduce 190) ((T_STRING_CAST) reduce 190) ((T_SWITCH) reduce 190) ((T_THROW) reduce 190) ((T_TRAIT) reduce 190) ((T_TRAIT_C) reduce 190) ((T_TRY) reduce 190) ((T_UNSET) reduce 190) ((T_UNSET_CAST) reduce 190) ((T_USE) reduce 190) ((T_VARIABLE) reduce 190) ((T_VOID_CAST) reduce 190) ((T_WHILE) reduce 190) ((T_YIELD) reduce 190) ((T_YIELD_FROM) reduce 190) (("[") reduce 190) (("`") reduce 190) (("{") reduce 190) (("~") reduce 190)) 31570 ((("!") reduce 250) (("\"") reduce 250) (("$") reduce 250) (($) reduce 250) (("(") reduce 250) (("+") reduce 250) (("-") reduce 250) ((";") reduce 250) (("@") reduce 250) ((T_ABSTRACT) reduce 250) ((T_ARRAY) reduce 250) ((T_ARRAY_CAST) reduce 250) ((T_ATTRIBUTE) reduce 250) ((T_BOOL_CAST) reduce 250) ((T_BREAK) reduce 250) ((T_CLASS) reduce 250) ((T_CLASS_C) reduce 250) ((T_CLONE) reduce 250) ((T_CONST) reduce 250) ((T_CONSTANT_ENCAPSED_STRING) reduce 250) ((T_CONTINUE) reduce 250) ((T_DEC) reduce 250) ((T_DECLARE) reduce 250) ((T_DIR) reduce 250) ((T_DNUMBER) reduce 250) ((T_DO) reduce 250) ((T_DOUBLE_CAST) reduce 250) ((T_ECHO) reduce 250) ((T_EMPTY) reduce 250) ((T_ENUM) reduce 250) ((T_EVAL) reduce 250) ((T_EXIT) reduce 250) ((T_FILE) reduce 250) ((T_FINAL) reduce 250) ((T_FN) reduce 250) ((T_FOR) reduce 250) ((T_FOREACH) reduce 250) ((T_FUNCTION) reduce 250) ((T_FUNC_C) reduce 250) ((T_GLOBAL) reduce 250) ((T_GOTO) reduce 250) ((T_HALT_COMPILER) reduce 250) ((T_IF) reduce 250) ((T_INC) reduce 250) ((T_INCLUDE) reduce 250) ((T_INCLUDE_ONCE) reduce 250) ((T_INLINE_HTML) reduce 250) ((T_INTERFACE) reduce 250) ((T_INT_CAST) reduce 250) ((T_ISSET) reduce 250) ((T_LINE) reduce 250) ((T_LIST) reduce 250) ((T_LNUMBER) reduce 250) ((T_MATCH) reduce 250) ((T_METHOD_C) reduce 250) ((T_NAMESPACE) reduce 250) ((T_NAME_FULLY_QUALIFIED) reduce 250) ((T_NAME_QUALIFIED) reduce 250) ((T_NAME_RELATIVE) reduce 250) ((T_NEW) reduce 250) ((T_NS_C) reduce 250) ((T_OBJECT_CAST) reduce 250) ((T_PRINT) reduce 250) ((T_PROPERTY_C) reduce 250) ((T_READONLY) reduce 250) ((T_REQUIRE) reduce 250) ((T_REQUIRE_ONCE) reduce 250) ((T_RETURN) reduce 250) ((T_START_HEREDOC) reduce 250) ((T_STATIC) reduce 250) ((T_STRING) reduce 250) ((T_STRING_CAST) reduce 250) ((T_SWITCH) reduce 250) ((T_THROW) reduce 250) ((T_TRAIT) reduce 250) ((T_TRAIT_C) reduce 250) ((T_TRY) reduce 250) ((T_UNSET) reduce 250) ((T_UNSET_CAST) reduce 250) ((T_USE) reduce 250) ((T_VARIABLE) reduce 250) ((T_VOID_CAST) reduce 250) ((T_WHILE) reduce 250) ((T_YIELD) reduce 250) ((T_YIELD_FROM) reduce 250) (("[") reduce 250) (("`") reduce 250) (("{") reduce 250) (("~") reduce 250)) 31574 ((("!") reduce 251) (("\"") reduce 251) (($) reduce 251) (("$") reduce 251) (("(") reduce 251) (("+") reduce 251) (("-") reduce 251) ((";") reduce 251) (("@") reduce 251) ((T_ABSTRACT) reduce 251) ((T_ARRAY) reduce 251) ((T_ARRAY_CAST) reduce 251) ((T_ATTRIBUTE) reduce 251) ((T_BOOL_CAST) reduce 251) ((T_BREAK) reduce 251) ((T_CLASS) reduce 251) ((T_CLASS_C) reduce 251) ((T_CLONE) reduce 251) ((T_CONST) reduce 251) ((T_CONSTANT_ENCAPSED_STRING) reduce 251) ((T_CONTINUE) reduce 251) ((T_DEC) reduce 251) ((T_DECLARE) reduce 251) ((T_DIR) reduce 251) ((T_DNUMBER) reduce 251) ((T_DO) reduce 251) ((T_DOUBLE_CAST) reduce 251) ((T_ECHO) reduce 251) ((T_EMPTY) reduce 251) ((T_ENUM) reduce 251) ((T_EVAL) reduce 251) ((T_EXIT) reduce 251) ((T_FILE) reduce 251) ((T_FINAL) reduce 251) ((T_FN) reduce 251) ((T_FOR) reduce 251) ((T_FOREACH) reduce 251) ((T_FUNCTION) reduce 251) ((T_FUNC_C) reduce 251) ((T_GLOBAL) reduce 251) ((T_GOTO) reduce 251) ((T_HALT_COMPILER) reduce 251) ((T_IF) reduce 251) ((T_INC) reduce 251) ((T_INCLUDE) reduce 251) ((T_INCLUDE_ONCE) reduce 251) ((T_INLINE_HTML) reduce 251) ((T_INTERFACE) reduce 251) ((T_INT_CAST) reduce 251) ((T_ISSET) reduce 251) ((T_LINE) reduce 251) ((T_LIST) reduce 251) ((T_LNUMBER) reduce 251) ((T_MATCH) reduce 251) ((T_METHOD_C) reduce 251) ((T_NAMESPACE) reduce 251) ((T_NAME_FULLY_QUALIFIED) reduce 251) ((T_NAME_QUALIFIED) reduce 251) ((T_NAME_RELATIVE) reduce 251) ((T_NEW) reduce 251) ((T_NS_C) reduce 251) ((T_OBJECT_CAST) reduce 251) ((T_PRINT) reduce 251) ((T_PROPERTY_C) reduce 251) ((T_READONLY) reduce 251) ((T_REQUIRE) reduce 251) ((T_REQUIRE_ONCE) reduce 251) ((T_RETURN) reduce 251) ((T_START_HEREDOC) reduce 251) ((T_STATIC) reduce 251) ((T_STRING) reduce 251) ((T_STRING_CAST) reduce 251) ((T_SWITCH) reduce 251) ((T_THROW) reduce 251) ((T_TRAIT) reduce 251) ((T_TRAIT_C) reduce 251) ((T_TRY) reduce 251) ((T_UNSET) reduce 251) ((T_UNSET_CAST) reduce 251) ((T_USE) reduce 251) ((T_VARIABLE) reduce 251) ((T_VOID_CAST) reduce 251) ((T_WHILE) reduce 251) ((T_YIELD) reduce 251) ((T_YIELD_FROM) reduce 251) (("[") reduce 251) (("`") reduce 251) (("{") reduce 251) (("~") reduce 251)) 32120 ((("!") reduce 242) (("\"") reduce 242) (("$") reduce 242) (($) reduce 242) (("(") reduce 242) (("+") reduce 242) (("-") reduce 242) ((";") reduce 242) (("@") reduce 242) ((T_ABSTRACT) reduce 242) ((T_ARRAY) reduce 242) ((T_ARRAY_CAST) reduce 242) ((T_ATTRIBUTE) reduce 242) ((T_BOOL_CAST) reduce 242) ((T_BREAK) reduce 242) ((T_CLASS) reduce 242) ((T_CLASS_C) reduce 242) ((T_CLONE) reduce 242) ((T_CONST) reduce 242) ((T_CONSTANT_ENCAPSED_STRING) reduce 242) ((T_CONTINUE) reduce 242) ((T_DEC) reduce 242) ((T_DECLARE) reduce 242) ((T_DIR) reduce 242) ((T_DNUMBER) reduce 242) ((T_DO) reduce 242) ((T_DOUBLE_CAST) reduce 242) ((T_ECHO) reduce 242) ((T_EMPTY) reduce 242) ((T_ENUM) reduce 242) ((T_EVAL) reduce 242) ((T_EXIT) reduce 242) ((T_FILE) reduce 242) ((T_FINAL) reduce 242) ((T_FN) reduce 242) ((T_FOR) reduce 242) ((T_FOREACH) reduce 242) ((T_FUNCTION) reduce 242) ((T_FUNC_C) reduce 242) ((T_GLOBAL) reduce 242) ((T_GOTO) reduce 242) ((T_HALT_COMPILER) reduce 242) ((T_IF) reduce 242) ((T_INC) reduce 242) ((T_INCLUDE) reduce 242) ((T_INCLUDE_ONCE) reduce 242) ((T_INLINE_HTML) reduce 242) ((T_INTERFACE) reduce 242) ((T_INT_CAST) reduce 242) ((T_ISSET) reduce 242) ((T_LINE) reduce 242) ((T_LIST) reduce 242) ((T_LNUMBER) reduce 242) ((T_MATCH) reduce 242) ((T_METHOD_C) reduce 242) ((T_NAMESPACE) reduce 242) ((T_NAME_FULLY_QUALIFIED) reduce 242) ((T_NAME_QUALIFIED) reduce 242) ((T_NAME_RELATIVE) reduce 242) ((T_NEW) reduce 242) ((T_NS_C) reduce 242) ((T_OBJECT_CAST) reduce 242) ((T_PRINT) reduce 242) ((T_PROPERTY_C) reduce 242) ((T_READONLY) reduce 242) ((T_REQUIRE) reduce 242) ((T_REQUIRE_ONCE) reduce 242) ((T_RETURN) reduce 242) ((T_START_HEREDOC) reduce 242) ((T_STATIC) reduce 242) ((T_STRING) reduce 242) ((T_STRING_CAST) reduce 242) ((T_SWITCH) reduce 242) ((T_THROW) reduce 242) ((T_TRAIT) reduce 242) ((T_TRAIT_C) reduce 242) ((T_TRY) reduce 242) ((T_UNSET) reduce 242) ((T_UNSET_CAST) reduce 242) ((T_USE) reduce 242) ((T_VARIABLE) reduce 242) ((T_VOID_CAST) reduce 242) ((T_WHILE) reduce 242) ((T_YIELD) reduce 242) ((T_YIELD_FROM) reduce 242) (("[") reduce 242) (("`") reduce 242) (("{") reduce 242) (("~") reduce 242)) 32121 ((("!") reduce 150) (("\"") reduce 150) (($) reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONST) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_EMPTY) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAMESPACE) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_USE) reduce 150) ((T_VARIABLE) reduce 150) ((T_VOID_CAST) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 32124 ((("!") reduce 243) (("\"") reduce 243) (($) reduce 243) (("$") reduce 243) (("(") reduce 243) (("+") reduce 243) (("-") reduce 243) ((";") reduce 243) (("@") reduce 243) ((T_ABSTRACT) reduce 243) ((T_ARRAY) reduce 243) ((T_ARRAY_CAST) reduce 243) ((T_ATTRIBUTE) reduce 243) ((T_BOOL_CAST) reduce 243) ((T_BREAK) reduce 243) ((T_CLASS) reduce 243) ((T_CLASS_C) reduce 243) ((T_CLONE) reduce 243) ((T_CONST) reduce 243) ((T_CONSTANT_ENCAPSED_STRING) reduce 243) ((T_CONTINUE) reduce 243) ((T_DEC) reduce 243) ((T_DECLARE) reduce 243) ((T_DIR) reduce 243) ((T_DNUMBER) reduce 243) ((T_DO) reduce 243) ((T_DOUBLE_CAST) reduce 243) ((T_ECHO) reduce 243) ((T_EMPTY) reduce 243) ((T_ENUM) reduce 243) ((T_EVAL) reduce 243) ((T_EXIT) reduce 243) ((T_FILE) reduce 243) ((T_FINAL) reduce 243) ((T_FN) reduce 243) ((T_FOR) reduce 243) ((T_FOREACH) reduce 243) ((T_FUNCTION) reduce 243) ((T_FUNC_C) reduce 243) ((T_GLOBAL) reduce 243) ((T_GOTO) reduce 243) ((T_HALT_COMPILER) reduce 243) ((T_IF) reduce 243) ((T_INC) reduce 243) ((T_INCLUDE) reduce 243) ((T_INCLUDE_ONCE) reduce 243) ((T_INLINE_HTML) reduce 243) ((T_INTERFACE) reduce 243) ((T_INT_CAST) reduce 243) ((T_ISSET) reduce 243) ((T_LINE) reduce 243) ((T_LIST) reduce 243) ((T_LNUMBER) reduce 243) ((T_MATCH) reduce 243) ((T_METHOD_C) reduce 243) ((T_NAMESPACE) reduce 243) ((T_NAME_FULLY_QUALIFIED) reduce 243) ((T_NAME_QUALIFIED) reduce 243) ((T_NAME_RELATIVE) reduce 243) ((T_NEW) reduce 243) ((T_NS_C) reduce 243) ((T_OBJECT_CAST) reduce 243) ((T_PRINT) reduce 243) ((T_PROPERTY_C) reduce 243) ((T_READONLY) reduce 243) ((T_REQUIRE) reduce 243) ((T_REQUIRE_ONCE) reduce 243) ((T_RETURN) reduce 243) ((T_START_HEREDOC) reduce 243) ((T_STATIC) reduce 243) ((T_STRING) reduce 243) ((T_STRING_CAST) reduce 243) ((T_SWITCH) reduce 243) ((T_THROW) reduce 243) ((T_TRAIT) reduce 243) ((T_TRAIT_C) reduce 243) ((T_TRY) reduce 243) ((T_UNSET) reduce 243) ((T_UNSET_CAST) reduce 243) ((T_USE) reduce 243) ((T_VARIABLE) reduce 243) ((T_VOID_CAST) reduce 243) ((T_WHILE) reduce 243) ((T_YIELD) reduce 243) ((T_YIELD_FROM) reduce 243) (("[") reduce 243) (("`") reduce 243) (("{") reduce 243) (("~") reduce 243)) 32126 ((("!") reduce 170) (("\"") reduce 170) (($) reduce 170) (("$") reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONST) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_EMPTY) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAMESPACE) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_USE) reduce 170) ((T_VARIABLE) reduce 170) ((T_VOID_CAST) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("~") reduce 170)) 32127 (((T_NAME_FULLY_QUALIFIED) reduce 122) ((T_NAME_QUALIFIED) reduce 122) ((T_STRING) reduce 122)) 32128 (((T_NAME_FULLY_QUALIFIED) reduce 121) ((T_NAME_QUALIFIED) reduce 121) ((T_STRING) reduce 121)) 32129 (((",") reduce 91) ((";") reduce 91) ((T_AS) reduce 91) ((T_NS_SEPARATOR) reduce 91)) 32130 (((",") reduce 89) ((";") reduce 89) ((T_AS) reduce 89) ((T_NS_SEPARATOR) reduce 89)) 32131 (((",") reduce 88) ((";") reduce 88) ((T_AS) reduce 88) ((T_NS_SEPARATOR) reduce 88)) 32132 (((",") reduce 137) ((";") reduce 137) ((T_AS) shift) ((T_NS_SEPARATOR) shift)) 32134 (((",") reduce 90) ((";") reduce 90) ((T_AS) reduce 90) ((T_NS_SEPARATOR) reduce 90)) 32135 (((",") reduce 132) ((";") reduce 132)) 32137 (((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_STRING) shift)) 32142 ((("!") reduce 120) (("\"") reduce 120) (($) reduce 120) (("$") reduce 120) (("(") reduce 120) (("+") reduce 120) (("-") reduce 120) ((";") reduce 120) (("@") reduce 120) ((T_ABSTRACT) reduce 120) ((T_ARRAY) reduce 120) ((T_ARRAY_CAST) reduce 120) ((T_ATTRIBUTE) reduce 120) ((T_BOOL_CAST) reduce 120) ((T_BREAK) reduce 120) ((T_CLASS) reduce 120) ((T_CLASS_C) reduce 120) ((T_CLONE) reduce 120) ((T_CONST) reduce 120) ((T_CONSTANT_ENCAPSED_STRING) reduce 120) ((T_CONTINUE) reduce 120) ((T_DEC) reduce 120) ((T_DECLARE) reduce 120) ((T_DIR) reduce 120) ((T_DNUMBER) reduce 120) ((T_DO) reduce 120) ((T_DOUBLE_CAST) reduce 120) ((T_ECHO) reduce 120) ((T_EMPTY) reduce 120) ((T_ENUM) reduce 120) ((T_EVAL) reduce 120) ((T_EXIT) reduce 120) ((T_FILE) reduce 120) ((T_FINAL) reduce 120) ((T_FN) reduce 120) ((T_FOR) reduce 120) ((T_FOREACH) reduce 120) ((T_FUNCTION) reduce 120) ((T_FUNC_C) reduce 120) ((T_GLOBAL) reduce 120) ((T_GOTO) reduce 120) ((T_HALT_COMPILER) reduce 120) ((T_IF) reduce 120) ((T_INC) reduce 120) ((T_INCLUDE) reduce 120) ((T_INCLUDE_ONCE) reduce 120) ((T_INLINE_HTML) reduce 120) ((T_INTERFACE) reduce 120) ((T_INT_CAST) reduce 120) ((T_ISSET) reduce 120) ((T_LINE) reduce 120) ((T_LIST) reduce 120) ((T_LNUMBER) reduce 120) ((T_MATCH) reduce 120) ((T_METHOD_C) reduce 120) ((T_NAMESPACE) reduce 120) ((T_NAME_FULLY_QUALIFIED) reduce 120) ((T_NAME_QUALIFIED) reduce 120) ((T_NAME_RELATIVE) reduce 120) ((T_NEW) reduce 120) ((T_NS_C) reduce 120) ((T_OBJECT_CAST) reduce 120) ((T_PRINT) reduce 120) ((T_PROPERTY_C) reduce 120) ((T_READONLY) reduce 120) ((T_REQUIRE) reduce 120) ((T_REQUIRE_ONCE) reduce 120) ((T_RETURN) reduce 120) ((T_START_HEREDOC) reduce 120) ((T_STATIC) reduce 120) ((T_STRING) reduce 120) ((T_STRING_CAST) reduce 120) ((T_SWITCH) reduce 120) ((T_THROW) reduce 120) ((T_TRAIT) reduce 120) ((T_TRAIT_C) reduce 120) ((T_TRY) reduce 120) ((T_UNSET) reduce 120) ((T_UNSET_CAST) reduce 120) ((T_USE) reduce 120) ((T_VARIABLE) reduce 120) ((T_VOID_CAST) reduce 120) ((T_WHILE) reduce 120) ((T_YIELD) reduce 120) ((T_YIELD_FROM) reduce 120) (("[") reduce 120) (("`") reduce 120) (("{") reduce 120) (("~") reduce 120)) 32143 (((",") reduce 91) ((";") reduce 91) ((T_AS) reduce 91)) 32144 (((",") reduce 89) ((";") reduce 89) ((T_AS) reduce 89)) 32145 (((",") reduce 88) ((";") reduce 88) ((T_AS) reduce 88)) 32146 (((",") reduce 137) ((";") reduce 137) ((T_AS) shift)) 32147 (((",") reduce 90) ((";") reduce 90) ((T_AS) reduce 90)) 32148 (((",") reduce 131) ((";") reduce 131)) 32150 (((",") reduce 138) ((";") reduce 138)) 32152 (((T_NAME_QUALIFIED) shift) ((T_STRING) shift)) 32153 (((",") reduce 89) ((T_AS) reduce 89) (("}") reduce 89)) 32154 (((",") reduce 88) ((T_AS) reduce 88) (("}") reduce 88)) 32155 (((",") reduce 135) ((T_AS) shift) (("}") reduce 135)) 32156 (((",") reduce 130) (("}") reduce 130)) 32158 (((T_NAME_QUALIFIED) shift) ((T_STRING) shift) (("}") reduce 126)) 32160 (((";") reduce 123)) 32161 (((",") reduce 129) (("}") reduce 129)) 32163 (((",") reduce 136) (("}") reduce 136)) 32164 ((("!") reduce 118) (("\"") reduce 118) (($) reduce 118) (("$") reduce 118) (("(") reduce 118) (("+") reduce 118) (("-") reduce 118) ((";") reduce 118) (("@") reduce 118) ((T_ABSTRACT) reduce 118) ((T_ARRAY) reduce 118) ((T_ARRAY_CAST) reduce 118) ((T_ATTRIBUTE) reduce 118) ((T_BOOL_CAST) reduce 118) ((T_BREAK) reduce 118) ((T_CLASS) reduce 118) ((T_CLASS_C) reduce 118) ((T_CLONE) reduce 118) ((T_CONST) reduce 118) ((T_CONSTANT_ENCAPSED_STRING) reduce 118) ((T_CONTINUE) reduce 118) ((T_DEC) reduce 118) ((T_DECLARE) reduce 118) ((T_DIR) reduce 118) ((T_DNUMBER) reduce 118) ((T_DO) reduce 118) ((T_DOUBLE_CAST) reduce 118) ((T_ECHO) reduce 118) ((T_EMPTY) reduce 118) ((T_ENUM) reduce 118) ((T_EVAL) reduce 118) ((T_EXIT) reduce 118) ((T_FILE) reduce 118) ((T_FINAL) reduce 118) ((T_FN) reduce 118) ((T_FOR) reduce 118) ((T_FOREACH) reduce 118) ((T_FUNCTION) reduce 118) ((T_FUNC_C) reduce 118) ((T_GLOBAL) reduce 118) ((T_GOTO) reduce 118) ((T_HALT_COMPILER) reduce 118) ((T_IF) reduce 118) ((T_INC) reduce 118) ((T_INCLUDE) reduce 118) ((T_INCLUDE_ONCE) reduce 118) ((T_INLINE_HTML) reduce 118) ((T_INTERFACE) reduce 118) ((T_INT_CAST) reduce 118) ((T_ISSET) reduce 118) ((T_LINE) reduce 118) ((T_LIST) reduce 118) ((T_LNUMBER) reduce 118) ((T_MATCH) reduce 118) ((T_METHOD_C) reduce 118) ((T_NAMESPACE) reduce 118) ((T_NAME_FULLY_QUALIFIED) reduce 118) ((T_NAME_QUALIFIED) reduce 118) ((T_NAME_RELATIVE) reduce 118) ((T_NEW) reduce 118) ((T_NS_C) reduce 118) ((T_OBJECT_CAST) reduce 118) ((T_PRINT) reduce 118) ((T_PROPERTY_C) reduce 118) ((T_READONLY) reduce 118) ((T_REQUIRE) reduce 118) ((T_REQUIRE_ONCE) reduce 118) ((T_RETURN) reduce 118) ((T_START_HEREDOC) reduce 118) ((T_STATIC) reduce 118) ((T_STRING) reduce 118) ((T_STRING_CAST) reduce 118) ((T_SWITCH) reduce 118) ((T_THROW) reduce 118) ((T_TRAIT) reduce 118) ((T_TRAIT_C) reduce 118) ((T_TRY) reduce 118) ((T_UNSET) reduce 118) ((T_UNSET_CAST) reduce 118) ((T_USE) reduce 118) ((T_VARIABLE) reduce 118) ((T_VOID_CAST) reduce 118) ((T_WHILE) reduce 118) ((T_YIELD) reduce 118) ((T_YIELD_FROM) reduce 118) (("[") reduce 118) (("`") reduce 118) (("{") reduce 118) (("~") reduce 118)) 32165 ((("!") reduce 119) (("\"") reduce 119) (("$") reduce 119) (($) reduce 119) (("(") reduce 119) (("+") reduce 119) (("-") reduce 119) ((";") reduce 119) (("@") reduce 119) ((T_ABSTRACT) reduce 119) ((T_ARRAY) reduce 119) ((T_ARRAY_CAST) reduce 119) ((T_ATTRIBUTE) reduce 119) ((T_BOOL_CAST) reduce 119) ((T_BREAK) reduce 119) ((T_CLASS) reduce 119) ((T_CLASS_C) reduce 119) ((T_CLONE) reduce 119) ((T_CONST) reduce 119) ((T_CONSTANT_ENCAPSED_STRING) reduce 119) ((T_CONTINUE) reduce 119) ((T_DEC) reduce 119) ((T_DECLARE) reduce 119) ((T_DIR) reduce 119) ((T_DNUMBER) reduce 119) ((T_DO) reduce 119) ((T_DOUBLE_CAST) reduce 119) ((T_ECHO) reduce 119) ((T_EMPTY) reduce 119) ((T_ENUM) reduce 119) ((T_EVAL) reduce 119) ((T_EXIT) reduce 119) ((T_FILE) reduce 119) ((T_FINAL) reduce 119) ((T_FN) reduce 119) ((T_FOR) reduce 119) ((T_FOREACH) reduce 119) ((T_FUNCTION) reduce 119) ((T_FUNC_C) reduce 119) ((T_GLOBAL) reduce 119) ((T_GOTO) reduce 119) ((T_HALT_COMPILER) reduce 119) ((T_IF) reduce 119) ((T_INC) reduce 119) ((T_INCLUDE) reduce 119) ((T_INCLUDE_ONCE) reduce 119) ((T_INLINE_HTML) reduce 119) ((T_INTERFACE) reduce 119) ((T_INT_CAST) reduce 119) ((T_ISSET) reduce 119) ((T_LINE) reduce 119) ((T_LIST) reduce 119) ((T_LNUMBER) reduce 119) ((T_MATCH) reduce 119) ((T_METHOD_C) reduce 119) ((T_NAMESPACE) reduce 119) ((T_NAME_FULLY_QUALIFIED) reduce 119) ((T_NAME_QUALIFIED) reduce 119) ((T_NAME_RELATIVE) reduce 119) ((T_NEW) reduce 119) ((T_NS_C) reduce 119) ((T_OBJECT_CAST) reduce 119) ((T_PRINT) reduce 119) ((T_PROPERTY_C) reduce 119) ((T_READONLY) reduce 119) ((T_REQUIRE) reduce 119) ((T_REQUIRE_ONCE) reduce 119) ((T_RETURN) reduce 119) ((T_START_HEREDOC) reduce 119) ((T_STATIC) reduce 119) ((T_STRING) reduce 119) ((T_STRING_CAST) reduce 119) ((T_SWITCH) reduce 119) ((T_THROW) reduce 119) ((T_TRAIT) reduce 119) ((T_TRAIT_C) reduce 119) ((T_TRY) reduce 119) ((T_UNSET) reduce 119) ((T_UNSET_CAST) reduce 119) ((T_USE) reduce 119) ((T_VARIABLE) reduce 119) ((T_VOID_CAST) reduce 119) ((T_WHILE) reduce 119) ((T_YIELD) reduce 119) ((T_YIELD_FROM) reduce 119) (("[") reduce 119) (("`") reduce 119) (("{") reduce 119) (("~") reduce 119)) 32166 ((("!") reduce 117) (("\"") reduce 117) (("$") reduce 117) (($) reduce 117) (("(") reduce 117) (("+") reduce 117) (("-") reduce 117) ((";") reduce 117) (("@") reduce 117) ((T_ABSTRACT) reduce 117) ((T_ARRAY) reduce 117) ((T_ARRAY_CAST) reduce 117) ((T_ATTRIBUTE) reduce 117) ((T_BOOL_CAST) reduce 117) ((T_BREAK) reduce 117) ((T_CLASS) reduce 117) ((T_CLASS_C) reduce 117) ((T_CLONE) reduce 117) ((T_CONST) reduce 117) ((T_CONSTANT_ENCAPSED_STRING) reduce 117) ((T_CONTINUE) reduce 117) ((T_DEC) reduce 117) ((T_DECLARE) reduce 117) ((T_DIR) reduce 117) ((T_DNUMBER) reduce 117) ((T_DO) reduce 117) ((T_DOUBLE_CAST) reduce 117) ((T_ECHO) reduce 117) ((T_EMPTY) reduce 117) ((T_ENUM) reduce 117) ((T_EVAL) reduce 117) ((T_EXIT) reduce 117) ((T_FILE) reduce 117) ((T_FINAL) reduce 117) ((T_FN) reduce 117) ((T_FOR) reduce 117) ((T_FOREACH) reduce 117) ((T_FUNCTION) reduce 117) ((T_FUNC_C) reduce 117) ((T_GLOBAL) reduce 117) ((T_GOTO) reduce 117) ((T_HALT_COMPILER) reduce 117) ((T_IF) reduce 117) ((T_INC) reduce 117) ((T_INCLUDE) reduce 117) ((T_INCLUDE_ONCE) reduce 117) ((T_INLINE_HTML) reduce 117) ((T_INTERFACE) reduce 117) ((T_INT_CAST) reduce 117) ((T_ISSET) reduce 117) ((T_LINE) reduce 117) ((T_LIST) reduce 117) ((T_LNUMBER) reduce 117) ((T_MATCH) reduce 117) ((T_METHOD_C) reduce 117) ((T_NAMESPACE) reduce 117) ((T_NAME_FULLY_QUALIFIED) reduce 117) ((T_NAME_QUALIFIED) reduce 117) ((T_NAME_RELATIVE) reduce 117) ((T_NEW) reduce 117) ((T_NS_C) reduce 117) ((T_OBJECT_CAST) reduce 117) ((T_PRINT) reduce 117) ((T_PROPERTY_C) reduce 117) ((T_READONLY) reduce 117) ((T_REQUIRE) reduce 117) ((T_REQUIRE_ONCE) reduce 117) ((T_RETURN) reduce 117) ((T_START_HEREDOC) reduce 117) ((T_STATIC) reduce 117) ((T_STRING) reduce 117) ((T_STRING_CAST) reduce 117) ((T_SWITCH) reduce 117) ((T_THROW) reduce 117) ((T_TRAIT) reduce 117) ((T_TRAIT_C) reduce 117) ((T_TRY) reduce 117) ((T_UNSET) reduce 117) ((T_UNSET_CAST) reduce 117) ((T_USE) reduce 117) ((T_VARIABLE) reduce 117) ((T_VOID_CAST) reduce 117) ((T_WHILE) reduce 117) ((T_YIELD) reduce 117) ((T_YIELD_FROM) reduce 117) (("[") reduce 117) (("`") reduce 117) (("{") reduce 117) (("~") reduce 117)) 32168 (((T_CONST) shift) ((T_FUNCTION) shift) ((T_NAME_QUALIFIED) shift) ((T_STRING) shift)) 32169 (((T_NAME_QUALIFIED) reduce 122) ((T_STRING) reduce 122)) 32170 (((T_NAME_QUALIFIED) reduce 121) ((T_STRING) reduce 121)) 32171 (((",") reduce 128) (("}") reduce 128)) 32173 (((",") reduce 133) (("}") reduce 133)) 32175 (((",") reduce 134) (("}") reduce 134)) 32176 (((T_CONST) shift) ((T_FUNCTION) shift) ((T_NAME_QUALIFIED) shift) ((T_STRING) shift) (("}") reduce 126)) 32178 (((";") reduce 124)) 32179 (((",") reduce 127) (("}") reduce 127)) 32185 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (($) reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONST) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_EMPTY) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAMESPACE) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_USE) reduce 162) ((T_VARIABLE) reduce 162) ((T_VOID_CAST) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 32188 ((("!") reduce 171) (("\"") reduce 171) (($) reduce 171) (("$") reduce 171) (("(") reduce 171) (("+") reduce 171) (("-") reduce 171) ((";") reduce 171) (("@") reduce 171) ((T_ABSTRACT) reduce 171) ((T_ARRAY) reduce 171) ((T_ARRAY_CAST) reduce 171) ((T_ATTRIBUTE) reduce 171) ((T_BOOL_CAST) reduce 171) ((T_BREAK) reduce 171) ((T_CATCH) reduce 171) ((T_CLASS) reduce 171) ((T_CLASS_C) reduce 171) ((T_CLONE) reduce 171) ((T_CONST) reduce 171) ((T_CONSTANT_ENCAPSED_STRING) reduce 171) ((T_CONTINUE) reduce 171) ((T_DEC) reduce 171) ((T_DECLARE) reduce 171) ((T_DIR) reduce 171) ((T_DNUMBER) reduce 171) ((T_DO) reduce 171) ((T_DOUBLE_CAST) reduce 171) ((T_ECHO) reduce 171) ((T_EMPTY) reduce 171) ((T_ENUM) reduce 171) ((T_EVAL) reduce 171) ((T_EXIT) reduce 171) ((T_FILE) reduce 171) ((T_FINAL) reduce 171) ((T_FINALLY) reduce 171) ((T_FN) reduce 171) ((T_FOR) reduce 171) ((T_FOREACH) reduce 171) ((T_FUNCTION) reduce 171) ((T_FUNC_C) reduce 171) ((T_GLOBAL) reduce 171) ((T_GOTO) reduce 171) ((T_HALT_COMPILER) reduce 171) ((T_IF) reduce 171) ((T_INC) reduce 171) ((T_INCLUDE) reduce 171) ((T_INCLUDE_ONCE) reduce 171) ((T_INLINE_HTML) reduce 171) ((T_INTERFACE) reduce 171) ((T_INT_CAST) reduce 171) ((T_ISSET) reduce 171) ((T_LINE) reduce 171) ((T_LIST) reduce 171) ((T_LNUMBER) reduce 171) ((T_MATCH) reduce 171) ((T_METHOD_C) reduce 171) ((T_NAMESPACE) reduce 171) ((T_NAME_FULLY_QUALIFIED) reduce 171) ((T_NAME_QUALIFIED) reduce 171) ((T_NAME_RELATIVE) reduce 171) ((T_NEW) reduce 171) ((T_NS_C) reduce 171) ((T_OBJECT_CAST) reduce 171) ((T_PRINT) reduce 171) ((T_PROPERTY_C) reduce 171) ((T_READONLY) reduce 171) ((T_REQUIRE) reduce 171) ((T_REQUIRE_ONCE) reduce 171) ((T_RETURN) reduce 171) ((T_START_HEREDOC) reduce 171) ((T_STATIC) reduce 171) ((T_STRING) reduce 171) ((T_STRING_CAST) reduce 171) ((T_SWITCH) reduce 171) ((T_THROW) reduce 171) ((T_TRAIT) reduce 171) ((T_TRAIT_C) reduce 171) ((T_TRY) reduce 171) ((T_UNSET) reduce 171) ((T_UNSET_CAST) reduce 171) ((T_USE) reduce 171) ((T_VARIABLE) reduce 171) ((T_VOID_CAST) reduce 171) ((T_WHILE) reduce 171) ((T_YIELD) reduce 171) ((T_YIELD_FROM) reduce 171) (("[") reduce 171) (("`") reduce 171) (("{") reduce 171) (("~") reduce 171)) 32189 ((("!") reduce 177) (("\"") reduce 177) (("$") reduce 177) (($) reduce 177) (("(") reduce 177) (("+") reduce 177) (("-") reduce 177) ((";") reduce 177) (("@") reduce 177) ((T_ABSTRACT) reduce 177) ((T_ARRAY) reduce 177) ((T_ARRAY_CAST) reduce 177) ((T_ATTRIBUTE) reduce 177) ((T_BOOL_CAST) reduce 177) ((T_BREAK) reduce 177) ((T_CATCH) shift) ((T_CLASS) reduce 177) ((T_CLASS_C) reduce 177) ((T_CLONE) reduce 177) ((T_CONST) reduce 177) ((T_CONSTANT_ENCAPSED_STRING) reduce 177) ((T_CONTINUE) reduce 177) ((T_DEC) reduce 177) ((T_DECLARE) reduce 177) ((T_DIR) reduce 177) ((T_DNUMBER) reduce 177) ((T_DO) reduce 177) ((T_DOUBLE_CAST) reduce 177) ((T_ECHO) reduce 177) ((T_EMPTY) reduce 177) ((T_ENUM) reduce 177) ((T_EVAL) reduce 177) ((T_EXIT) reduce 177) ((T_FILE) reduce 177) ((T_FINAL) reduce 177) ((T_FINALLY) shift) ((T_FN) reduce 177) ((T_FOR) reduce 177) ((T_FOREACH) reduce 177) ((T_FUNCTION) reduce 177) ((T_FUNC_C) reduce 177) ((T_GLOBAL) reduce 177) ((T_GOTO) reduce 177) ((T_HALT_COMPILER) reduce 177) ((T_IF) reduce 177) ((T_INC) reduce 177) ((T_INCLUDE) reduce 177) ((T_INCLUDE_ONCE) reduce 177) ((T_INLINE_HTML) reduce 177) ((T_INTERFACE) reduce 177) ((T_INT_CAST) reduce 177) ((T_ISSET) reduce 177) ((T_LINE) reduce 177) ((T_LIST) reduce 177) ((T_LNUMBER) reduce 177) ((T_MATCH) reduce 177) ((T_METHOD_C) reduce 177) ((T_NAMESPACE) reduce 177) ((T_NAME_FULLY_QUALIFIED) reduce 177) ((T_NAME_QUALIFIED) reduce 177) ((T_NAME_RELATIVE) reduce 177) ((T_NEW) reduce 177) ((T_NS_C) reduce 177) ((T_OBJECT_CAST) reduce 177) ((T_PRINT) reduce 177) ((T_PROPERTY_C) reduce 177) ((T_READONLY) reduce 177) ((T_REQUIRE) reduce 177) ((T_REQUIRE_ONCE) reduce 177) ((T_RETURN) reduce 177) ((T_START_HEREDOC) reduce 177) ((T_STATIC) reduce 177) ((T_STRING) reduce 177) ((T_STRING_CAST) reduce 177) ((T_SWITCH) reduce 177) ((T_THROW) reduce 177) ((T_TRAIT) reduce 177) ((T_TRAIT_C) reduce 177) ((T_TRY) reduce 177) ((T_UNSET) reduce 177) ((T_UNSET_CAST) reduce 177) ((T_USE) reduce 177) ((T_VARIABLE) reduce 177) ((T_VOID_CAST) reduce 177) ((T_WHILE) reduce 177) ((T_YIELD) reduce 177) ((T_YIELD_FROM) reduce 177) (("[") reduce 177) (("`") reduce 177) (("{") reduce 177) (("~") reduce 177)) 32192 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (($) reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONST) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_EMPTY) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAMESPACE) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_USE) reduce 167) ((T_VARIABLE) reduce 167) ((T_VOID_CAST) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 32195 ((("!") reduce 178) (("\"") reduce 178) (("$") reduce 178) (($) reduce 178) (("(") reduce 178) (("+") reduce 178) (("-") reduce 178) ((";") reduce 178) (("@") reduce 178) ((T_ABSTRACT) reduce 178) ((T_ARRAY) reduce 178) ((T_ARRAY_CAST) reduce 178) ((T_ATTRIBUTE) reduce 178) ((T_BOOL_CAST) reduce 178) ((T_BREAK) reduce 178) ((T_CLASS) reduce 178) ((T_CLASS_C) reduce 178) ((T_CLONE) reduce 178) ((T_CONST) reduce 178) ((T_CONSTANT_ENCAPSED_STRING) reduce 178) ((T_CONTINUE) reduce 178) ((T_DEC) reduce 178) ((T_DECLARE) reduce 178) ((T_DIR) reduce 178) ((T_DNUMBER) reduce 178) ((T_DO) reduce 178) ((T_DOUBLE_CAST) reduce 178) ((T_ECHO) reduce 178) ((T_EMPTY) reduce 178) ((T_ENUM) reduce 178) ((T_EVAL) reduce 178) ((T_EXIT) reduce 178) ((T_FILE) reduce 178) ((T_FINAL) reduce 178) ((T_FN) reduce 178) ((T_FOR) reduce 178) ((T_FOREACH) reduce 178) ((T_FUNCTION) reduce 178) ((T_FUNC_C) reduce 178) ((T_GLOBAL) reduce 178) ((T_GOTO) reduce 178) ((T_HALT_COMPILER) reduce 178) ((T_IF) reduce 178) ((T_INC) reduce 178) ((T_INCLUDE) reduce 178) ((T_INCLUDE_ONCE) reduce 178) ((T_INLINE_HTML) reduce 178) ((T_INTERFACE) reduce 178) ((T_INT_CAST) reduce 178) ((T_ISSET) reduce 178) ((T_LINE) reduce 178) ((T_LIST) reduce 178) ((T_LNUMBER) reduce 178) ((T_MATCH) reduce 178) ((T_METHOD_C) reduce 178) ((T_NAMESPACE) reduce 178) ((T_NAME_FULLY_QUALIFIED) reduce 178) ((T_NAME_QUALIFIED) reduce 178) ((T_NAME_RELATIVE) reduce 178) ((T_NEW) reduce 178) ((T_NS_C) reduce 178) ((T_OBJECT_CAST) reduce 178) ((T_PRINT) reduce 178) ((T_PROPERTY_C) reduce 178) ((T_READONLY) reduce 178) ((T_REQUIRE) reduce 178) ((T_REQUIRE_ONCE) reduce 178) ((T_RETURN) reduce 178) ((T_START_HEREDOC) reduce 178) ((T_STATIC) reduce 178) ((T_STRING) reduce 178) ((T_STRING_CAST) reduce 178) ((T_SWITCH) reduce 178) ((T_THROW) reduce 178) ((T_TRAIT) reduce 178) ((T_TRAIT_C) reduce 178) ((T_TRY) reduce 178) ((T_UNSET) reduce 178) ((T_UNSET_CAST) reduce 178) ((T_USE) reduce 178) ((T_VARIABLE) reduce 178) ((T_VOID_CAST) reduce 178) ((T_WHILE) reduce 178) ((T_YIELD) reduce 178) ((T_YIELD_FROM) reduce 178) (("[") reduce 178) (("`") reduce 178) (("{") reduce 178) (("~") reduce 178)) 32202 ((("!") reduce 172) (("\"") reduce 172) (("$") reduce 172) (($) reduce 172) (("(") reduce 172) (("+") reduce 172) (("-") reduce 172) ((";") reduce 172) (("@") reduce 172) ((T_ABSTRACT) reduce 172) ((T_ARRAY) reduce 172) ((T_ARRAY_CAST) reduce 172) ((T_ATTRIBUTE) reduce 172) ((T_BOOL_CAST) reduce 172) ((T_BREAK) reduce 172) ((T_CATCH) reduce 172) ((T_CLASS) reduce 172) ((T_CLASS_C) reduce 172) ((T_CLONE) reduce 172) ((T_CONST) reduce 172) ((T_CONSTANT_ENCAPSED_STRING) reduce 172) ((T_CONTINUE) reduce 172) ((T_DEC) reduce 172) ((T_DECLARE) reduce 172) ((T_DIR) reduce 172) ((T_DNUMBER) reduce 172) ((T_DO) reduce 172) ((T_DOUBLE_CAST) reduce 172) ((T_ECHO) reduce 172) ((T_EMPTY) reduce 172) ((T_ENUM) reduce 172) ((T_EVAL) reduce 172) ((T_EXIT) reduce 172) ((T_FILE) reduce 172) ((T_FINAL) reduce 172) ((T_FINALLY) reduce 172) ((T_FN) reduce 172) ((T_FOR) reduce 172) ((T_FOREACH) reduce 172) ((T_FUNCTION) reduce 172) ((T_FUNC_C) reduce 172) ((T_GLOBAL) reduce 172) ((T_GOTO) reduce 172) ((T_HALT_COMPILER) reduce 172) ((T_IF) reduce 172) ((T_INC) reduce 172) ((T_INCLUDE) reduce 172) ((T_INCLUDE_ONCE) reduce 172) ((T_INLINE_HTML) reduce 172) ((T_INTERFACE) reduce 172) ((T_INT_CAST) reduce 172) ((T_ISSET) reduce 172) ((T_LINE) reduce 172) ((T_LIST) reduce 172) ((T_LNUMBER) reduce 172) ((T_MATCH) reduce 172) ((T_METHOD_C) reduce 172) ((T_NAMESPACE) reduce 172) ((T_NAME_FULLY_QUALIFIED) reduce 172) ((T_NAME_QUALIFIED) reduce 172) ((T_NAME_RELATIVE) reduce 172) ((T_NEW) reduce 172) ((T_NS_C) reduce 172) ((T_OBJECT_CAST) reduce 172) ((T_PRINT) reduce 172) ((T_PROPERTY_C) reduce 172) ((T_READONLY) reduce 172) ((T_REQUIRE) reduce 172) ((T_REQUIRE_ONCE) reduce 172) ((T_RETURN) reduce 172) ((T_START_HEREDOC) reduce 172) ((T_STATIC) reduce 172) ((T_STRING) reduce 172) ((T_STRING_CAST) reduce 172) ((T_SWITCH) reduce 172) ((T_THROW) reduce 172) ((T_TRAIT) reduce 172) ((T_TRAIT_C) reduce 172) ((T_TRY) reduce 172) ((T_UNSET) reduce 172) ((T_UNSET_CAST) reduce 172) ((T_USE) reduce 172) ((T_VARIABLE) reduce 172) ((T_VOID_CAST) reduce 172) ((T_WHILE) reduce 172) ((T_YIELD) reduce 172) ((T_YIELD_FROM) reduce 172) (("[") reduce 172) (("`") reduce 172) (("{") reduce 172) (("~") reduce 172)) 32207 ((("!") reduce 200) (("\"") reduce 200) (("$") reduce 200) (($) reduce 200) (("(") reduce 200) (("+") reduce 200) (("-") reduce 200) ((";") reduce 200) (("@") reduce 200) ((T_ABSTRACT) reduce 200) ((T_ARRAY) reduce 200) ((T_ARRAY_CAST) reduce 200) ((T_ATTRIBUTE) reduce 200) ((T_BOOL_CAST) reduce 200) ((T_BREAK) reduce 200) ((T_CLASS) reduce 200) ((T_CLASS_C) reduce 200) ((T_CLONE) reduce 200) ((T_CONST) reduce 200) ((T_CONSTANT_ENCAPSED_STRING) reduce 200) ((T_CONTINUE) reduce 200) ((T_DEC) reduce 200) ((T_DECLARE) reduce 200) ((T_DIR) reduce 200) ((T_DNUMBER) reduce 200) ((T_DO) reduce 200) ((T_DOUBLE_CAST) reduce 200) ((T_ECHO) reduce 200) ((T_EMPTY) reduce 200) ((T_ENUM) reduce 200) ((T_EVAL) reduce 200) ((T_EXIT) reduce 200) ((T_FILE) reduce 200) ((T_FINAL) reduce 200) ((T_FN) reduce 200) ((T_FOR) reduce 200) ((T_FOREACH) reduce 200) ((T_FUNCTION) reduce 200) ((T_FUNC_C) reduce 200) ((T_GLOBAL) reduce 200) ((T_GOTO) reduce 200) ((T_HALT_COMPILER) reduce 200) ((T_IF) reduce 200) ((T_INC) reduce 200) ((T_INCLUDE) reduce 200) ((T_INCLUDE_ONCE) reduce 200) ((T_INLINE_HTML) reduce 200) ((T_INTERFACE) reduce 200) ((T_INT_CAST) reduce 200) ((T_ISSET) reduce 200) ((T_LINE) reduce 200) ((T_LIST) reduce 200) ((T_LNUMBER) reduce 200) ((T_MATCH) reduce 200) ((T_METHOD_C) reduce 200) ((T_NAMESPACE) reduce 200) ((T_NAME_FULLY_QUALIFIED) reduce 200) ((T_NAME_QUALIFIED) reduce 200) ((T_NAME_RELATIVE) reduce 200) ((T_NEW) reduce 200) ((T_NS_C) reduce 200) ((T_OBJECT_CAST) reduce 200) ((T_PRINT) reduce 200) ((T_PROPERTY_C) reduce 200) ((T_READONLY) reduce 200) ((T_REQUIRE) reduce 200) ((T_REQUIRE_ONCE) reduce 200) ((T_RETURN) reduce 200) ((T_START_HEREDOC) reduce 200) ((T_STATIC) reduce 200) ((T_STRING) reduce 200) ((T_STRING_CAST) reduce 200) ((T_SWITCH) reduce 200) ((T_THROW) reduce 200) ((T_TRAIT) reduce 200) ((T_TRAIT_C) reduce 200) ((T_TRY) reduce 200) ((T_UNSET) reduce 200) ((T_UNSET_CAST) reduce 200) ((T_USE) reduce 200) ((T_VARIABLE) reduce 200) ((T_VOID_CAST) reduce 200) ((T_WHILE) reduce 200) ((T_YIELD) reduce 200) ((T_YIELD_FROM) reduce 200) (("[") reduce 200) (("`") reduce 200) (("{") reduce 200) (("~") reduce 200)) 32213 ((("!") reduce 153) (("\"") reduce 153) (($) reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONST) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_EMPTY) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAMESPACE) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_USE) reduce 153) ((T_VARIABLE) reduce 153) ((T_VOID_CAST) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 32217 ((("!") reduce 224) (("\"") reduce 224) (("$") reduce 224) (($) reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONST) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_EMPTY) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAMESPACE) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_USE) reduce 224) ((T_VARIABLE) reduce 224) ((T_VOID_CAST) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("~") reduce 224)) 32219 ((("!") reduce 225) (("\"") reduce 225) (($) reduce 225) (("$") reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONST) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_EMPTY) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAMESPACE) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_USE) reduce 225) ((T_VARIABLE) reduce 225) ((T_VOID_CAST) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("~") reduce 225)) 32223 ((("!") reduce 226) (("\"") reduce 226) (($) reduce 226) (("$") reduce 226) (("(") reduce 226) (("+") reduce 226) (("-") reduce 226) ((";") reduce 226) (("@") reduce 226) ((T_ABSTRACT) reduce 226) ((T_ARRAY) reduce 226) ((T_ARRAY_CAST) reduce 226) ((T_ATTRIBUTE) reduce 226) ((T_BOOL_CAST) reduce 226) ((T_BREAK) reduce 226) ((T_CLASS) reduce 226) ((T_CLASS_C) reduce 226) ((T_CLONE) reduce 226) ((T_CONST) reduce 226) ((T_CONSTANT_ENCAPSED_STRING) reduce 226) ((T_CONTINUE) reduce 226) ((T_DEC) reduce 226) ((T_DECLARE) reduce 226) ((T_DIR) reduce 226) ((T_DNUMBER) reduce 226) ((T_DO) reduce 226) ((T_DOUBLE_CAST) reduce 226) ((T_ECHO) reduce 226) ((T_EMPTY) reduce 226) ((T_ENUM) reduce 226) ((T_EVAL) reduce 226) ((T_EXIT) reduce 226) ((T_FILE) reduce 226) ((T_FINAL) reduce 226) ((T_FN) reduce 226) ((T_FOR) reduce 226) ((T_FOREACH) reduce 226) ((T_FUNCTION) reduce 226) ((T_FUNC_C) reduce 226) ((T_GLOBAL) reduce 226) ((T_GOTO) reduce 226) ((T_HALT_COMPILER) reduce 226) ((T_IF) reduce 226) ((T_INC) reduce 226) ((T_INCLUDE) reduce 226) ((T_INCLUDE_ONCE) reduce 226) ((T_INLINE_HTML) reduce 226) ((T_INTERFACE) reduce 226) ((T_INT_CAST) reduce 226) ((T_ISSET) reduce 226) ((T_LINE) reduce 226) ((T_LIST) reduce 226) ((T_LNUMBER) reduce 226) ((T_MATCH) reduce 226) ((T_METHOD_C) reduce 226) ((T_NAMESPACE) reduce 226) ((T_NAME_FULLY_QUALIFIED) reduce 226) ((T_NAME_QUALIFIED) reduce 226) ((T_NAME_RELATIVE) reduce 226) ((T_NEW) reduce 226) ((T_NS_C) reduce 226) ((T_OBJECT_CAST) reduce 226) ((T_PRINT) reduce 226) ((T_PROPERTY_C) reduce 226) ((T_READONLY) reduce 226) ((T_REQUIRE) reduce 226) ((T_REQUIRE_ONCE) reduce 226) ((T_RETURN) reduce 226) ((T_START_HEREDOC) reduce 226) ((T_STATIC) reduce 226) ((T_STRING) reduce 226) ((T_STRING_CAST) reduce 226) ((T_SWITCH) reduce 226) ((T_THROW) reduce 226) ((T_TRAIT) reduce 226) ((T_TRAIT_C) reduce 226) ((T_TRY) reduce 226) ((T_UNSET) reduce 226) ((T_UNSET_CAST) reduce 226) ((T_USE) reduce 226) ((T_VARIABLE) reduce 226) ((T_VOID_CAST) reduce 226) ((T_WHILE) reduce 226) ((T_YIELD) reduce 226) ((T_YIELD_FROM) reduce 226) (("[") reduce 226) (("`") reduce 226) (("{") reduce 226) (("~") reduce 226)) 32226 ((("!") reduce 227) (("\"") reduce 227) (("$") reduce 227) (($) reduce 227) (("(") reduce 227) (("+") reduce 227) (("-") reduce 227) ((";") reduce 227) (("@") reduce 227) ((T_ABSTRACT) reduce 227) ((T_ARRAY) reduce 227) ((T_ARRAY_CAST) reduce 227) ((T_ATTRIBUTE) reduce 227) ((T_BOOL_CAST) reduce 227) ((T_BREAK) reduce 227) ((T_CLASS) reduce 227) ((T_CLASS_C) reduce 227) ((T_CLONE) reduce 227) ((T_CONST) reduce 227) ((T_CONSTANT_ENCAPSED_STRING) reduce 227) ((T_CONTINUE) reduce 227) ((T_DEC) reduce 227) ((T_DECLARE) reduce 227) ((T_DIR) reduce 227) ((T_DNUMBER) reduce 227) ((T_DO) reduce 227) ((T_DOUBLE_CAST) reduce 227) ((T_ECHO) reduce 227) ((T_EMPTY) reduce 227) ((T_ENUM) reduce 227) ((T_EVAL) reduce 227) ((T_EXIT) reduce 227) ((T_FILE) reduce 227) ((T_FINAL) reduce 227) ((T_FN) reduce 227) ((T_FOR) reduce 227) ((T_FOREACH) reduce 227) ((T_FUNCTION) reduce 227) ((T_FUNC_C) reduce 227) ((T_GLOBAL) reduce 227) ((T_GOTO) reduce 227) ((T_HALT_COMPILER) reduce 227) ((T_IF) reduce 227) ((T_INC) reduce 227) ((T_INCLUDE) reduce 227) ((T_INCLUDE_ONCE) reduce 227) ((T_INLINE_HTML) reduce 227) ((T_INTERFACE) reduce 227) ((T_INT_CAST) reduce 227) ((T_ISSET) reduce 227) ((T_LINE) reduce 227) ((T_LIST) reduce 227) ((T_LNUMBER) reduce 227) ((T_MATCH) reduce 227) ((T_METHOD_C) reduce 227) ((T_NAMESPACE) reduce 227) ((T_NAME_FULLY_QUALIFIED) reduce 227) ((T_NAME_QUALIFIED) reduce 227) ((T_NAME_RELATIVE) reduce 227) ((T_NEW) reduce 227) ((T_NS_C) reduce 227) ((T_OBJECT_CAST) reduce 227) ((T_PRINT) reduce 227) ((T_PROPERTY_C) reduce 227) ((T_READONLY) reduce 227) ((T_REQUIRE) reduce 227) ((T_REQUIRE_ONCE) reduce 227) ((T_RETURN) reduce 227) ((T_START_HEREDOC) reduce 227) ((T_STATIC) reduce 227) ((T_STRING) reduce 227) ((T_STRING_CAST) reduce 227) ((T_SWITCH) reduce 227) ((T_THROW) reduce 227) ((T_TRAIT) reduce 227) ((T_TRAIT_C) reduce 227) ((T_TRY) reduce 227) ((T_UNSET) reduce 227) ((T_UNSET_CAST) reduce 227) ((T_USE) reduce 227) ((T_VARIABLE) reduce 227) ((T_VOID_CAST) reduce 227) ((T_WHILE) reduce 227) ((T_YIELD) reduce 227) ((T_YIELD_FROM) reduce 227) (("[") reduce 227) (("`") reduce 227) (("{") reduce 227) (("~") reduce 227)) 32228 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (($) reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONST) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_EMPTY) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAMESPACE) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_USE) reduce 169) ((T_VARIABLE) reduce 169) ((T_VOID_CAST) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 32230 ((("!") reduce 158) (("\"") reduce 158) (($) reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONST) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_EMPTY) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAMESPACE) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_USE) reduce 158) ((T_VARIABLE) reduce 158) ((T_VOID_CAST) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 32237 ((("!") reduce 156) (("\"") reduce 156) (($) reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONST) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_EMPTY) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAMESPACE) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_USE) reduce 156) ((T_VARIABLE) reduce 156) ((T_VOID_CAST) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 32279 (((";") reduce 74) (("{") reduce 74)) 32280 (((";") reduce 52) (("{") reduce 52)) 32281 (((";") reduce 27) (("{") reduce 27)) 32282 (((";") reduce 51) (("{") reduce 51)) 32283 (((";") reduce 53) (("{") reduce 53)) 32284 (((";") reduce 49) (("{") reduce 49)) 32285 (((";") reduce 29) (("{") reduce 29)) 32286 (((";") reduce 59) (("{") reduce 59)) 32287 (((";") reduce 60) (("{") reduce 60)) 32288 (((";") reduce 11) (("{") reduce 11)) 32289 (((";") reduce 42) (("{") reduce 42)) 32290 (((";") reduce 39) (("{") reduce 39)) 32291 (((";") reduce 25) (("{") reduce 25)) 32292 (((";") reduce 50) (("{") reduce 50)) 32293 (((";") reduce 66) (("{") reduce 66)) 32294 (((";") reduce 18) (("{") reduce 18)) 32295 (((";") reduce 17) (("{") reduce 17)) 32296 (((";") reduce 15) (("{") reduce 15)) 32297 (((";") reduce 14) (("{") reduce 14)) 32298 (((";") reduce 38) (("{") reduce 38)) 32299 (((";") reduce 26) (("{") reduce 26)) 32300 (((";") reduce 22) (("{") reduce 22)) 32301 (((";") reduce 24) (("{") reduce 24)) 32302 (((";") reduce 16) (("{") reduce 16)) 32303 (((";") reduce 48) (("{") reduce 48)) 32304 (((";") reduce 20) (("{") reduce 20)) 32305 (((";") reduce 70) (("{") reduce 70)) 32306 (((";") reduce 3) (("{") reduce 3)) 32307 (((";") reduce 12) (("{") reduce 12)) 32308 (((";") reduce 54) (("{") reduce 54)) 32309 (((";") reduce 65) (("{") reduce 65)) 32310 (((";") reduce 75) (("{") reduce 75)) 32311 (((";") reduce 30) (("{") reduce 30)) 32312 (((";") reduce 68) (("{") reduce 68)) 32313 (((";") reduce 21) (("{") reduce 21)) 32314 (((";") reduce 23) (("{") reduce 23)) 32315 (((";") reduce 41) (("{") reduce 41)) 32316 (((";") reduce 62) (("{") reduce 62)) 32317 (((";") reduce 34) (("{") reduce 34)) 32318 (((";") reduce 40) (("{") reduce 40)) 32319 (((";") reduce 13) (("{") reduce 13)) 32320 (((";") reduce 55) (("{") reduce 55)) 32321 (((";") reduce 1) (("{") reduce 1)) 32322 (((";") reduce 2) (("{") reduce 2)) 32323 (((";") reduce 9) (("{") reduce 9)) 32324 (((";") reduce 33) (("{") reduce 33)) 32325 (((";") reduce 58) (("{") reduce 58)) 32326 (((";") reduce 37) (("{") reduce 37)) 32327 (((";") reduce 64) (("{") reduce 64)) 32328 (((";") reduce 46) (("{") reduce 46)) 32329 (((";") reduce 8) (("{") reduce 8)) 32330 (((";") reduce 6) (("{") reduce 6)) 32331 (((";") reduce 7) (("{") reduce 7)) 32332 (((";") reduce 69) (("{") reduce 69)) 32333 (((";") reduce 63) (("{") reduce 63)) 32334 (((";") reduce 56) (("{") reduce 56)) 32335 (((";") reduce 87) (("{") reduce 87)) 32336 (((";") reduce 10) (("{") reduce 10)) 32337 (((";") reduce 67) (("{") reduce 67)) 32338 (((";") reduce 44) (("{") reduce 44)) 32339 (((";") reduce 76) (("{") reduce 76)) 32340 (((";") reduce 71) (("{") reduce 71)) 32341 (((";") reduce 77) (("{") reduce 77)) 32342 (((";") reduce 78) (("{") reduce 78)) 32343 (((";") reduce 79) (("{") reduce 79)) 32344 (((";") reduce 4) (("{") reduce 4)) 32345 (((";") reduce 5) (("{") reduce 5)) 32346 (((";") reduce 43) (("{") reduce 43)) 32347 (((";") reduce 73) (("{") reduce 73)) 32348 (((";") reduce 82) (("{") reduce 82)) 32349 (((";") reduce 47) (("{") reduce 47)) 32350 (((";") reduce 31) (("{") reduce 31)) 32351 (((";") reduce 57) (("{") reduce 57)) 32352 (((";") reduce 61) (("{") reduce 61)) 32353 (((";") reduce 28) (("{") reduce 28)) 32354 (((";") reduce 36) (("{") reduce 36)) 32355 (((";") reduce 32) (("{") reduce 32)) 32356 (((";") reduce 35) (("{") reduce 35)) 32357 (((";") reduce 19) (("{") reduce 19)) 32358 (((";") reduce 45) (("{") reduce 45)) 32359 (((";") reduce 86) (("{") reduce 86)) 32361 (((";") reduce 72) (("{") reduce 72)) 32362 (((";") reduce 83) (("{") reduce 83)) 32363 ((("!") reduce 85) (("\"") reduce 85) (("$") reduce 85) (("(") reduce 85) (("+") reduce 85) (("-") reduce 85) ((";") reduce 85) (("@") reduce 85) ((T_ABSTRACT) reduce 85) ((T_ARRAY) reduce 85) ((T_ARRAY_CAST) reduce 85) ((T_ATTRIBUTE) reduce 85) ((T_BOOL_CAST) reduce 85) ((T_BREAK) reduce 85) ((T_CLASS) reduce 85) ((T_CLASS_C) reduce 85) ((T_CLONE) reduce 85) ((T_CONST) reduce 85) ((T_CONSTANT_ENCAPSED_STRING) reduce 85) ((T_CONTINUE) reduce 85) ((T_DEC) reduce 85) ((T_DECLARE) reduce 85) ((T_DIR) reduce 85) ((T_DNUMBER) reduce 85) ((T_DO) reduce 85) ((T_DOUBLE_CAST) reduce 85) ((T_ECHO) reduce 85) ((T_EMPTY) reduce 85) ((T_ENUM) reduce 85) ((T_EVAL) reduce 85) ((T_EXIT) reduce 85) ((T_FILE) reduce 85) ((T_FINAL) reduce 85) ((T_FN) reduce 85) ((T_FOR) reduce 85) ((T_FOREACH) reduce 85) ((T_FUNCTION) reduce 85) ((T_FUNC_C) reduce 85) ((T_GLOBAL) reduce 85) ((T_GOTO) reduce 85) ((T_HALT_COMPILER) reduce 85) ((T_IF) reduce 85) ((T_INC) reduce 85) ((T_INCLUDE) reduce 85) ((T_INCLUDE_ONCE) reduce 85) ((T_INLINE_HTML) reduce 85) ((T_INTERFACE) reduce 85) ((T_INT_CAST) reduce 85) ((T_ISSET) reduce 85) ((T_LINE) reduce 85) ((T_LIST) reduce 85) ((T_LNUMBER) reduce 85) ((T_MATCH) reduce 85) ((T_METHOD_C) reduce 85) ((T_NAMESPACE) reduce 85) ((T_NAME_FULLY_QUALIFIED) reduce 85) ((T_NAME_QUALIFIED) reduce 85) ((T_NAME_RELATIVE) reduce 85) ((T_NEW) reduce 85) ((T_NS_C) reduce 85) ((T_OBJECT_CAST) reduce 85) ((T_PRINT) reduce 85) ((T_PROPERTY_C) reduce 85) ((T_READONLY) reduce 85) ((T_REQUIRE) reduce 85) ((T_REQUIRE_ONCE) reduce 85) ((T_RETURN) reduce 85) ((T_START_HEREDOC) reduce 85) ((T_STATIC) reduce 85) ((T_STRING) reduce 85) ((T_STRING_CAST) reduce 85) ((T_SWITCH) reduce 85) ((T_THROW) reduce 85) ((T_TRAIT) reduce 85) ((T_TRAIT_C) reduce 85) ((T_TRY) reduce 85) ((T_UNSET) reduce 85) ((T_UNSET_CAST) reduce 85) ((T_USE) reduce 85) ((T_VARIABLE) reduce 85) ((T_VOID_CAST) reduce 85) ((T_WHILE) reduce 85) ((T_YIELD) reduce 85) ((T_YIELD_FROM) reduce 85) (("[") reduce 85) (("`") reduce 85) (("{") reduce 85) (("}") reduce 85) (("~") reduce 85)) 32364 ((("!") shift) (("\"") shift) (("$") shift) (("(") shift) (("+") shift) (("-") shift) ((";") shift) (("@") shift) ((T_ABSTRACT) shift) ((T_ARRAY) shift) ((T_ARRAY_CAST) shift) ((T_ATTRIBUTE) shift) ((T_BOOL_CAST) shift) ((T_BREAK) shift) ((T_CLASS) shift) ((T_CLASS_C) shift) ((T_CLONE) shift) ((T_CONST) shift) ((T_CONSTANT_ENCAPSED_STRING) shift) ((T_CONTINUE) shift) ((T_DEC) shift) ((T_DECLARE) shift) ((T_DIR) shift) ((T_DNUMBER) shift) ((T_DO) shift) ((T_DOUBLE_CAST) shift) ((T_ECHO) shift) ((T_EMPTY) shift) ((T_ENUM) shift) ((T_EVAL) shift) ((T_EXIT) shift) ((T_FILE) shift) ((T_FINAL) shift) ((T_FN) shift) ((T_FOR) shift) ((T_FOREACH) shift) ((T_FUNCTION) shift) ((T_FUNC_C) shift) ((T_GLOBAL) shift) ((T_GOTO) shift) ((T_HALT_COMPILER) shift) ((T_IF) shift) ((T_INC) shift) ((T_INCLUDE) shift) ((T_INCLUDE_ONCE) shift) ((T_INLINE_HTML) shift) ((T_INTERFACE) shift) ((T_INT_CAST) shift) ((T_ISSET) shift) ((T_LINE) shift) ((T_LIST) shift) ((T_LNUMBER) shift) ((T_MATCH) shift) ((T_METHOD_C) shift) ((T_NAMESPACE) shift) ((T_NAME_FULLY_QUALIFIED) shift) ((T_NAME_QUALIFIED) shift) ((T_NAME_RELATIVE) shift) ((T_NEW) shift) ((T_NS_C) shift) ((T_OBJECT_CAST) shift) ((T_PRINT) shift) ((T_PROPERTY_C) shift) ((T_READONLY) shift) ((T_REQUIRE) shift) ((T_REQUIRE_ONCE) shift) ((T_RETURN) shift) ((T_START_HEREDOC) shift) ((T_STATIC) shift) ((T_STRING) shift) ((T_STRING_CAST) shift) ((T_SWITCH) shift) ((T_THROW) shift) ((T_TRAIT) shift) ((T_TRAIT_C) shift) ((T_TRY) shift) ((T_UNSET) shift) ((T_UNSET_CAST) shift) ((T_USE) shift) ((T_VARIABLE) shift) ((T_VOID_CAST) shift) ((T_WHILE) shift) ((T_YIELD) shift) ((T_YIELD_FROM) shift) (("[") shift) (("`") shift) (("{") shift) (("}") shift) (("~") shift)) 32365 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONST) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_EMPTY) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAMESPACE) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_USE) reduce 166) ((T_VARIABLE) reduce 166) ((T_VOID_CAST) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("}") reduce 166) (("~") reduce 166)) 32380 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONST) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_EMPTY) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAMESPACE) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_USE) reduce 160) ((T_VARIABLE) reduce 160) ((T_VOID_CAST) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("}") reduce 160) (("~") reduce 160)) 32393 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONST) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_EMPTY) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAMESPACE) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_USE) reduce 149) ((T_VARIABLE) reduce 149) ((T_VOID_CAST) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("}") reduce 149) (("~") reduce 149)) 32395 ((("!") reduce 108) (("\"") reduce 108) (("$") reduce 108) (("(") reduce 108) (("+") reduce 108) (("-") reduce 108) ((";") reduce 108) (("@") reduce 108) ((T_ABSTRACT) reduce 108) ((T_ARRAY) reduce 108) ((T_ARRAY_CAST) reduce 108) ((T_ATTRIBUTE) reduce 108) ((T_BOOL_CAST) reduce 108) ((T_BREAK) reduce 108) ((T_CLASS) reduce 108) ((T_CLASS_C) reduce 108) ((T_CLONE) reduce 108) ((T_CONST) reduce 108) ((T_CONSTANT_ENCAPSED_STRING) reduce 108) ((T_CONTINUE) reduce 108) ((T_DEC) reduce 108) ((T_DECLARE) reduce 108) ((T_DIR) reduce 108) ((T_DNUMBER) reduce 108) ((T_DO) reduce 108) ((T_DOUBLE_CAST) reduce 108) ((T_ECHO) reduce 108) ((T_EMPTY) reduce 108) ((T_ENUM) reduce 108) ((T_EVAL) reduce 108) ((T_EXIT) reduce 108) ((T_FILE) reduce 108) ((T_FINAL) reduce 108) ((T_FN) reduce 108) ((T_FOR) reduce 108) ((T_FOREACH) reduce 108) ((T_FUNCTION) reduce 108) ((T_FUNC_C) reduce 108) ((T_GLOBAL) reduce 108) ((T_GOTO) reduce 108) ((T_HALT_COMPILER) reduce 108) ((T_IF) reduce 108) ((T_INC) reduce 108) ((T_INCLUDE) reduce 108) ((T_INCLUDE_ONCE) reduce 108) ((T_INLINE_HTML) reduce 108) ((T_INTERFACE) reduce 108) ((T_INT_CAST) reduce 108) ((T_ISSET) reduce 108) ((T_LINE) reduce 108) ((T_LIST) reduce 108) ((T_LNUMBER) reduce 108) ((T_MATCH) reduce 108) ((T_METHOD_C) reduce 108) ((T_NAMESPACE) reduce 108) ((T_NAME_FULLY_QUALIFIED) reduce 108) ((T_NAME_QUALIFIED) reduce 108) ((T_NAME_RELATIVE) reduce 108) ((T_NEW) reduce 108) ((T_NS_C) reduce 108) ((T_OBJECT_CAST) reduce 108) ((T_PRINT) reduce 108) ((T_PROPERTY_C) reduce 108) ((T_READONLY) reduce 108) ((T_REQUIRE) reduce 108) ((T_REQUIRE_ONCE) reduce 108) ((T_RETURN) reduce 108) ((T_START_HEREDOC) reduce 108) ((T_STATIC) reduce 108) ((T_STRING) reduce 108) ((T_STRING_CAST) reduce 108) ((T_SWITCH) reduce 108) ((T_THROW) reduce 108) ((T_TRAIT) reduce 108) ((T_TRAIT_C) reduce 108) ((T_TRY) reduce 108) ((T_UNSET) reduce 108) ((T_UNSET_CAST) reduce 108) ((T_USE) reduce 108) ((T_VARIABLE) reduce 108) ((T_VOID_CAST) reduce 108) ((T_WHILE) reduce 108) ((T_YIELD) reduce 108) ((T_YIELD_FROM) reduce 108) (("[") reduce 108) (("`") reduce 108) (("{") reduce 108) (("}") reduce 108) (("~") reduce 108)) 32396 ((("!") reduce 111) (("\"") reduce 111) (("$") reduce 111) (("(") reduce 111) (("+") reduce 111) (("-") reduce 111) ((";") reduce 111) (("@") reduce 111) ((T_ABSTRACT) reduce 111) ((T_ARRAY) reduce 111) ((T_ARRAY_CAST) reduce 111) ((T_ATTRIBUTE) reduce 111) ((T_BOOL_CAST) reduce 111) ((T_BREAK) reduce 111) ((T_CLASS) reduce 111) ((T_CLASS_C) reduce 111) ((T_CLONE) reduce 111) ((T_CONST) reduce 111) ((T_CONSTANT_ENCAPSED_STRING) reduce 111) ((T_CONTINUE) reduce 111) ((T_DEC) reduce 111) ((T_DECLARE) reduce 111) ((T_DIR) reduce 111) ((T_DNUMBER) reduce 111) ((T_DO) reduce 111) ((T_DOUBLE_CAST) reduce 111) ((T_ECHO) reduce 111) ((T_EMPTY) reduce 111) ((T_ENUM) reduce 111) ((T_EVAL) reduce 111) ((T_EXIT) reduce 111) ((T_FILE) reduce 111) ((T_FINAL) reduce 111) ((T_FN) reduce 111) ((T_FOR) reduce 111) ((T_FOREACH) reduce 111) ((T_FUNCTION) reduce 111) ((T_FUNC_C) reduce 111) ((T_GLOBAL) reduce 111) ((T_GOTO) reduce 111) ((T_HALT_COMPILER) reduce 111) ((T_IF) reduce 111) ((T_INC) reduce 111) ((T_INCLUDE) reduce 111) ((T_INCLUDE_ONCE) reduce 111) ((T_INLINE_HTML) reduce 111) ((T_INTERFACE) reduce 111) ((T_INT_CAST) reduce 111) ((T_ISSET) reduce 111) ((T_LINE) reduce 111) ((T_LIST) reduce 111) ((T_LNUMBER) reduce 111) ((T_MATCH) reduce 111) ((T_METHOD_C) reduce 111) ((T_NAMESPACE) reduce 111) ((T_NAME_FULLY_QUALIFIED) reduce 111) ((T_NAME_QUALIFIED) reduce 111) ((T_NAME_RELATIVE) reduce 111) ((T_NEW) reduce 111) ((T_NS_C) reduce 111) ((T_OBJECT_CAST) reduce 111) ((T_PRINT) reduce 111) ((T_PROPERTY_C) reduce 111) ((T_READONLY) reduce 111) ((T_REQUIRE) reduce 111) ((T_REQUIRE_ONCE) reduce 111) ((T_RETURN) reduce 111) ((T_START_HEREDOC) reduce 111) ((T_STATIC) reduce 111) ((T_STRING) reduce 111) ((T_STRING_CAST) reduce 111) ((T_SWITCH) reduce 111) ((T_THROW) reduce 111) ((T_TRAIT) reduce 111) ((T_TRAIT_C) reduce 111) ((T_TRY) reduce 111) ((T_UNSET) reduce 111) ((T_UNSET_CAST) reduce 111) ((T_USE) reduce 111) ((T_VARIABLE) reduce 111) ((T_VOID_CAST) reduce 111) ((T_WHILE) reduce 111) ((T_YIELD) reduce 111) ((T_YIELD_FROM) reduce 111) (("[") reduce 111) (("`") reduce 111) (("{") reduce 111) (("}") reduce 111) (("~") reduce 111)) 32398 ((("!") reduce 104) (("\"") reduce 104) (("$") reduce 104) (("(") reduce 104) (("+") reduce 104) (("-") reduce 104) ((";") reduce 104) (("@") reduce 104) ((T_ABSTRACT) reduce 104) ((T_ARRAY) reduce 104) ((T_ARRAY_CAST) reduce 104) ((T_ATTRIBUTE) reduce 104) ((T_BOOL_CAST) reduce 104) ((T_BREAK) reduce 104) ((T_CLASS) reduce 104) ((T_CLASS_C) reduce 104) ((T_CLONE) reduce 104) ((T_CONST) reduce 104) ((T_CONSTANT_ENCAPSED_STRING) reduce 104) ((T_CONTINUE) reduce 104) ((T_DEC) reduce 104) ((T_DECLARE) reduce 104) ((T_DIR) reduce 104) ((T_DNUMBER) reduce 104) ((T_DO) reduce 104) ((T_DOUBLE_CAST) reduce 104) ((T_ECHO) reduce 104) ((T_EMPTY) reduce 104) ((T_ENUM) reduce 104) ((T_EVAL) reduce 104) ((T_EXIT) reduce 104) ((T_FILE) reduce 104) ((T_FINAL) reduce 104) ((T_FN) reduce 104) ((T_FOR) reduce 104) ((T_FOREACH) reduce 104) ((T_FUNCTION) reduce 104) ((T_FUNC_C) reduce 104) ((T_GLOBAL) reduce 104) ((T_GOTO) reduce 104) ((T_HALT_COMPILER) reduce 104) ((T_IF) reduce 104) ((T_INC) reduce 104) ((T_INCLUDE) reduce 104) ((T_INCLUDE_ONCE) reduce 104) ((T_INLINE_HTML) reduce 104) ((T_INTERFACE) reduce 104) ((T_INT_CAST) reduce 104) ((T_ISSET) reduce 104) ((T_LINE) reduce 104) ((T_LIST) reduce 104) ((T_LNUMBER) reduce 104) ((T_MATCH) reduce 104) ((T_METHOD_C) reduce 104) ((T_NAMESPACE) reduce 104) ((T_NAME_FULLY_QUALIFIED) reduce 104) ((T_NAME_QUALIFIED) reduce 104) ((T_NAME_RELATIVE) reduce 104) ((T_NEW) reduce 104) ((T_NS_C) reduce 104) ((T_OBJECT_CAST) reduce 104) ((T_PRINT) reduce 104) ((T_PROPERTY_C) reduce 104) ((T_READONLY) reduce 104) ((T_REQUIRE) reduce 104) ((T_REQUIRE_ONCE) reduce 104) ((T_RETURN) reduce 104) ((T_START_HEREDOC) reduce 104) ((T_STATIC) reduce 104) ((T_STRING) reduce 104) ((T_STRING_CAST) reduce 104) ((T_SWITCH) reduce 104) ((T_THROW) reduce 104) ((T_TRAIT) reduce 104) ((T_TRAIT_C) reduce 104) ((T_TRY) reduce 104) ((T_UNSET) reduce 104) ((T_UNSET_CAST) reduce 104) ((T_USE) reduce 104) ((T_VARIABLE) reduce 104) ((T_VOID_CAST) reduce 104) ((T_WHILE) reduce 104) ((T_YIELD) reduce 104) ((T_YIELD_FROM) reduce 104) (("[") reduce 104) (("`") reduce 104) (("{") reduce 104) (("}") reduce 104) (("~") reduce 104)) 32400 ((("!") reduce 107) (("\"") reduce 107) (("$") reduce 107) (("(") reduce 107) (("+") reduce 107) (("-") reduce 107) ((";") reduce 107) (("@") reduce 107) ((T_ABSTRACT) reduce 107) ((T_ARRAY) reduce 107) ((T_ARRAY_CAST) reduce 107) ((T_ATTRIBUTE) reduce 107) ((T_BOOL_CAST) reduce 107) ((T_BREAK) reduce 107) ((T_CLASS) reduce 107) ((T_CLASS_C) reduce 107) ((T_CLONE) reduce 107) ((T_CONST) reduce 107) ((T_CONSTANT_ENCAPSED_STRING) reduce 107) ((T_CONTINUE) reduce 107) ((T_DEC) reduce 107) ((T_DECLARE) reduce 107) ((T_DIR) reduce 107) ((T_DNUMBER) reduce 107) ((T_DO) reduce 107) ((T_DOUBLE_CAST) reduce 107) ((T_ECHO) reduce 107) ((T_EMPTY) reduce 107) ((T_ENUM) reduce 107) ((T_EVAL) reduce 107) ((T_EXIT) reduce 107) ((T_FILE) reduce 107) ((T_FINAL) reduce 107) ((T_FN) reduce 107) ((T_FOR) reduce 107) ((T_FOREACH) reduce 107) ((T_FUNCTION) reduce 107) ((T_FUNC_C) reduce 107) ((T_GLOBAL) reduce 107) ((T_GOTO) reduce 107) ((T_HALT_COMPILER) reduce 107) ((T_IF) reduce 107) ((T_INC) reduce 107) ((T_INCLUDE) reduce 107) ((T_INCLUDE_ONCE) reduce 107) ((T_INLINE_HTML) reduce 107) ((T_INTERFACE) reduce 107) ((T_INT_CAST) reduce 107) ((T_ISSET) reduce 107) ((T_LINE) reduce 107) ((T_LIST) reduce 107) ((T_LNUMBER) reduce 107) ((T_MATCH) reduce 107) ((T_METHOD_C) reduce 107) ((T_NAMESPACE) reduce 107) ((T_NAME_FULLY_QUALIFIED) reduce 107) ((T_NAME_QUALIFIED) reduce 107) ((T_NAME_RELATIVE) reduce 107) ((T_NEW) reduce 107) ((T_NS_C) reduce 107) ((T_OBJECT_CAST) reduce 107) ((T_PRINT) reduce 107) ((T_PROPERTY_C) reduce 107) ((T_READONLY) reduce 107) ((T_REQUIRE) reduce 107) ((T_REQUIRE_ONCE) reduce 107) ((T_RETURN) reduce 107) ((T_START_HEREDOC) reduce 107) ((T_STATIC) reduce 107) ((T_STRING) reduce 107) ((T_STRING_CAST) reduce 107) ((T_SWITCH) reduce 107) ((T_THROW) reduce 107) ((T_TRAIT) reduce 107) ((T_TRAIT_C) reduce 107) ((T_TRY) reduce 107) ((T_UNSET) reduce 107) ((T_UNSET_CAST) reduce 107) ((T_USE) reduce 107) ((T_VARIABLE) reduce 107) ((T_VOID_CAST) reduce 107) ((T_WHILE) reduce 107) ((T_YIELD) reduce 107) ((T_YIELD_FROM) reduce 107) (("[") reduce 107) (("`") reduce 107) (("{") reduce 107) (("}") reduce 107) (("~") reduce 107)) 32403 ((("!") reduce 103) (("\"") reduce 103) (("$") reduce 103) (("(") reduce 103) (("+") reduce 103) (("-") reduce 103) ((";") reduce 103) (("@") reduce 103) ((T_ABSTRACT) reduce 103) ((T_ARRAY) reduce 103) ((T_ARRAY_CAST) reduce 103) ((T_ATTRIBUTE) reduce 103) ((T_BOOL_CAST) reduce 103) ((T_BREAK) reduce 103) ((T_CLASS) reduce 103) ((T_CLASS_C) reduce 103) ((T_CLONE) reduce 103) ((T_CONST) reduce 103) ((T_CONSTANT_ENCAPSED_STRING) reduce 103) ((T_CONTINUE) reduce 103) ((T_DEC) reduce 103) ((T_DECLARE) reduce 103) ((T_DIR) reduce 103) ((T_DNUMBER) reduce 103) ((T_DO) reduce 103) ((T_DOUBLE_CAST) reduce 103) ((T_ECHO) reduce 103) ((T_EMPTY) reduce 103) ((T_ENUM) reduce 103) ((T_EVAL) reduce 103) ((T_EXIT) reduce 103) ((T_FILE) reduce 103) ((T_FINAL) reduce 103) ((T_FN) reduce 103) ((T_FOR) reduce 103) ((T_FOREACH) reduce 103) ((T_FUNCTION) reduce 103) ((T_FUNC_C) reduce 103) ((T_GLOBAL) reduce 103) ((T_GOTO) reduce 103) ((T_HALT_COMPILER) reduce 103) ((T_IF) reduce 103) ((T_INC) reduce 103) ((T_INCLUDE) reduce 103) ((T_INCLUDE_ONCE) reduce 103) ((T_INLINE_HTML) reduce 103) ((T_INTERFACE) reduce 103) ((T_INT_CAST) reduce 103) ((T_ISSET) reduce 103) ((T_LINE) reduce 103) ((T_LIST) reduce 103) ((T_LNUMBER) reduce 103) ((T_MATCH) reduce 103) ((T_METHOD_C) reduce 103) ((T_NAMESPACE) reduce 103) ((T_NAME_FULLY_QUALIFIED) reduce 103) ((T_NAME_QUALIFIED) reduce 103) ((T_NAME_RELATIVE) reduce 103) ((T_NEW) reduce 103) ((T_NS_C) reduce 103) ((T_OBJECT_CAST) reduce 103) ((T_PRINT) reduce 103) ((T_PROPERTY_C) reduce 103) ((T_READONLY) reduce 103) ((T_REQUIRE) reduce 103) ((T_REQUIRE_ONCE) reduce 103) ((T_RETURN) reduce 103) ((T_START_HEREDOC) reduce 103) ((T_STATIC) reduce 103) ((T_STRING) reduce 103) ((T_STRING_CAST) reduce 103) ((T_SWITCH) reduce 103) ((T_THROW) reduce 103) ((T_TRAIT) reduce 103) ((T_TRAIT_C) reduce 103) ((T_TRY) reduce 103) ((T_UNSET) reduce 103) ((T_UNSET_CAST) reduce 103) ((T_USE) reduce 103) ((T_VARIABLE) reduce 103) ((T_VOID_CAST) reduce 103) ((T_WHILE) reduce 103) ((T_YIELD) reduce 103) ((T_YIELD_FROM) reduce 103) (("[") reduce 103) (("`") reduce 103) (("{") reduce 103) (("}") reduce 103) (("~") reduce 103)) 32404 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONST) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_EMPTY) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAMESPACE) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_USE) reduce 148) ((T_VARIABLE) reduce 148) ((T_VOID_CAST) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("}") reduce 148) (("~") reduce 148)) 32405 ((("!") reduce 246) (("\"") reduce 246) (("$") reduce 246) (("(") reduce 246) (("+") reduce 246) (("-") reduce 246) ((";") reduce 246) (("@") reduce 246) ((T_ABSTRACT) reduce 246) ((T_ARRAY) reduce 246) ((T_ARRAY_CAST) reduce 246) ((T_ATTRIBUTE) reduce 246) ((T_BOOL_CAST) reduce 246) ((T_BREAK) reduce 246) ((T_CLASS) reduce 246) ((T_CLASS_C) reduce 246) ((T_CLONE) reduce 246) ((T_CONST) reduce 246) ((T_CONSTANT_ENCAPSED_STRING) reduce 246) ((T_CONTINUE) reduce 246) ((T_DEC) reduce 246) ((T_DECLARE) reduce 246) ((T_DIR) reduce 246) ((T_DNUMBER) reduce 246) ((T_DO) reduce 246) ((T_DOUBLE_CAST) reduce 246) ((T_ECHO) reduce 246) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) reduce 246) ((T_ENUM) reduce 246) ((T_EVAL) reduce 246) ((T_EXIT) reduce 246) ((T_FILE) reduce 246) ((T_FINAL) reduce 246) ((T_FN) reduce 246) ((T_FOR) reduce 246) ((T_FOREACH) reduce 246) ((T_FUNCTION) reduce 246) ((T_FUNC_C) reduce 246) ((T_GLOBAL) reduce 246) ((T_GOTO) reduce 246) ((T_HALT_COMPILER) reduce 246) ((T_IF) reduce 246) ((T_INC) reduce 246) ((T_INCLUDE) reduce 246) ((T_INCLUDE_ONCE) reduce 246) ((T_INLINE_HTML) reduce 246) ((T_INTERFACE) reduce 246) ((T_INT_CAST) reduce 246) ((T_ISSET) reduce 246) ((T_LINE) reduce 246) ((T_LIST) reduce 246) ((T_LNUMBER) reduce 246) ((T_MATCH) reduce 246) ((T_METHOD_C) reduce 246) ((T_NAMESPACE) reduce 246) ((T_NAME_FULLY_QUALIFIED) reduce 246) ((T_NAME_QUALIFIED) reduce 246) ((T_NAME_RELATIVE) reduce 246) ((T_NEW) reduce 246) ((T_NS_C) reduce 246) ((T_OBJECT_CAST) reduce 246) ((T_PRINT) reduce 246) ((T_PROPERTY_C) reduce 246) ((T_READONLY) reduce 246) ((T_REQUIRE) reduce 246) ((T_REQUIRE_ONCE) reduce 246) ((T_RETURN) reduce 246) ((T_START_HEREDOC) reduce 246) ((T_STATIC) reduce 246) ((T_STRING) reduce 246) ((T_STRING_CAST) reduce 246) ((T_SWITCH) reduce 246) ((T_THROW) reduce 246) ((T_TRAIT) reduce 246) ((T_TRAIT_C) reduce 246) ((T_TRY) reduce 246) ((T_UNSET) reduce 246) ((T_UNSET_CAST) reduce 246) ((T_USE) reduce 246) ((T_VARIABLE) reduce 246) ((T_VOID_CAST) reduce 246) ((T_WHILE) reduce 246) ((T_YIELD) reduce 246) ((T_YIELD_FROM) reduce 246) (("[") reduce 246) (("`") reduce 246) (("{") reduce 246) (("}") reduce 246) (("~") reduce 246)) 32406 ((("!") reduce 106) (("\"") reduce 106) (("$") reduce 106) (("(") reduce 106) (("+") reduce 106) (("-") reduce 106) ((";") reduce 106) (("@") reduce 106) ((T_ABSTRACT) reduce 106) ((T_ARRAY) reduce 106) ((T_ARRAY_CAST) reduce 106) ((T_ATTRIBUTE) reduce 106) ((T_BOOL_CAST) reduce 106) ((T_BREAK) reduce 106) ((T_CLASS) reduce 106) ((T_CLASS_C) reduce 106) ((T_CLONE) reduce 106) ((T_CONST) reduce 106) ((T_CONSTANT_ENCAPSED_STRING) reduce 106) ((T_CONTINUE) reduce 106) ((T_DEC) reduce 106) ((T_DECLARE) reduce 106) ((T_DIR) reduce 106) ((T_DNUMBER) reduce 106) ((T_DO) reduce 106) ((T_DOUBLE_CAST) reduce 106) ((T_ECHO) reduce 106) ((T_EMPTY) reduce 106) ((T_ENUM) reduce 106) ((T_EVAL) reduce 106) ((T_EXIT) reduce 106) ((T_FILE) reduce 106) ((T_FINAL) reduce 106) ((T_FN) reduce 106) ((T_FOR) reduce 106) ((T_FOREACH) reduce 106) ((T_FUNCTION) reduce 106) ((T_FUNC_C) reduce 106) ((T_GLOBAL) reduce 106) ((T_GOTO) reduce 106) ((T_HALT_COMPILER) reduce 106) ((T_IF) reduce 106) ((T_INC) reduce 106) ((T_INCLUDE) reduce 106) ((T_INCLUDE_ONCE) reduce 106) ((T_INLINE_HTML) reduce 106) ((T_INTERFACE) reduce 106) ((T_INT_CAST) reduce 106) ((T_ISSET) reduce 106) ((T_LINE) reduce 106) ((T_LIST) reduce 106) ((T_LNUMBER) reduce 106) ((T_MATCH) reduce 106) ((T_METHOD_C) reduce 106) ((T_NAMESPACE) reduce 106) ((T_NAME_FULLY_QUALIFIED) reduce 106) ((T_NAME_QUALIFIED) reduce 106) ((T_NAME_RELATIVE) reduce 106) ((T_NEW) reduce 106) ((T_NS_C) reduce 106) ((T_OBJECT_CAST) reduce 106) ((T_PRINT) reduce 106) ((T_PROPERTY_C) reduce 106) ((T_READONLY) reduce 106) ((T_REQUIRE) reduce 106) ((T_REQUIRE_ONCE) reduce 106) ((T_RETURN) reduce 106) ((T_START_HEREDOC) reduce 106) ((T_STATIC) reduce 106) ((T_STRING) reduce 106) ((T_STRING_CAST) reduce 106) ((T_SWITCH) reduce 106) ((T_THROW) reduce 106) ((T_TRAIT) reduce 106) ((T_TRAIT_C) reduce 106) ((T_TRY) reduce 106) ((T_UNSET) reduce 106) ((T_UNSET_CAST) reduce 106) ((T_USE) reduce 106) ((T_VARIABLE) reduce 106) ((T_VOID_CAST) reduce 106) ((T_WHILE) reduce 106) ((T_YIELD) reduce 106) ((T_YIELD_FROM) reduce 106) (("[") reduce 106) (("`") reduce 106) (("{") reduce 106) (("}") reduce 106) (("~") reduce 106)) 32407 ((("!") reduce 110) (("\"") reduce 110) (("$") reduce 110) (("(") reduce 110) (("+") reduce 110) (("-") reduce 110) ((";") reduce 110) (("@") reduce 110) ((T_ABSTRACT) reduce 110) ((T_ARRAY) reduce 110) ((T_ARRAY_CAST) reduce 110) ((T_ATTRIBUTE) reduce 110) ((T_BOOL_CAST) reduce 110) ((T_BREAK) reduce 110) ((T_CLASS) reduce 110) ((T_CLASS_C) reduce 110) ((T_CLONE) reduce 110) ((T_CONST) reduce 110) ((T_CONSTANT_ENCAPSED_STRING) reduce 110) ((T_CONTINUE) reduce 110) ((T_DEC) reduce 110) ((T_DECLARE) reduce 110) ((T_DIR) reduce 110) ((T_DNUMBER) reduce 110) ((T_DO) reduce 110) ((T_DOUBLE_CAST) reduce 110) ((T_ECHO) reduce 110) ((T_EMPTY) reduce 110) ((T_ENUM) reduce 110) ((T_EVAL) reduce 110) ((T_EXIT) reduce 110) ((T_FILE) reduce 110) ((T_FINAL) reduce 110) ((T_FN) reduce 110) ((T_FOR) reduce 110) ((T_FOREACH) reduce 110) ((T_FUNCTION) reduce 110) ((T_FUNC_C) reduce 110) ((T_GLOBAL) reduce 110) ((T_GOTO) reduce 110) ((T_HALT_COMPILER) reduce 110) ((T_IF) reduce 110) ((T_INC) reduce 110) ((T_INCLUDE) reduce 110) ((T_INCLUDE_ONCE) reduce 110) ((T_INLINE_HTML) reduce 110) ((T_INTERFACE) reduce 110) ((T_INT_CAST) reduce 110) ((T_ISSET) reduce 110) ((T_LINE) reduce 110) ((T_LIST) reduce 110) ((T_LNUMBER) reduce 110) ((T_MATCH) reduce 110) ((T_METHOD_C) reduce 110) ((T_NAMESPACE) reduce 110) ((T_NAME_FULLY_QUALIFIED) reduce 110) ((T_NAME_QUALIFIED) reduce 110) ((T_NAME_RELATIVE) reduce 110) ((T_NEW) reduce 110) ((T_NS_C) reduce 110) ((T_OBJECT_CAST) reduce 110) ((T_PRINT) reduce 110) ((T_PROPERTY_C) reduce 110) ((T_READONLY) reduce 110) ((T_REQUIRE) reduce 110) ((T_REQUIRE_ONCE) reduce 110) ((T_RETURN) reduce 110) ((T_START_HEREDOC) reduce 110) ((T_STATIC) reduce 110) ((T_STRING) reduce 110) ((T_STRING_CAST) reduce 110) ((T_SWITCH) reduce 110) ((T_THROW) reduce 110) ((T_TRAIT) reduce 110) ((T_TRAIT_C) reduce 110) ((T_TRY) reduce 110) ((T_UNSET) reduce 110) ((T_UNSET_CAST) reduce 110) ((T_USE) reduce 110) ((T_VARIABLE) reduce 110) ((T_VOID_CAST) reduce 110) ((T_WHILE) reduce 110) ((T_YIELD) reduce 110) ((T_YIELD_FROM) reduce 110) (("[") reduce 110) (("`") reduce 110) (("{") reduce 110) (("}") reduce 110) (("~") reduce 110)) 32408 ((("!") reduce 84) (("\"") reduce 84) (("$") reduce 84) (("(") reduce 84) (("+") reduce 84) (("-") reduce 84) ((";") reduce 84) (("@") reduce 84) ((T_ABSTRACT) reduce 84) ((T_ARRAY) reduce 84) ((T_ARRAY_CAST) reduce 84) ((T_ATTRIBUTE) reduce 84) ((T_BOOL_CAST) reduce 84) ((T_BREAK) reduce 84) ((T_CLASS) reduce 84) ((T_CLASS_C) reduce 84) ((T_CLONE) reduce 84) ((T_CONST) reduce 84) ((T_CONSTANT_ENCAPSED_STRING) reduce 84) ((T_CONTINUE) reduce 84) ((T_DEC) reduce 84) ((T_DECLARE) reduce 84) ((T_DIR) reduce 84) ((T_DNUMBER) reduce 84) ((T_DO) reduce 84) ((T_DOUBLE_CAST) reduce 84) ((T_ECHO) reduce 84) ((T_EMPTY) reduce 84) ((T_ENUM) reduce 84) ((T_EVAL) reduce 84) ((T_EXIT) reduce 84) ((T_FILE) reduce 84) ((T_FINAL) reduce 84) ((T_FN) reduce 84) ((T_FOR) reduce 84) ((T_FOREACH) reduce 84) ((T_FUNCTION) reduce 84) ((T_FUNC_C) reduce 84) ((T_GLOBAL) reduce 84) ((T_GOTO) reduce 84) ((T_HALT_COMPILER) reduce 84) ((T_IF) reduce 84) ((T_INC) reduce 84) ((T_INCLUDE) reduce 84) ((T_INCLUDE_ONCE) reduce 84) ((T_INLINE_HTML) reduce 84) ((T_INTERFACE) reduce 84) ((T_INT_CAST) reduce 84) ((T_ISSET) reduce 84) ((T_LINE) reduce 84) ((T_LIST) reduce 84) ((T_LNUMBER) reduce 84) ((T_MATCH) reduce 84) ((T_METHOD_C) reduce 84) ((T_NAMESPACE) reduce 84) ((T_NAME_FULLY_QUALIFIED) reduce 84) ((T_NAME_QUALIFIED) reduce 84) ((T_NAME_RELATIVE) reduce 84) ((T_NEW) reduce 84) ((T_NS_C) reduce 84) ((T_OBJECT_CAST) reduce 84) ((T_PRINT) reduce 84) ((T_PROPERTY_C) reduce 84) ((T_READONLY) reduce 84) ((T_REQUIRE) reduce 84) ((T_REQUIRE_ONCE) reduce 84) ((T_RETURN) reduce 84) ((T_START_HEREDOC) reduce 84) ((T_STATIC) reduce 84) ((T_STRING) reduce 84) ((T_STRING_CAST) reduce 84) ((T_SWITCH) reduce 84) ((T_THROW) reduce 84) ((T_TRAIT) reduce 84) ((T_TRAIT_C) reduce 84) ((T_TRY) reduce 84) ((T_UNSET) reduce 84) ((T_UNSET_CAST) reduce 84) ((T_USE) reduce 84) ((T_VARIABLE) reduce 84) ((T_VOID_CAST) reduce 84) ((T_WHILE) reduce 84) ((T_YIELD) reduce 84) ((T_YIELD_FROM) reduce 84) (("[") reduce 84) (("`") reduce 84) (("{") reduce 84) (("}") reduce 84) (("~") reduce 84)) 32409 ((("!") reduce 105) (("\"") reduce 105) (("$") reduce 105) (("(") reduce 105) (("+") reduce 105) (("-") reduce 105) ((";") reduce 105) (("@") reduce 105) ((T_ABSTRACT) reduce 105) ((T_ARRAY) reduce 105) ((T_ARRAY_CAST) reduce 105) ((T_ATTRIBUTE) reduce 105) ((T_BOOL_CAST) reduce 105) ((T_BREAK) reduce 105) ((T_CLASS) reduce 105) ((T_CLASS_C) reduce 105) ((T_CLONE) reduce 105) ((T_CONST) reduce 105) ((T_CONSTANT_ENCAPSED_STRING) reduce 105) ((T_CONTINUE) reduce 105) ((T_DEC) reduce 105) ((T_DECLARE) reduce 105) ((T_DIR) reduce 105) ((T_DNUMBER) reduce 105) ((T_DO) reduce 105) ((T_DOUBLE_CAST) reduce 105) ((T_ECHO) reduce 105) ((T_EMPTY) reduce 105) ((T_ENUM) reduce 105) ((T_EVAL) reduce 105) ((T_EXIT) reduce 105) ((T_FILE) reduce 105) ((T_FINAL) reduce 105) ((T_FN) reduce 105) ((T_FOR) reduce 105) ((T_FOREACH) reduce 105) ((T_FUNCTION) reduce 105) ((T_FUNC_C) reduce 105) ((T_GLOBAL) reduce 105) ((T_GOTO) reduce 105) ((T_HALT_COMPILER) reduce 105) ((T_IF) reduce 105) ((T_INC) reduce 105) ((T_INCLUDE) reduce 105) ((T_INCLUDE_ONCE) reduce 105) ((T_INLINE_HTML) reduce 105) ((T_INTERFACE) reduce 105) ((T_INT_CAST) reduce 105) ((T_ISSET) reduce 105) ((T_LINE) reduce 105) ((T_LIST) reduce 105) ((T_LNUMBER) reduce 105) ((T_MATCH) reduce 105) ((T_METHOD_C) reduce 105) ((T_NAMESPACE) reduce 105) ((T_NAME_FULLY_QUALIFIED) reduce 105) ((T_NAME_QUALIFIED) reduce 105) ((T_NAME_RELATIVE) reduce 105) ((T_NEW) reduce 105) ((T_NS_C) reduce 105) ((T_OBJECT_CAST) reduce 105) ((T_PRINT) reduce 105) ((T_PROPERTY_C) reduce 105) ((T_READONLY) reduce 105) ((T_REQUIRE) reduce 105) ((T_REQUIRE_ONCE) reduce 105) ((T_RETURN) reduce 105) ((T_START_HEREDOC) reduce 105) ((T_STATIC) reduce 105) ((T_STRING) reduce 105) ((T_STRING_CAST) reduce 105) ((T_SWITCH) reduce 105) ((T_THROW) reduce 105) ((T_TRAIT) reduce 105) ((T_TRAIT_C) reduce 105) ((T_TRY) reduce 105) ((T_UNSET) reduce 105) ((T_UNSET_CAST) reduce 105) ((T_USE) reduce 105) ((T_VARIABLE) reduce 105) ((T_VOID_CAST) reduce 105) ((T_WHILE) reduce 105) ((T_YIELD) reduce 105) ((T_YIELD_FROM) reduce 105) (("[") reduce 105) (("`") reduce 105) (("{") reduce 105) (("}") reduce 105) (("~") reduce 105)) 32411 ((("!") reduce 116) (("\"") reduce 116) (($) reduce 116) (("$") reduce 116) (("(") reduce 116) (("+") reduce 116) (("-") reduce 116) ((";") reduce 116) (("@") reduce 116) ((T_ABSTRACT) reduce 116) ((T_ARRAY) reduce 116) ((T_ARRAY_CAST) reduce 116) ((T_ATTRIBUTE) reduce 116) ((T_BOOL_CAST) reduce 116) ((T_BREAK) reduce 116) ((T_CLASS) reduce 116) ((T_CLASS_C) reduce 116) ((T_CLONE) reduce 116) ((T_CONST) reduce 116) ((T_CONSTANT_ENCAPSED_STRING) reduce 116) ((T_CONTINUE) reduce 116) ((T_DEC) reduce 116) ((T_DECLARE) reduce 116) ((T_DIR) reduce 116) ((T_DNUMBER) reduce 116) ((T_DO) reduce 116) ((T_DOUBLE_CAST) reduce 116) ((T_ECHO) reduce 116) ((T_EMPTY) reduce 116) ((T_ENUM) reduce 116) ((T_EVAL) reduce 116) ((T_EXIT) reduce 116) ((T_FILE) reduce 116) ((T_FINAL) reduce 116) ((T_FN) reduce 116) ((T_FOR) reduce 116) ((T_FOREACH) reduce 116) ((T_FUNCTION) reduce 116) ((T_FUNC_C) reduce 116) ((T_GLOBAL) reduce 116) ((T_GOTO) reduce 116) ((T_HALT_COMPILER) reduce 116) ((T_IF) reduce 116) ((T_INC) reduce 116) ((T_INCLUDE) reduce 116) ((T_INCLUDE_ONCE) reduce 116) ((T_INLINE_HTML) reduce 116) ((T_INTERFACE) reduce 116) ((T_INT_CAST) reduce 116) ((T_ISSET) reduce 116) ((T_LINE) reduce 116) ((T_LIST) reduce 116) ((T_LNUMBER) reduce 116) ((T_MATCH) reduce 116) ((T_METHOD_C) reduce 116) ((T_NAMESPACE) reduce 116) ((T_NAME_FULLY_QUALIFIED) reduce 116) ((T_NAME_QUALIFIED) reduce 116) ((T_NAME_RELATIVE) reduce 116) ((T_NEW) reduce 116) ((T_NS_C) reduce 116) ((T_OBJECT_CAST) reduce 116) ((T_PRINT) reduce 116) ((T_PROPERTY_C) reduce 116) ((T_READONLY) reduce 116) ((T_REQUIRE) reduce 116) ((T_REQUIRE_ONCE) reduce 116) ((T_RETURN) reduce 116) ((T_START_HEREDOC) reduce 116) ((T_STATIC) reduce 116) ((T_STRING) reduce 116) ((T_STRING_CAST) reduce 116) ((T_SWITCH) reduce 116) ((T_THROW) reduce 116) ((T_TRAIT) reduce 116) ((T_TRAIT_C) reduce 116) ((T_TRY) reduce 116) ((T_UNSET) reduce 116) ((T_UNSET_CAST) reduce 116) ((T_USE) reduce 116) ((T_VARIABLE) reduce 116) ((T_VOID_CAST) reduce 116) ((T_WHILE) reduce 116) ((T_YIELD) reduce 116) ((T_YIELD_FROM) reduce 116) (("[") reduce 116) (("`") reduce 116) (("{") reduce 116) (("~") reduce 116)) 32413 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONST) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_EMPTY) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAMESPACE) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_USE) reduce 147) ((T_VARIABLE) reduce 147) ((T_VOID_CAST) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("}") reduce 147) (("~") reduce 147)) 32419 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONST) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_ELSE) reduce 166) ((T_ELSEIF) reduce 166) ((T_EMPTY) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAMESPACE) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_USE) reduce 166) ((T_VARIABLE) reduce 166) ((T_VOID_CAST) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("}") reduce 166) (("~") reduce 166)) 32429 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONST) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_ELSE) reduce 160) ((T_ELSEIF) reduce 160) ((T_EMPTY) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAMESPACE) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_USE) reduce 160) ((T_VARIABLE) reduce 160) ((T_VOID_CAST) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("}") reduce 160) (("~") reduce 160)) 32438 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONST) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_ELSE) reduce 149) ((T_ELSEIF) reduce 149) ((T_EMPTY) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAMESPACE) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_USE) reduce 149) ((T_VARIABLE) reduce 149) ((T_VOID_CAST) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("}") reduce 149) (("~") reduce 149)) 32441 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONST) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_ELSE) reduce 148) ((T_ELSEIF) reduce 148) ((T_EMPTY) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAMESPACE) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_USE) reduce 148) ((T_VARIABLE) reduce 148) ((T_VOID_CAST) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("}") reduce 148) (("~") reduce 148)) 32443 ((("!") reduce 245) (("\"") reduce 245) (("$") reduce 245) (("(") reduce 245) (("+") reduce 245) (("-") reduce 245) ((";") reduce 245) (("@") reduce 245) ((T_ABSTRACT) reduce 245) ((T_ARRAY) reduce 245) ((T_ARRAY_CAST) reduce 245) ((T_ATTRIBUTE) reduce 245) ((T_BOOL_CAST) reduce 245) ((T_BREAK) reduce 245) ((T_CLASS) reduce 245) ((T_CLASS_C) reduce 245) ((T_CLONE) reduce 245) ((T_CONST) reduce 245) ((T_CONSTANT_ENCAPSED_STRING) reduce 245) ((T_CONTINUE) reduce 245) ((T_DEC) reduce 245) ((T_DECLARE) reduce 245) ((T_DIR) reduce 245) ((T_DNUMBER) reduce 245) ((T_DO) reduce 245) ((T_DOUBLE_CAST) reduce 245) ((T_ECHO) reduce 245) ((T_ELSE) reduce 245) ((T_ELSEIF) reduce 245) ((T_EMPTY) reduce 245) ((T_ENUM) reduce 245) ((T_EVAL) reduce 245) ((T_EXIT) reduce 245) ((T_FILE) reduce 245) ((T_FINAL) reduce 245) ((T_FN) reduce 245) ((T_FOR) reduce 245) ((T_FOREACH) reduce 245) ((T_FUNCTION) reduce 245) ((T_FUNC_C) reduce 245) ((T_GLOBAL) reduce 245) ((T_GOTO) reduce 245) ((T_HALT_COMPILER) reduce 245) ((T_IF) reduce 245) ((T_INC) reduce 245) ((T_INCLUDE) reduce 245) ((T_INCLUDE_ONCE) reduce 245) ((T_INLINE_HTML) reduce 245) ((T_INTERFACE) reduce 245) ((T_INT_CAST) reduce 245) ((T_ISSET) reduce 245) ((T_LINE) reduce 245) ((T_LIST) reduce 245) ((T_LNUMBER) reduce 245) ((T_MATCH) reduce 245) ((T_METHOD_C) reduce 245) ((T_NAMESPACE) reduce 245) ((T_NAME_FULLY_QUALIFIED) reduce 245) ((T_NAME_QUALIFIED) reduce 245) ((T_NAME_RELATIVE) reduce 245) ((T_NEW) reduce 245) ((T_NS_C) reduce 245) ((T_OBJECT_CAST) reduce 245) ((T_PRINT) reduce 245) ((T_PROPERTY_C) reduce 245) ((T_READONLY) reduce 245) ((T_REQUIRE) reduce 245) ((T_REQUIRE_ONCE) reduce 245) ((T_RETURN) reduce 245) ((T_START_HEREDOC) reduce 245) ((T_STATIC) reduce 245) ((T_STRING) reduce 245) ((T_STRING_CAST) reduce 245) ((T_SWITCH) reduce 245) ((T_THROW) reduce 245) ((T_TRAIT) reduce 245) ((T_TRAIT_C) reduce 245) ((T_TRY) reduce 245) ((T_UNSET) reduce 245) ((T_UNSET_CAST) reduce 245) ((T_USE) reduce 245) ((T_VARIABLE) reduce 245) ((T_VOID_CAST) reduce 245) ((T_WHILE) reduce 245) ((T_YIELD) reduce 245) ((T_YIELD_FROM) reduce 245) (("[") reduce 245) (("`") reduce 245) (("{") reduce 245) (("}") reduce 245) (("~") reduce 245)) 32446 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONST) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_ELSE) reduce 147) ((T_ELSEIF) reduce 147) ((T_EMPTY) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAMESPACE) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_USE) reduce 147) ((T_VARIABLE) reduce 147) ((T_VOID_CAST) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("}") reduce 147) (("~") reduce 147)) 32448 ((("!") reduce 247) (("\"") reduce 247) (("$") reduce 247) (("(") reduce 247) (("+") reduce 247) (("-") reduce 247) ((";") reduce 247) (("@") reduce 247) ((T_ABSTRACT) reduce 247) ((T_ARRAY) reduce 247) ((T_ARRAY_CAST) reduce 247) ((T_ATTRIBUTE) reduce 247) ((T_BOOL_CAST) reduce 247) ((T_BREAK) reduce 247) ((T_CLASS) reduce 247) ((T_CLASS_C) reduce 247) ((T_CLONE) reduce 247) ((T_CONST) reduce 247) ((T_CONSTANT_ENCAPSED_STRING) reduce 247) ((T_CONTINUE) reduce 247) ((T_DEC) reduce 247) ((T_DECLARE) reduce 247) ((T_DIR) reduce 247) ((T_DNUMBER) reduce 247) ((T_DO) reduce 247) ((T_DOUBLE_CAST) reduce 247) ((T_ECHO) reduce 247) ((T_ELSE) reduce 247) ((T_ELSEIF) reduce 247) ((T_EMPTY) reduce 247) ((T_ENUM) reduce 247) ((T_EVAL) reduce 247) ((T_EXIT) reduce 247) ((T_FILE) reduce 247) ((T_FINAL) reduce 247) ((T_FN) reduce 247) ((T_FOR) reduce 247) ((T_FOREACH) reduce 247) ((T_FUNCTION) reduce 247) ((T_FUNC_C) reduce 247) ((T_GLOBAL) reduce 247) ((T_GOTO) reduce 247) ((T_HALT_COMPILER) reduce 247) ((T_IF) reduce 247) ((T_INC) reduce 247) ((T_INCLUDE) reduce 247) ((T_INCLUDE_ONCE) reduce 247) ((T_INLINE_HTML) reduce 247) ((T_INTERFACE) reduce 247) ((T_INT_CAST) reduce 247) ((T_ISSET) reduce 247) ((T_LINE) reduce 247) ((T_LIST) reduce 247) ((T_LNUMBER) reduce 247) ((T_MATCH) reduce 247) ((T_METHOD_C) reduce 247) ((T_NAMESPACE) reduce 247) ((T_NAME_FULLY_QUALIFIED) reduce 247) ((T_NAME_QUALIFIED) reduce 247) ((T_NAME_RELATIVE) reduce 247) ((T_NEW) reduce 247) ((T_NS_C) reduce 247) ((T_OBJECT_CAST) reduce 247) ((T_PRINT) reduce 247) ((T_PROPERTY_C) reduce 247) ((T_READONLY) reduce 247) ((T_REQUIRE) reduce 247) ((T_REQUIRE_ONCE) reduce 247) ((T_RETURN) reduce 247) ((T_START_HEREDOC) reduce 247) ((T_STATIC) reduce 247) ((T_STRING) reduce 247) ((T_STRING_CAST) reduce 247) ((T_SWITCH) reduce 247) ((T_THROW) reduce 247) ((T_TRAIT) reduce 247) ((T_TRAIT_C) reduce 247) ((T_TRY) reduce 247) ((T_UNSET) reduce 247) ((T_UNSET_CAST) reduce 247) ((T_USE) reduce 247) ((T_VARIABLE) reduce 247) ((T_VOID_CAST) reduce 247) ((T_WHILE) reduce 247) ((T_YIELD) reduce 247) ((T_YIELD_FROM) reduce 247) (("[") reduce 247) (("`") reduce 247) (("{") reduce 247) (("}") reduce 247) (("~") reduce 247)) 32449 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONST) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_ELSE) reduce 161) ((T_ELSEIF) reduce 161) ((T_EMPTY) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAMESPACE) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_USE) reduce 161) ((T_VARIABLE) reduce 161) ((T_VOID_CAST) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("}") reduce 161) (("~") reduce 161)) 32452 ((("!") reduce 250) (("\"") reduce 250) (("$") reduce 250) (("(") reduce 250) (("+") reduce 250) (("-") reduce 250) ((";") reduce 250) (("@") reduce 250) ((T_ABSTRACT) reduce 250) ((T_ARRAY) reduce 250) ((T_ARRAY_CAST) reduce 250) ((T_ATTRIBUTE) reduce 250) ((T_BOOL_CAST) reduce 250) ((T_BREAK) reduce 250) ((T_CLASS) reduce 250) ((T_CLASS_C) reduce 250) ((T_CLONE) reduce 250) ((T_CONST) reduce 250) ((T_CONSTANT_ENCAPSED_STRING) reduce 250) ((T_CONTINUE) reduce 250) ((T_DEC) reduce 250) ((T_DECLARE) reduce 250) ((T_DIR) reduce 250) ((T_DNUMBER) reduce 250) ((T_DO) reduce 250) ((T_DOUBLE_CAST) reduce 250) ((T_ECHO) reduce 250) ((T_ELSE) reduce 250) ((T_ELSEIF) reduce 250) ((T_EMPTY) reduce 250) ((T_ENUM) reduce 250) ((T_EVAL) reduce 250) ((T_EXIT) reduce 250) ((T_FILE) reduce 250) ((T_FINAL) reduce 250) ((T_FN) reduce 250) ((T_FOR) reduce 250) ((T_FOREACH) reduce 250) ((T_FUNCTION) reduce 250) ((T_FUNC_C) reduce 250) ((T_GLOBAL) reduce 250) ((T_GOTO) reduce 250) ((T_HALT_COMPILER) reduce 250) ((T_IF) reduce 250) ((T_INC) reduce 250) ((T_INCLUDE) reduce 250) ((T_INCLUDE_ONCE) reduce 250) ((T_INLINE_HTML) reduce 250) ((T_INTERFACE) reduce 250) ((T_INT_CAST) reduce 250) ((T_ISSET) reduce 250) ((T_LINE) reduce 250) ((T_LIST) reduce 250) ((T_LNUMBER) reduce 250) ((T_MATCH) reduce 250) ((T_METHOD_C) reduce 250) ((T_NAMESPACE) reduce 250) ((T_NAME_FULLY_QUALIFIED) reduce 250) ((T_NAME_QUALIFIED) reduce 250) ((T_NAME_RELATIVE) reduce 250) ((T_NEW) reduce 250) ((T_NS_C) reduce 250) ((T_OBJECT_CAST) reduce 250) ((T_PRINT) reduce 250) ((T_PROPERTY_C) reduce 250) ((T_READONLY) reduce 250) ((T_REQUIRE) reduce 250) ((T_REQUIRE_ONCE) reduce 250) ((T_RETURN) reduce 250) ((T_START_HEREDOC) reduce 250) ((T_STATIC) reduce 250) ((T_STRING) reduce 250) ((T_STRING_CAST) reduce 250) ((T_SWITCH) reduce 250) ((T_THROW) reduce 250) ((T_TRAIT) reduce 250) ((T_TRAIT_C) reduce 250) ((T_TRY) reduce 250) ((T_UNSET) reduce 250) ((T_UNSET_CAST) reduce 250) ((T_USE) reduce 250) ((T_VARIABLE) reduce 250) ((T_VOID_CAST) reduce 250) ((T_WHILE) reduce 250) ((T_YIELD) reduce 250) ((T_YIELD_FROM) reduce 250) (("[") reduce 250) (("`") reduce 250) (("{") reduce 250) (("}") reduce 250) (("~") reduce 250)) 32456 ((("!") reduce 251) (("\"") reduce 251) (("$") reduce 251) (("(") reduce 251) (("+") reduce 251) (("-") reduce 251) ((";") reduce 251) (("@") reduce 251) ((T_ABSTRACT) reduce 251) ((T_ARRAY) reduce 251) ((T_ARRAY_CAST) reduce 251) ((T_ATTRIBUTE) reduce 251) ((T_BOOL_CAST) reduce 251) ((T_BREAK) reduce 251) ((T_CLASS) reduce 251) ((T_CLASS_C) reduce 251) ((T_CLONE) reduce 251) ((T_CONST) reduce 251) ((T_CONSTANT_ENCAPSED_STRING) reduce 251) ((T_CONTINUE) reduce 251) ((T_DEC) reduce 251) ((T_DECLARE) reduce 251) ((T_DIR) reduce 251) ((T_DNUMBER) reduce 251) ((T_DO) reduce 251) ((T_DOUBLE_CAST) reduce 251) ((T_ECHO) reduce 251) ((T_ELSE) reduce 251) ((T_ELSEIF) reduce 251) ((T_EMPTY) reduce 251) ((T_ENUM) reduce 251) ((T_EVAL) reduce 251) ((T_EXIT) reduce 251) ((T_FILE) reduce 251) ((T_FINAL) reduce 251) ((T_FN) reduce 251) ((T_FOR) reduce 251) ((T_FOREACH) reduce 251) ((T_FUNCTION) reduce 251) ((T_FUNC_C) reduce 251) ((T_GLOBAL) reduce 251) ((T_GOTO) reduce 251) ((T_HALT_COMPILER) reduce 251) ((T_IF) reduce 251) ((T_INC) reduce 251) ((T_INCLUDE) reduce 251) ((T_INCLUDE_ONCE) reduce 251) ((T_INLINE_HTML) reduce 251) ((T_INTERFACE) reduce 251) ((T_INT_CAST) reduce 251) ((T_ISSET) reduce 251) ((T_LINE) reduce 251) ((T_LIST) reduce 251) ((T_LNUMBER) reduce 251) ((T_MATCH) reduce 251) ((T_METHOD_C) reduce 251) ((T_NAMESPACE) reduce 251) ((T_NAME_FULLY_QUALIFIED) reduce 251) ((T_NAME_QUALIFIED) reduce 251) ((T_NAME_RELATIVE) reduce 251) ((T_NEW) reduce 251) ((T_NS_C) reduce 251) ((T_OBJECT_CAST) reduce 251) ((T_PRINT) reduce 251) ((T_PROPERTY_C) reduce 251) ((T_READONLY) reduce 251) ((T_REQUIRE) reduce 251) ((T_REQUIRE_ONCE) reduce 251) ((T_RETURN) reduce 251) ((T_START_HEREDOC) reduce 251) ((T_STATIC) reduce 251) ((T_STRING) reduce 251) ((T_STRING_CAST) reduce 251) ((T_SWITCH) reduce 251) ((T_THROW) reduce 251) ((T_TRAIT) reduce 251) ((T_TRAIT_C) reduce 251) ((T_TRY) reduce 251) ((T_UNSET) reduce 251) ((T_UNSET_CAST) reduce 251) ((T_USE) reduce 251) ((T_VARIABLE) reduce 251) ((T_VOID_CAST) reduce 251) ((T_WHILE) reduce 251) ((T_YIELD) reduce 251) ((T_YIELD_FROM) reduce 251) (("[") reduce 251) (("`") reduce 251) (("{") reduce 251) (("}") reduce 251) (("~") reduce 251)) 32461 ((("!") reduce 242) (("\"") reduce 242) (("$") reduce 242) (("(") reduce 242) (("+") reduce 242) (("-") reduce 242) ((";") reduce 242) (("@") reduce 242) ((T_ABSTRACT) reduce 242) ((T_ARRAY) reduce 242) ((T_ARRAY_CAST) reduce 242) ((T_ATTRIBUTE) reduce 242) ((T_BOOL_CAST) reduce 242) ((T_BREAK) reduce 242) ((T_CLASS) reduce 242) ((T_CLASS_C) reduce 242) ((T_CLONE) reduce 242) ((T_CONST) reduce 242) ((T_CONSTANT_ENCAPSED_STRING) reduce 242) ((T_CONTINUE) reduce 242) ((T_DEC) reduce 242) ((T_DECLARE) reduce 242) ((T_DIR) reduce 242) ((T_DNUMBER) reduce 242) ((T_DO) reduce 242) ((T_DOUBLE_CAST) reduce 242) ((T_ECHO) reduce 242) ((T_ELSE) reduce 242) ((T_ELSEIF) reduce 242) ((T_EMPTY) reduce 242) ((T_ENUM) reduce 242) ((T_EVAL) reduce 242) ((T_EXIT) reduce 242) ((T_FILE) reduce 242) ((T_FINAL) reduce 242) ((T_FN) reduce 242) ((T_FOR) reduce 242) ((T_FOREACH) reduce 242) ((T_FUNCTION) reduce 242) ((T_FUNC_C) reduce 242) ((T_GLOBAL) reduce 242) ((T_GOTO) reduce 242) ((T_HALT_COMPILER) reduce 242) ((T_IF) reduce 242) ((T_INC) reduce 242) ((T_INCLUDE) reduce 242) ((T_INCLUDE_ONCE) reduce 242) ((T_INLINE_HTML) reduce 242) ((T_INTERFACE) reduce 242) ((T_INT_CAST) reduce 242) ((T_ISSET) reduce 242) ((T_LINE) reduce 242) ((T_LIST) reduce 242) ((T_LNUMBER) reduce 242) ((T_MATCH) reduce 242) ((T_METHOD_C) reduce 242) ((T_NAMESPACE) reduce 242) ((T_NAME_FULLY_QUALIFIED) reduce 242) ((T_NAME_QUALIFIED) reduce 242) ((T_NAME_RELATIVE) reduce 242) ((T_NEW) reduce 242) ((T_NS_C) reduce 242) ((T_OBJECT_CAST) reduce 242) ((T_PRINT) reduce 242) ((T_PROPERTY_C) reduce 242) ((T_READONLY) reduce 242) ((T_REQUIRE) reduce 242) ((T_REQUIRE_ONCE) reduce 242) ((T_RETURN) reduce 242) ((T_START_HEREDOC) reduce 242) ((T_STATIC) reduce 242) ((T_STRING) reduce 242) ((T_STRING_CAST) reduce 242) ((T_SWITCH) reduce 242) ((T_THROW) reduce 242) ((T_TRAIT) reduce 242) ((T_TRAIT_C) reduce 242) ((T_TRY) reduce 242) ((T_UNSET) reduce 242) ((T_UNSET_CAST) reduce 242) ((T_USE) reduce 242) ((T_VARIABLE) reduce 242) ((T_VOID_CAST) reduce 242) ((T_WHILE) reduce 242) ((T_YIELD) reduce 242) ((T_YIELD_FROM) reduce 242) (("[") reduce 242) (("`") reduce 242) (("{") reduce 242) (("}") reduce 242) (("~") reduce 242)) 32462 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONST) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_ELSE) reduce 150) ((T_ELSEIF) reduce 150) ((T_EMPTY) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAMESPACE) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_USE) reduce 150) ((T_VARIABLE) reduce 150) ((T_VOID_CAST) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("}") reduce 150) (("~") reduce 150)) 32465 ((("!") reduce 243) (("\"") reduce 243) (("$") reduce 243) (("(") reduce 243) (("+") reduce 243) (("-") reduce 243) ((";") reduce 243) (("@") reduce 243) ((T_ABSTRACT) reduce 243) ((T_ARRAY) reduce 243) ((T_ARRAY_CAST) reduce 243) ((T_ATTRIBUTE) reduce 243) ((T_BOOL_CAST) reduce 243) ((T_BREAK) reduce 243) ((T_CLASS) reduce 243) ((T_CLASS_C) reduce 243) ((T_CLONE) reduce 243) ((T_CONST) reduce 243) ((T_CONSTANT_ENCAPSED_STRING) reduce 243) ((T_CONTINUE) reduce 243) ((T_DEC) reduce 243) ((T_DECLARE) reduce 243) ((T_DIR) reduce 243) ((T_DNUMBER) reduce 243) ((T_DO) reduce 243) ((T_DOUBLE_CAST) reduce 243) ((T_ECHO) reduce 243) ((T_ELSE) reduce 243) ((T_ELSEIF) reduce 243) ((T_EMPTY) reduce 243) ((T_ENUM) reduce 243) ((T_EVAL) reduce 243) ((T_EXIT) reduce 243) ((T_FILE) reduce 243) ((T_FINAL) reduce 243) ((T_FN) reduce 243) ((T_FOR) reduce 243) ((T_FOREACH) reduce 243) ((T_FUNCTION) reduce 243) ((T_FUNC_C) reduce 243) ((T_GLOBAL) reduce 243) ((T_GOTO) reduce 243) ((T_HALT_COMPILER) reduce 243) ((T_IF) reduce 243) ((T_INC) reduce 243) ((T_INCLUDE) reduce 243) ((T_INCLUDE_ONCE) reduce 243) ((T_INLINE_HTML) reduce 243) ((T_INTERFACE) reduce 243) ((T_INT_CAST) reduce 243) ((T_ISSET) reduce 243) ((T_LINE) reduce 243) ((T_LIST) reduce 243) ((T_LNUMBER) reduce 243) ((T_MATCH) reduce 243) ((T_METHOD_C) reduce 243) ((T_NAMESPACE) reduce 243) ((T_NAME_FULLY_QUALIFIED) reduce 243) ((T_NAME_QUALIFIED) reduce 243) ((T_NAME_RELATIVE) reduce 243) ((T_NEW) reduce 243) ((T_NS_C) reduce 243) ((T_OBJECT_CAST) reduce 243) ((T_PRINT) reduce 243) ((T_PROPERTY_C) reduce 243) ((T_READONLY) reduce 243) ((T_REQUIRE) reduce 243) ((T_REQUIRE_ONCE) reduce 243) ((T_RETURN) reduce 243) ((T_START_HEREDOC) reduce 243) ((T_STATIC) reduce 243) ((T_STRING) reduce 243) ((T_STRING_CAST) reduce 243) ((T_SWITCH) reduce 243) ((T_THROW) reduce 243) ((T_TRAIT) reduce 243) ((T_TRAIT_C) reduce 243) ((T_TRY) reduce 243) ((T_UNSET) reduce 243) ((T_UNSET_CAST) reduce 243) ((T_USE) reduce 243) ((T_VARIABLE) reduce 243) ((T_VOID_CAST) reduce 243) ((T_WHILE) reduce 243) ((T_YIELD) reduce 243) ((T_YIELD_FROM) reduce 243) (("[") reduce 243) (("`") reduce 243) (("{") reduce 243) (("}") reduce 243) (("~") reduce 243)) 32467 ((("!") reduce 170) (("\"") reduce 170) (("$") reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONST) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_ELSE) reduce 170) ((T_ELSEIF) reduce 170) ((T_EMPTY) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAMESPACE) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_USE) reduce 170) ((T_VARIABLE) reduce 170) ((T_VOID_CAST) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("}") reduce 170) (("~") reduce 170)) 32472 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONST) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_ELSE) reduce 162) ((T_ELSEIF) reduce 162) ((T_EMPTY) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAMESPACE) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_USE) reduce 162) ((T_VARIABLE) reduce 162) ((T_VOID_CAST) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("}") reduce 162) (("~") reduce 162)) 32475 ((("!") reduce 171) (("\"") reduce 171) (("$") reduce 171) (("(") reduce 171) (("+") reduce 171) (("-") reduce 171) ((";") reduce 171) (("@") reduce 171) ((T_ABSTRACT) reduce 171) ((T_ARRAY) reduce 171) ((T_ARRAY_CAST) reduce 171) ((T_ATTRIBUTE) reduce 171) ((T_BOOL_CAST) reduce 171) ((T_BREAK) reduce 171) ((T_CATCH) reduce 171) ((T_CLASS) reduce 171) ((T_CLASS_C) reduce 171) ((T_CLONE) reduce 171) ((T_CONST) reduce 171) ((T_CONSTANT_ENCAPSED_STRING) reduce 171) ((T_CONTINUE) reduce 171) ((T_DEC) reduce 171) ((T_DECLARE) reduce 171) ((T_DIR) reduce 171) ((T_DNUMBER) reduce 171) ((T_DO) reduce 171) ((T_DOUBLE_CAST) reduce 171) ((T_ECHO) reduce 171) ((T_ELSE) reduce 171) ((T_ELSEIF) reduce 171) ((T_EMPTY) reduce 171) ((T_ENUM) reduce 171) ((T_EVAL) reduce 171) ((T_EXIT) reduce 171) ((T_FILE) reduce 171) ((T_FINAL) reduce 171) ((T_FINALLY) reduce 171) ((T_FN) reduce 171) ((T_FOR) reduce 171) ((T_FOREACH) reduce 171) ((T_FUNCTION) reduce 171) ((T_FUNC_C) reduce 171) ((T_GLOBAL) reduce 171) ((T_GOTO) reduce 171) ((T_HALT_COMPILER) reduce 171) ((T_IF) reduce 171) ((T_INC) reduce 171) ((T_INCLUDE) reduce 171) ((T_INCLUDE_ONCE) reduce 171) ((T_INLINE_HTML) reduce 171) ((T_INTERFACE) reduce 171) ((T_INT_CAST) reduce 171) ((T_ISSET) reduce 171) ((T_LINE) reduce 171) ((T_LIST) reduce 171) ((T_LNUMBER) reduce 171) ((T_MATCH) reduce 171) ((T_METHOD_C) reduce 171) ((T_NAMESPACE) reduce 171) ((T_NAME_FULLY_QUALIFIED) reduce 171) ((T_NAME_QUALIFIED) reduce 171) ((T_NAME_RELATIVE) reduce 171) ((T_NEW) reduce 171) ((T_NS_C) reduce 171) ((T_OBJECT_CAST) reduce 171) ((T_PRINT) reduce 171) ((T_PROPERTY_C) reduce 171) ((T_READONLY) reduce 171) ((T_REQUIRE) reduce 171) ((T_REQUIRE_ONCE) reduce 171) ((T_RETURN) reduce 171) ((T_START_HEREDOC) reduce 171) ((T_STATIC) reduce 171) ((T_STRING) reduce 171) ((T_STRING_CAST) reduce 171) ((T_SWITCH) reduce 171) ((T_THROW) reduce 171) ((T_TRAIT) reduce 171) ((T_TRAIT_C) reduce 171) ((T_TRY) reduce 171) ((T_UNSET) reduce 171) ((T_UNSET_CAST) reduce 171) ((T_USE) reduce 171) ((T_VARIABLE) reduce 171) ((T_VOID_CAST) reduce 171) ((T_WHILE) reduce 171) ((T_YIELD) reduce 171) ((T_YIELD_FROM) reduce 171) (("[") reduce 171) (("`") reduce 171) (("{") reduce 171) (("}") reduce 171) (("~") reduce 171)) 32476 ((("!") reduce 177) (("\"") reduce 177) (("$") reduce 177) (("(") reduce 177) (("+") reduce 177) (("-") reduce 177) ((";") reduce 177) (("@") reduce 177) ((T_ABSTRACT) reduce 177) ((T_ARRAY) reduce 177) ((T_ARRAY_CAST) reduce 177) ((T_ATTRIBUTE) reduce 177) ((T_BOOL_CAST) reduce 177) ((T_BREAK) reduce 177) ((T_CATCH) shift) ((T_CLASS) reduce 177) ((T_CLASS_C) reduce 177) ((T_CLONE) reduce 177) ((T_CONST) reduce 177) ((T_CONSTANT_ENCAPSED_STRING) reduce 177) ((T_CONTINUE) reduce 177) ((T_DEC) reduce 177) ((T_DECLARE) reduce 177) ((T_DIR) reduce 177) ((T_DNUMBER) reduce 177) ((T_DO) reduce 177) ((T_DOUBLE_CAST) reduce 177) ((T_ECHO) reduce 177) ((T_ELSE) reduce 177) ((T_ELSEIF) reduce 177) ((T_EMPTY) reduce 177) ((T_ENUM) reduce 177) ((T_EVAL) reduce 177) ((T_EXIT) reduce 177) ((T_FILE) reduce 177) ((T_FINAL) reduce 177) ((T_FINALLY) shift) ((T_FN) reduce 177) ((T_FOR) reduce 177) ((T_FOREACH) reduce 177) ((T_FUNCTION) reduce 177) ((T_FUNC_C) reduce 177) ((T_GLOBAL) reduce 177) ((T_GOTO) reduce 177) ((T_HALT_COMPILER) reduce 177) ((T_IF) reduce 177) ((T_INC) reduce 177) ((T_INCLUDE) reduce 177) ((T_INCLUDE_ONCE) reduce 177) ((T_INLINE_HTML) reduce 177) ((T_INTERFACE) reduce 177) ((T_INT_CAST) reduce 177) ((T_ISSET) reduce 177) ((T_LINE) reduce 177) ((T_LIST) reduce 177) ((T_LNUMBER) reduce 177) ((T_MATCH) reduce 177) ((T_METHOD_C) reduce 177) ((T_NAMESPACE) reduce 177) ((T_NAME_FULLY_QUALIFIED) reduce 177) ((T_NAME_QUALIFIED) reduce 177) ((T_NAME_RELATIVE) reduce 177) ((T_NEW) reduce 177) ((T_NS_C) reduce 177) ((T_OBJECT_CAST) reduce 177) ((T_PRINT) reduce 177) ((T_PROPERTY_C) reduce 177) ((T_READONLY) reduce 177) ((T_REQUIRE) reduce 177) ((T_REQUIRE_ONCE) reduce 177) ((T_RETURN) reduce 177) ((T_START_HEREDOC) reduce 177) ((T_STATIC) reduce 177) ((T_STRING) reduce 177) ((T_STRING_CAST) reduce 177) ((T_SWITCH) reduce 177) ((T_THROW) reduce 177) ((T_TRAIT) reduce 177) ((T_TRAIT_C) reduce 177) ((T_TRY) reduce 177) ((T_UNSET) reduce 177) ((T_UNSET_CAST) reduce 177) ((T_USE) reduce 177) ((T_VARIABLE) reduce 177) ((T_VOID_CAST) reduce 177) ((T_WHILE) reduce 177) ((T_YIELD) reduce 177) ((T_YIELD_FROM) reduce 177) (("[") reduce 177) (("`") reduce 177) (("{") reduce 177) (("}") reduce 177) (("~") reduce 177)) 32479 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONST) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_ELSE) reduce 167) ((T_ELSEIF) reduce 167) ((T_EMPTY) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAMESPACE) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_USE) reduce 167) ((T_VARIABLE) reduce 167) ((T_VOID_CAST) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("}") reduce 167) (("~") reduce 167)) 32482 ((("!") reduce 178) (("\"") reduce 178) (("$") reduce 178) (("(") reduce 178) (("+") reduce 178) (("-") reduce 178) ((";") reduce 178) (("@") reduce 178) ((T_ABSTRACT) reduce 178) ((T_ARRAY) reduce 178) ((T_ARRAY_CAST) reduce 178) ((T_ATTRIBUTE) reduce 178) ((T_BOOL_CAST) reduce 178) ((T_BREAK) reduce 178) ((T_CLASS) reduce 178) ((T_CLASS_C) reduce 178) ((T_CLONE) reduce 178) ((T_CONST) reduce 178) ((T_CONSTANT_ENCAPSED_STRING) reduce 178) ((T_CONTINUE) reduce 178) ((T_DEC) reduce 178) ((T_DECLARE) reduce 178) ((T_DIR) reduce 178) ((T_DNUMBER) reduce 178) ((T_DO) reduce 178) ((T_DOUBLE_CAST) reduce 178) ((T_ECHO) reduce 178) ((T_ELSE) reduce 178) ((T_ELSEIF) reduce 178) ((T_EMPTY) reduce 178) ((T_ENUM) reduce 178) ((T_EVAL) reduce 178) ((T_EXIT) reduce 178) ((T_FILE) reduce 178) ((T_FINAL) reduce 178) ((T_FN) reduce 178) ((T_FOR) reduce 178) ((T_FOREACH) reduce 178) ((T_FUNCTION) reduce 178) ((T_FUNC_C) reduce 178) ((T_GLOBAL) reduce 178) ((T_GOTO) reduce 178) ((T_HALT_COMPILER) reduce 178) ((T_IF) reduce 178) ((T_INC) reduce 178) ((T_INCLUDE) reduce 178) ((T_INCLUDE_ONCE) reduce 178) ((T_INLINE_HTML) reduce 178) ((T_INTERFACE) reduce 178) ((T_INT_CAST) reduce 178) ((T_ISSET) reduce 178) ((T_LINE) reduce 178) ((T_LIST) reduce 178) ((T_LNUMBER) reduce 178) ((T_MATCH) reduce 178) ((T_METHOD_C) reduce 178) ((T_NAMESPACE) reduce 178) ((T_NAME_FULLY_QUALIFIED) reduce 178) ((T_NAME_QUALIFIED) reduce 178) ((T_NAME_RELATIVE) reduce 178) ((T_NEW) reduce 178) ((T_NS_C) reduce 178) ((T_OBJECT_CAST) reduce 178) ((T_PRINT) reduce 178) ((T_PROPERTY_C) reduce 178) ((T_READONLY) reduce 178) ((T_REQUIRE) reduce 178) ((T_REQUIRE_ONCE) reduce 178) ((T_RETURN) reduce 178) ((T_START_HEREDOC) reduce 178) ((T_STATIC) reduce 178) ((T_STRING) reduce 178) ((T_STRING_CAST) reduce 178) ((T_SWITCH) reduce 178) ((T_THROW) reduce 178) ((T_TRAIT) reduce 178) ((T_TRAIT_C) reduce 178) ((T_TRY) reduce 178) ((T_UNSET) reduce 178) ((T_UNSET_CAST) reduce 178) ((T_USE) reduce 178) ((T_VARIABLE) reduce 178) ((T_VOID_CAST) reduce 178) ((T_WHILE) reduce 178) ((T_YIELD) reduce 178) ((T_YIELD_FROM) reduce 178) (("[") reduce 178) (("`") reduce 178) (("{") reduce 178) (("}") reduce 178) (("~") reduce 178)) 32489 ((("!") reduce 172) (("\"") reduce 172) (("$") reduce 172) (("(") reduce 172) (("+") reduce 172) (("-") reduce 172) ((";") reduce 172) (("@") reduce 172) ((T_ABSTRACT) reduce 172) ((T_ARRAY) reduce 172) ((T_ARRAY_CAST) reduce 172) ((T_ATTRIBUTE) reduce 172) ((T_BOOL_CAST) reduce 172) ((T_BREAK) reduce 172) ((T_CATCH) reduce 172) ((T_CLASS) reduce 172) ((T_CLASS_C) reduce 172) ((T_CLONE) reduce 172) ((T_CONST) reduce 172) ((T_CONSTANT_ENCAPSED_STRING) reduce 172) ((T_CONTINUE) reduce 172) ((T_DEC) reduce 172) ((T_DECLARE) reduce 172) ((T_DIR) reduce 172) ((T_DNUMBER) reduce 172) ((T_DO) reduce 172) ((T_DOUBLE_CAST) reduce 172) ((T_ECHO) reduce 172) ((T_ELSE) reduce 172) ((T_ELSEIF) reduce 172) ((T_EMPTY) reduce 172) ((T_ENUM) reduce 172) ((T_EVAL) reduce 172) ((T_EXIT) reduce 172) ((T_FILE) reduce 172) ((T_FINAL) reduce 172) ((T_FINALLY) reduce 172) ((T_FN) reduce 172) ((T_FOR) reduce 172) ((T_FOREACH) reduce 172) ((T_FUNCTION) reduce 172) ((T_FUNC_C) reduce 172) ((T_GLOBAL) reduce 172) ((T_GOTO) reduce 172) ((T_HALT_COMPILER) reduce 172) ((T_IF) reduce 172) ((T_INC) reduce 172) ((T_INCLUDE) reduce 172) ((T_INCLUDE_ONCE) reduce 172) ((T_INLINE_HTML) reduce 172) ((T_INTERFACE) reduce 172) ((T_INT_CAST) reduce 172) ((T_ISSET) reduce 172) ((T_LINE) reduce 172) ((T_LIST) reduce 172) ((T_LNUMBER) reduce 172) ((T_MATCH) reduce 172) ((T_METHOD_C) reduce 172) ((T_NAMESPACE) reduce 172) ((T_NAME_FULLY_QUALIFIED) reduce 172) ((T_NAME_QUALIFIED) reduce 172) ((T_NAME_RELATIVE) reduce 172) ((T_NEW) reduce 172) ((T_NS_C) reduce 172) ((T_OBJECT_CAST) reduce 172) ((T_PRINT) reduce 172) ((T_PROPERTY_C) reduce 172) ((T_READONLY) reduce 172) ((T_REQUIRE) reduce 172) ((T_REQUIRE_ONCE) reduce 172) ((T_RETURN) reduce 172) ((T_START_HEREDOC) reduce 172) ((T_STATIC) reduce 172) ((T_STRING) reduce 172) ((T_STRING_CAST) reduce 172) ((T_SWITCH) reduce 172) ((T_THROW) reduce 172) ((T_TRAIT) reduce 172) ((T_TRAIT_C) reduce 172) ((T_TRY) reduce 172) ((T_UNSET) reduce 172) ((T_UNSET_CAST) reduce 172) ((T_USE) reduce 172) ((T_VARIABLE) reduce 172) ((T_VOID_CAST) reduce 172) ((T_WHILE) reduce 172) ((T_YIELD) reduce 172) ((T_YIELD_FROM) reduce 172) (("[") reduce 172) (("`") reduce 172) (("{") reduce 172) (("}") reduce 172) (("~") reduce 172)) 32494 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONST) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_ELSE) reduce 153) ((T_ELSEIF) reduce 153) ((T_EMPTY) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAMESPACE) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_USE) reduce 153) ((T_VARIABLE) reduce 153) ((T_VOID_CAST) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("}") reduce 153) (("~") reduce 153)) 32498 ((("!") reduce 224) (("\"") reduce 224) (("$") reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONST) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_ELSE) reduce 224) ((T_ELSEIF) reduce 224) ((T_EMPTY) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAMESPACE) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_USE) reduce 224) ((T_VARIABLE) reduce 224) ((T_VOID_CAST) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("}") reduce 224) (("~") reduce 224)) 32500 ((("!") reduce 225) (("\"") reduce 225) (("$") reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONST) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_ELSE) reduce 225) ((T_ELSEIF) reduce 225) ((T_EMPTY) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAMESPACE) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_USE) reduce 225) ((T_VARIABLE) reduce 225) ((T_VOID_CAST) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("}") reduce 225) (("~") reduce 225)) 32504 ((("!") reduce 226) (("\"") reduce 226) (("$") reduce 226) (("(") reduce 226) (("+") reduce 226) (("-") reduce 226) ((";") reduce 226) (("@") reduce 226) ((T_ABSTRACT) reduce 226) ((T_ARRAY) reduce 226) ((T_ARRAY_CAST) reduce 226) ((T_ATTRIBUTE) reduce 226) ((T_BOOL_CAST) reduce 226) ((T_BREAK) reduce 226) ((T_CLASS) reduce 226) ((T_CLASS_C) reduce 226) ((T_CLONE) reduce 226) ((T_CONST) reduce 226) ((T_CONSTANT_ENCAPSED_STRING) reduce 226) ((T_CONTINUE) reduce 226) ((T_DEC) reduce 226) ((T_DECLARE) reduce 226) ((T_DIR) reduce 226) ((T_DNUMBER) reduce 226) ((T_DO) reduce 226) ((T_DOUBLE_CAST) reduce 226) ((T_ECHO) reduce 226) ((T_ELSE) reduce 226) ((T_ELSEIF) reduce 226) ((T_EMPTY) reduce 226) ((T_ENUM) reduce 226) ((T_EVAL) reduce 226) ((T_EXIT) reduce 226) ((T_FILE) reduce 226) ((T_FINAL) reduce 226) ((T_FN) reduce 226) ((T_FOR) reduce 226) ((T_FOREACH) reduce 226) ((T_FUNCTION) reduce 226) ((T_FUNC_C) reduce 226) ((T_GLOBAL) reduce 226) ((T_GOTO) reduce 226) ((T_HALT_COMPILER) reduce 226) ((T_IF) reduce 226) ((T_INC) reduce 226) ((T_INCLUDE) reduce 226) ((T_INCLUDE_ONCE) reduce 226) ((T_INLINE_HTML) reduce 226) ((T_INTERFACE) reduce 226) ((T_INT_CAST) reduce 226) ((T_ISSET) reduce 226) ((T_LINE) reduce 226) ((T_LIST) reduce 226) ((T_LNUMBER) reduce 226) ((T_MATCH) reduce 226) ((T_METHOD_C) reduce 226) ((T_NAMESPACE) reduce 226) ((T_NAME_FULLY_QUALIFIED) reduce 226) ((T_NAME_QUALIFIED) reduce 226) ((T_NAME_RELATIVE) reduce 226) ((T_NEW) reduce 226) ((T_NS_C) reduce 226) ((T_OBJECT_CAST) reduce 226) ((T_PRINT) reduce 226) ((T_PROPERTY_C) reduce 226) ((T_READONLY) reduce 226) ((T_REQUIRE) reduce 226) ((T_REQUIRE_ONCE) reduce 226) ((T_RETURN) reduce 226) ((T_START_HEREDOC) reduce 226) ((T_STATIC) reduce 226) ((T_STRING) reduce 226) ((T_STRING_CAST) reduce 226) ((T_SWITCH) reduce 226) ((T_THROW) reduce 226) ((T_TRAIT) reduce 226) ((T_TRAIT_C) reduce 226) ((T_TRY) reduce 226) ((T_UNSET) reduce 226) ((T_UNSET_CAST) reduce 226) ((T_USE) reduce 226) ((T_VARIABLE) reduce 226) ((T_VOID_CAST) reduce 226) ((T_WHILE) reduce 226) ((T_YIELD) reduce 226) ((T_YIELD_FROM) reduce 226) (("[") reduce 226) (("`") reduce 226) (("{") reduce 226) (("}") reduce 226) (("~") reduce 226)) 32507 ((("!") reduce 227) (("\"") reduce 227) (("$") reduce 227) (("(") reduce 227) (("+") reduce 227) (("-") reduce 227) ((";") reduce 227) (("@") reduce 227) ((T_ABSTRACT) reduce 227) ((T_ARRAY) reduce 227) ((T_ARRAY_CAST) reduce 227) ((T_ATTRIBUTE) reduce 227) ((T_BOOL_CAST) reduce 227) ((T_BREAK) reduce 227) ((T_CLASS) reduce 227) ((T_CLASS_C) reduce 227) ((T_CLONE) reduce 227) ((T_CONST) reduce 227) ((T_CONSTANT_ENCAPSED_STRING) reduce 227) ((T_CONTINUE) reduce 227) ((T_DEC) reduce 227) ((T_DECLARE) reduce 227) ((T_DIR) reduce 227) ((T_DNUMBER) reduce 227) ((T_DO) reduce 227) ((T_DOUBLE_CAST) reduce 227) ((T_ECHO) reduce 227) ((T_ELSE) reduce 227) ((T_ELSEIF) reduce 227) ((T_EMPTY) reduce 227) ((T_ENUM) reduce 227) ((T_EVAL) reduce 227) ((T_EXIT) reduce 227) ((T_FILE) reduce 227) ((T_FINAL) reduce 227) ((T_FN) reduce 227) ((T_FOR) reduce 227) ((T_FOREACH) reduce 227) ((T_FUNCTION) reduce 227) ((T_FUNC_C) reduce 227) ((T_GLOBAL) reduce 227) ((T_GOTO) reduce 227) ((T_HALT_COMPILER) reduce 227) ((T_IF) reduce 227) ((T_INC) reduce 227) ((T_INCLUDE) reduce 227) ((T_INCLUDE_ONCE) reduce 227) ((T_INLINE_HTML) reduce 227) ((T_INTERFACE) reduce 227) ((T_INT_CAST) reduce 227) ((T_ISSET) reduce 227) ((T_LINE) reduce 227) ((T_LIST) reduce 227) ((T_LNUMBER) reduce 227) ((T_MATCH) reduce 227) ((T_METHOD_C) reduce 227) ((T_NAMESPACE) reduce 227) ((T_NAME_FULLY_QUALIFIED) reduce 227) ((T_NAME_QUALIFIED) reduce 227) ((T_NAME_RELATIVE) reduce 227) ((T_NEW) reduce 227) ((T_NS_C) reduce 227) ((T_OBJECT_CAST) reduce 227) ((T_PRINT) reduce 227) ((T_PROPERTY_C) reduce 227) ((T_READONLY) reduce 227) ((T_REQUIRE) reduce 227) ((T_REQUIRE_ONCE) reduce 227) ((T_RETURN) reduce 227) ((T_START_HEREDOC) reduce 227) ((T_STATIC) reduce 227) ((T_STRING) reduce 227) ((T_STRING_CAST) reduce 227) ((T_SWITCH) reduce 227) ((T_THROW) reduce 227) ((T_TRAIT) reduce 227) ((T_TRAIT_C) reduce 227) ((T_TRY) reduce 227) ((T_UNSET) reduce 227) ((T_UNSET_CAST) reduce 227) ((T_USE) reduce 227) ((T_VARIABLE) reduce 227) ((T_VOID_CAST) reduce 227) ((T_WHILE) reduce 227) ((T_YIELD) reduce 227) ((T_YIELD_FROM) reduce 227) (("[") reduce 227) (("`") reduce 227) (("{") reduce 227) (("}") reduce 227) (("~") reduce 227)) 32508 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONST) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_ELSE) reduce 169) ((T_ELSEIF) reduce 169) ((T_EMPTY) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAMESPACE) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_USE) reduce 169) ((T_VARIABLE) reduce 169) ((T_VOID_CAST) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("}") reduce 169) (("~") reduce 169)) 32510 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONST) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_ELSE) reduce 158) ((T_ELSEIF) reduce 158) ((T_EMPTY) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAMESPACE) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_USE) reduce 158) ((T_VARIABLE) reduce 158) ((T_VOID_CAST) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("}") reduce 158) (("~") reduce 158)) 32512 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONST) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_ELSE) reduce 156) ((T_ELSEIF) reduce 156) ((T_EMPTY) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAMESPACE) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_USE) reduce 156) ((T_VARIABLE) reduce 156) ((T_VOID_CAST) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("}") reduce 156) (("~") reduce 156)) 32514 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONST) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_ELSE) reduce 168) ((T_ELSEIF) reduce 168) ((T_EMPTY) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAMESPACE) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_USE) reduce 168) ((T_VARIABLE) reduce 168) ((T_VOID_CAST) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("}") reduce 168) (("~") reduce 168)) 32516 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONST) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_ELSE) reduce 157) ((T_ELSEIF) reduce 157) ((T_EMPTY) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAMESPACE) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_USE) reduce 157) ((T_VARIABLE) reduce 157) ((T_VOID_CAST) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("}") reduce 157) (("~") reduce 157)) 32526 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONST) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_ELSE) reduce 164) ((T_ELSEIF) reduce 164) ((T_EMPTY) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAMESPACE) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_USE) reduce 164) ((T_VARIABLE) reduce 164) ((T_VOID_CAST) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("}") reduce 164) (("~") reduce 164)) 32527 ((("!") reduce 220) (("\"") reduce 220) (("$") reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONST) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_ELSE) reduce 220) ((T_ELSEIF) reduce 220) ((T_EMPTY) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAMESPACE) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_USE) reduce 220) ((T_VARIABLE) reduce 220) ((T_VOID_CAST) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("}") reduce 220) (("~") reduce 220)) 32530 ((("!") reduce 221) (("\"") reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONST) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_ELSE) reduce 221) ((T_ELSEIF) reduce 221) ((T_EMPTY) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAMESPACE) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_USE) reduce 221) ((T_VARIABLE) reduce 221) ((T_VOID_CAST) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("}") reduce 221) (("~") reduce 221)) 32531 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONST) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_ELSE) reduce 163) ((T_ELSEIF) reduce 163) ((T_EMPTY) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAMESPACE) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_USE) reduce 163) ((T_VARIABLE) reduce 163) ((T_VOID_CAST) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("}") reduce 163) (("~") reduce 163)) 32540 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONST) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_ELSE) reduce 152) ((T_ELSEIF) reduce 152) ((T_EMPTY) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAMESPACE) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_USE) reduce 152) ((T_VARIABLE) reduce 152) ((T_VOID_CAST) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("}") reduce 152) (("~") reduce 152)) 32541 ((("!") reduce 218) (("\"") reduce 218) (("$") reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONST) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_ELSE) reduce 218) ((T_ELSEIF) reduce 218) ((T_EMPTY) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAMESPACE) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_USE) reduce 218) ((T_VARIABLE) reduce 218) ((T_VOID_CAST) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("}") reduce 218) (("~") reduce 218)) 32544 ((("!") reduce 219) (("\"") reduce 219) (("$") reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONST) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_ELSE) reduce 219) ((T_ELSEIF) reduce 219) ((T_EMPTY) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAMESPACE) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_USE) reduce 219) ((T_VARIABLE) reduce 219) ((T_VOID_CAST) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("}") reduce 219) (("~") reduce 219)) 32546 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONST) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_ELSE) reduce 159) ((T_ELSEIF) reduce 159) ((T_EMPTY) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAMESPACE) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_USE) reduce 159) ((T_VARIABLE) reduce 159) ((T_VOID_CAST) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("}") reduce 159) (("~") reduce 159)) 32552 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONST) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_ELSE) reduce 151) ((T_ELSEIF) reduce 151) ((T_EMPTY) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAMESPACE) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_USE) reduce 151) ((T_VARIABLE) reduce 151) ((T_VOID_CAST) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("}") reduce 151) (("~") reduce 151)) 32557 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONST) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_ELSE) reduce 165) ((T_ELSEIF) reduce 165) ((T_EMPTY) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAMESPACE) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_USE) reduce 165) ((T_VARIABLE) reduce 165) ((T_VOID_CAST) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("}") reduce 165) (("~") reduce 165)) 32558 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONST) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_ELSE) reduce 222) ((T_ELSEIF) reduce 222) ((T_EMPTY) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAMESPACE) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_USE) reduce 222) ((T_VARIABLE) reduce 222) ((T_VOID_CAST) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("}") reduce 222) (("~") reduce 222)) 32561 ((("!") reduce 223) (("\"") reduce 223) (("$") reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONST) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_ELSE) reduce 223) ((T_ELSEIF) reduce 223) ((T_EMPTY) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAMESPACE) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_USE) reduce 223) ((T_VARIABLE) reduce 223) ((T_VOID_CAST) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("}") reduce 223) (("~") reduce 223)) 32563 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONST) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_ELSE) reduce 155) ((T_ELSEIF) reduce 155) ((T_EMPTY) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAMESPACE) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_USE) reduce 155) ((T_VARIABLE) reduce 155) ((T_VOID_CAST) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("}") reduce 155) (("~") reduce 155)) 32565 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONST) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_ELSE) reduce 154) ((T_ELSEIF) reduce 154) ((T_EMPTY) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAMESPACE) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_USE) reduce 154) ((T_VARIABLE) reduce 154) ((T_VOID_CAST) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("}") reduce 154) (("~") reduce 154)) 32566 ((("!") reduce 247) (("\"") reduce 247) (("$") reduce 247) (("(") reduce 247) (("+") reduce 247) (("-") reduce 247) ((";") reduce 247) (("@") reduce 247) ((T_ABSTRACT) reduce 247) ((T_ARRAY) reduce 247) ((T_ARRAY_CAST) reduce 247) ((T_ATTRIBUTE) reduce 247) ((T_BOOL_CAST) reduce 247) ((T_BREAK) reduce 247) ((T_CLASS) reduce 247) ((T_CLASS_C) reduce 247) ((T_CLONE) reduce 247) ((T_CONST) reduce 247) ((T_CONSTANT_ENCAPSED_STRING) reduce 247) ((T_CONTINUE) reduce 247) ((T_DEC) reduce 247) ((T_DECLARE) reduce 247) ((T_DIR) reduce 247) ((T_DNUMBER) reduce 247) ((T_DO) reduce 247) ((T_DOUBLE_CAST) reduce 247) ((T_ECHO) reduce 247) ((T_EMPTY) reduce 247) ((T_ENUM) reduce 247) ((T_EVAL) reduce 247) ((T_EXIT) reduce 247) ((T_FILE) reduce 247) ((T_FINAL) reduce 247) ((T_FN) reduce 247) ((T_FOR) reduce 247) ((T_FOREACH) reduce 247) ((T_FUNCTION) reduce 247) ((T_FUNC_C) reduce 247) ((T_GLOBAL) reduce 247) ((T_GOTO) reduce 247) ((T_HALT_COMPILER) reduce 247) ((T_IF) reduce 247) ((T_INC) reduce 247) ((T_INCLUDE) reduce 247) ((T_INCLUDE_ONCE) reduce 247) ((T_INLINE_HTML) reduce 247) ((T_INTERFACE) reduce 247) ((T_INT_CAST) reduce 247) ((T_ISSET) reduce 247) ((T_LINE) reduce 247) ((T_LIST) reduce 247) ((T_LNUMBER) reduce 247) ((T_MATCH) reduce 247) ((T_METHOD_C) reduce 247) ((T_NAMESPACE) reduce 247) ((T_NAME_FULLY_QUALIFIED) reduce 247) ((T_NAME_QUALIFIED) reduce 247) ((T_NAME_RELATIVE) reduce 247) ((T_NEW) reduce 247) ((T_NS_C) reduce 247) ((T_OBJECT_CAST) reduce 247) ((T_PRINT) reduce 247) ((T_PROPERTY_C) reduce 247) ((T_READONLY) reduce 247) ((T_REQUIRE) reduce 247) ((T_REQUIRE_ONCE) reduce 247) ((T_RETURN) reduce 247) ((T_START_HEREDOC) reduce 247) ((T_STATIC) reduce 247) ((T_STRING) reduce 247) ((T_STRING_CAST) reduce 247) ((T_SWITCH) reduce 247) ((T_THROW) reduce 247) ((T_TRAIT) reduce 247) ((T_TRAIT_C) reduce 247) ((T_TRY) reduce 247) ((T_UNSET) reduce 247) ((T_UNSET_CAST) reduce 247) ((T_USE) reduce 247) ((T_VARIABLE) reduce 247) ((T_VOID_CAST) reduce 247) ((T_WHILE) reduce 247) ((T_YIELD) reduce 247) ((T_YIELD_FROM) reduce 247) (("[") reduce 247) (("`") reduce 247) (("{") reduce 247) (("}") reduce 247) (("~") reduce 247)) 32577 ((("!") reduce 494) (("\"") reduce 494) (("$") reduce 494) (("(") reduce 494) (("+") reduce 494) (("-") reduce 494) ((";") reduce 494) (("@") reduce 494) ((T_ABSTRACT) reduce 494) ((T_ARRAY) reduce 494) ((T_ARRAY_CAST) reduce 494) ((T_ATTRIBUTE) reduce 494) ((T_BOOL_CAST) reduce 494) ((T_BREAK) reduce 494) ((T_CLASS) reduce 494) ((T_CLASS_C) reduce 494) ((T_CLONE) reduce 494) ((T_CONST) reduce 494) ((T_CONSTANT_ENCAPSED_STRING) reduce 494) ((T_CONTINUE) reduce 494) ((T_DEC) reduce 494) ((T_DECLARE) reduce 494) ((T_DIR) reduce 494) ((T_DNUMBER) reduce 494) ((T_DO) reduce 494) ((T_DOUBLE_CAST) reduce 494) ((T_ECHO) reduce 494) ((T_EMPTY) reduce 494) ((T_ENUM) reduce 494) ((T_EVAL) reduce 494) ((T_EXIT) reduce 494) ((T_FILE) reduce 494) ((T_FINAL) reduce 494) ((T_FN) reduce 494) ((T_FOR) reduce 494) ((T_FOREACH) reduce 494) ((T_FUNCTION) reduce 494) ((T_FUNC_C) reduce 494) ((T_GLOBAL) reduce 494) ((T_GOTO) reduce 494) ((T_HALT_COMPILER) reduce 494) ((T_IF) reduce 494) ((T_INC) reduce 494) ((T_INCLUDE) reduce 494) ((T_INCLUDE_ONCE) reduce 494) ((T_INLINE_HTML) reduce 494) ((T_INTERFACE) reduce 494) ((T_INT_CAST) reduce 494) ((T_ISSET) reduce 494) ((T_LINE) reduce 494) ((T_LIST) reduce 494) ((T_LNUMBER) reduce 494) ((T_MATCH) reduce 494) ((T_METHOD_C) reduce 494) ((T_NAMESPACE) reduce 494) ((T_NAME_FULLY_QUALIFIED) reduce 494) ((T_NAME_QUALIFIED) reduce 494) ((T_NAME_RELATIVE) reduce 494) ((T_NEW) reduce 494) ((T_NS_C) reduce 494) ((T_OBJECT_CAST) reduce 494) ((T_PRINT) reduce 494) ((T_PROPERTY_C) reduce 494) ((T_READONLY) reduce 494) ((T_REQUIRE) reduce 494) ((T_REQUIRE_ONCE) reduce 494) ((T_RETURN) reduce 494) ((T_START_HEREDOC) reduce 494) ((T_STATIC) reduce 494) ((T_STRING) reduce 494) ((T_STRING_CAST) reduce 494) ((T_SWITCH) reduce 494) ((T_THROW) reduce 494) ((T_TRAIT) reduce 494) ((T_TRAIT_C) reduce 494) ((T_TRY) reduce 494) ((T_UNSET) reduce 494) ((T_UNSET_CAST) reduce 494) ((T_USE) reduce 494) ((T_VARIABLE) reduce 494) ((T_VOID_CAST) reduce 494) ((T_WHILE) reduce 494) ((T_YIELD) reduce 494) ((T_YIELD_FROM) reduce 494) (("[") reduce 494) (("`") reduce 494) (("{") reduce 494) (("}") reduce 494) (("~") reduce 494)) 32578 ((("!") reduce 184) (("\"") reduce 184) (("$") reduce 184) (("(") reduce 184) (("+") reduce 184) (("-") reduce 184) ((";") reduce 184) (("@") reduce 184) ((T_ABSTRACT) reduce 184) ((T_ARRAY) reduce 184) ((T_ARRAY_CAST) reduce 184) ((T_ATTRIBUTE) reduce 184) ((T_BOOL_CAST) reduce 184) ((T_BREAK) reduce 184) ((T_CLASS) reduce 184) ((T_CLASS_C) reduce 184) ((T_CLONE) reduce 184) ((T_CONST) reduce 184) ((T_CONSTANT_ENCAPSED_STRING) reduce 184) ((T_CONTINUE) reduce 184) ((T_DEC) reduce 184) ((T_DECLARE) reduce 184) ((T_DIR) reduce 184) ((T_DNUMBER) reduce 184) ((T_DO) reduce 184) ((T_DOUBLE_CAST) reduce 184) ((T_ECHO) reduce 184) ((T_EMPTY) reduce 184) ((T_ENUM) reduce 184) ((T_EVAL) reduce 184) ((T_EXIT) reduce 184) ((T_FILE) reduce 184) ((T_FINAL) reduce 184) ((T_FN) reduce 184) ((T_FOR) reduce 184) ((T_FOREACH) reduce 184) ((T_FUNCTION) reduce 184) ((T_FUNC_C) reduce 184) ((T_GLOBAL) reduce 184) ((T_GOTO) reduce 184) ((T_HALT_COMPILER) reduce 184) ((T_IF) reduce 184) ((T_INC) reduce 184) ((T_INCLUDE) reduce 184) ((T_INCLUDE_ONCE) reduce 184) ((T_INLINE_HTML) reduce 184) ((T_INTERFACE) reduce 184) ((T_INT_CAST) reduce 184) ((T_ISSET) reduce 184) ((T_LINE) reduce 184) ((T_LIST) reduce 184) ((T_LNUMBER) reduce 184) ((T_MATCH) reduce 184) ((T_METHOD_C) reduce 184) ((T_NAMESPACE) reduce 184) ((T_NAME_FULLY_QUALIFIED) reduce 184) ((T_NAME_QUALIFIED) reduce 184) ((T_NAME_RELATIVE) reduce 184) ((T_NEW) reduce 184) ((T_NS_C) reduce 184) ((T_OBJECT_CAST) reduce 184) ((T_PRINT) reduce 184) ((T_PROPERTY_C) reduce 184) ((T_READONLY) reduce 184) ((T_REQUIRE) reduce 184) ((T_REQUIRE_ONCE) reduce 184) ((T_RETURN) reduce 184) ((T_START_HEREDOC) reduce 184) ((T_STATIC) reduce 184) ((T_STRING) reduce 184) ((T_STRING_CAST) reduce 184) ((T_SWITCH) reduce 184) ((T_THROW) reduce 184) ((T_TRAIT) reduce 184) ((T_TRAIT_C) reduce 184) ((T_TRY) reduce 184) ((T_UNSET) reduce 184) ((T_UNSET_CAST) reduce 184) ((T_USE) reduce 184) ((T_VARIABLE) reduce 184) ((T_VOID_CAST) reduce 184) ((T_WHILE) reduce 184) ((T_YIELD) reduce 184) ((T_YIELD_FROM) reduce 184) (("[") reduce 184) (("`") reduce 184) (("{") reduce 184) (("}") reduce 184) (("~") reduce 184)) 32579 ((("!") reduce 161) (("\"") reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONST) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_EMPTY) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAMESPACE) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_USE) reduce 161) ((T_VARIABLE) reduce 161) ((T_VOID_CAST) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("}") reduce 161) (("~") reduce 161)) 32587 ((("!") reduce 189) (("\"") reduce 189) (("$") reduce 189) (("(") reduce 189) (("+") reduce 189) (("-") reduce 189) ((";") reduce 189) (("@") reduce 189) ((T_ABSTRACT) reduce 189) ((T_ARRAY) reduce 189) ((T_ARRAY_CAST) reduce 189) ((T_ATTRIBUTE) reduce 189) ((T_BOOL_CAST) reduce 189) ((T_BREAK) reduce 189) ((T_CLASS) reduce 189) ((T_CLASS_C) reduce 189) ((T_CLONE) reduce 189) ((T_CONST) reduce 189) ((T_CONSTANT_ENCAPSED_STRING) reduce 189) ((T_CONTINUE) reduce 189) ((T_DEC) reduce 189) ((T_DECLARE) reduce 189) ((T_DIR) reduce 189) ((T_DNUMBER) reduce 189) ((T_DO) reduce 189) ((T_DOUBLE_CAST) reduce 189) ((T_ECHO) reduce 189) ((T_EMPTY) reduce 189) ((T_ENUM) reduce 189) ((T_EVAL) reduce 189) ((T_EXIT) reduce 189) ((T_FILE) reduce 189) ((T_FINAL) reduce 189) ((T_FN) reduce 189) ((T_FOR) reduce 189) ((T_FOREACH) reduce 189) ((T_FUNCTION) reduce 189) ((T_FUNC_C) reduce 189) ((T_GLOBAL) reduce 189) ((T_GOTO) reduce 189) ((T_HALT_COMPILER) reduce 189) ((T_IF) reduce 189) ((T_INC) reduce 189) ((T_INCLUDE) reduce 189) ((T_INCLUDE_ONCE) reduce 189) ((T_INLINE_HTML) reduce 189) ((T_INTERFACE) reduce 189) ((T_INT_CAST) reduce 189) ((T_ISSET) reduce 189) ((T_LINE) reduce 189) ((T_LIST) reduce 189) ((T_LNUMBER) reduce 189) ((T_MATCH) reduce 189) ((T_METHOD_C) reduce 189) ((T_NAMESPACE) reduce 189) ((T_NAME_FULLY_QUALIFIED) reduce 189) ((T_NAME_QUALIFIED) reduce 189) ((T_NAME_RELATIVE) reduce 189) ((T_NEW) reduce 189) ((T_NS_C) reduce 189) ((T_OBJECT_CAST) reduce 189) ((T_PRINT) reduce 189) ((T_PROPERTY_C) reduce 189) ((T_READONLY) reduce 189) ((T_REQUIRE) reduce 189) ((T_REQUIRE_ONCE) reduce 189) ((T_RETURN) reduce 189) ((T_START_HEREDOC) reduce 189) ((T_STATIC) reduce 189) ((T_STRING) reduce 189) ((T_STRING_CAST) reduce 189) ((T_SWITCH) reduce 189) ((T_THROW) reduce 189) ((T_TRAIT) reduce 189) ((T_TRAIT_C) reduce 189) ((T_TRY) reduce 189) ((T_UNSET) reduce 189) ((T_UNSET_CAST) reduce 189) ((T_USE) reduce 189) ((T_VARIABLE) reduce 189) ((T_VOID_CAST) reduce 189) ((T_WHILE) reduce 189) ((T_YIELD) reduce 189) ((T_YIELD_FROM) reduce 189) (("[") reduce 189) (("`") reduce 189) (("{") reduce 189) (("}") reduce 189) (("~") reduce 189)) 32588 ((("!") reduce 112) (("\"") reduce 112) (("$") reduce 112) (("(") reduce 112) (("+") reduce 112) (("-") reduce 112) ((";") reduce 112) (("@") reduce 112) ((T_ABSTRACT) reduce 112) ((T_ARRAY) reduce 112) ((T_ARRAY_CAST) reduce 112) ((T_ATTRIBUTE) reduce 112) ((T_BOOL_CAST) reduce 112) ((T_BREAK) reduce 112) ((T_CLASS) reduce 112) ((T_CLASS_C) reduce 112) ((T_CLONE) reduce 112) ((T_CONST) reduce 112) ((T_CONSTANT_ENCAPSED_STRING) reduce 112) ((T_CONTINUE) reduce 112) ((T_DEC) reduce 112) ((T_DECLARE) reduce 112) ((T_DIR) reduce 112) ((T_DNUMBER) reduce 112) ((T_DO) reduce 112) ((T_DOUBLE_CAST) reduce 112) ((T_ECHO) reduce 112) ((T_EMPTY) reduce 112) ((T_ENUM) reduce 112) ((T_EVAL) reduce 112) ((T_EXIT) reduce 112) ((T_FILE) reduce 112) ((T_FINAL) reduce 112) ((T_FN) reduce 112) ((T_FOR) reduce 112) ((T_FOREACH) reduce 112) ((T_FUNCTION) reduce 112) ((T_FUNC_C) reduce 112) ((T_GLOBAL) reduce 112) ((T_GOTO) reduce 112) ((T_HALT_COMPILER) reduce 112) ((T_IF) reduce 112) ((T_INC) reduce 112) ((T_INCLUDE) reduce 112) ((T_INCLUDE_ONCE) reduce 112) ((T_INLINE_HTML) reduce 112) ((T_INTERFACE) reduce 112) ((T_INT_CAST) reduce 112) ((T_ISSET) reduce 112) ((T_LINE) reduce 112) ((T_LIST) reduce 112) ((T_LNUMBER) reduce 112) ((T_MATCH) reduce 112) ((T_METHOD_C) reduce 112) ((T_NAMESPACE) reduce 112) ((T_NAME_FULLY_QUALIFIED) reduce 112) ((T_NAME_QUALIFIED) reduce 112) ((T_NAME_RELATIVE) reduce 112) ((T_NEW) reduce 112) ((T_NS_C) reduce 112) ((T_OBJECT_CAST) reduce 112) ((T_PRINT) reduce 112) ((T_PROPERTY_C) reduce 112) ((T_READONLY) reduce 112) ((T_REQUIRE) reduce 112) ((T_REQUIRE_ONCE) reduce 112) ((T_RETURN) reduce 112) ((T_START_HEREDOC) reduce 112) ((T_STATIC) reduce 112) ((T_STRING) reduce 112) ((T_STRING_CAST) reduce 112) ((T_SWITCH) reduce 112) ((T_THROW) reduce 112) ((T_TRAIT) reduce 112) ((T_TRAIT_C) reduce 112) ((T_TRY) reduce 112) ((T_UNSET) reduce 112) ((T_UNSET_CAST) reduce 112) ((T_USE) reduce 112) ((T_VARIABLE) reduce 112) ((T_VOID_CAST) reduce 112) ((T_WHILE) reduce 112) ((T_YIELD) reduce 112) ((T_YIELD_FROM) reduce 112) (("[") reduce 112) (("`") reduce 112) (("{") reduce 112) (("}") reduce 112) (("~") reduce 112)) 32591 ((("!") reduce 250) (("\"") reduce 250) (("$") reduce 250) (("(") reduce 250) (("+") reduce 250) (("-") reduce 250) ((";") reduce 250) (("@") reduce 250) ((T_ABSTRACT) reduce 250) ((T_ARRAY) reduce 250) ((T_ARRAY_CAST) reduce 250) ((T_ATTRIBUTE) reduce 250) ((T_BOOL_CAST) reduce 250) ((T_BREAK) reduce 250) ((T_CLASS) reduce 250) ((T_CLASS_C) reduce 250) ((T_CLONE) reduce 250) ((T_CONST) reduce 250) ((T_CONSTANT_ENCAPSED_STRING) reduce 250) ((T_CONTINUE) reduce 250) ((T_DEC) reduce 250) ((T_DECLARE) reduce 250) ((T_DIR) reduce 250) ((T_DNUMBER) reduce 250) ((T_DO) reduce 250) ((T_DOUBLE_CAST) reduce 250) ((T_ECHO) reduce 250) ((T_EMPTY) reduce 250) ((T_ENUM) reduce 250) ((T_EVAL) reduce 250) ((T_EXIT) reduce 250) ((T_FILE) reduce 250) ((T_FINAL) reduce 250) ((T_FN) reduce 250) ((T_FOR) reduce 250) ((T_FOREACH) reduce 250) ((T_FUNCTION) reduce 250) ((T_FUNC_C) reduce 250) ((T_GLOBAL) reduce 250) ((T_GOTO) reduce 250) ((T_HALT_COMPILER) reduce 250) ((T_IF) reduce 250) ((T_INC) reduce 250) ((T_INCLUDE) reduce 250) ((T_INCLUDE_ONCE) reduce 250) ((T_INLINE_HTML) reduce 250) ((T_INTERFACE) reduce 250) ((T_INT_CAST) reduce 250) ((T_ISSET) reduce 250) ((T_LINE) reduce 250) ((T_LIST) reduce 250) ((T_LNUMBER) reduce 250) ((T_MATCH) reduce 250) ((T_METHOD_C) reduce 250) ((T_NAMESPACE) reduce 250) ((T_NAME_FULLY_QUALIFIED) reduce 250) ((T_NAME_QUALIFIED) reduce 250) ((T_NAME_RELATIVE) reduce 250) ((T_NEW) reduce 250) ((T_NS_C) reduce 250) ((T_OBJECT_CAST) reduce 250) ((T_PRINT) reduce 250) ((T_PROPERTY_C) reduce 250) ((T_READONLY) reduce 250) ((T_REQUIRE) reduce 250) ((T_REQUIRE_ONCE) reduce 250) ((T_RETURN) reduce 250) ((T_START_HEREDOC) reduce 250) ((T_STATIC) reduce 250) ((T_STRING) reduce 250) ((T_STRING_CAST) reduce 250) ((T_SWITCH) reduce 250) ((T_THROW) reduce 250) ((T_TRAIT) reduce 250) ((T_TRAIT_C) reduce 250) ((T_TRY) reduce 250) ((T_UNSET) reduce 250) ((T_UNSET_CAST) reduce 250) ((T_USE) reduce 250) ((T_VARIABLE) reduce 250) ((T_VOID_CAST) reduce 250) ((T_WHILE) reduce 250) ((T_YIELD) reduce 250) ((T_YIELD_FROM) reduce 250) (("[") reduce 250) (("`") reduce 250) (("{") reduce 250) (("}") reduce 250) (("~") reduce 250)) 32595 ((("!") reduce 251) (("\"") reduce 251) (("$") reduce 251) (("(") reduce 251) (("+") reduce 251) (("-") reduce 251) ((";") reduce 251) (("@") reduce 251) ((T_ABSTRACT) reduce 251) ((T_ARRAY) reduce 251) ((T_ARRAY_CAST) reduce 251) ((T_ATTRIBUTE) reduce 251) ((T_BOOL_CAST) reduce 251) ((T_BREAK) reduce 251) ((T_CLASS) reduce 251) ((T_CLASS_C) reduce 251) ((T_CLONE) reduce 251) ((T_CONST) reduce 251) ((T_CONSTANT_ENCAPSED_STRING) reduce 251) ((T_CONTINUE) reduce 251) ((T_DEC) reduce 251) ((T_DECLARE) reduce 251) ((T_DIR) reduce 251) ((T_DNUMBER) reduce 251) ((T_DO) reduce 251) ((T_DOUBLE_CAST) reduce 251) ((T_ECHO) reduce 251) ((T_EMPTY) reduce 251) ((T_ENUM) reduce 251) ((T_EVAL) reduce 251) ((T_EXIT) reduce 251) ((T_FILE) reduce 251) ((T_FINAL) reduce 251) ((T_FN) reduce 251) ((T_FOR) reduce 251) ((T_FOREACH) reduce 251) ((T_FUNCTION) reduce 251) ((T_FUNC_C) reduce 251) ((T_GLOBAL) reduce 251) ((T_GOTO) reduce 251) ((T_HALT_COMPILER) reduce 251) ((T_IF) reduce 251) ((T_INC) reduce 251) ((T_INCLUDE) reduce 251) ((T_INCLUDE_ONCE) reduce 251) ((T_INLINE_HTML) reduce 251) ((T_INTERFACE) reduce 251) ((T_INT_CAST) reduce 251) ((T_ISSET) reduce 251) ((T_LINE) reduce 251) ((T_LIST) reduce 251) ((T_LNUMBER) reduce 251) ((T_MATCH) reduce 251) ((T_METHOD_C) reduce 251) ((T_NAMESPACE) reduce 251) ((T_NAME_FULLY_QUALIFIED) reduce 251) ((T_NAME_QUALIFIED) reduce 251) ((T_NAME_RELATIVE) reduce 251) ((T_NEW) reduce 251) ((T_NS_C) reduce 251) ((T_OBJECT_CAST) reduce 251) ((T_PRINT) reduce 251) ((T_PROPERTY_C) reduce 251) ((T_READONLY) reduce 251) ((T_REQUIRE) reduce 251) ((T_REQUIRE_ONCE) reduce 251) ((T_RETURN) reduce 251) ((T_START_HEREDOC) reduce 251) ((T_STATIC) reduce 251) ((T_STRING) reduce 251) ((T_STRING_CAST) reduce 251) ((T_SWITCH) reduce 251) ((T_THROW) reduce 251) ((T_TRAIT) reduce 251) ((T_TRAIT_C) reduce 251) ((T_TRY) reduce 251) ((T_UNSET) reduce 251) ((T_UNSET_CAST) reduce 251) ((T_USE) reduce 251) ((T_VARIABLE) reduce 251) ((T_VOID_CAST) reduce 251) ((T_WHILE) reduce 251) ((T_YIELD) reduce 251) ((T_YIELD_FROM) reduce 251) (("[") reduce 251) (("`") reduce 251) (("{") reduce 251) (("}") reduce 251) (("~") reduce 251)) 32600 ((("!") reduce 242) (("\"") reduce 242) (("$") reduce 242) (("(") reduce 242) (("+") reduce 242) (("-") reduce 242) ((";") reduce 242) (("@") reduce 242) ((T_ABSTRACT) reduce 242) ((T_ARRAY) reduce 242) ((T_ARRAY_CAST) reduce 242) ((T_ATTRIBUTE) reduce 242) ((T_BOOL_CAST) reduce 242) ((T_BREAK) reduce 242) ((T_CLASS) reduce 242) ((T_CLASS_C) reduce 242) ((T_CLONE) reduce 242) ((T_CONST) reduce 242) ((T_CONSTANT_ENCAPSED_STRING) reduce 242) ((T_CONTINUE) reduce 242) ((T_DEC) reduce 242) ((T_DECLARE) reduce 242) ((T_DIR) reduce 242) ((T_DNUMBER) reduce 242) ((T_DO) reduce 242) ((T_DOUBLE_CAST) reduce 242) ((T_ECHO) reduce 242) ((T_EMPTY) reduce 242) ((T_ENUM) reduce 242) ((T_EVAL) reduce 242) ((T_EXIT) reduce 242) ((T_FILE) reduce 242) ((T_FINAL) reduce 242) ((T_FN) reduce 242) ((T_FOR) reduce 242) ((T_FOREACH) reduce 242) ((T_FUNCTION) reduce 242) ((T_FUNC_C) reduce 242) ((T_GLOBAL) reduce 242) ((T_GOTO) reduce 242) ((T_HALT_COMPILER) reduce 242) ((T_IF) reduce 242) ((T_INC) reduce 242) ((T_INCLUDE) reduce 242) ((T_INCLUDE_ONCE) reduce 242) ((T_INLINE_HTML) reduce 242) ((T_INTERFACE) reduce 242) ((T_INT_CAST) reduce 242) ((T_ISSET) reduce 242) ((T_LINE) reduce 242) ((T_LIST) reduce 242) ((T_LNUMBER) reduce 242) ((T_MATCH) reduce 242) ((T_METHOD_C) reduce 242) ((T_NAMESPACE) reduce 242) ((T_NAME_FULLY_QUALIFIED) reduce 242) ((T_NAME_QUALIFIED) reduce 242) ((T_NAME_RELATIVE) reduce 242) ((T_NEW) reduce 242) ((T_NS_C) reduce 242) ((T_OBJECT_CAST) reduce 242) ((T_PRINT) reduce 242) ((T_PROPERTY_C) reduce 242) ((T_READONLY) reduce 242) ((T_REQUIRE) reduce 242) ((T_REQUIRE_ONCE) reduce 242) ((T_RETURN) reduce 242) ((T_START_HEREDOC) reduce 242) ((T_STATIC) reduce 242) ((T_STRING) reduce 242) ((T_STRING_CAST) reduce 242) ((T_SWITCH) reduce 242) ((T_THROW) reduce 242) ((T_TRAIT) reduce 242) ((T_TRAIT_C) reduce 242) ((T_TRY) reduce 242) ((T_UNSET) reduce 242) ((T_UNSET_CAST) reduce 242) ((T_USE) reduce 242) ((T_VARIABLE) reduce 242) ((T_VOID_CAST) reduce 242) ((T_WHILE) reduce 242) ((T_YIELD) reduce 242) ((T_YIELD_FROM) reduce 242) (("[") reduce 242) (("`") reduce 242) (("{") reduce 242) (("}") reduce 242) (("~") reduce 242)) 32601 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONST) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_EMPTY) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAMESPACE) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_USE) reduce 150) ((T_VARIABLE) reduce 150) ((T_VOID_CAST) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("}") reduce 150) (("~") reduce 150)) 32604 ((("!") reduce 243) (("\"") reduce 243) (("$") reduce 243) (("(") reduce 243) (("+") reduce 243) (("-") reduce 243) ((";") reduce 243) (("@") reduce 243) ((T_ABSTRACT) reduce 243) ((T_ARRAY) reduce 243) ((T_ARRAY_CAST) reduce 243) ((T_ATTRIBUTE) reduce 243) ((T_BOOL_CAST) reduce 243) ((T_BREAK) reduce 243) ((T_CLASS) reduce 243) ((T_CLASS_C) reduce 243) ((T_CLONE) reduce 243) ((T_CONST) reduce 243) ((T_CONSTANT_ENCAPSED_STRING) reduce 243) ((T_CONTINUE) reduce 243) ((T_DEC) reduce 243) ((T_DECLARE) reduce 243) ((T_DIR) reduce 243) ((T_DNUMBER) reduce 243) ((T_DO) reduce 243) ((T_DOUBLE_CAST) reduce 243) ((T_ECHO) reduce 243) ((T_EMPTY) reduce 243) ((T_ENUM) reduce 243) ((T_EVAL) reduce 243) ((T_EXIT) reduce 243) ((T_FILE) reduce 243) ((T_FINAL) reduce 243) ((T_FN) reduce 243) ((T_FOR) reduce 243) ((T_FOREACH) reduce 243) ((T_FUNCTION) reduce 243) ((T_FUNC_C) reduce 243) ((T_GLOBAL) reduce 243) ((T_GOTO) reduce 243) ((T_HALT_COMPILER) reduce 243) ((T_IF) reduce 243) ((T_INC) reduce 243) ((T_INCLUDE) reduce 243) ((T_INCLUDE_ONCE) reduce 243) ((T_INLINE_HTML) reduce 243) ((T_INTERFACE) reduce 243) ((T_INT_CAST) reduce 243) ((T_ISSET) reduce 243) ((T_LINE) reduce 243) ((T_LIST) reduce 243) ((T_LNUMBER) reduce 243) ((T_MATCH) reduce 243) ((T_METHOD_C) reduce 243) ((T_NAMESPACE) reduce 243) ((T_NAME_FULLY_QUALIFIED) reduce 243) ((T_NAME_QUALIFIED) reduce 243) ((T_NAME_RELATIVE) reduce 243) ((T_NEW) reduce 243) ((T_NS_C) reduce 243) ((T_OBJECT_CAST) reduce 243) ((T_PRINT) reduce 243) ((T_PROPERTY_C) reduce 243) ((T_READONLY) reduce 243) ((T_REQUIRE) reduce 243) ((T_REQUIRE_ONCE) reduce 243) ((T_RETURN) reduce 243) ((T_START_HEREDOC) reduce 243) ((T_STATIC) reduce 243) ((T_STRING) reduce 243) ((T_STRING_CAST) reduce 243) ((T_SWITCH) reduce 243) ((T_THROW) reduce 243) ((T_TRAIT) reduce 243) ((T_TRAIT_C) reduce 243) ((T_TRY) reduce 243) ((T_UNSET) reduce 243) ((T_UNSET_CAST) reduce 243) ((T_USE) reduce 243) ((T_VARIABLE) reduce 243) ((T_VOID_CAST) reduce 243) ((T_WHILE) reduce 243) ((T_YIELD) reduce 243) ((T_YIELD_FROM) reduce 243) (("[") reduce 243) (("`") reduce 243) (("{") reduce 243) (("}") reduce 243) (("~") reduce 243)) 32606 ((("!") reduce 170) (("\"") reduce 170) (("$") reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONST) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_EMPTY) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAMESPACE) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_USE) reduce 170) ((T_VARIABLE) reduce 170) ((T_VOID_CAST) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("}") reduce 170) (("~") reduce 170)) 32612 ((("!") reduce 120) (("\"") reduce 120) (("$") reduce 120) (("(") reduce 120) (("+") reduce 120) (("-") reduce 120) ((";") reduce 120) (("@") reduce 120) ((T_ABSTRACT) reduce 120) ((T_ARRAY) reduce 120) ((T_ARRAY_CAST) reduce 120) ((T_ATTRIBUTE) reduce 120) ((T_BOOL_CAST) reduce 120) ((T_BREAK) reduce 120) ((T_CLASS) reduce 120) ((T_CLASS_C) reduce 120) ((T_CLONE) reduce 120) ((T_CONST) reduce 120) ((T_CONSTANT_ENCAPSED_STRING) reduce 120) ((T_CONTINUE) reduce 120) ((T_DEC) reduce 120) ((T_DECLARE) reduce 120) ((T_DIR) reduce 120) ((T_DNUMBER) reduce 120) ((T_DO) reduce 120) ((T_DOUBLE_CAST) reduce 120) ((T_ECHO) reduce 120) ((T_EMPTY) reduce 120) ((T_ENUM) reduce 120) ((T_EVAL) reduce 120) ((T_EXIT) reduce 120) ((T_FILE) reduce 120) ((T_FINAL) reduce 120) ((T_FN) reduce 120) ((T_FOR) reduce 120) ((T_FOREACH) reduce 120) ((T_FUNCTION) reduce 120) ((T_FUNC_C) reduce 120) ((T_GLOBAL) reduce 120) ((T_GOTO) reduce 120) ((T_HALT_COMPILER) reduce 120) ((T_IF) reduce 120) ((T_INC) reduce 120) ((T_INCLUDE) reduce 120) ((T_INCLUDE_ONCE) reduce 120) ((T_INLINE_HTML) reduce 120) ((T_INTERFACE) reduce 120) ((T_INT_CAST) reduce 120) ((T_ISSET) reduce 120) ((T_LINE) reduce 120) ((T_LIST) reduce 120) ((T_LNUMBER) reduce 120) ((T_MATCH) reduce 120) ((T_METHOD_C) reduce 120) ((T_NAMESPACE) reduce 120) ((T_NAME_FULLY_QUALIFIED) reduce 120) ((T_NAME_QUALIFIED) reduce 120) ((T_NAME_RELATIVE) reduce 120) ((T_NEW) reduce 120) ((T_NS_C) reduce 120) ((T_OBJECT_CAST) reduce 120) ((T_PRINT) reduce 120) ((T_PROPERTY_C) reduce 120) ((T_READONLY) reduce 120) ((T_REQUIRE) reduce 120) ((T_REQUIRE_ONCE) reduce 120) ((T_RETURN) reduce 120) ((T_START_HEREDOC) reduce 120) ((T_STATIC) reduce 120) ((T_STRING) reduce 120) ((T_STRING_CAST) reduce 120) ((T_SWITCH) reduce 120) ((T_THROW) reduce 120) ((T_TRAIT) reduce 120) ((T_TRAIT_C) reduce 120) ((T_TRY) reduce 120) ((T_UNSET) reduce 120) ((T_UNSET_CAST) reduce 120) ((T_USE) reduce 120) ((T_VARIABLE) reduce 120) ((T_VOID_CAST) reduce 120) ((T_WHILE) reduce 120) ((T_YIELD) reduce 120) ((T_YIELD_FROM) reduce 120) (("[") reduce 120) (("`") reduce 120) (("{") reduce 120) (("}") reduce 120) (("~") reduce 120)) 32613 ((("!") reduce 118) (("\"") reduce 118) (("$") reduce 118) (("(") reduce 118) (("+") reduce 118) (("-") reduce 118) ((";") reduce 118) (("@") reduce 118) ((T_ABSTRACT) reduce 118) ((T_ARRAY) reduce 118) ((T_ARRAY_CAST) reduce 118) ((T_ATTRIBUTE) reduce 118) ((T_BOOL_CAST) reduce 118) ((T_BREAK) reduce 118) ((T_CLASS) reduce 118) ((T_CLASS_C) reduce 118) ((T_CLONE) reduce 118) ((T_CONST) reduce 118) ((T_CONSTANT_ENCAPSED_STRING) reduce 118) ((T_CONTINUE) reduce 118) ((T_DEC) reduce 118) ((T_DECLARE) reduce 118) ((T_DIR) reduce 118) ((T_DNUMBER) reduce 118) ((T_DO) reduce 118) ((T_DOUBLE_CAST) reduce 118) ((T_ECHO) reduce 118) ((T_EMPTY) reduce 118) ((T_ENUM) reduce 118) ((T_EVAL) reduce 118) ((T_EXIT) reduce 118) ((T_FILE) reduce 118) ((T_FINAL) reduce 118) ((T_FN) reduce 118) ((T_FOR) reduce 118) ((T_FOREACH) reduce 118) ((T_FUNCTION) reduce 118) ((T_FUNC_C) reduce 118) ((T_GLOBAL) reduce 118) ((T_GOTO) reduce 118) ((T_HALT_COMPILER) reduce 118) ((T_IF) reduce 118) ((T_INC) reduce 118) ((T_INCLUDE) reduce 118) ((T_INCLUDE_ONCE) reduce 118) ((T_INLINE_HTML) reduce 118) ((T_INTERFACE) reduce 118) ((T_INT_CAST) reduce 118) ((T_ISSET) reduce 118) ((T_LINE) reduce 118) ((T_LIST) reduce 118) ((T_LNUMBER) reduce 118) ((T_MATCH) reduce 118) ((T_METHOD_C) reduce 118) ((T_NAMESPACE) reduce 118) ((T_NAME_FULLY_QUALIFIED) reduce 118) ((T_NAME_QUALIFIED) reduce 118) ((T_NAME_RELATIVE) reduce 118) ((T_NEW) reduce 118) ((T_NS_C) reduce 118) ((T_OBJECT_CAST) reduce 118) ((T_PRINT) reduce 118) ((T_PROPERTY_C) reduce 118) ((T_READONLY) reduce 118) ((T_REQUIRE) reduce 118) ((T_REQUIRE_ONCE) reduce 118) ((T_RETURN) reduce 118) ((T_START_HEREDOC) reduce 118) ((T_STATIC) reduce 118) ((T_STRING) reduce 118) ((T_STRING_CAST) reduce 118) ((T_SWITCH) reduce 118) ((T_THROW) reduce 118) ((T_TRAIT) reduce 118) ((T_TRAIT_C) reduce 118) ((T_TRY) reduce 118) ((T_UNSET) reduce 118) ((T_UNSET_CAST) reduce 118) ((T_USE) reduce 118) ((T_VARIABLE) reduce 118) ((T_VOID_CAST) reduce 118) ((T_WHILE) reduce 118) ((T_YIELD) reduce 118) ((T_YIELD_FROM) reduce 118) (("[") reduce 118) (("`") reduce 118) (("{") reduce 118) (("}") reduce 118) (("~") reduce 118)) 32614 ((("!") reduce 119) (("\"") reduce 119) (("$") reduce 119) (("(") reduce 119) (("+") reduce 119) (("-") reduce 119) ((";") reduce 119) (("@") reduce 119) ((T_ABSTRACT) reduce 119) ((T_ARRAY) reduce 119) ((T_ARRAY_CAST) reduce 119) ((T_ATTRIBUTE) reduce 119) ((T_BOOL_CAST) reduce 119) ((T_BREAK) reduce 119) ((T_CLASS) reduce 119) ((T_CLASS_C) reduce 119) ((T_CLONE) reduce 119) ((T_CONST) reduce 119) ((T_CONSTANT_ENCAPSED_STRING) reduce 119) ((T_CONTINUE) reduce 119) ((T_DEC) reduce 119) ((T_DECLARE) reduce 119) ((T_DIR) reduce 119) ((T_DNUMBER) reduce 119) ((T_DO) reduce 119) ((T_DOUBLE_CAST) reduce 119) ((T_ECHO) reduce 119) ((T_EMPTY) reduce 119) ((T_ENUM) reduce 119) ((T_EVAL) reduce 119) ((T_EXIT) reduce 119) ((T_FILE) reduce 119) ((T_FINAL) reduce 119) ((T_FN) reduce 119) ((T_FOR) reduce 119) ((T_FOREACH) reduce 119) ((T_FUNCTION) reduce 119) ((T_FUNC_C) reduce 119) ((T_GLOBAL) reduce 119) ((T_GOTO) reduce 119) ((T_HALT_COMPILER) reduce 119) ((T_IF) reduce 119) ((T_INC) reduce 119) ((T_INCLUDE) reduce 119) ((T_INCLUDE_ONCE) reduce 119) ((T_INLINE_HTML) reduce 119) ((T_INTERFACE) reduce 119) ((T_INT_CAST) reduce 119) ((T_ISSET) reduce 119) ((T_LINE) reduce 119) ((T_LIST) reduce 119) ((T_LNUMBER) reduce 119) ((T_MATCH) reduce 119) ((T_METHOD_C) reduce 119) ((T_NAMESPACE) reduce 119) ((T_NAME_FULLY_QUALIFIED) reduce 119) ((T_NAME_QUALIFIED) reduce 119) ((T_NAME_RELATIVE) reduce 119) ((T_NEW) reduce 119) ((T_NS_C) reduce 119) ((T_OBJECT_CAST) reduce 119) ((T_PRINT) reduce 119) ((T_PROPERTY_C) reduce 119) ((T_READONLY) reduce 119) ((T_REQUIRE) reduce 119) ((T_REQUIRE_ONCE) reduce 119) ((T_RETURN) reduce 119) ((T_START_HEREDOC) reduce 119) ((T_STATIC) reduce 119) ((T_STRING) reduce 119) ((T_STRING_CAST) reduce 119) ((T_SWITCH) reduce 119) ((T_THROW) reduce 119) ((T_TRAIT) reduce 119) ((T_TRAIT_C) reduce 119) ((T_TRY) reduce 119) ((T_UNSET) reduce 119) ((T_UNSET_CAST) reduce 119) ((T_USE) reduce 119) ((T_VARIABLE) reduce 119) ((T_VOID_CAST) reduce 119) ((T_WHILE) reduce 119) ((T_YIELD) reduce 119) ((T_YIELD_FROM) reduce 119) (("[") reduce 119) (("`") reduce 119) (("{") reduce 119) (("}") reduce 119) (("~") reduce 119)) 32615 ((("!") reduce 117) (("\"") reduce 117) (("$") reduce 117) (("(") reduce 117) (("+") reduce 117) (("-") reduce 117) ((";") reduce 117) (("@") reduce 117) ((T_ABSTRACT) reduce 117) ((T_ARRAY) reduce 117) ((T_ARRAY_CAST) reduce 117) ((T_ATTRIBUTE) reduce 117) ((T_BOOL_CAST) reduce 117) ((T_BREAK) reduce 117) ((T_CLASS) reduce 117) ((T_CLASS_C) reduce 117) ((T_CLONE) reduce 117) ((T_CONST) reduce 117) ((T_CONSTANT_ENCAPSED_STRING) reduce 117) ((T_CONTINUE) reduce 117) ((T_DEC) reduce 117) ((T_DECLARE) reduce 117) ((T_DIR) reduce 117) ((T_DNUMBER) reduce 117) ((T_DO) reduce 117) ((T_DOUBLE_CAST) reduce 117) ((T_ECHO) reduce 117) ((T_EMPTY) reduce 117) ((T_ENUM) reduce 117) ((T_EVAL) reduce 117) ((T_EXIT) reduce 117) ((T_FILE) reduce 117) ((T_FINAL) reduce 117) ((T_FN) reduce 117) ((T_FOR) reduce 117) ((T_FOREACH) reduce 117) ((T_FUNCTION) reduce 117) ((T_FUNC_C) reduce 117) ((T_GLOBAL) reduce 117) ((T_GOTO) reduce 117) ((T_HALT_COMPILER) reduce 117) ((T_IF) reduce 117) ((T_INC) reduce 117) ((T_INCLUDE) reduce 117) ((T_INCLUDE_ONCE) reduce 117) ((T_INLINE_HTML) reduce 117) ((T_INTERFACE) reduce 117) ((T_INT_CAST) reduce 117) ((T_ISSET) reduce 117) ((T_LINE) reduce 117) ((T_LIST) reduce 117) ((T_LNUMBER) reduce 117) ((T_MATCH) reduce 117) ((T_METHOD_C) reduce 117) ((T_NAMESPACE) reduce 117) ((T_NAME_FULLY_QUALIFIED) reduce 117) ((T_NAME_QUALIFIED) reduce 117) ((T_NAME_RELATIVE) reduce 117) ((T_NEW) reduce 117) ((T_NS_C) reduce 117) ((T_OBJECT_CAST) reduce 117) ((T_PRINT) reduce 117) ((T_PROPERTY_C) reduce 117) ((T_READONLY) reduce 117) ((T_REQUIRE) reduce 117) ((T_REQUIRE_ONCE) reduce 117) ((T_RETURN) reduce 117) ((T_START_HEREDOC) reduce 117) ((T_STATIC) reduce 117) ((T_STRING) reduce 117) ((T_STRING_CAST) reduce 117) ((T_SWITCH) reduce 117) ((T_THROW) reduce 117) ((T_TRAIT) reduce 117) ((T_TRAIT_C) reduce 117) ((T_TRY) reduce 117) ((T_UNSET) reduce 117) ((T_UNSET_CAST) reduce 117) ((T_USE) reduce 117) ((T_VARIABLE) reduce 117) ((T_VOID_CAST) reduce 117) ((T_WHILE) reduce 117) ((T_YIELD) reduce 117) ((T_YIELD_FROM) reduce 117) (("[") reduce 117) (("`") reduce 117) (("{") reduce 117) (("}") reduce 117) (("~") reduce 117)) 32620 ((("!") reduce 162) (("\"") reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONST) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_EMPTY) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAMESPACE) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_USE) reduce 162) ((T_VARIABLE) reduce 162) ((T_VOID_CAST) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("}") reduce 162) (("~") reduce 162)) 32623 ((("!") reduce 171) (("\"") reduce 171) (("$") reduce 171) (("(") reduce 171) (("+") reduce 171) (("-") reduce 171) ((";") reduce 171) (("@") reduce 171) ((T_ABSTRACT) reduce 171) ((T_ARRAY) reduce 171) ((T_ARRAY_CAST) reduce 171) ((T_ATTRIBUTE) reduce 171) ((T_BOOL_CAST) reduce 171) ((T_BREAK) reduce 171) ((T_CATCH) reduce 171) ((T_CLASS) reduce 171) ((T_CLASS_C) reduce 171) ((T_CLONE) reduce 171) ((T_CONST) reduce 171) ((T_CONSTANT_ENCAPSED_STRING) reduce 171) ((T_CONTINUE) reduce 171) ((T_DEC) reduce 171) ((T_DECLARE) reduce 171) ((T_DIR) reduce 171) ((T_DNUMBER) reduce 171) ((T_DO) reduce 171) ((T_DOUBLE_CAST) reduce 171) ((T_ECHO) reduce 171) ((T_EMPTY) reduce 171) ((T_ENUM) reduce 171) ((T_EVAL) reduce 171) ((T_EXIT) reduce 171) ((T_FILE) reduce 171) ((T_FINAL) reduce 171) ((T_FINALLY) reduce 171) ((T_FN) reduce 171) ((T_FOR) reduce 171) ((T_FOREACH) reduce 171) ((T_FUNCTION) reduce 171) ((T_FUNC_C) reduce 171) ((T_GLOBAL) reduce 171) ((T_GOTO) reduce 171) ((T_HALT_COMPILER) reduce 171) ((T_IF) reduce 171) ((T_INC) reduce 171) ((T_INCLUDE) reduce 171) ((T_INCLUDE_ONCE) reduce 171) ((T_INLINE_HTML) reduce 171) ((T_INTERFACE) reduce 171) ((T_INT_CAST) reduce 171) ((T_ISSET) reduce 171) ((T_LINE) reduce 171) ((T_LIST) reduce 171) ((T_LNUMBER) reduce 171) ((T_MATCH) reduce 171) ((T_METHOD_C) reduce 171) ((T_NAMESPACE) reduce 171) ((T_NAME_FULLY_QUALIFIED) reduce 171) ((T_NAME_QUALIFIED) reduce 171) ((T_NAME_RELATIVE) reduce 171) ((T_NEW) reduce 171) ((T_NS_C) reduce 171) ((T_OBJECT_CAST) reduce 171) ((T_PRINT) reduce 171) ((T_PROPERTY_C) reduce 171) ((T_READONLY) reduce 171) ((T_REQUIRE) reduce 171) ((T_REQUIRE_ONCE) reduce 171) ((T_RETURN) reduce 171) ((T_START_HEREDOC) reduce 171) ((T_STATIC) reduce 171) ((T_STRING) reduce 171) ((T_STRING_CAST) reduce 171) ((T_SWITCH) reduce 171) ((T_THROW) reduce 171) ((T_TRAIT) reduce 171) ((T_TRAIT_C) reduce 171) ((T_TRY) reduce 171) ((T_UNSET) reduce 171) ((T_UNSET_CAST) reduce 171) ((T_USE) reduce 171) ((T_VARIABLE) reduce 171) ((T_VOID_CAST) reduce 171) ((T_WHILE) reduce 171) ((T_YIELD) reduce 171) ((T_YIELD_FROM) reduce 171) (("[") reduce 171) (("`") reduce 171) (("{") reduce 171) (("}") reduce 171) (("~") reduce 171)) 32624 ((("!") reduce 177) (("\"") reduce 177) (("$") reduce 177) (("(") reduce 177) (("+") reduce 177) (("-") reduce 177) ((";") reduce 177) (("@") reduce 177) ((T_ABSTRACT) reduce 177) ((T_ARRAY) reduce 177) ((T_ARRAY_CAST) reduce 177) ((T_ATTRIBUTE) reduce 177) ((T_BOOL_CAST) reduce 177) ((T_BREAK) reduce 177) ((T_CATCH) shift) ((T_CLASS) reduce 177) ((T_CLASS_C) reduce 177) ((T_CLONE) reduce 177) ((T_CONST) reduce 177) ((T_CONSTANT_ENCAPSED_STRING) reduce 177) ((T_CONTINUE) reduce 177) ((T_DEC) reduce 177) ((T_DECLARE) reduce 177) ((T_DIR) reduce 177) ((T_DNUMBER) reduce 177) ((T_DO) reduce 177) ((T_DOUBLE_CAST) reduce 177) ((T_ECHO) reduce 177) ((T_EMPTY) reduce 177) ((T_ENUM) reduce 177) ((T_EVAL) reduce 177) ((T_EXIT) reduce 177) ((T_FILE) reduce 177) ((T_FINAL) reduce 177) ((T_FINALLY) shift) ((T_FN) reduce 177) ((T_FOR) reduce 177) ((T_FOREACH) reduce 177) ((T_FUNCTION) reduce 177) ((T_FUNC_C) reduce 177) ((T_GLOBAL) reduce 177) ((T_GOTO) reduce 177) ((T_HALT_COMPILER) reduce 177) ((T_IF) reduce 177) ((T_INC) reduce 177) ((T_INCLUDE) reduce 177) ((T_INCLUDE_ONCE) reduce 177) ((T_INLINE_HTML) reduce 177) ((T_INTERFACE) reduce 177) ((T_INT_CAST) reduce 177) ((T_ISSET) reduce 177) ((T_LINE) reduce 177) ((T_LIST) reduce 177) ((T_LNUMBER) reduce 177) ((T_MATCH) reduce 177) ((T_METHOD_C) reduce 177) ((T_NAMESPACE) reduce 177) ((T_NAME_FULLY_QUALIFIED) reduce 177) ((T_NAME_QUALIFIED) reduce 177) ((T_NAME_RELATIVE) reduce 177) ((T_NEW) reduce 177) ((T_NS_C) reduce 177) ((T_OBJECT_CAST) reduce 177) ((T_PRINT) reduce 177) ((T_PROPERTY_C) reduce 177) ((T_READONLY) reduce 177) ((T_REQUIRE) reduce 177) ((T_REQUIRE_ONCE) reduce 177) ((T_RETURN) reduce 177) ((T_START_HEREDOC) reduce 177) ((T_STATIC) reduce 177) ((T_STRING) reduce 177) ((T_STRING_CAST) reduce 177) ((T_SWITCH) reduce 177) ((T_THROW) reduce 177) ((T_TRAIT) reduce 177) ((T_TRAIT_C) reduce 177) ((T_TRY) reduce 177) ((T_UNSET) reduce 177) ((T_UNSET_CAST) reduce 177) ((T_USE) reduce 177) ((T_VARIABLE) reduce 177) ((T_VOID_CAST) reduce 177) ((T_WHILE) reduce 177) ((T_YIELD) reduce 177) ((T_YIELD_FROM) reduce 177) (("[") reduce 177) (("`") reduce 177) (("{") reduce 177) (("}") reduce 177) (("~") reduce 177)) 32627 ((("!") reduce 167) (("\"") reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONST) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_EMPTY) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAMESPACE) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_USE) reduce 167) ((T_VARIABLE) reduce 167) ((T_VOID_CAST) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("}") reduce 167) (("~") reduce 167)) 32630 ((("!") reduce 178) (("\"") reduce 178) (("$") reduce 178) (("(") reduce 178) (("+") reduce 178) (("-") reduce 178) ((";") reduce 178) (("@") reduce 178) ((T_ABSTRACT) reduce 178) ((T_ARRAY) reduce 178) ((T_ARRAY_CAST) reduce 178) ((T_ATTRIBUTE) reduce 178) ((T_BOOL_CAST) reduce 178) ((T_BREAK) reduce 178) ((T_CLASS) reduce 178) ((T_CLASS_C) reduce 178) ((T_CLONE) reduce 178) ((T_CONST) reduce 178) ((T_CONSTANT_ENCAPSED_STRING) reduce 178) ((T_CONTINUE) reduce 178) ((T_DEC) reduce 178) ((T_DECLARE) reduce 178) ((T_DIR) reduce 178) ((T_DNUMBER) reduce 178) ((T_DO) reduce 178) ((T_DOUBLE_CAST) reduce 178) ((T_ECHO) reduce 178) ((T_EMPTY) reduce 178) ((T_ENUM) reduce 178) ((T_EVAL) reduce 178) ((T_EXIT) reduce 178) ((T_FILE) reduce 178) ((T_FINAL) reduce 178) ((T_FN) reduce 178) ((T_FOR) reduce 178) ((T_FOREACH) reduce 178) ((T_FUNCTION) reduce 178) ((T_FUNC_C) reduce 178) ((T_GLOBAL) reduce 178) ((T_GOTO) reduce 178) ((T_HALT_COMPILER) reduce 178) ((T_IF) reduce 178) ((T_INC) reduce 178) ((T_INCLUDE) reduce 178) ((T_INCLUDE_ONCE) reduce 178) ((T_INLINE_HTML) reduce 178) ((T_INTERFACE) reduce 178) ((T_INT_CAST) reduce 178) ((T_ISSET) reduce 178) ((T_LINE) reduce 178) ((T_LIST) reduce 178) ((T_LNUMBER) reduce 178) ((T_MATCH) reduce 178) ((T_METHOD_C) reduce 178) ((T_NAMESPACE) reduce 178) ((T_NAME_FULLY_QUALIFIED) reduce 178) ((T_NAME_QUALIFIED) reduce 178) ((T_NAME_RELATIVE) reduce 178) ((T_NEW) reduce 178) ((T_NS_C) reduce 178) ((T_OBJECT_CAST) reduce 178) ((T_PRINT) reduce 178) ((T_PROPERTY_C) reduce 178) ((T_READONLY) reduce 178) ((T_REQUIRE) reduce 178) ((T_REQUIRE_ONCE) reduce 178) ((T_RETURN) reduce 178) ((T_START_HEREDOC) reduce 178) ((T_STATIC) reduce 178) ((T_STRING) reduce 178) ((T_STRING_CAST) reduce 178) ((T_SWITCH) reduce 178) ((T_THROW) reduce 178) ((T_TRAIT) reduce 178) ((T_TRAIT_C) reduce 178) ((T_TRY) reduce 178) ((T_UNSET) reduce 178) ((T_UNSET_CAST) reduce 178) ((T_USE) reduce 178) ((T_VARIABLE) reduce 178) ((T_VOID_CAST) reduce 178) ((T_WHILE) reduce 178) ((T_YIELD) reduce 178) ((T_YIELD_FROM) reduce 178) (("[") reduce 178) (("`") reduce 178) (("{") reduce 178) (("}") reduce 178) (("~") reduce 178)) 32637 ((("!") reduce 172) (("\"") reduce 172) (("$") reduce 172) (("(") reduce 172) (("+") reduce 172) (("-") reduce 172) ((";") reduce 172) (("@") reduce 172) ((T_ABSTRACT) reduce 172) ((T_ARRAY) reduce 172) ((T_ARRAY_CAST) reduce 172) ((T_ATTRIBUTE) reduce 172) ((T_BOOL_CAST) reduce 172) ((T_BREAK) reduce 172) ((T_CATCH) reduce 172) ((T_CLASS) reduce 172) ((T_CLASS_C) reduce 172) ((T_CLONE) reduce 172) ((T_CONST) reduce 172) ((T_CONSTANT_ENCAPSED_STRING) reduce 172) ((T_CONTINUE) reduce 172) ((T_DEC) reduce 172) ((T_DECLARE) reduce 172) ((T_DIR) reduce 172) ((T_DNUMBER) reduce 172) ((T_DO) reduce 172) ((T_DOUBLE_CAST) reduce 172) ((T_ECHO) reduce 172) ((T_EMPTY) reduce 172) ((T_ENUM) reduce 172) ((T_EVAL) reduce 172) ((T_EXIT) reduce 172) ((T_FILE) reduce 172) ((T_FINAL) reduce 172) ((T_FINALLY) reduce 172) ((T_FN) reduce 172) ((T_FOR) reduce 172) ((T_FOREACH) reduce 172) ((T_FUNCTION) reduce 172) ((T_FUNC_C) reduce 172) ((T_GLOBAL) reduce 172) ((T_GOTO) reduce 172) ((T_HALT_COMPILER) reduce 172) ((T_IF) reduce 172) ((T_INC) reduce 172) ((T_INCLUDE) reduce 172) ((T_INCLUDE_ONCE) reduce 172) ((T_INLINE_HTML) reduce 172) ((T_INTERFACE) reduce 172) ((T_INT_CAST) reduce 172) ((T_ISSET) reduce 172) ((T_LINE) reduce 172) ((T_LIST) reduce 172) ((T_LNUMBER) reduce 172) ((T_MATCH) reduce 172) ((T_METHOD_C) reduce 172) ((T_NAMESPACE) reduce 172) ((T_NAME_FULLY_QUALIFIED) reduce 172) ((T_NAME_QUALIFIED) reduce 172) ((T_NAME_RELATIVE) reduce 172) ((T_NEW) reduce 172) ((T_NS_C) reduce 172) ((T_OBJECT_CAST) reduce 172) ((T_PRINT) reduce 172) ((T_PROPERTY_C) reduce 172) ((T_READONLY) reduce 172) ((T_REQUIRE) reduce 172) ((T_REQUIRE_ONCE) reduce 172) ((T_RETURN) reduce 172) ((T_START_HEREDOC) reduce 172) ((T_STATIC) reduce 172) ((T_STRING) reduce 172) ((T_STRING_CAST) reduce 172) ((T_SWITCH) reduce 172) ((T_THROW) reduce 172) ((T_TRAIT) reduce 172) ((T_TRAIT_C) reduce 172) ((T_TRY) reduce 172) ((T_UNSET) reduce 172) ((T_UNSET_CAST) reduce 172) ((T_USE) reduce 172) ((T_VARIABLE) reduce 172) ((T_VOID_CAST) reduce 172) ((T_WHILE) reduce 172) ((T_YIELD) reduce 172) ((T_YIELD_FROM) reduce 172) (("[") reduce 172) (("`") reduce 172) (("{") reduce 172) (("}") reduce 172) (("~") reduce 172)) 32642 ((("!") reduce 200) (("\"") reduce 200) (("$") reduce 200) (("(") reduce 200) (("+") reduce 200) (("-") reduce 200) ((";") reduce 200) (("@") reduce 200) ((T_ABSTRACT) reduce 200) ((T_ARRAY) reduce 200) ((T_ARRAY_CAST) reduce 200) ((T_ATTRIBUTE) reduce 200) ((T_BOOL_CAST) reduce 200) ((T_BREAK) reduce 200) ((T_CLASS) reduce 200) ((T_CLASS_C) reduce 200) ((T_CLONE) reduce 200) ((T_CONST) reduce 200) ((T_CONSTANT_ENCAPSED_STRING) reduce 200) ((T_CONTINUE) reduce 200) ((T_DEC) reduce 200) ((T_DECLARE) reduce 200) ((T_DIR) reduce 200) ((T_DNUMBER) reduce 200) ((T_DO) reduce 200) ((T_DOUBLE_CAST) reduce 200) ((T_ECHO) reduce 200) ((T_EMPTY) reduce 200) ((T_ENUM) reduce 200) ((T_EVAL) reduce 200) ((T_EXIT) reduce 200) ((T_FILE) reduce 200) ((T_FINAL) reduce 200) ((T_FN) reduce 200) ((T_FOR) reduce 200) ((T_FOREACH) reduce 200) ((T_FUNCTION) reduce 200) ((T_FUNC_C) reduce 200) ((T_GLOBAL) reduce 200) ((T_GOTO) reduce 200) ((T_HALT_COMPILER) reduce 200) ((T_IF) reduce 200) ((T_INC) reduce 200) ((T_INCLUDE) reduce 200) ((T_INCLUDE_ONCE) reduce 200) ((T_INLINE_HTML) reduce 200) ((T_INTERFACE) reduce 200) ((T_INT_CAST) reduce 200) ((T_ISSET) reduce 200) ((T_LINE) reduce 200) ((T_LIST) reduce 200) ((T_LNUMBER) reduce 200) ((T_MATCH) reduce 200) ((T_METHOD_C) reduce 200) ((T_NAMESPACE) reduce 200) ((T_NAME_FULLY_QUALIFIED) reduce 200) ((T_NAME_QUALIFIED) reduce 200) ((T_NAME_RELATIVE) reduce 200) ((T_NEW) reduce 200) ((T_NS_C) reduce 200) ((T_OBJECT_CAST) reduce 200) ((T_PRINT) reduce 200) ((T_PROPERTY_C) reduce 200) ((T_READONLY) reduce 200) ((T_REQUIRE) reduce 200) ((T_REQUIRE_ONCE) reduce 200) ((T_RETURN) reduce 200) ((T_START_HEREDOC) reduce 200) ((T_STATIC) reduce 200) ((T_STRING) reduce 200) ((T_STRING_CAST) reduce 200) ((T_SWITCH) reduce 200) ((T_THROW) reduce 200) ((T_TRAIT) reduce 200) ((T_TRAIT_C) reduce 200) ((T_TRY) reduce 200) ((T_UNSET) reduce 200) ((T_UNSET_CAST) reduce 200) ((T_USE) reduce 200) ((T_VARIABLE) reduce 200) ((T_VOID_CAST) reduce 200) ((T_WHILE) reduce 200) ((T_YIELD) reduce 200) ((T_YIELD_FROM) reduce 200) (("[") reduce 200) (("`") reduce 200) (("{") reduce 200) (("}") reduce 200) (("~") reduce 200)) 32647 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONST) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_EMPTY) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAMESPACE) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_USE) reduce 153) ((T_VARIABLE) reduce 153) ((T_VOID_CAST) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("}") reduce 153) (("~") reduce 153)) 32651 ((("!") reduce 224) (("\"") reduce 224) (("$") reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONST) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_EMPTY) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAMESPACE) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_USE) reduce 224) ((T_VARIABLE) reduce 224) ((T_VOID_CAST) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("}") reduce 224) (("~") reduce 224)) 32653 ((("!") reduce 225) (("\"") reduce 225) (("$") reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONST) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_EMPTY) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAMESPACE) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_USE) reduce 225) ((T_VARIABLE) reduce 225) ((T_VOID_CAST) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("}") reduce 225) (("~") reduce 225)) 32657 ((("!") reduce 226) (("\"") reduce 226) (("$") reduce 226) (("(") reduce 226) (("+") reduce 226) (("-") reduce 226) ((";") reduce 226) (("@") reduce 226) ((T_ABSTRACT) reduce 226) ((T_ARRAY) reduce 226) ((T_ARRAY_CAST) reduce 226) ((T_ATTRIBUTE) reduce 226) ((T_BOOL_CAST) reduce 226) ((T_BREAK) reduce 226) ((T_CLASS) reduce 226) ((T_CLASS_C) reduce 226) ((T_CLONE) reduce 226) ((T_CONST) reduce 226) ((T_CONSTANT_ENCAPSED_STRING) reduce 226) ((T_CONTINUE) reduce 226) ((T_DEC) reduce 226) ((T_DECLARE) reduce 226) ((T_DIR) reduce 226) ((T_DNUMBER) reduce 226) ((T_DO) reduce 226) ((T_DOUBLE_CAST) reduce 226) ((T_ECHO) reduce 226) ((T_EMPTY) reduce 226) ((T_ENUM) reduce 226) ((T_EVAL) reduce 226) ((T_EXIT) reduce 226) ((T_FILE) reduce 226) ((T_FINAL) reduce 226) ((T_FN) reduce 226) ((T_FOR) reduce 226) ((T_FOREACH) reduce 226) ((T_FUNCTION) reduce 226) ((T_FUNC_C) reduce 226) ((T_GLOBAL) reduce 226) ((T_GOTO) reduce 226) ((T_HALT_COMPILER) reduce 226) ((T_IF) reduce 226) ((T_INC) reduce 226) ((T_INCLUDE) reduce 226) ((T_INCLUDE_ONCE) reduce 226) ((T_INLINE_HTML) reduce 226) ((T_INTERFACE) reduce 226) ((T_INT_CAST) reduce 226) ((T_ISSET) reduce 226) ((T_LINE) reduce 226) ((T_LIST) reduce 226) ((T_LNUMBER) reduce 226) ((T_MATCH) reduce 226) ((T_METHOD_C) reduce 226) ((T_NAMESPACE) reduce 226) ((T_NAME_FULLY_QUALIFIED) reduce 226) ((T_NAME_QUALIFIED) reduce 226) ((T_NAME_RELATIVE) reduce 226) ((T_NEW) reduce 226) ((T_NS_C) reduce 226) ((T_OBJECT_CAST) reduce 226) ((T_PRINT) reduce 226) ((T_PROPERTY_C) reduce 226) ((T_READONLY) reduce 226) ((T_REQUIRE) reduce 226) ((T_REQUIRE_ONCE) reduce 226) ((T_RETURN) reduce 226) ((T_START_HEREDOC) reduce 226) ((T_STATIC) reduce 226) ((T_STRING) reduce 226) ((T_STRING_CAST) reduce 226) ((T_SWITCH) reduce 226) ((T_THROW) reduce 226) ((T_TRAIT) reduce 226) ((T_TRAIT_C) reduce 226) ((T_TRY) reduce 226) ((T_UNSET) reduce 226) ((T_UNSET_CAST) reduce 226) ((T_USE) reduce 226) ((T_VARIABLE) reduce 226) ((T_VOID_CAST) reduce 226) ((T_WHILE) reduce 226) ((T_YIELD) reduce 226) ((T_YIELD_FROM) reduce 226) (("[") reduce 226) (("`") reduce 226) (("{") reduce 226) (("}") reduce 226) (("~") reduce 226)) 32660 ((("!") reduce 227) (("\"") reduce 227) (("$") reduce 227) (("(") reduce 227) (("+") reduce 227) (("-") reduce 227) ((";") reduce 227) (("@") reduce 227) ((T_ABSTRACT) reduce 227) ((T_ARRAY) reduce 227) ((T_ARRAY_CAST) reduce 227) ((T_ATTRIBUTE) reduce 227) ((T_BOOL_CAST) reduce 227) ((T_BREAK) reduce 227) ((T_CLASS) reduce 227) ((T_CLASS_C) reduce 227) ((T_CLONE) reduce 227) ((T_CONST) reduce 227) ((T_CONSTANT_ENCAPSED_STRING) reduce 227) ((T_CONTINUE) reduce 227) ((T_DEC) reduce 227) ((T_DECLARE) reduce 227) ((T_DIR) reduce 227) ((T_DNUMBER) reduce 227) ((T_DO) reduce 227) ((T_DOUBLE_CAST) reduce 227) ((T_ECHO) reduce 227) ((T_EMPTY) reduce 227) ((T_ENUM) reduce 227) ((T_EVAL) reduce 227) ((T_EXIT) reduce 227) ((T_FILE) reduce 227) ((T_FINAL) reduce 227) ((T_FN) reduce 227) ((T_FOR) reduce 227) ((T_FOREACH) reduce 227) ((T_FUNCTION) reduce 227) ((T_FUNC_C) reduce 227) ((T_GLOBAL) reduce 227) ((T_GOTO) reduce 227) ((T_HALT_COMPILER) reduce 227) ((T_IF) reduce 227) ((T_INC) reduce 227) ((T_INCLUDE) reduce 227) ((T_INCLUDE_ONCE) reduce 227) ((T_INLINE_HTML) reduce 227) ((T_INTERFACE) reduce 227) ((T_INT_CAST) reduce 227) ((T_ISSET) reduce 227) ((T_LINE) reduce 227) ((T_LIST) reduce 227) ((T_LNUMBER) reduce 227) ((T_MATCH) reduce 227) ((T_METHOD_C) reduce 227) ((T_NAMESPACE) reduce 227) ((T_NAME_FULLY_QUALIFIED) reduce 227) ((T_NAME_QUALIFIED) reduce 227) ((T_NAME_RELATIVE) reduce 227) ((T_NEW) reduce 227) ((T_NS_C) reduce 227) ((T_OBJECT_CAST) reduce 227) ((T_PRINT) reduce 227) ((T_PROPERTY_C) reduce 227) ((T_READONLY) reduce 227) ((T_REQUIRE) reduce 227) ((T_REQUIRE_ONCE) reduce 227) ((T_RETURN) reduce 227) ((T_START_HEREDOC) reduce 227) ((T_STATIC) reduce 227) ((T_STRING) reduce 227) ((T_STRING_CAST) reduce 227) ((T_SWITCH) reduce 227) ((T_THROW) reduce 227) ((T_TRAIT) reduce 227) ((T_TRAIT_C) reduce 227) ((T_TRY) reduce 227) ((T_UNSET) reduce 227) ((T_UNSET_CAST) reduce 227) ((T_USE) reduce 227) ((T_VARIABLE) reduce 227) ((T_VOID_CAST) reduce 227) ((T_WHILE) reduce 227) ((T_YIELD) reduce 227) ((T_YIELD_FROM) reduce 227) (("[") reduce 227) (("`") reduce 227) (("{") reduce 227) (("}") reduce 227) (("~") reduce 227)) 32661 ((("!") reduce 169) (("\"") reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONST) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_EMPTY) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAMESPACE) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_USE) reduce 169) ((T_VARIABLE) reduce 169) ((T_VOID_CAST) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("}") reduce 169) (("~") reduce 169)) 32663 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONST) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_EMPTY) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAMESPACE) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_USE) reduce 158) ((T_VARIABLE) reduce 158) ((T_VOID_CAST) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("}") reduce 158) (("~") reduce 158)) 32665 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONST) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_EMPTY) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAMESPACE) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_USE) reduce 156) ((T_VARIABLE) reduce 156) ((T_VOID_CAST) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("}") reduce 156) (("~") reduce 156)) 32669 ((("!") reduce 116) (("\"") reduce 116) (("$") reduce 116) (("(") reduce 116) (("+") reduce 116) (("-") reduce 116) ((";") reduce 116) (("@") reduce 116) ((T_ABSTRACT) reduce 116) ((T_ARRAY) reduce 116) ((T_ARRAY_CAST) reduce 116) ((T_ATTRIBUTE) reduce 116) ((T_BOOL_CAST) reduce 116) ((T_BREAK) reduce 116) ((T_CLASS) reduce 116) ((T_CLASS_C) reduce 116) ((T_CLONE) reduce 116) ((T_CONST) reduce 116) ((T_CONSTANT_ENCAPSED_STRING) reduce 116) ((T_CONTINUE) reduce 116) ((T_DEC) reduce 116) ((T_DECLARE) reduce 116) ((T_DIR) reduce 116) ((T_DNUMBER) reduce 116) ((T_DO) reduce 116) ((T_DOUBLE_CAST) reduce 116) ((T_ECHO) reduce 116) ((T_EMPTY) reduce 116) ((T_ENUM) reduce 116) ((T_EVAL) reduce 116) ((T_EXIT) reduce 116) ((T_FILE) reduce 116) ((T_FINAL) reduce 116) ((T_FN) reduce 116) ((T_FOR) reduce 116) ((T_FOREACH) reduce 116) ((T_FUNCTION) reduce 116) ((T_FUNC_C) reduce 116) ((T_GLOBAL) reduce 116) ((T_GOTO) reduce 116) ((T_HALT_COMPILER) reduce 116) ((T_IF) reduce 116) ((T_INC) reduce 116) ((T_INCLUDE) reduce 116) ((T_INCLUDE_ONCE) reduce 116) ((T_INLINE_HTML) reduce 116) ((T_INTERFACE) reduce 116) ((T_INT_CAST) reduce 116) ((T_ISSET) reduce 116) ((T_LINE) reduce 116) ((T_LIST) reduce 116) ((T_LNUMBER) reduce 116) ((T_MATCH) reduce 116) ((T_METHOD_C) reduce 116) ((T_NAMESPACE) reduce 116) ((T_NAME_FULLY_QUALIFIED) reduce 116) ((T_NAME_QUALIFIED) reduce 116) ((T_NAME_RELATIVE) reduce 116) ((T_NEW) reduce 116) ((T_NS_C) reduce 116) ((T_OBJECT_CAST) reduce 116) ((T_PRINT) reduce 116) ((T_PROPERTY_C) reduce 116) ((T_READONLY) reduce 116) ((T_REQUIRE) reduce 116) ((T_REQUIRE_ONCE) reduce 116) ((T_RETURN) reduce 116) ((T_START_HEREDOC) reduce 116) ((T_STATIC) reduce 116) ((T_STRING) reduce 116) ((T_STRING_CAST) reduce 116) ((T_SWITCH) reduce 116) ((T_THROW) reduce 116) ((T_TRAIT) reduce 116) ((T_TRAIT_C) reduce 116) ((T_TRY) reduce 116) ((T_UNSET) reduce 116) ((T_UNSET_CAST) reduce 116) ((T_USE) reduce 116) ((T_VARIABLE) reduce 116) ((T_VOID_CAST) reduce 116) ((T_WHILE) reduce 116) ((T_YIELD) reduce 116) ((T_YIELD_FROM) reduce 116) (("[") reduce 116) (("`") reduce 116) (("{") reduce 116) (("}") reduce 116) (("~") reduce 116)) 32670 ((("!") reduce 114) (("\"") reduce 114) (("$") reduce 114) (("(") reduce 114) (("+") reduce 114) (("-") reduce 114) ((";") reduce 114) (("@") reduce 114) ((T_ABSTRACT) reduce 114) ((T_ARRAY) reduce 114) ((T_ARRAY_CAST) reduce 114) ((T_ATTRIBUTE) reduce 114) ((T_BOOL_CAST) reduce 114) ((T_BREAK) reduce 114) ((T_CLASS) reduce 114) ((T_CLASS_C) reduce 114) ((T_CLONE) reduce 114) ((T_CONST) reduce 114) ((T_CONSTANT_ENCAPSED_STRING) reduce 114) ((T_CONTINUE) reduce 114) ((T_DEC) reduce 114) ((T_DECLARE) reduce 114) ((T_DIR) reduce 114) ((T_DNUMBER) reduce 114) ((T_DO) reduce 114) ((T_DOUBLE_CAST) reduce 114) ((T_ECHO) reduce 114) ((T_EMPTY) reduce 114) ((T_ENUM) reduce 114) ((T_EVAL) reduce 114) ((T_EXIT) reduce 114) ((T_FILE) reduce 114) ((T_FINAL) reduce 114) ((T_FN) reduce 114) ((T_FOR) reduce 114) ((T_FOREACH) reduce 114) ((T_FUNCTION) reduce 114) ((T_FUNC_C) reduce 114) ((T_GLOBAL) reduce 114) ((T_GOTO) reduce 114) ((T_HALT_COMPILER) reduce 114) ((T_IF) reduce 114) ((T_INC) reduce 114) ((T_INCLUDE) reduce 114) ((T_INCLUDE_ONCE) reduce 114) ((T_INLINE_HTML) reduce 114) ((T_INTERFACE) reduce 114) ((T_INT_CAST) reduce 114) ((T_ISSET) reduce 114) ((T_LINE) reduce 114) ((T_LIST) reduce 114) ((T_LNUMBER) reduce 114) ((T_MATCH) reduce 114) ((T_METHOD_C) reduce 114) ((T_NAMESPACE) reduce 114) ((T_NAME_FULLY_QUALIFIED) reduce 114) ((T_NAME_QUALIFIED) reduce 114) ((T_NAME_RELATIVE) reduce 114) ((T_NEW) reduce 114) ((T_NS_C) reduce 114) ((T_OBJECT_CAST) reduce 114) ((T_PRINT) reduce 114) ((T_PROPERTY_C) reduce 114) ((T_READONLY) reduce 114) ((T_REQUIRE) reduce 114) ((T_REQUIRE_ONCE) reduce 114) ((T_RETURN) reduce 114) ((T_START_HEREDOC) reduce 114) ((T_STATIC) reduce 114) ((T_STRING) reduce 114) ((T_STRING_CAST) reduce 114) ((T_SWITCH) reduce 114) ((T_THROW) reduce 114) ((T_TRAIT) reduce 114) ((T_TRAIT_C) reduce 114) ((T_TRY) reduce 114) ((T_UNSET) reduce 114) ((T_UNSET_CAST) reduce 114) ((T_USE) reduce 114) ((T_VARIABLE) reduce 114) ((T_VOID_CAST) reduce 114) ((T_WHILE) reduce 114) ((T_YIELD) reduce 114) ((T_YIELD_FROM) reduce 114) (("[") reduce 114) (("`") reduce 114) (("{") reduce 114) (("}") reduce 114) (("~") reduce 114)) 32673 ((("!") reduce 115) (("\"") reduce 115) (("$") reduce 115) (("(") reduce 115) (("+") reduce 115) (("-") reduce 115) ((";") reduce 115) (("@") reduce 115) ((T_ABSTRACT) reduce 115) ((T_ARRAY) reduce 115) ((T_ARRAY_CAST) reduce 115) ((T_ATTRIBUTE) reduce 115) ((T_BOOL_CAST) reduce 115) ((T_BREAK) reduce 115) ((T_CLASS) reduce 115) ((T_CLASS_C) reduce 115) ((T_CLONE) reduce 115) ((T_CONST) reduce 115) ((T_CONSTANT_ENCAPSED_STRING) reduce 115) ((T_CONTINUE) reduce 115) ((T_DEC) reduce 115) ((T_DECLARE) reduce 115) ((T_DIR) reduce 115) ((T_DNUMBER) reduce 115) ((T_DO) reduce 115) ((T_DOUBLE_CAST) reduce 115) ((T_ECHO) reduce 115) ((T_EMPTY) reduce 115) ((T_ENUM) reduce 115) ((T_EVAL) reduce 115) ((T_EXIT) reduce 115) ((T_FILE) reduce 115) ((T_FINAL) reduce 115) ((T_FN) reduce 115) ((T_FOR) reduce 115) ((T_FOREACH) reduce 115) ((T_FUNCTION) reduce 115) ((T_FUNC_C) reduce 115) ((T_GLOBAL) reduce 115) ((T_GOTO) reduce 115) ((T_HALT_COMPILER) reduce 115) ((T_IF) reduce 115) ((T_INC) reduce 115) ((T_INCLUDE) reduce 115) ((T_INCLUDE_ONCE) reduce 115) ((T_INLINE_HTML) reduce 115) ((T_INTERFACE) reduce 115) ((T_INT_CAST) reduce 115) ((T_ISSET) reduce 115) ((T_LINE) reduce 115) ((T_LIST) reduce 115) ((T_LNUMBER) reduce 115) ((T_MATCH) reduce 115) ((T_METHOD_C) reduce 115) ((T_NAMESPACE) reduce 115) ((T_NAME_FULLY_QUALIFIED) reduce 115) ((T_NAME_QUALIFIED) reduce 115) ((T_NAME_RELATIVE) reduce 115) ((T_NEW) reduce 115) ((T_NS_C) reduce 115) ((T_OBJECT_CAST) reduce 115) ((T_PRINT) reduce 115) ((T_PROPERTY_C) reduce 115) ((T_READONLY) reduce 115) ((T_REQUIRE) reduce 115) ((T_REQUIRE_ONCE) reduce 115) ((T_RETURN) reduce 115) ((T_START_HEREDOC) reduce 115) ((T_STATIC) reduce 115) ((T_STRING) reduce 115) ((T_STRING_CAST) reduce 115) ((T_SWITCH) reduce 115) ((T_THROW) reduce 115) ((T_TRAIT) reduce 115) ((T_TRAIT_C) reduce 115) ((T_TRY) reduce 115) ((T_UNSET) reduce 115) ((T_UNSET_CAST) reduce 115) ((T_USE) reduce 115) ((T_VARIABLE) reduce 115) ((T_VOID_CAST) reduce 115) ((T_WHILE) reduce 115) ((T_YIELD) reduce 115) ((T_YIELD_FROM) reduce 115) (("[") reduce 115) (("`") reduce 115) (("{") reduce 115) (("}") reduce 115) (("~") reduce 115)) 32679 ((("!") reduce 201) (("\"") reduce 201) (("$") reduce 201) (("(") reduce 201) (("+") reduce 201) (("-") reduce 201) ((";") reduce 201) (("@") reduce 201) ((T_ABSTRACT) reduce 201) ((T_ARRAY) reduce 201) ((T_ARRAY_CAST) reduce 201) ((T_ATTRIBUTE) reduce 201) ((T_BOOL_CAST) reduce 201) ((T_BREAK) reduce 201) ((T_CLASS) reduce 201) ((T_CLASS_C) reduce 201) ((T_CLONE) reduce 201) ((T_CONST) reduce 201) ((T_CONSTANT_ENCAPSED_STRING) reduce 201) ((T_CONTINUE) reduce 201) ((T_DEC) reduce 201) ((T_DECLARE) reduce 201) ((T_DIR) reduce 201) ((T_DNUMBER) reduce 201) ((T_DO) reduce 201) ((T_DOUBLE_CAST) reduce 201) ((T_ECHO) reduce 201) ((T_EMPTY) reduce 201) ((T_ENUM) reduce 201) ((T_EVAL) reduce 201) ((T_EXIT) reduce 201) ((T_FILE) reduce 201) ((T_FINAL) reduce 201) ((T_FN) reduce 201) ((T_FOR) reduce 201) ((T_FOREACH) reduce 201) ((T_FUNCTION) reduce 201) ((T_FUNC_C) reduce 201) ((T_GLOBAL) reduce 201) ((T_GOTO) reduce 201) ((T_HALT_COMPILER) reduce 201) ((T_IF) reduce 201) ((T_INC) reduce 201) ((T_INCLUDE) reduce 201) ((T_INCLUDE_ONCE) reduce 201) ((T_INLINE_HTML) reduce 201) ((T_INTERFACE) reduce 201) ((T_INT_CAST) reduce 201) ((T_ISSET) reduce 201) ((T_LINE) reduce 201) ((T_LIST) reduce 201) ((T_LNUMBER) reduce 201) ((T_MATCH) reduce 201) ((T_METHOD_C) reduce 201) ((T_NAMESPACE) reduce 201) ((T_NAME_FULLY_QUALIFIED) reduce 201) ((T_NAME_QUALIFIED) reduce 201) ((T_NAME_RELATIVE) reduce 201) ((T_NEW) reduce 201) ((T_NS_C) reduce 201) ((T_OBJECT_CAST) reduce 201) ((T_PRINT) reduce 201) ((T_PROPERTY_C) reduce 201) ((T_READONLY) reduce 201) ((T_REQUIRE) reduce 201) ((T_REQUIRE_ONCE) reduce 201) ((T_RETURN) reduce 201) ((T_START_HEREDOC) reduce 201) ((T_STATIC) reduce 201) ((T_STRING) reduce 201) ((T_STRING_CAST) reduce 201) ((T_SWITCH) reduce 201) ((T_THROW) reduce 201) ((T_TRAIT) reduce 201) ((T_TRAIT_C) reduce 201) ((T_TRY) reduce 201) ((T_UNSET) reduce 201) ((T_UNSET_CAST) reduce 201) ((T_USE) reduce 201) ((T_VARIABLE) reduce 201) ((T_VOID_CAST) reduce 201) ((T_WHILE) reduce 201) ((T_YIELD) reduce 201) ((T_YIELD_FROM) reduce 201) (("[") reduce 201) (("`") reduce 201) (("{") reduce 201) (("}") reduce 201) (("~") reduce 201)) 32683 ((("!") reduce 244) (("\"") reduce 244) (("$") reduce 244) (("(") reduce 244) (("+") reduce 244) (("-") reduce 244) ((";") reduce 244) (("@") reduce 244) ((T_ABSTRACT) reduce 244) ((T_ARRAY) reduce 244) ((T_ARRAY_CAST) reduce 244) ((T_ATTRIBUTE) reduce 244) ((T_BOOL_CAST) reduce 244) ((T_BREAK) reduce 244) ((T_CLASS) reduce 244) ((T_CLASS_C) reduce 244) ((T_CLONE) reduce 244) ((T_CONST) reduce 244) ((T_CONSTANT_ENCAPSED_STRING) reduce 244) ((T_CONTINUE) reduce 244) ((T_DEC) reduce 244) ((T_DECLARE) reduce 244) ((T_DIR) reduce 244) ((T_DNUMBER) reduce 244) ((T_DO) reduce 244) ((T_DOUBLE_CAST) reduce 244) ((T_ECHO) reduce 244) ((T_ELSE) reduce 244) ((T_ELSEIF) reduce 244) ((T_EMPTY) reduce 244) ((T_ENUM) reduce 244) ((T_EVAL) reduce 244) ((T_EXIT) reduce 244) ((T_FILE) reduce 244) ((T_FINAL) reduce 244) ((T_FN) reduce 244) ((T_FOR) reduce 244) ((T_FOREACH) reduce 244) ((T_FUNCTION) reduce 244) ((T_FUNC_C) reduce 244) ((T_GLOBAL) reduce 244) ((T_GOTO) reduce 244) ((T_HALT_COMPILER) reduce 244) ((T_IF) reduce 244) ((T_INC) reduce 244) ((T_INCLUDE) reduce 244) ((T_INCLUDE_ONCE) reduce 244) ((T_INLINE_HTML) reduce 244) ((T_INTERFACE) reduce 244) ((T_INT_CAST) reduce 244) ((T_ISSET) reduce 244) ((T_LINE) reduce 244) ((T_LIST) reduce 244) ((T_LNUMBER) reduce 244) ((T_MATCH) reduce 244) ((T_METHOD_C) reduce 244) ((T_NAMESPACE) reduce 244) ((T_NAME_FULLY_QUALIFIED) reduce 244) ((T_NAME_QUALIFIED) reduce 244) ((T_NAME_RELATIVE) reduce 244) ((T_NEW) reduce 244) ((T_NS_C) reduce 244) ((T_OBJECT_CAST) reduce 244) ((T_PRINT) reduce 244) ((T_PROPERTY_C) reduce 244) ((T_READONLY) reduce 244) ((T_REQUIRE) reduce 244) ((T_REQUIRE_ONCE) reduce 244) ((T_RETURN) reduce 244) ((T_START_HEREDOC) reduce 244) ((T_STATIC) reduce 244) ((T_STRING) reduce 244) ((T_STRING_CAST) reduce 244) ((T_SWITCH) reduce 244) ((T_THROW) reduce 244) ((T_TRAIT) reduce 244) ((T_TRAIT_C) reduce 244) ((T_TRY) reduce 244) ((T_UNSET) reduce 244) ((T_UNSET_CAST) reduce 244) ((T_USE) reduce 244) ((T_VARIABLE) reduce 244) ((T_VOID_CAST) reduce 244) ((T_WHILE) reduce 244) ((T_YIELD) reduce 244) ((T_YIELD_FROM) reduce 244) (("[") reduce 244) (("`") reduce 244) (("{") reduce 244) (("}") reduce 244) (("~") reduce 244)) 32686 ((("!") reduce 113) (("\"") reduce 113) (("$") reduce 113) (("(") reduce 113) (("+") reduce 113) (("-") reduce 113) ((";") reduce 113) (("@") reduce 113) ((T_ABSTRACT) reduce 113) ((T_ARRAY) reduce 113) ((T_ARRAY_CAST) reduce 113) ((T_ATTRIBUTE) reduce 113) ((T_BOOL_CAST) reduce 113) ((T_BREAK) reduce 113) ((T_CLASS) reduce 113) ((T_CLASS_C) reduce 113) ((T_CLONE) reduce 113) ((T_CONST) reduce 113) ((T_CONSTANT_ENCAPSED_STRING) reduce 113) ((T_CONTINUE) reduce 113) ((T_DEC) reduce 113) ((T_DECLARE) reduce 113) ((T_DIR) reduce 113) ((T_DNUMBER) reduce 113) ((T_DO) reduce 113) ((T_DOUBLE_CAST) reduce 113) ((T_ECHO) reduce 113) ((T_EMPTY) reduce 113) ((T_ENUM) reduce 113) ((T_EVAL) reduce 113) ((T_EXIT) reduce 113) ((T_FILE) reduce 113) ((T_FINAL) reduce 113) ((T_FN) reduce 113) ((T_FOR) reduce 113) ((T_FOREACH) reduce 113) ((T_FUNCTION) reduce 113) ((T_FUNC_C) reduce 113) ((T_GLOBAL) reduce 113) ((T_GOTO) reduce 113) ((T_HALT_COMPILER) reduce 113) ((T_IF) reduce 113) ((T_INC) reduce 113) ((T_INCLUDE) reduce 113) ((T_INCLUDE_ONCE) reduce 113) ((T_INLINE_HTML) reduce 113) ((T_INTERFACE) reduce 113) ((T_INT_CAST) reduce 113) ((T_ISSET) reduce 113) ((T_LINE) reduce 113) ((T_LIST) reduce 113) ((T_LNUMBER) reduce 113) ((T_MATCH) reduce 113) ((T_METHOD_C) reduce 113) ((T_NAMESPACE) reduce 113) ((T_NAME_FULLY_QUALIFIED) reduce 113) ((T_NAME_QUALIFIED) reduce 113) ((T_NAME_RELATIVE) reduce 113) ((T_NEW) reduce 113) ((T_NS_C) reduce 113) ((T_OBJECT_CAST) reduce 113) ((T_PRINT) reduce 113) ((T_PROPERTY_C) reduce 113) ((T_READONLY) reduce 113) ((T_REQUIRE) reduce 113) ((T_REQUIRE_ONCE) reduce 113) ((T_RETURN) reduce 113) ((T_START_HEREDOC) reduce 113) ((T_STATIC) reduce 113) ((T_STRING) reduce 113) ((T_STRING_CAST) reduce 113) ((T_SWITCH) reduce 113) ((T_THROW) reduce 113) ((T_TRAIT) reduce 113) ((T_TRAIT_C) reduce 113) ((T_TRY) reduce 113) ((T_UNSET) reduce 113) ((T_UNSET_CAST) reduce 113) ((T_USE) reduce 113) ((T_VARIABLE) reduce 113) ((T_VOID_CAST) reduce 113) ((T_WHILE) reduce 113) ((T_YIELD) reduce 113) ((T_YIELD_FROM) reduce 113) (("[") reduce 113) (("`") reduce 113) (("{") reduce 113) (("}") reduce 113) (("~") reduce 113)) 32688 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONST) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_EMPTY) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAMESPACE) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_USE) reduce 168) ((T_VARIABLE) reduce 168) ((T_VOID_CAST) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("}") reduce 168) (("~") reduce 168)) 32690 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONST) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_EMPTY) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAMESPACE) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_USE) reduce 157) ((T_VARIABLE) reduce 157) ((T_VOID_CAST) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("}") reduce 157) (("~") reduce 157)) 32700 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONST) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_EMPTY) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAMESPACE) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_USE) reduce 164) ((T_VARIABLE) reduce 164) ((T_VOID_CAST) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("}") reduce 164) (("~") reduce 164)) 32701 ((("!") reduce 220) (("\"") reduce 220) (("$") reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONST) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_EMPTY) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAMESPACE) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_USE) reduce 220) ((T_VARIABLE) reduce 220) ((T_VOID_CAST) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("}") reduce 220) (("~") reduce 220)) 32704 ((("!") reduce 221) (("\"") reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONST) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_EMPTY) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAMESPACE) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_USE) reduce 221) ((T_VARIABLE) reduce 221) ((T_VOID_CAST) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("}") reduce 221) (("~") reduce 221)) 32705 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONST) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_EMPTY) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAMESPACE) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_USE) reduce 163) ((T_VARIABLE) reduce 163) ((T_VOID_CAST) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("}") reduce 163) (("~") reduce 163)) 32714 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONST) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_EMPTY) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAMESPACE) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_USE) reduce 152) ((T_VARIABLE) reduce 152) ((T_VOID_CAST) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("}") reduce 152) (("~") reduce 152)) 32715 ((("!") reduce 218) (("\"") reduce 218) (("$") reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONST) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_EMPTY) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAMESPACE) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_USE) reduce 218) ((T_VARIABLE) reduce 218) ((T_VOID_CAST) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("}") reduce 218) (("~") reduce 218)) 32718 ((("!") reduce 219) (("\"") reduce 219) (("$") reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONST) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_EMPTY) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAMESPACE) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_USE) reduce 219) ((T_VARIABLE) reduce 219) ((T_VOID_CAST) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("}") reduce 219) (("~") reduce 219)) 32725 ((("!") reduce 202) (("\"") reduce 202) (("$") reduce 202) (("(") reduce 202) (("+") reduce 202) (("-") reduce 202) ((";") reduce 202) (("@") reduce 202) ((T_ABSTRACT) reduce 202) ((T_ARRAY) reduce 202) ((T_ARRAY_CAST) reduce 202) ((T_ATTRIBUTE) reduce 202) ((T_BOOL_CAST) reduce 202) ((T_BREAK) reduce 202) ((T_CLASS) reduce 202) ((T_CLASS_C) reduce 202) ((T_CLONE) reduce 202) ((T_CONST) reduce 202) ((T_CONSTANT_ENCAPSED_STRING) reduce 202) ((T_CONTINUE) reduce 202) ((T_DEC) reduce 202) ((T_DECLARE) reduce 202) ((T_DIR) reduce 202) ((T_DNUMBER) reduce 202) ((T_DO) reduce 202) ((T_DOUBLE_CAST) reduce 202) ((T_ECHO) reduce 202) ((T_EMPTY) reduce 202) ((T_ENUM) reduce 202) ((T_EVAL) reduce 202) ((T_EXIT) reduce 202) ((T_FILE) reduce 202) ((T_FINAL) reduce 202) ((T_FN) reduce 202) ((T_FOR) reduce 202) ((T_FOREACH) reduce 202) ((T_FUNCTION) reduce 202) ((T_FUNC_C) reduce 202) ((T_GLOBAL) reduce 202) ((T_GOTO) reduce 202) ((T_HALT_COMPILER) reduce 202) ((T_IF) reduce 202) ((T_INC) reduce 202) ((T_INCLUDE) reduce 202) ((T_INCLUDE_ONCE) reduce 202) ((T_INLINE_HTML) reduce 202) ((T_INTERFACE) reduce 202) ((T_INT_CAST) reduce 202) ((T_ISSET) reduce 202) ((T_LINE) reduce 202) ((T_LIST) reduce 202) ((T_LNUMBER) reduce 202) ((T_MATCH) reduce 202) ((T_METHOD_C) reduce 202) ((T_NAMESPACE) reduce 202) ((T_NAME_FULLY_QUALIFIED) reduce 202) ((T_NAME_QUALIFIED) reduce 202) ((T_NAME_RELATIVE) reduce 202) ((T_NEW) reduce 202) ((T_NS_C) reduce 202) ((T_OBJECT_CAST) reduce 202) ((T_PRINT) reduce 202) ((T_PROPERTY_C) reduce 202) ((T_READONLY) reduce 202) ((T_REQUIRE) reduce 202) ((T_REQUIRE_ONCE) reduce 202) ((T_RETURN) reduce 202) ((T_START_HEREDOC) reduce 202) ((T_STATIC) reduce 202) ((T_STRING) reduce 202) ((T_STRING_CAST) reduce 202) ((T_SWITCH) reduce 202) ((T_THROW) reduce 202) ((T_TRAIT) reduce 202) ((T_TRAIT_C) reduce 202) ((T_TRY) reduce 202) ((T_UNSET) reduce 202) ((T_UNSET_CAST) reduce 202) ((T_USE) reduce 202) ((T_VARIABLE) reduce 202) ((T_VOID_CAST) reduce 202) ((T_WHILE) reduce 202) ((T_YIELD) reduce 202) ((T_YIELD_FROM) reduce 202) (("[") reduce 202) (("`") reduce 202) (("{") reduce 202) (("}") reduce 202) (("~") reduce 202)) 32727 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONST) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_EMPTY) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAMESPACE) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_USE) reduce 159) ((T_VARIABLE) reduce 159) ((T_VOID_CAST) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("}") reduce 159) (("~") reduce 159)) 32733 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONST) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_EMPTY) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAMESPACE) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_USE) reduce 151) ((T_VARIABLE) reduce 151) ((T_VOID_CAST) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("}") reduce 151) (("~") reduce 151)) 32738 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONST) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_EMPTY) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAMESPACE) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_USE) reduce 165) ((T_VARIABLE) reduce 165) ((T_VOID_CAST) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("}") reduce 165) (("~") reduce 165)) 32739 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONST) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_EMPTY) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAMESPACE) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_USE) reduce 222) ((T_VARIABLE) reduce 222) ((T_VOID_CAST) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("}") reduce 222) (("~") reduce 222)) 32742 ((("!") reduce 223) (("\"") reduce 223) (("$") reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONST) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_EMPTY) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAMESPACE) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_USE) reduce 223) ((T_VARIABLE) reduce 223) ((T_VOID_CAST) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("}") reduce 223) (("~") reduce 223)) 32744 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONST) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_EMPTY) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAMESPACE) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_USE) reduce 155) ((T_VARIABLE) reduce 155) ((T_VOID_CAST) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("}") reduce 155) (("~") reduce 155)) 32746 ((("!") reduce 109) (("\"") reduce 109) (("$") reduce 109) (("(") reduce 109) (("+") reduce 109) (("-") reduce 109) ((";") reduce 109) (("@") reduce 109) ((T_ABSTRACT) reduce 109) ((T_ARRAY) reduce 109) ((T_ARRAY_CAST) reduce 109) ((T_ATTRIBUTE) reduce 109) ((T_BOOL_CAST) reduce 109) ((T_BREAK) reduce 109) ((T_CLASS) reduce 109) ((T_CLASS_C) reduce 109) ((T_CLONE) reduce 109) ((T_CONST) reduce 109) ((T_CONSTANT_ENCAPSED_STRING) reduce 109) ((T_CONTINUE) reduce 109) ((T_DEC) reduce 109) ((T_DECLARE) reduce 109) ((T_DIR) reduce 109) ((T_DNUMBER) reduce 109) ((T_DO) reduce 109) ((T_DOUBLE_CAST) reduce 109) ((T_ECHO) reduce 109) ((T_EMPTY) reduce 109) ((T_ENUM) reduce 109) ((T_EVAL) reduce 109) ((T_EXIT) reduce 109) ((T_FILE) reduce 109) ((T_FINAL) reduce 109) ((T_FN) reduce 109) ((T_FOR) reduce 109) ((T_FOREACH) reduce 109) ((T_FUNCTION) reduce 109) ((T_FUNC_C) reduce 109) ((T_GLOBAL) reduce 109) ((T_GOTO) reduce 109) ((T_HALT_COMPILER) reduce 109) ((T_IF) reduce 109) ((T_INC) reduce 109) ((T_INCLUDE) reduce 109) ((T_INCLUDE_ONCE) reduce 109) ((T_INLINE_HTML) reduce 109) ((T_INTERFACE) reduce 109) ((T_INT_CAST) reduce 109) ((T_ISSET) reduce 109) ((T_LINE) reduce 109) ((T_LIST) reduce 109) ((T_LNUMBER) reduce 109) ((T_MATCH) reduce 109) ((T_METHOD_C) reduce 109) ((T_NAMESPACE) reduce 109) ((T_NAME_FULLY_QUALIFIED) reduce 109) ((T_NAME_QUALIFIED) reduce 109) ((T_NAME_RELATIVE) reduce 109) ((T_NEW) reduce 109) ((T_NS_C) reduce 109) ((T_OBJECT_CAST) reduce 109) ((T_PRINT) reduce 109) ((T_PROPERTY_C) reduce 109) ((T_READONLY) reduce 109) ((T_REQUIRE) reduce 109) ((T_REQUIRE_ONCE) reduce 109) ((T_RETURN) reduce 109) ((T_START_HEREDOC) reduce 109) ((T_STATIC) reduce 109) ((T_STRING) reduce 109) ((T_STRING_CAST) reduce 109) ((T_SWITCH) reduce 109) ((T_THROW) reduce 109) ((T_TRAIT) reduce 109) ((T_TRAIT_C) reduce 109) ((T_TRY) reduce 109) ((T_UNSET) reduce 109) ((T_UNSET_CAST) reduce 109) ((T_USE) reduce 109) ((T_VARIABLE) reduce 109) ((T_VOID_CAST) reduce 109) ((T_WHILE) reduce 109) ((T_YIELD) reduce 109) ((T_YIELD_FROM) reduce 109) (("[") reduce 109) (("`") reduce 109) (("{") reduce 109) (("}") reduce 109) (("~") reduce 109)) 32753 ((("!") reduce 190) (("\"") reduce 190) (("$") reduce 190) (("(") reduce 190) (("+") reduce 190) (("-") reduce 190) ((";") reduce 190) (("@") reduce 190) ((T_ABSTRACT) reduce 190) ((T_ARRAY) reduce 190) ((T_ARRAY_CAST) reduce 190) ((T_ATTRIBUTE) reduce 190) ((T_BOOL_CAST) reduce 190) ((T_BREAK) reduce 190) ((T_CLASS) reduce 190) ((T_CLASS_C) reduce 190) ((T_CLONE) reduce 190) ((T_CONST) reduce 190) ((T_CONSTANT_ENCAPSED_STRING) reduce 190) ((T_CONTINUE) reduce 190) ((T_DEC) reduce 190) ((T_DECLARE) reduce 190) ((T_DIR) reduce 190) ((T_DNUMBER) reduce 190) ((T_DO) reduce 190) ((T_DOUBLE_CAST) reduce 190) ((T_ECHO) reduce 190) ((T_EMPTY) reduce 190) ((T_ENUM) reduce 190) ((T_EVAL) reduce 190) ((T_EXIT) reduce 190) ((T_FILE) reduce 190) ((T_FINAL) reduce 190) ((T_FN) reduce 190) ((T_FOR) reduce 190) ((T_FOREACH) reduce 190) ((T_FUNCTION) reduce 190) ((T_FUNC_C) reduce 190) ((T_GLOBAL) reduce 190) ((T_GOTO) reduce 190) ((T_HALT_COMPILER) reduce 190) ((T_IF) reduce 190) ((T_INC) reduce 190) ((T_INCLUDE) reduce 190) ((T_INCLUDE_ONCE) reduce 190) ((T_INLINE_HTML) reduce 190) ((T_INTERFACE) reduce 190) ((T_INT_CAST) reduce 190) ((T_ISSET) reduce 190) ((T_LINE) reduce 190) ((T_LIST) reduce 190) ((T_LNUMBER) reduce 190) ((T_MATCH) reduce 190) ((T_METHOD_C) reduce 190) ((T_NAMESPACE) reduce 190) ((T_NAME_FULLY_QUALIFIED) reduce 190) ((T_NAME_QUALIFIED) reduce 190) ((T_NAME_RELATIVE) reduce 190) ((T_NEW) reduce 190) ((T_NS_C) reduce 190) ((T_OBJECT_CAST) reduce 190) ((T_PRINT) reduce 190) ((T_PROPERTY_C) reduce 190) ((T_READONLY) reduce 190) ((T_REQUIRE) reduce 190) ((T_REQUIRE_ONCE) reduce 190) ((T_RETURN) reduce 190) ((T_START_HEREDOC) reduce 190) ((T_STATIC) reduce 190) ((T_STRING) reduce 190) ((T_STRING_CAST) reduce 190) ((T_SWITCH) reduce 190) ((T_THROW) reduce 190) ((T_TRAIT) reduce 190) ((T_TRAIT_C) reduce 190) ((T_TRY) reduce 190) ((T_UNSET) reduce 190) ((T_UNSET_CAST) reduce 190) ((T_USE) reduce 190) ((T_VARIABLE) reduce 190) ((T_VOID_CAST) reduce 190) ((T_WHILE) reduce 190) ((T_YIELD) reduce 190) ((T_YIELD_FROM) reduce 190) (("[") reduce 190) (("`") reduce 190) (("{") reduce 190) (("}") reduce 190) (("~") reduce 190)) 32755 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONST) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_EMPTY) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAMESPACE) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_USE) reduce 154) ((T_VARIABLE) reduce 154) ((T_VOID_CAST) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("}") reduce 154) (("~") reduce 154)) 32756 ((("!") reduce 114) (("\"") reduce 114) (("$") reduce 114) (($) reduce 114) (("(") reduce 114) (("+") reduce 114) (("-") reduce 114) ((";") reduce 114) (("@") reduce 114) ((T_ABSTRACT) reduce 114) ((T_ARRAY) reduce 114) ((T_ARRAY_CAST) reduce 114) ((T_ATTRIBUTE) reduce 114) ((T_BOOL_CAST) reduce 114) ((T_BREAK) reduce 114) ((T_CLASS) reduce 114) ((T_CLASS_C) reduce 114) ((T_CLONE) reduce 114) ((T_CONST) reduce 114) ((T_CONSTANT_ENCAPSED_STRING) reduce 114) ((T_CONTINUE) reduce 114) ((T_DEC) reduce 114) ((T_DECLARE) reduce 114) ((T_DIR) reduce 114) ((T_DNUMBER) reduce 114) ((T_DO) reduce 114) ((T_DOUBLE_CAST) reduce 114) ((T_ECHO) reduce 114) ((T_EMPTY) reduce 114) ((T_ENUM) reduce 114) ((T_EVAL) reduce 114) ((T_EXIT) reduce 114) ((T_FILE) reduce 114) ((T_FINAL) reduce 114) ((T_FN) reduce 114) ((T_FOR) reduce 114) ((T_FOREACH) reduce 114) ((T_FUNCTION) reduce 114) ((T_FUNC_C) reduce 114) ((T_GLOBAL) reduce 114) ((T_GOTO) reduce 114) ((T_HALT_COMPILER) reduce 114) ((T_IF) reduce 114) ((T_INC) reduce 114) ((T_INCLUDE) reduce 114) ((T_INCLUDE_ONCE) reduce 114) ((T_INLINE_HTML) reduce 114) ((T_INTERFACE) reduce 114) ((T_INT_CAST) reduce 114) ((T_ISSET) reduce 114) ((T_LINE) reduce 114) ((T_LIST) reduce 114) ((T_LNUMBER) reduce 114) ((T_MATCH) reduce 114) ((T_METHOD_C) reduce 114) ((T_NAMESPACE) reduce 114) ((T_NAME_FULLY_QUALIFIED) reduce 114) ((T_NAME_QUALIFIED) reduce 114) ((T_NAME_RELATIVE) reduce 114) ((T_NEW) reduce 114) ((T_NS_C) reduce 114) ((T_OBJECT_CAST) reduce 114) ((T_PRINT) reduce 114) ((T_PROPERTY_C) reduce 114) ((T_READONLY) reduce 114) ((T_REQUIRE) reduce 114) ((T_REQUIRE_ONCE) reduce 114) ((T_RETURN) reduce 114) ((T_START_HEREDOC) reduce 114) ((T_STATIC) reduce 114) ((T_STRING) reduce 114) ((T_STRING_CAST) reduce 114) ((T_SWITCH) reduce 114) ((T_THROW) reduce 114) ((T_TRAIT) reduce 114) ((T_TRAIT_C) reduce 114) ((T_TRY) reduce 114) ((T_UNSET) reduce 114) ((T_UNSET_CAST) reduce 114) ((T_USE) reduce 114) ((T_VARIABLE) reduce 114) ((T_VOID_CAST) reduce 114) ((T_WHILE) reduce 114) ((T_YIELD) reduce 114) ((T_YIELD_FROM) reduce 114) (("[") reduce 114) (("`") reduce 114) (("{") reduce 114) (("~") reduce 114)) 32759 ((("!") reduce 115) (("\"") reduce 115) (("$") reduce 115) (($) reduce 115) (("(") reduce 115) (("+") reduce 115) (("-") reduce 115) ((";") reduce 115) (("@") reduce 115) ((T_ABSTRACT) reduce 115) ((T_ARRAY) reduce 115) ((T_ARRAY_CAST) reduce 115) ((T_ATTRIBUTE) reduce 115) ((T_BOOL_CAST) reduce 115) ((T_BREAK) reduce 115) ((T_CLASS) reduce 115) ((T_CLASS_C) reduce 115) ((T_CLONE) reduce 115) ((T_CONST) reduce 115) ((T_CONSTANT_ENCAPSED_STRING) reduce 115) ((T_CONTINUE) reduce 115) ((T_DEC) reduce 115) ((T_DECLARE) reduce 115) ((T_DIR) reduce 115) ((T_DNUMBER) reduce 115) ((T_DO) reduce 115) ((T_DOUBLE_CAST) reduce 115) ((T_ECHO) reduce 115) ((T_EMPTY) reduce 115) ((T_ENUM) reduce 115) ((T_EVAL) reduce 115) ((T_EXIT) reduce 115) ((T_FILE) reduce 115) ((T_FINAL) reduce 115) ((T_FN) reduce 115) ((T_FOR) reduce 115) ((T_FOREACH) reduce 115) ((T_FUNCTION) reduce 115) ((T_FUNC_C) reduce 115) ((T_GLOBAL) reduce 115) ((T_GOTO) reduce 115) ((T_HALT_COMPILER) reduce 115) ((T_IF) reduce 115) ((T_INC) reduce 115) ((T_INCLUDE) reduce 115) ((T_INCLUDE_ONCE) reduce 115) ((T_INLINE_HTML) reduce 115) ((T_INTERFACE) reduce 115) ((T_INT_CAST) reduce 115) ((T_ISSET) reduce 115) ((T_LINE) reduce 115) ((T_LIST) reduce 115) ((T_LNUMBER) reduce 115) ((T_MATCH) reduce 115) ((T_METHOD_C) reduce 115) ((T_NAMESPACE) reduce 115) ((T_NAME_FULLY_QUALIFIED) reduce 115) ((T_NAME_QUALIFIED) reduce 115) ((T_NAME_RELATIVE) reduce 115) ((T_NEW) reduce 115) ((T_NS_C) reduce 115) ((T_OBJECT_CAST) reduce 115) ((T_PRINT) reduce 115) ((T_PROPERTY_C) reduce 115) ((T_READONLY) reduce 115) ((T_REQUIRE) reduce 115) ((T_REQUIRE_ONCE) reduce 115) ((T_RETURN) reduce 115) ((T_START_HEREDOC) reduce 115) ((T_STATIC) reduce 115) ((T_STRING) reduce 115) ((T_STRING_CAST) reduce 115) ((T_SWITCH) reduce 115) ((T_THROW) reduce 115) ((T_TRAIT) reduce 115) ((T_TRAIT_C) reduce 115) ((T_TRY) reduce 115) ((T_UNSET) reduce 115) ((T_UNSET_CAST) reduce 115) ((T_USE) reduce 115) ((T_VARIABLE) reduce 115) ((T_VOID_CAST) reduce 115) ((T_WHILE) reduce 115) ((T_YIELD) reduce 115) ((T_YIELD_FROM) reduce 115) (("[") reduce 115) (("`") reduce 115) (("{") reduce 115) (("~") reduce 115)) 32781 ((("!") reduce 201) (("\"") reduce 201) (($) reduce 201) (("$") reduce 201) (("(") reduce 201) (("+") reduce 201) (("-") reduce 201) ((";") reduce 201) (("@") reduce 201) ((T_ABSTRACT) reduce 201) ((T_ARRAY) reduce 201) ((T_ARRAY_CAST) reduce 201) ((T_ATTRIBUTE) reduce 201) ((T_BOOL_CAST) reduce 201) ((T_BREAK) reduce 201) ((T_CLASS) reduce 201) ((T_CLASS_C) reduce 201) ((T_CLONE) reduce 201) ((T_CONST) reduce 201) ((T_CONSTANT_ENCAPSED_STRING) reduce 201) ((T_CONTINUE) reduce 201) ((T_DEC) reduce 201) ((T_DECLARE) reduce 201) ((T_DIR) reduce 201) ((T_DNUMBER) reduce 201) ((T_DO) reduce 201) ((T_DOUBLE_CAST) reduce 201) ((T_ECHO) reduce 201) ((T_EMPTY) reduce 201) ((T_ENUM) reduce 201) ((T_EVAL) reduce 201) ((T_EXIT) reduce 201) ((T_FILE) reduce 201) ((T_FINAL) reduce 201) ((T_FN) reduce 201) ((T_FOR) reduce 201) ((T_FOREACH) reduce 201) ((T_FUNCTION) reduce 201) ((T_FUNC_C) reduce 201) ((T_GLOBAL) reduce 201) ((T_GOTO) reduce 201) ((T_HALT_COMPILER) reduce 201) ((T_IF) reduce 201) ((T_INC) reduce 201) ((T_INCLUDE) reduce 201) ((T_INCLUDE_ONCE) reduce 201) ((T_INLINE_HTML) reduce 201) ((T_INTERFACE) reduce 201) ((T_INT_CAST) reduce 201) ((T_ISSET) reduce 201) ((T_LINE) reduce 201) ((T_LIST) reduce 201) ((T_LNUMBER) reduce 201) ((T_MATCH) reduce 201) ((T_METHOD_C) reduce 201) ((T_NAMESPACE) reduce 201) ((T_NAME_FULLY_QUALIFIED) reduce 201) ((T_NAME_QUALIFIED) reduce 201) ((T_NAME_RELATIVE) reduce 201) ((T_NEW) reduce 201) ((T_NS_C) reduce 201) ((T_OBJECT_CAST) reduce 201) ((T_PRINT) reduce 201) ((T_PROPERTY_C) reduce 201) ((T_READONLY) reduce 201) ((T_REQUIRE) reduce 201) ((T_REQUIRE_ONCE) reduce 201) ((T_RETURN) reduce 201) ((T_START_HEREDOC) reduce 201) ((T_STATIC) reduce 201) ((T_STRING) reduce 201) ((T_STRING_CAST) reduce 201) ((T_SWITCH) reduce 201) ((T_THROW) reduce 201) ((T_TRAIT) reduce 201) ((T_TRAIT_C) reduce 201) ((T_TRY) reduce 201) ((T_UNSET) reduce 201) ((T_UNSET_CAST) reduce 201) ((T_USE) reduce 201) ((T_VARIABLE) reduce 201) ((T_VOID_CAST) reduce 201) ((T_WHILE) reduce 201) ((T_YIELD) reduce 201) ((T_YIELD_FROM) reduce 201) (("[") reduce 201) (("`") reduce 201) (("{") reduce 201) (("~") reduce 201)) 32788 ((("!") reduce 166) (("\"") reduce 166) (("$") reduce 166) (($) reduce 166) (("(") reduce 166) (("+") reduce 166) (("-") reduce 166) ((";") reduce 166) (("@") reduce 166) ((T_ABSTRACT) reduce 166) ((T_ARRAY) reduce 166) ((T_ARRAY_CAST) reduce 166) ((T_ATTRIBUTE) reduce 166) ((T_BOOL_CAST) reduce 166) ((T_BREAK) reduce 166) ((T_CLASS) reduce 166) ((T_CLASS_C) reduce 166) ((T_CLONE) reduce 166) ((T_CONST) reduce 166) ((T_CONSTANT_ENCAPSED_STRING) reduce 166) ((T_CONTINUE) reduce 166) ((T_DEC) reduce 166) ((T_DECLARE) reduce 166) ((T_DIR) reduce 166) ((T_DNUMBER) reduce 166) ((T_DO) reduce 166) ((T_DOUBLE_CAST) reduce 166) ((T_ECHO) reduce 166) ((T_ELSE) reduce 166) ((T_ELSEIF) reduce 166) ((T_EMPTY) reduce 166) ((T_ENUM) reduce 166) ((T_EVAL) reduce 166) ((T_EXIT) reduce 166) ((T_FILE) reduce 166) ((T_FINAL) reduce 166) ((T_FN) reduce 166) ((T_FOR) reduce 166) ((T_FOREACH) reduce 166) ((T_FUNCTION) reduce 166) ((T_FUNC_C) reduce 166) ((T_GLOBAL) reduce 166) ((T_GOTO) reduce 166) ((T_HALT_COMPILER) reduce 166) ((T_IF) reduce 166) ((T_INC) reduce 166) ((T_INCLUDE) reduce 166) ((T_INCLUDE_ONCE) reduce 166) ((T_INLINE_HTML) reduce 166) ((T_INTERFACE) reduce 166) ((T_INT_CAST) reduce 166) ((T_ISSET) reduce 166) ((T_LINE) reduce 166) ((T_LIST) reduce 166) ((T_LNUMBER) reduce 166) ((T_MATCH) reduce 166) ((T_METHOD_C) reduce 166) ((T_NAMESPACE) reduce 166) ((T_NAME_FULLY_QUALIFIED) reduce 166) ((T_NAME_QUALIFIED) reduce 166) ((T_NAME_RELATIVE) reduce 166) ((T_NEW) reduce 166) ((T_NS_C) reduce 166) ((T_OBJECT_CAST) reduce 166) ((T_PRINT) reduce 166) ((T_PROPERTY_C) reduce 166) ((T_READONLY) reduce 166) ((T_REQUIRE) reduce 166) ((T_REQUIRE_ONCE) reduce 166) ((T_RETURN) reduce 166) ((T_START_HEREDOC) reduce 166) ((T_STATIC) reduce 166) ((T_STRING) reduce 166) ((T_STRING_CAST) reduce 166) ((T_SWITCH) reduce 166) ((T_THROW) reduce 166) ((T_TRAIT) reduce 166) ((T_TRAIT_C) reduce 166) ((T_TRY) reduce 166) ((T_UNSET) reduce 166) ((T_UNSET_CAST) reduce 166) ((T_USE) reduce 166) ((T_VARIABLE) reduce 166) ((T_VOID_CAST) reduce 166) ((T_WHILE) reduce 166) ((T_YIELD) reduce 166) ((T_YIELD_FROM) reduce 166) (("[") reduce 166) (("`") reduce 166) (("{") reduce 166) (("~") reduce 166)) 32799 ((("!") reduce 160) (("\"") reduce 160) (("$") reduce 160) (($) reduce 160) (("(") reduce 160) (("+") reduce 160) (("-") reduce 160) ((";") reduce 160) (("@") reduce 160) ((T_ABSTRACT) reduce 160) ((T_ARRAY) reduce 160) ((T_ARRAY_CAST) reduce 160) ((T_ATTRIBUTE) reduce 160) ((T_BOOL_CAST) reduce 160) ((T_BREAK) reduce 160) ((T_CLASS) reduce 160) ((T_CLASS_C) reduce 160) ((T_CLONE) reduce 160) ((T_CONST) reduce 160) ((T_CONSTANT_ENCAPSED_STRING) reduce 160) ((T_CONTINUE) reduce 160) ((T_DEC) reduce 160) ((T_DECLARE) reduce 160) ((T_DIR) reduce 160) ((T_DNUMBER) reduce 160) ((T_DO) reduce 160) ((T_DOUBLE_CAST) reduce 160) ((T_ECHO) reduce 160) ((T_ELSE) reduce 160) ((T_ELSEIF) reduce 160) ((T_EMPTY) reduce 160) ((T_ENUM) reduce 160) ((T_EVAL) reduce 160) ((T_EXIT) reduce 160) ((T_FILE) reduce 160) ((T_FINAL) reduce 160) ((T_FN) reduce 160) ((T_FOR) reduce 160) ((T_FOREACH) reduce 160) ((T_FUNCTION) reduce 160) ((T_FUNC_C) reduce 160) ((T_GLOBAL) reduce 160) ((T_GOTO) reduce 160) ((T_HALT_COMPILER) reduce 160) ((T_IF) reduce 160) ((T_INC) reduce 160) ((T_INCLUDE) reduce 160) ((T_INCLUDE_ONCE) reduce 160) ((T_INLINE_HTML) reduce 160) ((T_INTERFACE) reduce 160) ((T_INT_CAST) reduce 160) ((T_ISSET) reduce 160) ((T_LINE) reduce 160) ((T_LIST) reduce 160) ((T_LNUMBER) reduce 160) ((T_MATCH) reduce 160) ((T_METHOD_C) reduce 160) ((T_NAMESPACE) reduce 160) ((T_NAME_FULLY_QUALIFIED) reduce 160) ((T_NAME_QUALIFIED) reduce 160) ((T_NAME_RELATIVE) reduce 160) ((T_NEW) reduce 160) ((T_NS_C) reduce 160) ((T_OBJECT_CAST) reduce 160) ((T_PRINT) reduce 160) ((T_PROPERTY_C) reduce 160) ((T_READONLY) reduce 160) ((T_REQUIRE) reduce 160) ((T_REQUIRE_ONCE) reduce 160) ((T_RETURN) reduce 160) ((T_START_HEREDOC) reduce 160) ((T_STATIC) reduce 160) ((T_STRING) reduce 160) ((T_STRING_CAST) reduce 160) ((T_SWITCH) reduce 160) ((T_THROW) reduce 160) ((T_TRAIT) reduce 160) ((T_TRAIT_C) reduce 160) ((T_TRY) reduce 160) ((T_UNSET) reduce 160) ((T_UNSET_CAST) reduce 160) ((T_USE) reduce 160) ((T_VARIABLE) reduce 160) ((T_VOID_CAST) reduce 160) ((T_WHILE) reduce 160) ((T_YIELD) reduce 160) ((T_YIELD_FROM) reduce 160) (("[") reduce 160) (("`") reduce 160) (("{") reduce 160) (("~") reduce 160)) 32808 ((("!") reduce 149) (("\"") reduce 149) (("$") reduce 149) (($) reduce 149) (("(") reduce 149) (("+") reduce 149) (("-") reduce 149) ((";") reduce 149) (("@") reduce 149) ((T_ABSTRACT) reduce 149) ((T_ARRAY) reduce 149) ((T_ARRAY_CAST) reduce 149) ((T_ATTRIBUTE) reduce 149) ((T_BOOL_CAST) reduce 149) ((T_BREAK) reduce 149) ((T_CLASS) reduce 149) ((T_CLASS_C) reduce 149) ((T_CLONE) reduce 149) ((T_CONST) reduce 149) ((T_CONSTANT_ENCAPSED_STRING) reduce 149) ((T_CONTINUE) reduce 149) ((T_DEC) reduce 149) ((T_DECLARE) reduce 149) ((T_DIR) reduce 149) ((T_DNUMBER) reduce 149) ((T_DO) reduce 149) ((T_DOUBLE_CAST) reduce 149) ((T_ECHO) reduce 149) ((T_ELSE) reduce 149) ((T_ELSEIF) reduce 149) ((T_EMPTY) reduce 149) ((T_ENUM) reduce 149) ((T_EVAL) reduce 149) ((T_EXIT) reduce 149) ((T_FILE) reduce 149) ((T_FINAL) reduce 149) ((T_FN) reduce 149) ((T_FOR) reduce 149) ((T_FOREACH) reduce 149) ((T_FUNCTION) reduce 149) ((T_FUNC_C) reduce 149) ((T_GLOBAL) reduce 149) ((T_GOTO) reduce 149) ((T_HALT_COMPILER) reduce 149) ((T_IF) reduce 149) ((T_INC) reduce 149) ((T_INCLUDE) reduce 149) ((T_INCLUDE_ONCE) reduce 149) ((T_INLINE_HTML) reduce 149) ((T_INTERFACE) reduce 149) ((T_INT_CAST) reduce 149) ((T_ISSET) reduce 149) ((T_LINE) reduce 149) ((T_LIST) reduce 149) ((T_LNUMBER) reduce 149) ((T_MATCH) reduce 149) ((T_METHOD_C) reduce 149) ((T_NAMESPACE) reduce 149) ((T_NAME_FULLY_QUALIFIED) reduce 149) ((T_NAME_QUALIFIED) reduce 149) ((T_NAME_RELATIVE) reduce 149) ((T_NEW) reduce 149) ((T_NS_C) reduce 149) ((T_OBJECT_CAST) reduce 149) ((T_PRINT) reduce 149) ((T_PROPERTY_C) reduce 149) ((T_READONLY) reduce 149) ((T_REQUIRE) reduce 149) ((T_REQUIRE_ONCE) reduce 149) ((T_RETURN) reduce 149) ((T_START_HEREDOC) reduce 149) ((T_STATIC) reduce 149) ((T_STRING) reduce 149) ((T_STRING_CAST) reduce 149) ((T_SWITCH) reduce 149) ((T_THROW) reduce 149) ((T_TRAIT) reduce 149) ((T_TRAIT_C) reduce 149) ((T_TRY) reduce 149) ((T_UNSET) reduce 149) ((T_UNSET_CAST) reduce 149) ((T_USE) reduce 149) ((T_VARIABLE) reduce 149) ((T_VOID_CAST) reduce 149) ((T_WHILE) reduce 149) ((T_YIELD) reduce 149) ((T_YIELD_FROM) reduce 149) (("[") reduce 149) (("`") reduce 149) (("{") reduce 149) (("~") reduce 149)) 32811 ((("!") reduce 148) (("\"") reduce 148) (("$") reduce 148) (($) reduce 148) (("(") reduce 148) (("+") reduce 148) (("-") reduce 148) ((";") reduce 148) (("@") reduce 148) ((T_ABSTRACT) reduce 148) ((T_ARRAY) reduce 148) ((T_ARRAY_CAST) reduce 148) ((T_ATTRIBUTE) reduce 148) ((T_BOOL_CAST) reduce 148) ((T_BREAK) reduce 148) ((T_CLASS) reduce 148) ((T_CLASS_C) reduce 148) ((T_CLONE) reduce 148) ((T_CONST) reduce 148) ((T_CONSTANT_ENCAPSED_STRING) reduce 148) ((T_CONTINUE) reduce 148) ((T_DEC) reduce 148) ((T_DECLARE) reduce 148) ((T_DIR) reduce 148) ((T_DNUMBER) reduce 148) ((T_DO) reduce 148) ((T_DOUBLE_CAST) reduce 148) ((T_ECHO) reduce 148) ((T_ELSE) reduce 148) ((T_ELSEIF) reduce 148) ((T_EMPTY) reduce 148) ((T_ENUM) reduce 148) ((T_EVAL) reduce 148) ((T_EXIT) reduce 148) ((T_FILE) reduce 148) ((T_FINAL) reduce 148) ((T_FN) reduce 148) ((T_FOR) reduce 148) ((T_FOREACH) reduce 148) ((T_FUNCTION) reduce 148) ((T_FUNC_C) reduce 148) ((T_GLOBAL) reduce 148) ((T_GOTO) reduce 148) ((T_HALT_COMPILER) reduce 148) ((T_IF) reduce 148) ((T_INC) reduce 148) ((T_INCLUDE) reduce 148) ((T_INCLUDE_ONCE) reduce 148) ((T_INLINE_HTML) reduce 148) ((T_INTERFACE) reduce 148) ((T_INT_CAST) reduce 148) ((T_ISSET) reduce 148) ((T_LINE) reduce 148) ((T_LIST) reduce 148) ((T_LNUMBER) reduce 148) ((T_MATCH) reduce 148) ((T_METHOD_C) reduce 148) ((T_NAMESPACE) reduce 148) ((T_NAME_FULLY_QUALIFIED) reduce 148) ((T_NAME_QUALIFIED) reduce 148) ((T_NAME_RELATIVE) reduce 148) ((T_NEW) reduce 148) ((T_NS_C) reduce 148) ((T_OBJECT_CAST) reduce 148) ((T_PRINT) reduce 148) ((T_PROPERTY_C) reduce 148) ((T_READONLY) reduce 148) ((T_REQUIRE) reduce 148) ((T_REQUIRE_ONCE) reduce 148) ((T_RETURN) reduce 148) ((T_START_HEREDOC) reduce 148) ((T_STATIC) reduce 148) ((T_STRING) reduce 148) ((T_STRING_CAST) reduce 148) ((T_SWITCH) reduce 148) ((T_THROW) reduce 148) ((T_TRAIT) reduce 148) ((T_TRAIT_C) reduce 148) ((T_TRY) reduce 148) ((T_UNSET) reduce 148) ((T_UNSET_CAST) reduce 148) ((T_USE) reduce 148) ((T_VARIABLE) reduce 148) ((T_VOID_CAST) reduce 148) ((T_WHILE) reduce 148) ((T_YIELD) reduce 148) ((T_YIELD_FROM) reduce 148) (("[") reduce 148) (("`") reduce 148) (("{") reduce 148) (("~") reduce 148)) 32812 ((("!") reduce 246) (("\"") reduce 246) (($) reduce 246) (("$") reduce 246) (("(") reduce 246) (("+") reduce 246) (("-") reduce 246) ((";") reduce 246) (("@") reduce 246) ((T_ABSTRACT) reduce 246) ((T_ARRAY) reduce 246) ((T_ARRAY_CAST) reduce 246) ((T_ATTRIBUTE) reduce 246) ((T_BOOL_CAST) reduce 246) ((T_BREAK) reduce 246) ((T_CLASS) reduce 246) ((T_CLASS_C) reduce 246) ((T_CLONE) reduce 246) ((T_CONST) reduce 246) ((T_CONSTANT_ENCAPSED_STRING) reduce 246) ((T_CONTINUE) reduce 246) ((T_DEC) reduce 246) ((T_DECLARE) reduce 246) ((T_DIR) reduce 246) ((T_DNUMBER) reduce 246) ((T_DO) reduce 246) ((T_DOUBLE_CAST) reduce 246) ((T_ECHO) reduce 246) ((T_ELSE) shift) ((T_ELSEIF) shift) ((T_EMPTY) reduce 246) ((T_ENUM) reduce 246) ((T_EVAL) reduce 246) ((T_EXIT) reduce 246) ((T_FILE) reduce 246) ((T_FINAL) reduce 246) ((T_FN) reduce 246) ((T_FOR) reduce 246) ((T_FOREACH) reduce 246) ((T_FUNCTION) reduce 246) ((T_FUNC_C) reduce 246) ((T_GLOBAL) reduce 246) ((T_GOTO) reduce 246) ((T_HALT_COMPILER) reduce 246) ((T_IF) reduce 246) ((T_INC) reduce 246) ((T_INCLUDE) reduce 246) ((T_INCLUDE_ONCE) reduce 246) ((T_INLINE_HTML) reduce 246) ((T_INTERFACE) reduce 246) ((T_INT_CAST) reduce 246) ((T_ISSET) reduce 246) ((T_LINE) reduce 246) ((T_LIST) reduce 246) ((T_LNUMBER) reduce 246) ((T_MATCH) reduce 246) ((T_METHOD_C) reduce 246) ((T_NAMESPACE) reduce 246) ((T_NAME_FULLY_QUALIFIED) reduce 246) ((T_NAME_QUALIFIED) reduce 246) ((T_NAME_RELATIVE) reduce 246) ((T_NEW) reduce 246) ((T_NS_C) reduce 246) ((T_OBJECT_CAST) reduce 246) ((T_PRINT) reduce 246) ((T_PROPERTY_C) reduce 246) ((T_READONLY) reduce 246) ((T_REQUIRE) reduce 246) ((T_REQUIRE_ONCE) reduce 246) ((T_RETURN) reduce 246) ((T_START_HEREDOC) reduce 246) ((T_STATIC) reduce 246) ((T_STRING) reduce 246) ((T_STRING_CAST) reduce 246) ((T_SWITCH) reduce 246) ((T_THROW) reduce 246) ((T_TRAIT) reduce 246) ((T_TRAIT_C) reduce 246) ((T_TRY) reduce 246) ((T_UNSET) reduce 246) ((T_UNSET_CAST) reduce 246) ((T_USE) reduce 246) ((T_VARIABLE) reduce 246) ((T_VOID_CAST) reduce 246) ((T_WHILE) reduce 246) ((T_YIELD) reduce 246) ((T_YIELD_FROM) reduce 246) (("[") reduce 246) (("`") reduce 246) (("{") reduce 246) (("~") reduce 246)) 32813 ((("!") reduce 244) (("\"") reduce 244) (($) reduce 244) (("$") reduce 244) (("(") reduce 244) (("+") reduce 244) (("-") reduce 244) ((";") reduce 244) (("@") reduce 244) ((T_ABSTRACT) reduce 244) ((T_ARRAY) reduce 244) ((T_ARRAY_CAST) reduce 244) ((T_ATTRIBUTE) reduce 244) ((T_BOOL_CAST) reduce 244) ((T_BREAK) reduce 244) ((T_CLASS) reduce 244) ((T_CLASS_C) reduce 244) ((T_CLONE) reduce 244) ((T_CONST) reduce 244) ((T_CONSTANT_ENCAPSED_STRING) reduce 244) ((T_CONTINUE) reduce 244) ((T_DEC) reduce 244) ((T_DECLARE) reduce 244) ((T_DIR) reduce 244) ((T_DNUMBER) reduce 244) ((T_DO) reduce 244) ((T_DOUBLE_CAST) reduce 244) ((T_ECHO) reduce 244) ((T_ELSE) reduce 244) ((T_ELSEIF) reduce 244) ((T_EMPTY) reduce 244) ((T_ENUM) reduce 244) ((T_EVAL) reduce 244) ((T_EXIT) reduce 244) ((T_FILE) reduce 244) ((T_FINAL) reduce 244) ((T_FN) reduce 244) ((T_FOR) reduce 244) ((T_FOREACH) reduce 244) ((T_FUNCTION) reduce 244) ((T_FUNC_C) reduce 244) ((T_GLOBAL) reduce 244) ((T_GOTO) reduce 244) ((T_HALT_COMPILER) reduce 244) ((T_IF) reduce 244) ((T_INC) reduce 244) ((T_INCLUDE) reduce 244) ((T_INCLUDE_ONCE) reduce 244) ((T_INLINE_HTML) reduce 244) ((T_INTERFACE) reduce 244) ((T_INT_CAST) reduce 244) ((T_ISSET) reduce 244) ((T_LINE) reduce 244) ((T_LIST) reduce 244) ((T_LNUMBER) reduce 244) ((T_MATCH) reduce 244) ((T_METHOD_C) reduce 244) ((T_NAMESPACE) reduce 244) ((T_NAME_FULLY_QUALIFIED) reduce 244) ((T_NAME_QUALIFIED) reduce 244) ((T_NAME_RELATIVE) reduce 244) ((T_NEW) reduce 244) ((T_NS_C) reduce 244) ((T_OBJECT_CAST) reduce 244) ((T_PRINT) reduce 244) ((T_PROPERTY_C) reduce 244) ((T_READONLY) reduce 244) ((T_REQUIRE) reduce 244) ((T_REQUIRE_ONCE) reduce 244) ((T_RETURN) reduce 244) ((T_START_HEREDOC) reduce 244) ((T_STATIC) reduce 244) ((T_STRING) reduce 244) ((T_STRING_CAST) reduce 244) ((T_SWITCH) reduce 244) ((T_THROW) reduce 244) ((T_TRAIT) reduce 244) ((T_TRAIT_C) reduce 244) ((T_TRY) reduce 244) ((T_UNSET) reduce 244) ((T_UNSET_CAST) reduce 244) ((T_USE) reduce 244) ((T_VARIABLE) reduce 244) ((T_VOID_CAST) reduce 244) ((T_WHILE) reduce 244) ((T_YIELD) reduce 244) ((T_YIELD_FROM) reduce 244) (("[") reduce 244) (("`") reduce 244) (("{") reduce 244) (("~") reduce 244)) 32816 ((("!") reduce 147) (("\"") reduce 147) (("$") reduce 147) (($) reduce 147) (("(") reduce 147) (("+") reduce 147) (("-") reduce 147) ((";") reduce 147) (("@") reduce 147) ((T_ABSTRACT) reduce 147) ((T_ARRAY) reduce 147) ((T_ARRAY_CAST) reduce 147) ((T_ATTRIBUTE) reduce 147) ((T_BOOL_CAST) reduce 147) ((T_BREAK) reduce 147) ((T_CLASS) reduce 147) ((T_CLASS_C) reduce 147) ((T_CLONE) reduce 147) ((T_CONST) reduce 147) ((T_CONSTANT_ENCAPSED_STRING) reduce 147) ((T_CONTINUE) reduce 147) ((T_DEC) reduce 147) ((T_DECLARE) reduce 147) ((T_DIR) reduce 147) ((T_DNUMBER) reduce 147) ((T_DO) reduce 147) ((T_DOUBLE_CAST) reduce 147) ((T_ECHO) reduce 147) ((T_ELSE) reduce 147) ((T_ELSEIF) reduce 147) ((T_EMPTY) reduce 147) ((T_ENUM) reduce 147) ((T_EVAL) reduce 147) ((T_EXIT) reduce 147) ((T_FILE) reduce 147) ((T_FINAL) reduce 147) ((T_FN) reduce 147) ((T_FOR) reduce 147) ((T_FOREACH) reduce 147) ((T_FUNCTION) reduce 147) ((T_FUNC_C) reduce 147) ((T_GLOBAL) reduce 147) ((T_GOTO) reduce 147) ((T_HALT_COMPILER) reduce 147) ((T_IF) reduce 147) ((T_INC) reduce 147) ((T_INCLUDE) reduce 147) ((T_INCLUDE_ONCE) reduce 147) ((T_INLINE_HTML) reduce 147) ((T_INTERFACE) reduce 147) ((T_INT_CAST) reduce 147) ((T_ISSET) reduce 147) ((T_LINE) reduce 147) ((T_LIST) reduce 147) ((T_LNUMBER) reduce 147) ((T_MATCH) reduce 147) ((T_METHOD_C) reduce 147) ((T_NAMESPACE) reduce 147) ((T_NAME_FULLY_QUALIFIED) reduce 147) ((T_NAME_QUALIFIED) reduce 147) ((T_NAME_RELATIVE) reduce 147) ((T_NEW) reduce 147) ((T_NS_C) reduce 147) ((T_OBJECT_CAST) reduce 147) ((T_PRINT) reduce 147) ((T_PROPERTY_C) reduce 147) ((T_READONLY) reduce 147) ((T_REQUIRE) reduce 147) ((T_REQUIRE_ONCE) reduce 147) ((T_RETURN) reduce 147) ((T_START_HEREDOC) reduce 147) ((T_STATIC) reduce 147) ((T_STRING) reduce 147) ((T_STRING_CAST) reduce 147) ((T_SWITCH) reduce 147) ((T_THROW) reduce 147) ((T_TRAIT) reduce 147) ((T_TRAIT_C) reduce 147) ((T_TRY) reduce 147) ((T_UNSET) reduce 147) ((T_UNSET_CAST) reduce 147) ((T_USE) reduce 147) ((T_VARIABLE) reduce 147) ((T_VOID_CAST) reduce 147) ((T_WHILE) reduce 147) ((T_YIELD) reduce 147) ((T_YIELD_FROM) reduce 147) (("[") reduce 147) (("`") reduce 147) (("{") reduce 147) (("~") reduce 147)) 32822 ((("!") reduce 245) (("\"") reduce 245) (($) reduce 245) (("$") reduce 245) (("(") reduce 245) (("+") reduce 245) (("-") reduce 245) ((";") reduce 245) (("@") reduce 245) ((T_ABSTRACT) reduce 245) ((T_ARRAY) reduce 245) ((T_ARRAY_CAST) reduce 245) ((T_ATTRIBUTE) reduce 245) ((T_BOOL_CAST) reduce 245) ((T_BREAK) reduce 245) ((T_CLASS) reduce 245) ((T_CLASS_C) reduce 245) ((T_CLONE) reduce 245) ((T_CONST) reduce 245) ((T_CONSTANT_ENCAPSED_STRING) reduce 245) ((T_CONTINUE) reduce 245) ((T_DEC) reduce 245) ((T_DECLARE) reduce 245) ((T_DIR) reduce 245) ((T_DNUMBER) reduce 245) ((T_DO) reduce 245) ((T_DOUBLE_CAST) reduce 245) ((T_ECHO) reduce 245) ((T_ELSE) reduce 245) ((T_ELSEIF) reduce 245) ((T_EMPTY) reduce 245) ((T_ENUM) reduce 245) ((T_EVAL) reduce 245) ((T_EXIT) reduce 245) ((T_FILE) reduce 245) ((T_FINAL) reduce 245) ((T_FN) reduce 245) ((T_FOR) reduce 245) ((T_FOREACH) reduce 245) ((T_FUNCTION) reduce 245) ((T_FUNC_C) reduce 245) ((T_GLOBAL) reduce 245) ((T_GOTO) reduce 245) ((T_HALT_COMPILER) reduce 245) ((T_IF) reduce 245) ((T_INC) reduce 245) ((T_INCLUDE) reduce 245) ((T_INCLUDE_ONCE) reduce 245) ((T_INLINE_HTML) reduce 245) ((T_INTERFACE) reduce 245) ((T_INT_CAST) reduce 245) ((T_ISSET) reduce 245) ((T_LINE) reduce 245) ((T_LIST) reduce 245) ((T_LNUMBER) reduce 245) ((T_MATCH) reduce 245) ((T_METHOD_C) reduce 245) ((T_NAMESPACE) reduce 245) ((T_NAME_FULLY_QUALIFIED) reduce 245) ((T_NAME_QUALIFIED) reduce 245) ((T_NAME_RELATIVE) reduce 245) ((T_NEW) reduce 245) ((T_NS_C) reduce 245) ((T_OBJECT_CAST) reduce 245) ((T_PRINT) reduce 245) ((T_PROPERTY_C) reduce 245) ((T_READONLY) reduce 245) ((T_REQUIRE) reduce 245) ((T_REQUIRE_ONCE) reduce 245) ((T_RETURN) reduce 245) ((T_START_HEREDOC) reduce 245) ((T_STATIC) reduce 245) ((T_STRING) reduce 245) ((T_STRING_CAST) reduce 245) ((T_SWITCH) reduce 245) ((T_THROW) reduce 245) ((T_TRAIT) reduce 245) ((T_TRAIT_C) reduce 245) ((T_TRY) reduce 245) ((T_UNSET) reduce 245) ((T_UNSET_CAST) reduce 245) ((T_USE) reduce 245) ((T_VARIABLE) reduce 245) ((T_VOID_CAST) reduce 245) ((T_WHILE) reduce 245) ((T_YIELD) reduce 245) ((T_YIELD_FROM) reduce 245) (("[") reduce 245) (("`") reduce 245) (("{") reduce 245) (("~") reduce 245)) 32823 ((("!") reduce 247) (("\"") reduce 247) (($) reduce 247) (("$") reduce 247) (("(") reduce 247) (("+") reduce 247) (("-") reduce 247) ((";") reduce 247) (("@") reduce 247) ((T_ABSTRACT) reduce 247) ((T_ARRAY) reduce 247) ((T_ARRAY_CAST) reduce 247) ((T_ATTRIBUTE) reduce 247) ((T_BOOL_CAST) reduce 247) ((T_BREAK) reduce 247) ((T_CLASS) reduce 247) ((T_CLASS_C) reduce 247) ((T_CLONE) reduce 247) ((T_CONST) reduce 247) ((T_CONSTANT_ENCAPSED_STRING) reduce 247) ((T_CONTINUE) reduce 247) ((T_DEC) reduce 247) ((T_DECLARE) reduce 247) ((T_DIR) reduce 247) ((T_DNUMBER) reduce 247) ((T_DO) reduce 247) ((T_DOUBLE_CAST) reduce 247) ((T_ECHO) reduce 247) ((T_ELSE) reduce 247) ((T_ELSEIF) reduce 247) ((T_EMPTY) reduce 247) ((T_ENUM) reduce 247) ((T_EVAL) reduce 247) ((T_EXIT) reduce 247) ((T_FILE) reduce 247) ((T_FINAL) reduce 247) ((T_FN) reduce 247) ((T_FOR) reduce 247) ((T_FOREACH) reduce 247) ((T_FUNCTION) reduce 247) ((T_FUNC_C) reduce 247) ((T_GLOBAL) reduce 247) ((T_GOTO) reduce 247) ((T_HALT_COMPILER) reduce 247) ((T_IF) reduce 247) ((T_INC) reduce 247) ((T_INCLUDE) reduce 247) ((T_INCLUDE_ONCE) reduce 247) ((T_INLINE_HTML) reduce 247) ((T_INTERFACE) reduce 247) ((T_INT_CAST) reduce 247) ((T_ISSET) reduce 247) ((T_LINE) reduce 247) ((T_LIST) reduce 247) ((T_LNUMBER) reduce 247) ((T_MATCH) reduce 247) ((T_METHOD_C) reduce 247) ((T_NAMESPACE) reduce 247) ((T_NAME_FULLY_QUALIFIED) reduce 247) ((T_NAME_QUALIFIED) reduce 247) ((T_NAME_RELATIVE) reduce 247) ((T_NEW) reduce 247) ((T_NS_C) reduce 247) ((T_OBJECT_CAST) reduce 247) ((T_PRINT) reduce 247) ((T_PROPERTY_C) reduce 247) ((T_READONLY) reduce 247) ((T_REQUIRE) reduce 247) ((T_REQUIRE_ONCE) reduce 247) ((T_RETURN) reduce 247) ((T_START_HEREDOC) reduce 247) ((T_STATIC) reduce 247) ((T_STRING) reduce 247) ((T_STRING_CAST) reduce 247) ((T_SWITCH) reduce 247) ((T_THROW) reduce 247) ((T_TRAIT) reduce 247) ((T_TRAIT_C) reduce 247) ((T_TRY) reduce 247) ((T_UNSET) reduce 247) ((T_UNSET_CAST) reduce 247) ((T_USE) reduce 247) ((T_VARIABLE) reduce 247) ((T_VOID_CAST) reduce 247) ((T_WHILE) reduce 247) ((T_YIELD) reduce 247) ((T_YIELD_FROM) reduce 247) (("[") reduce 247) (("`") reduce 247) (("{") reduce 247) (("~") reduce 247)) 32824 ((("!") reduce 161) (("\"") reduce 161) (($) reduce 161) (("$") reduce 161) (("(") reduce 161) (("+") reduce 161) (("-") reduce 161) ((";") reduce 161) (("@") reduce 161) ((T_ABSTRACT) reduce 161) ((T_ARRAY) reduce 161) ((T_ARRAY_CAST) reduce 161) ((T_ATTRIBUTE) reduce 161) ((T_BOOL_CAST) reduce 161) ((T_BREAK) reduce 161) ((T_CLASS) reduce 161) ((T_CLASS_C) reduce 161) ((T_CLONE) reduce 161) ((T_CONST) reduce 161) ((T_CONSTANT_ENCAPSED_STRING) reduce 161) ((T_CONTINUE) reduce 161) ((T_DEC) reduce 161) ((T_DECLARE) reduce 161) ((T_DIR) reduce 161) ((T_DNUMBER) reduce 161) ((T_DO) reduce 161) ((T_DOUBLE_CAST) reduce 161) ((T_ECHO) reduce 161) ((T_ELSE) reduce 161) ((T_ELSEIF) reduce 161) ((T_EMPTY) reduce 161) ((T_ENUM) reduce 161) ((T_EVAL) reduce 161) ((T_EXIT) reduce 161) ((T_FILE) reduce 161) ((T_FINAL) reduce 161) ((T_FN) reduce 161) ((T_FOR) reduce 161) ((T_FOREACH) reduce 161) ((T_FUNCTION) reduce 161) ((T_FUNC_C) reduce 161) ((T_GLOBAL) reduce 161) ((T_GOTO) reduce 161) ((T_HALT_COMPILER) reduce 161) ((T_IF) reduce 161) ((T_INC) reduce 161) ((T_INCLUDE) reduce 161) ((T_INCLUDE_ONCE) reduce 161) ((T_INLINE_HTML) reduce 161) ((T_INTERFACE) reduce 161) ((T_INT_CAST) reduce 161) ((T_ISSET) reduce 161) ((T_LINE) reduce 161) ((T_LIST) reduce 161) ((T_LNUMBER) reduce 161) ((T_MATCH) reduce 161) ((T_METHOD_C) reduce 161) ((T_NAMESPACE) reduce 161) ((T_NAME_FULLY_QUALIFIED) reduce 161) ((T_NAME_QUALIFIED) reduce 161) ((T_NAME_RELATIVE) reduce 161) ((T_NEW) reduce 161) ((T_NS_C) reduce 161) ((T_OBJECT_CAST) reduce 161) ((T_PRINT) reduce 161) ((T_PROPERTY_C) reduce 161) ((T_READONLY) reduce 161) ((T_REQUIRE) reduce 161) ((T_REQUIRE_ONCE) reduce 161) ((T_RETURN) reduce 161) ((T_START_HEREDOC) reduce 161) ((T_STATIC) reduce 161) ((T_STRING) reduce 161) ((T_STRING_CAST) reduce 161) ((T_SWITCH) reduce 161) ((T_THROW) reduce 161) ((T_TRAIT) reduce 161) ((T_TRAIT_C) reduce 161) ((T_TRY) reduce 161) ((T_UNSET) reduce 161) ((T_UNSET_CAST) reduce 161) ((T_USE) reduce 161) ((T_VARIABLE) reduce 161) ((T_VOID_CAST) reduce 161) ((T_WHILE) reduce 161) ((T_YIELD) reduce 161) ((T_YIELD_FROM) reduce 161) (("[") reduce 161) (("`") reduce 161) (("{") reduce 161) (("~") reduce 161)) 32827 ((("!") reduce 250) (("\"") reduce 250) (($) reduce 250) (("$") reduce 250) (("(") reduce 250) (("+") reduce 250) (("-") reduce 250) ((";") reduce 250) (("@") reduce 250) ((T_ABSTRACT) reduce 250) ((T_ARRAY) reduce 250) ((T_ARRAY_CAST) reduce 250) ((T_ATTRIBUTE) reduce 250) ((T_BOOL_CAST) reduce 250) ((T_BREAK) reduce 250) ((T_CLASS) reduce 250) ((T_CLASS_C) reduce 250) ((T_CLONE) reduce 250) ((T_CONST) reduce 250) ((T_CONSTANT_ENCAPSED_STRING) reduce 250) ((T_CONTINUE) reduce 250) ((T_DEC) reduce 250) ((T_DECLARE) reduce 250) ((T_DIR) reduce 250) ((T_DNUMBER) reduce 250) ((T_DO) reduce 250) ((T_DOUBLE_CAST) reduce 250) ((T_ECHO) reduce 250) ((T_ELSE) reduce 250) ((T_ELSEIF) reduce 250) ((T_EMPTY) reduce 250) ((T_ENUM) reduce 250) ((T_EVAL) reduce 250) ((T_EXIT) reduce 250) ((T_FILE) reduce 250) ((T_FINAL) reduce 250) ((T_FN) reduce 250) ((T_FOR) reduce 250) ((T_FOREACH) reduce 250) ((T_FUNCTION) reduce 250) ((T_FUNC_C) reduce 250) ((T_GLOBAL) reduce 250) ((T_GOTO) reduce 250) ((T_HALT_COMPILER) reduce 250) ((T_IF) reduce 250) ((T_INC) reduce 250) ((T_INCLUDE) reduce 250) ((T_INCLUDE_ONCE) reduce 250) ((T_INLINE_HTML) reduce 250) ((T_INTERFACE) reduce 250) ((T_INT_CAST) reduce 250) ((T_ISSET) reduce 250) ((T_LINE) reduce 250) ((T_LIST) reduce 250) ((T_LNUMBER) reduce 250) ((T_MATCH) reduce 250) ((T_METHOD_C) reduce 250) ((T_NAMESPACE) reduce 250) ((T_NAME_FULLY_QUALIFIED) reduce 250) ((T_NAME_QUALIFIED) reduce 250) ((T_NAME_RELATIVE) reduce 250) ((T_NEW) reduce 250) ((T_NS_C) reduce 250) ((T_OBJECT_CAST) reduce 250) ((T_PRINT) reduce 250) ((T_PROPERTY_C) reduce 250) ((T_READONLY) reduce 250) ((T_REQUIRE) reduce 250) ((T_REQUIRE_ONCE) reduce 250) ((T_RETURN) reduce 250) ((T_START_HEREDOC) reduce 250) ((T_STATIC) reduce 250) ((T_STRING) reduce 250) ((T_STRING_CAST) reduce 250) ((T_SWITCH) reduce 250) ((T_THROW) reduce 250) ((T_TRAIT) reduce 250) ((T_TRAIT_C) reduce 250) ((T_TRY) reduce 250) ((T_UNSET) reduce 250) ((T_UNSET_CAST) reduce 250) ((T_USE) reduce 250) ((T_VARIABLE) reduce 250) ((T_VOID_CAST) reduce 250) ((T_WHILE) reduce 250) ((T_YIELD) reduce 250) ((T_YIELD_FROM) reduce 250) (("[") reduce 250) (("`") reduce 250) (("{") reduce 250) (("~") reduce 250)) 32831 ((("!") reduce 251) (("\"") reduce 251) (("$") reduce 251) (($) reduce 251) (("(") reduce 251) (("+") reduce 251) (("-") reduce 251) ((";") reduce 251) (("@") reduce 251) ((T_ABSTRACT) reduce 251) ((T_ARRAY) reduce 251) ((T_ARRAY_CAST) reduce 251) ((T_ATTRIBUTE) reduce 251) ((T_BOOL_CAST) reduce 251) ((T_BREAK) reduce 251) ((T_CLASS) reduce 251) ((T_CLASS_C) reduce 251) ((T_CLONE) reduce 251) ((T_CONST) reduce 251) ((T_CONSTANT_ENCAPSED_STRING) reduce 251) ((T_CONTINUE) reduce 251) ((T_DEC) reduce 251) ((T_DECLARE) reduce 251) ((T_DIR) reduce 251) ((T_DNUMBER) reduce 251) ((T_DO) reduce 251) ((T_DOUBLE_CAST) reduce 251) ((T_ECHO) reduce 251) ((T_ELSE) reduce 251) ((T_ELSEIF) reduce 251) ((T_EMPTY) reduce 251) ((T_ENUM) reduce 251) ((T_EVAL) reduce 251) ((T_EXIT) reduce 251) ((T_FILE) reduce 251) ((T_FINAL) reduce 251) ((T_FN) reduce 251) ((T_FOR) reduce 251) ((T_FOREACH) reduce 251) ((T_FUNCTION) reduce 251) ((T_FUNC_C) reduce 251) ((T_GLOBAL) reduce 251) ((T_GOTO) reduce 251) ((T_HALT_COMPILER) reduce 251) ((T_IF) reduce 251) ((T_INC) reduce 251) ((T_INCLUDE) reduce 251) ((T_INCLUDE_ONCE) reduce 251) ((T_INLINE_HTML) reduce 251) ((T_INTERFACE) reduce 251) ((T_INT_CAST) reduce 251) ((T_ISSET) reduce 251) ((T_LINE) reduce 251) ((T_LIST) reduce 251) ((T_LNUMBER) reduce 251) ((T_MATCH) reduce 251) ((T_METHOD_C) reduce 251) ((T_NAMESPACE) reduce 251) ((T_NAME_FULLY_QUALIFIED) reduce 251) ((T_NAME_QUALIFIED) reduce 251) ((T_NAME_RELATIVE) reduce 251) ((T_NEW) reduce 251) ((T_NS_C) reduce 251) ((T_OBJECT_CAST) reduce 251) ((T_PRINT) reduce 251) ((T_PROPERTY_C) reduce 251) ((T_READONLY) reduce 251) ((T_REQUIRE) reduce 251) ((T_REQUIRE_ONCE) reduce 251) ((T_RETURN) reduce 251) ((T_START_HEREDOC) reduce 251) ((T_STATIC) reduce 251) ((T_STRING) reduce 251) ((T_STRING_CAST) reduce 251) ((T_SWITCH) reduce 251) ((T_THROW) reduce 251) ((T_TRAIT) reduce 251) ((T_TRAIT_C) reduce 251) ((T_TRY) reduce 251) ((T_UNSET) reduce 251) ((T_UNSET_CAST) reduce 251) ((T_USE) reduce 251) ((T_VARIABLE) reduce 251) ((T_VOID_CAST) reduce 251) ((T_WHILE) reduce 251) ((T_YIELD) reduce 251) ((T_YIELD_FROM) reduce 251) (("[") reduce 251) (("`") reduce 251) (("{") reduce 251) (("~") reduce 251)) 32836 ((("!") reduce 242) (("\"") reduce 242) (($) reduce 242) (("$") reduce 242) (("(") reduce 242) (("+") reduce 242) (("-") reduce 242) ((";") reduce 242) (("@") reduce 242) ((T_ABSTRACT) reduce 242) ((T_ARRAY) reduce 242) ((T_ARRAY_CAST) reduce 242) ((T_ATTRIBUTE) reduce 242) ((T_BOOL_CAST) reduce 242) ((T_BREAK) reduce 242) ((T_CLASS) reduce 242) ((T_CLASS_C) reduce 242) ((T_CLONE) reduce 242) ((T_CONST) reduce 242) ((T_CONSTANT_ENCAPSED_STRING) reduce 242) ((T_CONTINUE) reduce 242) ((T_DEC) reduce 242) ((T_DECLARE) reduce 242) ((T_DIR) reduce 242) ((T_DNUMBER) reduce 242) ((T_DO) reduce 242) ((T_DOUBLE_CAST) reduce 242) ((T_ECHO) reduce 242) ((T_ELSE) reduce 242) ((T_ELSEIF) reduce 242) ((T_EMPTY) reduce 242) ((T_ENUM) reduce 242) ((T_EVAL) reduce 242) ((T_EXIT) reduce 242) ((T_FILE) reduce 242) ((T_FINAL) reduce 242) ((T_FN) reduce 242) ((T_FOR) reduce 242) ((T_FOREACH) reduce 242) ((T_FUNCTION) reduce 242) ((T_FUNC_C) reduce 242) ((T_GLOBAL) reduce 242) ((T_GOTO) reduce 242) ((T_HALT_COMPILER) reduce 242) ((T_IF) reduce 242) ((T_INC) reduce 242) ((T_INCLUDE) reduce 242) ((T_INCLUDE_ONCE) reduce 242) ((T_INLINE_HTML) reduce 242) ((T_INTERFACE) reduce 242) ((T_INT_CAST) reduce 242) ((T_ISSET) reduce 242) ((T_LINE) reduce 242) ((T_LIST) reduce 242) ((T_LNUMBER) reduce 242) ((T_MATCH) reduce 242) ((T_METHOD_C) reduce 242) ((T_NAMESPACE) reduce 242) ((T_NAME_FULLY_QUALIFIED) reduce 242) ((T_NAME_QUALIFIED) reduce 242) ((T_NAME_RELATIVE) reduce 242) ((T_NEW) reduce 242) ((T_NS_C) reduce 242) ((T_OBJECT_CAST) reduce 242) ((T_PRINT) reduce 242) ((T_PROPERTY_C) reduce 242) ((T_READONLY) reduce 242) ((T_REQUIRE) reduce 242) ((T_REQUIRE_ONCE) reduce 242) ((T_RETURN) reduce 242) ((T_START_HEREDOC) reduce 242) ((T_STATIC) reduce 242) ((T_STRING) reduce 242) ((T_STRING_CAST) reduce 242) ((T_SWITCH) reduce 242) ((T_THROW) reduce 242) ((T_TRAIT) reduce 242) ((T_TRAIT_C) reduce 242) ((T_TRY) reduce 242) ((T_UNSET) reduce 242) ((T_UNSET_CAST) reduce 242) ((T_USE) reduce 242) ((T_VARIABLE) reduce 242) ((T_VOID_CAST) reduce 242) ((T_WHILE) reduce 242) ((T_YIELD) reduce 242) ((T_YIELD_FROM) reduce 242) (("[") reduce 242) (("`") reduce 242) (("{") reduce 242) (("~") reduce 242)) 32837 ((("!") reduce 150) (("\"") reduce 150) (("$") reduce 150) (($) reduce 150) (("(") reduce 150) (("+") reduce 150) (("-") reduce 150) ((";") reduce 150) (("@") reduce 150) ((T_ABSTRACT) reduce 150) ((T_ARRAY) reduce 150) ((T_ARRAY_CAST) reduce 150) ((T_ATTRIBUTE) reduce 150) ((T_BOOL_CAST) reduce 150) ((T_BREAK) reduce 150) ((T_CLASS) reduce 150) ((T_CLASS_C) reduce 150) ((T_CLONE) reduce 150) ((T_CONST) reduce 150) ((T_CONSTANT_ENCAPSED_STRING) reduce 150) ((T_CONTINUE) reduce 150) ((T_DEC) reduce 150) ((T_DECLARE) reduce 150) ((T_DIR) reduce 150) ((T_DNUMBER) reduce 150) ((T_DO) reduce 150) ((T_DOUBLE_CAST) reduce 150) ((T_ECHO) reduce 150) ((T_ELSE) reduce 150) ((T_ELSEIF) reduce 150) ((T_EMPTY) reduce 150) ((T_ENUM) reduce 150) ((T_EVAL) reduce 150) ((T_EXIT) reduce 150) ((T_FILE) reduce 150) ((T_FINAL) reduce 150) ((T_FN) reduce 150) ((T_FOR) reduce 150) ((T_FOREACH) reduce 150) ((T_FUNCTION) reduce 150) ((T_FUNC_C) reduce 150) ((T_GLOBAL) reduce 150) ((T_GOTO) reduce 150) ((T_HALT_COMPILER) reduce 150) ((T_IF) reduce 150) ((T_INC) reduce 150) ((T_INCLUDE) reduce 150) ((T_INCLUDE_ONCE) reduce 150) ((T_INLINE_HTML) reduce 150) ((T_INTERFACE) reduce 150) ((T_INT_CAST) reduce 150) ((T_ISSET) reduce 150) ((T_LINE) reduce 150) ((T_LIST) reduce 150) ((T_LNUMBER) reduce 150) ((T_MATCH) reduce 150) ((T_METHOD_C) reduce 150) ((T_NAMESPACE) reduce 150) ((T_NAME_FULLY_QUALIFIED) reduce 150) ((T_NAME_QUALIFIED) reduce 150) ((T_NAME_RELATIVE) reduce 150) ((T_NEW) reduce 150) ((T_NS_C) reduce 150) ((T_OBJECT_CAST) reduce 150) ((T_PRINT) reduce 150) ((T_PROPERTY_C) reduce 150) ((T_READONLY) reduce 150) ((T_REQUIRE) reduce 150) ((T_REQUIRE_ONCE) reduce 150) ((T_RETURN) reduce 150) ((T_START_HEREDOC) reduce 150) ((T_STATIC) reduce 150) ((T_STRING) reduce 150) ((T_STRING_CAST) reduce 150) ((T_SWITCH) reduce 150) ((T_THROW) reduce 150) ((T_TRAIT) reduce 150) ((T_TRAIT_C) reduce 150) ((T_TRY) reduce 150) ((T_UNSET) reduce 150) ((T_UNSET_CAST) reduce 150) ((T_USE) reduce 150) ((T_VARIABLE) reduce 150) ((T_VOID_CAST) reduce 150) ((T_WHILE) reduce 150) ((T_YIELD) reduce 150) ((T_YIELD_FROM) reduce 150) (("[") reduce 150) (("`") reduce 150) (("{") reduce 150) (("~") reduce 150)) 32840 ((("!") reduce 243) (("\"") reduce 243) (("$") reduce 243) (($) reduce 243) (("(") reduce 243) (("+") reduce 243) (("-") reduce 243) ((";") reduce 243) (("@") reduce 243) ((T_ABSTRACT) reduce 243) ((T_ARRAY) reduce 243) ((T_ARRAY_CAST) reduce 243) ((T_ATTRIBUTE) reduce 243) ((T_BOOL_CAST) reduce 243) ((T_BREAK) reduce 243) ((T_CLASS) reduce 243) ((T_CLASS_C) reduce 243) ((T_CLONE) reduce 243) ((T_CONST) reduce 243) ((T_CONSTANT_ENCAPSED_STRING) reduce 243) ((T_CONTINUE) reduce 243) ((T_DEC) reduce 243) ((T_DECLARE) reduce 243) ((T_DIR) reduce 243) ((T_DNUMBER) reduce 243) ((T_DO) reduce 243) ((T_DOUBLE_CAST) reduce 243) ((T_ECHO) reduce 243) ((T_ELSE) reduce 243) ((T_ELSEIF) reduce 243) ((T_EMPTY) reduce 243) ((T_ENUM) reduce 243) ((T_EVAL) reduce 243) ((T_EXIT) reduce 243) ((T_FILE) reduce 243) ((T_FINAL) reduce 243) ((T_FN) reduce 243) ((T_FOR) reduce 243) ((T_FOREACH) reduce 243) ((T_FUNCTION) reduce 243) ((T_FUNC_C) reduce 243) ((T_GLOBAL) reduce 243) ((T_GOTO) reduce 243) ((T_HALT_COMPILER) reduce 243) ((T_IF) reduce 243) ((T_INC) reduce 243) ((T_INCLUDE) reduce 243) ((T_INCLUDE_ONCE) reduce 243) ((T_INLINE_HTML) reduce 243) ((T_INTERFACE) reduce 243) ((T_INT_CAST) reduce 243) ((T_ISSET) reduce 243) ((T_LINE) reduce 243) ((T_LIST) reduce 243) ((T_LNUMBER) reduce 243) ((T_MATCH) reduce 243) ((T_METHOD_C) reduce 243) ((T_NAMESPACE) reduce 243) ((T_NAME_FULLY_QUALIFIED) reduce 243) ((T_NAME_QUALIFIED) reduce 243) ((T_NAME_RELATIVE) reduce 243) ((T_NEW) reduce 243) ((T_NS_C) reduce 243) ((T_OBJECT_CAST) reduce 243) ((T_PRINT) reduce 243) ((T_PROPERTY_C) reduce 243) ((T_READONLY) reduce 243) ((T_REQUIRE) reduce 243) ((T_REQUIRE_ONCE) reduce 243) ((T_RETURN) reduce 243) ((T_START_HEREDOC) reduce 243) ((T_STATIC) reduce 243) ((T_STRING) reduce 243) ((T_STRING_CAST) reduce 243) ((T_SWITCH) reduce 243) ((T_THROW) reduce 243) ((T_TRAIT) reduce 243) ((T_TRAIT_C) reduce 243) ((T_TRY) reduce 243) ((T_UNSET) reduce 243) ((T_UNSET_CAST) reduce 243) ((T_USE) reduce 243) ((T_VARIABLE) reduce 243) ((T_VOID_CAST) reduce 243) ((T_WHILE) reduce 243) ((T_YIELD) reduce 243) ((T_YIELD_FROM) reduce 243) (("[") reduce 243) (("`") reduce 243) (("{") reduce 243) (("~") reduce 243)) 32842 ((("!") reduce 170) (("\"") reduce 170) (("$") reduce 170) (($) reduce 170) (("(") reduce 170) (("+") reduce 170) (("-") reduce 170) ((";") reduce 170) (("@") reduce 170) ((T_ABSTRACT) reduce 170) ((T_ARRAY) reduce 170) ((T_ARRAY_CAST) reduce 170) ((T_ATTRIBUTE) reduce 170) ((T_BOOL_CAST) reduce 170) ((T_BREAK) reduce 170) ((T_CLASS) reduce 170) ((T_CLASS_C) reduce 170) ((T_CLONE) reduce 170) ((T_CONST) reduce 170) ((T_CONSTANT_ENCAPSED_STRING) reduce 170) ((T_CONTINUE) reduce 170) ((T_DEC) reduce 170) ((T_DECLARE) reduce 170) ((T_DIR) reduce 170) ((T_DNUMBER) reduce 170) ((T_DO) reduce 170) ((T_DOUBLE_CAST) reduce 170) ((T_ECHO) reduce 170) ((T_ELSE) reduce 170) ((T_ELSEIF) reduce 170) ((T_EMPTY) reduce 170) ((T_ENUM) reduce 170) ((T_EVAL) reduce 170) ((T_EXIT) reduce 170) ((T_FILE) reduce 170) ((T_FINAL) reduce 170) ((T_FN) reduce 170) ((T_FOR) reduce 170) ((T_FOREACH) reduce 170) ((T_FUNCTION) reduce 170) ((T_FUNC_C) reduce 170) ((T_GLOBAL) reduce 170) ((T_GOTO) reduce 170) ((T_HALT_COMPILER) reduce 170) ((T_IF) reduce 170) ((T_INC) reduce 170) ((T_INCLUDE) reduce 170) ((T_INCLUDE_ONCE) reduce 170) ((T_INLINE_HTML) reduce 170) ((T_INTERFACE) reduce 170) ((T_INT_CAST) reduce 170) ((T_ISSET) reduce 170) ((T_LINE) reduce 170) ((T_LIST) reduce 170) ((T_LNUMBER) reduce 170) ((T_MATCH) reduce 170) ((T_METHOD_C) reduce 170) ((T_NAMESPACE) reduce 170) ((T_NAME_FULLY_QUALIFIED) reduce 170) ((T_NAME_QUALIFIED) reduce 170) ((T_NAME_RELATIVE) reduce 170) ((T_NEW) reduce 170) ((T_NS_C) reduce 170) ((T_OBJECT_CAST) reduce 170) ((T_PRINT) reduce 170) ((T_PROPERTY_C) reduce 170) ((T_READONLY) reduce 170) ((T_REQUIRE) reduce 170) ((T_REQUIRE_ONCE) reduce 170) ((T_RETURN) reduce 170) ((T_START_HEREDOC) reduce 170) ((T_STATIC) reduce 170) ((T_STRING) reduce 170) ((T_STRING_CAST) reduce 170) ((T_SWITCH) reduce 170) ((T_THROW) reduce 170) ((T_TRAIT) reduce 170) ((T_TRAIT_C) reduce 170) ((T_TRY) reduce 170) ((T_UNSET) reduce 170) ((T_UNSET_CAST) reduce 170) ((T_USE) reduce 170) ((T_VARIABLE) reduce 170) ((T_VOID_CAST) reduce 170) ((T_WHILE) reduce 170) ((T_YIELD) reduce 170) ((T_YIELD_FROM) reduce 170) (("[") reduce 170) (("`") reduce 170) (("{") reduce 170) (("~") reduce 170)) 32847 ((("!") reduce 162) (("\"") reduce 162) (($) reduce 162) (("$") reduce 162) (("(") reduce 162) (("+") reduce 162) (("-") reduce 162) ((";") reduce 162) (("@") reduce 162) ((T_ABSTRACT) reduce 162) ((T_ARRAY) reduce 162) ((T_ARRAY_CAST) reduce 162) ((T_ATTRIBUTE) reduce 162) ((T_BOOL_CAST) reduce 162) ((T_BREAK) reduce 162) ((T_CLASS) reduce 162) ((T_CLASS_C) reduce 162) ((T_CLONE) reduce 162) ((T_CONST) reduce 162) ((T_CONSTANT_ENCAPSED_STRING) reduce 162) ((T_CONTINUE) reduce 162) ((T_DEC) reduce 162) ((T_DECLARE) reduce 162) ((T_DIR) reduce 162) ((T_DNUMBER) reduce 162) ((T_DO) reduce 162) ((T_DOUBLE_CAST) reduce 162) ((T_ECHO) reduce 162) ((T_ELSE) reduce 162) ((T_ELSEIF) reduce 162) ((T_EMPTY) reduce 162) ((T_ENUM) reduce 162) ((T_EVAL) reduce 162) ((T_EXIT) reduce 162) ((T_FILE) reduce 162) ((T_FINAL) reduce 162) ((T_FN) reduce 162) ((T_FOR) reduce 162) ((T_FOREACH) reduce 162) ((T_FUNCTION) reduce 162) ((T_FUNC_C) reduce 162) ((T_GLOBAL) reduce 162) ((T_GOTO) reduce 162) ((T_HALT_COMPILER) reduce 162) ((T_IF) reduce 162) ((T_INC) reduce 162) ((T_INCLUDE) reduce 162) ((T_INCLUDE_ONCE) reduce 162) ((T_INLINE_HTML) reduce 162) ((T_INTERFACE) reduce 162) ((T_INT_CAST) reduce 162) ((T_ISSET) reduce 162) ((T_LINE) reduce 162) ((T_LIST) reduce 162) ((T_LNUMBER) reduce 162) ((T_MATCH) reduce 162) ((T_METHOD_C) reduce 162) ((T_NAMESPACE) reduce 162) ((T_NAME_FULLY_QUALIFIED) reduce 162) ((T_NAME_QUALIFIED) reduce 162) ((T_NAME_RELATIVE) reduce 162) ((T_NEW) reduce 162) ((T_NS_C) reduce 162) ((T_OBJECT_CAST) reduce 162) ((T_PRINT) reduce 162) ((T_PROPERTY_C) reduce 162) ((T_READONLY) reduce 162) ((T_REQUIRE) reduce 162) ((T_REQUIRE_ONCE) reduce 162) ((T_RETURN) reduce 162) ((T_START_HEREDOC) reduce 162) ((T_STATIC) reduce 162) ((T_STRING) reduce 162) ((T_STRING_CAST) reduce 162) ((T_SWITCH) reduce 162) ((T_THROW) reduce 162) ((T_TRAIT) reduce 162) ((T_TRAIT_C) reduce 162) ((T_TRY) reduce 162) ((T_UNSET) reduce 162) ((T_UNSET_CAST) reduce 162) ((T_USE) reduce 162) ((T_VARIABLE) reduce 162) ((T_VOID_CAST) reduce 162) ((T_WHILE) reduce 162) ((T_YIELD) reduce 162) ((T_YIELD_FROM) reduce 162) (("[") reduce 162) (("`") reduce 162) (("{") reduce 162) (("~") reduce 162)) 32850 ((("!") reduce 171) (("\"") reduce 171) (("$") reduce 171) (($) reduce 171) (("(") reduce 171) (("+") reduce 171) (("-") reduce 171) ((";") reduce 171) (("@") reduce 171) ((T_ABSTRACT) reduce 171) ((T_ARRAY) reduce 171) ((T_ARRAY_CAST) reduce 171) ((T_ATTRIBUTE) reduce 171) ((T_BOOL_CAST) reduce 171) ((T_BREAK) reduce 171) ((T_CATCH) reduce 171) ((T_CLASS) reduce 171) ((T_CLASS_C) reduce 171) ((T_CLONE) reduce 171) ((T_CONST) reduce 171) ((T_CONSTANT_ENCAPSED_STRING) reduce 171) ((T_CONTINUE) reduce 171) ((T_DEC) reduce 171) ((T_DECLARE) reduce 171) ((T_DIR) reduce 171) ((T_DNUMBER) reduce 171) ((T_DO) reduce 171) ((T_DOUBLE_CAST) reduce 171) ((T_ECHO) reduce 171) ((T_ELSE) reduce 171) ((T_ELSEIF) reduce 171) ((T_EMPTY) reduce 171) ((T_ENUM) reduce 171) ((T_EVAL) reduce 171) ((T_EXIT) reduce 171) ((T_FILE) reduce 171) ((T_FINAL) reduce 171) ((T_FINALLY) reduce 171) ((T_FN) reduce 171) ((T_FOR) reduce 171) ((T_FOREACH) reduce 171) ((T_FUNCTION) reduce 171) ((T_FUNC_C) reduce 171) ((T_GLOBAL) reduce 171) ((T_GOTO) reduce 171) ((T_HALT_COMPILER) reduce 171) ((T_IF) reduce 171) ((T_INC) reduce 171) ((T_INCLUDE) reduce 171) ((T_INCLUDE_ONCE) reduce 171) ((T_INLINE_HTML) reduce 171) ((T_INTERFACE) reduce 171) ((T_INT_CAST) reduce 171) ((T_ISSET) reduce 171) ((T_LINE) reduce 171) ((T_LIST) reduce 171) ((T_LNUMBER) reduce 171) ((T_MATCH) reduce 171) ((T_METHOD_C) reduce 171) ((T_NAMESPACE) reduce 171) ((T_NAME_FULLY_QUALIFIED) reduce 171) ((T_NAME_QUALIFIED) reduce 171) ((T_NAME_RELATIVE) reduce 171) ((T_NEW) reduce 171) ((T_NS_C) reduce 171) ((T_OBJECT_CAST) reduce 171) ((T_PRINT) reduce 171) ((T_PROPERTY_C) reduce 171) ((T_READONLY) reduce 171) ((T_REQUIRE) reduce 171) ((T_REQUIRE_ONCE) reduce 171) ((T_RETURN) reduce 171) ((T_START_HEREDOC) reduce 171) ((T_STATIC) reduce 171) ((T_STRING) reduce 171) ((T_STRING_CAST) reduce 171) ((T_SWITCH) reduce 171) ((T_THROW) reduce 171) ((T_TRAIT) reduce 171) ((T_TRAIT_C) reduce 171) ((T_TRY) reduce 171) ((T_UNSET) reduce 171) ((T_UNSET_CAST) reduce 171) ((T_USE) reduce 171) ((T_VARIABLE) reduce 171) ((T_VOID_CAST) reduce 171) ((T_WHILE) reduce 171) ((T_YIELD) reduce 171) ((T_YIELD_FROM) reduce 171) (("[") reduce 171) (("`") reduce 171) (("{") reduce 171) (("~") reduce 171)) 32851 ((("!") reduce 177) (("\"") reduce 177) (($) reduce 177) (("$") reduce 177) (("(") reduce 177) (("+") reduce 177) (("-") reduce 177) ((";") reduce 177) (("@") reduce 177) ((T_ABSTRACT) reduce 177) ((T_ARRAY) reduce 177) ((T_ARRAY_CAST) reduce 177) ((T_ATTRIBUTE) reduce 177) ((T_BOOL_CAST) reduce 177) ((T_BREAK) reduce 177) ((T_CATCH) shift) ((T_CLASS) reduce 177) ((T_CLASS_C) reduce 177) ((T_CLONE) reduce 177) ((T_CONST) reduce 177) ((T_CONSTANT_ENCAPSED_STRING) reduce 177) ((T_CONTINUE) reduce 177) ((T_DEC) reduce 177) ((T_DECLARE) reduce 177) ((T_DIR) reduce 177) ((T_DNUMBER) reduce 177) ((T_DO) reduce 177) ((T_DOUBLE_CAST) reduce 177) ((T_ECHO) reduce 177) ((T_ELSE) reduce 177) ((T_ELSEIF) reduce 177) ((T_EMPTY) reduce 177) ((T_ENUM) reduce 177) ((T_EVAL) reduce 177) ((T_EXIT) reduce 177) ((T_FILE) reduce 177) ((T_FINAL) reduce 177) ((T_FINALLY) shift) ((T_FN) reduce 177) ((T_FOR) reduce 177) ((T_FOREACH) reduce 177) ((T_FUNCTION) reduce 177) ((T_FUNC_C) reduce 177) ((T_GLOBAL) reduce 177) ((T_GOTO) reduce 177) ((T_HALT_COMPILER) reduce 177) ((T_IF) reduce 177) ((T_INC) reduce 177) ((T_INCLUDE) reduce 177) ((T_INCLUDE_ONCE) reduce 177) ((T_INLINE_HTML) reduce 177) ((T_INTERFACE) reduce 177) ((T_INT_CAST) reduce 177) ((T_ISSET) reduce 177) ((T_LINE) reduce 177) ((T_LIST) reduce 177) ((T_LNUMBER) reduce 177) ((T_MATCH) reduce 177) ((T_METHOD_C) reduce 177) ((T_NAMESPACE) reduce 177) ((T_NAME_FULLY_QUALIFIED) reduce 177) ((T_NAME_QUALIFIED) reduce 177) ((T_NAME_RELATIVE) reduce 177) ((T_NEW) reduce 177) ((T_NS_C) reduce 177) ((T_OBJECT_CAST) reduce 177) ((T_PRINT) reduce 177) ((T_PROPERTY_C) reduce 177) ((T_READONLY) reduce 177) ((T_REQUIRE) reduce 177) ((T_REQUIRE_ONCE) reduce 177) ((T_RETURN) reduce 177) ((T_START_HEREDOC) reduce 177) ((T_STATIC) reduce 177) ((T_STRING) reduce 177) ((T_STRING_CAST) reduce 177) ((T_SWITCH) reduce 177) ((T_THROW) reduce 177) ((T_TRAIT) reduce 177) ((T_TRAIT_C) reduce 177) ((T_TRY) reduce 177) ((T_UNSET) reduce 177) ((T_UNSET_CAST) reduce 177) ((T_USE) reduce 177) ((T_VARIABLE) reduce 177) ((T_VOID_CAST) reduce 177) ((T_WHILE) reduce 177) ((T_YIELD) reduce 177) ((T_YIELD_FROM) reduce 177) (("[") reduce 177) (("`") reduce 177) (("{") reduce 177) (("~") reduce 177)) 32854 ((("!") reduce 167) (("\"") reduce 167) (($) reduce 167) (("$") reduce 167) (("(") reduce 167) (("+") reduce 167) (("-") reduce 167) ((";") reduce 167) (("@") reduce 167) ((T_ABSTRACT) reduce 167) ((T_ARRAY) reduce 167) ((T_ARRAY_CAST) reduce 167) ((T_ATTRIBUTE) reduce 167) ((T_BOOL_CAST) reduce 167) ((T_BREAK) reduce 167) ((T_CLASS) reduce 167) ((T_CLASS_C) reduce 167) ((T_CLONE) reduce 167) ((T_CONST) reduce 167) ((T_CONSTANT_ENCAPSED_STRING) reduce 167) ((T_CONTINUE) reduce 167) ((T_DEC) reduce 167) ((T_DECLARE) reduce 167) ((T_DIR) reduce 167) ((T_DNUMBER) reduce 167) ((T_DO) reduce 167) ((T_DOUBLE_CAST) reduce 167) ((T_ECHO) reduce 167) ((T_ELSE) reduce 167) ((T_ELSEIF) reduce 167) ((T_EMPTY) reduce 167) ((T_ENUM) reduce 167) ((T_EVAL) reduce 167) ((T_EXIT) reduce 167) ((T_FILE) reduce 167) ((T_FINAL) reduce 167) ((T_FN) reduce 167) ((T_FOR) reduce 167) ((T_FOREACH) reduce 167) ((T_FUNCTION) reduce 167) ((T_FUNC_C) reduce 167) ((T_GLOBAL) reduce 167) ((T_GOTO) reduce 167) ((T_HALT_COMPILER) reduce 167) ((T_IF) reduce 167) ((T_INC) reduce 167) ((T_INCLUDE) reduce 167) ((T_INCLUDE_ONCE) reduce 167) ((T_INLINE_HTML) reduce 167) ((T_INTERFACE) reduce 167) ((T_INT_CAST) reduce 167) ((T_ISSET) reduce 167) ((T_LINE) reduce 167) ((T_LIST) reduce 167) ((T_LNUMBER) reduce 167) ((T_MATCH) reduce 167) ((T_METHOD_C) reduce 167) ((T_NAMESPACE) reduce 167) ((T_NAME_FULLY_QUALIFIED) reduce 167) ((T_NAME_QUALIFIED) reduce 167) ((T_NAME_RELATIVE) reduce 167) ((T_NEW) reduce 167) ((T_NS_C) reduce 167) ((T_OBJECT_CAST) reduce 167) ((T_PRINT) reduce 167) ((T_PROPERTY_C) reduce 167) ((T_READONLY) reduce 167) ((T_REQUIRE) reduce 167) ((T_REQUIRE_ONCE) reduce 167) ((T_RETURN) reduce 167) ((T_START_HEREDOC) reduce 167) ((T_STATIC) reduce 167) ((T_STRING) reduce 167) ((T_STRING_CAST) reduce 167) ((T_SWITCH) reduce 167) ((T_THROW) reduce 167) ((T_TRAIT) reduce 167) ((T_TRAIT_C) reduce 167) ((T_TRY) reduce 167) ((T_UNSET) reduce 167) ((T_UNSET_CAST) reduce 167) ((T_USE) reduce 167) ((T_VARIABLE) reduce 167) ((T_VOID_CAST) reduce 167) ((T_WHILE) reduce 167) ((T_YIELD) reduce 167) ((T_YIELD_FROM) reduce 167) (("[") reduce 167) (("`") reduce 167) (("{") reduce 167) (("~") reduce 167)) 32857 ((("!") reduce 178) (("\"") reduce 178) (($) reduce 178) (("$") reduce 178) (("(") reduce 178) (("+") reduce 178) (("-") reduce 178) ((";") reduce 178) (("@") reduce 178) ((T_ABSTRACT) reduce 178) ((T_ARRAY) reduce 178) ((T_ARRAY_CAST) reduce 178) ((T_ATTRIBUTE) reduce 178) ((T_BOOL_CAST) reduce 178) ((T_BREAK) reduce 178) ((T_CLASS) reduce 178) ((T_CLASS_C) reduce 178) ((T_CLONE) reduce 178) ((T_CONST) reduce 178) ((T_CONSTANT_ENCAPSED_STRING) reduce 178) ((T_CONTINUE) reduce 178) ((T_DEC) reduce 178) ((T_DECLARE) reduce 178) ((T_DIR) reduce 178) ((T_DNUMBER) reduce 178) ((T_DO) reduce 178) ((T_DOUBLE_CAST) reduce 178) ((T_ECHO) reduce 178) ((T_ELSE) reduce 178) ((T_ELSEIF) reduce 178) ((T_EMPTY) reduce 178) ((T_ENUM) reduce 178) ((T_EVAL) reduce 178) ((T_EXIT) reduce 178) ((T_FILE) reduce 178) ((T_FINAL) reduce 178) ((T_FN) reduce 178) ((T_FOR) reduce 178) ((T_FOREACH) reduce 178) ((T_FUNCTION) reduce 178) ((T_FUNC_C) reduce 178) ((T_GLOBAL) reduce 178) ((T_GOTO) reduce 178) ((T_HALT_COMPILER) reduce 178) ((T_IF) reduce 178) ((T_INC) reduce 178) ((T_INCLUDE) reduce 178) ((T_INCLUDE_ONCE) reduce 178) ((T_INLINE_HTML) reduce 178) ((T_INTERFACE) reduce 178) ((T_INT_CAST) reduce 178) ((T_ISSET) reduce 178) ((T_LINE) reduce 178) ((T_LIST) reduce 178) ((T_LNUMBER) reduce 178) ((T_MATCH) reduce 178) ((T_METHOD_C) reduce 178) ((T_NAMESPACE) reduce 178) ((T_NAME_FULLY_QUALIFIED) reduce 178) ((T_NAME_QUALIFIED) reduce 178) ((T_NAME_RELATIVE) reduce 178) ((T_NEW) reduce 178) ((T_NS_C) reduce 178) ((T_OBJECT_CAST) reduce 178) ((T_PRINT) reduce 178) ((T_PROPERTY_C) reduce 178) ((T_READONLY) reduce 178) ((T_REQUIRE) reduce 178) ((T_REQUIRE_ONCE) reduce 178) ((T_RETURN) reduce 178) ((T_START_HEREDOC) reduce 178) ((T_STATIC) reduce 178) ((T_STRING) reduce 178) ((T_STRING_CAST) reduce 178) ((T_SWITCH) reduce 178) ((T_THROW) reduce 178) ((T_TRAIT) reduce 178) ((T_TRAIT_C) reduce 178) ((T_TRY) reduce 178) ((T_UNSET) reduce 178) ((T_UNSET_CAST) reduce 178) ((T_USE) reduce 178) ((T_VARIABLE) reduce 178) ((T_VOID_CAST) reduce 178) ((T_WHILE) reduce 178) ((T_YIELD) reduce 178) ((T_YIELD_FROM) reduce 178) (("[") reduce 178) (("`") reduce 178) (("{") reduce 178) (("~") reduce 178)) 32864 ((("!") reduce 172) (("\"") reduce 172) (($) reduce 172) (("$") reduce 172) (("(") reduce 172) (("+") reduce 172) (("-") reduce 172) ((";") reduce 172) (("@") reduce 172) ((T_ABSTRACT) reduce 172) ((T_ARRAY) reduce 172) ((T_ARRAY_CAST) reduce 172) ((T_ATTRIBUTE) reduce 172) ((T_BOOL_CAST) reduce 172) ((T_BREAK) reduce 172) ((T_CATCH) reduce 172) ((T_CLASS) reduce 172) ((T_CLASS_C) reduce 172) ((T_CLONE) reduce 172) ((T_CONST) reduce 172) ((T_CONSTANT_ENCAPSED_STRING) reduce 172) ((T_CONTINUE) reduce 172) ((T_DEC) reduce 172) ((T_DECLARE) reduce 172) ((T_DIR) reduce 172) ((T_DNUMBER) reduce 172) ((T_DO) reduce 172) ((T_DOUBLE_CAST) reduce 172) ((T_ECHO) reduce 172) ((T_ELSE) reduce 172) ((T_ELSEIF) reduce 172) ((T_EMPTY) reduce 172) ((T_ENUM) reduce 172) ((T_EVAL) reduce 172) ((T_EXIT) reduce 172) ((T_FILE) reduce 172) ((T_FINAL) reduce 172) ((T_FINALLY) reduce 172) ((T_FN) reduce 172) ((T_FOR) reduce 172) ((T_FOREACH) reduce 172) ((T_FUNCTION) reduce 172) ((T_FUNC_C) reduce 172) ((T_GLOBAL) reduce 172) ((T_GOTO) reduce 172) ((T_HALT_COMPILER) reduce 172) ((T_IF) reduce 172) ((T_INC) reduce 172) ((T_INCLUDE) reduce 172) ((T_INCLUDE_ONCE) reduce 172) ((T_INLINE_HTML) reduce 172) ((T_INTERFACE) reduce 172) ((T_INT_CAST) reduce 172) ((T_ISSET) reduce 172) ((T_LINE) reduce 172) ((T_LIST) reduce 172) ((T_LNUMBER) reduce 172) ((T_MATCH) reduce 172) ((T_METHOD_C) reduce 172) ((T_NAMESPACE) reduce 172) ((T_NAME_FULLY_QUALIFIED) reduce 172) ((T_NAME_QUALIFIED) reduce 172) ((T_NAME_RELATIVE) reduce 172) ((T_NEW) reduce 172) ((T_NS_C) reduce 172) ((T_OBJECT_CAST) reduce 172) ((T_PRINT) reduce 172) ((T_PROPERTY_C) reduce 172) ((T_READONLY) reduce 172) ((T_REQUIRE) reduce 172) ((T_REQUIRE_ONCE) reduce 172) ((T_RETURN) reduce 172) ((T_START_HEREDOC) reduce 172) ((T_STATIC) reduce 172) ((T_STRING) reduce 172) ((T_STRING_CAST) reduce 172) ((T_SWITCH) reduce 172) ((T_THROW) reduce 172) ((T_TRAIT) reduce 172) ((T_TRAIT_C) reduce 172) ((T_TRY) reduce 172) ((T_UNSET) reduce 172) ((T_UNSET_CAST) reduce 172) ((T_USE) reduce 172) ((T_VARIABLE) reduce 172) ((T_VOID_CAST) reduce 172) ((T_WHILE) reduce 172) ((T_YIELD) reduce 172) ((T_YIELD_FROM) reduce 172) (("[") reduce 172) (("`") reduce 172) (("{") reduce 172) (("~") reduce 172)) 32869 ((("!") reduce 153) (("\"") reduce 153) (("$") reduce 153) (($) reduce 153) (("(") reduce 153) (("+") reduce 153) (("-") reduce 153) ((";") reduce 153) (("@") reduce 153) ((T_ABSTRACT) reduce 153) ((T_ARRAY) reduce 153) ((T_ARRAY_CAST) reduce 153) ((T_ATTRIBUTE) reduce 153) ((T_BOOL_CAST) reduce 153) ((T_BREAK) reduce 153) ((T_CLASS) reduce 153) ((T_CLASS_C) reduce 153) ((T_CLONE) reduce 153) ((T_CONST) reduce 153) ((T_CONSTANT_ENCAPSED_STRING) reduce 153) ((T_CONTINUE) reduce 153) ((T_DEC) reduce 153) ((T_DECLARE) reduce 153) ((T_DIR) reduce 153) ((T_DNUMBER) reduce 153) ((T_DO) reduce 153) ((T_DOUBLE_CAST) reduce 153) ((T_ECHO) reduce 153) ((T_ELSE) reduce 153) ((T_ELSEIF) reduce 153) ((T_EMPTY) reduce 153) ((T_ENUM) reduce 153) ((T_EVAL) reduce 153) ((T_EXIT) reduce 153) ((T_FILE) reduce 153) ((T_FINAL) reduce 153) ((T_FN) reduce 153) ((T_FOR) reduce 153) ((T_FOREACH) reduce 153) ((T_FUNCTION) reduce 153) ((T_FUNC_C) reduce 153) ((T_GLOBAL) reduce 153) ((T_GOTO) reduce 153) ((T_HALT_COMPILER) reduce 153) ((T_IF) reduce 153) ((T_INC) reduce 153) ((T_INCLUDE) reduce 153) ((T_INCLUDE_ONCE) reduce 153) ((T_INLINE_HTML) reduce 153) ((T_INTERFACE) reduce 153) ((T_INT_CAST) reduce 153) ((T_ISSET) reduce 153) ((T_LINE) reduce 153) ((T_LIST) reduce 153) ((T_LNUMBER) reduce 153) ((T_MATCH) reduce 153) ((T_METHOD_C) reduce 153) ((T_NAMESPACE) reduce 153) ((T_NAME_FULLY_QUALIFIED) reduce 153) ((T_NAME_QUALIFIED) reduce 153) ((T_NAME_RELATIVE) reduce 153) ((T_NEW) reduce 153) ((T_NS_C) reduce 153) ((T_OBJECT_CAST) reduce 153) ((T_PRINT) reduce 153) ((T_PROPERTY_C) reduce 153) ((T_READONLY) reduce 153) ((T_REQUIRE) reduce 153) ((T_REQUIRE_ONCE) reduce 153) ((T_RETURN) reduce 153) ((T_START_HEREDOC) reduce 153) ((T_STATIC) reduce 153) ((T_STRING) reduce 153) ((T_STRING_CAST) reduce 153) ((T_SWITCH) reduce 153) ((T_THROW) reduce 153) ((T_TRAIT) reduce 153) ((T_TRAIT_C) reduce 153) ((T_TRY) reduce 153) ((T_UNSET) reduce 153) ((T_UNSET_CAST) reduce 153) ((T_USE) reduce 153) ((T_VARIABLE) reduce 153) ((T_VOID_CAST) reduce 153) ((T_WHILE) reduce 153) ((T_YIELD) reduce 153) ((T_YIELD_FROM) reduce 153) (("[") reduce 153) (("`") reduce 153) (("{") reduce 153) (("~") reduce 153)) 32873 ((("!") reduce 224) (("\"") reduce 224) (($) reduce 224) (("$") reduce 224) (("(") reduce 224) (("+") reduce 224) (("-") reduce 224) ((";") reduce 224) (("@") reduce 224) ((T_ABSTRACT) reduce 224) ((T_ARRAY) reduce 224) ((T_ARRAY_CAST) reduce 224) ((T_ATTRIBUTE) reduce 224) ((T_BOOL_CAST) reduce 224) ((T_BREAK) reduce 224) ((T_CLASS) reduce 224) ((T_CLASS_C) reduce 224) ((T_CLONE) reduce 224) ((T_CONST) reduce 224) ((T_CONSTANT_ENCAPSED_STRING) reduce 224) ((T_CONTINUE) reduce 224) ((T_DEC) reduce 224) ((T_DECLARE) reduce 224) ((T_DIR) reduce 224) ((T_DNUMBER) reduce 224) ((T_DO) reduce 224) ((T_DOUBLE_CAST) reduce 224) ((T_ECHO) reduce 224) ((T_ELSE) reduce 224) ((T_ELSEIF) reduce 224) ((T_EMPTY) reduce 224) ((T_ENUM) reduce 224) ((T_EVAL) reduce 224) ((T_EXIT) reduce 224) ((T_FILE) reduce 224) ((T_FINAL) reduce 224) ((T_FN) reduce 224) ((T_FOR) reduce 224) ((T_FOREACH) reduce 224) ((T_FUNCTION) reduce 224) ((T_FUNC_C) reduce 224) ((T_GLOBAL) reduce 224) ((T_GOTO) reduce 224) ((T_HALT_COMPILER) reduce 224) ((T_IF) reduce 224) ((T_INC) reduce 224) ((T_INCLUDE) reduce 224) ((T_INCLUDE_ONCE) reduce 224) ((T_INLINE_HTML) reduce 224) ((T_INTERFACE) reduce 224) ((T_INT_CAST) reduce 224) ((T_ISSET) reduce 224) ((T_LINE) reduce 224) ((T_LIST) reduce 224) ((T_LNUMBER) reduce 224) ((T_MATCH) reduce 224) ((T_METHOD_C) reduce 224) ((T_NAMESPACE) reduce 224) ((T_NAME_FULLY_QUALIFIED) reduce 224) ((T_NAME_QUALIFIED) reduce 224) ((T_NAME_RELATIVE) reduce 224) ((T_NEW) reduce 224) ((T_NS_C) reduce 224) ((T_OBJECT_CAST) reduce 224) ((T_PRINT) reduce 224) ((T_PROPERTY_C) reduce 224) ((T_READONLY) reduce 224) ((T_REQUIRE) reduce 224) ((T_REQUIRE_ONCE) reduce 224) ((T_RETURN) reduce 224) ((T_START_HEREDOC) reduce 224) ((T_STATIC) reduce 224) ((T_STRING) reduce 224) ((T_STRING_CAST) reduce 224) ((T_SWITCH) reduce 224) ((T_THROW) reduce 224) ((T_TRAIT) reduce 224) ((T_TRAIT_C) reduce 224) ((T_TRY) reduce 224) ((T_UNSET) reduce 224) ((T_UNSET_CAST) reduce 224) ((T_USE) reduce 224) ((T_VARIABLE) reduce 224) ((T_VOID_CAST) reduce 224) ((T_WHILE) reduce 224) ((T_YIELD) reduce 224) ((T_YIELD_FROM) reduce 224) (("[") reduce 224) (("`") reduce 224) (("{") reduce 224) (("~") reduce 224)) 32875 ((("!") reduce 225) (("\"") reduce 225) (("$") reduce 225) (($) reduce 225) (("(") reduce 225) (("+") reduce 225) (("-") reduce 225) ((";") reduce 225) (("@") reduce 225) ((T_ABSTRACT) reduce 225) ((T_ARRAY) reduce 225) ((T_ARRAY_CAST) reduce 225) ((T_ATTRIBUTE) reduce 225) ((T_BOOL_CAST) reduce 225) ((T_BREAK) reduce 225) ((T_CLASS) reduce 225) ((T_CLASS_C) reduce 225) ((T_CLONE) reduce 225) ((T_CONST) reduce 225) ((T_CONSTANT_ENCAPSED_STRING) reduce 225) ((T_CONTINUE) reduce 225) ((T_DEC) reduce 225) ((T_DECLARE) reduce 225) ((T_DIR) reduce 225) ((T_DNUMBER) reduce 225) ((T_DO) reduce 225) ((T_DOUBLE_CAST) reduce 225) ((T_ECHO) reduce 225) ((T_ELSE) reduce 225) ((T_ELSEIF) reduce 225) ((T_EMPTY) reduce 225) ((T_ENUM) reduce 225) ((T_EVAL) reduce 225) ((T_EXIT) reduce 225) ((T_FILE) reduce 225) ((T_FINAL) reduce 225) ((T_FN) reduce 225) ((T_FOR) reduce 225) ((T_FOREACH) reduce 225) ((T_FUNCTION) reduce 225) ((T_FUNC_C) reduce 225) ((T_GLOBAL) reduce 225) ((T_GOTO) reduce 225) ((T_HALT_COMPILER) reduce 225) ((T_IF) reduce 225) ((T_INC) reduce 225) ((T_INCLUDE) reduce 225) ((T_INCLUDE_ONCE) reduce 225) ((T_INLINE_HTML) reduce 225) ((T_INTERFACE) reduce 225) ((T_INT_CAST) reduce 225) ((T_ISSET) reduce 225) ((T_LINE) reduce 225) ((T_LIST) reduce 225) ((T_LNUMBER) reduce 225) ((T_MATCH) reduce 225) ((T_METHOD_C) reduce 225) ((T_NAMESPACE) reduce 225) ((T_NAME_FULLY_QUALIFIED) reduce 225) ((T_NAME_QUALIFIED) reduce 225) ((T_NAME_RELATIVE) reduce 225) ((T_NEW) reduce 225) ((T_NS_C) reduce 225) ((T_OBJECT_CAST) reduce 225) ((T_PRINT) reduce 225) ((T_PROPERTY_C) reduce 225) ((T_READONLY) reduce 225) ((T_REQUIRE) reduce 225) ((T_REQUIRE_ONCE) reduce 225) ((T_RETURN) reduce 225) ((T_START_HEREDOC) reduce 225) ((T_STATIC) reduce 225) ((T_STRING) reduce 225) ((T_STRING_CAST) reduce 225) ((T_SWITCH) reduce 225) ((T_THROW) reduce 225) ((T_TRAIT) reduce 225) ((T_TRAIT_C) reduce 225) ((T_TRY) reduce 225) ((T_UNSET) reduce 225) ((T_UNSET_CAST) reduce 225) ((T_USE) reduce 225) ((T_VARIABLE) reduce 225) ((T_VOID_CAST) reduce 225) ((T_WHILE) reduce 225) ((T_YIELD) reduce 225) ((T_YIELD_FROM) reduce 225) (("[") reduce 225) (("`") reduce 225) (("{") reduce 225) (("~") reduce 225)) 32879 ((("!") reduce 226) (("\"") reduce 226) (("$") reduce 226) (($) reduce 226) (("(") reduce 226) (("+") reduce 226) (("-") reduce 226) ((";") reduce 226) (("@") reduce 226) ((T_ABSTRACT) reduce 226) ((T_ARRAY) reduce 226) ((T_ARRAY_CAST) reduce 226) ((T_ATTRIBUTE) reduce 226) ((T_BOOL_CAST) reduce 226) ((T_BREAK) reduce 226) ((T_CLASS) reduce 226) ((T_CLASS_C) reduce 226) ((T_CLONE) reduce 226) ((T_CONST) reduce 226) ((T_CONSTANT_ENCAPSED_STRING) reduce 226) ((T_CONTINUE) reduce 226) ((T_DEC) reduce 226) ((T_DECLARE) reduce 226) ((T_DIR) reduce 226) ((T_DNUMBER) reduce 226) ((T_DO) reduce 226) ((T_DOUBLE_CAST) reduce 226) ((T_ECHO) reduce 226) ((T_ELSE) reduce 226) ((T_ELSEIF) reduce 226) ((T_EMPTY) reduce 226) ((T_ENUM) reduce 226) ((T_EVAL) reduce 226) ((T_EXIT) reduce 226) ((T_FILE) reduce 226) ((T_FINAL) reduce 226) ((T_FN) reduce 226) ((T_FOR) reduce 226) ((T_FOREACH) reduce 226) ((T_FUNCTION) reduce 226) ((T_FUNC_C) reduce 226) ((T_GLOBAL) reduce 226) ((T_GOTO) reduce 226) ((T_HALT_COMPILER) reduce 226) ((T_IF) reduce 226) ((T_INC) reduce 226) ((T_INCLUDE) reduce 226) ((T_INCLUDE_ONCE) reduce 226) ((T_INLINE_HTML) reduce 226) ((T_INTERFACE) reduce 226) ((T_INT_CAST) reduce 226) ((T_ISSET) reduce 226) ((T_LINE) reduce 226) ((T_LIST) reduce 226) ((T_LNUMBER) reduce 226) ((T_MATCH) reduce 226) ((T_METHOD_C) reduce 226) ((T_NAMESPACE) reduce 226) ((T_NAME_FULLY_QUALIFIED) reduce 226) ((T_NAME_QUALIFIED) reduce 226) ((T_NAME_RELATIVE) reduce 226) ((T_NEW) reduce 226) ((T_NS_C) reduce 226) ((T_OBJECT_CAST) reduce 226) ((T_PRINT) reduce 226) ((T_PROPERTY_C) reduce 226) ((T_READONLY) reduce 226) ((T_REQUIRE) reduce 226) ((T_REQUIRE_ONCE) reduce 226) ((T_RETURN) reduce 226) ((T_START_HEREDOC) reduce 226) ((T_STATIC) reduce 226) ((T_STRING) reduce 226) ((T_STRING_CAST) reduce 226) ((T_SWITCH) reduce 226) ((T_THROW) reduce 226) ((T_TRAIT) reduce 226) ((T_TRAIT_C) reduce 226) ((T_TRY) reduce 226) ((T_UNSET) reduce 226) ((T_UNSET_CAST) reduce 226) ((T_USE) reduce 226) ((T_VARIABLE) reduce 226) ((T_VOID_CAST) reduce 226) ((T_WHILE) reduce 226) ((T_YIELD) reduce 226) ((T_YIELD_FROM) reduce 226) (("[") reduce 226) (("`") reduce 226) (("{") reduce 226) (("~") reduce 226)) 32882 ((("!") reduce 227) (("\"") reduce 227) (($) reduce 227) (("$") reduce 227) (("(") reduce 227) (("+") reduce 227) (("-") reduce 227) ((";") reduce 227) (("@") reduce 227) ((T_ABSTRACT) reduce 227) ((T_ARRAY) reduce 227) ((T_ARRAY_CAST) reduce 227) ((T_ATTRIBUTE) reduce 227) ((T_BOOL_CAST) reduce 227) ((T_BREAK) reduce 227) ((T_CLASS) reduce 227) ((T_CLASS_C) reduce 227) ((T_CLONE) reduce 227) ((T_CONST) reduce 227) ((T_CONSTANT_ENCAPSED_STRING) reduce 227) ((T_CONTINUE) reduce 227) ((T_DEC) reduce 227) ((T_DECLARE) reduce 227) ((T_DIR) reduce 227) ((T_DNUMBER) reduce 227) ((T_DO) reduce 227) ((T_DOUBLE_CAST) reduce 227) ((T_ECHO) reduce 227) ((T_ELSE) reduce 227) ((T_ELSEIF) reduce 227) ((T_EMPTY) reduce 227) ((T_ENUM) reduce 227) ((T_EVAL) reduce 227) ((T_EXIT) reduce 227) ((T_FILE) reduce 227) ((T_FINAL) reduce 227) ((T_FN) reduce 227) ((T_FOR) reduce 227) ((T_FOREACH) reduce 227) ((T_FUNCTION) reduce 227) ((T_FUNC_C) reduce 227) ((T_GLOBAL) reduce 227) ((T_GOTO) reduce 227) ((T_HALT_COMPILER) reduce 227) ((T_IF) reduce 227) ((T_INC) reduce 227) ((T_INCLUDE) reduce 227) ((T_INCLUDE_ONCE) reduce 227) ((T_INLINE_HTML) reduce 227) ((T_INTERFACE) reduce 227) ((T_INT_CAST) reduce 227) ((T_ISSET) reduce 227) ((T_LINE) reduce 227) ((T_LIST) reduce 227) ((T_LNUMBER) reduce 227) ((T_MATCH) reduce 227) ((T_METHOD_C) reduce 227) ((T_NAMESPACE) reduce 227) ((T_NAME_FULLY_QUALIFIED) reduce 227) ((T_NAME_QUALIFIED) reduce 227) ((T_NAME_RELATIVE) reduce 227) ((T_NEW) reduce 227) ((T_NS_C) reduce 227) ((T_OBJECT_CAST) reduce 227) ((T_PRINT) reduce 227) ((T_PROPERTY_C) reduce 227) ((T_READONLY) reduce 227) ((T_REQUIRE) reduce 227) ((T_REQUIRE_ONCE) reduce 227) ((T_RETURN) reduce 227) ((T_START_HEREDOC) reduce 227) ((T_STATIC) reduce 227) ((T_STRING) reduce 227) ((T_STRING_CAST) reduce 227) ((T_SWITCH) reduce 227) ((T_THROW) reduce 227) ((T_TRAIT) reduce 227) ((T_TRAIT_C) reduce 227) ((T_TRY) reduce 227) ((T_UNSET) reduce 227) ((T_UNSET_CAST) reduce 227) ((T_USE) reduce 227) ((T_VARIABLE) reduce 227) ((T_VOID_CAST) reduce 227) ((T_WHILE) reduce 227) ((T_YIELD) reduce 227) ((T_YIELD_FROM) reduce 227) (("[") reduce 227) (("`") reduce 227) (("{") reduce 227) (("~") reduce 227)) 32883 ((("!") reduce 169) (("\"") reduce 169) (($) reduce 169) (("$") reduce 169) (("(") reduce 169) (("+") reduce 169) (("-") reduce 169) ((";") reduce 169) (("@") reduce 169) ((T_ABSTRACT) reduce 169) ((T_ARRAY) reduce 169) ((T_ARRAY_CAST) reduce 169) ((T_ATTRIBUTE) reduce 169) ((T_BOOL_CAST) reduce 169) ((T_BREAK) reduce 169) ((T_CLASS) reduce 169) ((T_CLASS_C) reduce 169) ((T_CLONE) reduce 169) ((T_CONST) reduce 169) ((T_CONSTANT_ENCAPSED_STRING) reduce 169) ((T_CONTINUE) reduce 169) ((T_DEC) reduce 169) ((T_DECLARE) reduce 169) ((T_DIR) reduce 169) ((T_DNUMBER) reduce 169) ((T_DO) reduce 169) ((T_DOUBLE_CAST) reduce 169) ((T_ECHO) reduce 169) ((T_ELSE) reduce 169) ((T_ELSEIF) reduce 169) ((T_EMPTY) reduce 169) ((T_ENUM) reduce 169) ((T_EVAL) reduce 169) ((T_EXIT) reduce 169) ((T_FILE) reduce 169) ((T_FINAL) reduce 169) ((T_FN) reduce 169) ((T_FOR) reduce 169) ((T_FOREACH) reduce 169) ((T_FUNCTION) reduce 169) ((T_FUNC_C) reduce 169) ((T_GLOBAL) reduce 169) ((T_GOTO) reduce 169) ((T_HALT_COMPILER) reduce 169) ((T_IF) reduce 169) ((T_INC) reduce 169) ((T_INCLUDE) reduce 169) ((T_INCLUDE_ONCE) reduce 169) ((T_INLINE_HTML) reduce 169) ((T_INTERFACE) reduce 169) ((T_INT_CAST) reduce 169) ((T_ISSET) reduce 169) ((T_LINE) reduce 169) ((T_LIST) reduce 169) ((T_LNUMBER) reduce 169) ((T_MATCH) reduce 169) ((T_METHOD_C) reduce 169) ((T_NAMESPACE) reduce 169) ((T_NAME_FULLY_QUALIFIED) reduce 169) ((T_NAME_QUALIFIED) reduce 169) ((T_NAME_RELATIVE) reduce 169) ((T_NEW) reduce 169) ((T_NS_C) reduce 169) ((T_OBJECT_CAST) reduce 169) ((T_PRINT) reduce 169) ((T_PROPERTY_C) reduce 169) ((T_READONLY) reduce 169) ((T_REQUIRE) reduce 169) ((T_REQUIRE_ONCE) reduce 169) ((T_RETURN) reduce 169) ((T_START_HEREDOC) reduce 169) ((T_STATIC) reduce 169) ((T_STRING) reduce 169) ((T_STRING_CAST) reduce 169) ((T_SWITCH) reduce 169) ((T_THROW) reduce 169) ((T_TRAIT) reduce 169) ((T_TRAIT_C) reduce 169) ((T_TRY) reduce 169) ((T_UNSET) reduce 169) ((T_UNSET_CAST) reduce 169) ((T_USE) reduce 169) ((T_VARIABLE) reduce 169) ((T_VOID_CAST) reduce 169) ((T_WHILE) reduce 169) ((T_YIELD) reduce 169) ((T_YIELD_FROM) reduce 169) (("[") reduce 169) (("`") reduce 169) (("{") reduce 169) (("~") reduce 169)) 32885 ((("!") reduce 158) (("\"") reduce 158) (("$") reduce 158) (($) reduce 158) (("(") reduce 158) (("+") reduce 158) (("-") reduce 158) ((";") reduce 158) (("@") reduce 158) ((T_ABSTRACT) reduce 158) ((T_ARRAY) reduce 158) ((T_ARRAY_CAST) reduce 158) ((T_ATTRIBUTE) reduce 158) ((T_BOOL_CAST) reduce 158) ((T_BREAK) reduce 158) ((T_CLASS) reduce 158) ((T_CLASS_C) reduce 158) ((T_CLONE) reduce 158) ((T_CONST) reduce 158) ((T_CONSTANT_ENCAPSED_STRING) reduce 158) ((T_CONTINUE) reduce 158) ((T_DEC) reduce 158) ((T_DECLARE) reduce 158) ((T_DIR) reduce 158) ((T_DNUMBER) reduce 158) ((T_DO) reduce 158) ((T_DOUBLE_CAST) reduce 158) ((T_ECHO) reduce 158) ((T_ELSE) reduce 158) ((T_ELSEIF) reduce 158) ((T_EMPTY) reduce 158) ((T_ENUM) reduce 158) ((T_EVAL) reduce 158) ((T_EXIT) reduce 158) ((T_FILE) reduce 158) ((T_FINAL) reduce 158) ((T_FN) reduce 158) ((T_FOR) reduce 158) ((T_FOREACH) reduce 158) ((T_FUNCTION) reduce 158) ((T_FUNC_C) reduce 158) ((T_GLOBAL) reduce 158) ((T_GOTO) reduce 158) ((T_HALT_COMPILER) reduce 158) ((T_IF) reduce 158) ((T_INC) reduce 158) ((T_INCLUDE) reduce 158) ((T_INCLUDE_ONCE) reduce 158) ((T_INLINE_HTML) reduce 158) ((T_INTERFACE) reduce 158) ((T_INT_CAST) reduce 158) ((T_ISSET) reduce 158) ((T_LINE) reduce 158) ((T_LIST) reduce 158) ((T_LNUMBER) reduce 158) ((T_MATCH) reduce 158) ((T_METHOD_C) reduce 158) ((T_NAMESPACE) reduce 158) ((T_NAME_FULLY_QUALIFIED) reduce 158) ((T_NAME_QUALIFIED) reduce 158) ((T_NAME_RELATIVE) reduce 158) ((T_NEW) reduce 158) ((T_NS_C) reduce 158) ((T_OBJECT_CAST) reduce 158) ((T_PRINT) reduce 158) ((T_PROPERTY_C) reduce 158) ((T_READONLY) reduce 158) ((T_REQUIRE) reduce 158) ((T_REQUIRE_ONCE) reduce 158) ((T_RETURN) reduce 158) ((T_START_HEREDOC) reduce 158) ((T_STATIC) reduce 158) ((T_STRING) reduce 158) ((T_STRING_CAST) reduce 158) ((T_SWITCH) reduce 158) ((T_THROW) reduce 158) ((T_TRAIT) reduce 158) ((T_TRAIT_C) reduce 158) ((T_TRY) reduce 158) ((T_UNSET) reduce 158) ((T_UNSET_CAST) reduce 158) ((T_USE) reduce 158) ((T_VARIABLE) reduce 158) ((T_VOID_CAST) reduce 158) ((T_WHILE) reduce 158) ((T_YIELD) reduce 158) ((T_YIELD_FROM) reduce 158) (("[") reduce 158) (("`") reduce 158) (("{") reduce 158) (("~") reduce 158)) 32887 ((("!") reduce 156) (("\"") reduce 156) (("$") reduce 156) (($) reduce 156) (("(") reduce 156) (("+") reduce 156) (("-") reduce 156) ((";") reduce 156) (("@") reduce 156) ((T_ABSTRACT) reduce 156) ((T_ARRAY) reduce 156) ((T_ARRAY_CAST) reduce 156) ((T_ATTRIBUTE) reduce 156) ((T_BOOL_CAST) reduce 156) ((T_BREAK) reduce 156) ((T_CLASS) reduce 156) ((T_CLASS_C) reduce 156) ((T_CLONE) reduce 156) ((T_CONST) reduce 156) ((T_CONSTANT_ENCAPSED_STRING) reduce 156) ((T_CONTINUE) reduce 156) ((T_DEC) reduce 156) ((T_DECLARE) reduce 156) ((T_DIR) reduce 156) ((T_DNUMBER) reduce 156) ((T_DO) reduce 156) ((T_DOUBLE_CAST) reduce 156) ((T_ECHO) reduce 156) ((T_ELSE) reduce 156) ((T_ELSEIF) reduce 156) ((T_EMPTY) reduce 156) ((T_ENUM) reduce 156) ((T_EVAL) reduce 156) ((T_EXIT) reduce 156) ((T_FILE) reduce 156) ((T_FINAL) reduce 156) ((T_FN) reduce 156) ((T_FOR) reduce 156) ((T_FOREACH) reduce 156) ((T_FUNCTION) reduce 156) ((T_FUNC_C) reduce 156) ((T_GLOBAL) reduce 156) ((T_GOTO) reduce 156) ((T_HALT_COMPILER) reduce 156) ((T_IF) reduce 156) ((T_INC) reduce 156) ((T_INCLUDE) reduce 156) ((T_INCLUDE_ONCE) reduce 156) ((T_INLINE_HTML) reduce 156) ((T_INTERFACE) reduce 156) ((T_INT_CAST) reduce 156) ((T_ISSET) reduce 156) ((T_LINE) reduce 156) ((T_LIST) reduce 156) ((T_LNUMBER) reduce 156) ((T_MATCH) reduce 156) ((T_METHOD_C) reduce 156) ((T_NAMESPACE) reduce 156) ((T_NAME_FULLY_QUALIFIED) reduce 156) ((T_NAME_QUALIFIED) reduce 156) ((T_NAME_RELATIVE) reduce 156) ((T_NEW) reduce 156) ((T_NS_C) reduce 156) ((T_OBJECT_CAST) reduce 156) ((T_PRINT) reduce 156) ((T_PROPERTY_C) reduce 156) ((T_READONLY) reduce 156) ((T_REQUIRE) reduce 156) ((T_REQUIRE_ONCE) reduce 156) ((T_RETURN) reduce 156) ((T_START_HEREDOC) reduce 156) ((T_STATIC) reduce 156) ((T_STRING) reduce 156) ((T_STRING_CAST) reduce 156) ((T_SWITCH) reduce 156) ((T_THROW) reduce 156) ((T_TRAIT) reduce 156) ((T_TRAIT_C) reduce 156) ((T_TRY) reduce 156) ((T_UNSET) reduce 156) ((T_UNSET_CAST) reduce 156) ((T_USE) reduce 156) ((T_VARIABLE) reduce 156) ((T_VOID_CAST) reduce 156) ((T_WHILE) reduce 156) ((T_YIELD) reduce 156) ((T_YIELD_FROM) reduce 156) (("[") reduce 156) (("`") reduce 156) (("{") reduce 156) (("~") reduce 156)) 32891 ((("!") reduce 244) (("\"") reduce 244) (("$") reduce 244) (($) reduce 244) (("(") reduce 244) (("+") reduce 244) (("-") reduce 244) ((";") reduce 244) (("@") reduce 244) ((T_ABSTRACT) reduce 244) ((T_ARRAY) reduce 244) ((T_ARRAY_CAST) reduce 244) ((T_ATTRIBUTE) reduce 244) ((T_BOOL_CAST) reduce 244) ((T_BREAK) reduce 244) ((T_CLASS) reduce 244) ((T_CLASS_C) reduce 244) ((T_CLONE) reduce 244) ((T_CONST) reduce 244) ((T_CONSTANT_ENCAPSED_STRING) reduce 244) ((T_CONTINUE) reduce 244) ((T_DEC) reduce 244) ((T_DECLARE) reduce 244) ((T_DIR) reduce 244) ((T_DNUMBER) reduce 244) ((T_DO) reduce 244) ((T_DOUBLE_CAST) reduce 244) ((T_ECHO) reduce 244) ((T_ELSE) reduce 244) ((T_ELSEIF) reduce 244) ((T_EMPTY) reduce 244) ((T_ENUM) reduce 244) ((T_EVAL) reduce 244) ((T_EXIT) reduce 244) ((T_FILE) reduce 244) ((T_FINAL) reduce 244) ((T_FN) reduce 244) ((T_FOR) reduce 244) ((T_FOREACH) reduce 244) ((T_FUNCTION) reduce 244) ((T_FUNC_C) reduce 244) ((T_GLOBAL) reduce 244) ((T_GOTO) reduce 244) ((T_HALT_COMPILER) reduce 244) ((T_IF) reduce 244) ((T_INC) reduce 244) ((T_INCLUDE) reduce 244) ((T_INCLUDE_ONCE) reduce 244) ((T_INLINE_HTML) reduce 244) ((T_INTERFACE) reduce 244) ((T_INT_CAST) reduce 244) ((T_ISSET) reduce 244) ((T_LINE) reduce 244) ((T_LIST) reduce 244) ((T_LNUMBER) reduce 244) ((T_MATCH) reduce 244) ((T_METHOD_C) reduce 244) ((T_NAMESPACE) reduce 244) ((T_NAME_FULLY_QUALIFIED) reduce 244) ((T_NAME_QUALIFIED) reduce 244) ((T_NAME_RELATIVE) reduce 244) ((T_NEW) reduce 244) ((T_NS_C) reduce 244) ((T_OBJECT_CAST) reduce 244) ((T_PRINT) reduce 244) ((T_PROPERTY_C) reduce 244) ((T_READONLY) reduce 244) ((T_REQUIRE) reduce 244) ((T_REQUIRE_ONCE) reduce 244) ((T_RETURN) reduce 244) ((T_START_HEREDOC) reduce 244) ((T_STATIC) reduce 244) ((T_STRING) reduce 244) ((T_STRING_CAST) reduce 244) ((T_SWITCH) reduce 244) ((T_THROW) reduce 244) ((T_TRAIT) reduce 244) ((T_TRAIT_C) reduce 244) ((T_TRY) reduce 244) ((T_UNSET) reduce 244) ((T_UNSET_CAST) reduce 244) ((T_USE) reduce 244) ((T_VARIABLE) reduce 244) ((T_VOID_CAST) reduce 244) ((T_WHILE) reduce 244) ((T_YIELD) reduce 244) ((T_YIELD_FROM) reduce 244) (("[") reduce 244) (("`") reduce 244) (("{") reduce 244) (("~") reduce 244)) 32893 ((("!") reduce 168) (("\"") reduce 168) (("$") reduce 168) (($) reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONST) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_ELSE) reduce 168) ((T_ELSEIF) reduce 168) ((T_EMPTY) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAMESPACE) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_USE) reduce 168) ((T_VARIABLE) reduce 168) ((T_VOID_CAST) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 32895 ((("!") reduce 157) (("\"") reduce 157) (("$") reduce 157) (($) reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONST) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_ELSE) reduce 157) ((T_ELSEIF) reduce 157) ((T_EMPTY) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAMESPACE) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_USE) reduce 157) ((T_VARIABLE) reduce 157) ((T_VOID_CAST) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 32905 ((("!") reduce 164) (("\"") reduce 164) (("$") reduce 164) (($) reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONST) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_ELSE) reduce 164) ((T_ELSEIF) reduce 164) ((T_EMPTY) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAMESPACE) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_USE) reduce 164) ((T_VARIABLE) reduce 164) ((T_VOID_CAST) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 32906 ((("!") reduce 220) (("\"") reduce 220) (($) reduce 220) (("$") reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONST) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_ELSE) reduce 220) ((T_ELSEIF) reduce 220) ((T_EMPTY) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAMESPACE) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_USE) reduce 220) ((T_VARIABLE) reduce 220) ((T_VOID_CAST) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("~") reduce 220)) 32909 ((("!") reduce 221) (("\"") reduce 221) (("$") reduce 221) (($) reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONST) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_ELSE) reduce 221) ((T_ELSEIF) reduce 221) ((T_EMPTY) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAMESPACE) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_USE) reduce 221) ((T_VARIABLE) reduce 221) ((T_VOID_CAST) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("~") reduce 221)) 32910 ((("!") reduce 163) (("\"") reduce 163) (("$") reduce 163) (($) reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONST) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_ELSE) reduce 163) ((T_ELSEIF) reduce 163) ((T_EMPTY) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAMESPACE) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_USE) reduce 163) ((T_VARIABLE) reduce 163) ((T_VOID_CAST) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 32919 ((("!") reduce 152) (("\"") reduce 152) (("$") reduce 152) (($) reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONST) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_ELSE) reduce 152) ((T_ELSEIF) reduce 152) ((T_EMPTY) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAMESPACE) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_USE) reduce 152) ((T_VARIABLE) reduce 152) ((T_VOID_CAST) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 32920 ((("!") reduce 218) (("\"") reduce 218) (($) reduce 218) (("$") reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONST) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_ELSE) reduce 218) ((T_ELSEIF) reduce 218) ((T_EMPTY) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAMESPACE) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_USE) reduce 218) ((T_VARIABLE) reduce 218) ((T_VOID_CAST) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("~") reduce 218)) 32923 ((("!") reduce 219) (("\"") reduce 219) (("$") reduce 219) (($) reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONST) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_ELSE) reduce 219) ((T_ELSEIF) reduce 219) ((T_EMPTY) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAMESPACE) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_USE) reduce 219) ((T_VARIABLE) reduce 219) ((T_VOID_CAST) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("~") reduce 219)) 32925 ((("!") reduce 159) (("\"") reduce 159) (("$") reduce 159) (($) reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONST) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_ELSE) reduce 159) ((T_ELSEIF) reduce 159) ((T_EMPTY) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAMESPACE) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_USE) reduce 159) ((T_VARIABLE) reduce 159) ((T_VOID_CAST) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 32931 ((("!") reduce 151) (("\"") reduce 151) (("$") reduce 151) (($) reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONST) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_ELSE) reduce 151) ((T_ELSEIF) reduce 151) ((T_EMPTY) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAMESPACE) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_USE) reduce 151) ((T_VARIABLE) reduce 151) ((T_VOID_CAST) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 32936 ((("!") reduce 165) (("\"") reduce 165) (("$") reduce 165) (($) reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONST) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_ELSE) reduce 165) ((T_ELSEIF) reduce 165) ((T_EMPTY) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAMESPACE) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_USE) reduce 165) ((T_VARIABLE) reduce 165) ((T_VOID_CAST) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 32937 ((("!") reduce 222) (("\"") reduce 222) (($) reduce 222) (("$") reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONST) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_ELSE) reduce 222) ((T_ELSEIF) reduce 222) ((T_EMPTY) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAMESPACE) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_USE) reduce 222) ((T_VARIABLE) reduce 222) ((T_VOID_CAST) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("~") reduce 222)) 32940 ((("!") reduce 223) (("\"") reduce 223) (("$") reduce 223) (($) reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONST) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_ELSE) reduce 223) ((T_ELSEIF) reduce 223) ((T_EMPTY) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAMESPACE) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_USE) reduce 223) ((T_VARIABLE) reduce 223) ((T_VOID_CAST) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("~") reduce 223)) 32942 ((("!") reduce 155) (("\"") reduce 155) (("$") reduce 155) (($) reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONST) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_ELSE) reduce 155) ((T_ELSEIF) reduce 155) ((T_EMPTY) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAMESPACE) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_USE) reduce 155) ((T_VARIABLE) reduce 155) ((T_VOID_CAST) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 32944 ((("!") reduce 154) (("\"") reduce 154) (("$") reduce 154) (($) reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONST) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_ELSE) reduce 154) ((T_ELSEIF) reduce 154) ((T_EMPTY) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAMESPACE) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_USE) reduce 154) ((T_VARIABLE) reduce 154) ((T_VOID_CAST) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 32947 ((("!") reduce 113) (("\"") reduce 113) (($) reduce 113) (("$") reduce 113) (("(") reduce 113) (("+") reduce 113) (("-") reduce 113) ((";") reduce 113) (("@") reduce 113) ((T_ABSTRACT) reduce 113) ((T_ARRAY) reduce 113) ((T_ARRAY_CAST) reduce 113) ((T_ATTRIBUTE) reduce 113) ((T_BOOL_CAST) reduce 113) ((T_BREAK) reduce 113) ((T_CLASS) reduce 113) ((T_CLASS_C) reduce 113) ((T_CLONE) reduce 113) ((T_CONST) reduce 113) ((T_CONSTANT_ENCAPSED_STRING) reduce 113) ((T_CONTINUE) reduce 113) ((T_DEC) reduce 113) ((T_DECLARE) reduce 113) ((T_DIR) reduce 113) ((T_DNUMBER) reduce 113) ((T_DO) reduce 113) ((T_DOUBLE_CAST) reduce 113) ((T_ECHO) reduce 113) ((T_EMPTY) reduce 113) ((T_ENUM) reduce 113) ((T_EVAL) reduce 113) ((T_EXIT) reduce 113) ((T_FILE) reduce 113) ((T_FINAL) reduce 113) ((T_FN) reduce 113) ((T_FOR) reduce 113) ((T_FOREACH) reduce 113) ((T_FUNCTION) reduce 113) ((T_FUNC_C) reduce 113) ((T_GLOBAL) reduce 113) ((T_GOTO) reduce 113) ((T_HALT_COMPILER) reduce 113) ((T_IF) reduce 113) ((T_INC) reduce 113) ((T_INCLUDE) reduce 113) ((T_INCLUDE_ONCE) reduce 113) ((T_INLINE_HTML) reduce 113) ((T_INTERFACE) reduce 113) ((T_INT_CAST) reduce 113) ((T_ISSET) reduce 113) ((T_LINE) reduce 113) ((T_LIST) reduce 113) ((T_LNUMBER) reduce 113) ((T_MATCH) reduce 113) ((T_METHOD_C) reduce 113) ((T_NAMESPACE) reduce 113) ((T_NAME_FULLY_QUALIFIED) reduce 113) ((T_NAME_QUALIFIED) reduce 113) ((T_NAME_RELATIVE) reduce 113) ((T_NEW) reduce 113) ((T_NS_C) reduce 113) ((T_OBJECT_CAST) reduce 113) ((T_PRINT) reduce 113) ((T_PROPERTY_C) reduce 113) ((T_READONLY) reduce 113) ((T_REQUIRE) reduce 113) ((T_REQUIRE_ONCE) reduce 113) ((T_RETURN) reduce 113) ((T_START_HEREDOC) reduce 113) ((T_STATIC) reduce 113) ((T_STRING) reduce 113) ((T_STRING_CAST) reduce 113) ((T_SWITCH) reduce 113) ((T_THROW) reduce 113) ((T_TRAIT) reduce 113) ((T_TRAIT_C) reduce 113) ((T_TRY) reduce 113) ((T_UNSET) reduce 113) ((T_UNSET_CAST) reduce 113) ((T_USE) reduce 113) ((T_VARIABLE) reduce 113) ((T_VOID_CAST) reduce 113) ((T_WHILE) reduce 113) ((T_YIELD) reduce 113) ((T_YIELD_FROM) reduce 113) (("[") reduce 113) (("`") reduce 113) (("{") reduce 113) (("~") reduce 113)) 32949 ((("!") reduce 168) (("\"") reduce 168) (($) reduce 168) (("$") reduce 168) (("(") reduce 168) (("+") reduce 168) (("-") reduce 168) ((";") reduce 168) (("@") reduce 168) ((T_ABSTRACT) reduce 168) ((T_ARRAY) reduce 168) ((T_ARRAY_CAST) reduce 168) ((T_ATTRIBUTE) reduce 168) ((T_BOOL_CAST) reduce 168) ((T_BREAK) reduce 168) ((T_CLASS) reduce 168) ((T_CLASS_C) reduce 168) ((T_CLONE) reduce 168) ((T_CONST) reduce 168) ((T_CONSTANT_ENCAPSED_STRING) reduce 168) ((T_CONTINUE) reduce 168) ((T_DEC) reduce 168) ((T_DECLARE) reduce 168) ((T_DIR) reduce 168) ((T_DNUMBER) reduce 168) ((T_DO) reduce 168) ((T_DOUBLE_CAST) reduce 168) ((T_ECHO) reduce 168) ((T_EMPTY) reduce 168) ((T_ENUM) reduce 168) ((T_EVAL) reduce 168) ((T_EXIT) reduce 168) ((T_FILE) reduce 168) ((T_FINAL) reduce 168) ((T_FN) reduce 168) ((T_FOR) reduce 168) ((T_FOREACH) reduce 168) ((T_FUNCTION) reduce 168) ((T_FUNC_C) reduce 168) ((T_GLOBAL) reduce 168) ((T_GOTO) reduce 168) ((T_HALT_COMPILER) reduce 168) ((T_IF) reduce 168) ((T_INC) reduce 168) ((T_INCLUDE) reduce 168) ((T_INCLUDE_ONCE) reduce 168) ((T_INLINE_HTML) reduce 168) ((T_INTERFACE) reduce 168) ((T_INT_CAST) reduce 168) ((T_ISSET) reduce 168) ((T_LINE) reduce 168) ((T_LIST) reduce 168) ((T_LNUMBER) reduce 168) ((T_MATCH) reduce 168) ((T_METHOD_C) reduce 168) ((T_NAMESPACE) reduce 168) ((T_NAME_FULLY_QUALIFIED) reduce 168) ((T_NAME_QUALIFIED) reduce 168) ((T_NAME_RELATIVE) reduce 168) ((T_NEW) reduce 168) ((T_NS_C) reduce 168) ((T_OBJECT_CAST) reduce 168) ((T_PRINT) reduce 168) ((T_PROPERTY_C) reduce 168) ((T_READONLY) reduce 168) ((T_REQUIRE) reduce 168) ((T_REQUIRE_ONCE) reduce 168) ((T_RETURN) reduce 168) ((T_START_HEREDOC) reduce 168) ((T_STATIC) reduce 168) ((T_STRING) reduce 168) ((T_STRING_CAST) reduce 168) ((T_SWITCH) reduce 168) ((T_THROW) reduce 168) ((T_TRAIT) reduce 168) ((T_TRAIT_C) reduce 168) ((T_TRY) reduce 168) ((T_UNSET) reduce 168) ((T_UNSET_CAST) reduce 168) ((T_USE) reduce 168) ((T_VARIABLE) reduce 168) ((T_VOID_CAST) reduce 168) ((T_WHILE) reduce 168) ((T_YIELD) reduce 168) ((T_YIELD_FROM) reduce 168) (("[") reduce 168) (("`") reduce 168) (("{") reduce 168) (("~") reduce 168)) 32951 ((("!") reduce 157) (("\"") reduce 157) (($) reduce 157) (("$") reduce 157) (("(") reduce 157) (("+") reduce 157) (("-") reduce 157) ((";") reduce 157) (("@") reduce 157) ((T_ABSTRACT) reduce 157) ((T_ARRAY) reduce 157) ((T_ARRAY_CAST) reduce 157) ((T_ATTRIBUTE) reduce 157) ((T_BOOL_CAST) reduce 157) ((T_BREAK) reduce 157) ((T_CLASS) reduce 157) ((T_CLASS_C) reduce 157) ((T_CLONE) reduce 157) ((T_CONST) reduce 157) ((T_CONSTANT_ENCAPSED_STRING) reduce 157) ((T_CONTINUE) reduce 157) ((T_DEC) reduce 157) ((T_DECLARE) reduce 157) ((T_DIR) reduce 157) ((T_DNUMBER) reduce 157) ((T_DO) reduce 157) ((T_DOUBLE_CAST) reduce 157) ((T_ECHO) reduce 157) ((T_EMPTY) reduce 157) ((T_ENUM) reduce 157) ((T_EVAL) reduce 157) ((T_EXIT) reduce 157) ((T_FILE) reduce 157) ((T_FINAL) reduce 157) ((T_FN) reduce 157) ((T_FOR) reduce 157) ((T_FOREACH) reduce 157) ((T_FUNCTION) reduce 157) ((T_FUNC_C) reduce 157) ((T_GLOBAL) reduce 157) ((T_GOTO) reduce 157) ((T_HALT_COMPILER) reduce 157) ((T_IF) reduce 157) ((T_INC) reduce 157) ((T_INCLUDE) reduce 157) ((T_INCLUDE_ONCE) reduce 157) ((T_INLINE_HTML) reduce 157) ((T_INTERFACE) reduce 157) ((T_INT_CAST) reduce 157) ((T_ISSET) reduce 157) ((T_LINE) reduce 157) ((T_LIST) reduce 157) ((T_LNUMBER) reduce 157) ((T_MATCH) reduce 157) ((T_METHOD_C) reduce 157) ((T_NAMESPACE) reduce 157) ((T_NAME_FULLY_QUALIFIED) reduce 157) ((T_NAME_QUALIFIED) reduce 157) ((T_NAME_RELATIVE) reduce 157) ((T_NEW) reduce 157) ((T_NS_C) reduce 157) ((T_OBJECT_CAST) reduce 157) ((T_PRINT) reduce 157) ((T_PROPERTY_C) reduce 157) ((T_READONLY) reduce 157) ((T_REQUIRE) reduce 157) ((T_REQUIRE_ONCE) reduce 157) ((T_RETURN) reduce 157) ((T_START_HEREDOC) reduce 157) ((T_STATIC) reduce 157) ((T_STRING) reduce 157) ((T_STRING_CAST) reduce 157) ((T_SWITCH) reduce 157) ((T_THROW) reduce 157) ((T_TRAIT) reduce 157) ((T_TRAIT_C) reduce 157) ((T_TRY) reduce 157) ((T_UNSET) reduce 157) ((T_UNSET_CAST) reduce 157) ((T_USE) reduce 157) ((T_VARIABLE) reduce 157) ((T_VOID_CAST) reduce 157) ((T_WHILE) reduce 157) ((T_YIELD) reduce 157) ((T_YIELD_FROM) reduce 157) (("[") reduce 157) (("`") reduce 157) (("{") reduce 157) (("~") reduce 157)) 32961 ((("!") reduce 164) (("\"") reduce 164) (($) reduce 164) (("$") reduce 164) (("(") reduce 164) (("+") reduce 164) (("-") reduce 164) ((";") reduce 164) (("@") reduce 164) ((T_ABSTRACT) reduce 164) ((T_ARRAY) reduce 164) ((T_ARRAY_CAST) reduce 164) ((T_ATTRIBUTE) reduce 164) ((T_BOOL_CAST) reduce 164) ((T_BREAK) reduce 164) ((T_CLASS) reduce 164) ((T_CLASS_C) reduce 164) ((T_CLONE) reduce 164) ((T_CONST) reduce 164) ((T_CONSTANT_ENCAPSED_STRING) reduce 164) ((T_CONTINUE) reduce 164) ((T_DEC) reduce 164) ((T_DECLARE) reduce 164) ((T_DIR) reduce 164) ((T_DNUMBER) reduce 164) ((T_DO) reduce 164) ((T_DOUBLE_CAST) reduce 164) ((T_ECHO) reduce 164) ((T_EMPTY) reduce 164) ((T_ENUM) reduce 164) ((T_EVAL) reduce 164) ((T_EXIT) reduce 164) ((T_FILE) reduce 164) ((T_FINAL) reduce 164) ((T_FN) reduce 164) ((T_FOR) reduce 164) ((T_FOREACH) reduce 164) ((T_FUNCTION) reduce 164) ((T_FUNC_C) reduce 164) ((T_GLOBAL) reduce 164) ((T_GOTO) reduce 164) ((T_HALT_COMPILER) reduce 164) ((T_IF) reduce 164) ((T_INC) reduce 164) ((T_INCLUDE) reduce 164) ((T_INCLUDE_ONCE) reduce 164) ((T_INLINE_HTML) reduce 164) ((T_INTERFACE) reduce 164) ((T_INT_CAST) reduce 164) ((T_ISSET) reduce 164) ((T_LINE) reduce 164) ((T_LIST) reduce 164) ((T_LNUMBER) reduce 164) ((T_MATCH) reduce 164) ((T_METHOD_C) reduce 164) ((T_NAMESPACE) reduce 164) ((T_NAME_FULLY_QUALIFIED) reduce 164) ((T_NAME_QUALIFIED) reduce 164) ((T_NAME_RELATIVE) reduce 164) ((T_NEW) reduce 164) ((T_NS_C) reduce 164) ((T_OBJECT_CAST) reduce 164) ((T_PRINT) reduce 164) ((T_PROPERTY_C) reduce 164) ((T_READONLY) reduce 164) ((T_REQUIRE) reduce 164) ((T_REQUIRE_ONCE) reduce 164) ((T_RETURN) reduce 164) ((T_START_HEREDOC) reduce 164) ((T_STATIC) reduce 164) ((T_STRING) reduce 164) ((T_STRING_CAST) reduce 164) ((T_SWITCH) reduce 164) ((T_THROW) reduce 164) ((T_TRAIT) reduce 164) ((T_TRAIT_C) reduce 164) ((T_TRY) reduce 164) ((T_UNSET) reduce 164) ((T_UNSET_CAST) reduce 164) ((T_USE) reduce 164) ((T_VARIABLE) reduce 164) ((T_VOID_CAST) reduce 164) ((T_WHILE) reduce 164) ((T_YIELD) reduce 164) ((T_YIELD_FROM) reduce 164) (("[") reduce 164) (("`") reduce 164) (("{") reduce 164) (("~") reduce 164)) 32962 ((("!") reduce 220) (("\"") reduce 220) (("$") reduce 220) (($) reduce 220) (("(") reduce 220) (("+") reduce 220) (("-") reduce 220) ((";") reduce 220) (("@") reduce 220) ((T_ABSTRACT) reduce 220) ((T_ARRAY) reduce 220) ((T_ARRAY_CAST) reduce 220) ((T_ATTRIBUTE) reduce 220) ((T_BOOL_CAST) reduce 220) ((T_BREAK) reduce 220) ((T_CLASS) reduce 220) ((T_CLASS_C) reduce 220) ((T_CLONE) reduce 220) ((T_CONST) reduce 220) ((T_CONSTANT_ENCAPSED_STRING) reduce 220) ((T_CONTINUE) reduce 220) ((T_DEC) reduce 220) ((T_DECLARE) reduce 220) ((T_DIR) reduce 220) ((T_DNUMBER) reduce 220) ((T_DO) reduce 220) ((T_DOUBLE_CAST) reduce 220) ((T_ECHO) reduce 220) ((T_EMPTY) reduce 220) ((T_ENUM) reduce 220) ((T_EVAL) reduce 220) ((T_EXIT) reduce 220) ((T_FILE) reduce 220) ((T_FINAL) reduce 220) ((T_FN) reduce 220) ((T_FOR) reduce 220) ((T_FOREACH) reduce 220) ((T_FUNCTION) reduce 220) ((T_FUNC_C) reduce 220) ((T_GLOBAL) reduce 220) ((T_GOTO) reduce 220) ((T_HALT_COMPILER) reduce 220) ((T_IF) reduce 220) ((T_INC) reduce 220) ((T_INCLUDE) reduce 220) ((T_INCLUDE_ONCE) reduce 220) ((T_INLINE_HTML) reduce 220) ((T_INTERFACE) reduce 220) ((T_INT_CAST) reduce 220) ((T_ISSET) reduce 220) ((T_LINE) reduce 220) ((T_LIST) reduce 220) ((T_LNUMBER) reduce 220) ((T_MATCH) reduce 220) ((T_METHOD_C) reduce 220) ((T_NAMESPACE) reduce 220) ((T_NAME_FULLY_QUALIFIED) reduce 220) ((T_NAME_QUALIFIED) reduce 220) ((T_NAME_RELATIVE) reduce 220) ((T_NEW) reduce 220) ((T_NS_C) reduce 220) ((T_OBJECT_CAST) reduce 220) ((T_PRINT) reduce 220) ((T_PROPERTY_C) reduce 220) ((T_READONLY) reduce 220) ((T_REQUIRE) reduce 220) ((T_REQUIRE_ONCE) reduce 220) ((T_RETURN) reduce 220) ((T_START_HEREDOC) reduce 220) ((T_STATIC) reduce 220) ((T_STRING) reduce 220) ((T_STRING_CAST) reduce 220) ((T_SWITCH) reduce 220) ((T_THROW) reduce 220) ((T_TRAIT) reduce 220) ((T_TRAIT_C) reduce 220) ((T_TRY) reduce 220) ((T_UNSET) reduce 220) ((T_UNSET_CAST) reduce 220) ((T_USE) reduce 220) ((T_VARIABLE) reduce 220) ((T_VOID_CAST) reduce 220) ((T_WHILE) reduce 220) ((T_YIELD) reduce 220) ((T_YIELD_FROM) reduce 220) (("[") reduce 220) (("`") reduce 220) (("{") reduce 220) (("~") reduce 220)) 32965 ((("!") reduce 221) (("\"") reduce 221) (($) reduce 221) (("$") reduce 221) (("(") reduce 221) (("+") reduce 221) (("-") reduce 221) ((";") reduce 221) (("@") reduce 221) ((T_ABSTRACT) reduce 221) ((T_ARRAY) reduce 221) ((T_ARRAY_CAST) reduce 221) ((T_ATTRIBUTE) reduce 221) ((T_BOOL_CAST) reduce 221) ((T_BREAK) reduce 221) ((T_CLASS) reduce 221) ((T_CLASS_C) reduce 221) ((T_CLONE) reduce 221) ((T_CONST) reduce 221) ((T_CONSTANT_ENCAPSED_STRING) reduce 221) ((T_CONTINUE) reduce 221) ((T_DEC) reduce 221) ((T_DECLARE) reduce 221) ((T_DIR) reduce 221) ((T_DNUMBER) reduce 221) ((T_DO) reduce 221) ((T_DOUBLE_CAST) reduce 221) ((T_ECHO) reduce 221) ((T_EMPTY) reduce 221) ((T_ENUM) reduce 221) ((T_EVAL) reduce 221) ((T_EXIT) reduce 221) ((T_FILE) reduce 221) ((T_FINAL) reduce 221) ((T_FN) reduce 221) ((T_FOR) reduce 221) ((T_FOREACH) reduce 221) ((T_FUNCTION) reduce 221) ((T_FUNC_C) reduce 221) ((T_GLOBAL) reduce 221) ((T_GOTO) reduce 221) ((T_HALT_COMPILER) reduce 221) ((T_IF) reduce 221) ((T_INC) reduce 221) ((T_INCLUDE) reduce 221) ((T_INCLUDE_ONCE) reduce 221) ((T_INLINE_HTML) reduce 221) ((T_INTERFACE) reduce 221) ((T_INT_CAST) reduce 221) ((T_ISSET) reduce 221) ((T_LINE) reduce 221) ((T_LIST) reduce 221) ((T_LNUMBER) reduce 221) ((T_MATCH) reduce 221) ((T_METHOD_C) reduce 221) ((T_NAMESPACE) reduce 221) ((T_NAME_FULLY_QUALIFIED) reduce 221) ((T_NAME_QUALIFIED) reduce 221) ((T_NAME_RELATIVE) reduce 221) ((T_NEW) reduce 221) ((T_NS_C) reduce 221) ((T_OBJECT_CAST) reduce 221) ((T_PRINT) reduce 221) ((T_PROPERTY_C) reduce 221) ((T_READONLY) reduce 221) ((T_REQUIRE) reduce 221) ((T_REQUIRE_ONCE) reduce 221) ((T_RETURN) reduce 221) ((T_START_HEREDOC) reduce 221) ((T_STATIC) reduce 221) ((T_STRING) reduce 221) ((T_STRING_CAST) reduce 221) ((T_SWITCH) reduce 221) ((T_THROW) reduce 221) ((T_TRAIT) reduce 221) ((T_TRAIT_C) reduce 221) ((T_TRY) reduce 221) ((T_UNSET) reduce 221) ((T_UNSET_CAST) reduce 221) ((T_USE) reduce 221) ((T_VARIABLE) reduce 221) ((T_VOID_CAST) reduce 221) ((T_WHILE) reduce 221) ((T_YIELD) reduce 221) ((T_YIELD_FROM) reduce 221) (("[") reduce 221) (("`") reduce 221) (("{") reduce 221) (("~") reduce 221)) 32966 ((("!") reduce 163) (("\"") reduce 163) (($) reduce 163) (("$") reduce 163) (("(") reduce 163) (("+") reduce 163) (("-") reduce 163) ((";") reduce 163) (("@") reduce 163) ((T_ABSTRACT) reduce 163) ((T_ARRAY) reduce 163) ((T_ARRAY_CAST) reduce 163) ((T_ATTRIBUTE) reduce 163) ((T_BOOL_CAST) reduce 163) ((T_BREAK) reduce 163) ((T_CLASS) reduce 163) ((T_CLASS_C) reduce 163) ((T_CLONE) reduce 163) ((T_CONST) reduce 163) ((T_CONSTANT_ENCAPSED_STRING) reduce 163) ((T_CONTINUE) reduce 163) ((T_DEC) reduce 163) ((T_DECLARE) reduce 163) ((T_DIR) reduce 163) ((T_DNUMBER) reduce 163) ((T_DO) reduce 163) ((T_DOUBLE_CAST) reduce 163) ((T_ECHO) reduce 163) ((T_EMPTY) reduce 163) ((T_ENUM) reduce 163) ((T_EVAL) reduce 163) ((T_EXIT) reduce 163) ((T_FILE) reduce 163) ((T_FINAL) reduce 163) ((T_FN) reduce 163) ((T_FOR) reduce 163) ((T_FOREACH) reduce 163) ((T_FUNCTION) reduce 163) ((T_FUNC_C) reduce 163) ((T_GLOBAL) reduce 163) ((T_GOTO) reduce 163) ((T_HALT_COMPILER) reduce 163) ((T_IF) reduce 163) ((T_INC) reduce 163) ((T_INCLUDE) reduce 163) ((T_INCLUDE_ONCE) reduce 163) ((T_INLINE_HTML) reduce 163) ((T_INTERFACE) reduce 163) ((T_INT_CAST) reduce 163) ((T_ISSET) reduce 163) ((T_LINE) reduce 163) ((T_LIST) reduce 163) ((T_LNUMBER) reduce 163) ((T_MATCH) reduce 163) ((T_METHOD_C) reduce 163) ((T_NAMESPACE) reduce 163) ((T_NAME_FULLY_QUALIFIED) reduce 163) ((T_NAME_QUALIFIED) reduce 163) ((T_NAME_RELATIVE) reduce 163) ((T_NEW) reduce 163) ((T_NS_C) reduce 163) ((T_OBJECT_CAST) reduce 163) ((T_PRINT) reduce 163) ((T_PROPERTY_C) reduce 163) ((T_READONLY) reduce 163) ((T_REQUIRE) reduce 163) ((T_REQUIRE_ONCE) reduce 163) ((T_RETURN) reduce 163) ((T_START_HEREDOC) reduce 163) ((T_STATIC) reduce 163) ((T_STRING) reduce 163) ((T_STRING_CAST) reduce 163) ((T_SWITCH) reduce 163) ((T_THROW) reduce 163) ((T_TRAIT) reduce 163) ((T_TRAIT_C) reduce 163) ((T_TRY) reduce 163) ((T_UNSET) reduce 163) ((T_UNSET_CAST) reduce 163) ((T_USE) reduce 163) ((T_VARIABLE) reduce 163) ((T_VOID_CAST) reduce 163) ((T_WHILE) reduce 163) ((T_YIELD) reduce 163) ((T_YIELD_FROM) reduce 163) (("[") reduce 163) (("`") reduce 163) (("{") reduce 163) (("~") reduce 163)) 32975 ((("!") reduce 152) (("\"") reduce 152) (($) reduce 152) (("$") reduce 152) (("(") reduce 152) (("+") reduce 152) (("-") reduce 152) ((";") reduce 152) (("@") reduce 152) ((T_ABSTRACT) reduce 152) ((T_ARRAY) reduce 152) ((T_ARRAY_CAST) reduce 152) ((T_ATTRIBUTE) reduce 152) ((T_BOOL_CAST) reduce 152) ((T_BREAK) reduce 152) ((T_CLASS) reduce 152) ((T_CLASS_C) reduce 152) ((T_CLONE) reduce 152) ((T_CONST) reduce 152) ((T_CONSTANT_ENCAPSED_STRING) reduce 152) ((T_CONTINUE) reduce 152) ((T_DEC) reduce 152) ((T_DECLARE) reduce 152) ((T_DIR) reduce 152) ((T_DNUMBER) reduce 152) ((T_DO) reduce 152) ((T_DOUBLE_CAST) reduce 152) ((T_ECHO) reduce 152) ((T_EMPTY) reduce 152) ((T_ENUM) reduce 152) ((T_EVAL) reduce 152) ((T_EXIT) reduce 152) ((T_FILE) reduce 152) ((T_FINAL) reduce 152) ((T_FN) reduce 152) ((T_FOR) reduce 152) ((T_FOREACH) reduce 152) ((T_FUNCTION) reduce 152) ((T_FUNC_C) reduce 152) ((T_GLOBAL) reduce 152) ((T_GOTO) reduce 152) ((T_HALT_COMPILER) reduce 152) ((T_IF) reduce 152) ((T_INC) reduce 152) ((T_INCLUDE) reduce 152) ((T_INCLUDE_ONCE) reduce 152) ((T_INLINE_HTML) reduce 152) ((T_INTERFACE) reduce 152) ((T_INT_CAST) reduce 152) ((T_ISSET) reduce 152) ((T_LINE) reduce 152) ((T_LIST) reduce 152) ((T_LNUMBER) reduce 152) ((T_MATCH) reduce 152) ((T_METHOD_C) reduce 152) ((T_NAMESPACE) reduce 152) ((T_NAME_FULLY_QUALIFIED) reduce 152) ((T_NAME_QUALIFIED) reduce 152) ((T_NAME_RELATIVE) reduce 152) ((T_NEW) reduce 152) ((T_NS_C) reduce 152) ((T_OBJECT_CAST) reduce 152) ((T_PRINT) reduce 152) ((T_PROPERTY_C) reduce 152) ((T_READONLY) reduce 152) ((T_REQUIRE) reduce 152) ((T_REQUIRE_ONCE) reduce 152) ((T_RETURN) reduce 152) ((T_START_HEREDOC) reduce 152) ((T_STATIC) reduce 152) ((T_STRING) reduce 152) ((T_STRING_CAST) reduce 152) ((T_SWITCH) reduce 152) ((T_THROW) reduce 152) ((T_TRAIT) reduce 152) ((T_TRAIT_C) reduce 152) ((T_TRY) reduce 152) ((T_UNSET) reduce 152) ((T_UNSET_CAST) reduce 152) ((T_USE) reduce 152) ((T_VARIABLE) reduce 152) ((T_VOID_CAST) reduce 152) ((T_WHILE) reduce 152) ((T_YIELD) reduce 152) ((T_YIELD_FROM) reduce 152) (("[") reduce 152) (("`") reduce 152) (("{") reduce 152) (("~") reduce 152)) 32976 ((("!") reduce 218) (("\"") reduce 218) (("$") reduce 218) (($) reduce 218) (("(") reduce 218) (("+") reduce 218) (("-") reduce 218) ((";") reduce 218) (("@") reduce 218) ((T_ABSTRACT) reduce 218) ((T_ARRAY) reduce 218) ((T_ARRAY_CAST) reduce 218) ((T_ATTRIBUTE) reduce 218) ((T_BOOL_CAST) reduce 218) ((T_BREAK) reduce 218) ((T_CLASS) reduce 218) ((T_CLASS_C) reduce 218) ((T_CLONE) reduce 218) ((T_CONST) reduce 218) ((T_CONSTANT_ENCAPSED_STRING) reduce 218) ((T_CONTINUE) reduce 218) ((T_DEC) reduce 218) ((T_DECLARE) reduce 218) ((T_DIR) reduce 218) ((T_DNUMBER) reduce 218) ((T_DO) reduce 218) ((T_DOUBLE_CAST) reduce 218) ((T_ECHO) reduce 218) ((T_EMPTY) reduce 218) ((T_ENUM) reduce 218) ((T_EVAL) reduce 218) ((T_EXIT) reduce 218) ((T_FILE) reduce 218) ((T_FINAL) reduce 218) ((T_FN) reduce 218) ((T_FOR) reduce 218) ((T_FOREACH) reduce 218) ((T_FUNCTION) reduce 218) ((T_FUNC_C) reduce 218) ((T_GLOBAL) reduce 218) ((T_GOTO) reduce 218) ((T_HALT_COMPILER) reduce 218) ((T_IF) reduce 218) ((T_INC) reduce 218) ((T_INCLUDE) reduce 218) ((T_INCLUDE_ONCE) reduce 218) ((T_INLINE_HTML) reduce 218) ((T_INTERFACE) reduce 218) ((T_INT_CAST) reduce 218) ((T_ISSET) reduce 218) ((T_LINE) reduce 218) ((T_LIST) reduce 218) ((T_LNUMBER) reduce 218) ((T_MATCH) reduce 218) ((T_METHOD_C) reduce 218) ((T_NAMESPACE) reduce 218) ((T_NAME_FULLY_QUALIFIED) reduce 218) ((T_NAME_QUALIFIED) reduce 218) ((T_NAME_RELATIVE) reduce 218) ((T_NEW) reduce 218) ((T_NS_C) reduce 218) ((T_OBJECT_CAST) reduce 218) ((T_PRINT) reduce 218) ((T_PROPERTY_C) reduce 218) ((T_READONLY) reduce 218) ((T_REQUIRE) reduce 218) ((T_REQUIRE_ONCE) reduce 218) ((T_RETURN) reduce 218) ((T_START_HEREDOC) reduce 218) ((T_STATIC) reduce 218) ((T_STRING) reduce 218) ((T_STRING_CAST) reduce 218) ((T_SWITCH) reduce 218) ((T_THROW) reduce 218) ((T_TRAIT) reduce 218) ((T_TRAIT_C) reduce 218) ((T_TRY) reduce 218) ((T_UNSET) reduce 218) ((T_UNSET_CAST) reduce 218) ((T_USE) reduce 218) ((T_VARIABLE) reduce 218) ((T_VOID_CAST) reduce 218) ((T_WHILE) reduce 218) ((T_YIELD) reduce 218) ((T_YIELD_FROM) reduce 218) (("[") reduce 218) (("`") reduce 218) (("{") reduce 218) (("~") reduce 218)) 32979 ((("!") reduce 219) (("\"") reduce 219) (($) reduce 219) (("$") reduce 219) (("(") reduce 219) (("+") reduce 219) (("-") reduce 219) ((";") reduce 219) (("@") reduce 219) ((T_ABSTRACT) reduce 219) ((T_ARRAY) reduce 219) ((T_ARRAY_CAST) reduce 219) ((T_ATTRIBUTE) reduce 219) ((T_BOOL_CAST) reduce 219) ((T_BREAK) reduce 219) ((T_CLASS) reduce 219) ((T_CLASS_C) reduce 219) ((T_CLONE) reduce 219) ((T_CONST) reduce 219) ((T_CONSTANT_ENCAPSED_STRING) reduce 219) ((T_CONTINUE) reduce 219) ((T_DEC) reduce 219) ((T_DECLARE) reduce 219) ((T_DIR) reduce 219) ((T_DNUMBER) reduce 219) ((T_DO) reduce 219) ((T_DOUBLE_CAST) reduce 219) ((T_ECHO) reduce 219) ((T_EMPTY) reduce 219) ((T_ENUM) reduce 219) ((T_EVAL) reduce 219) ((T_EXIT) reduce 219) ((T_FILE) reduce 219) ((T_FINAL) reduce 219) ((T_FN) reduce 219) ((T_FOR) reduce 219) ((T_FOREACH) reduce 219) ((T_FUNCTION) reduce 219) ((T_FUNC_C) reduce 219) ((T_GLOBAL) reduce 219) ((T_GOTO) reduce 219) ((T_HALT_COMPILER) reduce 219) ((T_IF) reduce 219) ((T_INC) reduce 219) ((T_INCLUDE) reduce 219) ((T_INCLUDE_ONCE) reduce 219) ((T_INLINE_HTML) reduce 219) ((T_INTERFACE) reduce 219) ((T_INT_CAST) reduce 219) ((T_ISSET) reduce 219) ((T_LINE) reduce 219) ((T_LIST) reduce 219) ((T_LNUMBER) reduce 219) ((T_MATCH) reduce 219) ((T_METHOD_C) reduce 219) ((T_NAMESPACE) reduce 219) ((T_NAME_FULLY_QUALIFIED) reduce 219) ((T_NAME_QUALIFIED) reduce 219) ((T_NAME_RELATIVE) reduce 219) ((T_NEW) reduce 219) ((T_NS_C) reduce 219) ((T_OBJECT_CAST) reduce 219) ((T_PRINT) reduce 219) ((T_PROPERTY_C) reduce 219) ((T_READONLY) reduce 219) ((T_REQUIRE) reduce 219) ((T_REQUIRE_ONCE) reduce 219) ((T_RETURN) reduce 219) ((T_START_HEREDOC) reduce 219) ((T_STATIC) reduce 219) ((T_STRING) reduce 219) ((T_STRING_CAST) reduce 219) ((T_SWITCH) reduce 219) ((T_THROW) reduce 219) ((T_TRAIT) reduce 219) ((T_TRAIT_C) reduce 219) ((T_TRY) reduce 219) ((T_UNSET) reduce 219) ((T_UNSET_CAST) reduce 219) ((T_USE) reduce 219) ((T_VARIABLE) reduce 219) ((T_VOID_CAST) reduce 219) ((T_WHILE) reduce 219) ((T_YIELD) reduce 219) ((T_YIELD_FROM) reduce 219) (("[") reduce 219) (("`") reduce 219) (("{") reduce 219) (("~") reduce 219)) 32998 ((("!") reduce 202) (("\"") reduce 202) (("$") reduce 202) (($) reduce 202) (("(") reduce 202) (("+") reduce 202) (("-") reduce 202) ((";") reduce 202) (("@") reduce 202) ((T_ABSTRACT) reduce 202) ((T_ARRAY) reduce 202) ((T_ARRAY_CAST) reduce 202) ((T_ATTRIBUTE) reduce 202) ((T_BOOL_CAST) reduce 202) ((T_BREAK) reduce 202) ((T_CLASS) reduce 202) ((T_CLASS_C) reduce 202) ((T_CLONE) reduce 202) ((T_CONST) reduce 202) ((T_CONSTANT_ENCAPSED_STRING) reduce 202) ((T_CONTINUE) reduce 202) ((T_DEC) reduce 202) ((T_DECLARE) reduce 202) ((T_DIR) reduce 202) ((T_DNUMBER) reduce 202) ((T_DO) reduce 202) ((T_DOUBLE_CAST) reduce 202) ((T_ECHO) reduce 202) ((T_EMPTY) reduce 202) ((T_ENUM) reduce 202) ((T_EVAL) reduce 202) ((T_EXIT) reduce 202) ((T_FILE) reduce 202) ((T_FINAL) reduce 202) ((T_FN) reduce 202) ((T_FOR) reduce 202) ((T_FOREACH) reduce 202) ((T_FUNCTION) reduce 202) ((T_FUNC_C) reduce 202) ((T_GLOBAL) reduce 202) ((T_GOTO) reduce 202) ((T_HALT_COMPILER) reduce 202) ((T_IF) reduce 202) ((T_INC) reduce 202) ((T_INCLUDE) reduce 202) ((T_INCLUDE_ONCE) reduce 202) ((T_INLINE_HTML) reduce 202) ((T_INTERFACE) reduce 202) ((T_INT_CAST) reduce 202) ((T_ISSET) reduce 202) ((T_LINE) reduce 202) ((T_LIST) reduce 202) ((T_LNUMBER) reduce 202) ((T_MATCH) reduce 202) ((T_METHOD_C) reduce 202) ((T_NAMESPACE) reduce 202) ((T_NAME_FULLY_QUALIFIED) reduce 202) ((T_NAME_QUALIFIED) reduce 202) ((T_NAME_RELATIVE) reduce 202) ((T_NEW) reduce 202) ((T_NS_C) reduce 202) ((T_OBJECT_CAST) reduce 202) ((T_PRINT) reduce 202) ((T_PROPERTY_C) reduce 202) ((T_READONLY) reduce 202) ((T_REQUIRE) reduce 202) ((T_REQUIRE_ONCE) reduce 202) ((T_RETURN) reduce 202) ((T_START_HEREDOC) reduce 202) ((T_STATIC) reduce 202) ((T_STRING) reduce 202) ((T_STRING_CAST) reduce 202) ((T_SWITCH) reduce 202) ((T_THROW) reduce 202) ((T_TRAIT) reduce 202) ((T_TRAIT_C) reduce 202) ((T_TRY) reduce 202) ((T_UNSET) reduce 202) ((T_UNSET_CAST) reduce 202) ((T_USE) reduce 202) ((T_VARIABLE) reduce 202) ((T_VOID_CAST) reduce 202) ((T_WHILE) reduce 202) ((T_YIELD) reduce 202) ((T_YIELD_FROM) reduce 202) (("[") reduce 202) (("`") reduce 202) (("{") reduce 202) (("~") reduce 202)) 33003 ((("!") reduce 159) (("\"") reduce 159) (($) reduce 159) (("$") reduce 159) (("(") reduce 159) (("+") reduce 159) (("-") reduce 159) ((";") reduce 159) (("@") reduce 159) ((T_ABSTRACT) reduce 159) ((T_ARRAY) reduce 159) ((T_ARRAY_CAST) reduce 159) ((T_ATTRIBUTE) reduce 159) ((T_BOOL_CAST) reduce 159) ((T_BREAK) reduce 159) ((T_CLASS) reduce 159) ((T_CLASS_C) reduce 159) ((T_CLONE) reduce 159) ((T_CONST) reduce 159) ((T_CONSTANT_ENCAPSED_STRING) reduce 159) ((T_CONTINUE) reduce 159) ((T_DEC) reduce 159) ((T_DECLARE) reduce 159) ((T_DIR) reduce 159) ((T_DNUMBER) reduce 159) ((T_DO) reduce 159) ((T_DOUBLE_CAST) reduce 159) ((T_ECHO) reduce 159) ((T_EMPTY) reduce 159) ((T_ENUM) reduce 159) ((T_EVAL) reduce 159) ((T_EXIT) reduce 159) ((T_FILE) reduce 159) ((T_FINAL) reduce 159) ((T_FN) reduce 159) ((T_FOR) reduce 159) ((T_FOREACH) reduce 159) ((T_FUNCTION) reduce 159) ((T_FUNC_C) reduce 159) ((T_GLOBAL) reduce 159) ((T_GOTO) reduce 159) ((T_HALT_COMPILER) reduce 159) ((T_IF) reduce 159) ((T_INC) reduce 159) ((T_INCLUDE) reduce 159) ((T_INCLUDE_ONCE) reduce 159) ((T_INLINE_HTML) reduce 159) ((T_INTERFACE) reduce 159) ((T_INT_CAST) reduce 159) ((T_ISSET) reduce 159) ((T_LINE) reduce 159) ((T_LIST) reduce 159) ((T_LNUMBER) reduce 159) ((T_MATCH) reduce 159) ((T_METHOD_C) reduce 159) ((T_NAMESPACE) reduce 159) ((T_NAME_FULLY_QUALIFIED) reduce 159) ((T_NAME_QUALIFIED) reduce 159) ((T_NAME_RELATIVE) reduce 159) ((T_NEW) reduce 159) ((T_NS_C) reduce 159) ((T_OBJECT_CAST) reduce 159) ((T_PRINT) reduce 159) ((T_PROPERTY_C) reduce 159) ((T_READONLY) reduce 159) ((T_REQUIRE) reduce 159) ((T_REQUIRE_ONCE) reduce 159) ((T_RETURN) reduce 159) ((T_START_HEREDOC) reduce 159) ((T_STATIC) reduce 159) ((T_STRING) reduce 159) ((T_STRING_CAST) reduce 159) ((T_SWITCH) reduce 159) ((T_THROW) reduce 159) ((T_TRAIT) reduce 159) ((T_TRAIT_C) reduce 159) ((T_TRY) reduce 159) ((T_UNSET) reduce 159) ((T_UNSET_CAST) reduce 159) ((T_USE) reduce 159) ((T_VARIABLE) reduce 159) ((T_VOID_CAST) reduce 159) ((T_WHILE) reduce 159) ((T_YIELD) reduce 159) ((T_YIELD_FROM) reduce 159) (("[") reduce 159) (("`") reduce 159) (("{") reduce 159) (("~") reduce 159)) 33010 ((("!") reduce 151) (("\"") reduce 151) (($) reduce 151) (("$") reduce 151) (("(") reduce 151) (("+") reduce 151) (("-") reduce 151) ((";") reduce 151) (("@") reduce 151) ((T_ABSTRACT) reduce 151) ((T_ARRAY) reduce 151) ((T_ARRAY_CAST) reduce 151) ((T_ATTRIBUTE) reduce 151) ((T_BOOL_CAST) reduce 151) ((T_BREAK) reduce 151) ((T_CLASS) reduce 151) ((T_CLASS_C) reduce 151) ((T_CLONE) reduce 151) ((T_CONST) reduce 151) ((T_CONSTANT_ENCAPSED_STRING) reduce 151) ((T_CONTINUE) reduce 151) ((T_DEC) reduce 151) ((T_DECLARE) reduce 151) ((T_DIR) reduce 151) ((T_DNUMBER) reduce 151) ((T_DO) reduce 151) ((T_DOUBLE_CAST) reduce 151) ((T_ECHO) reduce 151) ((T_EMPTY) reduce 151) ((T_ENUM) reduce 151) ((T_EVAL) reduce 151) ((T_EXIT) reduce 151) ((T_FILE) reduce 151) ((T_FINAL) reduce 151) ((T_FN) reduce 151) ((T_FOR) reduce 151) ((T_FOREACH) reduce 151) ((T_FUNCTION) reduce 151) ((T_FUNC_C) reduce 151) ((T_GLOBAL) reduce 151) ((T_GOTO) reduce 151) ((T_HALT_COMPILER) reduce 151) ((T_IF) reduce 151) ((T_INC) reduce 151) ((T_INCLUDE) reduce 151) ((T_INCLUDE_ONCE) reduce 151) ((T_INLINE_HTML) reduce 151) ((T_INTERFACE) reduce 151) ((T_INT_CAST) reduce 151) ((T_ISSET) reduce 151) ((T_LINE) reduce 151) ((T_LIST) reduce 151) ((T_LNUMBER) reduce 151) ((T_MATCH) reduce 151) ((T_METHOD_C) reduce 151) ((T_NAMESPACE) reduce 151) ((T_NAME_FULLY_QUALIFIED) reduce 151) ((T_NAME_QUALIFIED) reduce 151) ((T_NAME_RELATIVE) reduce 151) ((T_NEW) reduce 151) ((T_NS_C) reduce 151) ((T_OBJECT_CAST) reduce 151) ((T_PRINT) reduce 151) ((T_PROPERTY_C) reduce 151) ((T_READONLY) reduce 151) ((T_REQUIRE) reduce 151) ((T_REQUIRE_ONCE) reduce 151) ((T_RETURN) reduce 151) ((T_START_HEREDOC) reduce 151) ((T_STATIC) reduce 151) ((T_STRING) reduce 151) ((T_STRING_CAST) reduce 151) ((T_SWITCH) reduce 151) ((T_THROW) reduce 151) ((T_TRAIT) reduce 151) ((T_TRAIT_C) reduce 151) ((T_TRY) reduce 151) ((T_UNSET) reduce 151) ((T_UNSET_CAST) reduce 151) ((T_USE) reduce 151) ((T_VARIABLE) reduce 151) ((T_VOID_CAST) reduce 151) ((T_WHILE) reduce 151) ((T_YIELD) reduce 151) ((T_YIELD_FROM) reduce 151) (("[") reduce 151) (("`") reduce 151) (("{") reduce 151) (("~") reduce 151)) 33015 ((("!") reduce 165) (("\"") reduce 165) (($) reduce 165) (("$") reduce 165) (("(") reduce 165) (("+") reduce 165) (("-") reduce 165) ((";") reduce 165) (("@") reduce 165) ((T_ABSTRACT) reduce 165) ((T_ARRAY) reduce 165) ((T_ARRAY_CAST) reduce 165) ((T_ATTRIBUTE) reduce 165) ((T_BOOL_CAST) reduce 165) ((T_BREAK) reduce 165) ((T_CLASS) reduce 165) ((T_CLASS_C) reduce 165) ((T_CLONE) reduce 165) ((T_CONST) reduce 165) ((T_CONSTANT_ENCAPSED_STRING) reduce 165) ((T_CONTINUE) reduce 165) ((T_DEC) reduce 165) ((T_DECLARE) reduce 165) ((T_DIR) reduce 165) ((T_DNUMBER) reduce 165) ((T_DO) reduce 165) ((T_DOUBLE_CAST) reduce 165) ((T_ECHO) reduce 165) ((T_EMPTY) reduce 165) ((T_ENUM) reduce 165) ((T_EVAL) reduce 165) ((T_EXIT) reduce 165) ((T_FILE) reduce 165) ((T_FINAL) reduce 165) ((T_FN) reduce 165) ((T_FOR) reduce 165) ((T_FOREACH) reduce 165) ((T_FUNCTION) reduce 165) ((T_FUNC_C) reduce 165) ((T_GLOBAL) reduce 165) ((T_GOTO) reduce 165) ((T_HALT_COMPILER) reduce 165) ((T_IF) reduce 165) ((T_INC) reduce 165) ((T_INCLUDE) reduce 165) ((T_INCLUDE_ONCE) reduce 165) ((T_INLINE_HTML) reduce 165) ((T_INTERFACE) reduce 165) ((T_INT_CAST) reduce 165) ((T_ISSET) reduce 165) ((T_LINE) reduce 165) ((T_LIST) reduce 165) ((T_LNUMBER) reduce 165) ((T_MATCH) reduce 165) ((T_METHOD_C) reduce 165) ((T_NAMESPACE) reduce 165) ((T_NAME_FULLY_QUALIFIED) reduce 165) ((T_NAME_QUALIFIED) reduce 165) ((T_NAME_RELATIVE) reduce 165) ((T_NEW) reduce 165) ((T_NS_C) reduce 165) ((T_OBJECT_CAST) reduce 165) ((T_PRINT) reduce 165) ((T_PROPERTY_C) reduce 165) ((T_READONLY) reduce 165) ((T_REQUIRE) reduce 165) ((T_REQUIRE_ONCE) reduce 165) ((T_RETURN) reduce 165) ((T_START_HEREDOC) reduce 165) ((T_STATIC) reduce 165) ((T_STRING) reduce 165) ((T_STRING_CAST) reduce 165) ((T_SWITCH) reduce 165) ((T_THROW) reduce 165) ((T_TRAIT) reduce 165) ((T_TRAIT_C) reduce 165) ((T_TRY) reduce 165) ((T_UNSET) reduce 165) ((T_UNSET_CAST) reduce 165) ((T_USE) reduce 165) ((T_VARIABLE) reduce 165) ((T_VOID_CAST) reduce 165) ((T_WHILE) reduce 165) ((T_YIELD) reduce 165) ((T_YIELD_FROM) reduce 165) (("[") reduce 165) (("`") reduce 165) (("{") reduce 165) (("~") reduce 165)) 33016 ((("!") reduce 222) (("\"") reduce 222) (("$") reduce 222) (($) reduce 222) (("(") reduce 222) (("+") reduce 222) (("-") reduce 222) ((";") reduce 222) (("@") reduce 222) ((T_ABSTRACT) reduce 222) ((T_ARRAY) reduce 222) ((T_ARRAY_CAST) reduce 222) ((T_ATTRIBUTE) reduce 222) ((T_BOOL_CAST) reduce 222) ((T_BREAK) reduce 222) ((T_CLASS) reduce 222) ((T_CLASS_C) reduce 222) ((T_CLONE) reduce 222) ((T_CONST) reduce 222) ((T_CONSTANT_ENCAPSED_STRING) reduce 222) ((T_CONTINUE) reduce 222) ((T_DEC) reduce 222) ((T_DECLARE) reduce 222) ((T_DIR) reduce 222) ((T_DNUMBER) reduce 222) ((T_DO) reduce 222) ((T_DOUBLE_CAST) reduce 222) ((T_ECHO) reduce 222) ((T_EMPTY) reduce 222) ((T_ENUM) reduce 222) ((T_EVAL) reduce 222) ((T_EXIT) reduce 222) ((T_FILE) reduce 222) ((T_FINAL) reduce 222) ((T_FN) reduce 222) ((T_FOR) reduce 222) ((T_FOREACH) reduce 222) ((T_FUNCTION) reduce 222) ((T_FUNC_C) reduce 222) ((T_GLOBAL) reduce 222) ((T_GOTO) reduce 222) ((T_HALT_COMPILER) reduce 222) ((T_IF) reduce 222) ((T_INC) reduce 222) ((T_INCLUDE) reduce 222) ((T_INCLUDE_ONCE) reduce 222) ((T_INLINE_HTML) reduce 222) ((T_INTERFACE) reduce 222) ((T_INT_CAST) reduce 222) ((T_ISSET) reduce 222) ((T_LINE) reduce 222) ((T_LIST) reduce 222) ((T_LNUMBER) reduce 222) ((T_MATCH) reduce 222) ((T_METHOD_C) reduce 222) ((T_NAMESPACE) reduce 222) ((T_NAME_FULLY_QUALIFIED) reduce 222) ((T_NAME_QUALIFIED) reduce 222) ((T_NAME_RELATIVE) reduce 222) ((T_NEW) reduce 222) ((T_NS_C) reduce 222) ((T_OBJECT_CAST) reduce 222) ((T_PRINT) reduce 222) ((T_PROPERTY_C) reduce 222) ((T_READONLY) reduce 222) ((T_REQUIRE) reduce 222) ((T_REQUIRE_ONCE) reduce 222) ((T_RETURN) reduce 222) ((T_START_HEREDOC) reduce 222) ((T_STATIC) reduce 222) ((T_STRING) reduce 222) ((T_STRING_CAST) reduce 222) ((T_SWITCH) reduce 222) ((T_THROW) reduce 222) ((T_TRAIT) reduce 222) ((T_TRAIT_C) reduce 222) ((T_TRY) reduce 222) ((T_UNSET) reduce 222) ((T_UNSET_CAST) reduce 222) ((T_USE) reduce 222) ((T_VARIABLE) reduce 222) ((T_VOID_CAST) reduce 222) ((T_WHILE) reduce 222) ((T_YIELD) reduce 222) ((T_YIELD_FROM) reduce 222) (("[") reduce 222) (("`") reduce 222) (("{") reduce 222) (("~") reduce 222)) 33019 ((("!") reduce 223) (("\"") reduce 223) (($) reduce 223) (("$") reduce 223) (("(") reduce 223) (("+") reduce 223) (("-") reduce 223) ((";") reduce 223) (("@") reduce 223) ((T_ABSTRACT) reduce 223) ((T_ARRAY) reduce 223) ((T_ARRAY_CAST) reduce 223) ((T_ATTRIBUTE) reduce 223) ((T_BOOL_CAST) reduce 223) ((T_BREAK) reduce 223) ((T_CLASS) reduce 223) ((T_CLASS_C) reduce 223) ((T_CLONE) reduce 223) ((T_CONST) reduce 223) ((T_CONSTANT_ENCAPSED_STRING) reduce 223) ((T_CONTINUE) reduce 223) ((T_DEC) reduce 223) ((T_DECLARE) reduce 223) ((T_DIR) reduce 223) ((T_DNUMBER) reduce 223) ((T_DO) reduce 223) ((T_DOUBLE_CAST) reduce 223) ((T_ECHO) reduce 223) ((T_EMPTY) reduce 223) ((T_ENUM) reduce 223) ((T_EVAL) reduce 223) ((T_EXIT) reduce 223) ((T_FILE) reduce 223) ((T_FINAL) reduce 223) ((T_FN) reduce 223) ((T_FOR) reduce 223) ((T_FOREACH) reduce 223) ((T_FUNCTION) reduce 223) ((T_FUNC_C) reduce 223) ((T_GLOBAL) reduce 223) ((T_GOTO) reduce 223) ((T_HALT_COMPILER) reduce 223) ((T_IF) reduce 223) ((T_INC) reduce 223) ((T_INCLUDE) reduce 223) ((T_INCLUDE_ONCE) reduce 223) ((T_INLINE_HTML) reduce 223) ((T_INTERFACE) reduce 223) ((T_INT_CAST) reduce 223) ((T_ISSET) reduce 223) ((T_LINE) reduce 223) ((T_LIST) reduce 223) ((T_LNUMBER) reduce 223) ((T_MATCH) reduce 223) ((T_METHOD_C) reduce 223) ((T_NAMESPACE) reduce 223) ((T_NAME_FULLY_QUALIFIED) reduce 223) ((T_NAME_QUALIFIED) reduce 223) ((T_NAME_RELATIVE) reduce 223) ((T_NEW) reduce 223) ((T_NS_C) reduce 223) ((T_OBJECT_CAST) reduce 223) ((T_PRINT) reduce 223) ((T_PROPERTY_C) reduce 223) ((T_READONLY) reduce 223) ((T_REQUIRE) reduce 223) ((T_REQUIRE_ONCE) reduce 223) ((T_RETURN) reduce 223) ((T_START_HEREDOC) reduce 223) ((T_STATIC) reduce 223) ((T_STRING) reduce 223) ((T_STRING_CAST) reduce 223) ((T_SWITCH) reduce 223) ((T_THROW) reduce 223) ((T_TRAIT) reduce 223) ((T_TRAIT_C) reduce 223) ((T_TRY) reduce 223) ((T_UNSET) reduce 223) ((T_UNSET_CAST) reduce 223) ((T_USE) reduce 223) ((T_VARIABLE) reduce 223) ((T_VOID_CAST) reduce 223) ((T_WHILE) reduce 223) ((T_YIELD) reduce 223) ((T_YIELD_FROM) reduce 223) (("[") reduce 223) (("`") reduce 223) (("{") reduce 223) (("~") reduce 223)) 33022 ((("!") reduce 155) (("\"") reduce 155) (($) reduce 155) (("$") reduce 155) (("(") reduce 155) (("+") reduce 155) (("-") reduce 155) ((";") reduce 155) (("@") reduce 155) ((T_ABSTRACT) reduce 155) ((T_ARRAY) reduce 155) ((T_ARRAY_CAST) reduce 155) ((T_ATTRIBUTE) reduce 155) ((T_BOOL_CAST) reduce 155) ((T_BREAK) reduce 155) ((T_CLASS) reduce 155) ((T_CLASS_C) reduce 155) ((T_CLONE) reduce 155) ((T_CONST) reduce 155) ((T_CONSTANT_ENCAPSED_STRING) reduce 155) ((T_CONTINUE) reduce 155) ((T_DEC) reduce 155) ((T_DECLARE) reduce 155) ((T_DIR) reduce 155) ((T_DNUMBER) reduce 155) ((T_DO) reduce 155) ((T_DOUBLE_CAST) reduce 155) ((T_ECHO) reduce 155) ((T_EMPTY) reduce 155) ((T_ENUM) reduce 155) ((T_EVAL) reduce 155) ((T_EXIT) reduce 155) ((T_FILE) reduce 155) ((T_FINAL) reduce 155) ((T_FN) reduce 155) ((T_FOR) reduce 155) ((T_FOREACH) reduce 155) ((T_FUNCTION) reduce 155) ((T_FUNC_C) reduce 155) ((T_GLOBAL) reduce 155) ((T_GOTO) reduce 155) ((T_HALT_COMPILER) reduce 155) ((T_IF) reduce 155) ((T_INC) reduce 155) ((T_INCLUDE) reduce 155) ((T_INCLUDE_ONCE) reduce 155) ((T_INLINE_HTML) reduce 155) ((T_INTERFACE) reduce 155) ((T_INT_CAST) reduce 155) ((T_ISSET) reduce 155) ((T_LINE) reduce 155) ((T_LIST) reduce 155) ((T_LNUMBER) reduce 155) ((T_MATCH) reduce 155) ((T_METHOD_C) reduce 155) ((T_NAMESPACE) reduce 155) ((T_NAME_FULLY_QUALIFIED) reduce 155) ((T_NAME_QUALIFIED) reduce 155) ((T_NAME_RELATIVE) reduce 155) ((T_NEW) reduce 155) ((T_NS_C) reduce 155) ((T_OBJECT_CAST) reduce 155) ((T_PRINT) reduce 155) ((T_PROPERTY_C) reduce 155) ((T_READONLY) reduce 155) ((T_REQUIRE) reduce 155) ((T_REQUIRE_ONCE) reduce 155) ((T_RETURN) reduce 155) ((T_START_HEREDOC) reduce 155) ((T_STATIC) reduce 155) ((T_STRING) reduce 155) ((T_STRING_CAST) reduce 155) ((T_SWITCH) reduce 155) ((T_THROW) reduce 155) ((T_TRAIT) reduce 155) ((T_TRAIT_C) reduce 155) ((T_TRY) reduce 155) ((T_UNSET) reduce 155) ((T_UNSET_CAST) reduce 155) ((T_USE) reduce 155) ((T_VARIABLE) reduce 155) ((T_VOID_CAST) reduce 155) ((T_WHILE) reduce 155) ((T_YIELD) reduce 155) ((T_YIELD_FROM) reduce 155) (("[") reduce 155) (("`") reduce 155) (("{") reduce 155) (("~") reduce 155)) 33024 ((("!") reduce 109) (("\"") reduce 109) (($) reduce 109) (("$") reduce 109) (("(") reduce 109) (("+") reduce 109) (("-") reduce 109) ((";") reduce 109) (("@") reduce 109) ((T_ABSTRACT) reduce 109) ((T_ARRAY) reduce 109) ((T_ARRAY_CAST) reduce 109) ((T_ATTRIBUTE) reduce 109) ((T_BOOL_CAST) reduce 109) ((T_BREAK) reduce 109) ((T_CLASS) reduce 109) ((T_CLASS_C) reduce 109) ((T_CLONE) reduce 109) ((T_CONST) reduce 109) ((T_CONSTANT_ENCAPSED_STRING) reduce 109) ((T_CONTINUE) reduce 109) ((T_DEC) reduce 109) ((T_DECLARE) reduce 109) ((T_DIR) reduce 109) ((T_DNUMBER) reduce 109) ((T_DO) reduce 109) ((T_DOUBLE_CAST) reduce 109) ((T_ECHO) reduce 109) ((T_EMPTY) reduce 109) ((T_ENUM) reduce 109) ((T_EVAL) reduce 109) ((T_EXIT) reduce 109) ((T_FILE) reduce 109) ((T_FINAL) reduce 109) ((T_FN) reduce 109) ((T_FOR) reduce 109) ((T_FOREACH) reduce 109) ((T_FUNCTION) reduce 109) ((T_FUNC_C) reduce 109) ((T_GLOBAL) reduce 109) ((T_GOTO) reduce 109) ((T_HALT_COMPILER) reduce 109) ((T_IF) reduce 109) ((T_INC) reduce 109) ((T_INCLUDE) reduce 109) ((T_INCLUDE_ONCE) reduce 109) ((T_INLINE_HTML) reduce 109) ((T_INTERFACE) reduce 109) ((T_INT_CAST) reduce 109) ((T_ISSET) reduce 109) ((T_LINE) reduce 109) ((T_LIST) reduce 109) ((T_LNUMBER) reduce 109) ((T_MATCH) reduce 109) ((T_METHOD_C) reduce 109) ((T_NAMESPACE) reduce 109) ((T_NAME_FULLY_QUALIFIED) reduce 109) ((T_NAME_QUALIFIED) reduce 109) ((T_NAME_RELATIVE) reduce 109) ((T_NEW) reduce 109) ((T_NS_C) reduce 109) ((T_OBJECT_CAST) reduce 109) ((T_PRINT) reduce 109) ((T_PROPERTY_C) reduce 109) ((T_READONLY) reduce 109) ((T_REQUIRE) reduce 109) ((T_REQUIRE_ONCE) reduce 109) ((T_RETURN) reduce 109) ((T_START_HEREDOC) reduce 109) ((T_STATIC) reduce 109) ((T_STRING) reduce 109) ((T_STRING_CAST) reduce 109) ((T_SWITCH) reduce 109) ((T_THROW) reduce 109) ((T_TRAIT) reduce 109) ((T_TRAIT_C) reduce 109) ((T_TRY) reduce 109) ((T_UNSET) reduce 109) ((T_UNSET_CAST) reduce 109) ((T_USE) reduce 109) ((T_VARIABLE) reduce 109) ((T_VOID_CAST) reduce 109) ((T_WHILE) reduce 109) ((T_YIELD) reduce 109) ((T_YIELD_FROM) reduce 109) (("[") reduce 109) (("`") reduce 109) (("{") reduce 109) (("~") reduce 109)) 33044 ((("!") reduce 190) (("\"") reduce 190) (("$") reduce 190) (($) reduce 190) (("(") reduce 190) (("+") reduce 190) (("-") reduce 190) ((";") reduce 190) (("@") reduce 190) ((T_ABSTRACT) reduce 190) ((T_ARRAY) reduce 190) ((T_ARRAY_CAST) reduce 190) ((T_ATTRIBUTE) reduce 190) ((T_BOOL_CAST) reduce 190) ((T_BREAK) reduce 190) ((T_CLASS) reduce 190) ((T_CLASS_C) reduce 190) ((T_CLONE) reduce 190) ((T_CONST) reduce 190) ((T_CONSTANT_ENCAPSED_STRING) reduce 190) ((T_CONTINUE) reduce 190) ((T_DEC) reduce 190) ((T_DECLARE) reduce 190) ((T_DIR) reduce 190) ((T_DNUMBER) reduce 190) ((T_DO) reduce 190) ((T_DOUBLE_CAST) reduce 190) ((T_ECHO) reduce 190) ((T_EMPTY) reduce 190) ((T_ENUM) reduce 190) ((T_EVAL) reduce 190) ((T_EXIT) reduce 190) ((T_FILE) reduce 190) ((T_FINAL) reduce 190) ((T_FN) reduce 190) ((T_FOR) reduce 190) ((T_FOREACH) reduce 190) ((T_FUNCTION) reduce 190) ((T_FUNC_C) reduce 190) ((T_GLOBAL) reduce 190) ((T_GOTO) reduce 190) ((T_HALT_COMPILER) reduce 190) ((T_IF) reduce 190) ((T_INC) reduce 190) ((T_INCLUDE) reduce 190) ((T_INCLUDE_ONCE) reduce 190) ((T_INLINE_HTML) reduce 190) ((T_INTERFACE) reduce 190) ((T_INT_CAST) reduce 190) ((T_ISSET) reduce 190) ((T_LINE) reduce 190) ((T_LIST) reduce 190) ((T_LNUMBER) reduce 190) ((T_MATCH) reduce 190) ((T_METHOD_C) reduce 190) ((T_NAMESPACE) reduce 190) ((T_NAME_FULLY_QUALIFIED) reduce 190) ((T_NAME_QUALIFIED) reduce 190) ((T_NAME_RELATIVE) reduce 190) ((T_NEW) reduce 190) ((T_NS_C) reduce 190) ((T_OBJECT_CAST) reduce 190) ((T_PRINT) reduce 190) ((T_PROPERTY_C) reduce 190) ((T_READONLY) reduce 190) ((T_REQUIRE) reduce 190) ((T_REQUIRE_ONCE) reduce 190) ((T_RETURN) reduce 190) ((T_START_HEREDOC) reduce 190) ((T_STATIC) reduce 190) ((T_STRING) reduce 190) ((T_STRING_CAST) reduce 190) ((T_SWITCH) reduce 190) ((T_THROW) reduce 190) ((T_TRAIT) reduce 190) ((T_TRAIT_C) reduce 190) ((T_TRY) reduce 190) ((T_UNSET) reduce 190) ((T_UNSET_CAST) reduce 190) ((T_USE) reduce 190) ((T_VARIABLE) reduce 190) ((T_VOID_CAST) reduce 190) ((T_WHILE) reduce 190) ((T_YIELD) reduce 190) ((T_YIELD_FROM) reduce 190) (("[") reduce 190) (("`") reduce 190) (("{") reduce 190) (("~") reduce 190)) 33046 ((("!") reduce 154) (("\"") reduce 154) (($) reduce 154) (("$") reduce 154) (("(") reduce 154) (("+") reduce 154) (("-") reduce 154) ((";") reduce 154) (("@") reduce 154) ((T_ABSTRACT) reduce 154) ((T_ARRAY) reduce 154) ((T_ARRAY_CAST) reduce 154) ((T_ATTRIBUTE) reduce 154) ((T_BOOL_CAST) reduce 154) ((T_BREAK) reduce 154) ((T_CLASS) reduce 154) ((T_CLASS_C) reduce 154) ((T_CLONE) reduce 154) ((T_CONST) reduce 154) ((T_CONSTANT_ENCAPSED_STRING) reduce 154) ((T_CONTINUE) reduce 154) ((T_DEC) reduce 154) ((T_DECLARE) reduce 154) ((T_DIR) reduce 154) ((T_DNUMBER) reduce 154) ((T_DO) reduce 154) ((T_DOUBLE_CAST) reduce 154) ((T_ECHO) reduce 154) ((T_EMPTY) reduce 154) ((T_ENUM) reduce 154) ((T_EVAL) reduce 154) ((T_EXIT) reduce 154) ((T_FILE) reduce 154) ((T_FINAL) reduce 154) ((T_FN) reduce 154) ((T_FOR) reduce 154) ((T_FOREACH) reduce 154) ((T_FUNCTION) reduce 154) ((T_FUNC_C) reduce 154) ((T_GLOBAL) reduce 154) ((T_GOTO) reduce 154) ((T_HALT_COMPILER) reduce 154) ((T_IF) reduce 154) ((T_INC) reduce 154) ((T_INCLUDE) reduce 154) ((T_INCLUDE_ONCE) reduce 154) ((T_INLINE_HTML) reduce 154) ((T_INTERFACE) reduce 154) ((T_INT_CAST) reduce 154) ((T_ISSET) reduce 154) ((T_LINE) reduce 154) ((T_LIST) reduce 154) ((T_LNUMBER) reduce 154) ((T_MATCH) reduce 154) ((T_METHOD_C) reduce 154) ((T_NAMESPACE) reduce 154) ((T_NAME_FULLY_QUALIFIED) reduce 154) ((T_NAME_QUALIFIED) reduce 154) ((T_NAME_RELATIVE) reduce 154) ((T_NEW) reduce 154) ((T_NS_C) reduce 154) ((T_OBJECT_CAST) reduce 154) ((T_PRINT) reduce 154) ((T_PROPERTY_C) reduce 154) ((T_READONLY) reduce 154) ((T_REQUIRE) reduce 154) ((T_REQUIRE_ONCE) reduce 154) ((T_RETURN) reduce 154) ((T_START_HEREDOC) reduce 154) ((T_STATIC) reduce 154) ((T_STRING) reduce 154) ((T_STRING_CAST) reduce 154) ((T_SWITCH) reduce 154) ((T_THROW) reduce 154) ((T_TRAIT) reduce 154) ((T_TRAIT_C) reduce 154) ((T_TRY) reduce 154) ((T_UNSET) reduce 154) ((T_UNSET_CAST) reduce 154) ((T_USE) reduce 154) ((T_VARIABLE) reduce 154) ((T_VOID_CAST) reduce 154) ((T_WHILE) reduce 154) ((T_YIELD) reduce 154) ((T_YIELD_FROM) reduce 154) (("[") reduce 154) (("`") reduce 154) (("{") reduce 154) (("~") reduce 154)) 33050 (((T_ABSTRACT) reduce 100) ((T_ATTRIBUTE) reduce 100) ((T_CLASS) reduce 100) ((T_CONST) reduce 100) ((T_ENUM) reduce 100) ((T_FINAL) reduce 100) ((T_FN) reduce 100) ((T_FUNCTION) reduce 100) ((T_INTERFACE) reduce 100) ((T_READONLY) reduce 100) ((T_STATIC) reduce 100) ((T_TRAIT) reduce 100))))
  "The generated distinct action-tables.")

(defvar
  phps-mode-parser--goto-tables
  #s(hash-table test equal data (0 0 1 1 2 2 3 3 4 4 5 5 6 6 7 7 8 8 9 9 10 8 11 10 12 11 13 12 14 13 15 14 16 15 17 8 18 16 19 17 20 8 21 18 22 19 23 20 24 8 25 8 26 21 27 22 28 23 29 24 30 25 31 26 32 27 33 8 34 8 35 8 36 28 37 29 38 8 39 8 40 30 41 31 42 32 43 33 44 34 45 35 46 36 47 8 48 37 49 38 50 39 51 8 52 40 53 8 54 41 55 8 56 42 57 8 58 8 59 8 60 43 61 8 62 44 63 45 64 8 65 46 66 47 67 48 68 49 69 50 70 51 71 52 72 53 73 54 74 55 75 56 76 8 77 57 78 58 79 59 80 60 81 8 82 61 83 62 84 63 85 64 86 65 87 66 88 8 89 67 90 68 91 8 92 8 93 8 94 69 95 70 96 8 97 8 98 8 99 8 100 71 101 72 102 8 103 8 104 8 105 73 106 74 107 8 108 75 109 8 110 8 111 8 112 76 113 8 114 8 115 8 116 8 117 77 118 8 119 8 120 8 121 8 122 8 123 8 124 8 125 8 126 78 127 79 128 80 129 81 130 82 131 8 132 46 133 83 134 8 135 8 136 84 137 85 138 86 139 8 140 8 141 8 142 87 143 88 144 89 145 8 146 90 147 8 148 8 149 8 150 8 151 8 152 8 153 8 154 8 155 8 156 8 157 8 158 91 159 8 160 92 161 93 162 94 163 8 164 95 165 96 166 97 167 98 168 99 169 100 170 101 171 102 172 103 173 8 174 104 175 105 176 106 177 107 178 108 179 109 180 110 181 111 182 112 183 113 184 114 185 115 186 116 187 117 188 8 189 118 190 119 191 120 192 121 193 122 194 123 195 124 196 125 197 126 198 127 199 8 200 128 201 8 202 8 203 129 204 8 205 130 206 8 207 131 208 132 209 8 210 8 211 133 212 8 213 8 214 8 215 8 216 134 217 135 218 8 219 136 220 8 221 137 222 138 223 139 224 140 225 141 226 142 227 143 228 144 229 145 230 146 231 147 232 148 233 149 234 8 235 150 236 8 237 151 238 8 239 8 240 152 241 153 242 154 243 155 244 8 245 8 246 156 247 157 248 158 249 159 250 160 251 8 252 161 253 8 254 162 255 8 256 8 257 8 258 8 259 163 260 8 261 164 262 165 263 8 264 166 265 167 266 168 267 169 268 170 269 8 270 171 271 172 272 8 273 173 274 8 275 174 276 175 277 176 278 177 279 178 280 179 281 180 282 8 283 8 284 181 285 8 286 8 287 182 288 183 289 184 290 8 291 8 292 8 293 8 294 185 295 8 296 8 297 8 298 8 299 8 300 186 301 187 302 188 303 189 304 190 305 191 306 192 307 193 308 194 309 195 310 196 311 197 312 198 313 199 314 200 315 201 316 202 317 203 318 204 319 205 320 206 321 207 322 208 323 209 324 210 325 211 326 212 327 213 328 214 329 215 330 216 331 217 332 218 333 219 334 220 335 189 336 189 337 189 338 189 339 189 340 189 341 189 342 189 343 189 344 189 345 189 346 189 347 189 348 189 349 189 350 189 351 221 352 222 353 8 354 8 355 8 356 8 357 8 358 8 359 223 360 8 361 8 362 224 363 8 364 225 365 226 366 227 367 228 368 229 369 230 370 231 371 232 372 233 373 234 374 235 375 236 376 237 377 238 378 8 379 239 380 8 381 240 382 8 383 8 384 241 385 242 386 243 387 244 388 8 389 8 390 245 391 246 392 247 393 248 394 249 395 8 396 250 397 8 398 251 399 8 400 8 401 8 402 8 403 252 404 8 405 253 406 254 407 8 408 255 409 256 410 257 411 258 412 259 413 8 414 260 415 261 416 8 417 262 418 8 419 263 420 264 421 265 422 266 423 267 424 268 425 269 426 8 427 8 428 270 429 8 430 8 431 271 432 272 433 273 434 8 435 8 436 8 437 8 438 274 439 8 440 8 441 275 442 8 443 8 444 8 445 276 446 277 447 278 448 271 449 279 450 280 451 281 452 282 453 283 454 284 455 285 456 286 457 287 458 288 459 289 460 290 461 291 462 292 463 293 464 294 465 295 466 296 467 297 468 298 469 299 470 300 471 301 472 302 473 303 474 304 475 305 476 306 477 307 478 308 479 309 480 271 481 271 482 271 483 271 484 271 485 271 486 271 487 271 488 271 489 271 490 271 491 271 492 271 493 271 494 271 495 271 496 310 497 311 498 8 499 8 500 8 501 8 502 8 503 8 504 312 505 8 506 8 507 313 508 8 509 314 510 315 511 316 512 317 513 318 514 8 515 8 516 8 517 8 518 8 519 319 520 320 521 321 522 322 523 323 524 324 525 325 526 326 527 327 528 328 529 329 530 8 531 330 532 8 533 331 534 8 535 8 536 332 537 333 538 334 539 335 540 8 541 8 542 336 543 337 544 338 545 339 546 340 547 8 548 341 549 8 550 342 551 8 552 8 553 8 554 8 555 343 556 8 557 344 558 345 559 8 560 346 561 347 562 348 563 349 564 350 565 8 566 351 567 352 568 8 569 353 570 8 571 354 572 355 573 356 574 357 575 358 576 359 577 360 578 8 579 8 580 361 581 8 582 8 583 362 584 363 585 364 586 8 587 8 588 8 589 8 590 365 591 8 592 8 593 8 594 8 595 8 596 366 597 367 598 368 599 369 600 370 601 371 602 372 603 373 604 374 605 375 606 376 607 377 608 378 609 379 610 380 611 381 612 382 613 383 614 384 615 385 616 386 617 387 618 388 619 389 620 390 621 391 622 392 623 393 624 394 625 395 626 396 627 397 628 398 629 399 630 400 631 369 632 369 633 369 634 369 635 369 636 369 637 369 638 369 639 369 640 369 641 369 642 369 643 369 644 369 645 369 646 369 647 401 648 402 649 8 650 8 651 8 652 8 653 8 654 8 655 403 656 8 657 8 658 404 659 8 660 405 661 406 662 407 663 408 664 409 665 8 666 8 667 8 668 8 669 8 670 410 671 411 672 8 673 8 674 412 675 8 676 413 677 8 678 8 679 414 680 415 681 8 682 369 683 369 684 369 685 369 686 369 687 416 688 417 689 418 690 419 691 420 692 421 693 422 694 423 695 424 696 425 697 426 698 8 699 427 700 8 701 428 702 8 703 8 704 429 705 430 706 431 707 432 708 8 709 8 710 433 711 434 712 435 713 436 714 437 715 8 716 438 717 8 718 439 719 8 720 8 721 8 722 8 723 440 724 8 725 441 726 442 727 8 728 443 729 444 730 445 731 446 732 447 733 8 734 448 735 449 736 8 737 450 738 8 739 451 740 452 741 453 742 454 743 455 744 456 745 457 746 8 747 8 748 458 749 8 750 8 751 459 752 460 753 461 754 8 755 8 756 8 757 8 758 462 759 8 760 8 761 8 762 8 763 8 764 463 765 464 766 465 767 466 768 467 769 468 770 469 771 470 772 471 773 472 774 473 775 474 776 475 777 476 778 477 779 478 780 479 781 480 782 481 783 482 784 483 785 484 786 485 787 486 788 487 789 488 790 489 791 490 792 491 793 492 794 493 795 494 796 495 797 496 798 497 799 466 800 466 801 466 802 466 803 466 804 466 805 466 806 466 807 466 808 466 809 466 810 466 811 466 812 466 813 466 814 466 815 498 816 499 817 8 818 8 819 8 820 8 821 8 822 8 823 500 824 8 825 8 826 501 827 8 828 502 829 503 830 504 831 505 832 506 833 8 834 8 835 8 836 8 837 8 838 507 839 508 840 8 841 8 842 509 843 8 844 510 845 8 846 8 847 511 848 512 849 8 850 466 851 466 852 466 853 466 854 466 855 513 856 514 857 515 858 466 859 466 860 466 861 466 862 466 863 466 864 466 865 466 866 466 867 466 868 516 869 8 870 8 871 8 872 8 873 8 874 8 875 8 876 8 877 8 878 8 879 8 880 8 881 8 882 8 883 8 884 8 885 8 886 8 887 8 888 8 889 8 890 8 891 8 892 8 893 8 894 8 895 8 896 8 897 8 898 8 899 8 900 8 901 8 902 8 903 8 904 8 905 8 906 8 907 8 908 8 909 8 910 8 911 8 912 8 913 8 914 8 915 8 916 8 917 8 918 8 919 8 920 8 921 8 922 8 923 8 924 8 925 8 926 8 927 8 928 8 929 8 930 8 931 8 932 8 933 8 934 8 935 8 936 8 937 8 938 8 939 8 940 8 941 8 942 8 943 8 944 8 945 8 946 8 947 8 948 8 949 517 950 8 951 8 952 8 953 518 954 519 955 8 956 520 957 8 958 521 959 522 960 523 961 524 962 8 963 525 964 526 965 527 966 8 967 528 968 529 969 8 970 530 971 8 972 8 973 8 974 8 975 8 976 8 977 531 978 8 979 8 980 8 981 532 982 8 983 8 984 8 985 8 986 8 987 533 988 8 989 534 990 8 991 8 992 535 993 8 994 8 995 8 996 8 997 8 998 8 999 8 1000 536 1001 537 1002 8 1003 8 1004 8 1005 8 1006 8 1007 8 1008 8 1009 8 1010 8 1011 8 1012 8 1013 8 1014 8 1015 538 1016 539 1017 540 1018 8 1019 8 1020 8 1021 541 1022 542 1023 8 1024 543 1025 8 1026 8 1027 8 1028 8 1029 544 1030 8 1031 8 1032 8 1033 8 1034 8 1035 545 1036 8 1037 546 1038 547 1039 8 1040 8 1041 8 1042 8 1043 548 1044 549 1045 550 1046 8 1047 551 1048 552 1049 8 1050 553 1051 8 1052 554 1053 8 1054 8 1055 8 1056 8 1057 555 1058 8 1059 8 1060 8 1061 8 1062 556 1063 557 1064 558 1065 559 1066 8 1067 8 1068 560 1069 561 1070 562 1071 8 1072 8 1073 563 1074 8 1075 8 1076 564 1077 565 1078 566 1079 8 1080 567 1081 8 1082 8 1083 8 1084 568 1085 8 1086 8 1087 569 1088 8 1089 8 1090 8 1091 8 1092 8 1093 570 1094 571 1095 572 1096 528 1097 8 1098 531 1099 8 1100 535 1101 536 1102 537 1103 8 1104 8 1105 539 1106 540 1107 542 1108 8 1109 543 1110 544 1111 8 1112 545 1113 8 1114 547 1115 8 1116 548 1117 549 1118 550 1119 552 1120 8 1121 554 1122 8 1123 556 1124 564 1125 573 1126 574 1127 575 1128 576 1129 577 1130 578 1131 579 1132 580 1133 581 1134 582 1135 583 1136 584 1137 585 1138 586 1139 587 1140 588 1141 589 1142 590 1143 591 1144 592 1145 593 1146 594 1147 595 1148 596 1149 597 1150 598 1151 599 1152 600 1153 601 1154 602 1155 603 1156 564 1157 564 1158 564 1159 564 1160 564 1161 564 1162 564 1163 564 1164 564 1165 564 1166 564 1167 564 1168 564 1169 564 1170 564 1171 564 1172 604 1173 605 1174 8 1175 8 1176 8 1177 8 1178 8 1179 8 1180 606 1181 8 1182 8 1183 607 1184 8 1185 608 1186 609 1187 610 1188 611 1189 612 1190 8 1191 8 1192 8 1193 8 1194 8 1195 613 1196 614 1197 8 1198 8 1199 615 1200 8 1201 616 1202 8 1203 8 1204 617 1205 618 1206 8 1207 564 1208 564 1209 564 1210 564 1211 564 1212 619 1213 620 1214 621 1215 564 1216 564 1217 564 1218 564 1219 564 1220 564 1221 564 1222 564 1223 564 1224 564 1225 622 1226 623 1227 624 1228 625 1229 626 1230 627 1231 628 1232 629 1233 630 1234 631 1235 8 1236 632 1237 8 1238 633 1239 8 1240 8 1241 634 1242 635 1243 636 1244 637 1245 8 1246 8 1247 638 1248 639 1249 640 1250 641 1251 642 1252 8 1253 643 1254 8 1255 644 1256 8 1257 8 1258 8 1259 8 1260 645 1261 8 1262 646 1263 647 1264 8 1265 648 1266 649 1267 650 1268 651 1269 652 1270 8 1271 653 1272 654 1273 8 1274 655 1275 8 1276 656 1277 657 1278 658 1279 659 1280 660 1281 661 1282 662 1283 8 1284 8 1285 663 1286 8 1287 8 1288 664 1289 665 1290 666 1291 8 1292 8 1293 8 1294 8 1295 667 1296 8 1297 8 1298 8 1299 8 1300 8 1301 668 1302 669 1303 670 1304 671 1305 672 1306 673 1307 674 1308 675 1309 676 1310 677 1311 678 1312 679 1313 680 1314 681 1315 682 1316 683 1317 684 1318 685 1319 686 1320 687 1321 688 1322 689 1323 690 1324 691 1325 692 1326 693 1327 694 1328 695 1329 696 1330 697 1331 698 1332 699 1333 700 1334 701 1335 702 1336 671 1337 671 1338 671 1339 671 1340 671 1341 671 1342 671 1343 671 1344 671 1345 671 1346 671 1347 671 1348 671 1349 671 1350 671 1351 671 1352 703 1353 704 1354 8 1355 8 1356 8 1357 8 1358 8 1359 8 1360 705 1361 8 1362 8 1363 706 1364 8 1365 707 1366 708 1367 709 1368 710 1369 711 1370 8 1371 8 1372 8 1373 8 1374 8 1375 712 1376 713 1377 8 1378 8 1379 714 1380 8 1381 715 1382 8 1383 8 1384 716 1385 717 1386 8 1387 671 1388 671 1389 671 1390 671 1391 671 1392 718 1393 719 1394 720 1395 671 1396 671 1397 671 1398 671 1399 671 1400 671 1401 671 1402 671 1403 671 1404 671 1405 721 1406 8 1407 8 1408 8 1409 8 1410 8 1411 8 1412 8 1413 8 1414 8 1415 8 1416 8 1417 8 1418 8 1419 8 1420 8 1421 8 1422 8 1423 8 1424 8 1425 8 1426 8 1427 8 1428 8 1429 8 1430 8 1431 8 1432 8 1433 8 1434 8 1435 8 1436 8 1437 8 1438 8 1439 8 1440 8 1441 8 1442 8 1443 8 1444 8 1445 8 1446 8 1447 8 1448 8 1449 8 1450 8 1451 8 1452 8 1453 8 1454 8 1455 8 1456 8 1457 8 1458 8 1459 8 1460 8 1461 8 1462 8 1463 8 1464 8 1465 8 1466 8 1467 8 1468 8 1469 8 1470 8 1471 8 1472 8 1473 8 1474 8 1475 8 1476 8 1477 8 1478 8 1479 8 1480 8 1481 8 1482 8 1483 8 1484 8 1485 8 1486 722 1487 8 1488 8 1489 8 1490 723 1491 724 1492 8 1493 725 1494 8 1495 8 1496 726 1497 727 1498 728 1499 729 1500 8 1501 730 1502 8 1503 564 1504 8 1505 731 1506 732 1507 733 1508 734 1509 8 1510 735 1511 8 1512 736 1513 737 1514 738 1515 739 1516 740 1517 741 1518 742 1519 743 1520 744 1521 671 1522 671 1523 671 1524 671 1525 671 1526 671 1527 671 1528 671 1529 671 1530 671 1531 671 1532 671 1533 671 1534 8 1535 8 1536 745 1537 671 1538 746 1539 747 1540 748 1541 749 1542 8 1543 8 1544 8 1545 8 1546 8 1547 8 1548 8 1549 8 1550 8 1551 8 1552 750 1553 8 1554 8 1555 751 1556 8 1557 8 1558 8 1559 8 1560 752 1561 753 1562 754 1563 8 1564 8 1565 755 1566 8 1567 8 1568 8 1569 756 1570 8 1571 8 1572 8 1573 8 1574 757 1575 758 1576 8 1577 8 1578 8 1579 8 1580 8 1581 8 1582 8 1583 8 1584 8 1585 8 1586 8 1587 8 1588 8 1589 8 1590 8 1591 8 1592 8 1593 8 1594 8 1595 8 1596 8 1597 8 1598 8 1599 8 1600 8 1601 8 1602 8 1603 8 1604 8 1605 8 1606 8 1607 8 1608 8 1609 8 1610 8 1611 8 1612 8 1613 8 1614 8 1615 8 1616 8 1617 8 1618 8 1619 8 1620 8 1621 8 1622 8 1623 8 1624 8 1625 8 1626 8 1627 8 1628 8 1629 8 1630 8 1631 8 1632 8 1633 8 1634 8 1635 8 1636 8 1637 8 1638 8 1639 8 1640 8 1641 8 1642 8 1643 8 1644 8 1645 8 1646 8 1647 8 1648 8 1649 8 1650 8 1651 8 1652 8 1653 8 1654 8 1655 8 1656 759 1657 8 1658 8 1659 8 1660 760 1661 761 1662 8 1663 762 1664 8 1665 8 1666 763 1667 764 1668 765 1669 8 1670 8 1671 8 1672 766 1673 8 1674 767 1675 8 1676 768 1677 769 1678 8 1679 8 1680 8 1681 770 1682 771 1683 772 1684 773 1685 8 1686 8 1687 774 1688 8 1689 775 1690 776 1691 8 1692 8 1693 777 1694 8 1695 778 1696 779 1697 780 1698 781 1699 782 1700 783 1701 784 1702 785 1703 786 1704 787 1705 8 1706 788 1707 8 1708 789 1709 8 1710 8 1711 790 1712 791 1713 792 1714 793 1715 794 1716 8 1717 8 1718 795 1719 796 1720 797 1721 798 1722 799 1723 8 1724 800 1725 8 1726 801 1727 8 1728 8 1729 8 1730 8 1731 802 1732 8 1733 803 1734 804 1735 8 1736 805 1737 806 1738 807 1739 808 1740 809 1741 8 1742 810 1743 811 1744 8 1745 812 1746 8 1747 813 1748 814 1749 815 1750 816 1751 817 1752 818 1753 8 1754 819 1755 820 1756 821 1757 8 1758 8 1759 822 1760 8 1761 8 1762 823 1763 824 1764 825 1765 8 1766 8 1767 8 1768 8 1769 826 1770 8 1771 8 1772 827 1773 8 1774 8 1775 8 1776 8 1777 828 1778 829 1779 830 1780 831 1781 832 1782 833 1783 834 1784 835 1785 836 1786 837 1787 838 1788 839 1789 840 1790 841 1791 842 1792 843 1793 844 1794 845 1795 846 1796 847 1797 848 1798 849 1799 850 1800 851 1801 852 1802 853 1803 854 1804 855 1805 856 1806 857 1807 858 1808 859 1809 860 1810 861 1811 862 1812 863 1813 832 1814 832 1815 832 1816 832 1817 832 1818 832 1819 832 1820 832 1821 832 1822 832 1823 832 1824 832 1825 832 1826 832 1827 832 1828 832 1829 864 1830 865 1831 8 1832 8 1833 8 1834 8 1835 8 1836 8 1837 866 1838 8 1839 8 1840 867 1841 8 1842 868 1843 869 1844 870 1845 871 1846 872 1847 8 1848 8 1849 8 1850 8 1851 8 1852 873 1853 874 1854 8 1855 8 1856 875 1857 8 1858 876 1859 8 1860 8 1861 877 1862 878 1863 8 1864 832 1865 832 1866 832 1867 832 1868 832 1869 879 1870 880 1871 881 1872 832 1873 832 1874 832 1875 832 1876 832 1877 832 1878 832 1879 832 1880 832 1881 832 1882 882 1883 883 1884 884 1885 885 1886 886 1887 887 1888 888 1889 889 1890 890 1891 891 1892 892 1893 8 1894 893 1895 8 1896 894 1897 8 1898 8 1899 895 1900 896 1901 897 1902 898 1903 899 1904 8 1905 8 1906 900 1907 901 1908 902 1909 903 1910 904 1911 8 1912 905 1913 8 1914 906 1915 8 1916 8 1917 8 1918 8 1919 907 1920 8 1921 908 1922 909 1923 8 1924 910 1925 911 1926 912 1927 913 1928 914 1929 8 1930 915 1931 916 1932 8 1933 917 1934 8 1935 918 1936 919 1937 920 1938 921 1939 922 1940 923 1941 8 1942 924 1943 925 1944 926 1945 8 1946 8 1947 927 1948 8 1949 8 1950 928 1951 929 1952 930 1953 8 1954 8 1955 8 1956 8 1957 931 1958 8 1959 8 1960 932 1961 8 1962 8 1963 8 1964 8 1965 933 1966 934 1967 935 1968 936 1969 937 1970 938 1971 939 1972 940 1973 941 1974 942 1975 943 1976 944 1977 945 1978 946 1979 947 1980 948 1981 949 1982 950 1983 951 1984 952 1985 953 1986 954 1987 955 1988 956 1989 957 1990 958 1991 959 1992 960 1993 961 1994 962 1995 963 1996 964 1997 965 1998 966 1999 967 2000 968 2001 937 2002 937 2003 937 2004 937 2005 937 2006 937 2007 937 2008 937 2009 937 2010 937 2011 937 2012 937 2013 937 2014 937 2015 937 2016 937 2017 969 2018 970 2019 8 2020 8 2021 8 2022 8 2023 8 2024 8 2025 971 2026 8 2027 8 2028 972 2029 8 2030 973 2031 974 2032 975 2033 976 2034 977 2035 8 2036 8 2037 8 2038 8 2039 8 2040 978 2041 979 2042 8 2043 8 2044 980 2045 8 2046 981 2047 8 2048 8 2049 982 2050 983 2051 8 2052 937 2053 937 2054 937 2055 937 2056 937 2057 984 2058 985 2059 986 2060 937 2061 937 2062 937 2063 937 2064 937 2065 937 2066 937 2067 937 2068 937 2069 937 2070 987 2071 988 2072 989 2073 990 2074 937 2075 991 2076 8 2077 8 2078 8 2079 8 2080 8 2081 8 2082 8 2083 8 2084 8 2085 8 2086 8 2087 8 2088 8 2089 8 2090 8 2091 8 2092 8 2093 8 2094 8 2095 8 2096 8 2097 8 2098 8 2099 8 2100 8 2101 8 2102 8 2103 8 2104 8 2105 8 2106 8 2107 8 2108 8 2109 8 2110 8 2111 8 2112 8 2113 8 2114 8 2115 8 2116 8 2117 8 2118 8 2119 8 2120 8 2121 8 2122 8 2123 8 2124 8 2125 8 2126 8 2127 8 2128 8 2129 8 2130 8 2131 8 2132 8 2133 8 2134 8 2135 8 2136 8 2137 8 2138 8 2139 8 2140 8 2141 8 2142 8 2143 8 2144 8 2145 8 2146 8 2147 8 2148 8 2149 8 2150 8 2151 8 2152 8 2153 8 2154 8 2155 8 2156 992 2157 8 2158 8 2159 8 2160 993 2161 994 2162 8 2163 995 2164 8 2165 8 2166 996 2167 997 2168 998 2169 8 2170 8 2171 999 2172 1000 2173 1001 2174 1002 2175 8 2176 1003 2177 8 2178 1004 2179 1005 2180 1006 2181 1007 2182 1008 2183 1009 2184 1010 2185 1011 2186 1012 2187 937 2188 937 2189 937 2190 937 2191 937 2192 937 2193 937 2194 937 2195 937 2196 937 2197 937 2198 937 2199 937 2200 1013 2201 937 2202 1014 2203 1015 2204 8 2205 1016 2206 1017 2207 8 2208 1018 2209 8 2210 1019 2211 8 2212 8 2213 8 2214 1020 2215 1021 2216 1022 2217 8 2218 1023 2219 8 2220 8 2221 1024 2222 1025 2223 1026 2224 8 2225 8 2226 8 2227 1027 2228 1028 2229 8 2230 8 2231 8 2232 1029 2233 1030 2234 1031 2235 1032 2236 8 2237 8 2238 1033 2239 8 2240 1034 2241 1035 2242 8 2243 8 2244 1036 2245 8 2246 8 2247 8 2248 1037 2249 1038 2250 8 2251 1039 2252 8 2253 8 2254 1040 2255 1041 2256 1042 2257 1043 2258 1044 2259 1045 2260 1046 2261 1047 2262 1048 2263 1049 2264 1050 2265 8 2266 1051 2267 1052 2268 1053 2269 1054 2270 1055 2271 1056 2272 1057 2273 1058 2274 1059 2275 1060 2276 1061 2277 8 2278 1062 2279 1063 2280 8 2281 8 2282 8 2283 8 2284 8 2285 8 2286 8 2287 1064 2288 8 2289 1065 2290 8 2291 8 2292 1066 2293 1067 2294 1068 2295 1069 2296 8 2297 8 2298 8 2299 8 2300 8 2301 8 2302 8 2303 8 2304 8 2305 8 2306 8 2307 8 2308 8 2309 8 2310 8 2311 8 2312 8 2313 8 2314 1070 2315 8 2316 1071 2317 8 2318 1072 2319 8 2320 8 2321 8 2322 8 2323 8 2324 8 2325 8 2326 8 2327 8 2328 8 2329 8 2330 1073 2331 8 2332 1074 2333 8 2334 1075 2335 8 2336 8 2337 8 2338 8 2339 8 2340 8 2341 8 2342 8 2343 8 2344 8 2345 1076 2346 8 2347 8 2348 8 2349 8 2350 8 2351 8 2352 8 2353 8 2354 8 2355 8 2356 8 2357 1077 2358 8 2359 1078 2360 1079 2361 1080 2362 1081 2363 1082 2364 1083 2365 1084 2366 1085 2367 1086 2368 1087 2369 1088 2370 8 2371 1089 2372 8 2373 1090 2374 8 2375 8 2376 1091 2377 1092 2378 1093 2379 1094 2380 8 2381 8 2382 1095 2383 1096 2384 1097 2385 1098 2386 1099 2387 8 2388 1100 2389 8 2390 1101 2391 8 2392 8 2393 8 2394 8 2395 1102 2396 8 2397 1103 2398 1104 2399 8 2400 1105 2401 1106 2402 1107 2403 1108 2404 1109 2405 8 2406 1110 2407 1111 2408 8 2409 1112 2410 8 2411 1113 2412 1114 2413 1115 2414 1116 2415 1117 2416 1118 2417 1119 2418 1120 2419 8 2420 8 2421 1121 2422 8 2423 8 2424 1122 2425 1123 2426 1124 2427 8 2428 8 2429 8 2430 8 2431 1125 2432 8 2433 8 2434 8 2435 8 2436 8 2437 1126 2438 1127 2439 1128 2440 1129 2441 1122 2442 1130 2443 1131 2444 1132 2445 1133 2446 1134 2447 1135 2448 1136 2449 1137 2450 1138 2451 1139 2452 1140 2453 1141 2454 1142 2455 1143 2456 1144 2457 1145 2458 1146 2459 1147 2460 1148 2461 1149 2462 1150 2463 1151 2464 1152 2465 1153 2466 1154 2467 1155 2468 1156 2469 1157 2470 1158 2471 1159 2472 1160 2473 1122 2474 1122 2475 1122 2476 1122 2477 1122 2478 1122 2479 1122 2480 1122 2481 1122 2482 1122 2483 1122 2484 1122 2485 1122 2486 1122 2487 1122 2488 1122 2489 1161 2490 1162 2491 8 2492 8 2493 8 2494 8 2495 8 2496 8 2497 1163 2498 8 2499 8 2500 1164 2501 8 2502 1165 2503 1166 2504 1167 2505 1168 2506 1169 2507 8 2508 8 2509 8 2510 8 2511 8 2512 1170 2513 1171 2514 8 2515 8 2516 1172 2517 8 2518 1173 2519 8 2520 8 2521 1174 2522 1175 2523 8 2524 1122 2525 1122 2526 1122 2527 1122 2528 1122 2529 1176 2530 1177 2531 1178 2532 1122 2533 1122 2534 1122 2535 1122 2536 1122 2537 1122 2538 1122 2539 1122 2540 1122 2541 1122 2542 1179 2543 1180 2544 1181 2545 1182 2546 1122 2547 1183 2548 8 2549 8 2550 8 2551 8 2552 8 2553 8 2554 8 2555 8 2556 8 2557 8 2558 8 2559 8 2560 8 2561 8 2562 8 2563 8 2564 8 2565 8 2566 8 2567 8 2568 8 2569 8 2570 8 2571 8 2572 8 2573 8 2574 8 2575 8 2576 8 2577 8 2578 8 2579 8 2580 8 2581 8 2582 8 2583 8 2584 8 2585 8 2586 8 2587 8 2588 8 2589 8 2590 8 2591 8 2592 8 2593 8 2594 8 2595 8 2596 8 2597 8 2598 8 2599 8 2600 8 2601 8 2602 8 2603 8 2604 8 2605 8 2606 8 2607 8 2608 8 2609 8 2610 8 2611 8 2612 8 2613 8 2614 8 2615 8 2616 8 2617 8 2618 8 2619 8 2620 8 2621 8 2622 8 2623 8 2624 8 2625 8 2626 8 2627 8 2628 1184 2629 8 2630 8 2631 8 2632 1185 2633 1186 2634 8 2635 1187 2636 8 2637 8 2638 1188 2639 1189 2640 1190 2641 8 2642 8 2643 1191 2644 1192 2645 1193 2646 1194 2647 8 2648 1195 2649 8 2650 1196 2651 1197 2652 1198 2653 1199 2654 1200 2655 1201 2656 1202 2657 1203 2658 1204 2659 1122 2660 1122 2661 1122 2662 1122 2663 1122 2664 1122 2665 1122 2666 1122 2667 1122 2668 1122 2669 1122 2670 1122 2671 1122 2672 1205 2673 1122 2674 1206 2675 1207 2676 8 2677 1208 2678 1209 2679 8 2680 1210 2681 8 2682 1211 2683 8 2684 8 2685 8 2686 1212 2687 1213 2688 1214 2689 8 2690 1215 2691 8 2692 8 2693 1216 2694 1217 2695 1218 2696 8 2697 8 2698 8 2699 1219 2700 1220 2701 8 2702 8 2703 8 2704 1221 2705 1222 2706 1223 2707 1224 2708 8 2709 8 2710 1225 2711 8 2712 1226 2713 1227 2714 8 2715 8 2716 1228 2717 8 2718 8 2719 8 2720 1229 2721 1230 2722 8 2723 8 2724 1231 2725 1232 2726 1233 2727 1234 2728 1235 2729 1236 2730 1237 2731 1238 2732 1239 2733 1240 2734 1241 2735 8 2736 1242 2737 1243 2738 1244 2739 1245 2740 1246 2741 1247 2742 1248 2743 1249 2744 1250 2745 1251 2746 8 2747 1252 2748 8 2749 1253 2750 8 2751 1254 2752 1255 2753 8 2754 8 2755 8 2756 1256 2757 8 2758 8 2759 8 2760 1257 2761 1258 2762 1259 2763 1260 2764 8 2765 8 2766 1261 2767 8 2768 1262 2769 1263 2770 8 2771 8 2772 1264 2773 8 2774 1265 2775 1266 2776 8 2777 1267 2778 1268 2779 8 2780 1269 2781 8 2782 1270 2783 8 2784 8 2785 8 2786 1271 2787 1272 2788 1273 2789 8 2790 1274 2791 8 2792 8 2793 1275 2794 1276 2795 1277 2796 8 2797 8 2798 8 2799 1278 2800 1279 2801 8 2802 8 2803 8 2804 1280 2805 1281 2806 1282 2807 1283 2808 8 2809 8 2810 1284 2811 8 2812 1285 2813 1286 2814 8 2815 8 2816 1287 2817 8 2818 8 2819 8 2820 1288 2821 1289 2822 8 2823 1290 2824 1291 2825 1292 2826 1293 2827 1294 2828 1295 2829 8 2830 8 2831 8 2832 8 2833 1296 2834 1297 2835 1298 2836 1299 2837 8 2838 8 2839 8 2840 1300 2841 8 2842 8 2843 8 2844 8 2845 8 2846 1301 2847 1302 2848 8 2849 1303 2850 8 2851 1304 2852 1305 2853 8 2854 1306 2855 1307 2856 8 2857 1308 2858 1309 2859 8 2860 1310 2861 8 2862 1311 2863 8 2864 8 2865 1312 2866 1313 2867 1314 2868 1315 2869 8 2870 1316 2871 8 2872 8 2873 1317 2874 1318 2875 1319 2876 8 2877 8 2878 8 2879 8 2880 1320 2881 1321 2882 8 2883 8 2884 8 2885 1322 2886 1323 2887 1324 2888 1325 2889 8 2890 8 2891 1326 2892 8 2893 1327 2894 1328 2895 8 2896 8 2897 1329 2898 8 2899 8 2900 8 2901 1330 2902 1331 2903 8 2904 1332 2905 1333 2906 1334 2907 1122 2908 1122 2909 1335 2910 1336 2911 1122 2912 1122 2913 1122 2914 1122 2915 8 2916 1337 2917 1338 2918 1339 2919 8 2920 1340 2921 1341 2922 8 2923 8 2924 8 2925 8 2926 1342 2927 1343 2928 1344 2929 1345 2930 8 2931 8 2932 8 2933 8 2934 8 2935 1346 2936 1347 2937 8 2938 1348 2939 8 2940 1349 2941 1350 2942 8 2943 1351 2944 8 2945 1122 2946 1122 2947 8 2948 1122 2949 1122 2950 1352 2951 1353 2952 8 2953 8 2954 8 2955 8 2956 8 2957 8 2958 8 2959 1354 2960 1355 2961 8 2962 1356 2963 8 2964 1357 2965 1358 2966 8 2967 1359 2968 8 2969 1360 2970 1361 2971 1362 2972 1363 2973 1364 2974 8 2975 8 2976 8 2977 8 2978 8 2979 1365 2980 8 2981 8 2982 8 2983 1366 2984 1367 2985 8 2986 1368 2987 1369 2988 1370 2989 1371 2990 1372 2991 1373 2992 1374 2993 1375 2994 8 2995 1376 2996 1377 2997 8 2998 8 2999 8 3000 8 3001 8 3002 8 3003 8 3004 8 3005 8 3006 1378 3007 8 3008 8 3009 8 3010 1379 3011 1380 3012 8 3013 8 3014 8 3015 1381 3016 1382 3017 1383 3018 8 3019 1384 3020 1385 3021 8 3022 8 3023 8 3024 8 3025 8 3026 8 3027 1386 3028 8 3029 1387 3030 8 3031 1388 3032 1389 3033 1390 3034 1391 3035 1392 3036 8 3037 8 3038 8 3039 8 3040 8 3041 8 3042 8 3043 8 3044 8 3045 1393 3046 1394 3047 1395 3048 8 3049 8 3050 1396 3051 8 3052 1397 3053 8 3054 1398 3055 8 3056 8 3057 8 3058 8 3059 8 3060 8 3061 8 3062 8 3063 1399 3064 8 3065 8 3066 1400 3067 1401 3068 8 3069 1402 3070 8 3071 1403 3072 8 3073 1404 3074 1405 3075 1406 3076 1407 3077 1408 3078 1409 3079 1410 3080 1411 3081 1412 3082 1413 3083 8 3084 1414 3085 8 3086 1415 3087 8 3088 8 3089 1416 3090 1417 3091 1418 3092 1419 3093 8 3094 8 3095 1420 3096 1421 3097 1422 3098 1423 3099 1424 3100 8 3101 1425 3102 8 3103 1426 3104 8 3105 8 3106 8 3107 8 3108 1427 3109 8 3110 1428 3111 1429 3112 8 3113 1430 3114 1431 3115 1432 3116 1433 3117 1434 3118 8 3119 1435 3120 1436 3121 8 3122 1437 3123 8 3124 1438 3125 1439 3126 1440 3127 1441 3128 1442 3129 1443 3130 1444 3131 8 3132 8 3133 1445 3134 8 3135 8 3136 1446 3137 1447 3138 1448 3139 8 3140 8 3141 8 3142 8 3143 1449 3144 8 3145 8 3146 8 3147 8 3148 8 3149 1450 3150 1451 3151 1452 3152 1453 3153 1454 3154 1455 3155 1456 3156 1457 3157 1458 3158 1459 3159 1460 3160 1461 3161 1462 3162 1463 3163 1464 3164 1465 3165 1466 3166 1467 3167 1468 3168 1469 3169 1470 3170 1471 3171 1472 3172 1473 3173 1474 3174 1475 3175 1476 3176 1477 3177 1478 3178 1479 3179 1480 3180 1481 3181 1482 3182 1483 3183 1484 3184 1453 3185 1453 3186 1453 3187 1453 3188 1453 3189 1453 3190 1453 3191 1453 3192 1453 3193 1453 3194 1453 3195 1453 3196 1453 3197 1453 3198 1453 3199 1453 3200 1485 3201 1486 3202 8 3203 8 3204 8 3205 8 3206 8 3207 8 3208 1487 3209 8 3210 8 3211 1488 3212 8 3213 1489 3214 1490 3215 1491 3216 1492 3217 1493 3218 8 3219 8 3220 8 3221 8 3222 8 3223 1494 3224 1495 3225 8 3226 8 3227 1496 3228 8 3229 1497 3230 8 3231 8 3232 1498 3233 1499 3234 8 3235 1453 3236 1453 3237 1453 3238 1453 3239 1453 3240 1500 3241 1501 3242 1502 3243 1453 3244 1453 3245 1453 3246 1453 3247 1453 3248 1453 3249 1453 3250 1453 3251 1453 3252 1453 3253 1503 3254 8 3255 8 3256 8 3257 8 3258 8 3259 8 3260 8 3261 8 3262 8 3263 8 3264 8 3265 8 3266 8 3267 8 3268 8 3269 8 3270 8 3271 8 3272 8 3273 8 3274 8 3275 8 3276 8 3277 8 3278 8 3279 8 3280 8 3281 8 3282 8 3283 8 3284 8 3285 8 3286 8 3287 8 3288 8 3289 8 3290 8 3291 8 3292 8 3293 8 3294 8 3295 8 3296 8 3297 8 3298 8 3299 8 3300 8 3301 8 3302 8 3303 8 3304 8 3305 8 3306 8 3307 8 3308 8 3309 8 3310 8 3311 8 3312 8 3313 8 3314 8 3315 8 3316 8 3317 8 3318 8 3319 8 3320 8 3321 8 3322 8 3323 8 3324 8 3325 8 3326 8 3327 8 3328 8 3329 8 3330 8 3331 8 3332 8 3333 8 3334 1504 3335 8 3336 8 3337 8 3338 1505 3339 1506 3340 8 3341 1507 3342 8 3343 8 3344 1508 3345 1509 3346 1510 3347 8 3348 8 3349 1511 3350 1512 3351 1513 3352 1514 3353 8 3354 1515 3355 8 3356 1516 3357 1517 3358 1518 3359 1519 3360 1520 3361 1521 3362 1522 3363 1523 3364 1524 3365 1453 3366 1453 3367 1453 3368 1453 3369 1453 3370 1453 3371 1453 3372 1453 3373 1453 3374 1453 3375 1453 3376 1453 3377 1453 3378 1525 3379 1453 3380 1526 3381 1527 3382 8 3383 1528 3384 1529 3385 8 3386 1530 3387 8 3388 1531 3389 8 3390 8 3391 8 3392 1532 3393 1533 3394 1534 3395 8 3396 1535 3397 8 3398 8 3399 1536 3400 1537 3401 1538 3402 8 3403 8 3404 8 3405 1539 3406 1540 3407 8 3408 8 3409 8 3410 1541 3411 1542 3412 1543 3413 1544 3414 8 3415 8 3416 1545 3417 8 3418 1546 3419 1547 3420 8 3421 8 3422 1548 3423 8 3424 8 3425 8 3426 1549 3427 1550 3428 8 3429 8 3430 1551 3431 1552 3432 1553 3433 1554 3434 1555 3435 1556 3436 1557 3437 1558 3438 1559 3439 1560 3440 1561 3441 8 3442 1562 3443 1563 3444 1564 3445 1565 3446 1566 3447 1567 3448 1568 3449 1569 3450 1570 3451 1571 3452 8 3453 8 3454 1572 3455 8 3456 1573 3457 8 3458 1574 3459 1575 3460 8 3461 8 3462 8 3463 1576 3464 8 3465 8 3466 1577 3467 1578 3468 1579 3469 1580 3470 8 3471 8 3472 1581 3473 8 3474 1582 3475 1583 3476 8 3477 8 3478 1584 3479 8 3480 1585 3481 8 3482 1586 3483 1587 3484 1588 3485 1453 3486 1453 3487 1589 3488 1590 3489 1591 3490 1592 3491 1593 3492 1594 3493 1595 3494 1596 3495 1597 3496 1598 3497 8 3498 1599 3499 8 3500 1600 3501 8 3502 8 3503 1601 3504 1602 3505 1603 3506 1604 3507 8 3508 8 3509 1605 3510 1606 3511 1607 3512 1608 3513 1609 3514 8 3515 1610 3516 8 3517 1611 3518 8 3519 8 3520 8 3521 8 3522 1612 3523 8 3524 1613 3525 1614 3526 8 3527 1615 3528 1616 3529 1617 3530 1618 3531 1619 3532 8 3533 1620 3534 1621 3535 8 3536 1622 3537 8 3538 1623 3539 1624 3540 1625 3541 1626 3542 1627 3543 1628 3544 1629 3545 8 3546 8 3547 1630 3548 8 3549 8 3550 1631 3551 1632 3552 1633 3553 8 3554 8 3555 8 3556 8 3557 1634 3558 8 3559 8 3560 8 3561 8 3562 8 3563 1635 3564 1636 3565 1637 3566 1638 3567 1639 3568 1640 3569 1641 3570 1642 3571 1643 3572 1644 3573 1645 3574 1646 3575 1647 3576 1648 3577 1649 3578 1650 3579 1651 3580 1652 3581 1653 3582 1654 3583 1655 3584 1656 3585 1657 3586 1658 3587 1659 3588 1660 3589 1661 3590 1662 3591 1663 3592 1664 3593 1665 3594 1666 3595 1667 3596 1668 3597 1669 3598 1638 3599 1638 3600 1638 3601 1638 3602 1638 3603 1638 3604 1638 3605 1638 3606 1638 3607 1638 3608 1638 3609 1638 3610 1638 3611 1638 3612 1638 3613 1638 3614 1670 3615 1671 3616 8 3617 8 3618 8 3619 8 3620 8 3621 8 3622 1672 3623 8 3624 8 3625 1673 3626 8 3627 1674 3628 1675 3629 1676 3630 1677 3631 1678 3632 8 3633 8 3634 8 3635 8 3636 8 3637 1679 3638 1680 3639 8 3640 8 3641 1681 3642 8 3643 1682 3644 8 3645 8 3646 1683 3647 1684 3648 8 3649 1638 3650 1638 3651 1638 3652 1638 3653 1638 3654 1685 3655 1686 3656 1687 3657 1638 3658 1638 3659 1638 3660 1638 3661 1638 3662 1638 3663 1638 3664 1638 3665 1638 3666 1638 3667 1688 3668 8 3669 8 3670 8 3671 8 3672 8 3673 8 3674 8 3675 8 3676 8 3677 8 3678 8 3679 8 3680 8 3681 8 3682 8 3683 8 3684 8 3685 8 3686 8 3687 8 3688 8 3689 8 3690 8 3691 8 3692 8 3693 8 3694 8 3695 8 3696 8 3697 8 3698 8 3699 8 3700 8 3701 8 3702 8 3703 8 3704 8 3705 8 3706 8 3707 8 3708 8 3709 8 3710 8 3711 8 3712 8 3713 8 3714 8 3715 8 3716 8 3717 8 3718 8 3719 8 3720 8 3721 8 3722 8 3723 8 3724 8 3725 8 3726 8 3727 8 3728 8 3729 8 3730 8 3731 8 3732 8 3733 8 3734 8 3735 8 3736 8 3737 8 3738 8 3739 8 3740 8 3741 8 3742 8 3743 8 3744 8 3745 8 3746 8 3747 8 3748 1689 3749 8 3750 8 3751 8 3752 1690 3753 1691 3754 8 3755 1692 3756 8 3757 8 3758 1693 3759 1694 3760 1695 3761 8 3762 8 3763 1696 3764 1697 3765 1698 3766 1699 3767 8 3768 1700 3769 8 3770 1701 3771 1702 3772 1703 3773 1704 3774 1705 3775 1706 3776 1707 3777 1708 3778 1709 3779 1638 3780 1638 3781 1638 3782 1638 3783 1638 3784 1638 3785 1638 3786 1638 3787 1638 3788 1638 3789 1638 3790 1638 3791 1638 3792 1710 3793 1638 3794 1711 3795 1712 3796 8 3797 1713 3798 1714 3799 8 3800 1715 3801 8 3802 1716 3803 8 3804 8 3805 8 3806 1717 3807 1718 3808 1719 3809 8 3810 1720 3811 8 3812 8 3813 1721 3814 1722 3815 1723 3816 8 3817 8 3818 8 3819 1724 3820 1725 3821 8 3822 8 3823 8 3824 1726 3825 1727 3826 1728 3827 1729 3828 8 3829 8 3830 1730 3831 8 3832 1731 3833 1732 3834 8 3835 8 3836 1733 3837 8 3838 8 3839 8 3840 1734 3841 1735 3842 8 3843 8 3844 1736 3845 1737 3846 1738 3847 1739 3848 1740 3849 1741 3850 1742 3851 1743 3852 1744 3853 1745 3854 1746 3855 8 3856 1747 3857 1748 3858 1749 3859 1750 3860 1751 3861 1752 3862 1753 3863 1754 3864 1755 3865 1756 3866 8 3867 1757 3868 1453 3869 1758 3870 8 3871 1759 3872 8 3873 1760 3874 1761 3875 8 3876 8 3877 8 3878 1762 3879 8 3880 8 3881 1763 3882 1764 3883 1765 3884 1766 3885 8 3886 8 3887 1767 3888 8 3889 1768 3890 1769 3891 8 3892 8 3893 1770 3894 8 3895 1771 3896 8 3897 1772 3898 1773 3899 1774 3900 1638 3901 1638 3902 1775 3903 1776 3904 1638 3905 1638 3906 1638 3907 1638 3908 8 3909 1777 3910 8 3911 1778 3912 8 3913 8 3914 1638 3915 1638 3916 8 3917 1638 3918 1638 3919 1779 3920 8 3921 8 3922 8 3923 8 3924 8 3925 8 3926 1780 3927 1781 3928 1782 3929 1783 3930 8 3931 1784 3932 8 3933 1785 3934 1786 3935 1787 3936 1788 3937 1789 3938 8 3939 8 3940 8 3941 8 3942 8 3943 1790 3944 8 3945 8 3946 1791 3947 1792 3948 1793 3949 1794 3950 1795 3951 1796 3952 1797 3953 8 3954 1798 3955 8 3956 1799 3957 1800 3958 1801 3959 1802 3960 1803 3961 1804 3962 1805 3963 1806 3964 1807 3965 1808 3966 1809 3967 1810 3968 1811 3969 1812 3970 8 3971 1813 3972 8 3973 1814 3974 1815 3975 8 3976 8 3977 1816 3978 1817 3979 1818 3980 1819 3981 8 3982 8 3983 1820 3984 1821 3985 1822 3986 1823 3987 1824 3988 8 3989 1825 3990 8 3991 1826 3992 8 3993 8 3994 8 3995 8 3996 1827 3997 8 3998 1828 3999 1829 4000 8 4001 1830 4002 1831 4003 1832 4004 1833 4005 1834 4006 8 4007 1835 4008 1836 4009 8 4010 1837 4011 8 4012 1838 4013 1839 4014 1840 4015 1841 4016 1842 4017 1843 4018 1844 4019 8 4020 8 4021 1845 4022 8 4023 8 4024 1846 4025 1847 4026 1848 4027 8 4028 8 4029 8 4030 8 4031 8 4032 1849 4033 1850 4034 1851 4035 8 4036 8 4037 1852 4038 8 4039 8 4040 8 4041 1853 4042 1854 4043 1855 4044 1846 4045 1856 4046 1857 4047 1858 4048 1859 4049 1860 4050 1861 4051 1862 4052 1863 4053 1864 4054 1865 4055 1866 4056 1867 4057 1868 4058 1869 4059 1870 4060 1871 4061 1872 4062 1873 4063 1874 4064 1875 4065 1876 4066 1877 4067 1878 4068 1879 4069 1880 4070 1881 4071 1882 4072 1883 4073 1884 4074 1885 4075 1886 4076 1846 4077 1846 4078 1846 4079 1846 4080 1846 4081 1846 4082 1846 4083 1846 4084 1846 4085 1846 4086 1846 4087 1846 4088 1846 4089 1846 4090 1846 4091 1846 4092 1887 4093 1888 4094 8 4095 8 4096 8 4097 8 4098 8 4099 8 4100 1889 4101 8 4102 8 4103 1890 4104 8 4105 1891 4106 1892 4107 1893 4108 1894 4109 1895 4110 8 4111 8 4112 8 4113 8 4114 8 4115 1896 4116 1897 4117 8 4118 8 4119 1898 4120 8 4121 1899 4122 8 4123 8 4124 1900 4125 1901 4126 8 4127 1846 4128 1846 4129 1846 4130 1846 4131 1846 4132 1902 4133 1903 4134 1904 4135 1846 4136 1846 4137 1846 4138 1846 4139 1846 4140 1846 4141 1846 4142 1846 4143 1846 4144 1846 4145 1905 4146 8 4147 8 4148 8 4149 8 4150 8 4151 8 4152 8 4153 8 4154 8 4155 8 4156 8 4157 8 4158 8 4159 8 4160 8 4161 8 4162 8 4163 8 4164 8 4165 8 4166 8 4167 8 4168 8 4169 8 4170 8 4171 8 4172 8 4173 8 4174 8 4175 8 4176 8 4177 8 4178 8 4179 8 4180 8 4181 8 4182 8 4183 8 4184 8 4185 8 4186 8 4187 8 4188 8 4189 8 4190 8 4191 8 4192 8 4193 8 4194 8 4195 8 4196 8 4197 8 4198 8 4199 8 4200 8 4201 8 4202 8 4203 8 4204 8 4205 8 4206 8 4207 8 4208 8 4209 8 4210 8 4211 8 4212 8 4213 8 4214 8 4215 8 4216 8 4217 8 4218 8 4219 8 4220 8 4221 8 4222 8 4223 8 4224 8 4225 8 4226 1906 4227 8 4228 8 4229 8 4230 1907 4231 1908 4232 8 4233 1909 4234 8 4235 8 4236 1910 4237 1911 4238 1912 4239 8 4240 8 4241 1913 4242 1914 4243 1915 4244 1916 4245 8 4246 1917 4247 8 4248 1918 4249 1919 4250 1920 4251 1921 4252 1922 4253 1923 4254 1924 4255 1925 4256 1926 4257 1846 4258 1846 4259 1846 4260 1846 4261 1846 4262 1846 4263 1846 4264 1846 4265 1846 4266 1846 4267 1846 4268 1846 4269 1846 4270 1927 4271 1846 4272 1928 4273 1929 4274 8 4275 1930 4276 1931 4277 8 4278 1932 4279 8 4280 1933 4281 8 4282 8 4283 8 4284 1934 4285 1935 4286 1936 4287 8 4288 1937 4289 8 4290 8 4291 1938 4292 1939 4293 1940 4294 8 4295 8 4296 8 4297 1941 4298 1942 4299 8 4300 8 4301 8 4302 1943 4303 1944 4304 1945 4305 1946 4306 8 4307 8 4308 1947 4309 8 4310 1948 4311 1949 4312 8 4313 8 4314 1950 4315 8 4316 8 4317 8 4318 1951 4319 1952 4320 8 4321 1953 4322 8 4323 8 4324 8 4325 8 4326 1954 4327 1955 4328 1956 4329 1957 4330 1958 4331 1959 4332 1960 4333 1961 4334 1962 4335 1963 4336 1964 4337 1965 4338 1966 4339 8 4340 1967 4341 8 4342 1968 4343 8 4344 8 4345 1969 4346 1970 4347 1971 4348 1972 4349 8 4350 8 4351 1973 4352 1974 4353 1975 4354 1976 4355 1977 4356 8 4357 1978 4358 8 4359 1979 4360 8 4361 8 4362 8 4363 8 4364 1980 4365 8 4366 1981 4367 1982 4368 8 4369 1983 4370 1984 4371 1985 4372 1986 4373 1987 4374 8 4375 1988 4376 1989 4377 8 4378 1990 4379 8 4380 1991 4381 1992 4382 1993 4383 1994 4384 1995 4385 1996 4386 1997 4387 8 4388 8 4389 1998 4390 8 4391 8 4392 1999 4393 2000 4394 2001 4395 8 4396 8 4397 8 4398 8 4399 2002 4400 8 4401 8 4402 8 4403 8 4404 8 4405 2003 4406 2004 4407 2005 4408 1999 4409 2006 4410 2007 4411 2008 4412 2009 4413 2010 4414 2011 4415 2012 4416 2013 4417 2014 4418 2015 4419 2016 4420 2017 4421 2018 4422 2019 4423 2020 4424 2021 4425 2022 4426 2023 4427 2024 4428 2025 4429 2026 4430 2027 4431 2028 4432 2029 4433 2030 4434 2031 4435 2032 4436 2033 4437 2034 4438 2035 4439 2036 4440 1999 4441 1999 4442 1999 4443 1999 4444 1999 4445 1999 4446 1999 4447 1999 4448 1999 4449 1999 4450 1999 4451 1999 4452 1999 4453 1999 4454 1999 4455 1999 4456 2037 4457 2038 4458 8 4459 8 4460 8 4461 8 4462 8 4463 8 4464 2039 4465 8 4466 8 4467 2040 4468 8 4469 2041 4470 2042 4471 2043 4472 2044 4473 2045 4474 8 4475 8 4476 8 4477 8 4478 8 4479 2046 4480 2047 4481 8 4482 8 4483 2048 4484 8 4485 2049 4486 8 4487 8 4488 2050 4489 2051 4490 8 4491 1999 4492 1999 4493 1999 4494 1999 4495 1999 4496 2052 4497 2053 4498 2054 4499 1999 4500 1999 4501 1999 4502 1999 4503 1999 4504 1999 4505 1999 4506 1999 4507 1999 4508 1999 4509 2055 4510 8 4511 8 4512 8 4513 8 4514 8 4515 8 4516 8 4517 8 4518 8 4519 8 4520 8 4521 8 4522 8 4523 8 4524 8 4525 8 4526 8 4527 8 4528 8 4529 8 4530 8 4531 8 4532 8 4533 8 4534 8 4535 8 4536 8 4537 8 4538 8 4539 8 4540 8 4541 8 4542 8 4543 8 4544 8 4545 8 4546 8 4547 8 4548 8 4549 8 4550 8 4551 8 4552 8 4553 8 4554 8 4555 8 4556 8 4557 8 4558 8 4559 8 4560 8 4561 8 4562 8 4563 8 4564 8 4565 8 4566 8 4567 8 4568 8 4569 8 4570 8 4571 8 4572 8 4573 8 4574 8 4575 8 4576 8 4577 8 4578 8 4579 8 4580 8 4581 8 4582 8 4583 8 4584 8 4585 8 4586 8 4587 8 4588 8 4589 8 4590 2056 4591 8 4592 8 4593 8 4594 2057 4595 2058 4596 8 4597 2059 4598 8 4599 8 4600 2060 4601 2061 4602 2062 4603 8 4604 8 4605 2063 4606 2064 4607 2065 4608 2066 4609 8 4610 2067 4611 8 4612 2068 4613 2069 4614 2070 4615 2071 4616 2072 4617 2073 4618 2074 4619 2075 4620 2076 4621 1999 4622 1999 4623 1999 4624 1999 4625 1999 4626 1999 4627 1999 4628 1999 4629 1999 4630 1999 4631 1999 4632 1999 4633 1999 4634 2077 4635 1999 4636 2078 4637 2079 4638 8 4639 2080 4640 2081 4641 8 4642 2082 4643 8 4644 2083 4645 8 4646 8 4647 8 4648 2084 4649 2085 4650 2086 4651 8 4652 2087 4653 8 4654 8 4655 2088 4656 2089 4657 2090 4658 8 4659 8 4660 8 4661 2091 4662 2092 4663 8 4664 8 4665 8 4666 2093 4667 2094 4668 2095 4669 2096 4670 8 4671 8 4672 2097 4673 8 4674 2098 4675 2099 4676 8 4677 8 4678 2100 4679 8 4680 8 4681 8 4682 2101 4683 2102 4684 8 4685 8 4686 2103 4687 2104 4688 2105 4689 2106 4690 2107 4691 2108 4692 2109 4693 2110 4694 2111 4695 2112 4696 2113 4697 8 4698 2114 4699 2115 4700 2116 4701 2117 4702 2118 4703 2119 4704 2120 4705 2121 4706 2122 4707 2123 4708 8 4709 2124 4710 8 4711 2125 4712 8 4713 2126 4714 2127 4715 8 4716 8 4717 8 4718 2128 4719 8 4720 8 4721 2129 4722 2130 4723 2131 4724 2132 4725 8 4726 8 4727 2133 4728 8 4729 2134 4730 2135 4731 8 4732 8 4733 2136 4734 8 4735 2137 4736 8 4737 2138 4738 2139 4739 2140 4740 1999 4741 1999 4742 2141 4743 2142 4744 2143 4745 2144 4746 2145 4747 2146 4748 2147 4749 2148 4750 2149 4751 2150 4752 8 4753 2151 4754 8 4755 2152 4756 8 4757 8 4758 2153 4759 2154 4760 2155 4761 2156 4762 8 4763 8 4764 2157 4765 2158 4766 2159 4767 2160 4768 2161 4769 8 4770 2162 4771 8 4772 2163 4773 8 4774 8 4775 8 4776 8 4777 2164 4778 8 4779 2165 4780 2166 4781 8 4782 2167 4783 2168 4784 2169 4785 2170 4786 2171 4787 8 4788 2172 4789 2173 4790 8 4791 2174 4792 8 4793 2175 4794 2176 4795 2177 4796 2178 4797 2179 4798 2180 4799 2181 4800 8 4801 8 4802 2182 4803 8 4804 8 4805 2183 4806 2184 4807 2185 4808 8 4809 8 4810 8 4811 8 4812 2186 4813 8 4814 8 4815 8 4816 8 4817 8 4818 2187 4819 2188 4820 2189 4821 2190 4822 2191 4823 2192 4824 2193 4825 2194 4826 2195 4827 2196 4828 2197 4829 2198 4830 2199 4831 2200 4832 2201 4833 2202 4834 2203 4835 2204 4836 2205 4837 2206 4838 2207 4839 2208 4840 2209 4841 2210 4842 2211 4843 2212 4844 2213 4845 2214 4846 2215 4847 2216 4848 2217 4849 2218 4850 2219 4851 2220 4852 2221 4853 2190 4854 2190 4855 2190 4856 2190 4857 2190 4858 2190 4859 2190 4860 2190 4861 2190 4862 2190 4863 2190 4864 2190 4865 2190 4866 2190 4867 2190 4868 2190 4869 2222 4870 2223 4871 8 4872 8 4873 8 4874 8 4875 8 4876 8 4877 2224 4878 8 4879 8 4880 2225 4881 8 4882 2226 4883 2227 4884 2228 4885 2229 4886 2230 4887 8 4888 8 4889 8 4890 8 4891 8 4892 2231 4893 2232 4894 8 4895 8 4896 2233 4897 8 4898 2234 4899 8 4900 8 4901 2235 4902 2236 4903 8 4904 2190 4905 2190 4906 2190 4907 2190 4908 2190 4909 2237 4910 2238 4911 2239 4912 2190 4913 2190 4914 2190 4915 2190 4916 2190 4917 2190 4918 2190 4919 2190 4920 2190 4921 2190 4922 2240 4923 8 4924 8 4925 8 4926 8 4927 8 4928 8 4929 8 4930 8 4931 8 4932 8 4933 8 4934 8 4935 8 4936 8 4937 8 4938 8 4939 8 4940 8 4941 8 4942 8 4943 8 4944 8 4945 8 4946 8 4947 8 4948 8 4949 8 4950 8 4951 8 4952 8 4953 8 4954 8 4955 8 4956 8 4957 8 4958 8 4959 8 4960 8 4961 8 4962 8 4963 8 4964 8 4965 8 4966 8 4967 8 4968 8 4969 8 4970 8 4971 8 4972 8 4973 8 4974 8 4975 8 4976 8 4977 8 4978 8 4979 8 4980 8 4981 8 4982 8 4983 8 4984 8 4985 8 4986 8 4987 8 4988 8 4989 8 4990 8 4991 8 4992 8 4993 8 4994 8 4995 8 4996 8 4997 8 4998 8 4999 8 5000 8 5001 8 5002 8 5003 2241 5004 8 5005 8 5006 8 5007 2242 5008 2243 5009 8 5010 2244 5011 8 5012 8 5013 2245 5014 2246 5015 2247 5016 8 5017 8 5018 2248 5019 2249 5020 2250 5021 2251 5022 8 5023 2252 5024 8 5025 2253 5026 2254 5027 2255 5028 2256 5029 2257 5030 2258 5031 2259 5032 2260 5033 2261 5034 2190 5035 2190 5036 2190 5037 2190 5038 2190 5039 2190 5040 2190 5041 2190 5042 2190 5043 2190 5044 2190 5045 2190 5046 2190 5047 2262 5048 2190 5049 2263 5050 2264 5051 8 5052 2265 5053 2266 5054 8 5055 2267 5056 8 5057 2268 5058 8 5059 8 5060 8 5061 2269 5062 2270 5063 2271 5064 8 5065 2272 5066 8 5067 8 5068 2273 5069 2274 5070 2275 5071 8 5072 8 5073 8 5074 2276 5075 2277 5076 8 5077 8 5078 8 5079 2278 5080 2279 5081 2280 5082 2281 5083 8 5084 8 5085 2282 5086 8 5087 2283 5088 2284 5089 8 5090 8 5091 2285 5092 8 5093 8 5094 8 5095 2286 5096 2287 5097 8 5098 8 5099 2288 5100 2289 5101 2290 5102 2291 5103 2292 5104 2293 5105 2294 5106 2295 5107 2296 5108 2297 5109 2298 5110 8 5111 2299 5112 2300 5113 2301 5114 2302 5115 2303 5116 2304 5117 2305 5118 2306 5119 2307 5120 2308 5121 8 5122 2309 5123 1999 5124 2310 5125 8 5126 2311 5127 8 5128 2312 5129 2313 5130 8 5131 8 5132 8 5133 2314 5134 8 5135 8 5136 2315 5137 2316 5138 2317 5139 2318 5140 8 5141 8 5142 2319 5143 8 5144 2320 5145 2321 5146 8 5147 8 5148 2322 5149 8 5150 2323 5151 8 5152 2324 5153 2325 5154 2326 5155 2190 5156 2190 5157 2327 5158 2328 5159 2190 5160 2190 5161 2190 5162 2190 5163 8 5164 2329 5165 8 5166 2330 5167 8 5168 8 5169 2190 5170 2190 5171 8 5172 2190 5173 2190 5174 2331 5175 8 5176 8 5177 8 5178 8 5179 8 5180 8 5181 2332 5182 2333 5183 2334 5184 2335 5185 8 5186 2336 5187 8 5188 2337 5189 2338 5190 2339 5191 2340 5192 2341 5193 8 5194 8 5195 8 5196 8 5197 8 5198 2342 5199 8 5200 8 5201 2343 5202 2344 5203 2345 5204 2346 5205 2347 5206 2348 5207 2349 5208 8 5209 2350 5210 8 5211 2351 5212 2352 5213 2353 5214 2354 5215 2355 5216 8 5217 2356 5218 2357 5219 2358 5220 2359 5221 2190 5222 2360 5223 564 5224 8 5225 2361 5226 2362 5227 2363 5228 8 5229 8 5230 2190 5231 2190 5232 2190 5233 8 5234 2364 5235 8 5236 8 5237 8 5238 2365 5239 2366 5240 2367 5241 8 5242 8 5243 2368 5244 2369 5245 8 5246 2370 5247 2371 5248 8 5249 2190 5250 8 5251 2372 5252 8 5253 2190 5254 8 5255 1086 5256 8 5257 1089 5258 8 5259 2373 5260 1092 5261 1093 5262 1094 5263 8 5264 8 5265 1096 5266 1097 5267 1099 5268 8 5269 1129 5270 1101 5271 8 5272 1102 5273 8 5274 1104 5275 8 5276 1105 5277 1106 5278 1107 5279 1109 5280 8 5281 1111 5282 8 5283 1113 5284 8 5285 2374 5286 2375 5287 2376 5288 2377 5289 8 5290 8 5291 8 5292 2378 5293 8 5294 8 5295 2379 5296 2380 5297 2381 5298 2382 5299 2383 5300 8 5301 2384 5302 2385 5303 2386 5304 8 5305 1122 5306 1122 5307 2387 5308 8 5309 8 5310 8 5311 2388 5312 2389 5313 2390 5314 8 5315 8 5316 2391 5317 2392 5318 8 5319 2393 5320 2394 5321 8 5322 8 5323 2395 5324 8 5325 1122 5326 8 5327 2396 5328 2397 5329 2398 5330 2399 5331 8 5332 8 5333 2400 5334 8 5335 8 5336 2401 5337 2402 5338 8 5339 2190 5340 2190 5341 2403 5342 2404 5343 8 5344 2190 5345 2190 5346 2190 5347 2405 5348 8 5349 2406 5350 2407 5351 8 5352 2408 5353 2409 5354 2410 5355 2411 5356 2412 5357 8 5358 8 5359 8 5360 8 5361 2413 5362 2414 5363 2415 5364 2416 5365 8 5366 8 5367 8 5368 8 5369 2417 5370 2418 5371 8 5372 2419 5373 8 5374 2420 5375 2421 5376 8 5377 2422 5378 8 5379 2190 5380 1999 5381 1999 5382 1999 5383 8 5384 2423 5385 8 5386 2424 5387 8 5388 8 5389 1999 5390 1999 5391 8 5392 1999 5393 1999 5394 2425 5395 8 5396 8 5397 8 5398 8 5399 8 5400 8 5401 2426 5402 2427 5403 2428 5404 2429 5405 8 5406 2430 5407 8 5408 2431 5409 2432 5410 2433 5411 2434 5412 2435 5413 8 5414 8 5415 8 5416 8 5417 8 5418 2436 5419 8 5420 8 5421 2437 5422 2438 5423 2439 5424 2440 5425 2441 5426 2442 5427 2443 5428 8 5429 2444 5430 8 5431 2445 5432 2446 5433 2447 5434 2448 5435 2449 5436 8 5437 2450 5438 2451 5439 2452 5440 2453 5441 1999 5442 2454 5443 2455 5444 2456 5445 8 5446 1999 5447 1999 5448 1999 5449 8 5450 2457 5451 8 5452 8 5453 8 5454 2458 5455 2459 5456 2460 5457 8 5458 8 5459 2461 5460 2462 5461 8 5462 2463 5463 2464 5464 8 5465 1999 5466 8 5467 2465 5468 8 5469 1999 5470 8 5471 2466 5472 2467 5473 2468 5474 2469 5475 8 5476 8 5477 2470 5478 8 5479 8 5480 1999 5481 1999 5482 2471 5483 2472 5484 8 5485 1999 5486 1999 5487 1999 5488 2473 5489 8 5490 2474 5491 2475 5492 8 5493 2476 5494 8 5495 1999 5496 1846 5497 2477 5498 2478 5499 2479 5500 2480 5501 2481 5502 2482 5503 2483 5504 2484 5505 2485 5506 2486 5507 2487 5508 8 5509 2488 5510 2489 5511 2490 5512 2491 5513 2492 5514 2493 5515 2494 5516 2495 5517 2496 5518 2497 5519 8 5520 2498 5521 8 5522 2499 5523 8 5524 2500 5525 2501 5526 8 5527 8 5528 8 5529 2502 5530 8 5531 8 5532 2503 5533 2504 5534 2505 5535 2506 5536 8 5537 8 5538 2507 5539 8 5540 2508 5541 2509 5542 8 5543 8 5544 2510 5545 8 5546 2511 5547 8 5548 2512 5549 2513 5550 2514 5551 1846 5552 1846 5553 2515 5554 2516 5555 1846 5556 1846 5557 1846 5558 1846 5559 8 5560 2517 5561 8 5562 2518 5563 8 5564 8 5565 1846 5566 1846 5567 8 5568 1846 5569 1846 5570 2519 5571 8 5572 8 5573 8 5574 8 5575 8 5576 8 5577 2520 5578 2521 5579 2522 5580 2523 5581 8 5582 2524 5583 8 5584 2525 5585 2526 5586 2527 5587 2528 5588 2529 5589 8 5590 8 5591 8 5592 8 5593 8 5594 2530 5595 8 5596 8 5597 2531 5598 2532 5599 2533 5600 2534 5601 2535 5602 2536 5603 2537 5604 8 5605 2538 5606 8 5607 2539 5608 2540 5609 2541 5610 2542 5611 2543 5612 8 5613 2544 5614 2545 5615 2546 5616 2547 5617 1846 5618 2548 5619 2549 5620 2550 5621 8 5622 1846 5623 1846 5624 1846 5625 8 5626 2551 5627 8 5628 8 5629 8 5630 2552 5631 2553 5632 2554 5633 8 5634 8 5635 2555 5636 2556 5637 8 5638 2557 5639 2558 5640 8 5641 1846 5642 8 5643 2559 5644 2560 5645 1999 5646 8 5647 2561 5648 8 5649 1846 5650 8 5651 2562 5652 2563 5653 2564 5654 2565 5655 8 5656 8 5657 2566 5658 8 5659 8 5660 1846 5661 1846 5662 2567 5663 2568 5664 8 5665 1846 5666 1846 5667 1846 5668 2569 5669 8 5670 2570 5671 2571 5672 8 5673 2572 5674 8 5675 1846 5676 2573 5677 2574 5678 2575 5679 2576 5680 1638 5681 2577 5682 2578 5683 2579 5684 8 5685 1638 5686 1638 5687 1638 5688 8 5689 2580 5690 8 5691 8 5692 8 5693 2581 5694 2582 5695 2583 5696 8 5697 8 5698 2584 5699 2585 5700 8 5701 2586 5702 2587 5703 8 5704 1638 5705 8 5706 2588 5707 8 5708 1638 5709 8 5710 2589 5711 2590 5712 2591 5713 2592 5714 8 5715 8 5716 2593 5717 8 5718 8 5719 1638 5720 1638 5721 2594 5722 2595 5723 8 5724 1638 5725 1638 5726 1638 5727 2596 5728 8 5729 2597 5730 2598 5731 8 5732 2599 5733 8 5734 1638 5735 1453 5736 1453 5737 1453 5738 8 5739 2600 5740 8 5741 2601 5742 8 5743 8 5744 1453 5745 1453 5746 8 5747 1453 5748 1453 5749 2602 5750 8 5751 8 5752 8 5753 8 5754 8 5755 8 5756 2603 5757 2604 5758 2605 5759 2606 5760 8 5761 2607 5762 8 5763 2608 5764 2609 5765 2610 5766 2611 5767 2612 5768 8 5769 8 5770 8 5771 8 5772 8 5773 2613 5774 8 5775 8 5776 2614 5777 2615 5778 2616 5779 2617 5780 2618 5781 2619 5782 2620 5783 8 5784 2621 5785 8 5786 2622 5787 2623 5788 2624 5789 2625 5790 2626 5791 8 5792 2627 5793 2628 5794 2629 5795 2630 5796 1453 5797 2631 5798 2632 5799 2633 5800 8 5801 1453 5802 1453 5803 1453 5804 8 5805 2634 5806 8 5807 8 5808 8 5809 2635 5810 2636 5811 2637 5812 8 5813 8 5814 2638 5815 2639 5816 8 5817 2640 5818 2641 5819 8 5820 1453 5821 8 5822 2642 5823 8 5824 1453 5825 8 5826 2643 5827 2644 5828 2645 5829 2646 5830 8 5831 8 5832 2647 5833 8 5834 8 5835 1453 5836 1453 5837 2648 5838 2649 5839 8 5840 1453 5841 1453 5842 1453 5843 2650 5844 8 5845 2651 5846 2652 5847 8 5848 2653 5849 8 5850 1453 5851 2654 5852 2655 5853 2656 5854 2657 5855 8 5856 2658 5857 8 5858 8 5859 8 5860 8 5861 8 5862 8 5863 8 5864 8 5865 8 5866 8 5867 2659 5868 8 5869 2660 5870 8 5871 2661 5872 8 5873 8 5874 8 5875 8 5876 2662 5877 2663 5878 2664 5879 2665 5880 2666 5881 2667 5882 8 5883 2668 5884 2669 5885 2670 5886 2671 5887 8 5888 8 5889 2672 5890 2673 5891 2674 5892 2675 5893 2676 5894 2677 5895 2678 5896 8 5897 2679 5898 2680 5899 2681 5900 2682 5901 2683 5902 2684 5903 2685 5904 2686 5905 2687 5906 2688 5907 2689 5908 2690 5909 2691 5910 2692 5911 2693 5912 2694 5913 2695 5914 2696 5915 2697 5916 2698 5917 2699 5918 2700 5919 2701 5920 2702 5921 2703 5922 85 5923 85 5924 85 5925 85 5926 85 5927 85 5928 85 5929 85 5930 85 5931 85 5932 85 5933 85 5934 85 5935 85 5936 85 5937 85 5938 2704 5939 2705 5940 8 5941 8 5942 8 5943 8 5944 8 5945 8 5946 2706 5947 8 5948 8 5949 2707 5950 8 5951 2708 5952 2709 5953 2710 5954 2711 5955 2712 5956 8 5957 8 5958 8 5959 8 5960 8 5961 2713 5962 2714 5963 8 5964 8 5965 2715 5966 8 5967 2716 5968 8 5969 8 5970 2717 5971 2718 5972 8 5973 85 5974 85 5975 85 5976 85 5977 85 5978 2719 5979 2720 5980 2721 5981 85 5982 85 5983 85 5984 85 5985 85 5986 85 5987 85 5988 85 5989 85 5990 85 5991 2722 5992 8 5993 8 5994 8 5995 8 5996 8 5997 8 5998 8 5999 8 6000 8 6001 8 6002 2723 6003 2724 6004 8 6005 2725 6006 8 6007 8 6008 2726 6009 2727 6010 2728 6011 8 6012 8 6013 2729 6014 2730 6015 8 6016 8 6017 2731 6018 2732 6019 2733 6020 2734 6021 2735 6022 2736 6023 2737 6024 2738 6025 2739 6026 2740 6027 8 6028 2741 6029 8 6030 2742 6031 8 6032 8 6033 2743 6034 2744 6035 2745 6036 2746 6037 8 6038 8 6039 2747 6040 2748 6041 2749 6042 2750 6043 2751 6044 8 6045 2752 6046 8 6047 2753 6048 8 6049 8 6050 8 6051 8 6052 2754 6053 8 6054 2755 6055 2756 6056 8 6057 2757 6058 2758 6059 2759 6060 2760 6061 2761 6062 8 6063 2762 6064 2763 6065 8 6066 2764 6067 8 6068 2765 6069 2766 6070 2767 6071 2768 6072 2769 6073 2770 6074 2771 6075 8 6076 8 6077 2772 6078 8 6079 8 6080 2773 6081 2774 6082 2775 6083 8 6084 8 6085 8 6086 8 6087 2776 6088 8 6089 8 6090 8 6091 8 6092 8 6093 2777 6094 2778 6095 2779 6096 2780 6097 2781 6098 2782 6099 2783 6100 2784 6101 2785 6102 2786 6103 2787 6104 2788 6105 2789 6106 2790 6107 2791 6108 2792 6109 2793 6110 2794 6111 2795 6112 2796 6113 2797 6114 2798 6115 2799 6116 2800 6117 2801 6118 2802 6119 2803 6120 2804 6121 2805 6122 2806 6123 2807 6124 2808 6125 2809 6126 2810 6127 2811 6128 2780 6129 2780 6130 2780 6131 2780 6132 2780 6133 2780 6134 2780 6135 2780 6136 2780 6137 2780 6138 2780 6139 2780 6140 2780 6141 2780 6142 2780 6143 2780 6144 2812 6145 2813 6146 8 6147 8 6148 8 6149 8 6150 8 6151 8 6152 2814 6153 8 6154 8 6155 2815 6156 8 6157 2816 6158 2817 6159 2818 6160 2819 6161 2820 6162 8 6163 8 6164 8 6165 8 6166 8 6167 2821 6168 2822 6169 8 6170 8 6171 2823 6172 8 6173 2824 6174 8 6175 8 6176 2825 6177 2826 6178 8 6179 2780 6180 2780 6181 2780 6182 2780 6183 2780 6184 2827 6185 2828 6186 2829 6187 2780 6188 2780 6189 2780 6190 2780 6191 2780 6192 2780 6193 2780 6194 2780 6195 2780 6196 2780 6197 2830 6198 8 6199 8 6200 8 6201 8 6202 8 6203 8 6204 8 6205 8 6206 8 6207 8 6208 8 6209 8 6210 8 6211 8 6212 8 6213 8 6214 8 6215 8 6216 8 6217 8 6218 8 6219 8 6220 8 6221 8 6222 8 6223 8 6224 8 6225 8 6226 8 6227 8 6228 8 6229 8 6230 8 6231 8 6232 8 6233 8 6234 8 6235 8 6236 8 6237 8 6238 8 6239 8 6240 8 6241 8 6242 8 6243 8 6244 8 6245 8 6246 8 6247 8 6248 8 6249 8 6250 8 6251 8 6252 8 6253 8 6254 8 6255 8 6256 8 6257 8 6258 8 6259 8 6260 8 6261 8 6262 8 6263 8 6264 8 6265 8 6266 8 6267 8 6268 8 6269 8 6270 8 6271 8 6272 8 6273 8 6274 8 6275 8 6276 8 6277 8 6278 2831 6279 8 6280 8 6281 8 6282 2832 6283 2833 6284 8 6285 2834 6286 8 6287 8 6288 2835 6289 2836 6290 2837 6291 8 6292 8 6293 2838 6294 2839 6295 2840 6296 2841 6297 8 6298 2842 6299 8 6300 2843 6301 2844 6302 2845 6303 2846 6304 2847 6305 2848 6306 2849 6307 2850 6308 2851 6309 2780 6310 2780 6311 2780 6312 2780 6313 2780 6314 2780 6315 2780 6316 2780 6317 2780 6318 2780 6319 2780 6320 2780 6321 2780 6322 2852 6323 2780 6324 2853 6325 2854 6326 8 6327 2855 6328 2856 6329 8 6330 2857 6331 8 6332 2858 6333 8 6334 8 6335 8 6336 2859 6337 2860 6338 2861 6339 8 6340 2862 6341 8 6342 8 6343 2863 6344 2864 6345 2865 6346 8 6347 8 6348 8 6349 2866 6350 2867 6351 8 6352 8 6353 8 6354 2868 6355 2869 6356 2870 6357 2871 6358 8 6359 8 6360 2872 6361 8 6362 2873 6363 2874 6364 8 6365 8 6366 2875 6367 8 6368 8 6369 8 6370 2876 6371 2877 6372 8 6373 8 6374 2878 6375 2879 6376 2880 6377 2881 6378 2882 6379 2883 6380 2884 6381 2885 6382 2886 6383 2887 6384 2888 6385 8 6386 2889 6387 2890 6388 2891 6389 2892 6390 2893 6391 2894 6392 2895 6393 2896 6394 2897 6395 2898 6396 8 6397 2899 6398 2900 6399 2901 6400 8 6401 2902 6402 8 6403 2903 6404 8 6405 2904 6406 2905 6407 8 6408 8 6409 8 6410 2906 6411 8 6412 8 6413 2907 6414 2908 6415 2909 6416 2910 6417 8 6418 8 6419 2911 6420 8 6421 2912 6422 2913 6423 8 6424 8 6425 2914 6426 8 6427 2915 6428 8 6429 2916 6430 2917 6431 2918 6432 2780 6433 2780 6434 2919 6435 2920 6436 2921 6437 2922 6438 2923 6439 2924 6440 2925 6441 2926 6442 2927 6443 2928 6444 8 6445 2929 6446 8 6447 2930 6448 8 6449 8 6450 2931 6451 2932 6452 2933 6453 2934 6454 8 6455 8 6456 2935 6457 2936 6458 2937 6459 2938 6460 2939 6461 8 6462 2940 6463 8 6464 2941 6465 8 6466 8 6467 8 6468 8 6469 2942 6470 8 6471 2943 6472 2944 6473 8 6474 2945 6475 2946 6476 2947 6477 2948 6478 2949 6479 8 6480 2950 6481 2951 6482 8 6483 2952 6484 8 6485 2953 6486 2954 6487 2955 6488 2956 6489 2957 6490 2958 6491 2959 6492 8 6493 8 6494 2960 6495 8 6496 8 6497 2961 6498 2962 6499 2963 6500 8 6501 8 6502 8 6503 8 6504 2964 6505 8 6506 8 6507 8 6508 8 6509 8 6510 2965 6511 2966 6512 2967 6513 2968 6514 2969 6515 2970 6516 2971 6517 2972 6518 2973 6519 2974 6520 2975 6521 2976 6522 2977 6523 2978 6524 2979 6525 2980 6526 2981 6527 2982 6528 2983 6529 2984 6530 2985 6531 2986 6532 2987 6533 2988 6534 2989 6535 2990 6536 2991 6537 2992 6538 2993 6539 2994 6540 2995 6541 2996 6542 2997 6543 2998 6544 2999 6545 2968 6546 2968 6547 2968 6548 2968 6549 2968 6550 2968 6551 2968 6552 2968 6553 2968 6554 2968 6555 2968 6556 2968 6557 2968 6558 2968 6559 2968 6560 2968 6561 3000 6562 3001 6563 8 6564 8 6565 8 6566 8 6567 8 6568 8 6569 3002 6570 8 6571 8 6572 3003 6573 8 6574 3004 6575 3005 6576 3006 6577 3007 6578 3008 6579 8 6580 8 6581 8 6582 8 6583 8 6584 3009 6585 3010 6586 8 6587 8 6588 3011 6589 8 6590 3012 6591 8 6592 8 6593 3013 6594 3014 6595 8 6596 2968 6597 2968 6598 2968 6599 2968 6600 2968 6601 3015 6602 3016 6603 3017 6604 2968 6605 2968 6606 2968 6607 2968 6608 2968 6609 2968 6610 2968 6611 2968 6612 2968 6613 2968 6614 3018 6615 8 6616 8 6617 8 6618 8 6619 8 6620 8 6621 8 6622 8 6623 8 6624 8 6625 8 6626 8 6627 8 6628 8 6629 8 6630 8 6631 8 6632 8 6633 8 6634 8 6635 8 6636 8 6637 8 6638 8 6639 8 6640 8 6641 8 6642 8 6643 8 6644 8 6645 8 6646 8 6647 8 6648 8 6649 8 6650 8 6651 8 6652 8 6653 8 6654 8 6655 8 6656 8 6657 8 6658 8 6659 8 6660 8 6661 8 6662 8 6663 8 6664 8 6665 8 6666 8 6667 8 6668 8 6669 8 6670 8 6671 8 6672 8 6673 8 6674 8 6675 8 6676 8 6677 8 6678 8 6679 8 6680 8 6681 8 6682 8 6683 8 6684 8 6685 8 6686 8 6687 8 6688 8 6689 8 6690 8 6691 8 6692 8 6693 8 6694 8 6695 3019 6696 8 6697 8 6698 8 6699 3020 6700 3021 6701 8 6702 3022 6703 8 6704 8 6705 3023 6706 3024 6707 3025 6708 8 6709 8 6710 3026 6711 3027 6712 3028 6713 3029 6714 8 6715 3030 6716 8 6717 3031 6718 3032 6719 3033 6720 3034 6721 3035 6722 3036 6723 3037 6724 3038 6725 3039 6726 2968 6727 2968 6728 2968 6729 2968 6730 2968 6731 2968 6732 2968 6733 2968 6734 2968 6735 2968 6736 2968 6737 2968 6738 2968 6739 3040 6740 2968 6741 3041 6742 3042 6743 8 6744 3043 6745 3044 6746 8 6747 3045 6748 8 6749 3046 6750 8 6751 8 6752 8 6753 3047 6754 3048 6755 3049 6756 8 6757 3050 6758 8 6759 8 6760 3051 6761 3052 6762 3053 6763 8 6764 8 6765 8 6766 3054 6767 3055 6768 8 6769 8 6770 8 6771 3056 6772 3057 6773 3058 6774 3059 6775 8 6776 8 6777 3060 6778 8 6779 3061 6780 3062 6781 8 6782 8 6783 3063 6784 8 6785 8 6786 8 6787 3064 6788 3065 6789 8 6790 8 6791 3066 6792 3067 6793 3068 6794 3069 6795 3070 6796 3071 6797 3072 6798 3073 6799 3074 6800 3075 6801 3076 6802 8 6803 3077 6804 3078 6805 3079 6806 3080 6807 3081 6808 3082 6809 3083 6810 3084 6811 3085 6812 3086 6813 8 6814 3087 6815 2780 6816 3088 6817 8 6818 3089 6819 8 6820 3090 6821 3091 6822 8 6823 8 6824 8 6825 3092 6826 8 6827 8 6828 3093 6829 3094 6830 3095 6831 3096 6832 8 6833 8 6834 3097 6835 8 6836 3098 6837 3099 6838 8 6839 8 6840 3100 6841 8 6842 3101 6843 8 6844 3102 6845 3103 6846 3104 6847 2968 6848 2968 6849 3105 6850 3106 6851 2968 6852 2968 6853 2968 6854 2968 6855 8 6856 3107 6857 8 6858 3108 6859 8 6860 8 6861 2968 6862 2968 6863 8 6864 2968 6865 2968 6866 3109 6867 8 6868 8 6869 8 6870 8 6871 8 6872 8 6873 3110 6874 3111 6875 3112 6876 3113 6877 8 6878 3114 6879 8 6880 3115 6881 3116 6882 3117 6883 3118 6884 3119 6885 8 6886 8 6887 8 6888 8 6889 8 6890 3120 6891 8 6892 8 6893 3121 6894 3122 6895 3123 6896 3124 6897 3125 6898 3126 6899 3127 6900 8 6901 3128 6902 8 6903 3129 6904 3130 6905 3131 6906 3132 6907 3133 6908 8 6909 3134 6910 3135 6911 3136 6912 3137 6913 2968 6914 3138 6915 3139 6916 3140 6917 8 6918 2968 6919 2968 6920 2968 6921 8 6922 3141 6923 8 6924 8 6925 8 6926 3142 6927 3143 6928 3144 6929 8 6930 8 6931 3145 6932 3146 6933 8 6934 3147 6935 3148 6936 8 6937 2968 6938 8 6939 3149 6940 8 6941 2968 6942 8 6943 3150 6944 3151 6945 3152 6946 3153 6947 8 6948 8 6949 3154 6950 8 6951 8 6952 2968 6953 2968 6954 3155 6955 3156 6956 8 6957 2968 6958 2968 6959 2968 6960 3157 6961 8 6962 3158 6963 3159 6964 8 6965 3160 6966 8 6967 2968 6968 2780 6969 2780 6970 2780 6971 8 6972 3161 6973 8 6974 3162 6975 8 6976 8 6977 2780 6978 2780 6979 8 6980 2780 6981 2780 6982 3163 6983 8 6984 8 6985 8 6986 8 6987 8 6988 8 6989 3164 6990 3165 6991 3166 6992 3167 6993 8 6994 3168 6995 8 6996 3169 6997 3170 6998 3171 6999 3172 7000 3173 7001 8 7002 8 7003 8 7004 8 7005 8 7006 3174 7007 8 7008 8 7009 3175 7010 3176 7011 3177 7012 3178 7013 3179 7014 3180 7015 3181 7016 8 7017 3182 7018 8 7019 3183 7020 3184 7021 3185 7022 3186 7023 3187 7024 8 7025 3188 7026 3189 7027 3190 7028 3191 7029 2780 7030 3192 7031 3193 7032 3194 7033 8 7034 2780 7035 2780 7036 2780 7037 8 7038 3195 7039 8 7040 8 7041 8 7042 3196 7043 3197 7044 3198 7045 8 7046 8 7047 3199 7048 3200 7049 8 7050 3201 7051 3202 7052 8 7053 2780 7054 8 7055 3203 7056 8 7057 2780 7058 8 7059 3204 7060 3205 7061 3206 7062 3207 7063 8 7064 8 7065 3208 7066 8 7067 8 7068 2780 7069 2780 7070 3209 7071 3210 7072 8 7073 2780 7074 2780 7075 2780 7076 3211 7077 8 7078 3212 7079 3213 7080 8 7081 3214 7082 8 7083 2780 7084 3215 7085 8 7086 8 7087 8 7088 8 7089 8 7090 8 7091 8 7092 8 7093 8 7094 3216 7095 8 7096 8 7097 8 7098 3217 7099 8 7100 8 7101 8 7102 3218 7103 8 7104 8 7105 8 7106 8 7107 8 7108 8 7109 8 7110 8 7111 8 7112 8 7113 8 7114 8 7115 8 7116 8 7117 8 7118 8 7119 8 7120 8 7121 8 7122 8 7123 8 7124 8 7125 8 7126 8 7127 8 7128 8 7129 8 7130 8 7131 8 7132 8 7133 8 7134 8 7135 8 7136 8 7137 8 7138 8 7139 8 7140 8 7141 8 7142 8 7143 8 7144 8 7145 8 7146 8 7147 8 7148 8 7149 8 7150 8 7151 8 7152 8 7153 8 7154 8 7155 8 7156 8 7157 8 7158 8 7159 8 7160 8 7161 8 7162 8 7163 8 7164 8 7165 8 7166 8 7167 8 7168 8 7169 8 7170 8 7171 8 7172 8 7173 8 7174 8 7175 8 7176 8 7177 8 7178 8 7179 8 7180 8 7181 8 7182 3219 7183 8 7184 8 7185 3220 7186 3221 7187 3222 7188 3223 7189 3224 7190 3225 7191 3226 7192 8 7193 3227 7194 3228 7195 3229 7196 8 7197 8 7198 3230 7199 8 7200 8 7201 8 7202 8 7203 8 7204 8 7205 8 7206 3231 7207 8 7208 3232 7209 8 7210 8 7211 3233 7212 3234 7213 3235 7214 3236 7215 8 7216 8 7217 8 7218 8 7219 8 7220 8 7221 8 7222 8 7223 8 7224 8 7225 8 7226 3237 7227 8 7228 3238 7229 8 7230 3239 7231 8 7232 8 7233 8 7234 8 7235 8 7236 8 7237 8 7238 8 7239 8 7240 8 7241 3240 7242 8 7243 8 7244 8 7245 8 7246 8 7247 8 7248 8 7249 8 7250 8 7251 8 7252 8 7253 3241 7254 3242 7255 8 7256 8 7257 8 7258 8 7259 8 7260 8 7261 8 7262 8 7263 8 7264 8 7265 8 7266 8 7267 8 7268 8 7269 8 7270 8 7271 8 7272 8 7273 8 7274 8 7275 8 7276 8 7277 8 7278 8 7279 8 7280 8 7281 8 7282 8 7283 8 7284 8 7285 8 7286 8 7287 8 7288 8 7289 8 7290 8 7291 8 7292 8 7293 8 7294 8 7295 8 7296 8 7297 8 7298 8 7299 8 7300 8 7301 8 7302 8 7303 8 7304 8 7305 8 7306 8 7307 8 7308 8 7309 8 7310 8 7311 8 7312 8 7313 8 7314 8 7315 8 7316 8 7317 8 7318 8 7319 8 7320 8 7321 8 7322 8 7323 8 7324 8 7325 8 7326 3243 7327 8 7328 8 7329 8 7330 8 7331 8 7332 8 7333 8 7334 8 7335 8 7336 8 7337 8 7338 3244 7339 3245 7340 8 7341 8 7342 3246 7343 3247 7344 3248 7345 8 7346 3249 7347 3250 7348 3251 7349 3252 7350 8 7351 8 7352 8 7353 8 7354 8 7355 8 7356 8 7357 8 7358 8 7359 8 7360 8 7361 3253 7362 8 7363 3254 7364 8 7365 8 7366 8 7367 8 7368 3243 7369 3255 7370 3256 7371 8 7372 8 7373 3257 7374 3258 7375 8 7376 3259 7377 8 7378 8 7379 8 7380 8 7381 8 7382 8 7383 8 7384 8 7385 8 7386 8 7387 3260 7388 3261 7389 8 7390 3262 7391 8 7392 8 7393 8 7394 8 7395 8 7396 8 7397 8 7398 8 7399 8 7400 8 7401 8 7402 8 7403 8 7404 8 7405 8 7406 8 7407 8 7408 8 7409 8 7410 8 7411 3263 7412 8 7413 3264 7414 8 7415 8 7416 3265 7417 8 7418 8 7419 8 7420 8 7421 8 7422 8 7423 8 7424 8 7425 8 7426 8 7427 8 7428 8 7429 8 7430 8 7431 8 7432 8 7433 8 7434 8 7435 8 7436 8 7437 8 7438 8 7439 8 7440 8 7441 8 7442 8 7443 8 7444 8 7445 8 7446 8 7447 8 7448 8 7449 8 7450 8 7451 8 7452 8 7453 8 7454 8 7455 8 7456 8 7457 8 7458 8 7459 8 7460 8 7461 8 7462 8 7463 8 7464 8 7465 8 7466 8 7467 8 7468 8 7469 8 7470 8 7471 8 7472 8 7473 8 7474 8 7475 8 7476 8 7477 8 7478 8 7479 8 7480 8 7481 8 7482 8 7483 8 7484 8 7485 8 7486 8 7487 8 7488 8 7489 8 7490 8 7491 8 7492 8 7493 8 7494 8 7495 8 7496 8 7497 8 7498 8 7499 3266 7500 3267 7501 8 7502 8 7503 8 7504 8 7505 8 7506 3268 7507 3269 7508 3270 7509 3271 7510 8 7511 8 7512 3272 7513 8 7514 8 7515 8 7516 8 7517 8 7518 8 7519 8 7520 8 7521 8 7522 8 7523 8 7524 8 7525 8 7526 8 7527 8 7528 8 7529 8 7530 8 7531 8 7532 8 7533 8 7534 8 7535 8 7536 8 7537 8 7538 8 7539 8 7540 8 7541 8 7542 8 7543 8 7544 8 7545 8 7546 8 7547 8 7548 8 7549 8 7550 8 7551 8 7552 8 7553 8 7554 8 7555 8 7556 8 7557 8 7558 8 7559 8 7560 8 7561 8 7562 8 7563 8 7564 8 7565 8 7566 8 7567 8 7568 8 7569 8 7570 8 7571 8 7572 8 7573 8 7574 8 7575 8 7576 8 7577 8 7578 8 7579 8 7580 8 7581 8 7582 8 7583 8 7584 8 7585 8 7586 8 7587 8 7588 8 7589 8 7590 8 7591 8 7592 8 7593 8 7594 8 7595 3273 7596 8 7597 8 7598 8 7599 8 7600 8 7601 8 7602 8 7603 8 7604 8 7605 8 7606 8 7607 8 7608 8 7609 8 7610 8 7611 3274 7612 8 7613 8 7614 8 7615 8 7616 8 7617 8 7618 8 7619 8 7620 8 7621 8 7622 8 7623 8 7624 8 7625 8 7626 8 7627 8 7628 8 7629 8 7630 8 7631 8 7632 8 7633 8 7634 8 7635 8 7636 8 7637 8 7638 8 7639 8 7640 8 7641 8 7642 8 7643 8 7644 8 7645 8 7646 8 7647 8 7648 8 7649 8 7650 8 7651 8 7652 8 7653 8 7654 8 7655 8 7656 8 7657 8 7658 8 7659 8 7660 8 7661 8 7662 8 7663 8 7664 8 7665 8 7666 8 7667 8 7668 8 7669 8 7670 8 7671 8 7672 8 7673 8 7674 8 7675 8 7676 8 7677 8 7678 8 7679 8 7680 8 7681 8 7682 8 7683 8 7684 8 7685 8 7686 8 7687 8 7688 8 7689 8 7690 8 7691 8 7692 8 7693 8 7694 3275 7695 8 7696 8 7697 8 7698 8 7699 8 7700 8 7701 3276 7702 8 7703 3277 7704 8 7705 8 7706 3278 7707 8 7708 8 7709 8 7710 8 7711 8 7712 8 7713 8 7714 8 7715 8 7716 8 7717 8 7718 8 7719 8 7720 8 7721 8 7722 8 7723 8 7724 8 7725 8 7726 8 7727 8 7728 8 7729 8 7730 8 7731 8 7732 8 7733 8 7734 8 7735 8 7736 8 7737 8 7738 8 7739 8 7740 8 7741 8 7742 8 7743 8 7744 8 7745 8 7746 8 7747 8 7748 8 7749 8 7750 8 7751 8 7752 8 7753 8 7754 8 7755 8 7756 8 7757 8 7758 8 7759 8 7760 8 7761 8 7762 8 7763 8 7764 8 7765 8 7766 8 7767 8 7768 8 7769 8 7770 8 7771 8 7772 8 7773 8 7774 8 7775 8 7776 8 7777 8 7778 8 7779 8 7780 8 7781 8 7782 8 7783 8 7784 8 7785 8 7786 3279 7787 8 7788 8 7789 3280 7790 3281 7791 8 7792 85 7793 3282 7794 3283 7795 8 7796 8 7797 8 7798 8 7799 8 7800 8 7801 8 7802 3284 7803 8 7804 3285 7805 8 7806 8 7807 8 7808 8 7809 8 7810 8 7811 8 7812 8 7813 8 7814 3286 7815 3287 7816 3288 7817 8 7818 8 7819 8 7820 3289 7821 3290 7822 8 7823 3291 7824 8 7825 3292 7826 3293 7827 1181 7828 1122 7829 8 7830 1122 7831 8 7832 1122 7833 1122 7834 1122 7835 1122 7836 1122 7837 3294 7838 8 7839 3295 7840 3296 7841 8 7842 3297 7843 8 7844 1122 7845 3298 7846 8 7847 3299 7848 8 7849 3300 7850 3301 7851 8 7852 8 7853 8 7854 3302 7855 8 7856 8 7857 3303 7858 3304 7859 832 7860 3305 7861 3306 7862 3307 7863 3308 7864 8 7865 8 7866 3309 7867 8 7868 3310 7869 3311 7870 8 7871 8 7872 3312 7873 8 7874 8 7875 3313 7876 8 7877 3314 7878 3315 7879 3316 7880 937 7881 937 7882 3317 7883 3318 7884 937 7885 937 7886 937 7887 937 7888 8 7889 3319 7890 8 7891 3320 7892 8 7893 8 7894 937 7895 937 7896 8 7897 937 7898 937 7899 3321 7900 8 7901 8 7902 8 7903 8 7904 8 7905 8 7906 3322 7907 3323 7908 3324 7909 3325 7910 8 7911 3326 7912 8 7913 3327 7914 3328 7915 3329 7916 3330 7917 3331 7918 8 7919 8 7920 8 7921 8 7922 8 7923 3332 7924 8 7925 8 7926 3333 7927 3334 7928 3335 7929 3336 7930 3337 7931 3338 7932 3339 7933 8 7934 3340 7935 8 7936 3341 7937 3342 7938 3343 7939 3344 7940 3345 7941 8 7942 3346 7943 3347 7944 989 7945 3348 7946 3349 7947 3350 7948 8 7949 937 7950 937 7951 937 7952 8 7953 3351 7954 8 7955 8 7956 8 7957 3352 7958 3353 7959 3354 7960 8 7961 8 7962 3355 7963 3356 7964 8 7965 3357 7966 3358 7967 8 7968 1122 7969 937 7970 8 7971 3359 7972 8 7973 937 7974 8 7975 3360 7976 3361 7977 3362 7978 3363 7979 8 7980 8 7981 3364 7982 8 7983 8 7984 937 7985 937 7986 3365 7987 3366 7988 8 7989 937 7990 937 7991 937 7992 3367 7993 8 7994 3368 7995 3369 7996 8 7997 3370 7998 8 7999 937 8000 3371 8001 8 8002 8 8003 8 8004 8 8005 8 8006 8 8007 8 8008 8 8009 8 8010 8 8011 8 8012 8 8013 8 8014 8 8015 8 8016 8 8017 8 8018 8 8019 8 8020 8 8021 8 8022 8 8023 8 8024 8 8025 8 8026 8 8027 8 8028 8 8029 8 8030 8 8031 8 8032 8 8033 8 8034 8 8035 8 8036 8 8037 8 8038 8 8039 8 8040 8 8041 8 8042 8 8043 8 8044 8 8045 8 8046 8 8047 8 8048 8 8049 8 8050 8 8051 8 8052 8 8053 8 8054 8 8055 8 8056 8 8057 8 8058 8 8059 8 8060 8 8061 8 8062 8 8063 8 8064 8 8065 8 8066 8 8067 8 8068 8 8069 8 8070 8 8071 8 8072 8 8073 8 8074 8 8075 8 8076 8 8077 8 8078 8 8079 8 8080 8 8081 3372 8082 8 8083 8 8084 8 8085 3373 8086 3374 8087 8 8088 3375 8089 8 8090 8 8091 3376 8092 3377 8093 3378 8094 8 8095 8 8096 3379 8097 3380 8098 3381 8099 3382 8100 8 8101 3383 8102 8 8103 3384 8104 3385 8105 3386 8106 3387 8107 3388 8108 3389 8109 3390 8110 3391 8111 3392 8112 832 8113 832 8114 832 8115 832 8116 832 8117 832 8118 832 8119 832 8120 832 8121 832 8122 832 8123 832 8124 832 8125 3393 8126 832 8127 3394 8128 3395 8129 8 8130 3396 8131 3397 8132 8 8133 3398 8134 8 8135 3399 8136 8 8137 8 8138 8 8139 3400 8140 3401 8141 3402 8142 8 8143 3403 8144 8 8145 8 8146 3404 8147 3405 8148 3406 8149 8 8150 8 8151 8 8152 3407 8153 3408 8154 8 8155 8 8156 8 8157 3409 8158 3410 8159 3411 8160 3412 8161 8 8162 8 8163 3413 8164 8 8165 3414 8166 3415 8167 8 8168 8 8169 3416 8170 8 8171 8 8172 8 8173 3417 8174 3418 8175 8 8176 3419 8177 8 8178 8 8179 3420 8180 3421 8181 3422 8182 3423 8183 3424 8184 3425 8185 3426 8186 3427 8187 3428 8188 3429 8189 3430 8190 8 8191 3431 8192 3432 8193 3433 8194 3434 8195 3435 8196 3436 8197 3437 8198 3438 8199 3439 8200 3440 8201 3441 8202 3442 8203 3443 8204 3444 8205 3445 8206 3446 8207 3447 8208 3448 8209 3449 8210 8 8211 3450 8212 8 8213 3451 8214 8 8215 8 8216 3452 8217 3453 8218 3454 8219 3455 8220 8 8221 8 8222 3456 8223 3457 8224 3458 8225 3459 8226 3460 8227 8 8228 3461 8229 8 8230 3462 8231 8 8232 8 8233 8 8234 8 8235 3463 8236 8 8237 3464 8238 3465 8239 8 8240 3466 8241 3467 8242 3468 8243 3469 8244 3470 8245 8 8246 3471 8247 3472 8248 8 8249 3473 8250 8 8251 3474 8252 3475 8253 3476 8254 3477 8255 3478 8256 3479 8257 3480 8258 8 8259 8 8260 3481 8261 8 8262 8 8263 3482 8264 3483 8265 3484 8266 8 8267 8 8268 8 8269 8 8270 3485 8271 8 8272 8 8273 8 8274 8 8275 8 8276 3486 8277 3487 8278 3488 8279 3489 8280 3490 8281 3491 8282 3492 8283 3493 8284 3494 8285 3495 8286 3496 8287 3497 8288 3498 8289 3499 8290 3500 8291 3501 8292 3502 8293 3503 8294 3504 8295 3505 8296 3506 8297 3507 8298 3508 8299 3509 8300 3510 8301 3511 8302 3512 8303 3513 8304 3514 8305 3515 8306 3516 8307 3517 8308 3518 8309 3519 8310 3520 8311 3489 8312 3489 8313 3489 8314 3489 8315 3489 8316 3489 8317 3489 8318 3489 8319 3489 8320 3489 8321 3489 8322 3489 8323 3489 8324 3489 8325 3489 8326 3489 8327 3521 8328 3522 8329 8 8330 8 8331 8 8332 8 8333 8 8334 8 8335 3523 8336 8 8337 8 8338 3524 8339 8 8340 3525 8341 3526 8342 3527 8343 3528 8344 3529 8345 8 8346 8 8347 8 8348 8 8349 8 8350 3530 8351 3531 8352 8 8353 8 8354 3532 8355 8 8356 3533 8357 8 8358 8 8359 3534 8360 3535 8361 8 8362 3489 8363 3489 8364 3489 8365 3489 8366 3489 8367 3536 8368 3537 8369 3538 8370 3489 8371 3489 8372 3489 8373 3489 8374 3489 8375 3489 8376 3489 8377 3489 8378 3489 8379 3489 8380 3539 8381 8 8382 8 8383 8 8384 8 8385 8 8386 8 8387 8 8388 8 8389 8 8390 8 8391 8 8392 8 8393 8 8394 8 8395 8 8396 8 8397 8 8398 8 8399 8 8400 8 8401 8 8402 8 8403 8 8404 8 8405 8 8406 8 8407 8 8408 8 8409 8 8410 8 8411 8 8412 8 8413 8 8414 8 8415 8 8416 8 8417 8 8418 8 8419 8 8420 8 8421 8 8422 8 8423 8 8424 8 8425 8 8426 8 8427 8 8428 8 8429 8 8430 8 8431 8 8432 8 8433 8 8434 8 8435 8 8436 8 8437 8 8438 8 8439 8 8440 8 8441 8 8442 8 8443 8 8444 8 8445 8 8446 8 8447 8 8448 8 8449 8 8450 8 8451 8 8452 8 8453 8 8454 8 8455 8 8456 8 8457 8 8458 8 8459 8 8460 8 8461 3540 8462 8 8463 8 8464 8 8465 3541 8466 3542 8467 8 8468 3543 8469 8 8470 8 8471 3544 8472 3545 8473 3546 8474 8 8475 8 8476 3547 8477 3548 8478 3549 8479 3550 8480 8 8481 3551 8482 8 8483 3552 8484 3553 8485 3554 8486 3555 8487 3556 8488 3557 8489 3558 8490 3559 8491 3560 8492 3489 8493 3489 8494 3489 8495 3489 8496 3489 8497 3489 8498 3489 8499 3489 8500 3489 8501 3489 8502 3489 8503 3489 8504 3489 8505 3561 8506 3489 8507 3562 8508 3563 8509 8 8510 3564 8511 3565 8512 8 8513 3566 8514 8 8515 3567 8516 8 8517 8 8518 8 8519 3568 8520 3569 8521 3570 8522 8 8523 3571 8524 8 8525 8 8526 3572 8527 3573 8528 3574 8529 8 8530 8 8531 8 8532 3575 8533 3576 8534 8 8535 8 8536 8 8537 3577 8538 3578 8539 3579 8540 3580 8541 8 8542 8 8543 3581 8544 8 8545 3582 8546 3583 8547 8 8548 8 8549 3584 8550 8 8551 8 8552 8 8553 3585 8554 3586 8555 8 8556 8 8557 3587 8558 3588 8559 3589 8560 3590 8561 3591 8562 3592 8563 3593 8564 3594 8565 3595 8566 3596 8567 3597 8568 8 8569 3598 8570 3599 8571 3600 8572 3601 8573 3602 8574 3603 8575 3604 8576 3605 8577 3606 8578 3607 8579 8 8580 8 8581 3608 8582 8 8583 3609 8584 8 8585 3610 8586 3611 8587 8 8588 8 8589 8 8590 3612 8591 8 8592 8 8593 3613 8594 3614 8595 3615 8596 3616 8597 8 8598 8 8599 3617 8600 8 8601 3618 8602 3619 8603 8 8604 8 8605 3620 8606 8 8607 3621 8608 8 8609 3622 8610 3623 8611 3624 8612 3489 8613 3489 8614 3625 8615 3626 8616 3489 8617 3489 8618 3489 8619 3489 8620 8 8621 3627 8622 8 8623 3628 8624 8 8625 8 8626 3489 8627 3489 8628 8 8629 3489 8630 3489 8631 3629 8632 8 8633 8 8634 8 8635 8 8636 8 8637 8 8638 3630 8639 3631 8640 3632 8641 3633 8642 8 8643 3634 8644 8 8645 3635 8646 3636 8647 3637 8648 3638 8649 3639 8650 8 8651 8 8652 8 8653 8 8654 8 8655 3640 8656 8 8657 8 8658 3641 8659 3642 8660 3643 8661 3644 8662 3645 8663 3646 8664 3647 8665 8 8666 3648 8667 8 8668 3649 8669 3650 8670 3651 8671 3652 8672 3653 8673 8 8674 3654 8675 3655 8676 3656 8677 3657 8678 3489 8679 3658 8680 3659 8681 3660 8682 8 8683 3489 8684 3489 8685 3489 8686 8 8687 3661 8688 8 8689 8 8690 8 8691 3662 8692 3663 8693 3664 8694 8 8695 8 8696 3665 8697 3666 8698 8 8699 3667 8700 3668 8701 8 8702 3489 8703 8 8704 3669 8705 8 8706 3489 8707 8 8708 3670 8709 3671 8710 3672 8711 3673 8712 8 8713 8 8714 3674 8715 8 8716 8 8717 3489 8718 3489 8719 3675 8720 3676 8721 8 8722 3489 8723 3489 8724 3489 8725 3677 8726 8 8727 3678 8728 3679 8729 8 8730 3680 8731 8 8732 3489 8733 3681 8734 3682 8735 3683 8736 3684 8737 3685 8738 3686 8739 3687 8740 3688 8741 3689 8742 3690 8743 3691 8744 8 8745 3692 8746 8 8747 3693 8748 8 8749 8 8750 3694 8751 3695 8752 3696 8753 3697 8754 8 8755 8 8756 3698 8757 3699 8758 3700 8759 3701 8760 3702 8761 8 8762 3703 8763 8 8764 3704 8765 8 8766 8 8767 8 8768 8 8769 3705 8770 8 8771 3706 8772 3707 8773 8 8774 3708 8775 3709 8776 3710 8777 3711 8778 3712 8779 8 8780 3713 8781 3714 8782 8 8783 3715 8784 8 8785 3716 8786 3717 8787 3718 8788 3719 8789 3720 8790 3721 8791 3722 8792 3723 8793 8 8794 8 8795 3724 8796 8 8797 8 8798 3725 8799 3726 8800 3727 8801 8 8802 8 8803 8 8804 8 8805 3728 8806 8 8807 8 8808 8 8809 8 8810 8 8811 3729 8812 3730 8813 3731 8814 3732 8815 3725 8816 3733 8817 3734 8818 3735 8819 3736 8820 3737 8821 3738 8822 3739 8823 3740 8824 3741 8825 3742 8826 3743 8827 3744 8828 3745 8829 3746 8830 3747 8831 3748 8832 3749 8833 3750 8834 3751 8835 3752 8836 3753 8837 3754 8838 3755 8839 3756 8840 3757 8841 3758 8842 3759 8843 3760 8844 3761 8845 3762 8846 3763 8847 3725 8848 3725 8849 3725 8850 3725 8851 3725 8852 3725 8853 3725 8854 3725 8855 3725 8856 3725 8857 3725 8858 3725 8859 3725 8860 3725 8861 3725 8862 3725 8863 3764 8864 3765 8865 8 8866 8 8867 8 8868 8 8869 8 8870 8 8871 3766 8872 8 8873 8 8874 3767 8875 8 8876 3768 8877 3769 8878 3770 8879 3771 8880 3772 8881 8 8882 8 8883 8 8884 8 8885 8 8886 3773 8887 3774 8888 8 8889 8 8890 3775 8891 8 8892 3776 8893 8 8894 8 8895 3777 8896 3778 8897 8 8898 3725 8899 3725 8900 3725 8901 3725 8902 3725 8903 3779 8904 3780 8905 3781 8906 3725 8907 3725 8908 3725 8909 3725 8910 3725 8911 3725 8912 3725 8913 3725 8914 3725 8915 3725 8916 3782 8917 3783 8918 3784 8919 3785 8920 3725 8921 3786 8922 8 8923 8 8924 8 8925 8 8926 8 8927 8 8928 8 8929 8 8930 8 8931 8 8932 8 8933 8 8934 8 8935 8 8936 8 8937 8 8938 8 8939 8 8940 8 8941 8 8942 8 8943 8 8944 8 8945 8 8946 8 8947 8 8948 8 8949 8 8950 8 8951 8 8952 8 8953 8 8954 8 8955 8 8956 8 8957 8 8958 8 8959 8 8960 8 8961 8 8962 8 8963 8 8964 8 8965 8 8966 8 8967 8 8968 8 8969 8 8970 8 8971 8 8972 8 8973 8 8974 8 8975 8 8976 8 8977 8 8978 8 8979 8 8980 8 8981 8 8982 8 8983 8 8984 8 8985 8 8986 8 8987 8 8988 8 8989 8 8990 8 8991 8 8992 8 8993 8 8994 8 8995 8 8996 8 8997 8 8998 8 8999 8 9000 8 9001 8 9002 3787 9003 8 9004 8 9005 8 9006 3788 9007 3789 9008 8 9009 3790 9010 8 9011 8 9012 3791 9013 3792 9014 3793 9015 8 9016 8 9017 3794 9018 3795 9019 3796 9020 3797 9021 8 9022 3798 9023 8 9024 3799 9025 3800 9026 3801 9027 3802 9028 3803 9029 3804 9030 3805 9031 3806 9032 3807 9033 3725 9034 3725 9035 3725 9036 3725 9037 3725 9038 3725 9039 3725 9040 3725 9041 3725 9042 3725 9043 3725 9044 3725 9045 3725 9046 3808 9047 3725 9048 3809 9049 3810 9050 8 9051 3811 9052 3812 9053 8 9054 3813 9055 8 9056 3814 9057 8 9058 8 9059 8 9060 3815 9061 3816 9062 3817 9063 8 9064 3818 9065 8 9066 8 9067 3819 9068 3820 9069 3821 9070 8 9071 8 9072 8 9073 3822 9074 3823 9075 8 9076 8 9077 8 9078 3824 9079 3825 9080 3826 9081 3827 9082 8 9083 8 9084 3828 9085 8 9086 3829 9087 3830 9088 8 9089 8 9090 3831 9091 8 9092 8 9093 8 9094 3832 9095 3833 9096 8 9097 8 9098 3834 9099 3835 9100 3836 9101 3837 9102 3838 9103 3839 9104 3840 9105 3841 9106 3842 9107 3843 9108 3844 9109 8 9110 3845 9111 3846 9112 3847 9113 3848 9114 3849 9115 3850 9116 3851 9117 3852 9118 3853 9119 3854 9120 3855 9121 3856 9122 3857 9123 3858 9124 3859 9125 3860 9126 3861 9127 3862 9128 3863 9129 8 9130 3864 9131 8 9132 3865 9133 8 9134 8 9135 3866 9136 3867 9137 3868 9138 3869 9139 8 9140 8 9141 3870 9142 3871 9143 3872 9144 3873 9145 3874 9146 8 9147 3875 9148 8 9149 3876 9150 8 9151 8 9152 8 9153 8 9154 3877 9155 8 9156 3878 9157 3879 9158 8 9159 3880 9160 3881 9161 3882 9162 3883 9163 3884 9164 8 9165 3885 9166 3886 9167 8 9168 3887 9169 8 9170 3888 9171 3889 9172 3890 9173 3891 9174 3892 9175 3893 9176 3894 9177 8 9178 8 9179 3895 9180 8 9181 8 9182 3896 9183 3897 9184 3898 9185 8 9186 8 9187 8 9188 8 9189 3899 9190 8 9191 8 9192 8 9193 8 9194 8 9195 3900 9196 3901 9197 3902 9198 3903 9199 3904 9200 3905 9201 3906 9202 3907 9203 3908 9204 3909 9205 3910 9206 3911 9207 3912 9208 3913 9209 3914 9210 3915 9211 3916 9212 3917 9213 3918 9214 3919 9215 3920 9216 3921 9217 3922 9218 3923 9219 3924 9220 3925 9221 3926 9222 3927 9223 3928 9224 3929 9225 3930 9226 3931 9227 3932 9228 3933 9229 3934 9230 3903 9231 3903 9232 3903 9233 3903 9234 3903 9235 3903 9236 3903 9237 3903 9238 3903 9239 3903 9240 3903 9241 3903 9242 3903 9243 3903 9244 3903 9245 3903 9246 3935 9247 3936 9248 8 9249 8 9250 8 9251 8 9252 8 9253 8 9254 3937 9255 8 9256 8 9257 3938 9258 8 9259 3939 9260 3940 9261 3941 9262 3942 9263 3943 9264 8 9265 8 9266 8 9267 8 9268 8 9269 3944 9270 3945 9271 8 9272 8 9273 3946 9274 8 9275 3947 9276 8 9277 8 9278 3948 9279 3949 9280 8 9281 3903 9282 3903 9283 3903 9284 3903 9285 3903 9286 3950 9287 3951 9288 3952 9289 3903 9290 3903 9291 3903 9292 3903 9293 3903 9294 3903 9295 3903 9296 3903 9297 3903 9298 3903 9299 3953 9300 8 9301 8 9302 8 9303 8 9304 8 9305 8 9306 8 9307 8 9308 8 9309 8 9310 8 9311 8 9312 8 9313 8 9314 8 9315 8 9316 8 9317 8 9318 8 9319 8 9320 8 9321 8 9322 8 9323 8 9324 8 9325 8 9326 8 9327 8 9328 8 9329 8 9330 8 9331 8 9332 8 9333 8 9334 8 9335 8 9336 8 9337 8 9338 8 9339 8 9340 8 9341 8 9342 8 9343 8 9344 8 9345 8 9346 8 9347 8 9348 8 9349 8 9350 8 9351 8 9352 8 9353 8 9354 8 9355 8 9356 8 9357 8 9358 8 9359 8 9360 8 9361 8 9362 8 9363 8 9364 8 9365 8 9366 8 9367 8 9368 8 9369 8 9370 8 9371 8 9372 8 9373 8 9374 8 9375 8 9376 8 9377 8 9378 8 9379 8 9380 3954 9381 8 9382 8 9383 8 9384 3955 9385 3956 9386 8 9387 3957 9388 8 9389 8 9390 3958 9391 3959 9392 3960 9393 8 9394 8 9395 3961 9396 3962 9397 3963 9398 3964 9399 8 9400 3965 9401 8 9402 3966 9403 3967 9404 3968 9405 3969 9406 3970 9407 3971 9408 3972 9409 3973 9410 3974 9411 3903 9412 3903 9413 3903 9414 3903 9415 3903 9416 3903 9417 3903 9418 3903 9419 3903 9420 3903 9421 3903 9422 3903 9423 3903 9424 3975 9425 3903 9426 3976 9427 3977 9428 8 9429 3978 9430 3979 9431 8 9432 3980 9433 8 9434 3981 9435 8 9436 8 9437 8 9438 3982 9439 3983 9440 3984 9441 8 9442 3985 9443 8 9444 8 9445 3986 9446 3987 9447 3988 9448 8 9449 8 9450 8 9451 3989 9452 3990 9453 8 9454 8 9455 8 9456 3991 9457 3992 9458 3993 9459 3994 9460 8 9461 8 9462 3995 9463 8 9464 3996 9465 3997 9466 8 9467 8 9468 3998 9469 8 9470 8 9471 8 9472 3999 9473 4000 9474 8 9475 8 9476 4001 9477 4002 9478 4003 9479 4004 9480 4005 9481 4006 9482 4007 9483 4008 9484 4009 9485 4010 9486 4011 9487 8 9488 4012 9489 4013 9490 4014 9491 4015 9492 4016 9493 4017 9494 4018 9495 4019 9496 4020 9497 4021 9498 8 9499 8 9500 4022 9501 8 9502 4023 9503 8 9504 4024 9505 4025 9506 8 9507 8 9508 8 9509 4026 9510 8 9511 8 9512 4027 9513 4028 9514 4029 9515 4030 9516 8 9517 8 9518 4031 9519 8 9520 4032 9521 4033 9522 8 9523 8 9524 4034 9525 8 9526 4035 9527 8 9528 4036 9529 4037 9530 4038 9531 3903 9532 3903 9533 4039 9534 4040 9535 3903 9536 3903 9537 3903 9538 3903 9539 8 9540 4041 9541 8 9542 4042 9543 8 9544 8 9545 3903 9546 3903 9547 8 9548 3903 9549 3903 9550 4043 9551 8 9552 8 9553 8 9554 8 9555 8 9556 8 9557 4044 9558 4045 9559 4046 9560 4047 9561 8 9562 4048 9563 8 9564 4049 9565 4050 9566 4051 9567 4052 9568 4053 9569 8 9570 8 9571 8 9572 8 9573 8 9574 4054 9575 8 9576 8 9577 4055 9578 4056 9579 4057 9580 4058 9581 4059 9582 4060 9583 4061 9584 8 9585 4062 9586 8 9587 4063 9588 4064 9589 4065 9590 4066 9591 4067 9592 8 9593 4068 9594 4069 9595 4070 9596 4071 9597 3903 9598 4072 9599 4073 9600 4074 9601 8 9602 3903 9603 3903 9604 3903 9605 8 9606 4075 9607 8 9608 8 9609 8 9610 4076 9611 4077 9612 4078 9613 8 9614 8 9615 4079 9616 4080 9617 8 9618 4081 9619 4082 9620 8 9621 3903 9622 8 9623 4083 9624 8 9625 3903 9626 8 9627 4084 9628 4085 9629 4086 9630 4087 9631 8 9632 8 9633 4088 9634 8 9635 8 9636 3903 9637 3903 9638 4089 9639 4090 9640 8 9641 3903 9642 3903 9643 3903 9644 4091 9645 8 9646 4092 9647 4093 9648 8 9649 4094 9650 8 9651 3903 9652 4095 9653 8 9654 4096 9655 8 9656 4097 9657 4098 9658 8 9659 8 9660 8 9661 4099 9662 8 9663 8 9664 4100 9665 4101 9666 4102 9667 4103 9668 8 9669 8 9670 4104 9671 8 9672 4105 9673 4106 9674 8 9675 8 9676 4107 9677 8 9678 4108 9679 4109 9680 8 9681 4110 9682 4111 9683 8 9684 4112 9685 8 9686 4113 9687 8 9688 8 9689 8 9690 4114 9691 4115 9692 4116 9693 8 9694 4117 9695 8 9696 8 9697 4118 9698 4119 9699 4120 9700 8 9701 8 9702 8 9703 4121 9704 4122 9705 8 9706 8 9707 8 9708 4123 9709 4124 9710 4125 9711 4126 9712 8 9713 8 9714 4127 9715 8 9716 4128 9717 4129 9718 8 9719 8 9720 4130 9721 8 9722 8 9723 8 9724 4131 9725 4132 9726 8 9727 4133 9728 8 9729 4134 9730 4135 9731 4136 9732 3725 9733 3725 9734 4137 9735 4138 9736 3725 9737 3725 9738 3725 9739 3725 9740 8 9741 4139 9742 8 9743 4140 9744 8 9745 8 9746 3725 9747 3725 9748 8 9749 3725 9750 3725 9751 4141 9752 8 9753 8 9754 8 9755 8 9756 8 9757 8 9758 4142 9759 4143 9760 4144 9761 4145 9762 8 9763 4146 9764 8 9765 4147 9766 4148 9767 4149 9768 4150 9769 4151 9770 8 9771 8 9772 8 9773 8 9774 8 9775 4152 9776 8 9777 8 9778 4153 9779 4154 9780 4155 9781 4156 9782 4157 9783 4158 9784 4159 9785 8 9786 4160 9787 8 9788 4161 9789 4162 9790 4163 9791 4164 9792 4165 9793 8 9794 4166 9795 4167 9796 3784 9797 4168 9798 4169 9799 4170 9800 8 9801 3725 9802 3725 9803 3725 9804 8 9805 4171 9806 8 9807 8 9808 8 9809 4172 9810 4173 9811 4174 9812 8 9813 8 9814 4175 9815 4176 9816 8 9817 4177 9818 4178 9819 8 9820 3725 9821 8 9822 4179 9823 8 9824 3725 9825 8 9826 4180 9827 4181 9828 4182 9829 4183 9830 8 9831 8 9832 4184 9833 8 9834 8 9835 3725 9836 3725 9837 4185 9838 4186 9839 8 9840 3725 9841 3725 9842 3725 9843 4187 9844 8 9845 4188 9846 4189 9847 8 9848 4190 9849 8 9850 3725 9851 4191 9852 8 9853 4192 9854 8 9855 4193 9856 4194 9857 8 9858 8 9859 8 9860 4195 9861 8 9862 8 9863 8 9864 4196 9865 4197 9866 4198 9867 4199 9868 8 9869 8 9870 4200 9871 8 9872 4201 9873 4202 9874 8 9875 8 9876 4203 9877 8 9878 8 9879 4204 9880 8 9881 4205 9882 4206 9883 4207 9884 832 9885 832 9886 4208 9887 4209 9888 832 9889 832 9890 832 9891 832 9892 8 9893 4210 9894 8 9895 4211 9896 8 9897 8 9898 832 9899 832 9900 8 9901 832 9902 832 9903 4212 9904 8 9905 8 9906 8 9907 8 9908 8 9909 8 9910 4213 9911 4214 9912 4215 9913 4216 9914 8 9915 4217 9916 8 9917 4218 9918 4219 9919 4220 9920 4221 9921 4222 9922 8 9923 8 9924 8 9925 8 9926 8 9927 4223 9928 8 9929 8 9930 4224 9931 4225 9932 4226 9933 4227 9934 4228 9935 4229 9936 4230 9937 8 9938 4231 9939 8 9940 4232 9941 4233 9942 4234 9943 4235 9944 4236 9945 8 9946 4237 9947 4238 9948 3303 9949 4239 9950 4240 9951 4241 9952 8 9953 832 9954 832 9955 832 9956 8 9957 4242 9958 8 9959 8 9960 8 9961 4243 9962 4244 9963 4245 9964 8 9965 8 9966 4246 9967 4247 9968 8 9969 4248 9970 4249 9971 8 9972 3725 9973 832 9974 8 9975 4250 9976 8 9977 832 9978 8 9979 4251 9980 4252 9981 4253 9982 4254 9983 8 9984 8 9985 4255 9986 8 9987 8 9988 832 9989 832 9990 4256 9991 4257 9992 8 9993 832 9994 832 9995 832 9996 4258 9997 8 9998 4259 9999 4260 10000 8 10001 4261 10002 8 10003 832 10004 8 10005 4262 10006 4263 10007 8 10008 8 10009 8 10010 8 10011 8 10012 8 10013 8 10014 4264 10015 4265 10016 4266 10017 4267 10018 8 10019 4268 10020 8 10021 4269 10022 4270 10023 4271 10024 4272 10025 4273 10026 8 10027 8 10028 8 10029 8 10030 8 10031 4274 10032 4275 10033 8 10034 8 10035 4276 10036 8 10037 8 10038 4277 10039 4278 10040 4279 10041 8 10042 8 10043 4280 10044 8 10045 8 10046 4281 10047 4282 10048 4283 10049 4284 10050 4285 10051 4286 10052 4287 10053 8 10054 4288 10055 8 10056 8 10057 4289 10058 4290 10059 8 10060 4291 10061 4292 10062 8 10063 4293 10064 8 10065 8 10066 4294 10067 4295 10068 8 10069 4296 10070 8 10071 8 10072 4297 10073 4298 10074 4299 10075 4300 10076 4301 10077 4302 10078 4303 10079 4304 10080 4305 10081 4306 10082 4307 10083 8 10084 4308 10085 4309 10086 4310 10087 4311 10088 4312 10089 4313 10090 4314 10091 4315 10092 4316 10093 4317 10094 8 10095 4318 10096 564 10097 4319 10098 8 10099 4320 10100 8 10101 4321 10102 4322 10103 8 10104 8 10105 8 10106 4323 10107 8 10108 8 10109 4324 10110 4325 10111 4326 10112 4327 10113 8 10114 8 10115 4328 10116 8 10117 4329 10118 4330 10119 8 10120 8 10121 4331 10122 8 10123 4332 10124 8 10125 4333 10126 4334 10127 4335 10128 671 10129 671 10130 4336 10131 4337 10132 671 10133 671 10134 671 10135 671 10136 8 10137 4338 10138 8 10139 4339 10140 8 10141 8 10142 671 10143 671 10144 8 10145 671 10146 671 10147 4340 10148 8 10149 8 10150 8 10151 8 10152 8 10153 8 10154 4341 10155 4342 10156 4343 10157 4344 10158 8 10159 4345 10160 8 10161 4346 10162 4347 10163 4348 10164 4349 10165 4350 10166 8 10167 8 10168 8 10169 8 10170 8 10171 4351 10172 8 10173 8 10174 4352 10175 4353 10176 4354 10177 4355 10178 4356 10179 4357 10180 4358 10181 8 10182 4359 10183 8 10184 4360 10185 4361 10186 4362 10187 4363 10188 4364 10189 8 10190 4365 10191 4366 10192 4367 10193 4368 10194 671 10195 4369 10196 4370 10197 4371 10198 8 10199 671 10200 671 10201 671 10202 8 10203 4372 10204 8 10205 8 10206 8 10207 4373 10208 4374 10209 4375 10210 8 10211 8 10212 4376 10213 4377 10214 8 10215 4378 10216 4379 10217 8 10218 671 10219 8 10220 4380 10221 8 10222 671 10223 8 10224 4381 10225 4382 10226 4383 10227 4384 10228 8 10229 8 10230 4385 10231 8 10232 8 10233 671 10234 671 10235 4386 10236 4387 10237 8 10238 671 10239 671 10240 671 10241 4388 10242 8 10243 4389 10244 4390 10245 8 10246 4391 10247 8 10248 671 10249 564 10250 8 10251 564 10252 564 10253 4392 10254 8 10255 8 10256 4393 10257 4394 10258 4395 10259 8 10260 8 10261 564 10262 4396 10263 4397 10264 8 10265 8 10266 8 10267 8 10268 8 10269 8 10270 8 10271 4398 10272 4399 10273 4400 10274 4401 10275 8 10276 4402 10277 8 10278 4403 10279 4404 10280 4405 10281 4406 10282 4407 10283 8 10284 8 10285 8 10286 8 10287 8 10288 4408 10289 4409 10290 8 10291 8 10292 4410 10293 8 10294 8 10295 4411 10296 4412 10297 4413 10298 8 10299 8 10300 4414 10301 8 10302 8 10303 4415 10304 4416 10305 4417 10306 4418 10307 4419 10308 4420 10309 4421 10310 8 10311 4422 10312 8 10313 8 10314 4423 10315 4424 10316 8 10317 4425 10318 4426 10319 4427 10320 4428 10321 4429 10322 8 10323 4430 10324 4431 10325 4432 10326 4433 10327 564 10328 4434 10329 4435 10330 4436 10331 8 10332 564 10333 564 10334 4437 10335 8 10336 8 10337 8 10338 4438 10339 4439 10340 4440 10341 8 10342 8 10343 4441 10344 4442 10345 8 10346 4443 10347 4444 10348 8 10349 4445 10350 8 10351 564 10352 8 10353 4446 10354 4447 10355 4448 10356 4449 10357 8 10358 8 10359 4450 10360 8 10361 8 10362 4451 10363 4452 10364 8 10365 4453 10366 8 10367 8 10368 8 10369 8 10370 8 10371 8 10372 8 10373 8 10374 8 10375 8 10376 8 10377 8 10378 8 10379 8 10380 8 10381 8 10382 8 10383 8 10384 8 10385 8 10386 8 10387 8 10388 8 10389 8 10390 8 10391 8 10392 8 10393 8 10394 8 10395 8 10396 8 10397 8 10398 8 10399 8 10400 8 10401 8 10402 8 10403 8 10404 8 10405 8 10406 8 10407 8 10408 8 10409 8 10410 8 10411 8 10412 8 10413 8 10414 8 10415 8 10416 8 10417 8 10418 8 10419 8 10420 8 10421 8 10422 8 10423 8 10424 8 10425 8 10426 8 10427 8 10428 8 10429 8 10430 8 10431 8 10432 8 10433 8 10434 8 10435 8 10436 8 10437 8 10438 8 10439 8 10440 8 10441 8 10442 8 10443 8 10444 8 10445 8 10446 4454 10447 8 10448 8 10449 8 10450 4455 10451 4456 10452 8 10453 8 10454 4457 10455 4458 10456 4459 10457 4460 10458 8 10459 4461 10460 8 10461 4462 10462 4463 10463 4464 10464 4465 10465 4466 10466 4467 10467 4468 10468 4469 10469 4470 10470 564 10471 564 10472 564 10473 564 10474 564 10475 564 10476 564 10477 564 10478 564 10479 564 10480 564 10481 564 10482 564 10483 4471 10484 564 10485 4472 10486 4473 10487 4474 10488 8 10489 4475 10490 8 10491 4476 10492 8 10493 8 10494 8 10495 4477 10496 4478 10497 4479 10498 8 10499 8 10500 8 10501 4480 10502 4481 10503 8 10504 8 10505 8 10506 4482 10507 4483 10508 4484 10509 4485 10510 8 10511 4486 10512 8 10513 4487 10514 8 10515 8 10516 4488 10517 8 10518 8 10519 4489 10520 564 10521 4490 10522 4491 10523 4492 10524 4493 10525 4494 10526 4495 10527 4496 10528 4497 10529 4498 10530 4499 10531 4500 10532 8 10533 4501 10534 4502 10535 4503 10536 4504 10537 4505 10538 4506 10539 4507 10540 4508 10541 4509 10542 4510 10543 8 10544 4511 10545 8 10546 4512 10547 8 10548 4513 10549 4514 10550 8 10551 8 10552 8 10553 4515 10554 8 10555 8 10556 4516 10557 4517 10558 4518 10559 4519 10560 8 10561 8 10562 4520 10563 8 10564 4521 10565 4522 10566 8 10567 8 10568 4523 10569 8 10570 4524 10571 8 10572 4525 10573 4526 10574 4527 10575 564 10576 564 10577 564 10578 564 10579 4528 10580 8 10581 4529 10582 8 10583 8 10584 564 10585 564 10586 8 10587 8 10588 564 10589 8 10590 564 10591 564 10592 564 10593 564 10594 564 10595 4530 10596 8 10597 4531 10598 4532 10599 8 10600 4533 10601 8 10602 564 10603 4534 10604 4535 10605 4536 10606 4537 10607 8 10608 4538 10609 8 10610 4539 10611 4540 10612 4541 10613 4542 10614 4543 10615 4544 10616 4545 10617 4546 10618 4547 10619 466 10620 466 10621 466 10622 466 10623 466 10624 466 10625 466 10626 466 10627 466 10628 466 10629 466 10630 466 10631 466 10632 4548 10633 466 10634 4549 10635 4550 10636 8 10637 4551 10638 4552 10639 8 10640 4553 10641 8 10642 4554 10643 8 10644 8 10645 8 10646 4555 10647 4556 10648 4557 10649 8 10650 4558 10651 8 10652 8 10653 4559 10654 4560 10655 4561 10656 8 10657 8 10658 8 10659 4562 10660 4563 10661 8 10662 8 10663 8 10664 4564 10665 4565 10666 4566 10667 4567 10668 8 10669 8 10670 4568 10671 8 10672 4569 10673 4570 10674 8 10675 8 10676 4571 10677 8 10678 8 10679 8 10680 4572 10681 4573 10682 8 10683 8 10684 4574 10685 4575 10686 4576 10687 4577 10688 4578 10689 4579 10690 4580 10691 4581 10692 4582 10693 4583 10694 4584 10695 8 10696 4585 10697 4586 10698 4587 10699 4588 10700 4589 10701 4590 10702 4591 10703 4592 10704 4593 10705 4594 10706 4595 10707 4596 10708 4597 10709 4598 10710 4599 10711 4600 10712 4601 10713 4602 10714 4603 10715 8 10716 4604 10717 8 10718 4605 10719 8 10720 8 10721 4606 10722 4607 10723 4608 10724 4609 10725 8 10726 8 10727 4610 10728 4611 10729 4612 10730 4613 10731 4614 10732 8 10733 4615 10734 8 10735 4616 10736 8 10737 8 10738 8 10739 8 10740 4617 10741 8 10742 4618 10743 4619 10744 8 10745 4620 10746 4621 10747 4622 10748 4623 10749 4624 10750 8 10751 4625 10752 4626 10753 8 10754 4627 10755 8 10756 4628 10757 4629 10758 4630 10759 4631 10760 4632 10761 4633 10762 4634 10763 8 10764 8 10765 4635 10766 8 10767 8 10768 4636 10769 4637 10770 4638 10771 8 10772 8 10773 8 10774 8 10775 4639 10776 8 10777 8 10778 8 10779 8 10780 8 10781 4640 10782 4641 10783 4642 10784 4643 10785 4644 10786 4645 10787 4646 10788 4647 10789 4648 10790 4649 10791 4650 10792 4651 10793 4652 10794 4653 10795 4654 10796 4655 10797 4656 10798 4657 10799 4658 10800 4659 10801 4660 10802 4661 10803 4662 10804 4663 10805 4664 10806 4665 10807 4666 10808 4667 10809 4668 10810 4669 10811 4670 10812 4671 10813 4672 10814 4673 10815 4674 10816 4643 10817 4643 10818 4643 10819 4643 10820 4643 10821 4643 10822 4643 10823 4643 10824 4643 10825 4643 10826 4643 10827 4643 10828 4643 10829 4643 10830 4643 10831 4643 10832 4675 10833 4676 10834 8 10835 8 10836 8 10837 8 10838 8 10839 8 10840 4677 10841 8 10842 8 10843 4678 10844 8 10845 4679 10846 4680 10847 4681 10848 4682 10849 4683 10850 8 10851 8 10852 8 10853 8 10854 8 10855 4684 10856 4685 10857 8 10858 8 10859 4686 10860 8 10861 4687 10862 8 10863 8 10864 4688 10865 4689 10866 8 10867 4643 10868 4643 10869 4643 10870 4643 10871 4643 10872 4690 10873 4691 10874 4692 10875 4643 10876 4643 10877 4643 10878 4643 10879 4643 10880 4643 10881 4643 10882 4643 10883 4643 10884 4643 10885 4693 10886 8 10887 8 10888 8 10889 8 10890 8 10891 8 10892 8 10893 8 10894 8 10895 8 10896 8 10897 8 10898 8 10899 8 10900 8 10901 8 10902 8 10903 8 10904 8 10905 8 10906 8 10907 8 10908 8 10909 8 10910 8 10911 8 10912 8 10913 8 10914 8 10915 8 10916 8 10917 8 10918 8 10919 8 10920 8 10921 8 10922 8 10923 8 10924 8 10925 8 10926 8 10927 8 10928 8 10929 8 10930 8 10931 8 10932 8 10933 8 10934 8 10935 8 10936 8 10937 8 10938 8 10939 8 10940 8 10941 8 10942 8 10943 8 10944 8 10945 8 10946 8 10947 8 10948 8 10949 8 10950 8 10951 8 10952 8 10953 8 10954 8 10955 8 10956 8 10957 8 10958 8 10959 8 10960 8 10961 8 10962 8 10963 8 10964 8 10965 8 10966 4694 10967 8 10968 8 10969 8 10970 4695 10971 4696 10972 8 10973 4697 10974 8 10975 8 10976 4698 10977 4699 10978 4700 10979 8 10980 8 10981 4701 10982 4702 10983 4703 10984 4704 10985 8 10986 4705 10987 8 10988 4706 10989 4707 10990 4708 10991 4709 10992 4710 10993 4711 10994 4712 10995 4713 10996 4714 10997 4643 10998 4643 10999 4643 11000 4643 11001 4643 11002 4643 11003 4643 11004 4643 11005 4643 11006 4643 11007 4643 11008 4643 11009 4643 11010 4715 11011 4643 11012 4716 11013 4717 11014 8 11015 4718 11016 4719 11017 8 11018 4720 11019 8 11020 4721 11021 8 11022 8 11023 8 11024 4722 11025 4723 11026 4724 11027 8 11028 4725 11029 8 11030 8 11031 4726 11032 4727 11033 4728 11034 8 11035 8 11036 8 11037 4729 11038 4730 11039 8 11040 8 11041 8 11042 4731 11043 4732 11044 4733 11045 4734 11046 8 11047 8 11048 4735 11049 8 11050 4736 11051 4737 11052 8 11053 8 11054 4738 11055 8 11056 8 11057 8 11058 4739 11059 4740 11060 8 11061 8 11062 4741 11063 4742 11064 4743 11065 4744 11066 4745 11067 4746 11068 4747 11069 4748 11070 4749 11071 4750 11072 4751 11073 8 11074 4752 11075 4753 11076 4754 11077 4755 11078 4756 11079 4757 11080 4758 11081 4759 11082 4760 11083 4761 11084 8 11085 8 11086 4762 11087 8 11088 4763 11089 8 11090 4764 11091 4765 11092 8 11093 8 11094 8 11095 4766 11096 8 11097 8 11098 4767 11099 4768 11100 4769 11101 4770 11102 8 11103 8 11104 4771 11105 8 11106 4772 11107 4773 11108 8 11109 8 11110 4774 11111 8 11112 4775 11113 8 11114 4776 11115 4777 11116 4778 11117 4643 11118 4643 11119 4779 11120 4780 11121 4781 11122 4782 11123 4783 11124 4784 11125 4785 11126 4786 11127 4787 11128 4788 11129 8 11130 4789 11131 8 11132 4790 11133 8 11134 8 11135 4791 11136 4792 11137 4793 11138 4794 11139 8 11140 8 11141 4795 11142 4796 11143 4797 11144 4798 11145 4799 11146 8 11147 4800 11148 8 11149 4801 11150 8 11151 8 11152 8 11153 8 11154 4802 11155 8 11156 4803 11157 4804 11158 8 11159 4805 11160 4806 11161 4807 11162 4808 11163 4809 11164 8 11165 4810 11166 4811 11167 8 11168 4812 11169 8 11170 4813 11171 4814 11172 4815 11173 4816 11174 4817 11175 4818 11176 4819 11177 8 11178 8 11179 4820 11180 8 11181 8 11182 4821 11183 4822 11184 4823 11185 8 11186 8 11187 8 11188 8 11189 4824 11190 8 11191 8 11192 8 11193 8 11194 8 11195 4825 11196 4826 11197 4827 11198 4828 11199 4829 11200 4830 11201 4831 11202 4832 11203 4833 11204 4834 11205 4835 11206 4836 11207 4837 11208 4838 11209 4839 11210 4840 11211 4841 11212 4842 11213 4843 11214 4844 11215 4845 11216 4846 11217 4847 11218 4848 11219 4849 11220 4850 11221 4851 11222 4852 11223 4853 11224 4854 11225 4855 11226 4856 11227 4857 11228 4858 11229 4859 11230 4828 11231 4828 11232 4828 11233 4828 11234 4828 11235 4828 11236 4828 11237 4828 11238 4828 11239 4828 11240 4828 11241 4828 11242 4828 11243 4828 11244 4828 11245 4828 11246 4860 11247 4861 11248 8 11249 8 11250 8 11251 8 11252 8 11253 8 11254 4862 11255 8 11256 8 11257 4863 11258 8 11259 4864 11260 4865 11261 4866 11262 4867 11263 4868 11264 8 11265 8 11266 8 11267 8 11268 8 11269 4869 11270 4870 11271 8 11272 8 11273 4871 11274 8 11275 4872 11276 8 11277 8 11278 4873 11279 4874 11280 8 11281 4828 11282 4828 11283 4828 11284 4828 11285 4828 11286 4875 11287 4876 11288 4877 11289 4828 11290 4828 11291 4828 11292 4828 11293 4828 11294 4828 11295 4828 11296 4828 11297 4828 11298 4828 11299 4878 11300 8 11301 8 11302 8 11303 8 11304 8 11305 8 11306 8 11307 8 11308 8 11309 8 11310 8 11311 8 11312 8 11313 8 11314 8 11315 8 11316 8 11317 8 11318 8 11319 8 11320 8 11321 8 11322 8 11323 8 11324 8 11325 8 11326 8 11327 8 11328 8 11329 8 11330 8 11331 8 11332 8 11333 8 11334 8 11335 8 11336 8 11337 8 11338 8 11339 8 11340 8 11341 8 11342 8 11343 8 11344 8 11345 8 11346 8 11347 8 11348 8 11349 8 11350 8 11351 8 11352 8 11353 8 11354 8 11355 8 11356 8 11357 8 11358 8 11359 8 11360 8 11361 8 11362 8 11363 8 11364 8 11365 8 11366 8 11367 8 11368 8 11369 8 11370 8 11371 8 11372 8 11373 8 11374 8 11375 8 11376 8 11377 8 11378 8 11379 8 11380 4879 11381 8 11382 8 11383 8 11384 4880 11385 4881 11386 8 11387 4882 11388 8 11389 8 11390 4883 11391 4884 11392 4885 11393 8 11394 8 11395 4886 11396 4887 11397 4888 11398 4889 11399 8 11400 4890 11401 8 11402 4891 11403 4892 11404 4893 11405 4894 11406 4895 11407 4896 11408 4897 11409 4898 11410 4899 11411 4828 11412 4828 11413 4828 11414 4828 11415 4828 11416 4828 11417 4828 11418 4828 11419 4828 11420 4828 11421 4828 11422 4828 11423 4828 11424 4900 11425 4828 11426 4901 11427 4902 11428 8 11429 4903 11430 4904 11431 8 11432 4905 11433 8 11434 4906 11435 8 11436 8 11437 8 11438 4907 11439 4908 11440 4909 11441 8 11442 4910 11443 8 11444 8 11445 4911 11446 4912 11447 4913 11448 8 11449 8 11450 8 11451 4914 11452 4915 11453 8 11454 8 11455 8 11456 4916 11457 4917 11458 4918 11459 4919 11460 8 11461 8 11462 4920 11463 8 11464 4921 11465 4922 11466 8 11467 8 11468 4923 11469 8 11470 8 11471 8 11472 4924 11473 4925 11474 8 11475 8 11476 4926 11477 4927 11478 4928 11479 4929 11480 4930 11481 4931 11482 4932 11483 4933 11484 4934 11485 4935 11486 4936 11487 8 11488 4937 11489 4938 11490 4939 11491 4940 11492 4941 11493 4942 11494 4943 11495 4944 11496 4945 11497 4946 11498 8 11499 4947 11500 4643 11501 4948 11502 8 11503 4949 11504 8 11505 4950 11506 4951 11507 8 11508 8 11509 8 11510 4952 11511 8 11512 8 11513 4953 11514 4954 11515 4955 11516 4956 11517 8 11518 8 11519 4957 11520 8 11521 4958 11522 4959 11523 8 11524 8 11525 4960 11526 8 11527 4961 11528 8 11529 4962 11530 4963 11531 4964 11532 4828 11533 4828 11534 4965 11535 4966 11536 4828 11537 4828 11538 4828 11539 4828 11540 8 11541 4967 11542 8 11543 4968 11544 8 11545 8 11546 4828 11547 4828 11548 8 11549 4828 11550 4828 11551 4969 11552 8 11553 8 11554 8 11555 8 11556 8 11557 8 11558 4970 11559 4971 11560 4972 11561 4973 11562 8 11563 4974 11564 8 11565 4975 11566 4976 11567 4977 11568 4978 11569 4979 11570 8 11571 8 11572 8 11573 8 11574 8 11575 4980 11576 8 11577 8 11578 4981 11579 4982 11580 4983 11581 4984 11582 4985 11583 4986 11584 4987 11585 8 11586 4988 11587 8 11588 4989 11589 4990 11590 4991 11591 4992 11592 4993 11593 8 11594 4994 11595 4995 11596 4996 11597 4997 11598 4828 11599 4998 11600 4999 11601 5000 11602 8 11603 4828 11604 4828 11605 4828 11606 8 11607 5001 11608 8 11609 8 11610 8 11611 5002 11612 5003 11613 5004 11614 8 11615 8 11616 5005 11617 5006 11618 8 11619 5007 11620 5008 11621 8 11622 4828 11623 8 11624 5009 11625 8 11626 4828 11627 8 11628 5010 11629 5011 11630 5012 11631 5013 11632 8 11633 8 11634 5014 11635 8 11636 8 11637 4828 11638 4828 11639 5015 11640 5016 11641 8 11642 4828 11643 4828 11644 4828 11645 5017 11646 8 11647 5018 11648 5019 11649 8 11650 5020 11651 8 11652 4828 11653 4643 11654 4643 11655 4643 11656 8 11657 5021 11658 8 11659 5022 11660 8 11661 8 11662 4643 11663 4643 11664 8 11665 4643 11666 4643 11667 5023 11668 8 11669 8 11670 8 11671 8 11672 8 11673 8 11674 5024 11675 5025 11676 5026 11677 5027 11678 8 11679 5028 11680 8 11681 5029 11682 5030 11683 5031 11684 5032 11685 5033 11686 8 11687 8 11688 8 11689 8 11690 8 11691 5034 11692 8 11693 8 11694 5035 11695 5036 11696 5037 11697 5038 11698 5039 11699 5040 11700 5041 11701 8 11702 5042 11703 8 11704 5043 11705 5044 11706 5045 11707 5046 11708 5047 11709 8 11710 5048 11711 5049 11712 5050 11713 5051 11714 4643 11715 5052 11716 5053 11717 5054 11718 8 11719 4643 11720 4643 11721 4643 11722 8 11723 5055 11724 8 11725 8 11726 8 11727 5056 11728 5057 11729 5058 11730 8 11731 8 11732 5059 11733 5060 11734 8 11735 5061 11736 5062 11737 8 11738 4643 11739 8 11740 5063 11741 8 11742 4643 11743 8 11744 5064 11745 5065 11746 5066 11747 5067 11748 8 11749 8 11750 5068 11751 8 11752 8 11753 4643 11754 4643 11755 5069 11756 5070 11757 8 11758 4643 11759 4643 11760 4643 11761 5071 11762 8 11763 5072 11764 5073 11765 8 11766 5074 11767 8 11768 4643 11769 5075 11770 369 11771 5076 11772 8 11773 5077 11774 8 11775 5078 11776 5079 11777 8 11778 8 11779 8 11780 5080 11781 8 11782 8 11783 5081 11784 5082 11785 5083 11786 5084 11787 8 11788 8 11789 5085 11790 8 11791 5086 11792 5087 11793 8 11794 8 11795 5088 11796 8 11797 5089 11798 8 11799 5090 11800 5091 11801 5092 11802 466 11803 466 11804 5093 11805 5094 11806 5095 11807 5096 11808 5097 11809 5098 11810 5099 11811 5100 11812 5101 11813 5102 11814 8 11815 5103 11816 8 11817 5104 11818 8 11819 8 11820 5105 11821 5106 11822 5107 11823 5108 11824 8 11825 8 11826 5109 11827 5110 11828 5111 11829 5112 11830 5113 11831 8 11832 5114 11833 8 11834 5115 11835 8 11836 8 11837 8 11838 8 11839 5116 11840 8 11841 5117 11842 5118 11843 8 11844 5119 11845 5120 11846 5121 11847 5122 11848 5123 11849 8 11850 5124 11851 5125 11852 8 11853 5126 11854 8 11855 5127 11856 5128 11857 5129 11858 5130 11859 5131 11860 5132 11861 5133 11862 8 11863 8 11864 5134 11865 8 11866 8 11867 5135 11868 5136 11869 5137 11870 8 11871 8 11872 8 11873 8 11874 5138 11875 8 11876 8 11877 8 11878 8 11879 8 11880 5139 11881 5140 11882 5141 11883 5142 11884 5143 11885 5144 11886 5145 11887 5146 11888 5147 11889 5148 11890 5149 11891 5150 11892 5151 11893 5152 11894 5153 11895 5154 11896 5155 11897 5156 11898 5157 11899 5158 11900 5159 11901 5160 11902 5161 11903 5162 11904 5163 11905 5164 11906 5165 11907 5166 11908 5167 11909 5168 11910 5169 11911 5170 11912 5171 11913 5172 11914 5173 11915 5142 11916 5142 11917 5142 11918 5142 11919 5142 11920 5142 11921 5142 11922 5142 11923 5142 11924 5142 11925 5142 11926 5142 11927 5142 11928 5142 11929 5142 11930 5142 11931 5174 11932 5175 11933 8 11934 8 11935 8 11936 8 11937 8 11938 8 11939 5176 11940 8 11941 8 11942 5177 11943 8 11944 5178 11945 5179 11946 5180 11947 5181 11948 5182 11949 8 11950 8 11951 8 11952 8 11953 8 11954 5183 11955 5184 11956 8 11957 8 11958 5185 11959 8 11960 5186 11961 8 11962 8 11963 5187 11964 5188 11965 8 11966 5142 11967 5142 11968 5142 11969 5142 11970 5142 11971 5189 11972 5190 11973 5191 11974 5142 11975 5142 11976 5142 11977 5142 11978 5142 11979 5142 11980 5142 11981 5142 11982 5142 11983 5142 11984 5192 11985 8 11986 8 11987 8 11988 8 11989 8 11990 8 11991 8 11992 8 11993 8 11994 8 11995 8 11996 8 11997 8 11998 8 11999 8 12000 8 12001 8 12002 8 12003 8 12004 8 12005 8 12006 8 12007 8 12008 8 12009 8 12010 8 12011 8 12012 8 12013 8 12014 8 12015 8 12016 8 12017 8 12018 8 12019 8 12020 8 12021 8 12022 8 12023 8 12024 8 12025 8 12026 8 12027 8 12028 8 12029 8 12030 8 12031 8 12032 8 12033 8 12034 8 12035 8 12036 8 12037 8 12038 8 12039 8 12040 8 12041 8 12042 8 12043 8 12044 8 12045 8 12046 8 12047 8 12048 8 12049 8 12050 8 12051 8 12052 8 12053 8 12054 8 12055 8 12056 8 12057 8 12058 8 12059 8 12060 8 12061 8 12062 8 12063 8 12064 8 12065 5193 12066 8 12067 8 12068 8 12069 5194 12070 5195 12071 8 12072 5196 12073 8 12074 8 12075 5197 12076 5198 12077 5199 12078 8 12079 8 12080 5200 12081 5201 12082 5202 12083 5203 12084 8 12085 5204 12086 8 12087 5205 12088 5206 12089 5207 12090 5208 12091 5209 12092 5210 12093 5211 12094 5212 12095 5213 12096 5142 12097 5142 12098 5142 12099 5142 12100 5142 12101 5142 12102 5142 12103 5142 12104 5142 12105 5142 12106 5142 12107 5142 12108 5142 12109 5214 12110 5142 12111 5215 12112 5216 12113 8 12114 5217 12115 5218 12116 8 12117 5219 12118 8 12119 5220 12120 8 12121 8 12122 8 12123 5221 12124 5222 12125 5223 12126 8 12127 5224 12128 8 12129 8 12130 5225 12131 5226 12132 5227 12133 8 12134 8 12135 8 12136 5228 12137 5229 12138 8 12139 8 12140 8 12141 5230 12142 5231 12143 5232 12144 5233 12145 8 12146 8 12147 5234 12148 8 12149 5235 12150 5236 12151 8 12152 8 12153 5237 12154 8 12155 8 12156 8 12157 5238 12158 5239 12159 8 12160 8 12161 5240 12162 5241 12163 5242 12164 5243 12165 5244 12166 5245 12167 5246 12168 5247 12169 5248 12170 5249 12171 5250 12172 8 12173 5251 12174 5252 12175 5253 12176 5254 12177 5255 12178 5256 12179 5257 12180 5258 12181 5259 12182 5260 12183 8 12184 5261 12185 466 12186 5262 12187 8 12188 5263 12189 8 12190 5264 12191 5265 12192 8 12193 8 12194 8 12195 5266 12196 8 12197 8 12198 5267 12199 5268 12200 5269 12201 5270 12202 8 12203 8 12204 5271 12205 8 12206 5272 12207 5273 12208 8 12209 8 12210 5274 12211 8 12212 5275 12213 8 12214 5276 12215 5277 12216 5278 12217 5142 12218 5142 12219 5279 12220 5280 12221 5142 12222 5142 12223 5142 12224 5142 12225 8 12226 5281 12227 8 12228 5282 12229 8 12230 8 12231 5142 12232 5142 12233 8 12234 5142 12235 5142 12236 5283 12237 8 12238 8 12239 8 12240 8 12241 8 12242 8 12243 5284 12244 5285 12245 5286 12246 5287 12247 8 12248 5288 12249 8 12250 5289 12251 5290 12252 5291 12253 5292 12254 5293 12255 8 12256 8 12257 8 12258 8 12259 8 12260 5294 12261 8 12262 8 12263 5295 12264 5296 12265 5297 12266 5298 12267 5299 12268 5300 12269 5301 12270 8 12271 5302 12272 8 12273 5303 12274 5304 12275 5305 12276 5306 12277 5307 12278 8 12279 5308 12280 5309 12281 5310 12282 5311 12283 5142 12284 5312 12285 5313 12286 5314 12287 8 12288 5142 12289 5142 12290 5142 12291 8 12292 5315 12293 8 12294 8 12295 8 12296 5316 12297 5317 12298 5318 12299 8 12300 8 12301 5319 12302 5320 12303 8 12304 5321 12305 5322 12306 8 12307 5142 12308 8 12309 5323 12310 8 12311 5142 12312 8 12313 5324 12314 5325 12315 5326 12316 5327 12317 8 12318 8 12319 5328 12320 8 12321 8 12322 5142 12323 5142 12324 5329 12325 5330 12326 8 12327 5142 12328 5142 12329 5142 12330 5331 12331 8 12332 5332 12333 5333 12334 8 12335 5334 12336 8 12337 5142 12338 466 12339 466 12340 466 12341 8 12342 5335 12343 8 12344 5336 12345 8 12346 8 12347 466 12348 466 12349 8 12350 466 12351 466 12352 5337 12353 8 12354 8 12355 8 12356 8 12357 8 12358 8 12359 5338 12360 5339 12361 5340 12362 5341 12363 8 12364 5342 12365 8 12366 5343 12367 5344 12368 5345 12369 5346 12370 5347 12371 8 12372 8 12373 8 12374 8 12375 8 12376 5348 12377 8 12378 8 12379 5349 12380 5350 12381 5351 12382 5352 12383 5353 12384 5354 12385 5355 12386 8 12387 5356 12388 8 12389 5357 12390 5358 12391 5359 12392 5360 12393 5361 12394 8 12395 5362 12396 5363 12397 5364 12398 5365 12399 466 12400 5366 12401 5367 12402 5368 12403 8 12404 466 12405 466 12406 466 12407 8 12408 5369 12409 8 12410 8 12411 8 12412 5370 12413 5371 12414 5372 12415 8 12416 8 12417 5373 12418 5374 12419 8 12420 5375 12421 5376 12422 8 12423 466 12424 8 12425 5377 12426 8 12427 466 12428 8 12429 5378 12430 5379 12431 5380 12432 5381 12433 8 12434 8 12435 5382 12436 8 12437 8 12438 466 12439 466 12440 5383 12441 5384 12442 8 12443 466 12444 369 12445 466 12446 466 12447 5385 12448 8 12449 5386 12450 5387 12451 8 12452 5388 12453 8 12454 466 12455 369 12456 369 12457 369 12458 369 12459 369 12460 369 12461 369 12462 369 12463 5389 12464 8 12465 8 12466 8 12467 8 12468 8 12469 8 12470 8 12471 8 12472 8 12473 8 12474 8 12475 8 12476 8 12477 8 12478 8 12479 8 12480 8 12481 8 12482 8 12483 8 12484 8 12485 8 12486 8 12487 8 12488 8 12489 8 12490 8 12491 8 12492 8 12493 8 12494 8 12495 8 12496 8 12497 8 12498 8 12499 8 12500 8 12501 8 12502 8 12503 8 12504 8 12505 8 12506 8 12507 8 12508 8 12509 8 12510 8 12511 8 12512 8 12513 8 12514 8 12515 8 12516 8 12517 8 12518 8 12519 8 12520 8 12521 8 12522 8 12523 8 12524 8 12525 8 12526 8 12527 8 12528 8 12529 8 12530 8 12531 8 12532 8 12533 8 12534 8 12535 8 12536 8 12537 8 12538 8 12539 8 12540 8 12541 8 12542 8 12543 8 12544 5390 12545 8 12546 8 12547 8 12548 5391 12549 5392 12550 8 12551 5393 12552 8 12553 8 12554 5394 12555 5395 12556 5396 12557 8 12558 8 12559 5397 12560 5398 12561 5399 12562 5400 12563 8 12564 5401 12565 8 12566 5402 12567 5403 12568 5404 12569 5405 12570 5406 12571 5407 12572 5408 12573 5409 12574 5410 12575 369 12576 369 12577 369 12578 369 12579 369 12580 369 12581 369 12582 369 12583 369 12584 369 12585 369 12586 369 12587 369 12588 5411 12589 369 12590 5412 12591 5413 12592 8 12593 5414 12594 5415 12595 8 12596 5416 12597 8 12598 5417 12599 8 12600 8 12601 8 12602 5418 12603 5419 12604 5420 12605 8 12606 5421 12607 8 12608 8 12609 5422 12610 5423 12611 5424 12612 8 12613 8 12614 8 12615 5425 12616 5426 12617 8 12618 8 12619 8 12620 5427 12621 5428 12622 5429 12623 5430 12624 8 12625 8 12626 5431 12627 8 12628 5432 12629 5433 12630 8 12631 8 12632 5434 12633 8 12634 8 12635 8 12636 5435 12637 5436 12638 8 12639 8 12640 5437 12641 5438 12642 5439 12643 5440 12644 5441 12645 5442 12646 5443 12647 5444 12648 5445 12649 5446 12650 5447 12651 8 12652 5448 12653 5449 12654 5450 12655 5451 12656 5452 12657 5453 12658 5454 12659 5455 12660 5456 12661 5457 12662 5458 12663 5459 12664 5460 12665 5461 12666 5462 12667 5463 12668 5464 12669 5465 12670 5466 12671 8 12672 5467 12673 8 12674 5468 12675 8 12676 8 12677 5469 12678 5470 12679 5471 12680 5472 12681 8 12682 8 12683 5473 12684 5474 12685 5475 12686 5476 12687 5477 12688 8 12689 5478 12690 8 12691 5479 12692 8 12693 8 12694 8 12695 8 12696 5480 12697 8 12698 5481 12699 5482 12700 8 12701 5483 12702 5484 12703 5485 12704 5486 12705 5487 12706 8 12707 5488 12708 5489 12709 8 12710 5490 12711 8 12712 5491 12713 5492 12714 5493 12715 5494 12716 5495 12717 5496 12718 5497 12719 8 12720 8 12721 5498 12722 8 12723 8 12724 5499 12725 5500 12726 5501 12727 8 12728 8 12729 8 12730 8 12731 5502 12732 8 12733 8 12734 8 12735 8 12736 8 12737 5503 12738 5504 12739 5505 12740 5506 12741 5507 12742 5508 12743 5509 12744 5510 12745 5511 12746 5512 12747 5513 12748 5514 12749 5515 12750 5516 12751 5517 12752 5518 12753 5519 12754 5520 12755 5521 12756 5522 12757 5523 12758 5524 12759 5525 12760 5526 12761 5527 12762 5528 12763 5529 12764 5530 12765 5531 12766 5532 12767 5533 12768 5534 12769 5535 12770 5536 12771 5537 12772 5506 12773 5506 12774 5506 12775 5506 12776 5506 12777 5506 12778 5506 12779 5506 12780 5506 12781 5506 12782 5506 12783 5506 12784 5506 12785 5506 12786 5506 12787 5506 12788 5538 12789 5539 12790 8 12791 8 12792 8 12793 8 12794 8 12795 8 12796 5540 12797 8 12798 8 12799 5541 12800 8 12801 5542 12802 5543 12803 5544 12804 5545 12805 5546 12806 8 12807 8 12808 8 12809 8 12810 8 12811 5547 12812 5548 12813 8 12814 8 12815 5549 12816 8 12817 5550 12818 8 12819 8 12820 5551 12821 5552 12822 8 12823 5506 12824 5506 12825 5506 12826 5506 12827 5506 12828 5553 12829 5554 12830 5555 12831 5506 12832 5506 12833 5506 12834 5506 12835 5506 12836 5506 12837 5506 12838 5506 12839 5506 12840 5506 12841 5556 12842 8 12843 8 12844 8 12845 8 12846 8 12847 8 12848 8 12849 8 12850 8 12851 8 12852 8 12853 8 12854 8 12855 8 12856 8 12857 8 12858 8 12859 8 12860 8 12861 8 12862 8 12863 8 12864 8 12865 8 12866 8 12867 8 12868 8 12869 8 12870 8 12871 8 12872 8 12873 8 12874 8 12875 8 12876 8 12877 8 12878 8 12879 8 12880 8 12881 8 12882 8 12883 8 12884 8 12885 8 12886 8 12887 8 12888 8 12889 8 12890 8 12891 8 12892 8 12893 8 12894 8 12895 8 12896 8 12897 8 12898 8 12899 8 12900 8 12901 8 12902 8 12903 8 12904 8 12905 8 12906 8 12907 8 12908 8 12909 8 12910 8 12911 8 12912 8 12913 8 12914 8 12915 8 12916 8 12917 8 12918 8 12919 8 12920 8 12921 8 12922 5557 12923 8 12924 8 12925 8 12926 5558 12927 5559 12928 8 12929 5560 12930 8 12931 8 12932 5561 12933 5562 12934 5563 12935 8 12936 8 12937 5564 12938 5565 12939 5566 12940 5567 12941 8 12942 5568 12943 8 12944 5569 12945 5570 12946 5571 12947 5572 12948 5573 12949 5574 12950 5575 12951 5576 12952 5577 12953 5506 12954 5506 12955 5506 12956 5506 12957 5506 12958 5506 12959 5506 12960 5506 12961 5506 12962 5506 12963 5506 12964 5506 12965 5506 12966 5578 12967 5506 12968 5579 12969 5580 12970 8 12971 5581 12972 5582 12973 8 12974 5583 12975 8 12976 5584 12977 8 12978 8 12979 8 12980 5585 12981 5586 12982 5587 12983 8 12984 5588 12985 8 12986 8 12987 5589 12988 5590 12989 5591 12990 8 12991 8 12992 8 12993 5592 12994 5593 12995 8 12996 8 12997 8 12998 5594 12999 5595 13000 5596 13001 5597 13002 8 13003 8 13004 5598 13005 8 13006 5599 13007 5600 13008 8 13009 8 13010 5601 13011 8 13012 8 13013 8 13014 5602 13015 5603 13016 8 13017 8 13018 5604 13019 5605 13020 5606 13021 5607 13022 5608 13023 5609 13024 5610 13025 5611 13026 5612 13027 5613 13028 5614 13029 8 13030 5615 13031 5616 13032 5617 13033 5618 13034 5619 13035 5620 13036 5621 13037 5622 13038 5623 13039 5624 13040 8 13041 8 13042 5625 13043 8 13044 5626 13045 8 13046 5627 13047 5628 13048 8 13049 8 13050 8 13051 5629 13052 8 13053 8 13054 5630 13055 5631 13056 5632 13057 5633 13058 8 13059 8 13060 5634 13061 8 13062 5635 13063 5636 13064 8 13065 8 13066 5637 13067 8 13068 5638 13069 8 13070 5639 13071 5640 13072 5641 13073 5506 13074 5506 13075 5642 13076 5643 13077 5644 13078 5645 13079 5646 13080 5647 13081 5648 13082 5649 13083 5650 13084 5651 13085 8 13086 5652 13087 8 13088 5653 13089 8 13090 8 13091 5654 13092 5655 13093 5656 13094 5657 13095 8 13096 8 13097 5658 13098 5659 13099 5660 13100 5661 13101 5662 13102 8 13103 5663 13104 8 13105 5664 13106 8 13107 8 13108 8 13109 8 13110 5665 13111 8 13112 5666 13113 5667 13114 8 13115 5668 13116 5669 13117 5670 13118 5671 13119 5672 13120 8 13121 5673 13122 5674 13123 8 13124 5675 13125 8 13126 5676 13127 5677 13128 5678 13129 5679 13130 5680 13131 5681 13132 5682 13133 8 13134 8 13135 5683 13136 8 13137 8 13138 5684 13139 5685 13140 5686 13141 8 13142 8 13143 8 13144 8 13145 5687 13146 8 13147 8 13148 8 13149 8 13150 8 13151 5688 13152 5689 13153 5690 13154 5691 13155 5692 13156 5693 13157 5694 13158 5695 13159 5696 13160 5697 13161 5698 13162 5699 13163 5700 13164 5701 13165 5702 13166 5703 13167 5704 13168 5705 13169 5706 13170 5707 13171 5708 13172 5709 13173 5710 13174 5711 13175 5712 13176 5713 13177 5714 13178 5715 13179 5716 13180 5717 13181 5718 13182 5719 13183 5720 13184 5721 13185 5722 13186 5691 13187 5691 13188 5691 13189 5691 13190 5691 13191 5691 13192 5691 13193 5691 13194 5691 13195 5691 13196 5691 13197 5691 13198 5691 13199 5691 13200 5691 13201 5691 13202 5723 13203 5724 13204 8 13205 8 13206 8 13207 8 13208 8 13209 8 13210 5725 13211 8 13212 8 13213 5726 13214 8 13215 5727 13216 5728 13217 5729 13218 5730 13219 5731 13220 8 13221 8 13222 8 13223 8 13224 8 13225 5732 13226 5733 13227 8 13228 8 13229 5734 13230 8 13231 5735 13232 8 13233 8 13234 5736 13235 5737 13236 8 13237 5691 13238 5691 13239 5691 13240 5691 13241 5691 13242 5738 13243 5739 13244 5740 13245 5691 13246 5691 13247 5691 13248 5691 13249 5691 13250 5691 13251 5691 13252 5691 13253 5691 13254 5691 13255 5741 13256 8 13257 8 13258 8 13259 8 13260 8 13261 8 13262 8 13263 8 13264 8 13265 8 13266 8 13267 8 13268 8 13269 8 13270 8 13271 8 13272 8 13273 8 13274 8 13275 8 13276 8 13277 8 13278 8 13279 8 13280 8 13281 8 13282 8 13283 8 13284 8 13285 8 13286 8 13287 8 13288 8 13289 8 13290 8 13291 8 13292 8 13293 8 13294 8 13295 8 13296 8 13297 8 13298 8 13299 8 13300 8 13301 8 13302 8 13303 8 13304 8 13305 8 13306 8 13307 8 13308 8 13309 8 13310 8 13311 8 13312 8 13313 8 13314 8 13315 8 13316 8 13317 8 13318 8 13319 8 13320 8 13321 8 13322 8 13323 8 13324 8 13325 8 13326 8 13327 8 13328 8 13329 8 13330 8 13331 8 13332 8 13333 8 13334 8 13335 8 13336 5742 13337 8 13338 8 13339 8 13340 5743 13341 5744 13342 8 13343 5745 13344 8 13345 8 13346 5746 13347 5747 13348 5748 13349 8 13350 8 13351 5749 13352 5750 13353 5751 13354 5752 13355 8 13356 5753 13357 8 13358 5754 13359 5755 13360 5756 13361 5757 13362 5758 13363 5759 13364 5760 13365 5761 13366 5762 13367 5691 13368 5691 13369 5691 13370 5691 13371 5691 13372 5691 13373 5691 13374 5691 13375 5691 13376 5691 13377 5691 13378 5691 13379 5691 13380 5763 13381 5691 13382 5764 13383 5765 13384 8 13385 5766 13386 5767 13387 8 13388 5768 13389 8 13390 5769 13391 8 13392 8 13393 8 13394 5770 13395 5771 13396 5772 13397 8 13398 5773 13399 8 13400 8 13401 5774 13402 5775 13403 5776 13404 8 13405 8 13406 8 13407 5777 13408 5778 13409 8 13410 8 13411 8 13412 5779 13413 5780 13414 5781 13415 5782 13416 8 13417 8 13418 5783 13419 8 13420 5784 13421 5785 13422 8 13423 8 13424 5786 13425 8 13426 8 13427 8 13428 5787 13429 5788 13430 8 13431 8 13432 5789 13433 5790 13434 5791 13435 5792 13436 5793 13437 5794 13438 5795 13439 5796 13440 5797 13441 5798 13442 5799 13443 8 13444 5800 13445 5801 13446 5802 13447 5803 13448 5804 13449 5805 13450 5806 13451 5807 13452 5808 13453 5809 13454 8 13455 5810 13456 5506 13457 5811 13458 8 13459 5812 13460 8 13461 5813 13462 5814 13463 8 13464 8 13465 8 13466 5815 13467 8 13468 8 13469 5816 13470 5817 13471 5818 13472 5819 13473 8 13474 8 13475 5820 13476 8 13477 5821 13478 5822 13479 8 13480 8 13481 5823 13482 8 13483 5824 13484 8 13485 5825 13486 5826 13487 5827 13488 5691 13489 5691 13490 5828 13491 5829 13492 5691 13493 5691 13494 5691 13495 5691 13496 8 13497 5830 13498 8 13499 5831 13500 8 13501 8 13502 5691 13503 5691 13504 8 13505 5691 13506 5691 13507 5832 13508 8 13509 8 13510 8 13511 8 13512 8 13513 8 13514 5833 13515 5834 13516 5835 13517 5836 13518 8 13519 5837 13520 8 13521 5838 13522 5839 13523 5840 13524 5841 13525 5842 13526 8 13527 8 13528 8 13529 8 13530 8 13531 5843 13532 8 13533 8 13534 5844 13535 5845 13536 5846 13537 5847 13538 5848 13539 5849 13540 5850 13541 8 13542 5851 13543 8 13544 5852 13545 5853 13546 5854 13547 5855 13548 5856 13549 8 13550 5857 13551 5858 13552 5859 13553 5860 13554 5691 13555 5861 13556 5862 13557 5863 13558 8 13559 5691 13560 5691 13561 5691 13562 8 13563 5864 13564 8 13565 8 13566 8 13567 5865 13568 5866 13569 5867 13570 8 13571 8 13572 5868 13573 5869 13574 8 13575 5870 13576 5871 13577 8 13578 5691 13579 8 13580 5872 13581 8 13582 5691 13583 8 13584 5873 13585 5874 13586 5875 13587 5876 13588 8 13589 8 13590 5877 13591 8 13592 8 13593 5691 13594 5691 13595 5878 13596 5879 13597 8 13598 5691 13599 5691 13600 5691 13601 5880 13602 8 13603 5881 13604 5882 13605 8 13606 5883 13607 8 13608 5691 13609 5506 13610 5506 13611 5506 13612 8 13613 5884 13614 8 13615 5885 13616 8 13617 8 13618 5506 13619 5506 13620 8 13621 5506 13622 5506 13623 5886 13624 8 13625 8 13626 8 13627 8 13628 8 13629 8 13630 5887 13631 5888 13632 5889 13633 5890 13634 8 13635 5891 13636 8 13637 5892 13638 5893 13639 5894 13640 5895 13641 5896 13642 8 13643 8 13644 8 13645 8 13646 8 13647 5897 13648 8 13649 8 13650 5898 13651 5899 13652 5900 13653 5901 13654 5902 13655 5903 13656 5904 13657 8 13658 5905 13659 8 13660 5906 13661 5907 13662 5908 13663 5909 13664 5910 13665 8 13666 5911 13667 5912 13668 5913 13669 5914 13670 5506 13671 5915 13672 5916 13673 5917 13674 8 13675 5506 13676 5506 13677 5506 13678 8 13679 5918 13680 8 13681 8 13682 8 13683 5919 13684 5920 13685 5921 13686 8 13687 8 13688 5922 13689 5923 13690 8 13691 5924 13692 5925 13693 8 13694 5506 13695 8 13696 5926 13697 8 13698 5506 13699 8 13700 5927 13701 5928 13702 5929 13703 5930 13704 8 13705 8 13706 5931 13707 8 13708 8 13709 5506 13710 5506 13711 5932 13712 5933 13713 8 13714 5506 13715 5506 13716 5506 13717 5934 13718 8 13719 5935 13720 5936 13721 8 13722 5937 13723 8 13724 5506 13725 8 13726 5938 13727 8 13728 5939 13729 8 13730 5940 13731 5941 13732 8 13733 8 13734 8 13735 5942 13736 8 13737 8 13738 5943 13739 5944 13740 5945 13741 5946 13742 8 13743 8 13744 5947 13745 8 13746 5948 13747 5949 13748 8 13749 8 13750 5950 13751 8 13752 5951 13753 8 13754 5952 13755 5953 13756 5954 13757 369 13758 369 13759 5955 13760 5956 13761 5957 13762 5958 13763 5959 13764 5960 13765 5961 13766 5962 13767 5963 13768 5964 13769 8 13770 5965 13771 8 13772 5966 13773 8 13774 8 13775 5967 13776 5968 13777 5969 13778 5970 13779 8 13780 8 13781 5971 13782 5972 13783 5973 13784 5974 13785 5975 13786 8 13787 5976 13788 8 13789 5977 13790 8 13791 8 13792 8 13793 8 13794 5978 13795 8 13796 5979 13797 5980 13798 8 13799 5981 13800 5982 13801 5983 13802 5984 13803 5985 13804 8 13805 5986 13806 5987 13807 8 13808 5988 13809 8 13810 5989 13811 5990 13812 5991 13813 5992 13814 5993 13815 5994 13816 5995 13817 8 13818 8 13819 5996 13820 8 13821 8 13822 5997 13823 5998 13824 5999 13825 8 13826 8 13827 8 13828 8 13829 6000 13830 8 13831 8 13832 8 13833 8 13834 8 13835 6001 13836 6002 13837 6003 13838 6004 13839 6005 13840 6006 13841 6007 13842 6008 13843 6009 13844 6010 13845 6011 13846 6012 13847 6013 13848 6014 13849 6015 13850 6016 13851 6017 13852 6018 13853 6019 13854 6020 13855 6021 13856 6022 13857 6023 13858 6024 13859 6025 13860 6026 13861 6027 13862 6028 13863 6029 13864 6030 13865 6031 13866 6032 13867 6033 13868 6034 13869 6035 13870 6004 13871 6004 13872 6004 13873 6004 13874 6004 13875 6004 13876 6004 13877 6004 13878 6004 13879 6004 13880 6004 13881 6004 13882 6004 13883 6004 13884 6004 13885 6004 13886 6036 13887 6037 13888 8 13889 8 13890 8 13891 8 13892 8 13893 8 13894 6038 13895 8 13896 8 13897 6039 13898 8 13899 6040 13900 6041 13901 6042 13902 6043 13903 6044 13904 8 13905 8 13906 8 13907 8 13908 8 13909 6045 13910 6046 13911 8 13912 8 13913 6047 13914 8 13915 6048 13916 8 13917 8 13918 6049 13919 6050 13920 8 13921 6004 13922 6004 13923 6004 13924 6004 13925 6004 13926 6051 13927 6052 13928 6053 13929 6004 13930 6004 13931 6004 13932 6004 13933 6004 13934 6004 13935 6004 13936 6004 13937 6004 13938 6004 13939 6054 13940 8 13941 8 13942 8 13943 8 13944 8 13945 8 13946 8 13947 8 13948 8 13949 8 13950 8 13951 8 13952 8 13953 8 13954 8 13955 8 13956 8 13957 8 13958 8 13959 8 13960 8 13961 8 13962 8 13963 8 13964 8 13965 8 13966 8 13967 8 13968 8 13969 8 13970 8 13971 8 13972 8 13973 8 13974 8 13975 8 13976 8 13977 8 13978 8 13979 8 13980 8 13981 8 13982 8 13983 8 13984 8 13985 8 13986 8 13987 8 13988 8 13989 8 13990 8 13991 8 13992 8 13993 8 13994 8 13995 8 13996 8 13997 8 13998 8 13999 8 14000 8 14001 8 14002 8 14003 8 14004 8 14005 8 14006 8 14007 8 14008 8 14009 8 14010 8 14011 8 14012 8 14013 8 14014 8 14015 8 14016 8 14017 8 14018 8 14019 8 14020 6055 14021 8 14022 8 14023 8 14024 6056 14025 6057 14026 8 14027 6058 14028 8 14029 8 14030 6059 14031 6060 14032 6061 14033 8 14034 8 14035 6062 14036 6063 14037 6064 14038 6065 14039 8 14040 6066 14041 8 14042 6067 14043 6068 14044 6069 14045 6070 14046 6071 14047 6072 14048 6073 14049 6074 14050 6075 14051 6004 14052 6004 14053 6004 14054 6004 14055 6004 14056 6004 14057 6004 14058 6004 14059 6004 14060 6004 14061 6004 14062 6004 14063 6004 14064 6076 14065 6004 14066 6077 14067 6078 14068 8 14069 6079 14070 6080 14071 8 14072 6081 14073 8 14074 6082 14075 8 14076 8 14077 8 14078 6083 14079 6084 14080 6085 14081 8 14082 6086 14083 8 14084 8 14085 6087 14086 6088 14087 6089 14088 8 14089 8 14090 8 14091 6090 14092 6091 14093 8 14094 8 14095 8 14096 6092 14097 6093 14098 6094 14099 6095 14100 8 14101 8 14102 6096 14103 8 14104 6097 14105 6098 14106 8 14107 8 14108 6099 14109 8 14110 8 14111 8 14112 6100 14113 6101 14114 8 14115 8 14116 6102 14117 6103 14118 6104 14119 6105 14120 6106 14121 6107 14122 6108 14123 6109 14124 6110 14125 6111 14126 6112 14127 8 14128 6113 14129 6114 14130 6115 14131 6116 14132 6117 14133 6118 14134 6119 14135 6120 14136 6121 14137 6122 14138 8 14139 6123 14140 369 14141 6124 14142 8 14143 6125 14144 8 14145 6126 14146 6127 14147 8 14148 8 14149 8 14150 6128 14151 8 14152 8 14153 6129 14154 6130 14155 6131 14156 6132 14157 8 14158 8 14159 6133 14160 8 14161 6134 14162 6135 14163 8 14164 8 14165 6136 14166 8 14167 6137 14168 8 14169 6138 14170 6139 14171 6140 14172 6004 14173 6004 14174 6141 14175 6142 14176 6004 14177 6004 14178 6004 14179 6004 14180 8 14181 6143 14182 8 14183 6144 14184 8 14185 8 14186 6004 14187 6004 14188 8 14189 6004 14190 6004 14191 6145 14192 8 14193 8 14194 8 14195 8 14196 8 14197 8 14198 6146 14199 6147 14200 6148 14201 6149 14202 8 14203 6150 14204 8 14205 6151 14206 6152 14207 6153 14208 6154 14209 6155 14210 8 14211 8 14212 8 14213 8 14214 8 14215 6156 14216 8 14217 8 14218 6157 14219 6158 14220 6159 14221 6160 14222 6161 14223 6162 14224 6163 14225 8 14226 6164 14227 8 14228 6165 14229 6166 14230 6167 14231 6168 14232 6169 14233 8 14234 6170 14235 6171 14236 6172 14237 6173 14238 6004 14239 6174 14240 6175 14241 6176 14242 8 14243 6004 14244 6004 14245 6004 14246 8 14247 6177 14248 8 14249 8 14250 8 14251 6178 14252 6179 14253 6180 14254 8 14255 8 14256 6181 14257 6182 14258 8 14259 6183 14260 6184 14261 8 14262 6004 14263 8 14264 6185 14265 8 14266 6004 14267 8 14268 6186 14269 6187 14270 6188 14271 6189 14272 8 14273 8 14274 6190 14275 8 14276 8 14277 6004 14278 6004 14279 6191 14280 6192 14281 8 14282 6004 14283 6004 14284 6004 14285 6193 14286 8 14287 6194 14288 6195 14289 8 14290 6196 14291 8 14292 6004 14293 369 14294 369 14295 369 14296 8 14297 6197 14298 8 14299 6198 14300 8 14301 8 14302 369 14303 369 14304 8 14305 369 14306 369 14307 6199 14308 8 14309 8 14310 8 14311 8 14312 8 14313 8 14314 6200 14315 6201 14316 6202 14317 6203 14318 8 14319 6204 14320 8 14321 6205 14322 6206 14323 6207 14324 6208 14325 6209 14326 8 14327 8 14328 8 14329 8 14330 8 14331 6210 14332 8 14333 8 14334 6211 14335 6212 14336 6213 14337 6214 14338 6215 14339 6216 14340 6217 14341 8 14342 6218 14343 8 14344 6219 14345 6220 14346 6221 14347 6222 14348 6223 14349 8 14350 6224 14351 6225 14352 6226 14353 6227 14354 369 14355 6228 14356 6229 14357 6230 14358 8 14359 369 14360 369 14361 369 14362 8 14363 6231 14364 8 14365 8 14366 8 14367 6232 14368 6233 14369 6234 14370 8 14371 8 14372 6235 14373 6236 14374 8 14375 6237 14376 6238 14377 8 14378 369 14379 8 14380 6239 14381 8 14382 369 14383 8 14384 6240 14385 6241 14386 6242 14387 6243 14388 8 14389 8 14390 6244 14391 8 14392 8 14393 369 14394 369 14395 6245 14396 6246 14397 8 14398 369 14399 369 14400 369 14401 6247 14402 8 14403 6248 14404 6249 14405 8 14406 6250 14407 8 14408 369 14409 8 14410 6251 14411 8 14412 6252 14413 8 14414 8 14415 6253 14416 6254 14417 8 14418 271 14419 271 14420 271 14421 271 14422 271 14423 6255 14424 6256 14425 6257 14426 271 14427 271 14428 271 14429 271 14430 271 14431 271 14432 271 14433 271 14434 271 14435 271 14436 6258 14437 8 14438 8 14439 8 14440 8 14441 8 14442 8 14443 8 14444 8 14445 8 14446 8 14447 8 14448 8 14449 8 14450 8 14451 8 14452 8 14453 8 14454 8 14455 8 14456 8 14457 8 14458 8 14459 8 14460 8 14461 8 14462 8 14463 8 14464 8 14465 8 14466 8 14467 8 14468 8 14469 8 14470 8 14471 8 14472 8 14473 8 14474 8 14475 8 14476 8 14477 8 14478 8 14479 8 14480 8 14481 8 14482 8 14483 8 14484 8 14485 8 14486 8 14487 8 14488 8 14489 8 14490 8 14491 8 14492 8 14493 8 14494 8 14495 8 14496 8 14497 8 14498 8 14499 8 14500 8 14501 8 14502 8 14503 8 14504 8 14505 8 14506 8 14507 8 14508 8 14509 8 14510 8 14511 8 14512 8 14513 8 14514 8 14515 8 14516 8 14517 6259 14518 8 14519 8 14520 8 14521 6260 14522 6261 14523 8 14524 6262 14525 8 14526 8 14527 6263 14528 6264 14529 6265 14530 8 14531 8 14532 6266 14533 6267 14534 6268 14535 6269 14536 8 14537 6270 14538 8 14539 6271 14540 6272 14541 6273 14542 6274 14543 6275 14544 6276 14545 6277 14546 6278 14547 6279 14548 271 14549 271 14550 271 14551 271 14552 271 14553 271 14554 271 14555 271 14556 271 14557 271 14558 271 14559 271 14560 271 14561 6280 14562 271 14563 6281 14564 6282 14565 8 14566 6283 14567 6284 14568 8 14569 6285 14570 8 14571 6286 14572 8 14573 8 14574 8 14575 6287 14576 6288 14577 6289 14578 8 14579 6290 14580 8 14581 8 14582 6291 14583 6292 14584 6293 14585 8 14586 8 14587 8 14588 6294 14589 6295 14590 8 14591 8 14592 8 14593 6296 14594 6297 14595 6298 14596 6299 14597 8 14598 8 14599 6300 14600 8 14601 6301 14602 6302 14603 8 14604 8 14605 6303 14606 8 14607 8 14608 8 14609 6304 14610 6305 14611 8 14612 8 14613 8 14614 6306 14615 6307 14616 6308 14617 6309 14618 6310 14619 6311 14620 6312 14621 6313 14622 6314 14623 6315 14624 6316 14625 8 14626 6317 14627 6318 14628 6319 14629 6320 14630 6321 14631 6322 14632 6323 14633 6324 14634 6325 14635 6326 14636 6327 14637 6328 14638 6329 14639 6330 14640 6331 14641 6332 14642 6333 14643 6334 14644 6335 14645 8 14646 6336 14647 8 14648 6337 14649 8 14650 8 14651 6338 14652 6339 14653 6340 14654 6341 14655 8 14656 8 14657 6342 14658 6343 14659 6344 14660 6345 14661 6346 14662 8 14663 6347 14664 8 14665 6348 14666 8 14667 8 14668 8 14669 8 14670 6349 14671 8 14672 6350 14673 6351 14674 8 14675 6352 14676 6353 14677 6354 14678 6355 14679 6356 14680 8 14681 6357 14682 6358 14683 8 14684 6359 14685 8 14686 6360 14687 6361 14688 6362 14689 6363 14690 6364 14691 6365 14692 6366 14693 8 14694 8 14695 6367 14696 8 14697 8 14698 6368 14699 6369 14700 6370 14701 8 14702 8 14703 8 14704 8 14705 6371 14706 8 14707 8 14708 8 14709 8 14710 8 14711 6372 14712 6373 14713 6374 14714 6375 14715 6376 14716 6377 14717 6378 14718 6379 14719 6380 14720 6381 14721 6382 14722 6383 14723 6384 14724 6385 14725 6386 14726 6387 14727 6388 14728 6389 14729 6390 14730 6391 14731 6392 14732 6393 14733 6394 14734 6395 14735 6396 14736 6397 14737 6398 14738 6399 14739 6400 14740 6401 14741 6402 14742 6403 14743 6404 14744 6405 14745 6406 14746 6375 14747 6375 14748 6375 14749 6375 14750 6375 14751 6375 14752 6375 14753 6375 14754 6375 14755 6375 14756 6375 14757 6375 14758 6375 14759 6375 14760 6375 14761 6375 14762 6407 14763 6408 14764 8 14765 8 14766 8 14767 8 14768 8 14769 8 14770 6409 14771 8 14772 8 14773 6410 14774 8 14775 6411 14776 6412 14777 6413 14778 6414 14779 6415 14780 8 14781 8 14782 8 14783 8 14784 8 14785 6416 14786 6417 14787 8 14788 8 14789 6418 14790 8 14791 6419 14792 8 14793 8 14794 6420 14795 6421 14796 8 14797 6375 14798 6375 14799 6375 14800 6375 14801 6375 14802 6422 14803 6423 14804 6424 14805 6375 14806 6375 14807 6375 14808 6375 14809 6375 14810 6375 14811 6375 14812 6375 14813 6375 14814 6375 14815 6425 14816 8 14817 8 14818 8 14819 8 14820 8 14821 8 14822 8 14823 8 14824 8 14825 8 14826 8 14827 8 14828 8 14829 8 14830 8 14831 8 14832 8 14833 8 14834 8 14835 8 14836 8 14837 8 14838 8 14839 8 14840 8 14841 8 14842 8 14843 8 14844 8 14845 8 14846 8 14847 8 14848 8 14849 8 14850 8 14851 8 14852 8 14853 8 14854 8 14855 8 14856 8 14857 8 14858 8 14859 8 14860 8 14861 8 14862 8 14863 8 14864 8 14865 8 14866 8 14867 8 14868 8 14869 8 14870 8 14871 8 14872 8 14873 8 14874 8 14875 8 14876 8 14877 8 14878 8 14879 8 14880 8 14881 8 14882 8 14883 8 14884 8 14885 8 14886 8 14887 8 14888 8 14889 8 14890 8 14891 8 14892 8 14893 8 14894 8 14895 8 14896 6426 14897 8 14898 8 14899 8 14900 6427 14901 6428 14902 8 14903 6429 14904 8 14905 8 14906 6430 14907 6431 14908 6432 14909 8 14910 8 14911 6433 14912 6434 14913 6435 14914 6436 14915 8 14916 6437 14917 8 14918 6438 14919 6439 14920 6440 14921 6441 14922 6442 14923 6443 14924 6444 14925 6445 14926 6446 14927 6375 14928 6375 14929 6375 14930 6375 14931 6375 14932 6375 14933 6375 14934 6375 14935 6375 14936 6375 14937 6375 14938 6375 14939 6375 14940 6447 14941 6375 14942 6448 14943 6449 14944 8 14945 6450 14946 6451 14947 8 14948 6452 14949 8 14950 6453 14951 8 14952 8 14953 8 14954 6454 14955 6455 14956 6456 14957 8 14958 6457 14959 8 14960 8 14961 6458 14962 6459 14963 6460 14964 8 14965 8 14966 8 14967 6461 14968 6462 14969 8 14970 8 14971 8 14972 6463 14973 6464 14974 6465 14975 6466 14976 8 14977 8 14978 6467 14979 8 14980 6468 14981 6469 14982 8 14983 8 14984 6470 14985 8 14986 8 14987 8 14988 6471 14989 6472 14990 8 14991 8 14992 6473 14993 6474 14994 6475 14995 6476 14996 6477 14997 6478 14998 6479 14999 6480 15000 6481 15001 6482 15002 6483 15003 8 15004 6484 15005 6485 15006 6486 15007 6487 15008 6488 15009 6489 15010 6490 15011 6491 15012 6492 15013 6493 15014 8 15015 8 15016 6494 15017 8 15018 6495 15019 8 15020 6496 15021 6497 15022 8 15023 8 15024 8 15025 6498 15026 8 15027 8 15028 6499 15029 6500 15030 6501 15031 6502 15032 8 15033 8 15034 6503 15035 8 15036 6504 15037 6505 15038 8 15039 8 15040 6506 15041 8 15042 6507 15043 8 15044 6508 15045 6509 15046 6510 15047 6375 15048 6375 15049 6511 15050 6512 15051 6513 15052 6514 15053 6515 15054 6516 15055 6517 15056 6518 15057 6519 15058 6520 15059 8 15060 6521 15061 8 15062 6522 15063 8 15064 8 15065 6523 15066 6524 15067 6525 15068 6526 15069 8 15070 8 15071 6527 15072 6528 15073 6529 15074 6530 15075 6531 15076 8 15077 6532 15078 8 15079 6533 15080 8 15081 8 15082 8 15083 8 15084 6534 15085 8 15086 6535 15087 6536 15088 8 15089 6537 15090 6538 15091 6539 15092 6540 15093 6541 15094 8 15095 6542 15096 6543 15097 8 15098 6544 15099 8 15100 6545 15101 6546 15102 6547 15103 6548 15104 6549 15105 6550 15106 6551 15107 8 15108 8 15109 6552 15110 8 15111 8 15112 6553 15113 6554 15114 6555 15115 8 15116 8 15117 8 15118 8 15119 6556 15120 8 15121 8 15122 8 15123 8 15124 8 15125 6557 15126 6558 15127 6559 15128 6560 15129 6561 15130 6562 15131 6563 15132 6564 15133 6565 15134 6566 15135 6567 15136 6568 15137 6569 15138 6570 15139 6571 15140 6572 15141 6573 15142 6574 15143 6575 15144 6576 15145 6577 15146 6578 15147 6579 15148 6580 15149 6581 15150 6582 15151 6583 15152 6584 15153 6585 15154 6586 15155 6587 15156 6588 15157 6589 15158 6590 15159 6591 15160 6560 15161 6560 15162 6560 15163 6560 15164 6560 15165 6560 15166 6560 15167 6560 15168 6560 15169 6560 15170 6560 15171 6560 15172 6560 15173 6560 15174 6560 15175 6560 15176 6592 15177 6593 15178 8 15179 8 15180 8 15181 8 15182 8 15183 8 15184 6594 15185 8 15186 8 15187 6595 15188 8 15189 6596 15190 6597 15191 6598 15192 6599 15193 6600 15194 8 15195 8 15196 8 15197 8 15198 8 15199 6601 15200 6602 15201 8 15202 8 15203 6603 15204 8 15205 6604 15206 8 15207 8 15208 6605 15209 6606 15210 8 15211 6560 15212 6560 15213 6560 15214 6560 15215 6560 15216 6607 15217 6608 15218 6609 15219 6560 15220 6560 15221 6560 15222 6560 15223 6560 15224 6560 15225 6560 15226 6560 15227 6560 15228 6560 15229 6610 15230 8 15231 8 15232 8 15233 8 15234 8 15235 8 15236 8 15237 8 15238 8 15239 8 15240 8 15241 8 15242 8 15243 8 15244 8 15245 8 15246 8 15247 8 15248 8 15249 8 15250 8 15251 8 15252 8 15253 8 15254 8 15255 8 15256 8 15257 8 15258 8 15259 8 15260 8 15261 8 15262 8 15263 8 15264 8 15265 8 15266 8 15267 8 15268 8 15269 8 15270 8 15271 8 15272 8 15273 8 15274 8 15275 8 15276 8 15277 8 15278 8 15279 8 15280 8 15281 8 15282 8 15283 8 15284 8 15285 8 15286 8 15287 8 15288 8 15289 8 15290 8 15291 8 15292 8 15293 8 15294 8 15295 8 15296 8 15297 8 15298 8 15299 8 15300 8 15301 8 15302 8 15303 8 15304 8 15305 8 15306 8 15307 8 15308 8 15309 8 15310 6611 15311 8 15312 8 15313 8 15314 6612 15315 6613 15316 8 15317 6614 15318 8 15319 8 15320 6615 15321 6616 15322 6617 15323 8 15324 8 15325 6618 15326 6619 15327 6620 15328 6621 15329 8 15330 6622 15331 8 15332 6623 15333 6624 15334 6625 15335 6626 15336 6627 15337 6628 15338 6629 15339 6630 15340 6631 15341 6560 15342 6560 15343 6560 15344 6560 15345 6560 15346 6560 15347 6560 15348 6560 15349 6560 15350 6560 15351 6560 15352 6560 15353 6560 15354 6632 15355 6560 15356 6633 15357 6634 15358 8 15359 6635 15360 6636 15361 8 15362 6637 15363 8 15364 6638 15365 8 15366 8 15367 8 15368 6639 15369 6640 15370 6641 15371 8 15372 6642 15373 8 15374 8 15375 6643 15376 6644 15377 6645 15378 8 15379 8 15380 8 15381 6646 15382 6647 15383 8 15384 8 15385 8 15386 6648 15387 6649 15388 6650 15389 6651 15390 8 15391 8 15392 6652 15393 8 15394 6653 15395 6654 15396 8 15397 8 15398 6655 15399 8 15400 8 15401 8 15402 6656 15403 6657 15404 8 15405 8 15406 6658 15407 6659 15408 6660 15409 6661 15410 6662 15411 6663 15412 6664 15413 6665 15414 6666 15415 6667 15416 6668 15417 8 15418 6669 15419 6670 15420 6671 15421 6672 15422 6673 15423 6674 15424 6675 15425 6676 15426 6677 15427 6678 15428 8 15429 6679 15430 6375 15431 6680 15432 8 15433 6681 15434 8 15435 6682 15436 6683 15437 8 15438 8 15439 8 15440 6684 15441 8 15442 8 15443 6685 15444 6686 15445 6687 15446 6688 15447 8 15448 8 15449 6689 15450 8 15451 6690 15452 6691 15453 8 15454 8 15455 6692 15456 8 15457 6693 15458 8 15459 6694 15460 6695 15461 6696 15462 6560 15463 6560 15464 6697 15465 6698 15466 6560 15467 6560 15468 6560 15469 6560 15470 8 15471 6699 15472 8 15473 6700 15474 8 15475 8 15476 6560 15477 6560 15478 8 15479 6560 15480 6560 15481 6701 15482 8 15483 8 15484 8 15485 8 15486 8 15487 8 15488 6702 15489 6703 15490 6704 15491 6705 15492 8 15493 6706 15494 8 15495 6707 15496 6708 15497 6709 15498 6710 15499 6711 15500 8 15501 8 15502 8 15503 8 15504 8 15505 6712 15506 8 15507 8 15508 6713 15509 6714 15510 6715 15511 6716 15512 6717 15513 6718 15514 6719 15515 8 15516 6720 15517 8 15518 6721 15519 6722 15520 6723 15521 6724 15522 6725 15523 8 15524 6726 15525 6727 15526 6728 15527 6729 15528 6560 15529 6730 15530 6731 15531 6732 15532 8 15533 6560 15534 6560 15535 6560 15536 8 15537 6733 15538 8 15539 8 15540 8 15541 6734 15542 6735 15543 6736 15544 8 15545 8 15546 6737 15547 6738 15548 8 15549 6739 15550 6740 15551 8 15552 6560 15553 8 15554 6741 15555 8 15556 6560 15557 8 15558 6742 15559 6743 15560 6744 15561 6745 15562 8 15563 8 15564 6746 15565 8 15566 8 15567 6560 15568 6560 15569 6747 15570 6748 15571 8 15572 6560 15573 6560 15574 6560 15575 6749 15576 8 15577 6750 15578 6751 15579 8 15580 6752 15581 8 15582 6560 15583 6375 15584 6375 15585 6375 15586 8 15587 6753 15588 8 15589 6754 15590 8 15591 8 15592 6375 15593 6375 15594 8 15595 6375 15596 6375 15597 6755 15598 8 15599 8 15600 8 15601 8 15602 8 15603 8 15604 6756 15605 6757 15606 6758 15607 6759 15608 8 15609 6760 15610 8 15611 6761 15612 6762 15613 6763 15614 6764 15615 6765 15616 8 15617 8 15618 8 15619 8 15620 8 15621 6766 15622 8 15623 8 15624 6767 15625 6768 15626 6769 15627 6770 15628 6771 15629 6772 15630 6773 15631 8 15632 6774 15633 8 15634 6775 15635 6776 15636 6777 15637 6778 15638 6779 15639 8 15640 6780 15641 6781 15642 6782 15643 6783 15644 6375 15645 6784 15646 6785 15647 6786 15648 8 15649 6375 15650 6375 15651 6375 15652 8 15653 6787 15654 8 15655 8 15656 8 15657 6788 15658 6789 15659 6790 15660 8 15661 8 15662 6791 15663 6792 15664 8 15665 6793 15666 6794 15667 8 15668 6375 15669 8 15670 6795 15671 8 15672 6375 15673 8 15674 6796 15675 6797 15676 6798 15677 6799 15678 8 15679 8 15680 6800 15681 8 15682 8 15683 6375 15684 6375 15685 6801 15686 6802 15687 8 15688 6375 15689 6375 15690 6375 15691 6803 15692 8 15693 6804 15694 6805 15695 8 15696 6806 15697 8 15698 6375 15699 6807 15700 8 15701 6808 15702 8 15703 6809 15704 6810 15705 8 15706 8 15707 8 15708 6811 15709 8 15710 8 15711 6812 15712 6813 15713 6814 15714 6815 15715 8 15716 8 15717 6816 15718 8 15719 6817 15720 6818 15721 8 15722 8 15723 6819 15724 8 15725 6820 15726 8 15727 6821 15728 6822 15729 6823 15730 271 15731 271 15732 6824 15733 6825 15734 6826 15735 6827 15736 6828 15737 6829 15738 6830 15739 6831 15740 6832 15741 6833 15742 8 15743 6834 15744 8 15745 6835 15746 8 15747 8 15748 6836 15749 6837 15750 6838 15751 6839 15752 8 15753 8 15754 6840 15755 6841 15756 6842 15757 6843 15758 6844 15759 8 15760 6845 15761 8 15762 6846 15763 8 15764 8 15765 8 15766 8 15767 6847 15768 8 15769 6848 15770 6849 15771 8 15772 6850 15773 6851 15774 6852 15775 6853 15776 6854 15777 8 15778 6855 15779 6856 15780 8 15781 6857 15782 8 15783 6858 15784 6859 15785 6860 15786 6861 15787 6862 15788 6863 15789 6864 15790 8 15791 8 15792 6865 15793 8 15794 8 15795 6866 15796 6867 15797 6868 15798 8 15799 8 15800 8 15801 8 15802 6869 15803 8 15804 8 15805 8 15806 8 15807 8 15808 6870 15809 6871 15810 6872 15811 6873 15812 6874 15813 6875 15814 6876 15815 6877 15816 6878 15817 6879 15818 6880 15819 6881 15820 6882 15821 6883 15822 6884 15823 6885 15824 6886 15825 6887 15826 6888 15827 6889 15828 6890 15829 6891 15830 6892 15831 6893 15832 6894 15833 6895 15834 6896 15835 6897 15836 6898 15837 6899 15838 6900 15839 6901 15840 6902 15841 6903 15842 6904 15843 6873 15844 6873 15845 6873 15846 6873 15847 6873 15848 6873 15849 6873 15850 6873 15851 6873 15852 6873 15853 6873 15854 6873 15855 6873 15856 6873 15857 6873 15858 6873 15859 6905 15860 6906 15861 8 15862 8 15863 8 15864 8 15865 8 15866 8 15867 6907 15868 8 15869 8 15870 6908 15871 8 15872 6909 15873 6910 15874 6911 15875 6912 15876 6913 15877 8 15878 8 15879 8 15880 8 15881 8 15882 6914 15883 6915 15884 8 15885 8 15886 6916 15887 8 15888 6917 15889 8 15890 8 15891 6918 15892 6919 15893 8 15894 6873 15895 6873 15896 6873 15897 6873 15898 6873 15899 6920 15900 6921 15901 6922 15902 6873 15903 6873 15904 6873 15905 6873 15906 6873 15907 6873 15908 6873 15909 6873 15910 6873 15911 6873 15912 6923 15913 8 15914 8 15915 8 15916 8 15917 8 15918 8 15919 8 15920 8 15921 8 15922 8 15923 8 15924 8 15925 8 15926 8 15927 8 15928 8 15929 8 15930 8 15931 8 15932 8 15933 8 15934 8 15935 8 15936 8 15937 8 15938 8 15939 8 15940 8 15941 8 15942 8 15943 8 15944 8 15945 8 15946 8 15947 8 15948 8 15949 8 15950 8 15951 8 15952 8 15953 8 15954 8 15955 8 15956 8 15957 8 15958 8 15959 8 15960 8 15961 8 15962 8 15963 8 15964 8 15965 8 15966 8 15967 8 15968 8 15969 8 15970 8 15971 8 15972 8 15973 8 15974 8 15975 8 15976 8 15977 8 15978 8 15979 8 15980 8 15981 8 15982 8 15983 8 15984 8 15985 8 15986 8 15987 8 15988 8 15989 8 15990 8 15991 8 15992 8 15993 6924 15994 8 15995 8 15996 8 15997 6925 15998 6926 15999 8 16000 6927 16001 8 16002 8 16003 6928 16004 6929 16005 6930 16006 8 16007 8 16008 6931 16009 6932 16010 6933 16011 6934 16012 8 16013 6935 16014 8 16015 6936 16016 6937 16017 6938 16018 6939 16019 6940 16020 6941 16021 6942 16022 6943 16023 6944 16024 6873 16025 6873 16026 6873 16027 6873 16028 6873 16029 6873 16030 6873 16031 6873 16032 6873 16033 6873 16034 6873 16035 6873 16036 6873 16037 6945 16038 6873 16039 6946 16040 6947 16041 8 16042 6948 16043 6949 16044 8 16045 6950 16046 8 16047 6951 16048 8 16049 8 16050 8 16051 6952 16052 6953 16053 6954 16054 8 16055 6955 16056 8 16057 8 16058 6956 16059 6957 16060 6958 16061 8 16062 8 16063 8 16064 6959 16065 6960 16066 8 16067 8 16068 8 16069 6961 16070 6962 16071 6963 16072 6964 16073 8 16074 8 16075 6965 16076 8 16077 6966 16078 6967 16079 8 16080 8 16081 6968 16082 8 16083 8 16084 8 16085 6969 16086 6970 16087 8 16088 8 16089 6971 16090 6972 16091 6973 16092 6974 16093 6975 16094 6976 16095 6977 16096 6978 16097 6979 16098 6980 16099 6981 16100 8 16101 6982 16102 6983 16103 6984 16104 6985 16105 6986 16106 6987 16107 6988 16108 6989 16109 6990 16110 6991 16111 8 16112 6992 16113 271 16114 6993 16115 8 16116 6994 16117 8 16118 6995 16119 6996 16120 8 16121 8 16122 8 16123 6997 16124 8 16125 8 16126 6998 16127 6999 16128 7000 16129 7001 16130 8 16131 8 16132 7002 16133 8 16134 7003 16135 7004 16136 8 16137 8 16138 7005 16139 8 16140 7006 16141 8 16142 7007 16143 7008 16144 7009 16145 6873 16146 6873 16147 7010 16148 7011 16149 6873 16150 6873 16151 6873 16152 6873 16153 8 16154 7012 16155 8 16156 7013 16157 8 16158 8 16159 6873 16160 6873 16161 8 16162 6873 16163 6873 16164 7014 16165 8 16166 8 16167 8 16168 8 16169 8 16170 8 16171 7015 16172 7016 16173 7017 16174 7018 16175 8 16176 7019 16177 8 16178 7020 16179 7021 16180 7022 16181 7023 16182 7024 16183 8 16184 8 16185 8 16186 8 16187 8 16188 7025 16189 8 16190 8 16191 7026 16192 7027 16193 7028 16194 7029 16195 7030 16196 7031 16197 7032 16198 8 16199 7033 16200 8 16201 7034 16202 7035 16203 7036 16204 7037 16205 7038 16206 8 16207 7039 16208 7040 16209 7041 16210 7042 16211 6873 16212 7043 16213 7044 16214 7045 16215 8 16216 6873 16217 6873 16218 6873 16219 8 16220 7046 16221 8 16222 8 16223 8 16224 7047 16225 7048 16226 7049 16227 8 16228 8 16229 7050 16230 7051 16231 8 16232 7052 16233 7053 16234 8 16235 6873 16236 8 16237 7054 16238 8 16239 6873 16240 8 16241 7055 16242 7056 16243 7057 16244 7058 16245 8 16246 8 16247 7059 16248 8 16249 8 16250 6873 16251 6873 16252 7060 16253 7061 16254 8 16255 6873 16256 6873 16257 6873 16258 7062 16259 8 16260 7063 16261 7064 16262 8 16263 7065 16264 8 16265 6873 16266 271 16267 271 16268 271 16269 8 16270 7066 16271 8 16272 7067 16273 8 16274 8 16275 271 16276 271 16277 8 16278 271 16279 271 16280 7068 16281 8 16282 8 16283 8 16284 8 16285 8 16286 8 16287 7069 16288 7070 16289 7071 16290 7072 16291 8 16292 7073 16293 8 16294 7074 16295 7075 16296 7076 16297 7077 16298 7078 16299 8 16300 8 16301 8 16302 8 16303 8 16304 7079 16305 8 16306 8 16307 7080 16308 7081 16309 7082 16310 7083 16311 7084 16312 7085 16313 7086 16314 8 16315 7087 16316 8 16317 7088 16318 7089 16319 7090 16320 7091 16321 7092 16322 8 16323 7093 16324 7094 16325 7095 16326 7096 16327 271 16328 7097 16329 7098 16330 7099 16331 8 16332 271 16333 271 16334 271 16335 8 16336 7100 16337 8 16338 8 16339 8 16340 7101 16341 7102 16342 7103 16343 8 16344 8 16345 7104 16346 7105 16347 8 16348 7106 16349 7107 16350 8 16351 271 16352 8 16353 7108 16354 8 16355 271 16356 8 16357 7109 16358 7110 16359 7111 16360 7112 16361 8 16362 8 16363 7113 16364 8 16365 8 16366 271 16367 271 16368 7114 16369 7115 16370 8 16371 271 16372 271 16373 271 16374 7116 16375 8 16376 7117 16377 7118 16378 8 16379 7119 16380 8 16381 271 16382 8 16383 8 16384 8 16385 8 16386 7120 16387 7121 16388 8 16389 8 16390 7122 16391 8 16392 7123 16393 8 16394 8 16395 7124 16396 7125 16397 8 16398 189 16399 189 16400 189 16401 189 16402 189 16403 7126 16404 7127 16405 7128 16406 189 16407 189 16408 189 16409 189 16410 189 16411 189 16412 189 16413 189 16414 189 16415 189 16416 7129 16417 8 16418 8 16419 8 16420 8 16421 8 16422 8 16423 8 16424 8 16425 8 16426 8 16427 8 16428 8 16429 8 16430 8 16431 8 16432 8 16433 8 16434 8 16435 8 16436 8 16437 8 16438 8 16439 8 16440 8 16441 8 16442 8 16443 8 16444 8 16445 8 16446 8 16447 8 16448 8 16449 8 16450 8 16451 8 16452 8 16453 8 16454 8 16455 8 16456 8 16457 8 16458 8 16459 8 16460 8 16461 8 16462 8 16463 8 16464 8 16465 8 16466 8 16467 8 16468 8 16469 8 16470 8 16471 8 16472 8 16473 8 16474 8 16475 8 16476 8 16477 8 16478 8 16479 8 16480 8 16481 8 16482 8 16483 8 16484 8 16485 8 16486 8 16487 8 16488 8 16489 8 16490 8 16491 8 16492 8 16493 8 16494 8 16495 8 16496 8 16497 7130 16498 8 16499 8 16500 8 16501 7131 16502 7132 16503 8 16504 7133 16505 8 16506 8 16507 7134 16508 7135 16509 7136 16510 8 16511 8 16512 7137 16513 7138 16514 7139 16515 7140 16516 8 16517 7141 16518 8 16519 7142 16520 7143 16521 7144 16522 7145 16523 7146 16524 7147 16525 7148 16526 7149 16527 7150 16528 189 16529 189 16530 189 16531 189 16532 189 16533 189 16534 189 16535 189 16536 189 16537 189 16538 189 16539 189 16540 189 16541 7151 16542 189 16543 7152 16544 7153 16545 8 16546 7154 16547 7155 16548 8 16549 7156 16550 8 16551 7157 16552 8 16553 8 16554 8 16555 7158 16556 7159 16557 7160 16558 8 16559 7161 16560 8 16561 8 16562 7162 16563 7163 16564 7164 16565 8 16566 8 16567 8 16568 7165 16569 7166 16570 8 16571 8 16572 8 16573 7167 16574 7168 16575 7169 16576 7170 16577 8 16578 8 16579 7171 16580 8 16581 7172 16582 7173 16583 8 16584 8 16585 7174 16586 8 16587 8 16588 8 16589 7175 16590 7176 16591 8 16592 8 16593 7177 16594 7178 16595 7179 16596 7180 16597 7181 16598 7182 16599 7183 16600 7184 16601 7185 16602 7186 16603 7187 16604 8 16605 7188 16606 7189 16607 7190 16608 7191 16609 7192 16610 7193 16611 7194 16612 7195 16613 7196 16614 7197 16615 7198 16616 7199 16617 7200 16618 7201 16619 7202 16620 7203 16621 7204 16622 7205 16623 7206 16624 8 16625 7207 16626 8 16627 7208 16628 8 16629 8 16630 7209 16631 7210 16632 7211 16633 7212 16634 8 16635 8 16636 7213 16637 7214 16638 7215 16639 7216 16640 7217 16641 8 16642 7218 16643 8 16644 7219 16645 8 16646 8 16647 8 16648 8 16649 7220 16650 8 16651 7221 16652 7222 16653 8 16654 7223 16655 7224 16656 7225 16657 7226 16658 7227 16659 8 16660 7228 16661 7229 16662 8 16663 7230 16664 8 16665 7231 16666 7232 16667 7233 16668 7234 16669 7235 16670 7236 16671 7237 16672 8 16673 8 16674 7238 16675 8 16676 8 16677 7239 16678 7240 16679 7241 16680 8 16681 8 16682 8 16683 8 16684 7242 16685 8 16686 8 16687 8 16688 8 16689 8 16690 7243 16691 7244 16692 7245 16693 7246 16694 7247 16695 7248 16696 7249 16697 7250 16698 7251 16699 7252 16700 7253 16701 7254 16702 7255 16703 7256 16704 7257 16705 7258 16706 7259 16707 7260 16708 7261 16709 7262 16710 7263 16711 7264 16712 7265 16713 7266 16714 7267 16715 7268 16716 7269 16717 7270 16718 7271 16719 7272 16720 7273 16721 7274 16722 7275 16723 7276 16724 7277 16725 7246 16726 7246 16727 7246 16728 7246 16729 7246 16730 7246 16731 7246 16732 7246 16733 7246 16734 7246 16735 7246 16736 7246 16737 7246 16738 7246 16739 7246 16740 7246 16741 7278 16742 7279 16743 8 16744 8 16745 8 16746 8 16747 8 16748 8 16749 7280 16750 8 16751 8 16752 7281 16753 8 16754 7282 16755 7283 16756 7284 16757 7285 16758 7286 16759 8 16760 8 16761 8 16762 8 16763 8 16764 7287 16765 7288 16766 8 16767 8 16768 7289 16769 8 16770 7290 16771 8 16772 8 16773 7291 16774 7292 16775 8 16776 7246 16777 7246 16778 7246 16779 7246 16780 7246 16781 7293 16782 7294 16783 7295 16784 7246 16785 7246 16786 7246 16787 7246 16788 7246 16789 7246 16790 7246 16791 7246 16792 7246 16793 7246 16794 7296 16795 8 16796 8 16797 8 16798 8 16799 8 16800 8 16801 8 16802 8 16803 8 16804 8 16805 8 16806 8 16807 8 16808 8 16809 8 16810 8 16811 8 16812 8 16813 8 16814 8 16815 8 16816 8 16817 8 16818 8 16819 8 16820 8 16821 8 16822 8 16823 8 16824 8 16825 8 16826 8 16827 8 16828 8 16829 8 16830 8 16831 8 16832 8 16833 8 16834 8 16835 8 16836 8 16837 8 16838 8 16839 8 16840 8 16841 8 16842 8 16843 8 16844 8 16845 8 16846 8 16847 8 16848 8 16849 8 16850 8 16851 8 16852 8 16853 8 16854 8 16855 8 16856 8 16857 8 16858 8 16859 8 16860 8 16861 8 16862 8 16863 8 16864 8 16865 8 16866 8 16867 8 16868 8 16869 8 16870 8 16871 8 16872 8 16873 8 16874 8 16875 7297 16876 8 16877 8 16878 8 16879 7298 16880 7299 16881 8 16882 7300 16883 8 16884 8 16885 7301 16886 7302 16887 7303 16888 8 16889 8 16890 7304 16891 7305 16892 7306 16893 7307 16894 8 16895 7308 16896 8 16897 7309 16898 7310 16899 7311 16900 7312 16901 7313 16902 7314 16903 7315 16904 7316 16905 7317 16906 7246 16907 7246 16908 7246 16909 7246 16910 7246 16911 7246 16912 7246 16913 7246 16914 7246 16915 7246 16916 7246 16917 7246 16918 7246 16919 7318 16920 7246 16921 7319 16922 7320 16923 8 16924 7321 16925 7322 16926 8 16927 7323 16928 8 16929 7324 16930 8 16931 8 16932 8 16933 7325 16934 7326 16935 7327 16936 8 16937 7328 16938 8 16939 8 16940 7329 16941 7330 16942 7331 16943 8 16944 8 16945 8 16946 7332 16947 7333 16948 8 16949 8 16950 8 16951 7334 16952 7335 16953 7336 16954 7337 16955 8 16956 8 16957 7338 16958 8 16959 7339 16960 7340 16961 8 16962 8 16963 7341 16964 8 16965 8 16966 8 16967 7342 16968 7343 16969 8 16970 8 16971 7344 16972 7345 16973 7346 16974 7347 16975 7348 16976 7349 16977 7350 16978 7351 16979 7352 16980 7353 16981 7354 16982 8 16983 7355 16984 7356 16985 7357 16986 7358 16987 7359 16988 7360 16989 7361 16990 7362 16991 7363 16992 7364 16993 8 16994 8 16995 7365 16996 8 16997 7366 16998 8 16999 7367 17000 7368 17001 8 17002 8 17003 8 17004 7369 17005 8 17006 8 17007 7370 17008 7371 17009 7372 17010 7373 17011 8 17012 8 17013 7374 17014 8 17015 7375 17016 7376 17017 8 17018 8 17019 7377 17020 8 17021 7378 17022 8 17023 7379 17024 7380 17025 7381 17026 7246 17027 7246 17028 7382 17029 7383 17030 7384 17031 7385 17032 7386 17033 7387 17034 7388 17035 7389 17036 7390 17037 7391 17038 8 17039 7392 17040 8 17041 7393 17042 8 17043 8 17044 7394 17045 7395 17046 7396 17047 7397 17048 8 17049 8 17050 7398 17051 7399 17052 7400 17053 7401 17054 7402 17055 8 17056 7403 17057 8 17058 7404 17059 8 17060 8 17061 8 17062 8 17063 7405 17064 8 17065 7406 17066 7407 17067 8 17068 7408 17069 7409 17070 7410 17071 7411 17072 7412 17073 8 17074 7413 17075 7414 17076 8 17077 7415 17078 8 17079 7416 17080 7417 17081 7418 17082 7419 17083 7420 17084 7421 17085 7422 17086 8 17087 8 17088 7423 17089 8 17090 8 17091 7424 17092 7425 17093 7426 17094 8 17095 8 17096 8 17097 8 17098 7427 17099 8 17100 8 17101 8 17102 8 17103 8 17104 7428 17105 7429 17106 7430 17107 7431 17108 7432 17109 7433 17110 7434 17111 7435 17112 7436 17113 7437 17114 7438 17115 7439 17116 7440 17117 7441 17118 7442 17119 7443 17120 7444 17121 7445 17122 7446 17123 7447 17124 7448 17125 7449 17126 7450 17127 7451 17128 7452 17129 7453 17130 7454 17131 7455 17132 7456 17133 7457 17134 7458 17135 7459 17136 7460 17137 7461 17138 7462 17139 7431 17140 7431 17141 7431 17142 7431 17143 7431 17144 7431 17145 7431 17146 7431 17147 7431 17148 7431 17149 7431 17150 7431 17151 7431 17152 7431 17153 7431 17154 7431 17155 7463 17156 7464 17157 8 17158 8 17159 8 17160 8 17161 8 17162 8 17163 7465 17164 8 17165 8 17166 7466 17167 8 17168 7467 17169 7468 17170 7469 17171 7470 17172 7471 17173 8 17174 8 17175 8 17176 8 17177 8 17178 7472 17179 7473 17180 8 17181 8 17182 7474 17183 8 17184 7475 17185 8 17186 8 17187 7476 17188 7477 17189 8 17190 7431 17191 7431 17192 7431 17193 7431 17194 7431 17195 7478 17196 7479 17197 7480 17198 7431 17199 7431 17200 7431 17201 7431 17202 7431 17203 7431 17204 7431 17205 7431 17206 7431 17207 7431 17208 7481 17209 8 17210 8 17211 8 17212 8 17213 8 17214 8 17215 8 17216 8 17217 8 17218 8 17219 8 17220 8 17221 8 17222 8 17223 8 17224 8 17225 8 17226 8 17227 8 17228 8 17229 8 17230 8 17231 8 17232 8 17233 8 17234 8 17235 8 17236 8 17237 8 17238 8 17239 8 17240 8 17241 8 17242 8 17243 8 17244 8 17245 8 17246 8 17247 8 17248 8 17249 8 17250 8 17251 8 17252 8 17253 8 17254 8 17255 8 17256 8 17257 8 17258 8 17259 8 17260 8 17261 8 17262 8 17263 8 17264 8 17265 8 17266 8 17267 8 17268 8 17269 8 17270 8 17271 8 17272 8 17273 8 17274 8 17275 8 17276 8 17277 8 17278 8 17279 8 17280 8 17281 8 17282 8 17283 8 17284 8 17285 8 17286 8 17287 8 17288 8 17289 7482 17290 8 17291 8 17292 8 17293 7483 17294 7484 17295 8 17296 7485 17297 8 17298 8 17299 7486 17300 7487 17301 7488 17302 8 17303 8 17304 7489 17305 7490 17306 7491 17307 7492 17308 8 17309 7493 17310 8 17311 7494 17312 7495 17313 7496 17314 7497 17315 7498 17316 7499 17317 7500 17318 7501 17319 7502 17320 7431 17321 7431 17322 7431 17323 7431 17324 7431 17325 7431 17326 7431 17327 7431 17328 7431 17329 7431 17330 7431 17331 7431 17332 7431 17333 7503 17334 7431 17335 7504 17336 7505 17337 8 17338 7506 17339 7507 17340 8 17341 7508 17342 8 17343 7509 17344 8 17345 8 17346 8 17347 7510 17348 7511 17349 7512 17350 8 17351 7513 17352 8 17353 8 17354 7514 17355 7515 17356 7516 17357 8 17358 8 17359 8 17360 7517 17361 7518 17362 8 17363 8 17364 8 17365 7519 17366 7520 17367 7521 17368 7522 17369 8 17370 8 17371 7523 17372 8 17373 7524 17374 7525 17375 8 17376 8 17377 7526 17378 8 17379 8 17380 8 17381 7527 17382 7528 17383 8 17384 8 17385 7529 17386 7530 17387 7531 17388 7532 17389 7533 17390 7534 17391 7535 17392 7536 17393 7537 17394 7538 17395 7539 17396 8 17397 7540 17398 7541 17399 7542 17400 7543 17401 7544 17402 7545 17403 7546 17404 7547 17405 7548 17406 7549 17407 8 17408 7550 17409 7246 17410 7551 17411 8 17412 7552 17413 8 17414 7553 17415 7554 17416 8 17417 8 17418 8 17419 7555 17420 8 17421 8 17422 7556 17423 7557 17424 7558 17425 7559 17426 8 17427 8 17428 7560 17429 8 17430 7561 17431 7562 17432 8 17433 8 17434 7563 17435 8 17436 7564 17437 8 17438 7565 17439 7566 17440 7567 17441 7431 17442 7431 17443 7568 17444 7569 17445 7431 17446 7431 17447 7431 17448 7431 17449 8 17450 7570 17451 8 17452 7571 17453 8 17454 8 17455 7431 17456 7431 17457 8 17458 7431 17459 7431 17460 7572 17461 8 17462 8 17463 8 17464 8 17465 8 17466 8 17467 7573 17468 7574 17469 7575 17470 7576 17471 8 17472 7577 17473 8 17474 7578 17475 7579 17476 7580 17477 7581 17478 7582 17479 8 17480 8 17481 8 17482 8 17483 8 17484 7583 17485 8 17486 8 17487 7584 17488 7585 17489 7586 17490 7587 17491 7588 17492 7589 17493 7590 17494 8 17495 7591 17496 8 17497 7592 17498 7593 17499 7594 17500 7595 17501 7596 17502 8 17503 7597 17504 7598 17505 7599 17506 7600 17507 7431 17508 7601 17509 7602 17510 7603 17511 8 17512 7431 17513 7431 17514 7431 17515 8 17516 7604 17517 8 17518 8 17519 8 17520 7605 17521 7606 17522 7607 17523 8 17524 8 17525 7608 17526 7609 17527 8 17528 7610 17529 7611 17530 8 17531 7431 17532 8 17533 7612 17534 8 17535 7431 17536 8 17537 7613 17538 7614 17539 7615 17540 7616 17541 8 17542 8 17543 7617 17544 8 17545 8 17546 7431 17547 7431 17548 7618 17549 7619 17550 8 17551 7431 17552 7431 17553 7431 17554 7620 17555 8 17556 7621 17557 7622 17558 8 17559 7623 17560 8 17561 7431 17562 7246 17563 7246 17564 7246 17565 8 17566 7624 17567 8 17568 7625 17569 8 17570 8 17571 7246 17572 7246 17573 8 17574 7246 17575 7246 17576 7626 17577 8 17578 8 17579 8 17580 8 17581 8 17582 8 17583 7627 17584 7628 17585 7629 17586 7630 17587 8 17588 7631 17589 8 17590 7632 17591 7633 17592 7634 17593 7635 17594 7636 17595 8 17596 8 17597 8 17598 8 17599 8 17600 7637 17601 8 17602 8 17603 7638 17604 7639 17605 7640 17606 7641 17607 7642 17608 7643 17609 7644 17610 8 17611 7645 17612 8 17613 7646 17614 7647 17615 7648 17616 7649 17617 7650 17618 8 17619 7651 17620 7652 17621 7653 17622 7654 17623 7246 17624 7655 17625 7656 17626 7657 17627 8 17628 7246 17629 7246 17630 7246 17631 8 17632 7658 17633 8 17634 8 17635 8 17636 7659 17637 7660 17638 7661 17639 8 17640 8 17641 7662 17642 7663 17643 8 17644 7664 17645 7665 17646 8 17647 7246 17648 8 17649 7666 17650 8 17651 7246 17652 8 17653 7667 17654 7668 17655 7669 17656 7670 17657 8 17658 8 17659 7671 17660 8 17661 8 17662 7246 17663 7246 17664 7672 17665 7673 17666 8 17667 7246 17668 7246 17669 7246 17670 7674 17671 8 17672 7675 17673 7676 17674 8 17675 7677 17676 8 17677 7246 17678 7678 17679 8 17680 7679 17681 7680 17682 7681 17683 7682 17684 7683 17685 7684 17686 7685 17687 7686 17688 7687 17689 8 17690 7688 17691 7689 17692 7690 17693 7691 17694 7692 17695 7693 17696 7694 17697 7695 17698 8 17699 7696 17700 7697 17701 8 17702 7698 17703 8 17704 7699 17705 7700 17706 8 17707 7701 17708 8 17709 8 17710 7702 17711 7703 17712 7704 17713 8 17714 7705 17715 7706 17716 7707 17717 7708 17718 7709 17719 8 17720 7710 17721 7711 17722 7712 17723 7713 17724 7714 17725 7715 17726 7716 17727 7717 17728 7718 17729 7719 17730 7720 17731 7721 17732 7722 17733 8 17734 7723 17735 7724 17736 7725 17737 7726 17738 7727 17739 7728 17740 7729 17741 7730 17742 7731 17743 7732 17744 8 17745 7733 17746 8 17747 7734 17748 8 17749 7735 17750 8 17751 7736 17752 7737 17753 8 17754 8 17755 7738 17756 8 17757 8 17758 8 17759 8 17760 7739 17761 7740 17762 8 17763 7741 17764 7742 17765 7743 17766 7744 17767 7745 17768 8 17769 8 17770 8 17771 8 17772 8 17773 7746 17774 8 17775 8 17776 7747 17777 7748 17778 7749 17779 7750 17780 7751 17781 7752 17782 7753 17783 7754 17784 7755 17785 7756 17786 8 17787 8 17788 7757 17789 8 17790 7758 17791 7759 17792 7760 17793 7761 17794 7762 17795 7763 17796 8 17797 7764 17798 8 17799 8 17800 8 17801 8 17802 7765 17803 7766 17804 8 17805 7767 17806 7768 17807 8 17808 7769 17809 7770 17810 7771 17811 7772 17812 7773 17813 7774 17814 7775 17815 7776 17816 7777 17817 7778 17818 7779 17819 7780 17820 7781 17821 8 17822 7782 17823 7783 17824 7784 17825 7785 17826 7786 17827 7787 17828 7788 17829 7789 17830 7790 17831 7791 17832 8 17833 7792 17834 8 17835 7793 17836 8 17837 7794 17838 8 17839 7795 17840 7796 17841 8 17842 8 17843 7797 17844 8 17845 8 17846 8 17847 8 17848 7798 17849 7799 17850 8 17851 7800 17852 8 17853 7801 17854 7802 17855 7803 17856 7804 17857 7805 17858 7806 17859 7807 17860 7808 17861 7809 17862 7810 17863 7811 17864 8 17865 8 17866 7812 17867 7813 17868 7814 17869 7815 17870 7816 17871 7817 17872 7818 17873 8 17874 8 17875 7819 17876 7820 17877 8 17878 7821 17879 7822 17880 7823 17881 8 17882 7824 17883 7825 17884 7826 17885 7827 17886 8 17887 8 17888 7828 17889 8 17890 7829 17891 7830 17892 7831 17893 7832 17894 7833 17895 7834 17896 7835 17897 7836 17898 7837 17899 7838 17900 7839 17901 7840 17902 7841 17903 7842 17904 8 17905 7843 17906 7844 17907 7845 17908 7846 17909 7847 17910 7848 17911 7849 17912 7850 17913 7851 17914 7852 17915 8 17916 7853 17917 8 17918 7854 17919 8 17920 7855 17921 8 17922 7856 17923 7857 17924 8 17925 8 17926 7858 17927 8 17928 8 17929 8 17930 8 17931 7859 17932 7860 17933 8 17934 7861 17935 7862 17936 7863 17937 7864 17938 7865 17939 8 17940 7866 17941 7867 17942 7868 17943 7869 17944 7870 17945 7871 17946 7872 17947 7873 17948 7874 17949 8 17950 7875 17951 7876 17952 7877 17953 7878 17954 7879 17955 7880 17956 7881 17957 7882 17958 8 17959 7883 17960 7884 17961 8 17962 7885 17963 8 17964 7886 17965 7887 17966 8 17967 7888 17968 8 17969 8 17970 7889 17971 7890 17972 8 17973 7891 17974 7892 17975 7893 17976 8 17977 7894 17978 7895 17979 7896 17980 7897 17981 8 17982 8 17983 7898 17984 7899 17985 8 17986 7900 17987 8 17988 7901 17989 7902 17990 7903 17991 8 17992 7904 17993 7905 17994 8 17995 7906 17996 8 17997 7907 17998 8 17999 8 18000 8 18001 7908 18002 8 18003 7909 18004 7910 18005 7911 18006 8 18007 7912 18008 8 18009 8 18010 7913 18011 7914 18012 7915 18013 8 18014 8 18015 7916 18016 7917 18017 7918 18018 8 18019 8 18020 8 18021 7919 18022 7920 18023 8 18024 8 18025 8 18026 7921 18027 7922 18028 7923 18029 7924 18030 8 18031 8 18032 7925 18033 8 18034 7926 18035 7927 18036 8 18037 8 18038 7928 18039 8 18040 8 18041 8 18042 7929 18043 7930 18044 8 18045 7931 18046 7932 18047 7933 18048 7934 18049 7935 18050 7936 18051 8 18052 7937 18053 7938 18054 8 18055 7939 18056 8 18057 8 18058 8 18059 8 18060 8 18061 7940 18062 8 18063 8 18064 8 18065 7941 18066 7942 18067 8 18068 7943 18069 7944 18070 7945 18071 8 18072 7946 18073 7947 18074 7948 18075 7949 18076 8 18077 7950 18078 7951 18079 7952 18080 7953 18081 7954 18082 8 18083 7955 18084 7956 18085 7957 18086 8 18087 7958 18088 7959 18089 7960 18090 7961 18091 7962 18092 7963 18093 7964 18094 7965 18095 7966 18096 7967 18097 7968 18098 7969 18099 7970 18100 8 18101 7971 18102 7972 18103 7973 18104 7974 18105 7975 18106 7976 18107 7977 18108 7978 18109 7979 18110 7980 18111 8 18112 7981 18113 8 18114 7982 18115 8 18116 7983 18117 8 18118 7984 18119 7985 18120 8 18121 8 18122 7986 18123 8 18124 8 18125 8 18126 8 18127 7987 18128 7988 18129 8 18130 7989 18131 7990 18132 7991 18133 7992 18134 7993 18135 8 18136 7994 18137 7995 18138 7996 18139 7997 18140 7998 18141 7999 18142 8000 18143 8001 18144 8002 18145 8 18146 8003 18147 8004 18148 8005 18149 8006 18150 8007 18151 8008 18152 8009 18153 8010 18154 8 18155 8011 18156 8012 18157 8 18158 8013 18159 8 18160 8014 18161 8015 18162 8 18163 8016 18164 8 18165 8 18166 8017 18167 8018 18168 8 18169 8019 18170 8020 18171 8021 18172 8 18173 8022 18174 8023 18175 8024 18176 8025 18177 8 18178 8 18179 8026 18180 8027 18181 8 18182 8028 18183 8 18184 8029 18185 8030 18186 8031 18187 8032 18188 8 18189 8033 18190 8034 18191 8035 18192 8036 18193 8037 18194 8038 18195 8 18196 8039 18197 8040 18198 8 18199 8041 18200 8042 18201 8043 18202 8044 18203 8045 18204 8046 18205 8 18206 8047 18207 8048 18208 8049 18209 8050 18210 8 18211 8051 18212 8052 18213 8053 18214 8 18215 8054 18216 8 18217 8055 18218 8056 18219 8057 18220 8058 18221 8059 18222 8 18223 8060 18224 8061 18225 8062 18226 8 18227 8063 18228 8064 18229 8065 18230 8066 18231 8067 18232 8068 18233 8069 18234 8070 18235 8071 18236 8072 18237 8073 18238 8074 18239 8075 18240 8 18241 8076 18242 8077 18243 8078 18244 8079 18245 8080 18246 8081 18247 8082 18248 8083 18249 8084 18250 8085 18251 8 18252 8086 18253 8 18254 8087 18255 8 18256 8088 18257 8 18258 8089 18259 8090 18260 8 18261 8 18262 8091 18263 8 18264 8 18265 8 18266 8 18267 8092 18268 8093 18269 8 18270 8094 18271 8095 18272 8096 18273 8097 18274 8098 18275 8 18276 8099 18277 8100 18278 8101 18279 8102 18280 8103 18281 8104 18282 8105 18283 8106 18284 8107 18285 8 18286 8108 18287 8109 18288 8110 18289 8111 18290 8112 18291 8113 18292 8114 18293 8115 18294 8 18295 8116 18296 8117 18297 8 18298 8118 18299 8 18300 8119 18301 8120 18302 8 18303 8121 18304 8 18305 8 18306 8122 18307 8123 18308 8 18309 8124 18310 8125 18311 8126 18312 8 18313 8127 18314 8128 18315 8129 18316 8130 18317 8 18318 8 18319 8131 18320 8132 18321 8 18322 8133 18323 8 18324 8134 18325 8135 18326 8136 18327 8137 18328 8 18329 8138 18330 8139 18331 8140 18332 8141 18333 8142 18334 8143 18335 8 18336 8144 18337 8145 18338 8 18339 8146 18340 8147 18341 8148 18342 8149 18343 8150 18344 8151 18345 8 18346 8152 18347 8153 18348 8154 18349 8155 18350 8 18351 8156 18352 8157 18353 8158 18354 8 18355 8159 18356 8 18357 8160 18358 8161 18359 8162 18360 8 18361 8163 18362 8164 18363 8 18364 8 18365 8165 18366 8 18367 8 18368 8166 18369 8167 18370 8 18371 8 18372 8168 18373 8169 18374 8170 18375 8171 18376 8172 18377 8173 18378 8174 18379 8175 18380 8176 18381 8177 18382 8178 18383 8 18384 8179 18385 8 18386 8180 18387 8 18388 8 18389 8181 18390 8182 18391 8183 18392 8184 18393 8 18394 8 18395 8185 18396 8186 18397 8187 18398 8188 18399 8189 18400 8 18401 8190 18402 8 18403 8191 18404 8 18405 8 18406 8 18407 8 18408 8192 18409 8 18410 8193 18411 8194 18412 8 18413 8195 18414 8196 18415 8197 18416 8198 18417 8199 18418 8 18419 8200 18420 8201 18421 8 18422 8202 18423 8 18424 8203 18425 8204 18426 8205 18427 8206 18428 8207 18429 8208 18430 8209 18431 8 18432 8 18433 8210 18434 8 18435 8 18436 8211 18437 8212 18438 8213 18439 8 18440 8 18441 8 18442 8 18443 8214 18444 8 18445 8 18446 8 18447 8 18448 8 18449 8215 18450 8216 18451 8217 18452 8211 18453 8218 18454 8219 18455 8220 18456 8221 18457 8222 18458 8223 18459 8224 18460 8225 18461 8226 18462 8227 18463 8228 18464 8229 18465 8230 18466 8231 18467 8232 18468 8233 18469 8234 18470 8235 18471 8236 18472 8237 18473 8238 18474 8239 18475 8240 18476 8241 18477 8242 18478 8243 18479 8244 18480 8245 18481 8246 18482 8247 18483 8248 18484 8211 18485 8211 18486 8211 18487 8211 18488 8211 18489 8211 18490 8211 18491 8211 18492 8211 18493 8211 18494 8211 18495 8211 18496 8211 18497 8211 18498 8211 18499 8211 18500 8249 18501 8250 18502 8 18503 8 18504 8 18505 8 18506 8 18507 8 18508 8251 18509 8 18510 8 18511 8252 18512 8 18513 8253 18514 8254 18515 8255 18516 8256 18517 8257 18518 8 18519 8 18520 8 18521 8 18522 8 18523 8258 18524 8259 18525 8 18526 8 18527 8260 18528 8 18529 8261 18530 8 18531 8 18532 8262 18533 8263 18534 8 18535 8211 18536 8211 18537 8211 18538 8211 18539 8211 18540 8264 18541 8265 18542 8266 18543 8211 18544 8211 18545 8211 18546 8211 18547 8211 18548 8211 18549 8211 18550 8211 18551 8211 18552 8211 18553 8267 18554 8 18555 8 18556 8 18557 8 18558 8 18559 8 18560 8 18561 8 18562 8 18563 8 18564 8 18565 8 18566 8 18567 8 18568 8 18569 8 18570 8 18571 8 18572 8 18573 8 18574 8 18575 8 18576 8 18577 8 18578 8 18579 8 18580 8 18581 8 18582 8 18583 8 18584 8 18585 8 18586 8 18587 8 18588 8 18589 8 18590 8 18591 8 18592 8 18593 8 18594 8 18595 8 18596 8 18597 8 18598 8 18599 8 18600 8 18601 8 18602 8 18603 8 18604 8 18605 8 18606 8 18607 8 18608 8 18609 8 18610 8 18611 8 18612 8 18613 8 18614 8 18615 8 18616 8 18617 8 18618 8 18619 8 18620 8 18621 8 18622 8 18623 8 18624 8 18625 8 18626 8 18627 8 18628 8 18629 8 18630 8 18631 8 18632 8 18633 8 18634 8268 18635 8 18636 8 18637 8 18638 8269 18639 8270 18640 8 18641 8271 18642 8 18643 8 18644 8272 18645 8273 18646 8274 18647 8 18648 8 18649 8275 18650 8276 18651 8277 18652 8278 18653 8 18654 8279 18655 8 18656 8280 18657 8281 18658 8282 18659 8283 18660 8284 18661 8285 18662 8286 18663 8287 18664 8288 18665 8211 18666 8211 18667 8211 18668 8211 18669 8211 18670 8211 18671 8211 18672 8211 18673 8211 18674 8211 18675 8211 18676 8211 18677 8211 18678 8289 18679 8211 18680 8290 18681 8291 18682 8 18683 8292 18684 8293 18685 8 18686 8294 18687 8 18688 8295 18689 8 18690 8 18691 8 18692 8296 18693 8297 18694 8298 18695 8 18696 8299 18697 8 18698 8 18699 8300 18700 8301 18701 8302 18702 8 18703 8 18704 8 18705 8303 18706 8304 18707 8 18708 8 18709 8 18710 8305 18711 8306 18712 8307 18713 8308 18714 8 18715 8 18716 8309 18717 8 18718 8310 18719 8311 18720 8 18721 8 18722 8312 18723 8 18724 8 18725 8 18726 8313 18727 8314 18728 8 18729 8 18730 8315 18731 8316 18732 8317 18733 8318 18734 8319 18735 8320 18736 8321 18737 8322 18738 8323 18739 8324 18740 8325 18741 8 18742 8326 18743 8327 18744 8328 18745 8329 18746 8330 18747 8331 18748 8332 18749 8333 18750 8334 18751 8335 18752 8 18753 8336 18754 8 18755 8337 18756 8 18757 8338 18758 8339 18759 8 18760 8 18761 8 18762 8340 18763 8 18764 8 18765 8341 18766 8342 18767 8343 18768 8344 18769 8 18770 8 18771 8345 18772 8 18773 8346 18774 8347 18775 8 18776 8 18777 8348 18778 8 18779 8349 18780 8 18781 8350 18782 8351 18783 8352 18784 8211 18785 8211 18786 8353 18787 8354 18788 8355 18789 8356 18790 8357 18791 8358 18792 8359 18793 8360 18794 8361 18795 8362 18796 8 18797 8363 18798 8 18799 8364 18800 8 18801 8 18802 8365 18803 8366 18804 8367 18805 8368 18806 8 18807 8 18808 8369 18809 8370 18810 8371 18811 8372 18812 8373 18813 8 18814 8374 18815 8 18816 8375 18817 8 18818 8 18819 8 18820 8 18821 8376 18822 8 18823 8377 18824 8378 18825 8 18826 8379 18827 8380 18828 8381 18829 8382 18830 8383 18831 8 18832 8384 18833 8385 18834 8 18835 8386 18836 8 18837 8387 18838 8388 18839 8389 18840 8390 18841 8391 18842 8392 18843 8393 18844 8 18845 8 18846 8394 18847 8 18848 8 18849 8395 18850 8396 18851 8397 18852 8 18853 8 18854 8 18855 8 18856 8398 18857 8 18858 8 18859 8 18860 8 18861 8 18862 8399 18863 8400 18864 8401 18865 8402 18866 8403 18867 8404 18868 8405 18869 8406 18870 8407 18871 8408 18872 8409 18873 8410 18874 8411 18875 8412 18876 8413 18877 8414 18878 8415 18879 8416 18880 8417 18881 8418 18882 8419 18883 8420 18884 8421 18885 8422 18886 8423 18887 8424 18888 8425 18889 8426 18890 8427 18891 8428 18892 8429 18893 8430 18894 8431 18895 8432 18896 8433 18897 8402 18898 8402 18899 8402 18900 8402 18901 8402 18902 8402 18903 8402 18904 8402 18905 8402 18906 8402 18907 8402 18908 8402 18909 8402 18910 8402 18911 8402 18912 8402 18913 8434 18914 8435 18915 8 18916 8 18917 8 18918 8 18919 8 18920 8 18921 8436 18922 8 18923 8 18924 8437 18925 8 18926 8438 18927 8439 18928 8440 18929 8441 18930 8442 18931 8 18932 8 18933 8 18934 8 18935 8 18936 8443 18937 8444 18938 8 18939 8 18940 8445 18941 8 18942 8446 18943 8 18944 8 18945 8447 18946 8448 18947 8 18948 8402 18949 8402 18950 8402 18951 8402 18952 8402 18953 8449 18954 8450 18955 8451 18956 8402 18957 8402 18958 8402 18959 8402 18960 8402 18961 8402 18962 8402 18963 8402 18964 8402 18965 8402 18966 8452 18967 8 18968 8 18969 8 18970 8 18971 8 18972 8 18973 8 18974 8 18975 8 18976 8 18977 8 18978 8 18979 8 18980 8 18981 8 18982 8 18983 8 18984 8 18985 8 18986 8 18987 8 18988 8 18989 8 18990 8 18991 8 18992 8 18993 8 18994 8 18995 8 18996 8 18997 8 18998 8 18999 8 19000 8 19001 8 19002 8 19003 8 19004 8 19005 8 19006 8 19007 8 19008 8 19009 8 19010 8 19011 8 19012 8 19013 8 19014 8 19015 8 19016 8 19017 8 19018 8 19019 8 19020 8 19021 8 19022 8 19023 8 19024 8 19025 8 19026 8 19027 8 19028 8 19029 8 19030 8 19031 8 19032 8 19033 8 19034 8 19035 8 19036 8 19037 8 19038 8 19039 8 19040 8 19041 8 19042 8 19043 8 19044 8 19045 8 19046 8 19047 8453 19048 8 19049 8 19050 8 19051 8454 19052 8455 19053 8 19054 8456 19055 8 19056 8 19057 8457 19058 8458 19059 8459 19060 8 19061 8 19062 8460 19063 8461 19064 8462 19065 8463 19066 8 19067 8464 19068 8 19069 8465 19070 8466 19071 8467 19072 8468 19073 8469 19074 8470 19075 8471 19076 8472 19077 8473 19078 8402 19079 8402 19080 8402 19081 8402 19082 8402 19083 8402 19084 8402 19085 8402 19086 8402 19087 8402 19088 8402 19089 8402 19090 8402 19091 8474 19092 8402 19093 8475 19094 8476 19095 8 19096 8477 19097 8478 19098 8 19099 8479 19100 8 19101 8480 19102 8 19103 8 19104 8 19105 8481 19106 8482 19107 8483 19108 8 19109 8484 19110 8 19111 8 19112 8485 19113 8486 19114 8487 19115 8 19116 8 19117 8 19118 8488 19119 8489 19120 8 19121 8 19122 8 19123 8490 19124 8491 19125 8492 19126 8493 19127 8 19128 8 19129 8494 19130 8 19131 8495 19132 8496 19133 8 19134 8 19135 8497 19136 8 19137 8 19138 8 19139 8498 19140 8499 19141 8 19142 8 19143 8500 19144 8501 19145 8502 19146 8503 19147 8504 19148 8505 19149 8506 19150 8507 19151 8508 19152 8509 19153 8510 19154 8 19155 8511 19156 8512 19157 8513 19158 8514 19159 8515 19160 8516 19161 8517 19162 8518 19163 8519 19164 8520 19165 8 19166 8521 19167 8211 19168 8522 19169 8 19170 8523 19171 8 19172 8524 19173 8525 19174 8 19175 8 19176 8 19177 8526 19178 8 19179 8 19180 8527 19181 8528 19182 8529 19183 8530 19184 8 19185 8 19186 8531 19187 8 19188 8532 19189 8533 19190 8 19191 8 19192 8534 19193 8 19194 8535 19195 8 19196 8536 19197 8537 19198 8538 19199 8402 19200 8402 19201 8539 19202 8540 19203 8402 19204 8402 19205 8402 19206 8402 19207 8 19208 8541 19209 8 19210 8542 19211 8 19212 8 19213 8402 19214 8402 19215 8 19216 8402 19217 8402 19218 8543 19219 8 19220 8 19221 8 19222 8 19223 8 19224 8 19225 8544 19226 8545 19227 8546 19228 8547 19229 8 19230 8548 19231 8 19232 8549 19233 8550 19234 8551 19235 8552 19236 8553 19237 8 19238 8 19239 8 19240 8 19241 8 19242 8554 19243 8 19244 8 19245 8555 19246 8556 19247 8557 19248 8558 19249 8559 19250 8560 19251 8561 19252 8 19253 8562 19254 8 19255 8563 19256 8564 19257 8565 19258 8566 19259 8567 19260 8 19261 8568 19262 8569 19263 8570 19264 8571 19265 8402 19266 8572 19267 8573 19268 8574 19269 8 19270 8402 19271 8402 19272 8402 19273 8 19274 8575 19275 8 19276 8 19277 8 19278 8576 19279 8577 19280 8578 19281 8 19282 8 19283 8579 19284 8580 19285 8 19286 8581 19287 8582 19288 8 19289 8402 19290 8 19291 8583 19292 8 19293 8402 19294 8 19295 8584 19296 8585 19297 8586 19298 8587 19299 8 19300 8 19301 8588 19302 8 19303 8 19304 8402 19305 8402 19306 8589 19307 8590 19308 8 19309 8402 19310 8402 19311 8402 19312 8591 19313 8 19314 8592 19315 8593 19316 8 19317 8594 19318 8 19319 8402 19320 8211 19321 8211 19322 8211 19323 8 19324 8595 19325 8 19326 8596 19327 8 19328 8 19329 8211 19330 8211 19331 8 19332 8211 19333 8211 19334 8597 19335 8 19336 8 19337 8 19338 8 19339 8 19340 8 19341 8598 19342 8599 19343 8600 19344 8601 19345 8 19346 8602 19347 8 19348 8603 19349 8604 19350 8605 19351 8606 19352 8607 19353 8 19354 8 19355 8 19356 8 19357 8 19358 8608 19359 8 19360 8 19361 8609 19362 8610 19363 8611 19364 8612 19365 8613 19366 8614 19367 8615 19368 8 19369 8616 19370 8 19371 8617 19372 8618 19373 8619 19374 8620 19375 8621 19376 8 19377 8622 19378 8623 19379 8624 19380 8625 19381 8211 19382 8626 19383 8627 19384 8628 19385 8 19386 8211 19387 8211 19388 8211 19389 8 19390 8629 19391 8 19392 8 19393 8 19394 8630 19395 8631 19396 8632 19397 8 19398 8 19399 8633 19400 8634 19401 8 19402 8635 19403 8636 19404 8 19405 8211 19406 8 19407 8637 19408 8 19409 8211 19410 8 19411 8638 19412 8639 19413 8640 19414 8641 19415 8 19416 8 19417 8642 19418 8 19419 8 19420 8211 19421 8211 19422 8643 19423 8644 19424 8 19425 8211 19426 8211 19427 8211 19428 8645 19429 8 19430 8646 19431 8647 19432 8 19433 8648 19434 8 19435 8211 19436 85 19437 8649 19438 8 19439 8650 19440 8 19441 8651 19442 8 19443 8 19444 8652 19445 8 19446 8653 19447 8 19448 8 19449 8 19450 8654 19451 8655 19452 8 19453 8656 19454 8657 19455 8658 19456 8659 19457 8660 19458 8661 19459 8662 19460 8663 19461 8664 19462 8665 19463 8666 19464 8 19465 8667 19466 8 19467 8668 19468 8 19469 8 19470 8669 19471 8670 19472 8671 19473 8672 19474 8 19475 8 19476 8673 19477 8674 19478 8675 19479 8676 19480 8677 19481 8 19482 8678 19483 8 19484 8679 19485 8 19486 8 19487 8 19488 8 19489 8680 19490 8 19491 8681 19492 8682 19493 8 19494 8683 19495 8684 19496 8685 19497 8686 19498 8687 19499 8 19500 8688 19501 8689 19502 8 19503 8690 19504 8 19505 8691 19506 8692 19507 8693 19508 8694 19509 8695 19510 8696 19511 8697 19512 8 19513 8 19514 8698 19515 8 19516 8 19517 8699 19518 8700 19519 8701 19520 8 19521 8 19522 8 19523 8 19524 8702 19525 8 19526 8 19527 8 19528 8 19529 8 19530 8703 19531 8704 19532 8705 19533 8706 19534 8707 19535 8708 19536 8709 19537 8710 19538 8711 19539 8712 19540 8713 19541 8714 19542 8715 19543 8716 19544 8717 19545 8718 19546 8719 19547 8720 19548 8721 19549 8722 19550 8723 19551 8724 19552 8725 19553 8726 19554 8727 19555 8728 19556 8729 19557 8730 19558 8731 19559 8732 19560 8733 19561 8734 19562 8735 19563 8736 19564 8737 19565 8706 19566 8706 19567 8706 19568 8706 19569 8706 19570 8706 19571 8706 19572 8706 19573 8706 19574 8706 19575 8706 19576 8706 19577 8706 19578 8706 19579 8706 19580 8706 19581 8738 19582 8739 19583 8 19584 8 19585 8 19586 8 19587 8 19588 8 19589 8740 19590 8 19591 8 19592 8741 19593 8 19594 8742 19595 8743 19596 8744 19597 8745 19598 8746 19599 8 19600 8 19601 8 19602 8 19603 8 19604 8747 19605 8748 19606 8 19607 8 19608 8749 19609 8 19610 8750 19611 8 19612 8 19613 8751 19614 8752 19615 8 19616 8706 19617 8706 19618 8706 19619 8706 19620 8706 19621 8753 19622 8754 19623 8755 19624 8706 19625 8706 19626 8706 19627 8706 19628 8706 19629 8706 19630 8706 19631 8706 19632 8706 19633 8706 19634 8756 19635 8 19636 8 19637 8 19638 8 19639 8 19640 8 19641 8 19642 8 19643 8 19644 8 19645 8 19646 8 19647 8 19648 8 19649 8 19650 8 19651 8 19652 8 19653 8 19654 8 19655 8 19656 8 19657 8 19658 8 19659 8 19660 8 19661 8 19662 8 19663 8 19664 8 19665 8 19666 8 19667 8 19668 8 19669 8 19670 8 19671 8 19672 8 19673 8 19674 8 19675 8 19676 8 19677 8 19678 8 19679 8 19680 8 19681 8 19682 8 19683 8 19684 8 19685 8 19686 8 19687 8 19688 8 19689 8 19690 8 19691 8 19692 8 19693 8 19694 8 19695 8 19696 8 19697 8 19698 8 19699 8 19700 8 19701 8 19702 8 19703 8 19704 8 19705 8 19706 8 19707 8 19708 8 19709 8 19710 8 19711 8 19712 8 19713 8 19714 8 19715 8757 19716 8 19717 8 19718 8 19719 8758 19720 8759 19721 8 19722 8760 19723 8 19724 8 19725 8761 19726 8762 19727 8763 19728 8 19729 8 19730 8764 19731 8765 19732 8766 19733 8767 19734 8 19735 8768 19736 8 19737 8769 19738 8770 19739 8771 19740 8772 19741 8773 19742 8774 19743 8775 19744 8776 19745 8777 19746 8706 19747 8706 19748 8706 19749 8706 19750 8706 19751 8706 19752 8706 19753 8706 19754 8706 19755 8706 19756 8706 19757 8706 19758 8706 19759 8778 19760 8706 19761 8779 19762 8780 19763 8 19764 8781 19765 8782 19766 8 19767 8783 19768 8 19769 8784 19770 8 19771 8 19772 8 19773 8785 19774 8786 19775 8787 19776 8 19777 8788 19778 8 19779 8 19780 8789 19781 8790 19782 8791 19783 8 19784 8 19785 8 19786 8792 19787 8793 19788 8 19789 8 19790 8 19791 8794 19792 8795 19793 8796 19794 8797 19795 8 19796 8 19797 8798 19798 8 19799 8799 19800 8800 19801 8 19802 8 19803 8801 19804 8 19805 8 19806 8 19807 8802 19808 8803 19809 8 19810 8 19811 8804 19812 8805 19813 8806 19814 8807 19815 8808 19816 8809 19817 8810 19818 8811 19819 8812 19820 8813 19821 8814 19822 8 19823 8815 19824 8816 19825 8817 19826 8818 19827 8819 19828 8820 19829 8821 19830 8822 19831 8823 19832 8824 19833 8825 19834 8826 19835 8827 19836 8828 19837 8829 19838 8830 19839 8831 19840 8832 19841 8833 19842 8 19843 8834 19844 8 19845 8835 19846 8 19847 8 19848 8836 19849 8837 19850 8838 19851 8839 19852 8 19853 8 19854 8840 19855 8841 19856 8842 19857 8843 19858 8844 19859 8 19860 8845 19861 8 19862 8846 19863 8 19864 8 19865 8 19866 8 19867 8847 19868 8 19869 8848 19870 8849 19871 8 19872 8850 19873 8851 19874 8852 19875 8853 19876 8854 19877 8 19878 8855 19879 8856 19880 8 19881 8857 19882 8 19883 8858 19884 8859 19885 8860 19886 8861 19887 8862 19888 8863 19889 8864 19890 8 19891 8 19892 8865 19893 8 19894 8 19895 8866 19896 8867 19897 8868 19898 8 19899 8 19900 8 19901 8 19902 8869 19903 8 19904 8 19905 8 19906 8 19907 8 19908 8870 19909 8871 19910 8872 19911 8873 19912 8874 19913 8875 19914 8876 19915 8877 19916 8878 19917 8879 19918 8880 19919 8881 19920 8882 19921 8883 19922 8884 19923 8885 19924 8886 19925 8887 19926 8888 19927 8889 19928 8890 19929 8891 19930 8892 19931 8893 19932 8894 19933 8895 19934 8896 19935 8897 19936 8898 19937 8899 19938 8900 19939 8901 19940 8902 19941 8903 19942 8904 19943 8873 19944 8873 19945 8873 19946 8873 19947 8873 19948 8873 19949 8873 19950 8873 19951 8873 19952 8873 19953 8873 19954 8873 19955 8873 19956 8873 19957 8873 19958 8873 19959 8905 19960 8906 19961 8 19962 8 19963 8 19964 8 19965 8 19966 8 19967 8907 19968 8 19969 8 19970 8908 19971 8 19972 8909 19973 8910 19974 8911 19975 8912 19976 8913 19977 8 19978 8 19979 8 19980 8 19981 8 19982 8914 19983 8915 19984 8 19985 8 19986 8916 19987 8 19988 8917 19989 8 19990 8 19991 8918 19992 8919 19993 8 19994 8873 19995 8873 19996 8873 19997 8873 19998 8873 19999 8920 20000 8921 20001 8922 20002 8873 20003 8873 20004 8873 20005 8873 20006 8873 20007 8873 20008 8873 20009 8873 20010 8873 20011 8873 20012 8923 20013 8 20014 8 20015 8 20016 8 20017 8 20018 8 20019 8 20020 8 20021 8 20022 8 20023 8 20024 8 20025 8 20026 8 20027 8 20028 8 20029 8 20030 8 20031 8 20032 8 20033 8 20034 8 20035 8 20036 8 20037 8 20038 8 20039 8 20040 8 20041 8 20042 8 20043 8 20044 8 20045 8 20046 8 20047 8 20048 8 20049 8 20050 8 20051 8 20052 8 20053 8 20054 8 20055 8 20056 8 20057 8 20058 8 20059 8 20060 8 20061 8 20062 8 20063 8 20064 8 20065 8 20066 8 20067 8 20068 8 20069 8 20070 8 20071 8 20072 8 20073 8 20074 8 20075 8 20076 8 20077 8 20078 8 20079 8 20080 8 20081 8 20082 8 20083 8 20084 8 20085 8 20086 8 20087 8 20088 8 20089 8 20090 8 20091 8 20092 8 20093 8924 20094 8 20095 8 20096 8 20097 8925 20098 8926 20099 8 20100 8927 20101 8 20102 8 20103 8928 20104 8929 20105 8930 20106 8 20107 8 20108 8931 20109 8932 20110 8933 20111 8934 20112 8 20113 8935 20114 8 20115 8936 20116 8937 20117 8938 20118 8939 20119 8940 20120 8941 20121 8942 20122 8943 20123 8944 20124 8873 20125 8873 20126 8873 20127 8873 20128 8873 20129 8873 20130 8873 20131 8873 20132 8873 20133 8873 20134 8873 20135 8873 20136 8873 20137 8945 20138 8873 20139 8946 20140 8947 20141 8 20142 8948 20143 8949 20144 8 20145 8950 20146 8 20147 8951 20148 8 20149 8 20150 8 20151 8952 20152 8953 20153 8954 20154 8 20155 8955 20156 8 20157 8 20158 8956 20159 8957 20160 8958 20161 8 20162 8 20163 8 20164 8959 20165 8960 20166 8 20167 8 20168 8 20169 8961 20170 8962 20171 8963 20172 8964 20173 8 20174 8 20175 8965 20176 8 20177 8966 20178 8967 20179 8 20180 8 20181 8968 20182 8 20183 8 20184 8 20185 8969 20186 8970 20187 8 20188 8 20189 8971 20190 8972 20191 8973 20192 8974 20193 8975 20194 8976 20195 8977 20196 8978 20197 8979 20198 8980 20199 8981 20200 8 20201 8982 20202 8983 20203 8984 20204 8985 20205 8986 20206 8987 20207 8988 20208 8989 20209 8990 20210 8991 20211 8 20212 8 20213 8992 20214 8 20215 8993 20216 8 20217 8994 20218 8995 20219 8 20220 8 20221 8 20222 8996 20223 8 20224 8 20225 8997 20226 8998 20227 8999 20228 9000 20229 8 20230 8 20231 9001 20232 8 20233 9002 20234 9003 20235 8 20236 8 20237 9004 20238 8 20239 9005 20240 8 20241 9006 20242 9007 20243 9008 20244 8873 20245 8873 20246 9009 20247 9010 20248 9011 20249 9012 20250 9013 20251 9014 20252 9015 20253 9016 20254 9017 20255 9018 20256 8 20257 9019 20258 8 20259 9020 20260 8 20261 8 20262 9021 20263 9022 20264 9023 20265 9024 20266 8 20267 8 20268 9025 20269 9026 20270 9027 20271 9028 20272 9029 20273 8 20274 9030 20275 8 20276 9031 20277 8 20278 8 20279 8 20280 8 20281 9032 20282 8 20283 9033 20284 9034 20285 8 20286 9035 20287 9036 20288 9037 20289 9038 20290 9039 20291 8 20292 9040 20293 9041 20294 8 20295 9042 20296 8 20297 9043 20298 9044 20299 9045 20300 9046 20301 9047 20302 9048 20303 9049 20304 8 20305 8 20306 9050 20307 8 20308 8 20309 9051 20310 9052 20311 9053 20312 8 20313 8 20314 8 20315 8 20316 9054 20317 8 20318 8 20319 8 20320 8 20321 8 20322 9055 20323 9056 20324 9057 20325 9058 20326 9059 20327 9060 20328 9061 20329 9062 20330 9063 20331 9064 20332 9065 20333 9066 20334 9067 20335 9068 20336 9069 20337 9070 20338 9071 20339 9072 20340 9073 20341 9074 20342 9075 20343 9076 20344 9077 20345 9078 20346 9079 20347 9080 20348 9081 20349 9082 20350 9083 20351 9084 20352 9085 20353 9086 20354 9087 20355 9088 20356 9089 20357 9058 20358 9058 20359 9058 20360 9058 20361 9058 20362 9058 20363 9058 20364 9058 20365 9058 20366 9058 20367 9058 20368 9058 20369 9058 20370 9058 20371 9058 20372 9058 20373 9090 20374 9091 20375 8 20376 8 20377 8 20378 8 20379 8 20380 8 20381 9092 20382 8 20383 8 20384 9093 20385 8 20386 9094 20387 9095 20388 9096 20389 9097 20390 9098 20391 8 20392 8 20393 8 20394 8 20395 8 20396 9099 20397 9100 20398 8 20399 8 20400 9101 20401 8 20402 9102 20403 8 20404 8 20405 9103 20406 9104 20407 8 20408 9058 20409 9058 20410 9058 20411 9058 20412 9058 20413 9105 20414 9106 20415 9107 20416 9058 20417 9058 20418 9058 20419 9058 20420 9058 20421 9058 20422 9058 20423 9058 20424 9058 20425 9058 20426 9108 20427 8 20428 8 20429 8 20430 8 20431 8 20432 8 20433 8 20434 8 20435 8 20436 8 20437 8 20438 8 20439 8 20440 8 20441 8 20442 8 20443 8 20444 8 20445 8 20446 8 20447 8 20448 8 20449 8 20450 8 20451 8 20452 8 20453 8 20454 8 20455 8 20456 8 20457 8 20458 8 20459 8 20460 8 20461 8 20462 8 20463 8 20464 8 20465 8 20466 8 20467 8 20468 8 20469 8 20470 8 20471 8 20472 8 20473 8 20474 8 20475 8 20476 8 20477 8 20478 8 20479 8 20480 8 20481 8 20482 8 20483 8 20484 8 20485 8 20486 8 20487 8 20488 8 20489 8 20490 8 20491 8 20492 8 20493 8 20494 8 20495 8 20496 8 20497 8 20498 8 20499 8 20500 8 20501 8 20502 8 20503 8 20504 8 20505 8 20506 8 20507 9109 20508 8 20509 8 20510 8 20511 9110 20512 9111 20513 8 20514 9112 20515 8 20516 8 20517 9113 20518 9114 20519 9115 20520 8 20521 8 20522 9116 20523 9117 20524 9118 20525 9119 20526 8 20527 9120 20528 8 20529 9121 20530 9122 20531 9123 20532 9124 20533 9125 20534 9126 20535 9127 20536 9128 20537 9129 20538 9058 20539 9058 20540 9058 20541 9058 20542 9058 20543 9058 20544 9058 20545 9058 20546 9058 20547 9058 20548 9058 20549 9058 20550 9058 20551 9130 20552 9058 20553 9131 20554 9132 20555 8 20556 9133 20557 9134 20558 8 20559 9135 20560 8 20561 9136 20562 8 20563 8 20564 8 20565 9137 20566 9138 20567 9139 20568 8 20569 9140 20570 8 20571 8 20572 9141 20573 9142 20574 9143 20575 8 20576 8 20577 8 20578 9144 20579 9145 20580 8 20581 8 20582 8 20583 9146 20584 9147 20585 9148 20586 9149 20587 8 20588 8 20589 9150 20590 8 20591 9151 20592 9152 20593 8 20594 8 20595 9153 20596 8 20597 8 20598 8 20599 9154 20600 9155 20601 8 20602 8 20603 9156 20604 9157 20605 9158 20606 9159 20607 9160 20608 9161 20609 9162 20610 9163 20611 9164 20612 9165 20613 9166 20614 8 20615 9167 20616 9168 20617 9169 20618 9170 20619 9171 20620 9172 20621 9173 20622 9174 20623 9175 20624 9176 20625 8 20626 9177 20627 8873 20628 9178 20629 8 20630 9179 20631 8 20632 9180 20633 9181 20634 8 20635 8 20636 8 20637 9182 20638 8 20639 8 20640 9183 20641 9184 20642 9185 20643 9186 20644 8 20645 8 20646 9187 20647 8 20648 9188 20649 9189 20650 8 20651 8 20652 9190 20653 8 20654 9191 20655 8 20656 9192 20657 9193 20658 9194 20659 9058 20660 9058 20661 9195 20662 9196 20663 9058 20664 9058 20665 9058 20666 9058 20667 8 20668 9197 20669 8 20670 9198 20671 8 20672 8 20673 9058 20674 9058 20675 8 20676 9058 20677 9058 20678 9199 20679 8 20680 8 20681 8 20682 8 20683 8 20684 8 20685 9200 20686 9201 20687 9202 20688 9203 20689 8 20690 9204 20691 8 20692 9205 20693 9206 20694 9207 20695 9208 20696 9209 20697 8 20698 8 20699 8 20700 8 20701 8 20702 9210 20703 8 20704 8 20705 9211 20706 9212 20707 9213 20708 9214 20709 9215 20710 9216 20711 9217 20712 8 20713 9218 20714 8 20715 9219 20716 9220 20717 9221 20718 9222 20719 9223 20720 8 20721 9224 20722 9225 20723 9226 20724 9227 20725 9058 20726 9228 20727 9229 20728 9230 20729 8 20730 9058 20731 9058 20732 9058 20733 8 20734 9231 20735 8 20736 8 20737 8 20738 9232 20739 9233 20740 9234 20741 8 20742 8 20743 9235 20744 9236 20745 8 20746 9237 20747 9238 20748 8 20749 9058 20750 8 20751 9239 20752 8 20753 9058 20754 8 20755 9240 20756 9241 20757 9242 20758 9243 20759 8 20760 8 20761 9244 20762 8 20763 8 20764 9058 20765 9058 20766 9245 20767 9246 20768 8 20769 9058 20770 9058 20771 9058 20772 9247 20773 8 20774 9248 20775 9249 20776 8 20777 9250 20778 8 20779 9058 20780 8873 20781 8873 20782 8873 20783 8 20784 9251 20785 8 20786 9252 20787 8 20788 8 20789 8873 20790 8873 20791 8 20792 8873 20793 8873 20794 9253 20795 8 20796 8 20797 8 20798 8 20799 8 20800 8 20801 9254 20802 9255 20803 9256 20804 9257 20805 8 20806 9258 20807 8 20808 9259 20809 9260 20810 9261 20811 9262 20812 9263 20813 8 20814 8 20815 8 20816 8 20817 8 20818 9264 20819 8 20820 8 20821 9265 20822 9266 20823 9267 20824 9268 20825 9269 20826 9270 20827 9271 20828 8 20829 9272 20830 8 20831 9273 20832 9274 20833 9275 20834 9276 20835 9277 20836 8 20837 9278 20838 9279 20839 9280 20840 9281 20841 8873 20842 9282 20843 9283 20844 9284 20845 8 20846 8873 20847 8873 20848 8873 20849 8 20850 9285 20851 8 20852 8 20853 8 20854 9286 20855 9287 20856 9288 20857 8 20858 8 20859 9289 20860 9290 20861 8 20862 9291 20863 9292 20864 8 20865 8873 20866 8 20867 9293 20868 8 20869 8873 20870 8 20871 9294 20872 9295 20873 9296 20874 9297 20875 8 20876 8 20877 9298 20878 8 20879 8 20880 8873 20881 8873 20882 9299 20883 9300 20884 8 20885 8873 20886 8873 20887 8873 20888 9301 20889 8 20890 9302 20891 9303 20892 8 20893 9304 20894 8 20895 8873 20896 9305 20897 9306 20898 9307 20899 9308 20900 8 20901 9309 20902 9310 20903 9311 20904 9312 20905 8 20906 9313 20907 9314 20908 8 20909 9315 20910 8 20911 8 20912 8 20913 9316 20914 9317 20915 9318 20916 8 20917 9319 20918 8 20919 8 20920 9320 20921 9321 20922 9322 20923 8 20924 8 20925 8 20926 9323 20927 9324 20928 9325 20929 9326 20930 9327 20931 8 20932 9328 20933 9329 20934 9330 20935 9331 20936 8 20937 9332 20938 9333 20939 8 20940 9334 20941 8 20942 8 20943 8 20944 9335 20945 9336 20946 9337 20947 8 20948 9338 20949 8 20950 8 20951 9339 20952 9340 20953 9341 20954 8 20955 8 20956 8 20957 9342 20958 9343 20959 8 20960 8 20961 9344 20962 8 20963 9345 20964 9346 20965 9347 20966 9348 20967 9349 20968 9350 20969 9351 20970 9352 20971 9353 20972 9354 20973 9355 20974 9356 20975 9357 20976 9358 20977 8 20978 9359 20979 9360 20980 9361 20981 9362 20982 9363 20983 9364 20984 9365 20985 9366 20986 9367 20987 9368 20988 8 20989 9369 20990 8 20991 9370 20992 8 20993 9371 20994 8 20995 9372 20996 9373 20997 8 20998 8 20999 9374 21000 8 21001 8 21002 8 21003 8 21004 9375 21005 9376 21006 8 21007 9377 21008 9378 21009 9379 21010 9380 21011 9381 21012 8 21013 9382 21014 9383 21015 9384 21016 9385 21017 9386 21018 9387 21019 9388 21020 9389 21021 9390 21022 8 21023 9391 21024 9392 21025 9393 21026 9394 21027 9395 21028 9396 21029 9397 21030 9398 21031 8 21032 9399 21033 9400 21034 8 21035 9401 21036 8 21037 9402 21038 9403 21039 8 21040 9404 21041 8 21042 8 21043 9405 21044 9406 21045 8 21046 9407 21047 9408 21048 9409 21049 8 21050 9410 21051 9411 21052 9412 21053 9413 21054 8 21055 8 21056 9414 21057 9415 21058 8 21059 9416 21060 8 21061 9417 21062 9418 21063 9419 21064 9420 21065 8 21066 9421 21067 9422 21068 9423 21069 9424 21070 9425 21071 9426 21072 8 21073 9427 21074 9428 21075 8 21076 9429 21077 9430 21078 9431 21079 9432 21080 9433 21081 9434 21082 8 21083 9435 21084 9436 21085 9437 21086 9438 21087 8 21088 9439 21089 9440 21090 9441 21091 8 21092 9442 21093 8 21094 9443 21095 9444 21096 9445 21097 8 21098 9446 21099 9447 21100 8 21101 8 21102 9448 21103 8 21104 9449 21105 8 21106 9450 21107 8 21108 9451 21109 8 21110 9452 21111 9453 21112 9454 21113 9455 21114 9456 21115 9457 21116 9458 21117 9459 21118 9460 21119 8 21120 8 21121 9461 21122 9462 21123 8 21124 8 21125 9463 21126 9464 21127 8211 21128 9465 21129 9466 21130 9467 21131 9468 21132 8211 21133 8211 21134 9469 21135 9470 21136 8211 21137 9471 21138 9472 21139 9473 21140 9474 21141 8211 21142 9475 21143 9476 21144 9477 21145 9478 21146 9479 21147 9480 21148 9481 21149 9482 21150 9483 21151 9484 21152 8 21153 9485 21154 8 21155 9486 21156 8 21157 8 21158 9487 21159 9488 21160 9489 21161 9490 21162 8 21163 8 21164 9491 21165 9492 21166 9493 21167 9494 21168 9495 21169 8 21170 9496 21171 8 21172 9497 21173 8 21174 8 21175 8 21176 8 21177 9498 21178 8 21179 9499 21180 9500 21181 8 21182 9501 21183 9502 21184 9503 21185 9504 21186 9505 21187 8 21188 9506 21189 9507 21190 8 21191 9508 21192 8 21193 9509 21194 9510 21195 9511 21196 9512 21197 9513 21198 9514 21199 9515 21200 8 21201 8 21202 9516 21203 8 21204 8 21205 9517 21206 9518 21207 9519 21208 8 21209 8 21210 8 21211 8 21212 9520 21213 8 21214 8 21215 8 21216 8 21217 8 21218 9521 21219 9522 21220 9523 21221 9517 21222 9524 21223 9525 21224 9526 21225 9527 21226 9528 21227 9529 21228 9530 21229 9531 21230 9532 21231 9533 21232 9534 21233 9535 21234 9536 21235 9537 21236 9538 21237 9539 21238 9540 21239 9541 21240 9542 21241 9543 21242 9544 21243 9545 21244 9546 21245 9547 21246 9548 21247 9549 21248 9550 21249 9551 21250 9552 21251 9553 21252 9554 21253 9517 21254 9517 21255 9517 21256 9517 21257 9517 21258 9517 21259 9517 21260 9517 21261 9517 21262 9517 21263 9517 21264 9517 21265 9517 21266 9517 21267 9517 21268 9517 21269 9555 21270 9556 21271 8 21272 8 21273 8 21274 8 21275 8 21276 8 21277 9557 21278 8 21279 8 21280 9558 21281 8 21282 9559 21283 9560 21284 9561 21285 9562 21286 9563 21287 8 21288 8 21289 8 21290 8 21291 8 21292 9564 21293 9565 21294 8 21295 8 21296 9566 21297 8 21298 9567 21299 8 21300 8 21301 9568 21302 9569 21303 8 21304 9517 21305 9517 21306 9517 21307 9517 21308 9517 21309 9570 21310 9571 21311 9572 21312 9517 21313 9517 21314 9517 21315 9517 21316 9517 21317 9517 21318 9517 21319 9517 21320 9517 21321 9517 21322 9573 21323 8 21324 8 21325 8 21326 8 21327 8 21328 8 21329 8 21330 8 21331 8 21332 8 21333 8 21334 8 21335 8 21336 8 21337 8 21338 8 21339 8 21340 8 21341 8 21342 8 21343 8 21344 8 21345 8 21346 8 21347 8 21348 8 21349 8 21350 8 21351 8 21352 8 21353 8 21354 8 21355 8 21356 8 21357 8 21358 8 21359 8 21360 8 21361 8 21362 8 21363 8 21364 8 21365 8 21366 8 21367 8 21368 8 21369 8 21370 8 21371 8 21372 8 21373 8 21374 8 21375 8 21376 8 21377 8 21378 8 21379 8 21380 8 21381 8 21382 8 21383 8 21384 8 21385 8 21386 8 21387 8 21388 8 21389 8 21390 8 21391 8 21392 8 21393 8 21394 8 21395 8 21396 8 21397 8 21398 8 21399 8 21400 8 21401 8 21402 8 21403 9574 21404 8 21405 8 21406 8 21407 9575 21408 9576 21409 8 21410 9577 21411 8 21412 8 21413 9578 21414 9579 21415 9580 21416 8 21417 8 21418 9581 21419 9582 21420 9583 21421 9584 21422 8 21423 9585 21424 8 21425 9586 21426 9587 21427 9588 21428 9589 21429 9590 21430 9591 21431 9592 21432 9593 21433 9594 21434 9517 21435 9517 21436 9517 21437 9517 21438 9517 21439 9517 21440 9517 21441 9517 21442 9517 21443 9517 21444 9517 21445 9517 21446 9517 21447 9595 21448 9517 21449 9596 21450 9597 21451 8 21452 9598 21453 9599 21454 8 21455 9600 21456 8 21457 9601 21458 8 21459 8 21460 8 21461 9602 21462 9603 21463 9604 21464 8 21465 9605 21466 8 21467 8 21468 9606 21469 9607 21470 9608 21471 8 21472 8 21473 8 21474 9609 21475 9610 21476 8 21477 8 21478 8 21479 9611 21480 9612 21481 9613 21482 9614 21483 8 21484 8 21485 9615 21486 8 21487 9616 21488 9617 21489 8 21490 8 21491 9618 21492 8 21493 8 21494 8 21495 9619 21496 9620 21497 8 21498 8 21499 9621 21500 9622 21501 9623 21502 9624 21503 9625 21504 9626 21505 9627 21506 9628 21507 9629 21508 9630 21509 9631 21510 8 21511 9632 21512 9633 21513 9634 21514 9635 21515 9636 21516 9637 21517 9638 21518 9639 21519 9640 21520 9641 21521 9642 21522 9643 21523 9644 21524 9645 21525 9646 21526 9647 21527 9648 21528 9649 21529 9650 21530 8 21531 9651 21532 8 21533 9652 21534 8 21535 8 21536 9653 21537 9654 21538 9655 21539 9656 21540 8 21541 8 21542 9657 21543 9658 21544 9659 21545 9660 21546 9661 21547 8 21548 9662 21549 8 21550 9663 21551 8 21552 8 21553 8 21554 8 21555 9664 21556 8 21557 9665 21558 9666 21559 8 21560 9667 21561 9668 21562 9669 21563 9670 21564 9671 21565 8 21566 9672 21567 9673 21568 8 21569 9674 21570 8 21571 9675 21572 9676 21573 9677 21574 9678 21575 9679 21576 9680 21577 9681 21578 8 21579 8 21580 9682 21581 8 21582 8 21583 9683 21584 9684 21585 9685 21586 8 21587 8 21588 8 21589 8 21590 9686 21591 8 21592 8 21593 8 21594 8 21595 8 21596 9687 21597 9688 21598 9689 21599 9690 21600 9691 21601 9692 21602 9693 21603 9694 21604 9695 21605 9696 21606 9697 21607 9698 21608 9699 21609 9700 21610 9701 21611 9702 21612 9703 21613 9704 21614 9705 21615 9706 21616 9707 21617 9708 21618 9709 21619 9710 21620 9711 21621 9712 21622 9713 21623 9714 21624 9715 21625 9716 21626 9717 21627 9718 21628 9719 21629 9720 21630 9721 21631 9690 21632 9690 21633 9690 21634 9690 21635 9690 21636 9690 21637 9690 21638 9690 21639 9690 21640 9690 21641 9690 21642 9690 21643 9690 21644 9690 21645 9690 21646 9690 21647 9722 21648 9723 21649 8 21650 8 21651 8 21652 8 21653 8 21654 8 21655 9724 21656 8 21657 8 21658 9725 21659 8 21660 9726 21661 9727 21662 9728 21663 9729 21664 9730 21665 8 21666 8 21667 8 21668 8 21669 8 21670 9731 21671 9732 21672 8 21673 8 21674 9733 21675 8 21676 9734 21677 8 21678 8 21679 9735 21680 9736 21681 8 21682 9690 21683 9690 21684 9690 21685 9690 21686 9690 21687 9737 21688 9738 21689 9739 21690 9690 21691 9690 21692 9690 21693 9690 21694 9690 21695 9690 21696 9690 21697 9690 21698 9690 21699 9690 21700 9740 21701 8 21702 8 21703 8 21704 8 21705 8 21706 8 21707 8 21708 8 21709 8 21710 8 21711 8 21712 8 21713 8 21714 8 21715 8 21716 8 21717 8 21718 8 21719 8 21720 8 21721 8 21722 8 21723 8 21724 8 21725 8 21726 8 21727 8 21728 8 21729 8 21730 8 21731 8 21732 8 21733 8 21734 8 21735 8 21736 8 21737 8 21738 8 21739 8 21740 8 21741 8 21742 8 21743 8 21744 8 21745 8 21746 8 21747 8 21748 8 21749 8 21750 8 21751 8 21752 8 21753 8 21754 8 21755 8 21756 8 21757 8 21758 8 21759 8 21760 8 21761 8 21762 8 21763 8 21764 8 21765 8 21766 8 21767 8 21768 8 21769 8 21770 8 21771 8 21772 8 21773 8 21774 8 21775 8 21776 8 21777 8 21778 8 21779 8 21780 8 21781 9741 21782 8 21783 8 21784 8 21785 9742 21786 9743 21787 8 21788 9744 21789 8 21790 8 21791 9745 21792 9746 21793 9747 21794 8 21795 8 21796 9748 21797 9749 21798 9750 21799 9751 21800 8 21801 9752 21802 8 21803 9753 21804 9754 21805 9755 21806 9756 21807 9757 21808 9758 21809 9759 21810 9760 21811 9761 21812 9690 21813 9690 21814 9690 21815 9690 21816 9690 21817 9690 21818 9690 21819 9690 21820 9690 21821 9690 21822 9690 21823 9690 21824 9690 21825 9762 21826 9690 21827 9763 21828 9764 21829 8 21830 9765 21831 9766 21832 8 21833 9767 21834 8 21835 9768 21836 8 21837 8 21838 8 21839 9769 21840 9770 21841 9771 21842 8 21843 9772 21844 8 21845 8 21846 9773 21847 9774 21848 9775 21849 8 21850 8 21851 8 21852 9776 21853 9777 21854 8 21855 8 21856 8 21857 9778 21858 9779 21859 9780 21860 9781 21861 8 21862 8 21863 9782 21864 8 21865 9783 21866 9784 21867 8 21868 8 21869 9785 21870 8 21871 8 21872 8 21873 9786 21874 9787 21875 8 21876 8 21877 9788 21878 9789 21879 9790 21880 9791 21881 9792 21882 9793 21883 9794 21884 9795 21885 9796 21886 9797 21887 9798 21888 8 21889 9799 21890 9800 21891 9801 21892 9802 21893 9803 21894 9804 21895 9805 21896 9806 21897 9807 21898 9808 21899 8 21900 8 21901 9809 21902 8 21903 9810 21904 8 21905 9811 21906 9812 21907 8 21908 8 21909 8 21910 9813 21911 8 21912 8 21913 9814 21914 9815 21915 9816 21916 9817 21917 8 21918 8 21919 9818 21920 8 21921 9819 21922 9820 21923 8 21924 8 21925 9821 21926 8 21927 9822 21928 8 21929 9823 21930 9824 21931 9825 21932 9690 21933 9690 21934 9826 21935 9827 21936 9690 21937 9690 21938 9690 21939 9690 21940 8 21941 9828 21942 8 21943 9829 21944 8 21945 8 21946 9690 21947 9690 21948 8 21949 9690 21950 9690 21951 9830 21952 8 21953 8 21954 8 21955 8 21956 8 21957 8 21958 9831 21959 9832 21960 9833 21961 9834 21962 8 21963 9835 21964 8 21965 9836 21966 9837 21967 9838 21968 9839 21969 9840 21970 8 21971 8 21972 8 21973 8 21974 8 21975 9841 21976 8 21977 8 21978 9842 21979 9843 21980 9844 21981 9845 21982 9846 21983 9847 21984 9848 21985 8 21986 9849 21987 8 21988 9850 21989 9851 21990 9852 21991 9853 21992 9854 21993 8 21994 9855 21995 9856 21996 9857 21997 9858 21998 9690 21999 9859 22000 9860 22001 9861 22002 8 22003 9690 22004 9690 22005 9690 22006 8 22007 9862 22008 8 22009 8 22010 8 22011 9863 22012 9864 22013 9865 22014 8 22015 8 22016 9866 22017 9867 22018 8 22019 9868 22020 9869 22021 8 22022 9690 22023 8 22024 9870 22025 8 22026 9690 22027 8 22028 9871 22029 9872 22030 9873 22031 9874 22032 8 22033 8 22034 9875 22035 8 22036 8 22037 9690 22038 9690 22039 9876 22040 9877 22041 8 22042 9690 22043 9690 22044 9690 22045 9878 22046 8 22047 9879 22048 9880 22049 8 22050 9881 22051 8 22052 9690 22053 9882 22054 8 22055 9883 22056 8 22057 9884 22058 9885 22059 8 22060 8 22061 8 22062 9886 22063 8 22064 8 22065 9887 22066 9888 22067 9889 22068 9890 22069 8 22070 8 22071 9891 22072 8 22073 9892 22074 9893 22075 8 22076 8 22077 9894 22078 8 22079 9895 22080 8 22081 9896 22082 9897 22083 9898 22084 9517 22085 9517 22086 9899 22087 9900 22088 9901 22089 9902 22090 9903 22091 9904 22092 9905 22093 9906 22094 9907 22095 9908 22096 8 22097 9909 22098 8 22099 9910 22100 8 22101 8 22102 9911 22103 9912 22104 9913 22105 9914 22106 8 22107 8 22108 9915 22109 9916 22110 9917 22111 9918 22112 9919 22113 8 22114 9920 22115 8 22116 9921 22117 8 22118 8 22119 8 22120 8 22121 9922 22122 8 22123 9923 22124 9924 22125 8 22126 9925 22127 9926 22128 9927 22129 9928 22130 9929 22131 8 22132 9930 22133 9931 22134 8 22135 9932 22136 8 22137 9933 22138 9934 22139 9935 22140 9936 22141 9937 22142 9938 22143 9939 22144 8 22145 8 22146 9940 22147 8 22148 8 22149 9941 22150 9942 22151 9943 22152 8 22153 8 22154 8 22155 8 22156 9944 22157 8 22158 8 22159 8 22160 8 22161 8 22162 9945 22163 9946 22164 9947 22165 9948 22166 9949 22167 9950 22168 9951 22169 9952 22170 9953 22171 9954 22172 9955 22173 9956 22174 9957 22175 9958 22176 9959 22177 9960 22178 9961 22179 9962 22180 9963 22181 9964 22182 9965 22183 9966 22184 9967 22185 9968 22186 9969 22187 9970 22188 9971 22189 9972 22190 9973 22191 9974 22192 9975 22193 9976 22194 9977 22195 9978 22196 9979 22197 9948 22198 9948 22199 9948 22200 9948 22201 9948 22202 9948 22203 9948 22204 9948 22205 9948 22206 9948 22207 9948 22208 9948 22209 9948 22210 9948 22211 9948 22212 9948 22213 9980 22214 9981 22215 8 22216 8 22217 8 22218 8 22219 8 22220 8 22221 9982 22222 8 22223 8 22224 9983 22225 8 22226 9984 22227 9985 22228 9986 22229 9987 22230 9988 22231 8 22232 8 22233 8 22234 8 22235 8 22236 9989 22237 9990 22238 8 22239 8 22240 9991 22241 8 22242 9992 22243 8 22244 8 22245 9993 22246 9994 22247 8 22248 9948 22249 9948 22250 9948 22251 9948 22252 9948 22253 9995 22254 9996 22255 9997 22256 9948 22257 9948 22258 9948 22259 9948 22260 9948 22261 9948 22262 9948 22263 9948 22264 9948 22265 9948 22266 9998 22267 8 22268 8 22269 8 22270 8 22271 8 22272 8 22273 8 22274 8 22275 8 22276 8 22277 8 22278 8 22279 8 22280 8 22281 8 22282 8 22283 8 22284 8 22285 8 22286 8 22287 8 22288 8 22289 8 22290 8 22291 8 22292 8 22293 8 22294 8 22295 8 22296 8 22297 8 22298 8 22299 8 22300 8 22301 8 22302 8 22303 8 22304 8 22305 8 22306 8 22307 8 22308 8 22309 8 22310 8 22311 8 22312 8 22313 8 22314 8 22315 8 22316 8 22317 8 22318 8 22319 8 22320 8 22321 8 22322 8 22323 8 22324 8 22325 8 22326 8 22327 8 22328 8 22329 8 22330 8 22331 8 22332 8 22333 8 22334 8 22335 8 22336 8 22337 8 22338 8 22339 8 22340 8 22341 8 22342 8 22343 8 22344 8 22345 8 22346 8 22347 9999 22348 8 22349 8 22350 8 22351 10000 22352 10001 22353 8 22354 10002 22355 8 22356 8 22357 10003 22358 10004 22359 10005 22360 8 22361 8 22362 10006 22363 10007 22364 10008 22365 10009 22366 8 22367 10010 22368 8 22369 10011 22370 10012 22371 10013 22372 10014 22373 10015 22374 10016 22375 10017 22376 10018 22377 10019 22378 9948 22379 9948 22380 9948 22381 9948 22382 9948 22383 9948 22384 9948 22385 9948 22386 9948 22387 9948 22388 9948 22389 9948 22390 9948 22391 10020 22392 9948 22393 10021 22394 10022 22395 8 22396 10023 22397 10024 22398 8 22399 10025 22400 8 22401 10026 22402 8 22403 8 22404 8 22405 10027 22406 10028 22407 10029 22408 8 22409 10030 22410 8 22411 8 22412 10031 22413 10032 22414 10033 22415 8 22416 8 22417 8 22418 10034 22419 10035 22420 8 22421 8 22422 8 22423 10036 22424 10037 22425 10038 22426 10039 22427 8 22428 8 22429 10040 22430 8 22431 10041 22432 10042 22433 8 22434 8 22435 10043 22436 8 22437 8 22438 8 22439 10044 22440 10045 22441 8 22442 8 22443 10046 22444 10047 22445 10048 22446 10049 22447 10050 22448 10051 22449 10052 22450 10053 22451 10054 22452 10055 22453 10056 22454 8 22455 10057 22456 10058 22457 10059 22458 10060 22459 10061 22460 10062 22461 10063 22462 10064 22463 10065 22464 10066 22465 8 22466 10067 22467 9517 22468 10068 22469 8 22470 10069 22471 8 22472 10070 22473 10071 22474 8 22475 8 22476 8 22477 10072 22478 8 22479 8 22480 10073 22481 10074 22482 10075 22483 10076 22484 8 22485 8 22486 10077 22487 8 22488 10078 22489 10079 22490 8 22491 8 22492 10080 22493 8 22494 10081 22495 8 22496 10082 22497 10083 22498 10084 22499 9948 22500 9948 22501 10085 22502 10086 22503 9948 22504 9948 22505 9948 22506 9948 22507 8 22508 10087 22509 8 22510 10088 22511 8 22512 8 22513 9948 22514 9948 22515 8 22516 9948 22517 9948 22518 10089 22519 8 22520 8 22521 8 22522 8 22523 8 22524 8 22525 10090 22526 10091 22527 10092 22528 10093 22529 8 22530 10094 22531 8 22532 10095 22533 10096 22534 10097 22535 10098 22536 10099 22537 8 22538 8 22539 8 22540 8 22541 8 22542 10100 22543 8 22544 8 22545 10101 22546 10102 22547 10103 22548 10104 22549 10105 22550 10106 22551 10107 22552 8 22553 10108 22554 8 22555 10109 22556 10110 22557 10111 22558 10112 22559 10113 22560 8 22561 10114 22562 10115 22563 10116 22564 10117 22565 9948 22566 10118 22567 10119 22568 10120 22569 8 22570 9948 22571 9948 22572 9948 22573 8 22574 10121 22575 8 22576 8 22577 8 22578 10122 22579 10123 22580 10124 22581 8 22582 8 22583 10125 22584 10126 22585 8 22586 10127 22587 10128 22588 8 22589 9948 22590 8 22591 10129 22592 8 22593 9948 22594 8 22595 10130 22596 10131 22597 10132 22598 10133 22599 8 22600 8 22601 10134 22602 8 22603 8 22604 9948 22605 9948 22606 10135 22607 10136 22608 8 22609 9948 22610 9948 22611 9948 22612 10137 22613 8 22614 10138 22615 10139 22616 8 22617 10140 22618 8 22619 9948 22620 9517 22621 9517 22622 9517 22623 8 22624 10141 22625 8 22626 10142 22627 8 22628 8 22629 9517 22630 9517 22631 8 22632 9517 22633 9517 22634 10143 22635 8 22636 8 22637 8 22638 8 22639 8 22640 8 22641 10144 22642 10145 22643 10146 22644 10147 22645 8 22646 10148 22647 8 22648 10149 22649 10150 22650 10151 22651 10152 22652 10153 22653 8 22654 8 22655 8 22656 8 22657 8 22658 10154 22659 8 22660 8 22661 10155 22662 10156 22663 10157 22664 10158 22665 10159 22666 10160 22667 10161 22668 8 22669 10162 22670 8 22671 10163 22672 10164 22673 10165 22674 10166 22675 10167 22676 8 22677 10168 22678 10169 22679 10170 22680 10171 22681 9517 22682 10172 22683 10173 22684 10174 22685 8 22686 9517 22687 9517 22688 9517 22689 8 22690 10175 22691 8 22692 8 22693 8 22694 10176 22695 10177 22696 10178 22697 8 22698 8 22699 10179 22700 10180 22701 8 22702 10181 22703 10182 22704 8 22705 9517 22706 8 22707 10183 22708 8 22709 9517 22710 8 22711 10184 22712 10185 22713 10186 22714 10187 22715 8 22716 8 22717 10188 22718 8 22719 8 22720 9517 22721 9517 22722 10189 22723 10190 22724 8 22725 9517 22726 9517 22727 9517 22728 10191 22729 8 22730 10192 22731 10193 22732 8 22733 10194 22734 8 22735 9517 22736 10195 22737 10196 22738 1122 22739 10197 22740 10198 22741 10199 22742 10200 22743 1122 22744 1122 22745 10201 22746 10202 22747 8 22748 8 22749 10203 22750 8 22751 10204 22752 10205 22753 10206 22754 10207 22755 10208 22756 10209 22757 10210 22758 10211 22759 10212 22760 10213 22761 10214 22762 10215 22763 10216 22764 10217 22765 8 22766 10218 22767 10219 22768 10220 22769 10221 22770 10222 22771 10223 22772 10224 22773 10225 22774 10226 22775 10227 22776 8 22777 10228 22778 8 22779 10229 22780 8 22781 10230 22782 8 22783 10231 22784 10232 22785 8 22786 8 22787 10233 22788 8 22789 8 22790 8 22791 8 22792 10234 22793 10235 22794 8 22795 10236 22796 10237 22797 10238 22798 10239 22799 10240 22800 8 22801 10241 22802 10242 22803 10243 22804 10244 22805 10245 22806 10246 22807 10247 22808 10248 22809 10249 22810 8 22811 10250 22812 10251 22813 10252 22814 10253 22815 10254 22816 10255 22817 10256 22818 10257 22819 8 22820 10258 22821 10259 22822 8 22823 10260 22824 8 22825 10261 22826 10262 22827 8 22828 10263 22829 8 22830 8 22831 10264 22832 10265 22833 8 22834 10266 22835 10267 22836 10268 22837 8 22838 10269 22839 10270 22840 10271 22841 10272 22842 8 22843 8 22844 10273 22845 10274 22846 8 22847 10275 22848 8 22849 10276 22850 10277 22851 10278 22852 10279 22853 8 22854 10280 22855 10281 22856 10282 22857 10283 22858 10284 22859 10285 22860 8 22861 10286 22862 10287 22863 8 22864 10288 22865 10289 22866 10290 22867 10291 22868 10292 22869 10293 22870 8 22871 10294 22872 10295 22873 10296 22874 10297 22875 8 22876 10298 22877 10299 22878 10300 22879 8 22880 10301 22881 8 22882 10302 22883 10303 22884 10304 22885 8 22886 10305 22887 10306 22888 8 22889 8 22890 10307 22891 8 22892 10308 22893 8 22894 10309 22895 8 22896 10310 22897 8 22898 10311 22899 10312 22900 10313 22901 10314 22902 10315 22903 10316 22904 10317 22905 10318 22906 10319 22907 8 22908 8 22909 10320 22910 10321 22911 8 22912 8 22913 10322 22914 10323 22915 10324 22916 10325 22917 10326 22918 10327 22919 10328 22920 10329 22921 8 22922 8 22923 10330 22924 10331 22925 8 22926 8 22927 10332 22928 8211 22929 10333 22930 8 22931 8 22932 8 22933 10334 22934 10335 22935 10336 22936 10337 22937 10338 22938 10339 22939 10340 22940 10341 22941 10342 22942 10343 22943 8 22944 10344 22945 10345 22946 10346 22947 10347 22948 10348 22949 10349 22950 10350 22951 10351 22952 8 22953 10352 22954 10353 22955 8 22956 10354 22957 10355 22958 10356 22959 10357 22960 8 22961 10358 22962 10359 22963 10360 22964 10361 22965 8 22966 10362 22967 10363 22968 10364 22969 10365 22970 10366 22971 8 22972 10367 22973 10368 22974 10369 22975 10370 22976 10371 22977 10372 22978 10373 22979 10374 22980 10375 22981 8 22982 10376 22983 10377 22984 10378 22985 10379 22986 10380 22987 10381 22988 10382 22989 10383 22990 8 22991 10384 22992 10385 22993 8 22994 10386 22995 8 22996 10387 22997 10388 22998 8 22999 10389 23000 8 23001 8 23002 10390 23003 10391 23004 8 23005 10392 23006 10393 23007 10394 23008 8 23009 10395 23010 10396 23011 10397 23012 10398 23013 8 23014 8 23015 10399 23016 10400 23017 8 23018 10401 23019 8 23020 10402 23021 10403 23022 10404 23023 10405 23024 8 23025 10406 23026 10407 23027 10408 23028 10409 23029 10410 23030 10411 23031 8 23032 10412 23033 10413 23034 8 23035 10414 23036 10415 23037 10416 23038 10417 23039 10418 23040 10419 23041 8 23042 10420 23043 10421 23044 10422 23045 10423 23046 8 23047 10424 23048 10425 23049 10426 23050 8 23051 10427 23052 8 23053 10428 23054 10429 23055 10430 23056 8 23057 10431 23058 10432 23059 8 23060 8 23061 10433 23062 8 23063 10434 23064 8 23065 10435 23066 8 23067 10436 23068 8 23069 10437 23070 10438 23071 10439 23072 10440 23073 10441 23074 10442 23075 10443 23076 10444 23077 10445 23078 8 23079 8 23080 10446 23081 10447 23082 8 23083 8 23084 10448 23085 10449 23086 10450 23087 10451 23088 10452 23089 10453 23090 10454 23091 10455 23092 8 23093 8 23094 10456 23095 10457 23096 8 23097 10458 23098 8 23099 10459 23100 10460 23101 10461 23102 10462 23103 10463 23104 8 23105 10464 23106 10465 23107 8 23108 10466 23109 10467 23110 10468 23111 8 23112 10469 23113 8 23114 8 23115 10470 23116 8 23117 10471 23118 10472 23119 10473 23120 10474 23121 10475 23122 10476 23123 10477 23124 10478 23125 10479 23126 10480 23127 10481 23128 10482 23129 10483 23130 10484 23131 8 23132 10485 23133 10486 23134 10487 23135 10488 23136 10489 23137 10490 23138 10491 23139 10492 23140 10493 23141 10494 23142 8 23143 10495 23144 8 23145 10496 23146 8 23147 10497 23148 8 23149 10498 23150 10499 23151 8 23152 8 23153 10500 23154 8 23155 8 23156 8 23157 8 23158 10501 23159 10502 23160 8 23161 10503 23162 10504 23163 10505 23164 10506 23165 10507 23166 8 23167 10508 23168 10509 23169 10510 23170 10511 23171 10512 23172 10513 23173 10514 23174 10515 23175 10516 23176 8 23177 10517 23178 10518 23179 10519 23180 10520 23181 10521 23182 10522 23183 10523 23184 10524 23185 8 23186 10525 23187 10526 23188 8 23189 10527 23190 8 23191 10528 23192 10529 23193 8 23194 10530 23195 8 23196 8 23197 10531 23198 10532 23199 8 23200 10533 23201 10534 23202 10535 23203 8 23204 10536 23205 10537 23206 10538 23207 10539 23208 8 23209 8 23210 10540 23211 10541 23212 8 23213 10542 23214 8 23215 10543 23216 10544 23217 10545 23218 10546 23219 8 23220 10547 23221 10548 23222 10549 23223 10550 23224 10551 23225 10552 23226 8 23227 10553 23228 10554 23229 8 23230 10555 23231 10556 23232 10557 23233 10558 23234 10559 23235 10560 23236 8 23237 10561 23238 10562 23239 10563 23240 10564 23241 8 23242 10565 23243 10566 23244 10567 23245 8 23246 10568 23247 8 23248 10569 23249 10570 23250 10571 23251 8 23252 10572 23253 10573 23254 8 23255 8 23256 10574 23257 8 23258 10575 23259 8 23260 10576 23261 8 23262 10577 23263 8 23264 10578 23265 10579 23266 10580 23267 10581 23268 10582 23269 10583 23270 10584 23271 10585 23272 10586 23273 8 23274 8 23275 10587 23276 10588 23277 8 23278 8 23279 10589 23280 10590 23281 10591 23282 10592 23283 10593 23284 10594 23285 10595 23286 10596 23287 8 23288 8 23289 10597 23290 10598 23291 8 23292 10599 23293 8 23294 10600 23295 10601 23296 10602 23297 10603 23298 10604 23299 8 23300 10605 23301 10606 23302 10607 23303 10608 23304 8 23305 8 23306 10609 23307 10610 23308 8 23309 10611 23310 8 23311 10612 23312 8 23313 8 23314 10613 23315 10614 23316 10615 23317 10616 23318 10617 23319 10618 23320 10619 23321 10620 23322 10621 23323 10622 23324 10623 23325 8 23326 8 23327 10624 23328 10625 23329 10626 23330 10627 23331 10628 23332 10629 23333 10630 23334 8 23335 8 23336 10631 23337 10632 23338 8 23339 10633 23340 10634 23341 10635 23342 8 23343 10636 23344 10637 23345 10638 23346 10639 23347 8 23348 8 23349 10640 23350 10641 23351 8 23352 10642 23353 8 23354 10643 23355 10644 23356 10645 23357 10646 23358 8 23359 10647 23360 10648 23361 10649 23362 10650 23363 10651 23364 10652 23365 8 23366 10653 23367 10654 23368 8 23369 10655 23370 10656 23371 10657 23372 10658 23373 10659 23374 10660 23375 8 23376 10661 23377 10662 23378 10663 23379 10664 23380 8 23381 10665 23382 10666 23383 10667 23384 10668 23385 8 23386 10669 23387 10670 23388 10671 23389 8 23390 10672 23391 8 23392 10673 23393 10674 23394 10675 23395 8 23396 10676 23397 10677 23398 8 23399 8 23400 10678 23401 8 23402 10679 23403 8 23404 10680 23405 10681 23406 10682 23407 10683 23408 10684 23409 10685 23410 8 23411 3284 23412 10686 23413 10687 23414 10688 23415 10689 23416 8 23417 7709 23418 10690 23419 10691 23420 8 23421 10692 23422 8 23423 10693 23424 8 23425 10694 23426 10695 23427 10696 23428 10697 23429 10698 23430 10699 23431 10700 23432 10701 23433 10702 23434 8 23435 8 23436 10703 23437 10704 23438 8 23439 8 23440 10705 23441 10706 23442 10707 23443 10708 23444 10709 23445 10710 23446 10711 23447 10712 23448 8 23449 8 23450 10713 23451 10714 23452 8 23453 10715 23454 10716 23455 10717 23456 10718 23457 10719 23458 10720 23459 10721 23460 8 23461 10722 23462 8 23463 8 23464 8 23465 8 23466 8 23467 8 23468 8 23469 10723 23470 8 23471 10724 23472 8 23473 8 23474 10725 23475 10726 23476 10727 23477 10728 23478 8 23479 8 23480 8 23481 8 23482 8 23483 8 23484 8 23485 8 23486 8 23487 8 23488 8 23489 10729 23490 8 23491 10730 23492 8 23493 10731 23494 8 23495 8 23496 8 23497 8 23498 8 23499 8 23500 8 23501 8 23502 8 23503 8 23504 10732 23505 8 23506 8 23507 8 23508 8 23509 8 23510 8 23511 8 23512 8 23513 10733 23514 8 23515 10734 23516 10735 23517 10736 23518 10737 23519 10738 23520 8 23521 10739 23522 10740 23523 10741 23524 10742 23525 8 23526 8 23527 10743 23528 10744 23529 8 23530 10745 23531 8 23532 10746 23533 10747 23534 10748 23535 10749 23536 10750 23537 10751 23538 8 23539 10752 23540 8 23541 10753 23542 10754 23543 8 23544 10755 23545 8 23546 10756 23547 8 23548 8 23549 8 23550 10757 23551 10758 23552 8 23553 10759 23554 10760 23555 10761 23556 8 23557 10762 23558 10763 23559 10764 23560 10765 23561 8 23562 8 23563 10766 23564 10767 23565 8 23566 10768 23567 8 23568 10769 23569 10770 23570 10771 23571 10772 23572 8 23573 10773 23574 10774 23575 10775 23576 10776 23577 10777 23578 10778 23579 8 23580 10779 23581 10780 23582 8 23583 10781 23584 10782 23585 10783 23586 10784 23587 10785 23588 10786 23589 8 23590 10787 23591 10788 23592 10789 23593 10790 23594 8 23595 10791 23596 10792 23597 10793 23598 8 23599 10794 23600 8 23601 10795 23602 10796 23603 10797 23604 8 23605 10798 23606 10799 23607 8 23608 8 23609 10800 23610 8 23611 10801 23612 8 23613 10802 23614 10803 23615 10804 23616 8 23617 10805 23618 8 23619 10806 23620 8 23621 10807 23622 10808 23623 10809 23624 10810 23625 10811 23626 10812 23627 10813 23628 10814 23629 10815 23630 8 23631 8 23632 10816 23633 10817 23634 8 23635 8 23636 10818 23637 10819 23638 10820 23639 10821 23640 10822 23641 10823 23642 10824 23643 10825 23644 8 23645 8 23646 10826 23647 10827 23648 8 23649 10828 23650 8 23651 10829 23652 10830 23653 10831 23654 10832 23655 10833 23656 8 23657 10834 23658 10835 23659 10836 23660 10837 23661 8 23662 8 23663 10838 23664 10839 23665 8 23666 10840 23667 8 23668 10841 23669 8 23670 10842 23671 10843 23672 10844 23673 10845 23674 8 23675 8 23676 10846 23677 10847 23678 8 23679 10848 23680 8 23681 10849 23682 8 23683 8 23684 10850 23685 10851 23686 10852 23687 10853 23688 10854 23689 10855 23690 10856 23691 10857 23692 10858 23693 10859 23694 10860 23695 8 23696 8 23697 10861 23698 10862 23699 10863 23700 10864 23701 10865 23702 10866 23703 10867 23704 8 23705 8 23706 10868 23707 10869 23708 8 23709 10870 23710 10871 23711 10872 23712 8 23713 10873 23714 10874 23715 10875 23716 10876 23717 8 23718 8 23719 10877 23720 10878 23721 8 23722 10879 23723 8 23724 10880 23725 10881 23726 10882 23727 10883 23728 8 23729 10884 23730 10885 23731 10886 23732 10887 23733 10888 23734 10889 23735 8 23736 10890 23737 10891 23738 8 23739 10892 23740 10893 23741 10894 23742 10895 23743 10896 23744 10897 23745 8 23746 10898 23747 10899 23748 10900 23749 10901 23750 8 23751 10902 23752 10903 23753 10904 23754 10905 23755 8 23756 10906 23757 10907 23758 10908 23759 8 23760 10909 23761 8 23762 10910 23763 10911 23764 10912 23765 8 23766 10913 23767 10914 23768 8 23769 8 23770 10915 23771 8 23772 10916 23773 8 23774 10917 23775 10918 23776 10919 23777 10920 23778 10921 23779 8 23780 10922 23781 10923 23782 10924 23783 8 23784 10925 23785 10926 23786 8 23787 10927 23788 8 23789 10928 23790 8 23791 10929 23792 10930 23793 10931 23794 10932 23795 10933 23796 10934 23797 10935 23798 10936 23799 10937 23800 8 23801 8 23802 10938 23803 10939 23804 8 23805 8 23806 10940 23807 10941 23808 10942 23809 10943 23810 10944 23811 10945 23812 10946 23813 10947 23814 8 23815 8 23816 10948 23817 10949 23818 8 23819 10950 23820 10951 23821 10952 23822 10953 23823 10954 23824 10955 23825 8 23826 8 23827 10956 23828 8 23829 10957 23830 10958 23831 10959 23832 10960 23833 10961 23834 8 23835 10962 23836 10963 23837 10964 23838 10965 23839 8 23840 8 23841 10966 23842 10967 23843 8 23844 10968 23845 8 23846 10969 23847 10970 23848 10971 23849 10972 23850 10973 23851 10974 23852 8 23853 10975 23854 8 23855 1122 23856 9517 23857 8211 23858 10976 23859 8 23860 10977 23861 10978 23862 10979 23863 10980 23864 10981 23865 8 23866 10982 23867 10983 23868 10984 23869 10985 23870 8 23871 8 23872 10986 23873 10987 23874 8 23875 10988 23876 8 23877 10989 23878 8 23879 8 23880 10990 23881 10991 23882 10992 23883 10993 23884 10994 23885 10995 23886 10996 23887 10997 23888 10998 23889 10999 23890 11000 23891 8 23892 8 23893 11001 23894 11002 23895 11003 23896 11004 23897 11005 23898 11006 23899 11007 23900 8 23901 8 23902 11008 23903 11009 23904 8 23905 11010 23906 11011 23907 11012 23908 8 23909 11013 23910 11014 23911 11015 23912 11016 23913 8 23914 8 23915 11017 23916 11018 23917 8 23918 11019 23919 8 23920 11020 23921 11021 23922 11022 23923 11023 23924 8 23925 11024 23926 11025 23927 11026 23928 11027 23929 11028 23930 11029 23931 8 23932 11030 23933 11031 23934 8 23935 11032 23936 11033 23937 11034 23938 11035 23939 11036 23940 11037 23941 8 23942 11038 23943 11039 23944 11040 23945 11041 23946 8 23947 11042 23948 11043 23949 11044 23950 11045 23951 8 23952 11046 23953 11047 23954 11048 23955 8 23956 11049 23957 8 23958 11050 23959 11051 23960 11052 23961 8 23962 11053 23963 11054 23964 8 23965 8 23966 11055 23967 8 23968 11056 23969 8 23970 11057 23971 11058 23972 11059 23973 11060 23974 11061 23975 8 23976 11062 23977 11063 23978 11064 23979 8 23980 11065 23981 11066 23982 8 23983 11067 23984 8 23985 11068 23986 8 23987 11069 23988 11070 23989 11071 23990 11072 23991 11073 23992 11074 23993 11075 23994 11076 23995 11077 23996 8 23997 8 23998 11078 23999 11079 24000 8 24001 8 24002 11080 24003 11081 24004 11082 24005 11083 24006 11084 24007 11085 24008 11086 24009 11087 24010 8 24011 8 24012 11088 24013 11089 24014 8 24015 11090 24016 11091 24017 11092 24018 11093 24019 11094 24020 11095 24021 8 24022 8 24023 11096 24024 8 24025 11097 24026 11098 24027 11099 24028 11100 24029 11101 24030 8 24031 11102 24032 11103 24033 11104 24034 11105 24035 8 24036 8 24037 11106 24038 11107 24039 8 24040 11108 24041 8 24042 11109 24043 11110 24044 11111 24045 11112 24046 11113 24047 11114 24048 8 24049 11115 24050 8 24051 11116 24052 11117 24053 8 24054 8 24055 8 24056 11118 24057 11119 24058 11120 24059 11121 24060 8 24061 8 24062 11122 24063 8 24064 11123 24065 11124 24066 8 24067 8 24068 11125 24069 8 24070 8 24071 11126 24072 8 24073 8 24074 11127 24075 11128 24076 8 24077 8 24078 11129 24079 11130 24080 8 24081 8 24082 11131 24083 11132 24084 8 24085 8 24086 8 24087 11133 24088 11134 24089 11135 24090 11136 24091 8 24092 8 24093 11137 24094 8 24095 11138 24096 11139 24097 8 24098 8 24099 11140 24100 8 24101 8 24102 11141 24103 8 24104 8 24105 11142 24106 11143 24107 8 24108 8 24109 11144 24110 11145 24111 8 24112 11146 24113 8 24114 11147 24115 8 24116 11148 24117 11149 24118 8 24119 8 24120 8 24121 11150 24122 8 24123 8 24124 11151 24125 11152 24126 11153 24127 11154 24128 8 24129 8 24130 11155 24131 8 24132 11156 24133 11157 24134 8 24135 8 24136 11158 24137 8 24138 11159 24139 8 24140 11160 24141 11161 24142 11162 24143 8706 24144 8706 24145 11163 24146 11164 24147 11165 24148 11166 24149 11167 24150 11168 24151 11169 24152 11170 24153 11171 24154 11172 24155 8 24156 11173 24157 8 24158 11174 24159 8 24160 8 24161 11175 24162 11176 24163 11177 24164 11178 24165 8 24166 8 24167 11179 24168 11180 24169 11181 24170 11182 24171 11183 24172 8 24173 11184 24174 8 24175 11185 24176 8 24177 8 24178 8 24179 8 24180 11186 24181 8 24182 11187 24183 11188 24184 8 24185 11189 24186 11190 24187 11191 24188 11192 24189 11193 24190 8 24191 11194 24192 11195 24193 8 24194 11196 24195 8 24196 11197 24197 11198 24198 11199 24199 11200 24200 11201 24201 11202 24202 11203 24203 8 24204 8 24205 11204 24206 8 24207 8 24208 11205 24209 11206 24210 11207 24211 8 24212 8 24213 8 24214 8 24215 11208 24216 8 24217 8 24218 8 24219 8 24220 8 24221 11209 24222 11210 24223 11211 24224 11212 24225 11213 24226 11214 24227 11215 24228 11216 24229 11217 24230 11218 24231 11219 24232 11220 24233 11221 24234 11222 24235 11223 24236 11224 24237 11225 24238 11226 24239 11227 24240 11228 24241 11229 24242 11230 24243 11231 24244 11232 24245 11233 24246 11234 24247 11235 24248 11236 24249 11237 24250 11238 24251 11239 24252 11240 24253 11241 24254 11242 24255 11243 24256 11212 24257 11212 24258 11212 24259 11212 24260 11212 24261 11212 24262 11212 24263 11212 24264 11212 24265 11212 24266 11212 24267 11212 24268 11212 24269 11212 24270 11212 24271 11212 24272 11244 24273 11245 24274 8 24275 8 24276 8 24277 8 24278 8 24279 8 24280 11246 24281 8 24282 8 24283 11247 24284 8 24285 11248 24286 11249 24287 11250 24288 11251 24289 11252 24290 8 24291 8 24292 8 24293 8 24294 8 24295 11253 24296 11254 24297 8 24298 8 24299 11255 24300 8 24301 11256 24302 8 24303 8 24304 11257 24305 11258 24306 8 24307 11212 24308 11212 24309 11212 24310 11212 24311 11212 24312 11259 24313 11260 24314 11261 24315 11212 24316 11212 24317 11212 24318 11212 24319 11212 24320 11212 24321 11212 24322 11212 24323 11212 24324 11212 24325 11262 24326 8 24327 8 24328 8 24329 8 24330 8 24331 8 24332 8 24333 8 24334 8 24335 8 24336 8 24337 8 24338 8 24339 8 24340 8 24341 8 24342 8 24343 8 24344 8 24345 8 24346 8 24347 8 24348 8 24349 8 24350 8 24351 8 24352 8 24353 8 24354 8 24355 8 24356 8 24357 8 24358 8 24359 8 24360 8 24361 8 24362 8 24363 8 24364 8 24365 8 24366 8 24367 8 24368 8 24369 8 24370 8 24371 8 24372 8 24373 8 24374 8 24375 8 24376 8 24377 8 24378 8 24379 8 24380 8 24381 8 24382 8 24383 8 24384 8 24385 8 24386 8 24387 8 24388 8 24389 8 24390 8 24391 8 24392 8 24393 8 24394 8 24395 8 24396 8 24397 8 24398 8 24399 8 24400 8 24401 8 24402 8 24403 8 24404 8 24405 8 24406 11263 24407 8 24408 8 24409 8 24410 11264 24411 11265 24412 8 24413 11266 24414 8 24415 8 24416 11267 24417 11268 24418 11269 24419 8 24420 8 24421 11270 24422 11271 24423 11272 24424 11273 24425 8 24426 11274 24427 8 24428 11275 24429 11276 24430 11277 24431 11278 24432 11279 24433 11280 24434 11281 24435 11282 24436 11283 24437 11212 24438 11212 24439 11212 24440 11212 24441 11212 24442 11212 24443 11212 24444 11212 24445 11212 24446 11212 24447 11212 24448 11212 24449 11212 24450 11284 24451 11212 24452 11285 24453 11286 24454 8 24455 11287 24456 11288 24457 8 24458 11289 24459 8 24460 11290 24461 8 24462 8 24463 8 24464 11291 24465 11292 24466 11293 24467 8 24468 11294 24469 8 24470 8 24471 11295 24472 11296 24473 11297 24474 8 24475 8 24476 8 24477 11298 24478 11299 24479 8 24480 8 24481 8 24482 11300 24483 11301 24484 11302 24485 11303 24486 8 24487 8 24488 11304 24489 8 24490 11305 24491 11306 24492 8 24493 8 24494 11307 24495 8 24496 8 24497 8 24498 11308 24499 11309 24500 8 24501 8 24502 11310 24503 11311 24504 11312 24505 11313 24506 11314 24507 11315 24508 11316 24509 11317 24510 11318 24511 11319 24512 11320 24513 8 24514 11321 24515 11322 24516 11323 24517 11324 24518 11325 24519 11326 24520 11327 24521 11328 24522 11329 24523 11330 24524 8 24525 11331 24526 8706 24527 11332 24528 8 24529 11333 24530 8 24531 11334 24532 11335 24533 8 24534 8 24535 8 24536 11336 24537 8 24538 8 24539 11337 24540 11338 24541 11339 24542 11340 24543 8 24544 8 24545 11341 24546 8 24547 11342 24548 11343 24549 8 24550 8 24551 11344 24552 8 24553 11345 24554 8 24555 11346 24556 11347 24557 11348 24558 11212 24559 11212 24560 11349 24561 11350 24562 11212 24563 11212 24564 11212 24565 11212 24566 8 24567 11351 24568 8 24569 11352 24570 8 24571 8 24572 11212 24573 11212 24574 8 24575 11212 24576 11212 24577 11353 24578 8 24579 8 24580 8 24581 8 24582 8 24583 8 24584 11354 24585 11355 24586 11356 24587 11357 24588 8 24589 11358 24590 8 24591 11359 24592 11360 24593 11361 24594 11362 24595 11363 24596 8 24597 8 24598 8 24599 8 24600 8 24601 11364 24602 8 24603 8 24604 11365 24605 11366 24606 11367 24607 11368 24608 11369 24609 11370 24610 11371 24611 8 24612 11372 24613 8 24614 11373 24615 11374 24616 11375 24617 11376 24618 11377 24619 8 24620 11378 24621 11379 24622 11380 24623 11381 24624 11212 24625 11382 24626 11383 24627 11384 24628 8 24629 11212 24630 11212 24631 11212 24632 8 24633 11385 24634 8 24635 8 24636 8 24637 11386 24638 11387 24639 11388 24640 8 24641 8 24642 11389 24643 11390 24644 8 24645 11391 24646 11392 24647 8 24648 11212 24649 8 24650 11393 24651 8 24652 11212 24653 8 24654 11394 24655 11395 24656 11396 24657 11397 24658 8 24659 8 24660 11398 24661 8 24662 8 24663 11212 24664 11212 24665 11399 24666 11400 24667 8 24668 11212 24669 11212 24670 11212 24671 11401 24672 8 24673 11402 24674 11403 24675 8 24676 11404 24677 8 24678 11212 24679 8706 24680 8706 24681 8706 24682 8 24683 11405 24684 8 24685 11406 24686 8 24687 8 24688 8706 24689 8706 24690 8 24691 8706 24692 8706 24693 11407 24694 8 24695 8 24696 8 24697 8 24698 8 24699 8 24700 11408 24701 11409 24702 11410 24703 11411 24704 8 24705 11412 24706 8 24707 11413 24708 11414 24709 11415 24710 11416 24711 11417 24712 8 24713 8 24714 8 24715 8 24716 8 24717 11418 24718 8 24719 8 24720 11419 24721 11420 24722 11421 24723 11422 24724 11423 24725 11424 24726 11425 24727 8 24728 11426 24729 8 24730 11427 24731 11428 24732 11429 24733 11430 24734 11431 24735 8 24736 11432 24737 11433 24738 11434 24739 11435 24740 8706 24741 11436 24742 11437 24743 11438 24744 8 24745 8706 24746 8706 24747 8706 24748 8 24749 11439 24750 8 24751 8 24752 8 24753 11440 24754 11441 24755 11442 24756 8 24757 8 24758 11443 24759 11444 24760 8 24761 11445 24762 11446 24763 8 24764 8706 24765 8 24766 11447 24767 8 24768 8706 24769 8 24770 11448 24771 11449 24772 11450 24773 11451 24774 8 24775 8 24776 11452 24777 8 24778 8 24779 8706 24780 8706 24781 11453 24782 11454 24783 8 24784 8706 24785 8706 24786 8706 24787 11455 24788 8 24789 11456 24790 11457 24791 8 24792 11458 24793 8 24794 8706 24795 11459 24796 11460 24797 11461 24798 11462 24799 11463 24800 11464 24801 11465 24802 11466 24803 8 24804 8 24805 11467 24806 11468 24807 8 24808 11469 24809 8 24810 11470 24811 11471 24812 11472 24813 11473 24814 11474 24815 8 24816 11475 24817 11476 24818 11477 24819 11478 24820 8 24821 8 24822 11479 24823 11480 24824 8 24825 11481 24826 8 24827 11482 24828 8 24829 8 24830 11483 24831 11484 24832 11485 24833 11486 24834 11487 24835 11488 24836 11489 24837 11490 24838 11491 24839 11492 24840 11493 24841 8 24842 8 24843 11494 24844 11495 24845 11496 24846 11497 24847 11498 24848 11499 24849 11500 24850 8 24851 8 24852 11501 24853 11502 24854 8 24855 11503 24856 11504 24857 11505 24858 8 24859 11506 24860 11507 24861 11508 24862 11509 24863 8 24864 8 24865 11510 24866 11511 24867 8 24868 11512 24869 8 24870 11513 24871 11514 24872 11515 24873 11516 24874 8 24875 11517 24876 11518 24877 11519 24878 11520 24879 11521 24880 11522 24881 8 24882 11523 24883 11524 24884 8 24885 11525 24886 11526 24887 11527 24888 11528 24889 11529 24890 11530 24891 8 24892 11531 24893 11532 24894 11533 24895 11534 24896 8 24897 11535 24898 11536 24899 11537 24900 11538 24901 8 24902 11539 24903 11540 24904 11541 24905 8 24906 11542 24907 8 24908 11543 24909 11544 24910 11545 24911 8 24912 11546 24913 11547 24914 8 24915 8 24916 11548 24917 8 24918 11549 24919 8 24920 11550 24921 11551 24922 11552 24923 11553 24924 11554 24925 8 24926 11555 24927 11556 24928 11557 24929 8 24930 11558 24931 11559 24932 8 24933 11560 24934 8 24935 11561 24936 8 24937 11562 24938 11563 24939 11564 24940 11565 24941 11566 24942 11567 24943 11568 24944 11569 24945 11570 24946 8 24947 8 24948 11571 24949 11572 24950 8 24951 8 24952 11573 24953 11574 24954 11575 24955 11576 24956 11577 24957 11578 24958 11579 24959 11580 24960 8 24961 8 24962 11581 24963 11582 24964 8 24965 11583 24966 11584 24967 11585 24968 11586 24969 11587 24970 11588 24971 8 24972 11589 24973 8 24974 11590 24975 11591 24976 11592 24977 11593 24978 11594 24979 8 24980 11595 24981 11596 24982 11597 24983 11598 24984 8 24985 8 24986 11599 24987 11600 24988 8 24989 11601 24990 8 24991 11602 24992 11603 24993 11604 24994 11605 24995 11606 24996 11607 24997 8 24998 11608 24999 8 25000 8062 25001 11609 25002 11610 25003 11611 25004 11612 25005 11613 25006 11614 25007 11615 25008 11616 25009 11617 25010 11618 25011 8 25012 11619 25013 8 25014 11620 25015 8 25016 8 25017 11621 25018 11622 25019 11623 25020 11624 25021 8 25022 8 25023 11625 25024 11626 25025 11627 25026 11628 25027 11629 25028 8 25029 11630 25030 8 25031 11631 25032 8 25033 8 25034 8 25035 8 25036 11632 25037 8 25038 11633 25039 11634 25040 8 25041 11635 25042 11636 25043 11637 25044 11638 25045 11639 25046 8 25047 11640 25048 11641 25049 8 25050 11642 25051 8 25052 11643 25053 11644 25054 11645 25055 11646 25056 11647 25057 11648 25058 11649 25059 8 25060 8 25061 11650 25062 8 25063 8 25064 11651 25065 11652 25066 11653 25067 8 25068 8 25069 8 25070 8 25071 11654 25072 8 25073 8 25074 8 25075 8 25076 8 25077 11655 25078 11656 25079 11657 25080 11658 25081 11659 25082 11660 25083 11661 25084 11662 25085 11663 25086 11664 25087 11665 25088 11666 25089 11667 25090 11668 25091 11669 25092 11670 25093 11671 25094 11672 25095 11673 25096 11674 25097 11675 25098 11676 25099 11677 25100 11678 25101 11679 25102 11680 25103 11681 25104 11682 25105 11683 25106 11684 25107 11685 25108 11686 25109 11687 25110 11688 25111 11689 25112 11658 25113 11658 25114 11658 25115 11658 25116 11658 25117 11658 25118 11658 25119 11658 25120 11658 25121 11658 25122 11658 25123 11658 25124 11658 25125 11658 25126 11658 25127 11658 25128 11690 25129 11691 25130 8 25131 8 25132 8 25133 8 25134 8 25135 8 25136 11692 25137 8 25138 8 25139 11693 25140 8 25141 11694 25142 11695 25143 11696 25144 11697 25145 11698 25146 8 25147 8 25148 8 25149 8 25150 8 25151 11699 25152 11700 25153 8 25154 8 25155 11701 25156 8 25157 11702 25158 8 25159 8 25160 11703 25161 11704 25162 8 25163 11658 25164 11658 25165 11658 25166 11658 25167 11658 25168 11705 25169 11706 25170 11707 25171 11658 25172 11658 25173 11658 25174 11658 25175 11658 25176 11658 25177 11658 25178 11658 25179 11658 25180 11658 25181 11708 25182 8 25183 8 25184 8 25185 8 25186 8 25187 8 25188 8 25189 8 25190 8 25191 8 25192 8 25193 8 25194 8 25195 8 25196 8 25197 8 25198 8 25199 8 25200 8 25201 8 25202 8 25203 8 25204 8 25205 8 25206 8 25207 8 25208 8 25209 8 25210 8 25211 8 25212 8 25213 8 25214 8 25215 8 25216 8 25217 8 25218 8 25219 8 25220 8 25221 8 25222 8 25223 8 25224 8 25225 8 25226 8 25227 8 25228 8 25229 8 25230 8 25231 8 25232 8 25233 8 25234 8 25235 8 25236 8 25237 8 25238 8 25239 8 25240 8 25241 8 25242 8 25243 8 25244 8 25245 8 25246 8 25247 8 25248 8 25249 8 25250 8 25251 8 25252 8 25253 8 25254 8 25255 8 25256 8 25257 8 25258 8 25259 8 25260 8 25261 8 25262 11709 25263 8 25264 8 25265 8 25266 11710 25267 11711 25268 8 25269 11712 25270 8 25271 8 25272 11713 25273 11714 25274 11715 25275 8 25276 8 25277 11716 25278 11717 25279 11718 25280 11719 25281 8 25282 11720 25283 8 25284 11721 25285 11722 25286 11723 25287 11724 25288 11725 25289 11726 25290 11727 25291 11728 25292 11729 25293 11658 25294 11658 25295 11658 25296 11658 25297 11658 25298 11658 25299 11658 25300 11658 25301 11658 25302 11658 25303 11658 25304 11658 25305 11658 25306 11730 25307 11658 25308 11731 25309 11732 25310 8 25311 11733 25312 11734 25313 8 25314 11735 25315 8 25316 11736 25317 8 25318 8 25319 8 25320 11737 25321 11738 25322 11739 25323 8 25324 11740 25325 8 25326 8 25327 11741 25328 11742 25329 11743 25330 8 25331 8 25332 8 25333 11744 25334 11745 25335 8 25336 8 25337 8 25338 11746 25339 11747 25340 11748 25341 11749 25342 8 25343 8 25344 11750 25345 8 25346 11751 25347 11752 25348 8 25349 8 25350 11753 25351 8 25352 8 25353 8 25354 11754 25355 11755 25356 8 25357 8 25358 11756 25359 11757 25360 11758 25361 11759 25362 11760 25363 11761 25364 11762 25365 11763 25366 11764 25367 11765 25368 11766 25369 8 25370 11767 25371 11768 25372 11769 25373 11770 25374 11771 25375 11772 25376 11773 25377 11774 25378 11775 25379 11776 25380 11777 25381 11778 25382 11779 25383 11780 25384 11781 25385 11782 25386 11783 25387 11784 25388 11785 25389 8 25390 11786 25391 8 25392 11787 25393 8 25394 8 25395 11788 25396 11789 25397 11790 25398 11791 25399 8 25400 8 25401 11792 25402 11793 25403 11794 25404 11795 25405 11796 25406 8 25407 11797 25408 8 25409 11798 25410 8 25411 8 25412 8 25413 8 25414 11799 25415 8 25416 11800 25417 11801 25418 8 25419 11802 25420 11803 25421 11804 25422 11805 25423 11806 25424 8 25425 11807 25426 11808 25427 8 25428 11809 25429 8 25430 11810 25431 11811 25432 11812 25433 11813 25434 11814 25435 11815 25436 11816 25437 8 25438 8 25439 11817 25440 8 25441 8 25442 11818 25443 11819 25444 11820 25445 8 25446 8 25447 8 25448 8 25449 11821 25450 8 25451 8 25452 8 25453 8 25454 8 25455 11822 25456 11823 25457 11824 25458 11825 25459 11826 25460 11827 25461 11828 25462 11829 25463 11830 25464 11831 25465 11832 25466 11833 25467 11834 25468 11835 25469 11836 25470 11837 25471 11838 25472 11839 25473 11840 25474 11841 25475 11842 25476 11843 25477 11844 25478 11845 25479 11846 25480 11847 25481 11848 25482 11849 25483 11850 25484 11851 25485 11852 25486 11853 25487 11854 25488 11855 25489 11856 25490 11825 25491 11825 25492 11825 25493 11825 25494 11825 25495 11825 25496 11825 25497 11825 25498 11825 25499 11825 25500 11825 25501 11825 25502 11825 25503 11825 25504 11825 25505 11825 25506 11857 25507 11858 25508 8 25509 8 25510 8 25511 8 25512 8 25513 8 25514 11859 25515 8 25516 8 25517 11860 25518 8 25519 11861 25520 11862 25521 11863 25522 11864 25523 11865 25524 8 25525 8 25526 8 25527 8 25528 8 25529 11866 25530 11867 25531 8 25532 8 25533 11868 25534 8 25535 11869 25536 8 25537 8 25538 11870 25539 11871 25540 8 25541 11825 25542 11825 25543 11825 25544 11825 25545 11825 25546 11872 25547 11873 25548 11874 25549 11825 25550 11825 25551 11825 25552 11825 25553 11825 25554 11825 25555 11825 25556 11825 25557 11825 25558 11825 25559 11875 25560 8 25561 8 25562 8 25563 8 25564 8 25565 8 25566 8 25567 8 25568 8 25569 8 25570 8 25571 8 25572 8 25573 8 25574 8 25575 8 25576 8 25577 8 25578 8 25579 8 25580 8 25581 8 25582 8 25583 8 25584 8 25585 8 25586 8 25587 8 25588 8 25589 8 25590 8 25591 8 25592 8 25593 8 25594 8 25595 8 25596 8 25597 8 25598 8 25599 8 25600 8 25601 8 25602 8 25603 8 25604 8 25605 8 25606 8 25607 8 25608 8 25609 8 25610 8 25611 8 25612 8 25613 8 25614 8 25615 8 25616 8 25617 8 25618 8 25619 8 25620 8 25621 8 25622 8 25623 8 25624 8 25625 8 25626 8 25627 8 25628 8 25629 8 25630 8 25631 8 25632 8 25633 8 25634 8 25635 8 25636 8 25637 8 25638 8 25639 8 25640 11876 25641 8 25642 8 25643 8 25644 11877 25645 11878 25646 8 25647 11879 25648 8 25649 8 25650 11880 25651 11881 25652 11882 25653 8 25654 8 25655 11883 25656 11884 25657 11885 25658 11886 25659 8 25660 11887 25661 8 25662 11888 25663 11889 25664 11890 25665 11891 25666 11892 25667 11893 25668 11894 25669 11895 25670 11896 25671 11825 25672 11825 25673 11825 25674 11825 25675 11825 25676 11825 25677 11825 25678 11825 25679 11825 25680 11825 25681 11825 25682 11825 25683 11825 25684 11897 25685 11825 25686 11898 25687 11899 25688 8 25689 11900 25690 11901 25691 8 25692 11902 25693 8 25694 11903 25695 8 25696 8 25697 8 25698 11904 25699 11905 25700 11906 25701 8 25702 11907 25703 8 25704 8 25705 11908 25706 11909 25707 11910 25708 8 25709 8 25710 8 25711 11911 25712 11912 25713 8 25714 8 25715 8 25716 11913 25717 11914 25718 11915 25719 11916 25720 8 25721 8 25722 11917 25723 8 25724 11918 25725 11919 25726 8 25727 8 25728 11920 25729 8 25730 8 25731 8 25732 11921 25733 11922 25734 8 25735 8 25736 11923 25737 11924 25738 11925 25739 11926 25740 11927 25741 11928 25742 11929 25743 11930 25744 11931 25745 11932 25746 11933 25747 8 25748 11934 25749 11935 25750 11936 25751 11937 25752 11938 25753 11939 25754 11940 25755 11941 25756 11942 25757 11943 25758 8 25759 8 25760 11944 25761 8 25762 11945 25763 8 25764 11946 25765 11947 25766 8 25767 8 25768 8 25769 11948 25770 8 25771 8 25772 11949 25773 11950 25774 11951 25775 11952 25776 8 25777 8 25778 11953 25779 8 25780 11954 25781 11955 25782 8 25783 8 25784 11956 25785 8 25786 11957 25787 8 25788 11958 25789 11959 25790 11960 25791 11825 25792 11825 25793 11961 25794 11962 25795 11963 25796 11964 25797 11965 25798 11966 25799 11967 25800 11968 25801 11969 25802 11970 25803 8 25804 11971 25805 8 25806 11972 25807 8 25808 8 25809 11973 25810 11974 25811 11975 25812 11976 25813 8 25814 8 25815 11977 25816 11978 25817 11979 25818 11980 25819 11981 25820 8 25821 11982 25822 8 25823 11983 25824 8 25825 8 25826 8 25827 8 25828 11984 25829 8 25830 11985 25831 11986 25832 8 25833 11987 25834 11988 25835 11989 25836 11990 25837 11991 25838 8 25839 11992 25840 11993 25841 8 25842 11994 25843 8 25844 11995 25845 11996 25846 11997 25847 11998 25848 11999 25849 12000 25850 12001 25851 8 25852 8 25853 12002 25854 8 25855 8 25856 12003 25857 12004 25858 12005 25859 8 25860 8 25861 8 25862 8 25863 12006 25864 8 25865 8 25866 8 25867 8 25868 8 25869 12007 25870 12008 25871 12009 25872 12010 25873 12011 25874 12012 25875 12013 25876 12014 25877 12015 25878 12016 25879 12017 25880 12018 25881 12019 25882 12020 25883 12021 25884 12022 25885 12023 25886 12024 25887 12025 25888 12026 25889 12027 25890 12028 25891 12029 25892 12030 25893 12031 25894 12032 25895 12033 25896 12034 25897 12035 25898 12036 25899 12037 25900 12038 25901 12039 25902 12040 25903 12041 25904 12010 25905 12010 25906 12010 25907 12010 25908 12010 25909 12010 25910 12010 25911 12010 25912 12010 25913 12010 25914 12010 25915 12010 25916 12010 25917 12010 25918 12010 25919 12010 25920 12042 25921 12043 25922 8 25923 8 25924 8 25925 8 25926 8 25927 8 25928 12044 25929 8 25930 8 25931 12045 25932 8 25933 12046 25934 12047 25935 12048 25936 12049 25937 12050 25938 8 25939 8 25940 8 25941 8 25942 8 25943 12051 25944 12052 25945 8 25946 8 25947 12053 25948 8 25949 12054 25950 8 25951 8 25952 12055 25953 12056 25954 8 25955 12010 25956 12010 25957 12010 25958 12010 25959 12010 25960 12057 25961 12058 25962 12059 25963 12010 25964 12010 25965 12010 25966 12010 25967 12010 25968 12010 25969 12010 25970 12010 25971 12010 25972 12010 25973 12060 25974 8 25975 8 25976 8 25977 8 25978 8 25979 8 25980 8 25981 8 25982 8 25983 8 25984 8 25985 8 25986 8 25987 8 25988 8 25989 8 25990 8 25991 8 25992 8 25993 8 25994 8 25995 8 25996 8 25997 8 25998 8 25999 8 26000 8 26001 8 26002 8 26003 8 26004 8 26005 8 26006 8 26007 8 26008 8 26009 8 26010 8 26011 8 26012 8 26013 8 26014 8 26015 8 26016 8 26017 8 26018 8 26019 8 26020 8 26021 8 26022 8 26023 8 26024 8 26025 8 26026 8 26027 8 26028 8 26029 8 26030 8 26031 8 26032 8 26033 8 26034 8 26035 8 26036 8 26037 8 26038 8 26039 8 26040 8 26041 8 26042 8 26043 8 26044 8 26045 8 26046 8 26047 8 26048 8 26049 8 26050 8 26051 8 26052 8 26053 8 26054 12061 26055 8 26056 8 26057 8 26058 12062 26059 12063 26060 8 26061 12064 26062 8 26063 8 26064 12065 26065 12066 26066 12067 26067 8 26068 8 26069 12068 26070 12069 26071 12070 26072 12071 26073 8 26074 12072 26075 8 26076 12073 26077 12074 26078 12075 26079 12076 26080 12077 26081 12078 26082 12079 26083 12080 26084 12081 26085 12010 26086 12010 26087 12010 26088 12010 26089 12010 26090 12010 26091 12010 26092 12010 26093 12010 26094 12010 26095 12010 26096 12010 26097 12010 26098 12082 26099 12010 26100 12083 26101 12084 26102 8 26103 12085 26104 12086 26105 8 26106 12087 26107 8 26108 12088 26109 8 26110 8 26111 8 26112 12089 26113 12090 26114 12091 26115 8 26116 12092 26117 8 26118 8 26119 12093 26120 12094 26121 12095 26122 8 26123 8 26124 8 26125 12096 26126 12097 26127 8 26128 8 26129 8 26130 12098 26131 12099 26132 12100 26133 12101 26134 8 26135 8 26136 12102 26137 8 26138 12103 26139 12104 26140 8 26141 8 26142 12105 26143 8 26144 8 26145 8 26146 12106 26147 12107 26148 8 26149 8 26150 12108 26151 12109 26152 12110 26153 12111 26154 12112 26155 12113 26156 12114 26157 12115 26158 12116 26159 12117 26160 12118 26161 8 26162 12119 26163 12120 26164 12121 26165 12122 26166 12123 26167 12124 26168 12125 26169 12126 26170 12127 26171 12128 26172 8 26173 12129 26174 11825 26175 12130 26176 8 26177 12131 26178 8 26179 12132 26180 12133 26181 8 26182 8 26183 8 26184 12134 26185 8 26186 8 26187 12135 26188 12136 26189 12137 26190 12138 26191 8 26192 8 26193 12139 26194 8 26195 12140 26196 12141 26197 8 26198 8 26199 12142 26200 8 26201 12143 26202 8 26203 12144 26204 12145 26205 12146 26206 12010 26207 12010 26208 12147 26209 12148 26210 12010 26211 12010 26212 12010 26213 12010 26214 8 26215 12149 26216 8 26217 12150 26218 8 26219 8 26220 12010 26221 12010 26222 8 26223 12010 26224 12010 26225 12151 26226 8 26227 8 26228 8 26229 8 26230 8 26231 8 26232 12152 26233 12153 26234 12154 26235 12155 26236 8 26237 12156 26238 8 26239 12157 26240 12158 26241 12159 26242 12160 26243 12161 26244 8 26245 8 26246 8 26247 8 26248 8 26249 12162 26250 8 26251 8 26252 12163 26253 12164 26254 12165 26255 12166 26256 12167 26257 12168 26258 12169 26259 8 26260 12170 26261 8 26262 12171 26263 12172 26264 12173 26265 12174 26266 12175 26267 8 26268 12176 26269 12177 26270 12178 26271 12179 26272 12010 26273 12180 26274 12181 26275 12182 26276 8 26277 12010 26278 12010 26279 12010 26280 8 26281 12183 26282 8 26283 8 26284 8 26285 12184 26286 12185 26287 12186 26288 8 26289 8 26290 12187 26291 12188 26292 8 26293 12189 26294 12190 26295 8 26296 12010 26297 8 26298 12191 26299 8 26300 12010 26301 8 26302 12192 26303 12193 26304 12194 26305 12195 26306 8 26307 8 26308 12196 26309 8 26310 8 26311 12010 26312 12010 26313 12197 26314 12198 26315 8 26316 12010 26317 12010 26318 12010 26319 12199 26320 8 26321 12200 26322 12201 26323 8 26324 12202 26325 8 26326 12010 26327 11825 26328 11825 26329 11825 26330 8 26331 12203 26332 8 26333 12204 26334 8 26335 8 26336 11825 26337 11825 26338 8 26339 11825 26340 11825 26341 12205 26342 8 26343 8 26344 8 26345 8 26346 8 26347 8 26348 12206 26349 12207 26350 12208 26351 12209 26352 8 26353 12210 26354 8 26355 12211 26356 12212 26357 12213 26358 12214 26359 12215 26360 8 26361 8 26362 8 26363 8 26364 8 26365 12216 26366 8 26367 8 26368 12217 26369 12218 26370 12219 26371 12220 26372 12221 26373 12222 26374 12223 26375 8 26376 12224 26377 8 26378 12225 26379 12226 26380 12227 26381 12228 26382 12229 26383 8 26384 12230 26385 12231 26386 12232 26387 12233 26388 11825 26389 12234 26390 12235 26391 12236 26392 8 26393 11825 26394 11825 26395 11825 26396 8 26397 12237 26398 8 26399 8 26400 8 26401 12238 26402 12239 26403 12240 26404 8 26405 8 26406 12241 26407 12242 26408 8 26409 12243 26410 12244 26411 8 26412 11825 26413 8 26414 12245 26415 8 26416 11825 26417 8 26418 12246 26419 12247 26420 12248 26421 12249 26422 8 26423 8 26424 12250 26425 8 26426 8 26427 11825 26428 11825 26429 12251 26430 12252 26431 8 26432 11825 26433 11825 26434 11825 26435 12253 26436 8 26437 12254 26438 12255 26439 8 26440 12256 26441 8 26442 11825 26443 12257 26444 12258 26445 8062 26446 8062 26447 12259 26448 8 26449 12260 26450 8 26451 12261 26452 12262 26453 8 26454 8 26455 8 26456 12263 26457 8 26458 8 26459 12264 26460 12265 26461 12266 26462 12267 26463 8 26464 8 26465 12268 26466 8 26467 12269 26468 12270 26469 8 26470 8 26471 12271 26472 8 26473 12272 26474 8 26475 12273 26476 12274 26477 12275 26478 11658 26479 11658 26480 12276 26481 12277 26482 12278 26483 12279 26484 12280 26485 12281 26486 12282 26487 12283 26488 12284 26489 12285 26490 8 26491 12286 26492 8 26493 12287 26494 8 26495 8 26496 12288 26497 12289 26498 12290 26499 12291 26500 8 26501 8 26502 12292 26503 12293 26504 12294 26505 12295 26506 12296 26507 8 26508 12297 26509 8 26510 12298 26511 8 26512 8 26513 8 26514 8 26515 12299 26516 8 26517 12300 26518 12301 26519 8 26520 12302 26521 12303 26522 12304 26523 12305 26524 12306 26525 8 26526 12307 26527 12308 26528 8 26529 12309 26530 8 26531 12310 26532 12311 26533 12312 26534 12313 26535 12314 26536 12315 26537 12316 26538 8 26539 8 26540 12317 26541 8 26542 8 26543 12318 26544 12319 26545 12320 26546 8 26547 8 26548 8 26549 8 26550 12321 26551 8 26552 8 26553 8 26554 8 26555 8 26556 12322 26557 12323 26558 12324 26559 12325 26560 12326 26561 12327 26562 12328 26563 12329 26564 12330 26565 12331 26566 12332 26567 12333 26568 12334 26569 12335 26570 12336 26571 12337 26572 12338 26573 12339 26574 12340 26575 12341 26576 12342 26577 12343 26578 12344 26579 12345 26580 12346 26581 12347 26582 12348 26583 12349 26584 12350 26585 12351 26586 12352 26587 12353 26588 12354 26589 12355 26590 12356 26591 12325 26592 12325 26593 12325 26594 12325 26595 12325 26596 12325 26597 12325 26598 12325 26599 12325 26600 12325 26601 12325 26602 12325 26603 12325 26604 12325 26605 12325 26606 12325 26607 12357 26608 12358 26609 8 26610 8 26611 8 26612 8 26613 8 26614 8 26615 12359 26616 8 26617 8 26618 12360 26619 8 26620 12361 26621 12362 26622 12363 26623 12364 26624 12365 26625 8 26626 8 26627 8 26628 8 26629 8 26630 12366 26631 12367 26632 8 26633 8 26634 12368 26635 8 26636 12369 26637 8 26638 8 26639 12370 26640 12371 26641 8 26642 12325 26643 12325 26644 12325 26645 12325 26646 12325 26647 12372 26648 12373 26649 12374 26650 12325 26651 12325 26652 12325 26653 12325 26654 12325 26655 12325 26656 12325 26657 12325 26658 12325 26659 12325 26660 12375 26661 8 26662 8 26663 8 26664 8 26665 8 26666 8 26667 8 26668 8 26669 8 26670 8 26671 8 26672 8 26673 8 26674 8 26675 8 26676 8 26677 8 26678 8 26679 8 26680 8 26681 8 26682 8 26683 8 26684 8 26685 8 26686 8 26687 8 26688 8 26689 8 26690 8 26691 8 26692 8 26693 8 26694 8 26695 8 26696 8 26697 8 26698 8 26699 8 26700 8 26701 8 26702 8 26703 8 26704 8 26705 8 26706 8 26707 8 26708 8 26709 8 26710 8 26711 8 26712 8 26713 8 26714 8 26715 8 26716 8 26717 8 26718 8 26719 8 26720 8 26721 8 26722 8 26723 8 26724 8 26725 8 26726 8 26727 8 26728 8 26729 8 26730 8 26731 8 26732 8 26733 8 26734 8 26735 8 26736 8 26737 8 26738 8 26739 8 26740 8 26741 12376 26742 8 26743 8 26744 8 26745 12377 26746 12378 26747 8 26748 12379 26749 8 26750 8 26751 12380 26752 12381 26753 12382 26754 8 26755 8 26756 12383 26757 12384 26758 12385 26759 12386 26760 8 26761 12387 26762 8 26763 12388 26764 12389 26765 12390 26766 12391 26767 12392 26768 12393 26769 12394 26770 12395 26771 12396 26772 12325 26773 12325 26774 12325 26775 12325 26776 12325 26777 12325 26778 12325 26779 12325 26780 12325 26781 12325 26782 12325 26783 12325 26784 12325 26785 12397 26786 12325 26787 12398 26788 12399 26789 8 26790 12400 26791 12401 26792 8 26793 12402 26794 8 26795 12403 26796 8 26797 8 26798 8 26799 12404 26800 12405 26801 12406 26802 8 26803 12407 26804 8 26805 8 26806 12408 26807 12409 26808 12410 26809 8 26810 8 26811 8 26812 12411 26813 12412 26814 8 26815 8 26816 8 26817 12413 26818 12414 26819 12415 26820 12416 26821 8 26822 8 26823 12417 26824 8 26825 12418 26826 12419 26827 8 26828 8 26829 12420 26830 8 26831 8 26832 8 26833 12421 26834 12422 26835 8 26836 8 26837 12423 26838 12424 26839 12425 26840 12426 26841 12427 26842 12428 26843 12429 26844 12430 26845 12431 26846 12432 26847 12433 26848 8 26849 12434 26850 12435 26851 12436 26852 12437 26853 12438 26854 12439 26855 12440 26856 12441 26857 12442 26858 12443 26859 8 26860 12444 26861 11658 26862 12445 26863 8 26864 12446 26865 8 26866 12447 26867 12448 26868 8 26869 8 26870 8 26871 12449 26872 8 26873 8 26874 12450 26875 12451 26876 12452 26877 12453 26878 8 26879 8 26880 12454 26881 8 26882 12455 26883 12456 26884 8 26885 8 26886 12457 26887 8 26888 12458 26889 8 26890 12459 26891 12460 26892 12461 26893 12325 26894 12325 26895 12462 26896 12463 26897 12325 26898 12325 26899 12325 26900 12325 26901 8 26902 12464 26903 8 26904 12465 26905 8 26906 8 26907 12325 26908 12325 26909 8 26910 12325 26911 12325 26912 12466 26913 8 26914 8 26915 8 26916 8 26917 8 26918 8 26919 12467 26920 12468 26921 12469 26922 12470 26923 8 26924 12471 26925 8 26926 12472 26927 12473 26928 12474 26929 12475 26930 12476 26931 8 26932 8 26933 8 26934 8 26935 8 26936 12477 26937 8 26938 8 26939 12478 26940 12479 26941 12480 26942 12481 26943 12482 26944 12483 26945 12484 26946 8 26947 12485 26948 8 26949 12486 26950 12487 26951 12488 26952 12489 26953 12490 26954 8 26955 12491 26956 12492 26957 12493 26958 12494 26959 12325 26960 12495 26961 12496 26962 12497 26963 8 26964 12325 26965 12325 26966 12325 26967 8 26968 12498 26969 8 26970 8 26971 8 26972 12499 26973 12500 26974 12501 26975 8 26976 8 26977 12502 26978 12503 26979 8 26980 12504 26981 12505 26982 8 26983 12325 26984 8 26985 12506 26986 8 26987 12325 26988 8 26989 12507 26990 12508 26991 12509 26992 12510 26993 8 26994 8 26995 12511 26996 8 26997 8 26998 12325 26999 12325 27000 12512 27001 12513 27002 8 27003 12325 27004 12325 27005 12325 27006 12514 27007 8 27008 12515 27009 12516 27010 8 27011 12517 27012 8 27013 12325 27014 11658 27015 11658 27016 11658 27017 8 27018 12518 27019 8 27020 12519 27021 8 27022 8 27023 11658 27024 11658 27025 8 27026 11658 27027 11658 27028 12520 27029 8 27030 8 27031 8 27032 8 27033 8 27034 8 27035 12521 27036 12522 27037 12523 27038 12524 27039 8 27040 12525 27041 8 27042 12526 27043 12527 27044 12528 27045 12529 27046 12530 27047 8 27048 8 27049 8 27050 8 27051 8 27052 12531 27053 8 27054 8 27055 12532 27056 12533 27057 12534 27058 12535 27059 12536 27060 12537 27061 12538 27062 8 27063 12539 27064 8 27065 12540 27066 12541 27067 12542 27068 12543 27069 12544 27070 8 27071 12545 27072 12546 27073 12547 27074 12548 27075 11658 27076 12549 27077 12550 27078 12551 27079 8 27080 11658 27081 11658 27082 11658 27083 8 27084 12552 27085 8 27086 8 27087 8 27088 12553 27089 12554 27090 12555 27091 8 27092 8 27093 12556 27094 12557 27095 8 27096 12558 27097 12559 27098 8 27099 11658 27100 8 27101 12560 27102 8 27103 11658 27104 8 27105 12561 27106 12562 27107 12563 27108 12564 27109 8 27110 8 27111 12565 27112 8 27113 8 27114 11658 27115 11658 27116 12566 27117 12567 27118 8 27119 11658 27120 11658 27121 11658 27122 12568 27123 8 27124 12569 27125 12570 27126 8 27127 12571 27128 8 27129 11658 27130 12572 27131 12573 27132 8 27133 8 27134 12574 27135 8 27136 12575 27137 8 27138 12576 27139 8 27140 12577 27141 8 27142 12578 27143 12579 27144 12580 27145 12581 27146 12582 27147 12583 27148 12584 27149 12585 27150 12586 27151 8 27152 8 27153 12587 27154 12588 27155 8 27156 8 27157 12589 27158 12590 27159 12591 27160 12592 27161 12593 27162 12594 27163 12595 27164 12596 27165 8 27166 8 27167 12597 27168 12598 27169 8 27170 12599 27171 8 27172 12600 27173 12601 27174 12602 27175 12603 27176 12604 27177 8 27178 12605 27179 12606 27180 12607 27181 12608 27182 8 27183 8 27184 12609 27185 12610 27186 8 27187 12611 27188 8 27189 12612 27190 8 27191 8 27192 12613 27193 12614 27194 12615 27195 12616 27196 12617 27197 12618 27198 12619 27199 12620 27200 12621 27201 12622 27202 12623 27203 8 27204 8 27205 12624 27206 12625 27207 12626 27208 12627 27209 12628 27210 12629 27211 12630 27212 8 27213 8 27214 12631 27215 12632 27216 8 27217 12633 27218 12634 27219 12635 27220 8 27221 12636 27222 12637 27223 12638 27224 12639 27225 8 27226 8 27227 12640 27228 12641 27229 8 27230 12642 27231 8 27232 12643 27233 12644 27234 12645 27235 12646 27236 8 27237 12647 27238 12648 27239 12649 27240 12650 27241 12651 27242 12652 27243 8 27244 12653 27245 12654 27246 8 27247 12655 27248 12656 27249 12657 27250 12658 27251 12659 27252 12660 27253 8 27254 12661 27255 12662 27256 12663 27257 12664 27258 8 27259 12665 27260 12666 27261 12667 27262 12668 27263 8 27264 12669 27265 12670 27266 12671 27267 8 27268 12672 27269 8 27270 12673 27271 12674 27272 12675 27273 8 27274 12676 27275 12677 27276 8 27277 8 27278 12678 27279 8 27280 12679 27281 8 27282 12680 27283 12681 27284 12682 27285 12683 27286 12684 27287 8 27288 12685 27289 12686 27290 12687 27291 8 27292 12688 27293 12689 27294 8 27295 12690 27296 8 27297 12691 27298 8 27299 12692 27300 12693 27301 12694 27302 12695 27303 12696 27304 12697 27305 12698 27306 12699 27307 12700 27308 8 27309 8 27310 12701 27311 12702 27312 8 27313 8 27314 12703 27315 12704 27316 12705 27317 12706 27318 12707 27319 12708 27320 12709 27321 12710 27322 8 27323 8 27324 12711 27325 12712 27326 8 27327 12713 27328 12714 27329 12715 27330 12716 27331 12717 27332 12718 27333 8 27334 12719 27335 8 27336 12720 27337 12721 27338 12722 27339 12723 27340 12724 27341 8 27342 12725 27343 12726 27344 12727 27345 12728 27346 8 27347 8 27348 12729 27349 12730 27350 8 27351 12731 27352 8 27353 12732 27354 12733 27355 12734 27356 12735 27357 12736 27358 12737 27359 8 27360 12738 27361 8 27362 7957 27363 12739 27364 12740 27365 12741 27366 7957 27367 7957 27368 12742 27369 8 27370 12743 27371 12744 27372 12745 27373 8 27374 12746 27375 12747 27376 8 27377 8 27378 12748 27379 8 27380 12749 27381 8 27382 12750 27383 8 27384 12751 27385 8 27386 12752 27387 12753 27388 12754 27389 12755 27390 12756 27391 12757 27392 12758 27393 12759 27394 12760 27395 8 27396 8 27397 12761 27398 12762 27399 8 27400 8 27401 12763 27402 12764 27403 12765 27404 12766 27405 12767 27406 12768 27407 12769 27408 12770 27409 8 27410 8 27411 12771 27412 12772 27413 8 27414 12773 27415 8 27416 12774 27417 12775 27418 12776 27419 12777 27420 12778 27421 8 27422 12779 27423 12780 27424 12781 27425 12782 27426 8 27427 8 27428 12783 27429 12784 27430 8 27431 12785 27432 8 27433 12786 27434 8 27435 8 27436 12787 27437 12788 27438 12789 27439 12790 27440 12791 27441 12792 27442 12793 27443 12794 27444 12795 27445 12796 27446 12797 27447 8 27448 8 27449 12798 27450 12799 27451 12800 27452 12801 27453 12802 27454 12803 27455 12804 27456 8 27457 8 27458 12805 27459 12806 27460 8 27461 12807 27462 12808 27463 12809 27464 8 27465 12810 27466 12811 27467 12812 27468 12813 27469 8 27470 8 27471 12814 27472 12815 27473 8 27474 12816 27475 8 27476 12817 27477 12818 27478 12819 27479 12820 27480 8 27481 12821 27482 12822 27483 12823 27484 12824 27485 12825 27486 12826 27487 8 27488 12827 27489 12828 27490 8 27491 12829 27492 12830 27493 12831 27494 12832 27495 12833 27496 12834 27497 8 27498 12835 27499 12836 27500 12837 27501 12838 27502 8 27503 12839 27504 12840 27505 12841 27506 12842 27507 8 27508 12843 27509 12844 27510 12845 27511 8 27512 12846 27513 8 27514 12847 27515 12848 27516 12849 27517 8 27518 12850 27519 12851 27520 8 27521 8 27522 12852 27523 8 27524 12853 27525 8 27526 12854 27527 12855 27528 12856 27529 12857 27530 12858 27531 8 27532 12859 27533 12860 27534 12861 27535 8 27536 12862 27537 12863 27538 8 27539 12864 27540 8 27541 12865 27542 8 27543 12866 27544 12867 27545 12868 27546 12869 27547 12870 27548 12871 27549 12872 27550 12873 27551 12874 27552 8 27553 8 27554 12875 27555 12876 27556 8 27557 8 27558 12877 27559 12878 27560 12879 27561 12880 27562 12881 27563 12882 27564 12883 27565 12884 27566 8 27567 8 27568 12885 27569 12886 27570 8 27571 12887 27572 12888 27573 12889 27574 12890 27575 12891 27576 12892 27577 8 27578 8 27579 12893 27580 8 27581 12894 27582 12895 27583 12896 27584 12897 27585 12898 27586 8 27587 12899 27588 12900 27589 12901 27590 12902 27591 8 27592 8 27593 12903 27594 12904 27595 8 27596 12905 27597 8 27598 12906 27599 12907 27600 12908 27601 12909 27602 12910 27603 12911 27604 8 27605 12912 27606 8 27607 12913 27608 8 27609 12914 27610 12915 27611 12916 27612 12917 27613 8 27614 12918 27615 12919 27616 12920 27617 12921 27618 12922 27619 12923 27620 8 27621 12924 27622 12925 27623 8 27624 12926 27625 12927 27626 12928 27627 12929 27628 12930 27629 12931 27630 8 27631 12932 27632 12933 27633 12934 27634 12935 27635 8 27636 12936 27637 12937 27638 12938 27639 12939 27640 8 27641 12940 27642 12941 27643 12942 27644 8 27645 12943 27646 8 27647 12944 27648 12945 27649 12946 27650 8 27651 12947 27652 12948 27653 8 27654 8 27655 12949 27656 8 27657 12950 27658 8 27659 12951 27660 12952 27661 12953 27662 12954 27663 12955 27664 8 27665 12956 27666 12957 27667 8 27668 12958 27669 8 27670 12959 27671 8 27672 12960 27673 12961 27674 12962 27675 12963 27676 12964 27677 12965 27678 12966 27679 12967 27680 12968 27681 8 27682 8 27683 12969 27684 12970 27685 8 27686 8 27687 12971 27688 12972 27689 12973 27690 12974 27691 12975 27692 12976 27693 12977 27694 12978 27695 8 27696 8 27697 12979 27698 12980 27699 8 27700 12981 27701 12982 27702 12983 27703 12984 27704 12985 27705 12986 27706 8 27707 8 27708 12987 27709 8 27710 12988 27711 12989 27712 12990 27713 12991 27714 12992 27715 8 27716 12993 27717 12994 27718 12995 27719 12996 27720 8 27721 8 27722 12997 27723 12998 27724 8 27725 12999 27726 8 27727 13000 27728 13001 27729 13002 27730 13003 27731 13004 27732 13005 27733 8 27734 13006 27735 8 27736 13007 27737 13008 27738 13009 27739 13010 27740 8 27741 8 27742 13011 27743 13012 27744 8 27745 13013 27746 8 27747 13014 27748 13015 27749 13016 27750 13017 27751 8 27752 13018 27753 13019 27754 13020 27755 13021 27756 13022 27757 13023 27758 8 27759 13024 27760 13025 27761 8 27762 13026 27763 13027 27764 13028 27765 13029 27766 13030 27767 13031 27768 8 27769 13032 27770 13033 27771 13034 27772 13035 27773 8 27774 13036 27775 13037 27776 13038 27777 13039 27778 8 27779 13040 27780 13041 27781 13042 27782 8 27783 13043 27784 8 27785 13044 27786 13045 27787 13046 27788 8 27789 13047 27790 13048 27791 8 27792 8 27793 13049 27794 8 27795 13050 27796 8 27797 13051 27798 13052 27799 13053 27800 13054 27801 13055 27802 8 27803 13056 27804 13057 27805 13058 27806 8 27807 13059 27808 13060 27809 8 27810 13061 27811 8 27812 13062 27813 8 27814 13063 27815 13064 27816 13065 27817 13066 27818 13067 27819 13068 27820 13069 27821 13070 27822 13071 27823 8 27824 8 27825 13072 27826 13073 27827 8 27828 8 27829 13074 27830 13075 27831 13076 27832 13077 27833 13078 27834 13079 27835 13080 27836 13081 27837 8 27838 8 27839 13082 27840 13083 27841 8 27842 13084 27843 13085 27844 13086 27845 13087 27846 13088 27847 13089 27848 8 27849 13090 27850 8 27851 13091 27852 13092 27853 13093 27854 13094 27855 13095 27856 8 27857 13096 27858 13097 27859 13098 27860 13099 27861 8 27862 8 27863 13100 27864 13101 27865 8 27866 13102 27867 8 27868 13103 27869 13104 27870 13105 27871 13106 27872 13107 27873 13108 27874 8 27875 13109 27876 8 27877 13110 27878 13111 27879 13112 27880 8 27881 13113 27882 13114 27883 13115 27884 13116 27885 8 27886 8 27887 13117 27888 13118 27889 8 27890 13119 27891 8 27892 13120 27893 13121 27894 13122 27895 13123 27896 8 27897 13124 27898 13125 27899 13126 27900 13127 27901 13128 27902 13129 27903 8 27904 13130 27905 13131 27906 8 27907 13132 27908 13133 27909 13134 27910 13135 27911 13136 27912 13137 27913 8 27914 13138 27915 13139 27916 13140 27917 13141 27918 8 27919 13142 27920 13143 27921 13144 27922 8 27923 13145 27924 8 27925 13146 27926 13147 27927 13148 27928 8 27929 13149 27930 13150 27931 8 27932 8 27933 13151 27934 8 27935 13152 27936 8 27937 13153 27938 8 27939 13154 27940 8 27941 13155 27942 13156 27943 13157 27944 13158 27945 13159 27946 13160 27947 13161 27948 13162 27949 13163 27950 8 27951 8 27952 13164 27953 13165 27954 8 27955 8 27956 13166 27957 13167 27958 13168 27959 13169 27960 13170 27961 13171 27962 13172 27963 13173 27964 8 27965 8 27966 13174 27967 13175 27968 8 27969 13176 27970 8 27971 13177 27972 13178 27973 13179 27974 13180 27975 13181 27976 8 27977 13182 27978 13183 27979 13184 27980 13185 27981 8 27982 8 27983 13186 27984 13187 27985 8 27986 13188 27987 8 27988 13189 27989 8 27990 13190 27991 8 27992 13191 27993 8 27994 13192 27995 13193 27996 8 27997 8 27998 8 27999 13194 28000 8 28001 8 28002 13195 28003 13196 28004 13197 28005 13198 28006 8 28007 8 28008 13199 28009 8 28010 13200 28011 13201 28012 8 28013 8 28014 13202 28015 8 28016 13203 28017 8 28018 13204 28019 13205 28020 13206 28021 189 28022 189 28023 13207 28024 13208 28025 13209 28026 13210 28027 13211 28028 13212 28029 13213 28030 13214 28031 13215 28032 13216 28033 8 28034 13217 28035 8 28036 13218 28037 8 28038 8 28039 13219 28040 13220 28041 13221 28042 13222 28043 8 28044 8 28045 13223 28046 13224 28047 13225 28048 13226 28049 13227 28050 8 28051 13228 28052 8 28053 13229 28054 8 28055 8 28056 8 28057 8 28058 13230 28059 8 28060 13231 28061 13232 28062 8 28063 13233 28064 13234 28065 13235 28066 13236 28067 13237 28068 8 28069 13238 28070 13239 28071 8 28072 13240 28073 8 28074 13241 28075 13242 28076 13243 28077 13244 28078 13245 28079 13246 28080 13247 28081 8 28082 8 28083 13248 28084 8 28085 8 28086 13249 28087 13250 28088 13251 28089 8 28090 8 28091 8 28092 8 28093 13252 28094 8 28095 8 28096 8 28097 8 28098 8 28099 13253 28100 13254 28101 13255 28102 13256 28103 13257 28104 13258 28105 13259 28106 13260 28107 13261 28108 13262 28109 13263 28110 13264 28111 13265 28112 13266 28113 13267 28114 13268 28115 13269 28116 13270 28117 13271 28118 13272 28119 13273 28120 13274 28121 13275 28122 13276 28123 13277 28124 13278 28125 13279 28126 13280 28127 13281 28128 13282 28129 13283 28130 13284 28131 13285 28132 13286 28133 13287 28134 13256 28135 13256 28136 13256 28137 13256 28138 13256 28139 13256 28140 13256 28141 13256 28142 13256 28143 13256 28144 13256 28145 13256 28146 13256 28147 13256 28148 13256 28149 13256 28150 13288 28151 13289 28152 8 28153 8 28154 8 28155 8 28156 8 28157 8 28158 13290 28159 8 28160 8 28161 13291 28162 8 28163 13292 28164 13293 28165 13294 28166 13295 28167 13296 28168 8 28169 8 28170 8 28171 8 28172 8 28173 13297 28174 13298 28175 8 28176 8 28177 13299 28178 8 28179 13300 28180 8 28181 8 28182 13301 28183 13302 28184 8 28185 13256 28186 13256 28187 13256 28188 13256 28189 13256 28190 13303 28191 13304 28192 13305 28193 13256 28194 13256 28195 13256 28196 13256 28197 13256 28198 13256 28199 13256 28200 13256 28201 13256 28202 13256 28203 13306 28204 8 28205 8 28206 8 28207 8 28208 8 28209 8 28210 8 28211 8 28212 8 28213 8 28214 8 28215 8 28216 8 28217 8 28218 8 28219 8 28220 8 28221 8 28222 8 28223 8 28224 8 28225 8 28226 8 28227 8 28228 8 28229 8 28230 8 28231 8 28232 8 28233 8 28234 8 28235 8 28236 8 28237 8 28238 8 28239 8 28240 8 28241 8 28242 8 28243 8 28244 8 28245 8 28246 8 28247 8 28248 8 28249 8 28250 8 28251 8 28252 8 28253 8 28254 8 28255 8 28256 8 28257 8 28258 8 28259 8 28260 8 28261 8 28262 8 28263 8 28264 8 28265 8 28266 8 28267 8 28268 8 28269 8 28270 8 28271 8 28272 8 28273 8 28274 8 28275 8 28276 8 28277 8 28278 8 28279 8 28280 8 28281 8 28282 8 28283 8 28284 13307 28285 8 28286 8 28287 8 28288 13308 28289 13309 28290 8 28291 13310 28292 8 28293 8 28294 13311 28295 13312 28296 13313 28297 8 28298 8 28299 13314 28300 13315 28301 13316 28302 13317 28303 8 28304 13318 28305 8 28306 13319 28307 13320 28308 13321 28309 13322 28310 13323 28311 13324 28312 13325 28313 13326 28314 13327 28315 13256 28316 13256 28317 13256 28318 13256 28319 13256 28320 13256 28321 13256 28322 13256 28323 13256 28324 13256 28325 13256 28326 13256 28327 13256 28328 13328 28329 13256 28330 13329 28331 13330 28332 8 28333 13331 28334 13332 28335 8 28336 13333 28337 8 28338 13334 28339 8 28340 8 28341 8 28342 13335 28343 13336 28344 13337 28345 8 28346 13338 28347 8 28348 8 28349 13339 28350 13340 28351 13341 28352 8 28353 8 28354 8 28355 13342 28356 13343 28357 8 28358 8 28359 8 28360 13344 28361 13345 28362 13346 28363 13347 28364 8 28365 8 28366 13348 28367 8 28368 13349 28369 13350 28370 8 28371 8 28372 13351 28373 8 28374 8 28375 8 28376 13352 28377 13353 28378 8 28379 8 28380 13354 28381 13355 28382 13356 28383 13357 28384 13358 28385 13359 28386 13360 28387 13361 28388 13362 28389 13363 28390 13364 28391 8 28392 13365 28393 13366 28394 13367 28395 13368 28396 13369 28397 13370 28398 13371 28399 13372 28400 13373 28401 13374 28402 8 28403 13375 28404 189 28405 13376 28406 8 28407 13377 28408 8 28409 13378 28410 13379 28411 8 28412 8 28413 8 28414 13380 28415 8 28416 8 28417 13381 28418 13382 28419 13383 28420 13384 28421 8 28422 8 28423 13385 28424 8 28425 13386 28426 13387 28427 8 28428 8 28429 13388 28430 8 28431 13389 28432 8 28433 13390 28434 13391 28435 13392 28436 13256 28437 13256 28438 13393 28439 13394 28440 13256 28441 13256 28442 13256 28443 13256 28444 8 28445 13395 28446 8 28447 13396 28448 8 28449 8 28450 13256 28451 13256 28452 8 28453 13256 28454 13256 28455 13397 28456 8 28457 8 28458 8 28459 8 28460 8 28461 8 28462 13398 28463 13399 28464 13400 28465 13401 28466 8 28467 13402 28468 8 28469 13403 28470 13404 28471 13405 28472 13406 28473 13407 28474 8 28475 8 28476 8 28477 8 28478 8 28479 13408 28480 8 28481 8 28482 13409 28483 13410 28484 13411 28485 13412 28486 13413 28487 13414 28488 13415 28489 8 28490 13416 28491 8 28492 13417 28493 13418 28494 13419 28495 13420 28496 13421 28497 8 28498 13422 28499 13423 28500 13424 28501 13425 28502 13256 28503 13426 28504 13427 28505 13428 28506 8 28507 13256 28508 13256 28509 13256 28510 8 28511 13429 28512 8 28513 8 28514 8 28515 13430 28516 13431 28517 13432 28518 8 28519 8 28520 13433 28521 13434 28522 8 28523 13435 28524 13436 28525 8 28526 13256 28527 8 28528 13437 28529 8 28530 13256 28531 8 28532 13438 28533 13439 28534 13440 28535 13441 28536 8 28537 8 28538 13442 28539 8 28540 8 28541 13256 28542 13256 28543 13443 28544 13444 28545 8 28546 13256 28547 13256 28548 13256 28549 13445 28550 8 28551 13446 28552 13447 28553 8 28554 13448 28555 8 28556 13256 28557 189 28558 189 28559 189 28560 8 28561 13449 28562 8 28563 13450 28564 8 28565 8 28566 189 28567 189 28568 8 28569 189 28570 189 28571 13451 28572 8 28573 8 28574 8 28575 8 28576 8 28577 8 28578 13452 28579 13453 28580 13454 28581 13455 28582 8 28583 13456 28584 8 28585 13457 28586 13458 28587 13459 28588 13460 28589 13461 28590 8 28591 8 28592 8 28593 8 28594 8 28595 13462 28596 8 28597 8 28598 13463 28599 13464 28600 13465 28601 13466 28602 13467 28603 13468 28604 13469 28605 8 28606 13470 28607 8 28608 13471 28609 13472 28610 13473 28611 13474 28612 13475 28613 8 28614 13476 28615 13477 28616 13478 28617 13479 28618 189 28619 13480 28620 13481 28621 13482 28622 8 28623 189 28624 189 28625 189 28626 8 28627 13483 28628 8 28629 8 28630 8 28631 13484 28632 13485 28633 13486 28634 8 28635 8 28636 13487 28637 13488 28638 8 28639 13489 28640 13490 28641 8 28642 189 28643 8 28644 13491 28645 8 28646 189 28647 8 28648 13492 28649 13493 28650 13494 28651 13495 28652 8 28653 8 28654 13496 28655 8 28656 8 28657 189 28658 189 28659 13497 28660 13498 28661 8 28662 189 28663 189 28664 189 28665 13499 28666 8 28667 13500 28668 13501 28669 8 28670 13502 28671 8 28672 189 28673 8 28674 13503 28675 13504 28676 13505 28677 13506 28678 13507 28679 13508 28680 13509 28681 13510 28682 13511 28683 13512 28684 13513 28685 8 28686 8 28687 13514 28688 13515 28689 13516 28690 13517 28691 13518 28692 13519 28693 13520 28694 8 28695 8 28696 13521 28697 13522 28698 8 28699 13523 28700 13524 28701 13525 28702 8 28703 13526 28704 13527 28705 13528 28706 13529 28707 8 28708 8 28709 13530 28710 13531 28711 8 28712 13532 28713 8 28714 13533 28715 13534 28716 13535 28717 13536 28718 8 28719 13537 28720 13538 28721 13539 28722 13540 28723 13541 28724 13542 28725 8 28726 13543 28727 13544 28728 8 28729 13545 28730 13546 28731 13547 28732 13548 28733 13549 28734 13550 28735 8 28736 13551 28737 13552 28738 13553 28739 13554 28740 8 28741 13555 28742 13556 28743 13557 28744 13558 28745 8 28746 13559 28747 13560 28748 13561 28749 8 28750 13562 28751 8 28752 13563 28753 13564 28754 13565 28755 8 28756 13566 28757 13567 28758 8 28759 8 28760 13568 28761 8 28762 13569 28763 8 28764 13570 28765 13571 28766 13572 28767 13573 28768 13574 28769 8 28770 13575 28771 13576 28772 13577 28773 8 28774 13578 28775 13579 28776 8 28777 13580 28778 8 28779 13581 28780 8 28781 13582 28782 13583 28783 13584 28784 13585 28785 13586 28786 13587 28787 13588 28788 13589 28789 13590 28790 8 28791 8 28792 13591 28793 13592 28794 8 28795 8 28796 13593 28797 13594 28798 13595 28799 13596 28800 13597 28801 13598 28802 13599 28803 13600 28804 8 28805 8 28806 13601 28807 13602 28808 8 28809 13603 28810 13604 28811 13605 28812 13606 28813 13607 28814 13608 28815 8 28816 13609 28817 8 28818 13610 28819 13611 28820 13612 28821 13613 28822 13614 28823 8 28824 13615 28825 13616 28826 13617 28827 13618 28828 8 28829 8 28830 13619 28831 13620 28832 8 28833 13621 28834 8 28835 13622 28836 13623 28837 13624 28838 13625 28839 13626 28840 13627 28841 8 28842 13628 28843 8 28844 13629 28845 13630 28846 8 28847 13631 28848 8 28849 8 28850 8 28851 8 28852 8 28853 8 28854 8 28855 13632 28856 8 28857 13633 28858 8 28859 8 28860 13634 28861 13635 28862 13636 28863 13637 28864 8 28865 8 28866 8 28867 8 28868 8 28869 8 28870 8 28871 8 28872 8 28873 8 28874 8 28875 13638 28876 8 28877 13639 28878 8 28879 13640 28880 8 28881 8 28882 8 28883 8 28884 8 28885 8 28886 8 28887 8 28888 8 28889 8 28890 13641 28891 8 28892 8 28893 8 28894 8 28895 8 28896 8 28897 8 28898 8 28899 8 28900 8 28901 8 28902 13642 28903 8 28904 8 28905 8 28906 13643 28907 8 28908 13644 28909 13645 28910 13646 28911 8 28912 8 28913 8 28914 13647 28915 8 28916 8 28917 8 28918 8 28919 8 28920 8 28921 13648 28922 8 28923 13649 28924 8 28925 13650 28926 13651 28927 13652 28928 13653 28929 13654 28930 8 28931 8 28932 8 28933 8 28934 8 28935 8 28936 8 28937 8 28938 8 28939 13655 28940 8 28941 13656 28942 8 28943 13657 28944 8 28945 8 28946 8 28947 8 28948 8 28949 8 28950 8 28951 8 28952 8 28953 13658 28954 8 28955 13659 28956 13660 28957 13661 28958 13662 28959 8 28960 13663 28961 13664 28962 8 28963 13665 28964 13666 28965 13667 28966 13668 28967 13669 28968 13670 28969 13671 28970 13672 28971 13673 28972 13674 28973 8 28974 13675 28975 8 28976 13676 28977 8 28978 8 28979 13677 28980 13678 28981 13679 28982 13680 28983 8 28984 8 28985 13681 28986 13682 28987 13683 28988 13684 28989 13685 28990 8 28991 13686 28992 8 28993 13687 28994 8 28995 8 28996 8 28997 8 28998 13688 28999 8 29000 13689 29001 13690 29002 8 29003 13691 29004 13692 29005 13693 29006 13694 29007 13695 29008 8 29009 13696 29010 13697 29011 8 29012 13698 29013 8 29014 13699 29015 13700 29016 13701 29017 13702 29018 13703 29019 13704 29020 13705 29021 8 29022 8 29023 13706 29024 8 29025 8 29026 13707 29027 13708 29028 13709 29029 8 29030 8 29031 8 29032 8 29033 13710 29034 8 29035 8 29036 8 29037 8 29038 8 29039 13711 29040 13712 29041 13713 29042 13714 29043 13715 29044 13716 29045 13717 29046 13718 29047 13719 29048 13720 29049 13721 29050 13722 29051 13723 29052 13724 29053 13725 29054 13726 29055 13727 29056 13728 29057 13729 29058 13730 29059 13731 29060 13732 29061 13733 29062 13734 29063 13735 29064 13736 29065 13737 29066 13738 29067 13739 29068 13740 29069 13741 29070 13742 29071 13743 29072 13744 29073 13745 29074 13714 29075 13714 29076 13714 29077 13714 29078 13714 29079 13714 29080 13714 29081 13714 29082 13714 29083 13714 29084 13714 29085 13714 29086 13714 29087 13714 29088 13714 29089 13714 29090 13746 29091 13747 29092 8 29093 8 29094 8 29095 8 29096 8 29097 8 29098 13748 29099 8 29100 8 29101 13749 29102 8 29103 13750 29104 13751 29105 13752 29106 13753 29107 13754 29108 8 29109 8 29110 8 29111 8 29112 8 29113 13755 29114 13756 29115 8 29116 8 29117 13757 29118 8 29119 13758 29120 8 29121 8 29122 13759 29123 13760 29124 8 29125 13714 29126 13714 29127 13714 29128 13714 29129 13714 29130 13761 29131 13762 29132 13763 29133 13714 29134 13714 29135 13714 29136 13714 29137 13714 29138 13714 29139 13714 29140 13714 29141 13714 29142 13714 29143 13764 29144 8 29145 8 29146 8 29147 8 29148 8 29149 8 29150 8 29151 8 29152 8 29153 8 29154 8 29155 8 29156 8 29157 8 29158 8 29159 8 29160 8 29161 8 29162 8 29163 8 29164 8 29165 8 29166 8 29167 8 29168 8 29169 8 29170 8 29171 8 29172 8 29173 8 29174 8 29175 8 29176 8 29177 8 29178 8 29179 8 29180 8 29181 8 29182 8 29183 8 29184 8 29185 8 29186 8 29187 8 29188 8 29189 8 29190 8 29191 8 29192 8 29193 8 29194 8 29195 8 29196 8 29197 8 29198 8 29199 8 29200 8 29201 8 29202 8 29203 8 29204 8 29205 8 29206 8 29207 8 29208 8 29209 8 29210 8 29211 8 29212 8 29213 8 29214 8 29215 8 29216 8 29217 8 29218 8 29219 8 29220 8 29221 8 29222 8 29223 8 29224 13765 29225 8 29226 8 29227 8 29228 13766 29229 13767 29230 8 29231 13768 29232 8 29233 8 29234 13769 29235 13770 29236 13771 29237 8 29238 8 29239 13772 29240 13773 29241 13774 29242 13775 29243 8 29244 13776 29245 8 29246 13777 29247 13778 29248 13779 29249 13780 29250 13781 29251 13782 29252 13783 29253 13784 29254 13785 29255 13714 29256 13714 29257 13714 29258 13714 29259 13714 29260 13714 29261 13714 29262 13714 29263 13714 29264 13714 29265 13714 29266 13714 29267 13714 29268 13786 29269 13714 29270 13787 29271 13788 29272 8 29273 13789 29274 13790 29275 8 29276 13791 29277 8 29278 13792 29279 8 29280 8 29281 8 29282 13793 29283 13794 29284 13795 29285 8 29286 13796 29287 8 29288 8 29289 13797 29290 13798 29291 13799 29292 8 29293 8 29294 8 29295 13800 29296 13801 29297 8 29298 8 29299 8 29300 13802 29301 13803 29302 13804 29303 13805 29304 8 29305 8 29306 13806 29307 8 29308 13807 29309 13808 29310 8 29311 8 29312 13809 29313 8 29314 8 29315 8 29316 13810 29317 13811 29318 8 29319 8 29320 13812 29321 13813 29322 13814 29323 13815 29324 13816 29325 13817 29326 13818 29327 13819 29328 13820 29329 13821 29330 13822 29331 8 29332 13823 29333 13824 29334 13825 29335 13826 29336 13827 29337 13828 29338 13829 29339 13830 29340 13831 29341 13832 29342 8 29343 8 29344 13833 29345 8 29346 13834 29347 8 29348 13835 29349 13836 29350 8 29351 8 29352 8 29353 13837 29354 8 29355 8 29356 13838 29357 13839 29358 13840 29359 13841 29360 8 29361 8 29362 13842 29363 8 29364 13843 29365 13844 29366 8 29367 8 29368 13845 29369 8 29370 13846 29371 8 29372 13847 29373 13848 29374 13849 29375 13714 29376 13714 29377 13850 29378 13851 29379 13852 29380 13853 29381 13854 29382 13855 29383 13856 29384 13857 29385 13858 29386 13859 29387 8 29388 13860 29389 8 29390 13861 29391 8 29392 8 29393 13862 29394 13863 29395 13864 29396 13865 29397 8 29398 8 29399 13866 29400 13867 29401 13868 29402 13869 29403 13870 29404 8 29405 13871 29406 8 29407 13872 29408 8 29409 8 29410 8 29411 8 29412 13873 29413 8 29414 13874 29415 13875 29416 8 29417 13876 29418 13877 29419 13878 29420 13879 29421 13880 29422 8 29423 13881 29424 13882 29425 8 29426 13883 29427 8 29428 13884 29429 13885 29430 13886 29431 13887 29432 13888 29433 13889 29434 13890 29435 8 29436 8 29437 13891 29438 8 29439 8 29440 13892 29441 13893 29442 13894 29443 8 29444 8 29445 8 29446 8 29447 13895 29448 8 29449 8 29450 8 29451 8 29452 8 29453 13896 29454 13897 29455 13898 29456 13899 29457 13900 29458 13901 29459 13902 29460 13903 29461 13904 29462 13905 29463 13906 29464 13907 29465 13908 29466 13909 29467 13910 29468 13911 29469 13912 29470 13913 29471 13914 29472 13915 29473 13916 29474 13917 29475 13918 29476 13919 29477 13920 29478 13921 29479 13922 29480 13923 29481 13924 29482 13925 29483 13926 29484 13927 29485 13928 29486 13929 29487 13930 29488 13899 29489 13899 29490 13899 29491 13899 29492 13899 29493 13899 29494 13899 29495 13899 29496 13899 29497 13899 29498 13899 29499 13899 29500 13899 29501 13899 29502 13899 29503 13899 29504 13931 29505 13932 29506 8 29507 8 29508 8 29509 8 29510 8 29511 8 29512 13933 29513 8 29514 8 29515 13934 29516 8 29517 13935 29518 13936 29519 13937 29520 13938 29521 13939 29522 8 29523 8 29524 8 29525 8 29526 8 29527 13940 29528 13941 29529 8 29530 8 29531 13942 29532 8 29533 13943 29534 8 29535 8 29536 13944 29537 13945 29538 8 29539 13899 29540 13899 29541 13899 29542 13899 29543 13899 29544 13946 29545 13947 29546 13948 29547 13899 29548 13899 29549 13899 29550 13899 29551 13899 29552 13899 29553 13899 29554 13899 29555 13899 29556 13899 29557 13949 29558 8 29559 8 29560 8 29561 8 29562 8 29563 8 29564 8 29565 8 29566 8 29567 8 29568 8 29569 8 29570 8 29571 8 29572 8 29573 8 29574 8 29575 8 29576 8 29577 8 29578 8 29579 8 29580 8 29581 8 29582 8 29583 8 29584 8 29585 8 29586 8 29587 8 29588 8 29589 8 29590 8 29591 8 29592 8 29593 8 29594 8 29595 8 29596 8 29597 8 29598 8 29599 8 29600 8 29601 8 29602 8 29603 8 29604 8 29605 8 29606 8 29607 8 29608 8 29609 8 29610 8 29611 8 29612 8 29613 8 29614 8 29615 8 29616 8 29617 8 29618 8 29619 8 29620 8 29621 8 29622 8 29623 8 29624 8 29625 8 29626 8 29627 8 29628 8 29629 8 29630 8 29631 8 29632 8 29633 8 29634 8 29635 8 29636 8 29637 8 29638 13950 29639 8 29640 8 29641 8 29642 13951 29643 13952 29644 8 29645 13953 29646 8 29647 8 29648 13954 29649 13955 29650 13956 29651 8 29652 8 29653 13957 29654 13958 29655 13959 29656 13960 29657 8 29658 13961 29659 8 29660 13962 29661 13963 29662 13964 29663 13965 29664 13966 29665 13967 29666 13968 29667 13969 29668 13970 29669 13899 29670 13899 29671 13899 29672 13899 29673 13899 29674 13899 29675 13899 29676 13899 29677 13899 29678 13899 29679 13899 29680 13899 29681 13899 29682 13971 29683 13899 29684 13972 29685 13973 29686 8 29687 13974 29688 13975 29689 8 29690 13976 29691 8 29692 13977 29693 8 29694 8 29695 8 29696 13978 29697 13979 29698 13980 29699 8 29700 13981 29701 8 29702 8 29703 13982 29704 13983 29705 13984 29706 8 29707 8 29708 8 29709 13985 29710 13986 29711 8 29712 8 29713 8 29714 13987 29715 13988 29716 13989 29717 13990 29718 8 29719 8 29720 13991 29721 8 29722 13992 29723 13993 29724 8 29725 8 29726 13994 29727 8 29728 8 29729 8 29730 13995 29731 13996 29732 8 29733 8 29734 13997 29735 13998 29736 13999 29737 14000 29738 14001 29739 14002 29740 14003 29741 14004 29742 14005 29743 14006 29744 14007 29745 8 29746 14008 29747 14009 29748 14010 29749 14011 29750 14012 29751 14013 29752 14014 29753 14015 29754 14016 29755 14017 29756 8 29757 14018 29758 13714 29759 14019 29760 8 29761 14020 29762 8 29763 14021 29764 14022 29765 8 29766 8 29767 8 29768 14023 29769 8 29770 8 29771 14024 29772 14025 29773 14026 29774 14027 29775 8 29776 8 29777 14028 29778 8 29779 14029 29780 14030 29781 8 29782 8 29783 14031 29784 8 29785 14032 29786 8 29787 14033 29788 14034 29789 14035 29790 13899 29791 13899 29792 14036 29793 14037 29794 13899 29795 13899 29796 13899 29797 13899 29798 8 29799 14038 29800 8 29801 14039 29802 8 29803 8 29804 13899 29805 13899 29806 8 29807 13899 29808 13899 29809 14040 29810 8 29811 8 29812 8 29813 8 29814 8 29815 8 29816 14041 29817 14042 29818 14043 29819 14044 29820 8 29821 14045 29822 8 29823 14046 29824 14047 29825 14048 29826 14049 29827 14050 29828 8 29829 8 29830 8 29831 8 29832 8 29833 14051 29834 8 29835 8 29836 14052 29837 14053 29838 14054 29839 14055 29840 14056 29841 14057 29842 14058 29843 8 29844 14059 29845 8 29846 14060 29847 14061 29848 14062 29849 14063 29850 14064 29851 8 29852 14065 29853 14066 29854 14067 29855 14068 29856 13899 29857 14069 29858 14070 29859 14071 29860 8 29861 13899 29862 13899 29863 13899 29864 8 29865 14072 29866 8 29867 8 29868 8 29869 14073 29870 14074 29871 14075 29872 8 29873 8 29874 14076 29875 14077 29876 8 29877 14078 29878 14079 29879 8 29880 13899 29881 8 29882 14080 29883 8 29884 13899 29885 8 29886 14081 29887 14082 29888 14083 29889 14084 29890 8 29891 8 29892 14085 29893 8 29894 8 29895 13899 29896 13899 29897 14086 29898 14087 29899 8 29900 13899 29901 13899 29902 13899 29903 14088 29904 8 29905 14089 29906 14090 29907 8 29908 14091 29909 8 29910 13899 29911 13714 29912 13714 29913 13714 29914 8 29915 14092 29916 8 29917 14093 29918 8 29919 8 29920 13714 29921 13714 29922 8 29923 13714 29924 13714 29925 14094 29926 8 29927 8 29928 8 29929 8 29930 8 29931 8 29932 14095 29933 14096 29934 14097 29935 14098 29936 8 29937 14099 29938 8 29939 14100 29940 14101 29941 14102 29942 14103 29943 14104 29944 8 29945 8 29946 8 29947 8 29948 8 29949 14105 29950 8 29951 8 29952 14106 29953 14107 29954 14108 29955 14109 29956 14110 29957 14111 29958 14112 29959 8 29960 14113 29961 8 29962 14114 29963 14115 29964 14116 29965 14117 29966 14118 29967 8 29968 14119 29969 14120 29970 14121 29971 14122 29972 13714 29973 14123 29974 14124 29975 14125 29976 8 29977 13714 29978 13714 29979 13714 29980 8 29981 14126 29982 8 29983 8 29984 8 29985 14127 29986 14128 29987 14129 29988 8 29989 8 29990 14130 29991 14131 29992 8 29993 14132 29994 14133 29995 8 29996 13714 29997 8 29998 14134 29999 8 30000 13714 30001 8 30002 14135 30003 14136 30004 14137 30005 14138 30006 8 30007 8 30008 14139 30009 8 30010 8 30011 13714 30012 13714 30013 14140 30014 14141 30015 8 30016 13714 30017 13714 30018 13714 30019 14142 30020 8 30021 14143 30022 14144 30023 8 30024 14145 30025 8 30026 13714 30027 14146 30028 8 30029 8 30030 8 30031 8 30032 8 30033 8 30034 8 30035 8 30036 8 30037 14147 30038 8 30039 8 30040 8 30041 8 30042 8 30043 14148 30044 14149 30045 8 30046 14150 30047 8 30048 8 30049 14151 30050 14152 30051 14153 30052 8 30053 8 30054 14154 30055 14155 30056 8 30057 14156 30058 8 30059 14157 30060 8 30061 8 30062 8 30063 8 30064 8 30065 8 30066 8 30067 8 30068 8 30069 8 30070 8 30071 8 30072 8 30073 8 30074 8 30075 8 30076 8 30077 8 30078 8 30079 8 30080 8 30081 8 30082 8 30083 8 30084 8 30085 8 30086 8 30087 8 30088 8 30089 8 30090 8 30091 8 30092 8 30093 8 30094 8 30095 8 30096 8 30097 8 30098 8 30099 8 30100 8 30101 8 30102 8 30103 8 30104 8 30105 8 30106 8 30107 8 30108 8 30109 8 30110 8 30111 8 30112 8 30113 8 30114 8 30115 8 30116 8 30117 8 30118 8 30119 8 30120 8 30121 8 30122 8 30123 8 30124 8 30125 8 30126 8 30127 8 30128 8 30129 8 30130 8 30131 8 30132 8 30133 8 30134 8 30135 8 30136 8 30137 8 30138 8 30139 8 30140 14158 30141 8 30142 8 30143 8 30144 14159 30145 14160 30146 8 30147 8 30148 14161 30149 14162 30150 14163 30151 14164 30152 8 30153 14165 30154 8 30155 14166 30156 14167 30157 14168 30158 14169 30159 14170 30160 14171 30161 14172 30162 14173 30163 14174 30164 85 30165 85 30166 85 30167 85 30168 85 30169 85 30170 85 30171 85 30172 85 30173 85 30174 85 30175 85 30176 85 30177 14175 30178 85 30179 14176 30180 14177 30181 8 30182 14178 30183 14179 30184 8 30185 14180 30186 8 30187 14181 30188 8 30189 8 30190 8 30191 14182 30192 14183 30193 14184 30194 8 30195 14185 30196 8 30197 8 30198 14186 30199 14187 30200 14188 30201 8 30202 8 30203 8 30204 14189 30205 14190 30206 8 30207 8 30208 8 30209 14191 30210 14192 30211 14193 30212 14194 30213 8 30214 8 30215 14195 30216 8 30217 14196 30218 14197 30219 8 30220 8 30221 14198 30222 8 30223 8 30224 8 30225 14199 30226 14200 30227 8 30228 8 30229 14201 30230 14202 30231 14203 30232 14204 30233 14205 30234 8 30235 14206 30236 14207 30237 14208 30238 14209 30239 14210 30240 14211 30241 14212 30242 14213 30243 14214 30244 8 30245 14215 30246 14216 30247 14217 30248 14218 30249 14219 30250 14220 30251 14221 30252 14222 30253 8 30254 14223 30255 14224 30256 8 30257 14225 30258 8 30259 14226 30260 14227 30261 8 30262 14228 30263 8 30264 8 30265 14229 30266 14230 30267 8 30268 14231 30269 14232 30270 14233 30271 8 30272 14234 30273 14235 30274 14236 30275 14237 30276 8 30277 8 30278 14238 30279 14239 30280 8 30281 14240 30282 8 30283 14241 30284 14242 30285 14243 30286 14244 30287 8 30288 14245 30289 14246 30290 14247 30291 14248 30292 14249 30293 14250 30294 8 30295 14251 30296 14252 30297 8 30298 14253 30299 14254 30300 14255 30301 14256 30302 14257 30303 14258 30304 8 30305 14259 30306 14260 30307 14261 30308 14262 30309 8 30310 14263 30311 14264 30312 14265 30313 8 30314 14266 30315 8 30316 14267 30317 14268 30318 14269 30319 8 30320 14270 30321 14271 30322 8 30323 8 30324 14272 30325 8 30326 14273 30327 8 30328 14274 30329 8 30330 14275 30331 8 30332 14276 30333 14277 30334 14278 30335 14279 30336 14280 30337 14281 30338 14282 30339 14283 30340 14284 30341 8 30342 8 30343 14285 30344 14286 30345 8 30346 8 30347 14287 30348 14288 30349 14289 30350 14290 30351 14291 30352 14292 30353 14293 30354 14294 30355 8 30356 8 30357 14295 30358 14296 30359 8 30360 14297 30361 8 30362 14298 30363 14299 30364 14300 30365 14301 30366 14302 30367 8 30368 14303 30369 14304 30370 14305 30371 14306 30372 8 30373 8 30374 14307 30375 14308 30376 8 30377 14309 30378 8 30379 14310 30380 8 30381 8 30382 14311 30383 14312 30384 14313 30385 14314 30386 14315 30387 14316 30388 14317 30389 14318 30390 14319 30391 14320 30392 14321 30393 8 30394 14322 30395 14323 30396 14324 30397 14325 30398 14326 30399 14327 30400 14328 30401 14329 30402 14330 30403 14331 30404 14332 30405 14333 30406 14334 30407 14335 30408 14336 30409 14337 30410 14338 30411 14339 30412 14340 30413 8 30414 14341 30415 8 30416 14342 30417 8 30418 8 30419 14343 30420 14344 30421 14345 30422 14346 30423 8 30424 8 30425 14347 30426 14348 30427 14349 30428 14350 30429 14351 30430 8 30431 14352 30432 8 30433 14353 30434 8 30435 8 30436 8 30437 8 30438 14354 30439 8 30440 14355 30441 14356 30442 8 30443 14357 30444 14358 30445 14359 30446 14360 30447 14361 30448 8 30449 14362 30450 14363 30451 8 30452 14364 30453 8 30454 14365 30455 14366 30456 14367 30457 14368 30458 14369 30459 14370 30460 14371 30461 8 30462 8 30463 14372 30464 8 30465 8 30466 14373 30467 14374 30468 14375 30469 8 30470 8 30471 8 30472 8 30473 14376 30474 8 30475 8 30476 8 30477 8 30478 8 30479 14377 30480 14378 30481 14379 30482 14380 30483 14381 30484 14382 30485 14383 30486 14384 30487 14385 30488 14386 30489 14387 30490 14388 30491 14389 30492 14390 30493 14391 30494 14392 30495 14393 30496 14394 30497 14395 30498 14396 30499 14397 30500 14398 30501 14399 30502 14400 30503 14401 30504 14402 30505 14403 30506 14404 30507 14405 30508 14406 30509 14407 30510 14408 30511 14409 30512 14410 30513 14411 30514 14380 30515 14380 30516 14380 30517 14380 30518 14380 30519 14380 30520 14380 30521 14380 30522 14380 30523 14380 30524 14380 30525 14380 30526 14380 30527 14380 30528 14380 30529 14380 30530 14412 30531 14413 30532 8 30533 8 30534 8 30535 8 30536 8 30537 8 30538 14414 30539 8 30540 8 30541 14415 30542 8 30543 14416 30544 14417 30545 14418 30546 14419 30547 14420 30548 8 30549 8 30550 8 30551 8 30552 8 30553 14421 30554 14422 30555 8 30556 8 30557 14423 30558 8 30559 14424 30560 8 30561 8 30562 14425 30563 14426 30564 8 30565 14380 30566 14380 30567 14380 30568 14380 30569 14380 30570 14427 30571 14428 30572 14429 30573 14380 30574 14380 30575 14380 30576 14380 30577 14380 30578 14380 30579 14380 30580 14380 30581 14380 30582 14380 30583 14430 30584 8 30585 8 30586 8 30587 8 30588 8 30589 8 30590 8 30591 8 30592 8 30593 8 30594 8 30595 8 30596 8 30597 8 30598 8 30599 8 30600 8 30601 8 30602 8 30603 8 30604 8 30605 8 30606 8 30607 8 30608 8 30609 8 30610 8 30611 8 30612 8 30613 8 30614 8 30615 8 30616 8 30617 8 30618 8 30619 8 30620 8 30621 8 30622 8 30623 8 30624 8 30625 8 30626 8 30627 8 30628 8 30629 8 30630 8 30631 8 30632 8 30633 8 30634 8 30635 8 30636 8 30637 8 30638 8 30639 8 30640 8 30641 8 30642 8 30643 8 30644 8 30645 8 30646 8 30647 8 30648 8 30649 8 30650 8 30651 8 30652 8 30653 8 30654 8 30655 8 30656 8 30657 8 30658 8 30659 8 30660 8 30661 8 30662 8 30663 8 30664 14431 30665 8 30666 8 30667 8 30668 14432 30669 14433 30670 8 30671 14434 30672 8 30673 8 30674 14435 30675 14436 30676 14437 30677 8 30678 8 30679 14438 30680 14439 30681 14440 30682 14441 30683 8 30684 14442 30685 8 30686 14443 30687 14444 30688 14445 30689 14446 30690 14447 30691 14448 30692 14449 30693 14450 30694 14451 30695 14380 30696 14380 30697 14380 30698 14380 30699 14380 30700 14380 30701 14380 30702 14380 30703 14380 30704 14380 30705 14380 30706 14380 30707 14380 30708 14452 30709 14380 30710 14453 30711 14454 30712 8 30713 14455 30714 14456 30715 8 30716 14457 30717 8 30718 14458 30719 8 30720 8 30721 8 30722 14459 30723 14460 30724 14461 30725 8 30726 14462 30727 8 30728 8 30729 14463 30730 14464 30731 14465 30732 8 30733 8 30734 8 30735 14466 30736 14467 30737 8 30738 8 30739 8 30740 14468 30741 14469 30742 14470 30743 14471 30744 8 30745 8 30746 14472 30747 8 30748 14473 30749 14474 30750 8 30751 8 30752 14475 30753 8 30754 8 30755 8 30756 14476 30757 14477 30758 8 30759 8 30760 14478 30761 14479 30762 14480 30763 14481 30764 14482 30765 14483 30766 14484 30767 14485 30768 14486 30769 14487 30770 14488 30771 8 30772 14489 30773 14490 30774 14491 30775 14492 30776 14493 30777 14494 30778 14495 30779 14496 30780 14497 30781 14498 30782 8 30783 8 30784 14499 30785 8 30786 14500 30787 8 30788 14501 30789 14502 30790 8 30791 8 30792 8 30793 14503 30794 8 30795 8 30796 14504 30797 14505 30798 14506 30799 14507 30800 8 30801 8 30802 14508 30803 8 30804 14509 30805 14510 30806 8 30807 8 30808 14511 30809 8 30810 14512 30811 8 30812 14513 30813 14514 30814 14515 30815 14380 30816 14380 30817 14516 30818 14517 30819 14518 30820 14519 30821 14520 30822 14521 30823 14522 30824 14523 30825 14524 30826 14525 30827 8 30828 14526 30829 8 30830 14527 30831 8 30832 8 30833 14528 30834 14529 30835 14530 30836 14531 30837 8 30838 8 30839 14532 30840 14533 30841 14534 30842 14535 30843 14536 30844 8 30845 14537 30846 8 30847 14538 30848 8 30849 8 30850 8 30851 8 30852 14539 30853 8 30854 14540 30855 14541 30856 8 30857 14542 30858 14543 30859 14544 30860 14545 30861 14546 30862 8 30863 14547 30864 14548 30865 8 30866 14549 30867 8 30868 14550 30869 14551 30870 14552 30871 14553 30872 14554 30873 14555 30874 14556 30875 8 30876 8 30877 14557 30878 8 30879 8 30880 14558 30881 14559 30882 14560 30883 8 30884 8 30885 8 30886 8 30887 14561 30888 8 30889 8 30890 8 30891 8 30892 8 30893 14562 30894 14563 30895 14564 30896 14565 30897 14566 30898 14567 30899 14568 30900 14569 30901 14570 30902 14571 30903 14572 30904 14573 30905 14574 30906 14575 30907 14576 30908 14577 30909 14578 30910 14579 30911 14580 30912 14581 30913 14582 30914 14583 30915 14584 30916 14585 30917 14586 30918 14587 30919 14588 30920 14589 30921 14590 30922 14591 30923 14592 30924 14593 30925 14594 30926 14595 30927 14596 30928 14565 30929 14565 30930 14565 30931 14565 30932 14565 30933 14565 30934 14565 30935 14565 30936 14565 30937 14565 30938 14565 30939 14565 30940 14565 30941 14565 30942 14565 30943 14565 30944 14597 30945 14598 30946 8 30947 8 30948 8 30949 8 30950 8 30951 8 30952 14599 30953 8 30954 8 30955 14600 30956 8 30957 14601 30958 14602 30959 14603 30960 14604 30961 14605 30962 8 30963 8 30964 8 30965 8 30966 8 30967 14606 30968 14607 30969 8 30970 8 30971 14608 30972 8 30973 14609 30974 8 30975 8 30976 14610 30977 14611 30978 8 30979 14565 30980 14565 30981 14565 30982 14565 30983 14565 30984 14612 30985 14613 30986 14614 30987 14565 30988 14565 30989 14565 30990 14565 30991 14565 30992 14565 30993 14565 30994 14565 30995 14565 30996 14565 30997 14615 30998 8 30999 8 31000 8 31001 8 31002 8 31003 8 31004 8 31005 8 31006 8 31007 8 31008 8 31009 8 31010 8 31011 8 31012 8 31013 8 31014 8 31015 8 31016 8 31017 8 31018 8 31019 8 31020 8 31021 8 31022 8 31023 8 31024 8 31025 8 31026 8 31027 8 31028 8 31029 8 31030 8 31031 8 31032 8 31033 8 31034 8 31035 8 31036 8 31037 8 31038 8 31039 8 31040 8 31041 8 31042 8 31043 8 31044 8 31045 8 31046 8 31047 8 31048 8 31049 8 31050 8 31051 8 31052 8 31053 8 31054 8 31055 8 31056 8 31057 8 31058 8 31059 8 31060 8 31061 8 31062 8 31063 8 31064 8 31065 8 31066 8 31067 8 31068 8 31069 8 31070 8 31071 8 31072 8 31073 8 31074 8 31075 8 31076 8 31077 8 31078 14616 31079 8 31080 8 31081 8 31082 14617 31083 14618 31084 8 31085 14619 31086 8 31087 8 31088 14620 31089 14621 31090 14622 31091 8 31092 8 31093 14623 31094 14624 31095 14625 31096 14626 31097 8 31098 14627 31099 8 31100 14628 31101 14629 31102 14630 31103 14631 31104 14632 31105 14633 31106 14634 31107 14635 31108 14636 31109 14565 31110 14565 31111 14565 31112 14565 31113 14565 31114 14565 31115 14565 31116 14565 31117 14565 31118 14565 31119 14565 31120 14565 31121 14565 31122 14637 31123 14565 31124 14638 31125 14639 31126 8 31127 14640 31128 14641 31129 8 31130 14642 31131 8 31132 14643 31133 8 31134 8 31135 8 31136 14644 31137 14645 31138 14646 31139 8 31140 14647 31141 8 31142 8 31143 14648 31144 14649 31145 14650 31146 8 31147 8 31148 8 31149 14651 31150 14652 31151 8 31152 8 31153 8 31154 14653 31155 14654 31156 14655 31157 14656 31158 8 31159 8 31160 14657 31161 8 31162 14658 31163 14659 31164 8 31165 8 31166 14660 31167 8 31168 8 31169 8 31170 14661 31171 14662 31172 8 31173 8 31174 14663 31175 14664 31176 14665 31177 14666 31178 14667 31179 14668 31180 14669 31181 14670 31182 14671 31183 14672 31184 14673 31185 8 31186 14674 31187 14675 31188 14676 31189 14677 31190 14678 31191 14679 31192 14680 31193 14681 31194 14682 31195 14683 31196 8 31197 14684 31198 14380 31199 14685 31200 8 31201 14686 31202 8 31203 14687 31204 14688 31205 8 31206 8 31207 8 31208 14689 31209 8 31210 8 31211 14690 31212 14691 31213 14692 31214 14693 31215 8 31216 8 31217 14694 31218 8 31219 14695 31220 14696 31221 8 31222 8 31223 14697 31224 8 31225 14698 31226 8 31227 14699 31228 14700 31229 14701 31230 14565 31231 14565 31232 14702 31233 14703 31234 14565 31235 14565 31236 14565 31237 14565 31238 8 31239 14704 31240 8 31241 14705 31242 8 31243 8 31244 14565 31245 14565 31246 8 31247 14565 31248 14565 31249 14706 31250 8 31251 8 31252 8 31253 8 31254 8 31255 8 31256 14707 31257 14708 31258 14709 31259 14710 31260 8 31261 14711 31262 8 31263 14712 31264 14713 31265 14714 31266 14715 31267 14716 31268 8 31269 8 31270 8 31271 8 31272 8 31273 14717 31274 8 31275 8 31276 14718 31277 14719 31278 14720 31279 14721 31280 14722 31281 14723 31282 14724 31283 8 31284 14725 31285 8 31286 14726 31287 14727 31288 14728 31289 14729 31290 14730 31291 8 31292 14731 31293 14732 31294 14733 31295 14734 31296 14565 31297 14735 31298 14736 31299 14737 31300 8 31301 14565 31302 14565 31303 14565 31304 8 31305 14738 31306 8 31307 8 31308 8 31309 14739 31310 14740 31311 14741 31312 8 31313 8 31314 14742 31315 14743 31316 8 31317 14744 31318 14745 31319 8 31320 14565 31321 8 31322 14746 31323 8 31324 14565 31325 8 31326 14747 31327 14748 31328 14749 31329 14750 31330 8 31331 8 31332 14751 31333 8 31334 8 31335 14565 31336 14565 31337 14752 31338 14753 31339 8 31340 14565 31341 14565 31342 14565 31343 14754 31344 8 31345 14755 31346 14756 31347 8 31348 14757 31349 8 31350 14565 31351 14380 31352 14380 31353 14380 31354 8 31355 14758 31356 8 31357 14759 31358 8 31359 8 31360 14380 31361 14380 31362 8 31363 14380 31364 14380 31365 14760 31366 8 31367 8 31368 8 31369 8 31370 8 31371 8 31372 14761 31373 14762 31374 14763 31375 14764 31376 8 31377 14765 31378 8 31379 14766 31380 14767 31381 14768 31382 14769 31383 14770 31384 8 31385 8 31386 8 31387 8 31388 8 31389 14771 31390 8 31391 8 31392 14772 31393 14773 31394 14774 31395 14775 31396 14776 31397 14777 31398 14778 31399 8 31400 14779 31401 8 31402 14780 31403 14781 31404 14782 31405 14783 31406 14784 31407 8 31408 14785 31409 14786 31410 14787 31411 14788 31412 14380 31413 14789 31414 14790 31415 14791 31416 8 31417 14380 31418 14380 31419 14380 31420 8 31421 14792 31422 8 31423 8 31424 8 31425 14793 31426 14794 31427 14795 31428 8 31429 8 31430 14796 31431 14797 31432 8 31433 14798 31434 14799 31435 8 31436 14380 31437 8 31438 14800 31439 8 31440 14380 31441 8 31442 14801 31443 14802 31444 14803 31445 14804 31446 8 31447 8 31448 14805 31449 8 31450 8 31451 14380 31452 14380 31453 14806 31454 14807 31455 8 31456 14380 31457 14380 31458 14380 31459 14808 31460 8 31461 14809 31462 14810 31463 8 31464 14811 31465 8 31466 14380 31467 8 31468 14812 31469 8 31470 14813 31471 8 31472 14814 31473 14815 31474 8 31475 8 31476 14816 31477 14817 31478 14818 31479 14819 31480 14820 31481 14821 31482 14822 31483 8 31484 8 31485 14823 31486 14824 31487 14825 31488 14826 31489 14827 31490 8 31491 8 31492 8 31493 8 31494 14828 31495 8 31496 14829 31497 8 31498 8 31499 8 31500 14830 31501 14831 31502 14832 31503 14833 31504 14834 31505 14835 31506 14836 31507 14837 31508 14838 31509 14839 31510 14840 31511 8 31512 14841 31513 14842 31514 14843 31515 14844 31516 14845 31517 14846 31518 14847 31519 8 31520 14848 31521 14849 31522 14850 31523 14851 31524 8 31525 14852 31526 14853 31527 14854 31528 14855 31529 14856 31530 8 31531 14857 31532 14858 31533 14859 31534 14860 31535 14861 31536 14862 31537 8 31538 14863 31539 8 31540 14864 31541 14865 31542 8 31543 8 31544 14866 31545 14867 31546 8 31547 14868 31548 14869 31549 14870 31550 14871 31551 14872 31552 14873 31553 8 31554 14874 31555 14875 31556 14876 31557 14877 31558 8 31559 8 31560 14878 31561 8 31562 14879 31563 14880 31564 8 31565 8 31566 14881 31567 8 31568 14882 31569 14883 31570 8 31571 14884 31572 14885 31573 14886 31574 8 31575 14887 31576 8 31577 14888 31578 14889 31579 14890 31580 85 31581 85 31582 14891 31583 14892 31584 14893 31585 14894 31586 14895 31587 14896 31588 14897 31589 14898 31590 14899 31591 14900 31592 8 31593 14901 31594 8 31595 14902 31596 8 31597 8 31598 14903 31599 14904 31600 14905 31601 14906 31602 8 31603 8 31604 14907 31605 14908 31606 14909 31607 14910 31608 14911 31609 8 31610 14912 31611 8 31612 14913 31613 8 31614 8 31615 8 31616 8 31617 14914 31618 8 31619 14915 31620 14916 31621 8 31622 14917 31623 14918 31624 14919 31625 14920 31626 14921 31627 8 31628 14922 31629 14923 31630 8 31631 14924 31632 8 31633 14925 31634 14926 31635 14927 31636 14928 31637 14929 31638 14930 31639 14931 31640 8 31641 8 31642 14932 31643 8 31644 8 31645 14933 31646 14934 31647 14935 31648 8 31649 8 31650 8 31651 8 31652 14936 31653 8 31654 8 31655 8 31656 8 31657 8 31658 14937 31659 14938 31660 14939 31661 14940 31662 14941 31663 14942 31664 14943 31665 14944 31666 14945 31667 14946 31668 14947 31669 14948 31670 14949 31671 14950 31672 14951 31673 14952 31674 14953 31675 14954 31676 14955 31677 14956 31678 14957 31679 14958 31680 14959 31681 14960 31682 14961 31683 14962 31684 14963 31685 14964 31686 14965 31687 14966 31688 14967 31689 14968 31690 14969 31691 14970 31692 14971 31693 14940 31694 14940 31695 14940 31696 14940 31697 14940 31698 14940 31699 14940 31700 14940 31701 14940 31702 14940 31703 14940 31704 14940 31705 14940 31706 14940 31707 14940 31708 14940 31709 14972 31710 14973 31711 8 31712 8 31713 8 31714 8 31715 8 31716 8 31717 14974 31718 8 31719 8 31720 14975 31721 8 31722 14976 31723 14977 31724 14978 31725 14979 31726 14980 31727 8 31728 8 31729 8 31730 8 31731 8 31732 14981 31733 14982 31734 8 31735 8 31736 14983 31737 8 31738 14984 31739 8 31740 8 31741 14985 31742 14986 31743 8 31744 14940 31745 14940 31746 14940 31747 14940 31748 14940 31749 14987 31750 14988 31751 14989 31752 14940 31753 14940 31754 14940 31755 14940 31756 14940 31757 14940 31758 14940 31759 14940 31760 14940 31761 14940 31762 14990 31763 8 31764 8 31765 8 31766 8 31767 8 31768 8 31769 8 31770 8 31771 8 31772 8 31773 8 31774 8 31775 8 31776 8 31777 8 31778 8 31779 8 31780 8 31781 8 31782 8 31783 8 31784 8 31785 8 31786 8 31787 8 31788 8 31789 8 31790 8 31791 8 31792 8 31793 8 31794 8 31795 8 31796 8 31797 8 31798 8 31799 8 31800 8 31801 8 31802 8 31803 8 31804 8 31805 8 31806 8 31807 8 31808 8 31809 8 31810 8 31811 8 31812 8 31813 8 31814 8 31815 8 31816 8 31817 8 31818 8 31819 8 31820 8 31821 8 31822 8 31823 8 31824 8 31825 8 31826 8 31827 8 31828 8 31829 8 31830 8 31831 8 31832 8 31833 8 31834 8 31835 8 31836 8 31837 8 31838 8 31839 8 31840 8 31841 8 31842 8 31843 14991 31844 8 31845 8 31846 8 31847 14992 31848 14993 31849 8 31850 14994 31851 8 31852 8 31853 14995 31854 14996 31855 14997 31856 8 31857 8 31858 14998 31859 14999 31860 15000 31861 15001 31862 8 31863 15002 31864 8 31865 15003 31866 15004 31867 15005 31868 15006 31869 15007 31870 15008 31871 15009 31872 15010 31873 15011 31874 14940 31875 14940 31876 14940 31877 14940 31878 14940 31879 14940 31880 14940 31881 14940 31882 14940 31883 14940 31884 14940 31885 14940 31886 14940 31887 15012 31888 14940 31889 15013 31890 15014 31891 8 31892 15015 31893 15016 31894 8 31895 15017 31896 8 31897 15018 31898 8 31899 8 31900 8 31901 15019 31902 15020 31903 15021 31904 8 31905 15022 31906 8 31907 8 31908 15023 31909 15024 31910 15025 31911 8 31912 8 31913 8 31914 15026 31915 15027 31916 8 31917 8 31918 8 31919 15028 31920 15029 31921 15030 31922 15031 31923 8 31924 8 31925 15032 31926 8 31927 15033 31928 15034 31929 8 31930 8 31931 15035 31932 8 31933 8 31934 8 31935 15036 31936 15037 31937 8 31938 8 31939 15038 31940 15039 31941 15040 31942 15041 31943 15042 31944 15043 31945 15044 31946 15045 31947 15046 31948 15047 31949 15048 31950 8 31951 15049 31952 15050 31953 15051 31954 15052 31955 15053 31956 15054 31957 15055 31958 15056 31959 15057 31960 15058 31961 8 31962 15059 31963 85 31964 15060 31965 8 31966 15061 31967 8 31968 15062 31969 15063 31970 8 31971 8 31972 8 31973 15064 31974 8 31975 8 31976 15065 31977 15066 31978 15067 31979 15068 31980 8 31981 8 31982 15069 31983 8 31984 15070 31985 15071 31986 8 31987 8 31988 15072 31989 8 31990 15073 31991 8 31992 15074 31993 15075 31994 15076 31995 14940 31996 14940 31997 15077 31998 15078 31999 14940 32000 14940 32001 14940 32002 14940 32003 8 32004 15079 32005 8 32006 15080 32007 8 32008 8 32009 14940 32010 14940 32011 8 32012 14940 32013 14940 32014 15081 32015 8 32016 8 32017 8 32018 8 32019 8 32020 8 32021 15082 32022 15083 32023 15084 32024 15085 32025 8 32026 15086 32027 8 32028 15087 32029 15088 32030 15089 32031 15090 32032 15091 32033 8 32034 8 32035 8 32036 8 32037 8 32038 15092 32039 8 32040 8 32041 15093 32042 15094 32043 15095 32044 15096 32045 15097 32046 15098 32047 15099 32048 8 32049 15100 32050 8 32051 15101 32052 15102 32053 15103 32054 15104 32055 15105 32056 8 32057 15106 32058 15107 32059 15108 32060 15109 32061 14940 32062 15110 32063 15111 32064 15112 32065 8 32066 14940 32067 14940 32068 14940 32069 8 32070 15113 32071 8 32072 8 32073 8 32074 15114 32075 15115 32076 15116 32077 8 32078 8 32079 15117 32080 15118 32081 8 32082 15119 32083 15120 32084 8 32085 14940 32086 8 32087 15121 32088 8 32089 14940 32090 8 32091 15122 32092 15123 32093 15124 32094 15125 32095 8 32096 8 32097 15126 32098 8 32099 8 32100 14940 32101 14940 32102 15127 32103 15128 32104 8 32105 14940 32106 14940 32107 14940 32108 15129 32109 8 32110 15130 32111 15131 32112 8 32113 15132 32114 8 32115 14940 32116 15133 32117 15134 32118 15135 32119 15136 32120 8 32121 8 32122 15137 32123 15138 32124 8 32125 15139 32126 8 32127 8 32128 8 32129 8 32130 8 32131 8 32132 15140 32133 15141 32134 8 32135 8 32136 15142 32137 15143 32138 15144 32139 15145 32140 15146 32141 15147 32142 8 32143 8 32144 8 32145 8 32146 15148 32147 8 32148 8 32149 15149 32150 8 32151 15150 32152 15151 32153 8 32154 8 32155 15152 32156 8 32157 15153 32158 15154 32159 15155 32160 8 32161 8 32162 15156 32163 8 32164 8 32165 8 32166 8 32167 15157 32168 15158 32169 8 32170 8 32171 8 32172 15159 32173 8 32174 15160 32175 8 32176 15161 32177 15162 32178 8 32179 8 32180 85 32181 15163 32182 15164 32183 15165 32184 15166 32185 8 32186 15167 32187 15168 32188 15169 32189 15170 32190 15171 32191 15172 32192 8 32193 15173 32194 15174 32195 8 32196 15175 32197 15176 32198 15177 32199 15178 32200 15179 32201 15180 32202 8 32203 15181 32204 15182 32205 15183 32206 15184 32207 8 32208 85 32209 15185 32210 15186 32211 15187 32212 15188 32213 8 32214 15189 32215 15190 32216 15191 32217 8 32218 15192 32219 8 32220 15193 32221 15194 32222 15195 32223 8 32224 15196 32225 15197 32226 8 32227 85 32228 8 32229 15198 32230 8 32231 15199 32232 8 32233 15200 32234 8 32235 8 32236 15201 32237 8 32238 85 32239 85 32240 85 32241 85 32242 15202 32243 8 32244 8 32245 8 32246 8 32247 8 32248 8 32249 15203 32250 15204 32251 15205 32252 15206 32253 8 32254 15207 32255 8 32256 15208 32257 15209 32258 15210 32259 15211 32260 15212 32261 8 32262 8 32263 8 32264 8 32265 8 32266 15213 32267 8 32268 8 32269 15214 32270 15215 32271 15216 32272 15217 32273 15218 32274 15219 32275 15220 32276 8 32277 15221 32278 8 32279 8 32280 8 32281 8 32282 8 32283 8 32284 8 32285 8 32286 8 32287 8 32288 8 32289 8 32290 8 32291 8 32292 8 32293 8 32294 8 32295 8 32296 8 32297 8 32298 8 32299 8 32300 8 32301 8 32302 8 32303 8 32304 8 32305 8 32306 8 32307 8 32308 8 32309 8 32310 8 32311 8 32312 8 32313 8 32314 8 32315 8 32316 8 32317 8 32318 8 32319 8 32320 8 32321 8 32322 8 32323 8 32324 8 32325 8 32326 8 32327 8 32328 8 32329 8 32330 8 32331 8 32332 8 32333 8 32334 8 32335 8 32336 8 32337 8 32338 8 32339 8 32340 8 32341 8 32342 8 32343 8 32344 8 32345 8 32346 8 32347 8 32348 8 32349 8 32350 8 32351 8 32352 8 32353 8 32354 8 32355 8 32356 8 32357 8 32358 8 32359 8 32360 15222 32361 8 32362 8 32363 15223 32364 15224 32365 8 32366 15225 32367 15226 32368 15227 32369 15228 32370 15229 32371 15230 32372 15231 32373 15232 32374 15233 32375 15234 32376 15235 32377 15236 32378 15237 32379 15238 32380 8 32381 15239 32382 15240 32383 15241 32384 15242 32385 15243 32386 15244 32387 15245 32388 15246 32389 15247 32390 15248 32391 15249 32392 15250 32393 8 32394 15251 32395 8 32396 8 32397 15252 32398 8 32399 15253 32400 8 32401 15254 32402 15255 32403 8 32404 8 32405 15256 32406 8 32407 8 32408 8 32409 8 32410 15257 32411 8 32412 15258 32413 8 32414 15259 32415 15260 32416 15261 32417 15262 32418 15263 32419 8 32420 15264 32421 15265 32422 15266 32423 15267 32424 15268 32425 15269 32426 15270 32427 15271 32428 15272 32429 8 32430 15273 32431 15274 32432 15275 32433 15276 32434 15277 32435 15278 32436 15279 32437 15280 32438 8 32439 15281 32440 15282 32441 8 32442 15283 32443 8 32444 15284 32445 15285 32446 8 32447 15286 32448 8 32449 8 32450 15287 32451 15288 32452 8 32453 15289 32454 15290 32455 15291 32456 8 32457 15292 32458 15293 32459 15294 32460 15295 32461 8 32462 8 32463 15296 32464 15297 32465 8 32466 15298 32467 8 32468 15299 32469 15300 32470 15301 32471 15302 32472 8 32473 15303 32474 15304 32475 15305 32476 15306 32477 15307 32478 15308 32479 8 32480 15309 32481 15310 32482 8 32483 15311 32484 15312 32485 15313 32486 15314 32487 15315 32488 15316 32489 8 32490 15317 32491 15318 32492 15319 32493 15320 32494 8 32495 15321 32496 15322 32497 15323 32498 8 32499 15324 32500 8 32501 15325 32502 15326 32503 15327 32504 8 32505 15328 32506 15329 32507 8 32508 8 32509 15330 32510 8 32511 15331 32512 8 32513 15332 32514 8 32515 15333 32516 8 32517 15334 32518 15335 32519 15336 32520 15337 32521 15338 32522 15339 32523 15340 32524 15341 32525 15342 32526 8 32527 8 32528 15343 32529 15344 32530 8 32531 8 32532 15345 32533 15346 32534 15347 32535 15348 32536 15349 32537 15350 32538 15351 32539 15352 32540 8 32541 8 32542 15353 32543 15354 32544 8 32545 15355 32546 8 32547 15356 32548 15357 32549 15358 32550 15359 32551 15360 32552 8 32553 15361 32554 15362 32555 15363 32556 15364 32557 8 32558 8 32559 15365 32560 15366 32561 8 32562 15367 32563 8 32564 15368 32565 8 32566 8 32567 15369 32568 15370 32569 15371 32570 15372 32571 15373 32572 15374 32573 15375 32574 15376 32575 15377 32576 15378 32577 15379 32578 8 32579 8 32580 15380 32581 15381 32582 15382 32583 15383 32584 15384 32585 15385 32586 15386 32587 8 32588 8 32589 15387 32590 15388 32591 8 32592 15389 32593 15390 32594 15391 32595 8 32596 15392 32597 15393 32598 15394 32599 15395 32600 8 32601 8 32602 15396 32603 15397 32604 8 32605 15398 32606 8 32607 15399 32608 15400 32609 15401 32610 15402 32611 15403 32612 8 32613 8 32614 8 32615 8 32616 15404 32617 15405 32618 15406 32619 15407 32620 8 32621 15408 32622 15409 32623 15410 32624 15411 32625 15412 32626 15413 32627 8 32628 15414 32629 15415 32630 8 32631 15416 32632 15417 32633 15418 32634 15419 32635 15420 32636 15421 32637 8 32638 15422 32639 15423 32640 15424 32641 15425 32642 8 32643 15426 32644 15427 32645 15428 32646 15429 32647 8 32648 15430 32649 15431 32650 15432 32651 8 32652 15433 32653 8 32654 15434 32655 15435 32656 15436 32657 8 32658 15437 32659 15438 32660 8 32661 8 32662 15439 32663 8 32664 15440 32665 8 32666 15441 32667 15442 32668 15443 32669 8 32670 8 32671 15444 32672 15445 32673 8 32674 15446 32675 15447 32676 15448 32677 15449 32678 15450 32679 8 32680 15451 32681 15452 32682 15453 32683 8 32684 15454 32685 15455 32686 8 32687 15456 32688 8 32689 15457 32690 8 32691 15458 32692 15459 32693 15460 32694 15461 32695 15462 32696 15463 32697 15464 32698 15465 32699 15466 32700 8 32701 8 32702 15467 32703 15468 32704 8 32705 8 32706 15469 32707 15470 32708 15471 32709 15472 32710 15473 32711 15474 32712 15475 32713 15476 32714 8 32715 8 32716 15477 32717 15478 32718 8 32719 15479 32720 15480 32721 15481 32722 15482 32723 15483 32724 15484 32725 8 32726 15485 32727 8 32728 15486 32729 15487 32730 15488 32731 15489 32732 15490 32733 8 32734 15491 32735 15492 32736 15493 32737 15494 32738 8 32739 8 32740 15495 32741 15496 32742 8 32743 15497 32744 8 32745 15498 32746 8 32747 15499 32748 15500 32749 15501 32750 15502 32751 15503 32752 15504 32753 8 32754 15505 32755 8 32756 8 32757 15506 32758 15507 32759 8 32760 15508 32761 15509 32762 15510 32763 15511 32764 15512 32765 8 32766 15513 32767 15514 32768 15515 32769 15516 32770 85 32771 15517 32772 15518 32773 15519 32774 8 32775 85 32776 15520 32777 15521 32778 15522 32779 15523 32780 15524 32781 8 32782 85 32783 85 32784 8 32785 15525 32786 15526 32787 15527 32788 8 32789 15528 32790 15529 32791 15530 32792 15531 32793 15532 32794 15533 32795 15534 32796 15535 32797 15536 32798 15537 32799 8 32800 15538 32801 15539 32802 15540 32803 15541 32804 15542 32805 15543 32806 15544 32807 15545 32808 8 32809 15546 32810 15547 32811 8 32812 15548 32813 8 32814 15549 32815 15550 32816 8 32817 15551 32818 15552 32819 15553 32820 15554 32821 15555 32822 8 32823 8 32824 8 32825 15556 32826 15557 32827 8 32828 15558 32829 15559 32830 15560 32831 8 32832 15561 32833 15562 32834 15563 32835 15564 32836 8 32837 8 32838 15565 32839 15566 32840 8 32841 15567 32842 8 32843 15568 32844 15569 32845 15570 32846 15571 32847 8 32848 15572 32849 15573 32850 15574 32851 15575 32852 15576 32853 15577 32854 8 32855 15578 32856 15579 32857 8 32858 15580 32859 15581 32860 15582 32861 15583 32862 15584 32863 15585 32864 8 32865 15586 32866 15587 32867 15588 32868 15589 32869 8 32870 15590 32871 15591 32872 15592 32873 8 32874 15593 32875 8 32876 15594 32877 15595 32878 15596 32879 8 32880 15597 32881 15598 32882 8 32883 8 32884 15599 32885 8 32886 15600 32887 8 32888 15601 32889 15602 32890 15603 32891 8 32892 15604 32893 8 32894 15605 32895 8 32896 15606 32897 15607 32898 15608 32899 15609 32900 15610 32901 15611 32902 15612 32903 15613 32904 15614 32905 8 32906 8 32907 15615 32908 15616 32909 8 32910 8 32911 15617 32912 15618 32913 15619 32914 15620 32915 15621 32916 15622 32917 15623 32918 15624 32919 8 32920 8 32921 15625 32922 15626 32923 8 32924 15627 32925 8 32926 15628 32927 15629 32928 15630 32929 15631 32930 15632 32931 8 32932 15633 32933 15634 32934 15635 32935 15636 32936 8 32937 8 32938 15637 32939 15638 32940 8 32941 15639 32942 8 32943 15640 32944 8 32945 15641 32946 15642 32947 8 32948 15643 32949 8 32950 15644 32951 8 32952 15645 32953 15646 32954 15647 32955 15648 32956 15649 32957 15650 32958 15651 32959 15652 32960 15653 32961 8 32962 8 32963 15654 32964 15655 32965 8 32966 8 32967 15656 32968 15657 32969 15658 32970 15659 32971 15660 32972 15661 32973 15662 32974 15663 32975 8 32976 8 32977 15664 32978 15665 32979 8 32980 15666 32981 8 32982 8 32983 8 32984 15667 32985 15668 32986 15669 32987 8 32988 8 32989 15670 32990 15671 32991 8 32992 15672 32993 15673 32994 15674 32995 15675 32996 15676 32997 15677 32998 8 32999 15678 33000 15679 33001 8 33002 15680 33003 8 33004 85 33005 15681 33006 15682 33007 15683 33008 15684 33009 15685 33010 8 33011 15686 33012 15687 33013 15688 33014 15689 33015 8 33016 8 33017 15690 33018 15691 33019 8 33020 8 33021 15692 33022 8 33023 15693 33024 8 33025 15694 33026 8 33027 85 33028 8 33029 15695 33030 15696 33031 15697 33032 15698 33033 8 33034 8 33035 15699 33036 8 33037 8 33038 15700 33039 15701 33040 15702 33041 15703 33042 15704 33043 15705 33044 8 33045 15706 33046 8 33047 85 33048 15707 33049 15708 33050 8 33051 85 33052 15709 33053 15710 33054 8 33055 85 33056 85 33057 85 33058 15711 33059 8 33060 15712 33061 15713 33062 8 33063 15714 33064 8 33065 85))
  "The generated goto-tables.")

(defvar
  phps-mode-parser--distinct-goto-tables
  #s(hash-table test equal data (0 ((top_statement_list 1)) 1 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 8) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 15) (T_CLASS 16) (T_CLASS_C 17) (T_CLONE 18) (T_CONST 19) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21) (T_DEC 22) (T_DECLARE 23) (T_DIR 24) (T_DNUMBER 25) (T_DO 26) (T_DOUBLE_CAST 27) (T_ECHO 28) (T_EMPTY 29) (T_ENUM 30) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 36) (T_FOREACH 37) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 40) (T_GOTO 41) (T_HALT_COMPILER 42) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 47) (T_INTERFACE 48) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAMESPACE 56) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 68) (T_START_HEREDOC 69) (T_STATIC 70) (T_STRING 71) (T_STRING_CAST 72) (T_SWITCH 73) (T_THROW 74) (T_TRAIT 75) (T_TRAIT_C 76) (T_TRY 77) (T_UNSET 78) (T_UNSET_CAST 79) (T_USE 80) (T_VARIABLE 81) (T_VOID_CAST 82) (T_WHILE 83) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 88) (alt_if_stmt_without_else 89) (array_object_dereferenceable 90) (attribute 91) (attributed_statement 92) (attributed_top_statement 93) (attributes 94) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 98) (class_modifier 99) (class_modifiers 100) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 104) (expr 105) (fn 106) (fully_dereferenceable 107) #'108 (function_call 109) (function_declaration_statement 110) (if_stmt 111) (if_stmt_without_else 112) (inline_function 113) (interface_declaration_statement 114) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 122) (static_member 123) (top_statement 124) (trait_declaration_statement 125) (variable 126) (variable_class_name 127) ("{" 128) ("~" 129)) 2 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 33065) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 3 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 33063) (encaps_var 5357)) 4 (("$" 4) (T_VARIABLE 81) (simple_variable 33059) ("{" 33060)) 5 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 33058) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 6 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 33057) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 7 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 33056) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 8 nil 9 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 33055) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 10 (("(" 33052)) 11 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 33051) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 12 ((T_NAME_FULLY_QUALIFIED 5996) (T_NAME_QUALIFIED 5997) (T_NAME_RELATIVE 5998) (T_STATIC 5999) (T_STRING 6000) (attribute_decl 6001) (attribute_group 33048) (class_name 6003) (name 6004)) 13 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 33047) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 14 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 33045) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 15 ((T_STRING 33038)) 16 (("!" 2) ("\"" 3) ("$" 4) ("(" 33025) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (clone_argument_list 33026) (constant 102) (dereferenceable_scalar 103) (expr 33027) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 17 ((T_STRING 31538) (const_decl 31539) (const_list 33023)) 18 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 33021) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 19 (("\"" 1538) ("$" 30179) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 30180) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 30181) ("[" 1560) (array_object_dereferenceable 30182) (callable_expr 30183) (callable_variable 30184) (class_constant 1564) (class_name 30185) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 30186) (name 30187) (new_dereferenceable 1570) (simple_variable 30188) (static_member 30189) (variable 33020) (variable_class_name 30191)) 20 (("(" 33011)) 21 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22932) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22933) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22934) (T_DEC 22) (T_DECLARE 22935) (T_DIR 24) (T_DNUMBER 25) (T_DO 22936) (T_DOUBLE_CAST 27) (T_ECHO 22937) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22938) (T_FOREACH 22939) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22940) (T_GOTO 22941) (T_IF 22942) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22943) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22944) (T_START_HEREDOC 69) (T_STATIC 22945) (T_STRING 22946) (T_STRING_CAST 72) (T_SWITCH 22947) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22948) (T_UNSET 22949) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22950) (T_WHILE 22951) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22952) (alt_if_stmt_without_else 22953) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22954) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22955) (if_stmt_without_else 22956) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 33005) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22958) ("~" 129)) 22 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 33004) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 23 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (echo_expr 22926) (echo_expr_list 33002) (expr 22928) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 24 (("(" 32999)) 25 ((T_STRING 32992)) 26 (("(" 32989)) 27 (("(" 32980) (argument_list 32981) (ctor_arguments 32982)) 28 (("(" 32967)) 29 (("(" 32952)) 30 (("$" 19441) (T_VARIABLE 19442) (global_var 19443) (global_var_list 32950) (simple_variable 19445)) 31 ((T_STRING 32948)) 32 (("(" 32945)) 33 (("(" 32785)) 34 (("\"" 1538) ("$" 30179) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 30180) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 30181) ("[" 1560) (array_object_dereferenceable 30182) (callable_expr 30183) (callable_variable 30184) (class_constant 1564) (class_name 30185) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 30186) (name 30187) (new_dereferenceable 1570) (simple_variable 30188) (static_member 30189) (variable 32784) (variable_class_name 30191)) 35 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 32783) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 36 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 32782) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 37 ((T_STRING 32776)) 38 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 32775) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 39 (("(" 32771)) 40 (("(" 32766)) 41 (("(" 32760)) 42 ((T_ABSTRACT 32279) (T_ARRAY 32280) (T_AS 32281) (T_BREAK 32282) (T_CALLABLE 32283) (T_CASE 32284) (T_CATCH 32285) (T_CLASS 32286) (T_CLASS_C 32287) (T_CLONE 32288) (T_CONST 32289) (T_CONTINUE 32290) (T_DECLARE 32291) (T_DEFAULT 32292) (T_DIR 32293) (T_DO 32294) (T_ECHO 32295) (T_ELSE 32296) (T_ELSEIF 32297) (T_EMPTY 32298) (T_ENDDECLARE 32299) (T_ENDFOR 32300) (T_ENDFOREACH 32301) (T_ENDIF 32302) (T_ENDSWITCH 32303) (T_ENDWHILE 32304) (T_ENUM 32305) (T_EVAL 32306) (T_EXIT 32307) (T_EXTENDS 32308) (T_FILE 32309) (T_FINAL 32310) (T_FINALLY 32311) (T_FN 32312) (T_FOR 32313) (T_FOREACH 32314) (T_FUNCTION 32315) (T_FUNC_C 32316) (T_GLOBAL 32317) (T_GOTO 32318) (T_IF 32319) (T_IMPLEMENTS 32320) (T_INCLUDE 32321) (T_INCLUDE_ONCE 32322) (T_INSTANCEOF 32323) (T_INSTEADOF 32324) (T_INTERFACE 32325) (T_ISSET 32326) (T_LINE 32327) (T_LIST 32328) (T_LOGICAL_AND 32329) (T_LOGICAL_OR 32330) (T_LOGICAL_XOR 32331) (T_MATCH 32332) (T_METHOD_C 32333) (T_NAMESPACE 32334) (T_NAME_QUALIFIED 32335) (T_NEW 32336) (T_NS_C 32337) (T_PRINT 32338) (T_PRIVATE 32339) (T_PROPERTY_C 32340) (T_PROTECTED 32341) (T_PUBLIC 32342) (T_READONLY 32343) (T_REQUIRE 32344) (T_REQUIRE_ONCE 32345) (T_RETURN 32346) (T_STATIC 32347) (T_STRING 32348) (T_SWITCH 32349) (T_THROW 32350) (T_TRAIT 32351) (T_TRAIT_C 32352) (T_TRY 32353) (T_UNSET 32354) (T_USE 32355) (T_VAR 32356) (T_WHILE 32357) (T_YIELD 32358) (identifier 32359) (namespace_declaration_name 32360) (reserved_non_modifiers 32361) (semi_reserved 32362) ("{" 32363)) 43 (("$" 30179) ("(" 32242) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 32243) (T_NAME_QUALIFIED 32244) (T_NAME_RELATIVE 32245) (T_READONLY 2955) (T_STATIC 32246) (T_STRING 32247) (T_VARIABLE 30181) (anonymous_class 32248) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 32249) (attribute 2961) (attributes 32250) (class_modifier 2963) (class_name 32251) (class_name_reference 32252) (name 32253) (new_variable 32254) (simple_variable 32255)) 44 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 32241) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 45 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 32240) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 46 (("(" 30046) (argument_list 30047)) 47 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 32239) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 48 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 32238) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 49 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 32236) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 50 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 32231) (T_END_HEREDOC 32232) (T_VARIABLE 2920) (encaps_list 32233) (encaps_var 2922)) 51 ((T_FN 35) (T_FUNCTION 131) (T_VARIABLE 18365) (fn 106) #'138 (inline_function 18366) (static_var 18367) (static_var_list 32229)) 52 ((":" 32228)) 53 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 32227) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 54 (("(" 32209)) 55 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 32208) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 56 ((T_STRING 32203)) 57 (("{" 32186)) 58 (("(" 32181)) 59 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 32180) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 60 ((T_CONST 32127) (T_FUNCTION 32128) (T_NAME_FULLY_QUALIFIED 32129) (T_NAME_QUALIFIED 32130) (T_STRING 32131) (legacy_namespace_name 32132) (mixed_group_use_declaration 32133) (namespace_name 32134) (use_declaration 32135) (use_declarations 32136) (use_type 32137)) 61 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 32125) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 62 (("(" 32116)) 63 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31645) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 64 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 31581) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 65 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 31577) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 66 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 31575) (encaps_list 2828) (encaps_var 2829)) 67 ((T_ELSE 31568) (T_ELSEIF 17711) (T_ENDIF 31569)) 68 ((T_NULLSAFE_OBJECT_OPERATOR 31554) (T_OBJECT_OPERATOR 31555) ("[" 31556)) 69 ((T_ABSTRACT 10) (T_ATTRIBUTE 13) (T_CLASS 31485) (T_CONST 31486) (T_ENUM 31487) (T_FINAL 34) (T_FN 35) (T_FUNCTION 38) (T_INTERFACE 31488) (T_READONLY 2955) (T_STATIC 17797) (T_TRAIT 31489) (attribute 31490) (attributed_statement 31491) (attributed_top_statement 31492) (class_declaration_statement 31493) (class_modifier 99) (class_modifiers 31494) (enum_declaration_statement 31495) (fn 106) #'31496 (function_declaration_statement 31497) (inline_function 17800) (interface_declaration_statement 31498) (trait_declaration_statement 31499)) 70 (("(" 30046) (argument_list 31484)) 71 ((T_ABSTRACT 10) (T_CLASS 31476) (T_FINAL 34) (T_READONLY 2955) (class_modifier 17789)) 72 ((T_PAAMAYIM_NEKUDOTAYIM 31468)) 73 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 31467) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 74 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 30394)) 75 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17770) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17771) (ampersand 17772) (returns_ref 30382)) 76 ((T_ELSE 30229) (T_ELSEIF 30230)) 77 (("(" 30046) (argument_list 30228)) 78 (("=" 30148) (T_AND_EQUAL 30149) (T_COALESCE_EQUAL 30150) (T_CONCAT_EQUAL 30151) (T_DEC 30152) (T_DIV_EQUAL 30153) (T_INC 30154) (T_MINUS_EQUAL 30155) (T_MOD_EQUAL 30156) (T_MUL_EQUAL 30157) (T_OR_EQUAL 30158) (T_PLUS_EQUAL 30159) (T_POW_EQUAL 30160) (T_SL_EQUAL 30161) (T_SR_EQUAL 30162) (T_XOR_EQUAL 30163)) 79 ((T_PAAMAYIM_NEKUDOTAYIM 30059)) 80 ((inner_statement_list 30057)) 81 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 137) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 82 ((T_NAME_FULLY_QUALIFIED 5996) (T_NAME_QUALIFIED 5997) (T_NAME_RELATIVE 5998) (T_STATIC 5999) (T_STRING 6000) (attribute_decl 6001) (attribute_group 30054) (class_name 6003) (name 6004)) 83 ((T_FN 35) (T_FUNCTION 131) (fn 106) #'138 (inline_function 18366)) 84 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 17797) (attribute 2757) (fn 106) #'138 (inline_function 17800)) 85 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 86 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 142)) 87 ((backup_doc_comment 143)) 88 (("(" 144)) 89 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 164)) 90 ((T_NAME_FULLY_QUALIFIED 5996) (T_NAME_QUALIFIED 5997) (T_NAME_RELATIVE 5998) (T_STATIC 5999) (T_STRING 6000) (attribute_decl 6001) (attribute_group 30043) (class_name 6003) (name 6004)) 91 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 30041) (member_modifier 159) (non_empty_member_modifiers 160) (optional_cpp_modifiers 162) (parameter 30042)) 92 ((T_ABSTRACT 145) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (member_modifier 30040)) 93 (("," 30037) (possible_comma 30038)) 94 (("(" 3019) ("?" 28914) (T_ARRAY 28915) (T_CALLABLE 28916) (T_NAME_FULLY_QUALIFIED 28917) (T_NAME_QUALIFIED 28918) (T_NAME_RELATIVE 28919) (T_STRING 28920) (intersection_type_without_static 28921) (name 28922) (optional_type_without_static 28923) (type_expr_without_static 28924) (type_without_static 28925) (union_type_without_static 28926) (union_type_without_static_element 28927)) 95 ((")" 165)) 96 ((T_USE 166) (lexical_vars 167)) 97 (("(" 28902)) 98 ((":" 168) (return_type 169)) 99 (("(" 2278) ("?" 28847) (T_ARRAY 28848) (T_CALLABLE 28849) (T_NAME_FULLY_QUALIFIED 28850) (T_NAME_QUALIFIED 28851) (T_NAME_RELATIVE 28852) (T_STATIC 28853) (T_STRING 28854) (intersection_type 28855) (name 28856) (type 28857) (type_expr 28858) (type_without_static 28859) (union_type 28860) (union_type_element 28861)) 100 ((backup_fn_flags 170)) 101 (("{" 171)) 102 ((inner_statement_list 172)) 103 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 217) ("~" 129)) 104 ((T_NAME_FULLY_QUALIFIED 5996) (T_NAME_QUALIFIED 5997) (T_NAME_RELATIVE 5998) (T_STATIC 5999) (T_STRING 6000) (attribute_decl 6001) (attribute_group 28844) (class_name 6003) (name 6004)) 105 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 28842) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 106 ((T_STRING 28835)) 107 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 28833) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 108 (("(" 28824)) 109 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22932) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22933) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22934) (T_DEC 22) (T_DECLARE 22935) (T_DIR 24) (T_DNUMBER 25) (T_DO 22936) (T_DOUBLE_CAST 27) (T_ECHO 22937) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22938) (T_FOREACH 22939) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22940) (T_GOTO 22941) (T_IF 22942) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22943) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22944) (T_START_HEREDOC 69) (T_STATIC 22945) (T_STRING 22946) (T_STRING_CAST 72) (T_SWITCH 22947) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22948) (T_UNSET 22949) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22950) (T_WHILE 22951) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22952) (alt_if_stmt_without_else 22953) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22954) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22955) (if_stmt_without_else 22956) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 28818) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22958) ("~" 129)) 110 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (echo_expr 22926) (echo_expr_list 28816) (expr 22928) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 111 ((T_STRING 28809)) 112 (("(" 28796)) 113 (("(" 28781)) 114 (("$" 19441) (T_VARIABLE 19442) (global_var 19443) (global_var_list 28779) (simple_variable 19445)) 115 ((T_STRING 28777)) 116 (("(" 28774)) 117 (("(" 28770)) 118 ((T_STRING 28764)) 119 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 28762) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 120 ((T_FN 35) (T_FUNCTION 131) (T_VARIABLE 18365) (fn 106) #'138 (inline_function 18366) (static_var 18367) (static_var_list 28760)) 121 ((":" 28759)) 122 (("(" 28741)) 123 ((T_STRING 28736)) 124 (("{" 28719)) 125 (("(" 28714)) 126 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 28712) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 127 (("(" 28703)) 128 ((T_ELSE 28696) (T_ELSEIF 17711) (T_ENDIF 28697)) 129 ((T_ABSTRACT 10) (T_ATTRIBUTE 174) (T_CLASS 176) (T_ENUM 181) (T_FINAL 34) (T_FN 35) (T_FUNCTION 38) (T_INTERFACE 189) (T_READONLY 2955) (T_STATIC 17797) (T_TRAIT 194) (attribute 17798) (attributed_statement 28695) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (enum_declaration_statement 206) (fn 106) #'208 (function_declaration_statement 209) (inline_function 17800) (interface_declaration_statement 213) (trait_declaration_statement 215)) 130 ((T_ABSTRACT 10) (T_CLASS 28687) (T_FINAL 34) (T_READONLY 2955) (class_modifier 17789)) 131 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 28686) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 132 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17770) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17771) (ampersand 17772) (returns_ref 28674)) 133 ((T_ELSE 221) (T_ELSEIF 222)) 134 ((inner_statement_list 219)) 135 ((backup_fn_flags 218)) 136 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 220) ("~" 129)) 137 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 207) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 28673) (static_member 123) (variable 126) (variable_class_name 127) ("{" 216) ("~" 129)) 138 (("(" 223)) 139 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 287) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 140 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 28672) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 141 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 28670) (encaps_var 5357)) 142 (("$" 226) (T_VARIABLE 274) (simple_variable 28666) ("{" 28667)) 143 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 28665) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 144 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 28664) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 145 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 28663) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 146 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 28662) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 147 (("(" 28659)) 148 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 28658) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 149 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 28657) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 150 (("!" 224) ("\"" 225) ("$" 226) ("(" 28644) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (clone_argument_list 28645) (constant 285) (dereferenceable_scalar 286) (expr 28646) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 151 (("\"" 1538) ("$" 16543) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 16544) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 16545) ("[" 1560) (array_object_dereferenceable 16546) (callable_expr 16547) (callable_variable 16548) (class_constant 1564) (class_name 16549) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 16550) (name 16551) (new_dereferenceable 1570) (simple_variable 16552) (static_member 16553) (variable 28643) (variable_class_name 16555)) 152 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 28642) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 153 (("(" 28639)) 154 (("(" 28636)) 155 (("(" 28627) (argument_list 28628) (ctor_arguments 28629)) 156 (("\"" 1538) ("$" 16543) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 16544) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 16545) ("[" 1560) (array_object_dereferenceable 16546) (callable_expr 16547) (callable_variable 16548) (class_constant 1564) (class_name 16549) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 16550) (name 16551) (new_dereferenceable 1570) (simple_variable 16552) (static_member 16553) (variable 28626) (variable_class_name 16555)) 157 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 28625) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 158 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 28624) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 159 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 28623) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 160 (("(" 28619)) 161 (("(" 28614)) 162 (("(" 28608)) 163 (("$" 16543) ("(" 28571) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 28572) (T_NAME_QUALIFIED 28573) (T_NAME_RELATIVE 28574) (T_READONLY 2955) (T_STATIC 28575) (T_STRING 28576) (T_VARIABLE 16545) (anonymous_class 28577) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 28578) (attribute 2961) (attributes 28579) (class_modifier 2963) (class_name 28580) (class_name_reference 28581) (name 28582) (new_variable 28583) (simple_variable 28584)) 164 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 28570) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 165 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 28569) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 166 (("(" 16504) (argument_list 28568)) 167 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 28567) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 168 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 28566) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 169 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 28561) (T_END_HEREDOC 28562) (T_VARIABLE 2920) (encaps_list 28563) (encaps_var 2922)) 170 ((T_FN 35) (T_FUNCTION 131) (fn 288) #'289 (inline_function 28560)) 171 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 28559) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 172 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 28558) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 173 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 28557) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 174 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28086) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 175 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 28022) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 176 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 28018) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 177 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 28016) (encaps_list 2828) (encaps_var 2829)) 178 ((T_NULLSAFE_OBJECT_OPERATOR 28002) (T_OBJECT_OPERATOR 28003) ("[" 28004)) 179 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 27999) (attribute 2757) (fn 288) #'289 (inline_function 28000)) 180 (("(" 16504) (argument_list 27998)) 181 ((T_PAAMAYIM_NEKUDOTAYIM 27990)) 182 (("%" 304) (")" 17678) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 183 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 16605)) 184 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 16593)) 185 (("(" 16504) (argument_list 16592)) 186 (("=" 16512) (T_AND_EQUAL 16513) (T_COALESCE_EQUAL 16514) (T_CONCAT_EQUAL 16515) (T_DEC 16516) (T_DIV_EQUAL 16517) (T_INC 16518) (T_MINUS_EQUAL 16519) (T_MOD_EQUAL 16520) (T_MUL_EQUAL 16521) (T_OR_EQUAL 16522) (T_PLUS_EQUAL 16523) (T_POW_EQUAL 16524) (T_SL_EQUAL 16525) (T_SR_EQUAL 16526) (T_XOR_EQUAL 16527)) 187 ((T_PAAMAYIM_NEKUDOTAYIM 16416)) 188 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 303) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 189 (("%" 304) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 190 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16415) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 191 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16414) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 192 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16413) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 193 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16412) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 194 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16411) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 195 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16410) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 196 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16409) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 197 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16408) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 198 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 16403) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 16404) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 199 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16402) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 200 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16401) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 201 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16400) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 202 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16399) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 203 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16398) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 204 (("$" 351) ("(" 352) (T_NAME_FULLY_QUALIFIED 353) (T_NAME_QUALIFIED 354) (T_NAME_RELATIVE 355) (T_STATIC 356) (T_STRING 357) (T_VARIABLE 358) (class_name 359) (class_name_reference 360) (name 361) (new_variable 362) (simple_variable 363)) 205 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 350) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 206 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 349) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 207 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 348) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 208 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 347) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 209 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 346) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 210 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 345) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 211 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 344) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 212 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 343) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 213 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 342) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 214 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 341) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 215 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 340) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 216 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 339) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 217 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 338) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 218 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 337) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 219 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 336) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 220 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 335) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 221 (("$" 351) (T_VARIABLE 358) (simple_variable 16394) ("{" 16395)) 222 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16392) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 223 ((T_PAAMAYIM_NEKUDOTAYIM 16390)) 224 ((T_NULLSAFE_OBJECT_OPERATOR 364) (T_OBJECT_OPERATOR 365) (T_PAAMAYIM_NEKUDOTAYIM 366) ("[" 367)) 225 (("$" 351) (T_STRING 16383) (T_VARIABLE 358) (property_name 16389) (simple_variable 16385) ("{" 16386)) 226 (("$" 351) (T_STRING 16383) (T_VARIABLE 358) (property_name 16384) (simple_variable 16385) ("{" 16386)) 227 (("$" 351) (T_VARIABLE 358) (simple_variable 16382)) 228 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 441) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 229 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 16381) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 230 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 16379) (encaps_var 5357)) 231 (("$" 370) (T_VARIABLE 418) (simple_variable 16375) ("{" 16376)) 232 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16374) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 233 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 16373) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 234 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 16372) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 235 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 16371) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 236 (("(" 16368)) 237 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 16367) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 238 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 16366) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 239 (("!" 368) ("\"" 369) ("$" 370) ("(" 16353) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (clone_argument_list 16354) (constant 429) (dereferenceable_scalar 430) (expr 16355) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 240 (("\"" 1538) ("$" 14563) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 14564) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 14565) ("[" 1560) (array_object_dereferenceable 14566) (callable_expr 14567) (callable_variable 14568) (class_constant 1564) (class_name 14569) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 14570) (name 14571) (new_dereferenceable 1570) (simple_variable 14572) (static_member 14573) (variable 16352) (variable_class_name 14575)) 241 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 16351) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 242 (("(" 16348)) 243 (("(" 16345)) 244 (("(" 16336) (argument_list 16337) (ctor_arguments 16338)) 245 (("\"" 1538) ("$" 14563) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 14564) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 14565) ("[" 1560) (array_object_dereferenceable 14566) (callable_expr 14567) (callable_variable 14568) (class_constant 1564) (class_name 14569) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 14570) (name 14571) (new_dereferenceable 1570) (simple_variable 14572) (static_member 14573) (variable 16335) (variable_class_name 14575)) 246 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 16334) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 247 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 16333) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 248 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 16332) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 249 (("(" 16328)) 250 (("(" 16323)) 251 (("(" 16317)) 252 (("$" 14563) ("(" 16280) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 16281) (T_NAME_QUALIFIED 16282) (T_NAME_RELATIVE 16283) (T_READONLY 2955) (T_STATIC 16284) (T_STRING 16285) (T_VARIABLE 14565) (anonymous_class 16286) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 16287) (attribute 2961) (attributes 16288) (class_modifier 2963) (class_name 16289) (class_name_reference 16290) (name 16291) (new_variable 16292) (simple_variable 16293)) 253 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 16279) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 254 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 16278) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 255 (("(" 14524) (argument_list 16277)) 256 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 16276) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 257 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 16275) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 258 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 16270) (T_END_HEREDOC 16271) (T_VARIABLE 2920) (encaps_list 16272) (encaps_var 2922)) 259 ((T_FN 35) (T_FUNCTION 131) (fn 432) #'433 (inline_function 16269)) 260 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 16268) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 261 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 16267) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 262 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 16266) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 263 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 15795) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 264 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 15731) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 265 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 15727) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 266 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 15725) (encaps_list 2828) (encaps_var 2829)) 267 ((T_NULLSAFE_OBJECT_OPERATOR 15711) (T_OBJECT_OPERATOR 15712) ("[" 15713)) 268 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 15708) (attribute 2757) (fn 432) #'433 (inline_function 15709)) 269 (("(" 14524) (argument_list 15707)) 270 ((T_PAAMAYIM_NEKUDOTAYIM 15699)) 271 (("%" 449) ("*" 450) ("+" 451) ("-" 452) ("." 453) ("/" 454) ("<" 455) (">" 456) ("?" 457) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 458) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 459) (T_BOOLEAN_AND 460) (T_BOOLEAN_OR 461) (T_COALESCE 462) (T_INSTANCEOF 463) (T_IS_EQUAL 464) (T_IS_GREATER_OR_EQUAL 465) (T_IS_IDENTICAL 466) (T_IS_NOT_EQUAL 467) (T_IS_NOT_IDENTICAL 468) (T_IS_SMALLER_OR_EQUAL 469) (T_LOGICAL_AND 470) (T_LOGICAL_OR 471) (T_LOGICAL_XOR 472) (T_PIPE 473) (T_POW 474) (T_SL 475) (T_SPACESHIP 476) (T_SR 477) ("^" 478) ("|" 479)) 272 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 14626)) 273 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 14614)) 274 (("(" 14524) (argument_list 14613)) 275 (("]" 14612)) 276 (("=" 14532) (T_AND_EQUAL 14533) (T_COALESCE_EQUAL 14534) (T_CONCAT_EQUAL 14535) (T_DEC 14536) (T_DIV_EQUAL 14537) (T_INC 14538) (T_MINUS_EQUAL 14539) (T_MOD_EQUAL 14540) (T_MUL_EQUAL 14541) (T_OR_EQUAL 14542) (T_PLUS_EQUAL 14543) (T_POW_EQUAL 14544) (T_SL_EQUAL 14545) (T_SR_EQUAL 14546) (T_XOR_EQUAL 14547)) 277 ((T_PAAMAYIM_NEKUDOTAYIM 14436)) 278 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 448) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 279 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 14435) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 280 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 14434) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 281 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 14433) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 282 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 14432) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 283 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 14431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 284 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 14430) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 285 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 14429) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 286 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 14428) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 287 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 14423) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 14424) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 288 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 14422) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 289 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 14421) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 290 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 14420) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 291 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 14419) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 292 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 14418) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 293 (("$" 496) ("(" 497) (T_NAME_FULLY_QUALIFIED 498) (T_NAME_QUALIFIED 499) (T_NAME_RELATIVE 500) (T_STATIC 501) (T_STRING 502) (T_VARIABLE 503) (class_name 504) (class_name_reference 505) (name 506) (new_variable 507) (simple_variable 508)) 294 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 495) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 295 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 494) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 296 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 493) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 297 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 492) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 298 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 491) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 299 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 490) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 300 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 489) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 301 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 488) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 302 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 487) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 303 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 486) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 304 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 485) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 305 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 484) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 306 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 483) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 307 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 482) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 308 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 481) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 309 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 480) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 310 (("$" 496) (T_VARIABLE 503) (simple_variable 14414) ("{" 14415)) 311 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 14412) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 312 ((T_PAAMAYIM_NEKUDOTAYIM 14410)) 313 ((T_NULLSAFE_OBJECT_OPERATOR 509) (T_OBJECT_OPERATOR 510) (T_PAAMAYIM_NEKUDOTAYIM 511) ("[" 512)) 314 (("$" 496) (T_STRING 516) (T_VARIABLE 503) (property_name 14409) (simple_variable 518) ("{" 519)) 315 (("$" 496) (T_STRING 516) (T_VARIABLE 503) (property_name 517) (simple_variable 518) ("{" 519)) 316 (("$" 496) (T_VARIABLE 503) (simple_variable 515)) 317 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 513) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 318 (("]" 514)) 319 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 583) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 320 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 14408) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 321 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 14406) (encaps_var 5357)) 322 (("$" 522) (T_VARIABLE 570) (simple_variable 14402) ("{" 14403)) 323 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 14401) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 324 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 14400) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 325 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 14399) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 326 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 14398) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 327 (("(" 14395)) 328 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 14394) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 329 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 14393) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 330 (("!" 520) ("\"" 521) ("$" 522) ("(" 14380) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (clone_argument_list 14381) (constant 581) (dereferenceable_scalar 582) (expr 14382) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 331 (("\"" 1538) ("$" 12590) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 12591) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 12592) ("[" 1560) (array_object_dereferenceable 12593) (callable_expr 12594) (callable_variable 12595) (class_constant 1564) (class_name 12596) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 12597) (name 12598) (new_dereferenceable 1570) (simple_variable 12599) (static_member 12600) (variable 14379) (variable_class_name 12602)) 332 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 14378) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 333 (("(" 14375)) 334 (("(" 14372)) 335 (("(" 14363) (argument_list 14364) (ctor_arguments 14365)) 336 (("\"" 1538) ("$" 12590) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 12591) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 12592) ("[" 1560) (array_object_dereferenceable 12593) (callable_expr 12594) (callable_variable 12595) (class_constant 1564) (class_name 12596) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 12597) (name 12598) (new_dereferenceable 1570) (simple_variable 12599) (static_member 12600) (variable 14362) (variable_class_name 12602)) 337 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 14361) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 338 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 14360) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 339 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 14359) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 340 (("(" 14355)) 341 (("(" 14350)) 342 (("(" 14344)) 343 (("$" 12590) ("(" 14307) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 14308) (T_NAME_QUALIFIED 14309) (T_NAME_RELATIVE 14310) (T_READONLY 2955) (T_STATIC 14311) (T_STRING 14312) (T_VARIABLE 12592) (anonymous_class 14313) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 14314) (attribute 2961) (attributes 14315) (class_modifier 2963) (class_name 14316) (class_name_reference 14317) (name 14318) (new_variable 14319) (simple_variable 14320)) 344 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 14306) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 345 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 14305) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 346 (("(" 12551) (argument_list 14304)) 347 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 14303) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 348 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 14302) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 349 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 14297) (T_END_HEREDOC 14298) (T_VARIABLE 2920) (encaps_list 14299) (encaps_var 2922)) 350 ((T_FN 35) (T_FUNCTION 131) (fn 584) #'585 (inline_function 14296)) 351 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 14295) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 352 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 14294) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 353 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 14293) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 354 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 13822) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 355 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 13758) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 356 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 13754) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 357 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 13752) (encaps_list 2828) (encaps_var 2829)) 358 ((T_NULLSAFE_OBJECT_OPERATOR 13738) (T_OBJECT_OPERATOR 13739) ("[" 13740)) 359 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 13735) (attribute 2757) (fn 584) #'585 (inline_function 13736)) 360 (("(" 12551) (argument_list 13734)) 361 ((T_PAAMAYIM_NEKUDOTAYIM 13726)) 362 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 13725)) 363 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 12652)) 364 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 12640)) 365 (("(" 12551) (argument_list 12639)) 366 (("=" 12559) (T_AND_EQUAL 12560) (T_COALESCE_EQUAL 12561) (T_CONCAT_EQUAL 12562) (T_DEC 12563) (T_DIV_EQUAL 12564) (T_INC 12565) (T_MINUS_EQUAL 12566) (T_MOD_EQUAL 12567) (T_MUL_EQUAL 12568) (T_OR_EQUAL 12569) (T_PLUS_EQUAL 12570) (T_POW_EQUAL 12571) (T_SL_EQUAL 12572) (T_SR_EQUAL 12573) (T_XOR_EQUAL 12574)) 367 ((T_PAAMAYIM_NEKUDOTAYIM 12463)) 368 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 599) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 369 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630)) 370 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 12462) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 371 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 12461) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 372 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 12460) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 373 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 12459) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 374 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 12458) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 375 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 12457) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 376 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 12456) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 377 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 12455) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 378 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 693) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 751) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 379 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 686) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 380 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 685) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 381 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 684) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 382 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 683) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 383 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 682) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 384 (("$" 647) ("(" 648) (T_NAME_FULLY_QUALIFIED 649) (T_NAME_QUALIFIED 650) (T_NAME_RELATIVE 651) (T_STATIC 652) (T_STRING 653) (T_VARIABLE 654) (class_name 655) (class_name_reference 656) (name 657) (new_variable 658) (simple_variable 659)) 385 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 646) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 386 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 645) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 387 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 644) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 388 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 643) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 389 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 642) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 390 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 641) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 391 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 640) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 392 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 639) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 393 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 638) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 394 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 637) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 395 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 636) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 396 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 635) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 397 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 634) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 398 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 633) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 399 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 632) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 400 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 631) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 401 (("$" 647) (T_VARIABLE 654) (simple_variable 678) ("{" 679)) 402 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 676) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 403 ((T_PAAMAYIM_NEKUDOTAYIM 674)) 404 ((T_NULLSAFE_OBJECT_OPERATOR 660) (T_OBJECT_OPERATOR 661) (T_PAAMAYIM_NEKUDOTAYIM 662) ("[" 663)) 405 (("$" 647) (T_STRING 667) (T_VARIABLE 654) (property_name 673) (simple_variable 669) ("{" 670)) 406 (("$" 647) (T_STRING 667) (T_VARIABLE 654) (property_name 668) (simple_variable 669) ("{" 670)) 407 (("$" 647) (T_VARIABLE 654) (simple_variable 666)) 408 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 664) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 409 (("]" 665)) 410 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 671) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 411 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 672)) 412 (("$" 647) (T_VARIABLE 654) (simple_variable 675)) 413 (("%" 304) (")" 677) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 414 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 680) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 415 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 681)) 416 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 12454) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 417 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 12452) (encaps_var 5357)) 418 (("$" 689) (T_VARIABLE 738) (simple_variable 12448) ("{" 12449)) 419 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 12447) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 420 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 12446) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 421 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 12445) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 422 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 12444) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 423 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 12443) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 424 (("(" 12440)) 425 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 12439) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 426 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 12438) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 427 (("!" 687) ("\"" 688) ("$" 689) ("(" 12425) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (clone_argument_list 12426) (constant 749) (dereferenceable_scalar 750) (expr 12427) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 428 (("\"" 1538) ("$" 10634) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 10635) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 10636) ("[" 1560) (array_object_dereferenceable 10637) (callable_expr 10638) (callable_variable 10639) (class_constant 1564) (class_name 10640) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 10641) (name 10642) (new_dereferenceable 1570) (simple_variable 10643) (static_member 10644) (variable 12424) (variable_class_name 10646)) 429 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 12423) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 430 (("(" 12420)) 431 (("(" 12417)) 432 (("(" 12408) (argument_list 12409) (ctor_arguments 12410)) 433 (("\"" 1538) ("$" 10634) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 10635) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 10636) ("[" 1560) (array_object_dereferenceable 10637) (callable_expr 10638) (callable_variable 10639) (class_constant 1564) (class_name 10640) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 10641) (name 10642) (new_dereferenceable 1570) (simple_variable 10643) (static_member 10644) (variable 12407) (variable_class_name 10646)) 434 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 12406) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 435 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 12405) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 436 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 12404) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 437 (("(" 12400)) 438 (("(" 12395)) 439 (("(" 12389)) 440 (("$" 10634) ("(" 12352) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 12353) (T_NAME_QUALIFIED 12354) (T_NAME_RELATIVE 12355) (T_READONLY 2955) (T_STATIC 12356) (T_STRING 12357) (T_VARIABLE 10636) (anonymous_class 12358) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 12359) (attribute 2961) (attributes 12360) (class_modifier 2963) (class_name 12361) (class_name_reference 12362) (name 12363) (new_variable 12364) (simple_variable 12365)) 441 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 12351) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 442 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 12350) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 443 (("(" 956) (argument_list 12349)) 444 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 12348) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 445 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 12347) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 446 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 12342) (T_END_HEREDOC 12343) (T_VARIABLE 2920) (encaps_list 12344) (encaps_var 2922)) 447 ((T_FN 35) (T_FUNCTION 131) (fn 752) #'753 (inline_function 12341)) 448 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 12340) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 449 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 12339) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 450 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 12338) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 451 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 11867) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 452 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 11803) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 453 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 11799) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 454 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 11797) (encaps_list 2828) (encaps_var 2829)) 455 ((T_NULLSAFE_OBJECT_OPERATOR 11783) (T_OBJECT_OPERATOR 11784) ("[" 11785)) 456 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 11780) (attribute 2757) (fn 752) #'753 (inline_function 11781)) 457 (("(" 956) (argument_list 11779)) 458 ((T_PAAMAYIM_NEKUDOTAYIM 11771)) 459 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 11769) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 460 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 10696)) 461 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 10684)) 462 (("(" 956) (argument_list 10683)) 463 (("=" 10603) (T_AND_EQUAL 10604) (T_COALESCE_EQUAL 10605) (T_CONCAT_EQUAL 10606) (T_DEC 10607) (T_DIV_EQUAL 10608) (T_INC 10609) (T_MINUS_EQUAL 10610) (T_MOD_EQUAL 10611) (T_MUL_EQUAL 10612) (T_OR_EQUAL 10613) (T_PLUS_EQUAL 10614) (T_POW_EQUAL 10615) (T_SL_EQUAL 10616) (T_SR_EQUAL 10617) (T_XOR_EQUAL 10618)) 464 ((T_PAAMAYIM_NEKUDOTAYIM 868)) 465 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 767) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 466 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 467 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 867) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 468 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 866) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 469 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 865) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 470 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 864) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 471 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 863) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 472 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 862) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 473 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 861) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 474 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 860) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 475 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 855) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 856) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 476 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 854) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 477 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 853) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 478 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 852) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 479 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 851) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 480 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 850) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 481 (("$" 815) ("(" 816) (T_NAME_FULLY_QUALIFIED 817) (T_NAME_QUALIFIED 818) (T_NAME_RELATIVE 819) (T_STATIC 820) (T_STRING 821) (T_VARIABLE 822) (class_name 823) (class_name_reference 824) (name 825) (new_variable 826) (simple_variable 827)) 482 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 814) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 483 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 813) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 484 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 812) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 485 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 811) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 486 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 810) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 487 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 809) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 488 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 808) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 489 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 807) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 490 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 806) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 491 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 805) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 492 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 804) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 493 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 803) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 494 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 802) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 495 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 801) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 496 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 800) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 497 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 799) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 498 (("$" 815) (T_VARIABLE 822) (simple_variable 846) ("{" 847)) 499 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 844) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 500 ((T_PAAMAYIM_NEKUDOTAYIM 842)) 501 ((T_NULLSAFE_OBJECT_OPERATOR 828) (T_OBJECT_OPERATOR 829) (T_PAAMAYIM_NEKUDOTAYIM 830) ("[" 831)) 502 (("$" 815) (T_STRING 835) (T_VARIABLE 822) (property_name 841) (simple_variable 837) ("{" 838)) 503 (("$" 815) (T_STRING 835) (T_VARIABLE 822) (property_name 836) (simple_variable 837) ("{" 838)) 504 (("$" 815) (T_VARIABLE 822) (simple_variable 834)) 505 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 832) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 506 (("]" 833)) 507 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 839) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 508 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 840)) 509 (("$" 815) (T_VARIABLE 822) (simple_variable 843)) 510 (("%" 304) (")" 845) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 511 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 848) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 512 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 849)) 513 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 859) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 514 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 857) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 515 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 858) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 516 (("$" 689) (T_ABSTRACT 869) (T_ARRAY 870) (T_AS 871) (T_BREAK 872) (T_CALLABLE 873) (T_CASE 874) (T_CATCH 875) (T_CLASS 876) (T_CLASS_C 877) (T_CLONE 878) (T_CONST 879) (T_CONTINUE 880) (T_DECLARE 881) (T_DEFAULT 882) (T_DIR 883) (T_DO 884) (T_ECHO 885) (T_ELSE 886) (T_ELSEIF 887) (T_EMPTY 888) (T_ENDDECLARE 889) (T_ENDFOR 890) (T_ENDFOREACH 891) (T_ENDIF 892) (T_ENDSWITCH 893) (T_ENDWHILE 894) (T_ENUM 895) (T_EVAL 896) (T_EXIT 897) (T_EXTENDS 898) (T_FILE 899) (T_FINAL 900) (T_FINALLY 901) (T_FN 902) (T_FOR 903) (T_FOREACH 904) (T_FUNCTION 905) (T_FUNC_C 906) (T_GLOBAL 907) (T_GOTO 908) (T_IF 909) (T_IMPLEMENTS 910) (T_INCLUDE 911) (T_INCLUDE_ONCE 912) (T_INSTANCEOF 913) (T_INSTEADOF 914) (T_INTERFACE 915) (T_ISSET 916) (T_LINE 917) (T_LIST 918) (T_LOGICAL_AND 919) (T_LOGICAL_OR 920) (T_LOGICAL_XOR 921) (T_MATCH 922) (T_METHOD_C 923) (T_NAMESPACE 924) (T_NEW 925) (T_NS_C 926) (T_PRINT 927) (T_PRIVATE 928) (T_PROPERTY_C 929) (T_PROTECTED 930) (T_PUBLIC 931) (T_READONLY 932) (T_REQUIRE 933) (T_REQUIRE_ONCE 934) (T_RETURN 935) (T_STATIC 936) (T_STRING 937) (T_SWITCH 938) (T_THROW 939) (T_TRAIT 940) (T_TRAIT_C 941) (T_TRY 942) (T_UNSET 943) (T_USE 944) (T_VAR 945) (T_VARIABLE 738) (T_WHILE 946) (T_YIELD 947) (identifier 948) (member_name 949) (reserved_non_modifiers 950) (semi_reserved 951) (simple_variable 952) ("{" 953)) 517 (("(" 956) (argument_list 957)) 518 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 954) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 519 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 955)) 520 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 962) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 989) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 1087) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 521 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 10602) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 522 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 10600) (encaps_var 5357)) 523 (("$" 960) (T_VARIABLE 1060) (simple_variable 10596) ("{" 10597)) 524 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 10595) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 525 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 10594) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 526 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 10593) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 527 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 10592) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 528 (("(" 10362)) 529 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 10591) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 530 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 10590) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 531 (("!" 958) ("\"" 959) ("$" 960) ("(" 10349) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (clone_argument_list 10350) (constant 1074) (dereferenceable_scalar 1075) (expr 10351) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 532 (("\"" 1538) ("$" 10262) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 10485) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 10269) ("[" 1560) (array_object_dereferenceable 10486) (callable_expr 10487) (callable_variable 10488) (class_constant 1564) (class_name 10489) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 10490) (name 10491) (new_dereferenceable 1570) (simple_variable 10492) (static_member 10493) (variable 10589) (variable_class_name 10495)) 533 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 10588) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 534 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 10587) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 535 (("(" 10346)) 536 (("(" 10343)) 537 (("(" 10334) (argument_list 10335) (ctor_arguments 10336)) 538 (("\"" 1538) ("$" 10262) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 10485) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 10269) ("[" 1560) (array_object_dereferenceable 10486) (callable_expr 10487) (callable_variable 10488) (class_constant 1564) (class_name 10489) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 10490) (name 10491) (new_dereferenceable 1570) (simple_variable 10492) (static_member 10493) (variable 10586) (variable_class_name 10495)) 539 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 10333) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 540 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 10332) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 541 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 10585) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 542 (("(" 10328)) 543 (("(" 10323)) 544 (("(" 10317)) 545 (("$" 10262) ("(" 10263) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 10264) (T_NAME_QUALIFIED 10265) (T_NAME_RELATIVE 10266) (T_READONLY 2955) (T_STATIC 10267) (T_STRING 10268) (T_VARIABLE 10269) (anonymous_class 10270) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 10271) (attribute 2961) (attributes 10272) (class_modifier 2963) (class_name 10273) (class_name_reference 10274) (name 10275) (new_variable 10276) (simple_variable 10277)) 546 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 10584) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 547 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 10261) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 548 (("(" 10253) (argument_list 10254)) 549 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 10252) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 550 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 10251) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 551 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 10579) (T_END_HEREDOC 10580) (T_VARIABLE 2920) (encaps_list 10581) (encaps_var 2922)) 552 ((T_FN 35) (T_FUNCTION 131) (fn 1077) #'1078 (inline_function 10250)) 553 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 10578) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 554 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 10249) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 555 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 10577) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 556 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1288) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 557 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 10576) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 558 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 10572) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 559 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 10570) (encaps_list 2828) (encaps_var 2829)) 560 ((T_NULLSAFE_OBJECT_OPERATOR 10556) (T_OBJECT_OPERATOR 10557) ("[" 10558)) 561 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 10553) (attribute 2757) (fn 1077) #'1078 (inline_function 10554)) 562 (("(" 10253) (argument_list 10552)) 563 ((T_PAAMAYIM_NEKUDOTAYIM 10544)) 564 (("%" 1125) ("*" 1126) ("+" 1127) ("-" 1128) ("." 1129) ("/" 1130) ("<" 1131) (">" 1132) ("?" 1133) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1134) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1135) (T_BOOLEAN_AND 1136) (T_BOOLEAN_OR 1137) (T_COALESCE 1138) (T_INSTANCEOF 1139) (T_IS_EQUAL 1140) (T_IS_GREATER_OR_EQUAL 1141) (T_IS_IDENTICAL 1142) (T_IS_NOT_EQUAL 1143) (T_IS_NOT_IDENTICAL 1144) (T_IS_SMALLER_OR_EQUAL 1145) (T_LOGICAL_AND 1146) (T_LOGICAL_OR 1147) (T_LOGICAL_XOR 1148) (T_PIPE 1149) (T_POW 1150) (T_SL 1151) (T_SPACESHIP 1152) (T_SR 1153) ("^" 1154) ("|" 1155)) 565 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 10533)) 566 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 10521)) 567 ((":" 10519)) 568 (("(" 10253) (argument_list 10518)) 569 (("," 1498) (possible_comma 10516)) 570 (("=" 10454) (T_AND_EQUAL 10455) (T_COALESCE_EQUAL 10456) (T_CONCAT_EQUAL 10457) (T_DEC 10458) (T_DIV_EQUAL 10459) (T_INC 10460) (T_MINUS_EQUAL 10461) (T_MOD_EQUAL 10462) (T_MUL_EQUAL 10463) (T_OR_EQUAL 10464) (T_PLUS_EQUAL 10465) (T_POW_EQUAL 10466) (T_SL_EQUAL 10467) (T_SR_EQUAL 10468) (T_XOR_EQUAL 10469)) 571 ((T_PAAMAYIM_NEKUDOTAYIM 10365)) 572 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1124) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 573 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1224) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 574 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 575 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1222) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 576 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1221) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 577 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1220) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 578 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1219) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 579 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1218) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 580 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1217) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 581 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 1212) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 1213) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 582 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1211) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 583 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1210) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 584 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1209) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 585 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1208) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 586 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1207) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 587 (("$" 1172) ("(" 1173) (T_NAME_FULLY_QUALIFIED 1174) (T_NAME_QUALIFIED 1175) (T_NAME_RELATIVE 1176) (T_STATIC 1177) (T_STRING 1178) (T_VARIABLE 1179) (class_name 1180) (class_name_reference 1181) (name 1182) (new_variable 1183) (simple_variable 1184)) 588 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1171) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 589 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1170) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 590 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1169) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 591 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1168) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 592 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1167) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 593 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1166) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 594 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1165) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 595 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1164) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 596 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1163) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 597 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1162) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 598 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1161) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 599 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1160) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 600 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1159) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 601 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1158) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 602 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1157) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 603 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1156) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 604 (("$" 1172) (T_VARIABLE 1179) (simple_variable 1203) ("{" 1204)) 605 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 1201) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 606 ((T_PAAMAYIM_NEKUDOTAYIM 1199)) 607 ((T_NULLSAFE_OBJECT_OPERATOR 1185) (T_OBJECT_OPERATOR 1186) (T_PAAMAYIM_NEKUDOTAYIM 1187) ("[" 1188)) 608 (("$" 1172) (T_STRING 1192) (T_VARIABLE 1179) (property_name 1198) (simple_variable 1194) ("{" 1195)) 609 (("$" 1172) (T_STRING 1192) (T_VARIABLE 1179) (property_name 1193) (simple_variable 1194) ("{" 1195)) 610 (("$" 1172) (T_VARIABLE 1179) (simple_variable 1191)) 611 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 1189) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 612 (("]" 1190)) 613 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 1196) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 614 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 1197)) 615 (("$" 1172) (T_VARIABLE 1179) (simple_variable 1200)) 616 (("%" 304) (")" 1202) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 617 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 1205) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 618 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 1206)) 619 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1216) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 620 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 1214) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 621 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1215) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 622 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 10248) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 623 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 10246) (encaps_var 5357)) 624 (("$" 1227) (T_VARIABLE 1275) (simple_variable 10242) ("{" 10243)) 625 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 10241) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 626 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 10240) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 627 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 10239) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 628 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 10238) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 629 (("(" 10235)) 630 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 10234) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 631 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 10233) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 632 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 10220) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (clone_argument_list 10221) (constant 1286) (dereferenceable_scalar 1287) (expr 10222) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 633 (("\"" 1538) ("$" 1539) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 1555) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 1559) ("[" 1560) (array_object_dereferenceable 1561) (callable_expr 1562) (callable_variable 1563) (class_constant 1564) (class_name 1565) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 1568) (name 1569) (new_dereferenceable 1570) (simple_variable 1571) (static_member 1572) (variable 10219) (variable_class_name 1574)) 634 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 10218) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 635 (("(" 10215)) 636 (("(" 10212)) 637 (("(" 10203) (argument_list 10204) (ctor_arguments 10205)) 638 (("\"" 1538) ("$" 1539) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 1555) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 1559) ("[" 1560) (array_object_dereferenceable 1561) (callable_expr 1562) (callable_variable 1563) (class_constant 1564) (class_name 1565) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 1568) (name 1569) (new_dereferenceable 1570) (simple_variable 1571) (static_member 1572) (variable 10202) (variable_class_name 1574)) 639 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 10201) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 640 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 10200) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 641 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 10199) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 642 (("(" 10195)) 643 (("(" 10190)) 644 (("(" 10184)) 645 (("$" 1539) ("(" 10147) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 10148) (T_NAME_QUALIFIED 10149) (T_NAME_RELATIVE 10150) (T_READONLY 2955) (T_STATIC 10151) (T_STRING 10152) (T_VARIABLE 1559) (anonymous_class 10153) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 10154) (attribute 2961) (attributes 10155) (class_modifier 2963) (class_name 10156) (class_name_reference 10157) (name 10158) (new_variable 10159) (simple_variable 10160)) 646 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 10146) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 647 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 10145) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 648 (("(" 1493) (argument_list 10144)) 649 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 10143) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 650 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 10142) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 651 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 10137) (T_END_HEREDOC 10138) (T_VARIABLE 2920) (encaps_list 10139) (encaps_var 2922)) 652 ((T_FN 35) (T_FUNCTION 131) (fn 1289) #'1290 (inline_function 10136)) 653 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 10135) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 654 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 10134) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 655 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 10133) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 656 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 10130) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 657 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 10129) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 658 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 10125) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 659 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 10123) (encaps_list 2828) (encaps_var 2829)) 660 ((T_NULLSAFE_OBJECT_OPERATOR 10109) (T_OBJECT_OPERATOR 10110) ("[" 10111)) 661 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 10106) (attribute 2757) (fn 1289) #'1290 (inline_function 10107)) 662 (("(" 1493) (argument_list 10105)) 663 ((T_PAAMAYIM_NEKUDOTAYIM 10097)) 664 (("%" 1305) ("*" 1306) ("+" 1307) ("-" 1308) ("." 1309) ("/" 1310) ("<" 1311) (">" 1312) ("?" 1313) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1314) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1315) (T_BOOLEAN_AND 1316) (T_BOOLEAN_OR 1317) (T_COALESCE 1318) (T_DOUBLE_ARROW 10095) (T_INSTANCEOF 1319) (T_IS_EQUAL 1320) (T_IS_GREATER_OR_EQUAL 1321) (T_IS_IDENTICAL 1322) (T_IS_NOT_EQUAL 1323) (T_IS_NOT_IDENTICAL 1324) (T_IS_SMALLER_OR_EQUAL 1325) (T_LOGICAL_AND 1326) (T_LOGICAL_OR 1327) (T_LOGICAL_XOR 1328) (T_PIPE 1329) (T_POW 1330) (T_SL 1331) (T_SPACESHIP 1332) (T_SR 1333) ("^" 1334) ("|" 1335)) 665 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 10084)) 666 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 10072)) 667 (("(" 1493) (argument_list 10071)) 668 (("=" 1505) (T_AND_EQUAL 1506) (T_COALESCE_EQUAL 1507) (T_CONCAT_EQUAL 1508) (T_DEC 1509) (T_DIV_EQUAL 1510) (T_INC 1511) (T_MINUS_EQUAL 1512) (T_MOD_EQUAL 1513) (T_MUL_EQUAL 1514) (T_OR_EQUAL 1515) (T_PLUS_EQUAL 1516) (T_POW_EQUAL 1517) (T_SL_EQUAL 1518) (T_SR_EQUAL 1519) (T_XOR_EQUAL 1520)) 669 ((T_PAAMAYIM_NEKUDOTAYIM 1405)) 670 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1304) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 671 (("%" 1305) ("*" 1306) ("+" 1307) ("-" 1308) ("." 1309) ("/" 1310) ("<" 1311) (">" 1312) ("?" 1313) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1314) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1315) (T_BOOLEAN_AND 1316) (T_BOOLEAN_OR 1317) (T_COALESCE 1318) (T_INSTANCEOF 1319) (T_IS_EQUAL 1320) (T_IS_GREATER_OR_EQUAL 1321) (T_IS_IDENTICAL 1322) (T_IS_NOT_EQUAL 1323) (T_IS_NOT_IDENTICAL 1324) (T_IS_SMALLER_OR_EQUAL 1325) (T_LOGICAL_AND 1326) (T_LOGICAL_OR 1327) (T_LOGICAL_XOR 1328) (T_PIPE 1329) (T_POW 1330) (T_SL 1331) (T_SPACESHIP 1332) (T_SR 1333) ("^" 1334) ("|" 1335)) 672 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1404) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 673 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1403) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 674 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1402) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 675 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1401) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 676 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1400) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 677 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1399) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 678 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1398) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 679 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1397) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 680 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 1392) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 1393) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 681 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1391) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 682 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1390) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 683 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1389) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 684 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1388) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 685 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1387) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 686 (("$" 1352) ("(" 1353) (T_NAME_FULLY_QUALIFIED 1354) (T_NAME_QUALIFIED 1355) (T_NAME_RELATIVE 1356) (T_STATIC 1357) (T_STRING 1358) (T_VARIABLE 1359) (class_name 1360) (class_name_reference 1361) (name 1362) (new_variable 1363) (simple_variable 1364)) 687 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1351) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 688 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1350) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 689 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1349) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 690 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1348) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 691 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1347) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 692 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1346) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 693 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1345) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 694 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1344) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 695 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1343) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 696 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1342) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 697 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1341) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 698 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1340) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 699 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1339) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 700 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1338) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 701 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1337) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 702 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1336) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 703 (("$" 1352) (T_VARIABLE 1359) (simple_variable 1383) ("{" 1384)) 704 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 1381) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 705 ((T_PAAMAYIM_NEKUDOTAYIM 1379)) 706 ((T_NULLSAFE_OBJECT_OPERATOR 1365) (T_OBJECT_OPERATOR 1366) (T_PAAMAYIM_NEKUDOTAYIM 1367) ("[" 1368)) 707 (("$" 1352) (T_STRING 1372) (T_VARIABLE 1359) (property_name 1378) (simple_variable 1374) ("{" 1375)) 708 (("$" 1352) (T_STRING 1372) (T_VARIABLE 1359) (property_name 1373) (simple_variable 1374) ("{" 1375)) 709 (("$" 1352) (T_VARIABLE 1359) (simple_variable 1371)) 710 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 1369) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 711 (("]" 1370)) 712 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 1376) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 713 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 1377)) 714 (("$" 1352) (T_VARIABLE 1359) (simple_variable 1380)) 715 (("%" 304) (")" 1382) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 716 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 1385) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 717 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 1386)) 718 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1396) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 719 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 1394) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 720 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1395) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 721 (("$" 1227) (T_ABSTRACT 1406) (T_ARRAY 1407) (T_AS 1408) (T_BREAK 1409) (T_CALLABLE 1410) (T_CASE 1411) (T_CATCH 1412) (T_CLASS 1413) (T_CLASS_C 1414) (T_CLONE 1415) (T_CONST 1416) (T_CONTINUE 1417) (T_DECLARE 1418) (T_DEFAULT 1419) (T_DIR 1420) (T_DO 1421) (T_ECHO 1422) (T_ELSE 1423) (T_ELSEIF 1424) (T_EMPTY 1425) (T_ENDDECLARE 1426) (T_ENDFOR 1427) (T_ENDFOREACH 1428) (T_ENDIF 1429) (T_ENDSWITCH 1430) (T_ENDWHILE 1431) (T_ENUM 1432) (T_EVAL 1433) (T_EXIT 1434) (T_EXTENDS 1435) (T_FILE 1436) (T_FINAL 1437) (T_FINALLY 1438) (T_FN 1439) (T_FOR 1440) (T_FOREACH 1441) (T_FUNCTION 1442) (T_FUNC_C 1443) (T_GLOBAL 1444) (T_GOTO 1445) (T_IF 1446) (T_IMPLEMENTS 1447) (T_INCLUDE 1448) (T_INCLUDE_ONCE 1449) (T_INSTANCEOF 1450) (T_INSTEADOF 1451) (T_INTERFACE 1452) (T_ISSET 1453) (T_LINE 1454) (T_LIST 1455) (T_LOGICAL_AND 1456) (T_LOGICAL_OR 1457) (T_LOGICAL_XOR 1458) (T_MATCH 1459) (T_METHOD_C 1460) (T_NAMESPACE 1461) (T_NEW 1462) (T_NS_C 1463) (T_PRINT 1464) (T_PRIVATE 1465) (T_PROPERTY_C 1466) (T_PROTECTED 1467) (T_PUBLIC 1468) (T_READONLY 1469) (T_REQUIRE 1470) (T_REQUIRE_ONCE 1471) (T_RETURN 1472) (T_STATIC 1473) (T_STRING 1474) (T_SWITCH 1475) (T_THROW 1476) (T_TRAIT 1477) (T_TRAIT_C 1478) (T_TRY 1479) (T_UNSET 1480) (T_USE 1481) (T_VAR 1482) (T_VARIABLE 1275) (T_WHILE 1483) (T_YIELD 1484) (identifier 1485) (member_name 1486) (reserved_non_modifiers 1487) (semi_reserved 1488) (simple_variable 1489) ("{" 1490)) 722 (("(" 1493) (argument_list 1494)) 723 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 1491) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 724 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 1492)) 725 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 1495) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1496) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 1497) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 726 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 1504) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 727 (("," 1498) (possible_comma 1499)) 728 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1502) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 729 ((")" 1500)) 730 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 731 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (ampersand 1536) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1537) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 732 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1533) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 733 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1532) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 734 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1531) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 735 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1530) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 736 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1529) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 737 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1528) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 738 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1527) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 739 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1526) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 740 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1525) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 741 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1524) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 742 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1523) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 743 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1522) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 744 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 1521) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 745 (("\"" 1538) ("$" 1539) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 1555) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 1559) ("[" 1560) (array_object_dereferenceable 1561) (callable_expr 1562) (callable_variable 1563) (class_constant 1564) (class_name 1565) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 1568) (name 1569) (new_dereferenceable 1570) (simple_variable 1571) (static_member 1572) (variable 1573) (variable_class_name 1574)) 746 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 10069) (encaps_var 5357)) 747 (("$" 1539) (T_VARIABLE 1559) (simple_variable 10065) ("{" 10066)) 748 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 10063) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 749 (("(" 10060)) 750 (("$" 10005) ("(" 10006) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 10007) (T_NAME_QUALIFIED 10008) (T_NAME_RELATIVE 10009) (T_READONLY 2955) (T_STATIC 10010) (T_STRING 10011) (T_VARIABLE 10012) (anonymous_class 10013) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 10014) (attribute 2961) (attributes 10015) (class_modifier 2963) (class_name 10016) (class_name_reference 10017) (name 10018) (new_variable 10019) (simple_variable 10020)) 751 (("(" 1663) (argument_list 10004)) 752 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 1754) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 753 ((T_NULLSAFE_OBJECT_OPERATOR 1681) (T_OBJECT_OPERATOR 1682) ("[" 1683)) 754 (("(" 1663) (argument_list 1680)) 755 ((T_PAAMAYIM_NEKUDOTAYIM 1672)) 756 (("(" 1663) (argument_list 1671)) 757 ((T_PAAMAYIM_NEKUDOTAYIM 1575)) 758 (("$" 1539) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 1559) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 1656) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 1659) ("{" 1660)) 759 (("(" 1663) (argument_list 1664)) 760 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 1661) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 761 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 1662)) 762 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 1665) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1666) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 1667) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 763 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 1670) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 764 (("," 1498) (possible_comma 1668)) 765 ((")" 1669)) 766 (("$" 1539) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 1559) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 1674) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 1675) ("{" 1676)) 767 (("(" 1663) (argument_list 1679)) 768 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 1677) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 769 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 1678)) 770 (("$" 1539) (T_STRING 1686) (T_VARIABLE 1559) (property_name 1693) (simple_variable 1688) ("{" 1689)) 771 (("$" 1539) (T_STRING 1686) (T_VARIABLE 1559) (property_name 1687) (simple_variable 1688) ("{" 1689)) 772 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 1684) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 773 (("]" 1685)) 774 (("(" 1663) (argument_list 1692)) 775 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 1690) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 776 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 1691)) 777 (("(" 1663) (argument_list 1694)) 778 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 10003) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 779 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 10001) (encaps_var 5357)) 780 (("$" 1697) (T_VARIABLE 1746) (simple_variable 9997) ("{" 9998)) 781 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 9996) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 782 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 9995) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 783 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 9994) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 784 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 9993) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 785 (("(" 9990)) 786 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 9989) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 787 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 9988) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 788 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 9975) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (clone_argument_list 9976) (constant 1760) (dereferenceable_scalar 1761) (expr 9977) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 789 (("\"" 1538) ("$" 8127) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 8128) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 8129) ("[" 1560) (array_object_dereferenceable 8130) (callable_expr 8131) (callable_variable 8132) (class_constant 1564) (class_name 8133) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 8134) (name 8135) (new_dereferenceable 1570) (simple_variable 8136) (static_member 8137) (variable 9974) (variable_class_name 8139)) 790 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 9973) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 791 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 9972) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 792 (("(" 9969)) 793 (("(" 9966)) 794 (("(" 9957) (argument_list 9958) (ctor_arguments 9959)) 795 (("\"" 1538) ("$" 8127) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 8128) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 8129) ("[" 1560) (array_object_dereferenceable 8130) (callable_expr 8131) (callable_variable 8132) (class_constant 1564) (class_name 8133) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 8134) (name 8135) (new_dereferenceable 1570) (simple_variable 8136) (static_member 8137) (variable 9956) (variable_class_name 8139)) 796 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 9955) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 797 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 9954) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 798 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 9953) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 799 (("(" 9949)) 800 (("(" 9946)) 801 (("(" 9940)) 802 (("$" 8127) ("(" 9903) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 9904) (T_NAME_QUALIFIED 9905) (T_NAME_RELATIVE 9906) (T_READONLY 2955) (T_STATIC 9907) (T_STRING 9908) (T_VARIABLE 8129) (anonymous_class 9909) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 9910) (attribute 2961) (attributes 9911) (class_modifier 2963) (class_name 9912) (class_name_reference 9913) (name 9914) (new_variable 9915) (simple_variable 9916)) 803 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 9902) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 804 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 9901) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 805 (("(" 8088) (argument_list 9900)) 806 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 9899) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 807 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 9898) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 808 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 9893) (T_END_HEREDOC 9894) (T_VARIABLE 2920) (encaps_list 9895) (encaps_var 2922)) 809 ((T_FN 35) (T_FUNCTION 131) (fn 1763) #'1764 (inline_function 9892)) 810 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 9891) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 811 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 9890) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 812 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 9889) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 813 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 9886) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 814 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 9885) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 815 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 9881) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 816 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 9879) (encaps_list 2828) (encaps_var 2829)) 817 (("\"" 1538) ("$" 9678) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 9679) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 9680) ("[" 1560) (array_object_dereferenceable 9681) (callable_expr 9682) (callable_variable 9683) (class_constant 1564) (class_name 9684) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 9685) (name 9686) (new_dereferenceable 1570) (simple_variable 9687) (static_member 9688) (variable 9878) (variable_class_name 9690)) 818 ((T_NULLSAFE_OBJECT_OPERATOR 9864) (T_OBJECT_OPERATOR 9865) ("[" 9866)) 819 (("]" 9863)) 820 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 9860) (attribute 2757) (fn 1763) #'1764 (inline_function 9861)) 821 (("(" 8088) (argument_list 9859)) 822 ((T_PAAMAYIM_NEKUDOTAYIM 9851)) 823 (("%" 1782) ("*" 1783) ("+" 1784) ("-" 1785) ("." 1786) ("/" 1787) ("<" 1788) (">" 1789) ("?" 1790) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1791) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1792) (T_BOOLEAN_AND 1793) (T_BOOLEAN_OR 1794) (T_COALESCE 1795) (T_DOUBLE_ARROW 8733) (T_INSTANCEOF 1796) (T_IS_EQUAL 1797) (T_IS_GREATER_OR_EQUAL 1798) (T_IS_IDENTICAL 1799) (T_IS_NOT_EQUAL 1800) (T_IS_NOT_IDENTICAL 1801) (T_IS_SMALLER_OR_EQUAL 1802) (T_LOGICAL_AND 1803) (T_LOGICAL_OR 1804) (T_LOGICAL_XOR 1805) (T_PIPE 1806) (T_POW 1807) (T_SL 1808) (T_SPACESHIP 1809) (T_SR 1810) ("^" 1811) ("|" 1812)) 824 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 8191)) 825 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 8179)) 826 (("(" 8088) (argument_list 8178)) 827 (("," 8176)) 828 (("=" 8096) (T_AND_EQUAL 8097) (T_COALESCE_EQUAL 8098) (T_CONCAT_EQUAL 8099) (T_DEC 8100) (T_DIV_EQUAL 8101) (T_INC 8102) (T_MINUS_EQUAL 8103) (T_MOD_EQUAL 8104) (T_MUL_EQUAL 8105) (T_OR_EQUAL 8106) (T_PLUS_EQUAL 8107) (T_POW_EQUAL 8108) (T_SL_EQUAL 8109) (T_SR_EQUAL 8110) (T_XOR_EQUAL 8111)) 829 ((T_PAAMAYIM_NEKUDOTAYIM 8000)) 830 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1781) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 831 (("(" 1882)) 832 (("%" 1782) ("*" 1783) ("+" 1784) ("-" 1785) ("." 1786) ("/" 1787) ("<" 1788) (">" 1789) ("?" 1790) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1791) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1792) (T_BOOLEAN_AND 1793) (T_BOOLEAN_OR 1794) (T_COALESCE 1795) (T_INSTANCEOF 1796) (T_IS_EQUAL 1797) (T_IS_GREATER_OR_EQUAL 1798) (T_IS_IDENTICAL 1799) (T_IS_NOT_EQUAL 1800) (T_IS_NOT_IDENTICAL 1801) (T_IS_SMALLER_OR_EQUAL 1802) (T_LOGICAL_AND 1803) (T_LOGICAL_OR 1804) (T_LOGICAL_XOR 1805) (T_PIPE 1806) (T_POW 1807) (T_SL 1808) (T_SPACESHIP 1809) (T_SR 1810) ("^" 1811) ("|" 1812)) 833 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1881) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 834 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1880) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 835 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1879) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 836 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1878) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 837 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1877) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 838 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1876) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 839 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1875) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 840 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1874) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 841 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 1869) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 1870) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 842 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1868) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 843 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1867) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 844 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1866) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 845 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1865) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 846 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1864) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 847 (("$" 1829) ("(" 1830) (T_NAME_FULLY_QUALIFIED 1831) (T_NAME_QUALIFIED 1832) (T_NAME_RELATIVE 1833) (T_STATIC 1834) (T_STRING 1835) (T_VARIABLE 1836) (class_name 1837) (class_name_reference 1838) (name 1839) (new_variable 1840) (simple_variable 1841)) 848 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1828) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 849 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1827) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 850 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1826) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 851 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1825) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 852 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1824) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 853 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1823) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 854 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1822) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 855 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1821) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 856 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1820) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 857 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1819) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 858 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1818) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 859 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1817) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 860 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1816) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 861 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1815) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 862 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1814) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 863 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1813) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 864 (("$" 1829) (T_VARIABLE 1836) (simple_variable 1860) ("{" 1861)) 865 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 1858) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 866 ((T_PAAMAYIM_NEKUDOTAYIM 1856)) 867 ((T_NULLSAFE_OBJECT_OPERATOR 1842) (T_OBJECT_OPERATOR 1843) (T_PAAMAYIM_NEKUDOTAYIM 1844) ("[" 1845)) 868 (("$" 1829) (T_STRING 1849) (T_VARIABLE 1836) (property_name 1855) (simple_variable 1851) ("{" 1852)) 869 (("$" 1829) (T_STRING 1849) (T_VARIABLE 1836) (property_name 1850) (simple_variable 1851) ("{" 1852)) 870 (("$" 1829) (T_VARIABLE 1836) (simple_variable 1848)) 871 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 1846) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 872 (("]" 1847)) 873 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 1853) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 874 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 1854)) 875 (("$" 1829) (T_VARIABLE 1836) (simple_variable 1857)) 876 (("%" 304) (")" 1859) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 877 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 1862) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 878 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 1863)) 879 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1873) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 880 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 1871) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 881 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1872) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 882 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 1942) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 883 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 7999) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 884 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 7997) (encaps_var 5357)) 885 (("$" 1885) (T_VARIABLE 1934) (simple_variable 7993) ("{" 7994)) 886 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 7992) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 887 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 7991) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 888 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 7990) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 889 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 7989) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 890 (("(" 7986)) 891 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 7985) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 892 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 7984) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 893 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 7971) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (clone_argument_list 7972) (constant 1948) (dereferenceable_scalar 1949) (expr 7973) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 894 (("\"" 1538) ("$" 2202) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 2203) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 2204) ("[" 1560) (array_object_dereferenceable 2205) (callable_expr 2206) (callable_variable 2207) (class_constant 1564) (class_name 2208) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 2209) (name 2210) (new_dereferenceable 1570) (simple_variable 2211) (static_member 2212) (variable 7970) (variable_class_name 2214)) 895 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 7969) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 896 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 7968) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 897 (("(" 7965)) 898 (("(" 7962)) 899 (("(" 7953) (argument_list 7954) (ctor_arguments 7955)) 900 (("\"" 1538) ("$" 2202) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 2203) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 2204) ("[" 1560) (array_object_dereferenceable 2205) (callable_expr 2206) (callable_variable 2207) (class_constant 1564) (class_name 2208) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 2209) (name 2210) (new_dereferenceable 1570) (simple_variable 2211) (static_member 2212) (variable 7952) (variable_class_name 2214)) 901 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 7951) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 902 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 7950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 903 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 7949) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 904 (("(" 7945)) 905 (("(" 7942)) 906 (("(" 7936)) 907 (("$" 2202) ("(" 7899) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 7900) (T_NAME_QUALIFIED 7901) (T_NAME_RELATIVE 7902) (T_READONLY 2955) (T_STATIC 7903) (T_STRING 7904) (T_VARIABLE 2204) (anonymous_class 7905) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 7906) (attribute 2961) (attributes 7907) (class_modifier 2963) (class_name 7908) (class_name_reference 7909) (name 7910) (new_variable 7911) (simple_variable 7912)) 908 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 7898) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 909 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 7897) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 910 (("(" 2163) (argument_list 7896)) 911 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 7895) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 912 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 7894) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 913 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 7889) (T_END_HEREDOC 7890) (T_VARIABLE 2920) (encaps_list 7891) (encaps_var 2922)) 914 ((T_FN 35) (T_FUNCTION 131) (fn 1951) #'1952 (inline_function 7888)) 915 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 7887) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 916 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 7886) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 917 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 7885) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 918 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 7882) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 919 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 7881) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 920 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 7877) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 921 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 7875) (encaps_list 2828) (encaps_var 2829)) 922 (("\"" 1538) ("$" 2774) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 2775) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 2776) ("[" 1560) (array_object_dereferenceable 2777) (callable_expr 2778) (callable_variable 2779) (class_constant 1564) (class_name 2780) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 2781) (name 2782) (new_dereferenceable 1570) (simple_variable 2783) (static_member 2784) (variable 7874) (variable_class_name 2786)) 923 ((T_NULLSAFE_OBJECT_OPERATOR 7860) (T_OBJECT_OPERATOR 7861) ("[" 7862)) 924 ((")" 7857)) 925 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 7854) (attribute 2757) (fn 1951) #'1952 (inline_function 7855)) 926 (("(" 2163) (argument_list 7853)) 927 ((T_PAAMAYIM_NEKUDOTAYIM 7845)) 928 (("%" 1970) ("*" 1971) ("+" 1972) ("-" 1973) ("." 1974) ("/" 1975) ("<" 1976) (">" 1977) ("?" 1978) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1979) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1980) (T_BOOLEAN_AND 1981) (T_BOOLEAN_OR 1982) (T_COALESCE 1983) (T_DOUBLE_ARROW 2359) (T_INSTANCEOF 1984) (T_IS_EQUAL 1985) (T_IS_GREATER_OR_EQUAL 1986) (T_IS_IDENTICAL 1987) (T_IS_NOT_EQUAL 1988) (T_IS_NOT_IDENTICAL 1989) (T_IS_SMALLER_OR_EQUAL 1990) (T_LOGICAL_AND 1991) (T_LOGICAL_OR 1992) (T_LOGICAL_XOR 1993) (T_PIPE 1994) (T_POW 1995) (T_SL 1996) (T_SPACESHIP 1997) (T_SR 1998) ("^" 1999) ("|" 2000)) 929 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 2266)) 930 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 2254)) 931 (("(" 2163) (argument_list 2253)) 932 (("," 2251)) 933 (("=" 2171) (T_AND_EQUAL 2172) (T_COALESCE_EQUAL 2173) (T_CONCAT_EQUAL 2174) (T_DEC 2175) (T_DIV_EQUAL 2176) (T_INC 2177) (T_MINUS_EQUAL 2178) (T_MOD_EQUAL 2179) (T_MUL_EQUAL 2180) (T_OR_EQUAL 2181) (T_PLUS_EQUAL 2182) (T_POW_EQUAL 2183) (T_SL_EQUAL 2184) (T_SR_EQUAL 2185) (T_XOR_EQUAL 2186)) 934 ((T_PAAMAYIM_NEKUDOTAYIM 2075)) 935 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1969) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 936 (("(" 2070)) 937 (("%" 1970) ("*" 1971) ("+" 1972) ("-" 1973) ("." 1974) ("/" 1975) ("<" 1976) (">" 1977) ("?" 1978) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1979) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1980) (T_BOOLEAN_AND 1981) (T_BOOLEAN_OR 1982) (T_COALESCE 1983) (T_INSTANCEOF 1984) (T_IS_EQUAL 1985) (T_IS_GREATER_OR_EQUAL 1986) (T_IS_IDENTICAL 1987) (T_IS_NOT_EQUAL 1988) (T_IS_NOT_IDENTICAL 1989) (T_IS_SMALLER_OR_EQUAL 1990) (T_LOGICAL_AND 1991) (T_LOGICAL_OR 1992) (T_LOGICAL_XOR 1993) (T_PIPE 1994) (T_POW 1995) (T_SL 1996) (T_SPACESHIP 1997) (T_SR 1998) ("^" 1999) ("|" 2000)) 938 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2069) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 939 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2068) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 940 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2067) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 941 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2066) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 942 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2065) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 943 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2064) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 944 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2063) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 945 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2062) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 946 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 2057) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 2058) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 947 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2056) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 948 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2055) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 949 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2054) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 950 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2053) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 951 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2052) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 952 (("$" 2017) ("(" 2018) (T_NAME_FULLY_QUALIFIED 2019) (T_NAME_QUALIFIED 2020) (T_NAME_RELATIVE 2021) (T_STATIC 2022) (T_STRING 2023) (T_VARIABLE 2024) (class_name 2025) (class_name_reference 2026) (name 2027) (new_variable 2028) (simple_variable 2029)) 953 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2016) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 954 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2015) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 955 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2014) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 956 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2013) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 957 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2012) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 958 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2011) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 959 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2010) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 960 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2009) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 961 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2008) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 962 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2007) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 963 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2006) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 964 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2005) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 965 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2004) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 966 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2003) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 967 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2002) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 968 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2001) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 969 (("$" 2017) (T_VARIABLE 2024) (simple_variable 2048) ("{" 2049)) 970 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 2046) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 971 ((T_PAAMAYIM_NEKUDOTAYIM 2044)) 972 ((T_NULLSAFE_OBJECT_OPERATOR 2030) (T_OBJECT_OPERATOR 2031) (T_PAAMAYIM_NEKUDOTAYIM 2032) ("[" 2033)) 973 (("$" 2017) (T_STRING 2037) (T_VARIABLE 2024) (property_name 2043) (simple_variable 2039) ("{" 2040)) 974 (("$" 2017) (T_STRING 2037) (T_VARIABLE 2024) (property_name 2038) (simple_variable 2039) ("{" 2040)) 975 (("$" 2017) (T_VARIABLE 2024) (simple_variable 2036)) 976 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 2034) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 977 (("]" 2035)) 978 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 2041) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 979 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 2042)) 980 (("$" 2017) (T_VARIABLE 2024) (simple_variable 2045)) 981 (("%" 304) (")" 2047) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 982 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 2050) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 983 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 2051)) 984 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2061) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 985 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 2059) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 986 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2060) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 987 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 2071) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 988 ((")" 2072)) 989 (("=" 2073)) 990 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2074) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 991 (("$" 1885) (T_ABSTRACT 2076) (T_ARRAY 2077) (T_AS 2078) (T_BREAK 2079) (T_CALLABLE 2080) (T_CASE 2081) (T_CATCH 2082) (T_CLASS 2083) (T_CLASS_C 2084) (T_CLONE 2085) (T_CONST 2086) (T_CONTINUE 2087) (T_DECLARE 2088) (T_DEFAULT 2089) (T_DIR 2090) (T_DO 2091) (T_ECHO 2092) (T_ELSE 2093) (T_ELSEIF 2094) (T_EMPTY 2095) (T_ENDDECLARE 2096) (T_ENDFOR 2097) (T_ENDFOREACH 2098) (T_ENDIF 2099) (T_ENDSWITCH 2100) (T_ENDWHILE 2101) (T_ENUM 2102) (T_EVAL 2103) (T_EXIT 2104) (T_EXTENDS 2105) (T_FILE 2106) (T_FINAL 2107) (T_FINALLY 2108) (T_FN 2109) (T_FOR 2110) (T_FOREACH 2111) (T_FUNCTION 2112) (T_FUNC_C 2113) (T_GLOBAL 2114) (T_GOTO 2115) (T_IF 2116) (T_IMPLEMENTS 2117) (T_INCLUDE 2118) (T_INCLUDE_ONCE 2119) (T_INSTANCEOF 2120) (T_INSTEADOF 2121) (T_INTERFACE 2122) (T_ISSET 2123) (T_LINE 2124) (T_LIST 2125) (T_LOGICAL_AND 2126) (T_LOGICAL_OR 2127) (T_LOGICAL_XOR 2128) (T_MATCH 2129) (T_METHOD_C 2130) (T_NAMESPACE 2131) (T_NEW 2132) (T_NS_C 2133) (T_PRINT 2134) (T_PRIVATE 2135) (T_PROPERTY_C 2136) (T_PROTECTED 2137) (T_PUBLIC 2138) (T_READONLY 2139) (T_REQUIRE 2140) (T_REQUIRE_ONCE 2141) (T_RETURN 2142) (T_STATIC 2143) (T_STRING 2144) (T_SWITCH 2145) (T_THROW 2146) (T_TRAIT 2147) (T_TRAIT_C 2148) (T_TRY 2149) (T_UNSET 2150) (T_USE 2151) (T_VAR 2152) (T_VARIABLE 1934) (T_WHILE 2153) (T_YIELD 2154) (identifier 2155) (member_name 2156) (reserved_non_modifiers 2157) (semi_reserved 2158) (simple_variable 2159) ("{" 2160)) 992 (("(" 2163) (argument_list 2164)) 993 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 2161) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 994 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 2162)) 995 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 2165) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 2166) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 2167) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 996 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 2170) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 997 (("," 1498) (possible_comma 2168)) 998 ((")" 2169)) 999 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 2200) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2201) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 1000 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2199) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 1001 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2198) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 1002 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2197) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 1003 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2196) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 1004 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2195) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 1005 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2194) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 1006 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2193) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 1007 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2192) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 1008 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2191) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 1009 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2190) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 1010 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2189) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 1011 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2188) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 1012 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2187) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 1013 (("\"" 1538) ("$" 2202) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 2203) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 2204) ("[" 1560) (array_object_dereferenceable 2205) (callable_expr 2206) (callable_variable 2207) (class_constant 1564) (class_name 2208) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 2209) (name 2210) (new_dereferenceable 1570) (simple_variable 2211) (static_member 2212) (variable 2213) (variable_class_name 2214)) 1014 (("$" 2202) (T_VARIABLE 2204) (simple_variable 2247) ("{" 2248)) 1015 (("(" 2218) (argument_list 2246)) 1016 ((T_NULLSAFE_OBJECT_OPERATOR 2232) (T_OBJECT_OPERATOR 2233) ("[" 2234)) 1017 (("(" 2218) (argument_list 2231)) 1018 ((T_PAAMAYIM_NEKUDOTAYIM 2227)) 1019 (("(" 2218) (argument_list 2226)) 1020 ((T_PAAMAYIM_NEKUDOTAYIM 2215)) 1021 (("$" 2202) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 2204) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 2216) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 2217) ("{" 1660)) 1022 (("(" 2218) (argument_list 2219)) 1023 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 2220) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 2221) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 2222) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 1024 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 2225) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 1025 (("," 1498) (possible_comma 2223)) 1026 ((")" 2224)) 1027 (("$" 2202) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 2204) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 2228) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 2229) ("{" 1676)) 1028 (("(" 2218) (argument_list 2230)) 1029 (("$" 2202) (T_STRING 2237) (T_VARIABLE 2204) (property_name 2244) (simple_variable 2239) ("{" 2240)) 1030 (("$" 2202) (T_STRING 2237) (T_VARIABLE 2204) (property_name 2238) (simple_variable 2239) ("{" 2240)) 1031 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 2235) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 1032 (("]" 2236)) 1033 (("(" 2218) (argument_list 2243)) 1034 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 2241) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 1035 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 2242)) 1036 (("(" 2218) (argument_list 2245)) 1037 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 2249) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 1038 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 2250)) 1039 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (possible_array_pair 2252) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 1040 ((backup_doc_comment 2255)) 1041 (("(" 2256)) 1042 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 2257)) 1043 ((")" 2258)) 1044 ((T_USE 166) (lexical_vars 2259)) 1045 ((":" 168) (return_type 2260)) 1046 ((backup_fn_flags 2261)) 1047 (("{" 2262)) 1048 ((inner_statement_list 2263)) 1049 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 2264) ("~" 129)) 1050 ((backup_fn_flags 2265)) 1051 ((backup_doc_comment 2267)) 1052 (("(" 2268)) 1053 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 2269)) 1054 ((")" 2270)) 1055 ((":" 2271) (return_type 2272)) 1056 (("(" 2278) ("?" 2279) (T_ARRAY 2280) (T_CALLABLE 2281) (T_NAME_FULLY_QUALIFIED 2282) (T_NAME_QUALIFIED 2283) (T_NAME_RELATIVE 2284) (T_STATIC 2285) (T_STRING 2286) (intersection_type 2287) (name 2288) (type 2289) (type_expr 2290) (type_without_static 2291) (union_type 2292) (union_type_element 2293)) 1057 ((T_DOUBLE_ARROW 2273)) 1058 ((backup_fn_flags 2274)) 1059 ((backup_lex_pos 2275)) 1060 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 2276) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 1061 (("%" 1305) ("*" 1306) ("+" 1307) ("-" 1308) ("." 1309) ("/" 1310) ("<" 1311) (">" 1312) ("?" 1313) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1314) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1315) (T_BOOLEAN_AND 1316) (T_BOOLEAN_OR 1317) (T_COALESCE 1318) (T_INSTANCEOF 1319) (T_IS_EQUAL 1320) (T_IS_GREATER_OR_EQUAL 1321) (T_IS_IDENTICAL 1322) (T_IS_NOT_EQUAL 1323) (T_IS_NOT_IDENTICAL 1324) (T_IS_SMALLER_OR_EQUAL 1325) (T_LOGICAL_AND 1326) (T_LOGICAL_OR 1327) (T_LOGICAL_XOR 1328) (T_PIPE 1329) (T_POW 1330) (T_SL 1331) (T_SPACESHIP 1332) (T_SR 1333) ("^" 1334) (backup_fn_flags 2277) ("|" 1335)) 1062 ((T_ARRAY 2307) (T_CALLABLE 2308) (T_NAME_FULLY_QUALIFIED 2309) (T_NAME_QUALIFIED 2310) (T_NAME_RELATIVE 2311) (T_STATIC 2312) (T_STRING 2313) (intersection_type 2357) (name 2315) (type 2316) (type_without_static 2317)) 1063 ((T_ARRAY 2347) (T_CALLABLE 2348) (T_NAME_FULLY_QUALIFIED 2349) (T_NAME_QUALIFIED 2350) (T_NAME_RELATIVE 2351) (T_STATIC 2352) (T_STRING 2353) (name 2354) (type 2355) (type_without_static 2356)) 1064 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2345)) 1065 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2334)) 1066 (("|" 2332)) 1067 (("|" 2294)) 1068 (("(" 2295) (T_ARRAY 2296) (T_CALLABLE 2297) (T_NAME_FULLY_QUALIFIED 2298) (T_NAME_QUALIFIED 2299) (T_NAME_RELATIVE 2300) (T_STATIC 2301) (T_STRING 2302) (name 2303) (type 2304) (type_without_static 2305) (union_type_element 2306)) 1069 ((T_ARRAY 2307) (T_CALLABLE 2308) (T_NAME_FULLY_QUALIFIED 2309) (T_NAME_QUALIFIED 2310) (T_NAME_RELATIVE 2311) (T_STATIC 2312) (T_STRING 2313) (intersection_type 2314) (name 2315) (type 2316) (type_without_static 2317)) 1070 ((")" 2329) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2330)) 1071 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2318)) 1072 ((T_ARRAY 2319) (T_CALLABLE 2320) (T_NAME_FULLY_QUALIFIED 2321) (T_NAME_QUALIFIED 2322) (T_NAME_RELATIVE 2323) (T_STATIC 2324) (T_STRING 2325) (name 2326) (type 2327) (type_without_static 2328)) 1073 ((T_ARRAY 2319) (T_CALLABLE 2320) (T_NAME_FULLY_QUALIFIED 2321) (T_NAME_QUALIFIED 2322) (T_NAME_RELATIVE 2323) (T_STATIC 2324) (T_STRING 2325) (name 2326) (type 2331) (type_without_static 2328)) 1074 (("(" 2295) (T_ARRAY 2296) (T_CALLABLE 2297) (T_NAME_FULLY_QUALIFIED 2298) (T_NAME_QUALIFIED 2299) (T_NAME_RELATIVE 2300) (T_STATIC 2301) (T_STRING 2302) (name 2303) (type 2304) (type_without_static 2305) (union_type_element 2333)) 1075 ((T_ARRAY 2335) (T_CALLABLE 2336) (T_NAME_FULLY_QUALIFIED 2337) (T_NAME_QUALIFIED 2338) (T_NAME_RELATIVE 2339) (T_STATIC 2340) (T_STRING 2341) (name 2342) (type 2343) (type_without_static 2344)) 1076 ((T_ARRAY 2335) (T_CALLABLE 2336) (T_NAME_FULLY_QUALIFIED 2337) (T_NAME_QUALIFIED 2338) (T_NAME_RELATIVE 2339) (T_STATIC 2340) (T_STRING 2341) (name 2342) (type 2346) (type_without_static 2344)) 1077 ((")" 2358) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2330)) 1078 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2388) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (ampersand 2415) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2424) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1079 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 7844) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1080 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 7842) (encaps_var 5357)) 1081 (("$" 2362) (T_VARIABLE 2410) (simple_variable 7838) ("{" 7839)) 1082 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 7837) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 1083 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 7836) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1084 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 7835) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1085 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 7834) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1086 (("(" 5336)) 1087 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 7833) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1088 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 7832) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1089 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 5323) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (clone_argument_list 5324) (constant 2422) (dereferenceable_scalar 2423) (expr 5325) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1090 (("\"" 1538) ("$" 2674) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 2675) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 2676) ("[" 1560) (array_object_dereferenceable 2677) (callable_expr 2678) (callable_variable 2679) (class_constant 1564) (class_name 2680) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 2681) (name 2682) (new_dereferenceable 1570) (simple_variable 2683) (static_member 2684) (variable 7831) (variable_class_name 2686)) 1091 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 7830) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1092 (("(" 5319)) 1093 (("(" 5316)) 1094 (("(" 5307) (argument_list 5308) (ctor_arguments 5309)) 1095 (("\"" 1538) ("$" 2674) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 2675) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 2676) ("[" 1560) (array_object_dereferenceable 2677) (callable_expr 2678) (callable_variable 2679) (class_constant 1564) (class_name 2680) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 2681) (name 2682) (new_dereferenceable 1570) (simple_variable 2683) (static_member 2684) (variable 7829) (variable_class_name 2686)) 1096 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 5306) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1097 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 5305) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1098 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 7828) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1099 (("(" 5301)) 1100 (("(" 7825)) 1101 (("(" 5295)) 1102 (("$" 2674) ("(" 2950) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 2952) (T_NAME_QUALIFIED 2953) (T_NAME_RELATIVE 2954) (T_READONLY 2955) (T_STATIC 2956) (T_STRING 2957) (T_VARIABLE 2676) (anonymous_class 2958) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 2960) (attribute 2961) (attributes 2962) (class_modifier 2963) (class_name 2964) (class_name_reference 2965) (name 2966) (new_variable 2967) (simple_variable 2968)) 1103 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2949) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1104 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2948) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1105 (("(" 2635) (argument_list 2947)) 1106 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2946) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1107 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2945) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1108 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2918) (T_END_HEREDOC 2919) (T_VARIABLE 2920) (encaps_list 2921) (encaps_var 2922)) 1109 ((T_FN 35) (T_FUNCTION 131) (fn 2425) #'2426 (inline_function 2915)) 1110 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2914) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1111 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2913) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1112 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2912) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1113 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 2909) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 1114 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2908) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1115 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 2904) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 1116 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 2827) (encaps_list 2828) (encaps_var 2829)) 1117 (("\"" 1538) ("$" 2774) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 2775) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 2776) ("[" 1560) (array_object_dereferenceable 2777) (callable_expr 2778) (callable_variable 2779) (class_constant 1564) (class_name 2780) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 2781) (name 2782) (new_dereferenceable 1570) (simple_variable 2783) (static_member 2784) (variable 2785) (variable_class_name 2786)) 1118 ((T_NULLSAFE_OBJECT_OPERATOR 2760) (T_OBJECT_OPERATOR 2761) ("[" 2762)) 1119 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 2756) (attribute 2757) (fn 2425) #'2426 (inline_function 2758)) 1120 (("(" 2635) (argument_list 2755)) 1121 ((T_PAAMAYIM_NEKUDOTAYIM 2747)) 1122 (("%" 2442) ("*" 2443) ("+" 2444) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 1123 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 2736)) 1124 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 2724)) 1125 (("(" 2635) (argument_list 2723)) 1126 (("=" 2643) (T_AND_EQUAL 2644) (T_COALESCE_EQUAL 2645) (T_CONCAT_EQUAL 2646) (T_DEC 2647) (T_DIV_EQUAL 2648) (T_INC 2649) (T_MINUS_EQUAL 2650) (T_MOD_EQUAL 2651) (T_MUL_EQUAL 2652) (T_OR_EQUAL 2653) (T_PLUS_EQUAL 2654) (T_POW_EQUAL 2655) (T_SL_EQUAL 2656) (T_SR_EQUAL 2657) (T_XOR_EQUAL 2658)) 1127 ((T_PAAMAYIM_NEKUDOTAYIM 2547)) 1128 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2441) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1129 (("(" 2542)) 1130 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2541) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1131 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2540) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1132 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2539) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1133 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2538) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1134 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2537) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1135 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2536) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1136 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2535) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1137 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2534) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1138 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 2529) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 2530) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 1139 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2528) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1140 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2527) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1141 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2526) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1142 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2525) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1143 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2524) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1144 (("$" 2489) ("(" 2490) (T_NAME_FULLY_QUALIFIED 2491) (T_NAME_QUALIFIED 2492) (T_NAME_RELATIVE 2493) (T_STATIC 2494) (T_STRING 2495) (T_VARIABLE 2496) (class_name 2497) (class_name_reference 2498) (name 2499) (new_variable 2500) (simple_variable 2501)) 1145 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2488) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1146 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2487) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1147 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2486) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1148 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2485) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1149 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2484) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1150 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2483) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1151 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2482) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1152 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2481) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1153 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2480) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1154 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2479) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1155 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2478) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1156 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2477) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1157 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2476) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1158 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2475) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1159 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2474) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1160 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2473) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1161 (("$" 2489) (T_VARIABLE 2496) (simple_variable 2520) ("{" 2521)) 1162 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 2518) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 1163 ((T_PAAMAYIM_NEKUDOTAYIM 2516)) 1164 ((T_NULLSAFE_OBJECT_OPERATOR 2502) (T_OBJECT_OPERATOR 2503) (T_PAAMAYIM_NEKUDOTAYIM 2504) ("[" 2505)) 1165 (("$" 2489) (T_STRING 2509) (T_VARIABLE 2496) (property_name 2515) (simple_variable 2511) ("{" 2512)) 1166 (("$" 2489) (T_STRING 2509) (T_VARIABLE 2496) (property_name 2510) (simple_variable 2511) ("{" 2512)) 1167 (("$" 2489) (T_VARIABLE 2496) (simple_variable 2508)) 1168 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 2506) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 1169 (("]" 2507)) 1170 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 2513) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 1171 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 2514)) 1172 (("$" 2489) (T_VARIABLE 2496) (simple_variable 2517)) 1173 (("%" 304) (")" 2519) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 1174 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 2522) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 1175 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 2523)) 1176 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2533) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1177 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 2531) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 1178 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2532) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1179 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 2543) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 1180 ((")" 2544)) 1181 (("=" 2545)) 1182 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2546) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1183 (("$" 2362) (T_ABSTRACT 2548) (T_ARRAY 2549) (T_AS 2550) (T_BREAK 2551) (T_CALLABLE 2552) (T_CASE 2553) (T_CATCH 2554) (T_CLASS 2555) (T_CLASS_C 2556) (T_CLONE 2557) (T_CONST 2558) (T_CONTINUE 2559) (T_DECLARE 2560) (T_DEFAULT 2561) (T_DIR 2562) (T_DO 2563) (T_ECHO 2564) (T_ELSE 2565) (T_ELSEIF 2566) (T_EMPTY 2567) (T_ENDDECLARE 2568) (T_ENDFOR 2569) (T_ENDFOREACH 2570) (T_ENDIF 2571) (T_ENDSWITCH 2572) (T_ENDWHILE 2573) (T_ENUM 2574) (T_EVAL 2575) (T_EXIT 2576) (T_EXTENDS 2577) (T_FILE 2578) (T_FINAL 2579) (T_FINALLY 2580) (T_FN 2581) (T_FOR 2582) (T_FOREACH 2583) (T_FUNCTION 2584) (T_FUNC_C 2585) (T_GLOBAL 2586) (T_GOTO 2587) (T_IF 2588) (T_IMPLEMENTS 2589) (T_INCLUDE 2590) (T_INCLUDE_ONCE 2591) (T_INSTANCEOF 2592) (T_INSTEADOF 2593) (T_INTERFACE 2594) (T_ISSET 2595) (T_LINE 2596) (T_LIST 2597) (T_LOGICAL_AND 2598) (T_LOGICAL_OR 2599) (T_LOGICAL_XOR 2600) (T_MATCH 2601) (T_METHOD_C 2602) (T_NAMESPACE 2603) (T_NEW 2604) (T_NS_C 2605) (T_PRINT 2606) (T_PRIVATE 2607) (T_PROPERTY_C 2608) (T_PROTECTED 2609) (T_PUBLIC 2610) (T_READONLY 2611) (T_REQUIRE 2612) (T_REQUIRE_ONCE 2613) (T_RETURN 2614) (T_STATIC 2615) (T_STRING 2616) (T_SWITCH 2617) (T_THROW 2618) (T_TRAIT 2619) (T_TRAIT_C 2620) (T_TRY 2621) (T_UNSET 2622) (T_USE 2623) (T_VAR 2624) (T_VARIABLE 2410) (T_WHILE 2625) (T_YIELD 2626) (identifier 2627) (member_name 2628) (reserved_non_modifiers 2629) (semi_reserved 2630) (simple_variable 2631) ("{" 2632)) 1184 (("(" 2635) (argument_list 2636)) 1185 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 2633) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 1186 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 2634)) 1187 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 2637) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 2638) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 2639) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 1188 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 2642) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 1189 (("," 1498) (possible_comma 2640)) 1190 ((")" 2641)) 1191 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (ampersand 2672) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2673) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1192 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2671) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1193 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2670) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1194 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2669) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1195 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2668) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1196 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2667) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1197 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2666) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1198 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2665) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1199 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2664) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1200 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2663) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1201 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2662) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1202 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2661) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1203 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2660) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1204 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2659) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1205 (("\"" 1538) ("$" 2674) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 2675) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 2676) ("[" 1560) (array_object_dereferenceable 2677) (callable_expr 2678) (callable_variable 2679) (class_constant 1564) (class_name 2680) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 2681) (name 2682) (new_dereferenceable 1570) (simple_variable 2683) (static_member 2684) (variable 2685) (variable_class_name 2686)) 1206 (("$" 2674) (T_VARIABLE 2676) (simple_variable 2719) ("{" 2720)) 1207 (("(" 2690) (argument_list 2718)) 1208 ((T_NULLSAFE_OBJECT_OPERATOR 2704) (T_OBJECT_OPERATOR 2705) ("[" 2706)) 1209 (("(" 2690) (argument_list 2703)) 1210 ((T_PAAMAYIM_NEKUDOTAYIM 2699)) 1211 (("(" 2690) (argument_list 2698)) 1212 ((T_PAAMAYIM_NEKUDOTAYIM 2687)) 1213 (("$" 2674) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 2676) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 2688) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 2689) ("{" 1660)) 1214 (("(" 2690) (argument_list 2691)) 1215 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 2692) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 2693) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 2694) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 1216 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 2697) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 1217 (("," 1498) (possible_comma 2695)) 1218 ((")" 2696)) 1219 (("$" 2674) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 2676) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 2700) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 2701) ("{" 1676)) 1220 (("(" 2690) (argument_list 2702)) 1221 (("$" 2674) (T_STRING 2709) (T_VARIABLE 2676) (property_name 2716) (simple_variable 2711) ("{" 2712)) 1222 (("$" 2674) (T_STRING 2709) (T_VARIABLE 2676) (property_name 2710) (simple_variable 2711) ("{" 2712)) 1223 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 2707) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 1224 (("]" 2708)) 1225 (("(" 2690) (argument_list 2715)) 1226 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 2713) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 1227 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 2714)) 1228 (("(" 2690) (argument_list 2717)) 1229 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 2721) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 1230 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 2722)) 1231 ((backup_doc_comment 2725)) 1232 (("(" 2726)) 1233 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 2727)) 1234 ((")" 2728)) 1235 ((T_USE 166) (lexical_vars 2729)) 1236 ((":" 168) (return_type 2730)) 1237 ((backup_fn_flags 2731)) 1238 (("{" 2732)) 1239 ((inner_statement_list 2733)) 1240 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 2734) ("~" 129)) 1241 ((backup_fn_flags 2735)) 1242 ((backup_doc_comment 2737)) 1243 (("(" 2738)) 1244 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 2739)) 1245 ((")" 2740)) 1246 ((":" 2271) (return_type 2741)) 1247 ((T_DOUBLE_ARROW 2742)) 1248 ((backup_fn_flags 2743)) 1249 ((backup_lex_pos 2744)) 1250 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 2745) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 1251 (("%" 1125) ("*" 1126) ("+" 1127) ("-" 1128) ("." 1129) ("/" 1130) ("<" 1131) (">" 1132) ("?" 1133) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1134) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1135) (T_BOOLEAN_AND 1136) (T_BOOLEAN_OR 1137) (T_COALESCE 1138) (T_INSTANCEOF 1139) (T_IS_EQUAL 1140) (T_IS_GREATER_OR_EQUAL 1141) (T_IS_IDENTICAL 1142) (T_IS_NOT_EQUAL 1143) (T_IS_NOT_IDENTICAL 1144) (T_IS_SMALLER_OR_EQUAL 1145) (T_LOGICAL_AND 1146) (T_LOGICAL_OR 1147) (T_LOGICAL_XOR 1148) (T_PIPE 1149) (T_POW 1150) (T_SL 1151) (T_SPACESHIP 1152) (T_SR 1153) ("^" 1154) (backup_fn_flags 2746) ("|" 1155)) 1252 (("$" 2362) (T_ABSTRACT 2548) (T_ARRAY 2549) (T_AS 2550) (T_BREAK 2551) (T_CALLABLE 2552) (T_CASE 2553) (T_CATCH 2554) (T_CLASS 2555) (T_CLASS_C 2556) (T_CLONE 2557) (T_CONST 2558) (T_CONTINUE 2559) (T_DECLARE 2560) (T_DEFAULT 2561) (T_DIR 2562) (T_DO 2563) (T_ECHO 2564) (T_ELSE 2565) (T_ELSEIF 2566) (T_EMPTY 2567) (T_ENDDECLARE 2568) (T_ENDFOR 2569) (T_ENDFOREACH 2570) (T_ENDIF 2571) (T_ENDSWITCH 2572) (T_ENDWHILE 2573) (T_ENUM 2574) (T_EVAL 2575) (T_EXIT 2576) (T_EXTENDS 2577) (T_FILE 2578) (T_FINAL 2579) (T_FINALLY 2580) (T_FN 2581) (T_FOR 2582) (T_FOREACH 2583) (T_FUNCTION 2584) (T_FUNC_C 2585) (T_GLOBAL 2586) (T_GOTO 2587) (T_IF 2588) (T_IMPLEMENTS 2589) (T_INCLUDE 2590) (T_INCLUDE_ONCE 2591) (T_INSTANCEOF 2592) (T_INSTEADOF 2593) (T_INTERFACE 2594) (T_ISSET 2595) (T_LINE 2596) (T_LIST 2597) (T_LOGICAL_AND 2598) (T_LOGICAL_OR 2599) (T_LOGICAL_XOR 2600) (T_MATCH 2601) (T_METHOD_C 2602) (T_NAMESPACE 2603) (T_NEW 2604) (T_NS_C 2605) (T_PRINT 2606) (T_PRIVATE 2607) (T_PROPERTY_C 2608) (T_PROTECTED 2609) (T_PUBLIC 2610) (T_READONLY 2611) (T_REQUIRE 2612) (T_REQUIRE_ONCE 2613) (T_RETURN 2614) (T_STATIC 2615) (T_STRING 2616) (T_SWITCH 2617) (T_THROW 2618) (T_TRAIT 2619) (T_TRAIT_C 2620) (T_TRY 2621) (T_UNSET 2622) (T_USE 2623) (T_VAR 2624) (T_VARIABLE 2410) (T_WHILE 2625) (T_YIELD 2626) (identifier 2748) (member_name 2749) (reserved_non_modifiers 2629) (semi_reserved 2630) (simple_variable 2750) ("{" 2751)) 1253 (("(" 2635) (argument_list 2754)) 1254 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 2752) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 1255 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 2753)) 1256 ((T_FN 35) (T_FUNCTION 131) (fn 2425) #'2426 (inline_function 2759)) 1257 (("$" 2362) (T_STRING 2765) (T_VARIABLE 2410) (property_name 2772) (simple_variable 2767) ("{" 2768)) 1258 (("$" 2362) (T_STRING 2765) (T_VARIABLE 2410) (property_name 2766) (simple_variable 2767) ("{" 2768)) 1259 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 2763) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 1260 (("]" 2764)) 1261 (("(" 2635) (argument_list 2771)) 1262 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 2769) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 1263 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 2770)) 1264 (("(" 2635) (argument_list 2773)) 1265 (("$" 2774) (T_VARIABLE 2776) (simple_variable 2819) ("{" 2820)) 1266 (("(" 2790) (argument_list 2818)) 1267 ((T_NULLSAFE_OBJECT_OPERATOR 2804) (T_OBJECT_OPERATOR 2805) ("[" 2806)) 1268 (("(" 2790) (argument_list 2803)) 1269 ((T_PAAMAYIM_NEKUDOTAYIM 2799)) 1270 (("(" 2790) (argument_list 2798)) 1271 ((T_PAAMAYIM_NEKUDOTAYIM 2787)) 1272 (("$" 2774) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 2776) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 2788) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 2789) ("{" 1660)) 1273 (("(" 2790) (argument_list 2791)) 1274 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 2792) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 2793) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 2794) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 1275 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 2797) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 1276 (("," 1498) (possible_comma 2795)) 1277 ((")" 2796)) 1278 (("$" 2774) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 2776) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 2800) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 2801) ("{" 1676)) 1279 (("(" 2790) (argument_list 2802)) 1280 (("$" 2774) (T_STRING 2809) (T_VARIABLE 2776) (property_name 2816) (simple_variable 2811) ("{" 2812)) 1281 (("$" 2774) (T_STRING 2809) (T_VARIABLE 2776) (property_name 2810) (simple_variable 2811) ("{" 2812)) 1282 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 2807) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 1283 (("]" 2808)) 1284 (("(" 2790) (argument_list 2815)) 1285 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 2813) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 1286 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 2814)) 1287 (("(" 2790) (argument_list 2817)) 1288 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 2821) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 1289 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 2822)) 1290 (("\"" 1538) ("$" 2854) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 2855) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 2856) ("[" 1560) (array_object_dereferenceable 2857) (callable_expr 2858) (callable_variable 2859) (class_constant 1564) (class_name 2860) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 2861) (name 2862) (new_dereferenceable 1570) (simple_variable 2863) (static_member 2864) (variable 2865) (variable_class_name 2866)) 1291 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_STRING_VARNAME 2846) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 2847) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 1292 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_VARIABLE 2826) (encaps_var 2845)) 1293 ((T_NULLSAFE_OBJECT_OPERATOR 2833) (T_OBJECT_OPERATOR 2834) ("[" 2835)) 1294 (("`" 2832)) 1295 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2830) (T_VARIABLE 2826) (encaps_var 2831)) 1296 ((T_STRING 2844)) 1297 ((T_STRING 2843)) 1298 (("-" 2836) (T_NUM_STRING 2837) (T_STRING 2838) (T_VARIABLE 2839) (encaps_var_offset 2840)) 1299 ((T_NUM_STRING 2842)) 1300 (("]" 2841)) 1301 (("[" 2849) ("}" 2850)) 1302 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 2848)) 1303 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 2851) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 1304 (("%" 449) ("*" 450) ("+" 451) ("-" 452) ("." 453) ("/" 454) ("<" 455) (">" 456) ("?" 457) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 458) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 459) (T_BOOLEAN_AND 460) (T_BOOLEAN_OR 461) (T_COALESCE 462) (T_INSTANCEOF 463) (T_IS_EQUAL 464) (T_IS_GREATER_OR_EQUAL 465) (T_IS_IDENTICAL 466) (T_IS_NOT_EQUAL 467) (T_IS_NOT_IDENTICAL 468) (T_IS_SMALLER_OR_EQUAL 469) (T_LOGICAL_AND 470) (T_LOGICAL_OR 471) (T_LOGICAL_XOR 472) (T_PIPE 473) (T_POW 474) (T_SL 475) (T_SPACESHIP 476) (T_SR 477) ("]" 2852) ("^" 478) ("|" 479)) 1305 (("}" 2853)) 1306 (("$" 2854) (T_VARIABLE 2856) (simple_variable 2900) ("{" 2901)) 1307 (("(" 2870) (argument_list 2899)) 1308 ((T_NULLSAFE_OBJECT_OPERATOR 2885) (T_OBJECT_OPERATOR 2886) ("[" 2887)) 1309 (("(" 2870) (argument_list 2884)) 1310 ((T_PAAMAYIM_NEKUDOTAYIM 2880)) 1311 (("(" 2870) (argument_list 2879)) 1312 (("}" 2878)) 1313 ((T_PAAMAYIM_NEKUDOTAYIM 2867)) 1314 (("$" 2854) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 2856) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 2868) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 2869) ("{" 1660)) 1315 (("(" 2870) (argument_list 2871)) 1316 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 2872) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 2873) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 2874) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 1317 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 2877) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 1318 (("," 1498) (possible_comma 2875)) 1319 ((")" 2876)) 1320 (("$" 2854) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 2856) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 2881) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 2882) ("{" 1676)) 1321 (("(" 2870) (argument_list 2883)) 1322 (("$" 2854) (T_STRING 2890) (T_VARIABLE 2856) (property_name 2897) (simple_variable 2892) ("{" 2893)) 1323 (("$" 2854) (T_STRING 2890) (T_VARIABLE 2856) (property_name 2891) (simple_variable 2892) ("{" 2893)) 1324 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 2888) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 1325 (("]" 2889)) 1326 (("(" 2870) (argument_list 2896)) 1327 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 2894) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 1328 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 2895)) 1329 (("(" 2870) (argument_list 2898)) 1330 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 2902) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 1331 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 2903)) 1332 (("]" 2905)) 1333 (("=" 2906)) 1334 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2907) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1335 (("%" 1970) ("*" 1971) ("+" 1972) ("-" 1973) ("." 1974) ("/" 1975) ("<" 1976) (">" 1977) ("?" 1978) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1979) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1980) (T_BOOLEAN_AND 1981) (T_BOOLEAN_OR 1982) (T_COALESCE 1983) (T_DOUBLE_ARROW 2910) (T_INSTANCEOF 1984) (T_IS_EQUAL 1985) (T_IS_GREATER_OR_EQUAL 1986) (T_IS_IDENTICAL 1987) (T_IS_NOT_EQUAL 1988) (T_IS_NOT_IDENTICAL 1989) (T_IS_SMALLER_OR_EQUAL 1990) (T_LOGICAL_AND 1991) (T_LOGICAL_OR 1992) (T_LOGICAL_XOR 1993) (T_PIPE 1994) (T_POW 1995) (T_SL 1996) (T_SPACESHIP 1997) (T_SR 1998) ("^" 1999) ("|" 2000)) 1336 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 2911) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 1337 (("\"" 1538) ("$" 2854) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 2855) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 2856) ("[" 1560) (array_object_dereferenceable 2857) (callable_expr 2858) (callable_variable 2859) (class_constant 1564) (class_name 2860) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 2861) (name 2862) (new_dereferenceable 1570) (simple_variable 2863) (static_member 2864) (variable 2943) (variable_class_name 2866)) 1338 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_STRING_VARNAME 2935) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 2936) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 1339 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 2933) (T_VARIABLE 2920) (encaps_var 2934)) 1340 ((T_NULLSAFE_OBJECT_OPERATOR 2926) (T_OBJECT_OPERATOR 2927) ("[" 2928)) 1341 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 2924) (T_VARIABLE 2920) (encaps_var 2925)) 1342 ((T_STRING 2932)) 1343 ((T_STRING 2931)) 1344 (("-" 2836) (T_NUM_STRING 2837) (T_STRING 2838) (T_VARIABLE 2839) (encaps_var_offset 2929)) 1345 (("]" 2930)) 1346 (("[" 2938) ("}" 2939)) 1347 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 2937)) 1348 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 2940) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 1349 (("%" 449) ("*" 450) ("+" 451) ("-" 452) ("." 453) ("/" 454) ("<" 455) (">" 456) ("?" 457) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 458) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 459) (T_BOOLEAN_AND 460) (T_BOOLEAN_OR 461) (T_COALESCE 462) (T_INSTANCEOF 463) (T_IS_EQUAL 464) (T_IS_GREATER_OR_EQUAL 465) (T_IS_IDENTICAL 466) (T_IS_NOT_EQUAL 467) (T_IS_NOT_IDENTICAL 468) (T_IS_SMALLER_OR_EQUAL 469) (T_LOGICAL_AND 470) (T_LOGICAL_OR 471) (T_LOGICAL_XOR 472) (T_PIPE 473) (T_POW 474) (T_SL 475) (T_SPACESHIP 476) (T_SR 477) ("]" 2941) ("^" 478) ("|" 479)) 1350 (("}" 2942)) 1351 (("}" 2944)) 1352 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 7823) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 1353 ((T_NAME_FULLY_QUALIFIED 5996) (T_NAME_QUALIFIED 5997) (T_NAME_RELATIVE 5998) (T_STATIC 5999) (T_STRING 6000) (attribute_decl 6001) (attribute_group 7820) (class_name 6003) (name 6004)) 1354 ((T_ABSTRACT 10) (T_FINAL 34) (T_READONLY 2955) (class_modifier 7819)) 1355 ((T_CLASS 2983)) 1356 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 2981) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 2960) (attribute 2982) (class_modifier 2963)) 1357 ((T_PAAMAYIM_NEKUDOTAYIM 2979)) 1358 (("(" 2690) (argument_list 2978)) 1359 ((T_NULLSAFE_OBJECT_OPERATOR 2969) (T_OBJECT_OPERATOR 2970) (T_PAAMAYIM_NEKUDOTAYIM 2971) ("[" 2972)) 1360 (("$" 2674) (T_STRING 2709) (T_VARIABLE 2676) (property_name 2977) (simple_variable 2711) ("{" 2712)) 1361 (("$" 2674) (T_STRING 2709) (T_VARIABLE 2676) (property_name 2976) (simple_variable 2711) ("{" 2712)) 1362 (("$" 2674) (T_VARIABLE 2676) (simple_variable 2975)) 1363 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 2973) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 1364 (("]" 2974)) 1365 (("$" 2674) (T_VARIABLE 2676) (simple_variable 2980)) 1366 (("(" 2984) (argument_list 2985) (ctor_arguments 2986)) 1367 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 7813) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 7814) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 7815) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 1368 ((T_EXTENDS 2987) (extends_from 2988)) 1369 ((T_NAME_FULLY_QUALIFIED 7806) (T_NAME_QUALIFIED 7807) (T_NAME_RELATIVE 7808) (T_STATIC 7809) (T_STRING 7810) (class_name 7811) (name 7812)) 1370 ((T_IMPLEMENTS 2989) (implements_list 2990)) 1371 ((T_NAME_FULLY_QUALIFIED 7796) (T_NAME_QUALIFIED 7797) (T_NAME_RELATIVE 7798) (T_STATIC 7799) (T_STRING 7800) (class_name 7801) (class_name_list 7802) (name 7803)) 1372 ((backup_doc_comment 2991)) 1373 (("{" 2992)) 1374 ((class_statement_list 2993)) 1375 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 3018)) 1376 ((T_NAME_FULLY_QUALIFIED 5996) (T_NAME_QUALIFIED 5997) (T_NAME_RELATIVE 5998) (T_STATIC 5999) (T_STRING 6000) (attribute_decl 6001) (attribute_group 7793) (class_name 6003) (name 6004)) 1377 ((backup_doc_comment 7706)) 1378 ((T_NAME_FULLY_QUALIFIED 7405) (T_NAME_QUALIFIED 7406) (T_NAME_RELATIVE 7407) (T_STATIC 7408) (T_STRING 7409) (class_name 7410) (class_name_list 7411) (name 7412)) 1379 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_VAR 3007) (attribute 7403) (attributed_class_statement 7404) (class_const_modifiers 3011) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017)) 1380 ((T_CONST 7253)) 1381 ((T_FUNCTION 7097) #'7098) 1382 ((T_ABSTRACT 2994) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (member_modifier 7096)) 1383 (("(" 3019) ("?" 3020) (T_ARRAY 3021) (T_CALLABLE 3022) (T_NAME_FULLY_QUALIFIED 3023) (T_NAME_QUALIFIED 3024) (T_NAME_RELATIVE 3025) (T_STRING 3026) (intersection_type_without_static 3027) (name 3028) (optional_type_without_static 3029) (type_expr_without_static 3030) (type_without_static 3031) (union_type_without_static 3032) (union_type_without_static_element 3033)) 1384 ((T_ARRAY 2307) (T_CALLABLE 2308) (T_NAME_FULLY_QUALIFIED 2309) (T_NAME_QUALIFIED 2310) (T_NAME_RELATIVE 2311) (T_STRING 2313) (intersection_type_without_static 7094) (name 2315) (type_without_static 3046)) 1385 ((T_ARRAY 7086) (T_CALLABLE 7087) (T_NAME_FULLY_QUALIFIED 7088) (T_NAME_QUALIFIED 7089) (T_NAME_RELATIVE 7090) (T_STRING 7091) (name 7092) (type_without_static 7093)) 1386 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 7084)) 1387 ((T_VARIABLE 3063) (hooked_property 3064) (property 3065) (property_list 3066)) 1388 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3054)) 1389 (("|" 3052)) 1390 (("|" 3034)) 1391 (("(" 3035) (T_ARRAY 3036) (T_CALLABLE 3037) (T_NAME_FULLY_QUALIFIED 3038) (T_NAME_QUALIFIED 3039) (T_NAME_RELATIVE 3040) (T_STRING 3041) (name 3042) (type_without_static 3043) (union_type_without_static_element 3044)) 1392 ((T_ARRAY 2307) (T_CALLABLE 2308) (T_NAME_FULLY_QUALIFIED 2309) (T_NAME_QUALIFIED 2310) (T_NAME_RELATIVE 2311) (T_STRING 2313) (intersection_type_without_static 3045) (name 2315) (type_without_static 3046)) 1393 ((")" 3049) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3050)) 1394 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3047)) 1395 ((T_ARRAY 2319) (T_CALLABLE 2320) (T_NAME_FULLY_QUALIFIED 2321) (T_NAME_QUALIFIED 2322) (T_NAME_RELATIVE 2323) (T_STRING 2325) (name 2326) (type_without_static 3048)) 1396 ((T_ARRAY 2319) (T_CALLABLE 2320) (T_NAME_FULLY_QUALIFIED 2321) (T_NAME_QUALIFIED 2322) (T_NAME_RELATIVE 2323) (T_STRING 2325) (name 2326) (type_without_static 3051)) 1397 (("(" 3035) (T_ARRAY 3036) (T_CALLABLE 3037) (T_NAME_FULLY_QUALIFIED 3038) (T_NAME_QUALIFIED 3039) (T_NAME_RELATIVE 3040) (T_STRING 3041) (name 3042) (type_without_static 3043) (union_type_without_static_element 3053)) 1398 ((T_ARRAY 3055) (T_CALLABLE 3056) (T_NAME_FULLY_QUALIFIED 3057) (T_NAME_QUALIFIED 3058) (T_NAME_RELATIVE 3059) (T_STRING 3060) (name 3061) (type_without_static 3062)) 1399 (("=" 5851) (backup_doc_comment 5852)) 1400 (("," 3067) (";" 3068)) 1401 ((T_VARIABLE 3069) (property 3070)) 1402 (("=" 3071) (backup_doc_comment 3072)) 1403 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3136) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1404 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 5850) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1405 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 5848) (encaps_var 5357)) 1406 (("$" 3075) (T_VARIABLE 3123) (simple_variable 5844) ("{" 5845)) 1407 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 5843) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 1408 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 5842) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1409 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 5841) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1410 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 5840) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1411 (("(" 5837)) 1412 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 5836) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1413 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 5835) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1414 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 5822) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (clone_argument_list 5823) (constant 3134) (dereferenceable_scalar 3135) (expr 5824) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1415 (("\"" 1538) ("$" 3380) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 3381) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 3382) ("[" 1560) (array_object_dereferenceable 3383) (callable_expr 3384) (callable_variable 3385) (class_constant 1564) (class_name 3386) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 3387) (name 3388) (new_dereferenceable 1570) (simple_variable 3389) (static_member 3390) (variable 5821) (variable_class_name 3392)) 1416 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 5820) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1417 (("(" 5817)) 1418 (("(" 5814)) 1419 (("(" 5805) (argument_list 5806) (ctor_arguments 5807)) 1420 (("\"" 1538) ("$" 3380) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 3381) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 3382) ("[" 1560) (array_object_dereferenceable 3383) (callable_expr 3384) (callable_variable 3385) (class_constant 1564) (class_name 3386) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 3387) (name 3388) (new_dereferenceable 1570) (simple_variable 3389) (static_member 3390) (variable 5804) (variable_class_name 3392)) 1421 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 5803) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1422 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 5802) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1423 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 5801) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1424 (("(" 5797)) 1425 (("(" 5792)) 1426 (("(" 5786)) 1427 (("$" 3380) ("(" 5749) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 5750) (T_NAME_QUALIFIED 5751) (T_NAME_RELATIVE 5752) (T_READONLY 2955) (T_STATIC 5753) (T_STRING 5754) (T_VARIABLE 3382) (anonymous_class 5755) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 5756) (attribute 2961) (attributes 5757) (class_modifier 2963) (class_name 5758) (class_name_reference 5759) (name 5760) (new_variable 5761) (simple_variable 5762)) 1428 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 5748) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1429 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 5747) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1430 (("(" 3341) (argument_list 5746)) 1431 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 5745) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1432 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 5744) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1433 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 5739) (T_END_HEREDOC 5740) (T_VARIABLE 2920) (encaps_list 5741) (encaps_var 2922)) 1434 ((T_FN 35) (T_FUNCTION 131) (fn 3137) #'3138 (inline_function 5738)) 1435 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 5737) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1436 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 5736) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1437 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 5735) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1438 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3550) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1439 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3486) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1440 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 3482) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 1441 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 3480) (encaps_list 2828) (encaps_var 2829)) 1442 ((T_NULLSAFE_OBJECT_OPERATOR 3466) (T_OBJECT_OPERATOR 3467) ("[" 3468)) 1443 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 3463) (attribute 2757) (fn 3137) #'3138 (inline_function 3464)) 1444 (("(" 3341) (argument_list 3462)) 1445 ((T_PAAMAYIM_NEKUDOTAYIM 3454)) 1446 (("%" 3153) ("*" 3154) ("+" 3155) ("-" 3156) ("." 3157) ("/" 3158) ("<" 3159) (">" 3160) ("?" 3161) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 3162) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3163) (T_BOOLEAN_AND 3164) (T_BOOLEAN_OR 3165) (T_COALESCE 3166) (T_INSTANCEOF 3167) (T_IS_EQUAL 3168) (T_IS_GREATER_OR_EQUAL 3169) (T_IS_IDENTICAL 3170) (T_IS_NOT_EQUAL 3171) (T_IS_NOT_IDENTICAL 3172) (T_IS_SMALLER_OR_EQUAL 3173) (T_LOGICAL_AND 3174) (T_LOGICAL_OR 3175) (T_LOGICAL_XOR 3176) (T_PIPE 3177) (T_POW 3178) (T_SL 3179) (T_SPACESHIP 3180) (T_SR 3181) ("^" 3182) (backup_doc_comment 3453) ("|" 3183)) 1447 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 3442)) 1448 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 3430)) 1449 (("(" 3341) (argument_list 3429)) 1450 (("=" 3349) (T_AND_EQUAL 3350) (T_COALESCE_EQUAL 3351) (T_CONCAT_EQUAL 3352) (T_DEC 3353) (T_DIV_EQUAL 3354) (T_INC 3355) (T_MINUS_EQUAL 3356) (T_MOD_EQUAL 3357) (T_MUL_EQUAL 3358) (T_OR_EQUAL 3359) (T_PLUS_EQUAL 3360) (T_POW_EQUAL 3361) (T_SL_EQUAL 3362) (T_SR_EQUAL 3363) (T_XOR_EQUAL 3364)) 1451 ((T_PAAMAYIM_NEKUDOTAYIM 3253)) 1452 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3152) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1453 (("%" 3153) ("*" 3154) ("+" 3155) ("-" 3156) ("." 3157) ("/" 3158) ("<" 3159) (">" 3160) ("?" 3161) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 3162) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3163) (T_BOOLEAN_AND 3164) (T_BOOLEAN_OR 3165) (T_COALESCE 3166) (T_INSTANCEOF 3167) (T_IS_EQUAL 3168) (T_IS_GREATER_OR_EQUAL 3169) (T_IS_IDENTICAL 3170) (T_IS_NOT_EQUAL 3171) (T_IS_NOT_IDENTICAL 3172) (T_IS_SMALLER_OR_EQUAL 3173) (T_LOGICAL_AND 3174) (T_LOGICAL_OR 3175) (T_LOGICAL_XOR 3176) (T_PIPE 3177) (T_POW 3178) (T_SL 3179) (T_SPACESHIP 3180) (T_SR 3181) ("^" 3182) ("|" 3183)) 1454 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3252) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1455 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3251) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1456 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3250) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1457 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3249) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1458 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3248) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1459 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3247) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1460 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3246) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1461 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3245) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1462 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 3240) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 3241) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 1463 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3239) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1464 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3238) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1465 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3237) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1466 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3236) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1467 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3235) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1468 (("$" 3200) ("(" 3201) (T_NAME_FULLY_QUALIFIED 3202) (T_NAME_QUALIFIED 3203) (T_NAME_RELATIVE 3204) (T_STATIC 3205) (T_STRING 3206) (T_VARIABLE 3207) (class_name 3208) (class_name_reference 3209) (name 3210) (new_variable 3211) (simple_variable 3212)) 1469 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3199) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1470 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3198) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1471 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3197) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1472 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3196) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1473 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3195) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1474 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3194) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1475 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3193) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1476 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3192) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1477 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3191) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1478 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3190) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1479 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3189) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1480 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3188) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1481 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3187) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1482 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3186) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1483 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3185) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1484 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3184) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1485 (("$" 3200) (T_VARIABLE 3207) (simple_variable 3231) ("{" 3232)) 1486 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 3229) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 1487 ((T_PAAMAYIM_NEKUDOTAYIM 3227)) 1488 ((T_NULLSAFE_OBJECT_OPERATOR 3213) (T_OBJECT_OPERATOR 3214) (T_PAAMAYIM_NEKUDOTAYIM 3215) ("[" 3216)) 1489 (("$" 3200) (T_STRING 3220) (T_VARIABLE 3207) (property_name 3226) (simple_variable 3222) ("{" 3223)) 1490 (("$" 3200) (T_STRING 3220) (T_VARIABLE 3207) (property_name 3221) (simple_variable 3222) ("{" 3223)) 1491 (("$" 3200) (T_VARIABLE 3207) (simple_variable 3219)) 1492 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 3217) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 1493 (("]" 3218)) 1494 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 3224) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 1495 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 3225)) 1496 (("$" 3200) (T_VARIABLE 3207) (simple_variable 3228)) 1497 (("%" 304) (")" 3230) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 1498 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 3233) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 1499 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 3234)) 1500 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3244) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1501 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 3242) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 1502 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3243) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1503 (("$" 3075) (T_ABSTRACT 3254) (T_ARRAY 3255) (T_AS 3256) (T_BREAK 3257) (T_CALLABLE 3258) (T_CASE 3259) (T_CATCH 3260) (T_CLASS 3261) (T_CLASS_C 3262) (T_CLONE 3263) (T_CONST 3264) (T_CONTINUE 3265) (T_DECLARE 3266) (T_DEFAULT 3267) (T_DIR 3268) (T_DO 3269) (T_ECHO 3270) (T_ELSE 3271) (T_ELSEIF 3272) (T_EMPTY 3273) (T_ENDDECLARE 3274) (T_ENDFOR 3275) (T_ENDFOREACH 3276) (T_ENDIF 3277) (T_ENDSWITCH 3278) (T_ENDWHILE 3279) (T_ENUM 3280) (T_EVAL 3281) (T_EXIT 3282) (T_EXTENDS 3283) (T_FILE 3284) (T_FINAL 3285) (T_FINALLY 3286) (T_FN 3287) (T_FOR 3288) (T_FOREACH 3289) (T_FUNCTION 3290) (T_FUNC_C 3291) (T_GLOBAL 3292) (T_GOTO 3293) (T_IF 3294) (T_IMPLEMENTS 3295) (T_INCLUDE 3296) (T_INCLUDE_ONCE 3297) (T_INSTANCEOF 3298) (T_INSTEADOF 3299) (T_INTERFACE 3300) (T_ISSET 3301) (T_LINE 3302) (T_LIST 3303) (T_LOGICAL_AND 3304) (T_LOGICAL_OR 3305) (T_LOGICAL_XOR 3306) (T_MATCH 3307) (T_METHOD_C 3308) (T_NAMESPACE 3309) (T_NEW 3310) (T_NS_C 3311) (T_PRINT 3312) (T_PRIVATE 3313) (T_PROPERTY_C 3314) (T_PROTECTED 3315) (T_PUBLIC 3316) (T_READONLY 3317) (T_REQUIRE 3318) (T_REQUIRE_ONCE 3319) (T_RETURN 3320) (T_STATIC 3321) (T_STRING 3322) (T_SWITCH 3323) (T_THROW 3324) (T_TRAIT 3325) (T_TRAIT_C 3326) (T_TRY 3327) (T_UNSET 3328) (T_USE 3329) (T_VAR 3330) (T_VARIABLE 3123) (T_WHILE 3331) (T_YIELD 3332) (identifier 3333) (member_name 3334) (reserved_non_modifiers 3335) (semi_reserved 3336) (simple_variable 3337) ("{" 3338)) 1504 (("(" 3341) (argument_list 3342)) 1505 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 3339) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 1506 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 3340)) 1507 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 3343) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 3344) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 3345) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 1508 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 3348) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 1509 (("," 1498) (possible_comma 3346)) 1510 ((")" 3347)) 1511 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (ampersand 3378) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3379) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1512 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3377) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1513 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3376) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1514 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3375) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1515 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3374) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1516 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3373) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1517 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3372) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1518 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3371) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1519 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3370) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1520 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3369) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1521 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3368) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1522 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3367) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1523 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3366) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1524 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3365) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1525 (("\"" 1538) ("$" 3380) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 3381) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 3382) ("[" 1560) (array_object_dereferenceable 3383) (callable_expr 3384) (callable_variable 3385) (class_constant 1564) (class_name 3386) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 3387) (name 3388) (new_dereferenceable 1570) (simple_variable 3389) (static_member 3390) (variable 3391) (variable_class_name 3392)) 1526 (("$" 3380) (T_VARIABLE 3382) (simple_variable 3425) ("{" 3426)) 1527 (("(" 3396) (argument_list 3424)) 1528 ((T_NULLSAFE_OBJECT_OPERATOR 3410) (T_OBJECT_OPERATOR 3411) ("[" 3412)) 1529 (("(" 3396) (argument_list 3409)) 1530 ((T_PAAMAYIM_NEKUDOTAYIM 3405)) 1531 (("(" 3396) (argument_list 3404)) 1532 ((T_PAAMAYIM_NEKUDOTAYIM 3393)) 1533 (("$" 3380) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 3382) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 3394) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 3395) ("{" 1660)) 1534 (("(" 3396) (argument_list 3397)) 1535 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 3398) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 3399) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 3400) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 1536 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 3403) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 1537 (("," 1498) (possible_comma 3401)) 1538 ((")" 3402)) 1539 (("$" 3380) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 3382) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 3406) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 3407) ("{" 1676)) 1540 (("(" 3396) (argument_list 3408)) 1541 (("$" 3380) (T_STRING 3415) (T_VARIABLE 3382) (property_name 3422) (simple_variable 3417) ("{" 3418)) 1542 (("$" 3380) (T_STRING 3415) (T_VARIABLE 3382) (property_name 3416) (simple_variable 3417) ("{" 3418)) 1543 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 3413) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 1544 (("]" 3414)) 1545 (("(" 3396) (argument_list 3421)) 1546 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 3419) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 1547 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 3420)) 1548 (("(" 3396) (argument_list 3423)) 1549 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 3427) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 1550 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 3428)) 1551 ((backup_doc_comment 3431)) 1552 (("(" 3432)) 1553 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 3433)) 1554 ((")" 3434)) 1555 ((T_USE 166) (lexical_vars 3435)) 1556 ((":" 168) (return_type 3436)) 1557 ((backup_fn_flags 3437)) 1558 (("{" 3438)) 1559 ((inner_statement_list 3439)) 1560 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 3440) ("~" 129)) 1561 ((backup_fn_flags 3441)) 1562 ((backup_doc_comment 3443)) 1563 (("(" 3444)) 1564 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 3445)) 1565 ((")" 3446)) 1566 ((":" 2271) (return_type 3447)) 1567 ((T_DOUBLE_ARROW 3448)) 1568 ((backup_fn_flags 3449)) 1569 ((backup_lex_pos 3450)) 1570 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3451) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1571 (("%" 3153) ("*" 3154) ("+" 3155) ("-" 3156) ("." 3157) ("/" 3158) ("<" 3159) (">" 3160) ("?" 3161) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 3162) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3163) (T_BOOLEAN_AND 3164) (T_BOOLEAN_OR 3165) (T_COALESCE 3166) (T_INSTANCEOF 3167) (T_IS_EQUAL 3168) (T_IS_GREATER_OR_EQUAL 3169) (T_IS_IDENTICAL 3170) (T_IS_NOT_EQUAL 3171) (T_IS_NOT_IDENTICAL 3172) (T_IS_SMALLER_OR_EQUAL 3173) (T_LOGICAL_AND 3174) (T_LOGICAL_OR 3175) (T_LOGICAL_XOR 3176) (T_PIPE 3177) (T_POW 3178) (T_SL 3179) (T_SPACESHIP 3180) (T_SR 3181) ("^" 3182) (backup_fn_flags 3452) ("|" 3183)) 1572 (("$" 3075) (T_ABSTRACT 3254) (T_ARRAY 3255) (T_AS 3256) (T_BREAK 3257) (T_CALLABLE 3258) (T_CASE 3259) (T_CATCH 3260) (T_CLASS 3261) (T_CLASS_C 3262) (T_CLONE 3263) (T_CONST 3264) (T_CONTINUE 3265) (T_DECLARE 3266) (T_DEFAULT 3267) (T_DIR 3268) (T_DO 3269) (T_ECHO 3270) (T_ELSE 3271) (T_ELSEIF 3272) (T_EMPTY 3273) (T_ENDDECLARE 3274) (T_ENDFOR 3275) (T_ENDFOREACH 3276) (T_ENDIF 3277) (T_ENDSWITCH 3278) (T_ENDWHILE 3279) (T_ENUM 3280) (T_EVAL 3281) (T_EXIT 3282) (T_EXTENDS 3283) (T_FILE 3284) (T_FINAL 3285) (T_FINALLY 3286) (T_FN 3287) (T_FOR 3288) (T_FOREACH 3289) (T_FUNCTION 3290) (T_FUNC_C 3291) (T_GLOBAL 3292) (T_GOTO 3293) (T_IF 3294) (T_IMPLEMENTS 3295) (T_INCLUDE 3296) (T_INCLUDE_ONCE 3297) (T_INSTANCEOF 3298) (T_INSTEADOF 3299) (T_INTERFACE 3300) (T_ISSET 3301) (T_LINE 3302) (T_LIST 3303) (T_LOGICAL_AND 3304) (T_LOGICAL_OR 3305) (T_LOGICAL_XOR 3306) (T_MATCH 3307) (T_METHOD_C 3308) (T_NAMESPACE 3309) (T_NEW 3310) (T_NS_C 3311) (T_PRINT 3312) (T_PRIVATE 3313) (T_PROPERTY_C 3314) (T_PROTECTED 3315) (T_PUBLIC 3316) (T_READONLY 3317) (T_REQUIRE 3318) (T_REQUIRE_ONCE 3319) (T_RETURN 3320) (T_STATIC 3321) (T_STRING 3322) (T_SWITCH 3323) (T_THROW 3324) (T_TRAIT 3325) (T_TRAIT_C 3326) (T_TRY 3327) (T_UNSET 3328) (T_USE 3329) (T_VAR 3330) (T_VARIABLE 3123) (T_WHILE 3331) (T_YIELD 3332) (identifier 3455) (member_name 3456) (reserved_non_modifiers 3335) (semi_reserved 3336) (simple_variable 3457) ("{" 3458)) 1573 (("(" 3341) (argument_list 3461)) 1574 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 3459) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 1575 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 3460)) 1576 ((T_FN 35) (T_FUNCTION 131) (fn 3137) #'3138 (inline_function 3465)) 1577 (("$" 3075) (T_STRING 3471) (T_VARIABLE 3123) (property_name 3478) (simple_variable 3473) ("{" 3474)) 1578 (("$" 3075) (T_STRING 3471) (T_VARIABLE 3123) (property_name 3472) (simple_variable 3473) ("{" 3474)) 1579 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 3469) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 1580 (("]" 3470)) 1581 (("(" 3341) (argument_list 3477)) 1582 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 3475) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 1583 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 3476)) 1584 (("(" 3341) (argument_list 3479)) 1585 (("`" 3481)) 1586 (("]" 3483)) 1587 (("=" 3484)) 1588 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3485) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1589 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 5734) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1590 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 5732) (encaps_var 5357)) 1591 (("$" 3489) (T_VARIABLE 3537) (simple_variable 5728) ("{" 5729)) 1592 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 5727) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 1593 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 5726) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1594 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 5725) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1595 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 5724) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1596 (("(" 5721)) 1597 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 5720) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1598 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 5719) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1599 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 5706) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (clone_argument_list 5707) (constant 3548) (dereferenceable_scalar 3549) (expr 5708) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1600 (("\"" 1538) ("$" 3794) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 3795) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 3796) ("[" 1560) (array_object_dereferenceable 3797) (callable_expr 3798) (callable_variable 3799) (class_constant 1564) (class_name 3800) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 3801) (name 3802) (new_dereferenceable 1570) (simple_variable 3803) (static_member 3804) (variable 5705) (variable_class_name 3806)) 1601 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 5704) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1602 (("(" 5701)) 1603 (("(" 5698)) 1604 (("(" 5689) (argument_list 5690) (ctor_arguments 5691)) 1605 (("\"" 1538) ("$" 3794) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 3795) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 3796) ("[" 1560) (array_object_dereferenceable 3797) (callable_expr 3798) (callable_variable 3799) (class_constant 1564) (class_name 3800) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 3801) (name 3802) (new_dereferenceable 1570) (simple_variable 3803) (static_member 3804) (variable 5688) (variable_class_name 3806)) 1606 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 5687) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1607 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 5686) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1608 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 5685) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1609 (("(" 5681)) 1610 (("(" 5676)) 1611 (("(" 3956)) 1612 (("$" 3794) ("(" 3919) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 3920) (T_NAME_QUALIFIED 3921) (T_NAME_RELATIVE 3922) (T_READONLY 2955) (T_STATIC 3923) (T_STRING 3924) (T_VARIABLE 3796) (anonymous_class 3925) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 3926) (attribute 2961) (attributes 3927) (class_modifier 2963) (class_name 3928) (class_name_reference 3929) (name 3930) (new_variable 3931) (simple_variable 3932)) 1613 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3918) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1614 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3917) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1615 (("(" 3755) (argument_list 3916)) 1616 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3915) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1617 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3914) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1618 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 3909) (T_END_HEREDOC 3910) (T_VARIABLE 2920) (encaps_list 3911) (encaps_var 2922)) 1619 ((T_FN 35) (T_FUNCTION 131) (fn 3551) #'3552 (inline_function 3908)) 1620 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3907) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1621 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3906) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1622 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3905) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1623 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3902) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1624 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3901) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1625 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 3897) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 1626 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 3895) (encaps_list 2828) (encaps_var 2829)) 1627 ((T_NULLSAFE_OBJECT_OPERATOR 3881) (T_OBJECT_OPERATOR 3882) ("[" 3883)) 1628 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 3878) (attribute 2757) (fn 3551) #'3552 (inline_function 3879)) 1629 (("(" 3755) (argument_list 3877)) 1630 ((T_PAAMAYIM_NEKUDOTAYIM 3869)) 1631 (("%" 3567) ("*" 3568) ("+" 3569) ("-" 3570) ("." 3571) ("/" 3572) ("<" 3573) (">" 3574) ("?" 3575) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 3576) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3577) (T_BOOLEAN_AND 3578) (T_BOOLEAN_OR 3579) (T_COALESCE 3580) (T_DOUBLE_ARROW 3867) (T_INSTANCEOF 3581) (T_IS_EQUAL 3582) (T_IS_GREATER_OR_EQUAL 3583) (T_IS_IDENTICAL 3584) (T_IS_NOT_EQUAL 3585) (T_IS_NOT_IDENTICAL 3586) (T_IS_SMALLER_OR_EQUAL 3587) (T_LOGICAL_AND 3588) (T_LOGICAL_OR 3589) (T_LOGICAL_XOR 3590) (T_PIPE 3591) (T_POW 3592) (T_SL 3593) (T_SPACESHIP 3594) (T_SR 3595) ("^" 3596) ("|" 3597)) 1632 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 3856)) 1633 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 3844)) 1634 (("(" 3755) (argument_list 3843)) 1635 (("=" 3763) (T_AND_EQUAL 3764) (T_COALESCE_EQUAL 3765) (T_CONCAT_EQUAL 3766) (T_DEC 3767) (T_DIV_EQUAL 3768) (T_INC 3769) (T_MINUS_EQUAL 3770) (T_MOD_EQUAL 3771) (T_MUL_EQUAL 3772) (T_OR_EQUAL 3773) (T_PLUS_EQUAL 3774) (T_POW_EQUAL 3775) (T_SL_EQUAL 3776) (T_SR_EQUAL 3777) (T_XOR_EQUAL 3778)) 1636 ((T_PAAMAYIM_NEKUDOTAYIM 3667)) 1637 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3566) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1638 (("%" 3567) ("*" 3568) ("+" 3569) ("-" 3570) ("." 3571) ("/" 3572) ("<" 3573) (">" 3574) ("?" 3575) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 3576) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3577) (T_BOOLEAN_AND 3578) (T_BOOLEAN_OR 3579) (T_COALESCE 3580) (T_INSTANCEOF 3581) (T_IS_EQUAL 3582) (T_IS_GREATER_OR_EQUAL 3583) (T_IS_IDENTICAL 3584) (T_IS_NOT_EQUAL 3585) (T_IS_NOT_IDENTICAL 3586) (T_IS_SMALLER_OR_EQUAL 3587) (T_LOGICAL_AND 3588) (T_LOGICAL_OR 3589) (T_LOGICAL_XOR 3590) (T_PIPE 3591) (T_POW 3592) (T_SL 3593) (T_SPACESHIP 3594) (T_SR 3595) ("^" 3596) ("|" 3597)) 1639 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3666) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1640 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3665) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1641 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3664) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1642 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3663) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1643 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3662) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1644 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3661) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1645 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3660) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1646 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3659) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1647 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 3654) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 3655) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 1648 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3653) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1649 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3652) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1650 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3651) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1651 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3650) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1652 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3649) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1653 (("$" 3614) ("(" 3615) (T_NAME_FULLY_QUALIFIED 3616) (T_NAME_QUALIFIED 3617) (T_NAME_RELATIVE 3618) (T_STATIC 3619) (T_STRING 3620) (T_VARIABLE 3621) (class_name 3622) (class_name_reference 3623) (name 3624) (new_variable 3625) (simple_variable 3626)) 1654 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3613) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1655 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3612) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1656 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3611) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1657 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3610) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1658 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3609) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1659 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3608) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1660 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3607) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1661 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3606) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1662 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3605) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1663 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3604) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1664 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3603) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1665 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3602) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1666 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3601) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1667 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3600) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1668 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3599) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1669 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3598) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1670 (("$" 3614) (T_VARIABLE 3621) (simple_variable 3645) ("{" 3646)) 1671 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 3643) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 1672 ((T_PAAMAYIM_NEKUDOTAYIM 3641)) 1673 ((T_NULLSAFE_OBJECT_OPERATOR 3627) (T_OBJECT_OPERATOR 3628) (T_PAAMAYIM_NEKUDOTAYIM 3629) ("[" 3630)) 1674 (("$" 3614) (T_STRING 3634) (T_VARIABLE 3621) (property_name 3640) (simple_variable 3636) ("{" 3637)) 1675 (("$" 3614) (T_STRING 3634) (T_VARIABLE 3621) (property_name 3635) (simple_variable 3636) ("{" 3637)) 1676 (("$" 3614) (T_VARIABLE 3621) (simple_variable 3633)) 1677 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 3631) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 1678 (("]" 3632)) 1679 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 3638) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 1680 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 3639)) 1681 (("$" 3614) (T_VARIABLE 3621) (simple_variable 3642)) 1682 (("%" 304) (")" 3644) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 1683 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 3647) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 1684 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 3648)) 1685 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3658) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1686 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 3656) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 1687 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3657) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1688 (("$" 3489) (T_ABSTRACT 3668) (T_ARRAY 3669) (T_AS 3670) (T_BREAK 3671) (T_CALLABLE 3672) (T_CASE 3673) (T_CATCH 3674) (T_CLASS 3675) (T_CLASS_C 3676) (T_CLONE 3677) (T_CONST 3678) (T_CONTINUE 3679) (T_DECLARE 3680) (T_DEFAULT 3681) (T_DIR 3682) (T_DO 3683) (T_ECHO 3684) (T_ELSE 3685) (T_ELSEIF 3686) (T_EMPTY 3687) (T_ENDDECLARE 3688) (T_ENDFOR 3689) (T_ENDFOREACH 3690) (T_ENDIF 3691) (T_ENDSWITCH 3692) (T_ENDWHILE 3693) (T_ENUM 3694) (T_EVAL 3695) (T_EXIT 3696) (T_EXTENDS 3697) (T_FILE 3698) (T_FINAL 3699) (T_FINALLY 3700) (T_FN 3701) (T_FOR 3702) (T_FOREACH 3703) (T_FUNCTION 3704) (T_FUNC_C 3705) (T_GLOBAL 3706) (T_GOTO 3707) (T_IF 3708) (T_IMPLEMENTS 3709) (T_INCLUDE 3710) (T_INCLUDE_ONCE 3711) (T_INSTANCEOF 3712) (T_INSTEADOF 3713) (T_INTERFACE 3714) (T_ISSET 3715) (T_LINE 3716) (T_LIST 3717) (T_LOGICAL_AND 3718) (T_LOGICAL_OR 3719) (T_LOGICAL_XOR 3720) (T_MATCH 3721) (T_METHOD_C 3722) (T_NAMESPACE 3723) (T_NEW 3724) (T_NS_C 3725) (T_PRINT 3726) (T_PRIVATE 3727) (T_PROPERTY_C 3728) (T_PROTECTED 3729) (T_PUBLIC 3730) (T_READONLY 3731) (T_REQUIRE 3732) (T_REQUIRE_ONCE 3733) (T_RETURN 3734) (T_STATIC 3735) (T_STRING 3736) (T_SWITCH 3737) (T_THROW 3738) (T_TRAIT 3739) (T_TRAIT_C 3740) (T_TRY 3741) (T_UNSET 3742) (T_USE 3743) (T_VAR 3744) (T_VARIABLE 3537) (T_WHILE 3745) (T_YIELD 3746) (identifier 3747) (member_name 3748) (reserved_non_modifiers 3749) (semi_reserved 3750) (simple_variable 3751) ("{" 3752)) 1689 (("(" 3755) (argument_list 3756)) 1690 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 3753) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 1691 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 3754)) 1692 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 3757) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 3758) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 3759) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 1693 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 3762) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 1694 (("," 1498) (possible_comma 3760)) 1695 ((")" 3761)) 1696 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (ampersand 3792) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3793) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1697 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3791) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1698 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3790) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1699 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3789) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1700 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3788) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1701 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3787) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1702 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3786) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1703 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3785) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1704 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3784) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1705 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3783) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1706 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3782) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1707 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3781) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1708 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3780) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1709 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3779) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1710 (("\"" 1538) ("$" 3794) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 3795) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 3796) ("[" 1560) (array_object_dereferenceable 3797) (callable_expr 3798) (callable_variable 3799) (class_constant 1564) (class_name 3800) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 3801) (name 3802) (new_dereferenceable 1570) (simple_variable 3803) (static_member 3804) (variable 3805) (variable_class_name 3806)) 1711 (("$" 3794) (T_VARIABLE 3796) (simple_variable 3839) ("{" 3840)) 1712 (("(" 3810) (argument_list 3838)) 1713 ((T_NULLSAFE_OBJECT_OPERATOR 3824) (T_OBJECT_OPERATOR 3825) ("[" 3826)) 1714 (("(" 3810) (argument_list 3823)) 1715 ((T_PAAMAYIM_NEKUDOTAYIM 3819)) 1716 (("(" 3810) (argument_list 3818)) 1717 ((T_PAAMAYIM_NEKUDOTAYIM 3807)) 1718 (("$" 3794) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 3796) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 3808) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 3809) ("{" 1660)) 1719 (("(" 3810) (argument_list 3811)) 1720 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 3812) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 3813) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 3814) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 1721 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 3817) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 1722 (("," 1498) (possible_comma 3815)) 1723 ((")" 3816)) 1724 (("$" 3794) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 3796) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 3820) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 3821) ("{" 1676)) 1725 (("(" 3810) (argument_list 3822)) 1726 (("$" 3794) (T_STRING 3829) (T_VARIABLE 3796) (property_name 3836) (simple_variable 3831) ("{" 3832)) 1727 (("$" 3794) (T_STRING 3829) (T_VARIABLE 3796) (property_name 3830) (simple_variable 3831) ("{" 3832)) 1728 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 3827) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 1729 (("]" 3828)) 1730 (("(" 3810) (argument_list 3835)) 1731 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 3833) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 1732 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 3834)) 1733 (("(" 3810) (argument_list 3837)) 1734 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 3841) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 1735 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 3842)) 1736 ((backup_doc_comment 3845)) 1737 (("(" 3846)) 1738 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 3847)) 1739 ((")" 3848)) 1740 ((T_USE 166) (lexical_vars 3849)) 1741 ((":" 168) (return_type 3850)) 1742 ((backup_fn_flags 3851)) 1743 (("{" 3852)) 1744 ((inner_statement_list 3853)) 1745 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 3854) ("~" 129)) 1746 ((backup_fn_flags 3855)) 1747 ((backup_doc_comment 3857)) 1748 (("(" 3858)) 1749 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 3859)) 1750 ((")" 3860)) 1751 ((":" 2271) (return_type 3861)) 1752 ((T_DOUBLE_ARROW 3862)) 1753 ((backup_fn_flags 3863)) 1754 ((backup_lex_pos 3864)) 1755 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3865) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1756 (("%" 3567) ("*" 3568) ("+" 3569) ("-" 3570) ("." 3571) ("/" 3572) ("<" 3573) (">" 3574) ("?" 3575) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 3576) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3577) (T_BOOLEAN_AND 3578) (T_BOOLEAN_OR 3579) (T_COALESCE 3580) (T_INSTANCEOF 3581) (T_IS_EQUAL 3582) (T_IS_GREATER_OR_EQUAL 3583) (T_IS_IDENTICAL 3584) (T_IS_NOT_EQUAL 3585) (T_IS_NOT_IDENTICAL 3586) (T_IS_SMALLER_OR_EQUAL 3587) (T_LOGICAL_AND 3588) (T_LOGICAL_OR 3589) (T_LOGICAL_XOR 3590) (T_PIPE 3591) (T_POW 3592) (T_SL 3593) (T_SPACESHIP 3594) (T_SR 3595) ("^" 3596) (backup_fn_flags 3866) ("|" 3597)) 1757 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 3868) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 1758 (("$" 3489) (T_ABSTRACT 3668) (T_ARRAY 3669) (T_AS 3670) (T_BREAK 3671) (T_CALLABLE 3672) (T_CASE 3673) (T_CATCH 3674) (T_CLASS 3675) (T_CLASS_C 3676) (T_CLONE 3677) (T_CONST 3678) (T_CONTINUE 3679) (T_DECLARE 3680) (T_DEFAULT 3681) (T_DIR 3682) (T_DO 3683) (T_ECHO 3684) (T_ELSE 3685) (T_ELSEIF 3686) (T_EMPTY 3687) (T_ENDDECLARE 3688) (T_ENDFOR 3689) (T_ENDFOREACH 3690) (T_ENDIF 3691) (T_ENDSWITCH 3692) (T_ENDWHILE 3693) (T_ENUM 3694) (T_EVAL 3695) (T_EXIT 3696) (T_EXTENDS 3697) (T_FILE 3698) (T_FINAL 3699) (T_FINALLY 3700) (T_FN 3701) (T_FOR 3702) (T_FOREACH 3703) (T_FUNCTION 3704) (T_FUNC_C 3705) (T_GLOBAL 3706) (T_GOTO 3707) (T_IF 3708) (T_IMPLEMENTS 3709) (T_INCLUDE 3710) (T_INCLUDE_ONCE 3711) (T_INSTANCEOF 3712) (T_INSTEADOF 3713) (T_INTERFACE 3714) (T_ISSET 3715) (T_LINE 3716) (T_LIST 3717) (T_LOGICAL_AND 3718) (T_LOGICAL_OR 3719) (T_LOGICAL_XOR 3720) (T_MATCH 3721) (T_METHOD_C 3722) (T_NAMESPACE 3723) (T_NEW 3724) (T_NS_C 3725) (T_PRINT 3726) (T_PRIVATE 3727) (T_PROPERTY_C 3728) (T_PROTECTED 3729) (T_PUBLIC 3730) (T_READONLY 3731) (T_REQUIRE 3732) (T_REQUIRE_ONCE 3733) (T_RETURN 3734) (T_STATIC 3735) (T_STRING 3736) (T_SWITCH 3737) (T_THROW 3738) (T_TRAIT 3739) (T_TRAIT_C 3740) (T_TRY 3741) (T_UNSET 3742) (T_USE 3743) (T_VAR 3744) (T_VARIABLE 3537) (T_WHILE 3745) (T_YIELD 3746) (identifier 3870) (member_name 3871) (reserved_non_modifiers 3749) (semi_reserved 3750) (simple_variable 3872) ("{" 3873)) 1759 (("(" 3755) (argument_list 3876)) 1760 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 3874) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 1761 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 3875)) 1762 ((T_FN 35) (T_FUNCTION 131) (fn 3551) #'3552 (inline_function 3880)) 1763 (("$" 3489) (T_STRING 3886) (T_VARIABLE 3537) (property_name 3893) (simple_variable 3888) ("{" 3889)) 1764 (("$" 3489) (T_STRING 3886) (T_VARIABLE 3537) (property_name 3887) (simple_variable 3888) ("{" 3889)) 1765 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 3884) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 1766 (("]" 3885)) 1767 (("(" 3755) (argument_list 3892)) 1768 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 3890) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 1769 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 3891)) 1770 (("(" 3755) (argument_list 3894)) 1771 (("`" 3896)) 1772 (("]" 3898)) 1773 (("=" 3899)) 1774 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3900) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1775 (("%" 3567) ("*" 3568) ("+" 3569) ("-" 3570) ("." 3571) ("/" 3572) ("<" 3573) (">" 3574) ("?" 3575) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 3576) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3577) (T_BOOLEAN_AND 3578) (T_BOOLEAN_OR 3579) (T_COALESCE 3580) (T_DOUBLE_ARROW 3903) (T_INSTANCEOF 3581) (T_IS_EQUAL 3582) (T_IS_GREATER_OR_EQUAL 3583) (T_IS_IDENTICAL 3584) (T_IS_NOT_EQUAL 3585) (T_IS_NOT_IDENTICAL 3586) (T_IS_SMALLER_OR_EQUAL 3587) (T_LOGICAL_AND 3588) (T_LOGICAL_OR 3589) (T_LOGICAL_XOR 3590) (T_PIPE 3591) (T_POW 3592) (T_SL 3593) (T_SPACESHIP 3594) (T_SR 3595) ("^" 3596) ("|" 3597)) 1776 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 3904) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 1777 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 3913) (T_VARIABLE 2920) (encaps_var 2934)) 1778 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 3912) (T_VARIABLE 2920) (encaps_var 2925)) 1779 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 3954) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 1780 ((T_CLASS 3946)) 1781 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 3945) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 3926) (attribute 2982) (class_modifier 2963)) 1782 ((T_PAAMAYIM_NEKUDOTAYIM 3943)) 1783 (("(" 3810) (argument_list 3942)) 1784 ((T_NULLSAFE_OBJECT_OPERATOR 3933) (T_OBJECT_OPERATOR 3934) (T_PAAMAYIM_NEKUDOTAYIM 3935) ("[" 3936)) 1785 (("$" 3794) (T_STRING 3829) (T_VARIABLE 3796) (property_name 3941) (simple_variable 3831) ("{" 3832)) 1786 (("$" 3794) (T_STRING 3829) (T_VARIABLE 3796) (property_name 3940) (simple_variable 3831) ("{" 3832)) 1787 (("$" 3794) (T_VARIABLE 3796) (simple_variable 3939)) 1788 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 3937) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 1789 (("]" 3938)) 1790 (("$" 3794) (T_VARIABLE 3796) (simple_variable 3944)) 1791 (("(" 2984) (argument_list 2985) (ctor_arguments 3947)) 1792 ((T_EXTENDS 2987) (extends_from 3948)) 1793 ((T_IMPLEMENTS 2989) (implements_list 3949)) 1794 ((backup_doc_comment 3950)) 1795 (("{" 3951)) 1796 ((class_statement_list 3952)) 1797 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 3953)) 1798 (("%" 304) (")" 3955) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 1799 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 3957) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 1800 (("%" 304) (")" 3958) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 1801 (("{" 3959)) 1802 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 4033) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1803 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 5675) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1804 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 5673) (encaps_var 5357)) 1805 (("$" 3962) (T_VARIABLE 4011) (simple_variable 5669) ("{" 5670)) 1806 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 5668) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 1807 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 5667) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1808 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 5666) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1809 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 5665) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1810 (("(" 5662)) 1811 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 5661) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1812 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 5660) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1813 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 5647) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (clone_argument_list 5648) (constant 4022) (dereferenceable_scalar 4023) (expr 5649) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1814 (("\"" 1538) ("$" 4272) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 4273) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 4274) ("[" 1560) (array_object_dereferenceable 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 1564) (class_name 4278) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 4279) (name 4280) (new_dereferenceable 1570) (simple_variable 4281) (static_member 4282) (variable 5646) (variable_class_name 4284)) 1815 (("," 5642) (possible_comma 5643)) 1816 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 5641) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1817 (("(" 5638)) 1818 (("(" 5635)) 1819 (("(" 5626) (argument_list 5627) (ctor_arguments 5628)) 1820 (("\"" 1538) ("$" 4272) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 4273) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 4274) ("[" 1560) (array_object_dereferenceable 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 1564) (class_name 4278) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 4279) (name 4280) (new_dereferenceable 1570) (simple_variable 4281) (static_member 4282) (variable 5625) (variable_class_name 4284)) 1821 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 5624) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1822 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 5623) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1823 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 5622) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1824 (("(" 5618)) 1825 (("(" 5613)) 1826 (("(" 5607)) 1827 (("$" 4272) ("(" 5570) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 5571) (T_NAME_QUALIFIED 5572) (T_NAME_RELATIVE 5573) (T_READONLY 2955) (T_STATIC 5574) (T_STRING 5575) (T_VARIABLE 4274) (anonymous_class 5576) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 5577) (attribute 2961) (attributes 5578) (class_modifier 2963) (class_name 5579) (class_name_reference 5580) (name 5581) (new_variable 5582) (simple_variable 5583)) 1828 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 5569) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1829 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 5568) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1830 (("(" 4233) (argument_list 5567)) 1831 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 5566) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1832 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 5565) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1833 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 5560) (T_END_HEREDOC 5561) (T_VARIABLE 2920) (encaps_list 5562) (encaps_var 2922)) 1834 ((T_FN 35) (T_FUNCTION 131) (fn 4025) #'4026 (inline_function 5559)) 1835 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 5558) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1836 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 5557) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1837 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 5556) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1838 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 5553) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1839 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 5552) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1840 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 5548) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 1841 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 5546) (encaps_list 2828) (encaps_var 2829)) 1842 ((T_NULLSAFE_OBJECT_OPERATOR 5532) (T_OBJECT_OPERATOR 5533) ("[" 5534)) 1843 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 5529) (attribute 2757) (fn 4025) #'4026 (inline_function 5530)) 1844 (("(" 4233) (argument_list 5528)) 1845 ((T_PAAMAYIM_NEKUDOTAYIM 5520)) 1846 (("%" 4045) ("*" 4046) ("+" 4047) ("-" 4048) ("." 4049) ("/" 4050) ("<" 4051) (">" 4052) ("?" 4053) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 4054) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 4055) (T_BOOLEAN_AND 4056) (T_BOOLEAN_OR 4057) (T_COALESCE 4058) (T_INSTANCEOF 4059) (T_IS_EQUAL 4060) (T_IS_GREATER_OR_EQUAL 4061) (T_IS_IDENTICAL 4062) (T_IS_NOT_EQUAL 4063) (T_IS_NOT_IDENTICAL 4064) (T_IS_SMALLER_OR_EQUAL 4065) (T_LOGICAL_AND 4066) (T_LOGICAL_OR 4067) (T_LOGICAL_XOR 4068) (T_PIPE 4069) (T_POW 4070) (T_SL 4071) (T_SPACESHIP 4072) (T_SR 4073) ("^" 4074) ("|" 4075)) 1847 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 5509)) 1848 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 5497)) 1849 (("," 4326) (possible_comma 4327)) 1850 (("}" 4325)) 1851 (("(" 4233) (argument_list 4324)) 1852 (("," 4321) (possible_comma 4322)) 1853 (("=" 4241) (T_AND_EQUAL 4242) (T_COALESCE_EQUAL 4243) (T_CONCAT_EQUAL 4244) (T_DEC 4245) (T_DIV_EQUAL 4246) (T_INC 4247) (T_MINUS_EQUAL 4248) (T_MOD_EQUAL 4249) (T_MUL_EQUAL 4250) (T_OR_EQUAL 4251) (T_PLUS_EQUAL 4252) (T_POW_EQUAL 4253) (T_SL_EQUAL 4254) (T_SR_EQUAL 4255) (T_XOR_EQUAL 4256)) 1854 ((T_PAAMAYIM_NEKUDOTAYIM 4145)) 1855 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4044) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1856 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4144) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1857 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4143) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1858 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4142) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1859 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4141) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1860 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4140) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1861 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4139) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1862 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4138) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1863 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4137) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1864 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 4132) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 4133) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 1865 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4131) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1866 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4130) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1867 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4129) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1868 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4128) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1869 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4127) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1870 (("$" 4092) ("(" 4093) (T_NAME_FULLY_QUALIFIED 4094) (T_NAME_QUALIFIED 4095) (T_NAME_RELATIVE 4096) (T_STATIC 4097) (T_STRING 4098) (T_VARIABLE 4099) (class_name 4100) (class_name_reference 4101) (name 4102) (new_variable 4103) (simple_variable 4104)) 1871 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4091) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1872 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4090) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1873 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4089) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1874 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4088) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1875 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4087) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1876 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4086) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1877 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4085) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1878 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4084) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1879 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4083) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1880 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4082) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1881 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4081) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1882 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4080) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1883 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4079) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1884 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4078) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1885 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4077) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1886 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4076) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1887 (("$" 4092) (T_VARIABLE 4099) (simple_variable 4123) ("{" 4124)) 1888 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 4121) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 1889 ((T_PAAMAYIM_NEKUDOTAYIM 4119)) 1890 ((T_NULLSAFE_OBJECT_OPERATOR 4105) (T_OBJECT_OPERATOR 4106) (T_PAAMAYIM_NEKUDOTAYIM 4107) ("[" 4108)) 1891 (("$" 4092) (T_STRING 4112) (T_VARIABLE 4099) (property_name 4118) (simple_variable 4114) ("{" 4115)) 1892 (("$" 4092) (T_STRING 4112) (T_VARIABLE 4099) (property_name 4113) (simple_variable 4114) ("{" 4115)) 1893 (("$" 4092) (T_VARIABLE 4099) (simple_variable 4111)) 1894 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 4109) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 1895 (("]" 4110)) 1896 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 4116) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 1897 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 4117)) 1898 (("$" 4092) (T_VARIABLE 4099) (simple_variable 4120)) 1899 (("%" 304) (")" 4122) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 1900 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 4125) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 1901 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 4126)) 1902 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4136) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1903 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 4134) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 1904 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4135) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1905 (("$" 3962) (T_ABSTRACT 4146) (T_ARRAY 4147) (T_AS 4148) (T_BREAK 4149) (T_CALLABLE 4150) (T_CASE 4151) (T_CATCH 4152) (T_CLASS 4153) (T_CLASS_C 4154) (T_CLONE 4155) (T_CONST 4156) (T_CONTINUE 4157) (T_DECLARE 4158) (T_DEFAULT 4159) (T_DIR 4160) (T_DO 4161) (T_ECHO 4162) (T_ELSE 4163) (T_ELSEIF 4164) (T_EMPTY 4165) (T_ENDDECLARE 4166) (T_ENDFOR 4167) (T_ENDFOREACH 4168) (T_ENDIF 4169) (T_ENDSWITCH 4170) (T_ENDWHILE 4171) (T_ENUM 4172) (T_EVAL 4173) (T_EXIT 4174) (T_EXTENDS 4175) (T_FILE 4176) (T_FINAL 4177) (T_FINALLY 4178) (T_FN 4179) (T_FOR 4180) (T_FOREACH 4181) (T_FUNCTION 4182) (T_FUNC_C 4183) (T_GLOBAL 4184) (T_GOTO 4185) (T_IF 4186) (T_IMPLEMENTS 4187) (T_INCLUDE 4188) (T_INCLUDE_ONCE 4189) (T_INSTANCEOF 4190) (T_INSTEADOF 4191) (T_INTERFACE 4192) (T_ISSET 4193) (T_LINE 4194) (T_LIST 4195) (T_LOGICAL_AND 4196) (T_LOGICAL_OR 4197) (T_LOGICAL_XOR 4198) (T_MATCH 4199) (T_METHOD_C 4200) (T_NAMESPACE 4201) (T_NEW 4202) (T_NS_C 4203) (T_PRINT 4204) (T_PRIVATE 4205) (T_PROPERTY_C 4206) (T_PROTECTED 4207) (T_PUBLIC 4208) (T_READONLY 4209) (T_REQUIRE 4210) (T_REQUIRE_ONCE 4211) (T_RETURN 4212) (T_STATIC 4213) (T_STRING 4214) (T_SWITCH 4215) (T_THROW 4216) (T_TRAIT 4217) (T_TRAIT_C 4218) (T_TRY 4219) (T_UNSET 4220) (T_USE 4221) (T_VAR 4222) (T_VARIABLE 4011) (T_WHILE 4223) (T_YIELD 4224) (identifier 4225) (member_name 4226) (reserved_non_modifiers 4227) (semi_reserved 4228) (simple_variable 4229) ("{" 4230)) 1906 (("(" 4233) (argument_list 4234)) 1907 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 4231) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 1908 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 4232)) 1909 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 4235) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 4236) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 4237) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 1910 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 4240) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 1911 (("," 1498) (possible_comma 4238)) 1912 ((")" 4239)) 1913 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (ampersand 4270) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4271) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1914 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4269) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1915 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4268) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1916 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4267) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1917 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4266) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1918 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4265) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1919 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4264) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1920 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4263) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1921 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4262) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1922 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4261) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1923 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4260) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1924 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4259) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1925 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4258) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1926 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4257) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1927 (("\"" 1538) ("$" 4272) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 4273) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 4274) ("[" 1560) (array_object_dereferenceable 4275) (callable_expr 4276) (callable_variable 4277) (class_constant 1564) (class_name 4278) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 4279) (name 4280) (new_dereferenceable 1570) (simple_variable 4281) (static_member 4282) (variable 4283) (variable_class_name 4284)) 1928 (("$" 4272) (T_VARIABLE 4274) (simple_variable 4317) ("{" 4318)) 1929 (("(" 4288) (argument_list 4316)) 1930 ((T_NULLSAFE_OBJECT_OPERATOR 4302) (T_OBJECT_OPERATOR 4303) ("[" 4304)) 1931 (("(" 4288) (argument_list 4301)) 1932 ((T_PAAMAYIM_NEKUDOTAYIM 4297)) 1933 (("(" 4288) (argument_list 4296)) 1934 ((T_PAAMAYIM_NEKUDOTAYIM 4285)) 1935 (("$" 4272) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 4274) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 4286) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 4287) ("{" 1660)) 1936 (("(" 4288) (argument_list 4289)) 1937 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 4290) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 4291) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 4292) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 1938 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 4295) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 1939 (("," 1498) (possible_comma 4293)) 1940 ((")" 4294)) 1941 (("$" 4272) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 4274) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 4298) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 4299) ("{" 1676)) 1942 (("(" 4288) (argument_list 4300)) 1943 (("$" 4272) (T_STRING 4307) (T_VARIABLE 4274) (property_name 4314) (simple_variable 4309) ("{" 4310)) 1944 (("$" 4272) (T_STRING 4307) (T_VARIABLE 4274) (property_name 4308) (simple_variable 4309) ("{" 4310)) 1945 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 4305) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 1946 (("]" 4306)) 1947 (("(" 4288) (argument_list 4313)) 1948 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 4311) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 1949 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 4312)) 1950 (("(" 4288) (argument_list 4315)) 1951 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 4319) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 1952 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 4320)) 1953 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4323) (match_arm_cond_list 4032) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1954 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 5496) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 1955 ((T_DOUBLE_ARROW 4328)) 1956 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4392) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 1957 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 5495) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 1958 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 5493) (encaps_var 5357)) 1959 (("$" 4331) (T_VARIABLE 4379) (simple_variable 5489) ("{" 5490)) 1960 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 5488) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 1961 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 5487) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 1962 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 5486) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 1963 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 5485) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 1964 (("(" 5482)) 1965 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 5481) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 1966 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 5480) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 1967 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 5467) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (clone_argument_list 5468) (constant 4390) (dereferenceable_scalar 4391) (expr 5469) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 1968 (("\"" 1538) ("$" 4636) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 4637) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 4638) ("[" 1560) (array_object_dereferenceable 4639) (callable_expr 4640) (callable_variable 4641) (class_constant 1564) (class_name 4642) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 4643) (name 4644) (new_dereferenceable 1570) (simple_variable 4645) (static_member 4646) (variable 5466) (variable_class_name 4648)) 1969 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 5465) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 1970 (("(" 5462)) 1971 (("(" 5459)) 1972 (("(" 5450) (argument_list 5451) (ctor_arguments 5452)) 1973 (("\"" 1538) ("$" 4636) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 4637) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 4638) ("[" 1560) (array_object_dereferenceable 4639) (callable_expr 4640) (callable_variable 4641) (class_constant 1564) (class_name 4642) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 4643) (name 4644) (new_dereferenceable 1570) (simple_variable 4645) (static_member 4646) (variable 5449) (variable_class_name 4648)) 1974 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 5448) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 1975 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 5447) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 1976 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 5446) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 1977 (("(" 5442)) 1978 (("(" 5437)) 1979 (("(" 5431)) 1980 (("$" 4636) ("(" 5394) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 5395) (T_NAME_QUALIFIED 5396) (T_NAME_RELATIVE 5397) (T_READONLY 2955) (T_STATIC 5398) (T_STRING 5399) (T_VARIABLE 4638) (anonymous_class 5400) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 5401) (attribute 2961) (attributes 5402) (class_modifier 2963) (class_name 5403) (class_name_reference 5404) (name 5405) (new_variable 5406) (simple_variable 5407)) 1981 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 5393) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 1982 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 5392) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 1983 (("(" 4597) (argument_list 5391)) 1984 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 5390) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 1985 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 5389) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 1986 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 5384) (T_END_HEREDOC 5385) (T_VARIABLE 2920) (encaps_list 5386) (encaps_var 2922)) 1987 ((T_FN 35) (T_FUNCTION 131) (fn 4393) #'4394 (inline_function 5383)) 1988 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 5382) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 1989 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 5381) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 1990 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 5380) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 1991 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 4805) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 1992 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4741) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 1993 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 4737) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 1994 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 4735) (encaps_list 2828) (encaps_var 2829)) 1995 ((T_NULLSAFE_OBJECT_OPERATOR 4721) (T_OBJECT_OPERATOR 4722) ("[" 4723)) 1996 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 4718) (attribute 2757) (fn 4393) #'4394 (inline_function 4719)) 1997 (("(" 4597) (argument_list 4717)) 1998 ((T_PAAMAYIM_NEKUDOTAYIM 4709)) 1999 (("%" 4409) ("*" 4410) ("+" 4411) ("-" 4412) ("." 4413) ("/" 4414) ("<" 4415) (">" 4416) ("?" 4417) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 4418) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 4419) (T_BOOLEAN_AND 4420) (T_BOOLEAN_OR 4421) (T_COALESCE 4422) (T_INSTANCEOF 4423) (T_IS_EQUAL 4424) (T_IS_GREATER_OR_EQUAL 4425) (T_IS_IDENTICAL 4426) (T_IS_NOT_EQUAL 4427) (T_IS_NOT_IDENTICAL 4428) (T_IS_SMALLER_OR_EQUAL 4429) (T_LOGICAL_AND 4430) (T_LOGICAL_OR 4431) (T_LOGICAL_XOR 4432) (T_PIPE 4433) (T_POW 4434) (T_SL 4435) (T_SPACESHIP 4436) (T_SR 4437) ("^" 4438) ("|" 4439)) 2000 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 4698)) 2001 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 4686)) 2002 (("(" 4597) (argument_list 4685)) 2003 (("=" 4605) (T_AND_EQUAL 4606) (T_COALESCE_EQUAL 4607) (T_CONCAT_EQUAL 4608) (T_DEC 4609) (T_DIV_EQUAL 4610) (T_INC 4611) (T_MINUS_EQUAL 4612) (T_MOD_EQUAL 4613) (T_MUL_EQUAL 4614) (T_OR_EQUAL 4615) (T_PLUS_EQUAL 4616) (T_POW_EQUAL 4617) (T_SL_EQUAL 4618) (T_SR_EQUAL 4619) (T_XOR_EQUAL 4620)) 2004 ((T_PAAMAYIM_NEKUDOTAYIM 4509)) 2005 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4408) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2006 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4508) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2007 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4507) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2008 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4506) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2009 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4505) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2010 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4504) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2011 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4503) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2012 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4502) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2013 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4501) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2014 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 4496) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 4497) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 2015 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4495) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2016 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4494) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2017 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4493) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2018 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4492) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2019 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4491) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2020 (("$" 4456) ("(" 4457) (T_NAME_FULLY_QUALIFIED 4458) (T_NAME_QUALIFIED 4459) (T_NAME_RELATIVE 4460) (T_STATIC 4461) (T_STRING 4462) (T_VARIABLE 4463) (class_name 4464) (class_name_reference 4465) (name 4466) (new_variable 4467) (simple_variable 4468)) 2021 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4455) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2022 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4454) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2023 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4453) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2024 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4452) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2025 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4451) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2026 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4450) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2027 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4449) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2028 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4448) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2029 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4447) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2030 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4446) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2031 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4445) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2032 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4444) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2033 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4443) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2034 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4442) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2035 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4441) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2036 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4440) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2037 (("$" 4456) (T_VARIABLE 4463) (simple_variable 4487) ("{" 4488)) 2038 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 4485) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 2039 ((T_PAAMAYIM_NEKUDOTAYIM 4483)) 2040 ((T_NULLSAFE_OBJECT_OPERATOR 4469) (T_OBJECT_OPERATOR 4470) (T_PAAMAYIM_NEKUDOTAYIM 4471) ("[" 4472)) 2041 (("$" 4456) (T_STRING 4476) (T_VARIABLE 4463) (property_name 4482) (simple_variable 4478) ("{" 4479)) 2042 (("$" 4456) (T_STRING 4476) (T_VARIABLE 4463) (property_name 4477) (simple_variable 4478) ("{" 4479)) 2043 (("$" 4456) (T_VARIABLE 4463) (simple_variable 4475)) 2044 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 4473) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 2045 (("]" 4474)) 2046 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 4480) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 2047 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 4481)) 2048 (("$" 4456) (T_VARIABLE 4463) (simple_variable 4484)) 2049 (("%" 304) (")" 4486) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 2050 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 4489) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 2051 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 4490)) 2052 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4500) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2053 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 4498) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 2054 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4499) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2055 (("$" 4331) (T_ABSTRACT 4510) (T_ARRAY 4511) (T_AS 4512) (T_BREAK 4513) (T_CALLABLE 4514) (T_CASE 4515) (T_CATCH 4516) (T_CLASS 4517) (T_CLASS_C 4518) (T_CLONE 4519) (T_CONST 4520) (T_CONTINUE 4521) (T_DECLARE 4522) (T_DEFAULT 4523) (T_DIR 4524) (T_DO 4525) (T_ECHO 4526) (T_ELSE 4527) (T_ELSEIF 4528) (T_EMPTY 4529) (T_ENDDECLARE 4530) (T_ENDFOR 4531) (T_ENDFOREACH 4532) (T_ENDIF 4533) (T_ENDSWITCH 4534) (T_ENDWHILE 4535) (T_ENUM 4536) (T_EVAL 4537) (T_EXIT 4538) (T_EXTENDS 4539) (T_FILE 4540) (T_FINAL 4541) (T_FINALLY 4542) (T_FN 4543) (T_FOR 4544) (T_FOREACH 4545) (T_FUNCTION 4546) (T_FUNC_C 4547) (T_GLOBAL 4548) (T_GOTO 4549) (T_IF 4550) (T_IMPLEMENTS 4551) (T_INCLUDE 4552) (T_INCLUDE_ONCE 4553) (T_INSTANCEOF 4554) (T_INSTEADOF 4555) (T_INTERFACE 4556) (T_ISSET 4557) (T_LINE 4558) (T_LIST 4559) (T_LOGICAL_AND 4560) (T_LOGICAL_OR 4561) (T_LOGICAL_XOR 4562) (T_MATCH 4563) (T_METHOD_C 4564) (T_NAMESPACE 4565) (T_NEW 4566) (T_NS_C 4567) (T_PRINT 4568) (T_PRIVATE 4569) (T_PROPERTY_C 4570) (T_PROTECTED 4571) (T_PUBLIC 4572) (T_READONLY 4573) (T_REQUIRE 4574) (T_REQUIRE_ONCE 4575) (T_RETURN 4576) (T_STATIC 4577) (T_STRING 4578) (T_SWITCH 4579) (T_THROW 4580) (T_TRAIT 4581) (T_TRAIT_C 4582) (T_TRY 4583) (T_UNSET 4584) (T_USE 4585) (T_VAR 4586) (T_VARIABLE 4379) (T_WHILE 4587) (T_YIELD 4588) (identifier 4589) (member_name 4590) (reserved_non_modifiers 4591) (semi_reserved 4592) (simple_variable 4593) ("{" 4594)) 2056 (("(" 4597) (argument_list 4598)) 2057 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 4595) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 2058 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 4596)) 2059 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 4599) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 4600) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 4601) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2060 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 4604) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2061 (("," 1498) (possible_comma 4602)) 2062 ((")" 4603)) 2063 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (ampersand 4634) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4635) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2064 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4633) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2065 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4632) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2066 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4631) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2067 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4630) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2068 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4629) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2069 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4628) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2070 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4627) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2071 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4626) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2072 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4625) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2073 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4624) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2074 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4623) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2075 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4622) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2076 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4621) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2077 (("\"" 1538) ("$" 4636) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 4637) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 4638) ("[" 1560) (array_object_dereferenceable 4639) (callable_expr 4640) (callable_variable 4641) (class_constant 1564) (class_name 4642) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 4643) (name 4644) (new_dereferenceable 1570) (simple_variable 4645) (static_member 4646) (variable 4647) (variable_class_name 4648)) 2078 (("$" 4636) (T_VARIABLE 4638) (simple_variable 4681) ("{" 4682)) 2079 (("(" 4652) (argument_list 4680)) 2080 ((T_NULLSAFE_OBJECT_OPERATOR 4666) (T_OBJECT_OPERATOR 4667) ("[" 4668)) 2081 (("(" 4652) (argument_list 4665)) 2082 ((T_PAAMAYIM_NEKUDOTAYIM 4661)) 2083 (("(" 4652) (argument_list 4660)) 2084 ((T_PAAMAYIM_NEKUDOTAYIM 4649)) 2085 (("$" 4636) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 4638) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 4650) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 4651) ("{" 1660)) 2086 (("(" 4652) (argument_list 4653)) 2087 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 4654) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 4655) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 4656) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2088 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 4659) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2089 (("," 1498) (possible_comma 4657)) 2090 ((")" 4658)) 2091 (("$" 4636) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 4638) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 4662) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 4663) ("{" 1676)) 2092 (("(" 4652) (argument_list 4664)) 2093 (("$" 4636) (T_STRING 4671) (T_VARIABLE 4638) (property_name 4678) (simple_variable 4673) ("{" 4674)) 2094 (("$" 4636) (T_STRING 4671) (T_VARIABLE 4638) (property_name 4672) (simple_variable 4673) ("{" 4674)) 2095 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 4669) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 2096 (("]" 4670)) 2097 (("(" 4652) (argument_list 4677)) 2098 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 4675) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 2099 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 4676)) 2100 (("(" 4652) (argument_list 4679)) 2101 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 4683) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 2102 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 4684)) 2103 ((backup_doc_comment 4687)) 2104 (("(" 4688)) 2105 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 4689)) 2106 ((")" 4690)) 2107 ((T_USE 166) (lexical_vars 4691)) 2108 ((":" 168) (return_type 4692)) 2109 ((backup_fn_flags 4693)) 2110 (("{" 4694)) 2111 ((inner_statement_list 4695)) 2112 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 4696) ("~" 129)) 2113 ((backup_fn_flags 4697)) 2114 ((backup_doc_comment 4699)) 2115 (("(" 4700)) 2116 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 4701)) 2117 ((")" 4702)) 2118 ((":" 2271) (return_type 4703)) 2119 ((T_DOUBLE_ARROW 4704)) 2120 ((backup_fn_flags 4705)) 2121 ((backup_lex_pos 4706)) 2122 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4707) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2123 (("%" 4409) ("*" 4410) ("+" 4411) ("-" 4412) ("." 4413) ("/" 4414) ("<" 4415) (">" 4416) ("?" 4417) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 4418) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 4419) (T_BOOLEAN_AND 4420) (T_BOOLEAN_OR 4421) (T_COALESCE 4422) (T_INSTANCEOF 4423) (T_IS_EQUAL 4424) (T_IS_GREATER_OR_EQUAL 4425) (T_IS_IDENTICAL 4426) (T_IS_NOT_EQUAL 4427) (T_IS_NOT_IDENTICAL 4428) (T_IS_SMALLER_OR_EQUAL 4429) (T_LOGICAL_AND 4430) (T_LOGICAL_OR 4431) (T_LOGICAL_XOR 4432) (T_PIPE 4433) (T_POW 4434) (T_SL 4435) (T_SPACESHIP 4436) (T_SR 4437) ("^" 4438) (backup_fn_flags 4708) ("|" 4439)) 2124 (("$" 4331) (T_ABSTRACT 4510) (T_ARRAY 4511) (T_AS 4512) (T_BREAK 4513) (T_CALLABLE 4514) (T_CASE 4515) (T_CATCH 4516) (T_CLASS 4517) (T_CLASS_C 4518) (T_CLONE 4519) (T_CONST 4520) (T_CONTINUE 4521) (T_DECLARE 4522) (T_DEFAULT 4523) (T_DIR 4524) (T_DO 4525) (T_ECHO 4526) (T_ELSE 4527) (T_ELSEIF 4528) (T_EMPTY 4529) (T_ENDDECLARE 4530) (T_ENDFOR 4531) (T_ENDFOREACH 4532) (T_ENDIF 4533) (T_ENDSWITCH 4534) (T_ENDWHILE 4535) (T_ENUM 4536) (T_EVAL 4537) (T_EXIT 4538) (T_EXTENDS 4539) (T_FILE 4540) (T_FINAL 4541) (T_FINALLY 4542) (T_FN 4543) (T_FOR 4544) (T_FOREACH 4545) (T_FUNCTION 4546) (T_FUNC_C 4547) (T_GLOBAL 4548) (T_GOTO 4549) (T_IF 4550) (T_IMPLEMENTS 4551) (T_INCLUDE 4552) (T_INCLUDE_ONCE 4553) (T_INSTANCEOF 4554) (T_INSTEADOF 4555) (T_INTERFACE 4556) (T_ISSET 4557) (T_LINE 4558) (T_LIST 4559) (T_LOGICAL_AND 4560) (T_LOGICAL_OR 4561) (T_LOGICAL_XOR 4562) (T_MATCH 4563) (T_METHOD_C 4564) (T_NAMESPACE 4565) (T_NEW 4566) (T_NS_C 4567) (T_PRINT 4568) (T_PRIVATE 4569) (T_PROPERTY_C 4570) (T_PROTECTED 4571) (T_PUBLIC 4572) (T_READONLY 4573) (T_REQUIRE 4574) (T_REQUIRE_ONCE 4575) (T_RETURN 4576) (T_STATIC 4577) (T_STRING 4578) (T_SWITCH 4579) (T_THROW 4580) (T_TRAIT 4581) (T_TRAIT_C 4582) (T_TRY 4583) (T_UNSET 4584) (T_USE 4585) (T_VAR 4586) (T_VARIABLE 4379) (T_WHILE 4587) (T_YIELD 4588) (identifier 4710) (member_name 4711) (reserved_non_modifiers 4591) (semi_reserved 4592) (simple_variable 4712) ("{" 4713)) 2125 (("(" 4597) (argument_list 4716)) 2126 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 4714) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 2127 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 4715)) 2128 ((T_FN 35) (T_FUNCTION 131) (fn 4393) #'4394 (inline_function 4720)) 2129 (("$" 4331) (T_STRING 4726) (T_VARIABLE 4379) (property_name 4733) (simple_variable 4728) ("{" 4729)) 2130 (("$" 4331) (T_STRING 4726) (T_VARIABLE 4379) (property_name 4727) (simple_variable 4728) ("{" 4729)) 2131 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 4724) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 2132 (("]" 4725)) 2133 (("(" 4597) (argument_list 4732)) 2134 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 4730) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 2135 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 4731)) 2136 (("(" 4597) (argument_list 4734)) 2137 (("`" 4736)) 2138 (("]" 4738)) 2139 (("=" 4739)) 2140 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 4740) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2141 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 5379) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2142 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 5356) (encaps_var 5357)) 2143 (("$" 4744) (T_VARIABLE 4792) (simple_variable 5348) ("{" 5349)) 2144 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 5347) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 2145 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 5346) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2146 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 5345) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2147 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 5344) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2148 (("(" 5341)) 2149 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 5340) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2150 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 5339) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2151 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 5251) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (clone_argument_list 5252) (constant 4803) (dereferenceable_scalar 4804) (expr 5253) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2152 (("\"" 1538) ("$" 5049) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 5050) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 5051) ("[" 1560) (array_object_dereferenceable 5052) (callable_expr 5053) (callable_variable 5054) (class_constant 1564) (class_name 5055) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 5056) (name 5057) (new_dereferenceable 1570) (simple_variable 5058) (static_member 5059) (variable 5250) (variable_class_name 5061)) 2153 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 5249) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2154 (("(" 5246)) 2155 (("(" 5243)) 2156 (("(" 5234) (argument_list 5235) (ctor_arguments 5236)) 2157 (("\"" 1538) ("$" 5049) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 5050) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 5051) ("[" 1560) (array_object_dereferenceable 5052) (callable_expr 5053) (callable_variable 5054) (class_constant 1564) (class_name 5055) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 5056) (name 5057) (new_dereferenceable 1570) (simple_variable 5058) (static_member 5059) (variable 5233) (variable_class_name 5061)) 2158 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 5232) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2159 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 5231) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2160 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 5230) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2161 (("(" 5222)) 2162 (("(" 5217)) 2163 (("(" 5211)) 2164 (("$" 5049) ("(" 5174) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 5175) (T_NAME_QUALIFIED 5176) (T_NAME_RELATIVE 5177) (T_READONLY 2955) (T_STATIC 5178) (T_STRING 5179) (T_VARIABLE 5051) (anonymous_class 5180) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 5181) (attribute 2961) (attributes 5182) (class_modifier 2963) (class_name 5183) (class_name_reference 5184) (name 5185) (new_variable 5186) (simple_variable 5187)) 2165 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 5173) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2166 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 5172) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2167 (("(" 5010) (argument_list 5171)) 2168 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 5170) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2169 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 5169) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2170 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 5164) (T_END_HEREDOC 5165) (T_VARIABLE 2920) (encaps_list 5166) (encaps_var 2922)) 2171 ((T_FN 35) (T_FUNCTION 131) (fn 4806) #'4807 (inline_function 5163)) 2172 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 5162) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2173 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 5161) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2174 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 5160) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2175 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 5157) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2176 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 5156) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2177 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 5152) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 2178 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 5150) (encaps_list 2828) (encaps_var 2829)) 2179 ((T_NULLSAFE_OBJECT_OPERATOR 5136) (T_OBJECT_OPERATOR 5137) ("[" 5138)) 2180 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 5133) (attribute 2757) (fn 4806) #'4807 (inline_function 5134)) 2181 (("(" 5010) (argument_list 5132)) 2182 ((T_PAAMAYIM_NEKUDOTAYIM 5124)) 2183 (("%" 4822) ("*" 4823) ("+" 4824) ("-" 4825) ("." 4826) ("/" 4827) ("<" 4828) (">" 4829) ("?" 4830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 4831) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 4832) (T_BOOLEAN_AND 4833) (T_BOOLEAN_OR 4834) (T_COALESCE 4835) (T_DOUBLE_ARROW 5122) (T_INSTANCEOF 4836) (T_IS_EQUAL 4837) (T_IS_GREATER_OR_EQUAL 4838) (T_IS_IDENTICAL 4839) (T_IS_NOT_EQUAL 4840) (T_IS_NOT_IDENTICAL 4841) (T_IS_SMALLER_OR_EQUAL 4842) (T_LOGICAL_AND 4843) (T_LOGICAL_OR 4844) (T_LOGICAL_XOR 4845) (T_PIPE 4846) (T_POW 4847) (T_SL 4848) (T_SPACESHIP 4849) (T_SR 4850) ("^" 4851) ("|" 4852)) 2184 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 5111)) 2185 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 5099)) 2186 (("(" 5010) (argument_list 5098)) 2187 (("=" 5018) (T_AND_EQUAL 5019) (T_COALESCE_EQUAL 5020) (T_CONCAT_EQUAL 5021) (T_DEC 5022) (T_DIV_EQUAL 5023) (T_INC 5024) (T_MINUS_EQUAL 5025) (T_MOD_EQUAL 5026) (T_MUL_EQUAL 5027) (T_OR_EQUAL 5028) (T_PLUS_EQUAL 5029) (T_POW_EQUAL 5030) (T_SL_EQUAL 5031) (T_SR_EQUAL 5032) (T_XOR_EQUAL 5033)) 2188 ((T_PAAMAYIM_NEKUDOTAYIM 4922)) 2189 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 4821) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2190 (("%" 4822) ("*" 4823) ("+" 4824) ("-" 4825) ("." 4826) ("/" 4827) ("<" 4828) (">" 4829) ("?" 4830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 4831) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 4832) (T_BOOLEAN_AND 4833) (T_BOOLEAN_OR 4834) (T_COALESCE 4835) (T_INSTANCEOF 4836) (T_IS_EQUAL 4837) (T_IS_GREATER_OR_EQUAL 4838) (T_IS_IDENTICAL 4839) (T_IS_NOT_EQUAL 4840) (T_IS_NOT_IDENTICAL 4841) (T_IS_SMALLER_OR_EQUAL 4842) (T_LOGICAL_AND 4843) (T_LOGICAL_OR 4844) (T_LOGICAL_XOR 4845) (T_PIPE 4846) (T_POW 4847) (T_SL 4848) (T_SPACESHIP 4849) (T_SR 4850) ("^" 4851) ("|" 4852)) 2191 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 4921) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2192 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 4920) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2193 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 4919) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2194 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 4918) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2195 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 4917) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2196 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 4916) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2197 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 4915) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2198 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 4914) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2199 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 4909) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 4910) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 2200 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 4908) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2201 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 4907) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2202 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 4906) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2203 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 4905) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2204 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 4904) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2205 (("$" 4869) ("(" 4870) (T_NAME_FULLY_QUALIFIED 4871) (T_NAME_QUALIFIED 4872) (T_NAME_RELATIVE 4873) (T_STATIC 4874) (T_STRING 4875) (T_VARIABLE 4876) (class_name 4877) (class_name_reference 4878) (name 4879) (new_variable 4880) (simple_variable 4881)) 2206 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 4868) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2207 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 4867) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2208 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 4866) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2209 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 4865) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2210 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 4864) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2211 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 4863) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2212 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 4862) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2213 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 4861) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2214 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 4860) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2215 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 4859) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2216 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 4858) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2217 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 4857) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2218 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 4856) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2219 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 4855) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2220 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 4854) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2221 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 4853) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2222 (("$" 4869) (T_VARIABLE 4876) (simple_variable 4900) ("{" 4901)) 2223 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 4898) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 2224 ((T_PAAMAYIM_NEKUDOTAYIM 4896)) 2225 ((T_NULLSAFE_OBJECT_OPERATOR 4882) (T_OBJECT_OPERATOR 4883) (T_PAAMAYIM_NEKUDOTAYIM 4884) ("[" 4885)) 2226 (("$" 4869) (T_STRING 4889) (T_VARIABLE 4876) (property_name 4895) (simple_variable 4891) ("{" 4892)) 2227 (("$" 4869) (T_STRING 4889) (T_VARIABLE 4876) (property_name 4890) (simple_variable 4891) ("{" 4892)) 2228 (("$" 4869) (T_VARIABLE 4876) (simple_variable 4888)) 2229 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 4886) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 2230 (("]" 4887)) 2231 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 4893) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 2232 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 4894)) 2233 (("$" 4869) (T_VARIABLE 4876) (simple_variable 4897)) 2234 (("%" 304) (")" 4899) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 2235 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 4902) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 2236 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 4903)) 2237 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 4913) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2238 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 4911) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 2239 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 4912) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2240 (("$" 4744) (T_ABSTRACT 4923) (T_ARRAY 4924) (T_AS 4925) (T_BREAK 4926) (T_CALLABLE 4927) (T_CASE 4928) (T_CATCH 4929) (T_CLASS 4930) (T_CLASS_C 4931) (T_CLONE 4932) (T_CONST 4933) (T_CONTINUE 4934) (T_DECLARE 4935) (T_DEFAULT 4936) (T_DIR 4937) (T_DO 4938) (T_ECHO 4939) (T_ELSE 4940) (T_ELSEIF 4941) (T_EMPTY 4942) (T_ENDDECLARE 4943) (T_ENDFOR 4944) (T_ENDFOREACH 4945) (T_ENDIF 4946) (T_ENDSWITCH 4947) (T_ENDWHILE 4948) (T_ENUM 4949) (T_EVAL 4950) (T_EXIT 4951) (T_EXTENDS 4952) (T_FILE 4953) (T_FINAL 4954) (T_FINALLY 4955) (T_FN 4956) (T_FOR 4957) (T_FOREACH 4958) (T_FUNCTION 4959) (T_FUNC_C 4960) (T_GLOBAL 4961) (T_GOTO 4962) (T_IF 4963) (T_IMPLEMENTS 4964) (T_INCLUDE 4965) (T_INCLUDE_ONCE 4966) (T_INSTANCEOF 4967) (T_INSTEADOF 4968) (T_INTERFACE 4969) (T_ISSET 4970) (T_LINE 4971) (T_LIST 4972) (T_LOGICAL_AND 4973) (T_LOGICAL_OR 4974) (T_LOGICAL_XOR 4975) (T_MATCH 4976) (T_METHOD_C 4977) (T_NAMESPACE 4978) (T_NEW 4979) (T_NS_C 4980) (T_PRINT 4981) (T_PRIVATE 4982) (T_PROPERTY_C 4983) (T_PROTECTED 4984) (T_PUBLIC 4985) (T_READONLY 4986) (T_REQUIRE 4987) (T_REQUIRE_ONCE 4988) (T_RETURN 4989) (T_STATIC 4990) (T_STRING 4991) (T_SWITCH 4992) (T_THROW 4993) (T_TRAIT 4994) (T_TRAIT_C 4995) (T_TRY 4996) (T_UNSET 4997) (T_USE 4998) (T_VAR 4999) (T_VARIABLE 4792) (T_WHILE 5000) (T_YIELD 5001) (identifier 5002) (member_name 5003) (reserved_non_modifiers 5004) (semi_reserved 5005) (simple_variable 5006) ("{" 5007)) 2241 (("(" 5010) (argument_list 5011)) 2242 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 5008) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 2243 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 5009)) 2244 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 5012) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 5013) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 5014) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2245 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 5017) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2246 (("," 1498) (possible_comma 5015)) 2247 ((")" 5016)) 2248 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (ampersand 5047) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 5048) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2249 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 5046) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2250 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 5045) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2251 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 5044) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2252 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 5043) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2253 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 5042) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2254 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 5041) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2255 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 5040) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2256 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 5039) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2257 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 5038) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2258 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 5037) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2259 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 5036) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2260 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 5035) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2261 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 5034) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2262 (("\"" 1538) ("$" 5049) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 5050) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 5051) ("[" 1560) (array_object_dereferenceable 5052) (callable_expr 5053) (callable_variable 5054) (class_constant 1564) (class_name 5055) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 5056) (name 5057) (new_dereferenceable 1570) (simple_variable 5058) (static_member 5059) (variable 5060) (variable_class_name 5061)) 2263 (("$" 5049) (T_VARIABLE 5051) (simple_variable 5094) ("{" 5095)) 2264 (("(" 5065) (argument_list 5093)) 2265 ((T_NULLSAFE_OBJECT_OPERATOR 5079) (T_OBJECT_OPERATOR 5080) ("[" 5081)) 2266 (("(" 5065) (argument_list 5078)) 2267 ((T_PAAMAYIM_NEKUDOTAYIM 5074)) 2268 (("(" 5065) (argument_list 5073)) 2269 ((T_PAAMAYIM_NEKUDOTAYIM 5062)) 2270 (("$" 5049) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 5051) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 5063) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 5064) ("{" 1660)) 2271 (("(" 5065) (argument_list 5066)) 2272 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 5067) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 5068) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 5069) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2273 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 5072) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2274 (("," 1498) (possible_comma 5070)) 2275 ((")" 5071)) 2276 (("$" 5049) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 5051) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 5075) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 5076) ("{" 1676)) 2277 (("(" 5065) (argument_list 5077)) 2278 (("$" 5049) (T_STRING 5084) (T_VARIABLE 5051) (property_name 5091) (simple_variable 5086) ("{" 5087)) 2279 (("$" 5049) (T_STRING 5084) (T_VARIABLE 5051) (property_name 5085) (simple_variable 5086) ("{" 5087)) 2280 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 5082) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 2281 (("]" 5083)) 2282 (("(" 5065) (argument_list 5090)) 2283 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 5088) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 2284 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 5089)) 2285 (("(" 5065) (argument_list 5092)) 2286 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 5096) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 2287 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 5097)) 2288 ((backup_doc_comment 5100)) 2289 (("(" 5101)) 2290 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 5102)) 2291 ((")" 5103)) 2292 ((T_USE 166) (lexical_vars 5104)) 2293 ((":" 168) (return_type 5105)) 2294 ((backup_fn_flags 5106)) 2295 (("{" 5107)) 2296 ((inner_statement_list 5108)) 2297 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 5109) ("~" 129)) 2298 ((backup_fn_flags 5110)) 2299 ((backup_doc_comment 5112)) 2300 (("(" 5113)) 2301 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 5114)) 2302 ((")" 5115)) 2303 ((":" 2271) (return_type 5116)) 2304 ((T_DOUBLE_ARROW 5117)) 2305 ((backup_fn_flags 5118)) 2306 ((backup_lex_pos 5119)) 2307 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 5120) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2308 (("%" 4822) ("*" 4823) ("+" 4824) ("-" 4825) ("." 4826) ("/" 4827) ("<" 4828) (">" 4829) ("?" 4830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 4831) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 4832) (T_BOOLEAN_AND 4833) (T_BOOLEAN_OR 4834) (T_COALESCE 4835) (T_INSTANCEOF 4836) (T_IS_EQUAL 4837) (T_IS_GREATER_OR_EQUAL 4838) (T_IS_IDENTICAL 4839) (T_IS_NOT_EQUAL 4840) (T_IS_NOT_IDENTICAL 4841) (T_IS_SMALLER_OR_EQUAL 4842) (T_LOGICAL_AND 4843) (T_LOGICAL_OR 4844) (T_LOGICAL_XOR 4845) (T_PIPE 4846) (T_POW 4847) (T_SL 4848) (T_SPACESHIP 4849) (T_SR 4850) ("^" 4851) (backup_fn_flags 5121) ("|" 4852)) 2309 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 5123) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2310 (("$" 4744) (T_ABSTRACT 4923) (T_ARRAY 4924) (T_AS 4925) (T_BREAK 4926) (T_CALLABLE 4927) (T_CASE 4928) (T_CATCH 4929) (T_CLASS 4930) (T_CLASS_C 4931) (T_CLONE 4932) (T_CONST 4933) (T_CONTINUE 4934) (T_DECLARE 4935) (T_DEFAULT 4936) (T_DIR 4937) (T_DO 4938) (T_ECHO 4939) (T_ELSE 4940) (T_ELSEIF 4941) (T_EMPTY 4942) (T_ENDDECLARE 4943) (T_ENDFOR 4944) (T_ENDFOREACH 4945) (T_ENDIF 4946) (T_ENDSWITCH 4947) (T_ENDWHILE 4948) (T_ENUM 4949) (T_EVAL 4950) (T_EXIT 4951) (T_EXTENDS 4952) (T_FILE 4953) (T_FINAL 4954) (T_FINALLY 4955) (T_FN 4956) (T_FOR 4957) (T_FOREACH 4958) (T_FUNCTION 4959) (T_FUNC_C 4960) (T_GLOBAL 4961) (T_GOTO 4962) (T_IF 4963) (T_IMPLEMENTS 4964) (T_INCLUDE 4965) (T_INCLUDE_ONCE 4966) (T_INSTANCEOF 4967) (T_INSTEADOF 4968) (T_INTERFACE 4969) (T_ISSET 4970) (T_LINE 4971) (T_LIST 4972) (T_LOGICAL_AND 4973) (T_LOGICAL_OR 4974) (T_LOGICAL_XOR 4975) (T_MATCH 4976) (T_METHOD_C 4977) (T_NAMESPACE 4978) (T_NEW 4979) (T_NS_C 4980) (T_PRINT 4981) (T_PRIVATE 4982) (T_PROPERTY_C 4983) (T_PROTECTED 4984) (T_PUBLIC 4985) (T_READONLY 4986) (T_REQUIRE 4987) (T_REQUIRE_ONCE 4988) (T_RETURN 4989) (T_STATIC 4990) (T_STRING 4991) (T_SWITCH 4992) (T_THROW 4993) (T_TRAIT 4994) (T_TRAIT_C 4995) (T_TRY 4996) (T_UNSET 4997) (T_USE 4998) (T_VAR 4999) (T_VARIABLE 4792) (T_WHILE 5000) (T_YIELD 5001) (identifier 5125) (member_name 5126) (reserved_non_modifiers 5004) (semi_reserved 5005) (simple_variable 5127) ("{" 5128)) 2311 (("(" 5010) (argument_list 5131)) 2312 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 5129) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 2313 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 5130)) 2314 ((T_FN 35) (T_FUNCTION 131) (fn 4806) #'4807 (inline_function 5135)) 2315 (("$" 4744) (T_STRING 5141) (T_VARIABLE 4792) (property_name 5148) (simple_variable 5143) ("{" 5144)) 2316 (("$" 4744) (T_STRING 5141) (T_VARIABLE 4792) (property_name 5142) (simple_variable 5143) ("{" 5144)) 2317 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 5139) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 2318 (("]" 5140)) 2319 (("(" 5010) (argument_list 5147)) 2320 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 5145) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 2321 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 5146)) 2322 (("(" 5010) (argument_list 5149)) 2323 (("`" 5151)) 2324 (("]" 5153)) 2325 (("=" 5154)) 2326 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 5155) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2327 (("%" 4822) ("*" 4823) ("+" 4824) ("-" 4825) ("." 4826) ("/" 4827) ("<" 4828) (">" 4829) ("?" 4830) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 4831) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 4832) (T_BOOLEAN_AND 4833) (T_BOOLEAN_OR 4834) (T_COALESCE 4835) (T_DOUBLE_ARROW 5158) (T_INSTANCEOF 4836) (T_IS_EQUAL 4837) (T_IS_GREATER_OR_EQUAL 4838) (T_IS_IDENTICAL 4839) (T_IS_NOT_EQUAL 4840) (T_IS_NOT_IDENTICAL 4841) (T_IS_SMALLER_OR_EQUAL 4842) (T_LOGICAL_AND 4843) (T_LOGICAL_OR 4844) (T_LOGICAL_XOR 4845) (T_PIPE 4846) (T_POW 4847) (T_SL 4848) (T_SPACESHIP 4849) (T_SR 4850) ("^" 4851) ("|" 4852)) 2328 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 5159) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2329 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 5168) (T_VARIABLE 2920) (encaps_var 2934)) 2330 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 5167) (T_VARIABLE 2920) (encaps_var 2925)) 2331 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 5209) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 2332 ((T_CLASS 5201)) 2333 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 5200) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 5181) (attribute 2982) (class_modifier 2963)) 2334 ((T_PAAMAYIM_NEKUDOTAYIM 5198)) 2335 (("(" 5065) (argument_list 5197)) 2336 ((T_NULLSAFE_OBJECT_OPERATOR 5188) (T_OBJECT_OPERATOR 5189) (T_PAAMAYIM_NEKUDOTAYIM 5190) ("[" 5191)) 2337 (("$" 5049) (T_STRING 5084) (T_VARIABLE 5051) (property_name 5196) (simple_variable 5086) ("{" 5087)) 2338 (("$" 5049) (T_STRING 5084) (T_VARIABLE 5051) (property_name 5195) (simple_variable 5086) ("{" 5087)) 2339 (("$" 5049) (T_VARIABLE 5051) (simple_variable 5194)) 2340 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 5192) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 2341 (("]" 5193)) 2342 (("$" 5049) (T_VARIABLE 5051) (simple_variable 5199)) 2343 (("(" 2984) (argument_list 2985) (ctor_arguments 5202)) 2344 ((T_EXTENDS 2987) (extends_from 5203)) 2345 ((T_IMPLEMENTS 2989) (implements_list 5204)) 2346 ((backup_doc_comment 5205)) 2347 (("{" 5206)) 2348 ((class_statement_list 5207)) 2349 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 5208)) 2350 (("%" 304) (")" 5210) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 2351 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 5212) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 2352 (("%" 304) (")" 5213) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 2353 (("{" 5214)) 2354 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 5215) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 2355 (("}" 5216)) 2356 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 5218) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 2357 ((")" 5219)) 2358 (("=" 5220)) 2359 (("!" 4742) ("\"" 4743) ("$" 4744) ("(" 4745) ("+" 4746) ("-" 4747) ("@" 4748) (T_ARRAY 4749) (T_ARRAY_CAST 4750) (T_ATTRIBUTE 130) (T_BOOL_CAST 4751) (T_CLASS_C 4752) (T_CLONE 4753) (T_CONSTANT_ENCAPSED_STRING 4754) (T_DEC 4755) (T_DIR 4756) (T_DNUMBER 4757) (T_DOUBLE_CAST 4758) (T_EMPTY 4759) (T_EVAL 4760) (T_EXIT 4761) (T_FILE 4762) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4763) (T_INC 4764) (T_INCLUDE 4765) (T_INCLUDE_ONCE 4766) (T_INT_CAST 4767) (T_ISSET 4768) (T_LINE 4769) (T_LIST 4770) (T_LNUMBER 4771) (T_MATCH 4772) (T_METHOD_C 4773) (T_NAME_FULLY_QUALIFIED 4774) (T_NAME_QUALIFIED 4775) (T_NAME_RELATIVE 4776) (T_NEW 4777) (T_NS_C 4778) (T_OBJECT_CAST 4779) (T_PRINT 4780) (T_PROPERTY_C 4781) (T_READONLY 4782) (T_REQUIRE 4783) (T_REQUIRE_ONCE 4784) (T_START_HEREDOC 4785) (T_STATIC 4786) (T_STRING 4787) (T_STRING_CAST 4788) (T_THROW 4789) (T_TRAIT_C 4790) (T_UNSET_CAST 4791) (T_VARIABLE 4792) (T_YIELD 4793) (T_YIELD_FROM 4794) ("[" 4795) ("`" 4796) (array_object_dereferenceable 4797) (attribute 135) (attributes 4798) (callable_expr 4799) (callable_variable 4800) (class_constant 4801) (class_name 4802) (constant 4803) (dereferenceable_scalar 4804) (expr 5221) (fn 4806) (fully_dereferenceable 107) #'4807 (function_call 4808) (inline_function 4809) (internal_functions_in_yacc 4810) (match 4811) (name 4812) (new_dereferenceable 4813) (new_non_dereferenceable 4814) (scalar 4815) (simple_variable 4816) (static_member 4817) (variable 4818) (variable_class_name 4819) ("~" 4820)) 2360 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 5225) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2361 (("," 5226) (possible_comma 5227)) 2362 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5229) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2363 ((")" 5228)) 2364 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 5237) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 5238) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 5239) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2365 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 5242) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2366 (("," 1498) (possible_comma 5240)) 2367 ((")" 5241)) 2368 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 5244) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 2369 (("%" 304) (")" 5245) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 2370 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 5247) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 2371 (("%" 304) (")" 5248) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 2372 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 5254) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 5259) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 5285) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 5286) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 2373 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 5322) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2374 (("%" 2442) (")" 5291) ("*" 2443) ("+" 2444) ("," 5292) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 2375 (("," 5287) (possible_comma 5288)) 2376 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5290) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2377 ((")" 5289)) 2378 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 5293) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2379 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 5296) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 2380 (("%" 304) (")" 5297) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 2381 (("{" 5298)) 2382 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 5299) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 2383 (("}" 5300)) 2384 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 5302) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2385 (("," 5226) (possible_comma 5303)) 2386 ((")" 5304)) 2387 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 5310) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 5311) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 5312) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2388 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 5315) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2389 (("," 1498) (possible_comma 5313)) 2390 ((")" 5314)) 2391 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 5317) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 2392 (("%" 304) (")" 5318) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 2393 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 5320) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 2394 (("%" 304) (")" 5321) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 2395 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 5326) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 5327) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 5328) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 5329) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 2396 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 5335) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2397 (("%" 2442) (")" 5332) ("*" 2443) ("+" 2444) ("," 5333) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 2398 (("," 5287) (possible_comma 5330)) 2399 ((")" 5331)) 2400 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 5334) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2401 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 5337) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 2402 ((")" 5338)) 2403 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 5342) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 2404 ((")" 5343)) 2405 (("%" 304) (")" 5291) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 2406 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 5350) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 2407 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 5351)) 2408 (("\"" 1538) ("$" 2854) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 2855) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 2856) ("[" 1560) (array_object_dereferenceable 2857) (callable_expr 2858) (callable_variable 2859) (class_constant 1564) (class_name 2860) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 2861) (name 2862) (new_dereferenceable 1570) (simple_variable 2863) (static_member 2864) (variable 5377) (variable_class_name 2866)) 2409 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_STRING_VARNAME 5369) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 5370) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 2410 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_VARIABLE 5355) (encaps_var 5368)) 2411 ((T_NULLSAFE_OBJECT_OPERATOR 5361) (T_OBJECT_OPERATOR 5362) ("[" 5363)) 2412 (("\"" 5358) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 2413 ((T_STRING 5367)) 2414 ((T_STRING 5366)) 2415 (("-" 2836) (T_NUM_STRING 2837) (T_STRING 2838) (T_VARIABLE 2839) (encaps_var_offset 5364)) 2416 (("]" 5365)) 2417 (("[" 5372) ("}" 5373)) 2418 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 5371)) 2419 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 5374) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 2420 (("%" 449) ("*" 450) ("+" 451) ("-" 452) ("." 453) ("/" 454) ("<" 455) (">" 456) ("?" 457) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 458) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 459) (T_BOOLEAN_AND 460) (T_BOOLEAN_OR 461) (T_COALESCE 462) (T_INSTANCEOF 463) (T_IS_EQUAL 464) (T_IS_GREATER_OR_EQUAL 465) (T_IS_IDENTICAL 466) (T_IS_NOT_EQUAL 467) (T_IS_NOT_IDENTICAL 468) (T_IS_SMALLER_OR_EQUAL 469) (T_LOGICAL_AND 470) (T_LOGICAL_OR 471) (T_LOGICAL_XOR 472) (T_PIPE 473) (T_POW 474) (T_SL 475) (T_SPACESHIP 476) (T_SR 477) ("]" 5375) ("^" 478) ("|" 479)) 2421 (("}" 5376)) 2422 (("}" 5378)) 2423 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 5388) (T_VARIABLE 2920) (encaps_var 2934)) 2424 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 5387) (T_VARIABLE 2920) (encaps_var 2925)) 2425 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 5429) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 2426 ((T_CLASS 5421)) 2427 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 5420) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 5401) (attribute 2982) (class_modifier 2963)) 2428 ((T_PAAMAYIM_NEKUDOTAYIM 5418)) 2429 (("(" 4652) (argument_list 5417)) 2430 ((T_NULLSAFE_OBJECT_OPERATOR 5408) (T_OBJECT_OPERATOR 5409) (T_PAAMAYIM_NEKUDOTAYIM 5410) ("[" 5411)) 2431 (("$" 4636) (T_STRING 4671) (T_VARIABLE 4638) (property_name 5416) (simple_variable 4673) ("{" 4674)) 2432 (("$" 4636) (T_STRING 4671) (T_VARIABLE 4638) (property_name 5415) (simple_variable 4673) ("{" 4674)) 2433 (("$" 4636) (T_VARIABLE 4638) (simple_variable 5414)) 2434 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 5412) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 2435 (("]" 5413)) 2436 (("$" 4636) (T_VARIABLE 4638) (simple_variable 5419)) 2437 (("(" 2984) (argument_list 2985) (ctor_arguments 5422)) 2438 ((T_EXTENDS 2987) (extends_from 5423)) 2439 ((T_IMPLEMENTS 2989) (implements_list 5424)) 2440 ((backup_doc_comment 5425)) 2441 (("{" 5426)) 2442 ((class_statement_list 5427)) 2443 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 5428)) 2444 (("%" 304) (")" 5430) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 2445 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 5432) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 2446 (("%" 304) (")" 5433) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 2447 (("{" 5434)) 2448 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 5435) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 2449 (("}" 5436)) 2450 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 5438) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 2451 ((")" 5439)) 2452 (("=" 5440)) 2453 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 5441) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2454 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 5443) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2455 (("," 5226) (possible_comma 5444)) 2456 ((")" 5445)) 2457 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 5453) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 5454) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 5455) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2458 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 5458) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2459 (("," 1498) (possible_comma 5456)) 2460 ((")" 5457)) 2461 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 5460) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 2462 (("%" 304) (")" 5461) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 2463 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 5463) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 2464 (("%" 304) (")" 5464) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 2465 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 5470) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 5471) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 5472) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 5473) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 2466 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 5479) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2467 (("%" 2442) (")" 5476) ("*" 2443) ("+" 2444) ("," 5477) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 2468 (("," 5287) (possible_comma 5474)) 2469 ((")" 5475)) 2470 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 5478) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2471 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 5483) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 2472 ((")" 5484)) 2473 (("%" 304) (")" 5476) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 2474 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 5491) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 2475 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 5492)) 2476 (("\"" 5494) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 2477 ((backup_doc_comment 5498)) 2478 (("(" 5499)) 2479 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 5500)) 2480 ((")" 5501)) 2481 ((T_USE 166) (lexical_vars 5502)) 2482 ((":" 168) (return_type 5503)) 2483 ((backup_fn_flags 5504)) 2484 (("{" 5505)) 2485 ((inner_statement_list 5506)) 2486 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 5507) ("~" 129)) 2487 ((backup_fn_flags 5508)) 2488 ((backup_doc_comment 5510)) 2489 (("(" 5511)) 2490 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 5512)) 2491 ((")" 5513)) 2492 ((":" 2271) (return_type 5514)) 2493 ((T_DOUBLE_ARROW 5515)) 2494 ((backup_fn_flags 5516)) 2495 ((backup_lex_pos 5517)) 2496 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 5518) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 2497 (("%" 4045) ("*" 4046) ("+" 4047) ("-" 4048) ("." 4049) ("/" 4050) ("<" 4051) (">" 4052) ("?" 4053) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 4054) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 4055) (T_BOOLEAN_AND 4056) (T_BOOLEAN_OR 4057) (T_COALESCE 4058) (T_INSTANCEOF 4059) (T_IS_EQUAL 4060) (T_IS_GREATER_OR_EQUAL 4061) (T_IS_IDENTICAL 4062) (T_IS_NOT_EQUAL 4063) (T_IS_NOT_IDENTICAL 4064) (T_IS_SMALLER_OR_EQUAL 4065) (T_LOGICAL_AND 4066) (T_LOGICAL_OR 4067) (T_LOGICAL_XOR 4068) (T_PIPE 4069) (T_POW 4070) (T_SL 4071) (T_SPACESHIP 4072) (T_SR 4073) ("^" 4074) (backup_fn_flags 5519) ("|" 4075)) 2498 (("$" 3962) (T_ABSTRACT 4146) (T_ARRAY 4147) (T_AS 4148) (T_BREAK 4149) (T_CALLABLE 4150) (T_CASE 4151) (T_CATCH 4152) (T_CLASS 4153) (T_CLASS_C 4154) (T_CLONE 4155) (T_CONST 4156) (T_CONTINUE 4157) (T_DECLARE 4158) (T_DEFAULT 4159) (T_DIR 4160) (T_DO 4161) (T_ECHO 4162) (T_ELSE 4163) (T_ELSEIF 4164) (T_EMPTY 4165) (T_ENDDECLARE 4166) (T_ENDFOR 4167) (T_ENDFOREACH 4168) (T_ENDIF 4169) (T_ENDSWITCH 4170) (T_ENDWHILE 4171) (T_ENUM 4172) (T_EVAL 4173) (T_EXIT 4174) (T_EXTENDS 4175) (T_FILE 4176) (T_FINAL 4177) (T_FINALLY 4178) (T_FN 4179) (T_FOR 4180) (T_FOREACH 4181) (T_FUNCTION 4182) (T_FUNC_C 4183) (T_GLOBAL 4184) (T_GOTO 4185) (T_IF 4186) (T_IMPLEMENTS 4187) (T_INCLUDE 4188) (T_INCLUDE_ONCE 4189) (T_INSTANCEOF 4190) (T_INSTEADOF 4191) (T_INTERFACE 4192) (T_ISSET 4193) (T_LINE 4194) (T_LIST 4195) (T_LOGICAL_AND 4196) (T_LOGICAL_OR 4197) (T_LOGICAL_XOR 4198) (T_MATCH 4199) (T_METHOD_C 4200) (T_NAMESPACE 4201) (T_NEW 4202) (T_NS_C 4203) (T_PRINT 4204) (T_PRIVATE 4205) (T_PROPERTY_C 4206) (T_PROTECTED 4207) (T_PUBLIC 4208) (T_READONLY 4209) (T_REQUIRE 4210) (T_REQUIRE_ONCE 4211) (T_RETURN 4212) (T_STATIC 4213) (T_STRING 4214) (T_SWITCH 4215) (T_THROW 4216) (T_TRAIT 4217) (T_TRAIT_C 4218) (T_TRY 4219) (T_UNSET 4220) (T_USE 4221) (T_VAR 4222) (T_VARIABLE 4011) (T_WHILE 4223) (T_YIELD 4224) (identifier 5521) (member_name 5522) (reserved_non_modifiers 4227) (semi_reserved 4228) (simple_variable 5523) ("{" 5524)) 2499 (("(" 4233) (argument_list 5527)) 2500 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 5525) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 2501 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 5526)) 2502 ((T_FN 35) (T_FUNCTION 131) (fn 4025) #'4026 (inline_function 5531)) 2503 (("$" 3962) (T_STRING 5537) (T_VARIABLE 4011) (property_name 5544) (simple_variable 5539) ("{" 5540)) 2504 (("$" 3962) (T_STRING 5537) (T_VARIABLE 4011) (property_name 5538) (simple_variable 5539) ("{" 5540)) 2505 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 5535) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 2506 (("]" 5536)) 2507 (("(" 4233) (argument_list 5543)) 2508 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 5541) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 2509 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 5542)) 2510 (("(" 4233) (argument_list 5545)) 2511 (("`" 5547)) 2512 (("]" 5549)) 2513 (("=" 5550)) 2514 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 5551) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 2515 (("%" 4045) ("*" 4046) ("+" 4047) ("-" 4048) ("." 4049) ("/" 4050) ("<" 4051) (">" 4052) ("?" 4053) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 4054) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 4055) (T_BOOLEAN_AND 4056) (T_BOOLEAN_OR 4057) (T_COALESCE 4058) (T_DOUBLE_ARROW 5554) (T_INSTANCEOF 4059) (T_IS_EQUAL 4060) (T_IS_GREATER_OR_EQUAL 4061) (T_IS_IDENTICAL 4062) (T_IS_NOT_EQUAL 4063) (T_IS_NOT_IDENTICAL 4064) (T_IS_SMALLER_OR_EQUAL 4065) (T_LOGICAL_AND 4066) (T_LOGICAL_OR 4067) (T_LOGICAL_XOR 4068) (T_PIPE 4069) (T_POW 4070) (T_SL 4071) (T_SPACESHIP 4072) (T_SR 4073) ("^" 4074) ("|" 4075)) 2516 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 5555) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 2517 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 5564) (T_VARIABLE 2920) (encaps_var 2934)) 2518 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 5563) (T_VARIABLE 2920) (encaps_var 2925)) 2519 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 5605) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 2520 ((T_CLASS 5597)) 2521 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 5596) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 5577) (attribute 2982) (class_modifier 2963)) 2522 ((T_PAAMAYIM_NEKUDOTAYIM 5594)) 2523 (("(" 4288) (argument_list 5593)) 2524 ((T_NULLSAFE_OBJECT_OPERATOR 5584) (T_OBJECT_OPERATOR 5585) (T_PAAMAYIM_NEKUDOTAYIM 5586) ("[" 5587)) 2525 (("$" 4272) (T_STRING 4307) (T_VARIABLE 4274) (property_name 5592) (simple_variable 4309) ("{" 4310)) 2526 (("$" 4272) (T_STRING 4307) (T_VARIABLE 4274) (property_name 5591) (simple_variable 4309) ("{" 4310)) 2527 (("$" 4272) (T_VARIABLE 4274) (simple_variable 5590)) 2528 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 5588) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 2529 (("]" 5589)) 2530 (("$" 4272) (T_VARIABLE 4274) (simple_variable 5595)) 2531 (("(" 2984) (argument_list 2985) (ctor_arguments 5598)) 2532 ((T_EXTENDS 2987) (extends_from 5599)) 2533 ((T_IMPLEMENTS 2989) (implements_list 5600)) 2534 ((backup_doc_comment 5601)) 2535 (("{" 5602)) 2536 ((class_statement_list 5603)) 2537 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 5604)) 2538 (("%" 304) (")" 5606) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 2539 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 5608) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 2540 (("%" 304) (")" 5609) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 2541 (("{" 5610)) 2542 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 5611) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 2543 (("}" 5612)) 2544 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 5614) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 2545 ((")" 5615)) 2546 (("=" 5616)) 2547 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 5617) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 2548 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 5619) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2549 (("," 5226) (possible_comma 5620)) 2550 ((")" 5621)) 2551 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 5629) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 5630) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 5631) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2552 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 5634) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2553 (("," 1498) (possible_comma 5632)) 2554 ((")" 5633)) 2555 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 5636) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 2556 (("%" 304) (")" 5637) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 2557 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 5639) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 2558 (("%" 304) (")" 5640) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 2559 ((T_DOUBLE_ARROW 5644)) 2560 (("!" 4329) ("\"" 4330) ("$" 4331) ("(" 4332) ("+" 4333) ("-" 4334) ("@" 4335) (T_ARRAY 4336) (T_ARRAY_CAST 4337) (T_ATTRIBUTE 130) (T_BOOL_CAST 4338) (T_CLASS_C 4339) (T_CLONE 4340) (T_CONSTANT_ENCAPSED_STRING 4341) (T_DEC 4342) (T_DIR 4343) (T_DNUMBER 4344) (T_DOUBLE_CAST 4345) (T_EMPTY 4346) (T_EVAL 4347) (T_EXIT 4348) (T_FILE 4349) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 4350) (T_INC 4351) (T_INCLUDE 4352) (T_INCLUDE_ONCE 4353) (T_INT_CAST 4354) (T_ISSET 4355) (T_LINE 4356) (T_LIST 4357) (T_LNUMBER 4358) (T_MATCH 4359) (T_METHOD_C 4360) (T_NAME_FULLY_QUALIFIED 4361) (T_NAME_QUALIFIED 4362) (T_NAME_RELATIVE 4363) (T_NEW 4364) (T_NS_C 4365) (T_OBJECT_CAST 4366) (T_PRINT 4367) (T_PROPERTY_C 4368) (T_READONLY 4369) (T_REQUIRE 4370) (T_REQUIRE_ONCE 4371) (T_START_HEREDOC 4372) (T_STATIC 4373) (T_STRING 4374) (T_STRING_CAST 4375) (T_THROW 4376) (T_TRAIT_C 4377) (T_UNSET_CAST 4378) (T_VARIABLE 4379) (T_YIELD 4380) (T_YIELD_FROM 4381) ("[" 4382) ("`" 4383) (array_object_dereferenceable 4384) (attribute 135) (attributes 4385) (callable_expr 4386) (callable_variable 4387) (class_constant 4388) (class_name 4389) (constant 4390) (dereferenceable_scalar 4391) (expr 5645) (fn 4393) (fully_dereferenceable 107) #'4394 (function_call 4395) (inline_function 4396) (internal_functions_in_yacc 4397) (match 4398) (name 4399) (new_dereferenceable 4400) (new_non_dereferenceable 4401) (scalar 4402) (simple_variable 4403) (static_member 4404) (variable 4405) (variable_class_name 4406) ("~" 4407)) 2561 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 5650) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 5651) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 5652) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 5653) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 2562 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 5659) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2563 (("%" 2442) (")" 5656) ("*" 2443) ("+" 2444) ("," 5657) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 2564 (("," 5287) (possible_comma 5654)) 2565 ((")" 5655)) 2566 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 5658) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2567 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 5663) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 2568 ((")" 5664)) 2569 (("%" 304) (")" 5656) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 2570 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 5671) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 2571 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 5672)) 2572 (("\"" 5674) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 2573 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 5677) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 2574 ((")" 5678)) 2575 (("=" 5679)) 2576 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 5680) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 2577 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 5682) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2578 (("," 5226) (possible_comma 5683)) 2579 ((")" 5684)) 2580 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 5692) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 5693) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 5694) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2581 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 5697) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2582 (("," 1498) (possible_comma 5695)) 2583 ((")" 5696)) 2584 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 5699) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 2585 (("%" 304) (")" 5700) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 2586 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 5702) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 2587 (("%" 304) (")" 5703) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 2588 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 5709) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 5710) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 5711) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 5712) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 2589 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 5718) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2590 (("%" 2442) (")" 5715) ("*" 2443) ("+" 2444) ("," 5716) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 2591 (("," 5287) (possible_comma 5713)) 2592 ((")" 5714)) 2593 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 5717) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2594 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 5722) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 2595 ((")" 5723)) 2596 (("%" 304) (")" 5715) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 2597 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 5730) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 2598 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 5731)) 2599 (("\"" 5733) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 2600 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 5743) (T_VARIABLE 2920) (encaps_var 2934)) 2601 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 5742) (T_VARIABLE 2920) (encaps_var 2925)) 2602 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 5784) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 2603 ((T_CLASS 5776)) 2604 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 5775) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 5756) (attribute 2982) (class_modifier 2963)) 2605 ((T_PAAMAYIM_NEKUDOTAYIM 5773)) 2606 (("(" 3396) (argument_list 5772)) 2607 ((T_NULLSAFE_OBJECT_OPERATOR 5763) (T_OBJECT_OPERATOR 5764) (T_PAAMAYIM_NEKUDOTAYIM 5765) ("[" 5766)) 2608 (("$" 3380) (T_STRING 3415) (T_VARIABLE 3382) (property_name 5771) (simple_variable 3417) ("{" 3418)) 2609 (("$" 3380) (T_STRING 3415) (T_VARIABLE 3382) (property_name 5770) (simple_variable 3417) ("{" 3418)) 2610 (("$" 3380) (T_VARIABLE 3382) (simple_variable 5769)) 2611 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 5767) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 2612 (("]" 5768)) 2613 (("$" 3380) (T_VARIABLE 3382) (simple_variable 5774)) 2614 (("(" 2984) (argument_list 2985) (ctor_arguments 5777)) 2615 ((T_EXTENDS 2987) (extends_from 5778)) 2616 ((T_IMPLEMENTS 2989) (implements_list 5779)) 2617 ((backup_doc_comment 5780)) 2618 (("{" 5781)) 2619 ((class_statement_list 5782)) 2620 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 5783)) 2621 (("%" 304) (")" 5785) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 2622 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 5787) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 2623 (("%" 304) (")" 5788) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 2624 (("{" 5789)) 2625 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 5790) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 2626 (("}" 5791)) 2627 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 5793) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 2628 ((")" 5794)) 2629 (("=" 5795)) 2630 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 5796) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 2631 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 5798) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2632 (("," 5226) (possible_comma 5799)) 2633 ((")" 5800)) 2634 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 5808) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 5809) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 5810) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2635 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 5813) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2636 (("," 1498) (possible_comma 5811)) 2637 ((")" 5812)) 2638 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 5815) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 2639 (("%" 304) (")" 5816) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 2640 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 5818) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 2641 (("%" 304) (")" 5819) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 2642 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 5825) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 5826) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 5827) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 5828) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 2643 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 5834) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2644 (("%" 2442) (")" 5831) ("*" 2443) ("+" 2444) ("," 5832) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 2645 (("," 5287) (possible_comma 5829)) 2646 ((")" 5830)) 2647 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 5833) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2648 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 5838) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 2649 ((")" 5839)) 2650 (("%" 304) (")" 5831) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 2651 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 5846) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 2652 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 5847)) 2653 (("\"" 5849) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 2654 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6080) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2655 (("{" 5853)) 2656 ((property_hook_list 5854)) 2657 ((T_ABSTRACT 5855) (T_ATTRIBUTE 5856) (T_FINAL 5857) (T_PRIVATE 5858) (T_PRIVATE_SET 5859) (T_PROTECTED 5860) (T_PROTECTED_SET 5861) (T_PUBLIC 5862) (T_PUBLIC_SET 5863) (T_READONLY 5864) (T_STATIC 5865) (attribute 5866) (attributes 5867) (member_modifier 5868) (non_empty_member_modifiers 5869) (property_hook 5870) (property_hook_modifiers 5871) ("}" 5872)) 2658 ((T_NAME_FULLY_QUALIFIED 5996) (T_NAME_QUALIFIED 5997) (T_NAME_RELATIVE 5998) (T_STATIC 5999) (T_STRING 6000) (attribute_decl 6001) (attribute_group 6002) (class_name 6003) (name 6004)) 2659 ((T_ABSTRACT 5855) (T_ATTRIBUTE 5856) (T_FINAL 5857) (T_PRIVATE 5858) (T_PRIVATE_SET 5859) (T_PROTECTED 5860) (T_PROTECTED_SET 5861) (T_PUBLIC 5862) (T_PUBLIC_SET 5863) (T_READONLY 5864) (T_STATIC 5865) (attribute 5994) (member_modifier 5868) (non_empty_member_modifiers 5869) (property_hook 5995) (property_hook_modifiers 5871)) 2660 ((T_ABSTRACT 5855) (T_FINAL 5857) (T_PRIVATE 5858) (T_PRIVATE_SET 5859) (T_PROTECTED 5860) (T_PROTECTED_SET 5861) (T_PUBLIC 5862) (T_PUBLIC_SET 5863) (T_READONLY 5864) (T_STATIC 5865) (member_modifier 5993)) 2661 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5873) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5874) (ampersand 5875) (returns_ref 5876)) 2662 ((T_STRING 5877)) 2663 ((backup_doc_comment 5878)) 2664 (("(" 5879) (optional_parameter_list 5880)) 2665 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 5991)) 2666 ((backup_fn_flags 5881)) 2667 ((";" 5882) (T_DOUBLE_ARROW 5883) (property_hook_body 5884) ("{" 5885)) 2668 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 5889) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 2669 ((backup_fn_flags 5888)) 2670 ((inner_statement_list 5886)) 2671 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 5887) ("~" 129)) 2672 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 5896) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 2673 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 5990) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 2674 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 5989) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 2675 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 5988) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 2676 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 5987) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 2677 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 5986) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 2678 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 5985) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 2679 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 5984) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 2680 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 5983) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 2681 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 5978) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 5979) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 2682 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 5977) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 2683 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 5976) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 2684 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 5975) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 2685 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 5974) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 2686 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 5973) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 2687 (("$" 5938) ("(" 5939) (T_NAME_FULLY_QUALIFIED 5940) (T_NAME_QUALIFIED 5941) (T_NAME_RELATIVE 5942) (T_STATIC 5943) (T_STRING 5944) (T_VARIABLE 5945) (class_name 5946) (class_name_reference 5947) (name 5948) (new_variable 5949) (simple_variable 5950)) 2688 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 5937) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 2689 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 5936) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 2690 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 5935) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 2691 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 5934) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 2692 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 5933) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 2693 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 5932) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 2694 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 5931) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 2695 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 5930) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 2696 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 5929) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 2697 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 5928) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 2698 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 5927) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 2699 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 5926) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 2700 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 5925) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 2701 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 5924) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 2702 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 5923) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 2703 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 5922) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 2704 (("$" 5938) (T_VARIABLE 5945) (simple_variable 5969) ("{" 5970)) 2705 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 5967) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 2706 ((T_PAAMAYIM_NEKUDOTAYIM 5965)) 2707 ((T_NULLSAFE_OBJECT_OPERATOR 5951) (T_OBJECT_OPERATOR 5952) (T_PAAMAYIM_NEKUDOTAYIM 5953) ("[" 5954)) 2708 (("$" 5938) (T_STRING 5958) (T_VARIABLE 5945) (property_name 5964) (simple_variable 5960) ("{" 5961)) 2709 (("$" 5938) (T_STRING 5958) (T_VARIABLE 5945) (property_name 5959) (simple_variable 5960) ("{" 5961)) 2710 (("$" 5938) (T_VARIABLE 5945) (simple_variable 5957)) 2711 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 5955) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 2712 (("]" 5956)) 2713 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 5962) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 2714 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 5963)) 2715 (("$" 5938) (T_VARIABLE 5945) (simple_variable 5966)) 2716 (("%" 304) (")" 5968) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 2717 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 5971) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 2718 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 5972)) 2719 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 5982) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 2720 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 5980) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 2721 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 5981) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 2722 ((")" 5992)) 2723 (("," 6013) (possible_comma 6014)) 2724 (("(" 6005) (argument_list 6006)) 2725 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 6007) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 6008) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 6009) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2726 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 6012) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2727 (("," 1498) (possible_comma 6010)) 2728 ((")" 6011)) 2729 ((T_NAME_FULLY_QUALIFIED 5996) (T_NAME_QUALIFIED 5997) (T_NAME_RELATIVE 5998) (T_STATIC 5999) (T_STRING 6000) (attribute_decl 6016) (class_name 6003) (name 6004)) 2730 (("]" 6015)) 2731 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 7083) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2732 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 7081) (encaps_var 5357)) 2733 (("$" 6019) (T_VARIABLE 6067) (simple_variable 7077) ("{" 7078)) 2734 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 7076) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 2735 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 7075) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2736 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 7074) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2737 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 7073) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2738 (("(" 7070)) 2739 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 7069) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2740 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 7068) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2741 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 7055) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (clone_argument_list 7056) (constant 6078) (dereferenceable_scalar 6079) (expr 7057) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2742 (("\"" 1538) ("$" 6324) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 6325) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 6326) ("[" 1560) (array_object_dereferenceable 6327) (callable_expr 6328) (callable_variable 6329) (class_constant 1564) (class_name 6330) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 6331) (name 6332) (new_dereferenceable 1570) (simple_variable 6333) (static_member 6334) (variable 7054) (variable_class_name 6336)) 2743 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 7053) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2744 (("(" 7050)) 2745 (("(" 7047)) 2746 (("(" 7038) (argument_list 7039) (ctor_arguments 7040)) 2747 (("\"" 1538) ("$" 6324) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 6325) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 6326) ("[" 1560) (array_object_dereferenceable 6327) (callable_expr 6328) (callable_variable 6329) (class_constant 1564) (class_name 6330) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 6331) (name 6332) (new_dereferenceable 1570) (simple_variable 6333) (static_member 6334) (variable 7037) (variable_class_name 6336)) 2748 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 7036) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2749 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 7035) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2750 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 7034) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2751 (("(" 7030)) 2752 (("(" 7025)) 2753 (("(" 7019)) 2754 (("$" 6324) ("(" 6982) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 6983) (T_NAME_QUALIFIED 6984) (T_NAME_RELATIVE 6985) (T_READONLY 2955) (T_STATIC 6986) (T_STRING 6987) (T_VARIABLE 6326) (anonymous_class 6988) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 6989) (attribute 2961) (attributes 6990) (class_modifier 2963) (class_name 6991) (class_name_reference 6992) (name 6993) (new_variable 6994) (simple_variable 6995)) 2755 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6981) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2756 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6980) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2757 (("(" 6285) (argument_list 6979)) 2758 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6978) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2759 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6977) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2760 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 6972) (T_END_HEREDOC 6973) (T_VARIABLE 2920) (encaps_list 6974) (encaps_var 2922)) 2761 ((T_FN 35) (T_FUNCTION 131) (fn 6081) #'6082 (inline_function 6971)) 2762 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6970) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2763 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6969) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2764 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6968) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2765 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6497) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2766 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6433) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2767 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 6429) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 2768 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 6427) (encaps_list 2828) (encaps_var 2829)) 2769 ((T_NULLSAFE_OBJECT_OPERATOR 6413) (T_OBJECT_OPERATOR 6414) ("[" 6415)) 2770 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 6410) (attribute 2757) (fn 6081) #'6082 (inline_function 6411)) 2771 (("(" 6285) (argument_list 6409)) 2772 ((T_PAAMAYIM_NEKUDOTAYIM 6401)) 2773 (("%" 6097) ("*" 6098) ("+" 6099) ("-" 6100) ("." 6101) ("/" 6102) ("<" 6103) (">" 6104) ("?" 6105) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6106) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6107) (T_BOOLEAN_AND 6108) (T_BOOLEAN_OR 6109) (T_COALESCE 6110) (T_INSTANCEOF 6111) (T_IS_EQUAL 6112) (T_IS_GREATER_OR_EQUAL 6113) (T_IS_IDENTICAL 6114) (T_IS_NOT_EQUAL 6115) (T_IS_NOT_IDENTICAL 6116) (T_IS_SMALLER_OR_EQUAL 6117) (T_LOGICAL_AND 6118) (T_LOGICAL_OR 6119) (T_LOGICAL_XOR 6120) (T_PIPE 6121) (T_POW 6122) (T_SL 6123) (T_SPACESHIP 6124) (T_SR 6125) ("^" 6126) (backup_doc_comment 6397) ("|" 6127)) 2774 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 6386)) 2775 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 6374)) 2776 (("(" 6285) (argument_list 6373)) 2777 (("=" 6293) (T_AND_EQUAL 6294) (T_COALESCE_EQUAL 6295) (T_CONCAT_EQUAL 6296) (T_DEC 6297) (T_DIV_EQUAL 6298) (T_INC 6299) (T_MINUS_EQUAL 6300) (T_MOD_EQUAL 6301) (T_MUL_EQUAL 6302) (T_OR_EQUAL 6303) (T_PLUS_EQUAL 6304) (T_POW_EQUAL 6305) (T_SL_EQUAL 6306) (T_SR_EQUAL 6307) (T_XOR_EQUAL 6308)) 2778 ((T_PAAMAYIM_NEKUDOTAYIM 6197)) 2779 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6096) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2780 (("%" 6097) ("*" 6098) ("+" 6099) ("-" 6100) ("." 6101) ("/" 6102) ("<" 6103) (">" 6104) ("?" 6105) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6106) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6107) (T_BOOLEAN_AND 6108) (T_BOOLEAN_OR 6109) (T_COALESCE 6110) (T_INSTANCEOF 6111) (T_IS_EQUAL 6112) (T_IS_GREATER_OR_EQUAL 6113) (T_IS_IDENTICAL 6114) (T_IS_NOT_EQUAL 6115) (T_IS_NOT_IDENTICAL 6116) (T_IS_SMALLER_OR_EQUAL 6117) (T_LOGICAL_AND 6118) (T_LOGICAL_OR 6119) (T_LOGICAL_XOR 6120) (T_PIPE 6121) (T_POW 6122) (T_SL 6123) (T_SPACESHIP 6124) (T_SR 6125) ("^" 6126) ("|" 6127)) 2781 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6196) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2782 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6195) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2783 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6194) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2784 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6193) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2785 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6192) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2786 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6191) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2787 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6190) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2788 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6189) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2789 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 6184) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 6185) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 2790 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6183) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2791 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6182) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2792 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6181) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2793 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6180) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2794 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6179) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2795 (("$" 6144) ("(" 6145) (T_NAME_FULLY_QUALIFIED 6146) (T_NAME_QUALIFIED 6147) (T_NAME_RELATIVE 6148) (T_STATIC 6149) (T_STRING 6150) (T_VARIABLE 6151) (class_name 6152) (class_name_reference 6153) (name 6154) (new_variable 6155) (simple_variable 6156)) 2796 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6143) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2797 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6142) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2798 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6141) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2799 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6140) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2800 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6139) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2801 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6138) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2802 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6137) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2803 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6136) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2804 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6135) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2805 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6134) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2806 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6133) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2807 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6132) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2808 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6131) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2809 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6130) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2810 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6129) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2811 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6128) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2812 (("$" 6144) (T_VARIABLE 6151) (simple_variable 6175) ("{" 6176)) 2813 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 6173) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 2814 ((T_PAAMAYIM_NEKUDOTAYIM 6171)) 2815 ((T_NULLSAFE_OBJECT_OPERATOR 6157) (T_OBJECT_OPERATOR 6158) (T_PAAMAYIM_NEKUDOTAYIM 6159) ("[" 6160)) 2816 (("$" 6144) (T_STRING 6164) (T_VARIABLE 6151) (property_name 6170) (simple_variable 6166) ("{" 6167)) 2817 (("$" 6144) (T_STRING 6164) (T_VARIABLE 6151) (property_name 6165) (simple_variable 6166) ("{" 6167)) 2818 (("$" 6144) (T_VARIABLE 6151) (simple_variable 6163)) 2819 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 6161) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 2820 (("]" 6162)) 2821 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 6168) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 2822 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 6169)) 2823 (("$" 6144) (T_VARIABLE 6151) (simple_variable 6172)) 2824 (("%" 304) (")" 6174) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 2825 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 6177) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 2826 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 6178)) 2827 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6188) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2828 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 6186) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 2829 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6187) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2830 (("$" 6019) (T_ABSTRACT 6198) (T_ARRAY 6199) (T_AS 6200) (T_BREAK 6201) (T_CALLABLE 6202) (T_CASE 6203) (T_CATCH 6204) (T_CLASS 6205) (T_CLASS_C 6206) (T_CLONE 6207) (T_CONST 6208) (T_CONTINUE 6209) (T_DECLARE 6210) (T_DEFAULT 6211) (T_DIR 6212) (T_DO 6213) (T_ECHO 6214) (T_ELSE 6215) (T_ELSEIF 6216) (T_EMPTY 6217) (T_ENDDECLARE 6218) (T_ENDFOR 6219) (T_ENDFOREACH 6220) (T_ENDIF 6221) (T_ENDSWITCH 6222) (T_ENDWHILE 6223) (T_ENUM 6224) (T_EVAL 6225) (T_EXIT 6226) (T_EXTENDS 6227) (T_FILE 6228) (T_FINAL 6229) (T_FINALLY 6230) (T_FN 6231) (T_FOR 6232) (T_FOREACH 6233) (T_FUNCTION 6234) (T_FUNC_C 6235) (T_GLOBAL 6236) (T_GOTO 6237) (T_IF 6238) (T_IMPLEMENTS 6239) (T_INCLUDE 6240) (T_INCLUDE_ONCE 6241) (T_INSTANCEOF 6242) (T_INSTEADOF 6243) (T_INTERFACE 6244) (T_ISSET 6245) (T_LINE 6246) (T_LIST 6247) (T_LOGICAL_AND 6248) (T_LOGICAL_OR 6249) (T_LOGICAL_XOR 6250) (T_MATCH 6251) (T_METHOD_C 6252) (T_NAMESPACE 6253) (T_NEW 6254) (T_NS_C 6255) (T_PRINT 6256) (T_PRIVATE 6257) (T_PROPERTY_C 6258) (T_PROTECTED 6259) (T_PUBLIC 6260) (T_READONLY 6261) (T_REQUIRE 6262) (T_REQUIRE_ONCE 6263) (T_RETURN 6264) (T_STATIC 6265) (T_STRING 6266) (T_SWITCH 6267) (T_THROW 6268) (T_TRAIT 6269) (T_TRAIT_C 6270) (T_TRY 6271) (T_UNSET 6272) (T_USE 6273) (T_VAR 6274) (T_VARIABLE 6067) (T_WHILE 6275) (T_YIELD 6276) (identifier 6277) (member_name 6278) (reserved_non_modifiers 6279) (semi_reserved 6280) (simple_variable 6281) ("{" 6282)) 2831 (("(" 6285) (argument_list 6286)) 2832 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 6283) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 2833 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 6284)) 2834 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 6287) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 6288) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 6289) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2835 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 6292) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2836 (("," 1498) (possible_comma 6290)) 2837 ((")" 6291)) 2838 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (ampersand 6322) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6323) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2839 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6321) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2840 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6320) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2841 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6319) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2842 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6318) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2843 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6317) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2844 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6316) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2845 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6315) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2846 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6314) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2847 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6313) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2848 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6312) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2849 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6311) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2850 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6310) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2851 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6309) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2852 (("\"" 1538) ("$" 6324) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 6325) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 6326) ("[" 1560) (array_object_dereferenceable 6327) (callable_expr 6328) (callable_variable 6329) (class_constant 1564) (class_name 6330) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 6331) (name 6332) (new_dereferenceable 1570) (simple_variable 6333) (static_member 6334) (variable 6335) (variable_class_name 6336)) 2853 (("$" 6324) (T_VARIABLE 6326) (simple_variable 6369) ("{" 6370)) 2854 (("(" 6340) (argument_list 6368)) 2855 ((T_NULLSAFE_OBJECT_OPERATOR 6354) (T_OBJECT_OPERATOR 6355) ("[" 6356)) 2856 (("(" 6340) (argument_list 6353)) 2857 ((T_PAAMAYIM_NEKUDOTAYIM 6349)) 2858 (("(" 6340) (argument_list 6348)) 2859 ((T_PAAMAYIM_NEKUDOTAYIM 6337)) 2860 (("$" 6324) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 6326) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 6338) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 6339) ("{" 1660)) 2861 (("(" 6340) (argument_list 6341)) 2862 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 6342) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 6343) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 6344) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2863 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 6347) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 2864 (("," 1498) (possible_comma 6345)) 2865 ((")" 6346)) 2866 (("$" 6324) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 6326) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 6350) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 6351) ("{" 1676)) 2867 (("(" 6340) (argument_list 6352)) 2868 (("$" 6324) (T_STRING 6359) (T_VARIABLE 6326) (property_name 6366) (simple_variable 6361) ("{" 6362)) 2869 (("$" 6324) (T_STRING 6359) (T_VARIABLE 6326) (property_name 6360) (simple_variable 6361) ("{" 6362)) 2870 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 6357) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 2871 (("]" 6358)) 2872 (("(" 6340) (argument_list 6365)) 2873 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 6363) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 2874 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 6364)) 2875 (("(" 6340) (argument_list 6367)) 2876 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 6371) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 2877 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 6372)) 2878 ((backup_doc_comment 6375)) 2879 (("(" 6376)) 2880 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 6377)) 2881 ((")" 6378)) 2882 ((T_USE 166) (lexical_vars 6379)) 2883 ((":" 168) (return_type 6380)) 2884 ((backup_fn_flags 6381)) 2885 (("{" 6382)) 2886 ((inner_statement_list 6383)) 2887 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 6384) ("~" 129)) 2888 ((backup_fn_flags 6385)) 2889 ((backup_doc_comment 6387)) 2890 (("(" 6388)) 2891 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 6389)) 2892 ((")" 6390)) 2893 ((":" 2271) (return_type 6391)) 2894 ((T_DOUBLE_ARROW 6392)) 2895 ((backup_fn_flags 6393)) 2896 ((backup_lex_pos 6394)) 2897 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6395) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2898 (("%" 6097) ("*" 6098) ("+" 6099) ("-" 6100) ("." 6101) ("/" 6102) ("<" 6103) (">" 6104) ("?" 6105) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6106) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6107) (T_BOOLEAN_AND 6108) (T_BOOLEAN_OR 6109) (T_COALESCE 6110) (T_INSTANCEOF 6111) (T_IS_EQUAL 6112) (T_IS_GREATER_OR_EQUAL 6113) (T_IS_IDENTICAL 6114) (T_IS_NOT_EQUAL 6115) (T_IS_NOT_IDENTICAL 6116) (T_IS_SMALLER_OR_EQUAL 6117) (T_LOGICAL_AND 6118) (T_LOGICAL_OR 6119) (T_LOGICAL_XOR 6120) (T_PIPE 6121) (T_POW 6122) (T_SL 6123) (T_SPACESHIP 6124) (T_SR 6125) ("^" 6126) (backup_fn_flags 6396) ("|" 6127)) 2899 (("{" 6398)) 2900 ((property_hook_list 6399)) 2901 ((T_ABSTRACT 5855) (T_ATTRIBUTE 5856) (T_FINAL 5857) (T_PRIVATE 5858) (T_PRIVATE_SET 5859) (T_PROTECTED 5860) (T_PROTECTED_SET 5861) (T_PUBLIC 5862) (T_PUBLIC_SET 5863) (T_READONLY 5864) (T_STATIC 5865) (attribute 5866) (attributes 5867) (member_modifier 5868) (non_empty_member_modifiers 5869) (property_hook 5870) (property_hook_modifiers 5871) ("}" 6400)) 2902 (("$" 6019) (T_ABSTRACT 6198) (T_ARRAY 6199) (T_AS 6200) (T_BREAK 6201) (T_CALLABLE 6202) (T_CASE 6203) (T_CATCH 6204) (T_CLASS 6205) (T_CLASS_C 6206) (T_CLONE 6207) (T_CONST 6208) (T_CONTINUE 6209) (T_DECLARE 6210) (T_DEFAULT 6211) (T_DIR 6212) (T_DO 6213) (T_ECHO 6214) (T_ELSE 6215) (T_ELSEIF 6216) (T_EMPTY 6217) (T_ENDDECLARE 6218) (T_ENDFOR 6219) (T_ENDFOREACH 6220) (T_ENDIF 6221) (T_ENDSWITCH 6222) (T_ENDWHILE 6223) (T_ENUM 6224) (T_EVAL 6225) (T_EXIT 6226) (T_EXTENDS 6227) (T_FILE 6228) (T_FINAL 6229) (T_FINALLY 6230) (T_FN 6231) (T_FOR 6232) (T_FOREACH 6233) (T_FUNCTION 6234) (T_FUNC_C 6235) (T_GLOBAL 6236) (T_GOTO 6237) (T_IF 6238) (T_IMPLEMENTS 6239) (T_INCLUDE 6240) (T_INCLUDE_ONCE 6241) (T_INSTANCEOF 6242) (T_INSTEADOF 6243) (T_INTERFACE 6244) (T_ISSET 6245) (T_LINE 6246) (T_LIST 6247) (T_LOGICAL_AND 6248) (T_LOGICAL_OR 6249) (T_LOGICAL_XOR 6250) (T_MATCH 6251) (T_METHOD_C 6252) (T_NAMESPACE 6253) (T_NEW 6254) (T_NS_C 6255) (T_PRINT 6256) (T_PRIVATE 6257) (T_PROPERTY_C 6258) (T_PROTECTED 6259) (T_PUBLIC 6260) (T_READONLY 6261) (T_REQUIRE 6262) (T_REQUIRE_ONCE 6263) (T_RETURN 6264) (T_STATIC 6265) (T_STRING 6266) (T_SWITCH 6267) (T_THROW 6268) (T_TRAIT 6269) (T_TRAIT_C 6270) (T_TRY 6271) (T_UNSET 6272) (T_USE 6273) (T_VAR 6274) (T_VARIABLE 6067) (T_WHILE 6275) (T_YIELD 6276) (identifier 6402) (member_name 6403) (reserved_non_modifiers 6279) (semi_reserved 6280) (simple_variable 6404) ("{" 6405)) 2903 (("(" 6285) (argument_list 6408)) 2904 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 6406) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 2905 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 6407)) 2906 ((T_FN 35) (T_FUNCTION 131) (fn 6081) #'6082 (inline_function 6412)) 2907 (("$" 6019) (T_STRING 6418) (T_VARIABLE 6067) (property_name 6425) (simple_variable 6420) ("{" 6421)) 2908 (("$" 6019) (T_STRING 6418) (T_VARIABLE 6067) (property_name 6419) (simple_variable 6420) ("{" 6421)) 2909 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 6416) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 2910 (("]" 6417)) 2911 (("(" 6285) (argument_list 6424)) 2912 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 6422) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 2913 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 6423)) 2914 (("(" 6285) (argument_list 6426)) 2915 (("`" 6428)) 2916 (("]" 6430)) 2917 (("=" 6431)) 2918 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6432) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 2919 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6967) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2920 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 6965) (encaps_var 5357)) 2921 (("$" 6436) (T_VARIABLE 6484) (simple_variable 6961) ("{" 6962)) 2922 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 6960) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 2923 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6959) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2924 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6958) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2925 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6957) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2926 (("(" 6954)) 2927 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6953) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2928 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6952) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2929 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6939) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (clone_argument_list 6940) (constant 6495) (dereferenceable_scalar 6496) (expr 6941) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2930 (("\"" 1538) ("$" 6741) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 6742) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 6743) ("[" 1560) (array_object_dereferenceable 6744) (callable_expr 6745) (callable_variable 6746) (class_constant 1564) (class_name 6747) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 6748) (name 6749) (new_dereferenceable 1570) (simple_variable 6750) (static_member 6751) (variable 6938) (variable_class_name 6753)) 2931 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6937) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2932 (("(" 6934)) 2933 (("(" 6931)) 2934 (("(" 6922) (argument_list 6923) (ctor_arguments 6924)) 2935 (("\"" 1538) ("$" 6741) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 6742) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 6743) ("[" 1560) (array_object_dereferenceable 6744) (callable_expr 6745) (callable_variable 6746) (class_constant 1564) (class_name 6747) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 6748) (name 6749) (new_dereferenceable 1570) (simple_variable 6750) (static_member 6751) (variable 6921) (variable_class_name 6753)) 2936 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6920) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2937 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6919) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2938 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6918) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2939 (("(" 6914)) 2940 (("(" 6909)) 2941 (("(" 6903)) 2942 (("$" 6741) ("(" 6866) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 6867) (T_NAME_QUALIFIED 6868) (T_NAME_RELATIVE 6869) (T_READONLY 2955) (T_STATIC 6870) (T_STRING 6871) (T_VARIABLE 6743) (anonymous_class 6872) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 6873) (attribute 2961) (attributes 6874) (class_modifier 2963) (class_name 6875) (class_name_reference 6876) (name 6877) (new_variable 6878) (simple_variable 6879)) 2943 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6865) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2944 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6864) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2945 (("(" 6702) (argument_list 6863)) 2946 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6862) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2947 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6861) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2948 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 6856) (T_END_HEREDOC 6857) (T_VARIABLE 2920) (encaps_list 6858) (encaps_var 2922)) 2949 ((T_FN 35) (T_FUNCTION 131) (fn 6498) #'6499 (inline_function 6855)) 2950 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6854) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2951 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6853) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2952 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6852) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2953 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6849) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2954 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6848) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2955 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 6844) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 2956 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 6842) (encaps_list 2828) (encaps_var 2829)) 2957 ((T_NULLSAFE_OBJECT_OPERATOR 6828) (T_OBJECT_OPERATOR 6829) ("[" 6830)) 2958 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 6825) (attribute 2757) (fn 6498) #'6499 (inline_function 6826)) 2959 (("(" 6702) (argument_list 6824)) 2960 ((T_PAAMAYIM_NEKUDOTAYIM 6816)) 2961 (("%" 6514) ("*" 6515) ("+" 6516) ("-" 6517) ("." 6518) ("/" 6519) ("<" 6520) (">" 6521) ("?" 6522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6523) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6524) (T_BOOLEAN_AND 6525) (T_BOOLEAN_OR 6526) (T_COALESCE 6527) (T_DOUBLE_ARROW 6814) (T_INSTANCEOF 6528) (T_IS_EQUAL 6529) (T_IS_GREATER_OR_EQUAL 6530) (T_IS_IDENTICAL 6531) (T_IS_NOT_EQUAL 6532) (T_IS_NOT_IDENTICAL 6533) (T_IS_SMALLER_OR_EQUAL 6534) (T_LOGICAL_AND 6535) (T_LOGICAL_OR 6536) (T_LOGICAL_XOR 6537) (T_PIPE 6538) (T_POW 6539) (T_SL 6540) (T_SPACESHIP 6541) (T_SR 6542) ("^" 6543) ("|" 6544)) 2962 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 6803)) 2963 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 6791)) 2964 (("(" 6702) (argument_list 6790)) 2965 (("=" 6710) (T_AND_EQUAL 6711) (T_COALESCE_EQUAL 6712) (T_CONCAT_EQUAL 6713) (T_DEC 6714) (T_DIV_EQUAL 6715) (T_INC 6716) (T_MINUS_EQUAL 6717) (T_MOD_EQUAL 6718) (T_MUL_EQUAL 6719) (T_OR_EQUAL 6720) (T_PLUS_EQUAL 6721) (T_POW_EQUAL 6722) (T_SL_EQUAL 6723) (T_SR_EQUAL 6724) (T_XOR_EQUAL 6725)) 2966 ((T_PAAMAYIM_NEKUDOTAYIM 6614)) 2967 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6513) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2968 (("%" 6514) ("*" 6515) ("+" 6516) ("-" 6517) ("." 6518) ("/" 6519) ("<" 6520) (">" 6521) ("?" 6522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6523) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6524) (T_BOOLEAN_AND 6525) (T_BOOLEAN_OR 6526) (T_COALESCE 6527) (T_INSTANCEOF 6528) (T_IS_EQUAL 6529) (T_IS_GREATER_OR_EQUAL 6530) (T_IS_IDENTICAL 6531) (T_IS_NOT_EQUAL 6532) (T_IS_NOT_IDENTICAL 6533) (T_IS_SMALLER_OR_EQUAL 6534) (T_LOGICAL_AND 6535) (T_LOGICAL_OR 6536) (T_LOGICAL_XOR 6537) (T_PIPE 6538) (T_POW 6539) (T_SL 6540) (T_SPACESHIP 6541) (T_SR 6542) ("^" 6543) ("|" 6544)) 2969 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6613) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2970 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6612) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2971 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6611) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2972 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6610) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2973 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6609) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2974 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6608) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2975 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6607) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2976 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6606) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2977 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 6601) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 6602) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 2978 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6600) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2979 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6599) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2980 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6598) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2981 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6597) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2982 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6596) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2983 (("$" 6561) ("(" 6562) (T_NAME_FULLY_QUALIFIED 6563) (T_NAME_QUALIFIED 6564) (T_NAME_RELATIVE 6565) (T_STATIC 6566) (T_STRING 6567) (T_VARIABLE 6568) (class_name 6569) (class_name_reference 6570) (name 6571) (new_variable 6572) (simple_variable 6573)) 2984 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6560) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2985 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6559) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2986 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6558) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2987 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6557) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2988 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6556) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2989 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6555) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2990 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6554) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2991 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6553) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2992 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6552) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2993 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6551) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2994 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6550) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2995 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6549) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2996 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6548) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2997 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6547) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2998 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6546) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 2999 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6545) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 3000 (("$" 6561) (T_VARIABLE 6568) (simple_variable 6592) ("{" 6593)) 3001 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 6590) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 3002 ((T_PAAMAYIM_NEKUDOTAYIM 6588)) 3003 ((T_NULLSAFE_OBJECT_OPERATOR 6574) (T_OBJECT_OPERATOR 6575) (T_PAAMAYIM_NEKUDOTAYIM 6576) ("[" 6577)) 3004 (("$" 6561) (T_STRING 6581) (T_VARIABLE 6568) (property_name 6587) (simple_variable 6583) ("{" 6584)) 3005 (("$" 6561) (T_STRING 6581) (T_VARIABLE 6568) (property_name 6582) (simple_variable 6583) ("{" 6584)) 3006 (("$" 6561) (T_VARIABLE 6568) (simple_variable 6580)) 3007 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 6578) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 3008 (("]" 6579)) 3009 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 6585) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 3010 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 6586)) 3011 (("$" 6561) (T_VARIABLE 6568) (simple_variable 6589)) 3012 (("%" 304) (")" 6591) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 3013 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 6594) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 3014 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 6595)) 3015 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6605) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 3016 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 6603) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 3017 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6604) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 3018 (("$" 6436) (T_ABSTRACT 6615) (T_ARRAY 6616) (T_AS 6617) (T_BREAK 6618) (T_CALLABLE 6619) (T_CASE 6620) (T_CATCH 6621) (T_CLASS 6622) (T_CLASS_C 6623) (T_CLONE 6624) (T_CONST 6625) (T_CONTINUE 6626) (T_DECLARE 6627) (T_DEFAULT 6628) (T_DIR 6629) (T_DO 6630) (T_ECHO 6631) (T_ELSE 6632) (T_ELSEIF 6633) (T_EMPTY 6634) (T_ENDDECLARE 6635) (T_ENDFOR 6636) (T_ENDFOREACH 6637) (T_ENDIF 6638) (T_ENDSWITCH 6639) (T_ENDWHILE 6640) (T_ENUM 6641) (T_EVAL 6642) (T_EXIT 6643) (T_EXTENDS 6644) (T_FILE 6645) (T_FINAL 6646) (T_FINALLY 6647) (T_FN 6648) (T_FOR 6649) (T_FOREACH 6650) (T_FUNCTION 6651) (T_FUNC_C 6652) (T_GLOBAL 6653) (T_GOTO 6654) (T_IF 6655) (T_IMPLEMENTS 6656) (T_INCLUDE 6657) (T_INCLUDE_ONCE 6658) (T_INSTANCEOF 6659) (T_INSTEADOF 6660) (T_INTERFACE 6661) (T_ISSET 6662) (T_LINE 6663) (T_LIST 6664) (T_LOGICAL_AND 6665) (T_LOGICAL_OR 6666) (T_LOGICAL_XOR 6667) (T_MATCH 6668) (T_METHOD_C 6669) (T_NAMESPACE 6670) (T_NEW 6671) (T_NS_C 6672) (T_PRINT 6673) (T_PRIVATE 6674) (T_PROPERTY_C 6675) (T_PROTECTED 6676) (T_PUBLIC 6677) (T_READONLY 6678) (T_REQUIRE 6679) (T_REQUIRE_ONCE 6680) (T_RETURN 6681) (T_STATIC 6682) (T_STRING 6683) (T_SWITCH 6684) (T_THROW 6685) (T_TRAIT 6686) (T_TRAIT_C 6687) (T_TRY 6688) (T_UNSET 6689) (T_USE 6690) (T_VAR 6691) (T_VARIABLE 6484) (T_WHILE 6692) (T_YIELD 6693) (identifier 6694) (member_name 6695) (reserved_non_modifiers 6696) (semi_reserved 6697) (simple_variable 6698) ("{" 6699)) 3019 (("(" 6702) (argument_list 6703)) 3020 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 6700) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 3021 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 6701)) 3022 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 6704) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 6705) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 6706) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3023 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 6709) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3024 (("," 1498) (possible_comma 6707)) 3025 ((")" 6708)) 3026 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (ampersand 6739) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6740) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 3027 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6738) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 3028 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6737) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 3029 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6736) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 3030 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6735) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 3031 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6734) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 3032 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6733) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 3033 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6732) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 3034 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6731) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 3035 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6730) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 3036 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6729) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 3037 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6728) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 3038 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6727) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 3039 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6726) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 3040 (("\"" 1538) ("$" 6741) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 6742) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 6743) ("[" 1560) (array_object_dereferenceable 6744) (callable_expr 6745) (callable_variable 6746) (class_constant 1564) (class_name 6747) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 6748) (name 6749) (new_dereferenceable 1570) (simple_variable 6750) (static_member 6751) (variable 6752) (variable_class_name 6753)) 3041 (("$" 6741) (T_VARIABLE 6743) (simple_variable 6786) ("{" 6787)) 3042 (("(" 6757) (argument_list 6785)) 3043 ((T_NULLSAFE_OBJECT_OPERATOR 6771) (T_OBJECT_OPERATOR 6772) ("[" 6773)) 3044 (("(" 6757) (argument_list 6770)) 3045 ((T_PAAMAYIM_NEKUDOTAYIM 6766)) 3046 (("(" 6757) (argument_list 6765)) 3047 ((T_PAAMAYIM_NEKUDOTAYIM 6754)) 3048 (("$" 6741) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 6743) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 6755) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 6756) ("{" 1660)) 3049 (("(" 6757) (argument_list 6758)) 3050 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 6759) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 6760) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 6761) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3051 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 6764) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3052 (("," 1498) (possible_comma 6762)) 3053 ((")" 6763)) 3054 (("$" 6741) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 6743) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 6767) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 6768) ("{" 1676)) 3055 (("(" 6757) (argument_list 6769)) 3056 (("$" 6741) (T_STRING 6776) (T_VARIABLE 6743) (property_name 6783) (simple_variable 6778) ("{" 6779)) 3057 (("$" 6741) (T_STRING 6776) (T_VARIABLE 6743) (property_name 6777) (simple_variable 6778) ("{" 6779)) 3058 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 6774) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 3059 (("]" 6775)) 3060 (("(" 6757) (argument_list 6782)) 3061 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 6780) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 3062 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 6781)) 3063 (("(" 6757) (argument_list 6784)) 3064 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 6788) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 3065 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 6789)) 3066 ((backup_doc_comment 6792)) 3067 (("(" 6793)) 3068 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 6794)) 3069 ((")" 6795)) 3070 ((T_USE 166) (lexical_vars 6796)) 3071 ((":" 168) (return_type 6797)) 3072 ((backup_fn_flags 6798)) 3073 (("{" 6799)) 3074 ((inner_statement_list 6800)) 3075 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 6801) ("~" 129)) 3076 ((backup_fn_flags 6802)) 3077 ((backup_doc_comment 6804)) 3078 (("(" 6805)) 3079 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 6806)) 3080 ((")" 6807)) 3081 ((":" 2271) (return_type 6808)) 3082 ((T_DOUBLE_ARROW 6809)) 3083 ((backup_fn_flags 6810)) 3084 ((backup_lex_pos 6811)) 3085 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6812) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 3086 (("%" 6514) ("*" 6515) ("+" 6516) ("-" 6517) ("." 6518) ("/" 6519) ("<" 6520) (">" 6521) ("?" 6522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6523) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6524) (T_BOOLEAN_AND 6525) (T_BOOLEAN_OR 6526) (T_COALESCE 6527) (T_INSTANCEOF 6528) (T_IS_EQUAL 6529) (T_IS_GREATER_OR_EQUAL 6530) (T_IS_IDENTICAL 6531) (T_IS_NOT_EQUAL 6532) (T_IS_NOT_IDENTICAL 6533) (T_IS_SMALLER_OR_EQUAL 6534) (T_LOGICAL_AND 6535) (T_LOGICAL_OR 6536) (T_LOGICAL_XOR 6537) (T_PIPE 6538) (T_POW 6539) (T_SL 6540) (T_SPACESHIP 6541) (T_SR 6542) ("^" 6543) (backup_fn_flags 6813) ("|" 6544)) 3087 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 6815) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 3088 (("$" 6436) (T_ABSTRACT 6615) (T_ARRAY 6616) (T_AS 6617) (T_BREAK 6618) (T_CALLABLE 6619) (T_CASE 6620) (T_CATCH 6621) (T_CLASS 6622) (T_CLASS_C 6623) (T_CLONE 6624) (T_CONST 6625) (T_CONTINUE 6626) (T_DECLARE 6627) (T_DEFAULT 6628) (T_DIR 6629) (T_DO 6630) (T_ECHO 6631) (T_ELSE 6632) (T_ELSEIF 6633) (T_EMPTY 6634) (T_ENDDECLARE 6635) (T_ENDFOR 6636) (T_ENDFOREACH 6637) (T_ENDIF 6638) (T_ENDSWITCH 6639) (T_ENDWHILE 6640) (T_ENUM 6641) (T_EVAL 6642) (T_EXIT 6643) (T_EXTENDS 6644) (T_FILE 6645) (T_FINAL 6646) (T_FINALLY 6647) (T_FN 6648) (T_FOR 6649) (T_FOREACH 6650) (T_FUNCTION 6651) (T_FUNC_C 6652) (T_GLOBAL 6653) (T_GOTO 6654) (T_IF 6655) (T_IMPLEMENTS 6656) (T_INCLUDE 6657) (T_INCLUDE_ONCE 6658) (T_INSTANCEOF 6659) (T_INSTEADOF 6660) (T_INTERFACE 6661) (T_ISSET 6662) (T_LINE 6663) (T_LIST 6664) (T_LOGICAL_AND 6665) (T_LOGICAL_OR 6666) (T_LOGICAL_XOR 6667) (T_MATCH 6668) (T_METHOD_C 6669) (T_NAMESPACE 6670) (T_NEW 6671) (T_NS_C 6672) (T_PRINT 6673) (T_PRIVATE 6674) (T_PROPERTY_C 6675) (T_PROTECTED 6676) (T_PUBLIC 6677) (T_READONLY 6678) (T_REQUIRE 6679) (T_REQUIRE_ONCE 6680) (T_RETURN 6681) (T_STATIC 6682) (T_STRING 6683) (T_SWITCH 6684) (T_THROW 6685) (T_TRAIT 6686) (T_TRAIT_C 6687) (T_TRY 6688) (T_UNSET 6689) (T_USE 6690) (T_VAR 6691) (T_VARIABLE 6484) (T_WHILE 6692) (T_YIELD 6693) (identifier 6817) (member_name 6818) (reserved_non_modifiers 6696) (semi_reserved 6697) (simple_variable 6819) ("{" 6820)) 3089 (("(" 6702) (argument_list 6823)) 3090 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 6821) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 3091 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 6822)) 3092 ((T_FN 35) (T_FUNCTION 131) (fn 6498) #'6499 (inline_function 6827)) 3093 (("$" 6436) (T_STRING 6833) (T_VARIABLE 6484) (property_name 6840) (simple_variable 6835) ("{" 6836)) 3094 (("$" 6436) (T_STRING 6833) (T_VARIABLE 6484) (property_name 6834) (simple_variable 6835) ("{" 6836)) 3095 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 6831) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 3096 (("]" 6832)) 3097 (("(" 6702) (argument_list 6839)) 3098 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 6837) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 3099 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 6838)) 3100 (("(" 6702) (argument_list 6841)) 3101 (("`" 6843)) 3102 (("]" 6845)) 3103 (("=" 6846)) 3104 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6847) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 3105 (("%" 6514) ("*" 6515) ("+" 6516) ("-" 6517) ("." 6518) ("/" 6519) ("<" 6520) (">" 6521) ("?" 6522) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 6523) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 6524) (T_BOOLEAN_AND 6525) (T_BOOLEAN_OR 6526) (T_COALESCE 6527) (T_DOUBLE_ARROW 6850) (T_INSTANCEOF 6528) (T_IS_EQUAL 6529) (T_IS_GREATER_OR_EQUAL 6530) (T_IS_IDENTICAL 6531) (T_IS_NOT_EQUAL 6532) (T_IS_NOT_IDENTICAL 6533) (T_IS_SMALLER_OR_EQUAL 6534) (T_LOGICAL_AND 6535) (T_LOGICAL_OR 6536) (T_LOGICAL_XOR 6537) (T_PIPE 6538) (T_POW 6539) (T_SL 6540) (T_SPACESHIP 6541) (T_SR 6542) ("^" 6543) ("|" 6544)) 3106 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6851) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 3107 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 6860) (T_VARIABLE 2920) (encaps_var 2934)) 3108 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 6859) (T_VARIABLE 2920) (encaps_var 2925)) 3109 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 6901) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 3110 ((T_CLASS 6893)) 3111 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 6892) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 6873) (attribute 2982) (class_modifier 2963)) 3112 ((T_PAAMAYIM_NEKUDOTAYIM 6890)) 3113 (("(" 6757) (argument_list 6889)) 3114 ((T_NULLSAFE_OBJECT_OPERATOR 6880) (T_OBJECT_OPERATOR 6881) (T_PAAMAYIM_NEKUDOTAYIM 6882) ("[" 6883)) 3115 (("$" 6741) (T_STRING 6776) (T_VARIABLE 6743) (property_name 6888) (simple_variable 6778) ("{" 6779)) 3116 (("$" 6741) (T_STRING 6776) (T_VARIABLE 6743) (property_name 6887) (simple_variable 6778) ("{" 6779)) 3117 (("$" 6741) (T_VARIABLE 6743) (simple_variable 6886)) 3118 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 6884) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 3119 (("]" 6885)) 3120 (("$" 6741) (T_VARIABLE 6743) (simple_variable 6891)) 3121 (("(" 2984) (argument_list 2985) (ctor_arguments 6894)) 3122 ((T_EXTENDS 2987) (extends_from 6895)) 3123 ((T_IMPLEMENTS 2989) (implements_list 6896)) 3124 ((backup_doc_comment 6897)) 3125 (("{" 6898)) 3126 ((class_statement_list 6899)) 3127 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 6900)) 3128 (("%" 304) (")" 6902) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 3129 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 6904) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 3130 (("%" 304) (")" 6905) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 3131 (("{" 6906)) 3132 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 6907) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 3133 (("}" 6908)) 3134 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 6910) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 3135 ((")" 6911)) 3136 (("=" 6912)) 3137 (("!" 6434) ("\"" 6435) ("$" 6436) ("(" 6437) ("+" 6438) ("-" 6439) ("@" 6440) (T_ARRAY 6441) (T_ARRAY_CAST 6442) (T_ATTRIBUTE 130) (T_BOOL_CAST 6443) (T_CLASS_C 6444) (T_CLONE 6445) (T_CONSTANT_ENCAPSED_STRING 6446) (T_DEC 6447) (T_DIR 6448) (T_DNUMBER 6449) (T_DOUBLE_CAST 6450) (T_EMPTY 6451) (T_EVAL 6452) (T_EXIT 6453) (T_FILE 6454) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6455) (T_INC 6456) (T_INCLUDE 6457) (T_INCLUDE_ONCE 6458) (T_INT_CAST 6459) (T_ISSET 6460) (T_LINE 6461) (T_LIST 6462) (T_LNUMBER 6463) (T_MATCH 6464) (T_METHOD_C 6465) (T_NAME_FULLY_QUALIFIED 6466) (T_NAME_QUALIFIED 6467) (T_NAME_RELATIVE 6468) (T_NEW 6469) (T_NS_C 6470) (T_OBJECT_CAST 6471) (T_PRINT 6472) (T_PROPERTY_C 6473) (T_READONLY 6474) (T_REQUIRE 6475) (T_REQUIRE_ONCE 6476) (T_START_HEREDOC 6477) (T_STATIC 6478) (T_STRING 6479) (T_STRING_CAST 6480) (T_THROW 6481) (T_TRAIT_C 6482) (T_UNSET_CAST 6483) (T_VARIABLE 6484) (T_YIELD 6485) (T_YIELD_FROM 6486) ("[" 6487) ("`" 6488) (array_object_dereferenceable 6489) (attribute 135) (attributes 6490) (callable_expr 6491) (callable_variable 6492) (class_constant 6493) (class_name 6494) (constant 6495) (dereferenceable_scalar 6496) (expr 6913) (fn 6498) (fully_dereferenceable 107) #'6499 (function_call 6500) (inline_function 6501) (internal_functions_in_yacc 6502) (match 6503) (name 6504) (new_dereferenceable 6505) (new_non_dereferenceable 6506) (scalar 6507) (simple_variable 6508) (static_member 6509) (variable 6510) (variable_class_name 6511) ("~" 6512)) 3138 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 6915) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3139 (("," 5226) (possible_comma 6916)) 3140 ((")" 6917)) 3141 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 6925) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 6926) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 6927) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3142 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 6930) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3143 (("," 1498) (possible_comma 6928)) 3144 ((")" 6929)) 3145 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 6932) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 3146 (("%" 304) (")" 6933) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 3147 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 6935) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 3148 (("%" 304) (")" 6936) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 3149 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 6942) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 6943) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 6944) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 6945) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 3150 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 6951) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3151 (("%" 2442) (")" 6948) ("*" 2443) ("+" 2444) ("," 6949) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 3152 (("," 5287) (possible_comma 6946)) 3153 ((")" 6947)) 3154 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 6950) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3155 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 6955) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 3156 ((")" 6956)) 3157 (("%" 304) (")" 6948) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 3158 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 6963) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 3159 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 6964)) 3160 (("\"" 6966) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 3161 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 6976) (T_VARIABLE 2920) (encaps_var 2934)) 3162 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 6975) (T_VARIABLE 2920) (encaps_var 2925)) 3163 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 7017) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 3164 ((T_CLASS 7009)) 3165 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 7008) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 6989) (attribute 2982) (class_modifier 2963)) 3166 ((T_PAAMAYIM_NEKUDOTAYIM 7006)) 3167 (("(" 6340) (argument_list 7005)) 3168 ((T_NULLSAFE_OBJECT_OPERATOR 6996) (T_OBJECT_OPERATOR 6997) (T_PAAMAYIM_NEKUDOTAYIM 6998) ("[" 6999)) 3169 (("$" 6324) (T_STRING 6359) (T_VARIABLE 6326) (property_name 7004) (simple_variable 6361) ("{" 6362)) 3170 (("$" 6324) (T_STRING 6359) (T_VARIABLE 6326) (property_name 7003) (simple_variable 6361) ("{" 6362)) 3171 (("$" 6324) (T_VARIABLE 6326) (simple_variable 7002)) 3172 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 7000) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 3173 (("]" 7001)) 3174 (("$" 6324) (T_VARIABLE 6326) (simple_variable 7007)) 3175 (("(" 2984) (argument_list 2985) (ctor_arguments 7010)) 3176 ((T_EXTENDS 2987) (extends_from 7011)) 3177 ((T_IMPLEMENTS 2989) (implements_list 7012)) 3178 ((backup_doc_comment 7013)) 3179 (("{" 7014)) 3180 ((class_statement_list 7015)) 3181 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 7016)) 3182 (("%" 304) (")" 7018) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 3183 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 7020) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 3184 (("%" 304) (")" 7021) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 3185 (("{" 7022)) 3186 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 7023) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 3187 (("}" 7024)) 3188 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 7026) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 3189 ((")" 7027)) 3190 (("=" 7028)) 3191 (("!" 6017) ("\"" 6018) ("$" 6019) ("(" 6020) ("+" 6021) ("-" 6022) ("@" 6023) (T_ARRAY 6024) (T_ARRAY_CAST 6025) (T_ATTRIBUTE 130) (T_BOOL_CAST 6026) (T_CLASS_C 6027) (T_CLONE 6028) (T_CONSTANT_ENCAPSED_STRING 6029) (T_DEC 6030) (T_DIR 6031) (T_DNUMBER 6032) (T_DOUBLE_CAST 6033) (T_EMPTY 6034) (T_EVAL 6035) (T_EXIT 6036) (T_FILE 6037) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 6038) (T_INC 6039) (T_INCLUDE 6040) (T_INCLUDE_ONCE 6041) (T_INT_CAST 6042) (T_ISSET 6043) (T_LINE 6044) (T_LIST 6045) (T_LNUMBER 6046) (T_MATCH 6047) (T_METHOD_C 6048) (T_NAME_FULLY_QUALIFIED 6049) (T_NAME_QUALIFIED 6050) (T_NAME_RELATIVE 6051) (T_NEW 6052) (T_NS_C 6053) (T_OBJECT_CAST 6054) (T_PRINT 6055) (T_PROPERTY_C 6056) (T_READONLY 6057) (T_REQUIRE 6058) (T_REQUIRE_ONCE 6059) (T_START_HEREDOC 6060) (T_STATIC 6061) (T_STRING 6062) (T_STRING_CAST 6063) (T_THROW 6064) (T_TRAIT_C 6065) (T_UNSET_CAST 6066) (T_VARIABLE 6067) (T_YIELD 6068) (T_YIELD_FROM 6069) ("[" 6070) ("`" 6071) (array_object_dereferenceable 6072) (attribute 135) (attributes 6073) (callable_expr 6074) (callable_variable 6075) (class_constant 6076) (class_name 6077) (constant 6078) (dereferenceable_scalar 6079) (expr 7029) (fn 6081) (fully_dereferenceable 107) #'6082 (function_call 6083) (inline_function 6084) (internal_functions_in_yacc 6085) (match 6086) (name 6087) (new_dereferenceable 6088) (new_non_dereferenceable 6089) (scalar 6090) (simple_variable 6091) (static_member 6092) (variable 6093) (variable_class_name 6094) ("~" 6095)) 3192 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 7031) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3193 (("," 5226) (possible_comma 7032)) 3194 ((")" 7033)) 3195 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 7041) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 7042) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 7043) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3196 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 7046) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3197 (("," 1498) (possible_comma 7044)) 3198 ((")" 7045)) 3199 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 7048) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 3200 (("%" 304) (")" 7049) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 3201 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 7051) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 3202 (("%" 304) (")" 7052) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 3203 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 7058) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 7059) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 7060) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 7061) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 3204 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 7067) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3205 (("%" 2442) (")" 7064) ("*" 2443) ("+" 2444) ("," 7065) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 3206 (("," 5287) (possible_comma 7062)) 3207 ((")" 7063)) 3208 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 7066) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3209 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 7071) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 3210 ((")" 7072)) 3211 (("%" 304) (")" 7064) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 3212 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 7079) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 3213 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 7080)) 3214 (("\"" 7082) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 3215 ((T_ARRAY 3055) (T_CALLABLE 3056) (T_NAME_FULLY_QUALIFIED 3057) (T_NAME_QUALIFIED 3058) (T_NAME_RELATIVE 3059) (T_STRING 3060) (name 3061) (type_without_static 7085)) 3216 ((")" 7095) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3050)) 3217 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 7099) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 7100) (ampersand 7101) (returns_ref 7102)) 3218 ((T_ABSTRACT 7103) (T_ARRAY 7104) (T_AS 7105) (T_BREAK 7106) (T_CALLABLE 7107) (T_CASE 7108) (T_CATCH 7109) (T_CLASS 7110) (T_CLASS_C 7111) (T_CLONE 7112) (T_CONST 7113) (T_CONTINUE 7114) (T_DECLARE 7115) (T_DEFAULT 7116) (T_DIR 7117) (T_DO 7118) (T_ECHO 7119) (T_ELSE 7120) (T_ELSEIF 7121) (T_EMPTY 7122) (T_ENDDECLARE 7123) (T_ENDFOR 7124) (T_ENDFOREACH 7125) (T_ENDIF 7126) (T_ENDSWITCH 7127) (T_ENDWHILE 7128) (T_ENUM 7129) (T_EVAL 7130) (T_EXIT 7131) (T_EXTENDS 7132) (T_FILE 7133) (T_FINAL 7134) (T_FINALLY 7135) (T_FN 7136) (T_FOR 7137) (T_FOREACH 7138) (T_FUNCTION 7139) (T_FUNC_C 7140) (T_GLOBAL 7141) (T_GOTO 7142) (T_IF 7143) (T_IMPLEMENTS 7144) (T_INCLUDE 7145) (T_INCLUDE_ONCE 7146) (T_INSTANCEOF 7147) (T_INSTEADOF 7148) (T_INTERFACE 7149) (T_ISSET 7150) (T_LINE 7151) (T_LIST 7152) (T_LOGICAL_AND 7153) (T_LOGICAL_OR 7154) (T_LOGICAL_XOR 7155) (T_MATCH 7156) (T_METHOD_C 7157) (T_NAMESPACE 7158) (T_NEW 7159) (T_NS_C 7160) (T_PRINT 7161) (T_PRIVATE 7162) (T_PROPERTY_C 7163) (T_PROTECTED 7164) (T_PUBLIC 7165) (T_READONLY 7166) (T_REQUIRE 7167) (T_REQUIRE_ONCE 7168) (T_RETURN 7169) (T_STATIC 7170) (T_STRING 7171) (T_SWITCH 7172) (T_THROW 7173) (T_TRAIT 7174) (T_TRAIT_C 7175) (T_TRY 7176) (T_UNSET 7177) (T_USE 7178) (T_VAR 7179) (T_WHILE 7180) (T_YIELD 7181) (identifier 7182) (reserved_non_modifiers 7183) (semi_reserved 7184)) 3219 ((backup_doc_comment 7185)) 3220 (("(" 7186)) 3221 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 7187)) 3222 ((")" 7188)) 3223 ((":" 7189) (return_type 7190)) 3224 (("(" 2278) ("?" 7198) (T_ARRAY 7199) (T_CALLABLE 7200) (T_NAME_FULLY_QUALIFIED 7201) (T_NAME_QUALIFIED 7202) (T_NAME_RELATIVE 7203) (T_STATIC 7204) (T_STRING 7205) (intersection_type 7206) (name 7207) (type 7208) (type_expr 7209) (type_without_static 7210) (union_type 7211) (union_type_element 7212)) 3225 ((backup_fn_flags 7191)) 3226 ((";" 7192) (method_body 7193) ("{" 7194)) 3227 ((backup_fn_flags 7197)) 3228 ((inner_statement_list 7195)) 3229 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 7196) ("~" 129)) 3230 ((T_ARRAY 7243) (T_CALLABLE 7244) (T_NAME_FULLY_QUALIFIED 7245) (T_NAME_QUALIFIED 7246) (T_NAME_RELATIVE 7247) (T_STATIC 7248) (T_STRING 7249) (name 7250) (type 7251) (type_without_static 7252)) 3231 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 7241)) 3232 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 7230)) 3233 (("|" 7228)) 3234 (("|" 7213)) 3235 (("(" 7214) (T_ARRAY 7215) (T_CALLABLE 7216) (T_NAME_FULLY_QUALIFIED 7217) (T_NAME_QUALIFIED 7218) (T_NAME_RELATIVE 7219) (T_STATIC 7220) (T_STRING 7221) (name 7222) (type 7223) (type_without_static 7224) (union_type_element 7225)) 3236 ((T_ARRAY 2307) (T_CALLABLE 2308) (T_NAME_FULLY_QUALIFIED 2309) (T_NAME_QUALIFIED 2310) (T_NAME_RELATIVE 2311) (T_STATIC 2312) (T_STRING 2313) (intersection_type 7226) (name 2315) (type 2316) (type_without_static 2317)) 3237 ((")" 7227) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2330)) 3238 (("(" 7214) (T_ARRAY 7215) (T_CALLABLE 7216) (T_NAME_FULLY_QUALIFIED 7217) (T_NAME_QUALIFIED 7218) (T_NAME_RELATIVE 7219) (T_STATIC 7220) (T_STRING 7221) (name 7222) (type 7223) (type_without_static 7224) (union_type_element 7229)) 3239 ((T_ARRAY 7231) (T_CALLABLE 7232) (T_NAME_FULLY_QUALIFIED 7233) (T_NAME_QUALIFIED 7234) (T_NAME_RELATIVE 7235) (T_STATIC 7236) (T_STRING 7237) (name 7238) (type 7239) (type_without_static 7240)) 3240 ((T_ARRAY 7231) (T_CALLABLE 7232) (T_NAME_FULLY_QUALIFIED 7233) (T_NAME_QUALIFIED 7234) (T_NAME_RELATIVE 7235) (T_STATIC 7236) (T_STRING 7237) (name 7238) (type 7242) (type_without_static 7240)) 3241 (("(" 2278) ("?" 7254) (T_ABSTRACT 7255) (T_ARRAY 7256) (T_AS 7257) (T_BREAK 7258) (T_CALLABLE 7259) (T_CASE 7260) (T_CATCH 7261) (T_CLASS 7262) (T_CLASS_C 7263) (T_CLONE 7264) (T_CONST 7265) (T_CONTINUE 7266) (T_DECLARE 7267) (T_DEFAULT 7268) (T_DIR 7269) (T_DO 7270) (T_ECHO 7271) (T_ELSE 7272) (T_ELSEIF 7273) (T_EMPTY 7274) (T_ENDDECLARE 7275) (T_ENDFOR 7276) (T_ENDFOREACH 7277) (T_ENDIF 7278) (T_ENDSWITCH 7279) (T_ENDWHILE 7280) (T_ENUM 7281) (T_EVAL 7282) (T_EXIT 7283) (T_EXTENDS 7284) (T_FILE 7285) (T_FINAL 7286) (T_FINALLY 7287) (T_FN 7288) (T_FOR 7289) (T_FOREACH 7290) (T_FUNCTION 7291) (T_FUNC_C 7292) (T_GLOBAL 7293) (T_GOTO 7294) (T_IF 7295) (T_IMPLEMENTS 7296) (T_INCLUDE 7297) (T_INCLUDE_ONCE 7298) (T_INSTANCEOF 7299) (T_INSTEADOF 7300) (T_INTERFACE 7301) (T_ISSET 7302) (T_LINE 7303) (T_LIST 7304) (T_LOGICAL_AND 7305) (T_LOGICAL_OR 7306) (T_LOGICAL_XOR 7307) (T_MATCH 7308) (T_METHOD_C 7309) (T_NAMESPACE 7310) (T_NAME_FULLY_QUALIFIED 7311) (T_NAME_QUALIFIED 7312) (T_NAME_RELATIVE 7313) (T_NEW 7314) (T_NS_C 7315) (T_PRINT 7316) (T_PRIVATE 7317) (T_PROPERTY_C 7318) (T_PROTECTED 7319) (T_PUBLIC 7320) (T_READONLY 7321) (T_REQUIRE 7322) (T_REQUIRE_ONCE 7323) (T_RETURN 7324) (T_STATIC 7325) (T_STRING 7326) (T_SWITCH 7327) (T_THROW 7328) (T_TRAIT 7329) (T_TRAIT_C 7330) (T_TRY 7331) (T_UNSET 7332) (T_USE 7333) (T_VAR 7334) (T_WHILE 7335) (T_YIELD 7336) (class_const_decl 7337) (class_const_list 7338) (intersection_type 7339) (name 7340) (reserved_non_modifiers 7341) (semi_reserved 7342) (type 7343) (type_expr 7344) (type_without_static 7345) (union_type 7346) (union_type_element 7347)) 3242 ((T_ARRAY 7393) (T_CALLABLE 7394) (T_NAME_FULLY_QUALIFIED 7395) (T_NAME_QUALIFIED 7396) (T_NAME_RELATIVE 7397) (T_STATIC 7398) (T_STRING 7399) (name 7400) (type 7401) (type_without_static 7402)) 3243 (("=" 7373)) 3244 (("," 7370) (";" 7392)) 3245 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 7390)) 3246 (("=" 7387)) 3247 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 7376)) 3248 ((T_ABSTRACT 7255) (T_ARRAY 7365) (T_AS 7257) (T_BREAK 7258) (T_CALLABLE 7366) (T_CASE 7260) (T_CATCH 7261) (T_CLASS 7262) (T_CLASS_C 7263) (T_CLONE 7264) (T_CONST 7265) (T_CONTINUE 7266) (T_DECLARE 7267) (T_DEFAULT 7268) (T_DIR 7269) (T_DO 7270) (T_ECHO 7271) (T_ELSE 7272) (T_ELSEIF 7273) (T_EMPTY 7274) (T_ENDDECLARE 7275) (T_ENDFOR 7276) (T_ENDFOREACH 7277) (T_ENDIF 7278) (T_ENDSWITCH 7279) (T_ENDWHILE 7280) (T_ENUM 7281) (T_EVAL 7282) (T_EXIT 7283) (T_EXTENDS 7284) (T_FILE 7285) (T_FINAL 7286) (T_FINALLY 7287) (T_FN 7288) (T_FOR 7289) (T_FOREACH 7290) (T_FUNCTION 7291) (T_FUNC_C 7292) (T_GLOBAL 7293) (T_GOTO 7294) (T_IF 7295) (T_IMPLEMENTS 7296) (T_INCLUDE 7297) (T_INCLUDE_ONCE 7298) (T_INSTANCEOF 7299) (T_INSTEADOF 7300) (T_INTERFACE 7301) (T_ISSET 7302) (T_LINE 7303) (T_LIST 7304) (T_LOGICAL_AND 7305) (T_LOGICAL_OR 7306) (T_LOGICAL_XOR 7307) (T_MATCH 7308) (T_METHOD_C 7309) (T_NAMESPACE 7310) (T_NEW 7314) (T_NS_C 7315) (T_PRINT 7316) (T_PRIVATE 7317) (T_PROPERTY_C 7318) (T_PROTECTED 7319) (T_PUBLIC 7320) (T_READONLY 7321) (T_REQUIRE 7322) (T_REQUIRE_ONCE 7323) (T_RETURN 7324) (T_STATIC 7367) (T_STRING 7368) (T_SWITCH 7327) (T_THROW 7328) (T_TRAIT 7329) (T_TRAIT_C 7330) (T_TRY 7331) (T_UNSET 7332) (T_USE 7333) (T_VAR 7334) (T_WHILE 7335) (T_YIELD 7336) (class_const_decl 7337) (class_const_list 7369) (reserved_non_modifiers 7341) (semi_reserved 7342)) 3249 (("|" 7363)) 3250 (("|" 7348)) 3251 (("(" 7349) (T_ARRAY 7350) (T_CALLABLE 7351) (T_NAME_FULLY_QUALIFIED 7352) (T_NAME_QUALIFIED 7353) (T_NAME_RELATIVE 7354) (T_STATIC 7355) (T_STRING 7356) (name 7357) (type 7358) (type_without_static 7359) (union_type_element 7360)) 3252 ((T_ARRAY 2307) (T_CALLABLE 2308) (T_NAME_FULLY_QUALIFIED 2309) (T_NAME_QUALIFIED 2310) (T_NAME_RELATIVE 2311) (T_STATIC 2312) (T_STRING 2313) (intersection_type 7361) (name 2315) (type 2316) (type_without_static 2317)) 3253 ((")" 7362) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2330)) 3254 (("(" 7349) (T_ARRAY 7350) (T_CALLABLE 7351) (T_NAME_FULLY_QUALIFIED 7352) (T_NAME_QUALIFIED 7353) (T_NAME_RELATIVE 7354) (T_STATIC 7355) (T_STRING 7356) (name 7357) (type 7358) (type_without_static 7359) (union_type_element 7364)) 3255 (("," 7370) (";" 7371)) 3256 ((T_ABSTRACT 7255) (T_ARRAY 7365) (T_AS 7257) (T_BREAK 7258) (T_CALLABLE 7366) (T_CASE 7260) (T_CATCH 7261) (T_CLASS 7262) (T_CLASS_C 7263) (T_CLONE 7264) (T_CONST 7265) (T_CONTINUE 7266) (T_DECLARE 7267) (T_DEFAULT 7268) (T_DIR 7269) (T_DO 7270) (T_ECHO 7271) (T_ELSE 7272) (T_ELSEIF 7273) (T_EMPTY 7274) (T_ENDDECLARE 7275) (T_ENDFOR 7276) (T_ENDFOREACH 7277) (T_ENDIF 7278) (T_ENDSWITCH 7279) (T_ENDWHILE 7280) (T_ENUM 7281) (T_EVAL 7282) (T_EXIT 7283) (T_EXTENDS 7284) (T_FILE 7285) (T_FINAL 7286) (T_FINALLY 7287) (T_FN 7288) (T_FOR 7289) (T_FOREACH 7290) (T_FUNCTION 7291) (T_FUNC_C 7292) (T_GLOBAL 7293) (T_GOTO 7294) (T_IF 7295) (T_IMPLEMENTS 7296) (T_INCLUDE 7297) (T_INCLUDE_ONCE 7298) (T_INSTANCEOF 7299) (T_INSTEADOF 7300) (T_INTERFACE 7301) (T_ISSET 7302) (T_LINE 7303) (T_LIST 7304) (T_LOGICAL_AND 7305) (T_LOGICAL_OR 7306) (T_LOGICAL_XOR 7307) (T_MATCH 7308) (T_METHOD_C 7309) (T_NAMESPACE 7310) (T_NEW 7314) (T_NS_C 7315) (T_PRINT 7316) (T_PRIVATE 7317) (T_PROPERTY_C 7318) (T_PROTECTED 7319) (T_PUBLIC 7320) (T_READONLY 7321) (T_REQUIRE 7322) (T_REQUIRE_ONCE 7323) (T_RETURN 7324) (T_STATIC 7367) (T_STRING 7368) (T_SWITCH 7327) (T_THROW 7328) (T_TRAIT 7329) (T_TRAIT_C 7330) (T_TRY 7331) (T_UNSET 7332) (T_USE 7333) (T_VAR 7334) (T_WHILE 7335) (T_YIELD 7336) (class_const_decl 7372) (reserved_non_modifiers 7341) (semi_reserved 7342)) 3257 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 7374) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 3258 (("%" 3153) ("*" 3154) ("+" 3155) ("-" 3156) ("." 3157) ("/" 3158) ("<" 3159) (">" 3160) ("?" 3161) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 3162) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3163) (T_BOOLEAN_AND 3164) (T_BOOLEAN_OR 3165) (T_COALESCE 3166) (T_INSTANCEOF 3167) (T_IS_EQUAL 3168) (T_IS_GREATER_OR_EQUAL 3169) (T_IS_IDENTICAL 3170) (T_IS_NOT_EQUAL 3171) (T_IS_NOT_IDENTICAL 3172) (T_IS_SMALLER_OR_EQUAL 3173) (T_LOGICAL_AND 3174) (T_LOGICAL_OR 3175) (T_LOGICAL_XOR 3176) (T_PIPE 3177) (T_POW 3178) (T_SL 3179) (T_SPACESHIP 3180) (T_SR 3181) ("^" 3182) (backup_doc_comment 7375) ("|" 3183)) 3259 ((T_ARRAY 7377) (T_CALLABLE 7378) (T_NAME_FULLY_QUALIFIED 7379) (T_NAME_QUALIFIED 7380) (T_NAME_RELATIVE 7381) (T_STATIC 7382) (T_STRING 7383) (name 7384) (type 7385) (type_without_static 7386)) 3260 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 7388) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 3261 (("%" 3153) ("*" 3154) ("+" 3155) ("-" 3156) ("." 3157) ("/" 3158) ("<" 3159) (">" 3160) ("?" 3161) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 3162) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3163) (T_BOOLEAN_AND 3164) (T_BOOLEAN_OR 3165) (T_COALESCE 3166) (T_INSTANCEOF 3167) (T_IS_EQUAL 3168) (T_IS_GREATER_OR_EQUAL 3169) (T_IS_IDENTICAL 3170) (T_IS_NOT_EQUAL 3171) (T_IS_NOT_IDENTICAL 3172) (T_IS_SMALLER_OR_EQUAL 3173) (T_LOGICAL_AND 3174) (T_LOGICAL_OR 3175) (T_LOGICAL_XOR 3176) (T_PIPE 3177) (T_POW 3178) (T_SL 3179) (T_SPACESHIP 3180) (T_SR 3181) ("^" 3182) (backup_doc_comment 7389) ("|" 3183)) 3262 ((T_ARRAY 7377) (T_CALLABLE 7378) (T_NAME_FULLY_QUALIFIED 7379) (T_NAME_QUALIFIED 7380) (T_NAME_RELATIVE 7381) (T_STATIC 7382) (T_STRING 7383) (name 7384) (type 7391) (type_without_static 7386)) 3263 (("," 7413) (";" 7414) (trait_adaptations 7415) ("{" 7416)) 3264 ((T_NAME_FULLY_QUALIFIED 7405) (T_NAME_QUALIFIED 7406) (T_NAME_RELATIVE 7407) (T_STATIC 7408) (T_STRING 7409) (class_name 7705) (name 7412)) 3265 ((T_ABSTRACT 7417) (T_ARRAY 7418) (T_AS 7419) (T_BREAK 7420) (T_CALLABLE 7421) (T_CASE 7422) (T_CATCH 7423) (T_CLASS 7424) (T_CLASS_C 7425) (T_CLONE 7426) (T_CONST 7427) (T_CONTINUE 7428) (T_DECLARE 7429) (T_DEFAULT 7430) (T_DIR 7431) (T_DO 7432) (T_ECHO 7433) (T_ELSE 7434) (T_ELSEIF 7435) (T_EMPTY 7436) (T_ENDDECLARE 7437) (T_ENDFOR 7438) (T_ENDFOREACH 7439) (T_ENDIF 7440) (T_ENDSWITCH 7441) (T_ENDWHILE 7442) (T_ENUM 7443) (T_EVAL 7444) (T_EXIT 7445) (T_EXTENDS 7446) (T_FILE 7447) (T_FINAL 7448) (T_FINALLY 7449) (T_FN 7450) (T_FOR 7451) (T_FOREACH 7452) (T_FUNCTION 7453) (T_FUNC_C 7454) (T_GLOBAL 7455) (T_GOTO 7456) (T_IF 7457) (T_IMPLEMENTS 7458) (T_INCLUDE 7459) (T_INCLUDE_ONCE 7460) (T_INSTANCEOF 7461) (T_INSTEADOF 7462) (T_INTERFACE 7463) (T_ISSET 7464) (T_LINE 7465) (T_LIST 7466) (T_LOGICAL_AND 7467) (T_LOGICAL_OR 7468) (T_LOGICAL_XOR 7469) (T_MATCH 7470) (T_METHOD_C 7471) (T_NAMESPACE 7472) (T_NAME_FULLY_QUALIFIED 7473) (T_NAME_QUALIFIED 7474) (T_NAME_RELATIVE 7475) (T_NEW 7476) (T_NS_C 7477) (T_PRINT 7478) (T_PRIVATE 7479) (T_PROPERTY_C 7480) (T_PROTECTED 7481) (T_PUBLIC 7482) (T_READONLY 7483) (T_REQUIRE 7484) (T_REQUIRE_ONCE 7485) (T_RETURN 7486) (T_STATIC 7487) (T_STRING 7488) (T_SWITCH 7489) (T_THROW 7490) (T_TRAIT 7491) (T_TRAIT_C 7492) (T_TRY 7493) (T_UNSET 7494) (T_USE 7495) (T_VAR 7496) (T_WHILE 7497) (T_YIELD 7498) (absolute_trait_method_reference 7499) (class_name 7500) (identifier 7501) (name 7502) (reserved_non_modifiers 7503) (semi_reserved 7504) (trait_adaptation 7505) (trait_adaptation_list 7506) (trait_alias 7507) (trait_method_reference 7508) (trait_precedence 7509) ("}" 7510)) 3266 ((T_INSTEADOF 7694)) 3267 ((T_PAAMAYIM_NEKUDOTAYIM 7611)) 3268 ((T_ABSTRACT 7417) (T_ARRAY 7418) (T_AS 7419) (T_BREAK 7420) (T_CALLABLE 7421) (T_CASE 7422) (T_CATCH 7423) (T_CLASS 7424) (T_CLASS_C 7425) (T_CLONE 7426) (T_CONST 7427) (T_CONTINUE 7428) (T_DECLARE 7429) (T_DEFAULT 7430) (T_DIR 7431) (T_DO 7432) (T_ECHO 7433) (T_ELSE 7434) (T_ELSEIF 7435) (T_EMPTY 7436) (T_ENDDECLARE 7437) (T_ENDFOR 7438) (T_ENDFOREACH 7439) (T_ENDIF 7440) (T_ENDSWITCH 7441) (T_ENDWHILE 7442) (T_ENUM 7443) (T_EVAL 7444) (T_EXIT 7445) (T_EXTENDS 7446) (T_FILE 7447) (T_FINAL 7448) (T_FINALLY 7449) (T_FN 7450) (T_FOR 7451) (T_FOREACH 7452) (T_FUNCTION 7453) (T_FUNC_C 7454) (T_GLOBAL 7455) (T_GOTO 7456) (T_IF 7457) (T_IMPLEMENTS 7458) (T_INCLUDE 7459) (T_INCLUDE_ONCE 7460) (T_INSTANCEOF 7461) (T_INSTEADOF 7462) (T_INTERFACE 7463) (T_ISSET 7464) (T_LINE 7465) (T_LIST 7466) (T_LOGICAL_AND 7467) (T_LOGICAL_OR 7468) (T_LOGICAL_XOR 7469) (T_MATCH 7470) (T_METHOD_C 7471) (T_NAMESPACE 7472) (T_NAME_FULLY_QUALIFIED 7473) (T_NAME_QUALIFIED 7474) (T_NAME_RELATIVE 7475) (T_NEW 7476) (T_NS_C 7477) (T_PRINT 7478) (T_PRIVATE 7479) (T_PROPERTY_C 7480) (T_PROTECTED 7481) (T_PUBLIC 7482) (T_READONLY 7483) (T_REQUIRE 7484) (T_REQUIRE_ONCE 7485) (T_RETURN 7486) (T_STATIC 7487) (T_STRING 7488) (T_SWITCH 7489) (T_THROW 7490) (T_TRAIT 7491) (T_TRAIT_C 7492) (T_TRY 7493) (T_UNSET 7494) (T_USE 7495) (T_VAR 7496) (T_WHILE 7497) (T_YIELD 7498) (absolute_trait_method_reference 7499) (class_name 7500) (identifier 7501) (name 7502) (reserved_non_modifiers 7503) (semi_reserved 7504) (trait_adaptation 7609) (trait_alias 7507) (trait_method_reference 7508) (trait_precedence 7509) ("}" 7610)) 3269 ((";" 7608)) 3270 ((T_AS 7512)) 3271 ((";" 7511)) 3272 ((T_ABSTRACT 7513) (T_ARRAY 7514) (T_AS 7515) (T_BREAK 7516) (T_CALLABLE 7517) (T_CASE 7518) (T_CATCH 7519) (T_CLASS 7520) (T_CLASS_C 7521) (T_CLONE 7522) (T_CONST 7523) (T_CONTINUE 7524) (T_DECLARE 7525) (T_DEFAULT 7526) (T_DIR 7527) (T_DO 7528) (T_ECHO 7529) (T_ELSE 7530) (T_ELSEIF 7531) (T_EMPTY 7532) (T_ENDDECLARE 7533) (T_ENDFOR 7534) (T_ENDFOREACH 7535) (T_ENDIF 7536) (T_ENDSWITCH 7537) (T_ENDWHILE 7538) (T_ENUM 7539) (T_EVAL 7540) (T_EXIT 7541) (T_EXTENDS 7542) (T_FILE 7543) (T_FINAL 7544) (T_FINALLY 7545) (T_FN 7546) (T_FOR 7547) (T_FOREACH 7548) (T_FUNCTION 7549) (T_FUNC_C 7550) (T_GLOBAL 7551) (T_GOTO 7552) (T_IF 7553) (T_IMPLEMENTS 7554) (T_INCLUDE 7555) (T_INCLUDE_ONCE 7556) (T_INSTANCEOF 7557) (T_INSTEADOF 7558) (T_INTERFACE 7559) (T_ISSET 7560) (T_LINE 7561) (T_LIST 7562) (T_LOGICAL_AND 7563) (T_LOGICAL_OR 7564) (T_LOGICAL_XOR 7565) (T_MATCH 7566) (T_METHOD_C 7567) (T_NAMESPACE 7568) (T_NEW 7569) (T_NS_C 7570) (T_PRINT 7571) (T_PRIVATE 7572) (T_PRIVATE_SET 7573) (T_PROPERTY_C 7574) (T_PROTECTED 7575) (T_PROTECTED_SET 7576) (T_PUBLIC 7577) (T_PUBLIC_SET 7578) (T_READONLY 7579) (T_REQUIRE 7580) (T_REQUIRE_ONCE 7581) (T_RETURN 7582) (T_STATIC 7583) (T_STRING 7584) (T_SWITCH 7585) (T_THROW 7586) (T_TRAIT 7587) (T_TRAIT_C 7588) (T_TRY 7589) (T_UNSET 7590) (T_USE 7591) (T_VAR 7592) (T_WHILE 7593) (T_YIELD 7594) (member_modifier 7595) (reserved_non_modifiers 7596)) 3273 ((T_ABSTRACT 7597) (T_ARRAY 7514) (T_AS 7515) (T_BREAK 7516) (T_CALLABLE 7517) (T_CASE 7518) (T_CATCH 7519) (T_CLASS 7520) (T_CLASS_C 7521) (T_CLONE 7522) (T_CONST 7523) (T_CONTINUE 7524) (T_DECLARE 7525) (T_DEFAULT 7526) (T_DIR 7527) (T_DO 7528) (T_ECHO 7529) (T_ELSE 7530) (T_ELSEIF 7531) (T_EMPTY 7532) (T_ENDDECLARE 7533) (T_ENDFOR 7534) (T_ENDFOREACH 7535) (T_ENDIF 7536) (T_ENDSWITCH 7537) (T_ENDWHILE 7538) (T_ENUM 7539) (T_EVAL 7540) (T_EXIT 7541) (T_EXTENDS 7542) (T_FILE 7543) (T_FINAL 7598) (T_FINALLY 7545) (T_FN 7546) (T_FOR 7547) (T_FOREACH 7548) (T_FUNCTION 7549) (T_FUNC_C 7550) (T_GLOBAL 7551) (T_GOTO 7552) (T_IF 7553) (T_IMPLEMENTS 7554) (T_INCLUDE 7555) (T_INCLUDE_ONCE 7556) (T_INSTANCEOF 7557) (T_INSTEADOF 7558) (T_INTERFACE 7559) (T_ISSET 7560) (T_LINE 7561) (T_LIST 7562) (T_LOGICAL_AND 7563) (T_LOGICAL_OR 7564) (T_LOGICAL_XOR 7565) (T_MATCH 7566) (T_METHOD_C 7567) (T_NAMESPACE 7568) (T_NEW 7569) (T_NS_C 7570) (T_PRINT 7571) (T_PRIVATE 7599) (T_PROPERTY_C 7574) (T_PROTECTED 7600) (T_PUBLIC 7601) (T_READONLY 7602) (T_REQUIRE 7580) (T_REQUIRE_ONCE 7581) (T_RETURN 7582) (T_STATIC 7603) (T_STRING 7604) (T_SWITCH 7585) (T_THROW 7586) (T_TRAIT 7587) (T_TRAIT_C 7588) (T_TRY 7589) (T_UNSET 7590) (T_USE 7591) (T_VAR 7592) (T_WHILE 7593) (T_YIELD 7594) (identifier 7605) (reserved_non_modifiers 7606) (semi_reserved 7607)) 3274 ((T_ABSTRACT 7612) (T_ARRAY 7613) (T_AS 7614) (T_BREAK 7615) (T_CALLABLE 7616) (T_CASE 7617) (T_CATCH 7618) (T_CLASS 7619) (T_CLASS_C 7620) (T_CLONE 7621) (T_CONST 7622) (T_CONTINUE 7623) (T_DECLARE 7624) (T_DEFAULT 7625) (T_DIR 7626) (T_DO 7627) (T_ECHO 7628) (T_ELSE 7629) (T_ELSEIF 7630) (T_EMPTY 7631) (T_ENDDECLARE 7632) (T_ENDFOR 7633) (T_ENDFOREACH 7634) (T_ENDIF 7635) (T_ENDSWITCH 7636) (T_ENDWHILE 7637) (T_ENUM 7638) (T_EVAL 7639) (T_EXIT 7640) (T_EXTENDS 7641) (T_FILE 7642) (T_FINAL 7643) (T_FINALLY 7644) (T_FN 7645) (T_FOR 7646) (T_FOREACH 7647) (T_FUNCTION 7648) (T_FUNC_C 7649) (T_GLOBAL 7650) (T_GOTO 7651) (T_IF 7652) (T_IMPLEMENTS 7653) (T_INCLUDE 7654) (T_INCLUDE_ONCE 7655) (T_INSTANCEOF 7656) (T_INSTEADOF 7657) (T_INTERFACE 7658) (T_ISSET 7659) (T_LINE 7660) (T_LIST 7661) (T_LOGICAL_AND 7662) (T_LOGICAL_OR 7663) (T_LOGICAL_XOR 7664) (T_MATCH 7665) (T_METHOD_C 7666) (T_NAMESPACE 7667) (T_NEW 7668) (T_NS_C 7669) (T_PRINT 7670) (T_PRIVATE 7671) (T_PROPERTY_C 7672) (T_PROTECTED 7673) (T_PUBLIC 7674) (T_READONLY 7675) (T_REQUIRE 7676) (T_REQUIRE_ONCE 7677) (T_RETURN 7678) (T_STATIC 7679) (T_STRING 7680) (T_SWITCH 7681) (T_THROW 7682) (T_TRAIT 7683) (T_TRAIT_C 7684) (T_TRY 7685) (T_UNSET 7686) (T_USE 7687) (T_VAR 7688) (T_WHILE 7689) (T_YIELD 7690) (identifier 7691) (reserved_non_modifiers 7692) (semi_reserved 7693)) 3275 ((T_NAME_FULLY_QUALIFIED 7695) (T_NAME_QUALIFIED 7696) (T_NAME_RELATIVE 7697) (T_STATIC 7698) (T_STRING 7699) (class_name 7700) (class_name_list 7701) (name 7702)) 3276 (("," 7703)) 3277 ((T_NAME_FULLY_QUALIFIED 7695) (T_NAME_QUALIFIED 7696) (T_NAME_RELATIVE 7697) (T_STATIC 7698) (T_STRING 7699) (class_name 7704) (name 7702)) 3278 ((T_ABSTRACT 7707) (T_ARRAY 7708) (T_AS 7709) (T_BREAK 7710) (T_CALLABLE 7711) (T_CASE 7712) (T_CATCH 7713) (T_CLASS 7714) (T_CLASS_C 7715) (T_CLONE 7716) (T_CONST 7717) (T_CONTINUE 7718) (T_DECLARE 7719) (T_DEFAULT 7720) (T_DIR 7721) (T_DO 7722) (T_ECHO 7723) (T_ELSE 7724) (T_ELSEIF 7725) (T_EMPTY 7726) (T_ENDDECLARE 7727) (T_ENDFOR 7728) (T_ENDFOREACH 7729) (T_ENDIF 7730) (T_ENDSWITCH 7731) (T_ENDWHILE 7732) (T_ENUM 7733) (T_EVAL 7734) (T_EXIT 7735) (T_EXTENDS 7736) (T_FILE 7737) (T_FINAL 7738) (T_FINALLY 7739) (T_FN 7740) (T_FOR 7741) (T_FOREACH 7742) (T_FUNCTION 7743) (T_FUNC_C 7744) (T_GLOBAL 7745) (T_GOTO 7746) (T_IF 7747) (T_IMPLEMENTS 7748) (T_INCLUDE 7749) (T_INCLUDE_ONCE 7750) (T_INSTANCEOF 7751) (T_INSTEADOF 7752) (T_INTERFACE 7753) (T_ISSET 7754) (T_LINE 7755) (T_LIST 7756) (T_LOGICAL_AND 7757) (T_LOGICAL_OR 7758) (T_LOGICAL_XOR 7759) (T_MATCH 7760) (T_METHOD_C 7761) (T_NAMESPACE 7762) (T_NEW 7763) (T_NS_C 7764) (T_PRINT 7765) (T_PRIVATE 7766) (T_PROPERTY_C 7767) (T_PROTECTED 7768) (T_PUBLIC 7769) (T_READONLY 7770) (T_REQUIRE 7771) (T_REQUIRE_ONCE 7772) (T_RETURN 7773) (T_STATIC 7774) (T_STRING 7775) (T_SWITCH 7776) (T_THROW 7777) (T_TRAIT 7778) (T_TRAIT_C 7779) (T_TRY 7780) (T_UNSET 7781) (T_USE 7782) (T_VAR 7783) (T_WHILE 7784) (T_YIELD 7785) (identifier 7786) (reserved_non_modifiers 7787) (semi_reserved 7788)) 3279 (("=" 7789) (enum_case_expr 7790)) 3280 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 7792) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 3281 ((";" 7791)) 3282 (("," 6013) (possible_comma 7794)) 3283 (("]" 7795)) 3284 (("," 7804)) 3285 ((T_NAME_FULLY_QUALIFIED 7796) (T_NAME_QUALIFIED 7797) (T_NAME_RELATIVE 7798) (T_STATIC 7799) (T_STRING 7800) (class_name 7805) (name 7803)) 3286 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 7818) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3287 (("," 1498) (possible_comma 7816)) 3288 ((")" 7817)) 3289 (("," 6013) (possible_comma 7821)) 3290 (("]" 7822)) 3291 (("%" 304) (")" 7824) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 3292 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 7826) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 3293 ((")" 7827)) 3294 (("%" 304) (")" 5332) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 3295 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 7840) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 3296 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 7841)) 3297 (("\"" 7843) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 3298 (("$" 1885) (T_ABSTRACT 2076) (T_ARRAY 2077) (T_AS 2078) (T_BREAK 2079) (T_CALLABLE 2080) (T_CASE 2081) (T_CATCH 2082) (T_CLASS 2083) (T_CLASS_C 2084) (T_CLONE 2085) (T_CONST 2086) (T_CONTINUE 2087) (T_DECLARE 2088) (T_DEFAULT 2089) (T_DIR 2090) (T_DO 2091) (T_ECHO 2092) (T_ELSE 2093) (T_ELSEIF 2094) (T_EMPTY 2095) (T_ENDDECLARE 2096) (T_ENDFOR 2097) (T_ENDFOREACH 2098) (T_ENDIF 2099) (T_ENDSWITCH 2100) (T_ENDWHILE 2101) (T_ENUM 2102) (T_EVAL 2103) (T_EXIT 2104) (T_EXTENDS 2105) (T_FILE 2106) (T_FINAL 2107) (T_FINALLY 2108) (T_FN 2109) (T_FOR 2110) (T_FOREACH 2111) (T_FUNCTION 2112) (T_FUNC_C 2113) (T_GLOBAL 2114) (T_GOTO 2115) (T_IF 2116) (T_IMPLEMENTS 2117) (T_INCLUDE 2118) (T_INCLUDE_ONCE 2119) (T_INSTANCEOF 2120) (T_INSTEADOF 2121) (T_INTERFACE 2122) (T_ISSET 2123) (T_LINE 2124) (T_LIST 2125) (T_LOGICAL_AND 2126) (T_LOGICAL_OR 2127) (T_LOGICAL_XOR 2128) (T_MATCH 2129) (T_METHOD_C 2130) (T_NAMESPACE 2131) (T_NEW 2132) (T_NS_C 2133) (T_PRINT 2134) (T_PRIVATE 2135) (T_PROPERTY_C 2136) (T_PROTECTED 2137) (T_PUBLIC 2138) (T_READONLY 2139) (T_REQUIRE 2140) (T_REQUIRE_ONCE 2141) (T_RETURN 2142) (T_STATIC 2143) (T_STRING 2144) (T_SWITCH 2145) (T_THROW 2146) (T_TRAIT 2147) (T_TRAIT_C 2148) (T_TRY 2149) (T_UNSET 2150) (T_USE 2151) (T_VAR 2152) (T_VARIABLE 1934) (T_WHILE 2153) (T_YIELD 2154) (identifier 7846) (member_name 7847) (reserved_non_modifiers 2157) (semi_reserved 2158) (simple_variable 7848) ("{" 7849)) 3299 (("(" 2163) (argument_list 7852)) 3300 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 7850) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 3301 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 7851)) 3302 ((T_FN 35) (T_FUNCTION 131) (fn 1951) #'1952 (inline_function 7856)) 3303 (("=" 7858)) 3304 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 7859) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 3305 (("$" 1885) (T_STRING 7865) (T_VARIABLE 1934) (property_name 7872) (simple_variable 7867) ("{" 7868)) 3306 (("$" 1885) (T_STRING 7865) (T_VARIABLE 1934) (property_name 7866) (simple_variable 7867) ("{" 7868)) 3307 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 7863) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 3308 (("]" 7864)) 3309 (("(" 2163) (argument_list 7871)) 3310 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 7869) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 3311 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 7870)) 3312 (("(" 2163) (argument_list 7873)) 3313 (("`" 7876)) 3314 (("]" 7878)) 3315 (("=" 7879)) 3316 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 7880) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 3317 (("%" 1970) ("*" 1971) ("+" 1972) ("-" 1973) ("." 1974) ("/" 1975) ("<" 1976) (">" 1977) ("?" 1978) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1979) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1980) (T_BOOLEAN_AND 1981) (T_BOOLEAN_OR 1982) (T_COALESCE 1983) (T_DOUBLE_ARROW 7883) (T_INSTANCEOF 1984) (T_IS_EQUAL 1985) (T_IS_GREATER_OR_EQUAL 1986) (T_IS_IDENTICAL 1987) (T_IS_NOT_EQUAL 1988) (T_IS_NOT_IDENTICAL 1989) (T_IS_SMALLER_OR_EQUAL 1990) (T_LOGICAL_AND 1991) (T_LOGICAL_OR 1992) (T_LOGICAL_XOR 1993) (T_PIPE 1994) (T_POW 1995) (T_SL 1996) (T_SPACESHIP 1997) (T_SR 1998) ("^" 1999) ("|" 2000)) 3318 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1968) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (array_object_dereferenceable 1940) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 7884) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 3319 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 7893) (T_VARIABLE 2920) (encaps_var 2934)) 3320 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 7892) (T_VARIABLE 2920) (encaps_var 2925)) 3321 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 7934) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 3322 ((T_CLASS 7926)) 3323 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 7925) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 7906) (attribute 2982) (class_modifier 2963)) 3324 ((T_PAAMAYIM_NEKUDOTAYIM 7923)) 3325 (("(" 2218) (argument_list 7922)) 3326 ((T_NULLSAFE_OBJECT_OPERATOR 7913) (T_OBJECT_OPERATOR 7914) (T_PAAMAYIM_NEKUDOTAYIM 7915) ("[" 7916)) 3327 (("$" 2202) (T_STRING 2237) (T_VARIABLE 2204) (property_name 7921) (simple_variable 2239) ("{" 2240)) 3328 (("$" 2202) (T_STRING 2237) (T_VARIABLE 2204) (property_name 7920) (simple_variable 2239) ("{" 2240)) 3329 (("$" 2202) (T_VARIABLE 2204) (simple_variable 7919)) 3330 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 7917) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 3331 (("]" 7918)) 3332 (("$" 2202) (T_VARIABLE 2204) (simple_variable 7924)) 3333 (("(" 2984) (argument_list 2985) (ctor_arguments 7927)) 3334 ((T_EXTENDS 2987) (extends_from 7928)) 3335 ((T_IMPLEMENTS 2989) (implements_list 7929)) 3336 ((backup_doc_comment 7930)) 3337 (("{" 7931)) 3338 ((class_statement_list 7932)) 3339 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 7933)) 3340 (("%" 304) (")" 7935) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 3341 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 7937) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 3342 (("%" 304) (")" 7938) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 3343 (("{" 7939)) 3344 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 7940) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 3345 (("}" 7941)) 3346 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 7943) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 3347 ((")" 7944)) 3348 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 7946) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3349 (("," 5226) (possible_comma 7947)) 3350 ((")" 7948)) 3351 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 7956) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 7957) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 7958) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3352 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 7961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3353 (("," 1498) (possible_comma 7959)) 3354 ((")" 7960)) 3355 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 7963) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 3356 (("%" 304) (")" 7964) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 3357 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 7966) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 3358 (("%" 304) (")" 7967) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 3359 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 7974) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 7975) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 7976) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 7977) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 3360 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 7983) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3361 (("%" 2442) (")" 7980) ("*" 2443) ("+" 2444) ("," 7981) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 3362 (("," 5287) (possible_comma 7978)) 3363 ((")" 7979)) 3364 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 7982) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3365 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 7987) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 3366 ((")" 7988)) 3367 (("%" 304) (")" 7980) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 3368 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 7995) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 3369 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 7996)) 3370 (("\"" 7998) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 3371 (("$" 1697) (T_ABSTRACT 8001) (T_ARRAY 8002) (T_AS 8003) (T_BREAK 8004) (T_CALLABLE 8005) (T_CASE 8006) (T_CATCH 8007) (T_CLASS 8008) (T_CLASS_C 8009) (T_CLONE 8010) (T_CONST 8011) (T_CONTINUE 8012) (T_DECLARE 8013) (T_DEFAULT 8014) (T_DIR 8015) (T_DO 8016) (T_ECHO 8017) (T_ELSE 8018) (T_ELSEIF 8019) (T_EMPTY 8020) (T_ENDDECLARE 8021) (T_ENDFOR 8022) (T_ENDFOREACH 8023) (T_ENDIF 8024) (T_ENDSWITCH 8025) (T_ENDWHILE 8026) (T_ENUM 8027) (T_EVAL 8028) (T_EXIT 8029) (T_EXTENDS 8030) (T_FILE 8031) (T_FINAL 8032) (T_FINALLY 8033) (T_FN 8034) (T_FOR 8035) (T_FOREACH 8036) (T_FUNCTION 8037) (T_FUNC_C 8038) (T_GLOBAL 8039) (T_GOTO 8040) (T_IF 8041) (T_IMPLEMENTS 8042) (T_INCLUDE 8043) (T_INCLUDE_ONCE 8044) (T_INSTANCEOF 8045) (T_INSTEADOF 8046) (T_INTERFACE 8047) (T_ISSET 8048) (T_LINE 8049) (T_LIST 8050) (T_LOGICAL_AND 8051) (T_LOGICAL_OR 8052) (T_LOGICAL_XOR 8053) (T_MATCH 8054) (T_METHOD_C 8055) (T_NAMESPACE 8056) (T_NEW 8057) (T_NS_C 8058) (T_PRINT 8059) (T_PRIVATE 8060) (T_PROPERTY_C 8061) (T_PROTECTED 8062) (T_PUBLIC 8063) (T_READONLY 8064) (T_REQUIRE 8065) (T_REQUIRE_ONCE 8066) (T_RETURN 8067) (T_STATIC 8068) (T_STRING 8069) (T_SWITCH 8070) (T_THROW 8071) (T_TRAIT 8072) (T_TRAIT_C 8073) (T_TRY 8074) (T_UNSET 8075) (T_USE 8076) (T_VAR 8077) (T_VARIABLE 1746) (T_WHILE 8078) (T_YIELD 8079) (identifier 8080) (member_name 8081) (reserved_non_modifiers 8082) (semi_reserved 8083) (simple_variable 8084) ("{" 8085)) 3372 (("(" 8088) (argument_list 8089)) 3373 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 8086) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 3374 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 8087)) 3375 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 8090) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 8091) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 8092) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3376 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 8095) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3377 (("," 1498) (possible_comma 8093)) 3378 ((")" 8094)) 3379 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 8125) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 8126) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 3380 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 8124) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 3381 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 8123) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 3382 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 8122) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 3383 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 8121) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 3384 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 8120) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 3385 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 8119) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 3386 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 8118) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 3387 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 8117) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 3388 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 8116) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 3389 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 8115) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 3390 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 8114) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 3391 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 8113) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 3392 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 8112) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 3393 (("\"" 1538) ("$" 8127) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 8128) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 8129) ("[" 1560) (array_object_dereferenceable 8130) (callable_expr 8131) (callable_variable 8132) (class_constant 1564) (class_name 8133) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 8134) (name 8135) (new_dereferenceable 1570) (simple_variable 8136) (static_member 8137) (variable 8138) (variable_class_name 8139)) 3394 (("$" 8127) (T_VARIABLE 8129) (simple_variable 8172) ("{" 8173)) 3395 (("(" 8143) (argument_list 8171)) 3396 ((T_NULLSAFE_OBJECT_OPERATOR 8157) (T_OBJECT_OPERATOR 8158) ("[" 8159)) 3397 (("(" 8143) (argument_list 8156)) 3398 ((T_PAAMAYIM_NEKUDOTAYIM 8152)) 3399 (("(" 8143) (argument_list 8151)) 3400 ((T_PAAMAYIM_NEKUDOTAYIM 8140)) 3401 (("$" 8127) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 8129) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 8141) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 8142) ("{" 1660)) 3402 (("(" 8143) (argument_list 8144)) 3403 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 8145) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 8146) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 8147) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3404 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 8150) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3405 (("," 1498) (possible_comma 8148)) 3406 ((")" 8149)) 3407 (("$" 8127) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 8129) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 8153) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 8154) ("{" 1676)) 3408 (("(" 8143) (argument_list 8155)) 3409 (("$" 8127) (T_STRING 8162) (T_VARIABLE 8129) (property_name 8169) (simple_variable 8164) ("{" 8165)) 3410 (("$" 8127) (T_STRING 8162) (T_VARIABLE 8129) (property_name 8163) (simple_variable 8164) ("{" 8165)) 3411 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 8160) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 3412 (("]" 8161)) 3413 (("(" 8143) (argument_list 8168)) 3414 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 8166) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 3415 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 8167)) 3416 (("(" 8143) (argument_list 8170)) 3417 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 8174) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 3418 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 8175)) 3419 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (possible_array_pair 8177) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 3420 ((backup_doc_comment 8180)) 3421 (("(" 8181)) 3422 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 8182)) 3423 ((")" 8183)) 3424 ((T_USE 166) (lexical_vars 8184)) 3425 ((":" 168) (return_type 8185)) 3426 ((backup_fn_flags 8186)) 3427 (("{" 8187)) 3428 ((inner_statement_list 8188)) 3429 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 8189) ("~" 129)) 3430 ((backup_fn_flags 8190)) 3431 ((backup_doc_comment 8192)) 3432 (("(" 8193)) 3433 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 8194)) 3434 ((")" 8195)) 3435 ((":" 2271) (return_type 8196)) 3436 ((T_DOUBLE_ARROW 8197)) 3437 ((backup_fn_flags 8198)) 3438 ((backup_lex_pos 8199)) 3439 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8263) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3440 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8732) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3441 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 8730) (encaps_var 5357)) 3442 (("$" 8202) (T_VARIABLE 8250) (simple_variable 8726) ("{" 8727)) 3443 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 8725) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 3444 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8724) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3445 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8723) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3446 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8722) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3447 (("(" 8719)) 3448 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8718) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3449 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8717) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3450 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8704) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (clone_argument_list 8705) (constant 8261) (dereferenceable_scalar 8262) (expr 8706) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3451 (("\"" 1538) ("$" 8507) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 8508) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 8509) ("[" 1560) (array_object_dereferenceable 8510) (callable_expr 8511) (callable_variable 8512) (class_constant 1564) (class_name 8513) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 8514) (name 8515) (new_dereferenceable 1570) (simple_variable 8516) (static_member 8517) (variable 8703) (variable_class_name 8519)) 3452 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8702) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3453 (("(" 8699)) 3454 (("(" 8696)) 3455 (("(" 8687) (argument_list 8688) (ctor_arguments 8689)) 3456 (("\"" 1538) ("$" 8507) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 8508) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 8509) ("[" 1560) (array_object_dereferenceable 8510) (callable_expr 8511) (callable_variable 8512) (class_constant 1564) (class_name 8513) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 8514) (name 8515) (new_dereferenceable 1570) (simple_variable 8516) (static_member 8517) (variable 8686) (variable_class_name 8519)) 3457 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8685) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3458 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8684) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3459 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8683) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3460 (("(" 8679)) 3461 (("(" 8674)) 3462 (("(" 8668)) 3463 (("$" 8507) ("(" 8631) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 8632) (T_NAME_QUALIFIED 8633) (T_NAME_RELATIVE 8634) (T_READONLY 2955) (T_STATIC 8635) (T_STRING 8636) (T_VARIABLE 8509) (anonymous_class 8637) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 8638) (attribute 2961) (attributes 8639) (class_modifier 2963) (class_name 8640) (class_name_reference 8641) (name 8642) (new_variable 8643) (simple_variable 8644)) 3464 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8630) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3465 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8629) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3466 (("(" 8468) (argument_list 8628)) 3467 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8627) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3468 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8626) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3469 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 8621) (T_END_HEREDOC 8622) (T_VARIABLE 2920) (encaps_list 8623) (encaps_var 2922)) 3470 ((T_FN 35) (T_FUNCTION 131) (fn 8264) #'8265 (inline_function 8620)) 3471 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8619) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3472 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8618) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3473 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8617) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3474 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8614) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3475 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8613) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3476 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 8609) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 3477 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 8607) (encaps_list 2828) (encaps_var 2829)) 3478 ((T_NULLSAFE_OBJECT_OPERATOR 8593) (T_OBJECT_OPERATOR 8594) ("[" 8595)) 3479 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 8590) (attribute 2757) (fn 8264) #'8265 (inline_function 8591)) 3480 (("(" 8468) (argument_list 8589)) 3481 ((T_PAAMAYIM_NEKUDOTAYIM 8581)) 3482 (("%" 8280) ("*" 8281) ("+" 8282) ("-" 8283) ("." 8284) ("/" 8285) ("<" 8286) (">" 8287) ("?" 8288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 8289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 8290) (T_BOOLEAN_AND 8291) (T_BOOLEAN_OR 8292) (T_COALESCE 8293) (T_INSTANCEOF 8294) (T_IS_EQUAL 8295) (T_IS_GREATER_OR_EQUAL 8296) (T_IS_IDENTICAL 8297) (T_IS_NOT_EQUAL 8298) (T_IS_NOT_IDENTICAL 8299) (T_IS_SMALLER_OR_EQUAL 8300) (T_LOGICAL_AND 8301) (T_LOGICAL_OR 8302) (T_LOGICAL_XOR 8303) (T_PIPE 8304) (T_POW 8305) (T_SL 8306) (T_SPACESHIP 8307) (T_SR 8308) ("^" 8309) (backup_fn_flags 8580) ("|" 8310)) 3483 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 8569)) 3484 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 8557)) 3485 (("(" 8468) (argument_list 8556)) 3486 (("=" 8476) (T_AND_EQUAL 8477) (T_COALESCE_EQUAL 8478) (T_CONCAT_EQUAL 8479) (T_DEC 8480) (T_DIV_EQUAL 8481) (T_INC 8482) (T_MINUS_EQUAL 8483) (T_MOD_EQUAL 8484) (T_MUL_EQUAL 8485) (T_OR_EQUAL 8486) (T_PLUS_EQUAL 8487) (T_POW_EQUAL 8488) (T_SL_EQUAL 8489) (T_SR_EQUAL 8490) (T_XOR_EQUAL 8491)) 3487 ((T_PAAMAYIM_NEKUDOTAYIM 8380)) 3488 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8279) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3489 (("%" 8280) ("*" 8281) ("+" 8282) ("-" 8283) ("." 8284) ("/" 8285) ("<" 8286) (">" 8287) ("?" 8288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 8289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 8290) (T_BOOLEAN_AND 8291) (T_BOOLEAN_OR 8292) (T_COALESCE 8293) (T_INSTANCEOF 8294) (T_IS_EQUAL 8295) (T_IS_GREATER_OR_EQUAL 8296) (T_IS_IDENTICAL 8297) (T_IS_NOT_EQUAL 8298) (T_IS_NOT_IDENTICAL 8299) (T_IS_SMALLER_OR_EQUAL 8300) (T_LOGICAL_AND 8301) (T_LOGICAL_OR 8302) (T_LOGICAL_XOR 8303) (T_PIPE 8304) (T_POW 8305) (T_SL 8306) (T_SPACESHIP 8307) (T_SR 8308) ("^" 8309) ("|" 8310)) 3490 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8379) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3491 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8378) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3492 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8377) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3493 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8376) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3494 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8375) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3495 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8374) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3496 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8373) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3497 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8372) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3498 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 8367) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 8368) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 3499 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8366) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3500 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8365) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3501 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8364) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3502 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8363) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3503 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8362) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3504 (("$" 8327) ("(" 8328) (T_NAME_FULLY_QUALIFIED 8329) (T_NAME_QUALIFIED 8330) (T_NAME_RELATIVE 8331) (T_STATIC 8332) (T_STRING 8333) (T_VARIABLE 8334) (class_name 8335) (class_name_reference 8336) (name 8337) (new_variable 8338) (simple_variable 8339)) 3505 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8326) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3506 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8325) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3507 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8324) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3508 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8323) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3509 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8322) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3510 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8321) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3511 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8320) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3512 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8319) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3513 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8318) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3514 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8317) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3515 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8316) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3516 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8315) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3517 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8314) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3518 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8313) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3519 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8312) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3520 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8311) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3521 (("$" 8327) (T_VARIABLE 8334) (simple_variable 8358) ("{" 8359)) 3522 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 8356) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 3523 ((T_PAAMAYIM_NEKUDOTAYIM 8354)) 3524 ((T_NULLSAFE_OBJECT_OPERATOR 8340) (T_OBJECT_OPERATOR 8341) (T_PAAMAYIM_NEKUDOTAYIM 8342) ("[" 8343)) 3525 (("$" 8327) (T_STRING 8347) (T_VARIABLE 8334) (property_name 8353) (simple_variable 8349) ("{" 8350)) 3526 (("$" 8327) (T_STRING 8347) (T_VARIABLE 8334) (property_name 8348) (simple_variable 8349) ("{" 8350)) 3527 (("$" 8327) (T_VARIABLE 8334) (simple_variable 8346)) 3528 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 8344) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 3529 (("]" 8345)) 3530 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 8351) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 3531 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 8352)) 3532 (("$" 8327) (T_VARIABLE 8334) (simple_variable 8355)) 3533 (("%" 304) (")" 8357) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 3534 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 8360) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 3535 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 8361)) 3536 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8371) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3537 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 8369) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 3538 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8370) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3539 (("$" 8202) (T_ABSTRACT 8381) (T_ARRAY 8382) (T_AS 8383) (T_BREAK 8384) (T_CALLABLE 8385) (T_CASE 8386) (T_CATCH 8387) (T_CLASS 8388) (T_CLASS_C 8389) (T_CLONE 8390) (T_CONST 8391) (T_CONTINUE 8392) (T_DECLARE 8393) (T_DEFAULT 8394) (T_DIR 8395) (T_DO 8396) (T_ECHO 8397) (T_ELSE 8398) (T_ELSEIF 8399) (T_EMPTY 8400) (T_ENDDECLARE 8401) (T_ENDFOR 8402) (T_ENDFOREACH 8403) (T_ENDIF 8404) (T_ENDSWITCH 8405) (T_ENDWHILE 8406) (T_ENUM 8407) (T_EVAL 8408) (T_EXIT 8409) (T_EXTENDS 8410) (T_FILE 8411) (T_FINAL 8412) (T_FINALLY 8413) (T_FN 8414) (T_FOR 8415) (T_FOREACH 8416) (T_FUNCTION 8417) (T_FUNC_C 8418) (T_GLOBAL 8419) (T_GOTO 8420) (T_IF 8421) (T_IMPLEMENTS 8422) (T_INCLUDE 8423) (T_INCLUDE_ONCE 8424) (T_INSTANCEOF 8425) (T_INSTEADOF 8426) (T_INTERFACE 8427) (T_ISSET 8428) (T_LINE 8429) (T_LIST 8430) (T_LOGICAL_AND 8431) (T_LOGICAL_OR 8432) (T_LOGICAL_XOR 8433) (T_MATCH 8434) (T_METHOD_C 8435) (T_NAMESPACE 8436) (T_NEW 8437) (T_NS_C 8438) (T_PRINT 8439) (T_PRIVATE 8440) (T_PROPERTY_C 8441) (T_PROTECTED 8442) (T_PUBLIC 8443) (T_READONLY 8444) (T_REQUIRE 8445) (T_REQUIRE_ONCE 8446) (T_RETURN 8447) (T_STATIC 8448) (T_STRING 8449) (T_SWITCH 8450) (T_THROW 8451) (T_TRAIT 8452) (T_TRAIT_C 8453) (T_TRY 8454) (T_UNSET 8455) (T_USE 8456) (T_VAR 8457) (T_VARIABLE 8250) (T_WHILE 8458) (T_YIELD 8459) (identifier 8460) (member_name 8461) (reserved_non_modifiers 8462) (semi_reserved 8463) (simple_variable 8464) ("{" 8465)) 3540 (("(" 8468) (argument_list 8469)) 3541 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 8466) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 3542 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 8467)) 3543 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 8470) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 8471) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 8472) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3544 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 8475) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3545 (("," 1498) (possible_comma 8473)) 3546 ((")" 8474)) 3547 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (ampersand 8505) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8506) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3548 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8504) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3549 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8503) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3550 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8502) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3551 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8501) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3552 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8500) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3553 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8499) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3554 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8498) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3555 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8497) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3556 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8496) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3557 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8495) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3558 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8494) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3559 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8493) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3560 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8492) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3561 (("\"" 1538) ("$" 8507) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 8508) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 8509) ("[" 1560) (array_object_dereferenceable 8510) (callable_expr 8511) (callable_variable 8512) (class_constant 1564) (class_name 8513) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 8514) (name 8515) (new_dereferenceable 1570) (simple_variable 8516) (static_member 8517) (variable 8518) (variable_class_name 8519)) 3562 (("$" 8507) (T_VARIABLE 8509) (simple_variable 8552) ("{" 8553)) 3563 (("(" 8523) (argument_list 8551)) 3564 ((T_NULLSAFE_OBJECT_OPERATOR 8537) (T_OBJECT_OPERATOR 8538) ("[" 8539)) 3565 (("(" 8523) (argument_list 8536)) 3566 ((T_PAAMAYIM_NEKUDOTAYIM 8532)) 3567 (("(" 8523) (argument_list 8531)) 3568 ((T_PAAMAYIM_NEKUDOTAYIM 8520)) 3569 (("$" 8507) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 8509) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 8521) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 8522) ("{" 1660)) 3570 (("(" 8523) (argument_list 8524)) 3571 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 8525) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 8526) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 8527) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3572 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 8530) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3573 (("," 1498) (possible_comma 8528)) 3574 ((")" 8529)) 3575 (("$" 8507) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 8509) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 8533) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 8534) ("{" 1676)) 3576 (("(" 8523) (argument_list 8535)) 3577 (("$" 8507) (T_STRING 8542) (T_VARIABLE 8509) (property_name 8549) (simple_variable 8544) ("{" 8545)) 3578 (("$" 8507) (T_STRING 8542) (T_VARIABLE 8509) (property_name 8543) (simple_variable 8544) ("{" 8545)) 3579 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 8540) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 3580 (("]" 8541)) 3581 (("(" 8523) (argument_list 8548)) 3582 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 8546) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 3583 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 8547)) 3584 (("(" 8523) (argument_list 8550)) 3585 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 8554) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 3586 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 8555)) 3587 ((backup_doc_comment 8558)) 3588 (("(" 8559)) 3589 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 8560)) 3590 ((")" 8561)) 3591 ((T_USE 166) (lexical_vars 8562)) 3592 ((":" 168) (return_type 8563)) 3593 ((backup_fn_flags 8564)) 3594 (("{" 8565)) 3595 ((inner_statement_list 8566)) 3596 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 8567) ("~" 129)) 3597 ((backup_fn_flags 8568)) 3598 ((backup_doc_comment 8570)) 3599 (("(" 8571)) 3600 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 8572)) 3601 ((")" 8573)) 3602 ((":" 2271) (return_type 8574)) 3603 ((T_DOUBLE_ARROW 8575)) 3604 ((backup_fn_flags 8576)) 3605 ((backup_lex_pos 8577)) 3606 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8578) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3607 (("%" 8280) ("*" 8281) ("+" 8282) ("-" 8283) ("." 8284) ("/" 8285) ("<" 8286) (">" 8287) ("?" 8288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 8289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 8290) (T_BOOLEAN_AND 8291) (T_BOOLEAN_OR 8292) (T_COALESCE 8293) (T_INSTANCEOF 8294) (T_IS_EQUAL 8295) (T_IS_GREATER_OR_EQUAL 8296) (T_IS_IDENTICAL 8297) (T_IS_NOT_EQUAL 8298) (T_IS_NOT_IDENTICAL 8299) (T_IS_SMALLER_OR_EQUAL 8300) (T_LOGICAL_AND 8301) (T_LOGICAL_OR 8302) (T_LOGICAL_XOR 8303) (T_PIPE 8304) (T_POW 8305) (T_SL 8306) (T_SPACESHIP 8307) (T_SR 8308) ("^" 8309) (backup_fn_flags 8579) ("|" 8310)) 3608 (("$" 8202) (T_ABSTRACT 8381) (T_ARRAY 8382) (T_AS 8383) (T_BREAK 8384) (T_CALLABLE 8385) (T_CASE 8386) (T_CATCH 8387) (T_CLASS 8388) (T_CLASS_C 8389) (T_CLONE 8390) (T_CONST 8391) (T_CONTINUE 8392) (T_DECLARE 8393) (T_DEFAULT 8394) (T_DIR 8395) (T_DO 8396) (T_ECHO 8397) (T_ELSE 8398) (T_ELSEIF 8399) (T_EMPTY 8400) (T_ENDDECLARE 8401) (T_ENDFOR 8402) (T_ENDFOREACH 8403) (T_ENDIF 8404) (T_ENDSWITCH 8405) (T_ENDWHILE 8406) (T_ENUM 8407) (T_EVAL 8408) (T_EXIT 8409) (T_EXTENDS 8410) (T_FILE 8411) (T_FINAL 8412) (T_FINALLY 8413) (T_FN 8414) (T_FOR 8415) (T_FOREACH 8416) (T_FUNCTION 8417) (T_FUNC_C 8418) (T_GLOBAL 8419) (T_GOTO 8420) (T_IF 8421) (T_IMPLEMENTS 8422) (T_INCLUDE 8423) (T_INCLUDE_ONCE 8424) (T_INSTANCEOF 8425) (T_INSTEADOF 8426) (T_INTERFACE 8427) (T_ISSET 8428) (T_LINE 8429) (T_LIST 8430) (T_LOGICAL_AND 8431) (T_LOGICAL_OR 8432) (T_LOGICAL_XOR 8433) (T_MATCH 8434) (T_METHOD_C 8435) (T_NAMESPACE 8436) (T_NEW 8437) (T_NS_C 8438) (T_PRINT 8439) (T_PRIVATE 8440) (T_PROPERTY_C 8441) (T_PROTECTED 8442) (T_PUBLIC 8443) (T_READONLY 8444) (T_REQUIRE 8445) (T_REQUIRE_ONCE 8446) (T_RETURN 8447) (T_STATIC 8448) (T_STRING 8449) (T_SWITCH 8450) (T_THROW 8451) (T_TRAIT 8452) (T_TRAIT_C 8453) (T_TRY 8454) (T_UNSET 8455) (T_USE 8456) (T_VAR 8457) (T_VARIABLE 8250) (T_WHILE 8458) (T_YIELD 8459) (identifier 8582) (member_name 8583) (reserved_non_modifiers 8462) (semi_reserved 8463) (simple_variable 8584) ("{" 8585)) 3609 (("(" 8468) (argument_list 8588)) 3610 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 8586) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 3611 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 8587)) 3612 ((T_FN 35) (T_FUNCTION 131) (fn 8264) #'8265 (inline_function 8592)) 3613 (("$" 8202) (T_STRING 8598) (T_VARIABLE 8250) (property_name 8605) (simple_variable 8600) ("{" 8601)) 3614 (("$" 8202) (T_STRING 8598) (T_VARIABLE 8250) (property_name 8599) (simple_variable 8600) ("{" 8601)) 3615 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 8596) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 3616 (("]" 8597)) 3617 (("(" 8468) (argument_list 8604)) 3618 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 8602) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 3619 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 8603)) 3620 (("(" 8468) (argument_list 8606)) 3621 (("`" 8608)) 3622 (("]" 8610)) 3623 (("=" 8611)) 3624 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8612) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3625 (("%" 8280) ("*" 8281) ("+" 8282) ("-" 8283) ("." 8284) ("/" 8285) ("<" 8286) (">" 8287) ("?" 8288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 8289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 8290) (T_BOOLEAN_AND 8291) (T_BOOLEAN_OR 8292) (T_COALESCE 8293) (T_DOUBLE_ARROW 8615) (T_INSTANCEOF 8294) (T_IS_EQUAL 8295) (T_IS_GREATER_OR_EQUAL 8296) (T_IS_IDENTICAL 8297) (T_IS_NOT_EQUAL 8298) (T_IS_NOT_IDENTICAL 8299) (T_IS_SMALLER_OR_EQUAL 8300) (T_LOGICAL_AND 8301) (T_LOGICAL_OR 8302) (T_LOGICAL_XOR 8303) (T_PIPE 8304) (T_POW 8305) (T_SL 8306) (T_SPACESHIP 8307) (T_SR 8308) ("^" 8309) ("|" 8310)) 3626 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8616) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3627 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 8625) (T_VARIABLE 2920) (encaps_var 2934)) 3628 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 8624) (T_VARIABLE 2920) (encaps_var 2925)) 3629 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 8666) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 3630 ((T_CLASS 8658)) 3631 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 8657) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 8638) (attribute 2982) (class_modifier 2963)) 3632 ((T_PAAMAYIM_NEKUDOTAYIM 8655)) 3633 (("(" 8523) (argument_list 8654)) 3634 ((T_NULLSAFE_OBJECT_OPERATOR 8645) (T_OBJECT_OPERATOR 8646) (T_PAAMAYIM_NEKUDOTAYIM 8647) ("[" 8648)) 3635 (("$" 8507) (T_STRING 8542) (T_VARIABLE 8509) (property_name 8653) (simple_variable 8544) ("{" 8545)) 3636 (("$" 8507) (T_STRING 8542) (T_VARIABLE 8509) (property_name 8652) (simple_variable 8544) ("{" 8545)) 3637 (("$" 8507) (T_VARIABLE 8509) (simple_variable 8651)) 3638 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 8649) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 3639 (("]" 8650)) 3640 (("$" 8507) (T_VARIABLE 8509) (simple_variable 8656)) 3641 (("(" 2984) (argument_list 2985) (ctor_arguments 8659)) 3642 ((T_EXTENDS 2987) (extends_from 8660)) 3643 ((T_IMPLEMENTS 2989) (implements_list 8661)) 3644 ((backup_doc_comment 8662)) 3645 (("{" 8663)) 3646 ((class_statement_list 8664)) 3647 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 8665)) 3648 (("%" 304) (")" 8667) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 3649 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 8669) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 3650 (("%" 304) (")" 8670) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 3651 (("{" 8671)) 3652 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 8672) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 3653 (("}" 8673)) 3654 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 8675) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 3655 ((")" 8676)) 3656 (("=" 8677)) 3657 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 8678) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3658 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 8680) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3659 (("," 5226) (possible_comma 8681)) 3660 ((")" 8682)) 3661 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 8690) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 8691) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 8692) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3662 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 8695) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3663 (("," 1498) (possible_comma 8693)) 3664 ((")" 8694)) 3665 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 8697) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 3666 (("%" 304) (")" 8698) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 3667 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 8700) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 3668 (("%" 304) (")" 8701) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 3669 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 8707) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 8708) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 8709) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 8710) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 3670 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 8716) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3671 (("%" 2442) (")" 8713) ("*" 2443) ("+" 2444) ("," 8714) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 3672 (("," 5287) (possible_comma 8711)) 3673 ((")" 8712)) 3674 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 8715) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3675 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 8720) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 3676 ((")" 8721)) 3677 (("%" 304) (")" 8713) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 3678 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 8728) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 3679 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 8729)) 3680 (("\"" 8731) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 3681 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8762) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (ampersand 8789) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 8798) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3682 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 9850) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3683 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 9848) (encaps_var 5357)) 3684 (("$" 8736) (T_VARIABLE 8784) (simple_variable 9844) ("{" 9845)) 3685 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 9843) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 3686 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 9842) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3687 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 9841) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3688 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 9840) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3689 (("(" 9837)) 3690 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 9836) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3691 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 9835) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3692 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 9822) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (clone_argument_list 9823) (constant 8796) (dereferenceable_scalar 8797) (expr 9824) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3693 (("\"" 1538) ("$" 9048) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 9049) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 9050) ("[" 1560) (array_object_dereferenceable 9051) (callable_expr 9052) (callable_variable 9053) (class_constant 1564) (class_name 9054) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 9055) (name 9056) (new_dereferenceable 1570) (simple_variable 9057) (static_member 9058) (variable 9821) (variable_class_name 9060)) 3694 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 9820) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3695 (("(" 9817)) 3696 (("(" 9814)) 3697 (("(" 9805) (argument_list 9806) (ctor_arguments 9807)) 3698 (("\"" 1538) ("$" 9048) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 9049) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 9050) ("[" 1560) (array_object_dereferenceable 9051) (callable_expr 9052) (callable_variable 9053) (class_constant 1564) (class_name 9054) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 9055) (name 9056) (new_dereferenceable 1570) (simple_variable 9057) (static_member 9058) (variable 9804) (variable_class_name 9060)) 3699 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 9803) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3700 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 9802) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3701 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 9801) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3702 (("(" 9797)) 3703 (("(" 9794)) 3704 (("(" 9788)) 3705 (("$" 9048) ("(" 9751) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 9752) (T_NAME_QUALIFIED 9753) (T_NAME_RELATIVE 9754) (T_READONLY 2955) (T_STATIC 9755) (T_STRING 9756) (T_VARIABLE 9050) (anonymous_class 9757) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 9758) (attribute 2961) (attributes 9759) (class_modifier 2963) (class_name 9760) (class_name_reference 9761) (name 9762) (new_variable 9763) (simple_variable 9764)) 3706 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 9750) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3707 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 9749) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3708 (("(" 9009) (argument_list 9748)) 3709 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 9747) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3710 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 9746) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3711 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 9741) (T_END_HEREDOC 9742) (T_VARIABLE 2920) (encaps_list 9743) (encaps_var 2922)) 3712 ((T_FN 35) (T_FUNCTION 131) (fn 8799) #'8800 (inline_function 9740)) 3713 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 9739) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3714 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 9738) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3715 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 9737) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3716 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 9734) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 3717 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 9733) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3718 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 9729) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 3719 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 9727) (encaps_list 2828) (encaps_var 2829)) 3720 (("\"" 1538) ("$" 9678) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 9679) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 9680) ("[" 1560) (array_object_dereferenceable 9681) (callable_expr 9682) (callable_variable 9683) (class_constant 1564) (class_name 9684) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 9685) (name 9686) (new_dereferenceable 1570) (simple_variable 9687) (static_member 9688) (variable 9689) (variable_class_name 9690)) 3721 ((T_NULLSAFE_OBJECT_OPERATOR 9664) (T_OBJECT_OPERATOR 9665) ("[" 9666)) 3722 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 9661) (attribute 2757) (fn 8799) #'8800 (inline_function 9662)) 3723 (("(" 9009) (argument_list 9660)) 3724 ((T_PAAMAYIM_NEKUDOTAYIM 9652)) 3725 (("%" 8816) ("*" 8817) ("+" 8818) ("-" 8819) ("." 8820) ("/" 8821) ("<" 8822) (">" 8823) ("?" 8824) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 8825) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 8826) (T_BOOLEAN_AND 8827) (T_BOOLEAN_OR 8828) (T_COALESCE 8829) (T_INSTANCEOF 8830) (T_IS_EQUAL 8831) (T_IS_GREATER_OR_EQUAL 8832) (T_IS_IDENTICAL 8833) (T_IS_NOT_EQUAL 8834) (T_IS_NOT_IDENTICAL 8835) (T_IS_SMALLER_OR_EQUAL 8836) (T_LOGICAL_AND 8837) (T_LOGICAL_OR 8838) (T_LOGICAL_XOR 8839) (T_PIPE 8840) (T_POW 8841) (T_SL 8842) (T_SPACESHIP 8843) (T_SR 8844) ("^" 8845) ("|" 8846)) 3726 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 9110)) 3727 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 9098)) 3728 (("(" 9009) (argument_list 9097)) 3729 (("=" 9017) (T_AND_EQUAL 9018) (T_COALESCE_EQUAL 9019) (T_CONCAT_EQUAL 9020) (T_DEC 9021) (T_DIV_EQUAL 9022) (T_INC 9023) (T_MINUS_EQUAL 9024) (T_MOD_EQUAL 9025) (T_MUL_EQUAL 9026) (T_OR_EQUAL 9027) (T_PLUS_EQUAL 9028) (T_POW_EQUAL 9029) (T_SL_EQUAL 9030) (T_SR_EQUAL 9031) (T_XOR_EQUAL 9032)) 3730 ((T_PAAMAYIM_NEKUDOTAYIM 8921)) 3731 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 8815) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3732 (("(" 8916)) 3733 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 8915) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3734 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 8914) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3735 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 8913) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3736 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 8912) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3737 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 8911) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3738 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 8910) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3739 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 8909) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3740 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 8908) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3741 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 8903) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 8904) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 3742 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 8902) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3743 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 8901) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3744 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 8900) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3745 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 8899) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3746 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 8898) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3747 (("$" 8863) ("(" 8864) (T_NAME_FULLY_QUALIFIED 8865) (T_NAME_QUALIFIED 8866) (T_NAME_RELATIVE 8867) (T_STATIC 8868) (T_STRING 8869) (T_VARIABLE 8870) (class_name 8871) (class_name_reference 8872) (name 8873) (new_variable 8874) (simple_variable 8875)) 3748 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 8862) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3749 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 8861) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3750 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 8860) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3751 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 8859) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3752 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 8858) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3753 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 8857) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3754 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 8856) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3755 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 8855) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3756 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 8854) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3757 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 8853) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3758 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 8852) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3759 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 8851) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3760 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 8850) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3761 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 8849) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3762 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 8848) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3763 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 8847) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3764 (("$" 8863) (T_VARIABLE 8870) (simple_variable 8894) ("{" 8895)) 3765 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 8892) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 3766 ((T_PAAMAYIM_NEKUDOTAYIM 8890)) 3767 ((T_NULLSAFE_OBJECT_OPERATOR 8876) (T_OBJECT_OPERATOR 8877) (T_PAAMAYIM_NEKUDOTAYIM 8878) ("[" 8879)) 3768 (("$" 8863) (T_STRING 8883) (T_VARIABLE 8870) (property_name 8889) (simple_variable 8885) ("{" 8886)) 3769 (("$" 8863) (T_STRING 8883) (T_VARIABLE 8870) (property_name 8884) (simple_variable 8885) ("{" 8886)) 3770 (("$" 8863) (T_VARIABLE 8870) (simple_variable 8882)) 3771 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 8880) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 3772 (("]" 8881)) 3773 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 8887) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 3774 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 8888)) 3775 (("$" 8863) (T_VARIABLE 8870) (simple_variable 8891)) 3776 (("%" 304) (")" 8893) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 3777 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 8896) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 3778 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 8897)) 3779 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 8907) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3780 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 8905) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 3781 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 8906) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3782 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 8917) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 3783 ((")" 8918)) 3784 (("=" 8919)) 3785 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 8920) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3786 (("$" 8736) (T_ABSTRACT 8922) (T_ARRAY 8923) (T_AS 8924) (T_BREAK 8925) (T_CALLABLE 8926) (T_CASE 8927) (T_CATCH 8928) (T_CLASS 8929) (T_CLASS_C 8930) (T_CLONE 8931) (T_CONST 8932) (T_CONTINUE 8933) (T_DECLARE 8934) (T_DEFAULT 8935) (T_DIR 8936) (T_DO 8937) (T_ECHO 8938) (T_ELSE 8939) (T_ELSEIF 8940) (T_EMPTY 8941) (T_ENDDECLARE 8942) (T_ENDFOR 8943) (T_ENDFOREACH 8944) (T_ENDIF 8945) (T_ENDSWITCH 8946) (T_ENDWHILE 8947) (T_ENUM 8948) (T_EVAL 8949) (T_EXIT 8950) (T_EXTENDS 8951) (T_FILE 8952) (T_FINAL 8953) (T_FINALLY 8954) (T_FN 8955) (T_FOR 8956) (T_FOREACH 8957) (T_FUNCTION 8958) (T_FUNC_C 8959) (T_GLOBAL 8960) (T_GOTO 8961) (T_IF 8962) (T_IMPLEMENTS 8963) (T_INCLUDE 8964) (T_INCLUDE_ONCE 8965) (T_INSTANCEOF 8966) (T_INSTEADOF 8967) (T_INTERFACE 8968) (T_ISSET 8969) (T_LINE 8970) (T_LIST 8971) (T_LOGICAL_AND 8972) (T_LOGICAL_OR 8973) (T_LOGICAL_XOR 8974) (T_MATCH 8975) (T_METHOD_C 8976) (T_NAMESPACE 8977) (T_NEW 8978) (T_NS_C 8979) (T_PRINT 8980) (T_PRIVATE 8981) (T_PROPERTY_C 8982) (T_PROTECTED 8983) (T_PUBLIC 8984) (T_READONLY 8985) (T_REQUIRE 8986) (T_REQUIRE_ONCE 8987) (T_RETURN 8988) (T_STATIC 8989) (T_STRING 8990) (T_SWITCH 8991) (T_THROW 8992) (T_TRAIT 8993) (T_TRAIT_C 8994) (T_TRY 8995) (T_UNSET 8996) (T_USE 8997) (T_VAR 8998) (T_VARIABLE 8784) (T_WHILE 8999) (T_YIELD 9000) (identifier 9001) (member_name 9002) (reserved_non_modifiers 9003) (semi_reserved 9004) (simple_variable 9005) ("{" 9006)) 3787 (("(" 9009) (argument_list 9010)) 3788 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 9007) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 3789 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 9008)) 3790 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 9011) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 9012) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 9013) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3791 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 9016) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3792 (("," 1498) (possible_comma 9014)) 3793 ((")" 9015)) 3794 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (ampersand 9046) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 9047) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3795 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 9045) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3796 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 9044) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3797 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 9043) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3798 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 9042) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3799 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 9041) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3800 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 9040) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3801 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 9039) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3802 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 9038) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3803 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 9037) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3804 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 9036) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3805 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 9035) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3806 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 9034) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3807 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 9033) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 3808 (("\"" 1538) ("$" 9048) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 9049) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 9050) ("[" 1560) (array_object_dereferenceable 9051) (callable_expr 9052) (callable_variable 9053) (class_constant 1564) (class_name 9054) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 9055) (name 9056) (new_dereferenceable 1570) (simple_variable 9057) (static_member 9058) (variable 9059) (variable_class_name 9060)) 3809 (("$" 9048) (T_VARIABLE 9050) (simple_variable 9093) ("{" 9094)) 3810 (("(" 9064) (argument_list 9092)) 3811 ((T_NULLSAFE_OBJECT_OPERATOR 9078) (T_OBJECT_OPERATOR 9079) ("[" 9080)) 3812 (("(" 9064) (argument_list 9077)) 3813 ((T_PAAMAYIM_NEKUDOTAYIM 9073)) 3814 (("(" 9064) (argument_list 9072)) 3815 ((T_PAAMAYIM_NEKUDOTAYIM 9061)) 3816 (("$" 9048) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 9050) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 9062) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 9063) ("{" 1660)) 3817 (("(" 9064) (argument_list 9065)) 3818 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 9066) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 9067) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 9068) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3819 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 9071) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3820 (("," 1498) (possible_comma 9069)) 3821 ((")" 9070)) 3822 (("$" 9048) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 9050) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 9074) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 9075) ("{" 1676)) 3823 (("(" 9064) (argument_list 9076)) 3824 (("$" 9048) (T_STRING 9083) (T_VARIABLE 9050) (property_name 9090) (simple_variable 9085) ("{" 9086)) 3825 (("$" 9048) (T_STRING 9083) (T_VARIABLE 9050) (property_name 9084) (simple_variable 9085) ("{" 9086)) 3826 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 9081) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 3827 (("]" 9082)) 3828 (("(" 9064) (argument_list 9089)) 3829 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 9087) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 3830 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 9088)) 3831 (("(" 9064) (argument_list 9091)) 3832 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 9095) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 3833 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 9096)) 3834 ((backup_doc_comment 9099)) 3835 (("(" 9100)) 3836 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 9101)) 3837 ((")" 9102)) 3838 ((T_USE 166) (lexical_vars 9103)) 3839 ((":" 168) (return_type 9104)) 3840 ((backup_fn_flags 9105)) 3841 (("{" 9106)) 3842 ((inner_statement_list 9107)) 3843 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 9108) ("~" 129)) 3844 ((backup_fn_flags 9109)) 3845 ((backup_doc_comment 9111)) 3846 (("(" 9112)) 3847 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 9113)) 3848 ((")" 9114)) 3849 ((":" 2271) (return_type 9115)) 3850 ((T_DOUBLE_ARROW 9116)) 3851 ((backup_fn_flags 9117)) 3852 ((backup_lex_pos 9118)) 3853 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9182) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3854 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9651) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3855 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 9649) (encaps_var 5357)) 3856 (("$" 9121) (T_VARIABLE 9169) (simple_variable 9645) ("{" 9646)) 3857 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 9644) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 3858 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9643) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3859 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9642) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3860 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9641) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3861 (("(" 9638)) 3862 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9637) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3863 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9636) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3864 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9623) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (clone_argument_list 9624) (constant 9180) (dereferenceable_scalar 9181) (expr 9625) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3865 (("\"" 1538) ("$" 9426) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 9427) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 9428) ("[" 1560) (array_object_dereferenceable 9429) (callable_expr 9430) (callable_variable 9431) (class_constant 1564) (class_name 9432) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 9433) (name 9434) (new_dereferenceable 1570) (simple_variable 9435) (static_member 9436) (variable 9622) (variable_class_name 9438)) 3866 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9621) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3867 (("(" 9618)) 3868 (("(" 9615)) 3869 (("(" 9606) (argument_list 9607) (ctor_arguments 9608)) 3870 (("\"" 1538) ("$" 9426) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 9427) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 9428) ("[" 1560) (array_object_dereferenceable 9429) (callable_expr 9430) (callable_variable 9431) (class_constant 1564) (class_name 9432) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 9433) (name 9434) (new_dereferenceable 1570) (simple_variable 9435) (static_member 9436) (variable 9605) (variable_class_name 9438)) 3871 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9604) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3872 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9603) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3873 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9602) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3874 (("(" 9598)) 3875 (("(" 9593)) 3876 (("(" 9587)) 3877 (("$" 9426) ("(" 9550) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 9551) (T_NAME_QUALIFIED 9552) (T_NAME_RELATIVE 9553) (T_READONLY 2955) (T_STATIC 9554) (T_STRING 9555) (T_VARIABLE 9428) (anonymous_class 9556) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 9557) (attribute 2961) (attributes 9558) (class_modifier 2963) (class_name 9559) (class_name_reference 9560) (name 9561) (new_variable 9562) (simple_variable 9563)) 3878 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9549) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3879 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9548) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3880 (("(" 9387) (argument_list 9547)) 3881 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9546) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3882 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9545) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3883 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 9540) (T_END_HEREDOC 9541) (T_VARIABLE 2920) (encaps_list 9542) (encaps_var 2922)) 3884 ((T_FN 35) (T_FUNCTION 131) (fn 9183) #'9184 (inline_function 9539)) 3885 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9538) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3886 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9537) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3887 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9536) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3888 (("!" 8200) ("\"" 8201) ("$" 8202) ("(" 8203) ("+" 8204) ("-" 8205) ("@" 8206) (T_ARRAY 8207) (T_ARRAY_CAST 8208) (T_ATTRIBUTE 130) (T_BOOL_CAST 8209) (T_CLASS_C 8210) (T_CLONE 8211) (T_CONSTANT_ENCAPSED_STRING 8212) (T_DEC 8213) (T_DIR 8214) (T_DNUMBER 8215) (T_DOUBLE_CAST 8216) (T_EMPTY 8217) (T_EVAL 8218) (T_EXIT 8219) (T_FILE 8220) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8221) (T_INC 8222) (T_INCLUDE 8223) (T_INCLUDE_ONCE 8224) (T_INT_CAST 8225) (T_ISSET 8226) (T_LINE 8227) (T_LIST 8228) (T_LNUMBER 8229) (T_MATCH 8230) (T_METHOD_C 8231) (T_NAME_FULLY_QUALIFIED 8232) (T_NAME_QUALIFIED 8233) (T_NAME_RELATIVE 8234) (T_NEW 8235) (T_NS_C 8236) (T_OBJECT_CAST 8237) (T_PRINT 8238) (T_PROPERTY_C 8239) (T_READONLY 8240) (T_REQUIRE 8241) (T_REQUIRE_ONCE 8242) (T_START_HEREDOC 8243) (T_STATIC 8244) (T_STRING 8245) (T_STRING_CAST 8246) (T_THROW 8247) (T_TRAIT_C 8248) (T_UNSET_CAST 8249) (T_VARIABLE 8250) (T_YIELD 8251) (T_YIELD_FROM 8252) ("[" 8253) ("`" 8254) (array_object_dereferenceable 8255) (attribute 135) (attributes 8256) (callable_expr 8257) (callable_variable 8258) (class_constant 8259) (class_name 8260) (constant 8261) (dereferenceable_scalar 8262) (expr 9533) (fn 8264) (fully_dereferenceable 107) #'8265 (function_call 8266) (inline_function 8267) (internal_functions_in_yacc 8268) (match 8269) (name 8270) (new_dereferenceable 8271) (new_non_dereferenceable 8272) (scalar 8273) (simple_variable 8274) (static_member 8275) (variable 8276) (variable_class_name 8277) ("~" 8278)) 3889 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9532) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3890 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 9528) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 3891 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 9526) (encaps_list 2828) (encaps_var 2829)) 3892 ((T_NULLSAFE_OBJECT_OPERATOR 9512) (T_OBJECT_OPERATOR 9513) ("[" 9514)) 3893 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 9509) (attribute 2757) (fn 9183) #'9184 (inline_function 9510)) 3894 (("(" 9387) (argument_list 9508)) 3895 ((T_PAAMAYIM_NEKUDOTAYIM 9500)) 3896 (("%" 9199) ("*" 9200) ("+" 9201) ("-" 9202) ("." 9203) ("/" 9204) ("<" 9205) (">" 9206) ("?" 9207) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 9208) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 9209) (T_BOOLEAN_AND 9210) (T_BOOLEAN_OR 9211) (T_COALESCE 9212) (T_INSTANCEOF 9213) (T_IS_EQUAL 9214) (T_IS_GREATER_OR_EQUAL 9215) (T_IS_IDENTICAL 9216) (T_IS_NOT_EQUAL 9217) (T_IS_NOT_IDENTICAL 9218) (T_IS_SMALLER_OR_EQUAL 9219) (T_LOGICAL_AND 9220) (T_LOGICAL_OR 9221) (T_LOGICAL_XOR 9222) (T_PIPE 9223) (T_POW 9224) (T_SL 9225) (T_SPACESHIP 9226) (T_SR 9227) ("^" 9228) (backup_fn_flags 9499) ("|" 9229)) 3897 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 9488)) 3898 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 9476)) 3899 (("(" 9387) (argument_list 9475)) 3900 (("=" 9395) (T_AND_EQUAL 9396) (T_COALESCE_EQUAL 9397) (T_CONCAT_EQUAL 9398) (T_DEC 9399) (T_DIV_EQUAL 9400) (T_INC 9401) (T_MINUS_EQUAL 9402) (T_MOD_EQUAL 9403) (T_MUL_EQUAL 9404) (T_OR_EQUAL 9405) (T_PLUS_EQUAL 9406) (T_POW_EQUAL 9407) (T_SL_EQUAL 9408) (T_SR_EQUAL 9409) (T_XOR_EQUAL 9410)) 3901 ((T_PAAMAYIM_NEKUDOTAYIM 9299)) 3902 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9198) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3903 (("%" 9199) ("*" 9200) ("+" 9201) ("-" 9202) ("." 9203) ("/" 9204) ("<" 9205) (">" 9206) ("?" 9207) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 9208) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 9209) (T_BOOLEAN_AND 9210) (T_BOOLEAN_OR 9211) (T_COALESCE 9212) (T_INSTANCEOF 9213) (T_IS_EQUAL 9214) (T_IS_GREATER_OR_EQUAL 9215) (T_IS_IDENTICAL 9216) (T_IS_NOT_EQUAL 9217) (T_IS_NOT_IDENTICAL 9218) (T_IS_SMALLER_OR_EQUAL 9219) (T_LOGICAL_AND 9220) (T_LOGICAL_OR 9221) (T_LOGICAL_XOR 9222) (T_PIPE 9223) (T_POW 9224) (T_SL 9225) (T_SPACESHIP 9226) (T_SR 9227) ("^" 9228) ("|" 9229)) 3904 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9298) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3905 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9297) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3906 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9296) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3907 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9295) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3908 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9294) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3909 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9293) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3910 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9292) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3911 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9291) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3912 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 9286) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 9287) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 3913 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9285) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3914 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9284) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3915 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9283) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3916 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9282) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3917 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9281) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3918 (("$" 9246) ("(" 9247) (T_NAME_FULLY_QUALIFIED 9248) (T_NAME_QUALIFIED 9249) (T_NAME_RELATIVE 9250) (T_STATIC 9251) (T_STRING 9252) (T_VARIABLE 9253) (class_name 9254) (class_name_reference 9255) (name 9256) (new_variable 9257) (simple_variable 9258)) 3919 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9245) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3920 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9244) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3921 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9243) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3922 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9242) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3923 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9241) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3924 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9240) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3925 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9239) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3926 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9238) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3927 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9237) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3928 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9236) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3929 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9235) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3930 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9234) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3931 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9233) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3932 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9232) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3933 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9231) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3934 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9230) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3935 (("$" 9246) (T_VARIABLE 9253) (simple_variable 9277) ("{" 9278)) 3936 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 9275) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 3937 ((T_PAAMAYIM_NEKUDOTAYIM 9273)) 3938 ((T_NULLSAFE_OBJECT_OPERATOR 9259) (T_OBJECT_OPERATOR 9260) (T_PAAMAYIM_NEKUDOTAYIM 9261) ("[" 9262)) 3939 (("$" 9246) (T_STRING 9266) (T_VARIABLE 9253) (property_name 9272) (simple_variable 9268) ("{" 9269)) 3940 (("$" 9246) (T_STRING 9266) (T_VARIABLE 9253) (property_name 9267) (simple_variable 9268) ("{" 9269)) 3941 (("$" 9246) (T_VARIABLE 9253) (simple_variable 9265)) 3942 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 9263) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 3943 (("]" 9264)) 3944 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 9270) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 3945 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 9271)) 3946 (("$" 9246) (T_VARIABLE 9253) (simple_variable 9274)) 3947 (("%" 304) (")" 9276) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 3948 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 9279) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 3949 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 9280)) 3950 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9290) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3951 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 9288) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 3952 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9289) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3953 (("$" 9121) (T_ABSTRACT 9300) (T_ARRAY 9301) (T_AS 9302) (T_BREAK 9303) (T_CALLABLE 9304) (T_CASE 9305) (T_CATCH 9306) (T_CLASS 9307) (T_CLASS_C 9308) (T_CLONE 9309) (T_CONST 9310) (T_CONTINUE 9311) (T_DECLARE 9312) (T_DEFAULT 9313) (T_DIR 9314) (T_DO 9315) (T_ECHO 9316) (T_ELSE 9317) (T_ELSEIF 9318) (T_EMPTY 9319) (T_ENDDECLARE 9320) (T_ENDFOR 9321) (T_ENDFOREACH 9322) (T_ENDIF 9323) (T_ENDSWITCH 9324) (T_ENDWHILE 9325) (T_ENUM 9326) (T_EVAL 9327) (T_EXIT 9328) (T_EXTENDS 9329) (T_FILE 9330) (T_FINAL 9331) (T_FINALLY 9332) (T_FN 9333) (T_FOR 9334) (T_FOREACH 9335) (T_FUNCTION 9336) (T_FUNC_C 9337) (T_GLOBAL 9338) (T_GOTO 9339) (T_IF 9340) (T_IMPLEMENTS 9341) (T_INCLUDE 9342) (T_INCLUDE_ONCE 9343) (T_INSTANCEOF 9344) (T_INSTEADOF 9345) (T_INTERFACE 9346) (T_ISSET 9347) (T_LINE 9348) (T_LIST 9349) (T_LOGICAL_AND 9350) (T_LOGICAL_OR 9351) (T_LOGICAL_XOR 9352) (T_MATCH 9353) (T_METHOD_C 9354) (T_NAMESPACE 9355) (T_NEW 9356) (T_NS_C 9357) (T_PRINT 9358) (T_PRIVATE 9359) (T_PROPERTY_C 9360) (T_PROTECTED 9361) (T_PUBLIC 9362) (T_READONLY 9363) (T_REQUIRE 9364) (T_REQUIRE_ONCE 9365) (T_RETURN 9366) (T_STATIC 9367) (T_STRING 9368) (T_SWITCH 9369) (T_THROW 9370) (T_TRAIT 9371) (T_TRAIT_C 9372) (T_TRY 9373) (T_UNSET 9374) (T_USE 9375) (T_VAR 9376) (T_VARIABLE 9169) (T_WHILE 9377) (T_YIELD 9378) (identifier 9379) (member_name 9380) (reserved_non_modifiers 9381) (semi_reserved 9382) (simple_variable 9383) ("{" 9384)) 3954 (("(" 9387) (argument_list 9388)) 3955 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 9385) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 3956 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 9386)) 3957 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 9389) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 9390) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 9391) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3958 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 9394) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3959 (("," 1498) (possible_comma 9392)) 3960 ((")" 9393)) 3961 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (ampersand 9424) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9425) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3962 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9423) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3963 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9422) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3964 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9421) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3965 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9420) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3966 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9419) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3967 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9418) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3968 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9417) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3969 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9416) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3970 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9415) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3971 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9414) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3972 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9413) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3973 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9412) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3974 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9411) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 3975 (("\"" 1538) ("$" 9426) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 9427) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 9428) ("[" 1560) (array_object_dereferenceable 9429) (callable_expr 9430) (callable_variable 9431) (class_constant 1564) (class_name 9432) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 9433) (name 9434) (new_dereferenceable 1570) (simple_variable 9435) (static_member 9436) (variable 9437) (variable_class_name 9438)) 3976 (("$" 9426) (T_VARIABLE 9428) (simple_variable 9471) ("{" 9472)) 3977 (("(" 9442) (argument_list 9470)) 3978 ((T_NULLSAFE_OBJECT_OPERATOR 9456) (T_OBJECT_OPERATOR 9457) ("[" 9458)) 3979 (("(" 9442) (argument_list 9455)) 3980 ((T_PAAMAYIM_NEKUDOTAYIM 9451)) 3981 (("(" 9442) (argument_list 9450)) 3982 ((T_PAAMAYIM_NEKUDOTAYIM 9439)) 3983 (("$" 9426) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 9428) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 9440) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 9441) ("{" 1660)) 3984 (("(" 9442) (argument_list 9443)) 3985 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 9444) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 9445) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 9446) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3986 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 9449) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 3987 (("," 1498) (possible_comma 9447)) 3988 ((")" 9448)) 3989 (("$" 9426) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 9428) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 9452) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 9453) ("{" 1676)) 3990 (("(" 9442) (argument_list 9454)) 3991 (("$" 9426) (T_STRING 9461) (T_VARIABLE 9428) (property_name 9468) (simple_variable 9463) ("{" 9464)) 3992 (("$" 9426) (T_STRING 9461) (T_VARIABLE 9428) (property_name 9462) (simple_variable 9463) ("{" 9464)) 3993 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 9459) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 3994 (("]" 9460)) 3995 (("(" 9442) (argument_list 9467)) 3996 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 9465) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 3997 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 9466)) 3998 (("(" 9442) (argument_list 9469)) 3999 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 9473) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 4000 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 9474)) 4001 ((backup_doc_comment 9477)) 4002 (("(" 9478)) 4003 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 9479)) 4004 ((")" 9480)) 4005 ((T_USE 166) (lexical_vars 9481)) 4006 ((":" 168) (return_type 9482)) 4007 ((backup_fn_flags 9483)) 4008 (("{" 9484)) 4009 ((inner_statement_list 9485)) 4010 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 9486) ("~" 129)) 4011 ((backup_fn_flags 9487)) 4012 ((backup_doc_comment 9489)) 4013 (("(" 9490)) 4014 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 9491)) 4015 ((")" 9492)) 4016 ((":" 2271) (return_type 9493)) 4017 ((T_DOUBLE_ARROW 9494)) 4018 ((backup_fn_flags 9495)) 4019 ((backup_lex_pos 9496)) 4020 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9497) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 4021 (("%" 9199) ("*" 9200) ("+" 9201) ("-" 9202) ("." 9203) ("/" 9204) ("<" 9205) (">" 9206) ("?" 9207) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 9208) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 9209) (T_BOOLEAN_AND 9210) (T_BOOLEAN_OR 9211) (T_COALESCE 9212) (T_INSTANCEOF 9213) (T_IS_EQUAL 9214) (T_IS_GREATER_OR_EQUAL 9215) (T_IS_IDENTICAL 9216) (T_IS_NOT_EQUAL 9217) (T_IS_NOT_IDENTICAL 9218) (T_IS_SMALLER_OR_EQUAL 9219) (T_LOGICAL_AND 9220) (T_LOGICAL_OR 9221) (T_LOGICAL_XOR 9222) (T_PIPE 9223) (T_POW 9224) (T_SL 9225) (T_SPACESHIP 9226) (T_SR 9227) ("^" 9228) (backup_fn_flags 9498) ("|" 9229)) 4022 (("$" 9121) (T_ABSTRACT 9300) (T_ARRAY 9301) (T_AS 9302) (T_BREAK 9303) (T_CALLABLE 9304) (T_CASE 9305) (T_CATCH 9306) (T_CLASS 9307) (T_CLASS_C 9308) (T_CLONE 9309) (T_CONST 9310) (T_CONTINUE 9311) (T_DECLARE 9312) (T_DEFAULT 9313) (T_DIR 9314) (T_DO 9315) (T_ECHO 9316) (T_ELSE 9317) (T_ELSEIF 9318) (T_EMPTY 9319) (T_ENDDECLARE 9320) (T_ENDFOR 9321) (T_ENDFOREACH 9322) (T_ENDIF 9323) (T_ENDSWITCH 9324) (T_ENDWHILE 9325) (T_ENUM 9326) (T_EVAL 9327) (T_EXIT 9328) (T_EXTENDS 9329) (T_FILE 9330) (T_FINAL 9331) (T_FINALLY 9332) (T_FN 9333) (T_FOR 9334) (T_FOREACH 9335) (T_FUNCTION 9336) (T_FUNC_C 9337) (T_GLOBAL 9338) (T_GOTO 9339) (T_IF 9340) (T_IMPLEMENTS 9341) (T_INCLUDE 9342) (T_INCLUDE_ONCE 9343) (T_INSTANCEOF 9344) (T_INSTEADOF 9345) (T_INTERFACE 9346) (T_ISSET 9347) (T_LINE 9348) (T_LIST 9349) (T_LOGICAL_AND 9350) (T_LOGICAL_OR 9351) (T_LOGICAL_XOR 9352) (T_MATCH 9353) (T_METHOD_C 9354) (T_NAMESPACE 9355) (T_NEW 9356) (T_NS_C 9357) (T_PRINT 9358) (T_PRIVATE 9359) (T_PROPERTY_C 9360) (T_PROTECTED 9361) (T_PUBLIC 9362) (T_READONLY 9363) (T_REQUIRE 9364) (T_REQUIRE_ONCE 9365) (T_RETURN 9366) (T_STATIC 9367) (T_STRING 9368) (T_SWITCH 9369) (T_THROW 9370) (T_TRAIT 9371) (T_TRAIT_C 9372) (T_TRY 9373) (T_UNSET 9374) (T_USE 9375) (T_VAR 9376) (T_VARIABLE 9169) (T_WHILE 9377) (T_YIELD 9378) (identifier 9501) (member_name 9502) (reserved_non_modifiers 9381) (semi_reserved 9382) (simple_variable 9503) ("{" 9504)) 4023 (("(" 9387) (argument_list 9507)) 4024 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 9505) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 4025 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 9506)) 4026 ((T_FN 35) (T_FUNCTION 131) (fn 9183) #'9184 (inline_function 9511)) 4027 (("$" 9121) (T_STRING 9517) (T_VARIABLE 9169) (property_name 9524) (simple_variable 9519) ("{" 9520)) 4028 (("$" 9121) (T_STRING 9517) (T_VARIABLE 9169) (property_name 9518) (simple_variable 9519) ("{" 9520)) 4029 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 9515) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 4030 (("]" 9516)) 4031 (("(" 9387) (argument_list 9523)) 4032 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 9521) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 4033 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 9522)) 4034 (("(" 9387) (argument_list 9525)) 4035 (("`" 9527)) 4036 (("]" 9529)) 4037 (("=" 9530)) 4038 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9531) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 4039 (("%" 8280) ("*" 8281) ("+" 8282) ("-" 8283) ("." 8284) ("/" 8285) ("<" 8286) (">" 8287) ("?" 8288) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 8289) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 8290) (T_BOOLEAN_AND 8291) (T_BOOLEAN_OR 8292) (T_COALESCE 8293) (T_DOUBLE_ARROW 9534) (T_INSTANCEOF 8294) (T_IS_EQUAL 8295) (T_IS_GREATER_OR_EQUAL 8296) (T_IS_IDENTICAL 8297) (T_IS_NOT_EQUAL 8298) (T_IS_NOT_IDENTICAL 8299) (T_IS_SMALLER_OR_EQUAL 8300) (T_LOGICAL_AND 8301) (T_LOGICAL_OR 8302) (T_LOGICAL_XOR 8303) (T_PIPE 8304) (T_POW 8305) (T_SL 8306) (T_SPACESHIP 8307) (T_SR 8308) ("^" 8309) ("|" 8310)) 4040 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9535) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 4041 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 9544) (T_VARIABLE 2920) (encaps_var 2934)) 4042 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 9543) (T_VARIABLE 2920) (encaps_var 2925)) 4043 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 9585) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 4044 ((T_CLASS 9577)) 4045 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 9576) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 9557) (attribute 2982) (class_modifier 2963)) 4046 ((T_PAAMAYIM_NEKUDOTAYIM 9574)) 4047 (("(" 9442) (argument_list 9573)) 4048 ((T_NULLSAFE_OBJECT_OPERATOR 9564) (T_OBJECT_OPERATOR 9565) (T_PAAMAYIM_NEKUDOTAYIM 9566) ("[" 9567)) 4049 (("$" 9426) (T_STRING 9461) (T_VARIABLE 9428) (property_name 9572) (simple_variable 9463) ("{" 9464)) 4050 (("$" 9426) (T_STRING 9461) (T_VARIABLE 9428) (property_name 9571) (simple_variable 9463) ("{" 9464)) 4051 (("$" 9426) (T_VARIABLE 9428) (simple_variable 9570)) 4052 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 9568) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 4053 (("]" 9569)) 4054 (("$" 9426) (T_VARIABLE 9428) (simple_variable 9575)) 4055 (("(" 2984) (argument_list 2985) (ctor_arguments 9578)) 4056 ((T_EXTENDS 2987) (extends_from 9579)) 4057 ((T_IMPLEMENTS 2989) (implements_list 9580)) 4058 ((backup_doc_comment 9581)) 4059 (("{" 9582)) 4060 ((class_statement_list 9583)) 4061 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 9584)) 4062 (("%" 304) (")" 9586) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 4063 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 9588) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 4064 (("%" 304) (")" 9589) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 4065 (("{" 9590)) 4066 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 9591) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 4067 (("}" 9592)) 4068 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 9594) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 4069 ((")" 9595)) 4070 (("=" 9596)) 4071 (("!" 9119) ("\"" 9120) ("$" 9121) ("(" 9122) ("+" 9123) ("-" 9124) ("@" 9125) (T_ARRAY 9126) (T_ARRAY_CAST 9127) (T_ATTRIBUTE 130) (T_BOOL_CAST 9128) (T_CLASS_C 9129) (T_CLONE 9130) (T_CONSTANT_ENCAPSED_STRING 9131) (T_DEC 9132) (T_DIR 9133) (T_DNUMBER 9134) (T_DOUBLE_CAST 9135) (T_EMPTY 9136) (T_EVAL 9137) (T_EXIT 9138) (T_FILE 9139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 9140) (T_INC 9141) (T_INCLUDE 9142) (T_INCLUDE_ONCE 9143) (T_INT_CAST 9144) (T_ISSET 9145) (T_LINE 9146) (T_LIST 9147) (T_LNUMBER 9148) (T_MATCH 9149) (T_METHOD_C 9150) (T_NAME_FULLY_QUALIFIED 9151) (T_NAME_QUALIFIED 9152) (T_NAME_RELATIVE 9153) (T_NEW 9154) (T_NS_C 9155) (T_OBJECT_CAST 9156) (T_PRINT 9157) (T_PROPERTY_C 9158) (T_READONLY 9159) (T_REQUIRE 9160) (T_REQUIRE_ONCE 9161) (T_START_HEREDOC 9162) (T_STATIC 9163) (T_STRING 9164) (T_STRING_CAST 9165) (T_THROW 9166) (T_TRAIT_C 9167) (T_UNSET_CAST 9168) (T_VARIABLE 9169) (T_YIELD 9170) (T_YIELD_FROM 9171) ("[" 9172) ("`" 9173) (array_object_dereferenceable 9174) (attribute 135) (attributes 9175) (callable_expr 9176) (callable_variable 9177) (class_constant 9178) (class_name 9179) (constant 9180) (dereferenceable_scalar 9181) (expr 9597) (fn 9183) (fully_dereferenceable 107) #'9184 (function_call 9185) (inline_function 9186) (internal_functions_in_yacc 9187) (match 9188) (name 9189) (new_dereferenceable 9190) (new_non_dereferenceable 9191) (scalar 9192) (simple_variable 9193) (static_member 9194) (variable 9195) (variable_class_name 9196) ("~" 9197)) 4072 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 9599) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4073 (("," 5226) (possible_comma 9600)) 4074 ((")" 9601)) 4075 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 9609) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 9610) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 9611) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4076 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 9614) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4077 (("," 1498) (possible_comma 9612)) 4078 ((")" 9613)) 4079 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 9616) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 4080 (("%" 304) (")" 9617) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 4081 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 9619) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 4082 (("%" 304) (")" 9620) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 4083 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 9626) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 9627) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 9628) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 9629) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 4084 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 9635) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4085 (("%" 2442) (")" 9632) ("*" 2443) ("+" 2444) ("," 9633) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 4086 (("," 5287) (possible_comma 9630)) 4087 ((")" 9631)) 4088 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 9634) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4089 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 9639) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 4090 ((")" 9640)) 4091 (("%" 304) (")" 9632) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 4092 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 9647) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 4093 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 9648)) 4094 (("\"" 9650) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 4095 (("$" 8736) (T_ABSTRACT 8922) (T_ARRAY 8923) (T_AS 8924) (T_BREAK 8925) (T_CALLABLE 8926) (T_CASE 8927) (T_CATCH 8928) (T_CLASS 8929) (T_CLASS_C 8930) (T_CLONE 8931) (T_CONST 8932) (T_CONTINUE 8933) (T_DECLARE 8934) (T_DEFAULT 8935) (T_DIR 8936) (T_DO 8937) (T_ECHO 8938) (T_ELSE 8939) (T_ELSEIF 8940) (T_EMPTY 8941) (T_ENDDECLARE 8942) (T_ENDFOR 8943) (T_ENDFOREACH 8944) (T_ENDIF 8945) (T_ENDSWITCH 8946) (T_ENDWHILE 8947) (T_ENUM 8948) (T_EVAL 8949) (T_EXIT 8950) (T_EXTENDS 8951) (T_FILE 8952) (T_FINAL 8953) (T_FINALLY 8954) (T_FN 8955) (T_FOR 8956) (T_FOREACH 8957) (T_FUNCTION 8958) (T_FUNC_C 8959) (T_GLOBAL 8960) (T_GOTO 8961) (T_IF 8962) (T_IMPLEMENTS 8963) (T_INCLUDE 8964) (T_INCLUDE_ONCE 8965) (T_INSTANCEOF 8966) (T_INSTEADOF 8967) (T_INTERFACE 8968) (T_ISSET 8969) (T_LINE 8970) (T_LIST 8971) (T_LOGICAL_AND 8972) (T_LOGICAL_OR 8973) (T_LOGICAL_XOR 8974) (T_MATCH 8975) (T_METHOD_C 8976) (T_NAMESPACE 8977) (T_NEW 8978) (T_NS_C 8979) (T_PRINT 8980) (T_PRIVATE 8981) (T_PROPERTY_C 8982) (T_PROTECTED 8983) (T_PUBLIC 8984) (T_READONLY 8985) (T_REQUIRE 8986) (T_REQUIRE_ONCE 8987) (T_RETURN 8988) (T_STATIC 8989) (T_STRING 8990) (T_SWITCH 8991) (T_THROW 8992) (T_TRAIT 8993) (T_TRAIT_C 8994) (T_TRY 8995) (T_UNSET 8996) (T_USE 8997) (T_VAR 8998) (T_VARIABLE 8784) (T_WHILE 8999) (T_YIELD 9000) (identifier 9653) (member_name 9654) (reserved_non_modifiers 9003) (semi_reserved 9004) (simple_variable 9655) ("{" 9656)) 4096 (("(" 9009) (argument_list 9659)) 4097 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 9657) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 4098 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 9658)) 4099 ((T_FN 35) (T_FUNCTION 131) (fn 8799) #'8800 (inline_function 9663)) 4100 (("$" 8736) (T_STRING 9669) (T_VARIABLE 8784) (property_name 9676) (simple_variable 9671) ("{" 9672)) 4101 (("$" 8736) (T_STRING 9669) (T_VARIABLE 8784) (property_name 9670) (simple_variable 9671) ("{" 9672)) 4102 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 9667) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 4103 (("]" 9668)) 4104 (("(" 9009) (argument_list 9675)) 4105 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 9673) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 4106 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 9674)) 4107 (("(" 9009) (argument_list 9677)) 4108 (("$" 9678) (T_VARIABLE 9680) (simple_variable 9723) ("{" 9724)) 4109 (("(" 9694) (argument_list 9722)) 4110 ((T_NULLSAFE_OBJECT_OPERATOR 9708) (T_OBJECT_OPERATOR 9709) ("[" 9710)) 4111 (("(" 9694) (argument_list 9707)) 4112 ((T_PAAMAYIM_NEKUDOTAYIM 9703)) 4113 (("(" 9694) (argument_list 9702)) 4114 ((T_PAAMAYIM_NEKUDOTAYIM 9691)) 4115 (("$" 9678) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 9680) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 9692) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 9693) ("{" 1660)) 4116 (("(" 9694) (argument_list 9695)) 4117 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 9696) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 9697) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 9698) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4118 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 9701) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4119 (("," 1498) (possible_comma 9699)) 4120 ((")" 9700)) 4121 (("$" 9678) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 9680) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 9704) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 9705) ("{" 1676)) 4122 (("(" 9694) (argument_list 9706)) 4123 (("$" 9678) (T_STRING 9713) (T_VARIABLE 9680) (property_name 9720) (simple_variable 9715) ("{" 9716)) 4124 (("$" 9678) (T_STRING 9713) (T_VARIABLE 9680) (property_name 9714) (simple_variable 9715) ("{" 9716)) 4125 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 9711) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 4126 (("]" 9712)) 4127 (("(" 9694) (argument_list 9719)) 4128 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 9717) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 4129 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 9718)) 4130 (("(" 9694) (argument_list 9721)) 4131 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 9725) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 4132 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 9726)) 4133 (("`" 9728)) 4134 (("]" 9730)) 4135 (("=" 9731)) 4136 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 9732) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 4137 (("%" 1782) ("*" 1783) ("+" 1784) ("-" 1785) ("." 1786) ("/" 1787) ("<" 1788) (">" 1789) ("?" 1790) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1791) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1792) (T_BOOLEAN_AND 1793) (T_BOOLEAN_OR 1794) (T_COALESCE 1795) (T_DOUBLE_ARROW 9735) (T_INSTANCEOF 1796) (T_IS_EQUAL 1797) (T_IS_GREATER_OR_EQUAL 1798) (T_IS_IDENTICAL 1799) (T_IS_NOT_EQUAL 1800) (T_IS_NOT_IDENTICAL 1801) (T_IS_SMALLER_OR_EQUAL 1802) (T_LOGICAL_AND 1803) (T_LOGICAL_OR 1804) (T_LOGICAL_XOR 1805) (T_PIPE 1806) (T_POW 1807) (T_SL 1808) (T_SPACESHIP 1809) (T_SR 1810) ("^" 1811) ("|" 1812)) 4138 (("!" 8734) ("\"" 8735) ("$" 8736) ("(" 8737) ("+" 8738) ("-" 8739) ("@" 8740) (T_ARRAY 8741) (T_ARRAY_CAST 8742) (T_ATTRIBUTE 130) (T_BOOL_CAST 8743) (T_CLASS_C 8744) (T_CLONE 8745) (T_CONSTANT_ENCAPSED_STRING 8746) (T_DEC 8747) (T_DIR 8748) (T_DNUMBER 8749) (T_DOUBLE_CAST 8750) (T_EMPTY 8751) (T_EVAL 8752) (T_EXIT 8753) (T_FILE 8754) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 8755) (T_INC 8756) (T_INCLUDE 8757) (T_INCLUDE_ONCE 8758) (T_INT_CAST 8759) (T_ISSET 8760) (T_LINE 8761) (T_LIST 8814) (T_LNUMBER 8763) (T_MATCH 8764) (T_METHOD_C 8765) (T_NAME_FULLY_QUALIFIED 8766) (T_NAME_QUALIFIED 8767) (T_NAME_RELATIVE 8768) (T_NEW 8769) (T_NS_C 8770) (T_OBJECT_CAST 8771) (T_PRINT 8772) (T_PROPERTY_C 8773) (T_READONLY 8774) (T_REQUIRE 8775) (T_REQUIRE_ONCE 8776) (T_START_HEREDOC 8777) (T_STATIC 8778) (T_STRING 8779) (T_STRING_CAST 8780) (T_THROW 8781) (T_TRAIT_C 8782) (T_UNSET_CAST 8783) (T_VARIABLE 8784) (T_YIELD 8785) (T_YIELD_FROM 8786) ("[" 8787) ("`" 8788) (array_object_dereferenceable 8790) (attribute 135) (attributes 8791) (callable_expr 8792) (callable_variable 8793) (class_constant 8794) (class_name 8795) (constant 8796) (dereferenceable_scalar 8797) (expr 9736) (fn 8799) (fully_dereferenceable 107) #'8800 (function_call 8801) (inline_function 8802) (internal_functions_in_yacc 8803) (match 8804) (name 8805) (new_dereferenceable 8806) (new_non_dereferenceable 8807) (scalar 8808) (simple_variable 8809) (static_member 8810) (variable 8811) (variable_class_name 8812) ("~" 8813)) 4139 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 9745) (T_VARIABLE 2920) (encaps_var 2934)) 4140 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 9744) (T_VARIABLE 2920) (encaps_var 2925)) 4141 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 9786) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 4142 ((T_CLASS 9778)) 4143 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 9777) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 9758) (attribute 2982) (class_modifier 2963)) 4144 ((T_PAAMAYIM_NEKUDOTAYIM 9775)) 4145 (("(" 9064) (argument_list 9774)) 4146 ((T_NULLSAFE_OBJECT_OPERATOR 9765) (T_OBJECT_OPERATOR 9766) (T_PAAMAYIM_NEKUDOTAYIM 9767) ("[" 9768)) 4147 (("$" 9048) (T_STRING 9083) (T_VARIABLE 9050) (property_name 9773) (simple_variable 9085) ("{" 9086)) 4148 (("$" 9048) (T_STRING 9083) (T_VARIABLE 9050) (property_name 9772) (simple_variable 9085) ("{" 9086)) 4149 (("$" 9048) (T_VARIABLE 9050) (simple_variable 9771)) 4150 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 9769) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 4151 (("]" 9770)) 4152 (("$" 9048) (T_VARIABLE 9050) (simple_variable 9776)) 4153 (("(" 2984) (argument_list 2985) (ctor_arguments 9779)) 4154 ((T_EXTENDS 2987) (extends_from 9780)) 4155 ((T_IMPLEMENTS 2989) (implements_list 9781)) 4156 ((backup_doc_comment 9782)) 4157 (("{" 9783)) 4158 ((class_statement_list 9784)) 4159 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 9785)) 4160 (("%" 304) (")" 9787) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 4161 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 9789) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 4162 (("%" 304) (")" 9790) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 4163 (("{" 9791)) 4164 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 9792) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 4165 (("}" 9793)) 4166 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 9795) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 4167 ((")" 9796)) 4168 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 9798) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4169 (("," 5226) (possible_comma 9799)) 4170 ((")" 9800)) 4171 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 9808) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 9809) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 9810) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4172 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 9813) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4173 (("," 1498) (possible_comma 9811)) 4174 ((")" 9812)) 4175 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 9815) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 4176 (("%" 304) (")" 9816) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 4177 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 9818) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 4178 (("%" 304) (")" 9819) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 4179 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 9825) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 9826) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 9827) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 9828) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 4180 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 9834) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4181 (("%" 2442) (")" 9831) ("*" 2443) ("+" 2444) ("," 9832) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 4182 (("," 5287) (possible_comma 9829)) 4183 ((")" 9830)) 4184 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 9833) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4185 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 9838) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 4186 ((")" 9839)) 4187 (("%" 304) (")" 9831) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 4188 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 9846) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 4189 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 9847)) 4190 (("\"" 9849) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 4191 (("$" 1697) (T_ABSTRACT 8001) (T_ARRAY 8002) (T_AS 8003) (T_BREAK 8004) (T_CALLABLE 8005) (T_CASE 8006) (T_CATCH 8007) (T_CLASS 8008) (T_CLASS_C 8009) (T_CLONE 8010) (T_CONST 8011) (T_CONTINUE 8012) (T_DECLARE 8013) (T_DEFAULT 8014) (T_DIR 8015) (T_DO 8016) (T_ECHO 8017) (T_ELSE 8018) (T_ELSEIF 8019) (T_EMPTY 8020) (T_ENDDECLARE 8021) (T_ENDFOR 8022) (T_ENDFOREACH 8023) (T_ENDIF 8024) (T_ENDSWITCH 8025) (T_ENDWHILE 8026) (T_ENUM 8027) (T_EVAL 8028) (T_EXIT 8029) (T_EXTENDS 8030) (T_FILE 8031) (T_FINAL 8032) (T_FINALLY 8033) (T_FN 8034) (T_FOR 8035) (T_FOREACH 8036) (T_FUNCTION 8037) (T_FUNC_C 8038) (T_GLOBAL 8039) (T_GOTO 8040) (T_IF 8041) (T_IMPLEMENTS 8042) (T_INCLUDE 8043) (T_INCLUDE_ONCE 8044) (T_INSTANCEOF 8045) (T_INSTEADOF 8046) (T_INTERFACE 8047) (T_ISSET 8048) (T_LINE 8049) (T_LIST 8050) (T_LOGICAL_AND 8051) (T_LOGICAL_OR 8052) (T_LOGICAL_XOR 8053) (T_MATCH 8054) (T_METHOD_C 8055) (T_NAMESPACE 8056) (T_NEW 8057) (T_NS_C 8058) (T_PRINT 8059) (T_PRIVATE 8060) (T_PROPERTY_C 8061) (T_PROTECTED 8062) (T_PUBLIC 8063) (T_READONLY 8064) (T_REQUIRE 8065) (T_REQUIRE_ONCE 8066) (T_RETURN 8067) (T_STATIC 8068) (T_STRING 8069) (T_SWITCH 8070) (T_THROW 8071) (T_TRAIT 8072) (T_TRAIT_C 8073) (T_TRY 8074) (T_UNSET 8075) (T_USE 8076) (T_VAR 8077) (T_VARIABLE 1746) (T_WHILE 8078) (T_YIELD 8079) (identifier 9852) (member_name 9853) (reserved_non_modifiers 8082) (semi_reserved 8083) (simple_variable 9854) ("{" 9855)) 4192 (("(" 8088) (argument_list 9858)) 4193 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 9856) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 4194 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 9857)) 4195 ((T_FN 35) (T_FUNCTION 131) (fn 1763) #'1764 (inline_function 9862)) 4196 (("$" 1697) (T_STRING 9869) (T_VARIABLE 1746) (property_name 9876) (simple_variable 9871) ("{" 9872)) 4197 (("$" 1697) (T_STRING 9869) (T_VARIABLE 1746) (property_name 9870) (simple_variable 9871) ("{" 9872)) 4198 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 9867) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 4199 (("]" 9868)) 4200 (("(" 8088) (argument_list 9875)) 4201 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 9873) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 4202 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 9874)) 4203 (("(" 8088) (argument_list 9877)) 4204 (("`" 9880)) 4205 (("]" 9882)) 4206 (("=" 9883)) 4207 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 9884) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 4208 (("%" 1782) ("*" 1783) ("+" 1784) ("-" 1785) ("." 1786) ("/" 1787) ("<" 1788) (">" 1789) ("?" 1790) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1791) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1792) (T_BOOLEAN_AND 1793) (T_BOOLEAN_OR 1794) (T_COALESCE 1795) (T_DOUBLE_ARROW 9887) (T_INSTANCEOF 1796) (T_IS_EQUAL 1797) (T_IS_GREATER_OR_EQUAL 1798) (T_IS_IDENTICAL 1799) (T_IS_NOT_EQUAL 1800) (T_IS_NOT_IDENTICAL 1801) (T_IS_SMALLER_OR_EQUAL 1802) (T_LOGICAL_AND 1803) (T_LOGICAL_OR 1804) (T_LOGICAL_XOR 1805) (T_PIPE 1806) (T_POW 1807) (T_SL 1808) (T_SPACESHIP 1809) (T_SR 1810) ("^" 1811) ("|" 1812)) 4209 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1780) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (array_object_dereferenceable 1752) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 9888) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 4210 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 9897) (T_VARIABLE 2920) (encaps_var 2934)) 4211 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 9896) (T_VARIABLE 2920) (encaps_var 2925)) 4212 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 9938) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 4213 ((T_CLASS 9930)) 4214 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 9929) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 9910) (attribute 2982) (class_modifier 2963)) 4215 ((T_PAAMAYIM_NEKUDOTAYIM 9927)) 4216 (("(" 8143) (argument_list 9926)) 4217 ((T_NULLSAFE_OBJECT_OPERATOR 9917) (T_OBJECT_OPERATOR 9918) (T_PAAMAYIM_NEKUDOTAYIM 9919) ("[" 9920)) 4218 (("$" 8127) (T_STRING 8162) (T_VARIABLE 8129) (property_name 9925) (simple_variable 8164) ("{" 8165)) 4219 (("$" 8127) (T_STRING 8162) (T_VARIABLE 8129) (property_name 9924) (simple_variable 8164) ("{" 8165)) 4220 (("$" 8127) (T_VARIABLE 8129) (simple_variable 9923)) 4221 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 9921) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 4222 (("]" 9922)) 4223 (("$" 8127) (T_VARIABLE 8129) (simple_variable 9928)) 4224 (("(" 2984) (argument_list 2985) (ctor_arguments 9931)) 4225 ((T_EXTENDS 2987) (extends_from 9932)) 4226 ((T_IMPLEMENTS 2989) (implements_list 9933)) 4227 ((backup_doc_comment 9934)) 4228 (("{" 9935)) 4229 ((class_statement_list 9936)) 4230 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 9937)) 4231 (("%" 304) (")" 9939) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 4232 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 9941) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 4233 (("%" 304) (")" 9942) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 4234 (("{" 9943)) 4235 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 9944) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 4236 (("}" 9945)) 4237 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 9947) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 4238 ((")" 9948)) 4239 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 9950) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4240 (("," 5226) (possible_comma 9951)) 4241 ((")" 9952)) 4242 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 9960) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 9961) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 9962) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4243 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 9965) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4244 (("," 1498) (possible_comma 9963)) 4245 ((")" 9964)) 4246 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 9967) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 4247 (("%" 304) (")" 9968) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 4248 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 9970) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 4249 (("%" 304) (")" 9971) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 4250 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 9978) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 9979) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 9980) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 9981) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 4251 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 9987) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4252 (("%" 2442) (")" 9984) ("*" 2443) ("+" 2444) ("," 9985) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 4253 (("," 5287) (possible_comma 9982)) 4254 ((")" 9983)) 4255 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 9986) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4256 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 9991) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 4257 ((")" 9992)) 4258 (("%" 304) (")" 9984) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 4259 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 9999) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 4260 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 10000)) 4261 (("\"" 10002) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 4262 (("$" 10005) (T_VARIABLE 10012) (simple_variable 10056) ("{" 10057)) 4263 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 10054) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 4264 ((T_CLASS 10046)) 4265 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 10045) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 10014) (attribute 2982) (class_modifier 2963)) 4266 ((T_PAAMAYIM_NEKUDOTAYIM 10043)) 4267 (("(" 10035) (argument_list 10036)) 4268 ((T_NULLSAFE_OBJECT_OPERATOR 10021) (T_OBJECT_OPERATOR 10022) (T_PAAMAYIM_NEKUDOTAYIM 10023) ("[" 10024)) 4269 (("$" 10005) (T_STRING 10028) (T_VARIABLE 10012) (property_name 10034) (simple_variable 10030) ("{" 10031)) 4270 (("$" 10005) (T_STRING 10028) (T_VARIABLE 10012) (property_name 10029) (simple_variable 10030) ("{" 10031)) 4271 (("$" 10005) (T_VARIABLE 10012) (simple_variable 10027)) 4272 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 10025) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 4273 (("]" 10026)) 4274 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 10032) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 4275 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 10033)) 4276 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 10037) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 10038) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 10039) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4277 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 10042) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4278 (("," 1498) (possible_comma 10040)) 4279 ((")" 10041)) 4280 (("$" 10005) (T_VARIABLE 10012) (simple_variable 10044)) 4281 (("(" 2984) (argument_list 2985) (ctor_arguments 10047)) 4282 ((T_EXTENDS 2987) (extends_from 10048)) 4283 ((T_IMPLEMENTS 2989) (implements_list 10049)) 4284 ((backup_doc_comment 10050)) 4285 (("{" 10051)) 4286 ((class_statement_list 10052)) 4287 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 10053)) 4288 (("%" 304) (")" 10055) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 4289 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 10058) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 4290 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 10059)) 4291 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 10061) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 4292 ((")" 10062)) 4293 (("%" 304) (")" 10064) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 4294 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 10067) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 4295 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 10068)) 4296 (("\"" 10070) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 4297 ((backup_doc_comment 10073)) 4298 (("(" 10074)) 4299 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 10075)) 4300 ((")" 10076)) 4301 ((T_USE 166) (lexical_vars 10077)) 4302 ((":" 168) (return_type 10078)) 4303 ((backup_fn_flags 10079)) 4304 (("{" 10080)) 4305 ((inner_statement_list 10081)) 4306 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 10082) ("~" 129)) 4307 ((backup_fn_flags 10083)) 4308 ((backup_doc_comment 10085)) 4309 (("(" 10086)) 4310 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 10087)) 4311 ((")" 10088)) 4312 ((":" 2271) (return_type 10089)) 4313 ((T_DOUBLE_ARROW 10090)) 4314 ((backup_fn_flags 10091)) 4315 ((backup_lex_pos 10092)) 4316 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 10093) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 4317 (("%" 1305) ("*" 1306) ("+" 1307) ("-" 1308) ("." 1309) ("/" 1310) ("<" 1311) (">" 1312) ("?" 1313) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1314) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1315) (T_BOOLEAN_AND 1316) (T_BOOLEAN_OR 1317) (T_COALESCE 1318) (T_INSTANCEOF 1319) (T_IS_EQUAL 1320) (T_IS_GREATER_OR_EQUAL 1321) (T_IS_IDENTICAL 1322) (T_IS_NOT_EQUAL 1323) (T_IS_NOT_IDENTICAL 1324) (T_IS_SMALLER_OR_EQUAL 1325) (T_LOGICAL_AND 1326) (T_LOGICAL_OR 1327) (T_LOGICAL_XOR 1328) (T_PIPE 1329) (T_POW 1330) (T_SL 1331) (T_SPACESHIP 1332) (T_SR 1333) ("^" 1334) (backup_fn_flags 10094) ("|" 1335)) 4318 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 10096) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4319 (("$" 1227) (T_ABSTRACT 1406) (T_ARRAY 1407) (T_AS 1408) (T_BREAK 1409) (T_CALLABLE 1410) (T_CASE 1411) (T_CATCH 1412) (T_CLASS 1413) (T_CLASS_C 1414) (T_CLONE 1415) (T_CONST 1416) (T_CONTINUE 1417) (T_DECLARE 1418) (T_DEFAULT 1419) (T_DIR 1420) (T_DO 1421) (T_ECHO 1422) (T_ELSE 1423) (T_ELSEIF 1424) (T_EMPTY 1425) (T_ENDDECLARE 1426) (T_ENDFOR 1427) (T_ENDFOREACH 1428) (T_ENDIF 1429) (T_ENDSWITCH 1430) (T_ENDWHILE 1431) (T_ENUM 1432) (T_EVAL 1433) (T_EXIT 1434) (T_EXTENDS 1435) (T_FILE 1436) (T_FINAL 1437) (T_FINALLY 1438) (T_FN 1439) (T_FOR 1440) (T_FOREACH 1441) (T_FUNCTION 1442) (T_FUNC_C 1443) (T_GLOBAL 1444) (T_GOTO 1445) (T_IF 1446) (T_IMPLEMENTS 1447) (T_INCLUDE 1448) (T_INCLUDE_ONCE 1449) (T_INSTANCEOF 1450) (T_INSTEADOF 1451) (T_INTERFACE 1452) (T_ISSET 1453) (T_LINE 1454) (T_LIST 1455) (T_LOGICAL_AND 1456) (T_LOGICAL_OR 1457) (T_LOGICAL_XOR 1458) (T_MATCH 1459) (T_METHOD_C 1460) (T_NAMESPACE 1461) (T_NEW 1462) (T_NS_C 1463) (T_PRINT 1464) (T_PRIVATE 1465) (T_PROPERTY_C 1466) (T_PROTECTED 1467) (T_PUBLIC 1468) (T_READONLY 1469) (T_REQUIRE 1470) (T_REQUIRE_ONCE 1471) (T_RETURN 1472) (T_STATIC 1473) (T_STRING 1474) (T_SWITCH 1475) (T_THROW 1476) (T_TRAIT 1477) (T_TRAIT_C 1478) (T_TRY 1479) (T_UNSET 1480) (T_USE 1481) (T_VAR 1482) (T_VARIABLE 1275) (T_WHILE 1483) (T_YIELD 1484) (identifier 10098) (member_name 10099) (reserved_non_modifiers 1487) (semi_reserved 1488) (simple_variable 10100) ("{" 10101)) 4320 (("(" 1493) (argument_list 10104)) 4321 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 10102) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 4322 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 10103)) 4323 ((T_FN 35) (T_FUNCTION 131) (fn 1289) #'1290 (inline_function 10108)) 4324 (("$" 1227) (T_STRING 10114) (T_VARIABLE 1275) (property_name 10121) (simple_variable 10116) ("{" 10117)) 4325 (("$" 1227) (T_STRING 10114) (T_VARIABLE 1275) (property_name 10115) (simple_variable 10116) ("{" 10117)) 4326 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 10112) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 4327 (("]" 10113)) 4328 (("(" 1493) (argument_list 10120)) 4329 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 10118) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 4330 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 10119)) 4331 (("(" 1493) (argument_list 10122)) 4332 (("`" 10124)) 4333 (("]" 10126)) 4334 (("=" 10127)) 4335 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 10128) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 4336 (("%" 1305) ("*" 1306) ("+" 1307) ("-" 1308) ("." 1309) ("/" 1310) ("<" 1311) (">" 1312) ("?" 1313) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1314) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1315) (T_BOOLEAN_AND 1316) (T_BOOLEAN_OR 1317) (T_COALESCE 1318) (T_DOUBLE_ARROW 10131) (T_INSTANCEOF 1319) (T_IS_EQUAL 1320) (T_IS_GREATER_OR_EQUAL 1321) (T_IS_IDENTICAL 1322) (T_IS_NOT_EQUAL 1323) (T_IS_NOT_IDENTICAL 1324) (T_IS_SMALLER_OR_EQUAL 1325) (T_LOGICAL_AND 1326) (T_LOGICAL_OR 1327) (T_LOGICAL_XOR 1328) (T_PIPE 1329) (T_POW 1330) (T_SL 1331) (T_SPACESHIP 1332) (T_SR 1333) ("^" 1334) ("|" 1335)) 4337 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 10132) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 4338 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 10141) (T_VARIABLE 2920) (encaps_var 2934)) 4339 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 10140) (T_VARIABLE 2920) (encaps_var 2925)) 4340 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 10182) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 4341 ((T_CLASS 10174)) 4342 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 10173) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 10154) (attribute 2982) (class_modifier 2963)) 4343 ((T_PAAMAYIM_NEKUDOTAYIM 10171)) 4344 (("(" 1663) (argument_list 10170)) 4345 ((T_NULLSAFE_OBJECT_OPERATOR 10161) (T_OBJECT_OPERATOR 10162) (T_PAAMAYIM_NEKUDOTAYIM 10163) ("[" 10164)) 4346 (("$" 1539) (T_STRING 1686) (T_VARIABLE 1559) (property_name 10169) (simple_variable 1688) ("{" 1689)) 4347 (("$" 1539) (T_STRING 1686) (T_VARIABLE 1559) (property_name 10168) (simple_variable 1688) ("{" 1689)) 4348 (("$" 1539) (T_VARIABLE 1559) (simple_variable 10167)) 4349 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 10165) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 4350 (("]" 10166)) 4351 (("$" 1539) (T_VARIABLE 1559) (simple_variable 10172)) 4352 (("(" 2984) (argument_list 2985) (ctor_arguments 10175)) 4353 ((T_EXTENDS 2987) (extends_from 10176)) 4354 ((T_IMPLEMENTS 2989) (implements_list 10177)) 4355 ((backup_doc_comment 10178)) 4356 (("{" 10179)) 4357 ((class_statement_list 10180)) 4358 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 10181)) 4359 (("%" 304) (")" 10183) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 4360 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 10185) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 4361 (("%" 304) (")" 10186) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 4362 (("{" 10187)) 4363 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 10188) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 4364 (("}" 10189)) 4365 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 10191) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 4366 ((")" 10192)) 4367 (("=" 10193)) 4368 (("!" 1225) ("\"" 1226) ("$" 1227) ("(" 1228) ("+" 1229) ("-" 1230) ("@" 1231) (T_ARRAY 1232) (T_ARRAY_CAST 1233) (T_ATTRIBUTE 130) (T_BOOL_CAST 1234) (T_CLASS_C 1235) (T_CLONE 1236) (T_CONSTANT_ENCAPSED_STRING 1237) (T_DEC 1238) (T_DIR 1239) (T_DNUMBER 1240) (T_DOUBLE_CAST 1241) (T_EMPTY 1242) (T_EVAL 1243) (T_EXIT 1244) (T_FILE 1245) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1246) (T_INC 1247) (T_INCLUDE 1248) (T_INCLUDE_ONCE 1249) (T_INT_CAST 1250) (T_ISSET 1251) (T_LINE 1252) (T_LIST 1253) (T_LNUMBER 1254) (T_MATCH 1255) (T_METHOD_C 1256) (T_NAME_FULLY_QUALIFIED 1257) (T_NAME_QUALIFIED 1258) (T_NAME_RELATIVE 1259) (T_NEW 1260) (T_NS_C 1261) (T_OBJECT_CAST 1262) (T_PRINT 1263) (T_PROPERTY_C 1264) (T_READONLY 1265) (T_REQUIRE 1266) (T_REQUIRE_ONCE 1267) (T_START_HEREDOC 1268) (T_STATIC 1269) (T_STRING 1270) (T_STRING_CAST 1271) (T_THROW 1272) (T_TRAIT_C 1273) (T_UNSET_CAST 1274) (T_VARIABLE 1275) (T_YIELD 1276) (T_YIELD_FROM 1277) ("[" 1278) ("`" 1279) (array_object_dereferenceable 1280) (attribute 135) (attributes 1281) (callable_expr 1282) (callable_variable 1283) (class_constant 1284) (class_name 1285) (constant 1286) (dereferenceable_scalar 1287) (expr 10194) (fn 1289) (fully_dereferenceable 107) #'1290 (function_call 1291) (inline_function 1292) (internal_functions_in_yacc 1293) (match 1294) (name 1295) (new_dereferenceable 1296) (new_non_dereferenceable 1297) (scalar 1298) (simple_variable 1299) (static_member 1300) (variable 1301) (variable_class_name 1302) ("~" 1303)) 4369 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 10196) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4370 (("," 5226) (possible_comma 10197)) 4371 ((")" 10198)) 4372 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 10206) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 10207) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 10208) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4373 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 10211) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4374 (("," 1498) (possible_comma 10209)) 4375 ((")" 10210)) 4376 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 10213) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 4377 (("%" 304) (")" 10214) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 4378 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 10216) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 4379 (("%" 304) (")" 10217) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 4380 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 10223) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 10224) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 10225) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 10226) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 4381 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 10232) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4382 (("%" 2442) (")" 10229) ("*" 2443) ("+" 2444) ("," 10230) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 4383 (("," 5287) (possible_comma 10227)) 4384 ((")" 10228)) 4385 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 10231) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4386 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 10236) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 4387 ((")" 10237)) 4388 (("%" 304) (")" 10229) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 4389 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 10244) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 4390 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 10245)) 4391 (("\"" 10247) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 4392 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 10255) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 10256) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 10257) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4393 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 10260) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4394 (("," 1498) (possible_comma 10258)) 4395 ((")" 10259)) 4396 (("$" 10262) (T_VARIABLE 10269) (simple_variable 10313) ("{" 10314)) 4397 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 10311) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 4398 ((T_CLASS 10303)) 4399 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 10302) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 10271) (attribute 2982) (class_modifier 2963)) 4400 ((T_PAAMAYIM_NEKUDOTAYIM 10300)) 4401 (("(" 10292) (argument_list 10293)) 4402 ((T_NULLSAFE_OBJECT_OPERATOR 10278) (T_OBJECT_OPERATOR 10279) (T_PAAMAYIM_NEKUDOTAYIM 10280) ("[" 10281)) 4403 (("$" 10262) (T_STRING 10285) (T_VARIABLE 10269) (property_name 10291) (simple_variable 10287) ("{" 10288)) 4404 (("$" 10262) (T_STRING 10285) (T_VARIABLE 10269) (property_name 10286) (simple_variable 10287) ("{" 10288)) 4405 (("$" 10262) (T_VARIABLE 10269) (simple_variable 10284)) 4406 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 10282) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 4407 (("]" 10283)) 4408 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 10289) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 4409 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 10290)) 4410 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 10294) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 10295) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 10296) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4411 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 10299) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4412 (("," 1498) (possible_comma 10297)) 4413 ((")" 10298)) 4414 (("$" 10262) (T_VARIABLE 10269) (simple_variable 10301)) 4415 (("(" 2984) (argument_list 2985) (ctor_arguments 10304)) 4416 ((T_EXTENDS 2987) (extends_from 10305)) 4417 ((T_IMPLEMENTS 2989) (implements_list 10306)) 4418 ((backup_doc_comment 10307)) 4419 (("{" 10308)) 4420 ((class_statement_list 10309)) 4421 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 10310)) 4422 (("%" 304) (")" 10312) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 4423 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 10315) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 4424 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 10316)) 4425 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 10318) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 4426 (("%" 304) (")" 10319) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 4427 (("{" 10320)) 4428 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 10321) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 4429 (("}" 10322)) 4430 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 10324) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 4431 ((")" 10325)) 4432 (("=" 10326)) 4433 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 10327) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4434 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 10329) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4435 (("," 5226) (possible_comma 10330)) 4436 ((")" 10331)) 4437 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 10337) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 10338) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 10339) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4438 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 10342) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4439 (("," 1498) (possible_comma 10340)) 4440 ((")" 10341)) 4441 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 10344) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 4442 (("%" 304) (")" 10345) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 4443 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 10347) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 4444 (("%" 304) (")" 10348) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 4445 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 10352) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 10353) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 10354) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 10355) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 4446 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 10361) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4447 (("%" 2442) (")" 10358) ("*" 2443) ("+" 2444) ("," 10359) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 4448 (("," 5287) (possible_comma 10356)) 4449 ((")" 10357)) 4450 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 10360) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4451 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 10363) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 4452 ((")" 10364)) 4453 (("$" 960) (T_ABSTRACT 10366) (T_ARRAY 10367) (T_AS 10368) (T_BREAK 10369) (T_CALLABLE 10370) (T_CASE 10371) (T_CATCH 10372) (T_CLASS 10373) (T_CLASS_C 10374) (T_CLONE 10375) (T_CONST 10376) (T_CONTINUE 10377) (T_DECLARE 10378) (T_DEFAULT 10379) (T_DIR 10380) (T_DO 10381) (T_ECHO 10382) (T_ELSE 10383) (T_ELSEIF 10384) (T_EMPTY 10385) (T_ENDDECLARE 10386) (T_ENDFOR 10387) (T_ENDFOREACH 10388) (T_ENDIF 10389) (T_ENDSWITCH 10390) (T_ENDWHILE 10391) (T_ENUM 10392) (T_EVAL 10393) (T_EXIT 10394) (T_EXTENDS 10395) (T_FILE 10396) (T_FINAL 10397) (T_FINALLY 10398) (T_FN 10399) (T_FOR 10400) (T_FOREACH 10401) (T_FUNCTION 10402) (T_FUNC_C 10403) (T_GLOBAL 10404) (T_GOTO 10405) (T_IF 10406) (T_IMPLEMENTS 10407) (T_INCLUDE 10408) (T_INCLUDE_ONCE 10409) (T_INSTANCEOF 10410) (T_INSTEADOF 10411) (T_INTERFACE 10412) (T_ISSET 10413) (T_LINE 10414) (T_LIST 10415) (T_LOGICAL_AND 10416) (T_LOGICAL_OR 10417) (T_LOGICAL_XOR 10418) (T_MATCH 10419) (T_METHOD_C 10420) (T_NAMESPACE 10421) (T_NEW 10422) (T_NS_C 10423) (T_PRINT 10424) (T_PRIVATE 10425) (T_PROPERTY_C 10426) (T_PROTECTED 10427) (T_PUBLIC 10428) (T_READONLY 10429) (T_REQUIRE 10430) (T_REQUIRE_ONCE 10431) (T_RETURN 10432) (T_STATIC 10433) (T_STRING 10434) (T_SWITCH 10435) (T_THROW 10436) (T_TRAIT 10437) (T_TRAIT_C 10438) (T_TRY 10439) (T_UNSET 10440) (T_USE 10441) (T_VAR 10442) (T_VARIABLE 1060) (T_WHILE 10443) (T_YIELD 10444) (identifier 10445) (member_name 10446) (reserved_non_modifiers 10447) (semi_reserved 10448) (simple_variable 10449) ("{" 10450)) 4454 (("(" 10253) (argument_list 10453)) 4455 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 10451) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 4456 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 10452)) 4457 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (ampersand 10483) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 10484) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4458 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 10482) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4459 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 10481) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4460 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 10480) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4461 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 10479) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4462 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 10478) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4463 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 10477) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4464 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 10476) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4465 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 10475) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4466 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 10474) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4467 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 10473) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4468 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 10472) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4469 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 10471) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4470 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 10470) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4471 (("\"" 1538) ("$" 10262) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 10485) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 10269) ("[" 1560) (array_object_dereferenceable 10486) (callable_expr 10487) (callable_variable 10488) (class_constant 1564) (class_name 10489) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 10490) (name 10491) (new_dereferenceable 1570) (simple_variable 10492) (static_member 10493) (variable 10494) (variable_class_name 10495)) 4472 (("(" 10292) (argument_list 10515)) 4473 ((T_NULLSAFE_OBJECT_OPERATOR 10506) (T_OBJECT_OPERATOR 10507) ("[" 10508)) 4474 (("(" 10292) (argument_list 10505)) 4475 ((T_PAAMAYIM_NEKUDOTAYIM 10501)) 4476 (("(" 10292) (argument_list 10500)) 4477 ((T_PAAMAYIM_NEKUDOTAYIM 10496)) 4478 (("$" 10262) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 10269) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 10497) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 10498) ("{" 1660)) 4479 (("(" 10292) (argument_list 10499)) 4480 (("$" 10262) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 10269) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 10502) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 10503) ("{" 1676)) 4481 (("(" 10292) (argument_list 10504)) 4482 (("$" 10262) (T_STRING 10285) (T_VARIABLE 10269) (property_name 10513) (simple_variable 10287) ("{" 10288)) 4483 (("$" 10262) (T_STRING 10285) (T_VARIABLE 10269) (property_name 10511) (simple_variable 10287) ("{" 10288)) 4484 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 10509) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 4485 (("]" 10510)) 4486 (("(" 10292) (argument_list 10512)) 4487 (("(" 10292) (argument_list 10514)) 4488 ((")" 10517)) 4489 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 10520) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4490 ((backup_doc_comment 10522)) 4491 (("(" 10523)) 4492 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 10524)) 4493 ((")" 10525)) 4494 ((T_USE 166) (lexical_vars 10526)) 4495 ((":" 168) (return_type 10527)) 4496 ((backup_fn_flags 10528)) 4497 (("{" 10529)) 4498 ((inner_statement_list 10530)) 4499 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 10531) ("~" 129)) 4500 ((backup_fn_flags 10532)) 4501 ((backup_doc_comment 10534)) 4502 (("(" 10535)) 4503 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 10536)) 4504 ((")" 10537)) 4505 ((":" 2271) (return_type 10538)) 4506 ((T_DOUBLE_ARROW 10539)) 4507 ((backup_fn_flags 10540)) 4508 ((backup_lex_pos 10541)) 4509 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 10542) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4510 (("%" 1125) ("*" 1126) ("+" 1127) ("-" 1128) ("." 1129) ("/" 1130) ("<" 1131) (">" 1132) ("?" 1133) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1134) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1135) (T_BOOLEAN_AND 1136) (T_BOOLEAN_OR 1137) (T_COALESCE 1138) (T_INSTANCEOF 1139) (T_IS_EQUAL 1140) (T_IS_GREATER_OR_EQUAL 1141) (T_IS_IDENTICAL 1142) (T_IS_NOT_EQUAL 1143) (T_IS_NOT_IDENTICAL 1144) (T_IS_SMALLER_OR_EQUAL 1145) (T_LOGICAL_AND 1146) (T_LOGICAL_OR 1147) (T_LOGICAL_XOR 1148) (T_PIPE 1149) (T_POW 1150) (T_SL 1151) (T_SPACESHIP 1152) (T_SR 1153) ("^" 1154) (backup_fn_flags 10543) ("|" 1155)) 4511 (("$" 960) (T_ABSTRACT 10366) (T_ARRAY 10367) (T_AS 10368) (T_BREAK 10369) (T_CALLABLE 10370) (T_CASE 10371) (T_CATCH 10372) (T_CLASS 10373) (T_CLASS_C 10374) (T_CLONE 10375) (T_CONST 10376) (T_CONTINUE 10377) (T_DECLARE 10378) (T_DEFAULT 10379) (T_DIR 10380) (T_DO 10381) (T_ECHO 10382) (T_ELSE 10383) (T_ELSEIF 10384) (T_EMPTY 10385) (T_ENDDECLARE 10386) (T_ENDFOR 10387) (T_ENDFOREACH 10388) (T_ENDIF 10389) (T_ENDSWITCH 10390) (T_ENDWHILE 10391) (T_ENUM 10392) (T_EVAL 10393) (T_EXIT 10394) (T_EXTENDS 10395) (T_FILE 10396) (T_FINAL 10397) (T_FINALLY 10398) (T_FN 10399) (T_FOR 10400) (T_FOREACH 10401) (T_FUNCTION 10402) (T_FUNC_C 10403) (T_GLOBAL 10404) (T_GOTO 10405) (T_IF 10406) (T_IMPLEMENTS 10407) (T_INCLUDE 10408) (T_INCLUDE_ONCE 10409) (T_INSTANCEOF 10410) (T_INSTEADOF 10411) (T_INTERFACE 10412) (T_ISSET 10413) (T_LINE 10414) (T_LIST 10415) (T_LOGICAL_AND 10416) (T_LOGICAL_OR 10417) (T_LOGICAL_XOR 10418) (T_MATCH 10419) (T_METHOD_C 10420) (T_NAMESPACE 10421) (T_NEW 10422) (T_NS_C 10423) (T_PRINT 10424) (T_PRIVATE 10425) (T_PROPERTY_C 10426) (T_PROTECTED 10427) (T_PUBLIC 10428) (T_READONLY 10429) (T_REQUIRE 10430) (T_REQUIRE_ONCE 10431) (T_RETURN 10432) (T_STATIC 10433) (T_STRING 10434) (T_SWITCH 10435) (T_THROW 10436) (T_TRAIT 10437) (T_TRAIT_C 10438) (T_TRY 10439) (T_UNSET 10440) (T_USE 10441) (T_VAR 10442) (T_VARIABLE 1060) (T_WHILE 10443) (T_YIELD 10444) (identifier 10545) (member_name 10546) (reserved_non_modifiers 10447) (semi_reserved 10448) (simple_variable 10547) ("{" 10548)) 4512 (("(" 10253) (argument_list 10551)) 4513 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 10549) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 4514 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 10550)) 4515 ((T_FN 35) (T_FUNCTION 131) (fn 1077) #'1078 (inline_function 10555)) 4516 (("$" 960) (T_STRING 10561) (T_VARIABLE 1060) (property_name 10568) (simple_variable 10563) ("{" 10564)) 4517 (("$" 960) (T_STRING 10561) (T_VARIABLE 1060) (property_name 10562) (simple_variable 10563) ("{" 10564)) 4518 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 10559) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 4519 (("]" 10560)) 4520 (("(" 10253) (argument_list 10567)) 4521 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 10565) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 4522 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 10566)) 4523 (("(" 10253) (argument_list 10569)) 4524 (("`" 10571)) 4525 (("]" 10573)) 4526 (("=" 10574)) 4527 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 10575) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4528 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 10583) (T_VARIABLE 2920) (encaps_var 2934)) 4529 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 10582) (T_VARIABLE 2920) (encaps_var 2925)) 4530 (("%" 304) (")" 10358) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 4531 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 10598) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 4532 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 10599)) 4533 (("\"" 10601) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 4534 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (ampersand 10632) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 10633) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 4535 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 10631) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 4536 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 10630) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 4537 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 10629) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 4538 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 10628) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 4539 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 10627) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 4540 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 10626) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 4541 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 10625) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 4542 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 10624) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 4543 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 10623) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 4544 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 10622) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 4545 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 10621) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 4546 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 10620) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 4547 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 10619) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 4548 (("\"" 1538) ("$" 10634) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 10635) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 10636) ("[" 1560) (array_object_dereferenceable 10637) (callable_expr 10638) (callable_variable 10639) (class_constant 1564) (class_name 10640) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 10641) (name 10642) (new_dereferenceable 1570) (simple_variable 10643) (static_member 10644) (variable 10645) (variable_class_name 10646)) 4549 (("$" 10634) (T_VARIABLE 10636) (simple_variable 10679) ("{" 10680)) 4550 (("(" 10650) (argument_list 10678)) 4551 ((T_NULLSAFE_OBJECT_OPERATOR 10664) (T_OBJECT_OPERATOR 10665) ("[" 10666)) 4552 (("(" 10650) (argument_list 10663)) 4553 ((T_PAAMAYIM_NEKUDOTAYIM 10659)) 4554 (("(" 10650) (argument_list 10658)) 4555 ((T_PAAMAYIM_NEKUDOTAYIM 10647)) 4556 (("$" 10634) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 10636) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 10648) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 10649) ("{" 1660)) 4557 (("(" 10650) (argument_list 10651)) 4558 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 10652) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 10653) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 10654) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4559 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 10657) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4560 (("," 1498) (possible_comma 10655)) 4561 ((")" 10656)) 4562 (("$" 10634) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 10636) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 10660) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 10661) ("{" 1676)) 4563 (("(" 10650) (argument_list 10662)) 4564 (("$" 10634) (T_STRING 10669) (T_VARIABLE 10636) (property_name 10676) (simple_variable 10671) ("{" 10672)) 4565 (("$" 10634) (T_STRING 10669) (T_VARIABLE 10636) (property_name 10670) (simple_variable 10671) ("{" 10672)) 4566 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 10667) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 4567 (("]" 10668)) 4568 (("(" 10650) (argument_list 10675)) 4569 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 10673) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 4570 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 10674)) 4571 (("(" 10650) (argument_list 10677)) 4572 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 10681) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 4573 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 10682)) 4574 ((backup_doc_comment 10685)) 4575 (("(" 10686)) 4576 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 10687)) 4577 ((")" 10688)) 4578 ((T_USE 166) (lexical_vars 10689)) 4579 ((":" 168) (return_type 10690)) 4580 ((backup_fn_flags 10691)) 4581 (("{" 10692)) 4582 ((inner_statement_list 10693)) 4583 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 10694) ("~" 129)) 4584 ((backup_fn_flags 10695)) 4585 ((backup_doc_comment 10697)) 4586 (("(" 10698)) 4587 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 10699)) 4588 ((")" 10700)) 4589 ((":" 2271) (return_type 10701)) 4590 ((T_DOUBLE_ARROW 10702)) 4591 ((backup_fn_flags 10703)) 4592 ((backup_lex_pos 10704)) 4593 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 10768) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4594 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 11768) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4595 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 11766) (encaps_var 5357)) 4596 (("$" 10707) (T_VARIABLE 10755) (simple_variable 11762) ("{" 11763)) 4597 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 11761) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 4598 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 11760) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4599 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 11759) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4600 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 11758) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4601 (("(" 11755)) 4602 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 11754) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4603 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 11753) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4604 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 11740) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (clone_argument_list 11741) (constant 10766) (dereferenceable_scalar 10767) (expr 11742) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4605 (("\"" 1538) ("$" 11012) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 11013) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 11014) ("[" 1560) (array_object_dereferenceable 11015) (callable_expr 11016) (callable_variable 11017) (class_constant 1564) (class_name 11018) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 11019) (name 11020) (new_dereferenceable 1570) (simple_variable 11021) (static_member 11022) (variable 11739) (variable_class_name 11024)) 4606 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 11738) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4607 (("(" 11735)) 4608 (("(" 11732)) 4609 (("(" 11723) (argument_list 11724) (ctor_arguments 11725)) 4610 (("\"" 1538) ("$" 11012) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 11013) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 11014) ("[" 1560) (array_object_dereferenceable 11015) (callable_expr 11016) (callable_variable 11017) (class_constant 1564) (class_name 11018) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 11019) (name 11020) (new_dereferenceable 1570) (simple_variable 11021) (static_member 11022) (variable 11722) (variable_class_name 11024)) 4611 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 11721) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4612 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 11720) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4613 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 11719) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4614 (("(" 11715)) 4615 (("(" 11710)) 4616 (("(" 11704)) 4617 (("$" 11012) ("(" 11667) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 11668) (T_NAME_QUALIFIED 11669) (T_NAME_RELATIVE 11670) (T_READONLY 2955) (T_STATIC 11671) (T_STRING 11672) (T_VARIABLE 11014) (anonymous_class 11673) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 11674) (attribute 2961) (attributes 11675) (class_modifier 2963) (class_name 11676) (class_name_reference 11677) (name 11678) (new_variable 11679) (simple_variable 11680)) 4618 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 11666) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4619 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 11665) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4620 (("(" 10973) (argument_list 11664)) 4621 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 11663) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4622 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 11662) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4623 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 11657) (T_END_HEREDOC 11658) (T_VARIABLE 2920) (encaps_list 11659) (encaps_var 2922)) 4624 ((T_FN 35) (T_FUNCTION 131) (fn 10769) #'10770 (inline_function 11656)) 4625 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 11655) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4626 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 11654) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4627 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 11653) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4628 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11182) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4629 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 11118) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4630 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 11114) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 4631 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 11112) (encaps_list 2828) (encaps_var 2829)) 4632 ((T_NULLSAFE_OBJECT_OPERATOR 11098) (T_OBJECT_OPERATOR 11099) ("[" 11100)) 4633 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 11095) (attribute 2757) (fn 10769) #'10770 (inline_function 11096)) 4634 (("(" 10973) (argument_list 11094)) 4635 ((T_PAAMAYIM_NEKUDOTAYIM 11086)) 4636 (("%" 10785) ("*" 10786) ("+" 10787) ("-" 10788) ("." 10789) ("/" 10790) ("<" 10791) (">" 10792) ("?" 10793) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 10794) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 10795) (T_BOOLEAN_AND 10796) (T_BOOLEAN_OR 10797) (T_COALESCE 10798) (T_INSTANCEOF 10799) (T_IS_EQUAL 10800) (T_IS_GREATER_OR_EQUAL 10801) (T_IS_IDENTICAL 10802) (T_IS_NOT_EQUAL 10803) (T_IS_NOT_IDENTICAL 10804) (T_IS_SMALLER_OR_EQUAL 10805) (T_LOGICAL_AND 10806) (T_LOGICAL_OR 10807) (T_LOGICAL_XOR 10808) (T_PIPE 10809) (T_POW 10810) (T_SL 10811) (T_SPACESHIP 10812) (T_SR 10813) ("^" 10814) (backup_fn_flags 11085) ("|" 10815)) 4637 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 11074)) 4638 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 11062)) 4639 (("(" 10973) (argument_list 11061)) 4640 (("=" 10981) (T_AND_EQUAL 10982) (T_COALESCE_EQUAL 10983) (T_CONCAT_EQUAL 10984) (T_DEC 10985) (T_DIV_EQUAL 10986) (T_INC 10987) (T_MINUS_EQUAL 10988) (T_MOD_EQUAL 10989) (T_MUL_EQUAL 10990) (T_OR_EQUAL 10991) (T_PLUS_EQUAL 10992) (T_POW_EQUAL 10993) (T_SL_EQUAL 10994) (T_SR_EQUAL 10995) (T_XOR_EQUAL 10996)) 4641 ((T_PAAMAYIM_NEKUDOTAYIM 10885)) 4642 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 10784) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4643 (("%" 10785) ("*" 10786) ("+" 10787) ("-" 10788) ("." 10789) ("/" 10790) ("<" 10791) (">" 10792) ("?" 10793) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 10794) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 10795) (T_BOOLEAN_AND 10796) (T_BOOLEAN_OR 10797) (T_COALESCE 10798) (T_INSTANCEOF 10799) (T_IS_EQUAL 10800) (T_IS_GREATER_OR_EQUAL 10801) (T_IS_IDENTICAL 10802) (T_IS_NOT_EQUAL 10803) (T_IS_NOT_IDENTICAL 10804) (T_IS_SMALLER_OR_EQUAL 10805) (T_LOGICAL_AND 10806) (T_LOGICAL_OR 10807) (T_LOGICAL_XOR 10808) (T_PIPE 10809) (T_POW 10810) (T_SL 10811) (T_SPACESHIP 10812) (T_SR 10813) ("^" 10814) ("|" 10815)) 4644 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 10884) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4645 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 10883) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4646 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 10882) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4647 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 10881) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4648 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 10880) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4649 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 10879) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4650 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 10878) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4651 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 10877) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4652 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 10872) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 10873) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 4653 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 10871) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4654 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 10870) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4655 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 10869) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4656 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 10868) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4657 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 10867) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4658 (("$" 10832) ("(" 10833) (T_NAME_FULLY_QUALIFIED 10834) (T_NAME_QUALIFIED 10835) (T_NAME_RELATIVE 10836) (T_STATIC 10837) (T_STRING 10838) (T_VARIABLE 10839) (class_name 10840) (class_name_reference 10841) (name 10842) (new_variable 10843) (simple_variable 10844)) 4659 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 10831) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4660 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 10830) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4661 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 10829) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4662 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 10828) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4663 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 10827) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4664 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 10826) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4665 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 10825) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4666 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 10824) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4667 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 10823) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4668 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 10822) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4669 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 10821) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4670 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 10820) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4671 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 10819) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4672 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 10818) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4673 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 10817) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4674 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 10816) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4675 (("$" 10832) (T_VARIABLE 10839) (simple_variable 10863) ("{" 10864)) 4676 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 10861) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 4677 ((T_PAAMAYIM_NEKUDOTAYIM 10859)) 4678 ((T_NULLSAFE_OBJECT_OPERATOR 10845) (T_OBJECT_OPERATOR 10846) (T_PAAMAYIM_NEKUDOTAYIM 10847) ("[" 10848)) 4679 (("$" 10832) (T_STRING 10852) (T_VARIABLE 10839) (property_name 10858) (simple_variable 10854) ("{" 10855)) 4680 (("$" 10832) (T_STRING 10852) (T_VARIABLE 10839) (property_name 10853) (simple_variable 10854) ("{" 10855)) 4681 (("$" 10832) (T_VARIABLE 10839) (simple_variable 10851)) 4682 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 10849) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 4683 (("]" 10850)) 4684 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 10856) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 4685 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 10857)) 4686 (("$" 10832) (T_VARIABLE 10839) (simple_variable 10860)) 4687 (("%" 304) (")" 10862) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 4688 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 10865) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 4689 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 10866)) 4690 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 10876) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4691 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 10874) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 4692 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 10875) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4693 (("$" 10707) (T_ABSTRACT 10886) (T_ARRAY 10887) (T_AS 10888) (T_BREAK 10889) (T_CALLABLE 10890) (T_CASE 10891) (T_CATCH 10892) (T_CLASS 10893) (T_CLASS_C 10894) (T_CLONE 10895) (T_CONST 10896) (T_CONTINUE 10897) (T_DECLARE 10898) (T_DEFAULT 10899) (T_DIR 10900) (T_DO 10901) (T_ECHO 10902) (T_ELSE 10903) (T_ELSEIF 10904) (T_EMPTY 10905) (T_ENDDECLARE 10906) (T_ENDFOR 10907) (T_ENDFOREACH 10908) (T_ENDIF 10909) (T_ENDSWITCH 10910) (T_ENDWHILE 10911) (T_ENUM 10912) (T_EVAL 10913) (T_EXIT 10914) (T_EXTENDS 10915) (T_FILE 10916) (T_FINAL 10917) (T_FINALLY 10918) (T_FN 10919) (T_FOR 10920) (T_FOREACH 10921) (T_FUNCTION 10922) (T_FUNC_C 10923) (T_GLOBAL 10924) (T_GOTO 10925) (T_IF 10926) (T_IMPLEMENTS 10927) (T_INCLUDE 10928) (T_INCLUDE_ONCE 10929) (T_INSTANCEOF 10930) (T_INSTEADOF 10931) (T_INTERFACE 10932) (T_ISSET 10933) (T_LINE 10934) (T_LIST 10935) (T_LOGICAL_AND 10936) (T_LOGICAL_OR 10937) (T_LOGICAL_XOR 10938) (T_MATCH 10939) (T_METHOD_C 10940) (T_NAMESPACE 10941) (T_NEW 10942) (T_NS_C 10943) (T_PRINT 10944) (T_PRIVATE 10945) (T_PROPERTY_C 10946) (T_PROTECTED 10947) (T_PUBLIC 10948) (T_READONLY 10949) (T_REQUIRE 10950) (T_REQUIRE_ONCE 10951) (T_RETURN 10952) (T_STATIC 10953) (T_STRING 10954) (T_SWITCH 10955) (T_THROW 10956) (T_TRAIT 10957) (T_TRAIT_C 10958) (T_TRY 10959) (T_UNSET 10960) (T_USE 10961) (T_VAR 10962) (T_VARIABLE 10755) (T_WHILE 10963) (T_YIELD 10964) (identifier 10965) (member_name 10966) (reserved_non_modifiers 10967) (semi_reserved 10968) (simple_variable 10969) ("{" 10970)) 4694 (("(" 10973) (argument_list 10974)) 4695 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 10971) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 4696 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 10972)) 4697 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 10975) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 10976) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 10977) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4698 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 10980) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4699 (("," 1498) (possible_comma 10978)) 4700 ((")" 10979)) 4701 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (ampersand 11010) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 11011) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4702 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 11009) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4703 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 11008) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4704 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 11007) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4705 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 11006) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4706 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 11005) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4707 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 11004) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4708 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 11003) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4709 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 11002) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4710 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 11001) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4711 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 11000) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4712 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 10999) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4713 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 10998) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4714 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 10997) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4715 (("\"" 1538) ("$" 11012) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 11013) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 11014) ("[" 1560) (array_object_dereferenceable 11015) (callable_expr 11016) (callable_variable 11017) (class_constant 1564) (class_name 11018) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 11019) (name 11020) (new_dereferenceable 1570) (simple_variable 11021) (static_member 11022) (variable 11023) (variable_class_name 11024)) 4716 (("$" 11012) (T_VARIABLE 11014) (simple_variable 11057) ("{" 11058)) 4717 (("(" 11028) (argument_list 11056)) 4718 ((T_NULLSAFE_OBJECT_OPERATOR 11042) (T_OBJECT_OPERATOR 11043) ("[" 11044)) 4719 (("(" 11028) (argument_list 11041)) 4720 ((T_PAAMAYIM_NEKUDOTAYIM 11037)) 4721 (("(" 11028) (argument_list 11036)) 4722 ((T_PAAMAYIM_NEKUDOTAYIM 11025)) 4723 (("$" 11012) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 11014) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 11026) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 11027) ("{" 1660)) 4724 (("(" 11028) (argument_list 11029)) 4725 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 11030) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 11031) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 11032) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4726 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 11035) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4727 (("," 1498) (possible_comma 11033)) 4728 ((")" 11034)) 4729 (("$" 11012) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 11014) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 11038) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 11039) ("{" 1676)) 4730 (("(" 11028) (argument_list 11040)) 4731 (("$" 11012) (T_STRING 11047) (T_VARIABLE 11014) (property_name 11054) (simple_variable 11049) ("{" 11050)) 4732 (("$" 11012) (T_STRING 11047) (T_VARIABLE 11014) (property_name 11048) (simple_variable 11049) ("{" 11050)) 4733 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 11045) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 4734 (("]" 11046)) 4735 (("(" 11028) (argument_list 11053)) 4736 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 11051) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 4737 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 11052)) 4738 (("(" 11028) (argument_list 11055)) 4739 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 11059) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 4740 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 11060)) 4741 ((backup_doc_comment 11063)) 4742 (("(" 11064)) 4743 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 11065)) 4744 ((")" 11066)) 4745 ((T_USE 166) (lexical_vars 11067)) 4746 ((":" 168) (return_type 11068)) 4747 ((backup_fn_flags 11069)) 4748 (("{" 11070)) 4749 ((inner_statement_list 11071)) 4750 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 11072) ("~" 129)) 4751 ((backup_fn_flags 11073)) 4752 ((backup_doc_comment 11075)) 4753 (("(" 11076)) 4754 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 11077)) 4755 ((")" 11078)) 4756 ((":" 2271) (return_type 11079)) 4757 ((T_DOUBLE_ARROW 11080)) 4758 ((backup_fn_flags 11081)) 4759 ((backup_lex_pos 11082)) 4760 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 11083) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4761 (("%" 10785) ("*" 10786) ("+" 10787) ("-" 10788) ("." 10789) ("/" 10790) ("<" 10791) (">" 10792) ("?" 10793) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 10794) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 10795) (T_BOOLEAN_AND 10796) (T_BOOLEAN_OR 10797) (T_COALESCE 10798) (T_INSTANCEOF 10799) (T_IS_EQUAL 10800) (T_IS_GREATER_OR_EQUAL 10801) (T_IS_IDENTICAL 10802) (T_IS_NOT_EQUAL 10803) (T_IS_NOT_IDENTICAL 10804) (T_IS_SMALLER_OR_EQUAL 10805) (T_LOGICAL_AND 10806) (T_LOGICAL_OR 10807) (T_LOGICAL_XOR 10808) (T_PIPE 10809) (T_POW 10810) (T_SL 10811) (T_SPACESHIP 10812) (T_SR 10813) ("^" 10814) (backup_fn_flags 11084) ("|" 10815)) 4762 (("$" 10707) (T_ABSTRACT 10886) (T_ARRAY 10887) (T_AS 10888) (T_BREAK 10889) (T_CALLABLE 10890) (T_CASE 10891) (T_CATCH 10892) (T_CLASS 10893) (T_CLASS_C 10894) (T_CLONE 10895) (T_CONST 10896) (T_CONTINUE 10897) (T_DECLARE 10898) (T_DEFAULT 10899) (T_DIR 10900) (T_DO 10901) (T_ECHO 10902) (T_ELSE 10903) (T_ELSEIF 10904) (T_EMPTY 10905) (T_ENDDECLARE 10906) (T_ENDFOR 10907) (T_ENDFOREACH 10908) (T_ENDIF 10909) (T_ENDSWITCH 10910) (T_ENDWHILE 10911) (T_ENUM 10912) (T_EVAL 10913) (T_EXIT 10914) (T_EXTENDS 10915) (T_FILE 10916) (T_FINAL 10917) (T_FINALLY 10918) (T_FN 10919) (T_FOR 10920) (T_FOREACH 10921) (T_FUNCTION 10922) (T_FUNC_C 10923) (T_GLOBAL 10924) (T_GOTO 10925) (T_IF 10926) (T_IMPLEMENTS 10927) (T_INCLUDE 10928) (T_INCLUDE_ONCE 10929) (T_INSTANCEOF 10930) (T_INSTEADOF 10931) (T_INTERFACE 10932) (T_ISSET 10933) (T_LINE 10934) (T_LIST 10935) (T_LOGICAL_AND 10936) (T_LOGICAL_OR 10937) (T_LOGICAL_XOR 10938) (T_MATCH 10939) (T_METHOD_C 10940) (T_NAMESPACE 10941) (T_NEW 10942) (T_NS_C 10943) (T_PRINT 10944) (T_PRIVATE 10945) (T_PROPERTY_C 10946) (T_PROTECTED 10947) (T_PUBLIC 10948) (T_READONLY 10949) (T_REQUIRE 10950) (T_REQUIRE_ONCE 10951) (T_RETURN 10952) (T_STATIC 10953) (T_STRING 10954) (T_SWITCH 10955) (T_THROW 10956) (T_TRAIT 10957) (T_TRAIT_C 10958) (T_TRY 10959) (T_UNSET 10960) (T_USE 10961) (T_VAR 10962) (T_VARIABLE 10755) (T_WHILE 10963) (T_YIELD 10964) (identifier 11087) (member_name 11088) (reserved_non_modifiers 10967) (semi_reserved 10968) (simple_variable 11089) ("{" 11090)) 4763 (("(" 10973) (argument_list 11093)) 4764 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 11091) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 4765 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 11092)) 4766 ((T_FN 35) (T_FUNCTION 131) (fn 10769) #'10770 (inline_function 11097)) 4767 (("$" 10707) (T_STRING 11103) (T_VARIABLE 10755) (property_name 11110) (simple_variable 11105) ("{" 11106)) 4768 (("$" 10707) (T_STRING 11103) (T_VARIABLE 10755) (property_name 11104) (simple_variable 11105) ("{" 11106)) 4769 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 11101) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 4770 (("]" 11102)) 4771 (("(" 10973) (argument_list 11109)) 4772 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 11107) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 4773 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 11108)) 4774 (("(" 10973) (argument_list 11111)) 4775 (("`" 11113)) 4776 (("]" 11115)) 4777 (("=" 11116)) 4778 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 11117) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4779 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11652) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4780 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 11650) (encaps_var 5357)) 4781 (("$" 11121) (T_VARIABLE 11169) (simple_variable 11646) ("{" 11647)) 4782 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 11645) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 4783 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11644) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4784 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11643) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4785 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11642) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4786 (("(" 11639)) 4787 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11638) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4788 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11637) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4789 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11624) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (clone_argument_list 11625) (constant 11180) (dereferenceable_scalar 11181) (expr 11626) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4790 (("\"" 1538) ("$" 11426) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 11427) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 11428) ("[" 1560) (array_object_dereferenceable 11429) (callable_expr 11430) (callable_variable 11431) (class_constant 1564) (class_name 11432) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 11433) (name 11434) (new_dereferenceable 1570) (simple_variable 11435) (static_member 11436) (variable 11623) (variable_class_name 11438)) 4791 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11622) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4792 (("(" 11619)) 4793 (("(" 11616)) 4794 (("(" 11607) (argument_list 11608) (ctor_arguments 11609)) 4795 (("\"" 1538) ("$" 11426) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 11427) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 11428) ("[" 1560) (array_object_dereferenceable 11429) (callable_expr 11430) (callable_variable 11431) (class_constant 1564) (class_name 11432) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 11433) (name 11434) (new_dereferenceable 1570) (simple_variable 11435) (static_member 11436) (variable 11606) (variable_class_name 11438)) 4796 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11605) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4797 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11604) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4798 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11603) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4799 (("(" 11599)) 4800 (("(" 11594)) 4801 (("(" 11588)) 4802 (("$" 11426) ("(" 11551) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 11552) (T_NAME_QUALIFIED 11553) (T_NAME_RELATIVE 11554) (T_READONLY 2955) (T_STATIC 11555) (T_STRING 11556) (T_VARIABLE 11428) (anonymous_class 11557) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 11558) (attribute 2961) (attributes 11559) (class_modifier 2963) (class_name 11560) (class_name_reference 11561) (name 11562) (new_variable 11563) (simple_variable 11564)) 4803 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11550) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4804 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11549) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4805 (("(" 11387) (argument_list 11548)) 4806 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11547) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4807 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11546) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4808 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 11541) (T_END_HEREDOC 11542) (T_VARIABLE 2920) (encaps_list 11543) (encaps_var 2922)) 4809 ((T_FN 35) (T_FUNCTION 131) (fn 11183) #'11184 (inline_function 11540)) 4810 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11539) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4811 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11538) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4812 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11537) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4813 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11534) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4814 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11533) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4815 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 11529) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 4816 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 11527) (encaps_list 2828) (encaps_var 2829)) 4817 ((T_NULLSAFE_OBJECT_OPERATOR 11513) (T_OBJECT_OPERATOR 11514) ("[" 11515)) 4818 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 11510) (attribute 2757) (fn 11183) #'11184 (inline_function 11511)) 4819 (("(" 11387) (argument_list 11509)) 4820 ((T_PAAMAYIM_NEKUDOTAYIM 11501)) 4821 (("%" 11199) ("*" 11200) ("+" 11201) ("-" 11202) ("." 11203) ("/" 11204) ("<" 11205) (">" 11206) ("?" 11207) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 11208) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 11209) (T_BOOLEAN_AND 11210) (T_BOOLEAN_OR 11211) (T_COALESCE 11212) (T_DOUBLE_ARROW 11499) (T_INSTANCEOF 11213) (T_IS_EQUAL 11214) (T_IS_GREATER_OR_EQUAL 11215) (T_IS_IDENTICAL 11216) (T_IS_NOT_EQUAL 11217) (T_IS_NOT_IDENTICAL 11218) (T_IS_SMALLER_OR_EQUAL 11219) (T_LOGICAL_AND 11220) (T_LOGICAL_OR 11221) (T_LOGICAL_XOR 11222) (T_PIPE 11223) (T_POW 11224) (T_SL 11225) (T_SPACESHIP 11226) (T_SR 11227) ("^" 11228) ("|" 11229)) 4822 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 11488)) 4823 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 11476)) 4824 (("(" 11387) (argument_list 11475)) 4825 (("=" 11395) (T_AND_EQUAL 11396) (T_COALESCE_EQUAL 11397) (T_CONCAT_EQUAL 11398) (T_DEC 11399) (T_DIV_EQUAL 11400) (T_INC 11401) (T_MINUS_EQUAL 11402) (T_MOD_EQUAL 11403) (T_MUL_EQUAL 11404) (T_OR_EQUAL 11405) (T_PLUS_EQUAL 11406) (T_POW_EQUAL 11407) (T_SL_EQUAL 11408) (T_SR_EQUAL 11409) (T_XOR_EQUAL 11410)) 4826 ((T_PAAMAYIM_NEKUDOTAYIM 11299)) 4827 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11198) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4828 (("%" 11199) ("*" 11200) ("+" 11201) ("-" 11202) ("." 11203) ("/" 11204) ("<" 11205) (">" 11206) ("?" 11207) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 11208) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 11209) (T_BOOLEAN_AND 11210) (T_BOOLEAN_OR 11211) (T_COALESCE 11212) (T_INSTANCEOF 11213) (T_IS_EQUAL 11214) (T_IS_GREATER_OR_EQUAL 11215) (T_IS_IDENTICAL 11216) (T_IS_NOT_EQUAL 11217) (T_IS_NOT_IDENTICAL 11218) (T_IS_SMALLER_OR_EQUAL 11219) (T_LOGICAL_AND 11220) (T_LOGICAL_OR 11221) (T_LOGICAL_XOR 11222) (T_PIPE 11223) (T_POW 11224) (T_SL 11225) (T_SPACESHIP 11226) (T_SR 11227) ("^" 11228) ("|" 11229)) 4829 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11298) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4830 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11297) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4831 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11296) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4832 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11295) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4833 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11294) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4834 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11293) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4835 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11292) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4836 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11291) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4837 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 11286) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 11287) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 4838 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11285) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4839 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11284) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4840 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11283) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4841 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11282) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4842 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11281) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4843 (("$" 11246) ("(" 11247) (T_NAME_FULLY_QUALIFIED 11248) (T_NAME_QUALIFIED 11249) (T_NAME_RELATIVE 11250) (T_STATIC 11251) (T_STRING 11252) (T_VARIABLE 11253) (class_name 11254) (class_name_reference 11255) (name 11256) (new_variable 11257) (simple_variable 11258)) 4844 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11245) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4845 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11244) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4846 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11243) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4847 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11242) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4848 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11241) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4849 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11240) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4850 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11239) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4851 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11238) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4852 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11237) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4853 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11236) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4854 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11235) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4855 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11234) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4856 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11233) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4857 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11232) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4858 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11231) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4859 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11230) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4860 (("$" 11246) (T_VARIABLE 11253) (simple_variable 11277) ("{" 11278)) 4861 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 11275) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 4862 ((T_PAAMAYIM_NEKUDOTAYIM 11273)) 4863 ((T_NULLSAFE_OBJECT_OPERATOR 11259) (T_OBJECT_OPERATOR 11260) (T_PAAMAYIM_NEKUDOTAYIM 11261) ("[" 11262)) 4864 (("$" 11246) (T_STRING 11266) (T_VARIABLE 11253) (property_name 11272) (simple_variable 11268) ("{" 11269)) 4865 (("$" 11246) (T_STRING 11266) (T_VARIABLE 11253) (property_name 11267) (simple_variable 11268) ("{" 11269)) 4866 (("$" 11246) (T_VARIABLE 11253) (simple_variable 11265)) 4867 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 11263) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 4868 (("]" 11264)) 4869 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 11270) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 4870 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 11271)) 4871 (("$" 11246) (T_VARIABLE 11253) (simple_variable 11274)) 4872 (("%" 304) (")" 11276) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 4873 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 11279) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 4874 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 11280)) 4875 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11290) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4876 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 11288) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 4877 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11289) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4878 (("$" 11121) (T_ABSTRACT 11300) (T_ARRAY 11301) (T_AS 11302) (T_BREAK 11303) (T_CALLABLE 11304) (T_CASE 11305) (T_CATCH 11306) (T_CLASS 11307) (T_CLASS_C 11308) (T_CLONE 11309) (T_CONST 11310) (T_CONTINUE 11311) (T_DECLARE 11312) (T_DEFAULT 11313) (T_DIR 11314) (T_DO 11315) (T_ECHO 11316) (T_ELSE 11317) (T_ELSEIF 11318) (T_EMPTY 11319) (T_ENDDECLARE 11320) (T_ENDFOR 11321) (T_ENDFOREACH 11322) (T_ENDIF 11323) (T_ENDSWITCH 11324) (T_ENDWHILE 11325) (T_ENUM 11326) (T_EVAL 11327) (T_EXIT 11328) (T_EXTENDS 11329) (T_FILE 11330) (T_FINAL 11331) (T_FINALLY 11332) (T_FN 11333) (T_FOR 11334) (T_FOREACH 11335) (T_FUNCTION 11336) (T_FUNC_C 11337) (T_GLOBAL 11338) (T_GOTO 11339) (T_IF 11340) (T_IMPLEMENTS 11341) (T_INCLUDE 11342) (T_INCLUDE_ONCE 11343) (T_INSTANCEOF 11344) (T_INSTEADOF 11345) (T_INTERFACE 11346) (T_ISSET 11347) (T_LINE 11348) (T_LIST 11349) (T_LOGICAL_AND 11350) (T_LOGICAL_OR 11351) (T_LOGICAL_XOR 11352) (T_MATCH 11353) (T_METHOD_C 11354) (T_NAMESPACE 11355) (T_NEW 11356) (T_NS_C 11357) (T_PRINT 11358) (T_PRIVATE 11359) (T_PROPERTY_C 11360) (T_PROTECTED 11361) (T_PUBLIC 11362) (T_READONLY 11363) (T_REQUIRE 11364) (T_REQUIRE_ONCE 11365) (T_RETURN 11366) (T_STATIC 11367) (T_STRING 11368) (T_SWITCH 11369) (T_THROW 11370) (T_TRAIT 11371) (T_TRAIT_C 11372) (T_TRY 11373) (T_UNSET 11374) (T_USE 11375) (T_VAR 11376) (T_VARIABLE 11169) (T_WHILE 11377) (T_YIELD 11378) (identifier 11379) (member_name 11380) (reserved_non_modifiers 11381) (semi_reserved 11382) (simple_variable 11383) ("{" 11384)) 4879 (("(" 11387) (argument_list 11388)) 4880 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 11385) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 4881 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 11386)) 4882 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 11389) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 11390) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 11391) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4883 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 11394) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4884 (("," 1498) (possible_comma 11392)) 4885 ((")" 11393)) 4886 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (ampersand 11424) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11425) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4887 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11423) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4888 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11422) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4889 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11421) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4890 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11420) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4891 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11419) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4892 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11418) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4893 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11417) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4894 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11416) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4895 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11415) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4896 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11414) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4897 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11413) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4898 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11412) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4899 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11411) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4900 (("\"" 1538) ("$" 11426) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 11427) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 11428) ("[" 1560) (array_object_dereferenceable 11429) (callable_expr 11430) (callable_variable 11431) (class_constant 1564) (class_name 11432) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 11433) (name 11434) (new_dereferenceable 1570) (simple_variable 11435) (static_member 11436) (variable 11437) (variable_class_name 11438)) 4901 (("$" 11426) (T_VARIABLE 11428) (simple_variable 11471) ("{" 11472)) 4902 (("(" 11442) (argument_list 11470)) 4903 ((T_NULLSAFE_OBJECT_OPERATOR 11456) (T_OBJECT_OPERATOR 11457) ("[" 11458)) 4904 (("(" 11442) (argument_list 11455)) 4905 ((T_PAAMAYIM_NEKUDOTAYIM 11451)) 4906 (("(" 11442) (argument_list 11450)) 4907 ((T_PAAMAYIM_NEKUDOTAYIM 11439)) 4908 (("$" 11426) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 11428) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 11440) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 11441) ("{" 1660)) 4909 (("(" 11442) (argument_list 11443)) 4910 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 11444) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 11445) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 11446) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4911 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 11449) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4912 (("," 1498) (possible_comma 11447)) 4913 ((")" 11448)) 4914 (("$" 11426) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 11428) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 11452) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 11453) ("{" 1676)) 4915 (("(" 11442) (argument_list 11454)) 4916 (("$" 11426) (T_STRING 11461) (T_VARIABLE 11428) (property_name 11468) (simple_variable 11463) ("{" 11464)) 4917 (("$" 11426) (T_STRING 11461) (T_VARIABLE 11428) (property_name 11462) (simple_variable 11463) ("{" 11464)) 4918 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 11459) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 4919 (("]" 11460)) 4920 (("(" 11442) (argument_list 11467)) 4921 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 11465) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 4922 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 11466)) 4923 (("(" 11442) (argument_list 11469)) 4924 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 11473) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 4925 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 11474)) 4926 ((backup_doc_comment 11477)) 4927 (("(" 11478)) 4928 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 11479)) 4929 ((")" 11480)) 4930 ((T_USE 166) (lexical_vars 11481)) 4931 ((":" 168) (return_type 11482)) 4932 ((backup_fn_flags 11483)) 4933 (("{" 11484)) 4934 ((inner_statement_list 11485)) 4935 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 11486) ("~" 129)) 4936 ((backup_fn_flags 11487)) 4937 ((backup_doc_comment 11489)) 4938 (("(" 11490)) 4939 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 11491)) 4940 ((")" 11492)) 4941 ((":" 2271) (return_type 11493)) 4942 ((T_DOUBLE_ARROW 11494)) 4943 ((backup_fn_flags 11495)) 4944 ((backup_lex_pos 11496)) 4945 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11497) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4946 (("%" 11199) ("*" 11200) ("+" 11201) ("-" 11202) ("." 11203) ("/" 11204) ("<" 11205) (">" 11206) ("?" 11207) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 11208) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 11209) (T_BOOLEAN_AND 11210) (T_BOOLEAN_OR 11211) (T_COALESCE 11212) (T_INSTANCEOF 11213) (T_IS_EQUAL 11214) (T_IS_GREATER_OR_EQUAL 11215) (T_IS_IDENTICAL 11216) (T_IS_NOT_EQUAL 11217) (T_IS_NOT_IDENTICAL 11218) (T_IS_SMALLER_OR_EQUAL 11219) (T_LOGICAL_AND 11220) (T_LOGICAL_OR 11221) (T_LOGICAL_XOR 11222) (T_PIPE 11223) (T_POW 11224) (T_SL 11225) (T_SPACESHIP 11226) (T_SR 11227) ("^" 11228) (backup_fn_flags 11498) ("|" 11229)) 4947 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 11500) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 4948 (("$" 11121) (T_ABSTRACT 11300) (T_ARRAY 11301) (T_AS 11302) (T_BREAK 11303) (T_CALLABLE 11304) (T_CASE 11305) (T_CATCH 11306) (T_CLASS 11307) (T_CLASS_C 11308) (T_CLONE 11309) (T_CONST 11310) (T_CONTINUE 11311) (T_DECLARE 11312) (T_DEFAULT 11313) (T_DIR 11314) (T_DO 11315) (T_ECHO 11316) (T_ELSE 11317) (T_ELSEIF 11318) (T_EMPTY 11319) (T_ENDDECLARE 11320) (T_ENDFOR 11321) (T_ENDFOREACH 11322) (T_ENDIF 11323) (T_ENDSWITCH 11324) (T_ENDWHILE 11325) (T_ENUM 11326) (T_EVAL 11327) (T_EXIT 11328) (T_EXTENDS 11329) (T_FILE 11330) (T_FINAL 11331) (T_FINALLY 11332) (T_FN 11333) (T_FOR 11334) (T_FOREACH 11335) (T_FUNCTION 11336) (T_FUNC_C 11337) (T_GLOBAL 11338) (T_GOTO 11339) (T_IF 11340) (T_IMPLEMENTS 11341) (T_INCLUDE 11342) (T_INCLUDE_ONCE 11343) (T_INSTANCEOF 11344) (T_INSTEADOF 11345) (T_INTERFACE 11346) (T_ISSET 11347) (T_LINE 11348) (T_LIST 11349) (T_LOGICAL_AND 11350) (T_LOGICAL_OR 11351) (T_LOGICAL_XOR 11352) (T_MATCH 11353) (T_METHOD_C 11354) (T_NAMESPACE 11355) (T_NEW 11356) (T_NS_C 11357) (T_PRINT 11358) (T_PRIVATE 11359) (T_PROPERTY_C 11360) (T_PROTECTED 11361) (T_PUBLIC 11362) (T_READONLY 11363) (T_REQUIRE 11364) (T_REQUIRE_ONCE 11365) (T_RETURN 11366) (T_STATIC 11367) (T_STRING 11368) (T_SWITCH 11369) (T_THROW 11370) (T_TRAIT 11371) (T_TRAIT_C 11372) (T_TRY 11373) (T_UNSET 11374) (T_USE 11375) (T_VAR 11376) (T_VARIABLE 11169) (T_WHILE 11377) (T_YIELD 11378) (identifier 11502) (member_name 11503) (reserved_non_modifiers 11381) (semi_reserved 11382) (simple_variable 11504) ("{" 11505)) 4949 (("(" 11387) (argument_list 11508)) 4950 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 11506) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 4951 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 11507)) 4952 ((T_FN 35) (T_FUNCTION 131) (fn 11183) #'11184 (inline_function 11512)) 4953 (("$" 11121) (T_STRING 11518) (T_VARIABLE 11169) (property_name 11525) (simple_variable 11520) ("{" 11521)) 4954 (("$" 11121) (T_STRING 11518) (T_VARIABLE 11169) (property_name 11519) (simple_variable 11520) ("{" 11521)) 4955 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 11516) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 4956 (("]" 11517)) 4957 (("(" 11387) (argument_list 11524)) 4958 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 11522) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 4959 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 11523)) 4960 (("(" 11387) (argument_list 11526)) 4961 (("`" 11528)) 4962 (("]" 11530)) 4963 (("=" 11531)) 4964 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11532) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4965 (("%" 11199) ("*" 11200) ("+" 11201) ("-" 11202) ("." 11203) ("/" 11204) ("<" 11205) (">" 11206) ("?" 11207) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 11208) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 11209) (T_BOOLEAN_AND 11210) (T_BOOLEAN_OR 11211) (T_COALESCE 11212) (T_DOUBLE_ARROW 11535) (T_INSTANCEOF 11213) (T_IS_EQUAL 11214) (T_IS_GREATER_OR_EQUAL 11215) (T_IS_IDENTICAL 11216) (T_IS_NOT_EQUAL 11217) (T_IS_NOT_IDENTICAL 11218) (T_IS_SMALLER_OR_EQUAL 11219) (T_LOGICAL_AND 11220) (T_LOGICAL_OR 11221) (T_LOGICAL_XOR 11222) (T_PIPE 11223) (T_POW 11224) (T_SL 11225) (T_SPACESHIP 11226) (T_SR 11227) ("^" 11228) ("|" 11229)) 4966 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11536) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4967 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 11545) (T_VARIABLE 2920) (encaps_var 2934)) 4968 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 11544) (T_VARIABLE 2920) (encaps_var 2925)) 4969 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 11586) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 4970 ((T_CLASS 11578)) 4971 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 11577) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 11558) (attribute 2982) (class_modifier 2963)) 4972 ((T_PAAMAYIM_NEKUDOTAYIM 11575)) 4973 (("(" 11442) (argument_list 11574)) 4974 ((T_NULLSAFE_OBJECT_OPERATOR 11565) (T_OBJECT_OPERATOR 11566) (T_PAAMAYIM_NEKUDOTAYIM 11567) ("[" 11568)) 4975 (("$" 11426) (T_STRING 11461) (T_VARIABLE 11428) (property_name 11573) (simple_variable 11463) ("{" 11464)) 4976 (("$" 11426) (T_STRING 11461) (T_VARIABLE 11428) (property_name 11572) (simple_variable 11463) ("{" 11464)) 4977 (("$" 11426) (T_VARIABLE 11428) (simple_variable 11571)) 4978 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 11569) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 4979 (("]" 11570)) 4980 (("$" 11426) (T_VARIABLE 11428) (simple_variable 11576)) 4981 (("(" 2984) (argument_list 2985) (ctor_arguments 11579)) 4982 ((T_EXTENDS 2987) (extends_from 11580)) 4983 ((T_IMPLEMENTS 2989) (implements_list 11581)) 4984 ((backup_doc_comment 11582)) 4985 (("{" 11583)) 4986 ((class_statement_list 11584)) 4987 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 11585)) 4988 (("%" 304) (")" 11587) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 4989 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 11589) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 4990 (("%" 304) (")" 11590) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 4991 (("{" 11591)) 4992 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 11592) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 4993 (("}" 11593)) 4994 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 11595) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 4995 ((")" 11596)) 4996 (("=" 11597)) 4997 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 11598) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 4998 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 11600) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 4999 (("," 5226) (possible_comma 11601)) 5000 ((")" 11602)) 5001 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 11610) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 11611) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 11612) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5002 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 11615) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5003 (("," 1498) (possible_comma 11613)) 5004 ((")" 11614)) 5005 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 11617) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 5006 (("%" 304) (")" 11618) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 5007 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 11620) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 5008 (("%" 304) (")" 11621) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 5009 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 11627) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 11628) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 11629) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 11630) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 5010 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 11636) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5011 (("%" 2442) (")" 11633) ("*" 2443) ("+" 2444) ("," 11634) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 5012 (("," 5287) (possible_comma 11631)) 5013 ((")" 11632)) 5014 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 11635) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5015 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 11640) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 5016 ((")" 11641)) 5017 (("%" 304) (")" 11633) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 5018 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 11648) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5019 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 11649)) 5020 (("\"" 11651) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 5021 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 11661) (T_VARIABLE 2920) (encaps_var 2934)) 5022 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 11660) (T_VARIABLE 2920) (encaps_var 2925)) 5023 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 11702) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 5024 ((T_CLASS 11694)) 5025 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 11693) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 11674) (attribute 2982) (class_modifier 2963)) 5026 ((T_PAAMAYIM_NEKUDOTAYIM 11691)) 5027 (("(" 11028) (argument_list 11690)) 5028 ((T_NULLSAFE_OBJECT_OPERATOR 11681) (T_OBJECT_OPERATOR 11682) (T_PAAMAYIM_NEKUDOTAYIM 11683) ("[" 11684)) 5029 (("$" 11012) (T_STRING 11047) (T_VARIABLE 11014) (property_name 11689) (simple_variable 11049) ("{" 11050)) 5030 (("$" 11012) (T_STRING 11047) (T_VARIABLE 11014) (property_name 11688) (simple_variable 11049) ("{" 11050)) 5031 (("$" 11012) (T_VARIABLE 11014) (simple_variable 11687)) 5032 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 11685) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 5033 (("]" 11686)) 5034 (("$" 11012) (T_VARIABLE 11014) (simple_variable 11692)) 5035 (("(" 2984) (argument_list 2985) (ctor_arguments 11695)) 5036 ((T_EXTENDS 2987) (extends_from 11696)) 5037 ((T_IMPLEMENTS 2989) (implements_list 11697)) 5038 ((backup_doc_comment 11698)) 5039 (("{" 11699)) 5040 ((class_statement_list 11700)) 5041 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 11701)) 5042 (("%" 304) (")" 11703) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 5043 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 11705) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 5044 (("%" 304) (")" 11706) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 5045 (("{" 11707)) 5046 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 11708) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 5047 (("}" 11709)) 5048 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 11711) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 5049 ((")" 11712)) 5050 (("=" 11713)) 5051 (("!" 10705) ("\"" 10706) ("$" 10707) ("(" 10708) ("+" 10709) ("-" 10710) ("@" 10711) (T_ARRAY 10712) (T_ARRAY_CAST 10713) (T_ATTRIBUTE 130) (T_BOOL_CAST 10714) (T_CLASS_C 10715) (T_CLONE 10716) (T_CONSTANT_ENCAPSED_STRING 10717) (T_DEC 10718) (T_DIR 10719) (T_DNUMBER 10720) (T_DOUBLE_CAST 10721) (T_EMPTY 10722) (T_EVAL 10723) (T_EXIT 10724) (T_FILE 10725) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 10726) (T_INC 10727) (T_INCLUDE 10728) (T_INCLUDE_ONCE 10729) (T_INT_CAST 10730) (T_ISSET 10731) (T_LINE 10732) (T_LIST 10733) (T_LNUMBER 10734) (T_MATCH 10735) (T_METHOD_C 10736) (T_NAME_FULLY_QUALIFIED 10737) (T_NAME_QUALIFIED 10738) (T_NAME_RELATIVE 10739) (T_NEW 10740) (T_NS_C 10741) (T_OBJECT_CAST 10742) (T_PRINT 10743) (T_PROPERTY_C 10744) (T_READONLY 10745) (T_REQUIRE 10746) (T_REQUIRE_ONCE 10747) (T_START_HEREDOC 10748) (T_STATIC 10749) (T_STRING 10750) (T_STRING_CAST 10751) (T_THROW 10752) (T_TRAIT_C 10753) (T_UNSET_CAST 10754) (T_VARIABLE 10755) (T_YIELD 10756) (T_YIELD_FROM 10757) ("[" 10758) ("`" 10759) (array_object_dereferenceable 10760) (attribute 135) (attributes 10761) (callable_expr 10762) (callable_variable 10763) (class_constant 10764) (class_name 10765) (constant 10766) (dereferenceable_scalar 10767) (expr 11714) (fn 10769) (fully_dereferenceable 107) #'10770 (function_call 10771) (inline_function 10772) (internal_functions_in_yacc 10773) (match 10774) (name 10775) (new_dereferenceable 10776) (new_non_dereferenceable 10777) (scalar 10778) (simple_variable 10779) (static_member 10780) (variable 10781) (variable_class_name 10782) ("~" 10783)) 5052 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 11716) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5053 (("," 5226) (possible_comma 11717)) 5054 ((")" 11718)) 5055 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 11726) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 11727) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 11728) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5056 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 11731) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5057 (("," 1498) (possible_comma 11729)) 5058 ((")" 11730)) 5059 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 11733) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 5060 (("%" 304) (")" 11734) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 5061 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 11736) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 5062 (("%" 304) (")" 11737) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 5063 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 11743) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 11744) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 11745) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 11746) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 5064 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 11752) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5065 (("%" 2442) (")" 11749) ("*" 2443) ("+" 2444) ("," 11750) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 5066 (("," 5287) (possible_comma 11747)) 5067 ((")" 11748)) 5068 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 11751) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5069 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 11756) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 5070 ((")" 11757)) 5071 (("%" 304) (")" 11749) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 5072 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 11764) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5073 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 11765)) 5074 (("\"" 11767) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 5075 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 11770) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5076 (("$" 689) (T_ABSTRACT 869) (T_ARRAY 870) (T_AS 871) (T_BREAK 872) (T_CALLABLE 873) (T_CASE 874) (T_CATCH 875) (T_CLASS 876) (T_CLASS_C 877) (T_CLONE 878) (T_CONST 879) (T_CONTINUE 880) (T_DECLARE 881) (T_DEFAULT 882) (T_DIR 883) (T_DO 884) (T_ECHO 885) (T_ELSE 886) (T_ELSEIF 887) (T_EMPTY 888) (T_ENDDECLARE 889) (T_ENDFOR 890) (T_ENDFOREACH 891) (T_ENDIF 892) (T_ENDSWITCH 893) (T_ENDWHILE 894) (T_ENUM 895) (T_EVAL 896) (T_EXIT 897) (T_EXTENDS 898) (T_FILE 899) (T_FINAL 900) (T_FINALLY 901) (T_FN 902) (T_FOR 903) (T_FOREACH 904) (T_FUNCTION 905) (T_FUNC_C 906) (T_GLOBAL 907) (T_GOTO 908) (T_IF 909) (T_IMPLEMENTS 910) (T_INCLUDE 911) (T_INCLUDE_ONCE 912) (T_INSTANCEOF 913) (T_INSTEADOF 914) (T_INTERFACE 915) (T_ISSET 916) (T_LINE 917) (T_LIST 918) (T_LOGICAL_AND 919) (T_LOGICAL_OR 920) (T_LOGICAL_XOR 921) (T_MATCH 922) (T_METHOD_C 923) (T_NAMESPACE 924) (T_NEW 925) (T_NS_C 926) (T_PRINT 927) (T_PRIVATE 928) (T_PROPERTY_C 929) (T_PROTECTED 930) (T_PUBLIC 931) (T_READONLY 932) (T_REQUIRE 933) (T_REQUIRE_ONCE 934) (T_RETURN 935) (T_STATIC 936) (T_STRING 937) (T_SWITCH 938) (T_THROW 939) (T_TRAIT 940) (T_TRAIT_C 941) (T_TRY 942) (T_UNSET 943) (T_USE 944) (T_VAR 945) (T_VARIABLE 738) (T_WHILE 946) (T_YIELD 947) (identifier 11772) (member_name 11773) (reserved_non_modifiers 950) (semi_reserved 951) (simple_variable 11774) ("{" 11775)) 5077 (("(" 956) (argument_list 11778)) 5078 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 11776) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5079 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 11777)) 5080 ((T_FN 35) (T_FUNCTION 131) (fn 752) #'753 (inline_function 11782)) 5081 (("$" 689) (T_STRING 11788) (T_VARIABLE 738) (property_name 11795) (simple_variable 11790) ("{" 11791)) 5082 (("$" 689) (T_STRING 11788) (T_VARIABLE 738) (property_name 11789) (simple_variable 11790) ("{" 11791)) 5083 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 11786) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 5084 (("]" 11787)) 5085 (("(" 956) (argument_list 11794)) 5086 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 11792) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5087 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 11793)) 5088 (("(" 956) (argument_list 11796)) 5089 (("`" 11798)) 5090 (("]" 11800)) 5091 (("=" 11801)) 5092 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 11802) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 5093 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 12337) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5094 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 12335) (encaps_var 5357)) 5095 (("$" 11806) (T_VARIABLE 11854) (simple_variable 12331) ("{" 12332)) 5096 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 12330) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 5097 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 12329) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5098 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 12328) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5099 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 12327) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5100 (("(" 12324)) 5101 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 12323) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5102 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 12322) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5103 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 12309) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (clone_argument_list 12310) (constant 11865) (dereferenceable_scalar 11866) (expr 12311) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5104 (("\"" 1538) ("$" 12111) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 12112) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 12113) ("[" 1560) (array_object_dereferenceable 12114) (callable_expr 12115) (callable_variable 12116) (class_constant 1564) (class_name 12117) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 12118) (name 12119) (new_dereferenceable 1570) (simple_variable 12120) (static_member 12121) (variable 12308) (variable_class_name 12123)) 5105 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 12307) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5106 (("(" 12304)) 5107 (("(" 12301)) 5108 (("(" 12292) (argument_list 12293) (ctor_arguments 12294)) 5109 (("\"" 1538) ("$" 12111) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 12112) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 12113) ("[" 1560) (array_object_dereferenceable 12114) (callable_expr 12115) (callable_variable 12116) (class_constant 1564) (class_name 12117) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 12118) (name 12119) (new_dereferenceable 1570) (simple_variable 12120) (static_member 12121) (variable 12291) (variable_class_name 12123)) 5110 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 12290) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5111 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 12289) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5112 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 12288) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5113 (("(" 12284)) 5114 (("(" 12279)) 5115 (("(" 12273)) 5116 (("$" 12111) ("(" 12236) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 12237) (T_NAME_QUALIFIED 12238) (T_NAME_RELATIVE 12239) (T_READONLY 2955) (T_STATIC 12240) (T_STRING 12241) (T_VARIABLE 12113) (anonymous_class 12242) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 12243) (attribute 2961) (attributes 12244) (class_modifier 2963) (class_name 12245) (class_name_reference 12246) (name 12247) (new_variable 12248) (simple_variable 12249)) 5117 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 12235) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5118 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 12234) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5119 (("(" 12072) (argument_list 12233)) 5120 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 12232) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5121 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 12231) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5122 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 12226) (T_END_HEREDOC 12227) (T_VARIABLE 2920) (encaps_list 12228) (encaps_var 2922)) 5123 ((T_FN 35) (T_FUNCTION 131) (fn 11868) #'11869 (inline_function 12225)) 5124 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 12224) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5125 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 12223) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5126 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 12222) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5127 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 12219) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5128 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 12218) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5129 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 12214) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 5130 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 12212) (encaps_list 2828) (encaps_var 2829)) 5131 ((T_NULLSAFE_OBJECT_OPERATOR 12198) (T_OBJECT_OPERATOR 12199) ("[" 12200)) 5132 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 12195) (attribute 2757) (fn 11868) #'11869 (inline_function 12196)) 5133 (("(" 12072) (argument_list 12194)) 5134 ((T_PAAMAYIM_NEKUDOTAYIM 12186)) 5135 (("%" 11884) ("*" 11885) ("+" 11886) ("-" 11887) ("." 11888) ("/" 11889) ("<" 11890) (">" 11891) ("?" 11892) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 11893) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 11894) (T_BOOLEAN_AND 11895) (T_BOOLEAN_OR 11896) (T_COALESCE 11897) (T_DOUBLE_ARROW 12184) (T_INSTANCEOF 11898) (T_IS_EQUAL 11899) (T_IS_GREATER_OR_EQUAL 11900) (T_IS_IDENTICAL 11901) (T_IS_NOT_EQUAL 11902) (T_IS_NOT_IDENTICAL 11903) (T_IS_SMALLER_OR_EQUAL 11904) (T_LOGICAL_AND 11905) (T_LOGICAL_OR 11906) (T_LOGICAL_XOR 11907) (T_PIPE 11908) (T_POW 11909) (T_SL 11910) (T_SPACESHIP 11911) (T_SR 11912) ("^" 11913) ("|" 11914)) 5136 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 12173)) 5137 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 12161)) 5138 (("(" 12072) (argument_list 12160)) 5139 (("=" 12080) (T_AND_EQUAL 12081) (T_COALESCE_EQUAL 12082) (T_CONCAT_EQUAL 12083) (T_DEC 12084) (T_DIV_EQUAL 12085) (T_INC 12086) (T_MINUS_EQUAL 12087) (T_MOD_EQUAL 12088) (T_MUL_EQUAL 12089) (T_OR_EQUAL 12090) (T_PLUS_EQUAL 12091) (T_POW_EQUAL 12092) (T_SL_EQUAL 12093) (T_SR_EQUAL 12094) (T_XOR_EQUAL 12095)) 5140 ((T_PAAMAYIM_NEKUDOTAYIM 11984)) 5141 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 11883) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5142 (("%" 11884) ("*" 11885) ("+" 11886) ("-" 11887) ("." 11888) ("/" 11889) ("<" 11890) (">" 11891) ("?" 11892) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 11893) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 11894) (T_BOOLEAN_AND 11895) (T_BOOLEAN_OR 11896) (T_COALESCE 11897) (T_INSTANCEOF 11898) (T_IS_EQUAL 11899) (T_IS_GREATER_OR_EQUAL 11900) (T_IS_IDENTICAL 11901) (T_IS_NOT_EQUAL 11902) (T_IS_NOT_IDENTICAL 11903) (T_IS_SMALLER_OR_EQUAL 11904) (T_LOGICAL_AND 11905) (T_LOGICAL_OR 11906) (T_LOGICAL_XOR 11907) (T_PIPE 11908) (T_POW 11909) (T_SL 11910) (T_SPACESHIP 11911) (T_SR 11912) ("^" 11913) ("|" 11914)) 5143 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 11983) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5144 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 11982) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5145 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 11981) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5146 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 11980) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5147 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 11979) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5148 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 11978) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5149 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 11977) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5150 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 11976) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5151 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 11971) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 11972) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 5152 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 11970) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5153 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 11969) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5154 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 11968) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5155 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 11967) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5156 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 11966) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5157 (("$" 11931) ("(" 11932) (T_NAME_FULLY_QUALIFIED 11933) (T_NAME_QUALIFIED 11934) (T_NAME_RELATIVE 11935) (T_STATIC 11936) (T_STRING 11937) (T_VARIABLE 11938) (class_name 11939) (class_name_reference 11940) (name 11941) (new_variable 11942) (simple_variable 11943)) 5158 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 11930) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5159 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 11929) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5160 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 11928) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5161 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 11927) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5162 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 11926) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5163 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 11925) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5164 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 11924) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5165 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 11923) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5166 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 11922) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5167 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 11921) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5168 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 11920) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5169 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 11919) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5170 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 11918) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5171 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 11917) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5172 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 11916) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5173 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 11915) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5174 (("$" 11931) (T_VARIABLE 11938) (simple_variable 11962) ("{" 11963)) 5175 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 11960) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 5176 ((T_PAAMAYIM_NEKUDOTAYIM 11958)) 5177 ((T_NULLSAFE_OBJECT_OPERATOR 11944) (T_OBJECT_OPERATOR 11945) (T_PAAMAYIM_NEKUDOTAYIM 11946) ("[" 11947)) 5178 (("$" 11931) (T_STRING 11951) (T_VARIABLE 11938) (property_name 11957) (simple_variable 11953) ("{" 11954)) 5179 (("$" 11931) (T_STRING 11951) (T_VARIABLE 11938) (property_name 11952) (simple_variable 11953) ("{" 11954)) 5180 (("$" 11931) (T_VARIABLE 11938) (simple_variable 11950)) 5181 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 11948) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 5182 (("]" 11949)) 5183 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 11955) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5184 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 11956)) 5185 (("$" 11931) (T_VARIABLE 11938) (simple_variable 11959)) 5186 (("%" 304) (")" 11961) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 5187 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 11964) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5188 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 11965)) 5189 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 11975) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5190 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 11973) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 5191 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 11974) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5192 (("$" 11806) (T_ABSTRACT 11985) (T_ARRAY 11986) (T_AS 11987) (T_BREAK 11988) (T_CALLABLE 11989) (T_CASE 11990) (T_CATCH 11991) (T_CLASS 11992) (T_CLASS_C 11993) (T_CLONE 11994) (T_CONST 11995) (T_CONTINUE 11996) (T_DECLARE 11997) (T_DEFAULT 11998) (T_DIR 11999) (T_DO 12000) (T_ECHO 12001) (T_ELSE 12002) (T_ELSEIF 12003) (T_EMPTY 12004) (T_ENDDECLARE 12005) (T_ENDFOR 12006) (T_ENDFOREACH 12007) (T_ENDIF 12008) (T_ENDSWITCH 12009) (T_ENDWHILE 12010) (T_ENUM 12011) (T_EVAL 12012) (T_EXIT 12013) (T_EXTENDS 12014) (T_FILE 12015) (T_FINAL 12016) (T_FINALLY 12017) (T_FN 12018) (T_FOR 12019) (T_FOREACH 12020) (T_FUNCTION 12021) (T_FUNC_C 12022) (T_GLOBAL 12023) (T_GOTO 12024) (T_IF 12025) (T_IMPLEMENTS 12026) (T_INCLUDE 12027) (T_INCLUDE_ONCE 12028) (T_INSTANCEOF 12029) (T_INSTEADOF 12030) (T_INTERFACE 12031) (T_ISSET 12032) (T_LINE 12033) (T_LIST 12034) (T_LOGICAL_AND 12035) (T_LOGICAL_OR 12036) (T_LOGICAL_XOR 12037) (T_MATCH 12038) (T_METHOD_C 12039) (T_NAMESPACE 12040) (T_NEW 12041) (T_NS_C 12042) (T_PRINT 12043) (T_PRIVATE 12044) (T_PROPERTY_C 12045) (T_PROTECTED 12046) (T_PUBLIC 12047) (T_READONLY 12048) (T_REQUIRE 12049) (T_REQUIRE_ONCE 12050) (T_RETURN 12051) (T_STATIC 12052) (T_STRING 12053) (T_SWITCH 12054) (T_THROW 12055) (T_TRAIT 12056) (T_TRAIT_C 12057) (T_TRY 12058) (T_UNSET 12059) (T_USE 12060) (T_VAR 12061) (T_VARIABLE 11854) (T_WHILE 12062) (T_YIELD 12063) (identifier 12064) (member_name 12065) (reserved_non_modifiers 12066) (semi_reserved 12067) (simple_variable 12068) ("{" 12069)) 5193 (("(" 12072) (argument_list 12073)) 5194 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 12070) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5195 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 12071)) 5196 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 12074) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 12075) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 12076) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5197 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 12079) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5198 (("," 1498) (possible_comma 12077)) 5199 ((")" 12078)) 5200 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (ampersand 12109) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 12110) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5201 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 12108) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5202 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 12107) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5203 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 12106) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5204 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 12105) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5205 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 12104) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5206 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 12103) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5207 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 12102) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5208 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 12101) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5209 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 12100) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5210 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 12099) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5211 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 12098) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5212 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 12097) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5213 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 12096) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5214 (("\"" 1538) ("$" 12111) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 12112) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 12113) ("[" 1560) (array_object_dereferenceable 12114) (callable_expr 12115) (callable_variable 12116) (class_constant 1564) (class_name 12117) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 12118) (name 12119) (new_dereferenceable 1570) (simple_variable 12120) (static_member 12121) (variable 12122) (variable_class_name 12123)) 5215 (("$" 12111) (T_VARIABLE 12113) (simple_variable 12156) ("{" 12157)) 5216 (("(" 12127) (argument_list 12155)) 5217 ((T_NULLSAFE_OBJECT_OPERATOR 12141) (T_OBJECT_OPERATOR 12142) ("[" 12143)) 5218 (("(" 12127) (argument_list 12140)) 5219 ((T_PAAMAYIM_NEKUDOTAYIM 12136)) 5220 (("(" 12127) (argument_list 12135)) 5221 ((T_PAAMAYIM_NEKUDOTAYIM 12124)) 5222 (("$" 12111) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 12113) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 12125) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 12126) ("{" 1660)) 5223 (("(" 12127) (argument_list 12128)) 5224 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 12129) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 12130) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 12131) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5225 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 12134) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5226 (("," 1498) (possible_comma 12132)) 5227 ((")" 12133)) 5228 (("$" 12111) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 12113) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 12137) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 12138) ("{" 1676)) 5229 (("(" 12127) (argument_list 12139)) 5230 (("$" 12111) (T_STRING 12146) (T_VARIABLE 12113) (property_name 12153) (simple_variable 12148) ("{" 12149)) 5231 (("$" 12111) (T_STRING 12146) (T_VARIABLE 12113) (property_name 12147) (simple_variable 12148) ("{" 12149)) 5232 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 12144) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 5233 (("]" 12145)) 5234 (("(" 12127) (argument_list 12152)) 5235 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 12150) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5236 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 12151)) 5237 (("(" 12127) (argument_list 12154)) 5238 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 12158) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5239 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 12159)) 5240 ((backup_doc_comment 12162)) 5241 (("(" 12163)) 5242 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 12164)) 5243 ((")" 12165)) 5244 ((T_USE 166) (lexical_vars 12166)) 5245 ((":" 168) (return_type 12167)) 5246 ((backup_fn_flags 12168)) 5247 (("{" 12169)) 5248 ((inner_statement_list 12170)) 5249 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 12171) ("~" 129)) 5250 ((backup_fn_flags 12172)) 5251 ((backup_doc_comment 12174)) 5252 (("(" 12175)) 5253 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 12176)) 5254 ((")" 12177)) 5255 ((":" 2271) (return_type 12178)) 5256 ((T_DOUBLE_ARROW 12179)) 5257 ((backup_fn_flags 12180)) 5258 ((backup_lex_pos 12181)) 5259 (("!" 11119) ("\"" 11120) ("$" 11121) ("(" 11122) ("+" 11123) ("-" 11124) ("@" 11125) (T_ARRAY 11126) (T_ARRAY_CAST 11127) (T_ATTRIBUTE 130) (T_BOOL_CAST 11128) (T_CLASS_C 11129) (T_CLONE 11130) (T_CONSTANT_ENCAPSED_STRING 11131) (T_DEC 11132) (T_DIR 11133) (T_DNUMBER 11134) (T_DOUBLE_CAST 11135) (T_EMPTY 11136) (T_EVAL 11137) (T_EXIT 11138) (T_FILE 11139) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11140) (T_INC 11141) (T_INCLUDE 11142) (T_INCLUDE_ONCE 11143) (T_INT_CAST 11144) (T_ISSET 11145) (T_LINE 11146) (T_LIST 11147) (T_LNUMBER 11148) (T_MATCH 11149) (T_METHOD_C 11150) (T_NAME_FULLY_QUALIFIED 11151) (T_NAME_QUALIFIED 11152) (T_NAME_RELATIVE 11153) (T_NEW 11154) (T_NS_C 11155) (T_OBJECT_CAST 11156) (T_PRINT 11157) (T_PROPERTY_C 11158) (T_READONLY 11159) (T_REQUIRE 11160) (T_REQUIRE_ONCE 11161) (T_START_HEREDOC 11162) (T_STATIC 11163) (T_STRING 11164) (T_STRING_CAST 11165) (T_THROW 11166) (T_TRAIT_C 11167) (T_UNSET_CAST 11168) (T_VARIABLE 11169) (T_YIELD 11170) (T_YIELD_FROM 11171) ("[" 11172) ("`" 11173) (array_object_dereferenceable 11174) (attribute 135) (attributes 11175) (callable_expr 11176) (callable_variable 11177) (class_constant 11178) (class_name 11179) (constant 11180) (dereferenceable_scalar 11181) (expr 12182) (fn 11183) (fully_dereferenceable 107) #'11184 (function_call 11185) (inline_function 11186) (internal_functions_in_yacc 11187) (match 11188) (name 11189) (new_dereferenceable 11190) (new_non_dereferenceable 11191) (scalar 11192) (simple_variable 11193) (static_member 11194) (variable 11195) (variable_class_name 11196) ("~" 11197)) 5260 (("%" 11199) ("*" 11200) ("+" 11201) ("-" 11202) ("." 11203) ("/" 11204) ("<" 11205) (">" 11206) ("?" 11207) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 11208) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 11209) (T_BOOLEAN_AND 11210) (T_BOOLEAN_OR 11211) (T_COALESCE 11212) (T_INSTANCEOF 11213) (T_IS_EQUAL 11214) (T_IS_GREATER_OR_EQUAL 11215) (T_IS_IDENTICAL 11216) (T_IS_NOT_EQUAL 11217) (T_IS_NOT_IDENTICAL 11218) (T_IS_SMALLER_OR_EQUAL 11219) (T_LOGICAL_AND 11220) (T_LOGICAL_OR 11221) (T_LOGICAL_XOR 11222) (T_PIPE 11223) (T_POW 11224) (T_SL 11225) (T_SPACESHIP 11226) (T_SR 11227) ("^" 11228) (backup_fn_flags 12183) ("|" 11229)) 5261 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 12185) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 5262 (("$" 11806) (T_ABSTRACT 11985) (T_ARRAY 11986) (T_AS 11987) (T_BREAK 11988) (T_CALLABLE 11989) (T_CASE 11990) (T_CATCH 11991) (T_CLASS 11992) (T_CLASS_C 11993) (T_CLONE 11994) (T_CONST 11995) (T_CONTINUE 11996) (T_DECLARE 11997) (T_DEFAULT 11998) (T_DIR 11999) (T_DO 12000) (T_ECHO 12001) (T_ELSE 12002) (T_ELSEIF 12003) (T_EMPTY 12004) (T_ENDDECLARE 12005) (T_ENDFOR 12006) (T_ENDFOREACH 12007) (T_ENDIF 12008) (T_ENDSWITCH 12009) (T_ENDWHILE 12010) (T_ENUM 12011) (T_EVAL 12012) (T_EXIT 12013) (T_EXTENDS 12014) (T_FILE 12015) (T_FINAL 12016) (T_FINALLY 12017) (T_FN 12018) (T_FOR 12019) (T_FOREACH 12020) (T_FUNCTION 12021) (T_FUNC_C 12022) (T_GLOBAL 12023) (T_GOTO 12024) (T_IF 12025) (T_IMPLEMENTS 12026) (T_INCLUDE 12027) (T_INCLUDE_ONCE 12028) (T_INSTANCEOF 12029) (T_INSTEADOF 12030) (T_INTERFACE 12031) (T_ISSET 12032) (T_LINE 12033) (T_LIST 12034) (T_LOGICAL_AND 12035) (T_LOGICAL_OR 12036) (T_LOGICAL_XOR 12037) (T_MATCH 12038) (T_METHOD_C 12039) (T_NAMESPACE 12040) (T_NEW 12041) (T_NS_C 12042) (T_PRINT 12043) (T_PRIVATE 12044) (T_PROPERTY_C 12045) (T_PROTECTED 12046) (T_PUBLIC 12047) (T_READONLY 12048) (T_REQUIRE 12049) (T_REQUIRE_ONCE 12050) (T_RETURN 12051) (T_STATIC 12052) (T_STRING 12053) (T_SWITCH 12054) (T_THROW 12055) (T_TRAIT 12056) (T_TRAIT_C 12057) (T_TRY 12058) (T_UNSET 12059) (T_USE 12060) (T_VAR 12061) (T_VARIABLE 11854) (T_WHILE 12062) (T_YIELD 12063) (identifier 12187) (member_name 12188) (reserved_non_modifiers 12066) (semi_reserved 12067) (simple_variable 12189) ("{" 12190)) 5263 (("(" 12072) (argument_list 12193)) 5264 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 12191) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5265 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 12192)) 5266 ((T_FN 35) (T_FUNCTION 131) (fn 11868) #'11869 (inline_function 12197)) 5267 (("$" 11806) (T_STRING 12203) (T_VARIABLE 11854) (property_name 12210) (simple_variable 12205) ("{" 12206)) 5268 (("$" 11806) (T_STRING 12203) (T_VARIABLE 11854) (property_name 12204) (simple_variable 12205) ("{" 12206)) 5269 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 12201) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 5270 (("]" 12202)) 5271 (("(" 12072) (argument_list 12209)) 5272 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 12207) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5273 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 12208)) 5274 (("(" 12072) (argument_list 12211)) 5275 (("`" 12213)) 5276 (("]" 12215)) 5277 (("=" 12216)) 5278 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 12217) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5279 (("%" 11884) ("*" 11885) ("+" 11886) ("-" 11887) ("." 11888) ("/" 11889) ("<" 11890) (">" 11891) ("?" 11892) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 11893) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 11894) (T_BOOLEAN_AND 11895) (T_BOOLEAN_OR 11896) (T_COALESCE 11897) (T_DOUBLE_ARROW 12220) (T_INSTANCEOF 11898) (T_IS_EQUAL 11899) (T_IS_GREATER_OR_EQUAL 11900) (T_IS_IDENTICAL 11901) (T_IS_NOT_EQUAL 11902) (T_IS_NOT_IDENTICAL 11903) (T_IS_SMALLER_OR_EQUAL 11904) (T_LOGICAL_AND 11905) (T_LOGICAL_OR 11906) (T_LOGICAL_XOR 11907) (T_PIPE 11908) (T_POW 11909) (T_SL 11910) (T_SPACESHIP 11911) (T_SR 11912) ("^" 11913) ("|" 11914)) 5280 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 12221) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5281 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 12230) (T_VARIABLE 2920) (encaps_var 2934)) 5282 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 12229) (T_VARIABLE 2920) (encaps_var 2925)) 5283 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 12271) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 5284 ((T_CLASS 12263)) 5285 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 12262) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 12243) (attribute 2982) (class_modifier 2963)) 5286 ((T_PAAMAYIM_NEKUDOTAYIM 12260)) 5287 (("(" 12127) (argument_list 12259)) 5288 ((T_NULLSAFE_OBJECT_OPERATOR 12250) (T_OBJECT_OPERATOR 12251) (T_PAAMAYIM_NEKUDOTAYIM 12252) ("[" 12253)) 5289 (("$" 12111) (T_STRING 12146) (T_VARIABLE 12113) (property_name 12258) (simple_variable 12148) ("{" 12149)) 5290 (("$" 12111) (T_STRING 12146) (T_VARIABLE 12113) (property_name 12257) (simple_variable 12148) ("{" 12149)) 5291 (("$" 12111) (T_VARIABLE 12113) (simple_variable 12256)) 5292 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 12254) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 5293 (("]" 12255)) 5294 (("$" 12111) (T_VARIABLE 12113) (simple_variable 12261)) 5295 (("(" 2984) (argument_list 2985) (ctor_arguments 12264)) 5296 ((T_EXTENDS 2987) (extends_from 12265)) 5297 ((T_IMPLEMENTS 2989) (implements_list 12266)) 5298 ((backup_doc_comment 12267)) 5299 (("{" 12268)) 5300 ((class_statement_list 12269)) 5301 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 12270)) 5302 (("%" 304) (")" 12272) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 5303 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 12274) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 5304 (("%" 304) (")" 12275) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 5305 (("{" 12276)) 5306 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 12277) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 5307 (("}" 12278)) 5308 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 12280) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 5309 ((")" 12281)) 5310 (("=" 12282)) 5311 (("!" 11804) ("\"" 11805) ("$" 11806) ("(" 11807) ("+" 11808) ("-" 11809) ("@" 11810) (T_ARRAY 11811) (T_ARRAY_CAST 11812) (T_ATTRIBUTE 130) (T_BOOL_CAST 11813) (T_CLASS_C 11814) (T_CLONE 11815) (T_CONSTANT_ENCAPSED_STRING 11816) (T_DEC 11817) (T_DIR 11818) (T_DNUMBER 11819) (T_DOUBLE_CAST 11820) (T_EMPTY 11821) (T_EVAL 11822) (T_EXIT 11823) (T_FILE 11824) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 11825) (T_INC 11826) (T_INCLUDE 11827) (T_INCLUDE_ONCE 11828) (T_INT_CAST 11829) (T_ISSET 11830) (T_LINE 11831) (T_LIST 11832) (T_LNUMBER 11833) (T_MATCH 11834) (T_METHOD_C 11835) (T_NAME_FULLY_QUALIFIED 11836) (T_NAME_QUALIFIED 11837) (T_NAME_RELATIVE 11838) (T_NEW 11839) (T_NS_C 11840) (T_OBJECT_CAST 11841) (T_PRINT 11842) (T_PROPERTY_C 11843) (T_READONLY 11844) (T_REQUIRE 11845) (T_REQUIRE_ONCE 11846) (T_START_HEREDOC 11847) (T_STATIC 11848) (T_STRING 11849) (T_STRING_CAST 11850) (T_THROW 11851) (T_TRAIT_C 11852) (T_UNSET_CAST 11853) (T_VARIABLE 11854) (T_YIELD 11855) (T_YIELD_FROM 11856) ("[" 11857) ("`" 11858) (array_object_dereferenceable 11859) (attribute 135) (attributes 11860) (callable_expr 11861) (callable_variable 11862) (class_constant 11863) (class_name 11864) (constant 11865) (dereferenceable_scalar 11866) (expr 12283) (fn 11868) (fully_dereferenceable 107) #'11869 (function_call 11870) (inline_function 11871) (internal_functions_in_yacc 11872) (match 11873) (name 11874) (new_dereferenceable 11875) (new_non_dereferenceable 11876) (scalar 11877) (simple_variable 11878) (static_member 11879) (variable 11880) (variable_class_name 11881) ("~" 11882)) 5312 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 12285) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5313 (("," 5226) (possible_comma 12286)) 5314 ((")" 12287)) 5315 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 12295) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 12296) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 12297) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5316 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 12300) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5317 (("," 1498) (possible_comma 12298)) 5318 ((")" 12299)) 5319 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 12302) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 5320 (("%" 304) (")" 12303) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 5321 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 12305) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 5322 (("%" 304) (")" 12306) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 5323 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 12312) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 12313) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 12314) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 12315) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 5324 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 12321) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5325 (("%" 2442) (")" 12318) ("*" 2443) ("+" 2444) ("," 12319) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 5326 (("," 5287) (possible_comma 12316)) 5327 ((")" 12317)) 5328 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 12320) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5329 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 12325) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 5330 ((")" 12326)) 5331 (("%" 304) (")" 12318) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 5332 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 12333) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5333 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 12334)) 5334 (("\"" 12336) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 5335 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 12346) (T_VARIABLE 2920) (encaps_var 2934)) 5336 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 12345) (T_VARIABLE 2920) (encaps_var 2925)) 5337 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 12387) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 5338 ((T_CLASS 12379)) 5339 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 12378) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 12359) (attribute 2982) (class_modifier 2963)) 5340 ((T_PAAMAYIM_NEKUDOTAYIM 12376)) 5341 (("(" 10650) (argument_list 12375)) 5342 ((T_NULLSAFE_OBJECT_OPERATOR 12366) (T_OBJECT_OPERATOR 12367) (T_PAAMAYIM_NEKUDOTAYIM 12368) ("[" 12369)) 5343 (("$" 10634) (T_STRING 10669) (T_VARIABLE 10636) (property_name 12374) (simple_variable 10671) ("{" 10672)) 5344 (("$" 10634) (T_STRING 10669) (T_VARIABLE 10636) (property_name 12373) (simple_variable 10671) ("{" 10672)) 5345 (("$" 10634) (T_VARIABLE 10636) (simple_variable 12372)) 5346 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 12370) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 5347 (("]" 12371)) 5348 (("$" 10634) (T_VARIABLE 10636) (simple_variable 12377)) 5349 (("(" 2984) (argument_list 2985) (ctor_arguments 12380)) 5350 ((T_EXTENDS 2987) (extends_from 12381)) 5351 ((T_IMPLEMENTS 2989) (implements_list 12382)) 5352 ((backup_doc_comment 12383)) 5353 (("{" 12384)) 5354 ((class_statement_list 12385)) 5355 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 12386)) 5356 (("%" 304) (")" 12388) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 5357 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 12390) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 5358 (("%" 304) (")" 12391) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 5359 (("{" 12392)) 5360 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 12393) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 5361 (("}" 12394)) 5362 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 12396) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 5363 ((")" 12397)) 5364 (("=" 12398)) 5365 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 12399) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 5366 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 12401) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5367 (("," 5226) (possible_comma 12402)) 5368 ((")" 12403)) 5369 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 12411) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 12412) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 12413) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5370 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 12416) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5371 (("," 1498) (possible_comma 12414)) 5372 ((")" 12415)) 5373 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 12418) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 5374 (("%" 304) (")" 12419) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 5375 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 12421) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 5376 (("%" 304) (")" 12422) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 5377 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 12428) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 12429) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 12430) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 12431) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 5378 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 12437) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5379 (("%" 2442) (")" 12434) ("*" 2443) ("+" 2444) ("," 12435) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 5380 (("," 5287) (possible_comma 12432)) 5381 ((")" 12433)) 5382 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 12436) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5383 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 12441) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 5384 ((")" 12442)) 5385 (("%" 304) (")" 12434) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 5386 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 12450) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5387 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 12451)) 5388 (("\"" 12453) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 5389 (("$" 522) (T_ABSTRACT 12464) (T_ARRAY 12465) (T_AS 12466) (T_BREAK 12467) (T_CALLABLE 12468) (T_CASE 12469) (T_CATCH 12470) (T_CLASS 12471) (T_CLASS_C 12472) (T_CLONE 12473) (T_CONST 12474) (T_CONTINUE 12475) (T_DECLARE 12476) (T_DEFAULT 12477) (T_DIR 12478) (T_DO 12479) (T_ECHO 12480) (T_ELSE 12481) (T_ELSEIF 12482) (T_EMPTY 12483) (T_ENDDECLARE 12484) (T_ENDFOR 12485) (T_ENDFOREACH 12486) (T_ENDIF 12487) (T_ENDSWITCH 12488) (T_ENDWHILE 12489) (T_ENUM 12490) (T_EVAL 12491) (T_EXIT 12492) (T_EXTENDS 12493) (T_FILE 12494) (T_FINAL 12495) (T_FINALLY 12496) (T_FN 12497) (T_FOR 12498) (T_FOREACH 12499) (T_FUNCTION 12500) (T_FUNC_C 12501) (T_GLOBAL 12502) (T_GOTO 12503) (T_IF 12504) (T_IMPLEMENTS 12505) (T_INCLUDE 12506) (T_INCLUDE_ONCE 12507) (T_INSTANCEOF 12508) (T_INSTEADOF 12509) (T_INTERFACE 12510) (T_ISSET 12511) (T_LINE 12512) (T_LIST 12513) (T_LOGICAL_AND 12514) (T_LOGICAL_OR 12515) (T_LOGICAL_XOR 12516) (T_MATCH 12517) (T_METHOD_C 12518) (T_NAMESPACE 12519) (T_NEW 12520) (T_NS_C 12521) (T_PRINT 12522) (T_PRIVATE 12523) (T_PROPERTY_C 12524) (T_PROTECTED 12525) (T_PUBLIC 12526) (T_READONLY 12527) (T_REQUIRE 12528) (T_REQUIRE_ONCE 12529) (T_RETURN 12530) (T_STATIC 12531) (T_STRING 12532) (T_SWITCH 12533) (T_THROW 12534) (T_TRAIT 12535) (T_TRAIT_C 12536) (T_TRY 12537) (T_UNSET 12538) (T_USE 12539) (T_VAR 12540) (T_VARIABLE 570) (T_WHILE 12541) (T_YIELD 12542) (identifier 12543) (member_name 12544) (reserved_non_modifiers 12545) (semi_reserved 12546) (simple_variable 12547) ("{" 12548)) 5390 (("(" 12551) (argument_list 12552)) 5391 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 12549) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5392 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 12550)) 5393 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 12553) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 12554) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 12555) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5394 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 12558) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5395 (("," 1498) (possible_comma 12556)) 5396 ((")" 12557)) 5397 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (ampersand 12588) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 12589) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5398 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 12587) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5399 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 12586) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5400 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 12585) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5401 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 12584) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5402 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 12583) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5403 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 12582) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5404 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 12581) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5405 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 12580) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5406 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 12579) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5407 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 12578) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5408 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 12577) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5409 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 12576) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5410 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 12575) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5411 (("\"" 1538) ("$" 12590) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 12591) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 12592) ("[" 1560) (array_object_dereferenceable 12593) (callable_expr 12594) (callable_variable 12595) (class_constant 1564) (class_name 12596) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 12597) (name 12598) (new_dereferenceable 1570) (simple_variable 12599) (static_member 12600) (variable 12601) (variable_class_name 12602)) 5412 (("$" 12590) (T_VARIABLE 12592) (simple_variable 12635) ("{" 12636)) 5413 (("(" 12606) (argument_list 12634)) 5414 ((T_NULLSAFE_OBJECT_OPERATOR 12620) (T_OBJECT_OPERATOR 12621) ("[" 12622)) 5415 (("(" 12606) (argument_list 12619)) 5416 ((T_PAAMAYIM_NEKUDOTAYIM 12615)) 5417 (("(" 12606) (argument_list 12614)) 5418 ((T_PAAMAYIM_NEKUDOTAYIM 12603)) 5419 (("$" 12590) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 12592) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 12604) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 12605) ("{" 1660)) 5420 (("(" 12606) (argument_list 12607)) 5421 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 12608) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 12609) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 12610) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5422 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 12613) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5423 (("," 1498) (possible_comma 12611)) 5424 ((")" 12612)) 5425 (("$" 12590) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 12592) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 12616) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 12617) ("{" 1676)) 5426 (("(" 12606) (argument_list 12618)) 5427 (("$" 12590) (T_STRING 12625) (T_VARIABLE 12592) (property_name 12632) (simple_variable 12627) ("{" 12628)) 5428 (("$" 12590) (T_STRING 12625) (T_VARIABLE 12592) (property_name 12626) (simple_variable 12627) ("{" 12628)) 5429 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 12623) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 5430 (("]" 12624)) 5431 (("(" 12606) (argument_list 12631)) 5432 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 12629) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5433 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 12630)) 5434 (("(" 12606) (argument_list 12633)) 5435 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 12637) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5436 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 12638)) 5437 ((backup_doc_comment 12641)) 5438 (("(" 12642)) 5439 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 12643)) 5440 ((")" 12644)) 5441 ((T_USE 166) (lexical_vars 12645)) 5442 ((":" 168) (return_type 12646)) 5443 ((backup_fn_flags 12647)) 5444 (("{" 12648)) 5445 ((inner_statement_list 12649)) 5446 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 12650) ("~" 129)) 5447 ((backup_fn_flags 12651)) 5448 ((backup_doc_comment 12653)) 5449 (("(" 12654)) 5450 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 12655)) 5451 ((")" 12656)) 5452 ((":" 2271) (return_type 12657)) 5453 ((T_DOUBLE_ARROW 12658)) 5454 ((backup_fn_flags 12659)) 5455 ((backup_lex_pos 12660)) 5456 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12724) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5457 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 13724) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5458 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 13722) (encaps_var 5357)) 5459 (("$" 12663) (T_VARIABLE 12711) (simple_variable 13718) ("{" 13719)) 5460 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 13717) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 5461 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 13716) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5462 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 13715) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5463 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 13714) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5464 (("(" 13711)) 5465 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 13710) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5466 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 13709) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5467 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 13696) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (clone_argument_list 13697) (constant 12722) (dereferenceable_scalar 12723) (expr 13698) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5468 (("\"" 1538) ("$" 12968) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 12969) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 12970) ("[" 1560) (array_object_dereferenceable 12971) (callable_expr 12972) (callable_variable 12973) (class_constant 1564) (class_name 12974) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 12975) (name 12976) (new_dereferenceable 1570) (simple_variable 12977) (static_member 12978) (variable 13695) (variable_class_name 12980)) 5469 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 13694) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5470 (("(" 13691)) 5471 (("(" 13688)) 5472 (("(" 13679) (argument_list 13680) (ctor_arguments 13681)) 5473 (("\"" 1538) ("$" 12968) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 12969) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 12970) ("[" 1560) (array_object_dereferenceable 12971) (callable_expr 12972) (callable_variable 12973) (class_constant 1564) (class_name 12974) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 12975) (name 12976) (new_dereferenceable 1570) (simple_variable 12977) (static_member 12978) (variable 13678) (variable_class_name 12980)) 5474 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 13677) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5475 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 13676) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5476 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 13675) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5477 (("(" 13671)) 5478 (("(" 13666)) 5479 (("(" 13660)) 5480 (("$" 12968) ("(" 13623) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 13624) (T_NAME_QUALIFIED 13625) (T_NAME_RELATIVE 13626) (T_READONLY 2955) (T_STATIC 13627) (T_STRING 13628) (T_VARIABLE 12970) (anonymous_class 13629) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 13630) (attribute 2961) (attributes 13631) (class_modifier 2963) (class_name 13632) (class_name_reference 13633) (name 13634) (new_variable 13635) (simple_variable 13636)) 5481 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 13622) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5482 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 13621) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5483 (("(" 12929) (argument_list 13620)) 5484 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 13619) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5485 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 13618) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5486 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 13613) (T_END_HEREDOC 13614) (T_VARIABLE 2920) (encaps_list 13615) (encaps_var 2922)) 5487 ((T_FN 35) (T_FUNCTION 131) (fn 12725) #'12726 (inline_function 13612)) 5488 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 13611) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5489 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 13610) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5490 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 13609) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5491 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13138) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5492 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 13074) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5493 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 13070) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 5494 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 13068) (encaps_list 2828) (encaps_var 2829)) 5495 ((T_NULLSAFE_OBJECT_OPERATOR 13054) (T_OBJECT_OPERATOR 13055) ("[" 13056)) 5496 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 13051) (attribute 2757) (fn 12725) #'12726 (inline_function 13052)) 5497 (("(" 12929) (argument_list 13050)) 5498 ((T_PAAMAYIM_NEKUDOTAYIM 13042)) 5499 (("%" 12741) ("*" 12742) ("+" 12743) ("-" 12744) ("." 12745) ("/" 12746) ("<" 12747) (">" 12748) ("?" 12749) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 12750) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 12751) (T_BOOLEAN_AND 12752) (T_BOOLEAN_OR 12753) (T_COALESCE 12754) (T_INSTANCEOF 12755) (T_IS_EQUAL 12756) (T_IS_GREATER_OR_EQUAL 12757) (T_IS_IDENTICAL 12758) (T_IS_NOT_EQUAL 12759) (T_IS_NOT_IDENTICAL 12760) (T_IS_SMALLER_OR_EQUAL 12761) (T_LOGICAL_AND 12762) (T_LOGICAL_OR 12763) (T_LOGICAL_XOR 12764) (T_PIPE 12765) (T_POW 12766) (T_SL 12767) (T_SPACESHIP 12768) (T_SR 12769) ("^" 12770) (backup_fn_flags 13041) ("|" 12771)) 5500 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 13030)) 5501 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 13018)) 5502 (("(" 12929) (argument_list 13017)) 5503 (("=" 12937) (T_AND_EQUAL 12938) (T_COALESCE_EQUAL 12939) (T_CONCAT_EQUAL 12940) (T_DEC 12941) (T_DIV_EQUAL 12942) (T_INC 12943) (T_MINUS_EQUAL 12944) (T_MOD_EQUAL 12945) (T_MUL_EQUAL 12946) (T_OR_EQUAL 12947) (T_PLUS_EQUAL 12948) (T_POW_EQUAL 12949) (T_SL_EQUAL 12950) (T_SR_EQUAL 12951) (T_XOR_EQUAL 12952)) 5504 ((T_PAAMAYIM_NEKUDOTAYIM 12841)) 5505 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12740) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5506 (("%" 12741) ("*" 12742) ("+" 12743) ("-" 12744) ("." 12745) ("/" 12746) ("<" 12747) (">" 12748) ("?" 12749) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 12750) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 12751) (T_BOOLEAN_AND 12752) (T_BOOLEAN_OR 12753) (T_COALESCE 12754) (T_INSTANCEOF 12755) (T_IS_EQUAL 12756) (T_IS_GREATER_OR_EQUAL 12757) (T_IS_IDENTICAL 12758) (T_IS_NOT_EQUAL 12759) (T_IS_NOT_IDENTICAL 12760) (T_IS_SMALLER_OR_EQUAL 12761) (T_LOGICAL_AND 12762) (T_LOGICAL_OR 12763) (T_LOGICAL_XOR 12764) (T_PIPE 12765) (T_POW 12766) (T_SL 12767) (T_SPACESHIP 12768) (T_SR 12769) ("^" 12770) ("|" 12771)) 5507 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12840) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5508 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12839) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5509 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12838) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5510 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12837) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5511 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12836) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5512 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12835) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5513 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12834) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5514 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12833) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5515 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 12828) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 12829) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 5516 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12827) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5517 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12826) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5518 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12825) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5519 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12824) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5520 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12823) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5521 (("$" 12788) ("(" 12789) (T_NAME_FULLY_QUALIFIED 12790) (T_NAME_QUALIFIED 12791) (T_NAME_RELATIVE 12792) (T_STATIC 12793) (T_STRING 12794) (T_VARIABLE 12795) (class_name 12796) (class_name_reference 12797) (name 12798) (new_variable 12799) (simple_variable 12800)) 5522 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12787) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5523 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12786) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5524 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12785) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5525 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12784) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5526 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12783) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5527 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12782) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5528 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12781) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5529 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12780) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5530 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12779) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5531 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12778) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5532 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12777) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5533 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12776) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5534 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12775) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5535 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12774) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5536 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12773) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5537 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12772) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5538 (("$" 12788) (T_VARIABLE 12795) (simple_variable 12819) ("{" 12820)) 5539 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 12817) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 5540 ((T_PAAMAYIM_NEKUDOTAYIM 12815)) 5541 ((T_NULLSAFE_OBJECT_OPERATOR 12801) (T_OBJECT_OPERATOR 12802) (T_PAAMAYIM_NEKUDOTAYIM 12803) ("[" 12804)) 5542 (("$" 12788) (T_STRING 12808) (T_VARIABLE 12795) (property_name 12814) (simple_variable 12810) ("{" 12811)) 5543 (("$" 12788) (T_STRING 12808) (T_VARIABLE 12795) (property_name 12809) (simple_variable 12810) ("{" 12811)) 5544 (("$" 12788) (T_VARIABLE 12795) (simple_variable 12807)) 5545 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 12805) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 5546 (("]" 12806)) 5547 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 12812) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5548 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 12813)) 5549 (("$" 12788) (T_VARIABLE 12795) (simple_variable 12816)) 5550 (("%" 304) (")" 12818) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 5551 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 12821) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5552 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 12822)) 5553 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12832) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5554 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 12830) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 5555 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12831) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5556 (("$" 12663) (T_ABSTRACT 12842) (T_ARRAY 12843) (T_AS 12844) (T_BREAK 12845) (T_CALLABLE 12846) (T_CASE 12847) (T_CATCH 12848) (T_CLASS 12849) (T_CLASS_C 12850) (T_CLONE 12851) (T_CONST 12852) (T_CONTINUE 12853) (T_DECLARE 12854) (T_DEFAULT 12855) (T_DIR 12856) (T_DO 12857) (T_ECHO 12858) (T_ELSE 12859) (T_ELSEIF 12860) (T_EMPTY 12861) (T_ENDDECLARE 12862) (T_ENDFOR 12863) (T_ENDFOREACH 12864) (T_ENDIF 12865) (T_ENDSWITCH 12866) (T_ENDWHILE 12867) (T_ENUM 12868) (T_EVAL 12869) (T_EXIT 12870) (T_EXTENDS 12871) (T_FILE 12872) (T_FINAL 12873) (T_FINALLY 12874) (T_FN 12875) (T_FOR 12876) (T_FOREACH 12877) (T_FUNCTION 12878) (T_FUNC_C 12879) (T_GLOBAL 12880) (T_GOTO 12881) (T_IF 12882) (T_IMPLEMENTS 12883) (T_INCLUDE 12884) (T_INCLUDE_ONCE 12885) (T_INSTANCEOF 12886) (T_INSTEADOF 12887) (T_INTERFACE 12888) (T_ISSET 12889) (T_LINE 12890) (T_LIST 12891) (T_LOGICAL_AND 12892) (T_LOGICAL_OR 12893) (T_LOGICAL_XOR 12894) (T_MATCH 12895) (T_METHOD_C 12896) (T_NAMESPACE 12897) (T_NEW 12898) (T_NS_C 12899) (T_PRINT 12900) (T_PRIVATE 12901) (T_PROPERTY_C 12902) (T_PROTECTED 12903) (T_PUBLIC 12904) (T_READONLY 12905) (T_REQUIRE 12906) (T_REQUIRE_ONCE 12907) (T_RETURN 12908) (T_STATIC 12909) (T_STRING 12910) (T_SWITCH 12911) (T_THROW 12912) (T_TRAIT 12913) (T_TRAIT_C 12914) (T_TRY 12915) (T_UNSET 12916) (T_USE 12917) (T_VAR 12918) (T_VARIABLE 12711) (T_WHILE 12919) (T_YIELD 12920) (identifier 12921) (member_name 12922) (reserved_non_modifiers 12923) (semi_reserved 12924) (simple_variable 12925) ("{" 12926)) 5557 (("(" 12929) (argument_list 12930)) 5558 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 12927) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5559 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 12928)) 5560 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 12931) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 12932) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 12933) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5561 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 12936) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5562 (("," 1498) (possible_comma 12934)) 5563 ((")" 12935)) 5564 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (ampersand 12966) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12967) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5565 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12965) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5566 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12964) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5567 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12963) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5568 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12962) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5569 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12961) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5570 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12960) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5571 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12959) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5572 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12958) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5573 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12957) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5574 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12956) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5575 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12955) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5576 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12954) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5577 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 12953) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5578 (("\"" 1538) ("$" 12968) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 12969) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 12970) ("[" 1560) (array_object_dereferenceable 12971) (callable_expr 12972) (callable_variable 12973) (class_constant 1564) (class_name 12974) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 12975) (name 12976) (new_dereferenceable 1570) (simple_variable 12977) (static_member 12978) (variable 12979) (variable_class_name 12980)) 5579 (("$" 12968) (T_VARIABLE 12970) (simple_variable 13013) ("{" 13014)) 5580 (("(" 12984) (argument_list 13012)) 5581 ((T_NULLSAFE_OBJECT_OPERATOR 12998) (T_OBJECT_OPERATOR 12999) ("[" 13000)) 5582 (("(" 12984) (argument_list 12997)) 5583 ((T_PAAMAYIM_NEKUDOTAYIM 12993)) 5584 (("(" 12984) (argument_list 12992)) 5585 ((T_PAAMAYIM_NEKUDOTAYIM 12981)) 5586 (("$" 12968) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 12970) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 12982) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 12983) ("{" 1660)) 5587 (("(" 12984) (argument_list 12985)) 5588 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 12986) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 12987) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 12988) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5589 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 12991) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5590 (("," 1498) (possible_comma 12989)) 5591 ((")" 12990)) 5592 (("$" 12968) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 12970) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 12994) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 12995) ("{" 1676)) 5593 (("(" 12984) (argument_list 12996)) 5594 (("$" 12968) (T_STRING 13003) (T_VARIABLE 12970) (property_name 13010) (simple_variable 13005) ("{" 13006)) 5595 (("$" 12968) (T_STRING 13003) (T_VARIABLE 12970) (property_name 13004) (simple_variable 13005) ("{" 13006)) 5596 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 13001) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 5597 (("]" 13002)) 5598 (("(" 12984) (argument_list 13009)) 5599 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 13007) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5600 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 13008)) 5601 (("(" 12984) (argument_list 13011)) 5602 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 13015) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5603 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 13016)) 5604 ((backup_doc_comment 13019)) 5605 (("(" 13020)) 5606 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 13021)) 5607 ((")" 13022)) 5608 ((T_USE 166) (lexical_vars 13023)) 5609 ((":" 168) (return_type 13024)) 5610 ((backup_fn_flags 13025)) 5611 (("{" 13026)) 5612 ((inner_statement_list 13027)) 5613 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 13028) ("~" 129)) 5614 ((backup_fn_flags 13029)) 5615 ((backup_doc_comment 13031)) 5616 (("(" 13032)) 5617 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 13033)) 5618 ((")" 13034)) 5619 ((":" 2271) (return_type 13035)) 5620 ((T_DOUBLE_ARROW 13036)) 5621 ((backup_fn_flags 13037)) 5622 ((backup_lex_pos 13038)) 5623 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 13039) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5624 (("%" 12741) ("*" 12742) ("+" 12743) ("-" 12744) ("." 12745) ("/" 12746) ("<" 12747) (">" 12748) ("?" 12749) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 12750) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 12751) (T_BOOLEAN_AND 12752) (T_BOOLEAN_OR 12753) (T_COALESCE 12754) (T_INSTANCEOF 12755) (T_IS_EQUAL 12756) (T_IS_GREATER_OR_EQUAL 12757) (T_IS_IDENTICAL 12758) (T_IS_NOT_EQUAL 12759) (T_IS_NOT_IDENTICAL 12760) (T_IS_SMALLER_OR_EQUAL 12761) (T_LOGICAL_AND 12762) (T_LOGICAL_OR 12763) (T_LOGICAL_XOR 12764) (T_PIPE 12765) (T_POW 12766) (T_SL 12767) (T_SPACESHIP 12768) (T_SR 12769) ("^" 12770) (backup_fn_flags 13040) ("|" 12771)) 5625 (("$" 12663) (T_ABSTRACT 12842) (T_ARRAY 12843) (T_AS 12844) (T_BREAK 12845) (T_CALLABLE 12846) (T_CASE 12847) (T_CATCH 12848) (T_CLASS 12849) (T_CLASS_C 12850) (T_CLONE 12851) (T_CONST 12852) (T_CONTINUE 12853) (T_DECLARE 12854) (T_DEFAULT 12855) (T_DIR 12856) (T_DO 12857) (T_ECHO 12858) (T_ELSE 12859) (T_ELSEIF 12860) (T_EMPTY 12861) (T_ENDDECLARE 12862) (T_ENDFOR 12863) (T_ENDFOREACH 12864) (T_ENDIF 12865) (T_ENDSWITCH 12866) (T_ENDWHILE 12867) (T_ENUM 12868) (T_EVAL 12869) (T_EXIT 12870) (T_EXTENDS 12871) (T_FILE 12872) (T_FINAL 12873) (T_FINALLY 12874) (T_FN 12875) (T_FOR 12876) (T_FOREACH 12877) (T_FUNCTION 12878) (T_FUNC_C 12879) (T_GLOBAL 12880) (T_GOTO 12881) (T_IF 12882) (T_IMPLEMENTS 12883) (T_INCLUDE 12884) (T_INCLUDE_ONCE 12885) (T_INSTANCEOF 12886) (T_INSTEADOF 12887) (T_INTERFACE 12888) (T_ISSET 12889) (T_LINE 12890) (T_LIST 12891) (T_LOGICAL_AND 12892) (T_LOGICAL_OR 12893) (T_LOGICAL_XOR 12894) (T_MATCH 12895) (T_METHOD_C 12896) (T_NAMESPACE 12897) (T_NEW 12898) (T_NS_C 12899) (T_PRINT 12900) (T_PRIVATE 12901) (T_PROPERTY_C 12902) (T_PROTECTED 12903) (T_PUBLIC 12904) (T_READONLY 12905) (T_REQUIRE 12906) (T_REQUIRE_ONCE 12907) (T_RETURN 12908) (T_STATIC 12909) (T_STRING 12910) (T_SWITCH 12911) (T_THROW 12912) (T_TRAIT 12913) (T_TRAIT_C 12914) (T_TRY 12915) (T_UNSET 12916) (T_USE 12917) (T_VAR 12918) (T_VARIABLE 12711) (T_WHILE 12919) (T_YIELD 12920) (identifier 13043) (member_name 13044) (reserved_non_modifiers 12923) (semi_reserved 12924) (simple_variable 13045) ("{" 13046)) 5626 (("(" 12929) (argument_list 13049)) 5627 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 13047) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5628 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 13048)) 5629 ((T_FN 35) (T_FUNCTION 131) (fn 12725) #'12726 (inline_function 13053)) 5630 (("$" 12663) (T_STRING 13059) (T_VARIABLE 12711) (property_name 13066) (simple_variable 13061) ("{" 13062)) 5631 (("$" 12663) (T_STRING 13059) (T_VARIABLE 12711) (property_name 13060) (simple_variable 13061) ("{" 13062)) 5632 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 13057) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 5633 (("]" 13058)) 5634 (("(" 12929) (argument_list 13065)) 5635 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 13063) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5636 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 13064)) 5637 (("(" 12929) (argument_list 13067)) 5638 (("`" 13069)) 5639 (("]" 13071)) 5640 (("=" 13072)) 5641 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 13073) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5642 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13608) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5643 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 13606) (encaps_var 5357)) 5644 (("$" 13077) (T_VARIABLE 13125) (simple_variable 13602) ("{" 13603)) 5645 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 13601) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 5646 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13600) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5647 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13599) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5648 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13598) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5649 (("(" 13595)) 5650 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13594) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5651 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13593) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5652 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13580) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (clone_argument_list 13581) (constant 13136) (dereferenceable_scalar 13137) (expr 13582) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5653 (("\"" 1538) ("$" 13382) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 13383) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 13384) ("[" 1560) (array_object_dereferenceable 13385) (callable_expr 13386) (callable_variable 13387) (class_constant 1564) (class_name 13388) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 13389) (name 13390) (new_dereferenceable 1570) (simple_variable 13391) (static_member 13392) (variable 13579) (variable_class_name 13394)) 5654 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13578) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5655 (("(" 13575)) 5656 (("(" 13572)) 5657 (("(" 13563) (argument_list 13564) (ctor_arguments 13565)) 5658 (("\"" 1538) ("$" 13382) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 13383) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 13384) ("[" 1560) (array_object_dereferenceable 13385) (callable_expr 13386) (callable_variable 13387) (class_constant 1564) (class_name 13388) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 13389) (name 13390) (new_dereferenceable 1570) (simple_variable 13391) (static_member 13392) (variable 13562) (variable_class_name 13394)) 5659 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13561) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5660 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13560) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5661 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13559) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5662 (("(" 13555)) 5663 (("(" 13550)) 5664 (("(" 13544)) 5665 (("$" 13382) ("(" 13507) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 13508) (T_NAME_QUALIFIED 13509) (T_NAME_RELATIVE 13510) (T_READONLY 2955) (T_STATIC 13511) (T_STRING 13512) (T_VARIABLE 13384) (anonymous_class 13513) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 13514) (attribute 2961) (attributes 13515) (class_modifier 2963) (class_name 13516) (class_name_reference 13517) (name 13518) (new_variable 13519) (simple_variable 13520)) 5666 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13506) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5667 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13505) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5668 (("(" 13343) (argument_list 13504)) 5669 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13503) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5670 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13502) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5671 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 13497) (T_END_HEREDOC 13498) (T_VARIABLE 2920) (encaps_list 13499) (encaps_var 2922)) 5672 ((T_FN 35) (T_FUNCTION 131) (fn 13139) #'13140 (inline_function 13496)) 5673 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13495) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5674 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13494) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5675 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13493) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5676 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13490) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5677 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13489) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5678 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 13485) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 5679 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 13483) (encaps_list 2828) (encaps_var 2829)) 5680 ((T_NULLSAFE_OBJECT_OPERATOR 13469) (T_OBJECT_OPERATOR 13470) ("[" 13471)) 5681 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 13466) (attribute 2757) (fn 13139) #'13140 (inline_function 13467)) 5682 (("(" 13343) (argument_list 13465)) 5683 ((T_PAAMAYIM_NEKUDOTAYIM 13457)) 5684 (("%" 13155) ("*" 13156) ("+" 13157) ("-" 13158) ("." 13159) ("/" 13160) ("<" 13161) (">" 13162) ("?" 13163) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 13164) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 13165) (T_BOOLEAN_AND 13166) (T_BOOLEAN_OR 13167) (T_COALESCE 13168) (T_DOUBLE_ARROW 13455) (T_INSTANCEOF 13169) (T_IS_EQUAL 13170) (T_IS_GREATER_OR_EQUAL 13171) (T_IS_IDENTICAL 13172) (T_IS_NOT_EQUAL 13173) (T_IS_NOT_IDENTICAL 13174) (T_IS_SMALLER_OR_EQUAL 13175) (T_LOGICAL_AND 13176) (T_LOGICAL_OR 13177) (T_LOGICAL_XOR 13178) (T_PIPE 13179) (T_POW 13180) (T_SL 13181) (T_SPACESHIP 13182) (T_SR 13183) ("^" 13184) ("|" 13185)) 5685 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 13444)) 5686 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 13432)) 5687 (("(" 13343) (argument_list 13431)) 5688 (("=" 13351) (T_AND_EQUAL 13352) (T_COALESCE_EQUAL 13353) (T_CONCAT_EQUAL 13354) (T_DEC 13355) (T_DIV_EQUAL 13356) (T_INC 13357) (T_MINUS_EQUAL 13358) (T_MOD_EQUAL 13359) (T_MUL_EQUAL 13360) (T_OR_EQUAL 13361) (T_PLUS_EQUAL 13362) (T_POW_EQUAL 13363) (T_SL_EQUAL 13364) (T_SR_EQUAL 13365) (T_XOR_EQUAL 13366)) 5689 ((T_PAAMAYIM_NEKUDOTAYIM 13255)) 5690 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13154) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5691 (("%" 13155) ("*" 13156) ("+" 13157) ("-" 13158) ("." 13159) ("/" 13160) ("<" 13161) (">" 13162) ("?" 13163) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 13164) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 13165) (T_BOOLEAN_AND 13166) (T_BOOLEAN_OR 13167) (T_COALESCE 13168) (T_INSTANCEOF 13169) (T_IS_EQUAL 13170) (T_IS_GREATER_OR_EQUAL 13171) (T_IS_IDENTICAL 13172) (T_IS_NOT_EQUAL 13173) (T_IS_NOT_IDENTICAL 13174) (T_IS_SMALLER_OR_EQUAL 13175) (T_LOGICAL_AND 13176) (T_LOGICAL_OR 13177) (T_LOGICAL_XOR 13178) (T_PIPE 13179) (T_POW 13180) (T_SL 13181) (T_SPACESHIP 13182) (T_SR 13183) ("^" 13184) ("|" 13185)) 5692 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13254) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5693 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13253) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5694 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13252) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5695 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13251) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5696 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13250) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5697 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13249) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5698 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13248) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5699 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13247) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5700 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 13242) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 13243) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 5701 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13241) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5702 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13240) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5703 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13239) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5704 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13238) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5705 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13237) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5706 (("$" 13202) ("(" 13203) (T_NAME_FULLY_QUALIFIED 13204) (T_NAME_QUALIFIED 13205) (T_NAME_RELATIVE 13206) (T_STATIC 13207) (T_STRING 13208) (T_VARIABLE 13209) (class_name 13210) (class_name_reference 13211) (name 13212) (new_variable 13213) (simple_variable 13214)) 5707 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13201) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5708 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13200) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5709 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13199) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5710 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13198) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5711 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13197) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5712 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13196) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5713 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13195) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5714 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13194) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5715 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13193) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5716 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13192) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5717 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13191) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5718 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13190) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5719 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13189) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5720 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13188) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5721 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13187) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5722 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13186) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5723 (("$" 13202) (T_VARIABLE 13209) (simple_variable 13233) ("{" 13234)) 5724 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 13231) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 5725 ((T_PAAMAYIM_NEKUDOTAYIM 13229)) 5726 ((T_NULLSAFE_OBJECT_OPERATOR 13215) (T_OBJECT_OPERATOR 13216) (T_PAAMAYIM_NEKUDOTAYIM 13217) ("[" 13218)) 5727 (("$" 13202) (T_STRING 13222) (T_VARIABLE 13209) (property_name 13228) (simple_variable 13224) ("{" 13225)) 5728 (("$" 13202) (T_STRING 13222) (T_VARIABLE 13209) (property_name 13223) (simple_variable 13224) ("{" 13225)) 5729 (("$" 13202) (T_VARIABLE 13209) (simple_variable 13221)) 5730 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 13219) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 5731 (("]" 13220)) 5732 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 13226) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5733 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 13227)) 5734 (("$" 13202) (T_VARIABLE 13209) (simple_variable 13230)) 5735 (("%" 304) (")" 13232) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 5736 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 13235) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5737 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 13236)) 5738 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13246) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5739 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 13244) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 5740 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13245) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5741 (("$" 13077) (T_ABSTRACT 13256) (T_ARRAY 13257) (T_AS 13258) (T_BREAK 13259) (T_CALLABLE 13260) (T_CASE 13261) (T_CATCH 13262) (T_CLASS 13263) (T_CLASS_C 13264) (T_CLONE 13265) (T_CONST 13266) (T_CONTINUE 13267) (T_DECLARE 13268) (T_DEFAULT 13269) (T_DIR 13270) (T_DO 13271) (T_ECHO 13272) (T_ELSE 13273) (T_ELSEIF 13274) (T_EMPTY 13275) (T_ENDDECLARE 13276) (T_ENDFOR 13277) (T_ENDFOREACH 13278) (T_ENDIF 13279) (T_ENDSWITCH 13280) (T_ENDWHILE 13281) (T_ENUM 13282) (T_EVAL 13283) (T_EXIT 13284) (T_EXTENDS 13285) (T_FILE 13286) (T_FINAL 13287) (T_FINALLY 13288) (T_FN 13289) (T_FOR 13290) (T_FOREACH 13291) (T_FUNCTION 13292) (T_FUNC_C 13293) (T_GLOBAL 13294) (T_GOTO 13295) (T_IF 13296) (T_IMPLEMENTS 13297) (T_INCLUDE 13298) (T_INCLUDE_ONCE 13299) (T_INSTANCEOF 13300) (T_INSTEADOF 13301) (T_INTERFACE 13302) (T_ISSET 13303) (T_LINE 13304) (T_LIST 13305) (T_LOGICAL_AND 13306) (T_LOGICAL_OR 13307) (T_LOGICAL_XOR 13308) (T_MATCH 13309) (T_METHOD_C 13310) (T_NAMESPACE 13311) (T_NEW 13312) (T_NS_C 13313) (T_PRINT 13314) (T_PRIVATE 13315) (T_PROPERTY_C 13316) (T_PROTECTED 13317) (T_PUBLIC 13318) (T_READONLY 13319) (T_REQUIRE 13320) (T_REQUIRE_ONCE 13321) (T_RETURN 13322) (T_STATIC 13323) (T_STRING 13324) (T_SWITCH 13325) (T_THROW 13326) (T_TRAIT 13327) (T_TRAIT_C 13328) (T_TRY 13329) (T_UNSET 13330) (T_USE 13331) (T_VAR 13332) (T_VARIABLE 13125) (T_WHILE 13333) (T_YIELD 13334) (identifier 13335) (member_name 13336) (reserved_non_modifiers 13337) (semi_reserved 13338) (simple_variable 13339) ("{" 13340)) 5742 (("(" 13343) (argument_list 13344)) 5743 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 13341) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5744 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 13342)) 5745 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 13345) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 13346) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 13347) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5746 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 13350) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5747 (("," 1498) (possible_comma 13348)) 5748 ((")" 13349)) 5749 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (ampersand 13380) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13381) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5750 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13379) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5751 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13378) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5752 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13377) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5753 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13376) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5754 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13375) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5755 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13374) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5756 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13373) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5757 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13372) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5758 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13371) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5759 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13370) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5760 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13369) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5761 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13368) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5762 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13367) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5763 (("\"" 1538) ("$" 13382) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 13383) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 13384) ("[" 1560) (array_object_dereferenceable 13385) (callable_expr 13386) (callable_variable 13387) (class_constant 1564) (class_name 13388) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 13389) (name 13390) (new_dereferenceable 1570) (simple_variable 13391) (static_member 13392) (variable 13393) (variable_class_name 13394)) 5764 (("$" 13382) (T_VARIABLE 13384) (simple_variable 13427) ("{" 13428)) 5765 (("(" 13398) (argument_list 13426)) 5766 ((T_NULLSAFE_OBJECT_OPERATOR 13412) (T_OBJECT_OPERATOR 13413) ("[" 13414)) 5767 (("(" 13398) (argument_list 13411)) 5768 ((T_PAAMAYIM_NEKUDOTAYIM 13407)) 5769 (("(" 13398) (argument_list 13406)) 5770 ((T_PAAMAYIM_NEKUDOTAYIM 13395)) 5771 (("$" 13382) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 13384) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 13396) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 13397) ("{" 1660)) 5772 (("(" 13398) (argument_list 13399)) 5773 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 13400) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 13401) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 13402) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5774 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 13405) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5775 (("," 1498) (possible_comma 13403)) 5776 ((")" 13404)) 5777 (("$" 13382) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 13384) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 13408) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 13409) ("{" 1676)) 5778 (("(" 13398) (argument_list 13410)) 5779 (("$" 13382) (T_STRING 13417) (T_VARIABLE 13384) (property_name 13424) (simple_variable 13419) ("{" 13420)) 5780 (("$" 13382) (T_STRING 13417) (T_VARIABLE 13384) (property_name 13418) (simple_variable 13419) ("{" 13420)) 5781 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 13415) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 5782 (("]" 13416)) 5783 (("(" 13398) (argument_list 13423)) 5784 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 13421) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5785 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 13422)) 5786 (("(" 13398) (argument_list 13425)) 5787 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 13429) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5788 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 13430)) 5789 ((backup_doc_comment 13433)) 5790 (("(" 13434)) 5791 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 13435)) 5792 ((")" 13436)) 5793 ((T_USE 166) (lexical_vars 13437)) 5794 ((":" 168) (return_type 13438)) 5795 ((backup_fn_flags 13439)) 5796 (("{" 13440)) 5797 ((inner_statement_list 13441)) 5798 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 13442) ("~" 129)) 5799 ((backup_fn_flags 13443)) 5800 ((backup_doc_comment 13445)) 5801 (("(" 13446)) 5802 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 13447)) 5803 ((")" 13448)) 5804 ((":" 2271) (return_type 13449)) 5805 ((T_DOUBLE_ARROW 13450)) 5806 ((backup_fn_flags 13451)) 5807 ((backup_lex_pos 13452)) 5808 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13453) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5809 (("%" 13155) ("*" 13156) ("+" 13157) ("-" 13158) ("." 13159) ("/" 13160) ("<" 13161) (">" 13162) ("?" 13163) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 13164) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 13165) (T_BOOLEAN_AND 13166) (T_BOOLEAN_OR 13167) (T_COALESCE 13168) (T_INSTANCEOF 13169) (T_IS_EQUAL 13170) (T_IS_GREATER_OR_EQUAL 13171) (T_IS_IDENTICAL 13172) (T_IS_NOT_EQUAL 13173) (T_IS_NOT_IDENTICAL 13174) (T_IS_SMALLER_OR_EQUAL 13175) (T_LOGICAL_AND 13176) (T_LOGICAL_OR 13177) (T_LOGICAL_XOR 13178) (T_PIPE 13179) (T_POW 13180) (T_SL 13181) (T_SPACESHIP 13182) (T_SR 13183) ("^" 13184) (backup_fn_flags 13454) ("|" 13185)) 5810 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 13456) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5811 (("$" 13077) (T_ABSTRACT 13256) (T_ARRAY 13257) (T_AS 13258) (T_BREAK 13259) (T_CALLABLE 13260) (T_CASE 13261) (T_CATCH 13262) (T_CLASS 13263) (T_CLASS_C 13264) (T_CLONE 13265) (T_CONST 13266) (T_CONTINUE 13267) (T_DECLARE 13268) (T_DEFAULT 13269) (T_DIR 13270) (T_DO 13271) (T_ECHO 13272) (T_ELSE 13273) (T_ELSEIF 13274) (T_EMPTY 13275) (T_ENDDECLARE 13276) (T_ENDFOR 13277) (T_ENDFOREACH 13278) (T_ENDIF 13279) (T_ENDSWITCH 13280) (T_ENDWHILE 13281) (T_ENUM 13282) (T_EVAL 13283) (T_EXIT 13284) (T_EXTENDS 13285) (T_FILE 13286) (T_FINAL 13287) (T_FINALLY 13288) (T_FN 13289) (T_FOR 13290) (T_FOREACH 13291) (T_FUNCTION 13292) (T_FUNC_C 13293) (T_GLOBAL 13294) (T_GOTO 13295) (T_IF 13296) (T_IMPLEMENTS 13297) (T_INCLUDE 13298) (T_INCLUDE_ONCE 13299) (T_INSTANCEOF 13300) (T_INSTEADOF 13301) (T_INTERFACE 13302) (T_ISSET 13303) (T_LINE 13304) (T_LIST 13305) (T_LOGICAL_AND 13306) (T_LOGICAL_OR 13307) (T_LOGICAL_XOR 13308) (T_MATCH 13309) (T_METHOD_C 13310) (T_NAMESPACE 13311) (T_NEW 13312) (T_NS_C 13313) (T_PRINT 13314) (T_PRIVATE 13315) (T_PROPERTY_C 13316) (T_PROTECTED 13317) (T_PUBLIC 13318) (T_READONLY 13319) (T_REQUIRE 13320) (T_REQUIRE_ONCE 13321) (T_RETURN 13322) (T_STATIC 13323) (T_STRING 13324) (T_SWITCH 13325) (T_THROW 13326) (T_TRAIT 13327) (T_TRAIT_C 13328) (T_TRY 13329) (T_UNSET 13330) (T_USE 13331) (T_VAR 13332) (T_VARIABLE 13125) (T_WHILE 13333) (T_YIELD 13334) (identifier 13458) (member_name 13459) (reserved_non_modifiers 13337) (semi_reserved 13338) (simple_variable 13460) ("{" 13461)) 5812 (("(" 13343) (argument_list 13464)) 5813 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 13462) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5814 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 13463)) 5815 ((T_FN 35) (T_FUNCTION 131) (fn 13139) #'13140 (inline_function 13468)) 5816 (("$" 13077) (T_STRING 13474) (T_VARIABLE 13125) (property_name 13481) (simple_variable 13476) ("{" 13477)) 5817 (("$" 13077) (T_STRING 13474) (T_VARIABLE 13125) (property_name 13475) (simple_variable 13476) ("{" 13477)) 5818 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 13472) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 5819 (("]" 13473)) 5820 (("(" 13343) (argument_list 13480)) 5821 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 13478) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5822 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 13479)) 5823 (("(" 13343) (argument_list 13482)) 5824 (("`" 13484)) 5825 (("]" 13486)) 5826 (("=" 13487)) 5827 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13488) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5828 (("%" 13155) ("*" 13156) ("+" 13157) ("-" 13158) ("." 13159) ("/" 13160) ("<" 13161) (">" 13162) ("?" 13163) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 13164) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 13165) (T_BOOLEAN_AND 13166) (T_BOOLEAN_OR 13167) (T_COALESCE 13168) (T_DOUBLE_ARROW 13491) (T_INSTANCEOF 13169) (T_IS_EQUAL 13170) (T_IS_GREATER_OR_EQUAL 13171) (T_IS_IDENTICAL 13172) (T_IS_NOT_EQUAL 13173) (T_IS_NOT_IDENTICAL 13174) (T_IS_SMALLER_OR_EQUAL 13175) (T_LOGICAL_AND 13176) (T_LOGICAL_OR 13177) (T_LOGICAL_XOR 13178) (T_PIPE 13179) (T_POW 13180) (T_SL 13181) (T_SPACESHIP 13182) (T_SR 13183) ("^" 13184) ("|" 13185)) 5829 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13492) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5830 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 13501) (T_VARIABLE 2920) (encaps_var 2934)) 5831 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 13500) (T_VARIABLE 2920) (encaps_var 2925)) 5832 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 13542) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 5833 ((T_CLASS 13534)) 5834 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 13533) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 13514) (attribute 2982) (class_modifier 2963)) 5835 ((T_PAAMAYIM_NEKUDOTAYIM 13531)) 5836 (("(" 13398) (argument_list 13530)) 5837 ((T_NULLSAFE_OBJECT_OPERATOR 13521) (T_OBJECT_OPERATOR 13522) (T_PAAMAYIM_NEKUDOTAYIM 13523) ("[" 13524)) 5838 (("$" 13382) (T_STRING 13417) (T_VARIABLE 13384) (property_name 13529) (simple_variable 13419) ("{" 13420)) 5839 (("$" 13382) (T_STRING 13417) (T_VARIABLE 13384) (property_name 13528) (simple_variable 13419) ("{" 13420)) 5840 (("$" 13382) (T_VARIABLE 13384) (simple_variable 13527)) 5841 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 13525) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 5842 (("]" 13526)) 5843 (("$" 13382) (T_VARIABLE 13384) (simple_variable 13532)) 5844 (("(" 2984) (argument_list 2985) (ctor_arguments 13535)) 5845 ((T_EXTENDS 2987) (extends_from 13536)) 5846 ((T_IMPLEMENTS 2989) (implements_list 13537)) 5847 ((backup_doc_comment 13538)) 5848 (("{" 13539)) 5849 ((class_statement_list 13540)) 5850 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 13541)) 5851 (("%" 304) (")" 13543) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 5852 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 13545) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 5853 (("%" 304) (")" 13546) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 5854 (("{" 13547)) 5855 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 13548) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 5856 (("}" 13549)) 5857 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 13551) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 5858 ((")" 13552)) 5859 (("=" 13553)) 5860 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 13554) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 5861 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 13556) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5862 (("," 5226) (possible_comma 13557)) 5863 ((")" 13558)) 5864 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 13566) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 13567) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 13568) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5865 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 13571) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5866 (("," 1498) (possible_comma 13569)) 5867 ((")" 13570)) 5868 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 13573) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 5869 (("%" 304) (")" 13574) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 5870 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 13576) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 5871 (("%" 304) (")" 13577) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 5872 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 13583) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 13584) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 13585) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 13586) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 5873 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 13592) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5874 (("%" 2442) (")" 13589) ("*" 2443) ("+" 2444) ("," 13590) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 5875 (("," 5287) (possible_comma 13587)) 5876 ((")" 13588)) 5877 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 13591) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5878 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 13596) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 5879 ((")" 13597)) 5880 (("%" 304) (")" 13589) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 5881 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 13604) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5882 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 13605)) 5883 (("\"" 13607) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 5884 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 13617) (T_VARIABLE 2920) (encaps_var 2934)) 5885 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 13616) (T_VARIABLE 2920) (encaps_var 2925)) 5886 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 13658) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 5887 ((T_CLASS 13650)) 5888 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 13649) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 13630) (attribute 2982) (class_modifier 2963)) 5889 ((T_PAAMAYIM_NEKUDOTAYIM 13647)) 5890 (("(" 12984) (argument_list 13646)) 5891 ((T_NULLSAFE_OBJECT_OPERATOR 13637) (T_OBJECT_OPERATOR 13638) (T_PAAMAYIM_NEKUDOTAYIM 13639) ("[" 13640)) 5892 (("$" 12968) (T_STRING 13003) (T_VARIABLE 12970) (property_name 13645) (simple_variable 13005) ("{" 13006)) 5893 (("$" 12968) (T_STRING 13003) (T_VARIABLE 12970) (property_name 13644) (simple_variable 13005) ("{" 13006)) 5894 (("$" 12968) (T_VARIABLE 12970) (simple_variable 13643)) 5895 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 13641) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 5896 (("]" 13642)) 5897 (("$" 12968) (T_VARIABLE 12970) (simple_variable 13648)) 5898 (("(" 2984) (argument_list 2985) (ctor_arguments 13651)) 5899 ((T_EXTENDS 2987) (extends_from 13652)) 5900 ((T_IMPLEMENTS 2989) (implements_list 13653)) 5901 ((backup_doc_comment 13654)) 5902 (("{" 13655)) 5903 ((class_statement_list 13656)) 5904 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 13657)) 5905 (("%" 304) (")" 13659) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 5906 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 13661) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 5907 (("%" 304) (")" 13662) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 5908 (("{" 13663)) 5909 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 13664) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 5910 (("}" 13665)) 5911 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 13667) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 5912 ((")" 13668)) 5913 (("=" 13669)) 5914 (("!" 12661) ("\"" 12662) ("$" 12663) ("(" 12664) ("+" 12665) ("-" 12666) ("@" 12667) (T_ARRAY 12668) (T_ARRAY_CAST 12669) (T_ATTRIBUTE 130) (T_BOOL_CAST 12670) (T_CLASS_C 12671) (T_CLONE 12672) (T_CONSTANT_ENCAPSED_STRING 12673) (T_DEC 12674) (T_DIR 12675) (T_DNUMBER 12676) (T_DOUBLE_CAST 12677) (T_EMPTY 12678) (T_EVAL 12679) (T_EXIT 12680) (T_FILE 12681) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 12682) (T_INC 12683) (T_INCLUDE 12684) (T_INCLUDE_ONCE 12685) (T_INT_CAST 12686) (T_ISSET 12687) (T_LINE 12688) (T_LIST 12689) (T_LNUMBER 12690) (T_MATCH 12691) (T_METHOD_C 12692) (T_NAME_FULLY_QUALIFIED 12693) (T_NAME_QUALIFIED 12694) (T_NAME_RELATIVE 12695) (T_NEW 12696) (T_NS_C 12697) (T_OBJECT_CAST 12698) (T_PRINT 12699) (T_PROPERTY_C 12700) (T_READONLY 12701) (T_REQUIRE 12702) (T_REQUIRE_ONCE 12703) (T_START_HEREDOC 12704) (T_STATIC 12705) (T_STRING 12706) (T_STRING_CAST 12707) (T_THROW 12708) (T_TRAIT_C 12709) (T_UNSET_CAST 12710) (T_VARIABLE 12711) (T_YIELD 12712) (T_YIELD_FROM 12713) ("[" 12714) ("`" 12715) (array_object_dereferenceable 12716) (attribute 135) (attributes 12717) (callable_expr 12718) (callable_variable 12719) (class_constant 12720) (class_name 12721) (constant 12722) (dereferenceable_scalar 12723) (expr 13670) (fn 12725) (fully_dereferenceable 107) #'12726 (function_call 12727) (inline_function 12728) (internal_functions_in_yacc 12729) (match 12730) (name 12731) (new_dereferenceable 12732) (new_non_dereferenceable 12733) (scalar 12734) (simple_variable 12735) (static_member 12736) (variable 12737) (variable_class_name 12738) ("~" 12739)) 5915 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 13672) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5916 (("," 5226) (possible_comma 13673)) 5917 ((")" 13674)) 5918 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 13682) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 13683) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 13684) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5919 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 13687) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5920 (("," 1498) (possible_comma 13685)) 5921 ((")" 13686)) 5922 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 13689) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 5923 (("%" 304) (")" 13690) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 5924 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 13692) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 5925 (("%" 304) (")" 13693) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 5926 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 13699) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 13700) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 13701) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 13702) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 5927 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 13708) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5928 (("%" 2442) (")" 13705) ("*" 2443) ("+" 2444) ("," 13706) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 5929 (("," 5287) (possible_comma 13703)) 5930 ((")" 13704)) 5931 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 13707) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 5932 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 13712) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 5933 ((")" 13713)) 5934 (("%" 304) (")" 13705) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 5935 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 13720) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5936 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 13721)) 5937 (("\"" 13723) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 5938 (("$" 522) (T_ABSTRACT 12464) (T_ARRAY 12465) (T_AS 12466) (T_BREAK 12467) (T_CALLABLE 12468) (T_CASE 12469) (T_CATCH 12470) (T_CLASS 12471) (T_CLASS_C 12472) (T_CLONE 12473) (T_CONST 12474) (T_CONTINUE 12475) (T_DECLARE 12476) (T_DEFAULT 12477) (T_DIR 12478) (T_DO 12479) (T_ECHO 12480) (T_ELSE 12481) (T_ELSEIF 12482) (T_EMPTY 12483) (T_ENDDECLARE 12484) (T_ENDFOR 12485) (T_ENDFOREACH 12486) (T_ENDIF 12487) (T_ENDSWITCH 12488) (T_ENDWHILE 12489) (T_ENUM 12490) (T_EVAL 12491) (T_EXIT 12492) (T_EXTENDS 12493) (T_FILE 12494) (T_FINAL 12495) (T_FINALLY 12496) (T_FN 12497) (T_FOR 12498) (T_FOREACH 12499) (T_FUNCTION 12500) (T_FUNC_C 12501) (T_GLOBAL 12502) (T_GOTO 12503) (T_IF 12504) (T_IMPLEMENTS 12505) (T_INCLUDE 12506) (T_INCLUDE_ONCE 12507) (T_INSTANCEOF 12508) (T_INSTEADOF 12509) (T_INTERFACE 12510) (T_ISSET 12511) (T_LINE 12512) (T_LIST 12513) (T_LOGICAL_AND 12514) (T_LOGICAL_OR 12515) (T_LOGICAL_XOR 12516) (T_MATCH 12517) (T_METHOD_C 12518) (T_NAMESPACE 12519) (T_NEW 12520) (T_NS_C 12521) (T_PRINT 12522) (T_PRIVATE 12523) (T_PROPERTY_C 12524) (T_PROTECTED 12525) (T_PUBLIC 12526) (T_READONLY 12527) (T_REQUIRE 12528) (T_REQUIRE_ONCE 12529) (T_RETURN 12530) (T_STATIC 12531) (T_STRING 12532) (T_SWITCH 12533) (T_THROW 12534) (T_TRAIT 12535) (T_TRAIT_C 12536) (T_TRY 12537) (T_UNSET 12538) (T_USE 12539) (T_VAR 12540) (T_VARIABLE 570) (T_WHILE 12541) (T_YIELD 12542) (identifier 13727) (member_name 13728) (reserved_non_modifiers 12545) (semi_reserved 12546) (simple_variable 13729) ("{" 13730)) 5939 (("(" 12551) (argument_list 13733)) 5940 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 13731) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5941 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 13732)) 5942 ((T_FN 35) (T_FUNCTION 131) (fn 584) #'585 (inline_function 13737)) 5943 (("$" 522) (T_STRING 13743) (T_VARIABLE 570) (property_name 13750) (simple_variable 13745) ("{" 13746)) 5944 (("$" 522) (T_STRING 13743) (T_VARIABLE 570) (property_name 13744) (simple_variable 13745) ("{" 13746)) 5945 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 13741) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 5946 (("]" 13742)) 5947 (("(" 12551) (argument_list 13749)) 5948 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 13747) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5949 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 13748)) 5950 (("(" 12551) (argument_list 13751)) 5951 (("`" 13753)) 5952 (("]" 13755)) 5953 (("=" 13756)) 5954 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 13757) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 5955 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 14292) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 5956 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 14290) (encaps_var 5357)) 5957 (("$" 13761) (T_VARIABLE 13809) (simple_variable 14286) ("{" 14287)) 5958 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 14285) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 5959 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 14284) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 5960 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 14283) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 5961 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 14282) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 5962 (("(" 14279)) 5963 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 14278) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 5964 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 14277) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 5965 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 14264) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (clone_argument_list 14265) (constant 13820) (dereferenceable_scalar 13821) (expr 14266) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 5966 (("\"" 1538) ("$" 14066) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 14067) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 14068) ("[" 1560) (array_object_dereferenceable 14069) (callable_expr 14070) (callable_variable 14071) (class_constant 1564) (class_name 14072) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 14073) (name 14074) (new_dereferenceable 1570) (simple_variable 14075) (static_member 14076) (variable 14263) (variable_class_name 14078)) 5967 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 14262) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 5968 (("(" 14259)) 5969 (("(" 14256)) 5970 (("(" 14247) (argument_list 14248) (ctor_arguments 14249)) 5971 (("\"" 1538) ("$" 14066) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 14067) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 14068) ("[" 1560) (array_object_dereferenceable 14069) (callable_expr 14070) (callable_variable 14071) (class_constant 1564) (class_name 14072) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 14073) (name 14074) (new_dereferenceable 1570) (simple_variable 14075) (static_member 14076) (variable 14246) (variable_class_name 14078)) 5972 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 14245) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 5973 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 14244) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 5974 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 14243) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 5975 (("(" 14239)) 5976 (("(" 14234)) 5977 (("(" 14228)) 5978 (("$" 14066) ("(" 14191) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 14192) (T_NAME_QUALIFIED 14193) (T_NAME_RELATIVE 14194) (T_READONLY 2955) (T_STATIC 14195) (T_STRING 14196) (T_VARIABLE 14068) (anonymous_class 14197) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 14198) (attribute 2961) (attributes 14199) (class_modifier 2963) (class_name 14200) (class_name_reference 14201) (name 14202) (new_variable 14203) (simple_variable 14204)) 5979 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 14190) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 5980 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 14189) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 5981 (("(" 14027) (argument_list 14188)) 5982 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 14187) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 5983 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 14186) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 5984 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 14181) (T_END_HEREDOC 14182) (T_VARIABLE 2920) (encaps_list 14183) (encaps_var 2922)) 5985 ((T_FN 35) (T_FUNCTION 131) (fn 13823) #'13824 (inline_function 14180)) 5986 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 14179) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 5987 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 14178) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 5988 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 14177) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 5989 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 14174) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 5990 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 14173) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 5991 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 14169) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 5992 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 14167) (encaps_list 2828) (encaps_var 2829)) 5993 ((T_NULLSAFE_OBJECT_OPERATOR 14153) (T_OBJECT_OPERATOR 14154) ("[" 14155)) 5994 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 14150) (attribute 2757) (fn 13823) #'13824 (inline_function 14151)) 5995 (("(" 14027) (argument_list 14149)) 5996 ((T_PAAMAYIM_NEKUDOTAYIM 14141)) 5997 (("%" 13839) ("*" 13840) ("+" 13841) ("-" 13842) ("." 13843) ("/" 13844) ("<" 13845) (">" 13846) ("?" 13847) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 13848) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 13849) (T_BOOLEAN_AND 13850) (T_BOOLEAN_OR 13851) (T_COALESCE 13852) (T_DOUBLE_ARROW 14139) (T_INSTANCEOF 13853) (T_IS_EQUAL 13854) (T_IS_GREATER_OR_EQUAL 13855) (T_IS_IDENTICAL 13856) (T_IS_NOT_EQUAL 13857) (T_IS_NOT_IDENTICAL 13858) (T_IS_SMALLER_OR_EQUAL 13859) (T_LOGICAL_AND 13860) (T_LOGICAL_OR 13861) (T_LOGICAL_XOR 13862) (T_PIPE 13863) (T_POW 13864) (T_SL 13865) (T_SPACESHIP 13866) (T_SR 13867) ("^" 13868) ("|" 13869)) 5998 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 14128)) 5999 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 14116)) 6000 (("(" 14027) (argument_list 14115)) 6001 (("=" 14035) (T_AND_EQUAL 14036) (T_COALESCE_EQUAL 14037) (T_CONCAT_EQUAL 14038) (T_DEC 14039) (T_DIV_EQUAL 14040) (T_INC 14041) (T_MINUS_EQUAL 14042) (T_MOD_EQUAL 14043) (T_MUL_EQUAL 14044) (T_OR_EQUAL 14045) (T_PLUS_EQUAL 14046) (T_POW_EQUAL 14047) (T_SL_EQUAL 14048) (T_SR_EQUAL 14049) (T_XOR_EQUAL 14050)) 6002 ((T_PAAMAYIM_NEKUDOTAYIM 13939)) 6003 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 13838) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6004 (("%" 13839) ("*" 13840) ("+" 13841) ("-" 13842) ("." 13843) ("/" 13844) ("<" 13845) (">" 13846) ("?" 13847) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 13848) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 13849) (T_BOOLEAN_AND 13850) (T_BOOLEAN_OR 13851) (T_COALESCE 13852) (T_INSTANCEOF 13853) (T_IS_EQUAL 13854) (T_IS_GREATER_OR_EQUAL 13855) (T_IS_IDENTICAL 13856) (T_IS_NOT_EQUAL 13857) (T_IS_NOT_IDENTICAL 13858) (T_IS_SMALLER_OR_EQUAL 13859) (T_LOGICAL_AND 13860) (T_LOGICAL_OR 13861) (T_LOGICAL_XOR 13862) (T_PIPE 13863) (T_POW 13864) (T_SL 13865) (T_SPACESHIP 13866) (T_SR 13867) ("^" 13868) ("|" 13869)) 6005 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 13938) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6006 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 13937) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6007 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 13936) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6008 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 13935) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6009 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 13934) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6010 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 13933) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6011 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 13932) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6012 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 13931) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6013 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 13926) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 13927) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 6014 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 13925) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6015 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 13924) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6016 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 13923) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6017 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 13922) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6018 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 13921) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6019 (("$" 13886) ("(" 13887) (T_NAME_FULLY_QUALIFIED 13888) (T_NAME_QUALIFIED 13889) (T_NAME_RELATIVE 13890) (T_STATIC 13891) (T_STRING 13892) (T_VARIABLE 13893) (class_name 13894) (class_name_reference 13895) (name 13896) (new_variable 13897) (simple_variable 13898)) 6020 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 13885) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6021 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 13884) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6022 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 13883) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6023 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 13882) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6024 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 13881) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6025 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 13880) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6026 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 13879) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6027 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 13878) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6028 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 13877) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6029 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 13876) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6030 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 13875) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6031 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 13874) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6032 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 13873) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6033 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 13872) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6034 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 13871) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6035 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 13870) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6036 (("$" 13886) (T_VARIABLE 13893) (simple_variable 13917) ("{" 13918)) 6037 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 13915) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 6038 ((T_PAAMAYIM_NEKUDOTAYIM 13913)) 6039 ((T_NULLSAFE_OBJECT_OPERATOR 13899) (T_OBJECT_OPERATOR 13900) (T_PAAMAYIM_NEKUDOTAYIM 13901) ("[" 13902)) 6040 (("$" 13886) (T_STRING 13906) (T_VARIABLE 13893) (property_name 13912) (simple_variable 13908) ("{" 13909)) 6041 (("$" 13886) (T_STRING 13906) (T_VARIABLE 13893) (property_name 13907) (simple_variable 13908) ("{" 13909)) 6042 (("$" 13886) (T_VARIABLE 13893) (simple_variable 13905)) 6043 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 13903) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 6044 (("]" 13904)) 6045 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 13910) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 6046 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 13911)) 6047 (("$" 13886) (T_VARIABLE 13893) (simple_variable 13914)) 6048 (("%" 304) (")" 13916) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 6049 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 13919) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 6050 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 13920)) 6051 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 13930) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6052 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 13928) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 6053 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 13929) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6054 (("$" 13761) (T_ABSTRACT 13940) (T_ARRAY 13941) (T_AS 13942) (T_BREAK 13943) (T_CALLABLE 13944) (T_CASE 13945) (T_CATCH 13946) (T_CLASS 13947) (T_CLASS_C 13948) (T_CLONE 13949) (T_CONST 13950) (T_CONTINUE 13951) (T_DECLARE 13952) (T_DEFAULT 13953) (T_DIR 13954) (T_DO 13955) (T_ECHO 13956) (T_ELSE 13957) (T_ELSEIF 13958) (T_EMPTY 13959) (T_ENDDECLARE 13960) (T_ENDFOR 13961) (T_ENDFOREACH 13962) (T_ENDIF 13963) (T_ENDSWITCH 13964) (T_ENDWHILE 13965) (T_ENUM 13966) (T_EVAL 13967) (T_EXIT 13968) (T_EXTENDS 13969) (T_FILE 13970) (T_FINAL 13971) (T_FINALLY 13972) (T_FN 13973) (T_FOR 13974) (T_FOREACH 13975) (T_FUNCTION 13976) (T_FUNC_C 13977) (T_GLOBAL 13978) (T_GOTO 13979) (T_IF 13980) (T_IMPLEMENTS 13981) (T_INCLUDE 13982) (T_INCLUDE_ONCE 13983) (T_INSTANCEOF 13984) (T_INSTEADOF 13985) (T_INTERFACE 13986) (T_ISSET 13987) (T_LINE 13988) (T_LIST 13989) (T_LOGICAL_AND 13990) (T_LOGICAL_OR 13991) (T_LOGICAL_XOR 13992) (T_MATCH 13993) (T_METHOD_C 13994) (T_NAMESPACE 13995) (T_NEW 13996) (T_NS_C 13997) (T_PRINT 13998) (T_PRIVATE 13999) (T_PROPERTY_C 14000) (T_PROTECTED 14001) (T_PUBLIC 14002) (T_READONLY 14003) (T_REQUIRE 14004) (T_REQUIRE_ONCE 14005) (T_RETURN 14006) (T_STATIC 14007) (T_STRING 14008) (T_SWITCH 14009) (T_THROW 14010) (T_TRAIT 14011) (T_TRAIT_C 14012) (T_TRY 14013) (T_UNSET 14014) (T_USE 14015) (T_VAR 14016) (T_VARIABLE 13809) (T_WHILE 14017) (T_YIELD 14018) (identifier 14019) (member_name 14020) (reserved_non_modifiers 14021) (semi_reserved 14022) (simple_variable 14023) ("{" 14024)) 6055 (("(" 14027) (argument_list 14028)) 6056 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 14025) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 6057 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 14026)) 6058 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 14029) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 14030) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 14031) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 6059 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 14034) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 6060 (("," 1498) (possible_comma 14032)) 6061 ((")" 14033)) 6062 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (ampersand 14064) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 14065) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6063 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 14063) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6064 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 14062) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6065 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 14061) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6066 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 14060) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6067 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 14059) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6068 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 14058) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6069 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 14057) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6070 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 14056) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6071 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 14055) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6072 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 14054) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6073 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 14053) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6074 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 14052) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6075 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 14051) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6076 (("\"" 1538) ("$" 14066) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 14067) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 14068) ("[" 1560) (array_object_dereferenceable 14069) (callable_expr 14070) (callable_variable 14071) (class_constant 1564) (class_name 14072) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 14073) (name 14074) (new_dereferenceable 1570) (simple_variable 14075) (static_member 14076) (variable 14077) (variable_class_name 14078)) 6077 (("$" 14066) (T_VARIABLE 14068) (simple_variable 14111) ("{" 14112)) 6078 (("(" 14082) (argument_list 14110)) 6079 ((T_NULLSAFE_OBJECT_OPERATOR 14096) (T_OBJECT_OPERATOR 14097) ("[" 14098)) 6080 (("(" 14082) (argument_list 14095)) 6081 ((T_PAAMAYIM_NEKUDOTAYIM 14091)) 6082 (("(" 14082) (argument_list 14090)) 6083 ((T_PAAMAYIM_NEKUDOTAYIM 14079)) 6084 (("$" 14066) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 14068) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 14080) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 14081) ("{" 1660)) 6085 (("(" 14082) (argument_list 14083)) 6086 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 14084) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 14085) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 14086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 6087 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 14089) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 6088 (("," 1498) (possible_comma 14087)) 6089 ((")" 14088)) 6090 (("$" 14066) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 14068) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 14092) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 14093) ("{" 1676)) 6091 (("(" 14082) (argument_list 14094)) 6092 (("$" 14066) (T_STRING 14101) (T_VARIABLE 14068) (property_name 14108) (simple_variable 14103) ("{" 14104)) 6093 (("$" 14066) (T_STRING 14101) (T_VARIABLE 14068) (property_name 14102) (simple_variable 14103) ("{" 14104)) 6094 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 14099) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 6095 (("]" 14100)) 6096 (("(" 14082) (argument_list 14107)) 6097 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 14105) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 6098 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 14106)) 6099 (("(" 14082) (argument_list 14109)) 6100 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 14113) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 6101 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 14114)) 6102 ((backup_doc_comment 14117)) 6103 (("(" 14118)) 6104 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 14119)) 6105 ((")" 14120)) 6106 ((T_USE 166) (lexical_vars 14121)) 6107 ((":" 168) (return_type 14122)) 6108 ((backup_fn_flags 14123)) 6109 (("{" 14124)) 6110 ((inner_statement_list 14125)) 6111 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 14126) ("~" 129)) 6112 ((backup_fn_flags 14127)) 6113 ((backup_doc_comment 14129)) 6114 (("(" 14130)) 6115 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 14131)) 6116 ((")" 14132)) 6117 ((":" 2271) (return_type 14133)) 6118 ((T_DOUBLE_ARROW 14134)) 6119 ((backup_fn_flags 14135)) 6120 ((backup_lex_pos 14136)) 6121 (("!" 13075) ("\"" 13076) ("$" 13077) ("(" 13078) ("+" 13079) ("-" 13080) ("@" 13081) (T_ARRAY 13082) (T_ARRAY_CAST 13083) (T_ATTRIBUTE 130) (T_BOOL_CAST 13084) (T_CLASS_C 13085) (T_CLONE 13086) (T_CONSTANT_ENCAPSED_STRING 13087) (T_DEC 13088) (T_DIR 13089) (T_DNUMBER 13090) (T_DOUBLE_CAST 13091) (T_EMPTY 13092) (T_EVAL 13093) (T_EXIT 13094) (T_FILE 13095) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13096) (T_INC 13097) (T_INCLUDE 13098) (T_INCLUDE_ONCE 13099) (T_INT_CAST 13100) (T_ISSET 13101) (T_LINE 13102) (T_LIST 13103) (T_LNUMBER 13104) (T_MATCH 13105) (T_METHOD_C 13106) (T_NAME_FULLY_QUALIFIED 13107) (T_NAME_QUALIFIED 13108) (T_NAME_RELATIVE 13109) (T_NEW 13110) (T_NS_C 13111) (T_OBJECT_CAST 13112) (T_PRINT 13113) (T_PROPERTY_C 13114) (T_READONLY 13115) (T_REQUIRE 13116) (T_REQUIRE_ONCE 13117) (T_START_HEREDOC 13118) (T_STATIC 13119) (T_STRING 13120) (T_STRING_CAST 13121) (T_THROW 13122) (T_TRAIT_C 13123) (T_UNSET_CAST 13124) (T_VARIABLE 13125) (T_YIELD 13126) (T_YIELD_FROM 13127) ("[" 13128) ("`" 13129) (array_object_dereferenceable 13130) (attribute 135) (attributes 13131) (callable_expr 13132) (callable_variable 13133) (class_constant 13134) (class_name 13135) (constant 13136) (dereferenceable_scalar 13137) (expr 14137) (fn 13139) (fully_dereferenceable 107) #'13140 (function_call 13141) (inline_function 13142) (internal_functions_in_yacc 13143) (match 13144) (name 13145) (new_dereferenceable 13146) (new_non_dereferenceable 13147) (scalar 13148) (simple_variable 13149) (static_member 13150) (variable 13151) (variable_class_name 13152) ("~" 13153)) 6122 (("%" 13155) ("*" 13156) ("+" 13157) ("-" 13158) ("." 13159) ("/" 13160) ("<" 13161) (">" 13162) ("?" 13163) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 13164) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 13165) (T_BOOLEAN_AND 13166) (T_BOOLEAN_OR 13167) (T_COALESCE 13168) (T_INSTANCEOF 13169) (T_IS_EQUAL 13170) (T_IS_GREATER_OR_EQUAL 13171) (T_IS_IDENTICAL 13172) (T_IS_NOT_EQUAL 13173) (T_IS_NOT_IDENTICAL 13174) (T_IS_SMALLER_OR_EQUAL 13175) (T_LOGICAL_AND 13176) (T_LOGICAL_OR 13177) (T_LOGICAL_XOR 13178) (T_PIPE 13179) (T_POW 13180) (T_SL 13181) (T_SPACESHIP 13182) (T_SR 13183) ("^" 13184) (backup_fn_flags 14138) ("|" 13185)) 6123 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 14140) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 6124 (("$" 13761) (T_ABSTRACT 13940) (T_ARRAY 13941) (T_AS 13942) (T_BREAK 13943) (T_CALLABLE 13944) (T_CASE 13945) (T_CATCH 13946) (T_CLASS 13947) (T_CLASS_C 13948) (T_CLONE 13949) (T_CONST 13950) (T_CONTINUE 13951) (T_DECLARE 13952) (T_DEFAULT 13953) (T_DIR 13954) (T_DO 13955) (T_ECHO 13956) (T_ELSE 13957) (T_ELSEIF 13958) (T_EMPTY 13959) (T_ENDDECLARE 13960) (T_ENDFOR 13961) (T_ENDFOREACH 13962) (T_ENDIF 13963) (T_ENDSWITCH 13964) (T_ENDWHILE 13965) (T_ENUM 13966) (T_EVAL 13967) (T_EXIT 13968) (T_EXTENDS 13969) (T_FILE 13970) (T_FINAL 13971) (T_FINALLY 13972) (T_FN 13973) (T_FOR 13974) (T_FOREACH 13975) (T_FUNCTION 13976) (T_FUNC_C 13977) (T_GLOBAL 13978) (T_GOTO 13979) (T_IF 13980) (T_IMPLEMENTS 13981) (T_INCLUDE 13982) (T_INCLUDE_ONCE 13983) (T_INSTANCEOF 13984) (T_INSTEADOF 13985) (T_INTERFACE 13986) (T_ISSET 13987) (T_LINE 13988) (T_LIST 13989) (T_LOGICAL_AND 13990) (T_LOGICAL_OR 13991) (T_LOGICAL_XOR 13992) (T_MATCH 13993) (T_METHOD_C 13994) (T_NAMESPACE 13995) (T_NEW 13996) (T_NS_C 13997) (T_PRINT 13998) (T_PRIVATE 13999) (T_PROPERTY_C 14000) (T_PROTECTED 14001) (T_PUBLIC 14002) (T_READONLY 14003) (T_REQUIRE 14004) (T_REQUIRE_ONCE 14005) (T_RETURN 14006) (T_STATIC 14007) (T_STRING 14008) (T_SWITCH 14009) (T_THROW 14010) (T_TRAIT 14011) (T_TRAIT_C 14012) (T_TRY 14013) (T_UNSET 14014) (T_USE 14015) (T_VAR 14016) (T_VARIABLE 13809) (T_WHILE 14017) (T_YIELD 14018) (identifier 14142) (member_name 14143) (reserved_non_modifiers 14021) (semi_reserved 14022) (simple_variable 14144) ("{" 14145)) 6125 (("(" 14027) (argument_list 14148)) 6126 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 14146) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 6127 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 14147)) 6128 ((T_FN 35) (T_FUNCTION 131) (fn 13823) #'13824 (inline_function 14152)) 6129 (("$" 13761) (T_STRING 14158) (T_VARIABLE 13809) (property_name 14165) (simple_variable 14160) ("{" 14161)) 6130 (("$" 13761) (T_STRING 14158) (T_VARIABLE 13809) (property_name 14159) (simple_variable 14160) ("{" 14161)) 6131 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 14156) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 6132 (("]" 14157)) 6133 (("(" 14027) (argument_list 14164)) 6134 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 14162) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 6135 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 14163)) 6136 (("(" 14027) (argument_list 14166)) 6137 (("`" 14168)) 6138 (("]" 14170)) 6139 (("=" 14171)) 6140 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 14172) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6141 (("%" 13839) ("*" 13840) ("+" 13841) ("-" 13842) ("." 13843) ("/" 13844) ("<" 13845) (">" 13846) ("?" 13847) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 13848) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 13849) (T_BOOLEAN_AND 13850) (T_BOOLEAN_OR 13851) (T_COALESCE 13852) (T_DOUBLE_ARROW 14175) (T_INSTANCEOF 13853) (T_IS_EQUAL 13854) (T_IS_GREATER_OR_EQUAL 13855) (T_IS_IDENTICAL 13856) (T_IS_NOT_EQUAL 13857) (T_IS_NOT_IDENTICAL 13858) (T_IS_SMALLER_OR_EQUAL 13859) (T_LOGICAL_AND 13860) (T_LOGICAL_OR 13861) (T_LOGICAL_XOR 13862) (T_PIPE 13863) (T_POW 13864) (T_SL 13865) (T_SPACESHIP 13866) (T_SR 13867) ("^" 13868) ("|" 13869)) 6142 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 14176) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6143 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 14185) (T_VARIABLE 2920) (encaps_var 2934)) 6144 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 14184) (T_VARIABLE 2920) (encaps_var 2925)) 6145 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 14226) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 6146 ((T_CLASS 14218)) 6147 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 14217) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 14198) (attribute 2982) (class_modifier 2963)) 6148 ((T_PAAMAYIM_NEKUDOTAYIM 14215)) 6149 (("(" 14082) (argument_list 14214)) 6150 ((T_NULLSAFE_OBJECT_OPERATOR 14205) (T_OBJECT_OPERATOR 14206) (T_PAAMAYIM_NEKUDOTAYIM 14207) ("[" 14208)) 6151 (("$" 14066) (T_STRING 14101) (T_VARIABLE 14068) (property_name 14213) (simple_variable 14103) ("{" 14104)) 6152 (("$" 14066) (T_STRING 14101) (T_VARIABLE 14068) (property_name 14212) (simple_variable 14103) ("{" 14104)) 6153 (("$" 14066) (T_VARIABLE 14068) (simple_variable 14211)) 6154 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 14209) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 6155 (("]" 14210)) 6156 (("$" 14066) (T_VARIABLE 14068) (simple_variable 14216)) 6157 (("(" 2984) (argument_list 2985) (ctor_arguments 14219)) 6158 ((T_EXTENDS 2987) (extends_from 14220)) 6159 ((T_IMPLEMENTS 2989) (implements_list 14221)) 6160 ((backup_doc_comment 14222)) 6161 (("{" 14223)) 6162 ((class_statement_list 14224)) 6163 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 14225)) 6164 (("%" 304) (")" 14227) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 6165 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 14229) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 6166 (("%" 304) (")" 14230) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 6167 (("{" 14231)) 6168 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 14232) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 6169 (("}" 14233)) 6170 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 14235) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 6171 ((")" 14236)) 6172 (("=" 14237)) 6173 (("!" 13759) ("\"" 13760) ("$" 13761) ("(" 13762) ("+" 13763) ("-" 13764) ("@" 13765) (T_ARRAY 13766) (T_ARRAY_CAST 13767) (T_ATTRIBUTE 130) (T_BOOL_CAST 13768) (T_CLASS_C 13769) (T_CLONE 13770) (T_CONSTANT_ENCAPSED_STRING 13771) (T_DEC 13772) (T_DIR 13773) (T_DNUMBER 13774) (T_DOUBLE_CAST 13775) (T_EMPTY 13776) (T_EVAL 13777) (T_EXIT 13778) (T_FILE 13779) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 13780) (T_INC 13781) (T_INCLUDE 13782) (T_INCLUDE_ONCE 13783) (T_INT_CAST 13784) (T_ISSET 13785) (T_LINE 13786) (T_LIST 13787) (T_LNUMBER 13788) (T_MATCH 13789) (T_METHOD_C 13790) (T_NAME_FULLY_QUALIFIED 13791) (T_NAME_QUALIFIED 13792) (T_NAME_RELATIVE 13793) (T_NEW 13794) (T_NS_C 13795) (T_OBJECT_CAST 13796) (T_PRINT 13797) (T_PROPERTY_C 13798) (T_READONLY 13799) (T_REQUIRE 13800) (T_REQUIRE_ONCE 13801) (T_START_HEREDOC 13802) (T_STATIC 13803) (T_STRING 13804) (T_STRING_CAST 13805) (T_THROW 13806) (T_TRAIT_C 13807) (T_UNSET_CAST 13808) (T_VARIABLE 13809) (T_YIELD 13810) (T_YIELD_FROM 13811) ("[" 13812) ("`" 13813) (array_object_dereferenceable 13814) (attribute 135) (attributes 13815) (callable_expr 13816) (callable_variable 13817) (class_constant 13818) (class_name 13819) (constant 13820) (dereferenceable_scalar 13821) (expr 14238) (fn 13823) (fully_dereferenceable 107) #'13824 (function_call 13825) (inline_function 13826) (internal_functions_in_yacc 13827) (match 13828) (name 13829) (new_dereferenceable 13830) (new_non_dereferenceable 13831) (scalar 13832) (simple_variable 13833) (static_member 13834) (variable 13835) (variable_class_name 13836) ("~" 13837)) 6174 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 14240) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 6175 (("," 5226) (possible_comma 14241)) 6176 ((")" 14242)) 6177 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 14250) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 14251) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 14252) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 6178 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 14255) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 6179 (("," 1498) (possible_comma 14253)) 6180 ((")" 14254)) 6181 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 14257) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 6182 (("%" 304) (")" 14258) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 6183 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 14260) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 6184 (("%" 304) (")" 14261) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 6185 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 14267) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 14268) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 14269) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 14270) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 6186 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 14276) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 6187 (("%" 2442) (")" 14273) ("*" 2443) ("+" 2444) ("," 14274) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 6188 (("," 5287) (possible_comma 14271)) 6189 ((")" 14272)) 6190 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 14275) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 6191 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 14280) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 6192 ((")" 14281)) 6193 (("%" 304) (")" 14273) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 6194 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 14288) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 6195 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 14289)) 6196 (("\"" 14291) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 6197 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 14301) (T_VARIABLE 2920) (encaps_var 2934)) 6198 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 14300) (T_VARIABLE 2920) (encaps_var 2925)) 6199 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 14342) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 6200 ((T_CLASS 14334)) 6201 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 14333) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 14314) (attribute 2982) (class_modifier 2963)) 6202 ((T_PAAMAYIM_NEKUDOTAYIM 14331)) 6203 (("(" 12606) (argument_list 14330)) 6204 ((T_NULLSAFE_OBJECT_OPERATOR 14321) (T_OBJECT_OPERATOR 14322) (T_PAAMAYIM_NEKUDOTAYIM 14323) ("[" 14324)) 6205 (("$" 12590) (T_STRING 12625) (T_VARIABLE 12592) (property_name 14329) (simple_variable 12627) ("{" 12628)) 6206 (("$" 12590) (T_STRING 12625) (T_VARIABLE 12592) (property_name 14328) (simple_variable 12627) ("{" 12628)) 6207 (("$" 12590) (T_VARIABLE 12592) (simple_variable 14327)) 6208 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 14325) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 6209 (("]" 14326)) 6210 (("$" 12590) (T_VARIABLE 12592) (simple_variable 14332)) 6211 (("(" 2984) (argument_list 2985) (ctor_arguments 14335)) 6212 ((T_EXTENDS 2987) (extends_from 14336)) 6213 ((T_IMPLEMENTS 2989) (implements_list 14337)) 6214 ((backup_doc_comment 14338)) 6215 (("{" 14339)) 6216 ((class_statement_list 14340)) 6217 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 14341)) 6218 (("%" 304) (")" 14343) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 6219 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 14345) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 6220 (("%" 304) (")" 14346) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 6221 (("{" 14347)) 6222 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 14348) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 6223 (("}" 14349)) 6224 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 14351) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 6225 ((")" 14352)) 6226 (("=" 14353)) 6227 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 14354) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 6228 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 14356) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 6229 (("," 5226) (possible_comma 14357)) 6230 ((")" 14358)) 6231 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 14366) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 14367) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 14368) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 6232 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 14371) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 6233 (("," 1498) (possible_comma 14369)) 6234 ((")" 14370)) 6235 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 14373) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 6236 (("%" 304) (")" 14374) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 6237 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 14376) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 6238 (("%" 304) (")" 14377) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 6239 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 14383) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 14384) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 14385) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 14386) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 6240 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 14392) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 6241 (("%" 2442) (")" 14389) ("*" 2443) ("+" 2444) ("," 14390) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 6242 (("," 5287) (possible_comma 14387)) 6243 ((")" 14388)) 6244 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 14391) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 6245 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 14396) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 6246 ((")" 14397)) 6247 (("%" 304) (")" 14389) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 6248 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 14404) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 6249 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 14405)) 6250 (("\"" 14407) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 6251 (("$" 496) (T_VARIABLE 503) (simple_variable 14411)) 6252 (("%" 304) (")" 14413) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 6253 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 14416) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 6254 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 14417)) 6255 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 14427) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 6256 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 14425) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 6257 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 14426) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 6258 (("$" 370) (T_ABSTRACT 14437) (T_ARRAY 14438) (T_AS 14439) (T_BREAK 14440) (T_CALLABLE 14441) (T_CASE 14442) (T_CATCH 14443) (T_CLASS 14444) (T_CLASS_C 14445) (T_CLONE 14446) (T_CONST 14447) (T_CONTINUE 14448) (T_DECLARE 14449) (T_DEFAULT 14450) (T_DIR 14451) (T_DO 14452) (T_ECHO 14453) (T_ELSE 14454) (T_ELSEIF 14455) (T_EMPTY 14456) (T_ENDDECLARE 14457) (T_ENDFOR 14458) (T_ENDFOREACH 14459) (T_ENDIF 14460) (T_ENDSWITCH 14461) (T_ENDWHILE 14462) (T_ENUM 14463) (T_EVAL 14464) (T_EXIT 14465) (T_EXTENDS 14466) (T_FILE 14467) (T_FINAL 14468) (T_FINALLY 14469) (T_FN 14470) (T_FOR 14471) (T_FOREACH 14472) (T_FUNCTION 14473) (T_FUNC_C 14474) (T_GLOBAL 14475) (T_GOTO 14476) (T_IF 14477) (T_IMPLEMENTS 14478) (T_INCLUDE 14479) (T_INCLUDE_ONCE 14480) (T_INSTANCEOF 14481) (T_INSTEADOF 14482) (T_INTERFACE 14483) (T_ISSET 14484) (T_LINE 14485) (T_LIST 14486) (T_LOGICAL_AND 14487) (T_LOGICAL_OR 14488) (T_LOGICAL_XOR 14489) (T_MATCH 14490) (T_METHOD_C 14491) (T_NAMESPACE 14492) (T_NEW 14493) (T_NS_C 14494) (T_PRINT 14495) (T_PRIVATE 14496) (T_PROPERTY_C 14497) (T_PROTECTED 14498) (T_PUBLIC 14499) (T_READONLY 14500) (T_REQUIRE 14501) (T_REQUIRE_ONCE 14502) (T_RETURN 14503) (T_STATIC 14504) (T_STRING 14505) (T_SWITCH 14506) (T_THROW 14507) (T_TRAIT 14508) (T_TRAIT_C 14509) (T_TRY 14510) (T_UNSET 14511) (T_USE 14512) (T_VAR 14513) (T_VARIABLE 418) (T_WHILE 14514) (T_YIELD 14515) (identifier 14516) (member_name 14517) (reserved_non_modifiers 14518) (semi_reserved 14519) (simple_variable 14520) ("{" 14521)) 6259 (("(" 14524) (argument_list 14525)) 6260 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 14522) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 6261 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 14523)) 6262 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 14526) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 14527) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 14528) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 6263 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 14531) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 6264 (("," 1498) (possible_comma 14529)) 6265 ((")" 14530)) 6266 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (ampersand 14561) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 14562) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 6267 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 14560) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 6268 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 14559) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 6269 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 14558) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 6270 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 14557) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 6271 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 14556) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 6272 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 14555) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 6273 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 14554) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 6274 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 14553) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 6275 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 14552) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 6276 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 14551) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 6277 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 14550) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 6278 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 14549) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 6279 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 14548) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 6280 (("\"" 1538) ("$" 14563) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 14564) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 14565) ("[" 1560) (array_object_dereferenceable 14566) (callable_expr 14567) (callable_variable 14568) (class_constant 1564) (class_name 14569) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 14570) (name 14571) (new_dereferenceable 1570) (simple_variable 14572) (static_member 14573) (variable 14574) (variable_class_name 14575)) 6281 (("$" 14563) (T_VARIABLE 14565) (simple_variable 14608) ("{" 14609)) 6282 (("(" 14579) (argument_list 14607)) 6283 ((T_NULLSAFE_OBJECT_OPERATOR 14593) (T_OBJECT_OPERATOR 14594) ("[" 14595)) 6284 (("(" 14579) (argument_list 14592)) 6285 ((T_PAAMAYIM_NEKUDOTAYIM 14588)) 6286 (("(" 14579) (argument_list 14587)) 6287 ((T_PAAMAYIM_NEKUDOTAYIM 14576)) 6288 (("$" 14563) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 14565) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 14577) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 14578) ("{" 1660)) 6289 (("(" 14579) (argument_list 14580)) 6290 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 14581) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 14582) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 14583) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 6291 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 14586) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 6292 (("," 1498) (possible_comma 14584)) 6293 ((")" 14585)) 6294 (("$" 14563) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 14565) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 14589) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 14590) ("{" 1676)) 6295 (("(" 14579) (argument_list 14591)) 6296 (("$" 14563) (T_STRING 14598) (T_VARIABLE 14565) (property_name 14605) (simple_variable 14600) ("{" 14601)) 6297 (("$" 14563) (T_STRING 14598) (T_VARIABLE 14565) (property_name 14599) (simple_variable 14600) ("{" 14601)) 6298 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 14596) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 6299 (("]" 14597)) 6300 (("(" 14579) (argument_list 14604)) 6301 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 14602) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 6302 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 14603)) 6303 (("(" 14579) (argument_list 14606)) 6304 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 14610) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 6305 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 14611)) 6306 ((backup_doc_comment 14615)) 6307 (("(" 14616)) 6308 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 14617)) 6309 ((")" 14618)) 6310 ((T_USE 166) (lexical_vars 14619)) 6311 ((":" 168) (return_type 14620)) 6312 ((backup_fn_flags 14621)) 6313 (("{" 14622)) 6314 ((inner_statement_list 14623)) 6315 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 14624) ("~" 129)) 6316 ((backup_fn_flags 14625)) 6317 ((backup_doc_comment 14627)) 6318 (("(" 14628)) 6319 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 14629)) 6320 ((")" 14630)) 6321 ((":" 2271) (return_type 14631)) 6322 ((T_DOUBLE_ARROW 14632)) 6323 ((backup_fn_flags 14633)) 6324 ((backup_lex_pos 14634)) 6325 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14698) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6326 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 15698) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6327 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 15696) (encaps_var 5357)) 6328 (("$" 14637) (T_VARIABLE 14685) (simple_variable 15692) ("{" 15693)) 6329 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 15691) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 6330 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 15690) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6331 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 15689) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6332 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 15688) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6333 (("(" 15685)) 6334 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 15684) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6335 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 15683) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6336 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 15670) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (clone_argument_list 15671) (constant 14696) (dereferenceable_scalar 14697) (expr 15672) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6337 (("\"" 1538) ("$" 14942) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 14943) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 14944) ("[" 1560) (array_object_dereferenceable 14945) (callable_expr 14946) (callable_variable 14947) (class_constant 1564) (class_name 14948) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 14949) (name 14950) (new_dereferenceable 1570) (simple_variable 14951) (static_member 14952) (variable 15669) (variable_class_name 14954)) 6338 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 15668) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6339 (("(" 15665)) 6340 (("(" 15662)) 6341 (("(" 15653) (argument_list 15654) (ctor_arguments 15655)) 6342 (("\"" 1538) ("$" 14942) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 14943) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 14944) ("[" 1560) (array_object_dereferenceable 14945) (callable_expr 14946) (callable_variable 14947) (class_constant 1564) (class_name 14948) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 14949) (name 14950) (new_dereferenceable 1570) (simple_variable 14951) (static_member 14952) (variable 15652) (variable_class_name 14954)) 6343 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 15651) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6344 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 15650) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6345 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 15649) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6346 (("(" 15645)) 6347 (("(" 15640)) 6348 (("(" 15634)) 6349 (("$" 14942) ("(" 15597) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 15598) (T_NAME_QUALIFIED 15599) (T_NAME_RELATIVE 15600) (T_READONLY 2955) (T_STATIC 15601) (T_STRING 15602) (T_VARIABLE 14944) (anonymous_class 15603) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 15604) (attribute 2961) (attributes 15605) (class_modifier 2963) (class_name 15606) (class_name_reference 15607) (name 15608) (new_variable 15609) (simple_variable 15610)) 6350 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 15596) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6351 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 15595) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6352 (("(" 14903) (argument_list 15594)) 6353 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 15593) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6354 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 15592) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6355 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 15587) (T_END_HEREDOC 15588) (T_VARIABLE 2920) (encaps_list 15589) (encaps_var 2922)) 6356 ((T_FN 35) (T_FUNCTION 131) (fn 14699) #'14700 (inline_function 15586)) 6357 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 15585) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6358 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 15584) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6359 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 15583) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6360 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15112) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6361 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 15048) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6362 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 15044) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 6363 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 15042) (encaps_list 2828) (encaps_var 2829)) 6364 ((T_NULLSAFE_OBJECT_OPERATOR 15028) (T_OBJECT_OPERATOR 15029) ("[" 15030)) 6365 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 15025) (attribute 2757) (fn 14699) #'14700 (inline_function 15026)) 6366 (("(" 14903) (argument_list 15024)) 6367 ((T_PAAMAYIM_NEKUDOTAYIM 15016)) 6368 (("%" 14715) ("*" 14716) ("+" 14717) ("-" 14718) ("." 14719) ("/" 14720) ("<" 14721) (">" 14722) ("?" 14723) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 14724) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 14725) (T_BOOLEAN_AND 14726) (T_BOOLEAN_OR 14727) (T_COALESCE 14728) (T_INSTANCEOF 14729) (T_IS_EQUAL 14730) (T_IS_GREATER_OR_EQUAL 14731) (T_IS_IDENTICAL 14732) (T_IS_NOT_EQUAL 14733) (T_IS_NOT_IDENTICAL 14734) (T_IS_SMALLER_OR_EQUAL 14735) (T_LOGICAL_AND 14736) (T_LOGICAL_OR 14737) (T_LOGICAL_XOR 14738) (T_PIPE 14739) (T_POW 14740) (T_SL 14741) (T_SPACESHIP 14742) (T_SR 14743) ("^" 14744) (backup_fn_flags 15015) ("|" 14745)) 6369 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 15004)) 6370 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 14992)) 6371 (("(" 14903) (argument_list 14991)) 6372 (("=" 14911) (T_AND_EQUAL 14912) (T_COALESCE_EQUAL 14913) (T_CONCAT_EQUAL 14914) (T_DEC 14915) (T_DIV_EQUAL 14916) (T_INC 14917) (T_MINUS_EQUAL 14918) (T_MOD_EQUAL 14919) (T_MUL_EQUAL 14920) (T_OR_EQUAL 14921) (T_PLUS_EQUAL 14922) (T_POW_EQUAL 14923) (T_SL_EQUAL 14924) (T_SR_EQUAL 14925) (T_XOR_EQUAL 14926)) 6373 ((T_PAAMAYIM_NEKUDOTAYIM 14815)) 6374 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14714) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6375 (("%" 14715) ("*" 14716) ("+" 14717) ("-" 14718) ("." 14719) ("/" 14720) ("<" 14721) (">" 14722) ("?" 14723) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 14724) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 14725) (T_BOOLEAN_AND 14726) (T_BOOLEAN_OR 14727) (T_COALESCE 14728) (T_INSTANCEOF 14729) (T_IS_EQUAL 14730) (T_IS_GREATER_OR_EQUAL 14731) (T_IS_IDENTICAL 14732) (T_IS_NOT_EQUAL 14733) (T_IS_NOT_IDENTICAL 14734) (T_IS_SMALLER_OR_EQUAL 14735) (T_LOGICAL_AND 14736) (T_LOGICAL_OR 14737) (T_LOGICAL_XOR 14738) (T_PIPE 14739) (T_POW 14740) (T_SL 14741) (T_SPACESHIP 14742) (T_SR 14743) ("^" 14744) ("|" 14745)) 6376 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14814) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6377 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14813) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6378 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14812) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6379 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14811) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6380 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14810) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6381 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14809) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6382 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14808) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6383 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14807) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6384 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 14802) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 14803) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 6385 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14801) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6386 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14800) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6387 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14799) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6388 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14798) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6389 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14797) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6390 (("$" 14762) ("(" 14763) (T_NAME_FULLY_QUALIFIED 14764) (T_NAME_QUALIFIED 14765) (T_NAME_RELATIVE 14766) (T_STATIC 14767) (T_STRING 14768) (T_VARIABLE 14769) (class_name 14770) (class_name_reference 14771) (name 14772) (new_variable 14773) (simple_variable 14774)) 6391 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14761) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6392 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14760) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6393 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14759) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6394 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14758) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6395 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14757) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6396 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14756) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6397 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14755) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6398 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14754) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6399 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14753) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6400 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14752) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6401 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14751) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6402 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14750) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6403 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14749) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6404 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14748) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6405 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14747) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6406 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14746) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6407 (("$" 14762) (T_VARIABLE 14769) (simple_variable 14793) ("{" 14794)) 6408 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 14791) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 6409 ((T_PAAMAYIM_NEKUDOTAYIM 14789)) 6410 ((T_NULLSAFE_OBJECT_OPERATOR 14775) (T_OBJECT_OPERATOR 14776) (T_PAAMAYIM_NEKUDOTAYIM 14777) ("[" 14778)) 6411 (("$" 14762) (T_STRING 14782) (T_VARIABLE 14769) (property_name 14788) (simple_variable 14784) ("{" 14785)) 6412 (("$" 14762) (T_STRING 14782) (T_VARIABLE 14769) (property_name 14783) (simple_variable 14784) ("{" 14785)) 6413 (("$" 14762) (T_VARIABLE 14769) (simple_variable 14781)) 6414 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 14779) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 6415 (("]" 14780)) 6416 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 14786) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 6417 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 14787)) 6418 (("$" 14762) (T_VARIABLE 14769) (simple_variable 14790)) 6419 (("%" 304) (")" 14792) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 6420 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 14795) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 6421 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 14796)) 6422 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14806) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6423 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 14804) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 6424 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14805) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6425 (("$" 14637) (T_ABSTRACT 14816) (T_ARRAY 14817) (T_AS 14818) (T_BREAK 14819) (T_CALLABLE 14820) (T_CASE 14821) (T_CATCH 14822) (T_CLASS 14823) (T_CLASS_C 14824) (T_CLONE 14825) (T_CONST 14826) (T_CONTINUE 14827) (T_DECLARE 14828) (T_DEFAULT 14829) (T_DIR 14830) (T_DO 14831) (T_ECHO 14832) (T_ELSE 14833) (T_ELSEIF 14834) (T_EMPTY 14835) (T_ENDDECLARE 14836) (T_ENDFOR 14837) (T_ENDFOREACH 14838) (T_ENDIF 14839) (T_ENDSWITCH 14840) (T_ENDWHILE 14841) (T_ENUM 14842) (T_EVAL 14843) (T_EXIT 14844) (T_EXTENDS 14845) (T_FILE 14846) (T_FINAL 14847) (T_FINALLY 14848) (T_FN 14849) (T_FOR 14850) (T_FOREACH 14851) (T_FUNCTION 14852) (T_FUNC_C 14853) (T_GLOBAL 14854) (T_GOTO 14855) (T_IF 14856) (T_IMPLEMENTS 14857) (T_INCLUDE 14858) (T_INCLUDE_ONCE 14859) (T_INSTANCEOF 14860) (T_INSTEADOF 14861) (T_INTERFACE 14862) (T_ISSET 14863) (T_LINE 14864) (T_LIST 14865) (T_LOGICAL_AND 14866) (T_LOGICAL_OR 14867) (T_LOGICAL_XOR 14868) (T_MATCH 14869) (T_METHOD_C 14870) (T_NAMESPACE 14871) (T_NEW 14872) (T_NS_C 14873) (T_PRINT 14874) (T_PRIVATE 14875) (T_PROPERTY_C 14876) (T_PROTECTED 14877) (T_PUBLIC 14878) (T_READONLY 14879) (T_REQUIRE 14880) (T_REQUIRE_ONCE 14881) (T_RETURN 14882) (T_STATIC 14883) (T_STRING 14884) (T_SWITCH 14885) (T_THROW 14886) (T_TRAIT 14887) (T_TRAIT_C 14888) (T_TRY 14889) (T_UNSET 14890) (T_USE 14891) (T_VAR 14892) (T_VARIABLE 14685) (T_WHILE 14893) (T_YIELD 14894) (identifier 14895) (member_name 14896) (reserved_non_modifiers 14897) (semi_reserved 14898) (simple_variable 14899) ("{" 14900)) 6426 (("(" 14903) (argument_list 14904)) 6427 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 14901) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 6428 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 14902)) 6429 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 14905) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 14906) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 14907) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 6430 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 14910) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 6431 (("," 1498) (possible_comma 14908)) 6432 ((")" 14909)) 6433 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (ampersand 14940) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14941) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6434 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14939) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6435 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14938) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6436 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14937) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6437 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14936) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6438 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14935) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6439 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14934) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6440 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14933) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6441 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14932) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6442 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14931) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6443 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14930) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6444 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14929) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6445 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14928) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6446 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 14927) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6447 (("\"" 1538) ("$" 14942) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 14943) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 14944) ("[" 1560) (array_object_dereferenceable 14945) (callable_expr 14946) (callable_variable 14947) (class_constant 1564) (class_name 14948) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 14949) (name 14950) (new_dereferenceable 1570) (simple_variable 14951) (static_member 14952) (variable 14953) (variable_class_name 14954)) 6448 (("$" 14942) (T_VARIABLE 14944) (simple_variable 14987) ("{" 14988)) 6449 (("(" 14958) (argument_list 14986)) 6450 ((T_NULLSAFE_OBJECT_OPERATOR 14972) (T_OBJECT_OPERATOR 14973) ("[" 14974)) 6451 (("(" 14958) (argument_list 14971)) 6452 ((T_PAAMAYIM_NEKUDOTAYIM 14967)) 6453 (("(" 14958) (argument_list 14966)) 6454 ((T_PAAMAYIM_NEKUDOTAYIM 14955)) 6455 (("$" 14942) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 14944) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 14956) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 14957) ("{" 1660)) 6456 (("(" 14958) (argument_list 14959)) 6457 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 14960) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 14961) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 14962) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 6458 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 14965) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 6459 (("," 1498) (possible_comma 14963)) 6460 ((")" 14964)) 6461 (("$" 14942) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 14944) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 14968) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 14969) ("{" 1676)) 6462 (("(" 14958) (argument_list 14970)) 6463 (("$" 14942) (T_STRING 14977) (T_VARIABLE 14944) (property_name 14984) (simple_variable 14979) ("{" 14980)) 6464 (("$" 14942) (T_STRING 14977) (T_VARIABLE 14944) (property_name 14978) (simple_variable 14979) ("{" 14980)) 6465 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 14975) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 6466 (("]" 14976)) 6467 (("(" 14958) (argument_list 14983)) 6468 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 14981) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 6469 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 14982)) 6470 (("(" 14958) (argument_list 14985)) 6471 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 14989) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 6472 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 14990)) 6473 ((backup_doc_comment 14993)) 6474 (("(" 14994)) 6475 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 14995)) 6476 ((")" 14996)) 6477 ((T_USE 166) (lexical_vars 14997)) 6478 ((":" 168) (return_type 14998)) 6479 ((backup_fn_flags 14999)) 6480 (("{" 15000)) 6481 ((inner_statement_list 15001)) 6482 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 15002) ("~" 129)) 6483 ((backup_fn_flags 15003)) 6484 ((backup_doc_comment 15005)) 6485 (("(" 15006)) 6486 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 15007)) 6487 ((")" 15008)) 6488 ((":" 2271) (return_type 15009)) 6489 ((T_DOUBLE_ARROW 15010)) 6490 ((backup_fn_flags 15011)) 6491 ((backup_lex_pos 15012)) 6492 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 15013) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6493 (("%" 14715) ("*" 14716) ("+" 14717) ("-" 14718) ("." 14719) ("/" 14720) ("<" 14721) (">" 14722) ("?" 14723) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 14724) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 14725) (T_BOOLEAN_AND 14726) (T_BOOLEAN_OR 14727) (T_COALESCE 14728) (T_INSTANCEOF 14729) (T_IS_EQUAL 14730) (T_IS_GREATER_OR_EQUAL 14731) (T_IS_IDENTICAL 14732) (T_IS_NOT_EQUAL 14733) (T_IS_NOT_IDENTICAL 14734) (T_IS_SMALLER_OR_EQUAL 14735) (T_LOGICAL_AND 14736) (T_LOGICAL_OR 14737) (T_LOGICAL_XOR 14738) (T_PIPE 14739) (T_POW 14740) (T_SL 14741) (T_SPACESHIP 14742) (T_SR 14743) ("^" 14744) (backup_fn_flags 15014) ("|" 14745)) 6494 (("$" 14637) (T_ABSTRACT 14816) (T_ARRAY 14817) (T_AS 14818) (T_BREAK 14819) (T_CALLABLE 14820) (T_CASE 14821) (T_CATCH 14822) (T_CLASS 14823) (T_CLASS_C 14824) (T_CLONE 14825) (T_CONST 14826) (T_CONTINUE 14827) (T_DECLARE 14828) (T_DEFAULT 14829) (T_DIR 14830) (T_DO 14831) (T_ECHO 14832) (T_ELSE 14833) (T_ELSEIF 14834) (T_EMPTY 14835) (T_ENDDECLARE 14836) (T_ENDFOR 14837) (T_ENDFOREACH 14838) (T_ENDIF 14839) (T_ENDSWITCH 14840) (T_ENDWHILE 14841) (T_ENUM 14842) (T_EVAL 14843) (T_EXIT 14844) (T_EXTENDS 14845) (T_FILE 14846) (T_FINAL 14847) (T_FINALLY 14848) (T_FN 14849) (T_FOR 14850) (T_FOREACH 14851) (T_FUNCTION 14852) (T_FUNC_C 14853) (T_GLOBAL 14854) (T_GOTO 14855) (T_IF 14856) (T_IMPLEMENTS 14857) (T_INCLUDE 14858) (T_INCLUDE_ONCE 14859) (T_INSTANCEOF 14860) (T_INSTEADOF 14861) (T_INTERFACE 14862) (T_ISSET 14863) (T_LINE 14864) (T_LIST 14865) (T_LOGICAL_AND 14866) (T_LOGICAL_OR 14867) (T_LOGICAL_XOR 14868) (T_MATCH 14869) (T_METHOD_C 14870) (T_NAMESPACE 14871) (T_NEW 14872) (T_NS_C 14873) (T_PRINT 14874) (T_PRIVATE 14875) (T_PROPERTY_C 14876) (T_PROTECTED 14877) (T_PUBLIC 14878) (T_READONLY 14879) (T_REQUIRE 14880) (T_REQUIRE_ONCE 14881) (T_RETURN 14882) (T_STATIC 14883) (T_STRING 14884) (T_SWITCH 14885) (T_THROW 14886) (T_TRAIT 14887) (T_TRAIT_C 14888) (T_TRY 14889) (T_UNSET 14890) (T_USE 14891) (T_VAR 14892) (T_VARIABLE 14685) (T_WHILE 14893) (T_YIELD 14894) (identifier 15017) (member_name 15018) (reserved_non_modifiers 14897) (semi_reserved 14898) (simple_variable 15019) ("{" 15020)) 6495 (("(" 14903) (argument_list 15023)) 6496 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 15021) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 6497 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 15022)) 6498 ((T_FN 35) (T_FUNCTION 131) (fn 14699) #'14700 (inline_function 15027)) 6499 (("$" 14637) (T_STRING 15033) (T_VARIABLE 14685) (property_name 15040) (simple_variable 15035) ("{" 15036)) 6500 (("$" 14637) (T_STRING 15033) (T_VARIABLE 14685) (property_name 15034) (simple_variable 15035) ("{" 15036)) 6501 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 15031) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 6502 (("]" 15032)) 6503 (("(" 14903) (argument_list 15039)) 6504 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 15037) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 6505 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 15038)) 6506 (("(" 14903) (argument_list 15041)) 6507 (("`" 15043)) 6508 (("]" 15045)) 6509 (("=" 15046)) 6510 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 15047) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6511 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15582) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6512 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 15580) (encaps_var 5357)) 6513 (("$" 15051) (T_VARIABLE 15099) (simple_variable 15576) ("{" 15577)) 6514 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 15575) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 6515 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15574) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6516 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15573) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6517 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15572) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6518 (("(" 15569)) 6519 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15568) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6520 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15567) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6521 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15554) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (clone_argument_list 15555) (constant 15110) (dereferenceable_scalar 15111) (expr 15556) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6522 (("\"" 1538) ("$" 15356) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 15357) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 15358) ("[" 1560) (array_object_dereferenceable 15359) (callable_expr 15360) (callable_variable 15361) (class_constant 1564) (class_name 15362) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 15363) (name 15364) (new_dereferenceable 1570) (simple_variable 15365) (static_member 15366) (variable 15553) (variable_class_name 15368)) 6523 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15552) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6524 (("(" 15549)) 6525 (("(" 15546)) 6526 (("(" 15537) (argument_list 15538) (ctor_arguments 15539)) 6527 (("\"" 1538) ("$" 15356) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 15357) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 15358) ("[" 1560) (array_object_dereferenceable 15359) (callable_expr 15360) (callable_variable 15361) (class_constant 1564) (class_name 15362) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 15363) (name 15364) (new_dereferenceable 1570) (simple_variable 15365) (static_member 15366) (variable 15536) (variable_class_name 15368)) 6528 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15535) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6529 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15534) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6530 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15533) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6531 (("(" 15529)) 6532 (("(" 15524)) 6533 (("(" 15518)) 6534 (("$" 15356) ("(" 15481) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 15482) (T_NAME_QUALIFIED 15483) (T_NAME_RELATIVE 15484) (T_READONLY 2955) (T_STATIC 15485) (T_STRING 15486) (T_VARIABLE 15358) (anonymous_class 15487) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 15488) (attribute 2961) (attributes 15489) (class_modifier 2963) (class_name 15490) (class_name_reference 15491) (name 15492) (new_variable 15493) (simple_variable 15494)) 6535 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15480) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6536 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15479) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6537 (("(" 15317) (argument_list 15478)) 6538 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15477) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6539 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15476) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6540 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 15471) (T_END_HEREDOC 15472) (T_VARIABLE 2920) (encaps_list 15473) (encaps_var 2922)) 6541 ((T_FN 35) (T_FUNCTION 131) (fn 15113) #'15114 (inline_function 15470)) 6542 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15469) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6543 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15468) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6544 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15467) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6545 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15464) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6546 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15463) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6547 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 15459) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 6548 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 15457) (encaps_list 2828) (encaps_var 2829)) 6549 ((T_NULLSAFE_OBJECT_OPERATOR 15443) (T_OBJECT_OPERATOR 15444) ("[" 15445)) 6550 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 15440) (attribute 2757) (fn 15113) #'15114 (inline_function 15441)) 6551 (("(" 15317) (argument_list 15439)) 6552 ((T_PAAMAYIM_NEKUDOTAYIM 15431)) 6553 (("%" 15129) ("*" 15130) ("+" 15131) ("-" 15132) ("." 15133) ("/" 15134) ("<" 15135) (">" 15136) ("?" 15137) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 15138) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 15139) (T_BOOLEAN_AND 15140) (T_BOOLEAN_OR 15141) (T_COALESCE 15142) (T_DOUBLE_ARROW 15429) (T_INSTANCEOF 15143) (T_IS_EQUAL 15144) (T_IS_GREATER_OR_EQUAL 15145) (T_IS_IDENTICAL 15146) (T_IS_NOT_EQUAL 15147) (T_IS_NOT_IDENTICAL 15148) (T_IS_SMALLER_OR_EQUAL 15149) (T_LOGICAL_AND 15150) (T_LOGICAL_OR 15151) (T_LOGICAL_XOR 15152) (T_PIPE 15153) (T_POW 15154) (T_SL 15155) (T_SPACESHIP 15156) (T_SR 15157) ("^" 15158) ("|" 15159)) 6554 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 15418)) 6555 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 15406)) 6556 (("(" 15317) (argument_list 15405)) 6557 (("=" 15325) (T_AND_EQUAL 15326) (T_COALESCE_EQUAL 15327) (T_CONCAT_EQUAL 15328) (T_DEC 15329) (T_DIV_EQUAL 15330) (T_INC 15331) (T_MINUS_EQUAL 15332) (T_MOD_EQUAL 15333) (T_MUL_EQUAL 15334) (T_OR_EQUAL 15335) (T_PLUS_EQUAL 15336) (T_POW_EQUAL 15337) (T_SL_EQUAL 15338) (T_SR_EQUAL 15339) (T_XOR_EQUAL 15340)) 6558 ((T_PAAMAYIM_NEKUDOTAYIM 15229)) 6559 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15128) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6560 (("%" 15129) ("*" 15130) ("+" 15131) ("-" 15132) ("." 15133) ("/" 15134) ("<" 15135) (">" 15136) ("?" 15137) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 15138) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 15139) (T_BOOLEAN_AND 15140) (T_BOOLEAN_OR 15141) (T_COALESCE 15142) (T_INSTANCEOF 15143) (T_IS_EQUAL 15144) (T_IS_GREATER_OR_EQUAL 15145) (T_IS_IDENTICAL 15146) (T_IS_NOT_EQUAL 15147) (T_IS_NOT_IDENTICAL 15148) (T_IS_SMALLER_OR_EQUAL 15149) (T_LOGICAL_AND 15150) (T_LOGICAL_OR 15151) (T_LOGICAL_XOR 15152) (T_PIPE 15153) (T_POW 15154) (T_SL 15155) (T_SPACESHIP 15156) (T_SR 15157) ("^" 15158) ("|" 15159)) 6561 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15228) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6562 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15227) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6563 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15226) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6564 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15225) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6565 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15224) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6566 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15223) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6567 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15222) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6568 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15221) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6569 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 15216) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 15217) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 6570 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15215) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6571 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15214) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6572 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15213) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6573 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15212) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6574 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15211) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6575 (("$" 15176) ("(" 15177) (T_NAME_FULLY_QUALIFIED 15178) (T_NAME_QUALIFIED 15179) (T_NAME_RELATIVE 15180) (T_STATIC 15181) (T_STRING 15182) (T_VARIABLE 15183) (class_name 15184) (class_name_reference 15185) (name 15186) (new_variable 15187) (simple_variable 15188)) 6576 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15175) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6577 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15174) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6578 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15173) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6579 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15172) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6580 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15171) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6581 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15170) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6582 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15169) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6583 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15168) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6584 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15167) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6585 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15166) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6586 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15165) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6587 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15164) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6588 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15163) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6589 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15162) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6590 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15161) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6591 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15160) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6592 (("$" 15176) (T_VARIABLE 15183) (simple_variable 15207) ("{" 15208)) 6593 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 15205) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 6594 ((T_PAAMAYIM_NEKUDOTAYIM 15203)) 6595 ((T_NULLSAFE_OBJECT_OPERATOR 15189) (T_OBJECT_OPERATOR 15190) (T_PAAMAYIM_NEKUDOTAYIM 15191) ("[" 15192)) 6596 (("$" 15176) (T_STRING 15196) (T_VARIABLE 15183) (property_name 15202) (simple_variable 15198) ("{" 15199)) 6597 (("$" 15176) (T_STRING 15196) (T_VARIABLE 15183) (property_name 15197) (simple_variable 15198) ("{" 15199)) 6598 (("$" 15176) (T_VARIABLE 15183) (simple_variable 15195)) 6599 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 15193) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 6600 (("]" 15194)) 6601 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 15200) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 6602 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 15201)) 6603 (("$" 15176) (T_VARIABLE 15183) (simple_variable 15204)) 6604 (("%" 304) (")" 15206) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 6605 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 15209) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 6606 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 15210)) 6607 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15220) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6608 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 15218) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 6609 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15219) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6610 (("$" 15051) (T_ABSTRACT 15230) (T_ARRAY 15231) (T_AS 15232) (T_BREAK 15233) (T_CALLABLE 15234) (T_CASE 15235) (T_CATCH 15236) (T_CLASS 15237) (T_CLASS_C 15238) (T_CLONE 15239) (T_CONST 15240) (T_CONTINUE 15241) (T_DECLARE 15242) (T_DEFAULT 15243) (T_DIR 15244) (T_DO 15245) (T_ECHO 15246) (T_ELSE 15247) (T_ELSEIF 15248) (T_EMPTY 15249) (T_ENDDECLARE 15250) (T_ENDFOR 15251) (T_ENDFOREACH 15252) (T_ENDIF 15253) (T_ENDSWITCH 15254) (T_ENDWHILE 15255) (T_ENUM 15256) (T_EVAL 15257) (T_EXIT 15258) (T_EXTENDS 15259) (T_FILE 15260) (T_FINAL 15261) (T_FINALLY 15262) (T_FN 15263) (T_FOR 15264) (T_FOREACH 15265) (T_FUNCTION 15266) (T_FUNC_C 15267) (T_GLOBAL 15268) (T_GOTO 15269) (T_IF 15270) (T_IMPLEMENTS 15271) (T_INCLUDE 15272) (T_INCLUDE_ONCE 15273) (T_INSTANCEOF 15274) (T_INSTEADOF 15275) (T_INTERFACE 15276) (T_ISSET 15277) (T_LINE 15278) (T_LIST 15279) (T_LOGICAL_AND 15280) (T_LOGICAL_OR 15281) (T_LOGICAL_XOR 15282) (T_MATCH 15283) (T_METHOD_C 15284) (T_NAMESPACE 15285) (T_NEW 15286) (T_NS_C 15287) (T_PRINT 15288) (T_PRIVATE 15289) (T_PROPERTY_C 15290) (T_PROTECTED 15291) (T_PUBLIC 15292) (T_READONLY 15293) (T_REQUIRE 15294) (T_REQUIRE_ONCE 15295) (T_RETURN 15296) (T_STATIC 15297) (T_STRING 15298) (T_SWITCH 15299) (T_THROW 15300) (T_TRAIT 15301) (T_TRAIT_C 15302) (T_TRY 15303) (T_UNSET 15304) (T_USE 15305) (T_VAR 15306) (T_VARIABLE 15099) (T_WHILE 15307) (T_YIELD 15308) (identifier 15309) (member_name 15310) (reserved_non_modifiers 15311) (semi_reserved 15312) (simple_variable 15313) ("{" 15314)) 6611 (("(" 15317) (argument_list 15318)) 6612 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 15315) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 6613 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 15316)) 6614 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 15319) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 15320) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 15321) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 6615 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 15324) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 6616 (("," 1498) (possible_comma 15322)) 6617 ((")" 15323)) 6618 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (ampersand 15354) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15355) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6619 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15353) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6620 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15352) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6621 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15351) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6622 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15350) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6623 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15349) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6624 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15348) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6625 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15347) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6626 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15346) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6627 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15345) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6628 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15344) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6629 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15343) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6630 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15342) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6631 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15341) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6632 (("\"" 1538) ("$" 15356) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 15357) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 15358) ("[" 1560) (array_object_dereferenceable 15359) (callable_expr 15360) (callable_variable 15361) (class_constant 1564) (class_name 15362) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 15363) (name 15364) (new_dereferenceable 1570) (simple_variable 15365) (static_member 15366) (variable 15367) (variable_class_name 15368)) 6633 (("$" 15356) (T_VARIABLE 15358) (simple_variable 15401) ("{" 15402)) 6634 (("(" 15372) (argument_list 15400)) 6635 ((T_NULLSAFE_OBJECT_OPERATOR 15386) (T_OBJECT_OPERATOR 15387) ("[" 15388)) 6636 (("(" 15372) (argument_list 15385)) 6637 ((T_PAAMAYIM_NEKUDOTAYIM 15381)) 6638 (("(" 15372) (argument_list 15380)) 6639 ((T_PAAMAYIM_NEKUDOTAYIM 15369)) 6640 (("$" 15356) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 15358) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 15370) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 15371) ("{" 1660)) 6641 (("(" 15372) (argument_list 15373)) 6642 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 15374) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 15375) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 15376) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 6643 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 15379) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 6644 (("," 1498) (possible_comma 15377)) 6645 ((")" 15378)) 6646 (("$" 15356) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 15358) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 15382) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 15383) ("{" 1676)) 6647 (("(" 15372) (argument_list 15384)) 6648 (("$" 15356) (T_STRING 15391) (T_VARIABLE 15358) (property_name 15398) (simple_variable 15393) ("{" 15394)) 6649 (("$" 15356) (T_STRING 15391) (T_VARIABLE 15358) (property_name 15392) (simple_variable 15393) ("{" 15394)) 6650 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 15389) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 6651 (("]" 15390)) 6652 (("(" 15372) (argument_list 15397)) 6653 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 15395) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 6654 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 15396)) 6655 (("(" 15372) (argument_list 15399)) 6656 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 15403) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 6657 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 15404)) 6658 ((backup_doc_comment 15407)) 6659 (("(" 15408)) 6660 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 15409)) 6661 ((")" 15410)) 6662 ((T_USE 166) (lexical_vars 15411)) 6663 ((":" 168) (return_type 15412)) 6664 ((backup_fn_flags 15413)) 6665 (("{" 15414)) 6666 ((inner_statement_list 15415)) 6667 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 15416) ("~" 129)) 6668 ((backup_fn_flags 15417)) 6669 ((backup_doc_comment 15419)) 6670 (("(" 15420)) 6671 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 15421)) 6672 ((")" 15422)) 6673 ((":" 2271) (return_type 15423)) 6674 ((T_DOUBLE_ARROW 15424)) 6675 ((backup_fn_flags 15425)) 6676 ((backup_lex_pos 15426)) 6677 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15427) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6678 (("%" 15129) ("*" 15130) ("+" 15131) ("-" 15132) ("." 15133) ("/" 15134) ("<" 15135) (">" 15136) ("?" 15137) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 15138) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 15139) (T_BOOLEAN_AND 15140) (T_BOOLEAN_OR 15141) (T_COALESCE 15142) (T_INSTANCEOF 15143) (T_IS_EQUAL 15144) (T_IS_GREATER_OR_EQUAL 15145) (T_IS_IDENTICAL 15146) (T_IS_NOT_EQUAL 15147) (T_IS_NOT_IDENTICAL 15148) (T_IS_SMALLER_OR_EQUAL 15149) (T_LOGICAL_AND 15150) (T_LOGICAL_OR 15151) (T_LOGICAL_XOR 15152) (T_PIPE 15153) (T_POW 15154) (T_SL 15155) (T_SPACESHIP 15156) (T_SR 15157) ("^" 15158) (backup_fn_flags 15428) ("|" 15159)) 6679 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 15430) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6680 (("$" 15051) (T_ABSTRACT 15230) (T_ARRAY 15231) (T_AS 15232) (T_BREAK 15233) (T_CALLABLE 15234) (T_CASE 15235) (T_CATCH 15236) (T_CLASS 15237) (T_CLASS_C 15238) (T_CLONE 15239) (T_CONST 15240) (T_CONTINUE 15241) (T_DECLARE 15242) (T_DEFAULT 15243) (T_DIR 15244) (T_DO 15245) (T_ECHO 15246) (T_ELSE 15247) (T_ELSEIF 15248) (T_EMPTY 15249) (T_ENDDECLARE 15250) (T_ENDFOR 15251) (T_ENDFOREACH 15252) (T_ENDIF 15253) (T_ENDSWITCH 15254) (T_ENDWHILE 15255) (T_ENUM 15256) (T_EVAL 15257) (T_EXIT 15258) (T_EXTENDS 15259) (T_FILE 15260) (T_FINAL 15261) (T_FINALLY 15262) (T_FN 15263) (T_FOR 15264) (T_FOREACH 15265) (T_FUNCTION 15266) (T_FUNC_C 15267) (T_GLOBAL 15268) (T_GOTO 15269) (T_IF 15270) (T_IMPLEMENTS 15271) (T_INCLUDE 15272) (T_INCLUDE_ONCE 15273) (T_INSTANCEOF 15274) (T_INSTEADOF 15275) (T_INTERFACE 15276) (T_ISSET 15277) (T_LINE 15278) (T_LIST 15279) (T_LOGICAL_AND 15280) (T_LOGICAL_OR 15281) (T_LOGICAL_XOR 15282) (T_MATCH 15283) (T_METHOD_C 15284) (T_NAMESPACE 15285) (T_NEW 15286) (T_NS_C 15287) (T_PRINT 15288) (T_PRIVATE 15289) (T_PROPERTY_C 15290) (T_PROTECTED 15291) (T_PUBLIC 15292) (T_READONLY 15293) (T_REQUIRE 15294) (T_REQUIRE_ONCE 15295) (T_RETURN 15296) (T_STATIC 15297) (T_STRING 15298) (T_SWITCH 15299) (T_THROW 15300) (T_TRAIT 15301) (T_TRAIT_C 15302) (T_TRY 15303) (T_UNSET 15304) (T_USE 15305) (T_VAR 15306) (T_VARIABLE 15099) (T_WHILE 15307) (T_YIELD 15308) (identifier 15432) (member_name 15433) (reserved_non_modifiers 15311) (semi_reserved 15312) (simple_variable 15434) ("{" 15435)) 6681 (("(" 15317) (argument_list 15438)) 6682 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 15436) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 6683 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 15437)) 6684 ((T_FN 35) (T_FUNCTION 131) (fn 15113) #'15114 (inline_function 15442)) 6685 (("$" 15051) (T_STRING 15448) (T_VARIABLE 15099) (property_name 15455) (simple_variable 15450) ("{" 15451)) 6686 (("$" 15051) (T_STRING 15448) (T_VARIABLE 15099) (property_name 15449) (simple_variable 15450) ("{" 15451)) 6687 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 15446) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 6688 (("]" 15447)) 6689 (("(" 15317) (argument_list 15454)) 6690 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 15452) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 6691 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 15453)) 6692 (("(" 15317) (argument_list 15456)) 6693 (("`" 15458)) 6694 (("]" 15460)) 6695 (("=" 15461)) 6696 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15462) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6697 (("%" 15129) ("*" 15130) ("+" 15131) ("-" 15132) ("." 15133) ("/" 15134) ("<" 15135) (">" 15136) ("?" 15137) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 15138) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 15139) (T_BOOLEAN_AND 15140) (T_BOOLEAN_OR 15141) (T_COALESCE 15142) (T_DOUBLE_ARROW 15465) (T_INSTANCEOF 15143) (T_IS_EQUAL 15144) (T_IS_GREATER_OR_EQUAL 15145) (T_IS_IDENTICAL 15146) (T_IS_NOT_EQUAL 15147) (T_IS_NOT_IDENTICAL 15148) (T_IS_SMALLER_OR_EQUAL 15149) (T_LOGICAL_AND 15150) (T_LOGICAL_OR 15151) (T_LOGICAL_XOR 15152) (T_PIPE 15153) (T_POW 15154) (T_SL 15155) (T_SPACESHIP 15156) (T_SR 15157) ("^" 15158) ("|" 15159)) 6698 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15466) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6699 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 15475) (T_VARIABLE 2920) (encaps_var 2934)) 6700 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 15474) (T_VARIABLE 2920) (encaps_var 2925)) 6701 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 15516) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 6702 ((T_CLASS 15508)) 6703 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 15507) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 15488) (attribute 2982) (class_modifier 2963)) 6704 ((T_PAAMAYIM_NEKUDOTAYIM 15505)) 6705 (("(" 15372) (argument_list 15504)) 6706 ((T_NULLSAFE_OBJECT_OPERATOR 15495) (T_OBJECT_OPERATOR 15496) (T_PAAMAYIM_NEKUDOTAYIM 15497) ("[" 15498)) 6707 (("$" 15356) (T_STRING 15391) (T_VARIABLE 15358) (property_name 15503) (simple_variable 15393) ("{" 15394)) 6708 (("$" 15356) (T_STRING 15391) (T_VARIABLE 15358) (property_name 15502) (simple_variable 15393) ("{" 15394)) 6709 (("$" 15356) (T_VARIABLE 15358) (simple_variable 15501)) 6710 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 15499) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 6711 (("]" 15500)) 6712 (("$" 15356) (T_VARIABLE 15358) (simple_variable 15506)) 6713 (("(" 2984) (argument_list 2985) (ctor_arguments 15509)) 6714 ((T_EXTENDS 2987) (extends_from 15510)) 6715 ((T_IMPLEMENTS 2989) (implements_list 15511)) 6716 ((backup_doc_comment 15512)) 6717 (("{" 15513)) 6718 ((class_statement_list 15514)) 6719 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 15515)) 6720 (("%" 304) (")" 15517) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 6721 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 15519) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 6722 (("%" 304) (")" 15520) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 6723 (("{" 15521)) 6724 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 15522) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 6725 (("}" 15523)) 6726 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 15525) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 6727 ((")" 15526)) 6728 (("=" 15527)) 6729 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 15528) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6730 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 15530) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 6731 (("," 5226) (possible_comma 15531)) 6732 ((")" 15532)) 6733 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 15540) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 15541) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 15542) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 6734 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 15545) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 6735 (("," 1498) (possible_comma 15543)) 6736 ((")" 15544)) 6737 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 15547) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 6738 (("%" 304) (")" 15548) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 6739 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 15550) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 6740 (("%" 304) (")" 15551) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 6741 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 15557) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 15558) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 15559) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 15560) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 6742 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 15566) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 6743 (("%" 2442) (")" 15563) ("*" 2443) ("+" 2444) ("," 15564) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 6744 (("," 5287) (possible_comma 15561)) 6745 ((")" 15562)) 6746 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 15565) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 6747 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 15570) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 6748 ((")" 15571)) 6749 (("%" 304) (")" 15563) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 6750 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 15578) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 6751 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 15579)) 6752 (("\"" 15581) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 6753 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 15591) (T_VARIABLE 2920) (encaps_var 2934)) 6754 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 15590) (T_VARIABLE 2920) (encaps_var 2925)) 6755 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 15632) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 6756 ((T_CLASS 15624)) 6757 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 15623) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 15604) (attribute 2982) (class_modifier 2963)) 6758 ((T_PAAMAYIM_NEKUDOTAYIM 15621)) 6759 (("(" 14958) (argument_list 15620)) 6760 ((T_NULLSAFE_OBJECT_OPERATOR 15611) (T_OBJECT_OPERATOR 15612) (T_PAAMAYIM_NEKUDOTAYIM 15613) ("[" 15614)) 6761 (("$" 14942) (T_STRING 14977) (T_VARIABLE 14944) (property_name 15619) (simple_variable 14979) ("{" 14980)) 6762 (("$" 14942) (T_STRING 14977) (T_VARIABLE 14944) (property_name 15618) (simple_variable 14979) ("{" 14980)) 6763 (("$" 14942) (T_VARIABLE 14944) (simple_variable 15617)) 6764 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 15615) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 6765 (("]" 15616)) 6766 (("$" 14942) (T_VARIABLE 14944) (simple_variable 15622)) 6767 (("(" 2984) (argument_list 2985) (ctor_arguments 15625)) 6768 ((T_EXTENDS 2987) (extends_from 15626)) 6769 ((T_IMPLEMENTS 2989) (implements_list 15627)) 6770 ((backup_doc_comment 15628)) 6771 (("{" 15629)) 6772 ((class_statement_list 15630)) 6773 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 15631)) 6774 (("%" 304) (")" 15633) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 6775 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 15635) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 6776 (("%" 304) (")" 15636) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 6777 (("{" 15637)) 6778 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 15638) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 6779 (("}" 15639)) 6780 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 15641) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 6781 ((")" 15642)) 6782 (("=" 15643)) 6783 (("!" 14635) ("\"" 14636) ("$" 14637) ("(" 14638) ("+" 14639) ("-" 14640) ("@" 14641) (T_ARRAY 14642) (T_ARRAY_CAST 14643) (T_ATTRIBUTE 130) (T_BOOL_CAST 14644) (T_CLASS_C 14645) (T_CLONE 14646) (T_CONSTANT_ENCAPSED_STRING 14647) (T_DEC 14648) (T_DIR 14649) (T_DNUMBER 14650) (T_DOUBLE_CAST 14651) (T_EMPTY 14652) (T_EVAL 14653) (T_EXIT 14654) (T_FILE 14655) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 14656) (T_INC 14657) (T_INCLUDE 14658) (T_INCLUDE_ONCE 14659) (T_INT_CAST 14660) (T_ISSET 14661) (T_LINE 14662) (T_LIST 14663) (T_LNUMBER 14664) (T_MATCH 14665) (T_METHOD_C 14666) (T_NAME_FULLY_QUALIFIED 14667) (T_NAME_QUALIFIED 14668) (T_NAME_RELATIVE 14669) (T_NEW 14670) (T_NS_C 14671) (T_OBJECT_CAST 14672) (T_PRINT 14673) (T_PROPERTY_C 14674) (T_READONLY 14675) (T_REQUIRE 14676) (T_REQUIRE_ONCE 14677) (T_START_HEREDOC 14678) (T_STATIC 14679) (T_STRING 14680) (T_STRING_CAST 14681) (T_THROW 14682) (T_TRAIT_C 14683) (T_UNSET_CAST 14684) (T_VARIABLE 14685) (T_YIELD 14686) (T_YIELD_FROM 14687) ("[" 14688) ("`" 14689) (array_object_dereferenceable 14690) (attribute 135) (attributes 14691) (callable_expr 14692) (callable_variable 14693) (class_constant 14694) (class_name 14695) (constant 14696) (dereferenceable_scalar 14697) (expr 15644) (fn 14699) (fully_dereferenceable 107) #'14700 (function_call 14701) (inline_function 14702) (internal_functions_in_yacc 14703) (match 14704) (name 14705) (new_dereferenceable 14706) (new_non_dereferenceable 14707) (scalar 14708) (simple_variable 14709) (static_member 14710) (variable 14711) (variable_class_name 14712) ("~" 14713)) 6784 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 15646) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 6785 (("," 5226) (possible_comma 15647)) 6786 ((")" 15648)) 6787 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 15656) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 15657) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 15658) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 6788 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 15661) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 6789 (("," 1498) (possible_comma 15659)) 6790 ((")" 15660)) 6791 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 15663) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 6792 (("%" 304) (")" 15664) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 6793 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 15666) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 6794 (("%" 304) (")" 15667) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 6795 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 15673) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 15674) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 15675) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 15676) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 6796 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 15682) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 6797 (("%" 2442) (")" 15679) ("*" 2443) ("+" 2444) ("," 15680) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 6798 (("," 5287) (possible_comma 15677)) 6799 ((")" 15678)) 6800 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 15681) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 6801 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 15686) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 6802 ((")" 15687)) 6803 (("%" 304) (")" 15679) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 6804 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 15694) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 6805 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 15695)) 6806 (("\"" 15697) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 6807 (("$" 370) (T_ABSTRACT 14437) (T_ARRAY 14438) (T_AS 14439) (T_BREAK 14440) (T_CALLABLE 14441) (T_CASE 14442) (T_CATCH 14443) (T_CLASS 14444) (T_CLASS_C 14445) (T_CLONE 14446) (T_CONST 14447) (T_CONTINUE 14448) (T_DECLARE 14449) (T_DEFAULT 14450) (T_DIR 14451) (T_DO 14452) (T_ECHO 14453) (T_ELSE 14454) (T_ELSEIF 14455) (T_EMPTY 14456) (T_ENDDECLARE 14457) (T_ENDFOR 14458) (T_ENDFOREACH 14459) (T_ENDIF 14460) (T_ENDSWITCH 14461) (T_ENDWHILE 14462) (T_ENUM 14463) (T_EVAL 14464) (T_EXIT 14465) (T_EXTENDS 14466) (T_FILE 14467) (T_FINAL 14468) (T_FINALLY 14469) (T_FN 14470) (T_FOR 14471) (T_FOREACH 14472) (T_FUNCTION 14473) (T_FUNC_C 14474) (T_GLOBAL 14475) (T_GOTO 14476) (T_IF 14477) (T_IMPLEMENTS 14478) (T_INCLUDE 14479) (T_INCLUDE_ONCE 14480) (T_INSTANCEOF 14481) (T_INSTEADOF 14482) (T_INTERFACE 14483) (T_ISSET 14484) (T_LINE 14485) (T_LIST 14486) (T_LOGICAL_AND 14487) (T_LOGICAL_OR 14488) (T_LOGICAL_XOR 14489) (T_MATCH 14490) (T_METHOD_C 14491) (T_NAMESPACE 14492) (T_NEW 14493) (T_NS_C 14494) (T_PRINT 14495) (T_PRIVATE 14496) (T_PROPERTY_C 14497) (T_PROTECTED 14498) (T_PUBLIC 14499) (T_READONLY 14500) (T_REQUIRE 14501) (T_REQUIRE_ONCE 14502) (T_RETURN 14503) (T_STATIC 14504) (T_STRING 14505) (T_SWITCH 14506) (T_THROW 14507) (T_TRAIT 14508) (T_TRAIT_C 14509) (T_TRY 14510) (T_UNSET 14511) (T_USE 14512) (T_VAR 14513) (T_VARIABLE 418) (T_WHILE 14514) (T_YIELD 14515) (identifier 15700) (member_name 15701) (reserved_non_modifiers 14518) (semi_reserved 14519) (simple_variable 15702) ("{" 15703)) 6808 (("(" 14524) (argument_list 15706)) 6809 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 15704) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 6810 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 15705)) 6811 ((T_FN 35) (T_FUNCTION 131) (fn 432) #'433 (inline_function 15710)) 6812 (("$" 370) (T_STRING 15716) (T_VARIABLE 418) (property_name 15723) (simple_variable 15718) ("{" 15719)) 6813 (("$" 370) (T_STRING 15716) (T_VARIABLE 418) (property_name 15717) (simple_variable 15718) ("{" 15719)) 6814 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 15714) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 6815 (("]" 15715)) 6816 (("(" 14524) (argument_list 15722)) 6817 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 15720) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 6818 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 15721)) 6819 (("(" 14524) (argument_list 15724)) 6820 (("`" 15726)) 6821 (("]" 15728)) 6822 (("=" 15729)) 6823 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 15730) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 6824 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 16265) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6825 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 16263) (encaps_var 5357)) 6826 (("$" 15734) (T_VARIABLE 15782) (simple_variable 16259) ("{" 16260)) 6827 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16258) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 6828 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 16257) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6829 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 16256) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6830 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 16255) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6831 (("(" 16252)) 6832 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 16251) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6833 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 16250) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6834 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 16237) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (clone_argument_list 16238) (constant 15793) (dereferenceable_scalar 15794) (expr 16239) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6835 (("\"" 1538) ("$" 16039) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 16040) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 16041) ("[" 1560) (array_object_dereferenceable 16042) (callable_expr 16043) (callable_variable 16044) (class_constant 1564) (class_name 16045) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 16046) (name 16047) (new_dereferenceable 1570) (simple_variable 16048) (static_member 16049) (variable 16236) (variable_class_name 16051)) 6836 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 16235) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6837 (("(" 16232)) 6838 (("(" 16229)) 6839 (("(" 16220) (argument_list 16221) (ctor_arguments 16222)) 6840 (("\"" 1538) ("$" 16039) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 16040) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 16041) ("[" 1560) (array_object_dereferenceable 16042) (callable_expr 16043) (callable_variable 16044) (class_constant 1564) (class_name 16045) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 16046) (name 16047) (new_dereferenceable 1570) (simple_variable 16048) (static_member 16049) (variable 16219) (variable_class_name 16051)) 6841 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 16218) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6842 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 16217) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6843 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 16216) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6844 (("(" 16212)) 6845 (("(" 16207)) 6846 (("(" 16201)) 6847 (("$" 16039) ("(" 16164) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 16165) (T_NAME_QUALIFIED 16166) (T_NAME_RELATIVE 16167) (T_READONLY 2955) (T_STATIC 16168) (T_STRING 16169) (T_VARIABLE 16041) (anonymous_class 16170) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 16171) (attribute 2961) (attributes 16172) (class_modifier 2963) (class_name 16173) (class_name_reference 16174) (name 16175) (new_variable 16176) (simple_variable 16177)) 6848 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 16163) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6849 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 16162) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6850 (("(" 16000) (argument_list 16161)) 6851 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 16160) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6852 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 16159) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6853 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 16154) (T_END_HEREDOC 16155) (T_VARIABLE 2920) (encaps_list 16156) (encaps_var 2922)) 6854 ((T_FN 35) (T_FUNCTION 131) (fn 15796) #'15797 (inline_function 16153)) 6855 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 16152) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6856 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 16151) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6857 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 16150) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6858 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 16147) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6859 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 16146) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6860 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 16142) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 6861 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 16140) (encaps_list 2828) (encaps_var 2829)) 6862 ((T_NULLSAFE_OBJECT_OPERATOR 16126) (T_OBJECT_OPERATOR 16127) ("[" 16128)) 6863 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 16123) (attribute 2757) (fn 15796) #'15797 (inline_function 16124)) 6864 (("(" 16000) (argument_list 16122)) 6865 ((T_PAAMAYIM_NEKUDOTAYIM 16114)) 6866 (("%" 15812) ("*" 15813) ("+" 15814) ("-" 15815) ("." 15816) ("/" 15817) ("<" 15818) (">" 15819) ("?" 15820) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 15821) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 15822) (T_BOOLEAN_AND 15823) (T_BOOLEAN_OR 15824) (T_COALESCE 15825) (T_DOUBLE_ARROW 16112) (T_INSTANCEOF 15826) (T_IS_EQUAL 15827) (T_IS_GREATER_OR_EQUAL 15828) (T_IS_IDENTICAL 15829) (T_IS_NOT_EQUAL 15830) (T_IS_NOT_IDENTICAL 15831) (T_IS_SMALLER_OR_EQUAL 15832) (T_LOGICAL_AND 15833) (T_LOGICAL_OR 15834) (T_LOGICAL_XOR 15835) (T_PIPE 15836) (T_POW 15837) (T_SL 15838) (T_SPACESHIP 15839) (T_SR 15840) ("^" 15841) ("|" 15842)) 6867 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 16101)) 6868 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 16089)) 6869 (("(" 16000) (argument_list 16088)) 6870 (("=" 16008) (T_AND_EQUAL 16009) (T_COALESCE_EQUAL 16010) (T_CONCAT_EQUAL 16011) (T_DEC 16012) (T_DIV_EQUAL 16013) (T_INC 16014) (T_MINUS_EQUAL 16015) (T_MOD_EQUAL 16016) (T_MUL_EQUAL 16017) (T_OR_EQUAL 16018) (T_PLUS_EQUAL 16019) (T_POW_EQUAL 16020) (T_SL_EQUAL 16021) (T_SR_EQUAL 16022) (T_XOR_EQUAL 16023)) 6871 ((T_PAAMAYIM_NEKUDOTAYIM 15912)) 6872 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 15811) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6873 (("%" 15812) ("*" 15813) ("+" 15814) ("-" 15815) ("." 15816) ("/" 15817) ("<" 15818) (">" 15819) ("?" 15820) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 15821) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 15822) (T_BOOLEAN_AND 15823) (T_BOOLEAN_OR 15824) (T_COALESCE 15825) (T_INSTANCEOF 15826) (T_IS_EQUAL 15827) (T_IS_GREATER_OR_EQUAL 15828) (T_IS_IDENTICAL 15829) (T_IS_NOT_EQUAL 15830) (T_IS_NOT_IDENTICAL 15831) (T_IS_SMALLER_OR_EQUAL 15832) (T_LOGICAL_AND 15833) (T_LOGICAL_OR 15834) (T_LOGICAL_XOR 15835) (T_PIPE 15836) (T_POW 15837) (T_SL 15838) (T_SPACESHIP 15839) (T_SR 15840) ("^" 15841) ("|" 15842)) 6874 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 15911) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6875 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 15910) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6876 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 15909) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6877 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 15908) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6878 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 15907) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6879 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 15906) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6880 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 15905) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6881 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 15904) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6882 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 15899) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 15900) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 6883 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 15898) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6884 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 15897) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6885 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 15896) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6886 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 15895) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6887 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 15894) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6888 (("$" 15859) ("(" 15860) (T_NAME_FULLY_QUALIFIED 15861) (T_NAME_QUALIFIED 15862) (T_NAME_RELATIVE 15863) (T_STATIC 15864) (T_STRING 15865) (T_VARIABLE 15866) (class_name 15867) (class_name_reference 15868) (name 15869) (new_variable 15870) (simple_variable 15871)) 6889 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 15858) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6890 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 15857) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6891 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 15856) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6892 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 15855) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6893 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 15854) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6894 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 15853) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6895 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 15852) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6896 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 15851) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6897 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 15850) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6898 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 15849) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6899 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 15848) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6900 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 15847) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6901 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 15846) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6902 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 15845) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6903 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 15844) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6904 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 15843) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6905 (("$" 15859) (T_VARIABLE 15866) (simple_variable 15890) ("{" 15891)) 6906 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 15888) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 6907 ((T_PAAMAYIM_NEKUDOTAYIM 15886)) 6908 ((T_NULLSAFE_OBJECT_OPERATOR 15872) (T_OBJECT_OPERATOR 15873) (T_PAAMAYIM_NEKUDOTAYIM 15874) ("[" 15875)) 6909 (("$" 15859) (T_STRING 15879) (T_VARIABLE 15866) (property_name 15885) (simple_variable 15881) ("{" 15882)) 6910 (("$" 15859) (T_STRING 15879) (T_VARIABLE 15866) (property_name 15880) (simple_variable 15881) ("{" 15882)) 6911 (("$" 15859) (T_VARIABLE 15866) (simple_variable 15878)) 6912 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 15876) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 6913 (("]" 15877)) 6914 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 15883) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 6915 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 15884)) 6916 (("$" 15859) (T_VARIABLE 15866) (simple_variable 15887)) 6917 (("%" 304) (")" 15889) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 6918 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 15892) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 6919 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 15893)) 6920 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 15903) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6921 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 15901) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 6922 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 15902) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6923 (("$" 15734) (T_ABSTRACT 15913) (T_ARRAY 15914) (T_AS 15915) (T_BREAK 15916) (T_CALLABLE 15917) (T_CASE 15918) (T_CATCH 15919) (T_CLASS 15920) (T_CLASS_C 15921) (T_CLONE 15922) (T_CONST 15923) (T_CONTINUE 15924) (T_DECLARE 15925) (T_DEFAULT 15926) (T_DIR 15927) (T_DO 15928) (T_ECHO 15929) (T_ELSE 15930) (T_ELSEIF 15931) (T_EMPTY 15932) (T_ENDDECLARE 15933) (T_ENDFOR 15934) (T_ENDFOREACH 15935) (T_ENDIF 15936) (T_ENDSWITCH 15937) (T_ENDWHILE 15938) (T_ENUM 15939) (T_EVAL 15940) (T_EXIT 15941) (T_EXTENDS 15942) (T_FILE 15943) (T_FINAL 15944) (T_FINALLY 15945) (T_FN 15946) (T_FOR 15947) (T_FOREACH 15948) (T_FUNCTION 15949) (T_FUNC_C 15950) (T_GLOBAL 15951) (T_GOTO 15952) (T_IF 15953) (T_IMPLEMENTS 15954) (T_INCLUDE 15955) (T_INCLUDE_ONCE 15956) (T_INSTANCEOF 15957) (T_INSTEADOF 15958) (T_INTERFACE 15959) (T_ISSET 15960) (T_LINE 15961) (T_LIST 15962) (T_LOGICAL_AND 15963) (T_LOGICAL_OR 15964) (T_LOGICAL_XOR 15965) (T_MATCH 15966) (T_METHOD_C 15967) (T_NAMESPACE 15968) (T_NEW 15969) (T_NS_C 15970) (T_PRINT 15971) (T_PRIVATE 15972) (T_PROPERTY_C 15973) (T_PROTECTED 15974) (T_PUBLIC 15975) (T_READONLY 15976) (T_REQUIRE 15977) (T_REQUIRE_ONCE 15978) (T_RETURN 15979) (T_STATIC 15980) (T_STRING 15981) (T_SWITCH 15982) (T_THROW 15983) (T_TRAIT 15984) (T_TRAIT_C 15985) (T_TRY 15986) (T_UNSET 15987) (T_USE 15988) (T_VAR 15989) (T_VARIABLE 15782) (T_WHILE 15990) (T_YIELD 15991) (identifier 15992) (member_name 15993) (reserved_non_modifiers 15994) (semi_reserved 15995) (simple_variable 15996) ("{" 15997)) 6924 (("(" 16000) (argument_list 16001)) 6925 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 15998) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 6926 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 15999)) 6927 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 16002) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 16003) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 16004) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 6928 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 16007) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 6929 (("," 1498) (possible_comma 16005)) 6930 ((")" 16006)) 6931 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (ampersand 16037) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 16038) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6932 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 16036) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6933 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 16035) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6934 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 16034) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6935 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 16033) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6936 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 16032) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6937 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 16031) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6938 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 16030) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6939 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 16029) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6940 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 16028) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6941 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 16027) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6942 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 16026) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6943 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 16025) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6944 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 16024) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 6945 (("\"" 1538) ("$" 16039) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 16040) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 16041) ("[" 1560) (array_object_dereferenceable 16042) (callable_expr 16043) (callable_variable 16044) (class_constant 1564) (class_name 16045) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 16046) (name 16047) (new_dereferenceable 1570) (simple_variable 16048) (static_member 16049) (variable 16050) (variable_class_name 16051)) 6946 (("$" 16039) (T_VARIABLE 16041) (simple_variable 16084) ("{" 16085)) 6947 (("(" 16055) (argument_list 16083)) 6948 ((T_NULLSAFE_OBJECT_OPERATOR 16069) (T_OBJECT_OPERATOR 16070) ("[" 16071)) 6949 (("(" 16055) (argument_list 16068)) 6950 ((T_PAAMAYIM_NEKUDOTAYIM 16064)) 6951 (("(" 16055) (argument_list 16063)) 6952 ((T_PAAMAYIM_NEKUDOTAYIM 16052)) 6953 (("$" 16039) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 16041) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 16053) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 16054) ("{" 1660)) 6954 (("(" 16055) (argument_list 16056)) 6955 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 16057) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 16058) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 16059) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 6956 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 16062) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 6957 (("," 1498) (possible_comma 16060)) 6958 ((")" 16061)) 6959 (("$" 16039) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 16041) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 16065) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 16066) ("{" 1676)) 6960 (("(" 16055) (argument_list 16067)) 6961 (("$" 16039) (T_STRING 16074) (T_VARIABLE 16041) (property_name 16081) (simple_variable 16076) ("{" 16077)) 6962 (("$" 16039) (T_STRING 16074) (T_VARIABLE 16041) (property_name 16075) (simple_variable 16076) ("{" 16077)) 6963 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 16072) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 6964 (("]" 16073)) 6965 (("(" 16055) (argument_list 16080)) 6966 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 16078) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 6967 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 16079)) 6968 (("(" 16055) (argument_list 16082)) 6969 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 16086) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 6970 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 16087)) 6971 ((backup_doc_comment 16090)) 6972 (("(" 16091)) 6973 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 16092)) 6974 ((")" 16093)) 6975 ((T_USE 166) (lexical_vars 16094)) 6976 ((":" 168) (return_type 16095)) 6977 ((backup_fn_flags 16096)) 6978 (("{" 16097)) 6979 ((inner_statement_list 16098)) 6980 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 16099) ("~" 129)) 6981 ((backup_fn_flags 16100)) 6982 ((backup_doc_comment 16102)) 6983 (("(" 16103)) 6984 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 16104)) 6985 ((")" 16105)) 6986 ((":" 2271) (return_type 16106)) 6987 ((T_DOUBLE_ARROW 16107)) 6988 ((backup_fn_flags 16108)) 6989 ((backup_lex_pos 16109)) 6990 (("!" 15049) ("\"" 15050) ("$" 15051) ("(" 15052) ("+" 15053) ("-" 15054) ("@" 15055) (T_ARRAY 15056) (T_ARRAY_CAST 15057) (T_ATTRIBUTE 130) (T_BOOL_CAST 15058) (T_CLASS_C 15059) (T_CLONE 15060) (T_CONSTANT_ENCAPSED_STRING 15061) (T_DEC 15062) (T_DIR 15063) (T_DNUMBER 15064) (T_DOUBLE_CAST 15065) (T_EMPTY 15066) (T_EVAL 15067) (T_EXIT 15068) (T_FILE 15069) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15070) (T_INC 15071) (T_INCLUDE 15072) (T_INCLUDE_ONCE 15073) (T_INT_CAST 15074) (T_ISSET 15075) (T_LINE 15076) (T_LIST 15077) (T_LNUMBER 15078) (T_MATCH 15079) (T_METHOD_C 15080) (T_NAME_FULLY_QUALIFIED 15081) (T_NAME_QUALIFIED 15082) (T_NAME_RELATIVE 15083) (T_NEW 15084) (T_NS_C 15085) (T_OBJECT_CAST 15086) (T_PRINT 15087) (T_PROPERTY_C 15088) (T_READONLY 15089) (T_REQUIRE 15090) (T_REQUIRE_ONCE 15091) (T_START_HEREDOC 15092) (T_STATIC 15093) (T_STRING 15094) (T_STRING_CAST 15095) (T_THROW 15096) (T_TRAIT_C 15097) (T_UNSET_CAST 15098) (T_VARIABLE 15099) (T_YIELD 15100) (T_YIELD_FROM 15101) ("[" 15102) ("`" 15103) (array_object_dereferenceable 15104) (attribute 135) (attributes 15105) (callable_expr 15106) (callable_variable 15107) (class_constant 15108) (class_name 15109) (constant 15110) (dereferenceable_scalar 15111) (expr 16110) (fn 15113) (fully_dereferenceable 107) #'15114 (function_call 15115) (inline_function 15116) (internal_functions_in_yacc 15117) (match 15118) (name 15119) (new_dereferenceable 15120) (new_non_dereferenceable 15121) (scalar 15122) (simple_variable 15123) (static_member 15124) (variable 15125) (variable_class_name 15126) ("~" 15127)) 6991 (("%" 15129) ("*" 15130) ("+" 15131) ("-" 15132) ("." 15133) ("/" 15134) ("<" 15135) (">" 15136) ("?" 15137) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 15138) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 15139) (T_BOOLEAN_AND 15140) (T_BOOLEAN_OR 15141) (T_COALESCE 15142) (T_INSTANCEOF 15143) (T_IS_EQUAL 15144) (T_IS_GREATER_OR_EQUAL 15145) (T_IS_IDENTICAL 15146) (T_IS_NOT_EQUAL 15147) (T_IS_NOT_IDENTICAL 15148) (T_IS_SMALLER_OR_EQUAL 15149) (T_LOGICAL_AND 15150) (T_LOGICAL_OR 15151) (T_LOGICAL_XOR 15152) (T_PIPE 15153) (T_POW 15154) (T_SL 15155) (T_SPACESHIP 15156) (T_SR 15157) ("^" 15158) (backup_fn_flags 16111) ("|" 15159)) 6992 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 16113) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 6993 (("$" 15734) (T_ABSTRACT 15913) (T_ARRAY 15914) (T_AS 15915) (T_BREAK 15916) (T_CALLABLE 15917) (T_CASE 15918) (T_CATCH 15919) (T_CLASS 15920) (T_CLASS_C 15921) (T_CLONE 15922) (T_CONST 15923) (T_CONTINUE 15924) (T_DECLARE 15925) (T_DEFAULT 15926) (T_DIR 15927) (T_DO 15928) (T_ECHO 15929) (T_ELSE 15930) (T_ELSEIF 15931) (T_EMPTY 15932) (T_ENDDECLARE 15933) (T_ENDFOR 15934) (T_ENDFOREACH 15935) (T_ENDIF 15936) (T_ENDSWITCH 15937) (T_ENDWHILE 15938) (T_ENUM 15939) (T_EVAL 15940) (T_EXIT 15941) (T_EXTENDS 15942) (T_FILE 15943) (T_FINAL 15944) (T_FINALLY 15945) (T_FN 15946) (T_FOR 15947) (T_FOREACH 15948) (T_FUNCTION 15949) (T_FUNC_C 15950) (T_GLOBAL 15951) (T_GOTO 15952) (T_IF 15953) (T_IMPLEMENTS 15954) (T_INCLUDE 15955) (T_INCLUDE_ONCE 15956) (T_INSTANCEOF 15957) (T_INSTEADOF 15958) (T_INTERFACE 15959) (T_ISSET 15960) (T_LINE 15961) (T_LIST 15962) (T_LOGICAL_AND 15963) (T_LOGICAL_OR 15964) (T_LOGICAL_XOR 15965) (T_MATCH 15966) (T_METHOD_C 15967) (T_NAMESPACE 15968) (T_NEW 15969) (T_NS_C 15970) (T_PRINT 15971) (T_PRIVATE 15972) (T_PROPERTY_C 15973) (T_PROTECTED 15974) (T_PUBLIC 15975) (T_READONLY 15976) (T_REQUIRE 15977) (T_REQUIRE_ONCE 15978) (T_RETURN 15979) (T_STATIC 15980) (T_STRING 15981) (T_SWITCH 15982) (T_THROW 15983) (T_TRAIT 15984) (T_TRAIT_C 15985) (T_TRY 15986) (T_UNSET 15987) (T_USE 15988) (T_VAR 15989) (T_VARIABLE 15782) (T_WHILE 15990) (T_YIELD 15991) (identifier 16115) (member_name 16116) (reserved_non_modifiers 15994) (semi_reserved 15995) (simple_variable 16117) ("{" 16118)) 6994 (("(" 16000) (argument_list 16121)) 6995 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 16119) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 6996 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 16120)) 6997 ((T_FN 35) (T_FUNCTION 131) (fn 15796) #'15797 (inline_function 16125)) 6998 (("$" 15734) (T_STRING 16131) (T_VARIABLE 15782) (property_name 16138) (simple_variable 16133) ("{" 16134)) 6999 (("$" 15734) (T_STRING 16131) (T_VARIABLE 15782) (property_name 16132) (simple_variable 16133) ("{" 16134)) 7000 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 16129) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 7001 (("]" 16130)) 7002 (("(" 16000) (argument_list 16137)) 7003 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 16135) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 7004 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 16136)) 7005 (("(" 16000) (argument_list 16139)) 7006 (("`" 16141)) 7007 (("]" 16143)) 7008 (("=" 16144)) 7009 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 16145) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 7010 (("%" 15812) ("*" 15813) ("+" 15814) ("-" 15815) ("." 15816) ("/" 15817) ("<" 15818) (">" 15819) ("?" 15820) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 15821) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 15822) (T_BOOLEAN_AND 15823) (T_BOOLEAN_OR 15824) (T_COALESCE 15825) (T_DOUBLE_ARROW 16148) (T_INSTANCEOF 15826) (T_IS_EQUAL 15827) (T_IS_GREATER_OR_EQUAL 15828) (T_IS_IDENTICAL 15829) (T_IS_NOT_EQUAL 15830) (T_IS_NOT_IDENTICAL 15831) (T_IS_SMALLER_OR_EQUAL 15832) (T_LOGICAL_AND 15833) (T_LOGICAL_OR 15834) (T_LOGICAL_XOR 15835) (T_PIPE 15836) (T_POW 15837) (T_SL 15838) (T_SPACESHIP 15839) (T_SR 15840) ("^" 15841) ("|" 15842)) 7011 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 16149) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 7012 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 16158) (T_VARIABLE 2920) (encaps_var 2934)) 7013 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 16157) (T_VARIABLE 2920) (encaps_var 2925)) 7014 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16199) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7015 ((T_CLASS 16191)) 7016 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 16190) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 16171) (attribute 2982) (class_modifier 2963)) 7017 ((T_PAAMAYIM_NEKUDOTAYIM 16188)) 7018 (("(" 16055) (argument_list 16187)) 7019 ((T_NULLSAFE_OBJECT_OPERATOR 16178) (T_OBJECT_OPERATOR 16179) (T_PAAMAYIM_NEKUDOTAYIM 16180) ("[" 16181)) 7020 (("$" 16039) (T_STRING 16074) (T_VARIABLE 16041) (property_name 16186) (simple_variable 16076) ("{" 16077)) 7021 (("$" 16039) (T_STRING 16074) (T_VARIABLE 16041) (property_name 16185) (simple_variable 16076) ("{" 16077)) 7022 (("$" 16039) (T_VARIABLE 16041) (simple_variable 16184)) 7023 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 16182) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 7024 (("]" 16183)) 7025 (("$" 16039) (T_VARIABLE 16041) (simple_variable 16189)) 7026 (("(" 2984) (argument_list 2985) (ctor_arguments 16192)) 7027 ((T_EXTENDS 2987) (extends_from 16193)) 7028 ((T_IMPLEMENTS 2989) (implements_list 16194)) 7029 ((backup_doc_comment 16195)) 7030 (("{" 16196)) 7031 ((class_statement_list 16197)) 7032 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 16198)) 7033 (("%" 304) (")" 16200) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 7034 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16202) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7035 (("%" 304) (")" 16203) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 7036 (("{" 16204)) 7037 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 16205) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 7038 (("}" 16206)) 7039 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 16208) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 7040 ((")" 16209)) 7041 (("=" 16210)) 7042 (("!" 15732) ("\"" 15733) ("$" 15734) ("(" 15735) ("+" 15736) ("-" 15737) ("@" 15738) (T_ARRAY 15739) (T_ARRAY_CAST 15740) (T_ATTRIBUTE 130) (T_BOOL_CAST 15741) (T_CLASS_C 15742) (T_CLONE 15743) (T_CONSTANT_ENCAPSED_STRING 15744) (T_DEC 15745) (T_DIR 15746) (T_DNUMBER 15747) (T_DOUBLE_CAST 15748) (T_EMPTY 15749) (T_EVAL 15750) (T_EXIT 15751) (T_FILE 15752) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 15753) (T_INC 15754) (T_INCLUDE 15755) (T_INCLUDE_ONCE 15756) (T_INT_CAST 15757) (T_ISSET 15758) (T_LINE 15759) (T_LIST 15760) (T_LNUMBER 15761) (T_MATCH 15762) (T_METHOD_C 15763) (T_NAME_FULLY_QUALIFIED 15764) (T_NAME_QUALIFIED 15765) (T_NAME_RELATIVE 15766) (T_NEW 15767) (T_NS_C 15768) (T_OBJECT_CAST 15769) (T_PRINT 15770) (T_PROPERTY_C 15771) (T_READONLY 15772) (T_REQUIRE 15773) (T_REQUIRE_ONCE 15774) (T_START_HEREDOC 15775) (T_STATIC 15776) (T_STRING 15777) (T_STRING_CAST 15778) (T_THROW 15779) (T_TRAIT_C 15780) (T_UNSET_CAST 15781) (T_VARIABLE 15782) (T_YIELD 15783) (T_YIELD_FROM 15784) ("[" 15785) ("`" 15786) (array_object_dereferenceable 15787) (attribute 135) (attributes 15788) (callable_expr 15789) (callable_variable 15790) (class_constant 15791) (class_name 15792) (constant 15793) (dereferenceable_scalar 15794) (expr 16211) (fn 15796) (fully_dereferenceable 107) #'15797 (function_call 15798) (inline_function 15799) (internal_functions_in_yacc 15800) (match 15801) (name 15802) (new_dereferenceable 15803) (new_non_dereferenceable 15804) (scalar 15805) (simple_variable 15806) (static_member 15807) (variable 15808) (variable_class_name 15809) ("~" 15810)) 7043 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 16213) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 7044 (("," 5226) (possible_comma 16214)) 7045 ((")" 16215)) 7046 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 16223) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 16224) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 16225) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 7047 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 16228) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 7048 (("," 1498) (possible_comma 16226)) 7049 ((")" 16227)) 7050 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16230) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7051 (("%" 304) (")" 16231) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 7052 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16233) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7053 (("%" 304) (")" 16234) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 7054 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 16240) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 16241) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 16242) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 16243) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 7055 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 16249) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 7056 (("%" 2442) (")" 16246) ("*" 2443) ("+" 2444) ("," 16247) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 7057 (("," 5287) (possible_comma 16244)) 7058 ((")" 16245)) 7059 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 16248) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 7060 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 16253) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 7061 ((")" 16254)) 7062 (("%" 304) (")" 16246) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 7063 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 16261) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 7064 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 16262)) 7065 (("\"" 16264) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 7066 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 16274) (T_VARIABLE 2920) (encaps_var 2934)) 7067 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 16273) (T_VARIABLE 2920) (encaps_var 2925)) 7068 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16315) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7069 ((T_CLASS 16307)) 7070 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 16306) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 16287) (attribute 2982) (class_modifier 2963)) 7071 ((T_PAAMAYIM_NEKUDOTAYIM 16304)) 7072 (("(" 14579) (argument_list 16303)) 7073 ((T_NULLSAFE_OBJECT_OPERATOR 16294) (T_OBJECT_OPERATOR 16295) (T_PAAMAYIM_NEKUDOTAYIM 16296) ("[" 16297)) 7074 (("$" 14563) (T_STRING 14598) (T_VARIABLE 14565) (property_name 16302) (simple_variable 14600) ("{" 14601)) 7075 (("$" 14563) (T_STRING 14598) (T_VARIABLE 14565) (property_name 16301) (simple_variable 14600) ("{" 14601)) 7076 (("$" 14563) (T_VARIABLE 14565) (simple_variable 16300)) 7077 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 16298) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 7078 (("]" 16299)) 7079 (("$" 14563) (T_VARIABLE 14565) (simple_variable 16305)) 7080 (("(" 2984) (argument_list 2985) (ctor_arguments 16308)) 7081 ((T_EXTENDS 2987) (extends_from 16309)) 7082 ((T_IMPLEMENTS 2989) (implements_list 16310)) 7083 ((backup_doc_comment 16311)) 7084 (("{" 16312)) 7085 ((class_statement_list 16313)) 7086 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 16314)) 7087 (("%" 304) (")" 16316) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 7088 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16318) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7089 (("%" 304) (")" 16319) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 7090 (("{" 16320)) 7091 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 16321) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 7092 (("}" 16322)) 7093 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 16324) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 7094 ((")" 16325)) 7095 (("=" 16326)) 7096 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 16327) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 7097 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 16329) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 7098 (("," 5226) (possible_comma 16330)) 7099 ((")" 16331)) 7100 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 16339) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 16340) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 16341) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 7101 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 16344) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 7102 (("," 1498) (possible_comma 16342)) 7103 ((")" 16343)) 7104 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16346) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7105 (("%" 304) (")" 16347) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 7106 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16349) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7107 (("%" 304) (")" 16350) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 7108 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 16356) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 16357) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 16358) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 16359) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 7109 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 16365) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 7110 (("%" 2442) (")" 16362) ("*" 2443) ("+" 2444) ("," 16363) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 7111 (("," 5287) (possible_comma 16360)) 7112 ((")" 16361)) 7113 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 16364) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 7114 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 16369) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 7115 ((")" 16370)) 7116 (("%" 304) (")" 16362) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 7117 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 16377) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 7118 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 16378)) 7119 (("\"" 16380) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 7120 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 16387) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 7121 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 16388)) 7122 (("$" 351) (T_VARIABLE 358) (simple_variable 16391)) 7123 (("%" 304) (")" 16393) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 7124 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 16396) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 7125 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 16397)) 7126 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16407) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7127 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 16405) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 7128 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16406) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7129 (("$" 226) (T_ABSTRACT 16417) (T_ARRAY 16418) (T_AS 16419) (T_BREAK 16420) (T_CALLABLE 16421) (T_CASE 16422) (T_CATCH 16423) (T_CLASS 16424) (T_CLASS_C 16425) (T_CLONE 16426) (T_CONST 16427) (T_CONTINUE 16428) (T_DECLARE 16429) (T_DEFAULT 16430) (T_DIR 16431) (T_DO 16432) (T_ECHO 16433) (T_ELSE 16434) (T_ELSEIF 16435) (T_EMPTY 16436) (T_ENDDECLARE 16437) (T_ENDFOR 16438) (T_ENDFOREACH 16439) (T_ENDIF 16440) (T_ENDSWITCH 16441) (T_ENDWHILE 16442) (T_ENUM 16443) (T_EVAL 16444) (T_EXIT 16445) (T_EXTENDS 16446) (T_FILE 16447) (T_FINAL 16448) (T_FINALLY 16449) (T_FN 16450) (T_FOR 16451) (T_FOREACH 16452) (T_FUNCTION 16453) (T_FUNC_C 16454) (T_GLOBAL 16455) (T_GOTO 16456) (T_IF 16457) (T_IMPLEMENTS 16458) (T_INCLUDE 16459) (T_INCLUDE_ONCE 16460) (T_INSTANCEOF 16461) (T_INSTEADOF 16462) (T_INTERFACE 16463) (T_ISSET 16464) (T_LINE 16465) (T_LIST 16466) (T_LOGICAL_AND 16467) (T_LOGICAL_OR 16468) (T_LOGICAL_XOR 16469) (T_MATCH 16470) (T_METHOD_C 16471) (T_NAMESPACE 16472) (T_NEW 16473) (T_NS_C 16474) (T_PRINT 16475) (T_PRIVATE 16476) (T_PROPERTY_C 16477) (T_PROTECTED 16478) (T_PUBLIC 16479) (T_READONLY 16480) (T_REQUIRE 16481) (T_REQUIRE_ONCE 16482) (T_RETURN 16483) (T_STATIC 16484) (T_STRING 16485) (T_SWITCH 16486) (T_THROW 16487) (T_TRAIT 16488) (T_TRAIT_C 16489) (T_TRY 16490) (T_UNSET 16491) (T_USE 16492) (T_VAR 16493) (T_VARIABLE 274) (T_WHILE 16494) (T_YIELD 16495) (identifier 16496) (member_name 16497) (reserved_non_modifiers 16498) (semi_reserved 16499) (simple_variable 16500) ("{" 16501)) 7130 (("(" 16504) (argument_list 16505)) 7131 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 16502) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 7132 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 16503)) 7133 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 16506) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 16507) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 16508) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 7134 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 16511) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 7135 (("," 1498) (possible_comma 16509)) 7136 ((")" 16510)) 7137 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (ampersand 16541) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16542) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7138 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16540) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7139 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16539) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7140 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16538) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7141 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16537) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7142 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16536) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7143 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16535) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7144 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16534) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7145 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16533) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7146 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16532) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7147 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16531) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7148 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16530) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7149 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16529) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7150 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16528) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7151 (("\"" 1538) ("$" 16543) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 16544) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 16545) ("[" 1560) (array_object_dereferenceable 16546) (callable_expr 16547) (callable_variable 16548) (class_constant 1564) (class_name 16549) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 16550) (name 16551) (new_dereferenceable 1570) (simple_variable 16552) (static_member 16553) (variable 16554) (variable_class_name 16555)) 7152 (("$" 16543) (T_VARIABLE 16545) (simple_variable 16588) ("{" 16589)) 7153 (("(" 16559) (argument_list 16587)) 7154 ((T_NULLSAFE_OBJECT_OPERATOR 16573) (T_OBJECT_OPERATOR 16574) ("[" 16575)) 7155 (("(" 16559) (argument_list 16572)) 7156 ((T_PAAMAYIM_NEKUDOTAYIM 16568)) 7157 (("(" 16559) (argument_list 16567)) 7158 ((T_PAAMAYIM_NEKUDOTAYIM 16556)) 7159 (("$" 16543) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 16545) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 16557) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 16558) ("{" 1660)) 7160 (("(" 16559) (argument_list 16560)) 7161 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 16561) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 16562) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 16563) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 7162 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 16566) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 7163 (("," 1498) (possible_comma 16564)) 7164 ((")" 16565)) 7165 (("$" 16543) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 16545) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 16569) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 16570) ("{" 1676)) 7166 (("(" 16559) (argument_list 16571)) 7167 (("$" 16543) (T_STRING 16578) (T_VARIABLE 16545) (property_name 16585) (simple_variable 16580) ("{" 16581)) 7168 (("$" 16543) (T_STRING 16578) (T_VARIABLE 16545) (property_name 16579) (simple_variable 16580) ("{" 16581)) 7169 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 16576) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 7170 (("]" 16577)) 7171 (("(" 16559) (argument_list 16584)) 7172 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 16582) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 7173 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 16583)) 7174 (("(" 16559) (argument_list 16586)) 7175 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 16590) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 7176 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 16591)) 7177 ((backup_doc_comment 16594)) 7178 (("(" 16595)) 7179 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 16596)) 7180 ((")" 16597)) 7181 ((T_USE 166) (lexical_vars 16598)) 7182 ((":" 168) (return_type 16599)) 7183 ((backup_fn_flags 16600)) 7184 (("{" 16601)) 7185 ((inner_statement_list 16602)) 7186 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 16603) ("~" 129)) 7187 ((backup_fn_flags 16604)) 7188 ((backup_doc_comment 16606)) 7189 (("(" 16607)) 7190 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 16608)) 7191 ((")" 16609)) 7192 ((":" 2271) (return_type 16610)) 7193 ((T_DOUBLE_ARROW 16611)) 7194 ((backup_fn_flags 16612)) 7195 ((backup_lex_pos 16613)) 7196 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16677) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7197 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 17677) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7198 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 17675) (encaps_var 5357)) 7199 (("$" 16616) (T_VARIABLE 16664) (simple_variable 17671) ("{" 17672)) 7200 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 17670) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7201 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 17669) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7202 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 17668) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7203 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 17667) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7204 (("(" 17664)) 7205 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 17663) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7206 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 17662) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7207 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 17649) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (clone_argument_list 17650) (constant 16675) (dereferenceable_scalar 16676) (expr 17651) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7208 (("\"" 1538) ("$" 16921) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 16922) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 16923) ("[" 1560) (array_object_dereferenceable 16924) (callable_expr 16925) (callable_variable 16926) (class_constant 1564) (class_name 16927) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 16928) (name 16929) (new_dereferenceable 1570) (simple_variable 16930) (static_member 16931) (variable 17648) (variable_class_name 16933)) 7209 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 17647) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7210 (("(" 17644)) 7211 (("(" 17641)) 7212 (("(" 17632) (argument_list 17633) (ctor_arguments 17634)) 7213 (("\"" 1538) ("$" 16921) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 16922) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 16923) ("[" 1560) (array_object_dereferenceable 16924) (callable_expr 16925) (callable_variable 16926) (class_constant 1564) (class_name 16927) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 16928) (name 16929) (new_dereferenceable 1570) (simple_variable 16930) (static_member 16931) (variable 17631) (variable_class_name 16933)) 7214 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 17630) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7215 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 17629) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7216 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 17628) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7217 (("(" 17624)) 7218 (("(" 17619)) 7219 (("(" 17613)) 7220 (("$" 16921) ("(" 17576) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 17577) (T_NAME_QUALIFIED 17578) (T_NAME_RELATIVE 17579) (T_READONLY 2955) (T_STATIC 17580) (T_STRING 17581) (T_VARIABLE 16923) (anonymous_class 17582) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 17583) (attribute 2961) (attributes 17584) (class_modifier 2963) (class_name 17585) (class_name_reference 17586) (name 17587) (new_variable 17588) (simple_variable 17589)) 7221 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 17575) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7222 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 17574) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7223 (("(" 16882) (argument_list 17573)) 7224 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 17572) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7225 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 17571) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7226 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 17566) (T_END_HEREDOC 17567) (T_VARIABLE 2920) (encaps_list 17568) (encaps_var 2922)) 7227 ((T_FN 35) (T_FUNCTION 131) (fn 16678) #'16679 (inline_function 17565)) 7228 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 17564) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7229 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 17563) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7230 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 17562) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7231 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17091) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7232 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 17027) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7233 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 17023) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 7234 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 17021) (encaps_list 2828) (encaps_var 2829)) 7235 ((T_NULLSAFE_OBJECT_OPERATOR 17007) (T_OBJECT_OPERATOR 17008) ("[" 17009)) 7236 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 17004) (attribute 2757) (fn 16678) #'16679 (inline_function 17005)) 7237 (("(" 16882) (argument_list 17003)) 7238 ((T_PAAMAYIM_NEKUDOTAYIM 16995)) 7239 (("%" 16694) ("*" 16695) ("+" 16696) ("-" 16697) ("." 16698) ("/" 16699) ("<" 16700) (">" 16701) ("?" 16702) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 16703) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 16704) (T_BOOLEAN_AND 16705) (T_BOOLEAN_OR 16706) (T_COALESCE 16707) (T_INSTANCEOF 16708) (T_IS_EQUAL 16709) (T_IS_GREATER_OR_EQUAL 16710) (T_IS_IDENTICAL 16711) (T_IS_NOT_EQUAL 16712) (T_IS_NOT_IDENTICAL 16713) (T_IS_SMALLER_OR_EQUAL 16714) (T_LOGICAL_AND 16715) (T_LOGICAL_OR 16716) (T_LOGICAL_XOR 16717) (T_PIPE 16718) (T_POW 16719) (T_SL 16720) (T_SPACESHIP 16721) (T_SR 16722) ("^" 16723) (backup_fn_flags 16994) ("|" 16724)) 7240 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 16983)) 7241 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 16971)) 7242 (("(" 16882) (argument_list 16970)) 7243 (("=" 16890) (T_AND_EQUAL 16891) (T_COALESCE_EQUAL 16892) (T_CONCAT_EQUAL 16893) (T_DEC 16894) (T_DIV_EQUAL 16895) (T_INC 16896) (T_MINUS_EQUAL 16897) (T_MOD_EQUAL 16898) (T_MUL_EQUAL 16899) (T_OR_EQUAL 16900) (T_PLUS_EQUAL 16901) (T_POW_EQUAL 16902) (T_SL_EQUAL 16903) (T_SR_EQUAL 16904) (T_XOR_EQUAL 16905)) 7244 ((T_PAAMAYIM_NEKUDOTAYIM 16794)) 7245 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16693) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7246 (("%" 16694) ("*" 16695) ("+" 16696) ("-" 16697) ("." 16698) ("/" 16699) ("<" 16700) (">" 16701) ("?" 16702) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 16703) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 16704) (T_BOOLEAN_AND 16705) (T_BOOLEAN_OR 16706) (T_COALESCE 16707) (T_INSTANCEOF 16708) (T_IS_EQUAL 16709) (T_IS_GREATER_OR_EQUAL 16710) (T_IS_IDENTICAL 16711) (T_IS_NOT_EQUAL 16712) (T_IS_NOT_IDENTICAL 16713) (T_IS_SMALLER_OR_EQUAL 16714) (T_LOGICAL_AND 16715) (T_LOGICAL_OR 16716) (T_LOGICAL_XOR 16717) (T_PIPE 16718) (T_POW 16719) (T_SL 16720) (T_SPACESHIP 16721) (T_SR 16722) ("^" 16723) ("|" 16724)) 7247 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16793) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7248 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16792) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7249 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16791) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7250 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16790) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7251 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16789) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7252 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16788) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7253 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16787) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7254 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16786) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7255 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 16781) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 16782) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 7256 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16780) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7257 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16779) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7258 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16778) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7259 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16777) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7260 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16776) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7261 (("$" 16741) ("(" 16742) (T_NAME_FULLY_QUALIFIED 16743) (T_NAME_QUALIFIED 16744) (T_NAME_RELATIVE 16745) (T_STATIC 16746) (T_STRING 16747) (T_VARIABLE 16748) (class_name 16749) (class_name_reference 16750) (name 16751) (new_variable 16752) (simple_variable 16753)) 7262 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16740) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7263 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16739) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7264 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16738) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7265 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16737) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7266 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16736) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7267 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16735) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7268 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16734) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7269 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16733) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7270 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16732) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7271 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16731) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7272 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16730) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7273 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16729) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7274 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16728) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7275 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16727) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7276 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16726) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7277 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16725) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7278 (("$" 16741) (T_VARIABLE 16748) (simple_variable 16772) ("{" 16773)) 7279 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 16770) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7280 ((T_PAAMAYIM_NEKUDOTAYIM 16768)) 7281 ((T_NULLSAFE_OBJECT_OPERATOR 16754) (T_OBJECT_OPERATOR 16755) (T_PAAMAYIM_NEKUDOTAYIM 16756) ("[" 16757)) 7282 (("$" 16741) (T_STRING 16761) (T_VARIABLE 16748) (property_name 16767) (simple_variable 16763) ("{" 16764)) 7283 (("$" 16741) (T_STRING 16761) (T_VARIABLE 16748) (property_name 16762) (simple_variable 16763) ("{" 16764)) 7284 (("$" 16741) (T_VARIABLE 16748) (simple_variable 16760)) 7285 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 16758) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 7286 (("]" 16759)) 7287 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 16765) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 7288 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 16766)) 7289 (("$" 16741) (T_VARIABLE 16748) (simple_variable 16769)) 7290 (("%" 304) (")" 16771) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 7291 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 16774) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 7292 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 16775)) 7293 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16785) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7294 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 16783) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 7295 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16784) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7296 (("$" 16616) (T_ABSTRACT 16795) (T_ARRAY 16796) (T_AS 16797) (T_BREAK 16798) (T_CALLABLE 16799) (T_CASE 16800) (T_CATCH 16801) (T_CLASS 16802) (T_CLASS_C 16803) (T_CLONE 16804) (T_CONST 16805) (T_CONTINUE 16806) (T_DECLARE 16807) (T_DEFAULT 16808) (T_DIR 16809) (T_DO 16810) (T_ECHO 16811) (T_ELSE 16812) (T_ELSEIF 16813) (T_EMPTY 16814) (T_ENDDECLARE 16815) (T_ENDFOR 16816) (T_ENDFOREACH 16817) (T_ENDIF 16818) (T_ENDSWITCH 16819) (T_ENDWHILE 16820) (T_ENUM 16821) (T_EVAL 16822) (T_EXIT 16823) (T_EXTENDS 16824) (T_FILE 16825) (T_FINAL 16826) (T_FINALLY 16827) (T_FN 16828) (T_FOR 16829) (T_FOREACH 16830) (T_FUNCTION 16831) (T_FUNC_C 16832) (T_GLOBAL 16833) (T_GOTO 16834) (T_IF 16835) (T_IMPLEMENTS 16836) (T_INCLUDE 16837) (T_INCLUDE_ONCE 16838) (T_INSTANCEOF 16839) (T_INSTEADOF 16840) (T_INTERFACE 16841) (T_ISSET 16842) (T_LINE 16843) (T_LIST 16844) (T_LOGICAL_AND 16845) (T_LOGICAL_OR 16846) (T_LOGICAL_XOR 16847) (T_MATCH 16848) (T_METHOD_C 16849) (T_NAMESPACE 16850) (T_NEW 16851) (T_NS_C 16852) (T_PRINT 16853) (T_PRIVATE 16854) (T_PROPERTY_C 16855) (T_PROTECTED 16856) (T_PUBLIC 16857) (T_READONLY 16858) (T_REQUIRE 16859) (T_REQUIRE_ONCE 16860) (T_RETURN 16861) (T_STATIC 16862) (T_STRING 16863) (T_SWITCH 16864) (T_THROW 16865) (T_TRAIT 16866) (T_TRAIT_C 16867) (T_TRY 16868) (T_UNSET 16869) (T_USE 16870) (T_VAR 16871) (T_VARIABLE 16664) (T_WHILE 16872) (T_YIELD 16873) (identifier 16874) (member_name 16875) (reserved_non_modifiers 16876) (semi_reserved 16877) (simple_variable 16878) ("{" 16879)) 7297 (("(" 16882) (argument_list 16883)) 7298 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 16880) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 7299 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 16881)) 7300 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 16884) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 16885) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 16886) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 7301 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 16889) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 7302 (("," 1498) (possible_comma 16887)) 7303 ((")" 16888)) 7304 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (ampersand 16919) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16920) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7305 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16918) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7306 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16917) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7307 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16916) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7308 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16915) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7309 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16914) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7310 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16913) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7311 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16912) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7312 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16911) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7313 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16910) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7314 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16909) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7315 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16908) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7316 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16907) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7317 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16906) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7318 (("\"" 1538) ("$" 16921) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 16922) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 16923) ("[" 1560) (array_object_dereferenceable 16924) (callable_expr 16925) (callable_variable 16926) (class_constant 1564) (class_name 16927) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 16928) (name 16929) (new_dereferenceable 1570) (simple_variable 16930) (static_member 16931) (variable 16932) (variable_class_name 16933)) 7319 (("$" 16921) (T_VARIABLE 16923) (simple_variable 16966) ("{" 16967)) 7320 (("(" 16937) (argument_list 16965)) 7321 ((T_NULLSAFE_OBJECT_OPERATOR 16951) (T_OBJECT_OPERATOR 16952) ("[" 16953)) 7322 (("(" 16937) (argument_list 16950)) 7323 ((T_PAAMAYIM_NEKUDOTAYIM 16946)) 7324 (("(" 16937) (argument_list 16945)) 7325 ((T_PAAMAYIM_NEKUDOTAYIM 16934)) 7326 (("$" 16921) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 16923) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 16935) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 16936) ("{" 1660)) 7327 (("(" 16937) (argument_list 16938)) 7328 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 16939) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 16940) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 16941) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 7329 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 16944) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 7330 (("," 1498) (possible_comma 16942)) 7331 ((")" 16943)) 7332 (("$" 16921) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 16923) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 16947) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 16948) ("{" 1676)) 7333 (("(" 16937) (argument_list 16949)) 7334 (("$" 16921) (T_STRING 16956) (T_VARIABLE 16923) (property_name 16963) (simple_variable 16958) ("{" 16959)) 7335 (("$" 16921) (T_STRING 16956) (T_VARIABLE 16923) (property_name 16957) (simple_variable 16958) ("{" 16959)) 7336 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 16954) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 7337 (("]" 16955)) 7338 (("(" 16937) (argument_list 16962)) 7339 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 16960) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 7340 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 16961)) 7341 (("(" 16937) (argument_list 16964)) 7342 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 16968) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 7343 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 16969)) 7344 ((backup_doc_comment 16972)) 7345 (("(" 16973)) 7346 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 16974)) 7347 ((")" 16975)) 7348 ((T_USE 166) (lexical_vars 16976)) 7349 ((":" 168) (return_type 16977)) 7350 ((backup_fn_flags 16978)) 7351 (("{" 16979)) 7352 ((inner_statement_list 16980)) 7353 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 16981) ("~" 129)) 7354 ((backup_fn_flags 16982)) 7355 ((backup_doc_comment 16984)) 7356 (("(" 16985)) 7357 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 16986)) 7358 ((")" 16987)) 7359 ((":" 2271) (return_type 16988)) 7360 ((T_DOUBLE_ARROW 16989)) 7361 ((backup_fn_flags 16990)) 7362 ((backup_lex_pos 16991)) 7363 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 16992) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7364 (("%" 16694) ("*" 16695) ("+" 16696) ("-" 16697) ("." 16698) ("/" 16699) ("<" 16700) (">" 16701) ("?" 16702) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 16703) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 16704) (T_BOOLEAN_AND 16705) (T_BOOLEAN_OR 16706) (T_COALESCE 16707) (T_INSTANCEOF 16708) (T_IS_EQUAL 16709) (T_IS_GREATER_OR_EQUAL 16710) (T_IS_IDENTICAL 16711) (T_IS_NOT_EQUAL 16712) (T_IS_NOT_IDENTICAL 16713) (T_IS_SMALLER_OR_EQUAL 16714) (T_LOGICAL_AND 16715) (T_LOGICAL_OR 16716) (T_LOGICAL_XOR 16717) (T_PIPE 16718) (T_POW 16719) (T_SL 16720) (T_SPACESHIP 16721) (T_SR 16722) ("^" 16723) (backup_fn_flags 16993) ("|" 16724)) 7365 (("$" 16616) (T_ABSTRACT 16795) (T_ARRAY 16796) (T_AS 16797) (T_BREAK 16798) (T_CALLABLE 16799) (T_CASE 16800) (T_CATCH 16801) (T_CLASS 16802) (T_CLASS_C 16803) (T_CLONE 16804) (T_CONST 16805) (T_CONTINUE 16806) (T_DECLARE 16807) (T_DEFAULT 16808) (T_DIR 16809) (T_DO 16810) (T_ECHO 16811) (T_ELSE 16812) (T_ELSEIF 16813) (T_EMPTY 16814) (T_ENDDECLARE 16815) (T_ENDFOR 16816) (T_ENDFOREACH 16817) (T_ENDIF 16818) (T_ENDSWITCH 16819) (T_ENDWHILE 16820) (T_ENUM 16821) (T_EVAL 16822) (T_EXIT 16823) (T_EXTENDS 16824) (T_FILE 16825) (T_FINAL 16826) (T_FINALLY 16827) (T_FN 16828) (T_FOR 16829) (T_FOREACH 16830) (T_FUNCTION 16831) (T_FUNC_C 16832) (T_GLOBAL 16833) (T_GOTO 16834) (T_IF 16835) (T_IMPLEMENTS 16836) (T_INCLUDE 16837) (T_INCLUDE_ONCE 16838) (T_INSTANCEOF 16839) (T_INSTEADOF 16840) (T_INTERFACE 16841) (T_ISSET 16842) (T_LINE 16843) (T_LIST 16844) (T_LOGICAL_AND 16845) (T_LOGICAL_OR 16846) (T_LOGICAL_XOR 16847) (T_MATCH 16848) (T_METHOD_C 16849) (T_NAMESPACE 16850) (T_NEW 16851) (T_NS_C 16852) (T_PRINT 16853) (T_PRIVATE 16854) (T_PROPERTY_C 16855) (T_PROTECTED 16856) (T_PUBLIC 16857) (T_READONLY 16858) (T_REQUIRE 16859) (T_REQUIRE_ONCE 16860) (T_RETURN 16861) (T_STATIC 16862) (T_STRING 16863) (T_SWITCH 16864) (T_THROW 16865) (T_TRAIT 16866) (T_TRAIT_C 16867) (T_TRY 16868) (T_UNSET 16869) (T_USE 16870) (T_VAR 16871) (T_VARIABLE 16664) (T_WHILE 16872) (T_YIELD 16873) (identifier 16996) (member_name 16997) (reserved_non_modifiers 16876) (semi_reserved 16877) (simple_variable 16998) ("{" 16999)) 7366 (("(" 16882) (argument_list 17002)) 7367 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 17000) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 7368 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 17001)) 7369 ((T_FN 35) (T_FUNCTION 131) (fn 16678) #'16679 (inline_function 17006)) 7370 (("$" 16616) (T_STRING 17012) (T_VARIABLE 16664) (property_name 17019) (simple_variable 17014) ("{" 17015)) 7371 (("$" 16616) (T_STRING 17012) (T_VARIABLE 16664) (property_name 17013) (simple_variable 17014) ("{" 17015)) 7372 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 17010) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 7373 (("]" 17011)) 7374 (("(" 16882) (argument_list 17018)) 7375 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 17016) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 7376 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 17017)) 7377 (("(" 16882) (argument_list 17020)) 7378 (("`" 17022)) 7379 (("]" 17024)) 7380 (("=" 17025)) 7381 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 17026) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7382 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17561) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7383 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 17559) (encaps_var 5357)) 7384 (("$" 17030) (T_VARIABLE 17078) (simple_variable 17555) ("{" 17556)) 7385 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 17554) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7386 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17553) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7387 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17552) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7388 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17551) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7389 (("(" 17548)) 7390 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17547) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7391 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17546) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7392 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17533) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (clone_argument_list 17534) (constant 17089) (dereferenceable_scalar 17090) (expr 17535) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7393 (("\"" 1538) ("$" 17335) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 17336) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 17337) ("[" 1560) (array_object_dereferenceable 17338) (callable_expr 17339) (callable_variable 17340) (class_constant 1564) (class_name 17341) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 17342) (name 17343) (new_dereferenceable 1570) (simple_variable 17344) (static_member 17345) (variable 17532) (variable_class_name 17347)) 7394 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17531) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7395 (("(" 17528)) 7396 (("(" 17525)) 7397 (("(" 17516) (argument_list 17517) (ctor_arguments 17518)) 7398 (("\"" 1538) ("$" 17335) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 17336) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 17337) ("[" 1560) (array_object_dereferenceable 17338) (callable_expr 17339) (callable_variable 17340) (class_constant 1564) (class_name 17341) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 17342) (name 17343) (new_dereferenceable 1570) (simple_variable 17344) (static_member 17345) (variable 17515) (variable_class_name 17347)) 7399 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17514) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7400 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17513) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7401 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17512) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7402 (("(" 17508)) 7403 (("(" 17503)) 7404 (("(" 17497)) 7405 (("$" 17335) ("(" 17460) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 17461) (T_NAME_QUALIFIED 17462) (T_NAME_RELATIVE 17463) (T_READONLY 2955) (T_STATIC 17464) (T_STRING 17465) (T_VARIABLE 17337) (anonymous_class 17466) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 17467) (attribute 2961) (attributes 17468) (class_modifier 2963) (class_name 17469) (class_name_reference 17470) (name 17471) (new_variable 17472) (simple_variable 17473)) 7406 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17459) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7407 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17458) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7408 (("(" 17296) (argument_list 17457)) 7409 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17456) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7410 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17455) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7411 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 17450) (T_END_HEREDOC 17451) (T_VARIABLE 2920) (encaps_list 17452) (encaps_var 2922)) 7412 ((T_FN 35) (T_FUNCTION 131) (fn 17092) #'17093 (inline_function 17449)) 7413 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17448) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7414 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17447) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7415 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17446) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7416 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17443) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7417 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17442) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7418 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 17438) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 7419 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 17436) (encaps_list 2828) (encaps_var 2829)) 7420 ((T_NULLSAFE_OBJECT_OPERATOR 17422) (T_OBJECT_OPERATOR 17423) ("[" 17424)) 7421 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 17419) (attribute 2757) (fn 17092) #'17093 (inline_function 17420)) 7422 (("(" 17296) (argument_list 17418)) 7423 ((T_PAAMAYIM_NEKUDOTAYIM 17410)) 7424 (("%" 17108) ("*" 17109) ("+" 17110) ("-" 17111) ("." 17112) ("/" 17113) ("<" 17114) (">" 17115) ("?" 17116) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17117) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17118) (T_BOOLEAN_AND 17119) (T_BOOLEAN_OR 17120) (T_COALESCE 17121) (T_DOUBLE_ARROW 17408) (T_INSTANCEOF 17122) (T_IS_EQUAL 17123) (T_IS_GREATER_OR_EQUAL 17124) (T_IS_IDENTICAL 17125) (T_IS_NOT_EQUAL 17126) (T_IS_NOT_IDENTICAL 17127) (T_IS_SMALLER_OR_EQUAL 17128) (T_LOGICAL_AND 17129) (T_LOGICAL_OR 17130) (T_LOGICAL_XOR 17131) (T_PIPE 17132) (T_POW 17133) (T_SL 17134) (T_SPACESHIP 17135) (T_SR 17136) ("^" 17137) ("|" 17138)) 7425 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 17397)) 7426 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 17385)) 7427 (("(" 17296) (argument_list 17384)) 7428 (("=" 17304) (T_AND_EQUAL 17305) (T_COALESCE_EQUAL 17306) (T_CONCAT_EQUAL 17307) (T_DEC 17308) (T_DIV_EQUAL 17309) (T_INC 17310) (T_MINUS_EQUAL 17311) (T_MOD_EQUAL 17312) (T_MUL_EQUAL 17313) (T_OR_EQUAL 17314) (T_PLUS_EQUAL 17315) (T_POW_EQUAL 17316) (T_SL_EQUAL 17317) (T_SR_EQUAL 17318) (T_XOR_EQUAL 17319)) 7429 ((T_PAAMAYIM_NEKUDOTAYIM 17208)) 7430 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17107) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7431 (("%" 17108) ("*" 17109) ("+" 17110) ("-" 17111) ("." 17112) ("/" 17113) ("<" 17114) (">" 17115) ("?" 17116) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17117) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17118) (T_BOOLEAN_AND 17119) (T_BOOLEAN_OR 17120) (T_COALESCE 17121) (T_INSTANCEOF 17122) (T_IS_EQUAL 17123) (T_IS_GREATER_OR_EQUAL 17124) (T_IS_IDENTICAL 17125) (T_IS_NOT_EQUAL 17126) (T_IS_NOT_IDENTICAL 17127) (T_IS_SMALLER_OR_EQUAL 17128) (T_LOGICAL_AND 17129) (T_LOGICAL_OR 17130) (T_LOGICAL_XOR 17131) (T_PIPE 17132) (T_POW 17133) (T_SL 17134) (T_SPACESHIP 17135) (T_SR 17136) ("^" 17137) ("|" 17138)) 7432 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17207) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7433 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17206) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7434 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17205) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7435 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17204) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7436 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17203) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7437 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17202) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7438 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17201) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7439 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17200) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7440 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 17195) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 17196) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 7441 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17194) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7442 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17193) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7443 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17192) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7444 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17191) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7445 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17190) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7446 (("$" 17155) ("(" 17156) (T_NAME_FULLY_QUALIFIED 17157) (T_NAME_QUALIFIED 17158) (T_NAME_RELATIVE 17159) (T_STATIC 17160) (T_STRING 17161) (T_VARIABLE 17162) (class_name 17163) (class_name_reference 17164) (name 17165) (new_variable 17166) (simple_variable 17167)) 7447 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17154) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7448 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17153) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7449 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17152) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7450 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17151) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7451 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17150) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7452 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17149) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7453 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17148) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7454 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17147) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7455 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17146) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7456 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17145) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7457 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17144) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7458 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17143) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7459 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17142) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7460 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17141) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7461 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17140) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7462 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17139) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7463 (("$" 17155) (T_VARIABLE 17162) (simple_variable 17186) ("{" 17187)) 7464 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 17184) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7465 ((T_PAAMAYIM_NEKUDOTAYIM 17182)) 7466 ((T_NULLSAFE_OBJECT_OPERATOR 17168) (T_OBJECT_OPERATOR 17169) (T_PAAMAYIM_NEKUDOTAYIM 17170) ("[" 17171)) 7467 (("$" 17155) (T_STRING 17175) (T_VARIABLE 17162) (property_name 17181) (simple_variable 17177) ("{" 17178)) 7468 (("$" 17155) (T_STRING 17175) (T_VARIABLE 17162) (property_name 17176) (simple_variable 17177) ("{" 17178)) 7469 (("$" 17155) (T_VARIABLE 17162) (simple_variable 17174)) 7470 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 17172) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 7471 (("]" 17173)) 7472 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 17179) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 7473 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 17180)) 7474 (("$" 17155) (T_VARIABLE 17162) (simple_variable 17183)) 7475 (("%" 304) (")" 17185) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 7476 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 17188) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 7477 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 17189)) 7478 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17199) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7479 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 17197) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 7480 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17198) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7481 (("$" 17030) (T_ABSTRACT 17209) (T_ARRAY 17210) (T_AS 17211) (T_BREAK 17212) (T_CALLABLE 17213) (T_CASE 17214) (T_CATCH 17215) (T_CLASS 17216) (T_CLASS_C 17217) (T_CLONE 17218) (T_CONST 17219) (T_CONTINUE 17220) (T_DECLARE 17221) (T_DEFAULT 17222) (T_DIR 17223) (T_DO 17224) (T_ECHO 17225) (T_ELSE 17226) (T_ELSEIF 17227) (T_EMPTY 17228) (T_ENDDECLARE 17229) (T_ENDFOR 17230) (T_ENDFOREACH 17231) (T_ENDIF 17232) (T_ENDSWITCH 17233) (T_ENDWHILE 17234) (T_ENUM 17235) (T_EVAL 17236) (T_EXIT 17237) (T_EXTENDS 17238) (T_FILE 17239) (T_FINAL 17240) (T_FINALLY 17241) (T_FN 17242) (T_FOR 17243) (T_FOREACH 17244) (T_FUNCTION 17245) (T_FUNC_C 17246) (T_GLOBAL 17247) (T_GOTO 17248) (T_IF 17249) (T_IMPLEMENTS 17250) (T_INCLUDE 17251) (T_INCLUDE_ONCE 17252) (T_INSTANCEOF 17253) (T_INSTEADOF 17254) (T_INTERFACE 17255) (T_ISSET 17256) (T_LINE 17257) (T_LIST 17258) (T_LOGICAL_AND 17259) (T_LOGICAL_OR 17260) (T_LOGICAL_XOR 17261) (T_MATCH 17262) (T_METHOD_C 17263) (T_NAMESPACE 17264) (T_NEW 17265) (T_NS_C 17266) (T_PRINT 17267) (T_PRIVATE 17268) (T_PROPERTY_C 17269) (T_PROTECTED 17270) (T_PUBLIC 17271) (T_READONLY 17272) (T_REQUIRE 17273) (T_REQUIRE_ONCE 17274) (T_RETURN 17275) (T_STATIC 17276) (T_STRING 17277) (T_SWITCH 17278) (T_THROW 17279) (T_TRAIT 17280) (T_TRAIT_C 17281) (T_TRY 17282) (T_UNSET 17283) (T_USE 17284) (T_VAR 17285) (T_VARIABLE 17078) (T_WHILE 17286) (T_YIELD 17287) (identifier 17288) (member_name 17289) (reserved_non_modifiers 17290) (semi_reserved 17291) (simple_variable 17292) ("{" 17293)) 7482 (("(" 17296) (argument_list 17297)) 7483 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 17294) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 7484 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 17295)) 7485 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 17298) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 17299) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 17300) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 7486 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 17303) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 7487 (("," 1498) (possible_comma 17301)) 7488 ((")" 17302)) 7489 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (ampersand 17333) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17334) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7490 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17332) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7491 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17331) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7492 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17330) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7493 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17329) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7494 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17328) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7495 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17327) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7496 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17326) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7497 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17325) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7498 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17324) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7499 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17323) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7500 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17322) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7501 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17321) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7502 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17320) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7503 (("\"" 1538) ("$" 17335) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 17336) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 17337) ("[" 1560) (array_object_dereferenceable 17338) (callable_expr 17339) (callable_variable 17340) (class_constant 1564) (class_name 17341) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 17342) (name 17343) (new_dereferenceable 1570) (simple_variable 17344) (static_member 17345) (variable 17346) (variable_class_name 17347)) 7504 (("$" 17335) (T_VARIABLE 17337) (simple_variable 17380) ("{" 17381)) 7505 (("(" 17351) (argument_list 17379)) 7506 ((T_NULLSAFE_OBJECT_OPERATOR 17365) (T_OBJECT_OPERATOR 17366) ("[" 17367)) 7507 (("(" 17351) (argument_list 17364)) 7508 ((T_PAAMAYIM_NEKUDOTAYIM 17360)) 7509 (("(" 17351) (argument_list 17359)) 7510 ((T_PAAMAYIM_NEKUDOTAYIM 17348)) 7511 (("$" 17335) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 17337) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 17349) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 17350) ("{" 1660)) 7512 (("(" 17351) (argument_list 17352)) 7513 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 17353) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 17354) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 17355) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 7514 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 17358) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 7515 (("," 1498) (possible_comma 17356)) 7516 ((")" 17357)) 7517 (("$" 17335) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 17337) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 17361) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 17362) ("{" 1676)) 7518 (("(" 17351) (argument_list 17363)) 7519 (("$" 17335) (T_STRING 17370) (T_VARIABLE 17337) (property_name 17377) (simple_variable 17372) ("{" 17373)) 7520 (("$" 17335) (T_STRING 17370) (T_VARIABLE 17337) (property_name 17371) (simple_variable 17372) ("{" 17373)) 7521 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 17368) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 7522 (("]" 17369)) 7523 (("(" 17351) (argument_list 17376)) 7524 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 17374) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 7525 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 17375)) 7526 (("(" 17351) (argument_list 17378)) 7527 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 17382) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 7528 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 17383)) 7529 ((backup_doc_comment 17386)) 7530 (("(" 17387)) 7531 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 17388)) 7532 ((")" 17389)) 7533 ((T_USE 166) (lexical_vars 17390)) 7534 ((":" 168) (return_type 17391)) 7535 ((backup_fn_flags 17392)) 7536 (("{" 17393)) 7537 ((inner_statement_list 17394)) 7538 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 17395) ("~" 129)) 7539 ((backup_fn_flags 17396)) 7540 ((backup_doc_comment 17398)) 7541 (("(" 17399)) 7542 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 17400)) 7543 ((")" 17401)) 7544 ((":" 2271) (return_type 17402)) 7545 ((T_DOUBLE_ARROW 17403)) 7546 ((backup_fn_flags 17404)) 7547 ((backup_lex_pos 17405)) 7548 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17406) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7549 (("%" 17108) ("*" 17109) ("+" 17110) ("-" 17111) ("." 17112) ("/" 17113) ("<" 17114) (">" 17115) ("?" 17116) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17117) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17118) (T_BOOLEAN_AND 17119) (T_BOOLEAN_OR 17120) (T_COALESCE 17121) (T_INSTANCEOF 17122) (T_IS_EQUAL 17123) (T_IS_GREATER_OR_EQUAL 17124) (T_IS_IDENTICAL 17125) (T_IS_NOT_EQUAL 17126) (T_IS_NOT_IDENTICAL 17127) (T_IS_SMALLER_OR_EQUAL 17128) (T_LOGICAL_AND 17129) (T_LOGICAL_OR 17130) (T_LOGICAL_XOR 17131) (T_PIPE 17132) (T_POW 17133) (T_SL 17134) (T_SPACESHIP 17135) (T_SR 17136) ("^" 17137) (backup_fn_flags 17407) ("|" 17138)) 7550 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 17409) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7551 (("$" 17030) (T_ABSTRACT 17209) (T_ARRAY 17210) (T_AS 17211) (T_BREAK 17212) (T_CALLABLE 17213) (T_CASE 17214) (T_CATCH 17215) (T_CLASS 17216) (T_CLASS_C 17217) (T_CLONE 17218) (T_CONST 17219) (T_CONTINUE 17220) (T_DECLARE 17221) (T_DEFAULT 17222) (T_DIR 17223) (T_DO 17224) (T_ECHO 17225) (T_ELSE 17226) (T_ELSEIF 17227) (T_EMPTY 17228) (T_ENDDECLARE 17229) (T_ENDFOR 17230) (T_ENDFOREACH 17231) (T_ENDIF 17232) (T_ENDSWITCH 17233) (T_ENDWHILE 17234) (T_ENUM 17235) (T_EVAL 17236) (T_EXIT 17237) (T_EXTENDS 17238) (T_FILE 17239) (T_FINAL 17240) (T_FINALLY 17241) (T_FN 17242) (T_FOR 17243) (T_FOREACH 17244) (T_FUNCTION 17245) (T_FUNC_C 17246) (T_GLOBAL 17247) (T_GOTO 17248) (T_IF 17249) (T_IMPLEMENTS 17250) (T_INCLUDE 17251) (T_INCLUDE_ONCE 17252) (T_INSTANCEOF 17253) (T_INSTEADOF 17254) (T_INTERFACE 17255) (T_ISSET 17256) (T_LINE 17257) (T_LIST 17258) (T_LOGICAL_AND 17259) (T_LOGICAL_OR 17260) (T_LOGICAL_XOR 17261) (T_MATCH 17262) (T_METHOD_C 17263) (T_NAMESPACE 17264) (T_NEW 17265) (T_NS_C 17266) (T_PRINT 17267) (T_PRIVATE 17268) (T_PROPERTY_C 17269) (T_PROTECTED 17270) (T_PUBLIC 17271) (T_READONLY 17272) (T_REQUIRE 17273) (T_REQUIRE_ONCE 17274) (T_RETURN 17275) (T_STATIC 17276) (T_STRING 17277) (T_SWITCH 17278) (T_THROW 17279) (T_TRAIT 17280) (T_TRAIT_C 17281) (T_TRY 17282) (T_UNSET 17283) (T_USE 17284) (T_VAR 17285) (T_VARIABLE 17078) (T_WHILE 17286) (T_YIELD 17287) (identifier 17411) (member_name 17412) (reserved_non_modifiers 17290) (semi_reserved 17291) (simple_variable 17413) ("{" 17414)) 7552 (("(" 17296) (argument_list 17417)) 7553 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 17415) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 7554 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 17416)) 7555 ((T_FN 35) (T_FUNCTION 131) (fn 17092) #'17093 (inline_function 17421)) 7556 (("$" 17030) (T_STRING 17427) (T_VARIABLE 17078) (property_name 17434) (simple_variable 17429) ("{" 17430)) 7557 (("$" 17030) (T_STRING 17427) (T_VARIABLE 17078) (property_name 17428) (simple_variable 17429) ("{" 17430)) 7558 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 17425) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 7559 (("]" 17426)) 7560 (("(" 17296) (argument_list 17433)) 7561 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 17431) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 7562 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 17432)) 7563 (("(" 17296) (argument_list 17435)) 7564 (("`" 17437)) 7565 (("]" 17439)) 7566 (("=" 17440)) 7567 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17441) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7568 (("%" 17108) ("*" 17109) ("+" 17110) ("-" 17111) ("." 17112) ("/" 17113) ("<" 17114) (">" 17115) ("?" 17116) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17117) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17118) (T_BOOLEAN_AND 17119) (T_BOOLEAN_OR 17120) (T_COALESCE 17121) (T_DOUBLE_ARROW 17444) (T_INSTANCEOF 17122) (T_IS_EQUAL 17123) (T_IS_GREATER_OR_EQUAL 17124) (T_IS_IDENTICAL 17125) (T_IS_NOT_EQUAL 17126) (T_IS_NOT_IDENTICAL 17127) (T_IS_SMALLER_OR_EQUAL 17128) (T_LOGICAL_AND 17129) (T_LOGICAL_OR 17130) (T_LOGICAL_XOR 17131) (T_PIPE 17132) (T_POW 17133) (T_SL 17134) (T_SPACESHIP 17135) (T_SR 17136) ("^" 17137) ("|" 17138)) 7569 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17445) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7570 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 17454) (T_VARIABLE 2920) (encaps_var 2934)) 7571 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 17453) (T_VARIABLE 2920) (encaps_var 2925)) 7572 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 17495) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7573 ((T_CLASS 17487)) 7574 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 17486) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 17467) (attribute 2982) (class_modifier 2963)) 7575 ((T_PAAMAYIM_NEKUDOTAYIM 17484)) 7576 (("(" 17351) (argument_list 17483)) 7577 ((T_NULLSAFE_OBJECT_OPERATOR 17474) (T_OBJECT_OPERATOR 17475) (T_PAAMAYIM_NEKUDOTAYIM 17476) ("[" 17477)) 7578 (("$" 17335) (T_STRING 17370) (T_VARIABLE 17337) (property_name 17482) (simple_variable 17372) ("{" 17373)) 7579 (("$" 17335) (T_STRING 17370) (T_VARIABLE 17337) (property_name 17481) (simple_variable 17372) ("{" 17373)) 7580 (("$" 17335) (T_VARIABLE 17337) (simple_variable 17480)) 7581 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 17478) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 7582 (("]" 17479)) 7583 (("$" 17335) (T_VARIABLE 17337) (simple_variable 17485)) 7584 (("(" 2984) (argument_list 2985) (ctor_arguments 17488)) 7585 ((T_EXTENDS 2987) (extends_from 17489)) 7586 ((T_IMPLEMENTS 2989) (implements_list 17490)) 7587 ((backup_doc_comment 17491)) 7588 (("{" 17492)) 7589 ((class_statement_list 17493)) 7590 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 17494)) 7591 (("%" 304) (")" 17496) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 7592 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 17498) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7593 (("%" 304) (")" 17499) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 7594 (("{" 17500)) 7595 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 17501) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 7596 (("}" 17502)) 7597 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 17504) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 7598 ((")" 17505)) 7599 (("=" 17506)) 7600 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 17507) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 7601 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 17509) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 7602 (("," 5226) (possible_comma 17510)) 7603 ((")" 17511)) 7604 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 17519) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 17520) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 17521) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 7605 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 17524) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 7606 (("," 1498) (possible_comma 17522)) 7607 ((")" 17523)) 7608 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 17526) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7609 (("%" 304) (")" 17527) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 7610 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 17529) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7611 (("%" 304) (")" 17530) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 7612 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 17536) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 17537) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 17538) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 17539) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 7613 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 17545) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 7614 (("%" 2442) (")" 17542) ("*" 2443) ("+" 2444) ("," 17543) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 7615 (("," 5287) (possible_comma 17540)) 7616 ((")" 17541)) 7617 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 17544) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 7618 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 17549) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 7619 ((")" 17550)) 7620 (("%" 304) (")" 17542) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 7621 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 17557) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 7622 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 17558)) 7623 (("\"" 17560) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 7624 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 17570) (T_VARIABLE 2920) (encaps_var 2934)) 7625 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 17569) (T_VARIABLE 2920) (encaps_var 2925)) 7626 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 17611) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7627 ((T_CLASS 17603)) 7628 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 17602) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 17583) (attribute 2982) (class_modifier 2963)) 7629 ((T_PAAMAYIM_NEKUDOTAYIM 17600)) 7630 (("(" 16937) (argument_list 17599)) 7631 ((T_NULLSAFE_OBJECT_OPERATOR 17590) (T_OBJECT_OPERATOR 17591) (T_PAAMAYIM_NEKUDOTAYIM 17592) ("[" 17593)) 7632 (("$" 16921) (T_STRING 16956) (T_VARIABLE 16923) (property_name 17598) (simple_variable 16958) ("{" 16959)) 7633 (("$" 16921) (T_STRING 16956) (T_VARIABLE 16923) (property_name 17597) (simple_variable 16958) ("{" 16959)) 7634 (("$" 16921) (T_VARIABLE 16923) (simple_variable 17596)) 7635 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 17594) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 7636 (("]" 17595)) 7637 (("$" 16921) (T_VARIABLE 16923) (simple_variable 17601)) 7638 (("(" 2984) (argument_list 2985) (ctor_arguments 17604)) 7639 ((T_EXTENDS 2987) (extends_from 17605)) 7640 ((T_IMPLEMENTS 2989) (implements_list 17606)) 7641 ((backup_doc_comment 17607)) 7642 (("{" 17608)) 7643 ((class_statement_list 17609)) 7644 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 17610)) 7645 (("%" 304) (")" 17612) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 7646 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 17614) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7647 (("%" 304) (")" 17615) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 7648 (("{" 17616)) 7649 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 17617) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 7650 (("}" 17618)) 7651 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 17620) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 7652 ((")" 17621)) 7653 (("=" 17622)) 7654 (("!" 16614) ("\"" 16615) ("$" 16616) ("(" 16617) ("+" 16618) ("-" 16619) ("@" 16620) (T_ARRAY 16621) (T_ARRAY_CAST 16622) (T_ATTRIBUTE 130) (T_BOOL_CAST 16623) (T_CLASS_C 16624) (T_CLONE 16625) (T_CONSTANT_ENCAPSED_STRING 16626) (T_DEC 16627) (T_DIR 16628) (T_DNUMBER 16629) (T_DOUBLE_CAST 16630) (T_EMPTY 16631) (T_EVAL 16632) (T_EXIT 16633) (T_FILE 16634) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 16635) (T_INC 16636) (T_INCLUDE 16637) (T_INCLUDE_ONCE 16638) (T_INT_CAST 16639) (T_ISSET 16640) (T_LINE 16641) (T_LIST 16642) (T_LNUMBER 16643) (T_MATCH 16644) (T_METHOD_C 16645) (T_NAME_FULLY_QUALIFIED 16646) (T_NAME_QUALIFIED 16647) (T_NAME_RELATIVE 16648) (T_NEW 16649) (T_NS_C 16650) (T_OBJECT_CAST 16651) (T_PRINT 16652) (T_PROPERTY_C 16653) (T_READONLY 16654) (T_REQUIRE 16655) (T_REQUIRE_ONCE 16656) (T_START_HEREDOC 16657) (T_STATIC 16658) (T_STRING 16659) (T_STRING_CAST 16660) (T_THROW 16661) (T_TRAIT_C 16662) (T_UNSET_CAST 16663) (T_VARIABLE 16664) (T_YIELD 16665) (T_YIELD_FROM 16666) ("[" 16667) ("`" 16668) (array_object_dereferenceable 16669) (attribute 135) (attributes 16670) (callable_expr 16671) (callable_variable 16672) (class_constant 16673) (class_name 16674) (constant 16675) (dereferenceable_scalar 16676) (expr 17623) (fn 16678) (fully_dereferenceable 107) #'16679 (function_call 16680) (inline_function 16681) (internal_functions_in_yacc 16682) (match 16683) (name 16684) (new_dereferenceable 16685) (new_non_dereferenceable 16686) (scalar 16687) (simple_variable 16688) (static_member 16689) (variable 16690) (variable_class_name 16691) ("~" 16692)) 7655 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 17625) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 7656 (("," 5226) (possible_comma 17626)) 7657 ((")" 17627)) 7658 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 17635) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 17636) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 17637) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 7659 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 17640) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 7660 (("," 1498) (possible_comma 17638)) 7661 ((")" 17639)) 7662 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 17642) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7663 (("%" 304) (")" 17643) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 7664 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 17645) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7665 (("%" 304) (")" 17646) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 7666 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 17652) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 17653) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 17654) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 17655) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 7667 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 17661) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 7668 (("%" 2442) (")" 17658) ("*" 2443) ("+" 2444) ("," 17659) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 7669 (("," 5287) (possible_comma 17656)) 7670 ((")" 17657)) 7671 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 17660) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 7672 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 17665) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 7673 ((")" 17666)) 7674 (("%" 304) (")" 17658) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 7675 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 17673) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 7676 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 17674)) 7677 (("\"" 17676) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 7678 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17679) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 17680) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17681) (T_DEC 22) (T_DECLARE 17682) (T_DIR 24) (T_DNUMBER 25) (T_DO 17683) (T_DOUBLE_CAST 27) (T_ECHO 17684) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17685) (T_FOREACH 17686) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 17687) (T_GOTO 17688) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17689) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17690) (T_START_HEREDOC 69) (T_STATIC 17691) (T_STRING 17692) (T_STRING_CAST 72) (T_SWITCH 17693) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17694) (T_UNSET 17695) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17696) (T_WHILE 17697) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17698) (alt_if_stmt_without_else 17699) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 17700) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 17701) (if_stmt_without_else 17702) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17703) (static_member 123) (variable 126) (variable_class_name 127) ("{" 17704) ("~" 129)) 7679 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 27988) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 7680 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 27986) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 7681 (("(" 27977)) 7682 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22932) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22933) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22934) (T_DEC 22) (T_DECLARE 22935) (T_DIR 24) (T_DNUMBER 25) (T_DO 22936) (T_DOUBLE_CAST 27) (T_ECHO 22937) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22938) (T_FOREACH 22939) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22940) (T_GOTO 22941) (T_IF 22942) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22943) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22944) (T_START_HEREDOC 69) (T_STATIC 22945) (T_STRING 22946) (T_STRING_CAST 72) (T_SWITCH 22947) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22948) (T_UNSET 22949) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22950) (T_WHILE 22951) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22952) (alt_if_stmt_without_else 22953) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22954) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22955) (if_stmt_without_else 22956) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27971) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22958) ("~" 129)) 7683 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (echo_expr 22926) (echo_expr_list 27969) (expr 22928) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 7684 (("(" 27956)) 7685 (("(" 27941)) 7686 (("$" 19441) (T_VARIABLE 19442) (global_var 19443) (global_var_list 27939) (simple_variable 19445)) 7687 ((T_STRING 27937)) 7688 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 27935) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 7689 ((T_FN 35) (T_FUNCTION 131) (T_VARIABLE 18365) (fn 106) #'138 (inline_function 18366) (static_var 18367) (static_var_list 27933)) 7690 ((":" 27932)) 7691 (("(" 27914)) 7692 (("{" 27897)) 7693 (("(" 27892)) 7694 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 27890) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 7695 (("(" 27881)) 7696 ((T_ELSE 17710) (T_ELSEIF 17711) (T_ENDIF 17712)) 7697 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 17709) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 7698 ((T_ELSE 17707) (T_ELSEIF 222)) 7699 ((inner_statement_list 17705)) 7700 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 17706) ("~" 129)) 7701 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17679) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 17680) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17681) (T_DEC 22) (T_DECLARE 17682) (T_DIR 24) (T_DNUMBER 25) (T_DO 17683) (T_DOUBLE_CAST 27) (T_ECHO 17684) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17685) (T_FOREACH 17686) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 17687) (T_GOTO 17688) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17689) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17690) (T_START_HEREDOC 69) (T_STATIC 17691) (T_STRING 17692) (T_STRING_CAST 72) (T_SWITCH 17693) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17694) (T_UNSET 17695) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17696) (T_WHILE 17697) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17698) (alt_if_stmt_without_else 17699) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 17700) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 17701) (if_stmt_without_else 17702) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17708) (static_member 123) (variable 126) (variable_class_name 127) ("{" 17704) ("~" 129)) 7702 ((":" 27877)) 7703 (("(" 17714)) 7704 ((";" 17713)) 7705 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 17715) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7706 (("%" 304) (")" 17716) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 7707 ((":" 17717)) 7708 ((inner_statement_list 17718)) 7709 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17719) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17720) (T_CLASS 17721) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17722) (T_DEC 22) (T_DECLARE 17723) (T_DIR 24) (T_DNUMBER 25) (T_DO 17724) (T_DOUBLE_CAST 27) (T_ECHO 17725) (T_EMPTY 29) (T_ENUM 17726) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17727) (T_FOREACH 17728) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17729) (T_GOTO 17730) (T_HALT_COMPILER 17731) (T_IF 17732) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17733) (T_INTERFACE 17734) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17735) (T_START_HEREDOC 69) (T_STATIC 17736) (T_STRING 17737) (T_STRING_CAST 72) (T_SWITCH 17738) (T_THROW 74) (T_TRAIT 17739) (T_TRAIT_C 76) (T_TRY 17740) (T_UNSET 17741) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17742) (T_WHILE 17743) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17744) (alt_if_stmt_without_else 17745) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17746) (attributes 17747) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17748) (class_modifier 99) (class_modifiers 17749) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17750) (expr 17751) (fn 106) (fully_dereferenceable 107) #'17752 (function_call 109) (function_declaration_statement 17753) (if_stmt 17754) (if_stmt_without_else 17755) (inline_function 113) (inner_statement 17756) (interface_declaration_statement 17757) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17758) (static_member 123) (trait_declaration_statement 17759) (variable 126) (variable_class_name 127) ("{" 17760) ("~" 129)) 7710 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 27875) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 7711 ((T_STRING 27868)) 7712 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 27866) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 7713 (("(" 27857)) 7714 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22932) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22933) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22934) (T_DEC 22) (T_DECLARE 22935) (T_DIR 24) (T_DNUMBER 25) (T_DO 22936) (T_DOUBLE_CAST 27) (T_ECHO 22937) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22938) (T_FOREACH 22939) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22940) (T_GOTO 22941) (T_IF 22942) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22943) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22944) (T_START_HEREDOC 69) (T_STATIC 22945) (T_STRING 22946) (T_STRING_CAST 72) (T_SWITCH 22947) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22948) (T_UNSET 22949) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22950) (T_WHILE 22951) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22952) (alt_if_stmt_without_else 22953) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22954) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22955) (if_stmt_without_else 22956) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27851) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22958) ("~" 129)) 7715 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (echo_expr 22926) (echo_expr_list 27849) (expr 22928) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 7716 ((T_STRING 27842)) 7717 (("(" 27829)) 7718 (("(" 27814)) 7719 (("$" 19441) (T_VARIABLE 19442) (global_var 19443) (global_var_list 27812) (simple_variable 19445)) 7720 ((T_STRING 27810)) 7721 (("(" 27807)) 7722 (("(" 27803)) 7723 ((T_STRING 27797)) 7724 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 27795) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 7725 ((T_FN 35) (T_FUNCTION 131) (T_VARIABLE 18365) (fn 106) #'138 (inline_function 18366) (static_var 18367) (static_var_list 27793)) 7726 ((":" 27792)) 7727 (("(" 27774)) 7728 ((T_STRING 27769)) 7729 (("{" 27752)) 7730 (("(" 27747)) 7731 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 27745) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 7732 (("(" 27736)) 7733 ((T_ELSE 17802) (T_ELSEIF 17711) (T_ENDIF 17803)) 7734 ((T_ABSTRACT 10) (T_ATTRIBUTE 174) (T_CLASS 17721) (T_ENUM 17726) (T_FINAL 34) (T_FN 35) (T_FUNCTION 38) (T_INTERFACE 17734) (T_READONLY 2955) (T_STATIC 17797) (T_TRAIT 17739) (attribute 17798) (attributed_statement 17799) (class_declaration_statement 17748) (class_modifier 99) (class_modifiers 17749) (enum_declaration_statement 17750) (fn 106) #'17752 (function_declaration_statement 17753) (inline_function 17800) (interface_declaration_statement 17757) (trait_declaration_statement 17759)) 7735 ((T_ABSTRACT 10) (T_CLASS 17788) (T_FINAL 34) (T_READONLY 2955) (class_modifier 17789)) 7736 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 17787) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 7737 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17770) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17771) (ampersand 17772) (returns_ref 17773)) 7738 ((T_ELSE 17763) (T_ELSEIF 17764)) 7739 ((inner_statement_list 17761)) 7740 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 17762) ("~" 129)) 7741 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17719) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 17720) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17722) (T_DEC 22) (T_DECLARE 17723) (T_DIR 24) (T_DNUMBER 25) (T_DO 17724) (T_DOUBLE_CAST 27) (T_ECHO 17725) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17727) (T_FOREACH 17728) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 17729) (T_GOTO 17730) (T_IF 17732) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17733) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17735) (T_START_HEREDOC 69) (T_STATIC 17736) (T_STRING 17737) (T_STRING_CAST 72) (T_SWITCH 17738) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17740) (T_UNSET 17741) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17742) (T_WHILE 17743) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17744) (alt_if_stmt_without_else 17745) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 17751) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 17754) (if_stmt_without_else 17755) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17769) (static_member 123) (variable 126) (variable_class_name 127) ("{" 17760) ("~" 129)) 7742 (("(" 17765)) 7743 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 17766) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7744 (("%" 304) (")" 17767) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 7745 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17719) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 17720) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17722) (T_DEC 22) (T_DECLARE 17723) (T_DIR 24) (T_DNUMBER 25) (T_DO 17724) (T_DOUBLE_CAST 27) (T_ECHO 17725) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17727) (T_FOREACH 17728) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 17729) (T_GOTO 17730) (T_IF 17732) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17733) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17735) (T_START_HEREDOC 69) (T_STATIC 17736) (T_STRING 17737) (T_STRING_CAST 72) (T_SWITCH 17738) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17740) (T_UNSET 17741) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17742) (T_WHILE 17743) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17744) (alt_if_stmt_without_else 17745) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 17751) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 17754) (if_stmt_without_else 17755) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17768) (static_member 123) (variable 126) (variable_class_name 127) ("{" 17760) ("~" 129)) 7746 ((T_READONLY 17774) (T_STRING 17775) (backup_doc_comment 143) (function_name 17776)) 7747 ((backup_doc_comment 17777)) 7748 (("(" 17778)) 7749 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 17779)) 7750 ((")" 17780)) 7751 ((":" 168) (return_type 17781)) 7752 ((backup_fn_flags 17782)) 7753 (("{" 17783)) 7754 ((inner_statement_list 17784)) 7755 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 17785) ("~" 129)) 7756 ((backup_fn_flags 17786)) 7757 ((T_STRING 17790)) 7758 ((T_EXTENDS 2987) (extends_from 17791)) 7759 ((T_IMPLEMENTS 2989) (implements_list 17792)) 7760 ((backup_doc_comment 17793)) 7761 (("{" 17794)) 7762 ((class_statement_list 17795)) 7763 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 17796)) 7764 ((T_FN 35) (T_FUNCTION 131) (fn 106) #'138 (inline_function 17801)) 7765 ((":" 17805)) 7766 ((";" 17804)) 7767 ((inner_statement_list 17806)) 7768 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17807) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17808) (T_CLASS 17809) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17810) (T_DEC 22) (T_DECLARE 17811) (T_DIR 24) (T_DNUMBER 25) (T_DO 17812) (T_DOUBLE_CAST 27) (T_ECHO 17813) (T_EMPTY 29) (T_ENDIF 17814) (T_ENUM 17815) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17816) (T_FOREACH 17817) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17818) (T_GOTO 17819) (T_HALT_COMPILER 17820) (T_IF 17732) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17821) (T_INTERFACE 17822) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17823) (T_START_HEREDOC 69) (T_STATIC 17824) (T_STRING 17825) (T_STRING_CAST 72) (T_SWITCH 17826) (T_THROW 74) (T_TRAIT 17827) (T_TRAIT_C 76) (T_TRY 17828) (T_UNSET 17829) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17830) (T_WHILE 17831) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17832) (alt_if_stmt_without_else 17833) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17834) (attributes 17835) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17836) (class_modifier 99) (class_modifiers 17837) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17838) (expr 17839) (fn 106) (fully_dereferenceable 107) #'17840 (function_call 109) (function_declaration_statement 17841) (if_stmt 17842) (if_stmt_without_else 17843) (inline_function 113) (inner_statement 17844) (interface_declaration_statement 17845) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17846) (static_member 123) (trait_declaration_statement 17847) (variable 126) (variable_class_name 127) ("{" 17848) ("~" 129)) 7769 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 27734) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 7770 ((T_STRING 27727)) 7771 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 27725) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 7772 (("(" 27716)) 7773 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22932) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22933) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22934) (T_DEC 22) (T_DECLARE 22935) (T_DIR 24) (T_DNUMBER 25) (T_DO 22936) (T_DOUBLE_CAST 27) (T_ECHO 22937) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22938) (T_FOREACH 22939) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22940) (T_GOTO 22941) (T_IF 22942) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22943) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22944) (T_START_HEREDOC 69) (T_STATIC 22945) (T_STRING 22946) (T_STRING_CAST 72) (T_SWITCH 22947) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22948) (T_UNSET 22949) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22950) (T_WHILE 22951) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22952) (alt_if_stmt_without_else 22953) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22954) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22955) (if_stmt_without_else 22956) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27710) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22958) ("~" 129)) 7774 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (echo_expr 22926) (echo_expr_list 27708) (expr 22928) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 7775 ((";" 27707)) 7776 ((T_STRING 27700)) 7777 (("(" 27687)) 7778 (("(" 27672)) 7779 (("$" 19441) (T_VARIABLE 19442) (global_var 19443) (global_var_list 27670) (simple_variable 19445)) 7780 ((T_STRING 27668)) 7781 (("(" 27665)) 7782 ((T_STRING 27659)) 7783 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 27657) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 7784 ((T_FN 35) (T_FUNCTION 131) (T_VARIABLE 18365) (fn 106) #'138 (inline_function 18366) (static_var 18367) (static_var_list 27655)) 7785 ((":" 27654)) 7786 (("(" 27636)) 7787 ((T_STRING 27631)) 7788 (("{" 27614)) 7789 (("(" 27609)) 7790 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 27607) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 7791 (("(" 17882)) 7792 ((T_ELSE 17875) (T_ELSEIF 17711) (T_ENDIF 17876)) 7793 ((T_ABSTRACT 10) (T_ATTRIBUTE 174) (T_CLASS 17809) (T_ENUM 17815) (T_FINAL 34) (T_FN 35) (T_FUNCTION 38) (T_INTERFACE 17822) (T_READONLY 2955) (T_STATIC 17797) (T_TRAIT 17827) (attribute 17798) (attributed_statement 17874) (class_declaration_statement 17836) (class_modifier 99) (class_modifiers 17837) (enum_declaration_statement 17838) (fn 106) #'17840 (function_declaration_statement 17841) (inline_function 17800) (interface_declaration_statement 17845) (trait_declaration_statement 17847)) 7794 ((T_ABSTRACT 10) (T_CLASS 17866) (T_FINAL 34) (T_READONLY 2955) (class_modifier 17789)) 7795 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 17865) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 7796 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17770) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17771) (ampersand 17772) (returns_ref 17853)) 7797 ((T_ELSE 17851) (T_ELSEIF 17764)) 7798 ((inner_statement_list 17849)) 7799 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 17850) ("~" 129)) 7800 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17807) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 17808) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17810) (T_DEC 22) (T_DECLARE 17811) (T_DIR 24) (T_DNUMBER 25) (T_DO 17812) (T_DOUBLE_CAST 27) (T_ECHO 17813) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17816) (T_FOREACH 17817) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 17818) (T_GOTO 17819) (T_IF 17732) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17821) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17823) (T_START_HEREDOC 69) (T_STATIC 17824) (T_STRING 17825) (T_STRING_CAST 72) (T_SWITCH 17826) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17828) (T_UNSET 17829) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17830) (T_WHILE 17831) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17832) (alt_if_stmt_without_else 17833) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 17839) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 17842) (if_stmt_without_else 17843) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17852) (static_member 123) (variable 126) (variable_class_name 127) ("{" 17848) ("~" 129)) 7801 ((T_READONLY 17774) (T_STRING 17775) (backup_doc_comment 143) (function_name 17854)) 7802 ((backup_doc_comment 17855)) 7803 (("(" 17856)) 7804 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 17857)) 7805 ((")" 17858)) 7806 ((":" 168) (return_type 17859)) 7807 ((backup_fn_flags 17860)) 7808 (("{" 17861)) 7809 ((inner_statement_list 17862)) 7810 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 17863) ("~" 129)) 7811 ((backup_fn_flags 17864)) 7812 ((T_STRING 17867)) 7813 ((T_EXTENDS 2987) (extends_from 17868)) 7814 ((T_IMPLEMENTS 2989) (implements_list 17869)) 7815 ((backup_doc_comment 17870)) 7816 (("{" 17871)) 7817 ((class_statement_list 17872)) 7818 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 17873)) 7819 ((":" 17878)) 7820 ((";" 17877)) 7821 ((inner_statement_list 17879)) 7822 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17807) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17808) (T_CLASS 17809) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17810) (T_DEC 22) (T_DECLARE 17811) (T_DIR 24) (T_DNUMBER 25) (T_DO 17812) (T_DOUBLE_CAST 27) (T_ECHO 17813) (T_EMPTY 29) (T_ENDIF 17880) (T_ENUM 17815) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17816) (T_FOREACH 17817) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17818) (T_GOTO 17819) (T_HALT_COMPILER 17820) (T_IF 17732) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17821) (T_INTERFACE 17822) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17823) (T_START_HEREDOC 69) (T_STATIC 17824) (T_STRING 17825) (T_STRING_CAST 72) (T_SWITCH 17826) (T_THROW 74) (T_TRAIT 17827) (T_TRAIT_C 76) (T_TRY 17828) (T_UNSET 17829) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17830) (T_WHILE 17831) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17832) (alt_if_stmt_without_else 17833) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17834) (attributes 17835) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17836) (class_modifier 99) (class_modifiers 17837) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17838) (expr 17839) (fn 106) (fully_dereferenceable 107) #'17840 (function_call 109) (function_declaration_statement 17841) (if_stmt 17842) (if_stmt_without_else 17843) (inline_function 113) (inner_statement 17844) (interface_declaration_statement 17845) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17846) (static_member 123) (trait_declaration_statement 17847) (variable 126) (variable_class_name 127) ("{" 17848) ("~" 129)) 7823 ((";" 17881)) 7824 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 17883) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7825 (("%" 304) (")" 17884) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 7826 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 17885) (";" 17807) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 17808) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17810) (T_DEC 22) (T_DECLARE 17811) (T_DIR 24) (T_DNUMBER 25) (T_DO 17812) (T_DOUBLE_CAST 27) (T_ECHO 17813) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17816) (T_FOREACH 17817) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 17818) (T_GOTO 17819) (T_IF 17732) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17821) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17823) (T_START_HEREDOC 69) (T_STATIC 17824) (T_STRING 17825) (T_STRING_CAST 72) (T_SWITCH 17826) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17828) (T_UNSET 17829) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17830) (T_WHILE 17831) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17832) (alt_if_stmt_without_else 17833) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 17839) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 17842) (if_stmt_without_else 17843) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17886) (static_member 123) (variable 126) (variable_class_name 127) (while_statement 17887) ("{" 17848) ("~" 129)) 7827 ((inner_statement_list 17888)) 7828 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17889) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17890) (T_CLASS 17891) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17892) (T_DEC 22) (T_DECLARE 17893) (T_DIR 24) (T_DNUMBER 25) (T_DO 17894) (T_DOUBLE_CAST 27) (T_ECHO 17895) (T_EMPTY 29) (T_ENDWHILE 17896) (T_ENUM 17897) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17898) (T_FOREACH 17899) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17900) (T_GOTO 17901) (T_HALT_COMPILER 17902) (T_IF 17903) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17904) (T_INTERFACE 17905) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17906) (T_START_HEREDOC 69) (T_STATIC 17907) (T_STRING 17908) (T_STRING_CAST 72) (T_SWITCH 17909) (T_THROW 74) (T_TRAIT 17910) (T_TRAIT_C 76) (T_TRY 17911) (T_UNSET 17912) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17913) (T_WHILE 17914) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17915) (alt_if_stmt_without_else 17916) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17917) (attributes 17918) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17919) (class_modifier 99) (class_modifiers 17920) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17921) (expr 17922) (fn 106) (fully_dereferenceable 107) #'17923 (function_call 109) (function_declaration_statement 17924) (if_stmt 17925) (if_stmt_without_else 17926) (inline_function 113) (inner_statement 17927) (interface_declaration_statement 17928) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17929) (static_member 123) (trait_declaration_statement 17930) (variable 126) (variable_class_name 127) ("{" 17931) ("~" 129)) 7829 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 27605) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 7830 ((T_STRING 27598)) 7831 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 27596) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 7832 (("(" 27587)) 7833 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22932) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22933) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22934) (T_DEC 22) (T_DECLARE 22935) (T_DIR 24) (T_DNUMBER 25) (T_DO 22936) (T_DOUBLE_CAST 27) (T_ECHO 22937) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22938) (T_FOREACH 22939) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22940) (T_GOTO 22941) (T_IF 22942) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22943) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22944) (T_START_HEREDOC 69) (T_STATIC 22945) (T_STRING 22946) (T_STRING_CAST 72) (T_SWITCH 22947) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22948) (T_UNSET 22949) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22950) (T_WHILE 22951) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22952) (alt_if_stmt_without_else 22953) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22954) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22955) (if_stmt_without_else 22956) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27581) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22958) ("~" 129)) 7834 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (echo_expr 22926) (echo_expr_list 27579) (expr 22928) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 7835 ((";" 27578)) 7836 ((T_STRING 27571)) 7837 (("(" 27558)) 7838 (("(" 27543)) 7839 (("$" 19441) (T_VARIABLE 19442) (global_var 19443) (global_var_list 27541) (simple_variable 19445)) 7840 ((T_STRING 27539)) 7841 (("(" 27536)) 7842 (("(" 27532)) 7843 ((T_STRING 27526)) 7844 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 27524) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 7845 ((T_FN 35) (T_FUNCTION 131) (T_VARIABLE 18365) (fn 106) #'138 (inline_function 18366) (static_var 18367) (static_var_list 27522)) 7846 ((":" 27521)) 7847 (("(" 27503)) 7848 ((T_STRING 27498)) 7849 (("{" 27481)) 7850 (("(" 27476)) 7851 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 27474) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 7852 (("(" 27465)) 7853 ((T_ELSE 27458) (T_ELSEIF 17711) (T_ENDIF 27459)) 7854 ((T_ABSTRACT 10) (T_ATTRIBUTE 174) (T_CLASS 17891) (T_ENUM 17897) (T_FINAL 34) (T_FN 35) (T_FUNCTION 38) (T_INTERFACE 17905) (T_READONLY 2955) (T_STATIC 17797) (T_TRAIT 17910) (attribute 17798) (attributed_statement 27457) (class_declaration_statement 17919) (class_modifier 99) (class_modifiers 17920) (enum_declaration_statement 17921) (fn 106) #'17923 (function_declaration_statement 17924) (inline_function 17800) (interface_declaration_statement 17928) (trait_declaration_statement 17930)) 7855 ((T_ABSTRACT 10) (T_CLASS 27449) (T_FINAL 34) (T_READONLY 2955) (class_modifier 17789)) 7856 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 27448) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 7857 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17770) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17771) (ampersand 17772) (returns_ref 27436)) 7858 ((T_ELSE 17934) (T_ELSEIF 17935)) 7859 ((inner_statement_list 17932)) 7860 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 17933) ("~" 129)) 7861 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17889) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 17890) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17892) (T_DEC 22) (T_DECLARE 17893) (T_DIR 24) (T_DNUMBER 25) (T_DO 17894) (T_DOUBLE_CAST 27) (T_ECHO 17895) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17898) (T_FOREACH 17899) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 17900) (T_GOTO 17901) (T_IF 17903) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17904) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17906) (T_START_HEREDOC 69) (T_STATIC 17907) (T_STRING 17908) (T_STRING_CAST 72) (T_SWITCH 17909) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17911) (T_UNSET 17912) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17913) (T_WHILE 17914) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17915) (alt_if_stmt_without_else 17916) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 17922) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 17925) (if_stmt_without_else 17926) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27435) (static_member 123) (variable 126) (variable_class_name 127) ("{" 17931) ("~" 129)) 7862 (("(" 17936)) 7863 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 17937) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7864 (("%" 304) (")" 17938) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 7865 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17939) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 17940) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17941) (T_DEC 22) (T_DECLARE 17942) (T_DIR 24) (T_DNUMBER 25) (T_DO 17943) (T_DOUBLE_CAST 27) (T_ECHO 17944) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17945) (T_FOREACH 17946) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 17947) (T_GOTO 17948) (T_IF 17903) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17949) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17950) (T_START_HEREDOC 69) (T_STATIC 17951) (T_STRING 17952) (T_STRING_CAST 72) (T_SWITCH 17953) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17954) (T_UNSET 17955) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17956) (T_WHILE 17957) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17958) (alt_if_stmt_without_else 17959) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 17960) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 17961) (if_stmt_without_else 17962) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17963) (static_member 123) (variable 126) (variable_class_name 127) ("{" 17964) ("~" 129)) 7866 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 27433) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 7867 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 27431) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 7868 (("(" 27422)) 7869 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22932) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22933) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22934) (T_DEC 22) (T_DECLARE 22935) (T_DIR 24) (T_DNUMBER 25) (T_DO 22936) (T_DOUBLE_CAST 27) (T_ECHO 22937) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22938) (T_FOREACH 22939) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22940) (T_GOTO 22941) (T_IF 22942) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22943) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22944) (T_START_HEREDOC 69) (T_STATIC 22945) (T_STRING 22946) (T_STRING_CAST 72) (T_SWITCH 22947) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22948) (T_UNSET 22949) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22950) (T_WHILE 22951) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22952) (alt_if_stmt_without_else 22953) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22954) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22955) (if_stmt_without_else 22956) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27416) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22958) ("~" 129)) 7870 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (echo_expr 22926) (echo_expr_list 27414) (expr 22928) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 7871 (("(" 27401)) 7872 (("(" 27386)) 7873 (("$" 19441) (T_VARIABLE 19442) (global_var 19443) (global_var_list 27384) (simple_variable 19445)) 7874 ((T_STRING 27382)) 7875 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 27380) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 7876 ((T_FN 35) (T_FUNCTION 131) (T_VARIABLE 18365) (fn 106) #'138 (inline_function 18366) (static_var 18367) (static_var_list 27378)) 7877 ((":" 27377)) 7878 (("(" 18072)) 7879 (("{" 18045)) 7880 (("(" 17988)) 7881 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 17986) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 7882 (("(" 17977)) 7883 ((T_ELSE 17970) (T_ELSEIF 17711) (T_ENDIF 17971)) 7884 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 17969) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 7885 ((T_ELSE 17967) (T_ELSEIF 17935)) 7886 ((inner_statement_list 17965)) 7887 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 17966) ("~" 129)) 7888 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17939) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 17940) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17941) (T_DEC 22) (T_DECLARE 17942) (T_DIR 24) (T_DNUMBER 25) (T_DO 17943) (T_DOUBLE_CAST 27) (T_ECHO 17944) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17945) (T_FOREACH 17946) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 17947) (T_GOTO 17948) (T_IF 17903) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17949) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17950) (T_START_HEREDOC 69) (T_STATIC 17951) (T_STRING 17952) (T_STRING_CAST 72) (T_SWITCH 17953) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17954) (T_UNSET 17955) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17956) (T_WHILE 17957) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17958) (alt_if_stmt_without_else 17959) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 17960) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 17961) (if_stmt_without_else 17962) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17968) (static_member 123) (variable 126) (variable_class_name 127) ("{" 17964) ("~" 129)) 7889 ((":" 17973)) 7890 ((";" 17972)) 7891 ((inner_statement_list 17974)) 7892 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17807) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17808) (T_CLASS 17809) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17810) (T_DEC 22) (T_DECLARE 17811) (T_DIR 24) (T_DNUMBER 25) (T_DO 17812) (T_DOUBLE_CAST 27) (T_ECHO 17813) (T_EMPTY 29) (T_ENDIF 17975) (T_ENUM 17815) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17816) (T_FOREACH 17817) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17818) (T_GOTO 17819) (T_HALT_COMPILER 17820) (T_IF 17732) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17821) (T_INTERFACE 17822) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17823) (T_START_HEREDOC 69) (T_STATIC 17824) (T_STRING 17825) (T_STRING_CAST 72) (T_SWITCH 17826) (T_THROW 74) (T_TRAIT 17827) (T_TRAIT_C 76) (T_TRY 17828) (T_UNSET 17829) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17830) (T_WHILE 17831) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17832) (alt_if_stmt_without_else 17833) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17834) (attributes 17835) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17836) (class_modifier 99) (class_modifiers 17837) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17838) (expr 17839) (fn 106) (fully_dereferenceable 107) #'17840 (function_call 109) (function_declaration_statement 17841) (if_stmt 17842) (if_stmt_without_else 17843) (inline_function 113) (inner_statement 17844) (interface_declaration_statement 17845) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17846) (static_member 123) (trait_declaration_statement 17847) (variable 126) (variable_class_name 127) ("{" 17848) ("~" 129)) 7893 ((";" 17976)) 7894 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 17978) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7895 (("%" 304) (")" 17979) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 7896 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 17980) (";" 17939) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 17940) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17941) (T_DEC 22) (T_DECLARE 17942) (T_DIR 24) (T_DNUMBER 25) (T_DO 17943) (T_DOUBLE_CAST 27) (T_ECHO 17944) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17945) (T_FOREACH 17946) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 17947) (T_GOTO 17948) (T_IF 17903) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17949) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17950) (T_START_HEREDOC 69) (T_STATIC 17951) (T_STRING 17952) (T_STRING_CAST 72) (T_SWITCH 17953) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17954) (T_UNSET 17955) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17956) (T_WHILE 17957) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17958) (alt_if_stmt_without_else 17959) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 17960) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 17961) (if_stmt_without_else 17962) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17981) (static_member 123) (variable 126) (variable_class_name 127) (while_statement 17982) ("{" 17964) ("~" 129)) 7897 ((inner_statement_list 17983)) 7898 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17889) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17890) (T_CLASS 17891) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17892) (T_DEC 22) (T_DECLARE 17893) (T_DIR 24) (T_DNUMBER 25) (T_DO 17894) (T_DOUBLE_CAST 27) (T_ECHO 17895) (T_EMPTY 29) (T_ENDWHILE 17984) (T_ENUM 17897) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17898) (T_FOREACH 17899) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17900) (T_GOTO 17901) (T_HALT_COMPILER 17902) (T_IF 17903) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17904) (T_INTERFACE 17905) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17906) (T_START_HEREDOC 69) (T_STATIC 17907) (T_STRING 17908) (T_STRING_CAST 72) (T_SWITCH 17909) (T_THROW 74) (T_TRAIT 17910) (T_TRAIT_C 76) (T_TRY 17911) (T_UNSET 17912) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17913) (T_WHILE 17914) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17915) (alt_if_stmt_without_else 17916) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17917) (attributes 17918) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17919) (class_modifier 99) (class_modifiers 17920) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17921) (expr 17922) (fn 106) (fully_dereferenceable 107) #'17923 (function_call 109) (function_declaration_statement 17924) (if_stmt 17925) (if_stmt_without_else 17926) (inline_function 113) (inner_statement 17927) (interface_declaration_statement 17928) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17929) (static_member 123) (trait_declaration_statement 17930) (variable 126) (variable_class_name 127) ("{" 17931) ("~" 129)) 7899 ((";" 17985)) 7900 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 17987) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 7901 (("\"" 1538) ("$" 17989) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 17990) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 17991) ("[" 1560) (array_object_dereferenceable 17992) (callable_expr 17993) (callable_variable 17994) (class_constant 1564) (class_name 17995) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 17996) (name 17997) (new_dereferenceable 1570) (simple_variable 17998) (static_member 17999) (unset_variable 18000) (unset_variables 18001) (variable 18002) (variable_class_name 18003)) 7902 (("$" 17989) (T_VARIABLE 17991) (simple_variable 18041) ("{" 18042)) 7903 (("(" 18007) (argument_list 18040)) 7904 ((T_NULLSAFE_OBJECT_OPERATOR 18026) (T_OBJECT_OPERATOR 18027) ("[" 18028)) 7905 (("(" 18007) (argument_list 18025)) 7906 ((T_PAAMAYIM_NEKUDOTAYIM 18021)) 7907 (("(" 18007) (argument_list 18020)) 7908 (("," 18015) (possible_comma 18016)) 7909 ((T_PAAMAYIM_NEKUDOTAYIM 18004)) 7910 (("$" 17989) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 17991) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 18005) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 18006) ("{" 1660)) 7911 (("(" 18007) (argument_list 18008)) 7912 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 18009) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 18010) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 18011) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 7913 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 18014) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 7914 (("," 1498) (possible_comma 18012)) 7915 ((")" 18013)) 7916 (("\"" 1538) ("$" 17989) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 17990) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 17991) ("[" 1560) (array_object_dereferenceable 17992) (callable_expr 17993) (callable_variable 17994) (class_constant 1564) (class_name 17995) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 17996) (name 17997) (new_dereferenceable 1570) (simple_variable 17998) (static_member 17999) (unset_variable 18019) (variable 18002) (variable_class_name 18003)) 7917 ((")" 18017)) 7918 ((";" 18018)) 7919 (("$" 17989) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 17991) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 18022) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 18023) ("{" 1676)) 7920 (("(" 18007) (argument_list 18024)) 7921 (("$" 17989) (T_STRING 18031) (T_VARIABLE 17991) (property_name 18038) (simple_variable 18033) ("{" 18034)) 7922 (("$" 17989) (T_STRING 18031) (T_VARIABLE 17991) (property_name 18032) (simple_variable 18033) ("{" 18034)) 7923 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 18029) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 7924 (("]" 18030)) 7925 (("(" 18007) (argument_list 18037)) 7926 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 18035) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 7927 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 18036)) 7928 (("(" 18007) (argument_list 18039)) 7929 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 18043) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 7930 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 18044)) 7931 ((inner_statement_list 18046)) 7932 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 18047) ("~" 129)) 7933 ((catch_list 18048)) 7934 ((T_CATCH 18049) (T_FINALLY 18050) (finally_statement 18051)) 7935 (("(" 18055)) 7936 (("{" 18052)) 7937 ((inner_statement_list 18053)) 7938 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 18054) ("~" 129)) 7939 ((T_NAME_FULLY_QUALIFIED 18056) (T_NAME_QUALIFIED 18057) (T_NAME_RELATIVE 18058) (T_STATIC 18059) (T_STRING 18060) (catch_name_list 18061) (class_name 18062) (name 18063)) 7940 ((T_VARIABLE 18064) (optional_variable 18065) ("|" 18066)) 7941 ((")" 18068)) 7942 ((T_NAME_FULLY_QUALIFIED 18056) (T_NAME_QUALIFIED 18057) (T_NAME_RELATIVE 18058) (T_STATIC 18059) (T_STRING 18060) (class_name 18067) (name 18063)) 7943 (("{" 18069)) 7944 ((inner_statement_list 18070)) 7945 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 18071) ("~" 129)) 7946 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 18073) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7947 (("%" 304) (")" 18074) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 7948 ((":" 18075) (switch_case_list 18076) ("{" 18077)) 7949 ((";" 27370) (case_list 27371)) 7950 ((";" 18078) (case_list 18079)) 7951 ((case_list 27368)) 7952 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 18082)) 7953 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 27363) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 7954 ((":" 18083) (";" 18084)) 7955 ((inner_statement_list 27362)) 7956 ((inner_statement_list 18085)) 7957 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 18086) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 18087) (T_CLASS 18088) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18089) (T_DEC 22) (T_DECLARE 18090) (T_DIR 24) (T_DNUMBER 25) (T_DO 18091) (T_DOUBLE_CAST 27) (T_ECHO 18092) (T_EMPTY 29) (T_ENUM 18093) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 18094) (T_FOREACH 18095) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 18096) (T_GOTO 18097) (T_HALT_COMPILER 18098) (T_IF 18099) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18100) (T_INTERFACE 18101) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 18102) (T_START_HEREDOC 69) (T_STATIC 18103) (T_STRING 18104) (T_STRING_CAST 72) (T_SWITCH 18105) (T_THROW 74) (T_TRAIT 18106) (T_TRAIT_C 76) (T_TRY 18107) (T_UNSET 18108) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 18109) (T_WHILE 18110) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 18111) (alt_if_stmt_without_else 18112) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 18113) (attributes 18114) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 18115) (class_modifier 99) (class_modifiers 18116) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 18117) (expr 18118) (fn 106) (fully_dereferenceable 107) #'18119 (function_call 109) (function_declaration_statement 18120) (if_stmt 18121) (if_stmt_without_else 18122) (inline_function 113) (inner_statement 18123) (interface_declaration_statement 18124) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 18125) (static_member 123) (trait_declaration_statement 18126) (variable 126) (variable_class_name 127) ("{" 18127) ("~" 129)) 7958 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 27360) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 7959 ((T_STRING 27353)) 7960 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 27351) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 7961 (("(" 27342)) 7962 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22932) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22933) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22934) (T_DEC 22) (T_DECLARE 22935) (T_DIR 24) (T_DNUMBER 25) (T_DO 22936) (T_DOUBLE_CAST 27) (T_ECHO 22937) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22938) (T_FOREACH 22939) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22940) (T_GOTO 22941) (T_IF 22942) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22943) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22944) (T_START_HEREDOC 69) (T_STATIC 22945) (T_STRING 22946) (T_STRING_CAST 72) (T_SWITCH 22947) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22948) (T_UNSET 22949) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22950) (T_WHILE 22951) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22952) (alt_if_stmt_without_else 22953) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22954) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22955) (if_stmt_without_else 22956) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27336) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22958) ("~" 129)) 7963 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (echo_expr 22926) (echo_expr_list 27334) (expr 22928) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 7964 ((T_STRING 27327)) 7965 (("(" 27314)) 7966 (("(" 27299)) 7967 (("$" 19441) (T_VARIABLE 19442) (global_var 19443) (global_var_list 27297) (simple_variable 19445)) 7968 ((T_STRING 27295)) 7969 (("(" 27292)) 7970 (("(" 27288)) 7971 ((T_STRING 27282)) 7972 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 27280) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 7973 ((T_FN 35) (T_FUNCTION 131) (T_VARIABLE 18365) (fn 106) #'138 (inline_function 18366) (static_var 18367) (static_var_list 27278)) 7974 ((":" 27277)) 7975 (("(" 27259)) 7976 ((T_STRING 27254)) 7977 (("{" 27237)) 7978 (("(" 27232)) 7979 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 27230) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 7980 (("(" 27221)) 7981 ((T_ELSE 27214) (T_ELSEIF 17711) (T_ENDIF 27215)) 7982 ((T_ABSTRACT 10) (T_ATTRIBUTE 174) (T_CLASS 18088) (T_ENUM 18093) (T_FINAL 34) (T_FN 35) (T_FUNCTION 38) (T_INTERFACE 18101) (T_READONLY 2955) (T_STATIC 17797) (T_TRAIT 18106) (attribute 17798) (attributed_statement 27213) (class_declaration_statement 18115) (class_modifier 99) (class_modifiers 18116) (enum_declaration_statement 18117) (fn 106) #'18119 (function_declaration_statement 18120) (inline_function 17800) (interface_declaration_statement 18124) (trait_declaration_statement 18126)) 7983 ((T_ABSTRACT 10) (T_CLASS 27205) (T_FINAL 34) (T_READONLY 2955) (class_modifier 17789)) 7984 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 27204) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 7985 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17770) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17771) (ampersand 17772) (returns_ref 27192)) 7986 ((T_ELSE 18130) (T_ELSEIF 18131)) 7987 ((inner_statement_list 18128)) 7988 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 18129) ("~" 129)) 7989 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 18086) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 18087) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18089) (T_DEC 22) (T_DECLARE 18090) (T_DIR 24) (T_DNUMBER 25) (T_DO 18091) (T_DOUBLE_CAST 27) (T_ECHO 18092) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18094) (T_FOREACH 18095) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 18096) (T_GOTO 18097) (T_IF 18099) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18100) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 18102) (T_START_HEREDOC 69) (T_STATIC 18103) (T_STRING 18104) (T_STRING_CAST 72) (T_SWITCH 18105) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 18107) (T_UNSET 18108) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 18109) (T_WHILE 18110) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 18111) (alt_if_stmt_without_else 18112) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 18118) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 18121) (if_stmt_without_else 18122) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27191) (static_member 123) (variable 126) (variable_class_name 127) ("{" 18127) ("~" 129)) 7990 (("(" 18132)) 7991 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 18133) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 7992 (("%" 304) (")" 18134) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 7993 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 18135) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 18136) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18137) (T_DEC 22) (T_DECLARE 18138) (T_DIR 24) (T_DNUMBER 25) (T_DO 18139) (T_DOUBLE_CAST 27) (T_ECHO 18140) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18141) (T_FOREACH 18142) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 18143) (T_GOTO 18144) (T_IF 18099) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18145) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 18146) (T_START_HEREDOC 69) (T_STATIC 18147) (T_STRING 18148) (T_STRING_CAST 72) (T_SWITCH 18149) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 18150) (T_UNSET 18151) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 18152) (T_WHILE 18153) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 18154) (alt_if_stmt_without_else 18155) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 18156) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 18157) (if_stmt_without_else 18158) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 18159) (static_member 123) (variable 126) (variable_class_name 127) ("{" 18160) ("~" 129)) 7994 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 27189) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 7995 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 27187) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 7996 (("(" 27178)) 7997 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22932) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22933) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22934) (T_DEC 22) (T_DECLARE 22935) (T_DIR 24) (T_DNUMBER 25) (T_DO 22936) (T_DOUBLE_CAST 27) (T_ECHO 22937) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22938) (T_FOREACH 22939) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22940) (T_GOTO 22941) (T_IF 22942) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22943) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22944) (T_START_HEREDOC 69) (T_STATIC 22945) (T_STRING 22946) (T_STRING_CAST 72) (T_SWITCH 22947) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22948) (T_UNSET 22949) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22950) (T_WHILE 22951) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22952) (alt_if_stmt_without_else 22953) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22954) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22955) (if_stmt_without_else 22956) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27172) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22958) ("~" 129)) 7998 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (echo_expr 22926) (echo_expr_list 27170) (expr 22928) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 7999 (("(" 27157)) 8000 (("(" 27142)) 8001 (("$" 19441) (T_VARIABLE 19442) (global_var 19443) (global_var_list 27140) (simple_variable 19445)) 8002 ((T_STRING 27138)) 8003 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 27136) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 8004 ((T_FN 35) (T_FUNCTION 131) (T_VARIABLE 18365) (fn 106) #'138 (inline_function 18366) (static_var 18367) (static_var_list 27134)) 8005 ((":" 27133)) 8006 (("(" 18206)) 8007 (("{" 18189)) 8008 (("(" 18184)) 8009 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 18182) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 8010 (("(" 18173)) 8011 ((T_ELSE 18166) (T_ELSEIF 17711) (T_ENDIF 18167)) 8012 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 18165) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 8013 ((T_ELSE 18163) (T_ELSEIF 18131)) 8014 ((inner_statement_list 18161)) 8015 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 18162) ("~" 129)) 8016 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 18135) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 18136) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18137) (T_DEC 22) (T_DECLARE 18138) (T_DIR 24) (T_DNUMBER 25) (T_DO 18139) (T_DOUBLE_CAST 27) (T_ECHO 18140) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18141) (T_FOREACH 18142) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 18143) (T_GOTO 18144) (T_IF 18099) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18145) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 18146) (T_START_HEREDOC 69) (T_STATIC 18147) (T_STRING 18148) (T_STRING_CAST 72) (T_SWITCH 18149) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 18150) (T_UNSET 18151) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 18152) (T_WHILE 18153) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 18154) (alt_if_stmt_without_else 18155) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 18156) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 18157) (if_stmt_without_else 18158) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 18164) (static_member 123) (variable 126) (variable_class_name 127) ("{" 18160) ("~" 129)) 8017 ((":" 18169)) 8018 ((";" 18168)) 8019 ((inner_statement_list 18170)) 8020 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17807) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17808) (T_CLASS 17809) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17810) (T_DEC 22) (T_DECLARE 17811) (T_DIR 24) (T_DNUMBER 25) (T_DO 17812) (T_DOUBLE_CAST 27) (T_ECHO 17813) (T_EMPTY 29) (T_ENDIF 18171) (T_ENUM 17815) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17816) (T_FOREACH 17817) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17818) (T_GOTO 17819) (T_HALT_COMPILER 17820) (T_IF 17732) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17821) (T_INTERFACE 17822) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17823) (T_START_HEREDOC 69) (T_STATIC 17824) (T_STRING 17825) (T_STRING_CAST 72) (T_SWITCH 17826) (T_THROW 74) (T_TRAIT 17827) (T_TRAIT_C 76) (T_TRY 17828) (T_UNSET 17829) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17830) (T_WHILE 17831) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17832) (alt_if_stmt_without_else 17833) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17834) (attributes 17835) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17836) (class_modifier 99) (class_modifiers 17837) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17838) (expr 17839) (fn 106) (fully_dereferenceable 107) #'17840 (function_call 109) (function_declaration_statement 17841) (if_stmt 17842) (if_stmt_without_else 17843) (inline_function 113) (inner_statement 17844) (interface_declaration_statement 17845) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17846) (static_member 123) (trait_declaration_statement 17847) (variable 126) (variable_class_name 127) ("{" 17848) ("~" 129)) 8021 ((";" 18172)) 8022 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 18174) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 8023 (("%" 304) (")" 18175) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 8024 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 18176) (";" 18135) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 18136) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18137) (T_DEC 22) (T_DECLARE 18138) (T_DIR 24) (T_DNUMBER 25) (T_DO 18139) (T_DOUBLE_CAST 27) (T_ECHO 18140) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18141) (T_FOREACH 18142) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 18143) (T_GOTO 18144) (T_IF 18099) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18145) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 18146) (T_START_HEREDOC 69) (T_STATIC 18147) (T_STRING 18148) (T_STRING_CAST 72) (T_SWITCH 18149) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 18150) (T_UNSET 18151) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 18152) (T_WHILE 18153) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 18154) (alt_if_stmt_without_else 18155) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 18156) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 18157) (if_stmt_without_else 18158) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 18177) (static_member 123) (variable 126) (variable_class_name 127) (while_statement 18178) ("{" 18160) ("~" 129)) 8025 ((inner_statement_list 18179)) 8026 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17889) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17890) (T_CLASS 17891) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17892) (T_DEC 22) (T_DECLARE 17893) (T_DIR 24) (T_DNUMBER 25) (T_DO 17894) (T_DOUBLE_CAST 27) (T_ECHO 17895) (T_EMPTY 29) (T_ENDWHILE 18180) (T_ENUM 17897) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17898) (T_FOREACH 17899) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17900) (T_GOTO 17901) (T_HALT_COMPILER 17902) (T_IF 17903) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17904) (T_INTERFACE 17905) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17906) (T_START_HEREDOC 69) (T_STATIC 17907) (T_STRING 17908) (T_STRING_CAST 72) (T_SWITCH 17909) (T_THROW 74) (T_TRAIT 17910) (T_TRAIT_C 76) (T_TRY 17911) (T_UNSET 17912) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17913) (T_WHILE 17914) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17915) (alt_if_stmt_without_else 17916) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17917) (attributes 17918) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17919) (class_modifier 99) (class_modifiers 17920) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17921) (expr 17922) (fn 106) (fully_dereferenceable 107) #'17923 (function_call 109) (function_declaration_statement 17924) (if_stmt 17925) (if_stmt_without_else 17926) (inline_function 113) (inner_statement 17927) (interface_declaration_statement 17928) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17929) (static_member 123) (trait_declaration_statement 17930) (variable 126) (variable_class_name 127) ("{" 17931) ("~" 129)) 8027 ((";" 18181)) 8028 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 18183) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 8029 (("\"" 1538) ("$" 17989) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 17990) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 17991) ("[" 1560) (array_object_dereferenceable 17992) (callable_expr 17993) (callable_variable 17994) (class_constant 1564) (class_name 17995) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 17996) (name 17997) (new_dereferenceable 1570) (simple_variable 17998) (static_member 17999) (unset_variable 18000) (unset_variables 18185) (variable 18002) (variable_class_name 18003)) 8030 (("," 18015) (possible_comma 18186)) 8031 ((")" 18187)) 8032 ((";" 18188)) 8033 ((inner_statement_list 18190)) 8034 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 18191) ("~" 129)) 8035 ((catch_list 18192)) 8036 ((T_CATCH 18193) (T_FINALLY 18194) (finally_statement 18195)) 8037 (("(" 18199)) 8038 (("{" 18196)) 8039 ((inner_statement_list 18197)) 8040 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 18198) ("~" 129)) 8041 ((T_NAME_FULLY_QUALIFIED 18056) (T_NAME_QUALIFIED 18057) (T_NAME_RELATIVE 18058) (T_STATIC 18059) (T_STRING 18060) (catch_name_list 18200) (class_name 18062) (name 18063)) 8042 ((T_VARIABLE 18064) (optional_variable 18201) ("|" 18066)) 8043 ((")" 18202)) 8044 (("{" 18203)) 8045 ((inner_statement_list 18204)) 8046 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 18205) ("~" 129)) 8047 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 18207) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 8048 (("%" 304) (")" 18208) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 8049 ((":" 18209) (switch_case_list 18210) ("{" 18211)) 8050 ((";" 18217) (case_list 18218)) 8051 ((";" 18212) (case_list 18213)) 8052 ((case_list 18215)) 8053 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 18214)) 8054 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 18216)) 8055 ((case_list 27130)) 8056 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 18221)) 8057 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25064) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 8058 ((":" 18223) (";" 18224)) 8059 ((";" 18222)) 8060 ((inner_statement_list 25000)) 8061 ((inner_statement_list 18225)) 8062 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 18226) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 18227) (T_CLASS 18228) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18229) (T_DEC 22) (T_DECLARE 18230) (T_DIR 24) (T_DNUMBER 25) (T_DO 18231) (T_DOUBLE_CAST 27) (T_ECHO 18232) (T_EMPTY 29) (T_ENUM 18233) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 18234) (T_FOREACH 18235) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 18236) (T_GOTO 18237) (T_HALT_COMPILER 18238) (T_IF 18239) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18240) (T_INTERFACE 18241) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 18242) (T_START_HEREDOC 69) (T_STATIC 18243) (T_STRING 18244) (T_STRING_CAST 72) (T_SWITCH 18245) (T_THROW 74) (T_TRAIT 18246) (T_TRAIT_C 76) (T_TRY 18247) (T_UNSET 18248) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 18249) (T_WHILE 18250) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 18251) (alt_if_stmt_without_else 18252) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 18253) (attributes 18254) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 18255) (class_modifier 99) (class_modifiers 18256) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 18257) (expr 18258) (fn 106) (fully_dereferenceable 107) #'18259 (function_call 109) (function_declaration_statement 18260) (if_stmt 18261) (if_stmt_without_else 18262) (inline_function 113) (inner_statement 18263) (interface_declaration_statement 18264) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 18265) (static_member 123) (trait_declaration_statement 18266) (variable 126) (variable_class_name 127) ("{" 18267) ("~" 129)) 8063 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 24998) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 8064 ((T_STRING 24991)) 8065 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 24989) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 8066 (("(" 24980)) 8067 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22932) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22933) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22934) (T_DEC 22) (T_DECLARE 22935) (T_DIR 24) (T_DNUMBER 25) (T_DO 22936) (T_DOUBLE_CAST 27) (T_ECHO 22937) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22938) (T_FOREACH 22939) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22940) (T_GOTO 22941) (T_IF 22942) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22943) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22944) (T_START_HEREDOC 69) (T_STATIC 22945) (T_STRING 22946) (T_STRING_CAST 72) (T_SWITCH 22947) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22948) (T_UNSET 22949) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22950) (T_WHILE 22951) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22952) (alt_if_stmt_without_else 22953) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22954) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22955) (if_stmt_without_else 22956) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 24974) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22958) ("~" 129)) 8068 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (echo_expr 22926) (echo_expr_list 24972) (expr 22928) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8069 ((T_STRING 24965)) 8070 (("(" 24952)) 8071 (("(" 24937)) 8072 (("$" 19441) (T_VARIABLE 19442) (global_var 19443) (global_var_list 24935) (simple_variable 19445)) 8073 ((T_STRING 24933)) 8074 (("(" 24930)) 8075 (("(" 24926)) 8076 ((T_STRING 24920)) 8077 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 24918) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 8078 ((T_FN 35) (T_FUNCTION 131) (T_VARIABLE 18365) (fn 106) #'138 (inline_function 18366) (static_var 18367) (static_var_list 24916)) 8079 ((":" 24915)) 8080 (("(" 24897)) 8081 ((T_STRING 24892)) 8082 (("{" 24875)) 8083 (("(" 24870)) 8084 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 24868) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 8085 (("(" 24859)) 8086 ((T_ELSE 24852) (T_ELSEIF 17711) (T_ENDIF 24853)) 8087 ((T_ABSTRACT 10) (T_ATTRIBUTE 174) (T_CLASS 18228) (T_ENUM 18233) (T_FINAL 34) (T_FN 35) (T_FUNCTION 38) (T_INTERFACE 18241) (T_READONLY 2955) (T_STATIC 17797) (T_TRAIT 18246) (attribute 17798) (attributed_statement 24851) (class_declaration_statement 18255) (class_modifier 99) (class_modifiers 18256) (enum_declaration_statement 18257) (fn 106) #'18259 (function_declaration_statement 18260) (inline_function 17800) (interface_declaration_statement 18264) (trait_declaration_statement 18266)) 8088 ((T_ABSTRACT 10) (T_CLASS 24843) (T_FINAL 34) (T_READONLY 2955) (class_modifier 17789)) 8089 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 24842) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 8090 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17770) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17771) (ampersand 17772) (returns_ref 24830)) 8091 ((T_ELSE 18270) (T_ELSEIF 18271)) 8092 ((inner_statement_list 18268)) 8093 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 18269) ("~" 129)) 8094 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 18226) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 18227) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18229) (T_DEC 22) (T_DECLARE 18230) (T_DIR 24) (T_DNUMBER 25) (T_DO 18231) (T_DOUBLE_CAST 27) (T_ECHO 18232) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18234) (T_FOREACH 18235) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 18236) (T_GOTO 18237) (T_IF 18239) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18240) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 18242) (T_START_HEREDOC 69) (T_STATIC 18243) (T_STRING 18244) (T_STRING_CAST 72) (T_SWITCH 18245) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 18247) (T_UNSET 18248) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 18249) (T_WHILE 18250) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 18251) (alt_if_stmt_without_else 18252) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 18258) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 18261) (if_stmt_without_else 18262) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 24829) (static_member 123) (variable 126) (variable_class_name 127) ("{" 18267) ("~" 129)) 8095 (("(" 18272)) 8096 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 18273) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 8097 (("%" 304) (")" 18274) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 8098 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 18275) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 18276) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18277) (T_DEC 22) (T_DECLARE 18278) (T_DIR 24) (T_DNUMBER 25) (T_DO 18279) (T_DOUBLE_CAST 27) (T_ECHO 18280) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18281) (T_FOREACH 18282) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 18283) (T_GOTO 18284) (T_IF 18239) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18285) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 18286) (T_START_HEREDOC 69) (T_STATIC 18287) (T_STRING 18288) (T_STRING_CAST 72) (T_SWITCH 18289) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 18290) (T_UNSET 18291) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 18292) (T_WHILE 18293) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 18294) (alt_if_stmt_without_else 18295) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 18296) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 18297) (if_stmt_without_else 18298) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 18299) (static_member 123) (variable 126) (variable_class_name 127) ("{" 18300) ("~" 129)) 8099 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 24827) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 8100 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 24825) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 8101 (("(" 24816)) 8102 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22932) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22933) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22934) (T_DEC 22) (T_DECLARE 22935) (T_DIR 24) (T_DNUMBER 25) (T_DO 22936) (T_DOUBLE_CAST 27) (T_ECHO 22937) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22938) (T_FOREACH 22939) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22940) (T_GOTO 22941) (T_IF 22942) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22943) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22944) (T_START_HEREDOC 69) (T_STATIC 22945) (T_STRING 22946) (T_STRING_CAST 72) (T_SWITCH 22947) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22948) (T_UNSET 22949) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22950) (T_WHILE 22951) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22952) (alt_if_stmt_without_else 22953) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22954) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22955) (if_stmt_without_else 22956) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 24810) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22958) ("~" 129)) 8103 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (echo_expr 22926) (echo_expr_list 24808) (expr 22928) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8104 (("(" 24795)) 8105 (("(" 19453)) 8106 (("$" 19441) (T_VARIABLE 19442) (global_var 19443) (global_var_list 19444) (simple_variable 19445)) 8107 ((T_STRING 19439)) 8108 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 19437) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 8109 ((T_FN 35) (T_FUNCTION 131) (T_VARIABLE 18365) (fn 106) #'138 (inline_function 18366) (static_var 18367) (static_var_list 18368)) 8110 ((":" 18364)) 8111 (("(" 18346)) 8112 (("{" 18329)) 8113 (("(" 18324)) 8114 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 18322) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 8115 (("(" 18313)) 8116 ((T_ELSE 18306) (T_ELSEIF 17711) (T_ENDIF 18307)) 8117 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 18305) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 8118 ((T_ELSE 18303) (T_ELSEIF 18271)) 8119 ((inner_statement_list 18301)) 8120 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 18302) ("~" 129)) 8121 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 18275) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 18276) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18277) (T_DEC 22) (T_DECLARE 18278) (T_DIR 24) (T_DNUMBER 25) (T_DO 18279) (T_DOUBLE_CAST 27) (T_ECHO 18280) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18281) (T_FOREACH 18282) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 18283) (T_GOTO 18284) (T_IF 18239) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18285) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 18286) (T_START_HEREDOC 69) (T_STATIC 18287) (T_STRING 18288) (T_STRING_CAST 72) (T_SWITCH 18289) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 18290) (T_UNSET 18291) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 18292) (T_WHILE 18293) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 18294) (alt_if_stmt_without_else 18295) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 18296) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 18297) (if_stmt_without_else 18298) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 18304) (static_member 123) (variable 126) (variable_class_name 127) ("{" 18300) ("~" 129)) 8122 ((":" 18309)) 8123 ((";" 18308)) 8124 ((inner_statement_list 18310)) 8125 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17807) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17808) (T_CLASS 17809) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17810) (T_DEC 22) (T_DECLARE 17811) (T_DIR 24) (T_DNUMBER 25) (T_DO 17812) (T_DOUBLE_CAST 27) (T_ECHO 17813) (T_EMPTY 29) (T_ENDIF 18311) (T_ENUM 17815) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17816) (T_FOREACH 17817) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17818) (T_GOTO 17819) (T_HALT_COMPILER 17820) (T_IF 17732) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17821) (T_INTERFACE 17822) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17823) (T_START_HEREDOC 69) (T_STATIC 17824) (T_STRING 17825) (T_STRING_CAST 72) (T_SWITCH 17826) (T_THROW 74) (T_TRAIT 17827) (T_TRAIT_C 76) (T_TRY 17828) (T_UNSET 17829) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17830) (T_WHILE 17831) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17832) (alt_if_stmt_without_else 17833) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17834) (attributes 17835) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17836) (class_modifier 99) (class_modifiers 17837) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17838) (expr 17839) (fn 106) (fully_dereferenceable 107) #'17840 (function_call 109) (function_declaration_statement 17841) (if_stmt 17842) (if_stmt_without_else 17843) (inline_function 113) (inner_statement 17844) (interface_declaration_statement 17845) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17846) (static_member 123) (trait_declaration_statement 17847) (variable 126) (variable_class_name 127) ("{" 17848) ("~" 129)) 8126 ((";" 18312)) 8127 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 18314) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 8128 (("%" 304) (")" 18315) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 8129 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 18316) (";" 18275) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 18276) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18277) (T_DEC 22) (T_DECLARE 18278) (T_DIR 24) (T_DNUMBER 25) (T_DO 18279) (T_DOUBLE_CAST 27) (T_ECHO 18280) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18281) (T_FOREACH 18282) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 18283) (T_GOTO 18284) (T_IF 18239) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18285) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 18286) (T_START_HEREDOC 69) (T_STATIC 18287) (T_STRING 18288) (T_STRING_CAST 72) (T_SWITCH 18289) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 18290) (T_UNSET 18291) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 18292) (T_WHILE 18293) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 18294) (alt_if_stmt_without_else 18295) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 18296) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 18297) (if_stmt_without_else 18298) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 18317) (static_member 123) (variable 126) (variable_class_name 127) (while_statement 18318) ("{" 18300) ("~" 129)) 8130 ((inner_statement_list 18319)) 8131 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17889) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17890) (T_CLASS 17891) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17892) (T_DEC 22) (T_DECLARE 17893) (T_DIR 24) (T_DNUMBER 25) (T_DO 17894) (T_DOUBLE_CAST 27) (T_ECHO 17895) (T_EMPTY 29) (T_ENDWHILE 18320) (T_ENUM 17897) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17898) (T_FOREACH 17899) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17900) (T_GOTO 17901) (T_HALT_COMPILER 17902) (T_IF 17903) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17904) (T_INTERFACE 17905) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17906) (T_START_HEREDOC 69) (T_STATIC 17907) (T_STRING 17908) (T_STRING_CAST 72) (T_SWITCH 17909) (T_THROW 74) (T_TRAIT 17910) (T_TRAIT_C 76) (T_TRY 17911) (T_UNSET 17912) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17913) (T_WHILE 17914) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17915) (alt_if_stmt_without_else 17916) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17917) (attributes 17918) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17919) (class_modifier 99) (class_modifiers 17920) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17921) (expr 17922) (fn 106) (fully_dereferenceable 107) #'17923 (function_call 109) (function_declaration_statement 17924) (if_stmt 17925) (if_stmt_without_else 17926) (inline_function 113) (inner_statement 17927) (interface_declaration_statement 17928) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17929) (static_member 123) (trait_declaration_statement 17930) (variable 126) (variable_class_name 127) ("{" 17931) ("~" 129)) 8132 ((";" 18321)) 8133 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 18323) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 8134 (("\"" 1538) ("$" 17989) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 17990) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 17991) ("[" 1560) (array_object_dereferenceable 17992) (callable_expr 17993) (callable_variable 17994) (class_constant 1564) (class_name 17995) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 17996) (name 17997) (new_dereferenceable 1570) (simple_variable 17998) (static_member 17999) (unset_variable 18000) (unset_variables 18325) (variable 18002) (variable_class_name 18003)) 8135 (("," 18015) (possible_comma 18326)) 8136 ((")" 18327)) 8137 ((";" 18328)) 8138 ((inner_statement_list 18330)) 8139 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 18331) ("~" 129)) 8140 ((catch_list 18332)) 8141 ((T_CATCH 18333) (T_FINALLY 18334) (finally_statement 18335)) 8142 (("(" 18339)) 8143 (("{" 18336)) 8144 ((inner_statement_list 18337)) 8145 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 18338) ("~" 129)) 8146 ((T_NAME_FULLY_QUALIFIED 18056) (T_NAME_QUALIFIED 18057) (T_NAME_RELATIVE 18058) (T_STATIC 18059) (T_STRING 18060) (catch_name_list 18340) (class_name 18062) (name 18063)) 8147 ((T_VARIABLE 18064) (optional_variable 18341) ("|" 18066)) 8148 ((")" 18342)) 8149 (("{" 18343)) 8150 ((inner_statement_list 18344)) 8151 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 18345) ("~" 129)) 8152 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 18347) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 8153 (("%" 304) (")" 18348) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 8154 ((":" 18349) (switch_case_list 18350) ("{" 18351)) 8155 ((";" 18357) (case_list 18358)) 8156 ((";" 18352) (case_list 18353)) 8157 ((case_list 18355)) 8158 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 18354)) 8159 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 18356)) 8160 ((case_list 18361)) 8161 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 18359)) 8162 ((";" 18360)) 8163 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 18362)) 8164 ((";" 18363)) 8165 (("=" 18372)) 8166 (("," 18369) (";" 18370)) 8167 ((T_VARIABLE 18365) (static_var 18371)) 8168 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18436) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8169 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 19435) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8170 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 19433) (encaps_var 5357)) 8171 (("$" 18375) (T_VARIABLE 18423) (simple_variable 19429) ("{" 19430)) 8172 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 19428) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 8173 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 19427) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8174 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 19426) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8175 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 19425) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8176 (("(" 19422)) 8177 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 19421) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8178 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 19420) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8179 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 19407) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (clone_argument_list 19408) (constant 18434) (dereferenceable_scalar 18435) (expr 19409) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8180 (("\"" 1538) ("$" 18680) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 18681) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 18682) ("[" 1560) (array_object_dereferenceable 18683) (callable_expr 18684) (callable_variable 18685) (class_constant 1564) (class_name 18686) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 18687) (name 18688) (new_dereferenceable 1570) (simple_variable 18689) (static_member 18690) (variable 19406) (variable_class_name 18692)) 8181 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 19405) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8182 (("(" 19402)) 8183 (("(" 19399)) 8184 (("(" 19390) (argument_list 19391) (ctor_arguments 19392)) 8185 (("\"" 1538) ("$" 18680) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 18681) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 18682) ("[" 1560) (array_object_dereferenceable 18683) (callable_expr 18684) (callable_variable 18685) (class_constant 1564) (class_name 18686) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 18687) (name 18688) (new_dereferenceable 1570) (simple_variable 18689) (static_member 18690) (variable 19389) (variable_class_name 18692)) 8186 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 19388) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8187 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 19387) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8188 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 19386) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8189 (("(" 19382)) 8190 (("(" 19377)) 8191 (("(" 19371)) 8192 (("$" 18680) ("(" 19334) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 19335) (T_NAME_QUALIFIED 19336) (T_NAME_RELATIVE 19337) (T_READONLY 2955) (T_STATIC 19338) (T_STRING 19339) (T_VARIABLE 18682) (anonymous_class 19340) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 19341) (attribute 2961) (attributes 19342) (class_modifier 2963) (class_name 19343) (class_name_reference 19344) (name 19345) (new_variable 19346) (simple_variable 19347)) 8193 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 19333) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8194 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 19332) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8195 (("(" 18641) (argument_list 19331)) 8196 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 19330) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8197 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 19329) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8198 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 19324) (T_END_HEREDOC 19325) (T_VARIABLE 2920) (encaps_list 19326) (encaps_var 2922)) 8199 ((T_FN 35) (T_FUNCTION 131) (fn 18437) #'18438 (inline_function 19323)) 8200 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 19322) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8201 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 19321) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8202 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 19320) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8203 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 18849) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8204 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18785) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8205 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 18781) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 8206 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 18779) (encaps_list 2828) (encaps_var 2829)) 8207 ((T_NULLSAFE_OBJECT_OPERATOR 18765) (T_OBJECT_OPERATOR 18766) ("[" 18767)) 8208 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 18762) (attribute 2757) (fn 18437) #'18438 (inline_function 18763)) 8209 (("(" 18641) (argument_list 18761)) 8210 ((T_PAAMAYIM_NEKUDOTAYIM 18753)) 8211 (("%" 18453) ("*" 18454) ("+" 18455) ("-" 18456) ("." 18457) ("/" 18458) ("<" 18459) (">" 18460) ("?" 18461) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 18462) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 18463) (T_BOOLEAN_AND 18464) (T_BOOLEAN_OR 18465) (T_COALESCE 18466) (T_INSTANCEOF 18467) (T_IS_EQUAL 18468) (T_IS_GREATER_OR_EQUAL 18469) (T_IS_IDENTICAL 18470) (T_IS_NOT_EQUAL 18471) (T_IS_NOT_IDENTICAL 18472) (T_IS_SMALLER_OR_EQUAL 18473) (T_LOGICAL_AND 18474) (T_LOGICAL_OR 18475) (T_LOGICAL_XOR 18476) (T_PIPE 18477) (T_POW 18478) (T_SL 18479) (T_SPACESHIP 18480) (T_SR 18481) ("^" 18482) ("|" 18483)) 8212 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 18742)) 8213 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 18730)) 8214 (("(" 18641) (argument_list 18729)) 8215 (("=" 18649) (T_AND_EQUAL 18650) (T_COALESCE_EQUAL 18651) (T_CONCAT_EQUAL 18652) (T_DEC 18653) (T_DIV_EQUAL 18654) (T_INC 18655) (T_MINUS_EQUAL 18656) (T_MOD_EQUAL 18657) (T_MUL_EQUAL 18658) (T_OR_EQUAL 18659) (T_PLUS_EQUAL 18660) (T_POW_EQUAL 18661) (T_SL_EQUAL 18662) (T_SR_EQUAL 18663) (T_XOR_EQUAL 18664)) 8216 ((T_PAAMAYIM_NEKUDOTAYIM 18553)) 8217 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18452) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8218 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18552) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8219 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18551) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8220 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18550) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8221 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18549) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8222 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18548) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8223 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18547) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8224 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18546) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8225 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18545) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8226 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 18540) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 18541) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 8227 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18539) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8228 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18538) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8229 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18537) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8230 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18536) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8231 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18535) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8232 (("$" 18500) ("(" 18501) (T_NAME_FULLY_QUALIFIED 18502) (T_NAME_QUALIFIED 18503) (T_NAME_RELATIVE 18504) (T_STATIC 18505) (T_STRING 18506) (T_VARIABLE 18507) (class_name 18508) (class_name_reference 18509) (name 18510) (new_variable 18511) (simple_variable 18512)) 8233 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18499) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8234 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18498) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8235 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18497) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8236 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18496) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8237 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18495) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8238 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18494) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8239 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18493) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8240 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18492) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8241 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18491) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8242 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18490) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8243 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18489) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8244 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18488) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8245 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18487) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8246 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18486) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8247 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18485) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8248 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18484) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8249 (("$" 18500) (T_VARIABLE 18507) (simple_variable 18531) ("{" 18532)) 8250 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 18529) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 8251 ((T_PAAMAYIM_NEKUDOTAYIM 18527)) 8252 ((T_NULLSAFE_OBJECT_OPERATOR 18513) (T_OBJECT_OPERATOR 18514) (T_PAAMAYIM_NEKUDOTAYIM 18515) ("[" 18516)) 8253 (("$" 18500) (T_STRING 18520) (T_VARIABLE 18507) (property_name 18526) (simple_variable 18522) ("{" 18523)) 8254 (("$" 18500) (T_STRING 18520) (T_VARIABLE 18507) (property_name 18521) (simple_variable 18522) ("{" 18523)) 8255 (("$" 18500) (T_VARIABLE 18507) (simple_variable 18519)) 8256 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 18517) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 8257 (("]" 18518)) 8258 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 18524) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 8259 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 18525)) 8260 (("$" 18500) (T_VARIABLE 18507) (simple_variable 18528)) 8261 (("%" 304) (")" 18530) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 8262 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 18533) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 8263 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 18534)) 8264 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18544) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8265 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 18542) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 8266 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18543) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8267 (("$" 18375) (T_ABSTRACT 18554) (T_ARRAY 18555) (T_AS 18556) (T_BREAK 18557) (T_CALLABLE 18558) (T_CASE 18559) (T_CATCH 18560) (T_CLASS 18561) (T_CLASS_C 18562) (T_CLONE 18563) (T_CONST 18564) (T_CONTINUE 18565) (T_DECLARE 18566) (T_DEFAULT 18567) (T_DIR 18568) (T_DO 18569) (T_ECHO 18570) (T_ELSE 18571) (T_ELSEIF 18572) (T_EMPTY 18573) (T_ENDDECLARE 18574) (T_ENDFOR 18575) (T_ENDFOREACH 18576) (T_ENDIF 18577) (T_ENDSWITCH 18578) (T_ENDWHILE 18579) (T_ENUM 18580) (T_EVAL 18581) (T_EXIT 18582) (T_EXTENDS 18583) (T_FILE 18584) (T_FINAL 18585) (T_FINALLY 18586) (T_FN 18587) (T_FOR 18588) (T_FOREACH 18589) (T_FUNCTION 18590) (T_FUNC_C 18591) (T_GLOBAL 18592) (T_GOTO 18593) (T_IF 18594) (T_IMPLEMENTS 18595) (T_INCLUDE 18596) (T_INCLUDE_ONCE 18597) (T_INSTANCEOF 18598) (T_INSTEADOF 18599) (T_INTERFACE 18600) (T_ISSET 18601) (T_LINE 18602) (T_LIST 18603) (T_LOGICAL_AND 18604) (T_LOGICAL_OR 18605) (T_LOGICAL_XOR 18606) (T_MATCH 18607) (T_METHOD_C 18608) (T_NAMESPACE 18609) (T_NEW 18610) (T_NS_C 18611) (T_PRINT 18612) (T_PRIVATE 18613) (T_PROPERTY_C 18614) (T_PROTECTED 18615) (T_PUBLIC 18616) (T_READONLY 18617) (T_REQUIRE 18618) (T_REQUIRE_ONCE 18619) (T_RETURN 18620) (T_STATIC 18621) (T_STRING 18622) (T_SWITCH 18623) (T_THROW 18624) (T_TRAIT 18625) (T_TRAIT_C 18626) (T_TRY 18627) (T_UNSET 18628) (T_USE 18629) (T_VAR 18630) (T_VARIABLE 18423) (T_WHILE 18631) (T_YIELD 18632) (identifier 18633) (member_name 18634) (reserved_non_modifiers 18635) (semi_reserved 18636) (simple_variable 18637) ("{" 18638)) 8268 (("(" 18641) (argument_list 18642)) 8269 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 18639) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 8270 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 18640)) 8271 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 18643) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 18644) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 18645) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 8272 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 18648) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 8273 (("," 1498) (possible_comma 18646)) 8274 ((")" 18647)) 8275 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (ampersand 18678) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18679) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8276 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18677) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8277 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18676) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8278 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18675) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8279 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18674) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8280 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18673) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8281 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18672) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8282 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18671) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8283 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18670) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8284 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18669) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8285 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18668) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8286 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18667) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8287 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18666) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8288 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18665) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8289 (("\"" 1538) ("$" 18680) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 18681) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 18682) ("[" 1560) (array_object_dereferenceable 18683) (callable_expr 18684) (callable_variable 18685) (class_constant 1564) (class_name 18686) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 18687) (name 18688) (new_dereferenceable 1570) (simple_variable 18689) (static_member 18690) (variable 18691) (variable_class_name 18692)) 8290 (("$" 18680) (T_VARIABLE 18682) (simple_variable 18725) ("{" 18726)) 8291 (("(" 18696) (argument_list 18724)) 8292 ((T_NULLSAFE_OBJECT_OPERATOR 18710) (T_OBJECT_OPERATOR 18711) ("[" 18712)) 8293 (("(" 18696) (argument_list 18709)) 8294 ((T_PAAMAYIM_NEKUDOTAYIM 18705)) 8295 (("(" 18696) (argument_list 18704)) 8296 ((T_PAAMAYIM_NEKUDOTAYIM 18693)) 8297 (("$" 18680) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 18682) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 18694) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 18695) ("{" 1660)) 8298 (("(" 18696) (argument_list 18697)) 8299 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 18698) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 18699) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 18700) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 8300 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 18703) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 8301 (("," 1498) (possible_comma 18701)) 8302 ((")" 18702)) 8303 (("$" 18680) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 18682) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 18706) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 18707) ("{" 1676)) 8304 (("(" 18696) (argument_list 18708)) 8305 (("$" 18680) (T_STRING 18715) (T_VARIABLE 18682) (property_name 18722) (simple_variable 18717) ("{" 18718)) 8306 (("$" 18680) (T_STRING 18715) (T_VARIABLE 18682) (property_name 18716) (simple_variable 18717) ("{" 18718)) 8307 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 18713) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 8308 (("]" 18714)) 8309 (("(" 18696) (argument_list 18721)) 8310 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 18719) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 8311 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 18720)) 8312 (("(" 18696) (argument_list 18723)) 8313 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 18727) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 8314 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 18728)) 8315 ((backup_doc_comment 18731)) 8316 (("(" 18732)) 8317 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 18733)) 8318 ((")" 18734)) 8319 ((T_USE 166) (lexical_vars 18735)) 8320 ((":" 168) (return_type 18736)) 8321 ((backup_fn_flags 18737)) 8322 (("{" 18738)) 8323 ((inner_statement_list 18739)) 8324 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 18740) ("~" 129)) 8325 ((backup_fn_flags 18741)) 8326 ((backup_doc_comment 18743)) 8327 (("(" 18744)) 8328 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 18745)) 8329 ((")" 18746)) 8330 ((":" 2271) (return_type 18747)) 8331 ((T_DOUBLE_ARROW 18748)) 8332 ((backup_fn_flags 18749)) 8333 ((backup_lex_pos 18750)) 8334 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 18751) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 8335 (("%" 3153) ("*" 3154) ("+" 3155) ("-" 3156) ("." 3157) ("/" 3158) ("<" 3159) (">" 3160) ("?" 3161) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 3162) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3163) (T_BOOLEAN_AND 3164) (T_BOOLEAN_OR 3165) (T_COALESCE 3166) (T_INSTANCEOF 3167) (T_IS_EQUAL 3168) (T_IS_GREATER_OR_EQUAL 3169) (T_IS_IDENTICAL 3170) (T_IS_NOT_EQUAL 3171) (T_IS_NOT_IDENTICAL 3172) (T_IS_SMALLER_OR_EQUAL 3173) (T_LOGICAL_AND 3174) (T_LOGICAL_OR 3175) (T_LOGICAL_XOR 3176) (T_PIPE 3177) (T_POW 3178) (T_SL 3179) (T_SPACESHIP 3180) (T_SR 3181) ("^" 3182) (backup_fn_flags 18752) ("|" 3183)) 8336 (("$" 18375) (T_ABSTRACT 18554) (T_ARRAY 18555) (T_AS 18556) (T_BREAK 18557) (T_CALLABLE 18558) (T_CASE 18559) (T_CATCH 18560) (T_CLASS 18561) (T_CLASS_C 18562) (T_CLONE 18563) (T_CONST 18564) (T_CONTINUE 18565) (T_DECLARE 18566) (T_DEFAULT 18567) (T_DIR 18568) (T_DO 18569) (T_ECHO 18570) (T_ELSE 18571) (T_ELSEIF 18572) (T_EMPTY 18573) (T_ENDDECLARE 18574) (T_ENDFOR 18575) (T_ENDFOREACH 18576) (T_ENDIF 18577) (T_ENDSWITCH 18578) (T_ENDWHILE 18579) (T_ENUM 18580) (T_EVAL 18581) (T_EXIT 18582) (T_EXTENDS 18583) (T_FILE 18584) (T_FINAL 18585) (T_FINALLY 18586) (T_FN 18587) (T_FOR 18588) (T_FOREACH 18589) (T_FUNCTION 18590) (T_FUNC_C 18591) (T_GLOBAL 18592) (T_GOTO 18593) (T_IF 18594) (T_IMPLEMENTS 18595) (T_INCLUDE 18596) (T_INCLUDE_ONCE 18597) (T_INSTANCEOF 18598) (T_INSTEADOF 18599) (T_INTERFACE 18600) (T_ISSET 18601) (T_LINE 18602) (T_LIST 18603) (T_LOGICAL_AND 18604) (T_LOGICAL_OR 18605) (T_LOGICAL_XOR 18606) (T_MATCH 18607) (T_METHOD_C 18608) (T_NAMESPACE 18609) (T_NEW 18610) (T_NS_C 18611) (T_PRINT 18612) (T_PRIVATE 18613) (T_PROPERTY_C 18614) (T_PROTECTED 18615) (T_PUBLIC 18616) (T_READONLY 18617) (T_REQUIRE 18618) (T_REQUIRE_ONCE 18619) (T_RETURN 18620) (T_STATIC 18621) (T_STRING 18622) (T_SWITCH 18623) (T_THROW 18624) (T_TRAIT 18625) (T_TRAIT_C 18626) (T_TRY 18627) (T_UNSET 18628) (T_USE 18629) (T_VAR 18630) (T_VARIABLE 18423) (T_WHILE 18631) (T_YIELD 18632) (identifier 18754) (member_name 18755) (reserved_non_modifiers 18635) (semi_reserved 18636) (simple_variable 18756) ("{" 18757)) 8337 (("(" 18641) (argument_list 18760)) 8338 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 18758) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 8339 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 18759)) 8340 ((T_FN 35) (T_FUNCTION 131) (fn 18437) #'18438 (inline_function 18764)) 8341 (("$" 18375) (T_STRING 18770) (T_VARIABLE 18423) (property_name 18777) (simple_variable 18772) ("{" 18773)) 8342 (("$" 18375) (T_STRING 18770) (T_VARIABLE 18423) (property_name 18771) (simple_variable 18772) ("{" 18773)) 8343 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 18768) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 8344 (("]" 18769)) 8345 (("(" 18641) (argument_list 18776)) 8346 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 18774) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 8347 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 18775)) 8348 (("(" 18641) (argument_list 18778)) 8349 (("`" 18780)) 8350 (("]" 18782)) 8351 (("=" 18783)) 8352 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 18784) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8353 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 19319) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8354 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 19317) (encaps_var 5357)) 8355 (("$" 18788) (T_VARIABLE 18836) (simple_variable 19313) ("{" 19314)) 8356 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 19312) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 8357 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 19311) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8358 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 19310) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8359 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 19309) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8360 (("(" 19306)) 8361 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 19305) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8362 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 19304) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8363 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 19291) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (clone_argument_list 19292) (constant 18847) (dereferenceable_scalar 18848) (expr 19293) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8364 (("\"" 1538) ("$" 19093) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 19094) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 19095) ("[" 1560) (array_object_dereferenceable 19096) (callable_expr 19097) (callable_variable 19098) (class_constant 1564) (class_name 19099) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 19100) (name 19101) (new_dereferenceable 1570) (simple_variable 19102) (static_member 19103) (variable 19290) (variable_class_name 19105)) 8365 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 19289) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8366 (("(" 19286)) 8367 (("(" 19283)) 8368 (("(" 19274) (argument_list 19275) (ctor_arguments 19276)) 8369 (("\"" 1538) ("$" 19093) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 19094) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 19095) ("[" 1560) (array_object_dereferenceable 19096) (callable_expr 19097) (callable_variable 19098) (class_constant 1564) (class_name 19099) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 19100) (name 19101) (new_dereferenceable 1570) (simple_variable 19102) (static_member 19103) (variable 19273) (variable_class_name 19105)) 8370 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 19272) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8371 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 19271) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8372 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 19270) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8373 (("(" 19266)) 8374 (("(" 19261)) 8375 (("(" 19255)) 8376 (("$" 19093) ("(" 19218) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 19219) (T_NAME_QUALIFIED 19220) (T_NAME_RELATIVE 19221) (T_READONLY 2955) (T_STATIC 19222) (T_STRING 19223) (T_VARIABLE 19095) (anonymous_class 19224) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 19225) (attribute 2961) (attributes 19226) (class_modifier 2963) (class_name 19227) (class_name_reference 19228) (name 19229) (new_variable 19230) (simple_variable 19231)) 8377 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 19217) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8378 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 19216) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8379 (("(" 19054) (argument_list 19215)) 8380 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 19214) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8381 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 19213) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8382 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 19208) (T_END_HEREDOC 19209) (T_VARIABLE 2920) (encaps_list 19210) (encaps_var 2922)) 8383 ((T_FN 35) (T_FUNCTION 131) (fn 18850) #'18851 (inline_function 19207)) 8384 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 19206) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8385 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 19205) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8386 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 19204) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8387 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 19201) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8388 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 19200) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8389 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 19196) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 8390 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 19194) (encaps_list 2828) (encaps_var 2829)) 8391 ((T_NULLSAFE_OBJECT_OPERATOR 19180) (T_OBJECT_OPERATOR 19181) ("[" 19182)) 8392 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 19177) (attribute 2757) (fn 18850) #'18851 (inline_function 19178)) 8393 (("(" 19054) (argument_list 19176)) 8394 ((T_PAAMAYIM_NEKUDOTAYIM 19168)) 8395 (("%" 18866) ("*" 18867) ("+" 18868) ("-" 18869) ("." 18870) ("/" 18871) ("<" 18872) (">" 18873) ("?" 18874) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 18875) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 18876) (T_BOOLEAN_AND 18877) (T_BOOLEAN_OR 18878) (T_COALESCE 18879) (T_DOUBLE_ARROW 19166) (T_INSTANCEOF 18880) (T_IS_EQUAL 18881) (T_IS_GREATER_OR_EQUAL 18882) (T_IS_IDENTICAL 18883) (T_IS_NOT_EQUAL 18884) (T_IS_NOT_IDENTICAL 18885) (T_IS_SMALLER_OR_EQUAL 18886) (T_LOGICAL_AND 18887) (T_LOGICAL_OR 18888) (T_LOGICAL_XOR 18889) (T_PIPE 18890) (T_POW 18891) (T_SL 18892) (T_SPACESHIP 18893) (T_SR 18894) ("^" 18895) ("|" 18896)) 8396 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 19155)) 8397 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 19143)) 8398 (("(" 19054) (argument_list 19142)) 8399 (("=" 19062) (T_AND_EQUAL 19063) (T_COALESCE_EQUAL 19064) (T_CONCAT_EQUAL 19065) (T_DEC 19066) (T_DIV_EQUAL 19067) (T_INC 19068) (T_MINUS_EQUAL 19069) (T_MOD_EQUAL 19070) (T_MUL_EQUAL 19071) (T_OR_EQUAL 19072) (T_PLUS_EQUAL 19073) (T_POW_EQUAL 19074) (T_SL_EQUAL 19075) (T_SR_EQUAL 19076) (T_XOR_EQUAL 19077)) 8400 ((T_PAAMAYIM_NEKUDOTAYIM 18966)) 8401 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 18865) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8402 (("%" 18866) ("*" 18867) ("+" 18868) ("-" 18869) ("." 18870) ("/" 18871) ("<" 18872) (">" 18873) ("?" 18874) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 18875) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 18876) (T_BOOLEAN_AND 18877) (T_BOOLEAN_OR 18878) (T_COALESCE 18879) (T_INSTANCEOF 18880) (T_IS_EQUAL 18881) (T_IS_GREATER_OR_EQUAL 18882) (T_IS_IDENTICAL 18883) (T_IS_NOT_EQUAL 18884) (T_IS_NOT_IDENTICAL 18885) (T_IS_SMALLER_OR_EQUAL 18886) (T_LOGICAL_AND 18887) (T_LOGICAL_OR 18888) (T_LOGICAL_XOR 18889) (T_PIPE 18890) (T_POW 18891) (T_SL 18892) (T_SPACESHIP 18893) (T_SR 18894) ("^" 18895) ("|" 18896)) 8403 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 18965) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8404 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 18964) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8405 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 18963) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8406 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 18962) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8407 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 18961) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8408 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 18960) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8409 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 18959) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8410 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 18958) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8411 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 18953) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 18954) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 8412 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 18952) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8413 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 18951) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8414 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 18950) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8415 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 18949) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8416 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 18948) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8417 (("$" 18913) ("(" 18914) (T_NAME_FULLY_QUALIFIED 18915) (T_NAME_QUALIFIED 18916) (T_NAME_RELATIVE 18917) (T_STATIC 18918) (T_STRING 18919) (T_VARIABLE 18920) (class_name 18921) (class_name_reference 18922) (name 18923) (new_variable 18924) (simple_variable 18925)) 8418 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 18912) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8419 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 18911) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8420 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 18910) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8421 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 18909) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8422 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 18908) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8423 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 18907) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8424 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 18906) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8425 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 18905) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8426 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 18904) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8427 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 18903) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8428 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 18902) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8429 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 18901) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8430 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 18900) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8431 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 18899) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8432 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 18898) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8433 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 18897) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8434 (("$" 18913) (T_VARIABLE 18920) (simple_variable 18944) ("{" 18945)) 8435 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 18942) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 8436 ((T_PAAMAYIM_NEKUDOTAYIM 18940)) 8437 ((T_NULLSAFE_OBJECT_OPERATOR 18926) (T_OBJECT_OPERATOR 18927) (T_PAAMAYIM_NEKUDOTAYIM 18928) ("[" 18929)) 8438 (("$" 18913) (T_STRING 18933) (T_VARIABLE 18920) (property_name 18939) (simple_variable 18935) ("{" 18936)) 8439 (("$" 18913) (T_STRING 18933) (T_VARIABLE 18920) (property_name 18934) (simple_variable 18935) ("{" 18936)) 8440 (("$" 18913) (T_VARIABLE 18920) (simple_variable 18932)) 8441 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 18930) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 8442 (("]" 18931)) 8443 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 18937) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 8444 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 18938)) 8445 (("$" 18913) (T_VARIABLE 18920) (simple_variable 18941)) 8446 (("%" 304) (")" 18943) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 8447 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 18946) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 8448 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 18947)) 8449 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 18957) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8450 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 18955) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 8451 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 18956) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8452 (("$" 18788) (T_ABSTRACT 18967) (T_ARRAY 18968) (T_AS 18969) (T_BREAK 18970) (T_CALLABLE 18971) (T_CASE 18972) (T_CATCH 18973) (T_CLASS 18974) (T_CLASS_C 18975) (T_CLONE 18976) (T_CONST 18977) (T_CONTINUE 18978) (T_DECLARE 18979) (T_DEFAULT 18980) (T_DIR 18981) (T_DO 18982) (T_ECHO 18983) (T_ELSE 18984) (T_ELSEIF 18985) (T_EMPTY 18986) (T_ENDDECLARE 18987) (T_ENDFOR 18988) (T_ENDFOREACH 18989) (T_ENDIF 18990) (T_ENDSWITCH 18991) (T_ENDWHILE 18992) (T_ENUM 18993) (T_EVAL 18994) (T_EXIT 18995) (T_EXTENDS 18996) (T_FILE 18997) (T_FINAL 18998) (T_FINALLY 18999) (T_FN 19000) (T_FOR 19001) (T_FOREACH 19002) (T_FUNCTION 19003) (T_FUNC_C 19004) (T_GLOBAL 19005) (T_GOTO 19006) (T_IF 19007) (T_IMPLEMENTS 19008) (T_INCLUDE 19009) (T_INCLUDE_ONCE 19010) (T_INSTANCEOF 19011) (T_INSTEADOF 19012) (T_INTERFACE 19013) (T_ISSET 19014) (T_LINE 19015) (T_LIST 19016) (T_LOGICAL_AND 19017) (T_LOGICAL_OR 19018) (T_LOGICAL_XOR 19019) (T_MATCH 19020) (T_METHOD_C 19021) (T_NAMESPACE 19022) (T_NEW 19023) (T_NS_C 19024) (T_PRINT 19025) (T_PRIVATE 19026) (T_PROPERTY_C 19027) (T_PROTECTED 19028) (T_PUBLIC 19029) (T_READONLY 19030) (T_REQUIRE 19031) (T_REQUIRE_ONCE 19032) (T_RETURN 19033) (T_STATIC 19034) (T_STRING 19035) (T_SWITCH 19036) (T_THROW 19037) (T_TRAIT 19038) (T_TRAIT_C 19039) (T_TRY 19040) (T_UNSET 19041) (T_USE 19042) (T_VAR 19043) (T_VARIABLE 18836) (T_WHILE 19044) (T_YIELD 19045) (identifier 19046) (member_name 19047) (reserved_non_modifiers 19048) (semi_reserved 19049) (simple_variable 19050) ("{" 19051)) 8453 (("(" 19054) (argument_list 19055)) 8454 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 19052) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 8455 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 19053)) 8456 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 19056) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 19057) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 19058) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 8457 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 19061) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 8458 (("," 1498) (possible_comma 19059)) 8459 ((")" 19060)) 8460 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (ampersand 19091) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 19092) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8461 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 19090) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8462 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 19089) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8463 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 19088) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8464 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 19087) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8465 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 19086) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8466 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 19085) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8467 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 19084) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8468 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 19083) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8469 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 19082) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8470 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 19081) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8471 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 19080) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8472 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 19079) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8473 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 19078) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8474 (("\"" 1538) ("$" 19093) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 19094) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 19095) ("[" 1560) (array_object_dereferenceable 19096) (callable_expr 19097) (callable_variable 19098) (class_constant 1564) (class_name 19099) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 19100) (name 19101) (new_dereferenceable 1570) (simple_variable 19102) (static_member 19103) (variable 19104) (variable_class_name 19105)) 8475 (("$" 19093) (T_VARIABLE 19095) (simple_variable 19138) ("{" 19139)) 8476 (("(" 19109) (argument_list 19137)) 8477 ((T_NULLSAFE_OBJECT_OPERATOR 19123) (T_OBJECT_OPERATOR 19124) ("[" 19125)) 8478 (("(" 19109) (argument_list 19122)) 8479 ((T_PAAMAYIM_NEKUDOTAYIM 19118)) 8480 (("(" 19109) (argument_list 19117)) 8481 ((T_PAAMAYIM_NEKUDOTAYIM 19106)) 8482 (("$" 19093) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 19095) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 19107) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 19108) ("{" 1660)) 8483 (("(" 19109) (argument_list 19110)) 8484 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 19111) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 19112) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 19113) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 8485 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 19116) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 8486 (("," 1498) (possible_comma 19114)) 8487 ((")" 19115)) 8488 (("$" 19093) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 19095) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 19119) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 19120) ("{" 1676)) 8489 (("(" 19109) (argument_list 19121)) 8490 (("$" 19093) (T_STRING 19128) (T_VARIABLE 19095) (property_name 19135) (simple_variable 19130) ("{" 19131)) 8491 (("$" 19093) (T_STRING 19128) (T_VARIABLE 19095) (property_name 19129) (simple_variable 19130) ("{" 19131)) 8492 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 19126) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 8493 (("]" 19127)) 8494 (("(" 19109) (argument_list 19134)) 8495 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 19132) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 8496 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 19133)) 8497 (("(" 19109) (argument_list 19136)) 8498 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 19140) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 8499 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 19141)) 8500 ((backup_doc_comment 19144)) 8501 (("(" 19145)) 8502 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 19146)) 8503 ((")" 19147)) 8504 ((T_USE 166) (lexical_vars 19148)) 8505 ((":" 168) (return_type 19149)) 8506 ((backup_fn_flags 19150)) 8507 (("{" 19151)) 8508 ((inner_statement_list 19152)) 8509 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 19153) ("~" 129)) 8510 ((backup_fn_flags 19154)) 8511 ((backup_doc_comment 19156)) 8512 (("(" 19157)) 8513 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 19158)) 8514 ((")" 19159)) 8515 ((":" 2271) (return_type 19160)) 8516 ((T_DOUBLE_ARROW 19161)) 8517 ((backup_fn_flags 19162)) 8518 ((backup_lex_pos 19163)) 8519 (("!" 3487) ("\"" 3488) ("$" 3489) ("(" 3490) ("+" 3491) ("-" 3492) ("@" 3493) (T_ARRAY 3494) (T_ARRAY_CAST 3495) (T_ATTRIBUTE 130) (T_BOOL_CAST 3496) (T_CLASS_C 3497) (T_CLONE 3498) (T_CONSTANT_ENCAPSED_STRING 3499) (T_DEC 3500) (T_DIR 3501) (T_DNUMBER 3502) (T_DOUBLE_CAST 3503) (T_EMPTY 3504) (T_EVAL 3505) (T_EXIT 3506) (T_FILE 3507) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3508) (T_INC 3509) (T_INCLUDE 3510) (T_INCLUDE_ONCE 3511) (T_INT_CAST 3512) (T_ISSET 3513) (T_LINE 3514) (T_LIST 3515) (T_LNUMBER 3516) (T_MATCH 3517) (T_METHOD_C 3518) (T_NAME_FULLY_QUALIFIED 3519) (T_NAME_QUALIFIED 3520) (T_NAME_RELATIVE 3521) (T_NEW 3522) (T_NS_C 3523) (T_OBJECT_CAST 3524) (T_PRINT 3525) (T_PROPERTY_C 3526) (T_READONLY 3527) (T_REQUIRE 3528) (T_REQUIRE_ONCE 3529) (T_START_HEREDOC 3530) (T_STATIC 3531) (T_STRING 3532) (T_STRING_CAST 3533) (T_THROW 3534) (T_TRAIT_C 3535) (T_UNSET_CAST 3536) (T_VARIABLE 3537) (T_YIELD 3538) (T_YIELD_FROM 3539) ("[" 3540) ("`" 3541) (array_object_dereferenceable 3542) (attribute 135) (attributes 3543) (callable_expr 3544) (callable_variable 3545) (class_constant 3546) (class_name 3547) (constant 3548) (dereferenceable_scalar 3549) (expr 19164) (fn 3551) (fully_dereferenceable 107) #'3552 (function_call 3553) (inline_function 3554) (internal_functions_in_yacc 3555) (match 3556) (name 3557) (new_dereferenceable 3558) (new_non_dereferenceable 3559) (scalar 3560) (simple_variable 3561) (static_member 3562) (variable 3563) (variable_class_name 3564) ("~" 3565)) 8520 (("%" 3567) ("*" 3568) ("+" 3569) ("-" 3570) ("." 3571) ("/" 3572) ("<" 3573) (">" 3574) ("?" 3575) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 3576) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3577) (T_BOOLEAN_AND 3578) (T_BOOLEAN_OR 3579) (T_COALESCE 3580) (T_INSTANCEOF 3581) (T_IS_EQUAL 3582) (T_IS_GREATER_OR_EQUAL 3583) (T_IS_IDENTICAL 3584) (T_IS_NOT_EQUAL 3585) (T_IS_NOT_IDENTICAL 3586) (T_IS_SMALLER_OR_EQUAL 3587) (T_LOGICAL_AND 3588) (T_LOGICAL_OR 3589) (T_LOGICAL_XOR 3590) (T_PIPE 3591) (T_POW 3592) (T_SL 3593) (T_SPACESHIP 3594) (T_SR 3595) ("^" 3596) (backup_fn_flags 19165) ("|" 3597)) 8521 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 19167) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8522 (("$" 18788) (T_ABSTRACT 18967) (T_ARRAY 18968) (T_AS 18969) (T_BREAK 18970) (T_CALLABLE 18971) (T_CASE 18972) (T_CATCH 18973) (T_CLASS 18974) (T_CLASS_C 18975) (T_CLONE 18976) (T_CONST 18977) (T_CONTINUE 18978) (T_DECLARE 18979) (T_DEFAULT 18980) (T_DIR 18981) (T_DO 18982) (T_ECHO 18983) (T_ELSE 18984) (T_ELSEIF 18985) (T_EMPTY 18986) (T_ENDDECLARE 18987) (T_ENDFOR 18988) (T_ENDFOREACH 18989) (T_ENDIF 18990) (T_ENDSWITCH 18991) (T_ENDWHILE 18992) (T_ENUM 18993) (T_EVAL 18994) (T_EXIT 18995) (T_EXTENDS 18996) (T_FILE 18997) (T_FINAL 18998) (T_FINALLY 18999) (T_FN 19000) (T_FOR 19001) (T_FOREACH 19002) (T_FUNCTION 19003) (T_FUNC_C 19004) (T_GLOBAL 19005) (T_GOTO 19006) (T_IF 19007) (T_IMPLEMENTS 19008) (T_INCLUDE 19009) (T_INCLUDE_ONCE 19010) (T_INSTANCEOF 19011) (T_INSTEADOF 19012) (T_INTERFACE 19013) (T_ISSET 19014) (T_LINE 19015) (T_LIST 19016) (T_LOGICAL_AND 19017) (T_LOGICAL_OR 19018) (T_LOGICAL_XOR 19019) (T_MATCH 19020) (T_METHOD_C 19021) (T_NAMESPACE 19022) (T_NEW 19023) (T_NS_C 19024) (T_PRINT 19025) (T_PRIVATE 19026) (T_PROPERTY_C 19027) (T_PROTECTED 19028) (T_PUBLIC 19029) (T_READONLY 19030) (T_REQUIRE 19031) (T_REQUIRE_ONCE 19032) (T_RETURN 19033) (T_STATIC 19034) (T_STRING 19035) (T_SWITCH 19036) (T_THROW 19037) (T_TRAIT 19038) (T_TRAIT_C 19039) (T_TRY 19040) (T_UNSET 19041) (T_USE 19042) (T_VAR 19043) (T_VARIABLE 18836) (T_WHILE 19044) (T_YIELD 19045) (identifier 19169) (member_name 19170) (reserved_non_modifiers 19048) (semi_reserved 19049) (simple_variable 19171) ("{" 19172)) 8523 (("(" 19054) (argument_list 19175)) 8524 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 19173) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 8525 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 19174)) 8526 ((T_FN 35) (T_FUNCTION 131) (fn 18850) #'18851 (inline_function 19179)) 8527 (("$" 18788) (T_STRING 19185) (T_VARIABLE 18836) (property_name 19192) (simple_variable 19187) ("{" 19188)) 8528 (("$" 18788) (T_STRING 19185) (T_VARIABLE 18836) (property_name 19186) (simple_variable 19187) ("{" 19188)) 8529 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 19183) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 8530 (("]" 19184)) 8531 (("(" 19054) (argument_list 19191)) 8532 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 19189) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 8533 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 19190)) 8534 (("(" 19054) (argument_list 19193)) 8535 (("`" 19195)) 8536 (("]" 19197)) 8537 (("=" 19198)) 8538 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 19199) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8539 (("%" 18866) ("*" 18867) ("+" 18868) ("-" 18869) ("." 18870) ("/" 18871) ("<" 18872) (">" 18873) ("?" 18874) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 18875) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 18876) (T_BOOLEAN_AND 18877) (T_BOOLEAN_OR 18878) (T_COALESCE 18879) (T_DOUBLE_ARROW 19202) (T_INSTANCEOF 18880) (T_IS_EQUAL 18881) (T_IS_GREATER_OR_EQUAL 18882) (T_IS_IDENTICAL 18883) (T_IS_NOT_EQUAL 18884) (T_IS_NOT_IDENTICAL 18885) (T_IS_SMALLER_OR_EQUAL 18886) (T_LOGICAL_AND 18887) (T_LOGICAL_OR 18888) (T_LOGICAL_XOR 18889) (T_PIPE 18890) (T_POW 18891) (T_SL 18892) (T_SPACESHIP 18893) (T_SR 18894) ("^" 18895) ("|" 18896)) 8540 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 19203) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8541 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 19212) (T_VARIABLE 2920) (encaps_var 2934)) 8542 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 19211) (T_VARIABLE 2920) (encaps_var 2925)) 8543 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 19253) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 8544 ((T_CLASS 19245)) 8545 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 19244) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 19225) (attribute 2982) (class_modifier 2963)) 8546 ((T_PAAMAYIM_NEKUDOTAYIM 19242)) 8547 (("(" 19109) (argument_list 19241)) 8548 ((T_NULLSAFE_OBJECT_OPERATOR 19232) (T_OBJECT_OPERATOR 19233) (T_PAAMAYIM_NEKUDOTAYIM 19234) ("[" 19235)) 8549 (("$" 19093) (T_STRING 19128) (T_VARIABLE 19095) (property_name 19240) (simple_variable 19130) ("{" 19131)) 8550 (("$" 19093) (T_STRING 19128) (T_VARIABLE 19095) (property_name 19239) (simple_variable 19130) ("{" 19131)) 8551 (("$" 19093) (T_VARIABLE 19095) (simple_variable 19238)) 8552 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 19236) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 8553 (("]" 19237)) 8554 (("$" 19093) (T_VARIABLE 19095) (simple_variable 19243)) 8555 (("(" 2984) (argument_list 2985) (ctor_arguments 19246)) 8556 ((T_EXTENDS 2987) (extends_from 19247)) 8557 ((T_IMPLEMENTS 2989) (implements_list 19248)) 8558 ((backup_doc_comment 19249)) 8559 (("{" 19250)) 8560 ((class_statement_list 19251)) 8561 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 19252)) 8562 (("%" 304) (")" 19254) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 8563 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 19256) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 8564 (("%" 304) (")" 19257) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 8565 (("{" 19258)) 8566 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 19259) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 8567 (("}" 19260)) 8568 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 19262) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 8569 ((")" 19263)) 8570 (("=" 19264)) 8571 (("!" 18786) ("\"" 18787) ("$" 18788) ("(" 18789) ("+" 18790) ("-" 18791) ("@" 18792) (T_ARRAY 18793) (T_ARRAY_CAST 18794) (T_ATTRIBUTE 130) (T_BOOL_CAST 18795) (T_CLASS_C 18796) (T_CLONE 18797) (T_CONSTANT_ENCAPSED_STRING 18798) (T_DEC 18799) (T_DIR 18800) (T_DNUMBER 18801) (T_DOUBLE_CAST 18802) (T_EMPTY 18803) (T_EVAL 18804) (T_EXIT 18805) (T_FILE 18806) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18807) (T_INC 18808) (T_INCLUDE 18809) (T_INCLUDE_ONCE 18810) (T_INT_CAST 18811) (T_ISSET 18812) (T_LINE 18813) (T_LIST 18814) (T_LNUMBER 18815) (T_MATCH 18816) (T_METHOD_C 18817) (T_NAME_FULLY_QUALIFIED 18818) (T_NAME_QUALIFIED 18819) (T_NAME_RELATIVE 18820) (T_NEW 18821) (T_NS_C 18822) (T_OBJECT_CAST 18823) (T_PRINT 18824) (T_PROPERTY_C 18825) (T_READONLY 18826) (T_REQUIRE 18827) (T_REQUIRE_ONCE 18828) (T_START_HEREDOC 18829) (T_STATIC 18830) (T_STRING 18831) (T_STRING_CAST 18832) (T_THROW 18833) (T_TRAIT_C 18834) (T_UNSET_CAST 18835) (T_VARIABLE 18836) (T_YIELD 18837) (T_YIELD_FROM 18838) ("[" 18839) ("`" 18840) (array_object_dereferenceable 18841) (attribute 135) (attributes 18842) (callable_expr 18843) (callable_variable 18844) (class_constant 18845) (class_name 18846) (constant 18847) (dereferenceable_scalar 18848) (expr 19265) (fn 18850) (fully_dereferenceable 107) #'18851 (function_call 18852) (inline_function 18853) (internal_functions_in_yacc 18854) (match 18855) (name 18856) (new_dereferenceable 18857) (new_non_dereferenceable 18858) (scalar 18859) (simple_variable 18860) (static_member 18861) (variable 18862) (variable_class_name 18863) ("~" 18864)) 8572 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 19267) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 8573 (("," 5226) (possible_comma 19268)) 8574 ((")" 19269)) 8575 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 19277) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 19278) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 19279) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 8576 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 19282) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 8577 (("," 1498) (possible_comma 19280)) 8578 ((")" 19281)) 8579 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 19284) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 8580 (("%" 304) (")" 19285) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 8581 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 19287) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 8582 (("%" 304) (")" 19288) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 8583 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 19294) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 19295) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 19296) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 19297) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 8584 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 19303) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 8585 (("%" 2442) (")" 19300) ("*" 2443) ("+" 2444) ("," 19301) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 8586 (("," 5287) (possible_comma 19298)) 8587 ((")" 19299)) 8588 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 19302) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 8589 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 19307) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 8590 ((")" 19308)) 8591 (("%" 304) (")" 19300) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 8592 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 19315) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 8593 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 19316)) 8594 (("\"" 19318) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 8595 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 19328) (T_VARIABLE 2920) (encaps_var 2934)) 8596 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 19327) (T_VARIABLE 2920) (encaps_var 2925)) 8597 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 19369) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 8598 ((T_CLASS 19361)) 8599 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 19360) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 19341) (attribute 2982) (class_modifier 2963)) 8600 ((T_PAAMAYIM_NEKUDOTAYIM 19358)) 8601 (("(" 18696) (argument_list 19357)) 8602 ((T_NULLSAFE_OBJECT_OPERATOR 19348) (T_OBJECT_OPERATOR 19349) (T_PAAMAYIM_NEKUDOTAYIM 19350) ("[" 19351)) 8603 (("$" 18680) (T_STRING 18715) (T_VARIABLE 18682) (property_name 19356) (simple_variable 18717) ("{" 18718)) 8604 (("$" 18680) (T_STRING 18715) (T_VARIABLE 18682) (property_name 19355) (simple_variable 18717) ("{" 18718)) 8605 (("$" 18680) (T_VARIABLE 18682) (simple_variable 19354)) 8606 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 19352) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 8607 (("]" 19353)) 8608 (("$" 18680) (T_VARIABLE 18682) (simple_variable 19359)) 8609 (("(" 2984) (argument_list 2985) (ctor_arguments 19362)) 8610 ((T_EXTENDS 2987) (extends_from 19363)) 8611 ((T_IMPLEMENTS 2989) (implements_list 19364)) 8612 ((backup_doc_comment 19365)) 8613 (("{" 19366)) 8614 ((class_statement_list 19367)) 8615 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 19368)) 8616 (("%" 304) (")" 19370) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 8617 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 19372) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 8618 (("%" 304) (")" 19373) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 8619 (("{" 19374)) 8620 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 19375) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 8621 (("}" 19376)) 8622 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 19378) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 8623 ((")" 19379)) 8624 (("=" 19380)) 8625 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 19381) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 8626 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 19383) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 8627 (("," 5226) (possible_comma 19384)) 8628 ((")" 19385)) 8629 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 19393) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 19394) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 19395) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 8630 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 19398) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 8631 (("," 1498) (possible_comma 19396)) 8632 ((")" 19397)) 8633 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 19400) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 8634 (("%" 304) (")" 19401) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 8635 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 19403) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 8636 (("%" 304) (")" 19404) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 8637 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 19410) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 19411) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 19412) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 19413) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 8638 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 19419) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 8639 (("%" 2442) (")" 19416) ("*" 2443) ("+" 2444) ("," 19417) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 8640 (("," 5287) (possible_comma 19414)) 8641 ((")" 19415)) 8642 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 19418) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 8643 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 19423) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 8644 ((")" 19424)) 8645 (("%" 304) (")" 19416) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 8646 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 19431) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 8647 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 19432)) 8648 (("\"" 19434) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 8649 ((";" 19438)) 8650 ((";" 19440)) 8651 (("$" 19441) (T_VARIABLE 19442) (simple_variable 19449) ("{" 19450)) 8652 (("," 19446) (";" 19447)) 8653 (("$" 19441) (T_VARIABLE 19442) (global_var 19448) (simple_variable 19445)) 8654 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 19451) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 8655 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 19452)) 8656 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19517) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8657 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 24794) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8658 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 24792) (encaps_var 5357)) 8659 (("$" 19456) (T_VARIABLE 19504) (simple_variable 24788) ("{" 24789)) 8660 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 24787) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 8661 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 24786) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8662 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 24785) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8663 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 24784) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8664 (("(" 24781)) 8665 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 24780) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8666 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 24779) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8667 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 24766) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (clone_argument_list 24767) (constant 19515) (dereferenceable_scalar 19516) (expr 24768) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8668 (("\"" 1538) ("$" 19761) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 19762) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 19763) ("[" 1560) (array_object_dereferenceable 19764) (callable_expr 19765) (callable_variable 19766) (class_constant 1564) (class_name 19767) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 19768) (name 19769) (new_dereferenceable 1570) (simple_variable 19770) (static_member 19771) (variable 24765) (variable_class_name 19773)) 8669 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 24764) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8670 (("(" 24761)) 8671 (("(" 24758)) 8672 (("(" 24749) (argument_list 24750) (ctor_arguments 24751)) 8673 (("\"" 1538) ("$" 19761) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 19762) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 19763) ("[" 1560) (array_object_dereferenceable 19764) (callable_expr 19765) (callable_variable 19766) (class_constant 1564) (class_name 19767) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 19768) (name 19769) (new_dereferenceable 1570) (simple_variable 19770) (static_member 19771) (variable 24748) (variable_class_name 19773)) 8674 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 24747) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8675 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 24746) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8676 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 24745) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8677 (("(" 24741)) 8678 (("(" 24736)) 8679 (("(" 24730)) 8680 (("$" 19761) ("(" 24693) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 24694) (T_NAME_QUALIFIED 24695) (T_NAME_RELATIVE 24696) (T_READONLY 2955) (T_STATIC 24697) (T_STRING 24698) (T_VARIABLE 19763) (anonymous_class 24699) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 24700) (attribute 2961) (attributes 24701) (class_modifier 2963) (class_name 24702) (class_name_reference 24703) (name 24704) (new_variable 24705) (simple_variable 24706)) 8681 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 24692) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8682 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 24691) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8683 (("(" 19722) (argument_list 24690)) 8684 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 24689) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8685 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 24688) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8686 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 24683) (T_END_HEREDOC 24684) (T_VARIABLE 2920) (encaps_list 24685) (encaps_var 2922)) 8687 ((T_FN 35) (T_FUNCTION 131) (fn 19518) #'19519 (inline_function 24682)) 8688 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 24681) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8689 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 24680) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8690 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 24679) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8691 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24208) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 8692 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 24144) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8693 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 24140) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 8694 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 24138) (encaps_list 2828) (encaps_var 2829)) 8695 ((T_NULLSAFE_OBJECT_OPERATOR 24124) (T_OBJECT_OPERATOR 24125) ("[" 24126)) 8696 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 24121) (attribute 2757) (fn 19518) #'19519 (inline_function 24122)) 8697 (("(" 19722) (argument_list 24120)) 8698 ((T_PAAMAYIM_NEKUDOTAYIM 24112)) 8699 (("%" 19534) ("*" 19535) ("+" 19536) ("-" 19537) ("." 19538) ("/" 19539) ("<" 19540) (">" 19541) ("?" 19542) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19543) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19544) (T_AS 20896) (T_BOOLEAN_AND 19545) (T_BOOLEAN_OR 19546) (T_COALESCE 19547) (T_INSTANCEOF 19548) (T_IS_EQUAL 19549) (T_IS_GREATER_OR_EQUAL 19550) (T_IS_IDENTICAL 19551) (T_IS_NOT_EQUAL 19552) (T_IS_NOT_IDENTICAL 19553) (T_IS_SMALLER_OR_EQUAL 19554) (T_LOGICAL_AND 19555) (T_LOGICAL_OR 19556) (T_LOGICAL_XOR 19557) (T_PIPE 19558) (T_POW 19559) (T_SL 19560) (T_SPACESHIP 19561) (T_SR 19562) ("^" 19563) ("|" 19564)) 8700 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 19823)) 8701 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 19811)) 8702 (("(" 19722) (argument_list 19810)) 8703 (("=" 19730) (T_AND_EQUAL 19731) (T_COALESCE_EQUAL 19732) (T_CONCAT_EQUAL 19733) (T_DEC 19734) (T_DIV_EQUAL 19735) (T_INC 19736) (T_MINUS_EQUAL 19737) (T_MOD_EQUAL 19738) (T_MUL_EQUAL 19739) (T_OR_EQUAL 19740) (T_PLUS_EQUAL 19741) (T_POW_EQUAL 19742) (T_SL_EQUAL 19743) (T_SR_EQUAL 19744) (T_XOR_EQUAL 19745)) 8704 ((T_PAAMAYIM_NEKUDOTAYIM 19634)) 8705 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19533) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8706 (("%" 19534) ("*" 19535) ("+" 19536) ("-" 19537) ("." 19538) ("/" 19539) ("<" 19540) (">" 19541) ("?" 19542) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19543) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19544) (T_BOOLEAN_AND 19545) (T_BOOLEAN_OR 19546) (T_COALESCE 19547) (T_INSTANCEOF 19548) (T_IS_EQUAL 19549) (T_IS_GREATER_OR_EQUAL 19550) (T_IS_IDENTICAL 19551) (T_IS_NOT_EQUAL 19552) (T_IS_NOT_IDENTICAL 19553) (T_IS_SMALLER_OR_EQUAL 19554) (T_LOGICAL_AND 19555) (T_LOGICAL_OR 19556) (T_LOGICAL_XOR 19557) (T_PIPE 19558) (T_POW 19559) (T_SL 19560) (T_SPACESHIP 19561) (T_SR 19562) ("^" 19563) ("|" 19564)) 8707 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19633) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8708 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19632) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8709 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19631) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8710 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19630) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8711 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19629) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8712 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19628) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8713 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19627) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8714 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19626) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8715 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 19621) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 19622) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 8716 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19620) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8717 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19619) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8718 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19618) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8719 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19617) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8720 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19616) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8721 (("$" 19581) ("(" 19582) (T_NAME_FULLY_QUALIFIED 19583) (T_NAME_QUALIFIED 19584) (T_NAME_RELATIVE 19585) (T_STATIC 19586) (T_STRING 19587) (T_VARIABLE 19588) (class_name 19589) (class_name_reference 19590) (name 19591) (new_variable 19592) (simple_variable 19593)) 8722 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19580) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8723 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19579) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8724 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19578) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8725 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19577) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8726 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19576) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8727 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19575) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8728 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19574) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8729 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19573) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8730 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19572) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8731 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19571) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8732 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19570) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8733 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19569) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8734 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19568) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8735 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19567) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8736 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19566) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8737 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19565) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8738 (("$" 19581) (T_VARIABLE 19588) (simple_variable 19612) ("{" 19613)) 8739 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 19610) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 8740 ((T_PAAMAYIM_NEKUDOTAYIM 19608)) 8741 ((T_NULLSAFE_OBJECT_OPERATOR 19594) (T_OBJECT_OPERATOR 19595) (T_PAAMAYIM_NEKUDOTAYIM 19596) ("[" 19597)) 8742 (("$" 19581) (T_STRING 19601) (T_VARIABLE 19588) (property_name 19607) (simple_variable 19603) ("{" 19604)) 8743 (("$" 19581) (T_STRING 19601) (T_VARIABLE 19588) (property_name 19602) (simple_variable 19603) ("{" 19604)) 8744 (("$" 19581) (T_VARIABLE 19588) (simple_variable 19600)) 8745 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 19598) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 8746 (("]" 19599)) 8747 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 19605) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 8748 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 19606)) 8749 (("$" 19581) (T_VARIABLE 19588) (simple_variable 19609)) 8750 (("%" 304) (")" 19611) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 8751 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 19614) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 8752 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 19615)) 8753 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19625) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8754 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 19623) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 8755 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19624) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8756 (("$" 19456) (T_ABSTRACT 19635) (T_ARRAY 19636) (T_AS 19637) (T_BREAK 19638) (T_CALLABLE 19639) (T_CASE 19640) (T_CATCH 19641) (T_CLASS 19642) (T_CLASS_C 19643) (T_CLONE 19644) (T_CONST 19645) (T_CONTINUE 19646) (T_DECLARE 19647) (T_DEFAULT 19648) (T_DIR 19649) (T_DO 19650) (T_ECHO 19651) (T_ELSE 19652) (T_ELSEIF 19653) (T_EMPTY 19654) (T_ENDDECLARE 19655) (T_ENDFOR 19656) (T_ENDFOREACH 19657) (T_ENDIF 19658) (T_ENDSWITCH 19659) (T_ENDWHILE 19660) (T_ENUM 19661) (T_EVAL 19662) (T_EXIT 19663) (T_EXTENDS 19664) (T_FILE 19665) (T_FINAL 19666) (T_FINALLY 19667) (T_FN 19668) (T_FOR 19669) (T_FOREACH 19670) (T_FUNCTION 19671) (T_FUNC_C 19672) (T_GLOBAL 19673) (T_GOTO 19674) (T_IF 19675) (T_IMPLEMENTS 19676) (T_INCLUDE 19677) (T_INCLUDE_ONCE 19678) (T_INSTANCEOF 19679) (T_INSTEADOF 19680) (T_INTERFACE 19681) (T_ISSET 19682) (T_LINE 19683) (T_LIST 19684) (T_LOGICAL_AND 19685) (T_LOGICAL_OR 19686) (T_LOGICAL_XOR 19687) (T_MATCH 19688) (T_METHOD_C 19689) (T_NAMESPACE 19690) (T_NEW 19691) (T_NS_C 19692) (T_PRINT 19693) (T_PRIVATE 19694) (T_PROPERTY_C 19695) (T_PROTECTED 19696) (T_PUBLIC 19697) (T_READONLY 19698) (T_REQUIRE 19699) (T_REQUIRE_ONCE 19700) (T_RETURN 19701) (T_STATIC 19702) (T_STRING 19703) (T_SWITCH 19704) (T_THROW 19705) (T_TRAIT 19706) (T_TRAIT_C 19707) (T_TRY 19708) (T_UNSET 19709) (T_USE 19710) (T_VAR 19711) (T_VARIABLE 19504) (T_WHILE 19712) (T_YIELD 19713) (identifier 19714) (member_name 19715) (reserved_non_modifiers 19716) (semi_reserved 19717) (simple_variable 19718) ("{" 19719)) 8757 (("(" 19722) (argument_list 19723)) 8758 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 19720) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 8759 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 19721)) 8760 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 19724) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 19725) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 19726) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 8761 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 19729) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 8762 (("," 1498) (possible_comma 19727)) 8763 ((")" 19728)) 8764 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (ampersand 19759) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19760) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8765 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19758) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8766 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19757) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8767 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19756) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8768 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19755) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8769 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19754) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8770 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19753) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8771 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19752) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8772 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19751) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8773 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19750) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8774 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19749) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8775 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19748) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8776 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19747) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8777 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 19746) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 8778 (("\"" 1538) ("$" 19761) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 19762) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 19763) ("[" 1560) (array_object_dereferenceable 19764) (callable_expr 19765) (callable_variable 19766) (class_constant 1564) (class_name 19767) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 19768) (name 19769) (new_dereferenceable 1570) (simple_variable 19770) (static_member 19771) (variable 19772) (variable_class_name 19773)) 8779 (("$" 19761) (T_VARIABLE 19763) (simple_variable 19806) ("{" 19807)) 8780 (("(" 19777) (argument_list 19805)) 8781 ((T_NULLSAFE_OBJECT_OPERATOR 19791) (T_OBJECT_OPERATOR 19792) ("[" 19793)) 8782 (("(" 19777) (argument_list 19790)) 8783 ((T_PAAMAYIM_NEKUDOTAYIM 19786)) 8784 (("(" 19777) (argument_list 19785)) 8785 ((T_PAAMAYIM_NEKUDOTAYIM 19774)) 8786 (("$" 19761) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 19763) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 19775) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 19776) ("{" 1660)) 8787 (("(" 19777) (argument_list 19778)) 8788 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 19779) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 19780) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 19781) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 8789 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 19784) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 8790 (("," 1498) (possible_comma 19782)) 8791 ((")" 19783)) 8792 (("$" 19761) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 19763) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 19787) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 19788) ("{" 1676)) 8793 (("(" 19777) (argument_list 19789)) 8794 (("$" 19761) (T_STRING 19796) (T_VARIABLE 19763) (property_name 19803) (simple_variable 19798) ("{" 19799)) 8795 (("$" 19761) (T_STRING 19796) (T_VARIABLE 19763) (property_name 19797) (simple_variable 19798) ("{" 19799)) 8796 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 19794) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 8797 (("]" 19795)) 8798 (("(" 19777) (argument_list 19802)) 8799 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 19800) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 8800 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 19801)) 8801 (("(" 19777) (argument_list 19804)) 8802 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 19808) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 8803 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 19809)) 8804 ((backup_doc_comment 19812)) 8805 (("(" 19813)) 8806 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 19814)) 8807 ((")" 19815)) 8808 ((T_USE 166) (lexical_vars 19816)) 8809 ((":" 168) (return_type 19817)) 8810 ((backup_fn_flags 19818)) 8811 (("{" 19819)) 8812 ((inner_statement_list 19820)) 8813 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 19821) ("~" 129)) 8814 ((backup_fn_flags 19822)) 8815 ((backup_doc_comment 19824)) 8816 (("(" 19825)) 8817 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 19826)) 8818 ((")" 19827)) 8819 ((":" 2271) (return_type 19828)) 8820 ((T_DOUBLE_ARROW 19829)) 8821 ((backup_fn_flags 19830)) 8822 ((backup_lex_pos 19831)) 8823 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 19895) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8824 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20895) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8825 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 20893) (encaps_var 5357)) 8826 (("$" 19834) (T_VARIABLE 19882) (simple_variable 20889) ("{" 20890)) 8827 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 20888) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 8828 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20887) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8829 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20886) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8830 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20885) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8831 (("(" 20882)) 8832 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20881) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8833 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20880) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8834 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 20867) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (clone_argument_list 20868) (constant 19893) (dereferenceable_scalar 19894) (expr 20869) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8835 (("\"" 1538) ("$" 20139) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20140) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20141) ("[" 1560) (array_object_dereferenceable 20142) (callable_expr 20143) (callable_variable 20144) (class_constant 1564) (class_name 20145) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 20146) (name 20147) (new_dereferenceable 1570) (simple_variable 20148) (static_member 20149) (variable 20866) (variable_class_name 20151)) 8836 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20865) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8837 (("(" 20862)) 8838 (("(" 20859)) 8839 (("(" 20850) (argument_list 20851) (ctor_arguments 20852)) 8840 (("\"" 1538) ("$" 20139) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20140) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20141) ("[" 1560) (array_object_dereferenceable 20142) (callable_expr 20143) (callable_variable 20144) (class_constant 1564) (class_name 20145) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 20146) (name 20147) (new_dereferenceable 1570) (simple_variable 20148) (static_member 20149) (variable 20849) (variable_class_name 20151)) 8841 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20848) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8842 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20847) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8843 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20846) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8844 (("(" 20842)) 8845 (("(" 20837)) 8846 (("(" 20831)) 8847 (("$" 20139) ("(" 20794) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 20795) (T_NAME_QUALIFIED 20796) (T_NAME_RELATIVE 20797) (T_READONLY 2955) (T_STATIC 20798) (T_STRING 20799) (T_VARIABLE 20141) (anonymous_class 20800) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 20801) (attribute 2961) (attributes 20802) (class_modifier 2963) (class_name 20803) (class_name_reference 20804) (name 20805) (new_variable 20806) (simple_variable 20807)) 8848 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20793) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8849 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20792) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8850 (("(" 20100) (argument_list 20791)) 8851 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20790) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8852 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20789) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8853 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 20784) (T_END_HEREDOC 20785) (T_VARIABLE 2920) (encaps_list 20786) (encaps_var 2922)) 8854 ((T_FN 35) (T_FUNCTION 131) (fn 19896) #'19897 (inline_function 20783)) 8855 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20782) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8856 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20781) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8857 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20780) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8858 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20309) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 8859 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20245) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8860 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 20241) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 8861 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 20239) (encaps_list 2828) (encaps_var 2829)) 8862 ((T_NULLSAFE_OBJECT_OPERATOR 20225) (T_OBJECT_OPERATOR 20226) ("[" 20227)) 8863 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 20222) (attribute 2757) (fn 19896) #'19897 (inline_function 20223)) 8864 (("(" 20100) (argument_list 20221)) 8865 ((T_PAAMAYIM_NEKUDOTAYIM 20213)) 8866 (("%" 19912) ("*" 19913) ("+" 19914) ("-" 19915) ("." 19916) ("/" 19917) ("<" 19918) (">" 19919) ("?" 19920) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19921) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19922) (T_BOOLEAN_AND 19923) (T_BOOLEAN_OR 19924) (T_COALESCE 19925) (T_INSTANCEOF 19926) (T_IS_EQUAL 19927) (T_IS_GREATER_OR_EQUAL 19928) (T_IS_IDENTICAL 19929) (T_IS_NOT_EQUAL 19930) (T_IS_NOT_IDENTICAL 19931) (T_IS_SMALLER_OR_EQUAL 19932) (T_LOGICAL_AND 19933) (T_LOGICAL_OR 19934) (T_LOGICAL_XOR 19935) (T_PIPE 19936) (T_POW 19937) (T_SL 19938) (T_SPACESHIP 19939) (T_SR 19940) ("^" 19941) (backup_fn_flags 20212) ("|" 19942)) 8867 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 20201)) 8868 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 20189)) 8869 (("(" 20100) (argument_list 20188)) 8870 (("=" 20108) (T_AND_EQUAL 20109) (T_COALESCE_EQUAL 20110) (T_CONCAT_EQUAL 20111) (T_DEC 20112) (T_DIV_EQUAL 20113) (T_INC 20114) (T_MINUS_EQUAL 20115) (T_MOD_EQUAL 20116) (T_MUL_EQUAL 20117) (T_OR_EQUAL 20118) (T_PLUS_EQUAL 20119) (T_POW_EQUAL 20120) (T_SL_EQUAL 20121) (T_SR_EQUAL 20122) (T_XOR_EQUAL 20123)) 8871 ((T_PAAMAYIM_NEKUDOTAYIM 20012)) 8872 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 19911) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8873 (("%" 19912) ("*" 19913) ("+" 19914) ("-" 19915) ("." 19916) ("/" 19917) ("<" 19918) (">" 19919) ("?" 19920) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19921) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19922) (T_BOOLEAN_AND 19923) (T_BOOLEAN_OR 19924) (T_COALESCE 19925) (T_INSTANCEOF 19926) (T_IS_EQUAL 19927) (T_IS_GREATER_OR_EQUAL 19928) (T_IS_IDENTICAL 19929) (T_IS_NOT_EQUAL 19930) (T_IS_NOT_IDENTICAL 19931) (T_IS_SMALLER_OR_EQUAL 19932) (T_LOGICAL_AND 19933) (T_LOGICAL_OR 19934) (T_LOGICAL_XOR 19935) (T_PIPE 19936) (T_POW 19937) (T_SL 19938) (T_SPACESHIP 19939) (T_SR 19940) ("^" 19941) ("|" 19942)) 8874 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20011) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8875 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20010) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8876 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20009) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8877 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20008) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8878 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20007) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8879 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20006) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8880 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20005) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8881 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20004) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8882 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 19999) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 20000) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 8883 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 19998) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8884 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 19997) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8885 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 19996) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8886 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 19995) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8887 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 19994) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8888 (("$" 19959) ("(" 19960) (T_NAME_FULLY_QUALIFIED 19961) (T_NAME_QUALIFIED 19962) (T_NAME_RELATIVE 19963) (T_STATIC 19964) (T_STRING 19965) (T_VARIABLE 19966) (class_name 19967) (class_name_reference 19968) (name 19969) (new_variable 19970) (simple_variable 19971)) 8889 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 19958) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8890 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 19957) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8891 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 19956) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8892 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 19955) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8893 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 19954) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8894 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 19953) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8895 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 19952) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8896 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 19951) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8897 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 19950) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8898 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 19949) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8899 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 19948) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8900 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 19947) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8901 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 19946) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8902 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 19945) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8903 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 19944) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8904 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 19943) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8905 (("$" 19959) (T_VARIABLE 19966) (simple_variable 19990) ("{" 19991)) 8906 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 19988) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 8907 ((T_PAAMAYIM_NEKUDOTAYIM 19986)) 8908 ((T_NULLSAFE_OBJECT_OPERATOR 19972) (T_OBJECT_OPERATOR 19973) (T_PAAMAYIM_NEKUDOTAYIM 19974) ("[" 19975)) 8909 (("$" 19959) (T_STRING 19979) (T_VARIABLE 19966) (property_name 19985) (simple_variable 19981) ("{" 19982)) 8910 (("$" 19959) (T_STRING 19979) (T_VARIABLE 19966) (property_name 19980) (simple_variable 19981) ("{" 19982)) 8911 (("$" 19959) (T_VARIABLE 19966) (simple_variable 19978)) 8912 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 19976) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 8913 (("]" 19977)) 8914 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 19983) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 8915 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 19984)) 8916 (("$" 19959) (T_VARIABLE 19966) (simple_variable 19987)) 8917 (("%" 304) (")" 19989) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 8918 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 19992) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 8919 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 19993)) 8920 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20003) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8921 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 20001) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 8922 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20002) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8923 (("$" 19834) (T_ABSTRACT 20013) (T_ARRAY 20014) (T_AS 20015) (T_BREAK 20016) (T_CALLABLE 20017) (T_CASE 20018) (T_CATCH 20019) (T_CLASS 20020) (T_CLASS_C 20021) (T_CLONE 20022) (T_CONST 20023) (T_CONTINUE 20024) (T_DECLARE 20025) (T_DEFAULT 20026) (T_DIR 20027) (T_DO 20028) (T_ECHO 20029) (T_ELSE 20030) (T_ELSEIF 20031) (T_EMPTY 20032) (T_ENDDECLARE 20033) (T_ENDFOR 20034) (T_ENDFOREACH 20035) (T_ENDIF 20036) (T_ENDSWITCH 20037) (T_ENDWHILE 20038) (T_ENUM 20039) (T_EVAL 20040) (T_EXIT 20041) (T_EXTENDS 20042) (T_FILE 20043) (T_FINAL 20044) (T_FINALLY 20045) (T_FN 20046) (T_FOR 20047) (T_FOREACH 20048) (T_FUNCTION 20049) (T_FUNC_C 20050) (T_GLOBAL 20051) (T_GOTO 20052) (T_IF 20053) (T_IMPLEMENTS 20054) (T_INCLUDE 20055) (T_INCLUDE_ONCE 20056) (T_INSTANCEOF 20057) (T_INSTEADOF 20058) (T_INTERFACE 20059) (T_ISSET 20060) (T_LINE 20061) (T_LIST 20062) (T_LOGICAL_AND 20063) (T_LOGICAL_OR 20064) (T_LOGICAL_XOR 20065) (T_MATCH 20066) (T_METHOD_C 20067) (T_NAMESPACE 20068) (T_NEW 20069) (T_NS_C 20070) (T_PRINT 20071) (T_PRIVATE 20072) (T_PROPERTY_C 20073) (T_PROTECTED 20074) (T_PUBLIC 20075) (T_READONLY 20076) (T_REQUIRE 20077) (T_REQUIRE_ONCE 20078) (T_RETURN 20079) (T_STATIC 20080) (T_STRING 20081) (T_SWITCH 20082) (T_THROW 20083) (T_TRAIT 20084) (T_TRAIT_C 20085) (T_TRY 20086) (T_UNSET 20087) (T_USE 20088) (T_VAR 20089) (T_VARIABLE 19882) (T_WHILE 20090) (T_YIELD 20091) (identifier 20092) (member_name 20093) (reserved_non_modifiers 20094) (semi_reserved 20095) (simple_variable 20096) ("{" 20097)) 8924 (("(" 20100) (argument_list 20101)) 8925 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 20098) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 8926 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 20099)) 8927 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 20102) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 20103) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 20104) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 8928 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 20107) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 8929 (("," 1498) (possible_comma 20105)) 8930 ((")" 20106)) 8931 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (ampersand 20137) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20138) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8932 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20136) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8933 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20135) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8934 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20134) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8935 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20133) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8936 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20132) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8937 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20131) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8938 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20130) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8939 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20129) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8940 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20128) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8941 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20127) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8942 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20126) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8943 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20125) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8944 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20124) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8945 (("\"" 1538) ("$" 20139) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20140) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20141) ("[" 1560) (array_object_dereferenceable 20142) (callable_expr 20143) (callable_variable 20144) (class_constant 1564) (class_name 20145) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 20146) (name 20147) (new_dereferenceable 1570) (simple_variable 20148) (static_member 20149) (variable 20150) (variable_class_name 20151)) 8946 (("$" 20139) (T_VARIABLE 20141) (simple_variable 20184) ("{" 20185)) 8947 (("(" 20155) (argument_list 20183)) 8948 ((T_NULLSAFE_OBJECT_OPERATOR 20169) (T_OBJECT_OPERATOR 20170) ("[" 20171)) 8949 (("(" 20155) (argument_list 20168)) 8950 ((T_PAAMAYIM_NEKUDOTAYIM 20164)) 8951 (("(" 20155) (argument_list 20163)) 8952 ((T_PAAMAYIM_NEKUDOTAYIM 20152)) 8953 (("$" 20139) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 20141) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 20153) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 20154) ("{" 1660)) 8954 (("(" 20155) (argument_list 20156)) 8955 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 20157) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 20158) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 20159) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 8956 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 20162) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 8957 (("," 1498) (possible_comma 20160)) 8958 ((")" 20161)) 8959 (("$" 20139) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 20141) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 20165) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 20166) ("{" 1676)) 8960 (("(" 20155) (argument_list 20167)) 8961 (("$" 20139) (T_STRING 20174) (T_VARIABLE 20141) (property_name 20181) (simple_variable 20176) ("{" 20177)) 8962 (("$" 20139) (T_STRING 20174) (T_VARIABLE 20141) (property_name 20175) (simple_variable 20176) ("{" 20177)) 8963 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 20172) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 8964 (("]" 20173)) 8965 (("(" 20155) (argument_list 20180)) 8966 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 20178) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 8967 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 20179)) 8968 (("(" 20155) (argument_list 20182)) 8969 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 20186) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 8970 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 20187)) 8971 ((backup_doc_comment 20190)) 8972 (("(" 20191)) 8973 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 20192)) 8974 ((")" 20193)) 8975 ((T_USE 166) (lexical_vars 20194)) 8976 ((":" 168) (return_type 20195)) 8977 ((backup_fn_flags 20196)) 8978 (("{" 20197)) 8979 ((inner_statement_list 20198)) 8980 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 20199) ("~" 129)) 8981 ((backup_fn_flags 20200)) 8982 ((backup_doc_comment 20202)) 8983 (("(" 20203)) 8984 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 20204)) 8985 ((")" 20205)) 8986 ((":" 2271) (return_type 20206)) 8987 ((T_DOUBLE_ARROW 20207)) 8988 ((backup_fn_flags 20208)) 8989 ((backup_lex_pos 20209)) 8990 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20210) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 8991 (("%" 19912) ("*" 19913) ("+" 19914) ("-" 19915) ("." 19916) ("/" 19917) ("<" 19918) (">" 19919) ("?" 19920) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19921) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19922) (T_BOOLEAN_AND 19923) (T_BOOLEAN_OR 19924) (T_COALESCE 19925) (T_INSTANCEOF 19926) (T_IS_EQUAL 19927) (T_IS_GREATER_OR_EQUAL 19928) (T_IS_IDENTICAL 19929) (T_IS_NOT_EQUAL 19930) (T_IS_NOT_IDENTICAL 19931) (T_IS_SMALLER_OR_EQUAL 19932) (T_LOGICAL_AND 19933) (T_LOGICAL_OR 19934) (T_LOGICAL_XOR 19935) (T_PIPE 19936) (T_POW 19937) (T_SL 19938) (T_SPACESHIP 19939) (T_SR 19940) ("^" 19941) (backup_fn_flags 20211) ("|" 19942)) 8992 (("$" 19834) (T_ABSTRACT 20013) (T_ARRAY 20014) (T_AS 20015) (T_BREAK 20016) (T_CALLABLE 20017) (T_CASE 20018) (T_CATCH 20019) (T_CLASS 20020) (T_CLASS_C 20021) (T_CLONE 20022) (T_CONST 20023) (T_CONTINUE 20024) (T_DECLARE 20025) (T_DEFAULT 20026) (T_DIR 20027) (T_DO 20028) (T_ECHO 20029) (T_ELSE 20030) (T_ELSEIF 20031) (T_EMPTY 20032) (T_ENDDECLARE 20033) (T_ENDFOR 20034) (T_ENDFOREACH 20035) (T_ENDIF 20036) (T_ENDSWITCH 20037) (T_ENDWHILE 20038) (T_ENUM 20039) (T_EVAL 20040) (T_EXIT 20041) (T_EXTENDS 20042) (T_FILE 20043) (T_FINAL 20044) (T_FINALLY 20045) (T_FN 20046) (T_FOR 20047) (T_FOREACH 20048) (T_FUNCTION 20049) (T_FUNC_C 20050) (T_GLOBAL 20051) (T_GOTO 20052) (T_IF 20053) (T_IMPLEMENTS 20054) (T_INCLUDE 20055) (T_INCLUDE_ONCE 20056) (T_INSTANCEOF 20057) (T_INSTEADOF 20058) (T_INTERFACE 20059) (T_ISSET 20060) (T_LINE 20061) (T_LIST 20062) (T_LOGICAL_AND 20063) (T_LOGICAL_OR 20064) (T_LOGICAL_XOR 20065) (T_MATCH 20066) (T_METHOD_C 20067) (T_NAMESPACE 20068) (T_NEW 20069) (T_NS_C 20070) (T_PRINT 20071) (T_PRIVATE 20072) (T_PROPERTY_C 20073) (T_PROTECTED 20074) (T_PUBLIC 20075) (T_READONLY 20076) (T_REQUIRE 20077) (T_REQUIRE_ONCE 20078) (T_RETURN 20079) (T_STATIC 20080) (T_STRING 20081) (T_SWITCH 20082) (T_THROW 20083) (T_TRAIT 20084) (T_TRAIT_C 20085) (T_TRY 20086) (T_UNSET 20087) (T_USE 20088) (T_VAR 20089) (T_VARIABLE 19882) (T_WHILE 20090) (T_YIELD 20091) (identifier 20214) (member_name 20215) (reserved_non_modifiers 20094) (semi_reserved 20095) (simple_variable 20216) ("{" 20217)) 8993 (("(" 20100) (argument_list 20220)) 8994 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 20218) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 8995 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 20219)) 8996 ((T_FN 35) (T_FUNCTION 131) (fn 19896) #'19897 (inline_function 20224)) 8997 (("$" 19834) (T_STRING 20230) (T_VARIABLE 19882) (property_name 20237) (simple_variable 20232) ("{" 20233)) 8998 (("$" 19834) (T_STRING 20230) (T_VARIABLE 19882) (property_name 20231) (simple_variable 20232) ("{" 20233)) 8999 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 20228) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 9000 (("]" 20229)) 9001 (("(" 20100) (argument_list 20236)) 9002 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 20234) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 9003 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 20235)) 9004 (("(" 20100) (argument_list 20238)) 9005 (("`" 20240)) 9006 (("]" 20242)) 9007 (("=" 20243)) 9008 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20244) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 9009 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20779) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9010 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 20777) (encaps_var 5357)) 9011 (("$" 20248) (T_VARIABLE 20296) (simple_variable 20773) ("{" 20774)) 9012 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 20772) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 9013 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20771) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9014 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20770) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9015 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20769) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9016 (("(" 20766)) 9017 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20765) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9018 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20764) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9019 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20751) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (clone_argument_list 20752) (constant 20307) (dereferenceable_scalar 20308) (expr 20753) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9020 (("\"" 1538) ("$" 20553) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20554) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20555) ("[" 1560) (array_object_dereferenceable 20556) (callable_expr 20557) (callable_variable 20558) (class_constant 1564) (class_name 20559) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 20560) (name 20561) (new_dereferenceable 1570) (simple_variable 20562) (static_member 20563) (variable 20750) (variable_class_name 20565)) 9021 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20749) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9022 (("(" 20746)) 9023 (("(" 20743)) 9024 (("(" 20734) (argument_list 20735) (ctor_arguments 20736)) 9025 (("\"" 1538) ("$" 20553) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20554) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20555) ("[" 1560) (array_object_dereferenceable 20556) (callable_expr 20557) (callable_variable 20558) (class_constant 1564) (class_name 20559) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 20560) (name 20561) (new_dereferenceable 1570) (simple_variable 20562) (static_member 20563) (variable 20733) (variable_class_name 20565)) 9026 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20732) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9027 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20731) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9028 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20730) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9029 (("(" 20726)) 9030 (("(" 20721)) 9031 (("(" 20715)) 9032 (("$" 20553) ("(" 20678) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 20679) (T_NAME_QUALIFIED 20680) (T_NAME_RELATIVE 20681) (T_READONLY 2955) (T_STATIC 20682) (T_STRING 20683) (T_VARIABLE 20555) (anonymous_class 20684) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 20685) (attribute 2961) (attributes 20686) (class_modifier 2963) (class_name 20687) (class_name_reference 20688) (name 20689) (new_variable 20690) (simple_variable 20691)) 9033 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20677) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9034 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20676) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9035 (("(" 20514) (argument_list 20675)) 9036 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20674) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9037 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20673) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9038 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 20668) (T_END_HEREDOC 20669) (T_VARIABLE 2920) (encaps_list 20670) (encaps_var 2922)) 9039 ((T_FN 35) (T_FUNCTION 131) (fn 20310) #'20311 (inline_function 20667)) 9040 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20666) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9041 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20665) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9042 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20664) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9043 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20661) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9044 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20660) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9045 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 20656) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 9046 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 20654) (encaps_list 2828) (encaps_var 2829)) 9047 ((T_NULLSAFE_OBJECT_OPERATOR 20640) (T_OBJECT_OPERATOR 20641) ("[" 20642)) 9048 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 20637) (attribute 2757) (fn 20310) #'20311 (inline_function 20638)) 9049 (("(" 20514) (argument_list 20636)) 9050 ((T_PAAMAYIM_NEKUDOTAYIM 20628)) 9051 (("%" 20326) ("*" 20327) ("+" 20328) ("-" 20329) ("." 20330) ("/" 20331) ("<" 20332) (">" 20333) ("?" 20334) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 20335) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 20336) (T_BOOLEAN_AND 20337) (T_BOOLEAN_OR 20338) (T_COALESCE 20339) (T_DOUBLE_ARROW 20626) (T_INSTANCEOF 20340) (T_IS_EQUAL 20341) (T_IS_GREATER_OR_EQUAL 20342) (T_IS_IDENTICAL 20343) (T_IS_NOT_EQUAL 20344) (T_IS_NOT_IDENTICAL 20345) (T_IS_SMALLER_OR_EQUAL 20346) (T_LOGICAL_AND 20347) (T_LOGICAL_OR 20348) (T_LOGICAL_XOR 20349) (T_PIPE 20350) (T_POW 20351) (T_SL 20352) (T_SPACESHIP 20353) (T_SR 20354) ("^" 20355) ("|" 20356)) 9052 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 20615)) 9053 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 20603)) 9054 (("(" 20514) (argument_list 20602)) 9055 (("=" 20522) (T_AND_EQUAL 20523) (T_COALESCE_EQUAL 20524) (T_CONCAT_EQUAL 20525) (T_DEC 20526) (T_DIV_EQUAL 20527) (T_INC 20528) (T_MINUS_EQUAL 20529) (T_MOD_EQUAL 20530) (T_MUL_EQUAL 20531) (T_OR_EQUAL 20532) (T_PLUS_EQUAL 20533) (T_POW_EQUAL 20534) (T_SL_EQUAL 20535) (T_SR_EQUAL 20536) (T_XOR_EQUAL 20537)) 9056 ((T_PAAMAYIM_NEKUDOTAYIM 20426)) 9057 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20325) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9058 (("%" 20326) ("*" 20327) ("+" 20328) ("-" 20329) ("." 20330) ("/" 20331) ("<" 20332) (">" 20333) ("?" 20334) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 20335) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 20336) (T_BOOLEAN_AND 20337) (T_BOOLEAN_OR 20338) (T_COALESCE 20339) (T_INSTANCEOF 20340) (T_IS_EQUAL 20341) (T_IS_GREATER_OR_EQUAL 20342) (T_IS_IDENTICAL 20343) (T_IS_NOT_EQUAL 20344) (T_IS_NOT_IDENTICAL 20345) (T_IS_SMALLER_OR_EQUAL 20346) (T_LOGICAL_AND 20347) (T_LOGICAL_OR 20348) (T_LOGICAL_XOR 20349) (T_PIPE 20350) (T_POW 20351) (T_SL 20352) (T_SPACESHIP 20353) (T_SR 20354) ("^" 20355) ("|" 20356)) 9059 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20425) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9060 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20424) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9061 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20423) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9062 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20422) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9063 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20421) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9064 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20420) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9065 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20419) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9066 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20418) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9067 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 20413) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 20414) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 9068 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20412) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9069 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20411) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9070 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20410) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9071 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20409) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9072 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20408) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9073 (("$" 20373) ("(" 20374) (T_NAME_FULLY_QUALIFIED 20375) (T_NAME_QUALIFIED 20376) (T_NAME_RELATIVE 20377) (T_STATIC 20378) (T_STRING 20379) (T_VARIABLE 20380) (class_name 20381) (class_name_reference 20382) (name 20383) (new_variable 20384) (simple_variable 20385)) 9074 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20372) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9075 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20371) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9076 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20370) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9077 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20369) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9078 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20368) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9079 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20367) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9080 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20366) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9081 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20365) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9082 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20364) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9083 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20363) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9084 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20362) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9085 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20361) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9086 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20360) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9087 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20359) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9088 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20358) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9089 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20357) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9090 (("$" 20373) (T_VARIABLE 20380) (simple_variable 20404) ("{" 20405)) 9091 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 20402) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 9092 ((T_PAAMAYIM_NEKUDOTAYIM 20400)) 9093 ((T_NULLSAFE_OBJECT_OPERATOR 20386) (T_OBJECT_OPERATOR 20387) (T_PAAMAYIM_NEKUDOTAYIM 20388) ("[" 20389)) 9094 (("$" 20373) (T_STRING 20393) (T_VARIABLE 20380) (property_name 20399) (simple_variable 20395) ("{" 20396)) 9095 (("$" 20373) (T_STRING 20393) (T_VARIABLE 20380) (property_name 20394) (simple_variable 20395) ("{" 20396)) 9096 (("$" 20373) (T_VARIABLE 20380) (simple_variable 20392)) 9097 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 20390) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 9098 (("]" 20391)) 9099 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 20397) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 9100 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 20398)) 9101 (("$" 20373) (T_VARIABLE 20380) (simple_variable 20401)) 9102 (("%" 304) (")" 20403) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 9103 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 20406) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 9104 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 20407)) 9105 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20417) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9106 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 20415) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 9107 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20416) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9108 (("$" 20248) (T_ABSTRACT 20427) (T_ARRAY 20428) (T_AS 20429) (T_BREAK 20430) (T_CALLABLE 20431) (T_CASE 20432) (T_CATCH 20433) (T_CLASS 20434) (T_CLASS_C 20435) (T_CLONE 20436) (T_CONST 20437) (T_CONTINUE 20438) (T_DECLARE 20439) (T_DEFAULT 20440) (T_DIR 20441) (T_DO 20442) (T_ECHO 20443) (T_ELSE 20444) (T_ELSEIF 20445) (T_EMPTY 20446) (T_ENDDECLARE 20447) (T_ENDFOR 20448) (T_ENDFOREACH 20449) (T_ENDIF 20450) (T_ENDSWITCH 20451) (T_ENDWHILE 20452) (T_ENUM 20453) (T_EVAL 20454) (T_EXIT 20455) (T_EXTENDS 20456) (T_FILE 20457) (T_FINAL 20458) (T_FINALLY 20459) (T_FN 20460) (T_FOR 20461) (T_FOREACH 20462) (T_FUNCTION 20463) (T_FUNC_C 20464) (T_GLOBAL 20465) (T_GOTO 20466) (T_IF 20467) (T_IMPLEMENTS 20468) (T_INCLUDE 20469) (T_INCLUDE_ONCE 20470) (T_INSTANCEOF 20471) (T_INSTEADOF 20472) (T_INTERFACE 20473) (T_ISSET 20474) (T_LINE 20475) (T_LIST 20476) (T_LOGICAL_AND 20477) (T_LOGICAL_OR 20478) (T_LOGICAL_XOR 20479) (T_MATCH 20480) (T_METHOD_C 20481) (T_NAMESPACE 20482) (T_NEW 20483) (T_NS_C 20484) (T_PRINT 20485) (T_PRIVATE 20486) (T_PROPERTY_C 20487) (T_PROTECTED 20488) (T_PUBLIC 20489) (T_READONLY 20490) (T_REQUIRE 20491) (T_REQUIRE_ONCE 20492) (T_RETURN 20493) (T_STATIC 20494) (T_STRING 20495) (T_SWITCH 20496) (T_THROW 20497) (T_TRAIT 20498) (T_TRAIT_C 20499) (T_TRY 20500) (T_UNSET 20501) (T_USE 20502) (T_VAR 20503) (T_VARIABLE 20296) (T_WHILE 20504) (T_YIELD 20505) (identifier 20506) (member_name 20507) (reserved_non_modifiers 20508) (semi_reserved 20509) (simple_variable 20510) ("{" 20511)) 9109 (("(" 20514) (argument_list 20515)) 9110 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 20512) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 9111 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 20513)) 9112 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 20516) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 20517) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 20518) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 9113 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 20521) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 9114 (("," 1498) (possible_comma 20519)) 9115 ((")" 20520)) 9116 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (ampersand 20551) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20552) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9117 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20550) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9118 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20549) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9119 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20548) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9120 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20547) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9121 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20546) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9122 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20545) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9123 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20544) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9124 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20543) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9125 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20542) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9126 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20541) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9127 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20540) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9128 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20539) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9129 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20538) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9130 (("\"" 1538) ("$" 20553) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20554) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20555) ("[" 1560) (array_object_dereferenceable 20556) (callable_expr 20557) (callable_variable 20558) (class_constant 1564) (class_name 20559) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 20560) (name 20561) (new_dereferenceable 1570) (simple_variable 20562) (static_member 20563) (variable 20564) (variable_class_name 20565)) 9131 (("$" 20553) (T_VARIABLE 20555) (simple_variable 20598) ("{" 20599)) 9132 (("(" 20569) (argument_list 20597)) 9133 ((T_NULLSAFE_OBJECT_OPERATOR 20583) (T_OBJECT_OPERATOR 20584) ("[" 20585)) 9134 (("(" 20569) (argument_list 20582)) 9135 ((T_PAAMAYIM_NEKUDOTAYIM 20578)) 9136 (("(" 20569) (argument_list 20577)) 9137 ((T_PAAMAYIM_NEKUDOTAYIM 20566)) 9138 (("$" 20553) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 20555) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 20567) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 20568) ("{" 1660)) 9139 (("(" 20569) (argument_list 20570)) 9140 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 20571) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 20572) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 20573) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 9141 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 20576) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 9142 (("," 1498) (possible_comma 20574)) 9143 ((")" 20575)) 9144 (("$" 20553) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 20555) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 20579) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 20580) ("{" 1676)) 9145 (("(" 20569) (argument_list 20581)) 9146 (("$" 20553) (T_STRING 20588) (T_VARIABLE 20555) (property_name 20595) (simple_variable 20590) ("{" 20591)) 9147 (("$" 20553) (T_STRING 20588) (T_VARIABLE 20555) (property_name 20589) (simple_variable 20590) ("{" 20591)) 9148 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 20586) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 9149 (("]" 20587)) 9150 (("(" 20569) (argument_list 20594)) 9151 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 20592) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 9152 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 20593)) 9153 (("(" 20569) (argument_list 20596)) 9154 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 20600) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 9155 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 20601)) 9156 ((backup_doc_comment 20604)) 9157 (("(" 20605)) 9158 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 20606)) 9159 ((")" 20607)) 9160 ((T_USE 166) (lexical_vars 20608)) 9161 ((":" 168) (return_type 20609)) 9162 ((backup_fn_flags 20610)) 9163 (("{" 20611)) 9164 ((inner_statement_list 20612)) 9165 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 20613) ("~" 129)) 9166 ((backup_fn_flags 20614)) 9167 ((backup_doc_comment 20616)) 9168 (("(" 20617)) 9169 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 20618)) 9170 ((")" 20619)) 9171 ((":" 2271) (return_type 20620)) 9172 ((T_DOUBLE_ARROW 20621)) 9173 ((backup_fn_flags 20622)) 9174 ((backup_lex_pos 20623)) 9175 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20624) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9176 (("%" 20326) ("*" 20327) ("+" 20328) ("-" 20329) ("." 20330) ("/" 20331) ("<" 20332) (">" 20333) ("?" 20334) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 20335) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 20336) (T_BOOLEAN_AND 20337) (T_BOOLEAN_OR 20338) (T_COALESCE 20339) (T_INSTANCEOF 20340) (T_IS_EQUAL 20341) (T_IS_GREATER_OR_EQUAL 20342) (T_IS_IDENTICAL 20343) (T_IS_NOT_EQUAL 20344) (T_IS_NOT_IDENTICAL 20345) (T_IS_SMALLER_OR_EQUAL 20346) (T_LOGICAL_AND 20347) (T_LOGICAL_OR 20348) (T_LOGICAL_XOR 20349) (T_PIPE 20350) (T_POW 20351) (T_SL 20352) (T_SPACESHIP 20353) (T_SR 20354) ("^" 20355) (backup_fn_flags 20625) ("|" 20356)) 9177 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20627) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 9178 (("$" 20248) (T_ABSTRACT 20427) (T_ARRAY 20428) (T_AS 20429) (T_BREAK 20430) (T_CALLABLE 20431) (T_CASE 20432) (T_CATCH 20433) (T_CLASS 20434) (T_CLASS_C 20435) (T_CLONE 20436) (T_CONST 20437) (T_CONTINUE 20438) (T_DECLARE 20439) (T_DEFAULT 20440) (T_DIR 20441) (T_DO 20442) (T_ECHO 20443) (T_ELSE 20444) (T_ELSEIF 20445) (T_EMPTY 20446) (T_ENDDECLARE 20447) (T_ENDFOR 20448) (T_ENDFOREACH 20449) (T_ENDIF 20450) (T_ENDSWITCH 20451) (T_ENDWHILE 20452) (T_ENUM 20453) (T_EVAL 20454) (T_EXIT 20455) (T_EXTENDS 20456) (T_FILE 20457) (T_FINAL 20458) (T_FINALLY 20459) (T_FN 20460) (T_FOR 20461) (T_FOREACH 20462) (T_FUNCTION 20463) (T_FUNC_C 20464) (T_GLOBAL 20465) (T_GOTO 20466) (T_IF 20467) (T_IMPLEMENTS 20468) (T_INCLUDE 20469) (T_INCLUDE_ONCE 20470) (T_INSTANCEOF 20471) (T_INSTEADOF 20472) (T_INTERFACE 20473) (T_ISSET 20474) (T_LINE 20475) (T_LIST 20476) (T_LOGICAL_AND 20477) (T_LOGICAL_OR 20478) (T_LOGICAL_XOR 20479) (T_MATCH 20480) (T_METHOD_C 20481) (T_NAMESPACE 20482) (T_NEW 20483) (T_NS_C 20484) (T_PRINT 20485) (T_PRIVATE 20486) (T_PROPERTY_C 20487) (T_PROTECTED 20488) (T_PUBLIC 20489) (T_READONLY 20490) (T_REQUIRE 20491) (T_REQUIRE_ONCE 20492) (T_RETURN 20493) (T_STATIC 20494) (T_STRING 20495) (T_SWITCH 20496) (T_THROW 20497) (T_TRAIT 20498) (T_TRAIT_C 20499) (T_TRY 20500) (T_UNSET 20501) (T_USE 20502) (T_VAR 20503) (T_VARIABLE 20296) (T_WHILE 20504) (T_YIELD 20505) (identifier 20629) (member_name 20630) (reserved_non_modifiers 20508) (semi_reserved 20509) (simple_variable 20631) ("{" 20632)) 9179 (("(" 20514) (argument_list 20635)) 9180 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 20633) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 9181 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 20634)) 9182 ((T_FN 35) (T_FUNCTION 131) (fn 20310) #'20311 (inline_function 20639)) 9183 (("$" 20248) (T_STRING 20645) (T_VARIABLE 20296) (property_name 20652) (simple_variable 20647) ("{" 20648)) 9184 (("$" 20248) (T_STRING 20645) (T_VARIABLE 20296) (property_name 20646) (simple_variable 20647) ("{" 20648)) 9185 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 20643) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 9186 (("]" 20644)) 9187 (("(" 20514) (argument_list 20651)) 9188 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 20649) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 9189 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 20650)) 9190 (("(" 20514) (argument_list 20653)) 9191 (("`" 20655)) 9192 (("]" 20657)) 9193 (("=" 20658)) 9194 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20659) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9195 (("%" 20326) ("*" 20327) ("+" 20328) ("-" 20329) ("." 20330) ("/" 20331) ("<" 20332) (">" 20333) ("?" 20334) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 20335) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 20336) (T_BOOLEAN_AND 20337) (T_BOOLEAN_OR 20338) (T_COALESCE 20339) (T_DOUBLE_ARROW 20662) (T_INSTANCEOF 20340) (T_IS_EQUAL 20341) (T_IS_GREATER_OR_EQUAL 20342) (T_IS_IDENTICAL 20343) (T_IS_NOT_EQUAL 20344) (T_IS_NOT_IDENTICAL 20345) (T_IS_SMALLER_OR_EQUAL 20346) (T_LOGICAL_AND 20347) (T_LOGICAL_OR 20348) (T_LOGICAL_XOR 20349) (T_PIPE 20350) (T_POW 20351) (T_SL 20352) (T_SPACESHIP 20353) (T_SR 20354) ("^" 20355) ("|" 20356)) 9196 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20663) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9197 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 20672) (T_VARIABLE 2920) (encaps_var 2934)) 9198 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 20671) (T_VARIABLE 2920) (encaps_var 2925)) 9199 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 20713) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 9200 ((T_CLASS 20705)) 9201 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 20704) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 20685) (attribute 2982) (class_modifier 2963)) 9202 ((T_PAAMAYIM_NEKUDOTAYIM 20702)) 9203 (("(" 20569) (argument_list 20701)) 9204 ((T_NULLSAFE_OBJECT_OPERATOR 20692) (T_OBJECT_OPERATOR 20693) (T_PAAMAYIM_NEKUDOTAYIM 20694) ("[" 20695)) 9205 (("$" 20553) (T_STRING 20588) (T_VARIABLE 20555) (property_name 20700) (simple_variable 20590) ("{" 20591)) 9206 (("$" 20553) (T_STRING 20588) (T_VARIABLE 20555) (property_name 20699) (simple_variable 20590) ("{" 20591)) 9207 (("$" 20553) (T_VARIABLE 20555) (simple_variable 20698)) 9208 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 20696) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 9209 (("]" 20697)) 9210 (("$" 20553) (T_VARIABLE 20555) (simple_variable 20703)) 9211 (("(" 2984) (argument_list 2985) (ctor_arguments 20706)) 9212 ((T_EXTENDS 2987) (extends_from 20707)) 9213 ((T_IMPLEMENTS 2989) (implements_list 20708)) 9214 ((backup_doc_comment 20709)) 9215 (("{" 20710)) 9216 ((class_statement_list 20711)) 9217 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 20712)) 9218 (("%" 304) (")" 20714) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 9219 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 20716) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 9220 (("%" 304) (")" 20717) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 9221 (("{" 20718)) 9222 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 20719) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 9223 (("}" 20720)) 9224 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 20722) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 9225 ((")" 20723)) 9226 (("=" 20724)) 9227 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 20725) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 9228 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 20727) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 9229 (("," 5226) (possible_comma 20728)) 9230 ((")" 20729)) 9231 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 20737) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 20738) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 20739) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 9232 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 20742) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 9233 (("," 1498) (possible_comma 20740)) 9234 ((")" 20741)) 9235 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 20744) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 9236 (("%" 304) (")" 20745) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 9237 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 20747) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 9238 (("%" 304) (")" 20748) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 9239 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 20754) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 20755) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 20756) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 20757) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 9240 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 20763) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 9241 (("%" 2442) (")" 20760) ("*" 2443) ("+" 2444) ("," 20761) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 9242 (("," 5287) (possible_comma 20758)) 9243 ((")" 20759)) 9244 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 20762) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 9245 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 20767) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 9246 ((")" 20768)) 9247 (("%" 304) (")" 20760) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 9248 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 20775) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 9249 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 20776)) 9250 (("\"" 20778) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 9251 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 20788) (T_VARIABLE 2920) (encaps_var 2934)) 9252 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 20787) (T_VARIABLE 2920) (encaps_var 2925)) 9253 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 20829) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 9254 ((T_CLASS 20821)) 9255 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 20820) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 20801) (attribute 2982) (class_modifier 2963)) 9256 ((T_PAAMAYIM_NEKUDOTAYIM 20818)) 9257 (("(" 20155) (argument_list 20817)) 9258 ((T_NULLSAFE_OBJECT_OPERATOR 20808) (T_OBJECT_OPERATOR 20809) (T_PAAMAYIM_NEKUDOTAYIM 20810) ("[" 20811)) 9259 (("$" 20139) (T_STRING 20174) (T_VARIABLE 20141) (property_name 20816) (simple_variable 20176) ("{" 20177)) 9260 (("$" 20139) (T_STRING 20174) (T_VARIABLE 20141) (property_name 20815) (simple_variable 20176) ("{" 20177)) 9261 (("$" 20139) (T_VARIABLE 20141) (simple_variable 20814)) 9262 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 20812) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 9263 (("]" 20813)) 9264 (("$" 20139) (T_VARIABLE 20141) (simple_variable 20819)) 9265 (("(" 2984) (argument_list 2985) (ctor_arguments 20822)) 9266 ((T_EXTENDS 2987) (extends_from 20823)) 9267 ((T_IMPLEMENTS 2989) (implements_list 20824)) 9268 ((backup_doc_comment 20825)) 9269 (("{" 20826)) 9270 ((class_statement_list 20827)) 9271 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 20828)) 9272 (("%" 304) (")" 20830) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 9273 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 20832) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 9274 (("%" 304) (")" 20833) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 9275 (("{" 20834)) 9276 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 20835) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 9277 (("}" 20836)) 9278 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 20838) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 9279 ((")" 20839)) 9280 (("=" 20840)) 9281 (("!" 19832) ("\"" 19833) ("$" 19834) ("(" 19835) ("+" 19836) ("-" 19837) ("@" 19838) (T_ARRAY 19839) (T_ARRAY_CAST 19840) (T_ATTRIBUTE 130) (T_BOOL_CAST 19841) (T_CLASS_C 19842) (T_CLONE 19843) (T_CONSTANT_ENCAPSED_STRING 19844) (T_DEC 19845) (T_DIR 19846) (T_DNUMBER 19847) (T_DOUBLE_CAST 19848) (T_EMPTY 19849) (T_EVAL 19850) (T_EXIT 19851) (T_FILE 19852) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19853) (T_INC 19854) (T_INCLUDE 19855) (T_INCLUDE_ONCE 19856) (T_INT_CAST 19857) (T_ISSET 19858) (T_LINE 19859) (T_LIST 19860) (T_LNUMBER 19861) (T_MATCH 19862) (T_METHOD_C 19863) (T_NAME_FULLY_QUALIFIED 19864) (T_NAME_QUALIFIED 19865) (T_NAME_RELATIVE 19866) (T_NEW 19867) (T_NS_C 19868) (T_OBJECT_CAST 19869) (T_PRINT 19870) (T_PROPERTY_C 19871) (T_READONLY 19872) (T_REQUIRE 19873) (T_REQUIRE_ONCE 19874) (T_START_HEREDOC 19875) (T_STATIC 19876) (T_STRING 19877) (T_STRING_CAST 19878) (T_THROW 19879) (T_TRAIT_C 19880) (T_UNSET_CAST 19881) (T_VARIABLE 19882) (T_YIELD 19883) (T_YIELD_FROM 19884) ("[" 19885) ("`" 19886) (array_object_dereferenceable 19887) (attribute 135) (attributes 19888) (callable_expr 19889) (callable_variable 19890) (class_constant 19891) (class_name 19892) (constant 19893) (dereferenceable_scalar 19894) (expr 20841) (fn 19896) (fully_dereferenceable 107) #'19897 (function_call 19898) (inline_function 19899) (internal_functions_in_yacc 19900) (match 19901) (name 19902) (new_dereferenceable 19903) (new_non_dereferenceable 19904) (scalar 19905) (simple_variable 19906) (static_member 19907) (variable 19908) (variable_class_name 19909) ("~" 19910)) 9282 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 20843) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 9283 (("," 5226) (possible_comma 20844)) 9284 ((")" 20845)) 9285 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 20853) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 20854) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 20855) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 9286 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 20858) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 9287 (("," 1498) (possible_comma 20856)) 9288 ((")" 20857)) 9289 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 20860) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 9290 (("%" 304) (")" 20861) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 9291 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 20863) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 9292 (("%" 304) (")" 20864) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 9293 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 20870) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 20871) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 20872) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 20873) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 9294 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 20879) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 9295 (("%" 2442) (")" 20876) ("*" 2443) ("+" 2444) ("," 20877) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 9296 (("," 5287) (possible_comma 20874)) 9297 ((")" 20875)) 9298 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 20878) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 9299 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 20883) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 9300 ((")" 20884)) 9301 (("%" 304) (")" 20876) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 9302 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 20891) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 9303 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 20892)) 9304 (("\"" 20894) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 9305 (("\"" 1538) ("$" 20897) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20898) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20899) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20900) ("[" 20901) (ampersand 20902) (array_object_dereferenceable 20903) (callable_expr 20904) (callable_variable 20905) (class_constant 1564) (class_name 20906) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 20907) (fully_dereferenceable 107) (function_call 20908) (name 20909) (new_dereferenceable 1570) (simple_variable 20910) (static_member 20911) (variable 20912) (variable_class_name 20913)) 9306 (("$" 20897) (T_VARIABLE 20900) (simple_variable 24108) ("{" 24109)) 9307 (("(" 24105)) 9308 (("(" 20917) (argument_list 24104)) 9309 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 24102) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 9310 (("\"" 1538) ("$" 20897) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20899) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20900) ("[" 1560) (array_object_dereferenceable 20903) (callable_expr 20904) (callable_variable 20905) (class_constant 1564) (class_name 20906) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 20908) (name 20909) (new_dereferenceable 1570) (simple_variable 20910) (static_member 20911) (variable 24101) (variable_class_name 20913)) 9311 ((T_NULLSAFE_OBJECT_OPERATOR 24087) (T_OBJECT_OPERATOR 24088) ("[" 24089)) 9312 (("(" 20917) (argument_list 24086)) 9313 ((T_PAAMAYIM_NEKUDOTAYIM 24082)) 9314 ((")" 20926) (T_DOUBLE_ARROW 20927)) 9315 (("(" 20917) (argument_list 20925)) 9316 ((T_PAAMAYIM_NEKUDOTAYIM 20914)) 9317 (("$" 20897) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 20900) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 20915) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 20916) ("{" 1660)) 9318 (("(" 20917) (argument_list 20918)) 9319 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 20919) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 20920) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 20921) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 9320 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 20924) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 9321 (("," 1498) (possible_comma 20922)) 9322 ((")" 20923)) 9323 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 20958) (";" 18275) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 18276) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18277) (T_DEC 22) (T_DECLARE 18278) (T_DIR 24) (T_DNUMBER 25) (T_DO 18279) (T_DOUBLE_CAST 27) (T_ECHO 18280) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18281) (T_FOREACH 18282) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 18283) (T_GOTO 18284) (T_IF 18239) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18285) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 18286) (T_START_HEREDOC 69) (T_STATIC 18287) (T_STRING 18288) (T_STRING_CAST 72) (T_SWITCH 18289) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 18290) (T_UNSET 18291) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 18292) (T_WHILE 18293) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 18294) (alt_if_stmt_without_else 18295) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 18296) (fn 106) (foreach_statement 24081) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 18297) (if_stmt_without_else 18298) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 20960) (static_member 123) (variable 126) (variable_class_name 127) ("{" 18300) ("~" 129)) 9324 (("\"" 1538) ("$" 20928) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20929) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20930) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20931) ("[" 20932) (ampersand 20933) (array_object_dereferenceable 20934) (callable_expr 20935) (callable_variable 20936) (class_constant 1564) (class_name 20937) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 20938) (fully_dereferenceable 107) (function_call 20939) (name 20940) (new_dereferenceable 1570) (simple_variable 20941) (static_member 20942) (variable 20943) (variable_class_name 20944)) 9325 (("$" 20928) (T_VARIABLE 20931) (simple_variable 24077) ("{" 24078)) 9326 (("(" 24074)) 9327 (("(" 20948) (argument_list 24073)) 9328 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 24071) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 9329 (("\"" 1538) ("$" 20928) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20930) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20931) ("[" 1560) (array_object_dereferenceable 20934) (callable_expr 20935) (callable_variable 20936) (class_constant 1564) (class_name 20937) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 20939) (name 20940) (new_dereferenceable 1570) (simple_variable 20941) (static_member 20942) (variable 24070) (variable_class_name 20944)) 9330 ((T_NULLSAFE_OBJECT_OPERATOR 24056) (T_OBJECT_OPERATOR 24057) ("[" 24058)) 9331 (("(" 20948) (argument_list 24055)) 9332 ((T_PAAMAYIM_NEKUDOTAYIM 24051)) 9333 ((")" 20957)) 9334 (("(" 20948) (argument_list 20956)) 9335 ((T_PAAMAYIM_NEKUDOTAYIM 20945)) 9336 (("$" 20928) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 20931) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 20946) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 20947) ("{" 1660)) 9337 (("(" 20948) (argument_list 20949)) 9338 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 20950) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 20951) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 20952) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 9339 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 20955) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 9340 (("," 1498) (possible_comma 20953)) 9341 ((")" 20954)) 9342 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 20958) (";" 18275) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 18276) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18277) (T_DEC 22) (T_DECLARE 18278) (T_DIR 24) (T_DNUMBER 25) (T_DO 18279) (T_DOUBLE_CAST 27) (T_ECHO 18280) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18281) (T_FOREACH 18282) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 18283) (T_GOTO 18284) (T_IF 18239) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18285) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 18286) (T_START_HEREDOC 69) (T_STATIC 18287) (T_STRING 18288) (T_STRING_CAST 72) (T_SWITCH 18289) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 18290) (T_UNSET 18291) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 18292) (T_WHILE 18293) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 18294) (alt_if_stmt_without_else 18295) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 18296) (fn 106) (foreach_statement 20959) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 18297) (if_stmt_without_else 18298) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 20960) (static_member 123) (variable 126) (variable_class_name 127) ("{" 18300) ("~" 129)) 9343 ((inner_statement_list 20961)) 9344 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20962) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 20963) (T_CLASS 20964) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20965) (T_DEC 22) (T_DECLARE 20966) (T_DIR 24) (T_DNUMBER 25) (T_DO 20967) (T_DOUBLE_CAST 27) (T_ECHO 20968) (T_EMPTY 29) (T_ENDFOREACH 20969) (T_ENUM 20970) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20971) (T_FOREACH 20972) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20973) (T_GOTO 20974) (T_HALT_COMPILER 20975) (T_IF 20976) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20977) (T_INTERFACE 20978) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20979) (T_START_HEREDOC 69) (T_STATIC 20980) (T_STRING 20981) (T_STRING_CAST 72) (T_SWITCH 20982) (T_THROW 74) (T_TRAIT 20983) (T_TRAIT_C 76) (T_TRY 20984) (T_UNSET 20985) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 20986) (T_WHILE 20987) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 20988) (alt_if_stmt_without_else 20989) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 20990) (attributes 20991) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 20992) (class_modifier 99) (class_modifiers 20993) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 20994) (expr 20995) (fn 106) (fully_dereferenceable 107) #'20996 (function_call 109) (function_declaration_statement 20997) (if_stmt 20998) (if_stmt_without_else 20999) (inline_function 113) (inner_statement 21000) (interface_declaration_statement 21001) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 21002) (static_member 123) (trait_declaration_statement 21003) (variable 126) (variable_class_name 127) ("{" 21004) ("~" 129)) 9345 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 24049) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 9346 ((T_STRING 24042)) 9347 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 24040) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 9348 (("(" 24031)) 9349 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22932) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22933) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22934) (T_DEC 22) (T_DECLARE 22935) (T_DIR 24) (T_DNUMBER 25) (T_DO 22936) (T_DOUBLE_CAST 27) (T_ECHO 22937) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22938) (T_FOREACH 22939) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22940) (T_GOTO 22941) (T_IF 22942) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22943) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22944) (T_START_HEREDOC 69) (T_STATIC 22945) (T_STRING 22946) (T_STRING_CAST 72) (T_SWITCH 22947) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22948) (T_UNSET 22949) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22950) (T_WHILE 22951) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22952) (alt_if_stmt_without_else 22953) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22954) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22955) (if_stmt_without_else 22956) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 24025) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22958) ("~" 129)) 9350 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (echo_expr 22926) (echo_expr_list 24023) (expr 22928) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 9351 ((";" 24022)) 9352 ((T_STRING 24015)) 9353 (("(" 24002)) 9354 (("(" 23987)) 9355 (("$" 19441) (T_VARIABLE 19442) (global_var 19443) (global_var_list 23985) (simple_variable 19445)) 9356 ((T_STRING 23983)) 9357 (("(" 23980)) 9358 (("(" 23976)) 9359 ((T_STRING 23970)) 9360 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 23968) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 9361 ((T_FN 35) (T_FUNCTION 131) (T_VARIABLE 18365) (fn 106) #'138 (inline_function 18366) (static_var 18367) (static_var_list 23966)) 9362 ((":" 23965)) 9363 (("(" 23947)) 9364 ((T_STRING 23942)) 9365 (("{" 23925)) 9366 (("(" 23920)) 9367 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 23918) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 9368 (("(" 23909)) 9369 ((T_ELSE 23902) (T_ELSEIF 17711) (T_ENDIF 23903)) 9370 ((T_ABSTRACT 10) (T_ATTRIBUTE 174) (T_CLASS 20964) (T_ENUM 20970) (T_FINAL 34) (T_FN 35) (T_FUNCTION 38) (T_INTERFACE 20978) (T_READONLY 2955) (T_STATIC 17797) (T_TRAIT 20983) (attribute 17798) (attributed_statement 23901) (class_declaration_statement 20992) (class_modifier 99) (class_modifiers 20993) (enum_declaration_statement 20994) (fn 106) #'20996 (function_declaration_statement 20997) (inline_function 17800) (interface_declaration_statement 21001) (trait_declaration_statement 21003)) 9371 ((T_ABSTRACT 10) (T_CLASS 23893) (T_FINAL 34) (T_READONLY 2955) (class_modifier 17789)) 9372 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 23892) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 9373 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17770) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17771) (ampersand 17772) (returns_ref 23880)) 9374 ((T_ELSE 21007) (T_ELSEIF 21008)) 9375 ((inner_statement_list 21005)) 9376 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 21006) ("~" 129)) 9377 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20962) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 20963) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20965) (T_DEC 22) (T_DECLARE 20966) (T_DIR 24) (T_DNUMBER 25) (T_DO 20967) (T_DOUBLE_CAST 27) (T_ECHO 20968) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20971) (T_FOREACH 20972) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 20973) (T_GOTO 20974) (T_IF 20976) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20977) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20979) (T_START_HEREDOC 69) (T_STATIC 20980) (T_STRING 20981) (T_STRING_CAST 72) (T_SWITCH 20982) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20984) (T_UNSET 20985) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 20986) (T_WHILE 20987) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 20988) (alt_if_stmt_without_else 20989) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 20995) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 20998) (if_stmt_without_else 20999) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23879) (static_member 123) (variable 126) (variable_class_name 127) ("{" 21004) ("~" 129)) 9378 (("(" 21009)) 9379 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 21010) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 9380 (("%" 304) (")" 21011) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 9381 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 21012) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 21013) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21014) (T_DEC 22) (T_DECLARE 21015) (T_DIR 24) (T_DNUMBER 25) (T_DO 21016) (T_DOUBLE_CAST 27) (T_ECHO 21017) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 21018) (T_FOREACH 21019) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 21020) (T_GOTO 21021) (T_IF 20976) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21022) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21023) (T_START_HEREDOC 69) (T_STATIC 21024) (T_STRING 21025) (T_STRING_CAST 72) (T_SWITCH 21026) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 21027) (T_UNSET 21028) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 21029) (T_WHILE 21030) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 21031) (alt_if_stmt_without_else 21032) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 21033) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 21034) (if_stmt_without_else 21035) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 21036) (static_member 123) (variable 126) (variable_class_name 127) ("{" 21037) ("~" 129)) 9382 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 23877) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 9383 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 23875) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 9384 (("(" 23866)) 9385 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22932) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22933) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22934) (T_DEC 22) (T_DECLARE 22935) (T_DIR 24) (T_DNUMBER 25) (T_DO 22936) (T_DOUBLE_CAST 27) (T_ECHO 22937) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22938) (T_FOREACH 22939) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22940) (T_GOTO 22941) (T_IF 22942) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22943) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22944) (T_START_HEREDOC 69) (T_STATIC 22945) (T_STRING 22946) (T_STRING_CAST 72) (T_SWITCH 22947) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22948) (T_UNSET 22949) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22950) (T_WHILE 22951) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22952) (alt_if_stmt_without_else 22953) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22954) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22955) (if_stmt_without_else 22956) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23860) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22958) ("~" 129)) 9386 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (echo_expr 22926) (echo_expr_list 23858) (expr 22928) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 9387 (("(" 21125)) 9388 (("(" 21110)) 9389 (("$" 19441) (T_VARIABLE 19442) (global_var 19443) (global_var_list 21108) (simple_variable 19445)) 9390 ((T_STRING 21106)) 9391 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 21104) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 9392 ((T_FN 35) (T_FUNCTION 131) (T_VARIABLE 18365) (fn 106) #'138 (inline_function 18366) (static_var 18367) (static_var_list 21102)) 9393 ((":" 21101)) 9394 (("(" 21083)) 9395 (("{" 21066)) 9396 (("(" 21061)) 9397 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 21059) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 9398 (("(" 21050)) 9399 ((T_ELSE 21043) (T_ELSEIF 17711) (T_ENDIF 21044)) 9400 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 21042) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 9401 ((T_ELSE 21040) (T_ELSEIF 21008)) 9402 ((inner_statement_list 21038)) 9403 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 21039) ("~" 129)) 9404 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 21012) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 21013) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21014) (T_DEC 22) (T_DECLARE 21015) (T_DIR 24) (T_DNUMBER 25) (T_DO 21016) (T_DOUBLE_CAST 27) (T_ECHO 21017) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 21018) (T_FOREACH 21019) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 21020) (T_GOTO 21021) (T_IF 20976) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21022) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21023) (T_START_HEREDOC 69) (T_STATIC 21024) (T_STRING 21025) (T_STRING_CAST 72) (T_SWITCH 21026) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 21027) (T_UNSET 21028) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 21029) (T_WHILE 21030) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 21031) (alt_if_stmt_without_else 21032) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 21033) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 21034) (if_stmt_without_else 21035) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 21041) (static_member 123) (variable 126) (variable_class_name 127) ("{" 21037) ("~" 129)) 9405 ((":" 21046)) 9406 ((";" 21045)) 9407 ((inner_statement_list 21047)) 9408 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17807) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17808) (T_CLASS 17809) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17810) (T_DEC 22) (T_DECLARE 17811) (T_DIR 24) (T_DNUMBER 25) (T_DO 17812) (T_DOUBLE_CAST 27) (T_ECHO 17813) (T_EMPTY 29) (T_ENDIF 21048) (T_ENUM 17815) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17816) (T_FOREACH 17817) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17818) (T_GOTO 17819) (T_HALT_COMPILER 17820) (T_IF 17732) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17821) (T_INTERFACE 17822) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17823) (T_START_HEREDOC 69) (T_STATIC 17824) (T_STRING 17825) (T_STRING_CAST 72) (T_SWITCH 17826) (T_THROW 74) (T_TRAIT 17827) (T_TRAIT_C 76) (T_TRY 17828) (T_UNSET 17829) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17830) (T_WHILE 17831) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17832) (alt_if_stmt_without_else 17833) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17834) (attributes 17835) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17836) (class_modifier 99) (class_modifiers 17837) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17838) (expr 17839) (fn 106) (fully_dereferenceable 107) #'17840 (function_call 109) (function_declaration_statement 17841) (if_stmt 17842) (if_stmt_without_else 17843) (inline_function 113) (inner_statement 17844) (interface_declaration_statement 17845) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17846) (static_member 123) (trait_declaration_statement 17847) (variable 126) (variable_class_name 127) ("{" 17848) ("~" 129)) 9409 ((";" 21049)) 9410 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 21051) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 9411 (("%" 304) (")" 21052) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 9412 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21053) (";" 21012) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 21013) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21014) (T_DEC 22) (T_DECLARE 21015) (T_DIR 24) (T_DNUMBER 25) (T_DO 21016) (T_DOUBLE_CAST 27) (T_ECHO 21017) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 21018) (T_FOREACH 21019) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 21020) (T_GOTO 21021) (T_IF 20976) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21022) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21023) (T_START_HEREDOC 69) (T_STATIC 21024) (T_STRING 21025) (T_STRING_CAST 72) (T_SWITCH 21026) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 21027) (T_UNSET 21028) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 21029) (T_WHILE 21030) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 21031) (alt_if_stmt_without_else 21032) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 21033) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 21034) (if_stmt_without_else 21035) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 21054) (static_member 123) (variable 126) (variable_class_name 127) (while_statement 21055) ("{" 21037) ("~" 129)) 9413 ((inner_statement_list 21056)) 9414 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17889) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17890) (T_CLASS 17891) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17892) (T_DEC 22) (T_DECLARE 17893) (T_DIR 24) (T_DNUMBER 25) (T_DO 17894) (T_DOUBLE_CAST 27) (T_ECHO 17895) (T_EMPTY 29) (T_ENDWHILE 21057) (T_ENUM 17897) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17898) (T_FOREACH 17899) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17900) (T_GOTO 17901) (T_HALT_COMPILER 17902) (T_IF 17903) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17904) (T_INTERFACE 17905) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17906) (T_START_HEREDOC 69) (T_STATIC 17907) (T_STRING 17908) (T_STRING_CAST 72) (T_SWITCH 17909) (T_THROW 74) (T_TRAIT 17910) (T_TRAIT_C 76) (T_TRY 17911) (T_UNSET 17912) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17913) (T_WHILE 17914) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17915) (alt_if_stmt_without_else 17916) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17917) (attributes 17918) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17919) (class_modifier 99) (class_modifiers 17920) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17921) (expr 17922) (fn 106) (fully_dereferenceable 107) #'17923 (function_call 109) (function_declaration_statement 17924) (if_stmt 17925) (if_stmt_without_else 17926) (inline_function 113) (inner_statement 17927) (interface_declaration_statement 17928) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17929) (static_member 123) (trait_declaration_statement 17930) (variable 126) (variable_class_name 127) ("{" 17931) ("~" 129)) 9415 ((";" 21058)) 9416 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 21060) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 9417 (("\"" 1538) ("$" 17989) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 17990) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 17991) ("[" 1560) (array_object_dereferenceable 17992) (callable_expr 17993) (callable_variable 17994) (class_constant 1564) (class_name 17995) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 17996) (name 17997) (new_dereferenceable 1570) (simple_variable 17998) (static_member 17999) (unset_variable 18000) (unset_variables 21062) (variable 18002) (variable_class_name 18003)) 9418 (("," 18015) (possible_comma 21063)) 9419 ((")" 21064)) 9420 ((";" 21065)) 9421 ((inner_statement_list 21067)) 9422 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 21068) ("~" 129)) 9423 ((catch_list 21069)) 9424 ((T_CATCH 21070) (T_FINALLY 21071) (finally_statement 21072)) 9425 (("(" 21076)) 9426 (("{" 21073)) 9427 ((inner_statement_list 21074)) 9428 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 21075) ("~" 129)) 9429 ((T_NAME_FULLY_QUALIFIED 18056) (T_NAME_QUALIFIED 18057) (T_NAME_RELATIVE 18058) (T_STATIC 18059) (T_STRING 18060) (catch_name_list 21077) (class_name 18062) (name 18063)) 9430 ((T_VARIABLE 18064) (optional_variable 21078) ("|" 18066)) 9431 ((")" 21079)) 9432 (("{" 21080)) 9433 ((inner_statement_list 21081)) 9434 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 21082) ("~" 129)) 9435 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 21084) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 9436 (("%" 304) (")" 21085) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 9437 ((":" 21086) (switch_case_list 21087) ("{" 21088)) 9438 ((";" 21094) (case_list 21095)) 9439 ((";" 21089) (case_list 21090)) 9440 ((case_list 21092)) 9441 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 21091)) 9442 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 21093)) 9443 ((case_list 21098)) 9444 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 21096)) 9445 ((";" 21097)) 9446 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 21099)) 9447 ((";" 21100)) 9448 (("," 18369) (";" 21103)) 9449 ((";" 21105)) 9450 ((";" 21107)) 9451 (("," 19446) (";" 21109)) 9452 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 21111) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 9453 (("%" 19534) ("*" 19535) ("+" 19536) ("-" 19537) ("." 19538) ("/" 19539) ("<" 19540) (">" 19541) ("?" 19542) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19543) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19544) (T_AS 21112) (T_BOOLEAN_AND 19545) (T_BOOLEAN_OR 19546) (T_COALESCE 19547) (T_INSTANCEOF 19548) (T_IS_EQUAL 19549) (T_IS_GREATER_OR_EQUAL 19550) (T_IS_IDENTICAL 19551) (T_IS_NOT_EQUAL 19552) (T_IS_NOT_IDENTICAL 19553) (T_IS_SMALLER_OR_EQUAL 19554) (T_LOGICAL_AND 19555) (T_LOGICAL_OR 19556) (T_LOGICAL_XOR 19557) (T_PIPE 19558) (T_POW 19559) (T_SL 19560) (T_SPACESHIP 19561) (T_SR 19562) ("^" 19563) ("|" 19564)) 9454 (("\"" 1538) ("$" 20897) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20898) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20899) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20900) ("[" 20901) (ampersand 20902) (array_object_dereferenceable 20903) (callable_expr 20904) (callable_variable 20905) (class_constant 1564) (class_name 20906) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 21113) (fully_dereferenceable 107) (function_call 20908) (name 20909) (new_dereferenceable 1570) (simple_variable 20910) (static_member 20911) (variable 20912) (variable_class_name 20913)) 9455 ((")" 21114) (T_DOUBLE_ARROW 21115)) 9456 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21118) (";" 21012) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 21013) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21014) (T_DEC 22) (T_DECLARE 21015) (T_DIR 24) (T_DNUMBER 25) (T_DO 21016) (T_DOUBLE_CAST 27) (T_ECHO 21017) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 21018) (T_FOREACH 21019) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 21020) (T_GOTO 21021) (T_IF 20976) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21022) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21023) (T_START_HEREDOC 69) (T_STATIC 21024) (T_STRING 21025) (T_STRING_CAST 72) (T_SWITCH 21026) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 21027) (T_UNSET 21028) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 21029) (T_WHILE 21030) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 21031) (alt_if_stmt_without_else 21032) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 21033) (fn 106) (foreach_statement 21124) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 21034) (if_stmt_without_else 21035) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 21120) (static_member 123) (variable 126) (variable_class_name 127) ("{" 21037) ("~" 129)) 9457 (("\"" 1538) ("$" 20928) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20929) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20930) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20931) ("[" 20932) (ampersand 20933) (array_object_dereferenceable 20934) (callable_expr 20935) (callable_variable 20936) (class_constant 1564) (class_name 20937) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 21116) (fully_dereferenceable 107) (function_call 20939) (name 20940) (new_dereferenceable 1570) (simple_variable 20941) (static_member 20942) (variable 20943) (variable_class_name 20944)) 9458 ((")" 21117)) 9459 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 21118) (";" 21012) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 21013) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21014) (T_DEC 22) (T_DECLARE 21015) (T_DIR 24) (T_DNUMBER 25) (T_DO 21016) (T_DOUBLE_CAST 27) (T_ECHO 21017) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 21018) (T_FOREACH 21019) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 21020) (T_GOTO 21021) (T_IF 20976) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21022) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21023) (T_START_HEREDOC 69) (T_STATIC 21024) (T_STRING 21025) (T_STRING_CAST 72) (T_SWITCH 21026) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 21027) (T_UNSET 21028) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 21029) (T_WHILE 21030) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 21031) (alt_if_stmt_without_else 21032) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 21033) (fn 106) (foreach_statement 21119) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 21034) (if_stmt_without_else 21035) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 21120) (static_member 123) (variable 126) (variable_class_name 127) ("{" 21037) ("~" 129)) 9460 ((inner_statement_list 21121)) 9461 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20962) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 20963) (T_CLASS 20964) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20965) (T_DEC 22) (T_DECLARE 20966) (T_DIR 24) (T_DNUMBER 25) (T_DO 20967) (T_DOUBLE_CAST 27) (T_ECHO 20968) (T_EMPTY 29) (T_ENDFOREACH 21122) (T_ENUM 20970) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20971) (T_FOREACH 20972) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20973) (T_GOTO 20974) (T_HALT_COMPILER 20975) (T_IF 20976) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20977) (T_INTERFACE 20978) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20979) (T_START_HEREDOC 69) (T_STATIC 20980) (T_STRING 20981) (T_STRING_CAST 72) (T_SWITCH 20982) (T_THROW 74) (T_TRAIT 20983) (T_TRAIT_C 76) (T_TRY 20984) (T_UNSET 20985) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 20986) (T_WHILE 20987) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 20988) (alt_if_stmt_without_else 20989) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 20990) (attributes 20991) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 20992) (class_modifier 99) (class_modifiers 20993) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 20994) (expr 20995) (fn 106) (fully_dereferenceable 107) #'20996 (function_call 109) (function_declaration_statement 20997) (if_stmt 20998) (if_stmt_without_else 20999) (inline_function 113) (inner_statement 21000) (interface_declaration_statement 21001) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 21002) (static_member 123) (trait_declaration_statement 21003) (variable 126) (variable_class_name 127) ("{" 21004) ("~" 129)) 9462 ((";" 21123)) 9463 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21126) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21127) (fn 18437) (for_exprs 21128) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21129) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 9464 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 23857) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 9465 ((";" 21134)) 9466 (("," 21130)) 9467 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21131) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21132) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 9468 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21133) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 9469 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21135) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21136) (fn 18437) (for_cond_exprs 21137) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21138) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 9470 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 23856) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9471 ((";" 22736)) 9472 (("," 21139)) 9473 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21140) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21141) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 9474 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21205) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9475 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 22735) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9476 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 22733) (encaps_var 5357)) 9477 (("$" 21144) (T_VARIABLE 21192) (simple_variable 22729) ("{" 22730)) 9478 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 22728) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 9479 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 22727) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9480 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 22726) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9481 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 22725) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9482 (("(" 22722)) 9483 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 22721) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9484 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 22720) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9485 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 22707) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (clone_argument_list 22708) (constant 21203) (dereferenceable_scalar 21204) (expr 22709) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9486 (("\"" 1538) ("$" 21449) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 21450) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 21451) ("[" 1560) (array_object_dereferenceable 21452) (callable_expr 21453) (callable_variable 21454) (class_constant 1564) (class_name 21455) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 21456) (name 21457) (new_dereferenceable 1570) (simple_variable 21458) (static_member 21459) (variable 22706) (variable_class_name 21461)) 9487 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 22705) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9488 (("(" 22702)) 9489 (("(" 22699)) 9490 (("(" 22690) (argument_list 22691) (ctor_arguments 22692)) 9491 (("\"" 1538) ("$" 21449) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 21450) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 21451) ("[" 1560) (array_object_dereferenceable 21452) (callable_expr 21453) (callable_variable 21454) (class_constant 1564) (class_name 21455) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 21456) (name 21457) (new_dereferenceable 1570) (simple_variable 21458) (static_member 21459) (variable 22689) (variable_class_name 21461)) 9492 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 22688) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9493 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 22687) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9494 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 22686) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9495 (("(" 22682)) 9496 (("(" 22677)) 9497 (("(" 22671)) 9498 (("$" 21449) ("(" 22634) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 22635) (T_NAME_QUALIFIED 22636) (T_NAME_RELATIVE 22637) (T_READONLY 2955) (T_STATIC 22638) (T_STRING 22639) (T_VARIABLE 21451) (anonymous_class 22640) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 22641) (attribute 2961) (attributes 22642) (class_modifier 2963) (class_name 22643) (class_name_reference 22644) (name 22645) (new_variable 22646) (simple_variable 22647)) 9499 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 22633) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9500 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 22632) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9501 (("(" 21410) (argument_list 22631)) 9502 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 22630) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9503 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 22629) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9504 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 22624) (T_END_HEREDOC 22625) (T_VARIABLE 2920) (encaps_list 22626) (encaps_var 2922)) 9505 ((T_FN 35) (T_FUNCTION 131) (fn 21206) #'21207 (inline_function 22623)) 9506 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 22622) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9507 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 22621) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9508 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 22620) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9509 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22149) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9510 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 22085) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9511 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 22081) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 9512 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 22079) (encaps_list 2828) (encaps_var 2829)) 9513 ((T_NULLSAFE_OBJECT_OPERATOR 22065) (T_OBJECT_OPERATOR 22066) ("[" 22067)) 9514 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 22062) (attribute 2757) (fn 21206) #'21207 (inline_function 22063)) 9515 (("(" 21410) (argument_list 22061)) 9516 ((T_PAAMAYIM_NEKUDOTAYIM 22053)) 9517 (("%" 21222) ("*" 21223) ("+" 21224) ("-" 21225) ("." 21226) ("/" 21227) ("<" 21228) (">" 21229) ("?" 21230) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 21231) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 21232) (T_BOOLEAN_AND 21233) (T_BOOLEAN_OR 21234) (T_COALESCE 21235) (T_INSTANCEOF 21236) (T_IS_EQUAL 21237) (T_IS_GREATER_OR_EQUAL 21238) (T_IS_IDENTICAL 21239) (T_IS_NOT_EQUAL 21240) (T_IS_NOT_IDENTICAL 21241) (T_IS_SMALLER_OR_EQUAL 21242) (T_LOGICAL_AND 21243) (T_LOGICAL_OR 21244) (T_LOGICAL_XOR 21245) (T_PIPE 21246) (T_POW 21247) (T_SL 21248) (T_SPACESHIP 21249) (T_SR 21250) ("^" 21251) ("|" 21252)) 9518 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 21511)) 9519 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 21499)) 9520 (("(" 21410) (argument_list 21498)) 9521 (("=" 21418) (T_AND_EQUAL 21419) (T_COALESCE_EQUAL 21420) (T_CONCAT_EQUAL 21421) (T_DEC 21422) (T_DIV_EQUAL 21423) (T_INC 21424) (T_MINUS_EQUAL 21425) (T_MOD_EQUAL 21426) (T_MUL_EQUAL 21427) (T_OR_EQUAL 21428) (T_PLUS_EQUAL 21429) (T_POW_EQUAL 21430) (T_SL_EQUAL 21431) (T_SR_EQUAL 21432) (T_XOR_EQUAL 21433)) 9522 ((T_PAAMAYIM_NEKUDOTAYIM 21322)) 9523 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21221) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9524 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21321) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9525 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21320) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9526 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21319) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9527 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21318) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9528 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21317) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9529 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21316) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9530 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21315) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9531 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21314) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9532 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 21309) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 21310) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 9533 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21308) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9534 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21307) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9535 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21306) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9536 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21305) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9537 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21304) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9538 (("$" 21269) ("(" 21270) (T_NAME_FULLY_QUALIFIED 21271) (T_NAME_QUALIFIED 21272) (T_NAME_RELATIVE 21273) (T_STATIC 21274) (T_STRING 21275) (T_VARIABLE 21276) (class_name 21277) (class_name_reference 21278) (name 21279) (new_variable 21280) (simple_variable 21281)) 9539 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21268) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9540 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21267) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9541 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21266) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9542 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21265) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9543 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21264) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9544 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21263) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9545 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21262) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9546 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21261) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9547 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21260) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9548 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21259) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9549 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21258) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9550 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21257) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9551 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21256) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9552 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21255) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9553 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21254) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9554 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21253) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9555 (("$" 21269) (T_VARIABLE 21276) (simple_variable 21300) ("{" 21301)) 9556 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 21298) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 9557 ((T_PAAMAYIM_NEKUDOTAYIM 21296)) 9558 ((T_NULLSAFE_OBJECT_OPERATOR 21282) (T_OBJECT_OPERATOR 21283) (T_PAAMAYIM_NEKUDOTAYIM 21284) ("[" 21285)) 9559 (("$" 21269) (T_STRING 21289) (T_VARIABLE 21276) (property_name 21295) (simple_variable 21291) ("{" 21292)) 9560 (("$" 21269) (T_STRING 21289) (T_VARIABLE 21276) (property_name 21290) (simple_variable 21291) ("{" 21292)) 9561 (("$" 21269) (T_VARIABLE 21276) (simple_variable 21288)) 9562 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 21286) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 9563 (("]" 21287)) 9564 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 21293) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 9565 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 21294)) 9566 (("$" 21269) (T_VARIABLE 21276) (simple_variable 21297)) 9567 (("%" 304) (")" 21299) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 9568 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 21302) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 9569 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 21303)) 9570 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21313) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9571 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 21311) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 9572 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21312) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9573 (("$" 21144) (T_ABSTRACT 21323) (T_ARRAY 21324) (T_AS 21325) (T_BREAK 21326) (T_CALLABLE 21327) (T_CASE 21328) (T_CATCH 21329) (T_CLASS 21330) (T_CLASS_C 21331) (T_CLONE 21332) (T_CONST 21333) (T_CONTINUE 21334) (T_DECLARE 21335) (T_DEFAULT 21336) (T_DIR 21337) (T_DO 21338) (T_ECHO 21339) (T_ELSE 21340) (T_ELSEIF 21341) (T_EMPTY 21342) (T_ENDDECLARE 21343) (T_ENDFOR 21344) (T_ENDFOREACH 21345) (T_ENDIF 21346) (T_ENDSWITCH 21347) (T_ENDWHILE 21348) (T_ENUM 21349) (T_EVAL 21350) (T_EXIT 21351) (T_EXTENDS 21352) (T_FILE 21353) (T_FINAL 21354) (T_FINALLY 21355) (T_FN 21356) (T_FOR 21357) (T_FOREACH 21358) (T_FUNCTION 21359) (T_FUNC_C 21360) (T_GLOBAL 21361) (T_GOTO 21362) (T_IF 21363) (T_IMPLEMENTS 21364) (T_INCLUDE 21365) (T_INCLUDE_ONCE 21366) (T_INSTANCEOF 21367) (T_INSTEADOF 21368) (T_INTERFACE 21369) (T_ISSET 21370) (T_LINE 21371) (T_LIST 21372) (T_LOGICAL_AND 21373) (T_LOGICAL_OR 21374) (T_LOGICAL_XOR 21375) (T_MATCH 21376) (T_METHOD_C 21377) (T_NAMESPACE 21378) (T_NEW 21379) (T_NS_C 21380) (T_PRINT 21381) (T_PRIVATE 21382) (T_PROPERTY_C 21383) (T_PROTECTED 21384) (T_PUBLIC 21385) (T_READONLY 21386) (T_REQUIRE 21387) (T_REQUIRE_ONCE 21388) (T_RETURN 21389) (T_STATIC 21390) (T_STRING 21391) (T_SWITCH 21392) (T_THROW 21393) (T_TRAIT 21394) (T_TRAIT_C 21395) (T_TRY 21396) (T_UNSET 21397) (T_USE 21398) (T_VAR 21399) (T_VARIABLE 21192) (T_WHILE 21400) (T_YIELD 21401) (identifier 21402) (member_name 21403) (reserved_non_modifiers 21404) (semi_reserved 21405) (simple_variable 21406) ("{" 21407)) 9574 (("(" 21410) (argument_list 21411)) 9575 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 21408) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 9576 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 21409)) 9577 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 21412) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 21413) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 21414) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 9578 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 21417) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 9579 (("," 1498) (possible_comma 21415)) 9580 ((")" 21416)) 9581 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (ampersand 21447) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21448) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9582 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21446) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9583 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21445) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9584 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21444) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9585 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21443) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9586 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21442) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9587 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21441) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9588 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21440) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9589 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21439) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9590 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21438) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9591 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21437) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9592 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21436) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9593 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21435) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9594 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 21434) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9595 (("\"" 1538) ("$" 21449) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 21450) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 21451) ("[" 1560) (array_object_dereferenceable 21452) (callable_expr 21453) (callable_variable 21454) (class_constant 1564) (class_name 21455) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 21456) (name 21457) (new_dereferenceable 1570) (simple_variable 21458) (static_member 21459) (variable 21460) (variable_class_name 21461)) 9596 (("$" 21449) (T_VARIABLE 21451) (simple_variable 21494) ("{" 21495)) 9597 (("(" 21465) (argument_list 21493)) 9598 ((T_NULLSAFE_OBJECT_OPERATOR 21479) (T_OBJECT_OPERATOR 21480) ("[" 21481)) 9599 (("(" 21465) (argument_list 21478)) 9600 ((T_PAAMAYIM_NEKUDOTAYIM 21474)) 9601 (("(" 21465) (argument_list 21473)) 9602 ((T_PAAMAYIM_NEKUDOTAYIM 21462)) 9603 (("$" 21449) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 21451) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 21463) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 21464) ("{" 1660)) 9604 (("(" 21465) (argument_list 21466)) 9605 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 21467) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 21468) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 21469) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 9606 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 21472) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 9607 (("," 1498) (possible_comma 21470)) 9608 ((")" 21471)) 9609 (("$" 21449) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 21451) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 21475) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 21476) ("{" 1676)) 9610 (("(" 21465) (argument_list 21477)) 9611 (("$" 21449) (T_STRING 21484) (T_VARIABLE 21451) (property_name 21491) (simple_variable 21486) ("{" 21487)) 9612 (("$" 21449) (T_STRING 21484) (T_VARIABLE 21451) (property_name 21485) (simple_variable 21486) ("{" 21487)) 9613 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 21482) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 9614 (("]" 21483)) 9615 (("(" 21465) (argument_list 21490)) 9616 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 21488) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 9617 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 21489)) 9618 (("(" 21465) (argument_list 21492)) 9619 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 21496) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 9620 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 21497)) 9621 ((backup_doc_comment 21500)) 9622 (("(" 21501)) 9623 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 21502)) 9624 ((")" 21503)) 9625 ((T_USE 166) (lexical_vars 21504)) 9626 ((":" 168) (return_type 21505)) 9627 ((backup_fn_flags 21506)) 9628 (("{" 21507)) 9629 ((inner_statement_list 21508)) 9630 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 21509) ("~" 129)) 9631 ((backup_fn_flags 21510)) 9632 ((backup_doc_comment 21512)) 9633 (("(" 21513)) 9634 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 21514)) 9635 ((")" 21515)) 9636 ((":" 2271) (return_type 21516)) 9637 ((T_DOUBLE_ARROW 21517)) 9638 ((backup_fn_flags 21518)) 9639 ((backup_lex_pos 21519)) 9640 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21583) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9641 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 22052) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9642 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 22050) (encaps_var 5357)) 9643 (("$" 21522) (T_VARIABLE 21570) (simple_variable 22046) ("{" 22047)) 9644 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 22045) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 9645 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 22044) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9646 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 22043) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9647 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 22042) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9648 (("(" 22039)) 9649 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 22038) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9650 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 22037) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9651 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 22024) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (clone_argument_list 22025) (constant 21581) (dereferenceable_scalar 21582) (expr 22026) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9652 (("\"" 1538) ("$" 21827) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 21828) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 21829) ("[" 1560) (array_object_dereferenceable 21830) (callable_expr 21831) (callable_variable 21832) (class_constant 1564) (class_name 21833) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 21834) (name 21835) (new_dereferenceable 1570) (simple_variable 21836) (static_member 21837) (variable 22023) (variable_class_name 21839)) 9653 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 22022) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9654 (("(" 22019)) 9655 (("(" 22016)) 9656 (("(" 22007) (argument_list 22008) (ctor_arguments 22009)) 9657 (("\"" 1538) ("$" 21827) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 21828) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 21829) ("[" 1560) (array_object_dereferenceable 21830) (callable_expr 21831) (callable_variable 21832) (class_constant 1564) (class_name 21833) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 21834) (name 21835) (new_dereferenceable 1570) (simple_variable 21836) (static_member 21837) (variable 22006) (variable_class_name 21839)) 9658 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 22005) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9659 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 22004) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9660 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 22003) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9661 (("(" 21999)) 9662 (("(" 21994)) 9663 (("(" 21988)) 9664 (("$" 21827) ("(" 21951) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 21952) (T_NAME_QUALIFIED 21953) (T_NAME_RELATIVE 21954) (T_READONLY 2955) (T_STATIC 21955) (T_STRING 21956) (T_VARIABLE 21829) (anonymous_class 21957) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 21958) (attribute 2961) (attributes 21959) (class_modifier 2963) (class_name 21960) (class_name_reference 21961) (name 21962) (new_variable 21963) (simple_variable 21964)) 9665 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21950) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9666 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21949) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9667 (("(" 21788) (argument_list 21948)) 9668 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21947) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9669 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21946) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9670 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 21941) (T_END_HEREDOC 21942) (T_VARIABLE 2920) (encaps_list 21943) (encaps_var 2922)) 9671 ((T_FN 35) (T_FUNCTION 131) (fn 21584) #'21585 (inline_function 21940)) 9672 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21939) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9673 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21938) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9674 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21937) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9675 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 21934) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 9676 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21933) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9677 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 21929) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 9678 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 21927) (encaps_list 2828) (encaps_var 2829)) 9679 ((T_NULLSAFE_OBJECT_OPERATOR 21913) (T_OBJECT_OPERATOR 21914) ("[" 21915)) 9680 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 21910) (attribute 2757) (fn 21584) #'21585 (inline_function 21911)) 9681 (("(" 21788) (argument_list 21909)) 9682 ((T_PAAMAYIM_NEKUDOTAYIM 21901)) 9683 (("%" 21600) ("*" 21601) ("+" 21602) ("-" 21603) ("." 21604) ("/" 21605) ("<" 21606) (">" 21607) ("?" 21608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 21609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 21610) (T_BOOLEAN_AND 21611) (T_BOOLEAN_OR 21612) (T_COALESCE 21613) (T_INSTANCEOF 21614) (T_IS_EQUAL 21615) (T_IS_GREATER_OR_EQUAL 21616) (T_IS_IDENTICAL 21617) (T_IS_NOT_EQUAL 21618) (T_IS_NOT_IDENTICAL 21619) (T_IS_SMALLER_OR_EQUAL 21620) (T_LOGICAL_AND 21621) (T_LOGICAL_OR 21622) (T_LOGICAL_XOR 21623) (T_PIPE 21624) (T_POW 21625) (T_SL 21626) (T_SPACESHIP 21627) (T_SR 21628) ("^" 21629) (backup_fn_flags 21900) ("|" 21630)) 9684 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 21889)) 9685 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 21877)) 9686 (("(" 21788) (argument_list 21876)) 9687 (("=" 21796) (T_AND_EQUAL 21797) (T_COALESCE_EQUAL 21798) (T_CONCAT_EQUAL 21799) (T_DEC 21800) (T_DIV_EQUAL 21801) (T_INC 21802) (T_MINUS_EQUAL 21803) (T_MOD_EQUAL 21804) (T_MUL_EQUAL 21805) (T_OR_EQUAL 21806) (T_PLUS_EQUAL 21807) (T_POW_EQUAL 21808) (T_SL_EQUAL 21809) (T_SR_EQUAL 21810) (T_XOR_EQUAL 21811)) 9688 ((T_PAAMAYIM_NEKUDOTAYIM 21700)) 9689 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21599) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9690 (("%" 21600) ("*" 21601) ("+" 21602) ("-" 21603) ("." 21604) ("/" 21605) ("<" 21606) (">" 21607) ("?" 21608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 21609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 21610) (T_BOOLEAN_AND 21611) (T_BOOLEAN_OR 21612) (T_COALESCE 21613) (T_INSTANCEOF 21614) (T_IS_EQUAL 21615) (T_IS_GREATER_OR_EQUAL 21616) (T_IS_IDENTICAL 21617) (T_IS_NOT_EQUAL 21618) (T_IS_NOT_IDENTICAL 21619) (T_IS_SMALLER_OR_EQUAL 21620) (T_LOGICAL_AND 21621) (T_LOGICAL_OR 21622) (T_LOGICAL_XOR 21623) (T_PIPE 21624) (T_POW 21625) (T_SL 21626) (T_SPACESHIP 21627) (T_SR 21628) ("^" 21629) ("|" 21630)) 9691 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21699) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9692 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21698) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9693 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21697) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9694 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21696) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9695 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21695) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9696 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21694) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9697 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21693) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9698 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21692) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9699 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 21687) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 21688) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 9700 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21686) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9701 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21685) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9702 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21684) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9703 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21683) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9704 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21682) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9705 (("$" 21647) ("(" 21648) (T_NAME_FULLY_QUALIFIED 21649) (T_NAME_QUALIFIED 21650) (T_NAME_RELATIVE 21651) (T_STATIC 21652) (T_STRING 21653) (T_VARIABLE 21654) (class_name 21655) (class_name_reference 21656) (name 21657) (new_variable 21658) (simple_variable 21659)) 9706 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21646) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9707 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21645) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9708 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21644) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9709 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21643) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9710 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21642) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9711 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21641) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9712 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21640) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9713 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21639) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9714 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21638) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9715 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21637) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9716 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21636) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9717 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21635) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9718 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21634) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9719 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21633) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9720 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21632) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9721 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21631) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9722 (("$" 21647) (T_VARIABLE 21654) (simple_variable 21678) ("{" 21679)) 9723 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 21676) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 9724 ((T_PAAMAYIM_NEKUDOTAYIM 21674)) 9725 ((T_NULLSAFE_OBJECT_OPERATOR 21660) (T_OBJECT_OPERATOR 21661) (T_PAAMAYIM_NEKUDOTAYIM 21662) ("[" 21663)) 9726 (("$" 21647) (T_STRING 21667) (T_VARIABLE 21654) (property_name 21673) (simple_variable 21669) ("{" 21670)) 9727 (("$" 21647) (T_STRING 21667) (T_VARIABLE 21654) (property_name 21668) (simple_variable 21669) ("{" 21670)) 9728 (("$" 21647) (T_VARIABLE 21654) (simple_variable 21666)) 9729 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 21664) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 9730 (("]" 21665)) 9731 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 21671) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 9732 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 21672)) 9733 (("$" 21647) (T_VARIABLE 21654) (simple_variable 21675)) 9734 (("%" 304) (")" 21677) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 9735 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 21680) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 9736 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 21681)) 9737 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21691) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9738 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 21689) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 9739 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21690) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9740 (("$" 21522) (T_ABSTRACT 21701) (T_ARRAY 21702) (T_AS 21703) (T_BREAK 21704) (T_CALLABLE 21705) (T_CASE 21706) (T_CATCH 21707) (T_CLASS 21708) (T_CLASS_C 21709) (T_CLONE 21710) (T_CONST 21711) (T_CONTINUE 21712) (T_DECLARE 21713) (T_DEFAULT 21714) (T_DIR 21715) (T_DO 21716) (T_ECHO 21717) (T_ELSE 21718) (T_ELSEIF 21719) (T_EMPTY 21720) (T_ENDDECLARE 21721) (T_ENDFOR 21722) (T_ENDFOREACH 21723) (T_ENDIF 21724) (T_ENDSWITCH 21725) (T_ENDWHILE 21726) (T_ENUM 21727) (T_EVAL 21728) (T_EXIT 21729) (T_EXTENDS 21730) (T_FILE 21731) (T_FINAL 21732) (T_FINALLY 21733) (T_FN 21734) (T_FOR 21735) (T_FOREACH 21736) (T_FUNCTION 21737) (T_FUNC_C 21738) (T_GLOBAL 21739) (T_GOTO 21740) (T_IF 21741) (T_IMPLEMENTS 21742) (T_INCLUDE 21743) (T_INCLUDE_ONCE 21744) (T_INSTANCEOF 21745) (T_INSTEADOF 21746) (T_INTERFACE 21747) (T_ISSET 21748) (T_LINE 21749) (T_LIST 21750) (T_LOGICAL_AND 21751) (T_LOGICAL_OR 21752) (T_LOGICAL_XOR 21753) (T_MATCH 21754) (T_METHOD_C 21755) (T_NAMESPACE 21756) (T_NEW 21757) (T_NS_C 21758) (T_PRINT 21759) (T_PRIVATE 21760) (T_PROPERTY_C 21761) (T_PROTECTED 21762) (T_PUBLIC 21763) (T_READONLY 21764) (T_REQUIRE 21765) (T_REQUIRE_ONCE 21766) (T_RETURN 21767) (T_STATIC 21768) (T_STRING 21769) (T_SWITCH 21770) (T_THROW 21771) (T_TRAIT 21772) (T_TRAIT_C 21773) (T_TRY 21774) (T_UNSET 21775) (T_USE 21776) (T_VAR 21777) (T_VARIABLE 21570) (T_WHILE 21778) (T_YIELD 21779) (identifier 21780) (member_name 21781) (reserved_non_modifiers 21782) (semi_reserved 21783) (simple_variable 21784) ("{" 21785)) 9741 (("(" 21788) (argument_list 21789)) 9742 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 21786) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 9743 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 21787)) 9744 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 21790) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 21791) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 21792) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 9745 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 21795) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 9746 (("," 1498) (possible_comma 21793)) 9747 ((")" 21794)) 9748 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (ampersand 21825) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21826) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9749 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21824) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9750 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21823) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9751 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21822) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9752 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21821) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9753 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21820) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9754 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21819) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9755 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21818) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9756 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21817) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9757 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21816) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9758 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21815) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9759 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21814) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9760 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21813) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9761 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21812) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9762 (("\"" 1538) ("$" 21827) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 21828) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 21829) ("[" 1560) (array_object_dereferenceable 21830) (callable_expr 21831) (callable_variable 21832) (class_constant 1564) (class_name 21833) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 21834) (name 21835) (new_dereferenceable 1570) (simple_variable 21836) (static_member 21837) (variable 21838) (variable_class_name 21839)) 9763 (("$" 21827) (T_VARIABLE 21829) (simple_variable 21872) ("{" 21873)) 9764 (("(" 21843) (argument_list 21871)) 9765 ((T_NULLSAFE_OBJECT_OPERATOR 21857) (T_OBJECT_OPERATOR 21858) ("[" 21859)) 9766 (("(" 21843) (argument_list 21856)) 9767 ((T_PAAMAYIM_NEKUDOTAYIM 21852)) 9768 (("(" 21843) (argument_list 21851)) 9769 ((T_PAAMAYIM_NEKUDOTAYIM 21840)) 9770 (("$" 21827) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 21829) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 21841) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 21842) ("{" 1660)) 9771 (("(" 21843) (argument_list 21844)) 9772 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 21845) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 21846) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 21847) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 9773 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 21850) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 9774 (("," 1498) (possible_comma 21848)) 9775 ((")" 21849)) 9776 (("$" 21827) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 21829) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 21853) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 21854) ("{" 1676)) 9777 (("(" 21843) (argument_list 21855)) 9778 (("$" 21827) (T_STRING 21862) (T_VARIABLE 21829) (property_name 21869) (simple_variable 21864) ("{" 21865)) 9779 (("$" 21827) (T_STRING 21862) (T_VARIABLE 21829) (property_name 21863) (simple_variable 21864) ("{" 21865)) 9780 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 21860) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 9781 (("]" 21861)) 9782 (("(" 21843) (argument_list 21868)) 9783 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 21866) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 9784 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 21867)) 9785 (("(" 21843) (argument_list 21870)) 9786 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 21874) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 9787 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 21875)) 9788 ((backup_doc_comment 21878)) 9789 (("(" 21879)) 9790 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 21880)) 9791 ((")" 21881)) 9792 ((T_USE 166) (lexical_vars 21882)) 9793 ((":" 168) (return_type 21883)) 9794 ((backup_fn_flags 21884)) 9795 (("{" 21885)) 9796 ((inner_statement_list 21886)) 9797 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 21887) ("~" 129)) 9798 ((backup_fn_flags 21888)) 9799 ((backup_doc_comment 21890)) 9800 (("(" 21891)) 9801 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 21892)) 9802 ((")" 21893)) 9803 ((":" 2271) (return_type 21894)) 9804 ((T_DOUBLE_ARROW 21895)) 9805 ((backup_fn_flags 21896)) 9806 ((backup_lex_pos 21897)) 9807 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21898) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9808 (("%" 21600) ("*" 21601) ("+" 21602) ("-" 21603) ("." 21604) ("/" 21605) ("<" 21606) (">" 21607) ("?" 21608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 21609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 21610) (T_BOOLEAN_AND 21611) (T_BOOLEAN_OR 21612) (T_COALESCE 21613) (T_INSTANCEOF 21614) (T_IS_EQUAL 21615) (T_IS_GREATER_OR_EQUAL 21616) (T_IS_IDENTICAL 21617) (T_IS_NOT_EQUAL 21618) (T_IS_NOT_IDENTICAL 21619) (T_IS_SMALLER_OR_EQUAL 21620) (T_LOGICAL_AND 21621) (T_LOGICAL_OR 21622) (T_LOGICAL_XOR 21623) (T_PIPE 21624) (T_POW 21625) (T_SL 21626) (T_SPACESHIP 21627) (T_SR 21628) ("^" 21629) (backup_fn_flags 21899) ("|" 21630)) 9809 (("$" 21522) (T_ABSTRACT 21701) (T_ARRAY 21702) (T_AS 21703) (T_BREAK 21704) (T_CALLABLE 21705) (T_CASE 21706) (T_CATCH 21707) (T_CLASS 21708) (T_CLASS_C 21709) (T_CLONE 21710) (T_CONST 21711) (T_CONTINUE 21712) (T_DECLARE 21713) (T_DEFAULT 21714) (T_DIR 21715) (T_DO 21716) (T_ECHO 21717) (T_ELSE 21718) (T_ELSEIF 21719) (T_EMPTY 21720) (T_ENDDECLARE 21721) (T_ENDFOR 21722) (T_ENDFOREACH 21723) (T_ENDIF 21724) (T_ENDSWITCH 21725) (T_ENDWHILE 21726) (T_ENUM 21727) (T_EVAL 21728) (T_EXIT 21729) (T_EXTENDS 21730) (T_FILE 21731) (T_FINAL 21732) (T_FINALLY 21733) (T_FN 21734) (T_FOR 21735) (T_FOREACH 21736) (T_FUNCTION 21737) (T_FUNC_C 21738) (T_GLOBAL 21739) (T_GOTO 21740) (T_IF 21741) (T_IMPLEMENTS 21742) (T_INCLUDE 21743) (T_INCLUDE_ONCE 21744) (T_INSTANCEOF 21745) (T_INSTEADOF 21746) (T_INTERFACE 21747) (T_ISSET 21748) (T_LINE 21749) (T_LIST 21750) (T_LOGICAL_AND 21751) (T_LOGICAL_OR 21752) (T_LOGICAL_XOR 21753) (T_MATCH 21754) (T_METHOD_C 21755) (T_NAMESPACE 21756) (T_NEW 21757) (T_NS_C 21758) (T_PRINT 21759) (T_PRIVATE 21760) (T_PROPERTY_C 21761) (T_PROTECTED 21762) (T_PUBLIC 21763) (T_READONLY 21764) (T_REQUIRE 21765) (T_REQUIRE_ONCE 21766) (T_RETURN 21767) (T_STATIC 21768) (T_STRING 21769) (T_SWITCH 21770) (T_THROW 21771) (T_TRAIT 21772) (T_TRAIT_C 21773) (T_TRY 21774) (T_UNSET 21775) (T_USE 21776) (T_VAR 21777) (T_VARIABLE 21570) (T_WHILE 21778) (T_YIELD 21779) (identifier 21902) (member_name 21903) (reserved_non_modifiers 21782) (semi_reserved 21783) (simple_variable 21904) ("{" 21905)) 9810 (("(" 21788) (argument_list 21908)) 9811 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 21906) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 9812 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 21907)) 9813 ((T_FN 35) (T_FUNCTION 131) (fn 21584) #'21585 (inline_function 21912)) 9814 (("$" 21522) (T_STRING 21918) (T_VARIABLE 21570) (property_name 21925) (simple_variable 21920) ("{" 21921)) 9815 (("$" 21522) (T_STRING 21918) (T_VARIABLE 21570) (property_name 21919) (simple_variable 21920) ("{" 21921)) 9816 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 21916) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 9817 (("]" 21917)) 9818 (("(" 21788) (argument_list 21924)) 9819 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 21922) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 9820 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 21923)) 9821 (("(" 21788) (argument_list 21926)) 9822 (("`" 21928)) 9823 (("]" 21930)) 9824 (("=" 21931)) 9825 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21932) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9826 (("%" 4045) ("*" 4046) ("+" 4047) ("-" 4048) ("." 4049) ("/" 4050) ("<" 4051) (">" 4052) ("?" 4053) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 4054) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 4055) (T_BOOLEAN_AND 4056) (T_BOOLEAN_OR 4057) (T_COALESCE 4058) (T_DOUBLE_ARROW 21935) (T_INSTANCEOF 4059) (T_IS_EQUAL 4060) (T_IS_GREATER_OR_EQUAL 4061) (T_IS_IDENTICAL 4062) (T_IS_NOT_EQUAL 4063) (T_IS_NOT_IDENTICAL 4064) (T_IS_SMALLER_OR_EQUAL 4065) (T_LOGICAL_AND 4066) (T_LOGICAL_OR 4067) (T_LOGICAL_XOR 4068) (T_PIPE 4069) (T_POW 4070) (T_SL 4071) (T_SPACESHIP 4072) (T_SR 4073) ("^" 4074) ("|" 4075)) 9827 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21936) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9828 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 21945) (T_VARIABLE 2920) (encaps_var 2934)) 9829 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 21944) (T_VARIABLE 2920) (encaps_var 2925)) 9830 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 21986) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 9831 ((T_CLASS 21978)) 9832 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 21977) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 21958) (attribute 2982) (class_modifier 2963)) 9833 ((T_PAAMAYIM_NEKUDOTAYIM 21975)) 9834 (("(" 21843) (argument_list 21974)) 9835 ((T_NULLSAFE_OBJECT_OPERATOR 21965) (T_OBJECT_OPERATOR 21966) (T_PAAMAYIM_NEKUDOTAYIM 21967) ("[" 21968)) 9836 (("$" 21827) (T_STRING 21862) (T_VARIABLE 21829) (property_name 21973) (simple_variable 21864) ("{" 21865)) 9837 (("$" 21827) (T_STRING 21862) (T_VARIABLE 21829) (property_name 21972) (simple_variable 21864) ("{" 21865)) 9838 (("$" 21827) (T_VARIABLE 21829) (simple_variable 21971)) 9839 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 21969) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 9840 (("]" 21970)) 9841 (("$" 21827) (T_VARIABLE 21829) (simple_variable 21976)) 9842 (("(" 2984) (argument_list 2985) (ctor_arguments 21979)) 9843 ((T_EXTENDS 2987) (extends_from 21980)) 9844 ((T_IMPLEMENTS 2989) (implements_list 21981)) 9845 ((backup_doc_comment 21982)) 9846 (("{" 21983)) 9847 ((class_statement_list 21984)) 9848 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 21985)) 9849 (("%" 304) (")" 21987) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 9850 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 21989) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 9851 (("%" 304) (")" 21990) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 9852 (("{" 21991)) 9853 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 21992) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 9854 (("}" 21993)) 9855 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 21995) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 9856 ((")" 21996)) 9857 (("=" 21997)) 9858 (("!" 21520) ("\"" 21521) ("$" 21522) ("(" 21523) ("+" 21524) ("-" 21525) ("@" 21526) (T_ARRAY 21527) (T_ARRAY_CAST 21528) (T_ATTRIBUTE 130) (T_BOOL_CAST 21529) (T_CLASS_C 21530) (T_CLONE 21531) (T_CONSTANT_ENCAPSED_STRING 21532) (T_DEC 21533) (T_DIR 21534) (T_DNUMBER 21535) (T_DOUBLE_CAST 21536) (T_EMPTY 21537) (T_EVAL 21538) (T_EXIT 21539) (T_FILE 21540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21541) (T_INC 21542) (T_INCLUDE 21543) (T_INCLUDE_ONCE 21544) (T_INT_CAST 21545) (T_ISSET 21546) (T_LINE 21547) (T_LIST 21548) (T_LNUMBER 21549) (T_MATCH 21550) (T_METHOD_C 21551) (T_NAME_FULLY_QUALIFIED 21552) (T_NAME_QUALIFIED 21553) (T_NAME_RELATIVE 21554) (T_NEW 21555) (T_NS_C 21556) (T_OBJECT_CAST 21557) (T_PRINT 21558) (T_PROPERTY_C 21559) (T_READONLY 21560) (T_REQUIRE 21561) (T_REQUIRE_ONCE 21562) (T_START_HEREDOC 21563) (T_STATIC 21564) (T_STRING 21565) (T_STRING_CAST 21566) (T_THROW 21567) (T_TRAIT_C 21568) (T_UNSET_CAST 21569) (T_VARIABLE 21570) (T_YIELD 21571) (T_YIELD_FROM 21572) ("[" 21573) ("`" 21574) (array_object_dereferenceable 21575) (attribute 135) (attributes 21576) (callable_expr 21577) (callable_variable 21578) (class_constant 21579) (class_name 21580) (constant 21581) (dereferenceable_scalar 21582) (expr 21998) (fn 21584) (fully_dereferenceable 107) #'21585 (function_call 21586) (inline_function 21587) (internal_functions_in_yacc 21588) (match 21589) (name 21590) (new_dereferenceable 21591) (new_non_dereferenceable 21592) (scalar 21593) (simple_variable 21594) (static_member 21595) (variable 21596) (variable_class_name 21597) ("~" 21598)) 9859 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 22000) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 9860 (("," 5226) (possible_comma 22001)) 9861 ((")" 22002)) 9862 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 22010) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 22011) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 22012) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 9863 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 22015) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 9864 (("," 1498) (possible_comma 22013)) 9865 ((")" 22014)) 9866 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 22017) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 9867 (("%" 304) (")" 22018) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 9868 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 22020) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 9869 (("%" 304) (")" 22021) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 9870 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 22027) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 22028) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 22029) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 22030) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 9871 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 22036) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 9872 (("%" 2442) (")" 22033) ("*" 2443) ("+" 2444) ("," 22034) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 9873 (("," 5287) (possible_comma 22031)) 9874 ((")" 22032)) 9875 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 22035) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 9876 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 22040) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 9877 ((")" 22041)) 9878 (("%" 304) (")" 22033) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 9879 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 22048) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 9880 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 22049)) 9881 (("\"" 22051) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 9882 (("$" 21144) (T_ABSTRACT 21323) (T_ARRAY 21324) (T_AS 21325) (T_BREAK 21326) (T_CALLABLE 21327) (T_CASE 21328) (T_CATCH 21329) (T_CLASS 21330) (T_CLASS_C 21331) (T_CLONE 21332) (T_CONST 21333) (T_CONTINUE 21334) (T_DECLARE 21335) (T_DEFAULT 21336) (T_DIR 21337) (T_DO 21338) (T_ECHO 21339) (T_ELSE 21340) (T_ELSEIF 21341) (T_EMPTY 21342) (T_ENDDECLARE 21343) (T_ENDFOR 21344) (T_ENDFOREACH 21345) (T_ENDIF 21346) (T_ENDSWITCH 21347) (T_ENDWHILE 21348) (T_ENUM 21349) (T_EVAL 21350) (T_EXIT 21351) (T_EXTENDS 21352) (T_FILE 21353) (T_FINAL 21354) (T_FINALLY 21355) (T_FN 21356) (T_FOR 21357) (T_FOREACH 21358) (T_FUNCTION 21359) (T_FUNC_C 21360) (T_GLOBAL 21361) (T_GOTO 21362) (T_IF 21363) (T_IMPLEMENTS 21364) (T_INCLUDE 21365) (T_INCLUDE_ONCE 21366) (T_INSTANCEOF 21367) (T_INSTEADOF 21368) (T_INTERFACE 21369) (T_ISSET 21370) (T_LINE 21371) (T_LIST 21372) (T_LOGICAL_AND 21373) (T_LOGICAL_OR 21374) (T_LOGICAL_XOR 21375) (T_MATCH 21376) (T_METHOD_C 21377) (T_NAMESPACE 21378) (T_NEW 21379) (T_NS_C 21380) (T_PRINT 21381) (T_PRIVATE 21382) (T_PROPERTY_C 21383) (T_PROTECTED 21384) (T_PUBLIC 21385) (T_READONLY 21386) (T_REQUIRE 21387) (T_REQUIRE_ONCE 21388) (T_RETURN 21389) (T_STATIC 21390) (T_STRING 21391) (T_SWITCH 21392) (T_THROW 21393) (T_TRAIT 21394) (T_TRAIT_C 21395) (T_TRY 21396) (T_UNSET 21397) (T_USE 21398) (T_VAR 21399) (T_VARIABLE 21192) (T_WHILE 21400) (T_YIELD 21401) (identifier 22054) (member_name 22055) (reserved_non_modifiers 21404) (semi_reserved 21405) (simple_variable 22056) ("{" 22057)) 9883 (("(" 21410) (argument_list 22060)) 9884 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 22058) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 9885 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 22059)) 9886 ((T_FN 35) (T_FUNCTION 131) (fn 21206) #'21207 (inline_function 22064)) 9887 (("$" 21144) (T_STRING 22070) (T_VARIABLE 21192) (property_name 22077) (simple_variable 22072) ("{" 22073)) 9888 (("$" 21144) (T_STRING 22070) (T_VARIABLE 21192) (property_name 22071) (simple_variable 22072) ("{" 22073)) 9889 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 22068) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 9890 (("]" 22069)) 9891 (("(" 21410) (argument_list 22076)) 9892 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 22074) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 9893 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 22075)) 9894 (("(" 21410) (argument_list 22078)) 9895 (("`" 22080)) 9896 (("]" 22082)) 9897 (("=" 22083)) 9898 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 22084) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 9899 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22619) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9900 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 22617) (encaps_var 5357)) 9901 (("$" 22088) (T_VARIABLE 22136) (simple_variable 22613) ("{" 22614)) 9902 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 22612) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 9903 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22611) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9904 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22610) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9905 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22609) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9906 (("(" 22606)) 9907 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22605) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9908 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22604) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9909 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22591) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (clone_argument_list 22592) (constant 22147) (dereferenceable_scalar 22148) (expr 22593) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9910 (("\"" 1538) ("$" 22393) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 22394) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 22395) ("[" 1560) (array_object_dereferenceable 22396) (callable_expr 22397) (callable_variable 22398) (class_constant 1564) (class_name 22399) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 22400) (name 22401) (new_dereferenceable 1570) (simple_variable 22402) (static_member 22403) (variable 22590) (variable_class_name 22405)) 9911 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22589) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9912 (("(" 22586)) 9913 (("(" 22583)) 9914 (("(" 22574) (argument_list 22575) (ctor_arguments 22576)) 9915 (("\"" 1538) ("$" 22393) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 22394) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 22395) ("[" 1560) (array_object_dereferenceable 22396) (callable_expr 22397) (callable_variable 22398) (class_constant 1564) (class_name 22399) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 22400) (name 22401) (new_dereferenceable 1570) (simple_variable 22402) (static_member 22403) (variable 22573) (variable_class_name 22405)) 9916 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22572) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9917 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22571) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9918 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22570) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9919 (("(" 22566)) 9920 (("(" 22561)) 9921 (("(" 22555)) 9922 (("$" 22393) ("(" 22518) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 22519) (T_NAME_QUALIFIED 22520) (T_NAME_RELATIVE 22521) (T_READONLY 2955) (T_STATIC 22522) (T_STRING 22523) (T_VARIABLE 22395) (anonymous_class 22524) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 22525) (attribute 2961) (attributes 22526) (class_modifier 2963) (class_name 22527) (class_name_reference 22528) (name 22529) (new_variable 22530) (simple_variable 22531)) 9923 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22517) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9924 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22516) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9925 (("(" 22354) (argument_list 22515)) 9926 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22514) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9927 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22513) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9928 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 22508) (T_END_HEREDOC 22509) (T_VARIABLE 2920) (encaps_list 22510) (encaps_var 2922)) 9929 ((T_FN 35) (T_FUNCTION 131) (fn 22150) #'22151 (inline_function 22507)) 9930 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22506) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9931 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22505) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9932 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22504) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9933 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22501) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9934 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22500) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9935 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 22496) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 9936 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 22494) (encaps_list 2828) (encaps_var 2829)) 9937 ((T_NULLSAFE_OBJECT_OPERATOR 22480) (T_OBJECT_OPERATOR 22481) ("[" 22482)) 9938 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 22477) (attribute 2757) (fn 22150) #'22151 (inline_function 22478)) 9939 (("(" 22354) (argument_list 22476)) 9940 ((T_PAAMAYIM_NEKUDOTAYIM 22468)) 9941 (("%" 22166) ("*" 22167) ("+" 22168) ("-" 22169) ("." 22170) ("/" 22171) ("<" 22172) (">" 22173) ("?" 22174) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 22175) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 22176) (T_BOOLEAN_AND 22177) (T_BOOLEAN_OR 22178) (T_COALESCE 22179) (T_DOUBLE_ARROW 22466) (T_INSTANCEOF 22180) (T_IS_EQUAL 22181) (T_IS_GREATER_OR_EQUAL 22182) (T_IS_IDENTICAL 22183) (T_IS_NOT_EQUAL 22184) (T_IS_NOT_IDENTICAL 22185) (T_IS_SMALLER_OR_EQUAL 22186) (T_LOGICAL_AND 22187) (T_LOGICAL_OR 22188) (T_LOGICAL_XOR 22189) (T_PIPE 22190) (T_POW 22191) (T_SL 22192) (T_SPACESHIP 22193) (T_SR 22194) ("^" 22195) ("|" 22196)) 9942 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 22455)) 9943 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 22443)) 9944 (("(" 22354) (argument_list 22442)) 9945 (("=" 22362) (T_AND_EQUAL 22363) (T_COALESCE_EQUAL 22364) (T_CONCAT_EQUAL 22365) (T_DEC 22366) (T_DIV_EQUAL 22367) (T_INC 22368) (T_MINUS_EQUAL 22369) (T_MOD_EQUAL 22370) (T_MUL_EQUAL 22371) (T_OR_EQUAL 22372) (T_PLUS_EQUAL 22373) (T_POW_EQUAL 22374) (T_SL_EQUAL 22375) (T_SR_EQUAL 22376) (T_XOR_EQUAL 22377)) 9946 ((T_PAAMAYIM_NEKUDOTAYIM 22266)) 9947 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22165) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9948 (("%" 22166) ("*" 22167) ("+" 22168) ("-" 22169) ("." 22170) ("/" 22171) ("<" 22172) (">" 22173) ("?" 22174) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 22175) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 22176) (T_BOOLEAN_AND 22177) (T_BOOLEAN_OR 22178) (T_COALESCE 22179) (T_INSTANCEOF 22180) (T_IS_EQUAL 22181) (T_IS_GREATER_OR_EQUAL 22182) (T_IS_IDENTICAL 22183) (T_IS_NOT_EQUAL 22184) (T_IS_NOT_IDENTICAL 22185) (T_IS_SMALLER_OR_EQUAL 22186) (T_LOGICAL_AND 22187) (T_LOGICAL_OR 22188) (T_LOGICAL_XOR 22189) (T_PIPE 22190) (T_POW 22191) (T_SL 22192) (T_SPACESHIP 22193) (T_SR 22194) ("^" 22195) ("|" 22196)) 9949 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22265) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9950 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22264) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9951 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22263) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9952 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22262) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9953 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22261) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9954 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22260) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9955 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22259) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9956 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22258) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9957 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 22253) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 22254) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 9958 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22252) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9959 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22251) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9960 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22250) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9961 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22249) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9962 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22248) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9963 (("$" 22213) ("(" 22214) (T_NAME_FULLY_QUALIFIED 22215) (T_NAME_QUALIFIED 22216) (T_NAME_RELATIVE 22217) (T_STATIC 22218) (T_STRING 22219) (T_VARIABLE 22220) (class_name 22221) (class_name_reference 22222) (name 22223) (new_variable 22224) (simple_variable 22225)) 9964 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22212) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9965 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22211) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9966 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22210) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9967 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22209) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9968 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22208) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9969 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22207) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9970 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22206) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9971 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22205) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9972 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22204) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9973 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22203) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9974 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22202) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9975 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22201) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9976 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22200) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9977 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22199) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9978 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22198) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9979 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22197) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9980 (("$" 22213) (T_VARIABLE 22220) (simple_variable 22244) ("{" 22245)) 9981 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 22242) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 9982 ((T_PAAMAYIM_NEKUDOTAYIM 22240)) 9983 ((T_NULLSAFE_OBJECT_OPERATOR 22226) (T_OBJECT_OPERATOR 22227) (T_PAAMAYIM_NEKUDOTAYIM 22228) ("[" 22229)) 9984 (("$" 22213) (T_STRING 22233) (T_VARIABLE 22220) (property_name 22239) (simple_variable 22235) ("{" 22236)) 9985 (("$" 22213) (T_STRING 22233) (T_VARIABLE 22220) (property_name 22234) (simple_variable 22235) ("{" 22236)) 9986 (("$" 22213) (T_VARIABLE 22220) (simple_variable 22232)) 9987 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 22230) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 9988 (("]" 22231)) 9989 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 22237) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 9990 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 22238)) 9991 (("$" 22213) (T_VARIABLE 22220) (simple_variable 22241)) 9992 (("%" 304) (")" 22243) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 9993 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 22246) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 9994 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 22247)) 9995 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22257) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9996 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 22255) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 9997 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22256) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 9998 (("$" 22088) (T_ABSTRACT 22267) (T_ARRAY 22268) (T_AS 22269) (T_BREAK 22270) (T_CALLABLE 22271) (T_CASE 22272) (T_CATCH 22273) (T_CLASS 22274) (T_CLASS_C 22275) (T_CLONE 22276) (T_CONST 22277) (T_CONTINUE 22278) (T_DECLARE 22279) (T_DEFAULT 22280) (T_DIR 22281) (T_DO 22282) (T_ECHO 22283) (T_ELSE 22284) (T_ELSEIF 22285) (T_EMPTY 22286) (T_ENDDECLARE 22287) (T_ENDFOR 22288) (T_ENDFOREACH 22289) (T_ENDIF 22290) (T_ENDSWITCH 22291) (T_ENDWHILE 22292) (T_ENUM 22293) (T_EVAL 22294) (T_EXIT 22295) (T_EXTENDS 22296) (T_FILE 22297) (T_FINAL 22298) (T_FINALLY 22299) (T_FN 22300) (T_FOR 22301) (T_FOREACH 22302) (T_FUNCTION 22303) (T_FUNC_C 22304) (T_GLOBAL 22305) (T_GOTO 22306) (T_IF 22307) (T_IMPLEMENTS 22308) (T_INCLUDE 22309) (T_INCLUDE_ONCE 22310) (T_INSTANCEOF 22311) (T_INSTEADOF 22312) (T_INTERFACE 22313) (T_ISSET 22314) (T_LINE 22315) (T_LIST 22316) (T_LOGICAL_AND 22317) (T_LOGICAL_OR 22318) (T_LOGICAL_XOR 22319) (T_MATCH 22320) (T_METHOD_C 22321) (T_NAMESPACE 22322) (T_NEW 22323) (T_NS_C 22324) (T_PRINT 22325) (T_PRIVATE 22326) (T_PROPERTY_C 22327) (T_PROTECTED 22328) (T_PUBLIC 22329) (T_READONLY 22330) (T_REQUIRE 22331) (T_REQUIRE_ONCE 22332) (T_RETURN 22333) (T_STATIC 22334) (T_STRING 22335) (T_SWITCH 22336) (T_THROW 22337) (T_TRAIT 22338) (T_TRAIT_C 22339) (T_TRY 22340) (T_UNSET 22341) (T_USE 22342) (T_VAR 22343) (T_VARIABLE 22136) (T_WHILE 22344) (T_YIELD 22345) (identifier 22346) (member_name 22347) (reserved_non_modifiers 22348) (semi_reserved 22349) (simple_variable 22350) ("{" 22351)) 9999 (("(" 22354) (argument_list 22355)) 10000 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 22352) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 10001 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 22353)) 10002 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 22356) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 22357) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 22358) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 10003 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 22361) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 10004 (("," 1498) (possible_comma 22359)) 10005 ((")" 22360)) 10006 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (ampersand 22391) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22392) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 10007 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22390) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 10008 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22389) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 10009 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22388) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 10010 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22387) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 10011 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22386) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 10012 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22385) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 10013 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22384) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 10014 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22383) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 10015 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22382) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 10016 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22381) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 10017 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22380) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 10018 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22379) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 10019 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22378) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 10020 (("\"" 1538) ("$" 22393) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 22394) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 22395) ("[" 1560) (array_object_dereferenceable 22396) (callable_expr 22397) (callable_variable 22398) (class_constant 1564) (class_name 22399) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 22400) (name 22401) (new_dereferenceable 1570) (simple_variable 22402) (static_member 22403) (variable 22404) (variable_class_name 22405)) 10021 (("$" 22393) (T_VARIABLE 22395) (simple_variable 22438) ("{" 22439)) 10022 (("(" 22409) (argument_list 22437)) 10023 ((T_NULLSAFE_OBJECT_OPERATOR 22423) (T_OBJECT_OPERATOR 22424) ("[" 22425)) 10024 (("(" 22409) (argument_list 22422)) 10025 ((T_PAAMAYIM_NEKUDOTAYIM 22418)) 10026 (("(" 22409) (argument_list 22417)) 10027 ((T_PAAMAYIM_NEKUDOTAYIM 22406)) 10028 (("$" 22393) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 22395) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 22407) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 22408) ("{" 1660)) 10029 (("(" 22409) (argument_list 22410)) 10030 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 22411) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 22412) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 22413) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 10031 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 22416) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 10032 (("," 1498) (possible_comma 22414)) 10033 ((")" 22415)) 10034 (("$" 22393) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 22395) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 22419) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 22420) ("{" 1676)) 10035 (("(" 22409) (argument_list 22421)) 10036 (("$" 22393) (T_STRING 22428) (T_VARIABLE 22395) (property_name 22435) (simple_variable 22430) ("{" 22431)) 10037 (("$" 22393) (T_STRING 22428) (T_VARIABLE 22395) (property_name 22429) (simple_variable 22430) ("{" 22431)) 10038 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 22426) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 10039 (("]" 22427)) 10040 (("(" 22409) (argument_list 22434)) 10041 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 22432) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 10042 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 22433)) 10043 (("(" 22409) (argument_list 22436)) 10044 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 22440) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 10045 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 22441)) 10046 ((backup_doc_comment 22444)) 10047 (("(" 22445)) 10048 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 22446)) 10049 ((")" 22447)) 10050 ((T_USE 166) (lexical_vars 22448)) 10051 ((":" 168) (return_type 22449)) 10052 ((backup_fn_flags 22450)) 10053 (("{" 22451)) 10054 ((inner_statement_list 22452)) 10055 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 22453) ("~" 129)) 10056 ((backup_fn_flags 22454)) 10057 ((backup_doc_comment 22456)) 10058 (("(" 22457)) 10059 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 22458)) 10060 ((")" 22459)) 10061 ((":" 2271) (return_type 22460)) 10062 ((T_DOUBLE_ARROW 22461)) 10063 ((backup_fn_flags 22462)) 10064 ((backup_lex_pos 22463)) 10065 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 22464) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 10066 (("%" 4045) ("*" 4046) ("+" 4047) ("-" 4048) ("." 4049) ("/" 4050) ("<" 4051) (">" 4052) ("?" 4053) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 4054) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 4055) (T_BOOLEAN_AND 4056) (T_BOOLEAN_OR 4057) (T_COALESCE 4058) (T_INSTANCEOF 4059) (T_IS_EQUAL 4060) (T_IS_GREATER_OR_EQUAL 4061) (T_IS_IDENTICAL 4062) (T_IS_NOT_EQUAL 4063) (T_IS_NOT_IDENTICAL 4064) (T_IS_SMALLER_OR_EQUAL 4065) (T_LOGICAL_AND 4066) (T_LOGICAL_OR 4067) (T_LOGICAL_XOR 4068) (T_PIPE 4069) (T_POW 4070) (T_SL 4071) (T_SPACESHIP 4072) (T_SR 4073) ("^" 4074) (backup_fn_flags 22465) ("|" 4075)) 10067 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 22467) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 10068 (("$" 22088) (T_ABSTRACT 22267) (T_ARRAY 22268) (T_AS 22269) (T_BREAK 22270) (T_CALLABLE 22271) (T_CASE 22272) (T_CATCH 22273) (T_CLASS 22274) (T_CLASS_C 22275) (T_CLONE 22276) (T_CONST 22277) (T_CONTINUE 22278) (T_DECLARE 22279) (T_DEFAULT 22280) (T_DIR 22281) (T_DO 22282) (T_ECHO 22283) (T_ELSE 22284) (T_ELSEIF 22285) (T_EMPTY 22286) (T_ENDDECLARE 22287) (T_ENDFOR 22288) (T_ENDFOREACH 22289) (T_ENDIF 22290) (T_ENDSWITCH 22291) (T_ENDWHILE 22292) (T_ENUM 22293) (T_EVAL 22294) (T_EXIT 22295) (T_EXTENDS 22296) (T_FILE 22297) (T_FINAL 22298) (T_FINALLY 22299) (T_FN 22300) (T_FOR 22301) (T_FOREACH 22302) (T_FUNCTION 22303) (T_FUNC_C 22304) (T_GLOBAL 22305) (T_GOTO 22306) (T_IF 22307) (T_IMPLEMENTS 22308) (T_INCLUDE 22309) (T_INCLUDE_ONCE 22310) (T_INSTANCEOF 22311) (T_INSTEADOF 22312) (T_INTERFACE 22313) (T_ISSET 22314) (T_LINE 22315) (T_LIST 22316) (T_LOGICAL_AND 22317) (T_LOGICAL_OR 22318) (T_LOGICAL_XOR 22319) (T_MATCH 22320) (T_METHOD_C 22321) (T_NAMESPACE 22322) (T_NEW 22323) (T_NS_C 22324) (T_PRINT 22325) (T_PRIVATE 22326) (T_PROPERTY_C 22327) (T_PROTECTED 22328) (T_PUBLIC 22329) (T_READONLY 22330) (T_REQUIRE 22331) (T_REQUIRE_ONCE 22332) (T_RETURN 22333) (T_STATIC 22334) (T_STRING 22335) (T_SWITCH 22336) (T_THROW 22337) (T_TRAIT 22338) (T_TRAIT_C 22339) (T_TRY 22340) (T_UNSET 22341) (T_USE 22342) (T_VAR 22343) (T_VARIABLE 22136) (T_WHILE 22344) (T_YIELD 22345) (identifier 22469) (member_name 22470) (reserved_non_modifiers 22348) (semi_reserved 22349) (simple_variable 22471) ("{" 22472)) 10069 (("(" 22354) (argument_list 22475)) 10070 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 22473) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 10071 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 22474)) 10072 ((T_FN 35) (T_FUNCTION 131) (fn 22150) #'22151 (inline_function 22479)) 10073 (("$" 22088) (T_STRING 22485) (T_VARIABLE 22136) (property_name 22492) (simple_variable 22487) ("{" 22488)) 10074 (("$" 22088) (T_STRING 22485) (T_VARIABLE 22136) (property_name 22486) (simple_variable 22487) ("{" 22488)) 10075 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 22483) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 10076 (("]" 22484)) 10077 (("(" 22354) (argument_list 22491)) 10078 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 22489) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 10079 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 22490)) 10080 (("(" 22354) (argument_list 22493)) 10081 (("`" 22495)) 10082 (("]" 22497)) 10083 (("=" 22498)) 10084 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22499) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 10085 (("%" 22166) ("*" 22167) ("+" 22168) ("-" 22169) ("." 22170) ("/" 22171) ("<" 22172) (">" 22173) ("?" 22174) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 22175) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 22176) (T_BOOLEAN_AND 22177) (T_BOOLEAN_OR 22178) (T_COALESCE 22179) (T_DOUBLE_ARROW 22502) (T_INSTANCEOF 22180) (T_IS_EQUAL 22181) (T_IS_GREATER_OR_EQUAL 22182) (T_IS_IDENTICAL 22183) (T_IS_NOT_EQUAL 22184) (T_IS_NOT_IDENTICAL 22185) (T_IS_SMALLER_OR_EQUAL 22186) (T_LOGICAL_AND 22187) (T_LOGICAL_OR 22188) (T_LOGICAL_XOR 22189) (T_PIPE 22190) (T_POW 22191) (T_SL 22192) (T_SPACESHIP 22193) (T_SR 22194) ("^" 22195) ("|" 22196)) 10086 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22503) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 10087 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 22512) (T_VARIABLE 2920) (encaps_var 2934)) 10088 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 22511) (T_VARIABLE 2920) (encaps_var 2925)) 10089 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 22553) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 10090 ((T_CLASS 22545)) 10091 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 22544) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 22525) (attribute 2982) (class_modifier 2963)) 10092 ((T_PAAMAYIM_NEKUDOTAYIM 22542)) 10093 (("(" 22409) (argument_list 22541)) 10094 ((T_NULLSAFE_OBJECT_OPERATOR 22532) (T_OBJECT_OPERATOR 22533) (T_PAAMAYIM_NEKUDOTAYIM 22534) ("[" 22535)) 10095 (("$" 22393) (T_STRING 22428) (T_VARIABLE 22395) (property_name 22540) (simple_variable 22430) ("{" 22431)) 10096 (("$" 22393) (T_STRING 22428) (T_VARIABLE 22395) (property_name 22539) (simple_variable 22430) ("{" 22431)) 10097 (("$" 22393) (T_VARIABLE 22395) (simple_variable 22538)) 10098 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 22536) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 10099 (("]" 22537)) 10100 (("$" 22393) (T_VARIABLE 22395) (simple_variable 22543)) 10101 (("(" 2984) (argument_list 2985) (ctor_arguments 22546)) 10102 ((T_EXTENDS 2987) (extends_from 22547)) 10103 ((T_IMPLEMENTS 2989) (implements_list 22548)) 10104 ((backup_doc_comment 22549)) 10105 (("{" 22550)) 10106 ((class_statement_list 22551)) 10107 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 22552)) 10108 (("%" 304) (")" 22554) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 10109 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 22556) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 10110 (("%" 304) (")" 22557) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 10111 (("{" 22558)) 10112 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 22559) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 10113 (("}" 22560)) 10114 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 22562) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 10115 ((")" 22563)) 10116 (("=" 22564)) 10117 (("!" 22086) ("\"" 22087) ("$" 22088) ("(" 22089) ("+" 22090) ("-" 22091) ("@" 22092) (T_ARRAY 22093) (T_ARRAY_CAST 22094) (T_ATTRIBUTE 130) (T_BOOL_CAST 22095) (T_CLASS_C 22096) (T_CLONE 22097) (T_CONSTANT_ENCAPSED_STRING 22098) (T_DEC 22099) (T_DIR 22100) (T_DNUMBER 22101) (T_DOUBLE_CAST 22102) (T_EMPTY 22103) (T_EVAL 22104) (T_EXIT 22105) (T_FILE 22106) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 22107) (T_INC 22108) (T_INCLUDE 22109) (T_INCLUDE_ONCE 22110) (T_INT_CAST 22111) (T_ISSET 22112) (T_LINE 22113) (T_LIST 22114) (T_LNUMBER 22115) (T_MATCH 22116) (T_METHOD_C 22117) (T_NAME_FULLY_QUALIFIED 22118) (T_NAME_QUALIFIED 22119) (T_NAME_RELATIVE 22120) (T_NEW 22121) (T_NS_C 22122) (T_OBJECT_CAST 22123) (T_PRINT 22124) (T_PROPERTY_C 22125) (T_READONLY 22126) (T_REQUIRE 22127) (T_REQUIRE_ONCE 22128) (T_START_HEREDOC 22129) (T_STATIC 22130) (T_STRING 22131) (T_STRING_CAST 22132) (T_THROW 22133) (T_TRAIT_C 22134) (T_UNSET_CAST 22135) (T_VARIABLE 22136) (T_YIELD 22137) (T_YIELD_FROM 22138) ("[" 22139) ("`" 22140) (array_object_dereferenceable 22141) (attribute 135) (attributes 22142) (callable_expr 22143) (callable_variable 22144) (class_constant 22145) (class_name 22146) (constant 22147) (dereferenceable_scalar 22148) (expr 22565) (fn 22150) (fully_dereferenceable 107) #'22151 (function_call 22152) (inline_function 22153) (internal_functions_in_yacc 22154) (match 22155) (name 22156) (new_dereferenceable 22157) (new_non_dereferenceable 22158) (scalar 22159) (simple_variable 22160) (static_member 22161) (variable 22162) (variable_class_name 22163) ("~" 22164)) 10118 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 22567) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 10119 (("," 5226) (possible_comma 22568)) 10120 ((")" 22569)) 10121 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 22577) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 22578) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 22579) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 10122 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 22582) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 10123 (("," 1498) (possible_comma 22580)) 10124 ((")" 22581)) 10125 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 22584) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 10126 (("%" 304) (")" 22585) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 10127 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 22587) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 10128 (("%" 304) (")" 22588) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 10129 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 22594) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 22595) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 22596) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 22597) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 10130 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 22603) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 10131 (("%" 2442) (")" 22600) ("*" 2443) ("+" 2444) ("," 22601) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 10132 (("," 5287) (possible_comma 22598)) 10133 ((")" 22599)) 10134 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 22602) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 10135 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 22607) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 10136 ((")" 22608)) 10137 (("%" 304) (")" 22600) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 10138 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 22615) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 10139 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 22616)) 10140 (("\"" 22618) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 10141 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 22628) (T_VARIABLE 2920) (encaps_var 2934)) 10142 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 22627) (T_VARIABLE 2920) (encaps_var 2925)) 10143 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 22669) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 10144 ((T_CLASS 22661)) 10145 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 22660) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 22641) (attribute 2982) (class_modifier 2963)) 10146 ((T_PAAMAYIM_NEKUDOTAYIM 22658)) 10147 (("(" 21465) (argument_list 22657)) 10148 ((T_NULLSAFE_OBJECT_OPERATOR 22648) (T_OBJECT_OPERATOR 22649) (T_PAAMAYIM_NEKUDOTAYIM 22650) ("[" 22651)) 10149 (("$" 21449) (T_STRING 21484) (T_VARIABLE 21451) (property_name 22656) (simple_variable 21486) ("{" 21487)) 10150 (("$" 21449) (T_STRING 21484) (T_VARIABLE 21451) (property_name 22655) (simple_variable 21486) ("{" 21487)) 10151 (("$" 21449) (T_VARIABLE 21451) (simple_variable 22654)) 10152 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 22652) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 10153 (("]" 22653)) 10154 (("$" 21449) (T_VARIABLE 21451) (simple_variable 22659)) 10155 (("(" 2984) (argument_list 2985) (ctor_arguments 22662)) 10156 ((T_EXTENDS 2987) (extends_from 22663)) 10157 ((T_IMPLEMENTS 2989) (implements_list 22664)) 10158 ((backup_doc_comment 22665)) 10159 (("{" 22666)) 10160 ((class_statement_list 22667)) 10161 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 22668)) 10162 (("%" 304) (")" 22670) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 10163 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 22672) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 10164 (("%" 304) (")" 22673) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 10165 (("{" 22674)) 10166 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 22675) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 10167 (("}" 22676)) 10168 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 22678) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 10169 ((")" 22679)) 10170 (("=" 22680)) 10171 (("!" 21142) ("\"" 21143) ("$" 21144) ("(" 21145) ("+" 21146) ("-" 21147) ("@" 21148) (T_ARRAY 21149) (T_ARRAY_CAST 21150) (T_ATTRIBUTE 130) (T_BOOL_CAST 21151) (T_CLASS_C 21152) (T_CLONE 21153) (T_CONSTANT_ENCAPSED_STRING 21154) (T_DEC 21155) (T_DIR 21156) (T_DNUMBER 21157) (T_DOUBLE_CAST 21158) (T_EMPTY 21159) (T_EVAL 21160) (T_EXIT 21161) (T_FILE 21162) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 21163) (T_INC 21164) (T_INCLUDE 21165) (T_INCLUDE_ONCE 21166) (T_INT_CAST 21167) (T_ISSET 21168) (T_LINE 21169) (T_LIST 21170) (T_LNUMBER 21171) (T_MATCH 21172) (T_METHOD_C 21173) (T_NAME_FULLY_QUALIFIED 21174) (T_NAME_QUALIFIED 21175) (T_NAME_RELATIVE 21176) (T_NEW 21177) (T_NS_C 21178) (T_OBJECT_CAST 21179) (T_PRINT 21180) (T_PROPERTY_C 21181) (T_READONLY 21182) (T_REQUIRE 21183) (T_REQUIRE_ONCE 21184) (T_START_HEREDOC 21185) (T_STATIC 21186) (T_STRING 21187) (T_STRING_CAST 21188) (T_THROW 21189) (T_TRAIT_C 21190) (T_UNSET_CAST 21191) (T_VARIABLE 21192) (T_YIELD 21193) (T_YIELD_FROM 21194) ("[" 21195) ("`" 21196) (array_object_dereferenceable 21197) (attribute 135) (attributes 21198) (callable_expr 21199) (callable_variable 21200) (class_constant 21201) (class_name 21202) (constant 21203) (dereferenceable_scalar 21204) (expr 22681) (fn 21206) (fully_dereferenceable 107) #'21207 (function_call 21208) (inline_function 21209) (internal_functions_in_yacc 21210) (match 21211) (name 21212) (new_dereferenceable 21213) (new_non_dereferenceable 21214) (scalar 21215) (simple_variable 21216) (static_member 21217) (variable 21218) (variable_class_name 21219) ("~" 21220)) 10172 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 22683) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 10173 (("," 5226) (possible_comma 22684)) 10174 ((")" 22685)) 10175 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 22693) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 22694) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 22695) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 10176 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 22698) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 10177 (("," 1498) (possible_comma 22696)) 10178 ((")" 22697)) 10179 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 22700) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 10180 (("%" 304) (")" 22701) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 10181 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 22703) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 10182 (("%" 304) (")" 22704) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 10183 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 22710) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 22711) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 22712) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 22713) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 10184 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 22719) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 10185 (("%" 2442) (")" 22716) ("*" 2443) ("+" 2444) ("," 22717) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 10186 (("," 5287) (possible_comma 22714)) 10187 ((")" 22715)) 10188 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 22718) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 10189 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 22723) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 10190 ((")" 22724)) 10191 (("%" 304) (")" 22716) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 10192 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 22731) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 10193 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 22732)) 10194 (("\"" 22734) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 10195 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_VOID_CAST 22737) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 22738) (fn 2425) (for_exprs 22739) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_for_exprs 22740) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 10196 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 23855) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 10197 ((")" 22745)) 10198 (("," 22741)) 10199 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_VOID_CAST 22742) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 22743) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 10200 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 22744) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 10201 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 22746) (";" 21012) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 21013) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21014) (T_DEC 22) (T_DECLARE 21015) (T_DIR 24) (T_DNUMBER 25) (T_DO 21016) (T_DOUBLE_CAST 27) (T_ECHO 21017) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 21018) (T_FOREACH 21019) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 21020) (T_GOTO 21021) (T_IF 20976) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21022) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21023) (T_START_HEREDOC 69) (T_STATIC 21024) (T_STRING 21025) (T_STRING_CAST 72) (T_SWITCH 21026) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 21027) (T_UNSET 21028) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 21029) (T_WHILE 21030) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 21031) (alt_if_stmt_without_else 21032) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 21033) (fn 106) (for_statement 22747) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 21034) (if_stmt_without_else 21035) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 22748) (static_member 123) (variable 126) (variable_class_name 127) ("{" 21037) ("~" 129)) 10202 ((inner_statement_list 22749)) 10203 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22750) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 22751) (T_CLASS 22752) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22753) (T_DEC 22) (T_DECLARE 22754) (T_DIR 24) (T_DNUMBER 25) (T_DO 22755) (T_DOUBLE_CAST 27) (T_ECHO 22756) (T_EMPTY 29) (T_ENDFOR 22757) (T_ENUM 22758) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 22759) (T_FOREACH 22760) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 22761) (T_GOTO 22762) (T_HALT_COMPILER 22763) (T_IF 22764) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22765) (T_INTERFACE 22766) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22767) (T_START_HEREDOC 69) (T_STATIC 22768) (T_STRING 22769) (T_STRING_CAST 72) (T_SWITCH 22770) (T_THROW 74) (T_TRAIT 22771) (T_TRAIT_C 76) (T_TRY 22772) (T_UNSET 22773) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22774) (T_WHILE 22775) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22776) (alt_if_stmt_without_else 22777) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 22778) (attributes 22779) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 22780) (class_modifier 99) (class_modifiers 22781) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 22782) (expr 22783) (fn 106) (fully_dereferenceable 107) #'22784 (function_call 109) (function_declaration_statement 22785) (if_stmt 22786) (if_stmt_without_else 22787) (inline_function 113) (inner_statement 22788) (interface_declaration_statement 22789) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 22790) (static_member 123) (trait_declaration_statement 22791) (variable 126) (variable_class_name 127) ("{" 22792) ("~" 129)) 10204 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 23853) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 10205 ((T_STRING 23846)) 10206 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 23844) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 10207 (("(" 23835)) 10208 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22932) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22933) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22934) (T_DEC 22) (T_DECLARE 22935) (T_DIR 24) (T_DNUMBER 25) (T_DO 22936) (T_DOUBLE_CAST 27) (T_ECHO 22937) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22938) (T_FOREACH 22939) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22940) (T_GOTO 22941) (T_IF 22942) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22943) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22944) (T_START_HEREDOC 69) (T_STATIC 22945) (T_STRING 22946) (T_STRING_CAST 72) (T_SWITCH 22947) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22948) (T_UNSET 22949) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22950) (T_WHILE 22951) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22952) (alt_if_stmt_without_else 22953) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22954) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22955) (if_stmt_without_else 22956) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23829) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22958) ("~" 129)) 10209 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (echo_expr 22926) (echo_expr_list 23827) (expr 22928) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 10210 ((";" 23826)) 10211 ((T_STRING 23819)) 10212 (("(" 23806)) 10213 (("(" 23791)) 10214 (("$" 19441) (T_VARIABLE 19442) (global_var 19443) (global_var_list 23789) (simple_variable 19445)) 10215 ((T_STRING 23787)) 10216 (("(" 23784)) 10217 (("(" 23780)) 10218 ((T_STRING 23774)) 10219 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 23772) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 10220 ((T_FN 35) (T_FUNCTION 131) (T_VARIABLE 18365) (fn 106) #'138 (inline_function 18366) (static_var 18367) (static_var_list 23770)) 10221 ((":" 23769)) 10222 (("(" 23751)) 10223 ((T_STRING 23746)) 10224 (("{" 23729)) 10225 (("(" 23724)) 10226 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 23722) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 10227 (("(" 23713)) 10228 ((T_ELSE 23706) (T_ELSEIF 17711) (T_ENDIF 23707)) 10229 ((T_ABSTRACT 10) (T_ATTRIBUTE 174) (T_CLASS 22752) (T_ENUM 22758) (T_FINAL 34) (T_FN 35) (T_FUNCTION 38) (T_INTERFACE 22766) (T_READONLY 2955) (T_STATIC 17797) (T_TRAIT 22771) (attribute 17798) (attributed_statement 23705) (class_declaration_statement 22780) (class_modifier 99) (class_modifiers 22781) (enum_declaration_statement 22782) (fn 106) #'22784 (function_declaration_statement 22785) (inline_function 17800) (interface_declaration_statement 22789) (trait_declaration_statement 22791)) 10230 ((T_ABSTRACT 10) (T_CLASS 23697) (T_FINAL 34) (T_READONLY 2955) (class_modifier 17789)) 10231 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 23696) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 10232 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17770) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17771) (ampersand 17772) (returns_ref 23684)) 10233 ((T_ELSE 22795) (T_ELSEIF 22796)) 10234 ((inner_statement_list 22793)) 10235 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 22794) ("~" 129)) 10236 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22750) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22751) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22753) (T_DEC 22) (T_DECLARE 22754) (T_DIR 24) (T_DNUMBER 25) (T_DO 22755) (T_DOUBLE_CAST 27) (T_ECHO 22756) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22759) (T_FOREACH 22760) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22761) (T_GOTO 22762) (T_IF 22764) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22765) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22767) (T_START_HEREDOC 69) (T_STATIC 22768) (T_STRING 22769) (T_STRING_CAST 72) (T_SWITCH 22770) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22772) (T_UNSET 22773) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22774) (T_WHILE 22775) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22776) (alt_if_stmt_without_else 22777) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22783) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22786) (if_stmt_without_else 22787) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23683) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22792) ("~" 129)) 10237 (("(" 22797)) 10238 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 22798) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 10239 (("%" 304) (")" 22799) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 10240 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22800) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22801) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22802) (T_DEC 22) (T_DECLARE 22803) (T_DIR 24) (T_DNUMBER 25) (T_DO 22804) (T_DOUBLE_CAST 27) (T_ECHO 22805) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22806) (T_FOREACH 22807) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22808) (T_GOTO 22809) (T_IF 22764) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22810) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22811) (T_START_HEREDOC 69) (T_STATIC 22812) (T_STRING 22813) (T_STRING_CAST 72) (T_SWITCH 22814) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22815) (T_UNSET 22816) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22817) (T_WHILE 22818) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22819) (alt_if_stmt_without_else 22820) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22821) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22822) (if_stmt_without_else 22823) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 22824) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22825) ("~" 129)) 10241 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 23681) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 10242 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 23679) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 10243 (("(" 23670)) 10244 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22932) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22933) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22934) (T_DEC 22) (T_DECLARE 22935) (T_DIR 24) (T_DNUMBER 25) (T_DO 22936) (T_DOUBLE_CAST 27) (T_ECHO 22937) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22938) (T_FOREACH 22939) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22940) (T_GOTO 22941) (T_IF 22942) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22943) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22944) (T_START_HEREDOC 69) (T_STATIC 22945) (T_STRING 22946) (T_STRING_CAST 72) (T_SWITCH 22947) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22948) (T_UNSET 22949) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22950) (T_WHILE 22951) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22952) (alt_if_stmt_without_else 22953) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22954) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22955) (if_stmt_without_else 22956) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 22957) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22958) ("~" 129)) 10245 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (echo_expr 22926) (echo_expr_list 22927) (expr 22928) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 10246 (("(" 22913)) 10247 (("(" 22898)) 10248 (("$" 19441) (T_VARIABLE 19442) (global_var 19443) (global_var_list 22896) (simple_variable 19445)) 10249 ((T_STRING 22894)) 10250 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 22892) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 10251 ((T_FN 35) (T_FUNCTION 131) (T_VARIABLE 18365) (fn 106) #'138 (inline_function 18366) (static_var 18367) (static_var_list 22890)) 10252 ((":" 22889)) 10253 (("(" 22871)) 10254 (("{" 22854)) 10255 (("(" 22849)) 10256 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22847) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 10257 (("(" 22838)) 10258 ((T_ELSE 22831) (T_ELSEIF 17711) (T_ENDIF 22832)) 10259 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 22830) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 10260 ((T_ELSE 22828) (T_ELSEIF 22796)) 10261 ((inner_statement_list 22826)) 10262 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 22827) ("~" 129)) 10263 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22800) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22801) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22802) (T_DEC 22) (T_DECLARE 22803) (T_DIR 24) (T_DNUMBER 25) (T_DO 22804) (T_DOUBLE_CAST 27) (T_ECHO 22805) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22806) (T_FOREACH 22807) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22808) (T_GOTO 22809) (T_IF 22764) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22810) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22811) (T_START_HEREDOC 69) (T_STATIC 22812) (T_STRING 22813) (T_STRING_CAST 72) (T_SWITCH 22814) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22815) (T_UNSET 22816) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22817) (T_WHILE 22818) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22819) (alt_if_stmt_without_else 22820) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22821) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22822) (if_stmt_without_else 22823) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 22829) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22825) ("~" 129)) 10264 ((":" 22834)) 10265 ((";" 22833)) 10266 ((inner_statement_list 22835)) 10267 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17807) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17808) (T_CLASS 17809) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17810) (T_DEC 22) (T_DECLARE 17811) (T_DIR 24) (T_DNUMBER 25) (T_DO 17812) (T_DOUBLE_CAST 27) (T_ECHO 17813) (T_EMPTY 29) (T_ENDIF 22836) (T_ENUM 17815) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17816) (T_FOREACH 17817) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17818) (T_GOTO 17819) (T_HALT_COMPILER 17820) (T_IF 17732) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17821) (T_INTERFACE 17822) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17823) (T_START_HEREDOC 69) (T_STATIC 17824) (T_STRING 17825) (T_STRING_CAST 72) (T_SWITCH 17826) (T_THROW 74) (T_TRAIT 17827) (T_TRAIT_C 76) (T_TRY 17828) (T_UNSET 17829) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17830) (T_WHILE 17831) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17832) (alt_if_stmt_without_else 17833) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17834) (attributes 17835) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17836) (class_modifier 99) (class_modifiers 17837) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17838) (expr 17839) (fn 106) (fully_dereferenceable 107) #'17840 (function_call 109) (function_declaration_statement 17841) (if_stmt 17842) (if_stmt_without_else 17843) (inline_function 113) (inner_statement 17844) (interface_declaration_statement 17845) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17846) (static_member 123) (trait_declaration_statement 17847) (variable 126) (variable_class_name 127) ("{" 17848) ("~" 129)) 10268 ((";" 22837)) 10269 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 22839) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 10270 (("%" 304) (")" 22840) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 10271 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 22841) (";" 22800) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22801) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22802) (T_DEC 22) (T_DECLARE 22803) (T_DIR 24) (T_DNUMBER 25) (T_DO 22804) (T_DOUBLE_CAST 27) (T_ECHO 22805) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22806) (T_FOREACH 22807) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22808) (T_GOTO 22809) (T_IF 22764) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22810) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22811) (T_START_HEREDOC 69) (T_STATIC 22812) (T_STRING 22813) (T_STRING_CAST 72) (T_SWITCH 22814) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22815) (T_UNSET 22816) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22817) (T_WHILE 22818) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22819) (alt_if_stmt_without_else 22820) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22821) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22822) (if_stmt_without_else 22823) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 22842) (static_member 123) (variable 126) (variable_class_name 127) (while_statement 22843) ("{" 22825) ("~" 129)) 10272 ((inner_statement_list 22844)) 10273 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17889) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17890) (T_CLASS 17891) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17892) (T_DEC 22) (T_DECLARE 17893) (T_DIR 24) (T_DNUMBER 25) (T_DO 17894) (T_DOUBLE_CAST 27) (T_ECHO 17895) (T_EMPTY 29) (T_ENDWHILE 22845) (T_ENUM 17897) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17898) (T_FOREACH 17899) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17900) (T_GOTO 17901) (T_HALT_COMPILER 17902) (T_IF 17903) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17904) (T_INTERFACE 17905) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17906) (T_START_HEREDOC 69) (T_STATIC 17907) (T_STRING 17908) (T_STRING_CAST 72) (T_SWITCH 17909) (T_THROW 74) (T_TRAIT 17910) (T_TRAIT_C 76) (T_TRY 17911) (T_UNSET 17912) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17913) (T_WHILE 17914) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17915) (alt_if_stmt_without_else 17916) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17917) (attributes 17918) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17919) (class_modifier 99) (class_modifiers 17920) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17921) (expr 17922) (fn 106) (fully_dereferenceable 107) #'17923 (function_call 109) (function_declaration_statement 17924) (if_stmt 17925) (if_stmt_without_else 17926) (inline_function 113) (inner_statement 17927) (interface_declaration_statement 17928) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17929) (static_member 123) (trait_declaration_statement 17930) (variable 126) (variable_class_name 127) ("{" 17931) ("~" 129)) 10274 ((";" 22846)) 10275 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 22848) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 10276 (("\"" 1538) ("$" 17989) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 17990) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 17991) ("[" 1560) (array_object_dereferenceable 17992) (callable_expr 17993) (callable_variable 17994) (class_constant 1564) (class_name 17995) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 17996) (name 17997) (new_dereferenceable 1570) (simple_variable 17998) (static_member 17999) (unset_variable 18000) (unset_variables 22850) (variable 18002) (variable_class_name 18003)) 10277 (("," 18015) (possible_comma 22851)) 10278 ((")" 22852)) 10279 ((";" 22853)) 10280 ((inner_statement_list 22855)) 10281 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 22856) ("~" 129)) 10282 ((catch_list 22857)) 10283 ((T_CATCH 22858) (T_FINALLY 22859) (finally_statement 22860)) 10284 (("(" 22864)) 10285 (("{" 22861)) 10286 ((inner_statement_list 22862)) 10287 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 22863) ("~" 129)) 10288 ((T_NAME_FULLY_QUALIFIED 18056) (T_NAME_QUALIFIED 18057) (T_NAME_RELATIVE 18058) (T_STATIC 18059) (T_STRING 18060) (catch_name_list 22865) (class_name 18062) (name 18063)) 10289 ((T_VARIABLE 18064) (optional_variable 22866) ("|" 18066)) 10290 ((")" 22867)) 10291 (("{" 22868)) 10292 ((inner_statement_list 22869)) 10293 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 22870) ("~" 129)) 10294 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 22872) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 10295 (("%" 304) (")" 22873) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 10296 ((":" 22874) (switch_case_list 22875) ("{" 22876)) 10297 ((";" 22882) (case_list 22883)) 10298 ((";" 22877) (case_list 22878)) 10299 ((case_list 22880)) 10300 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 22879)) 10301 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 22881)) 10302 ((case_list 22886)) 10303 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 22884)) 10304 ((";" 22885)) 10305 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 22887)) 10306 ((";" 22888)) 10307 (("," 18369) (";" 22891)) 10308 ((";" 22893)) 10309 ((";" 22895)) 10310 (("," 19446) (";" 22897)) 10311 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 22899) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 10312 (("%" 19534) ("*" 19535) ("+" 19536) ("-" 19537) ("." 19538) ("/" 19539) ("<" 19540) (">" 19541) ("?" 19542) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19543) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19544) (T_AS 22900) (T_BOOLEAN_AND 19545) (T_BOOLEAN_OR 19546) (T_COALESCE 19547) (T_INSTANCEOF 19548) (T_IS_EQUAL 19549) (T_IS_GREATER_OR_EQUAL 19550) (T_IS_IDENTICAL 19551) (T_IS_NOT_EQUAL 19552) (T_IS_NOT_IDENTICAL 19553) (T_IS_SMALLER_OR_EQUAL 19554) (T_LOGICAL_AND 19555) (T_LOGICAL_OR 19556) (T_LOGICAL_XOR 19557) (T_PIPE 19558) (T_POW 19559) (T_SL 19560) (T_SPACESHIP 19561) (T_SR 19562) ("^" 19563) ("|" 19564)) 10313 (("\"" 1538) ("$" 20897) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20898) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20899) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20900) ("[" 20901) (ampersand 20902) (array_object_dereferenceable 20903) (callable_expr 20904) (callable_variable 20905) (class_constant 1564) (class_name 20906) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 22901) (fully_dereferenceable 107) (function_call 20908) (name 20909) (new_dereferenceable 1570) (simple_variable 20910) (static_member 20911) (variable 20912) (variable_class_name 20913)) 10314 ((")" 22902) (T_DOUBLE_ARROW 22903)) 10315 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 22906) (";" 22800) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22801) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22802) (T_DEC 22) (T_DECLARE 22803) (T_DIR 24) (T_DNUMBER 25) (T_DO 22804) (T_DOUBLE_CAST 27) (T_ECHO 22805) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22806) (T_FOREACH 22807) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22808) (T_GOTO 22809) (T_IF 22764) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22810) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22811) (T_START_HEREDOC 69) (T_STATIC 22812) (T_STRING 22813) (T_STRING_CAST 72) (T_SWITCH 22814) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22815) (T_UNSET 22816) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22817) (T_WHILE 22818) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22819) (alt_if_stmt_without_else 22820) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22821) (fn 106) (foreach_statement 22912) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22822) (if_stmt_without_else 22823) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 22908) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22825) ("~" 129)) 10316 (("\"" 1538) ("$" 20928) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20929) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20930) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20931) ("[" 20932) (ampersand 20933) (array_object_dereferenceable 20934) (callable_expr 20935) (callable_variable 20936) (class_constant 1564) (class_name 20937) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 22904) (fully_dereferenceable 107) (function_call 20939) (name 20940) (new_dereferenceable 1570) (simple_variable 20941) (static_member 20942) (variable 20943) (variable_class_name 20944)) 10317 ((")" 22905)) 10318 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 22906) (";" 22800) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22801) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22802) (T_DEC 22) (T_DECLARE 22803) (T_DIR 24) (T_DNUMBER 25) (T_DO 22804) (T_DOUBLE_CAST 27) (T_ECHO 22805) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22806) (T_FOREACH 22807) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22808) (T_GOTO 22809) (T_IF 22764) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22810) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22811) (T_START_HEREDOC 69) (T_STATIC 22812) (T_STRING 22813) (T_STRING_CAST 72) (T_SWITCH 22814) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22815) (T_UNSET 22816) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22817) (T_WHILE 22818) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22819) (alt_if_stmt_without_else 22820) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22821) (fn 106) (foreach_statement 22907) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22822) (if_stmt_without_else 22823) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 22908) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22825) ("~" 129)) 10319 ((inner_statement_list 22909)) 10320 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20962) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 20963) (T_CLASS 20964) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20965) (T_DEC 22) (T_DECLARE 20966) (T_DIR 24) (T_DNUMBER 25) (T_DO 20967) (T_DOUBLE_CAST 27) (T_ECHO 20968) (T_EMPTY 29) (T_ENDFOREACH 22910) (T_ENUM 20970) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20971) (T_FOREACH 20972) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20973) (T_GOTO 20974) (T_HALT_COMPILER 20975) (T_IF 20976) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20977) (T_INTERFACE 20978) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20979) (T_START_HEREDOC 69) (T_STATIC 20980) (T_STRING 20981) (T_STRING_CAST 72) (T_SWITCH 20982) (T_THROW 74) (T_TRAIT 20983) (T_TRAIT_C 76) (T_TRY 20984) (T_UNSET 20985) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 20986) (T_WHILE 20987) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 20988) (alt_if_stmt_without_else 20989) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 20990) (attributes 20991) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 20992) (class_modifier 99) (class_modifiers 20993) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 20994) (expr 20995) (fn 106) (fully_dereferenceable 107) #'20996 (function_call 109) (function_declaration_statement 20997) (if_stmt 20998) (if_stmt_without_else 20999) (inline_function 113) (inner_statement 21000) (interface_declaration_statement 21001) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 21002) (static_member 123) (trait_declaration_statement 21003) (variable 126) (variable_class_name 127) ("{" 21004) ("~" 129)) 10321 ((";" 22911)) 10322 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21126) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21127) (fn 18437) (for_exprs 22914) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21129) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 10323 ((";" 22915)) 10324 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21135) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21136) (fn 18437) (for_cond_exprs 22916) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21138) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 10325 ((";" 22917)) 10326 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_VOID_CAST 22737) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 22738) (fn 2425) (for_exprs 22918) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_for_exprs 22740) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 10327 ((")" 22919)) 10328 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 22920) (";" 22800) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22801) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22802) (T_DEC 22) (T_DECLARE 22803) (T_DIR 24) (T_DNUMBER 25) (T_DO 22804) (T_DOUBLE_CAST 27) (T_ECHO 22805) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22806) (T_FOREACH 22807) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22808) (T_GOTO 22809) (T_IF 22764) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22810) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22811) (T_START_HEREDOC 69) (T_STATIC 22812) (T_STRING 22813) (T_STRING_CAST 72) (T_SWITCH 22814) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22815) (T_UNSET 22816) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22817) (T_WHILE 22818) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22819) (alt_if_stmt_without_else 22820) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22821) (fn 106) (for_statement 22921) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22822) (if_stmt_without_else 22823) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 22922) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22825) ("~" 129)) 10329 ((inner_statement_list 22923)) 10330 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22750) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 22751) (T_CLASS 22752) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22753) (T_DEC 22) (T_DECLARE 22754) (T_DIR 24) (T_DNUMBER 25) (T_DO 22755) (T_DOUBLE_CAST 27) (T_ECHO 22756) (T_EMPTY 29) (T_ENDFOR 22924) (T_ENUM 22758) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 22759) (T_FOREACH 22760) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 22761) (T_GOTO 22762) (T_HALT_COMPILER 22763) (T_IF 22764) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22765) (T_INTERFACE 22766) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22767) (T_START_HEREDOC 69) (T_STATIC 22768) (T_STRING 22769) (T_STRING_CAST 72) (T_SWITCH 22770) (T_THROW 74) (T_TRAIT 22771) (T_TRAIT_C 76) (T_TRY 22772) (T_UNSET 22773) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22774) (T_WHILE 22775) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22776) (alt_if_stmt_without_else 22777) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 22778) (attributes 22779) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 22780) (class_modifier 99) (class_modifiers 22781) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 22782) (expr 22783) (fn 106) (fully_dereferenceable 107) #'22784 (function_call 109) (function_declaration_statement 22785) (if_stmt 22786) (if_stmt_without_else 22787) (inline_function 113) (inner_statement 22788) (interface_declaration_statement 22789) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 22790) (static_member 123) (trait_declaration_statement 22791) (variable 126) (variable_class_name 127) ("{" 22792) ("~" 129)) 10331 ((";" 22925)) 10332 (("," 22929) (";" 22930)) 10333 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (echo_expr 22931) (expr 22928) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 10334 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 23668) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 10335 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 23666) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 10336 (("(" 23657)) 10337 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22932) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22933) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22934) (T_DEC 22) (T_DECLARE 22935) (T_DIR 24) (T_DNUMBER 25) (T_DO 22936) (T_DOUBLE_CAST 27) (T_ECHO 22937) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22938) (T_FOREACH 22939) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22940) (T_GOTO 22941) (T_IF 22942) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22943) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22944) (T_START_HEREDOC 69) (T_STATIC 22945) (T_STRING 22946) (T_STRING_CAST 72) (T_SWITCH 22947) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22948) (T_UNSET 22949) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22950) (T_WHILE 22951) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22952) (alt_if_stmt_without_else 22953) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22954) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22955) (if_stmt_without_else 22956) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23651) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22958) ("~" 129)) 10338 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (echo_expr 22926) (echo_expr_list 23649) (expr 22928) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 10339 (("(" 23636)) 10340 (("(" 23621)) 10341 (("$" 19441) (T_VARIABLE 19442) (global_var 19443) (global_var_list 23619) (simple_variable 19445)) 10342 ((T_STRING 23617)) 10343 (("(" 23613)) 10344 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 23611) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 10345 ((T_FN 35) (T_FUNCTION 131) (T_VARIABLE 18365) (fn 106) #'138 (inline_function 18366) (static_var 18367) (static_var_list 23609)) 10346 ((":" 23608)) 10347 (("(" 23590)) 10348 (("{" 23573)) 10349 (("(" 23568)) 10350 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 23566) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 10351 (("(" 23557)) 10352 ((T_ELSE 23550) (T_ELSEIF 17711) (T_ENDIF 23551)) 10353 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 23549) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 10354 ((T_ELSE 22966) (T_ELSEIF 22967)) 10355 ((T_WHILE 22961)) 10356 ((inner_statement_list 22959)) 10357 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 22960) ("~" 129)) 10358 (("(" 22962)) 10359 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 22963) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 10360 (("%" 304) (")" 22964) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 10361 ((";" 22965)) 10362 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22932) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22933) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22934) (T_DEC 22) (T_DECLARE 22935) (T_DIR 24) (T_DNUMBER 25) (T_DO 22936) (T_DOUBLE_CAST 27) (T_ECHO 22937) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22938) (T_FOREACH 22939) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22940) (T_GOTO 22941) (T_IF 22942) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22943) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22944) (T_START_HEREDOC 69) (T_STATIC 22945) (T_STRING 22946) (T_STRING_CAST 72) (T_SWITCH 22947) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22948) (T_UNSET 22949) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22950) (T_WHILE 22951) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22952) (alt_if_stmt_without_else 22953) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22954) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22955) (if_stmt_without_else 22956) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23548) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22958) ("~" 129)) 10363 (("(" 22968)) 10364 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 22969) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 10365 (("%" 304) (")" 22970) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 10366 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22971) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22972) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22973) (T_DEC 22) (T_DECLARE 22974) (T_DIR 24) (T_DNUMBER 25) (T_DO 22975) (T_DOUBLE_CAST 27) (T_ECHO 22976) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22977) (T_FOREACH 22978) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22979) (T_GOTO 22980) (T_IF 22942) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22981) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22982) (T_START_HEREDOC 69) (T_STATIC 22983) (T_STRING 22984) (T_STRING_CAST 72) (T_SWITCH 22985) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22986) (T_UNSET 22987) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22988) (T_WHILE 22989) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22990) (alt_if_stmt_without_else 22991) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22992) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22993) (if_stmt_without_else 22994) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 22995) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22996) ("~" 129)) 10367 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 23546) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 10368 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 23544) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 10369 (("(" 23105)) 10370 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22932) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22933) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22934) (T_DEC 22) (T_DECLARE 22935) (T_DIR 24) (T_DNUMBER 25) (T_DO 22936) (T_DOUBLE_CAST 27) (T_ECHO 22937) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22938) (T_FOREACH 22939) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22940) (T_GOTO 22941) (T_IF 22942) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22943) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22944) (T_START_HEREDOC 69) (T_STATIC 22945) (T_STRING 22946) (T_STRING_CAST 72) (T_SWITCH 22947) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22948) (T_UNSET 22949) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22950) (T_WHILE 22951) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22952) (alt_if_stmt_without_else 22953) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22954) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22955) (if_stmt_without_else 22956) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23099) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22958) ("~" 129)) 10371 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (echo_expr 22926) (echo_expr_list 23097) (expr 22928) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 10372 (("(" 23084)) 10373 (("(" 23069)) 10374 (("$" 19441) (T_VARIABLE 19442) (global_var 19443) (global_var_list 23067) (simple_variable 19445)) 10375 ((T_STRING 23065)) 10376 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 23063) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 10377 ((T_FN 35) (T_FUNCTION 131) (T_VARIABLE 18365) (fn 106) #'138 (inline_function 18366) (static_var 18367) (static_var_list 23061)) 10378 ((":" 23060)) 10379 (("(" 23042)) 10380 (("{" 23025)) 10381 (("(" 23020)) 10382 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 23018) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 10383 (("(" 23009)) 10384 ((T_ELSE 23002) (T_ELSEIF 17711) (T_ENDIF 23003)) 10385 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 23001) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 10386 ((T_ELSE 22999) (T_ELSEIF 22967)) 10387 ((inner_statement_list 22997)) 10388 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 22998) ("~" 129)) 10389 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22971) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22972) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22973) (T_DEC 22) (T_DECLARE 22974) (T_DIR 24) (T_DNUMBER 25) (T_DO 22975) (T_DOUBLE_CAST 27) (T_ECHO 22976) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22977) (T_FOREACH 22978) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22979) (T_GOTO 22980) (T_IF 22942) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22981) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22982) (T_START_HEREDOC 69) (T_STATIC 22983) (T_STRING 22984) (T_STRING_CAST 72) (T_SWITCH 22985) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22986) (T_UNSET 22987) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22988) (T_WHILE 22989) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22990) (alt_if_stmt_without_else 22991) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22992) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22993) (if_stmt_without_else 22994) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23000) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22996) ("~" 129)) 10390 ((":" 23005)) 10391 ((";" 23004)) 10392 ((inner_statement_list 23006)) 10393 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17807) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17808) (T_CLASS 17809) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17810) (T_DEC 22) (T_DECLARE 17811) (T_DIR 24) (T_DNUMBER 25) (T_DO 17812) (T_DOUBLE_CAST 27) (T_ECHO 17813) (T_EMPTY 29) (T_ENDIF 23007) (T_ENUM 17815) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17816) (T_FOREACH 17817) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17818) (T_GOTO 17819) (T_HALT_COMPILER 17820) (T_IF 17732) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17821) (T_INTERFACE 17822) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17823) (T_START_HEREDOC 69) (T_STATIC 17824) (T_STRING 17825) (T_STRING_CAST 72) (T_SWITCH 17826) (T_THROW 74) (T_TRAIT 17827) (T_TRAIT_C 76) (T_TRY 17828) (T_UNSET 17829) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17830) (T_WHILE 17831) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17832) (alt_if_stmt_without_else 17833) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17834) (attributes 17835) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17836) (class_modifier 99) (class_modifiers 17837) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17838) (expr 17839) (fn 106) (fully_dereferenceable 107) #'17840 (function_call 109) (function_declaration_statement 17841) (if_stmt 17842) (if_stmt_without_else 17843) (inline_function 113) (inner_statement 17844) (interface_declaration_statement 17845) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17846) (static_member 123) (trait_declaration_statement 17847) (variable 126) (variable_class_name 127) ("{" 17848) ("~" 129)) 10394 ((";" 23008)) 10395 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 23010) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 10396 (("%" 304) (")" 23011) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 10397 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23012) (";" 22971) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22972) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22973) (T_DEC 22) (T_DECLARE 22974) (T_DIR 24) (T_DNUMBER 25) (T_DO 22975) (T_DOUBLE_CAST 27) (T_ECHO 22976) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22977) (T_FOREACH 22978) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22979) (T_GOTO 22980) (T_IF 22942) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22981) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22982) (T_START_HEREDOC 69) (T_STATIC 22983) (T_STRING 22984) (T_STRING_CAST 72) (T_SWITCH 22985) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22986) (T_UNSET 22987) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22988) (T_WHILE 22989) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22990) (alt_if_stmt_without_else 22991) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22992) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22993) (if_stmt_without_else 22994) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23013) (static_member 123) (variable 126) (variable_class_name 127) (while_statement 23014) ("{" 22996) ("~" 129)) 10398 ((inner_statement_list 23015)) 10399 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17889) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17890) (T_CLASS 17891) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17892) (T_DEC 22) (T_DECLARE 17893) (T_DIR 24) (T_DNUMBER 25) (T_DO 17894) (T_DOUBLE_CAST 27) (T_ECHO 17895) (T_EMPTY 29) (T_ENDWHILE 23016) (T_ENUM 17897) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17898) (T_FOREACH 17899) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17900) (T_GOTO 17901) (T_HALT_COMPILER 17902) (T_IF 17903) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17904) (T_INTERFACE 17905) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17906) (T_START_HEREDOC 69) (T_STATIC 17907) (T_STRING 17908) (T_STRING_CAST 72) (T_SWITCH 17909) (T_THROW 74) (T_TRAIT 17910) (T_TRAIT_C 76) (T_TRY 17911) (T_UNSET 17912) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17913) (T_WHILE 17914) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17915) (alt_if_stmt_without_else 17916) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17917) (attributes 17918) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17919) (class_modifier 99) (class_modifiers 17920) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17921) (expr 17922) (fn 106) (fully_dereferenceable 107) #'17923 (function_call 109) (function_declaration_statement 17924) (if_stmt 17925) (if_stmt_without_else 17926) (inline_function 113) (inner_statement 17927) (interface_declaration_statement 17928) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17929) (static_member 123) (trait_declaration_statement 17930) (variable 126) (variable_class_name 127) ("{" 17931) ("~" 129)) 10400 ((";" 23017)) 10401 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 23019) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 10402 (("\"" 1538) ("$" 17989) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 17990) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 17991) ("[" 1560) (array_object_dereferenceable 17992) (callable_expr 17993) (callable_variable 17994) (class_constant 1564) (class_name 17995) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 17996) (name 17997) (new_dereferenceable 1570) (simple_variable 17998) (static_member 17999) (unset_variable 18000) (unset_variables 23021) (variable 18002) (variable_class_name 18003)) 10403 (("," 18015) (possible_comma 23022)) 10404 ((")" 23023)) 10405 ((";" 23024)) 10406 ((inner_statement_list 23026)) 10407 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 23027) ("~" 129)) 10408 ((catch_list 23028)) 10409 ((T_CATCH 23029) (T_FINALLY 23030) (finally_statement 23031)) 10410 (("(" 23035)) 10411 (("{" 23032)) 10412 ((inner_statement_list 23033)) 10413 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 23034) ("~" 129)) 10414 ((T_NAME_FULLY_QUALIFIED 18056) (T_NAME_QUALIFIED 18057) (T_NAME_RELATIVE 18058) (T_STATIC 18059) (T_STRING 18060) (catch_name_list 23036) (class_name 18062) (name 18063)) 10415 ((T_VARIABLE 18064) (optional_variable 23037) ("|" 18066)) 10416 ((")" 23038)) 10417 (("{" 23039)) 10418 ((inner_statement_list 23040)) 10419 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 23041) ("~" 129)) 10420 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 23043) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 10421 (("%" 304) (")" 23044) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 10422 ((":" 23045) (switch_case_list 23046) ("{" 23047)) 10423 ((";" 23053) (case_list 23054)) 10424 ((";" 23048) (case_list 23049)) 10425 ((case_list 23051)) 10426 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 23050)) 10427 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 23052)) 10428 ((case_list 23057)) 10429 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 23055)) 10430 ((";" 23056)) 10431 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 23058)) 10432 ((";" 23059)) 10433 (("," 18369) (";" 23062)) 10434 ((";" 23064)) 10435 ((";" 23066)) 10436 (("," 19446) (";" 23068)) 10437 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 23070) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 10438 (("%" 19534) ("*" 19535) ("+" 19536) ("-" 19537) ("." 19538) ("/" 19539) ("<" 19540) (">" 19541) ("?" 19542) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19543) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19544) (T_AS 23071) (T_BOOLEAN_AND 19545) (T_BOOLEAN_OR 19546) (T_COALESCE 19547) (T_INSTANCEOF 19548) (T_IS_EQUAL 19549) (T_IS_GREATER_OR_EQUAL 19550) (T_IS_IDENTICAL 19551) (T_IS_NOT_EQUAL 19552) (T_IS_NOT_IDENTICAL 19553) (T_IS_SMALLER_OR_EQUAL 19554) (T_LOGICAL_AND 19555) (T_LOGICAL_OR 19556) (T_LOGICAL_XOR 19557) (T_PIPE 19558) (T_POW 19559) (T_SL 19560) (T_SPACESHIP 19561) (T_SR 19562) ("^" 19563) ("|" 19564)) 10439 (("\"" 1538) ("$" 20897) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20898) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20899) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20900) ("[" 20901) (ampersand 20902) (array_object_dereferenceable 20903) (callable_expr 20904) (callable_variable 20905) (class_constant 1564) (class_name 20906) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 23072) (fully_dereferenceable 107) (function_call 20908) (name 20909) (new_dereferenceable 1570) (simple_variable 20910) (static_member 20911) (variable 20912) (variable_class_name 20913)) 10440 ((")" 23073) (T_DOUBLE_ARROW 23074)) 10441 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23077) (";" 22971) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22972) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22973) (T_DEC 22) (T_DECLARE 22974) (T_DIR 24) (T_DNUMBER 25) (T_DO 22975) (T_DOUBLE_CAST 27) (T_ECHO 22976) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22977) (T_FOREACH 22978) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22979) (T_GOTO 22980) (T_IF 22942) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22981) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22982) (T_START_HEREDOC 69) (T_STATIC 22983) (T_STRING 22984) (T_STRING_CAST 72) (T_SWITCH 22985) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22986) (T_UNSET 22987) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22988) (T_WHILE 22989) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22990) (alt_if_stmt_without_else 22991) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22992) (fn 106) (foreach_statement 23083) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22993) (if_stmt_without_else 22994) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23079) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22996) ("~" 129)) 10442 (("\"" 1538) ("$" 20928) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20929) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20930) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20931) ("[" 20932) (ampersand 20933) (array_object_dereferenceable 20934) (callable_expr 20935) (callable_variable 20936) (class_constant 1564) (class_name 20937) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 23075) (fully_dereferenceable 107) (function_call 20939) (name 20940) (new_dereferenceable 1570) (simple_variable 20941) (static_member 20942) (variable 20943) (variable_class_name 20944)) 10443 ((")" 23076)) 10444 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23077) (";" 22971) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22972) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22973) (T_DEC 22) (T_DECLARE 22974) (T_DIR 24) (T_DNUMBER 25) (T_DO 22975) (T_DOUBLE_CAST 27) (T_ECHO 22976) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22977) (T_FOREACH 22978) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22979) (T_GOTO 22980) (T_IF 22942) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22981) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22982) (T_START_HEREDOC 69) (T_STATIC 22983) (T_STRING 22984) (T_STRING_CAST 72) (T_SWITCH 22985) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22986) (T_UNSET 22987) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22988) (T_WHILE 22989) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22990) (alt_if_stmt_without_else 22991) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22992) (fn 106) (foreach_statement 23078) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22993) (if_stmt_without_else 22994) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23079) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22996) ("~" 129)) 10445 ((inner_statement_list 23080)) 10446 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20962) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 20963) (T_CLASS 20964) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20965) (T_DEC 22) (T_DECLARE 20966) (T_DIR 24) (T_DNUMBER 25) (T_DO 20967) (T_DOUBLE_CAST 27) (T_ECHO 20968) (T_EMPTY 29) (T_ENDFOREACH 23081) (T_ENUM 20970) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20971) (T_FOREACH 20972) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20973) (T_GOTO 20974) (T_HALT_COMPILER 20975) (T_IF 20976) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20977) (T_INTERFACE 20978) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20979) (T_START_HEREDOC 69) (T_STATIC 20980) (T_STRING 20981) (T_STRING_CAST 72) (T_SWITCH 20982) (T_THROW 74) (T_TRAIT 20983) (T_TRAIT_C 76) (T_TRY 20984) (T_UNSET 20985) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 20986) (T_WHILE 20987) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 20988) (alt_if_stmt_without_else 20989) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 20990) (attributes 20991) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 20992) (class_modifier 99) (class_modifiers 20993) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 20994) (expr 20995) (fn 106) (fully_dereferenceable 107) #'20996 (function_call 109) (function_declaration_statement 20997) (if_stmt 20998) (if_stmt_without_else 20999) (inline_function 113) (inner_statement 21000) (interface_declaration_statement 21001) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 21002) (static_member 123) (trait_declaration_statement 21003) (variable 126) (variable_class_name 127) ("{" 21004) ("~" 129)) 10447 ((";" 23082)) 10448 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21126) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21127) (fn 18437) (for_exprs 23085) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21129) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 10449 ((";" 23086)) 10450 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21135) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21136) (fn 18437) (for_cond_exprs 23087) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21138) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 10451 ((";" 23088)) 10452 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_VOID_CAST 22737) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 22738) (fn 2425) (for_exprs 23089) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_for_exprs 22740) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 10453 ((")" 23090)) 10454 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23091) (";" 22971) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22972) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22973) (T_DEC 22) (T_DECLARE 22974) (T_DIR 24) (T_DNUMBER 25) (T_DO 22975) (T_DOUBLE_CAST 27) (T_ECHO 22976) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22977) (T_FOREACH 22978) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22979) (T_GOTO 22980) (T_IF 22942) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22981) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22982) (T_START_HEREDOC 69) (T_STATIC 22983) (T_STRING 22984) (T_STRING_CAST 72) (T_SWITCH 22985) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22986) (T_UNSET 22987) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22988) (T_WHILE 22989) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22990) (alt_if_stmt_without_else 22991) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22992) (fn 106) (for_statement 23092) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22993) (if_stmt_without_else 22994) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23093) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22996) ("~" 129)) 10455 ((inner_statement_list 23094)) 10456 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22750) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 22751) (T_CLASS 22752) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22753) (T_DEC 22) (T_DECLARE 22754) (T_DIR 24) (T_DNUMBER 25) (T_DO 22755) (T_DOUBLE_CAST 27) (T_ECHO 22756) (T_EMPTY 29) (T_ENDFOR 23095) (T_ENUM 22758) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 22759) (T_FOREACH 22760) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 22761) (T_GOTO 22762) (T_HALT_COMPILER 22763) (T_IF 22764) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22765) (T_INTERFACE 22766) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22767) (T_START_HEREDOC 69) (T_STATIC 22768) (T_STRING 22769) (T_STRING_CAST 72) (T_SWITCH 22770) (T_THROW 74) (T_TRAIT 22771) (T_TRAIT_C 76) (T_TRY 22772) (T_UNSET 22773) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22774) (T_WHILE 22775) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22776) (alt_if_stmt_without_else 22777) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 22778) (attributes 22779) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 22780) (class_modifier 99) (class_modifiers 22781) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 22782) (expr 22783) (fn 106) (fully_dereferenceable 107) #'22784 (function_call 109) (function_declaration_statement 22785) (if_stmt 22786) (if_stmt_without_else 22787) (inline_function 113) (inner_statement 22788) (interface_declaration_statement 22789) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 22790) (static_member 123) (trait_declaration_statement 22791) (variable 126) (variable_class_name 127) ("{" 22792) ("~" 129)) 10457 ((";" 23096)) 10458 (("," 22929) (";" 23098)) 10459 ((T_WHILE 23100)) 10460 (("(" 23101)) 10461 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 23102) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 10462 (("%" 304) (")" 23103) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 10463 ((";" 23104)) 10464 ((T_STRING 23106) (const_decl 23107) (const_list 23108)) 10465 (("=" 23541)) 10466 ((")" 23109) ("," 23110)) 10467 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23112) (";" 22971) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22972) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22973) (T_DEC 22) (T_DECLARE 22974) (T_DIR 24) (T_DNUMBER 25) (T_DO 22975) (T_DOUBLE_CAST 27) (T_ECHO 22976) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22977) (T_FOREACH 22978) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22979) (T_GOTO 22980) (T_IF 22942) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22981) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22982) (T_START_HEREDOC 69) (T_STATIC 22983) (T_STRING 22984) (T_STRING_CAST 72) (T_SWITCH 22985) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22986) (T_UNSET 22987) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22988) (T_WHILE 22989) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22990) (alt_if_stmt_without_else 22991) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (declare_statement 23113) (dereferenceable_scalar 103) (expr 22992) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22993) (if_stmt_without_else 22994) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23114) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22996) ("~" 129)) 10468 ((T_STRING 23106) (const_decl 23111)) 10469 ((inner_statement_list 23115)) 10470 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 23116) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 23117) (T_CLASS 23118) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 23119) (T_DEC 22) (T_DECLARE 23120) (T_DIR 24) (T_DNUMBER 25) (T_DO 23121) (T_DOUBLE_CAST 27) (T_ECHO 23122) (T_EMPTY 29) (T_ENDDECLARE 23123) (T_ENUM 23124) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 23125) (T_FOREACH 23126) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 23127) (T_GOTO 23128) (T_HALT_COMPILER 23129) (T_IF 23130) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 23131) (T_INTERFACE 23132) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 23133) (T_START_HEREDOC 69) (T_STATIC 23134) (T_STRING 23135) (T_STRING_CAST 72) (T_SWITCH 23136) (T_THROW 74) (T_TRAIT 23137) (T_TRAIT_C 76) (T_TRY 23138) (T_UNSET 23139) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 23140) (T_WHILE 23141) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 23142) (alt_if_stmt_without_else 23143) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 23144) (attributes 23145) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 23146) (class_modifier 99) (class_modifiers 23147) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 23148) (expr 23149) (fn 106) (fully_dereferenceable 107) #'23150 (function_call 109) (function_declaration_statement 23151) (if_stmt 23152) (if_stmt_without_else 23153) (inline_function 113) (inner_statement 23154) (interface_declaration_statement 23155) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23156) (static_member 123) (trait_declaration_statement 23157) (variable 126) (variable_class_name 127) ("{" 23158) ("~" 129)) 10471 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 23539) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 10472 ((T_STRING 23532)) 10473 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 23530) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 10474 (("(" 23521)) 10475 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22932) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22933) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22934) (T_DEC 22) (T_DECLARE 22935) (T_DIR 24) (T_DNUMBER 25) (T_DO 22936) (T_DOUBLE_CAST 27) (T_ECHO 22937) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22938) (T_FOREACH 22939) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22940) (T_GOTO 22941) (T_IF 22942) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22943) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22944) (T_START_HEREDOC 69) (T_STATIC 22945) (T_STRING 22946) (T_STRING_CAST 72) (T_SWITCH 22947) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22948) (T_UNSET 22949) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22950) (T_WHILE 22951) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22952) (alt_if_stmt_without_else 22953) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22954) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22955) (if_stmt_without_else 22956) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23515) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22958) ("~" 129)) 10476 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (echo_expr 22926) (echo_expr_list 23513) (expr 22928) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 10477 ((";" 23512)) 10478 ((T_STRING 23453)) 10479 (("(" 23440)) 10480 (("(" 23425)) 10481 (("$" 19441) (T_VARIABLE 19442) (global_var 19443) (global_var_list 23423) (simple_variable 19445)) 10482 ((T_STRING 23421)) 10483 (("(" 23418)) 10484 (("(" 23412)) 10485 ((T_STRING 23404)) 10486 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 23402) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 10487 ((T_FN 35) (T_FUNCTION 131) (T_VARIABLE 18365) (fn 106) #'138 (inline_function 18366) (static_var 18367) (static_var_list 23400)) 10488 ((":" 23399)) 10489 (("(" 23381)) 10490 ((T_STRING 23376)) 10491 (("{" 23359)) 10492 (("(" 23354)) 10493 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 23352) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 10494 (("(" 23343)) 10495 ((T_ELSE 23336) (T_ELSEIF 17711) (T_ENDIF 23337)) 10496 ((T_ABSTRACT 10) (T_ATTRIBUTE 174) (T_CLASS 23118) (T_ENUM 23124) (T_FINAL 34) (T_FN 35) (T_FUNCTION 38) (T_INTERFACE 23132) (T_READONLY 2955) (T_STATIC 17797) (T_TRAIT 23137) (attribute 17798) (attributed_statement 23335) (class_declaration_statement 23146) (class_modifier 99) (class_modifiers 23147) (enum_declaration_statement 23148) (fn 106) #'23150 (function_declaration_statement 23151) (inline_function 17800) (interface_declaration_statement 23155) (trait_declaration_statement 23157)) 10497 ((T_ABSTRACT 10) (T_CLASS 23327) (T_FINAL 34) (T_READONLY 2955) (class_modifier 17789)) 10498 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 23326) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 10499 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17770) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17771) (ampersand 17772) (returns_ref 23314)) 10500 ((T_ELSE 23161) (T_ELSEIF 23162)) 10501 ((inner_statement_list 23159)) 10502 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 23160) ("~" 129)) 10503 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 23116) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 23117) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 23119) (T_DEC 22) (T_DECLARE 23120) (T_DIR 24) (T_DNUMBER 25) (T_DO 23121) (T_DOUBLE_CAST 27) (T_ECHO 23122) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 23125) (T_FOREACH 23126) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 23127) (T_GOTO 23128) (T_IF 23130) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 23131) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 23133) (T_START_HEREDOC 69) (T_STATIC 23134) (T_STRING 23135) (T_STRING_CAST 72) (T_SWITCH 23136) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 23138) (T_UNSET 23139) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 23140) (T_WHILE 23141) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 23142) (alt_if_stmt_without_else 23143) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 23149) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 23152) (if_stmt_without_else 23153) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23313) (static_member 123) (variable 126) (variable_class_name 127) ("{" 23158) ("~" 129)) 10504 (("(" 23163)) 10505 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 23164) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 10506 (("%" 304) (")" 23165) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 10507 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 23166) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 23167) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 23168) (T_DEC 22) (T_DECLARE 23169) (T_DIR 24) (T_DNUMBER 25) (T_DO 23170) (T_DOUBLE_CAST 27) (T_ECHO 23171) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 23172) (T_FOREACH 23173) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 23174) (T_GOTO 23175) (T_IF 23130) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 23176) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 23177) (T_START_HEREDOC 69) (T_STATIC 23178) (T_STRING 23179) (T_STRING_CAST 72) (T_SWITCH 23180) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 23181) (T_UNSET 23182) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 23183) (T_WHILE 23184) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 23185) (alt_if_stmt_without_else 23186) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 23187) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 23188) (if_stmt_without_else 23189) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23190) (static_member 123) (variable 126) (variable_class_name 127) ("{" 23191) ("~" 129)) 10508 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 23311) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 10509 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 23309) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 10510 (("(" 23300)) 10511 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22932) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22933) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22934) (T_DEC 22) (T_DECLARE 22935) (T_DIR 24) (T_DNUMBER 25) (T_DO 22936) (T_DOUBLE_CAST 27) (T_ECHO 22937) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22938) (T_FOREACH 22939) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22940) (T_GOTO 22941) (T_IF 22942) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22943) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22944) (T_START_HEREDOC 69) (T_STATIC 22945) (T_STRING 22946) (T_STRING_CAST 72) (T_SWITCH 22947) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22948) (T_UNSET 22949) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22950) (T_WHILE 22951) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22952) (alt_if_stmt_without_else 22953) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22954) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22955) (if_stmt_without_else 22956) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23294) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22958) ("~" 129)) 10512 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (echo_expr 22926) (echo_expr_list 23292) (expr 22928) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 10513 (("(" 23279)) 10514 (("(" 23264)) 10515 (("$" 19441) (T_VARIABLE 19442) (global_var 19443) (global_var_list 23262) (simple_variable 19445)) 10516 ((T_STRING 23260)) 10517 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 23258) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 10518 ((T_FN 35) (T_FUNCTION 131) (T_VARIABLE 18365) (fn 106) #'138 (inline_function 18366) (static_var 18367) (static_var_list 23256)) 10519 ((":" 23255)) 10520 (("(" 23237)) 10521 (("{" 23220)) 10522 (("(" 23215)) 10523 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 23213) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 10524 (("(" 23204)) 10525 ((T_ELSE 23197) (T_ELSEIF 17711) (T_ENDIF 23198)) 10526 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 23196) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 10527 ((T_ELSE 23194) (T_ELSEIF 23162)) 10528 ((inner_statement_list 23192)) 10529 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 23193) ("~" 129)) 10530 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 23166) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 23167) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 23168) (T_DEC 22) (T_DECLARE 23169) (T_DIR 24) (T_DNUMBER 25) (T_DO 23170) (T_DOUBLE_CAST 27) (T_ECHO 23171) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 23172) (T_FOREACH 23173) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 23174) (T_GOTO 23175) (T_IF 23130) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 23176) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 23177) (T_START_HEREDOC 69) (T_STATIC 23178) (T_STRING 23179) (T_STRING_CAST 72) (T_SWITCH 23180) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 23181) (T_UNSET 23182) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 23183) (T_WHILE 23184) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 23185) (alt_if_stmt_without_else 23186) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 23187) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 23188) (if_stmt_without_else 23189) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23195) (static_member 123) (variable 126) (variable_class_name 127) ("{" 23191) ("~" 129)) 10531 ((":" 23200)) 10532 ((";" 23199)) 10533 ((inner_statement_list 23201)) 10534 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17807) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17808) (T_CLASS 17809) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17810) (T_DEC 22) (T_DECLARE 17811) (T_DIR 24) (T_DNUMBER 25) (T_DO 17812) (T_DOUBLE_CAST 27) (T_ECHO 17813) (T_EMPTY 29) (T_ENDIF 23202) (T_ENUM 17815) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17816) (T_FOREACH 17817) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17818) (T_GOTO 17819) (T_HALT_COMPILER 17820) (T_IF 17732) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17821) (T_INTERFACE 17822) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17823) (T_START_HEREDOC 69) (T_STATIC 17824) (T_STRING 17825) (T_STRING_CAST 72) (T_SWITCH 17826) (T_THROW 74) (T_TRAIT 17827) (T_TRAIT_C 76) (T_TRY 17828) (T_UNSET 17829) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17830) (T_WHILE 17831) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17832) (alt_if_stmt_without_else 17833) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17834) (attributes 17835) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17836) (class_modifier 99) (class_modifiers 17837) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17838) (expr 17839) (fn 106) (fully_dereferenceable 107) #'17840 (function_call 109) (function_declaration_statement 17841) (if_stmt 17842) (if_stmt_without_else 17843) (inline_function 113) (inner_statement 17844) (interface_declaration_statement 17845) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17846) (static_member 123) (trait_declaration_statement 17847) (variable 126) (variable_class_name 127) ("{" 17848) ("~" 129)) 10535 ((";" 23203)) 10536 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 23205) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 10537 (("%" 304) (")" 23206) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 10538 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23207) (";" 23166) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 23167) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 23168) (T_DEC 22) (T_DECLARE 23169) (T_DIR 24) (T_DNUMBER 25) (T_DO 23170) (T_DOUBLE_CAST 27) (T_ECHO 23171) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 23172) (T_FOREACH 23173) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 23174) (T_GOTO 23175) (T_IF 23130) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 23176) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 23177) (T_START_HEREDOC 69) (T_STATIC 23178) (T_STRING 23179) (T_STRING_CAST 72) (T_SWITCH 23180) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 23181) (T_UNSET 23182) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 23183) (T_WHILE 23184) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 23185) (alt_if_stmt_without_else 23186) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 23187) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 23188) (if_stmt_without_else 23189) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23208) (static_member 123) (variable 126) (variable_class_name 127) (while_statement 23209) ("{" 23191) ("~" 129)) 10539 ((inner_statement_list 23210)) 10540 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17889) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17890) (T_CLASS 17891) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17892) (T_DEC 22) (T_DECLARE 17893) (T_DIR 24) (T_DNUMBER 25) (T_DO 17894) (T_DOUBLE_CAST 27) (T_ECHO 17895) (T_EMPTY 29) (T_ENDWHILE 23211) (T_ENUM 17897) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17898) (T_FOREACH 17899) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17900) (T_GOTO 17901) (T_HALT_COMPILER 17902) (T_IF 17903) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17904) (T_INTERFACE 17905) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17906) (T_START_HEREDOC 69) (T_STATIC 17907) (T_STRING 17908) (T_STRING_CAST 72) (T_SWITCH 17909) (T_THROW 74) (T_TRAIT 17910) (T_TRAIT_C 76) (T_TRY 17911) (T_UNSET 17912) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17913) (T_WHILE 17914) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17915) (alt_if_stmt_without_else 17916) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17917) (attributes 17918) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17919) (class_modifier 99) (class_modifiers 17920) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17921) (expr 17922) (fn 106) (fully_dereferenceable 107) #'17923 (function_call 109) (function_declaration_statement 17924) (if_stmt 17925) (if_stmt_without_else 17926) (inline_function 113) (inner_statement 17927) (interface_declaration_statement 17928) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17929) (static_member 123) (trait_declaration_statement 17930) (variable 126) (variable_class_name 127) ("{" 17931) ("~" 129)) 10541 ((";" 23212)) 10542 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 23214) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 10543 (("\"" 1538) ("$" 17989) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 17990) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 17991) ("[" 1560) (array_object_dereferenceable 17992) (callable_expr 17993) (callable_variable 17994) (class_constant 1564) (class_name 17995) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 17996) (name 17997) (new_dereferenceable 1570) (simple_variable 17998) (static_member 17999) (unset_variable 18000) (unset_variables 23216) (variable 18002) (variable_class_name 18003)) 10544 (("," 18015) (possible_comma 23217)) 10545 ((")" 23218)) 10546 ((";" 23219)) 10547 ((inner_statement_list 23221)) 10548 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 23222) ("~" 129)) 10549 ((catch_list 23223)) 10550 ((T_CATCH 23224) (T_FINALLY 23225) (finally_statement 23226)) 10551 (("(" 23230)) 10552 (("{" 23227)) 10553 ((inner_statement_list 23228)) 10554 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 23229) ("~" 129)) 10555 ((T_NAME_FULLY_QUALIFIED 18056) (T_NAME_QUALIFIED 18057) (T_NAME_RELATIVE 18058) (T_STATIC 18059) (T_STRING 18060) (catch_name_list 23231) (class_name 18062) (name 18063)) 10556 ((T_VARIABLE 18064) (optional_variable 23232) ("|" 18066)) 10557 ((")" 23233)) 10558 (("{" 23234)) 10559 ((inner_statement_list 23235)) 10560 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 23236) ("~" 129)) 10561 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 23238) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 10562 (("%" 304) (")" 23239) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 10563 ((":" 23240) (switch_case_list 23241) ("{" 23242)) 10564 ((";" 23248) (case_list 23249)) 10565 ((";" 23243) (case_list 23244)) 10566 ((case_list 23246)) 10567 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 23245)) 10568 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 23247)) 10569 ((case_list 23252)) 10570 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 23250)) 10571 ((";" 23251)) 10572 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 23253)) 10573 ((";" 23254)) 10574 (("," 18369) (";" 23257)) 10575 ((";" 23259)) 10576 ((";" 23261)) 10577 (("," 19446) (";" 23263)) 10578 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 23265) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 10579 (("%" 19534) ("*" 19535) ("+" 19536) ("-" 19537) ("." 19538) ("/" 19539) ("<" 19540) (">" 19541) ("?" 19542) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19543) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19544) (T_AS 23266) (T_BOOLEAN_AND 19545) (T_BOOLEAN_OR 19546) (T_COALESCE 19547) (T_INSTANCEOF 19548) (T_IS_EQUAL 19549) (T_IS_GREATER_OR_EQUAL 19550) (T_IS_IDENTICAL 19551) (T_IS_NOT_EQUAL 19552) (T_IS_NOT_IDENTICAL 19553) (T_IS_SMALLER_OR_EQUAL 19554) (T_LOGICAL_AND 19555) (T_LOGICAL_OR 19556) (T_LOGICAL_XOR 19557) (T_PIPE 19558) (T_POW 19559) (T_SL 19560) (T_SPACESHIP 19561) (T_SR 19562) ("^" 19563) ("|" 19564)) 10580 (("\"" 1538) ("$" 20897) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20898) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20899) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20900) ("[" 20901) (ampersand 20902) (array_object_dereferenceable 20903) (callable_expr 20904) (callable_variable 20905) (class_constant 1564) (class_name 20906) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 23267) (fully_dereferenceable 107) (function_call 20908) (name 20909) (new_dereferenceable 1570) (simple_variable 20910) (static_member 20911) (variable 20912) (variable_class_name 20913)) 10581 ((")" 23268) (T_DOUBLE_ARROW 23269)) 10582 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23272) (";" 23166) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 23167) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 23168) (T_DEC 22) (T_DECLARE 23169) (T_DIR 24) (T_DNUMBER 25) (T_DO 23170) (T_DOUBLE_CAST 27) (T_ECHO 23171) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 23172) (T_FOREACH 23173) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 23174) (T_GOTO 23175) (T_IF 23130) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 23176) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 23177) (T_START_HEREDOC 69) (T_STATIC 23178) (T_STRING 23179) (T_STRING_CAST 72) (T_SWITCH 23180) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 23181) (T_UNSET 23182) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 23183) (T_WHILE 23184) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 23185) (alt_if_stmt_without_else 23186) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 23187) (fn 106) (foreach_statement 23278) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 23188) (if_stmt_without_else 23189) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23274) (static_member 123) (variable 126) (variable_class_name 127) ("{" 23191) ("~" 129)) 10583 (("\"" 1538) ("$" 20928) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20929) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20930) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20931) ("[" 20932) (ampersand 20933) (array_object_dereferenceable 20934) (callable_expr 20935) (callable_variable 20936) (class_constant 1564) (class_name 20937) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 23270) (fully_dereferenceable 107) (function_call 20939) (name 20940) (new_dereferenceable 1570) (simple_variable 20941) (static_member 20942) (variable 20943) (variable_class_name 20944)) 10584 ((")" 23271)) 10585 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23272) (";" 23166) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 23167) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 23168) (T_DEC 22) (T_DECLARE 23169) (T_DIR 24) (T_DNUMBER 25) (T_DO 23170) (T_DOUBLE_CAST 27) (T_ECHO 23171) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 23172) (T_FOREACH 23173) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 23174) (T_GOTO 23175) (T_IF 23130) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 23176) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 23177) (T_START_HEREDOC 69) (T_STATIC 23178) (T_STRING 23179) (T_STRING_CAST 72) (T_SWITCH 23180) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 23181) (T_UNSET 23182) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 23183) (T_WHILE 23184) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 23185) (alt_if_stmt_without_else 23186) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 23187) (fn 106) (foreach_statement 23273) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 23188) (if_stmt_without_else 23189) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23274) (static_member 123) (variable 126) (variable_class_name 127) ("{" 23191) ("~" 129)) 10586 ((inner_statement_list 23275)) 10587 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20962) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 20963) (T_CLASS 20964) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20965) (T_DEC 22) (T_DECLARE 20966) (T_DIR 24) (T_DNUMBER 25) (T_DO 20967) (T_DOUBLE_CAST 27) (T_ECHO 20968) (T_EMPTY 29) (T_ENDFOREACH 23276) (T_ENUM 20970) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20971) (T_FOREACH 20972) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20973) (T_GOTO 20974) (T_HALT_COMPILER 20975) (T_IF 20976) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20977) (T_INTERFACE 20978) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20979) (T_START_HEREDOC 69) (T_STATIC 20980) (T_STRING 20981) (T_STRING_CAST 72) (T_SWITCH 20982) (T_THROW 74) (T_TRAIT 20983) (T_TRAIT_C 76) (T_TRY 20984) (T_UNSET 20985) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 20986) (T_WHILE 20987) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 20988) (alt_if_stmt_without_else 20989) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 20990) (attributes 20991) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 20992) (class_modifier 99) (class_modifiers 20993) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 20994) (expr 20995) (fn 106) (fully_dereferenceable 107) #'20996 (function_call 109) (function_declaration_statement 20997) (if_stmt 20998) (if_stmt_without_else 20999) (inline_function 113) (inner_statement 21000) (interface_declaration_statement 21001) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 21002) (static_member 123) (trait_declaration_statement 21003) (variable 126) (variable_class_name 127) ("{" 21004) ("~" 129)) 10588 ((";" 23277)) 10589 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21126) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21127) (fn 18437) (for_exprs 23280) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21129) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 10590 ((";" 23281)) 10591 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21135) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21136) (fn 18437) (for_cond_exprs 23282) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21138) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 10592 ((";" 23283)) 10593 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_VOID_CAST 22737) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 22738) (fn 2425) (for_exprs 23284) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_for_exprs 22740) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 10594 ((")" 23285)) 10595 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23286) (";" 23166) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 23167) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 23168) (T_DEC 22) (T_DECLARE 23169) (T_DIR 24) (T_DNUMBER 25) (T_DO 23170) (T_DOUBLE_CAST 27) (T_ECHO 23171) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 23172) (T_FOREACH 23173) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 23174) (T_GOTO 23175) (T_IF 23130) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 23176) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 23177) (T_START_HEREDOC 69) (T_STATIC 23178) (T_STRING 23179) (T_STRING_CAST 72) (T_SWITCH 23180) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 23181) (T_UNSET 23182) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 23183) (T_WHILE 23184) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 23185) (alt_if_stmt_without_else 23186) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 23187) (fn 106) (for_statement 23287) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 23188) (if_stmt_without_else 23189) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23288) (static_member 123) (variable 126) (variable_class_name 127) ("{" 23191) ("~" 129)) 10596 ((inner_statement_list 23289)) 10597 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22750) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 22751) (T_CLASS 22752) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22753) (T_DEC 22) (T_DECLARE 22754) (T_DIR 24) (T_DNUMBER 25) (T_DO 22755) (T_DOUBLE_CAST 27) (T_ECHO 22756) (T_EMPTY 29) (T_ENDFOR 23290) (T_ENUM 22758) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 22759) (T_FOREACH 22760) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 22761) (T_GOTO 22762) (T_HALT_COMPILER 22763) (T_IF 22764) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22765) (T_INTERFACE 22766) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22767) (T_START_HEREDOC 69) (T_STATIC 22768) (T_STRING 22769) (T_STRING_CAST 72) (T_SWITCH 22770) (T_THROW 74) (T_TRAIT 22771) (T_TRAIT_C 76) (T_TRY 22772) (T_UNSET 22773) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22774) (T_WHILE 22775) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22776) (alt_if_stmt_without_else 22777) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 22778) (attributes 22779) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 22780) (class_modifier 99) (class_modifiers 22781) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 22782) (expr 22783) (fn 106) (fully_dereferenceable 107) #'22784 (function_call 109) (function_declaration_statement 22785) (if_stmt 22786) (if_stmt_without_else 22787) (inline_function 113) (inner_statement 22788) (interface_declaration_statement 22789) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 22790) (static_member 123) (trait_declaration_statement 22791) (variable 126) (variable_class_name 127) ("{" 22792) ("~" 129)) 10598 ((";" 23291)) 10599 (("," 22929) (";" 23293)) 10600 ((T_WHILE 23295)) 10601 (("(" 23296)) 10602 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 23297) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 10603 (("%" 304) (")" 23298) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 10604 ((";" 23299)) 10605 ((T_STRING 23106) (const_decl 23107) (const_list 23301)) 10606 ((")" 23302) ("," 23110)) 10607 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23303) (";" 23166) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 23167) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 23168) (T_DEC 22) (T_DECLARE 23169) (T_DIR 24) (T_DNUMBER 25) (T_DO 23170) (T_DOUBLE_CAST 27) (T_ECHO 23171) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 23172) (T_FOREACH 23173) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 23174) (T_GOTO 23175) (T_IF 23130) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 23176) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 23177) (T_START_HEREDOC 69) (T_STATIC 23178) (T_STRING 23179) (T_STRING_CAST 72) (T_SWITCH 23180) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 23181) (T_UNSET 23182) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 23183) (T_WHILE 23184) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 23185) (alt_if_stmt_without_else 23186) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (declare_statement 23304) (dereferenceable_scalar 103) (expr 23187) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 23188) (if_stmt_without_else 23189) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23305) (static_member 123) (variable 126) (variable_class_name 127) ("{" 23191) ("~" 129)) 10608 ((inner_statement_list 23306)) 10609 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 23116) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 23117) (T_CLASS 23118) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 23119) (T_DEC 22) (T_DECLARE 23120) (T_DIR 24) (T_DNUMBER 25) (T_DO 23121) (T_DOUBLE_CAST 27) (T_ECHO 23122) (T_EMPTY 29) (T_ENDDECLARE 23307) (T_ENUM 23124) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 23125) (T_FOREACH 23126) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 23127) (T_GOTO 23128) (T_HALT_COMPILER 23129) (T_IF 23130) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 23131) (T_INTERFACE 23132) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 23133) (T_START_HEREDOC 69) (T_STATIC 23134) (T_STRING 23135) (T_STRING_CAST 72) (T_SWITCH 23136) (T_THROW 74) (T_TRAIT 23137) (T_TRAIT_C 76) (T_TRY 23138) (T_UNSET 23139) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 23140) (T_WHILE 23141) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 23142) (alt_if_stmt_without_else 23143) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 23144) (attributes 23145) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 23146) (class_modifier 99) (class_modifiers 23147) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 23148) (expr 23149) (fn 106) (fully_dereferenceable 107) #'23150 (function_call 109) (function_declaration_statement 23151) (if_stmt 23152) (if_stmt_without_else 23153) (inline_function 113) (inner_statement 23154) (interface_declaration_statement 23155) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23156) (static_member 123) (trait_declaration_statement 23157) (variable 126) (variable_class_name 127) ("{" 23158) ("~" 129)) 10610 ((";" 23308)) 10611 ((";" 23310)) 10612 ((";" 23312)) 10613 ((T_READONLY 17774) (T_STRING 17775) (backup_doc_comment 143) (function_name 23315)) 10614 ((backup_doc_comment 23316)) 10615 (("(" 23317)) 10616 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 23318)) 10617 ((")" 23319)) 10618 ((":" 168) (return_type 23320)) 10619 ((backup_fn_flags 23321)) 10620 (("{" 23322)) 10621 ((inner_statement_list 23323)) 10622 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 23324) ("~" 129)) 10623 ((backup_fn_flags 23325)) 10624 ((T_STRING 23328)) 10625 ((T_EXTENDS 2987) (extends_from 23329)) 10626 ((T_IMPLEMENTS 2989) (implements_list 23330)) 10627 ((backup_doc_comment 23331)) 10628 (("{" 23332)) 10629 ((class_statement_list 23333)) 10630 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 23334)) 10631 ((":" 23339)) 10632 ((";" 23338)) 10633 ((inner_statement_list 23340)) 10634 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17807) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17808) (T_CLASS 17809) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17810) (T_DEC 22) (T_DECLARE 17811) (T_DIR 24) (T_DNUMBER 25) (T_DO 17812) (T_DOUBLE_CAST 27) (T_ECHO 17813) (T_EMPTY 29) (T_ENDIF 23341) (T_ENUM 17815) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17816) (T_FOREACH 17817) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17818) (T_GOTO 17819) (T_HALT_COMPILER 17820) (T_IF 17732) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17821) (T_INTERFACE 17822) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17823) (T_START_HEREDOC 69) (T_STATIC 17824) (T_STRING 17825) (T_STRING_CAST 72) (T_SWITCH 17826) (T_THROW 74) (T_TRAIT 17827) (T_TRAIT_C 76) (T_TRY 17828) (T_UNSET 17829) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17830) (T_WHILE 17831) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17832) (alt_if_stmt_without_else 17833) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17834) (attributes 17835) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17836) (class_modifier 99) (class_modifiers 17837) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17838) (expr 17839) (fn 106) (fully_dereferenceable 107) #'17840 (function_call 109) (function_declaration_statement 17841) (if_stmt 17842) (if_stmt_without_else 17843) (inline_function 113) (inner_statement 17844) (interface_declaration_statement 17845) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17846) (static_member 123) (trait_declaration_statement 17847) (variable 126) (variable_class_name 127) ("{" 17848) ("~" 129)) 10635 ((";" 23342)) 10636 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 23344) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 10637 (("%" 304) (")" 23345) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 10638 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23346) (";" 23116) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 23117) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 23119) (T_DEC 22) (T_DECLARE 23120) (T_DIR 24) (T_DNUMBER 25) (T_DO 23121) (T_DOUBLE_CAST 27) (T_ECHO 23122) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 23125) (T_FOREACH 23126) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 23127) (T_GOTO 23128) (T_IF 23130) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 23131) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 23133) (T_START_HEREDOC 69) (T_STATIC 23134) (T_STRING 23135) (T_STRING_CAST 72) (T_SWITCH 23136) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 23138) (T_UNSET 23139) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 23140) (T_WHILE 23141) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 23142) (alt_if_stmt_without_else 23143) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 23149) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 23152) (if_stmt_without_else 23153) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23347) (static_member 123) (variable 126) (variable_class_name 127) (while_statement 23348) ("{" 23158) ("~" 129)) 10639 ((inner_statement_list 23349)) 10640 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17889) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17890) (T_CLASS 17891) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17892) (T_DEC 22) (T_DECLARE 17893) (T_DIR 24) (T_DNUMBER 25) (T_DO 17894) (T_DOUBLE_CAST 27) (T_ECHO 17895) (T_EMPTY 29) (T_ENDWHILE 23350) (T_ENUM 17897) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17898) (T_FOREACH 17899) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17900) (T_GOTO 17901) (T_HALT_COMPILER 17902) (T_IF 17903) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17904) (T_INTERFACE 17905) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17906) (T_START_HEREDOC 69) (T_STATIC 17907) (T_STRING 17908) (T_STRING_CAST 72) (T_SWITCH 17909) (T_THROW 74) (T_TRAIT 17910) (T_TRAIT_C 76) (T_TRY 17911) (T_UNSET 17912) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17913) (T_WHILE 17914) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17915) (alt_if_stmt_without_else 17916) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17917) (attributes 17918) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17919) (class_modifier 99) (class_modifiers 17920) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17921) (expr 17922) (fn 106) (fully_dereferenceable 107) #'17923 (function_call 109) (function_declaration_statement 17924) (if_stmt 17925) (if_stmt_without_else 17926) (inline_function 113) (inner_statement 17927) (interface_declaration_statement 17928) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17929) (static_member 123) (trait_declaration_statement 17930) (variable 126) (variable_class_name 127) ("{" 17931) ("~" 129)) 10641 ((";" 23351)) 10642 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 23353) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 10643 (("\"" 1538) ("$" 17989) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 17990) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 17991) ("[" 1560) (array_object_dereferenceable 17992) (callable_expr 17993) (callable_variable 17994) (class_constant 1564) (class_name 17995) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 17996) (name 17997) (new_dereferenceable 1570) (simple_variable 17998) (static_member 17999) (unset_variable 18000) (unset_variables 23355) (variable 18002) (variable_class_name 18003)) 10644 (("," 18015) (possible_comma 23356)) 10645 ((")" 23357)) 10646 ((";" 23358)) 10647 ((inner_statement_list 23360)) 10648 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 23361) ("~" 129)) 10649 ((catch_list 23362)) 10650 ((T_CATCH 23363) (T_FINALLY 23364) (finally_statement 23365)) 10651 (("(" 23369)) 10652 (("{" 23366)) 10653 ((inner_statement_list 23367)) 10654 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 23368) ("~" 129)) 10655 ((T_NAME_FULLY_QUALIFIED 18056) (T_NAME_QUALIFIED 18057) (T_NAME_RELATIVE 18058) (T_STATIC 18059) (T_STRING 18060) (catch_name_list 23370) (class_name 18062) (name 18063)) 10656 ((T_VARIABLE 18064) (optional_variable 23371) ("|" 18066)) 10657 ((")" 23372)) 10658 (("{" 23373)) 10659 ((inner_statement_list 23374)) 10660 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 23375) ("~" 129)) 10661 ((backup_doc_comment 23377)) 10662 (("{" 23378)) 10663 ((class_statement_list 23379)) 10664 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 23380)) 10665 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 23382) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 10666 (("%" 304) (")" 23383) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 10667 ((":" 23384) (switch_case_list 23385) ("{" 23386)) 10668 ((";" 23392) (case_list 23393)) 10669 ((";" 23387) (case_list 23388)) 10670 ((case_list 23390)) 10671 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 23389)) 10672 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 23391)) 10673 ((case_list 23396)) 10674 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 23394)) 10675 ((";" 23395)) 10676 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 23397)) 10677 ((";" 23398)) 10678 (("," 18369) (";" 23401)) 10679 ((";" 23403)) 10680 ((T_EXTENDS 23405) (interface_extends_list 23406)) 10681 ((T_NAME_FULLY_QUALIFIED 7796) (T_NAME_QUALIFIED 7797) (T_NAME_RELATIVE 7798) (T_STATIC 7799) (T_STRING 7800) (class_name 7801) (class_name_list 23411) (name 7803)) 10682 ((backup_doc_comment 23407)) 10683 (("{" 23408)) 10684 ((class_statement_list 23409)) 10685 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 23410)) 10686 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 23413) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 10687 (("%" 304) (")" 23414) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 10688 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23415) (";" 23166) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 23167) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 23168) (T_DEC 22) (T_DECLARE 23169) (T_DIR 24) (T_DNUMBER 25) (T_DO 23170) (T_DOUBLE_CAST 27) (T_ECHO 23171) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 23172) (T_FOREACH 23173) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 23174) (T_GOTO 23175) (T_IF 23130) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 23176) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 23177) (T_START_HEREDOC 69) (T_STATIC 23178) (T_STRING 23179) (T_STRING_CAST 72) (T_SWITCH 23180) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 23181) (T_UNSET 23182) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 23183) (T_WHILE 23184) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 23185) (alt_if_stmt_without_else 23186) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 23187) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 23188) (if_stmt_without_else 23189) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23416) (static_member 123) (variable 126) (variable_class_name 127) ("{" 23191) ("~" 129)) 10689 ((inner_statement_list 23417)) 10690 ((")" 23419)) 10691 ((";" 23420)) 10692 ((";" 23422)) 10693 (("," 19446) (";" 23424)) 10694 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 23426) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 10695 (("%" 19534) ("*" 19535) ("+" 19536) ("-" 19537) ("." 19538) ("/" 19539) ("<" 19540) (">" 19541) ("?" 19542) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19543) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19544) (T_AS 23427) (T_BOOLEAN_AND 19545) (T_BOOLEAN_OR 19546) (T_COALESCE 19547) (T_INSTANCEOF 19548) (T_IS_EQUAL 19549) (T_IS_GREATER_OR_EQUAL 19550) (T_IS_IDENTICAL 19551) (T_IS_NOT_EQUAL 19552) (T_IS_NOT_IDENTICAL 19553) (T_IS_SMALLER_OR_EQUAL 19554) (T_LOGICAL_AND 19555) (T_LOGICAL_OR 19556) (T_LOGICAL_XOR 19557) (T_PIPE 19558) (T_POW 19559) (T_SL 19560) (T_SPACESHIP 19561) (T_SR 19562) ("^" 19563) ("|" 19564)) 10696 (("\"" 1538) ("$" 20897) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20898) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20899) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20900) ("[" 20901) (ampersand 20902) (array_object_dereferenceable 20903) (callable_expr 20904) (callable_variable 20905) (class_constant 1564) (class_name 20906) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 23428) (fully_dereferenceable 107) (function_call 20908) (name 20909) (new_dereferenceable 1570) (simple_variable 20910) (static_member 20911) (variable 20912) (variable_class_name 20913)) 10697 ((")" 23429) (T_DOUBLE_ARROW 23430)) 10698 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23433) (";" 23116) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 23117) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 23119) (T_DEC 22) (T_DECLARE 23120) (T_DIR 24) (T_DNUMBER 25) (T_DO 23121) (T_DOUBLE_CAST 27) (T_ECHO 23122) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 23125) (T_FOREACH 23126) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 23127) (T_GOTO 23128) (T_IF 23130) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 23131) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 23133) (T_START_HEREDOC 69) (T_STATIC 23134) (T_STRING 23135) (T_STRING_CAST 72) (T_SWITCH 23136) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 23138) (T_UNSET 23139) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 23140) (T_WHILE 23141) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 23142) (alt_if_stmt_without_else 23143) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 23149) (fn 106) (foreach_statement 23439) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 23152) (if_stmt_without_else 23153) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23435) (static_member 123) (variable 126) (variable_class_name 127) ("{" 23158) ("~" 129)) 10699 (("\"" 1538) ("$" 20928) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20929) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20930) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20931) ("[" 20932) (ampersand 20933) (array_object_dereferenceable 20934) (callable_expr 20935) (callable_variable 20936) (class_constant 1564) (class_name 20937) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 23431) (fully_dereferenceable 107) (function_call 20939) (name 20940) (new_dereferenceable 1570) (simple_variable 20941) (static_member 20942) (variable 20943) (variable_class_name 20944)) 10700 ((")" 23432)) 10701 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23433) (";" 23116) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 23117) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 23119) (T_DEC 22) (T_DECLARE 23120) (T_DIR 24) (T_DNUMBER 25) (T_DO 23121) (T_DOUBLE_CAST 27) (T_ECHO 23122) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 23125) (T_FOREACH 23126) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 23127) (T_GOTO 23128) (T_IF 23130) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 23131) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 23133) (T_START_HEREDOC 69) (T_STATIC 23134) (T_STRING 23135) (T_STRING_CAST 72) (T_SWITCH 23136) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 23138) (T_UNSET 23139) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 23140) (T_WHILE 23141) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 23142) (alt_if_stmt_without_else 23143) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 23149) (fn 106) (foreach_statement 23434) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 23152) (if_stmt_without_else 23153) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23435) (static_member 123) (variable 126) (variable_class_name 127) ("{" 23158) ("~" 129)) 10702 ((inner_statement_list 23436)) 10703 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20962) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 20963) (T_CLASS 20964) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20965) (T_DEC 22) (T_DECLARE 20966) (T_DIR 24) (T_DNUMBER 25) (T_DO 20967) (T_DOUBLE_CAST 27) (T_ECHO 20968) (T_EMPTY 29) (T_ENDFOREACH 23437) (T_ENUM 20970) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20971) (T_FOREACH 20972) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20973) (T_GOTO 20974) (T_HALT_COMPILER 20975) (T_IF 20976) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20977) (T_INTERFACE 20978) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20979) (T_START_HEREDOC 69) (T_STATIC 20980) (T_STRING 20981) (T_STRING_CAST 72) (T_SWITCH 20982) (T_THROW 74) (T_TRAIT 20983) (T_TRAIT_C 76) (T_TRY 20984) (T_UNSET 20985) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 20986) (T_WHILE 20987) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 20988) (alt_if_stmt_without_else 20989) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 20990) (attributes 20991) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 20992) (class_modifier 99) (class_modifiers 20993) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 20994) (expr 20995) (fn 106) (fully_dereferenceable 107) #'20996 (function_call 109) (function_declaration_statement 20997) (if_stmt 20998) (if_stmt_without_else 20999) (inline_function 113) (inner_statement 21000) (interface_declaration_statement 21001) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 21002) (static_member 123) (trait_declaration_statement 21003) (variable 126) (variable_class_name 127) ("{" 21004) ("~" 129)) 10704 ((";" 23438)) 10705 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21126) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21127) (fn 18437) (for_exprs 23441) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21129) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 10706 ((";" 23442)) 10707 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21135) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21136) (fn 18437) (for_cond_exprs 23443) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21138) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 10708 ((";" 23444)) 10709 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_VOID_CAST 22737) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 22738) (fn 2425) (for_exprs 23445) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_for_exprs 22740) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 10710 ((")" 23446)) 10711 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23447) (";" 23116) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 23117) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 23119) (T_DEC 22) (T_DECLARE 23120) (T_DIR 24) (T_DNUMBER 25) (T_DO 23121) (T_DOUBLE_CAST 27) (T_ECHO 23122) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 23125) (T_FOREACH 23126) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 23127) (T_GOTO 23128) (T_IF 23130) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 23131) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 23133) (T_START_HEREDOC 69) (T_STATIC 23134) (T_STRING 23135) (T_STRING_CAST 72) (T_SWITCH 23136) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 23138) (T_UNSET 23139) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 23140) (T_WHILE 23141) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 23142) (alt_if_stmt_without_else 23143) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 23149) (fn 106) (for_statement 23448) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 23152) (if_stmt_without_else 23153) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23449) (static_member 123) (variable 126) (variable_class_name 127) ("{" 23158) ("~" 129)) 10712 ((inner_statement_list 23450)) 10713 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22750) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 22751) (T_CLASS 22752) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22753) (T_DEC 22) (T_DECLARE 22754) (T_DIR 24) (T_DNUMBER 25) (T_DO 22755) (T_DOUBLE_CAST 27) (T_ECHO 22756) (T_EMPTY 29) (T_ENDFOR 23451) (T_ENUM 22758) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 22759) (T_FOREACH 22760) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 22761) (T_GOTO 22762) (T_HALT_COMPILER 22763) (T_IF 22764) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22765) (T_INTERFACE 22766) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22767) (T_START_HEREDOC 69) (T_STATIC 22768) (T_STRING 22769) (T_STRING_CAST 72) (T_SWITCH 22770) (T_THROW 74) (T_TRAIT 22771) (T_TRAIT_C 76) (T_TRY 22772) (T_UNSET 22773) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22774) (T_WHILE 22775) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22776) (alt_if_stmt_without_else 22777) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 22778) (attributes 22779) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 22780) (class_modifier 99) (class_modifiers 22781) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 22782) (expr 22783) (fn 106) (fully_dereferenceable 107) #'22784 (function_call 109) (function_declaration_statement 22785) (if_stmt 22786) (if_stmt_without_else 22787) (inline_function 113) (inner_statement 22788) (interface_declaration_statement 22789) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 22790) (static_member 123) (trait_declaration_statement 22791) (variable 126) (variable_class_name 127) ("{" 22792) ("~" 129)) 10714 ((";" 23452)) 10715 ((":" 23454) (enum_backing_type 23455)) 10716 (("(" 2278) ("?" 23461) (T_ARRAY 23462) (T_CALLABLE 23463) (T_NAME_FULLY_QUALIFIED 23464) (T_NAME_QUALIFIED 23465) (T_NAME_RELATIVE 23466) (T_STATIC 23467) (T_STRING 23468) (intersection_type 23469) (name 23470) (type 23471) (type_expr 23472) (type_without_static 23473) (union_type 23474) (union_type_element 23475)) 10717 ((T_IMPLEMENTS 2989) (implements_list 23456)) 10718 ((backup_doc_comment 23457)) 10719 (("{" 23458)) 10720 ((class_statement_list 23459)) 10721 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 23460)) 10722 ((T_ARRAY 23506) (T_CALLABLE 23507) (T_NAME_FULLY_QUALIFIED 7806) (T_NAME_QUALIFIED 7807) (T_NAME_RELATIVE 7808) (T_STATIC 23508) (T_STRING 7810) (name 23509) (type 23510) (type_without_static 23511)) 10723 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 23504)) 10724 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 23493)) 10725 (("|" 23491)) 10726 (("|" 23476)) 10727 (("(" 23477) (T_ARRAY 23478) (T_CALLABLE 23479) (T_NAME_FULLY_QUALIFIED 23480) (T_NAME_QUALIFIED 23481) (T_NAME_RELATIVE 23482) (T_STATIC 23483) (T_STRING 23484) (name 23485) (type 23486) (type_without_static 23487) (union_type_element 23488)) 10728 ((T_ARRAY 2307) (T_CALLABLE 2308) (T_NAME_FULLY_QUALIFIED 2309) (T_NAME_QUALIFIED 2310) (T_NAME_RELATIVE 2311) (T_STATIC 2312) (T_STRING 2313) (intersection_type 23489) (name 2315) (type 2316) (type_without_static 2317)) 10729 ((")" 23490) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2330)) 10730 (("(" 23477) (T_ARRAY 23478) (T_CALLABLE 23479) (T_NAME_FULLY_QUALIFIED 23480) (T_NAME_QUALIFIED 23481) (T_NAME_RELATIVE 23482) (T_STATIC 23483) (T_STRING 23484) (name 23485) (type 23486) (type_without_static 23487) (union_type_element 23492)) 10731 ((T_ARRAY 23494) (T_CALLABLE 23495) (T_NAME_FULLY_QUALIFIED 23496) (T_NAME_QUALIFIED 23497) (T_NAME_RELATIVE 23498) (T_STATIC 23499) (T_STRING 23500) (name 23501) (type 23502) (type_without_static 23503)) 10732 ((T_ARRAY 23494) (T_CALLABLE 23495) (T_NAME_FULLY_QUALIFIED 23496) (T_NAME_QUALIFIED 23497) (T_NAME_RELATIVE 23498) (T_STATIC 23499) (T_STRING 23500) (name 23501) (type 23505) (type_without_static 23503)) 10733 (("," 22929) (";" 23514)) 10734 ((T_WHILE 23516)) 10735 (("(" 23517)) 10736 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 23518) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 10737 (("%" 304) (")" 23519) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 10738 ((";" 23520)) 10739 ((T_STRING 23106) (const_decl 23107) (const_list 23522)) 10740 ((")" 23523) ("," 23110)) 10741 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23524) (";" 23116) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 23117) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 23119) (T_DEC 22) (T_DECLARE 23120) (T_DIR 24) (T_DNUMBER 25) (T_DO 23121) (T_DOUBLE_CAST 27) (T_ECHO 23122) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 23125) (T_FOREACH 23126) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 23127) (T_GOTO 23128) (T_IF 23130) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 23131) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 23133) (T_START_HEREDOC 69) (T_STATIC 23134) (T_STRING 23135) (T_STRING_CAST 72) (T_SWITCH 23136) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 23138) (T_UNSET 23139) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 23140) (T_WHILE 23141) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 23142) (alt_if_stmt_without_else 23143) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (declare_statement 23525) (dereferenceable_scalar 103) (expr 23149) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 23152) (if_stmt_without_else 23153) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23526) (static_member 123) (variable 126) (variable_class_name 127) ("{" 23158) ("~" 129)) 10742 ((inner_statement_list 23527)) 10743 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 23116) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 23117) (T_CLASS 23118) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 23119) (T_DEC 22) (T_DECLARE 23120) (T_DIR 24) (T_DNUMBER 25) (T_DO 23121) (T_DOUBLE_CAST 27) (T_ECHO 23122) (T_EMPTY 29) (T_ENDDECLARE 23528) (T_ENUM 23124) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 23125) (T_FOREACH 23126) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 23127) (T_GOTO 23128) (T_HALT_COMPILER 23129) (T_IF 23130) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 23131) (T_INTERFACE 23132) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 23133) (T_START_HEREDOC 69) (T_STATIC 23134) (T_STRING 23135) (T_STRING_CAST 72) (T_SWITCH 23136) (T_THROW 74) (T_TRAIT 23137) (T_TRAIT_C 76) (T_TRY 23138) (T_UNSET 23139) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 23140) (T_WHILE 23141) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 23142) (alt_if_stmt_without_else 23143) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 23144) (attributes 23145) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 23146) (class_modifier 99) (class_modifiers 23147) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 23148) (expr 23149) (fn 106) (fully_dereferenceable 107) #'23150 (function_call 109) (function_declaration_statement 23151) (if_stmt 23152) (if_stmt_without_else 23153) (inline_function 113) (inner_statement 23154) (interface_declaration_statement 23155) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23156) (static_member 123) (trait_declaration_statement 23157) (variable 126) (variable_class_name 127) ("{" 23158) ("~" 129)) 10744 ((";" 23529)) 10745 ((";" 23531)) 10746 ((T_EXTENDS 2987) (extends_from 23533)) 10747 ((T_IMPLEMENTS 2989) (implements_list 23534)) 10748 ((backup_doc_comment 23535)) 10749 (("{" 23536)) 10750 ((class_statement_list 23537)) 10751 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 23538)) 10752 ((";" 23540)) 10753 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 23542) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 10754 (("%" 1125) ("*" 1126) ("+" 1127) ("-" 1128) ("." 1129) ("/" 1130) ("<" 1131) (">" 1132) ("?" 1133) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1134) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1135) (T_BOOLEAN_AND 1136) (T_BOOLEAN_OR 1137) (T_COALESCE 1138) (T_INSTANCEOF 1139) (T_IS_EQUAL 1140) (T_IS_GREATER_OR_EQUAL 1141) (T_IS_IDENTICAL 1142) (T_IS_NOT_EQUAL 1143) (T_IS_NOT_IDENTICAL 1144) (T_IS_SMALLER_OR_EQUAL 1145) (T_LOGICAL_AND 1146) (T_LOGICAL_OR 1147) (T_LOGICAL_XOR 1148) (T_PIPE 1149) (T_POW 1150) (T_SL 1151) (T_SPACESHIP 1152) (T_SR 1153) ("^" 1154) (backup_doc_comment 23543) ("|" 1155)) 10755 ((";" 23545)) 10756 ((";" 23547)) 10757 ((":" 23553)) 10758 ((";" 23552)) 10759 ((inner_statement_list 23554)) 10760 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17807) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17808) (T_CLASS 17809) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17810) (T_DEC 22) (T_DECLARE 17811) (T_DIR 24) (T_DNUMBER 25) (T_DO 17812) (T_DOUBLE_CAST 27) (T_ECHO 17813) (T_EMPTY 29) (T_ENDIF 23555) (T_ENUM 17815) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17816) (T_FOREACH 17817) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17818) (T_GOTO 17819) (T_HALT_COMPILER 17820) (T_IF 17732) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17821) (T_INTERFACE 17822) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17823) (T_START_HEREDOC 69) (T_STATIC 17824) (T_STRING 17825) (T_STRING_CAST 72) (T_SWITCH 17826) (T_THROW 74) (T_TRAIT 17827) (T_TRAIT_C 76) (T_TRY 17828) (T_UNSET 17829) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17830) (T_WHILE 17831) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17832) (alt_if_stmt_without_else 17833) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17834) (attributes 17835) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17836) (class_modifier 99) (class_modifiers 17837) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17838) (expr 17839) (fn 106) (fully_dereferenceable 107) #'17840 (function_call 109) (function_declaration_statement 17841) (if_stmt 17842) (if_stmt_without_else 17843) (inline_function 113) (inner_statement 17844) (interface_declaration_statement 17845) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17846) (static_member 123) (trait_declaration_statement 17847) (variable 126) (variable_class_name 127) ("{" 17848) ("~" 129)) 10761 ((";" 23556)) 10762 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 23558) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 10763 (("%" 304) (")" 23559) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 10764 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23560) (";" 22932) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22933) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22934) (T_DEC 22) (T_DECLARE 22935) (T_DIR 24) (T_DNUMBER 25) (T_DO 22936) (T_DOUBLE_CAST 27) (T_ECHO 22937) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22938) (T_FOREACH 22939) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22940) (T_GOTO 22941) (T_IF 22942) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22943) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22944) (T_START_HEREDOC 69) (T_STATIC 22945) (T_STRING 22946) (T_STRING_CAST 72) (T_SWITCH 22947) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22948) (T_UNSET 22949) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22950) (T_WHILE 22951) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22952) (alt_if_stmt_without_else 22953) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22954) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22955) (if_stmt_without_else 22956) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23561) (static_member 123) (variable 126) (variable_class_name 127) (while_statement 23562) ("{" 22958) ("~" 129)) 10765 ((inner_statement_list 23563)) 10766 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17889) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17890) (T_CLASS 17891) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17892) (T_DEC 22) (T_DECLARE 17893) (T_DIR 24) (T_DNUMBER 25) (T_DO 17894) (T_DOUBLE_CAST 27) (T_ECHO 17895) (T_EMPTY 29) (T_ENDWHILE 23564) (T_ENUM 17897) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17898) (T_FOREACH 17899) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17900) (T_GOTO 17901) (T_HALT_COMPILER 17902) (T_IF 17903) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17904) (T_INTERFACE 17905) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17906) (T_START_HEREDOC 69) (T_STATIC 17907) (T_STRING 17908) (T_STRING_CAST 72) (T_SWITCH 17909) (T_THROW 74) (T_TRAIT 17910) (T_TRAIT_C 76) (T_TRY 17911) (T_UNSET 17912) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17913) (T_WHILE 17914) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17915) (alt_if_stmt_without_else 17916) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17917) (attributes 17918) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17919) (class_modifier 99) (class_modifiers 17920) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17921) (expr 17922) (fn 106) (fully_dereferenceable 107) #'17923 (function_call 109) (function_declaration_statement 17924) (if_stmt 17925) (if_stmt_without_else 17926) (inline_function 113) (inner_statement 17927) (interface_declaration_statement 17928) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17929) (static_member 123) (trait_declaration_statement 17930) (variable 126) (variable_class_name 127) ("{" 17931) ("~" 129)) 10767 ((";" 23565)) 10768 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 23567) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 10769 (("\"" 1538) ("$" 17989) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 17990) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 17991) ("[" 1560) (array_object_dereferenceable 17992) (callable_expr 17993) (callable_variable 17994) (class_constant 1564) (class_name 17995) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 17996) (name 17997) (new_dereferenceable 1570) (simple_variable 17998) (static_member 17999) (unset_variable 18000) (unset_variables 23569) (variable 18002) (variable_class_name 18003)) 10770 (("," 18015) (possible_comma 23570)) 10771 ((")" 23571)) 10772 ((";" 23572)) 10773 ((inner_statement_list 23574)) 10774 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 23575) ("~" 129)) 10775 ((catch_list 23576)) 10776 ((T_CATCH 23577) (T_FINALLY 23578) (finally_statement 23579)) 10777 (("(" 23583)) 10778 (("{" 23580)) 10779 ((inner_statement_list 23581)) 10780 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 23582) ("~" 129)) 10781 ((T_NAME_FULLY_QUALIFIED 18056) (T_NAME_QUALIFIED 18057) (T_NAME_RELATIVE 18058) (T_STATIC 18059) (T_STRING 18060) (catch_name_list 23584) (class_name 18062) (name 18063)) 10782 ((T_VARIABLE 18064) (optional_variable 23585) ("|" 18066)) 10783 ((")" 23586)) 10784 (("{" 23587)) 10785 ((inner_statement_list 23588)) 10786 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 23589) ("~" 129)) 10787 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 23591) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 10788 (("%" 304) (")" 23592) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 10789 ((":" 23593) (switch_case_list 23594) ("{" 23595)) 10790 ((";" 23601) (case_list 23602)) 10791 ((";" 23596) (case_list 23597)) 10792 ((case_list 23599)) 10793 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 23598)) 10794 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 23600)) 10795 ((case_list 23605)) 10796 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 23603)) 10797 ((";" 23604)) 10798 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 23606)) 10799 ((";" 23607)) 10800 (("," 18369) (";" 23610)) 10801 ((";" 23612)) 10802 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 23614) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 10803 (("%" 304) (")" 23615) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 10804 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23415) (";" 22971) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22972) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22973) (T_DEC 22) (T_DECLARE 22974) (T_DIR 24) (T_DNUMBER 25) (T_DO 22975) (T_DOUBLE_CAST 27) (T_ECHO 22976) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22977) (T_FOREACH 22978) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22979) (T_GOTO 22980) (T_IF 22942) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22981) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22982) (T_START_HEREDOC 69) (T_STATIC 22983) (T_STRING 22984) (T_STRING_CAST 72) (T_SWITCH 22985) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22986) (T_UNSET 22987) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22988) (T_WHILE 22989) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22990) (alt_if_stmt_without_else 22991) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22992) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22993) (if_stmt_without_else 22994) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23616) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22996) ("~" 129)) 10805 ((";" 23618)) 10806 (("," 19446) (";" 23620)) 10807 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 23622) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 10808 (("%" 19534) ("*" 19535) ("+" 19536) ("-" 19537) ("." 19538) ("/" 19539) ("<" 19540) (">" 19541) ("?" 19542) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19543) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19544) (T_AS 23623) (T_BOOLEAN_AND 19545) (T_BOOLEAN_OR 19546) (T_COALESCE 19547) (T_INSTANCEOF 19548) (T_IS_EQUAL 19549) (T_IS_GREATER_OR_EQUAL 19550) (T_IS_IDENTICAL 19551) (T_IS_NOT_EQUAL 19552) (T_IS_NOT_IDENTICAL 19553) (T_IS_SMALLER_OR_EQUAL 19554) (T_LOGICAL_AND 19555) (T_LOGICAL_OR 19556) (T_LOGICAL_XOR 19557) (T_PIPE 19558) (T_POW 19559) (T_SL 19560) (T_SPACESHIP 19561) (T_SR 19562) ("^" 19563) ("|" 19564)) 10809 (("\"" 1538) ("$" 20897) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20898) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20899) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20900) ("[" 20901) (ampersand 20902) (array_object_dereferenceable 20903) (callable_expr 20904) (callable_variable 20905) (class_constant 1564) (class_name 20906) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 23624) (fully_dereferenceable 107) (function_call 20908) (name 20909) (new_dereferenceable 1570) (simple_variable 20910) (static_member 20911) (variable 20912) (variable_class_name 20913)) 10810 ((")" 23625) (T_DOUBLE_ARROW 23626)) 10811 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23629) (";" 22932) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22933) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22934) (T_DEC 22) (T_DECLARE 22935) (T_DIR 24) (T_DNUMBER 25) (T_DO 22936) (T_DOUBLE_CAST 27) (T_ECHO 22937) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22938) (T_FOREACH 22939) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22940) (T_GOTO 22941) (T_IF 22942) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22943) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22944) (T_START_HEREDOC 69) (T_STATIC 22945) (T_STRING 22946) (T_STRING_CAST 72) (T_SWITCH 22947) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22948) (T_UNSET 22949) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22950) (T_WHILE 22951) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22952) (alt_if_stmt_without_else 22953) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22954) (fn 106) (foreach_statement 23635) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22955) (if_stmt_without_else 22956) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23631) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22958) ("~" 129)) 10812 (("\"" 1538) ("$" 20928) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20929) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20930) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20931) ("[" 20932) (ampersand 20933) (array_object_dereferenceable 20934) (callable_expr 20935) (callable_variable 20936) (class_constant 1564) (class_name 20937) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 23627) (fully_dereferenceable 107) (function_call 20939) (name 20940) (new_dereferenceable 1570) (simple_variable 20941) (static_member 20942) (variable 20943) (variable_class_name 20944)) 10813 ((")" 23628)) 10814 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23629) (";" 22932) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22933) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22934) (T_DEC 22) (T_DECLARE 22935) (T_DIR 24) (T_DNUMBER 25) (T_DO 22936) (T_DOUBLE_CAST 27) (T_ECHO 22937) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22938) (T_FOREACH 22939) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22940) (T_GOTO 22941) (T_IF 22942) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22943) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22944) (T_START_HEREDOC 69) (T_STATIC 22945) (T_STRING 22946) (T_STRING_CAST 72) (T_SWITCH 22947) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22948) (T_UNSET 22949) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22950) (T_WHILE 22951) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22952) (alt_if_stmt_without_else 22953) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22954) (fn 106) (foreach_statement 23630) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22955) (if_stmt_without_else 22956) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23631) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22958) ("~" 129)) 10815 ((inner_statement_list 23632)) 10816 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20962) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 20963) (T_CLASS 20964) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20965) (T_DEC 22) (T_DECLARE 20966) (T_DIR 24) (T_DNUMBER 25) (T_DO 20967) (T_DOUBLE_CAST 27) (T_ECHO 20968) (T_EMPTY 29) (T_ENDFOREACH 23633) (T_ENUM 20970) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20971) (T_FOREACH 20972) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20973) (T_GOTO 20974) (T_HALT_COMPILER 20975) (T_IF 20976) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20977) (T_INTERFACE 20978) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20979) (T_START_HEREDOC 69) (T_STATIC 20980) (T_STRING 20981) (T_STRING_CAST 72) (T_SWITCH 20982) (T_THROW 74) (T_TRAIT 20983) (T_TRAIT_C 76) (T_TRY 20984) (T_UNSET 20985) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 20986) (T_WHILE 20987) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 20988) (alt_if_stmt_without_else 20989) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 20990) (attributes 20991) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 20992) (class_modifier 99) (class_modifiers 20993) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 20994) (expr 20995) (fn 106) (fully_dereferenceable 107) #'20996 (function_call 109) (function_declaration_statement 20997) (if_stmt 20998) (if_stmt_without_else 20999) (inline_function 113) (inner_statement 21000) (interface_declaration_statement 21001) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 21002) (static_member 123) (trait_declaration_statement 21003) (variable 126) (variable_class_name 127) ("{" 21004) ("~" 129)) 10817 ((";" 23634)) 10818 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21126) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21127) (fn 18437) (for_exprs 23637) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21129) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 10819 ((";" 23638)) 10820 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21135) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21136) (fn 18437) (for_cond_exprs 23639) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21138) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 10821 ((";" 23640)) 10822 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_VOID_CAST 22737) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 22738) (fn 2425) (for_exprs 23641) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_for_exprs 22740) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 10823 ((")" 23642)) 10824 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23643) (";" 22932) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22933) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22934) (T_DEC 22) (T_DECLARE 22935) (T_DIR 24) (T_DNUMBER 25) (T_DO 22936) (T_DOUBLE_CAST 27) (T_ECHO 22937) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22938) (T_FOREACH 22939) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22940) (T_GOTO 22941) (T_IF 22942) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22943) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22944) (T_START_HEREDOC 69) (T_STATIC 22945) (T_STRING 22946) (T_STRING_CAST 72) (T_SWITCH 22947) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22948) (T_UNSET 22949) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22950) (T_WHILE 22951) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22952) (alt_if_stmt_without_else 22953) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22954) (fn 106) (for_statement 23644) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22955) (if_stmt_without_else 22956) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23645) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22958) ("~" 129)) 10825 ((inner_statement_list 23646)) 10826 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22750) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 22751) (T_CLASS 22752) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22753) (T_DEC 22) (T_DECLARE 22754) (T_DIR 24) (T_DNUMBER 25) (T_DO 22755) (T_DOUBLE_CAST 27) (T_ECHO 22756) (T_EMPTY 29) (T_ENDFOR 23647) (T_ENUM 22758) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 22759) (T_FOREACH 22760) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 22761) (T_GOTO 22762) (T_HALT_COMPILER 22763) (T_IF 22764) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22765) (T_INTERFACE 22766) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22767) (T_START_HEREDOC 69) (T_STATIC 22768) (T_STRING 22769) (T_STRING_CAST 72) (T_SWITCH 22770) (T_THROW 74) (T_TRAIT 22771) (T_TRAIT_C 76) (T_TRY 22772) (T_UNSET 22773) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22774) (T_WHILE 22775) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22776) (alt_if_stmt_without_else 22777) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 22778) (attributes 22779) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 22780) (class_modifier 99) (class_modifiers 22781) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 22782) (expr 22783) (fn 106) (fully_dereferenceable 107) #'22784 (function_call 109) (function_declaration_statement 22785) (if_stmt 22786) (if_stmt_without_else 22787) (inline_function 113) (inner_statement 22788) (interface_declaration_statement 22789) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 22790) (static_member 123) (trait_declaration_statement 22791) (variable 126) (variable_class_name 127) ("{" 22792) ("~" 129)) 10827 ((";" 23648)) 10828 (("," 22929) (";" 23650)) 10829 ((T_WHILE 23652)) 10830 (("(" 23653)) 10831 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 23654) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 10832 (("%" 304) (")" 23655) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 10833 ((";" 23656)) 10834 ((T_STRING 23106) (const_decl 23107) (const_list 23658)) 10835 ((")" 23659) ("," 23110)) 10836 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23660) (";" 22932) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22933) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22934) (T_DEC 22) (T_DECLARE 22935) (T_DIR 24) (T_DNUMBER 25) (T_DO 22936) (T_DOUBLE_CAST 27) (T_ECHO 22937) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22938) (T_FOREACH 22939) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22940) (T_GOTO 22941) (T_IF 22942) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22943) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22944) (T_START_HEREDOC 69) (T_STATIC 22945) (T_STRING 22946) (T_STRING_CAST 72) (T_SWITCH 22947) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22948) (T_UNSET 22949) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22950) (T_WHILE 22951) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22952) (alt_if_stmt_without_else 22953) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (declare_statement 23661) (dereferenceable_scalar 103) (expr 22954) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22955) (if_stmt_without_else 22956) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23662) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22958) ("~" 129)) 10837 ((inner_statement_list 23663)) 10838 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 23116) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 23117) (T_CLASS 23118) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 23119) (T_DEC 22) (T_DECLARE 23120) (T_DIR 24) (T_DNUMBER 25) (T_DO 23121) (T_DOUBLE_CAST 27) (T_ECHO 23122) (T_EMPTY 29) (T_ENDDECLARE 23664) (T_ENUM 23124) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 23125) (T_FOREACH 23126) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 23127) (T_GOTO 23128) (T_HALT_COMPILER 23129) (T_IF 23130) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 23131) (T_INTERFACE 23132) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 23133) (T_START_HEREDOC 69) (T_STATIC 23134) (T_STRING 23135) (T_STRING_CAST 72) (T_SWITCH 23136) (T_THROW 74) (T_TRAIT 23137) (T_TRAIT_C 76) (T_TRY 23138) (T_UNSET 23139) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 23140) (T_WHILE 23141) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 23142) (alt_if_stmt_without_else 23143) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 23144) (attributes 23145) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 23146) (class_modifier 99) (class_modifiers 23147) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 23148) (expr 23149) (fn 106) (fully_dereferenceable 107) #'23150 (function_call 109) (function_declaration_statement 23151) (if_stmt 23152) (if_stmt_without_else 23153) (inline_function 113) (inner_statement 23154) (interface_declaration_statement 23155) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23156) (static_member 123) (trait_declaration_statement 23157) (variable 126) (variable_class_name 127) ("{" 23158) ("~" 129)) 10839 ((";" 23665)) 10840 ((";" 23667)) 10841 ((";" 23669)) 10842 ((T_STRING 23106) (const_decl 23107) (const_list 23671)) 10843 ((")" 23672) ("," 23110)) 10844 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23673) (";" 22800) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22801) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22802) (T_DEC 22) (T_DECLARE 22803) (T_DIR 24) (T_DNUMBER 25) (T_DO 22804) (T_DOUBLE_CAST 27) (T_ECHO 22805) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22806) (T_FOREACH 22807) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22808) (T_GOTO 22809) (T_IF 22764) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22810) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22811) (T_START_HEREDOC 69) (T_STATIC 22812) (T_STRING 22813) (T_STRING_CAST 72) (T_SWITCH 22814) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22815) (T_UNSET 22816) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22817) (T_WHILE 22818) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22819) (alt_if_stmt_without_else 22820) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (declare_statement 23674) (dereferenceable_scalar 103) (expr 22821) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22822) (if_stmt_without_else 22823) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23675) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22825) ("~" 129)) 10845 ((inner_statement_list 23676)) 10846 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 23116) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 23117) (T_CLASS 23118) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 23119) (T_DEC 22) (T_DECLARE 23120) (T_DIR 24) (T_DNUMBER 25) (T_DO 23121) (T_DOUBLE_CAST 27) (T_ECHO 23122) (T_EMPTY 29) (T_ENDDECLARE 23677) (T_ENUM 23124) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 23125) (T_FOREACH 23126) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 23127) (T_GOTO 23128) (T_HALT_COMPILER 23129) (T_IF 23130) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 23131) (T_INTERFACE 23132) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 23133) (T_START_HEREDOC 69) (T_STATIC 23134) (T_STRING 23135) (T_STRING_CAST 72) (T_SWITCH 23136) (T_THROW 74) (T_TRAIT 23137) (T_TRAIT_C 76) (T_TRY 23138) (T_UNSET 23139) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 23140) (T_WHILE 23141) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 23142) (alt_if_stmt_without_else 23143) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 23144) (attributes 23145) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 23146) (class_modifier 99) (class_modifiers 23147) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 23148) (expr 23149) (fn 106) (fully_dereferenceable 107) #'23150 (function_call 109) (function_declaration_statement 23151) (if_stmt 23152) (if_stmt_without_else 23153) (inline_function 113) (inner_statement 23154) (interface_declaration_statement 23155) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23156) (static_member 123) (trait_declaration_statement 23157) (variable 126) (variable_class_name 127) ("{" 23158) ("~" 129)) 10847 ((";" 23678)) 10848 ((";" 23680)) 10849 ((";" 23682)) 10850 ((T_READONLY 17774) (T_STRING 17775) (backup_doc_comment 143) (function_name 23685)) 10851 ((backup_doc_comment 23686)) 10852 (("(" 23687)) 10853 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 23688)) 10854 ((")" 23689)) 10855 ((":" 168) (return_type 23690)) 10856 ((backup_fn_flags 23691)) 10857 (("{" 23692)) 10858 ((inner_statement_list 23693)) 10859 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 23694) ("~" 129)) 10860 ((backup_fn_flags 23695)) 10861 ((T_STRING 23698)) 10862 ((T_EXTENDS 2987) (extends_from 23699)) 10863 ((T_IMPLEMENTS 2989) (implements_list 23700)) 10864 ((backup_doc_comment 23701)) 10865 (("{" 23702)) 10866 ((class_statement_list 23703)) 10867 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 23704)) 10868 ((":" 23709)) 10869 ((";" 23708)) 10870 ((inner_statement_list 23710)) 10871 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17807) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17808) (T_CLASS 17809) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17810) (T_DEC 22) (T_DECLARE 17811) (T_DIR 24) (T_DNUMBER 25) (T_DO 17812) (T_DOUBLE_CAST 27) (T_ECHO 17813) (T_EMPTY 29) (T_ENDIF 23711) (T_ENUM 17815) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17816) (T_FOREACH 17817) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17818) (T_GOTO 17819) (T_HALT_COMPILER 17820) (T_IF 17732) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17821) (T_INTERFACE 17822) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17823) (T_START_HEREDOC 69) (T_STATIC 17824) (T_STRING 17825) (T_STRING_CAST 72) (T_SWITCH 17826) (T_THROW 74) (T_TRAIT 17827) (T_TRAIT_C 76) (T_TRY 17828) (T_UNSET 17829) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17830) (T_WHILE 17831) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17832) (alt_if_stmt_without_else 17833) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17834) (attributes 17835) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17836) (class_modifier 99) (class_modifiers 17837) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17838) (expr 17839) (fn 106) (fully_dereferenceable 107) #'17840 (function_call 109) (function_declaration_statement 17841) (if_stmt 17842) (if_stmt_without_else 17843) (inline_function 113) (inner_statement 17844) (interface_declaration_statement 17845) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17846) (static_member 123) (trait_declaration_statement 17847) (variable 126) (variable_class_name 127) ("{" 17848) ("~" 129)) 10872 ((";" 23712)) 10873 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 23714) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 10874 (("%" 304) (")" 23715) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 10875 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23716) (";" 22750) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22751) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22753) (T_DEC 22) (T_DECLARE 22754) (T_DIR 24) (T_DNUMBER 25) (T_DO 22755) (T_DOUBLE_CAST 27) (T_ECHO 22756) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22759) (T_FOREACH 22760) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22761) (T_GOTO 22762) (T_IF 22764) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22765) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22767) (T_START_HEREDOC 69) (T_STATIC 22768) (T_STRING 22769) (T_STRING_CAST 72) (T_SWITCH 22770) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22772) (T_UNSET 22773) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22774) (T_WHILE 22775) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22776) (alt_if_stmt_without_else 22777) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22783) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22786) (if_stmt_without_else 22787) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23717) (static_member 123) (variable 126) (variable_class_name 127) (while_statement 23718) ("{" 22792) ("~" 129)) 10876 ((inner_statement_list 23719)) 10877 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17889) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17890) (T_CLASS 17891) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17892) (T_DEC 22) (T_DECLARE 17893) (T_DIR 24) (T_DNUMBER 25) (T_DO 17894) (T_DOUBLE_CAST 27) (T_ECHO 17895) (T_EMPTY 29) (T_ENDWHILE 23720) (T_ENUM 17897) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17898) (T_FOREACH 17899) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17900) (T_GOTO 17901) (T_HALT_COMPILER 17902) (T_IF 17903) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17904) (T_INTERFACE 17905) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17906) (T_START_HEREDOC 69) (T_STATIC 17907) (T_STRING 17908) (T_STRING_CAST 72) (T_SWITCH 17909) (T_THROW 74) (T_TRAIT 17910) (T_TRAIT_C 76) (T_TRY 17911) (T_UNSET 17912) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17913) (T_WHILE 17914) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17915) (alt_if_stmt_without_else 17916) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17917) (attributes 17918) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17919) (class_modifier 99) (class_modifiers 17920) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17921) (expr 17922) (fn 106) (fully_dereferenceable 107) #'17923 (function_call 109) (function_declaration_statement 17924) (if_stmt 17925) (if_stmt_without_else 17926) (inline_function 113) (inner_statement 17927) (interface_declaration_statement 17928) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17929) (static_member 123) (trait_declaration_statement 17930) (variable 126) (variable_class_name 127) ("{" 17931) ("~" 129)) 10878 ((";" 23721)) 10879 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 23723) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 10880 (("\"" 1538) ("$" 17989) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 17990) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 17991) ("[" 1560) (array_object_dereferenceable 17992) (callable_expr 17993) (callable_variable 17994) (class_constant 1564) (class_name 17995) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 17996) (name 17997) (new_dereferenceable 1570) (simple_variable 17998) (static_member 17999) (unset_variable 18000) (unset_variables 23725) (variable 18002) (variable_class_name 18003)) 10881 (("," 18015) (possible_comma 23726)) 10882 ((")" 23727)) 10883 ((";" 23728)) 10884 ((inner_statement_list 23730)) 10885 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 23731) ("~" 129)) 10886 ((catch_list 23732)) 10887 ((T_CATCH 23733) (T_FINALLY 23734) (finally_statement 23735)) 10888 (("(" 23739)) 10889 (("{" 23736)) 10890 ((inner_statement_list 23737)) 10891 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 23738) ("~" 129)) 10892 ((T_NAME_FULLY_QUALIFIED 18056) (T_NAME_QUALIFIED 18057) (T_NAME_RELATIVE 18058) (T_STATIC 18059) (T_STRING 18060) (catch_name_list 23740) (class_name 18062) (name 18063)) 10893 ((T_VARIABLE 18064) (optional_variable 23741) ("|" 18066)) 10894 ((")" 23742)) 10895 (("{" 23743)) 10896 ((inner_statement_list 23744)) 10897 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 23745) ("~" 129)) 10898 ((backup_doc_comment 23747)) 10899 (("{" 23748)) 10900 ((class_statement_list 23749)) 10901 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 23750)) 10902 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 23752) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 10903 (("%" 304) (")" 23753) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 10904 ((":" 23754) (switch_case_list 23755) ("{" 23756)) 10905 ((";" 23762) (case_list 23763)) 10906 ((";" 23757) (case_list 23758)) 10907 ((case_list 23760)) 10908 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 23759)) 10909 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 23761)) 10910 ((case_list 23766)) 10911 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 23764)) 10912 ((";" 23765)) 10913 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 23767)) 10914 ((";" 23768)) 10915 (("," 18369) (";" 23771)) 10916 ((";" 23773)) 10917 ((T_EXTENDS 23405) (interface_extends_list 23775)) 10918 ((backup_doc_comment 23776)) 10919 (("{" 23777)) 10920 ((class_statement_list 23778)) 10921 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 23779)) 10922 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 23781) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 10923 (("%" 304) (")" 23782) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 10924 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23415) (";" 22800) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22801) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22802) (T_DEC 22) (T_DECLARE 22803) (T_DIR 24) (T_DNUMBER 25) (T_DO 22804) (T_DOUBLE_CAST 27) (T_ECHO 22805) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22806) (T_FOREACH 22807) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22808) (T_GOTO 22809) (T_IF 22764) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22810) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22811) (T_START_HEREDOC 69) (T_STATIC 22812) (T_STRING 22813) (T_STRING_CAST 72) (T_SWITCH 22814) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22815) (T_UNSET 22816) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22817) (T_WHILE 22818) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22819) (alt_if_stmt_without_else 22820) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22821) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22822) (if_stmt_without_else 22823) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23783) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22825) ("~" 129)) 10925 ((")" 23785)) 10926 ((";" 23786)) 10927 ((";" 23788)) 10928 (("," 19446) (";" 23790)) 10929 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 23792) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 10930 (("%" 19534) ("*" 19535) ("+" 19536) ("-" 19537) ("." 19538) ("/" 19539) ("<" 19540) (">" 19541) ("?" 19542) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19543) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19544) (T_AS 23793) (T_BOOLEAN_AND 19545) (T_BOOLEAN_OR 19546) (T_COALESCE 19547) (T_INSTANCEOF 19548) (T_IS_EQUAL 19549) (T_IS_GREATER_OR_EQUAL 19550) (T_IS_IDENTICAL 19551) (T_IS_NOT_EQUAL 19552) (T_IS_NOT_IDENTICAL 19553) (T_IS_SMALLER_OR_EQUAL 19554) (T_LOGICAL_AND 19555) (T_LOGICAL_OR 19556) (T_LOGICAL_XOR 19557) (T_PIPE 19558) (T_POW 19559) (T_SL 19560) (T_SPACESHIP 19561) (T_SR 19562) ("^" 19563) ("|" 19564)) 10931 (("\"" 1538) ("$" 20897) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20898) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20899) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20900) ("[" 20901) (ampersand 20902) (array_object_dereferenceable 20903) (callable_expr 20904) (callable_variable 20905) (class_constant 1564) (class_name 20906) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 23794) (fully_dereferenceable 107) (function_call 20908) (name 20909) (new_dereferenceable 1570) (simple_variable 20910) (static_member 20911) (variable 20912) (variable_class_name 20913)) 10932 ((")" 23795) (T_DOUBLE_ARROW 23796)) 10933 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23799) (";" 22750) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22751) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22753) (T_DEC 22) (T_DECLARE 22754) (T_DIR 24) (T_DNUMBER 25) (T_DO 22755) (T_DOUBLE_CAST 27) (T_ECHO 22756) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22759) (T_FOREACH 22760) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22761) (T_GOTO 22762) (T_IF 22764) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22765) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22767) (T_START_HEREDOC 69) (T_STATIC 22768) (T_STRING 22769) (T_STRING_CAST 72) (T_SWITCH 22770) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22772) (T_UNSET 22773) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22774) (T_WHILE 22775) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22776) (alt_if_stmt_without_else 22777) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22783) (fn 106) (foreach_statement 23805) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22786) (if_stmt_without_else 22787) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23801) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22792) ("~" 129)) 10934 (("\"" 1538) ("$" 20928) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20929) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20930) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20931) ("[" 20932) (ampersand 20933) (array_object_dereferenceable 20934) (callable_expr 20935) (callable_variable 20936) (class_constant 1564) (class_name 20937) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 23797) (fully_dereferenceable 107) (function_call 20939) (name 20940) (new_dereferenceable 1570) (simple_variable 20941) (static_member 20942) (variable 20943) (variable_class_name 20944)) 10935 ((")" 23798)) 10936 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23799) (";" 22750) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22751) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22753) (T_DEC 22) (T_DECLARE 22754) (T_DIR 24) (T_DNUMBER 25) (T_DO 22755) (T_DOUBLE_CAST 27) (T_ECHO 22756) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22759) (T_FOREACH 22760) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22761) (T_GOTO 22762) (T_IF 22764) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22765) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22767) (T_START_HEREDOC 69) (T_STATIC 22768) (T_STRING 22769) (T_STRING_CAST 72) (T_SWITCH 22770) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22772) (T_UNSET 22773) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22774) (T_WHILE 22775) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22776) (alt_if_stmt_without_else 22777) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22783) (fn 106) (foreach_statement 23800) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22786) (if_stmt_without_else 22787) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23801) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22792) ("~" 129)) 10937 ((inner_statement_list 23802)) 10938 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20962) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 20963) (T_CLASS 20964) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20965) (T_DEC 22) (T_DECLARE 20966) (T_DIR 24) (T_DNUMBER 25) (T_DO 20967) (T_DOUBLE_CAST 27) (T_ECHO 20968) (T_EMPTY 29) (T_ENDFOREACH 23803) (T_ENUM 20970) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20971) (T_FOREACH 20972) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20973) (T_GOTO 20974) (T_HALT_COMPILER 20975) (T_IF 20976) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20977) (T_INTERFACE 20978) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20979) (T_START_HEREDOC 69) (T_STATIC 20980) (T_STRING 20981) (T_STRING_CAST 72) (T_SWITCH 20982) (T_THROW 74) (T_TRAIT 20983) (T_TRAIT_C 76) (T_TRY 20984) (T_UNSET 20985) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 20986) (T_WHILE 20987) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 20988) (alt_if_stmt_without_else 20989) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 20990) (attributes 20991) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 20992) (class_modifier 99) (class_modifiers 20993) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 20994) (expr 20995) (fn 106) (fully_dereferenceable 107) #'20996 (function_call 109) (function_declaration_statement 20997) (if_stmt 20998) (if_stmt_without_else 20999) (inline_function 113) (inner_statement 21000) (interface_declaration_statement 21001) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 21002) (static_member 123) (trait_declaration_statement 21003) (variable 126) (variable_class_name 127) ("{" 21004) ("~" 129)) 10939 ((";" 23804)) 10940 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21126) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21127) (fn 18437) (for_exprs 23807) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21129) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 10941 ((";" 23808)) 10942 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21135) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21136) (fn 18437) (for_cond_exprs 23809) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21138) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 10943 ((";" 23810)) 10944 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_VOID_CAST 22737) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 22738) (fn 2425) (for_exprs 23811) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_for_exprs 22740) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 10945 ((")" 23812)) 10946 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23813) (";" 22750) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22751) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22753) (T_DEC 22) (T_DECLARE 22754) (T_DIR 24) (T_DNUMBER 25) (T_DO 22755) (T_DOUBLE_CAST 27) (T_ECHO 22756) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22759) (T_FOREACH 22760) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22761) (T_GOTO 22762) (T_IF 22764) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22765) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22767) (T_START_HEREDOC 69) (T_STATIC 22768) (T_STRING 22769) (T_STRING_CAST 72) (T_SWITCH 22770) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22772) (T_UNSET 22773) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22774) (T_WHILE 22775) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22776) (alt_if_stmt_without_else 22777) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22783) (fn 106) (for_statement 23814) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22786) (if_stmt_without_else 22787) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23815) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22792) ("~" 129)) 10947 ((inner_statement_list 23816)) 10948 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22750) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 22751) (T_CLASS 22752) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22753) (T_DEC 22) (T_DECLARE 22754) (T_DIR 24) (T_DNUMBER 25) (T_DO 22755) (T_DOUBLE_CAST 27) (T_ECHO 22756) (T_EMPTY 29) (T_ENDFOR 23817) (T_ENUM 22758) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 22759) (T_FOREACH 22760) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 22761) (T_GOTO 22762) (T_HALT_COMPILER 22763) (T_IF 22764) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22765) (T_INTERFACE 22766) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22767) (T_START_HEREDOC 69) (T_STATIC 22768) (T_STRING 22769) (T_STRING_CAST 72) (T_SWITCH 22770) (T_THROW 74) (T_TRAIT 22771) (T_TRAIT_C 76) (T_TRY 22772) (T_UNSET 22773) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22774) (T_WHILE 22775) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22776) (alt_if_stmt_without_else 22777) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 22778) (attributes 22779) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 22780) (class_modifier 99) (class_modifiers 22781) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 22782) (expr 22783) (fn 106) (fully_dereferenceable 107) #'22784 (function_call 109) (function_declaration_statement 22785) (if_stmt 22786) (if_stmt_without_else 22787) (inline_function 113) (inner_statement 22788) (interface_declaration_statement 22789) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 22790) (static_member 123) (trait_declaration_statement 22791) (variable 126) (variable_class_name 127) ("{" 22792) ("~" 129)) 10949 ((";" 23818)) 10950 ((":" 23454) (enum_backing_type 23820)) 10951 ((T_IMPLEMENTS 2989) (implements_list 23821)) 10952 ((backup_doc_comment 23822)) 10953 (("{" 23823)) 10954 ((class_statement_list 23824)) 10955 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 23825)) 10956 (("," 22929) (";" 23828)) 10957 ((T_WHILE 23830)) 10958 (("(" 23831)) 10959 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 23832) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 10960 (("%" 304) (")" 23833) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 10961 ((";" 23834)) 10962 ((T_STRING 23106) (const_decl 23107) (const_list 23836)) 10963 ((")" 23837) ("," 23110)) 10964 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23838) (";" 22750) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22751) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22753) (T_DEC 22) (T_DECLARE 22754) (T_DIR 24) (T_DNUMBER 25) (T_DO 22755) (T_DOUBLE_CAST 27) (T_ECHO 22756) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22759) (T_FOREACH 22760) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22761) (T_GOTO 22762) (T_IF 22764) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22765) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22767) (T_START_HEREDOC 69) (T_STATIC 22768) (T_STRING 22769) (T_STRING_CAST 72) (T_SWITCH 22770) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22772) (T_UNSET 22773) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22774) (T_WHILE 22775) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22776) (alt_if_stmt_without_else 22777) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (declare_statement 23839) (dereferenceable_scalar 103) (expr 22783) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22786) (if_stmt_without_else 22787) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23840) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22792) ("~" 129)) 10965 ((inner_statement_list 23841)) 10966 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 23116) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 23117) (T_CLASS 23118) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 23119) (T_DEC 22) (T_DECLARE 23120) (T_DIR 24) (T_DNUMBER 25) (T_DO 23121) (T_DOUBLE_CAST 27) (T_ECHO 23122) (T_EMPTY 29) (T_ENDDECLARE 23842) (T_ENUM 23124) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 23125) (T_FOREACH 23126) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 23127) (T_GOTO 23128) (T_HALT_COMPILER 23129) (T_IF 23130) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 23131) (T_INTERFACE 23132) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 23133) (T_START_HEREDOC 69) (T_STATIC 23134) (T_STRING 23135) (T_STRING_CAST 72) (T_SWITCH 23136) (T_THROW 74) (T_TRAIT 23137) (T_TRAIT_C 76) (T_TRY 23138) (T_UNSET 23139) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 23140) (T_WHILE 23141) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 23142) (alt_if_stmt_without_else 23143) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 23144) (attributes 23145) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 23146) (class_modifier 99) (class_modifiers 23147) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 23148) (expr 23149) (fn 106) (fully_dereferenceable 107) #'23150 (function_call 109) (function_declaration_statement 23151) (if_stmt 23152) (if_stmt_without_else 23153) (inline_function 113) (inner_statement 23154) (interface_declaration_statement 23155) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23156) (static_member 123) (trait_declaration_statement 23157) (variable 126) (variable_class_name 127) ("{" 23158) ("~" 129)) 10967 ((";" 23843)) 10968 ((";" 23845)) 10969 ((T_EXTENDS 2987) (extends_from 23847)) 10970 ((T_IMPLEMENTS 2989) (implements_list 23848)) 10971 ((backup_doc_comment 23849)) 10972 (("{" 23850)) 10973 ((class_statement_list 23851)) 10974 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 23852)) 10975 ((";" 23854)) 10976 (("," 22929) (";" 23859)) 10977 ((T_WHILE 23861)) 10978 (("(" 23862)) 10979 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 23863) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 10980 (("%" 304) (")" 23864) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 10981 ((";" 23865)) 10982 ((T_STRING 23106) (const_decl 23107) (const_list 23867)) 10983 ((")" 23868) ("," 23110)) 10984 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23869) (";" 21012) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 21013) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21014) (T_DEC 22) (T_DECLARE 21015) (T_DIR 24) (T_DNUMBER 25) (T_DO 21016) (T_DOUBLE_CAST 27) (T_ECHO 21017) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 21018) (T_FOREACH 21019) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 21020) (T_GOTO 21021) (T_IF 20976) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21022) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21023) (T_START_HEREDOC 69) (T_STATIC 21024) (T_STRING 21025) (T_STRING_CAST 72) (T_SWITCH 21026) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 21027) (T_UNSET 21028) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 21029) (T_WHILE 21030) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 21031) (alt_if_stmt_without_else 21032) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (declare_statement 23870) (dereferenceable_scalar 103) (expr 21033) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 21034) (if_stmt_without_else 21035) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23871) (static_member 123) (variable 126) (variable_class_name 127) ("{" 21037) ("~" 129)) 10985 ((inner_statement_list 23872)) 10986 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 23116) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 23117) (T_CLASS 23118) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 23119) (T_DEC 22) (T_DECLARE 23120) (T_DIR 24) (T_DNUMBER 25) (T_DO 23121) (T_DOUBLE_CAST 27) (T_ECHO 23122) (T_EMPTY 29) (T_ENDDECLARE 23873) (T_ENUM 23124) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 23125) (T_FOREACH 23126) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 23127) (T_GOTO 23128) (T_HALT_COMPILER 23129) (T_IF 23130) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 23131) (T_INTERFACE 23132) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 23133) (T_START_HEREDOC 69) (T_STATIC 23134) (T_STRING 23135) (T_STRING_CAST 72) (T_SWITCH 23136) (T_THROW 74) (T_TRAIT 23137) (T_TRAIT_C 76) (T_TRY 23138) (T_UNSET 23139) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 23140) (T_WHILE 23141) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 23142) (alt_if_stmt_without_else 23143) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 23144) (attributes 23145) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 23146) (class_modifier 99) (class_modifiers 23147) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 23148) (expr 23149) (fn 106) (fully_dereferenceable 107) #'23150 (function_call 109) (function_declaration_statement 23151) (if_stmt 23152) (if_stmt_without_else 23153) (inline_function 113) (inner_statement 23154) (interface_declaration_statement 23155) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23156) (static_member 123) (trait_declaration_statement 23157) (variable 126) (variable_class_name 127) ("{" 23158) ("~" 129)) 10987 ((";" 23874)) 10988 ((";" 23876)) 10989 ((";" 23878)) 10990 ((T_READONLY 17774) (T_STRING 17775) (backup_doc_comment 143) (function_name 23881)) 10991 ((backup_doc_comment 23882)) 10992 (("(" 23883)) 10993 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 23884)) 10994 ((")" 23885)) 10995 ((":" 168) (return_type 23886)) 10996 ((backup_fn_flags 23887)) 10997 (("{" 23888)) 10998 ((inner_statement_list 23889)) 10999 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 23890) ("~" 129)) 11000 ((backup_fn_flags 23891)) 11001 ((T_STRING 23894)) 11002 ((T_EXTENDS 2987) (extends_from 23895)) 11003 ((T_IMPLEMENTS 2989) (implements_list 23896)) 11004 ((backup_doc_comment 23897)) 11005 (("{" 23898)) 11006 ((class_statement_list 23899)) 11007 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 23900)) 11008 ((":" 23905)) 11009 ((";" 23904)) 11010 ((inner_statement_list 23906)) 11011 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17807) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17808) (T_CLASS 17809) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17810) (T_DEC 22) (T_DECLARE 17811) (T_DIR 24) (T_DNUMBER 25) (T_DO 17812) (T_DOUBLE_CAST 27) (T_ECHO 17813) (T_EMPTY 29) (T_ENDIF 23907) (T_ENUM 17815) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17816) (T_FOREACH 17817) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17818) (T_GOTO 17819) (T_HALT_COMPILER 17820) (T_IF 17732) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17821) (T_INTERFACE 17822) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17823) (T_START_HEREDOC 69) (T_STATIC 17824) (T_STRING 17825) (T_STRING_CAST 72) (T_SWITCH 17826) (T_THROW 74) (T_TRAIT 17827) (T_TRAIT_C 76) (T_TRY 17828) (T_UNSET 17829) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17830) (T_WHILE 17831) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17832) (alt_if_stmt_without_else 17833) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17834) (attributes 17835) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17836) (class_modifier 99) (class_modifiers 17837) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17838) (expr 17839) (fn 106) (fully_dereferenceable 107) #'17840 (function_call 109) (function_declaration_statement 17841) (if_stmt 17842) (if_stmt_without_else 17843) (inline_function 113) (inner_statement 17844) (interface_declaration_statement 17845) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17846) (static_member 123) (trait_declaration_statement 17847) (variable 126) (variable_class_name 127) ("{" 17848) ("~" 129)) 11012 ((";" 23908)) 11013 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 23910) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 11014 (("%" 304) (")" 23911) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 11015 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23912) (";" 20962) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 20963) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20965) (T_DEC 22) (T_DECLARE 20966) (T_DIR 24) (T_DNUMBER 25) (T_DO 20967) (T_DOUBLE_CAST 27) (T_ECHO 20968) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20971) (T_FOREACH 20972) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 20973) (T_GOTO 20974) (T_IF 20976) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20977) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20979) (T_START_HEREDOC 69) (T_STATIC 20980) (T_STRING 20981) (T_STRING_CAST 72) (T_SWITCH 20982) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20984) (T_UNSET 20985) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 20986) (T_WHILE 20987) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 20988) (alt_if_stmt_without_else 20989) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 20995) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 20998) (if_stmt_without_else 20999) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23913) (static_member 123) (variable 126) (variable_class_name 127) (while_statement 23914) ("{" 21004) ("~" 129)) 11016 ((inner_statement_list 23915)) 11017 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17889) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17890) (T_CLASS 17891) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17892) (T_DEC 22) (T_DECLARE 17893) (T_DIR 24) (T_DNUMBER 25) (T_DO 17894) (T_DOUBLE_CAST 27) (T_ECHO 17895) (T_EMPTY 29) (T_ENDWHILE 23916) (T_ENUM 17897) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17898) (T_FOREACH 17899) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17900) (T_GOTO 17901) (T_HALT_COMPILER 17902) (T_IF 17903) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17904) (T_INTERFACE 17905) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17906) (T_START_HEREDOC 69) (T_STATIC 17907) (T_STRING 17908) (T_STRING_CAST 72) (T_SWITCH 17909) (T_THROW 74) (T_TRAIT 17910) (T_TRAIT_C 76) (T_TRY 17911) (T_UNSET 17912) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17913) (T_WHILE 17914) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17915) (alt_if_stmt_without_else 17916) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17917) (attributes 17918) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17919) (class_modifier 99) (class_modifiers 17920) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17921) (expr 17922) (fn 106) (fully_dereferenceable 107) #'17923 (function_call 109) (function_declaration_statement 17924) (if_stmt 17925) (if_stmt_without_else 17926) (inline_function 113) (inner_statement 17927) (interface_declaration_statement 17928) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17929) (static_member 123) (trait_declaration_statement 17930) (variable 126) (variable_class_name 127) ("{" 17931) ("~" 129)) 11018 ((";" 23917)) 11019 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 23919) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 11020 (("\"" 1538) ("$" 17989) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 17990) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 17991) ("[" 1560) (array_object_dereferenceable 17992) (callable_expr 17993) (callable_variable 17994) (class_constant 1564) (class_name 17995) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 17996) (name 17997) (new_dereferenceable 1570) (simple_variable 17998) (static_member 17999) (unset_variable 18000) (unset_variables 23921) (variable 18002) (variable_class_name 18003)) 11021 (("," 18015) (possible_comma 23922)) 11022 ((")" 23923)) 11023 ((";" 23924)) 11024 ((inner_statement_list 23926)) 11025 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 23927) ("~" 129)) 11026 ((catch_list 23928)) 11027 ((T_CATCH 23929) (T_FINALLY 23930) (finally_statement 23931)) 11028 (("(" 23935)) 11029 (("{" 23932)) 11030 ((inner_statement_list 23933)) 11031 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 23934) ("~" 129)) 11032 ((T_NAME_FULLY_QUALIFIED 18056) (T_NAME_QUALIFIED 18057) (T_NAME_RELATIVE 18058) (T_STATIC 18059) (T_STRING 18060) (catch_name_list 23936) (class_name 18062) (name 18063)) 11033 ((T_VARIABLE 18064) (optional_variable 23937) ("|" 18066)) 11034 ((")" 23938)) 11035 (("{" 23939)) 11036 ((inner_statement_list 23940)) 11037 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 23941) ("~" 129)) 11038 ((backup_doc_comment 23943)) 11039 (("{" 23944)) 11040 ((class_statement_list 23945)) 11041 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 23946)) 11042 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 23948) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 11043 (("%" 304) (")" 23949) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 11044 ((":" 23950) (switch_case_list 23951) ("{" 23952)) 11045 ((";" 23958) (case_list 23959)) 11046 ((";" 23953) (case_list 23954)) 11047 ((case_list 23956)) 11048 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 23955)) 11049 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 23957)) 11050 ((case_list 23962)) 11051 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 23960)) 11052 ((";" 23961)) 11053 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 23963)) 11054 ((";" 23964)) 11055 (("," 18369) (";" 23967)) 11056 ((";" 23969)) 11057 ((T_EXTENDS 23405) (interface_extends_list 23971)) 11058 ((backup_doc_comment 23972)) 11059 (("{" 23973)) 11060 ((class_statement_list 23974)) 11061 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 23975)) 11062 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 23977) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 11063 (("%" 304) (")" 23978) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 11064 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23415) (";" 21012) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 21013) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21014) (T_DEC 22) (T_DECLARE 21015) (T_DIR 24) (T_DNUMBER 25) (T_DO 21016) (T_DOUBLE_CAST 27) (T_ECHO 21017) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 21018) (T_FOREACH 21019) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 21020) (T_GOTO 21021) (T_IF 20976) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 21022) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 21023) (T_START_HEREDOC 69) (T_STATIC 21024) (T_STRING 21025) (T_STRING_CAST 72) (T_SWITCH 21026) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 21027) (T_UNSET 21028) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 21029) (T_WHILE 21030) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 21031) (alt_if_stmt_without_else 21032) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 21033) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 21034) (if_stmt_without_else 21035) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23979) (static_member 123) (variable 126) (variable_class_name 127) ("{" 21037) ("~" 129)) 11065 ((")" 23981)) 11066 ((";" 23982)) 11067 ((";" 23984)) 11068 (("," 19446) (";" 23986)) 11069 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 23988) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 11070 (("%" 19534) ("*" 19535) ("+" 19536) ("-" 19537) ("." 19538) ("/" 19539) ("<" 19540) (">" 19541) ("?" 19542) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19543) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19544) (T_AS 23989) (T_BOOLEAN_AND 19545) (T_BOOLEAN_OR 19546) (T_COALESCE 19547) (T_INSTANCEOF 19548) (T_IS_EQUAL 19549) (T_IS_GREATER_OR_EQUAL 19550) (T_IS_IDENTICAL 19551) (T_IS_NOT_EQUAL 19552) (T_IS_NOT_IDENTICAL 19553) (T_IS_SMALLER_OR_EQUAL 19554) (T_LOGICAL_AND 19555) (T_LOGICAL_OR 19556) (T_LOGICAL_XOR 19557) (T_PIPE 19558) (T_POW 19559) (T_SL 19560) (T_SPACESHIP 19561) (T_SR 19562) ("^" 19563) ("|" 19564)) 11071 (("\"" 1538) ("$" 20897) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20898) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20899) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20900) ("[" 20901) (ampersand 20902) (array_object_dereferenceable 20903) (callable_expr 20904) (callable_variable 20905) (class_constant 1564) (class_name 20906) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 23990) (fully_dereferenceable 107) (function_call 20908) (name 20909) (new_dereferenceable 1570) (simple_variable 20910) (static_member 20911) (variable 20912) (variable_class_name 20913)) 11072 ((")" 23991) (T_DOUBLE_ARROW 23992)) 11073 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23995) (";" 20962) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 20963) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20965) (T_DEC 22) (T_DECLARE 20966) (T_DIR 24) (T_DNUMBER 25) (T_DO 20967) (T_DOUBLE_CAST 27) (T_ECHO 20968) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20971) (T_FOREACH 20972) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 20973) (T_GOTO 20974) (T_IF 20976) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20977) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20979) (T_START_HEREDOC 69) (T_STATIC 20980) (T_STRING 20981) (T_STRING_CAST 72) (T_SWITCH 20982) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20984) (T_UNSET 20985) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 20986) (T_WHILE 20987) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 20988) (alt_if_stmt_without_else 20989) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 20995) (fn 106) (foreach_statement 24001) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 20998) (if_stmt_without_else 20999) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23997) (static_member 123) (variable 126) (variable_class_name 127) ("{" 21004) ("~" 129)) 11074 (("\"" 1538) ("$" 20928) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20929) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20930) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20931) ("[" 20932) (ampersand 20933) (array_object_dereferenceable 20934) (callable_expr 20935) (callable_variable 20936) (class_constant 1564) (class_name 20937) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 23993) (fully_dereferenceable 107) (function_call 20939) (name 20940) (new_dereferenceable 1570) (simple_variable 20941) (static_member 20942) (variable 20943) (variable_class_name 20944)) 11075 ((")" 23994)) 11076 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23995) (";" 20962) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 20963) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20965) (T_DEC 22) (T_DECLARE 20966) (T_DIR 24) (T_DNUMBER 25) (T_DO 20967) (T_DOUBLE_CAST 27) (T_ECHO 20968) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20971) (T_FOREACH 20972) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 20973) (T_GOTO 20974) (T_IF 20976) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20977) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20979) (T_START_HEREDOC 69) (T_STATIC 20980) (T_STRING 20981) (T_STRING_CAST 72) (T_SWITCH 20982) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20984) (T_UNSET 20985) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 20986) (T_WHILE 20987) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 20988) (alt_if_stmt_without_else 20989) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 20995) (fn 106) (foreach_statement 23996) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 20998) (if_stmt_without_else 20999) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23997) (static_member 123) (variable 126) (variable_class_name 127) ("{" 21004) ("~" 129)) 11077 ((inner_statement_list 23998)) 11078 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20962) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 20963) (T_CLASS 20964) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20965) (T_DEC 22) (T_DECLARE 20966) (T_DIR 24) (T_DNUMBER 25) (T_DO 20967) (T_DOUBLE_CAST 27) (T_ECHO 20968) (T_EMPTY 29) (T_ENDFOREACH 23999) (T_ENUM 20970) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20971) (T_FOREACH 20972) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20973) (T_GOTO 20974) (T_HALT_COMPILER 20975) (T_IF 20976) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20977) (T_INTERFACE 20978) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20979) (T_START_HEREDOC 69) (T_STATIC 20980) (T_STRING 20981) (T_STRING_CAST 72) (T_SWITCH 20982) (T_THROW 74) (T_TRAIT 20983) (T_TRAIT_C 76) (T_TRY 20984) (T_UNSET 20985) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 20986) (T_WHILE 20987) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 20988) (alt_if_stmt_without_else 20989) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 20990) (attributes 20991) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 20992) (class_modifier 99) (class_modifiers 20993) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 20994) (expr 20995) (fn 106) (fully_dereferenceable 107) #'20996 (function_call 109) (function_declaration_statement 20997) (if_stmt 20998) (if_stmt_without_else 20999) (inline_function 113) (inner_statement 21000) (interface_declaration_statement 21001) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 21002) (static_member 123) (trait_declaration_statement 21003) (variable 126) (variable_class_name 127) ("{" 21004) ("~" 129)) 11079 ((";" 24000)) 11080 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21126) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21127) (fn 18437) (for_exprs 24003) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21129) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 11081 ((";" 24004)) 11082 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21135) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21136) (fn 18437) (for_cond_exprs 24005) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21138) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 11083 ((";" 24006)) 11084 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_VOID_CAST 22737) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 22738) (fn 2425) (for_exprs 24007) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_for_exprs 22740) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 11085 ((")" 24008)) 11086 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 24009) (";" 20962) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 20963) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20965) (T_DEC 22) (T_DECLARE 20966) (T_DIR 24) (T_DNUMBER 25) (T_DO 20967) (T_DOUBLE_CAST 27) (T_ECHO 20968) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20971) (T_FOREACH 20972) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 20973) (T_GOTO 20974) (T_IF 20976) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20977) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20979) (T_START_HEREDOC 69) (T_STATIC 20980) (T_STRING 20981) (T_STRING_CAST 72) (T_SWITCH 20982) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20984) (T_UNSET 20985) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 20986) (T_WHILE 20987) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 20988) (alt_if_stmt_without_else 20989) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 20995) (fn 106) (for_statement 24010) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 20998) (if_stmt_without_else 20999) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 24011) (static_member 123) (variable 126) (variable_class_name 127) ("{" 21004) ("~" 129)) 11087 ((inner_statement_list 24012)) 11088 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22750) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 22751) (T_CLASS 22752) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22753) (T_DEC 22) (T_DECLARE 22754) (T_DIR 24) (T_DNUMBER 25) (T_DO 22755) (T_DOUBLE_CAST 27) (T_ECHO 22756) (T_EMPTY 29) (T_ENDFOR 24013) (T_ENUM 22758) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 22759) (T_FOREACH 22760) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 22761) (T_GOTO 22762) (T_HALT_COMPILER 22763) (T_IF 22764) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22765) (T_INTERFACE 22766) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22767) (T_START_HEREDOC 69) (T_STATIC 22768) (T_STRING 22769) (T_STRING_CAST 72) (T_SWITCH 22770) (T_THROW 74) (T_TRAIT 22771) (T_TRAIT_C 76) (T_TRY 22772) (T_UNSET 22773) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22774) (T_WHILE 22775) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22776) (alt_if_stmt_without_else 22777) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 22778) (attributes 22779) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 22780) (class_modifier 99) (class_modifiers 22781) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 22782) (expr 22783) (fn 106) (fully_dereferenceable 107) #'22784 (function_call 109) (function_declaration_statement 22785) (if_stmt 22786) (if_stmt_without_else 22787) (inline_function 113) (inner_statement 22788) (interface_declaration_statement 22789) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 22790) (static_member 123) (trait_declaration_statement 22791) (variable 126) (variable_class_name 127) ("{" 22792) ("~" 129)) 11089 ((";" 24014)) 11090 ((":" 23454) (enum_backing_type 24016)) 11091 ((T_IMPLEMENTS 2989) (implements_list 24017)) 11092 ((backup_doc_comment 24018)) 11093 (("{" 24019)) 11094 ((class_statement_list 24020)) 11095 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 24021)) 11096 (("," 22929) (";" 24024)) 11097 ((T_WHILE 24026)) 11098 (("(" 24027)) 11099 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 24028) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 11100 (("%" 304) (")" 24029) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 11101 ((";" 24030)) 11102 ((T_STRING 23106) (const_decl 23107) (const_list 24032)) 11103 ((")" 24033) ("," 23110)) 11104 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 24034) (";" 20962) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 20963) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20965) (T_DEC 22) (T_DECLARE 20966) (T_DIR 24) (T_DNUMBER 25) (T_DO 20967) (T_DOUBLE_CAST 27) (T_ECHO 20968) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 20971) (T_FOREACH 20972) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 20973) (T_GOTO 20974) (T_IF 20976) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20977) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20979) (T_START_HEREDOC 69) (T_STATIC 20980) (T_STRING 20981) (T_STRING_CAST 72) (T_SWITCH 20982) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 20984) (T_UNSET 20985) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 20986) (T_WHILE 20987) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 20988) (alt_if_stmt_without_else 20989) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (declare_statement 24035) (dereferenceable_scalar 103) (expr 20995) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 20998) (if_stmt_without_else 20999) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 24036) (static_member 123) (variable 126) (variable_class_name 127) ("{" 21004) ("~" 129)) 11105 ((inner_statement_list 24037)) 11106 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 23116) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 23117) (T_CLASS 23118) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 23119) (T_DEC 22) (T_DECLARE 23120) (T_DIR 24) (T_DNUMBER 25) (T_DO 23121) (T_DOUBLE_CAST 27) (T_ECHO 23122) (T_EMPTY 29) (T_ENDDECLARE 24038) (T_ENUM 23124) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 23125) (T_FOREACH 23126) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 23127) (T_GOTO 23128) (T_HALT_COMPILER 23129) (T_IF 23130) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 23131) (T_INTERFACE 23132) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 23133) (T_START_HEREDOC 69) (T_STATIC 23134) (T_STRING 23135) (T_STRING_CAST 72) (T_SWITCH 23136) (T_THROW 74) (T_TRAIT 23137) (T_TRAIT_C 76) (T_TRY 23138) (T_UNSET 23139) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 23140) (T_WHILE 23141) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 23142) (alt_if_stmt_without_else 23143) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 23144) (attributes 23145) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 23146) (class_modifier 99) (class_modifiers 23147) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 23148) (expr 23149) (fn 106) (fully_dereferenceable 107) #'23150 (function_call 109) (function_declaration_statement 23151) (if_stmt 23152) (if_stmt_without_else 23153) (inline_function 113) (inner_statement 23154) (interface_declaration_statement 23155) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23156) (static_member 123) (trait_declaration_statement 23157) (variable 126) (variable_class_name 127) ("{" 23158) ("~" 129)) 11107 ((";" 24039)) 11108 ((";" 24041)) 11109 ((T_EXTENDS 2987) (extends_from 24043)) 11110 ((T_IMPLEMENTS 2989) (implements_list 24044)) 11111 ((backup_doc_comment 24045)) 11112 (("{" 24046)) 11113 ((class_statement_list 24047)) 11114 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 24048)) 11115 ((";" 24050)) 11116 (("$" 20928) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 20931) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 24052) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 24053) ("{" 1676)) 11117 (("(" 20948) (argument_list 24054)) 11118 (("$" 20928) (T_STRING 24061) (T_VARIABLE 20931) (property_name 24068) (simple_variable 24063) ("{" 24064)) 11119 (("$" 20928) (T_STRING 24061) (T_VARIABLE 20931) (property_name 24062) (simple_variable 24063) ("{" 24064)) 11120 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 24059) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 11121 (("]" 24060)) 11122 (("(" 20948) (argument_list 24067)) 11123 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 24065) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 11124 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 24066)) 11125 (("(" 20948) (argument_list 24069)) 11126 (("]" 24072)) 11127 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 24075) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 11128 ((")" 24076)) 11129 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 24079) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 11130 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 24080)) 11131 (("$" 20897) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 20900) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 24083) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 24084) ("{" 1676)) 11132 (("(" 20917) (argument_list 24085)) 11133 (("$" 20897) (T_STRING 24092) (T_VARIABLE 20900) (property_name 24099) (simple_variable 24094) ("{" 24095)) 11134 (("$" 20897) (T_STRING 24092) (T_VARIABLE 20900) (property_name 24093) (simple_variable 24094) ("{" 24095)) 11135 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 24090) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 11136 (("]" 24091)) 11137 (("(" 20917) (argument_list 24098)) 11138 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 24096) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 11139 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 24097)) 11140 (("(" 20917) (argument_list 24100)) 11141 (("]" 24103)) 11142 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 24106) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 11143 ((")" 24107)) 11144 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 24110) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 11145 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 24111)) 11146 (("$" 19456) (T_ABSTRACT 19635) (T_ARRAY 19636) (T_AS 19637) (T_BREAK 19638) (T_CALLABLE 19639) (T_CASE 19640) (T_CATCH 19641) (T_CLASS 19642) (T_CLASS_C 19643) (T_CLONE 19644) (T_CONST 19645) (T_CONTINUE 19646) (T_DECLARE 19647) (T_DEFAULT 19648) (T_DIR 19649) (T_DO 19650) (T_ECHO 19651) (T_ELSE 19652) (T_ELSEIF 19653) (T_EMPTY 19654) (T_ENDDECLARE 19655) (T_ENDFOR 19656) (T_ENDFOREACH 19657) (T_ENDIF 19658) (T_ENDSWITCH 19659) (T_ENDWHILE 19660) (T_ENUM 19661) (T_EVAL 19662) (T_EXIT 19663) (T_EXTENDS 19664) (T_FILE 19665) (T_FINAL 19666) (T_FINALLY 19667) (T_FN 19668) (T_FOR 19669) (T_FOREACH 19670) (T_FUNCTION 19671) (T_FUNC_C 19672) (T_GLOBAL 19673) (T_GOTO 19674) (T_IF 19675) (T_IMPLEMENTS 19676) (T_INCLUDE 19677) (T_INCLUDE_ONCE 19678) (T_INSTANCEOF 19679) (T_INSTEADOF 19680) (T_INTERFACE 19681) (T_ISSET 19682) (T_LINE 19683) (T_LIST 19684) (T_LOGICAL_AND 19685) (T_LOGICAL_OR 19686) (T_LOGICAL_XOR 19687) (T_MATCH 19688) (T_METHOD_C 19689) (T_NAMESPACE 19690) (T_NEW 19691) (T_NS_C 19692) (T_PRINT 19693) (T_PRIVATE 19694) (T_PROPERTY_C 19695) (T_PROTECTED 19696) (T_PUBLIC 19697) (T_READONLY 19698) (T_REQUIRE 19699) (T_REQUIRE_ONCE 19700) (T_RETURN 19701) (T_STATIC 19702) (T_STRING 19703) (T_SWITCH 19704) (T_THROW 19705) (T_TRAIT 19706) (T_TRAIT_C 19707) (T_TRY 19708) (T_UNSET 19709) (T_USE 19710) (T_VAR 19711) (T_VARIABLE 19504) (T_WHILE 19712) (T_YIELD 19713) (identifier 24113) (member_name 24114) (reserved_non_modifiers 19716) (semi_reserved 19717) (simple_variable 24115) ("{" 24116)) 11147 (("(" 19722) (argument_list 24119)) 11148 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 24117) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 11149 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 24118)) 11150 ((T_FN 35) (T_FUNCTION 131) (fn 19518) #'19519 (inline_function 24123)) 11151 (("$" 19456) (T_STRING 24129) (T_VARIABLE 19504) (property_name 24136) (simple_variable 24131) ("{" 24132)) 11152 (("$" 19456) (T_STRING 24129) (T_VARIABLE 19504) (property_name 24130) (simple_variable 24131) ("{" 24132)) 11153 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 24127) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 11154 (("]" 24128)) 11155 (("(" 19722) (argument_list 24135)) 11156 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 24133) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 11157 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 24134)) 11158 (("(" 19722) (argument_list 24137)) 11159 (("`" 24139)) 11160 (("]" 24141)) 11161 (("=" 24142)) 11162 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 24143) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 11163 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24678) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11164 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 24676) (encaps_var 5357)) 11165 (("$" 24147) (T_VARIABLE 24195) (simple_variable 24672) ("{" 24673)) 11166 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 24671) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 11167 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24670) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11168 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24669) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11169 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24668) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11170 (("(" 24665)) 11171 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24664) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11172 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24663) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11173 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24650) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (clone_argument_list 24651) (constant 24206) (dereferenceable_scalar 24207) (expr 24652) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11174 (("\"" 1538) ("$" 24452) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 24453) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 24454) ("[" 1560) (array_object_dereferenceable 24455) (callable_expr 24456) (callable_variable 24457) (class_constant 1564) (class_name 24458) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 24459) (name 24460) (new_dereferenceable 1570) (simple_variable 24461) (static_member 24462) (variable 24649) (variable_class_name 24464)) 11175 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24648) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11176 (("(" 24645)) 11177 (("(" 24642)) 11178 (("(" 24633) (argument_list 24634) (ctor_arguments 24635)) 11179 (("\"" 1538) ("$" 24452) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 24453) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 24454) ("[" 1560) (array_object_dereferenceable 24455) (callable_expr 24456) (callable_variable 24457) (class_constant 1564) (class_name 24458) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 24459) (name 24460) (new_dereferenceable 1570) (simple_variable 24461) (static_member 24462) (variable 24632) (variable_class_name 24464)) 11180 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24631) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11181 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24630) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11182 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24629) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11183 (("(" 24625)) 11184 (("(" 24620)) 11185 (("(" 24614)) 11186 (("$" 24452) ("(" 24577) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 24578) (T_NAME_QUALIFIED 24579) (T_NAME_RELATIVE 24580) (T_READONLY 2955) (T_STATIC 24581) (T_STRING 24582) (T_VARIABLE 24454) (anonymous_class 24583) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 24584) (attribute 2961) (attributes 24585) (class_modifier 2963) (class_name 24586) (class_name_reference 24587) (name 24588) (new_variable 24589) (simple_variable 24590)) 11187 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24576) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11188 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24575) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11189 (("(" 24413) (argument_list 24574)) 11190 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24573) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11191 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24572) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11192 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 24567) (T_END_HEREDOC 24568) (T_VARIABLE 2920) (encaps_list 24569) (encaps_var 2922)) 11193 ((T_FN 35) (T_FUNCTION 131) (fn 24209) #'24210 (inline_function 24566)) 11194 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24565) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11195 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24564) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11196 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24563) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11197 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24560) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11198 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24559) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11199 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 24555) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 11200 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 24553) (encaps_list 2828) (encaps_var 2829)) 11201 ((T_NULLSAFE_OBJECT_OPERATOR 24539) (T_OBJECT_OPERATOR 24540) ("[" 24541)) 11202 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 24536) (attribute 2757) (fn 24209) #'24210 (inline_function 24537)) 11203 (("(" 24413) (argument_list 24535)) 11204 ((T_PAAMAYIM_NEKUDOTAYIM 24527)) 11205 (("%" 24225) ("*" 24226) ("+" 24227) ("-" 24228) ("." 24229) ("/" 24230) ("<" 24231) (">" 24232) ("?" 24233) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 24234) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 24235) (T_BOOLEAN_AND 24236) (T_BOOLEAN_OR 24237) (T_COALESCE 24238) (T_DOUBLE_ARROW 24525) (T_INSTANCEOF 24239) (T_IS_EQUAL 24240) (T_IS_GREATER_OR_EQUAL 24241) (T_IS_IDENTICAL 24242) (T_IS_NOT_EQUAL 24243) (T_IS_NOT_IDENTICAL 24244) (T_IS_SMALLER_OR_EQUAL 24245) (T_LOGICAL_AND 24246) (T_LOGICAL_OR 24247) (T_LOGICAL_XOR 24248) (T_PIPE 24249) (T_POW 24250) (T_SL 24251) (T_SPACESHIP 24252) (T_SR 24253) ("^" 24254) ("|" 24255)) 11206 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 24514)) 11207 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 24502)) 11208 (("(" 24413) (argument_list 24501)) 11209 (("=" 24421) (T_AND_EQUAL 24422) (T_COALESCE_EQUAL 24423) (T_CONCAT_EQUAL 24424) (T_DEC 24425) (T_DIV_EQUAL 24426) (T_INC 24427) (T_MINUS_EQUAL 24428) (T_MOD_EQUAL 24429) (T_MUL_EQUAL 24430) (T_OR_EQUAL 24431) (T_PLUS_EQUAL 24432) (T_POW_EQUAL 24433) (T_SL_EQUAL 24434) (T_SR_EQUAL 24435) (T_XOR_EQUAL 24436)) 11210 ((T_PAAMAYIM_NEKUDOTAYIM 24325)) 11211 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24224) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11212 (("%" 24225) ("*" 24226) ("+" 24227) ("-" 24228) ("." 24229) ("/" 24230) ("<" 24231) (">" 24232) ("?" 24233) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 24234) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 24235) (T_BOOLEAN_AND 24236) (T_BOOLEAN_OR 24237) (T_COALESCE 24238) (T_INSTANCEOF 24239) (T_IS_EQUAL 24240) (T_IS_GREATER_OR_EQUAL 24241) (T_IS_IDENTICAL 24242) (T_IS_NOT_EQUAL 24243) (T_IS_NOT_IDENTICAL 24244) (T_IS_SMALLER_OR_EQUAL 24245) (T_LOGICAL_AND 24246) (T_LOGICAL_OR 24247) (T_LOGICAL_XOR 24248) (T_PIPE 24249) (T_POW 24250) (T_SL 24251) (T_SPACESHIP 24252) (T_SR 24253) ("^" 24254) ("|" 24255)) 11213 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24324) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11214 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24323) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11215 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24322) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11216 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24321) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11217 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24320) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11218 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24319) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11219 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24318) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11220 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24317) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11221 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 24312) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 24313) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 11222 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24311) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11223 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24310) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11224 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24309) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11225 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24308) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11226 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24307) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11227 (("$" 24272) ("(" 24273) (T_NAME_FULLY_QUALIFIED 24274) (T_NAME_QUALIFIED 24275) (T_NAME_RELATIVE 24276) (T_STATIC 24277) (T_STRING 24278) (T_VARIABLE 24279) (class_name 24280) (class_name_reference 24281) (name 24282) (new_variable 24283) (simple_variable 24284)) 11228 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24271) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11229 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24270) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11230 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24269) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11231 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24268) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11232 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24267) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11233 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24266) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11234 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24265) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11235 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24264) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11236 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24263) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11237 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24262) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11238 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24261) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11239 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24260) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11240 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24259) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11241 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24258) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11242 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24257) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11243 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24256) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11244 (("$" 24272) (T_VARIABLE 24279) (simple_variable 24303) ("{" 24304)) 11245 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 24301) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 11246 ((T_PAAMAYIM_NEKUDOTAYIM 24299)) 11247 ((T_NULLSAFE_OBJECT_OPERATOR 24285) (T_OBJECT_OPERATOR 24286) (T_PAAMAYIM_NEKUDOTAYIM 24287) ("[" 24288)) 11248 (("$" 24272) (T_STRING 24292) (T_VARIABLE 24279) (property_name 24298) (simple_variable 24294) ("{" 24295)) 11249 (("$" 24272) (T_STRING 24292) (T_VARIABLE 24279) (property_name 24293) (simple_variable 24294) ("{" 24295)) 11250 (("$" 24272) (T_VARIABLE 24279) (simple_variable 24291)) 11251 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 24289) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 11252 (("]" 24290)) 11253 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 24296) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 11254 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 24297)) 11255 (("$" 24272) (T_VARIABLE 24279) (simple_variable 24300)) 11256 (("%" 304) (")" 24302) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 11257 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 24305) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 11258 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 24306)) 11259 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24316) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11260 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 24314) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 11261 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24315) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11262 (("$" 24147) (T_ABSTRACT 24326) (T_ARRAY 24327) (T_AS 24328) (T_BREAK 24329) (T_CALLABLE 24330) (T_CASE 24331) (T_CATCH 24332) (T_CLASS 24333) (T_CLASS_C 24334) (T_CLONE 24335) (T_CONST 24336) (T_CONTINUE 24337) (T_DECLARE 24338) (T_DEFAULT 24339) (T_DIR 24340) (T_DO 24341) (T_ECHO 24342) (T_ELSE 24343) (T_ELSEIF 24344) (T_EMPTY 24345) (T_ENDDECLARE 24346) (T_ENDFOR 24347) (T_ENDFOREACH 24348) (T_ENDIF 24349) (T_ENDSWITCH 24350) (T_ENDWHILE 24351) (T_ENUM 24352) (T_EVAL 24353) (T_EXIT 24354) (T_EXTENDS 24355) (T_FILE 24356) (T_FINAL 24357) (T_FINALLY 24358) (T_FN 24359) (T_FOR 24360) (T_FOREACH 24361) (T_FUNCTION 24362) (T_FUNC_C 24363) (T_GLOBAL 24364) (T_GOTO 24365) (T_IF 24366) (T_IMPLEMENTS 24367) (T_INCLUDE 24368) (T_INCLUDE_ONCE 24369) (T_INSTANCEOF 24370) (T_INSTEADOF 24371) (T_INTERFACE 24372) (T_ISSET 24373) (T_LINE 24374) (T_LIST 24375) (T_LOGICAL_AND 24376) (T_LOGICAL_OR 24377) (T_LOGICAL_XOR 24378) (T_MATCH 24379) (T_METHOD_C 24380) (T_NAMESPACE 24381) (T_NEW 24382) (T_NS_C 24383) (T_PRINT 24384) (T_PRIVATE 24385) (T_PROPERTY_C 24386) (T_PROTECTED 24387) (T_PUBLIC 24388) (T_READONLY 24389) (T_REQUIRE 24390) (T_REQUIRE_ONCE 24391) (T_RETURN 24392) (T_STATIC 24393) (T_STRING 24394) (T_SWITCH 24395) (T_THROW 24396) (T_TRAIT 24397) (T_TRAIT_C 24398) (T_TRY 24399) (T_UNSET 24400) (T_USE 24401) (T_VAR 24402) (T_VARIABLE 24195) (T_WHILE 24403) (T_YIELD 24404) (identifier 24405) (member_name 24406) (reserved_non_modifiers 24407) (semi_reserved 24408) (simple_variable 24409) ("{" 24410)) 11263 (("(" 24413) (argument_list 24414)) 11264 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 24411) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 11265 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 24412)) 11266 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 24415) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 24416) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 24417) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 11267 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 24420) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 11268 (("," 1498) (possible_comma 24418)) 11269 ((")" 24419)) 11270 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (ampersand 24450) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24451) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11271 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24449) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11272 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24448) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11273 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24447) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11274 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24446) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11275 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24445) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11276 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24444) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11277 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24443) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11278 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24442) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11279 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24441) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11280 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24440) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11281 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24439) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11282 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24438) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11283 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24437) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11284 (("\"" 1538) ("$" 24452) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 24453) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 24454) ("[" 1560) (array_object_dereferenceable 24455) (callable_expr 24456) (callable_variable 24457) (class_constant 1564) (class_name 24458) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 24459) (name 24460) (new_dereferenceable 1570) (simple_variable 24461) (static_member 24462) (variable 24463) (variable_class_name 24464)) 11285 (("$" 24452) (T_VARIABLE 24454) (simple_variable 24497) ("{" 24498)) 11286 (("(" 24468) (argument_list 24496)) 11287 ((T_NULLSAFE_OBJECT_OPERATOR 24482) (T_OBJECT_OPERATOR 24483) ("[" 24484)) 11288 (("(" 24468) (argument_list 24481)) 11289 ((T_PAAMAYIM_NEKUDOTAYIM 24477)) 11290 (("(" 24468) (argument_list 24476)) 11291 ((T_PAAMAYIM_NEKUDOTAYIM 24465)) 11292 (("$" 24452) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 24454) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 24466) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 24467) ("{" 1660)) 11293 (("(" 24468) (argument_list 24469)) 11294 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 24470) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 24471) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 24472) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 11295 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 24475) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 11296 (("," 1498) (possible_comma 24473)) 11297 ((")" 24474)) 11298 (("$" 24452) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 24454) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 24478) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 24479) ("{" 1676)) 11299 (("(" 24468) (argument_list 24480)) 11300 (("$" 24452) (T_STRING 24487) (T_VARIABLE 24454) (property_name 24494) (simple_variable 24489) ("{" 24490)) 11301 (("$" 24452) (T_STRING 24487) (T_VARIABLE 24454) (property_name 24488) (simple_variable 24489) ("{" 24490)) 11302 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 24485) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 11303 (("]" 24486)) 11304 (("(" 24468) (argument_list 24493)) 11305 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 24491) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 11306 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 24492)) 11307 (("(" 24468) (argument_list 24495)) 11308 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 24499) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 11309 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 24500)) 11310 ((backup_doc_comment 24503)) 11311 (("(" 24504)) 11312 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 24505)) 11313 ((")" 24506)) 11314 ((T_USE 166) (lexical_vars 24507)) 11315 ((":" 168) (return_type 24508)) 11316 ((backup_fn_flags 24509)) 11317 (("{" 24510)) 11318 ((inner_statement_list 24511)) 11319 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 24512) ("~" 129)) 11320 ((backup_fn_flags 24513)) 11321 ((backup_doc_comment 24515)) 11322 (("(" 24516)) 11323 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 24517)) 11324 ((")" 24518)) 11325 ((":" 2271) (return_type 24519)) 11326 ((T_DOUBLE_ARROW 24520)) 11327 ((backup_fn_flags 24521)) 11328 ((backup_lex_pos 24522)) 11329 (("!" 20246) ("\"" 20247) ("$" 20248) ("(" 20249) ("+" 20250) ("-" 20251) ("@" 20252) (T_ARRAY 20253) (T_ARRAY_CAST 20254) (T_ATTRIBUTE 130) (T_BOOL_CAST 20255) (T_CLASS_C 20256) (T_CLONE 20257) (T_CONSTANT_ENCAPSED_STRING 20258) (T_DEC 20259) (T_DIR 20260) (T_DNUMBER 20261) (T_DOUBLE_CAST 20262) (T_EMPTY 20263) (T_EVAL 20264) (T_EXIT 20265) (T_FILE 20266) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 20267) (T_INC 20268) (T_INCLUDE 20269) (T_INCLUDE_ONCE 20270) (T_INT_CAST 20271) (T_ISSET 20272) (T_LINE 20273) (T_LIST 20274) (T_LNUMBER 20275) (T_MATCH 20276) (T_METHOD_C 20277) (T_NAME_FULLY_QUALIFIED 20278) (T_NAME_QUALIFIED 20279) (T_NAME_RELATIVE 20280) (T_NEW 20281) (T_NS_C 20282) (T_OBJECT_CAST 20283) (T_PRINT 20284) (T_PROPERTY_C 20285) (T_READONLY 20286) (T_REQUIRE 20287) (T_REQUIRE_ONCE 20288) (T_START_HEREDOC 20289) (T_STATIC 20290) (T_STRING 20291) (T_STRING_CAST 20292) (T_THROW 20293) (T_TRAIT_C 20294) (T_UNSET_CAST 20295) (T_VARIABLE 20296) (T_YIELD 20297) (T_YIELD_FROM 20298) ("[" 20299) ("`" 20300) (array_object_dereferenceable 20301) (attribute 135) (attributes 20302) (callable_expr 20303) (callable_variable 20304) (class_constant 20305) (class_name 20306) (constant 20307) (dereferenceable_scalar 20308) (expr 24523) (fn 20310) (fully_dereferenceable 107) #'20311 (function_call 20312) (inline_function 20313) (internal_functions_in_yacc 20314) (match 20315) (name 20316) (new_dereferenceable 20317) (new_non_dereferenceable 20318) (scalar 20319) (simple_variable 20320) (static_member 20321) (variable 20322) (variable_class_name 20323) ("~" 20324)) 11330 (("%" 20326) ("*" 20327) ("+" 20328) ("-" 20329) ("." 20330) ("/" 20331) ("<" 20332) (">" 20333) ("?" 20334) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 20335) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 20336) (T_BOOLEAN_AND 20337) (T_BOOLEAN_OR 20338) (T_COALESCE 20339) (T_INSTANCEOF 20340) (T_IS_EQUAL 20341) (T_IS_GREATER_OR_EQUAL 20342) (T_IS_IDENTICAL 20343) (T_IS_NOT_EQUAL 20344) (T_IS_NOT_IDENTICAL 20345) (T_IS_SMALLER_OR_EQUAL 20346) (T_LOGICAL_AND 20347) (T_LOGICAL_OR 20348) (T_LOGICAL_XOR 20349) (T_PIPE 20350) (T_POW 20351) (T_SL 20352) (T_SPACESHIP 20353) (T_SR 20354) ("^" 20355) (backup_fn_flags 24524) ("|" 20356)) 11331 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 24526) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 11332 (("$" 24147) (T_ABSTRACT 24326) (T_ARRAY 24327) (T_AS 24328) (T_BREAK 24329) (T_CALLABLE 24330) (T_CASE 24331) (T_CATCH 24332) (T_CLASS 24333) (T_CLASS_C 24334) (T_CLONE 24335) (T_CONST 24336) (T_CONTINUE 24337) (T_DECLARE 24338) (T_DEFAULT 24339) (T_DIR 24340) (T_DO 24341) (T_ECHO 24342) (T_ELSE 24343) (T_ELSEIF 24344) (T_EMPTY 24345) (T_ENDDECLARE 24346) (T_ENDFOR 24347) (T_ENDFOREACH 24348) (T_ENDIF 24349) (T_ENDSWITCH 24350) (T_ENDWHILE 24351) (T_ENUM 24352) (T_EVAL 24353) (T_EXIT 24354) (T_EXTENDS 24355) (T_FILE 24356) (T_FINAL 24357) (T_FINALLY 24358) (T_FN 24359) (T_FOR 24360) (T_FOREACH 24361) (T_FUNCTION 24362) (T_FUNC_C 24363) (T_GLOBAL 24364) (T_GOTO 24365) (T_IF 24366) (T_IMPLEMENTS 24367) (T_INCLUDE 24368) (T_INCLUDE_ONCE 24369) (T_INSTANCEOF 24370) (T_INSTEADOF 24371) (T_INTERFACE 24372) (T_ISSET 24373) (T_LINE 24374) (T_LIST 24375) (T_LOGICAL_AND 24376) (T_LOGICAL_OR 24377) (T_LOGICAL_XOR 24378) (T_MATCH 24379) (T_METHOD_C 24380) (T_NAMESPACE 24381) (T_NEW 24382) (T_NS_C 24383) (T_PRINT 24384) (T_PRIVATE 24385) (T_PROPERTY_C 24386) (T_PROTECTED 24387) (T_PUBLIC 24388) (T_READONLY 24389) (T_REQUIRE 24390) (T_REQUIRE_ONCE 24391) (T_RETURN 24392) (T_STATIC 24393) (T_STRING 24394) (T_SWITCH 24395) (T_THROW 24396) (T_TRAIT 24397) (T_TRAIT_C 24398) (T_TRY 24399) (T_UNSET 24400) (T_USE 24401) (T_VAR 24402) (T_VARIABLE 24195) (T_WHILE 24403) (T_YIELD 24404) (identifier 24528) (member_name 24529) (reserved_non_modifiers 24407) (semi_reserved 24408) (simple_variable 24530) ("{" 24531)) 11333 (("(" 24413) (argument_list 24534)) 11334 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 24532) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 11335 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 24533)) 11336 ((T_FN 35) (T_FUNCTION 131) (fn 24209) #'24210 (inline_function 24538)) 11337 (("$" 24147) (T_STRING 24544) (T_VARIABLE 24195) (property_name 24551) (simple_variable 24546) ("{" 24547)) 11338 (("$" 24147) (T_STRING 24544) (T_VARIABLE 24195) (property_name 24545) (simple_variable 24546) ("{" 24547)) 11339 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 24542) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 11340 (("]" 24543)) 11341 (("(" 24413) (argument_list 24550)) 11342 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 24548) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 11343 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 24549)) 11344 (("(" 24413) (argument_list 24552)) 11345 (("`" 24554)) 11346 (("]" 24556)) 11347 (("=" 24557)) 11348 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24558) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11349 (("%" 24225) ("*" 24226) ("+" 24227) ("-" 24228) ("." 24229) ("/" 24230) ("<" 24231) (">" 24232) ("?" 24233) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 24234) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 24235) (T_BOOLEAN_AND 24236) (T_BOOLEAN_OR 24237) (T_COALESCE 24238) (T_DOUBLE_ARROW 24561) (T_INSTANCEOF 24239) (T_IS_EQUAL 24240) (T_IS_GREATER_OR_EQUAL 24241) (T_IS_IDENTICAL 24242) (T_IS_NOT_EQUAL 24243) (T_IS_NOT_IDENTICAL 24244) (T_IS_SMALLER_OR_EQUAL 24245) (T_LOGICAL_AND 24246) (T_LOGICAL_OR 24247) (T_LOGICAL_XOR 24248) (T_PIPE 24249) (T_POW 24250) (T_SL 24251) (T_SPACESHIP 24252) (T_SR 24253) ("^" 24254) ("|" 24255)) 11350 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24562) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11351 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 24571) (T_VARIABLE 2920) (encaps_var 2934)) 11352 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 24570) (T_VARIABLE 2920) (encaps_var 2925)) 11353 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 24612) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 11354 ((T_CLASS 24604)) 11355 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 24603) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 24584) (attribute 2982) (class_modifier 2963)) 11356 ((T_PAAMAYIM_NEKUDOTAYIM 24601)) 11357 (("(" 24468) (argument_list 24600)) 11358 ((T_NULLSAFE_OBJECT_OPERATOR 24591) (T_OBJECT_OPERATOR 24592) (T_PAAMAYIM_NEKUDOTAYIM 24593) ("[" 24594)) 11359 (("$" 24452) (T_STRING 24487) (T_VARIABLE 24454) (property_name 24599) (simple_variable 24489) ("{" 24490)) 11360 (("$" 24452) (T_STRING 24487) (T_VARIABLE 24454) (property_name 24598) (simple_variable 24489) ("{" 24490)) 11361 (("$" 24452) (T_VARIABLE 24454) (simple_variable 24597)) 11362 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 24595) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 11363 (("]" 24596)) 11364 (("$" 24452) (T_VARIABLE 24454) (simple_variable 24602)) 11365 (("(" 2984) (argument_list 2985) (ctor_arguments 24605)) 11366 ((T_EXTENDS 2987) (extends_from 24606)) 11367 ((T_IMPLEMENTS 2989) (implements_list 24607)) 11368 ((backup_doc_comment 24608)) 11369 (("{" 24609)) 11370 ((class_statement_list 24610)) 11371 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 24611)) 11372 (("%" 304) (")" 24613) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 11373 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 24615) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 11374 (("%" 304) (")" 24616) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 11375 (("{" 24617)) 11376 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 24618) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 11377 (("}" 24619)) 11378 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 24621) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 11379 ((")" 24622)) 11380 (("=" 24623)) 11381 (("!" 24145) ("\"" 24146) ("$" 24147) ("(" 24148) ("+" 24149) ("-" 24150) ("@" 24151) (T_ARRAY 24152) (T_ARRAY_CAST 24153) (T_ATTRIBUTE 130) (T_BOOL_CAST 24154) (T_CLASS_C 24155) (T_CLONE 24156) (T_CONSTANT_ENCAPSED_STRING 24157) (T_DEC 24158) (T_DIR 24159) (T_DNUMBER 24160) (T_DOUBLE_CAST 24161) (T_EMPTY 24162) (T_EVAL 24163) (T_EXIT 24164) (T_FILE 24165) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 24166) (T_INC 24167) (T_INCLUDE 24168) (T_INCLUDE_ONCE 24169) (T_INT_CAST 24170) (T_ISSET 24171) (T_LINE 24172) (T_LIST 24173) (T_LNUMBER 24174) (T_MATCH 24175) (T_METHOD_C 24176) (T_NAME_FULLY_QUALIFIED 24177) (T_NAME_QUALIFIED 24178) (T_NAME_RELATIVE 24179) (T_NEW 24180) (T_NS_C 24181) (T_OBJECT_CAST 24182) (T_PRINT 24183) (T_PROPERTY_C 24184) (T_READONLY 24185) (T_REQUIRE 24186) (T_REQUIRE_ONCE 24187) (T_START_HEREDOC 24188) (T_STATIC 24189) (T_STRING 24190) (T_STRING_CAST 24191) (T_THROW 24192) (T_TRAIT_C 24193) (T_UNSET_CAST 24194) (T_VARIABLE 24195) (T_YIELD 24196) (T_YIELD_FROM 24197) ("[" 24198) ("`" 24199) (array_object_dereferenceable 24200) (attribute 135) (attributes 24201) (callable_expr 24202) (callable_variable 24203) (class_constant 24204) (class_name 24205) (constant 24206) (dereferenceable_scalar 24207) (expr 24624) (fn 24209) (fully_dereferenceable 107) #'24210 (function_call 24211) (inline_function 24212) (internal_functions_in_yacc 24213) (match 24214) (name 24215) (new_dereferenceable 24216) (new_non_dereferenceable 24217) (scalar 24218) (simple_variable 24219) (static_member 24220) (variable 24221) (variable_class_name 24222) ("~" 24223)) 11382 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 24626) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 11383 (("," 5226) (possible_comma 24627)) 11384 ((")" 24628)) 11385 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 24636) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 24637) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 24638) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 11386 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 24641) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 11387 (("," 1498) (possible_comma 24639)) 11388 ((")" 24640)) 11389 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 24643) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 11390 (("%" 304) (")" 24644) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 11391 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 24646) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 11392 (("%" 304) (")" 24647) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 11393 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 24653) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 24654) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 24655) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 24656) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 11394 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 24662) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 11395 (("%" 2442) (")" 24659) ("*" 2443) ("+" 2444) ("," 24660) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 11396 (("," 5287) (possible_comma 24657)) 11397 ((")" 24658)) 11398 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 24661) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 11399 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 24666) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 11400 ((")" 24667)) 11401 (("%" 304) (")" 24659) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 11402 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 24674) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 11403 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 24675)) 11404 (("\"" 24677) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 11405 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 24687) (T_VARIABLE 2920) (encaps_var 2934)) 11406 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 24686) (T_VARIABLE 2920) (encaps_var 2925)) 11407 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 24728) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 11408 ((T_CLASS 24720)) 11409 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 24719) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 24700) (attribute 2982) (class_modifier 2963)) 11410 ((T_PAAMAYIM_NEKUDOTAYIM 24717)) 11411 (("(" 19777) (argument_list 24716)) 11412 ((T_NULLSAFE_OBJECT_OPERATOR 24707) (T_OBJECT_OPERATOR 24708) (T_PAAMAYIM_NEKUDOTAYIM 24709) ("[" 24710)) 11413 (("$" 19761) (T_STRING 19796) (T_VARIABLE 19763) (property_name 24715) (simple_variable 19798) ("{" 19799)) 11414 (("$" 19761) (T_STRING 19796) (T_VARIABLE 19763) (property_name 24714) (simple_variable 19798) ("{" 19799)) 11415 (("$" 19761) (T_VARIABLE 19763) (simple_variable 24713)) 11416 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 24711) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 11417 (("]" 24712)) 11418 (("$" 19761) (T_VARIABLE 19763) (simple_variable 24718)) 11419 (("(" 2984) (argument_list 2985) (ctor_arguments 24721)) 11420 ((T_EXTENDS 2987) (extends_from 24722)) 11421 ((T_IMPLEMENTS 2989) (implements_list 24723)) 11422 ((backup_doc_comment 24724)) 11423 (("{" 24725)) 11424 ((class_statement_list 24726)) 11425 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 24727)) 11426 (("%" 304) (")" 24729) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 11427 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 24731) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 11428 (("%" 304) (")" 24732) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 11429 (("{" 24733)) 11430 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 24734) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 11431 (("}" 24735)) 11432 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 24737) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 11433 ((")" 24738)) 11434 (("=" 24739)) 11435 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 24740) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 11436 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 24742) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 11437 (("," 5226) (possible_comma 24743)) 11438 ((")" 24744)) 11439 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 24752) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 24753) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 24754) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 11440 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 24757) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 11441 (("," 1498) (possible_comma 24755)) 11442 ((")" 24756)) 11443 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 24759) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 11444 (("%" 304) (")" 24760) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 11445 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 24762) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 11446 (("%" 304) (")" 24763) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 11447 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 24769) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 24770) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 24771) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 24772) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 11448 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 24778) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 11449 (("%" 2442) (")" 24775) ("*" 2443) ("+" 2444) ("," 24776) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 11450 (("," 5287) (possible_comma 24773)) 11451 ((")" 24774)) 11452 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 24777) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 11453 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 24782) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 11454 ((")" 24783)) 11455 (("%" 304) (")" 24775) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 11456 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 24790) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 11457 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 24791)) 11458 (("\"" 24793) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 11459 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21126) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21127) (fn 18437) (for_exprs 24796) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21129) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 11460 ((";" 24797)) 11461 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21135) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21136) (fn 18437) (for_cond_exprs 24798) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21138) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 11462 ((";" 24799)) 11463 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_VOID_CAST 22737) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 22738) (fn 2425) (for_exprs 24800) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_for_exprs 22740) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 11464 ((")" 24801)) 11465 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 24802) (";" 18275) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 18276) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18277) (T_DEC 22) (T_DECLARE 18278) (T_DIR 24) (T_DNUMBER 25) (T_DO 18279) (T_DOUBLE_CAST 27) (T_ECHO 18280) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18281) (T_FOREACH 18282) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 18283) (T_GOTO 18284) (T_IF 18239) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18285) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 18286) (T_START_HEREDOC 69) (T_STATIC 18287) (T_STRING 18288) (T_STRING_CAST 72) (T_SWITCH 18289) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 18290) (T_UNSET 18291) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 18292) (T_WHILE 18293) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 18294) (alt_if_stmt_without_else 18295) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 18296) (fn 106) (for_statement 24803) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 18297) (if_stmt_without_else 18298) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 24804) (static_member 123) (variable 126) (variable_class_name 127) ("{" 18300) ("~" 129)) 11466 ((inner_statement_list 24805)) 11467 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22750) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 22751) (T_CLASS 22752) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22753) (T_DEC 22) (T_DECLARE 22754) (T_DIR 24) (T_DNUMBER 25) (T_DO 22755) (T_DOUBLE_CAST 27) (T_ECHO 22756) (T_EMPTY 29) (T_ENDFOR 24806) (T_ENUM 22758) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 22759) (T_FOREACH 22760) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 22761) (T_GOTO 22762) (T_HALT_COMPILER 22763) (T_IF 22764) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22765) (T_INTERFACE 22766) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22767) (T_START_HEREDOC 69) (T_STATIC 22768) (T_STRING 22769) (T_STRING_CAST 72) (T_SWITCH 22770) (T_THROW 74) (T_TRAIT 22771) (T_TRAIT_C 76) (T_TRY 22772) (T_UNSET 22773) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22774) (T_WHILE 22775) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22776) (alt_if_stmt_without_else 22777) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 22778) (attributes 22779) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 22780) (class_modifier 99) (class_modifiers 22781) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 22782) (expr 22783) (fn 106) (fully_dereferenceable 107) #'22784 (function_call 109) (function_declaration_statement 22785) (if_stmt 22786) (if_stmt_without_else 22787) (inline_function 113) (inner_statement 22788) (interface_declaration_statement 22789) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 22790) (static_member 123) (trait_declaration_statement 22791) (variable 126) (variable_class_name 127) ("{" 22792) ("~" 129)) 11468 ((";" 24807)) 11469 (("," 22929) (";" 24809)) 11470 ((T_WHILE 24811)) 11471 (("(" 24812)) 11472 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 24813) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 11473 (("%" 304) (")" 24814) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 11474 ((";" 24815)) 11475 ((T_STRING 23106) (const_decl 23107) (const_list 24817)) 11476 ((")" 24818) ("," 23110)) 11477 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 24819) (";" 18275) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 18276) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18277) (T_DEC 22) (T_DECLARE 18278) (T_DIR 24) (T_DNUMBER 25) (T_DO 18279) (T_DOUBLE_CAST 27) (T_ECHO 18280) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18281) (T_FOREACH 18282) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 18283) (T_GOTO 18284) (T_IF 18239) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18285) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 18286) (T_START_HEREDOC 69) (T_STATIC 18287) (T_STRING 18288) (T_STRING_CAST 72) (T_SWITCH 18289) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 18290) (T_UNSET 18291) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 18292) (T_WHILE 18293) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 18294) (alt_if_stmt_without_else 18295) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (declare_statement 24820) (dereferenceable_scalar 103) (expr 18296) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 18297) (if_stmt_without_else 18298) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 24821) (static_member 123) (variable 126) (variable_class_name 127) ("{" 18300) ("~" 129)) 11478 ((inner_statement_list 24822)) 11479 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 23116) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 23117) (T_CLASS 23118) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 23119) (T_DEC 22) (T_DECLARE 23120) (T_DIR 24) (T_DNUMBER 25) (T_DO 23121) (T_DOUBLE_CAST 27) (T_ECHO 23122) (T_EMPTY 29) (T_ENDDECLARE 24823) (T_ENUM 23124) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 23125) (T_FOREACH 23126) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 23127) (T_GOTO 23128) (T_HALT_COMPILER 23129) (T_IF 23130) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 23131) (T_INTERFACE 23132) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 23133) (T_START_HEREDOC 69) (T_STATIC 23134) (T_STRING 23135) (T_STRING_CAST 72) (T_SWITCH 23136) (T_THROW 74) (T_TRAIT 23137) (T_TRAIT_C 76) (T_TRY 23138) (T_UNSET 23139) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 23140) (T_WHILE 23141) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 23142) (alt_if_stmt_without_else 23143) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 23144) (attributes 23145) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 23146) (class_modifier 99) (class_modifiers 23147) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 23148) (expr 23149) (fn 106) (fully_dereferenceable 107) #'23150 (function_call 109) (function_declaration_statement 23151) (if_stmt 23152) (if_stmt_without_else 23153) (inline_function 113) (inner_statement 23154) (interface_declaration_statement 23155) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23156) (static_member 123) (trait_declaration_statement 23157) (variable 126) (variable_class_name 127) ("{" 23158) ("~" 129)) 11480 ((";" 24824)) 11481 ((";" 24826)) 11482 ((";" 24828)) 11483 ((T_READONLY 17774) (T_STRING 17775) (backup_doc_comment 143) (function_name 24831)) 11484 ((backup_doc_comment 24832)) 11485 (("(" 24833)) 11486 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 24834)) 11487 ((")" 24835)) 11488 ((":" 168) (return_type 24836)) 11489 ((backup_fn_flags 24837)) 11490 (("{" 24838)) 11491 ((inner_statement_list 24839)) 11492 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 24840) ("~" 129)) 11493 ((backup_fn_flags 24841)) 11494 ((T_STRING 24844)) 11495 ((T_EXTENDS 2987) (extends_from 24845)) 11496 ((T_IMPLEMENTS 2989) (implements_list 24846)) 11497 ((backup_doc_comment 24847)) 11498 (("{" 24848)) 11499 ((class_statement_list 24849)) 11500 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 24850)) 11501 ((":" 24855)) 11502 ((";" 24854)) 11503 ((inner_statement_list 24856)) 11504 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17807) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17808) (T_CLASS 17809) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17810) (T_DEC 22) (T_DECLARE 17811) (T_DIR 24) (T_DNUMBER 25) (T_DO 17812) (T_DOUBLE_CAST 27) (T_ECHO 17813) (T_EMPTY 29) (T_ENDIF 24857) (T_ENUM 17815) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17816) (T_FOREACH 17817) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17818) (T_GOTO 17819) (T_HALT_COMPILER 17820) (T_IF 17732) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17821) (T_INTERFACE 17822) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17823) (T_START_HEREDOC 69) (T_STATIC 17824) (T_STRING 17825) (T_STRING_CAST 72) (T_SWITCH 17826) (T_THROW 74) (T_TRAIT 17827) (T_TRAIT_C 76) (T_TRY 17828) (T_UNSET 17829) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17830) (T_WHILE 17831) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17832) (alt_if_stmt_without_else 17833) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17834) (attributes 17835) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17836) (class_modifier 99) (class_modifiers 17837) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17838) (expr 17839) (fn 106) (fully_dereferenceable 107) #'17840 (function_call 109) (function_declaration_statement 17841) (if_stmt 17842) (if_stmt_without_else 17843) (inline_function 113) (inner_statement 17844) (interface_declaration_statement 17845) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17846) (static_member 123) (trait_declaration_statement 17847) (variable 126) (variable_class_name 127) ("{" 17848) ("~" 129)) 11505 ((";" 24858)) 11506 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 24860) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 11507 (("%" 304) (")" 24861) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 11508 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 24862) (";" 18226) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 18227) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18229) (T_DEC 22) (T_DECLARE 18230) (T_DIR 24) (T_DNUMBER 25) (T_DO 18231) (T_DOUBLE_CAST 27) (T_ECHO 18232) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18234) (T_FOREACH 18235) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 18236) (T_GOTO 18237) (T_IF 18239) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18240) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 18242) (T_START_HEREDOC 69) (T_STATIC 18243) (T_STRING 18244) (T_STRING_CAST 72) (T_SWITCH 18245) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 18247) (T_UNSET 18248) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 18249) (T_WHILE 18250) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 18251) (alt_if_stmt_without_else 18252) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 18258) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 18261) (if_stmt_without_else 18262) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 24863) (static_member 123) (variable 126) (variable_class_name 127) (while_statement 24864) ("{" 18267) ("~" 129)) 11509 ((inner_statement_list 24865)) 11510 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17889) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17890) (T_CLASS 17891) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17892) (T_DEC 22) (T_DECLARE 17893) (T_DIR 24) (T_DNUMBER 25) (T_DO 17894) (T_DOUBLE_CAST 27) (T_ECHO 17895) (T_EMPTY 29) (T_ENDWHILE 24866) (T_ENUM 17897) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17898) (T_FOREACH 17899) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17900) (T_GOTO 17901) (T_HALT_COMPILER 17902) (T_IF 17903) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17904) (T_INTERFACE 17905) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17906) (T_START_HEREDOC 69) (T_STATIC 17907) (T_STRING 17908) (T_STRING_CAST 72) (T_SWITCH 17909) (T_THROW 74) (T_TRAIT 17910) (T_TRAIT_C 76) (T_TRY 17911) (T_UNSET 17912) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17913) (T_WHILE 17914) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17915) (alt_if_stmt_without_else 17916) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17917) (attributes 17918) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17919) (class_modifier 99) (class_modifiers 17920) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17921) (expr 17922) (fn 106) (fully_dereferenceable 107) #'17923 (function_call 109) (function_declaration_statement 17924) (if_stmt 17925) (if_stmt_without_else 17926) (inline_function 113) (inner_statement 17927) (interface_declaration_statement 17928) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17929) (static_member 123) (trait_declaration_statement 17930) (variable 126) (variable_class_name 127) ("{" 17931) ("~" 129)) 11511 ((";" 24867)) 11512 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 24869) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 11513 (("\"" 1538) ("$" 17989) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 17990) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 17991) ("[" 1560) (array_object_dereferenceable 17992) (callable_expr 17993) (callable_variable 17994) (class_constant 1564) (class_name 17995) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 17996) (name 17997) (new_dereferenceable 1570) (simple_variable 17998) (static_member 17999) (unset_variable 18000) (unset_variables 24871) (variable 18002) (variable_class_name 18003)) 11514 (("," 18015) (possible_comma 24872)) 11515 ((")" 24873)) 11516 ((";" 24874)) 11517 ((inner_statement_list 24876)) 11518 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 24877) ("~" 129)) 11519 ((catch_list 24878)) 11520 ((T_CATCH 24879) (T_FINALLY 24880) (finally_statement 24881)) 11521 (("(" 24885)) 11522 (("{" 24882)) 11523 ((inner_statement_list 24883)) 11524 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 24884) ("~" 129)) 11525 ((T_NAME_FULLY_QUALIFIED 18056) (T_NAME_QUALIFIED 18057) (T_NAME_RELATIVE 18058) (T_STATIC 18059) (T_STRING 18060) (catch_name_list 24886) (class_name 18062) (name 18063)) 11526 ((T_VARIABLE 18064) (optional_variable 24887) ("|" 18066)) 11527 ((")" 24888)) 11528 (("{" 24889)) 11529 ((inner_statement_list 24890)) 11530 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 24891) ("~" 129)) 11531 ((backup_doc_comment 24893)) 11532 (("{" 24894)) 11533 ((class_statement_list 24895)) 11534 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 24896)) 11535 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 24898) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 11536 (("%" 304) (")" 24899) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 11537 ((":" 24900) (switch_case_list 24901) ("{" 24902)) 11538 ((";" 24908) (case_list 24909)) 11539 ((";" 24903) (case_list 24904)) 11540 ((case_list 24906)) 11541 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 24905)) 11542 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 24907)) 11543 ((case_list 24912)) 11544 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 24910)) 11545 ((";" 24911)) 11546 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 24913)) 11547 ((";" 24914)) 11548 (("," 18369) (";" 24917)) 11549 ((";" 24919)) 11550 ((T_EXTENDS 23405) (interface_extends_list 24921)) 11551 ((backup_doc_comment 24922)) 11552 (("{" 24923)) 11553 ((class_statement_list 24924)) 11554 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 24925)) 11555 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 24927) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 11556 (("%" 304) (")" 24928) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 11557 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23415) (";" 18275) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 18276) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18277) (T_DEC 22) (T_DECLARE 18278) (T_DIR 24) (T_DNUMBER 25) (T_DO 18279) (T_DOUBLE_CAST 27) (T_ECHO 18280) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18281) (T_FOREACH 18282) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 18283) (T_GOTO 18284) (T_IF 18239) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18285) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 18286) (T_START_HEREDOC 69) (T_STATIC 18287) (T_STRING 18288) (T_STRING_CAST 72) (T_SWITCH 18289) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 18290) (T_UNSET 18291) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 18292) (T_WHILE 18293) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 18294) (alt_if_stmt_without_else 18295) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 18296) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 18297) (if_stmt_without_else 18298) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 24929) (static_member 123) (variable 126) (variable_class_name 127) ("{" 18300) ("~" 129)) 11558 ((")" 24931)) 11559 ((";" 24932)) 11560 ((";" 24934)) 11561 (("," 19446) (";" 24936)) 11562 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 24938) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 11563 (("%" 19534) ("*" 19535) ("+" 19536) ("-" 19537) ("." 19538) ("/" 19539) ("<" 19540) (">" 19541) ("?" 19542) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19543) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19544) (T_AS 24939) (T_BOOLEAN_AND 19545) (T_BOOLEAN_OR 19546) (T_COALESCE 19547) (T_INSTANCEOF 19548) (T_IS_EQUAL 19549) (T_IS_GREATER_OR_EQUAL 19550) (T_IS_IDENTICAL 19551) (T_IS_NOT_EQUAL 19552) (T_IS_NOT_IDENTICAL 19553) (T_IS_SMALLER_OR_EQUAL 19554) (T_LOGICAL_AND 19555) (T_LOGICAL_OR 19556) (T_LOGICAL_XOR 19557) (T_PIPE 19558) (T_POW 19559) (T_SL 19560) (T_SPACESHIP 19561) (T_SR 19562) ("^" 19563) ("|" 19564)) 11564 (("\"" 1538) ("$" 20897) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20898) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20899) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20900) ("[" 20901) (ampersand 20902) (array_object_dereferenceable 20903) (callable_expr 20904) (callable_variable 20905) (class_constant 1564) (class_name 20906) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 24940) (fully_dereferenceable 107) (function_call 20908) (name 20909) (new_dereferenceable 1570) (simple_variable 20910) (static_member 20911) (variable 20912) (variable_class_name 20913)) 11565 ((")" 24941) (T_DOUBLE_ARROW 24942)) 11566 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 24945) (";" 18226) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 18227) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18229) (T_DEC 22) (T_DECLARE 18230) (T_DIR 24) (T_DNUMBER 25) (T_DO 18231) (T_DOUBLE_CAST 27) (T_ECHO 18232) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18234) (T_FOREACH 18235) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 18236) (T_GOTO 18237) (T_IF 18239) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18240) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 18242) (T_START_HEREDOC 69) (T_STATIC 18243) (T_STRING 18244) (T_STRING_CAST 72) (T_SWITCH 18245) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 18247) (T_UNSET 18248) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 18249) (T_WHILE 18250) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 18251) (alt_if_stmt_without_else 18252) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 18258) (fn 106) (foreach_statement 24951) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 18261) (if_stmt_without_else 18262) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 24947) (static_member 123) (variable 126) (variable_class_name 127) ("{" 18267) ("~" 129)) 11567 (("\"" 1538) ("$" 20928) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20929) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20930) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20931) ("[" 20932) (ampersand 20933) (array_object_dereferenceable 20934) (callable_expr 20935) (callable_variable 20936) (class_constant 1564) (class_name 20937) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 24943) (fully_dereferenceable 107) (function_call 20939) (name 20940) (new_dereferenceable 1570) (simple_variable 20941) (static_member 20942) (variable 20943) (variable_class_name 20944)) 11568 ((")" 24944)) 11569 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 24945) (";" 18226) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 18227) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18229) (T_DEC 22) (T_DECLARE 18230) (T_DIR 24) (T_DNUMBER 25) (T_DO 18231) (T_DOUBLE_CAST 27) (T_ECHO 18232) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18234) (T_FOREACH 18235) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 18236) (T_GOTO 18237) (T_IF 18239) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18240) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 18242) (T_START_HEREDOC 69) (T_STATIC 18243) (T_STRING 18244) (T_STRING_CAST 72) (T_SWITCH 18245) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 18247) (T_UNSET 18248) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 18249) (T_WHILE 18250) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 18251) (alt_if_stmt_without_else 18252) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 18258) (fn 106) (foreach_statement 24946) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 18261) (if_stmt_without_else 18262) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 24947) (static_member 123) (variable 126) (variable_class_name 127) ("{" 18267) ("~" 129)) 11570 ((inner_statement_list 24948)) 11571 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20962) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 20963) (T_CLASS 20964) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20965) (T_DEC 22) (T_DECLARE 20966) (T_DIR 24) (T_DNUMBER 25) (T_DO 20967) (T_DOUBLE_CAST 27) (T_ECHO 20968) (T_EMPTY 29) (T_ENDFOREACH 24949) (T_ENUM 20970) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20971) (T_FOREACH 20972) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20973) (T_GOTO 20974) (T_HALT_COMPILER 20975) (T_IF 20976) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20977) (T_INTERFACE 20978) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20979) (T_START_HEREDOC 69) (T_STATIC 20980) (T_STRING 20981) (T_STRING_CAST 72) (T_SWITCH 20982) (T_THROW 74) (T_TRAIT 20983) (T_TRAIT_C 76) (T_TRY 20984) (T_UNSET 20985) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 20986) (T_WHILE 20987) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 20988) (alt_if_stmt_without_else 20989) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 20990) (attributes 20991) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 20992) (class_modifier 99) (class_modifiers 20993) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 20994) (expr 20995) (fn 106) (fully_dereferenceable 107) #'20996 (function_call 109) (function_declaration_statement 20997) (if_stmt 20998) (if_stmt_without_else 20999) (inline_function 113) (inner_statement 21000) (interface_declaration_statement 21001) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 21002) (static_member 123) (trait_declaration_statement 21003) (variable 126) (variable_class_name 127) ("{" 21004) ("~" 129)) 11572 ((";" 24950)) 11573 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21126) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21127) (fn 18437) (for_exprs 24953) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21129) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 11574 ((";" 24954)) 11575 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21135) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21136) (fn 18437) (for_cond_exprs 24955) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21138) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 11576 ((";" 24956)) 11577 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_VOID_CAST 22737) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 22738) (fn 2425) (for_exprs 24957) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_for_exprs 22740) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 11578 ((")" 24958)) 11579 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 24959) (";" 18226) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 18227) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18229) (T_DEC 22) (T_DECLARE 18230) (T_DIR 24) (T_DNUMBER 25) (T_DO 18231) (T_DOUBLE_CAST 27) (T_ECHO 18232) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18234) (T_FOREACH 18235) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 18236) (T_GOTO 18237) (T_IF 18239) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18240) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 18242) (T_START_HEREDOC 69) (T_STATIC 18243) (T_STRING 18244) (T_STRING_CAST 72) (T_SWITCH 18245) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 18247) (T_UNSET 18248) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 18249) (T_WHILE 18250) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 18251) (alt_if_stmt_without_else 18252) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 18258) (fn 106) (for_statement 24960) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 18261) (if_stmt_without_else 18262) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 24961) (static_member 123) (variable 126) (variable_class_name 127) ("{" 18267) ("~" 129)) 11580 ((inner_statement_list 24962)) 11581 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22750) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 22751) (T_CLASS 22752) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22753) (T_DEC 22) (T_DECLARE 22754) (T_DIR 24) (T_DNUMBER 25) (T_DO 22755) (T_DOUBLE_CAST 27) (T_ECHO 22756) (T_EMPTY 29) (T_ENDFOR 24963) (T_ENUM 22758) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 22759) (T_FOREACH 22760) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 22761) (T_GOTO 22762) (T_HALT_COMPILER 22763) (T_IF 22764) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22765) (T_INTERFACE 22766) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22767) (T_START_HEREDOC 69) (T_STATIC 22768) (T_STRING 22769) (T_STRING_CAST 72) (T_SWITCH 22770) (T_THROW 74) (T_TRAIT 22771) (T_TRAIT_C 76) (T_TRY 22772) (T_UNSET 22773) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22774) (T_WHILE 22775) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22776) (alt_if_stmt_without_else 22777) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 22778) (attributes 22779) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 22780) (class_modifier 99) (class_modifiers 22781) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 22782) (expr 22783) (fn 106) (fully_dereferenceable 107) #'22784 (function_call 109) (function_declaration_statement 22785) (if_stmt 22786) (if_stmt_without_else 22787) (inline_function 113) (inner_statement 22788) (interface_declaration_statement 22789) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 22790) (static_member 123) (trait_declaration_statement 22791) (variable 126) (variable_class_name 127) ("{" 22792) ("~" 129)) 11582 ((";" 24964)) 11583 ((":" 23454) (enum_backing_type 24966)) 11584 ((T_IMPLEMENTS 2989) (implements_list 24967)) 11585 ((backup_doc_comment 24968)) 11586 (("{" 24969)) 11587 ((class_statement_list 24970)) 11588 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 24971)) 11589 (("," 22929) (";" 24973)) 11590 ((T_WHILE 24975)) 11591 (("(" 24976)) 11592 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 24977) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 11593 (("%" 304) (")" 24978) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 11594 ((";" 24979)) 11595 ((T_STRING 23106) (const_decl 23107) (const_list 24981)) 11596 ((")" 24982) ("," 23110)) 11597 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 24983) (";" 18226) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 18227) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18229) (T_DEC 22) (T_DECLARE 18230) (T_DIR 24) (T_DNUMBER 25) (T_DO 18231) (T_DOUBLE_CAST 27) (T_ECHO 18232) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18234) (T_FOREACH 18235) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 18236) (T_GOTO 18237) (T_IF 18239) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18240) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 18242) (T_START_HEREDOC 69) (T_STATIC 18243) (T_STRING 18244) (T_STRING_CAST 72) (T_SWITCH 18245) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 18247) (T_UNSET 18248) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 18249) (T_WHILE 18250) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 18251) (alt_if_stmt_without_else 18252) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (declare_statement 24984) (dereferenceable_scalar 103) (expr 18258) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 18261) (if_stmt_without_else 18262) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 24985) (static_member 123) (variable 126) (variable_class_name 127) ("{" 18267) ("~" 129)) 11598 ((inner_statement_list 24986)) 11599 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 23116) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 23117) (T_CLASS 23118) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 23119) (T_DEC 22) (T_DECLARE 23120) (T_DIR 24) (T_DNUMBER 25) (T_DO 23121) (T_DOUBLE_CAST 27) (T_ECHO 23122) (T_EMPTY 29) (T_ENDDECLARE 24987) (T_ENUM 23124) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 23125) (T_FOREACH 23126) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 23127) (T_GOTO 23128) (T_HALT_COMPILER 23129) (T_IF 23130) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 23131) (T_INTERFACE 23132) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 23133) (T_START_HEREDOC 69) (T_STATIC 23134) (T_STRING 23135) (T_STRING_CAST 72) (T_SWITCH 23136) (T_THROW 74) (T_TRAIT 23137) (T_TRAIT_C 76) (T_TRY 23138) (T_UNSET 23139) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 23140) (T_WHILE 23141) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 23142) (alt_if_stmt_without_else 23143) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 23144) (attributes 23145) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 23146) (class_modifier 99) (class_modifiers 23147) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 23148) (expr 23149) (fn 106) (fully_dereferenceable 107) #'23150 (function_call 109) (function_declaration_statement 23151) (if_stmt 23152) (if_stmt_without_else 23153) (inline_function 113) (inner_statement 23154) (interface_declaration_statement 23155) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23156) (static_member 123) (trait_declaration_statement 23157) (variable 126) (variable_class_name 127) ("{" 23158) ("~" 129)) 11600 ((";" 24988)) 11601 ((";" 24990)) 11602 ((T_EXTENDS 2987) (extends_from 24992)) 11603 ((T_IMPLEMENTS 2989) (implements_list 24993)) 11604 ((backup_doc_comment 24994)) 11605 (("{" 24995)) 11606 ((class_statement_list 24996)) 11607 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 24997)) 11608 ((";" 24999)) 11609 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 27129) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11610 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 27127) (encaps_var 5357)) 11611 (("$" 25003) (T_VARIABLE 25051) (simple_variable 27123) ("{" 27124)) 11612 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 27122) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 11613 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 27121) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11614 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 27120) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11615 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 27119) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11616 (("(" 27116)) 11617 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 27115) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11618 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 27114) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11619 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 27101) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (clone_argument_list 27102) (constant 25062) (dereferenceable_scalar 25063) (expr 27103) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11620 (("\"" 1538) ("$" 25308) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 25309) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 25310) ("[" 1560) (array_object_dereferenceable 25311) (callable_expr 25312) (callable_variable 25313) (class_constant 1564) (class_name 25314) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 25315) (name 25316) (new_dereferenceable 1570) (simple_variable 25317) (static_member 25318) (variable 27100) (variable_class_name 25320)) 11621 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 27099) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11622 (("(" 27096)) 11623 (("(" 27093)) 11624 (("(" 27084) (argument_list 27085) (ctor_arguments 27086)) 11625 (("\"" 1538) ("$" 25308) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 25309) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 25310) ("[" 1560) (array_object_dereferenceable 25311) (callable_expr 25312) (callable_variable 25313) (class_constant 1564) (class_name 25314) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 25315) (name 25316) (new_dereferenceable 1570) (simple_variable 25317) (static_member 25318) (variable 27083) (variable_class_name 25320)) 11626 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 27082) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11627 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 27081) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11628 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 27080) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11629 (("(" 27076)) 11630 (("(" 27071)) 11631 (("(" 27065)) 11632 (("$" 25308) ("(" 27028) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 27029) (T_NAME_QUALIFIED 27030) (T_NAME_RELATIVE 27031) (T_READONLY 2955) (T_STATIC 27032) (T_STRING 27033) (T_VARIABLE 25310) (anonymous_class 27034) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 27035) (attribute 2961) (attributes 27036) (class_modifier 2963) (class_name 27037) (class_name_reference 27038) (name 27039) (new_variable 27040) (simple_variable 27041)) 11633 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 27027) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11634 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 27026) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11635 (("(" 25269) (argument_list 27025)) 11636 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 27024) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11637 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 27023) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11638 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 27018) (T_END_HEREDOC 27019) (T_VARIABLE 2920) (encaps_list 27020) (encaps_var 2922)) 11639 ((T_FN 35) (T_FUNCTION 131) (fn 25065) #'25066 (inline_function 27017)) 11640 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 27016) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11641 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 27015) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11642 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 27014) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11643 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26543) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 11644 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 26479) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11645 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 26475) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 11646 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 26473) (encaps_list 2828) (encaps_var 2829)) 11647 ((T_NULLSAFE_OBJECT_OPERATOR 26459) (T_OBJECT_OPERATOR 26460) ("[" 26461)) 11648 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 26456) (attribute 2757) (fn 25065) #'25066 (inline_function 26457)) 11649 (("(" 25269) (argument_list 26455)) 11650 ((T_PAAMAYIM_NEKUDOTAYIM 26447)) 11651 (("%" 25081) ("*" 25082) ("+" 25083) ("-" 25084) ("." 25085) ("/" 25086) (":" 26443) (";" 26444) ("<" 25087) (">" 25088) ("?" 25089) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 25090) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 25091) (T_BOOLEAN_AND 25092) (T_BOOLEAN_OR 25093) (T_COALESCE 25094) (T_INSTANCEOF 25095) (T_IS_EQUAL 25096) (T_IS_GREATER_OR_EQUAL 25097) (T_IS_IDENTICAL 25098) (T_IS_NOT_EQUAL 25099) (T_IS_NOT_IDENTICAL 25100) (T_IS_SMALLER_OR_EQUAL 25101) (T_LOGICAL_AND 25102) (T_LOGICAL_OR 25103) (T_LOGICAL_XOR 25104) (T_PIPE 25105) (T_POW 25106) (T_SL 25107) (T_SPACESHIP 25108) (T_SR 25109) ("^" 25110) ("|" 25111)) 11652 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 25370)) 11653 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 25358)) 11654 (("(" 25269) (argument_list 25357)) 11655 (("=" 25277) (T_AND_EQUAL 25278) (T_COALESCE_EQUAL 25279) (T_CONCAT_EQUAL 25280) (T_DEC 25281) (T_DIV_EQUAL 25282) (T_INC 25283) (T_MINUS_EQUAL 25284) (T_MOD_EQUAL 25285) (T_MUL_EQUAL 25286) (T_OR_EQUAL 25287) (T_PLUS_EQUAL 25288) (T_POW_EQUAL 25289) (T_SL_EQUAL 25290) (T_SR_EQUAL 25291) (T_XOR_EQUAL 25292)) 11656 ((T_PAAMAYIM_NEKUDOTAYIM 25181)) 11657 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25080) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11658 (("%" 25081) ("*" 25082) ("+" 25083) ("-" 25084) ("." 25085) ("/" 25086) ("<" 25087) (">" 25088) ("?" 25089) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 25090) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 25091) (T_BOOLEAN_AND 25092) (T_BOOLEAN_OR 25093) (T_COALESCE 25094) (T_INSTANCEOF 25095) (T_IS_EQUAL 25096) (T_IS_GREATER_OR_EQUAL 25097) (T_IS_IDENTICAL 25098) (T_IS_NOT_EQUAL 25099) (T_IS_NOT_IDENTICAL 25100) (T_IS_SMALLER_OR_EQUAL 25101) (T_LOGICAL_AND 25102) (T_LOGICAL_OR 25103) (T_LOGICAL_XOR 25104) (T_PIPE 25105) (T_POW 25106) (T_SL 25107) (T_SPACESHIP 25108) (T_SR 25109) ("^" 25110) ("|" 25111)) 11659 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25180) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11660 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25179) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11661 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25178) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11662 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25177) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11663 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25176) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11664 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25175) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11665 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25174) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11666 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25173) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11667 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 25168) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 25169) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 11668 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25167) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11669 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25166) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11670 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25165) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11671 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25164) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11672 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25163) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11673 (("$" 25128) ("(" 25129) (T_NAME_FULLY_QUALIFIED 25130) (T_NAME_QUALIFIED 25131) (T_NAME_RELATIVE 25132) (T_STATIC 25133) (T_STRING 25134) (T_VARIABLE 25135) (class_name 25136) (class_name_reference 25137) (name 25138) (new_variable 25139) (simple_variable 25140)) 11674 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25127) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11675 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25126) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11676 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25125) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11677 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25124) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11678 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25123) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11679 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25122) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11680 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25121) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11681 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25120) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11682 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25119) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11683 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25118) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11684 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25117) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11685 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25116) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11686 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25115) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11687 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25114) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11688 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25113) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11689 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25112) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11690 (("$" 25128) (T_VARIABLE 25135) (simple_variable 25159) ("{" 25160)) 11691 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 25157) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 11692 ((T_PAAMAYIM_NEKUDOTAYIM 25155)) 11693 ((T_NULLSAFE_OBJECT_OPERATOR 25141) (T_OBJECT_OPERATOR 25142) (T_PAAMAYIM_NEKUDOTAYIM 25143) ("[" 25144)) 11694 (("$" 25128) (T_STRING 25148) (T_VARIABLE 25135) (property_name 25154) (simple_variable 25150) ("{" 25151)) 11695 (("$" 25128) (T_STRING 25148) (T_VARIABLE 25135) (property_name 25149) (simple_variable 25150) ("{" 25151)) 11696 (("$" 25128) (T_VARIABLE 25135) (simple_variable 25147)) 11697 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 25145) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 11698 (("]" 25146)) 11699 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 25152) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 11700 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 25153)) 11701 (("$" 25128) (T_VARIABLE 25135) (simple_variable 25156)) 11702 (("%" 304) (")" 25158) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 11703 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 25161) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 11704 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 25162)) 11705 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25172) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11706 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 25170) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 11707 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25171) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11708 (("$" 25003) (T_ABSTRACT 25182) (T_ARRAY 25183) (T_AS 25184) (T_BREAK 25185) (T_CALLABLE 25186) (T_CASE 25187) (T_CATCH 25188) (T_CLASS 25189) (T_CLASS_C 25190) (T_CLONE 25191) (T_CONST 25192) (T_CONTINUE 25193) (T_DECLARE 25194) (T_DEFAULT 25195) (T_DIR 25196) (T_DO 25197) (T_ECHO 25198) (T_ELSE 25199) (T_ELSEIF 25200) (T_EMPTY 25201) (T_ENDDECLARE 25202) (T_ENDFOR 25203) (T_ENDFOREACH 25204) (T_ENDIF 25205) (T_ENDSWITCH 25206) (T_ENDWHILE 25207) (T_ENUM 25208) (T_EVAL 25209) (T_EXIT 25210) (T_EXTENDS 25211) (T_FILE 25212) (T_FINAL 25213) (T_FINALLY 25214) (T_FN 25215) (T_FOR 25216) (T_FOREACH 25217) (T_FUNCTION 25218) (T_FUNC_C 25219) (T_GLOBAL 25220) (T_GOTO 25221) (T_IF 25222) (T_IMPLEMENTS 25223) (T_INCLUDE 25224) (T_INCLUDE_ONCE 25225) (T_INSTANCEOF 25226) (T_INSTEADOF 25227) (T_INTERFACE 25228) (T_ISSET 25229) (T_LINE 25230) (T_LIST 25231) (T_LOGICAL_AND 25232) (T_LOGICAL_OR 25233) (T_LOGICAL_XOR 25234) (T_MATCH 25235) (T_METHOD_C 25236) (T_NAMESPACE 25237) (T_NEW 25238) (T_NS_C 25239) (T_PRINT 25240) (T_PRIVATE 25241) (T_PROPERTY_C 25242) (T_PROTECTED 25243) (T_PUBLIC 25244) (T_READONLY 25245) (T_REQUIRE 25246) (T_REQUIRE_ONCE 25247) (T_RETURN 25248) (T_STATIC 25249) (T_STRING 25250) (T_SWITCH 25251) (T_THROW 25252) (T_TRAIT 25253) (T_TRAIT_C 25254) (T_TRY 25255) (T_UNSET 25256) (T_USE 25257) (T_VAR 25258) (T_VARIABLE 25051) (T_WHILE 25259) (T_YIELD 25260) (identifier 25261) (member_name 25262) (reserved_non_modifiers 25263) (semi_reserved 25264) (simple_variable 25265) ("{" 25266)) 11709 (("(" 25269) (argument_list 25270)) 11710 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 25267) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 11711 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 25268)) 11712 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 25271) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 25272) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 25273) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 11713 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 25276) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 11714 (("," 1498) (possible_comma 25274)) 11715 ((")" 25275)) 11716 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (ampersand 25306) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25307) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11717 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25305) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11718 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25304) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11719 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25303) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11720 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25302) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11721 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25301) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11722 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25300) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11723 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25299) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11724 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25298) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11725 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25297) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11726 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25296) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11727 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25295) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11728 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25294) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11729 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 25293) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 11730 (("\"" 1538) ("$" 25308) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 25309) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 25310) ("[" 1560) (array_object_dereferenceable 25311) (callable_expr 25312) (callable_variable 25313) (class_constant 1564) (class_name 25314) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 25315) (name 25316) (new_dereferenceable 1570) (simple_variable 25317) (static_member 25318) (variable 25319) (variable_class_name 25320)) 11731 (("$" 25308) (T_VARIABLE 25310) (simple_variable 25353) ("{" 25354)) 11732 (("(" 25324) (argument_list 25352)) 11733 ((T_NULLSAFE_OBJECT_OPERATOR 25338) (T_OBJECT_OPERATOR 25339) ("[" 25340)) 11734 (("(" 25324) (argument_list 25337)) 11735 ((T_PAAMAYIM_NEKUDOTAYIM 25333)) 11736 (("(" 25324) (argument_list 25332)) 11737 ((T_PAAMAYIM_NEKUDOTAYIM 25321)) 11738 (("$" 25308) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 25310) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 25322) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 25323) ("{" 1660)) 11739 (("(" 25324) (argument_list 25325)) 11740 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 25326) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 25327) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 25328) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 11741 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 25331) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 11742 (("," 1498) (possible_comma 25329)) 11743 ((")" 25330)) 11744 (("$" 25308) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 25310) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 25334) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 25335) ("{" 1676)) 11745 (("(" 25324) (argument_list 25336)) 11746 (("$" 25308) (T_STRING 25343) (T_VARIABLE 25310) (property_name 25350) (simple_variable 25345) ("{" 25346)) 11747 (("$" 25308) (T_STRING 25343) (T_VARIABLE 25310) (property_name 25344) (simple_variable 25345) ("{" 25346)) 11748 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 25341) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 11749 (("]" 25342)) 11750 (("(" 25324) (argument_list 25349)) 11751 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 25347) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 11752 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 25348)) 11753 (("(" 25324) (argument_list 25351)) 11754 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 25355) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 11755 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 25356)) 11756 ((backup_doc_comment 25359)) 11757 (("(" 25360)) 11758 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 25361)) 11759 ((")" 25362)) 11760 ((T_USE 166) (lexical_vars 25363)) 11761 ((":" 168) (return_type 25364)) 11762 ((backup_fn_flags 25365)) 11763 (("{" 25366)) 11764 ((inner_statement_list 25367)) 11765 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 25368) ("~" 129)) 11766 ((backup_fn_flags 25369)) 11767 ((backup_doc_comment 25371)) 11768 (("(" 25372)) 11769 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 25373)) 11770 ((")" 25374)) 11771 ((":" 2271) (return_type 25375)) 11772 ((T_DOUBLE_ARROW 25376)) 11773 ((backup_fn_flags 25377)) 11774 ((backup_lex_pos 25378)) 11775 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25442) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11776 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 26442) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11777 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 26440) (encaps_var 5357)) 11778 (("$" 25381) (T_VARIABLE 25429) (simple_variable 26436) ("{" 26437)) 11779 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 26435) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 11780 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 26434) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11781 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 26433) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11782 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 26432) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11783 (("(" 26429)) 11784 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 26428) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11785 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 26427) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11786 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 26414) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (clone_argument_list 26415) (constant 25440) (dereferenceable_scalar 25441) (expr 26416) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11787 (("\"" 1538) ("$" 25686) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 25687) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 25688) ("[" 1560) (array_object_dereferenceable 25689) (callable_expr 25690) (callable_variable 25691) (class_constant 1564) (class_name 25692) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 25693) (name 25694) (new_dereferenceable 1570) (simple_variable 25695) (static_member 25696) (variable 26413) (variable_class_name 25698)) 11788 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 26412) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11789 (("(" 26409)) 11790 (("(" 26406)) 11791 (("(" 26397) (argument_list 26398) (ctor_arguments 26399)) 11792 (("\"" 1538) ("$" 25686) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 25687) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 25688) ("[" 1560) (array_object_dereferenceable 25689) (callable_expr 25690) (callable_variable 25691) (class_constant 1564) (class_name 25692) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 25693) (name 25694) (new_dereferenceable 1570) (simple_variable 25695) (static_member 25696) (variable 26396) (variable_class_name 25698)) 11793 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 26395) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11794 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 26394) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11795 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 26393) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11796 (("(" 26389)) 11797 (("(" 26384)) 11798 (("(" 26378)) 11799 (("$" 25686) ("(" 26341) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 26342) (T_NAME_QUALIFIED 26343) (T_NAME_RELATIVE 26344) (T_READONLY 2955) (T_STATIC 26345) (T_STRING 26346) (T_VARIABLE 25688) (anonymous_class 26347) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 26348) (attribute 2961) (attributes 26349) (class_modifier 2963) (class_name 26350) (class_name_reference 26351) (name 26352) (new_variable 26353) (simple_variable 26354)) 11800 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 26340) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11801 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 26339) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11802 (("(" 25647) (argument_list 26338)) 11803 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 26337) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11804 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 26336) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11805 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 26331) (T_END_HEREDOC 26332) (T_VARIABLE 2920) (encaps_list 26333) (encaps_var 2922)) 11806 ((T_FN 35) (T_FUNCTION 131) (fn 25443) #'25444 (inline_function 26330)) 11807 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 26329) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11808 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 26328) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11809 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 26327) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11810 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 25856) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 11811 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25792) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11812 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 25788) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 11813 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 25786) (encaps_list 2828) (encaps_var 2829)) 11814 ((T_NULLSAFE_OBJECT_OPERATOR 25772) (T_OBJECT_OPERATOR 25773) ("[" 25774)) 11815 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 25769) (attribute 2757) (fn 25443) #'25444 (inline_function 25770)) 11816 (("(" 25647) (argument_list 25768)) 11817 ((T_PAAMAYIM_NEKUDOTAYIM 25760)) 11818 (("%" 25459) ("*" 25460) ("+" 25461) ("-" 25462) ("." 25463) ("/" 25464) ("<" 25465) (">" 25466) ("?" 25467) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 25468) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 25469) (T_BOOLEAN_AND 25470) (T_BOOLEAN_OR 25471) (T_COALESCE 25472) (T_INSTANCEOF 25473) (T_IS_EQUAL 25474) (T_IS_GREATER_OR_EQUAL 25475) (T_IS_IDENTICAL 25476) (T_IS_NOT_EQUAL 25477) (T_IS_NOT_IDENTICAL 25478) (T_IS_SMALLER_OR_EQUAL 25479) (T_LOGICAL_AND 25480) (T_LOGICAL_OR 25481) (T_LOGICAL_XOR 25482) (T_PIPE 25483) (T_POW 25484) (T_SL 25485) (T_SPACESHIP 25486) (T_SR 25487) ("^" 25488) (backup_fn_flags 25759) ("|" 25489)) 11819 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 25748)) 11820 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 25736)) 11821 (("(" 25647) (argument_list 25735)) 11822 (("=" 25655) (T_AND_EQUAL 25656) (T_COALESCE_EQUAL 25657) (T_CONCAT_EQUAL 25658) (T_DEC 25659) (T_DIV_EQUAL 25660) (T_INC 25661) (T_MINUS_EQUAL 25662) (T_MOD_EQUAL 25663) (T_MUL_EQUAL 25664) (T_OR_EQUAL 25665) (T_PLUS_EQUAL 25666) (T_POW_EQUAL 25667) (T_SL_EQUAL 25668) (T_SR_EQUAL 25669) (T_XOR_EQUAL 25670)) 11823 ((T_PAAMAYIM_NEKUDOTAYIM 25559)) 11824 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25458) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11825 (("%" 25459) ("*" 25460) ("+" 25461) ("-" 25462) ("." 25463) ("/" 25464) ("<" 25465) (">" 25466) ("?" 25467) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 25468) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 25469) (T_BOOLEAN_AND 25470) (T_BOOLEAN_OR 25471) (T_COALESCE 25472) (T_INSTANCEOF 25473) (T_IS_EQUAL 25474) (T_IS_GREATER_OR_EQUAL 25475) (T_IS_IDENTICAL 25476) (T_IS_NOT_EQUAL 25477) (T_IS_NOT_IDENTICAL 25478) (T_IS_SMALLER_OR_EQUAL 25479) (T_LOGICAL_AND 25480) (T_LOGICAL_OR 25481) (T_LOGICAL_XOR 25482) (T_PIPE 25483) (T_POW 25484) (T_SL 25485) (T_SPACESHIP 25486) (T_SR 25487) ("^" 25488) ("|" 25489)) 11826 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25558) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11827 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25557) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11828 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25556) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11829 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25555) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11830 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25554) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11831 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25553) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11832 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25552) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11833 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25551) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11834 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 25546) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 25547) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 11835 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25545) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11836 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25544) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11837 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25543) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11838 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25542) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11839 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25541) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11840 (("$" 25506) ("(" 25507) (T_NAME_FULLY_QUALIFIED 25508) (T_NAME_QUALIFIED 25509) (T_NAME_RELATIVE 25510) (T_STATIC 25511) (T_STRING 25512) (T_VARIABLE 25513) (class_name 25514) (class_name_reference 25515) (name 25516) (new_variable 25517) (simple_variable 25518)) 11841 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25505) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11842 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25504) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11843 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25503) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11844 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25502) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11845 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25501) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11846 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25500) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11847 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25499) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11848 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25498) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11849 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25497) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11850 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25496) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11851 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25495) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11852 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25494) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11853 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25493) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11854 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25492) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11855 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25491) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11856 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25490) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11857 (("$" 25506) (T_VARIABLE 25513) (simple_variable 25537) ("{" 25538)) 11858 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 25535) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 11859 ((T_PAAMAYIM_NEKUDOTAYIM 25533)) 11860 ((T_NULLSAFE_OBJECT_OPERATOR 25519) (T_OBJECT_OPERATOR 25520) (T_PAAMAYIM_NEKUDOTAYIM 25521) ("[" 25522)) 11861 (("$" 25506) (T_STRING 25526) (T_VARIABLE 25513) (property_name 25532) (simple_variable 25528) ("{" 25529)) 11862 (("$" 25506) (T_STRING 25526) (T_VARIABLE 25513) (property_name 25527) (simple_variable 25528) ("{" 25529)) 11863 (("$" 25506) (T_VARIABLE 25513) (simple_variable 25525)) 11864 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 25523) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 11865 (("]" 25524)) 11866 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 25530) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 11867 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 25531)) 11868 (("$" 25506) (T_VARIABLE 25513) (simple_variable 25534)) 11869 (("%" 304) (")" 25536) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 11870 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 25539) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 11871 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 25540)) 11872 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25550) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11873 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 25548) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 11874 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25549) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11875 (("$" 25381) (T_ABSTRACT 25560) (T_ARRAY 25561) (T_AS 25562) (T_BREAK 25563) (T_CALLABLE 25564) (T_CASE 25565) (T_CATCH 25566) (T_CLASS 25567) (T_CLASS_C 25568) (T_CLONE 25569) (T_CONST 25570) (T_CONTINUE 25571) (T_DECLARE 25572) (T_DEFAULT 25573) (T_DIR 25574) (T_DO 25575) (T_ECHO 25576) (T_ELSE 25577) (T_ELSEIF 25578) (T_EMPTY 25579) (T_ENDDECLARE 25580) (T_ENDFOR 25581) (T_ENDFOREACH 25582) (T_ENDIF 25583) (T_ENDSWITCH 25584) (T_ENDWHILE 25585) (T_ENUM 25586) (T_EVAL 25587) (T_EXIT 25588) (T_EXTENDS 25589) (T_FILE 25590) (T_FINAL 25591) (T_FINALLY 25592) (T_FN 25593) (T_FOR 25594) (T_FOREACH 25595) (T_FUNCTION 25596) (T_FUNC_C 25597) (T_GLOBAL 25598) (T_GOTO 25599) (T_IF 25600) (T_IMPLEMENTS 25601) (T_INCLUDE 25602) (T_INCLUDE_ONCE 25603) (T_INSTANCEOF 25604) (T_INSTEADOF 25605) (T_INTERFACE 25606) (T_ISSET 25607) (T_LINE 25608) (T_LIST 25609) (T_LOGICAL_AND 25610) (T_LOGICAL_OR 25611) (T_LOGICAL_XOR 25612) (T_MATCH 25613) (T_METHOD_C 25614) (T_NAMESPACE 25615) (T_NEW 25616) (T_NS_C 25617) (T_PRINT 25618) (T_PRIVATE 25619) (T_PROPERTY_C 25620) (T_PROTECTED 25621) (T_PUBLIC 25622) (T_READONLY 25623) (T_REQUIRE 25624) (T_REQUIRE_ONCE 25625) (T_RETURN 25626) (T_STATIC 25627) (T_STRING 25628) (T_SWITCH 25629) (T_THROW 25630) (T_TRAIT 25631) (T_TRAIT_C 25632) (T_TRY 25633) (T_UNSET 25634) (T_USE 25635) (T_VAR 25636) (T_VARIABLE 25429) (T_WHILE 25637) (T_YIELD 25638) (identifier 25639) (member_name 25640) (reserved_non_modifiers 25641) (semi_reserved 25642) (simple_variable 25643) ("{" 25644)) 11876 (("(" 25647) (argument_list 25648)) 11877 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 25645) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 11878 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 25646)) 11879 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 25649) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 25650) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 25651) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 11880 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 25654) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 11881 (("," 1498) (possible_comma 25652)) 11882 ((")" 25653)) 11883 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (ampersand 25684) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25685) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11884 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25683) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11885 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25682) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11886 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25681) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11887 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25680) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11888 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25679) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11889 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25678) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11890 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25677) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11891 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25676) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11892 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25675) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11893 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25674) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11894 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25673) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11895 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25672) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11896 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25671) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11897 (("\"" 1538) ("$" 25686) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 25687) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 25688) ("[" 1560) (array_object_dereferenceable 25689) (callable_expr 25690) (callable_variable 25691) (class_constant 1564) (class_name 25692) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 25693) (name 25694) (new_dereferenceable 1570) (simple_variable 25695) (static_member 25696) (variable 25697) (variable_class_name 25698)) 11898 (("$" 25686) (T_VARIABLE 25688) (simple_variable 25731) ("{" 25732)) 11899 (("(" 25702) (argument_list 25730)) 11900 ((T_NULLSAFE_OBJECT_OPERATOR 25716) (T_OBJECT_OPERATOR 25717) ("[" 25718)) 11901 (("(" 25702) (argument_list 25715)) 11902 ((T_PAAMAYIM_NEKUDOTAYIM 25711)) 11903 (("(" 25702) (argument_list 25710)) 11904 ((T_PAAMAYIM_NEKUDOTAYIM 25699)) 11905 (("$" 25686) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 25688) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 25700) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 25701) ("{" 1660)) 11906 (("(" 25702) (argument_list 25703)) 11907 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 25704) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 25705) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 25706) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 11908 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 25709) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 11909 (("," 1498) (possible_comma 25707)) 11910 ((")" 25708)) 11911 (("$" 25686) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 25688) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 25712) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 25713) ("{" 1676)) 11912 (("(" 25702) (argument_list 25714)) 11913 (("$" 25686) (T_STRING 25721) (T_VARIABLE 25688) (property_name 25728) (simple_variable 25723) ("{" 25724)) 11914 (("$" 25686) (T_STRING 25721) (T_VARIABLE 25688) (property_name 25722) (simple_variable 25723) ("{" 25724)) 11915 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 25719) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 11916 (("]" 25720)) 11917 (("(" 25702) (argument_list 25727)) 11918 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 25725) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 11919 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 25726)) 11920 (("(" 25702) (argument_list 25729)) 11921 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 25733) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 11922 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 25734)) 11923 ((backup_doc_comment 25737)) 11924 (("(" 25738)) 11925 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 25739)) 11926 ((")" 25740)) 11927 ((T_USE 166) (lexical_vars 25741)) 11928 ((":" 168) (return_type 25742)) 11929 ((backup_fn_flags 25743)) 11930 (("{" 25744)) 11931 ((inner_statement_list 25745)) 11932 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 25746) ("~" 129)) 11933 ((backup_fn_flags 25747)) 11934 ((backup_doc_comment 25749)) 11935 (("(" 25750)) 11936 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 25751)) 11937 ((")" 25752)) 11938 ((":" 2271) (return_type 25753)) 11939 ((T_DOUBLE_ARROW 25754)) 11940 ((backup_fn_flags 25755)) 11941 ((backup_lex_pos 25756)) 11942 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25757) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11943 (("%" 25459) ("*" 25460) ("+" 25461) ("-" 25462) ("." 25463) ("/" 25464) ("<" 25465) (">" 25466) ("?" 25467) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 25468) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 25469) (T_BOOLEAN_AND 25470) (T_BOOLEAN_OR 25471) (T_COALESCE 25472) (T_INSTANCEOF 25473) (T_IS_EQUAL 25474) (T_IS_GREATER_OR_EQUAL 25475) (T_IS_IDENTICAL 25476) (T_IS_NOT_EQUAL 25477) (T_IS_NOT_IDENTICAL 25478) (T_IS_SMALLER_OR_EQUAL 25479) (T_LOGICAL_AND 25480) (T_LOGICAL_OR 25481) (T_LOGICAL_XOR 25482) (T_PIPE 25483) (T_POW 25484) (T_SL 25485) (T_SPACESHIP 25486) (T_SR 25487) ("^" 25488) (backup_fn_flags 25758) ("|" 25489)) 11944 (("$" 25381) (T_ABSTRACT 25560) (T_ARRAY 25561) (T_AS 25562) (T_BREAK 25563) (T_CALLABLE 25564) (T_CASE 25565) (T_CATCH 25566) (T_CLASS 25567) (T_CLASS_C 25568) (T_CLONE 25569) (T_CONST 25570) (T_CONTINUE 25571) (T_DECLARE 25572) (T_DEFAULT 25573) (T_DIR 25574) (T_DO 25575) (T_ECHO 25576) (T_ELSE 25577) (T_ELSEIF 25578) (T_EMPTY 25579) (T_ENDDECLARE 25580) (T_ENDFOR 25581) (T_ENDFOREACH 25582) (T_ENDIF 25583) (T_ENDSWITCH 25584) (T_ENDWHILE 25585) (T_ENUM 25586) (T_EVAL 25587) (T_EXIT 25588) (T_EXTENDS 25589) (T_FILE 25590) (T_FINAL 25591) (T_FINALLY 25592) (T_FN 25593) (T_FOR 25594) (T_FOREACH 25595) (T_FUNCTION 25596) (T_FUNC_C 25597) (T_GLOBAL 25598) (T_GOTO 25599) (T_IF 25600) (T_IMPLEMENTS 25601) (T_INCLUDE 25602) (T_INCLUDE_ONCE 25603) (T_INSTANCEOF 25604) (T_INSTEADOF 25605) (T_INTERFACE 25606) (T_ISSET 25607) (T_LINE 25608) (T_LIST 25609) (T_LOGICAL_AND 25610) (T_LOGICAL_OR 25611) (T_LOGICAL_XOR 25612) (T_MATCH 25613) (T_METHOD_C 25614) (T_NAMESPACE 25615) (T_NEW 25616) (T_NS_C 25617) (T_PRINT 25618) (T_PRIVATE 25619) (T_PROPERTY_C 25620) (T_PROTECTED 25621) (T_PUBLIC 25622) (T_READONLY 25623) (T_REQUIRE 25624) (T_REQUIRE_ONCE 25625) (T_RETURN 25626) (T_STATIC 25627) (T_STRING 25628) (T_SWITCH 25629) (T_THROW 25630) (T_TRAIT 25631) (T_TRAIT_C 25632) (T_TRY 25633) (T_UNSET 25634) (T_USE 25635) (T_VAR 25636) (T_VARIABLE 25429) (T_WHILE 25637) (T_YIELD 25638) (identifier 25761) (member_name 25762) (reserved_non_modifiers 25641) (semi_reserved 25642) (simple_variable 25763) ("{" 25764)) 11945 (("(" 25647) (argument_list 25767)) 11946 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 25765) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 11947 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 25766)) 11948 ((T_FN 35) (T_FUNCTION 131) (fn 25443) #'25444 (inline_function 25771)) 11949 (("$" 25381) (T_STRING 25777) (T_VARIABLE 25429) (property_name 25784) (simple_variable 25779) ("{" 25780)) 11950 (("$" 25381) (T_STRING 25777) (T_VARIABLE 25429) (property_name 25778) (simple_variable 25779) ("{" 25780)) 11951 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 25775) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 11952 (("]" 25776)) 11953 (("(" 25647) (argument_list 25783)) 11954 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 25781) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 11955 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 25782)) 11956 (("(" 25647) (argument_list 25785)) 11957 (("`" 25787)) 11958 (("]" 25789)) 11959 (("=" 25790)) 11960 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 25791) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 11961 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 26326) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 11962 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 26324) (encaps_var 5357)) 11963 (("$" 25795) (T_VARIABLE 25843) (simple_variable 26320) ("{" 26321)) 11964 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 26319) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 11965 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 26318) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 11966 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 26317) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 11967 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 26316) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 11968 (("(" 26313)) 11969 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 26312) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 11970 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 26311) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 11971 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 26298) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (clone_argument_list 26299) (constant 25854) (dereferenceable_scalar 25855) (expr 26300) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 11972 (("\"" 1538) ("$" 26100) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 26101) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 26102) ("[" 1560) (array_object_dereferenceable 26103) (callable_expr 26104) (callable_variable 26105) (class_constant 1564) (class_name 26106) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 26107) (name 26108) (new_dereferenceable 1570) (simple_variable 26109) (static_member 26110) (variable 26297) (variable_class_name 26112)) 11973 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 26296) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 11974 (("(" 26293)) 11975 (("(" 26290)) 11976 (("(" 26281) (argument_list 26282) (ctor_arguments 26283)) 11977 (("\"" 1538) ("$" 26100) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 26101) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 26102) ("[" 1560) (array_object_dereferenceable 26103) (callable_expr 26104) (callable_variable 26105) (class_constant 1564) (class_name 26106) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 26107) (name 26108) (new_dereferenceable 1570) (simple_variable 26109) (static_member 26110) (variable 26280) (variable_class_name 26112)) 11978 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 26279) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 11979 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 26278) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 11980 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 26277) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 11981 (("(" 26273)) 11982 (("(" 26268)) 11983 (("(" 26262)) 11984 (("$" 26100) ("(" 26225) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 26226) (T_NAME_QUALIFIED 26227) (T_NAME_RELATIVE 26228) (T_READONLY 2955) (T_STATIC 26229) (T_STRING 26230) (T_VARIABLE 26102) (anonymous_class 26231) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 26232) (attribute 2961) (attributes 26233) (class_modifier 2963) (class_name 26234) (class_name_reference 26235) (name 26236) (new_variable 26237) (simple_variable 26238)) 11985 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 26224) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 11986 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 26223) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 11987 (("(" 26061) (argument_list 26222)) 11988 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 26221) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 11989 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 26220) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 11990 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 26215) (T_END_HEREDOC 26216) (T_VARIABLE 2920) (encaps_list 26217) (encaps_var 2922)) 11991 ((T_FN 35) (T_FUNCTION 131) (fn 25857) #'25858 (inline_function 26214)) 11992 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 26213) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 11993 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 26212) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 11994 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 26211) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 11995 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 26208) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 11996 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 26207) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 11997 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 26203) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 11998 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 26201) (encaps_list 2828) (encaps_var 2829)) 11999 ((T_NULLSAFE_OBJECT_OPERATOR 26187) (T_OBJECT_OPERATOR 26188) ("[" 26189)) 12000 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 26184) (attribute 2757) (fn 25857) #'25858 (inline_function 26185)) 12001 (("(" 26061) (argument_list 26183)) 12002 ((T_PAAMAYIM_NEKUDOTAYIM 26175)) 12003 (("%" 25873) ("*" 25874) ("+" 25875) ("-" 25876) ("." 25877) ("/" 25878) ("<" 25879) (">" 25880) ("?" 25881) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 25882) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 25883) (T_BOOLEAN_AND 25884) (T_BOOLEAN_OR 25885) (T_COALESCE 25886) (T_DOUBLE_ARROW 26173) (T_INSTANCEOF 25887) (T_IS_EQUAL 25888) (T_IS_GREATER_OR_EQUAL 25889) (T_IS_IDENTICAL 25890) (T_IS_NOT_EQUAL 25891) (T_IS_NOT_IDENTICAL 25892) (T_IS_SMALLER_OR_EQUAL 25893) (T_LOGICAL_AND 25894) (T_LOGICAL_OR 25895) (T_LOGICAL_XOR 25896) (T_PIPE 25897) (T_POW 25898) (T_SL 25899) (T_SPACESHIP 25900) (T_SR 25901) ("^" 25902) ("|" 25903)) 12004 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 26162)) 12005 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 26150)) 12006 (("(" 26061) (argument_list 26149)) 12007 (("=" 26069) (T_AND_EQUAL 26070) (T_COALESCE_EQUAL 26071) (T_CONCAT_EQUAL 26072) (T_DEC 26073) (T_DIV_EQUAL 26074) (T_INC 26075) (T_MINUS_EQUAL 26076) (T_MOD_EQUAL 26077) (T_MUL_EQUAL 26078) (T_OR_EQUAL 26079) (T_PLUS_EQUAL 26080) (T_POW_EQUAL 26081) (T_SL_EQUAL 26082) (T_SR_EQUAL 26083) (T_XOR_EQUAL 26084)) 12008 ((T_PAAMAYIM_NEKUDOTAYIM 25973)) 12009 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 25872) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12010 (("%" 25873) ("*" 25874) ("+" 25875) ("-" 25876) ("." 25877) ("/" 25878) ("<" 25879) (">" 25880) ("?" 25881) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 25882) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 25883) (T_BOOLEAN_AND 25884) (T_BOOLEAN_OR 25885) (T_COALESCE 25886) (T_INSTANCEOF 25887) (T_IS_EQUAL 25888) (T_IS_GREATER_OR_EQUAL 25889) (T_IS_IDENTICAL 25890) (T_IS_NOT_EQUAL 25891) (T_IS_NOT_IDENTICAL 25892) (T_IS_SMALLER_OR_EQUAL 25893) (T_LOGICAL_AND 25894) (T_LOGICAL_OR 25895) (T_LOGICAL_XOR 25896) (T_PIPE 25897) (T_POW 25898) (T_SL 25899) (T_SPACESHIP 25900) (T_SR 25901) ("^" 25902) ("|" 25903)) 12011 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 25972) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12012 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 25971) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12013 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 25970) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12014 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 25969) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12015 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 25968) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12016 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 25967) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12017 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 25966) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12018 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 25965) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12019 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 25960) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 25961) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 12020 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 25959) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12021 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 25958) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12022 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 25957) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12023 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 25956) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12024 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 25955) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12025 (("$" 25920) ("(" 25921) (T_NAME_FULLY_QUALIFIED 25922) (T_NAME_QUALIFIED 25923) (T_NAME_RELATIVE 25924) (T_STATIC 25925) (T_STRING 25926) (T_VARIABLE 25927) (class_name 25928) (class_name_reference 25929) (name 25930) (new_variable 25931) (simple_variable 25932)) 12026 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 25919) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12027 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 25918) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12028 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 25917) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12029 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 25916) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12030 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 25915) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12031 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 25914) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12032 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 25913) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12033 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 25912) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12034 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 25911) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12035 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 25910) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12036 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 25909) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12037 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 25908) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12038 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 25907) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12039 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 25906) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12040 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 25905) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12041 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 25904) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12042 (("$" 25920) (T_VARIABLE 25927) (simple_variable 25951) ("{" 25952)) 12043 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 25949) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 12044 ((T_PAAMAYIM_NEKUDOTAYIM 25947)) 12045 ((T_NULLSAFE_OBJECT_OPERATOR 25933) (T_OBJECT_OPERATOR 25934) (T_PAAMAYIM_NEKUDOTAYIM 25935) ("[" 25936)) 12046 (("$" 25920) (T_STRING 25940) (T_VARIABLE 25927) (property_name 25946) (simple_variable 25942) ("{" 25943)) 12047 (("$" 25920) (T_STRING 25940) (T_VARIABLE 25927) (property_name 25941) (simple_variable 25942) ("{" 25943)) 12048 (("$" 25920) (T_VARIABLE 25927) (simple_variable 25939)) 12049 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 25937) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 12050 (("]" 25938)) 12051 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 25944) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 12052 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 25945)) 12053 (("$" 25920) (T_VARIABLE 25927) (simple_variable 25948)) 12054 (("%" 304) (")" 25950) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 12055 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 25953) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 12056 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 25954)) 12057 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 25964) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12058 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 25962) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 12059 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 25963) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12060 (("$" 25795) (T_ABSTRACT 25974) (T_ARRAY 25975) (T_AS 25976) (T_BREAK 25977) (T_CALLABLE 25978) (T_CASE 25979) (T_CATCH 25980) (T_CLASS 25981) (T_CLASS_C 25982) (T_CLONE 25983) (T_CONST 25984) (T_CONTINUE 25985) (T_DECLARE 25986) (T_DEFAULT 25987) (T_DIR 25988) (T_DO 25989) (T_ECHO 25990) (T_ELSE 25991) (T_ELSEIF 25992) (T_EMPTY 25993) (T_ENDDECLARE 25994) (T_ENDFOR 25995) (T_ENDFOREACH 25996) (T_ENDIF 25997) (T_ENDSWITCH 25998) (T_ENDWHILE 25999) (T_ENUM 26000) (T_EVAL 26001) (T_EXIT 26002) (T_EXTENDS 26003) (T_FILE 26004) (T_FINAL 26005) (T_FINALLY 26006) (T_FN 26007) (T_FOR 26008) (T_FOREACH 26009) (T_FUNCTION 26010) (T_FUNC_C 26011) (T_GLOBAL 26012) (T_GOTO 26013) (T_IF 26014) (T_IMPLEMENTS 26015) (T_INCLUDE 26016) (T_INCLUDE_ONCE 26017) (T_INSTANCEOF 26018) (T_INSTEADOF 26019) (T_INTERFACE 26020) (T_ISSET 26021) (T_LINE 26022) (T_LIST 26023) (T_LOGICAL_AND 26024) (T_LOGICAL_OR 26025) (T_LOGICAL_XOR 26026) (T_MATCH 26027) (T_METHOD_C 26028) (T_NAMESPACE 26029) (T_NEW 26030) (T_NS_C 26031) (T_PRINT 26032) (T_PRIVATE 26033) (T_PROPERTY_C 26034) (T_PROTECTED 26035) (T_PUBLIC 26036) (T_READONLY 26037) (T_REQUIRE 26038) (T_REQUIRE_ONCE 26039) (T_RETURN 26040) (T_STATIC 26041) (T_STRING 26042) (T_SWITCH 26043) (T_THROW 26044) (T_TRAIT 26045) (T_TRAIT_C 26046) (T_TRY 26047) (T_UNSET 26048) (T_USE 26049) (T_VAR 26050) (T_VARIABLE 25843) (T_WHILE 26051) (T_YIELD 26052) (identifier 26053) (member_name 26054) (reserved_non_modifiers 26055) (semi_reserved 26056) (simple_variable 26057) ("{" 26058)) 12061 (("(" 26061) (argument_list 26062)) 12062 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 26059) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 12063 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 26060)) 12064 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 26063) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 26064) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 26065) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 12065 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 26068) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 12066 (("," 1498) (possible_comma 26066)) 12067 ((")" 26067)) 12068 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (ampersand 26098) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 26099) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12069 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 26097) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12070 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 26096) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12071 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 26095) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12072 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 26094) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12073 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 26093) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12074 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 26092) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12075 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 26091) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12076 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 26090) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12077 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 26089) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12078 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 26088) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12079 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 26087) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12080 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 26086) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12081 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 26085) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12082 (("\"" 1538) ("$" 26100) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 26101) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 26102) ("[" 1560) (array_object_dereferenceable 26103) (callable_expr 26104) (callable_variable 26105) (class_constant 1564) (class_name 26106) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 26107) (name 26108) (new_dereferenceable 1570) (simple_variable 26109) (static_member 26110) (variable 26111) (variable_class_name 26112)) 12083 (("$" 26100) (T_VARIABLE 26102) (simple_variable 26145) ("{" 26146)) 12084 (("(" 26116) (argument_list 26144)) 12085 ((T_NULLSAFE_OBJECT_OPERATOR 26130) (T_OBJECT_OPERATOR 26131) ("[" 26132)) 12086 (("(" 26116) (argument_list 26129)) 12087 ((T_PAAMAYIM_NEKUDOTAYIM 26125)) 12088 (("(" 26116) (argument_list 26124)) 12089 ((T_PAAMAYIM_NEKUDOTAYIM 26113)) 12090 (("$" 26100) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 26102) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 26114) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 26115) ("{" 1660)) 12091 (("(" 26116) (argument_list 26117)) 12092 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 26118) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 26119) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 26120) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 12093 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 26123) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 12094 (("," 1498) (possible_comma 26121)) 12095 ((")" 26122)) 12096 (("$" 26100) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 26102) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 26126) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 26127) ("{" 1676)) 12097 (("(" 26116) (argument_list 26128)) 12098 (("$" 26100) (T_STRING 26135) (T_VARIABLE 26102) (property_name 26142) (simple_variable 26137) ("{" 26138)) 12099 (("$" 26100) (T_STRING 26135) (T_VARIABLE 26102) (property_name 26136) (simple_variable 26137) ("{" 26138)) 12100 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 26133) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 12101 (("]" 26134)) 12102 (("(" 26116) (argument_list 26141)) 12103 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 26139) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 12104 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 26140)) 12105 (("(" 26116) (argument_list 26143)) 12106 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 26147) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 12107 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 26148)) 12108 ((backup_doc_comment 26151)) 12109 (("(" 26152)) 12110 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 26153)) 12111 ((")" 26154)) 12112 ((T_USE 166) (lexical_vars 26155)) 12113 ((":" 168) (return_type 26156)) 12114 ((backup_fn_flags 26157)) 12115 (("{" 26158)) 12116 ((inner_statement_list 26159)) 12117 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 26160) ("~" 129)) 12118 ((backup_fn_flags 26161)) 12119 ((backup_doc_comment 26163)) 12120 (("(" 26164)) 12121 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 26165)) 12122 ((")" 26166)) 12123 ((":" 2271) (return_type 26167)) 12124 ((T_DOUBLE_ARROW 26168)) 12125 ((backup_fn_flags 26169)) 12126 ((backup_lex_pos 26170)) 12127 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 26171) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12128 (("%" 25873) ("*" 25874) ("+" 25875) ("-" 25876) ("." 25877) ("/" 25878) ("<" 25879) (">" 25880) ("?" 25881) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 25882) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 25883) (T_BOOLEAN_AND 25884) (T_BOOLEAN_OR 25885) (T_COALESCE 25886) (T_INSTANCEOF 25887) (T_IS_EQUAL 25888) (T_IS_GREATER_OR_EQUAL 25889) (T_IS_IDENTICAL 25890) (T_IS_NOT_EQUAL 25891) (T_IS_NOT_IDENTICAL 25892) (T_IS_SMALLER_OR_EQUAL 25893) (T_LOGICAL_AND 25894) (T_LOGICAL_OR 25895) (T_LOGICAL_XOR 25896) (T_PIPE 25897) (T_POW 25898) (T_SL 25899) (T_SPACESHIP 25900) (T_SR 25901) ("^" 25902) (backup_fn_flags 26172) ("|" 25903)) 12129 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 26174) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 12130 (("$" 25795) (T_ABSTRACT 25974) (T_ARRAY 25975) (T_AS 25976) (T_BREAK 25977) (T_CALLABLE 25978) (T_CASE 25979) (T_CATCH 25980) (T_CLASS 25981) (T_CLASS_C 25982) (T_CLONE 25983) (T_CONST 25984) (T_CONTINUE 25985) (T_DECLARE 25986) (T_DEFAULT 25987) (T_DIR 25988) (T_DO 25989) (T_ECHO 25990) (T_ELSE 25991) (T_ELSEIF 25992) (T_EMPTY 25993) (T_ENDDECLARE 25994) (T_ENDFOR 25995) (T_ENDFOREACH 25996) (T_ENDIF 25997) (T_ENDSWITCH 25998) (T_ENDWHILE 25999) (T_ENUM 26000) (T_EVAL 26001) (T_EXIT 26002) (T_EXTENDS 26003) (T_FILE 26004) (T_FINAL 26005) (T_FINALLY 26006) (T_FN 26007) (T_FOR 26008) (T_FOREACH 26009) (T_FUNCTION 26010) (T_FUNC_C 26011) (T_GLOBAL 26012) (T_GOTO 26013) (T_IF 26014) (T_IMPLEMENTS 26015) (T_INCLUDE 26016) (T_INCLUDE_ONCE 26017) (T_INSTANCEOF 26018) (T_INSTEADOF 26019) (T_INTERFACE 26020) (T_ISSET 26021) (T_LINE 26022) (T_LIST 26023) (T_LOGICAL_AND 26024) (T_LOGICAL_OR 26025) (T_LOGICAL_XOR 26026) (T_MATCH 26027) (T_METHOD_C 26028) (T_NAMESPACE 26029) (T_NEW 26030) (T_NS_C 26031) (T_PRINT 26032) (T_PRIVATE 26033) (T_PROPERTY_C 26034) (T_PROTECTED 26035) (T_PUBLIC 26036) (T_READONLY 26037) (T_REQUIRE 26038) (T_REQUIRE_ONCE 26039) (T_RETURN 26040) (T_STATIC 26041) (T_STRING 26042) (T_SWITCH 26043) (T_THROW 26044) (T_TRAIT 26045) (T_TRAIT_C 26046) (T_TRY 26047) (T_UNSET 26048) (T_USE 26049) (T_VAR 26050) (T_VARIABLE 25843) (T_WHILE 26051) (T_YIELD 26052) (identifier 26176) (member_name 26177) (reserved_non_modifiers 26055) (semi_reserved 26056) (simple_variable 26178) ("{" 26179)) 12131 (("(" 26061) (argument_list 26182)) 12132 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 26180) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 12133 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 26181)) 12134 ((T_FN 35) (T_FUNCTION 131) (fn 25857) #'25858 (inline_function 26186)) 12135 (("$" 25795) (T_STRING 26192) (T_VARIABLE 25843) (property_name 26199) (simple_variable 26194) ("{" 26195)) 12136 (("$" 25795) (T_STRING 26192) (T_VARIABLE 25843) (property_name 26193) (simple_variable 26194) ("{" 26195)) 12137 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 26190) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 12138 (("]" 26191)) 12139 (("(" 26061) (argument_list 26198)) 12140 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 26196) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 12141 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 26197)) 12142 (("(" 26061) (argument_list 26200)) 12143 (("`" 26202)) 12144 (("]" 26204)) 12145 (("=" 26205)) 12146 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 26206) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12147 (("%" 25873) ("*" 25874) ("+" 25875) ("-" 25876) ("." 25877) ("/" 25878) ("<" 25879) (">" 25880) ("?" 25881) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 25882) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 25883) (T_BOOLEAN_AND 25884) (T_BOOLEAN_OR 25885) (T_COALESCE 25886) (T_DOUBLE_ARROW 26209) (T_INSTANCEOF 25887) (T_IS_EQUAL 25888) (T_IS_GREATER_OR_EQUAL 25889) (T_IS_IDENTICAL 25890) (T_IS_NOT_EQUAL 25891) (T_IS_NOT_IDENTICAL 25892) (T_IS_SMALLER_OR_EQUAL 25893) (T_LOGICAL_AND 25894) (T_LOGICAL_OR 25895) (T_LOGICAL_XOR 25896) (T_PIPE 25897) (T_POW 25898) (T_SL 25899) (T_SPACESHIP 25900) (T_SR 25901) ("^" 25902) ("|" 25903)) 12148 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 26210) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12149 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 26219) (T_VARIABLE 2920) (encaps_var 2934)) 12150 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 26218) (T_VARIABLE 2920) (encaps_var 2925)) 12151 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 26260) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 12152 ((T_CLASS 26252)) 12153 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 26251) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 26232) (attribute 2982) (class_modifier 2963)) 12154 ((T_PAAMAYIM_NEKUDOTAYIM 26249)) 12155 (("(" 26116) (argument_list 26248)) 12156 ((T_NULLSAFE_OBJECT_OPERATOR 26239) (T_OBJECT_OPERATOR 26240) (T_PAAMAYIM_NEKUDOTAYIM 26241) ("[" 26242)) 12157 (("$" 26100) (T_STRING 26135) (T_VARIABLE 26102) (property_name 26247) (simple_variable 26137) ("{" 26138)) 12158 (("$" 26100) (T_STRING 26135) (T_VARIABLE 26102) (property_name 26246) (simple_variable 26137) ("{" 26138)) 12159 (("$" 26100) (T_VARIABLE 26102) (simple_variable 26245)) 12160 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 26243) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 12161 (("]" 26244)) 12162 (("$" 26100) (T_VARIABLE 26102) (simple_variable 26250)) 12163 (("(" 2984) (argument_list 2985) (ctor_arguments 26253)) 12164 ((T_EXTENDS 2987) (extends_from 26254)) 12165 ((T_IMPLEMENTS 2989) (implements_list 26255)) 12166 ((backup_doc_comment 26256)) 12167 (("{" 26257)) 12168 ((class_statement_list 26258)) 12169 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 26259)) 12170 (("%" 304) (")" 26261) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 12171 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 26263) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 12172 (("%" 304) (")" 26264) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 12173 (("{" 26265)) 12174 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 26266) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 12175 (("}" 26267)) 12176 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 26269) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 12177 ((")" 26270)) 12178 (("=" 26271)) 12179 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 26272) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12180 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 26274) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 12181 (("," 5226) (possible_comma 26275)) 12182 ((")" 26276)) 12183 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 26284) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 26285) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 26286) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 12184 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 26289) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 12185 (("," 1498) (possible_comma 26287)) 12186 ((")" 26288)) 12187 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 26291) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 12188 (("%" 304) (")" 26292) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 12189 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 26294) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 12190 (("%" 304) (")" 26295) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 12191 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 26301) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 26302) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 26303) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 26304) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 12192 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 26310) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 12193 (("%" 2442) (")" 26307) ("*" 2443) ("+" 2444) ("," 26308) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 12194 (("," 5287) (possible_comma 26305)) 12195 ((")" 26306)) 12196 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 26309) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 12197 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 26314) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 12198 ((")" 26315)) 12199 (("%" 304) (")" 26307) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 12200 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 26322) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 12201 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 26323)) 12202 (("\"" 26325) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 12203 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 26335) (T_VARIABLE 2920) (encaps_var 2934)) 12204 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 26334) (T_VARIABLE 2920) (encaps_var 2925)) 12205 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 26376) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 12206 ((T_CLASS 26368)) 12207 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 26367) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 26348) (attribute 2982) (class_modifier 2963)) 12208 ((T_PAAMAYIM_NEKUDOTAYIM 26365)) 12209 (("(" 25702) (argument_list 26364)) 12210 ((T_NULLSAFE_OBJECT_OPERATOR 26355) (T_OBJECT_OPERATOR 26356) (T_PAAMAYIM_NEKUDOTAYIM 26357) ("[" 26358)) 12211 (("$" 25686) (T_STRING 25721) (T_VARIABLE 25688) (property_name 26363) (simple_variable 25723) ("{" 25724)) 12212 (("$" 25686) (T_STRING 25721) (T_VARIABLE 25688) (property_name 26362) (simple_variable 25723) ("{" 25724)) 12213 (("$" 25686) (T_VARIABLE 25688) (simple_variable 26361)) 12214 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 26359) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 12215 (("]" 26360)) 12216 (("$" 25686) (T_VARIABLE 25688) (simple_variable 26366)) 12217 (("(" 2984) (argument_list 2985) (ctor_arguments 26369)) 12218 ((T_EXTENDS 2987) (extends_from 26370)) 12219 ((T_IMPLEMENTS 2989) (implements_list 26371)) 12220 ((backup_doc_comment 26372)) 12221 (("{" 26373)) 12222 ((class_statement_list 26374)) 12223 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 26375)) 12224 (("%" 304) (")" 26377) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 12225 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 26379) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 12226 (("%" 304) (")" 26380) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 12227 (("{" 26381)) 12228 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 26382) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 12229 (("}" 26383)) 12230 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 26385) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 12231 ((")" 26386)) 12232 (("=" 26387)) 12233 (("!" 25379) ("\"" 25380) ("$" 25381) ("(" 25382) ("+" 25383) ("-" 25384) ("@" 25385) (T_ARRAY 25386) (T_ARRAY_CAST 25387) (T_ATTRIBUTE 130) (T_BOOL_CAST 25388) (T_CLASS_C 25389) (T_CLONE 25390) (T_CONSTANT_ENCAPSED_STRING 25391) (T_DEC 25392) (T_DIR 25393) (T_DNUMBER 25394) (T_DOUBLE_CAST 25395) (T_EMPTY 25396) (T_EVAL 25397) (T_EXIT 25398) (T_FILE 25399) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25400) (T_INC 25401) (T_INCLUDE 25402) (T_INCLUDE_ONCE 25403) (T_INT_CAST 25404) (T_ISSET 25405) (T_LINE 25406) (T_LIST 25407) (T_LNUMBER 25408) (T_MATCH 25409) (T_METHOD_C 25410) (T_NAME_FULLY_QUALIFIED 25411) (T_NAME_QUALIFIED 25412) (T_NAME_RELATIVE 25413) (T_NEW 25414) (T_NS_C 25415) (T_OBJECT_CAST 25416) (T_PRINT 25417) (T_PROPERTY_C 25418) (T_READONLY 25419) (T_REQUIRE 25420) (T_REQUIRE_ONCE 25421) (T_START_HEREDOC 25422) (T_STATIC 25423) (T_STRING 25424) (T_STRING_CAST 25425) (T_THROW 25426) (T_TRAIT_C 25427) (T_UNSET_CAST 25428) (T_VARIABLE 25429) (T_YIELD 25430) (T_YIELD_FROM 25431) ("[" 25432) ("`" 25433) (array_object_dereferenceable 25434) (attribute 135) (attributes 25435) (callable_expr 25436) (callable_variable 25437) (class_constant 25438) (class_name 25439) (constant 25440) (dereferenceable_scalar 25441) (expr 26388) (fn 25443) (fully_dereferenceable 107) #'25444 (function_call 25445) (inline_function 25446) (internal_functions_in_yacc 25447) (match 25448) (name 25449) (new_dereferenceable 25450) (new_non_dereferenceable 25451) (scalar 25452) (simple_variable 25453) (static_member 25454) (variable 25455) (variable_class_name 25456) ("~" 25457)) 12234 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 26390) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 12235 (("," 5226) (possible_comma 26391)) 12236 ((")" 26392)) 12237 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 26400) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 26401) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 26402) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 12238 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 26405) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 12239 (("," 1498) (possible_comma 26403)) 12240 ((")" 26404)) 12241 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 26407) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 12242 (("%" 304) (")" 26408) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 12243 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 26410) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 12244 (("%" 304) (")" 26411) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 12245 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 26417) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 26418) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 26419) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 26420) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 12246 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 26426) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 12247 (("%" 2442) (")" 26423) ("*" 2443) ("+" 2444) ("," 26424) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 12248 (("," 5287) (possible_comma 26421)) 12249 ((")" 26422)) 12250 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 26425) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 12251 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 26430) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 12252 ((")" 26431)) 12253 (("%" 304) (")" 26423) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 12254 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 26438) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 12255 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 26439)) 12256 (("\"" 26441) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 12257 ((inner_statement_list 26446)) 12258 ((inner_statement_list 26445)) 12259 (("$" 25003) (T_ABSTRACT 25182) (T_ARRAY 25183) (T_AS 25184) (T_BREAK 25185) (T_CALLABLE 25186) (T_CASE 25187) (T_CATCH 25188) (T_CLASS 25189) (T_CLASS_C 25190) (T_CLONE 25191) (T_CONST 25192) (T_CONTINUE 25193) (T_DECLARE 25194) (T_DEFAULT 25195) (T_DIR 25196) (T_DO 25197) (T_ECHO 25198) (T_ELSE 25199) (T_ELSEIF 25200) (T_EMPTY 25201) (T_ENDDECLARE 25202) (T_ENDFOR 25203) (T_ENDFOREACH 25204) (T_ENDIF 25205) (T_ENDSWITCH 25206) (T_ENDWHILE 25207) (T_ENUM 25208) (T_EVAL 25209) (T_EXIT 25210) (T_EXTENDS 25211) (T_FILE 25212) (T_FINAL 25213) (T_FINALLY 25214) (T_FN 25215) (T_FOR 25216) (T_FOREACH 25217) (T_FUNCTION 25218) (T_FUNC_C 25219) (T_GLOBAL 25220) (T_GOTO 25221) (T_IF 25222) (T_IMPLEMENTS 25223) (T_INCLUDE 25224) (T_INCLUDE_ONCE 25225) (T_INSTANCEOF 25226) (T_INSTEADOF 25227) (T_INTERFACE 25228) (T_ISSET 25229) (T_LINE 25230) (T_LIST 25231) (T_LOGICAL_AND 25232) (T_LOGICAL_OR 25233) (T_LOGICAL_XOR 25234) (T_MATCH 25235) (T_METHOD_C 25236) (T_NAMESPACE 25237) (T_NEW 25238) (T_NS_C 25239) (T_PRINT 25240) (T_PRIVATE 25241) (T_PROPERTY_C 25242) (T_PROTECTED 25243) (T_PUBLIC 25244) (T_READONLY 25245) (T_REQUIRE 25246) (T_REQUIRE_ONCE 25247) (T_RETURN 25248) (T_STATIC 25249) (T_STRING 25250) (T_SWITCH 25251) (T_THROW 25252) (T_TRAIT 25253) (T_TRAIT_C 25254) (T_TRY 25255) (T_UNSET 25256) (T_USE 25257) (T_VAR 25258) (T_VARIABLE 25051) (T_WHILE 25259) (T_YIELD 25260) (identifier 26448) (member_name 26449) (reserved_non_modifiers 25263) (semi_reserved 25264) (simple_variable 26450) ("{" 26451)) 12260 (("(" 25269) (argument_list 26454)) 12261 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 26452) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 12262 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 26453)) 12263 ((T_FN 35) (T_FUNCTION 131) (fn 25065) #'25066 (inline_function 26458)) 12264 (("$" 25003) (T_STRING 26464) (T_VARIABLE 25051) (property_name 26471) (simple_variable 26466) ("{" 26467)) 12265 (("$" 25003) (T_STRING 26464) (T_VARIABLE 25051) (property_name 26465) (simple_variable 26466) ("{" 26467)) 12266 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 26462) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 12267 (("]" 26463)) 12268 (("(" 25269) (argument_list 26470)) 12269 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 26468) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 12270 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 26469)) 12271 (("(" 25269) (argument_list 26472)) 12272 (("`" 26474)) 12273 (("]" 26476)) 12274 (("=" 26477)) 12275 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 26478) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 12276 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 27013) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12277 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 27011) (encaps_var 5357)) 12278 (("$" 26482) (T_VARIABLE 26530) (simple_variable 27007) ("{" 27008)) 12279 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 27006) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 12280 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 27005) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12281 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 27004) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12282 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 27003) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12283 (("(" 27000)) 12284 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26999) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12285 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26998) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12286 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26985) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (clone_argument_list 26986) (constant 26541) (dereferenceable_scalar 26542) (expr 26987) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12287 (("\"" 1538) ("$" 26787) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 26788) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 26789) ("[" 1560) (array_object_dereferenceable 26790) (callable_expr 26791) (callable_variable 26792) (class_constant 1564) (class_name 26793) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 26794) (name 26795) (new_dereferenceable 1570) (simple_variable 26796) (static_member 26797) (variable 26984) (variable_class_name 26799)) 12288 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26983) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12289 (("(" 26980)) 12290 (("(" 26977)) 12291 (("(" 26968) (argument_list 26969) (ctor_arguments 26970)) 12292 (("\"" 1538) ("$" 26787) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 26788) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 26789) ("[" 1560) (array_object_dereferenceable 26790) (callable_expr 26791) (callable_variable 26792) (class_constant 1564) (class_name 26793) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 26794) (name 26795) (new_dereferenceable 1570) (simple_variable 26796) (static_member 26797) (variable 26967) (variable_class_name 26799)) 12293 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26966) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12294 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26965) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12295 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26964) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12296 (("(" 26960)) 12297 (("(" 26955)) 12298 (("(" 26949)) 12299 (("$" 26787) ("(" 26912) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 26913) (T_NAME_QUALIFIED 26914) (T_NAME_RELATIVE 26915) (T_READONLY 2955) (T_STATIC 26916) (T_STRING 26917) (T_VARIABLE 26789) (anonymous_class 26918) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 26919) (attribute 2961) (attributes 26920) (class_modifier 2963) (class_name 26921) (class_name_reference 26922) (name 26923) (new_variable 26924) (simple_variable 26925)) 12300 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26911) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12301 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26910) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12302 (("(" 26748) (argument_list 26909)) 12303 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26908) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12304 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26907) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12305 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 26902) (T_END_HEREDOC 26903) (T_VARIABLE 2920) (encaps_list 26904) (encaps_var 2922)) 12306 ((T_FN 35) (T_FUNCTION 131) (fn 26544) #'26545 (inline_function 26901)) 12307 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26900) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12308 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26899) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12309 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26898) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12310 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26895) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12311 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26894) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12312 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 26890) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 12313 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 26888) (encaps_list 2828) (encaps_var 2829)) 12314 ((T_NULLSAFE_OBJECT_OPERATOR 26874) (T_OBJECT_OPERATOR 26875) ("[" 26876)) 12315 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 26871) (attribute 2757) (fn 26544) #'26545 (inline_function 26872)) 12316 (("(" 26748) (argument_list 26870)) 12317 ((T_PAAMAYIM_NEKUDOTAYIM 26862)) 12318 (("%" 26560) ("*" 26561) ("+" 26562) ("-" 26563) ("." 26564) ("/" 26565) ("<" 26566) (">" 26567) ("?" 26568) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 26569) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 26570) (T_BOOLEAN_AND 26571) (T_BOOLEAN_OR 26572) (T_COALESCE 26573) (T_DOUBLE_ARROW 26860) (T_INSTANCEOF 26574) (T_IS_EQUAL 26575) (T_IS_GREATER_OR_EQUAL 26576) (T_IS_IDENTICAL 26577) (T_IS_NOT_EQUAL 26578) (T_IS_NOT_IDENTICAL 26579) (T_IS_SMALLER_OR_EQUAL 26580) (T_LOGICAL_AND 26581) (T_LOGICAL_OR 26582) (T_LOGICAL_XOR 26583) (T_PIPE 26584) (T_POW 26585) (T_SL 26586) (T_SPACESHIP 26587) (T_SR 26588) ("^" 26589) ("|" 26590)) 12319 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 26849)) 12320 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 26837)) 12321 (("(" 26748) (argument_list 26836)) 12322 (("=" 26756) (T_AND_EQUAL 26757) (T_COALESCE_EQUAL 26758) (T_CONCAT_EQUAL 26759) (T_DEC 26760) (T_DIV_EQUAL 26761) (T_INC 26762) (T_MINUS_EQUAL 26763) (T_MOD_EQUAL 26764) (T_MUL_EQUAL 26765) (T_OR_EQUAL 26766) (T_PLUS_EQUAL 26767) (T_POW_EQUAL 26768) (T_SL_EQUAL 26769) (T_SR_EQUAL 26770) (T_XOR_EQUAL 26771)) 12323 ((T_PAAMAYIM_NEKUDOTAYIM 26660)) 12324 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26559) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12325 (("%" 26560) ("*" 26561) ("+" 26562) ("-" 26563) ("." 26564) ("/" 26565) ("<" 26566) (">" 26567) ("?" 26568) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 26569) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 26570) (T_BOOLEAN_AND 26571) (T_BOOLEAN_OR 26572) (T_COALESCE 26573) (T_INSTANCEOF 26574) (T_IS_EQUAL 26575) (T_IS_GREATER_OR_EQUAL 26576) (T_IS_IDENTICAL 26577) (T_IS_NOT_EQUAL 26578) (T_IS_NOT_IDENTICAL 26579) (T_IS_SMALLER_OR_EQUAL 26580) (T_LOGICAL_AND 26581) (T_LOGICAL_OR 26582) (T_LOGICAL_XOR 26583) (T_PIPE 26584) (T_POW 26585) (T_SL 26586) (T_SPACESHIP 26587) (T_SR 26588) ("^" 26589) ("|" 26590)) 12326 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26659) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12327 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26658) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12328 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26657) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12329 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26656) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12330 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26655) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12331 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26654) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12332 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26653) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12333 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26652) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12334 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 26647) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 26648) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 12335 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26646) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12336 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26645) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12337 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26644) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12338 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26643) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12339 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26642) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12340 (("$" 26607) ("(" 26608) (T_NAME_FULLY_QUALIFIED 26609) (T_NAME_QUALIFIED 26610) (T_NAME_RELATIVE 26611) (T_STATIC 26612) (T_STRING 26613) (T_VARIABLE 26614) (class_name 26615) (class_name_reference 26616) (name 26617) (new_variable 26618) (simple_variable 26619)) 12341 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26606) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12342 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26605) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12343 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26604) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12344 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26603) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12345 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26602) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12346 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26601) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12347 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26600) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12348 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26599) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12349 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26598) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12350 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26597) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12351 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26596) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12352 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26595) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12353 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26594) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12354 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26593) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12355 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26592) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12356 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26591) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12357 (("$" 26607) (T_VARIABLE 26614) (simple_variable 26638) ("{" 26639)) 12358 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 26636) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 12359 ((T_PAAMAYIM_NEKUDOTAYIM 26634)) 12360 ((T_NULLSAFE_OBJECT_OPERATOR 26620) (T_OBJECT_OPERATOR 26621) (T_PAAMAYIM_NEKUDOTAYIM 26622) ("[" 26623)) 12361 (("$" 26607) (T_STRING 26627) (T_VARIABLE 26614) (property_name 26633) (simple_variable 26629) ("{" 26630)) 12362 (("$" 26607) (T_STRING 26627) (T_VARIABLE 26614) (property_name 26628) (simple_variable 26629) ("{" 26630)) 12363 (("$" 26607) (T_VARIABLE 26614) (simple_variable 26626)) 12364 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 26624) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 12365 (("]" 26625)) 12366 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 26631) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 12367 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 26632)) 12368 (("$" 26607) (T_VARIABLE 26614) (simple_variable 26635)) 12369 (("%" 304) (")" 26637) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 12370 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 26640) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 12371 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 26641)) 12372 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26651) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12373 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 26649) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 12374 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26650) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12375 (("$" 26482) (T_ABSTRACT 26661) (T_ARRAY 26662) (T_AS 26663) (T_BREAK 26664) (T_CALLABLE 26665) (T_CASE 26666) (T_CATCH 26667) (T_CLASS 26668) (T_CLASS_C 26669) (T_CLONE 26670) (T_CONST 26671) (T_CONTINUE 26672) (T_DECLARE 26673) (T_DEFAULT 26674) (T_DIR 26675) (T_DO 26676) (T_ECHO 26677) (T_ELSE 26678) (T_ELSEIF 26679) (T_EMPTY 26680) (T_ENDDECLARE 26681) (T_ENDFOR 26682) (T_ENDFOREACH 26683) (T_ENDIF 26684) (T_ENDSWITCH 26685) (T_ENDWHILE 26686) (T_ENUM 26687) (T_EVAL 26688) (T_EXIT 26689) (T_EXTENDS 26690) (T_FILE 26691) (T_FINAL 26692) (T_FINALLY 26693) (T_FN 26694) (T_FOR 26695) (T_FOREACH 26696) (T_FUNCTION 26697) (T_FUNC_C 26698) (T_GLOBAL 26699) (T_GOTO 26700) (T_IF 26701) (T_IMPLEMENTS 26702) (T_INCLUDE 26703) (T_INCLUDE_ONCE 26704) (T_INSTANCEOF 26705) (T_INSTEADOF 26706) (T_INTERFACE 26707) (T_ISSET 26708) (T_LINE 26709) (T_LIST 26710) (T_LOGICAL_AND 26711) (T_LOGICAL_OR 26712) (T_LOGICAL_XOR 26713) (T_MATCH 26714) (T_METHOD_C 26715) (T_NAMESPACE 26716) (T_NEW 26717) (T_NS_C 26718) (T_PRINT 26719) (T_PRIVATE 26720) (T_PROPERTY_C 26721) (T_PROTECTED 26722) (T_PUBLIC 26723) (T_READONLY 26724) (T_REQUIRE 26725) (T_REQUIRE_ONCE 26726) (T_RETURN 26727) (T_STATIC 26728) (T_STRING 26729) (T_SWITCH 26730) (T_THROW 26731) (T_TRAIT 26732) (T_TRAIT_C 26733) (T_TRY 26734) (T_UNSET 26735) (T_USE 26736) (T_VAR 26737) (T_VARIABLE 26530) (T_WHILE 26738) (T_YIELD 26739) (identifier 26740) (member_name 26741) (reserved_non_modifiers 26742) (semi_reserved 26743) (simple_variable 26744) ("{" 26745)) 12376 (("(" 26748) (argument_list 26749)) 12377 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 26746) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 12378 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 26747)) 12379 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 26750) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 26751) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 26752) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 12380 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 26755) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 12381 (("," 1498) (possible_comma 26753)) 12382 ((")" 26754)) 12383 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (ampersand 26785) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26786) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12384 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26784) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12385 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26783) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12386 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26782) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12387 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26781) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12388 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26780) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12389 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26779) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12390 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26778) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12391 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26777) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12392 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26776) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12393 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26775) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12394 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26774) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12395 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26773) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12396 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26772) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12397 (("\"" 1538) ("$" 26787) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 26788) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 26789) ("[" 1560) (array_object_dereferenceable 26790) (callable_expr 26791) (callable_variable 26792) (class_constant 1564) (class_name 26793) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 26794) (name 26795) (new_dereferenceable 1570) (simple_variable 26796) (static_member 26797) (variable 26798) (variable_class_name 26799)) 12398 (("$" 26787) (T_VARIABLE 26789) (simple_variable 26832) ("{" 26833)) 12399 (("(" 26803) (argument_list 26831)) 12400 ((T_NULLSAFE_OBJECT_OPERATOR 26817) (T_OBJECT_OPERATOR 26818) ("[" 26819)) 12401 (("(" 26803) (argument_list 26816)) 12402 ((T_PAAMAYIM_NEKUDOTAYIM 26812)) 12403 (("(" 26803) (argument_list 26811)) 12404 ((T_PAAMAYIM_NEKUDOTAYIM 26800)) 12405 (("$" 26787) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 26789) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 26801) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 26802) ("{" 1660)) 12406 (("(" 26803) (argument_list 26804)) 12407 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 26805) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 26806) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 26807) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 12408 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 26810) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 12409 (("," 1498) (possible_comma 26808)) 12410 ((")" 26809)) 12411 (("$" 26787) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 26789) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 26813) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 26814) ("{" 1676)) 12412 (("(" 26803) (argument_list 26815)) 12413 (("$" 26787) (T_STRING 26822) (T_VARIABLE 26789) (property_name 26829) (simple_variable 26824) ("{" 26825)) 12414 (("$" 26787) (T_STRING 26822) (T_VARIABLE 26789) (property_name 26823) (simple_variable 26824) ("{" 26825)) 12415 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 26820) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 12416 (("]" 26821)) 12417 (("(" 26803) (argument_list 26828)) 12418 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 26826) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 12419 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 26827)) 12420 (("(" 26803) (argument_list 26830)) 12421 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 26834) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 12422 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 26835)) 12423 ((backup_doc_comment 26838)) 12424 (("(" 26839)) 12425 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 26840)) 12426 ((")" 26841)) 12427 ((T_USE 166) (lexical_vars 26842)) 12428 ((":" 168) (return_type 26843)) 12429 ((backup_fn_flags 26844)) 12430 (("{" 26845)) 12431 ((inner_statement_list 26846)) 12432 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 26847) ("~" 129)) 12433 ((backup_fn_flags 26848)) 12434 ((backup_doc_comment 26850)) 12435 (("(" 26851)) 12436 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 26852)) 12437 ((")" 26853)) 12438 ((":" 2271) (return_type 26854)) 12439 ((T_DOUBLE_ARROW 26855)) 12440 ((backup_fn_flags 26856)) 12441 ((backup_lex_pos 26857)) 12442 (("!" 25793) ("\"" 25794) ("$" 25795) ("(" 25796) ("+" 25797) ("-" 25798) ("@" 25799) (T_ARRAY 25800) (T_ARRAY_CAST 25801) (T_ATTRIBUTE 130) (T_BOOL_CAST 25802) (T_CLASS_C 25803) (T_CLONE 25804) (T_CONSTANT_ENCAPSED_STRING 25805) (T_DEC 25806) (T_DIR 25807) (T_DNUMBER 25808) (T_DOUBLE_CAST 25809) (T_EMPTY 25810) (T_EVAL 25811) (T_EXIT 25812) (T_FILE 25813) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25814) (T_INC 25815) (T_INCLUDE 25816) (T_INCLUDE_ONCE 25817) (T_INT_CAST 25818) (T_ISSET 25819) (T_LINE 25820) (T_LIST 25821) (T_LNUMBER 25822) (T_MATCH 25823) (T_METHOD_C 25824) (T_NAME_FULLY_QUALIFIED 25825) (T_NAME_QUALIFIED 25826) (T_NAME_RELATIVE 25827) (T_NEW 25828) (T_NS_C 25829) (T_OBJECT_CAST 25830) (T_PRINT 25831) (T_PROPERTY_C 25832) (T_READONLY 25833) (T_REQUIRE 25834) (T_REQUIRE_ONCE 25835) (T_START_HEREDOC 25836) (T_STATIC 25837) (T_STRING 25838) (T_STRING_CAST 25839) (T_THROW 25840) (T_TRAIT_C 25841) (T_UNSET_CAST 25842) (T_VARIABLE 25843) (T_YIELD 25844) (T_YIELD_FROM 25845) ("[" 25846) ("`" 25847) (array_object_dereferenceable 25848) (attribute 135) (attributes 25849) (callable_expr 25850) (callable_variable 25851) (class_constant 25852) (class_name 25853) (constant 25854) (dereferenceable_scalar 25855) (expr 26858) (fn 25857) (fully_dereferenceable 107) #'25858 (function_call 25859) (inline_function 25860) (internal_functions_in_yacc 25861) (match 25862) (name 25863) (new_dereferenceable 25864) (new_non_dereferenceable 25865) (scalar 25866) (simple_variable 25867) (static_member 25868) (variable 25869) (variable_class_name 25870) ("~" 25871)) 12443 (("%" 25873) ("*" 25874) ("+" 25875) ("-" 25876) ("." 25877) ("/" 25878) ("<" 25879) (">" 25880) ("?" 25881) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 25882) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 25883) (T_BOOLEAN_AND 25884) (T_BOOLEAN_OR 25885) (T_COALESCE 25886) (T_INSTANCEOF 25887) (T_IS_EQUAL 25888) (T_IS_GREATER_OR_EQUAL 25889) (T_IS_IDENTICAL 25890) (T_IS_NOT_EQUAL 25891) (T_IS_NOT_IDENTICAL 25892) (T_IS_SMALLER_OR_EQUAL 25893) (T_LOGICAL_AND 25894) (T_LOGICAL_OR 25895) (T_LOGICAL_XOR 25896) (T_PIPE 25897) (T_POW 25898) (T_SL 25899) (T_SPACESHIP 25900) (T_SR 25901) ("^" 25902) (backup_fn_flags 26859) ("|" 25903)) 12444 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 26861) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 12445 (("$" 26482) (T_ABSTRACT 26661) (T_ARRAY 26662) (T_AS 26663) (T_BREAK 26664) (T_CALLABLE 26665) (T_CASE 26666) (T_CATCH 26667) (T_CLASS 26668) (T_CLASS_C 26669) (T_CLONE 26670) (T_CONST 26671) (T_CONTINUE 26672) (T_DECLARE 26673) (T_DEFAULT 26674) (T_DIR 26675) (T_DO 26676) (T_ECHO 26677) (T_ELSE 26678) (T_ELSEIF 26679) (T_EMPTY 26680) (T_ENDDECLARE 26681) (T_ENDFOR 26682) (T_ENDFOREACH 26683) (T_ENDIF 26684) (T_ENDSWITCH 26685) (T_ENDWHILE 26686) (T_ENUM 26687) (T_EVAL 26688) (T_EXIT 26689) (T_EXTENDS 26690) (T_FILE 26691) (T_FINAL 26692) (T_FINALLY 26693) (T_FN 26694) (T_FOR 26695) (T_FOREACH 26696) (T_FUNCTION 26697) (T_FUNC_C 26698) (T_GLOBAL 26699) (T_GOTO 26700) (T_IF 26701) (T_IMPLEMENTS 26702) (T_INCLUDE 26703) (T_INCLUDE_ONCE 26704) (T_INSTANCEOF 26705) (T_INSTEADOF 26706) (T_INTERFACE 26707) (T_ISSET 26708) (T_LINE 26709) (T_LIST 26710) (T_LOGICAL_AND 26711) (T_LOGICAL_OR 26712) (T_LOGICAL_XOR 26713) (T_MATCH 26714) (T_METHOD_C 26715) (T_NAMESPACE 26716) (T_NEW 26717) (T_NS_C 26718) (T_PRINT 26719) (T_PRIVATE 26720) (T_PROPERTY_C 26721) (T_PROTECTED 26722) (T_PUBLIC 26723) (T_READONLY 26724) (T_REQUIRE 26725) (T_REQUIRE_ONCE 26726) (T_RETURN 26727) (T_STATIC 26728) (T_STRING 26729) (T_SWITCH 26730) (T_THROW 26731) (T_TRAIT 26732) (T_TRAIT_C 26733) (T_TRY 26734) (T_UNSET 26735) (T_USE 26736) (T_VAR 26737) (T_VARIABLE 26530) (T_WHILE 26738) (T_YIELD 26739) (identifier 26863) (member_name 26864) (reserved_non_modifiers 26742) (semi_reserved 26743) (simple_variable 26865) ("{" 26866)) 12446 (("(" 26748) (argument_list 26869)) 12447 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 26867) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 12448 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 26868)) 12449 ((T_FN 35) (T_FUNCTION 131) (fn 26544) #'26545 (inline_function 26873)) 12450 (("$" 26482) (T_STRING 26879) (T_VARIABLE 26530) (property_name 26886) (simple_variable 26881) ("{" 26882)) 12451 (("$" 26482) (T_STRING 26879) (T_VARIABLE 26530) (property_name 26880) (simple_variable 26881) ("{" 26882)) 12452 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 26877) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 12453 (("]" 26878)) 12454 (("(" 26748) (argument_list 26885)) 12455 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 26883) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 12456 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 26884)) 12457 (("(" 26748) (argument_list 26887)) 12458 (("`" 26889)) 12459 (("]" 26891)) 12460 (("=" 26892)) 12461 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26893) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12462 (("%" 26560) ("*" 26561) ("+" 26562) ("-" 26563) ("." 26564) ("/" 26565) ("<" 26566) (">" 26567) ("?" 26568) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 26569) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 26570) (T_BOOLEAN_AND 26571) (T_BOOLEAN_OR 26572) (T_COALESCE 26573) (T_DOUBLE_ARROW 26896) (T_INSTANCEOF 26574) (T_IS_EQUAL 26575) (T_IS_GREATER_OR_EQUAL 26576) (T_IS_IDENTICAL 26577) (T_IS_NOT_EQUAL 26578) (T_IS_NOT_IDENTICAL 26579) (T_IS_SMALLER_OR_EQUAL 26580) (T_LOGICAL_AND 26581) (T_LOGICAL_OR 26582) (T_LOGICAL_XOR 26583) (T_PIPE 26584) (T_POW 26585) (T_SL 26586) (T_SPACESHIP 26587) (T_SR 26588) ("^" 26589) ("|" 26590)) 12463 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26897) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12464 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 26906) (T_VARIABLE 2920) (encaps_var 2934)) 12465 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 26905) (T_VARIABLE 2920) (encaps_var 2925)) 12466 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 26947) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 12467 ((T_CLASS 26939)) 12468 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 26938) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 26919) (attribute 2982) (class_modifier 2963)) 12469 ((T_PAAMAYIM_NEKUDOTAYIM 26936)) 12470 (("(" 26803) (argument_list 26935)) 12471 ((T_NULLSAFE_OBJECT_OPERATOR 26926) (T_OBJECT_OPERATOR 26927) (T_PAAMAYIM_NEKUDOTAYIM 26928) ("[" 26929)) 12472 (("$" 26787) (T_STRING 26822) (T_VARIABLE 26789) (property_name 26934) (simple_variable 26824) ("{" 26825)) 12473 (("$" 26787) (T_STRING 26822) (T_VARIABLE 26789) (property_name 26933) (simple_variable 26824) ("{" 26825)) 12474 (("$" 26787) (T_VARIABLE 26789) (simple_variable 26932)) 12475 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 26930) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 12476 (("]" 26931)) 12477 (("$" 26787) (T_VARIABLE 26789) (simple_variable 26937)) 12478 (("(" 2984) (argument_list 2985) (ctor_arguments 26940)) 12479 ((T_EXTENDS 2987) (extends_from 26941)) 12480 ((T_IMPLEMENTS 2989) (implements_list 26942)) 12481 ((backup_doc_comment 26943)) 12482 (("{" 26944)) 12483 ((class_statement_list 26945)) 12484 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 26946)) 12485 (("%" 304) (")" 26948) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 12486 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 26950) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 12487 (("%" 304) (")" 26951) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 12488 (("{" 26952)) 12489 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 26953) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 12490 (("}" 26954)) 12491 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 26956) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 12492 ((")" 26957)) 12493 (("=" 26958)) 12494 (("!" 26480) ("\"" 26481) ("$" 26482) ("(" 26483) ("+" 26484) ("-" 26485) ("@" 26486) (T_ARRAY 26487) (T_ARRAY_CAST 26488) (T_ATTRIBUTE 130) (T_BOOL_CAST 26489) (T_CLASS_C 26490) (T_CLONE 26491) (T_CONSTANT_ENCAPSED_STRING 26492) (T_DEC 26493) (T_DIR 26494) (T_DNUMBER 26495) (T_DOUBLE_CAST 26496) (T_EMPTY 26497) (T_EVAL 26498) (T_EXIT 26499) (T_FILE 26500) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 26501) (T_INC 26502) (T_INCLUDE 26503) (T_INCLUDE_ONCE 26504) (T_INT_CAST 26505) (T_ISSET 26506) (T_LINE 26507) (T_LIST 26508) (T_LNUMBER 26509) (T_MATCH 26510) (T_METHOD_C 26511) (T_NAME_FULLY_QUALIFIED 26512) (T_NAME_QUALIFIED 26513) (T_NAME_RELATIVE 26514) (T_NEW 26515) (T_NS_C 26516) (T_OBJECT_CAST 26517) (T_PRINT 26518) (T_PROPERTY_C 26519) (T_READONLY 26520) (T_REQUIRE 26521) (T_REQUIRE_ONCE 26522) (T_START_HEREDOC 26523) (T_STATIC 26524) (T_STRING 26525) (T_STRING_CAST 26526) (T_THROW 26527) (T_TRAIT_C 26528) (T_UNSET_CAST 26529) (T_VARIABLE 26530) (T_YIELD 26531) (T_YIELD_FROM 26532) ("[" 26533) ("`" 26534) (array_object_dereferenceable 26535) (attribute 135) (attributes 26536) (callable_expr 26537) (callable_variable 26538) (class_constant 26539) (class_name 26540) (constant 26541) (dereferenceable_scalar 26542) (expr 26959) (fn 26544) (fully_dereferenceable 107) #'26545 (function_call 26546) (inline_function 26547) (internal_functions_in_yacc 26548) (match 26549) (name 26550) (new_dereferenceable 26551) (new_non_dereferenceable 26552) (scalar 26553) (simple_variable 26554) (static_member 26555) (variable 26556) (variable_class_name 26557) ("~" 26558)) 12495 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 26961) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 12496 (("," 5226) (possible_comma 26962)) 12497 ((")" 26963)) 12498 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 26971) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 26972) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 26973) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 12499 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 26976) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 12500 (("," 1498) (possible_comma 26974)) 12501 ((")" 26975)) 12502 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 26978) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 12503 (("%" 304) (")" 26979) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 12504 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 26981) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 12505 (("%" 304) (")" 26982) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 12506 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 26988) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 26989) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 26990) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 26991) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 12507 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 26997) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 12508 (("%" 2442) (")" 26994) ("*" 2443) ("+" 2444) ("," 26995) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 12509 (("," 5287) (possible_comma 26992)) 12510 ((")" 26993)) 12511 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 26996) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 12512 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 27001) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 12513 ((")" 27002)) 12514 (("%" 304) (")" 26994) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 12515 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 27009) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 12516 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 27010)) 12517 (("\"" 27012) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 12518 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 27022) (T_VARIABLE 2920) (encaps_var 2934)) 12519 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 27021) (T_VARIABLE 2920) (encaps_var 2925)) 12520 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 27063) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 12521 ((T_CLASS 27055)) 12522 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 27054) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 27035) (attribute 2982) (class_modifier 2963)) 12523 ((T_PAAMAYIM_NEKUDOTAYIM 27052)) 12524 (("(" 25324) (argument_list 27051)) 12525 ((T_NULLSAFE_OBJECT_OPERATOR 27042) (T_OBJECT_OPERATOR 27043) (T_PAAMAYIM_NEKUDOTAYIM 27044) ("[" 27045)) 12526 (("$" 25308) (T_STRING 25343) (T_VARIABLE 25310) (property_name 27050) (simple_variable 25345) ("{" 25346)) 12527 (("$" 25308) (T_STRING 25343) (T_VARIABLE 25310) (property_name 27049) (simple_variable 25345) ("{" 25346)) 12528 (("$" 25308) (T_VARIABLE 25310) (simple_variable 27048)) 12529 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 27046) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 12530 (("]" 27047)) 12531 (("$" 25308) (T_VARIABLE 25310) (simple_variable 27053)) 12532 (("(" 2984) (argument_list 2985) (ctor_arguments 27056)) 12533 ((T_EXTENDS 2987) (extends_from 27057)) 12534 ((T_IMPLEMENTS 2989) (implements_list 27058)) 12535 ((backup_doc_comment 27059)) 12536 (("{" 27060)) 12537 ((class_statement_list 27061)) 12538 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 27062)) 12539 (("%" 304) (")" 27064) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 12540 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 27066) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 12541 (("%" 304) (")" 27067) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 12542 (("{" 27068)) 12543 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 27069) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 12544 (("}" 27070)) 12545 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 27072) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 12546 ((")" 27073)) 12547 (("=" 27074)) 12548 (("!" 25001) ("\"" 25002) ("$" 25003) ("(" 25004) ("+" 25005) ("-" 25006) ("@" 25007) (T_ARRAY 25008) (T_ARRAY_CAST 25009) (T_ATTRIBUTE 130) (T_BOOL_CAST 25010) (T_CLASS_C 25011) (T_CLONE 25012) (T_CONSTANT_ENCAPSED_STRING 25013) (T_DEC 25014) (T_DIR 25015) (T_DNUMBER 25016) (T_DOUBLE_CAST 25017) (T_EMPTY 25018) (T_EVAL 25019) (T_EXIT 25020) (T_FILE 25021) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 25022) (T_INC 25023) (T_INCLUDE 25024) (T_INCLUDE_ONCE 25025) (T_INT_CAST 25026) (T_ISSET 25027) (T_LINE 25028) (T_LIST 25029) (T_LNUMBER 25030) (T_MATCH 25031) (T_METHOD_C 25032) (T_NAME_FULLY_QUALIFIED 25033) (T_NAME_QUALIFIED 25034) (T_NAME_RELATIVE 25035) (T_NEW 25036) (T_NS_C 25037) (T_OBJECT_CAST 25038) (T_PRINT 25039) (T_PROPERTY_C 25040) (T_READONLY 25041) (T_REQUIRE 25042) (T_REQUIRE_ONCE 25043) (T_START_HEREDOC 25044) (T_STATIC 25045) (T_STRING 25046) (T_STRING_CAST 25047) (T_THROW 25048) (T_TRAIT_C 25049) (T_UNSET_CAST 25050) (T_VARIABLE 25051) (T_YIELD 25052) (T_YIELD_FROM 25053) ("[" 25054) ("`" 25055) (array_object_dereferenceable 25056) (attribute 135) (attributes 25057) (callable_expr 25058) (callable_variable 25059) (class_constant 25060) (class_name 25061) (constant 25062) (dereferenceable_scalar 25063) (expr 27075) (fn 25065) (fully_dereferenceable 107) #'25066 (function_call 25067) (inline_function 25068) (internal_functions_in_yacc 25069) (match 25070) (name 25071) (new_dereferenceable 25072) (new_non_dereferenceable 25073) (scalar 25074) (simple_variable 25075) (static_member 25076) (variable 25077) (variable_class_name 25078) ("~" 25079)) 12549 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 27077) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 12550 (("," 5226) (possible_comma 27078)) 12551 ((")" 27079)) 12552 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 27087) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 27088) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 27089) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 12553 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 27092) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 12554 (("," 1498) (possible_comma 27090)) 12555 ((")" 27091)) 12556 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 27094) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 12557 (("%" 304) (")" 27095) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 12558 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 27097) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 12559 (("%" 304) (")" 27098) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 12560 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 27104) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 27105) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 27106) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 27107) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 12561 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 27113) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 12562 (("%" 2442) (")" 27110) ("*" 2443) ("+" 2444) ("," 27111) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 12563 (("," 5287) (possible_comma 27108)) 12564 ((")" 27109)) 12565 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 27112) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 12566 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 27117) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 12567 ((")" 27118)) 12568 (("%" 304) (")" 27110) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 12569 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 27125) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 12570 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 27126)) 12571 (("\"" 27128) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 12572 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 27131)) 12573 ((";" 27132)) 12574 (("," 18369) (";" 27135)) 12575 ((";" 27137)) 12576 ((";" 27139)) 12577 (("," 19446) (";" 27141)) 12578 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 27143) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 12579 (("%" 19534) ("*" 19535) ("+" 19536) ("-" 19537) ("." 19538) ("/" 19539) ("<" 19540) (">" 19541) ("?" 19542) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19543) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19544) (T_AS 27144) (T_BOOLEAN_AND 19545) (T_BOOLEAN_OR 19546) (T_COALESCE 19547) (T_INSTANCEOF 19548) (T_IS_EQUAL 19549) (T_IS_GREATER_OR_EQUAL 19550) (T_IS_IDENTICAL 19551) (T_IS_NOT_EQUAL 19552) (T_IS_NOT_IDENTICAL 19553) (T_IS_SMALLER_OR_EQUAL 19554) (T_LOGICAL_AND 19555) (T_LOGICAL_OR 19556) (T_LOGICAL_XOR 19557) (T_PIPE 19558) (T_POW 19559) (T_SL 19560) (T_SPACESHIP 19561) (T_SR 19562) ("^" 19563) ("|" 19564)) 12580 (("\"" 1538) ("$" 20897) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20898) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20899) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20900) ("[" 20901) (ampersand 20902) (array_object_dereferenceable 20903) (callable_expr 20904) (callable_variable 20905) (class_constant 1564) (class_name 20906) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 27145) (fully_dereferenceable 107) (function_call 20908) (name 20909) (new_dereferenceable 1570) (simple_variable 20910) (static_member 20911) (variable 20912) (variable_class_name 20913)) 12581 ((")" 27146) (T_DOUBLE_ARROW 27147)) 12582 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27150) (";" 18135) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 18136) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18137) (T_DEC 22) (T_DECLARE 18138) (T_DIR 24) (T_DNUMBER 25) (T_DO 18139) (T_DOUBLE_CAST 27) (T_ECHO 18140) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18141) (T_FOREACH 18142) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 18143) (T_GOTO 18144) (T_IF 18099) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18145) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 18146) (T_START_HEREDOC 69) (T_STATIC 18147) (T_STRING 18148) (T_STRING_CAST 72) (T_SWITCH 18149) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 18150) (T_UNSET 18151) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 18152) (T_WHILE 18153) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 18154) (alt_if_stmt_without_else 18155) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 18156) (fn 106) (foreach_statement 27156) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 18157) (if_stmt_without_else 18158) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27152) (static_member 123) (variable 126) (variable_class_name 127) ("{" 18160) ("~" 129)) 12583 (("\"" 1538) ("$" 20928) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20929) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20930) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20931) ("[" 20932) (ampersand 20933) (array_object_dereferenceable 20934) (callable_expr 20935) (callable_variable 20936) (class_constant 1564) (class_name 20937) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 27148) (fully_dereferenceable 107) (function_call 20939) (name 20940) (new_dereferenceable 1570) (simple_variable 20941) (static_member 20942) (variable 20943) (variable_class_name 20944)) 12584 ((")" 27149)) 12585 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27150) (";" 18135) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 18136) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18137) (T_DEC 22) (T_DECLARE 18138) (T_DIR 24) (T_DNUMBER 25) (T_DO 18139) (T_DOUBLE_CAST 27) (T_ECHO 18140) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18141) (T_FOREACH 18142) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 18143) (T_GOTO 18144) (T_IF 18099) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18145) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 18146) (T_START_HEREDOC 69) (T_STATIC 18147) (T_STRING 18148) (T_STRING_CAST 72) (T_SWITCH 18149) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 18150) (T_UNSET 18151) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 18152) (T_WHILE 18153) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 18154) (alt_if_stmt_without_else 18155) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 18156) (fn 106) (foreach_statement 27151) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 18157) (if_stmt_without_else 18158) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27152) (static_member 123) (variable 126) (variable_class_name 127) ("{" 18160) ("~" 129)) 12586 ((inner_statement_list 27153)) 12587 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20962) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 20963) (T_CLASS 20964) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20965) (T_DEC 22) (T_DECLARE 20966) (T_DIR 24) (T_DNUMBER 25) (T_DO 20967) (T_DOUBLE_CAST 27) (T_ECHO 20968) (T_EMPTY 29) (T_ENDFOREACH 27154) (T_ENUM 20970) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20971) (T_FOREACH 20972) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20973) (T_GOTO 20974) (T_HALT_COMPILER 20975) (T_IF 20976) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20977) (T_INTERFACE 20978) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20979) (T_START_HEREDOC 69) (T_STATIC 20980) (T_STRING 20981) (T_STRING_CAST 72) (T_SWITCH 20982) (T_THROW 74) (T_TRAIT 20983) (T_TRAIT_C 76) (T_TRY 20984) (T_UNSET 20985) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 20986) (T_WHILE 20987) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 20988) (alt_if_stmt_without_else 20989) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 20990) (attributes 20991) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 20992) (class_modifier 99) (class_modifiers 20993) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 20994) (expr 20995) (fn 106) (fully_dereferenceable 107) #'20996 (function_call 109) (function_declaration_statement 20997) (if_stmt 20998) (if_stmt_without_else 20999) (inline_function 113) (inner_statement 21000) (interface_declaration_statement 21001) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 21002) (static_member 123) (trait_declaration_statement 21003) (variable 126) (variable_class_name 127) ("{" 21004) ("~" 129)) 12588 ((";" 27155)) 12589 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21126) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21127) (fn 18437) (for_exprs 27158) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21129) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 12590 ((";" 27159)) 12591 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21135) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21136) (fn 18437) (for_cond_exprs 27160) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21138) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 12592 ((";" 27161)) 12593 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_VOID_CAST 22737) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 22738) (fn 2425) (for_exprs 27162) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_for_exprs 22740) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 12594 ((")" 27163)) 12595 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27164) (";" 18135) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 18136) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18137) (T_DEC 22) (T_DECLARE 18138) (T_DIR 24) (T_DNUMBER 25) (T_DO 18139) (T_DOUBLE_CAST 27) (T_ECHO 18140) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18141) (T_FOREACH 18142) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 18143) (T_GOTO 18144) (T_IF 18099) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18145) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 18146) (T_START_HEREDOC 69) (T_STATIC 18147) (T_STRING 18148) (T_STRING_CAST 72) (T_SWITCH 18149) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 18150) (T_UNSET 18151) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 18152) (T_WHILE 18153) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 18154) (alt_if_stmt_without_else 18155) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 18156) (fn 106) (for_statement 27165) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 18157) (if_stmt_without_else 18158) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27166) (static_member 123) (variable 126) (variable_class_name 127) ("{" 18160) ("~" 129)) 12596 ((inner_statement_list 27167)) 12597 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22750) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 22751) (T_CLASS 22752) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22753) (T_DEC 22) (T_DECLARE 22754) (T_DIR 24) (T_DNUMBER 25) (T_DO 22755) (T_DOUBLE_CAST 27) (T_ECHO 22756) (T_EMPTY 29) (T_ENDFOR 27168) (T_ENUM 22758) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 22759) (T_FOREACH 22760) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 22761) (T_GOTO 22762) (T_HALT_COMPILER 22763) (T_IF 22764) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22765) (T_INTERFACE 22766) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22767) (T_START_HEREDOC 69) (T_STATIC 22768) (T_STRING 22769) (T_STRING_CAST 72) (T_SWITCH 22770) (T_THROW 74) (T_TRAIT 22771) (T_TRAIT_C 76) (T_TRY 22772) (T_UNSET 22773) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22774) (T_WHILE 22775) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22776) (alt_if_stmt_without_else 22777) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 22778) (attributes 22779) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 22780) (class_modifier 99) (class_modifiers 22781) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 22782) (expr 22783) (fn 106) (fully_dereferenceable 107) #'22784 (function_call 109) (function_declaration_statement 22785) (if_stmt 22786) (if_stmt_without_else 22787) (inline_function 113) (inner_statement 22788) (interface_declaration_statement 22789) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 22790) (static_member 123) (trait_declaration_statement 22791) (variable 126) (variable_class_name 127) ("{" 22792) ("~" 129)) 12598 ((";" 27169)) 12599 (("," 22929) (";" 27171)) 12600 ((T_WHILE 27173)) 12601 (("(" 27174)) 12602 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 27175) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 12603 (("%" 304) (")" 27176) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 12604 ((";" 27177)) 12605 ((T_STRING 23106) (const_decl 23107) (const_list 27179)) 12606 ((")" 27180) ("," 23110)) 12607 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27181) (";" 18135) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 18136) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18137) (T_DEC 22) (T_DECLARE 18138) (T_DIR 24) (T_DNUMBER 25) (T_DO 18139) (T_DOUBLE_CAST 27) (T_ECHO 18140) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18141) (T_FOREACH 18142) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 18143) (T_GOTO 18144) (T_IF 18099) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18145) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 18146) (T_START_HEREDOC 69) (T_STATIC 18147) (T_STRING 18148) (T_STRING_CAST 72) (T_SWITCH 18149) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 18150) (T_UNSET 18151) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 18152) (T_WHILE 18153) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 18154) (alt_if_stmt_without_else 18155) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (declare_statement 27182) (dereferenceable_scalar 103) (expr 18156) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 18157) (if_stmt_without_else 18158) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27183) (static_member 123) (variable 126) (variable_class_name 127) ("{" 18160) ("~" 129)) 12608 ((inner_statement_list 27184)) 12609 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 23116) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 23117) (T_CLASS 23118) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 23119) (T_DEC 22) (T_DECLARE 23120) (T_DIR 24) (T_DNUMBER 25) (T_DO 23121) (T_DOUBLE_CAST 27) (T_ECHO 23122) (T_EMPTY 29) (T_ENDDECLARE 27185) (T_ENUM 23124) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 23125) (T_FOREACH 23126) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 23127) (T_GOTO 23128) (T_HALT_COMPILER 23129) (T_IF 23130) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 23131) (T_INTERFACE 23132) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 23133) (T_START_HEREDOC 69) (T_STATIC 23134) (T_STRING 23135) (T_STRING_CAST 72) (T_SWITCH 23136) (T_THROW 74) (T_TRAIT 23137) (T_TRAIT_C 76) (T_TRY 23138) (T_UNSET 23139) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 23140) (T_WHILE 23141) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 23142) (alt_if_stmt_without_else 23143) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 23144) (attributes 23145) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 23146) (class_modifier 99) (class_modifiers 23147) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 23148) (expr 23149) (fn 106) (fully_dereferenceable 107) #'23150 (function_call 109) (function_declaration_statement 23151) (if_stmt 23152) (if_stmt_without_else 23153) (inline_function 113) (inner_statement 23154) (interface_declaration_statement 23155) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23156) (static_member 123) (trait_declaration_statement 23157) (variable 126) (variable_class_name 127) ("{" 23158) ("~" 129)) 12610 ((";" 27186)) 12611 ((";" 27188)) 12612 ((";" 27190)) 12613 ((T_READONLY 17774) (T_STRING 17775) (backup_doc_comment 143) (function_name 27193)) 12614 ((backup_doc_comment 27194)) 12615 (("(" 27195)) 12616 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 27196)) 12617 ((")" 27197)) 12618 ((":" 168) (return_type 27198)) 12619 ((backup_fn_flags 27199)) 12620 (("{" 27200)) 12621 ((inner_statement_list 27201)) 12622 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 27202) ("~" 129)) 12623 ((backup_fn_flags 27203)) 12624 ((T_STRING 27206)) 12625 ((T_EXTENDS 2987) (extends_from 27207)) 12626 ((T_IMPLEMENTS 2989) (implements_list 27208)) 12627 ((backup_doc_comment 27209)) 12628 (("{" 27210)) 12629 ((class_statement_list 27211)) 12630 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 27212)) 12631 ((":" 27217)) 12632 ((";" 27216)) 12633 ((inner_statement_list 27218)) 12634 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17807) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17808) (T_CLASS 17809) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17810) (T_DEC 22) (T_DECLARE 17811) (T_DIR 24) (T_DNUMBER 25) (T_DO 17812) (T_DOUBLE_CAST 27) (T_ECHO 17813) (T_EMPTY 29) (T_ENDIF 27219) (T_ENUM 17815) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17816) (T_FOREACH 17817) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17818) (T_GOTO 17819) (T_HALT_COMPILER 17820) (T_IF 17732) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17821) (T_INTERFACE 17822) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17823) (T_START_HEREDOC 69) (T_STATIC 17824) (T_STRING 17825) (T_STRING_CAST 72) (T_SWITCH 17826) (T_THROW 74) (T_TRAIT 17827) (T_TRAIT_C 76) (T_TRY 17828) (T_UNSET 17829) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17830) (T_WHILE 17831) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17832) (alt_if_stmt_without_else 17833) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17834) (attributes 17835) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17836) (class_modifier 99) (class_modifiers 17837) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17838) (expr 17839) (fn 106) (fully_dereferenceable 107) #'17840 (function_call 109) (function_declaration_statement 17841) (if_stmt 17842) (if_stmt_without_else 17843) (inline_function 113) (inner_statement 17844) (interface_declaration_statement 17845) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17846) (static_member 123) (trait_declaration_statement 17847) (variable 126) (variable_class_name 127) ("{" 17848) ("~" 129)) 12635 ((";" 27220)) 12636 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 27222) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 12637 (("%" 304) (")" 27223) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 12638 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27224) (";" 18086) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 18087) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18089) (T_DEC 22) (T_DECLARE 18090) (T_DIR 24) (T_DNUMBER 25) (T_DO 18091) (T_DOUBLE_CAST 27) (T_ECHO 18092) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18094) (T_FOREACH 18095) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 18096) (T_GOTO 18097) (T_IF 18099) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18100) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 18102) (T_START_HEREDOC 69) (T_STATIC 18103) (T_STRING 18104) (T_STRING_CAST 72) (T_SWITCH 18105) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 18107) (T_UNSET 18108) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 18109) (T_WHILE 18110) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 18111) (alt_if_stmt_without_else 18112) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 18118) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 18121) (if_stmt_without_else 18122) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27225) (static_member 123) (variable 126) (variable_class_name 127) (while_statement 27226) ("{" 18127) ("~" 129)) 12639 ((inner_statement_list 27227)) 12640 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17889) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17890) (T_CLASS 17891) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17892) (T_DEC 22) (T_DECLARE 17893) (T_DIR 24) (T_DNUMBER 25) (T_DO 17894) (T_DOUBLE_CAST 27) (T_ECHO 17895) (T_EMPTY 29) (T_ENDWHILE 27228) (T_ENUM 17897) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17898) (T_FOREACH 17899) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17900) (T_GOTO 17901) (T_HALT_COMPILER 17902) (T_IF 17903) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17904) (T_INTERFACE 17905) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17906) (T_START_HEREDOC 69) (T_STATIC 17907) (T_STRING 17908) (T_STRING_CAST 72) (T_SWITCH 17909) (T_THROW 74) (T_TRAIT 17910) (T_TRAIT_C 76) (T_TRY 17911) (T_UNSET 17912) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17913) (T_WHILE 17914) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17915) (alt_if_stmt_without_else 17916) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17917) (attributes 17918) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17919) (class_modifier 99) (class_modifiers 17920) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17921) (expr 17922) (fn 106) (fully_dereferenceable 107) #'17923 (function_call 109) (function_declaration_statement 17924) (if_stmt 17925) (if_stmt_without_else 17926) (inline_function 113) (inner_statement 17927) (interface_declaration_statement 17928) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17929) (static_member 123) (trait_declaration_statement 17930) (variable 126) (variable_class_name 127) ("{" 17931) ("~" 129)) 12641 ((";" 27229)) 12642 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 27231) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 12643 (("\"" 1538) ("$" 17989) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 17990) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 17991) ("[" 1560) (array_object_dereferenceable 17992) (callable_expr 17993) (callable_variable 17994) (class_constant 1564) (class_name 17995) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 17996) (name 17997) (new_dereferenceable 1570) (simple_variable 17998) (static_member 17999) (unset_variable 18000) (unset_variables 27233) (variable 18002) (variable_class_name 18003)) 12644 (("," 18015) (possible_comma 27234)) 12645 ((")" 27235)) 12646 ((";" 27236)) 12647 ((inner_statement_list 27238)) 12648 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 27239) ("~" 129)) 12649 ((catch_list 27240)) 12650 ((T_CATCH 27241) (T_FINALLY 27242) (finally_statement 27243)) 12651 (("(" 27247)) 12652 (("{" 27244)) 12653 ((inner_statement_list 27245)) 12654 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 27246) ("~" 129)) 12655 ((T_NAME_FULLY_QUALIFIED 18056) (T_NAME_QUALIFIED 18057) (T_NAME_RELATIVE 18058) (T_STATIC 18059) (T_STRING 18060) (catch_name_list 27248) (class_name 18062) (name 18063)) 12656 ((T_VARIABLE 18064) (optional_variable 27249) ("|" 18066)) 12657 ((")" 27250)) 12658 (("{" 27251)) 12659 ((inner_statement_list 27252)) 12660 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 27253) ("~" 129)) 12661 ((backup_doc_comment 27255)) 12662 (("{" 27256)) 12663 ((class_statement_list 27257)) 12664 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 27258)) 12665 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 27260) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 12666 (("%" 304) (")" 27261) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 12667 ((":" 27262) (switch_case_list 27263) ("{" 27264)) 12668 ((";" 27270) (case_list 27271)) 12669 ((";" 27265) (case_list 27266)) 12670 ((case_list 27268)) 12671 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 27267)) 12672 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 27269)) 12673 ((case_list 27274)) 12674 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 27272)) 12675 ((";" 27273)) 12676 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 27275)) 12677 ((";" 27276)) 12678 (("," 18369) (";" 27279)) 12679 ((";" 27281)) 12680 ((T_EXTENDS 23405) (interface_extends_list 27283)) 12681 ((backup_doc_comment 27284)) 12682 (("{" 27285)) 12683 ((class_statement_list 27286)) 12684 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 27287)) 12685 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 27289) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 12686 (("%" 304) (")" 27290) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 12687 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23415) (";" 18135) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 18136) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18137) (T_DEC 22) (T_DECLARE 18138) (T_DIR 24) (T_DNUMBER 25) (T_DO 18139) (T_DOUBLE_CAST 27) (T_ECHO 18140) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18141) (T_FOREACH 18142) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 18143) (T_GOTO 18144) (T_IF 18099) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18145) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 18146) (T_START_HEREDOC 69) (T_STATIC 18147) (T_STRING 18148) (T_STRING_CAST 72) (T_SWITCH 18149) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 18150) (T_UNSET 18151) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 18152) (T_WHILE 18153) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 18154) (alt_if_stmt_without_else 18155) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 18156) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 18157) (if_stmt_without_else 18158) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27291) (static_member 123) (variable 126) (variable_class_name 127) ("{" 18160) ("~" 129)) 12688 ((")" 27293)) 12689 ((";" 27294)) 12690 ((";" 27296)) 12691 (("," 19446) (";" 27298)) 12692 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 27300) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 12693 (("%" 19534) ("*" 19535) ("+" 19536) ("-" 19537) ("." 19538) ("/" 19539) ("<" 19540) (">" 19541) ("?" 19542) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19543) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19544) (T_AS 27301) (T_BOOLEAN_AND 19545) (T_BOOLEAN_OR 19546) (T_COALESCE 19547) (T_INSTANCEOF 19548) (T_IS_EQUAL 19549) (T_IS_GREATER_OR_EQUAL 19550) (T_IS_IDENTICAL 19551) (T_IS_NOT_EQUAL 19552) (T_IS_NOT_IDENTICAL 19553) (T_IS_SMALLER_OR_EQUAL 19554) (T_LOGICAL_AND 19555) (T_LOGICAL_OR 19556) (T_LOGICAL_XOR 19557) (T_PIPE 19558) (T_POW 19559) (T_SL 19560) (T_SPACESHIP 19561) (T_SR 19562) ("^" 19563) ("|" 19564)) 12694 (("\"" 1538) ("$" 20897) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20898) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20899) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20900) ("[" 20901) (ampersand 20902) (array_object_dereferenceable 20903) (callable_expr 20904) (callable_variable 20905) (class_constant 1564) (class_name 20906) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 27302) (fully_dereferenceable 107) (function_call 20908) (name 20909) (new_dereferenceable 1570) (simple_variable 20910) (static_member 20911) (variable 20912) (variable_class_name 20913)) 12695 ((")" 27303) (T_DOUBLE_ARROW 27304)) 12696 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27307) (";" 18086) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 18087) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18089) (T_DEC 22) (T_DECLARE 18090) (T_DIR 24) (T_DNUMBER 25) (T_DO 18091) (T_DOUBLE_CAST 27) (T_ECHO 18092) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18094) (T_FOREACH 18095) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 18096) (T_GOTO 18097) (T_IF 18099) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18100) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 18102) (T_START_HEREDOC 69) (T_STATIC 18103) (T_STRING 18104) (T_STRING_CAST 72) (T_SWITCH 18105) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 18107) (T_UNSET 18108) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 18109) (T_WHILE 18110) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 18111) (alt_if_stmt_without_else 18112) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 18118) (fn 106) (foreach_statement 27313) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 18121) (if_stmt_without_else 18122) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27309) (static_member 123) (variable 126) (variable_class_name 127) ("{" 18127) ("~" 129)) 12697 (("\"" 1538) ("$" 20928) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20929) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20930) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20931) ("[" 20932) (ampersand 20933) (array_object_dereferenceable 20934) (callable_expr 20935) (callable_variable 20936) (class_constant 1564) (class_name 20937) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 27305) (fully_dereferenceable 107) (function_call 20939) (name 20940) (new_dereferenceable 1570) (simple_variable 20941) (static_member 20942) (variable 20943) (variable_class_name 20944)) 12698 ((")" 27306)) 12699 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27307) (";" 18086) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 18087) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18089) (T_DEC 22) (T_DECLARE 18090) (T_DIR 24) (T_DNUMBER 25) (T_DO 18091) (T_DOUBLE_CAST 27) (T_ECHO 18092) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18094) (T_FOREACH 18095) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 18096) (T_GOTO 18097) (T_IF 18099) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18100) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 18102) (T_START_HEREDOC 69) (T_STATIC 18103) (T_STRING 18104) (T_STRING_CAST 72) (T_SWITCH 18105) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 18107) (T_UNSET 18108) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 18109) (T_WHILE 18110) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 18111) (alt_if_stmt_without_else 18112) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 18118) (fn 106) (foreach_statement 27308) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 18121) (if_stmt_without_else 18122) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27309) (static_member 123) (variable 126) (variable_class_name 127) ("{" 18127) ("~" 129)) 12700 ((inner_statement_list 27310)) 12701 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20962) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 20963) (T_CLASS 20964) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20965) (T_DEC 22) (T_DECLARE 20966) (T_DIR 24) (T_DNUMBER 25) (T_DO 20967) (T_DOUBLE_CAST 27) (T_ECHO 20968) (T_EMPTY 29) (T_ENDFOREACH 27311) (T_ENUM 20970) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20971) (T_FOREACH 20972) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20973) (T_GOTO 20974) (T_HALT_COMPILER 20975) (T_IF 20976) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20977) (T_INTERFACE 20978) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20979) (T_START_HEREDOC 69) (T_STATIC 20980) (T_STRING 20981) (T_STRING_CAST 72) (T_SWITCH 20982) (T_THROW 74) (T_TRAIT 20983) (T_TRAIT_C 76) (T_TRY 20984) (T_UNSET 20985) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 20986) (T_WHILE 20987) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 20988) (alt_if_stmt_without_else 20989) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 20990) (attributes 20991) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 20992) (class_modifier 99) (class_modifiers 20993) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 20994) (expr 20995) (fn 106) (fully_dereferenceable 107) #'20996 (function_call 109) (function_declaration_statement 20997) (if_stmt 20998) (if_stmt_without_else 20999) (inline_function 113) (inner_statement 21000) (interface_declaration_statement 21001) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 21002) (static_member 123) (trait_declaration_statement 21003) (variable 126) (variable_class_name 127) ("{" 21004) ("~" 129)) 12702 ((";" 27312)) 12703 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21126) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21127) (fn 18437) (for_exprs 27315) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21129) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 12704 ((";" 27316)) 12705 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21135) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21136) (fn 18437) (for_cond_exprs 27317) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21138) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 12706 ((";" 27318)) 12707 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_VOID_CAST 22737) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 22738) (fn 2425) (for_exprs 27319) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_for_exprs 22740) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 12708 ((")" 27320)) 12709 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27321) (";" 18086) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 18087) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18089) (T_DEC 22) (T_DECLARE 18090) (T_DIR 24) (T_DNUMBER 25) (T_DO 18091) (T_DOUBLE_CAST 27) (T_ECHO 18092) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18094) (T_FOREACH 18095) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 18096) (T_GOTO 18097) (T_IF 18099) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18100) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 18102) (T_START_HEREDOC 69) (T_STATIC 18103) (T_STRING 18104) (T_STRING_CAST 72) (T_SWITCH 18105) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 18107) (T_UNSET 18108) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 18109) (T_WHILE 18110) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 18111) (alt_if_stmt_without_else 18112) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 18118) (fn 106) (for_statement 27322) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 18121) (if_stmt_without_else 18122) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27323) (static_member 123) (variable 126) (variable_class_name 127) ("{" 18127) ("~" 129)) 12710 ((inner_statement_list 27324)) 12711 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22750) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 22751) (T_CLASS 22752) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22753) (T_DEC 22) (T_DECLARE 22754) (T_DIR 24) (T_DNUMBER 25) (T_DO 22755) (T_DOUBLE_CAST 27) (T_ECHO 22756) (T_EMPTY 29) (T_ENDFOR 27325) (T_ENUM 22758) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 22759) (T_FOREACH 22760) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 22761) (T_GOTO 22762) (T_HALT_COMPILER 22763) (T_IF 22764) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22765) (T_INTERFACE 22766) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22767) (T_START_HEREDOC 69) (T_STATIC 22768) (T_STRING 22769) (T_STRING_CAST 72) (T_SWITCH 22770) (T_THROW 74) (T_TRAIT 22771) (T_TRAIT_C 76) (T_TRY 22772) (T_UNSET 22773) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22774) (T_WHILE 22775) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22776) (alt_if_stmt_without_else 22777) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 22778) (attributes 22779) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 22780) (class_modifier 99) (class_modifiers 22781) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 22782) (expr 22783) (fn 106) (fully_dereferenceable 107) #'22784 (function_call 109) (function_declaration_statement 22785) (if_stmt 22786) (if_stmt_without_else 22787) (inline_function 113) (inner_statement 22788) (interface_declaration_statement 22789) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 22790) (static_member 123) (trait_declaration_statement 22791) (variable 126) (variable_class_name 127) ("{" 22792) ("~" 129)) 12712 ((";" 27326)) 12713 ((":" 23454) (enum_backing_type 27328)) 12714 ((T_IMPLEMENTS 2989) (implements_list 27329)) 12715 ((backup_doc_comment 27330)) 12716 (("{" 27331)) 12717 ((class_statement_list 27332)) 12718 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 27333)) 12719 (("," 22929) (";" 27335)) 12720 ((T_WHILE 27337)) 12721 (("(" 27338)) 12722 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 27339) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 12723 (("%" 304) (")" 27340) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 12724 ((";" 27341)) 12725 ((T_STRING 23106) (const_decl 23107) (const_list 27343)) 12726 ((")" 27344) ("," 23110)) 12727 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27345) (";" 18086) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 18087) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 18089) (T_DEC 22) (T_DECLARE 18090) (T_DIR 24) (T_DNUMBER 25) (T_DO 18091) (T_DOUBLE_CAST 27) (T_ECHO 18092) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 18094) (T_FOREACH 18095) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 18096) (T_GOTO 18097) (T_IF 18099) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 18100) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 18102) (T_START_HEREDOC 69) (T_STATIC 18103) (T_STRING 18104) (T_STRING_CAST 72) (T_SWITCH 18105) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 18107) (T_UNSET 18108) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 18109) (T_WHILE 18110) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 18111) (alt_if_stmt_without_else 18112) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (declare_statement 27346) (dereferenceable_scalar 103) (expr 18118) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 18121) (if_stmt_without_else 18122) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27347) (static_member 123) (variable 126) (variable_class_name 127) ("{" 18127) ("~" 129)) 12728 ((inner_statement_list 27348)) 12729 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 23116) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 23117) (T_CLASS 23118) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 23119) (T_DEC 22) (T_DECLARE 23120) (T_DIR 24) (T_DNUMBER 25) (T_DO 23121) (T_DOUBLE_CAST 27) (T_ECHO 23122) (T_EMPTY 29) (T_ENDDECLARE 27349) (T_ENUM 23124) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 23125) (T_FOREACH 23126) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 23127) (T_GOTO 23128) (T_HALT_COMPILER 23129) (T_IF 23130) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 23131) (T_INTERFACE 23132) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 23133) (T_START_HEREDOC 69) (T_STATIC 23134) (T_STRING 23135) (T_STRING_CAST 72) (T_SWITCH 23136) (T_THROW 74) (T_TRAIT 23137) (T_TRAIT_C 76) (T_TRY 23138) (T_UNSET 23139) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 23140) (T_WHILE 23141) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 23142) (alt_if_stmt_without_else 23143) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 23144) (attributes 23145) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 23146) (class_modifier 99) (class_modifiers 23147) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 23148) (expr 23149) (fn 106) (fully_dereferenceable 107) #'23150 (function_call 109) (function_declaration_statement 23151) (if_stmt 23152) (if_stmt_without_else 23153) (inline_function 113) (inner_statement 23154) (interface_declaration_statement 23155) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23156) (static_member 123) (trait_declaration_statement 23157) (variable 126) (variable_class_name 127) ("{" 23158) ("~" 129)) 12730 ((";" 27350)) 12731 ((";" 27352)) 12732 ((T_EXTENDS 2987) (extends_from 27354)) 12733 ((T_IMPLEMENTS 2989) (implements_list 27355)) 12734 ((backup_doc_comment 27356)) 12735 (("{" 27357)) 12736 ((class_statement_list 27358)) 12737 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 27359)) 12738 ((";" 27361)) 12739 (("%" 25081) ("*" 25082) ("+" 25083) ("-" 25084) ("." 25085) ("/" 25086) (":" 27364) (";" 27365) ("<" 25087) (">" 25088) ("?" 25089) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 25090) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 25091) (T_BOOLEAN_AND 25092) (T_BOOLEAN_OR 25093) (T_COALESCE 25094) (T_INSTANCEOF 25095) (T_IS_EQUAL 25096) (T_IS_GREATER_OR_EQUAL 25097) (T_IS_IDENTICAL 25098) (T_IS_NOT_EQUAL 25099) (T_IS_NOT_IDENTICAL 25100) (T_IS_SMALLER_OR_EQUAL 25101) (T_LOGICAL_AND 25102) (T_LOGICAL_OR 25103) (T_LOGICAL_XOR 25104) (T_PIPE 25105) (T_POW 25106) (T_SL 25107) (T_SPACESHIP 25108) (T_SR 25109) ("^" 25110) ("|" 25111)) 12740 ((inner_statement_list 27367)) 12741 ((inner_statement_list 27366)) 12742 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 27369)) 12743 ((case_list 27374)) 12744 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 27372)) 12745 ((";" 27373)) 12746 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 27375)) 12747 ((";" 27376)) 12748 (("," 18369) (";" 27379)) 12749 ((";" 27381)) 12750 ((";" 27383)) 12751 (("," 19446) (";" 27385)) 12752 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 27387) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 12753 (("%" 19534) ("*" 19535) ("+" 19536) ("-" 19537) ("." 19538) ("/" 19539) ("<" 19540) (">" 19541) ("?" 19542) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19543) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19544) (T_AS 27388) (T_BOOLEAN_AND 19545) (T_BOOLEAN_OR 19546) (T_COALESCE 19547) (T_INSTANCEOF 19548) (T_IS_EQUAL 19549) (T_IS_GREATER_OR_EQUAL 19550) (T_IS_IDENTICAL 19551) (T_IS_NOT_EQUAL 19552) (T_IS_NOT_IDENTICAL 19553) (T_IS_SMALLER_OR_EQUAL 19554) (T_LOGICAL_AND 19555) (T_LOGICAL_OR 19556) (T_LOGICAL_XOR 19557) (T_PIPE 19558) (T_POW 19559) (T_SL 19560) (T_SPACESHIP 19561) (T_SR 19562) ("^" 19563) ("|" 19564)) 12754 (("\"" 1538) ("$" 20897) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20898) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20899) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20900) ("[" 20901) (ampersand 20902) (array_object_dereferenceable 20903) (callable_expr 20904) (callable_variable 20905) (class_constant 1564) (class_name 20906) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 27389) (fully_dereferenceable 107) (function_call 20908) (name 20909) (new_dereferenceable 1570) (simple_variable 20910) (static_member 20911) (variable 20912) (variable_class_name 20913)) 12755 ((")" 27390) (T_DOUBLE_ARROW 27391)) 12756 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27394) (";" 17939) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 17940) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17941) (T_DEC 22) (T_DECLARE 17942) (T_DIR 24) (T_DNUMBER 25) (T_DO 17943) (T_DOUBLE_CAST 27) (T_ECHO 17944) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17945) (T_FOREACH 17946) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 17947) (T_GOTO 17948) (T_IF 17903) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17949) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17950) (T_START_HEREDOC 69) (T_STATIC 17951) (T_STRING 17952) (T_STRING_CAST 72) (T_SWITCH 17953) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17954) (T_UNSET 17955) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17956) (T_WHILE 17957) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17958) (alt_if_stmt_without_else 17959) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 17960) (fn 106) (foreach_statement 27400) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 17961) (if_stmt_without_else 17962) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27396) (static_member 123) (variable 126) (variable_class_name 127) ("{" 17964) ("~" 129)) 12757 (("\"" 1538) ("$" 20928) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20929) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20930) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20931) ("[" 20932) (ampersand 20933) (array_object_dereferenceable 20934) (callable_expr 20935) (callable_variable 20936) (class_constant 1564) (class_name 20937) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 27392) (fully_dereferenceable 107) (function_call 20939) (name 20940) (new_dereferenceable 1570) (simple_variable 20941) (static_member 20942) (variable 20943) (variable_class_name 20944)) 12758 ((")" 27393)) 12759 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27394) (";" 17939) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 17940) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17941) (T_DEC 22) (T_DECLARE 17942) (T_DIR 24) (T_DNUMBER 25) (T_DO 17943) (T_DOUBLE_CAST 27) (T_ECHO 17944) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17945) (T_FOREACH 17946) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 17947) (T_GOTO 17948) (T_IF 17903) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17949) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17950) (T_START_HEREDOC 69) (T_STATIC 17951) (T_STRING 17952) (T_STRING_CAST 72) (T_SWITCH 17953) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17954) (T_UNSET 17955) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17956) (T_WHILE 17957) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17958) (alt_if_stmt_without_else 17959) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 17960) (fn 106) (foreach_statement 27395) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 17961) (if_stmt_without_else 17962) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27396) (static_member 123) (variable 126) (variable_class_name 127) ("{" 17964) ("~" 129)) 12760 ((inner_statement_list 27397)) 12761 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20962) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 20963) (T_CLASS 20964) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20965) (T_DEC 22) (T_DECLARE 20966) (T_DIR 24) (T_DNUMBER 25) (T_DO 20967) (T_DOUBLE_CAST 27) (T_ECHO 20968) (T_EMPTY 29) (T_ENDFOREACH 27398) (T_ENUM 20970) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20971) (T_FOREACH 20972) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20973) (T_GOTO 20974) (T_HALT_COMPILER 20975) (T_IF 20976) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20977) (T_INTERFACE 20978) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20979) (T_START_HEREDOC 69) (T_STATIC 20980) (T_STRING 20981) (T_STRING_CAST 72) (T_SWITCH 20982) (T_THROW 74) (T_TRAIT 20983) (T_TRAIT_C 76) (T_TRY 20984) (T_UNSET 20985) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 20986) (T_WHILE 20987) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 20988) (alt_if_stmt_without_else 20989) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 20990) (attributes 20991) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 20992) (class_modifier 99) (class_modifiers 20993) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 20994) (expr 20995) (fn 106) (fully_dereferenceable 107) #'20996 (function_call 109) (function_declaration_statement 20997) (if_stmt 20998) (if_stmt_without_else 20999) (inline_function 113) (inner_statement 21000) (interface_declaration_statement 21001) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 21002) (static_member 123) (trait_declaration_statement 21003) (variable 126) (variable_class_name 127) ("{" 21004) ("~" 129)) 12762 ((";" 27399)) 12763 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21126) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21127) (fn 18437) (for_exprs 27402) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21129) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 12764 ((";" 27403)) 12765 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21135) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21136) (fn 18437) (for_cond_exprs 27404) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21138) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 12766 ((";" 27405)) 12767 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_VOID_CAST 22737) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 22738) (fn 2425) (for_exprs 27406) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_for_exprs 22740) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 12768 ((")" 27407)) 12769 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27408) (";" 17939) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 17940) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17941) (T_DEC 22) (T_DECLARE 17942) (T_DIR 24) (T_DNUMBER 25) (T_DO 17943) (T_DOUBLE_CAST 27) (T_ECHO 17944) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17945) (T_FOREACH 17946) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 17947) (T_GOTO 17948) (T_IF 17903) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17949) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17950) (T_START_HEREDOC 69) (T_STATIC 17951) (T_STRING 17952) (T_STRING_CAST 72) (T_SWITCH 17953) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17954) (T_UNSET 17955) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17956) (T_WHILE 17957) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17958) (alt_if_stmt_without_else 17959) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 17960) (fn 106) (for_statement 27409) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 17961) (if_stmt_without_else 17962) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27410) (static_member 123) (variable 126) (variable_class_name 127) ("{" 17964) ("~" 129)) 12770 ((inner_statement_list 27411)) 12771 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22750) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 22751) (T_CLASS 22752) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22753) (T_DEC 22) (T_DECLARE 22754) (T_DIR 24) (T_DNUMBER 25) (T_DO 22755) (T_DOUBLE_CAST 27) (T_ECHO 22756) (T_EMPTY 29) (T_ENDFOR 27412) (T_ENUM 22758) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 22759) (T_FOREACH 22760) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 22761) (T_GOTO 22762) (T_HALT_COMPILER 22763) (T_IF 22764) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22765) (T_INTERFACE 22766) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22767) (T_START_HEREDOC 69) (T_STATIC 22768) (T_STRING 22769) (T_STRING_CAST 72) (T_SWITCH 22770) (T_THROW 74) (T_TRAIT 22771) (T_TRAIT_C 76) (T_TRY 22772) (T_UNSET 22773) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22774) (T_WHILE 22775) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22776) (alt_if_stmt_without_else 22777) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 22778) (attributes 22779) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 22780) (class_modifier 99) (class_modifiers 22781) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 22782) (expr 22783) (fn 106) (fully_dereferenceable 107) #'22784 (function_call 109) (function_declaration_statement 22785) (if_stmt 22786) (if_stmt_without_else 22787) (inline_function 113) (inner_statement 22788) (interface_declaration_statement 22789) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 22790) (static_member 123) (trait_declaration_statement 22791) (variable 126) (variable_class_name 127) ("{" 22792) ("~" 129)) 12772 ((";" 27413)) 12773 (("," 22929) (";" 27415)) 12774 ((T_WHILE 27417)) 12775 (("(" 27418)) 12776 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 27419) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 12777 (("%" 304) (")" 27420) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 12778 ((";" 27421)) 12779 ((T_STRING 23106) (const_decl 23107) (const_list 27423)) 12780 ((")" 27424) ("," 23110)) 12781 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27425) (";" 17939) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 17940) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17941) (T_DEC 22) (T_DECLARE 17942) (T_DIR 24) (T_DNUMBER 25) (T_DO 17943) (T_DOUBLE_CAST 27) (T_ECHO 17944) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17945) (T_FOREACH 17946) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 17947) (T_GOTO 17948) (T_IF 17903) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17949) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17950) (T_START_HEREDOC 69) (T_STATIC 17951) (T_STRING 17952) (T_STRING_CAST 72) (T_SWITCH 17953) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17954) (T_UNSET 17955) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17956) (T_WHILE 17957) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17958) (alt_if_stmt_without_else 17959) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (declare_statement 27426) (dereferenceable_scalar 103) (expr 17960) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 17961) (if_stmt_without_else 17962) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27427) (static_member 123) (variable 126) (variable_class_name 127) ("{" 17964) ("~" 129)) 12782 ((inner_statement_list 27428)) 12783 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 23116) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 23117) (T_CLASS 23118) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 23119) (T_DEC 22) (T_DECLARE 23120) (T_DIR 24) (T_DNUMBER 25) (T_DO 23121) (T_DOUBLE_CAST 27) (T_ECHO 23122) (T_EMPTY 29) (T_ENDDECLARE 27429) (T_ENUM 23124) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 23125) (T_FOREACH 23126) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 23127) (T_GOTO 23128) (T_HALT_COMPILER 23129) (T_IF 23130) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 23131) (T_INTERFACE 23132) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 23133) (T_START_HEREDOC 69) (T_STATIC 23134) (T_STRING 23135) (T_STRING_CAST 72) (T_SWITCH 23136) (T_THROW 74) (T_TRAIT 23137) (T_TRAIT_C 76) (T_TRY 23138) (T_UNSET 23139) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 23140) (T_WHILE 23141) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 23142) (alt_if_stmt_without_else 23143) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 23144) (attributes 23145) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 23146) (class_modifier 99) (class_modifiers 23147) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 23148) (expr 23149) (fn 106) (fully_dereferenceable 107) #'23150 (function_call 109) (function_declaration_statement 23151) (if_stmt 23152) (if_stmt_without_else 23153) (inline_function 113) (inner_statement 23154) (interface_declaration_statement 23155) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23156) (static_member 123) (trait_declaration_statement 23157) (variable 126) (variable_class_name 127) ("{" 23158) ("~" 129)) 12784 ((";" 27430)) 12785 ((";" 27432)) 12786 ((";" 27434)) 12787 ((T_READONLY 17774) (T_STRING 17775) (backup_doc_comment 143) (function_name 27437)) 12788 ((backup_doc_comment 27438)) 12789 (("(" 27439)) 12790 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 27440)) 12791 ((")" 27441)) 12792 ((":" 168) (return_type 27442)) 12793 ((backup_fn_flags 27443)) 12794 (("{" 27444)) 12795 ((inner_statement_list 27445)) 12796 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 27446) ("~" 129)) 12797 ((backup_fn_flags 27447)) 12798 ((T_STRING 27450)) 12799 ((T_EXTENDS 2987) (extends_from 27451)) 12800 ((T_IMPLEMENTS 2989) (implements_list 27452)) 12801 ((backup_doc_comment 27453)) 12802 (("{" 27454)) 12803 ((class_statement_list 27455)) 12804 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 27456)) 12805 ((":" 27461)) 12806 ((";" 27460)) 12807 ((inner_statement_list 27462)) 12808 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17807) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17808) (T_CLASS 17809) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17810) (T_DEC 22) (T_DECLARE 17811) (T_DIR 24) (T_DNUMBER 25) (T_DO 17812) (T_DOUBLE_CAST 27) (T_ECHO 17813) (T_EMPTY 29) (T_ENDIF 27463) (T_ENUM 17815) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17816) (T_FOREACH 17817) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17818) (T_GOTO 17819) (T_HALT_COMPILER 17820) (T_IF 17732) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17821) (T_INTERFACE 17822) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17823) (T_START_HEREDOC 69) (T_STATIC 17824) (T_STRING 17825) (T_STRING_CAST 72) (T_SWITCH 17826) (T_THROW 74) (T_TRAIT 17827) (T_TRAIT_C 76) (T_TRY 17828) (T_UNSET 17829) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17830) (T_WHILE 17831) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17832) (alt_if_stmt_without_else 17833) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17834) (attributes 17835) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17836) (class_modifier 99) (class_modifiers 17837) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17838) (expr 17839) (fn 106) (fully_dereferenceable 107) #'17840 (function_call 109) (function_declaration_statement 17841) (if_stmt 17842) (if_stmt_without_else 17843) (inline_function 113) (inner_statement 17844) (interface_declaration_statement 17845) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17846) (static_member 123) (trait_declaration_statement 17847) (variable 126) (variable_class_name 127) ("{" 17848) ("~" 129)) 12809 ((";" 27464)) 12810 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 27466) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 12811 (("%" 304) (")" 27467) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 12812 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27468) (";" 17889) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 17890) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17892) (T_DEC 22) (T_DECLARE 17893) (T_DIR 24) (T_DNUMBER 25) (T_DO 17894) (T_DOUBLE_CAST 27) (T_ECHO 17895) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17898) (T_FOREACH 17899) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 17900) (T_GOTO 17901) (T_IF 17903) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17904) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17906) (T_START_HEREDOC 69) (T_STATIC 17907) (T_STRING 17908) (T_STRING_CAST 72) (T_SWITCH 17909) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17911) (T_UNSET 17912) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17913) (T_WHILE 17914) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17915) (alt_if_stmt_without_else 17916) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 17922) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 17925) (if_stmt_without_else 17926) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27469) (static_member 123) (variable 126) (variable_class_name 127) (while_statement 27470) ("{" 17931) ("~" 129)) 12813 ((inner_statement_list 27471)) 12814 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17889) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17890) (T_CLASS 17891) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17892) (T_DEC 22) (T_DECLARE 17893) (T_DIR 24) (T_DNUMBER 25) (T_DO 17894) (T_DOUBLE_CAST 27) (T_ECHO 17895) (T_EMPTY 29) (T_ENDWHILE 27472) (T_ENUM 17897) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17898) (T_FOREACH 17899) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17900) (T_GOTO 17901) (T_HALT_COMPILER 17902) (T_IF 17903) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17904) (T_INTERFACE 17905) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17906) (T_START_HEREDOC 69) (T_STATIC 17907) (T_STRING 17908) (T_STRING_CAST 72) (T_SWITCH 17909) (T_THROW 74) (T_TRAIT 17910) (T_TRAIT_C 76) (T_TRY 17911) (T_UNSET 17912) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17913) (T_WHILE 17914) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17915) (alt_if_stmt_without_else 17916) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17917) (attributes 17918) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17919) (class_modifier 99) (class_modifiers 17920) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17921) (expr 17922) (fn 106) (fully_dereferenceable 107) #'17923 (function_call 109) (function_declaration_statement 17924) (if_stmt 17925) (if_stmt_without_else 17926) (inline_function 113) (inner_statement 17927) (interface_declaration_statement 17928) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17929) (static_member 123) (trait_declaration_statement 17930) (variable 126) (variable_class_name 127) ("{" 17931) ("~" 129)) 12815 ((";" 27473)) 12816 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 27475) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 12817 (("\"" 1538) ("$" 17989) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 17990) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 17991) ("[" 1560) (array_object_dereferenceable 17992) (callable_expr 17993) (callable_variable 17994) (class_constant 1564) (class_name 17995) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 17996) (name 17997) (new_dereferenceable 1570) (simple_variable 17998) (static_member 17999) (unset_variable 18000) (unset_variables 27477) (variable 18002) (variable_class_name 18003)) 12818 (("," 18015) (possible_comma 27478)) 12819 ((")" 27479)) 12820 ((";" 27480)) 12821 ((inner_statement_list 27482)) 12822 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 27483) ("~" 129)) 12823 ((catch_list 27484)) 12824 ((T_CATCH 27485) (T_FINALLY 27486) (finally_statement 27487)) 12825 (("(" 27491)) 12826 (("{" 27488)) 12827 ((inner_statement_list 27489)) 12828 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 27490) ("~" 129)) 12829 ((T_NAME_FULLY_QUALIFIED 18056) (T_NAME_QUALIFIED 18057) (T_NAME_RELATIVE 18058) (T_STATIC 18059) (T_STRING 18060) (catch_name_list 27492) (class_name 18062) (name 18063)) 12830 ((T_VARIABLE 18064) (optional_variable 27493) ("|" 18066)) 12831 ((")" 27494)) 12832 (("{" 27495)) 12833 ((inner_statement_list 27496)) 12834 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 27497) ("~" 129)) 12835 ((backup_doc_comment 27499)) 12836 (("{" 27500)) 12837 ((class_statement_list 27501)) 12838 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 27502)) 12839 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 27504) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 12840 (("%" 304) (")" 27505) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 12841 ((":" 27506) (switch_case_list 27507) ("{" 27508)) 12842 ((";" 27514) (case_list 27515)) 12843 ((";" 27509) (case_list 27510)) 12844 ((case_list 27512)) 12845 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 27511)) 12846 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 27513)) 12847 ((case_list 27518)) 12848 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 27516)) 12849 ((";" 27517)) 12850 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 27519)) 12851 ((";" 27520)) 12852 (("," 18369) (";" 27523)) 12853 ((";" 27525)) 12854 ((T_EXTENDS 23405) (interface_extends_list 27527)) 12855 ((backup_doc_comment 27528)) 12856 (("{" 27529)) 12857 ((class_statement_list 27530)) 12858 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 27531)) 12859 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 27533) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 12860 (("%" 304) (")" 27534) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 12861 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23415) (";" 17939) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 17940) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17941) (T_DEC 22) (T_DECLARE 17942) (T_DIR 24) (T_DNUMBER 25) (T_DO 17943) (T_DOUBLE_CAST 27) (T_ECHO 17944) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17945) (T_FOREACH 17946) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 17947) (T_GOTO 17948) (T_IF 17903) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17949) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17950) (T_START_HEREDOC 69) (T_STATIC 17951) (T_STRING 17952) (T_STRING_CAST 72) (T_SWITCH 17953) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17954) (T_UNSET 17955) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17956) (T_WHILE 17957) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17958) (alt_if_stmt_without_else 17959) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 17960) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 17961) (if_stmt_without_else 17962) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27535) (static_member 123) (variable 126) (variable_class_name 127) ("{" 17964) ("~" 129)) 12862 ((")" 27537)) 12863 ((";" 27538)) 12864 ((";" 27540)) 12865 (("," 19446) (";" 27542)) 12866 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 27544) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 12867 (("%" 19534) ("*" 19535) ("+" 19536) ("-" 19537) ("." 19538) ("/" 19539) ("<" 19540) (">" 19541) ("?" 19542) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19543) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19544) (T_AS 27545) (T_BOOLEAN_AND 19545) (T_BOOLEAN_OR 19546) (T_COALESCE 19547) (T_INSTANCEOF 19548) (T_IS_EQUAL 19549) (T_IS_GREATER_OR_EQUAL 19550) (T_IS_IDENTICAL 19551) (T_IS_NOT_EQUAL 19552) (T_IS_NOT_IDENTICAL 19553) (T_IS_SMALLER_OR_EQUAL 19554) (T_LOGICAL_AND 19555) (T_LOGICAL_OR 19556) (T_LOGICAL_XOR 19557) (T_PIPE 19558) (T_POW 19559) (T_SL 19560) (T_SPACESHIP 19561) (T_SR 19562) ("^" 19563) ("|" 19564)) 12868 (("\"" 1538) ("$" 20897) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20898) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20899) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20900) ("[" 20901) (ampersand 20902) (array_object_dereferenceable 20903) (callable_expr 20904) (callable_variable 20905) (class_constant 1564) (class_name 20906) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 27546) (fully_dereferenceable 107) (function_call 20908) (name 20909) (new_dereferenceable 1570) (simple_variable 20910) (static_member 20911) (variable 20912) (variable_class_name 20913)) 12869 ((")" 27547) (T_DOUBLE_ARROW 27548)) 12870 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27551) (";" 17889) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 17890) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17892) (T_DEC 22) (T_DECLARE 17893) (T_DIR 24) (T_DNUMBER 25) (T_DO 17894) (T_DOUBLE_CAST 27) (T_ECHO 17895) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17898) (T_FOREACH 17899) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 17900) (T_GOTO 17901) (T_IF 17903) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17904) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17906) (T_START_HEREDOC 69) (T_STATIC 17907) (T_STRING 17908) (T_STRING_CAST 72) (T_SWITCH 17909) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17911) (T_UNSET 17912) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17913) (T_WHILE 17914) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17915) (alt_if_stmt_without_else 17916) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 17922) (fn 106) (foreach_statement 27557) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 17925) (if_stmt_without_else 17926) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27553) (static_member 123) (variable 126) (variable_class_name 127) ("{" 17931) ("~" 129)) 12871 (("\"" 1538) ("$" 20928) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20929) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20930) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20931) ("[" 20932) (ampersand 20933) (array_object_dereferenceable 20934) (callable_expr 20935) (callable_variable 20936) (class_constant 1564) (class_name 20937) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 27549) (fully_dereferenceable 107) (function_call 20939) (name 20940) (new_dereferenceable 1570) (simple_variable 20941) (static_member 20942) (variable 20943) (variable_class_name 20944)) 12872 ((")" 27550)) 12873 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27551) (";" 17889) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 17890) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17892) (T_DEC 22) (T_DECLARE 17893) (T_DIR 24) (T_DNUMBER 25) (T_DO 17894) (T_DOUBLE_CAST 27) (T_ECHO 17895) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17898) (T_FOREACH 17899) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 17900) (T_GOTO 17901) (T_IF 17903) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17904) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17906) (T_START_HEREDOC 69) (T_STATIC 17907) (T_STRING 17908) (T_STRING_CAST 72) (T_SWITCH 17909) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17911) (T_UNSET 17912) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17913) (T_WHILE 17914) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17915) (alt_if_stmt_without_else 17916) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 17922) (fn 106) (foreach_statement 27552) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 17925) (if_stmt_without_else 17926) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27553) (static_member 123) (variable 126) (variable_class_name 127) ("{" 17931) ("~" 129)) 12874 ((inner_statement_list 27554)) 12875 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20962) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 20963) (T_CLASS 20964) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20965) (T_DEC 22) (T_DECLARE 20966) (T_DIR 24) (T_DNUMBER 25) (T_DO 20967) (T_DOUBLE_CAST 27) (T_ECHO 20968) (T_EMPTY 29) (T_ENDFOREACH 27555) (T_ENUM 20970) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20971) (T_FOREACH 20972) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20973) (T_GOTO 20974) (T_HALT_COMPILER 20975) (T_IF 20976) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20977) (T_INTERFACE 20978) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20979) (T_START_HEREDOC 69) (T_STATIC 20980) (T_STRING 20981) (T_STRING_CAST 72) (T_SWITCH 20982) (T_THROW 74) (T_TRAIT 20983) (T_TRAIT_C 76) (T_TRY 20984) (T_UNSET 20985) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 20986) (T_WHILE 20987) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 20988) (alt_if_stmt_without_else 20989) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 20990) (attributes 20991) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 20992) (class_modifier 99) (class_modifiers 20993) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 20994) (expr 20995) (fn 106) (fully_dereferenceable 107) #'20996 (function_call 109) (function_declaration_statement 20997) (if_stmt 20998) (if_stmt_without_else 20999) (inline_function 113) (inner_statement 21000) (interface_declaration_statement 21001) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 21002) (static_member 123) (trait_declaration_statement 21003) (variable 126) (variable_class_name 127) ("{" 21004) ("~" 129)) 12876 ((";" 27556)) 12877 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21126) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21127) (fn 18437) (for_exprs 27559) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21129) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 12878 ((";" 27560)) 12879 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21135) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21136) (fn 18437) (for_cond_exprs 27561) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21138) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 12880 ((";" 27562)) 12881 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_VOID_CAST 22737) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 22738) (fn 2425) (for_exprs 27563) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_for_exprs 22740) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 12882 ((")" 27564)) 12883 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27565) (";" 17889) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 17890) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17892) (T_DEC 22) (T_DECLARE 17893) (T_DIR 24) (T_DNUMBER 25) (T_DO 17894) (T_DOUBLE_CAST 27) (T_ECHO 17895) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17898) (T_FOREACH 17899) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 17900) (T_GOTO 17901) (T_IF 17903) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17904) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17906) (T_START_HEREDOC 69) (T_STATIC 17907) (T_STRING 17908) (T_STRING_CAST 72) (T_SWITCH 17909) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17911) (T_UNSET 17912) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17913) (T_WHILE 17914) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17915) (alt_if_stmt_without_else 17916) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 17922) (fn 106) (for_statement 27566) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 17925) (if_stmt_without_else 17926) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27567) (static_member 123) (variable 126) (variable_class_name 127) ("{" 17931) ("~" 129)) 12884 ((inner_statement_list 27568)) 12885 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22750) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 22751) (T_CLASS 22752) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22753) (T_DEC 22) (T_DECLARE 22754) (T_DIR 24) (T_DNUMBER 25) (T_DO 22755) (T_DOUBLE_CAST 27) (T_ECHO 22756) (T_EMPTY 29) (T_ENDFOR 27569) (T_ENUM 22758) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 22759) (T_FOREACH 22760) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 22761) (T_GOTO 22762) (T_HALT_COMPILER 22763) (T_IF 22764) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22765) (T_INTERFACE 22766) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22767) (T_START_HEREDOC 69) (T_STATIC 22768) (T_STRING 22769) (T_STRING_CAST 72) (T_SWITCH 22770) (T_THROW 74) (T_TRAIT 22771) (T_TRAIT_C 76) (T_TRY 22772) (T_UNSET 22773) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22774) (T_WHILE 22775) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22776) (alt_if_stmt_without_else 22777) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 22778) (attributes 22779) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 22780) (class_modifier 99) (class_modifiers 22781) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 22782) (expr 22783) (fn 106) (fully_dereferenceable 107) #'22784 (function_call 109) (function_declaration_statement 22785) (if_stmt 22786) (if_stmt_without_else 22787) (inline_function 113) (inner_statement 22788) (interface_declaration_statement 22789) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 22790) (static_member 123) (trait_declaration_statement 22791) (variable 126) (variable_class_name 127) ("{" 22792) ("~" 129)) 12886 ((";" 27570)) 12887 ((":" 23454) (enum_backing_type 27572)) 12888 ((T_IMPLEMENTS 2989) (implements_list 27573)) 12889 ((backup_doc_comment 27574)) 12890 (("{" 27575)) 12891 ((class_statement_list 27576)) 12892 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 27577)) 12893 (("," 22929) (";" 27580)) 12894 ((T_WHILE 27582)) 12895 (("(" 27583)) 12896 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 27584) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 12897 (("%" 304) (")" 27585) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 12898 ((";" 27586)) 12899 ((T_STRING 23106) (const_decl 23107) (const_list 27588)) 12900 ((")" 27589) ("," 23110)) 12901 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27590) (";" 17889) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 17890) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17892) (T_DEC 22) (T_DECLARE 17893) (T_DIR 24) (T_DNUMBER 25) (T_DO 17894) (T_DOUBLE_CAST 27) (T_ECHO 17895) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17898) (T_FOREACH 17899) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 17900) (T_GOTO 17901) (T_IF 17903) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17904) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17906) (T_START_HEREDOC 69) (T_STATIC 17907) (T_STRING 17908) (T_STRING_CAST 72) (T_SWITCH 17909) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17911) (T_UNSET 17912) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17913) (T_WHILE 17914) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17915) (alt_if_stmt_without_else 17916) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (declare_statement 27591) (dereferenceable_scalar 103) (expr 17922) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 17925) (if_stmt_without_else 17926) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27592) (static_member 123) (variable 126) (variable_class_name 127) ("{" 17931) ("~" 129)) 12902 ((inner_statement_list 27593)) 12903 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 23116) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 23117) (T_CLASS 23118) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 23119) (T_DEC 22) (T_DECLARE 23120) (T_DIR 24) (T_DNUMBER 25) (T_DO 23121) (T_DOUBLE_CAST 27) (T_ECHO 23122) (T_EMPTY 29) (T_ENDDECLARE 27594) (T_ENUM 23124) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 23125) (T_FOREACH 23126) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 23127) (T_GOTO 23128) (T_HALT_COMPILER 23129) (T_IF 23130) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 23131) (T_INTERFACE 23132) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 23133) (T_START_HEREDOC 69) (T_STATIC 23134) (T_STRING 23135) (T_STRING_CAST 72) (T_SWITCH 23136) (T_THROW 74) (T_TRAIT 23137) (T_TRAIT_C 76) (T_TRY 23138) (T_UNSET 23139) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 23140) (T_WHILE 23141) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 23142) (alt_if_stmt_without_else 23143) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 23144) (attributes 23145) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 23146) (class_modifier 99) (class_modifiers 23147) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 23148) (expr 23149) (fn 106) (fully_dereferenceable 107) #'23150 (function_call 109) (function_declaration_statement 23151) (if_stmt 23152) (if_stmt_without_else 23153) (inline_function 113) (inner_statement 23154) (interface_declaration_statement 23155) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23156) (static_member 123) (trait_declaration_statement 23157) (variable 126) (variable_class_name 127) ("{" 23158) ("~" 129)) 12904 ((";" 27595)) 12905 ((";" 27597)) 12906 ((T_EXTENDS 2987) (extends_from 27599)) 12907 ((T_IMPLEMENTS 2989) (implements_list 27600)) 12908 ((backup_doc_comment 27601)) 12909 (("{" 27602)) 12910 ((class_statement_list 27603)) 12911 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 27604)) 12912 ((";" 27606)) 12913 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 27608) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 12914 (("\"" 1538) ("$" 17989) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 17990) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 17991) ("[" 1560) (array_object_dereferenceable 17992) (callable_expr 17993) (callable_variable 17994) (class_constant 1564) (class_name 17995) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 17996) (name 17997) (new_dereferenceable 1570) (simple_variable 17998) (static_member 17999) (unset_variable 18000) (unset_variables 27610) (variable 18002) (variable_class_name 18003)) 12915 (("," 18015) (possible_comma 27611)) 12916 ((")" 27612)) 12917 ((";" 27613)) 12918 ((inner_statement_list 27615)) 12919 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 27616) ("~" 129)) 12920 ((catch_list 27617)) 12921 ((T_CATCH 27618) (T_FINALLY 27619) (finally_statement 27620)) 12922 (("(" 27624)) 12923 (("{" 27621)) 12924 ((inner_statement_list 27622)) 12925 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 27623) ("~" 129)) 12926 ((T_NAME_FULLY_QUALIFIED 18056) (T_NAME_QUALIFIED 18057) (T_NAME_RELATIVE 18058) (T_STATIC 18059) (T_STRING 18060) (catch_name_list 27625) (class_name 18062) (name 18063)) 12927 ((T_VARIABLE 18064) (optional_variable 27626) ("|" 18066)) 12928 ((")" 27627)) 12929 (("{" 27628)) 12930 ((inner_statement_list 27629)) 12931 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 27630) ("~" 129)) 12932 ((backup_doc_comment 27632)) 12933 (("{" 27633)) 12934 ((class_statement_list 27634)) 12935 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 27635)) 12936 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 27637) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 12937 (("%" 304) (")" 27638) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 12938 ((":" 27639) (switch_case_list 27640) ("{" 27641)) 12939 ((";" 27647) (case_list 27648)) 12940 ((";" 27642) (case_list 27643)) 12941 ((case_list 27645)) 12942 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 27644)) 12943 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 27646)) 12944 ((case_list 27651)) 12945 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 27649)) 12946 ((";" 27650)) 12947 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 27652)) 12948 ((";" 27653)) 12949 (("," 18369) (";" 27656)) 12950 ((";" 27658)) 12951 ((T_EXTENDS 23405) (interface_extends_list 27660)) 12952 ((backup_doc_comment 27661)) 12953 (("{" 27662)) 12954 ((class_statement_list 27663)) 12955 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 27664)) 12956 ((")" 27666)) 12957 ((";" 27667)) 12958 ((";" 27669)) 12959 (("," 19446) (";" 27671)) 12960 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 27673) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 12961 (("%" 19534) ("*" 19535) ("+" 19536) ("-" 19537) ("." 19538) ("/" 19539) ("<" 19540) (">" 19541) ("?" 19542) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19543) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19544) (T_AS 27674) (T_BOOLEAN_AND 19545) (T_BOOLEAN_OR 19546) (T_COALESCE 19547) (T_INSTANCEOF 19548) (T_IS_EQUAL 19549) (T_IS_GREATER_OR_EQUAL 19550) (T_IS_IDENTICAL 19551) (T_IS_NOT_EQUAL 19552) (T_IS_NOT_IDENTICAL 19553) (T_IS_SMALLER_OR_EQUAL 19554) (T_LOGICAL_AND 19555) (T_LOGICAL_OR 19556) (T_LOGICAL_XOR 19557) (T_PIPE 19558) (T_POW 19559) (T_SL 19560) (T_SPACESHIP 19561) (T_SR 19562) ("^" 19563) ("|" 19564)) 12962 (("\"" 1538) ("$" 20897) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20898) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20899) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20900) ("[" 20901) (ampersand 20902) (array_object_dereferenceable 20903) (callable_expr 20904) (callable_variable 20905) (class_constant 1564) (class_name 20906) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 27675) (fully_dereferenceable 107) (function_call 20908) (name 20909) (new_dereferenceable 1570) (simple_variable 20910) (static_member 20911) (variable 20912) (variable_class_name 20913)) 12963 ((")" 27676) (T_DOUBLE_ARROW 27677)) 12964 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27680) (";" 17807) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 17808) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17810) (T_DEC 22) (T_DECLARE 17811) (T_DIR 24) (T_DNUMBER 25) (T_DO 17812) (T_DOUBLE_CAST 27) (T_ECHO 17813) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17816) (T_FOREACH 17817) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 17818) (T_GOTO 17819) (T_IF 17732) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17821) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17823) (T_START_HEREDOC 69) (T_STATIC 17824) (T_STRING 17825) (T_STRING_CAST 72) (T_SWITCH 17826) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17828) (T_UNSET 17829) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17830) (T_WHILE 17831) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17832) (alt_if_stmt_without_else 17833) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 17839) (fn 106) (foreach_statement 27686) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 17842) (if_stmt_without_else 17843) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27682) (static_member 123) (variable 126) (variable_class_name 127) ("{" 17848) ("~" 129)) 12965 (("\"" 1538) ("$" 20928) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20929) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20930) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20931) ("[" 20932) (ampersand 20933) (array_object_dereferenceable 20934) (callable_expr 20935) (callable_variable 20936) (class_constant 1564) (class_name 20937) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 27678) (fully_dereferenceable 107) (function_call 20939) (name 20940) (new_dereferenceable 1570) (simple_variable 20941) (static_member 20942) (variable 20943) (variable_class_name 20944)) 12966 ((")" 27679)) 12967 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27680) (";" 17807) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 17808) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17810) (T_DEC 22) (T_DECLARE 17811) (T_DIR 24) (T_DNUMBER 25) (T_DO 17812) (T_DOUBLE_CAST 27) (T_ECHO 17813) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17816) (T_FOREACH 17817) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 17818) (T_GOTO 17819) (T_IF 17732) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17821) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17823) (T_START_HEREDOC 69) (T_STATIC 17824) (T_STRING 17825) (T_STRING_CAST 72) (T_SWITCH 17826) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17828) (T_UNSET 17829) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17830) (T_WHILE 17831) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17832) (alt_if_stmt_without_else 17833) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 17839) (fn 106) (foreach_statement 27681) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 17842) (if_stmt_without_else 17843) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27682) (static_member 123) (variable 126) (variable_class_name 127) ("{" 17848) ("~" 129)) 12968 ((inner_statement_list 27683)) 12969 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20962) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 20963) (T_CLASS 20964) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20965) (T_DEC 22) (T_DECLARE 20966) (T_DIR 24) (T_DNUMBER 25) (T_DO 20967) (T_DOUBLE_CAST 27) (T_ECHO 20968) (T_EMPTY 29) (T_ENDFOREACH 27684) (T_ENUM 20970) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20971) (T_FOREACH 20972) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20973) (T_GOTO 20974) (T_HALT_COMPILER 20975) (T_IF 20976) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20977) (T_INTERFACE 20978) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20979) (T_START_HEREDOC 69) (T_STATIC 20980) (T_STRING 20981) (T_STRING_CAST 72) (T_SWITCH 20982) (T_THROW 74) (T_TRAIT 20983) (T_TRAIT_C 76) (T_TRY 20984) (T_UNSET 20985) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 20986) (T_WHILE 20987) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 20988) (alt_if_stmt_without_else 20989) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 20990) (attributes 20991) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 20992) (class_modifier 99) (class_modifiers 20993) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 20994) (expr 20995) (fn 106) (fully_dereferenceable 107) #'20996 (function_call 109) (function_declaration_statement 20997) (if_stmt 20998) (if_stmt_without_else 20999) (inline_function 113) (inner_statement 21000) (interface_declaration_statement 21001) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 21002) (static_member 123) (trait_declaration_statement 21003) (variable 126) (variable_class_name 127) ("{" 21004) ("~" 129)) 12970 ((";" 27685)) 12971 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21126) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21127) (fn 18437) (for_exprs 27688) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21129) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 12972 ((";" 27689)) 12973 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21135) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21136) (fn 18437) (for_cond_exprs 27690) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21138) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 12974 ((";" 27691)) 12975 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_VOID_CAST 22737) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 22738) (fn 2425) (for_exprs 27692) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_for_exprs 22740) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 12976 ((")" 27693)) 12977 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27694) (";" 17807) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 17808) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17810) (T_DEC 22) (T_DECLARE 17811) (T_DIR 24) (T_DNUMBER 25) (T_DO 17812) (T_DOUBLE_CAST 27) (T_ECHO 17813) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17816) (T_FOREACH 17817) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 17818) (T_GOTO 17819) (T_IF 17732) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17821) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17823) (T_START_HEREDOC 69) (T_STATIC 17824) (T_STRING 17825) (T_STRING_CAST 72) (T_SWITCH 17826) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17828) (T_UNSET 17829) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17830) (T_WHILE 17831) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17832) (alt_if_stmt_without_else 17833) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 17839) (fn 106) (for_statement 27695) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 17842) (if_stmt_without_else 17843) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27696) (static_member 123) (variable 126) (variable_class_name 127) ("{" 17848) ("~" 129)) 12978 ((inner_statement_list 27697)) 12979 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22750) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 22751) (T_CLASS 22752) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22753) (T_DEC 22) (T_DECLARE 22754) (T_DIR 24) (T_DNUMBER 25) (T_DO 22755) (T_DOUBLE_CAST 27) (T_ECHO 22756) (T_EMPTY 29) (T_ENDFOR 27698) (T_ENUM 22758) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 22759) (T_FOREACH 22760) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 22761) (T_GOTO 22762) (T_HALT_COMPILER 22763) (T_IF 22764) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22765) (T_INTERFACE 22766) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22767) (T_START_HEREDOC 69) (T_STATIC 22768) (T_STRING 22769) (T_STRING_CAST 72) (T_SWITCH 22770) (T_THROW 74) (T_TRAIT 22771) (T_TRAIT_C 76) (T_TRY 22772) (T_UNSET 22773) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22774) (T_WHILE 22775) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22776) (alt_if_stmt_without_else 22777) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 22778) (attributes 22779) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 22780) (class_modifier 99) (class_modifiers 22781) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 22782) (expr 22783) (fn 106) (fully_dereferenceable 107) #'22784 (function_call 109) (function_declaration_statement 22785) (if_stmt 22786) (if_stmt_without_else 22787) (inline_function 113) (inner_statement 22788) (interface_declaration_statement 22789) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 22790) (static_member 123) (trait_declaration_statement 22791) (variable 126) (variable_class_name 127) ("{" 22792) ("~" 129)) 12980 ((";" 27699)) 12981 ((":" 23454) (enum_backing_type 27701)) 12982 ((T_IMPLEMENTS 2989) (implements_list 27702)) 12983 ((backup_doc_comment 27703)) 12984 (("{" 27704)) 12985 ((class_statement_list 27705)) 12986 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 27706)) 12987 (("," 22929) (";" 27709)) 12988 ((T_WHILE 27711)) 12989 (("(" 27712)) 12990 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 27713) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 12991 (("%" 304) (")" 27714) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 12992 ((";" 27715)) 12993 ((T_STRING 23106) (const_decl 23107) (const_list 27717)) 12994 ((")" 27718) ("," 23110)) 12995 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27719) (";" 17807) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 17808) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17810) (T_DEC 22) (T_DECLARE 17811) (T_DIR 24) (T_DNUMBER 25) (T_DO 17812) (T_DOUBLE_CAST 27) (T_ECHO 17813) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17816) (T_FOREACH 17817) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 17818) (T_GOTO 17819) (T_IF 17732) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17821) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17823) (T_START_HEREDOC 69) (T_STATIC 17824) (T_STRING 17825) (T_STRING_CAST 72) (T_SWITCH 17826) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17828) (T_UNSET 17829) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17830) (T_WHILE 17831) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17832) (alt_if_stmt_without_else 17833) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (declare_statement 27720) (dereferenceable_scalar 103) (expr 17839) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 17842) (if_stmt_without_else 17843) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27721) (static_member 123) (variable 126) (variable_class_name 127) ("{" 17848) ("~" 129)) 12996 ((inner_statement_list 27722)) 12997 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 23116) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 23117) (T_CLASS 23118) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 23119) (T_DEC 22) (T_DECLARE 23120) (T_DIR 24) (T_DNUMBER 25) (T_DO 23121) (T_DOUBLE_CAST 27) (T_ECHO 23122) (T_EMPTY 29) (T_ENDDECLARE 27723) (T_ENUM 23124) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 23125) (T_FOREACH 23126) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 23127) (T_GOTO 23128) (T_HALT_COMPILER 23129) (T_IF 23130) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 23131) (T_INTERFACE 23132) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 23133) (T_START_HEREDOC 69) (T_STATIC 23134) (T_STRING 23135) (T_STRING_CAST 72) (T_SWITCH 23136) (T_THROW 74) (T_TRAIT 23137) (T_TRAIT_C 76) (T_TRY 23138) (T_UNSET 23139) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 23140) (T_WHILE 23141) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 23142) (alt_if_stmt_without_else 23143) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 23144) (attributes 23145) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 23146) (class_modifier 99) (class_modifiers 23147) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 23148) (expr 23149) (fn 106) (fully_dereferenceable 107) #'23150 (function_call 109) (function_declaration_statement 23151) (if_stmt 23152) (if_stmt_without_else 23153) (inline_function 113) (inner_statement 23154) (interface_declaration_statement 23155) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23156) (static_member 123) (trait_declaration_statement 23157) (variable 126) (variable_class_name 127) ("{" 23158) ("~" 129)) 12998 ((";" 27724)) 12999 ((";" 27726)) 13000 ((T_EXTENDS 2987) (extends_from 27728)) 13001 ((T_IMPLEMENTS 2989) (implements_list 27729)) 13002 ((backup_doc_comment 27730)) 13003 (("{" 27731)) 13004 ((class_statement_list 27732)) 13005 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 27733)) 13006 ((";" 27735)) 13007 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 27737) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 13008 (("%" 304) (")" 27738) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 13009 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27739) (";" 17719) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 17720) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17722) (T_DEC 22) (T_DECLARE 17723) (T_DIR 24) (T_DNUMBER 25) (T_DO 17724) (T_DOUBLE_CAST 27) (T_ECHO 17725) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17727) (T_FOREACH 17728) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 17729) (T_GOTO 17730) (T_IF 17732) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17733) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17735) (T_START_HEREDOC 69) (T_STATIC 17736) (T_STRING 17737) (T_STRING_CAST 72) (T_SWITCH 17738) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17740) (T_UNSET 17741) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17742) (T_WHILE 17743) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17744) (alt_if_stmt_without_else 17745) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 17751) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 17754) (if_stmt_without_else 17755) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27740) (static_member 123) (variable 126) (variable_class_name 127) (while_statement 27741) ("{" 17760) ("~" 129)) 13010 ((inner_statement_list 27742)) 13011 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17889) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17890) (T_CLASS 17891) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17892) (T_DEC 22) (T_DECLARE 17893) (T_DIR 24) (T_DNUMBER 25) (T_DO 17894) (T_DOUBLE_CAST 27) (T_ECHO 17895) (T_EMPTY 29) (T_ENDWHILE 27743) (T_ENUM 17897) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17898) (T_FOREACH 17899) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17900) (T_GOTO 17901) (T_HALT_COMPILER 17902) (T_IF 17903) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17904) (T_INTERFACE 17905) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17906) (T_START_HEREDOC 69) (T_STATIC 17907) (T_STRING 17908) (T_STRING_CAST 72) (T_SWITCH 17909) (T_THROW 74) (T_TRAIT 17910) (T_TRAIT_C 76) (T_TRY 17911) (T_UNSET 17912) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17913) (T_WHILE 17914) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17915) (alt_if_stmt_without_else 17916) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17917) (attributes 17918) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17919) (class_modifier 99) (class_modifiers 17920) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17921) (expr 17922) (fn 106) (fully_dereferenceable 107) #'17923 (function_call 109) (function_declaration_statement 17924) (if_stmt 17925) (if_stmt_without_else 17926) (inline_function 113) (inner_statement 17927) (interface_declaration_statement 17928) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17929) (static_member 123) (trait_declaration_statement 17930) (variable 126) (variable_class_name 127) ("{" 17931) ("~" 129)) 13012 ((";" 27744)) 13013 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 27746) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 13014 (("\"" 1538) ("$" 17989) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 17990) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 17991) ("[" 1560) (array_object_dereferenceable 17992) (callable_expr 17993) (callable_variable 17994) (class_constant 1564) (class_name 17995) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 17996) (name 17997) (new_dereferenceable 1570) (simple_variable 17998) (static_member 17999) (unset_variable 18000) (unset_variables 27748) (variable 18002) (variable_class_name 18003)) 13015 (("," 18015) (possible_comma 27749)) 13016 ((")" 27750)) 13017 ((";" 27751)) 13018 ((inner_statement_list 27753)) 13019 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 27754) ("~" 129)) 13020 ((catch_list 27755)) 13021 ((T_CATCH 27756) (T_FINALLY 27757) (finally_statement 27758)) 13022 (("(" 27762)) 13023 (("{" 27759)) 13024 ((inner_statement_list 27760)) 13025 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 27761) ("~" 129)) 13026 ((T_NAME_FULLY_QUALIFIED 18056) (T_NAME_QUALIFIED 18057) (T_NAME_RELATIVE 18058) (T_STATIC 18059) (T_STRING 18060) (catch_name_list 27763) (class_name 18062) (name 18063)) 13027 ((T_VARIABLE 18064) (optional_variable 27764) ("|" 18066)) 13028 ((")" 27765)) 13029 (("{" 27766)) 13030 ((inner_statement_list 27767)) 13031 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 27768) ("~" 129)) 13032 ((backup_doc_comment 27770)) 13033 (("{" 27771)) 13034 ((class_statement_list 27772)) 13035 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 27773)) 13036 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 27775) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 13037 (("%" 304) (")" 27776) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 13038 ((":" 27777) (switch_case_list 27778) ("{" 27779)) 13039 ((";" 27785) (case_list 27786)) 13040 ((";" 27780) (case_list 27781)) 13041 ((case_list 27783)) 13042 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 27782)) 13043 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 27784)) 13044 ((case_list 27789)) 13045 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 27787)) 13046 ((";" 27788)) 13047 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 27790)) 13048 ((";" 27791)) 13049 (("," 18369) (";" 27794)) 13050 ((";" 27796)) 13051 ((T_EXTENDS 23405) (interface_extends_list 27798)) 13052 ((backup_doc_comment 27799)) 13053 (("{" 27800)) 13054 ((class_statement_list 27801)) 13055 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 27802)) 13056 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 27804) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 13057 (("%" 304) (")" 27805) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 13058 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23415) (";" 17719) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 17720) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17722) (T_DEC 22) (T_DECLARE 17723) (T_DIR 24) (T_DNUMBER 25) (T_DO 17724) (T_DOUBLE_CAST 27) (T_ECHO 17725) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17727) (T_FOREACH 17728) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 17729) (T_GOTO 17730) (T_IF 17732) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17733) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17735) (T_START_HEREDOC 69) (T_STATIC 17736) (T_STRING 17737) (T_STRING_CAST 72) (T_SWITCH 17738) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17740) (T_UNSET 17741) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17742) (T_WHILE 17743) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17744) (alt_if_stmt_without_else 17745) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 17751) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 17754) (if_stmt_without_else 17755) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27806) (static_member 123) (variable 126) (variable_class_name 127) ("{" 17760) ("~" 129)) 13059 ((")" 27808)) 13060 ((";" 27809)) 13061 ((";" 27811)) 13062 (("," 19446) (";" 27813)) 13063 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 27815) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 13064 (("%" 19534) ("*" 19535) ("+" 19536) ("-" 19537) ("." 19538) ("/" 19539) ("<" 19540) (">" 19541) ("?" 19542) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19543) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19544) (T_AS 27816) (T_BOOLEAN_AND 19545) (T_BOOLEAN_OR 19546) (T_COALESCE 19547) (T_INSTANCEOF 19548) (T_IS_EQUAL 19549) (T_IS_GREATER_OR_EQUAL 19550) (T_IS_IDENTICAL 19551) (T_IS_NOT_EQUAL 19552) (T_IS_NOT_IDENTICAL 19553) (T_IS_SMALLER_OR_EQUAL 19554) (T_LOGICAL_AND 19555) (T_LOGICAL_OR 19556) (T_LOGICAL_XOR 19557) (T_PIPE 19558) (T_POW 19559) (T_SL 19560) (T_SPACESHIP 19561) (T_SR 19562) ("^" 19563) ("|" 19564)) 13065 (("\"" 1538) ("$" 20897) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20898) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20899) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20900) ("[" 20901) (ampersand 20902) (array_object_dereferenceable 20903) (callable_expr 20904) (callable_variable 20905) (class_constant 1564) (class_name 20906) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 27817) (fully_dereferenceable 107) (function_call 20908) (name 20909) (new_dereferenceable 1570) (simple_variable 20910) (static_member 20911) (variable 20912) (variable_class_name 20913)) 13066 ((")" 27818) (T_DOUBLE_ARROW 27819)) 13067 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27822) (";" 17719) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 17720) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17722) (T_DEC 22) (T_DECLARE 17723) (T_DIR 24) (T_DNUMBER 25) (T_DO 17724) (T_DOUBLE_CAST 27) (T_ECHO 17725) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17727) (T_FOREACH 17728) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 17729) (T_GOTO 17730) (T_IF 17732) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17733) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17735) (T_START_HEREDOC 69) (T_STATIC 17736) (T_STRING 17737) (T_STRING_CAST 72) (T_SWITCH 17738) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17740) (T_UNSET 17741) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17742) (T_WHILE 17743) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17744) (alt_if_stmt_without_else 17745) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 17751) (fn 106) (foreach_statement 27828) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 17754) (if_stmt_without_else 17755) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27824) (static_member 123) (variable 126) (variable_class_name 127) ("{" 17760) ("~" 129)) 13068 (("\"" 1538) ("$" 20928) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20929) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20930) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20931) ("[" 20932) (ampersand 20933) (array_object_dereferenceable 20934) (callable_expr 20935) (callable_variable 20936) (class_constant 1564) (class_name 20937) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 27820) (fully_dereferenceable 107) (function_call 20939) (name 20940) (new_dereferenceable 1570) (simple_variable 20941) (static_member 20942) (variable 20943) (variable_class_name 20944)) 13069 ((")" 27821)) 13070 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27822) (";" 17719) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 17720) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17722) (T_DEC 22) (T_DECLARE 17723) (T_DIR 24) (T_DNUMBER 25) (T_DO 17724) (T_DOUBLE_CAST 27) (T_ECHO 17725) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17727) (T_FOREACH 17728) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 17729) (T_GOTO 17730) (T_IF 17732) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17733) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17735) (T_START_HEREDOC 69) (T_STATIC 17736) (T_STRING 17737) (T_STRING_CAST 72) (T_SWITCH 17738) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17740) (T_UNSET 17741) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17742) (T_WHILE 17743) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17744) (alt_if_stmt_without_else 17745) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 17751) (fn 106) (foreach_statement 27823) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 17754) (if_stmt_without_else 17755) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27824) (static_member 123) (variable 126) (variable_class_name 127) ("{" 17760) ("~" 129)) 13071 ((inner_statement_list 27825)) 13072 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20962) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 20963) (T_CLASS 20964) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20965) (T_DEC 22) (T_DECLARE 20966) (T_DIR 24) (T_DNUMBER 25) (T_DO 20967) (T_DOUBLE_CAST 27) (T_ECHO 20968) (T_EMPTY 29) (T_ENDFOREACH 27826) (T_ENUM 20970) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20971) (T_FOREACH 20972) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20973) (T_GOTO 20974) (T_HALT_COMPILER 20975) (T_IF 20976) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20977) (T_INTERFACE 20978) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20979) (T_START_HEREDOC 69) (T_STATIC 20980) (T_STRING 20981) (T_STRING_CAST 72) (T_SWITCH 20982) (T_THROW 74) (T_TRAIT 20983) (T_TRAIT_C 76) (T_TRY 20984) (T_UNSET 20985) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 20986) (T_WHILE 20987) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 20988) (alt_if_stmt_without_else 20989) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 20990) (attributes 20991) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 20992) (class_modifier 99) (class_modifiers 20993) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 20994) (expr 20995) (fn 106) (fully_dereferenceable 107) #'20996 (function_call 109) (function_declaration_statement 20997) (if_stmt 20998) (if_stmt_without_else 20999) (inline_function 113) (inner_statement 21000) (interface_declaration_statement 21001) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 21002) (static_member 123) (trait_declaration_statement 21003) (variable 126) (variable_class_name 127) ("{" 21004) ("~" 129)) 13073 ((";" 27827)) 13074 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21126) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21127) (fn 18437) (for_exprs 27830) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21129) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 13075 ((";" 27831)) 13076 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21135) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21136) (fn 18437) (for_cond_exprs 27832) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21138) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 13077 ((";" 27833)) 13078 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_VOID_CAST 22737) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 22738) (fn 2425) (for_exprs 27834) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_for_exprs 22740) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 13079 ((")" 27835)) 13080 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27836) (";" 17719) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 17720) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17722) (T_DEC 22) (T_DECLARE 17723) (T_DIR 24) (T_DNUMBER 25) (T_DO 17724) (T_DOUBLE_CAST 27) (T_ECHO 17725) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17727) (T_FOREACH 17728) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 17729) (T_GOTO 17730) (T_IF 17732) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17733) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17735) (T_START_HEREDOC 69) (T_STATIC 17736) (T_STRING 17737) (T_STRING_CAST 72) (T_SWITCH 17738) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17740) (T_UNSET 17741) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17742) (T_WHILE 17743) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17744) (alt_if_stmt_without_else 17745) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 17751) (fn 106) (for_statement 27837) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 17754) (if_stmt_without_else 17755) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27838) (static_member 123) (variable 126) (variable_class_name 127) ("{" 17760) ("~" 129)) 13081 ((inner_statement_list 27839)) 13082 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22750) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 22751) (T_CLASS 22752) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22753) (T_DEC 22) (T_DECLARE 22754) (T_DIR 24) (T_DNUMBER 25) (T_DO 22755) (T_DOUBLE_CAST 27) (T_ECHO 22756) (T_EMPTY 29) (T_ENDFOR 27840) (T_ENUM 22758) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 22759) (T_FOREACH 22760) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 22761) (T_GOTO 22762) (T_HALT_COMPILER 22763) (T_IF 22764) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22765) (T_INTERFACE 22766) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22767) (T_START_HEREDOC 69) (T_STATIC 22768) (T_STRING 22769) (T_STRING_CAST 72) (T_SWITCH 22770) (T_THROW 74) (T_TRAIT 22771) (T_TRAIT_C 76) (T_TRY 22772) (T_UNSET 22773) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22774) (T_WHILE 22775) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22776) (alt_if_stmt_without_else 22777) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 22778) (attributes 22779) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 22780) (class_modifier 99) (class_modifiers 22781) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 22782) (expr 22783) (fn 106) (fully_dereferenceable 107) #'22784 (function_call 109) (function_declaration_statement 22785) (if_stmt 22786) (if_stmt_without_else 22787) (inline_function 113) (inner_statement 22788) (interface_declaration_statement 22789) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 22790) (static_member 123) (trait_declaration_statement 22791) (variable 126) (variable_class_name 127) ("{" 22792) ("~" 129)) 13083 ((";" 27841)) 13084 ((":" 23454) (enum_backing_type 27843)) 13085 ((T_IMPLEMENTS 2989) (implements_list 27844)) 13086 ((backup_doc_comment 27845)) 13087 (("{" 27846)) 13088 ((class_statement_list 27847)) 13089 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 27848)) 13090 (("," 22929) (";" 27850)) 13091 ((T_WHILE 27852)) 13092 (("(" 27853)) 13093 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 27854) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 13094 (("%" 304) (")" 27855) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 13095 ((";" 27856)) 13096 ((T_STRING 23106) (const_decl 23107) (const_list 27858)) 13097 ((")" 27859) ("," 23110)) 13098 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27860) (";" 17719) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 17720) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17722) (T_DEC 22) (T_DECLARE 17723) (T_DIR 24) (T_DNUMBER 25) (T_DO 17724) (T_DOUBLE_CAST 27) (T_ECHO 17725) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17727) (T_FOREACH 17728) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 17729) (T_GOTO 17730) (T_IF 17732) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17733) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17735) (T_START_HEREDOC 69) (T_STATIC 17736) (T_STRING 17737) (T_STRING_CAST 72) (T_SWITCH 17738) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17740) (T_UNSET 17741) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17742) (T_WHILE 17743) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17744) (alt_if_stmt_without_else 17745) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (declare_statement 27861) (dereferenceable_scalar 103) (expr 17751) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 17754) (if_stmt_without_else 17755) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27862) (static_member 123) (variable 126) (variable_class_name 127) ("{" 17760) ("~" 129)) 13099 ((inner_statement_list 27863)) 13100 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 23116) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 23117) (T_CLASS 23118) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 23119) (T_DEC 22) (T_DECLARE 23120) (T_DIR 24) (T_DNUMBER 25) (T_DO 23121) (T_DOUBLE_CAST 27) (T_ECHO 23122) (T_EMPTY 29) (T_ENDDECLARE 27864) (T_ENUM 23124) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 23125) (T_FOREACH 23126) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 23127) (T_GOTO 23128) (T_HALT_COMPILER 23129) (T_IF 23130) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 23131) (T_INTERFACE 23132) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 23133) (T_START_HEREDOC 69) (T_STATIC 23134) (T_STRING 23135) (T_STRING_CAST 72) (T_SWITCH 23136) (T_THROW 74) (T_TRAIT 23137) (T_TRAIT_C 76) (T_TRY 23138) (T_UNSET 23139) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 23140) (T_WHILE 23141) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 23142) (alt_if_stmt_without_else 23143) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 23144) (attributes 23145) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 23146) (class_modifier 99) (class_modifiers 23147) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 23148) (expr 23149) (fn 106) (fully_dereferenceable 107) #'23150 (function_call 109) (function_declaration_statement 23151) (if_stmt 23152) (if_stmt_without_else 23153) (inline_function 113) (inner_statement 23154) (interface_declaration_statement 23155) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23156) (static_member 123) (trait_declaration_statement 23157) (variable 126) (variable_class_name 127) ("{" 23158) ("~" 129)) 13101 ((";" 27865)) 13102 ((";" 27867)) 13103 ((T_EXTENDS 2987) (extends_from 27869)) 13104 ((T_IMPLEMENTS 2989) (implements_list 27870)) 13105 ((backup_doc_comment 27871)) 13106 (("{" 27872)) 13107 ((class_statement_list 27873)) 13108 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 27874)) 13109 ((";" 27876)) 13110 ((inner_statement_list 27878)) 13111 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17807) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17808) (T_CLASS 17809) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17810) (T_DEC 22) (T_DECLARE 17811) (T_DIR 24) (T_DNUMBER 25) (T_DO 17812) (T_DOUBLE_CAST 27) (T_ECHO 17813) (T_EMPTY 29) (T_ENDIF 27879) (T_ENUM 17815) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17816) (T_FOREACH 17817) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17818) (T_GOTO 17819) (T_HALT_COMPILER 17820) (T_IF 17732) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17821) (T_INTERFACE 17822) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17823) (T_START_HEREDOC 69) (T_STATIC 17824) (T_STRING 17825) (T_STRING_CAST 72) (T_SWITCH 17826) (T_THROW 74) (T_TRAIT 17827) (T_TRAIT_C 76) (T_TRY 17828) (T_UNSET 17829) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17830) (T_WHILE 17831) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17832) (alt_if_stmt_without_else 17833) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17834) (attributes 17835) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17836) (class_modifier 99) (class_modifiers 17837) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17838) (expr 17839) (fn 106) (fully_dereferenceable 107) #'17840 (function_call 109) (function_declaration_statement 17841) (if_stmt 17842) (if_stmt_without_else 17843) (inline_function 113) (inner_statement 17844) (interface_declaration_statement 17845) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17846) (static_member 123) (trait_declaration_statement 17847) (variable 126) (variable_class_name 127) ("{" 17848) ("~" 129)) 13112 ((";" 27880)) 13113 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 27882) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 13114 (("%" 304) (")" 27883) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 13115 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27884) (";" 17679) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 17680) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17681) (T_DEC 22) (T_DECLARE 17682) (T_DIR 24) (T_DNUMBER 25) (T_DO 17683) (T_DOUBLE_CAST 27) (T_ECHO 17684) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17685) (T_FOREACH 17686) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 17687) (T_GOTO 17688) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17689) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17690) (T_START_HEREDOC 69) (T_STATIC 17691) (T_STRING 17692) (T_STRING_CAST 72) (T_SWITCH 17693) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17694) (T_UNSET 17695) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17696) (T_WHILE 17697) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17698) (alt_if_stmt_without_else 17699) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 17700) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 17701) (if_stmt_without_else 17702) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27885) (static_member 123) (variable 126) (variable_class_name 127) (while_statement 27886) ("{" 17704) ("~" 129)) 13116 ((inner_statement_list 27887)) 13117 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17889) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17890) (T_CLASS 17891) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17892) (T_DEC 22) (T_DECLARE 17893) (T_DIR 24) (T_DNUMBER 25) (T_DO 17894) (T_DOUBLE_CAST 27) (T_ECHO 17895) (T_EMPTY 29) (T_ENDWHILE 27888) (T_ENUM 17897) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17898) (T_FOREACH 17899) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17900) (T_GOTO 17901) (T_HALT_COMPILER 17902) (T_IF 17903) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17904) (T_INTERFACE 17905) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17906) (T_START_HEREDOC 69) (T_STATIC 17907) (T_STRING 17908) (T_STRING_CAST 72) (T_SWITCH 17909) (T_THROW 74) (T_TRAIT 17910) (T_TRAIT_C 76) (T_TRY 17911) (T_UNSET 17912) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17913) (T_WHILE 17914) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17915) (alt_if_stmt_without_else 17916) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17917) (attributes 17918) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17919) (class_modifier 99) (class_modifiers 17920) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17921) (expr 17922) (fn 106) (fully_dereferenceable 107) #'17923 (function_call 109) (function_declaration_statement 17924) (if_stmt 17925) (if_stmt_without_else 17926) (inline_function 113) (inner_statement 17927) (interface_declaration_statement 17928) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17929) (static_member 123) (trait_declaration_statement 17930) (variable 126) (variable_class_name 127) ("{" 17931) ("~" 129)) 13118 ((";" 27889)) 13119 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 27891) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 13120 (("\"" 1538) ("$" 17989) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 17990) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 17991) ("[" 1560) (array_object_dereferenceable 17992) (callable_expr 17993) (callable_variable 17994) (class_constant 1564) (class_name 17995) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 17996) (name 17997) (new_dereferenceable 1570) (simple_variable 17998) (static_member 17999) (unset_variable 18000) (unset_variables 27893) (variable 18002) (variable_class_name 18003)) 13121 (("," 18015) (possible_comma 27894)) 13122 ((")" 27895)) 13123 ((";" 27896)) 13124 ((inner_statement_list 27898)) 13125 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 27899) ("~" 129)) 13126 ((catch_list 27900)) 13127 ((T_CATCH 27901) (T_FINALLY 27902) (finally_statement 27903)) 13128 (("(" 27907)) 13129 (("{" 27904)) 13130 ((inner_statement_list 27905)) 13131 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 27906) ("~" 129)) 13132 ((T_NAME_FULLY_QUALIFIED 18056) (T_NAME_QUALIFIED 18057) (T_NAME_RELATIVE 18058) (T_STATIC 18059) (T_STRING 18060) (catch_name_list 27908) (class_name 18062) (name 18063)) 13133 ((T_VARIABLE 18064) (optional_variable 27909) ("|" 18066)) 13134 ((")" 27910)) 13135 (("{" 27911)) 13136 ((inner_statement_list 27912)) 13137 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 27913) ("~" 129)) 13138 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 27915) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 13139 (("%" 304) (")" 27916) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 13140 ((":" 27917) (switch_case_list 27918) ("{" 27919)) 13141 ((";" 27925) (case_list 27926)) 13142 ((";" 27920) (case_list 27921)) 13143 ((case_list 27923)) 13144 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 27922)) 13145 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 27924)) 13146 ((case_list 27929)) 13147 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 27927)) 13148 ((";" 27928)) 13149 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 27930)) 13150 ((";" 27931)) 13151 (("," 18369) (";" 27934)) 13152 ((";" 27936)) 13153 ((";" 27938)) 13154 (("," 19446) (";" 27940)) 13155 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 27942) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 13156 (("%" 19534) ("*" 19535) ("+" 19536) ("-" 19537) ("." 19538) ("/" 19539) ("<" 19540) (">" 19541) ("?" 19542) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19543) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19544) (T_AS 27943) (T_BOOLEAN_AND 19545) (T_BOOLEAN_OR 19546) (T_COALESCE 19547) (T_INSTANCEOF 19548) (T_IS_EQUAL 19549) (T_IS_GREATER_OR_EQUAL 19550) (T_IS_IDENTICAL 19551) (T_IS_NOT_EQUAL 19552) (T_IS_NOT_IDENTICAL 19553) (T_IS_SMALLER_OR_EQUAL 19554) (T_LOGICAL_AND 19555) (T_LOGICAL_OR 19556) (T_LOGICAL_XOR 19557) (T_PIPE 19558) (T_POW 19559) (T_SL 19560) (T_SPACESHIP 19561) (T_SR 19562) ("^" 19563) ("|" 19564)) 13157 (("\"" 1538) ("$" 20897) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20898) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20899) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20900) ("[" 20901) (ampersand 20902) (array_object_dereferenceable 20903) (callable_expr 20904) (callable_variable 20905) (class_constant 1564) (class_name 20906) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 27944) (fully_dereferenceable 107) (function_call 20908) (name 20909) (new_dereferenceable 1570) (simple_variable 20910) (static_member 20911) (variable 20912) (variable_class_name 20913)) 13158 ((")" 27945) (T_DOUBLE_ARROW 27946)) 13159 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27949) (";" 17679) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 17680) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17681) (T_DEC 22) (T_DECLARE 17682) (T_DIR 24) (T_DNUMBER 25) (T_DO 17683) (T_DOUBLE_CAST 27) (T_ECHO 17684) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17685) (T_FOREACH 17686) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 17687) (T_GOTO 17688) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17689) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17690) (T_START_HEREDOC 69) (T_STATIC 17691) (T_STRING 17692) (T_STRING_CAST 72) (T_SWITCH 17693) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17694) (T_UNSET 17695) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17696) (T_WHILE 17697) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17698) (alt_if_stmt_without_else 17699) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 17700) (fn 106) (foreach_statement 27955) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 17701) (if_stmt_without_else 17702) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27951) (static_member 123) (variable 126) (variable_class_name 127) ("{" 17704) ("~" 129)) 13160 (("\"" 1538) ("$" 20928) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20929) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20930) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20931) ("[" 20932) (ampersand 20933) (array_object_dereferenceable 20934) (callable_expr 20935) (callable_variable 20936) (class_constant 1564) (class_name 20937) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 27947) (fully_dereferenceable 107) (function_call 20939) (name 20940) (new_dereferenceable 1570) (simple_variable 20941) (static_member 20942) (variable 20943) (variable_class_name 20944)) 13161 ((")" 27948)) 13162 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27949) (";" 17679) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 17680) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17681) (T_DEC 22) (T_DECLARE 17682) (T_DIR 24) (T_DNUMBER 25) (T_DO 17683) (T_DOUBLE_CAST 27) (T_ECHO 17684) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17685) (T_FOREACH 17686) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 17687) (T_GOTO 17688) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17689) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17690) (T_START_HEREDOC 69) (T_STATIC 17691) (T_STRING 17692) (T_STRING_CAST 72) (T_SWITCH 17693) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17694) (T_UNSET 17695) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17696) (T_WHILE 17697) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17698) (alt_if_stmt_without_else 17699) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 17700) (fn 106) (foreach_statement 27950) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 17701) (if_stmt_without_else 17702) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27951) (static_member 123) (variable 126) (variable_class_name 127) ("{" 17704) ("~" 129)) 13163 ((inner_statement_list 27952)) 13164 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20962) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 20963) (T_CLASS 20964) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20965) (T_DEC 22) (T_DECLARE 20966) (T_DIR 24) (T_DNUMBER 25) (T_DO 20967) (T_DOUBLE_CAST 27) (T_ECHO 20968) (T_EMPTY 29) (T_ENDFOREACH 27953) (T_ENUM 20970) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20971) (T_FOREACH 20972) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20973) (T_GOTO 20974) (T_HALT_COMPILER 20975) (T_IF 20976) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20977) (T_INTERFACE 20978) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20979) (T_START_HEREDOC 69) (T_STATIC 20980) (T_STRING 20981) (T_STRING_CAST 72) (T_SWITCH 20982) (T_THROW 74) (T_TRAIT 20983) (T_TRAIT_C 76) (T_TRY 20984) (T_UNSET 20985) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 20986) (T_WHILE 20987) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 20988) (alt_if_stmt_without_else 20989) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 20990) (attributes 20991) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 20992) (class_modifier 99) (class_modifiers 20993) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 20994) (expr 20995) (fn 106) (fully_dereferenceable 107) #'20996 (function_call 109) (function_declaration_statement 20997) (if_stmt 20998) (if_stmt_without_else 20999) (inline_function 113) (inner_statement 21000) (interface_declaration_statement 21001) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 21002) (static_member 123) (trait_declaration_statement 21003) (variable 126) (variable_class_name 127) ("{" 21004) ("~" 129)) 13165 ((";" 27954)) 13166 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21126) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21127) (fn 18437) (for_exprs 27957) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21129) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 13167 ((";" 27958)) 13168 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21135) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21136) (fn 18437) (for_cond_exprs 27959) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21138) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 13169 ((";" 27960)) 13170 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_VOID_CAST 22737) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 22738) (fn 2425) (for_exprs 27961) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_for_exprs 22740) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 13171 ((")" 27962)) 13172 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27963) (";" 17679) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 17680) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17681) (T_DEC 22) (T_DECLARE 17682) (T_DIR 24) (T_DNUMBER 25) (T_DO 17683) (T_DOUBLE_CAST 27) (T_ECHO 17684) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17685) (T_FOREACH 17686) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 17687) (T_GOTO 17688) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17689) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17690) (T_START_HEREDOC 69) (T_STATIC 17691) (T_STRING 17692) (T_STRING_CAST 72) (T_SWITCH 17693) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17694) (T_UNSET 17695) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17696) (T_WHILE 17697) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17698) (alt_if_stmt_without_else 17699) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 17700) (fn 106) (for_statement 27964) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 17701) (if_stmt_without_else 17702) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27965) (static_member 123) (variable 126) (variable_class_name 127) ("{" 17704) ("~" 129)) 13173 ((inner_statement_list 27966)) 13174 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22750) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 22751) (T_CLASS 22752) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22753) (T_DEC 22) (T_DECLARE 22754) (T_DIR 24) (T_DNUMBER 25) (T_DO 22755) (T_DOUBLE_CAST 27) (T_ECHO 22756) (T_EMPTY 29) (T_ENDFOR 27967) (T_ENUM 22758) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 22759) (T_FOREACH 22760) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 22761) (T_GOTO 22762) (T_HALT_COMPILER 22763) (T_IF 22764) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22765) (T_INTERFACE 22766) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22767) (T_START_HEREDOC 69) (T_STATIC 22768) (T_STRING 22769) (T_STRING_CAST 72) (T_SWITCH 22770) (T_THROW 74) (T_TRAIT 22771) (T_TRAIT_C 76) (T_TRY 22772) (T_UNSET 22773) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22774) (T_WHILE 22775) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22776) (alt_if_stmt_without_else 22777) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 22778) (attributes 22779) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 22780) (class_modifier 99) (class_modifiers 22781) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 22782) (expr 22783) (fn 106) (fully_dereferenceable 107) #'22784 (function_call 109) (function_declaration_statement 22785) (if_stmt 22786) (if_stmt_without_else 22787) (inline_function 113) (inner_statement 22788) (interface_declaration_statement 22789) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 22790) (static_member 123) (trait_declaration_statement 22791) (variable 126) (variable_class_name 127) ("{" 22792) ("~" 129)) 13175 ((";" 27968)) 13176 (("," 22929) (";" 27970)) 13177 ((T_WHILE 27972)) 13178 (("(" 27973)) 13179 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 27974) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 13180 (("%" 304) (")" 27975) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 13181 ((";" 27976)) 13182 ((T_STRING 23106) (const_decl 23107) (const_list 27978)) 13183 ((")" 27979) ("," 23110)) 13184 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 27980) (";" 17679) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 17680) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17681) (T_DEC 22) (T_DECLARE 17682) (T_DIR 24) (T_DNUMBER 25) (T_DO 17683) (T_DOUBLE_CAST 27) (T_ECHO 17684) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17685) (T_FOREACH 17686) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 17687) (T_GOTO 17688) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17689) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17690) (T_START_HEREDOC 69) (T_STATIC 17691) (T_STRING 17692) (T_STRING_CAST 72) (T_SWITCH 17693) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17694) (T_UNSET 17695) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17696) (T_WHILE 17697) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17698) (alt_if_stmt_without_else 17699) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (declare_statement 27981) (dereferenceable_scalar 103) (expr 17700) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 17701) (if_stmt_without_else 17702) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 27982) (static_member 123) (variable 126) (variable_class_name 127) ("{" 17704) ("~" 129)) 13185 ((inner_statement_list 27983)) 13186 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 23116) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 23117) (T_CLASS 23118) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 23119) (T_DEC 22) (T_DECLARE 23120) (T_DIR 24) (T_DNUMBER 25) (T_DO 23121) (T_DOUBLE_CAST 27) (T_ECHO 23122) (T_EMPTY 29) (T_ENDDECLARE 27984) (T_ENUM 23124) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 23125) (T_FOREACH 23126) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 23127) (T_GOTO 23128) (T_HALT_COMPILER 23129) (T_IF 23130) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 23131) (T_INTERFACE 23132) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 23133) (T_START_HEREDOC 69) (T_STATIC 23134) (T_STRING 23135) (T_STRING_CAST 72) (T_SWITCH 23136) (T_THROW 74) (T_TRAIT 23137) (T_TRAIT_C 76) (T_TRY 23138) (T_UNSET 23139) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 23140) (T_WHILE 23141) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 23142) (alt_if_stmt_without_else 23143) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 23144) (attributes 23145) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 23146) (class_modifier 99) (class_modifiers 23147) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 23148) (expr 23149) (fn 106) (fully_dereferenceable 107) #'23150 (function_call 109) (function_declaration_statement 23151) (if_stmt 23152) (if_stmt_without_else 23153) (inline_function 113) (inner_statement 23154) (interface_declaration_statement 23155) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23156) (static_member 123) (trait_declaration_statement 23157) (variable 126) (variable_class_name 127) ("{" 23158) ("~" 129)) 13187 ((";" 27985)) 13188 ((";" 27987)) 13189 ((";" 27989)) 13190 (("$" 226) (T_ABSTRACT 16417) (T_ARRAY 16418) (T_AS 16419) (T_BREAK 16420) (T_CALLABLE 16421) (T_CASE 16422) (T_CATCH 16423) (T_CLASS 16424) (T_CLASS_C 16425) (T_CLONE 16426) (T_CONST 16427) (T_CONTINUE 16428) (T_DECLARE 16429) (T_DEFAULT 16430) (T_DIR 16431) (T_DO 16432) (T_ECHO 16433) (T_ELSE 16434) (T_ELSEIF 16435) (T_EMPTY 16436) (T_ENDDECLARE 16437) (T_ENDFOR 16438) (T_ENDFOREACH 16439) (T_ENDIF 16440) (T_ENDSWITCH 16441) (T_ENDWHILE 16442) (T_ENUM 16443) (T_EVAL 16444) (T_EXIT 16445) (T_EXTENDS 16446) (T_FILE 16447) (T_FINAL 16448) (T_FINALLY 16449) (T_FN 16450) (T_FOR 16451) (T_FOREACH 16452) (T_FUNCTION 16453) (T_FUNC_C 16454) (T_GLOBAL 16455) (T_GOTO 16456) (T_IF 16457) (T_IMPLEMENTS 16458) (T_INCLUDE 16459) (T_INCLUDE_ONCE 16460) (T_INSTANCEOF 16461) (T_INSTEADOF 16462) (T_INTERFACE 16463) (T_ISSET 16464) (T_LINE 16465) (T_LIST 16466) (T_LOGICAL_AND 16467) (T_LOGICAL_OR 16468) (T_LOGICAL_XOR 16469) (T_MATCH 16470) (T_METHOD_C 16471) (T_NAMESPACE 16472) (T_NEW 16473) (T_NS_C 16474) (T_PRINT 16475) (T_PRIVATE 16476) (T_PROPERTY_C 16477) (T_PROTECTED 16478) (T_PUBLIC 16479) (T_READONLY 16480) (T_REQUIRE 16481) (T_REQUIRE_ONCE 16482) (T_RETURN 16483) (T_STATIC 16484) (T_STRING 16485) (T_SWITCH 16486) (T_THROW 16487) (T_TRAIT 16488) (T_TRAIT_C 16489) (T_TRY 16490) (T_UNSET 16491) (T_USE 16492) (T_VAR 16493) (T_VARIABLE 274) (T_WHILE 16494) (T_YIELD 16495) (identifier 27991) (member_name 27992) (reserved_non_modifiers 16498) (semi_reserved 16499) (simple_variable 27993) ("{" 27994)) 13191 (("(" 16504) (argument_list 27997)) 13192 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 27995) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 13193 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 27996)) 13194 ((T_FN 35) (T_FUNCTION 131) (fn 288) #'289 (inline_function 28001)) 13195 (("$" 226) (T_STRING 28007) (T_VARIABLE 274) (property_name 28014) (simple_variable 28009) ("{" 28010)) 13196 (("$" 226) (T_STRING 28007) (T_VARIABLE 274) (property_name 28008) (simple_variable 28009) ("{" 28010)) 13197 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 28005) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 13198 (("]" 28006)) 13199 (("(" 16504) (argument_list 28013)) 13200 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 28011) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 13201 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 28012)) 13202 (("(" 16504) (argument_list 28015)) 13203 (("`" 28017)) 13204 (("]" 28019)) 13205 (("=" 28020)) 13206 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 28021) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 13207 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28556) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13208 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 28554) (encaps_var 5357)) 13209 (("$" 28025) (T_VARIABLE 28073) (simple_variable 28550) ("{" 28551)) 13210 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 28549) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 13211 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28548) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13212 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28547) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13213 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28546) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13214 (("(" 28543)) 13215 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28542) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13216 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28541) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13217 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28528) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (clone_argument_list 28529) (constant 28084) (dereferenceable_scalar 28085) (expr 28530) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13218 (("\"" 1538) ("$" 28330) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 28331) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 28332) ("[" 1560) (array_object_dereferenceable 28333) (callable_expr 28334) (callable_variable 28335) (class_constant 1564) (class_name 28336) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 28337) (name 28338) (new_dereferenceable 1570) (simple_variable 28339) (static_member 28340) (variable 28527) (variable_class_name 28342)) 13219 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28526) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13220 (("(" 28523)) 13221 (("(" 28520)) 13222 (("(" 28511) (argument_list 28512) (ctor_arguments 28513)) 13223 (("\"" 1538) ("$" 28330) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 28331) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 28332) ("[" 1560) (array_object_dereferenceable 28333) (callable_expr 28334) (callable_variable 28335) (class_constant 1564) (class_name 28336) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 28337) (name 28338) (new_dereferenceable 1570) (simple_variable 28339) (static_member 28340) (variable 28510) (variable_class_name 28342)) 13224 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28509) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13225 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28508) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13226 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28507) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13227 (("(" 28503)) 13228 (("(" 28498)) 13229 (("(" 28492)) 13230 (("$" 28330) ("(" 28455) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 28456) (T_NAME_QUALIFIED 28457) (T_NAME_RELATIVE 28458) (T_READONLY 2955) (T_STATIC 28459) (T_STRING 28460) (T_VARIABLE 28332) (anonymous_class 28461) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 28462) (attribute 2961) (attributes 28463) (class_modifier 2963) (class_name 28464) (class_name_reference 28465) (name 28466) (new_variable 28467) (simple_variable 28468)) 13231 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28454) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13232 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28453) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13233 (("(" 28291) (argument_list 28452)) 13234 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28451) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13235 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28450) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13236 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 28445) (T_END_HEREDOC 28446) (T_VARIABLE 2920) (encaps_list 28447) (encaps_var 2922)) 13237 ((T_FN 35) (T_FUNCTION 131) (fn 28087) #'28088 (inline_function 28444)) 13238 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28443) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13239 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28442) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13240 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28441) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13241 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28438) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13242 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28437) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13243 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 28433) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 13244 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 28431) (encaps_list 2828) (encaps_var 2829)) 13245 ((T_NULLSAFE_OBJECT_OPERATOR 28417) (T_OBJECT_OPERATOR 28418) ("[" 28419)) 13246 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 28414) (attribute 2757) (fn 28087) #'28088 (inline_function 28415)) 13247 (("(" 28291) (argument_list 28413)) 13248 ((T_PAAMAYIM_NEKUDOTAYIM 28405)) 13249 (("%" 28103) ("*" 28104) ("+" 28105) ("-" 28106) ("." 28107) ("/" 28108) ("<" 28109) (">" 28110) ("?" 28111) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 28112) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 28113) (T_BOOLEAN_AND 28114) (T_BOOLEAN_OR 28115) (T_COALESCE 28116) (T_DOUBLE_ARROW 28403) (T_INSTANCEOF 28117) (T_IS_EQUAL 28118) (T_IS_GREATER_OR_EQUAL 28119) (T_IS_IDENTICAL 28120) (T_IS_NOT_EQUAL 28121) (T_IS_NOT_IDENTICAL 28122) (T_IS_SMALLER_OR_EQUAL 28123) (T_LOGICAL_AND 28124) (T_LOGICAL_OR 28125) (T_LOGICAL_XOR 28126) (T_PIPE 28127) (T_POW 28128) (T_SL 28129) (T_SPACESHIP 28130) (T_SR 28131) ("^" 28132) ("|" 28133)) 13250 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 28392)) 13251 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 28380)) 13252 (("(" 28291) (argument_list 28379)) 13253 (("=" 28299) (T_AND_EQUAL 28300) (T_COALESCE_EQUAL 28301) (T_CONCAT_EQUAL 28302) (T_DEC 28303) (T_DIV_EQUAL 28304) (T_INC 28305) (T_MINUS_EQUAL 28306) (T_MOD_EQUAL 28307) (T_MUL_EQUAL 28308) (T_OR_EQUAL 28309) (T_PLUS_EQUAL 28310) (T_POW_EQUAL 28311) (T_SL_EQUAL 28312) (T_SR_EQUAL 28313) (T_XOR_EQUAL 28314)) 13254 ((T_PAAMAYIM_NEKUDOTAYIM 28203)) 13255 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28102) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13256 (("%" 28103) ("*" 28104) ("+" 28105) ("-" 28106) ("." 28107) ("/" 28108) ("<" 28109) (">" 28110) ("?" 28111) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 28112) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 28113) (T_BOOLEAN_AND 28114) (T_BOOLEAN_OR 28115) (T_COALESCE 28116) (T_INSTANCEOF 28117) (T_IS_EQUAL 28118) (T_IS_GREATER_OR_EQUAL 28119) (T_IS_IDENTICAL 28120) (T_IS_NOT_EQUAL 28121) (T_IS_NOT_IDENTICAL 28122) (T_IS_SMALLER_OR_EQUAL 28123) (T_LOGICAL_AND 28124) (T_LOGICAL_OR 28125) (T_LOGICAL_XOR 28126) (T_PIPE 28127) (T_POW 28128) (T_SL 28129) (T_SPACESHIP 28130) (T_SR 28131) ("^" 28132) ("|" 28133)) 13257 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28202) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13258 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28201) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13259 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28200) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13260 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28199) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13261 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28198) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13262 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28197) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13263 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28196) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13264 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28195) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13265 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 28190) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 28191) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 13266 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28189) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13267 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28188) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13268 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28187) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13269 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28186) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13270 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28185) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13271 (("$" 28150) ("(" 28151) (T_NAME_FULLY_QUALIFIED 28152) (T_NAME_QUALIFIED 28153) (T_NAME_RELATIVE 28154) (T_STATIC 28155) (T_STRING 28156) (T_VARIABLE 28157) (class_name 28158) (class_name_reference 28159) (name 28160) (new_variable 28161) (simple_variable 28162)) 13272 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28149) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13273 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28148) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13274 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28147) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13275 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28146) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13276 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28145) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13277 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28144) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13278 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28143) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13279 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28142) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13280 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28141) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13281 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28140) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13282 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28139) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13283 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28138) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13284 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28137) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13285 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28136) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13286 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28135) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13287 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28134) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13288 (("$" 28150) (T_VARIABLE 28157) (simple_variable 28181) ("{" 28182)) 13289 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 28179) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 13290 ((T_PAAMAYIM_NEKUDOTAYIM 28177)) 13291 ((T_NULLSAFE_OBJECT_OPERATOR 28163) (T_OBJECT_OPERATOR 28164) (T_PAAMAYIM_NEKUDOTAYIM 28165) ("[" 28166)) 13292 (("$" 28150) (T_STRING 28170) (T_VARIABLE 28157) (property_name 28176) (simple_variable 28172) ("{" 28173)) 13293 (("$" 28150) (T_STRING 28170) (T_VARIABLE 28157) (property_name 28171) (simple_variable 28172) ("{" 28173)) 13294 (("$" 28150) (T_VARIABLE 28157) (simple_variable 28169)) 13295 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 28167) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 13296 (("]" 28168)) 13297 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 28174) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 13298 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 28175)) 13299 (("$" 28150) (T_VARIABLE 28157) (simple_variable 28178)) 13300 (("%" 304) (")" 28180) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 13301 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 28183) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 13302 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 28184)) 13303 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28194) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13304 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 28192) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 13305 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28193) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13306 (("$" 28025) (T_ABSTRACT 28204) (T_ARRAY 28205) (T_AS 28206) (T_BREAK 28207) (T_CALLABLE 28208) (T_CASE 28209) (T_CATCH 28210) (T_CLASS 28211) (T_CLASS_C 28212) (T_CLONE 28213) (T_CONST 28214) (T_CONTINUE 28215) (T_DECLARE 28216) (T_DEFAULT 28217) (T_DIR 28218) (T_DO 28219) (T_ECHO 28220) (T_ELSE 28221) (T_ELSEIF 28222) (T_EMPTY 28223) (T_ENDDECLARE 28224) (T_ENDFOR 28225) (T_ENDFOREACH 28226) (T_ENDIF 28227) (T_ENDSWITCH 28228) (T_ENDWHILE 28229) (T_ENUM 28230) (T_EVAL 28231) (T_EXIT 28232) (T_EXTENDS 28233) (T_FILE 28234) (T_FINAL 28235) (T_FINALLY 28236) (T_FN 28237) (T_FOR 28238) (T_FOREACH 28239) (T_FUNCTION 28240) (T_FUNC_C 28241) (T_GLOBAL 28242) (T_GOTO 28243) (T_IF 28244) (T_IMPLEMENTS 28245) (T_INCLUDE 28246) (T_INCLUDE_ONCE 28247) (T_INSTANCEOF 28248) (T_INSTEADOF 28249) (T_INTERFACE 28250) (T_ISSET 28251) (T_LINE 28252) (T_LIST 28253) (T_LOGICAL_AND 28254) (T_LOGICAL_OR 28255) (T_LOGICAL_XOR 28256) (T_MATCH 28257) (T_METHOD_C 28258) (T_NAMESPACE 28259) (T_NEW 28260) (T_NS_C 28261) (T_PRINT 28262) (T_PRIVATE 28263) (T_PROPERTY_C 28264) (T_PROTECTED 28265) (T_PUBLIC 28266) (T_READONLY 28267) (T_REQUIRE 28268) (T_REQUIRE_ONCE 28269) (T_RETURN 28270) (T_STATIC 28271) (T_STRING 28272) (T_SWITCH 28273) (T_THROW 28274) (T_TRAIT 28275) (T_TRAIT_C 28276) (T_TRY 28277) (T_UNSET 28278) (T_USE 28279) (T_VAR 28280) (T_VARIABLE 28073) (T_WHILE 28281) (T_YIELD 28282) (identifier 28283) (member_name 28284) (reserved_non_modifiers 28285) (semi_reserved 28286) (simple_variable 28287) ("{" 28288)) 13307 (("(" 28291) (argument_list 28292)) 13308 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 28289) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 13309 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 28290)) 13310 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 28293) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 28294) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 28295) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 13311 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 28298) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 13312 (("," 1498) (possible_comma 28296)) 13313 ((")" 28297)) 13314 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (ampersand 28328) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28329) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13315 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28327) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13316 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28326) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13317 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28325) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13318 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28324) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13319 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28323) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13320 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28322) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13321 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28321) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13322 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28320) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13323 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28319) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13324 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28318) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13325 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28317) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13326 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28316) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13327 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28315) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13328 (("\"" 1538) ("$" 28330) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 28331) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 28332) ("[" 1560) (array_object_dereferenceable 28333) (callable_expr 28334) (callable_variable 28335) (class_constant 1564) (class_name 28336) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 28337) (name 28338) (new_dereferenceable 1570) (simple_variable 28339) (static_member 28340) (variable 28341) (variable_class_name 28342)) 13329 (("$" 28330) (T_VARIABLE 28332) (simple_variable 28375) ("{" 28376)) 13330 (("(" 28346) (argument_list 28374)) 13331 ((T_NULLSAFE_OBJECT_OPERATOR 28360) (T_OBJECT_OPERATOR 28361) ("[" 28362)) 13332 (("(" 28346) (argument_list 28359)) 13333 ((T_PAAMAYIM_NEKUDOTAYIM 28355)) 13334 (("(" 28346) (argument_list 28354)) 13335 ((T_PAAMAYIM_NEKUDOTAYIM 28343)) 13336 (("$" 28330) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 28332) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 28344) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 28345) ("{" 1660)) 13337 (("(" 28346) (argument_list 28347)) 13338 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 28348) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 28349) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 28350) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 13339 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 28353) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 13340 (("," 1498) (possible_comma 28351)) 13341 ((")" 28352)) 13342 (("$" 28330) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 28332) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 28356) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 28357) ("{" 1676)) 13343 (("(" 28346) (argument_list 28358)) 13344 (("$" 28330) (T_STRING 28365) (T_VARIABLE 28332) (property_name 28372) (simple_variable 28367) ("{" 28368)) 13345 (("$" 28330) (T_STRING 28365) (T_VARIABLE 28332) (property_name 28366) (simple_variable 28367) ("{" 28368)) 13346 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 28363) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 13347 (("]" 28364)) 13348 (("(" 28346) (argument_list 28371)) 13349 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 28369) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 13350 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 28370)) 13351 (("(" 28346) (argument_list 28373)) 13352 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 28377) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 13353 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 28378)) 13354 ((backup_doc_comment 28381)) 13355 (("(" 28382)) 13356 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 28383)) 13357 ((")" 28384)) 13358 ((T_USE 166) (lexical_vars 28385)) 13359 ((":" 168) (return_type 28386)) 13360 ((backup_fn_flags 28387)) 13361 (("{" 28388)) 13362 ((inner_statement_list 28389)) 13363 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 28390) ("~" 129)) 13364 ((backup_fn_flags 28391)) 13365 ((backup_doc_comment 28393)) 13366 (("(" 28394)) 13367 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 28395)) 13368 ((")" 28396)) 13369 ((":" 2271) (return_type 28397)) 13370 ((T_DOUBLE_ARROW 28398)) 13371 ((backup_fn_flags 28399)) 13372 ((backup_lex_pos 28400)) 13373 (("!" 17028) ("\"" 17029) ("$" 17030) ("(" 17031) ("+" 17032) ("-" 17033) ("@" 17034) (T_ARRAY 17035) (T_ARRAY_CAST 17036) (T_ATTRIBUTE 130) (T_BOOL_CAST 17037) (T_CLASS_C 17038) (T_CLONE 17039) (T_CONSTANT_ENCAPSED_STRING 17040) (T_DEC 17041) (T_DIR 17042) (T_DNUMBER 17043) (T_DOUBLE_CAST 17044) (T_EMPTY 17045) (T_EVAL 17046) (T_EXIT 17047) (T_FILE 17048) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 17049) (T_INC 17050) (T_INCLUDE 17051) (T_INCLUDE_ONCE 17052) (T_INT_CAST 17053) (T_ISSET 17054) (T_LINE 17055) (T_LIST 17056) (T_LNUMBER 17057) (T_MATCH 17058) (T_METHOD_C 17059) (T_NAME_FULLY_QUALIFIED 17060) (T_NAME_QUALIFIED 17061) (T_NAME_RELATIVE 17062) (T_NEW 17063) (T_NS_C 17064) (T_OBJECT_CAST 17065) (T_PRINT 17066) (T_PROPERTY_C 17067) (T_READONLY 17068) (T_REQUIRE 17069) (T_REQUIRE_ONCE 17070) (T_START_HEREDOC 17071) (T_STATIC 17072) (T_STRING 17073) (T_STRING_CAST 17074) (T_THROW 17075) (T_TRAIT_C 17076) (T_UNSET_CAST 17077) (T_VARIABLE 17078) (T_YIELD 17079) (T_YIELD_FROM 17080) ("[" 17081) ("`" 17082) (array_object_dereferenceable 17083) (attribute 135) (attributes 17084) (callable_expr 17085) (callable_variable 17086) (class_constant 17087) (class_name 17088) (constant 17089) (dereferenceable_scalar 17090) (expr 28401) (fn 17092) (fully_dereferenceable 107) #'17093 (function_call 17094) (inline_function 17095) (internal_functions_in_yacc 17096) (match 17097) (name 17098) (new_dereferenceable 17099) (new_non_dereferenceable 17100) (scalar 17101) (simple_variable 17102) (static_member 17103) (variable 17104) (variable_class_name 17105) ("~" 17106)) 13374 (("%" 17108) ("*" 17109) ("+" 17110) ("-" 17111) ("." 17112) ("/" 17113) ("<" 17114) (">" 17115) ("?" 17116) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17117) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17118) (T_BOOLEAN_AND 17119) (T_BOOLEAN_OR 17120) (T_COALESCE 17121) (T_INSTANCEOF 17122) (T_IS_EQUAL 17123) (T_IS_GREATER_OR_EQUAL 17124) (T_IS_IDENTICAL 17125) (T_IS_NOT_EQUAL 17126) (T_IS_NOT_IDENTICAL 17127) (T_IS_SMALLER_OR_EQUAL 17128) (T_LOGICAL_AND 17129) (T_LOGICAL_OR 17130) (T_LOGICAL_XOR 17131) (T_PIPE 17132) (T_POW 17133) (T_SL 17134) (T_SPACESHIP 17135) (T_SR 17136) ("^" 17137) (backup_fn_flags 28402) ("|" 17138)) 13375 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 28404) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 13376 (("$" 28025) (T_ABSTRACT 28204) (T_ARRAY 28205) (T_AS 28206) (T_BREAK 28207) (T_CALLABLE 28208) (T_CASE 28209) (T_CATCH 28210) (T_CLASS 28211) (T_CLASS_C 28212) (T_CLONE 28213) (T_CONST 28214) (T_CONTINUE 28215) (T_DECLARE 28216) (T_DEFAULT 28217) (T_DIR 28218) (T_DO 28219) (T_ECHO 28220) (T_ELSE 28221) (T_ELSEIF 28222) (T_EMPTY 28223) (T_ENDDECLARE 28224) (T_ENDFOR 28225) (T_ENDFOREACH 28226) (T_ENDIF 28227) (T_ENDSWITCH 28228) (T_ENDWHILE 28229) (T_ENUM 28230) (T_EVAL 28231) (T_EXIT 28232) (T_EXTENDS 28233) (T_FILE 28234) (T_FINAL 28235) (T_FINALLY 28236) (T_FN 28237) (T_FOR 28238) (T_FOREACH 28239) (T_FUNCTION 28240) (T_FUNC_C 28241) (T_GLOBAL 28242) (T_GOTO 28243) (T_IF 28244) (T_IMPLEMENTS 28245) (T_INCLUDE 28246) (T_INCLUDE_ONCE 28247) (T_INSTANCEOF 28248) (T_INSTEADOF 28249) (T_INTERFACE 28250) (T_ISSET 28251) (T_LINE 28252) (T_LIST 28253) (T_LOGICAL_AND 28254) (T_LOGICAL_OR 28255) (T_LOGICAL_XOR 28256) (T_MATCH 28257) (T_METHOD_C 28258) (T_NAMESPACE 28259) (T_NEW 28260) (T_NS_C 28261) (T_PRINT 28262) (T_PRIVATE 28263) (T_PROPERTY_C 28264) (T_PROTECTED 28265) (T_PUBLIC 28266) (T_READONLY 28267) (T_REQUIRE 28268) (T_REQUIRE_ONCE 28269) (T_RETURN 28270) (T_STATIC 28271) (T_STRING 28272) (T_SWITCH 28273) (T_THROW 28274) (T_TRAIT 28275) (T_TRAIT_C 28276) (T_TRY 28277) (T_UNSET 28278) (T_USE 28279) (T_VAR 28280) (T_VARIABLE 28073) (T_WHILE 28281) (T_YIELD 28282) (identifier 28406) (member_name 28407) (reserved_non_modifiers 28285) (semi_reserved 28286) (simple_variable 28408) ("{" 28409)) 13377 (("(" 28291) (argument_list 28412)) 13378 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 28410) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 13379 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 28411)) 13380 ((T_FN 35) (T_FUNCTION 131) (fn 28087) #'28088 (inline_function 28416)) 13381 (("$" 28025) (T_STRING 28422) (T_VARIABLE 28073) (property_name 28429) (simple_variable 28424) ("{" 28425)) 13382 (("$" 28025) (T_STRING 28422) (T_VARIABLE 28073) (property_name 28423) (simple_variable 28424) ("{" 28425)) 13383 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 28420) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 13384 (("]" 28421)) 13385 (("(" 28291) (argument_list 28428)) 13386 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 28426) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 13387 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 28427)) 13388 (("(" 28291) (argument_list 28430)) 13389 (("`" 28432)) 13390 (("]" 28434)) 13391 (("=" 28435)) 13392 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28436) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13393 (("%" 28103) ("*" 28104) ("+" 28105) ("-" 28106) ("." 28107) ("/" 28108) ("<" 28109) (">" 28110) ("?" 28111) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 28112) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 28113) (T_BOOLEAN_AND 28114) (T_BOOLEAN_OR 28115) (T_COALESCE 28116) (T_DOUBLE_ARROW 28439) (T_INSTANCEOF 28117) (T_IS_EQUAL 28118) (T_IS_GREATER_OR_EQUAL 28119) (T_IS_IDENTICAL 28120) (T_IS_NOT_EQUAL 28121) (T_IS_NOT_IDENTICAL 28122) (T_IS_SMALLER_OR_EQUAL 28123) (T_LOGICAL_AND 28124) (T_LOGICAL_OR 28125) (T_LOGICAL_XOR 28126) (T_PIPE 28127) (T_POW 28128) (T_SL 28129) (T_SPACESHIP 28130) (T_SR 28131) ("^" 28132) ("|" 28133)) 13394 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28440) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13395 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 28449) (T_VARIABLE 2920) (encaps_var 2934)) 13396 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 28448) (T_VARIABLE 2920) (encaps_var 2925)) 13397 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 28490) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 13398 ((T_CLASS 28482)) 13399 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 28481) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 28462) (attribute 2982) (class_modifier 2963)) 13400 ((T_PAAMAYIM_NEKUDOTAYIM 28479)) 13401 (("(" 28346) (argument_list 28478)) 13402 ((T_NULLSAFE_OBJECT_OPERATOR 28469) (T_OBJECT_OPERATOR 28470) (T_PAAMAYIM_NEKUDOTAYIM 28471) ("[" 28472)) 13403 (("$" 28330) (T_STRING 28365) (T_VARIABLE 28332) (property_name 28477) (simple_variable 28367) ("{" 28368)) 13404 (("$" 28330) (T_STRING 28365) (T_VARIABLE 28332) (property_name 28476) (simple_variable 28367) ("{" 28368)) 13405 (("$" 28330) (T_VARIABLE 28332) (simple_variable 28475)) 13406 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 28473) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 13407 (("]" 28474)) 13408 (("$" 28330) (T_VARIABLE 28332) (simple_variable 28480)) 13409 (("(" 2984) (argument_list 2985) (ctor_arguments 28483)) 13410 ((T_EXTENDS 2987) (extends_from 28484)) 13411 ((T_IMPLEMENTS 2989) (implements_list 28485)) 13412 ((backup_doc_comment 28486)) 13413 (("{" 28487)) 13414 ((class_statement_list 28488)) 13415 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 28489)) 13416 (("%" 304) (")" 28491) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 13417 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 28493) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 13418 (("%" 304) (")" 28494) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 13419 (("{" 28495)) 13420 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 28496) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 13421 (("}" 28497)) 13422 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 28499) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 13423 ((")" 28500)) 13424 (("=" 28501)) 13425 (("!" 28023) ("\"" 28024) ("$" 28025) ("(" 28026) ("+" 28027) ("-" 28028) ("@" 28029) (T_ARRAY 28030) (T_ARRAY_CAST 28031) (T_ATTRIBUTE 130) (T_BOOL_CAST 28032) (T_CLASS_C 28033) (T_CLONE 28034) (T_CONSTANT_ENCAPSED_STRING 28035) (T_DEC 28036) (T_DIR 28037) (T_DNUMBER 28038) (T_DOUBLE_CAST 28039) (T_EMPTY 28040) (T_EVAL 28041) (T_EXIT 28042) (T_FILE 28043) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28044) (T_INC 28045) (T_INCLUDE 28046) (T_INCLUDE_ONCE 28047) (T_INT_CAST 28048) (T_ISSET 28049) (T_LINE 28050) (T_LIST 28051) (T_LNUMBER 28052) (T_MATCH 28053) (T_METHOD_C 28054) (T_NAME_FULLY_QUALIFIED 28055) (T_NAME_QUALIFIED 28056) (T_NAME_RELATIVE 28057) (T_NEW 28058) (T_NS_C 28059) (T_OBJECT_CAST 28060) (T_PRINT 28061) (T_PROPERTY_C 28062) (T_READONLY 28063) (T_REQUIRE 28064) (T_REQUIRE_ONCE 28065) (T_START_HEREDOC 28066) (T_STATIC 28067) (T_STRING 28068) (T_STRING_CAST 28069) (T_THROW 28070) (T_TRAIT_C 28071) (T_UNSET_CAST 28072) (T_VARIABLE 28073) (T_YIELD 28074) (T_YIELD_FROM 28075) ("[" 28076) ("`" 28077) (array_object_dereferenceable 28078) (attribute 135) (attributes 28079) (callable_expr 28080) (callable_variable 28081) (class_constant 28082) (class_name 28083) (constant 28084) (dereferenceable_scalar 28085) (expr 28502) (fn 28087) (fully_dereferenceable 107) #'28088 (function_call 28089) (inline_function 28090) (internal_functions_in_yacc 28091) (match 28092) (name 28093) (new_dereferenceable 28094) (new_non_dereferenceable 28095) (scalar 28096) (simple_variable 28097) (static_member 28098) (variable 28099) (variable_class_name 28100) ("~" 28101)) 13426 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 28504) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 13427 (("," 5226) (possible_comma 28505)) 13428 ((")" 28506)) 13429 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 28514) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 28515) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 28516) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 13430 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 28519) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 13431 (("," 1498) (possible_comma 28517)) 13432 ((")" 28518)) 13433 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 28521) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 13434 (("%" 304) (")" 28522) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 13435 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 28524) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 13436 (("%" 304) (")" 28525) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 13437 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 28531) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 28532) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 28533) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 28534) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 13438 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 28540) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 13439 (("%" 2442) (")" 28537) ("*" 2443) ("+" 2444) ("," 28538) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 13440 (("," 5287) (possible_comma 28535)) 13441 ((")" 28536)) 13442 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 28539) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 13443 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 28544) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 13444 ((")" 28545)) 13445 (("%" 304) (")" 28537) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 13446 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 28552) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 13447 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 28553)) 13448 (("\"" 28555) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 13449 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 28565) (T_VARIABLE 2920) (encaps_var 2934)) 13450 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 28564) (T_VARIABLE 2920) (encaps_var 2925)) 13451 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 28606) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 13452 ((T_CLASS 28598)) 13453 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 28597) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 28578) (attribute 2982) (class_modifier 2963)) 13454 ((T_PAAMAYIM_NEKUDOTAYIM 28595)) 13455 (("(" 16559) (argument_list 28594)) 13456 ((T_NULLSAFE_OBJECT_OPERATOR 28585) (T_OBJECT_OPERATOR 28586) (T_PAAMAYIM_NEKUDOTAYIM 28587) ("[" 28588)) 13457 (("$" 16543) (T_STRING 16578) (T_VARIABLE 16545) (property_name 28593) (simple_variable 16580) ("{" 16581)) 13458 (("$" 16543) (T_STRING 16578) (T_VARIABLE 16545) (property_name 28592) (simple_variable 16580) ("{" 16581)) 13459 (("$" 16543) (T_VARIABLE 16545) (simple_variable 28591)) 13460 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 28589) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 13461 (("]" 28590)) 13462 (("$" 16543) (T_VARIABLE 16545) (simple_variable 28596)) 13463 (("(" 2984) (argument_list 2985) (ctor_arguments 28599)) 13464 ((T_EXTENDS 2987) (extends_from 28600)) 13465 ((T_IMPLEMENTS 2989) (implements_list 28601)) 13466 ((backup_doc_comment 28602)) 13467 (("{" 28603)) 13468 ((class_statement_list 28604)) 13469 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 28605)) 13470 (("%" 304) (")" 28607) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 13471 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 28609) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 13472 (("%" 304) (")" 28610) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 13473 (("{" 28611)) 13474 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 28612) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 13475 (("}" 28613)) 13476 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 28615) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 13477 ((")" 28616)) 13478 (("=" 28617)) 13479 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 28618) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 13480 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 28620) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 13481 (("," 5226) (possible_comma 28621)) 13482 ((")" 28622)) 13483 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 28630) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 28631) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 28632) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 13484 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 28635) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 13485 (("," 1498) (possible_comma 28633)) 13486 ((")" 28634)) 13487 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 28637) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 13488 (("%" 304) (")" 28638) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 13489 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 28640) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 13490 (("%" 304) (")" 28641) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 13491 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 28647) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 28648) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 28649) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 28650) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 13492 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 28656) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 13493 (("%" 2442) (")" 28653) ("*" 2443) ("+" 2444) ("," 28654) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 13494 (("," 5287) (possible_comma 28651)) 13495 ((")" 28652)) 13496 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 28655) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 13497 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 28660) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 13498 ((")" 28661)) 13499 (("%" 304) (")" 28653) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 13500 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 28668) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 13501 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 28669)) 13502 (("\"" 28671) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 13503 ((T_READONLY 17774) (T_STRING 17775) (backup_doc_comment 143) (function_name 28675)) 13504 ((backup_doc_comment 28676)) 13505 (("(" 28677)) 13506 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 28678)) 13507 ((")" 28679)) 13508 ((":" 168) (return_type 28680)) 13509 ((backup_fn_flags 28681)) 13510 (("{" 28682)) 13511 ((inner_statement_list 28683)) 13512 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 28684) ("~" 129)) 13513 ((backup_fn_flags 28685)) 13514 ((T_STRING 28688)) 13515 ((T_EXTENDS 2987) (extends_from 28689)) 13516 ((T_IMPLEMENTS 2989) (implements_list 28690)) 13517 ((backup_doc_comment 28691)) 13518 (("{" 28692)) 13519 ((class_statement_list 28693)) 13520 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 28694)) 13521 ((":" 28699)) 13522 ((";" 28698)) 13523 ((inner_statement_list 28700)) 13524 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17807) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17808) (T_CLASS 17809) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17810) (T_DEC 22) (T_DECLARE 17811) (T_DIR 24) (T_DNUMBER 25) (T_DO 17812) (T_DOUBLE_CAST 27) (T_ECHO 17813) (T_EMPTY 29) (T_ENDIF 28701) (T_ENUM 17815) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17816) (T_FOREACH 17817) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17818) (T_GOTO 17819) (T_HALT_COMPILER 17820) (T_IF 17732) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17821) (T_INTERFACE 17822) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17823) (T_START_HEREDOC 69) (T_STATIC 17824) (T_STRING 17825) (T_STRING_CAST 72) (T_SWITCH 17826) (T_THROW 74) (T_TRAIT 17827) (T_TRAIT_C 76) (T_TRY 17828) (T_UNSET 17829) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17830) (T_WHILE 17831) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17832) (alt_if_stmt_without_else 17833) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17834) (attributes 17835) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17836) (class_modifier 99) (class_modifiers 17837) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17838) (expr 17839) (fn 106) (fully_dereferenceable 107) #'17840 (function_call 109) (function_declaration_statement 17841) (if_stmt 17842) (if_stmt_without_else 17843) (inline_function 113) (inner_statement 17844) (interface_declaration_statement 17845) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17846) (static_member 123) (trait_declaration_statement 17847) (variable 126) (variable_class_name 127) ("{" 17848) ("~" 129)) 13525 ((";" 28702)) 13526 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 28704) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 13527 (("%" 304) (")" 28705) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 13528 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 28706) (";" 173) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 207) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 28707) (static_member 123) (variable 126) (variable_class_name 127) (while_statement 28708) ("{" 216) ("~" 129)) 13529 ((inner_statement_list 28709)) 13530 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17889) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17890) (T_CLASS 17891) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17892) (T_DEC 22) (T_DECLARE 17893) (T_DIR 24) (T_DNUMBER 25) (T_DO 17894) (T_DOUBLE_CAST 27) (T_ECHO 17895) (T_EMPTY 29) (T_ENDWHILE 28710) (T_ENUM 17897) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17898) (T_FOREACH 17899) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17900) (T_GOTO 17901) (T_HALT_COMPILER 17902) (T_IF 17903) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17904) (T_INTERFACE 17905) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17906) (T_START_HEREDOC 69) (T_STATIC 17907) (T_STRING 17908) (T_STRING_CAST 72) (T_SWITCH 17909) (T_THROW 74) (T_TRAIT 17910) (T_TRAIT_C 76) (T_TRY 17911) (T_UNSET 17912) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17913) (T_WHILE 17914) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17915) (alt_if_stmt_without_else 17916) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17917) (attributes 17918) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17919) (class_modifier 99) (class_modifiers 17920) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17921) (expr 17922) (fn 106) (fully_dereferenceable 107) #'17923 (function_call 109) (function_declaration_statement 17924) (if_stmt 17925) (if_stmt_without_else 17926) (inline_function 113) (inner_statement 17927) (interface_declaration_statement 17928) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17929) (static_member 123) (trait_declaration_statement 17930) (variable 126) (variable_class_name 127) ("{" 17931) ("~" 129)) 13531 ((";" 28711)) 13532 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 28713) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 13533 (("\"" 1538) ("$" 17989) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 17990) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 17991) ("[" 1560) (array_object_dereferenceable 17992) (callable_expr 17993) (callable_variable 17994) (class_constant 1564) (class_name 17995) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 17996) (name 17997) (new_dereferenceable 1570) (simple_variable 17998) (static_member 17999) (unset_variable 18000) (unset_variables 28715) (variable 18002) (variable_class_name 18003)) 13534 (("," 18015) (possible_comma 28716)) 13535 ((")" 28717)) 13536 ((";" 28718)) 13537 ((inner_statement_list 28720)) 13538 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 28721) ("~" 129)) 13539 ((catch_list 28722)) 13540 ((T_CATCH 28723) (T_FINALLY 28724) (finally_statement 28725)) 13541 (("(" 28729)) 13542 (("{" 28726)) 13543 ((inner_statement_list 28727)) 13544 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 28728) ("~" 129)) 13545 ((T_NAME_FULLY_QUALIFIED 18056) (T_NAME_QUALIFIED 18057) (T_NAME_RELATIVE 18058) (T_STATIC 18059) (T_STRING 18060) (catch_name_list 28730) (class_name 18062) (name 18063)) 13546 ((T_VARIABLE 18064) (optional_variable 28731) ("|" 18066)) 13547 ((")" 28732)) 13548 (("{" 28733)) 13549 ((inner_statement_list 28734)) 13550 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 28735) ("~" 129)) 13551 ((backup_doc_comment 28737)) 13552 (("{" 28738)) 13553 ((class_statement_list 28739)) 13554 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 28740)) 13555 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 28742) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 13556 (("%" 304) (")" 28743) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 13557 ((":" 28744) (switch_case_list 28745) ("{" 28746)) 13558 ((";" 28752) (case_list 28753)) 13559 ((";" 28747) (case_list 28748)) 13560 ((case_list 28750)) 13561 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 28749)) 13562 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 28751)) 13563 ((case_list 28756)) 13564 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 28754)) 13565 ((";" 28755)) 13566 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 28757)) 13567 ((";" 28758)) 13568 (("," 18369) (";" 28761)) 13569 ((";" 28763)) 13570 ((T_EXTENDS 23405) (interface_extends_list 28765)) 13571 ((backup_doc_comment 28766)) 13572 (("{" 28767)) 13573 ((class_statement_list 28768)) 13574 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 28769)) 13575 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 28771) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 13576 (("%" 304) (")" 28772) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 13577 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23415) (";" 17679) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 17680) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17681) (T_DEC 22) (T_DECLARE 17682) (T_DIR 24) (T_DNUMBER 25) (T_DO 17683) (T_DOUBLE_CAST 27) (T_ECHO 17684) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 17685) (T_FOREACH 17686) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 17687) (T_GOTO 17688) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17689) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17690) (T_START_HEREDOC 69) (T_STATIC 17691) (T_STRING 17692) (T_STRING_CAST 72) (T_SWITCH 17693) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 17694) (T_UNSET 17695) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17696) (T_WHILE 17697) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17698) (alt_if_stmt_without_else 17699) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 17700) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 17701) (if_stmt_without_else 17702) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 28773) (static_member 123) (variable 126) (variable_class_name 127) ("{" 17704) ("~" 129)) 13578 ((")" 28775)) 13579 ((";" 28776)) 13580 ((";" 28778)) 13581 (("," 19446) (";" 28780)) 13582 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 28782) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 13583 (("%" 19534) ("*" 19535) ("+" 19536) ("-" 19537) ("." 19538) ("/" 19539) ("<" 19540) (">" 19541) ("?" 19542) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19543) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19544) (T_AS 28783) (T_BOOLEAN_AND 19545) (T_BOOLEAN_OR 19546) (T_COALESCE 19547) (T_INSTANCEOF 19548) (T_IS_EQUAL 19549) (T_IS_GREATER_OR_EQUAL 19550) (T_IS_IDENTICAL 19551) (T_IS_NOT_EQUAL 19552) (T_IS_NOT_IDENTICAL 19553) (T_IS_SMALLER_OR_EQUAL 19554) (T_LOGICAL_AND 19555) (T_LOGICAL_OR 19556) (T_LOGICAL_XOR 19557) (T_PIPE 19558) (T_POW 19559) (T_SL 19560) (T_SPACESHIP 19561) (T_SR 19562) ("^" 19563) ("|" 19564)) 13584 (("\"" 1538) ("$" 20897) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20898) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20899) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20900) ("[" 20901) (ampersand 20902) (array_object_dereferenceable 20903) (callable_expr 20904) (callable_variable 20905) (class_constant 1564) (class_name 20906) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 28784) (fully_dereferenceable 107) (function_call 20908) (name 20909) (new_dereferenceable 1570) (simple_variable 20910) (static_member 20911) (variable 20912) (variable_class_name 20913)) 13585 ((")" 28785) (T_DOUBLE_ARROW 28786)) 13586 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 28789) (";" 173) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 207) (fn 106) (foreach_statement 28795) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 28791) (static_member 123) (variable 126) (variable_class_name 127) ("{" 216) ("~" 129)) 13587 (("\"" 1538) ("$" 20928) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20929) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20930) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20931) ("[" 20932) (ampersand 20933) (array_object_dereferenceable 20934) (callable_expr 20935) (callable_variable 20936) (class_constant 1564) (class_name 20937) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 28787) (fully_dereferenceable 107) (function_call 20939) (name 20940) (new_dereferenceable 1570) (simple_variable 20941) (static_member 20942) (variable 20943) (variable_class_name 20944)) 13588 ((")" 28788)) 13589 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 28789) (";" 173) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 207) (fn 106) (foreach_statement 28790) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 28791) (static_member 123) (variable 126) (variable_class_name 127) ("{" 216) ("~" 129)) 13590 ((inner_statement_list 28792)) 13591 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20962) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 20963) (T_CLASS 20964) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20965) (T_DEC 22) (T_DECLARE 20966) (T_DIR 24) (T_DNUMBER 25) (T_DO 20967) (T_DOUBLE_CAST 27) (T_ECHO 20968) (T_EMPTY 29) (T_ENDFOREACH 28793) (T_ENUM 20970) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20971) (T_FOREACH 20972) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20973) (T_GOTO 20974) (T_HALT_COMPILER 20975) (T_IF 20976) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20977) (T_INTERFACE 20978) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20979) (T_START_HEREDOC 69) (T_STATIC 20980) (T_STRING 20981) (T_STRING_CAST 72) (T_SWITCH 20982) (T_THROW 74) (T_TRAIT 20983) (T_TRAIT_C 76) (T_TRY 20984) (T_UNSET 20985) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 20986) (T_WHILE 20987) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 20988) (alt_if_stmt_without_else 20989) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 20990) (attributes 20991) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 20992) (class_modifier 99) (class_modifiers 20993) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 20994) (expr 20995) (fn 106) (fully_dereferenceable 107) #'20996 (function_call 109) (function_declaration_statement 20997) (if_stmt 20998) (if_stmt_without_else 20999) (inline_function 113) (inner_statement 21000) (interface_declaration_statement 21001) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 21002) (static_member 123) (trait_declaration_statement 21003) (variable 126) (variable_class_name 127) ("{" 21004) ("~" 129)) 13592 ((";" 28794)) 13593 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21126) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21127) (fn 18437) (for_exprs 28797) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21129) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 13594 ((";" 28798)) 13595 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21135) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21136) (fn 18437) (for_cond_exprs 28799) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21138) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 13596 ((";" 28800)) 13597 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_VOID_CAST 22737) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 22738) (fn 2425) (for_exprs 28801) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_for_exprs 22740) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 13598 ((")" 28802)) 13599 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 28803) (";" 173) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 207) (fn 106) (for_statement 28804) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 28805) (static_member 123) (variable 126) (variable_class_name 127) ("{" 216) ("~" 129)) 13600 ((inner_statement_list 28806)) 13601 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22750) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 22751) (T_CLASS 22752) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22753) (T_DEC 22) (T_DECLARE 22754) (T_DIR 24) (T_DNUMBER 25) (T_DO 22755) (T_DOUBLE_CAST 27) (T_ECHO 22756) (T_EMPTY 29) (T_ENDFOR 28807) (T_ENUM 22758) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 22759) (T_FOREACH 22760) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 22761) (T_GOTO 22762) (T_HALT_COMPILER 22763) (T_IF 22764) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22765) (T_INTERFACE 22766) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22767) (T_START_HEREDOC 69) (T_STATIC 22768) (T_STRING 22769) (T_STRING_CAST 72) (T_SWITCH 22770) (T_THROW 74) (T_TRAIT 22771) (T_TRAIT_C 76) (T_TRY 22772) (T_UNSET 22773) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22774) (T_WHILE 22775) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22776) (alt_if_stmt_without_else 22777) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 22778) (attributes 22779) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 22780) (class_modifier 99) (class_modifiers 22781) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 22782) (expr 22783) (fn 106) (fully_dereferenceable 107) #'22784 (function_call 109) (function_declaration_statement 22785) (if_stmt 22786) (if_stmt_without_else 22787) (inline_function 113) (inner_statement 22788) (interface_declaration_statement 22789) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 22790) (static_member 123) (trait_declaration_statement 22791) (variable 126) (variable_class_name 127) ("{" 22792) ("~" 129)) 13602 ((";" 28808)) 13603 ((":" 23454) (enum_backing_type 28810)) 13604 ((T_IMPLEMENTS 2989) (implements_list 28811)) 13605 ((backup_doc_comment 28812)) 13606 (("{" 28813)) 13607 ((class_statement_list 28814)) 13608 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 28815)) 13609 (("," 22929) (";" 28817)) 13610 ((T_WHILE 28819)) 13611 (("(" 28820)) 13612 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 28821) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 13613 (("%" 304) (")" 28822) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 13614 ((";" 28823)) 13615 ((T_STRING 23106) (const_decl 23107) (const_list 28825)) 13616 ((")" 28826) ("," 23110)) 13617 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 28827) (";" 173) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (declare_statement 28828) (dereferenceable_scalar 103) (expr 207) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 28829) (static_member 123) (variable 126) (variable_class_name 127) ("{" 216) ("~" 129)) 13618 ((inner_statement_list 28830)) 13619 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 23116) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 23117) (T_CLASS 23118) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 23119) (T_DEC 22) (T_DECLARE 23120) (T_DIR 24) (T_DNUMBER 25) (T_DO 23121) (T_DOUBLE_CAST 27) (T_ECHO 23122) (T_EMPTY 29) (T_ENDDECLARE 28831) (T_ENUM 23124) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 23125) (T_FOREACH 23126) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 23127) (T_GOTO 23128) (T_HALT_COMPILER 23129) (T_IF 23130) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 23131) (T_INTERFACE 23132) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 23133) (T_START_HEREDOC 69) (T_STATIC 23134) (T_STRING 23135) (T_STRING_CAST 72) (T_SWITCH 23136) (T_THROW 74) (T_TRAIT 23137) (T_TRAIT_C 76) (T_TRY 23138) (T_UNSET 23139) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 23140) (T_WHILE 23141) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 23142) (alt_if_stmt_without_else 23143) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 23144) (attributes 23145) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 23146) (class_modifier 99) (class_modifiers 23147) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 23148) (expr 23149) (fn 106) (fully_dereferenceable 107) #'23150 (function_call 109) (function_declaration_statement 23151) (if_stmt 23152) (if_stmt_without_else 23153) (inline_function 113) (inner_statement 23154) (interface_declaration_statement 23155) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23156) (static_member 123) (trait_declaration_statement 23157) (variable 126) (variable_class_name 127) ("{" 23158) ("~" 129)) 13620 ((";" 28832)) 13621 ((";" 28834)) 13622 ((T_EXTENDS 2987) (extends_from 28836)) 13623 ((T_IMPLEMENTS 2989) (implements_list 28837)) 13624 ((backup_doc_comment 28838)) 13625 (("{" 28839)) 13626 ((class_statement_list 28840)) 13627 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 28841)) 13628 ((";" 28843)) 13629 (("," 6013) (possible_comma 28845)) 13630 (("]" 28846)) 13631 ((T_ARRAY 28892) (T_CALLABLE 28893) (T_NAME_FULLY_QUALIFIED 28894) (T_NAME_QUALIFIED 28895) (T_NAME_RELATIVE 28896) (T_STATIC 28897) (T_STRING 28898) (name 28899) (type 28900) (type_without_static 28901)) 13632 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 28890)) 13633 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 28879)) 13634 (("|" 28877)) 13635 (("|" 28862)) 13636 (("(" 28863) (T_ARRAY 28864) (T_CALLABLE 28865) (T_NAME_FULLY_QUALIFIED 28866) (T_NAME_QUALIFIED 28867) (T_NAME_RELATIVE 28868) (T_STATIC 28869) (T_STRING 28870) (name 28871) (type 28872) (type_without_static 28873) (union_type_element 28874)) 13637 ((T_ARRAY 2307) (T_CALLABLE 2308) (T_NAME_FULLY_QUALIFIED 2309) (T_NAME_QUALIFIED 2310) (T_NAME_RELATIVE 2311) (T_STATIC 2312) (T_STRING 2313) (intersection_type 28875) (name 2315) (type 2316) (type_without_static 2317)) 13638 ((")" 28876) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2330)) 13639 (("(" 28863) (T_ARRAY 28864) (T_CALLABLE 28865) (T_NAME_FULLY_QUALIFIED 28866) (T_NAME_QUALIFIED 28867) (T_NAME_RELATIVE 28868) (T_STATIC 28869) (T_STRING 28870) (name 28871) (type 28872) (type_without_static 28873) (union_type_element 28878)) 13640 ((T_ARRAY 28880) (T_CALLABLE 28881) (T_NAME_FULLY_QUALIFIED 28882) (T_NAME_QUALIFIED 28883) (T_NAME_RELATIVE 28884) (T_STATIC 28885) (T_STRING 28886) (name 28887) (type 28888) (type_without_static 28889)) 13641 ((T_ARRAY 28880) (T_CALLABLE 28881) (T_NAME_FULLY_QUALIFIED 28882) (T_NAME_QUALIFIED 28883) (T_NAME_RELATIVE 28884) (T_STATIC 28885) (T_STRING 28886) (name 28887) (type 28891) (type_without_static 28889)) 13642 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 28903) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 28904) (T_VARIABLE 28905) (ampersand 28906) (lexical_var 28907) (lexical_var_list 28908)) 13643 ((T_VARIABLE 28913)) 13644 (("," 28909) (possible_comma 28910)) 13645 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 28903) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 28904) (T_VARIABLE 28905) (ampersand 28906) (lexical_var 28912)) 13646 ((")" 28911)) 13647 ((T_ARRAY 30029) (T_CALLABLE 30030) (T_NAME_FULLY_QUALIFIED 30031) (T_NAME_QUALIFIED 30032) (T_NAME_RELATIVE 30033) (T_STRING 30034) (name 30035) (type_without_static 30036)) 13648 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 30027)) 13649 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 28952) (is_reference 28953)) 13650 ((T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 28943)) 13651 (("|" 28941)) 13652 (("|" 28928)) 13653 (("(" 28929) (T_ARRAY 28930) (T_CALLABLE 28931) (T_NAME_FULLY_QUALIFIED 28932) (T_NAME_QUALIFIED 28933) (T_NAME_RELATIVE 28934) (T_STRING 28935) (name 28936) (type_without_static 28937) (union_type_without_static_element 28938)) 13654 ((T_ARRAY 2307) (T_CALLABLE 2308) (T_NAME_FULLY_QUALIFIED 2309) (T_NAME_QUALIFIED 2310) (T_NAME_RELATIVE 2311) (T_STRING 2313) (intersection_type_without_static 28939) (name 2315) (type_without_static 3046)) 13655 ((")" 28940) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3050)) 13656 (("(" 28929) (T_ARRAY 28930) (T_CALLABLE 28931) (T_NAME_FULLY_QUALIFIED 28932) (T_NAME_QUALIFIED 28933) (T_NAME_RELATIVE 28934) (T_STRING 28935) (name 28936) (type_without_static 28937) (union_type_without_static_element 28942)) 13657 ((T_ARRAY 28944) (T_CALLABLE 28945) (T_NAME_FULLY_QUALIFIED 28946) (T_NAME_QUALIFIED 28947) (T_NAME_RELATIVE 28948) (T_STRING 28949) (name 28950) (type_without_static 28951)) 13658 ((T_ELLIPSIS 28954) (is_variadic 28955)) 13659 ((T_VARIABLE 28956)) 13660 ((backup_doc_comment 28957)) 13661 (("=" 28958) (optional_property_hook_list 28959) ("{" 28960)) 13662 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29026) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13663 ((property_hook_list 28961)) 13664 ((T_ABSTRACT 5855) (T_ATTRIBUTE 5856) (T_FINAL 5857) (T_PRIVATE 5858) (T_PRIVATE_SET 5859) (T_PROTECTED 5860) (T_PROTECTED_SET 5861) (T_PUBLIC 5862) (T_PUBLIC_SET 5863) (T_READONLY 5864) (T_STATIC 5865) (attribute 5866) (attributes 5867) (member_modifier 5868) (non_empty_member_modifiers 5869) (property_hook 5870) (property_hook_modifiers 5871) ("}" 28962)) 13665 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 30026) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13666 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 30024) (encaps_var 5357)) 13667 (("$" 28965) (T_VARIABLE 29013) (simple_variable 30020) ("{" 30021)) 13668 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 30019) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 13669 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 30018) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13670 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 30017) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13671 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 30016) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13672 (("(" 30013)) 13673 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 30012) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13674 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 30011) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13675 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 29998) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (clone_argument_list 29999) (constant 29024) (dereferenceable_scalar 29025) (expr 30000) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13676 (("\"" 1538) ("$" 29270) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 29271) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 29272) ("[" 1560) (array_object_dereferenceable 29273) (callable_expr 29274) (callable_variable 29275) (class_constant 1564) (class_name 29276) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 29277) (name 29278) (new_dereferenceable 1570) (simple_variable 29279) (static_member 29280) (variable 29997) (variable_class_name 29282)) 13677 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29996) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13678 (("(" 29993)) 13679 (("(" 29990)) 13680 (("(" 29981) (argument_list 29982) (ctor_arguments 29983)) 13681 (("\"" 1538) ("$" 29270) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 29271) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 29272) ("[" 1560) (array_object_dereferenceable 29273) (callable_expr 29274) (callable_variable 29275) (class_constant 1564) (class_name 29276) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 29277) (name 29278) (new_dereferenceable 1570) (simple_variable 29279) (static_member 29280) (variable 29980) (variable_class_name 29282)) 13682 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29979) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13683 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29978) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13684 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29977) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13685 (("(" 29973)) 13686 (("(" 29968)) 13687 (("(" 29962)) 13688 (("$" 29270) ("(" 29925) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 29926) (T_NAME_QUALIFIED 29927) (T_NAME_RELATIVE 29928) (T_READONLY 2955) (T_STATIC 29929) (T_STRING 29930) (T_VARIABLE 29272) (anonymous_class 29931) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 29932) (attribute 2961) (attributes 29933) (class_modifier 2963) (class_name 29934) (class_name_reference 29935) (name 29936) (new_variable 29937) (simple_variable 29938)) 13689 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29924) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13690 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29923) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13691 (("(" 29231) (argument_list 29922)) 13692 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29921) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13693 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29920) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13694 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 29915) (T_END_HEREDOC 29916) (T_VARIABLE 2920) (encaps_list 29917) (encaps_var 2922)) 13695 ((T_FN 35) (T_FUNCTION 131) (fn 29027) #'29028 (inline_function 29914)) 13696 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29913) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13697 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29912) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13698 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29911) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13699 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29440) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13700 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29376) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13701 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 29372) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 13702 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 29370) (encaps_list 2828) (encaps_var 2829)) 13703 ((T_NULLSAFE_OBJECT_OPERATOR 29356) (T_OBJECT_OPERATOR 29357) ("[" 29358)) 13704 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 29353) (attribute 2757) (fn 29027) #'29028 (inline_function 29354)) 13705 (("(" 29231) (argument_list 29352)) 13706 ((T_PAAMAYIM_NEKUDOTAYIM 29344)) 13707 (("%" 29043) ("*" 29044) ("+" 29045) ("-" 29046) ("." 29047) ("/" 29048) ("<" 29049) (">" 29050) ("?" 29051) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 29052) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 29053) (T_BOOLEAN_AND 29054) (T_BOOLEAN_OR 29055) (T_COALESCE 29056) (T_INSTANCEOF 29057) (T_IS_EQUAL 29058) (T_IS_GREATER_OR_EQUAL 29059) (T_IS_IDENTICAL 29060) (T_IS_NOT_EQUAL 29061) (T_IS_NOT_IDENTICAL 29062) (T_IS_SMALLER_OR_EQUAL 29063) (T_LOGICAL_AND 29064) (T_LOGICAL_OR 29065) (T_LOGICAL_XOR 29066) (T_PIPE 29067) (T_POW 29068) (T_SL 29069) (T_SPACESHIP 29070) (T_SR 29071) ("^" 29072) (optional_property_hook_list 29343) ("{" 28960) ("|" 29073)) 13708 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 29332)) 13709 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 29320)) 13710 (("(" 29231) (argument_list 29319)) 13711 (("=" 29239) (T_AND_EQUAL 29240) (T_COALESCE_EQUAL 29241) (T_CONCAT_EQUAL 29242) (T_DEC 29243) (T_DIV_EQUAL 29244) (T_INC 29245) (T_MINUS_EQUAL 29246) (T_MOD_EQUAL 29247) (T_MUL_EQUAL 29248) (T_OR_EQUAL 29249) (T_PLUS_EQUAL 29250) (T_POW_EQUAL 29251) (T_SL_EQUAL 29252) (T_SR_EQUAL 29253) (T_XOR_EQUAL 29254)) 13712 ((T_PAAMAYIM_NEKUDOTAYIM 29143)) 13713 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29042) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13714 (("%" 29043) ("*" 29044) ("+" 29045) ("-" 29046) ("." 29047) ("/" 29048) ("<" 29049) (">" 29050) ("?" 29051) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 29052) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 29053) (T_BOOLEAN_AND 29054) (T_BOOLEAN_OR 29055) (T_COALESCE 29056) (T_INSTANCEOF 29057) (T_IS_EQUAL 29058) (T_IS_GREATER_OR_EQUAL 29059) (T_IS_IDENTICAL 29060) (T_IS_NOT_EQUAL 29061) (T_IS_NOT_IDENTICAL 29062) (T_IS_SMALLER_OR_EQUAL 29063) (T_LOGICAL_AND 29064) (T_LOGICAL_OR 29065) (T_LOGICAL_XOR 29066) (T_PIPE 29067) (T_POW 29068) (T_SL 29069) (T_SPACESHIP 29070) (T_SR 29071) ("^" 29072) ("|" 29073)) 13715 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29142) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13716 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29141) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13717 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29140) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13718 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29139) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13719 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29138) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13720 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29137) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13721 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29136) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13722 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29135) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13723 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 29130) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 29131) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 13724 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29129) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13725 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29128) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13726 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29127) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13727 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29126) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13728 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29125) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13729 (("$" 29090) ("(" 29091) (T_NAME_FULLY_QUALIFIED 29092) (T_NAME_QUALIFIED 29093) (T_NAME_RELATIVE 29094) (T_STATIC 29095) (T_STRING 29096) (T_VARIABLE 29097) (class_name 29098) (class_name_reference 29099) (name 29100) (new_variable 29101) (simple_variable 29102)) 13730 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29089) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13731 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29088) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13732 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29087) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13733 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29086) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13734 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29085) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13735 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29084) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13736 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29083) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13737 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29082) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13738 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29081) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13739 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29080) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13740 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29079) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13741 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29078) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13742 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29077) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13743 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29076) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13744 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29075) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13745 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29074) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13746 (("$" 29090) (T_VARIABLE 29097) (simple_variable 29121) ("{" 29122)) 13747 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 29119) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 13748 ((T_PAAMAYIM_NEKUDOTAYIM 29117)) 13749 ((T_NULLSAFE_OBJECT_OPERATOR 29103) (T_OBJECT_OPERATOR 29104) (T_PAAMAYIM_NEKUDOTAYIM 29105) ("[" 29106)) 13750 (("$" 29090) (T_STRING 29110) (T_VARIABLE 29097) (property_name 29116) (simple_variable 29112) ("{" 29113)) 13751 (("$" 29090) (T_STRING 29110) (T_VARIABLE 29097) (property_name 29111) (simple_variable 29112) ("{" 29113)) 13752 (("$" 29090) (T_VARIABLE 29097) (simple_variable 29109)) 13753 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 29107) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 13754 (("]" 29108)) 13755 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 29114) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 13756 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 29115)) 13757 (("$" 29090) (T_VARIABLE 29097) (simple_variable 29118)) 13758 (("%" 304) (")" 29120) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 13759 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 29123) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 13760 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 29124)) 13761 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29134) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13762 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 29132) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 13763 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29133) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13764 (("$" 28965) (T_ABSTRACT 29144) (T_ARRAY 29145) (T_AS 29146) (T_BREAK 29147) (T_CALLABLE 29148) (T_CASE 29149) (T_CATCH 29150) (T_CLASS 29151) (T_CLASS_C 29152) (T_CLONE 29153) (T_CONST 29154) (T_CONTINUE 29155) (T_DECLARE 29156) (T_DEFAULT 29157) (T_DIR 29158) (T_DO 29159) (T_ECHO 29160) (T_ELSE 29161) (T_ELSEIF 29162) (T_EMPTY 29163) (T_ENDDECLARE 29164) (T_ENDFOR 29165) (T_ENDFOREACH 29166) (T_ENDIF 29167) (T_ENDSWITCH 29168) (T_ENDWHILE 29169) (T_ENUM 29170) (T_EVAL 29171) (T_EXIT 29172) (T_EXTENDS 29173) (T_FILE 29174) (T_FINAL 29175) (T_FINALLY 29176) (T_FN 29177) (T_FOR 29178) (T_FOREACH 29179) (T_FUNCTION 29180) (T_FUNC_C 29181) (T_GLOBAL 29182) (T_GOTO 29183) (T_IF 29184) (T_IMPLEMENTS 29185) (T_INCLUDE 29186) (T_INCLUDE_ONCE 29187) (T_INSTANCEOF 29188) (T_INSTEADOF 29189) (T_INTERFACE 29190) (T_ISSET 29191) (T_LINE 29192) (T_LIST 29193) (T_LOGICAL_AND 29194) (T_LOGICAL_OR 29195) (T_LOGICAL_XOR 29196) (T_MATCH 29197) (T_METHOD_C 29198) (T_NAMESPACE 29199) (T_NEW 29200) (T_NS_C 29201) (T_PRINT 29202) (T_PRIVATE 29203) (T_PROPERTY_C 29204) (T_PROTECTED 29205) (T_PUBLIC 29206) (T_READONLY 29207) (T_REQUIRE 29208) (T_REQUIRE_ONCE 29209) (T_RETURN 29210) (T_STATIC 29211) (T_STRING 29212) (T_SWITCH 29213) (T_THROW 29214) (T_TRAIT 29215) (T_TRAIT_C 29216) (T_TRY 29217) (T_UNSET 29218) (T_USE 29219) (T_VAR 29220) (T_VARIABLE 29013) (T_WHILE 29221) (T_YIELD 29222) (identifier 29223) (member_name 29224) (reserved_non_modifiers 29225) (semi_reserved 29226) (simple_variable 29227) ("{" 29228)) 13765 (("(" 29231) (argument_list 29232)) 13766 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 29229) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 13767 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 29230)) 13768 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 29233) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 29234) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 29235) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 13769 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 29238) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 13770 (("," 1498) (possible_comma 29236)) 13771 ((")" 29237)) 13772 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (ampersand 29268) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29269) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13773 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29267) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13774 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29266) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13775 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29265) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13776 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29264) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13777 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29263) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13778 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29262) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13779 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29261) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13780 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29260) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13781 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29259) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13782 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29258) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13783 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29257) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13784 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29256) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13785 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29255) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13786 (("\"" 1538) ("$" 29270) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 29271) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 29272) ("[" 1560) (array_object_dereferenceable 29273) (callable_expr 29274) (callable_variable 29275) (class_constant 1564) (class_name 29276) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 29277) (name 29278) (new_dereferenceable 1570) (simple_variable 29279) (static_member 29280) (variable 29281) (variable_class_name 29282)) 13787 (("$" 29270) (T_VARIABLE 29272) (simple_variable 29315) ("{" 29316)) 13788 (("(" 29286) (argument_list 29314)) 13789 ((T_NULLSAFE_OBJECT_OPERATOR 29300) (T_OBJECT_OPERATOR 29301) ("[" 29302)) 13790 (("(" 29286) (argument_list 29299)) 13791 ((T_PAAMAYIM_NEKUDOTAYIM 29295)) 13792 (("(" 29286) (argument_list 29294)) 13793 ((T_PAAMAYIM_NEKUDOTAYIM 29283)) 13794 (("$" 29270) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 29272) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 29284) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 29285) ("{" 1660)) 13795 (("(" 29286) (argument_list 29287)) 13796 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 29288) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 29289) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 29290) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 13797 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 29293) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 13798 (("," 1498) (possible_comma 29291)) 13799 ((")" 29292)) 13800 (("$" 29270) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 29272) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 29296) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 29297) ("{" 1676)) 13801 (("(" 29286) (argument_list 29298)) 13802 (("$" 29270) (T_STRING 29305) (T_VARIABLE 29272) (property_name 29312) (simple_variable 29307) ("{" 29308)) 13803 (("$" 29270) (T_STRING 29305) (T_VARIABLE 29272) (property_name 29306) (simple_variable 29307) ("{" 29308)) 13804 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 29303) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 13805 (("]" 29304)) 13806 (("(" 29286) (argument_list 29311)) 13807 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 29309) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 13808 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 29310)) 13809 (("(" 29286) (argument_list 29313)) 13810 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 29317) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 13811 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 29318)) 13812 ((backup_doc_comment 29321)) 13813 (("(" 29322)) 13814 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 29323)) 13815 ((")" 29324)) 13816 ((T_USE 166) (lexical_vars 29325)) 13817 ((":" 168) (return_type 29326)) 13818 ((backup_fn_flags 29327)) 13819 (("{" 29328)) 13820 ((inner_statement_list 29329)) 13821 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 29330) ("~" 129)) 13822 ((backup_fn_flags 29331)) 13823 ((backup_doc_comment 29333)) 13824 (("(" 29334)) 13825 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 29335)) 13826 ((")" 29336)) 13827 ((":" 2271) (return_type 29337)) 13828 ((T_DOUBLE_ARROW 29338)) 13829 ((backup_fn_flags 29339)) 13830 ((backup_lex_pos 29340)) 13831 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29341) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13832 (("%" 29043) ("*" 29044) ("+" 29045) ("-" 29046) ("." 29047) ("/" 29048) ("<" 29049) (">" 29050) ("?" 29051) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 29052) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 29053) (T_BOOLEAN_AND 29054) (T_BOOLEAN_OR 29055) (T_COALESCE 29056) (T_INSTANCEOF 29057) (T_IS_EQUAL 29058) (T_IS_GREATER_OR_EQUAL 29059) (T_IS_IDENTICAL 29060) (T_IS_NOT_EQUAL 29061) (T_IS_NOT_IDENTICAL 29062) (T_IS_SMALLER_OR_EQUAL 29063) (T_LOGICAL_AND 29064) (T_LOGICAL_OR 29065) (T_LOGICAL_XOR 29066) (T_PIPE 29067) (T_POW 29068) (T_SL 29069) (T_SPACESHIP 29070) (T_SR 29071) ("^" 29072) (backup_fn_flags 29342) ("|" 29073)) 13833 (("$" 28965) (T_ABSTRACT 29144) (T_ARRAY 29145) (T_AS 29146) (T_BREAK 29147) (T_CALLABLE 29148) (T_CASE 29149) (T_CATCH 29150) (T_CLASS 29151) (T_CLASS_C 29152) (T_CLONE 29153) (T_CONST 29154) (T_CONTINUE 29155) (T_DECLARE 29156) (T_DEFAULT 29157) (T_DIR 29158) (T_DO 29159) (T_ECHO 29160) (T_ELSE 29161) (T_ELSEIF 29162) (T_EMPTY 29163) (T_ENDDECLARE 29164) (T_ENDFOR 29165) (T_ENDFOREACH 29166) (T_ENDIF 29167) (T_ENDSWITCH 29168) (T_ENDWHILE 29169) (T_ENUM 29170) (T_EVAL 29171) (T_EXIT 29172) (T_EXTENDS 29173) (T_FILE 29174) (T_FINAL 29175) (T_FINALLY 29176) (T_FN 29177) (T_FOR 29178) (T_FOREACH 29179) (T_FUNCTION 29180) (T_FUNC_C 29181) (T_GLOBAL 29182) (T_GOTO 29183) (T_IF 29184) (T_IMPLEMENTS 29185) (T_INCLUDE 29186) (T_INCLUDE_ONCE 29187) (T_INSTANCEOF 29188) (T_INSTEADOF 29189) (T_INTERFACE 29190) (T_ISSET 29191) (T_LINE 29192) (T_LIST 29193) (T_LOGICAL_AND 29194) (T_LOGICAL_OR 29195) (T_LOGICAL_XOR 29196) (T_MATCH 29197) (T_METHOD_C 29198) (T_NAMESPACE 29199) (T_NEW 29200) (T_NS_C 29201) (T_PRINT 29202) (T_PRIVATE 29203) (T_PROPERTY_C 29204) (T_PROTECTED 29205) (T_PUBLIC 29206) (T_READONLY 29207) (T_REQUIRE 29208) (T_REQUIRE_ONCE 29209) (T_RETURN 29210) (T_STATIC 29211) (T_STRING 29212) (T_SWITCH 29213) (T_THROW 29214) (T_TRAIT 29215) (T_TRAIT_C 29216) (T_TRY 29217) (T_UNSET 29218) (T_USE 29219) (T_VAR 29220) (T_VARIABLE 29013) (T_WHILE 29221) (T_YIELD 29222) (identifier 29345) (member_name 29346) (reserved_non_modifiers 29225) (semi_reserved 29226) (simple_variable 29347) ("{" 29348)) 13834 (("(" 29231) (argument_list 29351)) 13835 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 29349) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 13836 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 29350)) 13837 ((T_FN 35) (T_FUNCTION 131) (fn 29027) #'29028 (inline_function 29355)) 13838 (("$" 28965) (T_STRING 29361) (T_VARIABLE 29013) (property_name 29368) (simple_variable 29363) ("{" 29364)) 13839 (("$" 28965) (T_STRING 29361) (T_VARIABLE 29013) (property_name 29362) (simple_variable 29363) ("{" 29364)) 13840 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 29359) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 13841 (("]" 29360)) 13842 (("(" 29231) (argument_list 29367)) 13843 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 29365) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 13844 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 29366)) 13845 (("(" 29231) (argument_list 29369)) 13846 (("`" 29371)) 13847 (("]" 29373)) 13848 (("=" 29374)) 13849 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29375) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 13850 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29910) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13851 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 29908) (encaps_var 5357)) 13852 (("$" 29379) (T_VARIABLE 29427) (simple_variable 29904) ("{" 29905)) 13853 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 29903) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 13854 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29902) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13855 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29901) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13856 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29900) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13857 (("(" 29897)) 13858 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29896) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13859 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29895) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13860 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29882) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (clone_argument_list 29883) (constant 29438) (dereferenceable_scalar 29439) (expr 29884) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13861 (("\"" 1538) ("$" 29684) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 29685) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 29686) ("[" 1560) (array_object_dereferenceable 29687) (callable_expr 29688) (callable_variable 29689) (class_constant 1564) (class_name 29690) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 29691) (name 29692) (new_dereferenceable 1570) (simple_variable 29693) (static_member 29694) (variable 29881) (variable_class_name 29696)) 13862 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29880) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13863 (("(" 29877)) 13864 (("(" 29874)) 13865 (("(" 29865) (argument_list 29866) (ctor_arguments 29867)) 13866 (("\"" 1538) ("$" 29684) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 29685) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 29686) ("[" 1560) (array_object_dereferenceable 29687) (callable_expr 29688) (callable_variable 29689) (class_constant 1564) (class_name 29690) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 29691) (name 29692) (new_dereferenceable 1570) (simple_variable 29693) (static_member 29694) (variable 29864) (variable_class_name 29696)) 13867 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29863) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13868 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29862) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13869 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29861) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13870 (("(" 29857)) 13871 (("(" 29852)) 13872 (("(" 29846)) 13873 (("$" 29684) ("(" 29809) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 29810) (T_NAME_QUALIFIED 29811) (T_NAME_RELATIVE 29812) (T_READONLY 2955) (T_STATIC 29813) (T_STRING 29814) (T_VARIABLE 29686) (anonymous_class 29815) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 29816) (attribute 2961) (attributes 29817) (class_modifier 2963) (class_name 29818) (class_name_reference 29819) (name 29820) (new_variable 29821) (simple_variable 29822)) 13874 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29808) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13875 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29807) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13876 (("(" 29645) (argument_list 29806)) 13877 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29805) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13878 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29804) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13879 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 29799) (T_END_HEREDOC 29800) (T_VARIABLE 2920) (encaps_list 29801) (encaps_var 2922)) 13880 ((T_FN 35) (T_FUNCTION 131) (fn 29441) #'29442 (inline_function 29798)) 13881 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29797) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13882 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29796) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13883 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29795) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13884 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29792) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13885 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29791) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13886 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 29787) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 13887 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 29785) (encaps_list 2828) (encaps_var 2829)) 13888 ((T_NULLSAFE_OBJECT_OPERATOR 29771) (T_OBJECT_OPERATOR 29772) ("[" 29773)) 13889 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 29768) (attribute 2757) (fn 29441) #'29442 (inline_function 29769)) 13890 (("(" 29645) (argument_list 29767)) 13891 ((T_PAAMAYIM_NEKUDOTAYIM 29759)) 13892 (("%" 29457) ("*" 29458) ("+" 29459) ("-" 29460) ("." 29461) ("/" 29462) ("<" 29463) (">" 29464) ("?" 29465) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 29466) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 29467) (T_BOOLEAN_AND 29468) (T_BOOLEAN_OR 29469) (T_COALESCE 29470) (T_DOUBLE_ARROW 29757) (T_INSTANCEOF 29471) (T_IS_EQUAL 29472) (T_IS_GREATER_OR_EQUAL 29473) (T_IS_IDENTICAL 29474) (T_IS_NOT_EQUAL 29475) (T_IS_NOT_IDENTICAL 29476) (T_IS_SMALLER_OR_EQUAL 29477) (T_LOGICAL_AND 29478) (T_LOGICAL_OR 29479) (T_LOGICAL_XOR 29480) (T_PIPE 29481) (T_POW 29482) (T_SL 29483) (T_SPACESHIP 29484) (T_SR 29485) ("^" 29486) ("|" 29487)) 13893 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 29746)) 13894 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 29734)) 13895 (("(" 29645) (argument_list 29733)) 13896 (("=" 29653) (T_AND_EQUAL 29654) (T_COALESCE_EQUAL 29655) (T_CONCAT_EQUAL 29656) (T_DEC 29657) (T_DIV_EQUAL 29658) (T_INC 29659) (T_MINUS_EQUAL 29660) (T_MOD_EQUAL 29661) (T_MUL_EQUAL 29662) (T_OR_EQUAL 29663) (T_PLUS_EQUAL 29664) (T_POW_EQUAL 29665) (T_SL_EQUAL 29666) (T_SR_EQUAL 29667) (T_XOR_EQUAL 29668)) 13897 ((T_PAAMAYIM_NEKUDOTAYIM 29557)) 13898 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29456) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13899 (("%" 29457) ("*" 29458) ("+" 29459) ("-" 29460) ("." 29461) ("/" 29462) ("<" 29463) (">" 29464) ("?" 29465) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 29466) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 29467) (T_BOOLEAN_AND 29468) (T_BOOLEAN_OR 29469) (T_COALESCE 29470) (T_INSTANCEOF 29471) (T_IS_EQUAL 29472) (T_IS_GREATER_OR_EQUAL 29473) (T_IS_IDENTICAL 29474) (T_IS_NOT_EQUAL 29475) (T_IS_NOT_IDENTICAL 29476) (T_IS_SMALLER_OR_EQUAL 29477) (T_LOGICAL_AND 29478) (T_LOGICAL_OR 29479) (T_LOGICAL_XOR 29480) (T_PIPE 29481) (T_POW 29482) (T_SL 29483) (T_SPACESHIP 29484) (T_SR 29485) ("^" 29486) ("|" 29487)) 13900 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29556) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13901 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29555) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13902 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29554) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13903 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29553) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13904 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29552) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13905 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29551) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13906 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29550) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13907 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29549) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13908 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 29544) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 29545) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 13909 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29543) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13910 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29542) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13911 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29541) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13912 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29540) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13913 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29539) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13914 (("$" 29504) ("(" 29505) (T_NAME_FULLY_QUALIFIED 29506) (T_NAME_QUALIFIED 29507) (T_NAME_RELATIVE 29508) (T_STATIC 29509) (T_STRING 29510) (T_VARIABLE 29511) (class_name 29512) (class_name_reference 29513) (name 29514) (new_variable 29515) (simple_variable 29516)) 13915 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29503) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13916 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29502) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13917 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29501) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13918 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29500) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13919 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29499) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13920 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29498) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13921 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29497) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13922 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29496) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13923 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29495) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13924 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29494) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13925 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29493) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13926 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29492) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13927 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29491) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13928 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29490) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13929 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29489) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13930 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29488) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13931 (("$" 29504) (T_VARIABLE 29511) (simple_variable 29535) ("{" 29536)) 13932 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 29533) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 13933 ((T_PAAMAYIM_NEKUDOTAYIM 29531)) 13934 ((T_NULLSAFE_OBJECT_OPERATOR 29517) (T_OBJECT_OPERATOR 29518) (T_PAAMAYIM_NEKUDOTAYIM 29519) ("[" 29520)) 13935 (("$" 29504) (T_STRING 29524) (T_VARIABLE 29511) (property_name 29530) (simple_variable 29526) ("{" 29527)) 13936 (("$" 29504) (T_STRING 29524) (T_VARIABLE 29511) (property_name 29525) (simple_variable 29526) ("{" 29527)) 13937 (("$" 29504) (T_VARIABLE 29511) (simple_variable 29523)) 13938 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 29521) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 13939 (("]" 29522)) 13940 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 29528) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 13941 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 29529)) 13942 (("$" 29504) (T_VARIABLE 29511) (simple_variable 29532)) 13943 (("%" 304) (")" 29534) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 13944 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 29537) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 13945 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 29538)) 13946 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29548) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13947 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 29546) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 13948 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29547) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13949 (("$" 29379) (T_ABSTRACT 29558) (T_ARRAY 29559) (T_AS 29560) (T_BREAK 29561) (T_CALLABLE 29562) (T_CASE 29563) (T_CATCH 29564) (T_CLASS 29565) (T_CLASS_C 29566) (T_CLONE 29567) (T_CONST 29568) (T_CONTINUE 29569) (T_DECLARE 29570) (T_DEFAULT 29571) (T_DIR 29572) (T_DO 29573) (T_ECHO 29574) (T_ELSE 29575) (T_ELSEIF 29576) (T_EMPTY 29577) (T_ENDDECLARE 29578) (T_ENDFOR 29579) (T_ENDFOREACH 29580) (T_ENDIF 29581) (T_ENDSWITCH 29582) (T_ENDWHILE 29583) (T_ENUM 29584) (T_EVAL 29585) (T_EXIT 29586) (T_EXTENDS 29587) (T_FILE 29588) (T_FINAL 29589) (T_FINALLY 29590) (T_FN 29591) (T_FOR 29592) (T_FOREACH 29593) (T_FUNCTION 29594) (T_FUNC_C 29595) (T_GLOBAL 29596) (T_GOTO 29597) (T_IF 29598) (T_IMPLEMENTS 29599) (T_INCLUDE 29600) (T_INCLUDE_ONCE 29601) (T_INSTANCEOF 29602) (T_INSTEADOF 29603) (T_INTERFACE 29604) (T_ISSET 29605) (T_LINE 29606) (T_LIST 29607) (T_LOGICAL_AND 29608) (T_LOGICAL_OR 29609) (T_LOGICAL_XOR 29610) (T_MATCH 29611) (T_METHOD_C 29612) (T_NAMESPACE 29613) (T_NEW 29614) (T_NS_C 29615) (T_PRINT 29616) (T_PRIVATE 29617) (T_PROPERTY_C 29618) (T_PROTECTED 29619) (T_PUBLIC 29620) (T_READONLY 29621) (T_REQUIRE 29622) (T_REQUIRE_ONCE 29623) (T_RETURN 29624) (T_STATIC 29625) (T_STRING 29626) (T_SWITCH 29627) (T_THROW 29628) (T_TRAIT 29629) (T_TRAIT_C 29630) (T_TRY 29631) (T_UNSET 29632) (T_USE 29633) (T_VAR 29634) (T_VARIABLE 29427) (T_WHILE 29635) (T_YIELD 29636) (identifier 29637) (member_name 29638) (reserved_non_modifiers 29639) (semi_reserved 29640) (simple_variable 29641) ("{" 29642)) 13950 (("(" 29645) (argument_list 29646)) 13951 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 29643) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 13952 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 29644)) 13953 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 29647) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 29648) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 29649) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 13954 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 29652) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 13955 (("," 1498) (possible_comma 29650)) 13956 ((")" 29651)) 13957 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (ampersand 29682) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29683) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13958 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29681) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13959 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29680) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13960 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29679) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13961 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29678) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13962 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29677) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13963 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29676) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13964 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29675) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13965 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29674) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13966 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29673) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13967 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29672) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13968 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29671) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13969 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29670) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13970 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29669) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 13971 (("\"" 1538) ("$" 29684) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 29685) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 29686) ("[" 1560) (array_object_dereferenceable 29687) (callable_expr 29688) (callable_variable 29689) (class_constant 1564) (class_name 29690) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 29691) (name 29692) (new_dereferenceable 1570) (simple_variable 29693) (static_member 29694) (variable 29695) (variable_class_name 29696)) 13972 (("$" 29684) (T_VARIABLE 29686) (simple_variable 29729) ("{" 29730)) 13973 (("(" 29700) (argument_list 29728)) 13974 ((T_NULLSAFE_OBJECT_OPERATOR 29714) (T_OBJECT_OPERATOR 29715) ("[" 29716)) 13975 (("(" 29700) (argument_list 29713)) 13976 ((T_PAAMAYIM_NEKUDOTAYIM 29709)) 13977 (("(" 29700) (argument_list 29708)) 13978 ((T_PAAMAYIM_NEKUDOTAYIM 29697)) 13979 (("$" 29684) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 29686) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 29698) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 29699) ("{" 1660)) 13980 (("(" 29700) (argument_list 29701)) 13981 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 29702) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 29703) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 29704) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 13982 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 29707) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 13983 (("," 1498) (possible_comma 29705)) 13984 ((")" 29706)) 13985 (("$" 29684) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 29686) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 29710) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 29711) ("{" 1676)) 13986 (("(" 29700) (argument_list 29712)) 13987 (("$" 29684) (T_STRING 29719) (T_VARIABLE 29686) (property_name 29726) (simple_variable 29721) ("{" 29722)) 13988 (("$" 29684) (T_STRING 29719) (T_VARIABLE 29686) (property_name 29720) (simple_variable 29721) ("{" 29722)) 13989 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 29717) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 13990 (("]" 29718)) 13991 (("(" 29700) (argument_list 29725)) 13992 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 29723) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 13993 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 29724)) 13994 (("(" 29700) (argument_list 29727)) 13995 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 29731) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 13996 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 29732)) 13997 ((backup_doc_comment 29735)) 13998 (("(" 29736)) 13999 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 29737)) 14000 ((")" 29738)) 14001 ((T_USE 166) (lexical_vars 29739)) 14002 ((":" 168) (return_type 29740)) 14003 ((backup_fn_flags 29741)) 14004 (("{" 29742)) 14005 ((inner_statement_list 29743)) 14006 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 29744) ("~" 129)) 14007 ((backup_fn_flags 29745)) 14008 ((backup_doc_comment 29747)) 14009 (("(" 29748)) 14010 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 29749)) 14011 ((")" 29750)) 14012 ((":" 2271) (return_type 29751)) 14013 ((T_DOUBLE_ARROW 29752)) 14014 ((backup_fn_flags 29753)) 14015 ((backup_lex_pos 29754)) 14016 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29755) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 14017 (("%" 29457) ("*" 29458) ("+" 29459) ("-" 29460) ("." 29461) ("/" 29462) ("<" 29463) (">" 29464) ("?" 29465) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 29466) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 29467) (T_BOOLEAN_AND 29468) (T_BOOLEAN_OR 29469) (T_COALESCE 29470) (T_INSTANCEOF 29471) (T_IS_EQUAL 29472) (T_IS_GREATER_OR_EQUAL 29473) (T_IS_IDENTICAL 29474) (T_IS_NOT_EQUAL 29475) (T_IS_NOT_IDENTICAL 29476) (T_IS_SMALLER_OR_EQUAL 29477) (T_LOGICAL_AND 29478) (T_LOGICAL_OR 29479) (T_LOGICAL_XOR 29480) (T_PIPE 29481) (T_POW 29482) (T_SL 29483) (T_SPACESHIP 29484) (T_SR 29485) ("^" 29486) (backup_fn_flags 29756) ("|" 29487)) 14018 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29758) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 14019 (("$" 29379) (T_ABSTRACT 29558) (T_ARRAY 29559) (T_AS 29560) (T_BREAK 29561) (T_CALLABLE 29562) (T_CASE 29563) (T_CATCH 29564) (T_CLASS 29565) (T_CLASS_C 29566) (T_CLONE 29567) (T_CONST 29568) (T_CONTINUE 29569) (T_DECLARE 29570) (T_DEFAULT 29571) (T_DIR 29572) (T_DO 29573) (T_ECHO 29574) (T_ELSE 29575) (T_ELSEIF 29576) (T_EMPTY 29577) (T_ENDDECLARE 29578) (T_ENDFOR 29579) (T_ENDFOREACH 29580) (T_ENDIF 29581) (T_ENDSWITCH 29582) (T_ENDWHILE 29583) (T_ENUM 29584) (T_EVAL 29585) (T_EXIT 29586) (T_EXTENDS 29587) (T_FILE 29588) (T_FINAL 29589) (T_FINALLY 29590) (T_FN 29591) (T_FOR 29592) (T_FOREACH 29593) (T_FUNCTION 29594) (T_FUNC_C 29595) (T_GLOBAL 29596) (T_GOTO 29597) (T_IF 29598) (T_IMPLEMENTS 29599) (T_INCLUDE 29600) (T_INCLUDE_ONCE 29601) (T_INSTANCEOF 29602) (T_INSTEADOF 29603) (T_INTERFACE 29604) (T_ISSET 29605) (T_LINE 29606) (T_LIST 29607) (T_LOGICAL_AND 29608) (T_LOGICAL_OR 29609) (T_LOGICAL_XOR 29610) (T_MATCH 29611) (T_METHOD_C 29612) (T_NAMESPACE 29613) (T_NEW 29614) (T_NS_C 29615) (T_PRINT 29616) (T_PRIVATE 29617) (T_PROPERTY_C 29618) (T_PROTECTED 29619) (T_PUBLIC 29620) (T_READONLY 29621) (T_REQUIRE 29622) (T_REQUIRE_ONCE 29623) (T_RETURN 29624) (T_STATIC 29625) (T_STRING 29626) (T_SWITCH 29627) (T_THROW 29628) (T_TRAIT 29629) (T_TRAIT_C 29630) (T_TRY 29631) (T_UNSET 29632) (T_USE 29633) (T_VAR 29634) (T_VARIABLE 29427) (T_WHILE 29635) (T_YIELD 29636) (identifier 29760) (member_name 29761) (reserved_non_modifiers 29639) (semi_reserved 29640) (simple_variable 29762) ("{" 29763)) 14020 (("(" 29645) (argument_list 29766)) 14021 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 29764) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 14022 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 29765)) 14023 ((T_FN 35) (T_FUNCTION 131) (fn 29441) #'29442 (inline_function 29770)) 14024 (("$" 29379) (T_STRING 29776) (T_VARIABLE 29427) (property_name 29783) (simple_variable 29778) ("{" 29779)) 14025 (("$" 29379) (T_STRING 29776) (T_VARIABLE 29427) (property_name 29777) (simple_variable 29778) ("{" 29779)) 14026 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 29774) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 14027 (("]" 29775)) 14028 (("(" 29645) (argument_list 29782)) 14029 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 29780) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 14030 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 29781)) 14031 (("(" 29645) (argument_list 29784)) 14032 (("`" 29786)) 14033 (("]" 29788)) 14034 (("=" 29789)) 14035 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29790) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 14036 (("%" 29457) ("*" 29458) ("+" 29459) ("-" 29460) ("." 29461) ("/" 29462) ("<" 29463) (">" 29464) ("?" 29465) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 29466) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 29467) (T_BOOLEAN_AND 29468) (T_BOOLEAN_OR 29469) (T_COALESCE 29470) (T_DOUBLE_ARROW 29793) (T_INSTANCEOF 29471) (T_IS_EQUAL 29472) (T_IS_GREATER_OR_EQUAL 29473) (T_IS_IDENTICAL 29474) (T_IS_NOT_EQUAL 29475) (T_IS_NOT_IDENTICAL 29476) (T_IS_SMALLER_OR_EQUAL 29477) (T_LOGICAL_AND 29478) (T_LOGICAL_OR 29479) (T_LOGICAL_XOR 29480) (T_PIPE 29481) (T_POW 29482) (T_SL 29483) (T_SPACESHIP 29484) (T_SR 29485) ("^" 29486) ("|" 29487)) 14037 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29794) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 14038 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 29803) (T_VARIABLE 2920) (encaps_var 2934)) 14039 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 29802) (T_VARIABLE 2920) (encaps_var 2925)) 14040 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 29844) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 14041 ((T_CLASS 29836)) 14042 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 29835) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 29816) (attribute 2982) (class_modifier 2963)) 14043 ((T_PAAMAYIM_NEKUDOTAYIM 29833)) 14044 (("(" 29700) (argument_list 29832)) 14045 ((T_NULLSAFE_OBJECT_OPERATOR 29823) (T_OBJECT_OPERATOR 29824) (T_PAAMAYIM_NEKUDOTAYIM 29825) ("[" 29826)) 14046 (("$" 29684) (T_STRING 29719) (T_VARIABLE 29686) (property_name 29831) (simple_variable 29721) ("{" 29722)) 14047 (("$" 29684) (T_STRING 29719) (T_VARIABLE 29686) (property_name 29830) (simple_variable 29721) ("{" 29722)) 14048 (("$" 29684) (T_VARIABLE 29686) (simple_variable 29829)) 14049 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 29827) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 14050 (("]" 29828)) 14051 (("$" 29684) (T_VARIABLE 29686) (simple_variable 29834)) 14052 (("(" 2984) (argument_list 2985) (ctor_arguments 29837)) 14053 ((T_EXTENDS 2987) (extends_from 29838)) 14054 ((T_IMPLEMENTS 2989) (implements_list 29839)) 14055 ((backup_doc_comment 29840)) 14056 (("{" 29841)) 14057 ((class_statement_list 29842)) 14058 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 29843)) 14059 (("%" 304) (")" 29845) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 14060 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 29847) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 14061 (("%" 304) (")" 29848) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 14062 (("{" 29849)) 14063 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 29850) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 14064 (("}" 29851)) 14065 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 29853) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 14066 ((")" 29854)) 14067 (("=" 29855)) 14068 (("!" 29377) ("\"" 29378) ("$" 29379) ("(" 29380) ("+" 29381) ("-" 29382) ("@" 29383) (T_ARRAY 29384) (T_ARRAY_CAST 29385) (T_ATTRIBUTE 130) (T_BOOL_CAST 29386) (T_CLASS_C 29387) (T_CLONE 29388) (T_CONSTANT_ENCAPSED_STRING 29389) (T_DEC 29390) (T_DIR 29391) (T_DNUMBER 29392) (T_DOUBLE_CAST 29393) (T_EMPTY 29394) (T_EVAL 29395) (T_EXIT 29396) (T_FILE 29397) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 29398) (T_INC 29399) (T_INCLUDE 29400) (T_INCLUDE_ONCE 29401) (T_INT_CAST 29402) (T_ISSET 29403) (T_LINE 29404) (T_LIST 29405) (T_LNUMBER 29406) (T_MATCH 29407) (T_METHOD_C 29408) (T_NAME_FULLY_QUALIFIED 29409) (T_NAME_QUALIFIED 29410) (T_NAME_RELATIVE 29411) (T_NEW 29412) (T_NS_C 29413) (T_OBJECT_CAST 29414) (T_PRINT 29415) (T_PROPERTY_C 29416) (T_READONLY 29417) (T_REQUIRE 29418) (T_REQUIRE_ONCE 29419) (T_START_HEREDOC 29420) (T_STATIC 29421) (T_STRING 29422) (T_STRING_CAST 29423) (T_THROW 29424) (T_TRAIT_C 29425) (T_UNSET_CAST 29426) (T_VARIABLE 29427) (T_YIELD 29428) (T_YIELD_FROM 29429) ("[" 29430) ("`" 29431) (array_object_dereferenceable 29432) (attribute 135) (attributes 29433) (callable_expr 29434) (callable_variable 29435) (class_constant 29436) (class_name 29437) (constant 29438) (dereferenceable_scalar 29439) (expr 29856) (fn 29441) (fully_dereferenceable 107) #'29442 (function_call 29443) (inline_function 29444) (internal_functions_in_yacc 29445) (match 29446) (name 29447) (new_dereferenceable 29448) (new_non_dereferenceable 29449) (scalar 29450) (simple_variable 29451) (static_member 29452) (variable 29453) (variable_class_name 29454) ("~" 29455)) 14069 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 29858) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 14070 (("," 5226) (possible_comma 29859)) 14071 ((")" 29860)) 14072 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 29868) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 29869) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 29870) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 14073 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 29873) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 14074 (("," 1498) (possible_comma 29871)) 14075 ((")" 29872)) 14076 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 29875) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 14077 (("%" 304) (")" 29876) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 14078 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 29878) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 14079 (("%" 304) (")" 29879) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 14080 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 29885) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 29886) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 29887) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 29888) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 14081 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 29894) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 14082 (("%" 2442) (")" 29891) ("*" 2443) ("+" 2444) ("," 29892) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 14083 (("," 5287) (possible_comma 29889)) 14084 ((")" 29890)) 14085 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 29893) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 14086 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 29898) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 14087 ((")" 29899)) 14088 (("%" 304) (")" 29891) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 14089 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 29906) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 14090 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 29907)) 14091 (("\"" 29909) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 14092 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 29919) (T_VARIABLE 2920) (encaps_var 2934)) 14093 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 29918) (T_VARIABLE 2920) (encaps_var 2925)) 14094 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 29960) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 14095 ((T_CLASS 29952)) 14096 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 29951) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 29932) (attribute 2982) (class_modifier 2963)) 14097 ((T_PAAMAYIM_NEKUDOTAYIM 29949)) 14098 (("(" 29286) (argument_list 29948)) 14099 ((T_NULLSAFE_OBJECT_OPERATOR 29939) (T_OBJECT_OPERATOR 29940) (T_PAAMAYIM_NEKUDOTAYIM 29941) ("[" 29942)) 14100 (("$" 29270) (T_STRING 29305) (T_VARIABLE 29272) (property_name 29947) (simple_variable 29307) ("{" 29308)) 14101 (("$" 29270) (T_STRING 29305) (T_VARIABLE 29272) (property_name 29946) (simple_variable 29307) ("{" 29308)) 14102 (("$" 29270) (T_VARIABLE 29272) (simple_variable 29945)) 14103 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 29943) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 14104 (("]" 29944)) 14105 (("$" 29270) (T_VARIABLE 29272) (simple_variable 29950)) 14106 (("(" 2984) (argument_list 2985) (ctor_arguments 29953)) 14107 ((T_EXTENDS 2987) (extends_from 29954)) 14108 ((T_IMPLEMENTS 2989) (implements_list 29955)) 14109 ((backup_doc_comment 29956)) 14110 (("{" 29957)) 14111 ((class_statement_list 29958)) 14112 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 29959)) 14113 (("%" 304) (")" 29961) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 14114 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 29963) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 14115 (("%" 304) (")" 29964) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 14116 (("{" 29965)) 14117 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 29966) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 14118 (("}" 29967)) 14119 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 29969) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 14120 ((")" 29970)) 14121 (("=" 29971)) 14122 (("!" 28963) ("\"" 28964) ("$" 28965) ("(" 28966) ("+" 28967) ("-" 28968) ("@" 28969) (T_ARRAY 28970) (T_ARRAY_CAST 28971) (T_ATTRIBUTE 130) (T_BOOL_CAST 28972) (T_CLASS_C 28973) (T_CLONE 28974) (T_CONSTANT_ENCAPSED_STRING 28975) (T_DEC 28976) (T_DIR 28977) (T_DNUMBER 28978) (T_DOUBLE_CAST 28979) (T_EMPTY 28980) (T_EVAL 28981) (T_EXIT 28982) (T_FILE 28983) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 28984) (T_INC 28985) (T_INCLUDE 28986) (T_INCLUDE_ONCE 28987) (T_INT_CAST 28988) (T_ISSET 28989) (T_LINE 28990) (T_LIST 28991) (T_LNUMBER 28992) (T_MATCH 28993) (T_METHOD_C 28994) (T_NAME_FULLY_QUALIFIED 28995) (T_NAME_QUALIFIED 28996) (T_NAME_RELATIVE 28997) (T_NEW 28998) (T_NS_C 28999) (T_OBJECT_CAST 29000) (T_PRINT 29001) (T_PROPERTY_C 29002) (T_READONLY 29003) (T_REQUIRE 29004) (T_REQUIRE_ONCE 29005) (T_START_HEREDOC 29006) (T_STATIC 29007) (T_STRING 29008) (T_STRING_CAST 29009) (T_THROW 29010) (T_TRAIT_C 29011) (T_UNSET_CAST 29012) (T_VARIABLE 29013) (T_YIELD 29014) (T_YIELD_FROM 29015) ("[" 29016) ("`" 29017) (array_object_dereferenceable 29018) (attribute 135) (attributes 29019) (callable_expr 29020) (callable_variable 29021) (class_constant 29022) (class_name 29023) (constant 29024) (dereferenceable_scalar 29025) (expr 29972) (fn 29027) (fully_dereferenceable 107) #'29028 (function_call 29029) (inline_function 29030) (internal_functions_in_yacc 29031) (match 29032) (name 29033) (new_dereferenceable 29034) (new_non_dereferenceable 29035) (scalar 29036) (simple_variable 29037) (static_member 29038) (variable 29039) (variable_class_name 29040) ("~" 29041)) 14123 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 29974) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 14124 (("," 5226) (possible_comma 29975)) 14125 ((")" 29976)) 14126 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 29984) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 29985) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 29986) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 14127 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 29989) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 14128 (("," 1498) (possible_comma 29987)) 14129 ((")" 29988)) 14130 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 29991) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 14131 (("%" 304) (")" 29992) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 14132 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 29994) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 14133 (("%" 304) (")" 29995) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 14134 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 30001) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 30002) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 30003) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 30004) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 14135 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 30010) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 14136 (("%" 2442) (")" 30007) ("*" 2443) ("+" 2444) ("," 30008) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 14137 (("," 5287) (possible_comma 30005)) 14138 ((")" 30006)) 14139 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 30009) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 14140 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 30014) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 14141 ((")" 30015)) 14142 (("%" 304) (")" 30007) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 14143 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 30022) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 14144 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 30023)) 14145 (("\"" 30025) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 14146 ((T_ARRAY 28944) (T_CALLABLE 28945) (T_NAME_FULLY_QUALIFIED 28946) (T_NAME_QUALIFIED 28947) (T_NAME_RELATIVE 28948) (T_STRING 28949) (name 28950) (type_without_static 30028)) 14147 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 30039) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (optional_cpp_modifiers 162) (parameter 163)) 14148 (("," 6013) (possible_comma 30044)) 14149 (("]" 30045)) 14150 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 30048) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 30049) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 30050) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 14151 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 30053) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 14152 (("," 1498) (possible_comma 30051)) 14153 ((")" 30052)) 14154 (("," 6013) (possible_comma 30055)) 14155 (("]" 30056)) 14156 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 30058) ("~" 129)) 14157 (("$" 4) (T_ABSTRACT 30060) (T_ARRAY 30061) (T_AS 30062) (T_BREAK 30063) (T_CALLABLE 30064) (T_CASE 30065) (T_CATCH 30066) (T_CLASS 30067) (T_CLASS_C 30068) (T_CLONE 30069) (T_CONST 30070) (T_CONTINUE 30071) (T_DECLARE 30072) (T_DEFAULT 30073) (T_DIR 30074) (T_DO 30075) (T_ECHO 30076) (T_ELSE 30077) (T_ELSEIF 30078) (T_EMPTY 30079) (T_ENDDECLARE 30080) (T_ENDFOR 30081) (T_ENDFOREACH 30082) (T_ENDIF 30083) (T_ENDSWITCH 30084) (T_ENDWHILE 30085) (T_ENUM 30086) (T_EVAL 30087) (T_EXIT 30088) (T_EXTENDS 30089) (T_FILE 30090) (T_FINAL 30091) (T_FINALLY 30092) (T_FN 30093) (T_FOR 30094) (T_FOREACH 30095) (T_FUNCTION 30096) (T_FUNC_C 30097) (T_GLOBAL 30098) (T_GOTO 30099) (T_IF 30100) (T_IMPLEMENTS 30101) (T_INCLUDE 30102) (T_INCLUDE_ONCE 30103) (T_INSTANCEOF 30104) (T_INSTEADOF 30105) (T_INTERFACE 30106) (T_ISSET 30107) (T_LINE 30108) (T_LIST 30109) (T_LOGICAL_AND 30110) (T_LOGICAL_OR 30111) (T_LOGICAL_XOR 30112) (T_MATCH 30113) (T_METHOD_C 30114) (T_NAMESPACE 30115) (T_NEW 30116) (T_NS_C 30117) (T_PRINT 30118) (T_PRIVATE 30119) (T_PROPERTY_C 30120) (T_PROTECTED 30121) (T_PUBLIC 30122) (T_READONLY 30123) (T_REQUIRE 30124) (T_REQUIRE_ONCE 30125) (T_RETURN 30126) (T_STATIC 30127) (T_STRING 30128) (T_SWITCH 30129) (T_THROW 30130) (T_TRAIT 30131) (T_TRAIT_C 30132) (T_TRY 30133) (T_UNSET 30134) (T_USE 30135) (T_VAR 30136) (T_VARIABLE 81) (T_WHILE 30137) (T_YIELD 30138) (identifier 30139) (member_name 30140) (reserved_non_modifiers 30141) (semi_reserved 30142) (simple_variable 30143) ("{" 30144)) 14158 (("(" 30046) (argument_list 30147)) 14159 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 30145) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 14160 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 30146)) 14161 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (ampersand 30177) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 30178) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 14162 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 30176) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 14163 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 30175) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 14164 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 30174) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 14165 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 30173) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 14166 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 30172) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 14167 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 30171) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 14168 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 30170) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 14169 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 30169) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 14170 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 30168) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 14171 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 30167) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 14172 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 30166) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 14173 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 30165) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 14174 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 30164) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 14175 (("\"" 1538) ("$" 30179) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 30180) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 30181) ("[" 1560) (array_object_dereferenceable 30182) (callable_expr 30183) (callable_variable 30184) (class_constant 1564) (class_name 30185) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 30186) (name 30187) (new_dereferenceable 1570) (simple_variable 30188) (static_member 30189) (variable 30190) (variable_class_name 30191)) 14176 (("$" 30179) (T_VARIABLE 30181) (simple_variable 30224) ("{" 30225)) 14177 (("(" 30195) (argument_list 30223)) 14178 ((T_NULLSAFE_OBJECT_OPERATOR 30209) (T_OBJECT_OPERATOR 30210) ("[" 30211)) 14179 (("(" 30195) (argument_list 30208)) 14180 ((T_PAAMAYIM_NEKUDOTAYIM 30204)) 14181 (("(" 30195) (argument_list 30203)) 14182 ((T_PAAMAYIM_NEKUDOTAYIM 30192)) 14183 (("$" 30179) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 30181) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 30193) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 30194) ("{" 1660)) 14184 (("(" 30195) (argument_list 30196)) 14185 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 30197) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 30198) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 30199) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 14186 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 30202) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 14187 (("," 1498) (possible_comma 30200)) 14188 ((")" 30201)) 14189 (("$" 30179) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 30181) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 30205) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 30206) ("{" 1676)) 14190 (("(" 30195) (argument_list 30207)) 14191 (("$" 30179) (T_STRING 30214) (T_VARIABLE 30181) (property_name 30221) (simple_variable 30216) ("{" 30217)) 14192 (("$" 30179) (T_STRING 30214) (T_VARIABLE 30181) (property_name 30215) (simple_variable 30216) ("{" 30217)) 14193 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 30212) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 14194 (("]" 30213)) 14195 (("(" 30195) (argument_list 30220)) 14196 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 30218) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 14197 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 30219)) 14198 (("(" 30195) (argument_list 30222)) 14199 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 30226) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 14200 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 30227)) 14201 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 8) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 15) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21) (T_DEC 22) (T_DECLARE 23) (T_DIR 24) (T_DNUMBER 25) (T_DO 26) (T_DOUBLE_CAST 27) (T_ECHO 28) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 36) (T_FOREACH 37) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 40) (T_GOTO 41) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 47) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 68) (T_START_HEREDOC 69) (T_STATIC 70) (T_STRING 71) (T_STRING_CAST 72) (T_SWITCH 73) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 77) (T_UNSET 78) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 82) (T_WHILE 83) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 88) (alt_if_stmt_without_else 89) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 105) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 111) (if_stmt_without_else 112) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 30381) (static_member 123) (variable 126) (variable_class_name 127) ("{" 128) ("~" 129)) 14202 (("(" 30231)) 14203 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 30232) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 14204 (("%" 304) (")" 30233) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 14205 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 30234) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 30235) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 30236) (T_DEC 22) (T_DECLARE 30237) (T_DIR 24) (T_DNUMBER 25) (T_DO 30238) (T_DOUBLE_CAST 27) (T_ECHO 30239) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 30240) (T_FOREACH 30241) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 30242) (T_GOTO 30243) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 30244) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 30245) (T_START_HEREDOC 69) (T_STATIC 30246) (T_STRING 30247) (T_STRING_CAST 72) (T_SWITCH 30248) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 30249) (T_UNSET 30250) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 30251) (T_WHILE 30252) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 30253) (alt_if_stmt_without_else 30254) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 30255) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 30256) (if_stmt_without_else 30257) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 30258) (static_member 123) (variable 126) (variable_class_name 127) ("{" 30259) ("~" 129)) 14206 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 30379) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 14207 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 30377) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 14208 (("(" 30368)) 14209 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22932) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22933) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22934) (T_DEC 22) (T_DECLARE 22935) (T_DIR 24) (T_DNUMBER 25) (T_DO 22936) (T_DOUBLE_CAST 27) (T_ECHO 22937) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22938) (T_FOREACH 22939) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22940) (T_GOTO 22941) (T_IF 22942) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22943) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22944) (T_START_HEREDOC 69) (T_STATIC 22945) (T_STRING 22946) (T_STRING_CAST 72) (T_SWITCH 22947) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22948) (T_UNSET 22949) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22950) (T_WHILE 22951) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22952) (alt_if_stmt_without_else 22953) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22954) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22955) (if_stmt_without_else 22956) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 30362) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22958) ("~" 129)) 14210 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (echo_expr 22926) (echo_expr_list 30360) (expr 22928) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 14211 (("(" 30347)) 14212 (("(" 30332)) 14213 (("$" 19441) (T_VARIABLE 19442) (global_var 19443) (global_var_list 30330) (simple_variable 19445)) 14214 ((T_STRING 30328)) 14215 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 30326) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 14216 ((T_FN 35) (T_FUNCTION 131) (T_VARIABLE 18365) (fn 106) #'138 (inline_function 18366) (static_var 18367) (static_var_list 30324)) 14217 ((":" 30323)) 14218 (("(" 30305)) 14219 (("{" 30288)) 14220 (("(" 30283)) 14221 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 30281) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 14222 (("(" 30272)) 14223 ((T_ELSE 30265) (T_ELSEIF 17711) (T_ENDIF 30266)) 14224 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 30264) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 14225 ((T_ELSE 30262) (T_ELSEIF 30230)) 14226 ((inner_statement_list 30260)) 14227 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 30261) ("~" 129)) 14228 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 30234) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 30235) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 30236) (T_DEC 22) (T_DECLARE 30237) (T_DIR 24) (T_DNUMBER 25) (T_DO 30238) (T_DOUBLE_CAST 27) (T_ECHO 30239) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 30240) (T_FOREACH 30241) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 30242) (T_GOTO 30243) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 30244) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 30245) (T_START_HEREDOC 69) (T_STATIC 30246) (T_STRING 30247) (T_STRING_CAST 72) (T_SWITCH 30248) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 30249) (T_UNSET 30250) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 30251) (T_WHILE 30252) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 30253) (alt_if_stmt_without_else 30254) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 30255) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 30256) (if_stmt_without_else 30257) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 30263) (static_member 123) (variable 126) (variable_class_name 127) ("{" 30259) ("~" 129)) 14229 ((":" 30268)) 14230 ((";" 30267)) 14231 ((inner_statement_list 30269)) 14232 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17807) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17808) (T_CLASS 17809) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17810) (T_DEC 22) (T_DECLARE 17811) (T_DIR 24) (T_DNUMBER 25) (T_DO 17812) (T_DOUBLE_CAST 27) (T_ECHO 17813) (T_EMPTY 29) (T_ENDIF 30270) (T_ENUM 17815) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17816) (T_FOREACH 17817) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17818) (T_GOTO 17819) (T_HALT_COMPILER 17820) (T_IF 17732) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17821) (T_INTERFACE 17822) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17823) (T_START_HEREDOC 69) (T_STATIC 17824) (T_STRING 17825) (T_STRING_CAST 72) (T_SWITCH 17826) (T_THROW 74) (T_TRAIT 17827) (T_TRAIT_C 76) (T_TRY 17828) (T_UNSET 17829) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17830) (T_WHILE 17831) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17832) (alt_if_stmt_without_else 17833) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17834) (attributes 17835) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17836) (class_modifier 99) (class_modifiers 17837) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17838) (expr 17839) (fn 106) (fully_dereferenceable 107) #'17840 (function_call 109) (function_declaration_statement 17841) (if_stmt 17842) (if_stmt_without_else 17843) (inline_function 113) (inner_statement 17844) (interface_declaration_statement 17845) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17846) (static_member 123) (trait_declaration_statement 17847) (variable 126) (variable_class_name 127) ("{" 17848) ("~" 129)) 14233 ((";" 30271)) 14234 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 30273) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 14235 (("%" 304) (")" 30274) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 14236 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 30275) (";" 30234) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 30235) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 30236) (T_DEC 22) (T_DECLARE 30237) (T_DIR 24) (T_DNUMBER 25) (T_DO 30238) (T_DOUBLE_CAST 27) (T_ECHO 30239) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 30240) (T_FOREACH 30241) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 30242) (T_GOTO 30243) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 30244) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 30245) (T_START_HEREDOC 69) (T_STATIC 30246) (T_STRING 30247) (T_STRING_CAST 72) (T_SWITCH 30248) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 30249) (T_UNSET 30250) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 30251) (T_WHILE 30252) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 30253) (alt_if_stmt_without_else 30254) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 30255) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 30256) (if_stmt_without_else 30257) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 30276) (static_member 123) (variable 126) (variable_class_name 127) (while_statement 30277) ("{" 30259) ("~" 129)) 14237 ((inner_statement_list 30278)) 14238 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17889) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17890) (T_CLASS 17891) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17892) (T_DEC 22) (T_DECLARE 17893) (T_DIR 24) (T_DNUMBER 25) (T_DO 17894) (T_DOUBLE_CAST 27) (T_ECHO 17895) (T_EMPTY 29) (T_ENDWHILE 30279) (T_ENUM 17897) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17898) (T_FOREACH 17899) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17900) (T_GOTO 17901) (T_HALT_COMPILER 17902) (T_IF 17903) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17904) (T_INTERFACE 17905) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17906) (T_START_HEREDOC 69) (T_STATIC 17907) (T_STRING 17908) (T_STRING_CAST 72) (T_SWITCH 17909) (T_THROW 74) (T_TRAIT 17910) (T_TRAIT_C 76) (T_TRY 17911) (T_UNSET 17912) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17913) (T_WHILE 17914) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17915) (alt_if_stmt_without_else 17916) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17917) (attributes 17918) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17919) (class_modifier 99) (class_modifiers 17920) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17921) (expr 17922) (fn 106) (fully_dereferenceable 107) #'17923 (function_call 109) (function_declaration_statement 17924) (if_stmt 17925) (if_stmt_without_else 17926) (inline_function 113) (inner_statement 17927) (interface_declaration_statement 17928) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17929) (static_member 123) (trait_declaration_statement 17930) (variable 126) (variable_class_name 127) ("{" 17931) ("~" 129)) 14239 ((";" 30280)) 14240 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 30282) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 14241 (("\"" 1538) ("$" 17989) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 17990) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 17991) ("[" 1560) (array_object_dereferenceable 17992) (callable_expr 17993) (callable_variable 17994) (class_constant 1564) (class_name 17995) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 17996) (name 17997) (new_dereferenceable 1570) (simple_variable 17998) (static_member 17999) (unset_variable 18000) (unset_variables 30284) (variable 18002) (variable_class_name 18003)) 14242 (("," 18015) (possible_comma 30285)) 14243 ((")" 30286)) 14244 ((";" 30287)) 14245 ((inner_statement_list 30289)) 14246 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 30290) ("~" 129)) 14247 ((catch_list 30291)) 14248 ((T_CATCH 30292) (T_FINALLY 30293) (finally_statement 30294)) 14249 (("(" 30298)) 14250 (("{" 30295)) 14251 ((inner_statement_list 30296)) 14252 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 30297) ("~" 129)) 14253 ((T_NAME_FULLY_QUALIFIED 18056) (T_NAME_QUALIFIED 18057) (T_NAME_RELATIVE 18058) (T_STATIC 18059) (T_STRING 18060) (catch_name_list 30299) (class_name 18062) (name 18063)) 14254 ((T_VARIABLE 18064) (optional_variable 30300) ("|" 18066)) 14255 ((")" 30301)) 14256 (("{" 30302)) 14257 ((inner_statement_list 30303)) 14258 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 30304) ("~" 129)) 14259 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 30306) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 14260 (("%" 304) (")" 30307) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 14261 ((":" 30308) (switch_case_list 30309) ("{" 30310)) 14262 ((";" 30316) (case_list 30317)) 14263 ((";" 30311) (case_list 30312)) 14264 ((case_list 30314)) 14265 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 30313)) 14266 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 30315)) 14267 ((case_list 30320)) 14268 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 30318)) 14269 ((";" 30319)) 14270 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 30321)) 14271 ((";" 30322)) 14272 (("," 18369) (";" 30325)) 14273 ((";" 30327)) 14274 ((";" 30329)) 14275 (("," 19446) (";" 30331)) 14276 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 30333) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 14277 (("%" 19534) ("*" 19535) ("+" 19536) ("-" 19537) ("." 19538) ("/" 19539) ("<" 19540) (">" 19541) ("?" 19542) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19543) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19544) (T_AS 30334) (T_BOOLEAN_AND 19545) (T_BOOLEAN_OR 19546) (T_COALESCE 19547) (T_INSTANCEOF 19548) (T_IS_EQUAL 19549) (T_IS_GREATER_OR_EQUAL 19550) (T_IS_IDENTICAL 19551) (T_IS_NOT_EQUAL 19552) (T_IS_NOT_IDENTICAL 19553) (T_IS_SMALLER_OR_EQUAL 19554) (T_LOGICAL_AND 19555) (T_LOGICAL_OR 19556) (T_LOGICAL_XOR 19557) (T_PIPE 19558) (T_POW 19559) (T_SL 19560) (T_SPACESHIP 19561) (T_SR 19562) ("^" 19563) ("|" 19564)) 14278 (("\"" 1538) ("$" 20897) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20898) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20899) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20900) ("[" 20901) (ampersand 20902) (array_object_dereferenceable 20903) (callable_expr 20904) (callable_variable 20905) (class_constant 1564) (class_name 20906) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 30335) (fully_dereferenceable 107) (function_call 20908) (name 20909) (new_dereferenceable 1570) (simple_variable 20910) (static_member 20911) (variable 20912) (variable_class_name 20913)) 14279 ((")" 30336) (T_DOUBLE_ARROW 30337)) 14280 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 30340) (";" 30234) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 30235) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 30236) (T_DEC 22) (T_DECLARE 30237) (T_DIR 24) (T_DNUMBER 25) (T_DO 30238) (T_DOUBLE_CAST 27) (T_ECHO 30239) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 30240) (T_FOREACH 30241) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 30242) (T_GOTO 30243) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 30244) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 30245) (T_START_HEREDOC 69) (T_STATIC 30246) (T_STRING 30247) (T_STRING_CAST 72) (T_SWITCH 30248) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 30249) (T_UNSET 30250) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 30251) (T_WHILE 30252) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 30253) (alt_if_stmt_without_else 30254) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 30255) (fn 106) (foreach_statement 30346) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 30256) (if_stmt_without_else 30257) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 30342) (static_member 123) (variable 126) (variable_class_name 127) ("{" 30259) ("~" 129)) 14281 (("\"" 1538) ("$" 20928) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20929) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20930) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20931) ("[" 20932) (ampersand 20933) (array_object_dereferenceable 20934) (callable_expr 20935) (callable_variable 20936) (class_constant 1564) (class_name 20937) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 30338) (fully_dereferenceable 107) (function_call 20939) (name 20940) (new_dereferenceable 1570) (simple_variable 20941) (static_member 20942) (variable 20943) (variable_class_name 20944)) 14282 ((")" 30339)) 14283 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 30340) (";" 30234) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 30235) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 30236) (T_DEC 22) (T_DECLARE 30237) (T_DIR 24) (T_DNUMBER 25) (T_DO 30238) (T_DOUBLE_CAST 27) (T_ECHO 30239) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 30240) (T_FOREACH 30241) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 30242) (T_GOTO 30243) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 30244) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 30245) (T_START_HEREDOC 69) (T_STATIC 30246) (T_STRING 30247) (T_STRING_CAST 72) (T_SWITCH 30248) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 30249) (T_UNSET 30250) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 30251) (T_WHILE 30252) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 30253) (alt_if_stmt_without_else 30254) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 30255) (fn 106) (foreach_statement 30341) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 30256) (if_stmt_without_else 30257) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 30342) (static_member 123) (variable 126) (variable_class_name 127) ("{" 30259) ("~" 129)) 14284 ((inner_statement_list 30343)) 14285 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20962) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 20963) (T_CLASS 20964) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20965) (T_DEC 22) (T_DECLARE 20966) (T_DIR 24) (T_DNUMBER 25) (T_DO 20967) (T_DOUBLE_CAST 27) (T_ECHO 20968) (T_EMPTY 29) (T_ENDFOREACH 30344) (T_ENUM 20970) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20971) (T_FOREACH 20972) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20973) (T_GOTO 20974) (T_HALT_COMPILER 20975) (T_IF 20976) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20977) (T_INTERFACE 20978) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20979) (T_START_HEREDOC 69) (T_STATIC 20980) (T_STRING 20981) (T_STRING_CAST 72) (T_SWITCH 20982) (T_THROW 74) (T_TRAIT 20983) (T_TRAIT_C 76) (T_TRY 20984) (T_UNSET 20985) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 20986) (T_WHILE 20987) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 20988) (alt_if_stmt_without_else 20989) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 20990) (attributes 20991) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 20992) (class_modifier 99) (class_modifiers 20993) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 20994) (expr 20995) (fn 106) (fully_dereferenceable 107) #'20996 (function_call 109) (function_declaration_statement 20997) (if_stmt 20998) (if_stmt_without_else 20999) (inline_function 113) (inner_statement 21000) (interface_declaration_statement 21001) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 21002) (static_member 123) (trait_declaration_statement 21003) (variable 126) (variable_class_name 127) ("{" 21004) ("~" 129)) 14286 ((";" 30345)) 14287 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21126) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21127) (fn 18437) (for_exprs 30348) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21129) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 14288 ((";" 30349)) 14289 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21135) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21136) (fn 18437) (for_cond_exprs 30350) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21138) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 14290 ((";" 30351)) 14291 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_VOID_CAST 22737) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 22738) (fn 2425) (for_exprs 30352) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_for_exprs 22740) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 14292 ((")" 30353)) 14293 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 30354) (";" 30234) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 30235) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 30236) (T_DEC 22) (T_DECLARE 30237) (T_DIR 24) (T_DNUMBER 25) (T_DO 30238) (T_DOUBLE_CAST 27) (T_ECHO 30239) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 30240) (T_FOREACH 30241) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 30242) (T_GOTO 30243) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 30244) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 30245) (T_START_HEREDOC 69) (T_STATIC 30246) (T_STRING 30247) (T_STRING_CAST 72) (T_SWITCH 30248) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 30249) (T_UNSET 30250) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 30251) (T_WHILE 30252) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 30253) (alt_if_stmt_without_else 30254) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 30255) (fn 106) (for_statement 30355) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 30256) (if_stmt_without_else 30257) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 30356) (static_member 123) (variable 126) (variable_class_name 127) ("{" 30259) ("~" 129)) 14294 ((inner_statement_list 30357)) 14295 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22750) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 22751) (T_CLASS 22752) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22753) (T_DEC 22) (T_DECLARE 22754) (T_DIR 24) (T_DNUMBER 25) (T_DO 22755) (T_DOUBLE_CAST 27) (T_ECHO 22756) (T_EMPTY 29) (T_ENDFOR 30358) (T_ENUM 22758) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 22759) (T_FOREACH 22760) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 22761) (T_GOTO 22762) (T_HALT_COMPILER 22763) (T_IF 22764) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22765) (T_INTERFACE 22766) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22767) (T_START_HEREDOC 69) (T_STATIC 22768) (T_STRING 22769) (T_STRING_CAST 72) (T_SWITCH 22770) (T_THROW 74) (T_TRAIT 22771) (T_TRAIT_C 76) (T_TRY 22772) (T_UNSET 22773) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22774) (T_WHILE 22775) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22776) (alt_if_stmt_without_else 22777) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 22778) (attributes 22779) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 22780) (class_modifier 99) (class_modifiers 22781) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 22782) (expr 22783) (fn 106) (fully_dereferenceable 107) #'22784 (function_call 109) (function_declaration_statement 22785) (if_stmt 22786) (if_stmt_without_else 22787) (inline_function 113) (inner_statement 22788) (interface_declaration_statement 22789) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 22790) (static_member 123) (trait_declaration_statement 22791) (variable 126) (variable_class_name 127) ("{" 22792) ("~" 129)) 14296 ((";" 30359)) 14297 (("," 22929) (";" 30361)) 14298 ((T_WHILE 30363)) 14299 (("(" 30364)) 14300 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 30365) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 14301 (("%" 304) (")" 30366) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 14302 ((";" 30367)) 14303 ((T_STRING 23106) (const_decl 23107) (const_list 30369)) 14304 ((")" 30370) ("," 23110)) 14305 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 30371) (";" 30234) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 30235) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 30236) (T_DEC 22) (T_DECLARE 30237) (T_DIR 24) (T_DNUMBER 25) (T_DO 30238) (T_DOUBLE_CAST 27) (T_ECHO 30239) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 30240) (T_FOREACH 30241) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 30242) (T_GOTO 30243) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 30244) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 30245) (T_START_HEREDOC 69) (T_STATIC 30246) (T_STRING 30247) (T_STRING_CAST 72) (T_SWITCH 30248) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 30249) (T_UNSET 30250) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 30251) (T_WHILE 30252) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 30253) (alt_if_stmt_without_else 30254) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (declare_statement 30372) (dereferenceable_scalar 103) (expr 30255) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 30256) (if_stmt_without_else 30257) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 30373) (static_member 123) (variable 126) (variable_class_name 127) ("{" 30259) ("~" 129)) 14306 ((inner_statement_list 30374)) 14307 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 23116) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 23117) (T_CLASS 23118) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 23119) (T_DEC 22) (T_DECLARE 23120) (T_DIR 24) (T_DNUMBER 25) (T_DO 23121) (T_DOUBLE_CAST 27) (T_ECHO 23122) (T_EMPTY 29) (T_ENDDECLARE 30375) (T_ENUM 23124) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 23125) (T_FOREACH 23126) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 23127) (T_GOTO 23128) (T_HALT_COMPILER 23129) (T_IF 23130) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 23131) (T_INTERFACE 23132) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 23133) (T_START_HEREDOC 69) (T_STATIC 23134) (T_STRING 23135) (T_STRING_CAST 72) (T_SWITCH 23136) (T_THROW 74) (T_TRAIT 23137) (T_TRAIT_C 76) (T_TRY 23138) (T_UNSET 23139) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 23140) (T_WHILE 23141) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 23142) (alt_if_stmt_without_else 23143) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 23144) (attributes 23145) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 23146) (class_modifier 99) (class_modifiers 23147) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 23148) (expr 23149) (fn 106) (fully_dereferenceable 107) #'23150 (function_call 109) (function_declaration_statement 23151) (if_stmt 23152) (if_stmt_without_else 23153) (inline_function 113) (inner_statement 23154) (interface_declaration_statement 23155) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23156) (static_member 123) (trait_declaration_statement 23157) (variable 126) (variable_class_name 127) ("{" 23158) ("~" 129)) 14308 ((";" 30376)) 14309 ((";" 30378)) 14310 ((";" 30380)) 14311 ((T_READONLY 17774) (T_STRING 17775) (backup_doc_comment 143) (function_name 30383)) 14312 ((backup_doc_comment 30384)) 14313 (("(" 30385)) 14314 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 30386)) 14315 ((")" 30387)) 14316 ((":" 168) (return_type 30388)) 14317 ((backup_fn_flags 30389)) 14318 (("{" 30390)) 14319 ((inner_statement_list 30391)) 14320 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 30392) ("~" 129)) 14321 ((backup_fn_flags 30393)) 14322 ((backup_doc_comment 30395)) 14323 (("(" 30396)) 14324 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 30397)) 14325 ((")" 30398)) 14326 ((":" 2271) (return_type 30399)) 14327 ((T_DOUBLE_ARROW 30400)) 14328 ((backup_fn_flags 30401)) 14329 ((backup_lex_pos 30402)) 14330 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30466) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14331 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 31466) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14332 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 31464) (encaps_var 5357)) 14333 (("$" 30405) (T_VARIABLE 30453) (simple_variable 31460) ("{" 31461)) 14334 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 31459) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 14335 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 31458) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14336 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 31457) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14337 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 31456) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14338 (("(" 31453)) 14339 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 31452) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14340 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 31451) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14341 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 31438) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (clone_argument_list 31439) (constant 30464) (dereferenceable_scalar 30465) (expr 31440) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14342 (("\"" 1538) ("$" 30710) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 30711) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 30712) ("[" 1560) (array_object_dereferenceable 30713) (callable_expr 30714) (callable_variable 30715) (class_constant 1564) (class_name 30716) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 30717) (name 30718) (new_dereferenceable 1570) (simple_variable 30719) (static_member 30720) (variable 31437) (variable_class_name 30722)) 14343 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 31436) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14344 (("(" 31433)) 14345 (("(" 31430)) 14346 (("(" 31421) (argument_list 31422) (ctor_arguments 31423)) 14347 (("\"" 1538) ("$" 30710) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 30711) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 30712) ("[" 1560) (array_object_dereferenceable 30713) (callable_expr 30714) (callable_variable 30715) (class_constant 1564) (class_name 30716) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 30717) (name 30718) (new_dereferenceable 1570) (simple_variable 30719) (static_member 30720) (variable 31420) (variable_class_name 30722)) 14348 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 31419) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14349 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 31418) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14350 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 31417) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14351 (("(" 31413)) 14352 (("(" 31408)) 14353 (("(" 31402)) 14354 (("$" 30710) ("(" 31365) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 31366) (T_NAME_QUALIFIED 31367) (T_NAME_RELATIVE 31368) (T_READONLY 2955) (T_STATIC 31369) (T_STRING 31370) (T_VARIABLE 30712) (anonymous_class 31371) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 31372) (attribute 2961) (attributes 31373) (class_modifier 2963) (class_name 31374) (class_name_reference 31375) (name 31376) (new_variable 31377) (simple_variable 31378)) 14355 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 31364) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14356 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 31363) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14357 (("(" 30671) (argument_list 31362)) 14358 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 31361) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14359 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 31360) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14360 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 31355) (T_END_HEREDOC 31356) (T_VARIABLE 2920) (encaps_list 31357) (encaps_var 2922)) 14361 ((T_FN 35) (T_FUNCTION 131) (fn 30467) #'30468 (inline_function 31354)) 14362 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 31353) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14363 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 31352) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14364 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 31351) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14365 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 30880) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14366 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30816) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14367 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 30812) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 14368 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 30810) (encaps_list 2828) (encaps_var 2829)) 14369 ((T_NULLSAFE_OBJECT_OPERATOR 30796) (T_OBJECT_OPERATOR 30797) ("[" 30798)) 14370 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 30793) (attribute 2757) (fn 30467) #'30468 (inline_function 30794)) 14371 (("(" 30671) (argument_list 30792)) 14372 ((T_PAAMAYIM_NEKUDOTAYIM 30784)) 14373 (("%" 30483) ("*" 30484) ("+" 30485) ("-" 30486) ("." 30487) ("/" 30488) ("<" 30489) (">" 30490) ("?" 30491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 30492) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 30493) (T_BOOLEAN_AND 30494) (T_BOOLEAN_OR 30495) (T_COALESCE 30496) (T_INSTANCEOF 30497) (T_IS_EQUAL 30498) (T_IS_GREATER_OR_EQUAL 30499) (T_IS_IDENTICAL 30500) (T_IS_NOT_EQUAL 30501) (T_IS_NOT_IDENTICAL 30502) (T_IS_SMALLER_OR_EQUAL 30503) (T_LOGICAL_AND 30504) (T_LOGICAL_OR 30505) (T_LOGICAL_XOR 30506) (T_PIPE 30507) (T_POW 30508) (T_SL 30509) (T_SPACESHIP 30510) (T_SR 30511) ("^" 30512) (backup_fn_flags 30783) ("|" 30513)) 14374 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 30772)) 14375 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 30760)) 14376 (("(" 30671) (argument_list 30759)) 14377 (("=" 30679) (T_AND_EQUAL 30680) (T_COALESCE_EQUAL 30681) (T_CONCAT_EQUAL 30682) (T_DEC 30683) (T_DIV_EQUAL 30684) (T_INC 30685) (T_MINUS_EQUAL 30686) (T_MOD_EQUAL 30687) (T_MUL_EQUAL 30688) (T_OR_EQUAL 30689) (T_PLUS_EQUAL 30690) (T_POW_EQUAL 30691) (T_SL_EQUAL 30692) (T_SR_EQUAL 30693) (T_XOR_EQUAL 30694)) 14378 ((T_PAAMAYIM_NEKUDOTAYIM 30583)) 14379 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30482) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14380 (("%" 30483) ("*" 30484) ("+" 30485) ("-" 30486) ("." 30487) ("/" 30488) ("<" 30489) (">" 30490) ("?" 30491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 30492) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 30493) (T_BOOLEAN_AND 30494) (T_BOOLEAN_OR 30495) (T_COALESCE 30496) (T_INSTANCEOF 30497) (T_IS_EQUAL 30498) (T_IS_GREATER_OR_EQUAL 30499) (T_IS_IDENTICAL 30500) (T_IS_NOT_EQUAL 30501) (T_IS_NOT_IDENTICAL 30502) (T_IS_SMALLER_OR_EQUAL 30503) (T_LOGICAL_AND 30504) (T_LOGICAL_OR 30505) (T_LOGICAL_XOR 30506) (T_PIPE 30507) (T_POW 30508) (T_SL 30509) (T_SPACESHIP 30510) (T_SR 30511) ("^" 30512) ("|" 30513)) 14381 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30582) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14382 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30581) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14383 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30580) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14384 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30579) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14385 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30578) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14386 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30577) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14387 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30576) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14388 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30575) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14389 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 30570) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 30571) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 14390 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30569) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14391 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30568) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14392 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30567) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14393 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30566) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14394 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30565) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14395 (("$" 30530) ("(" 30531) (T_NAME_FULLY_QUALIFIED 30532) (T_NAME_QUALIFIED 30533) (T_NAME_RELATIVE 30534) (T_STATIC 30535) (T_STRING 30536) (T_VARIABLE 30537) (class_name 30538) (class_name_reference 30539) (name 30540) (new_variable 30541) (simple_variable 30542)) 14396 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30529) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14397 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30528) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14398 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30527) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14399 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30526) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14400 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30525) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14401 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30524) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14402 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30523) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14403 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30522) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14404 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30521) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14405 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30520) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14406 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30519) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14407 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30518) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14408 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30517) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14409 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30516) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14410 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30515) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14411 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30514) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14412 (("$" 30530) (T_VARIABLE 30537) (simple_variable 30561) ("{" 30562)) 14413 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 30559) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 14414 ((T_PAAMAYIM_NEKUDOTAYIM 30557)) 14415 ((T_NULLSAFE_OBJECT_OPERATOR 30543) (T_OBJECT_OPERATOR 30544) (T_PAAMAYIM_NEKUDOTAYIM 30545) ("[" 30546)) 14416 (("$" 30530) (T_STRING 30550) (T_VARIABLE 30537) (property_name 30556) (simple_variable 30552) ("{" 30553)) 14417 (("$" 30530) (T_STRING 30550) (T_VARIABLE 30537) (property_name 30551) (simple_variable 30552) ("{" 30553)) 14418 (("$" 30530) (T_VARIABLE 30537) (simple_variable 30549)) 14419 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 30547) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 14420 (("]" 30548)) 14421 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 30554) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 14422 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 30555)) 14423 (("$" 30530) (T_VARIABLE 30537) (simple_variable 30558)) 14424 (("%" 304) (")" 30560) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 14425 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 30563) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 14426 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 30564)) 14427 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30574) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14428 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 30572) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 14429 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30573) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14430 (("$" 30405) (T_ABSTRACT 30584) (T_ARRAY 30585) (T_AS 30586) (T_BREAK 30587) (T_CALLABLE 30588) (T_CASE 30589) (T_CATCH 30590) (T_CLASS 30591) (T_CLASS_C 30592) (T_CLONE 30593) (T_CONST 30594) (T_CONTINUE 30595) (T_DECLARE 30596) (T_DEFAULT 30597) (T_DIR 30598) (T_DO 30599) (T_ECHO 30600) (T_ELSE 30601) (T_ELSEIF 30602) (T_EMPTY 30603) (T_ENDDECLARE 30604) (T_ENDFOR 30605) (T_ENDFOREACH 30606) (T_ENDIF 30607) (T_ENDSWITCH 30608) (T_ENDWHILE 30609) (T_ENUM 30610) (T_EVAL 30611) (T_EXIT 30612) (T_EXTENDS 30613) (T_FILE 30614) (T_FINAL 30615) (T_FINALLY 30616) (T_FN 30617) (T_FOR 30618) (T_FOREACH 30619) (T_FUNCTION 30620) (T_FUNC_C 30621) (T_GLOBAL 30622) (T_GOTO 30623) (T_IF 30624) (T_IMPLEMENTS 30625) (T_INCLUDE 30626) (T_INCLUDE_ONCE 30627) (T_INSTANCEOF 30628) (T_INSTEADOF 30629) (T_INTERFACE 30630) (T_ISSET 30631) (T_LINE 30632) (T_LIST 30633) (T_LOGICAL_AND 30634) (T_LOGICAL_OR 30635) (T_LOGICAL_XOR 30636) (T_MATCH 30637) (T_METHOD_C 30638) (T_NAMESPACE 30639) (T_NEW 30640) (T_NS_C 30641) (T_PRINT 30642) (T_PRIVATE 30643) (T_PROPERTY_C 30644) (T_PROTECTED 30645) (T_PUBLIC 30646) (T_READONLY 30647) (T_REQUIRE 30648) (T_REQUIRE_ONCE 30649) (T_RETURN 30650) (T_STATIC 30651) (T_STRING 30652) (T_SWITCH 30653) (T_THROW 30654) (T_TRAIT 30655) (T_TRAIT_C 30656) (T_TRY 30657) (T_UNSET 30658) (T_USE 30659) (T_VAR 30660) (T_VARIABLE 30453) (T_WHILE 30661) (T_YIELD 30662) (identifier 30663) (member_name 30664) (reserved_non_modifiers 30665) (semi_reserved 30666) (simple_variable 30667) ("{" 30668)) 14431 (("(" 30671) (argument_list 30672)) 14432 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 30669) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 14433 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 30670)) 14434 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 30673) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 30674) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 30675) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 14435 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 30678) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 14436 (("," 1498) (possible_comma 30676)) 14437 ((")" 30677)) 14438 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (ampersand 30708) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30709) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14439 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30707) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14440 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30706) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14441 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30705) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14442 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30704) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14443 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30703) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14444 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30702) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14445 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30701) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14446 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30700) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14447 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30699) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14448 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30698) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14449 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30697) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14450 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30696) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14451 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30695) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14452 (("\"" 1538) ("$" 30710) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 30711) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 30712) ("[" 1560) (array_object_dereferenceable 30713) (callable_expr 30714) (callable_variable 30715) (class_constant 1564) (class_name 30716) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 30717) (name 30718) (new_dereferenceable 1570) (simple_variable 30719) (static_member 30720) (variable 30721) (variable_class_name 30722)) 14453 (("$" 30710) (T_VARIABLE 30712) (simple_variable 30755) ("{" 30756)) 14454 (("(" 30726) (argument_list 30754)) 14455 ((T_NULLSAFE_OBJECT_OPERATOR 30740) (T_OBJECT_OPERATOR 30741) ("[" 30742)) 14456 (("(" 30726) (argument_list 30739)) 14457 ((T_PAAMAYIM_NEKUDOTAYIM 30735)) 14458 (("(" 30726) (argument_list 30734)) 14459 ((T_PAAMAYIM_NEKUDOTAYIM 30723)) 14460 (("$" 30710) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 30712) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 30724) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 30725) ("{" 1660)) 14461 (("(" 30726) (argument_list 30727)) 14462 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 30728) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 30729) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 30730) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 14463 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 30733) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 14464 (("," 1498) (possible_comma 30731)) 14465 ((")" 30732)) 14466 (("$" 30710) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 30712) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 30736) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 30737) ("{" 1676)) 14467 (("(" 30726) (argument_list 30738)) 14468 (("$" 30710) (T_STRING 30745) (T_VARIABLE 30712) (property_name 30752) (simple_variable 30747) ("{" 30748)) 14469 (("$" 30710) (T_STRING 30745) (T_VARIABLE 30712) (property_name 30746) (simple_variable 30747) ("{" 30748)) 14470 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 30743) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 14471 (("]" 30744)) 14472 (("(" 30726) (argument_list 30751)) 14473 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 30749) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 14474 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 30750)) 14475 (("(" 30726) (argument_list 30753)) 14476 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 30757) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 14477 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 30758)) 14478 ((backup_doc_comment 30761)) 14479 (("(" 30762)) 14480 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 30763)) 14481 ((")" 30764)) 14482 ((T_USE 166) (lexical_vars 30765)) 14483 ((":" 168) (return_type 30766)) 14484 ((backup_fn_flags 30767)) 14485 (("{" 30768)) 14486 ((inner_statement_list 30769)) 14487 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 30770) ("~" 129)) 14488 ((backup_fn_flags 30771)) 14489 ((backup_doc_comment 30773)) 14490 (("(" 30774)) 14491 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 30775)) 14492 ((")" 30776)) 14493 ((":" 2271) (return_type 30777)) 14494 ((T_DOUBLE_ARROW 30778)) 14495 ((backup_fn_flags 30779)) 14496 ((backup_lex_pos 30780)) 14497 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30781) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14498 (("%" 30483) ("*" 30484) ("+" 30485) ("-" 30486) ("." 30487) ("/" 30488) ("<" 30489) (">" 30490) ("?" 30491) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 30492) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 30493) (T_BOOLEAN_AND 30494) (T_BOOLEAN_OR 30495) (T_COALESCE 30496) (T_INSTANCEOF 30497) (T_IS_EQUAL 30498) (T_IS_GREATER_OR_EQUAL 30499) (T_IS_IDENTICAL 30500) (T_IS_NOT_EQUAL 30501) (T_IS_NOT_IDENTICAL 30502) (T_IS_SMALLER_OR_EQUAL 30503) (T_LOGICAL_AND 30504) (T_LOGICAL_OR 30505) (T_LOGICAL_XOR 30506) (T_PIPE 30507) (T_POW 30508) (T_SL 30509) (T_SPACESHIP 30510) (T_SR 30511) ("^" 30512) (backup_fn_flags 30782) ("|" 30513)) 14499 (("$" 30405) (T_ABSTRACT 30584) (T_ARRAY 30585) (T_AS 30586) (T_BREAK 30587) (T_CALLABLE 30588) (T_CASE 30589) (T_CATCH 30590) (T_CLASS 30591) (T_CLASS_C 30592) (T_CLONE 30593) (T_CONST 30594) (T_CONTINUE 30595) (T_DECLARE 30596) (T_DEFAULT 30597) (T_DIR 30598) (T_DO 30599) (T_ECHO 30600) (T_ELSE 30601) (T_ELSEIF 30602) (T_EMPTY 30603) (T_ENDDECLARE 30604) (T_ENDFOR 30605) (T_ENDFOREACH 30606) (T_ENDIF 30607) (T_ENDSWITCH 30608) (T_ENDWHILE 30609) (T_ENUM 30610) (T_EVAL 30611) (T_EXIT 30612) (T_EXTENDS 30613) (T_FILE 30614) (T_FINAL 30615) (T_FINALLY 30616) (T_FN 30617) (T_FOR 30618) (T_FOREACH 30619) (T_FUNCTION 30620) (T_FUNC_C 30621) (T_GLOBAL 30622) (T_GOTO 30623) (T_IF 30624) (T_IMPLEMENTS 30625) (T_INCLUDE 30626) (T_INCLUDE_ONCE 30627) (T_INSTANCEOF 30628) (T_INSTEADOF 30629) (T_INTERFACE 30630) (T_ISSET 30631) (T_LINE 30632) (T_LIST 30633) (T_LOGICAL_AND 30634) (T_LOGICAL_OR 30635) (T_LOGICAL_XOR 30636) (T_MATCH 30637) (T_METHOD_C 30638) (T_NAMESPACE 30639) (T_NEW 30640) (T_NS_C 30641) (T_PRINT 30642) (T_PRIVATE 30643) (T_PROPERTY_C 30644) (T_PROTECTED 30645) (T_PUBLIC 30646) (T_READONLY 30647) (T_REQUIRE 30648) (T_REQUIRE_ONCE 30649) (T_RETURN 30650) (T_STATIC 30651) (T_STRING 30652) (T_SWITCH 30653) (T_THROW 30654) (T_TRAIT 30655) (T_TRAIT_C 30656) (T_TRY 30657) (T_UNSET 30658) (T_USE 30659) (T_VAR 30660) (T_VARIABLE 30453) (T_WHILE 30661) (T_YIELD 30662) (identifier 30785) (member_name 30786) (reserved_non_modifiers 30665) (semi_reserved 30666) (simple_variable 30787) ("{" 30788)) 14500 (("(" 30671) (argument_list 30791)) 14501 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 30789) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 14502 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 30790)) 14503 ((T_FN 35) (T_FUNCTION 131) (fn 30467) #'30468 (inline_function 30795)) 14504 (("$" 30405) (T_STRING 30801) (T_VARIABLE 30453) (property_name 30808) (simple_variable 30803) ("{" 30804)) 14505 (("$" 30405) (T_STRING 30801) (T_VARIABLE 30453) (property_name 30802) (simple_variable 30803) ("{" 30804)) 14506 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 30799) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 14507 (("]" 30800)) 14508 (("(" 30671) (argument_list 30807)) 14509 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 30805) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 14510 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 30806)) 14511 (("(" 30671) (argument_list 30809)) 14512 (("`" 30811)) 14513 (("]" 30813)) 14514 (("=" 30814)) 14515 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 30815) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14516 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 31350) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14517 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 31348) (encaps_var 5357)) 14518 (("$" 30819) (T_VARIABLE 30867) (simple_variable 31344) ("{" 31345)) 14519 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 31343) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 14520 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 31342) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14521 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 31341) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14522 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 31340) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14523 (("(" 31337)) 14524 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 31336) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14525 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 31335) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14526 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 31322) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (clone_argument_list 31323) (constant 30878) (dereferenceable_scalar 30879) (expr 31324) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14527 (("\"" 1538) ("$" 31124) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 31125) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 31126) ("[" 1560) (array_object_dereferenceable 31127) (callable_expr 31128) (callable_variable 31129) (class_constant 1564) (class_name 31130) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 31131) (name 31132) (new_dereferenceable 1570) (simple_variable 31133) (static_member 31134) (variable 31321) (variable_class_name 31136)) 14528 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 31320) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14529 (("(" 31317)) 14530 (("(" 31314)) 14531 (("(" 31305) (argument_list 31306) (ctor_arguments 31307)) 14532 (("\"" 1538) ("$" 31124) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 31125) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 31126) ("[" 1560) (array_object_dereferenceable 31127) (callable_expr 31128) (callable_variable 31129) (class_constant 1564) (class_name 31130) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 31131) (name 31132) (new_dereferenceable 1570) (simple_variable 31133) (static_member 31134) (variable 31304) (variable_class_name 31136)) 14533 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 31303) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14534 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 31302) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14535 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 31301) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14536 (("(" 31297)) 14537 (("(" 31292)) 14538 (("(" 31286)) 14539 (("$" 31124) ("(" 31249) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 31250) (T_NAME_QUALIFIED 31251) (T_NAME_RELATIVE 31252) (T_READONLY 2955) (T_STATIC 31253) (T_STRING 31254) (T_VARIABLE 31126) (anonymous_class 31255) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 31256) (attribute 2961) (attributes 31257) (class_modifier 2963) (class_name 31258) (class_name_reference 31259) (name 31260) (new_variable 31261) (simple_variable 31262)) 14540 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 31248) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14541 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 31247) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14542 (("(" 31085) (argument_list 31246)) 14543 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 31245) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14544 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 31244) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14545 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 31239) (T_END_HEREDOC 31240) (T_VARIABLE 2920) (encaps_list 31241) (encaps_var 2922)) 14546 ((T_FN 35) (T_FUNCTION 131) (fn 30881) #'30882 (inline_function 31238)) 14547 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 31237) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14548 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 31236) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14549 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 31235) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14550 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 31232) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14551 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 31231) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14552 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 31227) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 14553 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 31225) (encaps_list 2828) (encaps_var 2829)) 14554 ((T_NULLSAFE_OBJECT_OPERATOR 31211) (T_OBJECT_OPERATOR 31212) ("[" 31213)) 14555 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 31208) (attribute 2757) (fn 30881) #'30882 (inline_function 31209)) 14556 (("(" 31085) (argument_list 31207)) 14557 ((T_PAAMAYIM_NEKUDOTAYIM 31199)) 14558 (("%" 30897) ("*" 30898) ("+" 30899) ("-" 30900) ("." 30901) ("/" 30902) ("<" 30903) (">" 30904) ("?" 30905) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 30906) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 30907) (T_BOOLEAN_AND 30908) (T_BOOLEAN_OR 30909) (T_COALESCE 30910) (T_DOUBLE_ARROW 31197) (T_INSTANCEOF 30911) (T_IS_EQUAL 30912) (T_IS_GREATER_OR_EQUAL 30913) (T_IS_IDENTICAL 30914) (T_IS_NOT_EQUAL 30915) (T_IS_NOT_IDENTICAL 30916) (T_IS_SMALLER_OR_EQUAL 30917) (T_LOGICAL_AND 30918) (T_LOGICAL_OR 30919) (T_LOGICAL_XOR 30920) (T_PIPE 30921) (T_POW 30922) (T_SL 30923) (T_SPACESHIP 30924) (T_SR 30925) ("^" 30926) ("|" 30927)) 14559 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 31186)) 14560 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 31174)) 14561 (("(" 31085) (argument_list 31173)) 14562 (("=" 31093) (T_AND_EQUAL 31094) (T_COALESCE_EQUAL 31095) (T_CONCAT_EQUAL 31096) (T_DEC 31097) (T_DIV_EQUAL 31098) (T_INC 31099) (T_MINUS_EQUAL 31100) (T_MOD_EQUAL 31101) (T_MUL_EQUAL 31102) (T_OR_EQUAL 31103) (T_PLUS_EQUAL 31104) (T_POW_EQUAL 31105) (T_SL_EQUAL 31106) (T_SR_EQUAL 31107) (T_XOR_EQUAL 31108)) 14563 ((T_PAAMAYIM_NEKUDOTAYIM 30997)) 14564 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 30896) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14565 (("%" 30897) ("*" 30898) ("+" 30899) ("-" 30900) ("." 30901) ("/" 30902) ("<" 30903) (">" 30904) ("?" 30905) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 30906) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 30907) (T_BOOLEAN_AND 30908) (T_BOOLEAN_OR 30909) (T_COALESCE 30910) (T_INSTANCEOF 30911) (T_IS_EQUAL 30912) (T_IS_GREATER_OR_EQUAL 30913) (T_IS_IDENTICAL 30914) (T_IS_NOT_EQUAL 30915) (T_IS_NOT_IDENTICAL 30916) (T_IS_SMALLER_OR_EQUAL 30917) (T_LOGICAL_AND 30918) (T_LOGICAL_OR 30919) (T_LOGICAL_XOR 30920) (T_PIPE 30921) (T_POW 30922) (T_SL 30923) (T_SPACESHIP 30924) (T_SR 30925) ("^" 30926) ("|" 30927)) 14566 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 30996) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14567 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 30995) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14568 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 30994) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14569 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 30993) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14570 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 30992) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14571 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 30991) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14572 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 30990) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14573 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 30989) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14574 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 30984) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 30985) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 14575 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 30983) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14576 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 30982) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14577 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 30981) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14578 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 30980) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14579 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 30979) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14580 (("$" 30944) ("(" 30945) (T_NAME_FULLY_QUALIFIED 30946) (T_NAME_QUALIFIED 30947) (T_NAME_RELATIVE 30948) (T_STATIC 30949) (T_STRING 30950) (T_VARIABLE 30951) (class_name 30952) (class_name_reference 30953) (name 30954) (new_variable 30955) (simple_variable 30956)) 14581 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 30943) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14582 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 30942) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14583 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 30941) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14584 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 30940) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14585 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 30939) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14586 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 30938) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14587 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 30937) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14588 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 30936) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14589 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 30935) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14590 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 30934) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14591 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 30933) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14592 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 30932) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14593 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 30931) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14594 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 30930) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14595 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 30929) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14596 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 30928) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14597 (("$" 30944) (T_VARIABLE 30951) (simple_variable 30975) ("{" 30976)) 14598 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 30973) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 14599 ((T_PAAMAYIM_NEKUDOTAYIM 30971)) 14600 ((T_NULLSAFE_OBJECT_OPERATOR 30957) (T_OBJECT_OPERATOR 30958) (T_PAAMAYIM_NEKUDOTAYIM 30959) ("[" 30960)) 14601 (("$" 30944) (T_STRING 30964) (T_VARIABLE 30951) (property_name 30970) (simple_variable 30966) ("{" 30967)) 14602 (("$" 30944) (T_STRING 30964) (T_VARIABLE 30951) (property_name 30965) (simple_variable 30966) ("{" 30967)) 14603 (("$" 30944) (T_VARIABLE 30951) (simple_variable 30963)) 14604 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 30961) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 14605 (("]" 30962)) 14606 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 30968) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 14607 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 30969)) 14608 (("$" 30944) (T_VARIABLE 30951) (simple_variable 30972)) 14609 (("%" 304) (")" 30974) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 14610 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 30977) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 14611 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 30978)) 14612 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 30988) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14613 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 30986) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 14614 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 30987) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14615 (("$" 30819) (T_ABSTRACT 30998) (T_ARRAY 30999) (T_AS 31000) (T_BREAK 31001) (T_CALLABLE 31002) (T_CASE 31003) (T_CATCH 31004) (T_CLASS 31005) (T_CLASS_C 31006) (T_CLONE 31007) (T_CONST 31008) (T_CONTINUE 31009) (T_DECLARE 31010) (T_DEFAULT 31011) (T_DIR 31012) (T_DO 31013) (T_ECHO 31014) (T_ELSE 31015) (T_ELSEIF 31016) (T_EMPTY 31017) (T_ENDDECLARE 31018) (T_ENDFOR 31019) (T_ENDFOREACH 31020) (T_ENDIF 31021) (T_ENDSWITCH 31022) (T_ENDWHILE 31023) (T_ENUM 31024) (T_EVAL 31025) (T_EXIT 31026) (T_EXTENDS 31027) (T_FILE 31028) (T_FINAL 31029) (T_FINALLY 31030) (T_FN 31031) (T_FOR 31032) (T_FOREACH 31033) (T_FUNCTION 31034) (T_FUNC_C 31035) (T_GLOBAL 31036) (T_GOTO 31037) (T_IF 31038) (T_IMPLEMENTS 31039) (T_INCLUDE 31040) (T_INCLUDE_ONCE 31041) (T_INSTANCEOF 31042) (T_INSTEADOF 31043) (T_INTERFACE 31044) (T_ISSET 31045) (T_LINE 31046) (T_LIST 31047) (T_LOGICAL_AND 31048) (T_LOGICAL_OR 31049) (T_LOGICAL_XOR 31050) (T_MATCH 31051) (T_METHOD_C 31052) (T_NAMESPACE 31053) (T_NEW 31054) (T_NS_C 31055) (T_PRINT 31056) (T_PRIVATE 31057) (T_PROPERTY_C 31058) (T_PROTECTED 31059) (T_PUBLIC 31060) (T_READONLY 31061) (T_REQUIRE 31062) (T_REQUIRE_ONCE 31063) (T_RETURN 31064) (T_STATIC 31065) (T_STRING 31066) (T_SWITCH 31067) (T_THROW 31068) (T_TRAIT 31069) (T_TRAIT_C 31070) (T_TRY 31071) (T_UNSET 31072) (T_USE 31073) (T_VAR 31074) (T_VARIABLE 30867) (T_WHILE 31075) (T_YIELD 31076) (identifier 31077) (member_name 31078) (reserved_non_modifiers 31079) (semi_reserved 31080) (simple_variable 31081) ("{" 31082)) 14616 (("(" 31085) (argument_list 31086)) 14617 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 31083) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 14618 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 31084)) 14619 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 31087) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 31088) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 31089) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 14620 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 31092) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 14621 (("," 1498) (possible_comma 31090)) 14622 ((")" 31091)) 14623 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (ampersand 31122) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 31123) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14624 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 31121) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14625 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 31120) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14626 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 31119) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14627 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 31118) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14628 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 31117) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14629 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 31116) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14630 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 31115) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14631 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 31114) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14632 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 31113) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14633 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 31112) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14634 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 31111) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14635 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 31110) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14636 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 31109) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14637 (("\"" 1538) ("$" 31124) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 31125) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 31126) ("[" 1560) (array_object_dereferenceable 31127) (callable_expr 31128) (callable_variable 31129) (class_constant 1564) (class_name 31130) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 31131) (name 31132) (new_dereferenceable 1570) (simple_variable 31133) (static_member 31134) (variable 31135) (variable_class_name 31136)) 14638 (("$" 31124) (T_VARIABLE 31126) (simple_variable 31169) ("{" 31170)) 14639 (("(" 31140) (argument_list 31168)) 14640 ((T_NULLSAFE_OBJECT_OPERATOR 31154) (T_OBJECT_OPERATOR 31155) ("[" 31156)) 14641 (("(" 31140) (argument_list 31153)) 14642 ((T_PAAMAYIM_NEKUDOTAYIM 31149)) 14643 (("(" 31140) (argument_list 31148)) 14644 ((T_PAAMAYIM_NEKUDOTAYIM 31137)) 14645 (("$" 31124) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 31126) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 31138) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 31139) ("{" 1660)) 14646 (("(" 31140) (argument_list 31141)) 14647 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 31142) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 31143) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 31144) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 14648 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 31147) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 14649 (("," 1498) (possible_comma 31145)) 14650 ((")" 31146)) 14651 (("$" 31124) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 31126) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 31150) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 31151) ("{" 1676)) 14652 (("(" 31140) (argument_list 31152)) 14653 (("$" 31124) (T_STRING 31159) (T_VARIABLE 31126) (property_name 31166) (simple_variable 31161) ("{" 31162)) 14654 (("$" 31124) (T_STRING 31159) (T_VARIABLE 31126) (property_name 31160) (simple_variable 31161) ("{" 31162)) 14655 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 31157) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 14656 (("]" 31158)) 14657 (("(" 31140) (argument_list 31165)) 14658 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 31163) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 14659 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 31164)) 14660 (("(" 31140) (argument_list 31167)) 14661 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 31171) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 14662 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 31172)) 14663 ((backup_doc_comment 31175)) 14664 (("(" 31176)) 14665 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 31177)) 14666 ((")" 31178)) 14667 ((T_USE 166) (lexical_vars 31179)) 14668 ((":" 168) (return_type 31180)) 14669 ((backup_fn_flags 31181)) 14670 (("{" 31182)) 14671 ((inner_statement_list 31183)) 14672 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 31184) ("~" 129)) 14673 ((backup_fn_flags 31185)) 14674 ((backup_doc_comment 31187)) 14675 (("(" 31188)) 14676 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 31189)) 14677 ((")" 31190)) 14678 ((":" 2271) (return_type 31191)) 14679 ((T_DOUBLE_ARROW 31192)) 14680 ((backup_fn_flags 31193)) 14681 ((backup_lex_pos 31194)) 14682 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 31195) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14683 (("%" 30897) ("*" 30898) ("+" 30899) ("-" 30900) ("." 30901) ("/" 30902) ("<" 30903) (">" 30904) ("?" 30905) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 30906) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 30907) (T_BOOLEAN_AND 30908) (T_BOOLEAN_OR 30909) (T_COALESCE 30910) (T_INSTANCEOF 30911) (T_IS_EQUAL 30912) (T_IS_GREATER_OR_EQUAL 30913) (T_IS_IDENTICAL 30914) (T_IS_NOT_EQUAL 30915) (T_IS_NOT_IDENTICAL 30916) (T_IS_SMALLER_OR_EQUAL 30917) (T_LOGICAL_AND 30918) (T_LOGICAL_OR 30919) (T_LOGICAL_XOR 30920) (T_PIPE 30921) (T_POW 30922) (T_SL 30923) (T_SPACESHIP 30924) (T_SR 30925) ("^" 30926) (backup_fn_flags 31196) ("|" 30927)) 14684 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 31198) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14685 (("$" 30819) (T_ABSTRACT 30998) (T_ARRAY 30999) (T_AS 31000) (T_BREAK 31001) (T_CALLABLE 31002) (T_CASE 31003) (T_CATCH 31004) (T_CLASS 31005) (T_CLASS_C 31006) (T_CLONE 31007) (T_CONST 31008) (T_CONTINUE 31009) (T_DECLARE 31010) (T_DEFAULT 31011) (T_DIR 31012) (T_DO 31013) (T_ECHO 31014) (T_ELSE 31015) (T_ELSEIF 31016) (T_EMPTY 31017) (T_ENDDECLARE 31018) (T_ENDFOR 31019) (T_ENDFOREACH 31020) (T_ENDIF 31021) (T_ENDSWITCH 31022) (T_ENDWHILE 31023) (T_ENUM 31024) (T_EVAL 31025) (T_EXIT 31026) (T_EXTENDS 31027) (T_FILE 31028) (T_FINAL 31029) (T_FINALLY 31030) (T_FN 31031) (T_FOR 31032) (T_FOREACH 31033) (T_FUNCTION 31034) (T_FUNC_C 31035) (T_GLOBAL 31036) (T_GOTO 31037) (T_IF 31038) (T_IMPLEMENTS 31039) (T_INCLUDE 31040) (T_INCLUDE_ONCE 31041) (T_INSTANCEOF 31042) (T_INSTEADOF 31043) (T_INTERFACE 31044) (T_ISSET 31045) (T_LINE 31046) (T_LIST 31047) (T_LOGICAL_AND 31048) (T_LOGICAL_OR 31049) (T_LOGICAL_XOR 31050) (T_MATCH 31051) (T_METHOD_C 31052) (T_NAMESPACE 31053) (T_NEW 31054) (T_NS_C 31055) (T_PRINT 31056) (T_PRIVATE 31057) (T_PROPERTY_C 31058) (T_PROTECTED 31059) (T_PUBLIC 31060) (T_READONLY 31061) (T_REQUIRE 31062) (T_REQUIRE_ONCE 31063) (T_RETURN 31064) (T_STATIC 31065) (T_STRING 31066) (T_SWITCH 31067) (T_THROW 31068) (T_TRAIT 31069) (T_TRAIT_C 31070) (T_TRY 31071) (T_UNSET 31072) (T_USE 31073) (T_VAR 31074) (T_VARIABLE 30867) (T_WHILE 31075) (T_YIELD 31076) (identifier 31200) (member_name 31201) (reserved_non_modifiers 31079) (semi_reserved 31080) (simple_variable 31202) ("{" 31203)) 14686 (("(" 31085) (argument_list 31206)) 14687 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 31204) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 14688 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 31205)) 14689 ((T_FN 35) (T_FUNCTION 131) (fn 30881) #'30882 (inline_function 31210)) 14690 (("$" 30819) (T_STRING 31216) (T_VARIABLE 30867) (property_name 31223) (simple_variable 31218) ("{" 31219)) 14691 (("$" 30819) (T_STRING 31216) (T_VARIABLE 30867) (property_name 31217) (simple_variable 31218) ("{" 31219)) 14692 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 31214) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 14693 (("]" 31215)) 14694 (("(" 31085) (argument_list 31222)) 14695 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 31220) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 14696 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 31221)) 14697 (("(" 31085) (argument_list 31224)) 14698 (("`" 31226)) 14699 (("]" 31228)) 14700 (("=" 31229)) 14701 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 31230) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14702 (("%" 30897) ("*" 30898) ("+" 30899) ("-" 30900) ("." 30901) ("/" 30902) ("<" 30903) (">" 30904) ("?" 30905) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 30906) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 30907) (T_BOOLEAN_AND 30908) (T_BOOLEAN_OR 30909) (T_COALESCE 30910) (T_DOUBLE_ARROW 31233) (T_INSTANCEOF 30911) (T_IS_EQUAL 30912) (T_IS_GREATER_OR_EQUAL 30913) (T_IS_IDENTICAL 30914) (T_IS_NOT_EQUAL 30915) (T_IS_NOT_IDENTICAL 30916) (T_IS_SMALLER_OR_EQUAL 30917) (T_LOGICAL_AND 30918) (T_LOGICAL_OR 30919) (T_LOGICAL_XOR 30920) (T_PIPE 30921) (T_POW 30922) (T_SL 30923) (T_SPACESHIP 30924) (T_SR 30925) ("^" 30926) ("|" 30927)) 14703 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 31234) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14704 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 31243) (T_VARIABLE 2920) (encaps_var 2934)) 14705 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 31242) (T_VARIABLE 2920) (encaps_var 2925)) 14706 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 31284) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 14707 ((T_CLASS 31276)) 14708 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 31275) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 31256) (attribute 2982) (class_modifier 2963)) 14709 ((T_PAAMAYIM_NEKUDOTAYIM 31273)) 14710 (("(" 31140) (argument_list 31272)) 14711 ((T_NULLSAFE_OBJECT_OPERATOR 31263) (T_OBJECT_OPERATOR 31264) (T_PAAMAYIM_NEKUDOTAYIM 31265) ("[" 31266)) 14712 (("$" 31124) (T_STRING 31159) (T_VARIABLE 31126) (property_name 31271) (simple_variable 31161) ("{" 31162)) 14713 (("$" 31124) (T_STRING 31159) (T_VARIABLE 31126) (property_name 31270) (simple_variable 31161) ("{" 31162)) 14714 (("$" 31124) (T_VARIABLE 31126) (simple_variable 31269)) 14715 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 31267) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 14716 (("]" 31268)) 14717 (("$" 31124) (T_VARIABLE 31126) (simple_variable 31274)) 14718 (("(" 2984) (argument_list 2985) (ctor_arguments 31277)) 14719 ((T_EXTENDS 2987) (extends_from 31278)) 14720 ((T_IMPLEMENTS 2989) (implements_list 31279)) 14721 ((backup_doc_comment 31280)) 14722 (("{" 31281)) 14723 ((class_statement_list 31282)) 14724 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 31283)) 14725 (("%" 304) (")" 31285) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 14726 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 31287) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 14727 (("%" 304) (")" 31288) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 14728 (("{" 31289)) 14729 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 31290) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 14730 (("}" 31291)) 14731 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 31293) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 14732 ((")" 31294)) 14733 (("=" 31295)) 14734 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 31296) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 14735 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 31298) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 14736 (("," 5226) (possible_comma 31299)) 14737 ((")" 31300)) 14738 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 31308) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 31309) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 31310) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 14739 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 31313) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 14740 (("," 1498) (possible_comma 31311)) 14741 ((")" 31312)) 14742 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 31315) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 14743 (("%" 304) (")" 31316) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 14744 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 31318) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 14745 (("%" 304) (")" 31319) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 14746 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 31325) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 31326) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 31327) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 31328) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 14747 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 31334) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 14748 (("%" 2442) (")" 31331) ("*" 2443) ("+" 2444) ("," 31332) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 14749 (("," 5287) (possible_comma 31329)) 14750 ((")" 31330)) 14751 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 31333) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 14752 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 31338) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 14753 ((")" 31339)) 14754 (("%" 304) (")" 31331) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 14755 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 31346) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 14756 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 31347)) 14757 (("\"" 31349) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 14758 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 31359) (T_VARIABLE 2920) (encaps_var 2934)) 14759 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 31358) (T_VARIABLE 2920) (encaps_var 2925)) 14760 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 31400) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 14761 ((T_CLASS 31392)) 14762 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 31391) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 31372) (attribute 2982) (class_modifier 2963)) 14763 ((T_PAAMAYIM_NEKUDOTAYIM 31389)) 14764 (("(" 30726) (argument_list 31388)) 14765 ((T_NULLSAFE_OBJECT_OPERATOR 31379) (T_OBJECT_OPERATOR 31380) (T_PAAMAYIM_NEKUDOTAYIM 31381) ("[" 31382)) 14766 (("$" 30710) (T_STRING 30745) (T_VARIABLE 30712) (property_name 31387) (simple_variable 30747) ("{" 30748)) 14767 (("$" 30710) (T_STRING 30745) (T_VARIABLE 30712) (property_name 31386) (simple_variable 30747) ("{" 30748)) 14768 (("$" 30710) (T_VARIABLE 30712) (simple_variable 31385)) 14769 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 31383) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 14770 (("]" 31384)) 14771 (("$" 30710) (T_VARIABLE 30712) (simple_variable 31390)) 14772 (("(" 2984) (argument_list 2985) (ctor_arguments 31393)) 14773 ((T_EXTENDS 2987) (extends_from 31394)) 14774 ((T_IMPLEMENTS 2989) (implements_list 31395)) 14775 ((backup_doc_comment 31396)) 14776 (("{" 31397)) 14777 ((class_statement_list 31398)) 14778 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 31399)) 14779 (("%" 304) (")" 31401) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 14780 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 31403) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 14781 (("%" 304) (")" 31404) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 14782 (("{" 31405)) 14783 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 31406) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 14784 (("}" 31407)) 14785 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 31409) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 14786 ((")" 31410)) 14787 (("=" 31411)) 14788 (("!" 30403) ("\"" 30404) ("$" 30405) ("(" 30406) ("+" 30407) ("-" 30408) ("@" 30409) (T_ARRAY 30410) (T_ARRAY_CAST 30411) (T_ATTRIBUTE 130) (T_BOOL_CAST 30412) (T_CLASS_C 30413) (T_CLONE 30414) (T_CONSTANT_ENCAPSED_STRING 30415) (T_DEC 30416) (T_DIR 30417) (T_DNUMBER 30418) (T_DOUBLE_CAST 30419) (T_EMPTY 30420) (T_EVAL 30421) (T_EXIT 30422) (T_FILE 30423) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30424) (T_INC 30425) (T_INCLUDE 30426) (T_INCLUDE_ONCE 30427) (T_INT_CAST 30428) (T_ISSET 30429) (T_LINE 30430) (T_LIST 30431) (T_LNUMBER 30432) (T_MATCH 30433) (T_METHOD_C 30434) (T_NAME_FULLY_QUALIFIED 30435) (T_NAME_QUALIFIED 30436) (T_NAME_RELATIVE 30437) (T_NEW 30438) (T_NS_C 30439) (T_OBJECT_CAST 30440) (T_PRINT 30441) (T_PROPERTY_C 30442) (T_READONLY 30443) (T_REQUIRE 30444) (T_REQUIRE_ONCE 30445) (T_START_HEREDOC 30446) (T_STATIC 30447) (T_STRING 30448) (T_STRING_CAST 30449) (T_THROW 30450) (T_TRAIT_C 30451) (T_UNSET_CAST 30452) (T_VARIABLE 30453) (T_YIELD 30454) (T_YIELD_FROM 30455) ("[" 30456) ("`" 30457) (array_object_dereferenceable 30458) (attribute 135) (attributes 30459) (callable_expr 30460) (callable_variable 30461) (class_constant 30462) (class_name 30463) (constant 30464) (dereferenceable_scalar 30465) (expr 31412) (fn 30467) (fully_dereferenceable 107) #'30468 (function_call 30469) (inline_function 30470) (internal_functions_in_yacc 30471) (match 30472) (name 30473) (new_dereferenceable 30474) (new_non_dereferenceable 30475) (scalar 30476) (simple_variable 30477) (static_member 30478) (variable 30479) (variable_class_name 30480) ("~" 30481)) 14789 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 31414) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 14790 (("," 5226) (possible_comma 31415)) 14791 ((")" 31416)) 14792 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 31424) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 31425) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 31426) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 14793 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 31429) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 14794 (("," 1498) (possible_comma 31427)) 14795 ((")" 31428)) 14796 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 31431) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 14797 (("%" 304) (")" 31432) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 14798 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 31434) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 14799 (("%" 304) (")" 31435) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 14800 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 31441) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 31442) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 31443) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 31444) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 14801 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 31450) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 14802 (("%" 2442) (")" 31447) ("*" 2443) ("+" 2444) ("," 31448) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 14803 (("," 5287) (possible_comma 31445)) 14804 ((")" 31446)) 14805 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 31449) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 14806 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 31454) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 14807 ((")" 31455)) 14808 (("%" 304) (")" 31447) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 14809 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 31462) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 14810 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 31463)) 14811 (("\"" 31465) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 14812 (("$" 4) (T_ABSTRACT 30060) (T_ARRAY 30061) (T_AS 30062) (T_BREAK 30063) (T_CALLABLE 30064) (T_CASE 30065) (T_CATCH 30066) (T_CLASS 30067) (T_CLASS_C 30068) (T_CLONE 30069) (T_CONST 30070) (T_CONTINUE 30071) (T_DECLARE 30072) (T_DEFAULT 30073) (T_DIR 30074) (T_DO 30075) (T_ECHO 30076) (T_ELSE 30077) (T_ELSEIF 30078) (T_EMPTY 30079) (T_ENDDECLARE 30080) (T_ENDFOR 30081) (T_ENDFOREACH 30082) (T_ENDIF 30083) (T_ENDSWITCH 30084) (T_ENDWHILE 30085) (T_ENUM 30086) (T_EVAL 30087) (T_EXIT 30088) (T_EXTENDS 30089) (T_FILE 30090) (T_FINAL 30091) (T_FINALLY 30092) (T_FN 30093) (T_FOR 30094) (T_FOREACH 30095) (T_FUNCTION 30096) (T_FUNC_C 30097) (T_GLOBAL 30098) (T_GOTO 30099) (T_IF 30100) (T_IMPLEMENTS 30101) (T_INCLUDE 30102) (T_INCLUDE_ONCE 30103) (T_INSTANCEOF 30104) (T_INSTEADOF 30105) (T_INTERFACE 30106) (T_ISSET 30107) (T_LINE 30108) (T_LIST 30109) (T_LOGICAL_AND 30110) (T_LOGICAL_OR 30111) (T_LOGICAL_XOR 30112) (T_MATCH 30113) (T_METHOD_C 30114) (T_NAMESPACE 30115) (T_NEW 30116) (T_NS_C 30117) (T_PRINT 30118) (T_PRIVATE 30119) (T_PROPERTY_C 30120) (T_PROTECTED 30121) (T_PUBLIC 30122) (T_READONLY 30123) (T_REQUIRE 30124) (T_REQUIRE_ONCE 30125) (T_RETURN 30126) (T_STATIC 30127) (T_STRING 30128) (T_SWITCH 30129) (T_THROW 30130) (T_TRAIT 30131) (T_TRAIT_C 30132) (T_TRY 30133) (T_UNSET 30134) (T_USE 30135) (T_VAR 30136) (T_VARIABLE 81) (T_WHILE 30137) (T_YIELD 30138) (identifier 31469) (member_name 31470) (reserved_non_modifiers 30141) (semi_reserved 30142) (simple_variable 31471) ("{" 31472)) 14813 (("(" 30046) (argument_list 31475)) 14814 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 31473) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 14815 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 31474)) 14816 ((T_STRING 31477)) 14817 ((T_EXTENDS 2987) (extends_from 31478)) 14818 ((T_IMPLEMENTS 2989) (implements_list 31479)) 14819 ((backup_doc_comment 31480)) 14820 (("{" 31481)) 14821 ((class_statement_list 31482)) 14822 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 31483)) 14823 ((T_STRING 31547)) 14824 ((T_STRING 31538) (const_decl 31539) (const_list 31540)) 14825 ((T_STRING 31531)) 14826 ((T_STRING 31525)) 14827 ((T_STRING 31520)) 14828 ((T_ABSTRACT 10) (T_CLASS 31512) (T_FINAL 34) (T_READONLY 2955) (class_modifier 17789)) 14829 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17770) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17771) (ampersand 17772) (returns_ref 31500)) 14830 ((T_READONLY 17774) (T_STRING 17775) (backup_doc_comment 143) (function_name 31501)) 14831 ((backup_doc_comment 31502)) 14832 (("(" 31503)) 14833 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 31504)) 14834 ((")" 31505)) 14835 ((":" 168) (return_type 31506)) 14836 ((backup_fn_flags 31507)) 14837 (("{" 31508)) 14838 ((inner_statement_list 31509)) 14839 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 31510) ("~" 129)) 14840 ((backup_fn_flags 31511)) 14841 ((T_STRING 31513)) 14842 ((T_EXTENDS 2987) (extends_from 31514)) 14843 ((T_IMPLEMENTS 2989) (implements_list 31515)) 14844 ((backup_doc_comment 31516)) 14845 (("{" 31517)) 14846 ((class_statement_list 31518)) 14847 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 31519)) 14848 ((backup_doc_comment 31521)) 14849 (("{" 31522)) 14850 ((class_statement_list 31523)) 14851 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 31524)) 14852 ((T_EXTENDS 23405) (interface_extends_list 31526)) 14853 ((backup_doc_comment 31527)) 14854 (("{" 31528)) 14855 ((class_statement_list 31529)) 14856 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 31530)) 14857 ((":" 23454) (enum_backing_type 31532)) 14858 ((T_IMPLEMENTS 2989) (implements_list 31533)) 14859 ((backup_doc_comment 31534)) 14860 (("{" 31535)) 14861 ((class_statement_list 31536)) 14862 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 31537)) 14863 (("=" 31544)) 14864 (("," 31541) (";" 31542)) 14865 ((T_STRING 31538) (const_decl 31543)) 14866 (("!" 3073) ("\"" 3074) ("$" 3075) ("(" 3076) ("+" 3077) ("-" 3078) ("@" 3079) (T_ARRAY 3080) (T_ARRAY_CAST 3081) (T_ATTRIBUTE 130) (T_BOOL_CAST 3082) (T_CLASS_C 3083) (T_CLONE 3084) (T_CONSTANT_ENCAPSED_STRING 3085) (T_DEC 3086) (T_DIR 3087) (T_DNUMBER 3088) (T_DOUBLE_CAST 3089) (T_EMPTY 3090) (T_EVAL 3091) (T_EXIT 3092) (T_FILE 3093) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3094) (T_INC 3095) (T_INCLUDE 3096) (T_INCLUDE_ONCE 3097) (T_INT_CAST 3098) (T_ISSET 3099) (T_LINE 3100) (T_LIST 3101) (T_LNUMBER 3102) (T_MATCH 3103) (T_METHOD_C 3104) (T_NAME_FULLY_QUALIFIED 3105) (T_NAME_QUALIFIED 3106) (T_NAME_RELATIVE 3107) (T_NEW 3108) (T_NS_C 3109) (T_OBJECT_CAST 3110) (T_PRINT 3111) (T_PROPERTY_C 3112) (T_READONLY 3113) (T_REQUIRE 3114) (T_REQUIRE_ONCE 3115) (T_START_HEREDOC 3116) (T_STATIC 3117) (T_STRING 3118) (T_STRING_CAST 3119) (T_THROW 3120) (T_TRAIT_C 3121) (T_UNSET_CAST 3122) (T_VARIABLE 3123) (T_YIELD 3124) (T_YIELD_FROM 3125) ("[" 3126) ("`" 3127) (array_object_dereferenceable 3128) (attribute 135) (attributes 3129) (callable_expr 3130) (callable_variable 3131) (class_constant 3132) (class_name 3133) (constant 3134) (dereferenceable_scalar 3135) (expr 31545) (fn 3137) (fully_dereferenceable 107) #'3138 (function_call 3139) (inline_function 3140) (internal_functions_in_yacc 3141) (match 3142) (name 3143) (new_dereferenceable 3144) (new_non_dereferenceable 3145) (scalar 3146) (simple_variable 3147) (static_member 3148) (variable 3149) (variable_class_name 3150) ("~" 3151)) 14867 (("%" 3153) ("*" 3154) ("+" 3155) ("-" 3156) ("." 3157) ("/" 3158) ("<" 3159) (">" 3160) ("?" 3161) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 3162) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 3163) (T_BOOLEAN_AND 3164) (T_BOOLEAN_OR 3165) (T_COALESCE 3166) (T_INSTANCEOF 3167) (T_IS_EQUAL 3168) (T_IS_GREATER_OR_EQUAL 3169) (T_IS_IDENTICAL 3170) (T_IS_NOT_EQUAL 3171) (T_IS_NOT_IDENTICAL 3172) (T_IS_SMALLER_OR_EQUAL 3173) (T_LOGICAL_AND 3174) (T_LOGICAL_OR 3175) (T_LOGICAL_XOR 3176) (T_PIPE 3177) (T_POW 3178) (T_SL 3179) (T_SPACESHIP 3180) (T_SR 3181) ("^" 3182) (backup_doc_comment 31546) ("|" 3183)) 14868 ((T_EXTENDS 2987) (extends_from 31548)) 14869 ((T_IMPLEMENTS 2989) (implements_list 31549)) 14870 ((backup_doc_comment 31550)) 14871 (("{" 31551)) 14872 ((class_statement_list 31552)) 14873 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 31553)) 14874 (("$" 4) (T_STRING 31559) (T_VARIABLE 81) (property_name 31566) (simple_variable 31561) ("{" 31562)) 14875 (("$" 4) (T_STRING 31559) (T_VARIABLE 81) (property_name 31560) (simple_variable 31561) ("{" 31562)) 14876 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 31557) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 14877 (("]" 31558)) 14878 (("(" 30046) (argument_list 31565)) 14879 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 31563) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 14880 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 31564)) 14881 (("(" 30046) (argument_list 31567)) 14882 ((":" 31571)) 14883 ((";" 31570)) 14884 ((inner_statement_list 31572)) 14885 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17807) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17808) (T_CLASS 17809) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17810) (T_DEC 22) (T_DECLARE 17811) (T_DIR 24) (T_DNUMBER 25) (T_DO 17812) (T_DOUBLE_CAST 27) (T_ECHO 17813) (T_EMPTY 29) (T_ENDIF 31573) (T_ENUM 17815) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17816) (T_FOREACH 17817) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17818) (T_GOTO 17819) (T_HALT_COMPILER 17820) (T_IF 17732) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17821) (T_INTERFACE 17822) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17823) (T_START_HEREDOC 69) (T_STATIC 17824) (T_STRING 17825) (T_STRING_CAST 72) (T_SWITCH 17826) (T_THROW 74) (T_TRAIT 17827) (T_TRAIT_C 76) (T_TRY 17828) (T_UNSET 17829) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17830) (T_WHILE 17831) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17832) (alt_if_stmt_without_else 17833) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17834) (attributes 17835) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17836) (class_modifier 99) (class_modifiers 17837) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17838) (expr 17839) (fn 106) (fully_dereferenceable 107) #'17840 (function_call 109) (function_declaration_statement 17841) (if_stmt 17842) (if_stmt_without_else 17843) (inline_function 113) (inner_statement 17844) (interface_declaration_statement 17845) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17846) (static_member 123) (trait_declaration_statement 17847) (variable 126) (variable_class_name 127) ("{" 17848) ("~" 129)) 14886 ((";" 31574)) 14887 (("`" 31576)) 14888 (("]" 31578)) 14889 (("=" 31579)) 14890 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 31580) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 14891 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 32115) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14892 ((T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5354) (T_VARIABLE 5355) (encaps_list 32113) (encaps_var 5357)) 14893 (("$" 31584) (T_VARIABLE 31632) (simple_variable 32109) ("{" 32110)) 14894 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 32108) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 14895 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 32107) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14896 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 32106) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14897 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 32105) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14898 (("(" 32102)) 14899 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 32101) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14900 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 32100) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14901 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 32087) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (clone_argument_list 32088) (constant 31643) (dereferenceable_scalar 31644) (expr 32089) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14902 (("\"" 1538) ("$" 31889) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 31890) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 31891) ("[" 1560) (array_object_dereferenceable 31892) (callable_expr 31893) (callable_variable 31894) (class_constant 1564) (class_name 31895) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 31896) (name 31897) (new_dereferenceable 1570) (simple_variable 31898) (static_member 31899) (variable 32086) (variable_class_name 31901)) 14903 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 32085) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14904 (("(" 32082)) 14905 (("(" 32079)) 14906 (("(" 32070) (argument_list 32071) (ctor_arguments 32072)) 14907 (("\"" 1538) ("$" 31889) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 31890) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 31891) ("[" 1560) (array_object_dereferenceable 31892) (callable_expr 31893) (callable_variable 31894) (class_constant 1564) (class_name 31895) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 31896) (name 31897) (new_dereferenceable 1570) (simple_variable 31898) (static_member 31899) (variable 32069) (variable_class_name 31901)) 14908 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 32068) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14909 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 32067) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14910 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 32066) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14911 (("(" 32062)) 14912 (("(" 32057)) 14913 (("(" 32051)) 14914 (("$" 31889) ("(" 32014) (T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_NAME_FULLY_QUALIFIED 32015) (T_NAME_QUALIFIED 32016) (T_NAME_RELATIVE 32017) (T_READONLY 2955) (T_STATIC 32018) (T_STRING 32019) (T_VARIABLE 31891) (anonymous_class 32020) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 32021) (attribute 2961) (attributes 32022) (class_modifier 2963) (class_name 32023) (class_name_reference 32024) (name 32025) (new_variable 32026) (simple_variable 32027)) 14915 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 32013) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14916 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 32012) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14917 (("(" 31850) (argument_list 32011)) 14918 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 32010) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14919 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 32009) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14920 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 32004) (T_END_HEREDOC 32005) (T_VARIABLE 2920) (encaps_list 32006) (encaps_var 2922)) 14921 ((T_FN 35) (T_FUNCTION 131) (fn 31646) #'31647 (inline_function 32003)) 14922 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 32002) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14923 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 32001) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14924 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 32000) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14925 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31997) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14926 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31996) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14927 (("!" 1695) ("\"" 1696) ("$" 1697) ("(" 1698) ("+" 1699) ("-" 1700) ("@" 1701) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1702) (T_ARRAY_CAST 1703) (T_ATTRIBUTE 130) (T_BOOL_CAST 1704) (T_CLASS_C 1705) (T_CLONE 1706) (T_CONSTANT_ENCAPSED_STRING 1707) (T_DEC 1708) (T_DIR 1709) (T_DNUMBER 1710) (T_DOUBLE_CAST 1711) (T_ELLIPSIS 1712) (T_EMPTY 1713) (T_EVAL 1714) (T_EXIT 1715) (T_FILE 1716) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1717) (T_INC 1718) (T_INCLUDE 1719) (T_INCLUDE_ONCE 1720) (T_INT_CAST 1721) (T_ISSET 1722) (T_LINE 1723) (T_LIST 1724) (T_LNUMBER 1725) (T_MATCH 1726) (T_METHOD_C 1727) (T_NAME_FULLY_QUALIFIED 1728) (T_NAME_QUALIFIED 1729) (T_NAME_RELATIVE 1730) (T_NEW 1731) (T_NS_C 1732) (T_OBJECT_CAST 1733) (T_PRINT 1734) (T_PROPERTY_C 1735) (T_READONLY 1736) (T_REQUIRE 1737) (T_REQUIRE_ONCE 1738) (T_START_HEREDOC 1739) (T_STATIC 1740) (T_STRING 1741) (T_STRING_CAST 1742) (T_THROW 1743) (T_TRAIT_C 1744) (T_UNSET_CAST 1745) (T_VARIABLE 1746) (T_YIELD 1747) (T_YIELD_FROM 1748) ("[" 1749) ("`" 1750) (ampersand 1751) (array_object_dereferenceable 1752) (array_pair 1753) (array_pair_list 31992) (attribute 135) (attributes 1755) (callable_expr 1756) (callable_variable 1757) (class_constant 1758) (class_name 1759) (constant 1760) (dereferenceable_scalar 1761) (expr 1762) (fn 1763) (fully_dereferenceable 107) #'1764 (function_call 1765) (inline_function 1766) (internal_functions_in_yacc 1767) (match 1768) (name 1769) (new_dereferenceable 1770) (new_non_dereferenceable 1771) (non_empty_array_pair_list 1772) (possible_array_pair 1773) (scalar 1774) (simple_variable 1775) (static_member 1776) (variable 1777) (variable_class_name 1778) ("~" 1779)) 14928 ((T_CURLY_OPEN 2823) (T_DOLLAR_OPEN_CURLY_BRACES 2824) (T_ENCAPSED_AND_WHITESPACE 2825) (T_VARIABLE 2826) (backticks_expr 31990) (encaps_list 2828) (encaps_var 2829)) 14929 ((T_NULLSAFE_OBJECT_OPERATOR 31976) (T_OBJECT_OPERATOR 31977) ("[" 31978)) 14930 ((T_ATTRIBUTE 130) (T_FN 35) (T_FUNCTION 131) (T_STATIC 31973) (attribute 2757) (fn 31646) #'31647 (inline_function 31974)) 14931 (("(" 31850) (argument_list 31972)) 14932 ((T_PAAMAYIM_NEKUDOTAYIM 31964)) 14933 (("%" 31662) ("*" 31663) ("+" 31664) ("-" 31665) ("." 31666) ("/" 31667) ("<" 31668) (">" 31669) ("?" 31670) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 31671) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 31672) (T_BOOLEAN_AND 31673) (T_BOOLEAN_OR 31674) (T_COALESCE 31675) (T_DOUBLE_ARROW 31962) (T_INSTANCEOF 31676) (T_IS_EQUAL 31677) (T_IS_GREATER_OR_EQUAL 31678) (T_IS_IDENTICAL 31679) (T_IS_NOT_EQUAL 31680) (T_IS_NOT_IDENTICAL 31681) (T_IS_SMALLER_OR_EQUAL 31682) (T_LOGICAL_AND 31683) (T_LOGICAL_OR 31684) (T_LOGICAL_XOR 31685) (T_PIPE 31686) (T_POW 31687) (T_SL 31688) (T_SPACESHIP 31689) (T_SR 31690) ("^" 31691) ("|" 31692)) 14934 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 31951)) 14935 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 139) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 140) (ampersand 141) (returns_ref 31939)) 14936 (("(" 31850) (argument_list 31938)) 14937 (("=" 31858) (T_AND_EQUAL 31859) (T_COALESCE_EQUAL 31860) (T_CONCAT_EQUAL 31861) (T_DEC 31862) (T_DIV_EQUAL 31863) (T_INC 31864) (T_MINUS_EQUAL 31865) (T_MOD_EQUAL 31866) (T_MUL_EQUAL 31867) (T_OR_EQUAL 31868) (T_PLUS_EQUAL 31869) (T_POW_EQUAL 31870) (T_SL_EQUAL 31871) (T_SR_EQUAL 31872) (T_XOR_EQUAL 31873)) 14938 ((T_PAAMAYIM_NEKUDOTAYIM 31762)) 14939 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31661) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14940 (("%" 31662) ("*" 31663) ("+" 31664) ("-" 31665) ("." 31666) ("/" 31667) ("<" 31668) (">" 31669) ("?" 31670) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 31671) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 31672) (T_BOOLEAN_AND 31673) (T_BOOLEAN_OR 31674) (T_COALESCE 31675) (T_INSTANCEOF 31676) (T_IS_EQUAL 31677) (T_IS_GREATER_OR_EQUAL 31678) (T_IS_IDENTICAL 31679) (T_IS_NOT_EQUAL 31680) (T_IS_NOT_IDENTICAL 31681) (T_IS_SMALLER_OR_EQUAL 31682) (T_LOGICAL_AND 31683) (T_LOGICAL_OR 31684) (T_LOGICAL_XOR 31685) (T_PIPE 31686) (T_POW 31687) (T_SL 31688) (T_SPACESHIP 31689) (T_SR 31690) ("^" 31691) ("|" 31692)) 14941 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31761) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14942 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31760) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14943 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31759) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14944 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31758) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14945 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31757) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14946 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31756) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14947 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31755) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14948 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31754) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14949 (("!" 687) ("\"" 688) ("$" 689) ("(" 690) ("+" 691) ("-" 692) (":" 31749) ("@" 694) (T_ARRAY 695) (T_ARRAY_CAST 696) (T_ATTRIBUTE 130) (T_BOOL_CAST 697) (T_CLASS_C 698) (T_CLONE 699) (T_CONSTANT_ENCAPSED_STRING 700) (T_DEC 701) (T_DIR 702) (T_DNUMBER 703) (T_DOUBLE_CAST 704) (T_EMPTY 705) (T_EVAL 706) (T_EXIT 707) (T_FILE 708) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 709) (T_INC 710) (T_INCLUDE 711) (T_INCLUDE_ONCE 712) (T_INT_CAST 713) (T_ISSET 714) (T_LINE 715) (T_LIST 716) (T_LNUMBER 717) (T_MATCH 718) (T_METHOD_C 719) (T_NAME_FULLY_QUALIFIED 720) (T_NAME_QUALIFIED 721) (T_NAME_RELATIVE 722) (T_NEW 723) (T_NS_C 724) (T_OBJECT_CAST 725) (T_PRINT 726) (T_PROPERTY_C 727) (T_READONLY 728) (T_REQUIRE 729) (T_REQUIRE_ONCE 730) (T_START_HEREDOC 731) (T_STATIC 732) (T_STRING 733) (T_STRING_CAST 734) (T_THROW 735) (T_TRAIT_C 736) (T_UNSET_CAST 737) (T_VARIABLE 738) (T_YIELD 739) (T_YIELD_FROM 740) ("[" 741) ("`" 742) (array_object_dereferenceable 743) (attribute 135) (attributes 744) (callable_expr 745) (callable_variable 746) (class_constant 747) (class_name 748) (constant 749) (dereferenceable_scalar 750) (expr 31750) (fn 752) (fully_dereferenceable 107) #'753 (function_call 754) (inline_function 755) (internal_functions_in_yacc 756) (match 757) (name 758) (new_dereferenceable 759) (new_non_dereferenceable 760) (scalar 761) (simple_variable 762) (static_member 763) (variable 764) (variable_class_name 765) ("~" 766)) 14950 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31748) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14951 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31747) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14952 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31746) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14953 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31745) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14954 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31744) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14955 (("$" 31709) ("(" 31710) (T_NAME_FULLY_QUALIFIED 31711) (T_NAME_QUALIFIED 31712) (T_NAME_RELATIVE 31713) (T_STATIC 31714) (T_STRING 31715) (T_VARIABLE 31716) (class_name 31717) (class_name_reference 31718) (name 31719) (new_variable 31720) (simple_variable 31721)) 14956 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31708) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14957 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31707) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14958 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31706) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14959 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31705) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14960 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31704) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14961 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31703) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14962 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31702) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14963 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31701) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14964 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31700) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14965 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31699) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14966 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31698) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14967 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31697) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14968 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31696) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14969 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31695) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14970 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31694) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14971 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31693) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14972 (("$" 31709) (T_VARIABLE 31716) (simple_variable 31740) ("{" 31741)) 14973 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 31738) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 14974 ((T_PAAMAYIM_NEKUDOTAYIM 31736)) 14975 ((T_NULLSAFE_OBJECT_OPERATOR 31722) (T_OBJECT_OPERATOR 31723) (T_PAAMAYIM_NEKUDOTAYIM 31724) ("[" 31725)) 14976 (("$" 31709) (T_STRING 31729) (T_VARIABLE 31716) (property_name 31735) (simple_variable 31731) ("{" 31732)) 14977 (("$" 31709) (T_STRING 31729) (T_VARIABLE 31716) (property_name 31730) (simple_variable 31731) ("{" 31732)) 14978 (("$" 31709) (T_VARIABLE 31716) (simple_variable 31728)) 14979 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 31726) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 14980 (("]" 31727)) 14981 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 31733) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 14982 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 31734)) 14983 (("$" 31709) (T_VARIABLE 31716) (simple_variable 31737)) 14984 (("%" 304) (")" 31739) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 14985 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 31742) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 14986 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 31743)) 14987 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31753) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14988 (("%" 768) ("*" 769) ("+" 770) ("-" 771) ("." 772) ("/" 773) (":" 31751) ("<" 774) (">" 775) ("?" 776) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 777) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 778) (T_BOOLEAN_AND 779) (T_BOOLEAN_OR 780) (T_COALESCE 781) (T_INSTANCEOF 782) (T_IS_EQUAL 783) (T_IS_GREATER_OR_EQUAL 784) (T_IS_IDENTICAL 785) (T_IS_NOT_EQUAL 786) (T_IS_NOT_IDENTICAL 787) (T_IS_SMALLER_OR_EQUAL 788) (T_LOGICAL_AND 789) (T_LOGICAL_OR 790) (T_LOGICAL_XOR 791) (T_PIPE 792) (T_POW 793) (T_SL 794) (T_SPACESHIP 795) (T_SR 796) ("^" 797) ("|" 798)) 14989 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31752) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14990 (("$" 31584) (T_ABSTRACT 31763) (T_ARRAY 31764) (T_AS 31765) (T_BREAK 31766) (T_CALLABLE 31767) (T_CASE 31768) (T_CATCH 31769) (T_CLASS 31770) (T_CLASS_C 31771) (T_CLONE 31772) (T_CONST 31773) (T_CONTINUE 31774) (T_DECLARE 31775) (T_DEFAULT 31776) (T_DIR 31777) (T_DO 31778) (T_ECHO 31779) (T_ELSE 31780) (T_ELSEIF 31781) (T_EMPTY 31782) (T_ENDDECLARE 31783) (T_ENDFOR 31784) (T_ENDFOREACH 31785) (T_ENDIF 31786) (T_ENDSWITCH 31787) (T_ENDWHILE 31788) (T_ENUM 31789) (T_EVAL 31790) (T_EXIT 31791) (T_EXTENDS 31792) (T_FILE 31793) (T_FINAL 31794) (T_FINALLY 31795) (T_FN 31796) (T_FOR 31797) (T_FOREACH 31798) (T_FUNCTION 31799) (T_FUNC_C 31800) (T_GLOBAL 31801) (T_GOTO 31802) (T_IF 31803) (T_IMPLEMENTS 31804) (T_INCLUDE 31805) (T_INCLUDE_ONCE 31806) (T_INSTANCEOF 31807) (T_INSTEADOF 31808) (T_INTERFACE 31809) (T_ISSET 31810) (T_LINE 31811) (T_LIST 31812) (T_LOGICAL_AND 31813) (T_LOGICAL_OR 31814) (T_LOGICAL_XOR 31815) (T_MATCH 31816) (T_METHOD_C 31817) (T_NAMESPACE 31818) (T_NEW 31819) (T_NS_C 31820) (T_PRINT 31821) (T_PRIVATE 31822) (T_PROPERTY_C 31823) (T_PROTECTED 31824) (T_PUBLIC 31825) (T_READONLY 31826) (T_REQUIRE 31827) (T_REQUIRE_ONCE 31828) (T_RETURN 31829) (T_STATIC 31830) (T_STRING 31831) (T_SWITCH 31832) (T_THROW 31833) (T_TRAIT 31834) (T_TRAIT_C 31835) (T_TRY 31836) (T_UNSET 31837) (T_USE 31838) (T_VAR 31839) (T_VARIABLE 31632) (T_WHILE 31840) (T_YIELD 31841) (identifier 31842) (member_name 31843) (reserved_non_modifiers 31844) (semi_reserved 31845) (simple_variable 31846) ("{" 31847)) 14991 (("(" 31850) (argument_list 31851)) 14992 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 31848) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 14993 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 31849)) 14994 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 31852) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 31853) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 31854) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 14995 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 31857) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 14996 (("," 1498) (possible_comma 31855)) 14997 ((")" 31856)) 14998 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (ampersand 31887) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31888) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 14999 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31886) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 15000 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31885) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 15001 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31884) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 15002 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31883) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 15003 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31882) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 15004 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31881) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 15005 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31880) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 15006 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31879) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 15007 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31878) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 15008 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31877) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 15009 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31876) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 15010 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31875) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 15011 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31874) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 15012 (("\"" 1538) ("$" 31889) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 31890) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 31891) ("[" 1560) (array_object_dereferenceable 31892) (callable_expr 31893) (callable_variable 31894) (class_constant 1564) (class_name 31895) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 31896) (name 31897) (new_dereferenceable 1570) (simple_variable 31898) (static_member 31899) (variable 31900) (variable_class_name 31901)) 15013 (("$" 31889) (T_VARIABLE 31891) (simple_variable 31934) ("{" 31935)) 15014 (("(" 31905) (argument_list 31933)) 15015 ((T_NULLSAFE_OBJECT_OPERATOR 31919) (T_OBJECT_OPERATOR 31920) ("[" 31921)) 15016 (("(" 31905) (argument_list 31918)) 15017 ((T_PAAMAYIM_NEKUDOTAYIM 31914)) 15018 (("(" 31905) (argument_list 31913)) 15019 ((T_PAAMAYIM_NEKUDOTAYIM 31902)) 15020 (("$" 31889) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 31891) (T_WHILE 1653) (T_YIELD 1654) (identifier 1655) (member_name 31903) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 31904) ("{" 1660)) 15021 (("(" 31905) (argument_list 31906)) 15022 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 31907) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 31908) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 31909) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 15023 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 31912) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 15024 (("," 1498) (possible_comma 31910)) 15025 ((")" 31911)) 15026 (("$" 31889) (T_ABSTRACT 1576) (T_ARRAY 1577) (T_AS 1578) (T_BREAK 1579) (T_CALLABLE 1580) (T_CASE 1581) (T_CATCH 1582) (T_CLASS 1583) (T_CLASS_C 1584) (T_CLONE 1585) (T_CONST 1586) (T_CONTINUE 1587) (T_DECLARE 1588) (T_DEFAULT 1589) (T_DIR 1590) (T_DO 1591) (T_ECHO 1592) (T_ELSE 1593) (T_ELSEIF 1594) (T_EMPTY 1595) (T_ENDDECLARE 1596) (T_ENDFOR 1597) (T_ENDFOREACH 1598) (T_ENDIF 1599) (T_ENDSWITCH 1600) (T_ENDWHILE 1601) (T_ENUM 1602) (T_EVAL 1603) (T_EXIT 1604) (T_EXTENDS 1605) (T_FILE 1606) (T_FINAL 1607) (T_FINALLY 1608) (T_FN 1609) (T_FOR 1610) (T_FOREACH 1611) (T_FUNCTION 1612) (T_FUNC_C 1613) (T_GLOBAL 1614) (T_GOTO 1615) (T_IF 1616) (T_IMPLEMENTS 1617) (T_INCLUDE 1618) (T_INCLUDE_ONCE 1619) (T_INSTANCEOF 1620) (T_INSTEADOF 1621) (T_INTERFACE 1622) (T_ISSET 1623) (T_LINE 1624) (T_LIST 1625) (T_LOGICAL_AND 1626) (T_LOGICAL_OR 1627) (T_LOGICAL_XOR 1628) (T_MATCH 1629) (T_METHOD_C 1630) (T_NAMESPACE 1631) (T_NEW 1632) (T_NS_C 1633) (T_PRINT 1634) (T_PRIVATE 1635) (T_PROPERTY_C 1636) (T_PROTECTED 1637) (T_PUBLIC 1638) (T_READONLY 1639) (T_REQUIRE 1640) (T_REQUIRE_ONCE 1641) (T_RETURN 1642) (T_STATIC 1643) (T_STRING 1644) (T_SWITCH 1645) (T_THROW 1646) (T_TRAIT 1647) (T_TRAIT_C 1648) (T_TRY 1649) (T_UNSET 1650) (T_USE 1651) (T_VAR 1652) (T_VARIABLE 31891) (T_WHILE 1653) (T_YIELD 1654) (identifier 1673) (member_name 31915) (reserved_non_modifiers 1657) (semi_reserved 1658) (simple_variable 31916) ("{" 1676)) 15027 (("(" 31905) (argument_list 31917)) 15028 (("$" 31889) (T_STRING 31924) (T_VARIABLE 31891) (property_name 31931) (simple_variable 31926) ("{" 31927)) 15029 (("$" 31889) (T_STRING 31924) (T_VARIABLE 31891) (property_name 31925) (simple_variable 31926) ("{" 31927)) 15030 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 31922) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 15031 (("]" 31923)) 15032 (("(" 31905) (argument_list 31930)) 15033 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 31928) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 15034 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 31929)) 15035 (("(" 31905) (argument_list 31932)) 15036 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 31936) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 15037 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 31937)) 15038 ((backup_doc_comment 31940)) 15039 (("(" 31941)) 15040 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 31942)) 15041 ((")" 31943)) 15042 ((T_USE 166) (lexical_vars 31944)) 15043 ((":" 168) (return_type 31945)) 15044 ((backup_fn_flags 31946)) 15045 (("{" 31947)) 15046 ((inner_statement_list 31948)) 15047 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 31949) ("~" 129)) 15048 ((backup_fn_flags 31950)) 15049 ((backup_doc_comment 31952)) 15050 (("(" 31953)) 15051 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 31954)) 15052 ((")" 31955)) 15053 ((":" 2271) (return_type 31956)) 15054 ((T_DOUBLE_ARROW 31957)) 15055 ((backup_fn_flags 31958)) 15056 ((backup_lex_pos 31959)) 15057 (("!" 30817) ("\"" 30818) ("$" 30819) ("(" 30820) ("+" 30821) ("-" 30822) ("@" 30823) (T_ARRAY 30824) (T_ARRAY_CAST 30825) (T_ATTRIBUTE 130) (T_BOOL_CAST 30826) (T_CLASS_C 30827) (T_CLONE 30828) (T_CONSTANT_ENCAPSED_STRING 30829) (T_DEC 30830) (T_DIR 30831) (T_DNUMBER 30832) (T_DOUBLE_CAST 30833) (T_EMPTY 30834) (T_EVAL 30835) (T_EXIT 30836) (T_FILE 30837) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 30838) (T_INC 30839) (T_INCLUDE 30840) (T_INCLUDE_ONCE 30841) (T_INT_CAST 30842) (T_ISSET 30843) (T_LINE 30844) (T_LIST 30845) (T_LNUMBER 30846) (T_MATCH 30847) (T_METHOD_C 30848) (T_NAME_FULLY_QUALIFIED 30849) (T_NAME_QUALIFIED 30850) (T_NAME_RELATIVE 30851) (T_NEW 30852) (T_NS_C 30853) (T_OBJECT_CAST 30854) (T_PRINT 30855) (T_PROPERTY_C 30856) (T_READONLY 30857) (T_REQUIRE 30858) (T_REQUIRE_ONCE 30859) (T_START_HEREDOC 30860) (T_STATIC 30861) (T_STRING 30862) (T_STRING_CAST 30863) (T_THROW 30864) (T_TRAIT_C 30865) (T_UNSET_CAST 30866) (T_VARIABLE 30867) (T_YIELD 30868) (T_YIELD_FROM 30869) ("[" 30870) ("`" 30871) (array_object_dereferenceable 30872) (attribute 135) (attributes 30873) (callable_expr 30874) (callable_variable 30875) (class_constant 30876) (class_name 30877) (constant 30878) (dereferenceable_scalar 30879) (expr 31960) (fn 30881) (fully_dereferenceable 107) #'30882 (function_call 30883) (inline_function 30884) (internal_functions_in_yacc 30885) (match 30886) (name 30887) (new_dereferenceable 30888) (new_non_dereferenceable 30889) (scalar 30890) (simple_variable 30891) (static_member 30892) (variable 30893) (variable_class_name 30894) ("~" 30895)) 15058 (("%" 30897) ("*" 30898) ("+" 30899) ("-" 30900) ("." 30901) ("/" 30902) ("<" 30903) (">" 30904) ("?" 30905) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 30906) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 30907) (T_BOOLEAN_AND 30908) (T_BOOLEAN_OR 30909) (T_COALESCE 30910) (T_INSTANCEOF 30911) (T_IS_EQUAL 30912) (T_IS_GREATER_OR_EQUAL 30913) (T_IS_IDENTICAL 30914) (T_IS_NOT_EQUAL 30915) (T_IS_NOT_IDENTICAL 30916) (T_IS_SMALLER_OR_EQUAL 30917) (T_LOGICAL_AND 30918) (T_LOGICAL_OR 30919) (T_LOGICAL_XOR 30920) (T_PIPE 30921) (T_POW 30922) (T_SL 30923) (T_SPACESHIP 30924) (T_SR 30925) ("^" 30926) (backup_fn_flags 31961) ("|" 30927)) 15059 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 31963) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 15060 (("$" 31584) (T_ABSTRACT 31763) (T_ARRAY 31764) (T_AS 31765) (T_BREAK 31766) (T_CALLABLE 31767) (T_CASE 31768) (T_CATCH 31769) (T_CLASS 31770) (T_CLASS_C 31771) (T_CLONE 31772) (T_CONST 31773) (T_CONTINUE 31774) (T_DECLARE 31775) (T_DEFAULT 31776) (T_DIR 31777) (T_DO 31778) (T_ECHO 31779) (T_ELSE 31780) (T_ELSEIF 31781) (T_EMPTY 31782) (T_ENDDECLARE 31783) (T_ENDFOR 31784) (T_ENDFOREACH 31785) (T_ENDIF 31786) (T_ENDSWITCH 31787) (T_ENDWHILE 31788) (T_ENUM 31789) (T_EVAL 31790) (T_EXIT 31791) (T_EXTENDS 31792) (T_FILE 31793) (T_FINAL 31794) (T_FINALLY 31795) (T_FN 31796) (T_FOR 31797) (T_FOREACH 31798) (T_FUNCTION 31799) (T_FUNC_C 31800) (T_GLOBAL 31801) (T_GOTO 31802) (T_IF 31803) (T_IMPLEMENTS 31804) (T_INCLUDE 31805) (T_INCLUDE_ONCE 31806) (T_INSTANCEOF 31807) (T_INSTEADOF 31808) (T_INTERFACE 31809) (T_ISSET 31810) (T_LINE 31811) (T_LIST 31812) (T_LOGICAL_AND 31813) (T_LOGICAL_OR 31814) (T_LOGICAL_XOR 31815) (T_MATCH 31816) (T_METHOD_C 31817) (T_NAMESPACE 31818) (T_NEW 31819) (T_NS_C 31820) (T_PRINT 31821) (T_PRIVATE 31822) (T_PROPERTY_C 31823) (T_PROTECTED 31824) (T_PUBLIC 31825) (T_READONLY 31826) (T_REQUIRE 31827) (T_REQUIRE_ONCE 31828) (T_RETURN 31829) (T_STATIC 31830) (T_STRING 31831) (T_SWITCH 31832) (T_THROW 31833) (T_TRAIT 31834) (T_TRAIT_C 31835) (T_TRY 31836) (T_UNSET 31837) (T_USE 31838) (T_VAR 31839) (T_VARIABLE 31632) (T_WHILE 31840) (T_YIELD 31841) (identifier 31965) (member_name 31966) (reserved_non_modifiers 31844) (semi_reserved 31845) (simple_variable 31967) ("{" 31968)) 15061 (("(" 31850) (argument_list 31971)) 15062 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 31969) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 15063 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 31970)) 15064 ((T_FN 35) (T_FUNCTION 131) (fn 31646) #'31647 (inline_function 31975)) 15065 (("$" 31584) (T_STRING 31981) (T_VARIABLE 31632) (property_name 31988) (simple_variable 31983) ("{" 31984)) 15066 (("$" 31584) (T_STRING 31981) (T_VARIABLE 31632) (property_name 31982) (simple_variable 31983) ("{" 31984)) 15067 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 31979) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 15068 (("]" 31980)) 15069 (("(" 31850) (argument_list 31987)) 15070 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 31985) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 15071 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 31986)) 15072 (("(" 31850) (argument_list 31989)) 15073 (("`" 31991)) 15074 (("]" 31993)) 15075 (("=" 31994)) 15076 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31995) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 15077 (("%" 31662) ("*" 31663) ("+" 31664) ("-" 31665) ("." 31666) ("/" 31667) ("<" 31668) (">" 31669) ("?" 31670) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 31671) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 31672) (T_BOOLEAN_AND 31673) (T_BOOLEAN_OR 31674) (T_COALESCE 31675) (T_DOUBLE_ARROW 31998) (T_INSTANCEOF 31676) (T_IS_EQUAL 31677) (T_IS_GREATER_OR_EQUAL 31678) (T_IS_IDENTICAL 31679) (T_IS_NOT_EQUAL 31680) (T_IS_NOT_IDENTICAL 31681) (T_IS_SMALLER_OR_EQUAL 31682) (T_LOGICAL_AND 31683) (T_LOGICAL_OR 31684) (T_LOGICAL_XOR 31685) (T_PIPE 31686) (T_POW 31687) (T_SL 31688) (T_SPACESHIP 31689) (T_SR 31690) ("^" 31691) ("|" 31692)) 15078 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 31999) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 15079 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 32008) (T_VARIABLE 2920) (encaps_var 2934)) 15080 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 32007) (T_VARIABLE 2920) (encaps_var 2925)) 15081 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 32049) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 15082 ((T_CLASS 32041)) 15083 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 32040) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 32021) (attribute 2982) (class_modifier 2963)) 15084 ((T_PAAMAYIM_NEKUDOTAYIM 32038)) 15085 (("(" 31905) (argument_list 32037)) 15086 ((T_NULLSAFE_OBJECT_OPERATOR 32028) (T_OBJECT_OPERATOR 32029) (T_PAAMAYIM_NEKUDOTAYIM 32030) ("[" 32031)) 15087 (("$" 31889) (T_STRING 31924) (T_VARIABLE 31891) (property_name 32036) (simple_variable 31926) ("{" 31927)) 15088 (("$" 31889) (T_STRING 31924) (T_VARIABLE 31891) (property_name 32035) (simple_variable 31926) ("{" 31927)) 15089 (("$" 31889) (T_VARIABLE 31891) (simple_variable 32034)) 15090 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 32032) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 15091 (("]" 32033)) 15092 (("$" 31889) (T_VARIABLE 31891) (simple_variable 32039)) 15093 (("(" 2984) (argument_list 2985) (ctor_arguments 32042)) 15094 ((T_EXTENDS 2987) (extends_from 32043)) 15095 ((T_IMPLEMENTS 2989) (implements_list 32044)) 15096 ((backup_doc_comment 32045)) 15097 (("{" 32046)) 15098 ((class_statement_list 32047)) 15099 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 32048)) 15100 (("%" 304) (")" 32050) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 15101 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 32052) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 15102 (("%" 304) (")" 32053) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 15103 (("{" 32054)) 15104 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 32055) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 15105 (("}" 32056)) 15106 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 32058) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 15107 ((")" 32059)) 15108 (("=" 32060)) 15109 (("!" 31582) ("\"" 31583) ("$" 31584) ("(" 31585) ("+" 31586) ("-" 31587) ("@" 31588) (T_ARRAY 31589) (T_ARRAY_CAST 31590) (T_ATTRIBUTE 130) (T_BOOL_CAST 31591) (T_CLASS_C 31592) (T_CLONE 31593) (T_CONSTANT_ENCAPSED_STRING 31594) (T_DEC 31595) (T_DIR 31596) (T_DNUMBER 31597) (T_DOUBLE_CAST 31598) (T_EMPTY 31599) (T_EVAL 31600) (T_EXIT 31601) (T_FILE 31602) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 31603) (T_INC 31604) (T_INCLUDE 31605) (T_INCLUDE_ONCE 31606) (T_INT_CAST 31607) (T_ISSET 31608) (T_LINE 31609) (T_LIST 31610) (T_LNUMBER 31611) (T_MATCH 31612) (T_METHOD_C 31613) (T_NAME_FULLY_QUALIFIED 31614) (T_NAME_QUALIFIED 31615) (T_NAME_RELATIVE 31616) (T_NEW 31617) (T_NS_C 31618) (T_OBJECT_CAST 31619) (T_PRINT 31620) (T_PROPERTY_C 31621) (T_READONLY 31622) (T_REQUIRE 31623) (T_REQUIRE_ONCE 31624) (T_START_HEREDOC 31625) (T_STATIC 31626) (T_STRING 31627) (T_STRING_CAST 31628) (T_THROW 31629) (T_TRAIT_C 31630) (T_UNSET_CAST 31631) (T_VARIABLE 31632) (T_YIELD 31633) (T_YIELD_FROM 31634) ("[" 31635) ("`" 31636) (array_object_dereferenceable 31637) (attribute 135) (attributes 31638) (callable_expr 31639) (callable_variable 31640) (class_constant 31641) (class_name 31642) (constant 31643) (dereferenceable_scalar 31644) (expr 32061) (fn 31646) (fully_dereferenceable 107) #'31647 (function_call 31648) (inline_function 31649) (internal_functions_in_yacc 31650) (match 31651) (name 31652) (new_dereferenceable 31653) (new_non_dereferenceable 31654) (scalar 31655) (simple_variable 31656) (static_member 31657) (variable 31658) (variable_class_name 31659) ("~" 31660)) 15110 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 32063) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 15111 (("," 5226) (possible_comma 32064)) 15112 ((")" 32065)) 15113 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 32073) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 32074) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 32075) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 15114 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 32078) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 15115 (("," 1498) (possible_comma 32076)) 15116 ((")" 32077)) 15117 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 32080) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 15118 (("%" 304) (")" 32081) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 15119 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 32083) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 15120 (("%" 304) (")" 32084) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 15121 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 32090) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 32091) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 32092) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 32093) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 15122 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 32099) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 15123 (("%" 2442) (")" 32096) ("*" 2443) ("+" 2444) ("," 32097) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 15124 (("," 5287) (possible_comma 32094)) 15125 ((")" 32095)) 15126 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 32098) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 15127 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 32103) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 15128 ((")" 32104)) 15129 (("%" 304) (")" 32096) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 15130 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 32111) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 15131 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 32112)) 15132 (("\"" 32114) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360)) 15133 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 32117) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 15134 (("%" 304) (")" 32118) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 15135 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 32119) (";" 8) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 15) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21) (T_DEC 22) (T_DECLARE 23) (T_DIR 24) (T_DNUMBER 25) (T_DO 26) (T_DOUBLE_CAST 27) (T_ECHO 28) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 36) (T_FOREACH 37) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 40) (T_GOTO 41) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 47) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 68) (T_START_HEREDOC 69) (T_STATIC 70) (T_STRING 71) (T_STRING_CAST 72) (T_SWITCH 73) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 77) (T_UNSET 78) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 82) (T_WHILE 83) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 88) (alt_if_stmt_without_else 89) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 105) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 111) (if_stmt_without_else 112) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 32120) (static_member 123) (variable 126) (variable_class_name 127) (while_statement 32121) ("{" 128) ("~" 129)) 15136 ((inner_statement_list 32122)) 15137 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17889) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17890) (T_CLASS 17891) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17892) (T_DEC 22) (T_DECLARE 17893) (T_DIR 24) (T_DNUMBER 25) (T_DO 17894) (T_DOUBLE_CAST 27) (T_ECHO 17895) (T_EMPTY 29) (T_ENDWHILE 32123) (T_ENUM 17897) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17898) (T_FOREACH 17899) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17900) (T_GOTO 17901) (T_HALT_COMPILER 17902) (T_IF 17903) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17904) (T_INTERFACE 17905) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17906) (T_START_HEREDOC 69) (T_STATIC 17907) (T_STRING 17908) (T_STRING_CAST 72) (T_SWITCH 17909) (T_THROW 74) (T_TRAIT 17910) (T_TRAIT_C 76) (T_TRY 17911) (T_UNSET 17912) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17913) (T_WHILE 17914) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17915) (alt_if_stmt_without_else 17916) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17917) (attributes 17918) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17919) (class_modifier 99) (class_modifiers 17920) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17921) (expr 17922) (fn 106) (fully_dereferenceable 107) #'17923 (function_call 109) (function_declaration_statement 17924) (if_stmt 17925) (if_stmt_without_else 17926) (inline_function 113) (inner_statement 17927) (interface_declaration_statement 17928) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17929) (static_member 123) (trait_declaration_statement 17930) (variable 126) (variable_class_name 127) ("{" 17931) ("~" 129)) 15138 ((";" 32124)) 15139 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 32126) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 15140 ((T_AS 32149) (T_NS_SEPARATOR 32167)) 15141 ((";" 32166)) 15142 (("," 32141) (";" 32165)) 15143 ((T_NAME_FULLY_QUALIFIED 32129) (T_NAME_QUALIFIED 32130) (T_STRING 32131) (group_use_declaration 32138) (legacy_namespace_name 32139) (namespace_name 32134) (use_declaration 32135) (use_declarations 32140)) 15144 ((";" 32164)) 15145 ((T_AS 32149) (T_NS_SEPARATOR 32151)) 15146 (("," 32141) (";" 32142)) 15147 ((T_NAME_FULLY_QUALIFIED 32143) (T_NAME_QUALIFIED 32144) (T_STRING 32145) (legacy_namespace_name 32146) (namespace_name 32147) (use_declaration 32148)) 15148 ((T_AS 32149)) 15149 ((T_STRING 32150)) 15150 (("{" 32152)) 15151 ((T_NAME_QUALIFIED 32153) (T_STRING 32154) (namespace_name 32155) (unprefixed_use_declaration 32156) (unprefixed_use_declarations 32157)) 15152 ((T_AS 32162)) 15153 (("," 32158) (possible_comma 32159)) 15154 ((T_NAME_QUALIFIED 32153) (T_STRING 32154) (namespace_name 32155) (unprefixed_use_declaration 32161)) 15155 (("}" 32160)) 15156 ((T_STRING 32163)) 15157 (("{" 32168)) 15158 ((T_CONST 32169) (T_FUNCTION 32170) (T_NAME_QUALIFIED 32153) (T_STRING 32154) (inline_use_declaration 32171) (inline_use_declarations 32172) (namespace_name 32155) (unprefixed_use_declaration 32173) (use_type 32174)) 15159 (("," 32176) (possible_comma 32177)) 15160 ((T_NAME_QUALIFIED 32153) (T_STRING 32154) (namespace_name 32155) (unprefixed_use_declaration 32175)) 15161 ((T_CONST 32169) (T_FUNCTION 32170) (T_NAME_QUALIFIED 32153) (T_STRING 32154) (inline_use_declaration 32179) (namespace_name 32155) (unprefixed_use_declaration 32173) (use_type 32174)) 15162 (("}" 32178)) 15163 (("\"" 1538) ("$" 17989) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 17990) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 17991) ("[" 1560) (array_object_dereferenceable 17992) (callable_expr 17993) (callable_variable 17994) (class_constant 1564) (class_name 17995) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 17996) (name 17997) (new_dereferenceable 1570) (simple_variable 17998) (static_member 17999) (unset_variable 18000) (unset_variables 32182) (variable 18002) (variable_class_name 18003)) 15164 (("," 18015) (possible_comma 32183)) 15165 ((")" 32184)) 15166 ((";" 32185)) 15167 ((inner_statement_list 32187)) 15168 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 32188) ("~" 129)) 15169 ((catch_list 32189)) 15170 ((T_CATCH 32190) (T_FINALLY 32191) (finally_statement 32192)) 15171 (("(" 32196)) 15172 (("{" 32193)) 15173 ((inner_statement_list 32194)) 15174 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 32195) ("~" 129)) 15175 ((T_NAME_FULLY_QUALIFIED 18056) (T_NAME_QUALIFIED 18057) (T_NAME_RELATIVE 18058) (T_STATIC 18059) (T_STRING 18060) (catch_name_list 32197) (class_name 18062) (name 18063)) 15176 ((T_VARIABLE 18064) (optional_variable 32198) ("|" 18066)) 15177 ((")" 32199)) 15178 (("{" 32200)) 15179 ((inner_statement_list 32201)) 15180 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 32202) ("~" 129)) 15181 ((backup_doc_comment 32204)) 15182 (("{" 32205)) 15183 ((class_statement_list 32206)) 15184 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 32207)) 15185 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 32210) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 15186 (("%" 304) (")" 32211) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 15187 ((":" 32212) (switch_case_list 32213) ("{" 32214)) 15188 ((";" 32220) (case_list 32221)) 15189 ((";" 32215) (case_list 32216)) 15190 ((case_list 32218)) 15191 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 32217)) 15192 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 32219)) 15193 ((case_list 32224)) 15194 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 32222)) 15195 ((";" 32223)) 15196 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 32225)) 15197 ((";" 32226)) 15198 (("," 18369) (";" 32230)) 15199 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_END_HEREDOC 32235) (T_VARIABLE 2920) (encaps_var 2934)) 15200 ((T_CURLY_OPEN 2916) (T_DOLLAR_OPEN_CURLY_BRACES 2917) (T_ENCAPSED_AND_WHITESPACE 2923) (T_END_HEREDOC 32234) (T_VARIABLE 2920) (encaps_var 2925)) 15201 ((";" 32237)) 15202 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 32277) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 15203 ((T_CLASS 32269)) 15204 ((T_ABSTRACT 10) (T_ATTRIBUTE 2951) (T_FINAL 34) (T_READONLY 2955) (anonymous_class 32268) (anonymous_class_modifiers 2959) (anonymous_class_modifiers_optional 32249) (attribute 2982) (class_modifier 2963)) 15205 ((T_PAAMAYIM_NEKUDOTAYIM 32266)) 15206 (("(" 30195) (argument_list 32265)) 15207 ((T_NULLSAFE_OBJECT_OPERATOR 32256) (T_OBJECT_OPERATOR 32257) (T_PAAMAYIM_NEKUDOTAYIM 32258) ("[" 32259)) 15208 (("$" 30179) (T_STRING 30214) (T_VARIABLE 30181) (property_name 32264) (simple_variable 30216) ("{" 30217)) 15209 (("$" 30179) (T_STRING 30214) (T_VARIABLE 30181) (property_name 32263) (simple_variable 30216) ("{" 30217)) 15210 (("$" 30179) (T_VARIABLE 30181) (simple_variable 32262)) 15211 (("!" 368) ("\"" 369) ("$" 370) ("(" 371) ("+" 372) ("-" 373) ("@" 374) (T_ARRAY 375) (T_ARRAY_CAST 376) (T_ATTRIBUTE 130) (T_BOOL_CAST 377) (T_CLASS_C 378) (T_CLONE 379) (T_CONSTANT_ENCAPSED_STRING 380) (T_DEC 381) (T_DIR 382) (T_DNUMBER 383) (T_DOUBLE_CAST 384) (T_EMPTY 385) (T_EVAL 386) (T_EXIT 387) (T_FILE 388) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 389) (T_INC 390) (T_INCLUDE 391) (T_INCLUDE_ONCE 392) (T_INT_CAST 393) (T_ISSET 394) (T_LINE 395) (T_LIST 396) (T_LNUMBER 397) (T_MATCH 398) (T_METHOD_C 399) (T_NAME_FULLY_QUALIFIED 400) (T_NAME_QUALIFIED 401) (T_NAME_RELATIVE 402) (T_NEW 403) (T_NS_C 404) (T_OBJECT_CAST 405) (T_PRINT 406) (T_PROPERTY_C 407) (T_READONLY 408) (T_REQUIRE 409) (T_REQUIRE_ONCE 410) (T_START_HEREDOC 411) (T_STATIC 412) (T_STRING 413) (T_STRING_CAST 414) (T_THROW 415) (T_TRAIT_C 416) (T_UNSET_CAST 417) (T_VARIABLE 418) (T_YIELD 419) (T_YIELD_FROM 420) ("[" 421) ("`" 422) (array_object_dereferenceable 423) (attribute 135) (attributes 424) (callable_expr 425) (callable_variable 426) (class_constant 427) (class_name 428) (constant 429) (dereferenceable_scalar 430) (expr 431) (fn 432) (fully_dereferenceable 107) #'433 (function_call 434) (inline_function 435) (internal_functions_in_yacc 436) (match 437) (name 438) (new_dereferenceable 439) (new_non_dereferenceable 440) (optional_expr 32260) (scalar 442) (simple_variable 443) (static_member 444) (variable 445) (variable_class_name 446) ("~" 447)) 15212 (("]" 32261)) 15213 (("$" 30179) (T_VARIABLE 30181) (simple_variable 32267)) 15214 (("(" 2984) (argument_list 2985) (ctor_arguments 32270)) 15215 ((T_EXTENDS 2987) (extends_from 32271)) 15216 ((T_IMPLEMENTS 2989) (implements_list 32272)) 15217 ((backup_doc_comment 32273)) 15218 (("{" 32274)) 15219 ((class_statement_list 32275)) 15220 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 32276)) 15221 (("%" 304) (")" 32278) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 15222 ((";" 32756) ("{" 32757)) 15223 ((top_statement_list 32364)) 15224 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 32365) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 32366) (T_CLASS 32367) (T_CLASS_C 17) (T_CLONE 18) (T_CONST 32368) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 32369) (T_DEC 22) (T_DECLARE 32370) (T_DIR 24) (T_DNUMBER 25) (T_DO 32371) (T_DOUBLE_CAST 27) (T_ECHO 32372) (T_EMPTY 29) (T_ENUM 32373) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 32374) (T_FOREACH 32375) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 32376) (T_GOTO 32377) (T_HALT_COMPILER 32378) (T_IF 32379) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 32380) (T_INTERFACE 32381) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAMESPACE 32382) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 32383) (T_START_HEREDOC 69) (T_STATIC 32384) (T_STRING 32385) (T_STRING_CAST 72) (T_SWITCH 32386) (T_THROW 74) (T_TRAIT 32387) (T_TRAIT_C 76) (T_TRY 32388) (T_UNSET 32389) (T_UNSET_CAST 79) (T_USE 32390) (T_VARIABLE 81) (T_VOID_CAST 32391) (T_WHILE 32392) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 32393) (alt_if_stmt_without_else 32394) (array_object_dereferenceable 90) (attribute 91) (attributed_statement 32395) (attributed_top_statement 32396) (attributes 32397) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 32398) (class_modifier 99) (class_modifiers 32399) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 32400) (expr 32401) (fn 106) (fully_dereferenceable 107) #'32402 (function_call 109) (function_declaration_statement 32403) (if_stmt 32404) (if_stmt_without_else 32405) (inline_function 113) (interface_declaration_statement 32406) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 32407) (static_member 123) (top_statement 32408) (trait_declaration_statement 32409) (variable 126) (variable_class_name 127) ("{" 32410) ("}" 32411) ("~" 129)) 15225 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 32754) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 15226 ((T_STRING 32747)) 15227 ((T_STRING 31538) (const_decl 31539) (const_list 32745)) 15228 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 32743) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 15229 (("(" 32734)) 15230 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22932) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22933) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22934) (T_DEC 22) (T_DECLARE 22935) (T_DIR 24) (T_DNUMBER 25) (T_DO 22936) (T_DOUBLE_CAST 27) (T_ECHO 22937) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22938) (T_FOREACH 22939) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22940) (T_GOTO 22941) (T_IF 22942) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22943) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22944) (T_START_HEREDOC 69) (T_STATIC 22945) (T_STRING 22946) (T_STRING_CAST 72) (T_SWITCH 22947) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22948) (T_UNSET 22949) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22950) (T_WHILE 22951) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22952) (alt_if_stmt_without_else 22953) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22954) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22955) (if_stmt_without_else 22956) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 32728) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22958) ("~" 129)) 15231 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (echo_expr 22926) (echo_expr_list 32726) (expr 22928) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 15232 ((T_STRING 32719)) 15233 (("(" 32706)) 15234 (("(" 32691)) 15235 (("$" 19441) (T_VARIABLE 19442) (global_var 19443) (global_var_list 32689) (simple_variable 19445)) 15236 ((T_STRING 32687)) 15237 (("(" 32684)) 15238 (("(" 32680)) 15239 ((T_STRING 32674)) 15240 ((T_ABSTRACT 32279) (T_ARRAY 32280) (T_AS 32281) (T_BREAK 32282) (T_CALLABLE 32283) (T_CASE 32284) (T_CATCH 32285) (T_CLASS 32286) (T_CLASS_C 32287) (T_CLONE 32288) (T_CONST 32289) (T_CONTINUE 32290) (T_DECLARE 32291) (T_DEFAULT 32292) (T_DIR 32293) (T_DO 32294) (T_ECHO 32295) (T_ELSE 32296) (T_ELSEIF 32297) (T_EMPTY 32298) (T_ENDDECLARE 32299) (T_ENDFOR 32300) (T_ENDFOREACH 32301) (T_ENDIF 32302) (T_ENDSWITCH 32303) (T_ENDWHILE 32304) (T_ENUM 32305) (T_EVAL 32306) (T_EXIT 32307) (T_EXTENDS 32308) (T_FILE 32309) (T_FINAL 32310) (T_FINALLY 32311) (T_FN 32312) (T_FOR 32313) (T_FOREACH 32314) (T_FUNCTION 32315) (T_FUNC_C 32316) (T_GLOBAL 32317) (T_GOTO 32318) (T_IF 32319) (T_IMPLEMENTS 32320) (T_INCLUDE 32321) (T_INCLUDE_ONCE 32322) (T_INSTANCEOF 32323) (T_INSTEADOF 32324) (T_INTERFACE 32325) (T_ISSET 32326) (T_LINE 32327) (T_LIST 32328) (T_LOGICAL_AND 32329) (T_LOGICAL_OR 32330) (T_LOGICAL_XOR 32331) (T_MATCH 32332) (T_METHOD_C 32333) (T_NAMESPACE 32334) (T_NAME_QUALIFIED 32335) (T_NEW 32336) (T_NS_C 32337) (T_PRINT 32338) (T_PRIVATE 32339) (T_PROPERTY_C 32340) (T_PROTECTED 32341) (T_PUBLIC 32342) (T_READONLY 32343) (T_REQUIRE 32344) (T_REQUIRE_ONCE 32345) (T_RETURN 32346) (T_STATIC 32347) (T_STRING 32348) (T_SWITCH 32349) (T_THROW 32350) (T_TRAIT 32351) (T_TRAIT_C 32352) (T_TRY 32353) (T_UNSET 32354) (T_USE 32355) (T_VAR 32356) (T_WHILE 32357) (T_YIELD 32358) (identifier 32359) (namespace_declaration_name 32666) (reserved_non_modifiers 32361) (semi_reserved 32362) ("{" 32667)) 15241 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 32664) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 15242 ((T_FN 35) (T_FUNCTION 131) (T_VARIABLE 18365) (fn 106) #'138 (inline_function 18366) (static_var 18367) (static_var_list 32662)) 15243 ((":" 32661)) 15244 (("(" 32643)) 15245 ((T_STRING 32638)) 15246 (("{" 32621)) 15247 (("(" 32616)) 15248 ((T_CONST 32127) (T_FUNCTION 32128) (T_NAME_FULLY_QUALIFIED 32129) (T_NAME_QUALIFIED 32130) (T_STRING 32131) (legacy_namespace_name 32132) (mixed_group_use_declaration 32607) (namespace_name 32134) (use_declaration 32135) (use_declarations 32608) (use_type 32609)) 15249 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 32605) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 15250 (("(" 32596)) 15251 ((T_ELSE 32589) (T_ELSEIF 17711) (T_ENDIF 32590)) 15252 ((T_ABSTRACT 10) (T_ATTRIBUTE 13) (T_CLASS 32367) (T_CONST 32368) (T_ENUM 32373) (T_FINAL 34) (T_FN 35) (T_FUNCTION 38) (T_INTERFACE 32381) (T_READONLY 2955) (T_STATIC 17797) (T_TRAIT 32387) (attribute 31490) (attributed_statement 32395) (attributed_top_statement 32588) (class_declaration_statement 32398) (class_modifier 99) (class_modifiers 32399) (enum_declaration_statement 32400) (fn 106) #'32402 (function_declaration_statement 32403) (inline_function 17800) (interface_declaration_statement 32406) (trait_declaration_statement 32409)) 15253 ((T_ABSTRACT 10) (T_CLASS 32580) (T_FINAL 34) (T_READONLY 2955) (class_modifier 17789)) 15254 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 32579) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 15255 ((T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 17770) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 17771) (ampersand 17772) (returns_ref 32567)) 15256 ((T_ELSE 32414) (T_ELSEIF 32415)) 15257 ((inner_statement_list 32412)) 15258 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 32413) ("~" 129)) 15259 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 32365) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 32366) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 32369) (T_DEC 22) (T_DECLARE 32370) (T_DIR 24) (T_DNUMBER 25) (T_DO 32371) (T_DOUBLE_CAST 27) (T_ECHO 32372) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 32374) (T_FOREACH 32375) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 32376) (T_GOTO 32377) (T_IF 32379) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 32380) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 32383) (T_START_HEREDOC 69) (T_STATIC 32384) (T_STRING 32385) (T_STRING_CAST 72) (T_SWITCH 32386) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 32388) (T_UNSET 32389) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 32391) (T_WHILE 32392) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 32393) (alt_if_stmt_without_else 32394) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 32401) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 32404) (if_stmt_without_else 32405) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 32566) (static_member 123) (variable 126) (variable_class_name 127) ("{" 32410) ("~" 129)) 15260 (("(" 32416)) 15261 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 32417) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 15262 (("%" 304) (")" 32418) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 15263 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 32419) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 32420) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 32421) (T_DEC 22) (T_DECLARE 32422) (T_DIR 24) (T_DNUMBER 25) (T_DO 32423) (T_DOUBLE_CAST 27) (T_ECHO 32424) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 32425) (T_FOREACH 32426) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 32427) (T_GOTO 32428) (T_IF 32379) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 32429) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 32430) (T_START_HEREDOC 69) (T_STATIC 32431) (T_STRING 32432) (T_STRING_CAST 72) (T_SWITCH 32433) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 32434) (T_UNSET 32435) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 32436) (T_WHILE 32437) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 32438) (alt_if_stmt_without_else 32439) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 32440) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 32441) (if_stmt_without_else 32442) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 32443) (static_member 123) (variable 126) (variable_class_name 127) ("{" 32444) ("~" 129)) 15264 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 32564) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 15265 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 32562) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 15266 (("(" 32553)) 15267 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22932) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22933) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22934) (T_DEC 22) (T_DECLARE 22935) (T_DIR 24) (T_DNUMBER 25) (T_DO 22936) (T_DOUBLE_CAST 27) (T_ECHO 22937) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22938) (T_FOREACH 22939) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22940) (T_GOTO 22941) (T_IF 22942) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22943) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22944) (T_START_HEREDOC 69) (T_STATIC 22945) (T_STRING 22946) (T_STRING_CAST 72) (T_SWITCH 22947) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22948) (T_UNSET 22949) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22950) (T_WHILE 22951) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22952) (alt_if_stmt_without_else 22953) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22954) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22955) (if_stmt_without_else 22956) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 32547) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22958) ("~" 129)) 15268 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (echo_expr 22926) (echo_expr_list 32545) (expr 22928) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 15269 (("(" 32532)) 15270 (("(" 32517)) 15271 (("$" 19441) (T_VARIABLE 19442) (global_var 19443) (global_var_list 32515) (simple_variable 19445)) 15272 ((T_STRING 32513)) 15273 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 32511) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 15274 ((T_FN 35) (T_FUNCTION 131) (T_VARIABLE 18365) (fn 106) #'138 (inline_function 18366) (static_var 18367) (static_var_list 32509)) 15275 ((":" 32508)) 15276 (("(" 32490)) 15277 (("{" 32473)) 15278 (("(" 32468)) 15279 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 32466) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 15280 (("(" 32457)) 15281 ((T_ELSE 32450) (T_ELSEIF 17711) (T_ENDIF 32451)) 15282 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 32449) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 15283 ((T_ELSE 32447) (T_ELSEIF 32415)) 15284 ((inner_statement_list 32445)) 15285 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 32446) ("~" 129)) 15286 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 32419) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 32420) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 32421) (T_DEC 22) (T_DECLARE 32422) (T_DIR 24) (T_DNUMBER 25) (T_DO 32423) (T_DOUBLE_CAST 27) (T_ECHO 32424) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 32425) (T_FOREACH 32426) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 32427) (T_GOTO 32428) (T_IF 32379) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 32429) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 32430) (T_START_HEREDOC 69) (T_STATIC 32431) (T_STRING 32432) (T_STRING_CAST 72) (T_SWITCH 32433) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 32434) (T_UNSET 32435) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 32436) (T_WHILE 32437) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 32438) (alt_if_stmt_without_else 32439) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 32440) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 32441) (if_stmt_without_else 32442) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 32448) (static_member 123) (variable 126) (variable_class_name 127) ("{" 32444) ("~" 129)) 15287 ((":" 32453)) 15288 ((";" 32452)) 15289 ((inner_statement_list 32454)) 15290 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17807) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17808) (T_CLASS 17809) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17810) (T_DEC 22) (T_DECLARE 17811) (T_DIR 24) (T_DNUMBER 25) (T_DO 17812) (T_DOUBLE_CAST 27) (T_ECHO 17813) (T_EMPTY 29) (T_ENDIF 32455) (T_ENUM 17815) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17816) (T_FOREACH 17817) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17818) (T_GOTO 17819) (T_HALT_COMPILER 17820) (T_IF 17732) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17821) (T_INTERFACE 17822) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17823) (T_START_HEREDOC 69) (T_STATIC 17824) (T_STRING 17825) (T_STRING_CAST 72) (T_SWITCH 17826) (T_THROW 74) (T_TRAIT 17827) (T_TRAIT_C 76) (T_TRY 17828) (T_UNSET 17829) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17830) (T_WHILE 17831) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17832) (alt_if_stmt_without_else 17833) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17834) (attributes 17835) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17836) (class_modifier 99) (class_modifiers 17837) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17838) (expr 17839) (fn 106) (fully_dereferenceable 107) #'17840 (function_call 109) (function_declaration_statement 17841) (if_stmt 17842) (if_stmt_without_else 17843) (inline_function 113) (inner_statement 17844) (interface_declaration_statement 17845) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17846) (static_member 123) (trait_declaration_statement 17847) (variable 126) (variable_class_name 127) ("{" 17848) ("~" 129)) 15291 ((";" 32456)) 15292 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 32458) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 15293 (("%" 304) (")" 32459) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 15294 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 32460) (";" 32419) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 32420) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 32421) (T_DEC 22) (T_DECLARE 32422) (T_DIR 24) (T_DNUMBER 25) (T_DO 32423) (T_DOUBLE_CAST 27) (T_ECHO 32424) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 32425) (T_FOREACH 32426) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 32427) (T_GOTO 32428) (T_IF 32379) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 32429) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 32430) (T_START_HEREDOC 69) (T_STATIC 32431) (T_STRING 32432) (T_STRING_CAST 72) (T_SWITCH 32433) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 32434) (T_UNSET 32435) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 32436) (T_WHILE 32437) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 32438) (alt_if_stmt_without_else 32439) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 32440) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 32441) (if_stmt_without_else 32442) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 32461) (static_member 123) (variable 126) (variable_class_name 127) (while_statement 32462) ("{" 32444) ("~" 129)) 15295 ((inner_statement_list 32463)) 15296 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17889) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17890) (T_CLASS 17891) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17892) (T_DEC 22) (T_DECLARE 17893) (T_DIR 24) (T_DNUMBER 25) (T_DO 17894) (T_DOUBLE_CAST 27) (T_ECHO 17895) (T_EMPTY 29) (T_ENDWHILE 32464) (T_ENUM 17897) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17898) (T_FOREACH 17899) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17900) (T_GOTO 17901) (T_HALT_COMPILER 17902) (T_IF 17903) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17904) (T_INTERFACE 17905) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17906) (T_START_HEREDOC 69) (T_STATIC 17907) (T_STRING 17908) (T_STRING_CAST 72) (T_SWITCH 17909) (T_THROW 74) (T_TRAIT 17910) (T_TRAIT_C 76) (T_TRY 17911) (T_UNSET 17912) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17913) (T_WHILE 17914) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17915) (alt_if_stmt_without_else 17916) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17917) (attributes 17918) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17919) (class_modifier 99) (class_modifiers 17920) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17921) (expr 17922) (fn 106) (fully_dereferenceable 107) #'17923 (function_call 109) (function_declaration_statement 17924) (if_stmt 17925) (if_stmt_without_else 17926) (inline_function 113) (inner_statement 17927) (interface_declaration_statement 17928) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17929) (static_member 123) (trait_declaration_statement 17930) (variable 126) (variable_class_name 127) ("{" 17931) ("~" 129)) 15297 ((";" 32465)) 15298 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 32467) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 15299 (("\"" 1538) ("$" 17989) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 17990) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 17991) ("[" 1560) (array_object_dereferenceable 17992) (callable_expr 17993) (callable_variable 17994) (class_constant 1564) (class_name 17995) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 17996) (name 17997) (new_dereferenceable 1570) (simple_variable 17998) (static_member 17999) (unset_variable 18000) (unset_variables 32469) (variable 18002) (variable_class_name 18003)) 15300 (("," 18015) (possible_comma 32470)) 15301 ((")" 32471)) 15302 ((";" 32472)) 15303 ((inner_statement_list 32474)) 15304 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 32475) ("~" 129)) 15305 ((catch_list 32476)) 15306 ((T_CATCH 32477) (T_FINALLY 32478) (finally_statement 32479)) 15307 (("(" 32483)) 15308 (("{" 32480)) 15309 ((inner_statement_list 32481)) 15310 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 32482) ("~" 129)) 15311 ((T_NAME_FULLY_QUALIFIED 18056) (T_NAME_QUALIFIED 18057) (T_NAME_RELATIVE 18058) (T_STATIC 18059) (T_STRING 18060) (catch_name_list 32484) (class_name 18062) (name 18063)) 15312 ((T_VARIABLE 18064) (optional_variable 32485) ("|" 18066)) 15313 ((")" 32486)) 15314 (("{" 32487)) 15315 ((inner_statement_list 32488)) 15316 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 32489) ("~" 129)) 15317 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 32491) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 15318 (("%" 304) (")" 32492) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 15319 ((":" 32493) (switch_case_list 32494) ("{" 32495)) 15320 ((";" 32501) (case_list 32502)) 15321 ((";" 32496) (case_list 32497)) 15322 ((case_list 32499)) 15323 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 32498)) 15324 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 32500)) 15325 ((case_list 32505)) 15326 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 32503)) 15327 ((";" 32504)) 15328 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 32506)) 15329 ((";" 32507)) 15330 (("," 18369) (";" 32510)) 15331 ((";" 32512)) 15332 ((";" 32514)) 15333 (("," 19446) (";" 32516)) 15334 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 32518) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 15335 (("%" 19534) ("*" 19535) ("+" 19536) ("-" 19537) ("." 19538) ("/" 19539) ("<" 19540) (">" 19541) ("?" 19542) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19543) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19544) (T_AS 32519) (T_BOOLEAN_AND 19545) (T_BOOLEAN_OR 19546) (T_COALESCE 19547) (T_INSTANCEOF 19548) (T_IS_EQUAL 19549) (T_IS_GREATER_OR_EQUAL 19550) (T_IS_IDENTICAL 19551) (T_IS_NOT_EQUAL 19552) (T_IS_NOT_IDENTICAL 19553) (T_IS_SMALLER_OR_EQUAL 19554) (T_LOGICAL_AND 19555) (T_LOGICAL_OR 19556) (T_LOGICAL_XOR 19557) (T_PIPE 19558) (T_POW 19559) (T_SL 19560) (T_SPACESHIP 19561) (T_SR 19562) ("^" 19563) ("|" 19564)) 15336 (("\"" 1538) ("$" 20897) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20898) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20899) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20900) ("[" 20901) (ampersand 20902) (array_object_dereferenceable 20903) (callable_expr 20904) (callable_variable 20905) (class_constant 1564) (class_name 20906) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 32520) (fully_dereferenceable 107) (function_call 20908) (name 20909) (new_dereferenceable 1570) (simple_variable 20910) (static_member 20911) (variable 20912) (variable_class_name 20913)) 15337 ((")" 32521) (T_DOUBLE_ARROW 32522)) 15338 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 32525) (";" 32419) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 32420) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 32421) (T_DEC 22) (T_DECLARE 32422) (T_DIR 24) (T_DNUMBER 25) (T_DO 32423) (T_DOUBLE_CAST 27) (T_ECHO 32424) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 32425) (T_FOREACH 32426) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 32427) (T_GOTO 32428) (T_IF 32379) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 32429) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 32430) (T_START_HEREDOC 69) (T_STATIC 32431) (T_STRING 32432) (T_STRING_CAST 72) (T_SWITCH 32433) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 32434) (T_UNSET 32435) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 32436) (T_WHILE 32437) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 32438) (alt_if_stmt_without_else 32439) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 32440) (fn 106) (foreach_statement 32531) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 32441) (if_stmt_without_else 32442) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 32527) (static_member 123) (variable 126) (variable_class_name 127) ("{" 32444) ("~" 129)) 15339 (("\"" 1538) ("$" 20928) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20929) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20930) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20931) ("[" 20932) (ampersand 20933) (array_object_dereferenceable 20934) (callable_expr 20935) (callable_variable 20936) (class_constant 1564) (class_name 20937) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 32523) (fully_dereferenceable 107) (function_call 20939) (name 20940) (new_dereferenceable 1570) (simple_variable 20941) (static_member 20942) (variable 20943) (variable_class_name 20944)) 15340 ((")" 32524)) 15341 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 32525) (";" 32419) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 32420) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 32421) (T_DEC 22) (T_DECLARE 32422) (T_DIR 24) (T_DNUMBER 25) (T_DO 32423) (T_DOUBLE_CAST 27) (T_ECHO 32424) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 32425) (T_FOREACH 32426) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 32427) (T_GOTO 32428) (T_IF 32379) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 32429) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 32430) (T_START_HEREDOC 69) (T_STATIC 32431) (T_STRING 32432) (T_STRING_CAST 72) (T_SWITCH 32433) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 32434) (T_UNSET 32435) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 32436) (T_WHILE 32437) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 32438) (alt_if_stmt_without_else 32439) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 32440) (fn 106) (foreach_statement 32526) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 32441) (if_stmt_without_else 32442) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 32527) (static_member 123) (variable 126) (variable_class_name 127) ("{" 32444) ("~" 129)) 15342 ((inner_statement_list 32528)) 15343 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20962) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 20963) (T_CLASS 20964) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20965) (T_DEC 22) (T_DECLARE 20966) (T_DIR 24) (T_DNUMBER 25) (T_DO 20967) (T_DOUBLE_CAST 27) (T_ECHO 20968) (T_EMPTY 29) (T_ENDFOREACH 32529) (T_ENUM 20970) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20971) (T_FOREACH 20972) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20973) (T_GOTO 20974) (T_HALT_COMPILER 20975) (T_IF 20976) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20977) (T_INTERFACE 20978) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20979) (T_START_HEREDOC 69) (T_STATIC 20980) (T_STRING 20981) (T_STRING_CAST 72) (T_SWITCH 20982) (T_THROW 74) (T_TRAIT 20983) (T_TRAIT_C 76) (T_TRY 20984) (T_UNSET 20985) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 20986) (T_WHILE 20987) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 20988) (alt_if_stmt_without_else 20989) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 20990) (attributes 20991) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 20992) (class_modifier 99) (class_modifiers 20993) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 20994) (expr 20995) (fn 106) (fully_dereferenceable 107) #'20996 (function_call 109) (function_declaration_statement 20997) (if_stmt 20998) (if_stmt_without_else 20999) (inline_function 113) (inner_statement 21000) (interface_declaration_statement 21001) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 21002) (static_member 123) (trait_declaration_statement 21003) (variable 126) (variable_class_name 127) ("{" 21004) ("~" 129)) 15344 ((";" 32530)) 15345 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21126) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21127) (fn 18437) (for_exprs 32533) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21129) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 15346 ((";" 32534)) 15347 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21135) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21136) (fn 18437) (for_cond_exprs 32535) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21138) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 15348 ((";" 32536)) 15349 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_VOID_CAST 22737) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 22738) (fn 2425) (for_exprs 32537) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_for_exprs 22740) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 15350 ((")" 32538)) 15351 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 32539) (";" 32419) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 32420) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 32421) (T_DEC 22) (T_DECLARE 32422) (T_DIR 24) (T_DNUMBER 25) (T_DO 32423) (T_DOUBLE_CAST 27) (T_ECHO 32424) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 32425) (T_FOREACH 32426) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 32427) (T_GOTO 32428) (T_IF 32379) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 32429) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 32430) (T_START_HEREDOC 69) (T_STATIC 32431) (T_STRING 32432) (T_STRING_CAST 72) (T_SWITCH 32433) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 32434) (T_UNSET 32435) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 32436) (T_WHILE 32437) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 32438) (alt_if_stmt_without_else 32439) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 32440) (fn 106) (for_statement 32540) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 32441) (if_stmt_without_else 32442) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 32541) (static_member 123) (variable 126) (variable_class_name 127) ("{" 32444) ("~" 129)) 15352 ((inner_statement_list 32542)) 15353 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22750) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 22751) (T_CLASS 22752) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22753) (T_DEC 22) (T_DECLARE 22754) (T_DIR 24) (T_DNUMBER 25) (T_DO 22755) (T_DOUBLE_CAST 27) (T_ECHO 22756) (T_EMPTY 29) (T_ENDFOR 32543) (T_ENUM 22758) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 22759) (T_FOREACH 22760) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 22761) (T_GOTO 22762) (T_HALT_COMPILER 22763) (T_IF 22764) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22765) (T_INTERFACE 22766) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22767) (T_START_HEREDOC 69) (T_STATIC 22768) (T_STRING 22769) (T_STRING_CAST 72) (T_SWITCH 22770) (T_THROW 74) (T_TRAIT 22771) (T_TRAIT_C 76) (T_TRY 22772) (T_UNSET 22773) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22774) (T_WHILE 22775) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22776) (alt_if_stmt_without_else 22777) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 22778) (attributes 22779) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 22780) (class_modifier 99) (class_modifiers 22781) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 22782) (expr 22783) (fn 106) (fully_dereferenceable 107) #'22784 (function_call 109) (function_declaration_statement 22785) (if_stmt 22786) (if_stmt_without_else 22787) (inline_function 113) (inner_statement 22788) (interface_declaration_statement 22789) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 22790) (static_member 123) (trait_declaration_statement 22791) (variable 126) (variable_class_name 127) ("{" 22792) ("~" 129)) 15354 ((";" 32544)) 15355 (("," 22929) (";" 32546)) 15356 ((T_WHILE 32548)) 15357 (("(" 32549)) 15358 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 32550) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 15359 (("%" 304) (")" 32551) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 15360 ((";" 32552)) 15361 ((T_STRING 23106) (const_decl 23107) (const_list 32554)) 15362 ((")" 32555) ("," 23110)) 15363 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 32556) (";" 32419) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 32420) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 32421) (T_DEC 22) (T_DECLARE 32422) (T_DIR 24) (T_DNUMBER 25) (T_DO 32423) (T_DOUBLE_CAST 27) (T_ECHO 32424) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 32425) (T_FOREACH 32426) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 32427) (T_GOTO 32428) (T_IF 32379) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 32429) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 32430) (T_START_HEREDOC 69) (T_STATIC 32431) (T_STRING 32432) (T_STRING_CAST 72) (T_SWITCH 32433) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 32434) (T_UNSET 32435) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 32436) (T_WHILE 32437) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 32438) (alt_if_stmt_without_else 32439) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (declare_statement 32557) (dereferenceable_scalar 103) (expr 32440) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 32441) (if_stmt_without_else 32442) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 32558) (static_member 123) (variable 126) (variable_class_name 127) ("{" 32444) ("~" 129)) 15364 ((inner_statement_list 32559)) 15365 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 23116) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 23117) (T_CLASS 23118) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 23119) (T_DEC 22) (T_DECLARE 23120) (T_DIR 24) (T_DNUMBER 25) (T_DO 23121) (T_DOUBLE_CAST 27) (T_ECHO 23122) (T_EMPTY 29) (T_ENDDECLARE 32560) (T_ENUM 23124) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 23125) (T_FOREACH 23126) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 23127) (T_GOTO 23128) (T_HALT_COMPILER 23129) (T_IF 23130) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 23131) (T_INTERFACE 23132) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 23133) (T_START_HEREDOC 69) (T_STATIC 23134) (T_STRING 23135) (T_STRING_CAST 72) (T_SWITCH 23136) (T_THROW 74) (T_TRAIT 23137) (T_TRAIT_C 76) (T_TRY 23138) (T_UNSET 23139) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 23140) (T_WHILE 23141) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 23142) (alt_if_stmt_without_else 23143) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 23144) (attributes 23145) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 23146) (class_modifier 99) (class_modifiers 23147) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 23148) (expr 23149) (fn 106) (fully_dereferenceable 107) #'23150 (function_call 109) (function_declaration_statement 23151) (if_stmt 23152) (if_stmt_without_else 23153) (inline_function 113) (inner_statement 23154) (interface_declaration_statement 23155) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23156) (static_member 123) (trait_declaration_statement 23157) (variable 126) (variable_class_name 127) ("{" 23158) ("~" 129)) 15366 ((";" 32561)) 15367 ((";" 32563)) 15368 ((";" 32565)) 15369 ((T_READONLY 17774) (T_STRING 17775) (backup_doc_comment 143) (function_name 32568)) 15370 ((backup_doc_comment 32569)) 15371 (("(" 32570)) 15372 ((T_ABSTRACT 145) (T_ATTRIBUTE 146) (T_FINAL 147) (T_PRIVATE 148) (T_PRIVATE_SET 149) (T_PROTECTED 150) (T_PROTECTED_SET 151) (T_PUBLIC 152) (T_PUBLIC_SET 153) (T_READONLY 154) (T_STATIC 155) (attribute 156) (attributed_parameter 157) (attributes 158) (member_modifier 159) (non_empty_member_modifiers 160) (non_empty_parameter_list 161) (optional_cpp_modifiers 162) (parameter 163) (parameter_list 32571)) 15373 ((")" 32572)) 15374 ((":" 168) (return_type 32573)) 15375 ((backup_fn_flags 32574)) 15376 (("{" 32575)) 15377 ((inner_statement_list 32576)) 15378 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 32577) ("~" 129)) 15379 ((backup_fn_flags 32578)) 15380 ((T_STRING 32581)) 15381 ((T_EXTENDS 2987) (extends_from 32582)) 15382 ((T_IMPLEMENTS 2989) (implements_list 32583)) 15383 ((backup_doc_comment 32584)) 15384 (("{" 32585)) 15385 ((class_statement_list 32586)) 15386 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 32587)) 15387 ((":" 32592)) 15388 ((";" 32591)) 15389 ((inner_statement_list 32593)) 15390 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17807) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17808) (T_CLASS 17809) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17810) (T_DEC 22) (T_DECLARE 17811) (T_DIR 24) (T_DNUMBER 25) (T_DO 17812) (T_DOUBLE_CAST 27) (T_ECHO 17813) (T_EMPTY 29) (T_ENDIF 32594) (T_ENUM 17815) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17816) (T_FOREACH 17817) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17818) (T_GOTO 17819) (T_HALT_COMPILER 17820) (T_IF 17732) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17821) (T_INTERFACE 17822) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17823) (T_START_HEREDOC 69) (T_STATIC 17824) (T_STRING 17825) (T_STRING_CAST 72) (T_SWITCH 17826) (T_THROW 74) (T_TRAIT 17827) (T_TRAIT_C 76) (T_TRY 17828) (T_UNSET 17829) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17830) (T_WHILE 17831) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17832) (alt_if_stmt_without_else 17833) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17834) (attributes 17835) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17836) (class_modifier 99) (class_modifiers 17837) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17838) (expr 17839) (fn 106) (fully_dereferenceable 107) #'17840 (function_call 109) (function_declaration_statement 17841) (if_stmt 17842) (if_stmt_without_else 17843) (inline_function 113) (inner_statement 17844) (interface_declaration_statement 17845) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17846) (static_member 123) (trait_declaration_statement 17847) (variable 126) (variable_class_name 127) ("{" 17848) ("~" 129)) 15391 ((";" 32595)) 15392 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 32597) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 15393 (("%" 304) (")" 32598) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 15394 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 32599) (";" 32365) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 32366) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 32369) (T_DEC 22) (T_DECLARE 32370) (T_DIR 24) (T_DNUMBER 25) (T_DO 32371) (T_DOUBLE_CAST 27) (T_ECHO 32372) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 32374) (T_FOREACH 32375) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 32376) (T_GOTO 32377) (T_IF 32379) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 32380) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 32383) (T_START_HEREDOC 69) (T_STATIC 32384) (T_STRING 32385) (T_STRING_CAST 72) (T_SWITCH 32386) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 32388) (T_UNSET 32389) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 32391) (T_WHILE 32392) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 32393) (alt_if_stmt_without_else 32394) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 32401) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 32404) (if_stmt_without_else 32405) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 32600) (static_member 123) (variable 126) (variable_class_name 127) (while_statement 32601) ("{" 32410) ("~" 129)) 15395 ((inner_statement_list 32602)) 15396 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17889) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17890) (T_CLASS 17891) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17892) (T_DEC 22) (T_DECLARE 17893) (T_DIR 24) (T_DNUMBER 25) (T_DO 17894) (T_DOUBLE_CAST 27) (T_ECHO 17895) (T_EMPTY 29) (T_ENDWHILE 32603) (T_ENUM 17897) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17898) (T_FOREACH 17899) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17900) (T_GOTO 17901) (T_HALT_COMPILER 17902) (T_IF 17903) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17904) (T_INTERFACE 17905) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17906) (T_START_HEREDOC 69) (T_STATIC 17907) (T_STRING 17908) (T_STRING_CAST 72) (T_SWITCH 17909) (T_THROW 74) (T_TRAIT 17910) (T_TRAIT_C 76) (T_TRY 17911) (T_UNSET 17912) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17913) (T_WHILE 17914) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17915) (alt_if_stmt_without_else 17916) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17917) (attributes 17918) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17919) (class_modifier 99) (class_modifiers 17920) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17921) (expr 17922) (fn 106) (fully_dereferenceable 107) #'17923 (function_call 109) (function_declaration_statement 17924) (if_stmt 17925) (if_stmt_without_else 17926) (inline_function 113) (inner_statement 17927) (interface_declaration_statement 17928) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17929) (static_member 123) (trait_declaration_statement 17930) (variable 126) (variable_class_name 127) ("{" 17931) ("~" 129)) 15397 ((";" 32604)) 15398 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 32606) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 15399 ((";" 32615)) 15400 (("," 32141) (";" 32614)) 15401 ((T_NAME_FULLY_QUALIFIED 32129) (T_NAME_QUALIFIED 32130) (T_STRING 32131) (group_use_declaration 32610) (legacy_namespace_name 32139) (namespace_name 32134) (use_declaration 32135) (use_declarations 32611)) 15402 ((";" 32613)) 15403 (("," 32141) (";" 32612)) 15404 (("\"" 1538) ("$" 17989) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 17990) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 17991) ("[" 1560) (array_object_dereferenceable 17992) (callable_expr 17993) (callable_variable 17994) (class_constant 1564) (class_name 17995) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 17996) (name 17997) (new_dereferenceable 1570) (simple_variable 17998) (static_member 17999) (unset_variable 18000) (unset_variables 32617) (variable 18002) (variable_class_name 18003)) 15405 (("," 18015) (possible_comma 32618)) 15406 ((")" 32619)) 15407 ((";" 32620)) 15408 ((inner_statement_list 32622)) 15409 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 32623) ("~" 129)) 15410 ((catch_list 32624)) 15411 ((T_CATCH 32625) (T_FINALLY 32626) (finally_statement 32627)) 15412 (("(" 32631)) 15413 (("{" 32628)) 15414 ((inner_statement_list 32629)) 15415 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 32630) ("~" 129)) 15416 ((T_NAME_FULLY_QUALIFIED 18056) (T_NAME_QUALIFIED 18057) (T_NAME_RELATIVE 18058) (T_STATIC 18059) (T_STRING 18060) (catch_name_list 32632) (class_name 18062) (name 18063)) 15417 ((T_VARIABLE 18064) (optional_variable 32633) ("|" 18066)) 15418 ((")" 32634)) 15419 (("{" 32635)) 15420 ((inner_statement_list 32636)) 15421 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 32637) ("~" 129)) 15422 ((backup_doc_comment 32639)) 15423 (("{" 32640)) 15424 ((class_statement_list 32641)) 15425 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 32642)) 15426 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 32644) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 15427 (("%" 304) (")" 32645) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 15428 ((":" 32646) (switch_case_list 32647) ("{" 32648)) 15429 ((";" 32654) (case_list 32655)) 15430 ((";" 32649) (case_list 32650)) 15431 ((case_list 32652)) 15432 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 32651)) 15433 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 32653)) 15434 ((case_list 32658)) 15435 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 32656)) 15436 ((";" 32657)) 15437 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 32659)) 15438 ((";" 32660)) 15439 (("," 18369) (";" 32663)) 15440 ((";" 32665)) 15441 ((";" 32670) ("{" 32671)) 15442 ((top_statement_list 32668)) 15443 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 32365) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 32366) (T_CLASS 32367) (T_CLASS_C 17) (T_CLONE 18) (T_CONST 32368) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 32369) (T_DEC 22) (T_DECLARE 32370) (T_DIR 24) (T_DNUMBER 25) (T_DO 32371) (T_DOUBLE_CAST 27) (T_ECHO 32372) (T_EMPTY 29) (T_ENUM 32373) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 32374) (T_FOREACH 32375) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 32376) (T_GOTO 32377) (T_HALT_COMPILER 32378) (T_IF 32379) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 32380) (T_INTERFACE 32381) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAMESPACE 32382) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 32383) (T_START_HEREDOC 69) (T_STATIC 32384) (T_STRING 32385) (T_STRING_CAST 72) (T_SWITCH 32386) (T_THROW 74) (T_TRAIT 32387) (T_TRAIT_C 76) (T_TRY 32388) (T_UNSET 32389) (T_UNSET_CAST 79) (T_USE 32390) (T_VARIABLE 81) (T_VOID_CAST 32391) (T_WHILE 32392) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 32393) (alt_if_stmt_without_else 32394) (array_object_dereferenceable 90) (attribute 91) (attributed_statement 32395) (attributed_top_statement 32396) (attributes 32397) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 32398) (class_modifier 99) (class_modifiers 32399) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 32400) (expr 32401) (fn 106) (fully_dereferenceable 107) #'32402 (function_call 109) (function_declaration_statement 32403) (if_stmt 32404) (if_stmt_without_else 32405) (inline_function 113) (interface_declaration_statement 32406) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 32407) (static_member 123) (top_statement 32408) (trait_declaration_statement 32409) (variable 126) (variable_class_name 127) ("{" 32410) ("}" 32669) ("~" 129)) 15444 ((top_statement_list 32672)) 15445 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 32365) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 32366) (T_CLASS 32367) (T_CLASS_C 17) (T_CLONE 18) (T_CONST 32368) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 32369) (T_DEC 22) (T_DECLARE 32370) (T_DIR 24) (T_DNUMBER 25) (T_DO 32371) (T_DOUBLE_CAST 27) (T_ECHO 32372) (T_EMPTY 29) (T_ENUM 32373) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 32374) (T_FOREACH 32375) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 32376) (T_GOTO 32377) (T_HALT_COMPILER 32378) (T_IF 32379) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 32380) (T_INTERFACE 32381) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAMESPACE 32382) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 32383) (T_START_HEREDOC 69) (T_STATIC 32384) (T_STRING 32385) (T_STRING_CAST 72) (T_SWITCH 32386) (T_THROW 74) (T_TRAIT 32387) (T_TRAIT_C 76) (T_TRY 32388) (T_UNSET 32389) (T_UNSET_CAST 79) (T_USE 32390) (T_VARIABLE 81) (T_VOID_CAST 32391) (T_WHILE 32392) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 32393) (alt_if_stmt_without_else 32394) (array_object_dereferenceable 90) (attribute 91) (attributed_statement 32395) (attributed_top_statement 32396) (attributes 32397) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 32398) (class_modifier 99) (class_modifiers 32399) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 32400) (expr 32401) (fn 106) (fully_dereferenceable 107) #'32402 (function_call 109) (function_declaration_statement 32403) (if_stmt 32404) (if_stmt_without_else 32405) (inline_function 113) (interface_declaration_statement 32406) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 32407) (static_member 123) (top_statement 32408) (trait_declaration_statement 32409) (variable 126) (variable_class_name 127) ("{" 32410) ("}" 32673) ("~" 129)) 15446 ((T_EXTENDS 23405) (interface_extends_list 32675)) 15447 ((backup_doc_comment 32676)) 15448 (("{" 32677)) 15449 ((class_statement_list 32678)) 15450 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 32679)) 15451 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 32681) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 15452 (("%" 304) (")" 32682) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 15453 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23415) (";" 32419) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 32420) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 32421) (T_DEC 22) (T_DECLARE 32422) (T_DIR 24) (T_DNUMBER 25) (T_DO 32423) (T_DOUBLE_CAST 27) (T_ECHO 32424) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 32425) (T_FOREACH 32426) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 32427) (T_GOTO 32428) (T_IF 32379) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 32429) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 32430) (T_START_HEREDOC 69) (T_STATIC 32431) (T_STRING 32432) (T_STRING_CAST 72) (T_SWITCH 32433) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 32434) (T_UNSET 32435) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 32436) (T_WHILE 32437) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 32438) (alt_if_stmt_without_else 32439) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 32440) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 32441) (if_stmt_without_else 32442) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 32683) (static_member 123) (variable 126) (variable_class_name 127) ("{" 32444) ("~" 129)) 15454 ((")" 32685)) 15455 ((";" 32686)) 15456 ((";" 32688)) 15457 (("," 19446) (";" 32690)) 15458 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 32692) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 15459 (("%" 19534) ("*" 19535) ("+" 19536) ("-" 19537) ("." 19538) ("/" 19539) ("<" 19540) (">" 19541) ("?" 19542) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19543) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19544) (T_AS 32693) (T_BOOLEAN_AND 19545) (T_BOOLEAN_OR 19546) (T_COALESCE 19547) (T_INSTANCEOF 19548) (T_IS_EQUAL 19549) (T_IS_GREATER_OR_EQUAL 19550) (T_IS_IDENTICAL 19551) (T_IS_NOT_EQUAL 19552) (T_IS_NOT_IDENTICAL 19553) (T_IS_SMALLER_OR_EQUAL 19554) (T_LOGICAL_AND 19555) (T_LOGICAL_OR 19556) (T_LOGICAL_XOR 19557) (T_PIPE 19558) (T_POW 19559) (T_SL 19560) (T_SPACESHIP 19561) (T_SR 19562) ("^" 19563) ("|" 19564)) 15460 (("\"" 1538) ("$" 20897) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20898) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20899) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20900) ("[" 20901) (ampersand 20902) (array_object_dereferenceable 20903) (callable_expr 20904) (callable_variable 20905) (class_constant 1564) (class_name 20906) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 32694) (fully_dereferenceable 107) (function_call 20908) (name 20909) (new_dereferenceable 1570) (simple_variable 20910) (static_member 20911) (variable 20912) (variable_class_name 20913)) 15461 ((")" 32695) (T_DOUBLE_ARROW 32696)) 15462 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 32699) (";" 32365) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 32366) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 32369) (T_DEC 22) (T_DECLARE 32370) (T_DIR 24) (T_DNUMBER 25) (T_DO 32371) (T_DOUBLE_CAST 27) (T_ECHO 32372) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 32374) (T_FOREACH 32375) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 32376) (T_GOTO 32377) (T_IF 32379) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 32380) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 32383) (T_START_HEREDOC 69) (T_STATIC 32384) (T_STRING 32385) (T_STRING_CAST 72) (T_SWITCH 32386) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 32388) (T_UNSET 32389) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 32391) (T_WHILE 32392) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 32393) (alt_if_stmt_without_else 32394) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 32401) (fn 106) (foreach_statement 32705) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 32404) (if_stmt_without_else 32405) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 32701) (static_member 123) (variable 126) (variable_class_name 127) ("{" 32410) ("~" 129)) 15463 (("\"" 1538) ("$" 20928) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20929) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20930) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20931) ("[" 20932) (ampersand 20933) (array_object_dereferenceable 20934) (callable_expr 20935) (callable_variable 20936) (class_constant 1564) (class_name 20937) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 32697) (fully_dereferenceable 107) (function_call 20939) (name 20940) (new_dereferenceable 1570) (simple_variable 20941) (static_member 20942) (variable 20943) (variable_class_name 20944)) 15464 ((")" 32698)) 15465 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 32699) (";" 32365) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 32366) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 32369) (T_DEC 22) (T_DECLARE 32370) (T_DIR 24) (T_DNUMBER 25) (T_DO 32371) (T_DOUBLE_CAST 27) (T_ECHO 32372) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 32374) (T_FOREACH 32375) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 32376) (T_GOTO 32377) (T_IF 32379) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 32380) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 32383) (T_START_HEREDOC 69) (T_STATIC 32384) (T_STRING 32385) (T_STRING_CAST 72) (T_SWITCH 32386) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 32388) (T_UNSET 32389) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 32391) (T_WHILE 32392) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 32393) (alt_if_stmt_without_else 32394) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 32401) (fn 106) (foreach_statement 32700) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 32404) (if_stmt_without_else 32405) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 32701) (static_member 123) (variable 126) (variable_class_name 127) ("{" 32410) ("~" 129)) 15466 ((inner_statement_list 32702)) 15467 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20962) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 20963) (T_CLASS 20964) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20965) (T_DEC 22) (T_DECLARE 20966) (T_DIR 24) (T_DNUMBER 25) (T_DO 20967) (T_DOUBLE_CAST 27) (T_ECHO 20968) (T_EMPTY 29) (T_ENDFOREACH 32703) (T_ENUM 20970) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20971) (T_FOREACH 20972) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20973) (T_GOTO 20974) (T_HALT_COMPILER 20975) (T_IF 20976) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20977) (T_INTERFACE 20978) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20979) (T_START_HEREDOC 69) (T_STATIC 20980) (T_STRING 20981) (T_STRING_CAST 72) (T_SWITCH 20982) (T_THROW 74) (T_TRAIT 20983) (T_TRAIT_C 76) (T_TRY 20984) (T_UNSET 20985) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 20986) (T_WHILE 20987) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 20988) (alt_if_stmt_without_else 20989) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 20990) (attributes 20991) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 20992) (class_modifier 99) (class_modifiers 20993) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 20994) (expr 20995) (fn 106) (fully_dereferenceable 107) #'20996 (function_call 109) (function_declaration_statement 20997) (if_stmt 20998) (if_stmt_without_else 20999) (inline_function 113) (inner_statement 21000) (interface_declaration_statement 21001) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 21002) (static_member 123) (trait_declaration_statement 21003) (variable 126) (variable_class_name 127) ("{" 21004) ("~" 129)) 15468 ((";" 32704)) 15469 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21126) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21127) (fn 18437) (for_exprs 32707) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21129) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 15470 ((";" 32708)) 15471 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21135) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21136) (fn 18437) (for_cond_exprs 32709) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21138) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 15472 ((";" 32710)) 15473 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_VOID_CAST 22737) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 22738) (fn 2425) (for_exprs 32711) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_for_exprs 22740) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 15474 ((")" 32712)) 15475 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 32713) (";" 32365) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 32366) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 32369) (T_DEC 22) (T_DECLARE 32370) (T_DIR 24) (T_DNUMBER 25) (T_DO 32371) (T_DOUBLE_CAST 27) (T_ECHO 32372) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 32374) (T_FOREACH 32375) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 32376) (T_GOTO 32377) (T_IF 32379) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 32380) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 32383) (T_START_HEREDOC 69) (T_STATIC 32384) (T_STRING 32385) (T_STRING_CAST 72) (T_SWITCH 32386) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 32388) (T_UNSET 32389) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 32391) (T_WHILE 32392) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 32393) (alt_if_stmt_without_else 32394) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 32401) (fn 106) (for_statement 32714) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 32404) (if_stmt_without_else 32405) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 32715) (static_member 123) (variable 126) (variable_class_name 127) ("{" 32410) ("~" 129)) 15476 ((inner_statement_list 32716)) 15477 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22750) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 22751) (T_CLASS 22752) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22753) (T_DEC 22) (T_DECLARE 22754) (T_DIR 24) (T_DNUMBER 25) (T_DO 22755) (T_DOUBLE_CAST 27) (T_ECHO 22756) (T_EMPTY 29) (T_ENDFOR 32717) (T_ENUM 22758) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 22759) (T_FOREACH 22760) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 22761) (T_GOTO 22762) (T_HALT_COMPILER 22763) (T_IF 22764) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22765) (T_INTERFACE 22766) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22767) (T_START_HEREDOC 69) (T_STATIC 22768) (T_STRING 22769) (T_STRING_CAST 72) (T_SWITCH 22770) (T_THROW 74) (T_TRAIT 22771) (T_TRAIT_C 76) (T_TRY 22772) (T_UNSET 22773) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22774) (T_WHILE 22775) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22776) (alt_if_stmt_without_else 22777) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 22778) (attributes 22779) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 22780) (class_modifier 99) (class_modifiers 22781) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 22782) (expr 22783) (fn 106) (fully_dereferenceable 107) #'22784 (function_call 109) (function_declaration_statement 22785) (if_stmt 22786) (if_stmt_without_else 22787) (inline_function 113) (inner_statement 22788) (interface_declaration_statement 22789) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 22790) (static_member 123) (trait_declaration_statement 22791) (variable 126) (variable_class_name 127) ("{" 22792) ("~" 129)) 15478 ((";" 32718)) 15479 ((":" 23454) (enum_backing_type 32720)) 15480 ((T_IMPLEMENTS 2989) (implements_list 32721)) 15481 ((backup_doc_comment 32722)) 15482 (("{" 32723)) 15483 ((class_statement_list 32724)) 15484 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 32725)) 15485 (("," 22929) (";" 32727)) 15486 ((T_WHILE 32729)) 15487 (("(" 32730)) 15488 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 32731) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 15489 (("%" 304) (")" 32732) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 15490 ((";" 32733)) 15491 ((T_STRING 23106) (const_decl 23107) (const_list 32735)) 15492 ((")" 32736) ("," 23110)) 15493 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 32737) (";" 32365) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 32366) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 32369) (T_DEC 22) (T_DECLARE 32370) (T_DIR 24) (T_DNUMBER 25) (T_DO 32371) (T_DOUBLE_CAST 27) (T_ECHO 32372) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 32374) (T_FOREACH 32375) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 32376) (T_GOTO 32377) (T_IF 32379) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 32380) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 32383) (T_START_HEREDOC 69) (T_STATIC 32384) (T_STRING 32385) (T_STRING_CAST 72) (T_SWITCH 32386) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 32388) (T_UNSET 32389) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 32391) (T_WHILE 32392) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 32393) (alt_if_stmt_without_else 32394) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (declare_statement 32738) (dereferenceable_scalar 103) (expr 32401) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 32404) (if_stmt_without_else 32405) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 32739) (static_member 123) (variable 126) (variable_class_name 127) ("{" 32410) ("~" 129)) 15494 ((inner_statement_list 32740)) 15495 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 23116) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 23117) (T_CLASS 23118) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 23119) (T_DEC 22) (T_DECLARE 23120) (T_DIR 24) (T_DNUMBER 25) (T_DO 23121) (T_DOUBLE_CAST 27) (T_ECHO 23122) (T_EMPTY 29) (T_ENDDECLARE 32741) (T_ENUM 23124) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 23125) (T_FOREACH 23126) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 23127) (T_GOTO 23128) (T_HALT_COMPILER 23129) (T_IF 23130) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 23131) (T_INTERFACE 23132) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 23133) (T_START_HEREDOC 69) (T_STATIC 23134) (T_STRING 23135) (T_STRING_CAST 72) (T_SWITCH 23136) (T_THROW 74) (T_TRAIT 23137) (T_TRAIT_C 76) (T_TRY 23138) (T_UNSET 23139) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 23140) (T_WHILE 23141) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 23142) (alt_if_stmt_without_else 23143) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 23144) (attributes 23145) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 23146) (class_modifier 99) (class_modifiers 23147) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 23148) (expr 23149) (fn 106) (fully_dereferenceable 107) #'23150 (function_call 109) (function_declaration_statement 23151) (if_stmt 23152) (if_stmt_without_else 23153) (inline_function 113) (inner_statement 23154) (interface_declaration_statement 23155) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23156) (static_member 123) (trait_declaration_statement 23157) (variable 126) (variable_class_name 127) ("{" 23158) ("~" 129)) 15496 ((";" 32742)) 15497 ((";" 32744)) 15498 (("," 31541) (";" 32746)) 15499 ((T_EXTENDS 2987) (extends_from 32748)) 15500 ((T_IMPLEMENTS 2989) (implements_list 32749)) 15501 ((backup_doc_comment 32750)) 15502 (("{" 32751)) 15503 ((class_statement_list 32752)) 15504 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 32753)) 15505 ((";" 32755)) 15506 ((top_statement_list 32758)) 15507 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 32365) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 13) (T_BOOL_CAST 14) (T_BREAK 32366) (T_CLASS 32367) (T_CLASS_C 17) (T_CLONE 18) (T_CONST 32368) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 32369) (T_DEC 22) (T_DECLARE 32370) (T_DIR 24) (T_DNUMBER 25) (T_DO 32371) (T_DOUBLE_CAST 27) (T_ECHO 32372) (T_EMPTY 29) (T_ENUM 32373) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 32374) (T_FOREACH 32375) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 32376) (T_GOTO 32377) (T_HALT_COMPILER 32378) (T_IF 32379) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 32380) (T_INTERFACE 32381) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAMESPACE 32382) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 32383) (T_START_HEREDOC 69) (T_STATIC 32384) (T_STRING 32385) (T_STRING_CAST 72) (T_SWITCH 32386) (T_THROW 74) (T_TRAIT 32387) (T_TRAIT_C 76) (T_TRY 32388) (T_UNSET 32389) (T_UNSET_CAST 79) (T_USE 32390) (T_VARIABLE 81) (T_VOID_CAST 32391) (T_WHILE 32392) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 32393) (alt_if_stmt_without_else 32394) (array_object_dereferenceable 90) (attribute 91) (attributed_statement 32395) (attributed_top_statement 32396) (attributes 32397) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 32398) (class_modifier 99) (class_modifiers 32399) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 32400) (expr 32401) (fn 106) (fully_dereferenceable 107) #'32402 (function_call 109) (function_declaration_statement 32403) (if_stmt 32404) (if_stmt_without_else 32405) (inline_function 113) (interface_declaration_statement 32406) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 32407) (static_member 123) (top_statement 32408) (trait_declaration_statement 32409) (variable 126) (variable_class_name 127) ("{" 32410) ("}" 32759) ("~" 129)) 15508 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 32761) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 15509 (("%" 304) (")" 32762) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 15510 (("{" 32763)) 15511 (("!" 3960) ("\"" 3961) ("$" 3962) ("(" 3963) ("+" 3964) ("-" 3965) ("@" 3966) (T_ARRAY 3967) (T_ARRAY_CAST 3968) (T_ATTRIBUTE 130) (T_BOOL_CAST 3969) (T_CLASS_C 3970) (T_CLONE 3971) (T_CONSTANT_ENCAPSED_STRING 3972) (T_DEC 3973) (T_DEFAULT 3974) (T_DIR 3975) (T_DNUMBER 3976) (T_DOUBLE_CAST 3977) (T_EMPTY 3978) (T_EVAL 3979) (T_EXIT 3980) (T_FILE 3981) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 3982) (T_INC 3983) (T_INCLUDE 3984) (T_INCLUDE_ONCE 3985) (T_INT_CAST 3986) (T_ISSET 3987) (T_LINE 3988) (T_LIST 3989) (T_LNUMBER 3990) (T_MATCH 3991) (T_METHOD_C 3992) (T_NAME_FULLY_QUALIFIED 3993) (T_NAME_QUALIFIED 3994) (T_NAME_RELATIVE 3995) (T_NEW 3996) (T_NS_C 3997) (T_OBJECT_CAST 3998) (T_PRINT 3999) (T_PROPERTY_C 4000) (T_READONLY 4001) (T_REQUIRE 4002) (T_REQUIRE_ONCE 4003) (T_START_HEREDOC 4004) (T_STATIC 4005) (T_STRING 4006) (T_STRING_CAST 4007) (T_THROW 4008) (T_TRAIT_C 4009) (T_UNSET_CAST 4010) (T_VARIABLE 4011) (T_YIELD 4012) (T_YIELD_FROM 4013) ("[" 4014) ("`" 4015) (array_object_dereferenceable 4016) (attribute 135) (attributes 4017) (callable_expr 4018) (callable_variable 4019) (class_constant 4020) (class_name 4021) (constant 4022) (dereferenceable_scalar 4023) (expr 4024) (fn 4025) (fully_dereferenceable 107) #'4026 (function_call 4027) (inline_function 4028) (internal_functions_in_yacc 4029) (match 4030) (match_arm 4031) (match_arm_cond_list 4032) (match_arm_list 32764) (name 4034) (new_dereferenceable 4035) (new_non_dereferenceable 4036) (non_empty_match_arm_list 4037) (scalar 4038) (simple_variable 4039) (static_member 4040) (variable 4041) (variable_class_name 4042) ("~" 4043)) 15512 (("}" 32765)) 15513 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 32767) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 15514 ((")" 32768)) 15515 (("=" 32769)) 15516 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 32770) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 15517 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 5223) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (isset_variable 5224) (isset_variables 32772) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 15518 (("," 5226) (possible_comma 32773)) 15519 ((")" 32774)) 15520 ((T_EXTENDS 23405) (interface_extends_list 32777)) 15521 ((backup_doc_comment 32778)) 15522 (("{" 32779)) 15523 ((class_statement_list 32780)) 15524 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 32781)) 15525 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 32786) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 15526 (("%" 304) (")" 32787) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 15527 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23415) (";" 32788) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 32789) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 32790) (T_DEC 22) (T_DECLARE 32791) (T_DIR 24) (T_DNUMBER 25) (T_DO 32792) (T_DOUBLE_CAST 27) (T_ECHO 32793) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 32794) (T_FOREACH 32795) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 32796) (T_GOTO 32797) (T_IF 32798) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 32799) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 32800) (T_START_HEREDOC 69) (T_STATIC 32801) (T_STRING 32802) (T_STRING_CAST 72) (T_SWITCH 32803) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 32804) (T_UNSET 32805) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 32806) (T_WHILE 32807) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 32808) (alt_if_stmt_without_else 32809) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 32810) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 32811) (if_stmt_without_else 32812) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 32813) (static_member 123) (variable 126) (variable_class_name 127) ("{" 32814) ("~" 129)) 15528 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 32943) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 15529 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 32941) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 15530 (("(" 32932)) 15531 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22932) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 22933) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22934) (T_DEC 22) (T_DECLARE 22935) (T_DIR 24) (T_DNUMBER 25) (T_DO 22936) (T_DOUBLE_CAST 27) (T_ECHO 22937) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 22938) (T_FOREACH 22939) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 22940) (T_GOTO 22941) (T_IF 22942) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22943) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22944) (T_START_HEREDOC 69) (T_STATIC 22945) (T_STRING 22946) (T_STRING_CAST 72) (T_SWITCH 22947) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 22948) (T_UNSET 22949) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22950) (T_WHILE 22951) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22952) (alt_if_stmt_without_else 22953) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 22954) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 22955) (if_stmt_without_else 22956) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 32926) (static_member 123) (variable 126) (variable_class_name 127) ("{" 22958) ("~" 129)) 15532 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (echo_expr 22926) (echo_expr_list 32924) (expr 22928) (fn 18437) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 15533 (("(" 32911)) 15534 (("(" 32896)) 15535 (("$" 19441) (T_VARIABLE 19442) (global_var 19443) (global_var_list 32894) (simple_variable 19445)) 15536 ((T_STRING 32892)) 15537 (("(" 32888)) 15538 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 19436) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (optional_expr 32886) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 15539 ((T_FN 35) (T_FUNCTION 131) (T_VARIABLE 18365) (fn 106) #'138 (inline_function 18366) (static_var 18367) (static_var_list 32884)) 15540 ((":" 32883)) 15541 (("(" 32865)) 15542 (("{" 32848)) 15543 (("(" 32843)) 15544 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_DEC 22) (T_DIR 24) (T_DNUMBER 25) (T_DOUBLE_CAST 27) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 39) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_START_HEREDOC 69) (T_STATIC 133) (T_STRING 134) (T_STRING_CAST 72) (T_THROW 74) (T_TRAIT_C 76) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 32841) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (static_member 123) (variable 126) (variable_class_name 127) ("~" 129)) 15545 (("(" 32832)) 15546 ((T_ELSE 32825) (T_ELSEIF 17711) (T_ENDIF 32826)) 15547 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 32824) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 15548 ((T_ELSE 32817) (T_ELSEIF 32818)) 15549 ((inner_statement_list 32815)) 15550 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 32816) ("~" 129)) 15551 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 32788) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 32789) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 32790) (T_DEC 22) (T_DECLARE 32791) (T_DIR 24) (T_DNUMBER 25) (T_DO 32792) (T_DOUBLE_CAST 27) (T_ECHO 32793) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 32794) (T_FOREACH 32795) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 32796) (T_GOTO 32797) (T_IF 32798) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 32799) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 32800) (T_START_HEREDOC 69) (T_STATIC 32801) (T_STRING 32802) (T_STRING_CAST 72) (T_SWITCH 32803) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 32804) (T_UNSET 32805) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 32806) (T_WHILE 32807) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 32808) (alt_if_stmt_without_else 32809) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 32810) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 32811) (if_stmt_without_else 32812) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 32823) (static_member 123) (variable 126) (variable_class_name 127) ("{" 32814) ("~" 129)) 15552 (("(" 32819)) 15553 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 32820) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 15554 (("%" 304) (")" 32821) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 15555 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 32788) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 32789) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 32790) (T_DEC 22) (T_DECLARE 32791) (T_DIR 24) (T_DNUMBER 25) (T_DO 32792) (T_DOUBLE_CAST 27) (T_ECHO 32793) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 32794) (T_FOREACH 32795) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 32796) (T_GOTO 32797) (T_IF 32798) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 32799) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 32800) (T_START_HEREDOC 69) (T_STATIC 32801) (T_STRING 32802) (T_STRING_CAST 72) (T_SWITCH 32803) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 32804) (T_UNSET 32805) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 32806) (T_WHILE 32807) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 32808) (alt_if_stmt_without_else 32809) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 32810) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 32811) (if_stmt_without_else 32812) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 32822) (static_member 123) (variable 126) (variable_class_name 127) ("{" 32814) ("~" 129)) 15556 ((":" 32828)) 15557 ((";" 32827)) 15558 ((inner_statement_list 32829)) 15559 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17807) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17808) (T_CLASS 17809) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17810) (T_DEC 22) (T_DECLARE 17811) (T_DIR 24) (T_DNUMBER 25) (T_DO 17812) (T_DOUBLE_CAST 27) (T_ECHO 17813) (T_EMPTY 29) (T_ENDIF 32830) (T_ENUM 17815) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17816) (T_FOREACH 17817) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17818) (T_GOTO 17819) (T_HALT_COMPILER 17820) (T_IF 17732) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17821) (T_INTERFACE 17822) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17823) (T_START_HEREDOC 69) (T_STATIC 17824) (T_STRING 17825) (T_STRING_CAST 72) (T_SWITCH 17826) (T_THROW 74) (T_TRAIT 17827) (T_TRAIT_C 76) (T_TRY 17828) (T_UNSET 17829) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17830) (T_WHILE 17831) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17832) (alt_if_stmt_without_else 17833) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17834) (attributes 17835) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17836) (class_modifier 99) (class_modifiers 17837) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17838) (expr 17839) (fn 106) (fully_dereferenceable 107) #'17840 (function_call 109) (function_declaration_statement 17841) (if_stmt 17842) (if_stmt_without_else 17843) (inline_function 113) (inner_statement 17844) (interface_declaration_statement 17845) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17846) (static_member 123) (trait_declaration_statement 17847) (variable 126) (variable_class_name 127) ("{" 17848) ("~" 129)) 15560 ((";" 32831)) 15561 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 32833) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 15562 (("%" 304) (")" 32834) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 15563 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 32835) (";" 32788) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 32789) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 32790) (T_DEC 22) (T_DECLARE 32791) (T_DIR 24) (T_DNUMBER 25) (T_DO 32792) (T_DOUBLE_CAST 27) (T_ECHO 32793) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 32794) (T_FOREACH 32795) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 32796) (T_GOTO 32797) (T_IF 32798) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 32799) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 32800) (T_START_HEREDOC 69) (T_STATIC 32801) (T_STRING 32802) (T_STRING_CAST 72) (T_SWITCH 32803) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 32804) (T_UNSET 32805) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 32806) (T_WHILE 32807) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 32808) (alt_if_stmt_without_else 32809) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 32810) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 32811) (if_stmt_without_else 32812) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 32836) (static_member 123) (variable 126) (variable_class_name 127) (while_statement 32837) ("{" 32814) ("~" 129)) 15564 ((inner_statement_list 32838)) 15565 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 17889) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 17890) (T_CLASS 17891) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 17892) (T_DEC 22) (T_DECLARE 17893) (T_DIR 24) (T_DNUMBER 25) (T_DO 17894) (T_DOUBLE_CAST 27) (T_ECHO 17895) (T_EMPTY 29) (T_ENDWHILE 32839) (T_ENUM 17897) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 17898) (T_FOREACH 17899) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 17900) (T_GOTO 17901) (T_HALT_COMPILER 17902) (T_IF 17903) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 17904) (T_INTERFACE 17905) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 17906) (T_START_HEREDOC 69) (T_STATIC 17907) (T_STRING 17908) (T_STRING_CAST 72) (T_SWITCH 17909) (T_THROW 74) (T_TRAIT 17910) (T_TRAIT_C 76) (T_TRY 17911) (T_UNSET 17912) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 17913) (T_WHILE 17914) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 17915) (alt_if_stmt_without_else 17916) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 17917) (attributes 17918) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 17919) (class_modifier 99) (class_modifiers 17920) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 17921) (expr 17922) (fn 106) (fully_dereferenceable 107) #'17923 (function_call 109) (function_declaration_statement 17924) (if_stmt 17925) (if_stmt_without_else 17926) (inline_function 113) (inner_statement 17927) (interface_declaration_statement 17928) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 17929) (static_member 123) (trait_declaration_statement 17930) (variable 126) (variable_class_name 127) ("{" 17931) ("~" 129)) 15566 ((";" 32840)) 15567 (("%" 5890) ("*" 5891) ("+" 5892) ("-" 5893) ("." 5894) ("/" 5895) (";" 32842) ("<" 5897) (">" 5898) ("?" 5899) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 5900) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 5901) (T_BOOLEAN_AND 5902) (T_BOOLEAN_OR 5903) (T_COALESCE 5904) (T_INSTANCEOF 5905) (T_IS_EQUAL 5906) (T_IS_GREATER_OR_EQUAL 5907) (T_IS_IDENTICAL 5908) (T_IS_NOT_EQUAL 5909) (T_IS_NOT_IDENTICAL 5910) (T_IS_SMALLER_OR_EQUAL 5911) (T_LOGICAL_AND 5912) (T_LOGICAL_OR 5913) (T_LOGICAL_XOR 5914) (T_PIPE 5915) (T_POW 5916) (T_SL 5917) (T_SPACESHIP 5918) (T_SR 5919) ("^" 5920) ("|" 5921)) 15568 (("\"" 1538) ("$" 17989) ("(" 1540) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 17990) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 17991) ("[" 1560) (array_object_dereferenceable 17992) (callable_expr 17993) (callable_variable 17994) (class_constant 1564) (class_name 17995) (constant 1566) (dereferenceable_scalar 1567) (fully_dereferenceable 107) (function_call 17996) (name 17997) (new_dereferenceable 1570) (simple_variable 17998) (static_member 17999) (unset_variable 18000) (unset_variables 32844) (variable 18002) (variable_class_name 18003)) 15569 (("," 18015) (possible_comma 32845)) 15570 ((")" 32846)) 15571 ((";" 32847)) 15572 ((inner_statement_list 32849)) 15573 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 32850) ("~" 129)) 15574 ((catch_list 32851)) 15575 ((T_CATCH 32852) (T_FINALLY 32853) (finally_statement 32854)) 15576 (("(" 32858)) 15577 (("{" 32855)) 15578 ((inner_statement_list 32856)) 15579 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 32857) ("~" 129)) 15580 ((T_NAME_FULLY_QUALIFIED 18056) (T_NAME_QUALIFIED 18057) (T_NAME_RELATIVE 18058) (T_STATIC 18059) (T_STRING 18060) (catch_name_list 32859) (class_name 18062) (name 18063)) 15581 ((T_VARIABLE 18064) (optional_variable 32860) ("|" 18066)) 15582 ((")" 32861)) 15583 (("{" 32862)) 15584 ((inner_statement_list 32863)) 15585 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 173) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 175) (T_CLASS 176) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 177) (T_DEC 22) (T_DECLARE 178) (T_DIR 24) (T_DNUMBER 25) (T_DO 179) (T_DOUBLE_CAST 27) (T_ECHO 180) (T_EMPTY 29) (T_ENUM 181) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 182) (T_FOREACH 183) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 184) (T_GOTO 185) (T_HALT_COMPILER 186) (T_IF 187) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 188) (T_INTERFACE 189) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 190) (T_START_HEREDOC 69) (T_STATIC 191) (T_STRING 192) (T_STRING_CAST 72) (T_SWITCH 193) (T_THROW 74) (T_TRAIT 194) (T_TRAIT_C 76) (T_TRY 195) (T_UNSET 196) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 197) (T_WHILE 198) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 199) (alt_if_stmt_without_else 200) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 202) (attributes 203) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 204) (class_modifier 99) (class_modifiers 205) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 206) (expr 207) (fn 106) (fully_dereferenceable 107) #'208 (function_call 109) (function_declaration_statement 209) (if_stmt 210) (if_stmt_without_else 211) (inline_function 113) (inner_statement 212) (interface_declaration_statement 213) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 214) (static_member 123) (trait_declaration_statement 215) (variable 126) (variable_class_name 127) ("{" 216) ("}" 32864) ("~" 129)) 15586 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 32866) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 15587 (("%" 304) (")" 32867) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 15588 ((":" 32868) (switch_case_list 32869) ("{" 32870)) 15589 ((";" 32876) (case_list 32877)) 15590 ((";" 32871) (case_list 32872)) 15591 ((case_list 32874)) 15592 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 32873)) 15593 ((T_CASE 18080) (T_DEFAULT 18081) ("}" 32875)) 15594 ((case_list 32880)) 15595 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 32878)) 15596 ((";" 32879)) 15597 ((T_CASE 18219) (T_DEFAULT 18220) (T_ENDSWITCH 32881)) 15598 ((";" 32882)) 15599 (("," 18369) (";" 32885)) 15600 ((";" 32887)) 15601 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 32889) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 15602 (("%" 304) (")" 32890) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 15603 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 23415) (";" 30234) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 30235) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 30236) (T_DEC 22) (T_DECLARE 30237) (T_DIR 24) (T_DNUMBER 25) (T_DO 30238) (T_DOUBLE_CAST 27) (T_ECHO 30239) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 30240) (T_FOREACH 30241) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 30242) (T_GOTO 30243) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 30244) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 30245) (T_START_HEREDOC 69) (T_STATIC 30246) (T_STRING 30247) (T_STRING_CAST 72) (T_SWITCH 30248) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 30249) (T_UNSET 30250) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 30251) (T_WHILE 30252) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 30253) (alt_if_stmt_without_else 30254) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 30255) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 30256) (if_stmt_without_else 30257) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 32891) (static_member 123) (variable 126) (variable_class_name 127) ("{" 30259) ("~" 129)) 15604 ((";" 32893)) 15605 (("," 19446) (";" 32895)) 15606 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 32897) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 15607 (("%" 19534) ("*" 19535) ("+" 19536) ("-" 19537) ("." 19538) ("/" 19539) ("<" 19540) (">" 19541) ("?" 19542) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19543) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19544) (T_AS 32898) (T_BOOLEAN_AND 19545) (T_BOOLEAN_OR 19546) (T_COALESCE 19547) (T_INSTANCEOF 19548) (T_IS_EQUAL 19549) (T_IS_GREATER_OR_EQUAL 19550) (T_IS_IDENTICAL 19551) (T_IS_NOT_EQUAL 19552) (T_IS_NOT_IDENTICAL 19553) (T_IS_SMALLER_OR_EQUAL 19554) (T_LOGICAL_AND 19555) (T_LOGICAL_OR 19556) (T_LOGICAL_XOR 19557) (T_PIPE 19558) (T_POW 19559) (T_SL 19560) (T_SPACESHIP 19561) (T_SR 19562) ("^" 19563) ("|" 19564)) 15608 (("\"" 1538) ("$" 20897) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20898) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20899) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20900) ("[" 20901) (ampersand 20902) (array_object_dereferenceable 20903) (callable_expr 20904) (callable_variable 20905) (class_constant 1564) (class_name 20906) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 32899) (fully_dereferenceable 107) (function_call 20908) (name 20909) (new_dereferenceable 1570) (simple_variable 20910) (static_member 20911) (variable 20912) (variable_class_name 20913)) 15609 ((")" 32900) (T_DOUBLE_ARROW 32901)) 15610 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 32904) (";" 32788) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 32789) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 32790) (T_DEC 22) (T_DECLARE 32791) (T_DIR 24) (T_DNUMBER 25) (T_DO 32792) (T_DOUBLE_CAST 27) (T_ECHO 32793) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 32794) (T_FOREACH 32795) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 32796) (T_GOTO 32797) (T_IF 32798) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 32799) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 32800) (T_START_HEREDOC 69) (T_STATIC 32801) (T_STRING 32802) (T_STRING_CAST 72) (T_SWITCH 32803) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 32804) (T_UNSET 32805) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 32806) (T_WHILE 32807) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 32808) (alt_if_stmt_without_else 32809) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 32810) (fn 106) (foreach_statement 32910) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 32811) (if_stmt_without_else 32812) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 32906) (static_member 123) (variable 126) (variable_class_name 127) ("{" 32814) ("~" 129)) 15611 (("\"" 1538) ("$" 20928) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20929) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20930) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20931) ("[" 20932) (ampersand 20933) (array_object_dereferenceable 20934) (callable_expr 20935) (callable_variable 20936) (class_constant 1564) (class_name 20937) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 32902) (fully_dereferenceable 107) (function_call 20939) (name 20940) (new_dereferenceable 1570) (simple_variable 20941) (static_member 20942) (variable 20943) (variable_class_name 20944)) 15612 ((")" 32903)) 15613 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 32904) (";" 32788) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 32789) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 32790) (T_DEC 22) (T_DECLARE 32791) (T_DIR 24) (T_DNUMBER 25) (T_DO 32792) (T_DOUBLE_CAST 27) (T_ECHO 32793) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 32794) (T_FOREACH 32795) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 32796) (T_GOTO 32797) (T_IF 32798) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 32799) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 32800) (T_START_HEREDOC 69) (T_STATIC 32801) (T_STRING 32802) (T_STRING_CAST 72) (T_SWITCH 32803) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 32804) (T_UNSET 32805) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 32806) (T_WHILE 32807) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 32808) (alt_if_stmt_without_else 32809) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 32810) (fn 106) (foreach_statement 32905) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 32811) (if_stmt_without_else 32812) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 32906) (static_member 123) (variable 126) (variable_class_name 127) ("{" 32814) ("~" 129)) 15614 ((inner_statement_list 32907)) 15615 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20962) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 20963) (T_CLASS 20964) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20965) (T_DEC 22) (T_DECLARE 20966) (T_DIR 24) (T_DNUMBER 25) (T_DO 20967) (T_DOUBLE_CAST 27) (T_ECHO 20968) (T_EMPTY 29) (T_ENDFOREACH 32908) (T_ENUM 20970) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20971) (T_FOREACH 20972) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20973) (T_GOTO 20974) (T_HALT_COMPILER 20975) (T_IF 20976) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20977) (T_INTERFACE 20978) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20979) (T_START_HEREDOC 69) (T_STATIC 20980) (T_STRING 20981) (T_STRING_CAST 72) (T_SWITCH 20982) (T_THROW 74) (T_TRAIT 20983) (T_TRAIT_C 76) (T_TRY 20984) (T_UNSET 20985) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 20986) (T_WHILE 20987) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 20988) (alt_if_stmt_without_else 20989) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 20990) (attributes 20991) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 20992) (class_modifier 99) (class_modifiers 20993) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 20994) (expr 20995) (fn 106) (fully_dereferenceable 107) #'20996 (function_call 109) (function_declaration_statement 20997) (if_stmt 20998) (if_stmt_without_else 20999) (inline_function 113) (inner_statement 21000) (interface_declaration_statement 21001) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 21002) (static_member 123) (trait_declaration_statement 21003) (variable 126) (variable_class_name 127) ("{" 21004) ("~" 129)) 15616 ((";" 32909)) 15617 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21126) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21127) (fn 18437) (for_exprs 32912) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21129) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 15618 ((";" 32913)) 15619 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21135) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21136) (fn 18437) (for_cond_exprs 32914) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21138) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 15620 ((";" 32915)) 15621 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_VOID_CAST 22737) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 22738) (fn 2425) (for_exprs 32916) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_for_exprs 22740) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 15622 ((")" 32917)) 15623 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 32918) (";" 32788) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 32789) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 32790) (T_DEC 22) (T_DECLARE 32791) (T_DIR 24) (T_DNUMBER 25) (T_DO 32792) (T_DOUBLE_CAST 27) (T_ECHO 32793) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 32794) (T_FOREACH 32795) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 32796) (T_GOTO 32797) (T_IF 32798) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 32799) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 32800) (T_START_HEREDOC 69) (T_STATIC 32801) (T_STRING 32802) (T_STRING_CAST 72) (T_SWITCH 32803) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 32804) (T_UNSET 32805) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 32806) (T_WHILE 32807) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 32808) (alt_if_stmt_without_else 32809) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 32810) (fn 106) (for_statement 32919) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 32811) (if_stmt_without_else 32812) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 32920) (static_member 123) (variable 126) (variable_class_name 127) ("{" 32814) ("~" 129)) 15624 ((inner_statement_list 32921)) 15625 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22750) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 22751) (T_CLASS 22752) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22753) (T_DEC 22) (T_DECLARE 22754) (T_DIR 24) (T_DNUMBER 25) (T_DO 22755) (T_DOUBLE_CAST 27) (T_ECHO 22756) (T_EMPTY 29) (T_ENDFOR 32922) (T_ENUM 22758) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 22759) (T_FOREACH 22760) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 22761) (T_GOTO 22762) (T_HALT_COMPILER 22763) (T_IF 22764) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22765) (T_INTERFACE 22766) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22767) (T_START_HEREDOC 69) (T_STATIC 22768) (T_STRING 22769) (T_STRING_CAST 72) (T_SWITCH 22770) (T_THROW 74) (T_TRAIT 22771) (T_TRAIT_C 76) (T_TRY 22772) (T_UNSET 22773) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22774) (T_WHILE 22775) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22776) (alt_if_stmt_without_else 22777) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 22778) (attributes 22779) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 22780) (class_modifier 99) (class_modifiers 22781) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 22782) (expr 22783) (fn 106) (fully_dereferenceable 107) #'22784 (function_call 109) (function_declaration_statement 22785) (if_stmt 22786) (if_stmt_without_else 22787) (inline_function 113) (inner_statement 22788) (interface_declaration_statement 22789) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 22790) (static_member 123) (trait_declaration_statement 22791) (variable 126) (variable_class_name 127) ("{" 22792) ("~" 129)) 15626 ((";" 32923)) 15627 (("," 22929) (";" 32925)) 15628 ((T_WHILE 32927)) 15629 (("(" 32928)) 15630 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 32929) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 15631 (("%" 304) (")" 32930) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 15632 ((";" 32931)) 15633 ((T_STRING 23106) (const_decl 23107) (const_list 32933)) 15634 ((")" 32934) ("," 23110)) 15635 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 32935) (";" 32788) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 32789) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 32790) (T_DEC 22) (T_DECLARE 32791) (T_DIR 24) (T_DNUMBER 25) (T_DO 32792) (T_DOUBLE_CAST 27) (T_ECHO 32793) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 32794) (T_FOREACH 32795) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 32796) (T_GOTO 32797) (T_IF 32798) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 32799) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 32800) (T_START_HEREDOC 69) (T_STATIC 32801) (T_STRING 32802) (T_STRING_CAST 72) (T_SWITCH 32803) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 32804) (T_UNSET 32805) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 32806) (T_WHILE 32807) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 32808) (alt_if_stmt_without_else 32809) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (declare_statement 32936) (dereferenceable_scalar 103) (expr 32810) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 32811) (if_stmt_without_else 32812) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 32937) (static_member 123) (variable 126) (variable_class_name 127) ("{" 32814) ("~" 129)) 15636 ((inner_statement_list 32938)) 15637 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 23116) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 23117) (T_CLASS 23118) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 23119) (T_DEC 22) (T_DECLARE 23120) (T_DIR 24) (T_DNUMBER 25) (T_DO 23121) (T_DOUBLE_CAST 27) (T_ECHO 23122) (T_EMPTY 29) (T_ENDDECLARE 32939) (T_ENUM 23124) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 23125) (T_FOREACH 23126) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 23127) (T_GOTO 23128) (T_HALT_COMPILER 23129) (T_IF 23130) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 23131) (T_INTERFACE 23132) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 23133) (T_START_HEREDOC 69) (T_STATIC 23134) (T_STRING 23135) (T_STRING_CAST 72) (T_SWITCH 23136) (T_THROW 74) (T_TRAIT 23137) (T_TRAIT_C 76) (T_TRY 23138) (T_UNSET 23139) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 23140) (T_WHILE 23141) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 23142) (alt_if_stmt_without_else 23143) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 23144) (attributes 23145) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 23146) (class_modifier 99) (class_modifiers 23147) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 23148) (expr 23149) (fn 106) (fully_dereferenceable 107) #'23150 (function_call 109) (function_declaration_statement 23151) (if_stmt 23152) (if_stmt_without_else 23153) (inline_function 113) (inner_statement 23154) (interface_declaration_statement 23155) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23156) (static_member 123) (trait_declaration_statement 23157) (variable 126) (variable_class_name 127) ("{" 23158) ("~" 129)) 15638 ((";" 32940)) 15639 ((";" 32942)) 15640 ((";" 32944)) 15641 ((")" 32946)) 15642 ((";" 32947)) 15643 ((";" 32949)) 15644 (("," 19446) (";" 32951)) 15645 (("!" 19454) ("\"" 19455) ("$" 19456) ("(" 19457) ("+" 19458) ("-" 19459) ("@" 19460) (T_ARRAY 19461) (T_ARRAY_CAST 19462) (T_ATTRIBUTE 130) (T_BOOL_CAST 19463) (T_CLASS_C 19464) (T_CLONE 19465) (T_CONSTANT_ENCAPSED_STRING 19466) (T_DEC 19467) (T_DIR 19468) (T_DNUMBER 19469) (T_DOUBLE_CAST 19470) (T_EMPTY 19471) (T_EVAL 19472) (T_EXIT 19473) (T_FILE 19474) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 19475) (T_INC 19476) (T_INCLUDE 19477) (T_INCLUDE_ONCE 19478) (T_INT_CAST 19479) (T_ISSET 19480) (T_LINE 19481) (T_LIST 19482) (T_LNUMBER 19483) (T_MATCH 19484) (T_METHOD_C 19485) (T_NAME_FULLY_QUALIFIED 19486) (T_NAME_QUALIFIED 19487) (T_NAME_RELATIVE 19488) (T_NEW 19489) (T_NS_C 19490) (T_OBJECT_CAST 19491) (T_PRINT 19492) (T_PROPERTY_C 19493) (T_READONLY 19494) (T_REQUIRE 19495) (T_REQUIRE_ONCE 19496) (T_START_HEREDOC 19497) (T_STATIC 19498) (T_STRING 19499) (T_STRING_CAST 19500) (T_THROW 19501) (T_TRAIT_C 19502) (T_UNSET_CAST 19503) (T_VARIABLE 19504) (T_YIELD 19505) (T_YIELD_FROM 19506) ("[" 19507) ("`" 19508) (array_object_dereferenceable 19509) (attribute 135) (attributes 19510) (callable_expr 19511) (callable_variable 19512) (class_constant 19513) (class_name 19514) (constant 19515) (dereferenceable_scalar 19516) (expr 32953) (fn 19518) (fully_dereferenceable 107) #'19519 (function_call 19520) (inline_function 19521) (internal_functions_in_yacc 19522) (match 19523) (name 19524) (new_dereferenceable 19525) (new_non_dereferenceable 19526) (scalar 19527) (simple_variable 19528) (static_member 19529) (variable 19530) (variable_class_name 19531) ("~" 19532)) 15646 (("%" 19534) ("*" 19535) ("+" 19536) ("-" 19537) ("." 19538) ("/" 19539) ("<" 19540) (">" 19541) ("?" 19542) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 19543) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 19544) (T_AS 32954) (T_BOOLEAN_AND 19545) (T_BOOLEAN_OR 19546) (T_COALESCE 19547) (T_INSTANCEOF 19548) (T_IS_EQUAL 19549) (T_IS_GREATER_OR_EQUAL 19550) (T_IS_IDENTICAL 19551) (T_IS_NOT_EQUAL 19552) (T_IS_NOT_IDENTICAL 19553) (T_IS_SMALLER_OR_EQUAL 19554) (T_LOGICAL_AND 19555) (T_LOGICAL_OR 19556) (T_LOGICAL_XOR 19557) (T_PIPE 19558) (T_POW 19559) (T_SL 19560) (T_SPACESHIP 19561) (T_SR 19562) ("^" 19563) ("|" 19564)) 15647 (("\"" 1538) ("$" 20897) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20898) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20899) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20900) ("[" 20901) (ampersand 20902) (array_object_dereferenceable 20903) (callable_expr 20904) (callable_variable 20905) (class_constant 1564) (class_name 20906) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 32955) (fully_dereferenceable 107) (function_call 20908) (name 20909) (new_dereferenceable 1570) (simple_variable 20910) (static_member 20911) (variable 20912) (variable_class_name 20913)) 15648 ((")" 32956) (T_DOUBLE_ARROW 32957)) 15649 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 32960) (";" 8) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 15) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21) (T_DEC 22) (T_DECLARE 23) (T_DIR 24) (T_DNUMBER 25) (T_DO 26) (T_DOUBLE_CAST 27) (T_ECHO 28) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 36) (T_FOREACH 37) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 40) (T_GOTO 41) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 47) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 68) (T_START_HEREDOC 69) (T_STATIC 70) (T_STRING 71) (T_STRING_CAST 72) (T_SWITCH 73) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 77) (T_UNSET 78) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 82) (T_WHILE 83) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 88) (alt_if_stmt_without_else 89) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 105) (fn 106) (foreach_statement 32966) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 111) (if_stmt_without_else 112) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 32962) (static_member 123) (variable 126) (variable_class_name 127) ("{" 128) ("~" 129)) 15650 (("\"" 1538) ("$" 20928) ("(" 1540) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1541) (T_CLASS_C 1542) (T_CONSTANT_ENCAPSED_STRING 1543) (T_DIR 1544) (T_FILE 1545) (T_FUNC_C 1546) (T_LINE 1547) (T_LIST 20929) (T_METHOD_C 1548) (T_NAME_FULLY_QUALIFIED 1549) (T_NAME_QUALIFIED 1550) (T_NAME_RELATIVE 1551) (T_NEW 1552) (T_NS_C 1553) (T_PROPERTY_C 1554) (T_READONLY 20930) (T_STATIC 1556) (T_STRING 1557) (T_TRAIT_C 1558) (T_VARIABLE 20931) ("[" 20932) (ampersand 20933) (array_object_dereferenceable 20934) (callable_expr 20935) (callable_variable 20936) (class_constant 1564) (class_name 20937) (constant 1566) (dereferenceable_scalar 1567) (foreach_variable 32958) (fully_dereferenceable 107) (function_call 20939) (name 20940) (new_dereferenceable 1570) (simple_variable 20941) (static_member 20942) (variable 20943) (variable_class_name 20944)) 15651 ((")" 32959)) 15652 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 32960) (";" 8) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 15) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21) (T_DEC 22) (T_DECLARE 23) (T_DIR 24) (T_DNUMBER 25) (T_DO 26) (T_DOUBLE_CAST 27) (T_ECHO 28) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 36) (T_FOREACH 37) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 40) (T_GOTO 41) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 47) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 68) (T_START_HEREDOC 69) (T_STATIC 70) (T_STRING 71) (T_STRING_CAST 72) (T_SWITCH 73) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 77) (T_UNSET 78) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 82) (T_WHILE 83) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 88) (alt_if_stmt_without_else 89) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 105) (fn 106) (foreach_statement 32961) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 111) (if_stmt_without_else 112) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 32962) (static_member 123) (variable 126) (variable_class_name 127) ("{" 128) ("~" 129)) 15653 ((inner_statement_list 32963)) 15654 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 20962) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 20963) (T_CLASS 20964) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 20965) (T_DEC 22) (T_DECLARE 20966) (T_DIR 24) (T_DNUMBER 25) (T_DO 20967) (T_DOUBLE_CAST 27) (T_ECHO 20968) (T_EMPTY 29) (T_ENDFOREACH 32964) (T_ENUM 20970) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 20971) (T_FOREACH 20972) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 20973) (T_GOTO 20974) (T_HALT_COMPILER 20975) (T_IF 20976) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 20977) (T_INTERFACE 20978) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 20979) (T_START_HEREDOC 69) (T_STATIC 20980) (T_STRING 20981) (T_STRING_CAST 72) (T_SWITCH 20982) (T_THROW 74) (T_TRAIT 20983) (T_TRAIT_C 76) (T_TRY 20984) (T_UNSET 20985) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 20986) (T_WHILE 20987) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 20988) (alt_if_stmt_without_else 20989) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 20990) (attributes 20991) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 20992) (class_modifier 99) (class_modifiers 20993) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 20994) (expr 20995) (fn 106) (fully_dereferenceable 107) #'20996 (function_call 109) (function_declaration_statement 20997) (if_stmt 20998) (if_stmt_without_else 20999) (inline_function 113) (inner_statement 21000) (interface_declaration_statement 21001) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 21002) (static_member 123) (trait_declaration_statement 21003) (variable 126) (variable_class_name 127) ("{" 21004) ("~" 129)) 15655 ((";" 32965)) 15656 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21126) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21127) (fn 18437) (for_exprs 32968) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21129) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 15657 ((";" 32969)) 15658 (("!" 18373) ("\"" 18374) ("$" 18375) ("(" 18376) ("+" 18377) ("-" 18378) ("@" 18379) (T_ARRAY 18380) (T_ARRAY_CAST 18381) (T_ATTRIBUTE 130) (T_BOOL_CAST 18382) (T_CLASS_C 18383) (T_CLONE 18384) (T_CONSTANT_ENCAPSED_STRING 18385) (T_DEC 18386) (T_DIR 18387) (T_DNUMBER 18388) (T_DOUBLE_CAST 18389) (T_EMPTY 18390) (T_EVAL 18391) (T_EXIT 18392) (T_FILE 18393) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 18394) (T_INC 18395) (T_INCLUDE 18396) (T_INCLUDE_ONCE 18397) (T_INT_CAST 18398) (T_ISSET 18399) (T_LINE 18400) (T_LIST 18401) (T_LNUMBER 18402) (T_MATCH 18403) (T_METHOD_C 18404) (T_NAME_FULLY_QUALIFIED 18405) (T_NAME_QUALIFIED 18406) (T_NAME_RELATIVE 18407) (T_NEW 18408) (T_NS_C 18409) (T_OBJECT_CAST 18410) (T_PRINT 18411) (T_PROPERTY_C 18412) (T_READONLY 18413) (T_REQUIRE 18414) (T_REQUIRE_ONCE 18415) (T_START_HEREDOC 18416) (T_STATIC 18417) (T_STRING 18418) (T_STRING_CAST 18419) (T_THROW 18420) (T_TRAIT_C 18421) (T_UNSET_CAST 18422) (T_VARIABLE 18423) (T_VOID_CAST 21135) (T_YIELD 18424) (T_YIELD_FROM 18425) ("[" 18426) ("`" 18427) (array_object_dereferenceable 18428) (attribute 135) (attributes 18429) (callable_expr 18430) (callable_variable 18431) (class_constant 18432) (class_name 18433) (constant 18434) (dereferenceable_scalar 18435) (expr 21136) (fn 18437) (for_cond_exprs 32970) (fully_dereferenceable 107) #'18438 (function_call 18439) (inline_function 18440) (internal_functions_in_yacc 18441) (match 18442) (name 18443) (new_dereferenceable 18444) (new_non_dereferenceable 18445) (non_empty_for_exprs 21138) (scalar 18446) (simple_variable 18447) (static_member 18448) (variable 18449) (variable_class_name 18450) ("~" 18451)) 15659 ((";" 32971)) 15660 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) ("+" 2364) ("-" 2365) ("@" 2366) (T_ARRAY 2367) (T_ARRAY_CAST 2368) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_CLASS_C 2370) (T_CLONE 2371) (T_CONSTANT_ENCAPSED_STRING 2372) (T_DEC 2373) (T_DIR 2374) (T_DNUMBER 2375) (T_DOUBLE_CAST 2376) (T_EMPTY 2377) (T_EVAL 2378) (T_EXIT 2379) (T_FILE 2380) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 2381) (T_INC 2382) (T_INCLUDE 2383) (T_INCLUDE_ONCE 2384) (T_INT_CAST 2385) (T_ISSET 2386) (T_LINE 2387) (T_LIST 2440) (T_LNUMBER 2389) (T_MATCH 2390) (T_METHOD_C 2391) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 2395) (T_NS_C 2396) (T_OBJECT_CAST 2397) (T_PRINT 2398) (T_PROPERTY_C 2399) (T_READONLY 2400) (T_REQUIRE 2401) (T_REQUIRE_ONCE 2402) (T_START_HEREDOC 2403) (T_STATIC 2404) (T_STRING 2405) (T_STRING_CAST 2406) (T_THROW 2407) (T_TRAIT_C 2408) (T_UNSET_CAST 2409) (T_VARIABLE 2410) (T_VOID_CAST 22737) (T_YIELD 2411) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 22738) (fn 2425) (for_exprs 32972) (fully_dereferenceable 107) #'2426 (function_call 2427) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_for_exprs 22740) (scalar 2434) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 15661 ((")" 32973)) 15662 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 32974) (";" 8) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 15) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21) (T_DEC 22) (T_DECLARE 23) (T_DIR 24) (T_DNUMBER 25) (T_DO 26) (T_DOUBLE_CAST 27) (T_ECHO 28) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 36) (T_FOREACH 37) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 40) (T_GOTO 41) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 47) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 68) (T_START_HEREDOC 69) (T_STATIC 70) (T_STRING 71) (T_STRING_CAST 72) (T_SWITCH 73) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 77) (T_UNSET 78) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 82) (T_WHILE 83) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 88) (alt_if_stmt_without_else 89) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (dereferenceable_scalar 103) (expr 105) (fn 106) (for_statement 32975) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 111) (if_stmt_without_else 112) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 32976) (static_member 123) (variable 126) (variable_class_name 127) ("{" 128) ("~" 129)) 15663 ((inner_statement_list 32977)) 15664 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 22750) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 22751) (T_CLASS 22752) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 22753) (T_DEC 22) (T_DECLARE 22754) (T_DIR 24) (T_DNUMBER 25) (T_DO 22755) (T_DOUBLE_CAST 27) (T_ECHO 22756) (T_EMPTY 29) (T_ENDFOR 32978) (T_ENUM 22758) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 22759) (T_FOREACH 22760) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 22761) (T_GOTO 22762) (T_HALT_COMPILER 22763) (T_IF 22764) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 22765) (T_INTERFACE 22766) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 22767) (T_START_HEREDOC 69) (T_STATIC 22768) (T_STRING 22769) (T_STRING_CAST 72) (T_SWITCH 22770) (T_THROW 74) (T_TRAIT 22771) (T_TRAIT_C 76) (T_TRY 22772) (T_UNSET 22773) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 22774) (T_WHILE 22775) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 22776) (alt_if_stmt_without_else 22777) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 22778) (attributes 22779) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 22780) (class_modifier 99) (class_modifiers 22781) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 22782) (expr 22783) (fn 106) (fully_dereferenceable 107) #'22784 (function_call 109) (function_declaration_statement 22785) (if_stmt 22786) (if_stmt_without_else 22787) (inline_function 113) (inner_statement 22788) (interface_declaration_statement 22789) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 22790) (static_member 123) (trait_declaration_statement 22791) (variable 126) (variable_class_name 127) ("{" 22792) ("~" 129)) 15665 ((";" 32979)) 15666 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 32983) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 32984) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 1066) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (non_empty_argument_list 32985) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 15667 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 32988) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 15668 (("," 1498) (possible_comma 32986)) 15669 ((")" 32987)) 15670 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 32990) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 15671 (("%" 304) (")" 32991) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 15672 ((":" 23454) (enum_backing_type 32993)) 15673 ((T_IMPLEMENTS 2989) (implements_list 32994)) 15674 ((backup_doc_comment 32995)) 15675 (("{" 32996)) 15676 ((class_statement_list 32997)) 15677 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 32998)) 15678 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 33000) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 15679 (("%" 304) (")" 33001) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 15680 (("," 22929) (";" 33003)) 15681 ((T_WHILE 33006)) 15682 (("(" 33007)) 15683 (("!" 224) ("\"" 225) ("$" 226) ("(" 227) ("+" 228) ("-" 229) ("@" 230) (T_ARRAY 231) (T_ARRAY_CAST 232) (T_ATTRIBUTE 130) (T_BOOL_CAST 233) (T_CLASS_C 234) (T_CLONE 235) (T_CONSTANT_ENCAPSED_STRING 236) (T_DEC 237) (T_DIR 238) (T_DNUMBER 239) (T_DOUBLE_CAST 240) (T_EMPTY 241) (T_EVAL 242) (T_EXIT 243) (T_FILE 244) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 245) (T_INC 246) (T_INCLUDE 247) (T_INCLUDE_ONCE 248) (T_INT_CAST 249) (T_ISSET 250) (T_LINE 251) (T_LIST 252) (T_LNUMBER 253) (T_MATCH 254) (T_METHOD_C 255) (T_NAME_FULLY_QUALIFIED 256) (T_NAME_QUALIFIED 257) (T_NAME_RELATIVE 258) (T_NEW 259) (T_NS_C 260) (T_OBJECT_CAST 261) (T_PRINT 262) (T_PROPERTY_C 263) (T_READONLY 264) (T_REQUIRE 265) (T_REQUIRE_ONCE 266) (T_START_HEREDOC 267) (T_STATIC 268) (T_STRING 269) (T_STRING_CAST 270) (T_THROW 271) (T_TRAIT_C 272) (T_UNSET_CAST 273) (T_VARIABLE 274) (T_YIELD 275) (T_YIELD_FROM 276) ("[" 277) ("`" 278) (array_object_dereferenceable 279) (attribute 135) (attributes 280) (callable_expr 281) (callable_variable 282) (class_constant 283) (class_name 284) (constant 285) (dereferenceable_scalar 286) (expr 33008) (fn 288) (fully_dereferenceable 107) #'289 (function_call 290) (inline_function 291) (internal_functions_in_yacc 292) (match 293) (name 294) (new_dereferenceable 295) (new_non_dereferenceable 296) (scalar 297) (simple_variable 298) (static_member 299) (variable 300) (variable_class_name 301) ("~" 302)) 15684 (("%" 304) (")" 33009) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 15685 ((";" 33010)) 15686 ((T_STRING 23106) (const_decl 23107) (const_list 33012)) 15687 ((")" 33013) ("," 23110)) 15688 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (":" 33014) (";" 8) ("@" 9) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 130) (T_BOOL_CAST 14) (T_BREAK 15) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 21) (T_DEC 22) (T_DECLARE 23) (T_DIR 24) (T_DNUMBER 25) (T_DO 26) (T_DOUBLE_CAST 27) (T_ECHO 28) (T_EMPTY 29) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FN 35) (T_FOR 36) (T_FOREACH 37) (T_FUNCTION 131) (T_FUNC_C 39) (T_GLOBAL 40) (T_GOTO 41) (T_IF 43) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 47) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 132) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 68) (T_START_HEREDOC 69) (T_STATIC 70) (T_STRING 71) (T_STRING_CAST 72) (T_SWITCH 73) (T_THROW 74) (T_TRAIT_C 76) (T_TRY 77) (T_UNSET 78) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 82) (T_WHILE 83) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 88) (alt_if_stmt_without_else 89) (array_object_dereferenceable 90) (attribute 135) (attributes 136) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_name 101) (constant 102) (declare_statement 33015) (dereferenceable_scalar 103) (expr 105) (fn 106) (fully_dereferenceable 107) #'138 (function_call 109) (if_stmt 111) (if_stmt_without_else 112) (inline_function 113) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 33016) (static_member 123) (variable 126) (variable_class_name 127) ("{" 128) ("~" 129)) 15689 ((inner_statement_list 33017)) 15690 (("!" 2) ("\"" 3) ("$" 4) ("(" 5) ("+" 6) ("-" 7) (";" 23116) ("@" 9) (T_ABSTRACT 10) (T_ARRAY 11) (T_ARRAY_CAST 12) (T_ATTRIBUTE 174) (T_BOOL_CAST 14) (T_BREAK 23117) (T_CLASS 23118) (T_CLASS_C 17) (T_CLONE 18) (T_CONSTANT_ENCAPSED_STRING 20) (T_CONTINUE 23119) (T_DEC 22) (T_DECLARE 23120) (T_DIR 24) (T_DNUMBER 25) (T_DO 23121) (T_DOUBLE_CAST 27) (T_ECHO 23122) (T_EMPTY 29) (T_ENDDECLARE 33018) (T_ENUM 23124) (T_EVAL 31) (T_EXIT 32) (T_FILE 33) (T_FINAL 34) (T_FN 35) (T_FOR 23125) (T_FOREACH 23126) (T_FUNCTION 38) (T_FUNC_C 39) (T_GLOBAL 23127) (T_GOTO 23128) (T_HALT_COMPILER 23129) (T_IF 23130) (T_INC 44) (T_INCLUDE 45) (T_INCLUDE_ONCE 46) (T_INLINE_HTML 23131) (T_INTERFACE 23132) (T_INT_CAST 49) (T_ISSET 50) (T_LINE 51) (T_LIST 52) (T_LNUMBER 53) (T_MATCH 54) (T_METHOD_C 55) (T_NAME_FULLY_QUALIFIED 57) (T_NAME_QUALIFIED 58) (T_NAME_RELATIVE 59) (T_NEW 60) (T_NS_C 61) (T_OBJECT_CAST 62) (T_PRINT 63) (T_PROPERTY_C 64) (T_READONLY 65) (T_REQUIRE 66) (T_REQUIRE_ONCE 67) (T_RETURN 23133) (T_START_HEREDOC 69) (T_STATIC 23134) (T_STRING 23135) (T_STRING_CAST 72) (T_SWITCH 23136) (T_THROW 74) (T_TRAIT 23137) (T_TRAIT_C 76) (T_TRY 23138) (T_UNSET 23139) (T_UNSET_CAST 79) (T_VARIABLE 81) (T_VOID_CAST 23140) (T_WHILE 23141) (T_YIELD 84) (T_YIELD_FROM 85) ("[" 86) ("`" 87) (alt_if_stmt 23142) (alt_if_stmt_without_else 23143) (array_object_dereferenceable 90) (attribute 201) (attributed_statement 23144) (attributes 23145) (callable_expr 95) (callable_variable 96) (class_constant 97) (class_declaration_statement 23146) (class_modifier 99) (class_modifiers 23147) (class_name 101) (constant 102) (dereferenceable_scalar 103) (enum_declaration_statement 23148) (expr 23149) (fn 106) (fully_dereferenceable 107) #'23150 (function_call 109) (function_declaration_statement 23151) (if_stmt 23152) (if_stmt_without_else 23153) (inline_function 113) (inner_statement 23154) (interface_declaration_statement 23155) (internal_functions_in_yacc 115) (match 116) (name 117) (new_dereferenceable 118) (new_non_dereferenceable 119) (scalar 120) (simple_variable 121) (statement 23156) (static_member 123) (trait_declaration_statement 23157) (variable 126) (variable_class_name 127) ("{" 23158) ("~" 129)) 15691 ((";" 33019)) 15692 ((";" 33022)) 15693 (("," 31541) (";" 33024)) 15694 (("!" 2360) ("\"" 2361) ("$" 2362) ("(" 2363) (")" 33028) ("+" 2364) ("-" 2365) ("@" 2366) (T_ABSTRACT 966) (T_ARRAY 5255) (T_ARRAY_CAST 2368) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 2369) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 5256) (T_CLONE 5257) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 2372) (T_CONTINUE 980) (T_DEC 2373) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 5258) (T_DNUMBER 2375) (T_DO 986) (T_DOUBLE_CAST 2376) (T_ECHO 988) (T_ELLIPSIS 33029) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 5260) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 5261) (T_EXIT 5262) (T_EXTENDS 1002) (T_FILE 5263) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 5264) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 2382) (T_INCLUDE 5265) (T_INCLUDE_ONCE 5266) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 2385) (T_ISSET 5267) (T_LINE 5268) (T_LIST 5269) (T_LNUMBER 2389) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 5270) (T_METHOD_C 5271) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 2392) (T_NAME_QUALIFIED 2393) (T_NAME_RELATIVE 2394) (T_NEW 5272) (T_NS_C 5273) (T_OBJECT_CAST 2397) (T_PRINT 5274) (T_PRIVATE 1039) (T_PROPERTY_C 5275) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 5276) (T_REQUIRE 5277) (T_REQUIRE_ONCE 5278) (T_RETURN 1046) (T_START_HEREDOC 2403) (T_STATIC 5279) (T_STRING 5280) (T_STRING_CAST 2406) (T_SWITCH 1051) (T_THROW 5281) (T_TRAIT 1053) (T_TRAIT_C 5282) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 2409) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 2410) (T_WHILE 1061) (T_YIELD 5283) (T_YIELD_FROM 2412) ("[" 2413) ("`" 2414) (argument_no_expr 5284) (array_object_dereferenceable 2416) (attribute 135) (attributes 2417) (callable_expr 2418) (callable_variable 2419) (class_constant 2420) (class_name 2421) (constant 2422) (dereferenceable_scalar 2423) (expr 33030) (fn 2425) (fully_dereferenceable 107) #'2426 (function_call 2427) (identifier 1080) (inline_function 2428) (internal_functions_in_yacc 2429) (match 2430) (name 2431) (new_dereferenceable 2432) (new_non_dereferenceable 2433) (non_empty_clone_argument_list 33031) (reserved_non_modifiers 1088) (scalar 2434) (semi_reserved 1090) (simple_variable 2435) (static_member 2436) (variable 2437) (variable_class_name 2438) ("~" 2439)) 15695 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 33037) ("+" 963) ("-" 964) ("@" 965) (T_ARRAY 1096) (T_ARRAY_CAST 968) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_CLASS_C 1097) (T_CLONE 1098) (T_CONSTANT_ENCAPSED_STRING 979) (T_DEC 981) (T_DIR 1099) (T_DNUMBER 985) (T_DOUBLE_CAST 987) (T_EMPTY 1100) (T_EVAL 1101) (T_EXIT 1102) (T_FILE 1103) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1104) (T_INC 1015) (T_INCLUDE 1105) (T_INCLUDE_ONCE 1106) (T_INT_CAST 1021) (T_ISSET 1107) (T_LINE 1108) (T_LIST 1109) (T_LNUMBER 1025) (T_MATCH 1110) (T_METHOD_C 1111) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1112) (T_NS_C 1113) (T_OBJECT_CAST 1037) (T_PRINT 1114) (T_PROPERTY_C 1115) (T_READONLY 1116) (T_REQUIRE 1117) (T_REQUIRE_ONCE 1118) (T_START_HEREDOC 1047) (T_STATIC 1119) (T_STRING 1120) (T_STRING_CAST 1050) (T_THROW 1121) (T_TRAIT_C 1122) (T_UNSET_CAST 1057) (T_VARIABLE 1060) (T_YIELD 1123) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1503) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (scalar 1089) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 15696 (("%" 2442) (")" 33034) ("*" 2443) ("+" 2444) ("," 33035) ("-" 2445) ("." 2446) ("/" 2447) ("<" 2448) (">" 2449) ("?" 2450) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 2451) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 2452) (T_BOOLEAN_AND 2453) (T_BOOLEAN_OR 2454) (T_COALESCE 2455) (T_INSTANCEOF 2456) (T_IS_EQUAL 2457) (T_IS_GREATER_OR_EQUAL 2458) (T_IS_IDENTICAL 2459) (T_IS_NOT_EQUAL 2460) (T_IS_NOT_IDENTICAL 2461) (T_IS_SMALLER_OR_EQUAL 2462) (T_LOGICAL_AND 2463) (T_LOGICAL_OR 2464) (T_LOGICAL_XOR 2465) (T_PIPE 2466) (T_POW 2467) (T_SL 2468) (T_SPACESHIP 2469) (T_SR 2470) ("^" 2471) ("|" 2472)) 15697 (("," 5287) (possible_comma 33032)) 15698 ((")" 33033)) 15699 (("!" 958) ("\"" 959) ("$" 960) ("(" 961) (")" 33036) ("+" 963) ("-" 964) ("@" 965) (T_ABSTRACT 966) (T_ARRAY 967) (T_ARRAY_CAST 968) (T_AS 969) (T_ATTRIBUTE 130) (T_BOOL_CAST 970) (T_BREAK 971) (T_CALLABLE 972) (T_CASE 973) (T_CATCH 974) (T_CLASS 975) (T_CLASS_C 976) (T_CLONE 977) (T_CONST 978) (T_CONSTANT_ENCAPSED_STRING 979) (T_CONTINUE 980) (T_DEC 981) (T_DECLARE 982) (T_DEFAULT 983) (T_DIR 984) (T_DNUMBER 985) (T_DO 986) (T_DOUBLE_CAST 987) (T_ECHO 988) (T_ELLIPSIS 1501) (T_ELSE 990) (T_ELSEIF 991) (T_EMPTY 992) (T_ENDDECLARE 993) (T_ENDFOR 994) (T_ENDFOREACH 995) (T_ENDIF 996) (T_ENDSWITCH 997) (T_ENDWHILE 998) (T_ENUM 999) (T_EVAL 1000) (T_EXIT 1001) (T_EXTENDS 1002) (T_FILE 1003) (T_FINAL 1004) (T_FINALLY 1005) (T_FN 1006) (T_FOR 1007) (T_FOREACH 1008) (T_FUNCTION 1009) (T_FUNC_C 1010) (T_GLOBAL 1011) (T_GOTO 1012) (T_IF 1013) (T_IMPLEMENTS 1014) (T_INC 1015) (T_INCLUDE 1016) (T_INCLUDE_ONCE 1017) (T_INSTANCEOF 1018) (T_INSTEADOF 1019) (T_INTERFACE 1020) (T_INT_CAST 1021) (T_ISSET 1022) (T_LINE 1023) (T_LIST 1024) (T_LNUMBER 1025) (T_LOGICAL_AND 1026) (T_LOGICAL_OR 1027) (T_LOGICAL_XOR 1028) (T_MATCH 1029) (T_METHOD_C 1030) (T_NAMESPACE 1031) (T_NAME_FULLY_QUALIFIED 1032) (T_NAME_QUALIFIED 1033) (T_NAME_RELATIVE 1034) (T_NEW 1035) (T_NS_C 1036) (T_OBJECT_CAST 1037) (T_PRINT 1038) (T_PRIVATE 1039) (T_PROPERTY_C 1040) (T_PROTECTED 1041) (T_PUBLIC 1042) (T_READONLY 1043) (T_REQUIRE 1044) (T_REQUIRE_ONCE 1045) (T_RETURN 1046) (T_START_HEREDOC 1047) (T_STATIC 1048) (T_STRING 1049) (T_STRING_CAST 1050) (T_SWITCH 1051) (T_THROW 1052) (T_TRAIT 1053) (T_TRAIT_C 1054) (T_TRY 1055) (T_UNSET 1056) (T_UNSET_CAST 1057) (T_USE 1058) (T_VAR 1059) (T_VARIABLE 1060) (T_WHILE 1061) (T_YIELD 1062) (T_YIELD_FROM 1063) ("[" 1064) ("`" 1065) (argument 5294) (argument_no_expr 1067) (array_object_dereferenceable 1068) (attribute 135) (attributes 1069) (callable_expr 1070) (callable_variable 1071) (class_constant 1072) (class_name 1073) (constant 1074) (dereferenceable_scalar 1075) (expr 1076) (fn 1077) (fully_dereferenceable 107) #'1078 (function_call 1079) (identifier 1080) (inline_function 1081) (internal_functions_in_yacc 1082) (match 1083) (name 1084) (new_dereferenceable 1085) (new_non_dereferenceable 1086) (reserved_non_modifiers 1088) (scalar 1089) (semi_reserved 1090) (simple_variable 1091) (static_member 1092) (variable 1093) (variable_class_name 1094) ("~" 1095)) 15700 ((T_EXTENDS 2987) (extends_from 33039)) 15701 ((T_IMPLEMENTS 2989) (implements_list 33040)) 15702 ((backup_doc_comment 33041)) 15703 (("{" 33042)) 15704 ((class_statement_list 33043)) 15705 ((T_ABSTRACT 2994) (T_ATTRIBUTE 2995) (T_CASE 2996) (T_FINAL 2997) (T_PRIVATE 2998) (T_PRIVATE_SET 2999) (T_PROTECTED 3000) (T_PROTECTED_SET 3001) (T_PUBLIC 3002) (T_PUBLIC_SET 3003) (T_READONLY 3004) (T_STATIC 3005) (T_USE 3006) (T_VAR 3007) (attribute 3008) (attributed_class_statement 3009) (attributes 3010) (class_const_modifiers 3011) (class_statement 3012) (enum_case 3013) (member_modifier 3014) (method_modifiers 3015) (non_empty_member_modifiers 3016) (property_modifiers 3017) ("}" 33044)) 15706 ((";" 33046)) 15707 (("," 6013) (possible_comma 33049)) 15708 (("]" 33050)) 15709 (("!" 1883) ("\"" 1884) ("$" 1885) ("(" 1886) ("+" 1887) ("-" 1888) ("@" 1889) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 1534) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 1535) (T_ARRAY 1890) (T_ARRAY_CAST 1891) (T_ATTRIBUTE 130) (T_BOOL_CAST 1892) (T_CLASS_C 1893) (T_CLONE 1894) (T_CONSTANT_ENCAPSED_STRING 1895) (T_DEC 1896) (T_DIR 1897) (T_DNUMBER 1898) (T_DOUBLE_CAST 1899) (T_ELLIPSIS 1900) (T_EMPTY 1901) (T_EVAL 1902) (T_EXIT 1903) (T_FILE 1904) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 1905) (T_INC 1906) (T_INCLUDE 1907) (T_INCLUDE_ONCE 1908) (T_INT_CAST 1909) (T_ISSET 1910) (T_LINE 1911) (T_LIST 1912) (T_LNUMBER 1913) (T_MATCH 1914) (T_METHOD_C 1915) (T_NAME_FULLY_QUALIFIED 1916) (T_NAME_QUALIFIED 1917) (T_NAME_RELATIVE 1918) (T_NEW 1919) (T_NS_C 1920) (T_OBJECT_CAST 1921) (T_PRINT 1922) (T_PROPERTY_C 1923) (T_READONLY 1924) (T_REQUIRE 1925) (T_REQUIRE_ONCE 1926) (T_START_HEREDOC 1927) (T_STATIC 1928) (T_STRING 1929) (T_STRING_CAST 1930) (T_THROW 1931) (T_TRAIT_C 1932) (T_UNSET_CAST 1933) (T_VARIABLE 1934) (T_YIELD 1935) (T_YIELD_FROM 1936) ("[" 1937) ("`" 1938) (ampersand 1939) (array_object_dereferenceable 1940) (array_pair 1941) (array_pair_list 33053) (attribute 135) (attributes 1943) (callable_expr 1944) (callable_variable 1945) (class_constant 1946) (class_name 1947) (constant 1948) (dereferenceable_scalar 1949) (expr 1950) (fn 1951) (fully_dereferenceable 107) #'1952 (function_call 1953) (inline_function 1954) (internal_functions_in_yacc 1955) (match 1956) (name 1957) (new_dereferenceable 1958) (new_non_dereferenceable 1959) (non_empty_array_pair_list 1960) (possible_array_pair 1961) (scalar 1962) (simple_variable 1963) (static_member 1964) (variable 1965) (variable_class_name 1966) ("~" 1967)) 15710 ((")" 33054)) 15711 (("%" 304) (")" 33034) ("*" 305) ("+" 306) ("-" 307) ("." 308) ("/" 309) ("<" 310) (">" 311) ("?" 312) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 313) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 314) (T_BOOLEAN_AND 315) (T_BOOLEAN_OR 316) (T_COALESCE 317) (T_INSTANCEOF 318) (T_IS_EQUAL 319) (T_IS_GREATER_OR_EQUAL 320) (T_IS_IDENTICAL 321) (T_IS_NOT_EQUAL 322) (T_IS_NOT_IDENTICAL 323) (T_IS_SMALLER_OR_EQUAL 324) (T_LOGICAL_AND 325) (T_LOGICAL_OR 326) (T_LOGICAL_XOR 327) (T_PIPE 328) (T_POW 329) (T_SL 330) (T_SPACESHIP 331) (T_SR 332) ("^" 333) ("|" 334)) 15712 (("!" 520) ("\"" 521) ("$" 522) ("(" 523) ("+" 524) ("-" 525) ("@" 526) (T_ARRAY 527) (T_ARRAY_CAST 528) (T_ATTRIBUTE 130) (T_BOOL_CAST 529) (T_CLASS_C 530) (T_CLONE 531) (T_CONSTANT_ENCAPSED_STRING 532) (T_DEC 533) (T_DIR 534) (T_DNUMBER 535) (T_DOUBLE_CAST 536) (T_EMPTY 537) (T_EVAL 538) (T_EXIT 539) (T_FILE 540) (T_FN 35) (T_FUNCTION 131) (T_FUNC_C 541) (T_INC 542) (T_INCLUDE 543) (T_INCLUDE_ONCE 544) (T_INT_CAST 545) (T_ISSET 546) (T_LINE 547) (T_LIST 548) (T_LNUMBER 549) (T_MATCH 550) (T_METHOD_C 551) (T_NAME_FULLY_QUALIFIED 552) (T_NAME_QUALIFIED 553) (T_NAME_RELATIVE 554) (T_NEW 555) (T_NS_C 556) (T_OBJECT_CAST 557) (T_PRINT 558) (T_PROPERTY_C 559) (T_READONLY 560) (T_REQUIRE 561) (T_REQUIRE_ONCE 562) (T_START_HEREDOC 563) (T_STATIC 564) (T_STRING 565) (T_STRING_CAST 566) (T_THROW 567) (T_TRAIT_C 568) (T_UNSET_CAST 569) (T_VARIABLE 570) (T_YIELD 571) (T_YIELD_FROM 572) ("[" 573) ("`" 574) (array_object_dereferenceable 575) (attribute 135) (attributes 576) (callable_expr 577) (callable_variable 578) (class_constant 579) (class_name 580) (constant 581) (dereferenceable_scalar 582) (expr 33061) (fn 584) (fully_dereferenceable 107) #'585 (function_call 586) (inline_function 587) (internal_functions_in_yacc 588) (match 589) (name 590) (new_dereferenceable 591) (new_non_dereferenceable 592) (scalar 593) (simple_variable 594) (static_member 595) (variable 596) (variable_class_name 597) ("~" 598)) 15713 (("%" 600) ("*" 601) ("+" 602) ("-" 603) ("." 604) ("/" 605) ("<" 606) (">" 607) ("?" 608) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG 609) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG 610) (T_BOOLEAN_AND 611) (T_BOOLEAN_OR 612) (T_COALESCE 613) (T_INSTANCEOF 614) (T_IS_EQUAL 615) (T_IS_GREATER_OR_EQUAL 616) (T_IS_IDENTICAL 617) (T_IS_NOT_EQUAL 618) (T_IS_NOT_IDENTICAL 619) (T_IS_SMALLER_OR_EQUAL 620) (T_LOGICAL_AND 621) (T_LOGICAL_OR 622) (T_LOGICAL_XOR 623) (T_PIPE 624) (T_POW 625) (T_SL 626) (T_SPACESHIP 627) (T_SR 628) ("^" 629) ("|" 630) ("}" 33062)) 15714 (("\"" 33064) (T_CURLY_OPEN 5352) (T_DOLLAR_OPEN_CURLY_BRACES 5353) (T_ENCAPSED_AND_WHITESPACE 5359) (T_VARIABLE 5355) (encaps_var 5360))))
  "The generated distinct goto-tables.")

(defvar
  phps-mode-parser--table-productions-number-reverse
  #s(hash-table test equal data (0 ((start) (top_statement_list)) 1 ((reserved_non_modifiers) (T_INCLUDE)) 2 ((reserved_non_modifiers) (T_INCLUDE_ONCE)) 3 ((reserved_non_modifiers) (T_EVAL)) 4 ((reserved_non_modifiers) (T_REQUIRE)) 5 ((reserved_non_modifiers) (T_REQUIRE_ONCE)) 6 ((reserved_non_modifiers) (T_LOGICAL_OR)) 7 ((reserved_non_modifiers) (T_LOGICAL_XOR)) 8 ((reserved_non_modifiers) (T_LOGICAL_AND)) 9 ((reserved_non_modifiers) (T_INSTANCEOF)) 10 ((reserved_non_modifiers) (T_NEW)) 11 ((reserved_non_modifiers) (T_CLONE)) 12 ((reserved_non_modifiers) (T_EXIT)) 13 ((reserved_non_modifiers) (T_IF)) 14 ((reserved_non_modifiers) (T_ELSEIF)) 15 ((reserved_non_modifiers) (T_ELSE)) 16 ((reserved_non_modifiers) (T_ENDIF)) 17 ((reserved_non_modifiers) (T_ECHO)) 18 ((reserved_non_modifiers) (T_DO)) 19 ((reserved_non_modifiers) (T_WHILE)) 20 ((reserved_non_modifiers) (T_ENDWHILE)) 21 ((reserved_non_modifiers) (T_FOR)) 22 ((reserved_non_modifiers) (T_ENDFOR)) 23 ((reserved_non_modifiers) (T_FOREACH)) 24 ((reserved_non_modifiers) (T_ENDFOREACH)) 25 ((reserved_non_modifiers) (T_DECLARE)) 26 ((reserved_non_modifiers) (T_ENDDECLARE)) 27 ((reserved_non_modifiers) (T_AS)) 28 ((reserved_non_modifiers) (T_TRY)) 29 ((reserved_non_modifiers) (T_CATCH)) 30 ((reserved_non_modifiers) (T_FINALLY)) 31 ((reserved_non_modifiers) (T_THROW)) 32 ((reserved_non_modifiers) (T_USE)) 33 ((reserved_non_modifiers) (T_INSTEADOF)) 34 ((reserved_non_modifiers) (T_GLOBAL)) 35 ((reserved_non_modifiers) (T_VAR)) 36 ((reserved_non_modifiers) (T_UNSET)) 37 ((reserved_non_modifiers) (T_ISSET)) 38 ((reserved_non_modifiers) (T_EMPTY)) 39 ((reserved_non_modifiers) (T_CONTINUE)) 40 ((reserved_non_modifiers) (T_GOTO)) 41 ((reserved_non_modifiers) (T_FUNCTION)) 42 ((reserved_non_modifiers) (T_CONST)) 43 ((reserved_non_modifiers) (T_RETURN)) 44 ((reserved_non_modifiers) (T_PRINT)) 45 ((reserved_non_modifiers) (T_YIELD)) 46 ((reserved_non_modifiers) (T_LIST)) 47 ((reserved_non_modifiers) (T_SWITCH)) 48 ((reserved_non_modifiers) (T_ENDSWITCH)) 49 ((reserved_non_modifiers) (T_CASE)) 50 ((reserved_non_modifiers) (T_DEFAULT)) 51 ((reserved_non_modifiers) (T_BREAK)) 52 ((reserved_non_modifiers) (T_ARRAY)) 53 ((reserved_non_modifiers) (T_CALLABLE)) 54 ((reserved_non_modifiers) (T_EXTENDS)) 55 ((reserved_non_modifiers) (T_IMPLEMENTS)) 56 ((reserved_non_modifiers) (T_NAMESPACE)) 57 ((reserved_non_modifiers) (T_TRAIT)) 58 ((reserved_non_modifiers) (T_INTERFACE)) 59 ((reserved_non_modifiers) (T_CLASS)) 60 ((reserved_non_modifiers) (T_CLASS_C)) 61 ((reserved_non_modifiers) (T_TRAIT_C)) 62 ((reserved_non_modifiers) (T_FUNC_C)) 63 ((reserved_non_modifiers) (T_METHOD_C)) 64 ((reserved_non_modifiers) (T_LINE)) 65 ((reserved_non_modifiers) (T_FILE)) 66 ((reserved_non_modifiers) (T_DIR)) 67 ((reserved_non_modifiers) (T_NS_C)) 68 ((reserved_non_modifiers) (T_FN)) 69 ((reserved_non_modifiers) (T_MATCH)) 70 ((reserved_non_modifiers) (T_ENUM)) 71 ((reserved_non_modifiers) (T_PROPERTY_C)) 72 ((semi_reserved) (reserved_non_modifiers)) 73 ((semi_reserved) (T_STATIC)) 74 ((semi_reserved) (T_ABSTRACT)) 75 ((semi_reserved) (T_FINAL)) 76 ((semi_reserved) (T_PRIVATE)) 77 ((semi_reserved) (T_PROTECTED)) 78 ((semi_reserved) (T_PUBLIC)) 79 ((semi_reserved) (T_READONLY)) 80 ((ampersand) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG)) 81 ((ampersand) (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG)) 82 ((identifier) (T_STRING)) 83 ((identifier) (semi_reserved)) 84 ((top_statement_list) (top_statement_list top_statement)) 85 ((top_statement_list) (%empty)) 86 ((namespace_declaration_name) (identifier)) 87 ((namespace_declaration_name) (T_NAME_QUALIFIED)) 88 ((namespace_name) (T_STRING)) 89 ((namespace_name) (T_NAME_QUALIFIED)) 90 ((legacy_namespace_name) (namespace_name)) 91 ((legacy_namespace_name) (T_NAME_FULLY_QUALIFIED)) 92 ((name) (T_STRING)) 93 ((name) (T_NAME_QUALIFIED)) 94 ((name) (T_NAME_FULLY_QUALIFIED)) 95 ((name) (T_NAME_RELATIVE)) 96 ((attribute_decl) (class_name)) 97 ((attribute_decl) (class_name argument_list)) 98 ((attribute_group) (attribute_decl)) 99 ((attribute_group) (attribute_group "," attribute_decl)) 100 ((attribute) (T_ATTRIBUTE attribute_group possible_comma "]")) 101 ((attributes) (attribute)) 102 ((attributes) (attributes attribute)) 103 ((attributed_statement) (function_declaration_statement)) 104 ((attributed_statement) (class_declaration_statement)) 105 ((attributed_statement) (trait_declaration_statement)) 106 ((attributed_statement) (interface_declaration_statement)) 107 ((attributed_statement) (enum_declaration_statement)) 108 ((attributed_top_statement) (attributed_statement)) 109 ((attributed_top_statement) (T_CONST const_list ";")) 110 ((top_statement) (statement)) 111 ((top_statement) (attributed_top_statement)) 112 ((top_statement) (attributes attributed_top_statement)) 113 ((top_statement) (T_HALT_COMPILER "(" ")" ";")) 114 ((top_statement) (T_NAMESPACE namespace_declaration_name ";")) 115 ((top_statement) (T_NAMESPACE namespace_declaration_name "{" top_statement_list "}")) 116 ((top_statement) (T_NAMESPACE "{" top_statement_list "}")) 117 ((top_statement) (T_USE mixed_group_use_declaration ";")) 118 ((top_statement) (T_USE use_type group_use_declaration ";")) 119 ((top_statement) (T_USE use_declarations ";")) 120 ((top_statement) (T_USE use_type use_declarations ";")) 121 ((use_type) (T_FUNCTION)) 122 ((use_type) (T_CONST)) 123 ((group_use_declaration) (legacy_namespace_name T_NS_SEPARATOR "{" unprefixed_use_declarations possible_comma "}")) 124 ((mixed_group_use_declaration) (legacy_namespace_name T_NS_SEPARATOR "{" inline_use_declarations possible_comma "}")) 125 ((possible_comma) (%empty)) 126 ((possible_comma) (",")) 127 ((inline_use_declarations) (inline_use_declarations "," inline_use_declaration)) 128 ((inline_use_declarations) (inline_use_declaration)) 129 ((unprefixed_use_declarations) (unprefixed_use_declarations "," unprefixed_use_declaration)) 130 ((unprefixed_use_declarations) (unprefixed_use_declaration)) 131 ((use_declarations) (use_declarations "," use_declaration)) 132 ((use_declarations) (use_declaration)) 133 ((inline_use_declaration) (unprefixed_use_declaration)) 134 ((inline_use_declaration) (use_type unprefixed_use_declaration)) 135 ((unprefixed_use_declaration) (namespace_name)) 136 ((unprefixed_use_declaration) (namespace_name T_AS T_STRING)) 137 ((use_declaration) (legacy_namespace_name)) 138 ((use_declaration) (legacy_namespace_name T_AS T_STRING)) 139 ((const_list) (const_list "," const_decl)) 140 ((const_list) (const_decl)) 141 ((inner_statement_list) (inner_statement_list inner_statement)) 142 ((inner_statement_list) (%empty)) 143 ((inner_statement) (statement)) 144 ((inner_statement) (attributed_statement)) 145 ((inner_statement) (attributes attributed_statement)) 146 ((inner_statement) (T_HALT_COMPILER "(" ")" ";")) 147 ((statement) ("{" inner_statement_list "}")) 148 ((statement) (if_stmt)) 149 ((statement) (alt_if_stmt)) 150 ((statement) (T_WHILE "(" expr ")" while_statement)) 151 ((statement) (T_DO statement T_WHILE "(" expr ")" ";")) 152 ((statement) (T_FOR "(" for_exprs ";" for_cond_exprs ";" for_exprs ")" for_statement)) 153 ((statement) (T_SWITCH "(" expr ")" switch_case_list)) 154 ((statement) (T_BREAK optional_expr ";")) 155 ((statement) (T_CONTINUE optional_expr ";")) 156 ((statement) (T_RETURN optional_expr ";")) 157 ((statement) (T_GLOBAL global_var_list ";")) 158 ((statement) (T_STATIC static_var_list ";")) 159 ((statement) (T_ECHO echo_expr_list ";")) 160 ((statement) (T_INLINE_HTML)) 161 ((statement) (expr ";")) 162 ((statement) (T_UNSET "(" unset_variables possible_comma ")" ";")) 163 ((statement) (T_FOREACH "(" expr T_AS foreach_variable ")" foreach_statement)) 164 ((statement) (T_FOREACH "(" expr T_AS foreach_variable T_DOUBLE_ARROW foreach_variable ")" foreach_statement)) 165 ((statement) (T_DECLARE "(" const_list ")" declare_statement)) 166 ((statement) (";")) 167 ((statement) (T_TRY "{" inner_statement_list "}" catch_list finally_statement)) 168 ((statement) (T_GOTO T_STRING ";")) 169 ((statement) (T_STRING ":")) 170 ((statement) (T_VOID_CAST expr ";")) 171 ((catch_list) (%empty)) 172 ((catch_list) (catch_list T_CATCH "(" catch_name_list optional_variable ")" "{" inner_statement_list "}")) 173 ((catch_name_list) (class_name)) 174 ((catch_name_list) (catch_name_list "|" class_name)) 175 ((optional_variable) (%empty)) 176 ((optional_variable) (T_VARIABLE)) 177 ((finally_statement) (%empty)) 178 ((finally_statement) (T_FINALLY "{" inner_statement_list "}")) 179 ((unset_variables) (unset_variable)) 180 ((unset_variables) (unset_variables "," unset_variable)) 181 ((unset_variable) (variable)) 182 ((function_name) (T_STRING)) 183 ((function_name) (T_READONLY)) 184 ((function_declaration_statement) (function returns_ref function_name backup_doc_comment "(" parameter_list ")" return_type backup_fn_flags "{" inner_statement_list "}" backup_fn_flags)) 185 ((is_reference) (%empty)) 186 ((is_reference) (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG)) 187 ((is_variadic) (%empty)) 188 ((is_variadic) (T_ELLIPSIS)) 189 ((class_declaration_statement) (class_modifiers T_CLASS T_STRING extends_from implements_list backup_doc_comment "{" class_statement_list "}")) 190 ((class_declaration_statement) (T_CLASS T_STRING extends_from implements_list backup_doc_comment "{" class_statement_list "}")) 191 ((class_modifiers) (class_modifier)) 192 ((class_modifiers) (class_modifiers class_modifier)) 193 ((anonymous_class_modifiers) (class_modifier)) 194 ((anonymous_class_modifiers) (anonymous_class_modifiers class_modifier)) 195 ((anonymous_class_modifiers_optional) (%empty)) 196 ((anonymous_class_modifiers_optional) (anonymous_class_modifiers)) 197 ((class_modifier) (T_ABSTRACT)) 198 ((class_modifier) (T_FINAL)) 199 ((class_modifier) (T_READONLY)) 200 ((trait_declaration_statement) (T_TRAIT T_STRING backup_doc_comment "{" class_statement_list "}")) 201 ((interface_declaration_statement) (T_INTERFACE T_STRING interface_extends_list backup_doc_comment "{" class_statement_list "}")) 202 ((enum_declaration_statement) (T_ENUM T_STRING enum_backing_type implements_list backup_doc_comment "{" class_statement_list "}")) 203 ((enum_backing_type) (%empty)) 204 ((enum_backing_type) (":" type_expr)) 205 ((enum_case) (T_CASE backup_doc_comment identifier enum_case_expr ";")) 206 ((enum_case_expr) (%empty)) 207 ((enum_case_expr) ("=" expr)) 208 ((extends_from) (%empty)) 209 ((extends_from) (T_EXTENDS class_name)) 210 ((interface_extends_list) (%empty)) 211 ((interface_extends_list) (T_EXTENDS class_name_list)) 212 ((implements_list) (%empty)) 213 ((implements_list) (T_IMPLEMENTS class_name_list)) 214 ((foreach_variable) (variable)) 215 ((foreach_variable) (ampersand variable)) 216 ((foreach_variable) (T_LIST "(" array_pair_list ")")) 217 ((foreach_variable) ("[" array_pair_list "]")) 218 ((for_statement) (statement)) 219 ((for_statement) (":" inner_statement_list T_ENDFOR ";")) 220 ((foreach_statement) (statement)) 221 ((foreach_statement) (":" inner_statement_list T_ENDFOREACH ";")) 222 ((declare_statement) (statement)) 223 ((declare_statement) (":" inner_statement_list T_ENDDECLARE ";")) 224 ((switch_case_list) ("{" case_list "}")) 225 ((switch_case_list) ("{" ";" case_list "}")) 226 ((switch_case_list) (":" case_list T_ENDSWITCH ";")) 227 ((switch_case_list) (":" ";" case_list T_ENDSWITCH ";")) 228 ((case_list) (%empty)) 229 ((case_list) (case_list T_CASE expr ":" inner_statement_list)) 230 ((case_list) (case_list T_CASE expr ";" inner_statement_list)) 231 ((case_list) (case_list T_DEFAULT ":" inner_statement_list)) 232 ((case_list) (case_list T_DEFAULT ";" inner_statement_list)) 233 ((match) (T_MATCH "(" expr ")" "{" match_arm_list "}")) 234 ((match_arm_list) (%empty)) 235 ((match_arm_list) (non_empty_match_arm_list possible_comma)) 236 ((non_empty_match_arm_list) (match_arm)) 237 ((non_empty_match_arm_list) (non_empty_match_arm_list "," match_arm)) 238 ((match_arm) (match_arm_cond_list possible_comma T_DOUBLE_ARROW expr)) 239 ((match_arm) (T_DEFAULT possible_comma T_DOUBLE_ARROW expr)) 240 ((match_arm_cond_list) (expr)) 241 ((match_arm_cond_list) (match_arm_cond_list "," expr)) 242 ((while_statement) (statement)) 243 ((while_statement) (":" inner_statement_list T_ENDWHILE ";")) 244 ((if_stmt_without_else) (T_IF "(" expr ")" statement)) 245 ((if_stmt_without_else) (if_stmt_without_else T_ELSEIF "(" expr ")" statement)) 246 ((if_stmt) (if_stmt_without_else)) 247 ((if_stmt) (if_stmt_without_else T_ELSE statement)) 248 ((alt_if_stmt_without_else) (T_IF "(" expr ")" ":" inner_statement_list)) 249 ((alt_if_stmt_without_else) (alt_if_stmt_without_else T_ELSEIF "(" expr ")" ":" inner_statement_list)) 250 ((alt_if_stmt) (alt_if_stmt_without_else T_ENDIF ";")) 251 ((alt_if_stmt) (alt_if_stmt_without_else T_ELSE ":" inner_statement_list T_ENDIF ";")) 252 ((parameter_list) (non_empty_parameter_list possible_comma)) 253 ((parameter_list) (%empty)) 254 ((non_empty_parameter_list) (attributed_parameter)) 255 ((non_empty_parameter_list) (non_empty_parameter_list "," attributed_parameter)) 256 ((attributed_parameter) (attributes parameter)) 257 ((attributed_parameter) (parameter)) 258 ((optional_cpp_modifiers) (%empty)) 259 ((optional_cpp_modifiers) (non_empty_member_modifiers)) 260 ((parameter) (optional_cpp_modifiers optional_type_without_static is_reference is_variadic T_VARIABLE backup_doc_comment optional_property_hook_list)) 261 ((parameter) (optional_cpp_modifiers optional_type_without_static is_reference is_variadic T_VARIABLE backup_doc_comment "=" expr optional_property_hook_list)) 262 ((optional_type_without_static) (%empty)) 263 ((optional_type_without_static) (type_expr_without_static)) 264 ((type_expr) (type)) 265 ((type_expr) ("?" type)) 266 ((type_expr) (union_type)) 267 ((type_expr) (intersection_type)) 268 ((type) (type_without_static)) 269 ((type) (T_STATIC)) 270 ((union_type_element) (type)) 271 ((union_type_element) ("(" intersection_type ")")) 272 ((union_type) (union_type_element "|" union_type_element)) 273 ((union_type) (union_type "|" union_type_element)) 274 ((intersection_type) (type T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG type)) 275 ((intersection_type) (intersection_type T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG type)) 276 ((type_expr_without_static) (type_without_static)) 277 ((type_expr_without_static) ("?" type_without_static)) 278 ((type_expr_without_static) (union_type_without_static)) 279 ((type_expr_without_static) (intersection_type_without_static)) 280 ((type_without_static) (T_ARRAY)) 281 ((type_without_static) (T_CALLABLE)) 282 ((type_without_static) (name)) 283 ((union_type_without_static_element) (type_without_static)) 284 ((union_type_without_static_element) ("(" intersection_type_without_static ")")) 285 ((union_type_without_static) (union_type_without_static_element "|" union_type_without_static_element)) 286 ((union_type_without_static) (union_type_without_static "|" union_type_without_static_element)) 287 ((intersection_type_without_static) (type_without_static T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG type_without_static)) 288 ((intersection_type_without_static) (intersection_type_without_static T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG type_without_static)) 289 ((return_type) (%empty)) 290 ((return_type) (":" type_expr)) 291 ((argument_list) ("(" ")")) 292 ((argument_list) ("(" non_empty_argument_list possible_comma ")")) 293 ((argument_list) ("(" T_ELLIPSIS ")")) 294 ((non_empty_argument_list) (argument)) 295 ((non_empty_argument_list) (non_empty_argument_list "," argument)) 296 ((clone_argument_list) ("(" ")")) 297 ((clone_argument_list) ("(" non_empty_clone_argument_list possible_comma ")")) 298 ((clone_argument_list) ("(" expr "," ")")) 299 ((clone_argument_list) ("(" T_ELLIPSIS ")")) 300 ((non_empty_clone_argument_list) (expr "," argument)) 301 ((non_empty_clone_argument_list) (argument_no_expr)) 302 ((non_empty_clone_argument_list) (non_empty_clone_argument_list "," argument)) 303 ((argument_no_expr) (identifier ":" expr)) 304 ((argument_no_expr) (T_ELLIPSIS expr)) 305 ((argument) (expr)) 306 ((argument) (argument_no_expr)) 307 ((global_var_list) (global_var_list "," global_var)) 308 ((global_var_list) (global_var)) 309 ((global_var) (simple_variable)) 310 ((static_var_list) (static_var_list "," static_var)) 311 ((static_var_list) (static_var)) 312 ((static_var) (T_VARIABLE)) 313 ((static_var) (T_VARIABLE "=" expr)) 314 ((class_statement_list) (class_statement_list class_statement)) 315 ((class_statement_list) (%empty)) 316 ((attributed_class_statement) (property_modifiers optional_type_without_static property_list ";")) 317 ((attributed_class_statement) (property_modifiers optional_type_without_static hooked_property)) 318 ((attributed_class_statement) (class_const_modifiers T_CONST class_const_list ";")) 319 ((attributed_class_statement) (class_const_modifiers T_CONST type_expr class_const_list ";")) 320 ((attributed_class_statement) (method_modifiers function returns_ref identifier backup_doc_comment "(" parameter_list ")" return_type backup_fn_flags method_body backup_fn_flags)) 321 ((attributed_class_statement) (enum_case)) 322 ((class_statement) (attributed_class_statement)) 323 ((class_statement) (attributes attributed_class_statement)) 324 ((class_statement) (T_USE class_name_list trait_adaptations)) 325 ((class_name_list) (class_name)) 326 ((class_name_list) (class_name_list "," class_name)) 327 ((trait_adaptations) (";")) 328 ((trait_adaptations) ("{" "}")) 329 ((trait_adaptations) ("{" trait_adaptation_list "}")) 330 ((trait_adaptation_list) (trait_adaptation)) 331 ((trait_adaptation_list) (trait_adaptation_list trait_adaptation)) 332 ((trait_adaptation) (trait_precedence ";")) 333 ((trait_adaptation) (trait_alias ";")) 334 ((trait_precedence) (absolute_trait_method_reference T_INSTEADOF class_name_list)) 335 ((trait_alias) (trait_method_reference T_AS T_STRING)) 336 ((trait_alias) (trait_method_reference T_AS reserved_non_modifiers)) 337 ((trait_alias) (trait_method_reference T_AS member_modifier identifier)) 338 ((trait_alias) (trait_method_reference T_AS member_modifier)) 339 ((trait_method_reference) (identifier)) 340 ((trait_method_reference) (absolute_trait_method_reference)) 341 ((absolute_trait_method_reference) (class_name T_PAAMAYIM_NEKUDOTAYIM identifier)) 342 ((method_body) (";")) 343 ((method_body) ("{" inner_statement_list "}")) 344 ((property_modifiers) (non_empty_member_modifiers)) 345 ((property_modifiers) (T_VAR)) 346 ((method_modifiers) (%empty)) 347 ((method_modifiers) (non_empty_member_modifiers)) 348 ((class_const_modifiers) (%empty)) 349 ((class_const_modifiers) (non_empty_member_modifiers)) 350 ((non_empty_member_modifiers) (member_modifier)) 351 ((non_empty_member_modifiers) (non_empty_member_modifiers member_modifier)) 352 ((member_modifier) (T_PUBLIC)) 353 ((member_modifier) (T_PROTECTED)) 354 ((member_modifier) (T_PRIVATE)) 355 ((member_modifier) (T_PUBLIC_SET)) 356 ((member_modifier) (T_PROTECTED_SET)) 357 ((member_modifier) (T_PRIVATE_SET)) 358 ((member_modifier) (T_STATIC)) 359 ((member_modifier) (T_ABSTRACT)) 360 ((member_modifier) (T_FINAL)) 361 ((member_modifier) (T_READONLY)) 362 ((property_list) (property_list "," property)) 363 ((property_list) (property)) 364 ((property) (T_VARIABLE backup_doc_comment)) 365 ((property) (T_VARIABLE "=" expr backup_doc_comment)) 366 ((hooked_property) (T_VARIABLE backup_doc_comment "{" property_hook_list "}")) 367 ((hooked_property) (T_VARIABLE "=" expr backup_doc_comment "{" property_hook_list "}")) 368 ((property_hook_list) (%empty)) 369 ((property_hook_list) (property_hook_list property_hook)) 370 ((property_hook_list) (property_hook_list attributes property_hook)) 371 ((optional_property_hook_list) (%empty)) 372 ((optional_property_hook_list) ("{" property_hook_list "}")) 373 ((property_hook_modifiers) (%empty)) 374 ((property_hook_modifiers) (non_empty_member_modifiers)) 375 ((property_hook) (property_hook_modifiers returns_ref T_STRING backup_doc_comment optional_parameter_list backup_fn_flags property_hook_body backup_fn_flags)) 376 ((property_hook_body) (";")) 377 ((property_hook_body) ("{" inner_statement_list "}")) 378 ((property_hook_body) (T_DOUBLE_ARROW expr ";")) 379 ((optional_parameter_list) (%empty)) 380 ((optional_parameter_list) ("(" parameter_list ")")) 381 ((class_const_list) (class_const_list "," class_const_decl)) 382 ((class_const_list) (class_const_decl)) 383 ((class_const_decl) (T_STRING "=" expr backup_doc_comment)) 384 ((class_const_decl) (semi_reserved "=" expr backup_doc_comment)) 385 ((const_decl) (T_STRING "=" expr backup_doc_comment)) 386 ((echo_expr_list) (echo_expr_list "," echo_expr)) 387 ((echo_expr_list) (echo_expr)) 388 ((echo_expr) (expr)) 389 ((for_cond_exprs) (%empty)) 390 ((for_cond_exprs) (non_empty_for_exprs "," expr)) 391 ((for_cond_exprs) (expr)) 392 ((for_exprs) (%empty)) 393 ((for_exprs) (non_empty_for_exprs)) 394 ((non_empty_for_exprs) (non_empty_for_exprs "," expr)) 395 ((non_empty_for_exprs) (non_empty_for_exprs "," T_VOID_CAST expr)) 396 ((non_empty_for_exprs) (T_VOID_CAST expr)) 397 ((non_empty_for_exprs) (expr)) 398 ((anonymous_class) (anonymous_class_modifiers_optional T_CLASS ctor_arguments extends_from implements_list backup_doc_comment "{" class_statement_list "}")) 399 ((new_dereferenceable) (T_NEW class_name_reference argument_list)) 400 ((new_dereferenceable) (T_NEW anonymous_class)) 401 ((new_dereferenceable) (T_NEW attributes anonymous_class)) 402 ((new_non_dereferenceable) (T_NEW class_name_reference)) 403 ((expr) (variable)) 404 ((expr) (T_LIST "(" array_pair_list ")" "=" expr)) 405 ((expr) ("[" array_pair_list "]" "=" expr)) 406 ((expr) (variable "=" expr)) 407 ((expr) (variable "=" ampersand variable)) 408 ((expr) (T_CLONE clone_argument_list)) 409 ((expr) (T_CLONE expr)) 410 ((expr) (variable T_PLUS_EQUAL expr)) 411 ((expr) (variable T_MINUS_EQUAL expr)) 412 ((expr) (variable T_MUL_EQUAL expr)) 413 ((expr) (variable T_POW_EQUAL expr)) 414 ((expr) (variable T_DIV_EQUAL expr)) 415 ((expr) (variable T_CONCAT_EQUAL expr)) 416 ((expr) (variable T_MOD_EQUAL expr)) 417 ((expr) (variable T_AND_EQUAL expr)) 418 ((expr) (variable T_OR_EQUAL expr)) 419 ((expr) (variable T_XOR_EQUAL expr)) 420 ((expr) (variable T_SL_EQUAL expr)) 421 ((expr) (variable T_SR_EQUAL expr)) 422 ((expr) (variable T_COALESCE_EQUAL expr)) 423 ((expr) (variable T_INC)) 424 ((expr) (T_INC variable)) 425 ((expr) (variable T_DEC)) 426 ((expr) (T_DEC variable)) 427 ((expr) (expr T_BOOLEAN_OR expr)) 428 ((expr) (expr T_BOOLEAN_AND expr)) 429 ((expr) (expr T_LOGICAL_OR expr)) 430 ((expr) (expr T_LOGICAL_AND expr)) 431 ((expr) (expr T_LOGICAL_XOR expr)) 432 ((expr) (expr "|" expr)) 433 ((expr) (expr T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG expr)) 434 ((expr) (expr T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG expr)) 435 ((expr) (expr "^" expr)) 436 ((expr) (expr "." expr)) 437 ((expr) (expr "+" expr)) 438 ((expr) (expr "-" expr)) 439 ((expr) (expr "*" expr)) 440 ((expr) (expr T_POW expr)) 441 ((expr) (expr "/" expr)) 442 ((expr) (expr "%" expr)) 443 ((expr) (expr T_SL expr)) 444 ((expr) (expr T_SR expr)) 445 ((expr) ("+" expr)) 446 ((expr) ("-" expr)) 447 ((expr) ("!" expr)) 448 ((expr) ("~" expr)) 449 ((expr) (expr T_IS_IDENTICAL expr)) 450 ((expr) (expr T_IS_NOT_IDENTICAL expr)) 451 ((expr) (expr T_IS_EQUAL expr)) 452 ((expr) (expr T_IS_NOT_EQUAL expr)) 453 ((expr) (expr T_PIPE expr)) 454 ((expr) (expr "<" expr)) 455 ((expr) (expr T_IS_SMALLER_OR_EQUAL expr)) 456 ((expr) (expr ">" expr)) 457 ((expr) (expr T_IS_GREATER_OR_EQUAL expr)) 458 ((expr) (expr T_SPACESHIP expr)) 459 ((expr) (expr T_INSTANCEOF class_name_reference)) 460 ((expr) ("(" expr ")")) 461 ((expr) (new_dereferenceable)) 462 ((expr) (new_non_dereferenceable)) 463 ((expr) (expr "?" expr ":" expr)) 464 ((expr) (expr "?" ":" expr)) 465 ((expr) (expr T_COALESCE expr)) 466 ((expr) (internal_functions_in_yacc)) 467 ((expr) (T_INT_CAST expr)) 468 ((expr) (T_DOUBLE_CAST expr)) 469 ((expr) (T_STRING_CAST expr)) 470 ((expr) (T_ARRAY_CAST expr)) 471 ((expr) (T_OBJECT_CAST expr)) 472 ((expr) (T_BOOL_CAST expr)) 473 ((expr) (T_UNSET_CAST expr)) 474 ((expr) (T_EXIT ctor_arguments)) 475 ((expr) ("@" expr)) 476 ((expr) (scalar)) 477 ((expr) ("`" backticks_expr "`")) 478 ((expr) (T_PRINT expr)) 479 ((expr) (T_YIELD)) 480 ((expr) (T_YIELD expr)) 481 ((expr) (T_YIELD expr T_DOUBLE_ARROW expr)) 482 ((expr) (T_YIELD_FROM expr)) 483 ((expr) (T_THROW expr)) 484 ((expr) (inline_function)) 485 ((expr) (attributes inline_function)) 486 ((expr) (T_STATIC inline_function)) 487 ((expr) (attributes T_STATIC inline_function)) 488 ((expr) (match)) 489 ((inline_function) (function returns_ref backup_doc_comment "(" parameter_list ")" lexical_vars return_type backup_fn_flags "{" inner_statement_list "}" backup_fn_flags)) 490 ((inline_function) (fn returns_ref backup_doc_comment "(" parameter_list ")" return_type T_DOUBLE_ARROW backup_fn_flags backup_lex_pos expr backup_fn_flags)) 491 ((fn) (T_FN)) 492 ((function) (T_FUNCTION)) 493 ((backup_doc_comment) (%empty)) 494 ((backup_fn_flags) (%empty)) 495 ((backup_lex_pos) (%empty)) 496 ((returns_ref) (%empty)) 497 ((returns_ref) (ampersand)) 498 ((lexical_vars) (%empty)) 499 ((lexical_vars) (T_USE "(" lexical_var_list possible_comma ")")) 500 ((lexical_var_list) (lexical_var_list "," lexical_var)) 501 ((lexical_var_list) (lexical_var)) 502 ((lexical_var) (T_VARIABLE)) 503 ((lexical_var) (ampersand T_VARIABLE)) 504 ((function_call) (name argument_list)) 505 ((function_call) (T_READONLY argument_list)) 506 ((function_call) (class_name T_PAAMAYIM_NEKUDOTAYIM member_name argument_list)) 507 ((function_call) (variable_class_name T_PAAMAYIM_NEKUDOTAYIM member_name argument_list)) 508 ((function_call) (callable_expr argument_list)) 509 ((class_name) (T_STATIC)) 510 ((class_name) (name)) 511 ((class_name_reference) (class_name)) 512 ((class_name_reference) (new_variable)) 513 ((class_name_reference) ("(" expr ")")) 514 ((backticks_expr) (%empty)) 515 ((backticks_expr) (T_ENCAPSED_AND_WHITESPACE)) 516 ((backticks_expr) (encaps_list)) 517 ((ctor_arguments) (%empty)) 518 ((ctor_arguments) (argument_list)) 519 ((dereferenceable_scalar) (T_ARRAY "(" array_pair_list ")")) 520 ((dereferenceable_scalar) ("[" array_pair_list "]")) 521 ((dereferenceable_scalar) (T_CONSTANT_ENCAPSED_STRING)) 522 ((dereferenceable_scalar) ("\"" encaps_list "\"")) 523 ((scalar) (T_LNUMBER)) 524 ((scalar) (T_DNUMBER)) 525 ((scalar) (T_START_HEREDOC T_ENCAPSED_AND_WHITESPACE T_END_HEREDOC)) 526 ((scalar) (T_START_HEREDOC T_END_HEREDOC)) 527 ((scalar) (T_START_HEREDOC encaps_list T_END_HEREDOC)) 528 ((scalar) (dereferenceable_scalar)) 529 ((scalar) (constant)) 530 ((scalar) (class_constant)) 531 ((constant) (name)) 532 ((constant) (T_LINE)) 533 ((constant) (T_FILE)) 534 ((constant) (T_DIR)) 535 ((constant) (T_TRAIT_C)) 536 ((constant) (T_METHOD_C)) 537 ((constant) (T_FUNC_C)) 538 ((constant) (T_PROPERTY_C)) 539 ((constant) (T_NS_C)) 540 ((constant) (T_CLASS_C)) 541 ((class_constant) (class_name T_PAAMAYIM_NEKUDOTAYIM identifier)) 542 ((class_constant) (variable_class_name T_PAAMAYIM_NEKUDOTAYIM identifier)) 543 ((class_constant) (class_name T_PAAMAYIM_NEKUDOTAYIM "{" expr "}")) 544 ((class_constant) (variable_class_name T_PAAMAYIM_NEKUDOTAYIM "{" expr "}")) 545 ((optional_expr) (%empty)) 546 ((optional_expr) (expr)) 547 ((variable_class_name) (fully_dereferenceable)) 548 ((fully_dereferenceable) (variable)) 549 ((fully_dereferenceable) ("(" expr ")")) 550 ((fully_dereferenceable) (dereferenceable_scalar)) 551 ((fully_dereferenceable) (class_constant)) 552 ((fully_dereferenceable) (new_dereferenceable)) 553 ((array_object_dereferenceable) (fully_dereferenceable)) 554 ((array_object_dereferenceable) (constant)) 555 ((callable_expr) (callable_variable)) 556 ((callable_expr) ("(" expr ")")) 557 ((callable_expr) (dereferenceable_scalar)) 558 ((callable_expr) (new_dereferenceable)) 559 ((callable_variable) (simple_variable)) 560 ((callable_variable) (array_object_dereferenceable "[" optional_expr "]")) 561 ((callable_variable) (array_object_dereferenceable T_OBJECT_OPERATOR property_name argument_list)) 562 ((callable_variable) (array_object_dereferenceable T_NULLSAFE_OBJECT_OPERATOR property_name argument_list)) 563 ((callable_variable) (function_call)) 564 ((variable) (callable_variable)) 565 ((variable) (static_member)) 566 ((variable) (array_object_dereferenceable T_OBJECT_OPERATOR property_name)) 567 ((variable) (array_object_dereferenceable T_NULLSAFE_OBJECT_OPERATOR property_name)) 568 ((simple_variable) (T_VARIABLE)) 569 ((simple_variable) ("$" "{" expr "}")) 570 ((simple_variable) ("$" simple_variable)) 571 ((static_member) (class_name T_PAAMAYIM_NEKUDOTAYIM simple_variable)) 572 ((static_member) (variable_class_name T_PAAMAYIM_NEKUDOTAYIM simple_variable)) 573 ((new_variable) (simple_variable)) 574 ((new_variable) (new_variable "[" optional_expr "]")) 575 ((new_variable) (new_variable T_OBJECT_OPERATOR property_name)) 576 ((new_variable) (new_variable T_NULLSAFE_OBJECT_OPERATOR property_name)) 577 ((new_variable) (class_name T_PAAMAYIM_NEKUDOTAYIM simple_variable)) 578 ((new_variable) (new_variable T_PAAMAYIM_NEKUDOTAYIM simple_variable)) 579 ((member_name) (identifier)) 580 ((member_name) ("{" expr "}")) 581 ((member_name) (simple_variable)) 582 ((property_name) (T_STRING)) 583 ((property_name) ("{" expr "}")) 584 ((property_name) (simple_variable)) 585 ((array_pair_list) (non_empty_array_pair_list)) 586 ((possible_array_pair) (%empty)) 587 ((possible_array_pair) (array_pair)) 588 ((non_empty_array_pair_list) (non_empty_array_pair_list "," possible_array_pair)) 589 ((non_empty_array_pair_list) (possible_array_pair)) 590 ((array_pair) (expr T_DOUBLE_ARROW expr)) 591 ((array_pair) (expr)) 592 ((array_pair) (expr T_DOUBLE_ARROW ampersand variable)) 593 ((array_pair) (ampersand variable)) 594 ((array_pair) (T_ELLIPSIS expr)) 595 ((array_pair) (expr T_DOUBLE_ARROW T_LIST "(" array_pair_list ")")) 596 ((array_pair) (T_LIST "(" array_pair_list ")")) 597 ((encaps_list) (encaps_list encaps_var)) 598 ((encaps_list) (encaps_list T_ENCAPSED_AND_WHITESPACE)) 599 ((encaps_list) (encaps_var)) 600 ((encaps_list) (T_ENCAPSED_AND_WHITESPACE encaps_var)) 601 ((encaps_var) (T_VARIABLE)) 602 ((encaps_var) (T_VARIABLE "[" encaps_var_offset "]")) 603 ((encaps_var) (T_VARIABLE T_OBJECT_OPERATOR T_STRING)) 604 ((encaps_var) (T_VARIABLE T_NULLSAFE_OBJECT_OPERATOR T_STRING)) 605 ((encaps_var) (T_DOLLAR_OPEN_CURLY_BRACES expr "}")) 606 ((encaps_var) (T_DOLLAR_OPEN_CURLY_BRACES T_STRING_VARNAME "}")) 607 ((encaps_var) (T_DOLLAR_OPEN_CURLY_BRACES T_STRING_VARNAME "[" expr "]" "}")) 608 ((encaps_var) (T_CURLY_OPEN variable "}")) 609 ((encaps_var_offset) (T_STRING)) 610 ((encaps_var_offset) (T_NUM_STRING)) 611 ((encaps_var_offset) ("-" T_NUM_STRING)) 612 ((encaps_var_offset) (T_VARIABLE)) 613 ((internal_functions_in_yacc) (T_ISSET "(" isset_variables possible_comma ")")) 614 ((internal_functions_in_yacc) (T_EMPTY "(" expr ")")) 615 ((internal_functions_in_yacc) (T_INCLUDE expr)) 616 ((internal_functions_in_yacc) (T_INCLUDE_ONCE expr)) 617 ((internal_functions_in_yacc) (T_EVAL "(" expr ")")) 618 ((internal_functions_in_yacc) (T_REQUIRE expr)) 619 ((internal_functions_in_yacc) (T_REQUIRE_ONCE expr)) 620 ((isset_variables) (isset_variable)) 621 ((isset_variables) (isset_variables "," isset_variable)) 622 ((isset_variable) (expr))))
  "The hash-table indexed by production-number and value is production.")

(defvar
  phps-mode-parser--table-look-aheads
  #s(hash-table test equal data (("!") t ("\"") t ($) t ("$") t ("%") t ("(") t (")") t ("*") t ("+") t (",") t ("-") t (".") t ("/") t (":") t (";") t ("<") t ("=") t (">") t ("?") t ("@") t (PREC_ARROW_FUNCTION) t (T_ABSTRACT) t (T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG) t (T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG) t (T_AND_EQUAL) t (T_ARRAY) t (T_ARRAY_CAST) t (T_AS) t (T_ATTRIBUTE) t (T_BOOLEAN_AND) t (T_BOOLEAN_OR) t (T_BOOL_CAST) t (T_BREAK) t (T_CALLABLE) t (T_CASE) t (T_CATCH) t (T_CLASS) t (T_CLASS_C) t (T_CLONE) t (T_COALESCE) t (T_COALESCE_EQUAL) t (T_CONCAT_EQUAL) t (T_CONST) t (T_CONSTANT_ENCAPSED_STRING) t (T_CONTINUE) t (T_CURLY_OPEN) t (T_DEC) t (T_DECLARE) t (T_DEFAULT) t (T_DIR) t (T_DIV_EQUAL) t (T_DNUMBER) t (T_DO) t (T_DOLLAR_OPEN_CURLY_BRACES) t (T_DOUBLE_ARROW) t (T_DOUBLE_CAST) t (T_ECHO) t (T_ELLIPSIS) t (T_ELSE) t (T_ELSEIF) t (T_EMPTY) t (T_ENCAPSED_AND_WHITESPACE) t (T_ENDDECLARE) t (T_ENDFOR) t (T_ENDFOREACH) t (T_ENDIF) t (T_ENDSWITCH) t (T_ENDWHILE) t (T_END_HEREDOC) t (T_ENUM) t (T_EVAL) t (T_EXIT) t (T_EXTENDS) t (T_FILE) t (T_FINAL) t (T_FINALLY) t (T_FN) t (T_FOR) t (T_FOREACH) t (T_FUNCTION) t (T_FUNC_C) t (T_GLOBAL) t (T_GOTO) t (T_HALT_COMPILER) t (T_IF) t (T_IMPLEMENTS) t (T_INC) t (T_INCLUDE) t (T_INCLUDE_ONCE) t (T_INLINE_HTML) t (T_INSTANCEOF) t (T_INSTEADOF) t (T_INTERFACE) t (T_INT_CAST) t (T_ISSET) t (T_IS_EQUAL) t (T_IS_GREATER_OR_EQUAL) t (T_IS_IDENTICAL) t (T_IS_NOT_EQUAL) t (T_IS_NOT_IDENTICAL) t (T_IS_SMALLER_OR_EQUAL) t (T_LINE) t (T_LIST) t (T_LNUMBER) t (T_LOGICAL_AND) t (T_LOGICAL_OR) t (T_LOGICAL_XOR) t (T_MATCH) t (T_METHOD_C) t (T_MINUS_EQUAL) t (T_MOD_EQUAL) t (T_MUL_EQUAL) t (T_NAMESPACE) t (T_NAME_FULLY_QUALIFIED) t (T_NAME_QUALIFIED) t (T_NAME_RELATIVE) t (T_NEW) t (T_NOELSE) t (T_NS_C) t (T_NS_SEPARATOR) t (T_NULLSAFE_OBJECT_OPERATOR) t (T_NUM_STRING) t (T_OBJECT_CAST) t (T_OBJECT_OPERATOR) t (T_OR_EQUAL) t (T_PAAMAYIM_NEKUDOTAYIM) t (T_PIPE) t (T_PLUS_EQUAL) t (T_POW) t (T_POW_EQUAL) t (T_PRINT) t (T_PRIVATE) t (T_PRIVATE_SET) t (T_PROPERTY_C) t (T_PROTECTED) t (T_PROTECTED_SET) t (T_PUBLIC) t (T_PUBLIC_SET) t (T_READONLY) t (T_REQUIRE) t (T_REQUIRE_ONCE) t (T_RETURN) t (T_SL) t (T_SL_EQUAL) t (T_SPACESHIP) t (T_SR) t (T_SR_EQUAL) t (T_START_HEREDOC) t (T_STATIC) t (T_STRING) t (T_STRING_CAST) t (T_STRING_VARNAME) t (T_SWITCH) t (T_THROW) t (T_TRAIT) t (T_TRAIT_C) t (T_TRY) t (T_UNSET) t (T_UNSET_CAST) t (T_USE) t (T_VAR) t (T_VARIABLE) t (T_VOID_CAST) t (T_WHILE) t (T_XOR_EQUAL) t (T_YIELD) t (T_YIELD_FROM) t ("[") t ("]") t ("^") t ("`") t ("{") t ("|") t ("}") t ("~") t))
  "The hash-table of valid look-aheads.")

(defvar
  phps-mode-parser--table-terminal-p
  #s(hash-table test equal data (T_NUM_STRING t T_CURLY_OPEN t T_STRING_VARNAME t T_DOLLAR_OPEN_CURLY_BRACES t "$" t T_NULLSAFE_OBJECT_OPERATOR t T_OBJECT_OPERATOR t T_END_HEREDOC t T_START_HEREDOC t T_DNUMBER t T_LNUMBER t "\"" t T_CONSTANT_ENCAPSED_STRING t T_ENCAPSED_AND_WHITESPACE t PREC_ARROW_FUNCTION t T_YIELD_FROM t "`" t "@" t T_UNSET_CAST t T_BOOL_CAST t T_OBJECT_CAST t T_ARRAY_CAST t T_STRING_CAST t T_DOUBLE_CAST t T_INT_CAST t T_COALESCE t T_SPACESHIP t T_IS_GREATER_OR_EQUAL t ">" t T_IS_SMALLER_OR_EQUAL t "<" t T_PIPE t T_IS_NOT_EQUAL t T_IS_EQUAL t T_IS_NOT_IDENTICAL t T_IS_IDENTICAL t "!" t "~" t T_SR t T_SL t "%" t "/" t T_POW t "*" t "-" t "+" t "." t "^" t T_BOOLEAN_AND t T_BOOLEAN_OR t T_DEC t T_INC t T_COALESCE_EQUAL t T_SR_EQUAL t T_SL_EQUAL t T_XOR_EQUAL t T_OR_EQUAL t T_AND_EQUAL t T_MOD_EQUAL t T_CONCAT_EQUAL t T_DIV_EQUAL t T_POW_EQUAL t T_MUL_EQUAL t T_MINUS_EQUAL t T_PLUS_EQUAL t T_PRIVATE_SET t T_PROTECTED_SET t T_PUBLIC_SET t T_PAAMAYIM_NEKUDOTAYIM t "?" t T_NOELSE t "[" t "=" t T_ELLIPSIS t T_VARIABLE t "|" t T_VOID_CAST t ":" t T_DOUBLE_ARROW t T_INLINE_HTML t T_NS_SEPARATOR t "}" t "{" t ")" t "(" t T_HALT_COMPILER t ";" t "]" t T_ATTRIBUTE t "," t T_NAME_RELATIVE t T_NAME_FULLY_QUALIFIED t T_NAME_QUALIFIED t T_STRING t T_AMPERSAND_NOT_FOLLOWED_BY_VAR_OR_VARARG t T_AMPERSAND_FOLLOWED_BY_VAR_OR_VARARG t T_READONLY t T_PUBLIC t T_PROTECTED t T_PRIVATE t T_FINAL t T_ABSTRACT t T_STATIC t T_PROPERTY_C t T_ENUM t T_MATCH t T_FN t T_NS_C t T_DIR t T_FILE t T_LINE t T_METHOD_C t T_FUNC_C t T_TRAIT_C t T_CLASS_C t T_CLASS t T_INTERFACE t T_TRAIT t T_NAMESPACE t T_IMPLEMENTS t T_EXTENDS t T_CALLABLE t T_ARRAY t T_BREAK t T_DEFAULT t T_CASE t T_ENDSWITCH t T_SWITCH t T_LIST t T_YIELD t T_PRINT t T_RETURN t T_CONST t T_FUNCTION t T_GOTO t T_CONTINUE t T_EMPTY t T_ISSET t T_UNSET t T_VAR t T_GLOBAL t T_INSTEADOF t T_USE t T_THROW t T_FINALLY t T_CATCH t T_TRY t T_AS t T_ENDDECLARE t T_DECLARE t T_ENDFOREACH t T_FOREACH t T_ENDFOR t T_FOR t T_ENDWHILE t T_WHILE t T_DO t T_ECHO t T_ENDIF t T_ELSE t T_ELSEIF t T_IF t T_EXIT t T_CLONE t T_NEW t T_INSTANCEOF t T_LOGICAL_AND t T_LOGICAL_XOR t T_LOGICAL_OR t T_REQUIRE_ONCE t T_REQUIRE t T_EVAL t T_INCLUDE_ONCE t T_INCLUDE t))
  "The hash-table of valid terminals.")

(defvar
  phps-mode-parser--table-non-terminal-p
  #s(hash-table test equal data (isset_variable t isset_variables t encaps_var_offset t encaps_var t possible_array_pair t array_pair t non_empty_array_pair_list t static_member t function_call t property_name t array_object_dereferenceable t callable_variable t fully_dereferenceable t class_constant t constant t dereferenceable_scalar t encaps_list t new_variable t callable_expr t variable_class_name t member_name t lexical_var t lexical_var_list t backup_lex_pos t fn t lexical_vars t match t inline_function t backticks_expr t scalar t internal_functions_in_yacc t new_non_dereferenceable t new_dereferenceable t clone_argument_list t anonymous_class t class_name_reference t ctor_arguments t anonymous_class_modifiers_optional t non_empty_for_exprs t echo_expr t class_const_decl t property_hook_body t optional_parameter_list t property_hook_modifiers t property_hook t property_hook_list t property t member_modifier t trait_method_reference t absolute_trait_method_reference t trait_alias t trait_precedence t trait_adaptation t trait_adaptation_list t trait_adaptations t attributed_class_statement t enum_case t method_body t method_modifiers t class_const_list t class_const_modifiers t hooked_property t property_list t property_modifiers t class_statement t static_var t simple_variable t global_var t argument_no_expr t non_empty_clone_argument_list t argument t non_empty_argument_list t union_type_without_static_element t name t intersection_type_without_static t union_type_without_static t union_type_element t type_without_static t intersection_type t union_type t type t type_expr_without_static t optional_property_hook_list t is_variadic t is_reference t optional_type_without_static t optional_cpp_modifiers t non_empty_member_modifiers t parameter t attributed_parameter t non_empty_parameter_list t alt_if_stmt_without_else t if_stmt_without_else t match_arm_cond_list t match_arm t non_empty_match_arm_list t match_arm_list t case_list t array_pair_list t ampersand t class_name_list t enum_case_expr t type_expr t enum_backing_type t interface_extends_list t anonymous_class_modifiers t class_modifier t class_statement_list t implements_list t extends_from t class_modifiers t backup_fn_flags t return_type t parameter_list t backup_doc_comment t function_name t returns_ref t function t variable t unset_variable t optional_variable t catch_name_list t finally_statement t catch_list t declare_statement t foreach_statement t foreach_variable t unset_variables t echo_expr_list t static_var_list t global_var_list t optional_expr t switch_case_list t for_statement t for_cond_exprs t for_exprs t while_statement t expr t alt_if_stmt t if_stmt t inner_statement t inner_statement_list t const_decl t use_declaration t unprefixed_use_declaration t inline_use_declaration t inline_use_declarations t unprefixed_use_declarations t legacy_namespace_name t use_declarations t group_use_declaration t use_type t mixed_group_use_declaration t namespace_declaration_name t attributed_top_statement t statement t const_list t attributed_statement t enum_declaration_statement t interface_declaration_statement t trait_declaration_statement t class_declaration_statement t function_declaration_statement t attributes t attribute t possible_comma t attribute_group t attribute_decl t argument_list t class_name t namespace_name t identifier t top_statement t semi_reserved t reserved_non_modifiers t top_statement_list t))
  "The hash-table of valid non-terminals.")

(defvar
  phps-mode-parser--table-translations
  #s(hash-table test equal)
  "The hash-table of translations.")

(defvar
  phps-mode-parser--e-identifier
  '%empty
  "The e-identifier.")

(defvar
  phps-mode-parser--eof-identifier
  '$
  "The end-of-file-identifier.")

(defvar
  phps-mode-parser--look-ahead-number
  1
  "The look-ahead number.")

(defvar
  phps-mode-parser-lex-analyzer--index-init
  1
  "Initial value of index.")

(defvar
  phps-mode-parser-lex-analyzer--state-init
  '(ST_INITIAL)
  "Initial value of state.")


;;; Buffer-Local Variables:


(defvar-local
  phps-mode-parser-lex-analyzer--buffered-response
  nil
  "Buffered tokens of lex-analyzer.")

(defvar-local
  phps-mode-parser-lex-analyzer--index
  0
  "The current index of the lex-analyzer.")

(defvar-local
  phps-mode-parser-lex-analyzer--state
  nil
  "The current state of the lex-analyzer.")


;;; Variable Functions:


(defvar
  phps-mode-parser-lex-analyzer--get-function
  #'(lambda (token) (let ((start (car (cdr token))) (end (cdr (cdr token)))) (if (<= end (point-max)) (progn (buffer-substring-no-properties start end)))))
  "The lex-analyzer get function.")

(defvar
  phps-mode-parser-lex-analyzer--function
  #'(lambda (index old-state) (let* ((lexer-response (phps-mode-lexer--re2c index old-state)) (tokens (nth 0 lexer-response)) (move-to-index (nth 1 lexer-response)) (new-state (nth 2 lexer-response))) (if move-to-index nil (let ((token-type (car (car tokens)))) (cond ((or (equal token-type 'T_OPEN_TAG) (equal token-type 'T_COMMENT) (equal token-type 'T_DOC_COMMENT)) (setq move-to-index (cdr (cdr (car tokens))))) ((equal token-type 'T_OPEN_TAG_WITH_ECHO) (let* ((v (car tokens))) (setcar v 'T_ECHO))) ((equal token-type 'T_CLOSE_TAG) (let* ((v (car tokens))) (setcar v ";")))))) (list tokens move-to-index new-state)))
  "The lex-analyzer function.")

(defvar
  phps-mode-parser-lex-analyzer--reset-function
  nil
  "The lex-analyzer reset function.")


;;; Functions:


;;; Functions for Lex-Analyzer:


(defun
  phps-mode-parser-lex-analyzer--get-function (token)
  "Get information about TOKEN."
  (let ((meta-information))
    (condition-case
      error
      (progn
        (setq
          meta-information
          (funcall
            phps-mode-parser-lex-analyzer--get-function
            token)))
      (error (error
        "Lex-analyze failed to get token meta-data of %s, error: %s"
        token
        (car (cdr error)))))
    meta-information))

(defun
  phps-mode-parser-lex-analyzer--reset
  ()
  "Reset Lex-Analyzer."
  (setq
    phps-mode-parser-lex-analyzer--buffered-response
    (make-hash-table :test 'equal))
  (setq
    phps-mode-parser-lex-analyzer--index
    phps-mode-parser-lex-analyzer--index-init)
  (setq
    phps-mode-parser-lex-analyzer--state
    phps-mode-parser-lex-analyzer--state-init)
  (when
    phps-mode-parser-lex-analyzer--reset-function
    (funcall
      phps-mode-parser-lex-analyzer--reset-function)))

(defun
  phps-mode-parser-lex-analyzer--peek-next-look-ahead
  ()
  "Peek next look-ahead number of tokens via lex-analyzer."
  (let ((look-ahead)
        (look-ahead-length
          0)
        (index
          phps-mode-parser-lex-analyzer--index)
        (state
          phps-mode-parser-lex-analyzer--state)
        (k (max
            1
            phps-mode-parser--look-ahead-number)))
    (while (<
            look-ahead-length
            k)
      (let* ((result-list
              (phps-mode-parser-lex-analyzer--get-buffered-lex
               index
               state))
             (token
              (nth 0 result-list))
             (new-index
              (nth 2 result-list))
             (new-state
              (nth 3 result-list)))
        (push
         token
         look-ahead)
        (setq
         look-ahead-length
         (1+ look-ahead-length))
        (setq
         index
         new-index)
        (setq
         state
         new-state)))

    (nreverse look-ahead)))

(defun
  phps-mode-parser-lex-analyzer--pop-token ()
  "Pop next token via lex-analyzer."
  (let* ((result-list
          (phps-mode-parser-lex-analyzer--get-buffered-lex
           phps-mode-parser-lex-analyzer--index
           phps-mode-parser-lex-analyzer--state))
         (token
          (nth 0 result-list))
         (new-index
          (nth 2 result-list))
         (new-state
          (nth 3 result-list)))
    (setq-local
     phps-mode-parser-lex-analyzer--index
     new-index)
    (setq-local
     phps-mode-parser-lex-analyzer--state
     new-state)
    (list token)))

(defun phps-mode-parser-lex-analyzer--get-buffered-lex (index state)
  "Get next token in stream, use buffer to only call function when needed."
  (unless (gethash
           index
           phps-mode-parser-lex-analyzer--buffered-response)
    (let ((continue t)
          (tmp-index index)
          (tmp-state state))
      (while continue
        (condition-case error
            (progn
              (let* ((result-list
                      (funcall
                       phps-mode-parser-lex-analyzer--function
                       tmp-index
                       tmp-state))
                     (tokens
                      (nth 0 result-list))
                     (move-to-index-flag
                      (nth 1 result-list))
                     (new-state
                      (nth 2 result-list)))
                (if move-to-index-flag
                    (progn
                      (setq
                       tmp-index
                       move-to-index-flag)
                      (setq
                       tmp-state
                       new-state))

                  (if tokens

                      (progn
                        (unless (listp (car tokens))
                          (setq tokens (list tokens)))

                        (let* ((first-token (car tokens))
                               (first-token-start (car (cdr first-token)))
                               (first-token-end (cdr (cdr first-token))))
                          (when (< index first-token-start)
                            (let ((token-start index))
                              (while (< token-start first-token-start)
                                (puthash
                                 token-start
                                 (list
                                  first-token
                                  nil
                                  first-token-end
                                  new-state)
                                 phps-mode-parser-lex-analyzer--buffered-response)
                                (setq
                                 token-start
                                 (1+ token-start))))))

                        (dolist (token tokens)
                          (let ((token-start (car (cdr token)))
                                (token-end (cdr (cdr token))))
                            (puthash
                             token-start
                             (list token nil token-end new-state)
                             phps-mode-parser-lex-analyzer--buffered-response))))

                    ;; Fill up look-ahead with EOF-identifier if we found nothing
                    (puthash
                     index
                     (list (list phps-mode-parser--eof-identifier) nil (1+ index) nil)
                     phps-mode-parser-lex-analyzer--buffered-response))

                  (setq
                   continue
                   nil))))
          (error
           (error
            "Lex-analyze failed to get next token at: %s in state: %s, error: %s"
            index
            state
            error))))))
  (gethash
   index
   phps-mode-parser-lex-analyzer--buffered-response))


;;; Functions for Syntax-Analyzer / Parser:


(defun
  phps-mode-parser--get-grammar-production-by-number
  (production-number)
  "If PRODUCTION-NUMBER exist, return it's production."
  (gethash
   production-number
   phps-mode-parser--table-productions-number-reverse))

(defun
  phps-mode-parser--valid-symbol-p
  (symbol)
  "Return whether SYMBOL is valid or not."
  (let ((is-valid t))
    (unless (or
             (phps-mode-parser--valid-e-p symbol)
             (phps-mode-parser--valid-eof-p symbol)
             (phps-mode-parser--valid-non-terminal-p symbol)
             (phps-mode-parser--valid-terminal-p symbol))
      (setq is-valid nil))
    is-valid))

(defun
  phps-mode-parser--valid-e-p
  (symbol)
  "Return whether SYMBOL is the e identifier or not."
  (eq
   symbol
   phps-mode-parser--e-identifier))

(defun
  phps-mode-parser--valid-eof-p
  (symbol)
  "Return whether SYMBOL is the EOF identifier or not."
  (eq
    symbol
    phps-mode-parser--eof-identifier))

(defun
  phps-mode-parser--valid-non-terminal-p (symbol)
  "Return whether SYMBOL is a non-terminal in grammar or not."
  (gethash
   symbol
   phps-mode-parser--table-non-terminal-p))

(defun
  phps-mode-parser--valid-terminal-p (symbol)
  "Return whether SYMBOL is a terminal in grammar or not."
  (gethash
   symbol
   phps-mode-parser--table-terminal-p))

(defun
  phps-mode-parser--get-grammar-translation-by-number
  (production-number)
  "If translation for PRODUCTION-NUMBER exist, return it."
  (gethash
    production-number
    phps-mode-parser--table-translations))

(defun
  phps-mode-parser--parse
  (&optional
    perform-sdt
    input-tape-index
    pushdown-list
    output
    translation
    translation-symbol-table-list
    history)
  "Perform a LR-parse via lex-analyzer, optionally PERFORM-SDT means to perform
  syntax-directed translation and optioanlly start at INPUT-TAPE-INDEX with
  PUSHDOWN-LIST, OUTPUT, TRANSLATION, TRANSLATION-SYMBOL-TABLE-LIST and HISTORY."
  (unless input-tape-index
    (setq input-tape-index 1))
  (unless pushdown-list
    (push 0 pushdown-list))
  (let ((translation-symbol-table
         (make-hash-table :test 'equal)))
    (when translation-symbol-table-list
      (dolist
          (item translation-symbol-table-list)
        (puthash
         (nth 0 item)
         (nth 1 item)
         translation-symbol-table)))

    (if (and
         input-tape-index
         (> input-tape-index 1))
        (setq
         phps-mode-parser-lex-analyzer--index
         input-tape-index)
      (phps-mode-parser-lex-analyzer--reset))

    ;; Make sure tables exists
    (unless phps-mode-parser--action-tables
      (error "Missing action-tables for grammar!"))
    (unless phps-mode-parser--distinct-action-tables
      (error "Missing distinct GOTO-tables for grammar!"))
    (unless phps-mode-parser--goto-tables
      (error "Missing GOTO-tables for grammar!"))
    (unless phps-mode-parser--distinct-goto-tables
      (error "Missing distinct GOTO-tables for grammar!"))

    (let ((accept))
      (while (not accept)

        ;; (1) The look-ahead string u, consisting of the next k input symbols, is determined.
        (let ((look-ahead
               (phps-mode-parser-lex-analyzer--peek-next-look-ahead))
              (look-ahead-full))

          ;; Save token stream indexes in separate variable if needed later
          (setq look-ahead-full look-ahead)

          ;; Create simplified look-ahead for logic below
          (setq look-ahead nil)
          (dolist (look-ahead-item look-ahead-full)
            (if (listp look-ahead-item)
                (push (car look-ahead-item) look-ahead)
              (push look-ahead-item look-ahead)))
          (setq look-ahead (nreverse look-ahead))

          (let ((table-index
                 (car pushdown-list)))
            (let ((action-table-distinct-index
                   (gethash
                    table-index
                    phps-mode-parser--action-tables)))
              (let ((action-table
                     (gethash
                      action-table-distinct-index
                      phps-mode-parser--distinct-action-tables)))
              (unless action-table
                (error
                 "Action-table with index %s is empty! Push-down-list: %s"
                 table-index
                 pushdown-list))
              (let ((action-match nil)
                    (action-table-length (length action-table))
                    (action-index 0)
                    (possible-look-aheads))

                ;; (2) The parsing action f of the table on top of the pushdown list is applied to the lookahead string u.
                (while (and
                        (not action-match)
                        (< action-index action-table-length))
                  (let ((action (nth action-index action-table)))
                    (let ((action-look-ahead (car action)))
                      (push
                       action-look-ahead
                       possible-look-aheads)
                      (when
                          (equal
                           action-look-ahead
                           look-ahead)
                        (setq
                         action-match
                         (cdr action)))
                      (when
                          (and
                           (=
                            phps-mode-parser--look-ahead-number
                            0)
                           (not
                            action-look-ahead))
                        ;; LR(0) reduce actions occupy entire row
                        ;; and is applied regardless of look-ahead
                        (setq
                         action-match
                         (cdr action))))
                    (setq
                     action-index
                     (1+ action-index))))

                (unless action-match
                  ;; (c) If f(u) = error, we halt parsing (and, in practice
                  ;; transfer to an error recovery routine).
                  (signal
                    'error
                    (list
                      (format
                        "Invalid syntax! Expected one of %s found %s at %s"

                   possible-look-aheads
                   look-ahead
                   phps-mode-parser-lex-analyzer--index)
                   possible-look-aheads
                   look-ahead
                   phps-mode-parser-lex-analyzer--index)))

                (cond

                 ((equal action-match '(shift))
                  ;; (a) If f(u) = shift, then the next input symbol, say a
                  ;; is removed from the input and shifted onto the pushdown list.
                  ;; The goto function g of the table on top of the pushdown list
                  ;; is applied to a to determine the new table to be placed on
                  ;; top of the pushdown list. We then return to step(1). If
                  ;; there is no next input symbol or g(a) is undefined, halt
                  ;; and declare error.

                  (let ((a (list (car look-ahead)))
                        (a-full (list (car look-ahead-full))))
                      (let ((goto-table-distinct-index
                             (gethash
                              table-index
                              phps-mode-parser--goto-tables)))
                        (let ((goto-table
                               (gethash
                                goto-table-distinct-index
                                phps-mode-parser--distinct-goto-tables)))
                      (let ((goto-table-length (length goto-table))
                            (goto-index 0)
                            (searching-match t)
                            (next-index)
                            (possible-look-aheads))

                        (while (and
                                searching-match
                                (< goto-index goto-table-length))
                          (let ((goto-item (nth goto-index goto-table)))
                            (let ((goto-item-symbol (list (car goto-item)))
                                  (goto-item-next-index (car (cdr goto-item))))
                              (push goto-item-symbol possible-look-aheads)

                              (when (equal
                                     goto-item-symbol
                                     a)
                                (setq next-index goto-item-next-index)
                                (setq searching-match nil))))

                          (setq goto-index (1+ goto-index)))
                        (unless next-index
                          (error
                           "In shift, found no GOTO-item for %s at %s, expected one of %s"
                           a
                           phps-mode-parser-lex-analyzer--index
                           possible-look-aheads))

                        (push (car a-full) pushdown-list)
                        (push next-index pushdown-list)
                        (phps-mode-parser-lex-analyzer--pop-token))))))

                 ((equal (car action-match) 'reduce)
                  ;; (b) If f(u) = reduce i and production i is A -> a,
                  ;; then 2|a| symbols are removed from the top of the pushdown
                  ;; list, and production number i is placed in the output
                  ;; buffer. A new table T' is then exposed as the top table
                  ;; of the pushdown list, and the goto function of T' is applied
                  ;; to A to determine the next table to be placed on top of the
                  ;; pushdown list. We place A and this new table on top of the
                  ;; the pushdown list and return to step (1)

                  (let ((production-number (car (cdr action-match))))

                    (let ((production
                           (phps-mode-parser--get-grammar-production-by-number
                            production-number)))
                      (let ((production-lhs (car production))
                            (production-rhs (car (cdr production)))
                            (popped-items-contents))
                        (unless (equal
                                 production-rhs
                                 (list phps-mode-parser--e-identifier))
                          (let ((pop-items (* 2 (length production-rhs)))
                                (popped-items 0)
                                (popped-item))
                            (while (< popped-items pop-items)
                              (setq popped-item (pop pushdown-list))
                              (when (and
                                     (listp popped-item)
                                     (phps-mode-parser--valid-symbol-p
                                      (car popped-item)))
                                (push
                                 popped-item
                                 popped-items-contents))
                              (setq popped-items (1+ popped-items)))))
                        (push production-number output)

                        (when perform-sdt
                          (let ((popped-items-meta-contents)
                                (popped-items-terminals))
                            (setq
                             popped-items-contents
                             (reverse popped-items-contents))
                            ;; Collect arguments for translation
                            (dolist (popped-item popped-items-contents)
                              (if (and
                                   (listp popped-item)
                                   (cdr popped-item))
                                  ;; If item is a terminal, use it's literal value
                                  (progn
                                    (push
                                      (phps-mode-parser-lex-analyzer--get-function
                                       popped-item)
                                       popped-items-meta-contents)
                                      (push
                                       popped-item
                                       popped-items-terminals))

                                ;; If item is a non-terminal
                                (let ((temp-hash-key
                                       (format
                                        "%S"
                                         popped-item)))

                                ;; If we have a translation for symbol, pop one
                                ;; otherwise push nil on translation argument stack
                                (if (gethash
                                         temp-hash-key
                                         translation-symbol-table)
                                        (let ((symbol-translations
                                               (gethash
                                                temp-hash-key
                                                translation-symbol-table)))
                                          (let ((symbol-translation
                                                 (pop symbol-translations)))
                                            (push
                                             (car symbol-translation)
                                             popped-items-meta-contents)
                                            (push
                                             (car (cdr symbol-translation))
                                             popped-items-terminals)
                                            (puthash
                                             temp-hash-key
                                             symbol-translations
                                             translation-symbol-table)))
                                      (push
                                       nil
                                       popped-items-meta-contents)
                                      (push
                                       nil
                                       popped-items-terminals)))))

                              ;; If we just have one argument, pass it as a instead of a list
                              (when (= (length popped-items-meta-contents) 1)
                                (setq
                                 popped-items-meta-contents
                                 (car popped-items-meta-contents))
                                (setq
                                 popped-items-terminals
                                 (car popped-items-terminals)))

                              ;; Perform translation at reduction if specified
                              (if
                                  (phps-mode-parser--get-grammar-translation-by-number
                                   production-number)
                                    (condition-case conditions
                                        (let ((partial-translation
                                               (funcall
                                                (phps-mode-parser--get-grammar-translation-by-number
                                                 production-number)
                                                popped-items-meta-contents
                                                popped-items-terminals)))
                                          (let ((temp-hash-key
                                                 (format
                                                  "%S"
                                                  production-lhs)))
                                            (let ((symbol-translations
                                                   (gethash
                                                    temp-hash-key
                                                    translation-symbol-table)))
                                              (push
                                                (list
                                                 partial-translation
                                                 popped-items-terminals)
                                                 symbol-translations)
                                              (puthash
                                               temp-hash-key
                                               symbol-translations
                                               translation-symbol-table)
                                              (setq
                                               translation
                                               partial-translation))))

                                      (error
                                       (signal
                                        'error
                                        (list
                                         (format
                                          "Failed AST translation for production %S with error: %S"
                                          production-number
                                          conditions)))))


                                ;; When no translation is specified just use popped contents as translation
                                (let ((partial-translation
                                       (list
                                        popped-items-meta-contents
                                        popped-items-terminals)))
                                  (let ((temp-hash-key
                                         (format
                                          "%S"
                                          production-lhs)))
                                    (let ((symbol-translations
                                           (gethash
                                            temp-hash-key
                                            translation-symbol-table)))
                                      (push
                                       partial-translation
                                       symbol-translations)
                                      (puthash
                                       temp-hash-key
                                       symbol-translations
                                       translation-symbol-table)
                                      (setq
                                       translation
                                       (car partial-translation))))))))

                          (let ((new-table-index (car pushdown-list)))
                            (let ((goto-table-distinct-index
                                   (gethash
                                    new-table-index
                                    phps-mode-parser--goto-tables)))
                              (let ((goto-table
                                     (gethash
                                      goto-table-distinct-index
                                      phps-mode-parser--distinct-goto-tables)))
                                (let ((goto-table-length
                                       (length goto-table))
                                      (goto-index 0)
                                      (searching-match t)
                                      (next-index))

                                  (while (and
                                          searching-match
                                          (< goto-index goto-table-length))
                                    (let ((goto-item (nth goto-index goto-table)))
                                      (let ((goto-item-symbol (list (car goto-item)))
                                            (goto-item-next-index (car (cdr goto-item))))

                                        (when (equal
                                               goto-item-symbol
                                               production-lhs)
                                          (setq next-index goto-item-next-index)
                                          (setq searching-match nil))))

                                    (setq goto-index (1+ goto-index)))

                                  (when next-index
                                    (push production-lhs pushdown-list)
                                    (push next-index pushdown-list))))))))))

                   ((equal action-match '(accept))
                    ;;    (d) If f(u) = accept, we halt and declare the string
                    ;;    in the output buffer to be the right parse of the original
                    ;;    input string.

                    (setq accept t))
                   (t (error
                       "Invalid action-match: %s!"
                       action-match)))))))))
      (unless accept
        (error
         "Parsed entire string without getting accepting! Output: %s"
         (reverse output)))
      (when history
        (setq history (reverse history)))
      (when output
        (setq output (reverse output)))
      (let ((translation-symbol-table-list))
        (when translation-symbol-table
          (maphash
           (lambda (key value)
             (push
              `(,key ,value)
              translation-symbol-table-list))
           translation-symbol-table))
        (list
         output
         translation
         translation-symbol-table-list
         history)))))

(defun phps-mode-parser-parse
    (&optional
     input-tape-index
     pushdown-list
     output
     translation
     history)
  "Perform a LR-parse via lex-analyzer, optionally at INPUT-TAPE-INDEX with
  PUSHDOWN-LIST, OUTPUT, TRANSLATION and HISTORY."
  (let ((result
         (phps-mode-parser--parse
          nil
          input-tape-index
          pushdown-list
          output
          translation
          history)))
    (nth 0 result)))

(defun phps-mode-parser-translate
    (&optional
     input-tape-index
     pushdown-list
     output
     translation
     history)
  "Perform a LR-parse via lex-analyzer, optionally at INPUT-TAPE-INDEX with
  PUSHDOWN-LIST, OUTPUT, TRANSLATION and HISTORY."
  (let ((result
         (phps-mode-parser--parse
          t
          input-tape-index
          pushdown-list
          output
          translation
          history)))
    (nth 1 result)))

(provide 'phps-mode-parser)

;;; phps-mode-parser.el ends here
